package F5::ASMConfig::Client;

use strict;
use warnings;

use POSIX;
use IO::Socket::INET6;
use IO::Select;
use Time::HiRes;

use Thrift::BufferedTransport;
use Thrift::FramedTransport;
use Thrift::Socket;
use Thrift::BinaryProtocol;
use Thrift::MemoryBuffer;
use Scalar::Util;

my $DEFAULT_RPC_SERVER_PORT = 9781;
my $DEFAULT_EVENT_QUEUE_PORT = 9783;

my $DEFAULT_ASYNC_TIME_LIMIT = 120;
my $EVENT_SOCKET_TIMEOUT_IN_MS = 30000; # 30 seconds

use F5::ASMConfig::ASMConfig;
use base qw/F5::ASMConfig::ASMConfigClient/;

our $AUTOLOAD;

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;

    my $self = $class->SUPER::new();
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{gui_user_name} = '';
    if ($params{devices}) {
        $self->{devices} = $params{devices};
    }
    elsif ($params{device}) {
        $self->{device} = $params{device};
        $self->{host} = $params{device}->{mng_ip_address};
        $self->{ssl_ca_cert_pem} = $params{device}->{ca_cert_pem};
        $self->{gui_user_name} = $params{device}->{admin_user};
    }
    else {
        $self->{host} = $params{host} || '127.0.0.1';
    }

    $self->{rpc_server_port} = $params{rpc_server_port} || $DEFAULT_RPC_SERVER_PORT;
    $self->{event_queue_port} = $params{event_queue_port} || $DEFAULT_EVENT_QUEUE_PORT;
    $self->{username} = $params{username} || 'SYSTEM';
    $self->{async_calls} = $params{async_calls};
    $self->{current_async_call_index} = -1;
    if (defined $params{gui_user_name}) {
        $self->{gui_user_name} = $params{gui_user_name};
    }
    $self->{deploy_request_fields} = $params{deploy_request_fields};
    $self->{is_async} = $params{is_async};
    $self->{is_ssl} = $params{is_ssl};

    # If set, async_concurrent_calls allows sending multiple RPC calls to the same device, even before
    # the response is received. This would help to speed up large file transfers, for example, and reduce memory usage.
    # This is currently disabled by default due to socket buffering issues.
    $self->{async_concurrent_calls} = defined($params{async_concurrent_calls}) ? $params{async_concurrent_calls} : 0;
    $self->{concurrent_call_backlog} = [];

    $self->{async_rv_stack} = [];
    $self->{ssl_ca_cert_pem} ||= $params{ssl_ca_cert_pem};
    $self->{async_time_limit} = defined $params{async_time_limit} ? $params{async_time_limit} : $DEFAULT_ASYNC_TIME_LIMIT;
    $self->{async_spawn_fork} = 1;
    if (defined $ENV{ASM_ASYNC_SPAWN_FORK}) {
        $self->{async_spawn_fork} = $ENV{ASM_ASYNC_SPAWN_FORK};
    }
    elsif (defined $params{async_spawn_fork}) {
        $self->{async_spawn_fork} = $params{async_spawn_fork};
    }
    $self->{logger} = $params{logger};
    $self->{deploy_targets} = $params{deploy_targets};

    if (! $params{devices}) {
        $self->open_connection;
    }

    return $self;
}

#----------------------------------------------------------------------

sub open_connection {
    my($self) = @_;

    my $socket;
    my $buffer_size = 1024;
    if ($self->{is_ssl}) {
        my($key_file) = glob("/config/em/ssl.key/em-*.key");
        my($cert_file) = glob("/config/em/ssl.crt/em-*.crt");

        $socket = F5::Thrift::SSLSocket->new(PeerAddr => $self->{host},
                                             PeerPort => 4353,
                                             SSL_hostname => 'asmconfig.cgc.f5.com',
                                             SSL_use_cert => 1,
                                             SSL_key_file => $key_file,
                                             SSL_cert_file => $cert_file,
                                             SSL_ca_file => $self->{ssl_ca_cert_pem},
                                             Timeout => 10, # Allow 10 seconds to establish a connection
                                            );
        $buffer_size = 10_000_000; # 10 MB
    }
    else {
        $socket = Thrift::Socket->new({
                                       host => $self->{host},
                                       port => $self->{rpc_server_port},
                                       recvTimeout => 3600000,
                                       sendTimeout => 3600000,
                                      });
    }

    my $transport = Thrift::BufferedTransport->new($socket, $buffer_size, $buffer_size);
    $self->set_transport( Thrift::BinaryProtocol->new($transport));
    $transport->open();

    if ($self->{is_ssl}) {
        $self->{raw_socket} = $socket->{socket};
    }
    else {
        $self->{raw_socket} = ($socket->{handle}->handles)[0];
    }

    if ($self->{is_async}) {
        $self->{raw_socket}->blocking(0) if $self->{raw_socket};

        my @prepend_calls = (
                             {
                              call_sub => sub {
                                  my($client) = @_;
                                  $client->login($self->{username});
                              },
                              return_sub => sub {
                                  my($client, $rv) = @_;
                                  $client->{session} = $rv;
                              },
                              call_name => 'Login',
                             });

        if ($self->{is_ssl}) {
            push(@prepend_calls, {
                                  call_sub => sub {
                                      my($client) = @_;
                                      $client->full_buffering($client->{session});
                                  },
                                  call_name => 'Disabling buffering',
                                 });
        }

        unshift(@{$self->{async_calls}}, @prepend_calls);

        # Put raw socket aside and use MemoryBuffer for actual Thrift de/serialization
        $self->set_transport( Thrift::BinaryProtocol->new( Thrift::MemoryBuffer->new( 10_000_000)),
                              Thrift::BinaryProtocol->new( Thrift::MemoryBuffer->new( 10_000_000)));
    }
    else {
        $self->{session} = $self->login($self->{username});
        if ($self->{is_ssl}) {
            $self->full_buffering($self->{session});
        }
        $self->{session}->{gui_user_name} = $self->{gui_user_name};
    }
}

#----------------------------------------------------------------------

sub init_event_queue {
    my($self) = @_;
    $self->{event_socket} = IO::Socket::INET6->new(PeerAddr => $self->{host},
                                                   PeerPort => $self->{event_queue_port});

    $self->{event_socket}->send($self->{session}->{session_id} . "\n");

    $self->{event_select} = IO::Select->new($self->{event_socket});
    my $transport = Thrift::FramedTransport->new(Thrift::Socket->new({
                                                                      socket => $self->{event_socket},
                                                                      recvTimeout => $EVENT_SOCKET_TIMEOUT_IN_MS,
                                                                      sendTimeout => $EVENT_SOCKET_TIMEOUT_IN_MS,
                                                                     }));
    $self->{event_protocol} = Thrift::BinaryProtocol->new($transport);
}

#----------------------------------------------------------------------

sub has_waiting_events {
    my($self) = @_;
    if ($self->{event_select}->can_read(0)) {
        return 1;
    }
    return;
}

#----------------------------------------------------------------------

sub get_next_event {
    my($self) = @_;

    my $event = F5::ASMConfig::Event->new;
    $event->read($self->{event_protocol});
    return $event;
}

#----------------------------------------------------------------------

sub run_async_calls {
    my($self, %params) = @_;

    my $dbh = $params{dbh};

    # Dynamically load async-specific modules
    eval "use F5::ASMConfig::Entity::DeployRequest;";
    die "Can't load F5::ASMConfig::Entity::DeployRequest" if ($@);
    eval "use F5::DbUtils;";
    die "Can't load F5::DbUtils" if ($@);
    eval "use F5::IO::Selector;";
    die "Can't load F5::IO::Selector" if ($@);

    my $deploy_request;
    if ($dbh) {
        # Create new DeployRequest
        $deploy_request = F5::ASMConfig::Entity::DeployRequest->new(dbh => $dbh,
                                                                    object => $self->{deploy_request_fields});

        for my $device (@{$self->{devices}}) {
            $device->{deploy_request_status} = F5::ASMConfig::DeployRequestStatus->new(
                device_id => $device->{device_id},
                deploy_status => F5::ASMConfig::DeployStatus::DEPLOY_STATUS_PENDING,
                deploy_start_time => time(),);
            push(@{$deploy_request->{deploy_statuses}}, $device->{deploy_request_status});
 
            my @deploy_targets = grep {$_->{device}->{device_id} == $device->{device_id}} @{$self->{deploy_targets}};

            for my $deploy_target (@deploy_targets) {
                my $virtual_server_name = $deploy_target->{virtual_server}->{virtual_server_name};
                my $http_class_name = $deploy_target->{http_class}->{http_class_name};
                my $folder = $deploy_target->{virtual_server}->{folder} || $deploy_target->{http_class}->{folder};

                my $deploy_request_target = F5::ASMConfig::DeployRequestTarget->new(
                    device_id => $device->{device_id},
                    target_folder => $folder,
                    target_virtual_server_name => $virtual_server_name,
                    target_http_class_name => $http_class_name,);
                push(@{$deploy_request->{deploy_request_targets}}, $deploy_request_target);
            }
        }

        $deploy_request->store_add;
        $dbh->commit;
        $dbh->begin_work;
    }

    my $pid;
    if ($self->{async_spawn_fork}) {
        $pid = fork;
    }

    if ($pid) {
        return $deploy_request;
    }
    else {
        my $selector;

        eval {
            if ($dbh) {
                # disconnect dbh after fork
                my $orig_dbh = $dbh;
                $dbh = F5::DbUtils::get_dbh();
                $orig_dbh->{InactiveDestroy} = 1;

                $deploy_request->{dbh} = $dbh;
            }

            $selector = F5::IO::Selector->new;
            my $start_time = Time::HiRes::time();

            my @clients;
            for my $device (@{$self->{devices}}) {
                if ($dbh) {
                    $self->update_deploy_request_status(device => $device,
                                                        deploy_request => $deploy_request,
                                                        fields => {
                                                                   deploy_status => F5::ASMConfig::DeployStatus::DEPLOY_STATUS_RUNNING,
                                                                   last_action_time => time(),
                                                                   last_action_message => "Connecting",
                                                                  });
                }

                my $device_client = F5::ASMConfig::Client->new(device => $device,
                                                               is_ssl => $self->{is_ssl},
                                                               is_async => 1,
                                                               logger => $self->{logger},
                                                               async_calls => [ @{$self->{async_calls}} ],
                                                              );
                push(@clients, $device_client);

                my $device_socket = F5::IO::ASMConfig->new(client => $device_client,
                                                           logger => $self->{logger},
                                                           read_message_handler => sub {
                                                               my($handle, $message) = @_;

                                                               $handle->{client}->send_next_async_call(handle => $handle,
                                                                                                       message => $message,
                                                                                                       deploy_request => $deploy_request,
                                                                                                       selector => $selector,
                                                                                                      );
                                                           },
                                                           other_error_handler => sub {
                                                               my($handle, $error) = @_;
                                                               $self->{logger}->debug("Unknown error occurred (error $error) -- will retry") if $self->{logger};
                                                           },
                                                           socket_full_initial_handler => sub {
                                                               my($handle, $error) = @_;
                                                               $self->{logger}->debug("Socket full error triggered -- will retry") if $self->{logger};
                                                           },
                                                           selector => $selector,
                                                          );

                $device_client->send_next_async_call(handle => $device_socket,
                                                     deploy_request => $deploy_request,
                                                    );
            }

            my $count = 0;
            while (1) {
                my $poll_time = 1;
                my @rewrite_clients = grep { $_->{io_handle} && $_->{io_handle}->needs_rewrite() } @clients;
                if (@rewrite_clients) {
                    $poll_time = 0;
                }

                my @readers = $selector->can_read($poll_time);
                for my $reader (@readers) {
                    next if (grep { $reader eq $_ } @rewrite_clients);

                    eval {
                        my $last_read_rv = 1;
                        while ($last_read_rv) {
                            $last_read_rv = $reader->handle_read;

                            if ($reader->{client}->{async_concurrent_calls} && @{$reader->{client}->{async_rv_stack}}) {
                                # Re-attempt read, since SSL socket may not trigger polling if the data
                                # has already been read into the buffer
                                Time::HiRes::usleep(50_000);
                            }
                            else {
                                last;
                            }
                        }
                    };
                    if ($@) {
                        $selector->remove($reader);
                    }
                }

                for my $rewrite_client (@rewrite_clients) {
                    next if ($rewrite_client->{async_encountered_exception});
                    $self->{logger}->debug("Attempting rewrite on $rewrite_client->{host}") if $self->{logger};
                    $rewrite_client->{io_handle}->rewrite;
                }

                if (! $selector->has_handles) {
                    $self->{logger}->debug("All async clients have completed") if $self->{logger};
                    last;
                }

                if ($self->{async_time_limit} &&
                    Time::HiRes::time() > $start_time + $self->{async_time_limit}) {

                    $self->{logger}->debug("Time limit reached for async calls") if $self->{logger};
                    last;
                }
            }
        };
        if ($@) {
            my $msg = $@ || '';
            if (ref($@) && $@->{error_msg}) {
                $msg = $@->{error_msg};
            }

            if (ref($@) && $@->{message}) {
                $msg = $@->{message};
            }

            $self->{logger}->error("Async client failed: $msg") if $self->{logger};

            for my $device (map { $_->{client}->{device} } $selector->handles()) {
                $self->update_deploy_request_status(device => $device,
                                                    deploy_request => $deploy_request,
                                                    fields => {
                                                               deploy_status => F5::ASMConfig::DeployStatus::DEPLOY_STATUS_ERROR,
                                                               last_action_time => time(),
                                                               deploy_end_time => time(),
                                                               last_action_message => "Error encountered: $msg",
                                                              });
            }
        }


        if ($self->{async_spawn_fork}) {
            $self->{logger}->debug("Shutting down async call process") if $self->{logger};
            exit;
        }
    }
}

#----------------------------------------------------------------------

sub send_next_async_call {
    my($self, %params) = @_;

    my $message = $params{message};
    my $handle = $params{handle};
    my $deploy_request = $params{deploy_request};
    my $selector = $params{selector};

    if ($message) {
        $self->{logger}->trace("Received message for $self->{host}") if $self->{logger};

        my $previous_call = $self->{async_calls}->[$self->{current_async_call_index}];

        if ($message->{ex}) {
            if ($previous_call->{exception_sub}) {
                eval {
                    $previous_call->{exception_sub}->($self, $message->{ex});
                };
                if ($@) {
                    $self->{async_encountered_exception} = $@;
                }
            }
            else {
                $self->{async_encountered_exception} = $message->{ex};
            }
        }
        else {

            if (@{$self->{async_rv_stack}}) {
                if (@{$self->{concurrent_call_backlog}}) {
                    $self->{io_handle}->write_message( shift( @{$self->{concurrent_call_backlog}}));
                }
                # There are still more return values expected for this async_call
                return;
            }

            if ($message->{success} || $message->{void_return_value}) {
                if ($previous_call->{return_value}) {
                    $handle->{client}->{__async_rv}->{$previous_call->{return_value}} = $message->{success};
                }
                elsif ($previous_call->{return_sub}) {
                    eval {
                        $previous_call->{return_sub}->($self, $message->{success});
                    };
                    if ($@) {
                        $self->{async_encountered_exception} = $@;
                    }
                }
            }
        }
    }

    if ($self->{async_encountered_exception}) {
        my $error_msg = $self->{async_encountered_exception};
        if (ref($self->{async_encountered_exception})) {
            if ($self->{async_encountered_exception}->{error_message}) {
                $error_msg = $self->{async_encountered_exception}->{error_message};
            }
        }
        $self->{logger}->error("Async client encountered exception: $error_msg") if $self->{logger};

        if ($deploy_request) {
            $self->update_deploy_request_status(device => $self->{device},
                                                deploy_request => $deploy_request,
                                                fields => {
                                                           deploy_status => F5::ASMConfig::DeployStatus::DEPLOY_STATUS_ERROR,
                                                           last_action_time => time(),
                                                           deploy_end_time => time(),
                                                           last_action_message => "Error encountered: $error_msg",
                                                          });
        }
        $selector->remove($handle);
        $self->{raw_socket}->close if $self->{raw_socket};
        die "Async call encountered exception: $error_msg\n";
    }

    $self->{current_async_call_index}++;

    while (1) {
        if ($self->{current_async_call_index} > $#{$self->{async_calls}}) {
            if ($deploy_request) {
                $self->update_deploy_request_status(device => $self->{device},
                                                    deploy_request => $deploy_request,
                                                    fields => {
                                                               deploy_status => F5::ASMConfig::DeployStatus::DEPLOY_STATUS_SUCCESS,
                                                               last_action_time => time(),
                                                               deploy_end_time => time(),
                                                               last_action_message => "Complete",
                                                              });
            }

            $selector->remove($handle);
            $self->{raw_socket}->close if $self->{raw_socket};
            last;
        }

        my $call = $self->{async_calls}->[$self->{current_async_call_index}];

        my $call_name = $call->{call_name} || "Call # $self->{current_async_call_index}";
        $self->{logger}->trace("Calling $call_name") if $self->{logger};
        if ($deploy_request) {
            $self->update_deploy_request_status(device => $self->{device},
                                                deploy_request => $deploy_request,
                                                fields => {
                                                           deploy_status => F5::ASMConfig::DeployStatus::DEPLOY_STATUS_RUNNING,
                                                           last_action_time => time(),
                                                           last_action_message => $call_name,
                                                          });
        }
        if ($call->{call_sub}) {
            $call->{call_sub}->($self, $self->{__async_rv});
        }

        # Wait for response if we expect at least 1 return value; otherwise, continue to next call sub
        if (@{$self->{async_rv_stack}}) {
            last;
        }

        $self->{current_async_call_index}++;
    }
}

#----------------------------------------------------------------------

sub update_deploy_request_status {
    my($self, %params) = @_;

    my $deploy_request = $params{deploy_request};
    my $device = $params{device};
    my $fields = $params{fields};

    my $new_status = F5::ASMConfig::DeployRequestStatus->new(%{$device->{deploy_request_status}});
    for my $k (keys %$fields) {
        $new_status->{$k} = $fields->{$k};
    }
    $new_status->{device_id} = $device->{device_id};

    $deploy_request->store_update(store_obj => { deploy_statuses => [ $new_status ] }, call_flags => {append_nested_struct_on_update => 1});
}

#======================================================================

sub AUTOLOAD {
    my($self, @params) = @_;
    $AUTOLOAD =~ /::([^:]+)$/;
    my $method_call = $1;

    if ($self->{is_async}) {
        $self->initiate_call(method_call => $method_call,
                             rpc_params => \@params,
                            );

        if ($self->{async_concurrent_calls} || ! @{$self->{async_rv_stack}}) {
            $self->{io_handle}->write_message($self->{output}->{trans}->getBuffer);
            $self->{output}->{trans}->resetBuffer;
        }
        else {
            push(@{$self->{concurrent_call_backlog}}, $self->{output}->{trans}->getBuffer);
            $self->{output}->{trans}->resetBuffer;
        }

        push(@{$self->{async_rv_stack}}, $method_call);

        # If there are many client calls in a single call_sub, intervening (ignored) return
        # values need to be read and cleared from the buffer
        if ($self->{async_concurrent_calls} && scalar @{$self->{async_rv_stack}} > 1) {
            my $last_read_rv = 1;
            while ($last_read_rv && scalar @{$self->{async_rv_stack}} > 1) {
                $last_read_rv = $self->{io_handle}->handle_read(min_read_attempts => 5);
            }
        }
    }
    else {
        $method_call = "SUPER::$method_call";
        my $rv = $self->$method_call(@params);
    }
}

#======================================================================

package F5::IO::ASMConfig;

use base qw(F5::IO::Handle);

sub init {
    my($self, %params) = @_;
    $self->{client} = $params{client};
    $self->{socket} = $self->{client}->{raw_socket};

    # Link back from client to this IO::Handle
    $self->{client}->{io_handle} = $self;
    Scalar::Util::weaken($self->{client}->{io_handle});

    $self->SUPER::init(%params);
}

#----------------------------------------------------------------------

sub read_message {
    my($self, %params) = @_;

    my $read_attempt_count = 0;
    my $last_buffer_length = 0;
    my $min_read_attempts = $params{min_read_attempts};

    while (++$read_attempt_count < 500) {
        my $buf = $self->read(10_000_000);

        if (! $buf) {
            if ($min_read_attempts && $read_attempt_count < $min_read_attempts) {
                Time::HiRes::usleep(50_000);
                next;
            }
            else {
                return;
            }
        }

        if ($last_buffer_length && length($buf) == $last_buffer_length) {
            $self->unshift_to_read_buffer($buf);
            return;
        }

        $last_buffer_length = length($buf);
        $self->{client}->{input}->{trans}->resetBuffer($buf);
        my $rv;
        eval {
            $rv = $self->{client}->get_call_response(method_call => $self->{client}->{async_rv_stack}->[0],
                                                     raw_return_value => 1,
                                                    );
        };
        if ($@) {
            $self->unshift_to_read_buffer($buf);
        }
        else {
            shift(@{$self->{client}->{async_rv_stack}});

            # Unshift extra unneeded data from MemoryBuffer back to handle
            if ($self->{client}->{input}->{trans}->peek) {
                $self->unshift_to_read_buffer( $self->{client}->{input}->{trans}->read(10_000_000));
            }
            return $rv;
        }
    }

    return;
}

#----------------------------------------------------------------------

sub write_message {
    my($self, $message) = @_;

    $self->write($message);
}

#----------------------------------------------------------------------

1;
