package F5::ASMConfig::CorsBase;

use strict;
use warnings;

use F5::ASMConfig::Entity::AllowedMethod;

use base qw/F5::ASMConfig::Entity::Base/;


#----------------------------------------------------------------------

sub handle_cors {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags = $params{call_flags};
    my $field_values = $params{field_values};
    my $field_display_name = $params{field_display_name};

    if( $params{validate_value} && ($field_values->{enable_cors_enforcement} || $field_values->{flg_enable_cors_enforcement}) && $field_values->{flg_allow_cors} ) {
        if( exists $field_values->{allowed_origins_enforcement_mode} && $field_values->{allowed_origins_enforcement_mode} == F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_REPLACE_WITH()
            && ( !defined $field_values->{allowed_origins} || !@{ $field_values->{allowed_origins} } ) ) {
            if ($call_flags->{cascade_disable_flags}) {
                $store_obj->{allowed_origins_enforcement_mode} = $field_values->{allowed_origins_enforcement_mode} = F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_UNMODIFIED();
            }
            else {
                my $cors_allowed_origins_error_message = "At least one $field_display_name->{allowed_origins} should be defined.";
                die { error_message => $cors_allowed_origins_error_message,
                    display_message => $cors_allowed_origins_error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                };
            }
        }
        if( $field_values->{flg_allowed_methods_check} && ( !defined $field_values->{allowed_methods} || !@{ $field_values->{allowed_methods} } ) ) {
            if ($call_flags->{cascade_disable_flags}) {
                $store_obj->{flg_allowed_methods_check} = $field_values->{flg_methods_origins_check} = 0
            }
            else {
                my $cors_allowed_methods_error_message = "At least one $field_display_name->{allowed_methods} should be defined.";
                die { error_message => $cors_allowed_methods_error_message,
                    display_message => $cors_allowed_methods_error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                };
            }
        }
        for my $allowed_method (@{$field_values->{allowed_methods}}) {
            if ( !F5::ASMConfig::Entity::AllowedMethod->get_method(name => $allowed_method->{method_name}) ) {
                my $cors_allowed_methods_error_message = "Invalid method name: $allowed_method->{method_name}";
                die { error_message => $cors_allowed_methods_error_message,
                    display_message => $cors_allowed_methods_error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                };
            }
        }

        if( $field_values->{check_maximum_age} && !$field_values->{maximum_age} ) {
            if ($call_flags->{cascade_disable_flags}) {
                $store_obj->{check_maximum_age} = $field_values->{check_maximum_age} = 0
            }
            else {
                my $cors_maximum_age_error_message = "You must define $field_display_name->{maximum_age}.";
                die { error_message => $cors_maximum_age_error_message,
                    display_message => $cors_maximum_age_error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                };
            }
        }

        if( exists $field_values->{allowed_origins_enforcement_mode} && $field_values->{allowed_origins_enforcement_mode} == F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_UNMODIFIED() &&
            !$field_values->{flg_allowed_methods_check} &&
            !$field_values->{flg_allowed_headers_check} &&
            !$field_values->{flg_exposed_headers_check} &&
            !$field_values->{flg_check_credentials} &&
            !$field_values->{check_maximum_age} ) {

            if ($call_flags->{cascade_disable_flags}) {
                $store_obj->{enable_cors_enforcement} = $field_values->{enable_cors_enforcement} = 0;
                $store_obj->{flg_enable_cors_enforcement} = $field_values->{flg_enable_cors_enforcement} = 0;
            }
            else {
                my $cors_settings_error_message = "At least one HTML5 Cross-Domain Request Enforcement setting should be enabled.";
                die { error_message => $cors_settings_error_message,
                    display_message => $cors_settings_error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                };
            }
        }
    }

}

#----------------------------------------------------------------------

sub set_cors_values {
    my($self, %params) = @_;
    my $obj = $params{store_obj} || $self;
    return unless ref($obj);

    my $cors_enforcement_mode = $self->get_new_value(%params, field_name => 'cors_enforcement_mode');

    if ($cors_enforcement_mode) {
        if ($cors_enforcement_mode == F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_DISABLED()) {
            $obj->{enable_cors_enforcement} = 0;
            $obj->{flg_enable_cors_enforcement} = 0;
            $obj->{flg_allow_cors} = 0;
        }
        else {
            $obj->{enable_cors_enforcement} = 1;
            $obj->{flg_enable_cors_enforcement} = 1;
            if ($cors_enforcement_mode == F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_REMOVE_ALL_HEADERS()) {
                $obj->{flg_allow_cors} = 0;
            }
            else {
                $obj->{flg_allow_cors} = 1;
                if ( $self->element_type() == F5::ASMConfig::ElementType::_URL ) {

                    if ($cors_enforcement_mode == F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_REPLACE_HEADERS()) {
                        my $allowed_origins_enforcement_mode = $self->get_new_value(%params, field_name => 'allowed_origins_enforcement_mode');
                        if ($allowed_origins_enforcement_mode != F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_REPLACE_WITH() &&
                            $allowed_origins_enforcement_mode != F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_UNMODIFIED()) {
                            $obj->{allowed_origins_enforcement_mode} = F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_UNMODIFIED();
                        }
                    }
                    elsif ($cors_enforcement_mode == F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_ENFORCE()) {
                        $obj->{allowed_origins_enforcement_mode} = F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_ENFORCE();
                    }
                }
            }
        }
    }
}

#----------------------------------------------------------------------

sub map_virtual_cors_enforcement_mode {
    my ($self, %params) = @_;

    my $field_name = $params{field_name};
    my $handle_virtual_value = $params{handle_virtual_value};

    if ($handle_virtual_value) {
        $self->set_cors_values(%params);
    }
    else {
        my $proc_mode;
        if (! ($self->{enable_cors_enforcement} || $self->{flg_enable_cors_enforcement})) {
            $proc_mode = F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_DISABLED();
        }
        elsif (! $self->{flg_allow_cors}) {
            $proc_mode = F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_REMOVE_ALL_HEADERS();
        }
        elsif (exists $self->{allowed_origins_enforcement_mode} && $self->{allowed_origins_enforcement_mode} == F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_REPLACE_WITH()) {
            $proc_mode = F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_REPLACE_HEADERS();
        }
        elsif ( !exists $self->{allowed_origins_enforcement_mode} || $self->{allowed_origins_enforcement_mode} == F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_ENFORCE()) {
            $proc_mode = F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_ENFORCE();
        }
        else {
            $proc_mode = F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_REPLACE_HEADERS();
        }

        $self->{$field_name} = $proc_mode;
    }

}

#----------------------------------------------------------------------

sub update_default_object_by_pb_settings {
    my ($self, %params) = @_;

    $self->SUPER::update_default_object_by_pb_settings(%params);

    my $default_obj = $params{default_obj};
    my $policy = $params{policy};
    my $object = $params{object};

    if ( ! ($object->{enable_cors_enforcement} || $object->{flg_allow_cors}) ) {
        $default_obj->{cors_enforcement_mode} = F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_DISABLED();
        if ( $policy->{is_learning_enabled} ) {
           my $viol_index = F5::Violation::VIOL_CROSS_ORIGIN_REQUEST();
           my $violation_blocking_settings = $self->{dbh}->selectall_hashref(<<EndQuery, 'viol_index', undef, $policy->{policy_id}, $viol_index);
SELECT viol_index, flg_learn
  FROM PLC.PL_BLOCKING
 WHERE policy_id = ?
   AND viol_index = ?
EndQuery

            my $viol_learn_flg = $violation_blocking_settings->{$viol_index}->{flg_learn};
            if ( $viol_learn_flg ) {
                $default_obj->{cors_enforcement_mode} = F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_ENFORCE();
            }
        }
    }
}

#----------------------------------------------------------------------

return 1;
