package F5::ASMConfig::Entity::ConfigLog;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use F5::Utils::Rest;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::ConfigLog/;

use F5::CRC;

#----------------------------------------------------------------------

sub add_change_log_entry {
  # Do not save config log entry to PL_POLICY_CHANGES_LOG
}

#----------------------------------------------------------------------

sub get_complex_filter {
    my($class, %params) = @_;

    my $filter_field = $params{filter_field};
    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $plc_dbname = 'PLC';

    if ($filter_field eq 'search_from_first_stabilized_time') {
        my ($config_log_id) = $dbh->selectrow_array(<<EndQuery, undef, $policy->{policy_id}, F5::ASMConfig::ElementType::POLICY_ATTRIBUTES(), '%is stable%');
SELECT config_log_id
  FROM $plc_dbname.PL_CONFIG_LOG
 WHERE policy_id = ?
   AND element_type = ?
   AND description LIKE ?
 ORDER BY config_log_id ASC
 LIMIT 1
EndQuery

        return { where_clause => 'config_log_id >= ?',
                 where_values => [$config_log_id] };
    }
    return;
}

#----------------------------------------------------------------------

sub before_store {
  my($self, %params) = @_;
  my $store_obj = $params{store_obj};
  my $existing_obj = $params{existing_obj};
  my $call_flags = $params{call_flags} || {};

  $self->SUPER::before_store(%params);

  my $element = $self->get_new_value(%params, field_name => 'element', );
  my $policy = $params{policy} || $self->{policy};
  my $dbh = $params{dbh} || $self->{dbh};

  $store_obj->{rest_event_type} ||= $store_obj->{event_type};
  $store_obj->{rest_description} ||= $store_obj->{description};
  $store_obj->{rest_element_name} ||= $store_obj->{element_name};

  if (!$store_obj->{user_name} || $store_obj->{user_name} eq 'tsconfd') {
      $store_obj->{user_name} = 'SYSTEM';
  }
  elsif ($store_obj->{call_flags}->{partition_was_defined} && $store_obj->{user_name} eq 'RestClient') {
      $store_obj->{user_name} = 'GUI';
  }


  if ($store_obj->{element}) {

      my $element_type_info = $self->get_cached_element_type(element_type => $element->{element_type});
      if (my $element_class = $element_type_info->{class}) {
          my $element_type_name = $element_type_info->{element_type_name};
          my $entity = $element_class->new(%params, dbh => $dbh, policy => $policy, object => $element->{$element_type_name}, instantiate_entities => 1);
          my $changes_entity = $entity;
          if ($store_obj->{update_thrift_obj}) {
              $changes_entity = $element_class->new(%params, dbh => $dbh, policy => $policy, object => $store_obj->{update_thrift_obj}, instantiate_entities => 1);
          }
    
          my $description_entity = $entity;
    
          $store_obj->{rest_element_name} ||= $entity->get_rest_display_name(%params);
          my %call_flag_overrides = %{$call_flags};
    
          if (my $parent_diff_struct = $entity->parent_diff_struct(%params)) {
              $entity = $entity->get_diff_parent_entity(%params, parent_diff_struct => $parent_diff_struct, remove_non_unique_record_parts => 1);
              if ( !$parent_diff_struct->{display_values_from_child}) {
                  $description_entity = $entity;
                  $changes_entity = $changes_entity->get_diff_parent_entity(%params, parent_diff_struct => $parent_diff_struct, remove_non_unique_record_parts => 1);
              }
              my $subfield = $parent_diff_struct->{subfield};
              if (ref $entity->{$subfield} eq 'ARRAY') {  #Nested diff structs should appear as append/remove actions
                  if ($store_obj->{event_type} == F5::ASMConfig::EventType::DELETE()) {
                      $call_flag_overrides{remove_nested_struct_on_update} = 1;
                  }
                  elsif ($store_obj->{event_type} == F5::ASMConfig::EventType::ADD()) {
                      $call_flag_overrides{append_nested_struct_on_update} = 1;
                  }
              }
              $store_obj->{rest_event_type} = F5::ASMConfig::EventType::UPDATE();
              $call_flag_overrides{ignore_pb_unique_record_parts} = 1;
          }
    
          if (my $kind = $entity->get_kind(%params)) {
              $store_obj->{element_kind} = $kind;
              $store_obj->{rest_element_name} = $entity->get_rest_display_name(%params);
              $store_obj->{rest_element_type} = $entity->element_type(%params);
              $store_obj->{element_rest_uuid} = $entity->get_rest_uuid(%params);
              $store_obj->{rest_description} = ref($description_entity)->get_entity_changes_string(%params,
                                                                                                   call_flags => \%call_flag_overrides,
                                                                                                   entity => $changes_entity, 
                                                                                                   full_thrift_obj => $description_entity,
                                                                                                   get_rest_display_name => 1,
                                                                                                  ) || $store_obj->{description};
          }
      }
  }
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
    my($class_or_entity, %params) = @_;
    my $dbh = $params{dbh};
    my $rec = $params{rec} || $class_or_entity;
    my $policy = $params{policy};

    $class_or_entity->SUPER::map_fields_to_entity(%params);

    if ($rec->{element_kind}) {
        my $element_type_info = $class_or_entity->get_cached_element_type(element_type => $rec->{rest_element_type});
        my $element_class = $element_type_info->{class};
        $rec->{entity} = $element_class->new(dbh => $dbh, policy => $policy, object => {rest_uuid => $rec->{element_rest_uuid}});
        my $type_display_name = $rec->{entity}->get_element_type_display_name;
        $rec->{entity_type_display_name} = $type_display_name if $type_display_name;
    }
}

#----------------------------------------------------------------------

sub rest_search_filters {
    my ($class) = @_;
    return [ F5::ASMConfig::SearchFilter->new(field_name => 'rest_event_type',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GT(),
                                              field_value => 0),
           ];
}

#----------------------------------------------------------------------

my $change_number = 0;

sub get_rest_uuid {
    my ($self, %params) = @_;
    my $object = $params{object} || $self;
    my $rest_uuid = $object->{rest_uuid};
    if (!$rest_uuid) {
        #Create unique hash from PID, Change number for this process, and the event time
        my @values = ($$, $change_number++, $object->{event_time}); 
        $rest_uuid = F5::Utils::Rest::generate_uuid(join($F5::ASMConfig::Entity::Base::UNLIKELY_DELIMITER, @values));
    }
    return $rest_uuid;
}

#----------------------------------------------------------------------

1;

