package F5::ASMConfig::Entity::Extraction;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::Extraction/;

use F5::ASMConfig::Entity::Filetype;
use F5::ASMConfig::Entity::URL;
use F5::CRC;
use F5::DbUtils;

#----------------------------------------------------------------------

sub get_full_display_name {
    my ($self, %params) = @_;
    my @display_name_parts = ($self->get_field_display_string(field_name => 'extract_level', dbh => $params{dbh}),
                              $self->get_element_type_display_name,
                              "'" . $self->get_field_display_string(field_name => 'extract_name', dbh => $params{dbh}) . "'"
                             );
    if ($self->{extract_level} && $self->{extract_level} == F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_URL) {
        push (@display_name_parts, "on " . ($self->{url} ? $self->{url}->get_full_display_name(%params) : "UNNAMED"));
    }
    return join (" ", @display_name_parts);
}

#----------------------------------------------------------------------

sub get_retrieve_join_tables {
    my($class, %params) = @_;
    my %join_tables;

    my $plc_dbname = 'PLC';

    $join_tables{global_extracts} = <<EndQuery;
LEFT JOIN $plc_dbname.PL_GLOBAL_EXTRACTS global_extracts
       ON (main_table.id = global_extracts.extract_attrib_id)
EndQuery

    $join_tables{url_extracts} = <<EndQuery;
LEFT JOIN $plc_dbname.PL_OBJECT_EXTRACTS url_extracts
       ON (main_table.id = url_extracts.extract_attrib_id)
LEFT JOIN $plc_dbname.PL_OBJECTS url
       ON (url_extracts.obj_id = url.id)
LEFT JOIN $plc_dbname.PL_DYNAMIC_PARAM_EXTRACTS linked_parameter
       ON (main_table.id = linked_parameter.extract_attrib_id)
EndQuery

    return join('', map { $join_tables{$_} } $class->extern_db_prefixes(%params));
}

#----------------------------------------------------------------------

sub extern_db_prefixes {
    my($class, %params) = @_;

    if ($params{get_criteria} && $params{get_criteria}->{extract_level} && ! $params{get_criteria}->{extract_id}) {
        if ($params{get_criteria}->{extract_level} == F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_GLOBAL) {
            return qw(global_extracts);
        }
        if ($params{get_criteria}->{extract_level} == F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_URL) {
            return qw(url_extracts);
        }
    }
    else {
        return qw(global_extracts url_extracts);
    }
}

#----------------------------------------------------------------------

sub extern_unique_record_part_db_fields {
    my($class, %params) = @_;

    if ( $params{get_criteria} && $params{get_criteria}->{db_prefix} &&
         $params{get_criteria}->{db_prefix} eq 'url_extracts' ) {
        return ['obj_id', 'name'];
    }
    return [];
}

#----------------------------------------------------------------------

sub nested_struct_fields {
    my ($class_or_self, %params) = @_;
    my $level_obj = $params{store_obj} || $params{object} || $params{get_criteria} || $class_or_self;
    if ($level_obj && ref $level_obj && $level_obj->{extract_level} && $level_obj->{extract_level} == F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_URL()) {
        return [];
    }
    return ["applies_to_urls", "applies_to_filetypes"];
}


#----------------------------------------------------------------------

sub additional_field_mapping {
    my($class, %params) = @_;

    my $extract_level;
    if ($params{get_criteria} && ! $params{get_criteria}->{extract_id}) {
        $extract_level = $params{get_criteria}->{extract_level};
    }
    my %additional_field_mapping;
    my @url_groups;

    if (!$extract_level || $extract_level == F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_GLOBAL) {
        $additional_field_mapping{global_extracts__extract_attrib_id} = { extern_db_field_name => 'extract_attrib_id' };
        $additional_field_mapping{global_extracts__global_extract_id} = { extern_db_field_name => 'id' };
        $additional_field_mapping{global_extracts__applies_to_all} = { extern_db_field_name => 'flg_applies_to_all' };
        $additional_field_mapping{global_extracts__applies_to_regexp} = { extern_db_field_name => 'applies_to_regexp' };
    }

    if (!$extract_level || $extract_level == F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_URL) {
        $additional_field_mapping{url_extracts__extract_attrib_id} = { extern_db_field_name => 'extract_attrib_id' };
        $additional_field_mapping{linked_parameter__extract_attrib_id} = { extern_db_field_name => 'extract_attrib_id' };
        $additional_field_mapping{linked_parameter__param_id} = { extern_db_field_name => 'param_attrib_id' };
        push(@url_groups, 'url');
    }

    for my $url_group (@url_groups) {
        my %new_url_mapping =  (
                                "${url_group}__url_id" => { extern_db_field_name => 'id' },
                                "${url_group}__url_name" => { extern_db_field_name => 'uri' },
                                "${url_group}__app_id" => { extern_db_field_name => 'app_id' },
                                "${url_group}__method_name" => { extern_db_field_name => 'method_name' },
                                "${url_group}__is_wildcard" => { extern_db_field_name => 'flg_is_pattern' },
                               );
        %additional_field_mapping = (%additional_field_mapping,
                                     %new_url_mapping);
    }

    for my $k (keys %additional_field_mapping) {
        my @field_parts = split(/__/, $k);
        pop(@field_parts);
        $additional_field_mapping{$k}->{full_db_field_name} = join('__', @field_parts) . '.' . $additional_field_mapping{$k}->{extern_db_field_name};
    }

    my $extract_level_statement;
    if (! $extract_level) {
        $extract_level_statement = <<EndQuery;
CASE WHEN ( global_extracts.extract_attrib_id) THEN @{[F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_GLOBAL]} ELSE
        @{[F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_URL]} END
EndQuery
    }
    else {
        $extract_level_statement = $extract_level;
    }

    $additional_field_mapping{extract_level} = { full_db_field_name => $extract_level_statement,
                                               filter_using_having => 1,
                                             };

    return \%additional_field_mapping;
}

#----------------------------------------------------------------------

sub populate_from_join_row {
  my($class, %params) = @_;
  my $field_name = $params{field_name};
  my $rec = $params{rec};

  if ($field_name eq 'url' && $rec->{url_extracts__extract_attrib_id}) {
    $rec->{extract_name} = $rec->{url_extracts__extract_name};
    $rec->{url} = F5::ASMConfig::Entity::URL->new( dbh => $params{dbh},
                                                   policy => $params{policy},
                                                   object => {
                                                       url_id => $rec->{url__url_id},
                                                       url_name => $rec->{url__url_name},
                                                       app_id => $rec->{url__app_id},
                                                       method_name => $rec->{url__method_name},
                                                       is_wildcard => $rec->{url__is_wildcard},
                                                  });
  }
  elsif ($field_name eq 'linked_parameter' && $rec->{linked_parameter__param_id}) {
    $rec->{linked_parameter} = F5::ASMConfig::Entity::Parameter->new( dbh => $params{dbh},
                                                                   policy => $params{policy},
                                                                   get_criteria => {
                                                                         param_id => $rec->{linked_parameter__param_id},
                                                                   });
  }
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
  my($class, %params) = @_;

  $class->SUPER::map_fields_to_entity(%params);

  my $rec = $params{rec};

  if ($rec->{global_extracts__extract_attrib_id}) {
    $rec->{extract_name} = $rec->{global_extracts__extract_name};
    $rec->{applies_to_all} = $rec->{global_extracts__applies_to_all};
    $rec->{applies_to_regexp} = $rec->{global_extracts__applies_to_regexp};
    $rec->{global_extract_id} = $rec->{global_extracts__global_extract_id};
  }

}

#----------------------------------------------------------------------

sub get_external_entities {
    my ($self, %params) = @_;
    my @external_entities = @{$self->SUPER::get_external_entities(%params)};
    if ($self->{applies_to_urls}) {
        push (@external_entities, @{$self->{applies_to_urls}});
    }
    if ($self->{applies_to_filetypes}) {
        push (@external_entities, @{$self->{applies_to_filetypes}});
    }
    return \@external_entities;
}

#----------------------------------------------------------------------

sub unique_record_parts {
  my($class, %params) = @_;
  my $get_criteria = $params{get_criteria};
  my @unique_record_parts = grep { $class->fields->{$_}->{unique_record_part} } keys %{$class->fields};

  if ($params{all_parts} || ($get_criteria && $get_criteria->{extract_level})) {
    if ($params{all_parts} || ($get_criteria->{extract_level} == F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_URL)) {
      push(@unique_record_parts, 'url');
    }
  }

  return \@unique_record_parts;
}

#----------------------------------------------------------------------

sub before_store {
    my ($self, %params) = @_;

    return unless $params{validate_value};
    my @fields = qw(
                     applies_to_filetypes
                     applies_to_urls
                    );

    for my $field (@fields) {
       my $field_def = $self->fields->{$field};
       my $values = $self->get_new_value(%params,
                                         field_name => $field, );
       if ($values) {
           for my $value (@$values) {
               if (!defined($value->{is_wildcard})) {
                   $value = $field_def->{entity_type}->new(dbh => $self->{dbh},
                                                           policy => $self->{policy},
                                                           get_criteria => $value);                   
               }
               
               if ( defined $value->{policy} && (  $value->{policy}->{policy_id} != $self->{policy}->{policy_id}) ) {
                   my $display_field_name = $self->get_field_display_name_string(%params, field_name => $field);
                   my $message = "$display_field_name can not contain entities from other policies";
                   die { error_message => $message,
                         display_message => $message,
                         error_level     => 'info',
                         error_code      => F5::ASMConfig::ErrorCode::ILLEGAL_OPERATION(),
                       };
               }

               if ($value->{is_wildcard}) {
                   my $display_field_name = $self->get_field_display_name_string(%params, field_name => $field);
                   my $message = "$display_field_name can not contain wildcard entities";
                   die { error_message => $message,
                         display_message => $message,
                         error_level     => 'info',
                         error_code      => F5::ASMConfig::ErrorCode::ILLEGAL_OPERATION(),
                       };
               }

               if (defined $value->{flg_is_allowed} && !$value->{flg_is_allowed}) {
                   my $display_field_name = $self->get_field_display_name_string(%params, field_name => $field);
                   my $message = "$display_field_name can not contain disallowed entities";
                   die { error_message => $message,
                         display_message => $message,
                         error_level     => 'info',
                         error_code      => F5::ASMConfig::ErrorCode::ILLEGAL_OPERATION(),
                       };
               }

           }
       }
   }
}

#----------------------------------------------------------------------

sub validate_extraction_search_index {
    my ($self, %params) = @_;
    if ($self->get_new_value(%params, field_name => 'search_within_form')) {
        return $self->general_validate(%params);
    }
    return {
            completed_successfully => 1,
           };
}

#----------------------------------------------------------------------

sub after_store {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $store_obj = $params{store_obj};
  my $existing_obj = $params{existing_obj};

  my %extract_levels = (
                        F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_GLOBAL() => { extract_level => F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_GLOBAL,
                                                                                 table_name => 'PLC.PL_GLOBAL_EXTRACTS',
                                                                                 extra_field_func => sub {
                                                                                     my(%params) = @_;
                                                                                     my $db_rec = $params{extract_level_db_rec};
                                                                                     my $store_obj = $params{store_obj};
                                                                                     my $existing_obj = $params{existing_obj};

                                                                                     my %global_extract_fields = (
                                                                                                                  applies_to_all => { extern_db_field_name => 'flg_applies_to_all'},
                                                                                                                  applies_to_regexp => { extern_db_field_name => 'applies_to_regexp'},
                                                                                                                 );

                                                                                     for my $global_extract_field (keys %global_extract_fields) {
                                                                                         if ($self->has_value_changed(%params, check_fields => $global_extract_field)) {
                                                                                             $db_rec->{$global_extract_fields{$global_extract_field}->{extern_db_field_name}} =  $self->get_new_value(%params, field_name => $global_extract_field);
                                                                                         }
                                                                                     }
                                                                                 },
                                                                               },
                        F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_URL() => { extract_level => F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_URL,
                                                                              table_name => 'PLC.PL_OBJECT_EXTRACTS',
                                                                              extra_field_func => sub {
                                                                                my(%params) = @_;
                                                                                my $db_rec = $params{extract_level_db_rec};
                                                                                my $store_obj = $params{store_obj};
                                                                                my $existing_obj = $params{existing_obj};

                                                                                if ($store_obj->{url}) {
                                                                                    my $saved_url_name = $store_obj->{url}->{url_name} || '';

                                                                                    $store_obj->{url} = F5::ASMConfig::Entity::URL->new(dbh => $dbh,
                                                                                                                                        get_criteria => $store_obj->{url},
                                                                                                                                        minimal_load => 1,
                                                                                                                                        policy => $self->{policy});
                                                                                    unless ($store_obj->{url} && $store_obj->{url}->{url_id}) {
                                                                                        die "URL for Extraction not found ($saved_url_name) -- aborting\n";
                                                                                    }

                                                                                    if ($store_obj->{url}->{url_id} && 
                                                                                        (!$existing_obj->{url} ||
                                                                                         $store_obj->{url}->{url_id} ne $existing_obj->{url}->{url_id})) {

                                                                                        $db_rec->{obj_id} = $store_obj->{url}->{url_id};
                                                                                    }
                                                                                }
                                                                                elsif (!$existing_obj->{url}) {
                                                                                     die "URL for Extraction not found -- aborting\n";
                                                                                }
                                                                              },
                                                                            },
                       );

  my $extract_level_store_mode;

  if ($self->has_value_changed(%params, check_fields => 'extract_level')) {
    $extract_level_store_mode = 'add';

    if ($existing_obj->{extract_level}) {
      # Delete existing level row
      my $existing_extract_level_table_name = $extract_levels{$existing_obj->{extract_level}}->{table_name};

      $dbh->do(<<EndQuery, undef, $self->{extract_id});
DELETE FROM $existing_extract_level_table_name
      WHERE extract_attrib_id = ?
EndQuery
    }
  }
  else {
    $extract_level_store_mode = 'update';
  }

  my $extract_level_def = $extract_levels{$self->get_new_value(%params, field_name => 'extract_level')};
  my $extract_level_db_rec = {};
  if ($extract_level_def->{extra_field_func}) {
    $extract_level_def->{extra_field_func}->(%params, 
                                             extract_level_db_rec => $extract_level_db_rec);
  }

  my $fields = $self->fields;
  my %extract_level_fields = map { ($_ => $fields->{$_}) } grep { $fields->{$_}->{extern_db_field_name} } keys %$fields;

  for my $extract_level_field (keys %extract_level_fields) {
    if ($extract_level_store_mode eq 'add' || $self->has_value_changed(%params, check_fields => $extract_level_field)) {
      $extract_level_db_rec->{$extract_level_fields{$extract_level_field}->{extern_db_field_name}} =  $self->get_new_value(%params, field_name => $extract_level_field);
    }
  }

  my $db_change_made = 0;
  if (%$extract_level_db_rec) {
    if ($extract_level_store_mode eq 'add') {

      $extract_level_db_rec->{extract_attrib_id} = $self->{extract_id};
      $extract_level_db_rec->{policy_id} = $self->{policy}->{policy_id};
      my $global_extract_id;
      my $rv = F5::DbUtils::insert_row(dbh => $dbh,
                                       table_name => $extract_level_def->{table_name},
                                       rec => $extract_level_db_rec,
                                       last_insert_id => \$global_extract_id);
      $store_obj->{global_extract_id} = $global_extract_id;
      $db_change_made ||= $rv;
    }
    else {
      my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                       table_name => $extract_level_def->{table_name},
                                       rec => $extract_level_db_rec,
                                       where_fields => { extract_attrib_id => $self->{extract_id} });
      $db_change_made ||= $rv;
    }
  }

  if ($store_obj->{linked_parameter} && $store_obj->{linked_parameter}->{param_id} &&
      ( ! $existing_obj->{linked_parameter} || 
        $existing_obj->{linked_parameter}->{param_id} ne $store_obj->{linked_parameter}->{param_id})) {

    if ($existing_obj->{linked_parameter}) {
      $dbh->do(<<EndQuery, undef, $self->{extract_id});
DELETE FROM $self->{plc_dbname}.PL_DYNAMIC_PARAM_EXTRACTS
      WHERE extract_attrib_id = ?
EndQuery
    }

    my $rv = F5::DbUtils::insert_row(dbh => $dbh,
                                     table_name => 'PLC.PL_DYNAMIC_PARAM_EXTRACTS',
                                     rec => {
                                             policy_id => $self->{policy}->{policy_id},
                                             extract_attrib_id => $self->{extract_id},
                                             param_attrib_id => $store_obj->{linked_parameter}->{param_id},
                                            });
      $db_change_made ||= $rv;
  }

  return $db_change_made;
}

#----------------------------------------------------------------------

sub update_many_as_iterative {
    return 1;
}

#----------------------------------------------------------------------

sub empty_extraction_filters {
    my ($class_or_self, %params) = @_;
    my @empty_extraction_filters = ( F5::ASMConfig::SearchFilter->new(field_name => 'extract_level',
                                                                      field_value => F5::ASMConfig::ExtractLevel::EXTRACT_LEVEL_GLOBAL(),
                                                                      field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),),
                                     F5::ASMConfig::SearchFilter->new(field_name => 'global_extracts__applies_to_all',
                                                                      field_value => 0,
                                                                      field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),),
                                     F5::ASMConfig::SearchFilter->new(field_name => 'applies_to_filetypes',
                                                                      field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS(),),
                                     F5::ASMConfig::SearchFilter->new(field_name => 'applies_to_urls',
                                                                      field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS(),), );

    return \@empty_extraction_filters;
}

#----------------------------------------------------------------------

1;
