package F5::ASMConfig::Entity::Flow;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::Flow/;

#--------------------------------------------------------------------------

sub get_full_display_name {
    my ($self, %params) = @_;
    my $display_name = join (" ", $self->get_field_display_string(field_name => 'method', dbh => $params{dbh}),
                                  ($self->{from_url} && ( $self->{from_url}->{url_id} || $self->{from_url}->{url_name} ) ? $self->{from_url}->get_full_display_name(%params) . " to" : "(Entry Point)"),
                                  ($self->{to_url} && ( $self->{to_url}->{url_id} || $self->{to_url}->{url_name} ) ? $self->{to_url}->get_full_display_name(%params) : "UNNAMED"));

}

#--------------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $store_obj = $params{store_obj};
    return unless $params{validate_value};

    my $from_url = $self->get_new_value(%params, field_name => 'from_url');
    if ( $from_url && $from_url->{url_id} ) {
        if ( $from_url->{is_wildcard} ) {
            my $error_message = "Explicit URL must be specified.";
            die { error_message => $error_message,
                  display_message => $error_message,
                  error_level     => 'info',
                  error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                };
        }
    }
    else {
        $params{store_obj}->{from_url} = F5::ASMConfig::Entity::URL->new(dbh => $dbh, policy => $params{policy}, object => {url_id => 0});
    }

    my $to_url = $self->get_new_value(%params, field_name => 'to_url');
    if (! $to_url || ! ref($to_url) || ! $to_url->{url_id} || $to_url->{is_wildcard} ) {
         my $error_message = "Explicit URL must be specified";
         die { error_message => $error_message,
               display_message => $error_message,
               error_level     => 'info',
               error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
             };
    }
}

#--------------------------------------------------------------------------

sub get_retrieve_join_tables {
    my($class, %params) = @_;
    my $plc_dbname = 'PLC'; 

    my $retval = <<EndQuery;
LEFT JOIN $plc_dbname.PL_OBJECTS from_url
       ON (main_table.from_id = from_url.id)
LEFT JOIN $plc_dbname.PL_OBJECTS to_url
       ON (main_table.to_id = to_url.id)
EndQuery

    return $retval;
}

#----------------------------------------------------------------------

sub get_external_entities {
    my ($self, %params) = @_;
    if (!$self->{from_url} || $self->{from_url}->{url_id} == 0) {
        $params{exclude_fields} ||= [];
        push (@{$params{exclude_fields}}, 'from_url');
    }
    return $self->SUPER::get_external_entities(%params);
}

#----------------------------------------------------------------------

sub additional_field_mapping {
    my($class, %params) = @_;

    my %additional_field_mapping;
    my @url_groups = qw(from_url to_url);
    for my $url_group (@url_groups) {
        my %new_url_mapping =  (
                                "${url_group}__url_id" => { extern_db_field_name => 'id' },
                                "${url_group}__url_name" => { extern_db_field_name => 'uri' },
                                "${url_group}__app_id" => { extern_db_field_name => 'app_id' },
                                "${url_group}__method_name" => { extern_db_field_name => 'method_name' },
                                "${url_group}__is_wildcard" => { extern_db_field_name => 'flg_is_pattern' },
                               );
        %additional_field_mapping = (%additional_field_mapping,
                                     %new_url_mapping);
    }

    for my $k (keys %additional_field_mapping) {
        my @field_parts = split(/__/, $k);
        pop(@field_parts);
        $additional_field_mapping{$k}->{full_db_field_name} = join('__', @field_parts) . '.' . $additional_field_mapping{$k}->{extern_db_field_name};
    }

    return \%additional_field_mapping;
}
#----------------------------------------------------------------------

sub linked_entity_filters {
    my ($class_or_self, %params) = @_;
    my $entities = $params{entities} || [$class_or_self];
    my @entity_filters = ({filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'flow__flow_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types => ['F5::ASMConfig::Entity::Parameter'],});

    return @entity_filters;
}

#----------------------------------------------------------------------
sub map_fields_to_entity {
    my($class_or_entity, %params) = @_;
    my $dbh = $params{dbh} || $class_or_entity->{dbh};
    my $rec = $params{rec} || $class_or_entity;

    $class_or_entity->SUPER::map_fields_to_entity(%params);

    if ($rec->{flow_id}) {
        my $mandatory_params = F5::ASMConfig::Entity::Parameter->retrieve_all(dbh => $dbh,
                                                                              policy => $params{policy},
                                                                              page_size => 1,
                                                                              get_criteria => { flow => F5::ASMConfig::Flow->new(flow_id => $rec->{flow_id}),
                                                                                                flg_mandatory => 1 },
                                                                             );
                                                                          
        $rec->{num_mandatory_params} = $mandatory_params->total_record_count();
    }
}
#----------------------------------------------------------------------

1;
