package F5::ASMConfig::Entity::GraphqlProfile;

use strict;
use warnings;

use Memoize;

use F5::ASMConfig::ASMConfig;
use F5::DbUtils;
use F5::ASMConfig::ContentProfileBase;
use F5::ASMConfig::Entity::GraphqlDefenseAttribute;
use base qw/F5::ASMConfig::ContentProfileBase F5::ASMConfig::GraphqlProfile/;

#----------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags   = $params{call_flags};

    if ( defined($store_obj->{idl_files}) ) {
        my $nested_def = $self->fields->{idl_files};
        my $store_idl_files = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                                nested_recs => $store_obj->{idl_files},
                                                               );

        for my $key (keys %$store_idl_files) {
            if ( ! $store_idl_files->{$key}->{idl_file} ) {
                delete $store_idl_files->{$key}
            } elsif (defined($store_idl_files->{$key}->{import_url}) &&
                    ! length($store_idl_files->{$key}->{import_url})
                    ) {
                delete $store_idl_files->{$key}->{import_url};
            }
        }

        $store_obj->{idl_files} = [ values %$store_idl_files ];
    }

    $self->populate_defense_attributes(store_obj => $store_obj);
    if ($params{validate_value}) {
        $self->validate_defense_attributes(defense_settings => $store_obj->{graphql_defense});
    }

    for my $graphql_defense (@{$store_obj->{graphql_defense}}) {
        if (defined $graphql_defense->{def_attrib_value} &&
            $graphql_defense->{value_type} eq F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER) {
            $graphql_defense->{def_attrib_value} += 0;
        }
    }

    my @fields = qw(
                    block_disallowed_response_patterns
                    disallowed_response_patterns
                   );

   my %field_values;
   my %field_display_name;
   my $error_message;

   for my $field (@fields) {
      $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                    existing_obj => $existing_obj,
                                                    call_flags => $call_flags,
                                                    field_name => $field, );

      $field_display_name{$field} = $self->get_field_display_name_string(field_name => $field,%params);
   }

   if ( $field_values{block_disallowed_response_patterns} &&  (@{$field_values{disallowed_response_patterns}} == 0) ) {
       if ($call_flags->{cascade_disable_flags}) {
           $store_obj->{block_disallowed_response_patterns} = $field_values{block_disallowed_response_patterns} = 0;
       }
       else {
           $error_message =  "$field_display_name{disallowed_response_patterns} list must not be empty when $field_display_name{block_disallowed_response_patterns} is set";
       }
   }

   if ($error_message){
      die { error_message => $error_message,
            display_message => $error_message,
            error_level     => 'info',
            error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
          };
   }

}

#----------------------------------------------------------------------

sub after_nested_struct_store {
  my ($self, %params) = @_;
  my $dbh = $self->{dbh};
  my $store_obj = $params{store_obj};
  if ($params{store_mode} eq 'update') {
      $self->populate_defense_attributes(store_obj => $store_obj);
  }
}

#----------------------------------------------------------------------

sub after_nested_struct_load {
    my ($self, %params) = @_;
    $self->populate_defense_attributes();
    return;
}

#----------------------------------------------------------------------

sub populate_defense_attributes {
    my ($self, %params) = @_;
    my $store_obj = $params{store_obj} || $self;

    if (exists($store_obj->{graphql_defense})) {
        my $defense_attributes = $self->get_defense_attributes();
        for my $graphql_defense (@{$store_obj->{graphql_defense}}) {
            %$graphql_defense = ( %$graphql_defense, %{$defense_attributes->{$graphql_defense->{def_attrib_id}}} );
        }
    }
}

#----------------------------------------------------------------------

sub get_defense_attributes {
    my ($self, %params) = @_;
    my $dbh = $params{dbh} || $self->{dbh};

    my %retval = map {$_->{def_attrib_id} => $_} F5::ASMConfig::Entity::GraphqlDefenseAttribute->retrieve_all( dbh => $dbh, );

    return $self->install_accessor('get_defense_attributes', \%retval);
}

#----------------------------------------------------------------------

sub update_parent_staging {
    my ($self, %params) = @_;
    return $self->update_content_profile_staging(%params);
}

#----------------------------------------------------------------------

sub to_declarative {
    my ($self, %params) = @_;

    my $entity = $self->SUPER::to_declarative(%params);

    for my $idl_file (@{$entity->{idlFiles}}) {
        delete $idl_file->{idlFile}->{contents};
    }

    return $entity;
}

#----------------------------------------------------------------------

1;
