package F5::ASMConfig::Entity::Header;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::Header/;

#----------------------------------------------------------------------

sub update_many_as_iterative {
    return 1;
}

#----------------------------------------------------------------------

my %FACTORY_HEADERS = ( 'cookie' => { is_wildcard => 0,
                                      percent_normalization => 0,
                                      flg_check_signatures => 0,
                                      decode_value_as_base64 => F5::ASMConfig::Base64Handling::BASE64_HANDLING_DISABLED(),
                                      flg_is_base64 => 0,
                                      uri_normalization => 0,
                                      html_normalization => 0,
                                      enable_normalization_violation => 0,
                                      flg_mask_value => 0,
                                    },
                       'authorization' => { is_wildcard => 0,
                                            flg_is_base64 => 0,
                                          },
                       'referer' => { is_wildcard => 0 },
                       'transfer-encoding' => { is_wildcard => 0 },
                       '*' => { is_wildcard => 1,
                                is_mandatory => 0,
                                flg_is_base64 => 0, },
                      );


sub before_store {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};

    my @fields = qw(
                       header_name 
                       is_wildcard 
                       flg_check_signatures
                       flg_is_base64
                       decode_value_as_base64
                       percent_normalization
                       uri_normalization
                       html_normalization
                       enable_normalization_violation
                       flg_mask_value
                       is_mandatory
                  );

    my %field_values;
    my %field_display_name;

    if ($store_obj && defined $store_obj->{header_name}) {
        $store_obj->{header_name} = $self->normalize_header_name(field_value => $store_obj->{header_name})->{normalized_value};
    }

    for my $field (@fields) {
        $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                      existing_obj => $existing_obj,
                                                      field_name => $field, );

        $field_display_name{$field} = $self->get_field_display_name_string(%params, field_name => $field);
    }

    if ($params{validate_value} && $field_values{flg_is_base64} && ($field_values{percent_normalization} || $field_values{uri_normalization}) ) {
        my $base64_field = $self->is_feature_enabled(feature_name => $self->fields->{decode_value_as_base64}->{feature_name}) ? 'decode_value_as_base64' : 'flg_is_base64';
        my $normalization_field = $field_values{percent_normalization} ? 'percent_normalization' : 'uri_normalization';
        my $illegal_combined_fields_message = "$field_display_name{$base64_field} and $field_display_name{$normalization_field} cannot be enabled simultaneously";

        die { error_message => $illegal_combined_fields_message,
              display_message => $illegal_combined_fields_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
          };
    }

    #Import declarative first attempts to "add" all objects and falls back on update if there's a duplicate.
    #This causes an issue for default object population for factory headers when attempting to "add", so we only validate them on updates since they can't be deleted anyways.
    if ( $params{store_mode} eq 'update' && exists($FACTORY_HEADERS{$field_values{header_name}}) ) {
        my $factory_values = $FACTORY_HEADERS{$field_values{header_name}};
        if ($params{validate_value}) {
            for my $field (keys %{$factory_values}) {
                if (defined ($store_obj->{$field}) && $store_obj->{$field} != $factory_values->{$field}) {
                  my $message;
                  if ($field_values{header_name} eq 'cookie') {
                      $message = "Settings for the cookie header handling must be configured in cookie configuration.";
                  }
                  else {
                      if ($field eq 'flg_is_base64') {
                          $field = $self->is_feature_enabled(feature_name => $self->fields->{decode_value_as_base64}->{feature_name}) ? 'decode_value_as_base64' : 'flg_is_base64';
                      }
                      my $field_display_string = $self->get_field_display_string(%params, field_name => $field, field_value => $store_obj->{$field});
                      $message = "$field_display_name{$field} cannot be set to '$field_display_string' for header '$field_values{header_name}'.";
                  }

                  die { error_message => $message,
                        display_message => $message,
                        error_level     => 'info',
                        error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                  };
                }
            }
        }
        %$store_obj = (%$store_obj, %$factory_values);
    }
    elsif ($field_values{is_wildcard}) {
        if ($params{validate_value} && $field_values{is_mandatory}) {
            my $message = "Wildcard headers may not be mandatory.";
            die { error_message => $message,
                  display_message => $message,
                  error_level     => 'info',
                  error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                };
        }
        $store_obj->{is_mandatory} = 0; 
    }
    
    $self->adjust_wildcards(%params);

    if ( !$field_values{flg_check_signatures} ) {
        $store_obj->{enable_normalization_violation} = 0;
        $store_obj->{percent_normalization} = 0;
        $store_obj->{uri_normalization} = 0;
        $store_obj->{html_normalization} = 0;
    }
}

#----------------------------------------------------------------------------

sub map_fields_to_entity {
  my($class, %params) = @_;

  $class->SUPER::map_fields_to_entity(%params);

  my $rec = $params{rec};

  if (! defined $rec->{decode_value_as_base64}) {
      if ($rec->{flg_is_base64}) {
          $rec->{decode_value_as_base64} = F5::ASMConfig::Base64Handling::BASE64_HANDLING_REQUIRED();
      }
      elsif ($rec->{flg_detect_base64}) {
          $rec->{decode_value_as_base64} = F5::ASMConfig::Base64Handling::BASE64_HANDLING_ENABLED();
      }
      else {
          $rec->{decode_value_as_base64} = F5::ASMConfig::Base64Handling::BASE64_HANDLING_DISABLED();
      }
  }
}

#----------------------------------------------------------------------

sub is_deletable {
  my ($self, %params) = @_;

  if ( $self->{flg_is_default} ) {
      my $msg = "Default header '$self->{header_name}' cannot be deleted.";
      return { error_message => $msg,
               display_message => $msg,
               error_level     => 'info',
               error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
             };
  }
  return $self->SUPER::is_deletable(%params);
}

#----------------------------------------------------------------------

sub from_rest {
    my ($self, %params) = @_;
    my $rv = $self->SUPER::from_rest(%params);
    if ($rv->{flg_is_base64}) {
        #For Backwards compatibility
        $rv->{flg_detect_base64} = 0;
    }
    elsif (defined $rv->{decode_value_as_base64}) {
        $rv->{flg_is_base64} = ($rv->{decode_value_as_base64} == F5::ASMConfig::Base64Handling::BASE64_HANDLING_REQUIRED()) ? 1 : 0;
        $rv->{flg_detect_base64}  = ($rv->{decode_value_as_base64} == F5::ASMConfig::Base64Handling::BASE64_HANDLING_ENABLED()) ? 1 : 0;
    }
    return $rv;
}

#----------------------------------------------------------------------



1;

