package F5::ASMConfig::Entity::LoginObject;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::BruteForce;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::LoginObject/;

#----------------------------------------------------------------------
sub update_many_as_iterative {
    return 1;
}
#----------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;
    my $dbh = $self->{dbh};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $policy = $params{policy} || $self->{policy};

    return unless $params{validate_value};

    my $url = $self->get_new_value(%params, field_name => 'url');

    if (! $url || ! ref($url) || ! $url->{url_id}) {
        my $error_message = 'A valid URL must be specified';
        die { error_message => $error_message,
              display_message => $error_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
            };
    };

    if(! $url->{flg_is_allowed}){
        my $error_message = 'The specified URL [' . $url->{url_name} . ']  is Disallowed.';
        die { error_message => $error_message,
              display_message => $error_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
            };
    }

    my $brute_force = F5::ASMConfig::Entity::BruteForce->new(%params,
                                                             dbh => $self->{dbh},
                                                             logger => $self->{logger},
                                                             get_criteria => { url => $self->get_new_value(%params, field_name => 'url') });

    my $brute_force_display_name = F5::ASMConfig::Entity::BruteForce->get_element_type_display_name;
    my $login_object_display_name = $self->get_element_type_display_name;

    if ($brute_force && $self->get_new_value(%params, field_name => 'auth_type') == F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_NONE()) {
        my $auth_type_display = $self->get_field_display_name_string(%params, field_name => 'auth_type');
        $auth_type_display .= ' ' . $self->get_field_display_string(%params, field_name => 'auth_type');

        die "$login_object_display_name cannot be changed to $auth_type_display  while in use by $brute_force_display_name\n";
    }

    my @fields = qw(
                   cookie_contains
                   header_contains
                   header_omits
                   param_contains
                   response_contains
                   response_http_status
                   response_http_status_omits
                   response_omits
                   auth_type
                   username_param
                   password_param 
                   );

   my %field_values;
   my %field_display_name;

   for my $field (@fields) {
      $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                    existing_obj => $existing_obj,
                                                    field_name => $field, );

      $field_display_name{$field} = $self->get_field_display_name_string(%params, field_name => $field);
   }

   if ( $field_values{auth_type} == F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_FORM && 
        ( !$field_values{password_param}  || !$field_values{username_param})  ){
          my $error_message =  "$field_display_name{username_param} and $field_display_name{password_param} are required";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
              };
   }

   if ( ! ( length($field_values{cookie_contains})
            || length($field_values{header_contains})
            || length($field_values{header_omits})
            || length($field_values{param_contains})
            || length($field_values{response_contains})
            || length($field_values{response_http_status})
            || length($field_values{response_http_status_omits})
            || length($field_values{response_omits}) ) ){
          my $error_message =  "At least one of the Access Validation criteria must be defined";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
              };
   }

   if ( defined $field_values{response_contains} && defined $field_values{response_omits} && 
        length($field_values{response_contains}) && length($field_values{response_omits}) &&
        $field_values{response_contains} eq $field_values{response_omits} ){
          my $error_message =  "$field_display_name{response_contains} and $field_display_name{response_omits} must differ";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
              };
   }


   if ( defined $field_values{response_http_status} && defined $field_values{response_http_status_omits} &&
        length($field_values{response_http_status_omits}) && length($field_values{response_http_status_omits}) ){

          my %response_http_status_omits;
          for my $response_http_status_omit ( @{$field_values{response_http_status_omits}} ){
              $response_http_status_omits{$response_http_status_omit} = 1;
          }

          if ( $response_http_status_omits{$field_values{response_http_status}} ) {
              my $error_message =  "$field_display_name{response_http_status} and $field_display_name{response_http_status_omits} must differ";
              die { error_message => $error_message,
                    display_message => $error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                  };
          }
   }

   if (  $self->{auth_type} == F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_JSON() ) {
       my $url_content_profiles_iter = F5::ASMConfig::Entity::UrlContentProfile->retrieve_all(dbh => $dbh,
                                                                                              get_criteria => { url__url_id => $url->{url_id} },
                                                                                              search_filters => [F5::ASMConfig::SearchFilter->new(field_name => 'enforcement_type',
                                                                                                                                                  field_value => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON(),
                                                                                                                                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ())
                                                                                                                ],
                                                                                              );


       if (!$url_content_profiles_iter->{total_record_count}){
              my $error_message =  "There are no JSON profiles assigned to the URL";
              die { error_message => $error_message,
                    display_message => $error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                  };
       }

  }

  # XML/JSON are case sensitive, regardless of the policy case sensitivity
  if ($policy->{flg_is_case_insensitive} && defined $field_values{auth_type} &&
        $field_values{auth_type} != F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_JSON()) {

      # in case we are in transition from JSON to FORM auth_type,
      #     store_obj may not have these fields populated
      $store_obj->{username_param} ||= $field_values{username_param};
      $store_obj->{password_param} ||= $field_values{password_param};

      $store_obj->{username_param} =~ y/A-Z/a-z/;
      $store_obj->{password_param} =~ y/A-Z/a-z/;
  }

}

#----------------------------------------------------------------------
sub get_missing_external_entities {
    my ($self, %params) = @_;
    
    my @missing_entities;    
    if (  $self->{auth_type} == F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_JSON() && $self->{url}) {
            my $url_content_profile = F5::ASMConfig::Entity::UrlContentProfile->new( dbh => $self->{dbh},
                                                                                     policy => $self->{policy},
                                                                                     object => { header_name  => '*',
                                                                                                 header_value => '*',
                                                                                                 header_order => 0,
                                                                                                 enforcement_type =>  F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON(),
                                                                                                 url => $self->{url},
                                                                                               });
    
         @missing_entities = @{$url_content_profile->get_missing_external_entities(%params)};
         push (@missing_entities, $url_content_profile);
    }
    else{
        return $self->SUPER::get_missing_external_entities(%params);
    }

    return \@missing_entities;
}

#----------------------------------------------------------------------

sub get_case_insensitive_fields {
    my ($class, %params) = @_;
    return $class->SUPER::get_case_insensitive_fields(field_names => [ qw(password_param username_param) ]);
}

#----------------------------------------------------------------------

sub purify_store_object {
    my ($class_or_self, %params) = @_;
    my $object = $params{object} || $class_or_self;
    delete $object->{object_id};
    return $class_or_self->SUPER::purify_store_object(%params);
}

#----------------------------------------------------------------------

sub case_insensitive_search_filters {
    return [ F5::ASMConfig::SearchFilter->new(field_name => 'auth_type',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                                              field_value => F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_JSON()),
           ];
}

#----------------------------------------------------------------------

1;
