package F5::ASMConfig::Entity::Parameter;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::ContentProfile;
use List::Util qw(max);
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::Parameter/;

my $MAX_INT_LENGTH = 20;
#----------------------------------------------------------------------

sub get_full_display_name {
    my ($self, %params) = @_;
    my $display_name = $self->{param_name} || "UNNAMED";
    if ($self->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL) {
        $display_name .= " on " . ($self->{url} ? $self->{url}->get_full_display_name(dbh => $self->{dbh}) : "UNNAMED" );
    }
    elsif ($self->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW) {
         $display_name .= " on " . ($self->{flow} ? $self->{flow}->get_full_display_name(dbh => $self->{dbh}) : "UNNAMED" );
    }
    return $display_name;
}

#----------------------------------------------------------------------

#Prevent diff of "half" parameters just in case.
sub is_valid_diff_type {
    my ($class, %params) = @_;
    my $obj = $params{obj};
    if ($obj->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL) {
        return $obj->{url};
    }
    elsif ($obj->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW) {
        return $obj->{flow};
    }
    return 1;
}

#----------------------------------------------------------------------

sub get_retrieve_join_tables {
    my($class, %params) = @_;
    my %join_tables;

    my $plc_dbname = 'PLC';

    $join_tables{global_params} = <<EndQuery;
LEFT JOIN $plc_dbname.PL_GLOBAL_PARAMS global_params
       ON (main_table.id = global_params.param_attrib_id)
EndQuery

    $join_tables{url_params} = <<EndQuery;
LEFT JOIN $plc_dbname.PL_OBJECT_PARAMS url_params
       ON (main_table.id = url_params.param_attrib_id)
LEFT JOIN $plc_dbname.PL_OBJECTS url
       ON (url_params.obj_id = url.id)
EndQuery

    $join_tables{flow_params} = <<EndQuery;
LEFT JOIN $plc_dbname.PL_FLOW_PARAMS flow_params
       ON (main_table.id = flow_params.param_attrib_id)
LEFT JOIN $plc_dbname.PL_FLOWS flow
       ON (flow_params.flow_id = flow.id)
LEFT JOIN $plc_dbname.PL_OBJECTS flow__from_url
       ON (flow.from_id = flow__from_url.id)
LEFT JOIN $plc_dbname.PL_OBJECTS flow__to_url
       ON (flow.to_id = flow__to_url.id)
EndQuery

    return join('', (map { $join_tables{$_} } $class->extern_db_prefixes(%params)), $class->SUPER::get_retrieve_join_tables(%params));
}

#----------------------------------------------------------------------

sub extern_db_prefixes {
    my($class, %params) = @_;
    if ($params{get_criteria} && $params{get_criteria}->{param_level} && ! $params{get_criteria}->{param_id}) {
        if ($params{get_criteria}->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL) {
            return qw(global_params);
        }
        if ($params{get_criteria}->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL) {
            return qw(url_params);
        }
        if ($params{get_criteria}->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW) {
            return qw(flow_params);
        }
    }
    else {
        return qw(global_params url_params flow_params);
    }
}

#----------------------------------------------------------------------

sub extern_unique_record_part_db_fields {
    my($class, %params) = @_;

    if ( $params{get_criteria} && $params{get_criteria}->{db_prefix} ) {
        if ($params{get_criteria}->{db_prefix} eq 'url_params' ) {
            return ['obj_id', 'name'];
        }
        elsif ($params{get_criteria}->{db_prefix} eq 'flow_params' ) {
            return ['flow_id', 'name'];
        }
    }
    return [];
}

#----------------------------------------------------------------------

sub get_staging_subquery {
    my($class, %params) = @_;
    my $policy = $params{policy};
    my $want_suggestions = $params{has_suggestions} ? 1 : 0;

    my $plc_dbname = 'PLC';
    my $query = <<EndQuery;
  LEFT JOIN (
       SELECT DISTINCT P.rest_uuid AS union_rest_uuids, 1 AS matched FROM $plc_dbname.PL_PARAM_ATTRIBUTES AS P
          JOIN PLC.PL_SUGGESTIONS AS S ON (S.policy_id = ? AND S.suggestion_status = ? AND P.rest_uuid = S.element_rest_uuid AND S.element_type = ?)) params_suggestions
       ON (params_suggestions.union_rest_uuids = main_table.rest_uuid)
  LEFT JOIN (
       SELECT DISTINCT P1.rest_uuid AS union_rest_uuids, 1 AS matched FROM $plc_dbname.PL_PARAM_ATTRIBUTES AS P1
          JOIN $plc_dbname.PL_PARAM_CONTENT_PROFILES AS P2 ON (P1.policy_id = ? AND P2.param_attrib_id = P1.id)
          JOIN $plc_dbname.PL_CONTENT_PROFILES AS PR1 ON (P2.profile_id = PR1.id)
          JOIN $plc_dbname.PL_SUGGESTIONS AS S1 ON (S1.policy_id = ? AND S1.suggestion_status = ? AND PR1.rest_uuid = S1.element_rest_uuid AND S1.element_type IN (?, ?))) profile_suggestions
       ON (profile_suggestions.union_rest_uuids = main_table.rest_uuid)
EndQuery
    return { query => $query,
         bind_values => [ $policy->{policy_id}, F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING(), $class->element_type,
                          $policy->{policy_id},
                          $policy->{policy_id}, F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING(), F5::ASMConfig::ElementType::XML_PROFILE, F5::ASMConfig::ElementType::JSON_PROFILE],
         db_field => "CASE WHEN (params_suggestions.matched OR profile_suggestions.matched) THEN $want_suggestions ELSE NOT $want_suggestions END"};
}

#----------------------------------------------------------------------

sub additional_field_mapping {
    my($class, %params) = @_;

    my $param_level;
    if ($params{get_criteria} && ! $params{get_criteria}->{param_id}) {
        $param_level = $params{get_criteria}->{param_level};
    }
    my %additional_field_mapping;
    my @url_groups;

    if (!$param_level || $param_level == F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL) {
        $additional_field_mapping{global_params__param_attrib_id} = { extern_db_field_name => 'param_attrib_id' };
    }

    if (!$param_level || $param_level == F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL) {
        $additional_field_mapping{url_params__param_attrib_id} = { extern_db_field_name => 'param_attrib_id' };
        push(@url_groups, 'url');
    }

    if (!$param_level || $param_level == F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW) {
        $additional_field_mapping{flow_params__param_attrib_id} = { extern_db_field_name => 'param_attrib_id' };
        $additional_field_mapping{flow__flow_id} = { extern_db_field_name => 'id' };
        push(@url_groups, 'flow__to_url', 'flow__from_url');
        my %method_mapping = ("flow__method" => { extern_db_field_name => 'method'});
        %additional_field_mapping = (%additional_field_mapping,
                                     %method_mapping);
    }

    for my $url_group (@url_groups) {
        my %new_url_mapping =  (
                                "${url_group}__url_id" => { extern_db_field_name => 'id' },
                                "${url_group}__url_name" => { extern_db_field_name => 'uri' },
                                "${url_group}__app_id" => { extern_db_field_name => 'app_id' },
                                "${url_group}__is_wildcard" => { extern_db_field_name => 'flg_is_pattern' },
                                "${url_group}__method_name" => { extern_db_field_name => 'method_name' },
                                "${url_group}__rest_uuid" => { extern_db_field_name => 'rest_uuid' },
                               );
        %additional_field_mapping = (%additional_field_mapping,
                                     %new_url_mapping);
    }

    for my $k (keys %additional_field_mapping) {
        my @field_parts = split(/__/, $k);
        pop(@field_parts);
        $additional_field_mapping{$k}->{full_db_field_name} = join('__', @field_parts) . '.' . $additional_field_mapping{$k}->{extern_db_field_name};
    }

    my $param_level_statement;
    if (! $param_level) {
        $param_level_statement = <<EndQuery;
CASE WHEN ( flow_params.param_attrib_id ) THEN @{[F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW]}
    WHEN ( url_params.param_attrib_id ) THEN @{[F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL]}
     ELSE @{[F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL]} END
EndQuery
    }
    else {
        my @param_level_table_alias = $class->extern_db_prefixes(%params);
        $param_level_statement = "CASE WHEN (${param_level_table_alias[0]}.param_attrib_id) THEN $param_level ELSE NULL END";
    }

    $additional_field_mapping{param_level} = { full_db_field_name => $param_level_statement,
                                               filter_using_having => 1,
                                             };

    if ($params{extra_fields} && grep {$_ eq "extra_scanner_vulns"} (@{$params{extra_fields}})) {
        $additional_field_mapping{extra_scanner_vulns} = { nested_struct => 1,
                                                           entity_type => 'F5::ASMConfig::ExtraScannerVuln',
                                                           foreign_id_field => 'param_attrib_id',
                                                           parent_db_field_name => 'id',
                                                           parent_entity_field_name => 'param_id',
                                                           db_table_name => 'PLC.PL_SCANNER_VULNS',
                                                           fields => {
                                                                         vuln_id => { db_field_name => 'vuln_id', 
                                                                                      unique_record_part => 1,},                                                              
                                                                     },
                                                         };
    }

    return \%additional_field_mapping;
}

#----------------------------------------------------------------------

sub populate_from_join_row {
    my ($class, %params) = @_;
    my $rec = $params{rec};
    my $field_name = $params{field_name};

    if ($field_name eq 'url' && $rec->{url_params__param_attrib_id}) {
        $class->copy_fields_from_joined_row(row => $rec,
                                            prefix => 'url_params');
        $rec->{url} = F5::ASMConfig::Entity::URL->new(dbh => $params{dbh},
                                                      policy => $params{policy},
                                                      object => {
                                                          url_id => $rec->{url__url_id},
                                                          url_name => $rec->{url__url_name},
                                                          app_id => $rec->{url__app_id},
                                                          method_name => $rec->{url__method_name},
                                                          is_wildcard => $rec->{url__is_wildcard},
                                                          rest_uuid   => $rec->{url__rest_uuid},},
                                                   );
    }
    elsif ($field_name eq 'flow' && $rec->{flow_params__param_attrib_id}) {
        $class->copy_fields_from_joined_row(row => $rec,
                                            prefix => 'flow_params');
        my $from_url;
        if ($rec->{flow__from_url__url_id} == 0) {
            $from_url = F5::ASMConfig::Entity::URL->new(dbh => $params{dbh},
                                                        object => {url_id => 0});
        }
        else {
            $from_url = F5::ASMConfig::Entity::URL->new(dbh => $params{dbh},
                                                        policy => $params{policy},
                                                        object => {
                                                           url_id => $rec->{flow__from_url__url_id},
                                                           url_name => $rec->{flow__from_url__url_name},
                                                           app_id => $rec->{flow__from_url__app_id},
                                                           method_name => $rec->{flow__from_url__method_name},
                                                           is_wildcard => $rec->{flow__from_url__is_wildcard},
                                                           rest_uuid   => $rec->{flow__from_url__rest_uuid},},
                                                        );
        }
        my $to_url = F5::ASMConfig::Entity::URL->new(dbh => $params{dbh},
                                                     policy => $params{policy},
                                                     object => {
                                                         url_id => $rec->{flow__to_url__url_id},
                                                         url_name => $rec->{flow__to_url__url_name},
                                                         app_id => $rec->{flow__to_url__app_id},
                                                         method_name => $rec->{flow__from_url__method_name},
                                                         is_wildcard => $rec->{flow__to_url__is_wildcard},
                                                         rest_uuid   => $rec->{flow__to_url__rest_uuid},},
                                                   );
        $rec->{flow} = F5::ASMConfig::Entity::Flow->new(dbh => $params{dbh},
                                                        policy => $params{policy},
                                                        object => {
                                                            flow_id => $rec->{flow__flow_id},
                                                            from_url => $from_url,
                                                            to_url => $to_url,
                                                            method => $rec->{flow__method}},
                                                       );
    }
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
  my($class, %params) = @_;

  $class->SUPER::map_fields_to_entity(%params);

  my $rec = $params{rec};

  if ($rec->{global_params__param_attrib_id}) {
    $class->copy_fields_from_joined_row(row => $rec,
                                        prefix => 'global_params');
  }

  if (! defined $rec->{decode_value_as_base64}) {
      if ($rec->{flg_is_base64}) {
          $rec->{decode_value_as_base64} = F5::ASMConfig::Base64Handling::BASE64_HANDLING_REQUIRED();
      }
      elsif ($rec->{flg_detect_base64}) {
          $rec->{decode_value_as_base64} = F5::ASMConfig::Base64Handling::BASE64_HANDLING_ENABLED();
      }
      else {
          $rec->{decode_value_as_base64} = F5::ASMConfig::Base64Handling::BASE64_HANDLING_DISABLED();
      }
  }
}

#----------------------------------------------------------------------

sub copy_fields_from_joined_row {
  my($class, %params) = @_;
  my $row = $params{row};
  my $prefix = $params{prefix};

  $row->{param_name} = $row->{"${prefix}__param_name"};
  $row->{is_wildcard} = $row->{"${prefix}__is_wildcard"};
  $row->{wildcard_sequence} = $row->{"${prefix}__wildcard_sequence"};
}

#----------------------------------------------------------------------

sub unique_record_parts {
  my($class, %params) = @_;
  my $get_criteria = $params{get_criteria};
  my @unique_record_parts = grep { $class->fields->{$_}->{unique_record_part} } keys %{$class->fields};

  if ($params{all_parts} || ($get_criteria && $get_criteria->{param_level})) {
    if ($params{all_parts} || ($get_criteria->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL)) {
      push(@unique_record_parts, 'url');
    }
    if ($params{all_parts} || ($get_criteria->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW)) {
      push(@unique_record_parts, 'flow');
    }
  }

  return \@unique_record_parts;
}

#----------------------------------------------------------------------

sub rest_unique_record_parts {
  my ($class, %params) = @_;
  my $get_criteria = $params{get_criteria};
  my @rest_unique_record_parts = grep { $class->fields->{$_}->{rest_unique_record_part}
                                         || ($class->fields->{$_}->{unique_record_part} && !$class->fields->{$_}->{external_struct})
                                      } keys %{$class->fields};

  if ($params{all_parts} || ($get_criteria && $get_criteria->{param_level})) {
    if ($params{all_parts} || ($get_criteria->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL)) {
      push(@rest_unique_record_parts, 'url');
    }
    if ($params{all_parts} || ($get_criteria->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW)) {
      push(@rest_unique_record_parts, 'flow');
    }
  }

  @rest_unique_record_parts = sort { $a cmp $b } @rest_unique_record_parts;
  return \@rest_unique_record_parts;
}
#----------------------------------------------------------------------

sub before_store {
    my ($self, %params) = @_;
    
    my $dbh = $self->{dbh};
    
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags = $params{call_flags};
    my $policy = $params{policy} || $self->{policy};

    my $rv;
    my @modified_entities;
    
    if( $params{validate_value} && !$call_flags->{suggestion_rest_uuid}) {
       my @fields = qw(
                       param_value_type
                       param_data_type
                       param_content_profile
                       param_name
                       min_value
                       max_value
                       max_value_len
                       flg_check_max_value_len
                       flg_check_max_value
                       flg_check_min_value
                       flg_regexpr
                       regexpr_values
                       is_wildcard
                       flg_in_classification
                       parameter_location
                       array_serialization_format
                       flg_check_multiple_of_value
                       multiple_of_value
                       min_value_len
                       flg_check_min_value_len
                       flg_is_header
                       flg_mandatory
                       param_level
                       flg_is_cookie
                       object_serialization_style
                       flg_object_serialization_explode
                     );

       my %field_values;
       my %field_display_name;
    
       for my $field (@fields) {
          $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                        existing_obj => $existing_obj,
                                                        field_name => $field, );
    
          $field_display_name{$field} = $self->get_field_display_name_string(field_name => $field);
       }
       
       if ($params{store_mode} eq 'add') {
           if ($field_values{flg_is_header}) {
               $field_values{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_HEADER();
               $store_obj->{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_HEADER();
           }
           elsif ($field_values{flg_is_cookie}) {
               $field_values{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE();
               $store_obj->{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE();
           }

           if ($field_values{parameter_location}) {
               if ($field_values{parameter_location} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_HEADER()) {
                   $field_values{flg_is_header} = 1;
                   $store_obj->{flg_is_header} = 1;
               }
               elsif ($field_values{parameter_location} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE()) {
                   $field_values{flg_is_cookie} = 1;
                   $store_obj->{flg_is_cookie} = 1;
               }
           }
       }

       my $content_profile_valid = 1;
       my $content_profile_error_message = "";
       if( $call_flags->{rest_call} && ( $field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_XML_VALUE() || $field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_JSON_VALUE() ) ) {
           my $profile_type_title = ( $field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_XML_VALUE() ? "XML" : "JSON" );
           my $profile_type = ( $field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_XML_VALUE() ? F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML() : F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON() );
           $content_profile_error_message = "You must specify " . $profile_type_title  . " profile.";

           if ($params{store_mode} eq 'update' && ! $field_values{param_content_profile}) {
               $self->populate_diff_structs();
               $field_values{param_content_profile} = $self->get_new_value( store_obj => $store_obj,
                                                                            existing_obj => $existing_obj,
                                                                            field_name => 'param_content_profile');
           }

           #cases when profile not passed or passed by reference, id or selflink (in these cases get is performed to retrieve profile details)
           if( !$field_values{param_content_profile} || !ref($field_values{param_content_profile}) ||
               !$field_values{param_content_profile}->{content_profile} || !ref($field_values{param_content_profile}->{content_profile}) ||
               !$field_values{param_content_profile}->{content_profile}->{profile_name} ) {

               my $default_content_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                           dbh => $self->{dbh},
                                                                           policy => $policy,
                                                                           content_profile_type => $profile_type,
                                                                          );
               
               if ($default_content_profile->{__new_entity}) {
                   $default_content_profile->store_add(validate_value => $params{validate_value});
                   push(@modified_entities, {event_type => F5::ASMConfig::EventType::ADD, entity => $default_content_profile, existing_entity => $default_content_profile});
               }

               $store_obj->{param_content_profile} = F5::ASMConfig::Entity::ParameterContentProfile->new(dbh => $self->{dbh},
                                                                                                         object => {content_profile => $default_content_profile,
                                                                                                                    parameter => $self});
           }
           elsif( defined $field_values{param_content_profile}->{content_profile}->{content_profile_type} && $field_values{param_content_profile}->{content_profile}->{content_profile_type} != $profile_type )  {
               $content_profile_valid = 0;
           } 
           elsif(!defined $field_values{param_content_profile}->{content_profile}->{content_profile_type}) { #cases when name passed - make get and validate profile
               my $param_profile = F5::ASMConfig::Entity::ContentProfile->new(%params,
                                     dbh => $self->{dbh},
                                     logger => $self->{logger},
                                     get_criteria => { 
                                                      profile_name => $field_values{param_content_profile}->{content_profile}->{profile_name},
                                                      content_profile_type => $profile_type,
                                                     });
               if( !$param_profile ) { 
                   $content_profile_valid = 0;
               } 
               else {
                   $params{store_obj}->{param_content_profile}->{content_profile}->{content_profile_type} = $param_profile->{content_profile_type};
               }
               
           }
           
           if(!$content_profile_valid) {
               die { error_message => $content_profile_error_message,
                     display_message => $content_profile_error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                   };
           }
       }
       elsif ( $field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT() && ! $field_values{param_data_type}) {
           $params{store_obj}->{param_data_type} = F5::ASMConfig::ParamDataType::PARAM_DATA_ALPHA_NUMERIC();
       }

       if ($field_values{is_wildcard}) {
           my $parameter_location_name;
           if ($field_values{flg_is_header}) {
               $parameter_location_name = 'header';
           }
           elsif ($field_values{flg_is_cookie}) {
               $parameter_location_name = 'cookie';
           }
           elsif ($field_values{parameter_location} && $field_values{parameter_location} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_PATH()) {
               $parameter_location_name = 'path';
           }
           if ($parameter_location_name) {
               my $error_message = "Wildcard parameter with location \"$parameter_location_name\" is not allowed.";
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::MUTUALLY_EXCLUSIVE_FIELDS(),
                   };
           }
       }

       if ($field_values{flg_is_header}) {
           $field_values{param_name} = lc($field_values{param_name});
           my $error_message;
           if (length($field_values{param_name}) == 0) {
               $error_message = 'Empty parameter name with location "header" is not allowed.';
           }
           elsif ($field_values{param_name} eq "cookie") {
               $error_message = 'parameter name "Cookie" with location "header" is not allowed.';
           }
           elsif ($field_values{param_data_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_BINARY()) {
               $error_message = '"Binary" parameter data type is not allowed for parameter with location "header".';
           }

           if ($error_message) {
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::MUTUALLY_EXCLUSIVE_FIELDS(),
               };
           }

           $store_obj->{param_name} = $field_values{param_name};
       }

       if ($field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_ARRAY_VALUE() &&
           $field_values{array_serialization_format} == F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MULTI() &&
           $field_values{parameter_location} != F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_QUERY() &&
           $field_values{parameter_location} != F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE() &&
           $field_values{parameter_location} != F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_FORM_DATA()) {
           my $error_message = '"Multi" serialization format allowed only for parameter location "Query", "Cookie" or "Form Data".';
           die { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::SWAGGER_SERIALIZATION_ERRORS(),
               };
       }

       if ($field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_OBJECT_VALUE() &&
           $field_values{parameter_location} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE() &&
           $field_values{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_FORM &&
           $field_values{flg_object_serialization_explode} == 1) {
           my $error_message = 'Serialization method style "form" with explode "true" for cookie parameter of type object is not supported.';
           die { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::SWAGGER_SERIALIZATION_ERRORS(),
           };
       }

       if ($field_values{param_value_type} != F5::ASMConfig::ParamValueType::PARAM_VALUE_OBJECT_VALUE() && $field_values{object_serialization_style} &&
           $field_values{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_DEEP_OBJECT) {
           my $error_message = 'Serialization method style "deep object" supported only for parameter of type object.';
           die { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::SWAGGER_SERIALIZATION_ERRORS(),
           };
       }


       if ( ( $field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_DYNAMIC_CONTENT_VALUE ||
               $field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_DYNAMIC_PARAM_NAME ) && $field_values{is_wildcard} ){
           my $dynamic_param_error_message = "Wildcard parameter may not be Dynamic.\n";
           die { error_message => $dynamic_param_error_message,
                     display_message => $dynamic_param_error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                   };
       }

       if ( $params{store_mode} eq 'add' &&  $field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_DYNAMIC_PARAM_NAME ){

           my $param_criteria = { param_name => $field_values{param_name},
                                  param_value_type => F5::ASMConfig::ParamValueType::PARAM_VALUE_DYNAMIC_PARAM_NAME,
                                };


           my $dyn_param_name =  F5::ASMConfig::Entity::Parameter->new(policy => $policy,
                                                                       dbh => $self->{dbh},
                                                                       logger => $self->{logger},
                                                                       get_criteria => $param_criteria,
                                                                      );

           if ($dyn_param_name) {
               my $error_message = "A dynamic name parameter with the same name is already defined.";
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                 };
           }
       }

       if (($field_values{parameter_location} || 0) == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_PATH) {
           my $error_message;
           if ($field_values{flg_mandatory}) {
               $error_message = "Parameter with 'Path' location cannot be mandatory.";
           }
           elsif ($field_values{param_level} != F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL) {
               $error_message = "'Path' location allowed only for global parameter.";
           }

           if ($field_values{param_data_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_URI()) {
               $error_message = "Path location is not allowed with URI Data Type.";

           }

           if ($error_message) {
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                   };
           }
       }

       my $error_msg = '';
       if ($field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT()) {
           if ($field_values{flg_check_max_value_len} && defined $field_values{max_value_len} && 
                    ($field_values{max_value_len} < 0 || length ($field_values{max_value_len}) > 8)) {
               $error_msg = "'$field_display_name{max_value_len}' must have a positive numeric value and is also limited to 8 characters.";
           }
           elsif ($field_values{flg_check_min_value_len} && defined ($field_values{min_value_len}) && $field_values{min_value_len} < 0) {
               $error_msg = "'$field_display_name{min_value_len}' must have a positive numeric value.";
           }
           elsif ($field_values{flg_check_max_value_len} && ! defined ($field_values{max_value_len})) {
               $error_msg = "'$field_display_name{flg_check_max_value_len}' can not be enabled. Missing '$field_display_name{max_value_len}'.";
           }
           elsif ($field_values{flg_check_min_value_len} && $field_values{flg_check_max_value_len} &&
                  defined $field_values{min_value_len} && defined $field_values{max_value_len} &&
                  $field_values{min_value_len} > $field_values{max_value_len}) {
               $error_msg = "'$field_display_name{min_value_len}' can not be more than '$field_display_name{max_value_len}'.";
           }
           elsif ($field_values{param_data_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_ALPHA_NUMERIC) {
                if ($field_values{flg_regexpr} && (!defined $field_values{regexpr_values} || length ($field_values{regexpr_values}) == 0)) {
                    $error_msg = "'$field_display_name{regexpr_values}' must not be empty.";
                }
           }
           elsif ($field_values{param_data_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_BINARY) {
                if ($field_values{parameter_location} != F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_FORM_DATA() &&
                    $field_values{parameter_location} != F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_ANY) {
                    $error_msg = "'File Upload' data type allowed only for 'Form Data' location.";
                }
	   }
           elsif ($field_values{param_data_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_DECIMAL || $field_values{param_data_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_INTEGER) {
                if ( $field_values{flg_check_min_value} && defined $field_values{min_value} && length ($field_values{min_value}) > $MAX_INT_LENGTH) {
                    $error_msg = "'$field_display_name{min_value}' is limited to $MAX_INT_LENGTH characters.";
                }
                elsif ($field_values{flg_check_max_value} && defined $field_values{max_value} && length ($field_values{max_value}) > $MAX_INT_LENGTH) {
                    $error_msg = "'$field_display_name{max_value}' is limited to $MAX_INT_LENGTH characters.";
                }
                elsif ($field_values{flg_check_min_value} && $field_values{flg_check_max_value} &&
                         defined $field_values{min_value} && defined $field_values{max_value} && $field_values{min_value} > $field_values{max_value}) {
                    $error_msg = "'$field_display_name{min_value}' can not be more than '$field_display_name{max_value}'.";
                }
                elsif ($field_values{flg_check_max_value} && ! defined ($field_values{max_value})) {
                    $error_msg = "'$field_display_name{flg_check_max_value}' can not be enabled. Missing '$field_display_name{max_value}'.";
                }
                elsif ($field_values{flg_check_min_value} && ! defined ($field_values{min_value})) {
                    $error_msg = "'$field_display_name{flg_check_min_value}' can not be enabled. Missing '$field_display_name{min_value}'.";
                }
           }
       }

       if (!$error_msg && defined $field_values{multiple_of_value} && $field_values{multiple_of_value} <= 0) {
           $error_msg = "'$field_display_name{multiple_of_value}' must be greater than 0.";
       }

       if($error_msg) {
           die { error_message => $error_msg,
                 display_message => $error_msg,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
           };
       }
 
       if ($field_values{flg_in_classification}) {
           if ($params{store_mode} eq 'add') {
               $store_obj->{flg_in_classification} &&= $self->parameter_classification_flag(%params);
           }
           elsif ($self->has_value_changed(%params, check_fields => "flg_in_classification")){
               $store_obj->{flg_in_classification} = 0;
           }
       }
       if ($field_values{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_DYNAMIC_PARAM_NAME()) {
           $store_obj->{in_staging} = 0;
       }
   }

   if ($params{store_mode} eq 'add') {
       if ($call_flags->{suggestion_rest_uuid} &&
           !($params{thrift_obj} && defined($params{thrift_obj}->{flg_in_classification})) &&
           !($self->is_pure_wildcard(%params)) &&
           !($params{store_obj}->{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_XML_VALUE() 
           || $params{store_obj}->{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_JSON_VALUE() )) {
               #if this is being added due to a suggestion  and PB is at least in manual learning mode, inherit the classification status from PB_CONFIG
               my $flg_in_classification = $self->parameter_classification_flag(%params);
               $params{store_obj}->{flg_in_classification} = $flg_in_classification || 0;
               if ($flg_in_classification) {
                   $params{store_obj}->{param_value_type} = F5::ASMConfig::ParamValueType::PARAM_VALUE_IGNORE_VALUE();
                   $params{store_obj}->{param_data_type} = F5::ASMConfig::ParamDataType::PARAM_DATA_NONE();
               }
       }
   }
   elsif ( $self->has_value_changed(%params, check_fields => "param_value_type") &&
           $params{session}->{user_name} && $params{session}->{user_name} ne $F5::ASMConfig::Entity::Base::BIGIQ_USER_NAME ) {
       $params{store_obj}->{flg_in_classification} = 0;
   }


   $self->{param_name} =~ s/^\s+|\s+$//g; #remove leading and trailing spaces
   $self->sync_from_pb_config(%params, 
                              pb_config_field => "flg_parameters", 
                              table_name => F5::ASMConfig::Entity::PBConfigParameter->db_table());
   my $is_wildcard = $self->get_new_value(%params, field_name => 'is_wildcard');
   if ($params{validate_value} && $is_wildcard && !$self->{param_name}) {
       my $error_message = 'Wildcard parameters must have a name';
       die { error_message => $error_message,
             display_message => $error_message,
             error_level     => 'info',
             error_code      => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
           };        
   }

   $rv->{modified_entities} = \@modified_entities;
   return $rv;
}

#----------------------------------------------------------------------

sub after_store {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $store_obj = $params{store_obj};
  my $existing_obj = $params{existing_obj};
  my $policy = $params{policy};

  my %param_levels = (
                      F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL() => { param_level => F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL,
                                                                           table_name => 'PLC.PL_GLOBAL_PARAMS',
                                                                           extra_field_func => sub { 
                                                                             my(%params) = @_;
                                                                             my $db_rec = $params{param_level_db_rec};

                                                                             $self->adjust_wildcards(%params,
                                                                                                     db_table_name => 'PLC.PL_GLOBAL_PARAMS', 
                                                                                                    );
                                                                           },
                                                                         },
                      F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL() => { param_level => F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL,
                                                                        table_name => 'PLC.PL_OBJECT_PARAMS',
                                                                        extra_field_func => sub {
                                                                          my(%params) = @_;
                                                                          my $db_rec = $params{param_level_db_rec};
                                                                          my $store_obj = $params{store_obj};
                                                                          my $existing_obj = $params{existing_obj};

                                                                          my $url = $self->get_new_value(%params, field_name => 'url');
                                                                          if ($url) {
                                                                              my $saved_url_name = $store_obj->{url}->{url_name} || '';

                                                                              $store_obj->{url} = F5::ASMConfig::Entity::URL->new(dbh => $dbh,
                                                                                                                                  get_criteria => $url,
                                                                                                                                  minimal_load => 1,
                                                                                                                                  policy => $policy);

                                                                              unless ($store_obj->{url}) {
                                                                                  die "URL for Parameter not found ($saved_url_name) -- aborting\n";
                                                                              }

                                                                              if ($store_obj->{url}->{url_id} && 
                                                                                  (!$existing_obj->{url} ||
                                                                                   $store_obj->{url}->{url_id} ne $existing_obj->{url}->{url_id})) {

                                                                                  $db_rec->{obj_id} = $store_obj->{url}->{url_id};
                                                                              }
                                                                              $self->adjust_wildcards(%params,
                                                                                                      db_table_name => 'PLC.PL_OBJECT_PARAMS',
                                                                                                      where_values => { 'obj_id' => $db_rec->{obj_id} },
                                                                                                     );
                                                                          }
                                                                          else {
                                                                              die "URL for Parameter not found -- aborting\n";
                                                                          }
                                                                        },
                                                                      },
                      F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW() => { param_level => F5::ASMConfig::ParamLevel::PARAM_LEVEL_FLOW,
                                                                         table_name => 'PLC.PL_FLOW_PARAMS',
                                                                         extra_field_func => sub {
                                                                           my(%params) = @_;
                                                                           my $db_rec = $params{param_level_db_rec};

                                                                           if ($store_obj->{flow}) {
                                                                               $store_obj->{flow} = F5::ASMConfig::Entity::Flow->new(dbh => $dbh,
                                                                                                                                     get_criteria => $store_obj->{flow},
                                                                                                                                     minimal_load => 1,
                                                                                                                                     policy => $policy);

                                                                               unless ($store_obj->{flow}) {
                                                                                   die "Flow for Parameter not found -- aborting\n";
                                                                               }

                                                                               if ($store_obj->{flow}->{flow_id} &&
                                                                                   (!$existing_obj->{flow} ||
                                                                                    $store_obj->{flow}->{flow_id} ne $existing_obj->{flow}->{flow_id})) {

                                                                                   $db_rec->{flow_id} = $store_obj->{flow}->{flow_id};
                                                                               }
                                                                               $self->adjust_wildcards(%params,
                                                                                                       db_table_name => 'PLC.PL_FLOW_PARAMS',
                                                                                                       where_values => { 'flow_id' => $db_rec->{flow_id} },
                                                                                                      );
                                                                           }
                                                                         },
                                                                       },
                     );

  my $param_level_store_mode;

  if ($self->has_value_changed(%params, check_fields => 'param_level')  ||
                  $self->has_value_changed(%params, check_fields => 'url') ||
                     $self->has_value_changed(%params, check_fields => 'flow') ) {
      $param_level_store_mode = 'add';

      if ($existing_obj->{param_level}) {
          # Delete existing level row
          my $existing_param_level_table_name = $param_levels{$existing_obj->{param_level}}->{table_name};
          $dbh->do(<<EndQuery, undef, $self->{param_id});
DELETE FROM $existing_param_level_table_name
      WHERE param_attrib_id = ?
EndQuery

          # Reset wildcard order if changing param level
          if ($existing_obj->{is_wildcard}) {
              $existing_obj->{is_wildcard} = undef;
              $existing_obj->{wildcard_sequence} = undef;
              $store_obj->{is_wildcard} = 1;
              $store_obj->{wildcard_sequence} = undef;
          }
      }
  }
  else {
      $param_level_store_mode = 'update';
  }

  my $param_level_def = $param_levels{$self->get_new_value(%params, field_name => 'param_level')};
  my $param_level_db_rec = {};
  if ($param_level_def->{extra_field_func}) {
    $param_level_def->{extra_field_func}->(%params, 
                                           param_level_db_rec => $param_level_db_rec);
  }

  my $fields = $self->fields;
  my %param_level_fields = map { ($_ => $fields->{$_}) } grep { $fields->{$_}->{extern_db_field_name} && !$fields->{$_}->{extern_db_table} } keys %$fields;

  for my $param_level_field (keys %param_level_fields) {
    if (($param_level_store_mode eq 'add' || $self->has_value_changed(%params, check_fields => $param_level_field))
        && defined($self->get_new_value(%params, field_name => $param_level_field))) {
      $param_level_db_rec->{$param_level_fields{$param_level_field}->{extern_db_field_name}} = $self->get_new_value(%params, field_name => $param_level_field);
    }
  }

  my $db_table = $self->db_table;
  my $db_change_made = 0;
  if (%$param_level_db_rec) {
    if ($param_level_store_mode eq 'add') {

      $param_level_db_rec->{param_attrib_id} = $self->{param_id};
      $param_level_db_rec->{policy_id} = $policy->{policy_id};
      eval {
          my $rv = F5::DbUtils::insert_row(dbh => $dbh,
                                           table_name => $param_level_def->{table_name},
                                           rec => $param_level_db_rec);
          $db_change_made ||= $rv;
      };
      if (my $err = $@) {
          $dbh->do(<<EndQuery, undef, $self->{param_id});
    DELETE
      FROM $db_table
     WHERE id = ?
EndQuery

      delete $self->{param_id};
      die $@;
      }
    }
    else {
      my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                       table_name => $param_level_def->{table_name},
                                       rec => $param_level_db_rec,
                                       where_fields => { param_attrib_id => $self->{param_id} });
      $db_change_made ||= $rv;
    }
  }
  return $db_change_made;
}

#----------------------------------------------------------------------

sub linked_entity_filters {
    my ($class_or_self, %params) = @_;
    my $entities = $params{entities} || [$class_or_self];
    my @entity_filters = ({
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'linked_parameter__param_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types => ['F5::ASMConfig::Entity::Extraction'],
                          },
                          {
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'parameter__param_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types => ['F5::ASMConfig::Entity::ParameterContentProfile'],
                          },

                         );
    return @entity_filters;
}



#----------------------------------------------------------------------

sub is_pure_wildcard {
    my ($self, %params) = @_;
    return ($self->get_new_value(%params, field_name => 'param_name') &&
            $self->get_new_value(%params, field_name => 'param_name') eq "*" && 
            $self->get_new_value(%params, field_name => 'is_wildcard') && 
            $self->get_new_value(%params, field_name => 'param_level') == F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL);
}

#----------------------------------------------------------------------

sub clone_to_explicit {
    my ($self, %params) = @_;
    my $object = $params{object};
    if ($self->{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_IGNORE_VALUE() ||
        $self->{param_value_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_AUTO_DETECT()) {
        $object->{param_value_type} = F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT();
        $object->{param_data_type} = F5::ASMConfig::ParamDataType::PARAM_DATA_ALPHA_NUMERIC();
    }
    return $self->SUPER::clone_to_explicit(%params, object => $object);
}

#----------------------------------------------------------------------

sub get_missing_dependent_external_entities {
    my ($self, %params) = @_;
    my @missing_external_entities;

    my %content_profile_types = (F5::ASMConfig::ParamValueType::PARAM_VALUE_XML_VALUE() => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(),
                                 F5::ASMConfig::ParamValueType::PARAM_VALUE_JSON_VALUE() => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),
                                );


    if (my $content_profile_type = $content_profile_types{$self->{param_value_type}}) {

        $self->populate_diff_structs();
        my $param_content_profile = $self->{param_content_profile} || F5::ASMConfig::Entity::ParameterContentProfile->new(dbh => $self->{dbh}, 
                                                                                                                          policy => $self->{policy},
                                                                                                                          object => {parameter => F5::ASMConfig::Parameter->new($self)});
        if (!$param_content_profile->{content_profile} || $param_content_profile->{content_profile}->{content_profile_type} != $content_profile_types{$self->{param_value_type}}) {
            my $default_content_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                                     dbh => $self->{dbh},
                                                                                     policy => $self->{policy},
                                                                                     created_by => 'PolicyBuilder',
                                                                                     content_profile_type => $content_profile_type,
                                                                                 );

            if ($default_content_profile) {
                if ($default_content_profile->{__new_entity}) {
                    push (@missing_external_entities, $default_content_profile);
                }
                $param_content_profile->{content_profile} = $default_content_profile->uncast(); #In case of a new creation of a Content Profile Subclass

                push (@missing_external_entities, $param_content_profile);
            }
        }
    }
    return \@missing_external_entities;
}

#----------------------------------------------------------------------

sub parameter_classification_flag {
    my ($self, %params) = @_;
    my $policy = $params{policy} || $self->{policy};

    my $flg_in_classification = 0;
    ($flg_in_classification) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy->{policy_id},1);
SELECT flg_classify_parameters
  FROM PLC.PL_POLICY_BUILDER_CONFIG pb JOIN PLC.PL_POLICY_BUILDER_CONFIG_PARAMETER pbp ON pb.policy_id = pbp.policy_id
 WHERE pb.policy_id = ?
   AND flg_learning = ?
EndQuery

    return $flg_in_classification;
}

#----------------------------------------------------------------------

sub update_default_object_by_pb_settings {
    my ($self, %params) = @_;

    my $parameter = $params{object};

    if ($parameter->{flg_in_classification}) {
        if (!defined($parameter->{param_value_type}) && $self->parameter_classification_flag(%params)) {
            $parameter->{param_data_type} = F5::ASMConfig::ParamDataType::PARAM_DATA_NONE();
            $parameter->{param_value_type} = F5::ASMConfig::ParamValueType::PARAM_VALUE_IGNORE_VALUE();
        }
        elsif ($parameter->{param_value_type} != F5::ASMConfig::ParamValueType::PARAM_VALUE_IGNORE_VALUE()) {
            $parameter->{flg_in_classification} = 0;
        }
    }

    $self->SUPER::update_default_object_by_pb_settings(%params);
}

#----------------------------------------------------------------------

sub update_many_as_iterative {
    return 1;
}

#----------------------------------------------------------------------

my %profile_type_map = (
                        'json' => [ qw(json object openapi-array) ],
                        'xml' => [ qw(xml) ],
                       );

sub from_declarative {
    my ($self, %params) = @_;

    my $object = $params{object};
    my $field_name = $params{field_name};

    my $rv;
    if ($field_name eq 'param_content_profile' && $object->{valueType}) {
        for my $type (keys %profile_type_map) {
            if (grep { $_ eq $object->{valueType} } @{$profile_type_map{$type}}) {
                if (defined $object->{contentProfile} && ! defined $object->{contentProfile}->{contentProfile} ) {
                    $object->{contentProfile} = { contentProfile => $object->{contentProfile} };
                }
                $rv = { policy_reference => "$type-profiles" };
                last;
            }
        }
    }

    return $rv;
}

#----------------------------------------------------------------------

sub from_rest {
    my ($self, %params) = @_;
    my $rv = $self->SUPER::from_rest(%params);
    if ($rv->{flg_is_base64}) {
        #For Backwards compatibility
        $rv->{flg_detect_base64} = 0;
    }
    elsif (defined $rv->{decode_value_as_base64}) {
        $rv->{flg_is_base64} = ($rv->{decode_value_as_base64} == F5::ASMConfig::Base64Handling::BASE64_HANDLING_REQUIRED()) ? 1 : 0;
        $rv->{flg_detect_base64}  = ($rv->{decode_value_as_base64} == F5::ASMConfig::Base64Handling::BASE64_HANDLING_ENABLED()) ? 1 : 0;
    }
    return $rv;
}

#----------------------------------------------------------------------

1;
