package F5::ASMConfig::Entity::Policy;

use strict;
use warnings;

use File::Basename;
use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::Entity::AllowedMethod;
use F5::ASMConfig::Entity::L7Policy;
use F5::ASMConfig::Entity::Parameter;
use F5::ASMConfig::Entity::PolicyAttributes;
use F5::ASMConfig::Entity::PolicyTemplate;
use F5::ASMConfig::ASMConfig;
use F5::DbUtils;
use F5::Mcp::PolicyTarget;
use F5::Mcp::PolicyAction;
use F5::Utils::Rest;
use F5::ASMConfig::Entity::AllowedResponseCode;
use F5::ASMConfig::Entity::CustomXffHeader;
use F5::ASMConfig::Entity::PolicyNegsigAttributes;
use F5::ASMConfig::Entity::TrustXff;
use F5::ASMConfig::Entity::WebsocketUrl;
use F5::BigipUtils;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::Policy/;

use F5::ASMConfig::Cache;
use F5::CRC;
use File::stat;

my $SEC_LEVEL_STANDARD = 1;
my $DEFAULT_MAX_REQUEST_LENGTH = 8192;
my $CACHING_THRESHOLD = 10;
my $DEFAULT_TEMPLATE = 'POLICY_TEMPLATE_FUNDAMENTAL';
my $DEFAULT_PARENT_TEMPLATE = 'POLICY_TEMPLATE_FUNDAMENTAL';
my $BLANK_TEMPLATE = 'POLICY_TEMPLATE_BLANK';

#----------------------------------------------------------------------

sub policy {
    my ($self) = @_;
    return $self;
}

#----------------------------------------------------------------------

sub map_rest_virtual_servers {
    my ($self, %params) = @_;
    my $field_value = $params{field_value};
    my $direction = $params{direction};
    my $type = $params{type};
    my $rv;

    if ($field_value) {
        my @servers;
        if ($direction eq "to_rest") {
            for my $assoc (@$field_value) {
                if ($assoc->{assoc_type} == $type) {
                    push @servers , $assoc->{virtual_server}->{full_name} ;
                }
            }
        }
        else {
            @servers = map { F5::ASMConfig::VirtualServerAssoc->new(
                            virtual_server => F5::ASMConfig::VirtualServer->new(
                                full_name => $_)) } @$field_value;
        }
       $rv = \@servers;
    }
    return $rv;
}

#----------------------------------------------------------------------

sub map_rest_virtual_servers_complex {
    my ($self, %params) = @_;
    return $self->map_rest_virtual_servers(%params, type => F5::ASMConfig::VirtualServerAssocType::VIRTUAL_SERVER_ASSOC_TYPE_L7_COMPLEX);
}

#----------------------------------------------------------------------

sub map_rest_virtual_servers_simple {
    my ($self, %params) = @_;
    return $self->map_rest_virtual_servers(%params, type => F5::ASMConfig::VirtualServerAssocType::VIRTUAL_SERVER_ASSOC_TYPE_L7_SIMPLE);
}

#----------------------------------------------------------------------

sub get_retrieve_join_tables {
    my($class, %params) = @_;

    my $retval = $class->SUPER::get_retrieve_join_tables(%params);
   
    $retval .= <<EndQuery;
LEFT JOIN PLC.POLICY_TEMPLATES templ
       ON (plc__pl_policy_templates_settings.policy_template_id = templ.id)
LEFT JOIN PLC.PL_POLICY_BUILDER_CONFIG pb_config
       ON (main_table.id = pb_config.policy_id)
LEFT JOIN (SELECT username, from_date, policy_id FROM PLC.PL_POLICY_HISTORY WHERE to_date = 0 GROUP BY policy_id) pol_history
       ON (main_table.id = pol_history.policy_id)
LEFT JOIN PLC.PL_POLICIES parent_policy
       ON plc__pl_policy_parent.parent_policy_id = parent_policy.id
EndQuery

    return $retval;
}

#----------------------------------------------------------------------

sub retrieve_all {
    my ($self, %params) = @_;
    my $search_filters = $params{search_filters} || [];
    my $sort_orders = $params{sort_orders} || [];
    my $servers_filter;
    my $sort_by_policy_version_time;
    my $call_flags = $params{call_flags} || {};

    for my $search_filter (@$search_filters) {
        if (index($search_filter->{field_name}, 'extra_virtual_server_assocs') != -1) {
            $servers_filter = $search_filter;
        }
    }

    for my $sort_order (@$sort_orders) {
        if (index($sort_order->{field_name}, 'policy_version_time') != -1) {
            $sort_by_policy_version_time = $sort_order;
        }
    }

    if ( !$servers_filter && !$sort_by_policy_version_time) {
        return $self->SUPER::retrieve_all(%params);
    }

    my @selected_fields = ('policy_id', 'extra_virtual_server_assocs', 'policy_version_time');
    if (ref($call_flags->{rest_selected_fields})) {
        my $selected_fields_hash = $self->rest_selected_fields_to_selected_fields_hash(rest_selected_fields => $call_flags->{rest_selected_fields});
        push @selected_fields, keys %$selected_fields_hash;
    }

    if (ref($call_flags->{selected_fields})) {
        push @selected_fields, keys %{$call_flags->{selected_fields}};
    }

    my @rv = ();
    my $pre_rv_iterator = $self->SUPER::retrieve_all(%params,
                                                     return_as_raw_db_hashref => 0,
                                                     page_size => 0,
                                                     selected_fields => \@selected_fields,
                                                     extra_fields => ['extra_virtual_server_assocs']
                                                     );

    if ($servers_filter) {
        my $filter_field_name = $servers_filter->{field_name};
        my @name_fields = split('__', $filter_field_name);
        my $field_name = $name_fields[0];

        while (my $policy = $pre_rv_iterator->get_next()) {
            my $full_name = $servers_filter->{field_value}->{full_name};
            for my $virtual_server (@{$policy->{$field_name}}) {
                my $rv_filter = $policy->is_filter_match(search_filter => {field_value => $full_name, field_comparison => $servers_filter->{field_comparison}},
                                                        field_value => $virtual_server->{virtual_server}->{full_name},
                                                        match_as_positive_equivalent => 1);
                if ($rv_filter) {
                    push @rv , $policy;
                    last;
                }
            }
        }
    }
    elsif ($sort_by_policy_version_time) {
        while (my $policy = $pre_rv_iterator->get_next()) {
            push @rv , $policy;
        }
    }

    $self->apply_sort_order(result_set => \@rv, %params);
    my $total_record_count = scalar @rv;

    $self->apply_paging(result_set => \@rv, %params);

    if (wantarray) {
        return @rv;
    }
    else {
        my $iterator = F5::ASMConfig::Entity::Iterator->new(
                                                        get_next_function => sub {return shift(@rv);},
                                                        get_all_function => sub {return @rv;},
                                                        total_record_count => $total_record_count,
                                                        );
        return $iterator;
    }
}

#----------------------------------------------------------------------

sub get_api_protection_profile {
    my ($class_or_entity, %params) = @_;
    my $dbh = $params{dbh} || $class_or_entity->{dbh};
    my $policy = $params{rec} || $class_or_entity;

    my ($api_protection_profile) = @{F5::ASM::COAPI::Query->new('profile_api_protection', {security_policy => $policy->{policy_name}})};

    return $api_protection_profile;
}

#----------------------------------------------------------------------

sub additional_field_mapping {
    my($class, %params) = @_;

    my %additional_field_mapping = (

                                    "templ__icontrol_name" => { extern_db_field_name => 'icontrol_name',
                                                                full_db_field_name => 'IFNULL(templ.icontrol_name, "")', # must be defined for load_mcp_fields
                                                              },
                                    "templ__policy_template_name" => { extern_db_field_name => 'name',
                                                                full_db_field_name => 'templ.name',
                                                              },
                                    "templ__policy_template_id" => { extern_db_field_name => 'id',
                                                                     full_db_field_name => 'templ.id',
                                                                   },
                                    "templ__rest_uuid" => { extern_db_field_name => 'rest_uuid',
                                                            full_db_field_name => 'templ.rest_uuid',
                                                          },
                                    "parent_policy_name" => { extern_db_field_name => 'name',
                                                              full_db_field_name => 'IFNULL(parent_policy.name, "none")',
                                                            },
                                    "is_learning_enabled" => {
                                                full_db_field_name => 'CASE WHEN (pb_config.policy_id = main_table.id) THEN pb_config.flg_learning ELSE 0 END',
                                                'always_select_field' => 1,
                                                             },
                                    "is_pb_enabled" => {
                                                full_db_field_name => 'CASE WHEN (pb_config.policy_id = main_table.id) THEN pb_config.flg_enabled ELSE 0 END',
                                                'always_select_field' => 1,
                                                       },
                                    "modified_time" => { extern_db_field_name => 'from_date',
                                                         full_db_field_name => 'IFNULL(pol_history.from_date, 0)' },

                                    "modifier_name" => { extern_db_field_name => 'username',
                                                         full_db_field_name => 'IFNULL(pol_history.username,"")'},

                                   );

    return \%additional_field_mapping;
}

#----------------------------------------------------------------------

sub populate_version_fields {
    my($class_or_self, %params) = @_;
    my $policy = $params{policy} || $class_or_self;
    my $dbh = $params{dbh} || $class_or_self->{dbh};

    my ($config_log) = $dbh->selectrow_hashref(<<EndQuery, undef, $policy->{policy_id});
SELECT * FROM PLC.PL_CONFIG_LOG
 WHERE is_policy_version_event = 1
   AND policy_id = ?
 ORDER BY config_log_id DESC
 LIMIT 1
EndQuery

    if ($config_log) {
        $policy->{"policy_version_time"} = $config_log->{policy_version_time};
        $policy->{"policy_version_device_name"} = $config_log->{policy_version_device_name};
        $policy->{"policy_version_policy_name"} = $config_log->{policy_version_policy_name};
        $policy->{"policy_version_last_change"} = $config_log->{policy_version_last_change};
    }
}

#----------------------------------------------------------------------
#
sub map_fields_to_entity {
    my($class_or_entity, %params) = @_;
    my $dbh = $params{dbh} || $class_or_entity->{dbh};
    my $rec = $params{rec} || $class_or_entity;

    if ($rec->{"templ__policy_template_id"}) {
        $rec->{policy_template} = F5::ASMConfig::Entity::PolicyTemplate->new(object => {
                                                                                icontrol_name => $rec->{"templ__icontrol_name"}, 
                                                                                policy_template_name => $rec->{"templ__policy_template_name"}, 
                                                                                policy_template_id => $rec->{"templ__policy_template_id"},
                                                                                rest_uuid => $rec->{"templ__rest_uuid"},
                                                                             },
                                                                             dbh => $dbh,
                                                                            );
    }

    my %extra_fields = ();
    if ( $params{extra_fields} ) {
        %extra_fields = map { ( $_ => 1 ) } @{$params{extra_fields}};
    }
    elsif( $params{call_flags} && $params{call_flags}->{extra_fields} ){
         %extra_fields = map { ( $_ => 1 ) } @{$params{call_flags}->{extra_fields}};
    }

    if ( $rec->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()
         && ($extra_fields{extra_virtual_server_assocs} || $extra_fields{extra_virtual_servers})
         && (!$params{selected_fields_hash} || $params{selected_fields_hash}->{extra_virtual_server_assocs})) {
        my (@vs_assocs_simple, @vs_assocs_complex);
        my $do_cache = ($params{found_rows} && $params{found_rows} > $CACHING_THRESHOLD);
        my %policy_names;
        if ($rec->{is_active_policy}) {
            %policy_names = map { ( $_->{policy_name} => 1) } (@{F5::ASM::COAPI::Query->new("policy_action", 
                                                                                            { target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM, 
                                                                                              action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE, 
                                                                                              policy => $rec->{policy_name}},
                                                                                              do_cache => $do_cache)});
        }

        for my $policy_name (keys %policy_names) {
            my $l7_policy = F5::ASMConfig::Entity::L7Policy->new(get_criteria => { l7_policy_name => $policy_name} , dbh => $dbh );
            next unless $l7_policy; #Sometimes policy_actions are left in MCP without corresponding l7_policy entities.
            for my $virtual_server (@{$l7_policy->{virtual_servers}}) {
                my $vs_assoc = F5::ASMConfig::VirtualServerAssoc->new({
                                                                       virtual_server => $virtual_server,
                                                                       assoc_type => $l7_policy->{type},
                                                                       l7_policy_name => $l7_policy->{l7_policy_name},
                                                                      });

                if ( $l7_policy->{type} == F5::ASMConfig::VirtualServerAssocType::VIRTUAL_SERVER_ASSOC_TYPE_L7_SIMPLE ) {
                    push (@vs_assocs_simple,$vs_assoc);
                } else {
                    push (@vs_assocs_complex,$vs_assoc);
                }
            }
            if (!(@{$l7_policy->{virtual_servers}})) {
                push (@{$rec->{unassociated_l7_policies}}, $l7_policy);
            }
        }
        if ($extra_fields{extra_virtual_server_assocs}) {
            $rec->{extra_virtual_server_assocs} = \@vs_assocs_simple;
            $rec->{extra_virtual_server_assocs_complex} = \@vs_assocs_complex;
        }

        if ($extra_fields{extra_virtual_servers}) {
           @{$rec->{extra_virtual_servers}} = map { $_->{virtual_server} } ( @vs_assocs_simple , @vs_assocs_complex );
        }
    }

    if (!$params{selected_fields_hash} || $params{selected_fields_hash}->{is_modified}) {
        $rec->{is_modified} = $class_or_entity->is_modified(policy_id => $rec->{policy_id},
                                                            dbh => $params{dbh},
                                                           );
    }
    if ($rec->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() &&
        (!$params{selected_felds_hash} || $params{selected_fields_hash}->{child_policy_count})) {
        $rec->{child_policy_count} = scalar @{$class_or_entity->child_policy_ids(policy_id => $rec->{policy_id},
                                                                                 dbh => $params{dbh})};
    }

    $rec->{account_name} = $rec->{policy_name};  #For EM backwards compatibility

    if ( (!$params{selected_fields_hash} && !$params{minimal_load}) ||
         ($params{selected_fields_hash} && ($params{selected_fields_hash}->{policy_version_time} ||
                                            $params{selected_fields_hash}->{policy_version_device_name} ||
                                            $params{selected_fields_hash}->{policy_version_policy_name} ||
                                            $params{selected_fields_hash}->{policy_version_last_change})) ) {           
        $class_or_entity->populate_version_fields(policy => $rec, dbh => $dbh);
    }

    if ($extra_fields{can_update_protocol_free}) {
        $rec->{can_update_protocol_free} = $class_or_entity->can_update_protocol_free(%params);
    }

    if ($extra_fields{api_protection_profile}) {
        my $api_protection_profile = $class_or_entity->get_api_protection_profile(%params);
        if (ref($api_protection_profile) && $api_protection_profile->{name}) {
            $rec->{api_protection_profile} = $api_protection_profile->{name};
        }
    }

    $class_or_entity->SUPER::map_fields_to_entity(%params);
}

#----------------------------------------------------------------------

sub can_update_protocol_free {
    my ($class_or_entity, %params) = @_;
    my $dbh = $params{dbh} || $class_or_entity->{dbh};
    my $policy = $params{rec} || $class_or_entity;

    my ($urls) = $dbh->selectrow_array(<<EndQuery,undef,$policy->{policy_id});
SELECT count(*)
  FROM PLC.PL_OBJECTS
 WHERE policy_id = ?
 GROUP BY uri, method_name
HAVING count(*) >= 2
EndQuery

   if ($urls){
       return 0;
   }

   my ($websocket_urls) = $dbh->selectrow_array(<<EndQuery,undef,$policy->{policy_id});
SELECT count(*)
  FROM PLC.PL_WEBSOCKET_URLS
 WHERE policy_id = ?
 GROUP BY uri
HAVING count(*) >= 2
EndQuery

   if ($websocket_urls){
       return 0;
   }

   return 1;
}

#----------------------------------------------------------------------

sub lang_set_id {
    my ($self) = @_;
    my ($lang_set_id) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $self->{encoding});
SELECT lang_set_id
  FROM DCC.LANGUAGE_ENCODINGS le JOIN DCC.ENCODINGS e ON le.encoding_id = e.encoding_id
 WHERE le.flg_visible = 1 AND e.encoding = ?
EndQuery

    return $lang_set_id;
}

#----------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;

    my $store_mode = $params{store_mode};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};

    if ( $self->get_new_value( %params, field_name => 'flg_transparent_mode') == 0 ) {
        $store_obj->{flg_passive_mode} = 0;
    }
    
    my $error_message;
    if ($store_mode eq 'update') {
        if ($self->has_value_changed(%params, check_fields => ["encoding","flg_encoding_selected"])) {
            my $flg_encoding_selected = $self->get_new_value( %params, field_name => 'flg_encoding_selected', );
            my $encoding = $self->get_new_value( %params, field_name => 'encoding', );
            if ($self->has_value_changed(%params, check_fields => "flg_encoding_selected") && $flg_encoding_selected) {
                $params{store_obj}->{mcp_encoding} = $encoding; #TODO: Replace when better mcp field mapping is done
            }
            else {
                $error_message = "Security Policy language encoding can not be changed once it has been selected.";
            }
        }
        if ((exists($params{store_obj}->{parent_policy_id}) && $self->has_value_changed(%params, check_fields => ['parent_policy_id']) && !$params{store_obj}->{parent_policy_id})
            || ($self->has_value_changed(%params, check_fields => ['has_parent']) && !$self->get_new_value( %params, field_name => 'has_parent'))) {
                $params{store_obj}->{parent_policy_id} = 0;
                $params{store_obj}->{parent_policy_name} = 'none'; #MCP field
        }

        #Templates can only be removed, not added afterwards.
        if ((exists($params{store_obj}->{policy_template_id}) && $self->has_value_changed(%params, check_fields => ['policy_template_id']))) {
            if (defined($params{store_obj}->{policy_template_id})) {
                delete $params{store_obj}->{policy_template_id};
            }
            else {
                $params{store_obj}->{templ__icontrol_name} = ''; #Update MCP
                $params{set_fields_to_null}->{policy_template_id} = 1;
            }
        }
    }

    if ($self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()) {
        if ($store_obj->{parent_policy_id} && $self->has_value_changed(%params, check_fields => ['parent_policy_id'])) {
            my $parent_policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => { policy_id => $store_obj->{parent_policy_id} });
            if ($parent_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()) {
                $error_message = 'Security Policy cannot be set as a Parent of another Policy.';
            }
            $params{store_obj}->{parent_policy_name} = $parent_policy->{policy_name}; #Update MCP
        }
    }

    if ($self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        if ($store_obj->{is_active_policy}) {
            $error_message = 'Parent Policy cannot be Active.';
        }
        elsif ($store_obj->{parent_policy_id}) {
            $error_message = 'Parent Policy cannot have a Parent.';
        }
    }

    if ($error_message) {
        die { error_message => $error_message,
              display_message => $error_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
            };
    }
}

#----------------------------------------------------------------------
sub after_store {
    my($self, %params) = @_;
    # called from store_update and from store_add
    F5::ASMConfig::Cache::clear_cached_policy(dbh => $self->{dbh}, policy => $self);
    return; # no db change made
}
#----------------------------------------------------------------------

sub is_mcp_object {
    my ($self) = @_;
    return 1;
}

#----------------------------------------------------------------------

sub load_mcp_fields {
    my ($self, %params) = @_;
    my $rv = $self->SUPER::load_mcp_fields(%params);
    if (!$self->{app_id}) {
        delete $self->{app_id};
        delete $self->{strict_app_updates};
    }
    return $rv;
}

#----------------------------------------------------------------------

sub get_last_updated {
    my ($self, %params) = @_;
    my $config_log = F5::ASMConfig::Entity::ConfigLog->new(dbh => $self->{dbh},
                                                           policy => $self,
                                                           get_criteria => {},
                                                           sort_orders => [ { field_name => 'event_time',
                                                                              sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC()}],
                                                           search_filters => [{ field_name  => 'element_type',
                                                                                field_value => F5::ASMConfig::ElementType::SET_ACTIVE(),
                                                                                field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE()}],);
    my $last_updated_time = $config_log ? $config_log->{event_time}
                                        : 0;
    return $last_updated_time;
}
#----------------------------------------------------------------------

sub is_modified {
  my ($self, %params) = @_;
  my $policy_id = $params{policy_id} || $self->{policy_id};
  my $dbname = (ref $self && $self->{plc_dbname}) ? $self->{plc_dbname} 
                                                  : "PLC";
  my $dbh = $params{dbh} || $self->{dbh};
  my ($is_modified) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT MAX(update_time) 
  FROM $dbname.PL_POLICY_CHANGES_LOG 
 WHERE policy_id = ?
EndQuery

  return $is_modified ? 1 : 0;  
}

#----------------------------------------------------------------------

sub is_modified_by_pb {
  my ($self, %params) = @_;
  my $policy_id = $params{policy_id} || $self->{policy_id};
  my $dbname = (ref $self && $self->{plc_dbname}) ? $self->{plc_dbname}
                                                  : "PLC";
  my $dbh = $params{dbh} || $self->{dbh};

  my ($is_modified) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id, $F5::ASMConfig::Entity::Base::POLICY_BUILDER_USER_NAME);
SELECT MAX(update_time)
  FROM $dbname.PL_POLICY_CHANGES_LOG
 WHERE policy_id = ? AND user_name = ?
EndQuery

  return $is_modified ? 1 : 0;
}

#----------------------------------------------------------------------

#Copied from F5::PolicyHistory, consider moving all F5::Policy(History) functionality into Entity.
sub last_revision_number {
  my($class_or_self, %params) = @_;
  my $dbh = $params{dbh} || $class_or_self->{dbh};
  my $policy_id = $params{policy_id} || $class_or_self->{policy_id};
  my($last_revision_number) = $dbh->selectrow_array('SELECT MAX(revision) FROM PLC.PL_POLICY_HISTORY WHERE policy_id = ?',
                                                   undef, $policy_id);
  return $last_revision_number || 0;
}


#----------------------------------------------------------------------
sub clear_modification_log {
  my ($self, %params) = @_;
  my @where_clauses = ('policy_id = ?');
  my @bind_values = ($self->{policy_id});

  if ($params{max_modification_log_id}) {
    push(@where_clauses, 'policy_changes_log_id <= ?');
    push(@bind_values, $params{max_modification_log_id});
  }
  my $where_clauses = 'WHERE ' . join(' AND ', @where_clauses);

  $self->{dbh}->do(<<EndQuery, undef, @bind_values);
DELETE FROM $self->{plc_dbname}.PL_POLICY_CHANGES_LOG 
$where_clauses
EndQuery

  return; 
}

#----------------------------------------------------------------------

sub get_max_modification_log_id {
  my ($self) = @_;
  my ($max_modification_log_id) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $self->{policy_id});
SELECT MAX(policy_changes_log_id)
  FROM $self->{plc_dbname}.PL_POLICY_CHANGES_LOG
 WHERE policy_id = ?
EndQuery

  return $max_modification_log_id;
}
#----------------------------------------------------------------------

sub charset_config {
    my ($policy_id,  $databases, $dbh) = @_;

    my $plc_db = $databases->{PLC} || 'PLC';
    my $dcc_db = $databases->{DCC} || 'DCC';
    $dbh->do(<<EndQuery,undef,$policy_id, $policy_id);
REPLACE INTO $plc_db.PL_CHARSETS (policy_id, charset_type, charset, section_type, rest_uuid)
                SELECT ?, default_charsets.charset_type, default_charsets.charset, default_charsets.section_type, default_charsets.rest_uuid
                FROM $dcc_db.DEFAULT_CHARSETS default_charsets, $plc_db.PL_POLICIES policies,
                     $dcc_db.ENCODINGS encodings
                WHERE policies.id=? AND policies.encoding=encodings.encoding 
                AND encodings.encoding_id = default_charsets.encoding_id
EndQuery
}

#----------------------------------------------------------------------

sub policy_name_fixup {
    my ($class_or_self, %params) = @_;
    my $object = $params{object} || $class_or_self;
    my $session = $params{session};

    $object->{leaf_name} ||= $object->{policy_name};
    if ($object->{policy_name} && ($object->{policy_name} =~ m|^/|)) {
        my @parts = split('/', $object->{policy_name});
        $object->{partition_id} = $parts[1];
        $object->{leaf_name} = $parts[-1];
        $object->{folder_name} = dirname($object->{policy_name});
    }
    elsif ($object->{folder_name} && ($object->{folder_name} =~ m|^/([^/]*)|) ) {
        $object->{partition_id} = $1;
        $object->{policy_name} = "$object->{folder_name}/$object->{leaf_name}";
    }
    elsif (!F5::BigipUtils::is_docker()) {
        #Leave name as given in containerized environment
        $object->{partition_id} ||= $session->{partition} || "Common";
        $object->{folder_name} = "/$object->{partition_id}";
        $object->{policy_name} = "$object->{folder_name}/$object->{leaf_name}";
    }
}

#----------------------------------------------------------------------

sub store_add {
  my($self, %params) = @_;
  my $dbh = $self->{dbh};
  my $session = $params{session};
  my $call_flags = $params{call_flags} || {};
  my $databases = $params{databases} || {PLC => $self->{plc_dbname}, DCC => $self->{dcc_dbname}};
  my $validate_value = defined $params{validate_value} ? $params{validate_value} : 1;
  my $db_rec = {};
  my %set_fields_to_null = $self->get_fields_to_set_null(%params);
  my $populate_rest_uuid = defined $params{populate_rest_uuid} ? $params{populate_rest_uuid} :  $call_flags->{populate_rest_uuid};


  $self->policy_name_fixup(session => $session);
  my $partition = $self->{partition_id};

  if (!$self->{policy_type}) {
      $self->{policy_type} = F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY();
  }
 
  my $policy_name = $self->{policy_name};
  my $policy_type = $self->{policy_type};
  my $rest_uuid   = $self->set_rest_uuid();
  my $is_active_policy = $self->{is_active_policy};
  my $creator_name = $session->{gui_user_name} || $session->{user_name} || 'SYSTEM';
  $self->before_store(%params,
                      db_rec => $db_rec,
                      store_mode => 'add',
                      store_obj => $self,
                      existing_obj => {},
                      set_fields_to_null => \%set_fields_to_null,
                      validate_value => $validate_value,
                      );

  $self->validate_field(%params,
                        field_value => $policy_name, 
                        field_def => $self->fields->{policy_name},
                        field_name => 'policy_name',
                        dbh => $dbh,
                       );

  $self->{policy_name_crc} = F5::CRC::get_crc($policy_name); 
  if ($self->{encoding}) {
     $self->{encoding} = lc($self->{encoding});
     $self->validate_field(%params,
                           field_name  => 'encoding',
                           field_value => $self->{encoding}, 
                           field_def   => $self->fields->{encoding},
                           dbh         => $dbh);
  }

  $self->{account_id} ||= 0;

  if (!$self->{policy_template}) {
      #Will fetch according to policy_template_id if it is set, otherwise will take the default template
      my $default_icontrol_name = $self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() ? $DEFAULT_PARENT_TEMPLATE
                                                                                                          : $DEFAULT_TEMPLATE;
      $self->{policy_template} = F5::ASMConfig::PolicyTemplate->new(icontrol_name => $default_icontrol_name,
                                                                    policy_template_id => $self->{policy_template_id},);
  }

  my $encoding = $self->{flg_encoding_selected} ? $self->{encoding} : '';
  my $template = $self->{policy_template};
  my $parent_policy_id = $self->{parent_policy_id};
  my $parent_policy_name = $self->{parent_policy_name};

  $template = F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $self->{dbh}, get_criteria => $self->{policy_template} );
  if (!$self->{flg_encoding_selected} && !$self->{is_pb_enabled}
      && $template && $template->{template_defaults} && $template->{template_defaults}->{learningMode} eq 'manual' ) {
     my $error_message = "Auto detect encoding cannot be set with manual learning mode.";
     die { error_message => $error_message,
           display_message => $error_message,
           error_level     => 'info',
           error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
         };
  }

  my $res = $self->from_template(policy_template => $template,
                                 target_account_id => $self->{account_id},
                                 encoding => $encoding,
                                 flg_encoding_selected => $self->{flg_encoding_selected},
                                 policy_name => $policy_name,
                                 policy_type => $policy_type,
                                 partition => $partition,
                                 creator_name => $creator_name,
                                 skip_mcp => $call_flags->{skip_mcp},
                                 populate_rest_uuid => $populate_rest_uuid,
                              );

  if ($template) {
     $self->{policy_id} = $res->{policy_id};
     $self->{templ__icontrol_name} = $template->{icontrol_name}; #TODO: Replace when better mcp field mapping is done
     $self->associate_policy_template(%params,
                                      populate_rest_uuid => $populate_rest_uuid,
                                      template => $template);
  }

  if ($parent_policy_id && $parent_policy_name) {
      $dbh->do(<<EndQuery, undef, $res->{policy_id}, $parent_policy_id,);
REPLACE INTO $databases->{PLC}.PL_POLICY_PARENT (policy_id, parent_policy_id) VALUES (?, ?)
EndQuery
     $self->{parent_policy_name} = $parent_policy_name; #TODO: Replace when better mcp field mapping is done
  }

  $self->{is_active_policy} = $is_active_policy;

  if ($res->{encoding_changed} ) {
      charset_config($self->{policy_id}, $databases, $dbh);
  }

  #Act as update for this object since PolicyAttributes is the same table.
  my $diff_struct_store_rv = $self->diff_struct_store(store_obj => $self,
                               existing_obj => F5::ASMConfig::Entity::Policy->new(dbh => $dbh, object => {policy_id => $self->{policy_id}}),
                               store_mode => 'update',
                               %params);    
  $self->after_store(%params);

  $self->{mcp_encoding} = $self->{flg_encoding_selected} ? $self->{encoding} : ''; #TODO: Replace when better mcp field mapping is done
  $self->{templ__icontrol_name} ||= '';
  $self->{parent_policy_name} ||= 'none';
  $self->mcp_store(%params);

  #Refresh full fields to be returned
  $self->retrieve(%params,
                  get_criteria => $self,
                 );

  my $retval = { db_change_made => 1,
                 modified_entities => $diff_struct_store_rv->{modified_entities}
               };

  return $retval;
}
#----------------------------------------------------------------------

sub from_template {
    my ($self, %params) = @_;

    my $policy_template = $params{policy_template};
    if (!$policy_template) {
        die "Policy Template must be defined\n";
    }

    my $xml_template_file = $policy_template->{template_file};
    my $bin_template_file = $xml_template_file;
    $bin_template_file =~ s/\.xml/.bin/;
    my $is_bin_file_exists = 0;

    if ( -e $bin_template_file ){
        if (-e $xml_template_file) {
            my $bin_update_time = stat($bin_template_file)->mtime;
            my $template_update_time = stat($xml_template_file)->mtime;

            if ( $bin_update_time < $template_update_time ) {
                unlink $bin_template_file;
            }
            else {
                $is_bin_file_exists = 1;
            }
        } 
    }

    my $encoding = $params{encoding};
    my $use_bin_template = $is_bin_file_exists && $encoding && $encoding eq 'utf-8';

    my $res = $self->import_policy(%params,
                                   populate_rest_uuid => !$use_bin_template,
                                   filename => $use_bin_template ? $bin_template_file : $xml_template_file);

    if (!$is_bin_file_exists && $encoding && $encoding eq 'utf-8'
         && $params{policy_type} eq F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()) {
        F5::ImportExportPolicy->new(dbh => $self->{dbh})->export_policy(filename  => $bin_template_file,
                                                                        policy_id => $res->{policy_id},
                                                                        file_format => 'binary',
                                                                       );
    }

    return $res;
}

#----------------------------------------------------------------------

sub import_policy {
    my ($self, %params) = @_;
 
    my $result;
    require F5::ImportExportPolicy;
    eval {
      my $import_export = F5::ImportExportPolicy->new(dbh => $self->{dbh});
      $result = $import_export->import_policy(%params);
    };

    if($@  || !$result->{completed_successfully}) {
        my $err = $@;
        if ($result->{msgs}) {
            $err = join("\n", map { "$_->{msg_level}: $_->{msg_text}" } @{$result->{msgs}});
        }
        die "InternalError - import_policy failed: $err\n";
    }

    # Refresh policy fields after import
    $self->from_object(object => F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                                    get_criteria => { policy_id => $result->{policy_id}}));

    return $result;
}

#----------------------------------------------------------------------
sub delete{
   my ($self,%params) = @_;
   
   # Check that the policy is not active before deleting it
   if ($self->{is_active_policy}) {
       die "Policy $self->{policy_name} is active\n";
   }

   if ($self->{policy_type} = F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT && $self->{child_policy_count}) {
       die "Parent Policy has children\n";
   }

   my $dbh = $self->{dbh};
   my $policy_id = $self->{policy_id};
   my $policy_name = $self->{policy_name};

   my $rv = $self->SUPER::delete(%params);
   require F5::PolicyHistory;
   F5::PolicyHistory::delete_all_for_policy($dbh, $policy_id);
   $self->empty_policy_tables(import_tables_only => 0);

   require F5::XMLProfile;
   F5::XMLProfile::clean_unreferenced_profile_files(dbh => $dbh);
   require F5::JSONProfile;
   F5::JSONProfile::clean_unreferenced_profile_files(dbh => $dbh);
   require F5::GRPCProfile;
   F5::GRPCProfile::clean_unreferenced_profile_files(dbh => $dbh);
   require F5::GraphqlProfile;
   F5::GraphqlProfile::clean_unreferenced_profile_files(dbh => $dbh);
   require F5::EmServices;
   F5::EmServices->new(dbh => $dbh)->delete_from_policy_transactions($policy_id);

   F5::ASMConfig::Cache::clear_cached_policy(dbh => $dbh, policy => $self);
   F5::ASMConfig::Entity::AllowedMethod->cleanup_user_defined_methods(dbh => $self->{dbh});
   return $rv;
}
#----------------------------------------------------------------------
sub empty_policy_tables {
  my ($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $policy_id = $self->{policy_id};

  my @tables = F5::DbUtils::get_import_policy_tables($self->{plc_dbname});

  if (!$params{import_tables_only}) {
    push(@tables, F5::DbUtils::get_nonimport_policy_tables($self->{plc_dbname}));
  }

  for my $table (@tables) {
    $dbh->do("DELETE FROM $table WHERE policy_id = ?", undef, $policy_id);
  }

  return 1;
}
#----------------------------------------------------------------------

sub from_rest {
    my ($self, %params) = @_;
    my $rv = $self->SUPER::from_rest(%params);
    if (defined $rv->{learning_mode}) {
        $rv->{is_learning_enabled} = ($rv->{learning_mode} != F5::ASMConfig::LearningMode::LEARNING_MODE_NONE()) ? 1 : 0;
        $rv->{is_pb_enabled}  = ($rv->{learning_mode} == F5::ASMConfig::LearningMode::LEARNING_MODE_AUTOMATIC()) ? 1 : 0;
    }
    return $rv;
}

#----------------------------------------------------------------------

sub affects_policy {
   my ($self, %params) = @_;
   return 1;
}

#----------------------------------------------------------------------

sub child_policy_ids {
    my ($self, %params) = @_; 
    my $dbh = $params{dbh} || $self->{dbh};
    my $policy_id = $params{policy_id} || $self->{policy_id};
    my $child_policy_ids = ref($self) ? $self->{child_policy_ids} : undef;

    if (!$child_policy_ids) {    
        $child_policy_ids = F5::DbUtils::get_column_values(
                                               dbh => $dbh,
                                               field_name => 'policy_id',
                                               table_name => 'PLC.PL_POLICY_PARENT',
                                               where_fields => {'parent_policy_id' => $policy_id });
        if (ref($self)) {
            $self->{child_policy_ids} = $child_policy_ids;
        }
    }

    return $child_policy_ids;
}

#----------------------------------------------------------------------

sub clear_traffic_data {
    my ($self, %params) = @_;
    return if F5::BigipUtils::is_docker();
    $self->logical_delete_from_request_log(%params);
    $self->clear_reporting_events(%params);
}

#------------------------------------------------------------------------------------

sub clear_suggestions {
    my ($self) = @_;
    F5::ASMConfig::Entity::Suggestion->delete_many(policy => $self,
                                                   dbh => $self->{dbh},
                                                   get_criteria => {},
                                                   suppress_events => 1,);
    return;
}

#------------------------------------------------------------------------------------

sub clear_reporting_events {
    my ($self) = @_;

    my @table_names = (F5::DbUtils::get_event_backup_tables('DCC'),
                       F5::DbUtils::get_recover_stat_tables('DCC'),
                      );

    for my $table_name (@table_names) {
        if (grep { $_ eq 'account_id' } F5::DbUtils::get_db_fields($self->{dbh}, $table_name)) {

            $self->{dbh}->do(<<EndQuery, undef, $self->{account_id});
DELETE
  FROM $table_name
 WHERE account_id = ?
EndQuery
        }
    }

    return;
}

#--------------------------------
sub logical_delete_from_request_log {
    my ($self) = @_;

    $self->{dbh}->do(<<EndQuery, undef, $self->{policy_id});
REPLACE INTO PRX.REQUEST_LOG_PROPERTIES (request_log_id, flg_is_deleted)
SELECT id, 1
  FROM PRX.REQUEST_LOG
 WHERE policy_id = ?
EndQuery
    return;
}

#----------------------------------------------------------------------

sub logical_delete_from_incidents {
    my ($self) = @_;

    $self->{dbh}->do(<<EndQuery);
LOCK TABLES PRX.INCIDENTS WRITE
EndQuery

    $self->{dbh}->do(<<EndQuery, undef, $self->{policy_id});
DELETE FROM PRX.INCIDENTS WHERE policy_id = ?
EndQuery

    $self->{dbh}->do(<<EndQuery);
UNLOCK TABLES
EndQuery

    return;
}


#----------------------------------------------------------------------

sub get_policies {
    my ($self) = @_;
    return ($self);
}

#----------------------------------------------------------------------

my $LAYERED_POLICY_VERSION = "13.0.0";

sub rest_search_filters {
    my ($class, %params) = @_;
    if ($params{call_flags} && $params{call_flags}->{rest_version} &&
        F5::BigipVersionUtils::compare_versions(version1 => $LAYERED_POLICY_VERSION,
                                                version2 => $params{call_flags}->{rest_version}) > 0) {

        return [ F5::ASMConfig::SearchFilter->new(field_name => 'policy_type',
                                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                  field_value => F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()),
               ];
    }
    return [];
}

#----------------------------------------------------------------------

sub self_link {
    my ($self, %params) = @_;
    my $link = $params{base_link} || '';
    $link .= '/' . $self->element_rest_path_name() . "/" . $self->{rest_uuid};
    $link = $self->add_ver_to_self_link(%params,
                                        'link' => $link);
    return $link;
}

#----------------------------------------------------------------------

sub clean_policy_for_inheritance {
    my ($self, %params) = @_;
    my $policy = ref($self) ? $self : $params{policy};
    my $dbh = $params{dbh} || $policy->{dbh};
    return unless (ref($policy) && $policy->{policy_type} && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT());

    #Get rid of all objects that don't belong in a parent policy
    for my $entity_type (map { $_->{entity_type} } grep { $_->{no_inheritance} || $_->{no_inheritance_conditions} || $_->{clean_from_parent} } values %{$self->full_schema()}) {
        $entity_type->clean_from_parent(%params, dbh => $dbh, policy => $policy);
    }
    F5::ASMConfig::Entity::PolicySignatureSet->update_many(dbh => $dbh,
                                                           policy => $policy,
                                                           search_criteria => {flg_learn => 0},
                                                           store_obj => {flg_learn => 1});
}

#----------------------------------------------------------------------

sub associate_policy_template {
    my ($self, %params) = @_;
    my $dbh = $self->{dbh};
    my $databases = $params{databases} || {PLC => $self->{plc_dbname}, DCC => $self->{dcc_dbname}};

    my $template = $params{template} || F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $dbh, get_criteria => $self->{policy_template} );

    return if !$template;

    $dbh->do(<<EndQuery, undef, $self->{policy_id}, $template->{policy_template_id},);
REPLACE INTO $databases->{PLC}.PL_POLICY_TEMPLATES_SETTINGS (policy_id,policy_template_id) VALUES (?, ?)
EndQuery
    if ($template->{template_type} == F5::ASMConfig::TemplateType::TEMPLATE_TYPE_GENERIC()
        && !$params{defined_signature_sets}
        && $template->{icontrol_name} ne $BLANK_TEMPLATE) {
        require F5::NegativeSignatures;
        my $negative_signatures = F5::NegativeSignatures->new(dbh => $dbh, dbname => $databases->{PLC});
        my %sig_set_args = (policy_id => $self->{policy_id},
                            policy_negsig_attributes => $self->{policy_negsig_attributes},
                            is_in_declarative_export => $params{call_flags}->{is_in_declarative_export}, );
        if ($self->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
            $sig_set_args{flg_learn} = 1;
        }
        elsif (!$self->{is_learning_enabled}) {
            $sig_set_args{flg_learn} = 0;
        }
        $negative_signatures->add_default_sets_to_policy(%sig_set_args);
        if (!defined $params{populate_rest_uuid} || $params{populate_rest_uuid}) {
            F5::Utils::Rest::populate_uuids(dbh => $dbh,
                                            policy_id => $self->{policy_id},
                                            rest_entities => ['F5::ASMConfig::Entity::PolicySignatureSet'],
                                            );
        }
    }
}

#----------------------------------------------------------------------

1;
