package F5::ASMConfig::Entity::PolicySignature;

use strict;
use warnings;

use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::PolicySignature/;

#----------------------------------------------------------------------

sub get_staging_subquery {
    my($class, %params) = @_;
    my $policy = $params{policy};
    my $want_suggestions = $params{has_suggestions} ? 1 : 0;

    my $plc_dbname = 'PLC';
    my $query = <<EndQuery;
LEFT JOIN (SELECT DISTINCT PS.sig_id AS sig_ids, 1 AS matched FROM $plc_dbname.PL_SUGGESTIONS AS PS
                JOIN PLC.PL_SUGGESTIONS AS S ON (S.policy_id = ? AND S.suggestion_status = ? AND S.sig_id != 0 AND PS.sig_id = S.sig_id)) sig_suggestions
        ON (sig_suggestions.sig_ids = main_table.sig_id)
EndQuery
    return { query => $query,
         bind_values => [ $policy->{policy_id}, F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING() ],
         "db_field" => "CASE WHEN (sig_suggestions.matched) THEN $want_suggestions ELSE NOT $want_suggestions END"};
}


#----------------------------------------------------------------------

my %SUBQUERY_FIELDS = (is_in_policy => 1,
                       is_inherited => 1,
                       flg_alarm => 1,
                       flg_reject => 1,
                       flg_learn  => 1,);

sub get_retrieve_join_tables {
    my($class, %params) = @_;
    my $policy = $params{policy};
    my $extra_fields = $params{extra_fields} ? { map { ( $_ => 1 ) } @{$params{extra_fields}}} : {};
    my $search_filters = $params{search_filters} ? { map { ( $_->{field_name} => 1 ) } @{$params{search_filters}}} : {};

    my $query = '';
    my @bind_values;
    my $where_group_clauses;

    if ($params{search_filters} && (my ($search_filter) = grep { $_->{field_name} eq 'flg_disabled_differs_from_parent_policy_id' } @{$params{search_filters}})) {
        push (@bind_values, $search_filter->{field_value});
        $query .= <<EndQuery;
 JOIN PLC.PL_POLICY_NEGSIG_SIGNATURES parent_sigs 
  ON (parent_sigs.policy_id = ?
      AND parent_sigs.sig_id = main_table.sig_id
      AND parent_sigs.flg_disabled != main_table.flg_disabled)
EndQuery
    }

    if (!$params{only_if_needed_for_count} || ($params{search_filters} && (grep { $SUBQUERY_FIELDS{$_->{field_name}}} @{$params{search_filters}}))) {
        if ($policy && $policy->{policy_id}) {
            push (@bind_values, $policy->{policy_id});
            $where_group_clauses = <<EndQuery
        WHERE pol_sets.set_id = set_sigs.set_id
          AND pol_sets.policy_id = ?
     GROUP BY sig_id
EndQuery
            }
        else {
            $where_group_clauses = <<EndQuery
        WHERE pol_sets.set_id = set_sigs.set_id
     GROUP BY policy_id, sig_id
EndQuery
        }

        $query .= <<"EndQuery";
 LEFT JOIN (
            SELECT set_sigs.sig_id,
                   pol_sets.policy_id,
                   (1) AS is_in_policy,
                   MAX(pol_sets.is_inherited) AS is_inherited,
                   MAX(pol_sets.flg_alarm) AS flg_alarm,
                   MAX(pol_sets.flg_reject) AS flg_reject,
                   MAX(pol_sets.flg_learn) AS flg_learn
              FROM PLC.PL_POLICY_NEGSIG_SETS pol_sets,
                   PLC.NEGSIG_SET_SIGNATURES set_sigs
$where_group_clauses
) set_subquery
        ON (    main_table.sig_id = set_subquery.sig_id
            AND main_table.policy_id = set_subquery.policy_id)
EndQuery
    }

    if ( ($extra_fields && $extra_fields->{extra_has_suggestions}) || ($search_filters && $search_filters->{extra_has_suggestions}) ) {
        my $staging_subquery = $class->get_staging_subquery(%params,
                                                             policy => $params{policy},
                                                             has_suggestions => 1,);
        $query .= $staging_subquery->{query};
        push (@bind_values, @{$staging_subquery->{bind_values}});
    }

    return { query => $query,
             bind_values => \@bind_values,
           };
}

#----------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;

    my $store_mode = $params{store_mode};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};

    if ($store_mode eq 'update') {
    	if (($self->has_value_changed(%params, check_fields => 'flg_quarantined') &&
    	    $self->get_new_value(%params, field_name => 'flg_quarantined') == 0) ||    	    
    	    ($self->has_value_changed(%params, check_fields => 'flg_disabled') &&
    	    $self->get_new_value(%params, field_name => 'flg_disabled') == 1)) {
    	   $params{store_obj}->{previous_enforced_rule}  = '';    	    	    	    	    		 
    	   $params{store_obj}->{previous_enforced_rule_md5}  = '';    	    	    	    	    		 
    	}    	    	    	    	    		        
    }
}

#----------------------------------------------------------------------

sub update_staging {
    my ($class, %params) = @_;
    my $db_rec = $params{db_rec};
    my $store_obj = $params{store_obj};
    if ( (defined $store_obj->{flg_quarantined} && !$store_obj->{flg_quarantined}) ||
         ($store_obj->{flg_disabled})) {
        $store_obj->{previous_enforced_rule} = $db_rec->{previous_enforced_rule} = '';
        $store_obj->{previous_enforced_rule_md5} = $db_rec->{previous_enforced_rule_md5} = '';
    }

    return $class->SUPER::update_staging(%params);
}

#----------------------------------------------------------------------
# override remove_non_inheritance_fields from Base in order to store flg_quarantined in this case

sub remove_non_inheritance_fields {
    my ($self) = @_;
    my $flg_quarantined = $self->{flg_quarantined};
    $self->SUPER::remove_non_inheritance_fields();
    $self->{flg_quarantined} = $flg_quarantined;
}

#----------------------------------------------------------------------

sub is_in_policy_excluded {
    my($self, %params) = @_;
    my $call_flags = $params{call_flags};
    my $direction = $params{direction};
    my $rest_selected_fields = $params{rest_selected_fields} || {};

    if (! $direction || $direction eq 'from_rest') {
        # On calls from from_rest, consider inPolicy included in REST (for now)
        return 0;
    }
    if ($call_flags && $call_flags->{rest_selected_fields} && $call_flags->{rest_selected_fields}->{inPolicy}) {
        # inPolicy was explicitly specified in $select (or $filter), and must be included in search result
        return 0;
    }

    return 1;
}

#----------------------------------------------------------------------

sub catchup_search_filters{
    my ($class) = @_;
    return [ F5::ASMConfig::SearchFilter->new(field_name => 'is_in_policy',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                              field_value => 1,
                                              )
           ];
}

#----------------------------------------------------------------------

sub export_declarative {
    my ($class, %params) = @_;
    my $dbh = $params{dbh};
    my $policy = $params{policy};

    my $declarative_fields = $class->declarative_fields;
    my $signatures;
    if ($params{minimal_export}) {
        $signatures = {map { $_->{sig_id} => $_ } @{$class->extract_diff_entities(%params)}};
    }
    else {
        my $negsig_atributes = F5::ASMConfig::Entity::PolicyNegsigAttributes->new(%params,
                                                                                  get_criteria => { policy_id => $policy->{id} });
        my $db_fields = join(',', keys %$declarative_fields);
        my $db_table = $class->db_table;
        my $where_clause = 'flg_disabled = 1';
        if ($negsig_atributes->{flg_quarantine_signatures}) {
            $where_clause = $where_clause . ' OR flg_quarantined = 0';
        }
        $signatures = $dbh->selectall_hashref(<<EndQuery, 'sig_id', undef, $policy->{id});
SELECT $db_fields
  FROM $db_table
 WHERE policy_id = ? AND ($where_clause)
EndQuery
    }

    my @sig_ids = keys %$signatures;
    my @search_filters = (
                          F5::ASMConfig::SearchFilter->new(field_name => 'sig_id',
                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                           field_value_list => \@sig_ids),
                          F5::ASMConfig::SearchFilter->new(field_name => 'sig_tag',
                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                                                           field_value => ''),
                         );

    my $sig_declarative_fields = F5::ASMConfig::Entity::Signature->declarative_fields;
    my @tagged_sigs = F5::ASMConfig::Entity::Signature->retrieve_all(%params,
                                                                     selected_fields => [ keys %$sig_declarative_fields ],
                                                                     search_filters => \@search_filters,
                                                                    );
    my %tagged_sigs_map = map { $_->{sig_id} => $_ } @tagged_sigs;

    my @declarative_sigs;
    for my $sig_id (keys %$signatures) {
        my $policy_signature = $class->new(%params, object => $signatures->{$sig_id});
        my $declarative_sig = {};
        for my $field (keys %$declarative_fields) {
            my $value;
            if ($field eq 'sig_id') {
                my $signature_object = $tagged_sigs_map{$sig_id} || { sig_id => $sig_id };
                my $signature = F5::ASMConfig::Entity::Signature->new(dbh => $dbh,
                                                                      object => $signature_object)->to_declarative(%params);
                $declarative_sig = {%$declarative_sig, %$signature};
            }
            elsif (defined $policy_signature->{$field}) {
                my $rv = $policy_signature->map_declarative_field_value(field_name => $field);
                $declarative_sig->{$declarative_fields->{$field}->{rest_field_name}} = $rv;
            }
        }
        push (@declarative_sigs, $declarative_sig);
    }
    return { items => \@declarative_sigs };
}

#----------------------------------------------------------------------

1;
