package F5::ASMConfig::Entity::PolicySignatureRequirement;

use strict;
use warnings;

use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::PolicySignatureRequirement/;
use F5::DateTimeUtils;

#----------------------------------------------------------------------

sub before_store {
    my ($self, %params) = @_;

    my $store_mode   = $params{store_mode};
    my $store_obj    = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $dbh          = $params{dbh} || $self->{dbh};

    my @fields = qw(
                    min_revision_datetime
                    max_revision_datetime
                    sig_tag
                );
    my %field_display_name;
    my %field_values;

    for my $field (@fields) {
        $field_values{$field} = $self->get_new_value(
                                                        store_obj    => $store_obj,
                                                        existing_obj => $existing_obj,
                                                        field_name   => $field,
                                                    );
        $field_display_name{$field} = $self->get_field_display_name_string(%params, field_name => $field);
    }

    my $tag_display = $field_values{sig_tag} ? "tag '$field_values{sig_tag}'" : 'F5 Attack Signatures';
    my $sig_tags = F5::ASMConfig::Cache::get_cached_sig_tags(dbh => $dbh);

    if ($field_values{sig_tag} && !$sig_tags->{$field_values{sig_tag}}) {
        my $error_message = "'$field_values{sig_tag}' tag does not exist.";
        die {
            error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
            display_message => $error_message,
            error_message   => $error_message,
            error_level     => "info",
        };
    }

    if ($self->has_value_changed(%params, check_fields => 'min_revision_datetime')
        || $self->has_value_changed(%params, check_fields => 'max_revision_datetime'))
    {
        if ($field_values{min_revision_datetime} && $field_values{max_revision_datetime}) {
            if ($field_values{max_revision_datetime} < $field_values{min_revision_datetime}) {
                my $min_revision_datetime = F5::DateTimeUtils::convert_epoch_to_rfc3339($field_values{min_revision_datetime});
                my $max_revision_datetime = F5::DateTimeUtils::convert_epoch_to_rfc3339($field_values{max_revision_datetime});

                my $error_message = $field_display_name{min_revision_datetime}
                                    . "[$min_revision_datetime] for $tag_display should be same or earlier than "
                                    . $field_display_name{max_revision_datetime}
                                    . "[$max_revision_datetime]";
                die {
                    error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                    display_message => $error_message,
                    error_message   => $error_message,
                    error_level     => "info",
                };
            }
        }

        my $tag_timestamp = $dbh->selectrow_array(<<EndQuery, undef, $field_values{sig_tag});
SELECT sigfile_timestamp
  FROM PLC.SIGFILE_STATUS WHERE sig_tag = ?
  ORDER BY load_datetime DESC LIMIT 1;
EndQuery

        my $tag_timestamp_epoch = F5::DateTimeUtils::convert_datetime_to_epoch($tag_timestamp);

        if ($self->has_value_changed(%params, check_fields => 'min_revision_datetime')) {
            if (($field_values{min_revision_datetime} && $tag_timestamp)
                && ($field_values{min_revision_datetime} > $tag_timestamp_epoch))
            {
                my $min_revision_datetime = F5::DateTimeUtils::convert_epoch_to_rfc3339($field_values{min_revision_datetime});
                my $error_message = $field_display_name{min_revision_datetime}
                                    . "[$min_revision_datetime] for $tag_display should be same or earlier than '$tag_timestamp'";

                die {
                    error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                    display_message => $error_message,
                    error_message   => $error_message,
                    error_level     => "info",
                };
            }
        }
        if ($self->has_value_changed(%params, check_fields => 'max_revision_datetime')) {
            if (($field_values{max_revision_datetime} && $tag_timestamp)
                && ($field_values{max_revision_datetime} < $tag_timestamp_epoch))
            {
                my $max_revision_datetime = F5::DateTimeUtils::convert_epoch_to_rfc3339($field_values{max_revision_datetime});

                my $error_message = $field_display_name{max_revision_datetime}
                                    . "[$max_revision_datetime] for $tag_display should be same or later than '$tag_timestamp'";

                die {
                    error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                    display_message => $error_message,
                    error_message   => $error_message,
                    error_level     => "info",
                };
            }
        }
    }
}

1;

