package F5::ASMConfig::Entity::RequestLog;

use strict;
use warnings;

use Bit::Vector;

use File::Basename;

use F5::ObjUtils;
use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use F5::Violation::Utils;
use F5::DateTimeUtils;
use F5::IpAddr;
use F5::RequestLog;
use F5::BatchInsert;

use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::RequestLog/;
use constant ASMLOGD_CFG_FILE  =>  '/etc/ts/tools/asmlogd.cfg';
use URI::Escape;

use constant TEMP_TAG_FILTER_NAME => 'temp_tag_filter';

my $attack_types_cache        = [];
my $visible_reputations_cache = undef;

my %complex_entity_filters = ( 'F5::ASMConfig::Entity::Signature' => {full_field_prefix => 'violations/signature/',
                                                                      data_search_key => ['sig_id_orig'],},
                               'F5::ASMConfig::Entity::ThreatCampaign' => {full_field_prefix => 'violations/threatCampaign/',
                                                                           data_search_key => ['threat_campaign', 'name'],},);

my %EXISTS_COMPARISON_MAPPING = ( F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EXISTS() => '!= 0',
                                  F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS() => '= 0', );


our $REQUEST_LOG_COMPLEX_SEARCH_CANDIDATE_LIMIT = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'request_log', 'complex_search_candidate_limit') || 100_000;
my $protobuf_search_timeout = F5::Cfg::cfg_get_config_item(ASMLOGD_CFG_FILE, 'request_log', 'protobuf_search_timeout') || 50;

#----------------------------------------------------------------------

sub additional_where_clauses {
    my ($class, %params) = @_;
    my @additional_where_clauses;
    if ($params{hide_deleted} || ($params{call_flags} && $params{call_flags}->{hide_deleted})) {
        my $dbh = $params{dbh};
        my ($deleted_request_log_id) = $dbh->selectrow_array(<<EndQuery);
SELECT request_log_id FROM PRX.REQUEST_LOG_CLEARED
EndQuery

        if ($deleted_request_log_id) {
            push(@additional_where_clauses, "id > $deleted_request_log_id");
        }

        if (F5::DbUtils::exists_in_table(dbh => $dbh, 
                                         table_name => 'PRX.REQUEST_LOG_PROPERTIES',
                                         where_fields => { flg_is_deleted => 1 },)) {

            push (@additional_where_clauses, 
                  "(id NOT IN (SELECT log_delete_ids.request_log_id FROM PRX.REQUEST_LOG_PROPERTIES log_delete_ids WHERE log_delete_ids.flg_is_deleted = 1))");
        }
    }
    return @additional_where_clauses;
}


#----------------------------------------------------------------------

sub get_retrieve_join_tables {
    my($self, %params) = @_;
    my $search_filters = $params{search_filters} || [];
    my $selected_fields_hash = $params{selected_fields_hash} || {};
    my $fields = $self->fields;

    my ($tag_filter, $tag_filter_ids) = $self->tag_filter_data(%params);

    # Only add join table if absolutely needed (subclassed here to prevent potential impact on other entities)

    my $extern_field_is_used = 0;

    if (! $params{only_if_needed_for_count}) {
        if (! %$selected_fields_hash ||
            grep { $fields->{$_} && $fields->{$_}->{extern_db_table} } keys %$selected_fields_hash) {

            $extern_field_is_used = 1;
        }
    }

    $extern_field_is_used ||= grep { $fields->{$_->{field_name}} && $fields->{$_->{field_name}}->{extern_db_table} } @$search_filters;

    my $join_tables = '';
    if ($extern_field_is_used) {
        $join_tables = $self->SUPER::get_retrieve_join_tables(%params);
    }

    if ($tag_filter) {
        my $join_table_name = ', PRX.REQUEST_LOG_TAGS as rlta';
        if (ref($join_tables)) {
            $join_tables->{query} = $join_tables->{query}. $join_table_name;
        } else {
            $join_tables .= $join_table_name;
        }
    }

    return $join_tables;
}

#----------------------------------------------------------------------

sub rest_filters_to_search_filters {
    my($self, %params) = @_;

    # Strip out violation mask fields and handle here
    my $all_violation_filters = delete $params{rest_filters_by_field}->{violations};

    # Strip out signatures cves filter and handle here
    my $cve_filters = delete $params{rest_filters_by_field}->{signaturesCves};

    my $search_filters = $self->SUPER::rest_filters_to_search_filters(%params);

    # Add filter for violation
    if ($all_violation_filters) {
        my $complex_filter_field = 'viol_set_enforce_or_viol_set_staging';
        if (my($staging_filter) = grep { $_->{full_field} eq 'violations/enforcementState/isInStaging' } @$all_violation_filters) {
            $complex_filter_field = ($staging_filter->{value} eq 'true' ? 'viol_set_staging' : 'viol_set_enforce');
        }

        push(@$search_filters,
             $self->convert_sub_rest_filters_to_search_filters(%params,
                                                               entity_type => 'F5::ASMConfig::Entity::Violation',
                                                               rest_filters => $all_violation_filters,
                                                               complex_filter_field => $complex_filter_field,
                                                               full_field_prefix => 'violations/violation/'));

        for my $complex_filter_entity_type (keys %complex_entity_filters) {
            push(@$search_filters,
                 $self->convert_sub_rest_filters_to_search_filters(%params,
                                                                   entity_type => $complex_filter_entity_type,
                                                                   rest_filters => $all_violation_filters,
                                                                   full_field_prefix => $complex_entity_filters{$complex_filter_entity_type}->{full_field_prefix}));
        }
    }

    # Add filters for signatures cve
    for my $cve_filter (@$cve_filters) {
        my $cmp = $cve_filter->{cmp};
        my $cve_value = ($cmp eq 'eq' || $cmp eq 'ne') ? '%' . $cve_filter->{value} . '%' : $cve_filter->{value};
        if ($cve_filter->{cmp} eq 'eq' || $cve_filter->{cmp} eq 'like') {
            push (@$search_filters, F5::ASMConfig::SearchFilter->new(field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE(),
                                                                    field_value => $cve_value,
                                                                    field_name => 'sig_cves',));
        }
        elsif ($cve_filter->{cmp} eq 'ne' || $cve_filter->{cmp} eq 'not like') {
            push (@$search_filters, F5::ASMConfig::SearchFilter->new(field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE(),
                                                                    field_value => $cve_value,
                                                                    field_name => 'sig_cves',));

        }
    }

    for my $search_filter (@$search_filters) {
        if ($search_filter->{field_name} eq 'rest_uuid' &&
            $search_filter->{field_comparison} eq F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE() &&
            $search_filter->{field_value} =~ /^\%\d{4}$/) {

            $search_filter->{field_name} = 'support_id_suffix';
            $search_filter->{field_value} =~ s/^\%//;
        }
        elsif ($search_filter->{field_name} eq 'enforcement_state_attack_types') {

            $search_filter->{field_name} = 'attack_types_mask';
            my($obj) = F5::ASMConfig::Entity::AttackType->retrieve_all(dbh => $params{dbh},
                                                                       get_criteria => $search_filter->{field_value});
            # bitval for attack_type_masks is offset by 1 from actual attack_type_id
            $search_filter->{field_value} = $obj->{attack_type_id} - 1;
        }
        elsif ($search_filter->{field_name} eq 'client_ip_intelligence') {
            $search_filter->{field_name} = 'iprep_threat';
            my $category_icontrol_name = uc $search_filter->{field_value};
            $category_icontrol_name =~ s/-/_/g;
            my ($obj) = F5::ASMConfig::Entity::IpReputationCategoryAttributes->retrieve_all(dbh => $params{dbh},
                                                                       get_criteria => {category_icontrol_name => $category_icontrol_name});
            $search_filter->{field_value} = $obj->{category_id} - 1;
        }
    }

    return $search_filters;
}

#----------------------------------------------------------------------

sub convert_sub_rest_filters_to_search_filters {
    my($self, %params) = @_;
    my $entity_type = $params{entity_type};
    my $rest_filters = $params{rest_filters};
    my $complex_filter_field = $params{complex_filter_field} || $entity_type->rest_unique_id_field;
    my $full_field_prefix = $params{full_field_prefix};

    my @matching_filters = grep { $_->{full_field} =~ /^$full_field_prefix/ } @$rest_filters;

    if (@matching_filters) {
        for my $rest_filter (@matching_filters) {
            $rest_filter->{full_field} =~ s|^$full_field_prefix||;
            ($rest_filter->{rest_field}, $rest_filter->{nested_fields}) = split(m|/|, $rest_filter->{full_field}, 2)
        }
        my $search_filters = $entity_type->rest_filters_to_search_filters(%params,
                                                                          rest_filters_by_field => { map { $_->{rest_field} => [ $_ ] } @matching_filters });

        my @entities = $entity_type->retrieve_all(dbh => $params{dbh},
                                                  search_filters => $search_filters);

        if (@entities) {
            my $search_filter = F5::ASMConfig::SearchFilter->new(field_name => $complex_filter_field,
                                                    field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                    field_value_list => [ map { $_->get_rest_unique_id } @entities ]);
            $search_filter->{entity_type} = $entity_type;
            return $search_filter;
        }
        # Generic filter for an empty result
        return F5::ASMConfig::SearchFilter->new(field_name => 'request_log_id',
                                                field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                field_value_list => []);
    }
    return;
}

#----------------------------------------------------------------------

sub check_tag_filter {
    my (%params) = @_;

    my $search_filters = $params{search_filters} || [];

    my $tag_filter_found = 0;
    my $filter_tag_ids = [];
    for my $filter (@$search_filters) {
        if ($filter->{field_name} eq "tags__tag_id" || $filter->{field_name} eq "tags" && !$EXISTS_COMPARISON_MAPPING{$filter->{field_comparison}}) {
            $tag_filter_found = 1;
            $filter_tag_ids = $filter->{field_value_list} || [];
            last;
        }
    }

    return wantarray ? ($tag_filter_found, $filter_tag_ids) : $tag_filter_found;
}

#----------------------------------------------------------------------

sub remove_tag_filter {
    my (%params) = @_;

    my $search_filters = $params{search_filters} || [];

    my $new_filters = [];
    for my $filter (@$search_filters) {
        if ($filter->{field_name} ne "tags__tag_id" && $filter->{field_name} ne "tags") {
            push(@$new_filters, $filter);
        }
    }

    return $new_filters;
}

#----------------------------------------------------------------------

sub retrieve_all {
    my($class, %params) = @_;

    my $dbh = $params{dbh};
    my $search_filters = $params{search_filters} || [];

    my ($tag_filter_found, $tag_ids_ref) = check_tag_filter(%params);
    if ($tag_filter_found) {
        $params{search_filters} = remove_tag_filter(%params);

        push @{$params{search_filters}}, F5::ASMConfig::SearchFilter->new(field_name => TEMP_TAG_FILTER_NAME,
                                                    field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                    field_value_list => $tag_ids_ref);

        $search_filters = $params{search_filters} || [];
        $params{grouping_fields} = ['request_log_id'];
    }

    if (! check_filters_nonstage_violations(%params)) {
        return wantarray ? () : F5::ASMConfig::Entity::Iterator->new(get_next_function => sub {return undef;},
                                                                     get_all_function => sub {return ();},
                                                                     total_record_count => 0, );
    }

    my $bd_request_fields = $class->bd_request_fields();

    # Intercept full protobuf searches and pre-handle
    my @protobuf_searches;
    my @non_protobuf_searches;
    my @complex_search_filters;
    for my $search_filter (@$search_filters) {
        if ($search_filter->{entity_type} && $complex_entity_filters{$search_filter->{entity_type}}) {
            push(@complex_search_filters, $search_filter);
        }
        elsif ($bd_request_fields->{$search_filter->{field_name}}) {
            push(@protobuf_searches, $search_filter);
        }
        else {
            push(@non_protobuf_searches, $search_filter);
        }
    }

    if (@protobuf_searches || @complex_search_filters) {
        $dbh->do(<<EndQuery);
CREATE TEMPORARY TABLE IF NOT EXISTS PRX.REQUEST_LOG_PROTOBUF_SEARCH_RESULTS (
  request_log_id bigint(20) unsigned NOT NULL,
  PRIMARY KEY  (request_log_id));
EndQuery
        # Clear out temporary table in case it lingered from a prior search on the same session
        $dbh->do('DELETE FROM PRX.REQUEST_LOG_PROTOBUF_SEARCH_RESULTS');
        my $search_results_batch_insert = F5::BatchInsert->new(%params);

        @{$params{search_filters}} = @non_protobuf_searches;
        my $prematch_iter = $class->SUPER::retrieve_all(%params,
                                                        page_size => $REQUEST_LOG_COMPLEX_SEARCH_CANDIDATE_LIMIT,
                                                        start_page => undef,
                                                        selected_fields => [],
                                                        return_as_raw_db_hashref => 1,
                                                        skip_total_record_count => 1,
                                                       );

        my $iterations_end_time = time + $protobuf_search_timeout;
        while (my $rec = $prematch_iter->get_next) {
            $rec->{protobuf_bin} = F5::RequestLog->get_protobuf(rec => $rec);
            next unless defined $rec->{protobuf_bin};

            # First, check if all protobuf searches exist in raw buffer, whether positive or negative match.
            # Either way, this will require protobuf deserialization and final check against the correct field.
            # However, if even a single positive match is missed, this record will definitely not be a match,
            # and can be skipped.
            #
            # Matching matrix for raw protobuf check:
            #
            #                  | Found in raw protobuf         | Unfound in raw protobuf
            # -----------------+-------------------------------+--------------------------------------------
            # Positive match   | Need to verify if occurs in   | Will definitely not be found in desired field,
            #                  | desired field, deserialize    | record can be skipped as non-match
            #                  | and check specific field      | (instantly disqualified)
            # -----------------+----------------------------------------------------------------------------
            # Negative match   | Need to verify if occurs in   | Will definitely not be found in field it should
            #                  | field that it should *not* be | *not* be in, record is candidate for match
            #                  | in, deserialize and check     |
            #                  | specific field                |
            # -----------------+----------------------------------------------------------------------------
            # Attack signature | Due to protobuf numeric encoding, deserialize is always needed
            # -----------------+----------------------------------------------------------------------------
            #
            # If deserialize is needed, filter matching will continue in conventional way.

            my $found_in_protobuf_bin = 0;
            my $positive_match_missed = 0;
            if (@complex_search_filters) {
                # Signatures can only be searched from the protobuf, so automatic trigger of protobuf load.
                $found_in_protobuf_bin = 1;
            }

            my $bd_request_selected_fields = {};
            for my $protobuf_search (@protobuf_searches) {
                my $field_name = $protobuf_search->{field_name};
                $bd_request_selected_fields->{$field_name} = 1;

                if ($bd_request_fields->{$field_name}->{filter_value_in_protobuf}) {
                    $found_in_protobuf_bin = 1;
                }
                elsif ($EXISTS_COMPARISON_MAPPING{$protobuf_search->{field_comparison}}) {
                    $found_in_protobuf_bin = 1;
                }
                elsif ($class->is_filter_match(search_filter => $protobuf_search,
                                            field_value => $rec->{protobuf_bin},
                                            match_as_positive_equivalent => 1)) {
                    $found_in_protobuf_bin = 1;
                }
                elsif ($class->is_positive_match(search_filter => $protobuf_search)) {
                    $positive_match_missed = 1;
                    last;
                }
            }

            my $match_candidate = 1;
            if ($found_in_protobuf_bin && ! $positive_match_missed) {
                $class->populate_bd_request_fields(%params,
                                                   rec => $rec,
                                                   bd_request_selected_fields => $bd_request_selected_fields,
                                                   additional_bd_request_fields => [ qw(bd_data) ],
                                                   selected_fields_hash => undef,
                                                  );

                for my $protobuf_search (@protobuf_searches) {
                    if (! $class->is_filter_match(search_filter => $protobuf_search,
                                                  field_value => $rec->{$protobuf_search->{field_name}})) {
                        $match_candidate = 0;
                        last;
                    }
                }

                if ($match_candidate) {
                    for my $complex_filter (@complex_search_filters) {
                        my $filter_entity_type = $complex_filter->{entity_type};
                        if ($complex_entity_filters{$filter_entity_type}) {
                            # Check for search values throughout the entire unpacked protobuf (cheaper than
                            # populating violations)
                            my @request_values = F5::ObjUtils::get_deeply(ref => $rec->{bd_data},
                                                                          search_path => $complex_entity_filters{$filter_entity_type}->{data_search_key});
                            my %request_values = map { ($_ => 1) } @request_values;

                            if (! grep { $request_values{$_} } @{$complex_filter->{field_value_list}}) {
                                $match_candidate = 0;
                                last;
                            }
                        }
                    }
                }
            }

            if ($match_candidate && ! $positive_match_missed) {
                $search_results_batch_insert->insert_row(table_name => 'PRX.REQUEST_LOG_PROTOBUF_SEARCH_RESULTS',
                                                         do_insert_ignore => 1,
                                                         rec => $rec);
            }

            if (time > $iterations_end_time) {
                last;
            }
        }
        $search_results_batch_insert->flush;

        push(@{$params{search_filters}},
             F5::ASMConfig::SearchFilter->new(field_name => 'protobuf_search_results',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ()));

    }

    return $class->SUPER::retrieve_all(%params, tolerate_total_record_count_failure => 1);
}

#----------------------------------------------------------------------

sub check_filters_nonstage_violations {
    my (%params) = @_;

    my $search_filters = $params{search_filters} || [];
    for my $filter (@$search_filters) {
        if (grep {$filter->{field_name} eq $_} qw/viol_set_enforce
                                                  viol_set_staging
                                                  viol_set_enforce_or_viol_set_staging/) {
            my $filter_values = $filter->{field_value_list};
            for my $value (@$filter_values) {
                if (grep {$value == $_} (
                                        F5::Violation::VIOL_PARAMETER(),
                                        F5::Violation::VIOL_FILETYPE(),
                                        F5::Violation::VIOL_URL()
                                        )) {
                    if ($filter->{field_name} eq 'viol_set_staging') {
                        return undef;
                    }
                    elsif ($filter->{field_name} eq 'viol_set_enforce_or_viol_set_staging') {
                        $filter->{field_name} = 'viol_set_enforce';
                    }
                }
            }
        }
    }

    return 1;
}

#----------------------------------------------------------------------

my %bitval_fields = map { ($_ => 1) } qw(viol_set_enforce
                                         viol_set_staging
                                         viol_set_enforce_or_viol_set_staging
                                         attack_types_mask
                                         iprep_threat
                                       );

sub get_complex_filter {
    my($self, %params) = @_;

    my $filter_field = $params{filter_field};
    my $field_value = $params{field_value};
    my $field_value_list = $params{field_value_list};
    if (defined $field_value && ! defined $field_value_list) {
        $field_value_list = [ $field_value ];
    }

    if ($bitval_fields{$filter_field}) {

        my @bitval_clauses;
        my @filter_fields = ($filter_field);
        if ($filter_field eq 'viol_set_enforce_or_viol_set_staging') {
            @filter_fields = ('viol_set_enforce', 'viol_set_staging');
        }

        for my $bitval_field (@filter_fields) {
            if (my $exists_filter = $EXISTS_COMPARISON_MAPPING{$params{field_comparison}}) {
                push(@bitval_clauses, "main_table.$bitval_field $exists_filter");
            }

            for my $single_field_value (@$field_value_list) {
                # UDF bitval does not support string conversion and thus cannot be called with a placeholder
                push(@bitval_clauses,
                     "bitval(main_table.$bitval_field, $single_field_value) = 1");
            }
        }

        my $bitval_clauses = '(' . join(" OR ", @bitval_clauses) . ')';
        return { where_clause => $bitval_clauses };
    }
    elsif ($filter_field eq 'protobuf_search_results') {
        return { where_clause => 'main_table.id IN (SELECT request_log_id FROM PRX.REQUEST_LOG_PROTOBUF_SEARCH_RESULTS)' };
    }
    elsif ($filter_field eq TEMP_TAG_FILTER_NAME) {
        return {
            where_clause => ' main_table.id = rlta.request_log_id AND rlta.tag_id in ('. join(', ', map {'?'} @$field_value_list). ')',
            where_values => $field_value_list,
        };
    }
    return;
}

#----------------------------------------------------------------------

sub tag_filter_data {
    my($self, %params) = @_;

    my $search_filters = $params{search_filters} || [];

    my $tag_filter = 0;
    my $tag_filter_ids = [];
    for my $filter (@$search_filters) {
        if ($filter->{field_name} eq TEMP_TAG_FILTER_NAME) {
            $tag_filter = 1;
            $tag_filter_ids = $filter->{field_value_list};
            last;
        }
    }

    return ($tag_filter, $tag_filter_ids);
}

#----------------------------------------------------------------------

sub bd_request_fields {
  my ($class) = @_;
  my $all_fields = $class->all_fields;
  my %bd_request_fields = map { $_ => $all_fields->{$_} } grep { $all_fields->{$_}->{bd_request_field_name} } keys %{$all_fields};
  return $class->install_accessor('bd_request_fields', \%bd_request_fields);
}

#----------------------------------------------------------------------

sub validate_accept_status {
    my ($self, %params) = @_;

    my $display_message;

    if ($params{store_obj}) {
        my $new_accept_status = $self->get_new_value(%params);

        if ($self->{accept_status} == F5::ASMConfig::AcceptStatus::ACCEPT_STATUS_IN_PROGRESS() || 
            $self->{accept_status} == F5::ASMConfig::AcceptStatus::ACCEPT_STATUS_START() ||
            $new_accept_status != F5::ASMConfig::AcceptStatus::ACCEPT_STATUS_START()) {
            $display_message = "The Accept Status may only be manually changed from a finished state ('none','completed','failed') to 'start'";
        } 
        elsif (!$self->{request_policy}->{is_learning_enabled}) {
            $display_message = "Learning must be enabled on policy '$self->{request_policy}->{policy_name}' to accept requests";
        }
    }
    
    if ($display_message) {
        return { completed_successfully => 0,
                 display_message => $display_message};
    }
    return {completed_successfully => 1};
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
    my ( $class_or_entity, %params ) = @_;

    $class_or_entity->SUPER::map_fields_to_entity(%params);

    my $dbh    = $params{dbh};
    my $logger = $params{logger};
    my $rec    = $params{rec} || $class_or_entity;
    my $selected_fields_hash = $params{selected_fields_hash};
    my $call_flags = $params{call_flags};

    my $bd_request_fields = $class_or_entity->bd_request_fields();
    my $bd_request_selected_fields; 
    if ($selected_fields_hash) {
        for my $selected_field (keys %$selected_fields_hash) {
            if (my $field_def = $bd_request_fields->{$selected_field}) {
                $bd_request_selected_fields->{$field_def->{bd_request_field_name}} = 1;
            }
        }
    }

    if (!$rec->{protobuf_bin} && $rec->{request_log_id} && (!$selected_fields_hash || $bd_request_selected_fields)) {
        $rec->{protobuf_bin} = F5::RequestLog->get_protobuf(rec => $rec);
    }

    if ( $rec->{protobuf_bin} ) {
        $class_or_entity->populate_bd_request_fields(%params,
                                                     rec => $rec,
                                                     expand_fields => $call_flags->{expand_fields},
                                                     bd_request_selected_fields => $bd_request_selected_fields);
    }

    if (!$selected_fields_hash || $selected_fields_hash->{enforcement_state_attack_types}) {
        my @attackTypes = $class_or_entity->get_attack_types( dbh => $dbh, attack_types_mask => $rec->{attack_types_mask} );
        $rec->{enforcement_state_attack_types} = \@attackTypes;
    }

}

#----------------------------------------------------------------------

sub populate_bd_request_fields {
    my($class_or_entity, %params) = @_;
    my $rec = $params{rec};
    my $selected_fields_hash = $params{selected_fields_hash};
    my $bd_request_fields = $class_or_entity->bd_request_fields();
    my @additional_bd_request_fields;
    if ($params{additional_bd_request_fields}) {
        @additional_bd_request_fields = @{$params{additional_bd_request_fields}};
    }

    require F5::BdRequest;
    my $bd_request = F5::BdRequest->new( bd => $rec->{protobuf_bin},
                                         bd_request_selected_fields => $params{bd_request_selected_fields},
                                         expand_fields => $params{expand_fields},
                                         dbh => $params{dbh}, 
                                         logger => $params{logger},
                                         request_policy => $rec->{request_policy}, );
    for my $bd_request_field (keys %$bd_request_fields) {
        if (!$selected_fields_hash || $selected_fields_hash->{$bd_request_field}) {
            my $field_def = $bd_request_fields->{$bd_request_field};
            my $field_value = $bd_request->{$field_def->{bd_request_field_name}};
            if (my $func = $field_def->{format_field_func}) {
                $field_value = $class_or_entity->$func(%params,
                                                       field_value => $field_value);
            }
            $rec->{$bd_request_field} = $field_value;
        }
    }
    for my $bd_request_field (@additional_bd_request_fields) {
        $rec->{$bd_request_field} = $bd_request->{$bd_request_field};
    }
}

#----------------------------------------------------------------------

sub get_attack_types {
    my ( $self, %params ) = @_;

    my $dbh = $params{dbh};
    my $attack_types_mask = $params{attack_types_mask} || 0;

    my @attackTypes = ();
    my $vec = Bit::Vector->new_Dec( 64, $attack_types_mask );
    for my $pos ( $vec->Index_List_Read ) {
        my $attack_type_id = $pos + 1;
        my $attackType;
        if ( $attack_types_cache->[$attack_type_id] ) {
            $attackType = $attack_types_cache->[$attack_type_id];
        }
        else {
            $attackType = F5::ASMConfig::Entity::AttackType->new( dbh => $dbh, get_criteria => { attack_type_id => $attack_type_id } );
            $attack_types_cache->[$attack_type_id] = $attackType;
        }
        push( @attackTypes, $attackType );

    }

    return @attackTypes;
}

#----------------------------------------------------------------------

sub get_visible_reputations_cache {
    my ($self, %params) = @_;
    if ($visible_reputations_cache) {
        return $visible_reputations_cache;
    }
    my $dbh = $params{dbh} || $self->{dbh};
    $visible_reputations_cache = {};
    for my $reputation (
        F5::ASMConfig::Entity::IpReputationCategoryAttributes->retrieve_all( dbh => $dbh, get_criteria => { flg_is_visible => 1 } ) )
    {
        my $rep_name = lc $reputation->{category_icontrol_name};
        $rep_name =~ s/\_/\-/g;
        $visible_reputations_cache->{ $reputation->{category_id} } = $rep_name;
    }
    return $visible_reputations_cache;
}

#----------------------------------------------------------------------

sub convert_client_ip_intelligence {
    my ( $self, %params ) = @_;

    my $bit_mask  = $params{field_value};
    my $direction = $params{direction};
    my $dbh       = $params{dbh};

    return unless defined($bit_mask);
    my @intelligenceList = ();
    if ( $direction eq 'to_entity' ) {
        $visible_reputations_cache = get_visible_reputations_cache($self, %params);
        my $vec = Bit::Vector->new_Dec( 64, $bit_mask );
        for my $bit_offset ( $vec->Index_List_Read ) {
            my $pos = $bit_offset + 1;
            if ( $visible_reputations_cache->{$pos} ) {
                push( @intelligenceList, $visible_reputations_cache->{$pos} );
            }
        }
    }

    return \@intelligenceList;
}
#----------------------------------------------------------------------

sub uri_decode {
    my ( $self, %params ) = @_;

    my $field_value  = $params{field_value};

    return unless defined($field_value);
    my $converted_value = URI::Escape::uri_unescape($field_value);
    while ($converted_value ne $field_value) {
        $field_value = $converted_value;
        $converted_value = URI::Escape::uri_unescape($field_value);
    }

    return $converted_value;
}

#----------------------------------------------------------------------

sub map_suggestion_keys {
    my ( $self, %params ) = @_;

    my $field_value = $params{field_value};
    my $direction = $params{direction};
    my $dbh       = $params{dbh};
    my $policy    = $params{policy};

    return unless defined($field_value);
    my @suggestions;
    if ( $direction eq 'to_entity' ) {
        my @keys = split(',', $field_value);
        @suggestions = map { F5::ASMConfig::Suggestion->new(suggestion_key => $_) } @keys;
    }
    return \@suggestions;
}

#----------------------------------------------------------------------

sub map_signatures_cves {
    my ( $self, %params ) = @_;

    my $field_value = $params{field_value};
    my $direction = $params{direction};
    my $dbh       = $params{dbh};
    my $policy    = $params{policy};

    return unless defined($field_value);
    my @cves;
    if ( $direction eq 'to_entity' ) {
        my @values = split(',', $field_value);
        @cves = map { { name => $_ } } @values;
    }
    return \@cves;
}

#----------------------------------------------------------------------

sub get_rest_uuid {
    my ($self, %params) = @_;
    my $object = $params{object} || $self;
    my $rest_uuid = $object->{rest_uuid} || $object->{support_id} || $self->{rest_uuid} || $self->{support_id};
    return $rest_uuid;
}

#----------------------------------------------------------------------

sub self_link {
    my ($self, %params) = @_;
    my $link = $params{base_link} || '';
    $link .= '/' . $self->element_rest_path_name() . "/" . $self->get_rest_uuid;
    $link = $self->add_ver_to_self_link(%params,
                                        'link' => $link);
    return $link;
}

#----------------------------------------------------------------------

sub policy {
    my ($self) = @_;
    return F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $self->{request_policy});
}

#----------------------------------------------------------------------

sub policy_id_field {
    return '';
}

#----------------------------------------------------------------------

sub update_many_as_iterative {
    return 1;
}

#----------------------------------------------------------------------

1;

