package F5::ASMConfig::Entity::Section;

use strict;
use warnings;

use base qw(F5::ASMConfig::Entity::Base F5::ASMConfig::Section);

#----------------------------------------------------------------------

sub get_child_inheritance_status {
    my ($class_or_self, %params) = @_;
    my $parent_inheritance_status = $params{parent_inheritance_status} || $class_or_self->get_new_value(%params, field_name => 'parent_inheritance_status');
    my $child_inheritance_status = $params{child_inheritance_status} || $params{existing_obj}->{child_inheritance_status};
    my $call_flags = $params{call_flags};

    if ($parent_inheritance_status == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY()) {
        $child_inheritance_status = F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_MANDATORY();
    }
    elsif ($parent_inheritance_status == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NONE()) {
        $child_inheritance_status = F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_NA();
    }
    elsif ($parent_inheritance_status == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL()) {
        if ( ($call_flags->{is_in_add_policy} || $call_flags->{is_in_import_policy}) && !$call_flags->{retain_inheritance_settings} ) {
             $child_inheritance_status = F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_ACCEPTED();
        }
        elsif ( $child_inheritance_status == F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_MANDATORY() ) {
            #If switching from Mandatory then new child status is Accepted. Otherwise Child policy starts as Declined.
            $child_inheritance_status = F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_ACCEPTED();
        }
        elsif ( $child_inheritance_status == F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_NA() ) {
            $child_inheritance_status = F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_DECLINED();
        }
    }

    return $child_inheritance_status;
}

sub before_store {
    my($self, %params) = @_;
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags   = $params{call_flags};

    $self->SUPER::before_store(%params);

    #Can only set parent_inheritance_status if policy is Parent type, or change came from parent
    my $policy = $params{policy} || $self->{policy};
    if ( $policy->{policy_type} != F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() ) {
        if ( $self->has_value_changed(%params, check_fields => 'parent_inheritance_status') ) {
            if ($params{validate_value} && !$params{call_flags}->{inherited_call}) {
                my $field_display_name = $self->get_field_display_name_string(%params, field_name => 'parent_inheritance_status');
                my $error_message = "$field_display_name can only be modified in a Parent Policy.";
                die { error_message => $error_message,
                      display_message => $error_message,
                      error_level     => 'info',
                      error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                    };
                }
            $store_obj->{child_inheritance_status} = $self->get_child_inheritance_status(%params);
        }
        elsif ( $self->has_value_changed(%params, check_fields => 'child_inheritance_status')
             && $params{validate_value} && !$params{call_flags}->{inherited_call} ) {
            my $child_inheritance_status = $self->get_new_value(%params, field_name => 'child_inheritance_status');
            my $parent_inheritance_status = $self->get_new_value(%params, field_name => 'parent_inheritance_status');
            if ( !( $parent_inheritance_status == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL()
                    && ( $child_inheritance_status == F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_ACCEPTED()
                      || $child_inheritance_status == F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_DECLINED() ))) {
                my $field_display_name_child = $self->get_field_display_name_string(%params, field_name => 'child_inheritance_status');
                my $field_display_name_parent = $self->get_field_display_name_string(%params, field_name => 'parent_inheritance_status');
                my $error_message = "$field_display_name_child is not compatible with $field_display_name_parent";
                die { error_message => $error_message,
                      display_message => $error_message,
                      error_level     => 'info',
                      error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                    };
            }
        }
    }            
    else {
        if ( $self->has_value_changed(%params, check_fields => 'parent_inheritance_status') ) {
            my $parent_inheritance_status = $self->get_new_value(%params, field_name => 'parent_inheritance_status');
            if ( $parent_inheritance_status == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NA()
                 && $params{validate_value} && !$params{call_flags}->{inherited_call} ) {
                my $field_display_name = $self->get_field_display_name_string(%params, field_name => 'parent_inheritance_status');
                my $error_message = "Cannot set $field_display_name to Not Available";
                die { error_message => $error_message,
                      display_message => $error_message,
                      error_level     => 'info',
                      error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                    };
            }
        }
    }

    if ($params{validate_value} 
       && $self->has_value_changed(%params, check_fields => [qw(child_inheritance_status parent_inheritance_status)])
        && $existing_obj->{parent_inheritance_status} ==  F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NA()) {
        my $error_message = "Inheritance is not available for this section";
        die { error_message => $error_message,
              display_message => $error_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
           };
    }

}

#----------------------------------------------------------------------

sub after_nested_struct_load {
    my($self, %params) = @_;
    my $dbh = $params{dbh} || $self->{dbh};
    my $extra_fields_hash = $params{extra_fields_hash};

    if ($extra_fields_hash->{child_sections_with_inheritance_notification}) {
        my $policy = $params{policy} || $self->{policy};
        return unless $policy;
        if (!$policy->{policy_type}) {
            $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh, policy => $policy);
        }

        if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
            my @child_sections = F5::ASMConfig::Entity::Section->retrieve_all(
                                     dbh => $dbh,
                                     get_criteria => { section_type => $self->{section_type} },
                                     full_policy_populate => 1,
                                     search_filters => [ F5::ASMConfig::SearchFilter->new(field_name => 'policy_id',
                                                                                          field_value_list => $policy->child_policy_ids(),
                                                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),),
                                                         F5::ASMConfig::SearchFilter->new(field_name => 'has_inheritance_notification',
                                                                                          field_value => 1,
                                                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),),
                                                       ],
                                     );

           $self->{child_sections_with_inheritance_notification} = \@child_sections;
        }
    }
}

#----------------------------------------------------------------------

sub is_inherited {
    my ($self) = @_;
    return ($self->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY()
            || ($self->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL()
                && $self->{child_inheritance_status} == F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_ACCEPTED()));
}

#----------------------------------------------------------------------

sub section_entity_filters {
    my ($self, %params) = @_;

    my $is_inherited = $self->is_inherited;
    my @static_entity_types = map { $_->{entity_type} } grep { $_->{section_type} && $_->{section_type} == $self->{section_type} 
                                                               && !$_->{cascaded_inheritance} && !$_->{skip_inheritance} && !$_->{no_inheritance}} values %{$self->full_schema()};
    my @entity_filters = ({
                           entity_types => \@static_entity_types,
                          },
                          {
                           search_filters => [ F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'section_type',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                          field_value => $self->{section_type},}) ],
                           entity_types => ['F5::ASMConfig::Entity::Blocking', 'F5::ASMConfig::Entity::Charset'],
                          },
                          {
                           search_filters => [ F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'section_type',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                          field_value => $self->{section_type},}) ],
                           deassociate_update_field => 'inheritance_acceptability',
                           deassociate_update_value => F5::ASMConfig::InheritanceAcceptability::INHERITANCE_ACCEPTABILITY_ALLOWED(),
                           entity_types => ['F5::ASMConfig::Entity::Suggestion'],
                          },
                         );

    return \@entity_filters;

}

#----------------------------------------------------------------------

1;
