package F5::ASMConfig::Entity::SessionAwarenessDataPoint;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use F5::IpAddr;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::SessionAwarenessDataPoint/;

use F5::SessionDB;
use Math::BigInt;
use Socket;

my %data_point_action = (
                         'track_all' => F5::ASMConfig::SessionAwarenessAction::SESSION_ACTION_TRACK_ALL,
                         'block_illegal' => F5::ASMConfig::SessionAwarenessAction::SESSION_ACTION_BLOCK_ILLEGAL,
                         'block_all' => F5::ASMConfig::SessionAwarenessAction::SESSION_ACTION_BLOCK_ALL,
                        );
my %data_point_action_reverse = reverse %data_point_action;

my %data_point_session_db_value = (
                                   F5::ASMConfig::SessionAwarenessAction::SESSION_ACTION_TRACK_ALL() => 1,
                                   F5::ASMConfig::SessionAwarenessAction::SESSION_ACTION_BLOCK_ILLEGAL() => 2,
                                   F5::ASMConfig::SessionAwarenessAction::SESSION_ACTION_BLOCK_ALL() => 4,
                                  );
my %data_point_scope = (
                        'usr' => F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_USER,
                        'sid' => F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_SESSION,
                        'ip' => F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_IP,
                        'device' => F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_DEVICE,
                       );
my %data_point_scope_reverse = reverse %data_point_scope;
my $VERY_LARGE_TIMEOUT = 60*60*24*365*5; # 5 years!!
my $SESSION_DB_BATCH_SIZE = 150;
my $SESSION_DB_MAX_QUERY_SIZE = 100_000;
my $ACTION_BIN_NAME = 'asm_sessiondp';
my $WINDOW_BIN_NAME = 'asm_sessionsw';
my $MASTER_KEY_TIMEOUT = 0;
my $SESSION_VALUE_MIN_LENGTH = 8;

#----------------------------------------------------------------------

sub validate_data_point_value {
    my($self, %params) = @_;
    my $field_value = $params{field_value};
    my $store_obj = $params{store_obj};

    my $rv = $self->general_validate(%params);
    if (!$rv->{completed_successfully}){
        return $rv;
    }
    if (defined $field_value &&  ref($self)  ){
        my $data_point_scope = $self->get_new_value(%params, field_name => 'data_point_scope');
        if ( $data_point_scope  == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_SESSION() &&
                               length($field_value) < $SESSION_VALUE_MIN_LENGTH) {

               return {
                 completed_successfully => 0,
                 error_message => "The value is too short (minimum length = $SESSION_VALUE_MIN_LENGTH)",
               };
        }
        elsif ( $data_point_scope  == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_IP) {
            my $res = $self->normalize_ip(field_value => $field_value);
            if (!$res->{completed_successfully}){
              return {
                 completed_successfully => 0,
                 error_message => $res->{error_message},
               };
            }

        }
    }

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
    my($class, %params) = @_;
    my $rec = $params{rec};

    $class->SUPER::map_fields_to_entity(%params);
    $rec->{is_persistent} = 1;
}

#----------------------------------------------------------------------

sub get_default_obj {
    my($self,%params) = @_;

    my $retval = $self->SUPER::get_default_obj(%params);
    $retval->{is_persistent} = 0;
    return $retval;
}

#----------------------------------------------------------------------

sub retrieve_all {
    my($class, %params) = @_;

    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $get_criteria = $params{get_criteria} || {};
    my $fields = $class->fields;
    my $find_single_record = $params{find_single_record};

    my @search_filters;
    if ($params{search_filters}) {
        @search_filters = @{$params{search_filters}};
    }

    my @filter_fields;
    my @unique_record_parts = @{$class->unique_record_parts(get_criteria => $get_criteria)};
    if (($find_single_record || scalar @unique_record_parts) &&
        scalar @unique_record_parts == scalar grep { defined $get_criteria->{$_} } @unique_record_parts) {
        # Getting a single record - only filter on unique record parts
        @filter_fields = @unique_record_parts;
    }
    else {
        @filter_fields = grep { defined $get_criteria->{$_} } keys %$fields;
    }

    for my $filter_field (@filter_fields) {
        push(@search_filters,
             { field_name => $filter_field,
               field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
               field_value => $get_criteria->{$filter_field} });
    }

    my $policy = F5::ASMConfig::Cache::get_cached_policy(policy => $params{policy},
                                                         dbh => $dbh,
                                                        );

    my($is_persistent_filter) = grep { $_->{field_name} eq 'is_persistent' &&
                                         ($_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() ||
					 $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ()) &&
                                         $_->{field_value} == 1 } @search_filters;

    if ($is_persistent_filter || $params{for_populate_uuids} || !$policy->{is_active_policy}) {
        # Optimization to retrieve only from database if searching for persistent records OR in case the policy is inactive
        return $class->SUPER::retrieve_all(%params);
    }
    else {

        my @objs = $class->get_from_session_db(%params,
                                               search_filters => \@search_filters,
                                              );

        # Combine persistent records from database
        my %objs_as_hash = map { ($_->unique_record_parts_str => $_) } @objs;
        my $iter = $class->SUPER::retrieve_all(%params);

        while(my $db_obj = $iter->get_next) {
            $objs_as_hash{$db_obj->unique_record_parts_str}->{is_persistent} = 1;
            $objs_as_hash{$db_obj->unique_record_parts_str}->{data_point_id} = $db_obj->{data_point_id};
            $objs_as_hash{$db_obj->unique_record_parts_str}->{create_time} = $db_obj->{create_time};
        }

        $class->apply_search_filters(result_set => \@objs,
                                     %params,
                                     search_filters => \@search_filters,
                                    );

        $class->apply_sort_order(result_set => \@objs,
                                 %params);

        my $total_record_count = scalar @objs;
        $class->apply_paging(result_set => \@objs,
                                 %params);


        if (wantarray) {
            return @objs;
        }
        else {
            my $iterator = F5::ASMConfig::Entity::Iterator->new(
                                                                get_next_function => sub {return shift(@objs);},
                                                                get_all_function => sub {return @objs;},
                                                                total_record_count => $total_record_count,
                                                               );
            return $iterator;
        }
    }
}

#----------------------------------------------------------------------

sub store_update {
    my($self, %params) = @_;

    if ($self->has_value_changed(%params, existing_obj => $self, check_fields => 'is_persistent') &&
        $self->get_new_value(%params, existing_obj => $self, field_name => 'is_persistent')) {
        return $self->SUPER::store_add(%params);
    }
    return $self->SUPER::store_update(%params);
}

#----------------------------------------------------------------------

sub store_add {
    my($self, %params) = @_;
    my $dbh = $self->{dbh};
    my $fields = $self->fields;

    my $create_time = $params{create_time};
    my $in_reload_session_db = $params{in_reload_session_db} || 0;

    $self->{policy} = F5::ASMConfig::Cache::get_cached_policy(policy => ($params{policy} || $self->{policy}),
                                                              dbh => $dbh);

    if ( !$self->{policy}->{is_active_policy} ){
          my $error_message = "Session Awareness Data Point cannot be added to an inactive security policy";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
              };
    }

    if (!defined $self->{data_point_scope}) {
        $self->{data_point_scope} = $fields->{data_point_scope}->{default_value};
    }
    if (!defined $self->{data_point_action}) {
         $self->{data_point_action} = $fields->{data_point_action}->{default_value};
    } 

    if ($self->{data_point_action} == F5::ASMConfig::SessionAwarenessAction::SESSION_ACTION_BLOCK_ALL &&
        ! $params{save_to_session_db_only}) {

        $self->{is_persistent} = 1;
    }

    my $session_db = $params{session_db} || $self->get_new_session_db();
    my $session_db_rec = $self->get_session_db_rec;

    # Get session awareness settings from DCC, i.e. from last set active
    my $session_awareness_settings = $dbh->selectrow_hashref(<<EndQuery, undef, $self->{policy}->{policy_name_crc});
SELECT *
  FROM DCC.ACCOUNT_SESSION_AWARENESS_SETTINGS
 WHERE policy_name_crc = ?
EndQuery

    my $action = $data_point_action_reverse{$self->{data_point_action}};
    my $expiration_period = $session_awareness_settings->{"${action}_period"};

    # when reloading session db (say after a restart)
    #     do not restart the whole expiration_period from scratch
    if ($in_reload_session_db && defined $expiration_period && defined $create_time) {
        $expiration_period = $expiration_period - (time - $create_time);
    }

    if (! $expiration_period) {
        # Infinite duration
        $expiration_period = $VERY_LARGE_TIMEOUT * 2;
    }

    if (! $MASTER_KEY_TIMEOUT) {
        $MASTER_KEY_TIMEOUT = $dbh->selectrow_array(<<EndQuery, undef, 'sa_login_expiration_timeout');
SELECT current_value
  FROM DCC.INTERNAL_CONFIG_ITEMS
 WHERE field_name = ?
EndQuery
    }

    $session_db->add(key => $session_db_rec->{session_db_master_key},
                     %{$session_db_rec->{master_key_additional_values}},
                     expiration => $MASTER_KEY_TIMEOUT,
                     skip_if_exists => 1,
                    );

    $session_db->add(key => $session_db_rec->{session_db_action_key},
                     expiration => $expiration_period,
                     value => $session_db_rec->{data_point_session_db_value},
                     skip_if_exists => 1,
                    );

    if ($self->{is_persistent}) {
        $self->SUPER::store_add(%params);
    }
    else {
       $self->set_rest_uuid(); 
       $self->{create_time} = time();
    }

    return { db_change_made => 1 };
}

#----------------------------------------------------------------------

sub get_session_db_rec {
    my($self) = @_;
    # Note: despite the name, $self may be a non-object hashref, and should not be used directly
    # in method calls without verifying that it's an object.

    if ($self && ref($self) && ref($self) ne 'HASH') {
        my $fields = $self->fields;
        for my $field_name (keys %$fields) {
            my $field_def = $fields->{$field_name};

            if ($field_def->{validate_func}) {
                $self->validate_field(field_def => $field_def,
                                      field_name => $field_name,
                                      store_obj => $self,
                                      existing_obj => {},
                                      field_value => $self->{$field_name},
                                      dbh => $self->{dbh},
                                     );
            }
        }
    }

    my %retval = (
                  data_point_value => $self->{data_point_value},
                  data_point_scope => $data_point_scope_reverse{$self->{data_point_scope}},
                  master_key_additional_values => {},
                 );

    if ($self->{data_point_action}) {
        $retval{data_point_action} = $data_point_action_reverse{$self->{data_point_action}};
        $retval{data_point_session_db_value} = $data_point_session_db_value{$self->{data_point_action}};
    }

    if ($self->{data_point_scope} == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_USER) {
        my $username_as_hex = Math::BigInt->new(F5::CRC::get_crc($self->{data_point_value}))->as_hex;
        $username_as_hex =~ s/^0x//;
        $retval{data_point_value} = $username_as_hex;
        $retval{master_key_additional_values} = { value => $self->{data_point_value} };
    }
    if ($self->{data_point_scope} == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_IP) {
        my($ip_address, $route_domain) = split(/\%/, $self->{data_point_value});
        $route_domain ||= 0;
        $retval{data_point_value} = uc unpack('H*', F5::IpAddr::ip_pton($ip_address)) . "%" . $route_domain;
    }

    $retval{session_db_master_key} = "asm.session.dp:$retval{data_point_scope}:$retval{data_point_value}:$self->{policy}->{policy_name_crc}";
    $retval{session_db_window_key} = "asm.session.sw:$retval{data_point_scope}:$retval{data_point_value}:$self->{policy}->{policy_name_crc}";
    if ($self->{data_point_action}) {
        $retval{session_db_action_key} = "$retval{session_db_master_key}.$retval{data_point_action}";
    }

    return \%retval;
}

#----------------------------------------------------------------------

sub delete {
    my($self, %params) = @_;
    my $db_master_key_cleanup = defined $params{db_master_key_cleanup} ? $params{db_master_key_cleanup} : 0;

    my $session_db = $params{session_db} || $self->get_new_session_db();

    my $session_db_rec = $self->get_session_db_rec;
    if ($session_db_rec->{session_db_action_key}) {
        $session_db->delete(key => $session_db_rec->{session_db_action_key});
    }

    if ($session_db_rec->{session_db_window_key}) {
        # Delete master key of sliding window
        $session_db->delete(key => $session_db_rec->{session_db_window_key},
                            bin_name => $WINDOW_BIN_NAME,
                           );
    }

    if ($db_master_key_cleanup) {
        $self->db_master_key_cleanup(%params,
                                     session_db => $session_db,
                                     keys => [ $session_db_rec->{session_db_master_key} ]);
    }

    $self->{session_db_rec} = $session_db_rec;
 
    if ($self->{is_persistent}) {
        $self->SUPER::delete(%params);
    }
}

#----------------------------------------------------------------------

sub delete_many {
    my($class, %params) = @_;

    my $session_db = $class->get_new_session_db(batch_updates => 1,
                                        batch_size => $SESSION_DB_BATCH_SIZE,
                                       );
    my @entities;

    if ($params{master_keys}) {
        my $policy = $params{policy};
        my @master_db_recs = $session_db->get(key => 'asm.session.dp:*');
        if ($policy && $policy->{policy_name_crc}) {
            @master_db_recs = grep { $_->{key} =~ /^asm\.session\.dp:[^:]*:[^:]*:$policy->{policy_name_crc}$/ } @master_db_recs;
        }

        for my $master_db_rec (@master_db_recs) {
            $session_db->delete(key => $master_db_rec->{key});
        }

        my @window_db_recs = $session_db->get(key => 'asm.session.sw:*',
                                              bin_name => $WINDOW_BIN_NAME,
                                             );
        if ($policy && $policy->{policy_name_crc}) {
            @window_db_recs = grep { $_->{key} =~ /^asm\.session\.sw:[^:]*:[^:]*:$policy->{policy_name_crc}$/ } @window_db_recs;
        }

        for my $window_db_rec (@window_db_recs) {
            $session_db->delete(key => $window_db_rec->{key},
                                bin_name => $WINDOW_BIN_NAME,
                               );
        }

        my $search_criteria = {};
        if ($params{search_criteria}) {
            $search_criteria = { %{$params{search_criteria}} };
        }

        if ($policy && $policy->{policy_id} &&
            (! defined $search_criteria->{is_persistent} || $search_criteria->{is_persistent} == 1)) {

            $search_criteria->{is_persistent} = 1;
            @entities = $class->SUPER::delete_many(%params,
                                                   search_criteria => $search_criteria,
                                                  );
        }
    }
    else {

        @entities = $class->retrieve_all(%params,
                                         session_db => $session_db);

        for my $entity (@entities) {
            $entity->delete(%params,
                            session_db => $session_db,
                            db_master_key_cleanup => 0,
                           );
        }
    }

    $session_db->flush;

    return @entities;
}

#----------------------------------------------------------------------

sub db_master_key_cleanup {
    my($class, %params) = @_;

    my $keys = $params{keys};
    my $session_db = $params{session_db};

    return unless ($keys && @$keys);

   
    # filter out duplicates
    my %unique_keys = map { ( $_ => 1 ) } @$keys;

    my @session_db_recs = $session_db->get(key => [ map { "$_**" } (keys %unique_keys) ],
                                           extended => 1);
    my %master_key_count;

    for my $session_db_rec (@session_db_recs) {
        if ($session_db_rec->{key} =~ /^(asm\.session\.dp:.*?:.*?:\d+)/) {
            $master_key_count{$1}++;
        }
    }

    for my $orphan_master_key (grep { $master_key_count{$_} == 1 } keys %master_key_count) {
        $session_db->delete(key => $orphan_master_key);
    }
}

#----------------------------------------------------------------------

sub get_from_session_db {
    my ($class, %params) = @_;
    my $policy = F5::ASMConfig::Cache::get_cached_policy(policy => $params{policy},
                                                         dbh => $params{dbh});

    my $search_filters = $params{search_filters};   
 
    my @return_objs;

    my $session_db = $params{session_db} || $class->get_new_session_db();
    my @session_db_recs;

    my($scope_filter) = grep { $_->{field_name} eq 'data_point_scope' && 
                              ($_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() ||
                               $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() ||
                               $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN())} @$search_filters;
    my($value_filter) = grep { $_->{field_name} eq 'data_point_value' &&
                              ($_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() ||
                               $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() ||
                               $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN())} @$search_filters;

    my %dp_value_lookup;
    if ($scope_filter && $value_filter && $policy && $policy->{policy_name_crc}) {
        my @scopes = ($scope_filter->{field_value_list} ? @{$scope_filter->{field_value_list}} : ($scope_filter->{field_value}) );
        my @values = ($value_filter->{field_value_list} ? @{$value_filter->{field_value_list}} : ($value_filter->{field_value}) );
        # Scope and value are known - go directly to the desired record
        for my $scope (@scopes) {
            for my $value (@values) {
                my $session_db_rec = get_session_db_rec({ data_point_value => $value,
                                                          data_point_scope => $scope,
                                                          policy => $policy,
                                                        });
                push (@session_db_recs, "$session_db_rec->{session_db_master_key}**");
                $dp_value_lookup{$session_db_rec->{data_point_scope}}->{$session_db_rec->{data_point_value}} = $value;
            }
        }
    }
    else {
        # Get all top-level keys in session db
        for my $session_db_rec ($session_db->get(key => 'asm.session.dp:*')) {
            my($scope, $value, $policy_name_crc) = ($session_db_rec->{key} =~ /^asm\.session\.dp:(.*?):(.*?):(\d+)/);
            next if ($policy->{policy_name_crc} && $policy_name_crc != $policy->{policy_name_crc});

            push(@session_db_recs, "$session_db_rec->{key}**");
            $dp_value_lookup{$scope}->{$value} = $session_db_rec->{value};
        }
    }

    if ($#session_db_recs > $SESSION_DB_MAX_QUERY_SIZE) {
        splice(@session_db_recs, $SESSION_DB_MAX_QUERY_SIZE);
    }

    while (@session_db_recs) {
        my @keys_to_get = splice(@session_db_recs, 0, $SESSION_DB_BATCH_SIZE);
        my @data_points = $session_db->get(key => \@keys_to_get, extended => 1);
        for my $data_point (@data_points) {
            if ($data_point->{key} =~/^asm\.session\.dp:(.*?):(.*?):$policy->{policy_name_crc}\.(.*)/) {
                my ($scope, $value, $action) = ($1, $2, $3);
                my $rec = {
                           data_point_scope => $data_point_scope{$scope},
                           data_point_value => $value,
                           data_point_action => $data_point_action{$action},
                           dbh => $params{dbh},
                           logger => $params{logger},
                           is_persistent => 0,
                           create_time => (time - $data_point->{age}),
                           policy => $policy,
                           expiration_time => ($data_point->{timeout} > $VERY_LARGE_TIMEOUT) ? 0
                                                                                             : time  + ($data_point->{timeout} - $data_point->{inactivity_time}),
                          };

                if ($rec->{data_point_scope} == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_USER) {
                    $rec->{data_point_value} = $dp_value_lookup{$scope}->{$value};
                }
                elsif ($rec->{data_point_scope} == F5::ASMConfig::SessionAwarenessScope::SESSION_SCOPE_IP) {
                    my($ip_address_hex, $route_domain) = split(/\%/, $rec->{data_point_value});
                    $rec->{data_point_value} = F5::IpAddr::ip_ntop(pack('H*', $ip_address_hex));
                    if ($route_domain) {
                        $rec->{data_point_value} .= "%" . $route_domain;
                    }
                }
                my $obj = $class->new(object => $rec);
                $obj->set_rest_uuid();
                push(@return_objs, $obj);
            }
        }
    }

    return @return_objs;
}

#----------------------------------------------------------------------

sub reload_session_db {
    my($class, %params) = @_;
    my $dbh = $params{dbh};
    my $logger = $params{logger};

    my $session_db = $class->get_new_session_db(batch_updates => 1,
                                                batch_size => $SESSION_DB_BATCH_SIZE,
                                               );
    $logger->debug("Reloading session awareness data points to session DB") if $logger;

    if ($params{delete_existing}) {
        # clear out session db, such that only persistent DPs from the (new?) policy will be loaded
        $class->delete_many(%params,
                            search_criteria => { is_persistent => 0 },
                            master_keys => 1,
                           );
    }

    my %retrieve_opts;
    if ($params{policy}) {
        $retrieve_opts{policy} = $params{policy};
    }

    my $iter = $class->retrieve_all(dbh => $dbh,
                                    logger => $logger,
                                    get_criteria => { is_persistent => 1},
                                    %retrieve_opts,
                                   );
    while(my $rec = $iter->get_next) {
        $rec->{policy} = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh, policy => $rec->{policy});
        next if (! $rec->{policy}->{account_id});

        $rec->{is_persistent} = 0;
        $rec->store_add(session_db => $session_db,
                        save_to_session_db_only => 1,
                        in_reload_session_db => 1,
                        create_time => $rec->{create_time},
                       );
    }

    $session_db->flush;
}

#----------------------------------------------------------------------

sub get_new_session_db {
    my($class, %params) = @_;
    return F5::SessionDB->get_instance(default_bin_name => $ACTION_BIN_NAME,
                                       %params);
}

#----------------------------------------------------------------------

sub relayable {
    my ($class_or_self, %params) = @_;
    if (ref $class_or_self) {
        return $class_or_self->{is_persistent} ? F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_IS_RELAYABLE()
                                               : F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_IS_NOT_RELAYABLE();
    }
    else {
        return $class_or_self->SUPER::relayable(%params);
    }
}

#----------------------------------------------------------------------

sub suppress_events {
    my ($self, %params) = @_;
    if (ref ($self) && $self->{is_persistent}) {
        return 0;
    }
    return 1;
}

#----------------------------------------------------------------------

sub declarative_search_filters {
    return [ F5::ASMConfig::SearchFilter->new(field_name => 'is_persistent',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                              field_value => 1),
           ];
}

#----------------------------------------------------------------------

sub diff_search_filters {
    return [ F5::ASMConfig::SearchFilter->new(field_name => 'is_persistent',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                              field_value => 1),
           ];
}

#----------------------------------------------------------------------

sub rest_search_filters {
    return [];
}

#----------------------------------------------------------------------


1;
