package F5::ASMConfig::Entity::SessionAwarenessSettings;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use F5::Provisioning;
use F5::ASMConfig::Cache;

use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::SessionAwarenessSettings/;

#----------------------------------------------------------------------

sub before_store {
  my($self, %params) = @_;

  my $store_mode = $params{store_mode};
  my $store_obj = $params{store_obj};
  my $existing_obj = $params{existing_obj};

  return unless $params{validate_value};

  my @fields = qw(
                   application_username_type
                   session_awareness_login_objects
                   delay_blocking_violations
                  );

   my %field_values;
   my %field_display_name;

   for my $field (@fields) {
      $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                    existing_obj => $existing_obj,
                                                    field_name => $field, );

      $field_display_name{$field} = $self->get_field_display_name_string(%params,field_name => $field);
   }


    if ($self->has_value_changed(%params, check_fields => [ 'application_username_type' ]) 
        && ($field_values{application_username_type}  == F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_APM()) ){
       
         my $is_apm =  F5::Provisioning::is_module_enabled("apm");
         die  "To use this option, you must provision BIG-IP Access Policy Manager (APM).\n" if !$is_apm;
    }


   if ($field_values{application_username_type} == F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_NONE() ){
       $store_obj->{flg_block_only_authenticated} = 0;
       $store_obj->{flg_user_block_illegal_threshold_enabled} = 0;
       $store_obj->{flg_user_block_all_threshold_enabled} = 0;
       $store_obj->{flg_user_track_all_threshold_enabled} = 0;
   }elsif ($field_values{application_username_type} == F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_APM()){ 
       $store_obj->{flg_block_only_authenticated} = 0;
       $store_obj->{flg_sid_block_all_threshold_enabled} = 0;
       $store_obj->{flg_sid_track_all_threshold_enabled} = 0;
       $store_obj->{flg_sid_block_illegal_threshold_enabled} = 0;
   }

   my $login_objects_defined = exists $store_obj->{session_awareness_login_urls} || scalar @{$field_values{session_awareness_login_objects}};

   if  (  $field_values{application_username_type} != F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_LOGIN_PAGES() 
           && $login_objects_defined ) {
       $store_obj->{session_awareness_login_objects} = $field_values{session_awareness_login_objects} = [];
   }


   if ( $field_values{application_username_type} == F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_LOGIN_PAGES() 
        && !($login_objects_defined) ){
          my $error_message = "$field_display_name{session_awareness_login_objects} list cannot be empty when using $field_display_name{application_username_type} 'login-pages'";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
              };

   }

   my $cached_violations = F5::ASMConfig::Cache::get_cached_violations(dbh => $self->{dbh});
   for my $violation (@{$field_values{delay_blocking_violations}}) {
       my $cached_violation = $cached_violations->{$violation->{viol_index}};
       if ($cached_violation && $cached_violation->{flg_exclude_session_awareness}) {
           my $error_message = "Violation '$cached_violation->{description}' is invalid for $field_display_name{delay_blocking_violations}";
           die { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
               };
       }
   }

   for my $login_object ( @{$field_values{session_awareness_login_objects}}){
      if (! $login_object || ! ref($login_object) || ! $login_object->{object_id}){  
          my $error_message = "A valid Login Page must be specified in $field_display_name{session_awareness_login_objects}";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
              };
      }

      my $full_login_object = F5::ASMConfig::Entity::LoginObject->new(%params,
                                                                      dbh => $self->{dbh},
                                                                      logger => $self->{logger},
                                                                      get_criteria => { url => F5::ASMConfig::URL->new(url_id => $login_object->{object_id})});
 
 
      if ( !  $full_login_object || ! ref($full_login_object) ||  $full_login_object->{auth_type}  ==  F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_NONE()){
          my $error_message = "Login Page with authentication type 'None' is invalid for $field_display_name{session_awareness_login_objects}";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
              };
      }
   }

}
#----------------------------------------------------------------------

sub get_external_entities {
    my ($self, %params) = @_;
    my @external_entities = @{$self->SUPER::get_external_entities(%params)};
    if ($self->{session_awareness_login_urls}) {
        for my $login_url (@{$self->{session_awareness_login_urls}}) { 
            if (my $login_object = F5::ASMConfig::Entity::LoginObject->new(dbh => $self->{dbh},
                                                                           policy => $self->{policy},
                                                                           logger => $self->{logger},
                                                                           get_criteria => { url => $login_url->{url} })) {

                if (my @chained_entities = @{$login_object->get_external_entities()}) {
                    push (@external_entities, @chained_entities);
                }
                push (@external_entities, $login_object);
            }
        }
    }
    return \@external_entities;
}

#----------------------------------------------------------------------

1;
