package F5::ASMConfig::Entity::Signature;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::ASMConfig;
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::Signature/;
use F5::NegativeSignatures::Rule;
use F5::NegativeSignatures;
use F5::NegativeSignatures::Utils;

my  %system_name_cache;
my  %system_id_cache;
#----------------------------------------------------------------------

sub update_many_as_iterative {
    return 1;
}

#----------------------------------------------------------------------

sub delete_many_as_iterative {
    return 1;
}

#----------------------------------------------------------------------

sub rest_search_filters {
    my ($class) = @_;
    return [ F5::ASMConfig::SearchFilter->new(field_name => 'flg_is_hidden',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                                              field_value => 1),
           ];
}

#----------------------------------------------------------------------

sub linked_entity_filters {
    my ($class_or_self, %params) = @_;
    my $entities = $params{entities} || [$class_or_self];
    my @entity_filters = ({
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'sig_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types => ['F5::ASMConfig::Entity::PolicySignature'],
                          },);
    return @entity_filters;
}


#----------------------------------------------------------------------

sub before_store {
    my ($self, %params) = @_;
    my $store_obj = $params{store_obj};

    if ($self->has_value_changed(%params, check_fields => 'rule')) {
        $store_obj->{last_enforcement_change} = time();
    }
}

#----------------------------------------------------------------------

sub delete_linked_entities {
    my ($class_or_self, %params) = @_;

    #delete the policy signatures, and only mark the policies that have their sets included as modified
    my $dbh = $params{dbh};
    my $entities = $params{entities};

    if (ref ($class_or_self)) {
        $dbh      ||= $class_or_self->{dbh};
        $entities ||= [ $class_or_self ];
    }

    my @deleted_entities;

    my @sets = F5::ASMConfig::Entity::SignatureSet->retrieve_all(dbh => $dbh,
                                                                 search_filters => [ F5::ASMConfig::SearchFilter->new({
                                                                                                                field_name => 'sigs__sig_id',
                                                                                                                field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                                                                                field_value_list => [ map { $_->get_id() } @$entities ],}),
                                                                                   ],
                                                                 );

    my %affected_policies;
    if (@sets) {
        my @policies_sets = F5::ASMConfig::Entity::PolicySignatureSet->retrieve_all(dbh => $dbh,
                                                                                    search_filters => [ F5::ASMConfig::SearchFilter->new({
                                                                                                            field_name => 'set_id',
                                                                                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                                                                            field_value_list => [ map { $_->get_id() } @sets ],}),
                                                                                                      ],
                                                                                   );
        %affected_policies = map { ( $_->{policy}->{policy_id} => 1) } @policies_sets;
    }     
    my $policy_signatures_iter = F5::ASMConfig::Entity::PolicySignature->retrieve_all(dbh => $dbh,
                                                                                      search_filters => [ 
                                                                                          F5::ASMConfig::SearchFilter->new({field_name => 'sig_id',
                                                                                                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                                                                                            field_value_list => [ map { $_->get_id() } @$entities ],}),
                                                                                      ],);
    while (my $policy_signature = $policy_signatures_iter->get_next()) {
        push (@deleted_entities, $policy_signature);
        $policy_signature->delete(suppress_events => !$affected_policies{$policy_signature->{policy}->{policy_id}});
    }

    return @deleted_entities;
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
  my($class, %params) = @_;
  my $dbh = $params{dbh};
  my $rec = $params{rec};
  my $selected_fields_hash = $params{selected_fields_hash};

  $rec->{attack_type_name} = F5::NegativeSignatures::Utils->get_attack_type_name( attack_type_id =>  $rec->{attack_type_id}, dbh => $dbh);

  if ($rec->{rule} && $rec->{flg_is_user_defined} && (! $selected_fields_hash || $selected_fields_hash->{keywords})) {
      my $rule = F5::NegativeSignatures::Rule->new(rule_string => $rec->{rule},
                                                   apply_to => $rec->{apply_to},
                                                  );

      $rec->{keywords} = [ map { F5::ASMConfig::SignatureKeyword->new($_) } @{$rule->to_simple_keywords()} ];
  }
}

#----------------------------------------------------------------------
sub reload_cache {
  my($class, %params) = @_;

  my $dbh;
  if (ref($class)) {
    $dbh = $class->{dbh};
  }
  else {
    $dbh = $params{dbh};
  }
  %system_name_cache = ();
  %system_id_cache = ();
  my $plc_dbname = 'PLC';

  my $sth = $dbh->prepare(<<EndQuery);
SELECT system_display_name,system_id
  FROM $plc_dbname.NEGSIG_SYSTEMS
EndQuery
  $sth->execute;

  while (my $row = $sth->fetchrow_hashref) {
    $system_name_cache{$row->{system_display_name}} = $row->{system_id};
    $system_id_cache{$row->{system_id}} = $row->{system_display_name};
  }
}
#-----------------------------------------------------------------------
sub get_system_name {
  my($class, %params) = @_;

  my $system_id = $params{system_id};
  unless ($system_id_cache{$system_id}) {
    $class->reload_cache(%params);
  }

  return $system_id_cache{$system_id};
}
#------------------------------------------------------------------------
sub get_system_id {
  my($class, %params) = @_;

  my $system_name = $params{system_name};
  unless ($system_name_cache{$system_name}) {
    $class->reload_cache(%params);
  }

  return $system_name_cache{$system_name};
}
#------------------------------------------------------------------------
sub retrieve_nested_structures {
  my($self,%params) = @_;

  $self->SUPER::retrieve_nested_structures(%params);

  # set the interpolated link
  if ($self->{sig_references} && !$params{call_flags}->{rest_call}) {
      F5::NegativeSignatures->new(dbh => $self->{dbh})->map_sig_references( [$self ]);
  }

  if ($self->{systems}) {
   for my $sig_system (@{$self->{systems}}){
     $sig_system->{system_name} = $self->get_system_name( system_id => $sig_system->{system_id});
   }
  }
}

#------------------------------------------------------------------------
sub get_full_display_name {
    my ($self, %params) = @_;

    if ( !defined $self->{sig_tag} || $self->{sig_tag} eq '' ){
         return $self->{sig_name};
    }

    return $self->SUPER::get_full_display_name(%params);
}
#------------------------------------------------------------------------

sub to_declarative {
    my ($self, %params) = @_;

    my $rv;
    my $declarative_fields = $self->declarative_fields;
    if ($self->{flg_is_user_defined}) {
        $rv->{$declarative_fields->{sig_name}->{rest_field_name}} = $self->{sig_name};
        if ($self->{sig_tag}) {
            $rv->{$declarative_fields->{sig_tag}->{rest_field_name}} = $self->{sig_tag};
        }
    }
    else {
        $rv->{$declarative_fields->{sig_id}->{rest_field_name}} = $self->{sig_id};
    }

    return $rv;
}

#------------------------------------------------------------------------

1;
