package F5::ASMConfig::Entity::SsrfHost;
use base qw(F5::ASMConfig::Entity::Base F5::ASMConfig::SsrfHost);


use strict;
use warnings;

use F5::IpUtils;

#----------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};

    if ($params{validate_value} && $self->is_pure_wildcard(%params) &&
        $self->get_new_value(%params, field_name => 'action_type') == F5::ASMConfig::SsrfActionType::SSRF_ACTION_ALLOW()) {

        my $action_display_string = $self->get_field_display_string(%params, field_name => 'action_type', 'field_value' => $store_obj->{action_type});
        my $error_message = "Action $action_display_string is not allowed for host '*'.";
        die { error_message   => $error_message,
              display_message => $error_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
            };
    }

    if ($params{validate_value} && ($self->get_new_value(%params, field_name => 'action_type') == F5::ASMConfig::SsrfActionType::SSRF_ACTION_RESOLVE())) {
        my $input_host  = $self->get_new_value(%params, field_name => 'host');
        my $is_valid_ip = F5::IpUtils::is_valid_ip(field_value => $input_host);

        if ($is_valid_ip && $is_valid_ip->{completed_successfully}) {
            my $action_display_string = $self->get_field_display_string(%params, field_name => 'action_type', 'field_value' => $store_obj->{action_type});
            my $error_message = "'$action_display_string' is not applicable for '$input_host'";
            die { error_message   => $error_message,
                  display_message => $error_message,
                  error_level     => 'info',
                  error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                };
            }
    }

    if ($store_obj->{host} =~ /^\*\.(.*)/) {
        $store_obj->{is_wildcard} = 1;
    }
    else {
        $store_obj->{is_wildcard} = 0;
    }

}

#----------------------------------------------------------------------

sub is_pure_wildcard {
    my ($self, %params) = @_;
    return ($self->get_new_value(%params, field_name => 'host') &&
            $self->get_new_value(%params, field_name => 'host') eq "*");
}

#----------------------------------------------------------------------

sub delete {
    my ($self, %params) = @_;

    if ($self->{host} eq "*") {
        my $error_message = "Cannot delete '*' host.";
        die { error_message   => $error_message,
              display_message => $error_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
            };
    }

    return $self->SUPER::delete(%params);
}

#----------------------------------------------------------------------

sub delete_many_as_iterative {
    return 1
}

#----------------------------------------------------------------------

sub update_many_as_iterative {
    return 1;
}

#----------------------------------------------------------------------

sub retrieve_all {
    my ($class, %params) = @_;

    if ( $params{sort_orders} && @{$params{sort_orders}}) {
        return $class->SUPER::retrieve_all(%params);
    }

    my @ssrf_hosts_ns = $class->SUPER::retrieve_all(%params);
    my (@ip_list, @wildcard_domains, @explicit_domains, @pure_wildcards);

    foreach my $ssrf_host (@ssrf_hosts_ns) {
        my $is_valid_ip = F5::IpUtils::is_valid_ip(field_value => $ssrf_host->{'host'});
        if ($is_valid_ip && $is_valid_ip->{completed_successfully}) {
            push @ip_list, $ssrf_host;
            next;
        }

        if ($ssrf_host->{is_wildcard} && $ssrf_host->{host} ne '*') {
            push @wildcard_domains, $ssrf_host;
        }
        elsif ($ssrf_host->{host} eq "*") {
            push @pure_wildcards, $ssrf_host;
        }
        else {
            push @explicit_domains, $ssrf_host;
        }
    }

    my @ordered_ip_list = sort by_host_prefix_length @ip_list;
    my @ordered_wildcard_domains = sort by_number_of_dots_in_host @wildcard_domains;
    my @ordered_explicit_domains = sort by_number_of_dots_in_host @explicit_domains;

    my @ssrf_hosts = (@ordered_explicit_domains, @ordered_wildcard_domains, @ordered_ip_list, @pure_wildcards);
    my $total_record_count = scalar @ssrf_hosts;
    $class->apply_paging(result_set => \@ssrf_hosts, %params);

    if (wantarray) {
        return @ssrf_hosts;
    }
    else {
        my $iterator = F5::ASMConfig::Entity::Iterator->new(
                                                            get_next_function => sub {return shift(@ssrf_hosts);},
                                                            get_all_function => sub {return @ssrf_hosts;},
                                                            total_record_count => $total_record_count,
                                                           );
        return $iterator;
    }
}

#----------------------------------------------------------------------

sub by_number_of_dots_in_host {
    my $no_of_dots_a = () = $a->{host} =~ /\./gi;
    my $no_of_dots_b = () = $b->{host} =~ /\./gi;

    if ($no_of_dots_a > $no_of_dots_b) {
        return -1;
    }
    elsif ($no_of_dots_a == $no_of_dots_b) {
        return $a->{host} cmp $b->{host};
    }
    else {
        return 1;
    }
}

#----------------------------------------------------------------------

sub by_host_prefix_length {
    my $host_a = $a->{host};
    my $host_b = $b->{host};
    my ($host_a_prefix_length, $host_b_prefix_length) = (0, 0);

    if ($host_a =~ /(.*)\/(\d+)/) {
        $host_a_prefix_length = $2;
    }

    if ($host_b =~ /(.*)\/(\d+)/) {
        $host_b_prefix_length = $2;
    }

    if ($host_a_prefix_length > $host_b_prefix_length ) {
        return -1;
    }
    elsif ($host_a_prefix_length == $host_b_prefix_length) {
        return $a->{host} cmp $b->{host};
    }
    else {
        return 1;
    }
}

#----------------------------------------------------------------------

sub sort_domains {
    my ($domains) = @_;
    return sort by_number_of_dots_in_host @{$domains};
}

#----------------------------------------------------------------------

1;
