package F5::ASMConfig::Entity::URL;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::Entity::Policy;
use F5::ASMConfig::Entity::ContentProfile;
use F5::ASMConfig::Entity::Parameter;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::UrlContentProfile;
use base qw/F5::ASMConfig::CorsBase F5::ASMConfig::URL/;

use List::Util qw(max);

use F5::CRC;
use F5::DbUtils;
use F5::ParseUri;

our $ANY_URL_METHOD_NAME = '*';
my %PROFILE_ENFORCEMENT_TYPES = (
                          F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_XML() => {title => 'XML', profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML()},
                          F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON() => {title => 'JSON', profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON()},
                          F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GWT() => {title => 'GWT', profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GWT()},
                          F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRAPHQL() => {title => 'GraphQL', profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRAPHQL()},
                          F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRPC() => {title => 'GRPC', profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRPC()},
                       );

#----------------------------------------------------------------------
sub additional_field_mapping {
    my($class, %params) = @_;

    my %additional_field_mapping;
    if ($params{extra_fields} && grep {$_ eq "extra_scanner_vulns"} (@{$params{extra_fields}})) {  
        $additional_field_mapping{extra_scanner_vulns} = { nested_struct => 1,
                                                           entity_type => 'F5::ASMConfig::ExtraScannerVuln',
                                                           foreign_id_field => 'object_id',
                                                           parent_db_field_name => 'id',
                                                           parent_entity_field_name => 'url_id',
                                                           db_table_name => 'PLC.PL_SCANNER_VULNS',
                                                           fields => {
                                                                         vuln_id => { db_field_name => 'vuln_id', 
                                                                                      unique_record_part => 1,},
                                                                     },
                                                         };
    }

    return \%additional_field_mapping;

}

#----------------------------------------------------------------------

sub get_staging_subquery {
    my($class, %params) = @_;
    my $policy = $params{policy};
    my $want_suggestions = $params{has_suggestions} ? 1 : 0;

    my $plc_dbname = 'PLC';
    my $query = <<EndQuery;
  LEFT JOIN (
       SELECT DISTINCT O.rest_uuid AS union_rest_uuids, 1 AS matched FROM $plc_dbname.PL_OBJECTS AS O
          JOIN PLC.PL_SUGGESTIONS AS S ON (S.policy_id = ? AND S.suggestion_status = ? AND O.rest_uuid = S.element_rest_uuid AND S.element_type IN ( ? , ?))) url_suggestions
        ON (url_suggestions.union_rest_uuids = main_table.rest_uuid)
  LEFT JOIN (
       SELECT DISTINCT O1.rest_uuid AS union_rest_uuids, 1 AS matched FROM $plc_dbname.PL_OBJECTS AS O1
          JOIN $plc_dbname.PL_OBJECT_CONTENT_PROFILES AS O2 ON (O1.policy_id = ? AND O2.obj_id = O1.id)
          JOIN $plc_dbname.PL_CONTENT_PROFILES AS PR1 ON (O2.profile_id = PR1.id)
          JOIN $plc_dbname.PL_SUGGESTIONS AS S1 ON (S1.policy_id = ? AND S1.suggestion_status = ? AND PR1.rest_uuid = S1.element_rest_uuid AND S1.element_type IN (?, ?, ?, ?, ?))) profile_suggestions
       ON (profile_suggestions.union_rest_uuids = main_table.rest_uuid)
EndQuery

    return { query => $query,
         bind_values => [ $policy->{policy_id}, F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING(), $class->element_type, F5::ASMConfig::ElementType::URL_CONTENT_PROFILE,
                          $policy->{policy_id},
                          $policy->{policy_id}, F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING(), F5::ASMConfig::ElementType::XML_PROFILE, F5::ASMConfig::ElementType::JSON_PROFILE, F5::ASMConfig::ElementType::GWT_PROFILE, F5::ASMConfig::ElementType::GRAPHQL_PROFILE, F5::ASMConfig::ElementType::GRPC_PROFILE],
         db_field => "CASE WHEN (url_suggestions.matched OR profile_suggestions.matched) THEN $want_suggestions ELSE NOT $want_suggestions END"};
}

#----------------------------------------------------------------------

sub remove_non_unique_record_parts {
    my ($class_or_self, %params) = @_;
    my $object = $params{object} || $class_or_self;
    if ($object->{url_id}) {
        $class_or_self->SUPER::remove_non_unique_record_parts(%params);
    }
    return;
}

#----------------------------------------------------------------------

sub purify_store_object {
    my ($class_or_self, %params) = @_;
    my $object = $params{object} || $class_or_self;

    if ( $object->{positional_parameters} && @{$object->{positional_parameters}} ) {
        for my $positional_parameter (@{$object->{positional_parameters}}) {
            $positional_parameter->{param}->{param_id} ||= $positional_parameter->{param_id};
            delete $positional_parameter->{param_id};
            # purify_store_object doesn't support recursion for this type of nested structures
            F5::ASMConfig::Entity::Parameter->purify_store_object(object => $positional_parameter->{param},
                                                                  dbh => $params{dbh}, );
        }
    }

    if (defined($object->{url_id}) && !$object->{url_id}) {
        return $class_or_self;
    }
    else {
        return $class_or_self->SUPER::purify_store_object(%params);
    }    
}

#----------------------------------------------------------------------

sub before_store {
  my($self, %params) = @_;

  my $db_rec = $params{db_rec};
  my $store_mode = $params{store_mode};
  my $store_obj = $params{store_obj};
  my $existing_obj = $params{existing_obj};
  my $call_flags = $params{call_flags};
  my $policy = $params{policy} || $self->{policy};

  my $from_learning = $self->is_from_learning(%params);

  my @fields = qw(
                   flg_clickjacking_protection
                   clickjacking_type
                   clickjacking_from
                   is_wildcard
                   enable_cors_enforcement
                   flg_allow_cors
                   allowed_origins_enforcement_mode
                   flg_allowed_origins_check
                   allowed_origins
                   flg_allowed_methods_check
                   allowed_methods
                   flg_allowed_headers_check
                   flg_exposed_headers_check
                   flg_check_credentials
                   check_maximum_age
                   maximum_age
                   url_content_profiles
                   url_name
                   cors_enforcement_mode
                   flg_in_classification
                   flg_is_allowed
                   positional_parameters
                   method_name
                   flg_check_obj_methods
                   flg_mandatory_body
                 );

   my %field_values;
   my %field_display_name;

   for my $field (@fields) {
      $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                    existing_obj => $existing_obj,
                                                    call_flags => $call_flags,
                                                    field_name => $field, );

      $field_display_name{$field} = $self->get_field_display_name_string(%params, field_name => $field);
   }

   if ($store_mode eq 'add' && $store_obj->{positional_parameters} && @{$store_obj->{positional_parameters}}) {
       if(!$store_obj->{is_wildcard}) {
           my $error_message = "A URL containing positional parameters must be a wildcard URL";
           die { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
             };
       }

       my @url_parts = split(/\//,$store_obj->{url_name},-1);
       my %url_segment_indexes;
       for my $positional_parameter (@{$store_obj->{positional_parameters}}) {
           my $url_segment_index = $positional_parameter->{url_segment_index};
           my $param_name = ($positional_parameter->{param} && $positional_parameter->{param}->{param_name}) ? $positional_parameter->{param}->{param_name}
                                                                                                             : $positional_parameter->{param_name};
           my $param_id = ($positional_parameter->{param} && $positional_parameter->{param}->{param_id}) ? $positional_parameter->{param}->{param_id}
                                                                                                         : $positional_parameter->{param_id};

           if (!$param_id && !$param_name) {
               my $error_message = "$field_display_name{positional_parameters} missing valid parameter";
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                   };
           }

           my $param_criteria = $param_name ? { param_name => $param_name,
                                                param_level => F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL(), }
                                            : { param_id => $param_id };

           if (exists $url_segment_indexes{$url_segment_index}) {
               my $error_message = "$field_display_name{positional_parameters} must each have distinct indexes";
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                 };
           }
           else {
               $url_segment_indexes{$url_segment_index} = 1;
           }

           my $parameter = F5::ASMConfig::Entity::Parameter->new(policy => $policy,
                                                                 dbh => $self->{dbh},
                                                                 logger => $self->{logger},
                                                                 get_criteria => $param_criteria,
                                                                );
           if (!$parameter) {
               my $error_message = "A positional parameter with index of $url_segment_index could not be found";
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                 };
           }

           $positional_parameter->{param} = $parameter;

           if ($url_segment_index < 1 || $url_segment_index > $#url_parts) {
               my $error_message = "A positional parameter index of $url_segment_index is illegal for URL $store_obj->{url_name}";
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                 };
           }

           if ($parameter->{param_level} != F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL()) {
               my $error_message = "A positional parameter must be of global type";
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                 };
           }

           if ($parameter->{is_wildcard}) {
               my $error_message = "A positional parameter must be explicit";
               die { error_message => $error_message,
                     display_message => $error_message,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                 };
           }
       }
   }

   if ($field_values{method_name} &&
       $field_values{method_name} ne $ANY_URL_METHOD_NAME &&
       defined($field_values{cors_enforcement_mode}) &&
       $field_values{cors_enforcement_mode} != F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_DISABLED() &&
       $field_values{cors_enforcement_mode} != F5::ASMConfig::CorsEnforcementMode::CORS_ENFORCEMENT_MODE_ENFORCE()) {
           my $error_message = $field_display_name{cors_enforcement_mode}. " may only be set to 'disabled' or 'enforce' for method specific URLs.";
           die { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::MUTUALLY_EXCLUSIVE_FIELDS(), };
   }

   if ($field_values{method_name} && $field_values{method_name} ne $ANY_URL_METHOD_NAME && $field_values{flg_check_obj_methods}) {
       my $error_message = "URL Methods override is not allowed for method specific URLs.";
       die { error_message => $error_message,
             display_message => $error_message,
             error_level     => 'info',
             error_code      => F5::ASMConfig::ErrorCode::MUTUALLY_EXCLUSIVE_FIELDS(), };
   }

   if (defined $store_obj->{flg_allowed_origins_check} && ! defined $store_obj->{allowed_origins_enforcement_mode}) {
      $store_obj->{allowed_origins_enforcement_mode} = $field_values{flg_allowed_origins_check} ?
                                                        F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_REPLACE_WITH() :
                                                        F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_UNMODIFIED();
   }

   if( $call_flags->{rest_call} ) { #validate content profiles
       for my $url_content_profile (@{$field_values{url_content_profiles}}) {
           my $enforcement_type = $url_content_profile->{'enforcement_type'};
           if( $PROFILE_ENFORCEMENT_TYPES{$enforcement_type} ) { 
               my $profile_type_title = $PROFILE_ENFORCEMENT_TYPES{$enforcement_type}->{title};
               my $profile_type = $PROFILE_ENFORCEMENT_TYPES{$enforcement_type}->{profile_type};

               my $content_profile_valid = 1;
               my $content_profile_error_message = "You must specify " . $profile_type_title  . " profile.";
               
               my $content_profile = $url_content_profile->{content_profile};
               
               if( !$content_profile || !ref($content_profile) ||
                   !$content_profile->{profile_name} || 
                   ( defined $content_profile->{content_profile_type} && $content_profile->{content_profile_type} != $profile_type ) ) {
                   $content_profile_valid = 0;
               } 
               elsif(!defined $content_profile->{content_profile_type}) { #cases when name passed - make get and validate profile
                   my $url_profile = F5::ASMConfig::Entity::ContentProfile->new(%params,
                                          dbh => $self->{dbh},
                                          logger => $self->{logger},
                                          get_criteria => { 
                                                           profile_name => $content_profile->{profile_name},
                                                           content_profile_type => $profile_type,
                                                          });
                   
                   if( !$url_profile ) { 
                       $content_profile_valid = 0;
                   } 
                   else {
                       $content_profile->{content_profile_type} = $url_profile->{content_profile_type};
                   }
               }
              
               if(!$content_profile_valid) {
                    die { error_message => $content_profile_error_message,
                          display_message => $content_profile_error_message,
                          error_level     => 'info',
                          error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                        };
               }
           }
       }
   }

  if($self->has_value_changed(%params, check_fields => [ qw(is_wildcard) ])) {
       if($field_values{is_wildcard}) {
            my $logout_assoc = F5::ASMConfig::Entity::LogoutObject->new(%params,
                                    dbh => $self->{dbh},
                                    logger => $self->{logger},
                                    get_criteria => {
                                                     url => $self,
                                                    });

        if( $logout_assoc ) {
            my $error_message = "Logout URL cannot be wildcard";
            die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
            };
        }
     }
  }

  if ( !$field_values{is_wildcard} ){
     $store_obj->{flg_wildcard_includes_slash} = 0;
  }

 if ( $params{validate_value} && defined $field_values{flg_is_allowed} && !$field_values{flg_is_allowed} &&
       $self->is_pure_wildcard(%params, existing_obj => $existing_obj, store_obj => $store_obj) ){
          my $error_message = "Pure wildcard must be allowed";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
            };
  }

  if ( $params{validate_value} &&
       $field_values{flg_clickjacking_protection} &&  $field_values{clickjacking_type} == F5::ASMConfig::ClickjackingType::CLICKJACKING_ONLY_FROM() ) {

           my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($field_values{clickjacking_from});
           if ( ! ( (defined $app_id) && (defined $domain) )) {
                 die "The value '$field_values{clickjacking_from}' for '$field_display_name{clickjacking_from}' is invalid. The value should be a full valid URL\n";
           }
  }

  $self->handle_cors(%params, field_values => \%field_values, field_display_name => \%field_display_name);

  if ($store_mode eq 'add' || $self->has_value_changed(%params, check_fields => [ qw(url_name) ]) ) {
      $store_obj->{url_name} =~ s/^\s+|\s+$//g; #remove leading and trailing spaces
      if ( !$store_obj->{url_name} ){
           my $error_message = "$field_display_name{url_name} must be defined";
           die { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
           };
      }
      my $regex = $field_values{is_wildcard} ? '^(\/|\*|\[|\]|\?|\!)(.*)$'
                                             : '^(\/)(.*)$';
      if ($field_values{url_name} !~ /$regex/) {
          $store_obj->{url_name} = "/$store_obj->{url_name}";
      }
  }

  if ($field_values{flg_in_classification} && $call_flags->{rest_call}) {
      if ($params{store_mode} eq 'add') {
          $store_obj->{flg_in_classification} &&= $self->url_classification_flag(%params);
      }
      elsif ($self->has_value_changed(%params, check_fields => "flg_in_classification")){
          $store_obj->{flg_in_classification} = 0;
      }
  }


  my @modified_entities;
  if ( $params{store_mode} eq 'add' && !$store_obj->{url_content_profiles} && !$call_flags->{internal_flags_hash}->{skip_populate_defaults_on_add}){

      #if this is being added due to a suggestion and PB is at least in manual learning mode, inherit the classification status from PB_CONFIG
      my $flg_in_classification = $self->url_classification_flag(%params);
      if ( (!$from_learning || !$flg_in_classification)  && !($call_flags->{rest_call} && $store_obj->{flg_in_classification}) ) {
          $store_obj->{flg_in_classification} = 0;
          my $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $self->{policy});
          my $default_json_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                       dbh => $self->{dbh},
                                                                       policy => $self->{policy},
                                                                       content_profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),
                                                                   );
          if ($default_json_profile->{__new_entity}) {
              $default_json_profile->store_add(validate_value => $params{validate_value});
              push(@modified_entities, {event_type => F5::ASMConfig::EventType::ADD, entity => $default_json_profile, existing_entity => $default_json_profile});
          }

          my $default_xml_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                       dbh => $self->{dbh},
                                                                       policy => $self->{policy},
                                                                       content_profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(),
                                                                   );
          if ($default_xml_profile->{__new_entity}) {
              $default_xml_profile->store_add(validate_value => $params{validate_value});
              push(@modified_entities, {event_type => F5::ASMConfig::EventType::ADD, entity => $default_xml_profile, existing_entity => $default_xml_profile});
          }
      }
      else {
          $store_obj->{flg_in_classification} = 1;
      }
  }

  $self->adjust_wildcards(%params);
  $self->sync_from_pb_config(%params, 
                             pb_config_field => "flg_urls", 
                             table_name => F5::ASMConfig::Entity::PBConfigUrl->db_table(),);

  my $rv = {};
  $rv->{modified_entities} = \@modified_entities;
  return $rv;
}

#----------------------------------------------------------------------

sub after_store {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};
  my $store_obj = $params{store_obj};
  my $call_flags = $params{call_flags} || {};
  my $policy = $params{policy} || $self->{policy};

  my $from_learning = $self->is_from_learning(%params);

  if ($params{store_mode} eq 'add' && !$call_flags->{internal_flags_hash}->{skip_populate_defaults_on_add}) {
      if($store_obj->{url_content_profiles}) {
          my $found_default = 0;          
          for my $url_content_profile (@{$store_obj->{url_content_profiles}}) { 
              if(F5::ASMConfig::Entity::UrlContentProfile->is_default_url_content_profile(object => $url_content_profile)) {
                 $found_default = 1;
                 last;
              }
          }
          if(!$found_default) {
              my $error_message = "The default URL Content Profile (*:*) is mandatory.";
              die { error_message => $error_message,
                    display_message => $error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                  };
          }
      }
      else {
          if ( !$self->{flg_in_classification} ) {

              my $default_json_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                           dbh => $self->{dbh},
                                                                           policy => $policy,
                                                                           content_profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),
                                                                       );

             my $default_xml_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                           dbh => $self->{dbh},
                                                                           policy => $policy,
                                                                           content_profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(),
                                                                       );

             my $default_url_content_profiles = F5::ASMConfig::Entity::URL->get_default_url_content_profiles_db_rec(
                                                                                                                     default_xml_profile_id => $default_xml_profile->{profile_id},
                                                                                                                     default_json_profile_id => $default_json_profile->{profile_id},
                                                                                                                    );

             for my $default_url_content_profile ( @$default_url_content_profiles ){  

                    F5::DbUtils::insert_row(dbh => $dbh,
                                      table_name => "PLC.PL_OBJECT_CONTENT_PROFILES",
                                      ignore => 1,
                                      rec => {
                                              policy_id => $policy->{policy_id},
                                              obj_id => $store_obj->{url_id},
                                              profile_id => $default_url_content_profile->{profile_id}, 
                                              enforcement_type => $default_url_content_profile->{enforcement_type},
                                              header_name =>  $default_url_content_profile->{header_name},
                                              header_value => $default_url_content_profile->{header_value},
                                              header_order => $default_url_content_profile->{header_order},
                                             });
             }
          }
          else {

              F5::DbUtils::insert_row(dbh => $dbh,
                                      table_name => "PLC.PL_OBJECT_CONTENT_PROFILES",
                                      ignore => 1,
                                      rec => {
                                              policy_id => $policy->{policy_id},
                                              obj_id => $store_obj->{url_id},
                                              enforcement_type => 'apply_value_signatures',
                                             });


          }
      }
  }
}

#----------------------------------------------------------------------

sub linked_entity_filters {
    my ($class_or_self, %params) = @_;
    my $entities = $params{entities} || [$class_or_self];
    my @entity_filters = ({
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'to_url__url_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types => ['F5::ASMConfig::Entity::Flow'],
                          },
                          {
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'from_url__url_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types => ['F5::ASMConfig::Entity::Flow'],
                          },
                          {
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'url__url_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types => ['F5::ASMConfig::Entity::Parameter', 
                                            'F5::ASMConfig::Entity::Extraction', 
                                            'F5::ASMConfig::Entity::DynamicFlow',
                                            'F5::ASMConfig::Entity::BruteForce',
                                            'F5::ASMConfig::Entity::LoginObject'],
                          },
                          {
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'url__url_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types => ['F5::ASMConfig::Entity::UrlContentProfile'],
                           automatically_created => 1,
                           cascade_deleted => 1,
                          },
                          {
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'flow__to_url__url_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types  => ['F5::ASMConfig::Entity::Parameter'],
                          },
                          {
                           filter => F5::ASMConfig::SearchFilter->new({
                                                          field_name => 'flow__from_url__url_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                          field_value_list => [ map { $_->get_id() } @$entities ],}),
                           entity_types  => ['F5::ASMConfig::Entity::Parameter'],
                          },
                         );

    return @entity_filters;
}

#----------------------------------------------------------------------

sub after_nested_struct_load {
    my ($self, %params) = @_;
    $self->populate_positional_params_names();
    return;
}

#----------------------------------------------------------------------

sub populate_positional_params_names {
    my ($self, %params) = @_;
    my $store_obj = $params{store_obj} || $self;

    if (exists($store_obj->{positional_parameters})) {
        for my $positional_parameter (@{$store_obj->{positional_parameters}}) {
            if ( $positional_parameter->{param} && $positional_parameter->{param}->{param_name} ) {
                $positional_parameter->{param_name} = $positional_parameter->{param}->{param_name};
            }
        }
    }
}

#----------------------------------------------------------------------

sub diff_struct_store {
    my ($self, %params) = @_;
    my $rv = $self->SUPER::diff_struct_store(%params);

    if ( $params{store_mode} ne 'add' && $rv->{db_change_made} && $self->{flg_in_classification} && !$self->is_from_learning(%params)
         && $params{session}->{user_name} ne $F5::ASMConfig::Entity::Base::BIGIQ_USER_NAME ) {

        F5::DbUtils::update_row(dbh => $self->{dbh},
                                table_name => $self->db_table,
                                rec => {flg_in_classification => 0},
                                where_fields => {id => $params{existing_obj}->{url_id}});
    }

    return $rv;
}

#----------------------------------------------------------------------

sub is_pure_wildcard {
    my ($self, %params) = @_;
    return ($self->get_new_value(%params, field_name => 'url_name') &&
            $self->get_new_value(%params, field_name => 'url_name') eq "*" &&
            $self->get_new_value(%params, field_name => 'method_name') &&
            $self->get_new_value(%params, field_name => 'method_name') eq "*" &&
            $self->get_new_value(%params, field_name => 'is_wildcard')); 
}

#----------------------------------------------------------------------

sub map_fields_to_entity {
  my($class, %params) = @_;

  $class->SUPER::map_fields_to_entity(%params);

  my $rec = $params{rec};
  if (defined($rec->{allowed_origins_enforcement_mode}) && $rec->{allowed_origins_enforcement_mode} == F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_UNMODIFIED()) {
      $rec->{flg_allowed_origins_check} = 0;
  }
  else {
      $rec->{flg_allowed_origins_check} = 1;
  }
}

#----------------------------------------------------------------------

sub get_entity_changes_string {
    my ($class_or_self, %params) = @_;

    if ($params{for_suggestion}) { # When for_suggestion is true, we always receive self instead of class
        if ( defined $class_or_self->{flg_in_classification} && $class_or_self->{flg_in_classification} == 0 ){
            return "End Classification";
        }

        my $allowed_origins = defined $class_or_self->{allowed_origins} && ref($class_or_self->{allowed_origins}) eq 'ARRAY' ? $class_or_self->{allowed_origins} : undef;
        if ( $allowed_origins && !defined $class_or_self->{allowed_origins_enforcement_mode} ) {
            $class_or_self->{allowed_origins_enforcement_mode} = F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_ENFORCE();
        }
    }

    return $class_or_self->SUPER::get_entity_changes_string(%params);
}
#----------------------------------------------------------------------

sub get_default_url_content_profiles_db_rec {
  my ($class_or_self, %params) = @_;

  my $default_xml_profile_id = $params{default_xml_profile_id};
  my $default_json_profile_id = $params{default_json_profile_id};

  my $default_url_content_profiles = [ {  header_name  => '*',
                                          header_value => '*',
                                          enforcement_type => 'apply_value_signatures',
                                          header_order => 0,
                                       },

                                       {  header_name  => 'Content-Type',
                                          header_value => '*form*',
                                          enforcement_type => 'http',
                                          header_order => 1,
                                       },

                                       {  header_name  => 'Content-Type',
                                          header_value => '*json*',
                                          enforcement_type => 'json',
                                          profile_id   => $default_json_profile_id,
                                          header_order => 2,
                                       },

                                       {  header_name  => 'Content-Type',
                                          header_value => '*xml*',
                                          enforcement_type => 'xml',
                                          profile_id   => $default_xml_profile_id,
                                          header_order => 3,
                                       },
                                     ];

  return $default_url_content_profiles;
}

#----------------------------------------------------------------------

sub clone_to_explicit {
    my ($self, %params) = @_;

    my $call_flags = $params{call_flags} || {};

    my $entity = $self->SUPER::clone_to_explicit(%params);

    if ($call_flags->{suggestion_rest_uuid} && ref($entity->{url_content_profiles}) eq 'ARRAY') {
        my $flg_in_classification = $self->url_classification_flag(%params);
        if ($flg_in_classification) {
            delete $entity->{url_content_profiles};
        }
    }

    return $entity;
}

#----------------------------------------------------------------------

sub url_classification_flag {
    my ($self, %params) = @_;
    my $policy = $params{policy} || $self->{policy};

    my $flg_in_classification = 0;
    ($flg_in_classification) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy->{policy_id},1);
SELECT flg_classify_urls
  FROM PLC.PL_POLICY_BUILDER_CONFIG_URL pb_url,
       PLC.PL_POLICY_BUILDER_CONFIG pb
 WHERE pb.policy_id = ?
   AND pb.policy_id = pb_url.policy_id
   AND pb.flg_learning = ?
EndQuery

    return $flg_in_classification;
}

#----------------------------------------------------------------------

sub get_full_display_name {
    my ($self, %params) = @_;
    my $policy = $params{policy} || $self->{policy};

    if ($policy) {
        my $is_protocol_free = $policy->{flg_is_protocol_free};

        if ( !defined $is_protocol_free ) {
            $is_protocol_free =  $self->{dbh}->selectrow_array(<<EndQuery,undef,$policy->{policy_id});
SELECT flg_is_protocol_free
  FROM PLC.PL_POLICIES
 WHERE id = ?
EndQuery
        }

        my @display_name_elements;
        if (! $is_protocol_free ) {
            my $app_id = $self->{app_id} || $self->entity_config->{fields}->{app_id}->{default_value};
            my $protocol_name = $self->entity_config->{fields}->{app_id}->{display_value_map}->{ $app_id };
            push @display_name_elements, "[$protocol_name]";
        }
        if ($self->{method_name} && $self->{method_name} ne $ANY_URL_METHOD_NAME) {
            push @display_name_elements, $self->{method_name};
        }
        push @display_name_elements, $self->{url_name};

        return join(' ', @display_name_elements);
    }

    return $self->SUPER::get_full_display_name(%params);
}
#----------------------------------------------------------------------

sub from_declarative {
    my ($self, %params) = @_;

    my $object = $params{object};
    my $field_name = $params{field_name};

    my $rv;
    if ($field_name eq 'url_content_profiles') {
        for my $type ( qw(xml json gwt graphql grpc) ) {
            if ($object->{type} eq $type) {
                if (! defined $object->{contentProfile}) {
                    $object->{contentProfile} = { name => $object->{name} };
                    delete $object->{name};
                }
                $rv = { policy_reference => "$type-profiles" };
                last;
            }
        }
    }

    return $rv;
}

#----------------------------------------------------------------------

1;
