package F5::ASMConfig::Entity::UrlContentProfile;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::Entity::Policy;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::ContentProfileBase;
use List::Util qw(max);
use base qw/F5::ASMConfig::Entity::Base F5::ASMConfig::UrlContentProfile/;

use F5::CRC;

#----------------------------------------------------------------------------

sub get_full_display_name {
    my ($self, %params) = @_;
    my $header_name = $self->get_field_display_string(%params, field_name => 'header_name', dbh => $self->{dbh});
    my $header_value = $self->get_field_display_string(%params, field_name => 'header_value', dbh => $self->{dbh});
    my $enforcement_type = $self->get_field_display_string(%params, field_name => 'enforcement_type', dbh => $self->{dbh});

    my $display_name = "$header_name:$header_value - Type: $enforcement_type";

    if ($self->{content_profile}) {
        $display_name .= " - Profile '" . $self->get_field_display_string(%params,
                                                                          field_name => 'content_profile',
                                                                          dbh => $self->{dbh})
                                        . "'";
    }

    return $display_name;
}

#----------------------------------------------------------------------------

sub is_default_url_content_profile {
    my ($class_or_self, %params) = @_;
    my $object = $params{object} || $class_or_self;   

    return $object->{header_name} eq '*' && $object->{header_value} eq '*';
}

#----------------------------------------------------------------------------

sub is_deletable {
    my ($self, %params) = @_;
    if ($params{call_flags} && $params{call_flags}->{rest_call} && !$params{call_flags}->{is_in_diff_merge} && $self->is_default_url_content_profile) {
        my $error_message = 'The default URL Content Profile (*:*) is mandatory.';
        return { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
               };
    }
    return $self->SUPER::is_deletable(%params);
}

#----------------------------------------------------------------------------

sub delete {
    my ($self, %params) = @_;
    if ($self->{enforcement_type} == F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON() ){
        $self->validate_url_json_profiles(url => $self->{url});
    }
    return $self->SUPER::delete(%params);
}

#----------------------------------------------------------------------------

sub before_store {
  my($self, %params) = @_;

  my $call_flags = $params{call_flags} || {};
  my $dbh = $params{dbh} || $self->{dbh};
  my $policy = $params{policy} || $self->{policy};
  my $url = $self->get_new_value(%params, field_name => 'url');
  my $existing_obj = $params{existing_obj};
  my $rv = {};

  if (! $url || ! ref($url) || ! $url->{url_id}) {
      my $error_message = 'A valid URL must be specified';
      die { error_message => $error_message,
            display_message => $error_message,
            error_level     => 'info',
            error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
          };
  };

  my $store_obj = $params{store_obj};
  my $header_order = $self->get_new_value(%params, field_name => 'header_order');
  my $header_name = $self->get_new_value(%params, field_name => 'header_name');
  my $header_value = $self->get_new_value(%params, field_name => 'header_value');

  if ($header_name eq '*' && $header_value eq '*') {
      $store_obj->{header_order} = 0;
  }
  elsif ($params{validate_value} && $header_name =~ /([?*])/) {
      my $invalid_char = $1;
      my $field_display_name = $self->get_field_display_name_string(%params, field_name => 'header_name');
      my $error_message = "$field_display_name '$header_name' can not contain wildcard character '$invalid_char'.";
      die { error_message => $error_message,
            display_message => $error_message,
            error_level     => 'info',
            error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
          };
  }
  elsif ( !$header_order) {
      $store_obj->{header_order} = 1;
  }

  if ($self->has_value_changed(%params, check_fields => 'header_order')) {
      $self->order_field_adjust(%params,
                                order_field_name => 'header_order',
                                where_values => {
                                                 'obj_id' => $url->{url_id},
                                                },
                               );
  }

  if ($existing_obj->{enforcement_type} &&
      $existing_obj->{enforcement_type} == F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON() &&
      $self->has_value_changed(%params, check_fields => "enforcement_type")) {
          $self->validate_url_json_profiles(url => $url);
  }


  my %content_profile_types = (F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_XML()      => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(),
                               F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON()     => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),
                               F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GWT()      => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GWT(),
                               F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRAPHQL()  => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRAPHQL(),
                               F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRPC()     => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRPC(),
                              );
  my $enforcement_type = $self->get_new_value(%params, field_name => 'enforcement_type');
  my $content_profile = $self->get_new_value(%params, field_name => 'content_profile');
  my @modified_entities;
  if ($content_profile_types{$enforcement_type} &&
      (! $content_profile || ! $content_profile->{profile_id} )) {
      my $default_content_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                       dbh => $dbh,
                                                                       policy => $policy,
                                                                       created_by => 'PolicyBuilder',
                                                                       content_profile_type => $content_profile_types{$enforcement_type},
                                                                   );
      if ($default_content_profile->{__new_entity}) {
          $default_content_profile->store_add(validate_value => $params{validate_value});
          push(@modified_entities, {event_type => F5::ASMConfig::EventType::ADD, entity => $default_content_profile, existing_entity => $default_content_profile});
      }
      $store_obj->{content_profile} = $default_content_profile->uncast();
  }

  if ( (!$enforcement_type || ! $content_profile_types{$enforcement_type}) && 
      $content_profile) {

      $params{set_fields_to_null}->{content_profile} = 1;
      delete $store_obj->{content_profile};
  }
  $rv->{modified_entities} = \@modified_entities;
  return $rv;
}

#----------------------------------------------------------------------

sub get_missing_external_entities {
    my ($self, %params) = @_;
    my %content_profile_names = (F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_XML() => 'xml',
                                 F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON() => 'json',
                                 F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GWT() => 'gwt',
                                 F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRAPHQL() => 'graphql',
                                 F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRPC() => 'grpc',
                                );
    my %content_profile_types = (F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_XML() => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(),
                                 F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON() => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),
                                 F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GWT() => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GWT(),
                               F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRAPHQL()  => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRAPHQL(),
                                 F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRPC() => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRPC(),
                                );

    if ((my $profile_type_name = $content_profile_names{$self->{enforcement_type}}) && !$self->{content_profile}) {
        my $default_content_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                         dbh => $self->{dbh},
                                                                         policy => $self->{policy},
                                                                         created_by => 'PolicyBuilder',
                                                                         content_profile_type => $content_profile_types{$self->{enforcement_type}},
                                                                     );
        $self->{content_profile} = $default_content_profile;
    }
    return $self->SUPER::get_missing_external_entities(%params);
}

#----------------------------------------------------------------------
 sub validate_url_json_profiles {
    my ($self, %params) = @_;

    my $url = $params{url};
    my $json_login_page_url_iter = F5::ASMConfig::Entity::LoginObject->retrieve_all(dbh => $self->{dbh},
                                                                               get_criteria => { url__url_id => $url->{url_id} },
                                                                               search_filters => [F5::ASMConfig::SearchFilter->new(field_name => 'auth_type',
                                                                                                                                   field_value => F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_JSON(),
                                                                                                                                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ())
                                                                                                 ],
                                                                               );

   if ($json_login_page_url_iter->{total_record_count}){
        my $url_content_profiles_iter = F5::ASMConfig::Entity::UrlContentProfile->retrieve_all(dbh => $self->{dbh},
                                                                                              get_criteria => { url__url_id => $url->{url_id} },
                                                                                              search_filters => [ F5::ASMConfig::SearchFilter->new(field_name => 'enforcement_type',
                                                                                                                                                   field_value => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON(),
                                                                                                                                                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ())
                                                                                                                ,
                                                                                                                  F5::ASMConfig::SearchFilter->new(field_name => 'id',
                                                                                                                                                   field_value => $self->{id},
                                                                                                                                                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE())
                                                                                                                ],

                                                                                              );

       if (!$url_content_profiles_iter->{total_record_count}){
              my $error_message =  "URL is defined as JSON login page, JSON profile must be assigned to the URL";
              die { error_message => $error_message,
                    display_message => $error_message,
                    error_level     => 'info',
                    error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                  };
       }

   }
   return;
}
#----------------------------------------------------------------------

sub update_parent_staging {
    my ($self, %params) = @_;
    ## Update parent URL
    if (ref $self) { #Nothing should be using update_many_url_content_profiles, but just in case.
        my $url = $self->get_new_value(%params, field_name => 'url');
        $url = F5::ASMConfig::Entity::URL->new(dbh => $self->{dbh},
                                               policy => $self->{policy},
                                               object => $url,); #In case it is a Thrift object and not an Entity
        return $url->store_update(store_obj => {last_updated => time()}, 
                                  validate_value => 0);
    }
    return;
}

#----------------------------------------------------------------------

sub update_default_object_by_pb_settings {
    my ($self, %params) = @_;

    $self->SUPER::update_default_object_by_pb_settings(%params);

    my $default_obj = $params{default_obj};
    my $policy = $params{policy};
    my $dbh = $params{dbh} || $self->{dbh};

    my($is_pb_automatic) = $dbh->selectrow_array(<<EndQuery, undef, $policy->{policy_id});
SELECT flg_enabled
  FROM PLC.PL_POLICY_BUILDER_CONFIG
 WHERE policy_id = ?
EndQuery

    if ($is_pb_automatic) {
        $default_obj->{enforcement_type} = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_APPLY_VALUE_SIGNATURES();
    }
    else {
        $default_obj->{enforcement_type} = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_HTTP();
    }
}

#----------------------------------------------------------------------

1;
