package F5::ASMConfig::Entity::Webhook;

use strict;
use warnings;
use F5::LWP::UserAgent;
use F5::Cluster::Utils;
use F5::BigipVersionUtils;

use base qw(F5::ASMConfig::Entity::Base F5::ASMConfig::Webhook);

#----------------------------------------------------------------------

sub before_store {
    my ($self, %params) = @_;

    my $store_obj = $params{store_obj};

    if ($store_obj->{body}) {
        $self->parse_placeholders(body => $store_obj->{body});
    }
}

#----------------------------------------------------------------------

sub send {
    my ($self, %params) = @_;

    my $logger = $self->{logger};
    my $url = $self->{webhook_url};
    my $body = $self->get_webhook_body(%params);
    my $content_type = $self->{content_type};

    # Create a request
    my $request = HTTP::Request->new('POST', $url);
    $request->header('Content-Type' => $content_type);
    $request->content($body);

    # Create a user agent object
    my $user_agent = F5::LWP::UserAgent->new;

    # Send request
    my $rv = $user_agent->request($request);

    # Check the outcome of the response
    if (!$rv->is_success && $logger) {
        $logger->info("Failed to send webhook request: ");
        $logger->info($rv->status_line);
        $logger->info("[url = $url, body = $body, content-type = $content_type]");
    }

    return $rv;
}

#----------------------------------------------------------------------

my %general_fields = (
                      device => {
                                 hostname => undef,
                                 version => F5::BigipVersionUtils::bigip_version(),
                                }
                     );

sub get_webhook_body {
    my ($self, %params) = @_;

    my $element = $params{element};
    my $body = $self->{body};
    my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh}, object => $params{policy});

    my $entity_type = $self->get_cached_element_type(element_type => $element->{element_type})->{class};
    my $webhook_entity = $entity_type->new(dbh => $self->{dbh}, object => $element->{$entity_type->element_type_name});

    my $webhook_fields = $entity_type->webhook_fields;
    my $entity_namespace = $entity_type->element_rest_name;

    my $webhook_entity_retrieved = 0;
    my @placeholders = $body =~ /\{\{(.*?)\}\}/g;
    for my $placeholder (@placeholders) {
        my ($namespace, $field) = $placeholder =~ /(.*)\.(.*)/;

        if (!$namespace || !$field) {
            next;
        }

        # trim placeholder parts
        $namespace =~ s/^\s+|\s+$//g;
        $field =~ s/^\s+|\s+$//g;

        my $webhook_fields = {};
        my $entity;
        my $entity_field;

        if ($general_fields{$namespace}) {
            if (!defined $general_fields{$namespace}->{$field} && $placeholder =~ 'device.hostname') {
                $general_fields{device}->{hostname} = F5::Cluster::Utils::get_mcp_val('hostname');
            }
            my $value = $general_fields{$namespace}->{$field};
            $body =~ s/\{\{$placeholder\}\}/$value/gi;
            next;
        }
        elsif ($namespace eq 'policy') {
            $webhook_fields = F5::ASMConfig::Entity::Policy->webhook_fields;
            $entity = $policy;
            $entity_field = F5::ASMConfig::Entity::Policy->reverse_webhook_field_map->{$field};
        }
        elsif ($namespace eq $entity_namespace) {
            $webhook_fields = $entity_type->webhook_fields;
            $entity = $webhook_entity;
            $entity_field = $entity_type->reverse_webhook_field_map->{$field};
        }
        else {
            next;
        }

        if (!defined $entity->{$entity_field} && !$webhook_entity_retrieved) {
            my $retrieved_entity = $entity_type->new(dbh => $self->{dbh},
                                                     get_criteria => $element->{$entity_type->element_type_name}, );

            if (!$retrieved_entity) {
                $self->{logger}->info("Failed to retrieve $placeholder for webhook id $self->{webhook_id}") if $self->{logger};
                next;
            }

            $webhook_entity = $retrieved_entity;
            $entity = $retrieved_entity;
            $webhook_entity_retrieved = 1;
        }

        my $value;
        if ($webhook_fields->{$entity_field}->{external_struct}) {
            $value = $entity->get_field_display_string(dbh => $self->{dbh},
                                                       field_name => $entity_field);
        }
        else {
            $value = $entity->map_rest_field_value(dbh => $self->{dbh},
                                                   field_name => $entity_field,
                                                   field_value => $entity->{$entity_field},
                                                   field_def => $webhook_fields->{$entity_field},
                                                   direction => 'to_rest');
        }

        $body =~ s/\{\{$placeholder\}\}/$value/gi;
    }

    return $body;
}

#----------------------------------------------------------------------

sub parse_placeholders {
    my ($self, %params) = @_;

    my $body = $params{body};
    my @rest_entities = map { $self->full_schema->{$_}->{entity_type} } grep { $self->full_schema->{$_}->{rest_paths} } (keys %{$self->full_schema});
    my %entities_namespaces = map { $_->element_rest_name => $_ } @rest_entities;

    my $placeholders_fields;
    my @placeholders = $body =~ /\{\{(.*?)\}\}/g;
    for my $placeholder (@placeholders) {
        my ($namespace, $field) = $placeholder =~ /(.*)\.(.*)/;

        if (!$namespace || !$field) {
            next;
        }

        # trim placeholder parts
        $namespace =~ s/^\s+|\s+$//g;
        $field =~ s/^\s+|\s+$//g;

        if (!$placeholders_fields->{$namespace}) {
            if ($general_fields{$namespace}) {
                $placeholders_fields->{$namespace} = $general_fields{$namespace};
            }
            elsif (my $entity_type = $entities_namespaces{$namespace}) {
                my %webhook_fields = map { $_->{webhook_field_name} => 1 } values %{$entity_type->webhook_fields};
                $placeholders_fields->{$namespace} = \%webhook_fields;
            }
        }

        if ( !$placeholders_fields->{$namespace} || !exists $placeholders_fields->{$namespace}->{$field} ) {
            my $error_message = "$namespace.$field is not a supported field.";
            die { error_message => $error_message,
                  display_message => $error_message,
                  error_level     => 'info',
                  error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                };
        }
    }
}

#----------------------------------------------------------------------

1;
