package F5::ASMConfig::Entity::WebsocketUrl;

use strict;
use warnings;

use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::Entity::Policy;
use F5::ASMConfig::Entity::ContentProfile;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::PlainTextProfile;

use base qw(F5::ASMConfig::CorsBase F5::ASMConfig::WebsocketUrl);

#----------------------------------------------------------------------

my $CONTENT_PROFILE_FIELDS = [ { payload_flag_name => 'flg_websocket_message_format_json',
                                 payload_profile_field => 'websocket_json_profile',
                               },
                               { payload_flag_name => 'flg_websocket_message_format_text',
                                 payload_profile_field => 'websocket_plain_text_profile',
                               }
                              ];

sub before_store {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags = $params{call_flags};
    my $policy = $params{policy} || $self->{policy};
    my $rv;
    my @modified_entities;

    my @fields = qw(
                   enable_cors_enforcement
                   flg_allow_cors
                   flg_allowed_origins_check
                   allowed_origins
                   flg_allowed_headers_check
                   flg_exposed_headers_check
                   flg_check_credentials
                   check_maximum_age
                   maximum_age
                   flg_is_allowed
                   flg_in_classification
                 );

    my %field_values;
    my %field_display_name;
    for my $field (@fields) {
        $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                      existing_obj => $existing_obj,
                                                      call_flags => $call_flags,
                                                      field_name => $field, );
        $field_display_name{$field} = $self->get_field_display_name_string(%params, field_name => $field);
    }

    if (defined($field_values{flg_is_allowed}) && ! $field_values{flg_is_allowed} ) {
        $store_obj->{flg_check_payload} = 0;
        $store_obj->{flg_websocket_message_format_text} = 0;
        $field_values{flg_check_payload} = 0;
        $field_values{flg_websocket_message_format_text} = 0;

        if ( $self->is_pure_wildcard(%params, existing_obj => $existing_obj, store_obj => $store_obj) ){
            my $error_message = "Pure wildcard must be allowed";
            die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
            };
        }


    }

    if ( $params{store_mode} eq 'add' ) {
        my $flg_in_classification;
        if ($self->is_pure_wildcard(%params)) {
            $flg_in_classification = 0;
        }
        else {
            ($flg_in_classification) = $self->websocket_classification_flag(%params);
        }

        #if this is being added due to a suggestion and PB is at least in manual learning mode, inherit the classification state from PB_CONFIG
        if ( $call_flags->{suggestion_rest_uuid} ) {
            $store_obj->{flg_in_classification} = $flg_in_classification;
            $store_obj->{flg_check_payload} = $store_obj->{flg_in_classification} ? 0 : 1;
        }
        elsif ( $call_flags->{rest_call} ) {
            $store_obj->{flg_in_classification} &&= $flg_in_classification;
        }

        if ( $call_flags->{suggestion_rest_uuid} && !$call_flags->{cloned_from_parent} ) {
            $store_obj->{flg_websocket_message_format_text} = $flg_in_classification ? 0 : 1;
            $self->{flg_websocket_message_format_text} = $flg_in_classification ? 0 : 1;
        }
        elsif( $store_obj->{flg_in_classification} ) {
            $store_obj->{flg_check_payload} = 0;
            $store_obj->{flg_websocket_message_format_text} = 0;
            $self->{flg_websocket_message_format_text} = 0;
            $store_obj->{flg_websocket_message_format_json} = 0;
            $self->{flg_websocket_message_format_json} = 0;
            $store_obj->{flg_websocket_message_format_binary} = 0;
            $self->{flg_websocket_message_format_binary} = 0;
        }
    }

    #TODO: in case changing format flags from suggestion should remain "in classification" flag.
    #      need to test specific suggestion "set ON text flag and set OFF in classification" 
    if ( $params{store_mode} ne 'add' &&
         ! $call_flags->{suggestion_rest_uuid} &&
         (!$params{session}->{user_name} ||
           $params{session}->{user_name} ne $F5::ASMConfig::Entity::Base::BIGIQ_USER_NAME) &&
         $self->has_value_changed(%params,
                                  check_fields => [ qw(flg_websocket_message_format_text
                                                       flg_websocket_message_format_json
                                                       flg_websocket_message_format_binary
                                                       message_max_frame_size
                                                       message_max_frames
                                                       flg_check_payload
                                                       websocket_extension
                                                       flg_in_classification) ] ) ) {
        $store_obj->{flg_in_classification} = 0;
    }

    $field_values{flg_check_payload} = $self->get_new_value( store_obj => $store_obj,
                                                             existing_obj => $existing_obj,
                                                             call_flags => $call_flags,
                                                             field_name => 'flg_check_payload', );

    if ($params{validate_value} && $field_values{flg_is_allowed} && $field_values{flg_check_payload}
        && $self->has_value_changed(%params, check_fields => [ qw(flg_check_payload 
                                                                  flg_websocket_message_format_text 
                                                                  flg_websocket_message_format_json
                                                                  flg_websocket_message_format_binary 
                                                                  websocket_json_profile 
                                                                  websocket_plain_text_profile) ])) {
        my $any_payload_active = 0;
        for my $payload_flag_field (qw/flg_websocket_message_format_text flg_websocket_message_format_json flg_websocket_message_format_binary/) {
            if ($self->get_new_value( %params, field_name => $payload_flag_field, )) {
                $any_payload_active = 1;
                last;
            }
        }

        if (! $any_payload_active) {
            my $payload_display_name = $self->get_field_display_name_string(%params, field_name => 'flg_check_payload');
            my $error_message = "At least one payload type should be active when $payload_display_name is enabled.";
            die { error_message => $error_message,
                  display_message => $error_message,
                  error_level     => 'info',
                  error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                };
        }
    }

    for my $content_profile_fields (@$CONTENT_PROFILE_FIELDS) {
        my $payload_flag = $self->get_new_value(%params, field_name => $content_profile_fields->{payload_flag_name});
        my $payload_profile = $self->get_new_value(%params, field_name => $content_profile_fields->{payload_profile_field});
        if ($field_values{flg_check_payload} && $payload_flag && !$payload_profile) {
            my $profile_class = $self->fields->{$content_profile_fields->{payload_profile_field}}->{entity_type};
            my $default_content_profile = $profile_class->get_default_content_profile(
                                                                       dbh => $self->{dbh},
                                                                       policy => $policy,
                                                                       created_by => 'PolicyBuilder',
                                                                   );
            $store_obj->{$content_profile_fields->{payload_profile_field}} = $default_content_profile;
            if ($default_content_profile->{__new_entity}) {
                $default_content_profile->store_add(validate_value => $params{validate_value});
                push(@modified_entities, {event_type => F5::ASMConfig::EventType::ADD, entity => $default_content_profile, existing_entity => $default_content_profile});
            }
        }
        elsif ($payload_profile && ((!$field_values{flg_check_payload} && $self->has_value_changed(%params, check_fields => 'flg_check_payload')) ||
                       (!$payload_flag && $self->has_value_changed(%params, check_fields => [$content_profile_fields->{payload_flag_name}]))
                      )) {
            # disassociate from content profile only if value of flg_check_payload/payload flag changed to disabled
            $params{set_fields_to_null}->{$content_profile_fields->{payload_profile_field}} = 1;
            delete $store_obj->{$content_profile_fields->{payload_profile_field}};
        }
    }

    $self->handle_cors(%params, field_values => \%field_values, field_display_name => \%field_display_name);

    $self->adjust_wildcards(%params);
    $self->sync_from_pb_config(%params, 
                               pb_config_field => "flg_websocket_urls", 
                               table_name => F5::ASMConfig::Entity::PBConfigUrl->db_table(),);

    $rv->{modified_entities} = \@modified_entities;
    return $rv;
}

#----------------------------------------------------------------------

sub validate_payload_with_profile {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $call_flags = $params{call_flags};
    my $payload_flag_field = $params{payload_flag_name};
    my $payload_profile_field = $params{payload_profile_field};
    my $profile_type_class = $params{profile_type_class};

    if ( $self->has_value_changed(%params, check_fields => [ $payload_flag_field ]) ) {
        my $payload_flag = $self->get_new_value( store_obj => $store_obj,
                                              existing_obj => $existing_obj,
                                              call_flags => $call_flags,
                                              field_name => $payload_flag_field, );

        if ($payload_flag) {
            my $payload_profile = $self->get_new_value( store_obj => $store_obj,
                                                     existing_obj => $existing_obj,
                                                     call_flags => $call_flags,
                                                     field_name => $payload_profile_field, );
            my $validation_fail = 0;
            if ( ! $payload_profile ) {
                $validation_fail = 1;
            }
            elsif( $call_flags->{rest_call} ) {
                my $profile = $profile_type_class->new(%params,
                                                       dbh => $self->{dbh},
                                                       get_criteria => {
                                                           profile_name => $payload_profile->{profile_name},
                                                       });
                if ($profile) {
                    $store_obj->{$payload_profile_field} = $profile;
                }
                else {
                    $validation_fail = 1;
                }
            }

            if ($validation_fail) {
                my $payload_profile_display_name = $self->get_field_display_name_string(%params, field_name => $payload_profile_field);
                my $payload_flag_display_name = $self->get_field_display_name_string(%params, field_name => $payload_flag_field);
                my $content_profile_error_message = "$payload_profile_display_name is required when $payload_flag_display_name is enabled.";
                die { error_message => $content_profile_error_message,
                      display_message => $content_profile_error_message,
                      error_level     => 'info',
                      error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                    };
            }
        }
    }
}

#----------------------------------------------------------------------

sub get_missing_external_entities {
    my ($self, %params) = @_;

    my $existing_entity = $params{existing_entity};
    my $call_flags = $params{call_flags};

    my @missing_external_entities;

    my $pb_flg_in_classification;
    if ($self->is_pure_wildcard(%params)) {
        $pb_flg_in_classification = 0;
    }
    else {
        ($pb_flg_in_classification) = $self->websocket_classification_flag(%params);
    }

    if ( (! $existing_entity && ! $pb_flg_in_classification) ||
         ($existing_entity && $self->{flg_websocket_message_format_text} && ! $self->{websocket_plain_text_profile}) ) {
        if (! $call_flags->{cloned_from_parent}) {
            $self->{flg_websocket_message_format_text} = 1;
            my $text_content_profile = F5::ASMConfig::Entity::PlainTextProfile->get_default_content_profile(
                                                                                dbh => $self->{dbh},
                                                                                policy => $self->{policy},
                                                                                created_by => 'PolicyBuilder',
                                                                            );

            if ($text_content_profile) {
                $self->{websocket_plain_text_profile} = $text_content_profile;
                if ($text_content_profile->{__new_entity}) {
                    push (@missing_external_entities, $text_content_profile);
                }
            }
        }
    }

    if ( $pb_flg_in_classification &&
         $self->{flg_websocket_message_format_json} &&
         ! $self->{websocket_json_profile}) {
        if (! $call_flags->{cloned_from_parent}) {
            my $json_content_profile = F5::ASMConfig::Entity::JsonProfile->get_default_content_profile(
                                                                                dbh => $self->{dbh},
                                                                                policy => $self->{policy},
                                                                                created_by => 'PolicyBuilder',
                                                                            );

            if ($json_content_profile) {
                $self->{websocket_json_profile} = $json_content_profile;
                if ($json_content_profile->{__new_entity}) {
                    push (@missing_external_entities, $json_content_profile);
                }
            }
        }
    }

    return \@missing_external_entities;
}

#----------------------------------------------------------------------

sub is_pure_wildcard {
    my ($self, %params) = @_;
    return ($self->get_new_value(%params, field_name => 'websocket_url_name') &&
            $self->get_new_value(%params, field_name => 'websocket_url_name') eq "*" &&
            $self->get_new_value(%params, field_name => 'is_wildcard'));
}

#----------------------------------------------------------------------

sub get_missing_default_text_profile {
    my ($class, %params) = @_;

    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $store_obj = $params{store_obj};

    my $default_profile = F5::ASMConfig::Entity::PlainTextProfile->get_default_content_profile(
                                                                      dbh => $dbh,
                                                                      policy => $policy,
                                                                      created_by => 'PolicyBuilder',
                                                                   );

    if ($store_obj) {
        $store_obj->{websocket_plain_text_profile} = $default_profile;
    }

    return $default_profile;
}

#----------------------------------------------------------------------

sub websocket_classification_flag {
    my ($self, %params) = @_;
    my $policy = $params{policy} || $self->{policy};

    my $flg_in_classification = 0;
    ($flg_in_classification) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy->{policy_id},1);
SELECT flg_classify_websocket_urls
  FROM PLC.PL_POLICY_BUILDER_CONFIG_URL pbcu
    JOIN PLC.PL_POLICY_BUILDER_CONFIG pbc
    ON pbcu.policy_id = pbc.policy_id
  WHERE pbc.policy_id = ?
  AND flg_learning = ?
EndQuery

    return $flg_in_classification;
}

#----------------------------------------------------------------------
sub get_full_display_name {
    my ($self, %params) = @_;
    my $policy = $params{policy} || $self->{policy};
    my $is_protocol_free = $policy->{flg_is_protocol_free};

    if ( !defined $is_protocol_free ){
    $is_protocol_free =  $self->{dbh}->selectrow_array(<<EndQuery,undef,$policy->{policy_id});
SELECT flg_is_protocol_free
  FROM PLC.PL_POLICIES
 WHERE id = ?
EndQuery
    }

    if ( $is_protocol_free ) {
        return $self->{websocket_url_name};
    }

    return $self->SUPER::get_full_display_name(%params);
}
#----------------------------------------------------------------------
1;
