package F5::ASMConfig::Entity::WhitelistIp;
use base qw(F5::ASMConfig::Entity::Base F5::ASMConfig::WhitelistIp);


use strict;
use warnings;

use F5::IpAddr;

my $default_ipv4_mask = '255.255.255.255';
my $default_ipv6_mask = 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff';

#----------------------------------------------------------------------

sub before_store {
    my($self, %params) = @_;

    my $store_mode = $params{store_mode};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};

    return unless $params{validate_value};

     my @fields = qw(
                   ip_address
                   mask 
                  );

   my %field_values;
   my %field_display_name;

   for my $field (@fields) {
      $field_values{$field} = $self->get_new_value( store_obj => $store_obj,
                                                    existing_obj => $existing_obj,
                                                    field_name => $field, );

      $field_display_name{$field} = $self->get_field_display_name_string(%params,field_name => $field);
   }

   my $ip_address_is_valid_ipv4 = $self->is_valid_ipv4( field_value => $field_values{ip_address})->{completed_successfully};
   my $mask_is_valid_ipv4 = $self->is_valid_ipv4( field_value => $field_values{mask})->{completed_successfully};

   if ( $ip_address_is_valid_ipv4 ){
       if ( !$mask_is_valid_ipv4  ){
          my $error_message = "$field_display_name{mask} is not compatible with $field_display_name{ip_address}";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
          }; 
       }
   }
   elsif ( $mask_is_valid_ipv4 ){
       if ( $field_values{mask} == $default_ipv4_mask ){
           $store_obj->{mask} = $default_ipv6_mask;
           $field_values{mask} = $default_ipv6_mask;
       }
       else {
          my $error_message = "$field_display_name{mask} is not compatible with $field_display_name{ip_address}";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
          };
       }
   }

   if ($params{store_mode} eq 'add') {
       my $consistent = $self->is_ip_mask_consistent(ip_address => $field_values{ip_address}, mask => $field_values{mask});
       if (! $consistent) {
          my $error_message = "$field_display_name{ip_address} and $field_display_name{mask} combination is not valid";
          die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
          };
       }
   }

   if ($self->has_value_changed(%params, check_fields => 'flg_never_block') &&
       !($self->has_value_changed(%params, check_fields => 'block_ip'))) {
       $store_obj->{block_ip} = $store_obj->{flg_never_block} ? F5::ASMConfig::BlockIpType::BLOCK_IP_NEVER()
                                                              : F5::ASMConfig::BlockIpType::BLOCK_IP_POLICY_DEFAULT();
   }
}

#----------------------------------------------------------------------

sub is_ip_mask_consistent {
    my ($class_or_self, %params) = @_;

    my $ip_addr = $params{ip_address};
    my $mask = $params{mask};

    $ip_addr =~ s/\%.*//; #remove route domain
    my $ip_bin = F5::IpAddr::ip_pton($ip_addr);
    my $mask_bin = F5::IpAddr::ip_pton($mask);

    my $pref_up = $ip_bin | $mask_bin;
    my $mask_inv = ~$mask_bin;
    my $res_bin = $pref_up & $mask_inv;

    my @res_set = unpack("N*", $res_bin);
    my $consistent = 1;
    for my $res_entry (@res_set) {
        if ($res_entry > 0) {
            $consistent = 0;
        }
    }

    return $consistent;
}

#----------------------------------------------------------------------

sub retrieve_all {
    my ($class, %params) = @_;

    if ( ! grep {$_->{field_name} eq 'sort_len_and_num'} @{$params{sort_orders}}) {
        return $class->SUPER::retrieve_all(%params);
    }

    my @whiteListIps_ns = $class->SUPER::retrieve_all(%params);
    my @whiteListIps = sort {
        length($b->{ip_address}) <=> length($a->{ip_address}) ||
        $a->{ip_address} <=> $b->{ip_address}
    } @whiteListIps_ns;

    my $total_record_count = scalar @whiteListIps;
    $class->apply_paging(result_set => \@whiteListIps, %params);

    if (wantarray) {
        return @whiteListIps;
    }
    else {
        my $iterator = F5::ASMConfig::Entity::Iterator->new(
                                                            get_next_function => sub {return shift(@whiteListIps);},
                                                            get_all_function => sub {return @whiteListIps;},
                                                            total_record_count => $total_record_count,
                                                           );
        return $iterator;
    }
}

#----------------------------------------------------------------------

1;
