package F5::ASMConfig::Entity::XmlProfile;

use strict;
use warnings;

use Memoize;

use F5::ASMConfig::ASMConfig;
use F5::DbUtils;
use F5::ASMConfig::Entity::XmlDefenseAttribute;
use F5::ASMConfig::ContentProfileBase;
use base qw/F5::ASMConfig::ContentProfileBase F5::ASMConfig::XmlProfile/;

#------------------------------------------------------------------------

sub before_store {
    my ($self, %params) = @_;
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $store_mode = $params{store_mode};
    my $call_flags = $params{call_flags} || {};

    # can't have soap_methods when adding from rest
    if ($call_flags->{rest_call} && $store_mode eq 'add') {
        delete $store_obj->{soap_methods};
    }

    # when updating soap_methods from rest we are only allowed to toggle flg_allowed in existing methods
    if ($call_flags->{rest_call} && $store_mode eq 'update' && defined($store_obj->{soap_methods})) {
        my $nested_def = $self->fields->{soap_methods};
        my $store_methods = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                              nested_recs => $store_obj->{soap_methods},
                                                             );
        my $existing_methods = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                                 nested_recs => $existing_obj->{soap_methods},
                                                                 clone => 1,
                                                                );
        for my $key (keys %$store_methods) {
            if (exists $existing_methods->{$key} && defined $store_methods->{$key}->{flg_allowed}) {
                $existing_methods->{$key}->{flg_allowed} = $store_methods->{$key}->{flg_allowed};
            }
        }

        $store_obj->{soap_methods} = [ values %$existing_methods ];
    }

    if ( defined($store_obj->{validation_files})) {
        my $nested_def = $self->fields->{validation_files};
        my $store_validation_files = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                                       nested_recs => $store_obj->{validation_files},
                                                                      );

        for my $key (keys %$store_validation_files) {
            if ( ! $store_validation_files->{$key}->{xml_validation_file} ) {
                delete $store_validation_files->{$key}
            } elsif (defined($store_validation_files->{$key}->{import_url}) && 
                    ! length($store_validation_files->{$key}->{import_url})
                    ) {
                delete $store_validation_files->{$key}->{import_url};
            }
        }

        $store_obj->{validation_files} = [ values %$store_validation_files ];
    }

    my $profile_name = $self->get_new_value(%params, field_name => 'profile_name');

    if (  !defined $profile_name || !length($profile_name) ) {
        my $empty_profile_name_message = $self->get_field_display_name_string(field_name => 'profile_name') . ' is empty';
        die { error_message => $empty_profile_name_message,
              display_message => $empty_profile_name_message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
            };
    }

    $self->populate_defense_attributes(store_obj => $store_obj);

    if ($params{validate_value}) {
        $self->validate_defense_attributes(defense_settings => $store_obj->{xml_defense_settings});
    }
    for my $xml_defense (@{$store_obj->{xml_defense_settings}}) {
        if (defined $xml_defense->{def_attrib_value} &&
            $xml_defense->{value_type} eq F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER) {

            $xml_defense->{def_attrib_value} += 0;
        }
    }
}


#------------------------------------------------------------------------

sub after_nested_struct_store{
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $store_obj = $params{store_obj};
  if ($params{store_mode} eq 'update') {
      $self->populate_defense_attributes(store_obj => $store_obj);
  }
}

#----------------------------------------------------------------------

sub after_nested_struct_load {
    my ($self, %params) = @_;
    $self->populate_defense_attributes();
    return;
}

#----------------------------------------------------------------------

sub populate_defense_attributes {
    my ($self, %params) = @_;
    my $store_obj = $params{store_obj} || $self;

    if (exists($store_obj->{xml_defense_settings})) {
        my $defense_attributes = $self->get_defense_attributes();
        for my $xml_defense (@{$store_obj->{xml_defense_settings}}) {
            %$xml_defense = ( %$xml_defense, %{$defense_attributes->{$xml_defense->{def_attrib_id}}} );
        }
    }
}

#----------------------------------------------------------------------

sub get_defense_attributes {
    my ($self, %params) = @_;
    my $dbh = $params{dbh} || $self->{dbh};

    my %retval = map {$_->{def_attrib_id} => $_} F5::ASMConfig::Entity::XmlDefenseAttribute->retrieve_all( dbh => $dbh, );
    
    return $self->install_accessor('get_defense_attributes', \%retval);

}

#----------------------------------------------------------------------

sub update_parent_staging {
    my ($self, %params) = @_;
    return $self->update_content_profile_staging(%params);
}

#----------------------------------------------------------------------

sub to_declarative {
    my ($self, %params) = @_;

    my $entity = $self->SUPER::to_declarative(%params);

    for my $validation_file (@{$entity->{validationFiles}}) {
        delete $validation_file->{xmlValidationFile}->{contents};
    }
    return $entity;
}

#----------------------------------------------------------------------

1;
