package F5::ASMConfig::EventDispatcher;

use strict;
use warnings;

use File::Basename;
use Tie::RefHash;
use Storable qw(thaw);
use List::Util;

use F5::DbUtils;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Utils;
use F5::Cluster::Utils;
use F5::ObjUtils qw(freeze dclone);
use F5::ASMConfig::Entity::AsmConfigSettings;
use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::Cache;
use F5::BigIpLog;
use F5::DeviceGroup;
use F5::BigipUtils;
use F5::IO::Selector;
use F5::IO::Storable;
use F5::IO::ASMConfig::EventQueue;
use F5::DateTimeUtils;
use F5::Db::Single;

use Thrift::XS::BinaryProtocol;
use Thrift::XS::MemoryBuffer;

my $DEFAULT_EVENT_QUEUE_PORT = 9783;
my $DEFAULT_RPC_TO_EVENT_PORT = 9784;
my $CATCHUP_EVENT_LIMIT = 20;
my $RPC_TO_EVENT_LIMIT = 2;
my $MAX_DEFERRED_EVENT_SIZE = 20_000_000; # 20 MB
my $CLEAN_UNUSED_DEST_QUEUE_INTERVAL = 60;

my %topology_types = (
                      F5::ASMConfig::ElementType::FAILOVER_STATE() => 'failover_state',
                      F5::ASMConfig::ElementType::CLUSTER_STATE() => 'cluster_state',
                     );
my @SUBSCRIPTION_POLICY_FIELDS = qw(is_active_policy is_pb_enabled policy_type);

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);

  $SIG{PIPE} = 'IGNORE';

  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  F5::Db::Single->init(die_on_instantiate_failure => 1);
  $self->{logger} = $params{logger};
  $self->{logger}->debug('Initializing event dispatcher') if $self->{logger};
  $self->{selector} = F5::IO::Selector->new;

  $self->{config_file} = $params{config_file};

  # todo: Should catchup events handle multiple sync groups on EM? This may be a moot point.
  $self->{event_queue_port} = $params{event_queue_port} || F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'EventQueuePort') ||
    $DEFAULT_EVENT_QUEUE_PORT;
  $self->{rpc_to_event_port} = $params{rpc_to_event_port} || F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'RpcToEventPort') ||
    $DEFAULT_RPC_TO_EVENT_PORT;
  $self->{pulse_msg_interval} = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'pulse_msg_interval') || 30;
  $self->{last_dest_queue_connected} = '';
  $self->{events_received_count} = 0;

  my $bind_listener_local = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'AsmConfigHost') || 'local';
  my %listener_opts;
  if ($bind_listener_local eq 'local') {
      ### NOTE: This forces the socket routine to downgrade to vanilla INET4.
      $listener_opts{local_addr} = '127.0.0.1';
  }

  $self->{queue_server} = F5::IO::ASMConfig::EventQueue->new(server => 1,
                                                             port => $self->{event_queue_port},
                                                             read_message_handler => sub {
                                                                 my($connection, $session_id) = @_;
                                                                 $self->{logger}->debug("Read queue connection, session id = $session_id") if $self->{logger};
                                                                 $connection->{_session_id} = $session_id;
                                                                 $self->{last_dest_queue_connected} = $session_id;

                                                                 my $thrift_buffer = Thrift::XS::MemoryBuffer->new(100_000);
                                                                 $self->configure_dest_queue(session_id => $session_id,
                                                                                             socket => $connection,
                                                                                             thrift_buffer => $thrift_buffer,
                                                                                             thrift_serializer => Thrift::XS::BinaryProtocol->new($thrift_buffer),
                                                                                             is_active => 1,
                                                                                             was_queue_initialized => 1,
                                                                                            );
                                                             },
                                                             before_close_handler => sub {
                                                                 my($connection, $errno) = @_;
                                                                 $self->{logger}->debug("Closing event queue for session $connection->{_session_id}") if $self->{logger};
                                                                 delete $self->{dest_queues}->{$connection->{_session_id}};
                                                             },
                                                             socket_full_initial_handler => sub {
                                                                 my($connection, $errno) = @_;
                                                                 $self->{logger}->debug("Queue socket for session $connection->{_session_id} is full - deferring all writes") if $self->{logger};
                                                             },
                                                             flush_write_attempt_limit => 1,
                                                             logger => $self->{logger},
                                                             selector => $self->{selector},
                                                             %listener_opts,
                                                            );

  unless ($self->{queue_server}) {
    die "Can't bind to queue server port ($self->{event_queue_port})\n";
  }

  $self->{rpc_to_event_server} = F5::IO::Storable->new(server => 1,
                                                       port => $self->{rpc_to_event_port},
                                                       read_message_handler => sub {
                                                           $self->read_event(rpc_connection => $_[0],
                                                                             event => $_[1],
                                                                             event_size => $_[2],
                                                                            );
                                                       },
                                                       logger => $self->{logger},
                                                       selector => $self->{selector},
                                                       %listener_opts,
                                                      );

  unless ($self->{rpc_to_event_server}) {
    die "Can't bind to RPC/Event Dispatcher port ($self->{rpc_to_event_port})\n";
  }

  $self->{dest_queues} = {};
  $self->{event_id} = 1;
  $self->{throttled_subscriptions} = {};
  my $dbh = $params{dbh} || F5::Db::Single->instance;

  $self->update_system_topology(updated_state_key => "cluster_state",
                                cluster_state => F5::ASMConfig::ClusterState->new({
                                                                                 is_clustering_enabled => F5::Cluster::Utils::is_clustering_enabled() || 0,
                                                                                 is_master_blade => F5::Cluster::Utils::is_master_blade() || 0,
                                                                                 master_slot => F5::Cluster::Utils::get_master_blade() || 0,
                                                                                 this_slot => F5::Cluster::Utils::get_slot_number() || 0,
                                                                                }));

  $self->update_system_topology(updated_state_key => "failover_state",
                                failover_state => F5::ASMConfig::FailoverState->new({
                                                                                 is_active => F5::Cluster::Utils::is_failover_active(),
                                                                                 is_redundant => F5::Cluster::Utils::get_mcp_val("Failover.IsRedundant") eq "true" ? 1 : 0,
                                                                                }));
  if (F5::BigipUtils::is_device_bigip()) {
      $self->{flg_recommend_sync_when_policy_is_not_applied} = 
          F5::ASMConfig::Entity::AsmConfigSettings->new( dbh => $dbh)->retrieve(
                                                                                get_criteria => {}
                                                                               )->{flg_recommend_sync_when_policy_is_not_applied};
      $self->{flg_write_all_changes_to_syslog} = 
          F5::ASMConfig::Entity::AsmConfigSettings->new( dbh => $dbh)->retrieve(
                                                                                get_criteria => {}
                                                                               )->{flg_write_all_changes_to_syslog};
      $self->{device_group} = F5::DeviceGroup::get_device_group()->{name};
  }
  else  {
      $self->{flg_write_all_changes_to_syslog} = 1;
      $self->{flg_recommend_sync_when_policy_is_not_applied} = 0;
  }
  $self->{last_usage_of_policies} = {};

  $self->load_element_types;
}

#----------------------------------------------------------------------

sub load_element_types {
  my($self) = @_;

  my %element_types;

  my $module_path = F5::ASMConfig::Entity::Base::get_module_path();
  for my $class (glob($module_path . "/*.pm")) {
    next if ($class =~ m|/Base.pm$|);
    next if ($class =~ m|/Iterator.pm$|);
    $self->{logger}->trace("Loading element types from $class") if $self->{logger};

    $class =~ s|^.*/(F5.*)$|$1|;
    require $class;
    $class = basename($class);
    $class =~ s/.pm$//;
    $class = "F5::ASMConfig::Entity::$class";

    if ($class->can('element_type') && $class->element_type) {
      $element_types{$class->element_type} = $class;
    }
  }

  $self->{element_types} = \%element_types;
}

#----------------------------------------------------------------------

sub run_event_dispatch {
    my($self) = @_;

    my $last_pulse_time = time;

    while (1) {
        my $poll_time = undef;

        my @potential_poll_times;

        if (my @catchup_queues = $self->queues_needing_catchup) {
            if (grep { defined $_ && defined $_->{socket} && ! $_->{socket}->is_socket_full } @catchup_queues) {
                push(@potential_poll_times, 0);
            }
            else {
                push(@potential_poll_times, 5);
            }
        }

        if (my $next_throttled_event_time = $self->send_throttled_events) {
            push(@potential_poll_times, $next_throttled_event_time);
        }

        if ($self->{pulse_msg_interval}) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                                    start_time => $last_pulse_time));
        }

        if (@potential_poll_times) {
            $poll_time = List::Util::min(@potential_poll_times);
        }

        $self->{logger}->trace("About to poll for $poll_time seconds") if $self->{logger};
        my @readers = $self->{selector}->can_read($poll_time);

        for my $reader (@readers) {
            $reader->handle_read;
        }

        if ($self->{pulse_msg_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                               start_time => $last_pulse_time) == 0) {

            $self->log_pulse_message;
            $last_pulse_time = time;
        }

        $self->send_catchup_events;
        $self->clean_unused_dest_queues();
    }
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;

    my @additional_info;
    if ($self->{last_dest_queue_connected}) {
        push(@additional_info, "last queue connected = $self->{last_dest_queue_connected}");
    }

    my $additional_info = '';
    if (@additional_info) {
        $additional_info = '(' . join(', ', @additional_info) . ')';
    }
    $self->{logger}->info("$self->{events_received_count} events received since startup $additional_info") if $self->{logger};
}

#----------------------------------------------------------------------

sub read_event {
    my($self, %params) = @_;

    my $rpc_connection = $params{rpc_connection};
    my $event = $params{event};

    if ($event->{session} && $event->{session}->{session_id}) {
        $self->{logger}->debug("Received event from RPC channel (session $event->{session}->{session_id})") if $self->{logger};
    }

    $self->{events_received_count}++;
    $event->{event_id} = $self->{event_id};
    $self->{event_id}++;
    if ($self->is_subscription_event(event => $event)) {

        my $event_subscription = $event->{element}->{event_subscription};

        if ($event->{event_type} == F5::ASMConfig::EventType::ADD) {
            $self->add_subscription(subscription => $event_subscription,
                                    event => $event,
                                    session => $event->{session});
        }
        elsif ($event->{event_type} == F5::ASMConfig::EventType::DELETE) {
            $self->delete_subscription(subscription => $event_subscription,
                                       event => $event,
                                       policy => $event->{policy},
                                       session => $event->{session});
        }
    }
    else {
        $self->handle_event(%params);
    }
}

#----------------------------------------------------------------------

sub handle_event {
    my($self, %params) = @_;

    my $event = $params{event};
    my $event_size = $params{event_size};

    $self->apply_event_before(event => $event);
    $self->reset_subscriptions_to_catchup(event => $event);

    $self->{logger}->debug("Checking event for subscription matches") if $self->{logger};
    for my $dest_queue (values %{$self->{dest_queues}}) {
        if (!$dest_queue->{is_active} && $dest_queue->{was_queue_initialized}) {
            # dest_queue is dead and can be ignored
            next;
        }

        my @matched_subscription_ids;
        # send_result_as_event_type is passed directly from RPC on Storable struct (bypassing Thrift)
        if (($event->{send_result_as_event_type}  && $event->{session}->{session_id} eq $dest_queue->{session_id}) ||
            (! $event->{send_result_as_event_type}  &&
             (@matched_subscription_ids = $self->subscription_match(dest_queue => $dest_queue,
                                                                    event => $event)))) {
            if ($self->{logger} && $self->{logger}->is_debug()) {
                $self->{logger}->debug("Sending event to queue for session $dest_queue->{session_id} " . $self->get_event_debug_string($event));
            }

            $event->{event_subscription_id} = [ @matched_subscription_ids ];

            my $defer_this_event = 0;
            if (!$dest_queue->{was_queue_initialized} || 
                (@{$dest_queue->{catchup}} && !$self->is_subscription_event(event => $event)) || 
                @{$dest_queue->{deferred_events}}) {

                $defer_this_event = 1;
            }

            unless ($defer_this_event) {
                if (!$dest_queue->{socket}) {
                    $self->{logger}->debug("Event queue socket has gone away") if $self->{logger};
                    next;
                }
                elsif (! $dest_queue->{is_active} || $dest_queue->{socket}->is_socket_full()) {
                    $defer_this_event = 1;
                    }
                else {
                    my $rv = $self->write_event(event => $event,
                                                dest_queue => $dest_queue);

                    if (!$rv && $dest_queue->{is_active}) {
                        # Queue cannot accept right now, defer the event for later
                        $defer_this_event = 1;
                    }
                }
            }

            if ($defer_this_event) {
                my $total_deferred_event_size = List::Util::sum(@{$dest_queue->{deferred_event_sizes}}) || 0;
                if ($total_deferred_event_size > $MAX_DEFERRED_EVENT_SIZE) {
                    $self->{logger}->info("Session $dest_queue->{session_id} has too many deferred events - event dropped") if $self->{logger};
                }
                else {
                    $self->{logger}->debug("Session $dest_queue->{session_id} is still catching up - event deferred") if $self->{logger};
                    push(@{$dest_queue->{deferred_events}}, dclone($event));
                    push(@{$dest_queue->{deferred_event_sizes}}, $event_size);
                }
            }
        }
    }

    $self->apply_event_after(event => $event);
}

#----------------------------------------------------------------------

sub apply_event_before {
    my($self, %params) = @_;
    my $event = $params{event};

    if ($event->{element} && $event->{element}->{element_type}) {
        if ($event->{element}->{element_type} == F5::ASMConfig::ElementType::ASM_CONFIG_SETTINGS) {
            if (defined $event->{update_thrift_obj}->{flg_recommend_sync_when_policy_is_not_applied} ) {
                $self->{flg_recommend_sync_when_policy_is_not_applied} = $event->{update_thrift_obj}->{flg_recommend_sync_when_policy_is_not_applied};
            }
            if (defined $event->{update_thrift_obj}->{flg_write_all_changes_to_syslog} ) {
                $self->{flg_write_all_changes_to_syslog} = $event->{update_thrift_obj}->{flg_write_all_changes_to_syslog};
            }
        }
        elsif ($event->{element}->{element_type} == F5::ASMConfig::ElementType::DEVICE_GROUP) {
            $self->{device_group} = $event->{element}->{device_group}->{name};
        }

        if ($event->{element}->{element_type} == F5::ASMConfig::ElementType::CLUSTER_STATE) {
            $self->update_system_topology(updated_state_key => "cluster_state", event => $event);
        }

        if ($event->{element}->{element_type} == F5::ASMConfig::ElementType::FAILOVER_STATE) {
            $self->update_system_topology(updated_state_key => "failover_state", event => $event);
        }
    }
}

#----------------------------------------------------------------------

sub reset_subscriptions_to_catchup {
    my($self, %params) = @_;
    my $event = $params{event};

    if ($event->{element_type} && $event->{element_type} == F5::ASMConfig::ElementType::UCS_CONFIG_LOAD) {
        $self->{last_usage_of_policies} = {};
    }

    if ($event->{policy} && $event->{policy}->{policy_id}) {

        if ($event->{event_type} == F5::ASMConfig::EventType::DELETE &&
            $event->{element_type} == F5::ASMConfig::ElementType::POLICY &&
            $event->{policy} && $event->{policy}->{policy_id}) {

            delete $self->{last_usage_of_policies}->{$event->{policy}->{policy_id}};
        }

        my $last_usage_of_policy = $self->{last_usage_of_policies}->{$event->{policy}->{policy_id}} || {};

        my %restarted_subscriptions;
        for my $policy_flag (@SUBSCRIPTION_POLICY_FIELDS) {
            next unless defined $event->{policy}->{$policy_flag};

            if (($last_usage_of_policy->{$policy_flag} || 0) != $event->{policy}->{$policy_flag}) {
                # Flag has just been enabled -- check for subscriptions that would like to re-catchup

                for my $dest_queue (values %{$self->{dest_queues}}) {
                    for my $subscription (values %{$dest_queue->{subscriptions}}) {

                        if ($restarted_subscriptions{$subscription->{event_subscription_id}}) {
                            next;
                        }

                        if ($subscription->{resend_catchup_events_on_policy_state_change} &&
                            $subscription->{policy} &&
                            defined $subscription->{policy}->{$policy_flag} && 
                            $self->subscription_matches_policy_fields(subscription => $subscription,
                                                                      event => $event)) {

                            my $catchup_subscription = dclone($subscription);
                            $catchup_subscription->{policy}->{policy_id} = $event->{policy}->{policy_id};

                            $self->add_catchup_for_subscription(dest_queue => $dest_queue,
                                                                subscription => $catchup_subscription);

                            $self->send_catchup_start(dest_queue => $dest_queue,
                                                      subscription => $catchup_subscription);

                            $restarted_subscriptions{$subscription->{event_subscription_id}} = 1;
                        }
                    }
                }
            }
        }

        $self->{last_usage_of_policies}->{$event->{policy}->{policy_id}} = $event->{policy};
    }
    elsif ($event->{re_activate_catchup_subscriptions}) {

        for my $dest_queue (values %{$self->{dest_queues}}) {
            for my $subscription (values %{$dest_queue->{subscriptions}}) {
                if ($subscription->{resend_catchup_events_on_policy_state_change}) {

                    $self->add_catchup_for_subscription(dest_queue => $dest_queue,
                                                        subscription => $subscription);

                    $self->send_catchup_start(dest_queue => $dest_queue,
                                              subscription => $subscription);
                }
            }
        }
    }
}

#----------------------------------------------------------------------

sub subscription_matches_policy_fields {
    my($self, %params) = @_;
    my $event = $params{event};
    my $subscription = $params{subscription};

    if (! defined $subscription->{policy}) {
        return 1;
    }

    if (! $event->{policy}) {
        return 0;
    }

    for my $policy_field (('policy_id', @SUBSCRIPTION_POLICY_FIELDS)) {
        if (defined $subscription->{policy}->{$policy_field} &&
            $subscription->{policy}->{$policy_field} != $event->{policy}->{$policy_field}) {

            return 0;
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub apply_event_after {
    my($self, %params) = @_;
    my $event = $params{event};

    if ($self->{flg_write_all_changes_to_syslog} && defined $event->{syslog_message}) {
        F5::BigIpLog::bigip_ts_ts_info_msg_log($event->{syslog_message});
    }
}

#----------------------------------------------------------------------

sub update_system_topology {
    my($self, %params) = @_;

    my $updated_state_key = $params{updated_state_key};
    my $new_state;
    if ($params{event}) {
        $new_state = $params{event}->{element}->{$updated_state_key};
    }
    else {
        $new_state = $params{$updated_state_key};
    }

    if (hashes_differ($self->{$updated_state_key}, $new_state)) {
        $self->{$updated_state_key} = $new_state;
        if ($params{event}) {
            $params{event}->{override_event_sending_suppression} = 1;
        }
    }
    else {
        if ($params{event}) {
            $params{event}->{suppress_event_sending} = 1;
        }
    }

    if (($self->{cluster_state} && $self->{cluster_state}->{is_clustering_enabled} && ! $self->{cluster_state}->{is_master_blade}) || 
        ($self->{failover_state} && ! $self->{failover_state}->{is_active})) {
        $self->{suppress_event_sending} = 1;
    }
    else {
        if ($params{event} && $self->{suppress_event_sending}) {
            $params{event}->{re_activate_catchup_subscriptions} = 1;
        }
        $self->{suppress_event_sending} = 0;
    }
}

#----------------------------------------------------------------------

sub hashes_differ {
    my($hash_a, $hash_b) = @_;

    return 1 if (defined $hash_a != defined $hash_b);

    my %keys_a = map { ($_ => 1) } keys %$hash_a;
    my %keys_b = map { ($_ => 1) } keys %$hash_b;
    if (grep(!$keys_b{$_}, keys %keys_a)) {
        return 1;
    }

    for my $k (keys %$hash_a) {
        if ($hash_a->{$k} ne $hash_b->{$k}) {
            return 1;
        }
    }

    return;
}

#----------------------------------------------------------------------

sub configure_dest_queue {
  my($self, %params) = @_;

  my $session_id = $params{session_id};
  delete $params{session_id};

  unless ($self->{dest_queues}->{$session_id}) {
    $self->{dest_queues}->{$session_id} = {
                                           session_id => $session_id,
                                           socket => undef,
                                           thrift_serializer => undef,
                                           subscriptions => {},
                                           catchup => [],
                                           deferred_events => [],
                                           deferred_event_sizes => [],
                                           is_active => 0,
                                           was_queue_initialized => 0,
                                           create_time => time(),
                                          };
  }

  if (%params) {
    %{$self->{dest_queues}->{$session_id}} = (%{$self->{dest_queues}->{$session_id}},
                                              %params);
  }
}

#----------------------------------------------------------------------

sub subscription_match {
    my($self, %params) = @_;
    my $dest_queue = $params{dest_queue};
    my $event = $params{event};

    if ($event->{suppress_event_sending}) {
        return;
    }

    my @matching_subscriptions;
    my $newly_throttled_subscription;
    for my $subscription (values %{$dest_queue->{subscriptions}}) {

        if ($self->{suppress_event_sending} &&
            ! $event->{override_event_sending_suppression} &&
            ! $subscription->{send_topology_suppressed_events}) {
            next;
        }

        if ($self->is_subscription_event(event => $event)) {
            if ($subscription->{send_subscription_change_events} &&
                $event->{session}->{session_id} == $dest_queue->{session_id}) {

                push(@matching_subscriptions, $subscription->{event_subscription_id});
            }
            last;
        }

        if (! $self->subscription_matches_policy_fields(subscription => $subscription,
                                                       event => $event)) {
            next;
        }

        for my $element_event_type (@{$subscription->{element_event_types}}) {
            if ((($event->{element} && $event->{element}->{element_type} &&
                  $element_event_type->{element_type} == $event->{element}->{element_type}) ||
                 $element_event_type->{element_type} == F5::ASMConfig::ElementType::ALL_ELEMENT_TYPES) &&
                ($element_event_type->{event_type} == $event->{event_type} ||
                 $element_event_type->{event_type} == F5::ASMConfig::EventType::ALL_TYPES)) {

                if ($subscription->{rate_limit_per_policy}) {
                    my $policy_id = $event->{policy} ? $event->{policy}->{policy_id} : 0 || 0;
                    my $throttled_subscription_id = "$dest_queue->{session_id}__$subscription->{event_subscription_id}__$policy_id";

                    if (my $throttled_subscription = $self->{throttled_subscriptions}->{$throttled_subscription_id}) {
                        # There has already been an event matching this subscription within the rate limit interval

                        if (! $throttled_subscription->{next_event_to_send}) {
                            $newly_throttled_subscription = $throttled_subscription;
                        }

                        # An event for the throttled subscription has been sent in the last interval, so skip this event
                        $self->{logger}->trace("Skipping throttled event for session $dest_queue->{session_id} subscription $subscription->{event_subscription_id} (policy $policy_id)") if $self->{logger};
                        last;
                    }
                    else {
                        # Initial event for throttled subscription, which will be sent. Subsequent events after this one
                        # will be skipped (other than the event saved for sending after the interval has passed)
                        $self->{throttled_subscriptions}->{$throttled_subscription_id} = {
                                                                                          last_sent_event_time => time(),
                                                                                          subscription => $subscription,
                                                                                          dest_queue => $dest_queue,
                                                                                          policy_id => $policy_id,
                                                                                         };
                    }
                }

                push(@matching_subscriptions, $subscription->{event_subscription_id});
                last;
            }
        }
    }

    if (! @matching_subscriptions && $newly_throttled_subscription) {
        # This is the first skipped event due to throttling (i.e. the second event within the rate limit interval).
        # This event will be sent once the interval has passed, to ensure that Element Type/Policy triggers
        # will occur at some point that is definitely after the last ASM-config change was completed.
        # However, this event should only be saved until after the rate limit interval if it hasn't been
        # sent due to an unthrottled subscription. Also, it must only be sent once (for any one throttled
        # subscription).
        $self->{logger}->debug("Throttling events for session $dest_queue->{session_id} subscription $newly_throttled_subscription->{subscription}->{event_subscription_id} (policy $newly_throttled_subscription->{policy_id})") if $self->{logger};
        $newly_throttled_subscription->{next_event_to_send} = $event;
    }

    return @matching_subscriptions;
}

#----------------------------------------------------------------------

sub add_subscription {
  my($self, %params) = @_;

  my $subscription = $params{subscription};
  my $event = $params{event};
  my $session = $params{session};
  $event->{policy} = $subscription->{policy};

  $self->{logger}->debug("Adding subscription for session = $session->{session_id} " . $self->get_subscription_debug_string($subscription)) if $self->{logger};
  $self->configure_dest_queue(session_id => $session->{session_id});

  my $dest_queue = $self->{dest_queues}->{$session->{session_id}};
  $dest_queue->{subscriptions}->{$subscription->{event_subscription_id}} = $subscription;

  if ($subscription->{send_catchup_events}) {
      $self->add_catchup_for_subscription(dest_queue => $dest_queue,
                                          subscription => $subscription);
  }

  $self->handle_event(event => $event);
}

#----------------------------------------------------------------------

sub add_catchup_for_subscription {
    my($self, %params) = @_;
    my $subscription = $params{subscription};
    my $dest_queue = $params{dest_queue};

    # pabnag needs account_id on catchup start/complete events
    if ($subscription->{policy} && $subscription->{policy}->{policy_id} && ! $subscription->{policy}->{account_id}) {
        my $policy = F5::ASMConfig::Cache::get_cached_policy(use_singleton_dbh => 1,
                                                             policy => $subscription->{policy});
        $subscription->{policy}->{account_id} = $policy->{account_id};
    }

    for my $element_event_type (@{$subscription->{element_event_types}}) {
        push(@{$dest_queue->{catchup}},
             {
              element_type => $element_event_type->{element_type},
              minimal_load => $element_event_type->{minimal_load},
              event_subscription_id => $subscription->{event_subscription_id},
              policy => $subscription->{policy},
              subscription => $subscription,
             });
    }
}

#----------------------------------------------------------------------

sub send_catchup_complete {
    my($self, %params) = @_;

    return $self->send_catchup_event(%params,
                                     event_type => F5::ASMConfig::EventType::CATCHUP_COMPLETE,
                                     event_type_display => 'catchup complete',
                                    );
}

#----------------------------------------------------------------------

sub send_catchup_start {
    my($self, %params) = @_;

    return $self->send_catchup_event(%params,
                                     event_type => F5::ASMConfig::EventType::CATCHUP_START,
                                     event_type_display => 'catchup start',
                                    );
}

#----------------------------------------------------------------------

sub send_catchup_event {
  my($self, %params) = @_;
  my $subscription = $params{subscription};
  my $dest_queue = $params{dest_queue};

  my $event = F5::ASMConfig::Event->new({
                                         event_id => $self->{event_id}++,
                                         event_subscription_id => [ $subscription->{event_subscription_id} ],
                                         event_type => $params{event_type},
                                         policy => $subscription->{policy},
                                         element => F5::ASMConfig::Element->new({ element_type => 0}),
                                        });

  $self->{logger}->debug("Sending $params{event_type_display} to session $dest_queue->{session_id} for subscription $subscription->{event_subscription_id}") if $self->{logger};
  my $rv = $self->write_event(event => $event,
                              dest_queue => $dest_queue);
  unless ($rv) {
    $self->{logger}->debug("Write failed - queueing  $params{event_type_display} for session $dest_queue->{session_id}") if $self->{logger};
    unshift(@{$dest_queue->{deferred_events}}, $event);
    unshift(@{$dest_queue->{deferred_event_sizes}}, 0); # Catchup queue administrative messages don't count against the total
  }
}
#----------------------------------------------------------------------

sub delete_subscription {
    my($self, %params) = @_;

    my $subscription = $params{subscription};
    my $event = $params{event};
    my $policy = $params{policy};
    my $session = $params{session};

    my $dest_queue = $self->{dest_queues}->{$session->{session_id}};

    if ($subscription && $subscription->{event_subscription_id}) {
        $self->{logger}->debug("Deleting subscription $subscription->{event_subscription_id} for session = $session->{session_id}") if $self->{logger};
        $event->{policy} = $dest_queue->{subscriptions}->{$subscription->{event_subscription_id}}->{policy};
        $self->handle_event(event => $event);

        if ($dest_queue->{subscriptions}->{$subscription->{event_subscription_id}}->{send_catchup_events}) {
            for my $i (reverse 0..$#{$dest_queue->{catchup}}) {
                if ($subscription->{event_subscription_id} == $dest_queue->{catchup}->[$i]->{event_subscription_id}) {
                    splice(@{$dest_queue->{catchup}}, $i, 1);
                }
            }
        }

        delete $dest_queue->{subscriptions}->{$subscription->{event_subscription_id}};
    }
    else {
        my $delete_all_subscriptions = 0;
        if ($policy && $policy->{policy_id}) {
            $self->{logger}->debug("Deleting all subscriptions for session = $session->{session_id} on policy $policy->{policy_id}") if $self->{logger};
        }
        else {
            $self->{logger}->debug("Deleting all subscriptions for session = $session->{session_id} (all policies)") if $self->{logger};
            $delete_all_subscriptions = 1;
        }

        for my $subscription_id (keys %{$dest_queue->{subscriptions}}) {
            my $subscription = $dest_queue->{subscriptions}->{$subscription_id};

            if ($delete_all_subscriptions || 
                ($subscription->{policy} && $subscription->{policy}->{policy_id} == $policy->{policy_id})) {
                $self->{logger}->debug("Deleting subscription $subscription_id (" . $self->get_subscription_debug_string($subscription) . ")") if $self->{logger};
                $event->{element}->{event_subscription} = $dest_queue->{subscriptions}->{$subscription_id};
                $event->{policy} = $dest_queue->{subscriptions}->{$subscription_id}->{policy};
                $self->handle_event(event => $event);

                if ($subscription->{send_catchup_events}) {
                    for my $i (reverse 0..$#{$dest_queue->{catchup}}) {
                        if ($subscription_id == $dest_queue->{catchup}->[$i]->{event_subscription_id}) {
                            splice(@{$dest_queue->{catchup}}, $i, 1);
                        }
                    }
                }

                delete $dest_queue->{subscriptions}->{$subscription_id};
            }
        }
    }
}

#----------------------------------------------------------------------

sub queues_needing_catchup {
  my($self) = @_;

  return grep { @{$_->{catchup}} || @{$_->{deferred_events}} || ($_->{socket} && $_->{socket}->is_socket_full) } values %{$self->{dest_queues}};
}

#----------------------------------------------------------------------

sub send_catchup_events {
  my($self) = @_;

  for my $dest_queue (values %{$self->{dest_queues}}) {

      if ($dest_queue->{socket} && $dest_queue->{socket}->is_socket_full()) {
          # Attempt to resend partial messages to clear socket full
          $dest_queue->{socket}->rewrite();

          if ($dest_queue->{socket}->is_socket_full) {
              # Skip catchup events if socket is still full
              next;
          }
      }

      if ($self->{suppress_event_sending}) {
          for my $i (reverse 0..$#{$dest_queue->{catchup}}) {
              if (! $dest_queue->{catchup}->[$i]->{subscription}->{send_topology_suppressed_events}) {
                  splice(@{$dest_queue->{catchup}}, $i, 1);
              }
          }
      }

      my $catchup_send_count = 0;
      while ($dest_queue->{is_active} && @{$dest_queue->{catchup}} && $catchup_send_count < $CATCHUP_EVENT_LIMIT) {

          if (! $dest_queue->{socket} || $dest_queue->{socket}->is_socket_full()) {
              last;
          }

          my $current_catchup = $dest_queue->{catchup}->[0];

          if (!$current_catchup->{iterator}) {
              if ($self->{element_types}->{$current_catchup->{element_type}}) {
                  my $element_type_class = $self->{element_types}->{$current_catchup->{element_type}};

                  if ($element_type_class && $element_type_class->can('retrieve_all') && $element_type_class->db_table) {
                      my $minimal_load = defined($current_catchup->{minimal_load}) ? $current_catchup->{minimal_load}
                                                                                   : $current_catchup->{subscription}->{minimal_load};

                      my $search_filters = $element_type_class->catchup_search_filters;

                      $current_catchup->{iterator} = $element_type_class->retrieve_all(policy => $current_catchup->{policy},
                                                                                       minimal_load => $minimal_load,
                                                                                       logger => $self->{logger},
                                                                                       search_filters => $search_filters,
                                                                                       dbh => F5::Db::Single->instance);
                  }
              }
              elsif ($topology_types{$current_catchup->{element_type}}) {
                  my $get_next_sub;
                  {
                      my @topology_elements = ( F5::ASMConfig::Element->new({
                                                                             element_type => $current_catchup->{element_type},
                                                                             $topology_types{$current_catchup->{element_type}} => $self->{$topology_types{$current_catchup->{element_type}}},
                                                                            }) );
                      $get_next_sub = sub { shift @topology_elements };
                  }
                  $current_catchup->{iterator} = F5::ASMConfig::Entity::Iterator->new(get_next_function => $get_next_sub);
              }
          }

          my $is_catchup_complete = 0;
          if (!$current_catchup->{iterator}) {
              $is_catchup_complete = 1;
          }
          else {
              my @events;

              if ($current_catchup->{deferred_events}) {
                  @events = @{$current_catchup->{deferred_events}};
                  delete $current_catchup->{deferred_events};
              }
              else {
                  my %elements_by_policy;
                  my %policies;
                  for my $i (1..$current_catchup->{subscription}->{catchup_batch_size}) {
                      my $element = $current_catchup->{iterator}->get_next;

                      if ($element) {
                          if ($element->can('populate_for_event')) {
                              $element->populate_for_event();
                          }
                          my $policy_id = '';
                          if ($element->{policy}) {
                              if (! $element->{policy}->{policy_id} || ! $policies{$element->{policy}->{policy_id}}) {
                                  my $policy = F5::ASMConfig::Cache::get_cached_policy(use_singleton_dbh => 1,
                                                                                       policy => $element->{policy});
                                  $policies{$policy->{policy_id}} = $policy;
                                  $element->{policy} = $policy;
                              }
                              $policy_id = $element->{policy}->{policy_id};

                              if (! $self->subscription_matches_policy_fields(subscription => $current_catchup,
                                                                              event => { policy => $element->{policy} })) {
                                  next;
                              }
                          }
                          if (ref($element) ne 'F5::ASMConfig::Element') {
                              $element = F5::ASMConfig::Element->new({
                                                                      element_type => $element->element_type,
                                                                      $element->element_type_name => $element
                                                                     });
                          }

                          push(@{$elements_by_policy{$policy_id}}, $element);
                      }
                      else {
                          $is_catchup_complete = 1;
                          last;
                      }
                  }

                  for my $policy_id (keys %elements_by_policy) {

                      my $event_type = ($current_catchup->{subscription}->{send_catchup_events_as_add} ?
                                        F5::ASMConfig::EventType::ADD :
                                        F5::ASMConfig::EventType::CATCHUP_ADD);
                      my $event = F5::ASMConfig::Event->new({
                                                             event_id => $self->{event_id}++,
                                                             event_subscription_id => [ $current_catchup->{event_subscription_id} ],
                                                             event_type => $event_type,
                                                             policy => $policies{$policy_id},
                                                            });
                      if ($current_catchup->{subscription}->{catchup_batch_size} > 1 && 
                          scalar(@{$elements_by_policy{$policy_id}}) > 1) {

                          $event->{search_result} = F5::ASMConfig::SearchResult->new(element_list => $elements_by_policy{$policy_id});
                      }
                      else {
                          $event->{element} = $elements_by_policy{$policy_id}->[0];
                      }
                      push(@events, $event);
                  }

              }

              while (my $event = shift(@events)) {

                  if ($self->{logger} && $self->{logger}->is_debug()) {
                      $self->{logger}->debug("Sending catchup event to session $dest_queue->{session_id} for subscription $current_catchup->{event_subscription_id}, " . $self->get_event_debug_string($event));
                  }

                  my $rv = $self->write_event(event => $event,
                                              dest_queue => $dest_queue);
                  if (!$rv) {
                      $self->{logger}->debug("Catchup event deferred for session $dest_queue->{session_id}") if $self->{logger};
                      unshift(@events, $event);
                      $current_catchup->{deferred_events} = \@events;
                      last;
                  }
              }
          }

          if ($is_catchup_complete) {
              # Catchup complete
              shift(@{$dest_queue->{catchup}});
              F5::ASMConfig::Cache::soft_reset();

              unless (grep { $_->{event_subscription_id} eq $current_catchup->{event_subscription_id} } @{$dest_queue->{catchup}}) {
                  $self->send_catchup_complete(subscription => $current_catchup,
                                               dest_queue => $dest_queue,
                                              );
              }
          }

          $catchup_send_count++;
      }

      while ($dest_queue->{is_active} && ! @{$dest_queue->{catchup}} && @{$dest_queue->{deferred_events}} && $catchup_send_count < $CATCHUP_EVENT_LIMIT) {
          if (! $dest_queue->{socket} || $dest_queue->{socket}->is_socket_full()) {
              last;
          }

          my $event = shift(@{$dest_queue->{deferred_events}});
          my $event_size = shift(@{$dest_queue->{deferred_event_sizes}});

          if ($self->{logger} && $self->{logger}->is_debug()) {
              $self->{logger}->debug("Writing deferred event $event->{event_id} to session $dest_queue->{session_id} " . $self->get_event_debug_string($event));
          }
          my $rv = $self->write_event(event => $event,
                                      dest_queue => $dest_queue);
          $catchup_send_count++;
          unless ($rv) {
              $self->{logger}->debug("Write failed - requeueing event deferred for session $dest_queue->{session_id}") if $self->{logger};
              unshift(@{$dest_queue->{deferred_events}}, $event);
              unshift(@{$dest_queue->{deferred_event_sizes}}, $event_size);
              last;
          }
      }
  }
}

#----------------------------------------------------------------------

sub send_throttled_events {
    my($self) = @_;

    my $next_throttled_event_time;
    for my $throttled_subscription_id (keys %{$self->{throttled_subscriptions}}) {
        my $throttled_subscription = $self->{throttled_subscriptions}->{$throttled_subscription_id};

        my $subscription_event_time = F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $throttled_subscription->{subscription}->{rate_limit_per_policy},
                                                                                         start_time => $throttled_subscription->{last_sent_event_time});

        if ($subscription_event_time) {
            # The rate limit interval for this throttled subscription has not yet passed. The throttled subscription
            # that will send nearest in the future will be used to set the timeout for the select loop.
            if (! $next_throttled_event_time && $subscription_event_time < $next_throttled_event_time) {
                $next_throttled_event_time = $subscription_event_time;
            }
        }
        elsif ($throttled_subscription->{next_event_to_send}) {
            # This subscription has a waiting event to send. We send the event, but the
            # throttling metadata is preserved, in case there are more matching events
            # in the next interval.
            $self->{logger}->debug("Sending throttled event to queue for session $throttled_subscription->{dest_queue}->{session_id} " . $self->get_event_debug_string($throttled_subscription->{next_event_to_send})) if $self->{logger};
            $self->write_event(event => $throttled_subscription->{next_event_to_send},
                               dest_queue => $throttled_subscription->{dest_queue});
            $throttled_subscription->{last_sent_event_time} = time();
            delete $throttled_subscription->{next_event_to_send};
        }
        else {
            # The rate limit interval has passed without receiving another event, so
            # the subscription no longer needs to be throttled (for now)
            delete $self->{throttled_subscriptions}->{$throttled_subscription_id};
        }
    }

    return $next_throttled_event_time;
}

#----------------------------------------------------------------------

sub write_event {
  my($self, %params) = @_;
  my $event = $params{event};
  my $dest_queue = $params{dest_queue};

  $self->{logger}->trace($event) if $self->{logger};
  eval {
      $event->write($dest_queue->{thrift_serializer});
  };
  if ($@) {
      $self->{logger}->info("Event couldn't be serialized -- event will be skipped ($@)");
      $self->{logger}->info({event => $event});
      # Partially written thrift buffer (in the in-memory serializer) must be cleared out
      $dest_queue->{thrift_buffer}->read($dest_queue->{thrift_buffer}->available());
      # Return true value - otherwise, this event will be stuck as deferred and will loop endlessly
      return -1;
  };

  # Read entire contents of Thrift MemoryBuffer, which also clears it out completely
  my $buf = $dest_queue->{thrift_buffer}->read($dest_queue->{thrift_buffer}->available());

  return $dest_queue->{socket}->write_message($buf);
}

#----------------------------------------------------------------------

sub get_event_debug_string {
  my($self, $event) = @_;
  my @display_values;
  my %event_types = F5::ASMConfig::Utils::get_event_type_strings();
  push(@display_values, 'event_type = ' . $event_types{$event->event_type});
  if ($event->{policy} && $event->{policy}->{policy_id}) {
    push(@display_values, 'policy = ' . $event->{policy}->{policy_id});
  }

  my $element = $event->element;
  my $element_class;
  if ($element && ($element_class = $self->{element_types}->{$element->element_type})) {
    my $element_type_name = $element_class->element_type_name;
    my $entity = $element_class->new(dbh => F5::Db::Single->instance,
                                     object => $element->$element_type_name);

    push(@display_values, 'element = ' . $entity->get_debug_string);
  }

  return join(', ', @display_values);
}

#----------------------------------------------------------------------

sub get_subscription_debug_string {
  my($self, $subscription) = @_;
  my @display_values;
  my %event_types = F5::ASMConfig::Utils::get_event_type_strings();
  if ($subscription->policy) {
    if ($subscription->policy->{policy_id}) {
        push(@display_values, 'policy = ' . $subscription->policy->{policy_id});
    }
    else {
        if ($subscription->policy->{is_active_policy}) {
            push(@display_values, 'active policies');
        }
        if (defined($subscription->policy->{is_pb_enabled})) {
            push(@display_values, 'pb ' . ($subscription->policy->{is_pb_enabled} ? 'enabled ' : 'disabled ') . 'policies');
        }
        if (defined($subscription->policy->{policy_type})) {
            push(@display_values, F5::ASMConfig::Entity::Policy->get_field_display_string(field_name => 'policy_type',
                                                                                          field_value => $subscription->policy->{policy_type},) . ' policies');
        }
    }
  }
  else {
    push(@display_values, 'all policies');
  }

  for my $element_event_type (@{$subscription->element_event_types}) {
    my $element_type = $element_event_type->element_type;
    if ($self->{element_types}->{$element_type}) {
      push(@display_values, '[ element = ' . $self->{element_types}->{$element_type}->element_type_name . ', event = ' . $event_types{$element_event_type->event_type} . ']');
    }
  }

  return join(', ', @display_values);
}

#----------------------------------------------------------------------

sub is_subscription_event {
    my($self, %params) = @_;
    my $event = $params{event};

    if ($event->{element} && $event->{element}->{element_type} &&
        $event->{element}->{element_type} == F5::ASMConfig::ElementType::EVENT_SUBSCRIPTION) {
        return 1;
    }
    return;
}

#----------------------------------------------------------------------

sub clean_unused_dest_queues {
    my($self) = @_;

    for my $dest_queue (values %{$self->{dest_queues}}) {
        if (! $dest_queue->{was_queue_initialized} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $CLEAN_UNUSED_DEST_QUEUE_INTERVAL,
                                                               start_time => $dest_queue->{create_time}) == 0) {

            my $unused_interval = time() - $dest_queue->{create_time};
            $self->{logger}->debug("Session $dest_queue->{session_id} did not connect listener after $unused_interval seconds, discarding subscriptions and deferred events") if $self->{logger};
            delete $self->{dest_queues}->{$dest_queue->{session_id}};
        }
    }
}

#----------------------------------------------------------------------

1;
