package F5::ASMConfig::Handler::ActionItem;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::Entity::ActionItem;

our @handled_functions = (
                            {
                             function_name => 'delete_action_item',
                             relayable => 0,
                             write_access => 1,
                            },
                            {
                             function_name => 'delete_many_action_items',
                             relayable => 0,
                             write_access => 1,
                            },
                            {
                             function_name => 'get_action_item_notification_count',
                             relayable => 0,
                            },
                            {
                             function_name => 'search_action_item',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ActionItem';

#----------------------------------------------------------------------

sub delete_action_item {
  my ($self, $session, $call_flags, $placeholder, $action_item_criteria) = @_;
  my $action_item = $entity_class->new(get_criteria => $action_item_criteria, 
                                       dbh => $self->{dbh});
  return unless $action_item;
  if ($action_item->{action_item_type} == F5::ASMConfig::ActionItemType::POLICY_ACTION_ITEM_SUMMARY()) {
     return $self->handle_error(action => 'delete',
                                error_code => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                error_thrown => "Policy Summary items cannot be deleted",
                                call_flags => $call_flags,
                                error_level => 'info',
                               );
  }

  my $dismiss_time = $self->{action_item_handler}->get_dismiss_time(type => $action_item->{action_item_type});
  $action_item->{next_display_time} = time() + $dismiss_time;
  return $self->base_update(rpc_params => [$session, $call_flags, $placeholder, $action_item_criteria, $action_item],
                            entity_class => $entity_class,
                           );
}

#----------------------------------------------------------------------

sub delete_many_action_items {
    my ($self, $session, $call_flags, $placeholder, $action_item_criteria) = @_;

    $call_flags->{search_filters} ||= {};
    push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new(field_name => 'action_item_type',
                                                                             field_value => F5::ASMConfig::ActionItemType::POLICY_ACTION_ITEM_SUMMARY(),
                                                                             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                                                                         ));
    my $action_item_iter = $entity_class->retrieve_all(dbh => $self->{dbh},
                                                       call_flags => $call_flags,
                                                       search_filters => $call_flags->{search_filters},
                                                       get_criteria => $action_item_criteria);

    my $delete_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags, search_filters => []);
    my @entities_changed;
    while (my $action_item = $action_item_iter->get_next()) {
          my $rv = $self->delete_action_item($session,
                                             $delete_call_flags,
                                             $placeholder,
                                             $action_item);
          if ($rv) {
              push(@entities_changed, $rv);
          }
    }

    my $rv;
    if ($call_flags->{rest_call}) {  #Return a search result if this is a rest call
        $rv = $self->entities_to_search_result(entities => \@entities_changed,
                                               call_flags => $call_flags,
                                               entity_class => $entity_class,);
    }
    else {
        $rv = scalar @entities_changed;
    }

    my $retval = { return_value => $rv, entities_changed => \@entities_changed};                    
    return $retval;
}

#----------------------------------------------------------------------

sub get_action_item_notification_count {
    my ($self, $call_flags) = @_;
    my $dbh = $self->{dbh};
    my $time = time();
    my ($general_count) = $dbh->selectrow_array(<<EndQuery, undef, $time);
SELECT COUNT(*) FROM PLC.ACTION_ITEMS 
  WHERE next_display_time <= ?
EndQuery

    my ($policy_count) = $dbh->selectrow_array(<<EndQuery, undef, $time);
SELECT COUNT(*) FROM PLC.PL_POLICY_ACTION_ITEMS
  WHERE next_display_time <= ?
EndQuery

   return {return_value => $general_count + $policy_count};
}

#----------------------------------------------------------------------

sub search_action_item {
    my($self, $session, $call_flags, @other_rpc) = @_;

    $self->add_partition_search_filter(session    => $session,
                                       call_flags => $call_flags);  #This will be forwarded through to the call to find policies

    my $rv = $self->base_search(rpc_params => [$session, $call_flags, @other_rpc],
                                entity_class => $entity_class,
                               );
    return $rv;
}

#----------------------------------------------------------------------

1;
