package F5::ASMConfig::Handler::ApiProtectionProfileConfigTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ApiProtectionProfileConfigTask;

our @handled_functions = (
                            {
                             function_name => 'add_api_protection_profile_config_task',
                             criteria_index => 3,
                             store_index    => 4,
                            },
                            {
                              function_name => 'update_api_protection_profile_config_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            {
                             function_name => 'execute_api_protection_profile_config_task',
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                            },
                            {
                             function_name => 'delete_api_protection_profile_config_task',
                            },
                            {
                             function_name => 'delete_many_api_protection_profile_config_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ApiProtectionProfileConfigTask';

my $API_POLICY_TEMPLATE_NAME = 'API Security';

#----------------------------------------------------------------------

sub add_api_protection_profile_config_task {
    my($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                      rpc_params => \@rpc_params,
                                                      entity_class => $entity_class,
                                                      task_type => 'api_protection_profile_config',);
}

#----------------------------------------------------------------------

sub execute_api_protection_profile_config_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'api_protection_profile_config',
                                                          execute_func => sub { api_protection_profile_config(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub api_protection_profile_config {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};

    my $username = $task->{username};
    my $policy = $task->{target_policy};
    if ($policy) {
        $policy = F5::ASMConfig::Cache::get_cached_policy(policy => $policy, dbh => $self->{dbh});
    }

    my $policy_name = $task->{full_path} || $task->{policy_name} || $policy->{policy_name};
    my $partition = $session->{partition};
    if ($policy_name && ($policy_name =~ m|^/([^/]*)|)) {
        $partition = $1;
    }

    my $policy_template = F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $self->{dbh},
                                                                     get_criteria => {policy_template_name => $API_POLICY_TEMPLATE_NAME});
    my $filename = $policy_template->{template_file};
    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                              policy_name => $policy_name,
                                                                              policy_template => $policy_template,
                                                                              partition => $partition,
                                                                              force_policy_name_flg => defined ($task->{target_policy}),
                                                                              );

    eval {
        if ($import_export_properties->{force_policy_name_flg}) {
            if ($policy->{parent_policy_id} ) {
                $import_export_properties->{parent_policy} = F5::ASMConfig::Policy->new(policy_id =>  $policy->{parent_policy_id});
            }
        }

        $self->begin_work($self->{session});
        my $rv = $self->import_policy($session, $call_flags, $import_export_properties);
        $self->commit($self->{session});

        if ($rv->{msgs}) {
            my ($msg) = join ("\n", map { $_->{msg} } @{$rv->{msgs}});
            $task->{message} = $msg;
        }

        if (!$rv->{completed_successfully}) {
            $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
        }
        else {
            $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
            $task->{result_policy} = $rv->{policy};
        }
    };

    my $err = $@;
    if ($err) {
        die $err;
    }
}

#----------------------------------------------------------------------

sub update_api_protection_profile_config_task {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'api_protection_profile_config',);
}

#----------------------------------------------------------------------

sub delete_api_protection_profile_config_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_api_protection_profile_config_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;

