package F5::ASMConfig::Handler::BotDefenseEventLog;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::DbUtils;

our @handled_functions = (
                           {
                             function_name => 'search_bot_defense_event_log',
                             authz_limit_field => 'partition_id',
                           },
                           {
                             function_name => 'get_bot_defense_event_log',
                             authz_limit_field => 'partition_id',
                           },
                           {
                             function_name => 'delete_bot_defense_event_log',
                             authz_limit_field => 'partition_id',
                             criteria_index => 3,
                             store_index => 4,
                           },
                           {
                             function_name => 'delete_many_bot_defense_event_logs',
                             authz_limit_field => 'partition_id',
                             criteria_index => 3,
                             store_index => 4,
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::BotDefenseEventLog';
my $DELETE_BATCH_SIZE = 5000;

#----------------------------------------------------------------------

sub search_bot_defense_event_log {
    my($self, $session, $call_flags, undef, $bot_defense_event_log_criteria) = @_;
    $call_flags->{search_filters} ||= [];
    my $search_filters = $call_flags->{search_filters};

    if ($bot_defense_event_log_criteria) {
        my $obj = $entity_class->new(object => $bot_defense_event_log_criteria,
                                     dbh => $self->{dbh});
        push (@$search_filters, @{$obj->to_search_filters});
    }

    my $specific_search = grep { ($_->{field_name} eq 'support_id' || $_->{field_name} eq 'rest_uuid' || $_->{field_name} eq 'bot_defense_event_log_internal_id') &&
                                 ($_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() ||
                                  $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() ||
                                 $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()) } @$search_filters;

    if (!$specific_search) {
        $call_flags->{hide_deleted} = 1;
    }

    $self->add_partition_search_filter(session => $session,
                                       call_flags => $call_flags);

    my $rv = $self->base_search(rpc_params => [$session, $call_flags, undef, $bot_defense_event_log_criteria],
                                entity_class => $entity_class,
                               );
    return $rv;
}

#----------------------------------------------------------------------

sub delete_bot_defense_event_log {
    my($self, $session, $call_flags, undef, $bot_defense_event_log) = @_;

    my $class = ref($bot_defense_event_log) || "F5::ASMConfig::BotDefenseEventLog";
    my $store_bot_defense_event_log = $class->new();
    $store_bot_defense_event_log->{flg_is_deleted} = 1;
    my $retval =  $self->base_update(rpc_params => [$session, $call_flags, undef, $bot_defense_event_log, $store_bot_defense_event_log],
                                     entity_class => $entity_class,
                                    );
    return $retval;
}

#----------------------------------------------------------------------

my %STRICTLY_INCREASING_DB_FIELDS = ( support_id => 1,
                                      rest_uuid => 1,
                                      request_time => 1,
                                      bot_defense_event_log_internal_id => 1,
                                     );

sub delete_many_bot_defense_event_logs {
    my($self, $session, $call_flags, undef, $bot_defense_event_log_criteria) = @_;

    my $max_id;
    my $search_filters = $call_flags->{search_filters} || [];
    if ($bot_defense_event_log_criteria) {
        my $obj = $entity_class->new(object => $bot_defense_event_log_criteria,
                                     dbh => $self->{dbh});
        push (@$search_filters, @{$obj->to_search_filters});
    }

    if (!@{$search_filters} && !$call_flags->{page_size}) {
        $max_id = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT MAX(bot_defense_event_log_internal_id) FROM PRX.BOT_DEFENSE_EVENT_LOG
EndQuery
    }
    elsif (scalar @{$search_filters} == 1 && !$call_flags->{page_size}) {
        my ($search_filter) = @{$search_filters};
        if ($STRICTLY_INCREASING_DB_FIELDS{$search_filter->{field_name}} &&
            ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT() ||
             $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE())
           ) {
            my $db_field_name = $entity_class->fields->{$search_filter->{field_name}}->{db_field_name};
            my $comparison = ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT()) ? '<'
                                                                                                                           : '<=';
            $max_id = $self->{dbh}->selectrow_array(<<EndQuery, undef, $search_filter->{field_value});
SELECT MAX(bot_defense_event_log_internal_id) FROM PRX.BOT_DEFENSE_EVENT_LOG
  WHERE $db_field_name $comparison ?
EndQuery
        }
    }

    my $rv = 0;
    if ($max_id) {
        my $bot_defense_event_log_iterator = $entity_class->retrieve_all(dbh => $self->{dbh},
                                                                         page_size => 1,
                                                                         selected_fields => ['bot_defense_event_log_internal_id'],
                                                                         search_filters => [F5::ASMConfig::SearchFilter->new(field_name => 'bot_defense_event_log_internal_id',
                                                                                                                             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE(),
                                                                                                                             field_value => $max_id)],
                                                                         return_as_raw_db_hashref => 1,
                                                                        );
        $rv = $bot_defense_event_log_iterator->{total_record_count};

        # update if a row exists, insert otherwise
        F5::DbUtils::insert_row(dbh => $self->{dbh},
                                table_name => 'PRX.BOT_DEFENSE_EVENT_LOG_CLEARED',
                                rec => { bot_defense_event_log_id => $max_id },
                                on_duplicate_key_update_fields => [ qw( bot_defense_event_log_id ) ],
                               );
    }
    else {
        $call_flags->{hide_deleted} = 1;

        $self->add_partition_search_filter(session => $session,
                                           call_flags => $call_flags);

        my $bot_defense_event_log_iterator = $entity_class->retrieve_all(dbh => $self->{dbh},
                                                               sort_orders => $call_flags->{sort_orders},
                                                               page_size => $call_flags->{page_size},
                                                               start_page => $call_flags->{start_page},
                                                               get_criteria => $bot_defense_event_log_criteria,
                                                               selected_fields => ['bot_defense_event_log_internal_id'],
                                                               search_filters => $call_flags->{search_filters},
                                                               return_as_raw_db_hashref => 1,
                                                               call_flags => $call_flags,
                                                               skip_total_record_count => 1,
                                                              );
        my $batched_insertor = F5::BatchInsert->new(dbh => $self->{dbh},
                                                    logger => $self->{logger},
                                                    batch_size => $DELETE_BATCH_SIZE);

        while (my $bot_defense_event_log = $bot_defense_event_log_iterator->get_next()) {
           $batched_insertor->insert_row(table_name => 'PRX.BOT_DEFENSE_EVENT_LOG_PROPERTIES',
                                         rec => {bot_defense_event_log_id => $bot_defense_event_log->{bot_defense_event_log_internal_id},
                                                 flg_is_deleted => 1},
                                         do_replace => 1,
                                       );
           $rv++;
        }
        $batched_insertor->flush;
    }

    if ($call_flags->{rest_call}) {
        $rv = F5::ASMConfig::SearchResult->new(total_record_count => $rv);
    }
    return {return_value => $rv};
}

#----------------------------------------------------------------------

sub get_bot_defense_event_log {
    my($self, $session, $call_flags, undef, $bot_defense_event_log ) = @_;

    return $self->base_get(rpc_params => [$session, $call_flags, undef, $bot_defense_event_log ],
                                     entity_class => $entity_class,
                                    );
}

1;
