package F5::ASMConfig::Handler::BulkTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::BulkTask;
use F5::Utils::File;
use F5::Utils::Rest;
use F5::DateTimeUtils;
use MIME::Base64;
use JSON;
use URI;

our @handled_functions = (
                            {
                             function_name => 'add_bulk_task',
                             criteria_index => 3,
                             store_index    => 4,
                            },
                            {
                              function_name => 'update_bulk_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            { 
                             function_name => 'execute_bulk_task',
                            },
                            {
                              function_name => 'delete_bulk_task',
                            },
                            {
                             function_name => 'delete_many_bulk_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::BulkTask';
my $BULK_FILE_LOCATION = '/var/config/rest/bulk';
my $BULK_FILE_WORKER_URI = 'https://localhost/mgmt/shared/file-transfer/bulk';
my $ASM_REST_FILE_WORKER_URI = 'https://localhost/mgmt/tm/asm/file-transfer/downloads';


#----------------------------------------------------------------------

sub add_bulk_task {
  my($self, $session, $call_flags, $place_holder, $bulk_task) = @_;

  if (!defined($bulk_task->{flg_create_result_reference})) {
      #By default create a result file when the bulk endpoint is being used https://docs.f5net.com/x/K5DIEQ
      $bulk_task->{flg_create_result_reference} = $bulk_task->{bulk_deploy_file} ? 1 : 0;
  }

  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => [ $session, $call_flags, $place_holder, $bulk_task ],
                                                    entity_class => $entity_class,
                                                    task_type => 'bulk',);
}

#----------------------------------------------------------------------

sub update_bulk_task {
    my ($self, @rpc_params) = @_;   
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'bulk',);
}

#----------------------------------------------------------------------

sub delete_bulk_task {
  my($self, $session, $call_flags, $place_holder, $bulk_task) = @_;

  my $task = $entity_class->new(dbh => $self->{dbh}, get_criteria => $bulk_task);

  my $rv = F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => [ $session, $call_flags, $place_holder, $bulk_task ],
                                                         entity_class => $entity_class,
                                                        );

  if ($task && $task->{asm_bulk_deploy_file}) {
      unlink ("$F5::ASMConfig::Handler::TaskBase::REST_DIR/$task->{username}~$task->{asm_bulk_deploy_file}");
  }
  if ($task->{result_file} && $task->{result_file} =~ /\Q$ASM_REST_FILE_WORKER_URI\E/) {
      unlink ("$F5::ASMConfig::Handler::TaskBase::REST_DIR/$task->{username}~bulk-results-$task->{rest_uuid}.json");
  }
  return $rv;
}

#----------------------------------------------------------------------
sub command_to_rest_request {
    my ($command) = @_;
    my $uri = URI->new($command->{uri});
    my $body;
    if ($command->{body}) {
        $body = encode_json($command->{body}); #TODO: Test high ascii
    }
    my %rest_parameters = $uri->query_form();
    my $rest_path = $uri->path;
    if ((index($rest_path, 'bulk') > 0) && $command->{method} eq 'POST') {
        die "Recursive POST to bulk is not allowed.\n";
    } 
    return F5::ASMConfig::RestRequest->new(rest_path => $uri->path,
                                           rest_method => $command->{method},
                                           rest_body => $body,
                                           rest_parameters => \%rest_parameters,
                                           original_uri => $command->{uri},);
}
#----------------------------------------------------------------------

sub bulk {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};

    my @rest_requests;
    my $result_filename;
    my $result_fh;

    eval {
        my $deploy_file = $task->{bulk_deploy_file} || $task->{asm_bulk_deploy_file};
        if ($deploy_file) {
            my $filename = $task->{bulk_deploy_file} ? "$BULK_FILE_LOCATION/$deploy_file"
                                                     :  F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename});
            if (! -e $filename) {
                  $self->{logger}->info("Could not find $filename");
                  die  "Cannot find Deploy File '$deploy_file'\n";
            }
            my $bulk_deploy;
            eval {
                #TODO: Look into JSON::SL for partial parsing.
                $bulk_deploy = decode_json(F5::Utils::File::file_slurp($filename));
            };
            if ($@) {
                $self->{logger}->info("Cannot parse file '$filename': $@");
                die "Invalid bulkDeployFile '$deploy_file'\n";
            }
            @rest_requests = map { command_to_rest_request($_) } @{$bulk_deploy->{commands}};
        }
        else {
            @rest_requests = map { command_to_rest_request($_) } @{$task->{commands}};
        }

        if ($task->{flg_create_result_reference}) {
            $result_filename = "bulk-results-$task->{rest_uuid}.json";
            $task->{result_file} = $task->{bulk_deploy_file} ? "$BULK_FILE_WORKER_URI/$result_filename"
                                                             : "$ASM_REST_FILE_WORKER_URI/$result_filename";

            $result_filename = $task->{bulk_deploy_file} ? "$BULK_FILE_LOCATION/$result_filename"
                                                         :  F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $result_filename);
            open ($result_fh, '>', $result_filename);
            print $result_fh __result_header(task => $task);
        }

        $task->{timeout_seconds} ||= 0;
        if ($task->{timeout_seconds}) {
            alarm $task->{timeout_seconds};
        }

        $self->begin_work($self->{session});

        my $call_id = 0;        
        for my $rest_request (@rest_requests) {
            my $bulk_call_flags = F5::ASMConfig::CallFlags->new(bulk_call => 1);

            $call_id++;
            my $rv;
            eval {
                $self->{authz_checked} = 0;
                $rv = $self->rest_call($session, $bulk_call_flags, $rest_request);
            };
            my $ex = $@;
            die $ex if ($ex && $task->{is_transaction}); #all or nothing

            if ($result_filename) {
                print $result_fh __result_response(rest_request => $rest_request,
                                                   rv => $rv,
                                                   call_id => $call_id,
                                                   exception => $ex);
            }
        }

        if ($task->{timeout_seconds}) {
            alarm 0;
        }
        $self->commit($self->{session});
    };
    my $err = $@;

    if ($result_filename) {
        print $result_fh __result_footer(task => $task);
        close ($result_fh);
    }

    if ($err) {
        if (ref $err) {
            $task->{message} = $err->{error_message};
        }
        else {
            $task->{message} = $err;
        }

        if ($task->{timeout_seconds} && $task->{message} =~ /Timeout triggered/) {
            $task->{message} = "timeoutSeconds expired. ($task->{timeout_seconds})";
        }

        $self->{logger}->info("Bulk Task failed: $task->{message}");
        $task->{errors} = [$task->{message}];
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
        unlink ($result_filename) if ($result_filename);
        delete $task->{result_file};
    }
    else {
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
        
    }

    if ($result_filename) {
        $task->{file_size} = -s $result_filename;
        close ($result_fh);
    }
}


#----------------------------------------------------------------------
sub execute_bulk_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'bulk',
                                                          execute_func => sub { bulk(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub __result_header {
    my (%params) = @_;
    my $task = $params{task};

    my $datetime = F5::DateTimeUtils::convert_epoch_to_rfc3339($task->{start_time});
    return <<"EndHeader";
{
    "id": "$task->{rest_uuid}",
    "status": "FINISHED",
    "created": "$datetime",
    "timeout": $task->{timeout_seconds},
    "responses": [
EndHeader
}

sub  __result_response { 
    my (%params) = @_;
    my $rest_request = $params{rest_request};
    my $rv = $params{rv};
    my $exception = $params{exception};
    my $call_id = $params{call_id};

    my $comma = $call_id > 1 ? ',' : '';
    my $body = $exception ? $exception->{error_message} : $rv;
    my $status = $exception                              ? $exception->{rest_code}
               : ($rest_request->{rest_method} eq 'GET') ? 200
               : 201;

    return <<"EndResponse";
$comma {
            "method": "$rest_request->{rest_method}",
            "uri": "$rest_request->{original_uri}",
            "response" : {
                "status": $status,
                "body": $body
            }
       }
EndResponse
}

sub __result_footer {
    return ']}';
}

#----------------------------------------------------------------------

sub delete_many_bulk_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
