package F5::ASMConfig::Handler::DataProtection;

use strict;
use warnings;

use F5::Logger;
use F5::SetActive;
use F5::PrepareConf::DataProtection;
use JSON -convert_blessed_universally;
use F5::ObjUtils;
use MIME::Base64;
use F5::DataProtection;
use F5::SystemUtils;
use F5::ASMConfig::Entity::ImportDataProtectionTask;

use base qw/F5::ASMConfig::Handler::Base/;

our @handled_functions = (
                           {
                             function_name => 'update_data_protection',
                             relayable     => 1,
                             criteria_index => 3,
                             store_index   => 4,
                           },
                           {
                             function_name => 'import_data_protection',
                             relayable => 1,
                             filename_arguments => ['3.filename']
                           },
                           {
                             function_name => 'export_data_protection',
                             write_access => 1, 
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::DataProtection';
my $SECONDS_PER_MINUTE = 60;
my $GPG_CMD = '/usr/bin/gpg';

#----------------------------------------------------------------------

sub update_data_protection {
    my ($self, $session, $call_flags, $place_holder, $data_protection_criteria, $store_data_protection) = @_;

    my $data_protection = F5::ASMConfig::Entity::DataProtection->new(dbh => $self->{dbh},
                                                                     logger => $self->{logger},
                                                                     get_criteria => {});

    if ($data_protection->has_value_changed(existing_obj => $data_protection,
                                            store_obj => $store_data_protection,
                                            check_fields => 'data_protection_cookie_configurations')) {
        $store_data_protection->{regen_key} = 1;
        if ( !defined($store_data_protection->{grace_signing_interval}) ) {
            $store_data_protection->{grace_signing_interval} = F5::ASMConfig::Entity::ImportDataProtectionTask->fields->{grace_signing_interval}->{default_value};
        } 
        if ( !defined($store_data_protection->{grace_accepting_interval}) ) {
            $store_data_protection->{grace_accepting_interval} = F5::ASMConfig::Entity::ImportDataProtectionTask->fields->{grace_accepting_interval}->{default_value};
        }
    }

    my $now = time();
    #Used passed in key for CMI and for Import
    if ($store_data_protection->{regen_key} && ! $store_data_protection->{secure_key}) {
        my $new_key = F5::DataProtection::generate_key();     
        $store_data_protection->{secure_key} = $new_key;
    }

    my $security_context_changed = ($store_data_protection->{secure_key} && $store_data_protection->{secure_key} ne $data_protection->{secure_key});
    if ($security_context_changed) {
        $store_data_protection->{key_gen_time} = $now;
    }

    if ($call_flags->{rest_call}) {
        undef $store_data_protection->{grace_signing_end_time};
        undef $store_data_protection->{grace_accepting_end_time};
    }

    if (defined($store_data_protection->{grace_signing_interval})) {
        validate_interval(self => $self, obj => $store_data_protection, field => 'grace_signing_interval');
        $store_data_protection->{grace_signing_end_time} = $now + ($store_data_protection->{grace_signing_interval} * $SECONDS_PER_MINUTE);
    }
    if (defined($store_data_protection->{grace_accepting_interval})) {
        validate_interval(self => $self, obj => $store_data_protection, field => 'grace_accepting_interval');
        $store_data_protection->{grace_accepting_end_time} = $now + ($store_data_protection->{grace_accepting_interval} * $SECONDS_PER_MINUTE);
    }

    my $rv = $self->base_update(entity_class => $entity_class,
                                rpc_params => [ $session, $call_flags,$place_holder, $data_protection_criteria, $store_data_protection ],);

    my $lock_file =  $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                                 logger => $self->{logger});

    # keep the lock until after the set_active is complete
    $locker->lock();

    my $prepareconf_data_protection = F5::PrepareConf::DataProtection->new(
                                                                           dbh     =>  $self->{dbh},
                                                                           logger  =>  $self->{logger},
                                                                           );
    $prepareconf_data_protection->prepare(security_context_changed => $security_context_changed);

    # Use set active's built-in transaction
    if (!$call_flags->{skip_set_active}) {
        $self->{dbh}->commit;
        F5::SetActive::set_active(dbh => $self->{dbh},
                                  logger => $self->{logger},
                                  do_update_core => 1,
                                  send_sync => 1,
                                  asm_config_handler => $self);
        $self->{dbh}->begin_work;
    }
    $locker->unlock();

    return $rv;
}

#----------------------------------------------------------------------

sub import_data_protection {
    my ($self, $session, $call_flags, $place_holder, $data_protection_import_export) = @_;

    my $dbh = $self->{dbh};

    my $import_text;
    my $data_protection;
    if ($data_protection_import_export->{filename} && -e $data_protection_import_export->{filename}) {
        open (my $FH, '<', $data_protection_import_export->{filename});
        while (my $line = <$FH>) {
            $import_text .= $line;
        }
    }
    elsif ($data_protection_import_export->{import_text}) {
        $import_text = $data_protection_import_export->{import_text};
    }
    elsif ($data_protection_import_export->{data_protection_cookie_configurations}) {
        $data_protection = F5::ASMConfig::DataProtection->new($data_protection_import_export);
        $data_protection->{regen_key} = 1;
        
    }

    if (!$import_text && !$data_protection) {
        return $self->handle_error(action => "import",
                                   entity_class => $entity_class,
                                   thrift_obj  => $data_protection_import_export,
                                   error_code   => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                                   error_thrown => "Missing Data Protection Configuration or Import File/Text",
                                   error_level  => "info",
                                  );
    }

    if ($import_text) {
        #Decode text into DataProtectionObject
        eval {
            my $scrambled_text = MIME::Base64::decode_base64($import_text);
            die unless $scrambled_text;
            my $decoded_text   = F5::SystemUtils::run_system_tool(STDIN => $scrambled_text,
                                                                  command_line_params => [ $GPG_CMD, '--passphrase', $GPG_CMD, '--batch', '--decrypt' ])->{standard_output};
            $data_protection = F5::ASMConfig::DataProtection->new(from_json($decoded_text));
        };
        if ($@) {
            return $self->handle_error(action => "import",
                                       thrift_obj  => $data_protection_import_export,
                                       error_code   => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                                       error_thrown => "Invalid or Corrupt Data Protection Import File/Text",
                                       error_level  => "info",
                                      );
        }
    }

    #Recast nested struct after deserializing
    my @data_protection_cookie_configurations = map { F5::ASMConfig::DataProtectionCookieConfiguration->new($_) } @{$data_protection->{data_protection_cookie_configurations}};
    $data_protection->{data_protection_cookie_configurations} = \@data_protection_cookie_configurations;

    my $now = time();
    if (defined($data_protection_import_export->{grace_signing_interval})) {
        validate_interval(self => $self, obj => $data_protection_import_export, field => 'grace_signing_interval');
        $data_protection->{grace_signing_end_time} = $now + ($data_protection_import_export->{grace_signing_interval} * $SECONDS_PER_MINUTE);
        $data_protection->{grace_signing_interval} = $data_protection_import_export->{grace_signing_interval};
    }
    if (defined($data_protection_import_export->{grace_accepting_interval})) {
        validate_interval(self => $self, obj => $data_protection_import_export, field => 'grace_accepting_interval');
        $data_protection->{grace_accepting_end_time} = $now + ($data_protection_import_export->{grace_accepting_interval} * $SECONDS_PER_MINUTE);
	$data_protection->{grace_accepting_interval} = $data_protection_import_export->{grace_accepting_interval};
    }

    if ($data_protection_import_export->{preview_only}) {
        $call_flags->{do_not_relay} = 1;
        return { return_value => $data_protection };
    }

    return { return_value => $self->update_data_protection($session, $call_flags, $place_holder, $data_protection) };
}

#----------------------------------------------------------------------
sub export_data_protection {
    my($self, $session, $call_flags, $place_holder, $data_protection_import_export) = @_;
    my $data_protection = F5::ASMConfig::Entity::DataProtection->new(dbh => $self->{dbh}, 
                                                                     logger => $self->{logger},
                                                                     get_criteria => {});

    my @export_fields = qw(data_protection_cookie_configurations secure_key);
    my $export_data;
    for my $field (@export_fields) {
        $export_data->{$field} = $data_protection->{$field};
    }
    my $json = JSON->new->convert_blessed->allow_blessed;
    my $serialized_text = $json->encode($export_data);
    my $scrambled_text = F5::SystemUtils::run_system_tool(STDIN => $serialized_text,
                                                          command_line_params => [ $GPG_CMD, '--passphrase', $GPG_CMD, '--batch', '--symmetric' ])->{standard_output};
    my $encoded_text = MIME::Base64::encode_base64($scrambled_text);
    $data_protection_import_export->{import_text} = $encoded_text;

    my $tmpfile = "/tmp/export.$$";
    
    if ($data_protection_import_export->{filename}) {
        if (open (my $FH, ">", $data_protection_import_export->{filename})) {
            print $FH $data_protection_import_export->{import_text};
            close($FH);
        }
    }

    return { return_value => $data_protection_import_export };
}

#----------------------------------------------------------------------
sub validate_interval {
    my (%params) = @_;
    my $self = $params{self};
    my $obj = $params{obj};
    my $field = $params{field};

    if ($obj->{$field} < 0) {
       my $field_display_name = $entity_class->get_field_display_name_string(field_name  => $field);

       return $self->handle_error(action => "update",
                                  entity_class => $entity_class,
                                  thrift_obj  => $obj,
                                  error_code   => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                                  error_thrown => "Invalid field. $field_display_name must be positive.",
                                  error_level  => "info",
                                 );
    }
}

#----------------------------------------------------------------------

1;
