package F5::ASMConfig::Handler::DbTransaction;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;

our @handled_functions = (
                            {
                             function_name => 'begin_work',
                             skip_authz => 1,
                            },
                            {
                             function_name => 'rollback',
                             skip_authz => 1,
                            },
                            {
                             function_name => 'commit',
                             skip_authz => 1,
                            },
                         );

#----------------------------------------------------------------------

sub begin_work {
  my($self, $session) = @_;
  $self->{logger}->debug("Starting db transaction") if $self->{logger};

  # begin_work was already done by Handler
  push (@{$self->{relay_transaction_rpc_depth_stack}}, $self->{relay_transaction_rpc_depth});

  if(!$self->{in_transaction}) {
      $self->{transaction_rpc_entry_depth} = $self->{rpc_entry_depth};
      $self->{rpc_entry_depth} = 1;
      @{$self->{transaction_relay_queue}} = ();
  }

  $self->{relay_transaction_rpc_depth} = ($self->{rpc_entry_depth} - 1); 
  $self->{in_transaction}++;
  return { return_value => 1 };
}

#----------------------------------------------------------------------

sub rollback {
  my ($self, $session) = @_;

  # If this is a an error that is bubbling up from nested calls, only execute the actual rollback on top level (rpc depth = 1)
  if (!$self->{transaction_error} || $self->{rpc_entry_depth} == 1) {
      $self->{logger}->debug("Rolling back db transaction") if $self->{logger};
      if ( !$self->{dbh}->{AutoCommit} ) {
          $self->{dbh}->rollback;
      }
      $self->{in_transaction} = 0;
      $self->{transaction_error} = 0;
      if ($self->{transaction_rpc_entry_depth}) {
          $self->{rpc_entry_depth} = $self->{transaction_rpc_entry_depth};
          delete $self->{transaction_rpc_entry_depth};
      }
  
      @{$self->{transaction_relay_queue}} = ();
      $self->{relay_transaction_rpc_depth} = 0;
      $self->{relay_transaction_rpc_depth_stack} = ();
      if ($self->{transaction_locker}) {
          for my $locker (values %{$self->{transaction_locker}}) {
              $locker->unlock_all();
          }
          delete $self->{transaction_locker};
      }
  }

  return { return_value => 1 };
}

#----------------------------------------------------------------------

sub commit {
  my($self, $session) = @_;
  $self->{logger}->debug("Committing db transaction") if $self->{logger};

  if ($self->{in_transaction}) {
      $self->{in_transaction}--;
  }

  $self->{relay_transaction_rpc_depth} = pop (@{$self->{relay_transaction_rpc_depth_stack}});
  
  if (!$self->{in_transaction}) {
  # commit ineffective with AutoCommit enabled
      if ( !$self->{dbh}->{AutoCommit} ) {
          $self->{dbh}->commit;
      }
      if ($self->{transaction_rpc_entry_depth}) {
          $self->{rpc_entry_depth} = $self->{transaction_rpc_entry_depth};
          delete $self->{transaction_rpc_entry_depth};
      }
      for my $relay_params_ref (@{$self->{transaction_relay_queue}}) {
          $self->send_to_relay(%{$relay_params_ref});
      }
      @{$self->{transaction_relay_queue}} = ();

      if ($self->{update_mcp_timestamp}) {
          $self->update_mcp_timestamp();
          $self->{update_mcp_timestamp} = 0;
      }

      #Clean up transaction locker
      if ($self->{transaction_locker}) {
          for my $locker (values %{$self->{transaction_locker}}) {
              $locker->unlock_all();
          }
          delete $self->{transaction_locker};
      }
  }

  return { return_value => 1 };
}

#----------------------------------------------------------------------

1;
