package F5::ASMConfig::Handler::DeviceGroup;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use File::Basename;
use F5::DeviceGroup;
use F5::LockFactory;

our @handled_functions = (
                           {
                             function_name => "search_device_group",
                           },
                           {
                             function_name => "update_device_group",
                             criteria_index => 3,
                             store_index    => 4,
                             write_access   => 1,
                           },
                         );

my $entity_class = 'F5::ASMConfig::Entity::DeviceGroup';

#----------------------------------------------------------------------

sub update_device_group {
    my ($self, $session, $call_flags, $place_holder, $device_group_criteria, $store_device_group) = @_;

    my @events;
    my $mcp_device_groups = F5::DeviceGroup::get_device_groups_for_device($self->{shared_state}{local_device_name});
    my $mcp_device_group;
    if ($mcp_device_groups && @$mcp_device_groups) {
        ($mcp_device_group) = grep { $_->{name} eq $device_group_criteria->{name} } @$mcp_device_groups;
    }
    if (!$mcp_device_group) {
        $self->handle_error(action => 'update',
                            error_thrown => $F5::ASMConfig::Handler::NONEXISTENT_RECORD_ERROR,
                            call_flags => $call_flags,
                            thrift_obj => $device_group_criteria,
                           );
    }

    my $current_device_group = $self->{device_group};
    
    my $known_times_lock = F5::LockFactory::get_locker(type => "IPC",
                                                       ipc  => $self->{known_times_ipc});
    my $shared_state_lock = F5::LockFactory::get_locker(type => "IPC",
                                                        ipc  => $self->{shared_state_ipc});

    if ($current_device_group && defined($store_device_group->{asm_sync}) &&
        ($store_device_group->{asm_sync} ^ $current_device_group eq $device_group_criteria->{name}) ) {
        # If we turned off asm_sync on current device group or turned it on on a different one, turn off asm_sync on our current device group first

        $known_times_lock->lock();
        $shared_state_lock->lock();

        my ($old_device_group) = grep { $_->{name} eq $current_device_group } @$mcp_device_groups;
        eval {
            $old_device_group->{asm_sync} = 0;
            $old_device_group->save();
            F5::ASMConfig::Handler::SyncHandler::handle_device_group_change(self => $self,
                                                                            modified_device_group => $old_device_group->{name},
                                                                            update_topology => 1);
            
        };
        my $err = $@;
        $known_times_lock->unlock();
        $shared_state_lock->unlock();

        if ($err) {
            $self->handle_error(action => 'update',
                                error_thrown => $@,
                                error_code   => F5::ASMConfig::ErrorCode::COAPI_QUERY_FAILURE(),
                                call_flags => $call_flags,
                                thrift_obj => $device_group_criteria,
                               );
        }
    }

    if ($store_device_group->{asm_sync} && !$mcp_device_group->{asm_sync}) {

        #Turn on asm_sync flag then create then update_mcp_timestamp to trigger a push from this device to all others in the dg
        if ($current_device_group) {
            #If we are switching from an old device group give 2 seconds for the old connections to clear
            sleep 2;
        }

        $known_times_lock->lock();
        $shared_state_lock->lock();

        $mcp_device_group->{asm_sync} = $store_device_group->{asm_sync};        
        eval {
            $mcp_device_group->save();
            F5::ASMConfig::Handler::SyncHandler::handle_device_group_change(self => $self,
                                                                            modified_device_group => $mcp_device_group->{name},
                                                                            update_topology => 1);
            $self->{shared_state}{sync_init} = 0;
            $self->update_mcp_timestamp();
        };
        my $err = $@;
        $known_times_lock->unlock();
        $shared_state_lock->unlock();

        if ($err) {
            $self->handle_error(action => 'update',
                                error_thrown => $@,
                                error_code   => F5::ASMConfig::ErrorCode::COAPI_QUERY_FAILURE(),
                                call_flags => $call_flags,
                                thrift_obj => $device_group_criteria,
                               );

        }
    }
    # Update event_dispatcher and ActionItemHandler for the current DG state
    push @events, $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                      session => $session,
                                      element => F5::ASMConfig::Entity::DeviceGroup->new(dbh => $self->{dbh}, object => {name => $self->{device_group}} ),
                                      call_flags => $call_flags,
                                      user_name => $session->{user_name},
                                      suppress_save_to_log => 1,
                                     );
    return {return_value => F5::ASMConfig::DeviceGroup->new($mcp_device_group),
            events => \@events,
           };
}
#----------------------------------------------------------------------

sub search_device_group {
    my ($self, $session, $call_flags, $place_holder, $device_group_criteria) = @_;
    $self->{local_device_name} = F5::DeviceGroup::get_local_device_name();
    
    my $mcp_device_groups = F5::DeviceGroup::get_device_groups_for_device($self->{shared_state}{local_device_name});
    my @device_groups;
    if ($mcp_device_groups && @$mcp_device_groups) {
GROUP: for my $mcp_device_group (@$mcp_device_groups) {
           if ($device_group_criteria) {
               for my $criteria (keys %$device_group_criteria) {
                   if (defined $device_group_criteria->{$criteria} && defined($mcp_device_group->{$criteria})
                       && $device_group_criteria->{$criteria} ne $mcp_device_group->{$criteria}) {
                      next GROUP; 
                   }
               }
           }
           push @device_groups, F5::ASMConfig::DeviceGroup->new($mcp_device_group);
        }
    }

    my $result = $self->entities_to_search_result(entities => \@device_groups,
                                                  call_flags => $call_flags,
                                                  entity_class => $entity_class,
                                                 );
   return { return_value => $result };
}

#----------------------------------------------------------------------


1;
