package F5::ASMConfig::Handler::Diff;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::Entity::Diff;
use F5::ASMConfig::Entity::Policy;

our @handled_functions = (
                            {
                             function_name => 'add_diff',
                             relayable => 0,
                            },
                            {
                             function_name => 'auto_merge_diff',
                             relayable => 0,
                             skip_authz => 1, #handled by apply_diff_item
                            },

                         );

our $entity_class = 'F5::ASMConfig::Entity::Diff';

#----------------------------------------------------------------------

sub add_diff {
    my ($self, $session, $call_flags, $placeholder, $thrift_diff) = @_;
    my $first_policy  = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $thrift_diff->{first_policy});
    my $second_policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $thrift_diff->{second_policy});

    my @diff_validation_fields = qw(encoding
                                    flg_is_case_insensitive
                                    flg_is_protocol_free);

    for my $validation_field (@diff_validation_fields) {
        if (lc($first_policy->{$validation_field}) ne lc($second_policy->{$validation_field})) {
            my $error = "The policies have an incompatible '" . $first_policy->get_field_display_name_string(field_name => $validation_field)
                      . "' configuration";
            return $self->handle_error(action => "calculate",
                                       entity_class => $entity_class,
                                       thrift_obj  => $thrift_diff,
                                       error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                       error_thrown => $error,
                                       error_level  => "info",
                                );
        }
    }

    my $local_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags,
                                                         start_page => 2,
                                                         page_size => 4,
                                                         sort_orders => [ F5::ASMConfig::SortOrder->new(field_name => 'created_time',
                                                                                                        sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC()) ],
                                                         );
    $self->delete_many_diffs($session, $local_call_flags, $placeholder, $thrift_diff);

    $thrift_diff->{last_updated} = List::Util::max($first_policy->get_last_updated, $second_policy->get_last_updated) || time();
    $thrift_diff->{gui_user_name} = $session->{gui_user_name};
    $thrift_diff->{created_time} = time();
  
    my $rv = $self->base_add(rpc_params => [$session, $call_flags, $placeholder, $thrift_diff],
                             entity_class => $entity_class,);
    my $diff = $rv->{return_value};
  
    ## Get list of diffable entities
    my @diffable_entities = map {$_->{class}} grep { $_->{class} && $_->{class}->diffable } values %{$self->{element_types}};

    if ($first_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() ||
        $second_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()
       ) {
        @diffable_entities = grep {!$_->entity_config->{no_inheritance}} @diffable_entities;
    }
  
    #my @diffable_entities = qw(F5::ASMConfig::Entity::Filetype F5::ASMConfig::Entity::URL);
    ## Insert DiffItems for each list
    my @diff_summary;
    my @diff_items;
    for my $entity_type (@diffable_entities) {
        my $search_filters = $call_flags->{is_in_declarative_export} ? $entity_type->declarative_search_filters
                                                                     : $entity_type->diff_search_filters;

        if (($first_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() ||
            $second_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT())
            && $entity_type->entity_config->{no_inheritance_conditions}
           ) {
            push (@$search_filters, map { $entity_type->negate_search_filter(search_filter => F5::ASMConfig::SearchFilter->new($_)) } @{$entity_type->entity_config->{no_inheritance_conditions}});
        }

        my $first_iter = $entity_type->retrieve_all(policy => $first_policy, 
                                                    dbh => $self->{dbh}, 
                                                    sort_orders => $entity_type->natural_sort_orders,
                                                    search_filters => $search_filters,);
        my $second_iter = $entity_type->retrieve_all(policy => $second_policy, 
                                                     dbh => $self->{dbh},
                                                     sort_orders => $entity_type->natural_sort_orders,
                                                     search_filters => $search_filters,);
     
        #use batched inserter?      
   
        my $summary = {only_first_count  => 0,
                       only_second_count => 0,
                       conflict_count    => 0,
                       element_type      => $entity_type->element_type};
 
        my $first_obj = $first_iter->get_next();
        my $second_obj = $second_iter->get_next();

        my $strict_inheritance = ( $first_policy->{has_parent} || $second_policy->{has_parent} ) ? $entity_type->strict_inheritance() : 0;
        my $is_inherited_first = ( $first_policy->{has_parent} && $strict_inheritance ) ? F5::ASMConfig::Entity::Section->new(dbh => $self->{dbh},
                                                                                                                          policy => $first_policy,
                                                                                                                          get_criteria => {section_type => $entity_type->get_section_type()},
                                                                                                                         )->is_inherited() : 0;

        my $is_inherited_second = ( $second_policy->{has_parent} && $strict_inheritance ) ? F5::ASMConfig::Entity::Section->new(dbh => $self->{dbh},
                                                                                                                            policy => $second_policy,
                                                                                                                            get_criteria => {section_type => $entity_type->get_section_type()},
                                                                                                                            )->is_inherited() : 0;

        while ($first_obj || $second_obj) {
            my $diff_item;
            my $diff_type;
            if (!$first_obj) {
                $diff_type =  F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_SECOND();
            }
            elsif (!$second_obj) {
                $diff_type = F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_FIRST();
            }
            else {
                my $first_obj_cmp_str = $first_obj->unique_record_parts_str_for_diff;
                my $second_obj_cmp_str = $second_obj->unique_record_parts_str_for_diff;
                if ($first_obj_cmp_str eq $second_obj_cmp_str) {
                    $diff_type = F5::ASMConfig::DiffType::DIFF_TYPE_CONFLICT();
                }
                else {
                    $diff_type = ($first_obj_cmp_str lt $second_obj_cmp_str) ? F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_FIRST()
                                                                             : F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_SECOND();
                }
            }

            my $can_be_merged_first_to_second = ( $strict_inheritance && $is_inherited_second ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                               : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
            my $can_be_merged_second_to_first = ( $strict_inheritance && $is_inherited_first ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                              : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();

            if ($diff_type == F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_FIRST()) {
                if ($entity_type->is_valid_diff_type(diff_type => $diff_type, obj => $first_obj)) {
                    if ( $first_obj->{is_inherited} ) {
                        $can_be_merged_second_to_first = F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED();
                    }
                    $diff_item = F5::ASMConfig::Entity::DiffItem->new(dbh => $self->{dbh},
                                                                      object => { diff_id      => $diff->get_id,
                                                                                  diff_type    => $diff_type,
                                                                                  element_type => $entity_type->element_type,
                                                                                  element_kind => $first_obj->get_kind(),
                                                                                  element_name => $first_obj->get_full_display_name(),
                                                                                  first_element_id => $first_obj->get_unique_id(default => 0),
                                                                                  first_element_timestamp => $first_obj->get_last_updated(),
                                                                                  can_be_merged_first_to_second => $can_be_merged_first_to_second,
                                                                                  can_be_merged_second_to_first => $can_be_merged_second_to_first,
                                                                                },
                                                                     );
                    $summary->{only_first_count}++;
                }
                $first_obj = $first_iter->get_next();
            }
            elsif ($diff_type == F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_SECOND()) {
                if ($entity_type->is_valid_diff_type(diff_type => $diff_type, obj => $second_obj)) {
                    if ( $second_obj->{is_inherited} ) {
                        $can_be_merged_first_to_second = F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED();
                    }
                    $diff_item = F5::ASMConfig::Entity::DiffItem->new(dbh => $self->{dbh},
                                                                      object => { diff_id      => $diff->get_id,
                                                                                  diff_type    => $diff_type,
										  element_type => $entity_type->element_type,
										  element_kind => $second_obj->get_kind(),
										  element_name => $second_obj->get_full_display_name(),
										  second_element_id  => $second_obj->get_unique_id(default => 0),
										  second_element_timestamp => $second_obj->get_last_updated(),
										  can_be_merged_first_to_second => $can_be_merged_first_to_second,
										  can_be_merged_second_to_first => $can_be_merged_second_to_first,
                                                                                },
                                                                     );
                    $summary->{only_second_count}++;
                }
                $second_obj = $second_iter->get_next();
            }
            else {
              my @diff_item_details;
              my $incremental_id = 0;
              if ($entity_type->is_valid_diff_type(diff_type => $diff_type, obj => $first_obj)) {
                  my @diff_fields = $first_obj->get_differing_fields(obj2 => $second_obj);
                  for my $diff_field (@diff_fields) {
                      my ($field, $sub_field) = split(/\./, $diff_field, 2);
                      my @details = @{get_details(first_obj => $sub_field ? $first_obj->{$field} : $first_obj,
                                                  second_obj => $sub_field ? $second_obj->{$field} : $second_obj,
                                                  field => $sub_field ? $sub_field : $field,
                                                  diff_field => $diff_field,
                                                  incremental_id_ref => \$incremental_id,)};
                      push (@diff_item_details, @details) if @details;
                  }
              }
              if (@diff_item_details) {

                  my $allowed_merges_first_to_second = 0;
                  my $allowed_merges_second_to_first = 0;
                  for my $diff_item_detail (@diff_item_details) {
                      if ( $diff_item_detail->{can_be_merged_first_to_second} == F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED() ) {
                          $allowed_merges_first_to_second++;
                      }
                      if ( $diff_item_detail->{can_be_merged_second_to_first} == F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED() ) {
                          $allowed_merges_second_to_first++;
                      }
                  }

                  my $diff_items_len = scalar(@diff_item_details);
                  $can_be_merged_first_to_second = ( $allowed_merges_first_to_second == 0 ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                 : ( $allowed_merges_first_to_second == $diff_items_len ) ? F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED()
                                                 : F5::ASMConfig::AllowedMergeType::MERGE_BY_DETAILS();
                  $can_be_merged_second_to_first = ( $allowed_merges_second_to_first == 0 ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                 : ( $allowed_merges_second_to_first == $diff_items_len ) ? F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED()
                                                 : F5::ASMConfig::AllowedMergeType::MERGE_BY_DETAILS();

                  $diff_item = F5::ASMConfig::Entity::DiffItem->new(dbh => $self->{dbh},
                                                                    object => { diff_id      => $diff->get_id,
									        diff_type    => $diff_type,
										element_type => $entity_type->element_type,
										element_kind => $first_obj->get_kind(),
										element_name => $first_obj->get_full_display_name(),
										first_element_id  => $first_obj->get_unique_id(default => 0),
										second_element_id  => $second_obj->get_unique_id(default => 0),
										first_element_timestamp  => $first_obj->get_last_updated(),
										second_element_timestamp => $second_obj->get_last_updated(),
										diff_item_details => \@diff_item_details,
										can_be_merged_first_to_second => $can_be_merged_first_to_second,
										 can_be_merged_second_to_first => $can_be_merged_second_to_first,
                                                                              },
                                                                   );
                  $summary->{conflict_count}++;  
              }
              else {
                  $summary->{match_count}++;
              }
              $first_obj  = $first_iter->get_next();
              $second_obj = $second_iter->get_next();
          }
          if (defined $diff_item) {
             if (!$call_flags->{is_in_declarative_export}) {
                  $diff_item->store_add();
             }
             else {
                  push (@diff_items, $diff_item);
             }
          }
      }
      if ($summary->{only_first_count} || $summary->{only_second_count} || $summary->{conflict_count}) {
          push (@diff_summary, F5::ASMConfig::DiffSummary->new($summary));
      }
    }
    $diff->{diff_summary} = \@diff_summary;
    if ($call_flags->{is_in_declarative_export}) {
        $diff->{diff_items} = \@diff_items;
    }
    $rv->{return_value} = $diff;
    return $rv;
}

#----------------------------------------------------------------------

sub get_details {
    my (%params) = @_;
    my $first_obj  = $params{first_obj};
    my $second_obj = $params{second_obj};
    my $field      = $params{field};
    my $diff_field = $params{diff_field};

    my $diff_structs = $first_obj->diff_structs;
    my $field_def = $diff_structs->{$field} || $first_obj->fields->{$field};
#    field_def => $sub_field ? $field_def->{$sub_field} : $field_def,

    my $diff_class = ref($first_obj);

    my @diff_item_details;

    my $can_be_merged_first_to_second;
    my $can_be_merged_second_to_first;

    if ( !( $field_def->{exclude_from_inheritance} || $F5::ASMConfig::Entity::Base::common_inheritance_excludes->{$field} )) {
            $can_be_merged_first_to_second = ( $second_obj->{is_inherited} && $second_obj->{has_parent} ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                                          : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();

            $can_be_merged_second_to_first = ( $first_obj->{is_inherited} && $first_obj->{has_parent} ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                                        : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
    }
    else {
        $can_be_merged_first_to_second = F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
        $can_be_merged_second_to_first = F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
    }

    my $detail_id_ref = $params{incremental_id_ref};
    if ($field_def->{nested_struct}) {
        my $nested_def = $field_def;

        my $strict_inheritance = 0;
        my $is_inherited_first = 0;
        my $is_inherited_second = 0;
        if ( $field_def->{diff_struct} ) {
            if ( $first_obj->{policy}->{has_parent} || $second_obj->{policy}->{has_parent} ) {
                $strict_inheritance = $field_def->{entity_type}->strict_inheritance();
            }
            $is_inherited_first = $first_obj->{policy}->{has_parent} ? F5::ASMConfig::Entity::Section->new(dbh => $first_obj->{dbh},
                                                                                                           policy => $first_obj->{policy},
                                                                                                           get_criteria => {section_type => $first_obj->get_section_type()}
                                                                                                          )->is_inherited() : 0;
            $is_inherited_second = $second_obj->{policy}->{has_parent} ? F5::ASMConfig::Entity::Section->new(dbh => $second_obj->{dbh},
                                                                                                             policy => $second_obj->{policy},
                                                                                                             get_criteria => {section_type => $second_obj->get_section_type()}
                                                                                                            )->is_inherited() : 0;
        }
 
        my $first_nested_records  = $first_obj->{$field};
        my $second_nested_records = $second_obj->{$field};
        if (ref($first_nested_records) eq 'ARRAY') {
            $first_nested_records = $first_obj->hashify_nested_struct_list(nested_def => $nested_def,
                                                                           nested_recs => $first_nested_records);
        }
        if (ref($second_nested_records) eq 'ARRAY') {
            $second_nested_records = $second_obj->hashify_nested_struct_list(nested_def => $nested_def,
                                                                             nested_recs => $second_nested_records);
        }

        for my $nested_key (keys %$first_nested_records) {

            if ( $nested_def->{entity_type} ) {
                if ( $nested_def->{fields} && $nested_def->{fields}->{is_inherited} ) {
                    $can_be_merged_second_to_first = ( $first_nested_records->{$nested_key}->{is_inherited} ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                                              : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
                }
            }
            else {
                $can_be_merged_second_to_first = $first_obj->{is_inherited} ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                            : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
            }

            if (!$second_nested_records->{$nested_key}) {
                next if $field_def->{conflict_only};
                $can_be_merged_first_to_second = ( $strict_inheritance && $is_inherited_second ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                                 : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
                my ($first_entity_change) = $first_obj->get_entity_changes(fields      => [$field],
                                                                           field_values => { $field => [$first_nested_records->{$nested_key}] });
                my $field_display_name = $first_entity_change->{field_display_name};
                next unless $field_display_name;

                my $first_element = $diff_class->new(object=>{$field => [$first_nested_records->{$nested_key}]});
                my $second_element = $diff_class->new(object=>{$field => []});

                push (@diff_item_details,
                      F5::ASMConfig::DiffItemDetail->new({ field_name          => "$diff_field.$nested_key",
                                                           field_display_name  => $field_display_name,
                                                           first_diff_display  => $first_entity_change->{field_display_value},
                                                           first_element => $first_element,
                                                           second_element => $second_element,
                                                           can_be_merged_first_to_second => $can_be_merged_first_to_second,
                                                           can_be_merged_second_to_first => $can_be_merged_second_to_first,
                                                           detail_id_by_item => ++$$detail_id_ref,
                                                         }));
           }
           elsif ($nested_def->{entity_type}) {
               if ($nested_def->{fields} && $nested_def->{fields}->{is_inherited} ) {
                   $can_be_merged_first_to_second = ( $second_nested_records->{$nested_key}->{is_inherited} ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                                              : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
               }
               my $first_rec = $first_nested_records->{$nested_key};
               my $second_rec = $second_nested_records->{$nested_key};
               my @nested_field_names = $nested_def->{fields} ? grep { !$nested_def->{fields}->{$_}->{unique_record_part} && 
                                                                       !$nested_def->{fields}->{$_}->{exclude_from_diff} } keys %{$nested_def->{fields}}
                                                              : @{$nested_def->{entity_type}->diff_fields};

               for my $nested_field (@nested_field_names) {
                   if ((defined $first_rec->{$nested_field}) != (defined $second_rec->{$nested_field}) ||
                       (defined $first_rec->{$nested_field} &&
                        $first_rec->{$nested_field} ne $second_rec->{$nested_field})) {
                       if ((ref($first_rec->{$nested_field})  =~ /::Entity::/) &&
                           (ref($second_rec->{$nested_field}) =~ /::Entity::/)) {
                           next if $first_rec->{$nested_field}->unique_record_parts_str_for_diff eq $second_rec->{$nested_field}->unique_record_parts_str_for_diff;
                       }
                       my ($first_entity_change) = $first_obj->get_entity_changes(fields      => [$field],
                                                                                  field_values => { $field => [$first_nested_records->{$nested_key}] });
                       my ($second_entity_change) = $second_obj->get_entity_changes(fields      => [$field],
                                                                                    field_values => { $field => [$second_nested_records->{$nested_key}] });
                       my $field_display_name = $first_entity_change->{field_display_name} || $second_entity_change->{field_display_name};
                       next unless $field_display_name;

                       my $first_element = $diff_class->new(object=>{$field => [$first_nested_records->{$nested_key}]});
                       my $second_element = $diff_class->new(object=>{$field => [$second_nested_records->{$nested_key}]});

                       push (@diff_item_details,
                             F5::ASMConfig::DiffItemDetail->new({field_name          => "$diff_field.$nested_key",
                                                                 field_display_name  => $field_display_name,
                                                                 first_diff_display  => $first_entity_change->{field_display_value},
                                                                 second_diff_display => $second_entity_change->{field_display_value},
                                                                 first_element => $first_element,
                                                                 second_element => $second_element,
                                                                 can_be_merged_first_to_second => $can_be_merged_first_to_second,
                                                                 can_be_merged_second_to_first => $can_be_merged_second_to_first,
                                                                 detail_id_by_item => ++$$detail_id_ref,
                                                            }));

                       last;
                   }
               }
           }
        }
        for my $nested_key (keys %$second_nested_records) {
            if (!$first_nested_records->{$nested_key}) {
                next if $field_def->{conflict_only};

                if ( $nested_def->{entity_type} ) {
                    if ( $nested_def->{fields} && $nested_def->{fields}->{is_inherited} ) {
                        $can_be_merged_first_to_second = ( $second_nested_records->{$nested_key}->{is_inherited} ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                                                   : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
                    }
                }
                else {
                    $can_be_merged_first_to_second = $second_obj->{is_inherited} ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                 : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();
                }
                $can_be_merged_second_to_first = ( $strict_inheritance && $is_inherited_first ) ? F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()
                                                                                                : F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED();

                my ($second_entity_change) = $second_obj->get_entity_changes(fields      => [$field],
                                                                             field_values => { $field => [$second_nested_records->{$nested_key}] });
                my $field_display_name = $second_entity_change->{field_display_name};
                next unless $field_display_name;

                my $first_element = $diff_class->new(object=>{$field => []});
                my $second_element = $diff_class->new(object=>{$field => [$second_nested_records->{$nested_key}]});

                push (@diff_item_details,
                      F5::ASMConfig::DiffItemDetail->new({ field_name           => "$diff_field.$nested_key",
                                                           field_display_name   => $field_display_name,
                                                           second_diff_display  => $second_entity_change->{field_display_value},
                                                           first_element => $first_element,
                                                           second_element => $second_element,
                                                           can_be_merged_first_to_second => $can_be_merged_first_to_second,
                                                           can_be_merged_second_to_first => $can_be_merged_second_to_first,
                                                           detail_id_by_item => ++$$detail_id_ref,
                                                         }));
            }
        }
    }
    else {
        my ($first_entity_change) = $first_obj->get_entity_changes(fields      => [$field],);
        my ($second_entity_change) = $second_obj->get_entity_changes(fields      => [$field],);

        my $first_field = $first_entity_change->{field};
        my $second_field = $second_entity_change->{field};

        my $field_display_name = $first_entity_change->{field_display_name} || $second_entity_change->{field_display_name};
        if ($field_display_name) {
            my $first_element = $diff_class->new(object=>{$first_field => $first_obj->{$first_field}, $second_field => $first_obj->{$second_field}});
            my $second_element = $diff_class->new(object=>{$first_field => $second_obj->{$first_field}, $second_field => $second_obj->{$second_field}});
            push (@diff_item_details,
                  F5::ASMConfig::DiffItemDetail->new({ field_name          => $diff_field,
                                                       field_display_name  => $field_display_name,
                                                       first_diff_display  => $first_entity_change->{field_display_value},
                                                       second_diff_display => $second_entity_change->{field_display_value},
                                                       first_element => $first_element,
                                                       second_element => $second_element,
                                                       can_be_merged_first_to_second => $can_be_merged_first_to_second,
                                                       can_be_merged_second_to_first => $can_be_merged_second_to_first,
                                                       detail_id_by_item => ++$$detail_id_ref,
                                                    }));
       }
    }
    return \@diff_item_details;
}

#----------------------------------------------------------------------

sub auto_merge_diff {
    my ($self, $session, $call_flags, $placeholder, $thrift_diff, $auto_merge_flags) = @_;

    my $diff = $entity_class->new(dbh => $self->{dbh},
                                  logger => $self->{logger},
                                  get_criteria => $thrift_diff);

    if (!$diff) {
        return $self->handle_error(action => "auto_merge_diff",
                                   entity_class => $entity_class,
                                   thrift_obj  => $thrift_diff,
                                   error_code   => F5::ASMConfig::ErrorCode::POLICY_MERGE_ERROR,
                                   error_thrown => "Policy Merge error.  Specified Diff report not found.",
                                   error_level  => "info",
                                  );
    }

    $self->begin_work($self->{session});

    my $rv;
    my $has_inherited_fields = 0;
    if ($auto_merge_flags->{add_all_missing_to_first}) {
        $rv = $self->apply_diff_item($session, $call_flags, F5::ASMConfig::DiffItem->new({ diff_id   => $diff->get_id,
                                                                                     diff_type => F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_SECOND() }),
                                                            F5::ASMConfig::DiffFlags->new({ diff_direction => F5::ASMConfig::DiffDirection::DIFF_DIRECTION_SECOND_TO_FIRST(), }),
                               );
        $has_inherited_fields = $rv->{has_inherited_fields} ? 1 : $has_inherited_fields;
    }
    if ($auto_merge_flags->{add_all_missing_to_second}) {
        $rv = $self->apply_diff_item($session, $call_flags, F5::ASMConfig::DiffItem->new({ diff_id   => $diff->get_id,
                                                                                     diff_type => F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_FIRST() }),
                                                            F5::ASMConfig::DiffFlags->new({ diff_direction => F5::ASMConfig::DiffDirection::DIFF_DIRECTION_FIRST_TO_SECOND(), }),
                               );
        $has_inherited_fields = $rv->{has_inherited_fields} ? 1 : $has_inherited_fields;
    }
    if ($auto_merge_flags->{accept_conflicts_direction}) {
        $rv = $self->apply_diff_item($session, $call_flags, F5::ASMConfig::DiffItem->new({ diff_id   => $diff->get_id,
                                                                                     diff_type => F5::ASMConfig::DiffType::DIFF_TYPE_CONFLICT() }),
                                                            F5::ASMConfig::DiffFlags->new({ diff_direction => $auto_merge_flags->{accept_conflicts_direction} }),
                               );
        my $target_policy = ($auto_merge_flags->{accept_conflicts_direction} == F5::ASMConfig::DiffDirection::DIFF_DIRECTION_FIRST_TO_SECOND()) ? $diff->{second_policy}
                                                                                                                                                : $diff->{first_policy};
        $has_inherited_fields = $rv->{has_inherited_fields} ? 1 : $has_inherited_fields;
    }
    $self->commit($self->{session});

    my $retval = F5::ASMConfig::AutoMergeDiffResult->new({ completed_successfully => 1,
                                                           has_inherited_fields => $has_inherited_fields });
    return { return_value => $retval };
}

#----------------------------------------------------------------------

#TODO: return diff_summary on retrieve when present in extra_params  (possibly should be in Entity::Diff)

1;
