package F5::ASMConfig::Handler::DiffItem;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::Entity::DiffItem;
use F5::ASMConfig::Entity::Policy;

our @handled_functions = (
                            {
                             function_name => 'apply_diff_item',
                             relayable => 0,
                            },
                            {
                             function_name => 'update_diff_item',
                             relayable => 0,
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::DiffItem';

#----------------------------------------------------------------------

sub apply_diff_item {
    my ($self, $session, $call_flags, $placeholder, $thrift_diff_item, $thrift_diff_flags) = @_;
    my $diff_item_iter = $entity_class->retrieve_all(dbh => $self->{dbh},
                                                 get_criteria => $thrift_diff_item,
                                                 search_filters => $call_flags->{search_filters},);
    if (!$diff_item_iter->{total_record_count}) {
        my $retval = F5::ASMConfig::ApplyDiffItemResult->new({ completed_successfully => 1,
                                                               has_inherited_fields => 0 });
        return { return_value => $retval
               };
    }
    my $diff;
    my @thrift_diff_fields;
    if ($thrift_diff_item->{diff_item_details}) {
        @thrift_diff_fields = map { $_->{field_name} } @{$thrift_diff_item->{diff_item_details}};
    }

    my @events;
    my $seen_entities;
    my ($source, $target, $source_policy, $target_policy);
    my $is_merge_allowed_field;
    my $has_inherited_fields = 0;
    $self->begin_work($self->{session});
    while (my $diff_item = $diff_item_iter->get_next) {
        if (!$diff) {
            $diff = F5::ASMConfig::Entity::Diff->new(dbh => $self->{dbh},
                                                     get_criteria => { diff_id => $diff_item->{diff_id} });
            
            if ($thrift_diff_flags->{diff_direction} == F5::ASMConfig::DiffDirection::DIFF_DIRECTION_FIRST_TO_SECOND()) {
                $source = "first";
                $target = "second";
                $is_merge_allowed_field = "can_be_merged_first_to_second";
            }
            else {
                $source = "second";
                $target = "first";
                $is_merge_allowed_field = "can_be_merged_second_to_first";
            }
            $source_policy = $diff->{$source . "_policy"};
            $target_policy = $diff->{$target . "_policy"};
            $self->check_authz(write_access => 1, authz_objects => [$target_policy], function_name => "apply_diff_item");
        }

        if ($diff_item->{$is_merge_allowed_field} == F5::ASMConfig::AllowedMergeType::MERGE_NOT_ALLOWED()) {
            $has_inherited_fields = 1;
            next;
        }

        if ( $diff_item->{$is_merge_allowed_field} == F5::ASMConfig::AllowedMergeType::MERGE_BY_DETAILS() ) {
            $has_inherited_fields = 1;
            @{$diff_item->{diff_item_details}} = grep { $_->{$is_merge_allowed_field} == F5::ASMConfig::AllowedMergeType::MERGE_ALLOWED() } @{$diff_item->{diff_item_details}};
        }

        my ($source_entity, $target_entity);

        my $diff_class = $self->{element_types}->{$diff_item->{element_type}}->{class};

        if ( (my $id = $diff_item->{$source . "_element_id"}) || !$diff_class->unique_id_field) {
            $source_entity = $diff_class->new(dbh => $self->{dbh},
                                                get_criteria => $id ? { $diff_class->unique_id_field => $id } : {} ,
                                                policy => $source_policy);
        }
        if ((my $id = $diff_item->{$target . "_element_id"}) || !$diff_class->unique_id_field ) {
            $target_entity = $diff_class->new(dbh => $self->{dbh},
                                              get_criteria => $id ? { $diff_class->unique_id_field => $id } : {},
                                              policy => $target_policy);
        }

        my @diff_fields = @thrift_diff_fields ? @thrift_diff_fields
                                              : map { $_->{field_name} } @{$diff_item->{diff_item_details}};
        my @merge_list = @{$diff_class->get_modified_entities(source => $source_entity, target => $target_entity, fields => \@diff_fields)};

        my $merge_args = { 'add'    => {flags => {ignore_duplicate_record_on_add => 1,
                                                  internal_flags => [ 'skip_populate_defaults_on_add' ] }},
                           'delete' => {flags => {ignore_nonexistent_record_on_delete => 1}},
                           'update' => {flags => {add_on_updating_nonexistent_record => 1,
                                                  internal_flags => [ 'skip_populate_defaults_on_add' ] }},};

        for my $merge_item (@merge_list) {
            my $entity = $merge_item->{entity};
            if ($merge_item->{type} ne 'update' &&
                $seen_entities->{$merge_item->{type}}->{$entity->element_type}->{$entity->get_unique_id(default => 0)}++) {
                next;
            }
            
            my $function = $merge_item->{type} . "_" . $self->{element_types}->{$entity->element_type}->{entity_function_name};
            my $flags = {%$call_flags,
                         %{$merge_args->{$merge_item->{type}}->{flags}}, 
                         search_filters => $entity->diff_search_filters(),
                         set_fields_to_null => $merge_item->{set_fields_to_null},
                         force_virtual_field_update => 1};
            if ($merge_item->{type} ne 'delete') {
                $entity->remove_non_diff_fields();
            }
            $entity->{policy} = $target_policy;
            $entity->{policy_id} = $target_policy->{policy_id};

            if ($merge_item->{type} eq 'add') {
                #Retrieve first and only add if it does not exist in the target policy
                my $search_function =  "search_" . $self->{element_types}->{$entity->element_type}->{entity_function_name};
                my $search_result = $self->$search_function($session, $flags, $target_policy, $entity);
                next if ($search_result->{total_record_count});                
            }
            $self->$function($session, $flags, $target_policy, $entity);
        }   
        #Remove diff_item if all conflicts have been resolved
        if (@thrift_diff_fields && $diff_item->{diff_item_details}) {
            my %resolved_fields = map { ( $_ => 1) } @thrift_diff_fields;
            my @conflict_fields =  map { $_->{field_name} } @{$diff_item->{diff_item_details}};
            if ( grep { ! $resolved_fields{$_} } @conflict_fields ) {
                $target_entity->{dbh} = $self->{dbh};
                $diff_item->store_update(call_flags => {remove_nested_struct_on_update => 1},
                                         store_obj  => { diff_item_details              => $thrift_diff_item->{diff_item_details},
                                                         $target . "_element_timestamp" => $target_entity->get_last_updated(),} );
            }                      
            else {
                $diff_item->delete();
            }
        }
        else {
            $diff_item->delete();
        }
    }

    #Clean up affected diff items        
    if ($seen_entities->{add}){
        for my $element_type (keys %{$seen_entities->{add}}) {
            my $diff_type = ($thrift_diff_flags->{diff_direction} == F5::ASMConfig::DiffDirection::DIFF_DIRECTION_FIRST_TO_SECOND()) ? F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_FIRST()
                                                                                                                                     : F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_SECOND();
            $entity_class->delete_many(dbh => $self->{dbh},
                                       search_criteria => { element_type => $element_type, diff_type => $diff_type },
                                       call_flags => { search_filters => [ {
                                                       field_name => $source . "_element_id",
                                                       field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                       field_value_list => [  keys %{$seen_entities->{add}->{$element_type}} ],}] },);
        }
    }

    if ($seen_entities->{'delete'}){
        for my $element_type (keys %{$seen_entities->{'delete'}}) {
            my $diff_type = ($thrift_diff_flags->{diff_direction} == F5::ASMConfig::DiffDirection::DIFF_DIRECTION_FIRST_TO_SECOND()) ? F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_SECOND()
                                                                                                                                     : F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_FIRST();
            $entity_class->delete_many(dbh => $self->{dbh},
                                       search_criteria => { element_type => $element_type, diff_type => $diff_type },
                                       call_flags => { search_filters => [ {
                                                       field_name => $target . "_element_id",
                                                       field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                       field_value_list => [  keys %{$seen_entities->{'delete'}->{$element_type}} ],}] },);
        }
    }

    $self->commit($self->{session});

    my $retval = F5::ASMConfig::ApplyDiffItemResult->new({ completed_successfully => 1,
                                                           has_inherited_fields => $has_inherited_fields });
    return { return_value => $retval,
           };
}

#----------------------------------------------------------------------

sub update_diff_item {
    my($self, $session, $call_flags, $placeholder, $diff_item_criteria, $store_diff_item) = @_;

    my $local_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags, is_in_diff_merge => 1);
    my $diff_item = $entity_class->new(dbh => $self->{dbh},
                                       get_criteria => $diff_item_criteria);

    my $rv;
    if ( $store_diff_item->{diff_item_details} && $diff_item ) {
        my %details_by_direction;
        for my $detail (@{$store_diff_item->{diff_item_details}}) {
            if ( $detail->{merge_status} ) {
                my @field_names = map { $_->{field_name} } grep { $_->{detail_id_by_item} == $detail->{detail_id_by_item} } @{$diff_item->{diff_item_details}};
                if ( @field_names ) {
                    $detail->{field_name} = $field_names[0];
                    push(@{$details_by_direction{$detail->{merge_status}}}, $detail);
                }
            }
        }

        for my $direction (keys %details_by_direction) {
            $diff_item_criteria->{diff_item_details} = $details_by_direction{$direction};
            $self->apply_diff_item($session, $local_call_flags, $placeholder, $diff_item_criteria,
                                   F5::ASMConfig::DiffFlags->new({ diff_direction => $direction }), );
        }

        $rv = $entity_class->new(dbh => $self->{dbh},
                                 get_criteria => $diff_item_criteria);
    }

    if ( $store_diff_item->{merge_status} ) {
        delete $diff_item_criteria->{diff_item_details};
        $self->apply_diff_item($session, $local_call_flags, $placeholder, $diff_item_criteria,
                               F5::ASMConfig::DiffFlags->new({ diff_direction => $store_diff_item->{merge_status} }),
                              );
        $rv = $diff_item;
    }

    return {return_value => $rv};
}

#----------------------------------------------------------------------

1;
