package F5::ASMConfig::Handler::EmAsmPolicy;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::Entity::EmAsmPolicy;
use XML::Twig;
use File::Temp qw/tempfile/;
use POSIX;
use File::Copy;
use File::Path;
use File::Basename;
use F5::ASMConfig::Client;
use F5::DeviceGroup;
use F5::ASM::COAPI;

our @handled_functions = (
                           {
                             function_name => 'fetch_em_asm_policy',
                           },
                           {
                             function_name => 'export_em_asm_policy',
                           },
                           {
                             function_name => 'import_em_asm_policy',
                           },
                           {
                             function_name => 'deploy_em_asm_policy',
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::EmAsmPolicy';
my $policy_dir = "/ts/var/em/policies";

#----------------------------------------------------------------------

sub fetch_em_asm_policy {
    my($self, $session, $call_flags, $place_holder, $device, $policy) = @_;

    $device = $self->search_device($session, $call_flags,$device)->{element_list}->[0]->{device};
    my $client_to_remote = F5::ASMConfig::Client->new( username => 'my_test', host => $device->{mng_ip_address}, is_ssl => 1);
    $policy = $client_to_remote->search_policy($session, $call_flags,$policy)->{element_list}->[0]->{policy};

    if (!$policy->{flg_encoding_selected}) {
        my $error_message = "Security Policy \"$policy->{policy_name}\" was not fetched from device $device->{device_name}. Policy language was not defined.";
        $self->write_to_syslog(msg => $error_message);
        $self->handle_error(action => 'get',
                            error_code   => F5::ASMConfig::ErrorCode::INVALID_POLICY(),
                            error_thrown => $error_message,
                            call_flags => $call_flags,
                            thrift_obj => $policy,
                           ); 
    }

    my $remote_temp_filename = $client_to_remote->get_temp_filename($session, $call_flags);
    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new({filename => $remote_temp_filename});
    my $export_res = $client_to_remote->export_policy($session, $call_flags, $policy, $import_export_properties);

    my $CHUNK_SIZE = 2**19; # Half a MB
    my $total_chunks = ceil($export_res->{file_size} / $CHUNK_SIZE);
    my ($fh, $tmp_filename) = tempfile();
    close $fh;

    for my $chunk_index (1..$total_chunks) {
        my $chunk = $client_to_remote->retrieve_file_chunk($session, $call_flags, $export_res->{filename}, $CHUNK_SIZE, $chunk_index);
        open (my $out, ">>", $tmp_filename) or die "cannot append to policy file";
        binmode $out;
        print $out $chunk;
        close $out;
    }

    my $dbh = $self->{dbh};
    $call_flags->{add_on_updating_nonexistent_record} = 1;

    $device = $self->search_device($session, $call_flags, $device)->{element_list}->[0]->{device};

    create_policy_dir();

    my $new_policy_filename = $policy_dir . "/" . $device->{mng_ip_address} . "-" . $policy->{policy_name} . ".plc";
    move($tmp_filename, $new_policy_filename);
    unlink($tmp_filename);

    my $em_asm_policy = F5::ASMConfig::EmAsmPolicy->new({
                                                         policy_name => $policy->{policy_name},
                                                         description => $policy->{description},
                                                         device => $device,
                                                         http_class_name => $policy->{account_name},
                                                         flg_transparent_mode => $policy->{flg_transparent_mode},
                                                         policy_status => $policy->{account_name}
                                                                        ? 'Active'
                                                                        : 'Offline',
                                                         export_file_name => $new_policy_filename,
                                                       });

    my $rv = $self->base_update(
                                rpc_params => [$session, $call_flags, undef, $em_asm_policy],
                                entity_class => $entity_class
                               )->{return_value};

    $self->write_to_syslog(msg => "Security Policy \"$policy->{policy_name}\" was fetched from device $device->{device_name}.");

    my $retval = { return_value => $rv, };
    return $retval;
}

#----------------------------------------------------------------------

sub export_em_asm_policy {
    my($self, $session, $call_flags, $place_holder, $em_asm_policy, $import_export_policy) = @_;

    my $search_result = $self->search_em_asm_policy( $session, $call_flags, $em_asm_policy);

    if (! defined $search_result->{element_list}->[0]) {
        $self->handle_error(action => 'get',
                            error_thrown => $F5::ASMConfig::Handler::NONEXISTENT_RECORD_ERROR,
                            call_flags => $call_flags,
                            thrift_obj => $em_asm_policy,
                           );
    }

    my $policy = $search_result->{element_list}->[0]->{em_asm_policy};
    my $file_basename = basename($import_export_policy->{filename});
    my $target_filename = "/var/tmp/$file_basename";

    copy($policy->{export_file_name}, $target_filename);
    chmod (0644,$target_filename);

    return { return_value => 1, };
}

#----------------------------------------------------------------------

sub import_em_asm_policy{
    my($self, $session, $call_flags, $place_holder, $import_export_policy) = @_;

    my $file_basename = basename($import_export_policy->{filename});
# TODO: - naming, overwriting, uniqueness ...

    create_policy_dir();

    my $new_policy_filename = $policy_dir . "/" . $file_basename;

    my $em_asm_policy = F5::ASMConfig::EmAsmPolicy->new({export_file_name => $new_policy_filename,
                                                         policy_status => 'Offline',
                                                         device => F5::ASMConfig::Device->new({device_name => 'EM'}),
                                                        });

    copy($import_export_policy->{filename}, $new_policy_filename);

    my $import_policy_result = $self->import_policy(
        $session, 
        $call_flags, 
        F5::ASMConfig::ImportExportProperties->new({filename => $new_policy_filename, get_policy_attributes_only => 1}));

    $em_asm_policy->{policy_name} = $import_policy_result->{policy}->{policy_name};
    $em_asm_policy->{description} = $import_policy_result->{policy}->{description};
    $em_asm_policy->{policy_id} = $import_policy_result->{policy}->{policy_id};

    if (! $import_export_policy->{get_policy_attributes_only}) {
        $em_asm_policy = $self->update_em_asm_policy($session, 
                                                     F5::ASMConfig::CallFlags->new(add_on_updating_nonexistent_record=>1), 
                                                     $em_asm_policy);
    }

    return { return_value => $em_asm_policy, };
}

#----------------------------------------------------------------------

sub deploy_em_asm_policy{
    my($self, $session, $call_flags, $place_holder, $deploy_properties, $em_asm_policy) = @_;

    $call_flags->{originator} = basename(F5::DeviceGroup::get_local_device_name());

    my %target_devices;
    for my $deploy_target (@{$deploy_properties->{deploy_targets}}) {
        my $device = $self->get_device($session, $call_flags, $deploy_target->{device});
        if (! exists $target_devices{$device->{mng_ip_address}}) {
            $target_devices{$device->{mng_ip_address}} = $device;
        }
        $device = $target_devices{$device->{mng_ip_address}};
        $device->{virtual_server} = $deploy_target->{virtual_server};
    }

    $em_asm_policy = $self->get_em_asm_policy($session, $call_flags, $em_asm_policy);

    my $local_filename = $em_asm_policy->{export_file_name};
    my $CHUNK_SIZE = 2**19; # Half a MB
    my $file_size = -s $local_filename;
    my $total_chunks = ceil($file_size / $CHUNK_SIZE);

    my $multi_client = F5::ASMConfig::Client->new( 
        logger => $self->{logger},
        devices => [ values %target_devices ],
        is_async => 1,
        is_ssl => 1,
        deploy_request_fields => {
                                  deploy_type => F5::ASMConfig::ElementType::EM_ASM_POLICY(),
                                  em_asm_policy => $em_asm_policy,
                                 },
        deploy_targets => $deploy_properties->{deploy_targets},
        async_calls => [
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;

                                          if ($client->{device}->{virtual_server}) {
                                              $client->add_http_class(
                                                  $client->{session}, 
                                                  $call_flags, 
                                                  F5::ASMConfig::HttpClass->new(http_class_name => $em_asm_policy->{policy_name}, flg_is_asm_enabled => 1));
                                          }
                                         },
                         call_name => 'Create and Associate HTTP Class',
                         return_value => 'http_class',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;

                                          if ($client->{device}->{virtual_server}) {
                                              $client->update_virtual_server(
                                                  $client->{session}, 
                                                  F5::ASMConfig::CallFlags->new(append_nested_struct_on_update => 1), 
                                                  F5::ASMConfig::VirtualServer->new(
                                                      virtual_server_name => $client->{device}->{virtual_server}->{virtual_server_name},
                                                      http_classes => [$retvals->{http_class}] ));
                                          }
                                         },
                         call_name => 'Update Virtual Server',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;
                                          $client->get_temp_filename($client->{session}, $call_flags);
                                         },
                         call_name => 'Get Temp Filename',
                         return_value => 'tempfile_name',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;
                                          for my $chunk_index (1..$total_chunks) {
                                              open (my $in, "<", $local_filename);
                                              seek ($in, $CHUNK_SIZE*($chunk_index-1), 0);
                                              my $chunk;
                                              read $in, $chunk, $CHUNK_SIZE;
                                              close $in;
                                              $client->store_file_chunk($client->{session}, $call_flags, $chunk, $retvals->{tempfile_name});
                                          }
                                         },
                         call_name => 'Transfer Policy',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;

                                          $client->search_policy(
                                                                 $client->{session},
                                                                 $call_flags,
                                                                 F5::ASMConfig::Policy->new(policy_name => $em_asm_policy->{policy_name}));
                                         },
                         call_name => 'Search Policy',
                         return_value => 'existing_policy',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;

                                          my $force_policy_id;
                                          if ($retvals->{existing_policy}->{total_record_count}) {
                                              $force_policy_id = $retvals->{existing_policy}->{element_list}->[0]->{policy}->{policy_id};
                                          }
                                          else {
                                              $force_policy_id = $retvals->{http_class}->{policy_id};
                                          }
 
                                          $client->import_policy(
                                                                 $client->{session},
                                                                 $call_flags,
                                                                 F5::ASMConfig::ImportExportProperties->new(filename => $retvals->{tempfile_name},
                                                                                                            policy_name => $em_asm_policy->{policy_name},
                                                                                                            force_policy_name_flg => 1,
                                                                                                            force_policy_id => $force_policy_id,),
                                                                );
                                         },
                         call_name => 'Import Policy',
                         return_value => 'policy_import',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;

                                          my $policy = $retvals->{policy_import}->{policy};
                                          $policy->{description} = $em_asm_policy->{description};

                                          $client->update_policy($client->{session}, $call_flags, $policy);
                                         },
                         call_name => 'Set Policy Description',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;

                                          $client->unlink_temp_file($client->{session}, $call_flags, $retvals->{tempfile_name});
                                          $self->write_to_syslog(msg => "Security Policy \"$em_asm_policy->{policy_name}\" was deployed to device $client->{device}->{device_name}.");                                        

                                          my $existing_policy;
                                          if ($retvals->{existing_policy}->{total_record_count}) {
                                              $existing_policy = $retvals->{existing_policy}->{element_list}->[0]->{policy};
                                          }

                                          if ($retvals->{http_class}->{account_id} || ((defined $existing_policy) && $existing_policy->{is_active_policy})) {
                                              $client->set_active(
                                                                  $client->{session},
                                                                  $call_flags,
                                                                  $retvals->{policy_import}->{policy},
                                                                  F5::ASMConfig::SetActiveFlags->new());
                                          }
                                         },
                         call_name => 'Set Active Policy',
                        },
                    ],
    );

    return {return_value => $multi_client->run_async_calls(dbh => $self->{dbh})};
}


#----------------------------------------------------------------------

sub create_policy_dir{
    if (! -d $policy_dir) {
        mkpath($policy_dir);
        chmod (0755,"/ts/var/em");
        chmod (0755,$policy_dir);
    }
}

1;
