package F5::ASMConfig::Handler::ExportPolicyTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ExportPolicyTask;
use F5::Utils::Rest;
use MIME::Base64;

our @handled_functions = (
                            {
                             function_name => 'add_export_policy_task',
                             criteria_index => 3,
                             store_index    => 4,
                            },
                            {
                              function_name => 'update_export_policy_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            { 
                             function_name => 'execute_export_policy_task',
                            },
                            {
                             function_name => 'delete_export_policy_task',
                            },
                            {
                             function_name => 'delete_many_export_policy_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ExportPolicyTask';

#----------------------------------------------------------------------

sub add_export_policy_task {
  my($self, $session, $call_flags, $place_holder, $export_task) = @_;

  if (!defined($export_task->{is_base64}) && $export_task->{result_inline} 
      && $export_task->{file_format} == F5::ASMConfig::PolicyFileFormat::POLICY_FILE_FORMAT_BINARY()) {
      $export_task->{is_base64} = 1;
  }
  
  if ($export_task->{file_format} == F5::ASMConfig::PolicyFileFormat::POLICY_FILE_FORMAT_BINARY() 
     && $export_task->{result_inline} && !$export_task->{is_base64}) {
      return $self->handle_error(action => "add_export_policy_task",
                                   error_code => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                   entity_class => $entity_class,
                                   thrift_obj => $export_task,
                                   error_thrown => "Inline Binary export must be Base64 encoded",
                                   error_level => "info",
                                  );

  }

  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => [$session, $call_flags, $place_holder, $export_task],
                                                    entity_class => $entity_class,
                                                    task_type => 'export_policy',);
}

#----------------------------------------------------------------------

sub update_export_policy_task {
    my ($self, @rpc_params) = @_;   
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'export_policy',);
}

#----------------------------------------------------------------------

my %FILE_FORMAT_MAP = ( F5::ASMConfig::PolicyFileFormat::POLICY_FILE_FORMAT_BINARY() => 'plc',
                        F5::ASMConfig::PolicyFileFormat::POLICY_FILE_FORMAT_XML() => 'xml',
                        F5::ASMConfig::PolicyFileFormat::POLICY_FILE_FORMAT_JSON() => 'yaml',
                      );

sub export_policy {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};

    my $policy = $task->{target_policy};
    if (!$policy) {
        return $self->handle_error(action => 'execute',
                                   error_thrown => 'Policy not found',
                                   error_code => F5::ASMConfig::ErrorCode::MISSING_POLICY(),
                                   call_flags => $call_flags,
                                   entity_class => $entity_class,
                                   thrift_obj => $task,
                                  );
    }

    my $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => ($task->{filename} || $task->{rest_uuid}) );
    my $file_format = $FILE_FORMAT_MAP{$task->{file_format}};

    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(filename => $filename ,
                                                                              minimal_export => $task->{minimal},
                                                                              file_format => $file_format,
                                                                              policy_template => $task->{policy_template},
                                                                              export_suggestions => $task->{export_suggestions},
                                                                              include_scanner_tables => $task->{include_vulnerability_assessment_configuration_and_data});

    $self->begin_work($self->{session}) if $task->{minimal};
    my $rv = $self->export_policy($session, $call_flags, $policy, $import_export_properties);

    if ($rv->{msgs}) {
         my ($msg) = join ("\n", map { $_->{msg} } @{$rv->{msgs}});
         $task->{message} = $msg;
    }

    if (!defined $rv->{file_size}) {
        $self->rollback($self->{session}) if $task->{minimal};
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
    }
    else {
        $self->commit($self->{session}) if $task->{minimal};
        if ($task->{result_inline} || $task->{is_base64}) {
            my $file_contents;
            open (my $FH, '<', $filename);
            {
                local $/;
                $file_contents = <$FH>;
            }
            close ($FH);
            if ($task->{is_base64}) {
                $file_contents = MIME::Base64::encode_base64($file_contents);
                $rv->{file_size} = length($file_contents);
            }
            if ($task->{result_inline}) {
                $task->{exported_policy} = $file_contents;
            }
            if ($task->{filename}) {
                open (my $FH, '>', $filename);
                print $FH $file_contents;
                close ($FH);
            }
        }
        $task->{file_size} = $rv->{file_size};
        if (!$task->{filename}) {
            unlink($filename);
        }
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
    }
}


#----------------------------------------------------------------------

sub execute_export_policy_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'export_policy',
                                                          execute_func => sub { export_policy(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_export_policy_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_export_policy_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
