package F5::ASMConfig::Handler::ExportRequestsTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ExportRequestsTask;
use F5::PopulateTemplate;
use F5::Utils::Rest;
use File::Slurp;
use MIME::Base64;

our @handled_functions = (
                            {
                             function_name => 'add_export_requests_task',
                             criteria_index => 3,
                             store_index    => 4,
                            },
                            {
                              function_name => 'update_export_requests_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            {
                             function_name => 'execute_export_requests_task',
                            },
                            {
                             function_name => 'delete_export_requests_task',
                            },
                            {
                             function_name => 'delete_many_export_requests_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ExportRequestsTask';

#----------------------------------------------------------------------

sub add_export_requests_task {
  my($self, @rpc_params) = @_;
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => \@rpc_params,
                                                    entity_class => $entity_class,
                                                    task_type => 'export_requests',);
}

#----------------------------------------------------------------------

sub update_export_requests_task {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'export_requests',);
}

#----------------------------------------------------------------------

sub export_requests {
    my (%params) = @_;
    my $self       = $params{self};
    my $task       = $params{task};
    my $logger     = $self->{logger};

    my $filename;
    if ($task->{filename}) {
        $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => ($task->{filename} || $task->{rest_uuid}) );
    }

    my $populate_template = F5::PopulateTemplate->new(%params,
                                                      dbh => $self->{dbh},
                                                      logger => $logger,
                                                      asm_config_server => $self,
                                                     );

    my $export_res;
    eval {
        $export_res = $populate_template->populate_template(rest_path => '/mgmt/tm/asm/events/requests',
                                                            filter => $task->{filter},
                                                            order_by => $task->{order_by},
                                                            username => $task->{username},
                                                           );
    };

    if ($@) {
        $logger->error("Failed to create Request Export file ($@)");
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
        return;
    }

    $task->{file_size} = length($export_res);
    if ($task->{is_base64}) {
        $export_res = MIME::Base64::encode_base64($export_res);
    }
    if ($task->{result_inline}) {
        $task->{exported_requests} = $export_res;
    }
    else {
        File::Slurp::write_file($filename, $export_res);
    }

    $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
}

#----------------------------------------------------------------------

sub execute_export_requests_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'export_requests',
                                                          execute_func => sub { export_requests(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_export_requests_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_export_requests_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
