package F5::ASMConfig::Handler::ExportSuggestionsTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ExportSuggestionsTask;
use F5::PopulateTemplate;
use JSON;
use F5::Utils::Rest;
use File::Slurp;

our @handled_functions = (
                            {
                             function_name => 'add_export_suggestions_task',
                             criteria_index => 3,
                             store_index    => 4,
                            },
                            {
                              function_name => 'update_export_suggestions_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            {
                             function_name => 'execute_export_suggestions_task',
                            },
                            {
                             function_name => 'delete_export_suggestions_task',
                            },
                            {
                             function_name => 'delete_many_export_suggestions_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ExportSuggestionsTask';

#----------------------------------------------------------------------

sub add_export_suggestions_task {
  my($self, @rpc_params) = @_;
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => \@rpc_params,
                                                    entity_class => $entity_class,
                                                    task_type => 'export_suggestions',);
}

#----------------------------------------------------------------------

sub update_export_suggestions_task {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'export_suggestions',);
}

#----------------------------------------------------------------------

sub export_suggestions {
    my (%params) = @_;
    my $self       = $params{self};
    my $task       = $params{task};
    my $logger     = $self->{logger};

    my $filename;
    if ($task->{filename}) {
        $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => ($task->{filename} || $task->{rest_uuid}) );
    }

    my $policy = $task->{policy};

    eval {
        my $filters;
        if ($task->{item_filter}) {
            $filters = F5::ASMConfig::Entity::Suggestion->convert_rest_filter_string(%params,
                                                                                     dbh => $self->{dbh},
                                                                                     filter_string => $task->{filter});
        }

        my $rv = F5::ASMConfig::Entity::Suggestion->export_declarative(dbh => $self->{dbh},
                                                                       policy => $policy,
                                                                       search_filters => $filters);
        my $items = $rv->{items};
        my $suggestions = encode_json($items);
        if ($rv->{warnings} && @{$rv->{warnings}}) {
            $task->{message} = join('\n', @{$rv->{warnings}});
        }

        if ($task->{result_inline}) {
            $task->{exported_suggestions} = $suggestions;
        }
        else {
            $task->{file_size} = length($suggestions);
            File::Slurp::write_file($filename, $suggestions);
        }
    };
    if ($@) {
        $task->{message} = ref $@ ? $@->{error_message} : $@;
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
    }
    else {
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
    }
}

#----------------------------------------------------------------------

sub execute_export_suggestions_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'export_suggestions',
                                                          execute_func => sub { export_suggestions(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_export_suggestions_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_export_suggestions_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
