package F5::ASMConfig::Handler::GraphqlProfile;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::GraphqlProfile;

our @handled_functions = (
                           {
                             function_name => 'add_graphql_profile',
                             relayable => 1,
                             store_index => 3,
                           },
                           {
                             function_name => 'update_graphql_profile',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                           },
                           {
                             function_name => 'delete_graphql_profile',
                             relayable => 1,
                           },
                           {
                             function_name => 'handle_graphql_idl_files',
                           },
                           {
                             function_name => 'update_many_graphql_profiles',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::GraphqlProfile';
#---------------------------------------------------------

sub add_graphql_profile {
  my($self, $session, $call_flags, $policy,$graphql_profile) = @_;

  my $run_idl = defined($graphql_profile->{idl_files}) && @{$graphql_profile->{idl_files}};

  my $rv = $self->base_add(entity_class => $entity_class,
                           rpc_params => [ $session, $call_flags,$policy,$graphql_profile ]);
  my $profile = $rv->{return_value};
  if ($run_idl) {
       _handle_idl_files( self => $self,
                          profile => $profile,
                          call_flags => $call_flags,
                          policy => $policy,
                          action => 'add_graphql_profile',
                        );
  }
  return $rv;
}

#---------------------------------------------------------

sub update_graphql_profile {
  my($self, $session, $call_flags, $policy,$graphql_profile_criteria, $store_graphql_profile) = @_;


  my $run_idl;
  if (defined($store_graphql_profile->{idl_files})) {
      my $profile =  $entity_class->new(dbh => $self->{dbh},
                                        get_criteria => $graphql_profile_criteria,
                                        call_flags => $call_flags);
      if ($profile) {
          $run_idl = $profile->has_value_changed(existing_obj => $profile,
                                                 store_obj => $store_graphql_profile,
                                                 check_fields => ['idl_files']);
      }
      else {
          $run_idl = 1;
      }
  }
  my $rv = $self->base_update(entity_class => $entity_class,
                              rpc_params => [ $session, $call_flags,$policy,$graphql_profile_criteria, $store_graphql_profile ]);

  my $profile = $rv->{return_value};
  if ($run_idl) {
      _handle_idl_files( self => $self,
                                profile => $profile,
                                call_flags => $call_flags,
                                policy => $policy,
                                action => 'update_graphql_profile',
                               );
  }
  return $rv;
}


#---------------------------------------------------------

sub delete_graphql_profile {
 my($self, $session, $call_flags, $policy,$graphql_profile) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  my $rv = $self->base_delete(entity_class => $entity_class,
                              rpc_params => [ $session, $call_flags,$policy,$graphql_profile ],
                            );

  F5::GraphqlProfile::clean_unreferenced_profile_files(dbh => $dbh,
                                                   logger => $logger);
  return $rv;
}
#---------------------------------------------------------

sub handle_graphql_idl_files {
    my($self, $session, $call_flags, $policy,$graphql_profile) = @_;
    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $rv =  _handle_idl_files( self => $self,
                                 profile => $graphql_profile,
                                 policy => $policy,
                                 action => '',
                               );

    my $res = F5::ASMConfig::GraphqlIdlFilesRes->new($rv);
    my $retval = { return_value => $res };
    return $retval;
}

#---------------------------------------------------------

sub _handle_idl_files {
  my (%params) = @_;

  my $self = $params{self};
  my $profile = $params{profile};
  my $action = $params{action};
  my $call_flags = $params{call_flags};
  my $policy = $params{policy};

  if ($call_flags->{rest_call} && ! $call_flags->{async_rest_call}) {
      return $self->handle_error(action => $action,
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                 thrift_obj   => $profile,
                                 error_thrown => 'Can not set idlFiles directly. Please use the associate-idl-files-to-graphql-profile Task.',
                                 error_level  => 'info',
                                );
  }


  my $graphql_profile = F5::ASMConfig::Entity::GraphqlProfile->new(dbh => $self->{dbh}, object => $profile, policy => $policy);
  my $res = F5::GraphqlProfile::handle_idl_files(dbh => $self->{dbh},
                                                 logger => $self->{logger},
                                                 asm_config_handler => $self,
                                                 graphql_profile => $graphql_profile,
                                                 policy => $policy,
                                                );
  if (!$res->{completed_successfully}) {
      my $error_thrown = "Idl failed.";
      my $error_code = F5::ASMConfig::ErrorCode::INVALID_GRAPHQL_PROFILE();


      if ( defined $res->{messages} ) {
          $error_thrown .= "  " . (ref($res->{messages}) ? join (", ", @{$res->{messages}}) : $res->{messages});
      }

      return $self->handle_error(action => $action,
                                 entity_class => $entity_class,
                                 error_code   => $error_code,
                                 thrift_obj   => $profile,
                                 error_thrown => $error_thrown,
                                 error_level  => 'info',
                                );
  }
  return $res;
}
#---------------------------------------------------------

sub update_many_graphql_profiles {
  my($self, $session, $call_flags, $policy, $graphql_profile_criteria, $store_graphql_profile) = @_;

  # cannot update idl files in many graphql profiles - ever.
  # instead, clients need to update_graphql_profile.
  if (defined($store_graphql_profile->{idl_files})) {
      return $self->handle_error(action => 'update_many_graphql_profiles',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                 error_thrown => 'Can not set idlFiles directly. Please use the associate-graphql-idl-files-to-graphql-profile Task.',
                                 error_level  => 'info',
                                );
  }

  return $self->base_update_many(entity_class => $entity_class,
                                 rpc_params => [ $session, $call_flags, $policy, $graphql_profile_criteria, $store_graphql_profile ],);
}
#---------------------------------------------------------

1;
