package F5::ASMConfig::Handler::GrpcProfile;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::GRPCProfile;

our @handled_functions = (
                           {
                             function_name => 'add_grpc_profile',
                             relayable => 1,
                             store_index => 3,
                           },
                           {
                             function_name => 'update_grpc_profile',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                           },
                           {
                             function_name => 'delete_grpc_profile',
                             relayable => 1,
                           },
                           {
                             function_name => 'handle_grpc_idl_files',
                           },
                           {
                             function_name => 'update_many_grpc_profiles',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::GrpcProfile';
#---------------------------------------------------------

sub add_grpc_profile {
  my($self, $session, $call_flags, $policy,$grpc_profile) = @_;

  my $rv = $self->base_add(entity_class => $entity_class,
                           rpc_params => [ $session, $call_flags,$policy,$grpc_profile ]);

  my $saved_events = delete $self->{events};
  my $local_events = delete $rv->{events};

  my $profile = $rv->{return_value};
  _handle_idl_files( self => $self,
                     profile => $profile,
                     call_flags => $call_flags,
                     policy => $policy,
                     action => 'add_grpc_profile',
                   );
  if ($self->{events}) {
      push (@$local_events, @{delete($self->{events})});
  }
  $rv->{events} = $local_events;
  $self->{events} = $saved_events;

  return $rv;
}

#---------------------------------------------------------

sub update_grpc_profile {
  my($self, $session, $call_flags, $policy,$grpc_profile_criteria, $store_grpc_profile) = @_;


  my $run_validation;
  my $profile =  $entity_class->new(dbh => $self->{dbh},
                                    get_criteria => $grpc_profile_criteria,
                                    call_flags => $call_flags);

  if ($profile) {
      if (defined($store_grpc_profile->{idl_files})) {
          $run_validation = $profile->has_value_changed(existing_obj => $profile,
                                                        store_obj => $store_grpc_profile,
                                                        check_fields => ['idl_files']);
      }
  }
  else {
      $run_validation = 1;
  }

  my $rv = $self->base_update(entity_class => $entity_class,
                              rpc_params => [ $session, $call_flags,$policy,$grpc_profile_criteria, $store_grpc_profile ]);

  $profile = $rv->{return_value};
  if ($run_validation) {
      my $saved_events = delete $self->{events};
      my $local_events = delete $rv->{events};
      _handle_idl_files( self => $self,
                         profile => $profile,
                         call_flags => $call_flags,
                         policy => $policy,
                         action => 'update_grpc_profile',
                       );
      if ($self->{events}) {
          push (@$local_events, @{delete($self->{events})});
      }
      $rv->{events} = $local_events;
      $self->{events} = $saved_events;
  }
  return $rv;
}


#---------------------------------------------------------

sub delete_grpc_profile {
 my($self, $session, $call_flags, $policy,$grpc_profile) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  my $rv = $self->base_delete(entity_class => $entity_class,
                              rpc_params => [ $session, $call_flags,$policy,$grpc_profile ],
                            );

  F5::GRPCProfile::clean_unreferenced_profile_files(dbh => $dbh,
                                                    logger => $logger);
  return $rv;
}
#---------------------------------------------------------

sub handle_grpc_idl_files {
    my($self, $session, $call_flags, $policy,$grpc_profile) = @_;
    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $rv =  _handle_idl_files( self => $self,
                                 profile => $grpc_profile,
                                 policy => $policy,
                                 action => '',);

    my $res = F5::ASMConfig::GrpcIdlFilesRes->new($rv);
    my $retval = { return_value => $res };
    return $retval;
}

#---------------------------------------------------------

sub _handle_idl_files {
  my (%params) = @_;

  my $self = $params{self};
  my $profile = $params{profile};
  my $action = $params{action};
  my $call_flags = $params{call_flags};
  my $policy = $params{policy};


  my $grpc_profile = F5::ASMConfig::Entity::GrpcProfile->new(dbh => $self->{dbh}, object => $profile, policy => $policy);
  my $res = F5::GRPCProfile::handle_idl_files(dbh => $self->{dbh},
                                              logger => $self->{logger},
                                              asm_config_handler => $self,
                                              grpc_profile => $grpc_profile,
                                              policy => $policy,
                                              policy_name_crc => $policy->{policy_name_crc},
                                              profile_id => $profile->{profile_id},
                                              get_urls => 1,
                                             );
  if (!$res->{completed_successfully} || !$res->{urls} || ! @{$res->{urls}}) {
      my $error_thrown = "Validation failed.";
      my $error_code = F5::ASMConfig::ErrorCode::INVALID_GRPC_PROFILE();

      if ($res->{completed_successfully} && (! $res->{url} || ! @{$res->{urls}}) ) {
          $error_thrown = "gRPC Profile does not have any RPC services defined.";
      }
      elsif ( defined $res->{messages} ) {
          $error_thrown .= "  " . (ref($res->{messages}) ? join (",", @{$res->{messages}}) : $res->{messages});
      }

      return $self->handle_error(action => $action,
                                 entity_class => $entity_class,
                                 error_code   => $error_code,
                                 thrift_obj   => $profile,
                                 error_thrown => $error_thrown,
                                 error_level  => 'info',
                                );
  }
  elsif ($grpc_profile->{flg_associate_urls}) {
      my $created_by = "gRPC Profile $profile->{profile_name}";
      if ($action eq 'update_grpc_profile') {
          my $delete_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags, 
                                                                search_filters => [ F5::ASMConfig::SearchFilter->new(field_name => 'url_name',
                                                                                                                     field_value_list => $res->{services},
                                                                                                                     field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN()),
                                                                                    F5::ASMConfig::SearchFilter->new(field_name => 'created_by',
                                                                                                                     field_value => $created_by,
                                                                                                                     field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ())
                                                                                  ]);
          $self->delete_many_urls($self->{session}, $delete_call_flags, $policy, F5::ASMConfig::URL->new());
      }
      for my $service_name (@{$res->{services}}) {
          my @protocols = (0);
          if (!$policy->{flg_is_protocol_free}) {
              push @protocols, 1;
          }
          my $url_name = "$service_name/*";
          for my $protocol (@protocols) {
              my $url = F5::ASMConfig::URL->new({app_id => $protocol, url_name => $url_name, method_name => '*', is_wildcard => 1});
              my $url_entity  = F5::ASMConfig::Entity::URL->new(dbh => $self->{dbh}, policy => $policy, get_criteria => $url);
              if (!$url_entity) {
                  $url->{created_by} = $created_by;
                  $url->{url_content_profiles} = [ F5::ASMConfig::UrlContentProfile->new(header_name => 'Content-Type', header_value => 'application/grpc*', header_order => 1,
                                                                                         enforcement_type => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRPC(), content_profile => $grpc_profile),
                                                   F5::ASMConfig::UrlContentProfile->new(header_name => '*', header_value => '*', header_order => 0, 
                                                                                         enforcement_type => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_DISALLOW()) ];
                  $self->add_url($self->{session}, $call_flags, $policy, $url);
                  next;
              }

              my $url_content_profile = F5::ASMConfig::UrlContentProfile->new({ url => $url, enforcement_type => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRPC()});
              my $url_content_profile_entity = F5::ASMConfig::Entity::UrlContentProfile->new(dbh => $self->{dbh}, policy => $policy, 
                                                                                             get_criteria => $url_content_profile );
              if (!$url_content_profile_entity) {
                  $url_content_profile->{header_name} = 'Content-Type';
                  $url_content_profile->{header_value} = '*grpc*';
                  $url_content_profile->{header_order} = 1;
                  $url_content_profile->{content_profile} = $grpc_profile;
                  $self->add_url_content_profile($self->{session}, $call_flags, $policy, $url_content_profile);
              }
              elsif ($url_content_profile_entity->{content_profile}->{profile_name} ne $grpc_profile->{profile_name}) {
                  return $self->handle_error(action => $action,
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::INVALID_GRPC_PROFILE(),
                                 thrift_obj   => $profile,
                                 error_thrown => "Conflicting gRPC profiles '$url_content_profile_entity->{content_profile}->{profile_name}' and '$grpc_profile->{profile_name}' for URL $url_name",,
                                 error_level  => 'info',
                                );
              }
          }
      }
  }

  return $res;
}
#---------------------------------------------------------

sub update_many_grpc_profiles {
  my($self, $session, $call_flags, $policy, $grpc_profile_criteria, $store_grpc_profile) = @_;

  # cannot update idl files in many grpc profiles - ever.
  # instead, clients need to update_grpc_profile.
  if (defined($store_grpc_profile->{idl_files})) {
      return $self->handle_error(action => 'update_many_grpc_profiles',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                 error_thrown => 'Can not set idlFiles via bulk update.',
                                 error_level  => 'info',
                                );
  }

  return $self->base_update_many(entity_class => $entity_class,
                                 rpc_params => [ $session, $call_flags, $policy, $grpc_profile_criteria, $store_grpc_profile ],);
}
#---------------------------------------------------------

1;
