package F5::ASMConfig::Handler::ImportCertificateTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ImportCertificateTask;
use F5::PolicyTemplateDefaults;
use F5::Utils::Rest;
use File::Copy;

our @handled_functions = (
                            {
                             function_name => 'add_import_certificate_task',
                             criteria_index => 3,
                             store_index    => 4,
                             write_access   => 1,
                            },
                            {
                              function_name => 'update_import_certificate_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            {
                             function_name => 'execute_import_certificate_task',
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                            },
                            {
                             function_name => 'delete_import_certificate_task',
                            },
                            {
                             function_name => 'delete_many_import_certificate_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ImportCertificateTask';

#----------------------------------------------------------------------

sub add_import_certificate_task {
  my($self, @rpc_params) = @_;
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => \@rpc_params,
                                                    entity_class => $entity_class,
                                                    task_type => 'import_certificate',);
}

#----------------------------------------------------------------------

sub update_import_certificate_task {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'import_certificate',);
}

#----------------------------------------------------------------------

sub import_certificate {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};

    my $rv;
    my $delete_filename;

    my $filename = $task->{filename};
    my $file_contents = $task->{file_contents};
    if ($filename) {
        $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename} );
    }

    if ($file_contents) {
        if (!$filename) {
            $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{rest_uuid} );
            $delete_filename = 1;
        }
        open (my $FH, '>', $filename) || $self->handle_error(action => 'execute',
                                                             entity_class => $entity_class,
                                                             thrift_obj  => $task,
                                                             error_code  => F5::ASMConfig::ErrorCode::PERMISSION_DENIED(),
                                                             msg         => "Could not open $filename for writing.",
                                                             error_level  => "info", );
        print $FH $file_contents;
        close ($FH);
    }
    else {
        open (my $FH, '<', $filename);
        while (my $line = <$FH>) {
            $file_contents .= $line;
        }
    }

    my $flg_skip_validation = $task->{flg_skip_validation};
    my $existing_certificate = $task->{target_certificate};
    eval {
        my $result = $self->verify_certificate($session, $call_flags, $filename);
        my $throw_error = 0;
        my $error_message;
        my $expiration_date;
        if ($result->{completed_successfully}) {
            if($result->{is_expired} == 0 || ($result->{is_expired} == 1 && $flg_skip_validation)) {
                if ($result->{local_expiration_date}) {
                    $expiration_date = $result->{local_expiration_date};
                }
            }
            else {
                $throw_error = 1;
                $error_message = "Certificate expired. Please check 'Save Expired/Untrusted Certificate' to save it";
            }
            if($result->{is_untrusted} && !$flg_skip_validation) {
                $throw_error = 1;
                $error_message = "Certificate is untrusted. Please check 'Save Expired/Untrusted Certificate' to save it";
            }
        }
        else {
            $throw_error = 1;
            $error_message = $result->{error_msg};
        }

        if ($throw_error) {
            $self->handle_error(action => 'verify',
                                entity_class => $entity_class,
                                thrift_obj  => $task,
                                error_code  => F5::ASMConfig::ErrorCode::INVALID_CERTIFICATE(),
                                display_message => $error_message,
                                error_thrown => $error_message,
                                call_flags => $call_flags,
                                error_level  => "info", );
        }

        $expiration_date = F5::DateTimeUtils::convert_epoch_to_datetime($expiration_date);
        my $certificate = F5::ASMConfig::Entity::Certificate->new( object => {key_file => $file_contents,
                                                                              flg_skip_validation => $flg_skip_validation,
                                                                              expiration_date => $expiration_date, });

        $self->begin_work($self->{session});
        if ($existing_certificate) {
            if ( $task->{cert_type} && $task->{cert_type} ne $existing_certificate->{cert_type} ) {
                my $error_message = "Certificate Type cannot be modified.";
                $self->handle_error(action => 'import',
                                    entity_class => $entity_class,
                                    thrift_obj  => $task,
                                    error_code  => F5::ASMConfig::ErrorCode::ILLEGAL_OPERATION(),
                                    display_message => $error_message,
                                    error_thrown => $error_message,
                                    call_flags => $call_flags,
                                    error_level  => "info", );
            }
            $rv =  $self->update_certificate($session, undef, $existing_certificate, $certificate);
        }
        else {
            $certificate->{key_name} = $task->{key_name};
            $certificate->{cert_type} = $task->{cert_type};
            $rv = $self->add_certificate($session, undef, $certificate);
        }
        $self->commit($self->{session});

        $task->{result_certificate} = $rv;
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
    };
    my $err = $@;

    if ($delete_filename) {
        unlink($filename);
    }
    if ($err) {
        die $err;
    }
}

#----------------------------------------------------------------------

sub execute_import_certificate_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'import_certificate',
                                                          execute_func => sub { import_certificate(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_import_certificate_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_import_certificate_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
