package F5::ASMConfig::Handler::ImportDataProtectionTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ImportDataProtectionTask;
use F5::PolicyTemplateDefaults;
use F5::Utils::Rest;
use File::Copy;

our @handled_functions = (
                            {
                             function_name => 'add_import_data_protection_task',
                             criteria_index => 3,
                             store_index    => 4,
                            },
                            {
                              function_name => 'update_import_data_protection_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            {
                             function_name => 'execute_import_data_protection_task',
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                            },
                            {
                             function_name => 'delete_import_data_protection_task',
                            },
                            {
                             function_name => 'delete_many_import_data_protection_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ImportDataProtectionTask';

#----------------------------------------------------------------------

sub add_import_data_protection_task {
  my($self, @rpc_params) = @_;
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => \@rpc_params,
                                                    entity_class => $entity_class,
                                                    task_type => 'import_data_protection',);
}

#----------------------------------------------------------------------

sub update_import_data_protection_task {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'import_data_protection',);
}

#----------------------------------------------------------------------

sub import_data_protection {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};

    my $delete_filename;

    my $import_text = $task->{import_text};

    my $filename = $task->{filename};

    if ( !$import_text ) {
        if ($filename) {
            $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename} );
        }
        if ($task->{file_contents}) {
            if (!$filename) {
                $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{rest_uuid} );
                $delete_filename = 1;
            }
            open (my $FH, '>', $filename) || $self->handle_error(action => 'execute',
                                                                 entity_class => $entity_class,
                                                                 thrift_obj  => $task,
                                                                 error_code  => F5::ASMConfig::ErrorCode::PERMISSION_DENIED(),
                                                                 error_thrown => "Could not open $filename for writing.",
                                                                 error_level  => "info", );
            my $contents = $task->{file_contents};
            print $FH $contents;
            close ($FH);
        }

        if ($task->{filename} && ! -e $filename) {
            $self->handle_error(action => 'execute',
                                entity_class => $entity_class,
                                thrift_obj  => $task,
                                error_code  => F5::ASMConfig::ErrorCode::FILE_NOT_FOUND(),
                                display_message => "Could not find file: $task->{filename}",
                                error_thrown => "Could not find file: $filename",
                                error_level  => "info", );
        }
    }

    # Import Task now accepts imported Data Protection Context, or alternatively accepts Data Protection configuration used to regenerate a new secure key
    my $rv;
    my $data_protection_import_export = F5::ASMConfig::DataProtectionImportExport->new(filename => $filename,
                                                                                       import_text => $import_text,
                                                                                       data_protection_cookie_configurations => $task->{data_protection_cookie_configurations},
                                                                                       grace_signing_interval => $task->{grace_signing_interval},
                                                                                       grace_accepting_interval => $task->{grace_accepting_interval},
                                                                                      );
    $self->begin_work($self->{session});
    $rv = $self->import_data_protection($session, $call_flags, $data_protection_import_export);
    $self->commit($self->{session});

    $task->{result_data_protection} = $rv;
    $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();

    if ($delete_filename) {
        unlink($filename);
    }
}


#----------------------------------------------------------------------

sub execute_import_data_protection_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'import_data_protection',
                                                          execute_func => sub { import_data_protection(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_import_data_protection_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_import_data_protection_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
