package F5::ASMConfig::Handler::ImportOpenApiActions;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;

use F5::ASMConfig::ASMConfig;
use F5::DeviceGroup;
use F5::Mcp::CfgSrc;
use F5::License;

my $DEFAULT_POLICY_TEMPLATE_NAME = 'POLICY_TEMPLATE_API_SECURITY';


our @handled_functions = (
                           {
                             function_name => 'import_open_api_update_policy',
                             filename_arguments => ['3.filename'],
                             relayable     => 1,
                             writes_to_mcp => 0, #will be overridden if mcp will be affected (activate/deactivate/encoding change)
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'import_open_api_create_policy',
                             filename_arguments => ['3.filename'],
                             relayable => 1,
                             writes_to_mcp => 1,
                             skip_authz => 1,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                          );


#-------------------------------------------------------------------------

sub import_open_api_update_policy {
    my ($self, $session, $call_flags, $target_policy, $import_open_api_actions_fields) = @_;

    if (! F5::License::is_waf_licensed()) {
        my $error_message = "Open API is not licensed.";
        return $self->handle_error(action => "import_open_api_update_policy",
                                   error_code => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                   thrift_obj => $self,
                                   display_message => $error_message,
                                   error_thrown => $error_message,
                                   error_level => "info",);
    }

    my $open_api_profile_name = $import_open_api_actions_fields->{open_api_profile_name};
    if (! $open_api_profile_name) {
        $self->delete_many_urls($session, undef, $target_policy, {url_name => '*'});
        $self->delete_many_websocket_urls($session, undef, $target_policy, {websocket_url_name => '*'});
        $self->delete_many_parameters($session, undef, $target_policy, {param_name => '*'});
    }

    my $encoding = $import_open_api_actions_fields->{encoding} || $target_policy->{encoding} || 'utf-8';
    my $import_open_api = F5::ImportOpenApi->new(api_type => $import_open_api_actions_fields->{api_type},
                                                 username => $import_open_api_actions_fields->{username},
                                                 encoding => $encoding,
                                                 handler => $self,
                                                 session => $session,
                                                 call_flags => $call_flags,
                                                 logger => $self->{logger},
                                                 dbh => $self->{dbh},);

    $import_open_api->get_openapi_content(filename => $import_open_api_actions_fields->{filename},
                                          api_content_format => $import_open_api_actions_fields->{api_content_format},
                                          encoding => $encoding,
                                          is_base64 => $import_open_api_actions_fields->{is_base64},);

    my $policy_description = $import_open_api->{openapi_content}->{title} ||
                             $import_open_api->{openapi_content}->{info}->{title};
    if ($policy_description ne $target_policy->{description}) {
        my $update_policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh}, get_criteria => $target_policy,);
        $update_policy->store_update(store_obj => {description => $policy_description,},);
    }

    $import_open_api->process_openapi_content(policy => $target_policy,
                                              open_api_profile_name => $open_api_profile_name,
                                             );

    if (! $target_policy->{flg_is_protocol_free} && $target_policy->can_update_protocol_free(dbh => $self->{dbh},)) {
        $target_policy->{flg_is_protocol_free} = 1;
        $self->update_policy($session, $call_flags, $target_policy);
    }

    set_file_lastupdate($self,
                        session => $session,
                        call_flags => $call_flags,
                        policy => $target_policy,
                        filename => $import_open_api_actions_fields->{filename});

    my $rv = {policy => $target_policy};
    my $return_value = F5::ASMConfig::ImportPolicyRes->new($rv);
    return {return_value => $return_value};
}

#-------------------------------------------------------------------------

sub import_open_api_create_policy {
    my ($self, $session, $call_flags, $place_holder, $import_open_api_actions_fields) = @_;

    if (! F5::License::is_waf_licensed()) {
        my $error_message = "Open API is not licensed.";
        return $self->handle_error(action => "import_open_api_update_policy",
                                   error_code => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                   thrift_obj => $self,
                                   display_message => $error_message,
                                   error_thrown => $error_message,
                                   error_level => "info",);
    }

    my ($policy_name, $open_api_profile);
    my $open_api_profile_name = $import_open_api_actions_fields->{open_api_profile_name};

    my $force_create_policy = 0;
    if ($open_api_profile_name) {
        $force_create_policy = 1;
        $policy_name = $open_api_profile_name;
        my $open_api_profile_query = F5::ASM::COAPI::Query->new("profile_api_protection",
                                                                {name => $open_api_profile_name});
        if ($open_api_profile_query && @$open_api_profile_query) {
            $open_api_profile = $open_api_profile_query->[0];
            if ($open_api_profile->{config_source} == $F5::Mcp::CfgSrc::MCP_USER_CFG) {
                if ($open_api_profile->{security_policy_created} && $open_api_profile->{security_policy}) {
                    my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh}, get_criteria => {policy_name => $open_api_profile->{security_policy}});
                    if ($policy) {
                        my $return_value = F5::ASMConfig::ImportPolicyRes->new({policy => $policy});
                        return {return_value => $return_value};
                    }
                }
            }
            else {
                my $return_value = F5::ASMConfig::ImportPolicyRes->new({completed_successfully => 0});
                $self->{logger}->debug("ASM Policy cannot be created for default Open Api profile: $open_api_profile_name");
                return {return_value => $return_value};
            }
        }
    }
    else {
        $policy_name = $import_open_api_actions_fields->{policy_name};
    }

    my $filename = $import_open_api_actions_fields->{filename};
    if (! $force_create_policy && ! $filename && ! -e $filename) {
        $self->handle_error(action => 'import',
                            thrift_obj  => $self,
                            error_code  => F5::ASMConfig::ErrorCode::FILE_NOT_FOUND(),
                            msg         => "Cannot open Open Api file.",
                            error_level  => "info", );
    }

    $import_open_api_actions_fields->{policy_template} ||= { icontrol_name => $DEFAULT_POLICY_TEMPLATE_NAME };
    my $policy_template = F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $self->{dbh},
                                                                     get_criteria => $import_open_api_actions_fields->{policy_template});

    my $encoding = $import_open_api_actions_fields->{encoding} || 'utf-8';
    my $import_open_api;
    eval {
        $import_open_api = F5::ImportOpenApi->new(api_type => $import_open_api_actions_fields->{api_type},
                                                     username => $import_open_api_actions_fields->{username},
                                                     encoding => $encoding,
                                                     handler => $self,
                                                     session => $session,
                                                     call_flags => $call_flags,
                                                     logger => $self->{logger},
                                                     dbh => $self->{dbh},);

        $import_open_api->get_openapi_content(filename => $filename,
                                              api_content_format => $import_open_api_actions_fields->{api_content_format},
                                              encoding => $encoding,
                                              is_base64 => $import_open_api_actions_fields->{is_base64},);
    };

    my $process_open_api_content = 1;
    if ($@) {
        if ($force_create_policy) {
            $process_open_api_content = 0;
        }
        else {
            die $@;
        }
    }

    my $policy_description = $import_open_api_actions_fields->{policy_description};
    $policy_description ||= $import_open_api->{openapi_content}->{title} ||
                            $import_open_api->{openapi_content}->{info}->{title};

    if (!$import_open_api_actions_fields->{partition} && $policy_name && ($policy_name =~ m|^/([^/]*)|)) {
        $import_open_api_actions_fields->{partition} = $1;
    }

    my $policy = F5::ASMConfig::Policy->new(policy_name => $policy_name,
                                            policy_template => $policy_template,
                                            encoding => $encoding,
                                            flg_encoding_selected => 1,
                                            flg_is_protocol_free => 1,
                                            is_active_policy => 0,
                                            partition_id => $import_open_api_actions_fields->{partition},
                                            description => $policy_description,);
    my $result_policy = $self->add_policy($session, $call_flags, $policy);

    if ($open_api_profile_name && ! $self->{dbh}->{AutoCommit}) {
        # Since this import was triggered asynchronously by API Profile creation,
        # transactional MCP changes must be committed, even if the OpenAPI file subsequently
        # has errors. This will ensure that the API Profile is at least associated with an
        # active API Protection policy.
        $self->{dbh}->commit();
        $self->{dbh}->begin_work();
    }

    if (!$open_api_profile_name) {
        $self->delete_many_urls($session, undef, $result_policy, {url_name => '*'});
        $self->delete_many_parameters($session, undef, $result_policy, {param_name => '*'});
    }

    if ($process_open_api_content) {
        $import_open_api->process_openapi_content(policy => $result_policy,
                                                  open_api_profile_name => $open_api_profile_name,
                                                 );
        set_file_lastupdate($self,
                            session => $session,
                            call_flags => $call_flags,
                            policy => $result_policy,
                            filename => $filename);

        if ($open_api_profile_name) {
            $self->set_active($session, $call_flags, $result_policy, F5::ASMConfig::SetActiveFlags->new());
        }
    }

    if ($open_api_profile) {
        $open_api_profile->{security_policy} = $result_policy->{policy_name};
        $open_api_profile->{security_policy_created} = 1;
        if ($self->{rpc_entry_depth} == 1 && $self->{device_group}) {
            if (F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                         folder => $result_policy->mcp_folder(),
                                                         logger => $self->{logger},
                                                         create => $self->{is_relay_handler_process})) {
                $call_flags->{do_not_relay} = 1;
            }
            else {
                if ($call_flags->{skip_mcp} && $self->{is_relay_handler_process}) {
                    $call_flags->{skip_mcp} = 0;
                }
            }
        }

        $open_api_profile->transactional_action(call_flags => $call_flags, action => 'save', dbh => $self->{dbh});
    }

    return {return_value => F5::ASMConfig::ImportPolicyRes->new({ policy => $result_policy })};
}

#-------------------------------------------------------------------------

sub set_file_lastupdate {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh},
    my $policy = $params{policy};
    my $filename = $params{filename};
    my $session = $params{session};
    my $call_flags = $params{call_flags};

    if ($filename =~ /\~\d+?_(.*)/) {
        $filename = $1;
    }
    elsif ($filename =~ /\/\w+?\~(.*)/) {
        $filename = $1;
    }
    else {
        $filename =~ s|^.*\/||;
    }

    $self->delete_many_policy_open_api_files($session, $call_flags, $policy);

    my $api_file = F5::ASMConfig::PolicyOpenApiFile->new(policy => $policy,
                                                         filename => $filename,
                                                         upload_datetime => time(),);

    $self->add_policy_open_api_file($session, $call_flags, $policy, $api_file);
}

#-------------------------------------------------------------------------


1;

