package F5::ASMConfig::Handler::ImportOpenApiTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ImportOpenApiTask;
use F5::ASMConfig::Entity::PolicyTemplate;
use F5::Utils::Rest;
use MIME::Base64;

use F5::ImportOpenApi;

our @handled_functions = (
                            {
                             function_name => 'add_import_open_api_task',
                             criteria_index => 3,
                             store_index    => 4,
                             write_access   => 1,
                            },
                            {
                              function_name => 'update_import_open_api_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            {
                             function_name => 'execute_import_open_api_task',
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                            },
                            {
                             function_name => 'delete_import_open_api_task',
                            },
                            {
                             function_name => 'delete_many_import_open_api_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ImportOpenApiTask';

#----------------------------------------------------------------------

sub add_import_open_api_task {
  my($self, @rpc_params) = @_;
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => \@rpc_params,
                                                    entity_class => $entity_class,
                                                    task_type => 'import_open_api',);
}

#----------------------------------------------------------------------

sub update_import_open_api_task {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'import_open_api',);
}

#----------------------------------------------------------------------

sub import_open_api {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};

    my $filename = $task->{filename};
    my $contents = $task->{file_contents};
    if ($filename) {
        $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename});
    }
    if ($task->{file_contents}) {
        if (!$filename) {
            $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{rest_uuid});
        }
        open (my $FH, '>', $filename) || $self->handle_error(action => 'execute',
                                                             entity_class => $entity_class,
                                                             thrift_obj  => $task,
                                                             error_code  => F5::ASMConfig::ErrorCode::PERMISSION_DENIED(),
                                                             display_message => "Could not open $filename for writing.",
                                                             error_level  => "info", );
       if ($task->{is_base64}) {
           $task->{file_contents} = MIME::Base64::decode_base64($task->{file_contents});
           $task->{is_base64} = 0;
       }
       print $FH $task->{file_contents};
       close ($FH);
    }

    if (! $filename && ! -e $filename) {
        $self->handle_error(action => 'execute',
                            entity_class => $entity_class,
                            thrift_obj  => $task,
                            error_code  => F5::ASMConfig::ErrorCode::FILE_NOT_FOUND(),
                            display_message => "Cannot open Open API file.",
                            error_level  => "info", );
    }

    my $policy_name = $task->{full_path} || $task->{policy_name};
    if (!$task->{target_policy} && !$policy_name) {
        $self->handle_error(action => 'execute',
                            entity_class => $entity_class,
                            thrift_obj  => $task,
                            error_code  => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                            display_message => "Policy name is mandatory.",
                            error_level  => "info", );
    }

    eval {
        $self->begin_work($self->{session});
        if ($task->{target_policy}) {
            my $import_open_api_actions_fields = F5::ASMConfig::ImportOpenApiActionsFields->new(filename => $filename,
                                                                                                is_base64 => $task->{is_base64},
                                                                                                username => $task->{username},
                                                                                                api_type => $task->{api_type},
                                                                                                api_content_format => $task->{api_content_format},
                                                                                                encoding => $task->{encoding},);
            my $rv = $self->import_open_api_update_policy($session, $call_flags, $task->{target_policy}, $import_open_api_actions_fields);
            $task->{result_policy} = $rv->{policy};
        }
        else {
            my $partition = $session->{partition};
            if ($policy_name && ($policy_name =~ m|^/([^/]*)|)) {
                $partition = $1;
            }

            my $import_open_api_actions_fields = F5::ASMConfig::ImportOpenApiActionsFields->new(filename => $filename,
                                                                                                is_base64 => $task->{is_base64},
                                                                                                username => $task->{username},
                                                                                                policy_name => $policy_name,
                                                                                                policy_template => $task->{policy_template},
                                                                                                partition => $partition,
                                                                                                policy_description => $task->{policy_description},
                                                                                                api_type => $task->{api_type},
                                                                                                api_content_format => $task->{api_content_format},
                                                                                                encoding => $task->{encoding},);

            my $rv = $self->import_open_api_create_policy($session, $call_flags, $import_open_api_actions_fields);
            $task->{result_policy} = $rv->{policy};
        }
        $self->commit($self->{session});
    };

    if ($@) {
        $task->{message} = $@->{error_message};
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
    }
    else {
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
    }

    if ($task->{file_contents} && !$task->{filename}) {
        unlink($filename);
    }

}

#----------------------------------------------------------------------

sub execute_import_open_api_task {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'import_open_api',
                                                          execute_func => sub { import_open_api(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_import_open_api_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_import_open_api_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
