package F5::ASMConfig::Handler::ImportPolicyTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ImportPolicyTask;
use F5::PolicyHistory;
use F5::Utils::Rest;
use MIME::Base64;
use F5::ASM::COAPI;
use F5::YAML::Schema::Utils;
use F5::Utils::File;

our @handled_functions = (
                            {
                             function_name => 'add_import_policy_task',
                             criteria_index => 3,
                             store_index    => 4,
                             write_access => 1,
                            },
                            {
                              function_name => 'update_import_policy_task',
                              criteria_index => 3,
                              store_index => 4,
                              write_access => 1
                            },
                            { 
                             function_name => 'execute_import_policy_task',
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                            },
                            {
                             function_name => 'delete_import_policy_task',
                            },
                            {
                             function_name => 'delete_many_import_policy_tasks',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ImportPolicyTask';

#----------------------------------------------------------------------

sub add_import_policy_task {
  my($self, @rpc_params) = @_;
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => \@rpc_params,
                                                    entity_class => $entity_class,
                                                    task_type => 'import_policy',);
}

#----------------------------------------------------------------------

sub update_import_policy_task {
    my ($self, @rpc_params) = @_;   
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'import_policy',);
}

#----------------------------------------------------------------------

sub is_base64 {
    my (%params) = @_;
    my $file_contents = $params{file_contents};
    my $filename = $params{filename};
    if ($filename && -T $filename) {
        open(my $inf, '<', $filename);
        read($inf, $file_contents, 50);
        close $inf;
    }

    return ($file_contents && $file_contents !~ /<\?xml / && $file_contents !~ /{/ && $file_contents !~ /policy:/);
}

#----------------------------------------------------------------------

sub import_policy {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};
    my $policy_description = $task->{policy_description};

    my $delete_filename;

    my $filename = $task->{filename};
    my $file_contents = $task->{file_contents};
    my $file_reference = $task->{file_reference};

    if ($file_reference) {
       my $rv =  F5::Utils::File::get_file_from_link(user => $task->{username},
                                                     link => $file_reference);
       $task->{filename} = $rv->{filename};
       $filename = $task->{filename};
    }

    if ($filename) {
        $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename} );

        if (!$task->{is_base64} && is_base64(filename => $filename)) { #Auto-detect base64
            $task->{is_base64} = 1;
             $self->{logger}->debug("Detected Base64 File");
        }

        if ($task->{is_base64}) {
           #Use file_contents processing to decode this.
           $file_contents = F5::Utils::File::file_slurp($filename);
           undef $filename;
        }
    }

    if ($file_contents) {
        if (!$filename) {
            $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{rest_uuid} );
            $delete_filename = 1;
        }
        open (my $FH, '>', $filename) || $self->handle_error(action => 'execute',
                                                             entity_class => $entity_class,
                                                             thrift_obj  => $task,
                                                             error_code  => F5::ASMConfig::ErrorCode::PERMISSION_DENIED(),
                                                             msg         => "Could not open $filename for writing.",
                                                             error_level  => "info", );
       if ($task->{is_base64} || is_base64(file_contents => $file_contents)) {
           $file_contents = MIME::Base64::decode_base64($file_contents);
       }
       print $FH $file_contents;
       close ($FH);
    }

    my $rv;
    my $policy = $task->{target_policy};
    if (!$policy && $task->{policy_history}) {
        $policy = $task->{policy_history}->{policy};
    }
    if ($policy) {
        $policy = F5::ASMConfig::Cache::get_cached_policy(policy => $policy, dbh => $self->{dbh});
    }

    my $policy_name = $task->{full_path} || $task->{policy_name} || $policy->{policy_name};
    my $partition = $session->{partition};
    if ($policy_name && ($policy_name =~ m|^/([^/]*)|)) {
        $partition = $1;
    }
    
    if ($task->{source_policy}) {
        if (!$task->{source_policy}->is_modified && (!$task->{encoding} || $task->{encoding} eq $task->{source_policy}->{encoding})
                                                 && (!$task->{policy_type} || $task->{policy_type} eq $task->{source_policy}->{policy_type})) {
            #Shortcut!
            $filename = F5::PolicyHistory->construct_filename(revision => $task->{source_policy}->last_revision_number,
                                                              policy_id => $task->{source_policy}->{policy_id},)
        }

        if (!($filename && -e $filename )) {
            $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{rest_uuid} );
            $delete_filename = 1;

            my $file_format = 'binary';
            if ( $task->{encoding} && $task->{source_policy}->{encoding} &&
                  $task->{encoding} ne $task->{source_policy}->{encoding} ) {
                $file_format = 'xml';
            }

            $self->export_policy($session, $call_flags, $task->{source_policy},
                                 F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                            file_format => $file_format)
                                );
       }
       if (!$policy_description) {
           $policy_description = "Copied from $task->{source_policy}->{policy_name}";
       }
   }

    if (!$filename && $task->{policy_history}) {
        $filename = F5::PolicyHistory->construct_filename(revision => $task->{policy_history}->{revision},
                                                          policy_id => $task->{policy_history}->{policy_id},);
    }

    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                              policy_name => $policy_name,
                                                                              force_policy_name_flg => defined ($task->{target_policy}),
                                                                              policy_template => $task->{policy_template},
                                                                              encoding => $task->{encoding},
                                                                              flg_encoding_selected => $task->{flg_encoding_selected},
                                                                              policy_type => $task->{policy_type},
                                                                              parent_policy => $task->{parent_policy},
                                                                              retain_inheritance_settings => $task->{retain_inheritance_settings},
                                                                              get_policy_attributes_only => $task->{get_policy_attributes_only},
                                                                              partition => $partition,
                                                                              policy_description => $policy_description,
                                                                              );
    eval {
        if ($import_export_properties->{force_policy_name_flg}) {
            my @check_fields = qw(policy_type encoding flg_encoding_selected);
            my $err_field;
            if ($policy->{parent_policy_id} ) {
                if ($import_export_properties->{parent_policy} && 
                    ($import_export_properties->{parent_policy}->{policy_id} != $policy->{parent_policy_id})) {
                    $err_field = "parent_policy";
                }
                $import_export_properties->{parent_policy} = F5::ASMConfig::Policy->new(policy_id =>  $policy->{parent_policy_id});
            }
            for my $check_field (@check_fields) {
                last if $err_field;
                if (defined($import_export_properties->{$check_field}) && 
                      $import_export_properties->{$check_field} ne $policy->{$check_field}) {
                    $err_field = $check_field;
                }
                $import_export_properties->{$check_field} = $policy->{$check_field};
            }
     
            if ($err_field) {
                my $display_err_field = $task->get_field_display_name_string(field_name => $err_field, call_flags => $call_flags);
                return $self->handle_error( action => "import_policy",
                                            entity_class => $entity_class,
                                            error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                            thrift_obj   => $task,
                                            error_thrown => "Cannot change '$display_err_field' when replacing an existing Policy",
                                            error_level  => 'info',
                                            empty_string_value => $policy->{policy_name},
                                          );
            }
    
        }
    
        $self->begin_work($self->{session});
        $rv = $self->import_policy($session, $call_flags, $import_export_properties);
    
        if ($rv->{msgs}) {
            my ($msg) = join ("\n", map { $_->{msg} } @{$rv->{msgs}});
            $task->{message} = $msg;
        }
        
        if (!$rv->{completed_successfully}) {
            $self->rollback($self->{session});
            $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
        }
        else {
            $self->commit($self->{session});
            #TODO: Include validation report?
            $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
            if ($task->{get_policy_attributes_only}) {
                if (defined $rv->{flg_allow_parent_policy}){
                    $task->{flg_allow_parent_policy} = $rv->{flg_allow_parent_policy};
                }
                my $result_policy = F5::ASMConfig::Entity::Policy->new(object => $rv->{policy}, dbh => $self->{dbh});
                my $coder = JSON::XS->new->ascii->allow_nonref;
                $task->{result_policy_attributes} = $coder->encode($result_policy->to_rest(call_flags => {show_all_rest_fields => 1}));
            }
            else {
                $task->{result_policy} = $rv->{policy};
                if (!$task->{target_policy}) {
                    $call_flags->{writes_to_mcp} = 1;
                }
           }
        }
    };
    my $err = $@;

    if ($file_reference) {
        $task->{filename} = "";
    }

    if ($delete_filename) {
        unlink($filename);
    }

    if ($err) {
        die $err;
    }
}


#----------------------------------------------------------------------
sub execute_import_policy_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'import_policy',
                                                          execute_func => sub { import_policy(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_import_policy_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_import_policy_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
