package F5::ASMConfig::Handler::ImportPolicyTemplateTask;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::TaskBase/;
use F5::ASMConfig::Entity::ImportPolicyTemplateTask;
use F5::PolicyTemplateDefaults;
use F5::Utils::Rest;
use File::Copy;
use File::Basename;
use JSON::XS;

our @handled_functions = (
                            {
                             function_name => 'add_import_policy_template_task',
                             criteria_index => 3,
                             store_index    => 4,
                             write_access   => 1,
                            },
                            {
                              function_name => 'update_import_policy_template_task',
                              criteria_index => 3,
                              store_index => 4
                            },
                            { 
                             function_name => 'execute_import_policy_template_task',
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                            },
                            {
                             function_name => 'delete_import_policy_template_task',
                            },
                            {
                             function_name => 'delete_many_import_policy_template_tasks',
                            },

                         );

our $entity_class = 'F5::ASMConfig::Entity::ImportPolicyTemplateTask';

#----------------------------------------------------------------------

sub add_import_policy_template_task {
  my($self, @rpc_params) = @_;
  return F5::ASMConfig::Handler::TaskBase::add_task(self => $self,
                                                    rpc_params => \@rpc_params,
                                                    entity_class => $entity_class,
                                                    task_type => 'import_policy_template',);
}

#----------------------------------------------------------------------

sub update_import_policy_template_task {
    my ($self, @rpc_params) = @_;   
    return F5::ASMConfig::Handler::TaskBase::update_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                         task_type => 'import_policy_template',);
}

#----------------------------------------------------------------------
my $GUI_UPLOAD = '/shared/tmp';

sub import_policy_template {
    my (%params) = @_;
    my $self       = $params{self};
    my $session    = $params{session};
    my $call_flags = $params{call_flags};
    my $task       = $params{task};

    my $delete_filename;

    my $filename = $task->{filename};
    if ($filename) {
        $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{filename} );
        if (! -e $filename) {
            my $gui_filename = "$GUI_UPLOAD/" . basename($task->{filename});
            if (-e $gui_filename) {
                $filename = $gui_filename;
            }
        }
    }
    if ($task->{file_contents}) {
        if (!$filename) {
            $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{rest_uuid} );
            $delete_filename = 1;
        }
        open (my $FH, '>', $filename) || $self->handle_error(action => 'execute',
                                                             entity_class => $entity_class,
                                                             thrift_obj  => $task,
                                                             error_code  => F5::ASMConfig::ErrorCode::PERMISSION_DENIED(),
                                                             error_thrown => "Could not open $filename for writing.",
                                                             error_level  => "info", );
       my $contents = $task->{file_contents};
       if ($task->{is_base64}) {
           $contents = MIME::Base64::decode_base64($contents);
       }
       print $FH $contents;
       close ($FH);
    }

    if ($task->{filename} && ! -e $filename) {
        $self->handle_error(action => 'execute',
                            entity_class => $entity_class,
                            thrift_obj  => $task,
                            error_code  => F5::ASMConfig::ErrorCode::FILE_NOT_FOUND(),
                            display_message => "Could not find file: $task->{filename}",
                            error_thrown => "Could not find file: $filename",
                            error_level  => "info", );
    }

    my $policy = $task->{target_policy};
    my $existing_template = $task->{target_policy_template};
    $call_flags->{force_update_read_only} = 1;

    my $rv;
    eval {
        if ($existing_template) {
            my $err;
            if (!$existing_template->{flg_is_user_defined}) {
                $err = "Can not overwrite Factory Policy Template";
            }
            if ($policy &&
                $policy->{policy_type} && $existing_template->{policy_type} &&
                $policy->{policy_type} != $existing_template->{policy_type}) {
                $err =  'Can not overwrite an existing Policy Template with a Policy of different type';
            }
            if ($err) {
                return $self->handle_error(action => 'execute',
                                       error_thrown => $err,
                                       error_code => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                       call_flags => $call_flags,
                                       entity_class => $entity_class,
                                       thrift_obj => $task,
                                      );
            }

            if (defined($task->{description})) {
                $existing_template->{description} = $task->{description};
            }

            if ($policy) {
                $filename = F5::Utils::Rest::get_full_file_path(username => $task->{username}, filename => $task->{rest_uuid} );
                $delete_filename = 1;
                $self->export_policy($session, $call_flags, $policy, F5::ASMConfig::ImportExportProperties->new({
                                                                     filename => $filename,
                                                                     file_format => 'xml',
                                                                     }));
            }

            if (!$policy) {
                require F5::ImportExportPolicy::XML;
                F5::ImportExportPolicy::XML::check_schema($filename);
            }

            my $template_defaults_json = F5::PolicyTemplateDefaults::xml_template_defaults($filename);
            if ($template_defaults_json) {
                if (!$policy) {
                    my $new_defaults = JSON::XS->new->utf8(0)->decode($template_defaults_json);
                    my $existing_defaults = $existing_template->{template_defaults};
                    if ($existing_defaults->{type} && $new_defaults->{type} &&
                        $existing_defaults->{type} ne $new_defaults->{type}) {
                        return $self->handle_error(action => 'execute',
                                       error_thrown => 'Can not overwrite an existing Policy Template with a Policy of different type',
                                       error_code => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                       call_flags => $call_flags,
                                       entity_class => $entity_class,
                                       thrift_obj => $task,
                                      );
                    }
                }
                $existing_template->{template_defaults} = $template_defaults_json;
            }

            my $template_filename = "$existing_template->{policy_template_name}.xml";
            $template_filename =~ s/[ \/]/_/g;
            $template_filename = "$F5::ASMConfig::Handler::PolicyTemplate::TEMPLATE_DIR/$template_filename";
            move($filename, $template_filename);

            $existing_template->{last_updated} = time();

            $rv = $self->update_policy_template($session, $call_flags, $existing_template);
        }
        else {
            my $description = defined($task->{description})                ? $task->{description} 
                            : ($policy && defined($policy->{description})) ? $policy->{description}
                            : "";
            my $template_name = $task->{template_name} || $task->{template_title};
            my $policy_template = F5::ASMConfig::PolicyTemplate->new(policy_template_name => $template_name,
                                                                     description => $description,
                                                                    );
            $rv = $self->add_policy_template($session, $call_flags, $policy, $policy_template, $filename);
       }    

        $task->{result_policy_template} = $rv;
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED();
    };
    my $err = $@;

    if ($delete_filename) {
        unlink($filename);
    }

    if ($err) {
        die $err;
    }
}


#----------------------------------------------------------------------

sub execute_import_policy_template_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::execute_task(self => $self,
                                                          rpc_params => \@rpc_params,
                                                          entity_class => $entity_class,
                                                          task_type => 'import_policy_template',
                                                          execute_func => sub { import_policy_template(@_); },
                                                          );
}

#----------------------------------------------------------------------

sub delete_import_policy_template_task {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_task(self => $self,
                                                         rpc_params => \@rpc_params,
                                                         entity_class => $entity_class,
                                                        );
}

#----------------------------------------------------------------------

sub delete_many_import_policy_template_tasks {
    my ($self, @rpc_params) = @_;

    return F5::ASMConfig::Handler::TaskBase::delete_many_tasks(self => $self,
                                                               rpc_params => \@rpc_params,
                                                               entity_class => $entity_class,
                                                              );
}

#----------------------------------------------------------------------

1;
