package F5::ASMConfig::Handler::L7Policy;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::SetActive;
use F5::LockFactory;
use F5::DbUtils;

our @handled_functions = (
                           {
                             function_name => "activate_l7_asm_policies",
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                             relayable => 0, # l7_policy might not be present on the peer device
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::L7Policy';

#----------------------------------------------------------------------

sub activate_l7_asm_policies {
    my ($self, $session, $call_flags, $place_holder, $l7_policy_criteria) = @_;

    my $l7_policy = $self->base_get(rpc_params => [$session, $call_flags, $place_holder, $l7_policy_criteria],
                                    entity_class => $entity_class,
                                   )->{return_value};

    my $l7_policy_actions = $l7_policy->{actions};

    # actions may be empty, for the related l7_policy
    return { return_value => 1 }
        unless ref($l7_policy_actions) eq 'ARRAY' && scalar(@$l7_policy_actions);

    my $l7_policy_name = $l7_policy->{l7_policy_name};

    $self->begin_work($self->{session});
    for my $policy_action (@$l7_policy_actions) {
        # the 'policy' w/ in the action is the name of the asm policy referenced by that action
        # it may be empty if there is no asm policy in the action
        next unless ref($policy_action) && $policy_action->{policy};
        my $policy_name = $policy_action->{policy};

        my $msg = "ASM policy named '$policy_name' which is referenced by the LTM policy '$l7_policy_name'";

        my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                        get_criteria => {
                                                            policy_name => $policy_name,
                                                        },
                                                       );

        if (!ref($policy) || !$policy->{policy_id}) {
            return $self->handle_error(action => "activate_l7_asm_policies",
                                       error_code => F5::ASMConfig::ErrorCode::MISSING_POLICY(),
                                       error_thrown => "Could not find $msg",
                                       error_level => "error",
                                      );
        }

        # We should not be applying changes to a modified policy that is already active
        next if ($policy->{is_active_policy});

        $self->{logger}->info("Activating $msg") if $self->{logger};
        $call_flags->{writes_to_mcp} = 1;

        my $set_active_flags = F5::ASMConfig::SetActiveFlags->new($call_flags->{set_active_flags});
        $set_active_flags->{activate_policy} = 1;

        my $rv = $self->set_active($session, $call_flags, $policy, $set_active_flags);

        if (!$rv) {
            return $self->handle_error(action => "set_active",
                                       error_code => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                       error_thrown => "Could not activate $msg",
                                       error_level => "error",
                                      );
        }
    }
    $self->commit($self->{session});

    return { return_value => 1 };
}

#----------------------------------------------------------------------

1;
