package F5::ASMConfig::Handler::LoggingProfile;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::ASM::Tmsh;
use F5::ASM::COAPI;
use F5::BigipUtils;

use File::Temp qw/tempfile/;
use POSIX;
use File::Basename;


our @handled_functions = (
                            {
                                function_name => "search_logging_profile",
                            },
                            {
                                function_name => "fetch_logging_profile",
                            },
                            {
                                function_name => "deploy_logging_profile",
                            },
                            {
                                function_name => "append_logging_profile",
                                write_access => 1,
                            },
                            {
                                function_name => "associate_logging_profile",
                                write_access => 1,
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::LoggingProfile';

#----------------------------------------------------------------------

sub search_logging_profile {
    my($self, $session, $call_flags, $place_holder, $logging_profile) = @_;

    my @logging_profiles;

    my %logprof_names;
    my $app_query_results = F5::ASM::COAPI::Query->new('fw_log_application', {folder_name => '/Common'});
    my $dos_query_results = F5::ASM::COAPI::Query->new('fw_log_dos_application', {folder_name => '/Common', 
                                                                                  local_publisher => '/Common/local-db-publisher',
                                                                                  remote_publisher => ''});
    my $protocol_transfer_query_results = F5::ASM::COAPI::Query->new('fw_log_protocol_transfer', {folder_name => '/Common'});
    my $network_query_results = F5::ASM::COAPI::Query->new('fw_log_network', {folder_name => '/Common'});
    my $protocol_dns_query_results = F5::ASM::COAPI::Query->new('fw_log_protocol_dns', {folder_name => '/Common'});

    my $app_and_dos_query_results = [@$app_query_results, @$dos_query_results];
    for my $result (@$app_and_dos_query_results) {
        if ((! exists $logprof_names{$result->{profile_name}}) &&
            (! grep {$result->{profile_name} eq $_->{profile_name}} (@$protocol_transfer_query_results,
                                                                     @$network_query_results,
                                                                     @$protocol_dns_query_results))){
            $logprof_names{$result->{profile_name}} = $result;
        }
    }

    for my $name (keys %logprof_names) {
        my $logging_profile = F5::ASMConfig::LoggingProfile->new(logging_profile_name => basename($name));
        if (grep {$name eq $_->{profile_name}} @$app_query_results) {
            $logging_profile->{flg_has_application} = 1;
        }
        if (grep {$name eq $_->{profile_name}} @$dos_query_results) {
            $logging_profile->{flg_has_dos} = 1;
        }
        my @tmsh_cmd = ("tmsh", "list", "security", "log", "profile", $name);
        $logging_profile->{tmsh_data} = F5::ASM::Tmsh::Query->new(cmd => \@tmsh_cmd, logger => $self->{logger})->{standard_output};
        $logging_profile = F5::ASMConfig::LoggingProfile->new($logging_profile);

        if (F5::BigipUtils::is_device_em()) {
            my ($device_id) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $name);
SELECT device_id
  FROM SECURITY.LOGGING_PROFILES
 WHERE logging_profile_name = ?
EndQuery
            $logging_profile->{device_id} = $device_id;
        }
        push (@logging_profiles, $logging_profile);
    }


    my $result = $self->entities_to_search_result(entities => \@logging_profiles,
                                                  call_flags => $call_flags,
                                                  entity_class => $entity_class,);
    return { return_value => $result };
}

#----------------------------------------------------------------------

sub fetch_logging_profile {
    my($self, $session, $call_flags, $place_holder, $device, $logging_profile) = @_;

    my ($fh, $tmp_filename) = tempfile();
    print $fh $logging_profile->{tmsh_data};
    close $fh;

    my @tmsh_cmd = ("tmsh", "load", "sys", "config", "merge", "file",  $tmp_filename);
    F5::ASM::Tmsh::Query->new(cmd => \@tmsh_cmd, logger => $self->{logger});

    $self->{dbh}->do(<<EndQuery,undef,$logging_profile->{logging_profile_name}, $device->{device_id});
REPLACE INTO SECURITY.LOGGING_PROFILES
(logging_profile_name, device_id)
VALUES
(?,?)
EndQuery

    unlink($tmp_filename);

    $self->write_to_syslog(msg => "Logging Profile \"$logging_profile->{logging_profile_name}\" was fetched from device $device->{device_name}.");
    my $retval = { return_value => $logging_profile, };
    return $retval;
}

#----------------------------------------------------------------------

sub deploy_logging_profile {
    my($self, $session, $call_flags, $place_holder, $deploy_properties, $logging_profile) = @_;

    my ($fh, $tmp_filename) = tempfile();
    print $fh $logging_profile->{tmsh_data};
    close $fh;
    my $CHUNK_SIZE = 2**19; # Half a MB
    my $file_size = -s $tmp_filename;
    my $total_chunks = ceil($file_size / $CHUNK_SIZE);

    my %target_devices;
    for my $deploy_target (@{$deploy_properties->{deploy_targets}}) {
        my $device = $self->get_device($session, $call_flags, $deploy_target->{device});
        if (! exists $target_devices{$device->{mng_ip_address}}) {
            $target_devices{$device->{mng_ip_address}} = $device;
            $device->{virtual_servers} = [];
        }
        $device = $target_devices{$device->{mng_ip_address}};
        next unless defined $deploy_target->{virtual_server};

        push (@{$device->{virtual_servers}}, $deploy_target->{virtual_server}->{virtual_server_name});
    }

    my $multi_client = F5::ASMConfig::Client->new(
        logger => $self->{logger},
        devices => [ values %target_devices ],
        is_async => 1,
        is_ssl => 1,
        deploy_request_fields => {
                                  deploy_type => F5::ASMConfig::ElementType::LOGGING_PROFILE(),
                                  logging_profile_name => $logging_profile->{logging_profile_name},
                                 },
        deploy_targets => $deploy_properties->{deploy_targets},
        async_calls => [
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;
                                          $client->get_temp_filename($client->{session}, $call_flags);
                                         },
                         call_name => 'Get Temp Filename',
                         return_value => 'tempfile_name',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;
                                          for my $chunk_index (1..$total_chunks) {
                                              open (my $in, "<", $tmp_filename);
                                              seek ($in, $CHUNK_SIZE*($chunk_index-1), 0);
                                              my $chunk;
                                              read $in, $chunk, $CHUNK_SIZE;
                                              close $in;
                                              $client->store_file_chunk($client->{session}, $call_flags, $chunk, $retvals->{tempfile_name});
                                          }
                                         },
                         call_name => 'Transfer Logging Profile',
                        },
                        {
                         call_sub => sub {
                                          my($client, $retvals) = @_;
                                          $client->append_logging_profile($client->{session}, $call_flags, $retvals->{tempfile_name});
                                         },
                         call_name => 'Import Logging Profile',
                        },
                        {
                         call_sub => sub {
                             my($client, $retvals) = @_;
                             $self->write_to_syslog(msg => "Logging Profile \"$logging_profile->{logging_profile_name}\" was deployed to device $client->{device}->{device_name}.");

                             for my $virtual_server (@{$client->{device}->{virtual_servers}}) {

                                 $client->associate_logging_profile($client->{session}, $call_flags,
                                                                    $logging_profile->{logging_profile_name},
                                                                    [ $virtual_server ]);
                             }
                         },
                         call_name => 'Associate Logging Profile',
                        },
                    ],
    );

    return {return_value => $multi_client->run_async_calls(dbh => $self->{dbh})};
}

#----------------------------------------------------------------------

sub append_logging_profile {
    my($self, $session, $call_flags, $place_holder, $logging_profile_filename) = @_;

    my @error_messages;
    my @tmsh_cmd = ("tmsh", "load", "sys", "config", "merge", "file",  $logging_profile_filename);
    my $rv = F5::ASM::Tmsh::Query->new(cmd => \@tmsh_cmd, logger => $self->{logger});
    if (! $rv->{completed_successfully}) {
        push (@error_messages, $rv->{standard_error});
    }

    if (@error_messages) {
        my $error_message = join("\n", @error_messages);
        return $self->handle_error(action => 'append_logging_profile',
                                   error_thrown   => $error_message,
                                   display_message => $error_message,
                                   error_level     => 'error',
                                   error_code      => F5::ASMConfig::ErrorCode::COAPI_QUERY_FAILURE(),);
    }

    return { return_value => 1, };
}

#----------------------------------------------------------------------

sub associate_logging_profile {
    my($self, $session, $call_flags, $place_holder, $logging_profile_name, $virtual_server_names) = @_;

    my @error_messages;
    for my $virtual_server_name (@$virtual_server_names) {
        my @tmsh_cmd = ("tmsh", "modify", "ltm", "virtual", $virtual_server_name, "security-log-profiles", "replace-all-with", "{", $logging_profile_name, "}");
        my $rv = F5::ASM::Tmsh::Query->new(cmd => \@tmsh_cmd, logger => $self->{logger});
        if (! $rv->{completed_successfully}) {
            push (@error_messages, $rv->{standard_error});
        }
    }

    if (@error_messages) {
        my $error_message = join("\n", @error_messages);
        return $self->handle_error(action => 'associate_logging_profile',
                                   error_thrown   => $error_message,
                                   display_message => $error_message,
                                   error_level     => 'error',
                                   error_code      => F5::ASMConfig::ErrorCode::COAPI_QUERY_FAILURE(),);
    }

    return { return_value => 1, };
}

#----------------------------------------------------------------------

1;
