package F5::ASMConfig::Handler::McpDb;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::Deploy;
use F5::DbUtils;
use F5::Cluster::Utils;
use F5::LockFactory;
use F5::PrepareConf::BotDefense;
use F5::PrepareConf::Policy;
use F5::SetActive;
use Socket qw(:DEFAULT inet_pton inet_ntop);
use F5::License;

our @handled_functions = (
                          {
                              function_name => 'update_mcp_db',
                              relayable => 0,
                          },
                         );

my $LICENSE_STATE_CFG_FILE = F5::Deploy::map_path('/ts/files/site_1/config/full/license_state.cfg');

#------------------------------------------------------------------
sub update_mcp_db {
    my ($self, $session, $call_flags, $place_holder, $mcp_db, $set_active_flags) = @_;

    my $var_name = $mcp_db->{mcp_db_name};
    my $var_value = $mcp_db->{mcp_db_value};
    my $logger = $self->{logger};

    my $lock_file =  $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                             logger => $logger);

    if ($var_name=~/(Failover.State)/i) {
            $logger->info("sending value $var_value for $var_name");
            set_failover_conf($self, $session, $call_flags);
    } elsif ($var_name =~ /(hostname|admin_ip_addr|license_state|cluster.mgmtipaddr|asm\.credential_stuffing_service|cloud\.licensed\.bewaf)/i) {
            $locker->lock();
            update_enforcer(client => $self, var_name => $1, var_value => $var_value, flg_update_core => $set_active_flags->{do_update_core});
            $locker->unlock();
    } elsif ($var_name =~ /( cluster\.primary |
                             cluster\.blade\.green\.count |
                             platform\.chassis\.slot |
                             mysql\.password )/ix) {
            $locker->lock();
            set_cluster_conf(asm_config_handler => $self,
                             session => $session,
                             call_flags => $call_flags,
                             do_update_core => $set_active_flags->{do_update_core},
                             var_name => $var_name,
                             var_value => $var_value,
                            );
            $locker->unlock();
    } elsif ($var_name =~ /Log\.ASM\.(.*).Level/i) {
        F5::Logger::update_logging_file();
    } else {
            ## handling other bigdb variables should be handled here
            ## preferably in separate elsifs...
    }

    return {return_value => $mcp_db, events => []};
}

#------------------------------------------------------------------

sub update_enforcer {
    my (%args) = @_;

    my $client = $args{client};
    my $dbh = $client->{dbh};
    my $logger = $client->{logger};
    my $var_name = $args{var_name};
    my $var_value = $args{var_value};
    my $flg_update_core =$args{flg_update_core};

    if ($var_name=~/(asm\.credential_stuffing_service)/i) {
        my $flg_credential_stuffing_subscription = $var_value eq "enable" ? 1 : 0;
        my $credential_filename = '/var/ts/bloom/bloom_credentials_blob.dat';

        my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                         table_name => "DCC.CREDENTIAL_STUFFING_UPDATE_FILE",
                                         rec => { credential_filename => $credential_filename,
                                                  flg_credential_stuffing_subscription => $flg_credential_stuffing_subscription },
                                        );
        $flg_update_core ||= $rv;

        # A change to ASM.credential_stuffing_service is treated as a license change
        $var_name = 'license_state';
    }

    if ( $var_name eq 'license_state' ) {
        $var_value = F5::License::get_asm_license_mode();
        unlink $LICENSE_STATE_CFG_FILE if -f $LICENSE_STATE_CFG_FILE;
        $logger->info("update enforcer of variable $var_name with value $var_value");
        my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                         table_name => "DCC.LICENSE_STATE",
                                         rec => { state => $var_value },
                                        );
        $flg_update_core ||= $rv;
    }
    else{
        my $admin_ip_addr = '';
        my $admin_ip_addr2 = '';
        my $hostname = F5::Cluster::Utils::get_mcp_val('hostname');
        $client->{shared_state}->{local_hostname} = $hostname;

        if (  F5::Cluster::Utils::is_clustering_enabled() ){
            my $cluster_admin_ips = F5::ASM::COAPI::Query->new(q{cluster});
            $admin_ip_addr = $cluster_admin_ips->[0]{float_mgmt_ip} || '';
            $admin_ip_addr2 = $cluster_admin_ips->[0]{float_mgmt2_ip} || '';
        }
        else {
            my $admin_ips = F5::ASM::COAPI::Query->new(q{admin_ip});
            if ( defined $admin_ips->[1] ){
                if ( inet_pton(AF_INET6,$admin_ips->[0]{addr}) ){
                    $admin_ip_addr = $admin_ips->[1]{addr} ;
                    $admin_ip_addr2 = $admin_ips->[0]{addr};
                }
                else{
                    $admin_ip_addr = $admin_ips->[0]{addr} ;
                    $admin_ip_addr2 = $admin_ips->[1]{addr};
                }
            }
            elsif ( defined $admin_ips->[0] ) {
                $admin_ip_addr =  $admin_ips->[0]{addr};
                $admin_ip_addr2 = '';
            }
        }

        my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                         table_name => "DCC.BIG_IP_UNIT_INFO",
                                         rec => { mng_ip_addr    => $admin_ip_addr,
                                                  mng_ip_addr2   => $admin_ip_addr2,
                                                  unit_hostname  => $hostname,
                                                },
                                         );
       $flg_update_core ||= $rv;
    }

    # update enforcer with new configuration
    require F5::SetActive;

    if ($flg_update_core) {
        $logger->debug("Sending Update core ...");
        eval {
            my $prepareconf_policy;
            if ($var_name eq 'license_state' or $var_name eq 'cloud.licensed.bewaf') {
                my $prepare_account_tables = 0;

                my ($prev_waf_licensed) = $dbh->selectrow_array(<<EndQuery, undef);
SELECT is_waf_licensed
  FROM DCC.LICENSE_STATE
EndQuery

                my $is_waf_licensed = (F5::License::is_waf_licensed() || F5::License::is_credential_stuffing_licensed()) ? 1 : 0;
                if ($prev_waf_licensed != $is_waf_licensed) {
                    my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                                     table_name => "DCC.LICENSE_STATE",
                                                     rec => { is_waf_licensed => $is_waf_licensed,
                                                              last_updated => time(), },
                                                    );

                    $prepareconf_policy = F5::PrepareConf::Policy->new(dbh => $dbh,
                                                                      logger => $logger, );

                    $prepareconf_policy->prepare_deception_traps();

                    #set_sigs_active calls set_active inside
                    #this set active is optimized to handle only specific DCC tables (see: modified_dcc_tables)
                    my $sigfile_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
                    $sigfile_update->set_sigs_active(asm_config_handler => $client);

                    my $bot_defense = F5::PrepareConf::BotDefense->new(dbh => $dbh,
                                                                       logger => $logger,
                                                                      );
                    $bot_defense->prepare_bot_defense_profiles();

		    $prepareconf_policy->prepare_credential_stuffing_file();

		    $prepareconf_policy->prepare_deception_response_pages();

                    my $active_policies = $dbh->selectall_hashref(<<EndQuery, 'id');
SELECT account_id, id
  FROM PLC.PL_POLICIES
 WHERE id IN (SELECT policy_id FROM PLC.PL_JSON_VALIDATION_FILES)
   AND account_id > 0
EndQuery

                    for my $active_policy (values %$active_policies) {
                        $prepareconf_policy->prepare(account_id => $active_policy->{account_id},
                                                     policy_id => $active_policy->{id},
                                                     # must prepare 'objects' when preparing 'json_profiles'
                                                     policy_parts_to_prepare => ['json_profiles',
                                                                                 'objects',
                                                                                 'blocking',
                                                                                 'server_technology_responses',
                                                                                 'signature_responses',
                                                                                 'custom_deception_responses',
                                                                                 'deception_traps',
                                                                                ],
                                                    );
                    }

                    $prepare_account_tables = scalar keys %$active_policies;
                }

                $dbh->commit unless $dbh->{AutoCommit};
                F5::SetActive::set_active(dbh => $dbh,
                                          logger => $logger,
                                          do_update_core => 1,
                                          send_sync => 1,
                                          prepare_account_tables => $prepare_account_tables,
                                          asm_config_handler => $client,
                                         );
            } 
            else {
                F5::SetActive::set_active(dbh => $dbh,
                                          asm_config_handler => $client,
                                          do_update_core => 1,
                                          logger => $logger,
                                          prepare_account_tables => 0,
                                          );
            }

        };
        if ($@) {
             return $client->handle_error(action => 'update',
                                        error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                        error_thrown => "Could not set active: $@",
                                        error_level  => 'error',
                                       );
        }
    }
    else {
         $logger->info("Update core will be sent during load-complete...");
    }
}

#------------------------------------------------------------------

sub set_failover_conf {
    my ($client, $session, $call_flags) = @_;
    my $logger = $client->{logger};
    
    $logger->info(qq{set_failover_conf called});

    $client->update_failover_state( $session,
                                    $call_flags,
                                    F5::ASMConfig::FailoverState->new({
                                                                       is_active => F5::Cluster::Utils::is_failover_active(),
                                                                       is_redundant => F5::Cluster::Utils::get_mcp_val("Failover.IsRedundant") eq "true" ? 1 : 0,
                                                                      }));
}

#------------------------------------------------------------------

sub set_cluster_conf {
    my (%params) = @_;
    my $asm_config_handler = $params{asm_config_handler};
    my $dbh = $params{dbh} || $asm_config_handler->{dbh};
    my $logger = $params{logger} || $asm_config_handler->{logger};
    my $call_flags = $params{call_flags};
    my $session = $params{session};
    my $do_update_core = $params{do_update_core};
    my $var_name = $params{var_name};
    my $var_value = $params{var_value};
    
    $logger->info(qq{set_cluster_conf called});

    my $mysql_user = F5::DbUtils::get_mysql_user();
    $logger->info("mysql user of bd is $mysql_user");

    if ($var_name eq 'mysql.password' &&
          $var_value ne $mysql_user) {
        $logger->info("variable $var_name is ignored for user $var_value");
        return;
    }

    my $is_clustering_enabled = F5::Cluster::Utils::is_clustering_enabled();
    $logger->info("is_clustering_enabled value from mcp is $is_clustering_enabled");

    my $is_master_blade = F5::Cluster::Utils::is_master_blade();
    $logger->info("is_master_blade value from mcp is $is_master_blade");

    my $primary_slot_number = F5::Cluster::Utils::get_master_blade();
    $logger->info("master_blade value from mcp is $primary_slot_number");

    my $slot_number = F5::Cluster::Utils::get_slot_number();
    $logger->info("slot_number value from mcp is $slot_number");

    my $number_of_blades = F5::Cluster::Utils::get_mcp_val('Cluster.Blade.Green.Count');
    $logger->info("number_of_blades value from mcp is $number_of_blades");

    if (! $is_clustering_enabled) {
        $logger->info("Clustering disabled") if $logger;
        return;
    }

    $asm_config_handler->update_cluster_state($session,
                                              $call_flags,
                                              F5::ASMConfig::ClusterState->new({
                                                                                is_clustering_enabled => $is_clustering_enabled,
                                                                                is_master_blade => $is_master_blade,
                                                                                master_slot => $primary_slot_number || 0,
                                                                                this_slot => $slot_number || 0,
                                                                               }),);

    my $was_updated = 0;
    my $rv = F5::DbUtils::upsert_singleton(dbh => $dbh,
                                           table_name => 'DCC.SLOT_NAME',
                                           rec => { slot_name => "slot$slot_number",
                                                    slot_number => $slot_number }
                                          );
    $was_updated ||= $rv;

    $rv = F5::DbUtils::upsert_singleton(dbh => $dbh,
                                        table_name => 'DCC.CONF_PRIMARY_BLADE_RATE_LIMIT',
                                        rec => {primary_blade_slot => "slot$primary_slot_number",
                                                green_blade_count => $number_of_blades,
                                                mysql_password => F5::DbUtils::get_mysql_password(user => $mysql_user, encrypted => 1),
                                                ip_address => "127.3.0.$primary_slot_number"});
    $was_updated ||= $rv;

    if ($was_updated) {
        if ($do_update_core) {
            $logger->debug("Sending Update core ...");
            F5::SetActive::set_active(dbh => $dbh,
                                      asm_config_handler => $asm_config_handler,
                                      do_update_core => 1,
                                      logger => $logger,
                                      prepare_account_tables => 0,
                                     );
        } else {
            $logger->info("Update core will be sent during load-complete...") if $logger;
        }
    }
    else {
        $logger->debug("There was no change in clustering state") if $logger;
    }
}

1;
