package F5::ASMConfig::Handler::Policy;

use strict;
use warnings;

use F5::Logger;
use F5::DbUtils;
use F5::ImportExportPolicy;
use F5::Violation;
use F5::ASMConfig::Cache;
use F5::Mcp::PolicyTarget;
use F5::Mcp::PolicyAction;
use F5::Mcp::PolicyStatus;
use F5::DeviceGroup;

use List::Util qw(sum);
use F5::ObjUtils qw(dclone);

use base qw/F5::ASMConfig::Handler::Base/;

# same as in F5::ASMConfig::Handler::VirtualServerAssoc
use constant ASM_AUTO_L7_POLICY_NAME_PREFIX => 'asm_auto_l7_policy__';

our @handled_functions = (
                           {
                             function_name => 'add_policy',
                             relayable     => 1,
                             store_index   => 3,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },                           
                           {
                             function_name => 'delete_policy',
                             relayable => 1,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'update_policy',
                             relayable     => 1,
                             criteria_index => 3,
                             store_index   => 4,
                             writes_to_mcp => 0, #will be overridden if mcp will be affected (activate/deactivate/encoding change)
                           },
                           {
                             function_name => 'reconfigure_policy',
                             relayable => 1,
                             store_index => 3,
                           },
                           {
                             function_name => 'is_policy_modified',
                           },
                           {
                             function_name => 'import_policy',
                             relayable => 1,
                             filename_arguments => ['3.filename'],
                             writes_to_mcp => 1,
                             skip_authz => 1, #Done internally
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'export_policy',
                             write_access => 1, # like Export button in GUI
                           },
                           {
                             function_name => 'can_update_is_protocol_free_flag',
                           },
                           {
                             function_name => 'replace_policy',
                             relayable => 1,
                             authz_params => [3, 4],
                           },
                           {
                             function_name => 'search_policy',
                             authz_limit_field => 'policy_id',
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::Policy';
#----------------------------------------------------------------------

sub search_policy {
    my($self, $session, $call_flags, @other_rpc) = @_;

    $self->add_partition_search_filter(session    => $session,
                                       call_flags => $call_flags);

    my $rv = $self->base_search(rpc_params => [$session, $call_flags, @other_rpc],
                                entity_class => $entity_class,
                               );
    if ($self->{em_client}) {
        for my $element (@{$rv->{return_value}->{element_list}}) {
            $self->to_em_slash_replace(obj => $element->{policy}, field_name => 'policy_name');
        }
    }

    return $rv;
}

#----------------------------------------------------------------------

sub add_policy {
    my ($self, $session, $call_flags, $place_holder, $policy) = @_;

    my $action = "add";

    if ( defined($policy->{policy_type}) && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() &&
         defined($policy->{flg_encoding_selected}) && !$policy->{flg_encoding_selected} ) {
        return $self->handle_error(action => $action,
                                   error_code => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                   entity_class => $entity_class,
                                   thrift_obj => $policy,
                                   error_thrown => "Cannot create a Parent Policy without an encoding",
                                   error_level => "info",
                                  );
    }

    if (!defined($policy->{policy_type})) {
        $policy->{policy_type} = F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY();
    }

    if (!defined($policy->{is_active_policy}) && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()) {
        $policy->{is_active_policy} = (defined(  $policy->{extra_virtual_server_assocs} ) &&
                                        scalar(@{$policy->{extra_virtual_server_assocs}})
                                      ) ? 1 : 0 ;
    }

    my $parent_policy;
    if ($policy->{parent_policy_name} && $policy->{parent_policy_name} ne 'none') {
        $parent_policy = $entity_class->new(dbh => $self->{dbh},
                                            get_criteria => {policy_name => $policy->{parent_policy_name}});
        if (!$parent_policy) {
            return $self->handle_error(action => "$action",
                                       error_thrown => "Missing Parent Policy $policy->{parent_policy_name}",
                                       error_code => F5::ASMConfig::ErrorCode::INVALID_POLICY(),
                                       entity_class => $entity_class,
                                       thrift_obj => $policy,
                                      );
        }
        $policy->{parent_policy_id} = $parent_policy->{policy_id};
    }
    elsif ($policy->{parent_policy_id}) {
        $parent_policy = $entity_class->new(dbh => $self->{dbh},
                                            get_criteria => {policy_id => $policy->{parent_policy_id}});
    }

    if (($call_flags->{rest_call} && !$policy->{parent_policy_id})
        || $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() ) {

        if (!defined($policy->{flg_encoding_selected})) {
            $policy->{flg_encoding_selected} = 1;
        }
        $policy->{encoding} ||= "utf-8";
    }

    $entity_class->policy_name_fixup(object => $policy);

    if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()) {
        if (defined($parent_policy)) {
            if ( !$policy->{flg_encoding_selected} ) {
                $policy->{flg_encoding_selected} = 1;
                $policy->{encoding} = $parent_policy->{encoding};
            }
            elsif ( $policy->{encoding} ne $parent_policy->{encoding} ) {
                return $self->handle_error( action => $action,
                                            entity_class => $entity_class,
                                            error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                            thrift_obj   => $policy,
                                            error_thrown => "Security Policy and Parent Policy's Application Language must be the same. $policy->{policy_name} is $policy->{encoding}, $parent_policy->{policy_name} is $parent_policy->{encoding}.",
                                            error_level  => 'info',
                                            empty_string_value => $policy->{policy_name},
                                          );
            }

            if( !defined($policy->{flg_is_protocol_free}) ) {
                $policy->{flg_is_protocol_free} = $parent_policy->{flg_is_protocol_free};
            }

            if( !defined($policy->{flg_is_case_insensitive}) ) {
                $policy->{flg_is_case_insensitive} = $parent_policy->{flg_is_case_insensitive};
            }

            if ($parent_policy->{child_policy_count} >= $self->{max_child_policies}) {
                return $self->handle_error( action => $action,
                                            entity_class => $entity_class,
                                            error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                            thrift_obj   => $policy,
                                            error_thrown => "Parent Policies can only have $self->{max_child_policies} child policies.",
                                            error_level  => 'info',
                                            empty_string_value => $policy->{policy_name},
                                          );
            }
        }
        elsif (!$session->{user_name} || $session->{user_name} ne $F5::ASMConfig::Handler::TSCONFD_USER_NAME ) {
            my $parent_policy_settings = $self->get_parent_policy_settings($session, $call_flags);
            if ( $parent_policy_settings && $parent_policy_settings->{flg_is_parent_policy_mandatory} ) {
                my $parent_policy_rv = get_compatible_parent_policies(self => $self,
                                                                      session => $session,
                                                                      policy => $policy);
                if ($parent_policy_rv->{total_record_count}) {
                    return $self->handle_error( action => $action,
                                                entity_class => $entity_class,
                                                error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                                thrift_obj   => $policy,
                                                error_thrown => 'Parent Policy is mandatory',
                                                error_level  => 'info',
                                                empty_string_value => $policy->{policy_name},
                                              );
                }
            }
        }
    }

    if ($policy->{policy_template_id}) {
        my $policy_template = F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $self->{dbh},
                                                                         get_criteria => {policy_template_id => $policy->{policy_template_id}});
        if ($policy_template) {
            $policy->{policy_template} = $policy_template;
        }
    }

    _validate_policy_template(self => $self,
                              properties => $policy,
                              action => $action,
                              policy_template => $policy->{policy_template});

    my $lock_file = $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                             logger => $self->{logger},);

    $locker->lock();

    my $rv = $self->base_add(entity_class => $entity_class,
                             rpc_params => [ $session, { %$call_flags, suppress_save_to_log => 1 },$place_holder,$policy ],);
    my $new_policy = $rv->{return_value};
    my $was_policy_added = (exists $rv->{events}) ? 1 : 0 ;
    if ($was_policy_added) {
        $new_policy->clean_policy_for_inheritance();
        if ($new_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()) {
            if ($new_policy->{parent_policy_id}) {
                $call_flags->{reinherit} = 1;
            }
            $call_flags->{is_in_add_policy} = 1;
            _handle_parent_policy_change(self => $self,
                                         call_flags => $call_flags,
                                         policy => $new_policy,
                                         session => $session,);
        }
    }
    else { #Protect against ignored errors (tsconfd)
        $new_policy = $entity_class->new(dbh => $self->{dbh}, get_criteria => $policy);
        if (!$new_policy) {
            $locker->unlock();
            return $rv;
        }
    }

    if ($new_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        delete $policy->{parent_policy_id}; #Protect against GUI/tsconfd setting parent_policy_id = 0 by default.
    }

    my $do_mcp_store = 0;
    my $unset_do_not_relay = 0;

    if ($self->{rpc_entry_depth} == 1 && $was_policy_added && $self->{device_group} &&
        !F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                 folder => $new_policy->mcp_folder(),
                                                 logger => $self->{logger},
                                                 create => $self->{is_relay_handler_process})) {
        if ($call_flags->{skip_mcp} && $self->{is_relay_handler_process}) {
            $call_flags->{skip_mcp} = 0; #turn off in relay
            $do_mcp_store = 1;
        }
        $unset_do_not_relay = 1;
    }

    # Update fields (from tmsh or REST) that were ignored in store_add (although accepted in mcp_store)
    my $db_fields_updated = $new_policy->has_value_changed(existing_obj => $new_policy, store_obj => $policy, 
                                check_fields => [qw(is_pb_enabled is_learning_enabled is_active_policy parent_policy_id description flg_transparent_mode flg_passive_mode flg_is_protocol_free flg_is_case_insensitive extra_virtual_server_assocs flg_managed_by_bewaf)]);

    my $saved_events = delete $self->{events} || [];

    my $was_policy_updated = 0;
    if ($db_fields_updated) {
        $policy->{policy_name} = $new_policy->{policy_name}; # full path
        $policy->{policy_type} = $new_policy->{policy_type}; # populated type

        my $update_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags);
        $update_call_flags->{is_in_add_policy} = 1; # Flag to tell update_policy how to deal with inheritance properly. 
        $update_call_flags->{reinherit} = 1; # Flag to tell update_policy how to deal with inheritance properly. 
        $update_call_flags->{suppress_events} = 1;
        $update_call_flags->{force_update_read_only} = 1;
        $new_policy = $rv->{return_value} = $self->update_policy($session, $update_call_flags, $policy);
    }

    if ($was_policy_added && $new_policy->is_modified()) {
        if ($new_policy->{is_active_policy}) {
            $self->set_active($session, $call_flags, $new_policy, F5::ASMConfig::SetActiveFlags->new()); # apply active policy after change
        }
        else {
            $new_policy->clear_modification_log();
            $new_policy->{is_modified} = 0;
        }
    }

    $rv->{return_value} = $new_policy;
    if (!$call_flags->{suppress_events} && $was_policy_added) {
        my @add_events = $self->create_event(event_type => F5::ASMConfig::EventType::ADD,
                                            session    => $session,
                                            element => $rv->{return_value},
                                            policy  => $rv->{return_value},
                                            update_thrift_obj => $policy,
                                            call_flags => $call_flags,
                                            user_name  => $session->{user_name},
                                           );

        $rv->{return_value}->{policy_version_time} = $add_events[0]->{policy_version_time};
        $rv->{return_value}->{policy_version_device_name} = $add_events[0]->{policy_version_device_name};
        $rv->{return_value}->{policy_version_policy_name} = $add_events[0]->{policy_version_policy_name};
        $rv->{return_value}->{policy_version_last_change} = $add_events[0]->{policy_version_last_change};

        push @$saved_events, @add_events;
    }
    push @$saved_events, @{delete $self->{events} || []};
    $rv->{events} = $saved_events;
    

    if ($self->{device_group} && $session->{user_name} && $session->{user_name} ne $F5::ASMConfig::Handler::TSCONFD_USER_NAME) {
        $call_flags->{update_on_adding_duplicate_record} = 1;
    }

    if ($unset_do_not_relay) {
        $call_flags->{do_not_relay} = 0;
    }

    if ($do_mcp_store) {
        $new_policy->mcp_store();
    }

    $locker->unlock();
    return $rv;
}

#----------------------------------------------------------------------
sub _validate_policy_template {
    my (%params) = @_;
    my $self = $params{self};
    my $action = $params{action};
    my $policy_template = $params{policy_template};
    my $properties = $params{properties} || {};

    if (defined $policy_template) {
        if (!defined $policy_template->{icontrol_name} && !defined $policy_template->{icontrol_enum_order} &&
              !defined $policy_template->{policy_template_id} && !defined $policy_template->{policy_template_name}) {
            return $self->handle_error(action => "$action policy using",
                                       error_code => F5::ASMConfig::ErrorCode::INVALID_POLICY(),
                                       entity_class => "F5::ASMConfig::Entity::PolicyTemplate",
                                       thrift_obj => $policy_template,
                                       error_thrown => "Missing criteria fields",
                                       error_level => "info",
                                      );
        }

        my $policy_template_name = $policy_template->{icontrol_name} || $policy_template->{icontrol_enum_order} ||
                                   $policy_template->{policy_template_id} || $policy_template->{policy_template_name};

        $policy_template = F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $self->{dbh}, 
                                                                      logger => $self->{logger},
                                                                      get_criteria => $policy_template);

        if (!$policy_template) {
            $policy_template->{policy_template_name} = $policy_template_name; # display name

            return $self->handle_error(action => "$action policy using",
                                       error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                       entity_class => "F5::ASMConfig::Entity::PolicyTemplate",
                                       thrift_obj => $policy_template,
                                      );
        }

        if ($properties->{policy_type} && $properties->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY() &&
            $policy_template->{policy_type} &&  $policy_template->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
             return $self->handle_error(action => "$action policy using",
                                       error_code => F5::ASMConfig::ErrorCode::INVALID_POLICY(),
                                       entity_class => "F5::ASMConfig::Entity::PolicyTemplate",
                                       thrift_obj => $policy_template,
                                       error_thrown => "Cannot create a Security Policy from a Parent Policy Template",
                                       error_level => "info",
                                      );
        }
        $params{policy_template}->{icontrol_name} = $policy_template->{icontrol_name}; # for future use in reconfigure
        $params{policy_template}->{policy_template_id} = $policy_template->{policy_template_id}; # for future use in add_policy
        $params{policy_template}->{template_file} = $policy_template->{template_file}; # for future use in import_policy
    }
}

#----------------------------------------------------------------------

sub get_compatible_parent_policies {
    my  (%params) = @_;
    my $self = $params{self};
    my $policy = $params{policy};
    my $session = $params{session};

    my $partition = $policy->{partition_id} || $session->{partition};

    my $local_call_flags = F5::ASMConfig::CallFlags->new();
    my $search_filter = F5::ASMConfig::SearchFilter->new(field_name => "partition_id",
                                                         field_value_list => [ $partition,
                                                                               F5::ASMConfig::Constants::PARTITION_COMMON ],
                                                         field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN());

    push (@{$local_call_flags->{search_filters}}, $search_filter);
    my $policy_criteria = F5::ASMConfig::Policy->new({policy_type => F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()});

    my $rv = $self->base_search(rpc_params => [$session, $local_call_flags, undef, $policy_criteria],
                                entity_class => $entity_class);

    return $rv->{return_value};
}

#----------------------------------------------------------------------

sub update_policy {
    my ($self, $session, $call_flags, $place_holder, $policy_criteria, $store_policy) = @_;

    my @events;
    my $newly_updated_encoding;
    my $set_active_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags);
    if ($call_flags->{is_in_add_policy}){
        $set_active_call_flags->{suppress_events} = 0;
    }

    #Handler should not allow a call to directly update the account_id
    delete $store_policy->{account_id};

    # populate the VS assoc if store_policy has it populated or
    #    if the is_active_policy is defined (since this probably means that we're going to either activate or deactivate the policy)
    if ( !$call_flags->{extra_fields} &&
         (
            (defined $store_policy->{is_active_policy})
               ||
            (defined($store_policy->{extra_virtual_server_assocs}) || defined($store_policy->{extra_virtual_server_assocs_complex}))
         )
       ) {
        $call_flags->{extra_fields} = ['extra_virtual_server_assocs'];
    }
    my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                    logger => $self->{logger},
                                                    get_criteria => $policy_criteria,
                                                    extra_fields => $call_flags->{extra_fields},);

    if (!defined($policy)) {
        return $self->handle_error(action => "update",
                                   entity_class => $entity_class,
                                   thrift_obj  => $policy_criteria,
                                   error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                   error_thrown => "Policy doesn't exist",
                                   error_level  => "info",
                                   call_flags   => $call_flags,
                                  );
    }

    my $do_mcp_store = 0;
    my $unset_do_not_relay = 0;
    if ($self->{rpc_entry_depth} == 1 && $self->{device_group} &&
        !F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                 folder => $policy->mcp_folder(),
                                                 logger => $self->{logger},
                                                 create => $self->{is_relay_handler_process})) {
        if ($call_flags->{skip_mcp} && $self->{is_relay_handler_process}) {
            $call_flags->{skip_mcp} = 0; #turn off in relay
            $do_mcp_store = 1;
        }
        $unset_do_not_relay = 1;
    }

    if (defined($store_policy->{parent_policy_id}) && $store_policy->{parent_policy_id} == 0) {
        $store_policy->{has_parent} = 0;
        delete $store_policy->{parent_policy_id};
    }

    my $parent_policy;
    if ($store_policy->{parent_policy_name} && !defined($store_policy->{parent_policy_id})) {
        $parent_policy = $entity_class->new(dbh => $self->{dbh},
                                            get_criteria => { policy_name => $store_policy->{parent_policy_name} });
        if ($parent_policy && $parent_policy->{policy_id}) {
            $store_policy->{parent_policy_id} = $parent_policy->{policy_id};
        }
    }

    if ($policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => [qw(parent_policy_id has_parent)]) ) {
        #Attach, detach and modify are large db transactions that will require a db lock (to be released only after db commit)
        $self->{transaction_locker}->{$F5::DbUtils::LOCK_FILE} ||= F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
                                                                                               logger => $self->{logger},);
        $self->{transaction_locker}->{$F5::DbUtils::LOCK_FILE}->lock(single_lock => 1);
        #Refresh current policy
        $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                     logger => $self->{logger},
                                                     get_criteria => $policy_criteria,
                                                     extra_fields => $call_flags->{extra_fields},);
        if (!defined($policy)) {
            return $self->handle_error(action => "update",
                                   entity_class => $entity_class,
                                   thrift_obj  => $policy_criteria,
                                   error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                   error_thrown => "Policy doesn't exist",
                                   error_level  => "info",
                                  );
        }
    }

    if( !$parent_policy && ($store_policy->{parent_policy_id} || $policy->{parent_policy_id}) ) {
        my $parent_policy_id = $entity_class->get_new_value(existing_obj => $policy, store_obj => $store_policy, field_name => 'parent_policy_id');
        $parent_policy = $entity_class->new(dbh => $self->{dbh},
                                            get_criteria => { policy_id => $parent_policy_id });
    }

    if ( defined($store_policy->{has_parent}) && !$store_policy->{has_parent} && $policy->{has_parent} ) {
        undef($parent_policy);
        my $parent_policy_settings = $self->get_parent_policy_settings($session, $call_flags);
        if ( $parent_policy_settings && $parent_policy_settings->{flg_is_parent_policy_mandatory} ) {
            return $self->handle_error( action => 'update',
                                        entity_class => $entity_class,
                                        error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                        thrift_obj   => $policy,
                                        error_thrown => "Parent Policy is mandatory and cannot be removed.",
                                        error_level  => 'info',
                                        empty_string_value => $policy->{policy_name},
                                      );
        }
    }

    if ( defined($parent_policy) ) {

        if( $policy->{partition_id} && $parent_policy->{partition_id}
            && ($policy->{partition_id} ne $parent_policy->{partition_id}) &&  ($parent_policy->{partition_id} ne "Common") ) {
            my $action = $call_flags->{is_in_add_policy} ? "add" : "update";
            return $self->handle_error( action => $action,
                                        entity_class => $entity_class,
                                        error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                        thrift_obj   => $policy,
                                        error_thrown => "Security policy can have a Parent only in the same partition or in the /Common partition. $policy->{policy_name} is in $policy->{partition_id}, $parent_policy->{policy_name} is in $parent_policy->{partition_id}.",
                                        error_level  => 'info',
                                        empty_string_value => $policy->{policy_name},
                                      );
        }

        my $different_field;
        if ( $policy->{flg_encoding_selected} && $policy->{encoding} ne $parent_policy->{encoding} ) {
            $different_field = $policy->get_field_display_name_string(field_name => 'encoding', call_flags => $call_flags);
        }
        elsif ( $entity_class->get_new_value(existing_obj => $policy, store_obj => $store_policy, field_name => 'flg_is_protocol_free') != $parent_policy->{flg_is_protocol_free} ) {
            if ( $call_flags->{is_in_import_policy} ) {
                $store_policy->{flg_is_protocol_free} = $parent_policy->{flg_is_protocol_free};
            }
            else {
                $different_field = $policy->get_field_display_name_string(field_name => 'flg_is_protocol_free', call_flags => $call_flags);
            }
        }
        elsif ( $entity_class->get_new_value(existing_obj => $policy, store_obj => $store_policy, field_name => 'flg_is_case_insensitive') != $parent_policy->{flg_is_case_insensitive} ) {
            if ( $call_flags->{is_in_import_policy} ) {
                $store_policy->{flg_is_case_insensitive} = $parent_policy->{flg_is_case_insensitive};
            }
            else {
                $different_field = $policy->get_field_display_name_string(field_name => 'flg_is_case_insensitive', call_flags => $call_flags);
            }
        }

        my $action = $call_flags->{is_in_add_policy} ? "add" : "update";
        if ( defined($different_field) ) {
            return $self->handle_error( action => $action,
                                        entity_class => $entity_class,
                                        error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                        thrift_obj   => $policy,
                                        error_thrown => "Parent and Child policies cannot have different $different_field",
                                        error_level  => 'info',
                                        empty_string_value => $policy->{policy_name},
                                      );
        }

        if ($parent_policy->{child_policy_count} >= $self->{max_child_policies} &&
            $policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => ['parent_policy_id']) ) {

            return $self->handle_error( action => $action,
                                        entity_class => $entity_class,
                                        error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                        thrift_obj   => $policy,
                                        error_thrown => "Parent Policies can only have $self->{max_child_policies} child policies.",
                                        error_level  => 'info',
                                        empty_string_value => $policy->{policy_name},
                                      );
        }

        if (!$policy->{flg_encoding_selected}) {
            $store_policy->{flg_encoding_selected} = 1;
            $store_policy->{encoding} = $parent_policy->{encoding};
        }
    }

    if ( $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() && $policy->{child_policy_count} > 0) {
        my $changed_field;
        if ( $policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => 'flg_is_protocol_free') ) {
            $changed_field = $policy->get_field_display_name_string(field_name => 'flg_is_protocol_free', call_flags => $call_flags);
        }

        if ( $policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => 'flg_is_case_insensitive') ) {
            $changed_field = $policy->get_field_display_name_string(field_name => 'flg_is_case_insensitive', call_flags => $call_flags);
        }
        if ( defined($changed_field) ) {
            my $action = $call_flags->{is_in_add_policy} ? "add" : "update";
            return $self->handle_error( action => $action,
                                        entity_class => $entity_class,
                                        error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                        thrift_obj   => $policy,
                                        error_thrown => "Unable to change $changed_field for a Parent Policy with linked policies",
                                        error_level  => 'info',
                                        empty_string_value => $policy->{policy_name},
                                      );
        }
    }

    if ( defined $store_policy->{flg_is_protocol_free} ){

       my $current_flg_is_protocol_free = $policy->{flg_is_protocol_free};
  
       if ( $store_policy->{flg_is_protocol_free} && $current_flg_is_protocol_free == 0 ){

           if ( $call_flags->{is_in_add_policy} && defined($policy->{policy_template_id}) ) {
               my $policy_template = F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $self->{dbh},
                                                                            get_criteria => { policy_template_id => $policy->{policy_template_id} });
               if ( $policy_template->{template_type} == F5::ASMConfig::TemplateType::TEMPLATE_TYPE_GENERIC) {
                   F5::ASMConfig::Entity::URL->delete_many(dbh => $self->{dbh},
                                                           policy => $policy,
                                                           search_criteria => { app_id => 1 },
                                                           suppress_events => 1);
                   F5::ASMConfig::Entity::WebsocketUrl->delete_many(dbh => $self->{dbh},
                                                                    policy => $policy,
                                                                    search_criteria => { protocol => F5::ASMConfig::WebsocketProtocol::WS_PROTOCOL_WSS() },
                                                                    suppress_events => 1);
               }
           }

           my $can_do_update = $self->can_update_is_protocol_free_flag($session, $call_flags, $policy);

           if (! $can_do_update ){
               my $action = $call_flags->{is_in_add_policy} || $call_flags->{is_in_import_policy} ? "add" : "update";
               my $field_display_string = $policy->get_field_display_name_string(field_name => 'flg_is_protocol_free', call_flags => $call_flags);
               my $error_message = $action eq "add" ? "$field_display_string cannot be disabled to comply with Parent policy,"
                                                    : "$field_display_string cannot be disabled,";
               return $self->handle_error( action => $action,
                                           entity_class => $entity_class,
                                           error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                           error_thrown => "$error_message because there are two URLs with the path differing only on protocol",
                                           error_level  => 'info',
                                         );
           }


           my $iterator = F5::ASMConfig::Entity::URL->retrieve_all( dbh => $self->{dbh},
                                                                   policy => $policy,
                                                                   get_criteria => { app_id => 1 });
           my $url_store = F5::ASMConfig::URL->new(app_id => 0);

           while(my $url = $iterator->get_next){
               my $url_criteria = F5::ASMConfig::URL->new(url_id => $url->{url_id});    
               $self->update_url($session, $call_flags, $policy,$url_criteria,$url_store );
           }

           my $websocket_iterator = F5::ASMConfig::Entity::WebsocketUrl->retrieve_all( dbh => $self->{dbh},
                                                                                       policy => $policy,
                                                                                       get_criteria => { protocol => F5::ASMConfig::WebsocketProtocol::WS_PROTOCOL_WSS()});
           my $websocket_url_store = F5::ASMConfig::WebsocketUrl->new(protocol => F5::ASMConfig::WebsocketProtocol::WS_PROTOCOL_WS());
           while(my $websocket_url = $websocket_iterator->get_next){
               my $websocket_url_criteria = F5::ASMConfig::WebsocketUrl->new(websocket_url_id => $websocket_url->{websocket_url_id});
               $self->update_websocket_url($session, $call_flags, $policy,$websocket_url_criteria,$websocket_url_store );
           }
       }
    }

    if ( $store_policy->{flg_is_case_insensitive} &&
         $policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => 'flg_is_case_insensitive') ) {
        my $retval = handle_case_insensitive_flg_change( self => $self,
                                            call_flags => $call_flags,
                                            policy => $policy,
                                            session => $session );
        if ($retval->{events}) {
            push( @events, @{$retval->{events}} );
        }
    }

    if ($policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => [qw(is_pb_enabled is_learning_enabled)])) {
        $self->update_pbconfig($session, $call_flags, $policy, 
                                F5::ASMConfig::PBConfig->new(flg_enabled => $entity_class->get_new_value(existing_obj => $policy, store_obj => $store_policy, field_name => 'is_pb_enabled'),
                                                             flg_learning => $entity_class->get_new_value(existing_obj => $policy, store_obj => $store_policy, field_name => 'is_learning_enabled'))
                              );
    }

    if ($policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => [qw(encoding flg_encoding_selected)])) {
        $call_flags->{writes_to_mcp} = 1;

        if ($policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => 'encoding')) {
            $newly_updated_encoding = $store_policy->{encoding};
        }
    }

    if (!$call_flags->{remove_nested_struct_on_update} && 
        $store_policy->{extra_virtual_server_assocs} && @{$store_policy->{extra_virtual_server_assocs}}) {
        $store_policy->{is_active_policy} = 1;
    }

    my $set_active_flags;
    my $force_do_set_active;
    if (defined $store_policy->{is_active_policy} &&
          $store_policy->{is_active_policy} != $policy->{is_active_policy}) {
        $set_active_flags = F5::ASMConfig::SetActiveFlags->new($call_flags->{set_active_flags});
        if ($store_policy->{is_active_policy}) {
            $set_active_flags->{activate_policy} = 1;
            $call_flags->{writes_to_mcp} = 1;
            $force_do_set_active = 1;
        }
        else {
            # we're going to de-activate a policy
            #     can't do that if policy has complex vs assoc
            if (defined $policy->{extra_virtual_server_assocs_complex} &&
                scalar @{$policy->{extra_virtual_server_assocs_complex}}) {

                return $self->handle_error(action => "delete",
                                           error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                           error_thrown => "Cannot deactivate a security policy with manually configured virtual server association.",
                                           error_level  => "info",
                                          );
            }

            # we're going to de-activate a policy
            #    need to remove VS assoc to make that work
            # this is localized to REST (bug#529535) 
            #    because deactivation via GUI does not hit that bug
            if ($call_flags->{rest_call}) {
                $store_policy->{extra_virtual_server_assocs} = [];
                $store_policy->{extra_virtual_server_assocs_complex} = [];
            }

            #Do not use the implicit transaction and send an update core at the end of a successful transaction.
            $set_active_flags->{deactivate_policy} = 1;
            $set_active_flags->{send_sync} = 0;
            $set_active_flags->{do_transaction} = 0;
            $set_active_flags->{do_update_core} = 0;
            $set_active_flags->{delayed_update_core} = 1;
            $call_flags->{writes_to_mcp} = 1;
            $call_flags->{deactivate_policy} = 1;
            #Will be set by set_active, not in base_update
            delete $store_policy->{is_active_policy};
        }
    }

    my $detach_from_parent = $policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => ['has_parent']) &&
		             ! $policy->get_new_value(existing_obj => $policy, store_obj => $store_policy, field_name => 'has_parent');

    my $modify_parent = $policy->{parent_policy_id} && $store_policy->{parent_policy_id} &&
                        $policy->{parent_policy_id} != $store_policy->{parent_policy_id};

    my $attach_to_parent = $store_policy->{parent_policy_id} && 
		           $policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => ['parent_policy_id']) && 
                           ! $modify_parent;

    if ($detach_from_parent || $modify_parent || $attach_to_parent) {
        $call_flags->{writes_to_mcp} = 1;	
    }

    if($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY() && ($modify_parent || $detach_from_parent)) {
        # First unattach the original parent
        $policy->{has_parent} = 0;
        _handle_parent_policy_change(self => $self,
                                     call_flags => $call_flags,
                                     policy => $policy,
                                     session => $session,);
    }

    my $rv = $self->base_update(entity_class => $entity_class,
                                rpc_params => [ $session, $call_flags,$place_holder,$policy_criteria, $store_policy ],);
    #Ensure the order of events
    push (@{$self->{events}}, @{$rv->{events}});
    $rv->{events} = [];

    #restore policy which got deleted by base update
    $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                 logger => $self->{logger},
                                                 get_criteria => $policy_criteria,
                                                 extra_fields => $call_flags->{extra_fields},);

    if ($newly_updated_encoding) {
        # If the encoding has changed, charsets need be re-set to the default for the new encoding.
        # Presumably, this can only happen if the policy was initially Auto-detect Language.
        for my $charset_type (values %{F5::ASMConfig::Entity::Charset->full_schema->{CharsetType}}) {
            my $charset_retval = $self->get_default_charset_for_encoding($session, $call_flags,
                                                                         F5::ASMConfig::Charset->new(charset_type => $charset_type),
                                                                         $newly_updated_encoding);

            $self->update_charset($session, $call_flags, $policy, $charset_retval);
        }
    }

    if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY() && ($modify_parent || $attach_to_parent)) {
        _handle_parent_policy_change(self => $self,
                                     call_flags => $call_flags,
                                     policy => $policy,
                                     session => $session,);
    }

    my $refresh_policy = 0;
    my $has_active_changed = 0;
    if ($set_active_flags) {
        $has_active_changed = ((exists $set_active_flags->{deactivate_policy} && $set_active_flags->{deactivate_policy}) ||
                               (exists $set_active_flags->{activate_policy} && $set_active_flags->{activate_policy})) ? 1 : 0 ;
    }
    my @add_assocs = ();
    my @delete_assocs = ();
    my @existing_virtual_server_assocs = ();
    my $was_websec_prof_configured = 0;
    if ( (defined($store_policy->{extra_virtual_server_assocs}) || defined($store_policy->{extra_virtual_server_assocs_complex})) &&
         !($self->{is_relay_handler_process} && $self->{device_group} &&
           F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                   folder => $policy->mcp_folder(),
                                                   logger => $self->{logger},
                                                  ))
       ) {
        @existing_virtual_server_assocs = ( @{$policy->{extra_virtual_server_assocs}}, @{$policy->{extra_virtual_server_assocs_complex}});

        #If adding servers set active first, if removing set active second.
        if ($set_active_flags && !$call_flags->{remove_nested_struct_on_update} && @{$store_policy->{extra_virtual_server_assocs}}) {
            $self->set_active($session, $set_active_call_flags, $policy, $set_active_flags);
            undef $set_active_flags;
            $refresh_policy = 1;
        }
        if ($call_flags->{append_nested_struct_on_update}) {
            @add_assocs = @{$store_policy->{extra_virtual_server_assocs}};
        }
        elsif ($call_flags->{remove_nested_struct_on_update}) {
            @delete_assocs = @{$store_policy->{extra_virtual_server_assocs}};
        }
        else {
            my %existing_virtuals = map { ( $_->{virtual_server}->{full_name} => 1 ) } @existing_virtual_server_assocs;
            my %seen_virtuals;
            for my $assoc (@{$store_policy->{extra_virtual_server_assocs}}) {
                if (my $virtual = F5::ASMConfig::Entity::VirtualServer->new(get_criteria => $assoc->{virtual_server}, dbh => $self->{dbh})) {
                    $seen_virtuals{$virtual->{full_name}}++;
                    next if ($existing_virtuals{$virtual->{full_name}});

                    push (@add_assocs, $assoc);
                    $call_flags->{writes_to_mcp} = 1;
                }
            }
            # the complex assocs from store_policy need to be marked as seen as well
            for my $assoc (@{$policy->get_new_value(
                                 store_obj => $store_policy,
                                 existing_obj => $policy,
                                 field_name => 'extra_virtual_server_assocs_complex',
                                 call_flags => $call_flags,
                             )}) {
                if (my $virtual = F5::ASMConfig::Entity::VirtualServer->new(get_criteria => $assoc->{virtual_server}, dbh => $self->{dbh})) {
                    $seen_virtuals{$virtual->{full_name}}++;
                }
            }
            for my $assoc (@existing_virtual_server_assocs) {
                if (!$seen_virtuals{$assoc->{virtual_server}->{full_name}}) {
                    push (@delete_assocs, $assoc);
                    $call_flags->{writes_to_mcp} = 1; 
                }
            }
        }

        if (@add_assocs) {
            $was_websec_prof_configured = 1;
        }
        for my $assoc (@add_assocs) {
            my $rv = $self->add_virtual_server_assoc($session, $call_flags, $policy, $assoc);
        }
        for my $assoc (@delete_assocs) {
             my $rv = $self->delete_virtual_server_assoc($session, $call_flags, $policy, $assoc);
             if ($rv->{was_websec_prof_configured}) {
                 $was_websec_prof_configured = 1;
             }
        }
        if (@add_assocs || @delete_assocs) {
            $store_policy->{last_vs_assoc_update_time} = time();
        }

        if ($set_active_flags && $policy->{unassociated_l7_policies}) {
             #If attempting to deactivate policy, we need to remove all unassociated l7_policies that reference this asm policy as well
             for my $l7_policy (@{$policy->{unassociated_l7_policies}}) {
                 if ($l7_policy->{type} == F5::ASMConfig::VirtualServerAssocType::VIRTUAL_SERVER_ASSOC_TYPE_L7_COMPLEX ||
                     $l7_policy->{l7_policy_name} !~ ASM_AUTO_L7_POLICY_NAME_PREFIX) {
                     return $self->handle_error(action => "delete",
                                                error_code   => F5::ASMConfig::ErrorCode::MCP_ACTION_NOT_ALLOWED(),
                                                error_thrown => "Cannot remove manually configured L7 policy $l7_policy->{l7_policy_name}.",
                                                error_level  => "info",
                                                );
                 }
                 $l7_policy->delete();
             }
        }
    }

    # if we added/removed a websec profile to/from VS -
    # there will be a tsconfd event fired that will execute ramcache update
    # we tell set_active not to run another instance of ramcache update
    my $update_ramcache = ! $was_websec_prof_configured;

    if ($set_active_flags) {
        $set_active_flags->{update_ramcache} = $update_ramcache;
        $self->set_active($session, $set_active_call_flags, $policy, $set_active_flags);
        $refresh_policy = 1;
    }

    if ($refresh_policy) {
        $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $policy, force_reload => 1);
        $rv->{return_value} = $policy; #Update after set_active call
    }

    # after set_active
    #    see if we need to update flg_is_candidate_for_sweep in store_policy
    _update_flg_is_candidate_for_sweep(policy => $policy,
                                       store_policy => $store_policy,
                                       call_flags => $call_flags,
                                       logger => $self->{logger},
                                       active_changed => $has_active_changed,
                                       add_assocs_count => scalar @add_assocs,
                                       delete_assocs_count => scalar @delete_assocs,
                                       existing_assocs_count => scalar @existing_virtual_server_assocs,
                                      );

    my $suppress_event_fields = [qw(flg_is_candidate_for_sweep last_vs_assoc_update_time)];
    if ($policy->has_value_changed(existing_obj => $policy, store_obj => $store_policy, check_fields => $suppress_event_fields) ) {
        #Update these fields separately as they must not create audit logs or mark the policy as modified if they are the only changes.
        my $suppressed_event_store_obj = F5::ASMConfig::Policy->new();
        for my $suppress_event_field (@$suppress_event_fields) {
            $suppressed_event_store_obj->{$suppress_event_field} = delete $store_policy->{$suppress_event_field};
        }
        $policy->store_update(store_obj => $suppressed_event_store_obj,
                              suppress_events => 1,
                              skip_mcp => 1);

    }

    if (defined $store_policy->{is_modified} && ($store_policy->{is_modified} == 0)) {
       $force_do_set_active = 1;
    }

    if ($force_do_set_active && $policy->is_modified()) {
        if ($policy->{is_active_policy}) {
            $set_active_flags = F5::ASMConfig::SetActiveFlags->new($call_flags->{set_active_flags});
            $set_active_flags->{update_ramcache} = $update_ramcache;
            $self->set_active($session, $set_active_call_flags, $policy, $set_active_flags);
            $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $policy, force_reload => 1);
        }
        else {
            $policy->clear_modification_log();
            $policy->{is_modified} = 0;
        }
        $rv->{return_value} = $policy; #Update after set_active call
    }

    if (defined($store_policy->{extra_virtual_server_assocs}) || defined($store_policy->{extra_virtual_server_assocs_complex})) {
        F5::DbUtils::add_callback(dbh => $self->{dbh},
                                  callback => sub {
                                                    $entity_class->map_fields_to_entity(dbh => $self->{dbh}, rec => $rv->{return_value}, extra_fields => $call_flags->{extra_fields});
                                                  },
                                 );
    }

    if ( $set_active_flags && !$call_flags->{suppress_events}) {
        push( @events,
        $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                            session    => $session,
                            element => $policy,
                            policy =>  $policy,
                            update_thrift_obj => F5::ASMConfig::Policy->new(is_active_policy => $policy->{is_active_policy}),
                            call_flags => $call_flags,
                            user_name  => $session->{user_name},
                            do_policy_version_update => 0,
                           ));
    }

    if ($unset_do_not_relay) {
        $call_flags->{do_not_relay} = 0;
    }

    if ($do_mcp_store) {
        $policy->mcp_store();
    }

    push (@{$rv->{events}},@events);
    return $rv;
}
#----------------------------------------------------------------------

sub _update_flg_is_candidate_for_sweep {
    my (%params) = @_;

    my $logger = $params{logger};
    my $policy = $params{policy};
    my $store_policy = $params{store_policy};
    my $call_flags = $params{call_flags};
    my $active_changed = $params{active_changed} || 0;
    my $add_assocs_count = $params{add_assocs_count} || 0;
    my $delete_assocs_count = $params{delete_assocs_count} || 0;
    my $existing_assocs_count = $params{existing_assocs_count} || 0;

    my $vs_assoc_changed = $add_assocs_count + $delete_assocs_count;

    $logger->debug("Virtual Server association has " . ($vs_assoc_changed ? "" : "not ") . "changed for policy $policy->{policy_name}") if $logger;
    $logger->trace("Policy $policy->{policy_name} had $existing_assocs_count VS assocs, $add_assocs_count added, $delete_assocs_count deleted") if $logger;

    if (!$vs_assoc_changed && !$active_changed) {
        return;
    }

    my $vs_assoc_count = $existing_assocs_count + $add_assocs_count - $delete_assocs_count;
    $logger->debug("There are $vs_assoc_count Virtual Server associations for policy $policy->{policy_name}") if $logger;

    my $is_active = $policy->get_new_value(store_obj => $store_policy,
                                           existing_obj => $policy,
                                           field_name => 'is_active_policy',
                                           call_flags => $call_flags,
                                          );

    $logger->debug("Active state has " . ($active_changed ? "" : "not ") . "changed for policy $policy->{policy_name} - "
                   . ($is_active ? "active" : "inactive")) if $logger;

    my $candidate_for_sweep = 1;
    if ($active_changed && $is_active && !$vs_assoc_count) {
        $candidate_for_sweep = 0;
    }

    $store_policy->{flg_is_candidate_for_sweep} = $candidate_for_sweep;
    $logger->info("Policy $policy->{policy_name} is " . ($candidate_for_sweep ? "" : "not ") . "a candidate for active policy sweeper") if $logger;
}
#----------------------------------------------------------------------

sub handle_case_insensitive_flg_change {
    my (%params) = @_;
    my $self = $params{self};
    my $call_flags = $params{call_flags};
    my $policy = $params{policy};
    my $session = $params{session};

    my @entities_with_policy = $policy->get_policy_entities($call_flags, $policy);
    for my $entity (@entities_with_policy) {
        my @case_insensitive_fields = @{$entity->get_case_insensitive_fields};

        for my $case_insensitive_field (@case_insensitive_fields) {
            if ( $case_insensitive_field->{is_unique_record_part} ) {
                my $db_table = $case_insensitive_field->{db_table};
                my $db_field_name = $case_insensitive_field->{db_field_name};
                my $extern_db_field_name = $case_insensitive_field->{extern_db_field_name};
                my $dbh = $self->{dbh};
                my $join_tables_for_count = '';
                my @join_tables_for_count_bind_values;
                my $join_tables = defined($extern_db_field_name);
                my $case_insensitive_db_field = $join_tables ? $extern_db_field_name : $db_field_name;

                my $prefix_to_unique_record_parts;
                my @extern_prefixes = $join_tables ? $entity->extern_db_prefixes(%params) : ['main_table'];
                if ($join_tables) {
                   my $join_tables_for_count_rec = $entity->get_retrieve_join_tables( policy => $policy );
                   if (ref($join_tables_for_count_rec)) {
                       $join_tables_for_count = $join_tables_for_count_rec->{query};
                       @join_tables_for_count_bind_values = @{$join_tables_for_count_rec->{bind_values}};
                   }
                   else {
                       $join_tables_for_count = $join_tables_for_count_rec;
                   }

                   my @unique_record_parts;
                   for my $extern_prefix (@extern_prefixes) {
                       $prefix_to_unique_record_parts->{$extern_prefix} = \@{$entity->extern_unique_record_part_db_fields( get_criteria => { db_prefix => $extern_prefix } )};
                   }
                }
                else {
                    $prefix_to_unique_record_parts->{'main_table'} = \@{$case_insensitive_field->{unique_record_parts}};
                }

                my @count_bind_values = (@join_tables_for_count_bind_values, $policy->{policy_id});
                for my $prefix (keys %$prefix_to_unique_record_parts) {
                    my @unique_record_parts_for_count = grep {$_ ne $case_insensitive_db_field} @{$prefix_to_unique_record_parts->{$prefix}};
                    @unique_record_parts_for_count = map {"$prefix.$_"} @unique_record_parts_for_count;
                    my $unique_part = "$prefix.$case_insensitive_db_field";
                    if (!F5::DbUtils::is_sqlite($dbh)) {
                        $unique_part = "CONVERT($unique_part USING latin1)";
                    }
                    push(@unique_record_parts_for_count, "LOWER($unique_part)");
                    local $" = ',';

                    my @entities = @{$dbh->selectcol_arrayref(<<EndQuery,undef,@count_bind_values)};
SELECT $prefix.$case_insensitive_db_field
FROM $db_table AS main_table
$join_tables_for_count
WHERE main_table.policy_id = ? AND $prefix.$case_insensitive_db_field IS NOT NULL
GROUP BY @unique_record_parts_for_count
HAVING count(*) >= 2
EndQuery

                    if (@entities){
                        my $action = $call_flags->{is_in_add_policy} || $call_flags->{is_in_import_policy} ? "add" : "update";
                        my $field_display_string = $policy->get_field_display_name_string(field_name => 'flg_is_case_insensitive', call_flags => $call_flags);
                        my $print_entity_class = $entity->element_type_name_plural;
                        my $print_duplicates = join(', ', @entities);
                        my $error_message = $action eq "add" ? "$field_display_string cannot be disabled to comply with Parent policy,"
                                                             : "$field_display_string cannot be disabled,";
                        $error_message = $call_flags->{is_in_add_policy} ? "$error_message because the selected template contains two $print_entity_class with the same name differing only in case"
                                                                         : "$error_message because there are two $print_entity_class with the same name differing only in case";
                        return $self->handle_error( action => $action,
                                                    error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                                    thrift_obj   => $policy,
                                                    error_thrown => $error_message,
                                                    error_level  => 'info',
                                                    empty_string_value => $policy->{policy_name},
                                                  );
                    }
                }
            }
        }
    }

    my $retval;
    my $suppress_events = $call_flags->{is_in_add_policy};
    for my $entity (@entities_with_policy) {
        my @case_insensitive_fields_names = map { $_->{field_name} } @{$entity->get_case_insensitive_fields};
        next unless @case_insensitive_fields_names;
        my $iterator = $entity->retrieve_all(dbh => $self->{dbh},
                                             policy => $policy,
                                             selected_fields => \@case_insensitive_fields_names,
                                             search_filters => $entity->case_insensitive_search_filters(),
                                            );
        while (my $case_insensitive_entity = $iterator->get_next()) {
            my %field_name_to_value_lc;
            for my $case_insensitive_field_name (@case_insensitive_fields_names) {
                next unless defined $case_insensitive_entity->{$case_insensitive_field_name};
                my $case_insensitive_field_value = $case_insensitive_entity->{$case_insensitive_field_name};
                my $case_insensitive_field_value_lc = lc $case_insensitive_field_value;
                $field_name_to_value_lc{$case_insensitive_field_name} = $case_insensitive_field_value_lc;
            }
            my $store_entity = $entity->new(object => {%field_name_to_value_lc});
            my $call_flags_new = F5::ASMConfig::CallFlags->new(suppress_events => $suppress_events);
            my $rv = $self->base_update(entity_class => $entity,
                                        rpc_params => [ $session, $call_flags_new, $policy, $case_insensitive_entity, $store_entity ],);
            if ($rv->{events}) {
                push(@{$retval->{events}}, @{$rv->{events}});
            }
        }
    }

    return $retval;
}

#----------------------------------------------------------------------

sub _handle_parent_policy_change {
    my (%params) = @_;
    my $self = $params{self};
    my $call_flags = $params{call_flags};
    my $policy = $params{policy};
    my $session = $params{session};

    my $search_filter = F5::ASMConfig::SearchFilter->new(field_name => 'parent_inheritance_status',
                                                         field_value_list => [F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL(),
                                                                              F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY(),
                                                                              ],
                                                         field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN());

    my $suppress_events = $call_flags->{is_in_add_policy} || $call_flags->{reinherit} || 0;

    if ($policy->{has_parent}) {
        push (@{$search_filter->{field_value_list}}, F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NONE());
        my @parent_sections = F5::ASMConfig::Entity::Section->retrieve_all(dbh => $self->{dbh},
                                                                           search_filters => [$search_filter],
                                                                           policy => F5::ASMConfig::Policy->new(policy_id => $policy->{parent_policy_id}),
                                                                           get_criteria => {});
        for my $parent_section (@parent_sections) {
            my $child_inheritance_status = F5::ASMConfig::Entity::Section->get_child_inheritance_status(parent_inheritance_status => $parent_section->{parent_inheritance_status},
                                                                                                        call_flags => $call_flags,);
            
            my $store_section = F5::ASMConfig::Section->new(
                                  parent_inheritance_status => $parent_section->{parent_inheritance_status},
                                  child_inheritance_status => $child_inheritance_status,
                                  is_read => 0);

            if ($call_flags->{retain_inheritance_settings}) {
                delete $store_section->{is_read};
            }

            $self->update_section($session, 
                                  F5::ASMConfig::CallFlags->new(inherited_call => 1,
                                                                reinherit => $call_flags->{reinherit},
                                                                is_in_add_policy => $call_flags->{is_in_add_policy},
                                                                is_in_import_policy => $call_flags->{is_in_import_policy},
                                                                suppress_events => $suppress_events,), 
                                  $policy, 
                                  F5::ASMConfig::Section->new(section_type => $parent_section->{section_type}), 
                                  $store_section);
        }
    }
    else {
        my $section_call_flags = F5::ASMConfig::CallFlags->new(inherited_call => 1,
                                                               delete_disinherited_entities => $call_flags->{delete_disinherited_entities},
                                                               suppress_events => $suppress_events,
                                                               search_filters => [$search_filter],);
        $self->update_many_sections($session,
                                    $section_call_flags,
                                    $policy,
                                    F5::ASMConfig::Section->new(),
                                    F5::ASMConfig::Section->new(parent_inheritance_status => F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NONE(),
                                                                child_inheritance_status  => F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_NA(),));
    }
}

#----------------------------------------------------------------------

sub reconfigure_policy {
    my ($self, $session, $call_flags, $place_holder, $thrift_policy) = @_;

    my $action = "reconfigure";

    if (!defined $thrift_policy->{policy_name} ||
          !defined $thrift_policy->{policy_template} && !defined $thrift_policy->{flg_encoding_selected}) {
        return $self->handle_error(action => $action,
                                   error_code => F5::ASMConfig::ErrorCode::INVALID_POLICY(),
                                   entity_class => $entity_class,
                                   thrift_obj => $thrift_policy,
                                   error_thrown => "Missing mandatory fields for change: policy name and requested encoding and/or template",
                                   error_level => "info",
                                  );
    }

    _validate_policy_template(self => $self,
                              action => $action,
                              policy_template => $thrift_policy->{policy_template});

    my $lock_file = $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                             logger => $self->{logger},);

    my $policy; # existing or added one

    eval {
        $locker->lock();

        # Complex function - update MCP once at the end
        my $saved_skip_mcp = $call_flags->{skip_mcp};
        $call_flags->{skip_mcp} = 1;

        $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh},
                                                          policy => {policy_name => $thrift_policy->{policy_name}});

        if ($policy->{_failed_to_load}) {
            $locker->unlock();
            return $self->handle_error(action => $action,
                                       error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                       entity_class => $entity_class,
                                       thrift_obj => $thrift_policy,
                                      );
        }

        # NOTE: should be checked before possible return at no changes
        if ((!defined $thrift_policy->{flg_encoding_selected} ||
              ($thrift_policy->{flg_encoding_selected} && $policy->{flg_encoding_selected} && $thrift_policy->{encoding} eq $policy->{encoding}) ||
              (!$thrift_policy->{flg_encoding_selected} && !$policy->{flg_encoding_selected} && !$policy->{is_pb_enabled})) &&
              (!defined $thrift_policy->{policy_template} || 
              (defined $policy->{policy_template} && defined $policy->{policy_template}->{icontrol_name} &&
              $thrift_policy->{policy_template}->{icontrol_name} eq $policy->{policy_template}->{icontrol_name}))) {
            $self->{logger}->debug("Policy is correctly configured, no need to make changes");
            $locker->unlock();
            return;
        }

        # Preserve encoding for new policy from template
        if (!defined $thrift_policy->{flg_encoding_selected} && $policy->{flg_encoding_selected}) {
            $thrift_policy->{encoding} = $policy->{encoding};
            $thrift_policy->{flg_encoding_selected} = 1;
        }

        # Preserve active state
        $thrift_policy->{is_active_policy} = $policy->{is_active_policy} unless defined $thrift_policy->{is_active_policy};

        # Deactivate silently (in delete) before activating (in add)
        my $delete_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags);

        if ($thrift_policy->{is_active_policy}) {
            $delete_call_flags->{set_active_flags}->{do_update_core} = 0;
        }

        # Delete the former policy
        $self->delete_policy($session, $delete_call_flags, $policy);

        # Add a new policy with all specified fields, incl. template and encoding (overriding one from template file)
        $policy = $self->add_policy($session, $call_flags, $thrift_policy);

        $call_flags->{skip_mcp} = $saved_skip_mcp;
        if ($self->{rpc_entry_depth} == 1 && $self->{device_group} &&
            !F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                     folder => $policy->mcp_folder(),
                                                     logger => $self->{logger},
                                                     create => $self->{is_relay_handler_process})) {
            if ($call_flags->{skip_mcp} && $self->{is_relay_handler_process}) {
                $call_flags->{skip_mcp} = 0; #turn off in relay
            }
            $call_flags->{do_not_relay} = 0;
        }

        $policy->mcp_store(call_flags => $call_flags);
    };

    my $err = $@;
    $locker->unlock();

    if ($err) {
        die $err;
    }

    return { return_value => $policy };
}

#----------------------------------------------------------------------

sub is_policy_modified{

    my($self,$session, $call_flags, $place_holder, $thrift_policy) = @_;

    my $policy_id = $thrift_policy->{policy_id};
    my $dbh = $self->{dbh};

    my $is_modified =  $dbh->selectrow_arrayref("SELECT MAX(update_time) FROM $self->{plc_dbname}.PL_POLICY_CHANGES_LOG WHERE policy_id = ?",
                                              undef, $policy_id)->[0] || 0; 

   my $retval = { return_value => $is_modified,
               };
 
   return $retval;

}

#----------------------------------------------------------------------

sub import_policy{
  my($self, $session, $call_flags,$place_holder,$import_export_properties) = @_;

  my $dbh = $self->{dbh};
  my $action = "import";

  if ($import_export_properties->{parent_policy}) {
      my $parent_policy = F5::ASMConfig::Entity::Policy->new(dbh => $dbh,
                                                             get_criteria => $import_export_properties->{parent_policy});

      if (!$parent_policy) {
          return $self->handle_error(action => $action,
                                     entity_class => $entity_class,
                                     thrift_obj  => $parent_policy,
                                     error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                     error_thrown => "Policy doesn't exist",
                                     error_level  => "info",
                                    );
      }

      if ($parent_policy->{policy_type} && $parent_policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY()) {
          return $self->handle_error(action => $action,
                                     entity_class => $entity_class,
                                     thrift_obj  => $parent_policy,
                                     error_code => F5::ASMConfig::ErrorCode::INVALID_POLICY(),
                                     error_thrown => "Cannot set Security Policy '$parent_policy->{policy_name}' as a Parent Policy",
                                     error_level => "info",
                                    );
      }

      $import_export_properties->{parent_policy}->{policy_name} = $parent_policy->{policy_name};
      $import_export_properties->{parent_policy}->{policy_id} = $parent_policy->{policy_id};
  }

  my $extra_lock_on_import = ($self->{is_relay_handler_process} && 
                              $import_export_properties->{force_policy_name_flg} &&
                              $import_export_properties->{policy_name}) ? 1 : 0 ;

  my $lock_file = $F5::DbUtils::LOCK_FILE;
  my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                           logger => $self->{logger},);
  $locker->lock() if $extra_lock_on_import;

  my $import_export = F5::ImportExportPolicy->new(dbh => $dbh,
                                                  do_validation => $import_export_properties->{do_validation},
                                                  enforce_strictness => $import_export_properties->{enforce_strictness});

  if ($import_export_properties->{force_policy_name_flg}) {
      my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $dbh,
                                                      get_criteria => { policy_name => $import_export_properties->{policy_name}});
      if ($policy) {   
          $import_export_properties->{force_policy_id} = $policy->{policy_id};
      }
  }

  my $force_policy_id = $import_export_properties->{force_policy_id} ? $import_export_properties->{force_policy_id} : 0;
  my $overwrite = $force_policy_id ? 1 : 0;
  my $parent_policy_id;
  my %old_parent_policy_sections;

  if ($force_policy_id) {
      my $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh,
                                                           policy => {policy_id => $force_policy_id});

      if ($policy->{_failed_to_load}) {
          $force_policy_id = $overwrite = 0;
      }
      else {
          $self->check_authz(authz_objects => [$policy], function_name => "import_policy", write_access => 1,);
          $parent_policy_id = $policy->{parent_policy_id};

          if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() && $policy->{child_policy_count}) {
              my @sections = F5::ASMConfig::Entity::Section->retrieve_all(policy => $policy, dbh => $self->{dbh}, get_criteria => {});
              %old_parent_policy_sections = map { ($_->{section_type} => $_) } @sections;
          }
      }
  }
  else {
      $self->check_authz(authz_objects => [$import_export_properties], function_name => "import_policy", write_access => 1);  #Looks at policy_name or session's partition
  }

  # Replace filename with template filename if set
  my $filename = $import_export_properties->{filename};
  
  my $template = F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $self->{dbh}, get_criteria => $import_export_properties->{policy_template});
  if (defined ($template->{template_file})) {
      _validate_policy_template(self => $self,
                                properties => $import_export_properties,
                                action => $action,
                                policy_template => $template);
      $filename = $template->{template_file};
  }

  my $partition = $import_export_properties->{partition} || $session->{partition};
  my $creator_name = $session->{gui_user_name} || $session->{user_name} || 'SYSTEM';
  
  my $encoding = $import_export_properties->{flg_encoding_selected} ? $import_export_properties->{encoding} : '';
  #This flag indicates if we are in an Active/Active relayed policy import and automatic learning should be turned off locally.
  my $is_relayed_import = ($self->{is_relay_handler_process} &&
                           !$call_flags->{originator_is_failover_standby} &&
                           !F5::Cluster::Utils::is_failover_standby() &&
                           F5::DeviceGroup::is_device_active($call_flags->{originator}));

  my $rv  =   $import_export->import_policy(
                                            filename            => $filename,
                                            user_name           => $call_flags->{user_name},
                                            force               => $import_export_properties->{allow_unsupported_versions},
                                            overwrite           => $overwrite,
                                            force_policy_id     => $force_policy_id,
                                            policy_name         => $import_export_properties->{policy_name},
                                            force_policy_name_flg => $import_export_properties->{force_policy_name_flg},
                                            is_relayed_import   => $is_relayed_import,
                                            get_policy_attributes_only => $import_export_properties->{get_policy_attributes_only},
                                            partition           => $partition,
                                            asm_config_handler  => $self,
                                            skip_mcp            => $call_flags->{skip_mcp},
                                            encoding            => $encoding,
                                            allow_encoding_change => $import_export_properties->{allow_encoding_change},
                                            flg_encoding_selected => $import_export_properties->{flg_encoding_selected},
                                            policy_type => $import_export_properties->{policy_type},
                                            policy_description => $import_export_properties->{policy_description},
                                            creator_name => $creator_name,
                                            populate_rest_uuid => $call_flags->{populate_rest_uuid},
                                            parent_policy => $import_export_properties->{parent_policy},
                                            resolve_external_references => $import_export_properties->{resolve_external_references},
                                            content => $import_export_properties->{content},
                                            file_path_map => $import_export_properties->{file_path_map},
                                           );

  if ( $rv->{sync_files} ) {
      $call_flags->{sync_files} = $rv->{sync_files};
  }

  my $return_value = F5::ASMConfig::ImportPolicyRes->new($rv);

  if ($rv->{encoding_changed} ) {
      F5::ASMConfig::Entity::Policy::charset_config($rv->{policy_id}, {}, $dbh);
  }


  my @errors;
  if ($rv->{msgs}) { 
      for my $msg (@{$rv->{msgs}}) {
          my $err_msg = F5::ASMConfig::Message->new();
          $err_msg->{msg} = $msg->{msg_text};
          $err_msg->{level} = $msg->{msg_level};
          push @errors, $err_msg;
      }
  }
  $return_value->{msgs} = \@errors;

  if ($import_export_properties->{get_policy_attributes_only}) {
      $call_flags->{do_not_relay} = 1;
      $locker->unlock() if $extra_lock_on_import;
      return { return_value => $return_value };
  }

  my @events;

  if ($rv->{completed_successfully}) {
      my $ret_policy = F5::ASMConfig::Policy->new();
      $ret_policy->{policy_id}  = $rv->{policy_id};

      $return_value->{policy} = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                                   logger => $self->{logger},
                                                                   get_criteria => $ret_policy);


      if ( $self->{session} && defined $self->{session}->{has_admin_access} && !$self->{session}->{has_admin_access}
              && $return_value->{policy}->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() ) {
          return $self->handle_error(action => $action,
                                     entity_class => $entity_class,
                                     thrift_obj  => $return_value->{policy},
                                     error_code   => F5::ASMConfig::ErrorCode::PERMISSION_DENIED(),
                                     error_thrown => "User $session->{gui_user_name} does not have permission to call $action on a Parent Policy",
                                     error_level  => "info",
                                    );
      }

      $return_value->{policy}->clean_policy_for_inheritance();

      my $update_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags, 
                                                            suppress_events => 1, 
                                                            reinherit => 1,
                                                            is_in_import_policy => 1,
                                                            retain_inheritance_settings => $import_export_properties->{retain_inheritance_settings},
                                                            );

      if ( $return_value->{policy}->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY() ) {
          if (!$return_value->{policy}->{parent_policy_id}) {
              if ($import_export_properties->{parent_policy} || $parent_policy_id) {
                  $parent_policy_id ||= $import_export_properties->{parent_policy}->{policy_id};
                  #Imported child policy is always NONE/NA          
                  $return_value->{policy} = $self->update_policy($session, $update_call_flags, F5::ASMConfig::Policy->new(policy_id => $rv->{policy_id}),
                                                                                               F5::ASMConfig::Policy->new(parent_policy_id => $parent_policy_id));
              }
              else {
                  my $parent_policy_settings = $self->get_parent_policy_settings($session, $call_flags);
                  if ( $parent_policy_settings && $parent_policy_settings->{flg_is_parent_policy_mandatory} ) {
                      my $parent_policy_rv = get_compatible_parent_policies(self => $self,
                                                                            session => $session,
                                                                            policy => $return_value->{policy});
                      if ($parent_policy_rv->{total_record_count}) {
                          return $self->handle_error( action => $action,
                                                      entity_class => $entity_class,
                                                      error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                                      thrift_obj   => $ret_policy,
                                                      error_thrown => 'Parent Policy is mandatory',
                                                      error_level  => 'info',
                                                      empty_string_value => $ret_policy->{policy_name},
                                                    );
                      }
                  }
                  _handle_parent_policy_change(self => $self,
                                               call_flags => $update_call_flags,
                                               policy => $return_value->{policy},
                                               session => $session,);
              }
          }
          else {
              _handle_parent_policy_change(self => $self,
                                           call_flags => $update_call_flags,
                                           policy => $return_value->{policy},
                                           session => $session,);
          }
      }

      if (%old_parent_policy_sections) {
          #We have replaced a parent policy that has children. This policy may be entirely different so all the children need to fully reinherit
          # * -> M: Mark NONE, then MANDATORY.  This forces all children to reinherit
          # * -> N: Update all children to current parent status. Children that used to inherit will now stop
          # M -> O: Mark NONE, Mark MANDATORY, Mark OPTIONAL
          # N -> O: Mark OPTIONAL
          # O -> O: Take all Accepted children, Mark NONE, Mark MANDATORY, Mark OPTIONAL
          my @new_sections = F5::ASMConfig::Entity::Section->retrieve_all(policy => $return_value->{policy}, dbh => $self->{dbh}, get_criteria => {});
          my @child_policy_ids = @{$return_value->{policy}->child_policy_ids()};
          for my $section (@new_sections) {
              my $old_section = $old_parent_policy_sections{$section->{section_type}};
              next if ($old_section->{parent_inheritance_status} == $section->{parent_inheritance_status} &&
                       ($section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NA() ||
                        $section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NONE()));

              my $reinherit_search_filter = F5::ASMConfig::SearchFilter->new(field_name => 'policy_id',
                                                                             field_value_list => \@child_policy_ids,
                                                                             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN());

              my $reinherit_accepted_sections = ($section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL()
                                                 && $old_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL());
              if ($reinherit_accepted_sections) {
                  #Save accepted child_ids
                  my @accepting_sections = F5::ASMConfig::Entity::Section->retrieve_all(dbh => $self->{dbh},
                                               search_filters => [F5::ASMConfig::SearchFilter->new( field_name => 'section_type',
                                                                                                    field_value => $section->{section_type},
                                                                                                    field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),),
                                                                  F5::ASMConfig::SearchFilter->new( field_name => 'child_inheritance_status',
                                                                                                    field_value => F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_ACCEPTED(),
                                                                                                    field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),),
                                                                  $reinherit_search_filter,
                                                                 ],
                                               selected_fields => [],
                                               minimal_load => 1,
                                              );

                  next unless @accepting_sections;
                  #Perform Force Disinherit, Force Inherit, Update to Optional for all policies that had accepted this section.
                  my @accepting_child_policy_ids = map { $_->{policy_id} } @accepting_sections;
                  $reinherit_search_filter->{field_value_list} = \@accepting_child_policy_ids;
              }

              my $delete_disinherited_entities = ($section->{parent_inheritance_status} != F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NONE());
              my $reinherit_call_flags = F5::ASMConfig::CallFlags->new(inherited_call => 1,
                                                                       delete_disinherited_entities => $delete_disinherited_entities,
                                                                       search_filters => [$reinherit_search_filter]);

              if ($section->{parent_inheritance_status} != F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL()
                 || $old_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY()
                 || $reinherit_accepted_sections) {
                  #Update children to be NONE/NA to force disinherit
                  my $update_parent_inheritance_status = ($section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NA()) 
                                                        ? F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NA()
                                                        : F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NONE();
                  $self->update_many_sections($session, $reinherit_call_flags, undef, 
                                              F5::ASMConfig::Section->new(section_type => $section->{section_type}),
                                              F5::ASMConfig::Section->new(parent_inheritance_status => $update_parent_inheritance_status,
                                                                          child_inheritance_status => F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_NA()));
              }
              if ($section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY()
                 || ($old_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY()
                   && $section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL()) 
                 || $reinherit_accepted_sections) {
                  #Mark children as Mandatory to force inherit
                  $self->update_many_sections($session, $reinherit_call_flags, undef,
                                              F5::ASMConfig::Section->new(section_type => $section->{section_type}),
                                              F5::ASMConfig::Section->new(parent_inheritance_status => F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY(),
                                                                          child_inheritance_status => F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_MANDATORY()));
              }
              if ($section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL()) {
                  $self->update_many_sections($session, $reinherit_call_flags, undef,
                                              F5::ASMConfig::Section->new(section_type => $section->{section_type}),
                                              F5::ASMConfig::Section->new(parent_inheritance_status => F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_OPTIONAL()));
              }
          }
      }

      if ($self->{rpc_entry_depth} == 1 && $self->{device_group} &&
          !F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                    folder => $return_value->{policy}->mcp_folder(),
                                                    logger => $self->{logger},
                                                    create => $self->{is_relay_handler_process})) {
          if ($call_flags->{skip_mcp} && $self->{is_relay_handler_process}) {
              $call_flags->{skip_mcp} = 0; #turn off in relay
          }
          $call_flags->{do_not_relay} = 0;
      }

      $return_value->{policy}->mcp_store(call_flags => $call_flags); #Sync to MCP;

      if ($force_policy_id) {
          push(@events,
               $self->create_event(event_type => F5::ASMConfig::EventType::DELETE,
                                   session => $session,
                                   policy => $return_value->{policy},
                                   element => $return_value->{policy},
                                   call_flags => $call_flags,
                                   session => $session,
                                  ));
      }

      push(@events,
           $self->create_event(event_type => F5::ASMConfig::EventType::ADD,
                               session => $session,
                               policy => $return_value->{policy},
                               element => $return_value->{policy},
                               call_flags => $call_flags,
                               session => $session,
                               policy_version_time => $rv->{policy_version_time},
                               policy_version_device_name => $rv->{policy_version_device_name},
                               policy_version_policy_name => $rv->{policy_version_policy_name},
                               policy_version_last_change => $rv->{policy_version_last_change},
                              ));

      if ($self->{device_group}) {
          my $cloned_import_export_properties = dclone($import_export_properties);
          $cloned_import_export_properties->{force_policy_name_flg} = 1;
          $cloned_import_export_properties->{policy_name} = $return_value->{policy}->{policy_name};
          push (@{$self->{rpc_calls_to_relay}->{import_policy}}, {function_name => 'import_policy',
                                                                  rpc_params    => [$session, $call_flags, undef, $cloned_import_export_properties],
                                                                 });
      }
  }
  else {
      $call_flags->{do_not_relay} = 1;
  }

  $locker->unlock() if $extra_lock_on_import;
  my $retval = { return_value => $return_value, events => \@events, };
  return $retval;
}

#----------------------------------------------------------------------
sub export_policy{
  my($self, $session, $call_flags,$place_holder, $policy,$import_export_properties) = @_;

  my $dbh = $self->{dbh};

  my $import_export = F5::ImportExportPolicy->new(dbh => $dbh,
                                                  enforce_strictness => $import_export_properties->{enforce_strictness});

  $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                               logger => $self->{logger},
                                               get_criteria => $policy);
     
  if (!defined($policy)) {
      return $self->handle_error(action => "export",
                                 entity_class => $entity_class,
                                 thrift_obj  => $policy,
                                 error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                 error_thrown => "Policy doesn't exist",
                                 error_level  => "info",
                                );
  }

  my $rv = $import_export->export_policy(
                                         policy_id           => $policy->{policy_id},
                                         policy => $policy,
                                         asm_config_handler  => $self,
                                         %$import_export_properties,
                                        );

  $import_export_properties->{file_size} = -s $import_export_properties->{filename};

  my @warnings;
  if ($rv->{msgs}) {
      for my $msg (@{$rv->{msgs}}) {
          my $wrn_msg = F5::ASMConfig::Message->new();
          $wrn_msg->{msg} = $msg->{msg_text};
          $wrn_msg->{level} = $msg->{msg_level};
          push @warnings, $wrn_msg;
      }
  }

  $import_export_properties->{msgs} = \@warnings;


  if ($rv->{completed_successfully}) {
      return { return_value => F5::ASMConfig::ImportExportProperties->new({
                                filename  => $import_export_properties->{filename},
                                file_size => $import_export_properties->{file_size},
                                msgs      => $import_export_properties->{msgs},
                               }) };
  }
  else {
      my $error_msg = join(' ', map { $_->{msg_text} } @{$rv->{msgs}});
      return $self->handle_error(action => "export",
                                 entity_class => $entity_class,
                                 thrift_obj  => $policy,
                                 error_code   => F5::ASMConfig::ErrorCode::POLICY_EXPORT_ERROR(),
                                 error_thrown => $error_msg,
                                 error_level  => "info",
                                );
  }
}

#----------------------------------------------------------------------

sub delete_policy{
    my ($self, $session, $call_flags, $place_holder, $thrift_policy) = @_;

    my $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $thrift_policy);

    # the reason why we need to lock here is that tsconfd will call this method directly
    #   and without a lock we are bound to go into a deadlock because we are deleting
    #   from a bunch on tables in F5::ASMConfig::Entity::Policy::empty_policy_tables
    # that combined with a delete_policy call from another client, say REST, leads to a deadlock
    my $lock_file = $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                             logger => $self->{logger},);
    $locker->lock();

    # Deactivate active policy before actual deleting, to also delete account
    if ($policy->{is_active_policy}) {
        $policy->{is_active_policy} = 0;
        $self->update_policy($session, $call_flags, $policy);
    }

    if ($policy->{has_parent}) {
        #Clean suggestions for parent policies
        $self->delete_many_suggestions($session, 
                                       F5::ASMConfig::CallFlags->new(suppress_events => 1), 
                                       $policy, 
                                       F5::ASMConfig::Suggestion->new(inheritance_acceptability => F5::ASMConfig::InheritanceAcceptability::INHERITANCE_ACCEPTABILITY_LOCKED()));
    }

    if ($self->{rpc_entry_depth} == 1 && $self->{device_group} &&
        !F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                 folder => $policy->mcp_folder(),
                                                 logger => $self->{logger},)) {
        if ($call_flags->{skip_mcp} && $self->{is_relay_handler_process}) {
            $call_flags->{skip_mcp} = 0; #turn off in relay
        }
        $call_flags->{do_not_relay} = 0;
    }

    my $rv = $self->base_delete(entity_class => $entity_class,
                                rpc_params => [ $session, $call_flags,$place_holder,$thrift_policy ],
                               );

    if ($self->{device_group}) {
        push @{$call_flags->{ignored_errors}}, F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD()});
    }

    $locker->unlock();

    return $rv;
}

#----------------------------------------------------------------------

sub can_update_is_protocol_free_flag{
   my ($self, $session, $call_flags, $place_holder, $thrift_policy) = @_;  

   my $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $thrift_policy);
   my $retval = $policy->can_update_protocol_free();

   return {return_value => $retval };

}

#----------------------------------------------------------------------

sub replace_policy {
    my ($self, $session, $call_flags, $place_holder, $old_policy_thrift, $new_policy_thrift) = @_;
    my $old_policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh}, get_criteria => $old_policy_thrift, extra_fields => ['extra_virtual_server_assocs']);
    my $new_policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh}, get_criteria => $new_policy_thrift);

    if (!$old_policy  || !$new_policy) {
       return $self->handle_error(action => "replace",
                                  entity_class => $entity_class,
                                  thrift_obj  => ($new_policy ? $old_policy_thrift : $new_policy_thrift),
                                  error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                  error_thrown => "Policy doesn't exist",
                                  error_level  => "info",
                                 );
    }

    if ( !($self->{is_relay_handler_process} && $self->{device_group} &&
           F5::DeviceGroup::is_folder_synchronized(device_group => $self->{device_group},
                                                   folder => $old_policy->mcp_folder(),
                                                   logger => $self->{logger},
                                                  ))
       ) {
        for my $policy_action (@{F5::ASM::COAPI::Query->new("policy_action", {target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM,
                                                                              action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE,
                                                                              policy => $old_policy->{policy_name}})}) {

            my $original_l7_policy_status;
            my $l7_policy_name = $policy_action->{policy_name};
            my $l7_policy = F5::ASMConfig::Entity::L7Policy->new(get_criteria => { l7_policy_name => $l7_policy_name} , dbh => $self->{dbh} );
            if ($l7_policy->{status} != F5::Mcp::PolicyStatus::POLICY_STATUS_LEGACY) {
                $original_l7_policy_status = $l7_policy->{status};
                $l7_policy->store_update(store_obj => {as_legacy => 1,
                                                       status => F5::Mcp::PolicyStatus::POLICY_STATUS_LEGACY,
                                                      }
                                        );
            }

            $policy_action->{policy} = $new_policy->{policy_name};
            $policy_action->transactional_action(action => 'save', dbh => $self->{dbh});

            # after the l7 policy related, transactional action - return it's original status, if required
            if ($original_l7_policy_status) {
                my $internal_config_item = $self->get_internal_config_items(config_item_name => 'force_legacy_ltm_policy');
                if (!$internal_config_item->{current_value}) {
                    $l7_policy->store_update(store_obj => {as_legacy => 0,
                                                           status => $original_l7_policy_status,
                                                          }
                                            );
                }
            }
        }
    }

    $new_policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh}, get_criteria => $new_policy, extra_fields => ['extra_virtual_server_assocs']);

    return { return_value => $new_policy };
}

#----------------------------------------------------------------------

1;
