package F5::ASMConfig::Handler::PolicyHistory;

use strict;
use warnings;

use F5::ImportExportPolicy;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::Deploy;
use F5::ASMConfig::Entity::PolicyHistory;
my $POLICY_VERSIONS_DIR = F5::Deploy::map_path("/ts/dms/policy/policy_versions/");
use File::Path;
use File::Find;
use File::Copy;
use File::Basename;


our @handled_functions = (
                           {
                             function_name => 'save_policy_history',
                             relayable => 1,
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::PolicyHistory';

#----------------------------------------------------------------------
sub save_policy_history {
  my ($self, $session, $call_flags, $policy, $policy_history) = @_;
  my $dbh = $self->{dbh};

  my $changeover_time = time;
  if ($policy_history->{save_policy_history_frequency_limit}) {
      # Check if last save by this user was within save_policy_history_frequency_limit seconds
      my @search_filters = (F5::ASMConfig::SearchFilter->new(field_name => "from_date",
                                                             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT(),
                                                             field_value => $changeover_time),
                            F5::ASMConfig::SearchFilter->new(field_name => "from_date",
                                                             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GT(),
                                                             field_value => $changeover_time - $policy_history->{save_policy_history_frequency_limit}),
                           );
      my $previous_revision = $entity_class->new(dbh => $dbh, policy => $policy)->retrieve(get_criteria => $policy_history, 
                                                                                           search_filters => \@search_filters, 
                                                                                          );
      if ($previous_revision) {
          return {return_value => undef};
      }
  }

  my $current_revision = $policy->last_revision_number() + 1;

  for my $previous_history ($entity_class->retrieve_all(dbh => $dbh, policy => $policy, get_criteria => { to_date => 0 })) {
      $previous_history->store_update(store_obj       => {to_date => $changeover_time}, 
                                      suppress_events => 1);
  }

  my $filename = F5::PolicyHistory->construct_filename(revision => $current_revision,
                                                       policy_id => $policy->{policy_id});

  mkpath(dirname($filename));
  my $import_export = F5::ImportExportPolicy->new(dbh => $dbh);
  $import_export->export_policy(filename  => $filename,
                                policy_id => $policy->{policy_id},
                                include_scanner_tables => 1,
                               );
  #if an empty file has been created, delete it and return.
  if( -e $filename && -z $filename ){
      $self->{logger}->error("Failed saving policy history file.") if $self->{logger};
      unlink $filename;
      return  { return_value => undef };
  }

  $policy_history->{from_date} = $changeover_time;
  $policy_history->{to_date} = 0;
  $policy_history->{revision} = $current_revision;

  if(! defined $policy->{policy_version_time}) {
      $policy->populate_version_fields();
  }  

  $policy_history->{policy_version_time} = $policy->{policy_version_time};
  $policy_history->{policy_version_device_name} = $policy->{policy_version_device_name};
  $policy_history->{policy_version_policy_name} = $policy->{policy_version_policy_name};
  $policy_history->{policy_version_last_change} = $policy->{policy_version_last_change};

  my $policy_history_entity = $entity_class->new(dbh => $dbh, policy => $policy, object => $policy_history);
  $policy_history_entity->store_add(suppress_events => 1);

  ##if the effective user id is root, chown the new files/directory created
  if (! F5::BigipUtils::is_docker() && $> == 0) {
      my ($login, $pass,$apache_uid,$apache_gid);
      ($login,$pass,$apache_uid,$apache_gid) = getpwnam("apache")
        or die "Cannot getpwnam user apache";
     
      # We must not change the owner of DEFAULT_BASE_DIRECTORY itself, because it belongs to tsui rpm.
      find(sub {chown $apache_uid,$apache_gid,$_}, F5::PolicyHistory::DEFAULT_BASE_DIRECTORY);
  }

  return { return_value => $policy_history_entity,
         };

}

#----------------------------------------------------------------------
1;
