package F5::ASMConfig::Handler::PolicyServerTechnology;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::Entity::PolicyServerTechnology;
use F5::NegativeSignatures::Utils;

our @handled_functions = (
                            {
                             function_name => 'add_policy_server_technology',
                             relayable => 1,
                             criteria_index => 3,
                             store_index => 4
                            },
                            {
                             function_name => 'update_policy_server_technology',
                             relayable => 1,
                             criteria_index => 3,
                             store_index => 4
                            },
                            {
                             function_name => 'enable_sigs_by_system',
                            },
                            {
                             function_name => 'get_messages_only_for_add_policy_server_technology',
                            },
                         );

our $entity_class = 'F5::ASMConfig::Entity::PolicyServerTechnology';

#----------------------------------------------------------------------
sub add_policy_server_technology {
    my ($self, $session, $call_flags, $policy, $policy_server_technology) = @_;

    my %retval;
    my $dbh = $self->{dbh};

    my $rv = $self->base_add(entity_class => $entity_class,
                             rpc_params => [ $session, $call_flags,$policy, $policy_server_technology ],);

    if ($call_flags && $call_flags->{apply_side_effects} ) {
        apply_server_technology_side_effects( self                     => $self,
                                              session                  => $session,
                                              call_flags               => $call_flags,
                                              policy                   => $policy,
                                              callback_function        => 'add_policy_server_technology',
                                              policy_server_technology => $policy_server_technology );

    }

    return $rv;
}

#----------------------------------------------------------------------

sub update_policy_server_technology {
    my ($self, $session, $call_flags, $policy, $policy_server_technology_criteria, $store_policy_server_technology) = @_;

    my %retval;
    my $dbh = $self->{dbh};

    my $rv = $self->base_update(entity_class => $entity_class,
                                rpc_params => [ $session, $call_flags,$policy, $policy_server_technology_criteria, $store_policy_server_technology ],);

    if ($call_flags && $call_flags->{suggestion_rest_uuid} ) {
        apply_server_technology_side_effects( self                     => $self,
                                              session                  => $session,
                                              call_flags               => $call_flags,
                                              policy                   => $policy,
                                              callback_function        => 'update_policy_server_technology',
                                              policy_server_technology => $rv->{return_value} );

    }

    return $rv;
}

#----------------------------------------------------------------------
sub apply_server_technology_side_effects {
        my (%params) = @_;
        my $self = $params{self};
        my $session = $params{session},
        my $call_flags = $params{call_flags};
        my $policy = $params{policy};
        my $callback_function = $params{callback_function};
        my $policy_server_technology = $params{policy_server_technology};
        my $dbh = $self->{dbh};
        my $get_messages_only = $params{get_messages_only};
        my @retval_messages;

        my $main_server_technology = F5::ASMConfig::Entity::ServerTechnology->new(get_criteria => $policy_server_technology->{server_technology}, dbh => $self->{dbh});

        if ( !$main_server_technology ){
             if( $get_messages_only ){ # Actually main_server_technology can be undefined only in this case, otherwise base_add was failing in add_policy_server_technology
                 my $msg;
                 if ( $policy_server_technology && $policy_server_technology->{server_technology} && $policy_server_technology->{server_technology}->{server_technology_name} ){
                     $msg = "Server Technology $policy_server_technology->{server_technology}->{server_technology_name} is not defined in this system.";
                 }
                 else {
                     $msg = "Server Technology is not defined.";
                 }
                 push (@retval_messages, $msg);
                 return;
             }
        }

        if ( $get_messages_only ){
            my $msg = "Server Technology $main_server_technology->{server_technology_display_name} will be associated.";
            my $policy_server_technology = F5::ASMConfig::Entity::PolicyServerTechnology->new(get_criteria => { server_technology => $main_server_technology}, dbh => $self->{dbh}, policy => $policy);
            if ( $policy_server_technology ) {
                $msg = "Server Technology $main_server_technology->{server_technology_display_name} already defined in this policy.";
            }
            push (@retval_messages, $msg);
        }

        if ( !$call_flags || !defined $call_flags->{include_implied_server_technologies} || $call_flags->{include_implied_server_technologies} ) {
            for my $implied_server_technology ( @{$main_server_technology->{implied_server_technologies}} ) {
                $implied_server_technology = F5::ASMConfig::Entity::ServerTechnology->new(get_criteria => { server_technology_name => $implied_server_technology}, dbh => $self->{dbh});

                my $policy_implied_server_technology = F5::ASMConfig::Entity::PolicyServerTechnology->new(get_criteria => { server_technology => $implied_server_technology}, dbh => $self->{dbh}, policy => $policy);
                if (!$policy_implied_server_technology){
                    $policy_implied_server_technology = F5::ASMConfig::PolicyServerTechnology->new(server_technology => $implied_server_technology);
                        my $rv = $self->$callback_function($session, $call_flags, $policy, $policy_implied_server_technology);
                        if ($get_messages_only && $rv && @{$rv}){
                            push (@retval_messages, @{$rv});
                        }
                }
            }
        }
        my @policy_actions = $dbh->selectrow_array(<<EndQuery,  undef, $main_server_technology->{server_technology_id});
SELECT policy_action
  FROM PLC.SERVER_TECHNOLOGIES_POLICY_ACTIONS
 WHERE server_technology_id = ?
EndQuery

        for my $action ( @policy_actions ){
            my $method;
            my @internal_params;
            if ($action =~ /([^(]*)\((.*)\)/) {
                $method = $1;
                @internal_params = split (/,/, $2);
                my $rv = $self->$method( $session, $call_flags,$policy,$get_messages_only, @internal_params );
                if ( $get_messages_only && $rv &&  @{$rv} ) {
                     push(@retval_messages, @{$rv});
                }
           }
        }

    return \@retval_messages;
}

#----------------------------------------------------------------------
sub enable_sigs_by_system {
    my ($self, $session, $call_flags,$policy, $get_messages_only, $system_name) = @_;

    return F5::NegativeSignatures::Utils::create_and_assign_system_signatures_set( asm_config_handler => $self,
                                                                                   session => $session,
                                                                                   call_flags => $call_flags,
                                                                                   policy => $policy,
                                                                                   get_messages_only => $get_messages_only,
                                                                                   system_name => $system_name );
}
#----------------------------------------------------------------------

sub get_messages_only_for_add_policy_server_technology{
   my ($self, $session, $call_flags, $policy, $policy_server_technology) = @_;

   my $retval_messages = apply_server_technology_side_effects( self                     => $self,
                                                               session                  => $session,
                                                               call_flags               => $call_flags,
                                                               policy                   => $policy,
                                                               policy_server_technology => $policy_server_technology,
                                                               get_messages_only        => 1,
                                                               callback_function        => 'get_messages_only_for_add_policy_server_technology',
                                                              );

  return { return_value =>  $retval_messages} ;

}

#----------------------------------------------------------------------

1;
