package F5::ASMConfig::Handler::PolicySignature;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::Entity::PolicySignature;
use F5::NegativeSignatures;

our @handled_functions = (
                           {
                             function_name => 'search_policy_signature',
                           },
                           {
                             function_name => 'update_many_policy_signatures',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                           },
                           {
                             function_name => 'update_policy_signature',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                           },
                           {
                             function_name => 'get_policy_signature',
                           },

                         );

our $entity_class = 'F5::ASMConfig::Entity::PolicySignature';

#----------------------------------------------------------------------

sub update_policy_signature {
    my($self,$session, $call_flags, $policy, $policy_signature_criteria, $store_policy_signature) = @_;

    if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        delete $store_policy_signature->{flg_quarantined};
    }

    my $update_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags, skip_inheritance => 1);
    my $retval = $self->base_update(
                                    rpc_params => [$session, $update_call_flags, $policy, $policy_signature_criteria, $store_policy_signature],
                                    entity_class => $entity_class,
                                   );

    if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        if ($retval->{events}) {
            push (@{$self->{events}}, @{delete $retval->{events}});
        }
        $self->propagate_inheritance_rpc(
                                         function_name => 'update_many_policy_signatures',
                                         call_flags => $update_call_flags,
                                         policy => $policy,
                                         criteria_obj => $policy_signature_criteria,
                                         store_obj    => $store_policy_signature,
                                         use_policy_filter => 1,
                                         section_type => F5::ASMConfig::SectionType::SECTION_TYPE_ATTACK_SIGNATURES(),
                                        );
    }


    return $retval;

}

#----------------------------------------------------------------------

sub get_policy_signature {
    my ($self, $session, $call_flags, $policy, $policy_signature_criteria) = @_;

    if ($call_flags->{rest_call}) {
      return $self->base_get (
                               rpc_params => [$session, $call_flags, $policy, $policy_signature_criteria],
                               entity_class => $entity_class,
                              );
    }
    $call_flags->{page_size} = 1;
    my $rv = $self->search_policy_signature($session, $call_flags, $policy, $policy_signature_criteria);
    if (!$rv->{total_record_count}) {
        return $self->handle_error(action => 'get',
                                   error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                   call_flags => $call_flags,
                                   entity_class => $entity_class,
                                   thrift_obj => $policy_signature_criteria,
                                  );
    }
    return { return_value => $rv->{element_list}->[0]->{policy_signature},
             };
}

#----------------------------------------------------------------------
sub search_policy_signature {
  my($self, $session, $call_flags, $policy, $policy_signature_criteria) = @_;

  if ($call_flags->{rest_call}) {
      my %search_filter_fields = ();
      if ($call_flags->{search_filters} && @{$call_flags->{search_filters}}) {
          %search_filter_fields = map { ($_->{field_name} => 1 ) } grep { !$_->{derived_filter} } @{$call_flags->{search_filters}};
      }

      if (! $search_filter_fields{sig_id} && ! $search_filter_fields{is_in_policy}) {
          push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new({field_name => 'is_in_policy',
                                                                                    field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                                                    field_value => 1}));
      }
      if ($search_filter_fields{is_in_policy}) {
          # If inPolicy is specified explicitly in $filter, it must be included in the search result.
          # This is necessary in order to distinguish from the automatically added filter above.
          if (!$call_flags->{rest_selected_fields} || !%{$call_flags->{rest_selected_fields}}) {
              $call_flags->{rest_selected_fields}->{'*'} = {};
          }
          $call_flags->{rest_selected_fields}->{inPolicy} = {};

          my @is_in_policy_filters = grep {$_->{field_name} eq 'is_in_policy'} @{$call_flags->{search_filters}};
          @{$call_flags->{search_filters}} = grep {$_->{field_name} ne 'is_in_policy'} @{$call_flags->{search_filters}};

          my $in_match = 1;
          my $out_match = 1;
          for my $filter (@is_in_policy_filters) {
              $in_match &&= $entity_class->is_filter_match(search_filter => $filter, field_value => 1);
              $out_match &&= $entity_class->is_filter_match(search_filter => $filter, field_value => 0);
          }
          if ($in_match && $out_match) {
              #Add no filter as everything should match
          }
          elsif ($in_match) {
              push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new({field_name => 'is_in_policy',
                                                                                        field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                                                        field_value => 1}));
          }
          elsif ($out_match) {
              push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new({field_name => 'is_in_policy',
                                                                                    field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NULL(),
                                                                                    }));
          }
          else {
              # User searched for is_in_policy = 1 and is_in_policy = 0, which obviously matches nothing
              return {return_value =>  F5::ASMConfig::SearchResult->new({
                                                        total_record_count => 0,
                                                        max_page => 1,
                                                        result_record_count => 0,
                                                        start_page => 1,
                                                        element_list => [],
                                                       })
                     };
          }
      }

      return $self->base_search(
                                rpc_params => [$session, $call_flags, $policy, $policy_signature_criteria],
                                entity_class => $entity_class,
                               );
  }

  my $dbh = $self->{dbh};
  my $negsig = F5::NegativeSignatures->new(dbh => $dbh);

  my %filter;
  my ($start_page,$limit,@order_by,@extra_fields,@search_filters);

  if ($call_flags->{search_filters}) {
    @search_filters = @{$call_flags->{search_filters}};
  }

  if (defined($policy_signature_criteria)) {
    for my $field (keys %$policy_signature_criteria) {
      if (defined($policy_signature_criteria->{$field})){
        push @search_filters, {field_name => $field, 
                               field_value => $policy_signature_criteria->{$field},
                               field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ,};
      }
    }
  }
  my %convert_to_op = map {$_ => 1} qw (risk accuracy last_update quarantined_since);
  for my $search_filter ( @search_filters ) {
    my $field_name = $search_filter->{field_name};

    if ($field_name eq 'systems' ) {
      @{$filter{$field_name}} = @{$search_filter->{field_value_list}};
    }
    elsif ( $field_name eq 'sig_id') {
       if ( $search_filter->{field_value} ){
          $filter{$field_name} = ("$search_filter->{field_value}");
       }
       elsif ( $search_filter->{field_value_list} ){
         @{$filter{$field_name}} = @{$search_filter->{field_value_list}};
       }
    }
    elsif ($convert_to_op{$field_name}) {
      if ( $search_filter->{field_comparison} eq F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE  ){
         $filter{$field_name.'_op'} = 'le';
      }
      elsif ( $search_filter->{field_comparison} eq F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ ){
         $filter{$field_name.'_op'} = 'eq';
      }
      elsif ($search_filter->{field_comparison} eq F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GE ){
         $filter{$field_name.'_op'} = 'ge';
      }

      $filter{$field_name.'_val'} = $search_filter->{field_value};
    }
    elsif ( $field_name eq 'staging_or_tightening' ) {
      $filter{flg_quarantined} = 1;
    }
    elsif ( $field_name eq 'staging_or_tightening_has_suggestions' ) {
      $filter{flg_quarantined} = 1;
      $filter{has_suggestions} = 1;
    }
    elsif ( $field_name eq 'staging_or_tightening_past_period' ) {
       $filter{flg_quarantined} = 1;
       $filter{has_suggestions} = 0;
       $filter{$field_name} = $search_filter->{field_value};
    }
    elsif ( $field_name eq 'extra_has_suggestions' ) {
        $filter{has_suggestions} = $search_filter->{field_value};
    }
    elsif ( $field_name eq 'extra_updated_in_enforcement_readiness_period' ) {
        $filter{updated_in_enforcement_readiness_period} = $search_filter->{field_value};
    }
    elsif ( $field_name eq 'flg_has_cookiecontent_http' ) {
        $filter{flg_has_valuecontent_http} = $search_filter->{field_value};
    }
    else{
       $filter{$field_name} = $search_filter->{field_value}; 
    }
  }

        
  if ($call_flags->{extra_fields}){
    @extra_fields = @{$call_flags->{extra_fields}};
  }

  if ($call_flags->{page_size}){
    $limit = $call_flags->{page_size};
  }

  if ($call_flags->{start_page}){
    $start_page = $call_flags->{start_page};
  }

  if ($call_flags->{sort_orders}){
    for my $sort_order ( @{$call_flags->{sort_orders}}) {

      my $order_field = $sort_order->{field_name};
      my $sort_direction = 'asc';
  
      if ( $sort_order->{sort_direction} eq F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC){
         $sort_direction = 'desc';
      }

      push @order_by,"$order_field $sort_direction";
    }
  }

  my $sig_recs = $negsig->get_signatures({ policy_id => $policy->{policy_id}, 
                                           filter => \%filter , 
                                           extra_fields => \@extra_fields,
                                           fields => \@extra_fields,
                                           start_page => $start_page,
                                           limit => $limit,
                                           order_by => \@order_by });

  my $selected_fields_hash;
  if ($call_flags->{selected_fields}) {
      $selected_fields_hash = {};
      for my $selected_field (@{$call_flags->{selected_fields}}) {
          my ($field, $subfield) = split('__', $selected_field);
          $selected_fields_hash->{$field} ||= {};
          if ($subfield) {
             $selected_fields_hash->{$field}->{$subfield} = {};
          }
      }
  }

  my $policy_sig_fields = $entity_class->all_fields();
  my $sig_fields = F5::ASMConfig::Entity::Signature->all_fields();

  my @sig_list;
  for my $sig ( @{$sig_recs->{results}} ) {
   if (!$selected_fields_hash || exists($selected_fields_hash->{signature})) {
       $sig->{signature} = F5::ASMConfig::Signature->new(rest_uuid => $sig->{sig_rest_uuid},
                                                         sig_id => $sig->{sig_id},
                                                         sig_name => $sig->{sig_name},);
   }

   for my $field (keys %$sig) {
     #The $sig object is a mish-mash of all fields that may belong on the Signature itself or on the PolicySignature
     #If there are selected then remove any fields that are not in the selected hash.
     #This is must be done on two levels for fields that belong to the PolicySignture object or the Signature object
     #The Signature object may have the entire object selected (empty hash), or only certain fields within the signature.
     if ($selected_fields_hash && (($policy_sig_fields->{$field} && !exists($selected_fields_hash->{$field})) ||
                                   (!$policy_sig_fields->{$field} && $sig_fields->{$field} && 
                                       (!exists($selected_fields_hash->{signature}) || 
                                           %{$selected_fields_hash->{signature}} && !exists($selected_fields_hash->{signature}->{$field})))
                                   )) {
         delete $sig->{$field};
         next;
     }
     if ($field =~ /_sets$/){
       for my $set (@{$sig->{$field}}){
         $set = F5::ASMConfig::PolicySet->new($set);
       } 
     }
   }

   if ($call_flags->{extra_fields}){
     for my $field ( @{$call_flags->{extra_fields}}) {
      if ($field eq 'sig_references'){
        my @sig_refs;
        if($sig->{sig_references}){
        for my $sig_ref (@{$sig->{sig_references}}){
          push @sig_refs, F5::ASMConfig::SigReference->new($sig_ref);
        }
       $sig->{signature}->{sig_references} = \@sig_refs;
       delete $sig->{sig_references};
       }
      }
      elsif ( $field ne 'recent_incidents' &&
               $field ne 'extra_has_suggestions' &&
                $field ne 'extra_updated_in_enforcement_readiness_period'){
       $sig->{signature}->{$field} = $sig->{$field};
       delete $sig->{$field};
      }
     }
   }

   my $policy_sig = F5::ASMConfig::PolicySignature->new($sig);
   push @sig_list, $policy_sig;

  }

  my $search_result = $self->entities_to_search_result(entities => \@sig_list,
                                                       call_flags => $call_flags,
                                                       total_record_count => $sig_recs->{total_found},
                                                       entity_class => $entity_class,
                                                      );

  my $retval = { return_value => $search_result,
               };
  return $retval;

}

#----------------------------------------------------------------------
sub update_many_policy_signatures {
  my($self, $session, $call_flags, $policy, $policy_signature_criteria, $store_policy_signature) = @_;

  my $update_many_call_flags = F5::ASMConfig::CallFlags->new(search_filters => [], #Initialize as empty list
                                                             %$call_flags,
                                                             skip_inheritance => 1); ##Forward this whole call for inheritance.

  if ($call_flags->{inherited_call}) {
      push(@{$update_many_call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new( field_name => 'is_inherited',
                                                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ,
                                                                                           field_value => 1)) ;
  }

  if ($policy && (! $call_flags->{search_filters} || 
                  ! grep { $_->{field_name} eq 'sig_id' && ($_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()
                                                         || $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ()
                                                         || $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ())} @{$call_flags->{search_filters}})) {
      my $search_result = $self->search_policy_signature($session, $call_flags, $policy, $policy_signature_criteria);
      my @sig_ids = map { $_->{policy_signature}->{sig_id} } @{$search_result->{element_list}};

      push(@{$update_many_call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new(
                                                                                         field_name => 'sig_id',
                                                                                         field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN,
                                                                                         field_value_list => \@sig_ids)) ;
       #This new search filter replaces the need for using the given policy_signature_criteria, so we can blank it out.
       $policy_signature_criteria = F5::ASMConfig::PolicySignature->new();
      
  }

  my $retval = $self->base_update_many(
                                       rpc_params => [$session, $update_many_call_flags, $policy, $policy_signature_criteria, $store_policy_signature],
                                       entity_class => $entity_class,
                                      );
  
  if ($policy && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
      if ($retval->{events}) {
          push (@{$self->{events}}, @{delete $retval->{events}});
      }
      $self->propagate_inheritance_rpc(
                                       function_name => 'update_many_policy_signatures',
                                       call_flags => $update_many_call_flags,
                                       policy => $policy,
                                       criteria_obj => $policy_signature_criteria,
                                       store_obj    => $store_policy_signature,
                                       use_policy_filter => 1,
                                       section_type => F5::ASMConfig::SectionType::SECTION_TYPE_ATTACK_SIGNATURES(),
                                      );
  }

  return $retval;
}


#----------------------------------------------------------------------



1;
