package F5::ASMConfig::Handler::PolicyTemplate;

use strict;
use warnings;

use File::Temp qw/tempfile/;
use File::Copy;
use F5::Deploy;
use F5::ImportExportPolicy::Base;
use F5::ASM::COAPI;
use F5::PolicyTemplateDefaults;
use base qw/F5::ASMConfig::Handler::Base/;

our @handled_functions = (
                             {
                                 function_name => "add_policy_template",
                                 filename_arguments => ['4'],
                                 relayable => 1,
                                 store_index => 3,
                             },
                             {
                                 function_name => "delete_policy_template",
                                 relayable => 1,
                             },
                             {
                                 function_name => "delete_many_policy_templates",
                                 relayable => 1,
                             },
                             {
                                 function_name => "export_policy_template",
                             },
                         );

our $TEMPLATE_DIR = F5::Deploy::map_path('/ts/var/policy_templates');
our $entity_class = 'F5::ASMConfig::Entity::PolicyTemplate';

#----------------------------------------------------------------------

sub add_policy_template {
    my($self, $session, $call_flags, $policy, $policy_template, $xml_file_path) = @_;

    my $dbh = $self->{dbh};

    my $policy_template_name = $policy_template->{policy_template_name};

    my ($policy_template_name_exists) = $dbh->selectrow_array(<<EndQuery, undef, $policy_template_name);
SELECT count(*)
  FROM PLC.POLICY_TEMPLATES
 WHERE name = ?
EndQuery

    if ($policy_template_name_exists) {
        return $self->handle_error(action => 'add',
                                   error_code   => F5::ASMConfig::ErrorCode::DUPLICATE_RECORD(),
                                   display_message => "The policy template $policy_template_name already exists.",
                                   error_level  => 'info',
                                  );
    }

    if ($policy) {
        my ($fh, $filename) = tempfile();
        close $fh;
        $self->export_policy($session, $call_flags, $policy, F5::ASMConfig::ImportExportProperties->new({
                                                                 filename => $filename,
                                                                 file_format => 'xml',

                                                             }));
        $xml_file_path = $filename;
        $policy_template->{policy_type} = $policy->{policy_type};
        $policy_template->{description} ||= $policy->{description};
    }
    else {
        require F5::ImportExportPolicy::XML;
        F5::ImportExportPolicy::XML::check_schema($xml_file_path);
    }

    require F5::XML::Twig;
    my $twig = F5::XML::Twig->new();
    $twig->parsefile($xml_file_path);

    my $template_filename = "$policy_template_name.xml";
    $template_filename =~ s/[ \/]/_/g;
    $template_filename = "$TEMPLATE_DIR/$template_filename";

    move($xml_file_path, $template_filename);

    $policy_template->{policy_template_name} = $policy_template_name;
    $policy_template_name =~ s/ /_/g;
    $policy_template->{icontrol_name} = $policy_template_name;
    $policy_template->{template_file} = $template_filename;
    $policy_template->{flg_is_user_defined} = 1;
    $policy_template->{template_type} = F5::ASMConfig::TemplateType::TEMPLATE_TYPE_USER_DEFINED();
    $policy_template->{flg_is_visible} = 1;

    my $template_defaults_json = F5::PolicyTemplateDefaults::xml_template_defaults($template_filename);
    if ($template_defaults_json) {
        $policy_template->{template_defaults} = $template_defaults_json;
        if (!$policy) {
            my $defaults = JSON::XS->new->utf8(0)->decode($template_defaults_json);
            $policy_template->{policy_type} = ($defaults->{type} && $defaults->{type} eq 'parent') ? F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()
                                                                                                   : F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY();
        }
    }

    my $rv = $self->base_add(entity_class => $entity_class,
                             rpc_params => [ $session, $call_flags, undef, $policy_template ],);

    if (! $call_flags->{skip_mcp}) {
        if ($rv->{return_value}->{policy_template_id}) {
            if (F5::ASM::COAPI::Login('root')) {
                my $object = F5::ASM::COAPI::Object->new("asm_predefined_policy");
                $object->{name} = $policy_template->{icontrol_name};
                $object->{predefined_asm_id} = $rv->{return_value}->{policy_template_id};
                $object->save();
            }
        }
    }

    return $rv;
}

#----------------------------------------------------------------------

sub delete_policy_template {
    my($self, $session, $call_flags, $place_holder, $policy_template) = @_;

    $policy_template = $entity_class->new(dbh => $self->{dbh}, get_criteria => $policy_template);

    if ($policy_template->{template_type} != F5::ASMConfig::TemplateType::TEMPLATE_TYPE_USER_DEFINED()) {
        die { error_message => "Can't delete pre-defined policy template." };
    }

    my $policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh}, get_criteria => {policy_template_id => $policy_template->{policy_template_id}});
    while (my $policy = $policies->get_next()) {
        my $policy_call_flags = F5::ASMConfig::CallFlags->new(set_fields_to_null => [ qw(policy_template_id) ]);
        $policy->{policy_template_id} = undef;
        $policy->{templ__icontrol_name} = '';
        $self->update_policy($session, $call_flags, $policy, $policy);
    }


    my $icontrol_name = $policy_template->{icontrol_name};
    my $template_file = $policy_template->{template_file};
    my $rv = $self->base_delete(entity_class => $entity_class,
                                rpc_params => [ $session, $call_flags, $place_holder, $policy_template ],
                               );

    my ($result) = @{F5::ASM::COAPI::Query->new('asm_predefined_policy',{'name' => $icontrol_name} )};
    if ($result) {
        $result->transactional_action(action => 'delete', dbh => $self->{dbh}, 
                                      success_action => { func => sub { unlink($template_file); } },
                                     );
    }

    return $rv;
}

#----------------------------------------------------------------------

sub delete_many_policy_templates {
    my($self, $session, $call_flags, $place_holder, $policy_template_criteria) = @_;

    $call_flags->{search_filters} ||= [];
    push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new(field_name => 'template_type',
                                                                             field_value => F5::ASMConfig::TemplateType::TEMPLATE_TYPE_USER_DEFINED(),
                                                                             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ()));
                                                                     
    my @policy_templates = $entity_class->retrieve_all(dbh => $self->{dbh}, 
                                                       call_flags => $call_flags,
                                                       search_filters => $call_flags->{search_filters},
                                                       get_criteria => $policy_template_criteria);
    if (@policy_templates) {
        my @policy_template_ids = map { $_->{policy_template_id} } @policy_templates;

        my $policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh}, 
                                                                   search_filters => [F5::ASMConfig::SearchFilter->new( field_name => 'policy_template_id',
                                                                                                                        field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                                                                                        field_value_list => \@policy_template_ids )],
                                                                   minimal_load => 1,
                                                                   get_criteria => {});
        while (my $policy = $policies->get_next()) {
            my $policy_call_flags = F5::ASMConfig::CallFlags->new(set_fields_to_null => [ qw(policy_template_id) ]);
            $policy->{policy_template_id} = undef;
            $policy->{templ__icontrol_name} = '';
            $self->update_policy($session, $call_flags, $policy, $policy);
        }
    }

    my $rv = $self->base_delete_many(entity_class => $entity_class,
                                     rpc_params => [ $session, $call_flags, $place_holder, $policy_template_criteria ],
                                    );

    for my $policy_template (@policy_templates) {
        my ($result) = @{F5::ASM::COAPI::Query->new('asm_predefined_policy',{'name' => $policy_template->{icontrol_name}} )};
        if ($result) {
            $result->transactional_action(action => 'delete', dbh => $self->{dbh},
                                          success_action => { func => sub { unlink($policy_template->{template_file}); } },
                                         );
        }
    }

    return $rv;
}


#----------------------------------------------------------------------

1;
