package F5::ASMConfig::Handler::RequestLog;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::RequestLog;
use F5::SystemUtils;
use File::Path;
use File::Basename;
use F5::Cfg;

our @handled_functions = (
                           {
                             function_name => 'search_request_log',
                             authz_limit_field => 'request_policy__policy_id',
                           },
                           {
                             function_name => 'update_request_log',
                             authz_limit_field => 'request_policy__policy_id',
                             criteria_index => 3,
                             store_index => 4,
                           },
                           {
                             function_name => 'update_many_request_logs',
                             authz_limit_field => 'request_policy__policy_id',
                             criteria_index => 3,
                             store_index => 4,
                           },
                           {
                             function_name => 'delete_request_log',
                             authz_limit_field => 'request_policy__policy_id',
                             criteria_index => 3,
                             store_index => 4,
                           },
                           {
                             function_name => 'delete_many_request_logs',
                             authz_limit_field => 'request_policy__policy_id',
                             criteria_index => 3,
                             store_index => 4,
                           },
                           {
                             function_name => 'export_request_log',
                             authz_limit_field => 'request_policy__policy_id',
                           },
                           {
                             function_name => 'get_request_log',
                             authz_limit_field => 'request_policy__policy_id',
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::RequestLog';
my $REQUEST_DELETE_BATCH_SIZE = 5000;
my $DEFAULT_ROW_LIMIT = F5::Cfg::cfg_get_config_item('/etc/ts/tools/export_proxy_log.cfg', 'General', 'ProxyLogLimit');
my $TEMP_FILE_BASE_DIR = '/var/tmp';

#----------------------------------------------------------------------

sub search_request_log {
    my($self, $session, $call_flags, undef, $request_log_criteria) = @_;
    $call_flags->{search_filters} ||= [];
    my $search_filters = $call_flags->{search_filters};
    my $sort_orders = $call_flags->{sort_orders} || [];

    if ($request_log_criteria) {
        my $obj = $entity_class->new(object => $request_log_criteria,
                                     dbh => $self->{dbh});
        push (@$search_filters, @{$obj->to_search_filters});
    }

    my $specific_search = grep { ($_->{field_name} eq 'support_id' || $_->{field_name} eq 'rest_uuid' || $_->{field_name} eq 'request_log_id') &&
                                 ($_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() ||
                                  $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() ||
                                 $_->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()) } @$search_filters;

    if (!$specific_search) {
        $call_flags->{hide_deleted} = 1;
        push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new(field_name => 'flg_display',
                                                                                 field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                                                 field_value => 1));
    }

    $self->add_partition_search_filter(session    => $session,
                                       call_flags => $call_flags,
                                       search_field => 'request_policy__partition_id');

    if ( (scalar @{$sort_orders} == 1) && ($sort_orders->[0]->{field_name} eq 'log_time') ) {
            my $id_sort_order = F5::ASMConfig::SortOrder->new(field_name => 'request_log_id',
                                                              sort_direction => $sort_orders->[0]->{sort_direction});
            push(@{$sort_orders}, $id_sort_order);
    }

    my $rv = $self->base_search(rpc_params => [$session, $call_flags, undef, $request_log_criteria],
                                entity_class => $entity_class,
                               );
    return $rv;
}

#----------------------------------------------------------------------

sub create_accepted_requests {
    my (%params) = @_;
    my $self = $params{self};
    my $request_logs = $params{request_logs};
    my $store_request_log = $params{store_request_log};
    my $dbh = $params{dbh};

    for my $request_log (@$request_logs) {
        if (defined($store_request_log->{accept_status}) && $store_request_log->{accept_status} == F5::ASMConfig::AcceptStatus::ACCEPT_STATUS_START()) {
            my $protobuf_bin = $request_log->{protobuf_bin};             
            if (!$protobuf_bin) { 
                $protobuf_bin = F5::RequestLog->get_protobuf(rec => $request_log);
                next unless defined $protobuf_bin;
            }
            my $accepted_request = F5::ASMConfig::AcceptedRequest->new(support_id => $request_log->{support_id},
                                                                 protobuf_bin => $protobuf_bin,
                                                                 gui_user_name => $self->{session}->{gui_user_name},
                                                                 accept_status => F5::ASMConfig::AcceptStatus::ACCEPT_STATUS_START());

            if ( $request_log->{request_policy}->{has_parent} ){
                my $error = "This action is disabled for security policies with parent policy.";
                return $self->handle_error(action => 'add_accepted_request',
                                   entity_class => $entity_class,
                                   error_code   => F5::ASMConfig::ErrorCode::INVALID_ACTION_FOR_POLICY_TYPE(),
                                   error_thrown => $error,
                                   error_level  => 'error',
                                  );
            }
 
            $self->add_accepted_request($self->{session}, F5::ASMConfig::CallFlags->new(ignore_duplicate_record_on_add => 1), $request_log->{request_policy}, $accepted_request);
        }
    }
}

#----------------------------------------------------------------------
sub update_request_log {
    my($self, $session, $call_flags, undef, $request_log_criteria, $store_request_log) = @_;

    
    my $retval =  $self->base_update(rpc_params => [$session, $call_flags, undef, $request_log_criteria, $store_request_log],
                                     entity_class => $entity_class,
                                    );
    create_accepted_requests(dbh => $self->{dbh},
                          self => $self,
                          store_request_log => $store_request_log,
                          request_logs => [$retval->{return_value}]);
    return $retval;
}

#----------------------------------------------------------------------

sub update_many_request_logs {
    my($self, $session, $call_flags, undef, $request_log_criteria, $store_request_log) = @_;

    my $retval = $self->base_update_many(rpc_params => [$session, $call_flags, undef, $request_log_criteria, $store_request_log],
                                         entity_class => $entity_class,
                                        );
    create_accepted_requests(dbh => $self->{dbh},
                          self => $self,
                          store_request_log => $store_request_log,
                          request_logs => $retval->{entities_changed});
    return $retval;
}

#----------------------------------------------------------------------

sub delete_request_log {
    my($self, $session, $call_flags, undef, $request_log) = @_;

    my $class = ref($request_log) || "F5::ASMConfig::RequestLog";
    my $store_request_log = $class->new();
    $store_request_log->{flg_is_deleted} = 1;
    my $retval =  $self->base_update(rpc_params => [$session, $call_flags, undef, $request_log, $store_request_log],
                                     entity_class => $entity_class,
                                    );
    return $retval;
}

#----------------------------------------------------------------------

my %STRICTLY_INCREASING_DB_FIELDS = ( support_id => 1,
                                      rest_uuid => 1,
                                      log_time => 1,
                                      request_log_id => 1,
                                     );

sub delete_many_request_logs {
    my($self, $session, $call_flags, undef, $request_log_criteria) = @_;

    my $max_id; 
    my $search_filters = $call_flags->{search_filters} || [];
    if ($request_log_criteria) {
        my $obj = $entity_class->new(object => $request_log_criteria,
                                     dbh => $self->{dbh});
        push (@$search_filters, @{$obj->to_search_filters});
    }

    if (!@{$search_filters} && !$call_flags->{page_size}) {
        #Delete all
        $max_id = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT MAX(id) FROM PRX.REQUEST_LOG
EndQuery
    }
    elsif (scalar @{$search_filters} == 1 && !$call_flags->{page_size}) {
        my ($search_filter) = @{$search_filters};
        if ($STRICTLY_INCREASING_DB_FIELDS{$search_filter->{field_name}} &&
            ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT() ||
             $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE())
           ) {
            my $db_field_name = $entity_class->fields->{$search_filter->{field_name}}->{db_field_name};
            my $comparison = ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT()) ? '<' 
                                                                                                                           : '<=';
            $max_id = $self->{dbh}->selectrow_array(<<EndQuery, undef, $search_filter->{field_value});
SELECT MAX(id) FROM PRX.REQUEST_LOG
  WHERE $db_field_name $comparison ?
EndQuery
        }        
    } 

    my $rv = 0;
    if ($max_id) {
        my $request_log_iterator = $entity_class->retrieve_all(dbh => $self->{dbh},
                                                              page_size => 1,
                                                              selected_fields => ['request_log_id'],
                                                              search_filters => [F5::ASMConfig::SearchFilter->new(field_name => 'request_log_id',
                                                                                                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE(),
                                                                                                                  field_value => $max_id)],
                                                              return_as_raw_db_hashref => 1,
                                                             );
        $rv = $request_log_iterator->{total_record_count};

        $self->{dbh}->do(<<EndQuery, undef, $max_id);
UPDATE PRX.REQUEST_LOG_CLEARED
 SET request_log_id = ?;
EndQuery
    }
    else {
        $call_flags->{hide_deleted} = 1;
        push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new(field_name => 'flg_display',
                                                                                 field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                                                 field_value => 1));

        $self->add_partition_search_filter(session    => $session,
                                           call_flags => $call_flags,
                                           search_field => 'request_policy__partition_id');

        my $request_log_iterator = $entity_class->retrieve_all(dbh => $self->{dbh},
                                                               sort_orders => $call_flags->{sort_orders},
                                                               page_size => $call_flags->{page_size},
                                                               start_page => $call_flags->{start_page},
                                                               get_criteria => $request_log_criteria,
                                                               selected_fields => ['request_log_id'],
                                                               search_filters => $call_flags->{search_filters},
                                                               return_as_raw_db_hashref => 1,
                                                               call_flags => $call_flags,
                                                               skip_total_record_count => 1,
                                                              );
        my $batched_insertor = F5::BatchInsert->new(dbh => $self->{dbh},
                                                    logger => $self->{logger},
                                                    batch_size => $REQUEST_DELETE_BATCH_SIZE);

        while (my $request_log = $request_log_iterator->get_next()) {
           $batched_insertor->insert_row(table_name => 'PRX.REQUEST_LOG_PROPERTIES',
                                         rec => {request_log_id => $request_log->{request_log_id},
                                                 flg_is_deleted => 1},
                                         do_replace => 1,
                                       );
           $rv++;
        }
        $batched_insertor->flush;
    }

    if ($call_flags->{rest_call}) {
        $rv = F5::ASMConfig::SearchResult->new(total_record_count => $rv);
    }
    return {return_value => $rv};
}

#----------------------------------------------------------------------

sub extract_export_value {
    my (%params) = @_;
    my $field_def = $params{field_def};
    my $field_value = $params{field_value};

    if (!defined($field_value)) {
        $field_value = '\N';
    }
    else {
        if (my $ref = ref($field_value)) {
            my @values = ($field_value);
            if ($ref eq 'ARRAY') {
                @values = @$field_value;
            }

            if ($field_def->{export_field}) {
                @values = map {$_->{$field_def->{export_field}}} @values;
                if ($field_def->{fields} && $field_def->{fields}->{$field_def->{export_field}}) {
                    $field_def = $field_def->{fields}->{$field_def->{export_field}};
                }
            }

            if ($field_def->{entity_type}) {
                @values = map { $field_def->{entity_type}->new(dbh => $params{dbh},
                                                               object => $_)->get_full_display_name() } @values;
            }
            $field_value = join(", ", @values);
        }

        $field_value =~ s/\"/\"\"/g;
        $field_value =~ s/\n/\\/g;
        $field_value = qq{"$field_value"};
    }
    return $field_value;
}

#----------------------------------------------------------------------

sub export_request_log{
    my($self, $session, $call_flags,$place_holder, $export_request_log_properties) = @_;

    if ($export_request_log_properties->{filter}) {
        #Convert where $filter to search filters
        F5::ASMConfig::Handler::Rest::convert_odata(call_flags => $call_flags,
                                                    rest_parameters => {'$filter' => $export_request_log_properties->{filter}},
                                                    path_pointer => $F5::ASMConfig::Handler::Rest::path_mapper->{references}->{events}->{references}->{requests},
                                                    dbh => $self->{dbh},
                 );

    }

    my $file_size_limit = $export_request_log_properties->{file_size_limit};
    my $row_limit = $export_request_log_properties->{row_limit} || $DEFAULT_ROW_LIMIT;
    my @bundled_files = $export_request_log_properties->{bundled_files} || qw(/VERSION);
    my $target_file = $export_request_log_properties->{target_file};
    my $temp_dir = "$TEMP_FILE_BASE_DIR/export_request_log_$$";
    my $dump_file = "$temp_dir/requests.csv";
    mkpath($temp_dir);

    my $request_log_iterator = F5::ASMConfig::Entity::RequestLog->retrieve_all(dbh => $self->{dbh},
                                                                               call_flags => $call_flags,
                                                                               page_size => $row_limit,
                                                                               search_filters => $call_flags->{search_filters},);

    my @export_fields = @{F5::ASMConfig::Entity::RequestLog->export_fields()};
    my $field_defs = F5::ASMConfig::Entity::RequestLog->fields();
    open(my $outf, '>', $dump_file);
    #print field heading
    my $output_line = join (",", map { $field_defs->{$_}->{export_display_name} } @export_fields) . "\n";
    my $file_size = length($output_line);
    print $outf $output_line;

    while (my $request_log = $request_log_iterator->get_next()) {
        #print row data
        my @out_fields;
        for my $field (@export_fields) {
            my $value = extract_export_value(field_def => $field_defs->{$field},
                                             field_value => $request_log->{$field});
            push (@out_fields, $value);
        }
        $output_line = join(',', @out_fields) . "\n";
        $file_size += length($output_line);
        if ($file_size_limit && $file_size > $file_size_limit) {
            last;
        }
        print $outf $output_line;
    }

    close $outf;

    chmod(0666, $dump_file);

    for my $bundled_file (@bundled_files) {
        symlink($bundled_file, $temp_dir . '/' .basename($bundled_file));
    }
    my $rv = F5::SystemUtils::run_system_tool(command_line_params => [
                                                                       '/bin/tar',
                                                                       '-z',
                                                                       '-c',
                                                                       '-h',
                                                                       '-f', $target_file,
                                                                       '-C', $temp_dir,
                                                                       basename($dump_file),
                                                                       map { basename($_) } @bundled_files,
                                                                      ]);
    rmtree($temp_dir);
    if (! $rv->{completed_successfully}) {
        my $error_thrown = "Export request log failed: $rv->{standard_error}";
        return $self->handle_error( action => "export",
                                    error_code => F5::ASMConfig::ErrorCode::LOG_EXPORT_ERROR(),
                                    error_thrown => $error_thrown,
                                    call_flags => $call_flags,
                                    error_level => 'error',
                                  );
    }

    chmod(0666, $target_file);

    return { return_value => "Export request log completed successfully."};
}

#----------------------------------------------------------------------

sub get_request_log {
    my($self, $session, $call_flags, undef, $request_log) = @_;

    return $self->base_get(rpc_params => [$session, $call_flags, undef, $request_log],
                                     entity_class => $entity_class,
                                    );
}

#----------------------------------------------------------------------

1;
