package F5::ASMConfig::Handler::Rest;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::Deploy;
use F5::Utils::URI;
use F5::Utils::Rest;
use F5::BigipVersionUtils;
use Encode;
use YAML::XS;
use JSON -convert_blessed_universally;
use Scalar::Util qw(reftype);
use File::Basename;
use F5::Provisioning;
use F5::ASMConfig::Handler;

our @handled_functions = (
                          {
                              function_name => 'rest_call',
                              relayable => 0, 
                              skip_authz => 1, #Done during internal call
                          },
                          {
                              function_name => 'rest_request',
                              relayable => 0,
                              skip_authz => 1,
                          },
                          {
                              function_name => 'generate_rest_api_doc',
                          },
                          {
                           function_name => 'generate_openapi_doc',
                          },
                          {
                           function_name => 'generate_json_schema',
                          },
                         );

my $BASE_PATH = "tm/asm/";

my $collection_allowed_methods = { DELETE => 1,
                                   GET    => 1,
                                   PATCH => 1,
                                   POST  => 1,
                                 };

my $resource_allowed_methods =  { DELETE => 1,
                                  GET    => 1,
                                  PATCH => 1,
                                 };


our $path_mapper = YAML::XS::LoadFile(F5::Deploy::map_path('/usr/share/ts/share/asm_config/yaml/rest_path_map.yaml'));

my %direction_map = (
                        'asc' => F5::ASMConfig::SortDirection::SORT_DIRECTION_ASC(),
                        'desc' => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC(),
                    );

#------------------------------------------------------------------
sub rest_call {
    my $rv = rest_request(@_);
    if (ref($rv) && ref($rv->{return_value})) {
        $rv->{return_value} = $rv->{return_value}->{response_body};
    }
    return $rv;
}

#------------------------------------------------------------------
sub rest_request {
    my ($self, $session, $call_flags, $place_holder, $rest_request) = @_;
    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $path = $rest_request->{rest_path};
    my $do_transaction = (!$call_flags->{do_not_relay} && $rest_request->{rest_method} ne "GET");

    $path =~ s/.*$BASE_PATH//;
    my @path_parts = split("/", $path);
    my $path_pointer = $path_mapper;
    my $current_entity_type;
    my $rv = {};
    my $objects;
    my $current_obj;
    my $kind_path = 'tm/asm';
    my $is_collection;
    my $get_criteria;
    my $is_example;
    my $search_children;

    if ($path_parts[-1] eq 'example') {
        $is_example = pop @path_parts;
    }
    if ($path_parts[-1] eq 'children') {
        $search_children = pop @path_parts;
    }

    if ($rest_request->{rest_parameters}->{ignoredErrors}) {
        my @codes = split(",",$rest_request->{rest_parameters}->{ignoredErrors});
        my @ignored_errors = map { F5::ASMConfig::IgnoredError->new({error_code => $_}) } @codes;
        $call_flags->{rest_ignored_errors} = \@ignored_errors;
    }
    my $ignored_rv = F5::ASMConfig::RestResponse->new(response_body => $rest_request->{rest_body} || '');

    while (my ($collection_name) = (shift @path_parts)) {
        last unless $collection_name;
        $get_criteria = {};

        if ($path_pointer->{references} && $path_pointer->{references}->{$collection_name}) {
            $is_collection = 1;
            $kind_path .= "/" .$collection_name;
            $path_pointer = $path_pointer->{references}->{$collection_name};
            my $entity_type = $path_pointer->{entity_type};
            if (is_path_excluded(path_pointer => $path_pointer, 
                                 policy => $objects->{policy}, 
                                 call_flags => $call_flags,
                                 dbh => $self->{dbh})) {
                return $self->handle_error( error_code => F5::ASMConfig::ErrorCode::NOT_IMPLEMENTED(),
                                            error_thrown => "Invalid Path",
                                            error_level  => "info",
                                            ignored_rv => $ignored_rv,
                                            call_flags => $call_flags,
                                          );
            }

            #Will need to deal with external entities from longer paths
            if ($path_pointer->{extern_fields}) {
                for my $extern_field (keys %{$path_pointer->{extern_fields}}) {
                    $get_criteria->{$extern_field} = $objects->{$path_pointer->{extern_fields}->{$extern_field}};
                }
            }
            if ($path_pointer->{literal_fields}) {
                for my $literal_field (keys %{$path_pointer->{literal_fields}}) {
                    if (!defined($get_criteria->{$literal_field})) {
                        $get_criteria->{$literal_field} = $path_pointer->{literal_fields}->{$literal_field};
                    }
                }
            }

            undef $current_obj;
            my $identifier;
            if ($entity_type
                && ($path_pointer->{singleton} || (defined($identifier = shift @path_parts)) )) {
               $is_collection = 0;
               my $search_filters = $entity_type->rest_search_filters(session => $session);
               if (defined($identifier) && (length($identifier) || scalar @path_parts)) {
                   push (@$search_filters, F5::ASMConfig::SearchFilter->new(field_name => 'rest_uuid',
                                                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ(),
                                                                            field_value => $identifier));
               }

               if ($identifier && $identifier eq "example") {
                   $current_obj = $entity_type->new(dbh => $self->{dbh}, 
                                                    object => {rest_uuid => 'example'},
                                                    policy => $objects->{policy});
               }
               else {
                   $current_obj = $entity_type->new(dbh => $self->{dbh},
                                                    call_flags => F5::ASMConfig::CallFlags->new({rest_call => 1}),
                                                    selected_fields => [], #retrieve only id and unique record parts
                                                    get_criteria => $get_criteria, 
                                                    search_filters => $search_filters,
                                                    policy => $objects->{policy},
                                                   );
               }

               if (!$current_obj) {
                    return $self->handle_error(action => 'get',
                                        entity_class => $path_pointer->{entity_type},
                                        error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                        error_message => ($entity_type->element_type_name . " not found"),
                                        error_level => 'info',
                                        ignored_rv => $ignored_rv,
                                        call_flags => $call_flags,
                                       );
               }
               $objects->{$entity_type->element_type_name} = $current_obj;
            }
        }
        else {
            return $self->handle_error( error_code => F5::ASMConfig::ErrorCode::NOT_IMPLEMENTED(),
                                 error_thrown => "Invalid Path",
                                 error_level  => "info",
                                 ignored_rv => $ignored_rv,
                                 call_flags => $call_flags,
                               );
        }
    }

    if ($search_children && (!$objects->{policy} || $objects->{policy}->{policy_type} != F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT())) {
        return $self->handle_error( error_code => F5::ASMConfig::ErrorCode::NOT_IMPLEMENTED(),
                             error_thrown => "/children called with no Parent Policy in Path",
                             error_level  => "info",
                             ignored_rv => $ignored_rv,
                             call_flags => $call_flags,
                            );
    }

    my $allowed_methods;
    $is_collection ||= $search_children;
    if ($is_collection) {
        $allowed_methods = $path_pointer->{collection_allowed_methods} || $path_pointer->{allowed_methods} ||  $collection_allowed_methods;
        if (!$rest_request->{rest_parameters}->{'$top'} && $path_pointer->{default_top} && $path_pointer->{default_top}->{$rest_request->{rest_method}}) {
            $rest_request->{rest_parameters}->{'$top'} = $path_pointer->{default_top}->{$rest_request->{rest_method}};
        }
    }
    else {
        $allowed_methods = $path_pointer->{allowed_methods} || $resource_allowed_methods;
    }

    if ($search_children && $path_pointer->{singleton}) {
        #Single endpoints should never be deleted
        #TODO: Update many may be allowed in the future, but the framework is not currently expecting an update_many for these elements 
        #      and there may be unexpected results
        $allowed_methods = {%$allowed_methods, ( 'DELETE' => 0, 'PATCH' => 0)};
    }

    $rest_request->{rest_parameters}->{ver} ||=  F5::BigipVersionUtils::bigip_version();
    $call_flags->{rest_call} = 1; #set flag for special results (ie SearchResult for update_many calls

    $call_flags->{rest_version} = $rest_request->{rest_parameters}->{ver};
    if ($call_flags->{rest_version} !~ /^\d{1,4}\.\d{1,4}\.\d{1,4}(\.|$)/) {
        die {
            error_code  => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
            error_message => "Rest version format is incorrect.",
            error_level  => "info",
        };
    }
    my $cur_version = F5::BigipVersionUtils::bigip_version();
    if (F5::BigipVersionUtils::compare_versions(version1 => $cur_version, version2 => $call_flags->{rest_version}) < 0) {
        die {
            error_code  => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
            error_message => "Requested rest version is beyond current version.",
            error_level  => "info",
        };
    }

    if ( $rest_request->{rest_parameters}->{actAsComponent} &&
         $rest_request->{rest_parameters}->{actAsComponent} ne $F5::ASMConfig::Handler::TSCONFD_USER_NAME ) {
       $session->{user_name} = $rest_request->{rest_parameters}->{actAsComponent};
    }

    $call_flags->{show_all_rest_fields} = $rest_request->{rest_parameters}->{allfields};
    $call_flags->{show_rest_inheritance_fields} = $rest_request->{rest_parameters}->{inheritancefields};
    $call_flags->{append_nested_struct_on_update} = $rest_request->{rest_parameters}->{appendnested};
    $call_flags->{remove_nested_struct_on_update} = $rest_request->{rest_parameters}->{removenested};
    $call_flags->{delete_disinherited_entities} = $rest_request->{rest_parameters}->{deleteDisinheritedEntities};
    $call_flags->{update_on_adding_duplicate_record} = $rest_request->{rest_parameters}->{updateOnDuplicateRecord};
    $call_flags->{update_wildcard_entities} = $rest_request->{rest_parameters}->{updateWildcardEntities};
    $call_flags->{displace_only_target_wildcard_order} = $rest_request->{rest_parameters}->{displaceOnlyTargetWildcardOrder};
    $call_flags->{ignore_inherited_fields} = $rest_request->{rest_parameters}->{ignoreInheritedFields};
    $call_flags->{verbose_references} = $rest_request->{rest_parameters}->{verboseReferences};


    my %rest_parameters = %{$rest_request->{rest_parameters} || {} };

    convert_odata(call_flags => $call_flags,
                  rest_request => $rest_request,
                  path_pointer => $path_pointer,
                  dbh => $self->{dbh},
                  policy => $objects->{policy},
                 );
    $call_flags->{rest_selected_fields} ||= build_odata_field_tree('value' => $rest_parameters{'$select'});
    $call_flags->{expand_fields} ||= build_odata_field_tree('value' => $rest_parameters{'$expand'});

    if ($is_example) {
        my $response_body = rest_doc(path_pointer => $path_pointer,
                                     dbh => $self->{dbh},
                                     doc_type => "rest",
                                     rest_parameters => $rest_request->{rest_parameters},
                                     policy => $objects->{policy},
                                     path => $path)->{return_value};

        my $response_parameters = $path_pointer->{response_parameters} || {};
        my $response = F5::ASMConfig::RestResponse->new(response_body => $response_body,
                                                        %$response_parameters);
        return {return_value => $response, events => []};
    }

    if ($search_children && $objects->{policy} && $path_pointer->{entity_type}) {
        push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new( field_name => 'policy_id',
                                                                                  field_value_list => $objects->{policy}->child_policy_ids(),
                                                                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),));
        $call_flags->{full_policy_populate} = 1;
        delete $objects->{policy};
        undef $current_obj;
    }

    if ($path_pointer->{entity_type}) {
        push (@{$call_flags->{search_filters}}, @{$path_pointer->{entity_type}->rest_search_filters(session => $session,
                                                                                                    is_collection => $is_collection,
                                                                                                    call_flags => $call_flags,)});
        my $element_type_name = $path_pointer->{entity_type}->element_type_name;
        my $plural_element_type_name = $path_pointer->{entity_type}->element_type_name_plural;
        #TODO: extra fields need to be part of metadata.
        $call_flags->{extra_fields} = ["extra_virtual_server_assocs"];

        my $rpc_method;
        my $action;
        my $store_obj;
        my $criteria_obj = $current_obj;
        if ($rest_request->{rest_body}) {
            eval {
                $store_obj = $path_pointer->{entity_type}->new(dbh => $self->{dbh},
                                                               policy => $objects->{policy},
                                                               extract_policy_field => $path_pointer->{extract_policy_field},
                                                               call_flags => $call_flags,
                                                               rest_object => $rest_request->{rest_body},
                                                               references => $path_pointer->{references});
                if ($path_pointer->{extract_policy_field}) {
                    $objects->{policy} = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, 
                                                                                 policy => $store_obj->{policy});
                }
            };
            if ($@) {
                  return $self->handle_error(action => 'parse/validate',
                                      entity_class => $path_pointer->{entity_type},
                                      call_flags => $call_flags,
                                      thrift_obj  => $current_obj,
                                      ignored_rv => $ignored_rv,
                                      error_thrown => $@,
                                    ); 
            }

            for my $field (keys %$get_criteria) {
                $store_obj->{$field} = $get_criteria->{$field};  #Add foreign dependencies
            }
        }

        if ($rest_request->{rest_method} eq "GET") {
            $action = ($is_collection ? 'search' : 'get' );
            $rpc_method = $action . '_' . $element_type_name;
        }
        elsif ($rest_request->{rest_method} eq "DELETE") {
            $action = 'delete';
            $rpc_method = $action . '_' . ($is_collection ? "many_$plural_element_type_name" :  $element_type_name );
        }
        else {
            if ($rest_request->{rest_method} eq "POST") {
                if ($store_obj->{command}) {
                    $action = $store_obj->{command};
                    $rpc_method = "rest_command_$element_type_name";
                }
                else {
                    $action = 'add';
                    $rpc_method = $action . '_' . $element_type_name;
                }
                $objects->{$path_pointer->{entity_type}->element_type_name} ||= $store_obj;
                $criteria_obj ||= $store_obj;
            }
            elsif ($rest_request->{rest_method} eq "PATCH") {
                $action = 'update';
                $rpc_method = $action . '_' . ($is_collection ? "many_$plural_element_type_name" : $element_type_name );
            } 
        }       

        if ($allowed_methods->{$rest_request->{rest_method}} && $self->can($rpc_method)) {
            $criteria_obj ||= $path_pointer->{entity_type}->new(object => $get_criteria);
            my @rpc_params = ($session, $call_flags);
            if ($element_type_name ne "policy") {
                push (@rpc_params, $objects->{policy});
            }
            push (@rpc_params, $criteria_obj, $store_obj); 
            if ($do_transaction) {
                $self->begin_work($self->{session});  #This will handle all the events being sent at the right time and CMI relaying
            }
            $rv = $self->$rpc_method(@rpc_params);
            if ($do_transaction) {
                $self->commit($self->{session});
                delete $call_flags->{relayable}; #Never relay the rest_call itself:
            }
        }
        else {
            return $self->handle_error(action => $action,
                                entity_class => $path_pointer->{entity_type},
                                call_flags => $call_flags,
                                ignored_rv => $ignored_rv,
                                error_code   => F5::ASMConfig::ErrorCode::NOT_IMPLEMENTED(),
                                error_thrown => "Method not implemented.",
                                error_level  => "info",);
        }
        if (ref($rv) && (ref($rv) eq "F5::ASMConfig::SearchResult")) {
            $rv = convert_search_result(search_result => $rv,
                                        call_flags    => $call_flags,
                                        element_type_name => $element_type_name,
                                        entity_class => $path_pointer->{entity_type},
                                        policy => $objects->{policy},
                                        self => $self,
                                       );
        }
    }
    elsif (!$allowed_methods->{$rest_request->{rest_method}}) {
        return $self->handle_error(action => $rest_request->{rest_method},
                            call_flags => $call_flags,
                            error_code   => F5::ASMConfig::ErrorCode::NOT_IMPLEMENTED(),
                            ignored_rv => $ignored_rv,
                            error_thrown => "Method not implemented.",
                            error_level  => "info",);
    }

    my $filter          = $rest_parameters{'$filter'};

    $rv = add_references(rv => $rv,
                         path_pointer => $path_pointer,
                         rest_request => $rest_request,
                         kind_path   => $kind_path,
                         is_collection => $is_collection,
                         search_children => $search_children,
                         expand_fields   => $call_flags->{expand_fields},
                         call_flags      => $call_flags,
                         dbh             => $self->{dbh},
                         logger          => $self->{logger},                        
                         filter          => $filter,
                         policy          => $objects->{policy},
                         self            => $self,);

    if (%rest_parameters) {        
        my ($link_base, $params) = split(/\?/, $rv->{selfLink}, 2);
        $link_base .= "?";
        $rv->{selfLink} = $link_base . F5::Utils::URI::build_query_string($rest_request->{rest_parameters} );

        if ($rest_request->{rest_parameters}->{'$top'}) {
            #We are in a paging result, so build previous and next links
            my $skip = $rest_request->{rest_parameters}->{'$skip'};
            $rest_request->{rest_parameters}->{'$skip'} += $rest_request->{rest_parameters}->{'$top'};
            if ($rest_request->{rest_parameters}->{'$skip'} < ($rv->{totalItems} || 0)) {
                $rv->{nextLink} = $link_base . F5::Utils::URI::build_query_string($rest_request->{rest_parameters});
            }
            $rest_request->{rest_parameters}->{'$skip'} -= 2*$rest_request->{rest_parameters}->{'$top'};
            if ($rest_request->{rest_parameters}->{'$skip'} >= 0) {
                delete $rest_request->{parameters}->{'$skip'} if !$rest_request->{rest_parameters}->{'$skip'};
                $rv->{previousLink} = $link_base . F5::Utils::URI::build_query_string($rest_request->{rest_parameters});
            }
            $rest_request->{rest_parameters}->{'$skip'} = $skip; 
        }
    }

    my $json = JSON->new->convert_blessed->allow_blessed;
    my $response_body = $rv;
    if (!$rest_request->{want_object}) {
        $response_body = $json->encode($rv);
        eval { $response_body = decode('utf8', $response_body) };
    }
    my $response_parameters = $path_pointer->{response_parameters} || {};
    if ($call_flags->{rest_response_parameters}) {
        $response_parameters = {%$response_parameters, %{$call_flags->{rest_response_parameters}}};
    }
    my $response = F5::ASMConfig::RestResponse->new(response_body => $response_body,
                                                    %$response_parameters);

    return {return_value => $response, events => []};
}

#------------------------------------------------------------------

sub build_odata_field_tree {
    my (%params) = @_;
    my $value = $params{'value'};
    my $tree = {};
    if (defined($value)) {
        for my $field (split(",",  $value)) {
            chomp($field);
            my ($field_part, $remainder) = split("/", $field, 2);

             $tree->{$field_part} ||= {};
             $tree->{$field_part} = { %{$tree->{$field_part}}, 
                                      %{build_odata_field_tree('value' => $remainder)},
                                    }; 
        }
   }

   return $tree;
}

#------------------------------------------------------------------
sub add_references {
    my (%params) = @_;
    my $rv = $params{rv};
    my $original_rv = $params{rv};
    my $path_pointer = $params{path_pointer};
    my $rest_request = $params{rest_request};
    my $is_collection  = $params{is_collection};
    my $search_children = $params{search_children};
    my $kind_path = $params{kind_path};
    my $expand_fields = $params{expand_fields};
    my $call_flags = $params{call_flags};
    my $self = $params{self};
    my $filter = $params{filter};
    my $policy = $params{policy};
    my $rest_selected_fields = $call_flags->{rest_selected_fields};

    my @references;
    if ($rv->{items} && scalar @{$rv->{items}}) {
        if ($path_pointer->{skip_items_on_modify_collection} && $rest_request->{rest_method} ne "GET") {
            $rv->{items} = [];
        }
        else {
            my $entity_type = ref($rv->{items}->[0]);
            if (!$path_pointer->{extract_policy_field} && $entity_type =~ /ASMConfig::Entity/ && !$search_children) {
                #Exclude objects (like RequestLog or a search_children call) where each item can have a different policy, from pre-caching.
                my $policy = $params{policy} || $rv->{items}->[0]->{policy};
                $entity_type->prepare_rest_reference_cache(%params,
                                                          rest_selected_fields => $rest_selected_fields,
                                                          policy => $policy,
                                                          objects => $rv->{items});
            }

            for my $item (@{$rv->{items}}) {
                $item = add_references(%params, rv => $item);
            }
        }
    }
    elsif ($search_children) {
        #replace parent policy from child links
        my $rest_path = $rest_request->{rest_path};
        my $replace_string = $policy ? "/$policy->{rest_uuid}"
                           : ($rv->{policy} && $rv->element_type != F5::ASMConfig::ElementType::POLICY()) ? "/$rv->{policy}->{rest_uuid}"
                           : '';
        $rest_path =~ s|/policies/[^/]*(.*)/children|/policies$replace_string$1|;
        $rest_request = {%$rest_request, rest_path => $rest_path};
    }

    my $host = ($rest_request->{rest_host} || "https://localhost");
    my $link_base = $host . '/mgmt/tm/asm';
    my $rest_uuid = $rv->{rest_uuid};
    if (ref($rv) =~ /ASMConfig/) {
        if (ref($rv) =~ /ASMConfig::Entity/) {
            #field translation here
            $rv = $rv->to_rest(rest_selected_fields => $rest_selected_fields,
                               expand_fields   => $expand_fields,                               
                               dbh => $params{dbh},
                               logger => $params{logger},
                               base_link => $link_base,
                               call_flags => $call_flags,
                               extract_policy_field => $path_pointer->{extract_policy_field},
                              );
        }
        else {
            $rv->remove_undefined_fields();
            for my $field (keys %$rv) {
                my $value = delete($rv->{$field});
                #To camelCase.  TODO: Thrift objects will need their own mapping.
                $field = F5::Utils::Rest::to_camel_case($field);
                $rv->{$field} = $value;
            }
        }
    }

    $link_base = $host . $rest_request->{rest_path};
    if (defined($rest_uuid) && length($rest_uuid) && $is_collection) {
        $link_base .= "/$rest_uuid";
    }
    $rv->{selfLink} ||= $link_base;
    $rv->{kind} = $call_flags->{kind};

    if (!$rv->{kind}) {
        my $kind_append = ($rv->{items} ? "collection" : "") . "state";
        my $kind_key = 'kind_' . $kind_append;
        my $kind;
        if (exists $path_pointer->{kind}) { # Special case based on: tools/rest_mapping_base.yaml
            $kind = $path_pointer->{kind};
        } elsif (exists $path_pointer->{$kind_key}) { # kind_state/kind_collectionstate are set by 'generate_rest_data.pl'. So let's choose the right one.
            $kind = $path_pointer->{$kind_key};
        }
        else {
            $kind = $kind_path;
            $kind =~ s|/|:|g;
            if ($path_pointer->{entity_type}) {
               $kind .= ":" . $path_pointer->{entity_type}->element_rest_name . $kind_append;
            }
        }
        $rv->{kind} = $kind;
    }
    $rv->{kind} = lc($rv->{kind});
    return $rv if $call_flags->{skipReferences} || $rv->{items};

    if ($path_pointer->{references}) {
        for my $reference (keys %{$path_pointer->{references}}) {
            next if ($path_pointer->{references}->{$reference}->{hidden_path});
            if ($path_pointer->{references}->{$reference}->{rest_deprecated_version} &&
                F5::BigipVersionUtils::compare_versions(version1 => $path_pointer->{references}->{$reference}->{rest_deprecated_version},
                                                        version2 => $call_flags->{rest_version}) <= 0) {
                next;
            }
            if ($policy && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT() &&
                $path_pointer->{references}->{$reference}->{entity_type} && 
                $path_pointer->{references}->{$reference}->{entity_type}->entity_config->{no_inheritance}) {
                next;
            }
            if (is_path_excluded(path_pointer => $path_pointer->{references}->{$reference},
                                 policy => $params{policy},
                                 call_flags => $call_flags,
                                 dbh => $self->{dbh})) {
                next;
            }

            my $reference_name = $path_pointer->{references}->{$reference}->{reference_name};
            my $skip_ref = 0;

            if ( my $dependent_fields = $path_pointer->{references}->{$reference}->{dependent_fields}){
                for my $dependent_field ( keys %$dependent_fields ){
                    if ( $original_rv->{$dependent_field} != $dependent_fields->{$dependent_field} ){
                        $skip_ref = 1;
                        last;
                    }
                }
            } 
            next if $skip_ref;

            if ($reference_name) {
                if ( ! %$rest_selected_fields || exists($rest_selected_fields->{$reference_name}) || exists($rest_selected_fields->{$reference}) || exists($rest_selected_fields->{'*'})) {
                    my $nested_filter = '';
                    if ($filter) {
                        my @expand_filters;
                        my $filters = F5::Utils::Rest::parse_odata_filter(filter => $filter);

                        for my $f (@$filters) {
                            if ($f->{field} =~ s/^$reference\///) {
                                my $value = defined($f->{value_list}) ? ( '(' . join (",", @{$f->{value_list}}) . ')')
                                                                      : $f->{value};
                                push (@expand_filters, "$f->{field} $f->{'cmp'} $value" );
                            }
                        }
                        
                        $nested_filter = join(' and ', @expand_filters);
                    }
                  
                    if ( $expand_fields && (exists($expand_fields->{'*'}) || exists($expand_fields->{$reference_name}) || exists($expand_fields->{$reference}))) {
                        my $call_flag_overrides = {search_filters => [], 
                                                   page_size => 0,
                                                   expand_fields => ($expand_fields->{$reference_name} || 
                                                                     $expand_fields->{$reference} || 
                                                                     $expand_fields->{'*'}),
                                                  };
                        if (%$rest_selected_fields) {
                            $call_flag_overrides->{'rest_selected_fields'} = ($rest_selected_fields->{$reference} ||
                                                                         $rest_selected_fields->{$reference_name} ||
                                                                         $rest_selected_fields->{'*'});
                        }
                        my $rest_parameters = {ver => $call_flags->{rest_version}};
                        if ($nested_filter) {
                            $rest_parameters->{'$filter'} = $nested_filter;
                        }
                        my ($host, $path);
                        if ($rv->{selfLink} =~ m|(https://[^/]*)([^?]*)| ) {
                            ($host, $path) = ($1, $2);
                        }
                        my $rest_request = { rest_path => "$path/$reference",
                                             rest_host => $host,
                                             rest_method => 'GET',
                                             want_object => 1,
                                             rest_parameters => $rest_parameters,                                           
                                           };
                        ##Clear paging and search_filters in call_flags?
                        my $expanded_result = $self->rest_call($self->{session}, 
                                                              {%$call_flags, %$call_flag_overrides}, 
                                                              $rest_request);                        
                        $rv->{$reference} = $expanded_result->{items} ? $expanded_result->{items} : $expanded_result;
                    }
                    else {
                        if ($nested_filter) {
                            $nested_filter = "?" . F5::Utils::URI::build_query_string({'$filter' => $nested_filter,
                                                                                       ver => $call_flags->{rest_version},
                                                                                      });
                        }
                        $rv->{$reference_name} =  { 'link' => "$link_base/$reference$nested_filter?ver=$call_flags->{rest_version}"};
                        if ( ! $path_pointer->{references}->{$reference}->{singleton}) {
                            $rv->{$reference_name}->{isSubCollection} = JSON::true;
                        }
                    }
                }
            }
            else {
               push (@references, { reference => { 'link' => "$link_base/$reference?ver=$call_flags->{rest_version}" } });
            }
        }
    }
    if (@references) {
        $rv->{items} = \@references;
    }

    return $rv;
}

#------------------------------------------------------------------

sub convert_search_result {
    my (%params) = @_;
    my $self = $params{self};
    my $search_result = $params{search_result};
    my $call_flags = $params{call_flags};
    my $element_type_name = $params{element_type_name};
    my $entity_class = $params{entity_class};
    my $policy = $params{policy};
    my $rest_result;

    ##Deal with paging and add total items and links from total_record_count and call_flags

    $search_result->{total_record_count} = 0 unless defined $search_result->{total_record_count};
  
    my @items = map { $_->{$element_type_name} } @{$search_result->{element_list}};
    $rest_result->{items} = \@items;
    if ($call_flags->{page_size}) {
        $rest_result->{itemsPerPage} = int($call_flags->{page_size});
        my $page_index = $call_flags->{start_page} || 1;
        $rest_result->{pageIndex} = int($page_index);
        $rest_result->{startIndex} = int((($page_index - 1)* $call_flags->{page_size}) + 1);
        $rest_result->{totalItems} = int($search_result->{total_record_count});
        $rest_result->{totalPages} = int($search_result->{total_record_count} / $call_flags->{page_size});
        if ($search_result->{total_record_count} % $call_flags->{page_size}) {
            $rest_result->{totalPages} += 1;
        }
    }
    elsif ($search_result) {
        $rest_result->{totalItems} = int($search_result->{total_record_count});
    }

    if ($entity_class->strict_inheritance() && $policy && $policy->{has_parent}) {
        my $section = F5::ASMConfig::Entity::Section->new(dbh => $self->{dbh},
                                                          policy => $policy,
                                                          get_criteria => {section_type => $entity_class->get_section_type()});
        $rest_result->{isInherited} = $section->is_inherited() ? JSON::true : JSON::false;
    }
    return $rest_result;
}

#------------------------------------------------------------------

sub convert_odata {
    my (%params) = @_;
    my $rest_request = $params{rest_request};
    my $call_flags = $params{call_flags};
    my $path_pointer = $params{path_pointer};
 
    my $parameters = $params{rest_parameters} || $rest_request->{rest_parameters};

    return unless $parameters;

    my $top = $parameters->{'$top'};
    my $skip = $parameters->{'$skip'};
    my $orderby = $parameters->{'$orderby'};
    my $groupby = $parameters->{'groupby'};
    my $aggregatefields = $parameters->{'aggregatefields'};
    my $server_time = $parameters->{'servertime'};
    my $filter_string = $parameters->{'$filter'};
    my $inline_count = $parameters->{'$inlinecount'};

    if ($inline_count) {
        $call_flags->{skip_total_record_count} = lc($inline_count) eq 'none' ? 1
                                                                             : 0;
    }

    if ($server_time) {
        $call_flags->{server_time} = lc($server_time) eq 'true' ? 1
                                                                : 0;
    }

    if ( $call_flags->{update_wildcard_entities} ) {
        $call_flags->{update_wildcard_entities} = lc($call_flags->{update_wildcard_entities}) eq 'true' ? 1
                                                                                                        : 0;
    }
    else {
        $call_flags->{update_wildcard_entities} = 1;
    }

    if ( $call_flags->{ignore_inherited_fields} ) {
        $call_flags->{ignore_inherited_fields} = lc($call_flags->{ignore_inherited_fields}) eq 'true' ? 1
                                                                                                              : 0;
    }

    if (defined($top)) {
        if ($top ne int($top) || $top <= 0) {
            die {
                error_code  => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                error_message => "Invalid \$top value '$top'. \$top must be a positive integer.",
                error_level  => "info",
                };
        }
        if ($path_pointer->{max_top} && $path_pointer->{max_top}->{$rest_request->{rest_method}} &&
            $top > $path_pointer->{max_top}->{$rest_request->{rest_method}}) {
              die {
                 error_code   => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                 error_message => "Invalid \$top value for $rest_request->{rest_method}. Max \$top is $path_pointer->{max_top}->{$rest_request->{rest_method}}.",
                 error_level  => "info",
                };
        }
        $call_flags->{page_size} = $top;
    }

    if ($skip && $top) {
        if ($skip ne int($skip) || $skip < 0) {
            die {
                error_code  => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                error_message => "Invalid \$skip value '$skip'. \$skip must be a positive integer.",
                error_level  => "info",
                };
        }
        if ($skip % $top) {
            die {
                error_code  => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                error_message => "Invalid \$skip value '$skip'. \$skip must be a multiple of \$top.",
                error_level  => "info",
                };
        }
        $call_flags->{start_page} = int(($skip / $top) + 1);
    }

    if (my $entity_type = $params{path_pointer}->{entity_type}) {
        my %reverse_rest_field_map = %{$entity_type->reverse_rest_field_map()};
        if ($groupby) {
            my $fields = convert_odata_to_parts(value => $groupby,
                                                split_regex => ',');
            for my $field (@$fields) {
                if (my $entity_field = $reverse_rest_field_map{$field}) {
                    my $db_field_name = $entity_type->fields->{$entity_field}->{db_field_name} || $entity_type->fields->{$entity_field}->{full_db_field_name};
                    next unless $db_field_name;
                    push (@{$call_flags->{grouping_fields}}, $db_field_name);
                }
            }
        }

        if ($aggregatefields) {
            my $aggregate_fields;
            eval {
                $aggregate_fields = F5::Utils::Rest::parse_aggregate_fields(aggregatefields => $aggregatefields);
            };
            if ($@) {
                die {
                     error_code   => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                     error_message => $@,
                     error_level  => "info",
                };
            }
            for my $field (keys %$aggregate_fields) {
                if (my $entity_field = $reverse_rest_field_map{$field}) {
                    for my $alias (keys %{$aggregate_fields->{$field}}) {
                        $reverse_rest_field_map{"__metadata/$alias"} = $alias;
                        $call_flags->{aggregate_fields}->{$entity_field}->{$alias} = $aggregate_fields->{$field}->{$alias};
                    }
                }
            }
        }

        if ($orderby) {
            $reverse_rest_field_map{'__metadata/groupbyCount'} = 'grouping_count';
            my $orders = convert_odata_to_parts(value => $orderby,
                                                split_regex => ',');
            push (@{$call_flags->{sort_orders}},@{convert_odata_sort_order(
                                                                          sort_orders => $orders,
                                                                          path_pointer => $path_pointer,
                                                                          reverse_rest_field_map => \%reverse_rest_field_map,
                                                                          dbh => $params{dbh},
                                                                          policy => $params{policy},)});
        }
    }

    if ($filter_string) {
        my $filters;
        eval {
            $filters = F5::Utils::Rest::parse_odata_filter(filter => $filter_string);
        };
        if ($@) {
            die {
                 error_code   => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                 error_message => $@,
                 error_level  => "info",
                };
        }
        push (@{$call_flags->{search_filters}}, @{convert_odata_filter(
                                                                      filters => $filters,
                                                                      path_pointer => $path_pointer,
                                                                      dbh => $params{dbh},
                                                                      policy => $params{policy},)});
    }
}

#------------------------------------------------------------------
sub convert_odata_to_parts {
    my (%params) = @_;
    my $value = $params{value};
    my $split_regex = $params{split_regex};

    $value =~ s/[()'"]/ /g;
    $value =~ s/\s+/ /g; #reduce multiple whitespace to one space
    $value =~ s/^ //; #remove leading space
    my @values = split(/$split_regex/i, $value);

    return \@values;
}

#------------------------------------------------------------------

sub convert_odata_sort_order {
    my (%params) = @_;
    my $orders = $params{sort_orders};
    my $path_pointer = $params{path_pointer};
    my $dbh = $params{dbh};
    my $policy = $params{policy};

    my $entity_type = $params{path_pointer}->{entity_type};
    my $sort_orders = [];

    my $rest_sort_orders = [];
    for my $order (@$orders) {
        my ($full_field, $direction) = split(/ /, $order);
        my ($field, $nested_fields) = split(/\//, $full_field, 2);
        if ($direction && $direction_map{lc($direction)}) {
            $direction = $direction_map{lc($direction)};
        }
        else {
            $direction = F5::ASMConfig::SortDirection::SORT_DIRECTION_ASC();
        }
        push (@$rest_sort_orders, { full_field => $full_field,
                                    direction => $direction,
                                    rest_field => $field,
                                    nested_fields => $nested_fields });
    }

    if ($rest_sort_orders) {
        push (@$sort_orders, @{$entity_type->convert_rest_orders(dbh => $dbh,
                                                                 policy => $params{policy},
                                                                 reverse_rest_field_map => $params{reverse_rest_field_map},
                                                                 rest_orders => $rest_sort_orders)});

    }

    return $sort_orders;
}

#------------------------------------------------------------------

sub convert_odata_filter {
    my (%params) = @_;
    my $filters = $params{filters};
    my $path_pointer = $params{path_pointer};
    my $dbh = $params{dbh};
    my $policy = $params{policy};

    my $entity_type = $params{path_pointer}->{entity_type};
    my $search_filters = [];

    my $field_filters;
    my $reference_filters;
    for my $filter (@$filters) {
        my $full_field = $filter->{field};
        my ($field, $nested_fields) = split(/\//, $full_field, 2);

        if ($path_pointer->{references} && $path_pointer->{references}->{$field}) {           
            push (@{$reference_filters->{$field}}, {%$filter, (field => $nested_fields)});
        }
        else {
            my $cmp = $filter->{'cmp'};
            my $value = defined($filter->{value_list}) ? $filter->{value_list} : $filter->{value};
            push (@{$field_filters->{$field}}, { full_field => $full_field,
                                                 'cmp' => $cmp,
                                                 value => $value,
                                                 rest_field => $field,
                                                 nested_fields => $nested_fields,
                                                 is_grouped => $filter->{is_grouped}, });
        }
    }

    if ($reference_filters) {
        for my $field (keys %$reference_filters) {
            my $nested_pointer = $path_pointer->{references}->{$field};
            my $grouping_field = $nested_pointer->{group_by} || $path_pointer->{child_group_by};

            my $nested_filters = convert_odata_filter(%params,
                                                      path_pointer => $nested_pointer,
                                                      filters => $reference_filters->{$field},);

            if ($nested_pointer->{entity_type}) {
                my @nested_entities = $nested_pointer->{entity_type}->retrieve_all(dbh => $dbh,
                                                                                   policy => $policy,
                                                                                   selected_fields => [ $grouping_field ],
                                                                                   grouping_fields => [$grouping_field],
                                                                                   search_filters => $nested_filters,);
                my @values = map { $_->{$grouping_field} } @nested_entities;
                push (@$search_filters,  F5::ASMConfig::SearchFilter->new(field_name => $path_pointer->{child_group_by},
                                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                                          field_value_list => \@values));
            }
            else {
                push (@$search_filters, @$nested_filters);
            }                             
        }
    }

    if ($field_filters) {
        push (@$search_filters, @{$entity_type->rest_filters_to_search_filters(dbh => $dbh,
                                                                               policy => $params{policy},
                                                                               rest_filters_by_field => $field_filters)});
    }
    return $search_filters;
}

#------------------------------------------------------------------

sub is_path_excluded {
    my (%params) = @_;
    my $path_pointer = $params{path_pointer};
    my $exclude_from_rest_func = $path_pointer->{exclude_from_rest_func};
    if ($path_pointer->{entity_type} && $path_pointer->{entity_type}->feature_name
          && !$exclude_from_rest_func) {
        $path_pointer->{exclude_from_rest_func} = $exclude_from_rest_func = 'is_feature_disabled';
        $path_pointer->{feature_name} = $path_pointer->{entity_type}->feature_name;
    }
    if ($exclude_from_rest_func) {
        my $entity_type = $path_pointer->{entity_type} || 'F5::ASMConfig::Entity::Base';
        if ($entity_type->$exclude_from_rest_func(dbh => $params{dbh},
                                                  field_def => $path_pointer,
                                                  policy => $params{policy})) {
            return 1;
        }
    }
    return 0;
}
#------------------------------------------------------------------

my $SEPARATOR = "<br/>";

sub generate_rest_api_doc {
    my ($self, $session, $call_flags) = @_;
    my $dbh = $self->{dbh};
    
    my @paths = ({path_pointer => $path_mapper,
                  uri => "mgmt/tm/asm"});
    my $kind_path = 'tm/asm';
    my $ver = $call_flags->{rest_version} || F5::BigipVersionUtils::bigip_version();

    my $doc;
    my @elements;
    #Header
    my $header = <<EndHeader;
<html>
<style>
td, th { vertical-align: top; padding: 3px }
table { margin: 5px; }
table, tr, td { border-collapse: collapse; border: 1px solid black; }
tr td:first-child { min-width: 150px }
</style>
<title>ASM REST API</title>
<h1>ASM REST API</h1>
<a name="top"/>
EndHeader

    while (my $path = shift @paths) {
        my $path_pointer = $path->{path_pointer};
        if (is_path_excluded(path_pointer => $path_pointer,
                             dbh => $self->{dbh})) {
            next;
        }

        my $uri = $path->{uri};
        push(@elements, document_path(path_pointer => $path_pointer,
                                      doc_type => 'html',
                                      dbh => $dbh,
                                      ver => $ver,
                                      uri => $uri));

        if ($path_pointer->{references}) {
            map {push(@paths, {path_pointer => $path_pointer->{references}->{$_},
                               uri => $uri . ($path_pointer->{entity_type} ? "/MD5HASH" : '') . "/$_",}
                     )} sort keys %{$path_pointer->{references}};
        }
    }
    my $footer = "\n</html>";
   
    $doc = $header . elements_to_table(\@elements) . $footer;
    return {return_value => $doc};
}


#----------------------------------------------------------------------

my $DOC_PATH = F5::Deploy::map_path('/usr/share/ts/share/asm_config/yaml/doc');
my $common_fields_desc = {}; 

sub get_descriptions {
    my(%params) = @_;
    my $uri = $params{uri};

    my @uri_parts = split '/', $uri;
    my ($filename) = @uri_parts;
    my $descriptions_file = F5::Deploy::map_path("$DOC_PATH/$filename.yaml");
    my $entity_desc = (-e $descriptions_file) ? YAML::XS::LoadFile($descriptions_file) : undef;

    if (!keys %$common_fields_desc) {
        for my $common_desc_file (glob("$DOC_PATH/common/*.yaml")) {
            my $common_field = (fileparse($common_desc_file, '\..*'))[0];
            $common_fields_desc->{"common/$common_field.yaml"} = YAML::XS::LoadFile($common_desc_file)->{$common_field};
        }
    }

    my $rv;
    if ($entity_desc) {
        if ($uri ne 'policy') {
            $entity_desc = $entity_desc->{policy};
        }
        $rv->{full_desc} = $entity_desc;
        while (my $part = shift @uri_parts) {
            $entity_desc = $entity_desc->{$part};

        }
        if (defined $entity_desc) {
            $entity_desc = {%$entity_desc, %$common_fields_desc};
        }
    }

    $rv->{entity_desc} = $entity_desc;
    return $rv;
}

#----------------------------------------------------------------------

my $POLICY_BASE_SCHEMA = F5::Deploy::map_path('/usr/share/ts/share/policy_schema_base.yaml');
my $POLICY_OVERLAY = F5::Deploy::map_path('/usr/share/ts/config/declarative_policy_overlay.yaml');
my %common_overlays = ();

sub generate_json_schema {
    my ($self, $session, $call_flags, $placeholder, $generate_json_opts) = @_;
    my $dbh = $self->{dbh};
    my $ver = $call_flags->{rest_version} || F5::BigipVersionUtils::bigip_version();
    my $generate_md = $generate_json_opts->{generate_md};
    my $generate_rst = $generate_json_opts->{generate_rst};

    my $rv = {};
    if ($generate_rst) {
        $rv->{collections} = generate_rst_collections(dbh => $self->{dbh});
    }

    my $doc;
    my $schema = YAML::XS::LoadFile($POLICY_BASE_SCHEMA);
    my @md_tables;
    my @rst_tables;

    my $refs_map = $path_mapper->{references}->{policies}->{references};

    my $overlay;
    if (-f $POLICY_OVERLAY) {
        $overlay = YAML::XS::LoadFile($POLICY_OVERLAY)->{policy};
        if (!F5::BigipUtils::is_feature_enabled('allow_staging')) {
            $common_overlays{performStaging} = JSON::false;
            $common_overlays{learn} = JSON::false;
        }

        for my $field (keys %$overlay) {
            if (my $overlay_reftype = reftype($overlay->{$field})) {
                if ($overlay_reftype eq 'HASH' && $overlay->{$field}->{__schema_ignore}) {
                    delete $schema->{properties}->{policy}->{properties}->{$field};
                }
            }
        }
    }

    my $populate_factory_values = $generate_rst;

    my $policy_desc = get_descriptions(uri => 'policy')->{entity_desc};

    my $entity_type = 'F5::ASMConfig::Entity::Policy';
    my $policy_doc = document_fields(entity_type => $entity_type,
                                     dbh => $dbh,
                                     expand_all => 0,
                                     ver => $ver,
                                     fields_desc => $policy_desc,
                                     overlay => $overlay,
                                     populate_factory_values => $populate_factory_values,
                                     doc_type => 'json_schema')->{json_schema_rv}; 

    $schema->{properties}->{policy}->{properties} = {%{$schema->{properties}->{policy}->{properties}}, %$policy_doc};

    my $fields = $entity_type->fields;
    my $entity_field_map = $entity_type->rest_field_map();
    my @policy_required =  map { $entity_field_map->{$_} } grep { $fields->{$_}->{declarative_required_field} } keys %$fields;
    $schema->{properties}->{policy}->{required} = \@policy_required;
    $schema->{properties}->{policy}->{type} = 'object';

    # handling template since in imported policy expected identifier is 'name' field which isn't the unique record part
    $entity_type = 'F5::ASMConfig::Entity::PolicyTemplate';
    my @template_fields = grep { $entity_type->fields->{$_}->{declarative_required_field} } keys %{$entity_type->fields}; 
    my %fields_def = map { $_ => $entity_type->fields->{$_ } } @template_fields;
    my $template_obj = document_fields(entity_type => $entity_type,
                                       dbh => $dbh,
                                       fields => \%fields_def,
                                       expand_all => 0,
                                       ver => $ver,
                                       fields_desc => $policy_desc->{template},
                                       overlay => $overlay->{template},
                                       populate_factory_values => $populate_factory_values,
                                       doc_type => 'json_schema')->{json_schema_rv};

    if (keys %$template_obj) {
        my @template_required = map { $entity_type->rest_field_map()->{$_} } @template_fields;
        $schema->{properties}->{policy}->{properties}->{template}->{required} = \@template_required;
        $schema->{properties}->{policy}->{properties}->{template}->{type} = 'object';
        $schema->{properties}->{policy}->{properties}->{template}->{properties} = $template_obj;
        if ($overlay && $overlay->{template} && $overlay->{template}->{__schema_overlay} && $overlay->{template}->{__schema_overlay}->{doc}) {
            $schema->{properties}->{policy}->{properties}->{template}->{description} = $overlay->{template}->{__schema_overlay}->{doc};
        }
        elsif ($policy_desc && $policy_desc->{template}->{__schema_def}->{doc}) {
            $schema->{properties}->{policy}->{properties}->{template}->{description} = $policy_desc->{template}->{__schema_def}->{doc};
        }
    }

    my $path_pointer = $path_mapper->{references}->{policies};
    my @paths = map { {path_pointer => $path_pointer->{references}->{$_}, uri => $_} } sort keys %{$path_pointer->{references}};

    while (my $path = shift @paths) {
        $path_pointer = $path->{path_pointer};
        my $uri = $path->{uri};
        my $base_uri = $path->{base_uri};
        my $path_overlay;

        if ($overlay) {
            $path_overlay = $base_uri ? $overlay->{$base_uri}->{$uri} : $overlay->{$uri};
            if ($path_overlay && $path_overlay->{__schema_ignore}) {
                next;
            }
        }

        my $rest_deprecated_version = $path_pointer->{rest_deprecated_version};
        if ($rest_deprecated_version &&
                (F5::BigipVersionUtils::compare_versions(version1 => $rest_deprecated_version, version2 => $ver) <= 0)) {
            next;
        }
        if (is_path_excluded(path_pointer => $path_pointer,
                             dbh => $self->{dbh})) {
            next;
        }

        my $is_collection = !$path_pointer->{singleton};
        my $descriptions = get_descriptions(uri => $base_uri? "$base_uri/$uri" : $uri);
        my $entity_desc = $descriptions->{entity_desc};

        my $resource_methods = $path_pointer->{allowed_methods};
        my $collection_methods = $path_pointer->{collection_allowed_methods} || $resource_methods;

        if (my $entity_type = $path_pointer->{rest_entity_type} || $path_pointer->{entity_type}) {
            if ( $collection_methods && !($collection_methods->{PATCH} || $collection_methods->{POST}) ) {
                next if ($is_collection && !$resource_methods->{PATCH});
            }

            if (my $declarative_handler_func = $entity_type->entity_config->{declarative_handler_func}) {
                next if $declarative_handler_func eq 'handle_excluded_entity';
            }

    	    $doc = document_fields(entity_type => $entity_type,
	    		           dbh => $dbh,
			           expand_all => 0,
			           ver => $ver,
                                   fields_desc => $entity_desc,
                                   overlay => $path_overlay,
                                   populate_factory_values => $populate_factory_values,
			           doc_type => 'json_schema')->{json_schema_rv};

            if ($generate_md) {
                push (@md_tables, generate_md_tables(uri => $uri,
                                                     link => $base_uri ? "policy/$base_uri" . "_$uri" : "policy/$uri",
                                                     fields => $doc,));
            }

            if ($generate_rst) {
                push (@rst_tables, generate_rst_tables(uri => $uri,
                                                       link => $base_uri ? "policy/$base_uri" . "_$uri" : "policy/$uri",
                                                       level => $base_uri ? 2 : 1,
                                                       fields => $doc,));
            }

            $doc = { type => 'object',
                     properties => $doc };

            if ($path_overlay && $path_overlay->{__schema_overlay} && $path_overlay->{__schema_overlay} && $path_overlay->{__schema_overlay}->{doc}) {
                $doc->{description} = $path_overlay->{__schema_overlay}->{doc};
            }
            elsif ($entity_desc && $entity_desc->{__schema_def}->{doc}) {
                $doc->{description} = $entity_desc->{__schema_def}->{doc};
            }

            $fields = $entity_type->fields;
            $entity_field_map = $entity_type->rest_field_map();

            if ($is_collection) {
                my ($declarative_link_field) = grep { $fields->{$_}->{declarative_link} } keys %$fields;
                if ($declarative_link_field) {
                    my $declarative_link_def = { format => 'uri',
                                                 type => 'string' };
                    $doc->{properties}->{'$ref'} = $declarative_link_def;
                    $doc->{properties}->{link} = $declarative_link_def;
                }

                if ( !defined $collection_methods || defined $collection_methods->{DELETE} ) {
                    $doc->{properties}->{'$action'} = { type => 'string',
                                                        enum => [ 'delete' ], };
                }

                my $entity_required_fields = $entity_type->declarative_required_fields();

                if (@$entity_required_fields) {
                    if ($declarative_link_field && $fields->{$declarative_link_field}->{unique_record_part}) {
                        my $declarative_link_rest_name = $entity_field_map->{$declarative_link_field};
                        my @non_link_required_fields = grep { $_ ne $declarative_link_rest_name } @$entity_required_fields;
                        my @required_structs = ( $entity_required_fields,
                                                [ @non_link_required_fields, '$ref' ],
                                                [ @non_link_required_fields, 'link' ] );
                        for my $required_struct (@required_structs) {
                            push (@{$doc->{oneOf}}, { required => $required_struct });
                        }
                    }
                    else {
                        my @required_fields = grep { exists $doc->{properties}->{$_} && !$doc->{properties}->{$_}->{exclude_from_required_fields} } @$entity_required_fields;
                        if (@required_fields) {
                            $doc->{required} = \@required_fields;
                        }
                    }
                }

                map { delete $doc->{properties}->{$_}->{exclude_from_required_fields} } keys %{$doc->{properties}};

                $doc = { type => 'array',
                         items => $doc };
            }

            $doc = { oneOf => [ $doc,
                                { '$ref' => '#/definitions/reference' }] };

            if ($base_uri) {
                my $schema_pointer = $schema->{properties}->{policy}->{properties};
                my $overlay_pointer = $overlay;
                my $desc_pointer = $descriptions->{full_desc};
                my @base_uri_parts = split '/', $base_uri;
                while (my $part = shift @base_uri_parts) {
                    if (!defined $schema_pointer->{$part}) {
                        $schema_pointer->{$part} = { type => 'object',
                                                     properties => {} };

                    }
                    if (defined $desc_pointer && defined $desc_pointer->{$part} && defined defined $desc_pointer->{$part}->{__schema_def}) {
                        $schema_pointer->{$part}->{description} = $desc_pointer->{$part}->{__schema_def}->{doc};
                        $desc_pointer = $desc_pointer->{$part};
                    }
                    if ($overlay_pointer) {
                        if ($overlay_pointer->{$part} && $overlay_pointer->{$part}->{__schema_overlay} && $overlay_pointer->{$part}->{__schema_overlay}->{doc}) {
                            $schema_pointer->{$part}->{description} = $overlay_pointer->{$part}->{__schema_overlay}->{doc};
                        }
                        $overlay_pointer = $overlay_pointer->{$part};
                    }
                    $schema_pointer = $schema_pointer->{$part}->{properties};
                }
                $schema_pointer->{$uri} = $doc;
            }
            else {
                $schema->{properties}->{policy}->{properties}->{$uri} = $doc;
            }

            if (($generate_md || $generate_rst) && !$base_uri) {
                $schema->{properties}->{policy}->{properties}->{$uri}->{oneOf}->[0]->{'$ref'} = "#/properties/policy/properties/$uri/items";
            }
        }
        elsif ($path_pointer->{references}) {
            map { push(@paths, {path_pointer => $path_pointer->{references}->{$_},
                                uri => $_,
                                base_uri => $base_uri? "$base_uri/$path->{uri}" : $path->{uri}}) } sort keys %{$path_pointer->{references}};

            if ($generate_md || $generate_rst) {
                my $references;
                for my $reference (sort keys %{$path_pointer->{references}}) {
                    my $reference_def = $path_pointer->{references}->{$reference};
                    if ($overlay && $overlay->{$uri} && $overlay->{$uri}->{$reference}
                         && $overlay->{$uri}->{$reference}->{__schema_ignore}) {
                        next;
                    }
                    if (defined $reference_def->{entity_type}
                         && !F5::YAML::Schema::Utils::is_supported_entity(entity_type => $reference_def->{entity_type}) ) {
                        next;
                    }
                    $references->{$reference}->{type} = 'object';
                    my $reference_desc = get_descriptions(uri => $base_uri? "$base_uri/$path->{uri}/$reference" : "$path->{uri}/$reference");
                    if ($overlay && $overlay->{$uri} && $overlay->{$uri}->{$reference} && $overlay->{$uri}->{$reference}->{__schema_overlay} && $overlay->{$uri}->{$reference}->{__schema_overlay}->{doc}) {
                        $references->{$reference}->{description} = $overlay->{$uri}->{$reference}->{__schema_overlay}->{doc};
                    }
                    elsif (defined $reference_desc->{entity_desc} && defined $reference_desc->{entity_desc}->{__schema_def}) {
                        $references->{$reference}->{description} = $reference_desc->{entity_desc}->{__schema_def}->{doc};
                    }
                    if (!$path_pointer->{references}->{$reference}->{singleton}) {
                        $references->{$reference}->{type} = 'array';
                        $references->{$reference}->{items}->{type} = 'object';
                    }
                    $references->{$reference}->{'$ref'} = $base_uri ? "#/properties/policy/properties/$base_uri/$uri" . "_$reference/items"
                                                                    : "#/properties/policy/properties/$uri" . "_$reference/items";
                }

                push (@md_tables, generate_md_tables(uri => $uri,
                                                     link => $base_uri ? "policy/$base_uri" . "_$uri" : "policy/$uri",
                                                     fields => $references,)) if $generate_md;

                push (@rst_tables, generate_rst_tables(uri => $uri,
                                                       level => 1,
                                                       is_refs_table => 1,
                                                       link => $base_uri ? "policy/$base_uri" . "_$uri" : "policy/$uri",
                                                       fields => $references,)) if $generate_rst;

                $schema->{properties}->{policy}->{properties}->{$uri} = { '$ref' => "#/properties/policy/properties/$uri/items",
                                                            type => 'object',
                                                            properties => {} } if !$base_uri;
            }
        }
    }

    if ($generate_md || $generate_rst) {
        unshift (@md_tables, generate_md_tables(uri => 'policy',
                                                link => 'policy',
                                                fields => $schema->{properties}->{policy}->{properties},) ) if $generate_md;
        unshift (@rst_tables, generate_rst_tables(uri => 'policy',
                                                  link => 'policy',
                                                  is_refs_table => 1,
                                                  fields => $schema->{properties}->{policy}->{properties},) ) if $generate_rst;
        for my $uri (keys %{$schema->{properties}->{policy}->{properties}}) {
            if (defined $schema->{properties}->{policy}->{properties}->{$uri}->{oneOf}) {
                delete $schema->{properties}->{policy}->{properties}->{$uri}->{oneOf}->[0]->{'$ref'};
            }
            else {
                delete $schema->{properties}->{policy}->{properties}->{$uri}->{'$ref'};
            }
        }
    }

    my $md_header = "<!--- GENERATED FILE - DO NOT MODIFY --->\n\n# Declarative Policy";

    $rv = { %$rv,
            json_rv => $schema,
            rst_rv => join('', @rst_tables),
            md_rv => $md_header . join("\n\n---\n", @md_tables), };

    return { return_value => $rv };
}

#----------------------------------------------------------------------

sub get_required_fields {
    my(%params) = @_;
    my $entity_type = $params{entity_type};
    my $fields = $entity_type->fields;

    my @required_fields = grep { $fields->{$_}->{declarative_required_field} } keys %$fields;
    if (!@required_fields) {
        my @unique_record_parts = grep { !$fields->{$_}->{exclude_from_rest} } @{$entity_type->unique_record_parts};
        if (my ($name_field) = grep { $fields->{$_}->{is_name_field} } @unique_record_parts) {
            @required_fields = ( $name_field );
        }
        else {
            @required_fields = @unique_record_parts;
        }
    }

    my @required_fields_rv;
    for my $field (@required_fields) {
        my $rest_field_map = $entity_type->rest_field_map();
        if ($fields->{$field}->{declarative_flat_reference}) {
            $rest_field_map = $fields->{$field}->{entity_type}->rest_field_map();
            ($field) = @{$fields->{$field}->{entity_type}->unique_record_parts};
        }
        push (@required_fields_rv, $rest_field_map->{$field});
    }

    return \@required_fields_rv;
}

#----------------------------------------------------------------------

my $VIOLATION_FIELDS = <<EndHeader;
   :widths: 4 5 8

   * - Violation Name
     - Title
     - Description
EndHeader

my $SUB_VIOLATION_FIELDS = <<EndHeader;
   :widths: 5 8

   * - Sub Violation Name
     - Description
EndHeader

my $SERVER_TECHNOLOGY_FIELDS = <<EndHeader;
   :widths: 4 8 5

   * - Server Technology Name
     - Description
     - Implied Technologies
EndHeader

my @rst_collections = (
                       {
                        entities => [ {entity_type => 'F5::ASMConfig::Entity::Violation',
                                       policy_entity_type => 'F5::ASMConfig::Entity::Blocking'} ],
                        selected_fields => [ qw(viol_name description viol_description_text) ],
                        rst_header_fields => $VIOLATION_FIELDS,
                       },
                       {
                        entities => [ { entity_type => 'F5::ASMConfig::Entity::NormalizationCode',
                                        policy_entity_type => 'F5::ASMConfig::Entity::Normalization',
                                        title => 'Evasion Techniques Sub-Violations'},
                                      { entity_type => 'F5::ASMConfig::Entity::WssSubviolationCode',
                                        policy_entity_type => 'F5::ASMConfig::Entity::WssSubviolation',
                                        title => 'WSS Sub-Violations'},
                                      { entity_type => 'F5::ASMConfig::Entity::HttpCheckCode',
                                        policy_entity_type => 'F5::ASMConfig::Entity::HttpCheck',
                                        title => 'HTTP Compliance Sub-Violations'} ],
                        selected_fields => [ qw(description description_text) ],
                        rst_header_fields => $SUB_VIOLATION_FIELDS,
                       },
                       {
                        entities => [ { entity_type => 'F5::ASMConfig::Entity::ServerTechnology',
                                        policy_entity_type => 'F5::ASMConfig::Entity::PolicyServerTechnology' } ],
                        selected_fields => [ qw(server_technology_name description implied_server_technologies) ],
                        rst_header_fields => $SERVER_TECHNOLOGY_FIELDS,
                       },
                      );

sub generate_rst_collections {
    my(%params) = @_;
    my $dbh = $params{dbh};

    my $rv;
    my $overlay;
    if (-f $POLICY_OVERLAY) {
        $overlay = YAML::XS::LoadFile($POLICY_OVERLAY)->{policy};
    }

    for my $collection (@rst_collections) {
        my $table_header = <<EndHeader;
.. list-table::
   :header-rows: 1
$collection->{rst_header_fields}
EndHeader

        for my $entity (@{$collection->{entities}}) {
            my $table = $table_header;
            my $entity_type = $entity->{entity_type};
            my $search_filters;
            if ($overlay) {
                my $policy_entity_type = $entity->{policy_entity_type};
                my $path = $policy_entity_type->entity_config->{rest_paths}->[0]->{rest_path};
                my @path_parts = split("/", $path);
                my $overlay_pointer = $overlay;
                while (my $part = shift @path_parts) {
                    if (defined $overlay_pointer->{$part}) {
                        $overlay_pointer = $overlay_pointer->{$part};
                    }
                }

                my ($entity_unique_field) = @{$entity_type->declarative_unique_record_parts};
                my $field_name = $entity_type->rest_field_map->{$entity_unique_field};

                if (defined $overlay_pointer && defined $overlay_pointer->{$field_name} && defined $overlay_pointer->{$field_name}->{__schema_overlay}) {
                    my $values;
                    my $field_comparison;
                    if (defined $overlay_pointer->{$field_name}->{__schema_overlay}->{allowed_values}) {
                        $values = $overlay_pointer->{$field_name}->{__schema_overlay}->{allowed_values};
                        $field_comparison = F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN();
                    }
                    elsif (defined $overlay_pointer->{$field_name}->{__schema_overlay}->{excluded_values}) {
                        $values = $overlay_pointer->{$field_name}->{__schema_overlay}->{excluded_values};
                        $field_comparison = F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN();
                    }
                    if ($field_comparison) {
                        $search_filters = [ F5::ASMConfig::SearchFilter->new(field_name => $entity_unique_field,
                                                                             field_comparison => $field_comparison,
                                                                             field_value_list => $values) ];
                    }
                }
            }

            my @values = $entity_type->retrieve_all(dbh => $dbh,
                                                    selected_fields => $collection->{selected_fields},
                                                    search_filters => $search_filters);

            for my $value (@values) {
                my @row_values = map { ref($value->{$_}) eq 'ARRAY' ? join(", ", @{$value->{$_}})
                                                                    : $value->{$_} || '' } @{$collection->{selected_fields}};

                for my $val (@row_values) {
                    $val =~ s/\n-/\n  -/ig;
                    $val =~ s/\n/\n       /ig;
                }
                $table = "$table   * - ";
                $table = $table . join("\n     - ", map { defined $_ ? $_ : '' } @row_values) . "\n";
            }
            my $title = defined $entity->{title} ? $entity->{title} : $entity_type->element_type_display_name_plural;
            $title = "$title Description";
            my $headline_title_size = length $title;
            my $overline = '=' x $headline_title_size;

            my $rst = "$title\n$overline\n\n\n$table";
            $rv->{$entity_type->element_type_name} = $rst;
        }
    }
    return $rv;
}

#----------------------------------------------------------------------

my $RST_TABLE_HEADER = <<EndHeader;
.. list-table::
   :header-rows: 1
   :widths: 5 1 8 3

   * - Field Name
     - Type
     - Description
     - Allowed Values
EndHeader

my $REF_TABLE_HEADER = <<EndHeader;
.. list-table::
   :header-rows: 1
   :widths: 4 1 1 8 3

   * - Field Name
     - Reference
     - Type
     - Description
     - Allowed Values
EndHeader

my @RST_TITLE_SIZE_CHAR = ('*', '=', '-', '^', '"');

sub generate_rst_tables {
    my(%params) = @_;
    my $uri = $params{uri};
    my $fields = $params{fields};
    my $sub_table = $params{sub_table};
    my $link = $params{link};
    my $level = $params{level} || 0;
    my $is_refs_table = $params{is_refs_table};

    my $headline_title_char = ($level >= @RST_TITLE_SIZE_CHAR) ? $RST_TITLE_SIZE_CHAR[-1] : $RST_TITLE_SIZE_CHAR[$level];
    my $headline_title_size = length $uri;
    my $overline = $headline_title_char x $headline_title_size;

    my $table_headline = "$overline\n$uri\n$overline";

    if ($link) {
        $table_headline = ".. _$link:\n\n$table_headline";
    }

    my @references_tables;
    my @rows;
    for my $field (sort keys %$fields) {
        my $field_def = $fields->{$field};
        if ($uri eq 'policy') {
            if (defined $field_def->{items} && defined $field_def->{items}->{oneOf}) {
                $field_def = { items => $field_def->{items}->{oneOf}->[0], type => 'array' };
            }
            elsif (defined $field_def->{oneOf}) {
                $field_def = $field_def->{oneOf}->[0];
            }
        }
        my $type = $field_def->{type};

        my $row_rv = generate_rst_row(uri => $uri,
                                      link => $link,
                                      field => $field,
                                      populate_factory_values => 1,
                                      field_def => $field_def,
                                      is_refs_table => $is_refs_table,
                                      );

        if ($row_rv->{sub_table_link}) {
            push (@references_tables, generate_rst_tables(uri => $field,
                                                          link => $row_rv->{sub_table_link},
                                                          fields => $type eq 'object' ? $field_def->{properties}
                                                                                      : $field_def->{items}->{properties},
                                                          level => $level + 1, ));
        }

        push(@rows, $row_rv->{row});
    }

    my $table = "\n\n$table_headline\n\n" .($is_refs_table ? $REF_TABLE_HEADER : $RST_TABLE_HEADER) . join('', @rows);

    if (@references_tables) {
        $table = $table . join('', @references_tables);
    }

    return $table;
}

#----------------------------------------------------------------------

sub generate_rst_row {
    my(%params) = @_;
    my $uri = $params{uri};
    my $link = $params{link};
    my $field = $params{field};
    my $field_def = $params{field_def};

    my $field_name;
    my $type = $field_def->{type};

    if (defined $field_def->{minimum}) {
        $type = "$type\n       minimum: " . ($field_def->{minimum}+0);
    }

    if (defined $field_def->{maximum}) {
        $type = "$type\n       maximum: " . ($field_def->{maximum}+0);
    }

    my $allowed_values = $field_def->{enum};
    if (!$allowed_values && $params{populate_factory_values}) {
        $allowed_values = $field_def->{factory_values};
    }
    delete $field_def->{factory_values};

    my $allowed_values_str = '';

    if (!$type) {
       $type = "";
       for my $schema (@{$field_def->{oneOf}}) {
           $type = "$type\n       * $schema->{type}";
           if (defined $schema->{minimum}) {
               $type = "$type\n         minimum: " . ($schema->{minimum}+0);
           }

           if (defined $schema->{maximum}) {
               $type = "$type\n         maximum: " . ($schema->{maximum}+0);
           }
       }
       for my $schema (@{$field_def->{oneOf}}) {
           if ($schema->{type} eq 'integer') {
               push (@$allowed_values, 'Integer values');
           }
           elsif (defined $schema->{enum}) {
               push (@$allowed_values, "\"$schema->{enum}[0]\"");
           }
       }
    }

    my $description = $field_def->{description} || '';
    if (!$description && ($type eq 'array' && defined $field_def->{items} && defined $field_def->{items}->{description})) {
        $description = $field_def->{items}->{description};
    }
    $description =~ s/\n-/\n  -/ig;
    $description =~ s/\n/\n       /ig;

    my $has_reference;
    my $sub_table_link;
    my $add_referenced_table;

    if ( ($type eq 'object') || ($type eq 'array' && defined $field_def->{items} && $field_def->{items}->{type} eq 'object')) {
        $add_referenced_table = 1;
    }
    if ($type eq 'array' && defined $field_def->{items}) {
        my $item_type = $field_def->{items}->{type} || 'object';
        $type = "array of " . $item_type .  "s";
    }

    if ($add_referenced_table) {
        $sub_table_link = "$uri/$field";
        if ($field_def->{'$ref'}) {
            my $ref = $field_def->{'$ref'};
            $ref =~ s/#\/properties\/policy\/properties\///ig;
            $ref =~ s/\/items//ig;
            $ref =~ s/\/oneOf\/0//ig;
            $sub_table_link = "policy/$ref";
        }
        elsif($link) {
            $sub_table_link = "$link/$field";
        }
        $field_name = "`$field <$sub_table_link" . "_>`_";

        if (!$field_def->{'$ref'}) {
            $has_reference = 1;
        }
    }
    else {
        $field_name = "``$field``";
        if ($allowed_values) {
            $allowed_values_str = join("\n       ", map { $_ eq '*' ? "* \\$_" : "* $_" } @$allowed_values);
        }
    }

    my $is_reference = '';
    if ($params{is_refs_table}) {
        $is_reference = '     - ' . (defined $add_referenced_table ? 'Yes' : 'No');
    }

    my $row = <<EndRow;
   * - $field_name
$is_reference
     - $type
     - $description
     - $allowed_values_str
EndRow

    return { row => $row,
             sub_table_link => $has_reference ? $sub_table_link : undef,
           }
}

#----------------------------------------------------------------------

my $MD_EVAL_RST = "```eval_rst\n";
my $MD_END_RST = "```";

sub generate_md_tables {
    my(%params) = @_;
    my $uri = $params{uri};
    my $fields = $params{fields};
    my $sub_table = $params{sub_table};
    my $link = $params{link};

    my $headline_size_mark = $sub_table ? "#####" : "###";
    my $md_headline = "\n\n$headline_size_mark $uri\n\n";
    if ($link) {
        $md_headline = "\n\n$MD_EVAL_RST.. _$link:\n$MD_END_RST\n$md_headline\n\n"
    }

    my @references_tables;
    my @rows;
    for my $field (sort keys %$fields) {
        my $field_name;
        my $field_def = $fields->{$field};
        if ($uri eq 'policy') {
            if (defined $field_def->{items} && defined $field_def->{items}->{oneOf}) {
                $field_def = { items => $field_def->{items}->{oneOf}->[0], type => 'array' };
            }
            elsif (defined $field_def->{oneOf}) {
                $field_def = $field_def->{oneOf}->[0];
            }
        }

        my $type = $field_def->{type};

        my $row_rv = generate_rst_row(uri => $uri,
                                      link => $link,
                                      field => $field,
                                      field_def => $field_def,
                                      );

        if ($row_rv->{sub_table_link}) {
            push (@references_tables, generate_md_tables(uri => $field,
                                                         link => $row_rv->{sub_table_link},
                                                         fields => $type eq 'object' ? $field_def->{properties}
                                                                                     : $field_def->{items}->{properties},
                                                         sub_table => 1, ));
        }

        push(@rows, $row_rv->{row});
    }

    my $table =  $md_headline . $MD_EVAL_RST . $RST_TABLE_HEADER . join('', @rows) . $MD_END_RST;

    if (@references_tables) {
        $table = $table . join('', @references_tables);
    }

    return $table;
}

#----------------------------------------------------------------------

sub generate_openapi_doc {
    my ($self, $session, $call_flags, $generate_openapi_opts) = @_;
    my $dbh = $self->{dbh};
    
    my @paths = ({path_pointer => $path_mapper,
                  uri => "mgmt/tm/asm"});
    my $kind_path = 'tm/asm';
    my $ver = $call_flags->{rest_version} || F5::BigipVersionUtils::bigip_version();

    my $doc;
    my %schemas;

    while (my $path = shift @paths) {
        my $path_pointer = $path->{path_pointer};
        my $uri = $path->{uri};
        my $element_lines = document_path(path_pointer => $path_pointer,
                                          doc_type => 'openapi',
                                          dbh => $dbh,
                                          ver => $ver,
                                          uri => $uri);
        my $element_def;
        for my $element_line (@$element_lines) {
            $element_def->{$element_line->{'values'}->[0]} = $element_line->{'values'}->[1];
        }
        my $resource_name = $element_def->{'Resource Name'};
        my $resource_display_name = $resource_name;
        my $resource_id = '';
        if ($resource_name) {
            $resource_id = $resource_name . 'Id';
        }
        if (my $entity_type = ($path_pointer->{rest_entity_type} || $path_pointer->{entity_type})) {
            my $openapi_doc = { openapi => '3.0.0',
                                info => { title => $resource_name,
                                          description => $resource_name,
                                          version => '0.0.1',
                                        },
                                servers => [ { 'url' => 'http://localhost:443' } ],
                              };
            my $fields = document_fields(entity_type => ($path_pointer->{rest_entity_type} || $path_pointer->{entity_type}),
                                         dbh => $dbh,
                                         expand_all => 0,
                                         ver => $ver,
                                         doc_type => 'openapi')->{openapi_rv};
            $openapi_doc->{components}->{schemas}->{$resource_name}->{properties} = $fields;

            my $paths_def = {};
            if ($element_def->{Type} eq 'Collection') {
                my $list_result_def = { items => { 'type' => 'array',
                                                   'items' => { '$ref' => "#/components/schemas/$resource_name" },
                                                 },
                                        totalItems => { 'type' => 'integer' },
                                      };
                my $resource_list_name = $resource_name . '-list';
                $openapi_doc->{components}->{schemas}->{$resource_list_name}->{properties} = $list_result_def;

                my $collection_uri = '/' . $element_def->{'Namespace URI'};
                for my $collection_method (split(/$SEPARATOR/, lc $element_def->{'Collection Allowed Methods'})) {
                    my $method_def = {};
                    my %bulk_method_map = ('get' => 'search',
                                           'patch' => 'bulk update',
                                           'delete' => 'bulk delete');
                    if (exists $bulk_method_map{$collection_method}) {
                        $method_def->{summary} = ucfirst($bulk_method_map{$collection_method}) . " $resource_display_name";
                        $method_def->{description} = ucfirst($bulk_method_map{$collection_method}) . " $resource_display_name";
                        $method_def->{operationId} = "$bulk_method_map{$collection_method}-$resource_name";

                        $method_def->{responses}->{200}->{content}->{'application/json'}->{schema}->{'$ref'} = "#/components/schemas/$resource_list_name";
                        $method_def->{responses}->{200}->{description} = "Successfully executed $bulk_method_map{$collection_method} for $resource_name";
                    }
                    if ($collection_method eq 'patch') {
                        $method_def->{requestBody}->{content}->{'application/json'}->{schema}->{'$ref'} = "#/components/schemas/$resource_name";
                    }

                    if ($collection_method eq 'post') {
                        $method_def->{summary} = "Create new $resource_display_name";
                        $method_def->{description} = "Create new $resource_display_name";
                        $method_def->{operationId} = "create-$resource_name";
                        $method_def->{requestBody}->{content}->{'application/json'}->{schema}->{'$ref'} = "#/components/schemas/$resource_name";
                        $method_def->{responses}->{201}->{content}->{'application/json'}->{schema}->{'$ref'} = "#/components/schemas/$resource_name";
                        $method_def->{responses}->{201}->{description} = "Successfully added $resource_name";
                    }

                    if (%$method_def) {
                        $paths_def->{$collection_uri}->{$collection_method} = $method_def;
                    }
                }
            }
            my $resource_uri;
            if ($element_def->{Type} eq 'Collection') {
                $resource_uri = '/' . $element_def->{'Namespace URI'} . "/{" . $resource_id . "}";
            }
            else {
                $resource_uri = '/' . $element_def->{'Namespace URI'};
            }
            for my $resource_method (split(/$SEPARATOR/, lc $element_def->{'Resource Allowed Methods'})) {
                my $method_def = {};
                my %method_map = ('get' => 'retrieve',
                                  'patch' => 'update',
                                  'delete' => 'delete');
                if (exists $method_map{$resource_method}) {
                    $method_def->{summary} = ucfirst($method_map{$resource_method}) . " existing $resource_display_name";
                    $method_def->{description} = ucfirst($method_map{$resource_method}) . " existing new $resource_display_name";
                    $method_def->{operationId} = "$method_map{$resource_method}-$resource_name";
                    $method_def->{responses}->{200}->{content}->{'application/json'}->{schema}->{'$ref'} = "#/components/schemas/$resource_name";
                    $method_def->{responses}->{200}->{description} = "Successfully $method_map{$resource_method}d $resource_name";
                }
                if ($resource_method eq 'patch') {
                    $method_def->{requestBody}->{content}->{'application/json'}->{schema}->{'$ref'} = "#/components/schemas/$resource_name";
                }
                if (%$method_def) {
                    $paths_def->{$resource_uri}->{$resource_method} = $method_def;
                }
            }

            for my $path_parameter ($resource_uri =~ /\{([-\w]+?)\}/g) {
                push(@{$paths_def->{$resource_uri}->{parameters}},
                     {
                      'name' => $path_parameter,
                      'in' => 'path',
                      'required' => JSON::true,
                      'schema' => { 'type' => 'string' }});
            }
            $openapi_doc->{paths} = $paths_def;
            $schemas{$element_def->{'Namespace URI'}} = $openapi_doc;
        }

        if ($path_pointer->{references}) {
            my $entity_id_path_part = '';
            if ($path_pointer->{entity_type}) {
                my $entity_type = $path_pointer->{entity_type};
                my $entity_name = $entity_type->element_rest_name();
                $entity_id_path_part = '/{' . $entity_name . 'Id}';
            }
            map {push(@paths, {path_pointer => $path_pointer->{references}->{$_},
                               uri => $uri . ($path_pointer->{entity_type} ? $entity_id_path_part  : '') . "/$_",}
                     )} sort keys %{$path_pointer->{references}};
        }
    }

    return {return_value => \%schemas};
}

#------------------------------------------------------------------

sub elements_to_table {
    my ($elements_ref) = @_;
    my $body = '';
    my @index;
    my @tables;
    for my $element (@$elements_ref) {
        my $table = "<table border=1>";
        for my $row (@$element) {
            if ($row->{'index'}) {
                $table = "<hr/><a name=\"$row->{'index'}\"/><a href=\"#top$row->{'index'}\"><h3>$row->{'index'}</h3></a>" . $table;
                push (@index, "<li><a name =\"top$row->{'index'}\"/><a href=\"#$row->{'index'}\">$row->{'index'}</a></li>");
            }
            my $cell_open = $row->{header} ? "<th>" : "<td>";
            my $cell_close = $row->{header} ? "</th>" : "</td>";
            $table .= "<tr>" . join('', map { "$cell_open $_ $cell_close \n" } @{$row->{'values'}}) . "</tr>\n";
        }
        $table .="</table>\n";
        push (@tables, $table);
    }
    $body .= join('', @index);
    $body .= join($SEPARATOR, @tables);
    return $body;
}

#------------------------------------------------------------------

sub rest_doc {
    my (%params) = @_;
    my $path_pointer = $params{path_pointer};
    my $entity_type = $params{entity_type} || $path_pointer->{entity_type};
    my $rest_parameters = $params{rest_parameters} || {};
    my $policy = $params{policy};
    my $path = $params{path};

    my $rest_selected_fields = $params{rest_selected_fields} || build_odata_field_tree('value' => $rest_parameters->{'$select'});
    my $expand_fields   = $params{expand_fields}   || build_odata_field_tree('value' => $rest_parameters->{'$expand'});
    my $base_link = "https://localhost/mgmt/tm/asm";
    my $ver = $rest_parameters->{ver};

 
    my $doc;
    if ($entity_type) {
        my $self_link_params = {};
        if ($path_pointer->{extern_fields}) {
            %$self_link_params = map { ( $_ => {rest_uuid => 'example'} ) } keys %{$path_pointer->{extern_fields}};
        }
        if ($path_pointer->{literal_fields}) {
             %$self_link_params = (%$self_link_params, %{$path_pointer->{literal_fields}});
        }

        $doc = document_fields(entity_type => $entity_type,
                               dbh => $params{dbh},
                               expand_fields => $expand_fields,
                               rest_selected_fields => $rest_selected_fields,
                               policy => $policy,
                               base_link => $base_link,
                               self_link_params => $self_link_params,
                               ver => $ver,
                               doc_type => 'rest')->{rest_doc};
    }

    if ($path_pointer->{references}) {
        for my $reference (keys %{$path_pointer->{references}}) {
            next if ($path_pointer->{references}->{$reference}->{hidden_path});
            if (is_path_excluded(path_pointer => $path_pointer->{references}->{$reference},
                                 policy => $policy,
                                 dbh => $params{dbh})) {
                next;
            }
            my $reference_name = $path_pointer->{references}->{$reference}->{reference_name};
            if (%$rest_selected_fields && !($rest_selected_fields->{$reference} || $rest_selected_fields->{$reference_name} || $rest_selected_fields->{'*'})) {
                next;
            }
            if ($expand_fields->{$reference} || $expand_fields->{$reference_name} || $expand_fields->{'*'}) {
                my $new_path = "$path/$reference";
                if ($path_pointer->{references}->{$reference}->{entity_type}) {
                    $new_path .= "/example";
                }
                $doc->{$reference} = rest_doc(%params,
                                              path_pointer => $path_pointer->{references}->{$reference},
                                              expand_fields => ($expand_fields->{$reference} || $expand_fields->{$reference_name} || $expand_fields->{'*'}),
                                              rest_selected_fields => ($rest_selected_fields->{$reference} || $rest_selected_fields->{$reference_name} || $rest_selected_fields->{'*'}),
                                              path => "$new_path",
                                              want_object => 1,                                              
                                             );
            }
            else {
                $doc->{$reference_name || $reference} = { 'link' => "$base_link/$path/$reference?ver=$ver" };
            }
        }
    } 

    if ($params{want_object}) {
        return $doc;
    }

    my $json = JSON->new->convert_blessed->allow_blessed->canonical;
    my $response = decode('utf8', $json->encode($doc));
    return {return_value => $response, events => []};
}

#------------------------------------------------------------------

sub document_path {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $path_pointer = $params{path_pointer};
    my $uri = $params{uri};
    my $doc_type = $params{doc_type};
    my $ver = $params{ver};

    my @element;
    ##Name Space Entry
    ### URI
    push (@element, {'values' => ["Namespace URI", $uri],
                     'index' => $uri});
    #### Allowed Methods (Collection, Entity)
    
    my $collection_methods = $path_pointer->{collection_allowed_methods} || $path_pointer->{allowed_methods} || $collection_allowed_methods;
    if (my $entity_type = ($path_pointer->{rest_entity_type} || $path_pointer->{entity_type})) {
        my $resource_methods = $path_pointer->{allowed_methods} || $resource_allowed_methods;
        push (@element, {'values' => ["Type", $path_pointer->{singleton} ? "Resource" : "Collection"]});
        push (@element, {'values' => ["Resource Name", $entity_type->element_rest_name()]});
        push (@element, {'values' => ["Response Parameters", encode_json($path_pointer->{response_parameters})]}) if $path_pointer->{response_parameters};
        push (@element, {'values' => ["Resource Allowed Methods", join($SEPARATOR, keys %$resource_methods)]});
        push (@element, {'values' => ["Collection Allowed Methods", $path_pointer->{singleton} ? "N/A" : join($SEPARATOR, keys %$collection_methods)]});

        my $fields_table;
        if ($doc_type ne 'openapi') {
            $fields_table = document_fields(entity_type => $entity_type,
                                            dbh => $dbh,
                                            expand_all => 0,
                                            ver => $ver,
                                            doc_type => $doc_type)->{html_elements};
        }
        push (@element, {'values' => ["Fields", elements_to_table([$fields_table])]});
    }
    else {
        push (@element, {'values' => ["Type", "Namespace"]});
        push (@element, {'values' => ["Response Parameters", encode_json($path_pointer->{response_parameters})]}) if $path_pointer->{response_parameters};
        push (@element, {'values' => ["Allowed Methods", join($SEPARATOR, keys %$collection_methods)]});
    }
    #### Sub Name Spaces (with expanded names too)
    if ($path_pointer->{references}) {
        my @reference_elements = ({header => 1, 'values' => [qw(reference-name expanded-name)]});
        for my $reference (keys %{$path_pointer->{references}}) {
            if (is_path_excluded(path_pointer => $path_pointer->{references}->{$reference},
                                 dbh => $dbh)) {
                next;
            }
            my @values;
            if ($path_pointer->{references}->{$reference}->{reference_name}) {
               @values = ($path_pointer->{references}->{$reference}->{reference_name}, $reference);
            }
            else {
               @values = ($reference, "N/A");
            }
            push (@reference_elements, {'values' => \@values});
        }
        push (@element, {'values' => ["Sub-References", elements_to_table([\@reference_elements])]});
    }

    #### Entity Name + kind (if applicable)
    ##### Fields
    
    ##### Rest name, type (if applicable), mappings (if applicable)
    ##### If Reference print link and expanded name
    return \@element;
}

#------------------------------------------------------------------

sub document_fields {
    my(%params) = @_;
    my $entity_type = $params{entity_type};
    my $dbh = $params{dbh};
    my $doc_type = $params{doc_type};
    my $expand_all = $params{expand_all} || 0;
    my $fields = $params{fields} || $entity_type->example_fields;
    my $expand_fields = $params{expand_fields} || {};
    my $rest_selected_fields = $params{rest_selected_fields} || {};
    my $default_object = $params{default_object};
    my $policy = $params{policy};
    my $base_link = $params{base_link} || '';
    my $self_link_params = $params{self_link_params} || {};
    my $ver = $params{ver};
    my $fields_desc = $params{fields_desc};
    my $overlay = $params{overlay};
    my $populate_factory_values = $params{populate_factory_values};

    if (!$default_object) {
        $default_object = $entity_type->new(dbh => $dbh, preload_default => 1, all_fields => 1, policy => $policy, rest_call => 1);
    }


    my @html_elements = ({header => 1, 'values' => [qw(name type values/sub-fields)]});
    my $rest_doc = {};
    my $openapi_rv = {};
    my $json_schema_rv = {};

    my %REST_TYPE_MAP = ( 'int' => 'int',
                          'boolean' => 'boolean',
                          'float' => 'float',
                          'hex' => 'hexadecimal',
                          'timestamp' => 'rfc3339-date',
                          'datetime' => 'rfc3339-date',
                          'last_update_field' => 'microseconds',
                          'last_update_field_from_datetime' => 'microseconds',
                        );

    my $now = F5::DateTimeUtils::convert_epoch_to_datetime(time);
    my %DEFAULT_VALUES_MAP = ('int' => 0,
                              boolean => 0,
                              reverse_boolean => 1,
                              float => 0,
                              'hex' => 0,
                              'timestamp' => 0,
                              datetime => $now,
                              last_update_field => 0,
                              'last_update_field_from_datetime' => $now,
                             );
                               

    my %HTML_TYPE_MAP = ( 'int' => 'Integer',
                          'boolean' => 'Boolean',
                          'reverse_boolean' => 'Boolean',
                          'value_map' => 'Enum',
                          'float' => 'Float',
                          'hex' => 'Hexadecimal',
                          'timestamp' => 'RFC3339 Date',
                          'datetime' => 'RFC3339 Date',
                          'last_update_field' => 'Microseconds',
                          'last_update_field_from_datetime' => 'Microseconds',
                          'tightening_mode' => 'Enum',
                          'is_wildcard' => 'Enum',
                          'Struct' => 'Struct', 
                        );

    my %OPENAPI_TYPE_MAP = (
                            'int' => 'integer',
                            'boolean' => 'boolean',
                            'reverse_boolean' => 'boolean',
                            'value_map' => 'string', 
                            'float' => 'number',
                            'hex' => { type => 'string',
                                       description => 'hexidecimal value',
                                     },
                            'timestamp' => { type => 'string',
                                             format => 'date-time',
                                           },
                            'datetime' => { type => 'string',
                                            format => 'date-time',
                                          },
                            'last_update_field' => { type => 'integer',
                                                     description => 'Microseconds since epoch',
                                                   },
                            'last_update_field_from_datetime' => { type => 'integer',
                                                                   description => 'Microseconds since epoch',
                                                                 },
                            'tightening_mode' => 'string', # todo: should be enum
                            'is_wildcard' => 'string',
                            'Struct' => 'object', 
                           );

    my %json_schema_excludes = ( 
                                %$F5::ASMConfig::Entity::Base::common_diff_excludes,
                                rest_uuid => 1,
                                flg_in_classification => 1,
                                is_default => 1,
                                is_referenced => 1,
                                wildcard_sequence => 0,
                               );

    my $rest_field_map = $params{rest_field_map} || $entity_type->rest_field_map();
    my $field_value_types = $params{field_value_types} || $entity_type->field_value_types();

    my $groups = {};
    for my $field (sort keys %$fields) {
        my $field_def = $fields->{$field};

        if ($doc_type eq 'json_schema' && ($field_def->{extra_field} || $field_def->{exclude_from_declarative}) ) { 
            next;
        }

        if($field_def->{exclude_from_rest} || $field_def->{exclude_from_rest_example} || ($field_def->{rest_deprecated_version} && F5::BigipVersionUtils::compare_versions(version1 => $field_def->{rest_deprecated_version},
                                                                                                                                version2 => $ver) <= 0) ) {
            next;
        }

        if (my $exclude_from_rest_func = ($field_def->{exclude_from_rest_func} || $field_def->{rest_deprecated_func}) ) {
            if ($default_object->$exclude_from_rest_func(dbh => $params{dbh},
                                                         field_def => $field_def,
                                                         policy => $params{policy})) {
                next;
            }
        }

        if ($field_def->{rest_group} && !$params{group}) {
            if ( ! %$rest_selected_fields || exists($rest_selected_fields->{$field_def->{rest_group}}) || exists($rest_selected_fields->{'*'})) {
                $groups->{$field_def->{rest_group}}->{$field} = $field_def;
            }
            next;
        }

        if (my $rest_name = ($rest_field_map->{$field} || $field_def->{rest_field_name})) {
            my $value_type = $field_def->{rest_value_type} || $field_value_types->{$field} || '';
            my $expand_name = $field_def->{rest_expand_name} || '';

            if ( %$rest_selected_fields && !(exists($rest_selected_fields->{$rest_name}) || exists($rest_selected_fields->{$expand_name}) || exists($rest_selected_fields->{'*'}) )) {
                next;
            }

            my $field_overlay;
            if (exists $overlay->{$rest_name} || exists $common_overlays{$rest_name}) {
                $field_overlay = exists $overlay->{$rest_name} ? $overlay->{$rest_name} : $common_overlays{$rest_name};
                next if !defined $field_overlay;
                my $overlay_reftype = reftype($field_overlay);
                if (!$overlay_reftype || $overlay_reftype eq 'SCALAR'
                      || ($overlay_reftype eq 'HASH' && $field_overlay->{__schema_ignore}) ) {
                    if (my $dependent_fields_map = $entity_type->entity_config->{dependent_fields}->{$field}) {
                        my $field_value = $field_overlay;
                        if ($field_def->{rest_value_map}) {
                            my %value_map = reverse %{$field_def->{rest_value_map}};
                            $field_value = $value_map{$field_value};
                        }
                        for my $dependent_fields_value (keys %$dependent_fields_map) {
                            next if $dependent_fields_value eq $field_value;
                            my $dependent_fields = $dependent_fields_map->{$dependent_fields_value};
                            for my $dependent_field (@$dependent_fields) {
                                my $dependent_field_name = $rest_field_map->{$dependent_field} || $fields->{$dependent_field}->{rest_field_name};
                                next if !$dependent_field_name;
                                delete $json_schema_rv->{$dependent_field_name};
                                $json_schema_excludes{$dependent_field} = 1;
                            }
                        }
                    }
                    next if !$field_def->{unique_record_part};
                }
            }

            next if ($doc_type eq 'json_schema') && $json_schema_excludes{$field};

            my $html_obj = { type => $HTML_TYPE_MAP{$value_type} || 'String',
                             name => ($expand_name ? $expand_name . $SEPARATOR . "<b>Expands</b> to: $rest_name" : $rest_name),
                             value => $SEPARATOR,                             
                           };
            my $rest_default_value = defined($field_def->{rest_default_value}) ? $field_def->{rest_default_value} :
              defined ($field_def->{default_value}) ? $field_def->{default_value} :
              $default_object->{$field};
            my $rest_obj = { name => $rest_name,
                             type => $REST_TYPE_MAP{$value_type} || 'string',
                             value => $rest_default_value,
                           };
            if ($doc_type eq 'openapi' || $doc_type eq 'json_schema') {
                my $openapi_type = $OPENAPI_TYPE_MAP{$value_type} || 'string';
                if (!ref($openapi_type)) {
                    $openapi_type = { type => $openapi_type };
                }

                if ($doc_type eq 'openapi') {
                    for my $k (keys %$openapi_type) {
                        $rest_obj->{$k} = $openapi_type->{$k};
                    }
                }

                $rest_obj->{type} = $openapi_type->{type};
                if ($doc_type eq 'json_schema' && ($rest_obj->{type} eq 'integer' || $rest_obj->{type} eq 'number')) {
                    if (defined $field_def->{validate_min_val}) {
                        $rest_obj->{minimum} = $field_def->{validate_min_val} + 0;
                    }
                    if (defined $field_def->{validate_max_val}) {
                        $rest_obj->{maximum} = $field_def->{validate_max_val} + 0;
                    }
                    if (my $func = $field_def->{rest_value_func}) {
                        if (defined $rest_obj->{minimum}) {
                            $rest_obj->{minimum} = $entity_type->$func(field_value => $rest_obj->{minimum}, direction => 'to_rest');
                        }
                        if (defined $rest_obj->{maximum}) {
                            $rest_obj->{maximum} = $entity_type->$func(field_value => $rest_obj->{maximum}, direction => 'to_rest');
                        }
                    }
                }
            }

            my $map_rest_value_func = $field_def->{rest_value_func};
            my %struct_value_func = ("map_rest_external_struct" => 1,
                                     "map_rest_defense_attributes" => 1,);


            if ($expand_name && !$expand_all && !($expand_fields->{$rest_name} || $expand_fields->{$expand_name} || $expand_fields->{'*'}) ) {
                $rest_obj->{name} = $expand_name;
                my $expand_type = $field_def->{rest_entity_type} || $field_def->{entity_type};
                my %expand_object_args = ( object => {rest_uuid => "example"} );
                if ($expand_type) {
                    if ($expand_type->policy_id_field()) {
                        $expand_object_args{policy} = $policy || { rest_uuid => "example" };
                    }
                    my $external_reference = $expand_type->new(%expand_object_args, %$self_link_params)->self_link(base_link => $base_link,
                                                                                                                   ver => $ver,
                                                                                                                   dbh => $dbh);
                    if ($doc_type eq 'openapi') {
                        $rest_obj->{type} = 'object';
                        $rest_obj->{properties}->{link}->{type} = 'string';
                        $rest_obj->{properties}->{link}->{description} = "reference to $expand_type";
                        $rest_obj->{properties}->{link}->{default} = $external_reference;
                    }
                    elsif ($doc_type eq 'json_schema') {
                        if (my $declarative_handler_func = $expand_type->entity_config->{declarative_handler_func}) {
                            next if $declarative_handler_func eq 'handle_excluded_entity';
                        }

                        my $type_reference_name = $expand_type->element_type_name_plural;

                        $rest_obj->{type} = 'object';
                        $rest_obj->{name} = $field_def->{rest_field_name} || $entity_type->rest_field_map->{$field};

                        my $expand_type_path_pointer = $path_mapper->{references}->{policies}->{references}->{$expand_type->element_rest_path_name};
                        if ( $expand_type->policy_id_field() && $expand_type_path_pointer
                                && !$expand_type_path_pointer->{hidden_path} && !$field_def->{declarative_flat_reference} ) {
                            my $type_reference_name = $expand_type->element_rest_path_name;
                            $rest_obj->{'$ref'} = "#/properties/policy/properties/$type_reference_name/oneOf/0/items";
                        }
                        else {
                            my @expand_unique_record_parts = grep { !$expand_type->fields->{$_}->{exclude_from_rest} && !$expand_type->fields->{$_}->{exclude_from_declarative} } @{$expand_type->declarative_unique_record_parts};
                            my @declarative_required_fields = grep { $expand_type->fields->{$_}->{declarative_required_field} } keys %{$expand_type->fields};

                            my @expand_type_fields = (@expand_unique_record_parts, @declarative_required_fields);

                            if ($field_def->{declarative_flat_reference}) {
                                @expand_type_fields = (@expand_unique_record_parts);
                            }

                            my %expand_type_fields_def = map { $_ => $expand_type->fields->{$_} } @expand_type_fields;
			    my $reference_doc = document_fields(fields => \%expand_type_fields_def,
			                                        entity_type => $expand_type,
                                                                doc_type => $doc_type,
						                policy => $policy,
                                                                populate_factory_values => $populate_factory_values,
						                base_link => $base_link,
                                                                fields_desc => $fields_desc->{$rest_obj->{name}} || $fields_desc,
						                ver => $ver,
						                dbh => $dbh,);

                            my $expand_type_doc = $reference_doc->{json_schema_rv};
                            next if !%$expand_type_doc;

                            $rest_obj->{properties} = $expand_type_doc;
                            my @required_fields = map { $expand_type->rest_field_map()->{$_} } @expand_unique_record_parts;
                            my $set_unique_field = !$expand_type->policy_id_field();
                            if ( defined $field_def->{declarative_flat_reference} ) {
                                $set_unique_field = $field_def->{declarative_flat_reference};
                            }
                            if ($set_unique_field) {
                                $rest_obj->{declarative_unique_field} = \@required_fields;
                            }

                            if (@declarative_required_fields) {
                                $rest_obj->{exclude_from_required_fields} = 1;
                            }
                        }
                    }
                    else {
                        $rest_obj->{value} = { "link" => $external_reference };
                    }
                    $html_obj->{name} = $expand_name;
                    $html_obj->{type} = 'Struct';
                }
            }
            elsif ((my $nested_entity_type = ($field_def->{rest_entity_type} || $field_def->{entity_type})) &&
                   (!$map_rest_value_func || $struct_value_func{$map_rest_value_func})) {
                my $nested_fields = ($map_rest_value_func && $map_rest_value_func eq "map_rest_external_struct") ? {%{$nested_entity_type->fields}}
                                                                                                                 : $field_def->{fields} || {%{$nested_entity_type->fields}};
                my %nested_args;
                if ($map_rest_value_func && $map_rest_value_func eq "map_rest_defense_attributes") {
                    $default_object->populate_defense_attributes();
                    my $defense_attributes = $entity_type->get_defense_attributes(dbh => $dbh);
                    my $defenses = $default_object->{$field};
                    for my $defense (@$defenses) {
                        my $defense_attribute = $defense_attributes->{$defense->{def_attrib_id}};
                        $nested_args{rest_field_map}->{$defense->{def_attrib_id}} = F5::Utils::Rest::to_camel_case($defense->{def_attrib_name});
                        $defense_attribute->{default_value} = $defense->{def_attrib_value};
                        if ($defense->{value_type} == F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER()) {  
                            $nested_args{field_value_types}->{$defense->{def_attrib_id}} = 'int';
                            $defense_attribute->{rest_value_map} = {0 => 'any'};
                            if ( ! $defense->{flg_check_attrib}) {
                                $defense_attribute->{default_value} = 'any'; 
                            }
                            if (defined $defense->{min_value}) {
                                $defense_attribute->{validate_min_val} = $defense->{min_value};
                            }
                            if (defined $defense->{max_value}) {
                                $defense_attribute->{validate_max_val} = $defense->{max_value};
                            }
                        }
                        elsif ($defense->{value_type} == F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_BOOLEAN()) {
                            $nested_args{field_value_types}->{$defense->{def_attrib_id}} = 'boolean';
                        }                        
                    }
                    $nested_fields = $defense_attributes;
                }
                elsif ( ! $field_def->{external_struct}) {
                    my %nested_field_map = map { ( $_ => ($nested_fields->{$_}->{rest_field_name} || '') ) } keys %$nested_fields;
                    my %nested_field_types = map { ( $_ => ($nested_fields->{$_}->{rest_value_type} || '') ) } keys %$nested_fields;
                    $nested_args{rest_field_map} = \%nested_field_map;
                    $nested_args{field_value_types} = \%nested_field_types;
                }
                if ($nested_entity_type !~ /::Entity::/) {
                        $nested_args{entity_type} = $entity_type;
                }

                my $nested_desc;
                if ($fields_desc && $fields_desc->{$rest_name}) {
                    my $nested_field_desc = $fields_desc->{$rest_name}->{__schema_def}->{items};
                    if (my $referenced_doc = $nested_field_desc->{'$ref'}) {
                        $nested_desc = $fields_desc->{$referenced_doc};
                    }
                    else {
                        $nested_desc = $fields_desc->{$rest_name};
                    }
                }

                my $nested_doc = document_fields(fields => $nested_fields,
                                          entity_type => $nested_entity_type,
                                          expand_fields => ($expand_fields->{$rest_name} || $expand_fields->{$expand_name} || $expand_fields->{'*'}),
                                          rest_selected_fields => ($rest_selected_fields->{$rest_name} || $rest_selected_fields->{$expand_name} || $rest_selected_fields->{'*'}),
                                          expand_all => $expand_all,
                                          doc_type => $doc_type,
                                          policy => $policy,
                                          base_link => $base_link,
                                          fields_desc => $nested_desc,
                                          populate_factory_values => $populate_factory_values,
                                          %nested_args,
                                          ver => $ver,
                                          dbh => $dbh,);
                $html_obj->{type} = 'Struct';
                $html_obj->{value} = elements_to_table([$nested_doc->{html_elements}]);
                if ($doc_type eq 'openapi' || $doc_type eq 'json_schema') {
                    $rest_obj->{type} = 'object';
                    $rest_obj->{properties} = $nested_doc->{$doc_type . "_rv"};
                    if ($doc_type eq 'json_schema') {
                        my @unique_record_parts =  grep { $nested_fields->{$_}->{declarative_required_field} } keys %$nested_fields;
                        if (!@unique_record_parts) {
                            @unique_record_parts = grep { !$nested_fields->{$_}->{exclude_from_rest} && $nested_fields->{$_}->{unique_record_part} } keys %$nested_fields;
                        }
                        my @required_fields = map { $nested_args{rest_field_map}->{$_} || $nested_entity_type->rest_field_map->{$_} } @unique_record_parts;
                        @required_fields = grep { exists $rest_obj->{properties}->{$_} } @required_fields;
                        if (@required_fields) {
                             $rest_obj->{required} = \@required_fields;
                        }
                    }
                }
                else {
                    $rest_obj->{value} = $nested_doc->{rest_doc};
                }
            }
            elsif (my %map = %{$entity_type->get_rest_map(dbh => $dbh,
                                                          field_def => $field_def,
                                                          value_type => $value_type)}) {
                if ($field_def->{validate_allowed_values}) {
                    %map = map { ($_ => $map{$_} ) } @{$field_def->{validate_allowed_values}};
                }
                $html_obj->{value} = join($SEPARATOR, values %map);
                if ($rest_obj->{type} ne "boolean" && $rest_obj->{type} ne 'rfc3339-date' && $rest_obj->{name} ne 'enabled') {
                    my %reverse_map = reverse %map;
                    if ($doc_type eq 'openapi' || $doc_type eq 'json_schema') {
                        $rest_obj->{enum} = [sort keys %reverse_map];  #Using keys of reverse map (instead of values) to filter out duplicates
                    }
                    else {
                        $rest_doc->{$rest_obj->{name} . "Enums"} = [sort keys %reverse_map];  #Using keys of reverse map (instead of values) to filter out duplicates
                        if ($rest_obj->{type} eq 'int') {
                            push (@{$rest_doc->{$rest_obj->{name} . "Enums"}}, "integer values");
                        }
                        $rest_obj->{type} = 'enum';
                    }
                    if ($doc_type eq 'json_schema') {
                        if ($rest_obj->{type} eq 'integer') {
                            my $integer_schema = { type => 'integer'};
                            if (defined $rest_obj->{minimum}) {
                                $integer_schema->{minimum} = $rest_obj->{minimum};
                                delete $rest_obj->{minimum};
                            }
                            if (defined $rest_obj->{maximum}) {
                                $integer_schema->{maximum} = $rest_obj->{maximum};
                                delete $rest_obj->{maximum};
                            }
                            $rest_obj->{oneOf} = [
                                                  $integer_schema,
                                                  { type => 'string',
                                                    enum => $rest_obj->{enum} },
                                                  { type => 'string',
                                                    pattern => '\\d+$' },
                                                 ];
                            delete $rest_obj->{enum};
                        }
                        elsif (exists $field_def->{validate_foreign_value_excludes}) {
                            push( @{$rest_obj->{enum}}, @{$field_def->{validate_foreign_value_excludes}});
                        }
                    }
                }

                if (!defined($rest_obj->{value}) && $rest_obj->{type} eq "boolean") {
                   $rest_obj->{value} = 0;
                }

                if (defined($rest_obj->{value}) && defined($map{$rest_obj->{value}})) {
                    $rest_obj->{value} = $map{$rest_obj->{value}};                    
                }
                elsif (!defined($rest_obj->{value}) && $rest_obj->{type} ne 'rfc3339-date') {
                    ($rest_obj->{value}) = (values %map);
                }
            }
            elsif ($doc_type eq 'json_schema' && $field_def->{populate_declarative_values}) {
                my @field_values = $entity_type->retrieve_all(dbh => $dbh,
                                                              selected_fields => [ $field ],);
                my @allowed_values = map { $_->{$field} } @field_values;
                $rest_obj->{enum} = [ sort @allowed_values ];
            }
            elsif ($populate_factory_values && $field_def->{populate_factory_values}) {
                my @field_values = $entity_type->retrieve_all(dbh => $dbh,
                                                              selected_fields => [ $field ],);
                my @allowed_values = map { $_->{$field} } @field_values;
                $rest_obj->{factory_values} = [ sort @allowed_values ];
            }
            else {
                if (!defined $rest_obj->{value}) {
                     $rest_obj->{value} = defined($DEFAULT_VALUES_MAP{$value_type}) ? $DEFAULT_VALUES_MAP{$value_type} : '';
                }
                $rest_obj->{value} = $default_object->map_rest_field_value(field_def => $field_def,
                                                                           field_value => $rest_obj->{value},
                                                                           direction => 'to_rest',
                                                                           dbh => $dbh,
                                                                           field_name => $field,
                                                                          );
            }

            if ($doc_type ne 'openapi') {
                # todo: ????
                if ($field_def->{check_base64}) {
                    $rest_doc->{$rest_obj->{name} . 'Base64Encoded'} = $entity_type->get_rest_map(dbh => $dbh, value_type => 'boolean')->{0};
                }
            }

            if ($field_def->{nested_struct} && !$field_def->{map_rest_nested_as_single}) {
                $html_obj->{type} .= "[]";
                if (!ref($rest_obj->{value}) || ref($rest_obj->{value}) ne "ARRAY" ) {
                    $rest_obj->{value} = [$rest_obj->{value}];
                }
                if ($doc_type eq 'openapi' || $doc_type eq 'json_schema') {
                    my $orig_type = $rest_obj->{type};
                    $rest_obj->{type} = 'array';
                    $rest_obj->{items}->{type} = $orig_type;
                    if ($rest_obj->{properties}) {
                        $rest_obj->{items}->{properties} = $rest_obj->{properties};
                    }
                    if ($rest_obj->{required}) {
                        $rest_obj->{items}->{required} = $rest_obj->{required};
                    }
                    delete $rest_obj->{required};
                    delete $rest_obj->{properties};
                    delete $rest_obj->{declarative_unique_field};
                }
            }

            my $openapi_obj = { %$rest_obj };
            delete $openapi_obj->{name};
            if ($openapi_obj->{type} ne 'array' && $openapi_obj->{type} ne 'object') {
                $openapi_obj->{default} = $openapi_obj->{value};
            }
            delete $openapi_obj->{value};
            $openapi_rv->{$rest_obj->{name}} = $openapi_obj;

            if (defined $rest_obj->{oneOf}) {
                delete $rest_obj->{type};
            }

            my $field_name = $rest_obj->{name};
            if ($field_overlay && reftype($field_overlay) && (reftype($field_overlay) eq 'HASH') && $field_overlay->{__schema_overlay} && $field_overlay->{__schema_overlay}->{doc}) {
                $rest_obj->{description} = $field_overlay->{__schema_overlay}->{doc};
            }
            elsif ($fields_desc && $fields_desc->{$field_name} && $fields_desc->{$field_name}->{__schema_def}->{doc}) {
                $rest_obj->{description} = $fields_desc->{$field_name}->{__schema_def}->{doc};
            }


            if (defined $rest_obj->{declarative_unique_field}) {
                my @declarative_unique_fields = @{$rest_obj->{declarative_unique_field}};
                delete $rest_obj->{declarative_unique_field};
                for my $field_name (@declarative_unique_fields) {
                    if (!defined $json_schema_rv->{$field_name}) {
                        $json_schema_rv->{$field_name} = $rest_obj->{properties}->{$field_name};
                    }
                    if ($fields_desc && $fields_desc->{$field_name} && $fields_desc->{$field_name}->{__schema_def}->{doc}) {
                        $json_schema_rv->{$field_name}->{description} = $fields_desc->{$field_name}->{__schema_def}->{doc};
                    }
                    if (!defined $field_overlay) {
                        $field_overlay = $overlay->{$field_name};
                    }
                    delete $rest_obj->{properties}->{$field_name};
                }

                if (keys %{$rest_obj->{properties}}) {
                    $json_schema_rv->{$field_name} = { %$rest_obj };
                    delete $json_schema_rv->{$field_name}->{name};
                    delete $json_schema_rv->{$field_name}->{value};
                }

                if(scalar @declarative_unique_fields == 1) {
                    ($field_name) = @declarative_unique_fields;
                }
            }
            else {
                my $json_schema_obj = { %$rest_obj };
                delete $json_schema_obj->{name};
                delete $json_schema_obj->{value};
                $json_schema_rv->{$field_name} = $json_schema_obj;
            }

            if (defined $field_overlay) {
                my $overlay_reftype = reftype($field_overlay);
                if ($overlay_reftype && $overlay_reftype eq 'HASH') {
                    if (my $schema_overlay = $field_overlay->{__schema_overlay} ) {
                        if ($schema_overlay->{allowed_values}) {
                            $json_schema_rv->{$field_name}->{enum} = $schema_overlay->{allowed_values};
                            $json_schema_rv->{$field_name}->{type} = 'string';
                        }
                        elsif ($schema_overlay->{excluded_values}) {
                            my %overlay_excluded_values = map { $_ => 1 } @{$schema_overlay->{excluded_values}};
                            my @allowed_values = grep { !$overlay_excluded_values{$_} } @{$json_schema_rv->{$field_name}->{enum}};
                            $json_schema_rv->{$field_name}->{enum} = \@allowed_values;
                            $json_schema_rv->{$field_name}->{type} = 'string';
                        }
                    }
                }
            }
            $rest_doc->{$rest_obj->{name}} = $rest_obj->{value};

            push (@html_elements, {'values' => [$html_obj->{name}, $html_obj->{type}, $html_obj->{value}]}) if $doc_type eq 'html';
       }
    }
    for my $group (sort keys %$groups) {
       if ($overlay && exists $overlay->{$group}) {
           my $group_overlay = $overlay->{$group};
           if (!defined $group_overlay || (reftype($group_overlay) && reftype($group_overlay) eq 'HASH' && $group_overlay->{__schema_ignore})) {
               next;
           } 
       }

        my $nested_doc = document_fields(fields => $groups->{$group},
                                         entity_type => $entity_type,
                                         doc_type => $doc_type,
                                         expand_all => $expand_all,
                                         expand_fields => ($expand_fields->{$group} || $expand_fields->{'*'}),
                                         rest_selected_fields => ($rest_selected_fields->{$group} || $rest_selected_fields->{'*'}),
                                         group => 1,
                                         base_link => $base_link,
                                         fields_desc => $fields_desc ? $fields_desc->{$group} : undef,
                                         overlay => ($overlay && exists $overlay->{$group}) ? $overlay->{$group} : undef,
                                         populate_factory_values => $populate_factory_values,
                                         policy => $policy,
                                         ver => $ver,
                                         dbh => $dbh,);
        push (@html_elements, {'values' => [$group, "Struct",  elements_to_table([$nested_doc->{html_elements}])]}) if $doc_type eq 'html';

        my $group_obj = { type => 'object',
                          properties => $nested_doc->{$doc_type . "_rv"} };

        if ($fields_desc && $fields_desc->{$group}) {
            $group_obj->{description} = $fields_desc->{$group}->{__schema_def}->{doc};
        }

        $openapi_rv->{$group} = $group_obj;
        $json_schema_rv->{$group} = $group_obj;

        $rest_doc->{$group} = $nested_doc->{rest_doc};
    }

    return {html_elements => \@html_elements,
            rest_doc => $rest_doc,
            openapi_rv => $openapi_rv,
            json_schema_rv => $json_schema_rv,
           };
}

#------------------------------------------------------------------
1;
