package F5::ASMConfig::Handler::ScannerActions;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use YAML::XS;

use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Handler::ScannerVuln;
use F5::Scanner;
use F5::BigipUtils;

my %threat_and_severity = (
    urgent => F5::ASMConfig::SeverityLevel::SEVERITY_URGENT(),
    critical => F5::ASMConfig::SeverityLevel::SEVERITY_CRITICAL(),
    high => F5::ASMConfig::SeverityLevel::SEVERITY_HIGH(),
    medium => F5::ASMConfig::SeverityLevel::SEVERITY_MEDIUM(),
    low => F5::ASMConfig::SeverityLevel::SEVERITY_LOW(),
    info => F5::ASMConfig::SeverityLevel::SEVERITY_INFO(),
);

my %status = (
    open => F5::ASMConfig::ScannerVulnStatus::SCANNER_VULN_STATUS_OPEN(),
    closed => F5::ASMConfig::ScannerVulnStatus::SCANNER_VULN_STATUS_CLOSED(),
);

our @handled_functions = (
                           {
                             function_name => 'update_scanner_vuln_data',
                             relayable => 1,
                           },
                           {
                             function_name => 'get_scanner_vuln_url',
                             relayable => 0,
                           },
                           {
                             function_name => 'get_scanner_auth_url',
                             relayable => 0,
                           },
                           {
                             function_name => 'get_global_scanner_auth_url',
                             relayable => 0,
                           },
                           {
                             function_name => 'get_scanner_register_url',
                             relayable => 0,
                           },
                           {
                             function_name => 'get_global_scanner_register_url',
                             relayable => 0,
                           },
                           {
                             function_name => 'complete_scanner_auth',
                             relayable => 1,
                           },
                           {
                             function_name => 'complete_global_scanner_auth',
                             relayable => 1,
                           },
                           {
                             function_name => 'get_scanner_subscriptions',
                             relayable => 0,
                           },
                           {
                             function_name => 'enqueue_scanner_scan',
                             relayable => 0,
                           },
                           {
                             function_name => 'terminate_scanner_scan',
                             relayable => 0,
                           },
                           {
                             function_name => 'get_register_url',
                             relayable => 0,
                           },
                           {
                             function_name => 'get_scanner_more_info_url',
                             relayable => 0,
                           },
                           {
                             function_name => 'get_global_scanner_more_info_url',
                             relayable => 0,
                           },
                           {
                             function_name => 'get_scanner_more_info_url_with_subscription',
                             relayable => 0,
                           },
                          );

#-------------------------------------------------------------------------

sub get_scanner_auth_url {
    my ($self, $session, $call_flags, $policy, $redirect_url) = @_;
    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    my $url = $scanner->get_scanner_auth_url(bigip_redirect_url => $redirect_url);

    return { return_value => $url };
}

#-------------------------------------------------------------------------

sub get_global_scanner_auth_url {
    my ($self, $session, $call_flags, $place_holder, $redirect_url, $global_scanner_config) = @_;

    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   scanner_type => $global_scanner_config->{scanner_type},
                                   call_flags => $call_flags,
                                  );

    my $url = $scanner->get_scanner_auth_url(bigip_redirect_url => $redirect_url, scanner_type => $global_scanner_config->{scanner_type});

    return { return_value => $url };
}

#-------------------------------------------------------------------------

sub get_scanner_register_url {
    my ($self, $session, $call_flags, $policy, $redirect_url) = @_;
    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    my $url = $scanner->get_scanner_register_url(bigip_redirect_url => $redirect_url);

    return { return_value => $url };
}

#-------------------------------------------------------------------------

sub get_global_scanner_register_url {
    my ($self, $session, $call_flags, $policy, $redirect_url, $global_scanner_config) = @_;
    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   scanner_type => $global_scanner_config->{scanner_type});

    my $url = $scanner->get_scanner_register_url(bigip_redirect_url => $redirect_url, scanner_type => $global_scanner_config->{scanner_type});

    return { return_value => $url };
}

#-------------------------------------------------------------------------

sub complete_scanner_auth {
    my ($self, $session, $call_flags, $policy, $qs_map) = @_;

    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    my $retval = $scanner->complete_scanner_auth(qs_map => $qs_map);
    return { return_value => $retval };
}

#-------------------------------------------------------------------------

sub complete_global_scanner_auth {
    my ($self, $session, $call_flags, $place_holder, $qs_map) = @_;


    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                  );

    my $retval = $scanner->complete_global_scanner_auth(qs_map => $qs_map, client => $self);

    $self->update_many_scanner_configs(
            $session,
            $call_flags,
            F5::ASMConfig::Policy->new(),
            F5::ASMConfig::ScannerConfig->new(
                scanner_type => 'hailstorm',),
            F5::ASMConfig::ScannerConfig->new(
                api_key => '',
                scanner_site_name => '',
                flg_is_api_key_final => 0,
                flg_is_api_key_encrypted => 0,
                scanner_auth_data_1 => '',
                scanner_auth_data_2 => ''),);

    return { return_value => $retval };
}

#-------------------------------------------------------------------------

sub get_scanner_subscriptions {
    my ($self, $session, $call_flags, $policy, $scanner_config) = @_;

    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    my $scanner_subscriptions = $scanner->get_scanner_subscriptions(scanner_config => $scanner_config);

    my @subscriptions;

    for my $subscription (@$scanner_subscriptions) { 
        my @scans;

        for my $scan (@{$subscription->{subscription_scans}}) {
            push (@scans, F5::ASMConfig::ScannerSubscriptionScan->new($scan));
        }

        push (@subscriptions, F5::ASMConfig::ScannerSubscription->new({
                                  subscription_id => $subscription->{subscription_id}, 
                                  url => $subscription->{url}, 
                                  product_id => $subscription->{product_id}, 
                                  subscription_scans => \@scans}));
    }

    return { return_value => \@subscriptions };
}

#-------------------------------------------------------------------------

sub enqueue_scanner_scan {
    my ($self, $session, $call_flags, $policy, $subsciption_id) = @_;

    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);
    
    my $retval;
    eval {
        $retval = $scanner->enqueue_scanner_scan(subscription_id => $subsciption_id);
    };
    if ($@) {
        return $self->handle_error( error_code   => F5::ASMConfig::ErrorCode::SCANNER_SERVICE_ERROR(),
                                    error_thrown => $@,
                                    display_message => $@,
                                    error_level  => 'info' );
    }
    
    return { return_value => $retval };
}

#-------------------------------------------------------------------------

sub terminate_scanner_scan {
    my ($self, $session, $call_flags, $policy) = @_;

    my $scanner_config = $self->get_scanner_config($session, $call_flags, $policy, F5::ASMConfig::ScannerConfig->new,);
    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    my $retval = $scanner->terminate_scanner_scan(subscription_id => $scanner_config->{scanner_auth_data_1});

    return { return_value => $retval };
}

#-------------------------------------------------------------------------

sub get_scanner_vuln_url {
    my ($self, $session, $call_flags, $policy, $scanner_actions_fields) = @_;

    my $scanner_config = $self->get_scanner_config($session, $call_flags, $policy, F5::ASMConfig::ScannerConfig->new,);

   
    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    my $url = $scanner->get_full_url(
        vuln_ids => $scanner_actions_fields->{scanner_vuln_ids},
        url_type => 'vuln_status_url'
    );
    
    return { return_value => $url, events => []};
}

#-------------------------------------------------------------------------

sub update_scanner_vuln_data {
    my ($self, $session, $call_flags, $policy, $scanner_actions_fields) = @_;

    my $xml_data = $scanner_actions_fields->{scanner_vuln_data_xml};
    my $xml_file_path = $scanner_actions_fields->{xml_file_path};
    my $vuln_ids;
    my @events;

    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self, 
                                   session => $session, 
                                   call_flags => $call_flags, 
                                   policy => $policy,);

    my $call_flags_for_count = F5::ASMConfig::CallFlags->new;
    $call_flags_for_count->{grouping_fields} = [ 'scanner_vuln_id' ];
    
    my $retval;
    eval {
        $retval = $scanner->parse_xml(events => \@events, 
                                      scanner_actions_fields => $scanner_actions_fields);
    };
    if ($@) {
        if (ref($@) && $@->{error_message}) {
            if ($@->{error_code} == $F5::Scanner::CANT_DOWNLOAD_VULNS_FROM_SCANNER) {
                my $error_msg = 'Cannot process the scanner vulnerability data';
                return $self->handle_error(action => 'add',
                                   error_code   => F5::ASMConfig::ErrorCode::VULN_DOWNLOAD_ERROR(),
                                   error_thrown => $error_msg,
                                   display_message => $error_msg,
                                   error_level  => 'info',
                                  );
            }
            else {
                die $@->{error_message}
            }
        }
        else {
            die $@;
        } 
    }

    if (!$scanner_actions_fields->{get_host_names_only} && !F5::BigipUtils::is_docker()) {

        $self->{dbh}->do(<<"EndQuery", undef, $policy->{policy_id}, $policy->{policy_id});
UPDATE PLC.PL_SCANNER_VULNS
  JOIN (SELECT scanner_vuln_id, MIN(scanner_vuln_status+0) AS status_min
          FROM PLC.PL_SCANNER_VULNS 
         WHERE policy_id = ?
         GROUP BY scanner_vuln_id) AS temp_table
    ON PLC.PL_SCANNER_VULNS.scanner_vuln_id = temp_table.scanner_vuln_id
   AND PLC.PL_SCANNER_VULNS.policy_id = ?
   SET PLC.PL_SCANNER_VULNS.scanner_vuln_status = temp_table.status_min
EndQuery

        if (defined $retval->{error}) {
            my $error_msg = $retval->{error};
            $error_msg =~ s|\A(.*?) at\s/.*\z|$1|xms;
#//
            return $self->handle_error(action => 'add',
                                       error_code   => F5::ASMConfig::ErrorCode::VULN_FILE_BAD_FORMAT(),
                                       error_thrown => $retval->{error},
                                       display_message => "Vulnerability XML file is invalid ( $error_msg )",
                                       error_level  => 'info',
                                      );
        }
    
        push(@events,
               $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                   session    => $session,
                                   element => F5::ASMConfig::Element->new({
                                                                           element_type => F5::ASMConfig::ElementType::SCANNER_ACTIONS,
                                                                         }),
                                   policy => $policy,
                                   call_flags => $call_flags,
                                   user_name  => $session->{user_name},
                                   suppress_save_to_log => 1,
                                  ));
    }
    

    my @vuln_hosts;
    my $total_vuln_count = 0;

    for my $key(keys %{$retval->{vuln_hosts}}) {
        my $vuln_host = F5::ASMConfig::VulnHost->new(host_name => $key, vuln_count => $retval->{vuln_hosts}->{$key});
        
        if ( $key ) {
            my $norm_host_retval = F5::ASMConfig::Entity::Base->normalize_host(field_value => $vuln_host->{host_name});
            if ($norm_host_retval->{completed_successfully}) {
                push (@vuln_hosts, $vuln_host);
            }
        }
        else {
            push (@vuln_hosts, $vuln_host);
        }
    }

    for my $key(keys %{$retval->{updated_vuln_hosts}}) {
        $total_vuln_count += $retval->{updated_vuln_hosts}->{$key};
    }

    return {return_value => F5::ASMConfig::UpdateScannerVulnDataRes->new(
                                total_vuln_count => $total_vuln_count,
                                vuln_hosts => \@vuln_hosts,
                                vuln_file_path => $retval->{vuln_file_path},),
            events => \@events};
}
                
#-------------------------------------------------------------------------
                                   
sub get_register_url {         
    my ($self, $session, $call_flags, $policy, $redirect_url) = @_;

    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    return {return_value => $scanner->get_register_url(redirect_url => $redirect_url)};
}

#-------------------------------------------------------------------------

sub get_scanner_more_info_url {
    my ($self, $session, $call_flags, $policy) = @_;
    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    my $retval = $scanner->get_scanner_more_info_url();

    return { return_value => $retval };
}

#-------------------------------------------------------------------------

sub get_global_scanner_more_info_url {
    my ($self, $session, $call_flags, $global_scanner_config) = @_;
    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   scanner_type => $global_scanner_config->{scanner_type},
                                   call_flags => $call_flags,);

    my $retval = $scanner->get_scanner_more_info_url();

    return { return_value => $retval };
}

#-------------------------------------------------------------------------

sub get_scanner_more_info_url_with_subscription {
    my ($self, $session, $call_flags, $policy) = @_;
    my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                                   client => $self,
                                   session => $session,
                                   call_flags => $call_flags,
                                   policy => $policy,);

    my $retval = $scanner->get_scanner_more_info_url_with_subscription();

    return { return_value => $retval };
}

#----------------------------------------------------------------------

1;
