package F5::ASMConfig::Handler::ScannerSiteMapTransmission;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::DbUtils;
use Time::Local;
use POSIX;
use JSON;
use F5::Cluster::Utils;

our @handled_functions = (
                           {
                             function_name => "get_last_scanner_site_map_transmission",
                           },
                           {
                             function_name => "add_scanner_site_map_transmission",
                           },
                           {
                             function_name => "update_scanner_site_map_transmission",
                           },
                           {
                             function_name => "run_scheduled_scanner_site_map_updates",
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::ScannerSiteMapTransmission';

#----------------------------------------------------------------------
sub get_last_scanner_site_map_transmission {
    my($self, $session, $call_flags, $policy) = @_;

    my $last_scanner_site_map_transmission = $entity_class->new(dbh => $self->{dbh},
            policy => $policy,
            get_criteria => {},
            sort_orders => [{field_name => 'transmission_id', # update_time may be not unique
            sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC()}]);

    if (!$last_scanner_site_map_transmission) {
        return $self->handle_error(action => "get",
                call_flags => $call_flags,
                error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                entity_class => $entity_class);
    }

    return { return_value => $last_scanner_site_map_transmission };
}

#----------------------------------------------------------------------
sub add_scanner_site_map_transmission {
    my ($self, $session, $call_flags, $policy, $transmission) = @_;

    # validate no pending transmission in progress
    my $search_call_flags = F5::ASMConfig::CallFlags->new(search_filters => [ F5::ASMConfig::SearchFilter->new(
                field_name => 'status',
                field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN(),
                field_value_list => [F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_FAILED(),
                    F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_COMPLETED()],)],
        search_orders => [{field_name => 'transmission_id', # update_time may be not unique
                sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC()}],);
    my $search_result = $self->search_scanner_site_map_transmission($session, $search_call_flags, $policy,
        F5::ASMConfig::ScannerSiteMapTransmission->new());
    if (defined $search_result->{element_list}->[0]) {
        return $self->handle_error(action => "add",
            call_flags => $call_flags,
            error_code => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
            entity_class => $entity_class);
    }

    $transmission->{update_time} = time;
    return $self->base_add(entity_class => $entity_class,
        rpc_params => [ $session, $call_flags, $policy, $transmission ]);
}

#----------------------------------------------------------------------
sub update_scanner_site_map_transmission {
    my ($self, $session, $call_flags, $policy, $transmission_criteria, $store_transmission) = @_;

    my $logger = $self->{logger};

    if ($store_transmission->{action} == F5::ASMConfig::ScannerSiteMapAction::SCANNER_SITE_MAP_ACTION_CHECK_READINESS()
        && $store_transmission->{status} == F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_IN_PROGRESS()) {
        $logger->debug("Connect to the scanner sitemap and check readiness");

        # case need to check scanner site readiness
        eval {
            my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                client => $self,
                session => $session,
                call_flags => $call_flags,
                policy => $policy);
            my $response = $scanner->get_scanner_upload_status();
            if ($response->is_success) {
                my $content = $response->content ? from_json($response->content) : {};
                if ($content->{ready}) {
                    # scanner is ready - switch to prepare sitemap
                    $store_transmission->{action} = F5::ASMConfig::ScannerSiteMapAction::SCANNER_SITE_MAP_ACTION_PREPARING_SITE_MAP();
                    $store_transmission->{update_incremental} = $content->{incremental} ? 1 : 0;
                    $store_transmission->{msg} = '';
                } else {
                    # scanner not ready - not an error, report the reason
                    $store_transmission->{status} = F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_COMPLETED();
                    $store_transmission->{msg} = "Not ready. " . $content->{reason} || '';
                }
            } else {
                # scanner respond with an error - log errors if found
                my $rcode = $response->code;
                $store_transmission->{status} = F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_FAILED();
                $store_transmission->{msg} = "Failed to get scanner status (code $rcode)";
                $logger->error("Failed to get scanner status (code $rcode)");
                $logger->error($response->content);
            }
        };
        if ($@) {
            $logger->error($@);
            $store_transmission->{status} = F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_FAILED();
            $store_transmission->{msg} = "Failed to get scanner status";
        }
        $store_transmission->{update_time} = time;
    } elsif ($store_transmission->{action} == F5::ASMConfig::ScannerSiteMapAction::SCANNER_SITE_MAP_ACTION_POSTING_SITE_MAP()
        && $store_transmission->{status} == F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_IN_PROGRESS()) {

        # case need to post site map information to scanner site

        if ($store_transmission->{urls_reported} == 0) {
            $logger->debug("No new site map information to post to the scanner");
            $store_transmission->{status} = F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_COMPLETED();
            $store_transmission->{msg} = "No new site map information to post to the scanner";
        } else {
            $logger->debug("Posting site map information to the scanner required");
            eval {
                my $scanner = F5::Scanner->new(dbh => $self->{dbh},
                    client => $self,
                    session => $session,
                    call_flags => $call_flags,
                    policy => $policy);

                my $response = $scanner->post_scanner_site_map_data();
                if ($response->is_success) {
                    # site map information posted
                    $store_transmission->{status} = F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_COMPLETED();
                } else {
                    # scanner post request failed with error, report errors if found
                    my $rcode = $response->code;
                    $store_transmission->{status} = F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_FAILED();
                    $store_transmission->{msg} = "Failed to send site map information to scanner (code $rcode)";
                    $logger->error("Failed to send site map information to scanner (code $rcode)");
                    $logger->error($response->content);
                }
            };
            if ($@) {
                $logger->error($@);
                $store_transmission->{status} = F5::ASMConfig::ScannerSiteMapStatus::SCANNER_SITE_MAP_STATUS_FAILED();
                $store_transmission->{msg} = "Failed to send site map information to scanner";
            }
        }
        $store_transmission->{update_time} = time;
    }

    return $self->base_update(entity_class => $entity_class,
        rpc_params => [ $session, $call_flags, $policy, $transmission_criteria, $store_transmission ]);
}

#----------------------------------------------------------------------

sub is_scanner_connected {
    my ($self, %params) = @_;
}
#----------------------------------------------------------------------

sub run_scheduled_scanner_site_map_updates {
    my ($self, $session, $call_flags, $place_holder) = @_;

    if (( F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade()) ||
        F5::Cluster::Utils::is_failover_standby()) {
        return { return_value => 1 };
    }

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my ($sched_sec,$sched_min,$sched_hour,$sched_mday,$sched_mon,$sched_year,$sched_wday,$sched_yday,$sched_isdst) = localtime(time);
    my $sync_time = timelocal(0,0,$sched_hour,$sched_mday,$sched_mon,$sched_year);

    # get all policy IDs of active policies
    # that have the bidirectional feature enabled

    my $sql = <<EndQuery;
SELECT p.id
FROM   PLC.PL_POLICIES as p
JOIN   PLC.PL_SCANNER_CONFIG as s
ON     p.id = s.policy_id
WHERE  p.account_id > 0
AND    s.flg_bidir_enabled = 1
EndQuery

    my $policy_recs = $dbh->selectall_arrayref($sql ,{ Slice => {} });

    # for each active policy with the bidirectional feature enabled,
    # get its scanner config, and if necessary, initiate the procedure
    # of sending the site map info to the scanner.

    for my $policy_rec (@{$policy_recs}) {

        my $policy_id = $policy_rec->{id};
        my $policy = F5::ASMConfig::Policy->new({policy_id => $policy_id});
        my $get_call_flags = F5::ASMConfig::CallFlags->new();
        my $scanner_config = $self->get_scanner_config($session, $get_call_flags, $policy,);

        # skip if scanner is not supported or not connected
        if ($scanner_config->{scanner_type} =~ /^sentinel/) {
            if ($scanner_config->{scan_status} == F5::ASMConfig::ScanStatus::SCAN_STATUS_REGISTRATION_REQUESTED() ||
                ($scanner_config->{flg_is_api_key_final} == 0) ||
                ($scanner_config->{api_key} eq '')) {
                next;
            }
        } elsif ($scanner_config->{scanner_type} eq 'hailstorm') {
            if (($scanner_config->{scanner_site_name} eq '') ||
                ($scanner_config->{api_key} eq '')) {
                next;
            }
        } else {
            # unsupported scanner type
            next;
        }

        if ($scanner_config->{last_scheduled_sync_time} == $sync_time) {
            $logger->debug("Skipping scanner config sync for policy ($policy_id): already synced");
            next;
        }

        # check scanner schedule update frequency
        if ($scanner_config->{sync_frequency} == F5::ASMConfig::ScannerScheduleFrequency::SCANNER_SCHEDULE_FREQUENCY_DAILY()
            && $scanner_config->{sync_hour} == $sched_hour) {
            $logger->debug("Scanner config for policy ($policy_id): required sync - daily");
        } elsif ($scanner_config->{sync_frequency} == F5::ASMConfig::ScannerScheduleFrequency::SCANNER_SCHEDULE_FREQUENCY_WEEKLY()
            && $scanner_config->{sync_weekday} == $sched_wday
            && $scanner_config->{sync_hour} == $sched_hour) {
            $logger->debug("Scanner config for policy ($policy_id): required sync - weekly");
        } elsif ($scanner_config->{sync_frequency} == F5::ASMConfig::ScannerScheduleFrequency::SCANNER_SCHEDULE_FREQUENCY_MONTHLY()
            && $scanner_config->{sync_day_of_month} == $sched_mday
            && $scanner_config->{sync_hour} == $sched_hour) {
            $logger->debug("Scanner config for policy ($policy_id): required sync - monthly");
        } else {
            # nothing to do - skip it
            next;
        }

        # invoke sync on policy and update the scanner configuration
        my $add_call_flags = F5::ASMConfig::CallFlags->new();
        my $transmission = F5::ASMConfig::ScannerSiteMapTransmission->new();
        eval {
            $self->add_scanner_site_map_transmission($session, $add_call_flags, $policy, $transmission);
        };
        if ($@) {
            $logger->error("Failed to trigger scanner site map update for policy ($policy_id)");
            $logger->error($@);
        }

        # update scanner configuration with the last time we scheduled a sync request
        eval {
           $logger->debug("updating scanner configuration for policy ($policy_id) with scheduled sync request time ($sync_time)");
            my $updated_scanner_config = F5::ASMConfig::ScannerConfig->new({
                    scanner_config_id => $scanner_config->{scanner_config_id},
                    last_scheduled_sync_time => $sync_time,
                });
            $self->update_scanner_config($session, $call_flags, $policy, $updated_scanner_config, $updated_scanner_config);
        };
        if ($@) {
            $logger->error("Failed updating scanner configuration for policy ($policy_id) with last scheduled sync time ($sync_time)");
            $logger->error($@);
        }
    }
    return { return_value => 1 };
}

#----------------------------------------------------------------------
1;
