package F5::ASMConfig::Handler::Section;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;

our @handled_functions = (
                           {
                             function_name => 'update_section',
                             criteria_index => 3,
                             store_index => 4,
                             relayable => 1,
                           },
                           {
                             function_name => 'update_many_sections',
                             criteria_index => 3,
                             store_index => 4,
                             relayable => 1,
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::Section';

#----------------------------------------------------------------------

sub apply_inheritance_changes {
    my (%params) = @_;
    my $self = $params{self};
    my $existing_section = $params{existing_section};
    my $store_section = $params{store_section};
    my $dbh = $params{dbh};
    my $policy = $params{policy} || F5::ASMConfig::Cache::get_cached_policy(policy => $store_section->{policy});
    my $call_flags = $params{call_flags};

    if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        my $child_policy_ids = $policy->child_policy_ids();
        if ($store_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY()) {
            #On switch to Mandatory push all related entities to all children that are not already inheriting, otherwise allow children to manage their own inheritance.
            my @sections = F5::ASMConfig::Entity::Section->retrieve_all(dbh => $dbh,
                                                                        search_filters => [ 
                       F5::ASMConfig::SearchFilter->new( field_name => 'section_type',
                                                         field_value => $store_section->{section_type},
                                                         field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),),
                       F5::ASMConfig::SearchFilter->new( field_name => 'child_inheritance_status',
                                                         field_value_list => [
                                                             F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_ACCEPTED(),
                                                             F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_MANDATORY(),
                                                                             ],
                                                         field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN(),),
                       F5::ASMConfig::SearchFilter->new( field_name => 'policy_id',
                                                         field_value_list => $child_policy_ids,
                                                         field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),)
                                                                         ],
                                                                        selected_fields => [],
                                                                        minimal_load => 1,
                                                                       );
            my @child_policies = map { F5::ASMConfig::Policy->new(policy_id => $_->{policy_id}) } @sections;
            #Update all children who are newly inheriting this section
            copy_section_entities(parent_policy => $policy,
                                  child_policies => \@child_policies,
                                  sections => [ $store_section ],
                                  call_flags => $call_flags,
                                  self => $self,
                                  dbh => $dbh,);

        }

        my $child_call_flags = F5::ASMConfig::CallFlags->new(
                                   %$call_flags,
                                   inherited_call => 1, 
                                   skip_inheritance => ($store_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_MANDATORY())
                                  );
        my $child_section_criteria = F5::ASMConfig::Section->new(section_type => $store_section->{section_type});
        my $store_child_section = F5::ASMConfig::Section->new(parent_inheritance_status => $store_section->{parent_inheritance_status});
        for my $child_policy_id (@$child_policy_ids) {
            $self->update_section($self->{session},
                                  $child_call_flags, 
                                  F5::ASMConfig::Policy->new(policy_id => $child_policy_id),
                                  $child_section_criteria,
                                  $store_child_section);
        }
    }
    else {
        if ( !$call_flags->{skip_inheritance}
            && ($call_flags->{reinherit} 
                || ($existing_section->is_inherited() != $store_section->is_inherited()))
            ) {

            my $parent_policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh,
                                                                        policy => {policy_id => $policy->{parent_policy_id}});
            if (!$store_section->is_inherited() || $call_flags->{reinherit}) {
                #Uninherit Section from Policy
                my %entity_call_params = (dbh => $dbh,
                                          policy => $policy,
                                         );

                my $delete_disinherited_entities = $call_flags->{delete_disinherited_entities};

                #Update policy entities 
                for my $entity_filter (@{$store_section->section_entity_filters()}) {
                    my $call_flags = F5::ASMConfig::CallFlags->new(search_filters => $entity_filter->{search_filters},
                                                                   disinherit => 1,
                                                                   inherited_call => 1);
                    my $deassociate_update_field = $entity_filter->{deassociate_update_field} || 'is_inherited';
                    my $deassociate_update_value = $entity_filter->{deassociate_update_value} || 0;

                    for my $entity_type (@{$entity_filter->{entity_types}}) {
                        my $criteria_obj = $entity_type->new()->set_inherited();
                        my $update_function_name = $self->get_function_name(element_type => $entity_type->element_type,
                                                                            event_type => F5::ASMConfig::EventType::UPDATE(),
                                                                            is_many => 1,);
                        if ( $delete_disinherited_entities && $entity_type->can_delete_on_disinherit ) {
                            my $function_name = $self->get_function_name(element_type => $entity_type->element_type,
                                                                         event_type => F5::ASMConfig::EventType::DELETE(),
                                                                         is_many => 1,);

                            $self->$function_name($self->{session}, $call_flags, $policy, $criteria_obj);
                        }
                        my $store_obj = $entity_type->new(object => {$deassociate_update_field => $deassociate_update_value});
                        $self->$update_function_name($self->{session}, $call_flags, $policy, $criteria_obj, $store_obj);
                    }
                }

                if ($policy->{parent_policy_id} && $parent_policy) {
                    #Update suggestions on Parent Policy for this section to remove this policy from the suggesting children.
                    $self->update_many_suggestions($self->{session}, 
                                                   F5::ASMConfig::CallFlags->new(remove_nested_struct_on_update => 1),
                                                   $parent_policy,
                                                   F5::ASMConfig::Suggestion->new(section_type => $store_section->{section_type}),
                                                   F5::ASMConfig::Suggestion->new(suggesting_child_policy_ids => [$policy->{policy_id}]));
                }
            }
            if ($store_section->is_inherited()) {
                #Inherit Section to Policy
                copy_section_entities(parent_policy => $parent_policy,
                                      child_policies => [ $policy ],
                                      sections => [ $store_section ],
                                      call_flags => $call_flags,
                                      self => $self,
                                      dbh => $dbh,);
            }
        }
    }
}

sub copy_section_entities {
    my (%params) = @_;
    my $self = $params{self};
    my $parent_policy = $params{parent_policy};
    my $child_policies = $params{child_policies};
    my $sections = $params{sections};
    my $dbh = $params{dbh};
    my $call_flags = $params{call_flags};
 
    my $add_call_flags = F5::ASMConfig::CallFlags->new( add_on_updating_nonexistent_record => 1,
                                                        ignore_duplicate_record_on_add => 1,
                                                        log_note => "Inherited from $parent_policy->{policy_name}",
                                                        suppress_events => $call_flags->{suppress_events},
                                                        inherited_call => 1);

    for my $section (@$sections) {
        #Update all children who are newly inheriting this section
        for my $entity_filter (@{$section->section_entity_filters()}) {
            next if $entity_filter->{deassociate_update_field};
            for my $entity_type (@{$entity_filter->{entity_types}}) {
                my @parent_strict_identifiers;
                my $strict_inheritance = $entity_type->strict_inheritance();
                my $parent_entity_iter = $entity_type->retrieve_all(dbh => $dbh,
                                                                    policy => $parent_policy,
                                                                    search_filters => $entity_filter->{search_filters},);

                while (my $parent_entity = $parent_entity_iter->get_next()) {
                    if ($strict_inheritance) {
                        push (@parent_strict_identifiers, $parent_entity->{$entity_type->name_field()});
                    }
                    my @inherit_entities = (@{$parent_entity->get_external_entities(within_policy => 1)});
                    for my $inherit_entity (@inherit_entities) {
                        my $add_function_name =  $self->get_function_name(element_type => $inherit_entity->element_type,
                                                                          event_type => F5::ASMConfig::EventType::ADD(),);
                        $inherit_entity->purify_store_object();
                        for my $child_policy (@$child_policies) {
                            $inherit_entity->{policy} = $child_policy;
                            $inherit_entity->{policy_id} = $child_policy->{policy_id};
                            $self->$add_function_name($self->{session}, $add_call_flags, $child_policy, $inherit_entity);
                        }
                    }
                    my $add_function_name =  $self->get_function_name(element_type => $parent_entity->element_type,
                                                                      event_type => F5::ASMConfig::EventType::UPDATE(),);
                    $parent_entity->remove_non_inheritance_fields();
                    $parent_entity->set_inherited();
                    for my $child_policy (@$child_policies) {
                            $parent_entity->{policy} = $child_policy;
                            $parent_entity->{policy_id} = $child_policy->{policy_id};
                            $self->$add_function_name($self->{session}, $add_call_flags, $child_policy, $parent_entity);
                    }
                }
                if ($strict_inheritance) {
                    my $delete_function_name =  $self->get_function_name(element_type => $entity_type->element_type,
                                                                         event_type => F5::ASMConfig::EventType::DELETE(),
                                                                         is_many => 1,);
                    my $delete_call_flags = F5::ASMConfig::CallFlags->new(
                                                                           log_note => "Inherited from $parent_policy->{policy_name}",
                                                                           search_filters => [ F5::ASMConfig::SearchFilter->new(
                                                                               field_name => $entity_type->name_field(),
                                                                               field_value_list => \@parent_strict_identifiers,
                                                                               field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN())],
                                                                           suppress_events => $call_flags->{suppress_events},
                                                                           inherited_call => 1);
                    for my $child_policy (@$child_policies) {
                        $self->$delete_function_name($self->{session}, $delete_call_flags, $child_policy, $entity_type->new());
                    }
                }
            }
        }
        #Special case to make Signatures more efficient
        #This is executed after the updates to signature_sets
        if ($section->{section_type} == F5::ASMConfig::SectionType::SECTION_TYPE_ATTACK_SIGNATURES()) {            
            my $signature_call_flags = F5::ASMConfig::CallFlags->new(%$add_call_flags);
            my @child_policy_ids = map { $_->{policy_id} } @$child_policies;
            my $enabled_search_filter = F5::ASMConfig::SearchFilter->new({ field_name => 'flg_disabled', 
                                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                                                                         });
            $signature_call_flags->{search_filters} = [ 
                                               F5::ASMConfig::SearchFilter->new(
                                                          field_name => 'is_inherited',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                          field_value => 1,),
                                               F5::ASMConfig::SearchFilter->new(
                                                          field_name => 'flg_disabled_differs_from_parent_policy_id',
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                          field_value => $parent_policy->{policy_id}),
                                               F5::ASMConfig::SearchFilter->new(
                                                          field_name => 'policy_id',
                                                          field_value_list => \@child_policy_ids,
                                                          field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()),
                                               $enabled_search_filter
                                               ];
            for my $flg_disabled (0,1) {
                $enabled_search_filter->{field_value} = $flg_disabled;
                $self->update_many_policy_signatures($self->{session}, $signature_call_flags, undef, F5::ASMConfig::PolicySignature->new(), 
                                                                                                     F5::ASMConfig::PolicySignature->new(flg_disabled => $flg_disabled));
            }
        }

    }
}

#----------------------------------------------------------------------

sub update_section {
    my($self, $session, $call_flags, $policy, $section_criteria, $store_section) = @_;

    my $section = $entity_class->new(dbh => $self->{dbh}, policy => $policy, get_criteria => $section_criteria);

    if ( $section && $section->{child_inheritance_status} == F5::ASMConfig::SectionChildInheritanceStatus::SECTION_CHILD_INHERITANCE_STATUS_NA ) {
        undef $store_section->{comment};
    }    

    my $apply_inheritance_changes = !$section || ($section->has_value_changed(existing_obj => $section, store_obj => $store_section, 
                                                                                                          check_fields => [qw(child_inheritance_status parent_inheritance_status)]));

    if ( $apply_inheritance_changes &&
         ( $store_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NONE ||
           $store_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NA )) {
        $store_section->{comment} = '';
    }

    my $retval = $self->base_update(rpc_params => [$session, $call_flags, $policy, $section_criteria, $store_section],
                                    entity_class => $entity_class,
                                    );

    if (($retval->{events} && $apply_inheritance_changes) || $call_flags->{reinherit})  {
        $self->{section_policy_cache} = {};
        my $saved_events = delete $self->{events};
        my $local_events = delete $retval->{events};

        apply_inheritance_changes(dbh => $self->{dbh},
                       self => $self,
                       policy => $policy,
                       existing_section => $section,
                       store_section => $retval->{return_value},
                       call_flags => $call_flags,
                      );
        if ($self->{events}) {
            push (@$local_events, @{delete($self->{events})});
        }
        $retval->{events} = $local_events;
        $self->{events} = $saved_events;
    }
    return $retval;
}

#----------------------------------------------------------------------

sub update_many_sections {
    my($self, $session, $call_flags, $policy, $section_criteria, $store_section) = @_;

    if (!defined $call_flags){
        $call_flags = F5::ASMConfig::CallFlags->new();
    }

    if (!defined $call_flags->{search_filters}){
       $call_flags->{search_filters} = [];
    }

    if ( $store_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NONE() ||
         $store_section->{parent_inheritance_status} == F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NA() ) {
        $store_section->{comment} = '';
    }

    push(@{$call_flags->{search_filters}},
                 { field_name => 'parent_inheritance_status',
                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                   field_value => F5::ASMConfig::SectionParentInheritanceStatus::SECTION_PARENT_INHERITANCE_STATUS_NA() });

    my @sections = $entity_class->retrieve_all(dbh => $self->{dbh},
                                               policy => $policy,
                                               get_criteria => $section_criteria,
                                               search_filters => $call_flags->{search_filters});

    my $retval = $self->base_update_many(rpc_params => [$session, $call_flags, $policy, $section_criteria, $store_section],
                                         entity_class => $entity_class,);

    if ($store_section->{child_inheritance_status} || $store_section->{parent_inheritance_status}) {
        $self->{section_policy_cache} = {};
        my $saved_events = delete $self->{events};
        my $local_events = delete $retval->{events};
        my %existing_sections = map { ($_->{section_type} => $_) } @sections;
        for my $section (@{$retval->{entities_changed}}) {
            my $existing_section = $existing_sections{$section->{section_type}};
            if ($existing_section &&
                $section->has_value_changed(existing_obj => $existing_section, store_obj => $section, 
                                            check_fields => [qw(child_inheritance_status parent_inheritance_status)])) {

                apply_inheritance_changes(dbh => $self->{dbh},
                                          self => $self,
                                          policy => $policy,
                                          store_section => $section,
                                          existing_section => $existing_section,
                                          call_flags => $call_flags,
                                         );
            }
        }
        if ($self->{events}) {
            push (@$local_events, @{delete($self->{events})});
        }
        $retval->{events} = $local_events;
        $self->{events} = $saved_events;
    }

    return $retval;
}

#----------------------------------------------------------------------

1;
