package F5::ASMConfig::Handler::SetActive;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::PBConfig;
use F5::ASMConfig::Entity::SetActive;
use F5::ASMConfig::Entity::Base;
use F5::ASMConfig::Cache;
use F5::SetActive::Impl;
use F5::Utils::Account;

our @handled_functions = (
                           {
                             function_name => 'set_active',
                             relayable => 1,
                             update_mcp_timestamp => 0,
                             recommend_sync => 1,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                         );

#----------------------------------------------------------------------

sub set_active {
    my($self, $session, $call_flags, $policy, $set_active) = @_;

    my $policy_id;
    if ($policy->{policy_id}) {
        $policy_id = $policy->{policy_id};
        if ($policy->{_failed_to_load}) {
            my $error_msg = "Missing policy";
            $self->handle_error(action => 'set_active',
                                       error_thrown   => $error_msg,
                                       display_message => $error_msg,
                                       error_level     => 'info',
                                       error_code      => F5::ASMConfig::ErrorCode::MISSING_POLICY(),
            );
            return { return_value => 1};
        }
    }

    if ($policy->{policy_type} && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        return apply_parent_policy(self => $self,
                                   call_flags => $call_flags,
                                   policy => $policy,
                                   set_active => $set_active);
    }

    my %known_times_copy = %{$self->{known_times}};
    if (exists($call_flags->{known_times}) &&
        $policy->{account_id} &&
        ( !$session->{user_name} || ( $session->{user_name} ne "Policy Builder") )
       ) {
        #This is a CGC forwarded call.  Check the known times for sanity.
        ## if the times match, then awesomesauce.
        if ( (keys %{$call_flags->{known_times}} ne keys %known_times_copy) ||
               grep {$known_times_copy{$_} ne $call_flags->{known_times}{$_}} keys %known_times_copy) {
            if ($call_flags->{out_of_sync}) {
                $self->{logger}->debug("Received set_active. known_times hash does not match, but Sender was out_of_sync. Ignoring.");
                return { return_value => 1};
            }
            ## otherwise... squawk, go into emergency state and return... 
            if ($policy && $policy->{policy_id}) {
                my $request_full_sync = 0;
                if ($call_flags->{known_times}{$call_flags->{originator}} gt $known_times_copy{$call_flags->{originator}}) {
                    $request_full_sync = 1;
                    $self->{logger}->info("Received out of order call from $call_flags->{originator}.  Requesting full sync.");
                }
                elsif (!$call_flags->{originator_zero_padded_major_minor_version} ||
                     $call_flags->{originator_zero_padded_major_minor_version} ne $self->{zero_padded_major_minor_version}) {

                    $request_full_sync = 1;
                    $self->{logger}->info("Received a call where known_times hash does not match for devices running different product versions. Requesting full sync.");
                }
                if ($request_full_sync) {
                    $self->sync_request_ucs($session, $call_flags, $policy, $set_active);
                }
                else {
                    $self->{logger}->info("Received set_active request, but known_times hash does not match.  Requesting active policy sync.");
                    $self->sync_request_active_policy_ucs($session, $call_flags, $policy, $set_active);
               }
            }
            else {
                $self->{logger}->debug("Received non-policy set_active request, but known_times hash does not match. Unsure how to proceed.");
            }
            return { return_value => 1};
        }

    }

    $set_active->{username} ||= $session->{gui_user_name} || $session->{user_name};
    $set_active->{job} ||= $call_flags->{job};
    if (! defined $set_active->{ignore_if_policy_unmodified}) {
        $set_active->{ignore_if_policy_unmodified} = 1;
    }

    if (!$policy->{account_id}) {
        $set_active->{activate_policy} = 1;
    }

    if ($set_active->{activate_policy}) {
        F5::Utils::Account::validate_create_account(policy => $policy,
                                                    dbh => $self->{dbh},);
    }

    # Don't update MCP in intermediate commit
    my $saved_callbacks;
    if (! $set_active->{deactivate_policy}) {
        $saved_callbacks = $self->{dbh}->{private_f5_callbacks};
        $self->{dbh}->{private_f5_callbacks} = [];
    }

    if (!exists($set_active->{do_transaction})) {
        $set_active->{do_transaction} = 1;
    }

    if ($set_active->{do_transaction}) {
        # Use set active's built-in transaction
        $self->{dbh}->commit;
    }

    # always skip_mcp here, 
    #    do an mcp_store on the policy obj after restoring the saved callbacks
    my $rv;
    eval {
        $rv = F5::SetActive::Impl::set_active(dbh => $self->{dbh},
                                              logger => $self->{logger},
                                              policy_id => $policy_id,
                                              asm_config_handler => $self,
                                              skip_mcp => 1,
                                              %$set_active,
                                             );
    };
    my $set_active_error = $@;

    # 'begin_work' will generate a - 
    #     "DBD::mysql::db begin_work failed: Already in a transaction"
    # error, in case 'set_active' blew up and left an opened transaction (AutoCommit=0)
    # should 'begin_work' only if a transaction was closed by 'set_active' (AutoCommit=1)
    if ($set_active->{do_transaction} && $self->{dbh}->{AutoCommit}) {
        # Restart transaction for events
        $self->{dbh}->begin_work;
    }

    my $set_active_event;
    if ($set_active_error) {
        my $set_active_entity = F5::ASMConfig::Entity::SetActive->new(dbh => $self->{dbh},
                                                                      policy => $policy,);
        $call_flags->{do_not_relay} = 1;

        ($set_active_event) = $self->create_event(event_type => F5::ASMConfig::EventType::ERROR,
                                                  session => $session,
                                                  policy => $policy,
                                                  element => $set_active_entity,
                                                  call_flags => $call_flags,
                                                  user_name => $session->{user_name},
                                                  do_policy_version_update => 0,
                                                 );

        $self->handle_error(action => 'set_active',
                            error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                            error_thrown => $set_active_error,
                            events => [ $set_active_event ],
                            error_level => 'error',
                           );

        return { return_value => 1,
                 events => [ $set_active_event ]};
    }

    if ($policy && $policy->{policy_id}) {
        $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $policy, force_reload => 1);
    }

    # Restore callbacks after reset in begin_work
    if ($saved_callbacks) {
        $self->{dbh}->{private_f5_callbacks} = $saved_callbacks;
    }

    # after restoring the saved callbacks, persist the resulting policy obj into MCP
    if ($rv->{do_mcp_store}) {
        $policy->mcp_store(skip_mcp => $call_flags->{skip_mcp});
    }

    my @events;
    if ($policy && $policy->{policy_id}) {
        if ($self->{rpc_entry_depth} == 1) {
            #CMI: only add known_times to the call_flags on an actual set_active call, not on a nested call
            $call_flags->{known_times} = \%known_times_copy;
        }
        if ($rv->{completed_successfully}) {
            my $entity = F5::ASMConfig::Entity::PBConfig->new(dbh => $self->{dbh},
                                                              policy => $policy,
                                                              get_criteria => { } );
            if ($entity) {
                push(@events,
                     $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                         session => $session,
                                         policy => $policy,
                                         element => $entity,
                                         call_flags => $call_flags,
                                         user_name => $session->{user_name},
                                         suppress_save_to_log => 1,
                                        )
                    );
            }
        }

        my $set_active_entity = F5::ASMConfig::Entity::SetActive->new(dbh => $self->{dbh},
                                                                      policy => $policy,
                                                                      object => $rv,);

        ($set_active_event) = $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                                  session => $session,
                                                  policy => $policy,
                                                  element => $set_active_entity,
                                                  call_flags => $call_flags,
                                                  user_name => $session->{user_name},
                                                  do_policy_version_update => 0,
                                                 );
        push(@events, $set_active_event);
    }

    if ($rv->{completed_successfully}) {
        if (defined $session->{user_name} && $session->{user_name} eq 'Policy Builder') {
            if ($policy && $policy->{policy_name}) {
                $self->write_to_syslog(msg => "Policy Builder applied security policy \"$policy->{policy_name}\".");
            }
        }
        if ($set_active->{activate_policy} || $set_active->{deactivate_policy}) {
            $call_flags->{writes_to_mcp} = 1;
        }
        $self->{logger}->debug("Send event for set active") if $self->{logger};
        return { return_value => 1,
                 events => \@events };
    }
    else {
        $call_flags->{do_not_relay} = 1;
        $set_active_event->{event_type} = F5::ASMConfig::EventType::ERROR;

        if ($rv->{error_message} && $rv->{error_message} =~ /aborting/) {
            return { return_value => 1};
        }
        my $error_thrown = $rv->{error_message} || "Unknown error";
        my $error_level = $rv->{error_level} || "error";
        $self->handle_error(action => 'set_active',
                            error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                            error_thrown => $error_thrown,
                            events => [ $set_active_event ],
                            error_level => $error_level,
                           );
    }
    return { return_value => 1};
}

#----------------------------------------------------------------------
sub apply_parent_policy {
    my (%params) = @_;
    my $self       = $params{self};
    my $call_flags = $params{call_flags};
    my $policy     = $params{policy};

    my $lock_file =  $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                             logger => $self->{logger});

    $locker->lock();

    #We must not retrieve the is_modified field for every child policy. If they have no changes it will lock up MYSQL
    my $child_policy_ids = $policy->child_policy_ids();
    my $modified_child_policies = F5::DbUtils::get_distinct_values(dbh => $self->{dbh},
                                                                   table_name => 'PLC.PL_POLICY_CHANGES_LOG',
                                                                   fields => ['policy_id'],
                                                                   where_fields => { policy_id => $child_policy_ids});

    my @modified_child_policy_ids = keys %$modified_child_policies;

    my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => { policy_id => \@modified_child_policy_ids});
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    my ($max_change_log) = $self->{dbh}->selectrow_array(<<EndQuery, undef, @where_values);
SELECT MAX(policy_changes_log_id)
  FROM PLC.PL_POLICY_CHANGES_LOG
  WHERE $where_clause
EndQuery

    my $child_policies_iter = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh},
                                                                          selected_fields => [qw(is_active_policy)],
                                                                          search_filters => [
                                                                              F5::ASMConfig::SearchFilter->new(field_name => 'policy_id',
                                                                                                               field_value_list => \@modified_child_policy_ids,
                                                                                                               field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()),
                                                                              ],
                                                                          minimal_load => 1,
                                                                          get_criteria => { },
                                                                         );

    my $set_active_flags = F5::ASMConfig::SetActiveFlags->new(send_sync => 0,
                                                              do_transaction => 0,
                                                              do_update_core => 0,
                                                              do_save_policy_history => 0,
                                                              delayed_update_core => 1,); # Only perform a single update core at the very end

    # delayed_update_core should not do cluster sync cause we have suitable post_commit_hook(s) right below
    $call_flags->{skip_cluster_sync} = 1;

    my @events;
    while (my $child_policy = $child_policies_iter->get_next()) {
       if ($child_policy->{is_active_policy}) {
           $self->set_active($self->{session}, $call_flags, $child_policy, $set_active_flags);
           $set_active_flags->{checksums_policy_id} = $child_policy->{policy_id};
       }
    }

    push (@events, $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                       session => $self->{session},
                                       policy => $policy,
                                       element => F5::ASMConfig::Entity::SetActive->new(object => {completed_successfully => 1}, policy => $policy, dbh => $self->{dbh}),
                                       call_flags => $call_flags,
                                       user_name => $self->{session}->{user_name},
                                       do_policy_version_update => 0,
                                      ));

    if ($policy->is_modified) {
        my $user_name = $self->{session}->{gui_user_name} || $self->{session}->{user_name} || 'SYSTEM';
        $self->save_policy_history($self->{session},
                                   F5::ASMConfig::CallFlags->new({do_not_relay => 1}),
                                   $policy,
                                   F5::ASMConfig::PolicyHistory->new({user_name => $user_name}));
        $policy->clear_modification_log();
    }
    $locker->unlock();

    $self->add_post_commit_hook(name => 'mark_policies_unmodified',
                                func => sub { 
                                    my $rv = $self->{dbh}->do(<<EndQuery, undef, @where_values, $max_change_log);
DELETE FROM PLC.PL_POLICY_CHANGES_LOG
  WHERE $where_clause AND policy_changes_log_id <= ?
EndQuery
                                    return $rv;
               });

    my $cluster_sync = F5::Cluster::Sync->new(dbh => $self->{dbh}, logger => $self->{logger});
    if ($policy->{child_policy_count}) {
        $self->add_post_commit_hook(name => 'perform_full_sync',
                                    func => sub {
                                        $cluster_sync->send_sync(force_full_sync => 1);
                                    }
                                   );
    }
    else {
        $self->add_post_commit_hook(name => 'perform_parent_policy_sync',
                                    func => sub {
                                        $cluster_sync->send_sync(policy_id => $policy->{policy_id}, account_id => 0, policy_name_crc => $policy->{policy_name_crc});
                                    }
                                   );
    }

    return {return_value => 1,
            events => \@events};
}

#----------------------------------------------------------------------

1;
