package F5::ASMConfig::Handler::Sigfile;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::NegativeSignatures;
use F5::DbUtils;
use F5::NegativeSignatures::SigFile::Update;
use F5::NegativeSignatures::SigFile::AutoDownload;
use F5::NegativeSignatures::Signature qw(:constants);

our @handled_functions = (
                           {
                             function_name => 'update_sigfile',
                             filename_arguments => ['3.filename'],
                             write_access => 1,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'check_for_sigfile_update',
                           },
                           {
                             function_name => 'generate_sigfile',
                           },
                           {
                             function_name => 'get_sigfile_version',
                           },
                         );

#--------------------------------------------------------------------
sub update_sigfile{
  my($self, $session, $call_flags, $place_holder,$sig_file, $sigfile_update_settings) = @_;

  my %sigfile_update_settings_defaults = (
                                          check_interval => 0,
                                          allow_older_timestamp => 0,
                                         );

  if (! $sigfile_update_settings) {
      $sigfile_update_settings = {};
  }

  $call_flags->{skip_memory_check} = 1;

  for my $default_key (keys %sigfile_update_settings_defaults) {
      if (! defined $sigfile_update_settings->{$default_key}) {
          $sigfile_update_settings->{$default_key} = $sigfile_update_settings_defaults{$default_key};
      }
  }

  my %params = (
                dbh => $self->{dbh},
                automatic_mode => defined $sig_file->{filename} ? 0 : 1,
                filename => $sig_file->{filename} || "",
                file_contents_struct => $sig_file->{file_contents_struct},
                flg_is_user_defined => $sig_file->{flg_is_user_defined} || 0,
                username => $call_flags->{user_name} || "",
                in_wrapping_transaction => $self->{in_transaction},
                logger => $self->{logger},
                outbound_call_timeout => $call_flags->{outbound_call_timeout},
                originator => $call_flags->{originator},
                asm_config_handler => $self,
                flg_is_genesis_file => $sig_file->{flg_is_genesis_file},
                sigfile_format => uc($sig_file->{file_format} || ''),
                skip_set_active => $call_flags->{skip_set_active},
                skip_set_sigs_active => $call_flags->{skip_set_sigs_active},
                %$sigfile_update_settings,
               );

  my $rv;
  eval{
      my $sig = F5::NegativeSignatures::SigFile::Update->new(%params);
      if($params{update_type} && ! $params{im_file}) {
          $rv = $sig->apply_changes(%params);
          if(!$params{dry_run}) {
              $sig->post_process(%params);
          }
      }
      else {
          $rv = $sig->update(%params);
      }
  };

  if ($@) {
      my $orig_error = $@;
      my $error_thrown;
      if (ref($orig_error) && $orig_error->{error_code}) {
          $error_thrown = $orig_error;
      }
      elsif (ref($orig_error) && $orig_error->{error_msg}) {
          $error_thrown = $orig_error->{error_msg};
      }
      else {
          $error_thrown = "Signature file update failed. Error: $@";
      }

      #Error suppression added for EM+iControl's benefit
      if (ref($error_thrown) && (exists $error_thrown->{error_code}) &&
         ($error_thrown->{error_code} == F5::ASMConfig::ErrorCode::INVALID_SIGFILE_VERSION() ||
          $error_thrown->{error_code} == F5::ASMConfig::ErrorCode::OLDER_SIGFILE_TIMESTAMP())) {
          $rv = {completed_successfully => 0,
                 error_msg => $error_thrown->{display_message},
                 error_code => $error_thrown->{error_code}};
      }
      elsif (ref($error_thrown) && (exists $error_thrown->{error_msg})) {
         my $error_msg_secured = $error_thrown->{error_msg};
	 # Error message may look like this: "User signature file does not conform to schema (/ts/var/negsig/admin~OhDF4F2qSU1rjaLPfh-8bA:4: parser error : Prema..."
	 # We don't want the file location to be shown to user.
	 $error_msg_secured =~ s/\([^\s]+\s/\(/;
         $rv = {completed_successfully => 0,
                error_msg => $error_msg_secured};
      }     
      else {
          return $self->handle_error(action => 'update_sigfile',
                                     error_thrown => $error_thrown,
                                     error_level  => 'error',
                                    );
      }
  }

   my $return_value = F5::ASMConfig::Entity::SigfileUpdateRes->new(dbh => $self->{dbh}, object => $rv);
   #When doing a manual update with automatic delivery mode, update_mode = "auto-user"
   $return_value->{update_mode} = ($sigfile_update_settings->{update_mode} && $sigfile_update_settings->{update_mode} eq 'manual') ? F5::ASMConfig::SigfileUpdateMode::SIGFILE_UPDATE_MODE_MANUAL 
                                                                                                                                   : F5::ASMConfig::SigfileUpdateMode::SIGFILE_UPDATE_MODE_SCHEDULED;
  
   my @events;
   my $retval = { return_value => $return_value ,
                  events => \@events,
                 };

  if ($return_value->{completed_successfully} && !$return_value->{error_msg} && !$params{dry_run}) {

      my $iterator = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh}, get_criteria => { is_active_policy => 1 });

      while(my $policy = $iterator->get_next) {
         $self->send_policy_signature_set_event($session, $call_flags, $policy);
      }

      if (! $sigfile_update_settings->{check_interval}) {
          $call_flags->{relayable} = 1;
      }
      my $description = "Signatures updated successfully - $rv->{sig_count} signatures loaded ($rv->{sigs_updated_count} updated).";
      $description .= " Server Technologies updated successfully - $rv->{server_technologies_updated_count} server technologies updated, $rv->{server_technologies_added_count} server technologies added, $rv->{server_technologies_deleted_count}  server technologies deleted.";
      push(@{$retval->{events}},
           $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                               session    => $session,
                               element    => $return_value,
                               call_flags => $call_flags,
                               user_name  => $session->{user_name},
                               description => $description,
                              ));
  }
  elsif ($rv->{unscheduled} && !$params{dry_run}) {
       push(@{$retval->{events}},
           $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                               session    => $session,
                               element    => $return_value,
                               call_flags => $call_flags,
                               user_name  => $session->{user_name},
                               suppress_save_to_log => 1,
                              ));
  }
  return $retval;
}

#--------------------------------------------------------------------
sub check_for_sigfile_update{
   my($self, $session, $call_flags) = @_;

   my $dbh = $self->{dbh};
   my $auto_download = F5::NegativeSignatures::SigFile::AutoDownload->new(dbh => $dbh,
                                                                          outbound_call_timeout => $call_flags->{outbound_call_timeout},
                                                                          logger => $self->{logger},
                                                                         );

   my $is_newer_sigfile_available;

   eval {
     $is_newer_sigfile_available = $auto_download->is_newer_sigfile_available;
   };

   if ($@) {
       my $error_msg = "Could not reach the update server. Error: $@";
       return $self->handle_error(action => 'check_for_sigfile_update',
                                  error_code   => F5::ASMConfig::ErrorCode::UNREACHABLE_HOST(),
                                  error_thrown => $error_msg,
                                  error_level  => 'error',
                                  call_flags => $call_flags,
                                  ignored_rv => 0,
                                 );
   }

   my $retval = { return_value => $is_newer_sigfile_available};

   return $retval;
}
#--------------------------------------------------------------------

sub generate_sigfile{
   my($self, $session, $call_flags, $place_holder, $sigfile_spec) = @_;

   my $completed_successfully = 0;
   my $dbh = $self->{dbh};
   $call_flags->{search_filters} ||= [];

   for my $sigfile_filter_field (qw (flg_is_user_defined sig_tag) ) {
       if (defined $sigfile_spec->{$sigfile_filter_field}) {
           push (@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new(field_name => $sigfile_filter_field,
                                                                                    field_value => $sigfile_spec->{$sigfile_filter_field},
                                                                                    field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ())
                );
       }
   }
   my $iterator = F5::ASMConfig::Entity::Signature->retrieve_all(dbh => $dbh, 
                                                                 get_criteria => {}, 
                                                                 extra_fields => ['packed_rule_obj'],
                                                                 search_filters => $call_flags->{search_filters});
   my $sigfile = F5::NegativeSignatures::SigFile->new(dbh => $dbh,
                                                      sigfile_format => uc($sigfile_spec->{file_format} || 'XML'),
                                                      output_filename => $sigfile_spec->{filename},
                                                      sig_tag => $sigfile_spec->{tag},
                                                      flg_is_user_defined => $sigfile_spec->{flg_is_user_defined},
                                                      sig_report_fields => $call_flags->{selected_fields},
                                                     );

   $completed_successfully = $sigfile->create_sigfile(sig_iterator => $iterator);

   my $retval = { return_value => $completed_successfully };

   return $retval;
}
#--------------------------------------------------------------------
sub get_sigfile_version{
   my($self, $session, $call_flags) = @_;
	
   my $dbh = $self->{dbh};
   my $negative_sigs = F5::NegativeSignatures->new(dbh => $dbh);
   my $file_date = $negative_sigs->get_update_file_date(flg_any_version => 1);
   $file_date =~ tr| |_|;
   $file_date =~ s|[-/:]||g;

   my $retval = { return_value => $file_date};

   return $retval;
}
#--------------------------------------------------------------------
1;
