package F5::ASMConfig::Handler::Signature;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::Entity::Signature;
use F5::NegativeSignatures;
use F5::DbUtils;
use F5::NegativeSignatures::Signature;
use F5::DateTimeUtils;

our @handled_functions = (
                           {
                             function_name => 'add_signature',
                             relayable => 1,
                             store_index => 3,
			     exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'update_signature',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
		             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'delete_signature',
                             relayable => 1,
			     exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'delete_many_signatures',
                             relayable => 1,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'search_signature',
                           },

                         );

our $entity_class = 'F5::ASMConfig::Entity::Signature';
my $max_sig_name_length = 254;
my $max_sig_tag_length = 50;

#----------------------------------------------------------------------
sub add_signature{
  my($self, $session, $call_flags, $place_holder, $signature) = @_;

  if (!defined($signature->{sig_name}) || !length($signature->{sig_name}) || ! ($signature->{sig_name} =~ /\S/)) {
       return $self->handle_error(action => 'add',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                                 thrift_obj   => $signature,
                                 error_thrown => "Signature name is a required field",
                                );
  }

  if (length($signature->{sig_name}) > $max_sig_name_length ) {
       return $self->handle_error(action => 'add',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                                 thrift_obj   => $signature,
                                 error_thrown => "Maximum allowed length for signature name is $max_sig_name_length",
                                );
  }

  if (defined($signature->{sig_tag}) && length($signature->{sig_tag}) > $max_sig_tag_length ) {
       return $self->handle_error(action => 'add',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                                 thrift_obj   => $signature,
                                 error_thrown => "Signature tag is limited to 50 characters",
                                );
  }


  my $dbh = $self->{dbh};

  for my $unique_field (qw(sig_name sig_id)) {
      next unless $signature->{$unique_field};
      my %name_tag;
      if ( $unique_field eq "sig_name" ){
          $name_tag{sig_tag} = $signature->{sig_tag};
          $name_tag{sig_tag_policy} = $signature->{sig_tag_policy};
      }

      my $check_obj = F5::ASMConfig::Signature->new($unique_field => $signature->{$unique_field}, %name_tag);
      my $added_sig = F5::ASMConfig::Entity::Signature->new(dbh => $dbh,
                                                            get_criteria => $check_obj);
      if ($added_sig) {
          return $self->handle_error(action => 'add',
                                     entity_class => $entity_class,
                                     error_code   => F5::ASMConfig::ErrorCode::DUPLICATE_RECORD(),
                                     thrift_obj   => $check_obj,
                                     error_thrown => $F5::ASMConfig::Handler::DUPLICATE_FOUND_DBI_ERROR,
                                     error_level  => 'info',
                                     empty_string_value => '',
                                    );
      }
  }

  my %sig_fields = %$signature;
  $sig_fields{flg_is_user_defined} = 1;
  $sig_fields{attack_type} = defined $signature->{attack_type_name}? $signature->{attack_type_name} : $signature->{attack_type_id};
  $sig_fields{sig_desc} ||= '';
  $sig_fields{apply_to} ||= 'Request';
  $sig_fields{risk} ||= 1;
  $sig_fields{accuracy} ||= 1;
  

  if (!$sig_fields{attack_type}) {
       return $self->handle_error(action => 'add',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                                 thrift_obj   => $signature,
                                 error_thrown => "Attack Type is a required field",
                                );
  }

  if ($signature->{systems}){
      for my $sys (@{$signature->{systems}}){
        push @{$sig_fields{sig_systems}}, $sys->{system_id};
      }
  }

  my $update_if_change_response =  update_if_changed( self => $self,
                                     action => 'add_signature',
                                     sig_fields => \%sig_fields,
                                     dbh => $dbh,
                                     session => $session,
                                     call_flags => $call_flags,
                                   );
  return $update_if_change_response;

}
#----------------------------------------------------------------------
sub update_signature{
  my($self, $session, $call_flags, $place_holder, $signature_criteria, $store_signature) = @_;
  my $dbh =$self->{dbh};
 
  my $get_criteria;
  if ($signature_criteria->{sig_id}) {
      $get_criteria->{sig_id} = $signature_criteria->{sig_id};
  }
  else  {
      $get_criteria->{sig_name} = $signature_criteria->{sig_name};
  }

  my $updated_sig = F5::ASMConfig::Entity::Signature->new(dbh => $dbh,
                                                          get_criteria => $get_criteria);

  if (!$updated_sig) {
     return $self->handle_error(action => 'update',
                                entity_class => $entity_class,
                                error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                thrift_obj   => $signature_criteria,
                                error_thrown => $F5::ASMConfig::Handler::NONEXISTENT_RECORD_ERROR,
                                error_level  => 'info',
                               );
  }

  if (! $updated_sig->{flg_is_user_defined}){
     return $self->handle_error(action => 'update',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                 entity       => $updated_sig,
                                 thrift_obj   => $signature_criteria,
                                 error_thrown => "User may not change factory signature",
                                 error_level  => 'info',
                                );
  }

  if ($updated_sig->{sig_tag}) {
    my $tag_timestamp = F5::DateTimeUtils::convert_epoch_to_rfc3339(time());
    eval {
      my $sigfile = F5::NegativeSignatures::SigFile->new(dbh => $dbh);
      $sigfile->validate_signature_requirements_for_policies(
                                                              sig_tag => $updated_sig->{sig_tag},
                                                              timestamp => $tag_timestamp,
                                                              flg_is_user_defined => 1
                                                            );
    };
    if ($@) {
      my $error_msg = $@;
      if (ref($error_msg)) {
        $error_msg = $error_msg->{error_msg};
      }
      $error_msg =~ s/\n/ /g;
      return $self->handle_error(
                                  action       => 'update',
                                  entity_class => $entity_class,
                                  error_code   => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                                  entity       => $updated_sig,
                                  thrift_obj   => $signature_criteria,
                                  error_thrown => $error_msg,
                                  error_level  => 'error',
                                );
    }
  }

  if (! $store_signature->{sig_name}) {
       $store_signature->{sig_name} = $updated_sig->{sig_name};
  }


  my %sig_fields;
  if ($store_signature->{attack_type_name}){
    $sig_fields{attack_type} = $store_signature->{attack_type_name};
  }
  elsif ( $store_signature->{attack_type_id}){
    $sig_fields{attack_type} = $store_signature->{attack_type_id};
  }
  else{
    $sig_fields{attack_type} = $updated_sig->{attack_type_name};
  }

  my $fields = F5::ASMConfig::Entity::Signature->fields();

  my %all_sig_fields = map { ($_ => 1) } (keys %$store_signature, keys %$updated_sig);
  for my $field (keys %all_sig_fields) {
    if (defined $store_signature->{$field}
        && !(($call_flags->{rest_call} && $field eq 'rest_uuid') || ($fields->{$field}->{rest_unique_record_part} || $fields->{$field}->{unique_record_part} || $fields->{$field}->{read_only}))) {
       $sig_fields{$field} = $store_signature->{$field};
    }
    else {
       $sig_fields{$field} = $updated_sig->{$field};
    }
  }

  my @systems = defined $store_signature->{systems} ? @{$store_signature->{systems}} : 
                defined  $updated_sig->{systems} ? @{$updated_sig->{systems}} : (); 
      

  for my $sys (@systems){
      push @{$sig_fields{sig_systems}}, $sys->{system_id};
  }

  my $update_if_change_response = update_if_changed( self => $self,
                                                     action => 'update_signature',
                                                     sig_fields => \%sig_fields,
                                                     dbh => $dbh,
   			                             session => $session,
                                                     call_flags => $call_flags,
                                                   );
  return $update_if_change_response; 
}
#----------------------------------------------------------------------
sub update_if_changed{
  my (%params) = @_;

  my $self = $params{self};
  my $action = $params{action};
  my $sig_fields = $params{sig_fields};
  my $dbh = $params{dbh};
  my $session = $params{session};
  my $call_flags = $params{call_flags};
  my $rv;
  my $updated_sig;

  eval{
       $rv = F5::NegativeSignatures::Signature->update_if_changed(dbh => $dbh,
                                                                  sig_fields => $sig_fields,
                                                                  lookup_by => 'sig_id',
                                                                  find_affected_policies => 1,
                                                                 );
  };

  if ($@){
     return $self->handle_error(action => $action,
                                entity_class => $entity_class,
                                error_thrown => $@,
                               );
     
  }

  $updated_sig = F5::ASMConfig::Entity::Signature->new(dbh => $dbh,
                                                       get_criteria => { sig_id => $rv->{sig_id}});

  my @events;
  my $retval = { return_value => $updated_sig,
                 events => \@events,
                };

  if ($rv->{affected_policies}){
      for my $affected_policy (@{$rv->{affected_policies}}) {

          my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                       get_criteria => { policy_id => $affected_policy} );
          if ($policy->{is_active_policy}) {
              $self->send_policy_signature_set_event($session, $call_flags, $policy);
          }
      }

      eval {
          my $sigfile_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
          $sigfile_update->set_sigs_active(asm_config_handler => $self,
                                           single_sig_id => $updated_sig->{sig_id},
                                           affected_policies => $rv->{affected_policies},
                                          );
      };
      if ($@){
           return $self->handle_error(action => $action,
                                      entity_class => $entity_class,
                                      error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                      entity       => $updated_sig,
                                      error_thrown => "Could not set active: $@",
                                      error_level  => 'error',
                                     );
      }
  }
  return $retval;
}
#----------------------------------------------------------------------
sub delete_signature {
  my($self, $session, $call_flags, $place_holder, $signature) = @_;
  my $dbh = $self->{dbh};

  if (!$signature->{sig_id}) {
      $signature = $entity_class->new(dbh => $dbh,
                                      get_criteria => $signature,
                                      call_flags => $call_flags);
  }

  my @affected_policies = F5::NegativeSignatures->new(dbh => $dbh)->get_policies_using_signatures($signature->{sig_id});
  my $rv = $self->base_delete(entity_class => $entity_class,
                              rpc_params => [ $session, $call_flags, $place_holder, $signature],);

  if (@affected_policies) {
      eval {
         my $sigfile_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
         $sigfile_update->set_sigs_active(asm_config_handler => $self,
                                          affected_policies => \@affected_policies,);
      };
      if ($@){
          return $self->handle_error(action => 'delete',
                                     entity_class => $entity_class,
                                     error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                     thrift_obj   => $signature,
                                     error_thrown => "Could not set active: $@",
                                     error_level  => 'error',
                                    );
      }
  }
  
  return $rv;
}

#----------------------------------------------------------------------

sub delete_many_signatures {
  my($self, $session, $call_flags, $place_holder, $signature_criteria) = @_;
  my $dbh = $self->{dbh};

  if (!defined $call_flags){
    $call_flags = F5::ASMConfig::CallFlags->new();
  }

  if (!defined $call_flags->{search_filters}){
    $call_flags->{search_filters} = [];
  }

  push(@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new({ field_name => 'flg_is_user_defined',
                                                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                                            field_value => '1'}));

  my @affected_policies;
  if (! $call_flags->{skip_set_sigs_active}) {
      my @signatures = $entity_class->retrieve_all(dbh => $dbh,
                                                   search_filters => $call_flags->{search_filters},
                                                   selected_fields => ['sig_id'],
                                                   minimal_load => 1,
                                                   get_criteria => $signature_criteria,);
      my @sig_ids = map { $_->{sig_id} } @signatures;
      @affected_policies = F5::NegativeSignatures->new(dbh => $dbh)->get_policies_using_signatures(\@sig_ids);
  }

  my $rv = $self->base_delete_many(entity_class => $entity_class,
                                   rpc_params => [ $session, $call_flags, $place_holder, $signature_criteria],);

  if (@affected_policies) {
     eval {
         my $sigfile_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
         $sigfile_update->set_sigs_active(asm_config_handler => $self,
                                          affected_policies  => \@affected_policies,
                                          skip_set_active    => $call_flags->{skip_set_active});
      };
      if ($@) {
          return $self->handle_error(action => 'delete_many',
                                     entity_class => $entity_class,
                                     error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                     thrift_obj   => $signature_criteria,
                                     error_thrown => "Could not set active: $@",
                                     error_level  => 'error',
                                    );
      }
  }

  return $rv;
}
#----------------------------------------------------------------------

sub search_signature{
  my($self, $session, $call_flags, $place_holder, $signature_criteria) = @_;
  my $is_hidden_flag_defined = 0;

  if ( defined $signature_criteria->{flg_is_hidden} ) {
     $is_hidden_flag_defined = 1;
  }
  elsif (defined $call_flags->{search_filters}){
      for my $search_filter (@{$call_flags->{search_filters}}){
          if ( $search_filter->{field_name} eq 'flg_is_hidden' ) {
             $is_hidden_flag_defined = 1;
          }
      }
  }

  $signature_criteria->{flg_is_hidden} = $is_hidden_flag_defined ?  $signature_criteria->{flg_is_hidden} : 0;

  return $self->base_search(rpc_params => [$session, $call_flags, undef, $signature_criteria],
                            entity_class => $entity_class,
                           );
}
#----------------------------------------------------------------------
1;
