package F5::ASMConfig::Handler::SignatureSet;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::SignatureSet;
use F5::NegativeSignatures;
use F5::NegativeSignatures::Set;
use F5::NegativeSignatures::SigFile::Update;
use F5::ObjUtils qw(dclone);

our @handled_functions = (
                           {
                             function_name => 'add_signature_set',
                             relayable => 1,
                             store_index => 3,
			     exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'import_signature_set',
                             relayable => 1,
                             store_index => 3,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'update_signature_set',
                             relayable => 1,
                             criteria_index => 3,
                             store_index    => 4,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'delete_signature_set',
                             relayable => 1,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'delete_many_signature_sets',
                             relayable => 1,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::SignatureSet';

#----------------------------------------------------------------------
sub add_signature_set {
  my($self, $session, $call_flags, $place_holder, $signature_set) = @_;
   
  return add_signature_set_helper(self => $self,
                                  session => $session,
                                  call_flags => $call_flags,
                                  signature_set => $signature_set,
                                  mode => 'add');
}
#----------------------------------------------------------------------
sub import_signature_set {
  my($self, $session, $call_flags, $place_holder, $signature_set) = @_;

  return add_signature_set_helper(self => $self, 
	                          session => $session, 
			          call_flags => $call_flags, 
   				  signature_set => $signature_set,
				  mode => 'import');
}
#----------------------------------------------------------------------
sub add_signature_set_helper {
  my (%params) = @_;
  my $self = $params{self};
  my $session = $params{session};
  my $call_flags = $params{call_flags};
  my $signature_set = $params{signature_set};
  my $mode = $params{mode};
 
  if (!defined($signature_set->{set_name}) || !length($signature_set->{set_name}) || ! ($signature_set->{set_name} =~ /\S/) ) {
       return $self->handle_error(action => $mode,
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                                 thrift_obj   => $signature_set,
                                 error_thrown => "Signature set name is a required field",
                                );
  }

  if (defined $signature_set->{flg_is_user_defined} && !$signature_set->{flg_is_user_defined}){
      return $self->handle_error(action => $mode,
                                  entity_class => $entity_class,
                                  error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                  thrift_obj   => $signature_set,
                                  error_thrown => "User may not create factory signature set",
                                  error_level  => 'info',
                                 );
  }


  my $dbh =$self->{dbh};

  if ($mode eq 'add') {
      my $added_set = F5::ASMConfig::Entity::SignatureSet->new(dbh => $dbh,
                                                            get_criteria => { set_name => $signature_set->{set_name}});
      if($added_set) {
          return $self->handle_error(action => 'add',
                                     entity_class => $entity_class,
                                     error_code   => F5::ASMConfig::ErrorCode::DUPLICATE_RECORD(),
                                     entity       => $added_set,
                                     thrift_obj   => $signature_set,
                                     error_thrown => $F5::ASMConfig::Handler::DUPLICATE_FOUND_DBI_ERROR,
                                     error_level  => 'info',
                                    );
      }
  }

  my %set_fields = %$signature_set;
  delete $set_fields{systems};
  
  $set_fields{set_flg_is_user_defined} = 1;
  delete $set_fields{flg_is_user_defined};

  if ($set_fields{set_flg_is_user_defined}) {
      $set_fields{set_group_name} = "User-defined";
      $signature_set->{set_group_name} = "User-defined";
  }

  if ( !exists $set_fields{flg_is_hidden} ){
      $set_fields{flg_is_hidden} = 0;
  }

  if (defined $signature_set->{signature_set_filter}->[0]){
      my $sig_tag_op = $signature_set->{signature_set_filter}->[0]->{sig_tag_op};
      my $sig_tag_val = $signature_set->{signature_set_filter}->[0]->{sig_tag_val};

      if ( defined $sig_tag_op && $sig_tag_op eq 'eq' &&
           ( (not defined $sig_tag_val) || !length($sig_tag_val)  || !($sig_tag_val =~ /\S/) ) ) {
               return $self->handle_error(action => $mode,
                                          entity_class => $entity_class,
                                          error_code   => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                                          thrift_obj   => $signature_set,
                                          error_thrown => "Signature tag value in filter is a required field when defined signature tag filter 'eq'.",
                                );
      }

      %set_fields = (%set_fields, %{$signature_set->{signature_set_filter}->[0]});
  }

  delete $set_fields{signature_set_filter};

  $set_fields{attack_type} = $signature_set->{signature_set_filter}->[0]->{attack_type_id};
  delete $set_fields{attack_type_id};

  if ($signature_set->{systems}){
      for my $sys (@{$signature_set->{systems}}){
        push @{$set_fields{set_systems}}, $sys->{system_id};
      }
  }

  if ($set_fields{flg_is_manual}){ # manual set
     return add_manual_set($self, $session, $call_flags, $signature_set);
  }

  delete $set_fields{sigs};

  my $update_id_changed_response =  update_if_changed( self => $self,
                                                       action => $mode . '_signature_set',
                                                       set_fields => \%set_fields,
                                                       dbh => $dbh,
                                                       session => $session,
                                                       call_flags => $call_flags,
                                                       signature_set => $signature_set,
                                                       mode => $mode,
                                                     );

  return $update_id_changed_response;
}

#----------------------------------------------------------------------
sub add_manual_set{
  my ($self, $session, $call_flags, $signature_set) = @_;

  if (!defined($signature_set->{sigs}) || !@{$signature_set->{sigs}}) {
       return $self->handle_error(action => 'add',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                                 thrift_obj   => $signature_set,
                                 error_thrown => "At least one attack signature must be assigned.",
                                );
  }

  delete $signature_set->{signature_set_filter};
  $signature_set->{flg_is_user_defined} = 1;

  return $self->base_add(entity_class => $entity_class,
                         rpc_params => [ $session, $call_flags, undef, $signature_set ]);
}
#----------------------------------------------------------------------
sub update_signature_set{
  my($self, $session, $call_flags, $place_holder, $signature_set_criteria, $store_signature_set) = @_;
  my $dbh = $self->{dbh};
  my $new_signature_set = dclone($store_signature_set);
  my $get_criteria;
  if ($signature_set_criteria->{set_id}) {
      $get_criteria->{set_id} = $signature_set_criteria->{set_id};
  }
  else  {
      $get_criteria->{set_name} = $signature_set_criteria->{set_name};
  }

  my $original_set = F5::ASMConfig::Entity::SignatureSet->new(dbh => $dbh,
                                                             get_criteria => $get_criteria);
  if (!$original_set) {
      return $self->handle_error(action => 'update',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                 thrift_obj   => $signature_set_criteria,
                                 error_thrown => $F5::ASMConfig::Handler::NONEXISTENT_RECORD_ERROR,
                                 error_level  => 'info',
                                );
  }

  if (! $original_set->{flg_is_user_defined}){
      return $self->handle_error(action => 'update',
                                  entity_class => $entity_class,
                                  error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                  entity       => $original_set,
                                  thrift_obj   => $signature_set_criteria,
                                  error_thrown => "User may not change factory signature set", 
                                  error_level  => 'info',
                                 );
  }

  if (defined $new_signature_set->{flg_is_user_defined} && !$new_signature_set->{flg_is_user_defined}){
      return $self->handle_error(action => 'update',
                                  entity_class => $entity_class,
                                  error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                  entity       => $original_set,
                                  thrift_obj   => $signature_set_criteria,
                                  error_thrown => "User may not create factory signature set",
                                  error_level  => 'info',
                                 );
  }

  my $fields = F5::ASMConfig::Entity::SignatureSet->fields();
  for my $field (keys %$new_signature_set, keys %$original_set) {
    if (!defined($new_signature_set->{$field})
       || ($call_flags->{rest_call} && ( ($field eq 'rest_uuid' || $fields->{$field}->{unique_record_part} || $fields->{$field}->{rest_unique_record_part} || $fields->{$field}->{read_only})))) {
       $new_signature_set->{$field} = $original_set->{$field};
    }
  }

  if ($new_signature_set->{flg_is_manual}){ # manual set
     my $update_manual_response = update_manual_set(self => $self, 
                                                    session => $session, 
                                                    call_flags => $call_flags, 
                                                    signature_set => $new_signature_set,
                                                    original_set => $original_set,
                                                    signature_set_criteria => $signature_set_criteria );
    return $update_manual_response; 
  }

  if ($original_set->{sigs}){
    $dbh->do(<<EndQuery,undef,$original_set->{set_id});
DELETE FROM $self->{plc_dbname}.NEGSIG_SET_SIGNATURES
 WHERE set_id = ?
EndQuery
  }
  
  my %set_fields;

  $set_fields{attack_type} = $new_signature_set->{signature_set_filter}->[0]->{attack_type_id}; 

  $set_fields{set_flg_is_hidden} =  $new_signature_set->{flg_is_hidden};
  $set_fields{set_flg_is_user_defined} = 1;

  for my $field (keys %$new_signature_set){
    if ($field eq 'signature_set_filter' && defined $new_signature_set->{signature_set_filter}->[0]){
      if ($call_flags->{rest_call}) {
          my $original_set_filter = $original_set->{signature_set_filter} && $original_set->{signature_set_filter}->[0] ? $original_set->{signature_set_filter}->[0] : {};
          %set_fields = (%set_fields, %$original_set_filter);
          delete $set_fields{attack_type_id};
      }
      for my $filter_field ( keys %{$new_signature_set->{signature_set_filter}->[0]}){
        if ($filter_field ne 'attack_type_id'){
           $set_fields{$filter_field} = $new_signature_set->{signature_set_filter}->[0]->{$filter_field};
        }
      } 
    }
    elsif ($field ne 'systems' && $field ne 'flg_is_user_defined'){
       $set_fields{$field} = $new_signature_set->{$field};
    }
  }

  if (defined $new_signature_set->{systems} ){
     @{$set_fields{set_systems}} = map { $_->{system_id} } @{$new_signature_set->{systems}};
  }

  delete $set_fields{sigs};

  my $update_if_change_response = update_if_changed( self => $self,
                                                     action => 'update_signature_set',
                                                     set_fields => \%set_fields,
                                                     dbh => $dbh,
                                                     session => $session,
                                                     call_flags => $call_flags,
                                                     signature_set => $signature_set_criteria,
                                                   );
  return $update_if_change_response;
}
#----------------------------------------------------------------------
sub update_manual_set {

 my (%params) = @_;
 my $self = $params{self};
 my $dbh = $self->{dbh};
 my $session = $params{session};
 my $call_flags = $params{call_flags};
 my $signature_set = $params{signature_set};
 my $signature_set_criteria = $params{signature_set_criteria};
 my $original_set = $params{original_set};
 my $set_id = $original_set->{set_id};
 my $flg_set_changed = 0;

 if (! $original_set->{flg_is_manual}) { #type changed from filter based to manual
  $dbh->do(<<EndQuery,undef,$set_id);
DELETE FROM $self->{plc_dbname}.NEGSIG_SET_FILTERS
 WHERE set_id = ?
EndQuery
 $flg_set_changed = 1;
 }

 if (defined $original_set->{systems}){
  $dbh->do(<<EndQuery,undef,$set_id);
DELETE FROM $self->{plc_dbname}.NEGSIG_SET_FILTER_SYSTEMS
 WHERE set_id = ?
EndQuery
 $flg_set_changed = 1;
 }

 if ($original_set->{flg_is_manual}){ # original type was also manual
     if ($original_set->{sigs} && $signature_set->{sigs}){ # both lists are not empty
         if (scalar @{$original_set->{sigs}} != scalar @{$signature_set->{sigs}}){
             $flg_set_changed = 1;
         }
         else{
             my %original_sigs = map {$_,1} @{$original_set->{sigs}};

             for my $new_sig ( @{$signature_set->{sigs}} ) {
                 if (!$original_sigs{$new_sig}){
                     $flg_set_changed = 1;
                     last;
                 }
             }
         }
     }
     elsif( ($original_set->{sigs} && !$signature_set->{sigs}) || 
             (!$original_set->{sigs} && $signature_set->{sigs}) ){ # only one list is empty
         $flg_set_changed = 1;
     }
 }

 if (!defined($signature_set->{sigs}) || !@{$signature_set->{sigs}}) {
       return $self->handle_error(action => 'update',
                                 entity_class => $entity_class,
                                 error_code   => F5::ASMConfig::ErrorCode::MISSING_FIELD(),
                                 thrift_obj   => $signature_set,
                                 error_thrown => "At least one attack signature must be assigned.",
                                );
 }


  delete $signature_set->{signature_set_filter};
  delete $signature_set->{systems};
  $signature_set->{set_flg_is_user_defined} = 1;

  my $rv = $self->base_update(entity_class => $entity_class,
                           rpc_params => [ $session, $call_flags, undef, $signature_set_criteria, $signature_set ]);

 if ($flg_set_changed){

    my $negsig_set = F5::NegativeSignatures::Set->new(dbh => $dbh, set_id => $set_id);
    my @affected_policies = F5::NegativeSignatures->new($self->{dbh})->get_policies_using_sets($set_id);

    if (@affected_policies) {
        $negsig_set->mark_policies_modified(policies => \@affected_policies,
                                            modify_action => 'update');
   
        append_policy_events(  handler => $self,
                               retval => $rv, 
                               session => $session, 
                               call_flags => $call_flags,
                               event_type => F5::ASMConfig::EventType::UPDATE,
                               affected_policies => \@affected_policies,
                               set_id => $signature_set->{set_id});
  

        eval {
            my $sigfile_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
            $sigfile_update->set_sigs_active(asm_config_handler => $self,
                                             affected_policies => \@affected_policies,);
        };
        if ($@) {
            return $self->handle_error(action => 'update',
                                       entity_class => $entity_class,
                                       error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                       thrift_obj   => $signature_set,
                                       error_thrown => "Could not set active: $@",
                                       error_level  => 'error',
                                      );
        }
    }
 }
 
 return $rv;
}
#----------------------------------------------------------------------
sub delete_signature_set{
  my($self, $session, $call_flags, $place_holder, $signature_set) = @_;
  my $dbh = $self->{dbh};

  if (!$signature_set->{set_id}) {
      $signature_set = $entity_class->new(dbh => $dbh,
                                          get_criteria => $signature_set,
                                          call_flags => $call_flags);
                                                                
  }
  my @affected_policies = F5::NegativeSignatures->new($self->{dbh})->get_policies_using_sets($signature_set->{set_id});

  my $rv = $self->base_delete(entity_class => $entity_class,
                              rpc_params => [ $session, $call_flags, $place_holder, $signature_set],);

  if (@affected_policies) {
      eval {
         my $sigfile_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
         $sigfile_update->set_sigs_active(asm_config_handler => $self,
                                          affected_policies => \@affected_policies,);
      };
      if ($@) {
         return $self->handle_error(action => 'update',
                                     entity_class => $entity_class,
                                     error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                     thrift_obj   => $signature_set,
                                     error_thrown => "Could not set active: $@",
                                     error_level  => 'error',
                                    );
      }
  }

  return $rv;
}
#----------------------------------------------------------------------
sub delete_many_signature_sets{
  my($self, $session, $call_flags, $place_holder, $signature_set_criteria ) = @_;
  my $dbh = $self->{dbh};

  if (!defined $call_flags){
    $call_flags = F5::ASMConfig::CallFlags->new();
  }

  if (!defined $call_flags->{search_filters}){
    $call_flags->{search_filters} = [];
  }

  push(@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new({ field_name => 'flg_is_user_defined',
                                                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                                            field_value => '1'}));

  my @affected_policies;
  if (! $call_flags->{skip_set_sigs_active}) {
      my @signature_sets = $entity_class->retrieve_all(dbh => $self->{dbh},
                                                       call_flags => $call_flags,
                                                       search_filters => $call_flags->{search_filters},
                                                       selected_fields => ['set_id'],
                                                       minimal_load => 1,
                                                       get_criteria => $signature_set_criteria);
      if (@signature_sets) {
          my @sig_set_ids = map { $_->{set_id} } @signature_sets;
          @affected_policies = F5::NegativeSignatures->new($self->{dbh})->get_policies_using_sets(\@sig_set_ids);
      }
  }

  my $rv = $self->base_delete_many(entity_class => $entity_class,
                                   rpc_params => [ $session, $call_flags, $place_holder, $signature_set_criteria],);

  if (@affected_policies) {
      eval {
          my $sigfile_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
          $sigfile_update->set_sigs_active(asm_config_handler => $self,
                                           affected_policies => \@affected_policies,);
      };
      if ($@) {
         return $self->handle_error(action => 'update',
                                    entity_class => $entity_class,
                                    error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                    thrift_obj   => $signature_set_criteria,
                                    error_thrown => "Could not set active: $@",
                                    error_level  => 'error',
                                   );
      }
  }

  return $rv;
}

#----------------------------------------------------------------------
sub update_if_changed{
  my (%params) = @_;
  my $self = $params{self};
  my $action = $params{action};
  my $set_fields = $params{set_fields};
  my $dbh = $params{dbh};
  my $session = $params{session};
  my $call_flags = $params{call_flags};
  my $signature_set = $params{signature_set};
  my $mode = $params{mode};

  my $rv;
  my $updated_set;

  eval{
        $rv = F5::NegativeSignatures::Set->update_if_changed(dbh => $dbh,
                                                             set_fields => $set_fields,
 							     mode => $mode, 
                                                             find_affected_policies => 1,
	                                                     lookup_by => $mode eq 'add' ? 'set_id' : 'set_name');
  };

  if ($@){
     return $self->handle_error(action => $action,
                                entity_class => $entity_class,
                                thrift_obj   => $set_fields,
                                error_thrown => $@,
                                error_level  => 'error',
                               );
  }

  $updated_set = F5::ASMConfig::Entity::SignatureSet->new(dbh => $dbh,
                                                          get_criteria => { set_id => $rv->{new_set_id}});

  my @events;
  my $event_type =  $action eq 'update_signature_set' ? F5::ASMConfig::EventType::UPDATE :  F5::ASMConfig::EventType::ADD;
  push(@events,
         $self->create_event(event_type =>  $event_type,
                                session => $session,
                                element => F5::ASMConfig::Entity::SignatureSet->new(dbh => $dbh,
                                                                                    get_criteria => $signature_set),
                                call_flags => $call_flags,
                                user_name => $session->{user_name},
                             ));

  my $retval = { return_value => $updated_set,
                 events => \@events,
                };


  if ($rv->{affected_policies} && @{$rv->{affected_policies}}){
      append_policy_events(  handler => $self,
                             retval => $retval,
                             session => $session,
                             call_flags => $call_flags,
                             event_type => F5::ASMConfig::EventType::UPDATE,
                             affected_policies => $rv->{affected_policies},
                             set_id => $signature_set->{set_id});
      eval{
          my $sigfile_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
          $sigfile_update->set_sigs_active(asm_config_handler => $self,
                                           affected_policies => $rv->{affected_policies},);
      };
      if ($@){
           return $self->handle_error(action => $action,
                                      entity_class => $entity_class,
                                      error_code   => F5::ASMConfig::ErrorCode::SET_ACTIVE_FAIL(),
                                      entity       => $updated_set,
                                      error_thrown => "Could not set active: $@",
                                      error_level  => 'error',
                                     );
      }
  }

  return $retval;

}
#----------------------------------------------------------------------
sub append_policy_events {
  my(%params) = @_;
 
  my $handler = $params{handler};
  my $retval = $params{retval};
  my $event_type = defined $params{event_type} ? $params{event_type} : F5::ASMConfig::EventType::UPDATE ;
 
  my @affected_policies = $params{affected_policies} ? @{$params{affected_policies}} 
                                                     : F5::NegativeSignatures->new($handler->{dbh})->get_policies_using_sets($params{set_id});


  for my $affected_policy (@affected_policies) {
      my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $handler->{dbh},
                                                    get_criteria => { policy_id => $affected_policy});

      my $policy_sigset = F5::ASMConfig::Entity::PolicySignatureSet->new(dbh => $handler->{dbh}, 
                                                                         policy => $policy,
                                                                         get_criteria => { set_id => $params{set_id}});


      push(@{$retval->{events}},
          $handler->create_event(event_type => $event_type,
                                 session => $params{session},
                                 policy => $policy,
                                 element =>  $policy_sigset,
                                 call_flags => $params{call_flags},
                                 user_name => $params{session}->{user_name},
                             ));
   }
 
}
#----------------------------------------------------------------------
1;
