package F5::ASMConfig::Handler::Suggestion;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;

our @handled_functions = (
                           {
                             function_name => 'add_suggestion',
                             store_index => 3,
                             relayable => 0,
                           },
                           {
                             function_name => 'update_suggestion',
                             criteria_index => 3,
                             store_index => 4,
                             write_access => 1,
                             relayable => 0,
                           },
                           {
                             function_name => 'update_many_suggestions',
                             criteria_index => 3,
                             store_index => 4,
                             write_access => 1,
                             relayable => 0,
                           },
                           {
                             function_name => 'delete_suggestion',
                             write_access => 1,
                             relayable => F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_CHECK_RELAYABLE(),
                           },
                           {
                             function_name => 'delete_many_suggestions',
                             write_access => 1,
                             relayable => F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_CHECK_RELAYABLE(),
                           },
                         );

our $entity_class = 'F5::ASMConfig::Entity::Suggestion';

#----------------------------------------------------------------------

my %element_type_to_charset_mapping = (
                                  F5::ASMConfig::ElementType::HEADER() => F5::ASMConfig::CharsetType::CHARSET_TYPE_HEADER(),
                                  F5::ASMConfig::ElementType::XML_PROFILE() => F5::ASMConfig::CharsetType::CHARSET_TYPE_XML_PROFILE(),
                                  F5::ASMConfig::ElementType::JSON_PROFILE() => F5::ASMConfig::CharsetType::CHARSET_TYPE_JSON_PROFILE(),
                                  F5::ASMConfig::ElementType::GWT_PROFILE() => F5::ASMConfig::CharsetType::CHARSET_TYPE_GWT_PROFILE(),
                                  F5::ASMConfig::ElementType::PLAIN_TEXT_PROFILE() => F5::ASMConfig::CharsetType::CHARSET_TYPE_PLAIN_TEXT_PROFILE(),
                                  F5::ASMConfig::ElementType::PARAMETER() => F5::ASMConfig::CharsetType::CHARSET_TYPE_PARAM_VALUE(), #May also be CHARSET_TYPE_PARAM_NAME
                                  F5::ASMConfig::ElementType::_URL() => F5::ASMConfig::CharsetType::CHARSET_TYPE_URI(),
                                  F5::ASMConfig::ElementType::WEBSOCKET_URL() => F5::ASMConfig::CharsetType::CHARSET_TYPE_URI(),
                               );


sub accept_suggestions {
    my (%params) = @_;
    my $self = $params{self};
    my $suggestions = $params{suggestions};
    my $store_suggestion = $params{store_suggestion};
    my $inherited_call = $params{inherited_call};
    my $dbh = $params{dbh};
    my $suggestion_log_note = $params{suggestion_log_note};

    my $log_note = ($inherited_call ? "Parent " : "") . "Accepted Suggestion";

    if ( $suggestion_log_note ) {
        $log_note .= ": ".$suggestion_log_note;
    }

    for my $suggestion (@$suggestions) {
        my $policy = $params{policy} || F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh}, policy => $suggestion->{policy});
        if ($entity_class->is_suggestion_status_accept(object => $store_suggestion)) {

            # Change the suggestion key to make it distinct from a future suggestion for the same change.
            my $updated_suggestion = $self->update_suggestion($self->{session}, F5::ASMConfig::CallFlags->new(skip_inheritance => 1), $policy,
                                                               F5::ASMConfig::Suggestion->new(suggestion_key => $suggestion->{suggestion_key}), 
                                                               F5::ASMConfig::Suggestion->new(suggestion_key => ($suggestion->{suggestion_key} . time())));
            #Update original suggestion for return object and events.
            $suggestion->from_object(object => $updated_suggestion);

            my $call_flags = F5::ASMConfig::CallFlags->new( ignored_errors => [F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::KNOWN_DB_ERRORS()})],
                                                            append_nested_struct_on_update => ($suggestion->{suggestion_action} == F5::ASMConfig::SuggestionAction::SUGGESTION_ACTION_UPDATE_APPEND()),
                                                            remove_nested_struct_on_update => ($suggestion->{suggestion_action} == F5::ASMConfig::SuggestionAction::SUGGESTION_ACTION_UPDATE_REMOVE()),
                                                            add_on_updating_nonexistent_record => 1,
                                                            log_note => $log_note,
                                                            inherited_call => $inherited_call,
                                                            suggestion_rest_uuid => $suggestion->{rest_uuid},
                                                            cascade_disable_flags => 1,
                                                          );

            my $suggestion_element_type = $suggestion->{element}->{element_type};
            my $suggestion_element = $suggestion->{element}->{$self->{element_types}->{$suggestion_element_type}->{element_type_name}};
            my $element_class = $self->{element_types}->{$suggestion_element_type}->{class};
            my $entity;
            my $store_entity;

            if ($store_suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT_GLOBALLY()) {
                if (defined($suggestion->{metachar})) {
                    my $charset_type = $element_type_to_charset_mapping{$suggestion_element_type};
                    if ($suggestion_element_type == F5::ASMConfig::ElementType::PARAMETER() && $suggestion_element->{param_name_metachar_overrides}) {
                       $charset_type = F5::ASMConfig::CharsetType::CHARSET_TYPE_PARAM_NAME();
                    }

                    $element_class = 'F5::ASMConfig::Entity::Charset';
                    $suggestion_element = $element_class->new(dbh => $dbh, policy => $policy, 
                                                              object => { charset_type => $charset_type, 
                                                                          metachars => [ F5::ASMConfig::MetacharOverride->new(metachar => $suggestion->{metachar},
                                                                                                                              flg_is_allowed => 1) ] });
                    $suggestion_element_type = F5::ASMConfig::ElementType::CHARSET();
                }
                elsif ($suggestion->{sig_id}) {
                    $element_class = 'F5::ASMConfig::Entity::PolicySignature';
                    $suggestion_element = $element_class->new(dbh => $dbh, policy => $policy,
                                                              object => { sig_id => $suggestion->{sig_id},
                                                                          flg_disabled => 1 });
                    $suggestion_element_type = F5::ASMConfig::ElementType::POLICY_SIGNATURE();
                }
            }
            elsif ( $store_suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT_ON_MICROSERVICES() &&
                    defined($suggestion->{accepted_microservices}) && @{$suggestion->{accepted_microservices}} ){
                my $microservice_override;
                if ($suggestion_element_type == F5::ASMConfig::ElementType::HTTP_CHECK() && $suggestion_element->{http_check_code}) {
                     $microservice_override = 'http_check_overrides';
                }
                elsif ( $suggestion_element_type == F5::ASMConfig::ElementType::NORMALIZATION() && $suggestion_element->{normalization_code} ) {
                     $microservice_override = 'normalization_overrides';
                }
                elsif ( $suggestion_element_type == F5::ASMConfig::ElementType::Blocking() && $suggestion_element->{violation} ) {
                     $microservice_override = 'blocking_settings_overrides';
                }
                $element_class = 'F5::ASMConfig::Entity::Microservice';
                $entity = $element_class->new(dbh => $dbh, policy => $policy, object => {});
                $store_entity = $element_class->new(dbh => $dbh, policy => $policy,
                                                    object => {
                                                                $microservice_override => [ $suggestion_element ] ,
                                                              },
                                                   );
                $suggestion_element_type = F5::ASMConfig::ElementType::MICROSERVICE();
                $call_flags->{append_nested_struct_on_update} = 1;
                my @microservice_ids = map { $_->{microservice_id} } @{$suggestion->{accepted_microservices}};
                $call_flags->{search_filters} = [ F5::ASMConfig::SearchFilter->new(field_name => 'microservice_id',
                                                                                 field_value_list => \@microservice_ids,
                                                                                 field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),)
                                                ];
            }

            my $parent_entity;
            my $parent_diff_entity;      
            my $update_parent_function_name = $self->get_function_name(element_type => $suggestion_element_type,
                                                                       event_type => F5::ASMConfig::EventType::UPDATE(),);
            my $function_name;
            my $update_parent_diff_entity_function_name;

            #If this is a tightening suggestion (there is a parent wildcard entity, and the child does not yet exist), 
            #then clone the settings from the wildcard as the basis of the new entity.
            my $existing_entity = $element_class->new(dbh => $dbh, policy => $policy, get_criteria => $suggestion_element);
            if ($suggestion->{matched_wildcard_entity} && !$existing_entity) {

                if ($parent_entity = $element_class->new(dbh => $dbh, 
                                                         policy => $policy, 
                                                         get_criteria => $suggestion->{matched_wildcard_entity},)
                                                        ) {
                    if ($parent_entity->{has_tightening} == F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION()) {
                        $entity = $parent_entity->clone_to_explicit(dbh => $dbh, policy => $policy, object => $suggestion_element, call_flags => $call_flags);
                        $call_flags->{cloned_from_parent} = 1;
                    }
                    $parent_entity->{last_tightening_time} = time();
                }
            }           

            if (!$entity) {
                $entity = $element_class->new(dbh => $dbh, policy => $policy, object => $suggestion_element);
            }
            if (!$store_entity) {
                $store_entity = $entity;
            }

            if ($store_suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT_AND_STAGE()) {
                $entity->{in_staging} = 1;
                if (my $parent_diff_struct = $entity->parent_diff_struct(%params)) {
                    $parent_diff_entity = $entity->get_diff_parent_entity(%params, parent_diff_struct => $parent_diff_struct);
                    if (defined $parent_diff_entity->{in_staging} && !$parent_diff_entity->{in_staging}) {
                        $parent_diff_entity->{in_staging} = 1;
                        $update_parent_diff_entity_function_name = $self->get_function_name(element_type => $parent_diff_entity->element_type(),
                                                                                            event_type => F5::ASMConfig::EventType::UPDATE(),);
                    }
                }
            }

            my @missing_entities;
            my @missing_dependent_external_entities;
            if ($store_suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT_ON_MICROSERVICES()) {
                $function_name = $self->get_function_name(element_type => $suggestion_element_type,
                                                          event_type => F5::ASMConfig::EventType::UPDATE(),
                                                          is_many => 1);
            }
            elsif ($suggestion->{suggestion_action} == F5::ASMConfig::SuggestionAction::SUGGESTION_ACTION_DELETE()
                   || ($entity->parent_diff_struct && $suggestion->{suggestion_action} == F5::ASMConfig::SuggestionAction::SUGGESTION_ACTION_UPDATE_REMOVE())) {
                   $function_name = $self->get_function_name(element_type => $suggestion_element_type,
                                                             event_type => F5::ASMConfig::EventType::DELETE(),);
            }
            else {
                $function_name = $self->get_function_name(element_type => $suggestion_element_type,
                                                          event_type => F5::ASMConfig::EventType::UPDATE(),);
                #Make sure required external entities exist
                @missing_entities = @{$entity->get_missing_external_entities(existing_entity => $existing_entity, call_flags => $call_flags)};
                @missing_dependent_external_entities = @{$entity->get_missing_dependent_external_entities()};
            }

            $self->begin_work($self->{session});
            if ($parent_entity) {
                my $update_parent_call_flags = F5::ASMConfig::CallFlags->new(%$call_flags, suppress_events => 1); 
                $self->$update_parent_function_name($self->{session}, $update_parent_call_flags, $policy, $parent_entity);
            }
            for my $missing_entity (@missing_entities) {
                my $add_missing_entity_function_name = $self->get_function_name(element_type => $missing_entity->element_type(),
                                                                                event_type => F5::ASMConfig::EventType::UPDATE(),);
                $missing_entity->{in_staging} = $entity->{in_staging};

                $self->$add_missing_entity_function_name($self->{session}, $call_flags, $policy, $missing_entity);
            }
            if ($update_parent_diff_entity_function_name) {
                $self->$update_parent_diff_entity_function_name($self->{session}, $call_flags, $policy, $parent_diff_entity);
            }

            $self->$function_name($self->{session}, $call_flags, $policy, $entity, $store_entity);

            for my $missing_entity (@missing_dependent_external_entities) {
                my $add_missing_entity_function_name = $self->get_function_name(element_type => $missing_entity->element_type(),
                                                                                event_type => F5::ASMConfig::EventType::ADD(),);
                $missing_entity->{in_staging} = $entity->{in_staging};
                $self->$add_missing_entity_function_name($self->{session}, $call_flags, $policy, $missing_entity);
            }

            if ($store_suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT_GLOBALLY()) {
                #Clean up irrelevent suggestions
                my $suggestion_criteria_to_clean = F5::ASMConfig::Suggestion->new(suggestion_status => F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING(),
                                                                                  sig_id => $suggestion->{sig_id},
                                                                                  metachar => $suggestion->{metachar},
                                                                                  violation => $suggestion->{violation},
                                                                                 );
                if ($suggestion->{metachar}) {
                    $suggestion_criteria_to_clean->{element_type} = $suggestion->{element_type};
                };

                $self->delete_many_suggestions($self->{session}, $call_flags, $policy, $suggestion_criteria_to_clean);
            }

            $self->commit($self->{session});
        }
    }
}

#----------------------------------------------------------------------

sub add_suggestion {
    my($self, $session, $call_flags, $policy, $store_suggestion) = @_;

    my $retval =  $self->base_add(rpc_params => [$session, $call_flags, $policy, $store_suggestion],
                                  entity_class => $entity_class,
                                 );

    if ($retval->{events}) {
        push (@{$self->{events}}, @{delete $retval->{events}});
    }

    accept_suggestions(dbh => $self->{dbh},
                       self => $self,
                       policy => $policy,
                       inherited_call => $call_flags->{inherited_call},
                       store_suggestion => $store_suggestion,
                       suggestion_log_note => $call_flags->{log_note},
                       suggestions => [$retval->{return_value}],
                      );

    return $retval;
}

#----------------------------------------------------------------------
sub update_suggestion {
    my($self, $session, $call_flags, $policy, $suggestion_criteria, $store_suggestion) = @_;

    if ($self->{device_group} && $self->{shared_state}{is_auto_sync}) {
        #For case of accepting an ignored suggestion
        my $suggestion = $entity_class->new(dbh => $self->{dbh},
                                            policy => $policy,
                                            get_criteria => $suggestion_criteria);
        if ($suggestion && $suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_IGNORED()) {
            $call_flags->{relayable} = 1;  
        }
    }
    my $retval =  $self->base_update(rpc_params => [$session, $call_flags, $policy, $suggestion_criteria, $store_suggestion],
                                     entity_class => $entity_class,
                                    );

    if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()
        && $retval->{return_value}->{suggesting_child_policy_count} == 0) {
        return $self->base_delete(rpc_params => [$session, F5::ASMConfig::CallFlags->new(%$call_flags, skip_inheritance => 1), $policy, $suggestion_criteria],
                                  entity_class => $entity_class,
                                 );
    }


    if ($retval->{events}) {
        push (@{$self->{events}}, @{delete $retval->{events}});
    }
    accept_suggestions(dbh => $self->{dbh},
                       self => $self,
                       policy => $policy,
                       inherited_call => $call_flags->{inherited_call},
                       store_suggestion => $store_suggestion,
                       suggestion_log_note => $call_flags->{log_note},
                       suggestions => [$retval->{return_value}],
                      );
    if (defined($store_suggestion->{suggestion_status}) && $store_suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_IGNORED()) {
        clean_parent_suggestions(dbh => $self->{dbh},
                                 self => $self,
                                 policy => $policy,
                                 suggestions => [$retval->{return_value}],
                                );
        if ($self->{device_group}) {
            $call_flags->{relayable} = 1;
            $call_flags->{add_on_updating_nonexistent_record} = 1;
            push (@{$self->{rpc_calls_to_relay}->{update_suggestion}}, {function_name => 'update_suggestion',
                                                                                          rpc_params    => [$session, $call_flags, $policy, $retval->{return_value}, $retval->{return_value}],
                                                                       });
        }
    }

    return $retval;
}

#----------------------------------------------------------------------

sub update_many_suggestions {
    my($self, $session, $call_flags, $policy, $suggestion_criteria, $store_suggestion) = @_;

    if ($store_suggestion->{suggestion_status} && $store_suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT_GLOBALLY) {
        $call_flags->{search_filters} ||= [];
        push(@{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new(field_name => 'is_globally_acceptable',
                                                                                field_value => 1,
                                                                                field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ()));  
    }

    my $retval = $self->base_update_many(rpc_params => [$session, $call_flags, $policy, $suggestion_criteria, $store_suggestion],
                                         entity_class => $entity_class,);

    if ($retval->{events}) {
        if (defined($store_suggestion->{suggestion_status}) && $store_suggestion->{suggestion_status} == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_IGNORED()) {
            clean_parent_suggestions(dbh => $self->{dbh},
                                     self => $self,
                                     policy => $policy,
                                     suggestions => $retval->{entities_changed},
                                    );

            if ($self->{device_group}) {
                #These suggestions may not have been persistent until now so they may not exist on the remote device yet. 
                #Therefore each updated suggestion needs to be sent as an indivdual call with all of its fields
                
                $call_flags->{relayable} = 1;
                $call_flags->{add_on_updating_nonexistent_record} = 1;
                for my $event (@{$retval->{events}}) {
                    if ($event->{element} && $event->{element}->{suggestion}) {
                        push (@{$self->{rpc_calls_to_relay}->{update_many_suggestions}}, {function_name => 'update_suggestion',
                                                                                          rpc_params    => [$session, $call_flags, $policy, $event->{element}->{suggestion}, $event->{element}->{suggestion}],
                                                                                         }
                             );
                    }
                }
            }
        }
        push (@{$self->{events}}, @{delete $retval->{events}});
    }

    accept_suggestions(dbh => $self->{dbh},
                       self => $self,
                       policy => $policy,
                       inherited_call => $call_flags->{inherited_call},
                       store_suggestion => $store_suggestion,
                       suggestions => $retval->{entities_changed},
                       suggestion_log_note => $call_flags->{log_note},
                      );

    return $retval;
}

#----------------------------------------------------------------------

sub delete_suggestion {
    my ($self, $session, $call_flags, $policy, $suggestion_criteria) = @_;

    my $retval = $self->base_delete(rpc_params => [$session, $call_flags, $policy, $suggestion_criteria],
                                    entity_class => $entity_class,);

    clean_parent_suggestions(dbh => $self->{dbh},
                             self => $self,
                             policy => $policy,
                             suggestions => [$retval->{return_value}],
                            );

    return $retval;
}

#----------------------------------------------------------------------

sub delete_many_suggestions {
    my ($self, $session, $call_flags, $policy, $suggestion_criteria) = @_;

    my $retval = $self->base_delete_many(rpc_params => [$session, $call_flags, $policy, $suggestion_criteria],
                                         entity_class => $entity_class,);


    clean_parent_suggestions(dbh => $self->{dbh},
                             self => $self,
                             policy => $policy,
                             suggestions => $retval->{entities_changed},
                            );
    return $retval;
}

#----------------------------------------------------------------------

sub clean_parent_suggestions {
    my (%params) = @_;
    my $self = $params{self};
    my $suggestions = $params{suggestions};
    my $dbh = $params{dbh};
    my $policy = $params{policy};

    if ($policy->{has_parent}) {
        my @suggestion_keys;
        for my $suggestion (@$suggestions) {
            if ($suggestion->{inheritance_acceptability} == F5::ASMConfig::InheritanceAcceptability::INHERITANCE_ACCEPTABILITY_LOCKED() && $suggestion->{suggestion_key}) {
                push (@suggestion_keys, $suggestion->{suggestion_key});
            }
        }
        $self->update_many_suggestions($self->{session},
                                       F5::ASMConfig::CallFlags->new(remove_nested_struct_on_update => 1,
                                                 search_filters => [ F5::ASMConfig::SearchFilter->new(field_name => 'suggestion_key',
                                                                                                      field_value_list => \@suggestion_keys,
                                                                                                      field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),),
                                                                   ],),
                                       F5::ASMConfig::Policy->new(policy_id => $policy->{parent_policy_id}),
                                       F5::ASMConfig::Suggestion->new(),
                                       F5::ASMConfig::Suggestion->new(suggesting_child_policy_ids => [$policy->{policy_id}]));
    }
}

#----------------------------------------------------------------------

1;
