package F5::ASMConfig::Handler::SyncHandler;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::Deploy;
use F5::ConfigSync;
use F5::Policy;
use F5::BigipVersionUtils;
use F5::DeviceGroup;
use F5::Cgc::Channel::DeviceGroup;
use F5::Syslog;
use POSIX qw(strftime ceil);
use File::Basename;
use File::Temp qw/tempfile/;
use F5::LockFactory;
use F5::DateTimeUtils;
use F5::ObjUtils qw(freeze);
use F5::Cluster::Utils;

our @handled_functions = (
                           {
                             function_name => "sync_request_ucs",
                           },
                           {
                             function_name => "sync_request_active_policy_ucs",
                           },
                           {
                             function_name => "sync_request_data_protection",
                           },
                           {
                             function_name => "sync_send_ucs",
                           },
                           {
                             function_name => "sync_send_active_policy_ucs",
                           },
                           {
                             function_name => "sync_send_data_protection",
                           },                           
                           {
                             function_name => "sync_send_file",
                           },
                           {
                             function_name => "sync_receive_file_part",
                           }, 
                           {
                             function_name => "sync_ucs_file_complete",
                           },
                           {
                             function_name => "sync_data_protection_complete",
                           },
                           {
                             function_name => "sync_active_policy_file_complete",
                           },
                           {
                             function_name => "update_mcp_device_group",
                           },
                           {
                             function_name => "update_mcp_device_group_device",
                           },
                           {
                             function_name => "update_mcp_device",
                           },
                           {
                             function_name => "sync_sender_finished",
                           },
                           {
                             function_name => "send_sync",
                           },
                           {
                             function_name => "retrieve_file_chunk",
                           },
                           {
                             function_name => "store_file_chunk",
                           },
                           {
                             function_name => "get_temp_filename",
                           },
                           {
                             function_name => "unlink_temp_file",
                           },
                         );

my $SYNC_DIR = F5::Deploy::map_path("/ts/var/sync");
my $VAR_TMP = F5::Deploy::get_path_by_type("tmp");
my $SYNC_LOCK_FILE = F5::Deploy::map_path('/ts/lock/send_sync.lock');

unless (-e $SYNC_DIR) {
    mkdir($SYNC_DIR);
    chmod(0755, $SYNC_DIR);
}

my $default_rv = { return_value => 1,
                 events => [],
                 };

my $SYNC_SENDER_WAIT_TIME = 20 * 60; # 20 Minutes
my $SYNC_REREQUEST_TIME = $F5::ASMConfig::Handler::SYNC_REREQUEST_TIME;

#----------------------------------------------------------------------

sub sync_request_ucs {
    my ($self, $session, $call_flags, @other_params) = @_;
    $self->{shared_state}{out_of_sync} = 1;
    my $is_failover_standby = F5::Cluster::Utils::is_failover_standby();
    $call_flags->{sync_request_from_active_device} = !$is_failover_standby;
    $call_flags->{policy_versions_hash} = F5::ConfigSync::build_policy_versions_hash();
    return F5::ASMConfig::Handler::SyncHandler::sync_general_ucs_request($self,
                                                                         rpc_params => [$session, $call_flags, @other_params],
                                                                         remote_function_name => 'sync_send_ucs',
                                                                         remote_function_display_name => 'complete configuration',
                                                                        );
}

#----------------------------------------------------------------------

sub sync_request_active_policy_ucs {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::SyncHandler::sync_general_ucs_request($self,
                                                                     rpc_params => \@rpc_params,
                                                                     remote_function_name => 'sync_send_active_policy_ucs',
                                                                     remote_function_display_name => 'enforcement configuration',
                                                                    );
}

#----------------------------------------------------------------------

sub sync_request_data_protection {
    my ($self, @rpc_params) = @_;
    return F5::ASMConfig::Handler::SyncHandler::sync_general_ucs_request($self,
                                                                     rpc_params => \@rpc_params,
                                                                     remote_function_name => 'sync_send_data_protection',
                                                                     remote_function_display_name => 'Data Protection Configuration',
                                                                    );
}

#----------------------------------------------------------------------
sub sync_general_ucs_request {
    my($self, %params) = @_;

    my $remote_function_name = $params{remote_function_name};
    my $remote_function_display_name = $params{remote_function_display_name};
    my $rpc_params = $params{rpc_params};
    F5::ObjUtils::remove_unfreezable_references($rpc_params);
    my $call_flags = $rpc_params->[1];
    my $group_id = $call_flags->{originator} || $self->{device_group};

    if (F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade()) {
        $self->{logger}->info("Do not request sync from non-Primary blade") if $self->{logger};
        return $default_rv;
    }

    my $check_time = F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $SYNC_REREQUEST_TIME,
                                                                        start_time => $self->{shared_state}{last_sync_request_time},
                                                                       );
    if ($check_time) {
        $self->{logger}->info("Too soon to request sync again") if $self->{logger};
        $self->{shared_state}{last_sync_request_host} = $group_id;
        return $default_rv;
    }

    my $log_msg = "ASM is now entering sync recovery state. Requesting $remote_function_display_name from $group_id";
    $self->{logger}->info($log_msg) if $self->{logger};
    F5::Syslog::ts_syslog('info', $log_msg); 

    $call_flags->{recipient} = $self->{local_device_name};
    $self->send_to_relay(function_name => $remote_function_name,
                         rpc_params    => $rpc_params,
                         group_id      => $group_id,
                         priority      => "high",
                        );
    my $request_time = time();
    $self->{sync_info}{out_of_sync_start} ||= $request_time;
    $self->{shared_state}{emergency} = $request_time;
    $self->{shared_state}{sync_init} = 0;
    $self->{shared_state}{last_sync_request_time} = $request_time;
    $self->{shared_state}{last_sync_request_host} = $group_id;
    return $default_rv;
}

#----------------------------------------------------------------------

sub sync_send_ucs {
    my ($self, $session, $call_flags, $policy, $failed_params) = @_;

    my $export_filename;
    my $policy_versions_string = q{};
    if ($call_flags->{policy_versions_hash}) {
       $policy_versions_string = Digest::MD5->new->add(freeze($call_flags->{policy_versions_hash}))->hexdigest;
    }

    if ($call_flags->{originator_time}) {
        #Was a result of an device group sync event
        $export_filename = "$SYNC_DIR/sync_" . $call_flags->{originator_time} . "_${policy_versions_string}_full_update"; 
    }
    else {
        $export_filename = "$SYNC_DIR/sync_" . strftime("%Y%m%d_%H%M%S", localtime()) . "_${policy_versions_string}_full_update";
    }

    my $update_mcp = 0;
    $self->{shared_state}{sync_init} = 0;
    $self->{shared_state}{last_sync_request_time} = 0;

    unless (-e $export_filename) {
        my $lock_file = $F5::DbUtils::LOCK_FILE;
        my $db_locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                                    logger => $self->{logger},);
        my $known_times_lock = F5::LockFactory::get_locker(type => "IPC",
                                                           ipc  => $self->{known_times_ipc});
        $db_locker->lock();
        $known_times_lock->lock();
        if (!defined($self->{known_times}{$self->{local_device_name}})) {
            my $time = Time::HiRes::time();
            $self->{known_times}{$self->{local_device_name}} = $time;
            $update_mcp = 1;
            eval {
                $self->update_device_group_member(undef, 
                                                  F5::ASMConfig::CallFlags->new({add_on_updating_nonexistent_record => 1}), 
                                                  undef, 
                                                  F5::ASMConfig::DeviceGroupMember->new({ host => $self->{local_device_name},
                                                                                          known_time => $time}),
                                                  );
            };
        }
        # Set defined known time for recepient peer, as now it will have a known time of last config change 
        # (Prevent erroneous full sync request in auto-sync FO on non-ASMConfig change
        $self->{known_times}{$call_flags->{recipient}} ||= 0;
        eval {
        # Wrap up ucs file
            my $configsync = F5::ConfigSync->new(dbh          => $self->{dbh},
                                                logger       => $self->{logger},
                                                base_dir     => $SYNC_DIR,
                                                asm_config_handler => $self,
                                                call_flags => $call_flags,
                                                cmi_metadata => 1,);
            $configsync->save_to_file(filename => $export_filename);
        };
        my $e = $@;
        $known_times_lock->unlock();
        $db_locker->unlock();
        if ($e) {
           $self->{logger}->warn("Could not save configuration to send full sync to $call_flags->{originator}: $e") if $self->{logger};
           return $default_rv;
        }
    }
  
    F5::ASMConfig::Handler::SyncHandler::call_as_rpc_client($self, function_name => "send_sync",
                                                                   rpc_params => [$call_flags, $policy, $export_filename],);

    if ($update_mcp && $self->{shared_state}->{is_auto_sync}) {
        $self->update_mcp_timestamp();
    }

    return $default_rv;
}

#----------------------------------------------------------------------

sub sync_send_active_policy_ucs {
    my ($self, $session, $call_flags, $policy, $set_active) = @_;

    # Wrap up ucs file
    my $export_filename;
    my $account_id = $policy->{account_id};
    if ($account_id && ! $policy->{_failed_to_load}) {
        $export_filename = "$SYNC_DIR/sync_" . strftime("%Y%m%d_%H%M%S", localtime()) . "_active_policy";
        my $configsync = F5::ConfigSync->new(dbh => $self->{dbh},
                                             logger => $self->{logger},
                                             base_dir => $SYNC_DIR,
                                             active_policy_sync_only => 1,
                                             policy_history => 0,
                                             account_id     => $account_id,
                                             asm_config_handler => $self,);
        $configsync->save_to_file(filename => $export_filename);
    }
    F5::ASMConfig::Handler::SyncHandler::call_as_rpc_client($self, function_name => "send_sync",
                                                                   rpc_params => [$call_flags, $policy, $export_filename],);

    return $default_rv;
}

#----------------------------------------------------------------------

sub sync_send_data_protection {
    my ($self, $session, $call_flags, $policy) = @_;

    my $export_filename = "$SYNC_DIR/sync_" . strftime("%Y%m%d_%H%M%S", localtime()) . "_data_protection";
    $self->export_data_protection($session, $call_flags, $policy, F5::ASMConfig::DataProtectionImportExport->new(filename => $export_filename));
    F5::ASMConfig::Handler::SyncHandler::call_as_rpc_client($self, function_name => "send_sync",
                                                                   rpc_params => [$call_flags, $policy, $export_filename],);

    return $default_rv;

 
}

#----------------------------------------------------------------------

sub call_as_rpc_client {
    my ($self, %params) = @_;
    my $client = $self;
    if ($self->{is_relay_handler_process}) {
        $client = F5::ASMConfig::EasyClient->new(username => q{configsync});
    }
    my $session = $client->{session};
    my $function_name = $params{function_name};
    my $call_flags_copy = F5::ASMConfig::CallFlags->new(%{$params{rpc_params}->[0]},
                                                        is_asynchronous => 1);
    $params{rpc_params}->[0] = $call_flags_copy;
    eval {
        $client->$function_name($session, @{$params{rpc_params}});
    };
    if (my $err = $@) {
        if (ref $err) {
            $err = $err->{error_message};
        }
        $self->{logger}->error("Error handing function '$function_name' over to thrift server: $err") if $self->{logger};
    }
    return;
}

#----------------------------------------------------------------------

sub send_sync {
    my ($self, $session, $call_flags, $policy, $export_filename) = @_;
   
    my %connection_params = F5::ASMConfig::Handler::SyncHandler::get_connection_params($self, call_flags => $call_flags);

    $self->{locker} = F5::LockFactory::get_locker(lock_filename => $SYNC_LOCK_FILE . "." . basename($connection_params{group_id}),
                                                  logger => $self->{logger});
    my $got_lock = $self->{locker}->lock(dont_block => 1); #will unlock when process exits
    if (!$got_lock) {
       $self->{logger}->info("CMI sync file already being sent to '$connection_params{group_id}', aborting.") if $self->{logger};
       return $default_rv;
    }

    my $log_msg = "Sending CMI sync file '$export_filename' to '$connection_params{group_id}'";
    $self->{logger}->info($log_msg) if $self->{logger};
    F5::Syslog::ts_syslog('info', $log_msg);

    $self->sync_send_file($session, $call_flags, $policy, $export_filename);

    my $call_flags_copy = F5::ASMConfig::CallFlags->new(%{$call_flags},
                                                        ucs_version => F5::BigipVersionUtils::bigip_version(),
                                                        sync_sender_pid  => $$,
                                                       );
    my $fo_dg = F5::DeviceGroup::get_failover_device_group();
    if ($fo_dg->{name}) {
        $call_flags_copy->{failover_device_group} = $fo_dg->{name};
    }

    my $function_hash = { data_protection => "sync_data_protection_complete",
                          full_update     => "sync_ucs_file_complete",
                          active_policy   => "sync_active_policy_file_complete",};
    my $completed_function_name;
    if ($export_filename =~ /data_protection$/) {
        $completed_function_name = "sync_data_protection_complete";
    }
    elsif ($export_filename =~ /full_update$/) {
        $completed_function_name = "sync_ucs_file_complete";
    }
    else {
        $completed_function_name = "sync_active_policy_file_complete";
    }
    $self->send_to_relay(
                         function_name => $completed_function_name,
                         rpc_params => [$session, $call_flags_copy, $policy, $export_filename],
                         %connection_params,
                        );
    $self->{sleep_on_destroy} = $SYNC_SENDER_WAIT_TIME;
   
    my $num_devices_in_group = scalar @{$self->{shared_state}{devices_in_group}};
    if ($num_devices_in_group > 0 && $num_devices_in_group < 3 && -e $export_filename) {
        unlink($export_filename);
    }

    return $default_rv;
}

#----------------------------------------------------------------------


sub sync_send_file {
    my($self, $session, $call_flags, $policy, $filename) = @_;
    $self->{logger}->debug("In sync_send_file, and will be sending $filename") if $self->{logger};
    F5::ObjUtils::remove_unfreezable_references([$session, $call_flags, $policy]);
    my $CHUNK_SIZE = 2**19; # Half a MB

    my %connection_params = F5::ASMConfig::Handler::SyncHandler::get_connection_params($self, call_flags => $call_flags),
    my $file_size = -s $filename;
    my $total_chunks = ceil($file_size / $CHUNK_SIZE);
    open (my $fh, "<", $filename);
    binmode $fh;
    my $chunk;
    my $chunk_number = 1;
    while (read $fh, $chunk, $CHUNK_SIZE)  {
        $self->send_to_relay(function_name => "sync_receive_file_part",
                             rpc_params    => [$session, $call_flags, $policy,
                                               {filename      => $filename,
                                                data          => $chunk,
                                                chunk_number  => $chunk_number++,
                                                total_chunks  => $total_chunks,
                                                file_size     => $file_size,
                                               }],
                             %connection_params,
                             );
    }
    close ($fh);
    return;
}

#----------------------------------------------------------------------

sub sync_receive_file_part {
  my ($self, $session, $call_flags, $policy, $filepart) = @_;
  my $filename      = "$SYNC_DIR/" . basename($filepart->{filename});
  my $data          = $filepart->{data};
  my $chunk_number  = $filepart->{chunk_number};
  my $total_chunks  = $filepart->{total_chunks};

  my $mode = ($chunk_number == 1) ? ">" : ">>";
  open (my $fh, $mode, $filename);
  binmode $fh;
  print $fh $data;
  close ($fh);

  if ($chunk_number == $total_chunks) {
      F5::ASMConfig::Handler::SyncHandler::verify_received_file($self,
                                                                filename => $filename,
                                                                file_size => $filepart->{file_size},
                                                               );
  }
  if ($self->{shared_state}{emergency}) {
      $self->{shared_state}{emergency} = time(); #refresh emergency wait timer
  }
  return $default_rv;
}

#----------------------------------------------------------------------

sub get_temp_filename {
    my ($self, $session, $call_flags) = @_;
    my ($fh, $tmp_filename) = tempfile(DIR => $VAR_TMP);
    close $fh;

    return { return_value => $tmp_filename };
}
    
#----------------------------------------------------------------------

sub unlink_temp_file {
    my ($self, $session, $call_flags, $place_holder, $temp_filename) = @_;
    
    if ($temp_filename !~ /^$VAR_TMP/) {
        return $self->handle_error(action => 'delete',
                                   error_code   => F5::ASMConfig::ErrorCode::ILLEGAL_FILE_PERMISSIONS(),
                                   thrift_obj   => $temp_filename,
                                   error_level  => 'error',
                                  );
    }

    unlink ($temp_filename);
    return { return_value => 1 };
}

#----------------------------------------------------------------------

sub retrieve_file_chunk {
    my ($self, $session, $call_flags, $place_holder, $filename, $chunk_size, $chunk_index) = @_;

    if ($filename !~ /^$VAR_TMP/) {
        return $self->handle_error(action => 'retrieve',
                                   error_code   => F5::ASMConfig::ErrorCode::ILLEGAL_FILE_PERMISSIONS(),
                                   thrift_obj   => $filename,
                                   error_level  => 'error',
                                  );
    }

    
    my $rv;
    open my $in, "<", $filename;
    seek ($in, $chunk_size*($chunk_index-1), 0);
    read $in, $rv, $chunk_size;
    close $in;
    return { return_value => $rv };
}
    
#----------------------------------------------------------------------

sub store_file_chunk {
    my ($self, $session, $call_flags, $place_holder, $chunk, $temp_filename) = @_;
    
    if ($temp_filename !~ /^$VAR_TMP/) {
        return $self->handle_error(action => 'store_file',
                                   error_code   => F5::ASMConfig::ErrorCode::ILLEGAL_FILE_PERMISSIONS(),
                                   thrift_obj   => $temp_filename,
                                   error_level  => 'error',
                                  );
    }

    open (my $out, ">>", $temp_filename);
    binmode $out;
    print $out $chunk;
    close $out;

    return { return_value => 1 };
}

#----------------------------------------------------------------------

sub verify_received_file {
    my ($self, %params) = @_;
    my $filename      = $params{filename};
    my $expected_size = $params{file_size};
    my $actual_size   = -s $filename;

    if ($actual_size != $expected_size) {
        my $msg = "Expected size of $filename ($expected_size) does not match actual size ($actual_size)";
        $self->{logger}->info($msg) if $self->{logger};
        unlink ($filename);
        die "$msg\n";
    }

    return 1;
}

#---------------------------------------------------------------

sub sync_ucs_file_complete {
    my ($self, $session, $call_flags, $policy, $filename) = @_;

    my $group_id = $call_flags->{originator} || $self->{device_group};
    $call_flags->{filename} = $filename;
    $self->send_to_relay(function_name => "sync_sender_finished",
                         rpc_params    => [$session, $call_flags],
                         group_id      => $group_id,
                         priority      => "high",
                        );

    my $rv;
    if (! -e $filename) {
        $self->{logger}->info("Not processing load from incomplete file '$filename'.") if $self->{logger};
        return { return_value => $rv };
    }

    my $log_msg = "Recieved sync file '$filename' from $call_flags->{originator}";   
    $self->{logger}->info($log_msg) if $self->{logger};
    F5::Syslog::ts_syslog('info', $log_msg);

    my $configsync = F5::ConfigSync->new(dbh          => $self->{dbh},
                                         logger       => $self->{logger},
                                         base_dir     => "$SYNC_DIR/",
                                         ucs_version  => $call_flags->{ucs_version},
                                         asm_config_handler => $self,
                                         call_flags   => $call_flags,
                                         cmi_metadata => 1);
    #Modify session for policy logging
    $session->{user_name} = "SYSTEM";
    $session->{gui_user_name} = '';

    my $lock_file = $F5::DbUtils::LOCK_FILE;
    my $db_locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                                logger => $self->{logger},);
    my $known_times_lock = F5::LockFactory::get_locker(type => "IPC",
                                                        ipc  => $self->{known_times_ipc});
    $db_locker->lock();
    $known_times_lock->lock();

    my $last_config_load_time = time();

    eval {
        $self->{dbh}->commit;
        $rv = $configsync->load_from_file(filename        => $filename,
                                          check_asm_ready => 0,);
        $self->{dbh}->begin_work;
        if ($rv) {
            for my $host (keys %{$self->{known_times}}) {
                my $device_group_member = F5::ASMConfig::DeviceGroupMember->new({host => $host,
                                                                                 known_time => $self->{known_times}{$host}});
                eval {
                    $self->update_device_group_member(undef, F5::ASMConfig::CallFlags->new({add_on_updating_nonexistent_record => 1}), undef, $device_group_member);
                };
                $self->{logger}->info("Failed to update device group member: $@") if $@ && $self->{logger};
            }
        }

        $self->{sync_info}{out_of_sync_start} = 0;
        $self->{shared_state}{emergency} = 0;
        $self->{shared_state}{out_of_sync} = 0;
        $self->{shared_state}{last_sync_request_time} = 0;
        $self->{shared_state}{last_config_load_time} = $last_config_load_time;

        $log_msg = "Sync recovery state resolved successfully";
        $self->{logger}->info($log_msg) if $self->{logger};
        F5::Syslog::ts_syslog('info', $log_msg);
    };
    if ($@) {
        $self->{logger}->warn("Could not load configuration while in sync recovery state: $@") if $self->{logger};
    }
    $known_times_lock->unlock();
    $db_locker->unlock();

    if (-e $filename) {
        unlink($filename);
    }

    return { return_value => $rv, };
}

#---------------------------------------------------------------

sub sync_active_policy_file_complete {
    my ($self, $session, $call_flags, $policy, $filename) = @_;
    my $rv;
    my $group_id = $call_flags->{originator} || $self->{device_group};
    $call_flags->{filename} = $filename;
    $self->send_to_relay(function_name => "sync_sender_finished",
                         rpc_params    => [$session, $call_flags],
                         group_id      => $group_id,
                         priority      => "high",
                        );


    if (!$filename) {
        $self->{logger}->info("Active policy sync was cancelled by initiating device.") if $self->{logger};
        $rv = $default_rv;
    }
    else {
        my $log_msg = "Recieved sync file '$filename' from $call_flags->{originator}";
        $self->{logger}->info($log_msg) if $self->{logger};
        F5::Syslog::ts_syslog('info', $log_msg);

       #TODO: Check for existence of file 
        my $configsync = F5::ConfigSync->new(dbh                     => $self->{dbh},
                                             logger                  => $self->{logger},
                                             base_dir                => $SYNC_DIR,
                                             active_policy_sync_only => 1,
                                             policy_history          => 0,
                                             account_id              => $policy->{account_id},
                                             asm_config_handler => $self,);
        $self->{dbh}->commit;
        eval {
            $rv = $configsync->load_from_file(filename => $filename,
                                              check_asm_ready => 0,);
        };
        if ($@) {
            $self->{logger}->warn("Could not load Active policy configuration: $@") if $self->{logger};
        }
        if ($self->{dbh}->{AutoCommit}) {
            $self->{dbh}->begin_work;
        }
    }
    if (!$rv) {
       $self->{logger}->info("Active policy sync was unsuccessful, requesting full sync") if $self->{logger};
       $self->sync_request_ucs($session, $call_flags);        
    }
    else {
        $self->{sync_info}{out_of_sync_start} = 0;
        $self->{shared_state}{emergency} = 0;
        $self->{shared_state}{last_sync_request_time} = 0;
        my $log_msg = "Sync recovery state resolved successfully";
        $self->{logger}->info($log_msg) if $self->{logger};
        F5::Syslog::ts_syslog('info', $log_msg);
    }

    if (-e $filename) {
        unlink($filename);
    }

    return { return_value => $rv};
}

#---------------------------------------------------------------

sub sync_data_protection_complete {
    my ($self, $session, $call_flags, $policy, $filename) = @_;
    my $rv;
    my $group_id = $call_flags->{originator} || $self->{device_group};
    $call_flags->{filename} = $filename;
    $self->send_to_relay(function_name => "sync_sender_finished",
                         rpc_params    => [$session, $call_flags],
                         group_id      => $group_id,
                         priority      => "high",
                        );


    if (-e $filename) {
        my $log_msg = "Received sync file '$filename' from $call_flags->{originator}";
        $self->{logger}->info($log_msg) if $self->{logger};
        F5::Syslog::ts_syslog('info', $log_msg);
        $rv = $self->import_data_protection($session, $call_flags, $policy, F5::ASMConfig::DataProtectionImportExport->new({filename => $filename,
                                                                                                                           grace_signing_interval => 0,
                                                                                                                           grace_accepting_interval => 0}));
    }
    if (!$rv) {
       $self->{logger}->info("Data Protection sync was unsuccessful, requesting full sync") if $self->{logger};
       $self->sync_request_ucs($session, $call_flags);
    }
    else {
        $self->{sync_info}{out_of_sync_start} = 0;
        $self->{shared_state}{emergency} = 0;
        $self->{shared_state}{last_sync_request_time} = 0;
        my $log_msg = "Sync recovery state resolved successfully";
        $self->{logger}->info($log_msg) if $self->{logger};
        F5::Syslog::ts_syslog('info', $log_msg);
    }

    if (-e $filename) {
        unlink($filename);
    }

    return { return_value => $rv};

}

#---------------------------------------------------------------

sub sync_sender_finished {
    my ($self, $session, $call_flags) = @_;

    my $filename = $call_flags->{filename};
    if ($filename && -e $filename) {
        unlink($filename);
    }

    if ($call_flags->{sync_sender_pid}) {
        $self->{logger}->debug("Cleaning up sync sender $call_flags->{sync_sender_pid}") if $self->{logger};
        kill ("TERM", $call_flags->{sync_sender_pid});
    }
    return $default_rv;
}

#---------------------------------------------------------------

sub get_connection_params {
    my($self, %params) = @_;
    my $call_flags = $params{call_flags};

    my %retval;
    $retval{priority} = "low";
    $retval{group_id} = $self->{device_group};
    if ($call_flags->{recipient}) {
        $retval{priority} = "high";
        $retval{group_id} = $call_flags->{recipient};
    }

    return %retval;
}

#----------------------------------------------------------------------
# Updates for joining/leaving a device group or of a sync occured.
sub update_mcp_device_group_device {
    my ($self, $session, $call_flags, $place_holder, $device_group_device) = @_;

    # Did my name change?   
    if ($device_group_device->{originator} &&
        $device_group_device->{originator} eq $self->{local_device_name} &&
        $device_group_device->{device_group} eq "/Common/device_trust_group") {
        my $local_device_name = F5::DeviceGroup::get_local_device_name();
        if ($local_device_name ne $self->{local_device_name}) {
            $self->{logger}->info("Local device name has changed from $self->{local_device_name} to $local_device_name") if $self->{logger};
            $self->{shared_state}->{local_device_name} = $self->{local_device_name} = $local_device_name;
            return { return_value => 0 };
        }
    }
    
    my $known_times_lock = F5::LockFactory::get_locker(type => "IPC",
                                                       ipc  => $self->{known_times_ipc});
    my $shared_state_lock = F5::LockFactory::get_locker(type => "IPC",
                                                        ipc  => $self->{shared_state_ipc});
    $known_times_lock->lock();
    $shared_state_lock->lock();

    $self->{device_group} = $self->{shared_state}{device_group};
    # Is this about me?
    if (($self->{device_group} && $device_group_device->{device_group} ne $self->{device_group})
        || (!$self->{device_group} && $device_group_device->{device} ne $self->{local_device_name})) {
         $self->{logger}->debug("Update does not concern us.") if $self->{logger};
         return { return_value => 0 };
    }
    # Is this a sync event?
    if ($self->{device_group} && $device_group_device->{sync_type} && $device_group_device->{originator}) {
        if ($device_group_device->{device_group} ne $self->{device_group}
            || (!$self->{shared_state}{sync_init} && $device_group_device->{originator} eq $self->{local_device_name})
            || $device_group_device->{device} ne $self->{local_device_name}) {
             # Ignore incremental auto sync, or if this is not the ASM device group we are a part of, or we are the originator, or not the target device.
              my $display_string = join(", ", map { "$_: '$device_group_device->{$_}'"} keys (%$device_group_device)); 
              $self->{logger}->debug("Ignoring device_group_device sync event.  $display_string.") if $self->{logger};
              return { return_value => 0};
        }

        #If the sync type is Manual group, request a full sync for ASM
        if ($device_group_device->{sync_type} == F5::ASMConfig::SyncType::SYNC_TYPE_FULL_AUTO()
            || $device_group_device->{sync_type} == F5::ASMConfig::SyncType::SYNC_TYPE_INCREMENTAL_AUTO()) {
            #If incremental, check if we have a known_time for the originating device  
            #Otherwise ignore it and let the relay handler decide if we need to request a full sync by looking at the MCP device-sync events.
            my $has_device_sync = 0;
            if (my $folder_name = F5::DeviceGroup::get_asm_sync_folder($self->{device_group})) {
                my $device_sync = F5::ASM::COAPI::Query->new("asm_device_sync", {device_name =>  "$folder_name/" . basename($device_group_device->{originator})});
                $has_device_sync = ($device_sync && @$device_sync);
            }

            if ( !$self->{shared_state}{sync_init} &&
                 ($has_device_sync || defined($self->{known_times}{$device_group_device->{originator}}))
               ) {
                $self->{logger}->debug("Ignoring device_group_device sync event. Will detect via device-sync object") if $self->{logger};
                return { return_value => 0};    
            }
        }

        my $last_mcp_sync_time = $self->{shared_state}{last_mcp_sync_time};
        if (!$last_mcp_sync_time ||
            $last_mcp_sync_time lt $device_group_device->{originator_time}) {
            $self->{logger}->info("$device_group_device->{device} has received an config synchronization event from $device_group_device->{originator} for device group $device_group_device->{device_group}.") if $self->{logger};
            $self->{shared_state}{last_mcp_sync_time} = $device_group_device->{originator_time};
            if (!$last_mcp_sync_time) {
                #If this is an initial sync on joining a group, wait .5s for the relay handlers to start listening.
                Time::HiRes::usleep(500_000);
            }
            #Recast device_group_device as call_flags;
            $self->sync_request_ucs($session, F5::ASMConfig::CallFlags->new(%$device_group_device, %$call_flags))
        }
        return { return_value => 0};
    }

    if ($self->{device_group}) {
        my @old_devices_in_group = $self->{shared_state}{devices_in_group} ? @{$self->{shared_state}{devices_in_group}} : ();
        my @devices_in_group = F5::DeviceGroup::get_devices_in_device_group($self->{device_group});
        if ($#devices_in_group == $#old_devices_in_group) {
           my %old_devices = map {($_ => 1)} @old_devices_in_group;
           my @matches = grep {$old_devices{$_}} @devices_in_group;
           if ($#matches == $#devices_in_group) {
               $self->{logger}->debug("No change in device group membership. Returning") if $self->{logger};
               return { return_value => 0 };
           }
       }
       $self->{shared_state}{devices_in_group} = \@devices_in_group;
    }

    F5::ASMConfig::Handler::SyncHandler::handle_device_group_change(self => $self,
                                                                    modified_device_group => $device_group_device->{device_group},
                                                                    update_topology => 1,);

    $known_times_lock->unlock();
    $shared_state_lock->unlock();

    my @events;
    push(@events,
         $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                             session => $session,
                             element => F5::ASMConfig::Entity::DeviceGroup->new(dbh => $self->{dbh}, object => {name => $self->{device_group}}),
                             call_flags => $call_flags,
                             user_name => $session->{user_name},
                             suppress_save_to_log => 1,
                            ));
    return { return_value => $self->{device_group} ? 1 : 0,
             events => \@events };
}

#----------------------------------------------------------------------

sub handle_device_group_change {
    my (%params) = @_;
    my $self = $params{self};
    my $modified_device_group = $params{modified_device_group};
  
    my $known_times_lock = F5::LockFactory::get_locker(type => "IPC",
                                                       ipc  => $self->{known_times_ipc});
    my $shared_state_lock = F5::LockFactory::get_locker(type => "IPC",
                                                        ipc  => $self->{shared_state_ipc});
    $known_times_lock->lock();
    $shared_state_lock->lock();
    my $current_device_group = $self->{shared_state}{device_group};
    my $mcp_device_group = F5::DeviceGroup::get_device_group($self->{local_device_name});
    my $new_device_group = $mcp_device_group->{name};

    if (!$current_device_group && !$new_device_group) {
        $shared_state_lock->unlock();
        $known_times_lock->unlock();
        return 0;
    }

    my $device_group_changed = 0;
    eval {
        if ($current_device_group ne $new_device_group) {
            if ($modified_device_group && ($current_device_group eq $modified_device_group)) {
                # If our device group has changed, but the device_group of the event that triggered this call is the current device group and not the new device group
                # then process this call only as leaving the current group. Another event will come to trigger joining the new group.
                $new_device_group = "";
            }
            if ($current_device_group) {
                $self->{logger}->info("Device group has been updated.  '$self->{local_device_name}' IS NOT a member of an ASM device group: '$current_device_group'" ) if $self->{logger};
            }
            if ($new_device_group) {
                $self->{logger}->info("Device group has been updated.  '$self->{local_device_name}' IS a member of an ASM device group: '$new_device_group'" ) if $self->{logger};
            }
            $self->{device_group} = $new_device_group;
            $self->{shared_state}{device_group} = $self->{device_group};
            $self->{shared_state}{is_auto_sync} = $mcp_device_group->{autosync_enabled};
            $self->{shared_state}{last_mcp_sync_time} = 0;
            $self->{shared_state}{out_of_sync} = 0;
            $self->{shared_state}{emergency} = 0;
            $self->{shared_state}{last_sync_request_time} = 0;
            $self->{sync_info}{out_of_sync_start} = 0;

            if ($current_device_group) {
                $self->clean_device_group_state(device_group => $current_device_group); 
            }             

            if ($new_device_group) {
                my @devices_in_group = F5::DeviceGroup::get_devices_in_device_group($self->{device_group});
                $self->{shared_state}{devices_in_group} = \@devices_in_group;
                $self->{shared_state}{sync_init} = time();
            } 

            $params{update_topology} = 1;
            $device_group_changed = 1;
         }
         else {
             my @devices_in_group = @{$self->{shared_state}{devices_in_group}};
             my $best_peer = $self->find_best_peer(devices => \@devices_in_group, include_self => 1);
             my $is_group_leader = ($best_peer) ? ($best_peer->{device} eq $self->{local_device_name}) 
                                                : ($self->{local_device_name} eq $devices_in_group[0]) ;

             for my $known_host (keys %{$self->{known_times}}) {
                if (!grep {$_ eq $known_host} @devices_in_group) {
                    $self->{logger}->debug("$known_host is no longer in our device group") if $self->{logger};
                    delete $self->{known_times}->{$known_host};
                    F5::ASMConfig::Entity::DeviceGroupMember->delete_many(dbh => $self->{dbh}, search_criteria => F5::ASMConfig::DeviceGroupMember->new({host => $known_host}));
                    if ($is_group_leader) {
                        eval {
                            my $folder_name =  F5::DeviceGroup::get_asm_sync_folder_name($self->{device_group});
                            my $device_sync = F5::ASM::COAPI::Query->new("asm_device_sync", {device_name =>  "$folder_name/" . basename($known_host)});
                            if ($device_sync && @$device_sync) {
                                $device_sync->[0]->delete();
                                $self->{logger}->debug("Removed asm_device_sync for $known_host") if $self->{logger};
                            }
                        };
                    }
                }
            }
        }
        if ($current_device_group && $device_group_changed) {
            $self->{shared_state}{resubscribe} = 1;
            F5::Cgc::Channel::DeviceGroup::update_topology_self(logger => $self->{logger});
        }
        elsif ($params{update_topology}) {
            F5::Cgc::Channel::DeviceGroup::update_topology(logger => $self->{logger});
        }
    };
    if ($@) {
        $self->{logger}->info("Error receiving update from device_group: $@") if $self->{logger};
    }
    $shared_state_lock->unlock();
    $known_times_lock->unlock();

    return $device_group_changed;
}
#----------------------------------------------------------------------
sub update_mcp_device_group {
    my ($self, $session, $call_flags, $place_holder, $device_group) = @_;
    if (( $self->{device_group} && $device_group->{name} ne $self->{device_group}) || 
        (!$self->{device_group} && !$device_group->{asm_sync})) {
         $self->{logger}->debug("Update does not concern us.") if $self->{logger};
         return  { return_value => 0 };
    }

    my @events;
    if (defined ($device_group->{asm_sync}) 
        && (($device_group->{name} eq $self->{device_group} && !$device_group->{asm_sync})
            || (!$self->{device_group} && $device_group->{asm_sync}))) {
        if (F5::ASMConfig::Handler::SyncHandler::handle_device_group_change(self => $self,
                                                                            modified_device_group => $device_group->{name},)) {
            push(@events,
                 $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                                     session => $session,
                                     element => F5::ASMConfig::Entity::DeviceGroup->new(dbh => $self->{dbh}, object => {name => $self->{device_group}} ),
                                     call_flags => $call_flags,
                                     user_name => $session->{user_name},
                                     suppress_save_to_log => 1,
                                    )
                );
        }
    }
    elsif (defined ($device_group->{autosync_enabled})) {
        $self->{logger}->debug("Updating autosync state of group to $device_group->{autosync_enabled}") if $self->{logger};
        $self->{shared_state}{is_auto_sync} = $device_group->{autosync_enabled};
    }

    return { return_value => $self->{device_group} ? 1 : 0,
             events  => \@events };
}

#----------------------------------------------------------------------

sub update_mcp_device {
    my ($self, @rpc_params) = @_;
    my $local_device_name = $self->{local_device_name};
    my $configsync_ip = $self->{shared_state}->{configsync_ip};
    if (has_self_changed($self)) {
        my $shared_state_lock = F5::LockFactory::get_locker(type => "IPC",
                                                        ipc  => $self->{shared_state_ipc});
        $shared_state_lock->lock();
        $self->{shared_state}{resubscribe} = 1;
        F5::Cgc::Channel::DeviceGroup::update_topology_self(local_device_name => $local_device_name,
                                                            configsync_ip     => $configsync_ip,
                                                            logger            => $self->{logger},
                                                           );
        $shared_state_lock->unlock();
      
    }
    elsif (have_device_group_members_changed($self)) {
        F5::Cgc::Channel::DeviceGroup::update_topology(logger => $self->{logger});
    }
    return  { return_value => 1 };
}

#----------------------------------------------------------------------
sub has_local_name_changed {
    my ($self) = @_;
    my $current_local_device_name = $self->{shared_state}->{local_device_name};
    my $new_value = F5::DeviceGroup::get_local_device_name();
    if ($current_local_device_name ne $new_value) {
         $self->{logger}->info("Local device name has changed from $current_local_device_name to $new_value") if $self->{logger};
         $self->{shared_state}->{local_device_name} = $new_value;
         $self->{local_device_name} = $new_value;
         return 1;
    }
    return 0;
}

#----------------------------------------------------------------------
sub has_configsync_ip_changed {
    my ($self) = @_;
    my $current_configsync_ip = $self->{shared_state}->{configsync_ip};
    my $device = F5::DeviceGroup::get_device($self->{shared_state}->{local_device_name});
    my $new_value = $device->{configsync_ip};
    if ($current_configsync_ip ne $new_value) {
         $self->{logger}->info("Local device config-sync ip has changed from $current_configsync_ip to $new_value") if $self->{logger};
         $self->{shared_state}->{configsync_ip} = $new_value;
         return 1;
    }
    return 0;
}
 
#----------------------------------------------------------------------
sub has_self_changed {
    my ($self) = @_;
    my $name_changed = has_local_name_changed($self);
    my $ip_changed = has_configsync_ip_changed($self);
    return $name_changed || $ip_changed;
}
#----------------------------------------------------------------------
sub have_device_group_members_changed {
     my ($self) = @_;
     my $known_device_group_members_string = F5::Cgc::Channel::DeviceGroup::get_device_group_members_config_string() || $self->{shared_state}->{device_group_members_string} || '';
     my $current_device_group_members_string = '';
     my @devices_in_group = @{$self->{shared_state}{devices_in_group} || []};
     for my $device_name (sort @devices_in_group) {
         next if $device_name eq $self->{local_device_name};
         my $device = F5::DeviceGroup::get_device($device_name);
         $current_device_group_members_string .= F5::DeviceGroup::get_device_compare_string($device);
     }
     if ($current_device_group_members_string ne $known_device_group_members_string) {
         $self->{logger}->info("Device Group members have changed name or ip. $known_device_group_members_string to $current_device_group_members_string");
         $self->{shared_state}->{device_group_members_string} = $current_device_group_members_string;
         return 1;
     }
     return 0;
}
#----------------------------------------------------------------------
1;
