package F5::ASMConfig::Handler::SystemTools;
use strict;
use warnings;

use F5::Deploy;
use F5::CRC;
use F5::RE2;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::SystemUtils;
use F5::ASMConfig::ASMConfig;
use F5::Cfg;
use Time::Local;

our @handled_functions = (
                           {
                             function_name => 'xpath_validate',
                           },
                           {
                             function_name => 'is_valid_regexp',
                           },
                           {
                             function_name => 'get_crc',
                           },
                           {
                             function_name => 'test_regexp_match',
                           },
                           {
                             function_name => 'geoip_lookup',
                           },
                           {
                             function_name => 'verify_user_password',
                           },
                           {
                             function_name => 'max_account_exceeded',
                           },
                           {
                             function_name => 'get_timezone_diff',
                           },
                         );

#------------------------------------------------------
sub xpath_validate{
  my($self, $session, $call_flags,$place_holder,$xpath) = @_;
  my @params;
  push ( @params, F5::Deploy::map_path('/usr/share/ts/bin/xpath_validate'), $xpath);

  my $res =  F5::SystemUtils::run_system_tool( command_line_params => \@params);

  return { return_value => $res->{completed_successfully} } ;

}
#------------------------------------------------------
sub is_valid_regexp{
  my($self, $session, $call_flags,$place_holder,$regexp, $regexp_type) = @_;

  my $return_value;
  my $is_valid;
  my $error_message;
  if ($regexp_type == F5::ASMConfig::RegexpType::REGEXP_TYPE_RE2) {
      $return_value = F5::RE2::is_re2_valid($regexp) || 1;
      $is_valid = ( $return_value eq '1' ) ? 1 : 0;
      $error_message = ( $return_value eq '1' ) ? "" :  $return_value;
  }
  else {
      $regexp =~ s/\//\\\//g;
      eval {
          no warnings 'all';
          no strict;
          my $a =~ /$regexp/;
      };

      if($@){
        my $eval_error = $@;
        $eval_error =~ s/\s+ at \s+ \S+ \s+ line \s+ \d+ .*$//sx;
        $error_message = "Invalid PCRE: $regexp ($eval_error)";
        $is_valid = 0;
      }
      else{
        $is_valid = 1;
      }
  }

   my $retval = F5::ASMConfig::RegexpValidationRes->new(is_valid => $is_valid ,error_message => $error_message);


   return { return_value => $retval } ; 
}
#------------------------------------------------------
sub test_regexp_match{
  my($self, $session, $call_flags,$place_holder,$regexp, $regexp_type,$value) = @_;

  my $is_valid_regexp = $self->is_valid_regexp($session, $call_flags,$place_holder,$regexp, $regexp_type);

  if (! $is_valid_regexp ){
     die "The regular expression is not valid";
  }

  my $match ='';
  
  if  ($regexp_type == F5::ASMConfig::RegexpType::REGEXP_TYPE_RE2) {
     $match = F5::RE2::re2_matches($value,$regexp);
  }
  else{
     if ( $value =~ /$regexp/ ){
         $match = $&;
     }
  }

  return { return_value => $match};
}
#------------------------------------------------------
sub get_crc{
  my($self, $session, $call_flags,$place_holder, $str) = @_;

  my $crc =  F5::CRC::get_crc($str);

  return { return_value => $crc };
}
#------------------------------------------------------
sub geoip_lookup{
  my($self, $session, $call_flags,$place_holder,$ip) = @_;
  my @params;
  push ( @params,'geoip_lookup',$ip);

  my $res =  F5::SystemUtils::run_system_tool( command_line_params => \@params);

  my $geo_ip_res = F5::ASMConfig::GeoIpRes->new();
  
  if ($res->{standard_output} =~ /region_namesize of geoip database = (\d+),/g){
      $geo_ip_res->{size_of_geoip_database} = $1;
  }

  if ($res->{standard_output} =~ /geoip_seek = (.*)/g){
      $geo_ip_res->{geoip_seek} = $1;
  }

  if ($res->{standard_output} =~ /geoip record ip = (.*)/g){
      $geo_ip_res->{geoip_record_ip} = $1;
  }

  if ($res->{standard_output} =~ /country_code = (.*)/g){
      $geo_ip_res->{country_code} = $1;
  }

  if ($res->{standard_output} =~ /country_name = (.*)/g){
      $geo_ip_res->{country_name} = $1;
  }

  if ($res->{standard_output} =~ /region_name = (.*)/g){
      $geo_ip_res->{region_name} = $1;
  }

  return { return_value => $geo_ip_res } ;
}
#------------------------------------------------------
sub verify_user_password{
  my($self, $session, $call_flags, $place_holder, $username, $password) = @_;

  my @cmd = ("/sbin/unix_chkpwd", $username, "nonull");
  my $res = F5::SystemUtils::run_system_tool(command_line_params => \@cmd,
                                             STDIN => $password . "\000");

  return { return_value => $res->{completed_successfully} } ;
}
#------------------------------------------------------
sub max_account_exceeded {
  my($self, $session, $call_flags) = @_;

  my $max_account_exceeded  = 0;
  my $account_id = F5::DbUtils::find_lowest_unused_id(dbh => $self->{dbh},
                                                      table_name => 'DCC.ACCOUNTS',
                                                      id_name => 'account_id',
                                                      min_id_value => 1,
                                                    );

  # NOTE: 'MaxAccounts' value in 'dcc.cfg' is updated by F5::ProcessHandler::start_bd on startup.
  my $max_accounts = F5::Cfg::cfg_get_config_item('/etc/ts/dcc/dcc.cfg', 'GENERAL', 'MaxAccounts');
 
  if ($account_id > $max_accounts) {
      $max_account_exceeded  = $max_accounts;
  }

  return { return_value => $max_account_exceeded };
}
#------------------------------------------------------
sub get_timezone_diff {
  my($self, $session, $call_flags) = @_;

  my @t = localtime(time);
  my $timezone_diff = timegm(@t) - timelocal(@t);

  return { return_value => $timezone_diff };
}
#------------------------------------------------------
1;
