package F5::ASMConfig::Handler::TaskBase;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::Deploy;
use F5::ASMConfig::ASMConfig;
use Time::HiRes;

our @handled_functions = ();

our $REST_DIR = F5::Deploy::map_path("/ts/var/rest");

#----------------------------------------------------------------------

sub add_task {
    my (%params) = @_;
    my $self = $params{self};
    my ($session, $call_flags, $policy, @other_rpc_params) = @{$params{rpc_params}};
    my $entity_class = $params{entity_class};
    my $task_type = $params{task_type};

    my $rv = $self->base_add(rpc_params => [ $session, $call_flags, $policy, @other_rpc_params],
                                           entity_class => $entity_class,);

    my $task = $entity_class->new(object => {task_id => $rv->{return_value}->{task_id}});

    $call_flags->{async_rest_call} = 1;
    my $is_asynchronous = $call_flags->{bulk_call} ? 0 : 1;

    push (@{$self->{rpc_queue}}, {function_name => "update_${task_type}_task",
                                  is_asynchronous => $is_asynchronous,
                                  rpc_params => [$session, $call_flags, $policy, $entity_class->new(object => {%$task, status => F5::ASMConfig::TaskStatus::TASK_STATUS_STARTED()})]
                               });
       
    push (@{$self->{rpc_queue}}, {function_name => "execute_${task_type}_task",
                                  is_asynchronous => $is_asynchronous,
                                  rpc_params => [$session, $call_flags, $policy, $task]
                                });

    return $rv;
}

#----------------------------------------------------------------------

my %finished_task_status = ( F5::ASMConfig::TaskStatus::TASK_STATUS_COMPLETED() => 1,
                             F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE()   => 1, 
                           );

sub update_task {
    my (%params) = @_;
    my $self = $params{self};
    my ($session, $call_flags, $policy, $task_criteria, $store_task) = @{$params{rpc_params}};
    my $entity_class = $params{entity_class};
    my $task_type = $params{task_type};

    my $task = $entity_class->new(dbh => $self->{dbh},
                                  get_criteria => $task_criteria,
                                  policy => $policy,
                                  );
    if (!$task || $finished_task_status{$task->{status}}) {
        my $error = $task ? F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED() 
                          : F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD();
        return $self->handle_error(action => 'update',
                               error_thrown => 'Task already finished or no longer exists',
                               error_code => $error,
                               call_flags => $call_flags,
                               entity_class => $entity_class,
                               thrift_obj => $task_criteria,
                              );
    }

    my $task_status = $entity_class->get_new_value(store_obj => $store_task,
                                                   existing_obj => $task,
                                                   field_name => 'status');
    if ( $task_status == F5::ASMConfig::TaskStatus::TASK_STATUS_STARTED() ) {
        $store_task->{pid} = $$;
	my $execution_start_time = $entity_class->get_new_value(store_obj => $store_task,
                                                              existing_obj => $task,
                                                              field_name => 'execution_start_time');
	$store_task->{execution_start_time} = time unless ($execution_start_time);
    }

    return $self->base_update(rpc_params => [ $session, $call_flags, $policy, $task_criteria, $store_task],
                              entity_class => $entity_class,)
}

#----------------------------------------------------------------------

sub execute_task {
    my (%params) = @_;
    my $self = $params{self};
    my ($session, $call_flags, $policy, $thrift_task) = @{$params{rpc_params}};
    my $entity_class = $params{entity_class};
    my $task_type = $params{task_type};
    my $execute_func = $params{execute_func};

    my $task = $entity_class->new(dbh => $self->{dbh},
                               get_criteria => $thrift_task,
                               policy => $policy,
                              );
    if (!$task || !$task->{status} == F5::ASMConfig::TaskStatus::TASK_STATUS_STARTED()) {
        my $error = $task ? F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED()
                          : F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD();
        return $self->handle_error(action => 'execute',
                               error_thrown => 'Task already finished or no longer exists',
                               error_code => $error,
                               call_flags => $call_flags,
                               entity_class => $entity_class,
                               thrift_obj => $thrift_task,
                              );
    }

    eval {
        $execute_func->(self => $self, 
                        session => $session, 
                        call_flags => $call_flags,
                        task => $task,
                        policy => $policy,
                       
                       );
    };

    my ($time, $micro) = Time::HiRes::gettimeofday();
    $task->{end_time} = sprintf('%d.%06d',$time, $micro);


    if (my $err = $@) {
        $task->{status} = F5::ASMConfig::TaskStatus::TASK_STATUS_FAILURE();
         if (ref $err) {
            $task->{message} = $err->{error_message};
        }
        else {
            $task->{message} = $err;
        }
    }

    my $update_function = "update_${task_type}_task";

    return $self->$update_function($session, $call_flags, $policy, $task);

}

#----------------------------------------------------------------------

sub delete_task {
    my (%params) = @_;
    my $self = $params{self};
    my ($session, $call_flags, $policy, $thrift_task) = @{$params{rpc_params}};
    my $entity_class = $params{entity_class};

    my $deleted_task = $entity_class->new( dbh => $self->{dbh},
                                           policy => $policy,
                                           get_criteria => $thrift_task );


    if ( $deleted_task && $deleted_task->{status} == F5::ASMConfig::TaskStatus::TASK_STATUS_STARTED() ) {
        return $self->handle_error( error_code =>  F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                    error_thrown => "Could not delete running task.\n",
                                    error_level  => "info",
                                    thrift_obj => $thrift_task,
                                    call_flags => $call_flags,
                                  );
    }

    return $self->base_delete( rpc_params => [$session, $call_flags, $policy, $thrift_task],
                               entity_class => $entity_class,
                             );
}

#----------------------------------------------------------------------


sub delete_many_tasks {
    my (%params) = @_;
    my $self = $params{self};
    my ($session, $call_flags, $policy, $task_criteria) = @{$params{rpc_params}};
    my $entity_class = $params{entity_class};

    my $local_call_flags = F5::ASMConfig::CallFlags->new($call_flags);

    if (!defined $local_call_flags->{search_filters}){
        $local_call_flags->{search_filters} = [];
    }

    push(@{$local_call_flags->{search_filters}},
                 { field_name => 'status',
                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                   field_value => F5::ASMConfig::TaskStatus::TASK_STATUS_STARTED() });

    my $rv = $self->base_delete_many( entity_class => $entity_class,
                                      rpc_params => [ $session, $local_call_flags,$policy, $task_criteria ],);

    return $rv;
}

#----------------------------------------------------------------------

1;
