package F5::ASMConfig::Handler::ThreatCampaignFile;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::DbUtils;

use F5::ThreatCampaigns;
use F5::ThreatCampaignFile::Update;
use F5::ThreatCampaignFile::AutoDownload;

our @handled_functions = (
                           {
                             function_name => 'update_threat_campaign_file',
                             filename_arguments => ['3.filename'],
                             write_access => 1,
                             exclusive_lock_file => $F5::DbUtils::LOCK_FILE,
                           },
                           {
                             function_name => 'check_for_threat_campaign_file_update',
                           },
                           {
                             function_name => 'generate_threat_campaign_file',
                           },
                           {
                             function_name => 'get_threat_campaign_file_version',
                           },
                         );

#--------------------------------------------------------------------
sub update_threat_campaign_file{
  my($self, $session, $call_flags, $place_holder, $threat_campaign_file, $threat_campaign_file_update_settings) = @_;

  my %threat_campaign_file_update_settings_defaults = (
                                          check_interval => 0,
                                          allow_older_timestamp => 0,
                                         );

  if (! $threat_campaign_file_update_settings) {
      $threat_campaign_file_update_settings = {};
  }

  for my $default_key (keys %threat_campaign_file_update_settings_defaults) {
      if (! defined $threat_campaign_file_update_settings->{$default_key}) {
          $threat_campaign_file_update_settings->{$default_key} = $threat_campaign_file_update_settings_defaults{$default_key};
      }
  }

  my %params = (
                dbh => $self->{dbh},
                automatic_mode => defined $threat_campaign_file->{filename} ? 0 : 1,
                filename => $threat_campaign_file->{filename} || "",
                username => $call_flags->{user_name} || "",
                in_wrapping_transaction => $self->{in_transaction},
                logger => $self->{logger},
                outbound_call_timeout => $call_flags->{outbound_call_timeout},
                originator => $call_flags->{originator},
                asm_config_handler => $self,
                flg_is_genesis_file => $threat_campaign_file->{flg_is_genesis_file},
                threat_campaign_file_format => uc($threat_campaign_file->{file_format} || ''),
                skip_set_active => $call_flags->{skip_set_active},
                skip_set_sigs_active => $call_flags->{skip_set_sigs_active},
                %$threat_campaign_file_update_settings,
               );


  my $rv;
  eval {
         my $threat_campaign_update = F5::ThreatCampaignFile::Update->new(%params);
	 if($params{new_live_update}) {
	     $rv = $threat_campaign_update->apply_changes(%params);
	     if(!$params{dry_run}) {
	         $threat_campaign_update->post_process(%params);
	     }
	 }
	 else {
             $rv = $threat_campaign_update->update(%params);
         }
  };

  if ($@) {
      my $orig_error = $@;
      my $error_thrown;
      if (ref($orig_error) && $orig_error->{error_code}) {
          $error_thrown = $orig_error;
      }
      elsif (ref($orig_error) && $orig_error->{error_msg}) {
          $error_thrown = $orig_error->{error_msg};
      }
      else {
          $error_thrown = "Threat Campaign File update failed. Error: $@";
      }

      #Error suppression added for EM+iControl's benefit
      if (ref($error_thrown) && (exists $error_thrown->{error_code}) &&
         ($error_thrown->{error_code} == F5::ASMConfig::ErrorCode::INVALID_SIGFILE_VERSION() ||
          $error_thrown->{error_code} == F5::ASMConfig::ErrorCode::OLDER_SIGFILE_TIMESTAMP())) {
          $rv = {completed_successfully => 0,
                 error_msg => $error_thrown->{display_message},
                 error_code => $error_thrown->{error_code}};
      }
      elsif (ref($error_thrown) && (exists $error_thrown->{error_msg})) {
         my $error_msg_secured = $error_thrown->{error_msg};
         # Error message may look like this: "User signature file does not conform to schema (/ts/var/negsig/admin~OhDF4F2qSU1rjaLPfh-8bA:4: parser error : Prema..."
         # We don't want the file location to be shown to user.
         $error_msg_secured =~ s/\([^\s]+\s/\(/;
         $rv = {completed_successfully => 0,
                error_msg => $error_msg_secured};
      }
      else {
          return $self->handle_error(action => 'update_threat_campaign_file',
                                     error_thrown => $error_thrown,
                                     error_level  => 'error',
                                    );
      }
  }

   my $return_value = F5::ASMConfig::Entity::ThreatCampaignFileUpdateRes->new(dbh => $self->{dbh}, object => $rv);
   #When doing a manual update with automatic delivery mode, update_mode = "auto-user"
   $return_value->{update_mode} = ($threat_campaign_file_update_settings->{update_mode} && $threat_campaign_file_update_settings->{update_mode} eq 'manual') ? F5::ASMConfig::ThreatCampaignFileUpdateMode::THREAT_CAMPAIGN_FILE_UPDATE_MODE_MANUAL 
                                                                                                                                   : F5::ASMConfig::ThreatCampaignFileUpdateMode::THREAT_CAMPAIGN_FILE_UPDATE_MODE_SCHEDULED;
  
   my @events;
   my $retval = { return_value => $return_value ,
                  events => \@events,
                 };

  if ($return_value->{completed_successfully} && !$return_value->{error_msg} && !$params{dry_run}) {
      if (! $threat_campaign_file_update_settings->{check_interval}) {
          $call_flags->{relayable} = 1;
      }
      my $description = "Threat Campaigns updated successfully - $rv->{threat_campaigns_count} threat campaigns loaded ($rv->{threat_campaigns_updated_count} updated).";
      push(@{$retval->{events}},
           $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                               session    => $session,
                               element    => $return_value,
                               call_flags => $call_flags,
                               user_name  => $session->{user_name},
                               description => $description,
                              ));
  }
  elsif ($rv->{unscheduled} && !$params{dry_run}) {
       push(@{$retval->{events}},
           $self->create_event(event_type => F5::ASMConfig::EventType::UPDATE,
                               session    => $session,
                               element    => $return_value,
                               call_flags => $call_flags,
                               user_name  => $session->{user_name},
                               suppress_save_to_log => 1,
                              ));
  }
  return $retval;
}

#--------------------------------------------------------------------

sub check_for_threat_campaign_file_update{
   my($self, $session, $call_flags) = @_;

   my $dbh = $self->{dbh};
   my $auto_download = F5::ThreatCampaignFile::AutoDownload->new(dbh => $dbh,
                                                                 outbound_call_timeout => $call_flags->{outbound_call_timeout},
                                                                 logger => $self->{logger},
                                                                );

   my $is_newer_threat_campaign_file_available;

   eval {
     $is_newer_threat_campaign_file_available = $auto_download->is_newer_sigfile_available;
   };

   if ($@) {
       my $error_msg = "Could not reach the update server. Error: $@";
       return $self->handle_error(action => 'check_for_threat_campaign_file_update',
                                  error_code   => F5::ASMConfig::ErrorCode::UNREACHABLE_HOST(),
                                  error_thrown => $error_msg,
                                  error_level  => 'error',
                                  call_flags => $call_flags,
                                  ignored_rv => 0,
                                 );
   }

   my $retval = { return_value => $is_newer_threat_campaign_file_available};

   return $retval;
}

#--------------------------------------------------------------------

sub generate_threat_campaign_file {
   my($self, $session, $call_flags, $place_holder, $file_spec) = @_;

   my $completed_successfully = 0;
   my $dbh = $self->{dbh};
   $call_flags->{search_filters} ||= [];

   my $iterator = F5::ASMConfig::Entity::ThreatCampaign->retrieve_all(dbh => $dbh,
                                                                 get_criteria => {},
                                                                 );
   my $threat_campaign_file = F5::ThreatCampaignFile->new(dbh => $dbh,
                                             output_filename => $file_spec->{filename},
                                            );

   $completed_successfully = $threat_campaign_file->create_threat_campaign_file(threat_campaign_iterator => $iterator);

   my $retval = { return_value => $completed_successfully };

   return $retval;
}

#--------------------------------------------------------------------

sub get_threat_campaign_file_version {
   my($self, $session, $call_flags) = @_;
	
   my $dbh = $self->{dbh};
   my $file_date = F5::ThreatCampaigns::get_update_file_date(flg_any_version => 1,
                                                             dbh => $dbh);
   $file_date =~ tr| |_|;
   $file_date =~ s|[-/:]||g;

   my $retval = { return_value => $file_date};

   return $retval;
}

#--------------------------------------------------------------------

1;
