package F5::ASMConfig::Handler::Violation;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use F5::ASMConfig::ASMConfig;
use F5::SetActive;

our @handled_functions = (
                           {
                             function_name => 'add_violation',
                             relayable => 1,
                             store_index  => 3,
                           },
                           {
                             function_name => 'update_violation',
                             relayable => 1,
                             criteria_index => 3,
                             store_index => 4,
                           },
                           {
                             function_name => 'reset_defaults_violations_severities',
                             relayable => 1,
                           },
                           {
                             function_name => 'export_violations',
                             criteria_index => 3,
                           },
                           {
                             function_name => 'import_violations',
                             relayable => 1,
                             filename_arguments => ['5.filename'],
                             criteria_index => 3,
                             store_index => 4,
                           },

                         );

our $entity_class = 'F5::ASMConfig::Entity::Violation';

#----------------------------------------------------------------------

sub add_violation {
  my($self, $session, $call_flags, $policy,$violation) = @_;

  my $retval = $self->base_add(rpc_params => [$session, $call_flags, $policy,$violation],
                               entity_class => $entity_class,
                                       );

  set_violations_active($self);

  return $retval;
}

#----------------------------------------------------------------------

sub update_violation {
  my($self, $session, $call_flags, $policy,$violation_criteria, $store_violation) = @_;

  my $rv = $self->base_update(rpc_params => [$session, $call_flags, $policy,$violation_criteria, $store_violation],
                              entity_class => $entity_class,
                             );

  set_violations_active($self);

  return $rv;
}

#----------------------------------------------------------------------

sub reset_defaults_violations_severities{
  my($self,$session, $call_flags) = @_;
  my $dbh = $self->{dbh};

  $dbh->do("DELETE FROM PLC.VIOLATIONS_PROPERTIES WHERE viol_index < 99");

  my $query = <<"EndQuery";
INSERT INTO PLC.VIOLATIONS_PROPERTIES (viol_index, severity_id) 
SELECT viol_index,severity_id from PLC.VIOLATIONS_PROPERTIES_DEFAULTS
EndQuery

  $dbh->do($query);

  set_violations_active($self);

  return { return_value => 1 };
}

#----------------------------------------------------------------------

sub set_violations_active {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    my $lock_file =  $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                             logger => $logger);
    $locker->lock();

    my $prepareconf = F5::PrepareConf::Base->new(
                                                 dbh     =>  $dbh,
                                                 logger  =>  $logger,
                                                );

    $prepareconf->prepare_violations;
    $dbh->commit unless $dbh->{AutoCommit};

    F5::SetActive::set_active(dbh => $dbh,
                              logger => $logger,
                              do_update_core => 1,
                              send_sync => 1,
                              asm_config_handler => $self);

    $locker->unlock();
}

#----------------------------------------------------------------------

sub export_violations {
   my($self, $session, $call_flags, $place_holder, $violation_criteria, $export_flags) = @_;

   my $completed_successfully = 0;
   my $dbh = $self->{dbh};

   $violation_criteria->{flg_is_user_defined} = 1;
   $violation_criteria->{flg_is_deleted} = 0;
   my $xml_content = F5::ASMConfig::Entity::Violation->export_xml(dbh => $dbh, 
                                                                  get_criteria => $violation_criteria, 
                                                                  search_filters => $call_flags->{search_filters});
   my $outf;
   if(! open($outf, '>', $export_flags->{filename}) ) {
       return $self->handle_error(action => "export_violations",
                                  error_code => F5::ASMConfig::ErrorCode::FILE_NOT_FOUND(),
                                  error_thrown => "Could not open '$export_flags->{filename}' : $!",
                                  error_level  => 'error',
                                 );
    }


   print $outf $xml_content;
   close $outf;

   my $retval = F5::ASMConfig::ExportResult->new( filename => $export_flags->{filename});

  return { return_value => $retval };
}

#----------------------------------------------------------------------

sub import_violations {
   my($self, $session, $call_flags, $place_holder, $violation_criteria, $store_violation, $import_flags) = @_;

   my $completed_successfully = 0;
   my $dbh = $self->{dbh};

   my $import_result = F5::ASMConfig::Entity::Violation->import_xml(dbh => $dbh, 
                                                                    xml_filename => $import_flags->{filename},
                                                                    store_values => { flg_is_deleted => 0},
                                                                   );

   my $retval = F5::ASMConfig::ImportResult->new( filename => $import_flags->{filename});

   return { return_value => $retval };
}

#----------------------------------------------------------------------

1;
