package F5::ASMConfig::Handler::VirtualServerAssoc;

use strict;
use warnings;

use base qw/F5::ASMConfig::Handler::Base/;
use File::Basename;
use F5::ASM::COAPI;
use F5::BigipUtils;
use F5::Mcp::PolicyControls;
use F5::Mcp::PolicyRequires;
use F5::Mcp::PolicyTarget;
use F5::Mcp::PolicyAction;
use F5::Mcp::PolicyStatus;
use F5::Mcp::ProfileType;

# same as in F5::ASMConfig::Handler::Policy
use constant ASM_AUTO_L7_POLICY_NAME_PREFIX => 'asm_auto_l7_policy__';

our @handled_functions = (
                           {
                             function_name => 'add_virtual_server_assoc',
                             relayable => 1,
                             writes_to_mcp => 1,
                           },
                           {
                             function_name => 'delete_virtual_server_assoc',
                             relayable => 1,
                             writes_to_mcp => 1,
                           },
                         );

#----------------------------------------------------------------------

sub add_virtual_server_assoc {
    my ($self, $session, $call_flags, $policy, $virtual_server_assoc) = @_;
    #Create/Get L7 policy
    my $action = "add association for";
    my $virtual_server = F5::ASMConfig::Entity::VirtualServer->new(get_criteria => $virtual_server_assoc->{virtual_server}, dbh => $self->{dbh});
    if (!defined($virtual_server)) {
        return $self->handle_error(action => $action,
                                   entity_class => 'F5::ASMConfig::Entity::VirtualServer',
                                   thrift_obj  => $virtual_server_assoc->{virtual_server},
                                   error_code   => F5::ASMConfig::ErrorCode::MCP_NONEXISTENT_RECORD(),
                                  );
    }

    if (!$policy || $policy->{_failed_to_load}) {
        return $self->handle_error(action => $action,
                                   entity_class => 'F5::ASMConfig::Entity::Policy',
                                   thrift_obj  => $policy,
                                   error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                  );
    }

    if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        return $self->handle_error(action => $action,
                                   error_thrown => 'Cannot associate parent policy to virtual servers',
                                   entity_class => 'F5::ASMConfig::Entity::Policy',
                                   thrift_obj  => $policy,
                                   error_code   => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
                                  );
    }

    my $in_new_l7_policy_creation = ($virtual_server_assoc->{l7_policy_name} ? 0 : 1);
    $virtual_server_assoc->{assoc_type} ||= F5::ASMConfig::VirtualServerAssocType::VIRTUAL_SERVER_ASSOC_TYPE_L7_SIMPLE;
    $virtual_server_assoc->{l7_policy_name} ||= $virtual_server->{asm_enabled_l7_policy_name};

    my $l7_policy;
    my $default_rule_name = 'default';
    my $l7_policy_name_base = $virtual_server_assoc->{l7_policy_name} || ($virtual_server->{folder} . '/' . ASM_AUTO_L7_POLICY_NAME_PREFIX . $virtual_server->{virtual_server_name});
    # Truncate L7 policy name to ensure that composed name does not exceed maximum length
    # Truncate to 249 characters to allow space for 4 places of numeric suffix 
    # ("policy_...long_name_1234")
    $l7_policy_name_base = substr($l7_policy_name_base, 0, 249);

    my $original_l7_policy_status;
    my $l7_policy_name = $l7_policy_name_base;
    my $attempt = 1;
    while (!$l7_policy) {
        my $conflicts_with_existing_l7policy = 0;

        $l7_policy = F5::ASMConfig::Entity::L7Policy->new(dbh => $self->{dbh}, get_criteria => { l7_policy_name => $l7_policy_name });
        if (! $l7_policy && $self->{transactional_saves} && $self->{transactional_saves}->{new_l7_policies} &&
            $self->{transactional_saves}->{new_l7_policies}->{$l7_policy_name}) {

            $l7_policy = $self->{transactional_saves}->{new_l7_policies}->{$l7_policy_name};
        }
        if ($l7_policy) {
            if ($l7_policy->{type} == F5::ASMConfig::VirtualServerAssocType::VIRTUAL_SERVER_ASSOC_TYPE_L7_COMPLEX) {
                return $self->handle_error(action => "delete",
                                           error_code   => F5::ASMConfig::ErrorCode::MCP_ACTION_NOT_ALLOWED(),
                                           error_thrown => "Cannot modify Virtual Server Association for $virtual_server->{full_name} via manually configured L7 policy $l7_policy_name.",
                                           error_level  => "info",
                                          );
            }
            if ($l7_policy->{type} != $virtual_server_assoc->{assoc_type}) {
                if ( !$virtual_server_assoc->{l7_policy_name} ) {
                    $conflicts_with_existing_l7policy = 1;
                }
                else {
                    die "Cannot modify existing Policy $virtual_server_assoc->{l7_policy_name}.";
                }
            }
            elsif ($in_new_l7_policy_creation && $l7_policy->{virtual_servers} && @{$l7_policy->{virtual_servers}} &&
                   $l7_policy->{virtual_servers}->[0] && $l7_policy->{virtual_servers}->[0]->{full_name} &&
                   $l7_policy->{virtual_servers}->[0]->{full_name} ne $virtual_server->{full_name}) {

                $conflicts_with_existing_l7policy = 1;
            }
            elsif ($l7_policy->{status} != F5::Mcp::PolicyStatus::POLICY_STATUS_LEGACY) {
                $original_l7_policy_status = $l7_policy->{status};
                $l7_policy->store_update(store_obj => {as_legacy => 1,
                                                       status => F5::Mcp::PolicyStatus::POLICY_STATUS_LEGACY,
                                                      }
                                        );
            }
        }
        else {
            $l7_policy = F5::ASMConfig::Entity::L7Policy->new(dbh => $self->{dbh},
                                                              object => { l7_policy_name => $l7_policy_name,
                                                                          controls => [F5::Mcp::PolicyControls::L7_POLICY_CONTROLS_ASM],
                                                                          requires => [F5::Mcp::PolicyRequires::L7_POLICY_REQUIRES_HTTP],
                                                                          as_legacy => 1,
                                                                          status => F5::Mcp::PolicyStatus::POLICY_STATUS_LEGACY,
                                                                          strategy => '/Common/first-match' });
            $l7_policy->store_add();
            $self->{transactional_saves}->{new_l7_policies}->{$l7_policy_name} = $l7_policy;
        }
        if ($conflicts_with_existing_l7policy) {
            $attempt++;
            $l7_policy_name = $l7_policy_name_base . "_$attempt";
            undef $l7_policy;
        }
    }

    my $policy_action_query = F5::ASM::COAPI::Query->new("policy_action", {target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM,
                                                                           action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE,
                                                                           policy_name => $l7_policy_name});
    if ($policy_action_query && @$policy_action_query) {
        for my $policy_action (@$policy_action_query) {
            if (!$call_flags->{replaced_policy_name} || $policy_action->{policy} eq $call_flags->{replaced_policy_name}) {
                #For replace, make sure the policy action we are replacing is for the correct asm policy.
                $policy_action->{policy} = $policy->{policy_name};
                $policy_action->transactional_action(action => 'save', dbh => $self->{dbh});
            }
        }
    }
    else {
        #Add asm_policy to L7 policy
        my $policy_action = F5::ASM::COAPI::Object->new('policy_action');
        $policy_action->{rule_name} = $default_rule_name;
        $policy_action->{policy_name} = $l7_policy_name;
        $policy_action->{action_id} = 1;
        $policy_action->{target} = F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM;
        $policy_action->{action} = F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE;
        $policy_action->{policy} = $policy->{policy_name};
        $policy_action->transactional_action(action => 'save', dbh => $self->{dbh});

        # policy_rule must be added after policy_action due to rule validation
        my $rule = F5::ASM::COAPI::Object->new('policy_rule');
        $rule->{name} = $default_rule_name;
        $rule->{policy_name} = $l7_policy_name;
        $rule->{ordinal} = 1;
        $rule->transactional_action(action => 'save', dbh => $self->{dbh});
    }

    # after the last, l7 policy related, transactional action - return it's original status, if required
    if ($original_l7_policy_status) {
        my $internal_config_item = $self->get_internal_config_items(config_item_name => 'force_legacy_ltm_policy');
        if (!$internal_config_item->{current_value}) {
            $l7_policy->store_update(store_obj => {as_legacy => 0,
                                                   status => $original_l7_policy_status,
                                                  }
                                    );
        }
    }

    #Assign L7 Policy to VS (requires websecurity and http profiles)
    my $http_profile_query = F5::ASM::COAPI::Query->new('virtual_server_profile', {profile_type => F5::Mcp::ProfileType::PROFILE_HTTP,
                                                                                   vs_name => $virtual_server->{full_name}});
    if (!$http_profile_query || !@$http_profile_query) {
        my $http_profile = F5::ASM::COAPI::Object->new('virtual_server_profile');
        $http_profile->{vs_name} = $virtual_server->{full_name};
        $http_profile->{profile_name} = '/Common/http';
        $http_profile->transactional_action(action => 'save', dbh => $self->{dbh});
    }

    my $ws = F5::ASM::COAPI::Object->new('virtual_server_profile');
    $ws->{vs_name} = $virtual_server->{full_name};
    $ws->{profile_name} = '/Common/websecurity';
    $ws->transactional_action(action => 'save', dbh => $self->{dbh});

    my $vs_policy = F5::ASM::COAPI::Object->new('virtual_server_policy');
    $vs_policy->{policy_name} = $l7_policy_name;
    $vs_policy->{vs_name} = $virtual_server->{full_name};
    $vs_policy->transactional_action(action => 'save', dbh => $self->{dbh});

    $virtual_server_assoc->{virtual_server}->{flg_is_asm_enabled} = 1;
    $virtual_server_assoc->{l7_policy_name} = $l7_policy_name;
    return {return_value => $virtual_server_assoc};
}

#----------------------------------------------------------------------

sub delete_virtual_server_assoc {
    my ($self, $session, $call_flags, $policy, $virtual_server_assoc) = @_;
    my $action = "delete association for";
    my $virtual_server = F5::ASMConfig::Entity::VirtualServer->new(get_criteria => $virtual_server_assoc->{virtual_server});
    if (!defined($virtual_server)) {
        return $self->handle_error(action => $action,
                                   entity_class => 'F5::ASMConfig::Entity::VirtualServer',
                                   thrift_obj  => $virtual_server_assoc->{virtual_server},
                                   error_code   => F5::ASMConfig::ErrorCode::MCP_NONEXISTENT_RECORD(),
                                  );
    }
    if ($policy && $policy->{_failed_to_load} && !$call_flags->{ignore_nonexistent_record_on_delete}) {
        return $self->handle_error(action => $action,
                                   entity_class => 'F5::ASMConfig::Entity::Policy',
                                   thrift_obj  => $policy,
                                   error_code   => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD(),
                                  );
    }


    #TODO: handle removing http_class association?
    #Detach specified policy or l7 policy
    my $l7_policy;
    my $original_l7_policy_status;
    my $removed_l7_plc = 0;
    my $was_websec_prof_configured = 0;
    if (!$virtual_server_assoc->{l7_policy_name} && $policy) {
        if (!$policy->{extra_virtual_server_assocs}) {
            $policy->map_fields_to_entity(extra_fields => [qw(extra_virtual_server_assocs)]);
        }
        ($virtual_server_assoc) = grep { $_->{virtual_server}->{full_name} eq $virtual_server_assoc->{virtual_server}->{full_name} } @{$policy->{extra_virtual_server_assocs}};
    }
    if ($virtual_server_assoc->{l7_policy_name} &&
        ($l7_policy = F5::ASMConfig::Entity::L7Policy->new(dbh => $self->{dbh}, get_criteria => { l7_policy_name => $virtual_server_assoc->{l7_policy_name} }))) {
        if ($l7_policy->{type} == F5::ASMConfig::VirtualServerAssocType::VIRTUAL_SERVER_ASSOC_TYPE_L7_COMPLEX) {
            return $self->handle_error(action => "delete",
                                   error_code   => F5::ASMConfig::ErrorCode::MCP_ACTION_NOT_ALLOWED(),
                                   error_thrown => "Cannot remove Virtual Server Association via manually configured L7 policy $virtual_server_assoc->{l7_policy_name}.",
                                   error_level  => "info",
                                  );
        }
        if (my ($vs_policy) = @{F5::ASM::COAPI::Query->new('virtual_server_policy', {vs_name => $virtual_server->{full_name},
                                                                                     policy_name => $virtual_server_assoc->{l7_policy_name}})}) {
            $vs_policy->transactional_action(action => 'delete', dbh => $self->{dbh});
            $removed_l7_plc = 1;
        }

        #If it is not referenced by any other virtuals
        if (!(grep { $_->{full_name} ne $virtual_server->{full_name} } @{$l7_policy->{virtual_servers}})) {
            # if it was not manually created - delete the whole l7_policy
            if ($virtual_server_assoc->{l7_policy_name} =~ ASM_AUTO_L7_POLICY_NAME_PREFIX) {
                $l7_policy->delete();
            } # if it was manually created and we are trying todeactivate the policy - do not allow the action.
            else {
                # flip L7 policy to legacy (we revert it later)
                if ($l7_policy->{status} != F5::Mcp::PolicyStatus::POLICY_STATUS_LEGACY) {
                    $original_l7_policy_status = $l7_policy->{status};
                    $l7_policy->store_update(store_obj => {as_legacy => 1,
                                                           status => F5::Mcp::PolicyStatus::POLICY_STATUS_LEGACY,
                                                          }
                                            );
                }
   
                # set 'asm disable', in the relevant L7 policy/rule/action, instead of 'asm enable policy <asm_policy>', for the target <asm_policy>
                my $policy_action_query = F5::ASM::COAPI::Query->new("policy_action", {target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM,
                                                                                       action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE,
                                                                                       policy_name => $virtual_server_assoc->{l7_policy_name}});
                if ($policy_action_query && @$policy_action_query) {
                    for my $policy_action (@$policy_action_query) {
                        $self->{logger}->info("Detaching security policy $policy_action->{policy} from L7 policy $policy_action->{policy_name}") if $self->{logger};
                        $self->{logger}->debug("Setting 'asm disable' in action $policy_action->{action_id}, rule $policy_action->{rule_name}, L7 policy $policy_action->{policy_name}") if $self->{logger};
   
                        $policy_action->{target} = F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM;
                        $policy_action->{action} = F5::Mcp::PolicyAction::L7_POLICY_ACTION_DISABLE;
                        $policy_action->{policy} = '';
   
                        $policy_action->transactional_action(action => 'save', dbh => $self->{dbh});
                    }
                }
            }
        }
        # A VS can only have one l7_policy that controls ASM - 
        #     if we removed a l7_policy above - we can safely remove the websecurity profile as well
        if ($removed_l7_plc && (my ($ws) = @{F5::ASM::COAPI::Query->new('virtual_server_profile', {vs_name => $virtual_server->{full_name},
                                                                             profile_type => F5::Mcp::ProfileType::PROFILE_WEBSECURITY,})}
           )) {
                $ws->transactional_action(action => 'delete',
                                          dbh => $self->{dbh},
                                          ignore_error => 1,
                                          logger => $self->{logger},);
                $was_websec_prof_configured = 1;
        }
    }

    # after the last, l7 policy related, transactional action - return it's original status, if required
    if ($original_l7_policy_status) {
        my $internal_config_item = $self->get_internal_config_items(config_item_name => 'force_legacy_ltm_policy');
        if (!$internal_config_item->{current_value}) {
            $l7_policy->store_update(store_obj => {as_legacy => 0,
                                                   status => $original_l7_policy_status,
                                                  }
                                    );
        }
    }

    $virtual_server_assoc->{was_websec_prof_configured} = $was_websec_prof_configured;
    return {return_value => $virtual_server_assoc};
}

1;
