package F5::ASMConfig::Server;

use strict;
use warnings;

use Thrift::Socket;
use Thrift::BufferedTransport;
use Thrift::Server;
use F5::Deploy;
use F5::Utils::File;
use F5::ASMConfig::Handler;
use F5::ASMConfig::EventDispatcher;
use F5::Logger;
use F5::Cfg;
use F5::BigipUtils;
#use F5::DeviceGroup;

my $DEFAULT_CONFIG_FILE = '/etc/ts/tools/asm_config_server.cfg';
my $DEFAULT_RPC_SERVER_PORT = 9781;
my $DEFAULT_RPC_SERVER_HOST = 'local';

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);

  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{dbh} = $params{dbh};
  my $dbh = $self->{dbh};
  $self->{logger} = $params{logger};

  $self->{config_file} = $params{config_file} || $DEFAULT_CONFIG_FILE;
  if (my $rpc_log_level = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'rpc_log_level')) {
      $self->{logger}->log_level($rpc_log_level);
  }

  # $self->{device_settings} = F5::DeviceGroup::get_device_settings(dbh => $dbh);

  # if ($self->{device_settings}) {
  #     for my $k (qw(rpc_server_port rpc_to_event_port event_queue_port)) {
  #         $self->{$k} = $self->{device_settings}->{$k};
  #     }
  #     $self->{rpc_server_host} = $self->{device_settings}->{local_bind_addr};
  # }
  # else {
      $self->{rpc_server_port} = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'AsmConfigServerPort') ||
        $DEFAULT_RPC_SERVER_PORT;

      $self->{rpc_server_host} = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'AsmConfigHost') ||
        $DEFAULT_RPC_SERVER_HOST;
  # }
}

#----------------------------------------------------------------------

sub start {
  my($self) = @_;

  if (fork) {
    # Parent
    sleep 2;                    # Wait for event dispatcher to start
    $self->{logger}->debug('Starting RPC server') if $self->{logger};

    F5::Utils::File::cleanup_files(glob_pattern => F5::Deploy::map_path("/ts/lock/asm_config_server_async_for_*.lock"),
                                   max_retained_files => 0);

    my $handler = F5::ASMConfig::Handler->new(dbh => $self->{dbh},
                                              logger => $self->{logger},
                                              config_file => $self->{config_file},
                                              rpc_to_event_port => $self->{rpc_to_event_port},
                                              device_settings => $self->{device_settings},
                                              guarantee_event_sending => 1,
                                              send_events_to_dispatcher => 1,
                                             );
    if (! F5::BigipUtils::is_docker()) {
        eval {
            $handler->spawn_relay_handler();
        };
        if ($@) {
            $self->{logger}->warn("Relay handler shutdown: $@") if $self->{logger};
            die $@;
        }
    }

    my $processor = F5::ASMConfig::ASMConfigProcessor->new($handler);
    my $transport;
    if ($self->{rpc_server_host} eq 'local') {
       $transport = Thrift::ServerSocket->new( addr => '127.0.0.1' , port => $self->{rpc_server_port});
    }
    else {
       $transport = Thrift::ServerSocket->new(port => $self->{rpc_server_port});
   }

    my $server = Thrift::ForkingServer->new( $processor, $transport);
    $server->{serverEventHandler} = $handler;

    eval {
      $server->serve;
    };

    if ($@) {
      my $error_message = "RPC server aborted ";
      if (ref($@) && ref($@) eq 'Thrift::TException') {
        $error_message .= "($@->{message})";
      }
      else {
        $error_message .= "($@)";
      }
      $self->{logger}->error($error_message) if $self->{logger};
      die "$error_message\n";
    }

    # We should not get to this line
    die "RPC server ended unexpectedly\n";
  }
  else {
    $self->{logger}->debug('Starting event dispatcher') if $self->{logger};
    my $event_logger = F5::Logger->get_logger(logger_name => 'asmconfigevent');
    if (my $event_dispatcher_log_level = F5::Cfg::cfg_get_config_item($self->{config_file}, 'General', 'event_dispatcher_log_level')) {
        $event_logger->log_level($event_dispatcher_log_level);
    }

    my $event_dispatcher = F5::ASMConfig::EventDispatcher->new(dbh => $self->{dbh},
                                                               logger => $event_logger,
                                                               config_file => $self->{config_file},
                                                               event_queue_port => $self->{event_queue_port},
                                                               rpc_to_event_port => $self->{rpc_to_event_port},
                                                              );

    eval {
      $event_dispatcher->run_event_dispatch;
    };
    if ($@) {
      my $error_message = "Event dispatcher aborted ($@)";
      $self->{logger}->error($error_message) if $self->{logger};
      die "$error_message\n";
    }

    # We should not get to this line
    die "RPC server ended unexpectedly\n";
  }
}

#----------------------------------------------------------------------

1;
