package F5::ASMConfig::Validations;
use strict;
use warnings;

use F5::SystemUtils;
use F5::IpAddr;
use F5::IpUtils;
use Socket qw(:DEFAULT inet_pton inet_ntop);

my $IPV4_IRRELEVANT_BITS = 80;

#----------------------------------------------------------------------

# Overridden in Entity::Base
sub general_validate {
    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

sub validate_response_codes {
    my ($self, %params) = @_;

    my $field_value = $params{field_value};
    my $field_def = $params{field_def};
    my $field_name = $params{field_name};

    if( $field_value ) {
        my %allowed_codes = (
                             '1xx' => 1,
                             '2xx' => 1,
                             '3xx' => 1,
                             '4xx' => 1,
                             '5xx' => 1,
                            );
        if( !$allowed_codes{$field_value} && ($field_value < 100 || $field_value > 599) ) {
            return {
                completed_successfully => 0,
                error_message => "Legal values are 1xx, 2xx, 3xx, 4xx, 5xx or any integer from 100 to 599.",
            };
        }
    }
    return $self->general_validate(%params);
}

#----------------------------------------------------------------------

sub validate_xpath {
    my ($self, %params) = @_;
    my $field_value = $params{field_value};

    my $res =  F5::SystemUtils::run_system_tool( command_line_params => [ F5::Deploy::map_path('/usr/share/ts/bin/xpath_validate'), $field_value]);

    if ($res->{completed_successfully}) {
        return $self->general_validate(%params);
    }
    else {
        $res->{error_message} = "The value is not a valid xpath";
        return $res;
    }
}

#----------------------------------------------------------------------

sub validate_netmask {
    my ($self,%params) = @_;
    my $field_value = $params{field_value};

    my $bin_mask = unpack("B*", F5::IpAddr::ip_pton($field_value));
    if ($self->is_valid_ipv4(field_value => $field_value)) {
        $bin_mask = substr($bin_mask,$IPV4_IRRELEVANT_BITS);
    }

    if  ($bin_mask =~ /^1*0*$/) {
        return { completed_successfully => 1 };
    }
    else {
        return {
                   completed_successfully => 0,
                   error_message => 'The value must be a valid IP mask',
               }
    }
}

#----------------------------------------------------------------------

sub normalize_lc {
     my ($class_or_self, %params) = @_;
     my $value = $params{field_value};
     $value = lc($value);

     return {
              completed_successfully => 1,
              normalized_value => $value,
            };
}

#----------------------------------------------------------------------

sub normalize_ip {
  my ($class_or_self, %params) = @_;
  my $field_def = $params{field_def};
  my $ip_add = $params{field_value};
  my $remove_brackets = $params{remove_brackets};
  my $normalized_value =  $ip_add;
  my $port_number;

  if ($field_def->{allow_port_number}) {
      if ($normalized_value =~ s/:(\d+)$//) {
          $port_number = $1;
          if ($port_number < 1 || $port_number > 65535) {
              return {
                      completed_successfully => 0,
                      error_message => "Port number must be between 1 and 65535",
                     };
          }
      }
  }

  my $normalize_error_message = defined $params{normalize_error_message} ?
                                        $params{normalize_error_message} : 'The value must be a valid ip address';
  my $allow_ipv4 = defined $field_def->{allow_ipv4} ? $field_def->{allow_ipv4} : 1;
  my $allow_ipv6 = defined $field_def->{allow_ipv6} ? $field_def->{allow_ipv6} : 1;

  if (! defined $ip_add) {
      return {
              completed_successfully => 0,
              error_message => $normalize_error_message,
             };
  }

  if ( $remove_brackets ){
     $normalized_value = ( $ip_add =~ /^\[(.*)\]$/g) ? $1 :  $ip_add;
  }

  my $split_ip_route;
  eval {
      $split_ip_route = F5::IpUtils::split_ip_route( ip_address => $normalized_value );
  };

  if ($@){
    return {
              completed_successfully => 0,
              error_message => $@
            }
  }

  my $route_domain = $split_ip_route->{route_domain};
  my $norm_without_route = $split_ip_route->{ip_address};

  my $is_valid_ip = ( ($allow_ipv4 && defined inet_pton(AF_INET, $norm_without_route)) ||
                      ($allow_ipv6 && defined inet_pton(AF_INET6, $norm_without_route)) )  ? 1 : 0;


  my $error_message = $is_valid_ip ? '': $normalize_error_message ;

  if ( ! $is_valid_ip ) {
     return {
              completed_successfully => 0,
              error_message => $error_message
            }

  }

  # Normalize IPv6 address to exclude extra zeros
  $normalized_value = F5::IpAddr::ip_ntop(F5::IpAddr::ip_pton($norm_without_route));

  if ($route_domain){
     $normalized_value = $normalized_value."%".$route_domain;
  }

  if ($port_number) {
      # Re-append port number
      $normalized_value .= ":$port_number";
  }

  return {
            completed_successfully => 1,
            normalized_value => $normalized_value,
         }
}

#----------------------------------------------------------------------

sub is_valid_host_name {
    my ($self,%params) = @_;
    my $host = $params{field_value};
    my $field_def = $params{field_def};

    if ( $host eq '' && $field_def->{allow_empty_string} ) {
        return { completed_successfully => 1 };
    }

    if (($host =~ /^[a-z0-9]+([a-z0-9-]*[a-z0-9]+)?(\.([a-z0-9]+([a-z0-9-]*[a-z0-9]+)?)+)*$/i) &&
        (!($host =~ /.*\.(.*)/g && $1 !~ /[a-z]+/i))) { # Atleast one alpha character in TLD for valid domain.
        return { completed_successfully => 1 };
    }

    return {
            completed_successfully => 0,
            error_message => 'The value must be a valid host name',
           };
}

#----------------------------------------------------------------------

sub normalize_host{
  my ($self,%params) = @_;
  my $host = $params{field_value};
  my $field_def = $params{field_def};

  my $rv = $self->normalize_ip(field_value => $host,
                               remove_brackets => 1,
                               field_def => $field_def,);

  if ( $rv->{completed_successfully}){
     return {
             completed_successfully => 1,
             normalized_value => $rv->{normalized_value},
           }
  }

  $rv = $self->is_valid_host_name(field_value => $host,
                                  field_def => $field_def,);
  if ( $rv->{completed_successfully}) {
     return {
             completed_successfully => 1,
             normalized_value => $host,
           }
  }

  return {
            completed_successfully => 0,
            error_message => 'The value must be a valid host name or a valid IP address.',
         }

}

#----------------------------------------------------------------------

sub normalize_header_name{
  my ($self,%params) = @_;
  my $header_name = $params{field_value};
  $header_name =~ s/^\s+|\s+$//g; #remove leading and trailing spaces
  $header_name = lc($header_name);

  return {
           completed_successfully => 1,
           normalized_value => $header_name,
         }
}

#----------------------------------------------------------------------

sub normalize_url_name{
    my ($self,%params) = @_;
    my $url_name = $params{field_value};

    if ( defined($url_name) && length($url_name) && !($url_name =~ /^\/(.*)+/g) ){
        $url_name = '/'.$url_name;
    }

    return {
           completed_successfully => 1,
           normalized_value => $url_name,
         }

}

#----------------------------------------------------------------------

sub normalize_url_name_wildcard_supported{
    my ($self,%params) = @_;
    my $url_name = $params{field_value};

    if ( defined($url_name) && length($url_name) && !($url_name =~ /^\/(.*)+/g) && !($url_name =~ /^\*.*/g)){
        $url_name = '/'.$url_name;
    }

    return {
           completed_successfully => 1,
           normalized_value => $url_name,
         }

}

#----------------------------------------------------------------------

sub is_valid_ipv4{
  my ($self,%params) = @_;
  my $ip_add = $params{field_value};
  my $field_def = $params{field_def};

  if ( $ip_add eq '' && $field_def->{allow_empty_string} ) {
      return { completed_successfully => 1 };
  }

  my $split_ip_route;
  eval {
      $split_ip_route = F5::IpUtils::split_ip_route( ip_address => $ip_add );
  };

  if($@){
    return {
              completed_successfully => 0,
              error_message => $@
            }
  }

  my $is_valid_ip =   defined inet_pton(AF_INET, $split_ip_route->{ip_address}) ? 1 : 0;
  my $error_message = $is_valid_ip ? '': 'The value must be a valid ipv4 address';

  return {
            completed_successfully => $is_valid_ip,
            error_message => $error_message
         }
}

#----------------------------------------------------------------------

1;
