package F5::ASMReady;

use strict;
use warnings;

use F5::Deploy;
use F5::BigIpLog;
use F5::DateTimeUtils;
use F5::bss qw(READY_SEMA bss_release bss_is_released_no_block);
use File::Slurp;
use F5::BigipUtils;

use Fcntl qw/:flock/;           # import LOCK_* constants
use YAML::XS;

use constant {
    BSS_NAME => "asm",
};

# file indicating that ASM initialization complete
my $ASM_READY_FILE = F5::Deploy::map_path('/ts/recovery/asm_ready');

our $BD_MANIFEST_FILE = F5::Deploy::map_path('/var/ts/files/site_1/config/bd_cfg_manifest.dat');
our $BD_MANIFEST_LOAD_STATE_FILE = F5::Deploy::map_path('/var/ts/files/site_1/config/bd_cfg_manifest_load_state.json');

my $ASM_LOAD_COMPLETE_END_LOCK_FILE = F5::Deploy::map_path('/ts/recovery/asm_load_complete_end.lock');

my $ASM_UPGRADE_LOCK_FILE = F5::Deploy::map_path('/ts/recovery/asm_upgrade.lock');

use constant BD_MANIFEST_GRACE => 1800;
use constant BD_GRACE => 60;

our $lock_fh;

# --------------------------------------------
sub set_asm_ready {
    our $bss_release_once;
    if ( ! $bss_release_once ) {
        bss_release(BSS_NAME, READY_SEMA, 0) && die "error: bss_release(): $!";
        $bss_release_once++;
    }
    # create the file indicating that ASM started successfully
    open($lock_fh, ">", $ASM_READY_FILE );
    print $lock_fh $$;
    chmod(0644, $ASM_READY_FILE);
    flock($lock_fh, LOCK_EX);
}

# --------------------------------------------
sub is_asm_upgrade_lock {
    return YAML::XS::LoadFile($ASM_UPGRADE_LOCK_FILE) if (-f $ASM_UPGRADE_LOCK_FILE);
    return 0;
}

# --------------------------------------------
sub is_asm_ready {
    return 1 if F5::BigipUtils::is_docker();
    return bss_is_released_no_block(BSS_NAME, READY_SEMA);
}

# --------------------------------------------
sub wait_for_asm_to_be_ready{
   my %params = @_;

   my $waited = 0;
   my $iteration_sleep_time = $params{iterations} || 5;
   my $timeout = $params{timeout} || 180;

   while($waited < $timeout) {      
      if( is_asm_ready() ) {
        return 1;
      }
      $waited+= sleep($iteration_sleep_time);  
   }

   return 0;
}

# --------------------------------------------
sub remove_asm_ready_file {
    if ($lock_fh) {
        close($lock_fh);
    }
    unlink($ASM_READY_FILE) if -f $ASM_READY_FILE;
}

# --------------------------------------------
sub remove_bd_manifest_load_state_file {
    unlink($BD_MANIFEST_LOAD_STATE_FILE) if -f $BD_MANIFEST_LOAD_STATE_FILE;
}

# --------------------------------------------
sub wait_for_bd_manifest_load {
    my %params = @_;

    my $logger = $params{logger};
    my $timeout = BD_MANIFEST_GRACE;
    my $msg;

    if (! -f $BD_MANIFEST_FILE ) {
        $msg = 'Going to wait for bd manifest';
        $logger->info($msg) if $logger;
        F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm
        while ($timeout) {
            $timeout--;
            last if -f $BD_MANIFEST_FILE;
            sleep 1;
        }
    }

    if (-f $BD_MANIFEST_FILE) {
        $msg = 'Going to wait for bd to finish reading bd_manifest';
        $timeout = BD_GRACE;
        while  ($timeout) {
            $timeout--;
            last if -f $BD_MANIFEST_LOAD_STATE_FILE;
            sleep 1;
        }
    }

    if ( !$timeout ) {
        $msg = 'Gave up waiting for bd to finish reading manifest. Trying a normal asm start-up sequence anyway';
    }
    else {
        $msg = 'bd has finished reading manifest';
    }
    $logger->info($msg) if $logger;
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm
}

# --------------------------------------------
sub remove_asm_upgrade_lock {
    unlink $ASM_UPGRADE_LOCK_FILE if -f $ASM_UPGRADE_LOCK_FILE;
}

# --------------------------------------------
sub write_asm_upgrade_lock {
    File::Slurp::write_file( $ASM_UPGRADE_LOCK_FILE, YAML::XS::Dump { date => F5::DateTimeUtils::convert_epoch_to_datetime(time()), pid => $$ } );
}

# --------------------------------------------
sub remove_asm_load_complete_end_lock_file {
    unlink($ASM_LOAD_COMPLETE_END_LOCK_FILE) if -f $ASM_LOAD_COMPLETE_END_LOCK_FILE;
}

# --------------------------------------------
sub write_asm_load_complete_end_lock_file {
    open (my $outf, '>', $ASM_LOAD_COMPLETE_END_LOCK_FILE) or die "$!\n";
    print $outf "$$";
    close ($outf);
}

# --------------------------------------------
sub exists_asm_load_complete_end_lock_file {
    return (-f $ASM_LOAD_COMPLETE_END_LOCK_FILE) ? 1 : 0 ;
}

# --------------------------------------------
# wait for bd to start all threads
sub wait_for_bd_threads {
    my %params = @_;

    my $logger = $params{logger};
    my $nwd = $params{nwd};
    my $timeout = BD_GRACE;

    my $msg = 'Going to wait for bd to start all threads';
    $logger->info($msg) if $logger;
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm

    while ($timeout) {
        $timeout--;
        last if $nwd->verify(proc => 'bd');
        sleep 1;
    }

    if ( !$timeout ) {
        $msg = 'Gave up waiting for bd to start all threads. Trying a normal asm start-up sequence anyway';
    }
    else {
        $msg = 'bd has started all threads';
    }
    $logger->info($msg) if $logger;
    F5::BigIpLog::bigip_ts_ts_info_msg_log($msg); # /var/log/asm
}

# --------------------------------------------

1; 
