package F5::Acc::Entity::Column;

use strict;
use warnings;
use Carp;
use Class::DBI::Column;

# column interface:
#   * set_default_on_entity
#   * make_cdbi_column
#   * validate_required
#   * is_uniq
#   * validate
#   * public name
#   * is_key

sub new {
    my ($class, $entity_class, $name, $descriptor) = @_;
    my $self = bless {
        name          => $name,
        entity_class  => $entity_class,
        key           => delete($descriptor->{key})           || 0,
        uniq          => delete($descriptor->{uniq})          || 0,
        null          => delete($descriptor->{null})          || 0,
        shallow_clone => delete($descriptor->{shallow_clone})  || 0,
        alias         => delete($descriptor->{alias}),
    }, $class;
    $self->{default} = delete($descriptor->{default})
        if exists $descriptor->{default};
    croak "can't set uniq on non-native CDBI columns [". ref($class). "]"
        if $self->is_uniq && !$self->is_cdbi;
    return $self;
}

sub get_default {
    my ($self, $entity) = @_;
    my $default = $self->{default};
    if ($self->has_default)
        { return ref $default eq 'CODE'? $default->($entity): $default }
    return $self->_get_default($entity);
}

sub validate_value {
    my ($self, $entity, $value) = @_;
    if (@_ == 2) { $value = $entity->_attrs($self->name) }
    $self->call_before_set_handler($entity, $self->name, $value);
    return if !defined($value) && $self->{null};
    return if !$self->is_cdbi || $self->is_key;
    $self->invalid('', "can't set undef because column type can't be null")
        unless defined $value;
    $self->_validate($value);
}

# sets the column on the entity to the default of the column type
# if needed
sub validate_default {
    my ($self, $entity) = @_;
    return if
        $self->is_key   ||
        !$self->is_cdbi ||
        $entity->_attribute_exists($self->name);
    $entity->_attribute_set( $self->name, $self->get_default($entity) );
}

sub make_cdbi_column {
    my $self = shift;
    return () unless $self->is_cdbi;
    my $alias = $self->alias;
    my @params = (
        $self->name,
        $alias? ({accessor => $alias, mutator => $alias}): ()
    );
    return Class::DBI::Column->new(@params);
}

sub validate_required {
    my ($self, $entity) = @_;
    return unless $self->is_uniq;
    my $name = $self->name;
    croak "missing required parameter for create: $name" if
        !$entity->_attribute_exists($name) && !$self->has_default;
}

sub invalid {
    my ($self, $value, $message) = @_;
    croak("$message [$self->{name}=$value]");
}

sub is_key          { shift->{key} }
sub name            { shift->{name}  }
sub alias           { shift->{alias} }
sub is_uniq         { shift->{uniq}  }
sub entity_class    { shift->{entity_class} }
sub shallow_clone    { shift->{shallow_clone} }
sub has_default     { exists shift->{default} }
sub public_name     { $_[0]->{alias} || $_[0]->{name} }

# used to delegate calls from a wrapper to its wrapped
# for this we need to know delegating methods we need to
# install on the wrapper
sub get_col_methods { return (shift->public_name) }

# triggers ---------------------------------------------------------------------

# call a before_set handler on an entity for a specific column
# and call its general before_set handler, called for each column
sub call_before_set_handler {
    my ($self, $entity, $name, $value) = @_;
    my $before_set_method = "before_set_$name";
    $entity->$before_set_method($value) if $entity->can($before_set_method);
    $entity->before_set($name, $value) if $entity->can('before_set');
}

sub call_after_set_handler {
    my ($self, $entity, $value) = @_;
    return if !$self->is_cdbi || $self->is_key;
    my $name  = $self->name;
    $value = $entity->_attrs($name) if @_ == 2;
    my $after_set_method = "after_set_$name";
    $entity->$after_set_method($value) if $entity->can($after_set_method);
    $entity->after_set($name, $value) if $entity->can('after_set');
}

# template methods -------------------------------------------------------------

sub _get_default { croak "Abstract method called" }
sub _validate    { croak "Abstract method called" }

sub is_cdbi { 1 }

sub make_entity_column {}

# can this column be given to CDBI insert constructor
# or is it a column which must be set after insert?
sub can_insert { 1 }

# called on all columns that are not overriden when cloning
sub clone_deep {
    my ($self, $entity) = @_;
    my $col_name = $self->public_name;
    my $value    = $entity->$col_name;
    return ($col_name, $value);
}

1;

