package F5::Acc::Entity::ColumnList;

use strict;
use warnings;
use Carp;
use List::Util qw(first);
use F5::Acc::Entity::Util;
use F5::Acc::Entity::Column::Factory;

# initialization ---------------------------------------------------------------

sub new {
    my ($class, $entity_class, $descriptor) = @_;
    my $self = bless {
        entity_class => $entity_class,
        descriptor   => $descriptor,
    }, $class;
    $self->parse_descriptor;
    return $self;
}

sub parse_descriptor {
    local $_;
    my $self = shift;
    my %desc = %{ $self->descriptor };
    $self->{cols_hash} = { map { $_ =>
        F5::Acc::Entity::Column::Factory->make_column
            ($self->entity_class, $_, $desc{$_})
    } keys %desc };
    $self->setup_column_filters;
}

sub setup_column_filters {
    local $_;
    my $self = shift;
    my %cols = %{$self->cols_hash};
    my @cols = values %cols;
    my $ec   = $self->entity_class;
    eval {
        $self->{cols_key}     =[grep {  $_->is_key            } @cols];
        $self->{cols_no_key} = [grep { !$_->is_key            } @cols];
        $self->{cols_uniq}   = [grep {  $_->is_uniq           } @cols];
        $self->{cols_cdbi}   = [grep {  $_->is_cdbi           } $self->cols_no_key];
        $self->{col_cdbi}    = first {  $_->is_cdbi           } $self->cols_no_key;
        $self->{pub_cdbi}    = [map  {  $_->public_name       } $self->cols_cdbi];
        $self->{cols}        = [$self->cols_key, $self->cols_no_key]; # key 1st
        $self->{cols_public_name}  = {map  {  $_->public_name => $_ } $self->cols};
    };
    croak "Schema with no columns [$ec]"      unless @{$self->cols};
    croak "Schema with no key [$ec]"          unless @{$self->cols_key};
    croak "Schema with no cdbi columns [$ec]" unless $self->col_cdbi;
    croak "Can't setup column filters [$ec]: $@" if $@;
}

sub clone_deep {
    my ($self, $entity, %override_params) = @_;
    my %cols_hash = %{ $self->cols_hash };
    # don't want to clone overriden columns
    delete $cols_hash{$_} for keys %override_params;
    # definately not the primary key - TODO need to test that
    for my $col_key ($entity->schema->cols_key) {
        delete $cols_hash{$col_key->name};
    }
    my @insert_params;
    for my $col_name (keys %cols_hash) {
        my $col = $cols_hash{$col_name};
        my @col_keys_and_values = $col->clone_deep($entity);
        push @insert_params, @col_keys_and_values;
    }
    my %insert_params = (@insert_params, %override_params);
    my $clone = $entity->schema->entity_class->insert( {%insert_params} );
    return $clone;
}

# CDBI resets the entity columns after create and update
# this reloads it back assuming all columns are essential
# only parameter is entity
sub reload { pop->get( shift->col_cdbi->name ) }

# validation -------------------------------------------------------------------

# called before create
sub validate_entity {
    my ($self, $entity) = @_;
    $_->validate_required($entity) for $self->cols;
    $_-> validate_default($entity) for $self->cols;
    $_->   validate_value($entity) for $self->cols;
    $self->validate_duplicates($entity);
}

# called before set not before create
sub call_before_set_handler {
    my ($self, $entity, $name, $value) = @_;
    eval {
        my $col = $self->cols_hash->{$name};
        $col->validate_value($entity, $value);
        # if a change to a uniq column, make sure we are not a duplicate now
        $self->validate_duplicates($entity, $name, $value) if $col->is_uniq;
    };
    if ($@) {
        $entity->discard_changes;
        croak $@;
    }
}

# called after set not after create
sub call_after_set_handler {
    my ($self, $entity, $name) = @_;
    eval { $self->cols_hash->{$name}->call_after_set_handler($entity) };
    if ($@) {
        $entity->discard_changes;
        croak $@;
    }
}

# can be called before create, with no name or value
# or before set, with a name and a value to be set
sub validate_duplicates {
    my ($self, $entity, $name, $value, $cols_to_validate) = @_;
    local $_;
    my @cols = $cols_to_validate? 
        (map { $self->cols_hash->{$_} } @$cols_to_validate):
        $self->cols_uniq;
    return unless @cols; # no uniq column names, no duplicate
    my %values = map { $_->name => $entity->_attrs($_->name) } @cols;
    $values{$name} = $value if $name; # may be in before_set, entity not set yet
    my @results = $entity->search(%values);
    croak 'duplicate exists for entity '. ref($entity). ', fields: '. join(",", %values)
        if @results;
}

# class dbi setup --------------------------------------------------------------

sub make_entity_columns {
    local $_;
    my $self = shift;
    my $entity_class = $self->entity_class;
    
    $entity_class->columns
        (Essential => map { $_->make_cdbi_column } $self->cols);

    if($self->cols_key)
    {
        $entity_class->columns
            (Primary => map { $_->make_cdbi_column } @{$self->cols_key});
    }
    $_->make_entity_column for $self->cols;
    $self->make_column_triggers;
}

sub make_column_triggers {
    my $self = shift;
    my $entity_class = $self->entity_class;
    for my $col ($self->cols_no_key) {
        my $name = $col->name;
        $entity_class->add_trigger("before_set_$name" => sub {
            # thing is class for insert, object when in the call flow of set
            my ($thing, $value) = @_;
            # We block before_set that is triggered by CDBI insert.
            # Need to call before_set handler manually from before_create.
            # We block because CDBI calls before_set from insert on the
            # class, not the object. As a workaround, we call it on the
            # object from our own handle_before_create.
            return unless ref $thing; # block before_set from CDBI insert
            $self->call_before_set_handler($thing, $name, $value);
        });
        $entity_class->add_trigger("after_set_$name" => sub {
            my ($thing, $value) = @_;
            return unless ref $thing;
            $self->call_after_set_handler($thing, $name);
        });
    }
}

# accessors --------------------------------------------------------------------

sub entity_class { shift->{entity_class} }   # which this is a column list of
sub descriptor   { shift->{descriptor} }     # the columns descriptor
sub cols_key      { wantarray? @{shift->{cols_key}}: shift->{cols_key} }       # the primary key columns
sub col_cdbi     { shift->{col_cdbi} }       # a random cdbi column
sub cols_hash    { shift->{cols_hash} }      # hash of all columns keyed by name

sub get_column_by_public_name { shift->{cols_public_name}->{pop()} }

{
    local $_;
    my $class = __PACKAGE__;
    for my $key (qw(
        cols_no_key cols_uniq cols_cdbi pub_cdbi cols
    )) {
        install_sub(
            $class,
            $key,
            sub { wantarray? @{ shift->{$key} }: shift->{$key} },
        );
    }
}

sub public_cdbi_column_names { shift->pub_cdbi }

1;

__END__

Used by schema to manage the list of column definitions.

cloning related nomenclature:
    overriden- when cloning you may want to override some columns.
               these will not be deeply cloned, instead the clone
               will have the overriden value. e.g. if you override
               a has_a column, then the target object will NOT be 
               cloned. Instead the clone will be connected to the
               overriden value given in a parameter to clone_deep
               which you override will
    shallow_clone- don't deep clone these columns. E.g. PolicyEntity
                  policy_id

