package F5::Acc::Entity::Schema;

use strict;
use warnings;
use Carp;
use F5::Acc::Entity::ColumnList;

sub new {
    my ($class, $entity_class) = @_;
    my $self = bless { entity_class => $entity_class }, $class;
    $self->parse_schema;
    return $self;
}

sub parse_schema {
    my $self = shift;
    local $_;
    my %descriptor    = $self->entity_class->compute_schema_descriptor;
    $self->{table}    = $descriptor{table};
    $self->{column_list} = F5::Acc::Entity::ColumnList->new
        ($self->entity_class, $descriptor{columns} || {});
}

# triggers ---------------------------------------------------------------------

sub handle_before_create {
    my ($self, $entity) = @_;
    eval {
        $self->validate_entity($entity);
        $entity->before_create if $entity->can('before_create');
    };
    if ($@) {
        $entity->discard_changes;
        die $@;
    }
}

sub handle_after_create {
    my ($self, $entity) = @_;
    local $_;
    eval {
        $self->reload($entity);
        $_->call_after_set_handler($entity) for $self->cols_cdbi;
        $entity->after_create if $entity->can('after_create');
    };
    if ($@) {
        $entity->discard_changes;
        die $@;
    }
}

sub handle_before_update {
    my ($self, $entity) = @_;
    $entity->before_update if $entity->can('before_update');
}

# TODO: maybe these 2 should be in eval to stop people from
# deleting constrained entities?
sub handle_before_delete {
    my ($self, $entity) = @_;
    $entity->before_delete if $entity->can('before_delete');
}

sub handle_after_delete {
    my ($self, $entity) = @_;
    $entity->after_delete if $entity->can('after_delete');
}

# class dbi setup --------------------------------------------------------------

# configure CDBI for a new persistent class
sub make_entity_class {
    my $self = shift;
    local $_;
    my $entity_class = $self->entity_class;
    $entity_class->table($self->table);
    $entity_class->autoupdate(0);
    $self->make_entity_columns;
    $self->make_entity_triggers;
}

sub make_entity_triggers {
    my $self = shift;
    my $entity_class = $self->entity_class;
    for my $h (qw(create delete)) {
        my $before = "handle_before_$h";
        my $after  = "handle_after_$h";
        $entity_class->add_trigger("before_$h" => sub { $self->$before(@_) });
        $entity_class->add_trigger("after_$h"  => sub { $self->$after(@_) });
    }
    $entity_class->add_trigger
        (before_update => sub { $self->handle_before_update(@_) });
}

# accessors --------------------------------------------------------------------

sub entity_class { shift->{entity_class} }
sub table        { shift->{table} }
sub column_list  { shift->{column_list} }

# delegated --------------------------------------------------------------------

sub validate_entity           { shift->column_list->validate_entity(@_) }
sub validate_duplicates       { shift->column_list->validate_duplicates(@_) }
sub reload                    { shift->column_list->reload(@_) }
sub cols_cdbi                 { shift->column_list->cols_cdbi(@_) }
sub cols_key                  { shift->column_list->cols_key }
sub cols_hash                 { shift->column_list->cols_hash(@_) }
sub make_entity_columns       { shift->column_list->make_entity_columns(@_) }
sub public_cdbi_column_names  { shift->column_list->public_cdbi_column_names }
sub get_column_by_public_name { shift->column_list->get_column_by_public_name(@_) }
sub clone_deep                { shift->column_list->clone_deep(@_) }

1;

__END__

# a schema object is a descriptor of an entity
# each entity class has a schema object
# the schema object is built from a hash specified in the entity
# the hash defines columns, table name, etc.
# the schema is used by the entity (usually the base entity) for:
#  * initializing Class::DBI (using CDBI methods table, column, has_a, etc.)
#  * validating column values and calling validators on the entity
#    this is built using the CDBI trigger mechanism
# also implements its own trigger mechanism, on top of CDBI's, which
# fixes some CDBI trgger problems:
#  * need to register a trigger, instead of just create a correctly named
#    handler sub
#  * before_set triggers send the class instead of the entity on create
#  * after_set triggers send an empty entity on create




# TODO: this will only work if we mark the entity as changed in user handler
# CDBI ignore the value in the entity, and rewrites it because
# it uses triggers for validation. We use it also for things
# like normalization- e.g. adding a / to path. If we want
# changes to the entity done in user code to stick, we need
# to hack the values in @_.
# CDBI behaves like this because Class::Trigger clones the
# entity before firing the trigger
