

package F5::Acc::Exception;

use strict;
use overload
    q{""}    => sub {shift->{string}},
    fallback => 1;
use F5::Acc::Log;

# NOTE: iControlDispatch is aware of some of the keys to this table,
# so any changes to key names must be checked there.
my %exception_table = (
    # code 1 is used in throw() for unknown exception keys
    # codes 2-10 are reserved for iControl/dispatch exceptions
    UnableToLoadPackage => {
        code => 11,
        string => q{Package '%s' could not be loaded. },
    },
    NameEmpty => {
        code => 12,
        string => q{The name is empty. },
    },
    NameInvalid => {
        code => 13,
        string => q{The name %s is invalid. },
    },
    GroupNotFound => {
        code => 14,
        string => q{The Web Application Group %s is not found. },
    },
    InternalError => {
        code => 15,
        string => q{An internal error has occurred. %s },
    },
    WebAppNotFound => {
        code => 16,
        string => q{The Web Application %s is not found. },
    },
    NameTooLong => {
        code => 17,
        string => q{The %s name is too long. },
    },
    GroupNameDuplicate => {
        code => 18,
        string => q{The Web Application Group "%s" already exists. },
    },
    CannotReconfDisabledWebapp => {
        code => 19,
        string => q{Cannot reconfigure disabled webapp: %s . },
    },
    WebAppNameDuplicate => {
        code => 20,
        string => q{The Web Application "%s" already exists. },
    },
    DatabaseError => {
        code => 21,
        string => q{Database Exception. ( %s ) },
    },
    LanguageAlreadySet => {
        code => 22,
        string => q{Language is already set for Web Application %s. },
    },
    WebAppAlreadyInGroup => {
        code => 23,
        string => q{Web application %s is already a member of %s group. },
    },
    CannotUnsetLanguage => {
        code => 24,
        string => q{Cannot unset language to web application. },
    },
    TrafficSamplingIntervalInvalid => {
        code => 25,
        string => q{The interval of traffic sampling must be a positive integer. },
    },
    LanguageNotSet => {
        code => 26,
        string => q{The language of the web application was not set yet. },
    },
    PolNotOfWebApp => {
        code => 27,
        string => q{The policy %s does not belong to web application %s. },
    },
    WebappNotBelongingToGroup => {
        code => 28,
        string => q{The web application %s does not belong to group %s. },
    },
    CannotDeleteWebapp => {
        code => 29,
        string => q{Cannot delete web application %s. },
    },
    WebAppEnabled => {
        code => 30,
        string => q{Web application %s is enabled. },
    },
    SetActivePolicyFailed => {
        code => 31,
        string => q{Setting active policy has failed. %s },
    },
    ASMNotEnabled => {
        code => 32,
        string => q{ASM is not licensed/enabled.},
        throw_level => 'info',
    },
    PolicyNameDuplicate => {
        code => 33,
        string => q{The Policy "%s" already exists. },
    },
    PolicyNotFound => {
        code => 34,
        string => q{The Policy "%s" does not exist. },
    },
    BlockingNotFound => {
        code => 35,
        string => q{The violation of policy: %d code: %d does not exist. },
    },
    DisableBlocking => {
        code => 36,
        string => q{The disable blocking flag is on. Cannot set on blocking. },
    },
    PolicyActive => {
        code => 37,
        string => q{The policy %s is active. Cannot delete policy. },
    },
    PolicyLearningSet => {
        code => 38,
        string => q{Unable to delete the policy %s. It is currently the "Apply Learning To" policy. },
    },
    PolicyCookieLengthLimit => {
        code => 39,
        string => q{The cookie length can be maximum %d. },
    },
    PolicyHeaderLengthLimit => {
        code => 40,
        string => q{The header length can be maximum %d. },
    },
    CrawlerIsRunning => {
        code => 41,
        string => q{Policy builder is currently running in policy %s. },
    },
    LanguageIllegal => {
        code => 42,
        string => q{The language is set incorrectly. },
    },
    ParserStatusCodeNotFound => {
        code => 43,
        string => q{No such violation %d exists},
    },
    ParserStatusCodeSeverityNotFound => {
        code => 44,
        string => q{No such severity %d exists},
    },
    NegsigSetNotFound => {
        code => 45,
        string => q{No such negsig set %s exists},
    },
    AttackTypeNotFound => {
        code => 46,
        string => q{No such attack type %s exists},
    },
    NegsigSetDuplicate => {
        code => 47,
        string => q{Negsig set %s already exists},
    },
    UnknownLoggingProfile => {
        code => 46,
        string => q{No such logging profile %s exists},
    },
    WebApplicationTimeout => {
        code => 47,
        string => q{Timelimit for set_languege_extended exceeded. Web Application doesn't exist yet.},
    },
    WebApplicationIsDefault => {
        code => 48,
        string => q{Web Application exists by default in the system. Cannot be deleted.},
    },
    XmlProfileDoesntExist => {
        code => 49,
        string => q{The XML profile %s does't exist.},
    },
    XmlProfileWSDLDoesntExist => {
        code => 50,
        string => q{The XML profile %s doesn't have a WSDL file %s.},
    },
    XmlProfileWSDLInvalid => {
        code => 51,
        string => q{The WSDL file %s is invalid. %s},
    },
    WebObjectNotFound => {
        code => 52,
        string => q{The URL %s was not found},
    },
    ParameterNotFound => {
        code => 53,
        string => q{The parameter %s was not found},
    },
    ASMNotReady => {
        code => 54,
        string => q{ASM is not ready},
    },
    SetActiveLoggingProfile => {
        code => 55,
        string => q{Setting active logging profile has failed. %s },
    },
    ExportPolicyNoFileExists => {
        code => 56,
        string => q{The exported policy file doesn't exist. Export failed.},
    },
);


sub throw
{
    my $class = shift;
    my $key = shift;

    my $e = $exception_table{ $key }
        or return bless {
            code => 1,
            string => "Exception '$key' was thrown",
            key => $key },
        $class;

    my $exception_string = sprintf(qq/$e->{ string }/, @_);

    if (! exists $e->{throw_level} || $e->{throw_level} eq 'error') {
        F5::Acc::Log->Error($exception_string);
    } else {
        F5::Acc::Log->Info($exception_string);
    }

    return bless {
        code => $e->{code},
        string => $exception_string,
        key => $key,
    }, $class;
}

1;
