package F5::ActionItem::Base;
use strict;
use warnings;

use F5::ASMConfig::Entity::PolicyActionItem;
use F5::ASMConfig::Entity::ActionItem;

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);

  return $self;
}

#----------------------------------------------------------------------

sub install_accessor {
  my($self, $field, $val) = @_;

  no strict 'refs';
  no warnings 'redefine';
  my $class = ref($self) || $self;
  *{$class . '::' . $field} = sub {
    return $val;
  };
  return $val;
}

#----------------------------------------------------------------------

#STUB
sub check_condition {
    return;
}

#----------------------------------------------------------------------

sub get_staging_period {
    my ($class, %params) = @_;
    my $staging_period;
    if ($params{event} && $params{event}->{element} &&
        $params{event}->{element}->{element_type} == F5::ASMConfig::ElementType::POLICY_ATTRIBUTES) {
        $staging_period = $params{event}->{element}->{policy_attributes}->{quarantined_period};
    }
    else {
       ($staging_period) = $params{dbh}->selectrow_array(<<EndQuery, undef, $params{policy}->{policy_id});
SELECT quarantined_period
  FROM PLC.PL_POLICY_GENERAL
 WHERE policy_id = ?
EndQuery
    }
    return $staging_period;
}

#----------------------------------------------------------------------

sub check_condition_staging {
    my ($class, %params) = @_;
    return unless $params{policy};
    return unless $params{policy}->{flg_encoding_selected};

    my $staging_period = $class->get_staging_period(%params);
    return unless defined $staging_period;

    my $now = time();
    my $staging_start_check_time = $now - ($staging_period * 60*60*24);

    my $min_time = $class->staging_class->get_next_staging_or_tightening_past_period_time(%params);

    my $condition = $min_time && ($min_time <= $staging_start_check_time);
    if ( ! $condition && $min_time && $params{schedule} ) {
        my $trigger_time = time() + $min_time - $staging_start_check_time;
        $params{action_item_handler}->task_queue_add(task => {module    => $class,
                                                         policy_id => $params{policy}->{policy_id},
                                                         trigger_time => $trigger_time,},
                                               );
    }
    return $condition;

}

#----------------------------------------------------------------------

sub detect {
    my ($class, %params) = @_;
    my $handler = $params{action_item_handler};
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $event = $params{event};

    if ($class->is_pb_dependant() && $params{policy}->{is_pb_enabled}) {
        return;
    }

#TODO: Check in memory
    my $entity_class = $class->is_policy_action_item ? "F5::ASMConfig::Entity::PolicyActionItem"
                                                     : "F5::ASMConfig::Entity::ActionItem";
    if ($entity_class->new(dbh => $dbh, policy => $params{policy}, get_criteria => {action_item_type => $class->type()})) {
        return;
    }

    if ($params{force} || $class->check_condition(%params)) {
        my $now = time();
        $entity_class->new(dbh => $dbh,
                           policy => $params{policy},
                           object => {action_item_type => $class->type(),
                                      created_time => $now,
                                      next_display_time => $now,
                                      severity => $class->severity(),
                                      message => $handler->get_message(type => $class->type()),}
                          )->store_add();
    }
    return;
}

#----------------------------------------------------------------------

sub clean {
    my ($class, %params) = @_;

    my $handler = $params{action_item_handler};
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $event = $params{event};
    my $force = $params{force};
  
    if ($class->is_pb_dependant() && $params{policy}->{is_pb_enabled}) {
        $force = 1;
    }

    my $entity_class = $class->is_policy_action_item ? "F5::ASMConfig::Entity::PolicyActionItem"
                                                     : "F5::ASMConfig::Entity::ActionItem";

    my $action_item = $entity_class->new(dbh => $dbh,
                                         policy => $params{policy}, get_criteria => {action_item_type => $class->type()});
    if ($action_item) {
          if ($force || ! $class->check_condition(%params) ) {
              return $action_item->delete();
          }
    }
    return;
}


#----------------------------------------------------------------------

sub elements {
  my($class) = @_;
  no strict 'refs';

  my $elements = [];
  if (${$class . '::elements'}) {
      $elements = ${$class . '::elements'};
  }
  return $class->install_accessor('elements', $elements);
}

#----------------------------------------------------------------------

sub attributes {
  my($class) = @_;
  no strict 'refs';

  my $attributes;
  if (${$class . '::attributes'}) {
      $attributes = ${$class . '::attributes'};
  }
  return $class->install_accessor('attributes', $attributes);
}


#----------------------------------------------------------------------

sub type {
  my($class) = @_;
  return $class->attributes->{type};
}

#----------------------------------------------------------------------

sub staging_class {
  my ($class) = @_;
  return $class->attributes->{staging_class};
}

#----------------------------------------------------------------------

sub is_subscriber {
  my($class) = @_;
  return $class->attributes->{subscriber};
}


#----------------------------------------------------------------------

sub is_pb_dependant {
  my($class) = @_;
  return $class->attributes->{pb_dependant};
}

#----------------------------------------------------------------------

sub is_policy_action_item {
  my($class) = @_;
  return $class->attributes->{policy_action_item};
}

#----------------------------------------------------------------------

sub name {
  my($class) = @_;
  return $class->attributes->{name};
}

#----------------------------------------------------------------------

sub severity {
  my($class) = @_;
  return $class->attributes->{severity};
}

#----------------------------------------------------------------------

sub throttle_events {
    my($class) = @_;
    return $class->attributes->{throttle_events};
}

#----------------------------------------------------------------------

1;
