package F5::ActionItemHandler;

use strict;
use warnings;

use F5::Deploy;
use F5::ActionItem::Base;
use F5::DbUtils;
use F5::ModuleLoader;
use F5::ASMConfig::ASMConfig;
use YAML::XS;

my $ACTION_ITEM_CONFIG_FILENAME = F5::Deploy::map_path('/usr/share/ts/config/action_items.yaml');
our @DYNAMIC_MODULES = F5::ModuleLoader::load_from_base_class('F5::ActionItem::Base');
our $THROTTLED_EVENT_RATE_LIMT = 60;

my %singleton;
preload(\%singleton);

#----------------------------------------------------------------------

sub preload {
    my ($self) = @_;

    $self->{config} = YAML::XS::LoadFile($ACTION_ITEM_CONFIG_FILENAME);

    $self->{element_triggers} = {F5::ASMConfig::ElementType::UCS_CONFIG_LOAD => \@DYNAMIC_MODULES};
    $self->{table_triggers} = {};
    $self->{subscriptions} = { elements => [ ],
                               policy_elements => [ ],
                               pb_dependant_elements => [ ],
                             };
    $self->{type_cache} = {};
    $self->{name_cache} = {};
    $self->{policy_modules} = [];
    $self->{pb_dependant_modules} = [];

    my %subscription_elements = (F5::ASMConfig::ElementType::UCS_CONFIG_LOAD => 1,);
    my %policy_subscription_elements = (F5::ASMConfig::ElementType::SET_ACTIVE => 1,
                                        F5::ASMConfig::ElementType::DISABLED_ACTION_ITEM => 1,
                                        F5::ASMConfig::ElementType::PBCONFIG => 1,
                                        F5::ASMConfig::ElementType::POLICY => 1,
                                        F5::ASMConfig::ElementType::SUGGESTION => 1,
                                       );
    my %pb_dependant_subscription_elements;
    my %policy_throttled_subscription_elements;
    my %pb_dependant_throttled_subscription_elements;
    for my $module (@DYNAMIC_MODULES) {
        for my $element (@{$module->elements()}) {
            push(@{$self->{element_triggers}->{$element}}, $module);
            if ($module->is_subscriber()) {
                if ($module->is_pb_dependant) {
                    if ($module->throttle_events) {
                        $pb_dependant_throttled_subscription_elements{$element} = 1;
                    }
                    else {
                        $pb_dependant_subscription_elements{$element} = 1;
                    }
                }
                elsif ($module->is_policy_action_item) {
                    if ($module->throttle_events) {
                        $policy_throttled_subscription_elements{$element} = 1;
                    }
                    else {
                        $policy_subscription_elements{$element} = 1;
                    }
                }
                else {
                    $subscription_elements{$element} = 1;
                }
            }
        }
        if ($module->is_policy_action_item()) {
            push @{$self->{policy_modules}}, $module;
        }
        if ($module->is_pb_dependant()) {
            push @{$self->{pb_dependant_modules}}, $module;
        }
        $self->{type_cache}->{$module->type()} = $module;       
        $self->{name_cache}->{$module->name()} = $module;       
    }
    push @{$self->{subscriptions}->{elements}}, keys %subscription_elements;
    push @{$self->{subscriptions}->{policy_elements}}, keys %policy_subscription_elements;
    push @{$self->{subscriptions}->{pb_dependant_elements}}, keys %pb_dependant_subscription_elements;
    push @{$self->{subscriptions}->{policy_throttled_elements}}, keys %policy_throttled_subscription_elements;
    push @{$self->{subscriptions}->{pb_dependant_throttled_elements}}, keys %pb_dependant_throttled_subscription_elements;
}

#----------------------------------------------------------------------

sub get_instance {
    my ($class, %params) = @_;
    my $self = \%singleton;
    bless $self, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------


sub init {
    my ($self, %params) = @_;
    $self->{task_queue} = [];
    $self->{dbh} = $params{dbh};
    $self->{client} = $params{client};
    $self->{logger} = $params{logger};
    $self->{cache} = $params{cache};
    $self->{active_policies} = {};
    $self->{disabled_types} = {};
    $self->reinit(%params);
}

#----------------------------------------------------------------------

sub reinit {
    my ($self, %params) = @_;
    return unless $self->{cache};
    $self->{task_queue} = [];
    $self->reinit_active_policies(%params);
}

#----------------------------------------------------------------------

sub reinit_active_policies {
    my ($self, %params) = @_;
    return unless $self->{cache};
    ## Get info on active policies.
    my @active_policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh}, get_criteria => {is_active_policy => 1});
    $self->{active_policies} = {map { ($_->{policy_id} => $_) } @active_policies};
    $self->{account_policies} = {map { ($_->{account_id} => $_) } @active_policies};
    $self->{disabled_types} = $self->{dbh}->selectall_hashref(<<"EndQuery", [qw(policy_id action_item_type)]);
SELECT policy_id, action_item_type
 FROM PLC.PL_DISABLED_ACTION_ITEMS
EndQuery
}

#----------------------------------------------------------------------

sub refresh_disabled_types {
    my ($self, %params) = @_;
    return unless $self->{cache};
    my $policy = $params{policy};
    $self->{disabled_types}->{$policy->{policy_id}} = $self->{dbh}->selectall_hashref(<<"EndQuery", "action_item_type", undef, $policy->{policy_id});
SELECT action_item_type
 FROM PLC.PL_DISABLED_ACTION_ITEMS
  WHERE policy_id = ?
EndQuery
}

#----------------------------------------------------------------------

sub set_dbh {
    my ($self, %params) = @_;
    $self->{dbh} = $params{dbh};
}

#----------------------------------------------------------------------

sub set_client {
    my ($self, %params) = @_;
    $self->{client} = $params{client};
}

#----------------------------------------------------------------------
sub get_subscriptions {
    my ($self, %params) = @_;
    my @subscriptions;
    my @element_event_types = map { F5::ASMConfig::ElementEventType->new( { event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                            element_type => $_ })
                                  } @{$self->{subscriptions}->{elements}};
    push (@subscriptions, F5::ASMConfig::EventSubscription->new({element_event_types => \@element_event_types,
                                                                 send_topology_suppressed_events => 1,
                                                                 send_catchup_events => 0,
                                                                }));

    if (@{$self->{subscriptions}->{policy_elements}}) {
        my @element_event_types = map {
            F5::ASMConfig::ElementEventType->new( { event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                    element_type => $_ })
          } @{$self->{subscriptions}->{policy_elements}};
        push (@subscriptions, F5::ASMConfig::EventSubscription->new({element_event_types => \@element_event_types,
                                                                     policy => F5::ASMConfig::Policy->new({is_active_policy => 1}),
                                                                     send_topology_suppressed_events => 1,
                                                                     send_catchup_events => 0,
                                                                    }));
    }

    if (@{$self->{subscriptions}->{policy_throttled_elements}}) {
        my @element_event_types = map {
            F5::ASMConfig::ElementEventType->new( { event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                    element_type => $_ })
          } @{$self->{subscriptions}->{policy_throttled_elements}};
        push (@subscriptions, F5::ASMConfig::EventSubscription->new({element_event_types => \@element_event_types,
                                                                     policy => F5::ASMConfig::Policy->new({is_active_policy => 1}),
                                                                     send_topology_suppressed_events => 1,
                                                                     send_catchup_events => 0,
                                                                     rate_limit_per_policy => $THROTTLED_EVENT_RATE_LIMT,
                                                                    }));
    }

    if (@{$self->{subscriptions}->{pb_dependant_elements}}) {
        my @element_event_types = map {
            F5::ASMConfig::ElementEventType->new( { event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                    element_type => $_ })
          } @{$self->{subscriptions}->{pb_dependant_elements}};
        push (@subscriptions, F5::ASMConfig::EventSubscription->new({element_event_types => \@element_event_types,
                                                                     policy => F5::ASMConfig::Policy->new({is_active_policy => 1,
                                                                                                           is_pb_enabled => 0,}),
                                                                     send_topology_suppressed_events => 1,
                                                                     send_catchup_events => 0,
                                                                    }));
    }

    if (@{$self->{subscriptions}->{pb_dependant_throttled_elements}}) {
        my @element_event_types = map {
            F5::ASMConfig::ElementEventType->new( { event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                    element_type => $_ })
          } @{$self->{subscriptions}->{pb_dependant_throttled_elements}};
        push (@subscriptions, F5::ASMConfig::EventSubscription->new({element_event_types => \@element_event_types,
                                                                     policy => F5::ASMConfig::Policy->new({is_active_policy => 1,
                                                                                                           is_pb_enabled => 0,}),
                                                                     send_topology_suppressed_events => 1,
                                                                     send_catchup_events => 0,
                                                                     rate_limit_per_policy => $THROTTLED_EVENT_RATE_LIMT,
                                                                    }));
    }

    return \@subscriptions;
}

#----------------------------------------------------------------------

sub subscribe {
    my ($self, %params) = @_;
    my $client = $params{client} || $self->{client};

    for my $subscription (@{$self->get_subscriptions}) {
        $client->add_event_subscription($client->{session}, $subscription);
    }
    return;
}

#----------------------------------------------------------------------

sub get_affected_modules {
    my ($self, %params) = @_;
    my $event = $params{event};
    my $mode  = $params{mode};

    my @modules;
    if ($event && $event->{element}) {
        my $element = $event->{element};

        if ($self->{element_triggers}->{$element->{element_type}}) {
            @modules =  @{$self->{element_triggers}->{$element->{element_type}}};
        }

        if ($self->{element_triggers}->{F5::ASMConfig::ElementType::ALL_ELEMENT_TYPES}) {
            for my $module (@{$self->{element_triggers}->{F5::ASMConfig::ElementType::ALL_ELEMENT_TYPES}}) {
                if ( ! $module->is_policy_action_item || $event->{policy} ) {
                    push (@modules, $module);
                }
            }
        }

        if ($element->{element_type} == F5::ASMConfig::ElementType::DISABLED_ACTION_ITEM) {      
            @modules = ($self->{type_cache}->{$element->{disabled_action_item}->{action_item_type}});
        }
        elsif ($element->{element_type} == F5::ASMConfig::ElementType::SET_ACTIVE) {
            if ($element->{set_active}->{active_policy_was_changed}) {
                 @modules =  @{$self->{policy_modules}};
            }
        }
        elsif ($element->{element_type} == F5::ASMConfig::ElementType::POLICY && 
               $element->{policy}->{is_active_policy} &&
                ( ($event->{event_type} == F5::ASMConfig::EventType::DELETE && $mode eq 'clean') ||
                  ($event->{event_type} == F5::ASMConfig::EventType::ADD && $mode eq 'detect') )
              ) {
            @modules =  @{$self->{policy_modules}};
        }
        elsif ($element->{element_type} == F5::ASMConfig::ElementType::SUGGESTION) {
            my $suggestion_element_type = $element->{suggestion}->{element}->{element_type};
            if ($self->{element_triggers}->{$suggestion_element_type}) {
                push(@modules, @{$self->{element_triggers}->{$suggestion_element_type}});
            }
        }
    }
    else {
        @modules = @DYNAMIC_MODULES;
    }
    return @modules;
}

#----------------------------------------------------------------------


sub update_internals {
    my ($self, %params) = @_;
    return unless $self->{cache};
    my $event = $params{event};
    
    if ($event && $event->{element}) {
        my $element = $event->{element};
        my $policy = $event->{policy};
      
        if ($element->{element_type} == F5::ASMConfig::ElementType::DISABLED_ACTION_ITEM) {
            if ($event->{event_type} == F5::ASMConfig::EventType::ADD) {
                $self->{disabled_types}->{$policy->{policy_id}}->{$element->{disabled_action_item}->{action_item_type}} = 1;
            }
            elsif ($event->{event_type} == F5::ASMConfig::EventType::DELETE) {
                delete $self->{disabled_types}->{$policy->{policy_id}}->{$element->{disabled_action_item}->{action_item_type}};
            }
        }
        elsif ( ($element->{element_type} == F5::ASMConfig::ElementType::SET_ACTIVE && $element->{set_active}->{active_policy_was_changed}) ||
                ($element->{element_type} == F5::ASMConfig::ElementType::POLICY && $event->{event_type} == F5::ASMConfig::EventType::ADD && $policy->{is_active_policy})
              ) {
             if (my $old_policy = $self->{account_policies}->{$policy->{account_id}}) {
                 delete $self->{active_policies}->{$old_policy->{policy_id}};
                 delete $self->{disabled_types}->{$old_policy->{policy_id}};
             }
             $self->{account_policies}->{$policy->{account_id}} = $policy;                                          
             $self->refresh_disabled_types(policy => $policy);
        }
        elsif ($element->{element_type} == F5::ASMConfig::ElementType::POLICY &&
               $event->{event_type} == F5::ASMConfig::EventType::DELETE &&
               $policy->{is_active_policy}) {
            delete $self->{account_policies}->{$policy->{account_id}};
            delete $self->{active_policies}->{$policy->{policy_id}};
            delete $self->{disabled_types}->{$policy->{policy_id}};
            undef $policy;
        }
        elsif ($element->{element_type} == F5::ASMConfig::ElementType::UCS_CONFIG_LOAD) {
            $self->reinit(%params);
        }
        elsif ($element->{element_type} == F5::ASMConfig::ElementType::PBCONFIG) {
            $self->update_pb_config_state(policy => $policy,
                                          event => $event);
        }
        
        if ($policy && $policy->{is_active_policy}) {
            $self->{active_policies}->{$policy->{policy_id}} = $policy;
        }
    }
}

#----------------------------------------------------------------------

sub update_pb_config_state {
    my ($self, %params) = @_;
    my $policy = $params{policy};
    return unless $policy && $policy->{is_active_policy} && exists($self->{active_policies}->{$policy->{policy_id}});
    #Check for special case of toggling PBConfig flag

    if ($policy->{is_pb_enabled} != $self->{active_policies}->{$policy->{policy_id}}->{is_pb_enabled}) {
        my $modules = $self->{pb_dependant_modules};
        my $func = $policy->{is_pb_enabled} ? "clean" : "detect";
        for my $module (@$modules) {
            $module->$func(%params,
                           action_item_handler => $self,
                           dbh => $self->{dbh},
                           logger => $self->{logger},
                           );
        }
    }
}

#----------------------------------------------------------------------

sub detect {
    my ($self, %params) = @_;
    $self->update_internals(%params);
    my @modules = $self->get_affected_modules(mode => "detect", 
                                              %params);
    for my $module (@modules) {
        #Check if policy is still active, and type is not disabled 
        if ($module->is_policy_action_item()) {
            my @policies = $params{event} && $params{event}->{policy} ? ($params{event}->{policy})
                                                                      : values %{$self->{active_policies}};
            for my $policy (@policies) {
                next if exists $self->{disabled_types}->{$policy->{policy_id}}->{$module->type()};
                next if ! $policy->{is_active_policy};
                $module->detect(action_item_handler => $self,
                                policy => $policy,
                                dbh => $self->{dbh},
                                logger => $self->{logger},
                                schedule => 1,
                                %params);
            }
        }
        else {
           $module->detect(action_item_handler => $self,
                           dbh => $self->{dbh},
                           logger => $self->{logger},
                           schedule => 1,
                            %params);
        }
    }
}

#----------------------------------------------------------------------

sub clean {
    my ($self, %params) = @_;
    $self->update_internals(%params);
    my @modules = $self->get_affected_modules(mode => "clean",
                                              %params);

    my @policies = $params{event} && $params{event}->{policy} ? ($params{event}->{policy})
                                                              : values %{$self->{active_policies}};
    for my $module (@modules) {
        if ($module->is_policy_action_item()) {
            for my $policy (@policies) {
                $module->clean(action_item_handler => $self,
                               policy => $policy,
                               dbh => $self->{dbh},
                               logger => $self->{logger},
                               %params);
            }
       }
       else {
           $module->clean(action_item_handler => $self,
                          dbh => $self->{dbh},
                          logger => $self->{logger},
                          %params);
       }
    }
}


#----------------------------------------------------------------------

sub get_dismiss_time {
    my ($self, %params) = @_;
    my $dismiss_time = $self->{config}->{default}->{dismiss_time};
    if ($self->{type_cache}->{$params{type}}) {
        my $name = $self->{type_cache}->{$params{type}}->name();
        if ($self->{config}->{action_items}->{$name}->{dismiss_time}) {
            $dismiss_time = $self->{config}->{action_items}->{$name}->{dismiss_time};
        }
    }
    return $dismiss_time;
}

#----------------------------------------------------------------------

sub get_message {
    my ($self, %params) = @_;
    if ($self->{type_cache}->{$params{type}}) {
        my $name = $self->{type_cache}->{$params{type}}->name();
        return $self->{config}->{action_items}->{$name}->{message};
    }
    return;
}

#----------------------------------------------------------------------

sub check_task_queue {
    my ($self, %params) = @_;
    while ($self->{task_queue}->[0] && ($self->{task_queue}->[0]->{trigger_time} < time())) {
        my $task = shift @{$self->{task_queue}};
        my $module = $task->{module};
        if ($module->is_policy_action_item()) {
            my $policy_id = $task->{policy_id};
            my $active_policy = exists $self->{active_policies}->{$policy_id} ? $self->{active_policies}->{$policy_id} : undef;

            next unless $active_policy;
            next if exists $self->{disabled_types}->{$active_policy->{policy_id}}->{$module->type()};
            $module->detect(action_item_handler => $self,
                            policy => $active_policy,
                            dbh => $self->{dbh},
                            logger => $self->{logger},
                            schedule => 1,
                            %params);
        }
        else {
            $module->detect(action_item_handler => $self,
                            dbh => $self->{dbh},
                            logger => $self->{logger},
                            schedule => 1,
                            %params);
        }
    }
}

#----------------------------------------------------------------------

sub next_run_interval {
    my ($self, %params) = @_;
    if ($self->{task_queue}->[0]) {
        return ($self->{task_queue}->[0]->{trigger_time} - time);
    }
    return 60*60*24; #default to a day
}

#----------------------------------------------------------------------

sub task_queue_add {
    my ($self, %params) = @_;
    my $task = $params{task};
    my @queue = @{$self->{task_queue}};
    @queue = grep { ($_->{module} ne $task->{module}) || ($_->{policy_id} != $task->{policy_id}) } @queue;
    push @queue, $task;
    @queue = sort {$a->{trigger_time} <=> $b->{trigger_time}} @queue;
    $self->{task_queue} = \@queue;
    return;
}

#----------------------------------------------------------------------

1;
