package F5::Asmqkview;

use strict;
use warnings;

use File::Path;
use File::Find;
use List::Util;
use File::Copy;
use File::Basename;

use F5::DbUtils;
use F5::ConfigSync;
use F5::SystemUtils;
use F5::RequestLog;

use constant BYTES_IN_1_MB => 1024 * 1024;
my $DEFAULT_DUMP_DIR = '/var/tmp/asm_tables_dump';

sub run_asmqkview{
   my (%params) = @_;

   my $flg_add_request_log = defined $params{flg_add_request_log} ? $params{flg_add_request_log} : 0;
   my $request_log_size_limit = defined $params{request_log_size_limit} ? $params{request_log_size_limit} : 0;
   my $flg_include_traffic_data = defined $params{flg_include_traffic_data} ? $params{flg_include_traffic_data} : 0;
   my $flg_no_policy_history = defined $params{flg_no_policy_history} ? $params{flg_no_policy_history} : 0;
   my $dump_dir = defined $params{dump_dir} ? $params{dump_dir} : $DEFAULT_DUMP_DIR;
   
   File::Path::rmtree($dump_dir);

   my $mysql_user = F5::DbUtils::get_mysql_user();
   my $mysql_password = F5::DbUtils::get_mysql_password(user => $mysql_user);
   return unless ($mysql_password);

   $mysql_user = "-u$mysql_user";
   $mysql_password = "-p$mysql_password";

   mkpath($dump_dir);
   chmod(0777, $dump_dir);

   my @plc_non_config_sync_tables = remove_prefix(F5::DbUtils::get_non_config_sync_tables('PLC'));
   my $plc_non_configsync_dump_file = "$dump_dir/plc_non_configsync_dump.sql"; 

   my @mysql_command = (qw(/usr/bin/mysqldump), $mysql_user, $mysql_password,
                        qw(--comments --complete-insert --add-drop-table
                           --add-locks --create-options --quick --single-transaction
                           --extended-insert --lock-tables --set-charset --disable-keys));

   my @cmd = ( @mysql_command,
            '--result-file',$plc_non_configsync_dump_file,
            'PLC',@plc_non_config_sync_tables);

   run_system_command(@cmd);

   my @dcc_non_config_sync_tables = remove_prefix(F5::DbUtils::get_non_config_sync_tables('DCC'));
   my @dcc_active_tables = remove_prefix(F5::DbUtils::get_active_policy_tables('DCC'));

   @dcc_non_config_sync_tables = remove_tables_from_list( exclude_tables => \@dcc_active_tables,
                                                             tables_list => \@dcc_non_config_sync_tables);

   my $dcc_non_configsync_dump_file = "$dump_dir/dcc_non_configsync_dump.sql";

   @cmd = ( @mysql_command,
           '--result-file',$dcc_non_configsync_dump_file,
           'DCC',@dcc_non_config_sync_tables);

   run_system_command(@cmd);

   my @prx_asmqkview_tables = remove_prefix(F5::DbUtils::get_asmqkview_include_tables('PRX'));
   my $prx_asmqkview_dump_file = "$dump_dir/prx_asmqkview_dump.sql"; 

   @cmd = ( @mysql_command,
            '--result-file',$prx_asmqkview_dump_file,
            'PRX',@prx_asmqkview_tables);

   run_system_command(@cmd);

   if ($flg_add_request_log) {
       my $request_log_total_size = 0;
       my @request_log_files;
       find({
             wanted => sub {
                 return unless (-f $File::Find::name);

                 my $current_file_size = -s $File::Find::name;
                 if (! $request_log_size_limit || 
                     $request_log_total_size + $current_file_size < ($request_log_size_limit * BYTES_IN_1_MB)) {

                     push( @request_log_files, $File::Find::name );
                     $request_log_total_size += $current_file_size;
                 }
             },
             no_chdir => 1,
             preprocess => sub { return reverse sort @_ }, # add files in reverse lexical order, in case of size limit
            },
            $F5::RequestLog::PROTOBUF_FILE_PATH);

            if ( @request_log_files ) {
                 run_system_command('tar','-zcf',"$dump_dir/request_log.tgz", @request_log_files );
            }
   }

   for my $mysql_file (glob("/var/lib/mysql/mysqld.err*")) {
      File::Copy::copy($mysql_file, $dump_dir . "/" . basename($mysql_file));
   }

   save_asm_config(flg_no_policy_history => $flg_no_policy_history,
                   dump_dir => $dump_dir);

   dump_cmd_output(command_line_params => [ qw(/usr/share/dosl7/bin/dosl7d_mem_dump --csv --prvnt-db) ],
                   output_filename => "$dump_dir/dosl7.prvnt_db.csv");

   dump_cmd_output(command_line_params => [ qw(/usr/share/dosl7/bin/dosl7d_mem_dump --csv --arb-prvnt-db) ],
                   output_filename => "$dump_dir/dosl7.arb_prvnt_db.csv");


   if ($flg_include_traffic_data){
	my $tcpdump_dir = '/shared/dosl7/tcpdumps';
	my $tcpdump_new_dir = '/var/tmp/dosl7_tcpdumps';
	mkdir $tcpdump_new_dir;
	find({ wanted => sub {
			return unless -f $File::Find::name;
			my $new_filename = $File::Find::name;
			$new_filename =~ s/$tcpdump_dir/$tcpdump_new_dir/;
			link($File::Find::name, $new_filename);
			},
		no_chdir => 1,
		},
        $tcpdump_dir);
	}
}

#----------------------------------------------------------------------------
sub remove_prefix{
    my (@tables) = @_;

    for my $table (@tables) {
       $table =~ s/^.*?\.//;
    }

    return @tables;
}
#-----------------------------------------------------------------------------
sub save_asm_config{
  my (%params) = @_;

  my $flg_no_policy_history = $params{flg_no_policy_history};
  my $dump_dir = $params{dump_dir};

  my $configsync_base_dir = "/var/tmp/asmqkview__$$";
  mkpath($configsync_base_dir);
  my $configsync = F5::ConfigSync->new(
                                       no_private_keys => 1,
                                       policy_history => !$flg_no_policy_history,
                                       base_dir => $configsync_base_dir,
                                      );

  my $configsync_save_file = "/var/tmp/asmqkview__$$.tar.gz";
  eval {
     $configsync->save(save_tar_file => $configsync_save_file,
                       do_lock => 0,
                      );
     $configsync->cleanup(remove_temp_databases => 0);
  };

  if($@){
   warn "Failed saving bigip ucs file in conf.tgz: $@ \n";
   return;
  }

  move($configsync_save_file, "$dump_dir/conf.tgz");
  rmtree($configsync_base_dir);
}

#-----------------------------------------------------------------------------

sub run_system_command{
  my (@cmd) = @_;
  return  F5::SystemUtils::run_system_tool(command_line_params => \@cmd);
}

#-----------------------------------------------------------------------------

sub dump_cmd_output {
    my(%params) = @_;

    my $command_line_params = $params{command_line_params};
    my $output_filename = $params{output_filename};

    my $cmd_output = F5::SystemUtils::run_system_tool(command_line_params => $command_line_params,
                                                      timeout => 60,
                                                     );
    if ($cmd_output && $cmd_output->{completed_successfully}) {
        open(my $out_fh, '>', $output_filename);
        print $out_fh $cmd_output->{standard_output};
        close $out_fh;

        return 1;
    }

    return;
}

#-----------------------------------------------------------------------------

sub remove_tables_from_list{
  my (%params) = @_;
  my @exclude_tables = @{$params{exclude_tables}};
  my @tables_list = @{$params{tables_list}};

  my %exclude_tables = map { ($_ => 1) } @exclude_tables;

  my @final_list = grep { !$exclude_tables{$_} } @tables_list;

  return @final_list;
  
}

1;
