package F5::BatchInsert;
use strict;
use warnings;

use F5::DbUtils;
use F5::Db::Single;
#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = bless {}, $class;

  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{dbh} = $params{dbh};
  $self->{logger} = $params{logger};

  $self->{batch_size} = $params{batch_size} || 50;

  $self->{table_batch} = {};
  $self->{table_batch_insert_order} = [];
  $self->{table_prerequisites} = {};
  $self->{table_batch_oldest_record} = {};
  $self->{table_postinsert_functions} = {};
  $self->{table_deferred_ids} = {};
  $self->{table_opts} = {};
  $self->{db_handler} = $params{db_handler};
  $self->{row_handler} = $params{row_handler};
  # todo: add parameters for replace, etc?
}

#----------------------------------------------------------------------
sub update_row {
  my ($self, %params) = @_;

  my $dbh = $self->{dbh} || F5::Db::Single->instance;
  if ($self->{row_handler}) {
      $self->{row_handler}->(%params,
                             dbh => $dbh);
  }

  $self->{logger}->debug("Flushing to database") if $self->{logger};
  $self->flush();

  return F5::DbUtils::update_row(%params,
                                 dbh => $dbh);
}

#----------------------------------------------------------------------

sub insert_row {
  my($self, %params) = @_;

  my $table_name = $params{table_name};
  my $rec = { %{$params{rec}} };
  my $prerequisite_table_names = $params{prerequisite_table_names};

  if (defined($params{fields})) {
      $self->{table_opts}->{$table_name}->{fields} = $params{fields};
  }

  if (defined($params{do_insert_ignore})) {
    $self->{table_opts}->{$table_name}->{do_insert_ignore} = $params{do_insert_ignore};
  }

  if (defined($params{do_replace})) {
    $self->{table_opts}->{$table_name}->{do_replace} = $params{do_replace};
  }

  if (defined($params{set_date_fields_to_now})) {
    $self->{table_opts}->{$table_name}->{set_date_fields_to_now} = $params{set_date_fields_to_now};
  }

  if (defined($params{set_date_fields_to_now_unix})) {
      $self->{table_opts}->{$table_name}->{set_date_fields_to_now_unix} = $params{set_date_fields_to_now_unix};
  }

  unless ($self->{table_batch}->{$table_name}) {
    $self->{table_batch}->{$table_name} = [];
    push(@{$self->{table_batch_insert_order}}, $table_name);
  }
  unless ($self->{table_batch_oldest_record}->{$table_name}) {
    $self->{table_batch_oldest_record}->{$table_name} = time;
  }
  if ($prerequisite_table_names) {
      for my $prerequisite_table_name (@$prerequisite_table_names) {
          $self->{table_prerequisites}->{$table_name}->{$prerequisite_table_name} = 1;
      }
  }

  if ($self->{row_handler}) {
      $self->{row_handler}->(table_name => $table_name,
                             rec => $rec,
                             dbh => $self->{dbh});
  }

  push(@{$self->{table_batch}->{$table_name}}, $rec);

  if ($params{get_deferred_id}) {
    my $deferred_id = F5::BatchInsert::DeferredID->new(batch_insert => $self,
                                                       table_name => $table_name);
    unless ($self->{table_deferred_ids}->{$table_name}) {
      $self->{table_deferred_ids}->{$table_name} = [];
    }
    push(@{$self->{table_deferred_ids}->{$table_name}}, 
         {
          batch_index => $#{$self->{table_batch}->{$table_name}},
          deferred_id_obj => $deferred_id,
         });

    ${$params{get_deferred_id}} = $deferred_id;
  }

  if ($params{postinsert_function}) {
    unless ($self->{table_postinsert_functions}->{$table_name}) {
      $self->{table_postinsert_functions}->{$table_name} = [];
    }
    push(@{$self->{table_postinsert_functions}->{$table_name}}, 
         {
          batch_index => $#{$self->{table_batch}->{$table_name}},
          postinsert_function => $params{postinsert_function},
         });
  }

  if (scalar @{$self->{table_batch}->{$table_name}} >= $self->{batch_size}) {
    $self->insert_table_batch(table_name => $table_name);
  }
}

#----------------------------------------------------------------------

sub flush {
  my($self) = @_;
  for my $table_name (@{$self->{table_batch_insert_order}}) {
      $self->insert_table_batch(table_name => $table_name);
  }

  $self->{table_batch_insert_order} = [];
}

#----------------------------------------------------------------------

sub insert_table_batch {
    my($self, %params) = @_;

    my $table_name = $params{table_name};
    my $data = $self->{table_batch}->{$table_name};

    if ($data && @$data) {
        my $dbh = $self->{dbh} || F5::Db::Single->instance;

        if ($self->{table_prerequisites}->{$table_name}) {
            for my $prerequisite_table_name (keys %{$self->{table_prerequisites}->{$table_name}}) {
                $self->insert_table_batch(table_name => $prerequisite_table_name);
            }
        }

        # all fields used anywhere in the dataset
        my %dataset_fields;
        for my $row (@$data) {
            for my $k (keys %$row) {
                $dataset_fields{$k} = 1;
            }
        }

        # all db fields
        my $db_fields_href = F5::DbUtils::get_db_fields($dbh, $table_name, undef, 1);
        my @db_fields = keys %$db_fields_href;
        if ($self->{table_opts}->{$table_name}->{fields}) {
            my %db_fields = map { ($_ => 1) } @db_fields;
            @db_fields = grep { $db_fields{$_} } @{$self->{table_opts}->{$table_name}->{fields}};
        }

        # throw away irrelevant db fields
        my @fields = grep { $dataset_fields{$_} } @db_fields;

        local $" = ',';

        my $insert_type = "INSERT";
        if ($self->{table_opts}->{$table_name} && $self->{table_opts}->{$table_name}->{do_insert_ignore}) {
            $insert_type .= " " . F5::DbUtils::get_ignore_string($dbh);
        }
        elsif ($self->{table_opts}->{$table_name} && $self->{table_opts}->{$table_name}->{do_replace}) {
            $insert_type = "REPLACE";
        }
        my @set_date_fields_to_now = ();
        if ($self->{table_opts}->{$table_name} && $self->{table_opts}->{$table_name}->{set_date_fields_to_now}) {
            @set_date_fields_to_now = @{$self->{table_opts}->{$table_name}->{set_date_fields_to_now}};
        }

        my @set_date_fields_to_now_unix = ();
        if ($self->{table_opts}->{$table_name} && $self->{table_opts}->{$table_name}->{set_date_fields_to_now_unix}) {
            @set_date_fields_to_now_unix = @{$self->{table_opts}->{$table_name}->{set_date_fields_to_now_unix}};
        }

        my @all_placeholders = ();
        my @all_values = ();
        my @all_fields = (@fields, @set_date_fields_to_now, @set_date_fields_to_now_unix);

        for my $row (@$data) {
            my @row_placeholders = ();
            my @row_values = ();

            for my $field (@fields) {
                my $value = exists($row->{$field}) ? $row->{$field} : undef;
                # Resolve deferred ids, if needed
                if (exists $row->{$field} && ref($row->{$field}) eq 'F5::BatchInsert::DeferredID') {
                    $value = $row->{$field}->get_value;
                }

                if ( !defined($value) ) {
                    $value = $db_fields_href->{$field}->{default};
                }

                if ( !defined($value) )
                {
                    push (@row_placeholders, 'NULL');
                }
                else {
                    push (@row_placeholders, '?');
                    push (@row_values, $value);
                }
            }
            my $now_string = F5::DbUtils::get_now_string(dbh => $dbh);
            my $now_unix_string = F5::DbUtils::get_unix_timestamp_string(dbh => $dbh);
            push (@row_placeholders,
                  ($now_string) x scalar @set_date_fields_to_now,
                  ($now_unix_string) x scalar @set_date_fields_to_now_unix,
                 );

            push (@all_values, @row_values);
            push (@all_placeholders, "(@row_placeholders)");
        }

        eval {
            $dbh->do("$insert_type INTO $table_name (@all_fields) VALUES @all_placeholders", undef, @all_values);
        };

        if ($@) {
            my $error_msg = "Failed to insert to $table_name ($@)";
            $self->{logger}->error($error_msg) if $self->{logger};
            die "$error_msg\n";
        }

        if ($self->{table_deferred_ids}->{$table_name}) {
            my($initial_insert_id) = $dbh->last_insert_id(undef, undef, undef, undef);

            if (! $initial_insert_id) {
                # Fake an insert id for tables with non-autoincrement foreign keys (such as LRN_REQUESTS)
                $initial_insert_id = 1;
            }
            elsif (F5::DbUtils::is_sqlite($dbh)) {
                # SQLite returns the LAST id inserted for a multi-row, while MySQL returns the first
                $initial_insert_id -= $#{$self->{table_deferred_ids}->{$table_name}};
            }

            for my $deferred_id (@{$self->{table_deferred_ids}->{$table_name}}) {
                $deferred_id->{deferred_id_obj}->set_value($initial_insert_id + $deferred_id->{batch_index});
            }
            delete $self->{table_deferred_ids}->{$table_name};
        }

        if ($self->{table_postinsert_functions}->{$table_name}) {
            for my $postinsert_function (@{$self->{table_postinsert_functions}->{$table_name}}) {
                $postinsert_function->{postinsert_function}->();
            }
            delete $self->{table_postinsert_functions}->{$table_name};
        }
    }

    delete $self->{table_batch}->{$table_name};
    delete $self->{table_batch_oldest_record}->{$table_name};
    delete $self->{table_prerequisites}->{$table_name};

    if ($data && @$data && $self->{db_handler}) {
        $self->{db_handler}->inserted_rows(row_count => scalar @$data,
                                           table_name => $table_name,
                                          );
    }
}

#----------------------------------------------------------------------

sub set_dbh {
  my ($self, $dbh) = @_;
  $self->{dbh} = $dbh;
  if ($self->{db_handler}) {
      $self->{db_handler}->set_dbh($dbh);
  }
}

#----------------------------------------------------------------------

sub has_waiting_records {
  my($self) = @_;

  for my $table_name (keys %{$self->{table_batch}}) {
    if (@{$self->{table_batch}->{$table_name}}) {
      return 1;
    }
  }
  return 0;
}

#----------------------------------------------------------------------

sub flush_old_records {
  my($self, %params) = @_;
  my $checktime = time - $params{interval};
  my $num_tables_inserted = 0;
  for my $table_name (keys %{$self->{table_batch_oldest_record}}) {
    if ($self->{table_batch_oldest_record}->{$table_name} < $checktime) {
      $self->insert_table_batch(table_name => $table_name);
      $num_tables_inserted++;
    }
  }
  return $num_tables_inserted;
}

#----------------------------------------------------------------------

sub oldest_record_time {
    my($self) = @_;
    return List::Util::min(values %{$self->{table_batch_oldest_record}});
}

#----------------------------------------------------------------------

package F5::BatchInsert::DeferredID;

sub new {
  my($class, %params) = @_;
  my $self = bless { %params }, $class;

  return $self;
}

#----------------------------------------------------------------------

sub set_value {
  my($self, $val) = @_;
  $self->{value} = $val;
}

#----------------------------------------------------------------------

sub get_value {
  my($self) = @_;

  unless ($self->{value}) {
    $self->{batch_insert}->insert_table_batch(table_name => $self->{table_name});

    # If value hasn't populated by now, we have a real problem
    unless ($self->{value}) {
      die "Deferred ID from table $self->{table_name} wasn't populated";
    }
  }
  return $self->{value};
}

#----------------------------------------------------------------------

1;
