package F5::BdAgent;

use strict;
use warnings;

use Math::BigInt;
use POSIX qw(mkfifo);
use Bit::Vector;
use Time::HiRes qw( ualarm gettimeofday tv_interval );
use List::Util;

use F5::Deploy;
use F5::ASM::COAPI;
use F5::IO::Mtcl;
use F5::IO::Storable;
use F5::IO::Selector;
use F5::BigIpLog;
use F5::DbUtils;
use F5::IpAddr;
use F5::Db::Single;
use F5::ASMReady;
use F5::DateTimeUtils;
use F5::Logger;
use F5::Log::Job;

use constant {
    AGENT_MSG_CONFIG_END => 11,
    EV_TYPE_CONTENT => 1,

    MSG_QUEUE_MAX_SIZE => 500,
    MSG_QUEUE_HANDLE_PERIOD => 1.0, # floating seconds
    MSG_QUEUE_HANDLE_TIMEOUT => 500000, # microseconds

    DEBUG => 0,
    KEEP_BINARY_FILES => 0,
};

my $PULSE_MSG_INTERVAL = 30;
my $UPDATE_REQUEST_MAX_POLL_INTERVAL = 2;
my $UPDATE_REQUEST_CLEANING_AGE = 60*60; # 1 hour

my $FROM_BD_PIPE_NAME = F5::Deploy::map_path('/var/ts/bd/ts_agent_pipe');

my @core_msgs = qw(
                      CORE_MSG_ACK
                      CORE_MSG_ACK_MORE
                      CORE_MSG_ERROR
                      CORE_MSG_EVENT
                      CORE_MSG_LOG
                      CORE_MSG_INIT_READY
                 );
my %core_msgs = map { ($core_msgs[$_] => $_) } 0..$#core_msgs;

# event to handler method name (see 'BD event handler methods section' below)
my %EVENT_TYPE_TO_HANDLER = (
    'EV_TYPE_CORE_SYSTEM' => 'core_system_event_handler',
    'EV_TYPE_CONTENT' => 'core_content_event_handler',

    'EV_TYPE_CMD' => 'mng_cmd_event_handler',
    'EV_TYPE_INFO' => 'mng_info_event_handler',
    'EV_TYPE_SYSLOG' => 'mng_syslog_event_handler',
);

my %SYSLOG_SEVERITY = (
    0 => 'EMERG',
    1 => 'ALERT',
    2 => 'CRIT',
    3 => 'ERR',
    4 => 'WARNING',
    5 => 'NOTICE',
    6 => 'INFO',
);

# like in 'typedef enum _CORE_EVENT_CODE' from TrafficShield/common/include/core_events.h
my %CORE_EVENT_CODE_TO_MACRO = ( # event names in comment for convenience
    28 => 'TS_ERROR_MSG', # EV_CORE_ERROR
    29 => 'TS_ERROR_OUT_OF_MEMEORY', # EV_CORE_ERROR_OUT_OF_MEMEORY
    30 => 'TS_ERROR_DISK_FAILURE', # EV_CORE_ERROR_DISK_FAILURE
    31 => 'TS_ERROR_CARD_FAILURE', # EV_CORE_ERROR_CARD_FAILURE
    32 => 'TS_ERROR_SYS_FAILURE', # EV_CORE_ERROR_SYS_FAILURE
    33 => 'TS_ERROR_NETWORK_FAILURE', # EV_CORE_ERROR_NETWORK_FAILURE
    34 => 'TS_ERROR_SSL', # EV_CORE_ERROR_SSL
    35 => 'TS_ERROR_FULL_QUEUE', # EV_CORE_ERROR_FULL_QUEUE
    36 => 'TS_MANAGEMENT_COMMUNICATION_ERROR', # EV_CORE_MANAGEMENT_COMMUNICATION_ERROR
    38 => 'TS_CONFIGURATION_ERROR', # EV_CORE_CONFIGURATION_ERROR
    41 => 'TS_BAD_REQUEST', # EV_CORE_BAD_REQUEST
    42 => 'TS_BAD_RESPONSE', # EV_CORE_BAD_RESPONSE
    51 => 'TS_BD_FILTERED_HTML_EXCEEDS_MAX_LENGTH', # EV_CORE_BD_FILTERED_HTML_EXCEEDS_MAX_LENGTH
    52 => 'TS_BD_APPLICATION_ERROR_FILTERING_APPLIED', # EV_CORE_BD_APPLICATION_ERROR_FILTERING_APPLIED
    53 => 'TS_CONFIGURATION_ERROR', # EV_CORE_ILLEGAL_REG_EXP
    54 => 'TS_ERROR_HIGH_CPU', # EV_CORE_HIGH_CPU_UTILIZATION
    55 => 'TS_ERROR_SESSION_DB_OUT_OF_BOUNDS', # EV_CORE_SESSION_DB_OUT_OF_BOUNDS
    56 => 'TS_INFO_MSG', # EV_CORE_MISSING_ACCOUNT_CONFIG,
);

# like in 'typedef enum _CORE_EVENT_CODE' from TrafficShield/common/include/core_events.h
my %CORE_EVENT_CODE_TO_METHOD = ( # event names in comment for convenience
    56 => 'missing_account_config', # EV_CORE_MISSING_ACCOUNT_CONFIG
);

# like in 'typedef enum _EVENT_CODE' from TrafficShield/common/include/mng_events.h
my %MNG_EVENT_CODE_TO_MACRO = ( # event names in comment for convenience
    1001  => 'TS_ERROR_UNIT_STATE_CHANGED', # EV_UNIT_STATE_CHANGED,
    1003  => 'TS_INFO_MSG', # EV_UNIT_STARTED,
    1004  => 'TS_ERROR_UNIT_SHUTDOWN', # EV_UNIT_SHUTDOWN,
    1006  => 'TS_ERROR_WEBSERVER_FAILURE', # EV_WEBSERVER_FAILURE,
    1008  => 'TS_ERROR_BD_FAILURE', # EV_BD_FAILURE,
    1014  => 'TS_INFO_MSG', # EV_PROCESS_FAILURE,
    1030  => 'TS_ERROR_LEARNING_FAILED', # EV_LEARNING_REQUEST_FAILED,
);

# like in 'typedef enum _DCC_MSG' from TrafficShield/common/include/dcc_msg.h
my %DCC_EVENT_CODE_TO_MACRO = ( # event names in comment for convenience
    2 => 'TS_ERROR_UPDATE_CONFIG', # DCC_MSG_UPDATE_CONFIG
);

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->init_db(%params);

    $self->{f5_big_ip_log} = F5::BigIpLog->new;
    $self->{logger} = $params{logger};
    $self->{selector} = F5::IO::Selector->new;
    $self->init_core_channel(%params);
    $self->{startup_complete} = 1;
}

#----------------------------------------------------------------------

sub init_db {
    my($self, %params) = @_;

    $self->{dbh} = $params{dbh};
    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    my $ary_ref = $dbh->selectcol_arrayref("SELECT viol_index, description FROM PLC.VIOLATIONS", { Columns=>[1, 2] });
    $self->{sub_viol_descriptions} = {@{$ary_ref}};

    $ary_ref = $dbh->selectcol_arrayref("SELECT http_check_code, description FROM PLC.HTTP_CHECK_CODES", { Columns=>[1, 2] });
    $self->{sub_viol_http_descriptions} = {@{$ary_ref}};

    $ary_ref = $dbh->selectcol_arrayref("SELECT field_name, current_value FROM DCC.INTERNAL_CONFIG_ITEMS", { Columns=>[1, 2] });
    $self->{internal_config_items} = {@{$ary_ref}};

    $ary_ref = $dbh->selectcol_arrayref("SELECT subviolation_code, subviolation_description FROM PLC.XML_WS_SUBVIOLATIONS_CODES", { Columns=>[1, 2] });
    $self->{sub_viol_wss_descriptions} = {@{$ary_ref}};

    $ary_ref = $dbh->selectcol_arrayref("SELECT normalization_code, description FROM PLC.NORMALIZATION_CODES", { Columns=>[1, 2] });
    $self->{sub_viol_norm_descriptions} = {@{$ary_ref}};

    # When ASM starts (bd_agent) then we set the end_time to the current time
    $self->update_row(table_name => 'DCC.NEW_EVENT_SYSTEM',
                      rec => { },
                      set_date_fields_to_now_unix => [ 'l_time' ],
                      where_fields => { l_time => 0 }
                     );
}

#----------------------------------------------------------------------

sub init_core_channel {
    my($self, %params) = @_;

    $self->{msg_queue} = [];

    mkfifo($FROM_BD_PIPE_NAME, 0777);

    open(my $from_pipe_fh, '<', $FROM_BD_PIPE_NAME);
    $self->{from_pipe} = F5::IO::Mtcl->new(socket => $from_pipe_fh,
                                           logger => $self->{logger},
                                           read_message_handler => sub {
                                               $self->handle_bd_pipe_message(message => $_[1]);
                                           },
                                           selector => $self->{selector},
                                          );


    $self->{is_core_ready} = 0;
    $self->{startup_complete} = 0;

    # these counters get reset only on init
    $self->{bd_other_message_count} = 0;
}

#----------------------------------------------------------------------

sub start {
    my($self, %params) = @_;

    $self->{logger}->info('Starting bd_agent') if $self->{logger};

    my $last_pulse_time = time;

    while (1) {

        $self->handle_messages(last_pulse_time => $last_pulse_time);

        if ($PULSE_MSG_INTERVAL &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PULSE_MSG_INTERVAL,
                                                               start_time => $last_pulse_time) == 0) {

            $self->log_pulse_message;
            $last_pulse_time = time;
        }
    }
}

#----------------------------------------------------------------------

sub handle_messages {
    my($self, %params) = @_;
    my $poll_time = $self->get_poll_time(last_pulse_time => $params{last_pulse_time});
    $self->{logger}->trace("About to poll " . (defined $poll_time ? "for $poll_time seconds" : "until message received")) if $self->{logger};
    my @readers = $self->{selector}->can_read($poll_time);
    for my $reader (@readers) {
        $reader->handle_read;
    }
}

#----------------------------------------------------------------------

sub get_poll_time {
    my($self, %params) = @_;
    my @potential_poll_times;
    my $last_pulse_time = $params{last_pulse_time};

    if ($PULSE_MSG_INTERVAL) {
        push(@potential_poll_times,
             F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $PULSE_MSG_INTERVAL,
                                                                start_time => $last_pulse_time));
    }

    return (@potential_poll_times) ? List::Util::min(@potential_poll_times) : undef;
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;

    $self->{logger}->info("$self->{bd_other_message_count} BD messages handled since startup") if $self->{logger};
}

#----------------------------------------------------------------------

sub handle_bd_pipe_message {
    my($self, %params) = @_;

    my $message = $params{message};

    if ( DEBUG && $self->{logger} && $self->{logger}->is_trace()) {
        $self->{logger}->trace($message);
    }

    if (! defined $message->{msg_code}) {
        return;
    }

    if ($message->{msg_code} == $core_msgs{CORE_MSG_INIT_READY}) {
        $self->{is_core_ready} = 1;
        $self->{logger}->info("bd is ready") if $self->{logger};
    }
    else {
        # this happens first time in only
        $self->{start_time} = [gettimeofday()] unless $self->{start_time};

        $self->{bd_other_message_count}++;

        my $event_code = $message->get_event_code();
        $self->{logger}->debug("Got event code:[$event_code]") if $self->{logger};

        if ( $event_code == EV_TYPE_CONTENT ) {
            # oldest first
            push(@{$self->{msg_queue}},$message);

            # if queue reached MSG_QUEUE_MAX_SIZE -
            #    throw away 1 oldest messages of the head of the queue
            if ( scalar @{$self->{msg_queue}} > MSG_QUEUE_MAX_SIZE ) {
                 shift(@{$self->{msg_queue}});
            }

            if ( tv_interval($self->{start_time}) >= MSG_QUEUE_HANDLE_PERIOD ) {
                eval {
                    local $SIG{ALRM} = sub { die; };
                    ualarm(MSG_QUEUE_HANDLE_TIMEOUT);

                    # oldest handled first
                    while (my $queued_message = shift(@{$self->{msg_queue}}) ) {
                        $self->handle_message(event_code => $event_code,
                                              message => $queued_message,
                                             );
                    }

                    ualarm(0);
                };
                # reset last time stamp once - immediately after the eval above
                $self->{start_time} = [gettimeofday()];
            }
        }
        else {
            $self->handle_message(event_code => $event_code,
                                  message => $message,
                                 );
        }
    }
}

#----------------------------------------------------------------------

sub handle_message {
    my($self, %params) = @_;
    my $event_code = $params{event_code} ;
    my $message = $params{message} ;

    my $event;
    eval {
        $event = $message->get_event();
    };
    if ($@) {
        # every time the ualarm above fires, for EV_TYPE_CONTENT messages, we'll get here
        #     thus, we lower log level to debug for EV_TYPE_CONTENT messages
        my $method = ( $event_code == EV_TYPE_CONTENT ) ? 'debug' : 'error' ;
        $self->{logger}->$method("Event parsing failed for event code:[$event_code]") if $self->{logger};
        if ( DEBUG ) {
            $self->{logger}->trace($message) if $self->{logger};
        }
    }
    else {
        my $handler_method = $EVENT_TYPE_TO_HANDLER{$event->{bd_event}};
        $self->{logger}->debug("Event parsing succeeded for event code:[$event_code]. Chosen method is:[$handler_method].") if $self->{logger};

        if ( DEBUG ) {
            $self->dump_event( handler => $handler_method, event => $event);
        }

        # this will fire one of the BD event handler methods below
        $self->$handler_method($event->{$event->{bd_event}}, $event->{COMMON_EVENT_TYPE});
    }
}

#----------------------------------------------------------------------

sub get_debug_message {
    my($m) = @_;
    # unhandled messages - events go here
    my $message_copy = { %$m };
    if (length($message_copy->{data}) > 500) {
        $message_copy->{data_suffix} = substr($message_copy->{data}, -200);
        $message_copy->{data_prefix} = substr($message_copy->{data}, 0, 200);
        delete $message_copy->{data};
    }
    return $message_copy;
}

#----------------------------------------------------------------------

###########################################################
#    Start BD event handler methods section
###########################################################

sub missing_account_config {
    my($self, $event) = @_;

    my $account_name = $event->{desc};
    return unless $account_name;

    my $mcp_asm_policies = F5::ASM::COAPI::Query->new('asm_policy',{name => $account_name});
    return unless $mcp_asm_policies && scalar @$mcp_asm_policies;

    my $mcp_asm_policy = $mcp_asm_policies->[0];
    return unless exists $mcp_asm_policy->{is_active_policy};
    return if $mcp_asm_policy->{is_active_policy};

    $mcp_asm_policy->{is_active_policy} = 1;
    $mcp_asm_policy->save();

    return "bd_agent is asynchronously activating ASM security policy named '$account_name', per request from bd";
}

#----------------------------------------------------------------------

sub core_system_event_handler {
    my($self, $event, $common) = @_;

    my $desc = $event->{desc};

    my $method = $CORE_EVENT_CODE_TO_METHOD{$event->{code}} || 0;
    if ($method) {
        $desc = $self->$method($event);
    }

    if ($desc) {
        my $macro = $CORE_EVENT_CODE_TO_MACRO{$event->{code}} || 0;
        $self->big_ip_log($macro, $event->{code}, $desc);
    }
}

#----------------------------------------------------------------------

sub core_content_event_handler {
    my($self, $event, $common) = @_;

    my $severity = $SYSLOG_SEVERITY{$event->{severity_id}};
    my $status = ($event->{is_blocked}) ? 'blocked' : 'violation' ;
    my $asm_prefix = ($severity eq 'INFO') ? 'asm_' : '' ;
    my $macro = 'ts_' . $asm_prefix . 'request_' . $status . '_' . $severity;

    my $rejected_desc = $self->get_viol_desc($event->{viol_set_enforce}->as_bin());

    my $http_desc = $self->get_sub_viol_desc(sub_viol_value => $event->{non_rfc_sub_violations},
                                      db_map => $self->{sub_viol_http_descriptions},
                                     );

    my $wss_desc = $self->get_sub_viol_desc(sub_viol_value => $event->{wss_sub_violations},
                                     db_map => $self->{sub_viol_wss_descriptions},
                                    );

    my $et_desc = $self->get_sub_viol_desc(sub_viol_value => $event->{normalization_sub_violations},
                                    db_map => $self->{sub_viol_norm_descriptions},
                                   );

    my $str_app_id = ($event->{app_id}) ? 'HTTPS' : 'HTTP' ;


    my $src_ip = $self->ip_ntop($event->{source_ip_raw});
    my $dest_ip = $self->ip_ntop($event->{destination_ip_raw});
    my $xff_ip = $self->ip_ntop($event->{xff_ip_raw});

    # session_id shold be a lower-case hex string w/o the leading '0x'
    my $session_id = $event->{session_id}->as_hex(); # a signed hexadecimal string with prefixed 0x
    $session_id = lc(substr($session_id,2)); # lower-cased and '0x' stripped

    $self->big_ip_log($macro, undef,
        $rejected_desc,
        $http_desc,
        $wss_desc,
        $et_desc,
        $event->{virus_name} || "N/A",
        $event->{support_id}->bstr(),
        $src_ip,
        $xff_ip,
        $event->{client_remote_port},
        $dest_ip,
        $event->{client_local_port},
        $event->{rtdom_id},
        $event->{http_selector},
        $str_app_id,
        $event->{geo_location},
        $event->{first_byte},
        $event->{username} || "N/A",
        $session_id,
        $event->{violation_rate}
    );

    if ( DEBUG ) {
        $macro = 'ts_info_msg';
        $self->big_ip_log($macro, undef,
           'counter:['. $common->{counter} .'], '.
           'data_size:['.  $common->{data_size} .'], '.
           'event_id:['.  $common->{event_id}->bstr() .'], '.
           'event_type:['.  $common->{event_type} .'], '.
           'first_time:['.  $common->{first_time} .'], '.
           'last_time:['.  $common->{last_time} .'], '.
           'machine_id:['.  sprintf('%02X:%02X:%02X:%02X:%02X:%02X', @{$common->{machine_id}}) .'], '.
           'site_id:['.  $common->{site_id} .']'
        );
    }

}

#----------------------------------------------------------------------

sub mng_cmd_event_handler {
    my($self, $event, $common) = @_;

    my $rec = {
        f_time => $common->{first_time},
        l_time => $common->{last_time},
        counter => $common->{counter},
        mac_ID => sprintf('%02X:%02X:%02X:%02X:%02X:%02X', @{$common->{machine_id}}),
        event_crc => $common->{event_id}->bstr(),
        type => $common->{event_type},
        code => $event->{cmd_code},
        result_description => $event->{result_description},
    };

    $self->insert_row(table_name => 'DCC.NEW_EVENT_SYSTEM',
                      rec => $rec,
                     );

    my $macro = $DCC_EVENT_CODE_TO_MACRO{$event->{cmd_code}} || 0;
    $self->big_ip_log($macro, $event->{cmd_code}, $event->{result_description});
}

#----------------------------------------------------------------------

sub mng_info_event_handler {
    my($self, $event, $common) = @_;

    my $rec = {
        f_time => $common->{first_time},
        l_time => $common->{last_time},
        counter => $common->{counter},
        mac_ID => sprintf('%02X:%02X:%02X:%02X:%02X:%02X', @{$common->{machine_id}}),
        event_crc => $common->{event_id}->bstr(),
        type => $common->{event_type},
        code => $event->{code},
        state => $event->{state},
        result_description => $event->{desc},
    };

    $self->insert_row(table_name => 'DCC.NEW_EVENT_SYSTEM',
                      rec => $rec,
                     );

    my $macro = $MNG_EVENT_CODE_TO_MACRO{$event->{code}} || 0;
    $self->big_ip_log($macro, $event->{code}, $event->{desc});
}

#----------------------------------------------------------------------

sub mng_syslog_event_handler {
    my($self, $event, $common) = @_;

    my $rec = {
        f_time => $common->{first_time},
        l_time => $common->{last_time},
        counter => $common->{counter},
        mac_ID => sprintf('%02X:%02X:%02X:%02X:%02X:%02X', @{$common->{machine_id}}),
        event_crc => $common->{event_id}->bstr(),
        type => $common->{event_type},
        result_description => $event->{result_description},
    };

    $self->insert_row(table_name => 'DCC.NEW_EVENT_SYSTEM',
                      rec => $rec,
                     );
}


###########################################################
#    End BD event handler methods section
###########################################################

#----------------------------------------------------------------------

# MACRO is from TrafficShield/common/ts_bigip_log/bigip_ts_error_defs.in
# that MACRO, in lower case, is also an OO method in F5::BigIpLog
# so you can F5::BigIpLog->$method(ARGS) when $method = lc(MACRO);
# ARGS is an input per each '%' from the chosen MACRO - which is a C style printf modifier
# you can also view the prototype of $method here:
#     TrafficShield/mng/scripts/packages/xs_packages/BigIpLog/BigIpLog.xs
# there look for "bigip_ts_${method}_log".
# Note that BigIpLog.xs is auto-generated (you need to build tscommon and mng-scripts to get it)

sub big_ip_log {
    my ($self, $macro, $evt_code, @msg) = @_;
    $evt_code ||= 'none';

    if ( ! $macro ) {
        my $whowasi = (caller(1))[3];
        $self->{logger}->debug("Received unknown event number:[$evt_code] in method:[$whowasi].") if $self->{logger};
        return;
    }

    $macro = lc($macro);

    if ($self->{logger} && DEBUG ) {
        my $whowasi = (caller(1))[3];
        $self->{logger}->debug("Call details: macro:[$macro] caller:[$whowasi] event number:[$evt_code].");
        $self->{logger}->trace("ARGS to macro:[$macro] were:");
        $self->{logger}->trace(\@msg);
    }

    eval {
        $self->{f5_big_ip_log}->$macro(@msg);
    };
    if ($@) {
        $self->{logger}->debug('Error executing F5::BigIpLog->' . $macro . "(...) : [$@]") if $self->{logger};
    }
}

#----------------------------------------------------------------------
# called only when DEBUG is set to '1'

sub dump_event {
    my($self, %params) = @_;

    my $event = $params{event};
    my $handler = $params{handler};

    $self->{logger}->trace("The target handler is [$handler]. This is the event that I received:\n") if $self->{logger};
    $self->{logger}->trace($event) if $self->{logger};
}

#----------------------------------------------------------------------

sub get_sub_viol_desc {
    my($self, %params) = @_;

    my $sub_viol_value = $params{sub_viol_value};
    return "N/A" unless $sub_viol_value;

    my $db_map = $params{db_map};
    my $viol_vect = Bit::Vector->new_Dec(32, $sub_viol_value);

    my $desc = '';
    for my $bit ( $viol_vect->Index_List_Read() ) {
        $desc .= $db_map->{2 ** $bit} . ',';
    }
    chop($desc);

    return $desc || "N/A";
}

#----------------------------------------------------------------------

sub ip_ntop {
    my ($self, $ref) = @_;

    my $res = F5::IpAddr::ip_ntop(pack("L4", $ref->[0], $ref->[1], $ref->[2], $ref->[3]));

    return ( $res eq '::' ) ? "N/A" : $res ;
}

#----------------------------------------------------------------------

sub get_ip_binary_data {
    my ($self, $ref) = @_;

    return pack("L4", $ref->[0], $ref->[1], $ref->[2], $ref->[3]);
}

#----------------------------------------------------------------------

# tbd:sig
#     change bd_agent.h so that C::Include would unpack into a representation (char?)
#     that can be later used in F5::Violation::Utils::get_indexes_from_viol_set

sub get_viol_desc {
    my ($self, $bits) = @_;

    return "N/A" unless $bits; # will return on a bitstring that is all 0(s)

    my $rejected_desc;

    # $bits has a leading left '0b' notation, we remove it
    #    and left-pad it with '0's to have an even number of octets
    substr($bits, 0, 2, 0 x (8 - ((length($bits) - 2) % 8)));
    my $bits_len = length($bits);

    # traverse octets
    # each octet by itself is reversed - we reverse them back
    # the input is a C::Include->unpack product, not sure why the octets are reversed
    for (my $i=0 ; $i<$bits_len ; $i+=8) {
        my $octet = substr($bits, $i, 8);
        next if $octet eq '00000000'; # skip octet if no bits ON
        substr($bits, $i, 8, scalar reverse $octet);
    }

    # now, an index of each bit that it turned ON is a viol_index
    # Index_List_Read returns a list of indexes of bits that are turned ON in this octet (0-7)
    my $vec = Bit::Vector->new_Bin($bits_len, $bits);
    for my $bit ( $vec->Index_List_Read() ) {
        $rejected_desc .= $self->{sub_viol_descriptions}->{$bit} . ',';
    }

    chop($rejected_desc);
    return $rejected_desc || "N/A";
}

#----------------------------------------------------------------------

sub insert_row {
    my ($self, %params) = @_;
    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    F5::DbUtils::insert_row(dbh => $dbh, %params);
}

#----------------------------------------------------------------------

sub update_row {
    my ($self, %params) = @_;
    my $dbh = $self->{dbh} || F5::Db::Single->instance;

    F5::DbUtils::update_row(dbh => $dbh, %params);
}

#----------------------------------------------------------------------

1;
