package F5::BdCfg::Bundle;

use strict;
use warnings;

use Getopt::Long;
use JSON;
use File::Find;
use File::Spec;
use File::Path qw(make_path rmtree);
use Cwd qw(getcwd);
use YAML::XS;

use F5::ASMConfig::GuaranteedClient;
use F5::Deploy;
use F5::Utils::File;
use F5::PrepareFiles;
use F5::ConfigSet;

my $BD_CONFIG_DIR          = F5::Deploy::get_path_by_type("bd_config");
my $BD_CONFIG_TABLES_DIR   = F5::Deploy::get_path_by_type("bd_config_tables");
my $DATABASE_BASE_DIR      = F5::Deploy::get_path_by_type("database");
my $POLICY_BUNDLE_BASE_DIR = F5::Deploy::get_path_by_type("var") . "/policy_bundles";
my $OPT                    = F5::Deploy::get_path_by_type("top_level");
my $GLOBAL_FILE            = "$POLICY_BUNDLE_BASE_DIR/global";
my $CATALOG_FILE           = "catalog.yaml";
my $VERSION                = File::Slurp::read_file("$OPT/VERSION.common");
my $call_flags             = F5::ASMConfig::CallFlags->new( skip_set_active => 1, suppress_logging => 1, skip_set_sigs_active => 1 );

#----------------------------------------------------------------------

sub new {
    my ( $class, %params ) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my ( $self, %params ) = @_;

    $self->{workdir}    = getcwd;
    $self->{config_set} = F5::ConfigSet->new();

    # Backup initial databases
    F5::DbUtils::copy_sqlite_databases( suffix => 'initial', overwrite_existing => 0 );

}

#----------------------------------------------------------------------

sub get_client {
    my ($self) = @_;
    if ( $self->{client} ) {
        return $self->{client};
    }
    require F5::ASMConfig::GuaranteedClient;
    $self->{client} = F5::ASMConfig::GuaranteedClient::get_client( username => 'import_export_policy', dbh => $self->{dbh} );
    return $self->{client};
}

#----------------------------------------------------------------------

sub load_global_state {
    my ( $self, %params ) = @_;
    my $input_filename = File::Spec->rel2abs( $params{filename} );

    die { "error_message" => "Missing global settings/state file - '$input_filename'" } unless ( -f $input_filename );

    my $is_json = F5::Utils::File::is_file_json( filename => $input_filename );
    my $is_tgz  = F5::Utils::File::check_filetype( filename => $input_filename, valid_types => ['gzip'] )->{completed_successfully};

    if ($is_json) {
        $self->reset_global( skip_set_global_info => 1 );
        require F5::GlobalSettings;
        $self->{global_settings} = F5::GlobalSettings->new(
            client   => $self->get_client(),
            filename => $input_filename
        );
        if ( defined $self->{global_settings}->{load_from_file_successfully} && !$self->{global_settings}->{load_from_file_successfully} ) {
            die { "error_message" => "Failed to load WAF settings file - " . $self->{global_settings}->{load_from_file_failure_msg} };
        }
        my $apply_rv = $self->{global_settings}->apply( call_flags => $call_flags );

        if ( defined $apply_rv->{completed_successfully} && !$apply_rv->{completed_successfully} ) {
            die { "error_message" => "Failed to apply WAF settings - " . $apply_rv->{failure_msg} };
        }
        $self->set_global_state_info(
            user_defined_signatures => $apply_rv->{user_defined_signatures},
            cookie_seed             => $apply_rv->{configured_settings},
        );

        return { completed_successfully => 1 };
    }
    elsif ($is_tgz) {
        my $unzip_dir = "$POLICY_BUNDLE_BASE_DIR/load_global.$$";
        make_path($unzip_dir);
        F5::SystemUtils::run_system_tool( command_line_params => [ 'tar', '-zpx', '-f', $input_filename, '-C', $unzip_dir, ] );
        find(
            {
                wanted => sub {
                    if ( -f $File::Find::name ) {
                        my $actual_file_location = $File::Find::name;
                        if ( $_ eq "global" ) {
                            $actual_file_location = $GLOBAL_FILE;
                        }
                        else {
                            $actual_file_location =~ s/$unzip_dir/$OPT/;
                        }
                        F5::Utils::File::copy_files_recursively( from_dir => $File::Find::name, to_dir => $actual_file_location );
                    }
                },
            },
            $unzip_dir
        );
        rmtree($unzip_dir);

        return { completed_successfully => 1 };
    }
    else {
        die { "error_message" => "Invalid global file format" };
    }
}

#----------------------------------------------------------------------

sub create_policy_bundle {
    my ( $self, %params ) = @_;
    my $bundle_policy_name;
    my $policy_name;
    my $policies_to_bundle = $self->{config_set}->get_default_policy_files();
    my $policy_filename    = File::Spec->rel2abs( $params{policy_filename} );
    my $output_filename    = File::Spec->rel2abs( $params{output_file} );

    die { "error_message" => "Missing policy file - '$policy_filename'" } unless ( -f $policy_filename );
    $self->set_global_state_info();

    if ( exists $policies_to_bundle->{$policy_filename} ) {
        $bundle_policy_name = $policies_to_bundle->{$policy_filename};
    }
    else {
        $bundle_policy_name = $policies_to_bundle->{$policy_filename} = $self->extract_policy_name( filename => $policy_filename );
    }

    my %catalog_policies = (
        bundlePolicyName => $bundle_policy_name,
        policies         => {},
    );

    for my $policy_filename ( keys %$policies_to_bundle ) {
        my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(
            filename              => $policy_filename,
            force_policy_name_flg => 1,
            policy_name           => $policies_to_bundle->{$policy_filename}
        );
        my $import_rv;
        my $policy_name_crc;
        my $policy_file_md5;

        eval {
            $import_rv       = $self->get_client->import_policy( $self->get_client->{session}, $call_flags, $import_export_properties );
            $policy_name     = $import_rv->{policy_name};
            $policy_name_crc = $import_rv->{policy}->{policy_name_crc};
            $policy_file_md5 = F5::Utils::File::get_file_md5($policy_filename);
            if ( $import_rv && $import_rv->{msgs} ) {
                my @msgs = map { $_->{msg} } grep { $_->{level} eq 'warning' } @{ $import_rv->{msgs} };
                if (@msgs) {
                    $self->{warnings} = \@msgs;    # TODO: merge warnings from different policies
                }
            }
        };

        if ( !$import_rv || !$import_rv->{completed_successfully} ) {
            my $err = $@;
            if ( $import_rv && $import_rv->{msgs} ) {
                $err = join( "\n", map { $_->{msg} } @{ $import_rv->{msgs} } );
            }
            elsif ( ref $err ) {
                $err = $err->{display_message} || $err->{error_message};
            }
            die { "error_message" => "Failed to import policy from '$policy_filename - $err'" };
        }
        my $set_active_result = F5::SetActive::set_active(
            policy_id              => $import_rv->{policy_id},
            expanded_return_value  => 1,
            do_save_policy_history => 0,
            do_update_core         => 0,
            dbh                    => $self->{dbh},
            wait_for_enforcer      => 0,
            use_daemon             => 0,
            update_ramcache        => 0,
            asm_config_handler     => $self->get_client(),
            send_sync              => 0,
        );

        die { "error_message" => "Failed to compile policy - '$policy_name'" } unless $set_active_result->{completed_successfully};

        $catalog_policies{"policies"}{$policy_name} = {
            originalFilename => $policy_filename,
            policyNameCrc    => $policy_name_crc,
            fileMD5          => $policy_file_md5,
        };
        $catalog_policies{"policies"}{$policy_name}{"warnings"} = $self->{warnings} if $self->{warnings};
    }

    my $set_active_result = F5::SetActive::set_active(
        asm_config_handler    => $self->get_client(),
        dbh                   => $self->{dbh},
        do_update_core        => 1,
        expanded_return_value => 1,
        send_sync             => 0,
        wait_for_enforcer     => 0,
    );

    die { "error_message" => "Failed to prepare bundle files" } unless $set_active_result->{completed_successfully};

    my $latest_full_dir = F5::PrepareFiles::find_latest_full_dir();
    my $global_state    = YAML::XS::LoadFile($GLOBAL_FILE);
    my $global_hash     = F5::Utils::File::get_file_md5($GLOBAL_FILE);
    $global_state->{"global-state"}->{hash} = $global_hash;

    my $subbundle_root_tmp        = "$POLICY_BUNDLE_BASE_DIR/save_policy__$bundle_policy_name.$$";
    my $subbundle_global          = "$subbundle_root_tmp/subbundle/global/$global_hash";
    my $subbundle_data_protection = "$subbundle_global/files/data_protection/";
    my $gen_path                  = $latest_full_dir =~ s/$BD_CONFIG_DIR//r;
    make_path( $subbundle_global . $gen_path, $subbundle_data_protection );

    my $bd_cfg_manifest   = YAML::XS::LoadFile("$BD_CONFIG_TABLES_DIR/bd_cfg_manifest.yaml");
    my $latest_full_files = $bd_cfg_manifest->{generations}->[0]->{fullFiles};

    my $crc_map;
    for my $policy ( keys %{ $catalog_policies{"policies"} } ) {
        my $crc = $catalog_policies{"policies"}{$policy}{"policyNameCrc"};
        $crc_map->{$crc}->{subbundle_policy} = "$subbundle_root_tmp/subbundle/policies/$policy";
        make_path( $crc_map->{$crc}->{subbundle_policy} . $gen_path );
    }

    for my $file ( @{$latest_full_files} ) {
        my $file_policy_name_crc;
        if ( $file->{filename} =~ /__logprofile_\d+|vs_log_profile/ ) {
            next;
        }
        if ( $file->{filename} =~ /__account_(\d+)/ ) {
            $file_policy_name_crc = $1;
            if ( exists $crc_map->{$file_policy_name_crc} ) {
                File::Copy::cp( $file->{filename}, $crc_map->{$file_policy_name_crc}->{subbundle_policy} . $gen_path );
            }
            else {
                next;
            }
        }
        else {
            File::Copy::cp( $file->{filename}, $subbundle_global . $gen_path );
            undef $file_policy_name_crc;
        }
        if ( defined $file->{dependentFiles} ) {
            for my $dependency ( @{ $file->{dependentFiles} } ) {
                my $target_dir;
                if ( $dependency->{filename} =~ /data_protection/ ) {
                    $target_dir = $subbundle_data_protection;
                }
                elsif ( $dependency->{filename} =~ /bloom/ ) {
                    next;
                }
                elsif ( defined $file_policy_name_crc ) {
                    my $basedir = $dependency->{filename};
                    $basedir =~ s/$BD_CONFIG_DIR//;
                    $target_dir = $crc_map->{$file_policy_name_crc}->{subbundle_policy} . $basedir;
                }
                else {
                    # in case of new non-crc dependent files
                    $target_dir = "$subbundle_global/files/";
                }
                F5::Utils::File::copy_files_recursively( from_dir => $dependency->{filename}, to_dir => $target_dir );
            }
        }
    }

    my %catalog = ( %catalog_policies, %{$global_state} );
    $catalog{"timestamp"}       = F5::DateTimeUtils::convert_epoch_to_rfc3339( time() );
    $catalog{"softwareVersion"} = $VERSION;
    YAML::XS::DumpFile( "$subbundle_root_tmp/$CATALOG_FILE", \%catalog );

    my $rv = $self->archive_bundle(
        source_dir      => $subbundle_root_tmp,
        output_filename => $output_filename
    );

    rmtree($subbundle_root_tmp);
    return $rv;
}

#----------------------------------------------------------------------

sub create_logging_profile_bundle {
    my ( $self, %params ) = @_;
    my $logging_profile_filename = File::Spec->rel2abs( $params{logging_profile_filename} );
    my $output_filename          = File::Spec->rel2abs( $params{output_file} );

    die { "error_message" => "Missing logging profile file - '$logging_profile_filename'" } unless ( -f $logging_profile_filename );
    my $logging_profile;
    $logging_profile = F5::Utils::File::load_file(
        filename  => $logging_profile_filename,
        json_only => 1
    );
    my $logging_profile_name;
    if ( $logging_profile->{name} ) {
        $logging_profile_name = $logging_profile->{name};
    }
    else {
        require File::Basename;
        import File::Basename 'fileparse';
        $logging_profile_name = fileparse( $logging_profile_filename, '\..*' );
    }
    my $validate_chars = "-a-zA-Z0-9_";
    if ( $logging_profile_name =~ /([^$validate_chars])/ ) {
        die { "error_message" => "The value for logging profile name contains an illegal character (\"$1\")" };
    }
    my $logging_config = {
        logging => {
            filter  => $logging_profile->{filter}->{request_type},
            content => $logging_profile->{content},
        },
    };

    my %logging_profiles;
    $self->{config_set}->build_logging_config(
        policy           => { vs_info => { $logging_profile_name => $logging_config } },
        logging_profiles => \%logging_profiles
    );

    # take the first element, which is our log_profile config md5_hex
    my ($logging_profile_config_md5_hex) = ( keys %logging_profiles );

    require F5::CRC;
    my $logging_profile_name_crc = F5::CRC::get_crc32($logging_profile_config_md5_hex);
    my @logging_profiles_list    = values %logging_profiles;
    my $prepare_result           = $self->{config_set}->prepare_all_logging_profiles( logging_profiles => \@logging_profiles_list );
    my $set_active_result        = F5::SetActive::set_active(
        dbh                   => $self->{dbh},
        do_update_core        => 1,
        asm_config_handler    => $self->get_client(),
        expanded_return_value => 1,
        send_sync             => 0,
    );

    my $latest_full_dir = F5::PrepareFiles::find_latest_full_dir();
    my @full_files      = F5::PrepareFiles::get_full_files_by_dir(
        base_dir             => $latest_full_dir,
        log_profile_name_crc => $logging_profile_name_crc,
    );

    die { "error_message" => "Failed to compile logging profile - '$logging_profile_name'" }
      unless $set_active_result->{completed_successfully};

    my $logging_profile_label     = "$logging_profile_name" . "_$logging_profile_config_md5_hex";
    my $subbundle_root_tmp        = "$POLICY_BUNDLE_BASE_DIR/save_logging_profile__$logging_profile_label.$$";
    my $gen_path                  = $latest_full_dir =~ s/$BD_CONFIG_DIR//r;
    my $subbundle_logging_profile = "$subbundle_root_tmp/subbundle/logging_profiles/$logging_profile_label/$gen_path";
    make_path($subbundle_logging_profile);
    my $logging_profile_file_md5 = F5::Utils::File::get_file_md5($logging_profile_filename);

    my %catalog = (
        logging_profiles => {
            $logging_profile_label => {
                originalFilename => $logging_profile_filename,
                fileMD5          => $logging_profile_file_md5,
                logProfileCrc    => $logging_profile_name_crc,
            },
        },
    );

    $catalog{"softwareVersion"} = $VERSION;
    $catalog{"timestamp"}       = F5::DateTimeUtils::convert_epoch_to_rfc3339( time() );

    YAML::XS::DumpFile( "$subbundle_root_tmp/$CATALOG_FILE", \%catalog );

    for my $filename (@full_files) {
        File::Copy::cp( $filename, $subbundle_logging_profile ) if $filename =~ /logprofile_$logging_profile_name_crc/;
    }

    my $rv = $self->archive_bundle(
        source_dir      => $subbundle_root_tmp,
        output_filename => $output_filename
    );

    rmtree($subbundle_root_tmp);
    return $rv;
}

#----------------------------------------------------------------------

sub update_cookie_seed {
    my ( $self, %params ) = @_;
    require F5::GlobalSettings;
    my $seed            = $params{cookie_seed} || F5::GlobalSettings::generate_cookie_seed_string();
    my $global_settings = F5::GlobalSettings->new(
        cookie_protection__seed => $seed,
        client                  => $self->get_client()
    );
    $global_settings->apply( call_flags => $call_flags );
    return $seed;
}

#----------------------------------------------------------------------

sub set_global_state_info {
    my ( $self, %params ) = @_;

    my $is_global_import = 1 if %params;

    my $user_defined_signatures   = $params{user_defined_signatures} || ();
    my $attack_signatures_version = $self->{config_set}->get_update_version( update_type => "signatures", generate_version => 1 );
    $attack_signatures_version->{revisionDatetime} = delete $attack_signatures_version->{revision_datetime};
    $self->{attack_signatures_version}             = $attack_signatures_version;

    my $threat_campaigns_version = $self->{config_set}->get_update_version( update_type => "threat_campaigns", generate_version => 1 );
    if ( %{$threat_campaigns_version} ) {
        $threat_campaigns_version->{revisionDatetime} = delete $threat_campaigns_version->{revision_datetime};
    }
    $self->{threat_campaigns_version} = $threat_campaigns_version;

    my $global_state = {};
    my $global_state_cache_struct;

    if ( -e $GLOBAL_FILE && !$is_global_import ) {
        $global_state_cache_struct = YAML::XS::LoadFile($GLOBAL_FILE);
        my $global_state_cache = $global_state_cache_struct->{"global-state"};
        $global_state->{"cookie-protection"}       = $global_state_cache->{"cookie-protection"};
        $global_state->{"user-defined-signatures"} = $global_state_cache->{"user-defined-signatures"}
          if $global_state_cache->{"user-defined-signatures"};
    }
    else {
        make_path($POLICY_BUNDLE_BASE_DIR);
        my $cookie_seed = $params{cookie_seed} || $self->update_cookie_seed();
        $global_state->{"cookie-protection"}       = { "seed" => $cookie_seed };
        $global_state->{"user-defined-signatures"} = $user_defined_signatures if $user_defined_signatures;
    }

    $global_state->{"attack-signatures"} = $self->{attack_signatures_version} if %{ $self->{attack_signatures_version} };
    $global_state->{"threat-campaigns"}  = $self->{threat_campaigns_version}  if %{ $self->{threat_campaigns_version} };

    my $global_state_struct = { "global-state" => $global_state };

    unless ( F5::ConfigSet::are_hashes_equivalent( $global_state_struct, $global_state_cache_struct ) ) {
        YAML::XS::DumpFile( $GLOBAL_FILE, $global_state_struct );
    }
}

#----------------------------------------------------------------------

sub export_global_state {
    my ( $self, %params ) = @_;
    my $filesize;
    my $output_filename = File::Spec->rel2abs( $params{output_file} );

    $self->set_global_state_info();

    my $global_root_tmp = "$POLICY_BUNDLE_BASE_DIR/save_global.$$";
    my $global_db       = "$global_root_tmp/db";
    make_path($global_root_tmp);
    make_path($global_db);
    File::Copy::cp( "$DATABASE_BASE_DIR/PLC", $global_db );
    File::Copy::cp( "$DATABASE_BASE_DIR/DCC", $global_db );
    File::Copy::cp( $GLOBAL_FILE,             $global_root_tmp );

    my $archive_result =
      F5::SystemUtils::run_system_tool(
        command_line_params => [ 'tar', '-zpc', '-f', $output_filename, '-C', $global_root_tmp, 'db', 'global' ] );
    if ( $archive_result->{completed_successfully} ) {
        $filesize = -s $output_filename;
    }
    else {
        die { "error_message" => "Failed to save '$output_filename' - " . $archive_result->{standard_error} };
    }

    rmtree($global_root_tmp);

    return { completed_successfully => 1, filename => $output_filename, file_size => $filesize };
}

#----------------------------------------------------------------------

sub dump_bundle {
    my ( $self, %params ) = @_;
    my $input_filename = File::Spec->rel2abs( $params{filename}, $self->{workdir} );

    die { "error_message" => "Missing bundle file - '$input_filename'" } unless -f $input_filename;

    my $dump_catalog = F5::SystemUtils::run_system_tool( command_line_params => [ 'tar', '-O', '-xf', $input_filename, $CATALOG_FILE, ] );
    my $dump_global  = F5::SystemUtils::run_system_tool( command_line_params => [ 'tar', '-O', '-xf', $input_filename, 'global', ] );
    unless ( $dump_catalog->{completed_successfully} || $dump_global->{completed_successfully} ) {
        die { "error_message" => "Failed to open '$input_filename' - " . $dump_catalog->{standard_error} };
    }

    my $res = $dump_catalog->{standard_output} || $dump_global->{standard_output};

    return $res;

}

#----------------------------------------------------------------------

sub reset_global {
    my ( $self, %params ) = @_;

    F5::DbUtils::restore_sqlite_databases( suffix => 'initial' );

    # reinstall signatures updates
    $self->{config_set}->reinstall_all_updates( call_flags => $call_flags );

    # clean existing global_state_info
    unlink glob "$POLICY_BUNDLE_BASE_DIR/*";

    $self->set_global_state_info() unless $params{skip_set_global_info};

    # TODO: clean old bd_config files
}

#----------------------------------------------------------------------

sub extract_policy_name {
    my ( $self, %params ) = @_;
    my $filename = $params{filename};
    my $policy;
    require JSON;
    if ( F5::Utils::File::is_file_json( filename => $filename ) ) {
        $policy = JSON::from_json( File::Slurp::read_file($filename) );
    }
    if ( !ref $policy ) {
        die { "error_message" => "Bad file format - expected JSON" };
    }
    if ( !defined( $policy->{policy}->{name} ) ) {
        die { "error_message" => "Invalid file - expected declarative policy" };
    }
    return $policy->{policy}->{name};
}

#----------------------------------------------------------------------

sub archive_bundle {
    my ( $self, %params ) = @_;
    my $filesize;
    my $source_dir      = $params{source_dir};
    my $output_filename = $params{output_filename};
    my $archive_result =
      F5::SystemUtils::run_system_tool(
        command_line_params => [ 'tar', '-zpc', '-f', $output_filename, '-C', $source_dir, $CATALOG_FILE, 'subbundle' ] );
    if ( $archive_result->{completed_successfully} ) {
        $filesize = -s $output_filename;
    }
    else {
        die { "error_message" => "Failed to save '$output_filename' - " . $archive_result->{standard_error} };
    }

    return { completed_successfully => 1, filename => $output_filename, file_size => $filesize };
}

#----------------------------------------------------------------------

1;
