package F5::BdRequest;

use strict;
use warnings;

use Bit::Vector;

use bd::Main;
use F5::IpAddr;
use F5::Violation::Utils;
use F5::Utils::Rest;
use F5::ObjUtils qw(dclone);
use F5::ASMConfig::Entity::Violation;
use F5::ASMConfig::Entity::Signature;
use F5::ASMConfig::Entity::HttpCheck;
use F5::ASMConfig::Entity::Normalization;
use F5::ASMConfig::Entity::XmlDefenseAttribute;
use F5::ASMConfig::Entity::JsonDefenseAttribute;
use F5::ASMConfig::Entity::GwtDefenseAttribute;
use F5::ASMConfig::Entity::ContentProfile;
use F5::ASMConfig::Entity::GraphqlDefenseAttribute;
use F5::ASMConfig::Entity::Flow;
use F5::ASMConfig::Entity::Header;
use F5::ASMConfig::Entity::Cookie;
use F5::ASMConfig::Entity::Parameter;
use F5::ASMConfig::Entity::URL;
use F5::ASMConfig::Entity::Filetype;
use F5::ASMConfig::Entity::RequestLog;
use F5::Logger;
use F5::ClientReputationReason;
use F5::ASMConfig::Entity::Incident;


my $subviolations_cache = {};

my %entity_type_class_map = (
    header => ['F5::ASMConfig::Entity::Header', 'header_name'],
    cookie => ['F5::ASMConfig::Entity::Cookie', 'cookie_name'],
    param => ['F5::ASMConfig::Entity::Parameter', 'param_name'],
    url => ['F5::ASMConfig::Entity::URL', 'url_name'],
    filetype => ['F5::ASMConfig::Entity::Filetype', 'filetype_name'],
    websocketUrl => ['F5::ASMConfig::Entity::WebsocketUrl', 'websocket_url_name'],
);


my %viol_handling = (
    F5::Violation::VIOL_HTTP_RESPONSE_STATUS() => { extra_field_handler => 'check_http_response_status', },
    F5::Violation::VIOL_FLOW_MANDATORY_PARAMS() => {
        request_path_map => {
            'actual_value'   => [ 'number_of_mandatory_params' ],
            'expected_value' => [ 'request_additional_violation_info', 'max_mandatory_params' ],
        },
    },
    F5::Violation::VIOL_PARAMETER_MULTIPART_NULL_VALUE() => {
        ignore_entity_staging => 1,
    },
    F5::Violation::VIOL_ENCODING() => {
        ignore_entity_staging => 1,
        extra_field_handler   => 'get_encoding_fields',
    },
    F5::Violation::VIOL_ATTACK_SIGNATURE()     => {
        entity_instances_creator => 'signature_instances',
        extra_field_handler     => 'get_signature_fields',
        extract_profile_on_profile_entity => 1,
        look_in_request_entity  => 1,
    },
    F5::Violation::VIOL_EVASION() => {
        instances_creator    => 'evasion_instances',
        extra_field_handler  => 'get_evasion_subviolations',
        non_entity_violation => 1,
        field_map           => {
            'name' => 'observed_entity_name',
            'value' => 'observed_entity_value',
        },
    },
    F5::Violation::VIOL_PARAMETER_VALUE_METACHAR() => {
        entity_instances_creator   => 'param_value_metacharacter_instances',
        extra_field_handler => 'param_value_metacharacter_violation',
        extract_content_profile => 1,
        extract_profile_on_profile_entity => 1,
    },
    F5::Violation::VIOL_PARAMETER_NAME_METACHAR() => {
        multi_occurrence => 'name_metachar',
        entity_type      => 'param',
        field_map        => { 'instance' => 'metacharacter', },
    },
    F5::Violation::VIOL_PARAMETER_EMPTY_VALUE() => { entity_type => 'param', },
    F5::Violation::VIOL_PARAMETER()             => {
        entity_type         => 'param',
        extra_field_handler => 'check_tighten_wildcard',
    },
    F5::Violation::VIOL_PARAMETER_DATA_TYPE() => {
        entity_type => 'param',
        field_map   => { 'expected_data_type' => 'observed_entity_param_data_type', },
    },
    F5::Violation::VIOL_PARAMETER_VALUE_REGEXP()  => { field_path_map => { 'regexp' => [ 'param_additional_violation_info', 'reg_exp' ], }, },
    F5::Violation::VIOL_PARAMETER_NUMERIC_VALUE() => {
        field_path_map => {
            'max_numeric_value' => [ 'param_additional_violation_info', 'max_numeric_value' ],
            'min_numeric_value' => [ 'param_additional_violation_info', 'min_numeric_value' ],
            'multiple_of_value' => [ 'param_additional_violation_info', 'multiple_of_value' ],
            'exclusive_max' => [ 'param_additional_violation_info', 'flg_exclusive_max' ],
            'exclusive_min' => [ 'param_additional_violation_info', 'flg_exclusive_min' ],
        },
    },
    F5::Violation::VIOL_PARAMETER_VALUE_LENGTH() => {
        field_path_map => {
            'expected_value' => [ 'param_additional_violation_info', 'expected_value_len' ],
            'parameter_value_length_violation_type' => [ 'param_additional_violation_info', 'parameter_value_length_type' ],
        },
    },
    F5::Violation::VIOL_HEADER_LENGTH() => {
        entity_type    => 'header',
        field_path_map => {
            'header_len'                     => [ 'header_additional_violation_info', 'actual_len' ],
            'actual_value'                   => [ 'header_additional_violation_info', 'actual_len' ],
            'expected_value'                 => [ 'header_additional_violation_info', 'max_len' ],
            'observed_entity_header_content' => [ 'name', ],
        },
        block_violation_entry => ['observed_entity_name'],
    },
    F5::Violation::VIOL_COOKIE_LENGTH() => {
        viol_elem_map => {
            'cookie_len'                     => 'cookie_len',
            'actual_value'                   => 'cookie_len',
            'observed_entity_cookie_content' => 'cookie_content',
        },
        field_path_map => { 'expected_value' => [ 'request_additional_violation_info', 'max_cookie_len' ], },
        extra_field_handler => 'cookie_length',
    },
    F5::Violation::VIOL_URL_LENGTH() => {
        entity_type      => 'filetype',
        field_path_map   => { 'expected_value' => [ 'filetype_additional_violation_info', 'max_url_len' ], },
        request_path_map => {
            'url_len'      => [ 'filetype', 'url_len' ],
            'actual_value' => [ 'filetype', 'url_len' ],
        },
    },
    F5::Violation::VIOL_REQUEST_LENGTH() => {
        entity_type      => 'filetype',
        field_path_map   => { 'expected_value' => [ 'filetype_additional_violation_info', 'max_req_len' ], },
        request_path_map => {
            'req_len'      => [ 'filetype', 'req_len' ],
            'actual_value' => [ 'filetype', 'req_len' ],
        },
    },
    F5::Violation::VIOL_QUERY_STRING_LENGTH() => {
        entity_type      => 'filetype',
        field_path_map   => { 'expected_value' => [ 'filetype_additional_violation_info', 'max_qs_len' ], },
        request_path_map => {
            'qs_len'       => [ 'filetype', 'qs_len' ],
            'actual_value' => [ 'filetype', 'qs_len' ],
        },
    },
    F5::Violation::VIOL_POST_DATA_LENGTH() => {
        entity_type      => 'filetype',
        field_path_map   => { 'expected_value' => [ 'filetype_additional_violation_info', 'max_pd_len' ], },
        request_path_map => {
            'pd_len'       => [ 'filetype', 'pd_len' ],
            'actual_value' => [ 'filetype', 'pd_len' ],
        },
    },
    F5::Violation::VIOL_FILETYPE() => {
        extra_field_handler => 'check_tighten_wildcard',
        entity_type         => 'filetype',
        field_map           => {
            'url_len'                  => 'url_len',
            'qs_len'                   => 'qs_len',
            'pd_len'                   => 'pd_len',
            'req_len'                  => 'req_len',
            'flg_disallowed_file_type' => 'observed_entity_is_disallowed',
        },
    },
    F5::Violation::VIOL_HEADER_METACHAR() => {
        multi_occurrence => 'metachar',
        entity_type      => 'header',
        field_map        => { 'instance' => 'metacharacter', },
    },
    F5::Violation::VIOL_MANDATORY_HEADER() => { entity_type => 'header', },
    F5::Violation::VIOL_HEADER_REPEATED() => { entity_type => 'header', },
    F5::Violation::VIOL_MANDATORY_REQUEST_BODY() => {
        extra_field_handler => 'get_url_violation',
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_method'          => [ 'url', 'method' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_URL_METACHAR()     => {
        multi_occurrence => 'metachar',
        entity_type      => 'url',
        field_map        => { 'instance' => 'metacharacter', },
    },
    F5::Violation::VIOL_URL() => {
        extra_field_handler => 'get_url_violation',
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_method'          => [ 'url', 'method' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_COOKIE_EXPIRED() => {
        entity_type    => 'cookie',
        field_path_map => { 'expected_value' => [ 'cookie_additional_violation_info', 'expiration_time' ], },
        viol_elem_map  => {
            'observed_entity_value' => 'diff_time',
            'actual_value'          => 'diff_time',
        },
    },
    F5::Violation::VIOL_LOGIN_URL_EXPIRED() => {
        field_path_map => {
            'expected_value'        => [ 'request_additional_violation_info', 'login_expiration_interval' ],
            'observed_entity_value' => [ 'request_additional_violation_info', 'login_time_diff' ],
            'actual_value'          => [ 'request_additional_violation_info', 'login_time_diff' ],
        },
    },
    F5::Violation::VIOL_CSRF() => {
                                   extra_field_handler => 'get_csrf_details',
                                   field_path_map => {
                                                      'origin_header_domain'      => [ 'origin_header', 'domain' ],
                                                      'origin_header_protocol'    => [ 'origin_header', 'protocol' ],
                                                      'origin_header_port'        => [ 'origin_header', 'port' ],
                                                      'origin_header_description' => [ 'origin_header', 'description' ],
                                                      'referer_header_domain'      => [ 'referer_header', 'domain' ],
                                                      'referer_header_protocol'    => [ 'referer_header', 'protocol' ],
                                                      'referer_header_port'        => [ 'referer_header', 'port' ],
                                                      'referer_header_description' => [ 'referer_header', 'description' ],
                                                     },
                                  },
    F5::Violation::VIOL_CSRF_EXPIRED() => {
                                           extra_field_handler => 'get_csrf_details',
                                           field_path_map => {
                                                              'origin_header_domain'      => [ 'origin_header', 'domain' ],
                                                              'origin_header_protocol'    => [ 'origin_header', 'protocol' ],
                                                              'origin_header_port'        => [ 'origin_header', 'port' ],
                                                              'origin_header_description' => [ 'origin_header', 'description' ],
                                                              'referer_header_domain'      => [ 'referer_header', 'domain' ],
                                                              'referer_header_protocol'    => [ 'referer_header', 'protocol' ],
                                                              'referer_header_port'        => [ 'referer_header', 'port' ],
                                                              'referer_header_description' => [ 'referer_header', 'description' ],
                                                              'expected_value'        => [ 'request_additional_violation_info', 'login_expiration_interval' ],
                                                              'observed_entity_value' => [ 'request_additional_violation_info', 'login_time_diff' ],
                                                              'actual_value'          => [ 'request_additional_violation_info', 'login_time_diff' ],
                                                             },
    },
    F5::Violation::VIOL_REQUEST_MAX_LENGTH() => {
        field_path_map => {
            'expected_value'        => [ 'request_additional_violation_info', 'req_buffer_size' ],
            'observed_entity_value' => [ 'request_additional_violation_info', 'total_req_len' ],
            'actual_value'          => [ 'request_additional_violation_info', 'total_req_len' ],
        },
    },
    F5::Violation::VIOL_COOKIE_MALFORMED() => {
        entity_type => 'cookie',
        field_map   => { 'cookie_viol_desc' => 'observed_entity_violation_description', },
    },
    F5::Violation::VIOL_COOKIE_MODIFIED() => {
        entity_type    => ['cookie', 'request'],
        field_map      => { 'is_new_cookie' => 'is_new_cookie', },
        field_path_map => { 'too_many_cookies' => [ 'request_additional_violation_info', 'too_many_cookies_flg' ], },
    },
    F5::Violation::VIOL_DYNAMIC_SESSION() =>
      { field_path_map => { 'observed_entity_dynamic_url' => [ 'request_additional_violation_info', 'dynamic_session_url' ], }, },
    F5::Violation::VIOL_HTTP_PROTOCOL() => {
        extra_field_handler => 'get_http_subviolations',
        viol_elem_map       => { 'observed_entity_violation_description' => 'viol_desc', },
    },
    F5::Violation::VIOL_VIRUS()      => { viol_elem_map => { 'observed_entity_violation_description' => 'viol_desc', }, },
    F5::Violation::VIOL_DATA_GUARD() => {
        viol_elem_map => {
            'info_leakage_target' => 'target',
            'info_leakage_type'   => 'type_str',
            'snippet_buffer'      => 'context',
        },
        change_info_elem    => 'info_leakage',
        extra_field_handler => 'get_data_guard_violation',
    },

    F5::Violation::VIOL_COOKIE_MALFORMED()    => { extract_content_profile => 1, },
    F5::Violation::VIOL_GWT_MALFORMED()       => { extract_content_profile => 1, },
    F5::Violation::VIOL_GRAPHQL_MALFORMED()   => { extract_content_profile => 1, },
    F5::Violation::VIOL_GRAPHQL_INTROSPECTION_QUERY()   => { extract_content_profile => 1, },
    F5::Violation::VIOL_GRAPHQL_ERROR_RESPONSE() => {
        extract_content_profile => 1,
        field_map      => { 'detected_pattern' => 'detected_pattern', },
    },
    F5::Violation::VIOL_XML_SOAP_ATTACHMENT() => {
        extract_content_profile  => 1,
        content_profile_elements => {},
    },
    F5::Violation::VIOL_XML_WEB_SERVICES_SECURITY() => {
        extract_content_profile => 1,
        extra_field_handler     => 'get_wss_subviolations',
    },
    F5::Violation::VIOL_JSON_FORMAT() => {
        extra_field_handler => 'get_format_violations',
        entity_instances_creator => 'get_format_violations_instances',
    },
    F5::Violation::VIOL_XML_FORMAT() => {
        extra_field_handler => 'get_format_violations',
        entity_instances_creator => 'get_format_violations_instances',
    },
    F5::Violation::VIOL_GWT_FORMAT() => {
        extra_field_handler => 'get_format_violations',
        entity_instances_creator => 'get_format_violations_instances',
    },
    F5::Violation::VIOL_GRAPHQL_FORMAT() => {
        extra_field_handler => 'get_format_violations',
        entity_instances_creator => 'get_format_violations_instances',
    },

    F5::Violation::VIOL_XML_SOAP_METHOD() => {
        extra_field_handler        => 'get_xml_soap_method',
    },
    F5::Violation::VIOL_XML_MALFORMED()   => {
        extract_content_profile => 1,
    },
    F5::Violation::VIOL_JSON_MALFORMED()  => { extract_content_profile => 1, },

    F5::Violation::VIOL_XML_SCHEMA() => {
        extract_content_profile  => 1,
        extra_field_handler      => 'get_xml_schema',
        content_profile_elements => {
            'observed_entity_soap_method'    => 1,
            'observed_entity_soap_namespace' => 1,
        },
    },

    F5::Violation::VIOL_JSON_SCHEMA() => {
        extra_field_handler => 'get_json_schema',
        entity_instances_creator => 'json_schema_violation_instances',
    },

    F5::Violation::VIOL_REDIRECT()     => { extra_field_handler => 'get_viol_redirect', },
    F5::Violation::VIOL_BRUTE_FORCE() => {
        field_path_map => {
            'brute_force_failed_logins'                                     => [ 'brute_force_violation_details', 'failed_logins' ],
            'brute_force_credentials_stuffing_matches'                      => [ 'brute_force_violation_details', 'credentials_stuffing_matches' ],
            'brute_force_passed_captcha_challenges'                         => [ 'brute_force_violation_details', 'passed_captcha_challenges' ],
            'brute_force_passed_client_side_integrity_challenges'           => [ 'brute_force_violation_details', 'passed_client_side_integrity_challenges' ],
            'brute_force_detection_period'                                  => [ 'brute_force_violation_details', 'detection_period' ],
            'brute_force_prevention_duration'                               => [ 'brute_force_violation_details', 'prevention_duration' ],
            'brute_force_failed_logins_threshold'                           => [ 'brute_force_violation_details', 'failed_logins_threshold' ],
            'brute_force_passed_captcha_challenges_threshold'               => [ 'brute_force_violation_details', 'passed_captcha_challenges_threshold' ],
            'brute_force_credentials_stuffing_matches_threshold'            => [ 'brute_force_violation_details', 'credentials_stuffing_matches_threshold' ],
            'brute_force_passed_client_side_integrity_challenges_threshold' => [ 'brute_force_violation_details', 'passed_client_side_integrity_challenges_threshold' ],
            'brute_force_mitigation_action'                                 => [ 'brute_force_violation_details', 'mitigation_action' ],
            'brute_force_actual_mitigation_action'                          => [ 'brute_force_violation_details', 'actual_mitigation_action' ],
            'brute_force_mitigation_action_reason'                          => [ 'brute_force_violation_details', 'actual_action_reason' ],
            'brute_force_mitigated_entity'                                  => [ 'brute_force_violation_details', 'mitigated_entity' ],
        },
        extra_field_handler => 'get_brute_force_subviolations',
    },
    F5::Violation::VIOL_MALICIOUS_IP()      => { extra_field_handler => 'get_malicious_ip_subviolations', },
    F5::Violation::VIOL_FILE_UPLOAD()       => { viol_elem_map       => { 'observed_entity_violation_description' => 'uploaded_file_desc', }, },
    F5::Violation::VIOL_SESSION_AWARENESS() => { extra_field_handler => 'session_awareness_subviolations', },
    F5::Violation::VIOL_METHOD()            => { extra_field_handler => 'method_violation', },
    F5::Violation::VIOL_URL_CONTENT_TYPE()  => { extra_field_handler => 'url_content_type', },

    F5::Violation::VIOL_ASM_COOKIE_HIJACKING() => {
        extra_field_handler => 'asm_cookie_hijacking',
    },
    F5::Violation::VIOL_FILE_UPLOAD_IN_BODY() => {
        extra_field_handler => 'file_upload_in_body',
    },
    F5::Violation::VIOL_WEBSOCKET_FRAME_LENGTH() => {
        field_path_map => {
            'expected_value' => [ 'websocket_violation_info', 'frame_size', 'expected' ],
            'actual_value'   => [ 'websocket_violation_info', 'frame_size', 'actual' ],
        },
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_WEBSOCKET_TEXT_MESSAGE_NOT_ALLOWED() => {
        field_path_map => {
            'expected_value' => [ 'websocket_violation_info', 'type_not_allowed', 'expected' ],
            'actual_value'   => [ 'websocket_violation_info', 'type_not_allowed', 'actual' ],
        },
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_WEBSOCKET_BINARY_MESSAGE_LENGTH() => {
        field_path_map => {
            'expected_value' => [ 'websocket_violation_info', 'binary_message_size', 'expected' ],
            'actual_value'   => [ 'websocket_violation_info', 'binary_message_size', 'actual' ],
        },
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_WEBSOCKET_BINARY_MESSAGE_NOT_ALLOWED() => {
        field_path_map => {
            'expected_value' => [ 'websocket_violation_info', 'type_not_allowed', 'expected' ],
            'actual_value'   => [ 'websocket_violation_info', 'type_not_allowed', 'actual' ],
        },
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_CROSS_ORIGIN_REQUEST() => {
        field_path_map => {
            'origin_header_domain'      => [ 'origin_header', 'domain' ],
            'origin_header_protocol'    => [ 'origin_header', 'protocol' ],
            'origin_header_port'        => [ 'origin_header', 'port' ],
            'origin_header_description' => [ 'origin_header', 'description' ],
        },
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_method'          => [ 'url', 'method' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_WEBSOCKET_FRAME_MASKING() => {
        field_path_map => {
            'expected_value' => [ 'websocket_violation_info', 'masking_error', 'expected' ],
            'actual_value'   => [ 'websocket_violation_info', 'masking_error', 'actual' ],
        },
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
        },
    },
    F5::Violation::VIOL_WEBSOCKET_FRAMES_PER_MESSAGE_COUNT() => {
        field_path_map => {
            'expected_value' => [ 'websocket_violation_info', 'frames_per_message', 'expected' ],
            'actual_value'   => [ 'websocket_violation_info', 'frames_per_message', 'actual' ],
        },
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_WEBSOCKET_BAD_REQUEST() => {
        field_path_map => {
            'expected_value' => [ 'websocket_violation_info', 'bad_request', 'expected' ],
            'actual_value'   => [ 'websocket_violation_info', 'bad_request', 'actual' ],
        },
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
        },
    },
    F5::Violation::VIOL_WEBSOCKET_FRAMING_PROTOCOL() => {
        extra_field_handler => 'get_framing_protocol_violations',
    },
    F5::Violation::VIOL_WEBSOCKET_EXTENSION() => {
        extra_field_handler => 'get_websocket_extension_violations',
        request_path_map    => {
            'observed_entity_name'            => [ 'url', 'name' ],
            'observed_entity_is_disallowed'   => [ 'url', 'flg_disallowed_url' ],
            'enforcement_state_is_in_staging' => [ 'url', 'is_in_staging' ],
        },
    },
    F5::Violation::VIOL_PLAINTEXT_FORMAT() => {
        extra_field_handler => 'get_format_violations',
        entity_instances_creator => 'get_format_violations_instances',
    },
    F5::Violation::VIOL_THREAT_CAMPAIGN() => {
        extra_field_handler => 'threat_campaign_violations',
        entity_instances_creator => 'threat_campaign_instances',
    },
    F5::Violation::VIOL_PARAMETER_ARRAY_VALUE() => {
        extra_field_handler => 'parameter_array_violations',
        entity_instances_creator => 'parameter_array_instances',
    },
    F5::Violation::VIOL_MANDATORY_PARAMETER() => {
        extra_field_handler => 'mandatory_parameter_violations',
        entity_instances_creator => 'mandatory_parameter_instances',
    },
    F5::Violation::VIOL_PARAMETER_LOCATION() => {
    },
    F5::Violation::VIOL_BLOCKING_CONDITION() => {
        extra_field_handler => 'blocking_condition_violations',
        entity_instances_creator => 'blocking_condition_instances',
    },
    F5::Violation::VIOL_THREAT_ANALYSIS() => {
        extra_field_handler => 'threat_analysis_violations',
    },
    F5::Violation::VIOL_MALICIOUS_DEVICE() => {
        extra_field_handler => 'malicious_device_violations',
    },
    F5::Violation::VIOL_LEAKED_CREDENTIALS() => {
        request_path_map => {
	    'violation_username' => [ 'leaked_credentials_violation_details', 'username' ],
	    'leaked_credentials_mitigation_action' => [ 'leaked_credentials_violation_details', 'mitigation_action' ],
	},
    },
    F5::Violation::VIOL_CONVICTION() => {
        extra_field_handler => 'conviction_violations',
        entity_instances_creator => 'conviction_instances',
    },
    F5::Violation::VIOL_DATA_INTEGRITY() => {
        extra_field_handler => 'datasafe_violations',
        request_path_map => {
	    'datasafe_profile_name' => [ 'data_integrity_alert', 'profile_name' ],
	    'datasafe_configured_url_name' => [ 'data_integrity_alert', 'configured_url_name' ],
	    'datasafe_configured_view_id' => [ 'data_integrity_alert', 'configured_view_id' ],
	    'datasafe_score' => [ 'data_integrity_alert', 'score' ],
	    'datasafe_extra_manipulated_parameters_count' => [ 'data_integrity_alert', 'extra_manipulated_parameters_count' ],
	},
    },
    F5::Violation::VIOL_SERVER_SIDE_HOST() => {
        entity_type         => 'param',
        extra_field_handler => 'get_ssrf_details',
    },
);

use constant {
    JSON_SCHEMA_VALIDATION_ERROR_CODE_FALSE => 1,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_TYPE => 2,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_DISALLOW => 3,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_ADDITIONAL_PROPERTIES => 6,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_ANY_OF => 7,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_ONE_OF => 8,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_ALL_OF => 9,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_ITEMS => 10,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_ADDITIONAL_ITEMS => 11,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_REQUIRED => 12,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_DEPENDENCIES => 13,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MINIMUM => 14,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MAXIMUM => 15,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_EX_MINIMUM => 16,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_EX_MAXIMUM => 17,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MIN_ITEMS => 18,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MAX_ITEMS => 19,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MIN_PROPERTIES => 20,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MAX_PROPERTIES => 21,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MIN_LENGTH => 22,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MAX_LENGTH => 23,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_UNIQUE_ITEMS => 24,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_PATTERN => 25,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_ENUM => 26,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_FORMAT => 27,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_MULTIPLE_OF => 28,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_NOT => 29,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_PROPERTY_NAMES => 30,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_CONTAINS => 31,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_CONST => 32,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_IF_THEN => 33,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_IF_ELSE => 34,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_JSON_LOAD => 35,
    JSON_SCHEMA_VALIDATION_ERROR_CODE_REF_LOOP => 36,
};

my $json_schema_violation_error_code_to_violation_record = {
    JSON_SCHEMA_VALIDATION_ERROR_CODE_FALSE() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'False Violation',
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_TYPE() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Illegal type',
                      actual_value => $violation_details->{received_str},
                      expected_value => $violation_details->{expected_str},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_DISALLOW() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => "Disallowed type. Disallowed values: $violation_details->{expected_str}",
                      actual_value => $violation_details->{received_str},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_ADDITIONAL_PROPERTIES() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Illegal additional property',
                      property_name => $violation_details->{property_name},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_ANY_OF() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'AnyOf violation',
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_ONE_OF() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'OneOf violation',
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_ALL_OF() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'AllOf violation',
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_ITEMS() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Illegal items',
                      property_name => $violation_details->{property_name},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_ADDITIONAL_ITEMS() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => "Illegal additional items: $violation_details->{num_additional_items} additional Item(s) detected",
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_REQUIRED() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Required property missing',
                      property_name => $violation_details->{property_name},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_DEPENDENCIES() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Dependencies violation',
                      property_name => $violation_details->{property_name},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MINIMUM() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Value is less than minimum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MAXIMUM() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Value is greater than maximum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_EX_MINIMUM() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Value is less than or equal to exclusive minimum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_EX_MAXIMUM() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Value is greater than or equal to exclusive maximum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MIN_ITEMS() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Number of items is less than minimum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MAX_ITEMS() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Number of items is greater than maximum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MIN_PROPERTIES() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Number of properties is less than minimum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MAX_PROPERTIES() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Number of properties is greater than maximum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MIN_LENGTH() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Value length is less than minimum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MAX_LENGTH() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Value length is greater than maximum',
                      actual_value => $violation_details->{received},
                      expected_value => $violation_details->{expected},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_UNIQUE_ITEMS() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => "Array items are not unique. Items $violation_details->{first_non_unique_item} and $violation_details->{second_non_unique_item} are identical",
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_PATTERN() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => "Pattern violation $violation_details->{received_str} doesn't match $violation_details->{expected_str}",
                      actual_value => $violation_details->{received_str},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_ENUM() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Illegal enum value',
                      actual_value => $violation_details->{received_str},
                      expected_value => $violation_details->{expected_str},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_FORMAT() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => "Format violation: $violation_details->{received_str} doesn't match $violation_details->{expected_str}",
                      actual_value => $violation_details->{received_str},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_MULTIPLE_OF() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => "MultipleOf violation: $violation_details->{received} is not a multiple of $violation_details->{expected}",
                      actual_value => $violation_details->{received},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_NOT() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Not violation',
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_PROPERTY_NAMES() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Invalid property name',
                      property_name => $violation_details->{property_name},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_CONTAINS() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Contains violation',
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_CONST() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => "Const violation: $violation_details->{received_str} doesn't match $violation_details->{expected_str}",
                      actual_value => $violation_details->{received_str},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_IF_THEN() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'If-then violation',
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_IF_ELSE() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'If-else violation',
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_JSON_LOAD() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Failed to parse JSON payload',
                      property_name => $violation_details->{property_name},
                     };
          },

    JSON_SCHEMA_VALIDATION_ERROR_CODE_REF_LOOP() => sub {

              my (%params) = @_;
              my $violation_details = $params{violation_details};

              return {validation_error => 'Reference loop detected',
                      property_name => $violation_details->{property_name},
                     };
          },
};

#----------------------------------------------------------------------

sub new {
    my ( $class, %params ) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my ( $self, %params ) = @_;

    my $bd_rec                     = $params{bd};
    my $bd_request_selected_fields = $params{bd_request_selected_fields};
    $self->{expand_fields} = $params{expand_fields};
    $self->{dbh}            = $params{dbh};
    $self->{logger} = $params{logger} || F5::Logger->get_logger(logger_name => 'asm_config_server.pl');
    $self->{request_policy} = $params{request_policy};

    my $bd = bd::Main->new();
    $bd->unpack($bd_rec);
    my $bd_data = $bd->to_hashref();
    $self->{bd_data}  = $bd_data;
    $self->{request}  = $bd_data->{data}->{request};
    $self->{response} = $bd_data->{data}->{response};

    $self->{account_id}                        = $bd_data->{account_id};
    $self->{request_log_time}                  = $self->{request}->{log_time};
    $self->{request_severity_id}               = $self->{request}->{severity_id};
    $self->{request_host}                      = $self->{request}->{host_header};
    $self->{dest_ip}                           = $self->{request}->{dest_ip};
    $self->{max_request_header_length}         = $self->{request}->{max_header_len};
    $self->{login_result}                      = $self->{request}->{login_result};
    $self->{max_request_cookie_length}         = $self->{request}->{max_cookie_len};
    $self->{request_raw}                       = $self->{request}->{raw_req_data}->{raw_request};
    $self->{request_is_truncated}              = $self->{request}->{raw_req_data}->{flag_req_truncated} || 0;
    $self->{request_actual_size}               = $self->{request}->{raw_req_data}->{actual_request_len};
    $self->{is_base64_encoded}                 = $self->{request}->{raw_req_data}->{is_base64_encoded} || 0;
    $self->{enforcement_state_rating}          = $self->{request}->{violation_rating};
    $self->{request_schema}                    = $self->{request}->{url}->{app_id};
    $self->{auth_type}                         = $self->{request}->{authentication_type};
    $self->{request_log_request_content_type}  = $self->{request}->{header_content_type};
    $self->{captcha_result}                    = $self->{request}->{captcha_result};
    $self->{request_log_response_content_type} = $self->{response}->{header_content_type};
    $self->{response_actual_size}              = $self->{response}->{response_len};
    $self->{response_datetime}                 = $self->{response}->{response_time};
    $self->{response_is_truncated}             = $self->{response}->{is_truncated} || 0;
    $self->{response_is_base64_encoded}        = $self->{response}->{is_base64_encoded} || 0;
    $self->{client_type}                       = $self->{request}->{client_type};

    if ( !$bd_request_selected_fields || $bd_request_selected_fields->{source_ip} ) {
        $self->{source_ip} = $self->compose_source_ip();
    }
    if ( $bd_data->{data}->{response}->{raw_response} ) {
        $self->{response_raw} = $bd_data->{data}->{response}->{raw_response};
    }
    if ( $self->{response}->{exclusion_reason} ) {
        $self->{raw_response_exclusion_reason} = $self->{response}->{exclusion_reason};
    }
    if ( $self->{response}->{is_truncated} ) {
        $self->{raw_response_is_truncated} = $self->{response}->{is_truncated};
    }
    if ( $self->{response}->{is_compressed} ) {
        $self->{raw_response_was_compressed} = $self->{response}->{is_compressed};
        $self->{raw_response_compression_method} = $self->{response}->{compression_method};
    }
    if ( $self->{request}->{raw_req_data}->{is_compressed} ) {
        $self->{raw_request_was_compressed} = $self->{request}->{raw_req_data}->{is_compressed};
        $self->{raw_request_compression_method} = $self->{request}->{raw_req_data}->{compression_method};
    }
    if ( $self->{request}->{websocket_message_type} ) {
        $self->{websocket_message_type} = (keys %{$self->{request}->{websocket_message_type}})[0];
    }
    if ( defined($self->{request}->{websocket_message_direction}) ) {
        $self->{websocket_message_direction} = $self->{request}->{websocket_message_direction};
    }
    if ( defined($self->{request}->{blocking_exception_reason}) ) {
        $self->{blocking_exception_reason} = $self->{request}->{blocking_exception_reason};
    }

    if ( defined($self->{request}->{mobile_app_version}) ){
        $self->{mobile_app_version} = $self->{request}->{mobile_app_version};
    }

    if ( defined($self->{request}->{mobile_app_name}) ){
        $self->{mobile_app_name} = $self->{request}->{mobile_app_name};
    }

    if ( defined($self->{request}->{protocol_info}) ){
        $self->{protocol_info} = $self->{request}->{protocol_info};
    }

    my $additional_violation_info = $self->{request}->{request_additional_violation_info};
    if ( $additional_violation_info && $additional_violation_info->{session_awareness_flag} && $additional_violation_info->{session_awareness_scope} )
    {
        my $flag_list  = $additional_violation_info->{session_awareness_flag};
        my $scope_list = $additional_violation_info->{session_awareness_scope};

        my @session_awareness_data = ();
        for ( my $i=0; $i < scalar @$scope_list; ++$i ) {
            push( @session_awareness_data, { scope => $scope_list->[$i], flag => $flag_list->[$i] } );
        }

        $self->{session_awareness_data} = \@session_awareness_data;
    }

    $self->{is_in_staging} = 0;
    $self->{is_alarmed} = 0;
    if (
        !$bd_request_selected_fields
        || (   $bd_request_selected_fields->{violations}
            || $bd_request_selected_fields->{has_violations}
            || $bd_request_selected_fields->{enforcement_state_has_request_violations}
            || $bd_request_selected_fields->{enforcement_state_is_in_staging}
            || $bd_request_selected_fields->{enforcement_state_has_response_violations} )
      )
    {
        $self->{enforcement_state_has_request_violations} = 0;
        $self->{enforcement_state_has_response_violations} = 0;
        my ( $request_violations, $response_violations ) = $self->get_all_request_violations();
        $self->{violations} = [ @$request_violations, @$response_violations ];

        for my $violation ( @$request_violations ) {
            $self->{is_in_staging} = $self->{is_in_staging} || $violation->{enforcement_state_is_in_staging} || 0;
            my $viol_is_alarmed = $violation->{enforcement_state_is_in_staging} ? 0 : $violation->{enforcement_state_is_alarmed};
            $self->{is_alarmed} = $self->{is_alarmed}
                                  || $viol_is_alarmed
                                  || 0;
            $self->{enforcement_state_has_request_violations} = $self->{enforcement_state_has_request_violations} 
                                  || (! $violation->{enforcement_state_is_in_staging}
                                      && ($violation->{enforcement_state_is_alarmed} || $violation->{enforcement_state_is_blocked}))
                                  || 0;
        }

        for my $violation ( @$response_violations ) {
            $self->{is_in_staging} = $self->{is_in_staging} || $violation->{enforcement_state_is_in_staging} || 0;
            my $viol_is_alarmed = $violation->{enforcement_state_is_in_staging} ? 0 : $violation->{enforcement_state_is_alarmed};
            $self->{is_alarmed} = $self->{is_alarmed}
                                  || $viol_is_alarmed
                                  || 0;
            $self->{enforcement_state_has_response_violations} = $self->{enforcement_state_has_response_violations}
                                  || (! $violation->{enforcement_state_is_in_staging}
                                      && ($violation->{enforcement_state_is_alarmed} || $violation->{enforcement_state_is_blocked}))
                                  || 0;
        }

        $self->{has_violations} = ($self->{enforcement_state_has_request_violations} || $self->{enforcement_state_has_response_violations}) || 0;
    }

    if ($self->{response}->{injected_deception_trap}) {
        my @traps;
        if (ref($self->{response}->{injected_deception_trap}) eq 'ARRAY') {
            @traps = @{$self->{response}->{injected_deception_trap}};
        }
        elsif (ref($self->{response}->{injected_deception_trap}) eq 'HASH') {
            push(@traps, $self->{response}->{injected_deception_trap});
        }

        my @injected_traps;
        for my $trap (@traps) {
            my $attack_type;
            if ($trap->{attack_type_id}) {
                $attack_type = F5::ASMConfig::Entity::AttackType->new(dbh => $self->{dbh},
                                                                      minimal_load => 1,
                                                                      get_criteria => {
                                                                         attack_type_id => $trap->{attack_type_id},
                                                                      });

            }

            push (@injected_traps, F5::ASMConfig::DeceptionTrap->new(deception_trap_name => $trap->{deception_trap_name},
                                                                     deception_trap_display_name => $trap->{deception_trap_display_name},
                                                                     risk => $trap->{risk},
                                                                     score => $trap->{score},
                                                                     attack_type => $attack_type));
        }

        $self->{injected_deception_traps} = \@injected_traps;
    }
}

#----------------------------------------------------------------------

sub compose_source_ip {
    my ($self) = @_;

    my $protobuf_source_ip = $self->{request}->{source_ip};
    my $source_ip;
    if ( ref $protobuf_source_ip ) {
        $source_ip = F5::IpAddr::ip_ntop(
            pack( "L4", $protobuf_source_ip->{ip0}, $protobuf_source_ip->{ip1}, $protobuf_source_ip->{ip2}, $protobuf_source_ip->{ip3} ) );
        if ( $protobuf_source_ip->{rtdom_id} ) {
            $source_ip .= "%" . $protobuf_source_ip->{rtdom_id};
        }
    }
    return $source_ip;
}

#----------------------------------------------------------------------

sub get_all_request_violations {
    my ($self)   = @_;
    my $request  = $self->{request};
    my $response = $self->{response};

    my @request_violations  = ();
    my @response_violations = ();
    my @excluse_violations  = ();

    # Populate violation state for policy
    for my $mask (qw(learn alarm block)) {
        for my $viol_index ( F5::Violation::Utils::get_indexes_from_viol_set( $request->{ $mask . "_masks" }->{viol_set} ) ) {
            $self->{policy_viol_state}->{$mask}->{$viol_index} = 1;
        }
    }

    for my $entity_type (qw(header cookie param url filetype)) {
        if ( $request->{$entity_type} ) {
            my $subentities = $request->{$entity_type};
            if ( ref($subentities) ne 'ARRAY' ) {
                $subentities = [$subentities];
            }
            for my $subentity (@$subentities) {
                my @violations = $self->get_entity_violations(
                    entity_type  => $entity_type,
                    entity       => $subentity,
                    request      => $request,
                    location     => 'request',
                    exclude_viol => [],
                );
                push( @request_violations, @violations );
                push( @excluse_violations, @violations );
            }
        }
    }

    push(
        @response_violations,
        $self->get_entity_violations(
            entity_type  => 'response',
            entity       => $response,
            request      => $response,
            location     => 'response',
            exclude_viol => [],
        )
    );
    push( @excluse_violations, @response_violations );

    push(
        @request_violations,
        $self->get_entity_violations(
            entity_type  => 'request',
            entity       => $request,
            request      => $request,
            location     => 'request',
            exclude_viol => \@excluse_violations,
        )
    );

    my @viol_on_request = ();
    for my $req_viol (@request_violations) {
        if ( $req_viol->{entity_type} eq 'response' ) {
            push( @response_violations, $req_viol );
        }
        else {
            push( @viol_on_request, $req_viol );
        }
    }

    my %violations_by_viol_index;
    for my $viol_rec (@viol_on_request, @response_violations) {
        if ($viol_rec->{violation} && $viol_rec->{violation}->{viol_index}) {
            $violations_by_viol_index{$viol_rec->{violation}->{viol_index}} = 1;
        }
    }

    # Create truncated records for violations that occurred in the request, but otherwise have no violation record
    # (possibly due to BD truncation or by errant ecard_max_http_req_uri_len setting)
    for my $viol_index (F5::Violation::Utils::get_indexes_from_viol_set($request->{enforced_violations}->{viol_set})) {
        if (! $violations_by_viol_index{$viol_index}) {
            my $violation_rec = {};
            my $violation_obj = F5::ASMConfig::Entity::Violation->new( dbh => $self->{dbh}, get_criteria => { viol_index => $viol_index } );
            $violation_rec->{violation} = $violation_obj;
            $violation_rec->{violation_severity} = $violation_obj->{severity};

            $violation_rec->{is_truncated} = 1;
            push(@viol_on_request, $violation_rec);
        }
    }

    return ( \@viol_on_request, \@response_violations );
}

#----------------------------------------------------------------------

sub get_entity_violations {
    my ( $self, %params ) = @_;

    my $entity       = $params{entity};
    my $entity_type  = $params{entity_type};
    my $request      = $params{request};
    my $location     = $params{location};
    my $exclude_viol = $params{exclude_viol};
    my $context      = $entity_type;

    my $content_profile_info = $entity->{content_profile_info};
    if ( ref($content_profile_info) eq 'HASH' ) {
        $self->prepare_content_profile_data(
            entity               => $entity,
            content_profile_info => $content_profile_info,
        );
    }
    elsif ( ref($content_profile_info) eq 'ARRAY' ) {
        $self->prepare_repeated_content_profile_data(
            entity               => $entity,
            content_profile_info => $content_profile_info,
        );
    }

    my @violations;

    # todo: these apply to param, need to check if they apply for all entities
    my %observed_entity_fields = (
        'name'          => 'observed_entity_name',
        'value'         => 'observed_entity_value',
        'value_len'     => 'observed_entity_value_length',
        'context'       => 'observed_entity_context',
        'is_in_staging' => 'enforcement_state_is_in_staging',
        'is_staging'    => 'enforcement_state_is_in_staging',
        'sensitive'     => 'observed_entity_sensitive',
        'level'         => 'param_level',
        'auto_detected_type' => 'parameter_value_auto_detected_type',
    );

    my %observed_instance_fields = (
        'is_in_staging' => 'enforcement_state_is_in_staging',
        'is_staging'    => 'enforcement_state_is_in_staging',
        'actual_location' => 'observed_entity_context',
        'configured_location' => 'parameter_configured_location',
    );

    my %viol_state_for_field = (
        'violations_status_code' => {},
        'tightening_status_code' => {},
        'enforced_violations'    => '',
        'violations'             => '',
    );

    # Violations may be listed in violations_status_code, tightening_status_code, or in *_violation_info
    my %viol_set_indexes   = ();
    for my $violation_field (keys %viol_state_for_field) {
        if ( $entity->{$violation_field} && $entity->{$violation_field}->{viol_set} ) {
            for my $viol_index ( F5::Violation::Utils::get_indexes_from_viol_set( $entity->{$violation_field}->{viol_set} ) ) {
                $viol_set_indexes{$viol_index} = 1;
                if ( ref($viol_state_for_field{$violation_field}) ) {
                    $viol_state_for_field{$violation_field}->{$viol_index} = 1;
                }
            }
        }
    }

    my $tightening_viol_indexes   = $viol_state_for_field{tightening_status_code};
    my $violations_status_indexes = $viol_state_for_field{violations_status_code};

    # todo: Probably need to pass along snippet, etc.
    my @viol_info_indexes = ();
    for my $violation_info_field ( grep { /_violation_info$/ } keys %$entity ) {
        if ( ref( $entity->{$violation_info_field} ) && ref( $entity->{$violation_info_field} ) eq 'ARRAY' ) {
            my %found_viol_elem = ();

            #in case VIOL_XML_SOAP_METHOD only one instance should be created
            my $soap_method_viol_elem = undef;
            my @info_elements = @{ $entity->{$violation_info_field} };
            for (my $ii=0; $ii<scalar @info_elements; $ii++) {
                my $viol_info_elem = $info_elements[$ii];
                my $elem_viol_index = $viol_info_elem->{viol_index};
                if ($elem_viol_index == F5::Violation::VIOL_XML_SOAP_METHOD()) {
                    next if $soap_method_viol_elem;
                    $soap_method_viol_elem = $viol_info_elem;
                }
                if ($elem_viol_index) {
                    my $handling = $viol_handling{$elem_viol_index};
                    if ( ref($handling) && $found_viol_elem{$elem_viol_index} && ($handling->{multi_occurrence} || $handling->{entity_instances_creator}) ) {
                        next;
                    }
                    $found_viol_elem{$elem_viol_index} = 1;
                    delete $viol_set_indexes{$elem_viol_index};
                    if ( $handling->{change_info_elem} ) {
                        my $new_info_elems = $entity->{$handling->{change_info_elem}};
                        if ( ref($new_info_elems) eq 'ARRAY' ) {
                            $viol_info_elem = $new_info_elems->[$ii];
                            next unless ref($viol_info_elem);
                        }
                    }
                    push( @viol_info_indexes, [ $elem_viol_index, $viol_info_elem ] );
                }
            }
        }
    }

    my @viol_indexes = map { [ $_, undef ] } keys %viol_set_indexes;
    push( @viol_indexes, @viol_info_indexes );

    my %exclude_ids = ();
    for my $exc_viol (@$exclude_viol) {
        my $viol_index = $exc_viol->{viol_index};
        if ( !$viol_handling{$viol_index}->{look_in_request_entity} ) {
            $exclude_ids{$viol_index} = 1;
        }
    }

    @viol_indexes = grep { !exists $exclude_ids{ $_->[0] } } @viol_indexes;
    my $entities_by_order_count;
    for my $viol_pair (@viol_indexes) {
        my ( $viol_index, $viol_info_elem ) = @$viol_pair;
        my $viol_handling = $viol_handling{$viol_index} || {};
        my $viol_error_desc = $viol_info_elem->{error_desc};

        if ( $viol_handling->{entity_type} ) {
            if ( ref( $viol_handling->{entity_type} ) eq 'ARRAY' ) {
                my $type_found = 0;
                for my $type ( @{ $viol_handling->{entity_type} } ) {
                    if ( $entity_type eq $type ) {
                        $type_found = 1;
                        last;
                    }
                }
                next if !$type_found;
            }
            elsif ( $entity_type ne $viol_handling->{entity_type} ) {
                next;
            }
        }

        my @instances = ($entity);    # Exactly 1 instance of this violation
        if ( $viol_handling->{multi_occurrence} ) {
            if ( ref( $entity->{ $viol_handling->{multi_occurrence} } ) eq 'ARRAY' ) {
                @instances = @{ $entity->{ $viol_handling->{multi_occurrence} } };
            }
            else {
                @instances = ();      # There are no instances of repeated element -- therefore, no violations to add
            }
        }
        elsif ( $viol_handling->{entity_instances_creator} ) {
            my $instances_creator_func = $viol_handling->{entity_instances_creator};
            @instances = $self->$instances_creator_func(
                entity => $entity,
                viol_index => $viol_index,
            );

            for my $inst_field (keys %observed_instance_fields) {
                if (defined($entity->{$inst_field})) {
                    for my $inst (@instances) {
                        $inst->{$inst_field} = $entity->{$inst_field} if ! defined($inst->{$inst_field});
                    }
                }
            }

            # avoid entity fields pollution
            if (defined $entity->{_transient_fields}) {
                for my $entity_field (@{$entity->{_transient_fields}}) {
                    delete $entity->{$entity_field};
                }
                delete $entity->{_transient_fields};
            }
        }
        elsif ( ref($viol_info_elem) && ref( $viol_info_elem->{snippet} ) eq 'HASH' ) {
            @instances = ($viol_info_elem);
        }
        elsif ( ref( $viol_handling->{info_multi_occurrence_path} ) ) {
            my $path_elem;
            if (ref($viol_info_elem) && $viol_info_elem->{$viol_handling->{info_multi_occurrence_path}->[0]}) {
                $path_elem = $viol_info_elem;
            }
            else {
                $path_elem = $entity;
            }
            my $found_elem;
            for my $entry_name ( @{ $viol_handling->{info_multi_occurrence_path} } ) {
                $found_elem = $path_elem->{$entry_name};
                if ( ref( $path_elem->{$entry_name} ) eq 'HASH' ) {
                    $path_elem = $path_elem->{$entry_name};
                }
                elsif ( ref( $path_elem->{$entry_name} ) eq 'ARRAY' ) {
                    $path_elem = $path_elem->{$entry_name}[0];
                }
                else {
                    last;
                }
            }

            if ( ref($found_elem) eq 'ARRAY' ) {
                @instances = @$found_elem;
            }
            else {
                @instances = ();
            }
        }
        elsif ( $viol_handling->{instances_creator} ) {
            my $instances_creator_func = $viol_handling->{instances_creator};
            @instances = $self->$instances_creator_func(
                entity         => $entity,
                viol_info_elem => $viol_info_elem,
            );
        }

        for my $instance (@instances) {
            my $violation_rec = { viol_index => $viol_index, error_desc => $viol_error_desc };

            $violation_rec->{enforcement_state_is_blocked} = $self->{policy_viol_state}->{block}->{$viol_index} ? 1 : 0;
            $violation_rec->{enforcement_state_is_alarmed} = $self->{policy_viol_state}->{alarm}->{$viol_index} ? 1 : 0;
            $violation_rec->{enforcement_state_is_learned} = $self->{policy_viol_state}->{learn}->{$viol_index} ? 1 : 0;

            if ( ref($instance) eq 'HASH' && $instance->{is_detected_on_resp} ) {
                $violation_rec->{location}    = 'response';
                $violation_rec->{entity_type} = 'response';
            }
            else {
                $violation_rec->{location}    = $location;
                $violation_rec->{entity_type} = $context;
            }

            if ( ref $viol_info_elem  ) {
                if ( ref $viol_handling->{viol_elem_map} ) {
                    for my $viol_elem_key ( keys %{ $viol_handling->{viol_elem_map} } ) {
                        $violation_rec->{$viol_elem_key} = $viol_info_elem->{ $viol_handling->{viol_elem_map}->{$viol_elem_key} };
                    }
                } else {
                    my %custom_violation_keys_map = (                    
                        user_defined_violation_key => 'custom_key',
                        user_defined_violation_value => 'custom_value', 
                    );
                    for my $viol_elem_key ( keys %custom_violation_keys_map ) {
                        $violation_rec->{$viol_elem_key} = $viol_info_elem->{ $custom_violation_keys_map{$viol_elem_key} };
                    }
                }
            }

            if ( $viol_handling->{extract_content_profile} && ref($content_profile_info) eq 'HASH' ) {
                $self->set_content_profile_data(
                    content_profile_info     => $content_profile_info,
                    violation_rec            => $violation_rec,
                    entity_type              => $entity_type,
                    entity                   => $entity,
                    content_profile_elements => $viol_handling->{content_profile_elements},
                    instance                 => $instance,
                );
            }

            my $instance_content_profile;
            if (ref($instance) eq 'HASH' && $instance->{from_content_profile} && $instance->{content_profile_info}) {
                $instance_content_profile = $instance->{content_profile_info};
            }
            else {
                $instance_content_profile = ref($content_profile_info) eq 'HASH' ? $content_profile_info :
                                            ref($instance) eq 'HASH' ? $instance->{content_profile_info} : undef;
            }

            if ( $viol_handling->{extract_profile_on_profile_entity} && ref($instance_content_profile) eq 'HASH' && $instance->{from_content_profile}) {
                $self->set_content_profile_data(
                    content_profile_info     => $instance_content_profile,
                    violation_rec            => $violation_rec,
                    entity_type              => $entity_type,
                    entity                   => $entity,
                    content_profile_elements => $viol_handling->{content_profile_elements},
                    instance                 => $instance,
                );
            }

            #TODO: content profile handling should be redesigned
            #may be content profile should be handled at level of entity
            if ( $viol_handling->{extract_content_profile} && ref($content_profile_info) eq 'ARRAY'
                 && !($instance->{from_content_profile} && $instance->{content_profile_info})) {
                $self->set_repeated_content_profile_data(
                    content_profile_info     => $content_profile_info,
                    violation_rec            => $violation_rec,
                    entity_type              => $entity_type,
                    entity                   => $entity,
                    content_profile_elements => $viol_handling->{content_profile_elements},
                    instance                 => $instance,
                );
            }

            # todo: must be cached
            my $violation_obj = F5::ASMConfig::Entity::Violation->new( dbh => $self->{dbh}, get_criteria => { viol_index => $viol_index } );
            if ( !$violation_obj ) {
                next;
            }

            $violation_rec->{violation}          = $violation_obj;
            $violation_rec->{violation_severity} = $violation_obj->{severity};

            if ( ! $viol_handling->{non_entity_violation}) {
                for my $observed_entity_field ( keys %observed_entity_fields ) {
                    if ( defined $entity->{$observed_entity_field} ) {
                        $violation_rec->{ $observed_entity_fields{$observed_entity_field} } = $entity->{$observed_entity_field};
                    }
                }

                for my $observed_instance_field ( keys %observed_instance_fields ) {
                    if ( defined $entity->{$observed_instance_field} ) {
                        $violation_rec->{ $observed_instance_fields{$observed_instance_field} } = $entity->{$observed_instance_field};
                    }
                }

                if (ref($instance)) {
                    for my $observed_instance_field ( keys %observed_instance_fields ) {
                        if ( defined $instance->{$observed_instance_field} ) {
                            if ($observed_instance_field eq 'actual_location' &&
                                $instance->{$observed_instance_field} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_HEADER()) {
                                    $violation_rec->{actual_header_parameter} = 1;
                                    
                            }
                            elsif ($observed_instance_field eq 'actual_location' &&
                                $instance->{$observed_instance_field} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE()) {
                                    $violation_rec->{actual_cookie_parameter} = 1;
                                    
                            }
                            elsif ($observed_instance_field eq 'configured_location' &&
                                $instance->{$observed_instance_field} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_HEADER()) {
                                    $violation_rec->{configured_header_parameter} = 1;
                                    
                            }
                            elsif ($observed_instance_field eq 'configured_location' &&
                                $instance->{$observed_instance_field} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE()) {
                                    $violation_rec->{configured_cookie_parameter} = 1;
                                    
                            }
                            else {
                                $violation_rec->{ $observed_instance_fields{$observed_instance_field} } = $instance->{$observed_instance_field};
                            }
                        }
                    }
                }
            }

            if ( $viol_handling->{field_map} ) {
                for my $k ( keys %{ $viol_handling->{field_map} } ) {
                    if ( $k eq 'instance' ) {
                        # Special case -- instance value itself is significant
                        $violation_rec->{ $viol_handling->{field_map}->{$k} } = $instance;
                    }
                    elsif ( ref($instance) && exists $instance->{$k} ) {
                        $violation_rec->{ $viol_handling->{field_map}->{$k} } = $instance->{$k};
                    }
                    elsif ( exists $entity->{$k} ) {
                        $violation_rec->{ $viol_handling->{field_map}->{$k} } = $entity->{$k};
                    }
                }
            }

            if ( $viol_handling->{field_path_map} ) {
                $self->extract_path_map( violation_rec => $violation_rec, path_map => $viol_handling->{field_path_map}, entity => $entity );
            }

            if ( $viol_handling->{request_path_map} ) {
                $self->extract_path_map( violation_rec => $violation_rec, path_map => $viol_handling->{request_path_map}, entity => $request );
            }

            if ( ref($instance) && ref( $instance->{snippet} ) eq 'HASH' ) {
                $violation_rec->{snippet_buffer} = $instance->{snippet}->{buffer};
                $violation_rec->{snippet_offset} = $instance->{snippet}->{offset};
                $violation_rec->{snippet_length} = $instance->{snippet}->{length};
            }

            if ( $viol_handling->{extra_field_handler} ) {
                my $extra_field_handler_func = $viol_handling->{extra_field_handler};
                my $valid_violation          = $self->$extra_field_handler_func(
                    %params,
                    viol_index                => $viol_index,
                    violation_rec             => $violation_rec,
                    instance                  => $instance,
                    entity                    => $entity,
                    entity_type               => $entity_type,
                    viol_info_elem            => $viol_info_elem,
                    tightening_viol_indexes   => $tightening_viol_indexes,
                    violations_status_indexes => $violations_status_indexes,
                );
                next unless $valid_violation;
            }

            if ( $viol_handling->{ignore_entity_staging} ) {
                $violation_rec->{enforcement_state_is_in_staging} = undef;
            }

            if ( !defined( $violation_rec->{enforcement_state_is_in_staging} ) ) {
                $violation_rec->{enforcement_state_is_in_staging} = $tightening_viol_indexes->{$viol_index} ? 1 : 0;
            }
            elsif ( $violation_rec->{enforcement_state_is_in_staging} ) {
                $violation_rec->{_explict_staging} = 1;
                #this case when violation instance has explicit "in_staging" flag
            }

            if ($viol_index == F5::Violation::VIOL_ATTACK_SIGNATURE() &&
                ! defined($violation_rec->{signature_is_staging})) {
                $violation_rec->{signature_is_staging} = $violation_rec->{enforcement_state_is_in_staging};
            }

            if ( $violation_rec->{enforcement_state_is_in_staging} ) {
                $violation_rec->{enforcement_state_is_blocked} = 0;
                $violation_rec->{enforcement_state_is_alarmed} = 0;
            }

            if ( $entity->{wildcard}  && ! $viol_handling->{non_entity_violation}) {
                $self->set_policy_entity(
                    violation_rec => $violation_rec,
                    wildcard      => $entity->{wildcard},
                    staging       => $violation_rec->{enforcement_state_is_in_staging},
                    entity_type   => $entity_type,
                );
            }

            if ($entity_type eq 'request') {
                my $url_entity_type = $entity_type;
                if ($self->{bd_data}->{is_websockets}) {
                    $url_entity_type = 'websocketUrl';
                    $violation_rec->{entity_type} = $url_entity_type;
                }

                if ($entity->{url}->{wildcard}) {
                    if (! $violation_rec->{policyEntity}) {
                        $self->set_policy_entity(
                            violation_rec => $violation_rec,
                            wildcard      => $entity->{url}->{wildcard},
                            staging       => $violation_rec->{enforcement_state_is_in_staging},
                            entity_type   => $url_entity_type,
                        );
                    }
                    if (! defined($violation_rec->{observed_entity_name})) {
                        $violation_rec->{observed_entity_name} = $entity->{url}->{name};
                        $violation_rec->{observed_entity_method} = $entity->{url}->{method} if $entity->{url}->{method};
                    }
                }
            }

            if ( $viol_handling->{block_violation_entry} ) {
                for my $block_violation_entry ( @{ $viol_handling->{block_violation_entry} } ) {
                    delete $violation_rec->{$block_violation_entry};
                }
            }

            push( @violations, $violation_rec );
        }
    }

    #real violation has at least one flag 'ON'
    #staging flag does not affect on "real violation" decision
    my @real_violations = grep {
        $_->{enforcement_state_is_blocked} ||
        $_->{enforcement_state_is_alarmed} ||
        $_->{enforcement_state_is_learned} ||
        $_->{_explict_staging}
    } @violations;

    return @real_violations;
}

#----------------------------------------------------------------------

sub set_policy_entity {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};
    my $wildcard      = $params{wildcard};
    my $staging       = $params{staging};
    my $entity_type   = $params{entity_type};

    $violation_rec->{policy_entity_staging} = $staging;
    return unless $entity_type;

    $entity_type = $entity_type eq 'request' ? 'url' : $entity_type;
    my ($entity_class_name, $entity_name_string) = @{$entity_type_class_map{$entity_type}};

    return unless $entity_class_name;

    my $get_criteria = { $entity_name_string => $wildcard, is_wildcard => 1 };
    my $method_name = $F5::ASMConfig::Entity::URL::ANY_URL_METHOD_NAME;
    if ($self->{request}->{url} && $self->{request}->{url}->{method}) {
        $method_name = $self->{request}->{url}->{method};
    }
    if ($entity_type eq 'url' || $entity_type eq 'websocketUrl') {
        $get_criteria->{app_id} = $self->{request}->{url}->{app_id};
        if ($entity_type eq 'url') {
            $get_criteria->{method_name} = $method_name;
        }
    }

    if (defined $violation_rec->{param_level}) {
        $get_criteria->{param_level} = $violation_rec->{param_level};
        if ($violation_rec->{param_level} == F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL) {
            $get_criteria->{url} = F5::ASMConfig::Entity::URL->new( dbh => $self->{dbh},
                                                                    policy => $self->{request_policy},
                                                                    object => {url_name => $self->{request}->{url}->{name},
                                                                               method_name => $method_name,
                                                                                     app_id => $self->{request}->{url}->{app_id}, } );
        }
    }

    delete $violation_rec->{policy_entity_staging};
    my $minimal_load = !F5::Utils::Rest::is_expanded(expand_fields => $self->{expand_fields},
                                                     path => 'violations/policyEntity');

    $violation_rec->{policyEntity} = $entity_class_name->new( dbh => $self->{dbh},
                                                              policy => $self->{request_policy},
                                                              logger => $self->{logger},
                                                              minimal_load => $minimal_load,
                                                              get_criteria => $get_criteria );
}

#----------------------------------------------------------------------

sub set_repeated_content_profile_data {
    my ( $self, %params ) = @_;

    my $repeated_content_profile_info = $params{content_profile_info};
    for my $content_profile_info (@$repeated_content_profile_info) {
        $self->set_content_profile_data(%params, content_profile_info => $content_profile_info);
    }
}

#----------------------------------------------------------------------

sub set_content_profile_data {
    my ( $self, %params ) = @_;

    my $content_profile_info     = $params{content_profile_info};
    my $violation_rec            = $params{violation_rec};
    my $entity                   = $params{entity};
    my $entity_type              = $params{entity_type};
    my $content_profile_elements = $params{content_profile_elements};
    my $instance                 = $params{instance};

    $violation_rec->{content_profile} = $self->get_policy_content_profile(content_profile_info => $content_profile_info);
    $violation_rec->{content_profile_type} = $content_profile_info->{content_profile_type};
    if ( ref $content_profile_info->{xml_violation_info} ) {
        my @violation_info_list = grep { ($_->{viol_index} == $violation_rec->{viol_index}) && ($_->{error_desc} eq $violation_rec->{error_desc}) } @{$content_profile_info->{xml_violation_info}};
        if ($violation_rec->{viol_index} == F5::Violation::VIOL_PARAMETER_VALUE_METACHAR() ||
            $violation_rec->{viol_index} == F5::Violation::VIOL_PARAMETER_NAME_METACHAR() ||
            $violation_rec->{viol_index} == F5::Violation::VIOL_URL_METACHAR() ||
            $violation_rec->{viol_index} == F5::Violation::VIOL_HEADER_METACHAR()
        ) {
            my @metachar_violation_infos = grep {$_->{illegal_metachar} == $instance->{metacharacter}} @violation_info_list;
            if (@metachar_violation_infos) {
                @violation_info_list = @metachar_violation_infos;
            }
        }

        my ($xml_violation_info) = @violation_info_list;
        if ($xml_violation_info) {        
            $violation_rec->{content_profile_location_desc} = $xml_violation_info->{location_desc};
            $violation_rec->{snippet_buffer}                = $xml_violation_info->{data_buffer};
            if ($content_profile_elements->{observed_entity_soap_method}) {
                $violation_rec->{observed_entity_soap_method} = $xml_violation_info->{method_name};
            }
            if ($content_profile_elements->{observed_entity_soap_namespace}) {
                $violation_rec->{observed_entity_soap_namespace} = $xml_violation_info->{namespace_};
            }
            my $violation_description = join( "\n", grep { $_ } $xml_violation_info->{error_desc}, $xml_violation_info->{fault_details} );
            if ($violation_description) {
                $violation_rec->{observed_entity_violation_description} = $violation_description;
            }
        }
    }

    my %def_viol_info = (
        'defense_element_allowed_value'  => [ 'defense_violation_info', 'defense_element_allowed_value_str' ],
        'defense_element_name'           => [ 'defense_violation_info', 'defense_element_name' ],
        'defense_element_name_xpointer'  => [ 'defense_violation_info', 'defense_element_name_xpointer' ],
        'defense_element_value'          => [ 'defense_violation_info', 'defense_element_value_str' ],
    );

    if ($content_profile_elements) {
        for my $viol_info_key ( keys %def_viol_info ) {
            if ( !$content_profile_elements->{$viol_info_key} ) {
                delete $def_viol_info{$viol_info_key};
            }
        }
    }

    $self->extract_path_map( violation_rec => $violation_rec, path_map => \%def_viol_info, entity => $content_profile_info );

    $self->check_entity_type_request(
        violation_rec => $violation_rec,
        entity_type   => $entity_type,
        entity        => $entity,
    );

}

#----------------------------------------------------------------------

sub get_policy_content_profile {
    my ( $self, %params ) = @_;

    my $content_profile_info = $params{content_profile_info};

    my $profile_get_criteria;
    if ($content_profile_info->{profile_name} && $content_profile_info->{content_profile_type}) {
        $profile_get_criteria = {
            profile_name => $content_profile_info->{profile_name},
            content_profile_type => $content_profile_info->{content_profile_type},
        };
    }
    else {
        $profile_get_criteria = {
            profile_id => $content_profile_info->{content_profile_id},
        };
    }

    #TODO: should be moved to bd config or asmlogd
    my $minimal_load = !F5::Utils::Rest::is_expanded(expand_fields => $self->{expand_fields},
                                                     path => 'violations/contentProfile');
    my $content_profile = F5::ASMConfig::Entity::ContentProfile->new(
        dbh => $self->{dbh},
        logger => $self->{logger},
        policy => $self->{request_policy},
        minimal_load => $minimal_load,
        get_criteria => $profile_get_criteria,
    );

    return $content_profile;
}

#----------------------------------------------------------------------

sub prepare_repeated_content_profile_data {
    my ( $self, %params ) = @_;

    my $repeated_content_profile_info = $params{content_profile_info};
    for my $content_profile_info (@$repeated_content_profile_info) {
        $self->prepare_content_profile_data(%params, content_profile_info => $content_profile_info);
    }
}

#----------------------------------------------------------------------

sub prepare_content_profile_data {
    my ( $self, %params ) = @_;

    my $entity               = $params{entity};
    my $content_profile_info = $params{content_profile_info} || $entity->{content_profile_info};
    for my $content_profile_entry ( keys %$content_profile_info ) {
        if ( $content_profile_entry eq 'metachar' && ref( $content_profile_info->{$content_profile_entry} ) eq 'HASH' ) {
            for my $metachar_entry ( keys %{ $content_profile_info->{$content_profile_entry} } ) {
                $entity->{$content_profile_entry} = $content_profile_info->{$content_profile_entry}->{$metachar_entry};
            }
        }
        else {
            if ( ref($content_profile_info->{$content_profile_entry}) eq 'ARRAY' ) {
                if (! exists $entity->{$content_profile_entry}) {
                    $entity->{$content_profile_entry} = [];
                }
                elsif(ref($entity->{$content_profile_entry}) eq 'HASH') {
                    my $exist_entity_elem = $entity->{$content_profile_entry};
                    $entity->{$content_profile_entry} = [];
                    push ( @{$entity->{$content_profile_entry}}, $exist_entity_elem); 
                }

                for my $content_profile_multi_entry ( @{$content_profile_info->{$content_profile_entry}} ) {
                    if (ref($content_profile_multi_entry) eq 'HASH') {
                        $content_profile_multi_entry->{from_content_profile} = 1;
                    }
                    push( @{$entity->{$content_profile_entry}}, $content_profile_multi_entry);
                }
            }
            else {
                $entity->{$content_profile_entry} = $content_profile_info->{$content_profile_entry};
            }
        }
    }
}

#----------------------------------------------------------------------

sub extract_path_map {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};
    my $path_map      = $params{path_map};
    my $entity        = $params{entity};

    for my $rec_entry_name ( keys %$path_map ) {
        my @entity_value_path = @{ $path_map->{$rec_entry_name} };
        my $curr_path_elem    = $entity;
        for my $path_elem (@entity_value_path) {
            if ( ref($curr_path_elem) eq 'ARRAY' ) {
                $curr_path_elem = $curr_path_elem->[0];
            }
            if ( ref($curr_path_elem) ) {
                $curr_path_elem = $curr_path_elem->{$path_elem};
            }
            else {
                last;
            }
        }
        if ( defined $curr_path_elem ) {
            $violation_rec->{$rec_entry_name} = $curr_path_elem;
        }
    }
}

#----------------------------------------------------------------------

sub get_signature_fields {
    my ( $self, %params ) = @_;

    my $sig                     = $params{instance};
    my $violation_rec           = $params{violation_rec};
    my $viol_info_elem          = $params{viol_info_elem};
    my $entity                  = $params{entity};
    my $tightening_viol_indexes = $params{tightening_viol_indexes};
    my $viol_index              = $params{viol_index};

    return unless ref($sig);

    # sig_id_orig should have a real signature ID even if the previous enforced rule was caught
    my $minimal_load = !F5::Utils::Rest::is_expanded(expand_fields => $self->{expand_fields},
                                                     path => 'violations/signature');
    $violation_rec->{signature} = F5::ASMConfig::Entity::Signature->new( dbh => $self->{dbh},
                                                                         minimal_load => $minimal_load,
                                                                         get_criteria => { sig_id => $sig->{sig_id_orig} } );

    return undef unless $violation_rec->{signature};

    if (exists($entity->{is_staging})) {
        $violation_rec->{observed_entity_is_staging} = $entity->{is_staging};
    }
    elsif (exists($entity->{url}) && exists($entity->{url}->{is_in_staging})) {
        $violation_rec->{observed_entity_is_staging} = $entity->{url}->{is_in_staging};
    }
    elsif ($entity->{is_violation}) {
        $violation_rec->{observed_entity_is_staging} = 0;
    }

    if($sig->{sig_id} != $sig->{sig_id_orig}) { # Previous enforced rule was caught 
        $violation_rec->{signature_is_staging} = 0;
    }
    elsif ( $violation_rec->{signature}{flg_has_valuecontent_http} ||
         $violation_rec->{signature}{flg_has_valuecontent_gwt} ||
         $violation_rec->{signature}{flg_has_valuecontent_json} ||
         $violation_rec->{signature}{flg_has_valuecontent_xml} ||
         $violation_rec->{signature}{flg_has_valuecontent_plain_text}
         ) {

        if (defined($sig->{is_staging})) {
            $violation_rec->{signature_is_staging} = $sig->{is_staging};
        }
        elsif (ref($tightening_viol_indexes) && defined($tightening_viol_indexes->{$viol_index})) {
            $violation_rec->{signature_is_staging} = $tightening_viol_indexes->{$viol_index} ? 1 : 0;
        } 
        elsif (ref($entity) && defined($entity->{is_staging})) {
            $violation_rec->{signature_is_staging} = $entity->{is_staging};
        }
        elsif ( defined($violation_rec->{policy_entity_staging}) ) {
            $violation_rec->{signature_is_staging} = $violation_rec->{policy_entity_staging};
        }
    }
    else {
        $violation_rec->{signature_is_staging} = $sig->{is_staging} //
                                                 $sig->{is_in_staging} //
                                                 $entity->{is_staging} //
                                                 $entity->{is_in_staging};
    }
    $violation_rec->{enforcement_state_is_in_staging} = $violation_rec->{signature_is_staging} ? 1 : undef;
    if ($sig->{is_violation}) {
        $violation_rec->{enforcement_state_is_in_staging} = 0;
        $violation_rec->{enforcement_state_is_likely_false_positive} = 0;
    }
    else {
        $violation_rec->{enforcement_state_is_likely_false_positive} = ($sig->{is_likely_false_positive}) ? 1 : 0;
    }

    if (defined $sig->{block_mask}) {
        my @block_flag_entries = ( 'enforcement_state_is_blocked', 'enforcement_state_is_alarmed', 'enforcement_state_is_learned' );
        my $block_vec = Bit::Vector->new_Dec( 32, $sig->{block_mask} );
        my @vector_array;
        for my $pos ( $block_vec->Index_List_Read ) {
            $vector_array[$pos] = 1;
        }
        for my $block_flag_pos ( 0..$#block_flag_entries ) {
            my $block_flag_entry = $block_flag_entries[$block_flag_pos];
            if ($vector_array[$block_flag_pos]) {
                $violation_rec->{ $block_flag_entry } = 1;
            }
            else {
                $violation_rec->{ $block_flag_entry } = 0;
            }
        }
    }

    if (defined $sig->{context} && $sig->{context} !~ /profile/i) {
        $violation_rec->{entity_type} = $sig->{context};
    }

    my @snippets = ();
    if ( ref( $sig->{snippet} ) eq 'ARRAY' ) {
        for my $sig_snippet ( @{ $sig->{snippet} } ) {
            my $snip = {};
            $snip->{snippet_buffer} = $sig_snippet->{buffer};
            $snip->{snippet_offset} = $sig_snippet->{offset};
            $snip->{snippet_length} = $sig_snippet->{length};
            push( @snippets, $snip );
        }
    }
    elsif ( ref( $sig->{snippet} ) eq 'HASH' ) {
        my $snip = {};
        $snip->{snippet_buffer} = $sig->{snippet}->{buffer};
        $snip->{snippet_offset} = $sig->{snippet}->{offset};
        $snip->{snippet_length} = $sig->{snippet}->{length};
        push( @snippets, $snip );
    }
    elsif( $sig->{viol_desc} eq "Matched buffer may include sensitive data" ) {
        my $snip = {};
        $snip->{snippet_sensitive} = 1;
        $snip->{snippet_buffer} = '****';
        push( @snippets, $snip );
    }

    if ( scalar @snippets ) {
        $violation_rec->{attack_signature_snippets} = \@snippets;
        $violation_rec->{snippet_buffer} = $snippets[0]->{snippet_buffer};
        $violation_rec->{snippet_offset} = $snippets[0]->{snippet_offset};
        $violation_rec->{snippet_length} = $snippets[0]->{snippet_length};
        $violation_rec->{snippet_sensitive} = $snippets[0]->{snippet_sensitive};
    }

    return 1;
}

#----------------------------------------------------------------------

sub threat_campaign_violations {
    my ( $self, %params ) = @_;

    my $campaign      = $params{instance};
    my $violation_rec = $params{violation_rec};
    my $entity        = $params{entity};

    return unless ref($campaign);

    if (exists $campaign->{is_staging}) {
        $violation_rec->{enforcement_state_is_in_staging} = $campaign->{is_staging};
    }
    elsif (exists $campaign->{is_violation}) {
        $violation_rec->{enforcement_state_is_in_staging} = ! $campaign->{is_violation};
    }

    my $minimal_load = !F5::Utils::Rest::is_expanded(expand_fields => $self->{expand_fields},
                                                     path => 'violations/threatCampaign');
    my $campaign_entity = F5::ASMConfig::Entity::ThreatCampaign->new( dbh => $self->{dbh},
                                                minimal_load => $minimal_load,
                                                get_criteria => {
                                                    name => $campaign->{name},
                                                });

    return unless ref($campaign_entity);

    $violation_rec->{observed_entity_name} = $campaign_entity->{display_name};
    $violation_rec->{threat_campaign} = $campaign_entity;

    return 1;
}

#----------------------------------------------------------------------

sub get_wss_subviolations {
    my ( $self, %params ) = @_;

    my $instance       = $params{instance};
    my $violation_rec  = $params{violation_rec};
    my $xml_violation_info = $params{viol_info_elem};

    my %wss_subviolations = %{ $self->cache_subviolations( 'F5::ASMConfig::Entity::WssSubviolation', 'subviolation_code' ) };

    my $error_code = $xml_violation_info->{error_code};
    $violation_rec->{wss_subviolation}                      = $wss_subviolations{$error_code};
    $violation_rec->{observed_entity_violation_description} = $xml_violation_info->{error_desc};
    if ($xml_violation_info->{context}) {
        $violation_rec->{entity_type} = lc $xml_violation_info->{context};
    }

    return undef unless $violation_rec->{wss_subviolation};
    $violation_rec->{wss_subviolation}->{dbh} = $self->{dbh};

    return 1;
}

#----------------------------------------------------------------------

sub get_http_subviolations {
    my ( $self, %params ) = @_;

    my $instance       = $params{instance};
    my $violation_rec  = $params{violation_rec};
    my $viol_info_elem = $params{viol_info_elem};

    my %http_subviolations = %{ $self->cache_subviolations( 'F5::ASMConfig::Entity::HttpCheck', 'http_check_code' ) };

    return undef unless $viol_info_elem->{subviolation_status};
    $violation_rec->{http_subviolation} = $http_subviolations{ $viol_info_elem->{subviolation_status} };

    return undef unless $violation_rec->{http_subviolation};
    $violation_rec->{http_subviolation}->{dbh} = $self->{dbh};

    if ( $viol_info_elem->{subviolation_status} == 131072 ) {    #exceed num of headers
        if ( $viol_info_elem->{viol_desc} =~ /^Number of headers (\d+?) exceeded maximum limit of: (\d+)/ ) {
            $violation_rec->{expected_value} = $2;
            $violation_rec->{actual_value}   = $1;
        }
    }
    elsif ( $viol_info_elem->{subviolation_status} == 2 ) {      #Header name with no header value
        if ( $viol_info_elem->{viol_desc} =~ /^Header '(.+)' has no value/ ) {
            $violation_rec->{actual_value} = $1;
        }
    }
    elsif ( $viol_info_elem->{subviolation_status} == 2048 ) {    #Host header with IP value
        if ( $viol_info_elem->{viol_desc} =~ /^Host header with IP value: (.+)/ ) {
            $violation_rec->{actual_value} = $1;
        }
    }
    elsif ( $viol_info_elem->{subviolation_status} == 8192 ) {    #Bad HTTP version
        if ( $viol_info_elem->{viol_desc} =~ /^Bad HTTP version: (.+)/ ) {
            $violation_rec->{actual_value} = $1;
        }
    }
    elsif ( $viol_info_elem->{subviolation_status} == 262144 ) {    #Bad Host header
        if ( $viol_info_elem->{viol_desc} =~ /^Host header: (.+)/ ) {
            $violation_rec->{actual_value} = $1;
        }
    }
    elsif ( $viol_info_elem->{subviolation_status} == 524288 ) {    #exceed num of params
        if ( $viol_info_elem->{viol_desc} =~ /^Number of total parameters <(\d+?)> in/ ) {
            $violation_rec->{actual_value} = $1;
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub cache_subviolations {
    my ( $self, $subviolation_type, $code_name ) = @_;

    my $subviolation_type_cache = $subviolations_cache->{$subviolation_type};
    if ( !$subviolation_type_cache ) {
        $subviolation_type_cache = {};
        my @subviolations = $subviolation_type->retrieve_all( dbh => $self->{dbh}, logger => $self->{logger}, );
        for my $subv (@subviolations) {
            $subviolation_type_cache->{ $subv->{$code_name} } = $subv;
        }
        $subviolations_cache->{$subviolation_type} = $subviolation_type_cache;
    }

    return $subviolation_type_cache;
}

#----------------------------------------------------------------------

sub get_evasion_subviolations {
    my ( $self, %params ) = @_;

    my $entity        = $params{entity};
    my $instance      = $params{instance};
    my $violation_rec = $params{violation_rec};

    my $evasion_context = $instance->{context};
    if ($evasion_context) {
        if ( $evasion_context eq 'qs' ) {
            $violation_rec->{entity_type} = 'query-string';
            $violation_rec->{observed_entity_context} = 0;
        }
        elsif ( $evasion_context eq 'param_qs' ) {
            $violation_rec->{entity_type} = 'param';
            $violation_rec->{observed_entity_context} = 0;
        }
        elsif ( $evasion_context eq 'pd' ) {
            $violation_rec->{entity_type} = 'post-data';
            $violation_rec->{observed_entity_context} = 1;
        }
        elsif ( $evasion_context eq 'param_pd' ) {
            $violation_rec->{entity_type} = 'param';
            $violation_rec->{observed_entity_context} = 1;
        }
        elsif ( $evasion_context =~ /websocket/i ) {
            $violation_rec->{entity_type} = 'websocketUrl';
        }
    }

    if (defined $entity->{level}) {
        $violation_rec->{param_level} = $entity->{level};
    }

    my %subviolations = %{ $self->cache_subviolations( 'F5::ASMConfig::Entity::Normalization', 'normalization_code' ) };
    $violation_rec->{normalization_subviolation} = $subviolations{ $instance->{detected_evasion_bit} };

    return undef unless $violation_rec->{normalization_subviolation};
    $violation_rec->{normalization_subviolation}->{dbh} = $self->{dbh};

    return 1;
}

#----------------------------------------------------------------------

sub get_format_violations_instances {
    my ( $self, %params ) = @_;

    my $entity     = $params{entity};
    my $viol_index = $params{viol_index};

    my @instances;
    my @content_profiles;
    if (ref($entity->{content_profile_info}) eq 'HASH') {
        @content_profiles = ($entity->{content_profile_info});
    }
    elsif (ref($entity->{content_profile_info}) eq 'ARRAY') {
        @content_profiles = @{$entity->{content_profile_info}};
    }

    for my $content_profile_entry (@content_profiles) {
        my $starting_xml_violation_offset;
        next unless ref($content_profile_entry->{xml_violation_info}) eq 'ARRAY';
        my @xml_violation_infos = @{$content_profile_entry->{xml_violation_info}};
        for (my $idx=0; $idx<scalar @xml_violation_infos; ++$idx) {
            if ($xml_violation_infos[$idx]->{viol_index} == $viol_index) {
                $starting_xml_violation_offset = $idx;
                last;
            }
        }

        next unless defined($starting_xml_violation_offset);
        if ( ref($content_profile_entry->{defense_violation_info}) eq 'ARRAY') {
            my @defense_violation_info = @{$content_profile_entry->{defense_violation_info}};
            for (my $i=0; $i<scalar @defense_violation_info; ++$i) {
                my $xml_violation_info = $content_profile_entry->{xml_violation_info}[$i+$starting_xml_violation_offset] ||
                                         $content_profile_entry->{xml_violation_info}[-1];

                my $instance = dclone($defense_violation_info[$i]);
                my $new_entry = dclone($content_profile_entry);
                $new_entry->{defense_violation_info} = [];
                $new_entry->{xml_violation_info} = [$xml_violation_info];
                $instance->{content_profile_info} = $new_entry;
                push @instances, $instance;
            }
        }
    }

    return @instances;
}

#----------------------------------------------------------------------

sub get_format_violations {
    my ( $self, %params ) = @_;

    my $viol_index           = $params{viol_index};
    my $instance             = $params{instance};
    my $violation_rec        = $params{violation_rec};
    my $entity               = $params{entity};
    my $entity_type          = $params{entity_type};

    my $content_profile_info = $instance->{content_profile_info} || get_entity_profile_info($entity);
    return undef unless $content_profile_info;

    $violation_rec->{content_profile} = $self->get_policy_content_profile(content_profile_info => $content_profile_info);
    $violation_rec->{content_profile_type} = $content_profile_info->{content_profile_type};

    if ( ref $content_profile_info->{xml_violation_info} ) {
        my ($xml_violation_info) = grep { $_->{viol_index} == $violation_rec->{viol_index} } @{$content_profile_info->{xml_violation_info}};
        if ($xml_violation_info) {
            $violation_rec->{content_profile_location_desc} = $xml_violation_info->{location_desc};
            $violation_rec->{snippet_buffer}                = $xml_violation_info->{data_buffer};
            my $violation_desc = join( "\n", grep { $_ } $xml_violation_info->{error_desc}, $xml_violation_info->{fault_details} );
            if ($violation_desc) {
                $violation_rec->{observed_entity_violation_description} = $violation_desc;
            }
        }
    }

    $violation_rec->{defense_element_allowed_value} = $instance->{defense_element_allowed_value_str};
    $violation_rec->{defense_element_name}          = $instance->{defense_element_name};
    $violation_rec->{defense_element_name_xpointer} = $instance->{defense_element_name_xpointer};
    $violation_rec->{defense_element_value}         = $instance->{defense_element_value_str};

    # todo: must be cached
    my $defenseClass = 'F5::ASMConfig::Entity::XmlDefenseAttribute';
    if ($viol_index == F5::Violation::VIOL_GWT_FORMAT()) {
        $defenseClass = 'F5::ASMConfig::Entity::GwtDefenseAttribute';
    }
    elsif ($viol_index == F5::Violation::VIOL_JSON_FORMAT()) {
        $defenseClass = 'F5::ASMConfig::Entity::JsonDefenseAttribute';
    }
    elsif ($viol_index == F5::Violation::VIOL_PLAINTEXT_FORMAT()) {
        $defenseClass = 'F5::ASMConfig::Entity::PlainTextDefenseAttribute';
    }
    elsif ($viol_index == F5::Violation::VIOL_GRAPHQL_FORMAT()) {
        $defenseClass = 'F5::ASMConfig::Entity::GraphqlDefenseAttribute';
    }

    my $defenseAttribute = $defenseClass->new(
        dbh          => $self->{dbh},
        get_criteria => { def_attrib_xpath => $instance->{defense_element_name_xpointer} }
    );

    return undef unless $defenseAttribute;

    $violation_rec->{defense_failed} = $defenseAttribute->def_attrib_name;

    $self->check_entity_type_request(
        violation_rec => $violation_rec,
        entity_type   => $entity_type,
        entity        => $entity,
    );

    return 1;
}

#----------------------------------------------------------------------

sub get_entity_profile_info {
    my ($entity) = @_;

    my $content_profile_info;
    if (ref($entity->{content_profile_info}) eq 'HASH') {
        $content_profile_info = $entity->{content_profile_info};
    }
    elsif (ref($entity->{content_profile_info}) eq 'ARRAY') {
        $content_profile_info = $entity->{content_profile_info}[0];
    }

    return $content_profile_info;
}

#----------------------------------------------------------------------

sub check_entity_type_request {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};
    my $entity_type   = $params{entity_type};
    my $entity        = $params{entity};

    if ( $entity_type eq 'request' ) {
        $violation_rec->{entity_type}          = 'url';
        $violation_rec->{observed_entity_name} = $entity->{url}->{name};
        $violation_rec->{observed_entity_method} = $entity->{url}->{method} if $entity->{url}->{method};
        my $wildcard = $entity->{url}->{wildcard};
        if ($wildcard) {
            $self->set_policy_entity(
                violation_rec => $violation_rec,
                wildcard      => $wildcard,
                staging       => $entity->{url}->{is_in_staging},
                entity_type   => $entity_type,
            );
        }
    }

}

#----------------------------------------------------------------------

sub json_schema_violation_instances {
    my ($self, %params) = @_;

    my $entity = $params{entity};
    my $viol_index = $params{viol_index};

    # when entity is the request itself, the staging property will be inside the url, with in the request
    # need to bring it up to the entity (request) level, so that post-processing of the entity_instances_creator
    # logic would do the right thing
    if (defined $entity->{url} && defined $entity->{url}->{is_in_staging}) {
        $entity->{is_staging} = $entity->{url}->{is_in_staging};

        # avoid entity fields pollution
        $entity->{_transient_fields} = [ 'is_staging' ];
    }

    my $content_profile; # there is only one
    if (ref($entity->{content_profile_info}) eq 'ARRAY') {
        $content_profile = $entity->{content_profile_info}->[0];
    }
    else {
        $content_profile = $entity->{content_profile_info};
    }

    my @instances = ();
    return @instances unless ref($content_profile->{json_schema_violation_details}) eq 'ARRAY';

    my @desired_viols = grep { $_->{viol_index} == $viol_index } @{$content_profile->{json_schema_violation_details}};
    for my $viol (@desired_viols) {
        my $instance = dclone($content_profile);
        my $viol_clone = dclone($viol);
        $instance->{json_schema_violation_details} = [ $viol_clone ];
        push @instances, $instance;
    }

    $self->{logger}->trace(\@instances) if $self->{logger};

    # list of "content_profile_info" hashes
    # each with a "json_schema_violation_details" list of only one violation
    return @instances;
}

#----------------------------------------------------------------------

sub get_json_schema {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};
    my $instance = $params{instance};
    my $entity = $params{entity};

    $violation_rec->{content_profile} = $self->get_policy_content_profile(content_profile_info => $instance);
    $violation_rec->{content_profile_type} = $instance->{content_profile_type};

    my $violation_details = $instance->{json_schema_violation_details}->[0];
    if ($violation_details->{viol_index} && $violation_details->{viol_index} == F5::Violation::VIOL_JSON_SCHEMA()) {
        for my $detail (keys %$violation_details) {
            $violation_rec->{$detail} = $violation_details->{$detail} if defined $violation_details->{$detail};
        }

        if (exists $json_schema_violation_error_code_to_violation_record->{$violation_details->{error}}) {
            my $rv = $json_schema_violation_error_code_to_violation_record->{$violation_details->{error}}->(violation_details => $violation_details);
            for my $detail (keys %$rv) {
                $violation_rec->{$detail} = $rv->{$detail} if defined($rv->{$detail});
            }
        }
        else {
            my $error_message = "Unrecognized validation error code $violation_details->{error}";
            $violation_rec->{validation_error} = $error_message;
            $self->{logger}->error($error_message) if $self->{logger};
        }
    }

    $self->{logger}->trace($violation_rec) if $self->{logger};

    return 1;
}

#----------------------------------------------------------------------

sub get_xml_schema {
    my ( $self, %params ) = @_;

    my $violation_rec        = $params{violation_rec};
    my $instance             = $params{instance};
    my $entity               = $params{entity};

    my $content_profile_info = get_entity_profile_info($entity);
    return undef unless $content_profile_info;

    if (ref $content_profile_info->{xml_additional_violation_info}) {
        $violation_rec->{observed_entity_expected_soap_action} = $content_profile_info->{xml_additional_violation_info}->{expected_soap_action_header};
    }

    if ( ref $content_profile_info->{xml_violation_info} ) {
        for my $xml_violation_info ( @{$content_profile_info->{xml_violation_info}} ) {
            if ($xml_violation_info->{viol_index} == F5::Violation::VIOL_XML_SCHEMA()) {
                $violation_rec->{observed_entity_soap_method} = $xml_violation_info->{method_name} if $xml_violation_info->{method_name};
                $violation_rec->{observed_entity_soap_namespace} = $xml_violation_info->{namespace_} if $xml_violation_info->{namespace_};
                $violation_rec->{observed_entity_actual_soap_action} = $xml_violation_info->{soap_action_header_value} if $xml_violation_info->{soap_action_header_value};
                last;
            }
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub get_xml_soap_method {
    my ( $self, %params ) = @_;

    my $violation_rec        = $params{violation_rec};
    my $entity               = $params{entity};
    my $entity_type          = $params{entity_type};

    my $content_profile_info = get_entity_profile_info($entity);
    return undef unless $content_profile_info;

    $violation_rec->{content_profile} = $self->get_policy_content_profile(content_profile_info => $content_profile_info);
    if ($entity->{url} && defined($entity->{url}->{name})) {
        $violation_rec->{observed_entity_name} = $entity->{url}->{name};
        $violation_rec->{observed_entity_method} = $entity->{url}->{method} if $entity->{url}->{method};
    }

    if ( ref $content_profile_info->{xml_violation_info} ) {
        #in case VIOL_XML_SOAP_METHOD all xml_violation_info entries should be united into one
        my $xml_violation_info = {};
        for my $info_elem ( @{$content_profile_info->{xml_violation_info}} ) {
            if ($info_elem->{viol_index} == F5::Violation::VIOL_XML_SOAP_METHOD()) {
                for my $info_entry ( keys %$info_elem ) {
                    $xml_violation_info->{$info_entry} = $info_elem->{$info_entry};
                }
            }
        }

        $violation_rec->{observed_entity_soap_method} = $xml_violation_info->{method_name} if $xml_violation_info->{method_name};
        $violation_rec->{observed_entity_soap_namespace} = $xml_violation_info->{namespace_} if $xml_violation_info->{namespace_};
        $violation_rec->{observed_entity_actual_soap_action} = $xml_violation_info->{soap_action_header_value} if $xml_violation_info->{soap_action_header_value};
    }

    $self->check_entity_type_request(
        violation_rec => $violation_rec,
        entity_type   => $entity_type,
        entity        => $entity,
    );

    return 1;
}

#----------------------------------------------------------------------

sub session_awareness_subviolations {
    my ( $self, %params ) = @_;

    my $instance      = $params{instance};
    my $violation_rec = $params{violation_rec};

    my $additional_violation_info = $self->{request}->{request_additional_violation_info};
    if ( $additional_violation_info &&
         ref($additional_violation_info->{session_awareness_scope}) eq 'ARRAY' &&
         ref($additional_violation_info->{session_awareness_flag}) eq 'ARRAY' &&
         $#{$additional_violation_info->{session_awareness_flag}} == $#{$additional_violation_info->{session_awareness_scope}}
       ) {
        my %scopes = ();
        for my $i (0 .. $#{$additional_violation_info->{session_awareness_scope}} ) {
            my $flag = $additional_violation_info->{session_awareness_flag}->[$i];
            next unless $flag eq "block_all";

            my $scope = $additional_violation_info->{session_awareness_scope}->[$i];
            if ($scope eq 'sid') {
                $scopes{'session'} = 1;
            }
            else {
                $scopes{$scope} = 1;
            }
        }
        my @scope_list = keys %scopes;
        $violation_rec->{observed_entity_sa_lock_scopes} = join( ',', @scope_list);
    }

    return 1;
}

#----------------------------------------------------------------------

sub get_viol_redirect {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};

    $violation_rec->{observed_entity_redirection_domain} = $self->{response}->{redirection_domain};

    return check_tighten_wildcard(@_);
}

#----------------------------------------------------------------------

sub method_violation {
    my ( $self, %params ) = @_;

    my $instance       = $params{instance};
    my $violation_rec  = $params{violation_rec};
    my $viol_info_elem = $params{viol_info_elem};
    my $entity         = $params{entity};

    $violation_rec->{entity_type}            = 'method';
    $violation_rec->{observed_entity_method} = $instance->{method};
    $violation_rec->{observed_entity_violation_description} = $viol_info_elem->{viol_desc};

    if ($entity && $entity->{violating_entity_is_url} && $entity->{url} ){

        if ($entity->{url}->{name}) {
            $violation_rec->{observed_entity_name} = $entity->{url}->{name};
            $violation_rec->{observed_entity_method} = $entity->{url}->{method} if $entity->{url}->{method};
        }

        if ( $entity->{url}->{wildcard}){
                    $self->set_policy_entity(
                    violation_rec => $violation_rec,
                    wildcard      => $entity->{url}->{wildcard},
                    staging       => $violation_rec->{enforcement_state_is_in_staging},
                    entity_type   => 'url',
                );    
        }    
    }

    return 1;
}

#----------------------------------------------------------------------

sub url_content_type {
    my ( $self, %params ) = @_;

    my $instance       = $params{instance};
    my $violation_rec  = $params{violation_rec};

    if (! ref $instance->{header_content_profile}) {
        return;
    }

    my $default_content_profile = $instance->{header_content_profile}->{is_default_content_profile};
    if ($default_content_profile) {
        $violation_rec->{header_name} = '*';
    }
    else {
        $violation_rec->{header_name}          = $instance->{header_content_profile}->{name};
        $violation_rec->{header_actual_value}  = $instance->{header_content_profile}->{actual_value};
        $violation_rec->{header_matched_value} = $instance->{header_content_profile}->{value};
    }

    return 1;
}

#----------------------------------------------------------------------

sub get_brute_force_subviolations {
    my ( $self, %params ) = @_;

    my $instance       = $params{instance};
    my $violation_rec  = $params{violation_rec};
    my $viol_info_elem = $params{viol_info_elem};

    my $username            = $violation_rec->{brute_force_username};
    my $authentication_type = $self->{request}->{authentication_type};
    if ( $authentication_type == F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_HTTP_DIGEST ) {
        if ( $username =~ /Digest username="(.+?)"/ ) {
            $username = $1;
        }
    }
    elsif ( $authentication_type == F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_HTTP_BASIC ) {
        $username =~ s/^Basic\s//;
    }
    $violation_rec->{brute_force_username} = $username;

    return 1;
}

#----------------------------------------------------------------------

sub get_malicious_ip_subviolations {
    my ( $self, %params ) = @_;

    my $instance       = $params{instance};
    my $violation_rec  = $params{violation_rec};
    my $viol_info_elem = $params{viol_info_elem};

    my $ip_reps = $self->cache_subviolations( 'F5::ASMConfig::Entity::IpReputationCategoryAttributes', 'category_id' );

    my $status_vec = Bit::Vector->new_Dec( 64, $instance->{iprep}->{status} );
    my $alarm_vec  = Bit::Vector->new_Dec( 64, $instance->{iprep}->{alarm_mask} );
    my $block_vec  = Bit::Vector->new_Dec( 64, $instance->{iprep}->{block_mask} );

    for my $pos ( $status_vec->Index_List_Read ) {
        $violation_rec->{ip_rep_categories} ||= [];

        my $pos_bit         = $pos + 1;
        my $ip_rep_category = $ip_reps->{$pos_bit};

        my $alarm = 0;
        for my $alarm_pos ( $alarm_vec->Index_List_Read ) {
            if ( $alarm_pos == $pos ) {
                $alarm = 1;
                last;
            }
        }

        my $block = 0;
        for my $block_pos ( $block_vec->Index_List_Read ) {
            if ( $block_pos == $pos ) {
                $block = 1;
                last;
            }
        }

        my $category = {
            ip_rep_category_name => $ip_rep_category->{category_name},
            category_is_alarmed  => $alarm,
            category_is_blocked  => $block,
        };

        push @{ $violation_rec->{ip_rep_categories} }, $category;
    }

    return 1;
}

#----------------------------------------------------------------------

sub get_url_violation {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};

    return unless $self->check_tighten_wildcard(%params);

    if ((! exists($entity->{wildcard})) && $entity->{url} && $entity->{url}->{wildcard}) {
        $entity->{wildcard} = $entity->{url}->{wildcard};
    }

    return 1;
}

#----------------------------------------------------------------------

sub check_tighten_wildcard {
    my ( $self, %params ) = @_;

    my $instance                  = $params{instance};
    my $tightening_viol_indexes   = $params{tightening_viol_indexes};
    my $violations_status_indexes = $params{violations_status_indexes};
    my $viol_index                = $params{viol_index};
    my $entity_type               = $params{entity_type};

    if ( $tightening_viol_indexes->{$viol_index} && $instance->{wildcard} ) {
        return undef;
    }

    if ( $entity_type eq 'request' && $tightening_viol_indexes->{$viol_index} && !$violations_status_indexes->{$viol_index} ) {
        return undef;
    }

    return 1;
}

#----------------------------------------------------------------------

sub check_http_response_status {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};

    $violation_rec->{location}    = 'response';
    $violation_rec->{entity_type} = 'response';
    $violation_rec->{http_response_code} = $self->{response}->{status};

    return 1;
}

#----------------------------------------------------------------------

sub cookie_length {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};

    $violation_rec->{entity_type} = 'cookie';

    return 1;
}

#----------------------------------------------------------------------

sub signature_instances {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};

    my @instances;
    if (ref( $entity->{ attack_sig } ) eq 'ARRAY') {
        for my $new_inst (@{$entity->{ attack_sig }}) {
            if (ref($new_inst) && ! $new_inst->{from_content_profile}) {
                push @instances, $new_inst;
            }
        }
    }

    my @content_profiles;
    if (ref($entity->{content_profile_info}) eq 'ARRAY') {
        push @content_profiles, @{$entity->{content_profile_info}};
    }
    elsif (ref($entity->{content_profile_info}) eq 'HASH') {
        push @content_profiles, $entity->{content_profile_info};
    }
    for my $content_profile_info (@content_profiles) {
        if (ref($content_profile_info->{attack_sig}) eq 'ARRAY') {
            for my $instance (@{$content_profile_info->{attack_sig}}) {
                if ($instance->{sig_id}) {
                    my $instance_clone = dclone($instance);
                    $instance_clone->{content_profile_info} = $content_profile_info;
                    $instance_clone->{from_content_profile} = 1;
                    push @instances, $instance_clone;
                }
            }
        }
    }

    if (ref($entity->{url}) && ref($entity->{url}->{attack_sig}) eq 'ARRAY') {
        if (defined($entity->{url}->{is_in_staging})) {
            for my $inst (@{$entity->{url}->{attack_sig}}) {
                $inst->{is_in_staging} = $entity->{url}->{is_in_staging};
            }
        }
        push @instances, @{$entity->{url}->{attack_sig}};
    }

    return @instances;
}

#----------------------------------------------------------------------

sub threat_campaign_instances {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};

    $self->{_handled_threat_campaigns} ||= {};

    my @instances;
    if (ref( $entity->{ threat_campaign } ) eq 'ARRAY') {
        my $threat_campaigns = $entity->{ threat_campaign };
        # one instance for each TC
        for my $campaign (@$threat_campaigns) {
            my $campaign_name = $campaign->{name};
            if (! $self->{_handled_threat_campaigns}->{$campaign_name}) {
                $self->{_handled_threat_campaigns}->{$campaign_name} = 1;
                push(@instances, $campaign);
            }
        }
    }
    elsif (ref( $entity->{ threat_campaign } ) eq 'HASH') {
        push(@instances, $entity->{ threat_campaign });
    }

    return @instances;
}

#----------------------------------------------------------------------

sub evasion_instances {
    my ( $self, %params ) = @_;

    my $entity         = $params{entity};
    my $viol_info_elem = $params{viol_info_elem};

    my @instances = ();

    return @instances unless ref $viol_info_elem;

    my $refType = ref $viol_info_elem->{evasion_info};
    my @evasion_info_list;
    if ($refType eq 'ARRAY') {
        @evasion_info_list = @{$viol_info_elem->{evasion_info}};
    }
    elsif ($refType eq 'HASH') {
        @evasion_info_list = ($viol_info_elem->{evasion_info});
    }
    else {
        return @instances;
    }

    my $bit_field = 'detected_evasion_bit';
    for my $evasion_info (@evasion_info_list) {
        next unless ref $evasion_info;
        for my $instance (@{$evasion_info->{specific_evasion_details}}) {
            my $bit = $instance->{$bit_field};
            for my $index (F5::Violation::Utils::split_bitmask_values($bit)) {
                push (@instances, {%$instance, $bit_field => $index, context => $evasion_info->{context}});
            }
        }
    }

    return @instances;
}

#----------------------------------------------------------------------

sub param_value_metacharacter_instances {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};

    my %found_metachars = ();
    my @instances = ();
    if (ref($entity->{xml_violation_info}) eq 'HASH') {
        my $viol_info_elem = $entity->{xml_violation_info};
        if ($viol_info_elem->{viol_index} == F5::Violation::VIOL_PARAMETER_VALUE_METACHAR() && ! $viol_info_elem->{from_content_profile}) {
            push @instances, {
                metacharacter => $viol_info_elem->{illegal_metachar},
                viol_info_elem => $viol_info_elem,
            };
            $found_metachars{$viol_info_elem->{illegal_metachar}} = 1;
        }
    }
    elsif (ref($entity->{xml_violation_info}) eq 'ARRAY') {
        for my $viol_info_elem ( @{$entity->{xml_violation_info}} ) {
            if ($viol_info_elem->{viol_index} == F5::Violation::VIOL_PARAMETER_VALUE_METACHAR() && ! $viol_info_elem->{from_content_profile}) {
                push @instances, {
                    metacharacter => $viol_info_elem->{illegal_metachar},
                    viol_info_elem => $viol_info_elem,
                };
                $found_metachars{$viol_info_elem->{illegal_metachar}} = 1;
            }
        }
    }

    if (ref($entity->{content_profile_info}) eq 'ARRAY') {
        for my $content_profile_info (@{$entity->{content_profile_info}}) {
            my %found_profile_metachar;
            if (ref($content_profile_info->{xml_violation_info}) eq 'ARRAY') {
                for my $viol_info_elem ( @{$content_profile_info->{xml_violation_info}} ) {
                    if ($viol_info_elem->{viol_index} == F5::Violation::VIOL_PARAMETER_VALUE_METACHAR()) {
                        push @instances, {
                            metacharacter => $viol_info_elem->{illegal_metachar},
                            viol_info_elem => $viol_info_elem,
                            content_profile_info => $content_profile_info,
                            from_content_profile => 1,
                        };
                        $found_metachars{$viol_info_elem->{illegal_metachar}} = 1;
                        $found_profile_metachar{$viol_info_elem->{illegal_metachar}} = 1;
                    }
                }
            }

            if (ref($content_profile_info->{metachar}) && ref($content_profile_info->{metachar}->{value_metachar}) eq 'ARRAY') {
                for my $metachar (@{$content_profile_info->{metachar}->{value_metachar}}) {
                    if (! $found_profile_metachar{$metachar} ) {
                        push @instances, {
                            metacharacter => $metachar,
                            viol_info_elem => { data_buffer => chr($metachar) },
                            content_profile_info => $content_profile_info,
                            from_content_profile => 1,
                        };
                        $found_metachars{$metachar} = 1;
                    }
                }
            }
        }
    }

    for my $metachar_entry_name ('metachar', 'value_metachar') {
        if (ref($entity->{$metachar_entry_name}) eq 'ARRAY') {
            for my $metachar ( @{$entity->{$metachar_entry_name}} ) {
                if ( ! $found_metachars{$metachar} ) {
                    push @instances, {
                        metacharacter => $metachar,
                        viol_info_elem => { data_buffer => chr($metachar) },
                    };
                    $found_metachars{$metachar} = 1;
                }
            }
        }
    }

    return @instances;
}

#----------------------------------------------------------------------

sub parameter_array_instances {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};

    my @instances = ();
    if (ref($entity->{param_additional_violation_info}) && ref($entity->{param_additional_violation_info}->{array_violation_info})) {
        my $array_violation_info = $entity->{param_additional_violation_info}->{array_violation_info};
        if (ref($array_violation_info) eq 'ARRAY') {
            @instances = @$array_violation_info;
        }
        elsif (ref($array_violation_info) eq 'HASH') {
            @instances = ($array_violation_info);
        }
    }

    return @instances;
}

#----------------------------------------------------------------------

sub parameter_array_violations {
    my ( $self, %params ) = @_;

    my $instance = $params{instance};
    my $violation_rec = $params{violation_rec};

    return undef unless defined($instance->{reason});

    $violation_rec->{array_violation_reason} = $instance->{reason};
    $violation_rec->{actual_value} = $instance->{array_num_items};
    $violation_rec->{min_numeric_value} = $instance->{array_min_items};
    $violation_rec->{max_numeric_value} = $instance->{array_max_items};

    return 1;
}

#----------------------------------------------------------------------

sub mandatory_parameter_instances {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};

    my @instances = ();
    if (ref($entity->{request_additional_violation_info}) && ref($entity->{request_additional_violation_info}->{missing_mandatory_param})) {
        my $missing_mandatory_param = $entity->{request_additional_violation_info}->{missing_mandatory_param};
        if (ref($missing_mandatory_param) eq 'ARRAY') {
            @instances = @$missing_mandatory_param;
        }
        elsif (ref($missing_mandatory_param) eq 'HASH') {
            @instances = ($missing_mandatory_param);
        }
    }

    return @instances;
    
}

#----------------------------------------------------------------------

sub mandatory_parameter_violations {
    my ( $self, %params ) = @_;

    my $instance = $params{instance};
    my $violation_rec = $params{violation_rec};

    $violation_rec->{observed_entity_name} = $instance->{param_name};
    $violation_rec->{param_level} = $instance->{param_level};

    return 1;
}

#----------------------------------------------------------------------

sub param_value_metacharacter_violation {
    my ( $self, %params ) = @_;

    my $instance      = $params{instance};
    my $violation_rec = $params{violation_rec};

    if ( ref($instance) ) {
        $violation_rec->{metacharacter} = $instance->{metacharacter};
        if ($instance->{viol_info_elem} && $instance->{viol_info_elem}->{data_buffer}) {
            $violation_rec->{snippet_buffer} = $instance->{viol_info_elem}->{data_buffer};
        }
    }
    else {
        $violation_rec->{metacharacter} = $instance;
    }

    return 1;
}

#----------------------------------------------------------------------

sub threat_analysis_violations {
    my ( $self, %params ) = @_;

    my $threat_analysis = $params{instance}->{threat_analysis_enforcement};
    my $violation_rec = $params{violation_rec};

    $violation_rec->{threat_analysis_configured_enforcement_action} = $threat_analysis->{configured_enforcement_action};
    $violation_rec->{threat_analysis_actual_enforcement_action} = $threat_analysis->{actual_enforcement_action};
    $violation_rec->{threat_analysis_enforcement_action_reason} = $threat_analysis->{enforcement_action_reason};

    if ($threat_analysis->{actual_enforcement_action} != F5::ASMConfig::ThreatAnalysisEnforcementAction::THREAT_ANALYSIS_ENFORCEMENT_ACTION_DECEPTION) {
        return 1;
    }

    if ($threat_analysis->{server_technology}) {
        my $server_technology = F5::ASMConfig::Entity::ServerTechnology->new(dbh => $self->{dbh},
                                                                             minimal_load => 1,
                                                                             get_criteria => {
                                                                                 server_technology_id => $threat_analysis->{server_technology},
                                                                             });
        if (ref($server_technology)) {
            $violation_rec->{deception_response_server_technology} = $server_technology;
        }
    }

    if ($threat_analysis->{sig_id}) {
        my $sig = F5::ASMConfig::Entity::Signature->new(dbh => $self->{dbh},
                                                        minimal_load => 1,
                                                        get_criteria => {
                                                            sig_id => $threat_analysis->{sig_id},
                                                        });

        if (ref($sig)) {
            $violation_rec->{deception_response_signature} = $sig;
        }
    }

    $violation_rec->{deception_response_code} = $threat_analysis->{error_code};

    return 1;
}

#----------------------------------------------------------------------

sub blocking_condition_violations {
    my ( $self, %params ) = @_;

    my $blocking_condition = $params{instance};
    my $violation_rec = $params{violation_rec};

    $violation_rec->{blocking_condition} = $blocking_condition->{blocking_condition};
    $violation_rec->{blocking_condition_sig_id} = $blocking_condition->{sig_id};

    if (!$blocking_condition->{sig_id}) {
        return 1;
    }

    my $sig = F5::ASMConfig::Entity::Signature->new(dbh => $self->{dbh},
                                                    minimal_load => 1,
                                                    get_criteria => {
                                                        sig_id => $blocking_condition->{sig_id},
                                                    });

    if (ref($sig)) {
        $violation_rec->{blocking_condition_signature} = $sig;
    }

    return 1;
}

#----------------------------------------------------------------------

sub blocking_condition_instances {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};

    my @instances;
    if (ref($entity->{behavioral_blocking_condition}) eq 'ARRAY') {
        @instances = @{$entity->{behavioral_blocking_condition}};
    }
    elsif (ref($entity->{behavioral_blocking_condition}) eq 'HASH') {
        push(@instances, $entity->{behavioral_blocking_condition});
    }

    return @instances;
}

#----------------------------------------------------------------------

sub conviction_violations {
    my ( $self, %params ) = @_;

    my $deception_trap = $params{instance};
    my $violation_rec = $params{violation_rec};

    $violation_rec->{deception_trap_display_name} = $deception_trap->{deception_trap_display_name};
    $violation_rec->{deception_trap_name} = $deception_trap->{deception_trap_name};
    $violation_rec->{deception_trap_risk} = $deception_trap->{risk};
    $violation_rec->{deception_trap_score} = $deception_trap->{score};
    $violation_rec->{attack_type_id} = $deception_trap->{attack_type_id};

    if ($deception_trap->{attack_type_id}) {
        my $attack_type = F5::ASMConfig::Entity::AttackType->new(dbh => $self->{dbh},
                                                                 minimal_load => 1,
                                                                 get_criteria => {
                                                                     attack_type_id => $deception_trap->{attack_type_id},
                                                                 });

        if (ref($attack_type)) {
            $violation_rec->{deception_trap_attack_type} = $attack_type;
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub conviction_instances {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};

    my @instances;
    if (ref($entity->{deception_trap}) eq 'ARRAY') {
        @instances = @{$entity->{deception_trap}};
    }
    elsif (ref($entity->{deception_trap}) eq 'HASH') {
        push(@instances, $entity->{deception_trap});
    }

    return @instances;
}

#----------------------------------------------------------------------

sub malicious_device_violations {
    my ( $self, %params ) = @_;

    my $malicious_device = $params{request}->{malicious_device};
    my $violation_rec    = $params{violation_rec};

    if ($malicious_device->{incident_id}) {
        my $minimal_load = !F5::Utils::Rest::is_expanded(expand_fields => $self->{expand_fields},
                                                         path => 'violations/incident');

        my $incident = F5::ASMConfig::Entity::Incident->new(dbh => $self->{dbh},
                                                            minimal_load => $minimal_load,
                                                            get_criteria => {
                                                                incident_id => $malicious_device->{incident_id}
                                                            });
        if (ref($incident)) {
            $violation_rec->{incident} = $incident;
        }
    }

    $violation_rec->{source_type} = $malicious_device->{source_type};
    $violation_rec->{client_reputation_reasons} ||= [];

    my $heuristic_reason_mask = $malicious_device->{heuristic_reason_mask};
    return unless $heuristic_reason_mask;

    my $vec = Bit::Vector->new_Dec(64, $heuristic_reason_mask);
    for my $pos ($vec->Index_List_Read) {
        push @{$violation_rec->{client_reputation_reasons}}, $F5::ClientReputationReason::client_reputation_reason{$pos};
    }

    return 1;
}

#----------------------------------------------------------------------

sub get_data_guard_violation {
    my ( $self, %params ) = @_;

    my $viol_info_elem = $params{viol_info_elem};
    my $violation_rec  = $params{violation_rec};

    if ( $viol_info_elem->{format_id} ) {
        my $formats = $self->cache_subviolations( 'F5::ASMConfig::Entity::ResponseScrubbingFormat', 'format_id' );
        my $fid = $viol_info_elem->{format_id};
        if (ref($formats) && ref($formats->{$fid})) {
            my $format = $formats->{$fid};
            $violation_rec->{snippet_buffer} = $format->{format_name};
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub asm_cookie_hijacking {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};
    my $violation_rec  = $params{violation_rec};

    my $violation_info = $entity->{request_violation_info};
    if (ref $violation_info eq 'ARRAY') {
        for my $info_entry (@$violation_info) {
            if ($info_entry->{viol_index} == F5::Violation::VIOL_ASM_COOKIE_HIJACKING()) {
                $violation_rec->{actual_value} = $info_entry->{viol_desc};
                last;
            }
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub file_upload_in_body {
    my ( $self, %params ) = @_;

    my $entity = $params{entity};
    my $violation_rec  = $params{violation_rec};

    my $violation_info = $entity->{request_violation_info};
    if (ref $violation_info eq 'ARRAY') {
        for my $info_entry (@$violation_info) {
            if ($info_entry->{viol_index} == F5::Violation::VIOL_FILE_UPLOAD_IN_BODY()) {
                $violation_rec->{observed_entity_violation_description} = $info_entry->{uploaded_file_desc};
                last;
            }
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub get_framing_protocol_violations {
    my ( $self, %params ) = @_;

    my $entity        = $params{entity};
    my $violation_rec = $params{violation_rec};

    my %framing_error_descriptions = (
        1 => 'Continuation frame without start frame',
        2 => 'Start frame without ending the previous message fragmentation (interleaving fragmentation is not allowed!)',
        3 => 'Reserved frame opcode values',
        4 => 'Control frame size exceeded',
        5 => 'Control frame with fin flag off',
    );

    $violation_rec->{actual_value} = $framing_error_descriptions{ $entity->{websocket_violation_info}->{framing_error} };

    return 1;
}

#----------------------------------------------------------------------

sub get_websocket_extension_violations {
    my ( $self, %params ) = @_;

    my $entity        = $params{entity};
    my $violation_rec = $params{violation_rec};

    $violation_rec->{actual_value} = $entity->{websocket_violation_info}->{illegal_extensions}->{actual} ||
                                     $entity->{websocket_violation_info}->{bad_request}->{actual};

    return 1;
}

#----------------------------------------------------------------------

sub get_encoding_fields {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};

    my $is_websocket = $self->{bd_data}->{is_websockets};
    if ($is_websocket) {
        $violation_rec->{entity_type} = 'websocketUrl';
    }

    return 1;
}

#----------------------------------------------------------------------

sub get_csrf_details {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};
    my $request = $params{request};

    for my $k (keys %{$request->{csrf_violation_details}}) {
        $violation_rec->{"csrf_details_$k"} = $request->{csrf_violation_details}->{$k};
    }

    return 1;
}

#----------------------------------------------------------------------

sub datasafe_violations {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};

    $violation_rec->{manipulated_parameters} = $params{request}->{data_integrity_alert}->{manipulated_parameters};

    return 1;
}

#----------------------------------------------------------------------

sub get_ssrf_details {
    my ( $self, %params ) = @_;

    my $violation_rec = $params{violation_rec};
    my $ssrf_violation_details  = $params{entity}->{param_additional_violation_info}->{ssrf_violation_details};

    $violation_rec->{ssrf_details} = F5::ASMConfig::SsrfDetails->new($ssrf_violation_details);

    for my $k (keys %{$ssrf_violation_details->{snippet}}) {
        $violation_rec->{"snippet_$k"} = $ssrf_violation_details->{snippet}->{$k};
    }

    return 1;
}

#----------------------------------------------------------------------
1;
