package F5::BehavioralWafFile;

use strict;
use warnings;

use File::Basename;
use F5::Deploy;
use F5::Logger;
use File::Find;
use F5::Utils::File;
use F5::XML::Twig;
use F5::ASMConfig::EasyClient;
use F5::DbUtils;
use F5::Utils::String;
use File::Copy;
use File::Path;
use F5::ASM::COAPI;
use F5::CRC;
use F5::ASMConfig::Entity::AttackType;
use F5::SystemUtils;

my $TEMP_BASE_DIR = F5::Deploy::map_path("/ts/var/tmp");
my $INITIAL_DATASYNC_PROFILE_NAME = "/Common/datasync-global/datasync-global-profile-crypt-asm-dosl7-initial";
my $DECEPTION_TRAPS_FILE_NAME = "deception_traps.tar.gz";

sub new {
  my($class, %params) = @_;

  my $self = {};
  bless $self, $class;
  $self->init(%params);

  return $self;
}

#----------------------------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;

    $self->{dbh} = $params{dbh};
    $self->{logger} = $params{logger} || F5::Logger->new('/ts/log/behavioral_waf.log');
    $self->{filename} = $params{filename};
    $self->{base_dir} = "$TEMP_BASE_DIR/deception.tmp.$$";
}

#----------------------------------------------------------------------------------------

sub update_behavioral_waf_file {
    my ($self, %params) = @_;

    mkdir $self->{base_dir};

    eval {
        my $sys_rv = F5::SystemUtils::run_system_tool(command_line_params => ['tar','-zmxf', $self->{filename}, '-C', $self->{base_dir}],
                                                     find_command_path => 1);
        if (!$sys_rv->{completed_successfully}) {
           die "Failed to unpack $self->{filename}: '$sys_rv->{standard_error}'\n";
        }
        $self->load_deception_response_pages(%params);
        $self->load_deception_traps(%params);
    };
    rmtree($self->{base_dir});
    if ($@) {
        die { error_message => "Failed to update behavioral waf file: $@" };
    }
}

#----------------------------------------------------------------------------------------

sub load_deception_response_pages {
    my ($self, %params) = @_;
    my $update_dir = "$self->{base_dir}/deception_response_pages";

    my @files;
    find({ wanted => sub {
               if($File::Find::name =~ /\.yaml$/i && $File::Find::name !~ /timestamps/) {
                   push(@files, $File::Find::name);
               }
           }
         },
         $update_dir);

    my $server_technologies = {};
    my $deception_server_technologies = [];

    my $existing_responses = $self->{dbh}->selectall_hashref("SELECT deception_response_id FROM PLC.DECEPTION_RESPONSES", 'deception_response_id', undef);
    for my $file (@files) {
        my $update_file = YAML::XS::LoadFile($file);
        my $response = $update_file->{response_page};

        $self->{logger}->debug("Loading deception response page $response->{response_page_id}: $response->{status_code}") if $self->{logger};

        my $server_technology_id = defined $response->{server_technology} ? $server_technologies->{$response->{server_technology}} : 0;
        if (!defined $server_technology_id) {
            my $server_technology = F5::ASMConfig::Entity::ServerTechnology->new(dbh => $self->{dbh},
                                                                                 get_criteria => {
                                                                                     server_technology_name => $response->{server_technology}
                                                                                 });
            if (!ref($server_technology)) {
                $self->{logger}->info("Skipping response page $response->{response_page_id} -- $response->{server_technology} does not exists") if $self->{logger};
                next;
            }

            $server_technology_id = $server_technology->{server_technology_id};
            $server_technologies->{$response->{server_technology}} = $server_technology_id;
        }

        if (exists $existing_responses->{$response->{response_page_id}}) {
            delete $existing_responses->{$response->{response_page_id}};
        }

        my $response_rec = { deception_response_id => $response->{response_page_id},
                             server_technology_id => $server_technology_id,
                             response_body => $response->{body},
                             response_code => $response->{status_code},
                             response_headers => $response->{headers}};

        F5::DbUtils::insert_row(dbh => $self->{dbh},
                                table_name => 'PLC.DECEPTION_RESPONSES',
                                on_duplicate_key_update_fields => [ qw( response_code response_body response_headers ) ],
                                rec => $response_rec);

        if (!defined $response->{sig_id}) {
            push (@$deception_server_technologies, $server_technology_id);
            next;
        }
	
	$self->{logger}->debug("Loading response page signatures: @{$response->{sig_id}}") if $self->{logger};
        
	my $existing_responses_signatures =
            $self->{dbh}->selectall_hashref("SELECT sig_id FROM PLC.DECEPTION_RESPONSES_SIGNATURES where deception_response_id = ?",
                                            'sig_id', undef, $response->{response_page_id});
        for my $sig_id (@{$response->{sig_id}}) {
            if (exists $existing_responses_signatures->{$sig_id}) {
                delete $existing_responses_signatures->{$sig_id};
            }
            F5::DbUtils::insert_row(dbh => $self->{dbh},
                                    table_name => 'PLC.DECEPTION_RESPONSES_SIGNATURES',
                                    ignore => 1,
                                    rec => { deception_response_id => $response->{response_page_id},
                                             sig_id => $sig_id, });
        }

        # deleting redundant signatures relations
        my @deleted_response_signatures = keys %$existing_responses_signatures;
        if (@deleted_response_signatures){
            $self->{logger}->debug("Deleting deception response signatures @deleted_response_signatures") if $self->{logger};
            F5::DbUtils::delete_rows(dbh => $self->{dbh},
                                table_name => 'PLC.DECEPTION_RESPONSES_SIGNATURES',
                                where_fields => {
                                deception_response_id => $response->{response_page_id},
                                sig_id => \@deleted_response_signatures},
                                );
        }
    }

    my @deleted_responses = keys %$existing_responses;
    if (@deleted_responses) {
        $self->{logger}->debug("Deleting deception response pages @deleted_responses") if $self->{logger};
        F5::DbUtils::delete_rows(dbh => $self->{dbh},
                                 table_name => 'PLC.DECEPTION_RESPONSES',
                                 where_fields => { deception_response_id => \@deleted_responses },
                                );
    }

    F5::DbUtils::update_row(dbh => $self->{dbh},
                            table_name => 'PLC.SERVER_TECHNOLOGIES',
                            rec => { flg_has_deception_responses => 0 }, );

    if (@$deception_server_technologies) {
        F5::DbUtils::update_row(dbh => $self->{dbh},
                                table_name => 'PLC.SERVER_TECHNOLOGIES',
                                rec => { flg_has_deception_responses => 1 },
                                where_fields => { server_technology_id => $deception_server_technologies }, );
    }
}

#----------------------------------------------------------------------------------------

my $deception_enums_map = {
    risk => {
             high => F5::ASMConfig::DeceptionTrapRisk::DECEPTION_TRAP_RISK_HIGH(),
             medium => F5::ASMConfig::DeceptionTrapRisk::DECEPTION_TRAP_RISK_MEDIUM(),
             low => F5::ASMConfig::DeceptionTrapRisk::DECEPTION_TRAP_RISK_LOW(),
            },
    injection_type => {
                       cookie => F5::ASMConfig::InjectionType::INJECTION_TYPE_COOKIE(),
                       parameter => F5::ASMConfig::InjectionType::INJECTION_TYPE_PARAM(),
                       login_page_comment => F5::ASMConfig::InjectionType::INJECTION_TYPE_LOGIN_PAGE_COMMENT(),
                       login_page_field => F5::ASMConfig::InjectionType::INJECTION_TYPE_LOGIN_PAGE_FIELD(),
                       url_exists => F5::ASMConfig::InjectionType::INJECTION_TYPE_URL_EXISTS(),
                       url_not_exists => F5::ASMConfig::InjectionType::INJECTION_TYPE_URL_NOT_EXISTS(),
		       header => F5::ASMConfig::InjectionType::INJECTION_TYPE_HEADER(),
                      },
    conviction_source => {
                          cookie_name => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_COOKIE_NAME(),
                          cookie_value => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_COOKIE_VALUE(),
                          parameter_name => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_PARAM_NAME(),
                          parameter_value => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_PARAM_VALUE(),
                          header_name => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_HEADER_NAME(),
                          header_value => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_HEADER_VALUE(),
                          username => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_USERNAME(),
                          password => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_PASSWORD(),
                          url => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_URL(),
                          signature => F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_SIGNATURE(),
                        },
    logical_operation => {
                          eq => F5::ASMConfig::DeceptionTrapOperator::DECEPTION_TRAP_OPERATOR_EQUAL(),
                          ne => F5::ASMConfig::DeceptionTrapOperator::DECEPTION_TRAP_OPERATOR_NOT_EQUAL(),
                         },
    tag_injection_location => {
                               before => F5::ASMConfig::TagInjectionLocation::TAG_INJECTION_LOCATION_BEFORE(),
                               after => F5::ASMConfig::TagInjectionLocation::TAG_INJECTION_LOCATION_AFTER(),
                              },
};

sub load_deception_traps {
    my ($self, %params) = @_;

    my $sys_rv = F5::SystemUtils::run_system_tool(command_line_params => ['tar','-zmxf', "$self->{base_dir}/$DECEPTION_TRAPS_FILE_NAME", '-C', $self->{base_dir}],
                                                  find_command_path => 1);
    if (!$sys_rv->{completed_successfully}) {
        die "Failed to unpack $self->{base_dir}/$DECEPTION_TRAPS_FILE_NAME: '$sys_rv->{standard_error}'\n";
    }

    my $update_dir = "$self->{base_dir}/deception_traps";

    my @files;
    my $vars_file;
    find({ wanted => sub {
               if($File::Find::name =~ /\.yaml$/i && $File::Find::name !~ /timestamps/) {
                   if ($File::Find::name =~ /\/vars\//) {
                       $vars_file = $File::Find::name;
                   }
                   else {
                       push(@files, $File::Find::name);
                   }
               }
           }
         },
         $update_dir);

    my $variables = $self->get_random_values(%params,
                                             file => $vars_file);

    my $lazy_insert = F5::LazyInsert->new(dbh => $self->{dbh},
                                          logger => $self->{logger},
                                          batch_insert => F5::BatchInsert->new(dbh => $self->{dbh}),
                                         );

    $lazy_insert->handle_table(table_name => 'PLC.DECEPTION_TRAPS',
                               unique_key => 'deception_trap_id');

    $lazy_insert->handle_table(table_name => 'PLC.DECEPTION_TRAP_CONDITIONS',
                               ignore_fields => [ qw(deception_trap_condition_id last_updated) ],
                               unique_key => [ qw(deception_trap_id condition_source condition_operator condition_value) ] );

    $lazy_insert->handle_table(table_name => 'PLC.DECEPTION_TRAP_INJECTIONS',
                               ignore_fields => [ qw(deception_trap_injection_id) ],
                               unique_key => 'deception_trap_id');

    $lazy_insert->handle_table(table_name => 'PLC.DECEPTION_TRAP_INJECTION_HEADERS',
                               ignore_fields => [ qw(injection_header_id) ],
                               unique_key => [ qw(deception_trap_id header_name) ] );


    my $existing_traps = $self->{dbh}->selectall_hashref("SELECT deception_trap_id FROM PLC.DECEPTION_TRAPS", 'deception_trap_id', undef);

    my @url_filters = ();
    for my $file (@files) {
        my $update_file = YAML::XS::LoadFile($file);
        my $trap = $update_file->{deception_trap};

        next if !$trap;

        my $trap_rec = {};

        $self->{logger}->debug("Loading deception trap $trap->{deception_trap_id} : $trap->{name}") if $self->{logger};

        $trap_rec->{deception_trap_id} = $trap->{deception_trap_id};
        $trap_rec->{deception_trap_name} = $trap->{name};
        $trap_rec->{score} = $trap->{score};
        $trap_rec->{deception_trap_display_name} = $trap->{display_name};

        if (exists $existing_traps->{$trap->{deception_trap_id}}) {
            delete $existing_traps->{$trap->{deception_trap_id}};
        }

        my $attack_type = F5::ASMConfig::Entity::AttackType->new(dbh => $self->{dbh},
                                                                 get_criteria => {
                                                                     attack_type_display_name => $trap->{attack_type}
                                                                 });
        if ($attack_type) {
            $trap_rec->{attack_type_id} = $attack_type->{attack_type_id};
        }

        $trap_rec->{risk} = $deception_enums_map->{risk}->{$trap->{risk}};
        $lazy_insert->insert_row(table_name => 'PLC.DECEPTION_TRAPS',
                                 rec => $trap_rec);

        $self->inject_random_values(%params,
                                    variables => $variables,
                                    trap => $trap);

        if ($trap->{injection_type}) {
            my $tag_location = F5::ASMConfig::TagInjectionLocation::TAG_INJECTION_LOCATION_NA();
            if ($trap->{tag_injection_location}) {
                $tag_location = $deception_enums_map->{tag_injection_location}->{$trap->{tag_injection_location}};
            }
            $lazy_insert->insert_row(table_name => 'PLC.DECEPTION_TRAP_INJECTIONS',
                                     rec => {
                                             deception_trap_id => $trap->{deception_trap_id},
                                             injection_type => $deception_enums_map->{injection_type}->{$trap->{injection_type}},
                                             injection_value => $trap->{injection_value},
                                             injection_condition => $trap->{injection_key},
                                             tag_injection_location => $tag_location,
                                             tag_name => lc($trap->{injection_tag} || ''),
                                            }
                                    );
            push @url_filters, $trap->{injection_key} if ($trap->{injection_type} eq 'url_not_exists');
            for my $header (@{$trap->{injection_headers}}) {
                $lazy_insert->insert_row(table_name => 'PLC.DECEPTION_TRAP_INJECTION_HEADERS',
                                         rec => {
                                                 deception_trap_id => $trap->{deception_trap_id},
                                                 header_name => $header->{header_name},
                                                 header_value => $header->{header_val},
                                                }
                                        );
            }
        }

        for my $condition (@{$trap->{conviction_conditions}}) {
            if ($condition->{conviction_source} eq 'header_name') {
                $condition->{value} = lc($condition->{value});
            }
            $lazy_insert->insert_row(table_name => 'PLC.DECEPTION_TRAP_CONDITIONS',
                                     rec => {
                                             deception_trap_id => $trap->{deception_trap_id},
                                             condition_source => $deception_enums_map->{conviction_source}->{$condition->{conviction_source}},
                                             condition_operator => $deception_enums_map->{logical_operation}->{$condition->{logical_operation}},
                                             condition_value => $condition->{value},
                                            },
                                     set_date_fields_to_now_unix => [ 'last_updated' ],
                                    );
        }
    }

    $self->handle_mcp_url_filters(url_filters => \@url_filters);
 
    my @deleted_traps = keys %$existing_traps;
    if (@deleted_traps) {
        $self->{logger}->debug("Deleted traps: @deleted_traps") if $self->{logger};
    }

    $lazy_insert->flush();
}

#----------------------------------------------------------------------------------------

sub handle_mcp_url_filters {
    my ($self, %params) = @_;
    my $url_filters = $params{url_filters};
    my $obj = F5::ASM::COAPI::Object->new('app_cloud_security_service_config');
    $obj->set_value('app_cloud_security_service_name', '/Common/f5-global-cloud-app');
    $obj->set_value('name', '/Common/f5-global-cloud-app$.LiveUpdateBasicConfig.urlFilters');
    $obj->set_value('entity', '$.LiveUpdateBasicConfig.urlFilters');
    if (@$url_filters) {
	$obj->set_value('value', join(",",@$url_filters));
    }
    else {
        $obj->delete_on_save();
    } 
    eval {
      F5::ASM::COAPI::TransactionalSave([$obj]);
    };
    if ($@) {
       # an error may happen the profile could not be created, for example when the license expired
       $self->{logger}->info("cannot send updated objects to mcp: $@ $F5::COAPI::error") if $self->{logger};
    }
}

#----------------------------------------------------------------------------------------

sub inject_random_values {
    my ($self, %params) = @_;

    my $variables = $params{variables};
    my $trap = $params{trap};

    if ($trap->{injection_key}) {
        my @injection_key_vars = $trap->{injection_key} =~ /\{\{(.*?)\}\}/g;
        for my $var (@injection_key_vars) {
            my $value = $variables->{$var};
            $trap->{injection_key} =~ s/\{\{$var\}\}/$value/gi;
        }
    }

    if ($trap->{injection_value}) {
        my @injection_value_vars = $trap->{injection_value} =~ /\{\{(.*?)\}\}/g;
        for my $var (@injection_value_vars) {
            my $value = $variables->{$var};
            $trap->{injection_value} =~ s/\{\{$var\}\}/$value/gi;
        }
    }

    if ($trap->{conviction_conditions}) {
        for my $condition (@{$trap->{conviction_conditions}}) {
            my @condition_vars = $condition->{value} =~ /\{\{(.*?)\}\}/g;
            for my $var (@condition_vars) {
                my $value = $variables->{$var};
                $condition->{value} =~ s/\{\{$var\}\}/$value/gi;
            }
        }
    }
}

#----------------------------------------------------------------------------------------

sub get_random_values {
    my ($self, %params) = @_;
    my $filename = $params{file};

    my $datasyncd_profile = F5::ASM::COAPI::Query->new("datasync_global_profile", { name => $INITIAL_DATASYNC_PROFILE_NAME })->[0];

    if ($datasyncd_profile && defined $datasyncd_profile->{master_key}) {
        my $seed = F5::CRC::get_crc_any('get_crc32', $datasyncd_profile->{master_key});
        srand($seed)
    }

    my $existing_values = $self->{dbh}->selectall_hashref(<<EndQuery, 'var_name');
SELECT *
  FROM PLC.DECEPTION_TRAP_VARIABLES
EndQuery

    my $values = {};

    if ($filename) {
        my $variables_file = YAML::XS::LoadFile($filename);
        my $vars = $variables_file->{vars};

        for my $var (@$vars) {
            if (defined $existing_values->{$var->{name}}) {
                $values->{$var->{name}} = $existing_values->{$var->{name}}->{var_value};
                delete $existing_values->{$var->{name}};
                next;
            }
            $values->{$var->{name}} = $self->generate_value(%params,
                                                        var => $var);

            F5::DbUtils::insert_row(dbh => $self->{dbh},
                                    table_name => 'PLC.DECEPTION_TRAP_VARIABLES',
                                    rec => { var_name => $var->{name},
                                             var_value => $values->{$var->{name}} });
        }
    }

    my @deleted_vars = map { $_->{variable_id} } values %$existing_values;
    F5::DbUtils::delete_rows(dbh => $self->{dbh},
                             table_name => 'PLC.DECEPTION_TRAP_VARIABLES',
                             where_fields => { variable_id => \@deleted_vars },
                            );

    srand;

    return $values;
}

#----------------------------------------------------------------------------------------

sub generate_value {
    my ($self, %params) = @_;

    my $var = $params{var};

    my $value = '';
    if ($var->{list}) {
        my @list = @{$var->{list}};
        $value = $list[rand @list];
    }
    else {
        $value = F5::Utils::String::generate_random_value(length => $var->{length},
                                                          characters => $var->{characters});
    }

    return $value;
}

1;
