package F5::BigipUtils;

use strict;
use warnings;

use Memoize;
use YAML::XS;
use Scalar::Util qw(reftype);


use F5::Deploy;

memoize('is_device_bigip');
memoize('is_device_em');
memoize('is_install_mode');
memoize('is_docker');
memoize('is_feature_enabled');
memoize('waf_enabled_features');
memoize('get_default_log_level');
memoize('waf_cp_features');
memoize('get_feature_value');

my $WAF_CP_FEATURES_FILENAME = F5::Deploy::map_path('/etc/ts/tools/waf_cp_features.yaml');
my $WAF_CP_FEATURE_FIELDS_FILENAME = F5::Deploy::map_path('/etc/ts/tools/waf_cp_feature_fields.yaml');

#-------------------------------------------------------------------

sub is_install_mode {
    if ($ENV{F5_INSTALL_MODE} && $ENV{F5_INSTALL_MODE} eq 'install') {
        return 1;
    }

    return 0;
}

#-------------------------------------------------------------------

sub is_docker {
    return is_feature_enabled('standalone');
}

#-------------------------------------------------------------------

sub is_device_bigip {
    return (is_docker() || F5::BigipUtils::get_product() ) ? 1 : 0;
}

#-------------------------------------------------------------------

sub is_device_em {
    return (!is_docker() &&  F5::BigipUtils::get_product() eq 'EM' ) ? 1 : 0;
}

#-------------------------------------------------------------------

sub get_product {
    open my $version_file, "<", "/VERSION";
    my @lines = <$version_file>;
    close $version_file;
    my ($product) = map {$_, s/Product: //g} grep {/Product:/} @lines;
    chomp $product;

    return $product;
}

#-------------------------------------------------------------------

sub waf_enabled_features {
    my $waf_cp_features = waf_cp_features();
    my $feature_fields = YAML::XS::LoadFile($WAF_CP_FEATURE_FIELDS_FILENAME);
    my @enabled_features = grep { $waf_cp_features->{$_}  } @{$feature_fields->{features}};
    return @enabled_features;
}

#-------------------------------------------------------------------

sub waf_cp_features {
    if (! -f $WAF_CP_FEATURES_FILENAME) {
        return {};
    }
    return YAML::XS::LoadFile($WAF_CP_FEATURES_FILENAME);
}

#-------------------------------------------------------------------

sub get_feature_value {
    my ($feature_name, $default_value) = @_;

    my $waf_cp_features = waf_cp_features();
    return exists($waf_cp_features->{$feature_name}) ? $waf_cp_features->{$feature_name}
                                                     : $default_value;
}

#-------------------------------------------------------------------

sub get_traffic_engine {
    return get_feature_value('traffic_engine', 'tmm')
}

#-------------------------------------------------------------------

sub get_default_log_level {
    # Note: Log level can also be overridden by F5WAF_LOG_LEVEL
    # enviroment variable, which is handled by F5::Logger.
    return get_feature_value('default_log_level');
}

#-------------------------------------------------------------------

sub get_default_log_format_fields {
    return get_feature_value('default_log_format_fields');
}

#-------------------------------------------------------------------


sub get_grpc_log_format_fields {
    return get_feature_value('grpc_log_format_fields');
}

#-------------------------------------------------------------------

sub is_feature_enabled {
    my($feature_name) = @_;

    my $env_varname = "F5WAF_ENABLE_" . uc($feature_name);
    if (defined $ENV{$env_varname}) {
        return $ENV{$env_varname};
    }

    if (! -f $WAF_CP_FEATURES_FILENAME) {
        if ($feature_name eq 'standalone') {
            # Special handling for standalone flag, which should only be enabled in containers
            return 0;
        }
        else {
            return 1;
        }
    }

    my $waf_cp_features = waf_cp_features();
    if (exists $waf_cp_features->{$feature_name}) {
        return $waf_cp_features->{$feature_name} ? 1 : 0;
    }

    return 0;
}

#-------------------------------------------------------------------

sub get_docker_container_id {
    open(my $inf, '<', '/proc/self/cgroup');
    my $cgroup_line = <$inf>;
    $cgroup_line =~ m|/docker/(\w{12})|;
    return $1;
}

#-------------------------------------------------------------------

1;
