package F5::BigipVersionUtils;

use strict;
use warnings;

use Memoize;

use F5::Deploy;
use YAML::XS;

memoize('bigip_version');
memoize('current_version_codename');
memoize('get_codenames');
memoize('compare_versions');

my $CODENAMES_FILENAME = $ENV{ASM_CODENAMES_FILENAME} || F5::Deploy::map_path('/ts/share/codenames.txt');

#----------------------------------------------------------------------

sub find_by_version {
  my(%params) = @_;

  my %elements_by_version;
  for my $element (@{$params{elements}}) {
    next unless $element;
    my $version = $params{version_extract}->($element);
    next unless $version;
    $version = zero_pad_segments($version);

    $elements_by_version{$version} = $element;
  }

  my $latest_version_so_far;

  my $bigip_version = $params{bigip_version} || bigip_version();
  my $zero_pad_bigip_version = zero_pad_segments($bigip_version);

  for my $version (keys %elements_by_version) {
    if ($zero_pad_bigip_version ge $version &&
        (!defined $latest_version_so_far || 
         $version gt $latest_version_so_far)) {

      $latest_version_so_far = $version;
    }
  }

  if ($latest_version_so_far) {
    return $elements_by_version{$latest_version_so_far};
  }
  return;
}

#----------------------------------------------------------------------

sub bigip_version {
    my(%params) = @_;
    my $max_elements = defined $params{max_elements} ? $params{max_elements} : 3;
    my ($bigip_version, $bigip_build);

    if ($ENV{BIGIP_VERSION}) {
        $bigip_version = $ENV{BIGIP_VERSION};
    }
    else {
        ($bigip_version, $bigip_build) = read_version_file();
    }
    return max_elements(version => $bigip_version, max_elements => $max_elements);
}

#----------------------------------------------------------------------

sub max_elements {
    my(%params) = @_;
    my $version = $params{version};
    my $max_elements = $params{max_elements};
    if ($max_elements) {
        my @version_elements = split(/\./, $version);
        if ($max_elements <= $#version_elements) {
            splice(@version_elements, $max_elements);
        }
        return join('.', @version_elements);
    }
    return $version;
}

#----------------------------------------------------------------------

sub compare_versions {
     my (%params) = @_;
     my $version1 = zero_pad_segments($params{version1}, 3);
     my $version2 = zero_pad_segments($params{version2}, 3);
     return $version1 cmp $version2;
}

#----------------------------------------------------------------------

sub bigip_build {
  my ($bigip_version, $bigip_build) = read_version_file();

  return $bigip_build;
}

#----------------------------------------------------------------------

sub read_version_file {
  my $bigip_version = '';
  my $bigip_build = '';

  my $f5waf_version_file = F5::Deploy::get_path_by_type('share') . "/version.yaml";
  if (-e "/VERSION") {
      open(my $fh, "/VERSION");
      while (my $line = <$fh>) {
          if ($line =~ /^Version: ([\d.]+)/) {
              $bigip_version = $1;
          }
          elsif ($line =~ /^Build: ([\d.]+)/) {
              $bigip_build = $1;
          }
      }
      close $fh;
  }
  elsif (-e $f5waf_version_file) {
      my $version_data = YAML::XS::LoadFile($f5waf_version_file);
      $bigip_version = $version_data->{version};
      $bigip_build = $version_data->{build};
  }

  return ($bigip_version, $bigip_build);
}
#----------------------------------------------------------------------

sub zero_padded_major_minor {
  my($version) = @_;
  return zero_pad_segments($version,2);
}
#----------------------------------------------------------------------

sub zero_pad_segments {
  my($version, $places) = @_;
  return "" unless $version;
  my @segments = split(/\./, $version);
  if ($places) {
      splice(@segments, $places);
  }
  return join(".", map(sprintf("%.4d", $_), @segments));
}
#----------------------------------------------------------------------

sub zero_unpad_segments {
  my($version) = @_;
  my @segments = split(/\./, $version);
  return join(".",  map {$_+0} @{segments}[0..2]);
}

#----------------------------------------------------------------------

sub find_by_codename {
  my(%params) = @_;

  my %elements_by_codename;
  for my $element (@{$params{elements}}) {
    next unless $element;
    my $codename = $params{codename_extract}->($element);
    next unless $codename;

    $elements_by_codename{$codename} = $element;
  }

  my $latest_codename_so_far;

  my $bigip_codename = $params{bigip_codename} || current_version_codename($params{current_version_codename_filename});
  my %codenames_by_order = get_codenames_by_order();

  for my $codename (keys %elements_by_codename) {
    if ($codenames_by_order{$bigip_codename} >= $codenames_by_order{$codename} &&
        (!defined $latest_codename_so_far || 
         $codenames_by_order{$codename} > $codenames_by_order{$latest_codename_so_far})) {

      $latest_codename_so_far = $codename;
    }
  }

  if ($latest_codename_so_far) {
    return $elements_by_codename{$latest_codename_so_far};
  }
  return;
}

#----------------------------------------------------------------------

sub current_version_codename {
  my($filename) = @_;

  open(my $fh, '<', $filename);
  my $codename = <$fh>;
  chomp($codename);

  close $fh;

  return $codename;
}

#----------------------------------------------------------------------

sub get_codenames {
  my(%params) = @_;

  my %codenames;
  my @codenames;
  open(my $fh, '<', $CODENAMES_FILENAME);
  while(my $line = <$fh>) {
    my($codename, $version) = split(/=/, $line, 2);
    $codenames{$codename} = $version;
    push(@codenames, $codename);
  }
  close $fh;

  if ($params{output_style}) {
    if ($params{output_style} eq 'list') {
      return @codenames;
    }
  }
  else {
    return %codenames;
  }
}

#----------------------------------------------------------------------

sub get_codenames_by_order {
  my @codenames = get_codenames(output_style => 'list');

  my %codenames;
  for my $i (0..$#codenames) {
    $codenames{$codenames[$i]} = $i + 1;
  }

  return %codenames;
}

#----------------------------------------------------------------------

1;
