package F5::BotSignatures;

use strict;
use warnings;

use File::Slurp;
use File::Find;
use XML::Hash::LX;

use F5::Logger;
use F5::DbUtils;
use F5::LockFactory;

use F5::ASMConfig::Entity::PolicyBotSignature;
use F5::ASMConfig::Entity::BotSignature;
use F5::ASMConfig::Entity::Policy;
use F5::ASMConfig::Entity::Section;

use constant BOT_SIGNATURE_SIG_ID_START => 800000000;

our $BOT_SIGNATURE_ACTION_BLOCK = 'block';
our $BOT_SIGNATURE_ACTION_ALARM = 'alarm';
our $BOT_SIGNATURE_ACTION_DETECT = 'detect';
our $BOT_SIGNATURE_ACTION_IGNORE = 'ignore';

#----------------------------------------------------------------------------------------

sub load_from_xml {
    my (%params) = @_;

    my $dir = $params{dir};
    my $dbh = $params{dbh};
    my $logger = $params{logger} || F5::Logger->new('/ts/log/bot_signatures.log');

    $logger->info("Begin Loading Bot Signatures") if $logger;
    my $locker = F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
                                             logger => $logger,
                                            );
    $locker->lock();

    my $call_flags =  F5::ASMConfig::CallFlags->new();
    my $entity_class = 'F5::ASMConfig::Entity::BotSignature';
    my @bot_signatures = $entity_class->retrieve_all(dbh => $dbh,
                                                     call_flags => $call_flags,
                                                     logger => $logger,);
    my %existing_bot_signatures = map { ( $_->{bot_name} => $_ ) } @bot_signatures;

    if ( ! -d $dir ) {
        my $error_msg = "Cannot load Bot Signatures from unknown directory ($dir)";
        $logger->error($error_msg) if $logger;
        $locker->unlock();

        die "Cannot load Bot Signatures\n";
    }

    $logger->info("Loading Bot Signatures: $dir") if $logger;

    $call_flags->{add_on_updating_nonexistent_record} = 1;

    my $update_count = 0;
    my $add_count = 0;
    my $current_total_count = 0;

    find({ wanted => sub {
            my $sig_file = $File::Find::name;
            return unless (-f $sig_file);
            return unless ($sig_file =~ /\.xml$/);
            $logger->debug("Parsing Bot Signatures file $sig_file") if $logger;

            my $sig_xml = File::Slurp::read_file($sig_file);
            my $sig_hash = XML::Hash::LX::xml2hash($sig_xml);
            my $sig_rev = $sig_hash->{sig}->{rev};

            if ($sig_rev->{category_type} eq 'malicious') {
                $sig_rev->{bot_class} = F5::ASMConfig::BotClass::BOT_CLASS_MALICIOUS_BOT();
            }
            elsif ($sig_rev->{category_type} eq 'benign' && $sig_rev->{category} ne 'Search Engine') {
                $sig_rev->{bot_class} = F5::ASMConfig::BotClass::BOT_CLASS_UNTRUSTED_BOT();
            }
            else {
                $sig_rev->{bot_class} = F5::ASMConfig::BotClass::BOT_CLASS_TRUSTED_BOT();
            }
            $sig_rev->{bot_signature_id} = $sig_hash->{sig}->{'-id'};

            my $domains = [];
            if ((defined $sig_rev->{hostnames}) &&
                (ref ($sig_rev->{hostnames})) &&
                (defined $sig_rev->{hostnames}->{hostname})) {

                $domains = $sig_rev->{hostnames}->{hostname};
                if (! ref ($domains)) {
                    $domains = [$domains];
                }
            }
            $sig_rev->{hostnames} = $domains;

            $current_total_count++;
            my $bot_signature;
            eval {
                $bot_signature = F5::ASMConfig::Entity::BotSignature->new(dbh => $dbh, logger => $logger, object => $sig_rev);
            };
            if ($@) {
                $logger->error("Could not load Bot Signature $sig_file ($@)") if $logger;
                return;
            }

            my $existing_bot_signature = $existing_bot_signatures{ $bot_signature->{bot_name} };
            if ($existing_bot_signature) {
                $existing_bot_signature->{_preserved} = 1;
                my $rv = $existing_bot_signature->store_update(store_obj => $bot_signature,
                                                               suppress_events => 1,
                                                               call_flags => $call_flags,
                                                               logger => $logger,
                                                              );
                if ($rv->{db_change_made}) {
                    $update_count++;
                    $bot_signature = $existing_bot_signature;
                }
            }
            else {
                eval {
                    $bot_signature->store_add(store_obj => $bot_signature,
                                              call_flags => $call_flags,
                                              logger => $logger,
                                             );
                };
                my $err = $@;
                if ($err) {
                    if (ref($err)) {
                        $err = $err->{error_message};
                    }
                    $logger->error("Fail to create Bot Signature \"$bot_signature->{bot_name}\": $err") if $logger;
                    return;
                }
                $add_count++;
            }
        }
    }, $dir);

    my $delete_count = 0;
    for my $bot_signature_name (keys %existing_bot_signatures) {
        if (! $existing_bot_signatures{$bot_signature_name}->{_preserved}) {
            $logger->debug("Remove bot signature $existing_bot_signatures{$bot_signature_name}->{bot_signature_id}") if $logger;
            $existing_bot_signatures{$bot_signature_name}->delete();
            $delete_count++;
        }
    }

    $locker->unlock();
    $logger->info("Finish Loading BotSignatures") if $logger;

    return {
            completed_successfully => 1,
            bot_signatures_count => $current_total_count,
            bot_signatures_updated_count => $update_count,
            bot_signatures_added_count => $add_count,
            bot_signatures_deleted_count => $delete_count,
           };
}

#----------------------------------------------------------------------------------------
my %default_class_action = (
                            F5::ASMConfig::BotClass::BOT_CLASS_TRUSTED_BOT() => F5::ASMConfig::BotAction::BOT_ACTION_DETECT(),
                            F5::ASMConfig::BotClass::BOT_CLASS_UNTRUSTED_BOT() => F5::ASMConfig::BotAction::BOT_ACTION_ALARM(),
                            F5::ASMConfig::BotClass::BOT_CLASS_MALICIOUS_BOT() => F5::ASMConfig::BotAction::BOT_ACTION_BLOCK(),
                           );

sub get_bot_signatures_for_policy {
    my (%params) = @_;

    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $policy_id = $params{policy_id};

    my $bot_signatures = get_all_bot_signatures(%params);

    my $policy_bot_classes = $dbh->selectall_hashref(<<EndQuery, 'bot_class', undef, $policy_id);
SELECT *
  FROM PLC.PL_BOT_CLASSES
 WHERE policy_id = ?
EndQuery

    for my $bot_class (keys %default_class_action) {
        if (! defined $policy_bot_classes->{$bot_class} || ! defined $policy_bot_classes->{$bot_class}->{action}) {
            $policy_bot_classes->{$bot_class}->{action} = $default_class_action{$bot_class};
        }
    }

    my $policy_bot_signatures = $dbh->selectall_hashref(<<EndQuery, 'bot_signature_id', undef, $policy_id);
SELECT *
  FROM PLC.PL_BOT_SIGNATURES
 WHERE policy_id = ?
EndQuery

    for my $bot_signature (@$bot_signatures) {
        if ($policy_bot_signatures->{$bot_signature->{bot_signature_id}}) {
            $bot_signature->{action} = $policy_bot_signatures->{$bot_signature->{bot_signature_id}}->{action};
        }
        else {
            $bot_signature->{action} = $policy_bot_classes->{$bot_signature->{bot_class}}->{action};
        }
    }

    return $bot_signatures;
}

#----------------------------------------------------------------------------------------

sub get_all_bot_signatures {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $bot_signatures = $dbh->selectall_hashref(<<EndQuery, 'bot_signature_id');
SELECT *
  FROM PLC.BOT_SIGNATURES
EndQuery

    if ($params{include_hostnames}) {
        my $hosts = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT *
  FROM PLC.BOT_SIGNATURE_HOSTNAMES
EndQuery
        for my $host (@$hosts) {
            if (!defined $bot_signatures->{$host->{bot_signature_id}}->{hosts}) { 
                $bot_signatures->{ $host->{bot_signature_id}}->{hosts} = [];
            } 
            push (@{$bot_signatures->{$host->{bot_signature_id}}->{hosts}}, $host->{hostname});
        } 
    }
    my @bot_signatures_rv = values %$bot_signatures;
    for my $bot_signature (@bot_signatures_rv) { 
        $bot_signature->{sig_id}  = get_bot_signature_sig_id($bot_signature->{bot_signature_id});
    }
    return \@bot_signatures_rv;
}

#----------------------------------------------------------------------------------------

sub get_bot_signature_sig_id {
    my ($bot_signature_id) = @_;
    return $bot_signature_id + BOT_SIGNATURE_SIG_ID_START;
}

#----------------------------------------------------------------------------------------

1;
