package F5::Captcha;

use strict;
use warnings;

use File::Path;

our $CAPTCHA_OUT_DIR = F5::Deploy::get_path_by_type('bd_config_files') . '/captcha';

my $CAPTCHA_DEFAULT_GENERATE_COUNT = 120;
my $CAPTCHA_DEFAULT_GENERATE_BATCH_SIZE = 10;
my $CAPTCHA_SCRIPT_LOCATION = F5::Deploy::get_path_by_type('share') . '/captcha/securimage_show.php';

my $ALARM_AND_CAPTCHA_ACTION = 3;
my $ALARM_AND_CLIENT_SIDE_INTEGRITY_CAPTCHA_ACTION = 5;

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->{dbh} = $params{dbh};
    $self->{logger} = $params{logger};
    $self->{asmconfig_client} = $params{asmconfig_client}
}

#----------------------------------------------------------------------

sub get_captcha_command {
    my($self, %params) = @_;

    if (! exists $params{out_dir}) { die "Missing mandatory out_dir param\n"; }
    my $out_dir = $params{out_dir};
    my $generate_count = defined($params{generate_count}) ? $params{generate_count} : $CAPTCHA_DEFAULT_GENERATE_COUNT;
    my $batch_size = defined($params{batch_size}) ? $params{batch_size} : $CAPTCHA_DEFAULT_GENERATE_BATCH_SIZE;

    $batch_size = List::Util::min($batch_size, $generate_count);

    return ('php', $CAPTCHA_SCRIPT_LOCATION, '-o', $out_dir, '-p', $generate_count, '-b', $batch_size);
}

#----------------------------------------------------------------------

sub create_captcha_out_subdir {
    my($self, %params) = @_;

    my $out_dir_suffix = defined($params{out_dir_suffix}) ? $params{out_dir_suffix} :  "";
    my $timestamp = POSIX::strftime "%Y%m%d_%H%M%S", localtime;
    my $current_captcha_out_subdir = "$CAPTCHA_OUT_DIR/$timestamp$out_dir_suffix";
    mkpath($CAPTCHA_OUT_DIR) unless -d $CAPTCHA_OUT_DIR;
    mkpath($current_captcha_out_subdir);
    return $current_captcha_out_subdir;
}

#----------------------------------------------------------------------

sub generate_captchas {
    my($self, %params) = @_;

    my $check_if_captcha_in_use = defined($params{check_if_captcha_in_use}) ? $params{check_if_captcha_in_use} : 0;
    if (! $check_if_captcha_in_use && ! $self->is_captcha_in_use) {
        return;
    }

    my $generate_count = defined($params{generate_count}) ? $params{generate_count} : $CAPTCHA_DEFAULT_GENERATE_COUNT;
    my $do_set_active = defined($params{do_set_active}) ? $params{do_set_active} : 1;
    my $current_captcha_out_subdir = $self->create_captcha_out_subdir(out_dir_suffix => $params{out_dir_suffix});
    my @captcha_command = $self->get_captcha_command(out_dir => $current_captcha_out_subdir, generate_count => $generate_count);
    my $expected_generated_files_count = 3 * $generate_count; # gif, mp3, txt

    $self->{logger}->debug("Generate captchas: generating $generate_count captchas [". join(", ", @captcha_command) . "]") if $self->{logger};
    my $rv = F5::SystemUtils::run_system_tool(command_line_params => \@captcha_command);
    my @generated_files = glob("$current_captcha_out_subdir/*");
    my $generated_files_count = scalar @generated_files;
    my $code;
    if (! $rv->{completed_successfully} || $generated_files_count != $expected_generated_files_count) {
        $self->{logger}->error("Generate captchas: generator failure - generated $generated_files_count files out of $expected_generated_files_count ($generate_count captchas)") if $self->{logger};
        $code = 1;
    }
    else {
        $self->{logger}->info("Generate captchas: successfully generated $generate_count captchas") if $self->{logger};
        $code = 0;
    }

    if ($rv->{standard_error}) {
        $self->{logger}->error("Generate captchas: generator errors:\n$rv->{standard_error}") if $self->{logger};
    }
    if ($rv->{standard_output}) {
        $self->{logger}->trace("Generate captchas: generator output:\n$rv->{standard_output}") if $self->{logger};
    }


    # Reset db handle
    my $orig_dbh = $self->{dbh};
    $self->{dbh} = F5::DbUtils::get_dbh();
    $orig_dbh->{InactiveDestroy} = 1;

    if ($do_set_active) {
        my $base = F5::PrepareConf::Base->new(dbh => $self->{dbh}, logger => $self->{logger});
        F5::SetActive::set_active(dbh => $self->{dbh},
                                  logger => $self->{logger},
                                  do_update_core => 1,
                                  asm_config_handler => $self->{asmconfig_client});
    }

    return $code;
}

#----------------------------------------------------------------------

sub is_captcha_in_use {
    my($self) = @_;
    my $policy_brute_force_attribs = $self->{dbh}->selectall_hashref("SELECT * FROM PLC.PL_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES", 'policy_id');
    my %actions_by_flags = (flg_detect_distributed_brute_force_attack => 'detection_threshold_action',
                            flg_detect_credentials_stuffing_attack => 'detection_threshold_action',
                            flg_detect_login_attempts_from_the_same_user => 'login_attempts_from_the_same_user_action',
                            flg_identify_device => 'login_attempts_from_the_same_device_id_action',
                            flg_detect_login_attempts_from_the_same_ip => 'login_attempts_from_the_same_ip_action',
                            flg_client_side_integrity_bypass => 'client_side_integrity_bypass_threshold_action'
                           );
    for my $brute_force_attribs (values %$policy_brute_force_attribs) {
        my $current_policy = $brute_force_attribs;
        if ($current_policy->{flg_enabled}) {
            for my $flg (keys %actions_by_flags) {
                if ($current_policy->{$flg} &&
                ($current_policy->{$actions_by_flags{$flg}} == $ALARM_AND_CAPTCHA_ACTION || $current_policy->{$actions_by_flags{$flg}} == $ALARM_AND_CLIENT_SIDE_INTEGRITY_CAPTCHA_ACTION)) {
                    return 1;
                }
            }
        }
    }
    $self->{logger}->debug("Not generating captchas") if $self->{logger};
    return;
}

#----------------------------------------------------------------------

1;
