package F5::Certificates;

use strict;
use warnings;

use File::Copy;
use File::Basename;
use F5::DbUtils;
use Time::Local;
use IPC::Open3;
use Symbol;
use File::Spec;
use Date::Parse;
use F5::Deploy;

#------------------------------------------------------------------------------------
# Sub: verify_certificate
# Description: - verify certificate for structure/trust/expiration 
#              - check expiration date 
#              - extract expiration date
# Usage: verify_certificate(pem_file_content => $pem_file_content, logger => $logger);
#-------------------------------------------------------------------------------------
sub verify_certificate{
 my %params = @_;

 my $logger = $params{logger};
 my $pem_file_content = $params{pem_file_content};
 my $completed_successfully = 0;
 my $temp_cert_file = F5::Deploy::map_path("/ts/var/cert/temp_ssl_cert.pem");
 my $fh;

 if(! open($fh, '>', $temp_cert_file)){
       return {
          completed_successfully => $completed_successfully,
          error_message => "Could not open '$temp_cert_file': $!\n",
       }
 }
 print $fh $pem_file_content;
 close $fh;

 open(my $null1, ">", File::Spec->devnull);
 open(my $null2, File::Spec->devnull);

 my $openssl_err_fh = gensym;
 my $openssl_out_fh = gensym;

 my $openssl_pid = open3($null1, $openssl_out_fh, $openssl_err_fh,'openssl','verify', '-CApath', '/config/ssl/ssl.crt/',$temp_cert_file);
 my $openssl_err = join('', <$openssl_err_fh>);
 my $openssl_out = join('', <$openssl_out_fh>);
 waitpid($openssl_pid, 0);
 if ($?) {
     $logger->info("verify_certificate: $openssl_err") if $logger;
     return {
           completed_successfully => $completed_successfully,
           error_message => $openssl_out,
            }
 }

 my $validation_report = $openssl_out;
 my $is_untrusted;

 if($validation_report =~ /$temp_cert_file:.*error.*/s) {
    $validation_report = 'Untrusted:'.$validation_report;
    $is_untrusted = 1;
    $completed_successfully = 1;
 }elsif ($validation_report =~ /$temp_cert_file: OK/){
    $validation_report = 'Trusted';
    $completed_successfully = 1;
    $is_untrusted = 0;
 }

 my $expiration;
 my $flg_is_expired;
 my $local_expiration_date;
 my $expiration_str;


 $expiration = `openssl x509 -enddate -noout -in $temp_cert_file`;

 $expiration =~ s/notAfter=//g;
 $local_expiration_date = str2time($expiration);

 $expiration_str =  `openssl x509 -checkend 1 -in $temp_cert_file`;

 if($expiration_str =~ /Certificate will not expire/){
       $flg_is_expired = 0;
 }else{
       $flg_is_expired = 1;
 }

 unlink $temp_cert_file;

 return {
          completed_successfully => $completed_successfully,
          expiration_date_before_convert => $expiration,
          local_expiration_date => $local_expiration_date,
          validation_report => $validation_report,
          is_expired => $flg_is_expired,
          is_untrusted => $is_untrusted,
          error_message => $openssl_out,
        }
}
#----------------------------------------------------------------------------------------------------------
# Sub: upload_certificate
# Description: - based on ssl_cert_id, write the content of key_file blob into a pem file under /ts/var/cert
#              - the pem file name will be /ts/var/cert/cert_$ssl_cert_id.pem
#              - Update SSL_CERTIFICATE_FILE_PATH table with the full path for the pem file
#
# Usage: upload_certificate(ssl_cert_id => $ssl_cert_id, logger => $logger); 
#-----------------------------------------------------------------------------------------------------------

sub upload_certificate{
  my %params = @_;
  
  my $logger = $params{logger};
  my $ssl_cert_id = $params{ssl_cert_id};
  my $dbh = defined $params{dbh} ? $params{dbh} : F5::DbUtils::get_dbh();
  my $res = write_pem_file( ssl_cert_id => $ssl_cert_id, logger => $logger , dbh => $dbh);

 if($res->{completed_successfully}) {
     my $rv = update_file_path( ssl_cert_id => $ssl_cert_id, 
                                  key_file_path => $res->{certificate_filename},
                                  logger => $logger,
                                  dbh => $dbh);
    
    return $rv if(!$rv->{completed_successfully})   
 
 }

 return $res;
	
}
#----------------------------------------------------------------------
# internal functions
#----------------------------------------------------------------------
sub write_pem_file{
  my %params = @_;

  my $logger = $params{logger};
  my $ssl_cert_id = $params{ssl_cert_id};
  my $fh;

  my $certificate_filename = F5::Deploy::map_path("/ts/var/cert/ssl_cert_$ssl_cert_id.pem");

  my $dbh = $params{dbh} || F5::DbUtils::get_dbh();

  if(!$dbh){
   $logger->error("Unable to connect to database");
        return  {
              completed_successfully => 0,
              error_message => "Unable to connect to database",
             };
  }

  my $cert_properties =   $dbh->selectrow_hashref(<<EndQuery, { Slice => {} } , $ssl_cert_id);
SELECT * FROM PLC.SSL_CERTIFICATES
 WHERE ssl_cert_id = ?
EndQuery

 if($dbh->errstr){
   $logger->error("Can't read PLC.SSL_CERTIFICATES:".$dbh->errstr);
        return  {
              completed_successfully => 0,
              error_message => "PLC.SSL_CERTIFICATES:".$dbh->errstr,
             };
  }

  if(defined $cert_properties->{key_file}){
     if(! open($fh, '>', $certificate_filename)){
       return  {
              completed_successfully => 0,
              error_message => "Could not open $certificate_filename: $!",
        }
      }
     print $fh $cert_properties->{key_file};
     close $fh;

     chmod(0664, $certificate_filename);

     $logger->info("Certificate $certificate_filename written successfully.") if $logger;
     return  {
              completed_successfully => 1,
              certificate_filename => $certificate_filename,
             };
  }
  else{
     $logger->error("Certificate upload fail: can't find ssl_cert_id $ssl_cert_id.") if $logger;
     return  {
              completed_successfully => 0,
              error_message => "can't find ssl_cert_id $ssl_cert_id.",
             };

  }

}
#----------------------------------------------------------------------
sub update_file_path{
 my %params = @_;

 my $logger = $params{logger};
 my $ssl_cert_id = $params{ssl_cert_id};
 my $key_file_path = $params{key_file_path};

 my $dbh = $params{dbh} || F5::DbUtils::get_dbh();

    $dbh->do(<<EndQuery, undef, $ssl_cert_id,$key_file_path);
REPLACE INTO PLC.SSL_CERTIFICATE_FILES_PATH 
            (ssl_cert_id, key_file_path)  VALUES (?,?)
EndQuery

 if($dbh->errstr){
   $logger->error("Can't write to PLC.SSL_CERTIFICATE_FILES_PATH:".$dbh->errstr);
        return  {
              completed_successfully => 0,
              error_message => "Can't write to PLC.SSL_CERTIFICATE_FILES_PATH".$dbh->errstr,
             };
  }else{
     return  {
              completed_successfully => 1,
             };
  }
}
#----------------------------------------------------------------------
1;
