package F5::CfgConvert;

use warnings;
use strict;

use File::Basename;
use Bit::Vector;
use Fcntl 'SEEK_SET';
use MIME::Base64;
use Math::BigInt;

use F5::Deploy;
use F5::IpAddr;
use F5::BdCfgTables;
use F5::MessagePack;
use F5::Utils::Iterator;

our $BD_CFG_BASE_DIR = F5::Deploy::map_path("/usr/share/ts/share") . "/bd-cfg";
our $dcc_table_config;
our $MSG_SIGNATURE_DEFAULT = 98716425;
our $PREFIX_PACK_STRING = 'i i i';

my %types = (
             'binary' => {
                          pack_as => 'a',
                          field_length => 1,
                          fixup_func => sub {
                              my($val) = @_;
                              if (utf8::is_utf8($val)) {
                                  # Treat scalar as bytes, not utf8
                                  # See https://jira.pdsea.f5net.com/browse/MBIPWAF-1980
                                  utf8::encode($val);
                              }
                              return $val;
                          },
                         },
             'bitmap64' => {
                            pack_as => 'L L',
                            field_length => 8,
                            fixup_func => sub {
                                my $vec = Bit::Vector->new_Bin(64, shift);
                                return ($vec->Chunk_Read(32, 0), $vec->Chunk_Read(32, 32));
                            },
                           },
             'bitmap' => {
                            pack_as => 'L L',
                            field_length => 8,
                         },
             'byte' => {
                        pack_as => 'C',
                        field_length => 1,
                       },
             'byte256' => {
                           pack_as => 'c256',
                           field_length => 256,
                           fixup_func => sub {
                               return split(//, shift);
                            },
                       },
             'char' => {
                        pack_as => 'a',
                        field_length => 1,
                        fixup_func => sub {
                            my($val) = @_;
                            if (utf8::is_utf8($val)) {
                                # Treat scalar as bytes, not utf8
                                # See https://jira.pdsea.f5net.com/browse/MBIPWAF-1980
                                utf8::encode($val);
                            }
                            return $val;
                        },
                       },
             'crc' => {
                       pack_as => 'L L ',
                       fixup_func => sub {
                           my $vec = Bit::Vector->new_Dec(64, shift);
                           return ($vec->Chunk_Read(32, 0), $vec->Chunk_Read(32, 32));
                       },
                       field_length => 8,
                      },
             'crc16' => {
                         pack_as => 'S',
                         field_length => 2,
                        },
             'crc32' => {
                         pack_as => 'L',
                         field_length => 4,
                        },
             'double' => {
                          pack_as => 'd', # todo:rss: is this long double
                          field_length => 8,
                         },
             'int' => {
                       pack_as => 'i',
                       field_length => 4,
                       empty_value => 0,
                      },
             'ip' => {
                      pack_as => 'a16',
                      field_length => 16,
                      fixup_func => sub {
                          return F5::IpAddr::ip_pton(shift);
                      }
                     },
             'long' => {
                        pack_as => 'L',
                        field_length => 4,
                        empty_value => 0,
                       },
             'mac' => {
                       pack_as => 'H12',
                       field_length => 6,
                       fixup_func => sub {
                           my $val = shift;
                           $val =~ s/://g;
                           return $val;
                       },
                      },
             'short' => {
                         pack_as => 's',
                         field_length => 2,
                        },
             'unsigned_char' => {
                                 pack_as => 'C',
                                 field_length => 1,
                                },
             'unsigned_int' => {
                                pack_as => 'I',
                                field_length => 4,
                                empty_value => 0,
                               },
             'unsigned_short' => {
                                  pack_as => 'S',
                                  field_length => 2,
                                  empty_value => 0,
                                 },
            );

$types{bytes} = $types{binary};
$types{string} = $types{char};

my $BD_CONFIG_ACTION_ADD = 1; # Bd::Config::ConfigAction::ADD(),
my $BD_CONFIG_ACTION_UPDATE = 2; # Bd::Config::ConfigAction::UPDATE()
my $BD_CONFIG_ACTION_DELETE = 3; # Bd::Config::ConfigAction::DELETE()

my %PROTOBUF_OPERATION_MAP = (
                              '1' => $BD_CONFIG_ACTION_ADD,
                              '+1' => $BD_CONFIG_ACTION_ADD,
                              '-1' => $BD_CONFIG_ACTION_DELETE,
                              '0' => $BD_CONFIG_ACTION_UPDATE,
                             );

my %OPERATION_CODES = ('add' => '+1',
                       'delete' => '-1',
                       'update' => '0',
                      );

#----------------------------------------------------------------------

sub convert_table {
    my(%params) = @_;
    my $from_file = $params{from_file};

    my $dcc_table_config = F5::BdCfgTables::get_dcc_table_config();

    my $from_file_cfg = basename($from_file);
    $from_file_cfg =~ s/\.([^.]+)$//;
    my $file_suffix = $1;
    $from_file_cfg =~ s/__account_(\d+)$//;
    $from_file_cfg =~ s/__logprofile_(\d+)$//;
    $from_file_cfg =~ s/^.*__//;

    my %binary_convert_opts;
    if ($from_file_cfg eq 'bd_cfg_manifest') {
        $from_file_cfg = "bd_cfg_manifest_reference";
        $binary_convert_opts{recs} = [ { action => 'add',
                                         element => { bd_cfg_manifest_filename => $from_file },
                                       }
                                     ];
    }
    elsif ($file_suffix eq 'dat') {
        $from_file_cfg = "proto_file_reference";
        $binary_convert_opts{recs} = [ { action => 'add',
                                         element => { proto_bin_filename => $from_file },
                                       }
                                     ];
    }

    my($table_def) = grep { $from_file_cfg eq $_->{output_file} } @$dcc_table_config;

    return unless $table_def;

    return binary_convert(%params,
                          %binary_convert_opts,
                          table_def => $table_def,
                         );
}

#----------------------------------------------------------------------

sub binary_convert {
    my(%params) = @_;

    my $table_def = $params{table_def};
    my $recs = $params{recs};
    my $from_file = $params{from_file};
    my $to_file = $params{to_file};
    my $return_as_string = $params{return_as_string};
    my $cfg_logger = $params{cfg_logger};
    my $retval = '';
    my $outf;

    my $operation_count;

    $cfg_logger->trace("Converting BD config from $from_file") if $cfg_logger;
    if ($to_file) {
        open($outf, '>', $to_file);
        syswrite($outf, pack('L a36', $MSG_SIGNATURE_DEFAULT, ''));
        syswrite($outf, pack('i i i', 0, 0, 0)); # place holder for block count - only determined by the end
    }

    my $binary_convert_cfg = get_binary_convert_cfg(table_def => $table_def);
    my $pack_string = $binary_convert_cfg->{pack_string};
    my $row_length = $binary_convert_cfg->{row_length};

    my @ordered_field_names = map { $_->{name} } @{$table_def->{message_fields}};

    my $full_file_iterator;
    if ($recs) {
        $full_file_iterator = F5::Utils::Iterator->new(recs => $recs);
    }
    else {
        $full_file_iterator = F5::MessagePack::Iterator->new(filename => $from_file);
    }

    my $row_count = 0;
    while(my $row = $full_file_iterator->get_next()) {

        my @field_vals = map { $row->{element}->{$_} } @ordered_field_names;
        if (exists $row->{action}) {
            unshift @field_vals, $OPERATION_CODES{$row->{action}};
        }

        my $operation;
        if ($table_def->{send_all_data_in_update}) {
            $operation = $BD_CONFIG_ACTION_ADD;
        }
        else {
            my $rec_operation = $field_vals[0];
            $operation = $PROTOBUF_OPERATION_MAP{ $rec_operation };
            $operation ||= $BD_CONFIG_ACTION_ADD;
        }

        my $skip_line = 0;
        my $i = $#field_vals;
        while ($i >= 0) {
            my $field_def = $binary_convert_cfg->{field_def}->[$i];

            if ($field_def->{fixup_func}) {
                my @rv;
                eval {
                    @rv = $field_def->{fixup_func}->($field_vals[$i]);
                };
                if (my $err = $@) {
                    $cfg_logger->error("Error encountered while handling config line element = $i ($err)") if $cfg_logger;
                    $skip_line = 1;
                    last;
                }

                splice(@field_vals, $i, 1, @rv);
            }

            if ($field_vals[$i] eq '' && exists $field_def->{empty_value}) {
                $field_vals[$i] = $field_def->{empty_value};
            }

            $i--;
        }
        next if $skip_line;
        $operation_count->{$operation}++;

        my $encoded_row = pack($pack_string, @field_vals);
        $row_count++;

        if ($to_file) {
            syswrite($outf, $encoded_row);
        }

        if ($return_as_string) {
            $retval .= $encoded_row;
        }
    }
    $cfg_logger->trace("Completed BD config from $from_file") if $cfg_logger;

    if ($row_count == 0) {
        # Old cfg_conventer sends row_length = 0 when there are no records
        $row_length = 0;
    }

    my $prefix = pack($PREFIX_PACK_STRING,
                      $table_def->{dcc_table_id},
                      $row_length,
                      $row_count);
    if ($to_file) {
        # Rewrite block size and row count
        sysseek($outf, 40, SEEK_SET);
        syswrite($outf, $prefix, 12);

        close $outf;
    }

    if ($return_as_string) {
        $retval = $prefix . $retval;
    }

    my %metadata = (
        added_records => $operation_count->{ $BD_CONFIG_ACTION_ADD } || 0,
        updated_records => $operation_count->{ $BD_CONFIG_ACTION_UPDATE } || 0,
        deleted_records => $operation_count->{ $BD_CONFIG_ACTION_DELETE } || 0,
    );

    return wantarray ? ($retval, %metadata) : $retval ;
}

#----------------------------------------------------------------------

sub get_binary_convert_cfg {
    my(%params) = @_;

    my $table_def = $params{table_def};

    if (! $table_def->{binary_convert_cfg}) {

        my @field_def;
        my $pack_string;
        my $row_length;
        if (! $table_def->{send_all_data_in_update}) {
            $pack_string = 'c';
            $row_length = 1;
            push(@field_def, $types{'byte'});
        }

        for my $field (@{$table_def->{message_fields}}) {
            my $data_type = $field->{data_type};
            my $field_length;
            if ($types{$data_type}) {
                $field_length = $types{$data_type}->{field_length};
                $pack_string .= $types{$data_type}->{pack_as} . ' ';
            }
            else {
                $data_type =~ s/(\d+)$//;
                $field_length = $1;
                $pack_string .= $types{$data_type}->{pack_as} . $field_length . ' ';
            }
            $row_length += $field_length;

            push(@field_def, $types{$data_type});
        }

        $table_def->{binary_convert_cfg} = {
                                            field_def => \@field_def,
                                            row_length => $row_length,
                                            pack_string => $pack_string,
                                           };
    }

    return $table_def->{binary_convert_cfg};
}

#----------------------------------------------------------------------

1;
