package F5::Cgc::Channel::DeviceGroup;

use strict;
use warnings;

use F5::Logger;
use IPC::ShareLite;
use F5::DeviceGroup;
use F5::Cgc;

use base qw/F5::Cgc::Channel/;
my $initial_config = 0;
my $last_config_change;
my $shared_config_change;
my $topology_change_subscriber;
my $topology_change_publisher;
my $device_group_members_config_string = '';

my $STANDARD_TOPOLOGY_CHANGE = 1;
my $SELF_TOPOLOGY_CHANGE = 2;

our $ASM_CGC_LOW_PRI_TO_ALL = 'asm_cgc_low_pri_to_all';
our $ASM_CGC_HIGH_PRI_TO_SINGLE = 'asm_cgc_high_pri_to_single';
our $ASM_CGC_TOPOLOGY = 'asm_cgc_topology';
our $ASM_CGC_MCP_EVENTS = 'asm_cgc_mcp_events';
our $ASM_CGC_LOCAL_GROUP = 'ASM_local_ASM_channels_ASM';

our $F5_CGC_DAEMON_IP = '127.0.0.1';

my %services = (
                $ASM_CGC_LOW_PRI_TO_ALL => 6123,
                $ASM_CGC_HIGH_PRI_TO_SINGLE => 6126,
                $ASM_CGC_TOPOLOGY => 6127,
                $ASM_CGC_MCP_EVENTS => 6128,
               );

#----------------------------------------------------------------------

sub reset {
    my($class, %params) = @_;
    $initial_config = 0;
    undef $last_config_change;

    my $do_reset = $params{do_reset};
    if ($topology_change_subscriber) {
        $topology_change_subscriber->close();
        undef $topology_change_subscriber;
        $do_reset = 1;
    }
    if ($topology_change_publisher) {
        $topology_change_publisher->close();
        undef $topology_change_publisher;
        $do_reset = 1;
    }

    if ($do_reset) {
        F5::Cgc::reset();
    }
}

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  if (! $params{logger}) {
    $params{logger} = F5::Logger->get_logger(logger_name => 'asm_config_server.pl');
  }
  my $logger = $params{logger};
  reinit(%params);
  my $self = $class->SUPER::new(%params);
  if (! $self) {
      return;
  }
  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub has_config_changed {
    my (%params) = @_;
    my $shared_config_time = $shared_config_change->fetch || 0;
    return (defined $last_config_change && $last_config_change != $shared_config_time);
}

#----------------------------------------------------------------------

sub reinit {
    my (%params) = @_;
    my $logger = $params{logger};

    $logger->trace("Got reinit from '" . (caller(1))[3] . "' line " . (caller(1))[2]) if $logger;

    for my $iteration (1,2) {
        eval {
            startup(logger => $logger);
            my $shared_config_time = $shared_config_change->fetch || 0;
            if (defined $last_config_change && $last_config_change == $shared_config_time) {
                $logger->trace("Last configuration change at $last_config_change already handled -- skipping device group topology adjustment") if $logger;
                return;
            }

            # Create general "Groups" for publisher and subscriber
            my @endpoints;
            my $local_device_name = F5::DeviceGroup::get_local_device_name();

            if (! $local_device_name) {
                $logger->debug("Could not get the local device name -- skipping device group topology adjustment") if $logger;
                return;
            }

            my $mcp_device_group = F5::DeviceGroup::get_device_group($local_device_name);
            my $device_group = $mcp_device_group->{name};

            my $local_ip = $F5_CGC_DAEMON_IP;

            my @config = (
                          {group => $ASM_CGC_LOCAL_GROUP },
                          {'local' => "tcp://$local_ip" },
                         );
            if ($params{local_device_name} && $local_device_name ne $params{local_device_name}){
                push (@config, 
                     {
                      device_name => $params{local_device_name},
                      ip_address => "tcp://$local_ip",
                      groups => [$ASM_CGC_LOCAL_GROUP],
                     });           
            }

            my %devices;
            if ($device_group) {
                %devices = map { ($_ => [ $device_group ]) } F5::DeviceGroup::get_devices_in_device_group($device_group);
            }
            push @{$devices{$local_device_name}}, $ASM_CGC_LOCAL_GROUP;
            $logger->trace(\%devices) if $logger;

            $device_group_members_config_string = '';
            for my $device_name (sort keys %devices) {
                if (!$device_name) {
                    $logger->debug("Got an empty device name, skipping") if $logger;
                    next;
                }
                my $device = F5::DeviceGroup::get_device($device_name);
                my $ip_address = $device->{configsync_ip};

                my %extra_params;
                if ($local_device_name eq $device_name) {
                    $extra_params{is_local} = 1;
                    $ip_address = $local_ip;
                }
                elsif ($ip_address eq "::") {
                    next;
                }
                else {
                    $device_group_members_config_string .= F5::DeviceGroup::get_device_compare_string($device);
                }

                push(@config,
                     {
                      device_name => $device_name,
                      ip_address => "tcp://$ip_address",
                      groups => $devices{$device_name},
                      %extra_params,
                     });
            }
            $logger->trace(\@config) if $logger;
            F5::Cgc::update_hosts(config => \@config, logger => $logger);
            $last_config_change = $shared_config_time;
        };
        if ($@) {
            $logger->info("Failed to adjust device group topology adjustment on attempt $iteration ($@)") if $logger;
        }
        else {
            last;
        }
    }
}

#----------------------------------------------------------------------

sub get_device_group_members_config_string {
    return $device_group_members_config_string;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{logger} = $params{logger};
    $self->{received_messages} = {};
}

#----------------------------------------------------------------------

sub startup {
    my (%params) = @_;
    my $logger = $params{logger};

    if ($initial_config) {
        $logger->debug("Already initialized. Nothing to do") if $logger;
        return;
    }
    $initial_config = 1;

    $logger->debug("Going to initialize publisher and subscriber services") if $logger;
    for my $service (keys %services) {
        F5::Cgc::add_service(logger => $logger,
                             service_name => $service,
                             channel_type_1 => "publisher",
                             channel_port_1 => $services{$service},
                             channel_type_2 => "subscriber",
                             channel_port_2 => $services{$service},
                            );
    }

    $shared_config_change = IPC::ShareLite->new(
                                                -key => 6123,
                                                -create => 'yes',
                                                -destroy => 'no'
                                               );

    $logger->debug("Done initializing publisher and subscriber services") if $logger;
}

#----------------------------------------------------------------------

sub poll {
    my($self_or_class, %params) = @_;
    my $logger = $params{logger} || $self_or_class->{logger};

    reinit(logger => $logger);
    if (! $topology_change_subscriber) {
        $logger->debug("Creating a new topology change subscriber") if $logger;
        $topology_change_subscriber = F5::Cgc::Channel->new(group_id => $ASM_CGC_LOCAL_GROUP,
                                                            service => $ASM_CGC_TOPOLOGY,
                                                            channel_type => 'subscriber',
                                                            logger => $logger,
                                                            is_listen => 1);
    }

    my @waiting_channels;
    while (1) {
        @waiting_channels = $self_or_class->SUPER::poll(channels => [ $topology_change_subscriber,
                                                              @{$params{channels}} ],
                                                timeout => $params{timeout},
                                                logger => $logger,
                                              );
        if (@waiting_channels && $waiting_channels[0] eq $topology_change_subscriber) {
            $logger->debug("Polled a topology change subscriber, going to receive a message") if $logger;
            my $topology_message;
            eval {
                $topology_message = $topology_change_subscriber->receive;
            };
            if (my $err = $@) {
                $logger->error("Error on receiving a topology change message: $err") if $logger;
                $logger->trace(\%params) if $logger;
                die $err;
            }
            if ($topology_message) {
                $logger->debug("Received a topology change message, going to reinit") if $logger;
                reinit(logger => $logger);
                $topology_change_subscriber->close();
                undef $topology_change_subscriber;
                ${$params{resubscribe_ref}} = 1;
                return;
            }
        }
        else {
            last;
        }
    }
    $logger->debug("Done polling") if $logger;
    $logger->trace(\@waiting_channels) if $logger;
    return @waiting_channels;
}

#----------------------------------------------------------------------

sub send {
    my($self, @params) = @_;
    my $logger = $self->{logger};

    reinit(logger => $logger);
    return $self->SUPER::send(@params);
}

#----------------------------------------------------------------------

sub receive {
    my($self, @params) = @_;
    my $logger = $self->{logger};

    reinit(logger => $logger);
    return $self->SUPER::receive(@params);
}

#----------------------------------------------------------------------
sub update_topology {
    my (%params) = @_ ;
    my $logger = $params{logger};
    my $msg = $params{msg} || $STANDARD_TOPOLOGY_CHANGE;

    $logger->debug("Topology change type is " . ($msg == $STANDARD_TOPOLOGY_CHANGE ? "'standard'" : "'self'" )) if $logger;

    reinit(%params);

    my $now = time();
    $shared_config_change->store($now);
    if (! $topology_change_publisher) {
        $logger->debug("Creating new topology change publisher. Timestamp '$now'") if $logger;
        $topology_change_publisher = F5::Cgc::Channel->new(group_id => $ASM_CGC_LOCAL_GROUP,
                                                           service => $ASM_CGC_TOPOLOGY,
                                                           logger => $logger,
                                                           channel_type => 'publisher');
    }
    if ($topology_change_publisher) {
        $logger->debug("Sending topology change message now") if $logger;
        eval {
            $topology_change_publisher->send($msg);
        };
        if (my $err = $@) {
            $logger->error("Error on sending a topology change message: $err") if $logger;
            $logger->trace(\%params) if $logger;
            die $err;
        }
        else {
            $logger->debug("Successfully sent a topology change message") if $logger;
        }
    }
    else {
        $logger->error("Could not create a new topology change publisher") if $logger;
    }

    return;
}

#----------------------------------------------------------------------
sub update_topology_self {
    my (%params) = @_;
    return update_topology(msg => $SELF_TOPOLOGY_CHANGE,
                           %params);
}
#----------------------------------------------------------------------

sub DESTROY {
    if ($topology_change_publisher) {
        sleep 1;
    }

    if ($topology_change_publisher) {
        my $logger = F5::Logger->get_logger(logger_name => 'asm_config_server.pl');
        $logger->debug("Still have a topology change publisher at the end of DESTROY") if $logger;
    }
}

1;
