package F5::Cluster::Sync;

use strict;
use warnings;

use F5::Deploy;
use F5::Cluster::Utils;
use F5::ConfigSync;
use F5::PolicyHistory;
use F5::Policy;
use F5::SetActive;
use F5::BigipVersionUtils;
use F5::BigipUtils;
use F5::DbUtils;
use F5::Utils::File;
use F5::LockFactory;

use File::Basename;
use File::stat;
use File::Copy;
use POSIX qw(strftime);
use Config::IniFiles;
use Time::HiRes qw(gettimeofday);
use YAML::XS;

my $TEMP_DIR = F5::Deploy::map_path("/ts/var/cluster/temp");
my $SEND_DIR = F5::Deploy::map_path("/ts/var/cluster/send");
my $STATUS_DIR = F5::Deploy::map_path("/ts/var/cluster/status");

my $APPLY_SYNC_LOCK_FILE = F5::Deploy::map_path("/ts/lock/apply_sync.lock");

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;

  my $self = {};
  bless $self, $class;
  return $self->init(%params);
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  # logger can be passed from the caller, but ideally clustering logging should be used instead
  $self->{logger} = $params{logger} || F5::Logger->get_logger(logger_name => 'clustering');

  my $config = Config::IniFiles->new( -file => F5::Deploy::map_path('/usr/share/ts/config/clustering.ini'));
  $self->{policy_size_percent_threshold} = $config->val('Clustering', 'policy_size_percent_threshold');
  $self->{time_to_wait_for_startup} = $config->val('Clustering', 'time_to_wait_for_startup');
  $self->{copy_sync_file_to_temp_dir} = $config->val('Clustering', 'copy_sync_file_to_temp_dir', 1);

  $self->{dbh} = $params{dbh} || F5::DbUtils::get_dbh();

  return $self;

}

#----------------------------------------------------------------------

sub send_sync {
  my($self, %params) = @_;

  my $logger = $self->{logger};
  my $dbh = $self->{dbh};
  my $slot_number = F5::Cluster::Utils::get_slot_number();

  $logger->debug("About to create sync file") if $logger;
  $self->cleanup_sync_files;
  $self->cleanup_temp_files;

  if (F5::BigipUtils::is_docker() ||
      !F5::Cluster::Utils::is_clustering_enabled() ||
      !F5::Cluster::Utils::is_master_blade()) {

    $logger->debug("Sync file creation skipped") if $logger;
    return;
  }

  my $export_filename;
  my $sync_type;
  my $create_datetime = time;

  my $current_table_checksums = $self->generate_table_checksums;
  if ($params{policy_id} &&
      (! $params{force_full_sync}) && 
      are_hashes_equivalent($current_table_checksums, $self->get_db_table_checksums) && 
      $self->is_policy_update_more_efficient(policy_id => $params{policy_id})) {

    # Generate only the policy export file
    $sync_type = 'policy';
    $export_filename = "sync_" . filename_formatted_datetime($create_datetime) . "__slot_${slot_number}__account_" . $params{policy_name_crc} . "__policy_" . $params{policy_id};
    $logger->info("Sync file is policy export (policy_id = $params{policy_id}, filename = $export_filename)") if $logger;
  }
  else {
    # Generate complete UCS
    $sync_type = 'full';
    $export_filename = "sync_" . filename_formatted_datetime($create_datetime) . "__slot_${slot_number}__full_update";
    $logger->info("Sync file is full UCS (filename = $export_filename)") if $logger;
  }


  # Store in db on master
  my %sync_file_rec;
  $sync_file_rec{sync_filename} = "$SEND_DIR/$export_filename";
  $sync_file_rec{create_datetime} = mysql_formatted_datetime($create_datetime);
  $sync_file_rec{status} = 'primary';
  if ($sync_type eq 'policy') {
    $sync_file_rec{policy_id} = $params{policy_id};
    $sync_file_rec{policy_name_crc} = $params{policy_name_crc};
  }

  $self->insert_row(table => 'DCC.CLUSTER_SYNC_FILES',
                    rec => \%sync_file_rec);
 
  if ($sync_type eq 'policy') {
      my $policy = F5::Policy->new(policy_id => $params{policy_id},
                                   dbh => $dbh);
      if (-e $policy->history->get_last_revision_file) {
          copy($policy->history->get_last_revision_file, "$TEMP_DIR/$export_filename");
      }
      else {
          $policy->export_to_file("$TEMP_DIR/$export_filename");
      }
  }
  else {  #$sync_type eq 'full'
      my $configsync = F5::ConfigSync->new(dbh => $dbh,
                                           logger => $self->{logger},
                                           base_dir => "$TEMP_DIR/ucs_sync_file_temp.$$",
                                           include_cluster_tables => 1);
      $configsync->save_to_file(filename => "$TEMP_DIR/$export_filename");
      $self->save_table_checksums;
  }

  move("$TEMP_DIR/$export_filename", "$SEND_DIR/$export_filename");

  $logger->debug("Sync file creation complete") if $logger;
  return {
          completed_successfully => 1,
          sync_file_type => $sync_type,
          sync_filename => "$SEND_DIR/$export_filename",
         };
}

#----------------------------------------------------------------------

sub generate_table_checksums {
  my($self) = @_;
  my $dbh = $self->{dbh};

  my @ignore_tables = qw(PLC.PL_POLICY_HISTORY
                         PLC.PL_POLICIES
                         PLC.PL_POLICY_EM_TRANSACTIONS
                         DCC.ACCOUNTS
                       );
  # todo: DCC.ACCOUNTS really needs to be considered -- we need to remove policy_apply_date for that to happen
  my %ignore_tables = map { ($_ => 1) } @ignore_tables;

  my $table_list = join(',', grep(!$ignore_tables{$_}, F5::DbUtils::get_nonpolicy_config_sync_tables()));
  my $checksums = {};

  my $checksums_arrayref = $dbh->selectcol_arrayref("CHECKSUM TABLE $table_list", { Columns=>[1,2] });
  if ($checksums_arrayref) {
    $checksums = { @$checksums_arrayref};
  }
  return $checksums;
}

#----------------------------------------------------------------------

sub get_db_table_checksums {
  my($self) = @_;
  my $dbh = $self->{dbh};

  my $db_checksums = {};

  my $db_checksums_arrayref = $dbh->selectcol_arrayref("SELECT table_name, checksum_value FROM DCC.NON_POLICY_TABLE_CHECKSUMS", { Columns=>[1,2] });
  if ($db_checksums_arrayref) {
    $db_checksums = { @$db_checksums_arrayref };
  }
  return $db_checksums;
}

#----------------------------------------------------------------------

sub save_table_checksums {
  my($self) = @_;

  my $dbh = $self->{dbh};
  $dbh->do('DELETE FROM DCC.NON_POLICY_TABLE_CHECKSUMS');

  my $checksums = $self->generate_table_checksums;
  for my $table (keys %$checksums) {
    $dbh->do(<<EndQuery, undef, $table, $checksums->{$table});
INSERT INTO DCC.NON_POLICY_TABLE_CHECKSUMS 
  (table_name, checksum_value) 
  VALUES (?,?)
EndQuery
  }
}

#----------------------------------------------------------------------

sub are_hashes_equivalent {
  my($hash_a, $hash_b) = @_;
  my %keys_seen;
  for my $k ((keys %$hash_a, keys %$hash_b)) {
    $keys_seen{$k}++;
  }

  my @unmatched_keys = grep($keys_seen{$_} != 2, keys %keys_seen);
  return 0 if @unmatched_keys;

  for my $k (keys %$hash_a) {
    return 0 if ($hash_a->{$k} ne $hash_b->{$k});
  }
  return 1;
}

#----------------------------------------------------------------------

sub apply_sync {
  my($self, %params) = @_;

  my $logger = $self->{logger};
  my $dbh = $self->{dbh};
  my $max_sync_file_handled_count = $params{max_sync_file_handled_count};

  if (! F5::Cluster::Utils::is_clustering_enabled() ||
      F5::Cluster::Utils::is_master_blade()) {

      return;
  }

  my $locker = F5::LockFactory::get_locker(lock_filename => $APPLY_SYNC_LOCK_FILE,
                                           logger => $logger);
  $locker->lock;

  my $sync_dir = $params{sync_dir} || $SEND_DIR;

  $logger->debug("Checking for sync files (sync_dir = $sync_dir)") if $logger;

  my $sync_file_handled_count = 0;
  my $last_sync_file_handled = '';
  SYNC_FILE_QUEUE:
  while (! $max_sync_file_handled_count || $sync_file_handled_count < $max_sync_file_handled_count) {
      $self->cleanup_status_db;
      $self->cleanup_status_files;
      $self->cleanup_temp_files;
      my @sync_files = $self->get_sync_files;
      if (! @sync_files) {
          last;
      }
      my $sync_file = shift(@sync_files);

      $logger->info("Checking sync file (sync_file = $sync_file->{sync_filename}, " . (scalar @sync_files)  . " remaining sync files in queue)") if $logger;
      if (@sync_files) {
          $logger->trace("Remaining sync files in queue: " . join(', ', map { $_->{sync_filename} } @sync_files)) if $logger;
      }

      # Check for superseding sync file
      for my $other_sync_file (@sync_files) {
          if (($sync_file->{sync_type} eq 'policy' &&
               $other_sync_file->{sync_type} eq 'policy' &&
               $other_sync_file->{policy_name_crc} == $sync_file->{policy_name_crc}) ||
              $other_sync_file->{sync_type} eq 'full') {

              $logger->info("Sync file superseded by later file (other_sync_file = $other_sync_file->{sync_filename})") if $logger;

              $self->update_sync_status(sync_file => $sync_file,
                                        status => 'skipped');
              next SYNC_FILE_QUEUE;
          }
      }
      if ($self->{copy_sync_file_to_temp_dir}) {
          $sync_file->{sync_temp_filename} = $TEMP_DIR . "/" . basename($sync_file->{sync_filename});
          $logger->trace("Copying sync file to temp directory ($sync_file->{sync_temp_filename})") if $logger;
          my $sync_file_stat = stat($sync_file->{sync_filename});
          my $rv = copy($sync_file->{sync_filename}, $sync_file->{sync_temp_filename});
          my $copy_sync_file_stat = stat($sync_file->{sync_temp_filename});

          if (! $rv) {
              $logger->info("Sync file could not be copied ($!) -- skipping") if $logger;
              unlink $sync_file->{sync_temp_filename};
              next;
          }

          if (! $sync_file_stat || ! $copy_sync_file_stat || $sync_file_stat->size != $copy_sync_file_stat->size) {
              $logger->info("Sync file size discrepancy observed -- skipping") if $logger;
              unlink $sync_file->{sync_temp_filename};
              next;
          }
          $sync_file->{sync_load_source_filename} = $sync_file->{sync_temp_filename};
      }
      else {
          $sync_file->{sync_load_source_filename} = $sync_file->{sync_filename};
      }

      $sync_file_handled_count++;
      $last_sync_file_handled = $sync_file->{sync_filename};

      $self->update_sync_status(sync_file => $sync_file,
                                status => 'loading');

      my $time_waited_for_system_start = 0;
      my $iteration_sleep_time = 5;

      if ($sync_file->{sync_type} eq 'policy') {
          my $rv;
          my $db_locker;

          eval {
              $logger->info("Applying sync file $sync_file->{sync_filename} (policy)") if $logger;
              require F5::ImportExportPolicy;

              my $lock_file =  $F5::DbUtils::LOCK_FILE;
              $db_locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                                          logger => $self->{logger});

              $db_locker->lock();

              # Check that account is present on system
              my($db_policy_name_crc) = $dbh->selectrow_array(<<EndQuery, undef, $sync_file->{policy_name_crc}, $sync_file->{policy_id});
SELECT policy_name_crc
  FROM PLC.PL_POLICIES
 WHERE policy_name_crc = ?
   AND id = ?
EndQuery
              if (! defined $db_policy_name_crc) {
                  $logger->info("Account $sync_file->{policy_name_crc} was not found on this blade -- skipping sync policy file load") if $logger;
                  $self->update_sync_status(sync_file => $sync_file, status => 'skipped');
                  $rv = 1; # Skipped import should not be considered an error
              }
              else {

                  my $import_policy = F5::ImportExportPolicy->new(dbh => $dbh,
                                                                  logger => $logger);
    
                  $rv = $import_policy->import_policy(filename => $sync_file->{sync_load_source_filename},
                                                      policy_name_crc => $sync_file->{policy_name_crc},
                                                      force_policy_id => $sync_file->{policy_id},
                                                      force_policy_name_flg => 1,
                                                      overwrite => 1,
                                                     );
    
                  if (!$rv->{completed_successfully}) {
                      $db_locker->unlock();
                      die "Import policy failed\n";
                  }
    
                  my $policy = F5::Policy->new(
                                               policy_id   => $sync_file->{policy_id},
                                               dbh         => $dbh,
                                              );
                  $rv = F5::SetActive::set_active(
                                                  dbh                 => $dbh,
                                                  logger              => $logger,
                                                  policy              => $policy,
                                                  policy_name_crc          => $sync_file->{policy_name_crc},
                                                  send_sync           => 0,
                                                  asm_config_handler  => $params{asm_config_handler},
                                                 );
              }
          };

          if ((!$rv) || $@) {
              $logger->error("Applying sync file failed") if $logger;
              $logger->error($@) if $@ && $logger;
              $self->update_sync_status(sync_file => $sync_file, status => 'error');
              $db_locker->unlock();
              $locker->unlock;
              return;
          }
          $db_locker->unlock();
      }
      else {
          $logger->info("Applying sync file $sync_file->{sync_filename} (full UCS)") if $logger;
          my $configsync = F5::ConfigSync->new(dbh => $dbh,
                                               logger => $logger,
                                               base_dir => "$TEMP_DIR/ucs_sync_file_temp.$$",
                                               bigip_version => F5::BigipVersionUtils::bigip_version(),
                                               ucs_version => F5::BigipVersionUtils::bigip_version(),
                                               include_cluster_tables => 1,
                                               allow_upgrade => 0, # Cluster sync files should never lead to upgrade
                                               asm_config_handler => $params{asm_config_handler});
          my $rv;
          eval {
              $rv = $configsync->load_from_file(filename => $sync_file->{sync_load_source_filename},
                                                check_asm_ready => $params{check_asm_ready},
                                               );
          };
          if (!$rv) {
              $logger->error("Set active failed") if $logger;
              $logger->error($@) if $@ && $logger;
              $self->update_sync_status(sync_file => $sync_file, status => 'error');
              $locker->unlock;
              return;
          }
      }

      $logger->debug("Sync file completed") if $logger;
      $self->update_sync_status(sync_file => $sync_file,
                                status => 'completed');

      if ($self->{copy_sync_file_to_temp_dir}) {
          # Delete temp copy of sync file
          $logger->trace("Deleting sync file temp copy $sync_file->{sync_temp_filename}") if $logger;
          unlink $sync_file->{sync_temp_filename};
      }

  }

  $locker->unlock;
  return { sync_file_handled_count => $sync_file_handled_count,
           last_sync_file_handled => $last_sync_file_handled,
         }
}

#----------------------------------------------------------------------

sub get_sync_file_attribs {
  my($filename) = @_;

  if (basename($filename) =~ /__account_(\d+)__policy_(\d+)/) {
    return { sync_filename => $filename, 
             sync_type => 'policy',
             policy_name_crc => $1,
             policy_id => $2,
           };
  }
  elsif (basename($filename) =~ /full_update/) {
    return { sync_filename => $filename, 
             sync_type => 'full' };
  }
}

#----------------------------------------------------------------------

sub get_sync_files {
  my($self) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  # Compare files with database, add new files, delete old files
  my $db_sync_files = $dbh->selectall_hashref(<<EndQuery, 'sync_filename');
SELECT sync_filename, first_seen_datetime, sync_type, policy_name_crc, policy_id, status, completed_datetime
  FROM DCC.CLUSTER_SYNC_FILES
EndQuery

  my $master_blade = F5::Cluster::Utils::get_master_blade();
  my $current_blade = F5::Cluster::Utils::get_slot_number();
  my %fs_sync_files = map { ($_ => 1) } glob("$SEND_DIR/*");
  for my $fs_sync_file (keys %fs_sync_files) {

    my @filename_parts = split(/__/, basename($fs_sync_file));
    # Ignore sync files created on this blade
    if (grep { $_ eq "slot_$current_blade" } @filename_parts) {
        $logger->debug("Skipping sync file $fs_sync_file (originated on this blade)") if $logger;
        next;
    }

    unless (exists $db_sync_files->{$fs_sync_file}) {
      my $new_sync_file = get_sync_file_attribs($fs_sync_file);
      $new_sync_file->{first_seen_datetime} = mysql_formatted_datetime();
      $new_sync_file->{status} = 'new';

      eval {
          $self->insert_row(table => 'DCC.CLUSTER_SYNC_FILES',
                            rec => $new_sync_file);
      };
      if ($@) {
          $self->write_status_file(sync_file => $new_sync_file,
                                   status => 'error');
          next;
      }

      $self->write_status_file(sync_file => $new_sync_file,
                               status => 'queued');

      $db_sync_files->{$new_sync_file->{sync_filename}} = $new_sync_file;
    }
  }

  for my $db_sync_file (values %$db_sync_files) {
    unless(exists $fs_sync_files{$db_sync_file->{sync_filename}}) {
      $dbh->do('DELETE FROM DCC.CLUSTER_SYNC_FILES WHERE sync_filename = ?', 
               undef, $db_sync_file->{sync_filename});
      delete $db_sync_files->{$db_sync_file->{sync_filename}};
    }
  }

  my @uncompleted_sync_files = grep { $_->{status} eq 'new' } 
    sort { $a->{sync_filename} cmp $b->{sync_filename} }
    values %$db_sync_files;

  return @uncompleted_sync_files;
}

#----------------------------------------------------------------------

sub update_sync_status {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $sync_file = $params{sync_file};
  my $status = $params{status};
  my $completed_datetime = mysql_formatted_datetime();

  $dbh->do(<<EndQuery, undef, $status, $completed_datetime, $sync_file->{sync_filename});
UPDATE DCC.CLUSTER_SYNC_FILES
   SET status = ?,
       completed_datetime = ?
 WHERE sync_filename = ?
EndQuery

  $self->write_status_file(sync_file => $sync_file,
                           status => $status);

}

#----------------------------------------------------------------------

sub write_status_file {
  my($self, %params) = @_;

  my $sync_file = $params{sync_file};
  my $status = $params{status};

  my $slot_number = F5::Cluster::Utils::get_slot_number();
  my $microseconds = (gettimeofday)[1];
  my $status_filename = "status__slot_" . $slot_number . "__" . filename_formatted_datetime() . "__" . sprintf("%06d", $microseconds);

  my %status_rec;
  $status_rec{slot_number} = $slot_number;
  $status_rec{sync_filename} = $sync_file->{sync_filename};
  $status_rec{status} = $status;
  $status_rec{status_datetime} = mysql_formatted_datetime();

  open(my $status_fh, '>', "$TEMP_DIR/$status_filename");
  for my $key (keys %status_rec) {
    print $status_fh "$key=$status_rec{$key}\n";
  }
  close $status_fh;

  move("$TEMP_DIR/$status_filename", "$STATUS_DIR/$status_filename");
}

#----------------------------------------------------------------------

sub get_sync_status_file {
  my($self) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  # Compare files with database, add new files
  my %db_status_files = map { ($_ => 1) } @{$dbh->selectcol_arrayref(<<EndQuery)};
SELECT status_filename FROM DCC.CLUSTER_SYNC_FILE_STATUS
EndQuery

  for my $status_file (glob("$STATUS_DIR/*")) {
    unless (exists $db_status_files{$status_file}) {
      my %status_file_rec;
      $status_file_rec{status_filename} = $status_file;

      open(my $status_fh, '<', $status_file);
      while(my $line = <$status_fh>) {
        chomp($line);
        my($key, $val) = split(/=/, $line, 2);
        $status_file_rec{$key} = $val;
      }
      close $status_fh;

      my($sync_file_id) = $dbh->selectrow_array(<<EndQuery, undef, $status_file_rec{sync_filename});
SELECT sync_file_id 
  FROM DCC.CLUSTER_SYNC_FILES
 WHERE sync_filename = ?
EndQuery

      if (! $sync_file_id) {
          $logger->trace("Status file received for an unknown sync file ($status_file_rec{sync_filename})") if $logger;
          # It was suspected that csyncd had introduced latency in transferring sync files. In order to mitigate this,
          # cleanup_intervening_sync_files was used to remove unnecessary sync files, possibly before being sent to the
          # secondary. However, this was found not to be the case.
          # Furthermore, this line was logged excessively under expected behavior because all sync status files are re-handled
          # upon every newly received status file. However, older sync files have already been correctly cleaned.
          # Since this cannot be detected without removing the foreign key in CLUSTER_SYNC_FILE_STATUS and adding cleaning
          # to replace the CASCADE DELETE, this change has been abandoned to allow for backporting without schema impact.
          next;
      }
      $status_file_rec{sync_file_id} = $sync_file_id;
      delete $status_file_rec{sync_filename};

      $self->insert_row(table => 'DCC.CLUSTER_SYNC_FILE_STATUS',
                        rec => \%status_file_rec);
    }
  }

  return 1;
}

#----------------------------------------------------------------------

sub mysql_formatted_datetime {
  my($time) = @_;
  $time ||= time;
  return strftime("%Y-%m-%d %H:%M:%S", localtime($time));
}

#----------------------------------------------------------------------

sub filename_formatted_datetime {
  my($time) = @_;
  $time ||= time;
  return strftime("%Y%m%d_%H%M%S", localtime($time));
}

#----------------------------------------------------------------------

sub get_sync_status {
  my($self) = @_;

  if (!F5::Cluster::Utils::is_master_blade()) {
    return [ { slot_number => F5::Cluster::Utils::get_slot_number(),
               message_type => 'bold',
               status => 'Secondary',
               info => '',
             }];
  }

  my @status_order = ('error', 'not seen', 'queued', 'loading', 'completed');
  my $status_cfg = {
                    'error' => {
                                message_type => 'error',
                                status_text => 'Error',
                                detail => 'Error',
                               },
                    'not seen' => {
                                   message_type => 'wait',
                                   status_text => 'Waiting for reply',
                                   detail => 'Sync file %s not yet detected',
                                   superceded_by => { 'queued' => 1,
                                                      'loading' => 1 },
                                  },
                    'queued' => {
                                 message_type => 'wait',
                                 status_text => 'Queued',
                                 detail => 'Sync file %s queued',
                                 superceded_by => { 'loading' => 1 },
                                },
                    'loading' => {
                                  message_type => 'wait',
                                  status_text => 'Loading',
                                  detail => 'Sync file %s loading',
                                 },
                    'completed' => { 
                                    message_type => 'ok',
                                    status_text => 'Up to date',
                                    detail => '',
                                   },
                   };

  my $slots = $self->get_slot_statuses;

  my @slot_statuses = ( { slot_number => F5::Cluster::Utils::get_slot_number(),
                          message_type => 'bold',
                          status => 'Primary',
                          info => '',
                        });

  for my $slot_number (sort keys %$slots) {
    next if ($slot_number == F5::Cluster::Utils::get_slot_number());
    my $slot = $slots->{$slot_number};
    my $last_successful_update = '';

    my @slot_status_recs = @{$slot}{reverse sort keys %$slot};

    for my $slot_status_rec (@slot_status_recs) {
      if ($slot_status_rec->{status} eq 'completed') {
        $last_successful_update = "Last update = $slot_status_rec->{create_datetime}";
        last;
      }
    }


    for my $i (0..$#slot_status_recs) {
      my $status = $slot_status_recs[$i]->{status};

      if ($i < $#slot_status_recs &&
          exists $status_cfg->{$status}->{superceded_by} &&
          $status_cfg->{$status}->{superceded_by}->{$slot_status_recs[$i + 1]->{status}}) {
        # The current status is superceded by a prior one
        next;
      }

      my $occurrence_datetime = $slot_status_recs[$i]->{create_datetime};
      my $info = $last_successful_update;
      if ($status_cfg->{$status}->{detail}) {
        $info = sprintf($status_cfg->{$status}->{detail}, $occurrence_datetime) . "\n$info";
      }

      push(@slot_statuses,
           { slot_number => $slot_number,
             message_type => $status_cfg->{$status}->{message_type},
             status => $status_cfg->{$status}->{status_text},
             info => $info,
           });
      last;
    }
  }

  return \@slot_statuses;
}

#----------------------------------------------------------------------

sub get_slot_statuses {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my %slots;
    my($status_sth) = $dbh->prepare(<<EndQuery);
  SELECT sync_filename, create_datetime, sync_type, policy_name_crc, policy_id,
         slot_number, status.status, status_datetime, message
    FROM DCC.CLUSTER_SYNC_FILES files
         LEFT JOIN DCC.CLUSTER_SYNC_FILE_STATUS status
                ON files.sync_file_id = status.sync_file_id
ORDER BY create_datetime, status_datetime
EndQuery

    $status_sth->execute;
    while (my $status_rec = $status_sth->fetchrow_hashref) {
        if ($status_rec->{slot_number}) {
            $slots{$status_rec->{slot_number}}->{$status_rec->{sync_filename}} = { status => $status_rec->{status},
                                                                                   status_datetime => $status_rec->{status_datetime},
                                                                                   message => $status_rec->{message},
                                                                                   create_datetime => $status_rec->{create_datetime},
                                                                                 };
        }
        for my $slot_number (keys %slots) {
            if (!exists $slots{$slot_number}->{$status_rec->{sync_filename}}) {
                $slots{$slot_number}->{$status_rec->{sync_filename}} = { status => 'not seen',
                                                                         create_datetime => $status_rec->{create_datetime},
                                                                         status_datetime => $status_rec->{create_datetime},
                                                                       };
            }
        }
    }

    return \%slots;
}

#----------------------------------------------------------------------

sub insert_row {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $table = $params{table};
  my $rec = $params{rec};

  local $" = ',';
  my @fields = keys %$rec;
  my @placeholders = ('?') x scalar @fields;
  $dbh->do(<<EndQuery, undef, @{$rec}{@fields});
INSERT INTO $table (@fields) VALUES (@placeholders)
EndQuery
}

#----------------------------------------------------------------------

sub is_policy_update_more_efficient {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};
  my $policy_id = $params{policy_id};

  my $total_size = 0;
  my $policy_size;
  my $revisions_sth = $dbh->prepare(<<EndQuery);
SELECT policy_id, revision
  FROM PLC.PL_POLICY_HISTORY
 WHERE to_date = 0
EndQuery

  $revisions_sth->execute;
  while(my $revision_rec = $revisions_sth->fetchrow_hashref) {
    my $file_size = -s F5::PolicyHistory->construct_filename(%$revision_rec);
    $total_size += $file_size;
    if ($revision_rec->{policy_id} == $policy_id) {
      $policy_size = $file_size;
      $logger->debug("Policy size is $policy_size") if $logger;
    }
  }

  $logger->debug("Total policies size is $total_size") if $logger;
  if ( $total_size ) {
    if (($policy_size / $total_size) < $self->{policy_size_percent_threshold}) {
      $logger->debug("Policy size is less than threshold percent ($self->{policy_size_percent_threshold}) -- policy alone will be synced") if $logger;
      return 1;
    }
    else {
      $logger->debug("Policy size is greater than threshold percent ($self->{policy_size_percent_threshold}) -- UCS will be synced") if $logger;
      return 0;
    }
  }
  return 0;
}

#----------------------------------------------------------------------

sub cleanup_status_db {
    my($self) = @_;

    # Remove status updates from the future (if system clock moves backwards)
    $self->{dbh}->do(<<EndQuery, undef, mysql_formatted_datetime());
DELETE FROM DCC.CLUSTER_SYNC_FILE_STATUS
      WHERE status_datetime > ?
EndQuery
}

#----------------------------------------------------------------------


sub cleanup_status_files {
    my($self) = @_;
    F5::Utils::File::cleanup_files(glob_pattern => "$STATUS_DIR/status__slot_" . F5::Cluster::Utils::get_slot_number() . "__*",
                                   age => 1 * 60 * 60, # 1 hours
                                   consider_future_files_as_older => 1,
                                   max_retained_files => 100,
                                   min_retained_files => 2,
                                   max_size => 10_000_000, # 10 MB
                                   logger => $self->{logger},
                                  );
}

#----------------------------------------------------------------------
sub cleanup_temp_files {
    my($self) = @_;
    F5::Utils::File::cleanup_files(dir => $TEMP_DIR,
                                   age => 20 * 60,        # 20 minutes
                                   consider_future_files_as_older => 1,
                                   max_size => 500_000_000, # 500 MB
                                   logger => $self->{logger},
                                  );
}

#----------------------------------------------------------------------

sub cleanup_sync_files {
  my($self) = @_;

  return if F5::BigipUtils::is_docker();
  my $dbh = $self->{dbh};
  F5::Utils::File::cleanup_files(glob_pattern => "$SEND_DIR/*",
                                 age => 2 * 60 * 60, # 2 hours
                                 consider_future_files_as_older => 1,
                                 min_retained_files => 2,
                                 max_size => 500_000_000, # 10 MB
                                 logger => $self->{logger},
                                );
  F5::Utils::File::cleanup_files(glob_pattern => "$STATUS_DIR/status__slot_" . F5::Cluster::Utils::get_slot_number() . "__*",
                                 consider_future_files_as_older => 1,
                                 age => 1 * 60 * 60,     # 2 hours
                                 max_size => 10_000_000, # 10 MB
                                 logger => $self->{logger},
                                );
  $self->cleanup_completed_sync_files_by_status;

  my %db_sync_files = @{$dbh->selectcol_arrayref(<<EndQuery, { Columns=>[1,2] })};
SELECT sync_filename, sync_file_id
  FROM DCC.CLUSTER_SYNC_FILES
EndQuery

  my %fs_sync_files = map { ($_ => 1) } glob("$SEND_DIR/*");
  for my $db_sync_filename (keys %db_sync_files) {
      unless (exists $fs_sync_files{$db_sync_filename}) {
          $self->{logger}->trace("Sync file $db_sync_filename is missing from filesystem -- deleting from CLUSTER_SYNC_FILES") if $self->{logger};
          $dbh->do('DELETE FROM DCC.CLUSTER_SYNC_FILES WHERE sync_file_id = ?',
                   undef, $db_sync_files{$db_sync_filename});
      }
  }
}

#----------------------------------------------------------------------

sub cleanup_completed_sync_files_by_status {
    my($self) = @_;

    my @final_statuses = qw(completed error skipped);

    # Cleanup sync files that have already been completed on all secondary slots, except the most recent
    my $slot_statuses = $self->get_slot_statuses;
    delete $slot_statuses->{F5::Cluster::Utils::get_slot_number()};

    my @slot_numbers = keys %$slot_statuses;
    my %slot_completed_sync_files;
    for my $slot_status (values %$slot_statuses) {
        for my $sync_file (keys %$slot_status) {
            if (grep { $slot_status->{$sync_file}->{status} eq $_ } @final_statuses) {
                $slot_completed_sync_files{$sync_file}++;
            }
        }
    }
    my @completed_sync_files = grep { $slot_completed_sync_files{$_} == scalar @slot_numbers } keys %slot_completed_sync_files;
    @completed_sync_files = reverse sort @completed_sync_files;
    # Preserve most recent completed file
    my $preserved_sync_file = shift(@completed_sync_files);
    if ($preserved_sync_file) {
        $self->{logger}->trace("Preserving most recent sync file $preserved_sync_file") if $self->{logger};

        # Before cleaning completed sync files, clean all earlier sync files. There may be leftovers from
        # earlier primary blades that are no longer managed by the blade that created them.
        my %completed_sync_filename_lookup = map { ($_ => 1) } @completed_sync_files;
        my @sync_filenames = sort glob("$SEND_DIR/*");
        for my $sync_filename (@sync_filenames) {
            if ($sync_filename lt $preserved_sync_file && ! $completed_sync_filename_lookup{$sync_filename}) {
                $self->{logger}->info("Deleting earlier sync file $sync_filename") if $self->{logger};
                unlink $sync_filename;
            }
        }
    }

    for my $completed_sync_file (@completed_sync_files) {
        $self->{logger}->trace("Deleting completed sync file $completed_sync_file") if $self->{logger};
        unlink $completed_sync_file;
    }
}

#----------------------------------------------------------------------

sub cleanup_intervening_sync_files {
    my($self) = @_;

    my %slot_seen_sync_files;
    my $slot_statuses = $self->get_slot_statuses;
    delete $slot_statuses->{F5::Cluster::Utils::get_slot_number()};

    for my $slot_status (values %$slot_statuses) {
        for my $sync_file (keys %$slot_status) {
            $slot_seen_sync_files{$sync_file} = 1;
        }
    }

    # Only clean full updates (policy sync files are usually much smaller)
    my @sync_filenames = sort glob("$SEND_DIR/*__full_update");
    # Always preserve first and last full sync files
    shift(@sync_filenames);
    pop(@sync_filenames);
    for my $sync_filename (@sync_filenames) {
        if (! $slot_seen_sync_files{$sync_filename}) {
            $self->{logger}->trace("Deleting intervening sync file $sync_filename") if $self->{logger};
            unlink $sync_filename;
        }
    }
}

#----------------------------------------------------------------------

my $SYNC_DIR_FILE = F5::Deploy::map_path("/ts/var/cluster/sync_cfg.yaml");
my $SYNC_DIR = F5::Deploy::map_path("/ts/var/cluster");
my @DIR_NAMES = (qw/send temp status/);

sub set_sync_file_dir {
    my($self, %params) = @_;
    my $sync_dir = $params{sync_dir} || '';
    my $logger = $self->{logger};

    if ($sync_dir) {
        $sync_dir =~ s/\/$//;
        if (-d $sync_dir) {
            YAML::XS::DumpFile($SYNC_DIR_FILE, {sync_dir => $sync_dir});
        }
        else {
            die("specified path does not exists\n");
        }
    }

    if (-f $SYNC_DIR_FILE) {
        my $target_sync_dir = YAML::XS::LoadFile($SYNC_DIR_FILE)->{sync_dir};
        if ($target_sync_dir) {
            chomp($target_sync_dir);
            if (-l "$SYNC_DIR/send") {
                my $current_sync_dir = readlink "$SYNC_DIR/send";
                $current_sync_dir = dirname($current_sync_dir);
                if ( ($target_sync_dir eq $SYNC_DIR) || ($target_sync_dir ne $current_sync_dir) ) {
                    $logger->trace("old sync path: $current_sync_dir");
                    $logger->trace("new sync path: $target_sync_dir");
                    $self->set_sync_directories(target_sync_dir => $target_sync_dir,
                                                current_sync_dir => $current_sync_dir,
                                                do_unlink => 1,
                                                create_symlink => $target_sync_dir ne $current_sync_dir,);
                }
            }
            else {
                if ($target_sync_dir ne $SYNC_DIR) {
                    $logger->trace("old sync path: $SYNC_DIR");
                    $logger->trace("new sync path: $target_sync_dir");
                    $self->set_sync_directories(target_sync_dir => $target_sync_dir,
                                                current_sync_dir => $SYNC_DIR,
                                                do_unlink => 0,
                                                create_symlink => 1);
                }
            }
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub set_sync_directories {
    my($self, %params) = @_;

    my $target_sync_dir = $params{target_sync_dir};
    my $current_sync_dir = $params{current_sync_dir};
    my $do_unlink = $params{do_unlink};
    my $create_symlink = $params{create_symlink};

    for my $dir (@DIR_NAMES) {
        my $target = "$target_sync_dir/$dir";
        my $link = "$SYNC_DIR/$dir";
        if ($do_unlink) {
            unlink($link);
        }
        mkdir $target;
        chmod(0777, $target);
        for my $file (glob("$current_sync_dir/$dir/*")) {
            my $filename = basename($file);
            move("$current_sync_dir/$dir/$filename", "$target/$filename");
        }
        rmdir "$current_sync_dir/$dir";
        if ($create_symlink) {
            symlink($target, $link);
        }
    }
}

#----------------------------------------------------------------------

1;
