package F5::Cluster::Utils;

use strict;
use warnings;

use F5::Deploy;
use F5::SystemUtils;
use F5::ASM::COAPI;
use F5::BigipUtils;

use Memoize;
memoize('is_clustering_enabled');

#----------------------------------------------------------------------

sub is_clustering_enabled {
  my($self) = @_;
  my $cluster_env = get_mcp_val("Clustered.Environment");
  return 0 if (! defined $cluster_env);
  return ($cluster_env eq 'true') ? 1 : 0;
}

#----------------------------------------------------------------------

sub is_master_blade {
  my $cluster_primary = get_mcp_val("Cluster.Primary");
  return 0 if (! defined $cluster_primary);
  return ($cluster_primary eq 'true') ? 1 : 0;
}

#----------------------------------------------------------------------

sub is_failover_standby {
    my $failover_state = get_mcp_val("Failover.State");
    return 0 if (! defined $failover_state);
    return ($failover_state eq "standby") ? 1 : 0;
}

#----------------------------------------------------------------------

sub is_failover_active {
    if (F5::BigipUtils::is_feature_enabled('standalone')) {
        return 1;
    }
    my $failover_state = get_mcp_val("Failover.State");
    return 0 if (! defined $failover_state);
    return ($failover_state eq "active") ? 1 : 0;
}

#----------------------------------------------------------------------

sub get_master_blade {
  return get_mcp_val("Cluster.Primary.Slot");
}

#----------------------------------------------------------------------

sub get_slot_number {
  my($self) = @_;

  return get_mcp_val("Platform.Chassis.Slot");
}

#----------------------------------------------------------------------

sub get_num_slots {
    my $res = F5::Cluster::Utils::fetch_mcp_vector('cluster_mbr');
    my $num_slots = scalar @$res;
    return $num_slots ? $num_slots : 0;
}

#----------------------------------------------------------------------

sub set_mcp_val {
  my($key, $val) = @_;

  my ($obj) = @{F5::ASM::COAPI::Query->new("db_variable", {name => lc $key})};
  $obj->set_value(value => $val);
  $obj->save();
  return $val;
}

#----------------------------------------------------------------------

sub get_mcp_val {
  my($key) = @_;

  my ($obj) = @{F5::ASM::COAPI::Query->new("db_variable", {name => lc $key})};
  if ($obj) {
      if (defined $obj->{value} && $obj->{value} ne '') {
          # Empty string is the only "non-value" for db_variable
          return $obj->{value};
      }
      if (defined $obj->{default}) {
          return $obj->{default};
      }
  }
  return '';
}

#----------------------------------------------------------------------

sub fetch_mcp_vector {
    my($key) = @_;

    my $res;
    eval {
        $res = new F5::ASM::COAPI::Query($key);
    };
    return $res unless $@;
}

#----------------------------------------------------------------------

sub set_slot_ip
{
   my($self) = @_;

   require F5::DbUtils;

   my $dbh = F5::DbUtils::get_dbh();

   my $slot_ip = get_slot_ip();

   $dbh->do('UPDATE DCC.SITES SET slot_ip = ?', undef, $slot_ip);
}

#----------------------------------------------------------------------

sub get_slot_ip
{
    my($self) = @_;    

    return "127.3.0.".get_slot_number();
}

#----------------------------------------------------------------------
# Usage: 
#     F5::Cluster::Utils::get_bigstart_values(qw(arg1 arg2 arg3));
#     the arguments to this subroutine are the names of functions from '/usr/lib/bigstart/functions' that you need to be executed
#     note that these do not accept any values, so we only need the function name to execute it

# Returns: 
#    a HASH ref with name of function as the key and the returned value as the value 
#    an empty returned value, by the undelying function from 'bigstart/functions', is set to 'undef'

# Example:
#     perl -MF5::Cluster::Utils -MData::Dumper -e 'print Dumper F5::Cluster::Utils::get_bigstart_values(qw(get_number_physical_cpu is_ht_split blabla))'
#     $VAR1 = {
#           'is_ht_split' => '0',
#           'get_number_physical_cpu' => '2',
#           'blabla' => undef
#         };

sub get_bigstart_values {
    my (@params) = @_;

    my $res;
    if (-e '/usr/lib/bigstart/functions') {
        my $args = join(' ',@params);
        open (CMD, "-|", F5::Deploy::map_path("/usr/share/ts/bin/run_bigstart_function $args")) or die $! ;
        while ( my $line = <CMD> ) {
            chomp($line);
            my ($key,$val) = split('=',$line);
            $val = undef unless $val =~ /\S/;
            $res->{$key} = $val;
        }
        close CMD;
    }

    return $res;
}

#----------------------------------------------------------------------

sub get_qprop_value {
    my (%params) = @_;

    my $variable_name = $params{variable_name};
    return unless $variable_name;

    my $default_on_error = $params{default_on_error} || '';
    $default_on_error = ($default_on_error) ? "default_on_error=$default_on_error" : "default_on_error=unlikely_qprop_error" ;

    my $result = F5::SystemUtils::run_system_tool(command_line_params => [ '/bin/qprop', $variable_name, $default_on_error ])->{standard_output};
    chomp($result);

    if ($result && $result ne 'unlikely_qprop_error') {
        return $result;
    }

    return;
}
#----------------------------------------------------------------------

my %config_load_fail_status = ( 'none' => 1,
                                'platform' => 1,
                                'base' => 1,
                              );

sub mcp_config_loaded {
    my (%params) = @_;
    my $logger = $params{logger};

    my $mcp_state = F5::ASM::Tmsh::Query->new(cmd => [ qw(tmsh -a show sys mcp-state field-fmt) ],
                                              parse_response => 1);

    my $mcp_config_loaded = 0;
    if ($mcp_state && $mcp_state->{completed_successfully}) {
        if ($mcp_state->{response}->{'mcpd-state'} && $mcp_state->{response}->{'mcpd-state'}->{'phase'}) {
            my $phase = $mcp_state->{response}->{'mcpd-state'}->{'phase'};
            if (! $config_load_fail_status{$phase}) {
                $mcp_config_loaded = 1;
            }
        }
    }
    else {
        $logger->info("Failed to retrieve MCP state") if $logger;
        $logger->info($mcp_state->{standard_error}) if $logger;
    }

    return $mcp_config_loaded;
}

#----------------------------------------------------------------------

1;
