package F5::CommonUpgrade;

use strict;
use warnings;

use F5::Deploy;
use F5::CommonUpgrade::Base;
use F5::DbUtils;
use F5::ModuleLoader;
use F5::BigipVersionUtils;

my $ASM_DEFAULT_ROWS_CONFIG_FILENAME = F5::Deploy::map_path('/usr/share/ts/config/default_rows_config.yaml');
our @DYNAMIC_MODULES = F5::ModuleLoader::load_from_base_class('F5::CommonUpgrade::Base');

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = {};
    bless $self, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{dbh} = $params{dbh};
    $self->{logger} = $params{logger};
    $self->{upgrade_source} = $params{upgrade_source} || F5::ASMConfig::PolicyFileFormat::POLICY_FILE_FORMAT_BINARY();
    $self->{notification_handler} = $params{notification_handler};

    $self->set_old_version( $params{old_version});
    $self->set_new_version( $params{new_version} || F5::BigipVersionUtils::bigip_version());
    my $config_file = $ENV{ASM_DEFAULT_ROWS_CONFIG_FILENAME} || $ASM_DEFAULT_ROWS_CONFIG_FILENAME;
    $self->{default_rows_config} = YAML::XS::LoadFile($config_file);
}

#----------------------------------------------------------------------

sub set_old_version {
    my($self, $old_version) = @_;
    if ($old_version) {
        $old_version = F5::BigipVersionUtils::zero_pad_segments($old_version, 3);
    }
    $self->{old_version} = $old_version;

    $self->{active_modules} = [];
    for my $module (@DYNAMIC_MODULES) {
        if ($module->is_valid_version( $self->{old_version}, $self->{new_version})) {
            push(@{$self->{active_modules}}, $module);
        }
    }

    delete $self->{renamed_tables};
    $self->{row_fixup_callback_cache} = {};
    $self->{after_table_complete_callback_cache} = {};
}

#----------------------------------------------------------------------

sub set_new_version {
    my($self, $new_version) = @_;
    if ($new_version) {
        $new_version = F5::BigipVersionUtils::zero_pad_segments($new_version, 3);
    }
    $self->{new_version} = $new_version;
}

#----------------------------------------------------------------------

sub get_callback {
    my($self, %params) = @_;
    my $table_name = $params{table_name};
    my $callback   = $params{callback};
 
    my $cache = "${callback}_cache";
    if (exists $self->{$cache}->{$table_name}) {
        return $self->{$cache}->{$table_name};
    }

    my $callback_sub;
    my @modules;
    for my $module (@{$self->{active_modules}}) {
        if (grep { $_ eq $table_name } @{$module->tables_handled}) {

            if ($module->can($callback)) {
                push(@modules, $module);
            }
        }
    }

    if (@modules) {
        # Sort by max_version
        @modules = sort sort_by_callback_version @modules;
    }

    if (@modules) {
       $self->{logger}->debug("$callback callbacks on table $table_name will be " . join(',', @modules)) if $self->{logger};
        $callback_sub = sub {
            my(%inner_params) = @_;
            no strict 'refs';

            for my $module (@modules) {
                $module->$callback(%inner_params,
                                   common_upgrade => $self,
                                   orig_table_name => $table_name,
                                  );
            }
        };
    }

    $self->{$cache}->{$table_name} = $callback_sub;
    return $callback_sub;
}

#----------------------------------------------------------------------

sub get_row_fixup_callback {
    my($self, %params) = @_;
    return $self->get_callback(%params,
                               callback => "row_fixup");
}

#----------------------------------------------------------------------

sub get_after_table_complete_callback {
    my($self, %params) = @_;
    return $self->get_callback(%params,
                               callback => "after_table_complete");
}

#----------------------------------------------------------------------

sub sort_by_callback_version {
    my $a_max_version = $a->max_version || '99999'; # 99999 is larger than the largest possible real version
    my $b_max_version = $b->max_version || '99999';
    if ($a_max_version eq $b_max_version) {
        my $a_priority = $a->priority || 99999;
        my $b_priority = $b->priority || 99999;
        return $a_priority <=> $b_priority;
    }
    return $a_max_version cmp $b_max_version;
}

#----------------------------------------------------------------------

sub table_row_fixup {
    my($self, %params) = @_;

    my $table_row_fixup_func = $self->get_row_fixup_callback(%params);
    return unless $table_row_fixup_func;

    my $table_data = $params{table_data};
    for my $key (keys %$table_data) {
        my $row = $table_data->{$key};
        my $row_state_rv;

        $table_row_fixup_func->(%params,
                                common_upgrade => $self,
                                row_state_rv => \$row_state_rv,
                                row => $row);

        if ($row_state_rv && $row_state_rv eq 'skip') {
            delete $table_data->{$key};
        }
    }
}

#----------------------------------------------------------------------

sub after_table_complete {
    my($self, %params) = @_;

    my $after_table_complete_func = $self->get_after_table_complete_callback(%params);
    return unless $after_table_complete_func;

    $after_table_complete_func->(%params,
                                 common_upgrade => $self,
                                );
}

#----------------------------------------------------------------------

sub after_all_tables_complete {
    my($self, %params) = @_;

    for my $module (@{$self->{active_modules}}) {
        if ($module->can('after_all_tables_complete')) {
            $module->after_all_tables_complete(%params,
                                               common_upgrade => $self,
                                              );
        }
    }

}

#----------------------------------------------------------------------

sub add_default_rows {
    my($self, %params) = @_;

    my $policy_id = $params{policy_id};
    my $dbname = $params{dbname} || 'PLC';
    my $seen_tables = $params{seen_tables};
    my $only_tables = $params{only_tables};
    my $check_empty = $params{check_empty};
    my $is_upgrade = 0;
    if ($self->{old_version} && $self->{old_version} ne $self->{new_version}) {
        $is_upgrade = 1;
    }

    if (ref($seen_tables) eq 'ARRAY') {
        $seen_tables = { map { ($_ => 1) } @$seen_tables };
    }
    if (ref($only_tables) eq 'ARRAY') {
        $only_tables = { map { ($_ => 1) } @$only_tables };
    }

    my $dbh = $self->{dbh};

    my $single_row_rec;

    my @policy_ids;
    if ($policy_id eq '*') {
        @policy_ids = @{$dbh->selectcol_arrayref("SELECT id FROM $dbname.PL_POLICIES")};
    }
    else {
        @policy_ids = ( $policy_id );
    }

    if ($only_tables && !$params{skip_only_tables_check}) {
        my %known_policy_ids;
        for my $table (keys %$only_tables) {
            my $check_table_name = $table;
            $check_table_name =~ s/^.*\./$dbname./;
            eval {
                my @seen_ids = @{$dbh->selectcol_arrayref("SELECT distinct(policy_id) FROM $check_table_name")};
                map { $known_policy_ids{$_} = 1 } @seen_ids;
            };
        }
        @policy_ids = grep { !$known_policy_ids{$_} } @policy_ids;
        return if !@policy_ids;
    }

    $single_row_rec = [ map { { policy_id => $_ } } @policy_ids ];
    my %saved_insert_ids;

    for my $defaults_def (@{$self->{default_rows_config}}) {
        next if (defined($only_tables) && !$only_tables->{$defaults_def->{db_table}});
        my $insert_table_name = $defaults_def->{db_table};
        $insert_table_name =~ s/^.*\./$dbname./;

        if ($defaults_def->{insert_ignore} ||
            !$seen_tables->{$defaults_def->{db_table}}) {

            if ($defaults_def->{single_row}) {
                my @single_rows = @$single_row_rec;
                if ($is_upgrade && $defaults_def->{upgrade_values}) {
                    # Merge the hashes together.
                    @single_rows = map { { %$_, %{$defaults_def->{upgrade_values}} } } @single_rows;
                }

                my %additional_params;
                my $last_insert_id;
                if ($defaults_def->{save_insert_id}) {
                    $additional_params{last_insert_id} = \$last_insert_id;
                }

		if(@single_rows) {
                    F5::DbUtils::insert_row(dbh => $dbh,
                                            ignore => 1,
                                            table_name => $insert_table_name,
                                            rec => \@single_rows,
                                            %additional_params,
                                           );
		}
                if ($defaults_def->{save_insert_id}) {
                    for my $row (@single_rows) {
                        $saved_insert_ids{$defaults_def->{db_table}}->{$row->{policy_id}} = $last_insert_id++;
                    }
                }
            }
            else {
                my %copy_defaults_opts = (%$defaults_def,
                                          table_name => $insert_table_name);

                if ($copy_defaults_opts{insert_ignore}) {
                    $copy_defaults_opts{ignore} = 1;
                }

                if ($defaults_def->{add_foreign_key}) {
                    if (scalar @policy_ids > 1) {
                        # Since the saved_insert_ids are stored based on the results of INSERT IGNORE, the auto-incremented
                        # ids are only relevant for policies that did not have a record prior to the insertion. Thus,
                        # the calculated last_insert_id will not be aligned with the actual id. We must reload the
                        # actual ids from the parent table.
                        for my $foreign_key (keys %{$defaults_def->{add_foreign_key}}) {
                            my $foreign_table = $defaults_def->{add_foreign_key}->{$foreign_key};
                            my $reloaded_insert_ids = $dbh->selectcol_arrayref("SELECT policy_id, $foreign_key FROM $foreign_table", { Columns => [1,2] });
                            $saved_insert_ids{$foreign_table} = { @$reloaded_insert_ids };
                        }
                    }

                    for my $policy_id (@policy_ids) {
                        my %insert_values;
                        my $missing = 0;
                        for my $foreign_key (keys %{$defaults_def->{add_foreign_key}}) {
                            if ($saved_insert_ids{$defaults_def->{add_foreign_key}->{$foreign_key}}->{$policy_id}){
                                $insert_values{$foreign_key} = $saved_insert_ids{$defaults_def->{add_foreign_key}->{$foreign_key}}->{$policy_id};
                            }
                            else{
                              $missing = 1;
                              last;
                            }
                        }
                        next if $missing; 

                        F5::DbUtils::copy_defaults_table(dbh => $dbh,
                                                         policy_id => $policy_id,
                                                         insert_values => \%insert_values,
                                                         %copy_defaults_opts,
                                                        );
                    }
                }
                else {
                    F5::DbUtils::copy_defaults_table(dbh => $dbh,
                                                     policy_id => $policy_id,
                                                     %copy_defaults_opts,
                                                    );
                }
            }
        }
    }
}

#----------------------------------------------------------------------

sub get_new_table_name {
    my($self, %params) = @_;
    my $table_name = $params{table_name};

    if (! $self->{renamed_tables}) {
        $self->get_renamed_tables;
    }

    if (exists $self->{renamed_tables}->{$table_name}) {
        my $new_table_name = $self->{renamed_tables}->{$table_name};
        $new_table_name =~ s/^\w*?\.//;
        return $new_table_name;
    }

    return;
}

#----------------------------------------------------------------------

sub get_renamed_tables {
    my($self) = @_;

    unless ($self->{renamed_tables}) {

        my %renamed_tables;
        for my $module (@{$self->{active_modules}}) {
            %renamed_tables = (%renamed_tables,
                               %{$module->tables_renamed});
        }
        $self->{renamed_tables} = \%renamed_tables;
    }

    return $self->{renamed_tables};
}

#----------------------------------------------------------------------

sub is_old_version_before {
    my($self, $target_version) = @_;

    $target_version = F5::BigipVersionUtils::zero_pad_segments($target_version);
    if ($self->{old_version} lt $target_version) {
        return 1;
    }
    else {
        return 0;
    }
}

#----------------------------------------------------------------------

1;
