package F5::CommonUpgrade::Blocking;

use strict;
use warnings;

use base 'F5::CommonUpgrade::Base';

use F5::DbUtils;
use F5::FeatureEnable;
use F5::Violation;
use F5::Violation::Utils;
use F5::ResetReusedViolations;
use F5::HttpChecksUpgrade;
use F5::DbUtils;

our $tables_handled = [ qw(PLC.PL_BLOCKING) ];
our $max_version = '11.3.0';

my %handled_violations = ( F5::Violation::VIOL_CSRF() => 'PL_POLICY_CSRF',
                           F5::Violation::VIOL_DATA_GUARD() => 'PL_RESPONSE_SCRUBBING',);

sub row_fixup {
    my($self, %params) = @_;

    my $row = $params{row};
    my $idmap = $params{idmap};
    my $common_upgrade = $params{common_upgrade};

    if (exists $row->{code} && ! exists $row->{viol_index}) {
        $row->{viol_index} = F5::Violation::Utils::get_viol_index_from_code($row->{code});
    }

    my %reused_viols = map { ($_ => 1) } F5::ResetReusedViolations::get_reused_violations_by_bigip_version($common_upgrade->{old_version});
    if ($reused_viols{$row->{viol_index}}) {
        # The correct default value will get added after import/upgrade completes
        my $row_state_rv = $params{row_state_rv};
        $$row_state_rv = 'skip';
        return;
    }

    if (! exists $row->{flg_learn}) {
        $row->{flg_learn} = $row->{flg_alarm} || $row->{flg_reject};
        $idmap->{'__private_save'}->{pl_blocking__flg_learn_is_missing} = 1;
    }

    push (@{$idmap->{'__private_save'}->{pl_blocking}}, { %$row } );
}

#------------------------------------------------------------------------------

sub after_all_tables_complete {
    my($self, %params) = @_;

    my $dbh    = $params{dbh};
    my $idmap  = $params{idmap};
    my $policy_id = $params{policy_id};
    my $dbname = $params{dbname} || 'PLC';
    my $common_upgrade = $params{common_upgrade};

    for my $policy_reject (@{$idmap->{'__private_save'}->{pl_blocking}}) {
        next unless defined $handled_violations{$policy_reject->{viol_index}};
 
        if ( $common_upgrade->is_old_version_before('11.0.0') ) {
            F5::FeatureEnable::enable_feature(table_name => "$dbname." . $handled_violations{$policy_reject->{viol_index}},
                                              policy_id => $policy_id eq '*' ? $policy_reject->{policy_id} : $policy_id,
                                              dbh => $dbh,
                                              enable => $policy_reject->{flg_reject} || $policy_reject->{flg_learn} || $policy_reject->{flg_alarm});
        }
    }

    if ($common_upgrade->is_old_version_before('9.4.4')) {
        F5::HttpChecksUpgrade::populate_http_checks_defaults(dbh => $dbh,
                                                             dbname => $dbname,
                                                             logger => $params{logger},
                                                             policy_id => '*',
                                                             saved_blocking => $idmap->{'__private_save'}->{pl_blocking},
                                                            );

        if ($idmap->{'__private_save'}->{non_rfc_bitmask}) {
	    $dbh->do("UPDATE $dbname.PL_HTTP_CHECKS SET flg_is_enabled=1 WHERE http_check_code = http_check_code & ? ", undef,
                     $idmap->{'__private_save'}->{non_rfc_bitmask});
        }


    }

    if ($common_upgrade->is_old_version_before('9.4.2')) {
        my(%policy_is_transparent) = @{$dbh->selectcol_arrayref(<<EndQuery, { Columns => [1,2] })};
SELECT id, flg_disable_rejects
  FROM $dbname.PL_POLICIES
EndQuery

        for my $blocking_row (@{$idmap->{'__private_save'}->{pl_blocking}}) {
            my $policy_id = ($policy_id eq '*' ? $blocking_row->{policy_id} : $policy_id);

            if (exists $blocking_row->{flg_reject_last} && $policy_is_transparent{$policy_id}) {

                my $update_rec = { flg_reject => $blocking_row->{flg_reject_last} };
                if ($idmap->{'__private_save'}->{pl_blocking__flg_learn_is_missing}) {
                    $update_rec->{flg_learn} = $blocking_row->{flg_reject_last} || $blocking_row->{flg_alarm};
                }

                F5::DbUtils::update_row(dbh => $dbh,
                                        table_name => "$dbname.PL_BLOCKING",
                                        rec => $update_rec,
                                        where_fields => {
                                                         policy_id => $policy_id,
                                                         viol_index => $blocking_row->{viol_index},
                                                        },
                                       );
            }
        }
    }

    if ($idmap && $idmap->{'__private_save'}) {
        for my $private_save_key (qw(pl_blocking
                                     non_rfc_bitmask
                                     pl_blocking__flg_learn_is_missing
                                    )) {

            delete $idmap->{'__private_save'}->{$private_save_key};
        }
    }
}

#----------------------------------------------------------------------

1;
