package F5::CommonUpgrade::Cookies;

use base 'F5::CommonUpgrade::Base';

our $max_version = '11.4.0';
our $tables_handled = [ qw(PLC.PL_ALLOWED_COOKIES) ];
our $priority = 10; # must run before F5::CommonUpgrade::PBConfigCombined

sub row_fixup {
    my($self, %params) = @_;

    my $row = $params{row};
    #No pattern order existed before 11.0.0, let the after_table_complete handle the correct new order
    if( !exists $row->{pattern_order}){
        $row->{pattern_order} = undef;
    }

    if( $row->{flg_tightening}){
        $row->{flg_tightening} = F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION();
    }
    else{
        $row->{flg_tightening} = F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER();
    }

    if ( defined $row->{cookie_enforcement_mode} && $row->{cookie_enforcement_mode} eq 'allow'){
        $row->{flg_check_signatures} = 1;
        $row->{flg_is_staging} = 1;
    }
    else{
        $row->{flg_check_signatures} = 0;
    }
}

#----------------------------------------------------------------------

sub after_table_complete {
  my($self, %params) = @_;
  local $" = ',';

  my $dbh    = $params{dbh};
  my $logger = $params{logger};
  my $dbname = $params{dbname} || 'PLC';

  my $policy_where_clause = '';
  my $policy_and_clause = '';
  if ($params{policy_id}) {
    $policy_and_clause = "AND policy_id in ($params{policy_id})";
    $policy_where_clause = "WHERE policy_id in ($params{policy_id})";
  }
  my $cookies = $dbh->selectall_arrayref(<<END_SQL);
SELECT id, policy_id
  FROM $dbname.PL_ALLOWED_COOKIES
  WHERE flg_is_pattern = 1 AND pattern_order IS NULL $policy_and_clause
  ORDER BY length(cookie_name) DESC
END_SQL

  if ($cookies) {
      my $policy_max_order = $dbh->selectall_hashref(<<END_SQL, 'policy_id');
SELECT policy_id, MAX(pattern_order) as max
  FROM $dbname.PL_ALLOWED_COOKIES $policy_where_clause
  GROUP BY policy_id
END_SQL

      for my $cookie (@$cookies) {
          my ($id, $policy_id) = @$cookie;
          $policy_max_order->{$policy_id}{max}++;
          $dbh->do("UPDATE $dbname.PL_ALLOWED_COOKIES SET pattern_order = $policy_max_order->{$policy_id}{max} WHERE id = $id");
      }
  }
}

#----------------------------------------------------------------------

sub after_all_tables_complete {
    my($self, %params) = @_;
    my $idmap = $params{idmap};
    my $dbh = $params{dbh};
    my $dbname = $params{dbname} || 'PLC';

    my $policy_ids = ($params{policy_id} eq "*") ?  $dbh->selectcol_arrayref("SELECT id FROM $dbname.PL_POLICIES") : [$params{policy_id}];

    for my $policy_id (@{$policy_ids}) {
        my($max_pattern_order) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT MAX(pattern_order)
  FROM $dbname.PL_ALLOWED_COOKIES
 WHERE policy_id = ?
EndQuery

        my ($tightening_mode, $pb_enabled) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT pbconfig_cookie.flg_cookies, pbconfig.flg_enabled
  FROM $dbname.PL_POLICY_BUILDER_CONFIG pbconfig JOIN $dbname.PL_POLICY_BUILDER_CONFIG_COOKIE pbconfig_cookie 
      ON pbconfig.policy_id = pbconfig_cookie.policy_id
 WHERE pbconfig.policy_id = ?
EndQuery

        my $orig_wild_cookie =  $dbh->selectrow_hashref(<<EndQuery, undef, $policy_id,'*',1) || {};
SELECT pattern_order, cookie_enforcement_mode, last_tightening_time, flg_is_staging, flg_tightening
  FROM $dbname.PL_ALLOWED_COOKIES
 WHERE policy_id = ?
   AND cookie_name = ?
   AND flg_is_pattern = ?
EndQuery

        $idmap->{'__private_save'}->{orig_wild_cookie_enforcement_mode}->{$policy_id} = $orig_wild_cookie->{cookie_enforcement_mode};

        if (!$pb_enabled && defined ($orig_wild_cookie->{flg_tightening})) {
            $tightening_mode = $orig_wild_cookie->{flg_tightening};
        }
        $tightening_mode = ($tightening_mode ? F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION() : F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER());

        F5::DbUtils::insert_row(dbh => $dbh,
                                table_name => "$dbname.PL_ALLOWED_COOKIES",
                                rec => {
                                        policy_id               => $policy_id,
                                        cookie_name             => '*',
                                        flg_is_pattern          => 1,
                                        pattern_order           => (defined $max_pattern_order ? ($max_pattern_order+1) : 1 ), 
                                        cookie_enforcement_mode => 'allow',
                                        last_tightening_time    => time(),
                                        flg_is_staging          => 1,
                                        staged_since            => time(),
                                        last_updated            => time(),
                                        %$orig_wild_cookie,
                                        flg_tightening          => $tightening_mode,
                                       },
                                do_replace => 1,
                               );

        F5::DbUtils::update_row(dbh => $dbh,
                                table_name => "$dbname.PL_POLICY_BUILDER_CONFIG_COOKIE",
                                rec => { flg_cookies => $tightening_mode },
                                where_fields => {  policy_id => $policy_id, },
                               );
    }
}

#----------------------------------------------------------------------

1;
