package F5::ConfigSync;
use strict;
use warnings;

use POSIX;
use Digest::MD5;
use File::Path;
use File::Copy;
use File::Find;
use File::Basename;
use File::Slurp;

# Used to call tar externally
use IPC::Open3;
use Symbol;
use File::Spec;
use Time::HiRes;
use YAML::XS;
use List::Util;
use Fcntl;
use Tie::File;

use IO::Compress::Gzip qw(gzip $GzipError);
use IO::Uncompress::Gunzip qw(gunzip $GunzipError);

use F5::Deploy;
use F5::BigipUtils;
use F5::Db::Partition;
use F5::ASM::COAPI;
use F5::RequestLog;
use F5::ProcessTable;
use F5::Syslog;
use F5::LockFactory;
use F5::DbUtils;
use F5::DbUpgrade;
use F5::SetActive;
use F5::Utils::Account;
use F5::Utils::Rest;
use F5::Policy;
use F5::Logger;
use F5::Cluster::Sync;
use F5::Cluster::Utils;
use F5::BigipVersionUtils;
use F5::Provisioning;
use F5::ASMReady;
use F5::Certificates;
use F5::PrepareConf::Policy;
use F5::PrepareConf::Checksums;
use F5::PrepareConf::BotDefense;
use F5::ASMConfig::GuaranteedClient;
use F5::NegativeSignatures::SigFile;
use F5::NegativeSignatures::Collection;
use F5::Cfg;
use F5::Utils::File;

$SIG{PIPE}          = 'IGNORE';

my $ASM_DATA_DIR = "/var/asmdata1";
my $TRAFFIC_DATA = "$ASM_DATA_DIR/traffic_data";
my $TRAFFIC_DATA_SUFFIX = "traffic_data";
my $TRAFFIC_DATA_TOKEN = "token";
my $UCS_INSTALL_ORIGINATOR = "install_ucs";
my $UCS_SAVE_ON_LOAD_ENV_VAR = "CS_SAVE_ON_LOAD";

my $SCHEMA_HOME_DIR = '/ts/var/schema';
my $SCHEMA_BACKUP_DIR = '/var/ts/var/tmp/schema';

# default in case 'sys db ucs.asm.traffic_data.save' is not available
# disable | enable, default:enable
my $UCS_ASM_TRAFFIC_DATA_SAVE = 'enable';

# default in case 'sys db ucs.asm.traffic_data.load' is not available
# never | matching | always, default:matching
my $UCS_ASM_TRAFFIC_DATA_LOAD = 'matching';

my %PARTITIONED_TABLES_TO_PARTITION_ROW_LIMIT = (
    'PRX.REQUEST_LOG' => $F5::RequestLog::DB_PARTITION_ROW_LIMIT,
);

my $ACCOUNT_DATA_DIR = $F5::PrepareConf::Base::ACCOUNT_DATA_DIR;
my @account_data_dirs = ($ACCOUNT_DATA_DIR,
                         "/ts/var/policy_templates",
                         "/ts/var/schema",
                         "/var/ts/dms/server_technologies_logos",
                        );

my @db_file_references = (
                              { db_file_refs => [
                                                 { table_name => 'DCC.ACCOUNT_VALID_HOST_NAMES_REGEXES',
                                                   field => 'regexes_filename', },
                                                ],
                                dir => "$ACCOUNT_DATA_DIR/valid_host_names",
                              },
                              { db_file_refs => [
                                                 { table_name => 'DCC.JSON_VALIDATION_FILES',
                                                   field => 'schema_filename', },
                                                ],
                                dir => "$ACCOUNT_DATA_DIR/json",
                              },
                              { db_file_refs => [
                                                 { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                   field => 'schema_filename', },
                                                 { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                   field => 'defense_config_filename', },
                                                 { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                   field => 'request_xpath_filename', },
                                                 { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                   field => 'response_xpath_filename', },
                                                ],
                                dir => "$ACCOUNT_DATA_DIR/xml",
                              },
                              { db_file_refs => [
                                                 { table_name => 'DCC.ACCOUNT_NEGSIG_COLLECTIONS',
                                                   field => 'request_sig_collection_filename' },
                                                 { table_name => 'DCC.ACCOUNT_NEGSIG_COLLECTIONS',
                                                   field => 'response_sig_collection_filename' },
                                                ],
                                dir => "$F5::NegativeSignatures::Collection::ACCOUNT_DATA_DIR/negsig",
                              },
                              { db_file_refs => [
                                                 { table_name => 'DCC.DATA_PROTECTION',
                                                   field => 'conf_file' },
                                                 { table_name => 'DCC.DATA_PROTECTION',
                                                   field => 'grace_conf_file' },
                                                ],
                                dir => "$ACCOUNT_DATA_DIR/data_protection",
                              },
                              { db_file_refs => [
                                                 { table_name => 'DCC.CREDENTIAL_STUFFING_UPDATE_FILE',
                                                   field => 'credential_filename' },
                                                ],
                                dir => "$ACCOUNT_DATA_DIR/credentials",
                              },
                             );

my $UCS_POST_INSTALL_PATH = '/ts/var/install';
my $POLICY_VERSIONS_DIR = '/ts/dms/policy/policy_versions';
my $ASM_UCS_INSTALL_PROC_FILENAME = '/var/ts/var/install/ucs_install.pid';
my $COAPI_TRANSACTION_TIMEOUT = F5::Cfg::cfg_get_config_item('/etc/ts/tools/asm_config_server.cfg', 'GENERAL', 'McpQueryTimeout') || 600; # seconds

sub new {
  my($class, %params) = @_;

  my $self = {};
  bless $self, $class;
  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{mysql_dump_user} = 'root'; # for access to staging and ucs dbs
    $self->{mysql_dump_password} = F5::Cfg::get_mysql_password(user => $self->{mysql_dump_user});
    $self->{ucs_post_install} = $params{ucs_post_install};
    $self->{originator} = $params{originator} || '';

    $self->{base_dir} = $params{base_dir}         ? $params{base_dir}
                      : $params{ucs_post_install} ? $UCS_POST_INSTALL_PATH
                      :                             "/var/tmp";

    unless (-e $self->{base_dir}) {
      mkdir($self->{base_dir});
      chmod(0777, $self->{base_dir});
      $self->{created_temp_base_dir} = 1;
    }

    $self->{ucs_v1_dump_file} = "ts_db.sql.cstmp";
    $self->{dump_file_prefix} = 'ts_db';
    $self->{dump_file_suffix} = 'cstmp';
    $self->{cleanup_temporary_files} = defined $params{cleanup_temporary_files} ? $params{cleanup_temporary_files} : 1;
    $self->{bigip_version} = $params{bigip_version} || F5::BigipVersionUtils::bigip_version();
    $self->{bigip_build} = (defined $params{bigip_build} && length($params{bigip_build})) ? $params{bigip_build} : '0.0';
    $self->{ucs_version} = $params{ucs_version};
    if (!$self->{ucs_version} && -e $self->ucs_version_file) {
      $self->{ucs_version} = F5::Utils::File::file_slurp($self->ucs_version_file);
      chomp($self->{ucs_version});
    }
    elsif ($self->{ucs_version}) {
        $self->{ucs_build} = (defined $params{ucs_build} && length ($params{ucs_build})) ? $params{ucs_build} : '0.0';
        $self->{ucs_version} = F5::BigipVersionUtils::zero_pad_segments("$self->{ucs_version}.$self->{ucs_build}");
    }
    $self->{force_upgrade} = $params{force_upgrade};
    $self->{policy_history} = defined $params{policy_history} ? $params{policy_history} : 1;
    $self->{dbh} = $params{dbh};

    $self->{logger} = $params{logger} || F5::Logger->get_logger(logger_name => 'ts_configsync');
    $self->{include_cluster_tables} = $params{include_cluster_tables};
    $self->{no_private_keys} = defined $params{no_private_keys} ? $params{no_private_keys} : 0;
    $self->{account_id} = $params{account_id};
    $self->{active_policy_sync_only} = $params{active_policy_sync_only};
    $self->{asm_config_handler} = $params{asm_config_handler};
    $self->{call_flags} = $params{call_flags};
    $self->{cmi_metadata} = $params{cmi_metadata};
    $self->{allow_upgrade} = defined $params{allow_upgrade} ? $params{allow_upgrade} : 1;

    $self->{databases} = [
    {
        'ucs_dbname' => 'DCC_ucs',
        'staging_dbname' => 'DCC_staging',
        'bigip_dbname' => 'DCC',
        'ucs_v2_schema_filename' => $self->{dump_file_prefix}. ".DCC.schema.". $self->{dump_file_suffix},
        'ucs_v2_data_filename' => $self->{dump_file_prefix}. ".DCC.data.". $self->{dump_file_suffix},
    },
    {
        'ucs_dbname' => 'PLC_ucs',
        'staging_dbname' => 'PLC_staging',
        'bigip_dbname' => 'PLC',
        'ucs_v2_schema_filename' => $self->{dump_file_prefix}. ".PLC.schema.". $self->{dump_file_suffix},
        'ucs_v2_data_filename' => $self->{dump_file_prefix}. ".PLC.data.". $self->{dump_file_suffix},
    }];

    if ($self->{bigip_version}) {
      $self->{bigip_version} = F5::BigipVersionUtils::zero_pad_segments("$self->{bigip_version}.$self->{bigip_build}");
    }
}

#----------------------------------------------------------------------

sub save {
    my ($self, %params) = @_;

    my $logger = $self->{logger};
    my $provisioned = $params{provisioned};
    my $check_asm_ready = defined $params{check_asm_ready} ? $params{check_asm_ready} : 1;
    my $save_tar_file = $params{save_tar_file};
    my $do_lock = defined $params{do_lock} ? $params{do_lock} : 1;

    if ($self->{no_private_keys}){
       $logger->debug("Using Support - NO_PRIVATE_KEYS IS ON");
    }else{
       $logger->debug("Not using Support - NO_PRIVATE_KEYS IS OFF");
    }

    if(((defined $provisioned) && (! $provisioned)) ||
       (! F5::Provisioning::is_asm_on())) {
        $logger->debug("ASM configuration will not be saved, ASM is not on") if $logger;
        return;
    }

    if(! F5::Provisioning::was_tsconfig_run()) {
        $logger->info("ASM configuration will not be saved, ASM was never started although it is provisioned") if $logger;
        return;
    }

    if ($check_asm_ready && ! F5::ASMReady::wait_for_asm_to_be_ready() ){
      my $msg = 'gave up waiting for ASM to start, please try again later';
      F5::Syslog::ts_syslog( 'err', $msg);
      die $msg;
    }

    if (!$self->{dbh} || $self->{dbh}->{Username} ne "root") {
      if ($self->{dbh} && !$self->{dbh}->{AutoCommit}) {
          $self->{dbh}->commit;
          $self->{dbh}->begin_work; #keep orginial dbh in the same state it came in 
      }

      my %db_opts;
      
      if (defined $params{db_opts}){
          %db_opts = %{$params{db_opts}};
      }
      
      $self->{dbh} = F5::DbUtils::get_dbh(database_username => "root",
                                          db_opts => { RaiseError => $self->{dbh}->{RaiseError},
                                                       PrintError => $self->{dbh}->{PrintError},
                                                       %db_opts });
    }

    if (! $self->{dbh}) {
        my $msg = "ASM configuration could not be saved, mysql is unavailable";
        $logger->error($msg) if $logger;
        die "$msg\n";
    }
    my $locker;
    if ($do_lock) {
        $locker = F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
                                              logger => $logger);
        $locker->lock();
    }
    $self->_save(save_tar_file => $save_tar_file);
    if ($do_lock) {
        $locker->unlock();
    }
}

#----------------------------------------------------------------------

sub load {
    my ($self, %params) = @_;
    my $update_mcp_timestamp = $params{update_mcp_timestamp} || 0;

    my $logger = $self->{logger};
    my $check_asm_ready = (defined $params{check_asm_ready}) ? $params{check_asm_ready} : 1;

    my $use_config_server_for_set_active = ($check_asm_ready) ? 1 : 0;

    if(! $self->is_actual_load_possible($check_asm_ready)) {
      $self->move_ucs_files_to_post_install_path;
      return 1;
    }

    if (! F5::Provisioning::is_asm_on()) {
        $logger->debug("ASM configuration will not be loaded, ASM is not on") if $logger;
        return 1; # return success to prevent false 'F5::Cluster::Sync::apply_sync): Set active failed' errors on Hybrid DDOS (PBD)
    }

    if (!$self->{dbh} || $self->{dbh}->{Username} ne "root") {
      if ($self->{dbh} && !$self->{dbh}->{AutoCommit}) {
          $self->{dbh}->commit;
          $self->{dbh}->begin_work; #keep orginial dbh in the same state it came in
      }

      $self->{dbh} = F5::DbUtils::get_dbh(database_username => "root",
                                          db_opts => { RaiseError => $self->{dbh}->{RaiseError},
                                                       PrintError => $self->{dbh}->{PrintError} });
    }

    $self->{dbh}->do('SET @SAVE_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS');
    $self->{dbh}->do(q{SET @SAVE_SQL_MODE=@@SQL_MODE});
    $self->{dbh}->do('SET FOREIGN_KEY_CHECKS=0');
    $self->{dbh}->do(q{SET SQL_MODE='NO_AUTO_VALUE_ON_ZERO'});
    my $locker = F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
                                             logger => $logger,);
    $locker->lock();

    if ($self->{originator} eq $UCS_INSTALL_ORIGINATOR || $self->{ucs_post_install}) {
        F5::ASMReady::write_asm_upgrade_lock();
    }

    # save the pre-load contents of /ts/var/schema/ aside
    if (-d $SCHEMA_BACKUP_DIR) {
        F5::Utils::File::delete_files_recursively($SCHEMA_BACKUP_DIR);
    }
    if (-d $SCHEMA_HOME_DIR) {
        F5::Utils::File::copy_files_recursively(from_dir => $SCHEMA_HOME_DIR,
                                                to_dir => $SCHEMA_BACKUP_DIR);
    }

    my $rv = { completed_successfully => 1 };
    my $fatal_error_caught;
    eval {
      $rv = $self->_load(use_config_server_for_set_active => $use_config_server_for_set_active);
    };
    if ($@) {
      $fatal_error_caught = $@;
      $self->{dbh}->rollback() if !$self->{dbh}->{AutoCommit};
    }

    $self->{dbh}->do('SET FOREIGN_KEY_CHECKS=@SAVE_FOREIGN_KEY_CHECKS');
    $self->{dbh}->do(q{SET SQL_MODE=@SAVE_SQL_MODE});

    if ($fatal_error_caught) {
      if (-d $SCHEMA_BACKUP_DIR) {
          # restore the pre-load files if the load failed
          F5::Utils::File::move_files_recursively(from_dir => $SCHEMA_BACKUP_DIR,
                                                  to_dir => $SCHEMA_HOME_DIR);
      }
      $self->cleanup if $self->{cleanup_temporary_files};
      $locker->unlock();
      F5::ASMReady::remove_asm_upgrade_lock();
      die $fatal_error_caught;
    }

    my $accounts_to_delete = $self->{dbh}->selectcol_arrayref(<<EndQuery);
SELECT policy_name_crc 
  FROM DCC.ACCOUNTS accounts 
 WHERE policy_name NOT IN (SELECT name FROM PLC.PL_POLICIES)
EndQuery

    for my $policy_name_crc (@$accounts_to_delete) {
        F5::Utils::Account::delete_account(policy_name_crc => $policy_name_crc,
                                           dbh => $self->{dbh});
    }

    # before the lock removal
    if (!$rv || !$rv->{completed_successfully}) {
        if (-d $SCHEMA_BACKUP_DIR) {
            # restore the pre-load files if the load failed
            F5::Utils::File::move_files_recursively(from_dir => $SCHEMA_BACKUP_DIR,
                                                    to_dir => $SCHEMA_HOME_DIR);
        }
    }
    else {
        if (-d $SCHEMA_BACKUP_DIR) {
            # purge the pre-load files if the load completed successfully
            F5::Utils::File::delete_files_recursively($SCHEMA_BACKUP_DIR);
        }
    }

    $locker->unlock();

    $self->{dbh}->do(<<EndQuery,undef,time );
UPDATE DCC.INSTALL_UCS_INFO
   SET last_load_time = ?
EndQuery

    F5::ASMReady::remove_asm_upgrade_lock();

    if (!$rv || !$rv->{completed_successfully}) {
        return;
    }

    if (! $self->{include_cluster_tables} && !$self->{active_policy_sync_only} ) {
        my $call_flags = $self->{call_flags} || F5::ASMConfig::CallFlags->new;
        if ( $self->{ucs_post_install} ){
            $call_flags->{send_events_to_dispatcher} = 0;
        }

        eval {
            my $asm_config_client = $self->{asm_config_handler} || F5::ASMConfig::GuaranteedClient::get_client(username => 'configsync');
            $call_flags->{update_mcp_timestamp} = $update_mcp_timestamp;
            if ($asm_config_client) {
                if ($self->{cmi_metadata} || (-e $ASM_UCS_INSTALL_PROC_FILENAME && !$self->{ucs_post_install})) {
                    # Session Awareness Data Points have already been reloaded during UCS _load
                    $asm_config_client->load_complete( $asm_config_client->{session},
                                                       $call_flags,
                                                       F5::ASMConfig::LoadComplete->new(do_reload_session_awareness => 0),
                                                     );
                }

                $asm_config_client->send_ucs_config_load_event( $asm_config_client->{session},
                                                                $call_flags,
                                                                0, # placeholder to avoid POLICY_INDEX nastiness
                                                                $rv->{policy_versions});
            }
        };
        if ($@) {
            $logger->info("Couldn't connect to ASM Config ($@)") if $logger;
        }
    }

    $rv = $self->update_core($use_config_server_for_set_active);

    return unless ($rv);

    if (! $self->{ucs_post_install}) {

        my $cluster_sync = F5::Cluster::Sync->new(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  check_asm_ready => $check_asm_ready );
        if (!$cluster_sync) {
            $logger->error("Can't sync - ASM is not ready") if $logger;
        }
        else {
            $cluster_sync->send_sync;
        }
    }

    return 1;
}

#---------------------------------------------------------------------
sub _save
{
  my ($self, %params) = @_;
  
  my $dbh = $self->{dbh};
  my $logger = $self->{logger};
  my $save_tar_file = $params{save_tar_file};

  my $ucs_v3_file_dir = "$self->{dump_file_prefix}.save_dir_$$.$self->{dump_file_suffix}";

  my $dump_file_dir = "$self->{base_dir}/$ucs_v3_file_dir";
  #Clean up any remnants of previous saves from this process, if running in long-lived directory
  if (! $self->{created_temp_base_dir}) {
      $self->cleanup(dump_file_dir => $dump_file_dir,
                     clean_previous_runs => 1,
                    );
  }

  open(my $pointer_file_fh, '>', $self->ucs_v3_pointer_file);
  print $pointer_file_fh "$ucs_v3_file_dir\n";
  close $pointer_file_fh;

  mkdir($dump_file_dir);
  chmod(0777, $dump_file_dir);
  $logger->debug("Saving configuration to $dump_file_dir") if $logger;

  my $failsafe_start_filename = $self->get_dump_file_name("failsafe_start");
  my $failsafe_end_filename = $self->get_dump_file_name("failsafe_end");
  my $data_file_count = 0;
  my $schema_file_count = 0;
  open(my $failsafe_start_fh, '>', $failsafe_start_filename);
  print $failsafe_start_fh "Start file\n";
  close $failsafe_start_fh;

  my $ucs_start_time = Time::HiRes::time();
  for my $database (@{$self->{databases}}) {
    my @tables = (F5::DbUtils::get_active_policy_tables($database->{bigip_dbname}));

    push (@tables, F5::DbUtils::get_config_sync_tables($database->{bigip_dbname})) if !$self->{active_policy_sync_only};
    push (@tables, F5::DbUtils::get_cluster_tables($database->{bigip_dbname}))     if $self->{include_cluster_tables};
    my %export_table_where_clauses = F5::DbUtils::get_export_where_clauses_by_table($database->{bigip_dbname});
LINE: for my $table (@tables) {
         if ($self->{no_private_keys}){
             if($table eq 'PLC.SSL_CERTIFICATES' || $table eq 'PLC.SSL_CERTIFICATE_FILES_PATH'){
               $logger->debug("Skip $table table") if $logger;
               next LINE;
             }
         }
         $logger->debug("Exporting $table") if $logger;

         if (! -d $dump_file_dir) {
           my $err_msg = "Temporary directory disappeared during save (perhaps concurrent config save/install in progress?) - ASM configuration save aborted";
           $logger->error($err_msg) if $logger;
           die "$err_msg\n";
         }

         my @where_clauses = (1);
         if ($export_table_where_clauses{$table}) {
             push (@where_clauses, $export_table_where_clauses{$table});
         }
         if ($self->{active_policy_sync_only}) {  
             if (defined($self->{policy_name_crc}) && (grep {$_ eq "policy_name_crc"} F5::DbUtils::get_db_fields($dbh, $table))) {
                 push (@where_clauses, "policy_name_crc= $self->{policy_name_crc}");
             }
         }
         else {  
             my($table_name, $create_sql) = $dbh->selectrow_array("SHOW CREATE TABLE $table");
             open(my $table_sql_fh, '>', "$dump_file_dir/$self->{dump_file_prefix}.schema.$table.$self->{dump_file_suffix}");
             print $table_sql_fh $create_sql;
             close $table_sql_fh;
             $schema_file_count++;
         }
         my $where_clause = join(' AND ', @where_clauses);

         my $data_outfile = "$dump_file_dir/$self->{dump_file_prefix}.data.$table.$self->{dump_file_suffix}";
         my $rv = $dbh->do(<<EndQuery);
SELECT *
  FROM $table WHERE $where_clause
  INTO OUTFILE '$data_outfile'
  FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY '"'
EndQuery

         $data_file_count++;
         if (! defined $rv) {
           my $err_msg = "Could not export table data $table - ASM configuration save aborted";
           $logger->error($err_msg) if $logger;
           die "$err_msg\n";
         }
      }
  }

  eval {
      $self->save_traffic_data(dump_file_dir => $dump_file_dir);
  };
  if (my $err = $@) {
      $logger->error("Could not save traffic data") if $logger;
      $logger->error($err) if $logger;
  }

  if (! $self->{include_cluster_tables}) {
      # Save policy versions -- for cluster, this is handled by syncing PL_CONFIG_LOG itself
      my @preserve_fields = qw(policy_id
                               policy_name
                               policy_version_time
                               policy_version_device_name
                               policy_version_policy_name
                               policy_version_last_change
                             );

      my $iter = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $dbh, logger => $logger);
      my @policies;
      while(my $policy_entity = $iter->get_next) {
          my $policy_rec = {};
          for my $k (@preserve_fields) {
              $policy_rec->{$k} = $policy_entity->{$k};
          }
          push(@policies, $policy_rec);
      }

      YAML::XS::DumpFile($self->get_dump_file_name("policy_version_metadata"), \@policies);
  }

  if ($self->{cmi_metadata} && $self->{asm_config_handler}) {
      $logger->debug("Saving cmi metadata") if $logger;
      my $is_failover_standby = F5::Cluster::Utils::is_failover_standby();
      my %known_times_copy = %{$self->{asm_config_handler}->{known_times}};
      my $metadata = { ucs_end => $ucs_start_time,
                       known_times => \%known_times_copy,
                       sync_from_active_device => !$is_failover_standby,
                       sync_request_from_active_device => $self->{call_flags}->{sync_request_from_active_device},
                     };

      my $metadata_filename = $self->get_dump_file_name("cmi_metadata");
      YAML::XS::DumpFile($metadata_filename, $metadata);

      #We may need to reforward calls back to the target machine so it doesn't miss updates from itself that got overwritten by this ucs
      $self->{asm_config_handler}->{sync_info}{$self->{call_flags}->{originator}} = $ucs_start_time;
  }

  my $version_build_metadata = { install_sql_checksums => $self->get_install_sql_checksums() };
  YAML::XS::DumpFile($self->get_dump_file_name("version_build_metadata"), $version_build_metadata);

  # Failsafe end file will contain the total number of files exported
  $dbh->do(<<EndQuery, undef, $schema_file_count, $data_file_count);
SELECT 1,?,?
  INTO OUTFILE '$failsafe_end_filename'
  FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY '"'
EndQuery

  if ($save_tar_file) {
      $self->do_conf_tar(filename => $save_tar_file,
                         flg_cleanup => 0);
  }

  F5::Syslog::ts_syslog( 'info', 'ASM config saved');
  $logger->debug('Configuration save complete') if $logger;
}

#---------------------------------------------------------------------

sub get_install_sql_checksums {
    my($self, %params) = @_;

    my $install_sql_checksums = {};
    for my $database (@{$self->{databases}}) {
        open(my $sql_fh, '<', F5::DbUtils::get_sql_filename($database->{bigip_dbname}));
        $install_sql_checksums->{$database->{bigip_dbname}} = Digest::MD5->new->addfile($sql_fh)->hexdigest;
        close $sql_fh;
    }
    return $install_sql_checksums;
}

#---------------------------------------------------------------------

sub is_implicit_save {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  return ($ENV{$UCS_SAVE_ON_LOAD_ENV_VAR}) ? 1 : 0 ;
}

#---------------------------------------------------------------------

sub get_traffic_data_matching_token {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  my $policies_md5 = Digest::MD5->new;

  my $ary_ref = $dbh->selectall_arrayref("SELECT id,encoding,name FROM PLC.PL_POLICIES");

  for my $unique_plc_ref (@$ary_ref) {
    $policies_md5->add(@$unique_plc_ref);
  }

  return $policies_md5->hexdigest();
}

#---------------------------------------------------------------------

sub save_traffic_data {
  my($self, %params) = @_;

  my $dump_file_dir = $params{dump_file_dir};
  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  if ($self->is_implicit_save()) {
      $logger->info("Not saving traffic data during an implicit save ucs") if $logger;
      return;
  }

  if (! -d $ASM_DATA_DIR) {
      $logger->info("Not saving traffic data, could not find folder '$ASM_DATA_DIR'") if $logger;
      return;
  }

  # save - [ disable | enable ]
  my $do_save_traffic_data = F5::Cluster::Utils::get_mcp_val('ucs.asm.traffic_data.save') || $UCS_ASM_TRAFFIC_DATA_SAVE;

  if ($logger) {
      $logger->debug("Saving traffic data is " . ($do_save_traffic_data ? "enabled" : "disabled" ) );
      $logger->debug("Originator is '$self->{originator}'");
  }

  if ($self->{originator} eq $UCS_INSTALL_ORIGINATOR && $do_save_traffic_data eq 'enable') {
      $logger->debug("Will save traffic data") if $logger;

      mkdir($TRAFFIC_DATA) if (! -d $TRAFFIC_DATA);
      chmod(0777, $TRAFFIC_DATA);
      File::Slurp::write_file("$TRAFFIC_DATA/$TRAFFIC_DATA_TOKEN", $self->get_traffic_data_matching_token());

      for my $table (F5::DbUtils::get_request_log_upgrade_tables(),
                     F5::DbUtils::get_learning_suggestions_upgrade_tables()) {

          my $schema_file = "$self->{dump_file_prefix}.schema.$table.$self->{dump_file_suffix}";
          my $data_file = "$self->{dump_file_prefix}.data.$table.$self->{dump_file_suffix}";

          # partitions info is appended to the bottom of the 'REQUEST_LOG' CREATE TABLE schema in a remark /*...*/
          #    it contains information that may change over time:
          #       PARTITION p200001 VALUES LESS THAN (200001) ENGINE = MyISAM,
          # we get rid of it via strip_comments opt
          my $is_table_partitioned = (exists $PARTITIONED_TABLES_TO_PARTITION_ROW_LIMIT{$table}) ? 1 : 0 ;
          my $create_sql = F5::DbUtils::get_table_schema(dbh => $dbh,
                                                         table_name => $table,
                                                         strip_comments => $is_table_partitioned,
                                                        );

          my $schema_outfile = "$TRAFFIC_DATA/$schema_file";
          open(my $table_sql_fh, '>', $schema_outfile);
          print $table_sql_fh $create_sql;
          close $table_sql_fh;

          my $data_outfile = "$dump_file_dir/$data_file.$TRAFFIC_DATA_SUFFIX";
          my $rv = F5::DbUtils::select_into_file(dbh => $dbh,
                                                 table_name => $table,
                                                 file_name => $data_outfile,
                                                );

          if (! defined $rv || ! -f $data_outfile) {
            my $err_msg = "Could not export table data $table - traffic data save aborted";
            $logger->error($err_msg) if $logger;
            die "$err_msg\n";
          }

          unlink("$TRAFFIC_DATA/$data_file.gz") if (-f "$TRAFFIC_DATA/$data_file.gz");
          gzip($data_outfile => "$TRAFFIC_DATA/$data_file.gz") or die "gzip failed: $GzipError\n";
          unlink $data_outfile;

          $logger->info("Saved $table to '$TRAFFIC_DATA/$data_file.gz'") if $logger;
      }

      $logger->info("Done saving traffic data") if $logger;
  }
  else {
      $logger->info("Will not save traffic data") if $logger;
      $self->cleanup_traffic_data(%params);
  }
}

#----------------------------------------------------------------------

sub cleanup_traffic_data {
    my($self, %params) = @_;

    my $dump_file_dir = $params{dump_file_dir};
    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    $logger->debug("Cleaning traffic data files") if $logger;
    rmtree($TRAFFIC_DATA);

    if ($dump_file_dir) {
        find({
              no_chdir => 1,
              follow => 1,
              follow_skip => 2,
              wanted => sub {
                  my $filename = $File::Find::name;
                  if ($filename =~ /\.$TRAFFIC_DATA_SUFFIX$/) {
                      $logger->trace("Deleting $filename") if $logger;
                      unlink($filename);
                  }
              },
             },
             $dump_file_dir);
    }
}

#----------------------------------------------------------------------
sub _load {
  my($self, %params) = @_;

  my $use_config_server_for_set_active = $params{use_config_server_for_set_active};
  my $dbh = $self->{dbh}; 

  my $account_names = $self->get_account_names;
  my $load_ucs_rv;
  if (!$self->{active_policy_sync_only}) {
      #schema is not dumped or loaded for active_policy_sync
      $load_ucs_rv = $self->load_ucs_tables;
  }

  my $upgrade_rv = 1;
  my $do_upgrade = 0;
  if ($self->{active_policy_sync_only}) {
      $self->{logger}->info("Active policy sync performed -- config will be directly loaded") if $self->{logger};
  }
  else {
      if (! $self->check_db_version_match) {
          $do_upgrade = 1;
          $self->{logger}->info("UCS database version does not match current version -- config will be upgraded") if $self->{logger};
      }
      else {
          $self->{logger}->debug("UCS database version matches current version") if $self->{logger};
      }

      if (! $self->{bigip_version}) {
          $self->{logger}->debug("Skipping version match check") if $self->{logger};
      }
      elsif (! $self->{ucs_version}) {
          $do_upgrade = 1;
          $self->{logger}->info("UCS version is not specified -- config will be upgraded") if $self->{logger};
      }
      elsif ($self->{bigip_version} ne $self->{ucs_version}) {
          $do_upgrade = 1;
          $self->{logger}->info("Current version and UCS version differ (current version = $self->{bigip_version}, UCS version = $self->{ucs_version}) -- config will be upgraded)") if $self->{logger};
      }
      else {
          $self->{logger}->debug("Current version and UCS version match (version = $self->{bigip_version})") if $self->{logger};
      }

      if ($self->{force_upgrade}) {
          $do_upgrade = 1;
          $self->{logger}->info("Upgrade specifically indicated -- config will be upgraded") if $self->{logger};
      }

      if (! $load_ucs_rv) {
          $do_upgrade = 1;
          $self->{logger}->info("UCS tables are in legacy format -- config will be upgraded") if $self->{logger};
      }

      if (! $self->check_account_data_files_exist()) {
          $do_upgrade = 1;
          $self->{logger}->info("UCS file is missing account tables -- config will be upgraded ") if $self->{logger};
      }
  }

  if ($do_upgrade && ! $self->{allow_upgrade}) {
      my $err_msg = "Upgrade config load attempted, but was disallowed (possibly due to cluster load)";
      $self->{logger}->error($err_msg) if $self->{logger};
      die "$err_msg\n";
  }

  if ($self->{ucs_version}) {
    $self->{ucs_version_major_minor} = F5::BigipVersionUtils::zero_padded_major_minor($self->{ucs_version});
  }
  if ($self->{bigip_version}) {
    $self->{bigip_version_major_minor} = F5::BigipVersionUtils::zero_padded_major_minor($self->{bigip_version});
  }

  my $prepareconf = F5::PrepareConf::Base->new(
                                               dbh     =>  $dbh,
                                               logger  =>  $self->{logger},
                                              );
  $prepareconf->prepare_method_pool();

  if (! $do_upgrade) {
      $self->non_upgrade_load;
  }
  else {
      $upgrade_rv = $self->upgrade_load(use_config_server_for_set_active => 0);
  }

  my $populate_mcp = 1;
  if ($self->{cmi_metadata} && $self->{call_flags}->{failover_device_group}) {
      my $fo_dg = F5::DeviceGroup::get_failover_device_group();
      if ($fo_dg->{name} && $fo_dg->{autosync_enabled} && 
          $fo_dg->{name} eq $self->{call_flags}->{failover_device_group}) {
          $populate_mcp = 0;
      }
  }

  if ($populate_mcp) {
      eval {
          $self->populate_policies_to_mcp();
      };
      if ($@) {
          my $error_msg = $@;
          chomp($error_msg);
          $self->{logger}->error("Policies could not be populated to MCP ($error_msg)") if $self->{logger};
      }
  }

  $prepareconf->prepare_violations();

  $self->set_account_name_changed($account_names);

  if (! F5::Cluster::Utils::is_clustering_enabled() || F5::Cluster::Utils::is_master_blade() ) {
      eval {
          require F5::ASMConfig::Entity::SessionAwarenessDataPoint;
          F5::ASMConfig::Entity::SessionAwarenessDataPoint->reload_session_db(dbh => $dbh,
                                                                              logger => $self->{logger},
                                                                              delete_existing => 1,
                                                                             );
      };
      if ($@) {
          my $error_msg = $@;
          chomp($error_msg);
          $self->{logger}->error("Session Awarenesss data could not be loaded ($error_msg)") if $self->{logger};
      }
  }

  my $policy_version_file = $self->get_dump_file_name("policy_version_metadata");
  my $policy_versions = [];
  if ($policy_version_file && -f $policy_version_file) {
      my $policy_versions_raw = YAML::XS::LoadFile($policy_version_file);
      if ($policy_versions_raw && @$policy_versions_raw) {
          $policy_versions = [ map { F5::ASMConfig::Policy->new(%$_) } @$policy_versions_raw ];
      }
  }

  $self->cleanup if $self->{cleanup_temporary_files};

  upload_certificates( dbh => $dbh, logger => $self->{logger});
  copy_schemas_to_var();

  for my $db_file_reference (@db_file_references) {
      F5::DbUtils::delete_unreferenced_files(dbh => $dbh,
                                             logger => $self->{logger},
                                             db_file_refs => $db_file_reference->{db_file_refs},
                                             dir => $db_file_reference->{dir},);

  }

  return { completed_successfully => $upgrade_rv, 
           policy_versions => $policy_versions,
         };
}

#----------------------------------------------------------------------

sub check_db_version_match {
  my($self) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  my $version_build_metadata = {};
  my $version_build_metadata_filename = $self->get_dump_file_name("version_build_metadata");
  if (-f $version_build_metadata_filename) {
      $version_build_metadata = YAML::XS::LoadFile($self->get_dump_file_name("version_build_metadata"));
  }
  my $current_install_sql_checksums = $self->get_install_sql_checksums();
  my $is_install_sql_changed = 0;
  for my $dbname (keys %$current_install_sql_checksums) {
      if (! $version_build_metadata->{install_sql_checksums} ||
          ! $version_build_metadata->{install_sql_checksums}->{$dbname} ||
          $version_build_metadata->{install_sql_checksums}->{$dbname} ne $current_install_sql_checksums->{$dbname}) {

          $logger->info("Base install database definitions differ (database = $dbname)") if $logger;
          if ($version_build_metadata && $version_build_metadata->{install_sql_checksums} && $version_build_metadata->{install_sql_checksums}->{$dbname}) {
              $logger->debug("Base install SQL differs (database = $dbname, previous SQL MD5 = $version_build_metadata->{install_sql_checksums}->{$dbname}, current SQL MD5 = $current_install_sql_checksums->{$dbname})") if $logger;
          }
          $is_install_sql_changed = 1;
      }
  }

  if ($is_install_sql_changed) {
      return;
  }

  $logger->info('UCS and staging tables match') if $logger;
  return 1;
}

#----------------------------------------------------------------------

sub check_account_data_files_exist {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    my $are_all_account_data_files_present = 1;

    my %loaded_tables;
    my @account_data_db_fields = map { @{$_->{db_file_refs}} } @db_file_references;
    for my $account_data_db_field (@account_data_db_fields) {
        my ($base_dbname, $table_name) = split(/\./, $account_data_db_field->{table_name});
        my $dbname = $base_dbname . "_ucs";

        if (! F5::DbUtils::check_table_exists(dbh => $dbh,
                                              dbname => $dbname,
                                              table_name => $table_name)) {
            next;
        }

        if (! grep { $_ eq $account_data_db_field->{field} } F5::DbUtils::get_db_fields($dbh, "$dbname.$table_name")) {
            next;
        }

        # Temporarily load the contents of the table to UCS db (only once per table needed)
        if (! $loaded_tables{"$dbname.$table_name"}) {
            my $data_file = $self->get_dump_file_dir_from_pointer . "/$self->{dump_file_prefix}.data.$base_dbname.$table_name.$self->{dump_file_suffix}";
            my $rv = $dbh->do(<<EndQuery);
LOAD DATA INFILE '$data_file'
    INTO TABLE $dbname.$table_name
  FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY '"'
EndQuery

            if (! defined $rv) {
              my $err_msg = "Could not import table $dbname.$table_name data from file " . basename($data_file) . " - ASM configuration load aborted.";
              $logger->error($err_msg . " Data file - '$data_file'") if $logger;
              die "$err_msg\n";
            }

           $loaded_tables{"$dbname.$table_name"} = 1;
        }

        my $sth = $dbh->prepare("SELECT $account_data_db_field->{field} FROM $dbname.$table_name");
        $sth->execute;

        while (my $rec = $sth->fetchrow_arrayref) {

            if ($rec->[0] && ! -e $rec->[0]) {
                $logger->debug("Missing enforcer data file ($rec->[0])") if $logger;
                $are_all_account_data_files_present = 0;
            }
        }
    }

    # Clear out all temporarily loaded tables
    for my $loaded_table (keys %loaded_tables) {
        $dbh->do("DELETE FROM $loaded_table");
    }

    if ($are_all_account_data_files_present) {
        $logger->debug('All expected enforcer data files are present') if $logger;
    }
    return $are_all_account_data_files_present;
}

#---------------------------------------------------------------------
sub non_upgrade_load {
  my ($self, %params) = @_;

  $self->load_infiles_to_db(dest_dbname => 'bigip_dbname',
                            load_traffic_data => 1,
                           );
}
#----------------------------------------------------------------------

my $CMI_RELAY_SETTINGS = {
                           pb_config => { 
                                         table_name => 'PLC.PL_POLICY_BUILDER_CONFIG',
                                         field_name => 'flg_enabled',
                                         disable_only => 1,
                                        },
                           scanner_config => { table_name => 'PLC.PL_SCANNER_CONFIG',
                                               field_name => 'flg_bidir_enabled',
                                             },
                         };

sub load_infiles_to_db {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};
  my $dest_dbname = $params{dest_dbname} || '';
  my $load_traffic_data = $params{load_traffic_data};

  my $dump_file_dir = $self->get_dump_file_dir_from_pointer;

  $logger->debug("Loading data in $dest_dbname tables") if $logger;
  $dbh->begin_work;
  my $is_failover_standby = F5::Cluster::Utils::is_failover_standby();
  my $old_configs;
  if ( $self->{cmi_metadata} && !$is_failover_standby ) {
      for my $cmi_relay_setting (keys %$CMI_RELAY_SETTINGS) {
          my $table_name = $CMI_RELAY_SETTINGS->{$cmi_relay_setting}->{table_name};
          my $field_name = $CMI_RELAY_SETTINGS->{$cmi_relay_setting}->{field_name};
      
          $old_configs->{$cmi_relay_setting} = $dbh->selectall_hashref(<<"EndQuery", 'name');
SELECT name, $field_name
  FROM PLC.PL_POLICIES JOIN $table_name
   ON id = policy_id
EndQuery
      }
  }


  for my $database (@{$self->{databases}}) {
    my %tables_to_skip;
    my @data_files = glob("$dump_file_dir/$self->{dump_file_prefix}.data.$database->{bigip_dbname}.*");

    for my $data_file (@data_files) {
      if (! -e $data_file) {
        my $err_msg = "Temporary data file $data_file disappeared during save (perhaps concurrent config save/install in progress?) - ASM configuration save aborted";
        $logger->error($err_msg) if $logger;
        die "$err_msg\n";
      }

      my($table) = ($data_file =~ /$self->{dump_file_prefix}\.data\.$database->{bigip_dbname}\.(.*?)\.$self->{dump_file_suffix}/);
      if ($tables_to_skip{$table}) {
          $logger->debug("Skipping loading data for table '$table'");
          next;
      }

      my $dbname = ($dest_dbname ? $database->{$dest_dbname} : $database->{bigip_dbname});
      my $full_db_table = "$dbname.$table";

      my $active_policy_where = '';
      if ($self->{active_policy_sync_only}) {
          if (defined($self->{policy_name_crc}) && (grep {$_ eq "policy_name_crc"} F5::DbUtils::get_db_fields($dbh, $full_db_table))) {
              $active_policy_where = "WHERE policy_name_crc= $self->{policy_name_crc}";
          }
      }

      $logger->debug("Loading data to $full_db_table") if $logger;
      $dbh->do("DELETE FROM $full_db_table $active_policy_where");

      my $rv = $dbh->do(<<EndQuery);
LOAD DATA INFILE '$data_file'
    INTO TABLE $full_db_table
  FIELDS TERMINATED BY ',' OPTIONALLY ENCLOSED BY '"'
EndQuery

      if (! defined $rv) {
        my $err_msg = "Could not import table $full_db_table data from file " . basename($data_file) . " - ASM configuration load aborted";
        $logger->error($err_msg . " Data file - '$data_file'") if $logger;
        die "$err_msg\n";
      }
    }
  }

  my $sync_from_active_device = 0;
  my $sync_request_from_active_device = 0;

  if ($self->{asm_config_handler} && 
      $self->{cmi_metadata} && -s $self->get_dump_file_name("cmi_metadata")) {

      $logger->debug("Loading cmi metadata") if $logger;
      my $metadata = YAML::XS::LoadFile($self->get_dump_file_name("cmi_metadata"));
      %{$self->{asm_config_handler}->{known_times}} = %{$metadata->{known_times}};
      $self->{asm_config_handler}->{shared_state}->{recovery_time} = $metadata->{ucs_end};
      $self->{asm_config_handler}->{sync_info}->{originator} = $self->{call_flags}->{originator};
      $self->{asm_config_handler}->{sync_info}->{timestamp}  = $metadata->{ucs_end};
      $sync_from_active_device = $metadata->{sync_from_active_device};
      $sync_request_from_active_device = $metadata->{sync_request_from_active_device};
  }
  if ( $self->{cmi_metadata} && !$is_failover_standby && $sync_from_active_device 
       && $sync_request_from_active_device && F5::DeviceGroup::is_device_active($self->{call_flags}->{originator})) {
      for my $cmi_relay_setting (keys %$CMI_RELAY_SETTINGS) { 
          my $table_name = $CMI_RELAY_SETTINGS->{$cmi_relay_setting}->{table_name};
          my $field_name = $CMI_RELAY_SETTINGS->{$cmi_relay_setting}->{field_name};

          my $new_configs = $dbh->selectall_hashref(<<"EndQuery", 'id');
SELECT id, name, $field_name
  FROM PLC.PL_POLICIES JOIN $table_name
   ON id = policy_id
EndQuery

          my @policies_to_disable = grep {$new_configs->{$_}->{$field_name}} keys (%$new_configs);
          my @policies_to_enable  = grep {!$new_configs->{$_}->{$field_name} && $old_configs->{$cmi_relay_setting}->{$new_configs->{$_}->{name}}->{$field_name} } keys (%$new_configs);

          if (@policies_to_disable) {
              my $placeholders = join (',', map {"?"} @policies_to_disable);
              $self->{dbh}->do(<<EndQuery, undef, @policies_to_disable);
UPDATE $table_name
     SET $field_name = 0
     WHERE policy_id IN ($placeholders)
EndQuery
          }

          if (!$CMI_RELAY_SETTINGS->{$cmi_relay_setting}->{disable_only} && @policies_to_enable) {
              my $placeholders = join (',', map {"?"} @policies_to_enable);
              $self->{dbh}->do(<<EndQuery, undef, @policies_to_enable);
UPDATE $table_name
     SET $field_name = 1
     WHERE policy_id IN ($placeholders)
EndQuery
          }

      }
  }
  
  $dbh->commit;
  F5::Syslog::ts_syslog( 'info', 'ASM config loaded');

  if ($load_traffic_data) {
      eval {
          $self->load_traffic_data();
      };
      if (my $err = $@) {
          $logger->error("Could not load traffic data") if $logger;
          $logger->error($err) if $logger;
      }
  }

}

#---------------------------------------------------------------------

sub check_and_log_file_existence {
  my($self, %params) = @_;

  my $logger = $self->{logger};
  my $file = $params{file_name};
  my $non_zero_size = $params{non_zero_size};

  if (! -f $file) {
      my $err_msg = "Can't locate file '$file'" ;
      $logger->info($err_msg) if $logger;
      return 0;
  }

  if ($non_zero_size && -z $file) {
      my $err_msg = "File '$file' is empty" ;
      $logger->info($err_msg) if $logger;
      return 0;
  }

  return 1;
}

#---------------------------------------------------------------------

sub is_traffic_data_matching_ucs {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};
  my $token_file = $params{token_file} || '';

  return unless -f $token_file;

  return File::Slurp::read_file("$token_file") eq $self->get_traffic_data_matching_token();
}

#---------------------------------------------------------------------

sub load_traffic_data {
  my($self, %params) = @_;

  my $dump_file_dir = $self->get_dump_file_dir_from_pointer;
  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  if (! -d $ASM_DATA_DIR) {
      $logger->info("Not loading traffic data, could not find folder '$ASM_DATA_DIR'") if $logger;
      return;
  }

  my $is_supported_load_traffic_data = 1;
  if ($self->{ucs_version_major_minor} && $self->{bigip_version_major_minor} &&
      $self->{ucs_version_major_minor} ne $self->{bigip_version_major_minor}) {
      $logger->debug("Loading traffic data from version $self->{ucs_version} to version $self->{bigip_version} is not supported ") if $logger;;
      $is_supported_load_traffic_data = 0;
  }

  # load - [ never | matching | always ]
  my $do_load_traffic_data = F5::Cluster::Utils::get_mcp_val('ucs.asm.traffic_data.load') || $UCS_ASM_TRAFFIC_DATA_LOAD;
  my $traffic_data_dir_exists = (-d $TRAFFIC_DATA) ? 1 : 0 ;
  my $traffic_data_matching_ucs = $self->is_traffic_data_matching_ucs(token_file => "$TRAFFIC_DATA/$TRAFFIC_DATA_TOKEN");

  if ($logger) {
      $logger->debug("Traffic data directory " . ($traffic_data_dir_exists ? "exists" : "does not exist") );
      $logger->debug("Traffic data " . ($traffic_data_matching_ucs ? "mathches ucs" : "does not match ucs" ) );
      $logger->debug("Loading traffic data is " . ($do_load_traffic_data ? "enabled" : "disabled" ) );
      $logger->debug("Originator is '$self->{originator}'");
      $logger->debug("This is " . ($self->{ucs_post_install} ? "" : "not ") . "a ucs post install load" );
  }

  if (($self->{originator} eq $UCS_INSTALL_ORIGINATOR || $self->{ucs_post_install}) && $traffic_data_dir_exists &&
       $is_supported_load_traffic_data &&
      ($do_load_traffic_data eq 'always' ||
       ($do_load_traffic_data eq 'matching' && $traffic_data_matching_ucs)
      )
     ) {

      $logger->info("Will load traffic data") if $logger;

      for my $table (F5::DbUtils::get_request_log_upgrade_tables(),
                     F5::DbUtils::get_learning_suggestions_upgrade_tables()) {

          my $data_file = "$self->{dump_file_prefix}.data.$table.$self->{dump_file_suffix}";
          my $schema_file = "$self->{dump_file_prefix}.schema.$table.$self->{dump_file_suffix}";

          my $data_infile_gz = "$TRAFFIC_DATA/$data_file.gz";
          next unless $self->check_and_log_file_existence(file_name => $data_infile_gz);

          my $schema_infile = "$TRAFFIC_DATA/$schema_file";
          next unless $self->check_and_log_file_existence(file_name => $schema_infile,
                                                          non_zero_size => 1,
                                                         );

          my ($dbname_aliased,$table_name_aliased) = split('\.',$table);
          if (! F5::DbUtils::check_table_exists(dbh => $dbh,
                                                dbname => $dbname_aliased,
                                                table_name => $table_name_aliased)) {
              $logger->info("Can't locate table '$table'") if $logger;
              next;
          }

          my $schema_infile_sql = F5::Utils::File::file_slurp($schema_infile);

          # partitions info is appended to the bottom of the 'REQUEST_LOG' CREATE TABLE schema in a remark /*...*/
          #    it contains information that may change over time:
          #       PARTITION p200001 VALUES LESS THAN (200001) ENGINE = MyISAM,
          # we get rid of it via strip_comments opt
          my $is_table_partitioned = (exists $PARTITIONED_TABLES_TO_PARTITION_ROW_LIMIT{$table}) ? 1 : 0 ;
          my $create_sql = F5::DbUtils::get_table_schema(dbh => $dbh,
                                                         table_name => $table,
                                                         strip_comments => $is_table_partitioned,
                                                        );

          $create_sql =~ s/ AUTO_INCREMENT=\d+//;
          $schema_infile_sql =~ s/ AUTO_INCREMENT=\d+//;
          chomp($create_sql);
          chomp($schema_infile_sql);

          if (!$create_sql || !$schema_infile_sql || $create_sql ne $schema_infile_sql) {
              $logger->info("Database schema differs for table '$table' -- skipping load") if $logger;
              $logger->debug("file '$schema_infile' schema '$schema_infile_sql' differs from table '$table' schema '$create_sql'") if $logger;
              next;
          }

          my $data_infile = "$dump_file_dir/$data_file.$TRAFFIC_DATA_SUFFIX";
          unlink ($data_infile) if -f $data_infile;

          gunzip($data_infile_gz => $data_infile) or die "gunzip failed: $GunzipError\n";
          next unless $self->check_and_log_file_existence(file_name => $data_infile,
                                                          non_zero_size => 1,
                                                         );

          $logger->trace("DELETE FROM $table") if $logger;
          $dbh->do("DELETE FROM $table");

          if ($is_table_partitioned) {
              $logger->debug("The table '$table' is partitioned") if $logger;
              # The file is not loaded into memory, so this will work even for gigantic files.
              #    source - http://search.cpan.org/~toddr/Tie-File-1.00/lib/Tie/File.pm
              # memory => 0
              #    cache nothing since it's a one time access
              tie my @rows, 'Tie::File', $data_infile, mode => O_RDONLY, memory => 0 or die "error: $!\n";
              my $first_line = $rows[0];
              my $last_line = $rows[-1];
              untie @rows;

              my $first_id;
              if ($first_line =~ /^(\d+),/) {
                  $first_id = $1;
              }

              my $last_id;
              if ($last_line =~ /^(\d+),/) {
                  $last_id = $1;
              }

              # create as much empty partitions as required to contain all records from id=1 up to the largest id
              if ($first_id && $last_id) {
                  $logger->debug("ID range of table $table is from $first_id to $last_id") if $logger;

                  F5::Db::Partition::remove_partitioning(dbh => $dbh,
                                                         table_name => $table,
                                                         logger => $logger,
                                                        );

                  my $partition_row_limit = $PARTITIONED_TABLES_TO_PARTITION_ROW_LIMIT{$table};

                  # a partition name is 'pID' and it stores entries with ids in the range: (1;ID-1).
                  # if a partition's row limit (partition_row_limit) is say 100 rows, then partition names go like this:
                  #    p101 (ids:   1 -> 100)
                  #    p201 (ids: 101 -> 200)
                  # so, a partition's name is, in fact 'p' followed by the next partition's first id 'next_partition_first_id'
                  my $next_partition_first_id = $partition_row_limit * (POSIX::ceil($first_id / $partition_row_limit)) + 1;

                  $logger->trace("Setting up initial partitioning for table $table") if $logger;
                  my $partition_name = F5::Db::Partition::setup_partitioning(dbh => $dbh,
                                                                             table_name => $table,
                                                                             next_partition_first_id => $next_partition_first_id,
                                                                             logger => $logger,
                                                                            );

                  $logger->debug("Initial partition name for table $table is $partition_name") if $logger;

                  my $request_log = F5::RequestLog->new(dbh => $dbh,
                                                        logger => $logger,
                                                       );

                  # max_id is a bit misleading, cause in fact a partition will hold entries w/ IDs less then the value of 'max_id'
                  push(@{$request_log->{partitions}},
                       { PARTITION_NAME => $partition_name,
                         max_id => $next_partition_first_id });

                  $logger->trace("Adjusting required partitions for table $table") if $logger;
                  for my $i ($first_id .. $last_id) {
                      $request_log->adjust_partitions(next_request_log_id => $i,
                                                      table_name => $table,
                                                     );
                  }
              }
          }

          $logger->debug("Loading data from file $data_infile into table $table") if $logger;
          my $rv = F5::DbUtils::load_from_file(dbh => $dbh,
                                               table_name => $table,
                                               file_name => $data_infile,
                                              );


          # after loading the records and if the table is partitioned -
          #    there might be some empty partitions left.
          # drop empty partitions
          if ($is_table_partitioned) {
              $logger->debug("Deleting empty partitions from table $table") if $logger;

              my $partitions = F5::Db::Partition::retrieve_partitions_info(dbh => $dbh,
                                                                           table_name => $table,
                                                                           logger => $logger,
                                                                          );

              for my $partition (@{$partitions}) {
                  next if $partition->{TABLE_ROWS};
                  next unless $partition->{PARTITION_NAME};
                  $logger->debug("Deleting empty partition $partition->{PARTITION_NAME} from table $table") if $logger;

                  F5::Db::Partition::delete_partition(dbh => $dbh,
                                                      table_name => $table,
                                                      partition_name => $partition->{PARTITION_NAME},
                                                      logger => $logger,
                                                     );
              }
          }

          unlink $data_infile;

          if (! defined $rv) {
            my $err_msg = "Could not import table data $table - traffic data load aborted";
            $logger->error($err_msg) if $logger;
            die "$err_msg\n";
          }

          $logger->info("Done loading traffic data for table $table") if $logger;
      }

      $logger->info("Done loading traffic data") if $logger;
  }
  else {
      $logger->info("Will not load traffic data") if $logger;
  }
}

#---------------------------------------------------------------------

sub populate_policies_to_mcp {
    my ($self, %params) = @_;
    #Populate all policies to MCP
    
    # TBD: retrieve_all had this critreria:  get_criteria => { flg_is_hsl => 0 }
    # instead, we need to make sure that hsl policies are not in the db to start with
    my @asm_policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh},
                                                                   get_criteria => {} );
    my %asm_policy_hash = map { ( $_->{policy_name} => $_ ) } @asm_policies;
    my %mcp_folders = map { ( $_->{name} => $_ ) } @{F5::ASM::COAPI::Query->new('folder')};
    my @mcp_policies = @{F5::ASM::COAPI::Query->new('asm_policy')};
    for my $mcp_policy (@mcp_policies) {
        my $policy_name = $mcp_policy->{name};    
        if ( !$asm_policy_hash{$policy_name} 
             || ($mcp_policy->{is_active_policy} && !$asm_policy_hash{$policy_name}->{is_active_policy}) ) {
            #Attempt to unassociate from all L7 (ltm) policies, and delete policy from mcp
            eval {
                my %l7_policy_names = map { ( $_->{policy_name} => 1) } (@{F5::ASM::COAPI::Query->new("policy_action",
                                                                                                   { target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM,
                                                                                                     action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE,
                                                                                                     policy => $policy_name},
                                                                                                     )});
                for my $l7_policy_name (keys %l7_policy_names) {
                    my $l7_policy = F5::ASMConfig::Entity::L7Policy->new(get_criteria => { l7_policy_name => $l7_policy_name} , dbh => $self->{dbh} );
                    if (!$l7_policy || $l7_policy->{type} != F5::ASMConfig::VirtualServerAssocType::VIRTUAL_SERVER_ASSOC_TYPE_L7_SIMPLE()) {
                        next;
                    }

                    for my $virtual_server (@{$l7_policy->{virtual_servers}}) {
                        $self->{asm_config_handler} ||= F5::ASMConfig::GuaranteedClient::get_client(username => 'configsync');
                        $self->{asm_config_handler}->delete_virtual_server_assoc($self->{asm_config_handler}->{session}, 
                                                                                 F5::ASMConfig::CallFlags->new(ignore_nonexistent_record_on_delete => 1, do_not_relay => 1),
                                                                                 F5::ASMConfig::Policy->new(policy_name => $policy_name),
                                                                                 F5::ASMConfig::VirtualServerAssoc->new(virtual_server => $virtual_server,
                                                                                                                        assoc_type => $l7_policy->{type},
                                                                                                                        l7_policy_name => $l7_policy_name));
                    }
                }
                if (!$asm_policy_hash{$policy_name}) {
                    $mcp_policy->delete();
                }
            };
            if ($@) {
                my $error_msg = $@;
                chomp($error_msg);
                $self->{logger}->error("Policy: $policy_name could not be modified or deleted in MCP ($error_msg)") if $self->{logger};
            }
        }
    }

    for my $policy (@asm_policies) {
        my $folder = dirname($policy->{policy_name});
        if (!$mcp_folders{$folder}) {
            #recursively create any missing folders
            my @path_parts = split('/', $folder);
            my $path = shift @path_parts; #get rid of first empty match;
            for my $part (@path_parts) {
                $path .= "/$part";
                if (!$mcp_folders{$path}) {
                    my $obj = F5::ASM::COAPI::Object->new('folder');
                    $obj->{name} = $path;
                    $obj->save();
                    $mcp_folders{$path} = $obj;
                }
            }
        }
        $policy->{app_id} = $mcp_folders{$folder}->{app_id};
        eval {
            $policy->mcp_store();
        };
        if ($@) {
            my $error_msg = $@;
            chomp($error_msg);
            $self->{logger}->error("Policy: $policy->{policy_name} could not be saved to MCP ($error_msg)") if $self->{logger};
        }
    }
    
}
#---------------------------------------------------------------------

sub upgrade_load {
  my($self, %params) = @_;

  my $use_config_server_for_set_active = $params{use_config_server_for_set_active};
  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  if ($self->{bigip_version_major_minor} && $self->{ucs_version_major_minor} &&
      $self->{bigip_version_major_minor} lt $self->{ucs_version_major_minor} ) {
      my $error = "Invalid UCS.  Cannot load configuration from newer version ($self->{ucs_version})";
      $logger->error($error);
      die "$error\n";
  }

  $self->create_staging_dbs;
  $self->load_old_ucs_db_from_files;

  $dbh->begin_work;

  $logger->info('Starting upgrade') if $logger;

  my %external_idmap;
  F5::DbUpgrade::db_upgrade(new_verison => $self->{bigip_version},
			    old_version => $self->{ucs_version},
			    dbh => $dbh,
			    databases => $self->{databases},
                            external_idmap => \%external_idmap,
                           );

  #remove ucs DB
  for my $database (@{$self->{databases}}) {
    $self->remove_database($database->{ucs_dbname});
  }

  F5::PrepareConf::Checksums->new(dbh => $dbh)->empty_checksums;

  for my $database (@{$self->{databases}}) {
    my @tables = (F5::DbUtils::get_config_sync_tables($database->{bigip_dbname}),
                  F5::DbUtils::get_active_policy_tables($database->{bigip_dbname}));

    for my $table (@tables) {
      my $staging_table = $table;
      $staging_table =~ s/^$database->{bigip_dbname}/$database->{staging_dbname}/;
      $dbh->do("DELETE FROM $table");
      $dbh->do(<<EndQuery);
INSERT INTO $table
 SELECT *
   FROM $staging_table
EndQuery
    }
  }

  #remove staging DB
  for my $database (@{$self->{databases}}) {
    $self->remove_database($database->{staging_dbname});
  }

  $logger->debug("use_config_server_for_set_active=$use_config_server_for_set_active") if $logger;

  my $rv = $self->restore_active_policies(use_config_server_for_set_active => $use_config_server_for_set_active,
                                          external_idmap => \%external_idmap,
                                         );

  $self->{external_idmap} = \%external_idmap;

  # Save bigip.conf immediately before upgrade is completed (to preserve upgraded DoS profiles, logging profiles, etc.)
  F5::Utils::Rest::populate_uuids(dbh => $dbh,
                                  handler => $self->{asm_config_handler},
                                 );

  F5::ASM::COAPI::SaveAll(logger => $logger, timeout => $COAPI_TRANSACTION_TIMEOUT);

  $dbh->commit;

  $logger->info('Upgrade completed') if $logger;

  eval {
      $self->load_traffic_data();
  };
  if (my $err = $@) {
      $logger->error("Could not load traffic data") if $logger;
      $logger->error($err) if $logger;
  }

  return $rv;
}

#----------------------------------------------------------------------

sub create_staging_dbs {
  my($self) = @_;

  my $mysql_cmd_fh = $self->start_mysql_cmdline;

  for my $database (@{$self->{databases}}) {
    $self->remove_database($database->{staging_dbname});
    $self->create_database($database->{staging_dbname});

    print $mysql_cmd_fh "USE $database->{staging_dbname};\n";
    open(my $template_fh, '<', F5::DbUpgrade::get_sql_filename($database->{bigip_dbname}));
    while(my $line = <$template_fh>) {
      print $mysql_cmd_fh $line;
    }
    close $template_fh;
  }
  close $mysql_cmd_fh;
}

#----------------------------------------------------------------------

sub cleanup {
  my($self, %params) = @_;
  my $do_remove_temp_databases = defined $params{remove_temp_databases} ? $params{remove_temp_databases} : 1;
  my $clean_previous_runs = defined($params{clean_previous_runs}) ? $params{clean_previous_runs} : 0;

  my $logger = $self->{logger};

  $logger->debug('Cleaning up temporary files and tables') if $logger;

  if ($do_remove_temp_databases) {
      $self->remove_temp_dbs;
  }

  unlink("$self->{base_dir}/$self->{ucs_v1_dump_file}") if (-e "$self->{base_dir}/$self->{ucs_v1_dump_file}");

  for my $database (@{$self->{databases}}) {
    unlink("$self->{base_dir}/$database->{ucs_v2_schema_filename}") if (-e "$self->{base_dir}/$database->{ucs_v2_schema_filename}");
    unlink("$self->{base_dir}/$database->{ucs_v2_data_filename}") if (-e "$self->{base_dir}/$database->{ucs_v2_data_filename}");
  }

  my $dump_file_dir = $params{dump_file_dir} || $self->get_dump_file_dir_from_pointer;

  if ($dump_file_dir && -d $dump_file_dir) {
    $logger->trace("Cleaning up from dump_file_dir: $dump_file_dir") if $logger;
    rmtree($dump_file_dir);
    $self->cleanup_traffic_data(dump_file_dir => $dump_file_dir, %params);
  }

  if (-e $self->ucs_v3_pointer_file) {
    unlink $self->ucs_v3_pointer_file;
  }

  if (-e $self->ucs_version_file) {
    unlink $self->ucs_version_file;
  }


  if ($self->{created_temp_base_dir}) {
    rmtree($self->{base_dir});
  }

  if ($clean_previous_runs) {
      my(@previous_dirs) = glob("$self->{base_dir}/$self->{dump_file_prefix}.save_dir_*.$self->{dump_file_suffix}");
      for my $previous_dir (@previous_dirs) {
          $logger->trace("Cleaning up from previous run: $previous_dir") if $logger;
          rmtree($previous_dir);
      }
  }
}

#----------------------------------------------------------------------

# Starting from Negev, UCS contains separate files for tables and data.
# We load the tables schema first. If the bigip version and database schema
# match the UCS, the data files will be loaded directly into the live database.
# Otherwise, the data will be loaded to the UCS database and upgraded.

sub load_ucs_tables {
  my($self) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  $logger->debug('Creating UCS database') if $logger;

  for my $database (@{$self->{databases}}) {
    $self->remove_database($database->{ucs_dbname});
    $self->create_database($database->{ucs_dbname});
  }

  if (-e $self->ucs_v3_pointer_file) {
    $logger->debug('Loading v3 schemas') if $logger;
    my $dump_file_dir = $self->get_dump_file_dir_from_pointer;

    # Check failsafe files
    if (! $self->is_ucs_complete()) {

      if (-e "/ts/var/upgrade_failure_override") {
        warn "UCS contains incomplete ASM configuration, but override file exists -- some data may be incomplete\n";
      }
      else {
        my $err_msg = "UCS contains incomplete ASM configuration - aborting ASM load";
        $logger->error($err_msg) if $logger;
        die "$err_msg\n";
      }
    }

    $dbh->begin_work;

    for my $database (@{$self->{databases}}) {
      my @schema_files = glob("$dump_file_dir/$self->{dump_file_prefix}.schema.$database->{bigip_dbname}.*");

      for my $schema_file (@schema_files) {
        $logger->debug("Loading $schema_file to UCS database") if $logger;

        if (! -e $schema_file) {
          my $err_msg = "Temporary schema file $schema_file disappeared during save (perhaps concurrent config save/install in progress?) - ASM configuration save aborted";
          $logger->error($err_msg) if $logger;
          die "$err_msg\n";
        }

        my $schema_sql = F5::Utils::File::file_slurp($schema_file);

        $schema_sql =~ s/^(CREATE TABLE )/$1$database->{ucs_dbname}./;
        $dbh->do($schema_sql);
      }
    }
    $dbh->commit;
    return 1;
  }
  else {
    return;
  }
}

#----------------------------------------------------------------------

sub is_ucs_complete {
    my($self) = @_;
    my $dump_file_dir = $self->get_dump_file_dir_from_pointer;

    if (-s $self->get_dump_file_name("failsafe_start")) {
        # version with failsafe files

        if (-s $self->get_dump_file_name("failsafe_end")) {
            my $failsafe_end_result = F5::Utils::File::file_slurp($self->get_dump_file_name("failsafe_end"));
            chomp($failsafe_end_result);
            if ($failsafe_end_result eq '1') {
                # Previous version failsafe without file count -- consider as complete
                return 1;
            }

            my($success_flag, $failsafe_schema_file_count, $failsafe_data_file_count) = split(/,/, $failsafe_end_result);
            $failsafe_schema_file_count =~ s/\"//g;
            $failsafe_data_file_count =~ s/\"//g;
            my @schema_files = glob("$dump_file_dir/$self->{dump_file_prefix}.schema.*.$self->{dump_file_suffix}");
            my $schema_file_received_count = scalar @schema_files;
            my @data_files = glob("$dump_file_dir/$self->{dump_file_prefix}.data.*.$self->{dump_file_suffix}");
            my $data_file_received_count = scalar @data_files;
            if ($failsafe_schema_file_count == $schema_file_received_count &&
                $failsafe_data_file_count == $data_file_received_count) {

                return 1;
            }
            else {
                if ($failsafe_schema_file_count != $schema_file_received_count) {
                    $self->{logger}->warn("Schema file count in UCS differs from expected (expected = $failsafe_schema_file_count, actual = $schema_file_received_count)") if $self->{logger};
                }
                if ($failsafe_data_file_count != $data_file_received_count) {
                    $self->{logger}->warn("Data file count in UCS differs from expected (expected = $failsafe_data_file_count, actual = $data_file_received_count)") if $self->{logger};
                }
            }
        }
    }
    else {
        if (-s $self->get_dump_file_name("data.PLC.NEGSIG_SIGNATURES")) {
            # No failsafe files in this version, but NEGSIG_SIGNATURES table exists
            return 1;
        }
    }

    return;
}

#----------------------------------------------------------------------

sub load_old_ucs_db_from_files {
  my($self, %params) = @_;

  my $logger = $self->{logger};

  $logger->debug('Loading UCS data') if $logger;

  if (-e $self->ucs_v3_pointer_file) {
    # Table schemas have already been loaded, so now load just the data
    $self->load_infiles_to_db(dest_dbname => 'ucs_dbname',
                              load_traffic_data => 0,
                             );
  }
  elsif (-e "$self->{base_dir}/$self->{databases}->[0]->{ucs_v2_schema_filename}") {
    for my $database (@{$self->{databases}}) {
      $logger->debug("Loading UCS schema and data to $database->{ucs_dbname}") if $logger;
      $self->load_sql_file("$self->{base_dir}/$database->{ucs_v2_schema_filename}", $database->{ucs_dbname});
      $self->load_sql_file("$self->{base_dir}/$database->{ucs_v2_data_filename}", $database->{ucs_dbname});
    }
  }
  else {
    # Pre-Sonoma style UCS
    my %ucs_dbname;
    for my $database (@{$self->{databases}}) {
      $logger->debug("Loading UCS schema and data to $database->{ucs_dbname}") if $logger;
      $ucs_dbname{$database->{bigip_dbname}} = $database->{ucs_dbname};
    }

    change_file_inplace("$self->{base_dir}/$self->{ucs_v1_dump_file}",
			sub {
			  my $line = shift;
			  $line =~ s/^\s*(USE `?)(.*?)(`?;)/$1$ucs_dbname{$2}$3/;
			  return $line;
			}
		       );
    $self->load_sql_file("$self->{base_dir}/$self->{ucs_v1_dump_file}");
  }
}

#----------------------------------------------------------------------
sub load_sql_file {
  my($self, $filename, $dbname) = @_;
  my @cmd = ('mysql', "-u$self->{mysql_dump_user}", "-p$self->{mysql_dump_password}");
  my $cmd_fh;
  my $in_fh;

  open ($cmd_fh, '|-', @cmd)
    || die( qq/Failed to load: piped open '$cmd[0]': $!/ );

  print $cmd_fh "USE $dbname;\n" if ($dbname);

  open ($in_fh, '<', $filename)
    || die( qq/Failed to load: open file '$filename': $!/ );

  while ( <$in_fh> ) {
    print $cmd_fh $_
      || die( qq/Failed to load: print to piped cmd '$cmd[0]': $!/ );
  }

  close ($cmd_fh)
    || die( qq/Failed to load: close pipe '$cmd[0]': $! ($?)/ );
  close $in_fh;
}

#----------------------------------------------------------------------

sub change_file_inplace {
  my($filename, $change_func) = @_;
  my $tempfile = "$filename.tmp.$$";
  open(my $in_fh, '<', $filename);
  open(my $out_fh, '>', $tempfile);

  while(my $line = <$in_fh>) {
    $line = $change_func->($line);
    print $out_fh $line;
  }
  close $in_fh;
  close $out_fh;

  rename($tempfile, $filename);
}

#----------------------------------------------------------------------

sub remove_temp_dbs {
  my($self) = @_;
  for my $database (@{$self->{databases}}) {
    $self->remove_database($database->{ucs_dbname});
    $self->remove_database($database->{staging_dbname});
  }
}

#----------------------------------------------------------------------

sub start_mysql_cmdline {
  my ($self) = @_;

  my $mysql_cmd_fh;
  my @cmd = ('mysql', "-u$self->{mysql_dump_user}", "-p$self->{mysql_dump_password}");
  open $mysql_cmd_fh, '|-', @cmd;

  return $mysql_cmd_fh;
}

#----------------------------------------------------------------------

sub restore_active_policies {
  my($self, %params) = @_;
  my $use_config_server_for_set_active = $params{use_config_server_for_set_active};
  my $dbh = $self->{dbh};
  my $logger = $self->{logger};
  my $external_idmap = $params{external_idmap};
  my $rv = 1;

  my $update_core_logger = F5::Logger->get_logger(logger_name => 'update_core');

  my $active_policies   = $dbh->selectall_hashref('SELECT id, name FROM PLC.PL_POLICIES WHERE account_id > 0', 'id');
  my $whitehat_policies = $dbh->selectall_hashref('SELECT policy_id FROM PLC.PL_SCANNER_CONFIG WHERE scanner_type LIKE ?', 'policy_id', 
                                                  undef, 'sentinel%');
  my $child_policies    = $dbh->selectall_hashref('SELECT policy_id FROM PLC.PL_POLICY_PARENT WHERE parent_policy_id > 0', 'policy_id');

  my ($transaction_start_time) = $dbh->selectrow_array('SELECT NOW()');

  for my $active_policy (values %$active_policies) {
    my $active_policy_id = $active_policy->{id};

    next if ($active_policy_id <= 0);
    $logger->log(msg => "Restoring active policy - policy $active_policy->{name} (id = $active_policy_id)",
                 log_level => 'debug',
                 syslog => 1) if $logger;
    my $policy = F5::Policy->new(
			     policy_id => $active_policy_id,
			     dbh => $dbh,
			    );
    $external_idmap->{__private_save}->{force_whitehat} = defined($whitehat_policies->{$active_policy_id});
    $rv = $policy->set_active_from_history(ucs_version => $self->{ucs_version},
                                           logger => $update_core_logger,
                                           use_daemon => $use_config_server_for_set_active,
                                           transaction_start_time => $transaction_start_time,
                                           external_idmap => $external_idmap,
                                           base_dir => $self->{base_dir},
                                           asm_config_handler => $self->{asm_config_handler},
                                           has_parent => $child_policies->{$active_policy_id},
                                          );
    if (!$rv) {
      $logger->error("Restoring active policy - policy $active_policy->{name} (id = $active_policy_id)") if $logger;
    }
  }

  return $rv;
}

#----------------------------------------------------------------------

sub update_core {
  my($self, $use_daemon) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  $logger->debug("Updating core") if $logger;
  my $update_core_logger = F5::Logger->get_logger(logger_name => 'update_core');

  if (!$self->{active_policy_sync_only}) {
      my $hsl_profile_ids = $dbh->selectcol_arrayref("SELECT hsl_profile_id FROM PLC.HSL_PROFILES");
      for my $hsl_profile_id (@$hsl_profile_ids)
      {
         F5::SetActive::set_active (
            dbh            => $dbh,
            hsl_profile_id => $hsl_profile_id,
            send_sync      => 0,
            do_update_core => 0,
            do_hbcl => 0,
            use_daemon => $use_daemon,
            asm_config_handler => $self->{asm_config_handler},
         );
      }
  }
  if ( F5::ProcessTable::is_proc_up('dcc') ) {
    $logger->debug("calling F5::SetActive::set_active") if $logger;
    F5::SetActive::set_active(dbh => $dbh, 
                              logger => $logger, 
                              do_update_core => 1, 
                              update_core_only => $self->{active_policy_sync_only},
                              asm_config_handler => $self->{asm_config_handler},);
  } 
  else {
    $update_core_logger->info("DCC isn't up -- not sending update core");
  }

  my $bot_defense = F5::PrepareConf::BotDefense->new(dbh => $dbh,
                                                     logger => $logger,
                                                    );

  $bot_defense->prepare_bot_defense_profiles();

  return 1;
}

#----------------------------------------------------------------------

sub get_account_names {
  my($self) = @_;

  my $accounts = $self->{dbh}->selectall_hashref('SELECT policy_name_crc, policy_name FROM DCC.ACCOUNTS',
					 'policy_name_crc');
  return $accounts;
}

#----------------------------------------------------------------------

sub set_account_name_changed {
  my($self, $orig_account_names) = @_;

  my $dbh = $self->{dbh};
  my $update_time = time;
  my $new_account_names = $self->get_account_names;
  for my $policy_name_crc (keys %$new_account_names) {
    if ($orig_account_names->{$policy_name_crc} &&
	$orig_account_names->{$policy_name_crc}->{policy_name} ne $new_account_names->{$policy_name_crc}->{policy_name}) {
      $dbh->do('UPDATE DCC.ACCOUNTS SET policy_changed_time = ? WHERE policy_name_crc = ?',
	       undef, $update_time, $policy_name_crc);
    }
  }
}

#----------------------------------------------------------------------

sub remove_database {
  my($self, $dbname) = @_;

  # explicitly check if the database exists
  my ($schema_name_exists) = $self->{dbh}->selectrow_array("SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = ?", undef, $dbname);
  if (!$schema_name_exists) {
      return;
  }

  $self->{dbh}->do("DROP DATABASE $dbname");
}

#----------------------------------------------------------------------

sub create_database {
  my($self, $dbname) = @_;
  $self->{dbh}->do("CREATE DATABASE $dbname");
}

#----------------------------------------------------------------------

sub get_dump_file_dir_from_pointer {
  my($self) = @_;

  if (! -e $self->ucs_v3_pointer_file) {
      return;
  }
  my $ucs_v3_file_dir = F5::Utils::File::file_slurp($self->ucs_v3_pointer_file);
  chomp($ucs_v3_file_dir);

  return "$self->{base_dir}/$ucs_v3_file_dir";
}

#----------------------------------------------------------------------

sub ucs_v3_pointer_file {
  my($self) = @_;
  return "$self->{base_dir}/ts_db.save_dir_location.cstmp";
}

#----------------------------------------------------------------------

sub ucs_cmi_metadata_file {
  my ($self) = @_;
  return "$self->{base_dir}/ts_db.cmi_metadata.cstmp";
}
#----------------------------------------------------------------------

sub ucs_policy_versions_file {
  my ($self) = @_;
  return "$self->{base_dir}/ts_db.policy_versions.cstmp";
}

#----------------------------------------------------------------------

sub ucs_version_file {
  my($self) = @_;
  return "$self->{base_dir}/ts_db.ucs_version.cstmp";
}

#----------------------------------------------------------------------
sub save_to_file {
  my($self, %params) = @_;

  $self->save;
  return $self->do_conf_tar(filename=>$params{filename}, flg_cleanup=>1);
}
#----------------------------------------------------------------------
sub do_conf_tar{
  my($self, %params) = @_;
  my $logger = $self->{logger};
  my @account_data_dirs2change = @account_data_dirs;
  my @policy_version_files;
  my @not_included_policy_version_files;

  if ($self->{policy_history}){
      if ($self->{call_flags} && $self->{call_flags}->{policy_versions_hash}) {
          find(sub {
                   my $filename = $File::Find::name;
                   return unless -f $filename;
                   if ($self->{call_flags}->{policy_versions_hash}->{$filename}) {
                       open(FILE, $filename);
                       binmode(FILE);
                       my $md5sum = Digest::MD5->new->addfile(*FILE)->hexdigest;
                       if ($self->{call_flags}->{policy_versions_hash}->{$filename} ne $md5sum) {
                           push (@policy_version_files, $filename);
                       }
                       else {
                           push(@not_included_policy_version_files, $filename);
                       }
                   }
                   else {
                       push (@policy_version_files, $filename);
                   }
               }, $POLICY_VERSIONS_DIR);
           my $policy_versions =  { policy_versions => [@policy_version_files, @not_included_policy_version_files] };
           # Note: File named "policy_versions" actually contains a list of policy history files.
           # This is not to be confused with policy_version_metadata, which is really for policy versions.
           my $policy_versions_filename = $self->get_dump_file_name("policy_versions");
           YAML::XS::DumpFile($policy_versions_filename, $policy_versions);
      }
      else {
          push (@account_data_dirs2change, $POLICY_VERSIONS_DIR);
      }
  }

  my @tar_files = ($self->ucs_v3_pointer_file,
                   $self->get_dump_file_dir_from_pointer,
                   @account_data_dirs2change,
                   @policy_version_files,
                  );

  for my $f (@tar_files) {
    $f =~ s|^$self->{base_dir}/||;
  }

  open(my $null1, ">", File::Spec->devnull);
  open(my $null2, File::Spec->devnull);
  my $err_fh = gensym;
  my $pid = open3($null1, $null2, $err_fh, 
                  'tar', '-zpc', 
                  '-f', $params{filename}, 
                  '-C', $self->{base_dir},
                  @tar_files,
                 );
  waitpid($pid, 0);
  if ($params{flg_cleanup}){
     $self->cleanup;
  }

  return;
}

#----------------------------------------------------------------------

sub load_from_file {
    my($self, %params) = @_;
    return $self->load_with_account_data(%params)
}

#----------------------------------------------------------------------

sub load_with_account_data {
    my($self, %params) = @_;

    my $ACCOUNT_DATA_DIR_BACKUP_DIR = "/var/ts/var/tmp/account.backup.$$";

    my $logger = $self->{logger};
    my $locker = F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
                                              logger => $logger);
    $locker->lock();

    mkpath($ACCOUNT_DATA_DIR_BACKUP_DIR);
    # Delay cleanup until after account data files are moved
    $self->{cleanup_temporary_files} = 0;
    my @account_data_dirs = (@account_data_dirs, $POLICY_VERSIONS_DIR);
    my $rv;
    eval {
        if ($params{filename}) {
            system('tar', '-zpx', 
                   '-f', $params{filename}, 
                   '-C', $self->{base_dir}
                  );
        }

        my $policy_versions_filename = $self->get_dump_file_name("policy_versions");
        if (-e $policy_versions_filename) {
            my $policy_hash = YAML::XS::LoadFile($policy_versions_filename);
            my %version_files = map { ( $_ => 1 ) } @{$policy_hash->{policy_versions}};

            find({
                  wanted => sub {
                      my $filename = $File::Find::name;
                      return unless -f $filename;

                      my $base_dir_filename = $filename;
                      $base_dir_filename = $self->{base_dir} . "/" . $base_dir_filename;

                      if ($version_files{$filename} && ! -e $base_dir_filename) {
                          # This policy history file is still needed
                          # Create a hard link to the base_dir copy of the same file (most efficient)
                          mkpath(dirname($base_dir_filename));
                          my $rv = link($filename, $base_dir_filename);
                          if (! $rv) {
                              # POLICY_VERSIONS_DIR and base_dir are on different filesystems,
                              # copy instead of hard link
                              copy($filename, $base_dir_filename);
                          }
                      }
                  },
                  no_chdir => 1,
                  follow => 0,
                 },
                 $POLICY_VERSIONS_DIR);
        }

        F5::Utils::File::move_files_recursively(from_dir => $POLICY_VERSIONS_DIR,
                                                to_dir => "$ACCOUNT_DATA_DIR_BACKUP_DIR/$POLICY_VERSIONS_DIR");
        for my $account_data_dir (@account_data_dirs) {
            F5::Utils::File::move_files_recursively(from_dir => "$self->{base_dir}/$account_data_dir",
                                                    to_dir => $account_data_dir);
            rmtree("$self->{base_dir}/$account_data_dir");
        }

        $rv = $self->load(check_asm_ready => $params{check_asm_ready});

        rmtree($ACCOUNT_DATA_DIR_BACKUP_DIR);
    };
    my $err = $@;
    $self->cleanup;
    $locker->unlock();

    if ($err) {
        # Restore original policy versions
        F5::Utils::File::remove_files_only($POLICY_VERSIONS_DIR);
        F5::Utils::File::move_files_recursively(from_dir => "$ACCOUNT_DATA_DIR_BACKUP_DIR/$POLICY_VERSIONS_DIR",
                                                to_dir => $POLICY_VERSIONS_DIR);

        rmtree($ACCOUNT_DATA_DIR_BACKUP_DIR);
        die $err;
    }

    return $rv;
}

#----------------------------------------------------------------------

sub is_ucs_post_install_waiting {
  return (-e "$UCS_POST_INSTALL_PATH/ts_db.ucs_version.cstmp");
}

#----------------------------------------------------------------------

sub move_ucs_files_to_post_install_path {
  my($self) = @_;

  open(my $ucs_version_fh, '>', $self->ucs_version_file);
  print $ucs_version_fh "$self->{ucs_version}\n";
  close $ucs_version_fh;

  find({ no_chdir => 1,
         follow => 1,
         follow_skip => 2,
         wanted => sub {
           my $old_filename = $File::Find::name;

           return unless (-f $old_filename);
           return unless ($old_filename =~ m|^$self->{base_dir}/ts_db.[^/]+.cstmp|);

           my $new_filename = $old_filename;
           $new_filename =~ s|^$self->{base_dir}|$UCS_POST_INSTALL_PATH|;

           unless (-d dirname($new_filename)) {
             mkpath(dirname($new_filename));
             chmod(0755, dirname($new_filename));            
           }
           move($old_filename, $new_filename);
           chmod(0644, $new_filename);
         },
       },
       $self->{base_dir});

  # Copy account_data_dirs (rather than move), in case the files were used for the old (current) config
  my @account_data_dirs = (@account_data_dirs, $POLICY_VERSIONS_DIR);
  for my $account_data_dir (@account_data_dirs) {
      F5::Utils::File::copy_files_recursively(from_dir => $account_data_dir,
                                              to_dir => "$UCS_POST_INSTALL_PATH/$account_data_dir");
  }
}

#----------------------------------------------------------------------

sub is_actual_load_possible {
    my ($self, $check_asm_ready) = @_;
    if (F5::BigipUtils::is_install_mode()) {
        return 0;
    }
    
    if(! F5::Provisioning::was_tsconfig_run()) {
        return 0;
    }
    if($check_asm_ready && (! F5::ASMReady::is_asm_ready())) {
        return 0;
    }
    return 1;
}

#----------------------------------------------------------------------

sub get_dump_file_name {
  my($self, $filename_base) = @_;

  my $dump_file_dir = $self->get_dump_file_dir_from_pointer;
  if ($dump_file_dir) {
      return "$dump_file_dir/$self->{dump_file_prefix}.$filename_base.$self->{dump_file_suffix}";
  }
  else {
      return;
  }
}

#----------------------------------------------------------------------

sub upload_certificates{
  my(%params) = @_;

  my $dbh = $params{dbh};  
  my $logger = $params{logger};

  my $ssl_keys = $dbh->selectcol_arrayref(<<EndQuery, { Slice => {} });
SELECT ssl_cert_id 
  FROM PLC.SSL_CERTIFICATES
EndQuery

  for my $ssl_key( @$ssl_keys ) {
    F5::Certificates::upload_certificate(ssl_cert_id => $ssl_key, logger => $logger);
  }

  return;
}

#----------------------------------------------------------------------

sub build_policy_versions_hash {
    my $policy_versions_hash;
    find({wanted => sub {
                         my $filename = $File::Find::name;
                         return unless -f $filename;
                         open(FILE, $filename);
                         binmode(FILE);
                         $policy_versions_hash->{$filename} = Digest::MD5->new->addfile(*FILE)->hexdigest;
                         close(FILE);
                        },}, $POLICY_VERSIONS_DIR);

    return $policy_versions_hash;
}

#--------------------------------------------------------------------------------

sub copy_schemas_to_var {
    find({
          wanted => sub {
              my $orig_filename = $File::Find::name;
              if (-f $orig_filename && $orig_filename =~ /\.xsd$/) {
                  if (basename($orig_filename) eq 'common.xsd') {
                      F5::NegativeSignatures::SigFile->update_common_schema(new_schema_filename => $orig_filename);
                  }
                  else {
                      my $new_filename = $orig_filename;
                      my $source_dir = F5::Deploy::map_path("/ts/share");
                      my $target_dir = F5::Deploy::map_path("/ts/var/schema");
                      $new_filename =~ s|^$source_dir|$target_dir|;
                      # Always overwrite /ts/var/schema with authoritative copy
                      # (in case it was changed by dev updates or hotfixes)
                      copy($orig_filename, $new_filename);
                  }
              }
          },
          no_chdir => 1,
         },
         F5::Deploy::map_path('/ts/share'));
}

#--------------------------------------------------------------------------------

1;
