package F5::CredentialStuffingFile;

use strict;
use warnings;

use Digest::MD5;
use File::Copy;
use File::Path;
use File::Find;
use YAML::XS;
use F5::DbUtils;
use F5::SystemUtils;

my $TEMP_BASE_DIR = "/ts/var/tmp";
my $GENESIS_DIR = "/var/lib/hsqldb/live-update/update-files/";
my $GENESIS_BLOB_PATH = "/var/ts/bloom/genesis_bloom_credentials_blob.dat";
my $BSDTAR = "bsdtar";
my $ENC_FILE_PATH = "RPMS/packages/current_update_file.enc";
my $GPG_HOME_DIR = "/ts/share/negsig/gpg_asm_sigfile_installer";
my $PAYLOAD_FILE_NAME = "payload.tar.gz";
my $PAYLOAD_DIR = "PAYLOAD";
my $GPG = "gpg";
my $TAR = "tar";
my $PYTHON = "/bin/python";
my $LIVE_UPDATE_REPORT_GENESIS_STATUS = "/var/lib/hsqldb/live-update/live_update_mark_active_installation.py";

my $ACCOUNT_DATA_DIR = "/ts/var/account";
my $ACCOUNT_CREDENTIALS_BASE_PATH = "/ts/var/account/credentials/credentials_";

#----------------------------------------------------------------------------------------

sub extract_from_genesis {
    my $directory = "$TEMP_BASE_DIR/credential_stuffing.tmp.$$";
    mkpath $directory;
    my $genesis_filename;
    eval {
        my $live_update_genesis_yaml = YAML::XS::LoadFile('/etc/live-update-genesis.yaml');
        $genesis_filename = $live_update_genesis_yaml->{'Credential Stuffing'}->{'genesisFileName'};
        my $filename = $GENESIS_DIR . $genesis_filename;
        my $sys_rv = F5::SystemUtils::run_system_tool(command_line_params => [$BSDTAR, '--directory', $directory, '-xf', $filename, $ENC_FILE_PATH],
                                                      find_command_path => 1);
        if (!$sys_rv->{completed_successfully}) {
            die $sys_rv->{standard_error};
        }
        my $decrypted_filename = "$directory/$PAYLOAD_FILE_NAME";
        $sys_rv = F5::SystemUtils::run_system_tool(command_line_params => [$GPG, '--always-trust', "--yes", '--homedir', $GPG_HOME_DIR, 
                                                                           "--output", $decrypted_filename, "--decrypt", "$directory/$ENC_FILE_PATH"],
                                                   find_command_path => 1);
        if (!$sys_rv->{completed_successfully}) {
            die $sys_rv->{standard_error};
        }
        my $payload_directory = "$directory/$PAYLOAD_DIR";
        mkdir $payload_directory;
        $sys_rv = F5::SystemUtils::run_system_tool(command_line_params => [$TAR,'-zxf', $decrypted_filename, "-C", $payload_directory],
                                                   find_command_path => 1);
        if (!$sys_rv->{completed_successfully}) {
            die $sys_rv->{standard_error};
        }

        find({ wanted => sub {
                             my $blob_filename = $File::Find::name;
                             if(-f $blob_filename) {
                                 copy($blob_filename, $GENESIS_BLOB_PATH) or die "Failed to copy blob file: $!";
			     }
                         }
             }, "$payload_directory/credentials");
    };
    rmtree($directory);
    if ($@) {
        die "Could not extract credential stuffing from genesis: $@\n";
    }
    return $genesis_filename;
}

#----------------------------------------------------------------------------------------

sub load_from_genesis {
    my (%params) = @_;
    my $dbh = $params{dbh};

    my $genesis_filename = extract_from_genesis();
    if(defined $genesis_filename && $genesis_filename =~ /CredentialStuffing_(\d{8}_\d{6}).im/) {
        my $create_time = $1;
        load_credential_stuffing_file(filename => $GENESIS_BLOB_PATH,
                                      dbh => $dbh,
                                      copy_to_account_dir => 0);

	# Everthing went fine. Report installation status to live-update.
	system($PYTHON, $LIVE_UPDATE_REPORT_GENESIS_STATUS, "--component", "credential-stuffing", "--create-time", $create_time);
    }
    else {
	die "Failed to calculate create time from credential stuffing genesis file name";
    }
}

#----------------------------------------------------------------------------------------

sub load_credential_stuffing_file {
    my (%params) = @_;

    my $dbh = $params{dbh};
    my $logger = $params{logger} || F5::Logger->new('/ts/log/credential_stuffing_file.log');
    my $filename = $params{filename};
    my $copy_to_account_dir = defined $params{copy_to_account_dir} ? $params{copy_to_account_dir} : 1;

    my $table_name = 'PLC.CREDENTIAL_STUFFING_UPDATE_FILE';
    my $credential_filename = $filename;

    if($copy_to_account_dir) {
        open(my $fh, $filename);
        my $md5 = Digest::MD5->new->addfile($fh)->hexdigest;
        close $fh;

        $credential_filename = $ACCOUNT_CREDENTIALS_BASE_PATH . $md5;

        copy($filename, $credential_filename) or die "Failed to copy credential stuffing file: $!";

    }

    F5::DbUtils::delete_rows(dbh => $dbh,
                             where_fields => { 1 => 1 },
                             table_name => $table_name);

    my $rec = { credential_filename => $credential_filename };

    F5::DbUtils::insert_row(dbh => $dbh,
                            table_name => $table_name,
                            rec => $rec);

    if($copy_to_account_dir) {
        F5::DbUtils::delete_unreferenced_files(dbh          => $dbh,
                                               logger       => $logger,
                                               db_file_refs => [
                                                                    { table_name => $table_name,
                                                                      field     => 'credential_filename',
                                                                    },
                                                               ],
                                               dir          => "$ACCOUNT_DATA_DIR/credentials");
    }
}

#----------------------------------------------------------------------------------------

1;
