package F5::DPI::Sigfile::Update;

use strict;
use warnings;

use base qw(F5::Sigfile::Update);

use Time::Local;
use File::Basename;
use File::Path;
use File::Copy;
use File::Find;
use Data::Dumper; # for dumping hashes
use F5::Logger;
use F5::DPI::Sigfile::AutoDownload;
use F5::BigipVersionUtils;
use F5::ASM::COAPI;
use F5::ASM::Tmsh;
use F5::DateTimeUtils;
use File::Temp qw/tempfile tempdir/;

my %AUTO_UPDATE_INTERVAL_MAP = ( 1 => "daily",
                                 2 => "weekly",
                                 3 => "monthly",);

my %STATUS_MAP = ( NONE => 0,
                   SUCCESS => 1,
                   FAILURE => 2,
                   IN_PROGRESS => 3,);

my $DPI = "/var/libdata/dpi";
my $CONF = "$DPI/conf/classification_update.conf";

#----------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;
    $self->SUPER::init(%params);
    $self->{sigfile_type} = "DPI Signatures";
    # Location to put the downloaded IMs. This is not /var/libdata/dpi/im,
    # as that folder is csynced between blades only after IM is validated.
    $self->{save_dir} = $DPI;
    $self->{tempdir} = "";
    $self->{update_installed} = 0;

    # These dirs contain everything related to Hitless Upgrade. They should be
    # created in tmm.daemontools. However, they're in a rw-folder,
    # so we cannot rely on their existence and we should create them just in case.
    system("mkdir -p $DPI/{im,conf,reset}");
    system("touch $CONF");
    # Cleanup IM files from previous upgrade
    system("rm -f $DPI/im/*.im");
    $self->{lock_file} = '/ts/lock/dpi_signatures.lock';
}

#----------------------------------------------------------------------
sub get_schedule_info {
    my ($self, %params) = @_;
    my @schedules;
    my $query = F5::ASM::COAPI::Query->new("classification_signatures_update_scheduling");
    if ($query && @$query) {
        my $status_query = F5::ASM::COAPI::Query->new("classification_signatures_update_status");
        for my $schedule (@$query) {
            my $rv;
            $rv->{auto_update_interval} = $AUTO_UPDATE_INTERVAL_MAP{$schedule->{auto_update_interval}};
            $rv->{update_mode} = $schedule->{update_mode_automatic} ? "auto-interval"
                                                                    : "auto-user";
            $rv->{last_update_attempt} = 0;
            if ($status_query && @$status_query) {
                my ($update_status) = @$status_query;
                if ($update_status) {
                    $rv->{last_update_attempt} = $update_status->{last_attempt_datetime};
                }
            }
            push (@schedules, $rv);
        }
    }

    return @schedules;
}

sub retval {
    my($result, %params) = @_;
    my $rv = {completed_successfully => $result->{completed_successfully},
        msg => $result->{standard_error}, };
    return $rv;
}

#----------------------------------------------------------------------
sub apply_changes {
  my($self, %params) = @_;

  # Sigfile::Update::update() doesn't call process_im() if IM type is not ISO.
  eval {
      %params = $self->process_im(%params);
  };
  die($@) if $@;

  my $out=qx(/usr/bin/classification_upgrade_blade $params{filename} validate 2>&1);
  return { msg => $out } if $?;

  my $mcp_query = F5::ASM::COAPI::Query->new("classification_signature_version");
  ($mcp_query and @$mcp_query) or
      die("Could not open classification_signature_version for writing, aborting upgrade.");

  my ($sig_ver) = @$mcp_query;
  $self->{logger}->debug("Saving update_time: $params{update_time}");
  $sig_ver->{update_time} = $params{update_time};
  $sig_ver->{version} = $params{datever};
  $sig_ver->{cec_filename} = "libcec.so";
  $sig_ver->{cec_version} = $params{cec_version};
  $sig_ver->{im_version} = $params{datever_file};
  $sig_ver->save;

  # #487849: Upgrade via GUI returns a filename in the form "upload_xxxxx.tmp",
  # and *.tmp files are ignored by csyncd, hence we need to rename the IM.
  # #525639: Randomize the IM name to make it unique and prevent second csyncd
  # post trigger (csyncd bug) on the other blades.
  my $temp_suffix = qx(mktemp --dry-run XXXXXX);
  chomp($temp_suffix);
  my $synced_im = "$DPI/im/classification-update-$params{bigip}-$params{datever}-$temp_suffix.im";

  # Prevent second trigger on this blade by creating a temp file in synced dir, prior to copying
  qx(touch "$synced_im.tmp");

  # Sync IM to other devices/blades in device-group to proceed there.
  copy($params{filename}, $synced_im);
  $out=qx(/usr/bin/classification_sync_im.py $synced_im 2>&1);
  $self->{logger}->debug("Sync_im output: $out");

  # Cleanup file artifacts
  my @file_artifacts = ("$synced_im.tmp", "$params{filename}", "$params{filename}.tmp");
  my $number = unlink (@file_artifacts);
  $self->{logger}->debug("Unlinked file artifacts : $number");

  $self->{update_installed} = 1;
  return { completed_successfully => 1 };
}


#----------------------------------------------------------------------

sub start_update {
    my ($self, %params) = @_;
    $self->report_status();
}

#----------------------------------------------------------------------
sub report_status {
  my($self, %params) = @_;
  my $now = time();
  my $status;
  my $query = F5::ASM::COAPI::Query->new("classification_signatures_update_status");
  if ($query && @$query) {
      ($status) = @$query;
  }
  else {
      $status = F5::ASM::COAPI::Object->new("classification_signatures_update_status");
      $status->{name} = "/Common/status";
  }
  $status->{last_attempt_datetime} = $now;
  $status->{progress_status} = $STATUS_MAP{IN_PROGRESS};
  $status->{message} = $params{msg} || "";
  $status->{last_attempt_user} = $self->{username};
  $status->{last_attempt_automatic_mode} = $self->{check_interval};

  if (defined $params{completed_successfully} && $params{completed_successfully}) {
      $status->{progress_status} = $STATUS_MAP{SUCCESS};
      if ($self->{update_installed}) {
          $status->{last_update_user} = $self->{username};
          $status->{last_update_automatic_mode} = $self->{check_interval};
          $status->{last_update_datetime} = $now;
      }
  } else {
      $status->{progress_status} = $STATUS_MAP{FAILURE};
  }
  $status->save();

  if ($self->{update_installed}) {
      # Save the config only on successful upgrade. Don't bother saving
      # failed/not started statuses because it does more harm (#497741)
      # than good.
      my @tmsh_cmd = ("tmsh", "save", "sys", "config");
      F5::ASM::Tmsh::Query->new(cmd => \@tmsh_cmd, logger => $self->{logger});
  }

  #if !successfull write the error message to syslog? ( $params{msg} )
}

#----------------------------------------------------------------------

sub get_autodownload {
    my ($self, %params) = @_;
    return F5::DPI::Sigfile::AutoDownload->new(logger => $self->{logger},
            outbound_call_timeout => $self->{outbound_call_timeout},
            );
}

sub process_im {
    my ($self, %params) = @_;

    my $datever_file = qx(tar -tf $params{filename} | grep classification-update);
    my $cec_version = qx(tar -tf $params{filename} | grep libcec.so | sed 's/libcec.so.//');
    $self->{logger}->debug("Update file: $params{filename}, datever file: $datever_file");
    $datever_file =~ /classification-update-(.*)-(.*?)\./ or
        die(qq/IM filename does not match the required pattern, aborting upgrade./);

    my ($new_bigip, $new_datever) = ($1, $2);
    my $this_bigip = F5::BigipVersionUtils::bigip_version();
    $self->{logger}->debug("BigIP version: current: $this_bigip, new: $new_bigip");
    $self->{logger}->debug("Datever: $new_datever");

    $new_bigip eq $this_bigip or
        die("Package version is incompatible with this release, expected $this_bigip, got $new_bigip, aborting upgrade.");

    $new_datever =~ /(\d\d\d\d)(\d\d)(\d\d)_(\d\d)(\d\d)/ or
        die(qq/Cannot parse datever from the RPM name/);
    my ($year, $month, $day, $hour, $minute) = ($1, $2, $3, $4, $5);
    $params{update_time} = F5::DateTimeUtils::convert_datetime_to_epoch(
        "$year-$month-$day $hour:$minute:00");
    $self->{logger}->debug("Process_im: $params{update_time}");

    my $mcp_query = F5::ASM::COAPI::Query->new("classification_signature_version");
    ($mcp_query and @$mcp_query) or
        die("Could not read classification_signature_version, bad configuration? aborting upgrade.");

    my ($sig_ver) = @$mcp_query;
    $self->{logger}->debug(Dumper($sig_ver));

    # If installed datever is empty, it's a vanilla box and we proceed anyway.
    if ($sig_ver->{version} && $sig_ver->{version} eq $new_datever) {
        die("Cannot install the same package (version $new_datever), aborting upgrade.");
    }

    # Remove redundant new line char
    $datever_file =~ s/[\n\r]//g;
    $cec_version =~ s/[\n\r]//g;

    $params{bigip}  = $this_bigip;
    $params{datever}  = $new_datever;
    $params{datever_file} = $datever_file;
    $params{cec_version} = $cec_version;

    return %params;
}

sub DESTROY {
    my $self = shift;

    my %error_code = (
      0 => "Failure",
      1 => "Success",
    );

    # print STDERR "DPI.Sigfile.DESTROY: Update installed:$self->{update_installed}";
    $self->{logger}->info("DPI.Sigfile.DESTROY: Update installed:$error_code{$self->{update_installed}}");

    # Give suggestions for new commands to user in case of failure.
    my $error_help = "\nClassification hitless upgrade commands have been updated. Please check the following manual entries for details.\n" .
                     "tmsh\$ help ltm classification updates\n" .
                     "tmsh\$ help ltm classification auto-update\n\n";
    $self->{logger}->info($error_help);
}

#----------------------------------------------------------------------

# This function is run by classification_reset_upgrade script. It basically
# deletes the signature-version block, but without allowing doing the same
# from tmsh and avoiding schema changes.
sub reset_signature_version {
    my $mcp_query = F5::ASM::COAPI::Query->new("classification_signature_version");
    ($mcp_query and @$mcp_query) or
        die("Could not open classification_signature_version for writing.");
    my ($sig_ver) = @$mcp_query;
    $sig_ver->{update_time} = "";
    $sig_ver->{version} = "";
    $sig_ver->{cec_filename} = "libcec.so";
    $sig_ver->{cec_version} = "default";
    $sig_ver->{im_version} = "default";
    $sig_ver->save;
}

1;
