package F5::DateTimeUtils;

use strict;
use warnings;

use Time::Local;
use Carp;

sub convert_epoch_to_datetime {
    my ($epoch, $server_time) = @_;
    return unless defined($epoch);
    if (!defined($server_time)) {
        $server_time = 1;
    }

    my @date = $server_time ? localtime($epoch)
                            : gmtime($epoch);
    $date[5]+= 1900;
    $date[4]++;
    splice(@date, 6);
    return sprintf("%0.4d-%0.2d-%0.2d %0.2d:%0.2d:%0.2d", reverse @date);
}

#----------------------------------------------------------------------

sub convert_datetime_to_epoch_in_ms {
    return convert_datetime_to_epoch(@_) . '000';
}

#----------------------------------------------------------------------

sub convert_datetime_to_epoch {
    my ($datetime, $server_time) = @_;
    return unless defined($datetime);

    if (my @date_parts = ($datetime =~ m|^(\d{4})[-/]?(\d\d)[-/]?(\d\d)[ _](\d\d):?(\d\d):?(\d\d)|)) {
        $date_parts[1]--;
        return ($server_time) ? timegm(reverse @date_parts) : timelocal(reverse @date_parts) ;
    }
    return $datetime;
}

#----------------------------------------------------------------------

sub convert_datetime_to_file_timestamp {
    my ($datetime) = @_;
    return unless defined($datetime);

    if ($datetime =~ m|^(\d{4})[-/](\d\d)[-/](\d\d) (\d\d):(\d\d):(\d\d)|) {
        $datetime = $1 . $2 . $3 . '_' . $4 . $5 . $6;
    }

    return $datetime;
}

#----------------------------------------------------------------------

sub convert_epoch_to_file_timestamp {
    my $datetime = convert_epoch_to_datetime(@_);
    return convert_datetime_to_file_timestamp($datetime);
}

#----------------------------------------------------------------------

sub convert_datetime_to_rpm_timestamp {
    my ($datetime) = @_;
    return unless defined($datetime);

    if ($datetime =~ m|^(\d{4})[-/](\d\d)[-/](\d\d) (\d\d):(\d\d):(\d\d)|) {
        $datetime = $1 . "." . $2 . "." . $3;
    }
    return $datetime;
}

#----------------------------------------------------------------------

sub convert_epoch_to_rfc3339 {
    my ($epoch, $server_time) = @_;
    return unless defined($epoch);

    my $fractional;
    if ($epoch != int($epoch)) {
        ($epoch, $fractional) = split(/\./, $epoch);
    }

    my @date = $server_time ? localtime($epoch)
                            : gmtime($epoch);
    $date[5]+= 1900;
    $date[4]++;
    splice(@date, 6);
    my $rfc3339_time = sprintf("%0.4d-%0.2d-%0.2dT%0.2d:%0.2d:%0.2d", reverse @date);

    if ($fractional) {
        $rfc3339_time .= ".$fractional";
    }

    if ($server_time) {
        my @gmdate = gmtime($epoch);
        my $tz_hours = $date[2] - $gmdate[2];
        my $tz_mins = $date[1] - $gmdate[1];
        my $tz_diff = ($tz_hours * 60) + ($tz_mins);
        my $shift_hour = ($tz_hours * $tz_mins) < 0 ? 1 : 0;

        $rfc3339_time .= ($tz_diff < 0 ? "-" : "+") . sprintf("%0.2d:%0.2d", (abs($tz_hours) - $shift_hour), abs($tz_mins));
    }
    else {
        $rfc3339_time .= "Z";
    }
    return $rfc3339_time;

}

#----------------------------------------------------------------------

sub convert_rfc3339_to_epoch {
    my ($datetime) = @_;
    return unless defined($datetime);
    #From DateTime::Format::RFC3339
    $datetime = uc($datetime);

    my ($Y,$M,$D) = $datetime =~ s/^(\d{4})-(\d{2})-(\d{2})// && (0+$1,0+$2,0+$3)
       or croak("Incorrectly formatted date in '$datetime'");
    $M--;

    $datetime =~ s/^T//
      or croak("Incorrectly formatted datetime in '$datetime'");

    my ($h,$m,$s) = $datetime =~ s/^(\d{2}):(\d{2}):(\d{2})// && (0+$1,0+$2,0+$3)
       or croak("Incorrectly formatted time in '$datetime'");

    my ($tz_type, $tz_hour, $tz_min);
    if    ( $datetime =~ s/^Z//                     ) { $tz_type = '+', $tz_hour = 0; $tz_min = 0;    }
    elsif ( $datetime =~ s/^([+-])(\d{2}):(\d{2})// ) { ($tz_type, $tz_hour, $tz_min) = ($1, $2, $3); }
    else { croak("Missing time zone in '$datetime'"); }

    $datetime =~ /^\z/ or croak("Incorrectly formatted datetime in '$datetime'");

    if ($tz_type eq "-") {
        $tz_hour *= -1;
        $tz_min *= -1;
    }
    $h += $tz_hour;
    $m += $tz_min;
    my $time = timegm($s, $m, $h, $D, $M, ($Y - 1900));

    return $time;
}

#----------------------------------------------------------------------

sub convert_any_datetime_to_epoch {
    my($datetime) = @_;

    my $epoch;
    # Try rfc3339 first, since it's stricter
    eval {
        $epoch = convert_rfc3339_to_epoch($datetime);
    };
    if ($@ || ! $epoch) {
        $epoch = convert_datetime_to_epoch($datetime);
    }

    return $epoch;
}

#----------------------------------------------------------------------

sub get_time_until_maximum_interval {
    my (%params) = @_;
    my $start_time  = $params{start_time} || 0;
    my $current_time = $params{current_time} || time;
    my $minimum_interval = $params{minimum_interval} || 0;
    my $retval = $start_time + $params{maximum_interval} - $current_time;
    if ($retval > $params{maximum_interval} || $retval < $minimum_interval) {
        # If retval is longer than maximum (due to clock change) or expired
        $retval = $minimum_interval;
    }
    return $retval;
}

#----------------------------------------------------------------------

sub get_time_until_hour_minute {
    my (%params) = @_;

    my $minute_on_every_hour = $params{minute_on_every_hour};
    my $start_time = $params{start_time};
    my $current_time = $params{current_time} || time;

    my $time_since_last_run = $current_time - $start_time;

    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($current_time);
    my $run_interval = $minute_on_every_hour*60 - $min*60;

    #check if there was run during last minute
    if ($time_since_last_run <= 60) {
        $run_interval = 3600;
    }
    elsif ($min > $minute_on_every_hour) {
        $run_interval += 3600;
    }

    return $run_interval;
}

#----------------------------------------------------------------------

sub get_time_with_milliseconds {
    my($seconds, $microseconds) = Time::HiRes::gettimeofday();
    my $milliseconds = int($microseconds / 1000);
    return $seconds + ($milliseconds / 1000);
}

#----------------------------------------------------------------------

1;
