package F5::DbClean;

use strict;
use warnings;

use YAML::XS;
use List::Util qw(first);
use F5::Db::Single;
use F5::DateTimeUtils;

our $DEFAULT_PULSE_MSG_INTERVAL = 30;
our $WARN_ROW_LIMIT_EXCEEDED_BY_FACTOR = 2.5;

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = bless {}, $class;

  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{dbh} = $params{dbh};
  $self->{logger} = $params{logger};
  $self->{table_state} = {};

  $self->load_table_def(config_file => $params{config_file});

  $self->{after_clean} = $params{after_clean};
  $self->{pulse_msg_interval} = $self->{config}->{pulse_msg_interval} || $DEFAULT_PULSE_MSG_INTERVAL;
  $self->{tables_cleaned_count} = 0;
  $self->{tables_checked_count} = 0;
  $self->{last_table_cleaned} = undef;
  $self->{rows_cleaned_count} = 0;
}

#----------------------------------------------------------------------

sub load_table_def {
    my ($self, %params) = @_;
    my $config = YAML::XS::LoadFile($params{config_file});
    $self->{config} = $config;

    for my $table_def (values %{$self->{config}->{tables}}) {
        for my $k (keys %{$self->{config}->{default}}) {
            if (! exists $table_def->{$k}) {
                $table_def->{$k} = $self->{config}->{default}->{$k};
            }
        }
    }

    return $config;
}

#----------------------------------------------------------------------

sub set_dbh {
  my ($self, $dbh) = @_;
  $self->{dbh} = $dbh;
}

#----------------------------------------------------------------------

sub inserted_rows {
    my($self, %params) = @_;
    return $self->clean_table(%params,
                              inserted_row_count => $params{row_count},
                             );
}

#----------------------------------------------------------------------

sub clean_table {
    my($self, %params) = @_;

    my $table_name = $params{table_name};
    my $dbh = $self->{dbh} || F5::Db::Single->instance;
    my $delete_count_result = 0;
    my $delete_count;
    my @where_clauses;
    my @bind_values;
    my $where_clause = '';
    my $has_exceeded_limit = 0;
    my $row_limit_for_reduce;

    my $table_def = $self->{config}->{tables}->{$table_name};
    return 0 unless $table_def;

    if (! $self->{table_state}->{$table_name}) {
        $self->{table_state}->{$table_name} = { row_count => 0 };
    }
    my $table_state = $self->{table_state}->{$table_name};

    if (exists $table_def->{do_cleaning} && $table_def->{do_cleaning} == 0) {
        return;
    }
    $self->{tables_checked_count}++;

    if ($table_def->{cache_row_count} && defined($table_state->{row_count}) && $params{inserted_row_count}) {
        $table_state->{row_count} += $params{inserted_row_count};
        $self->{logger}->trace("Table $table_name has approximately $table_state->{row_count} rows (tracked by cache)") if $self->{logger};
    }

    if (! defined($table_state->{row_count}) ||
        ! $table_def->{cache_row_count} ||
        $table_state->{row_count} > $table_def->{row_limit}) {

        if ($table_def->{where_values}) {
            for my $k (keys %{$table_def->{where_values}}) {
                if (ref $table_def->{where_values}->{$k}) {
                    my @values = @{$table_def->{where_values}->{$k}};
                    my $placeholders = join(',', ('?') x scalar @values); 
                    push(@where_clauses, "$k IN ($placeholders)");
                    push(@bind_values, @values);
                }
                else {
                    push(@where_clauses, "$k = ?");
                    push(@bind_values, $table_def->{where_values}->{$k});
                }
            }
        }
        if ($table_def->{protected_row_subquery}) {
            my $protected_ids = $dbh->selectcol_arrayref($table_def->{protected_row_subquery});
            if ($protected_ids && @$protected_ids) {
                my $ids = join(",", @$protected_ids);
                push(@where_clauses,
                     "$table_def->{order_by} NOT IN ($ids)");
            }
        }

        if (@where_clauses) {
            $where_clause = 'WHERE ' . join(' AND ', @where_clauses);
        }

        # Confirm row count from database
        $table_state->{row_count} = $self->get_row_count(table_name => $table_name,
                                                         table_def => $table_def,
                                                         where_clause => $where_clause,
                                                         bind_values => \@bind_values,
                                                         dbh => $dbh,
                                                        );

        if ( !defined $table_state->{row_count} ) {
            $table_state->{row_count} = 0;
        }

        $self->{logger}->trace("Table $table_name currently has $table_state->{row_count} rows (row limit = $table_def->{row_limit})") if $self->{logger};

        if ( $table_state->{row_count} > $table_def->{row_limit} ) {
            $has_exceeded_limit = 1;
            $row_limit_for_reduce = $table_def->{row_limit};
            $self->{logger}->debug("Table $table_name has exceeded its limit of $table_def->{row_limit} rows and currently has $table_state->{row_count} rows.") if $self->{logger};
            if ($table_state->{row_count} > ($table_def->{row_limit} * $WARN_ROW_LIMIT_EXCEEDED_BY_FACTOR)) {
                $self->{logger}->log(msg => "Table $table_name has exceeded its limit of $table_def->{row_limit} and currently has $table_state->{row_count} rows",
                                     log_level => 'info',
                                     syslog => 1) if $self->{logger};
            }
        }

    }

    my $table_size;

    if ( $table_def->{size_limit_in_mb} ) {
        $table_size =  $self->get_table_size( dbh => $dbh, table_name => $table_name );
        $self->{logger}->trace("Table $table_name currently has $table_size MB (size limit = $table_def->{size_limit_in_mb} MB)") if $self->{logger};

        if ( $table_size > $table_def->{size_limit_in_mb} ) {
            $has_exceeded_limit = 1;
            $row_limit_for_reduce = $table_state->{row_count};
            $self->{logger}->debug("Table $table_name has exceeded its limit of $table_def->{size_limit} MB and currently has $table_size MB.") if $self->{logger};
        }
    }

    if ($has_exceeded_limit) {

        $delete_count = int($row_limit_for_reduce * ($table_def->{reduce_by_pct} / 100) );

        $self->{logger}->debug("Attempting to delete $delete_count rows") if $self->{logger};

        if (ref($table_def->{order_by})) {
            my %table_fields = map { ($_ => 1) } F5::DbUtils::get_db_fields($dbh, $table_name);
            $table_def->{order_by} = first { $table_fields{$_} } @{$table_def->{order_by}};
        }

        my $order_clause = '';
        if ($table_def->{order_by}) {
            $order_clause = "ORDER BY $table_def->{order_by}";
        }

        $delete_count_result = $dbh->do("DELETE FROM $table_name $where_clause $order_clause LIMIT $delete_count",
                                        undef, @bind_values);
        $self->{logger}->debug("$delete_count_result rows were cleaned from $table_name") if $self->{logger};
        $self->{tables_cleaned_count}++;
        $self->{rows_cleaned_count} += $delete_count_result;
        $self->{last_table_cleaned} = $table_name;
        if ($table_def->{cache_row_count}) {
            $table_state->{row_count} -= $delete_count_result;
        }

        if ($table_def->{clean_orphans}) {
            for my $orphan_table (@{$table_def->{clean_orphans}}) {
                $self->{logger}->debug("Cleaning orphaned records from $orphan_table->{table_name}") if $self->{logger};
                my $orphan_delete_count_result = $dbh->do(<<EndQuery);
DELETE $orphan_table->{table_name} orphan_table
  FROM $orphan_table->{table_name} orphan_table
  LEFT JOIN $table_name main_table
         ON (orphan_table.$orphan_table->{foreign_key} = main_table.$orphan_table->{foreign_key})
 WHERE main_table.$orphan_table->{foreign_key} IS NULL
EndQuery

                $self->{logger}->debug("$orphan_delete_count_result rows were cleaned from $orphan_table->{table_name}") if $self->{logger}; 
            }
        }

        if ($self->{after_clean}) {
            $self->{after_clean}->(table_name => $table_name);
        }
    }

    if (! $params{additional_limit_call} && $table_def->{additional_limits}) {
        for my $additional_limit (@{$table_def->{additional_limits}}) {
            $delete_count_result += $self->clean_table(%params,
                                                       additional_limit_call => 1,
                                                       %$additional_limit);
        }
    }

    $table_state->{last_run_time} = time;
    return $delete_count_result;
}

#----------------------------------------------------------------------

sub get_row_count {
    my($self, %params) = @_;

    my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;
    my $table_name = $params{table_name};
    my $table_def = $params{table_def};
    my $row_count;

    if ($table_def->{get_row_count_using_table_status} && ! $params{where_clause}) {
        my($db_name, $table_name_only) = split(/\./, $table_name);
        my $row = $dbh->selectrow_hashref("SHOW TABLE STATUS from $db_name LIKE ?", undef, $table_name_only);
        if ($row && defined $row->{Rows}) {
            ($row_count) = $row->{Rows};
        }
    }
    else {
        ($row_count) = $dbh->selectrow_array("SELECT count(*) FROM $table_name $params{where_clause}",
                                             undef, @{$params{bind_values}});
    }
    return $row_count;
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;
    my @additional_info;


    if ($self->{tables_cleaned_count}) {
        push(@additional_info, "tables cleaned = $self->{tables_cleaned_count}");
    }

    if ($self->{last_table_cleaned}) {
        push(@additional_info, "last table cleaned = $self->{last_table_cleaned}");
    }

    if ($self->{rows_cleaned_count}) {
        push(@additional_info, "rows cleaned = $self->{rows_cleaned_count}");
    }

    my $additional_info = '';
    if (@additional_info) {
        $additional_info = '(' . join(', ', @additional_info) . ')';
    }
    $self->{logger}->info("$self->{tables_checked_count} table checks performed since startup $additional_info") if $self->{logger};
}

#----------------------------------------------------------------------

sub run {
    my($self, %params) = @_;

    my $last_pulse_time = time;

    while (1) {

        my $poll_time = undef;
        my @potential_poll_times = ($self->{pulse_msg_interval});

        for my $table_name (keys %{$self->{config}->{tables}}) {
            my $table_clean_interval = $self->{config}->{tables}->{$table_name}->{interval_in_sec};
            my $table_last_clean_time = $self->{table_state}->{$table_name}->{last_run_time} || 0;

            my $next_clean_duration = F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $table_clean_interval,
                                                                                         start_time =>  $table_last_clean_time);
            if ($next_clean_duration == 0) {
                $self->{logger}->debug("Checking table $table_name") if $self->{logger};
                $self->clean_table(table_name => $table_name);
                push(@potential_poll_times, $table_clean_interval);
            }
            else {
                $self->{logger}->trace("Not checking table $table_name ($next_clean_duration seconds to next check)") if $self->{logger};
                push(@potential_poll_times, $next_clean_duration);
            }
        }

        if (@potential_poll_times) {
            $poll_time = List::Util::min(@potential_poll_times);
        }
        $self->{logger}->trace("About to sleep for $poll_time seconds") if $self->{logger};
        sleep($poll_time);

        if ($self->{pulse_msg_interval} && 
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                               start_time => $last_pulse_time) == 0) {

            $self->log_pulse_message;
            $last_pulse_time = time;
        }
    }
}

#----------------------------------------------------------------------

sub get_table_size {
    my($self, %params) = @_;
    my $dbh = $params{dbh} || $self->{dbh} || F5::Db::Single->instance;
    my $table_name = $params{table_name};

    my($db_name, $table_name_only) = split(/\./, $table_name);

    my $table_size = $dbh->selectrow_array(<<"EndQuery", undef, $db_name, $table_name_only );
SELECT round(((data_length + index_length) / 1024 / 1024), 2) as size
  FROM information_schema.TABLES
 WHERE table_schema = ?
   AND table_name = ?
EndQuery

    return $table_size;
}

#----------------------------------------------------------------------

1;
