package F5::DbFixup::DataProtection;

use strict;
use warnings;
use base 'F5::DbFixup::Default';
use F5::Utils::SharedMemory;
use F5::ASM::COAPI;
use F5::DeviceGroup;
use File::Basename;


# Package globals, handled by AUTOLOAD
our $fixup_db = 'PLC';
our $max_version = '0011.0004.0000.0000.0000';
our $excluded_tables = [ 'PLC.DATA_PROTECTION', 'PLC.DATA_PROTECTION_CONFIGURATION', 'DCC.DATA_PROTECTION'];

#Retain Data Protection Settings already in the system if loading a UCS from a version without those tables.
sub after_fixup {   
    my ($self, $dbh, $old_dbname, $new_dbname, $logger, $params) = @_;

    my $databases = $params->{databases};
    $logger->debug("Begin Data Protection fixup");

    for my $full_table (@$excluded_tables) {
        # PLC_staging/DCC_staging tables should be empty, but clear them out to be sure
        my($db_type, $table_name) = split(/\./, $full_table);

        $dbh->do(<<EndQuery, undef);
DELETE FROM $databases->{$db_type}->{staging_dbname}.$table_name
EndQuery

        $self->copy_table(dbh => $dbh,
                          old_dbname => $databases->{$db_type}->{bigip_dbname},
                          new_dbname => $databases->{$db_type}->{staging_dbname},
                          table_name => $table_name,
                          id_field_name => '',
                          logger => $logger,);
    }

    ## Check if we are in a Device Group.  
    my $local_device_name = F5::DeviceGroup::get_local_device_name();
    my $mcp_device_group = F5::DeviceGroup::get_device_group($local_device_name);
    if ($mcp_device_group) {
        my @devices_in_group = F5::DeviceGroup::get_devices_in_device_group($mcp_device_group->{name});
        if (@devices_in_group > 1) {
            ## Check if we are the first to upgrade.
            my $local_device = F5::DeviceGroup::get_device($local_device_name);
            my $upgraded_device = $local_device;
            for my $device_name (@devices_in_group) {
                next if $device_name eq $local_device_name;
                my $mcp_device = F5::DeviceGroup::get_device($device_name);
                if ($mcp_device->{version} ge $local_device->{version}) {
                     $upgraded_device = $mcp_device;
                     last;
                }
            }
            if ($upgraded_device->{name} eq $local_device_name) {
                ## If we are the first to upgrade, update MCP
                my $asm_sync_folder = F5::DeviceGroup::get_or_create_asm_sync_folder($mcp_device_group->{name});
                my $object = F5::ASM::COAPI::Object->new("asm_device_sync");
                $object->{device_name} =  $asm_sync_folder . "/" . basename($local_device_name);
                $object->{timestamp} = time();
                $object->save();
                $logger->info("$local_device_name is the first Device Group Member to upgrade to this version.  Marking device as updated.");
            }
            else {
               ## If we are not the first to upgrade, schedule a request for the DataProtection from a peer that has upgraded
               my ($shared_state_ipc, $shared_state) = F5::Utils::SharedMemory::get_shared_memory(key => $F5::Utils::SharedMemory::SHARED_STATE_KEY);
               $shared_state->{data_protection_peer} = $upgraded_device->{name};
               $logger->info("Scheduling request for Data Protection Configuration from $upgraded_device->{name}");
            }
        }
    }
    $logger->debug("End Data Protection fixup");
}

1;
