package F5::DbFixup::Default;

use strict;
use warnings;
use constant FIND_ONLY_INDEX_FIELDS => 1;
use constant FIND_ONLY_NONINDEX_FIELDS => 2;

use constant MAX_BATCH_SIZE => 5000;
use F5::DbUtils;
use File::Basename;
use F5::BatchInsert;

use constant DBFIXUP_AUTOMERGE_MODE_INSERT_ALL => 1;
use constant DBFIXUP_AUTOMERGE_MODE_UPDATE_BY_INDEX => 2;
use constant DBFIXUP_AUTOMERGE_MODE_INSERT_MISSING_BY_INDEX => 3;

our $AUTOLOAD;
# Global properties for all fixup modules and their default values
our %allowed_global_properties = (min_version => undef,
				  max_version => undef,
				  fixup_db => 'DCC',
				  excluded_tables => [],
				  priority => 10);

sub new {
  my ($class, @params) = @_;
  my $self = bless({}, $class);
  $self->init(@params) if $self->can('init');
  return $self;
}

#----------------------------------------------------------------------

sub is_valid_version {
  my($self, $old_version, $new_version) = @_;
  if ((!defined($self->min_version) || $old_version ge $self->min_version) &&
      (!defined($self->max_version) || $old_version lt $self->max_version)) {
    return 1;
  }
  else {
    return 0;
  }
}

#----------------------------------------------------------------------

sub automerge {
  my($class, %params) = @_;

  my $logger = $params{logger};
  my $dbh = $params{dbh};
  my $databases = $params{databases};
  my $old_dbname = $params{old_dbname};
  my $new_dbname = $params{new_dbname};
  my $abstract_dbname = $params{abstract_dbname};
  my $idmap = $params{idmap};
  my $table = $params{table};
  $table =~ s/^.*\.//;
  my $new_table_name = $table;
  $logger->info("Copying all data for table $table from database $old_dbname to $new_dbname") if $logger;
  my (@indexes, @fields);
  @indexes = @{$params{indexes}} if (defined($params{indexes}));
  @fields = @{$params{fields}} if (defined($params{fields}));
  my $old_index = $params{old_index};
  my $id_map_in = $params{id_map_in};
  my $copy_limit_condition = $params{copy_limit_condition};
  my $common_upgrade = $params{common_upgrade};
  my $batch_insertor = $params{batch_insertor} || F5::BatchInsert->new(dbh => $dbh);

  my $common_upgrade_fixup_callback;
  my $after_table_complete_callback;
  my $common_upgrade_new_table_name;

  if ($common_upgrade) {
      $common_upgrade_fixup_callback = $common_upgrade->get_row_fixup_callback(table_name => "$abstract_dbname.$table");
      $after_table_complete_callback = $common_upgrade->get_after_table_complete_callback(table_name => "$abstract_dbname.$table");
      my $common_upgrade_new_table_name = $common_upgrade->get_new_table_name(table_name => "$abstract_dbname.$table");
      if (defined( $common_upgrade_new_table_name)) {
          $new_table_name = $common_upgrade_new_table_name;
          $new_table_name =~ s/^.*\.//;
      }
  }

  my $factory_data_row_count;
  if ($new_table_name) {
      ($factory_data_row_count) = $dbh->selectrow_array(<<EndQuery);
SELECT COUNT(*)
  FROM $new_dbname.$new_table_name
EndQuery
  }

  if (! $factory_data_row_count) {
      # There are no factory records -- a simple copy table can be done
      my $fixup_callback = undef;

      if ($common_upgrade_fixup_callback) {
          $fixup_callback =sub {
                               my ($dbh, $logger, $data, $new_table_name, $foo, $idmap) = @_;
                               for my $key (keys %$data) {
                                   my $row = $data->{$key};
                                   my $row_state_rv;
                                   $common_upgrade_fixup_callback->(row => $row,
                                                                    idmap => $idmap,
                                                                    databases => $databases,
                                                                    dbh => $dbh,
                                                                    row_state_rv => \$row_state_rv,
                                                                    dbfixup_automerge_mode => DBFIXUP_AUTOMERGE_MODE_INSERT_ALL,
                                                                   );
                                   if ($row_state_rv && $row_state_rv eq 'skip') {
                                       delete $data->{$key};
                                   }
                               }
                           };
      }

      $class->copy_table(%params,
                         table_name => $params{table},
                         new_table_name => $new_table_name,
                         id_field_name => undef,
                         idmap => $idmap,
                         fixup_callback => $fixup_callback,
                        );

      if ($after_table_complete_callback) {
          $after_table_complete_callback->(dbname => $new_dbname,
                                   logger => $logger,
                                   dbh    => $dbh);
      }

      return 1;
  }

  unless (defined($params{indexes}) || defined($params{insert_indexes})) {
    # Figure out index fields.
    # The index must be the same for both source and destination or we bail.
    my @old_indexes = _get_fields($dbh, $old_dbname, $table, FIND_ONLY_INDEX_FIELDS);
    my @new_indexes = _get_fields($dbh, $new_dbname, $new_table_name, FIND_ONLY_INDEX_FIELDS);
    if (join(',', sort @old_indexes) ne join(',', sort @new_indexes)) {
      $logger->fatal("Indexes for table $table didn't match - automatic copy can't be done") if $logger;
      return 0;
    }
    else {
      @indexes = @old_indexes;
      if (@indexes) {
	$logger->debug("Found indexes for table $table - " . join(',', @indexes)) if $logger;
      }
    }
  }

  unless (defined($params{fields}) || defined($params{insert_fields})) {
    # Figure out fields in common to copy.
    my %field_count;
    for (_get_fields($dbh, $old_dbname, $table, FIND_ONLY_NONINDEX_FIELDS)) {
      $field_count{$_}++;
    }
    for (_get_fields($dbh, $new_dbname, $new_table_name, FIND_ONLY_NONINDEX_FIELDS)) {
      $field_count{$_}++;
    }
    @fields = grep($field_count{$_} == 2, keys %field_count);
    $logger->debug("Found non-index fields for table $table - " . join(',', @fields)) if $logger;
  }

  my @insert_fields = $params{insert_fields} ? @{$params{insert_fields}} : @fields;
  my @insert_indexes = $params{insert_indexes} ? @{$params{insert_indexes}} : @indexes;
  my @update_fields = $params{update_fields} ? @{$params{update_fields}} : @fields;
  my @update_indexes = $params{update_indexes} ? @{$params{update_indexes}} : @indexes;

  if ($id_map_in && grep($_ eq $params{map_field}, @update_indexes)) {
    # If there is an incoming id_map and the map field is an index field,
    # each row needs to be compared individually. This is a very costly operation.
    individual_row_merge(%params,
			 insert_indexes => \@insert_indexes,
			 insert_fields => \@insert_fields,
			 update_indexes => \@update_indexes,
			 update_fields => \@update_fields);
  } elsif (@insert_indexes) {
    if (@update_fields) {
      # Update changed values
      my $join_fields = join(',', @update_indexes);
      my $indexes = join(', ', map("OLDT.$_", @update_indexes));
      my $unmatched_field_condition = join(' OR ', map("BINARY NEWT.$_ != OLDT.$_", @update_fields));
      if ($copy_limit_condition) {
	$unmatched_field_condition .= " AND $copy_limit_condition";
      }
      my $find_changed_query = <<"EndQuery";
SELECT $indexes
FROM $new_dbname.$new_table_name NEWT JOIN $old_dbname.$table OLDT USING ($join_fields)
WHERE $unmatched_field_condition
EndQuery

      my $changed_rows_sth = $dbh->prepare($find_changed_query);
      $changed_rows_sth->execute();
      while (my $row = $changed_rows_sth->fetchrow_hashref) {
          log_with_row($logger, $table, $row, 'Overwriting');

          my $index_limit_condition = join(' AND ', map("$_ = ?", @update_indexes));
          my $old_rec = $dbh->selectrow_hashref(<<EndQuery, undef, @{$row}{@update_indexes});
SELECT * 
  FROM $old_dbname.$table 
 WHERE $index_limit_condition
EndQuery

          if ($common_upgrade_fixup_callback) {
              my $row_state_rv;
              $common_upgrade_fixup_callback->(row => $old_rec, idmap => $idmap,
                                               dbfixup_automerge_mode => DBFIXUP_AUTOMERGE_MODE_UPDATE_BY_INDEX,
                                               row_state_rv => \$row_state_rv,
                                              );
              if ($row_state_rv && $row_state_rv eq 'skip') {
                  next;
              }
          }

          F5::DbUtils::update_row(table_name => "$new_dbname.$new_table_name",
                                  dbh => $dbh,
                                  rec => $old_rec,
                                  fields => \@update_fields,
                                  where_fields => { map { ($_ => $row->{$_}) } @update_indexes });
      }
    }

    # Insert unmatched records from old DB into new DB, using only common fields.
    my $join_fields = join(',', @insert_indexes);
    my $all_fields = join(',', @insert_indexes, @insert_fields);
    my $old_db_fields = join(',', map("OLDT.$_", _get_fields($dbh, $old_dbname, $table)));
    if ($old_index) {
      $old_db_fields = "OLDT.$old_index,$old_db_fields";
    }
    my $nonexistant_row_condition = "NEWT.$insert_indexes[0] IS NULL";
    if ($copy_limit_condition) {
      $nonexistant_row_condition .= " AND $copy_limit_condition";
    }
    my $find_added_query = <<"EndQuery";
SELECT $old_db_fields
FROM $old_dbname.$table AS OLDT left outer join $new_dbname.$new_table_name AS NEWT USING ($join_fields)
WHERE $nonexistant_row_condition
EndQuery

    my $find_added_sth = $dbh->prepare($find_added_query);
    $find_added_sth->execute;
    while(my $row = $find_added_sth->fetchrow_hashref) {
      if ($id_map_in && defined($id_map_in->{$row->{$params{map_field}}})) {
	$logger->debug("Mapping $params{map_field} to $id_map_in->{$row->{$params{map_field}}} (from $row->{$params{map_field}})") if $logger;
	$row->{$params{map_field}} = $id_map_in->{$row->{$params{map_field}}};
      }

      if ($common_upgrade_fixup_callback) {
          my $row_state_rv;
          $common_upgrade_fixup_callback->(row => $row,
                                           idmap => $idmap,
                                           row_state_rv => \$row_state_rv,
                                           dbfixup_automerge_mode => DBFIXUP_AUTOMERGE_MODE_INSERT_MISSING_BY_INDEX,
                                          );

          if ($row_state_rv && $row_state_rv eq 'skip') {
              next;
          }
      }

      log_with_row($logger, $new_table_name, $row, 'Adding');
      $batch_insertor->insert_row(table_name => "$new_dbname.$new_table_name",
                                  rec => $row);
    }
  }
  else {
    # No primary key - dump all values in new table
    # Insert unmatched records from old DB into new DB, using only common fields.
    $logger->info("Adding all rows from old data for table $table") if $logger;
    my $all_fields = join(',', @insert_fields);
    my $old_db_fields = join(',', map("OLDT.$_", @insert_fields));
    my $where_clause = '';
    if ($copy_limit_condition) {
      $where_clause = "WHERE $copy_limit_condition";
    }
    my $get_all_rows_query = <<"EndQuery";
SELECT $old_db_fields
FROM $old_dbname.$table AS OLDT
$where_clause
EndQuery

    my $get_all_rows_sth = $dbh->prepare($get_all_rows_query);
    $get_all_rows_sth->execute;
    while(my $row = $get_all_rows_sth->fetchrow_hashref) {
      if ($id_map_in && defined($id_map_in->{$row->{$params{map_field}}})) {
	$logger->debug("Mapping $params{map_field} to $id_map_in->{$row->{$params{map_field}}} (from $row->{$params{map_field}})") if $logger;
	$row->{$params{map_field}} = $id_map_in->{$row->{$params{map_field}}};
      }
      log_with_row($logger, $new_table_name, $row, 'Adding');
      $batch_insertor->insert_row(table_name => "$new_dbname.$new_table_name",
                                  rec => $row);
    }
  }

  if (! $params{batch_insertor}) {
      $batch_insertor->flush;
  }

  if ($after_table_complete_callback) {
      $after_table_complete_callback->(dbh => $dbh,
                               logger => $logger,
                               dbname => $new_dbname,
                              );
  }

  return 1;
}

#----------------------------------------------------------------------

sub individual_row_merge {
  my (%params) = @_;
  my $logger = $params{logger};
  my $dbh = $params{dbh};
  my $old_dbname = $params{old_dbname};
  my $new_dbname = $params{new_dbname};
  my $table = $params{table};
  $table =~ s/^.*\.//;
  my @insert_indexes = @{$params{insert_indexes}};
  my @insert_fields = @{$params{insert_fields}};
  my @update_indexes = @{$params{update_indexes}};
  my @update_fields = @{$params{update_fields}};
  my $old_index = $params{old_index};
  my $id_map_in = $params{id_map_in};
  my $id_map_out = $params{id_map_out};
  my $copy_limit_condition = $params{copy_limit_condition};

  my $all_fields = join(',', @update_indexes, @update_fields);
  my $where_clause = '';
  if ($copy_limit_condition) {
    $where_clause = "WHERE $copy_limit_condition";
  }
  my $get_all_old_rows_query = <<"EndQuery";
SELECT $all_fields
FROM $old_dbname.$table AS OLDT
$where_clause
EndQuery

  my $get_all_old_rows_sth = $dbh->prepare($get_all_old_rows_query);
  $get_all_old_rows_sth->execute();
  while(my $row = $get_all_old_rows_sth->fetchrow_hashref) {
    if ($id_map_in && defined($id_map_in->{$row->{$params{map_field}}})) {
      $logger->debug("Mapping $params{map_field} to $id_map_in->{$row->{$params{map_field}}} (from $row->{$params{map_field}})") if $logger;
      $row->{$params{map_field}} = $id_map_in->{$row->{$params{map_field}}};
    }
    log_with_row($logger, $table, $row, 'Comparing');

    # Search for the existing row in the new DB
    my $join_fields = join(',', @update_indexes);
    my $index_limit_condition = join(' AND ', map("BINARY $_ = ?", @update_indexes));
    my $find_matching_row_query = <<"EndQuery";
SELECT $join_fields
FROM $new_dbname.$table
WHERE $index_limit_condition
EndQuery
    my $new_db_row = $dbh->selectall_arrayref($find_matching_row_query, undef, @{$row}{@update_indexes});

    if (@$new_db_row) {
      # Row exists -- update only the non-index fields
      if (@update_fields) {
	log_with_row($logger, $table, $row, 'Updating');
	my $update_field_list = join(', ', map("$_ = ?", @update_fields));
	my $update_matching_row_query = <<"EndQuery";
UPDATE $new_dbname.$table
SET $update_field_list
WHERE $index_limit_condition 
EndQuery

	$dbh->do($update_matching_row_query, undef, @{$row}{@update_fields, @update_indexes});
      }
    } else {
      # Row doesn't exist -- add it now
      log_with_row($logger, $table, $row, 'Adding');
      my $all_fields = join(',', @insert_indexes, @insert_fields);
      my $placeholders = '?,' x ($#insert_indexes + $#insert_fields + 1) . '?';
      my $update_matching_row_query = <<"EndQuery";
INSERT INTO $new_dbname.$table ($all_fields) VALUES ($placeholders)
EndQuery

      $dbh->do($update_matching_row_query, undef, @{$row}{@insert_indexes, @insert_fields});
    }
  }
}

#----------------------------------------------------------------------
sub _get_fields {
  my($dbh, $dbname, $table, $get_index) = @_;

  my $sth = $dbh->column_info(undef, $dbname, $table, '%');
  $sth->execute;
  my @primary_key;
  while(my $row = $sth->fetchrow_hashref) {
    if (!defined($get_index) || 
	($get_index == FIND_ONLY_NONINDEX_FIELDS && !$row->{mysql_is_pri_key}) ||
	($get_index == FIND_ONLY_INDEX_FIELDS && $row->{mysql_is_pri_key})) {
      push(@primary_key, $row->{COLUMN_NAME});
    }
  }
  return @primary_key;
}

#----------------------------------------------------------------------

# AUTOLOAD is used as a read-only accessor to package global

sub AUTOLOAD {
  my($self) = @_;
  $AUTOLOAD =~ /::([^:]*)$/;
  my $base_variable = $1;
  return unless $base_variable =~ /[^A-Z]/;
  unless (grep($_ eq $base_variable, keys %allowed_global_properties)) {
    warn "Illegal global property accessed - $base_variable\n";
    return undef;
  }
  my $retval = eval('$' . $AUTOLOAD);
  unless (defined($retval)) {
    $retval = $allowed_global_properties{$base_variable};
  }
  return $retval;
}

#----------------------------------------------------------------------

sub log_with_row {
  my($logger, $table, $row, $msg) = @_;

  $logger->info("$msg row with old data for table $table (" . join(',', map("$_ = " . (defined($row->{$_}) ? $row->{$_} : 'null'), grep(defined($_), keys %$row))) . ')') if $logger;
}

#----------------------------------------------------------------------

sub copy_table {
  my($self, %params) = @_;

  my $dbh = $params{dbh};
  my $old_dbname = $params{old_dbname};
  my $new_dbname = $params{new_dbname};
  my $logger = $params{logger};
  my $old_table_name = $params{old_table_name} || $params{table_name};
  my $new_table_name = defined $params{new_table_name} ? $params{new_table_name} : $params{table_name};
  my $fixup_callback = $params{fixup_callback};
  my $ignore = $params{ignore};
  my $idmap = $params{idmap};
  my $preserve_id = $params{preserve_id} || 0;
  my $id_field_name = exists($params{id_field_name}) ? $params{id_field_name} : 'id';
  my $where_clause = $params{where_clause} ? " WHERE ". $params{where_clause} : '';
  my $idmap_out = $params{idmap_out};

  $old_table_name =~ s/^.*\.//;
  if ($new_table_name) {
      $new_table_name =~ s/^.*\.//;
  }

  return unless  F5::DbUtils::check_table_exists(dbh => $dbh,
                                                  dbname => $old_dbname,
                                                  table_name => $old_table_name);

  my $default_batch_size = MAX_BATCH_SIZE;
  if (grep { /\.$new_table_name$/ } F5::DbUtils::get_large_record_tables($params{abstract_dbname})) {
      $default_batch_size = 1;
  }
  my $batch_size = $params{batch_size} || $default_batch_size;
  $idmap->{new_dbname} = $new_dbname;
  my $sth = $dbh->prepare("SELECT * FROM $old_dbname.$old_table_name $where_clause");
  $sth->execute;
  while(my $rows = $sth->fetchall_arrayref({}, $batch_size)) {
    last unless @$rows;

    my $data;
    if ($id_field_name) {
      $data = { map(($_->{$id_field_name} => $_), @$rows) };
    }
    else {
      $data = { map(($_ => $rows->[$_]), 0..$#$rows) };
    }

    $fixup_callback->($dbh, $logger, $data, $new_table_name, {}, $idmap) if $fixup_callback;

    if ($new_table_name && $new_dbname) {
      my $idmap_out_rv = F5::DbUtils::insert_data_to_table(dbh => $dbh,
                                                         logger => $logger,
                                                         data => $data,
                                                         ignore => $ignore,
                                                         table_name => "$new_dbname.$new_table_name",
                                                         preserve_id => $preserve_id,
                                                         id_field_name => $id_field_name,
                                                         idmap => $idmap,
                                                         idmap_save => defined $idmap_out,
                                                       );
      if ($idmap_out) {
        %$idmap_out = (%$idmap_out, %$idmap_out_rv);
      }
    }
    last unless $sth->{Active};
  }
}

#----------------------------------------------------------------------

sub copy_user_defined_values {
  my($self, %params) = @_;

  my $dbh = $params{dbh};
  my $user_defined_ids = $dbh->selectcol_arrayref($params{get_user_defined_ids_sql});

  for my $id (@{$user_defined_ids}) {

    for my $excluded_table (@{$params{tables}}) {
      my($table_name) = ($excluded_table =~ /^.*?\.(.*)$/);

      $self->copy_table(dbh => $dbh,
                        old_dbname => $params{old_dbname},
                        new_dbname => $params{new_dbname},
                        logger => $params{logger},
                        table_name => $table_name,
                        preserve_id => 1,
                        where_clause => "$params{id_field_name} = $id",
                        id_field_name => $params{id_field_name},
                       );
    }
  }
}

#----------------------------------------------------------------------

sub get_module_path {
  return dirname(__FILE__);
}

#----------------------------------------------------------------------

sub was_hsl {
    my($self, %params) = @_;
    my $dbh = $params{dbh};
    my $dbname = $params{dbname} || 'PLC';

    my $hsl_policy_id;

    if ($params{account_id}) {
        ($hsl_policy_id) = $dbh->selectrow_array(<<EndQuery, undef, $params{account_id});
SELECT id
  FROM $dbname.PL_POLICIES
  WHERE flg_is_hsl = 1
    AND account_id = ?
EndQuery
    }

    return $hsl_policy_id;
}

#----------------------------------------------------------------------

1;
