package F5::DbFixup::DosProfileMigration;
use warnings;
use strict;

use F5::DbUtils;
use F5::ASM::COAPI;
use F5::Mcp::PolicyTarget;
use F5::Mcp::PolicyAction;

our @ISA = qw(F5::DbFixup::Default);

# Package globals, handled by AUTOLOAD
our $max_version = '0011.0003.0000.0000.0000';
our $fixup_db = 'PLC';
our $priority = 99;

my %prevention_policy_bitmask_values = (
                                        flg_source_ip_client_side_defense => 1,
                                        flg_source_ip_rate_limit => 2,
                                        flg_url_rate_limit => 4,
                                        flg_url_client_side_defense => 8,
                                       );

my %OPERATION_MODE_MAPPING = ( OFF => 0,
                               TRANSPARENT => 1,
                               BLOCKING => 2,
                             );

my @defense_mechanism_flags = qw(
                                    source_ip_based_client_side_integrity_defense
                                    url_based_client_side_integrity_defense
                                    source_ip_based_rate_limiting
                                    url_based_rate_limiting
                               );

sub after_fixup {
    my ($self,$dbh,$old_dbname,$new_dbname, $logger, $params) = @_;
  
    my $idmap = $params->{idmap};
    my $databases = $params->{databases};
 
    $logger->debug("Begin Dos Profile fixup");

#Gather old info
#Migrate profiles of active policies

    my %enabled_policies;
    my @dos_profiles;
    if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'PL_DOS_SETTINGS')) {
        #Get settings for active profiles
        my $dosprof_sth = $dbh->prepare(<<EndQuery);
SELECT *
  FROM $old_dbname.PL_DOS_SETTINGS
 WHERE policy_id IN (SELECT active_policy_id FROM $databases->{DCC}->{ucs_dbname}.ACCOUNTS)
   AND account_detection = 1
EndQuery

        $dosprof_sth->execute();
        while (my $profile = $dosprof_sth->fetchrow_hashref) {
            $enabled_policies{$profile->{policy_id}} = 1;
            ## From F5::CommonUpgrade::DosSettingsDetectionMode (10.1.0)
            if (! defined($profile->{detection_mode}) ) {
                $profile->{detection_mode} = 1; # Default to "Latency"
            }
            if($params->{old_version} lt '0010.0001.0000.0000.0000') {
                ## From F5::CommonUpgrade::DosSettingsSplitClientside (10.1.0)
                if ($profile->{prevention_policy_mask} &&
                    $profile->{prevention_policy_mask} & $prevention_policy_bitmask_values{flg_source_ip_client_side_defense}) {
                    $profile->{prevention_policy_mask} |= $prevention_policy_bitmask_values{flg_url_client_side_defense};
                }            
                ## From F5::CommonUpgrade::DosSettingsCopyRateLimit (10.1.0)
                if ($profile->{rps_increase_rate_limit}) {
                    $profile->{rps_increase_rate_limit_object} = $profile->{rps_increase_rate_limit};
                    $profile->{rps_increase_rate_limit_ip} = $profile->{rps_increase_rate_limit};
                }
            }

            $profile->{trigger_irule} = 0;
            if ($params->{old_version} ge '0010.0002.0000.0000.0000') {
                ($profile->{trigger_irule}) = $dbh->selectrow_array(<<EndQuery, undef, $profile->{policy_id});
SELECT flg_irules
  FROM $old_dbname.PL_POLICIES
 WHERE id = ?
EndQuery
            }

            ## From F5::CommonUpgrade::DosSettingsSplitBitmask (10.2.0)
            if (exists($profile->{prevention_policy_mask})) {
                for my $bitmask_field (keys %prevention_policy_bitmask_values) {
                    $profile->{$bitmask_field} = ($profile->{prevention_policy_mask} & $prevention_policy_bitmask_values{$bitmask_field} ? 1 : 0);
                }
            }
            ## From F5::CommonUpgrade::DosMinimumSuspicious (11.1.0)
            if ($idmap && $idmap->{'__private_save'} && $idmap->{'__private_save'}->{deleted_internal_config_items}) {
                my $saved_internal_config = $idmap->{'__private_save'}->{deleted_internal_config_items};
                if (defined $saved_internal_config->{dos_min_detection_object_threshold}) {
                    $profile->{minimum_suspicious_object_threshold} = $saved_internal_config->{dos_min_detection_object_threshold};
                }
                if (defined $saved_internal_config->{dos_min_detection_ip_threshold}) {
                    $profile->{minimum_suspicious_ip_threshold} = $saved_internal_config->{dos_min_detection_ip_threshold};
                }
            }
            if (! $profile->{minimum_suspicious_object_threshold} ||
                $profile->{minimum_suspicious_object_threshold} > $profile->{suspicious_object_threshold}) {
                $profile->{minimum_suspicious_object_threshold} = $profile->{suspicious_object_threshold};
            }
            if ( !$profile->{minimum_suspicious_ip_threshold} ||
                $profile->{minimum_suspicious_ip_threshold} > $profile->{suspicious_ip_threshold}) {
                $profile->{minimum_suspicious_ip_threshold} = $profile->{suspicious_ip_threshold};
            }
    
            ## From F5::CommonUpgrade::IpUnification (11.2.0)
            if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'PL_DOS_IP_WHITE_LIST')) {
                $profile->{ip_white_list} = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $profile->{policy_id});
SELECT ip_mask AS mask, ip_address
  FROM $old_dbname.PL_DOS_IP_WHITE_LIST
 WHERE policy_id = ?
EndQuery
            }
            elsif (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'PL_IP_WHITELIST')) {
                $profile->{ip_white_list} = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $profile->{policy_id});
SELECT mask, ip_address
  FROM $old_dbname.PL_IP_WHITELIST
 WHERE policy_id = ?
   AND flg_ignore_anomaly = 1
EndQuery
            }
            push (@dos_profiles, $profile);
        }
    }

    ## Get Account Info
    my %policy_id_to_policy_name;
    my %l7policy_to_asm_policy;
    my $account_sth = $dbh->prepare(<<EndQuery);
SELECT active_policy_id, policy_name, accounts.account_id
  FROM $databases->{DCC}->{staging_dbname}.ACCOUNTS accounts
    JOIN $databases->{DCC}->{ucs_dbname}.ACCOUNTS old_accounts on accounts.account_id = old_accounts.account_id
EndQuery
    $account_sth->execute();
    while (my $account = $account_sth->fetchrow_hashref) {
        next unless $enabled_policies{$account->{active_policy_id}};
        if ((defined($idmap->{'__private_save'}->{hsl_accounts}) && $idmap->{'__private_save'}->{hsl_accounts}->{$account->{account_id}}) ||
            (!defined($idmap->{'__private_save'}->{hsl_accounts}) && F5::DbUtils::check_table_exists(dbh => $dbh,
                                                                                                     table_name => 'HSL_HTTP_PROFILES',
                                                                                                     dbname => $old_dbname))
                                                                                 && $self->was_hsl(dbh => $dbh,
                                                                                            account_id => $account->{account_id},
                                                                                            dbname => $old_dbname,)) {
            #Don't migrate from HSL accounts;
            next;
        }
        for my $policy_action (@{F5::ASM::COAPI::Query->new("policy_action", {target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM,
                                                                              action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE,
                                                                              policy => $account->{policy_name}})}) {
            $l7policy_to_asm_policy{$policy_action->{policy_name}} = $account->{policy_name};
        }

        $policy_id_to_policy_name{$account->{active_policy_id}} = $account->{policy_name};
    }

    my $SUFFIX = "--profile-dos";

    #Migrate profiles into MCP objects.
    for my $profile (@dos_profiles)  {
        next unless $policy_id_to_policy_name{$profile->{policy_id}};
        my $name = $policy_id_to_policy_name{$profile->{policy_id}} . $SUFFIX;

        my $dos_profile = F5::ASM::COAPI::Object->new('profile_dos');
        $dos_profile->{name} = $name;
        $dos_profile->save();
    
        my $dos_application = F5::ASM::COAPI::Object->new('dos_application');
        $dos_application->{profile_name} = $name;
        $dos_application->{name} = $name;

        # Reset fields which are set by default in new Dos profiles
        $dos_application->{tps_based_mode} = 0;
        $dos_application->{heavy_url_protection} = 0;
        $dos_application->{heavy_url_automatic_detection} = 0;

        my $detection = $profile->{detection_mode} ? "stress_based_"
                                                   : "tps_based_";
      
        $dos_application->{$detection . 'mode'} = $OPERATION_MODE_MAPPING{OFF};
        if ($profile->{account_detection}) {
            $dos_application->{$detection . 'mode'} = $profile->{transparent_mode} ? $OPERATION_MODE_MAPPING{TRANSPARENT}
                                                                                   : $OPERATION_MODE_MAPPING{BLOCKING};
        }
        $dos_application->{$detection . 'source_ip_based_client_side_integrity_defense'} = $profile->{flg_source_ip_client_side_defense};
        $dos_application->{$detection . 'url_based_client_side_integrity_defense'}       = $profile->{flg_url_client_side_defense};
        $dos_application->{$detection . 'source_ip_based_rate_limiting'} = $profile->{suspicious_ip_threshold};
        $dos_application->{$detection . 'url_based_rate_limiting'} = $profile->{suspicious_object_threshold};
        $dos_application->{$detection . 'ip_tps_increase_rate'} = $profile->{rps_increase_rate_limit_ip};
        $dos_application->{$detection . 'ip_maximum_tps'} = $profile->{suspicious_ip_threshold};
        $dos_application->{$detection . 'ip_minimum_tps'} = $profile->{minimum_suspicious_ip_threshold};
        $dos_application->{$detection . 'url_tps_increase_rate'} = $profile->{rps_increase_rate_limit_object};
        $dos_application->{$detection . 'url_maximum_tps'} = $profile->{suspicious_object_threshold};
        $dos_application->{$detection . 'url_minimum_tps'} = $profile->{minimum_suspicious_object_threshold};

        my $active_defense_mechanism_count = scalar grep { $dos_application->{$detection . $_} } @defense_mechanism_flags;

        # In case when 'max_prevention_interval' is less then 'active_defense_mechanism_count' (but larger then a '0') we'll encounter this error:
        #     01071780:3: The de-escalation period can be either zero or greater than or equal to the escalation period in TPS-based anomaly.
        # because 'escalation_period' will be set to '600' (since int() will result in a '0') 
        #    and 'deescalation_period' will be set to something smaller than that (but larger than '0')
        #
        # note that 'active_defense_mechanism_count' is in the range of (1,4)
        # thus, in case 'max_prevention_interval' is less then 'active_defense_mechanism_count' 
        #    then 'max_prevention_interval' is in the range of (1,3)
        #    in which case we can easily bump it up to be equal to 'active_defense_mechanism_count' and go on with the upgrade
        if ($profile->{max_prevention_interval} && $profile->{max_prevention_interval} < $active_defense_mechanism_count) {
            $profile->{max_prevention_interval} = $active_defense_mechanism_count;
        }

        $dos_application->{$detection . 'escalation_period'} = int($profile->{max_prevention_interval} / $active_defense_mechanism_count) || 600;
        $dos_application->{$detection . 'deescalation_period'} = $profile->{max_prevention_interval}+0 || 0;

        $dos_application->{latency_increase_rate} = $profile->{latency_increase_rate_limit};
        $dos_application->{maximum_latency} = $profile->{maximum_latency_threshold};
        $dos_application->{minimum_latency} = $profile->{minimum_latency_threshold};
 
        $dos_application->{trigger_irule} = $profile->{trigger_irule};
        $dos_application->save();

        if ($profile->{ip_white_list} && @{$profile->{ip_white_list}}) {
            for my $white_ip (@{$profile->{ip_white_list}}) {
                my $dos_white_ip = F5::ASM::COAPI::Object->new('dos_white_ip');
                $dos_white_ip->{parent_profile} = $name;
                $dos_white_ip->{parent_name} = $name;
                $dos_white_ip->{address} = $white_ip->{ip_address};
                $dos_white_ip->{netmask} = $white_ip->{mask};
                eval {
                    $dos_white_ip->save();
                };
                if ($@) {
                    $logger->error("Could not migrate DOS Allowlist ($name) - IP: $white_ip->{ip_address}, Mask: $white_ip->{mask}. Error: $@");
                }
            }
       }
    }

    # Assign profiles to VSs that had the policy assigned to them.
    my $virtual_servers = F5::ASM::COAPI::Query->new('virtual_server_policy');
    if ($virtual_servers && @$virtual_servers) {
        #use last http class (vs may have more than one attached)
        my %vs_dos_profiles;
        for my $virtual_server (@$virtual_servers) {
            if ($l7policy_to_asm_policy{$virtual_server->{policy_name}}) {
                $vs_dos_profiles{$virtual_server->{vs_name}} = $l7policy_to_asm_policy{$virtual_server->{policy_name}};
            }
        }
        for my $virtual_server (keys %vs_dos_profiles) {
            my $server_profile = F5::ASM::COAPI::Object->new('virtual_server_profile');
            $server_profile->{vs_name} = $virtual_server;
            $server_profile->{profile_name} = $vs_dos_profiles{$virtual_server} . $SUFFIX;
            $server_profile->save();
        }
    }

    $logger->debug("End Dos Profile fixup");
}

#----------------------------------------------------------------------

1;
