package F5::DbFixup::LoggingProfileMigration;
use warnings;
use strict;

use F5::DbUtils;
use F5::ASM::COAPI;
use F5::Mcp::ProfileType;
use F5::Mcp::PolicyTarget;
use F5::Mcp::PolicyAction;

our @ISA = qw(F5::DbFixup::Default);

# Package globals, handled by AUTOLOAD
our $max_version = '0011.0003.0000.0000.0000';
our $fixup_db = 'PLC';
our $priority = 99;

my $LOG_ILLEGAL_LOGPROF_ID = 1;
my $LOG_ALL_LOGPROF_ID = 2;
my $HSL_LOGPROF_ID = 299998;
my $SYSLOG_LEVEL_INFO = 6;
my $FW_LOG_ENTRY_LENGTH_1K = 1024;
my $FW_LOG_ENTRY_LENGTH_2K = 2048;
my $FW_LOG_RESPONSE_LOGGING_NONE = 1;

my %response_logging_map = ( 'none'    => 1,
                             'illegal' => 2,
                             'all'     => 3,);

my %remote_storage_map = ( 'none' => 0,
                           'remote' => 1,
                           'remote-logging' => 2,
                           'arcsight' => 3,
                           'remote-bigiq' => 4);

my %remote_storage_destination_map = ('remote' => 'log_destination_remote_syslog',
                                      'remote-logging' => 'log_destination_splunk',
                                      'remote-bigiq' => 'log_destination_bigiq',
                                      'arcsight' => 'log_destination_arcsight');


my %application_protocol_map = ( 'UDP' => 0,
                                 'TCP' => 1,
                                 'TCP-RFC3195' => 2);

my %destination_protocol_map = ( 'TCP' => 0,
                                 'TCP-RFC3195' => 0,   #Not used for PSM
                                 'UDP' => 1,);

my %filter_key_map = ( 'request_type' => 10,
                       'protocol'     => 11,
                       'filter_response_codes' => 12,
                       'filter_method_codes'  => 13,);

my %FW_LOG_STORAGE_FORMAT = ( PREDEFINED => 0,
                              USER_DEFINED => 1,);

sub after_fixup {
    my ($self,$dbh,$old_dbname,$new_dbname, $logger, $params) = @_;
    my $idmap = $params->{idmap};
    my $databases = $params->{databases};
    my @local_profiles;

    $logger->debug("Begin Logging Profile fixup");


#Gather old info
    my @logging_profiles;
    if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'LOGPROF_LOGGING_PROFILES')) {

        my $http_methods = F5::ASM::COAPI::Query->new('asm_http_method');
        my %http_method_map = map { ($_->{code} => $_->{leaf_name}) } @$http_methods;

        my $storage_fields = F5::ASM::COAPI::Query->new('fw_log_storage_field');
        my %storage_pool_map = map { ( $_->{id} => $_->{name} ) } @$storage_fields;

        #Only migrate user defined and hsl profiles (which can be non-default)
        my $logprof_sth = $dbh->prepare(<<EndQuery);
SELECT * 
  FROM $old_dbname.LOGPROF_LOGGING_PROFILES
 WHERE user_defined = 1
    OR logprof_id = ?
EndQuery

        $logprof_sth->execute($HSL_LOGPROF_ID);
        while (my $logprof = $logprof_sth->fetchrow_hashref) {
            $logprof->{target_system} ||= 'http_full';
            #Fixup all LOGPROF properties
            if (exists $logprof->{storage_type}) {
                # F5::DbFixup::LoggingProfilesUserDefined
                $logprof->{local_storage_type}  = $logprof->{storage_type} eq 'local' ? $logprof->{storage_type} : 'none';
                $logprof->{remote_storage_type} = $logprof->{storage_type} ne 'local' ? $logprof->{storage_type} : 'none';
            }

            my @nested_fields = ({ field => 'remote_loggers',
                                   table_name => 'LOGPROF_REMOTE_CONFIG',
                                   additional_where => 'AND protocol != ""',
                                 },
                                 { field => 'storage_format', table_name => 'LOGPROF_STORAGE_FORMAT_FIELDS', order_by => 'ORDER BY order_id'},
                                 { field => 'logging_profile_filter', table_name => 'LOGPROF_FILTERS', },
                                 { field => 'filter_method_codes', table_name => 'LOGPROF_FILTER_METHODS', },
                                 { field => 'filter_response_codes', table_name => 'LOGPROF_FILTER_RESPONSE_CODES', },
                                 { field => 'filter_pattern', table_name => 'LOGPROF_FILTER_PATTERNS', },
                                );
                                  
            for my $nested_field (@nested_fields) {
                my $additional_where = $nested_field->{additional_where} || "";
                my $order_by   = $nested_field->{order_by} || "";
                my $table_name =  $nested_field->{table_name};
                $logprof->{$nested_field->{field}} = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $logprof->{logprof_id});
SELECT * 
  FROM $old_dbname.$table_name
 WHERE logprof_id = ?
 $additional_where
 $order_by
EndQuery
            }

            next if ($logprof->{target_system} eq 'hsl' 
                    && ( ($logprof->{remote_storage_type} eq 'none') ||
                         (!$logprof->{remote_loggers} || !@{$logprof->{remote_loggers}})
                   ));

            if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'LOGPROF_FILTER_PATTERN_GROUPS') 
                && $logprof->{filter_pattern} && @{$logprof->{filter_pattern}}) {
                # F5::DbFixup::LoggingProfileSearchTarget
                for my $filter_pattern (@{$logprof->{filter_pattern}}) {
                    ($filter_pattern->{search_target}) = $dbh->selectrow_array(<<EndQuery, undef, $filter_pattern->{logprof_filter_pattern_group_id});
SELECT search_target
  FROM $old_dbname.LOGPROF_FILTER_PATTERN_GROUPS
 WHERE logprof_filter_pattern_group_id = ?
EndQuery
                }
            }
            
            if ($logprof->{logging_profile_filter}) {
                # (filter request_type and protocol default to "ALL") (F5::DbFixup::LoggingProfileFilterProtocol)
                for my $logging_profile_filter (@{$logprof->{logging_profile_filter}}) {
                    $logging_profile_filter->{request_type} ||= "ALL";
                    $logging_profile_filter->{protocol} ||= "ALL";
                }
            }
            #Remove spaces for MCP compatibility
            $logprof->{logprof_name} =~ s/ /_/g;
            push (@logging_profiles, $logprof);
        }

        # Create new (custom profiles) as objects in COAPI
        for my $logging_profile (@logging_profiles) {
            my @additional_objects;

            my $log_profile = F5::ASM::COAPI::Object->new('fw_log_profile');
            $log_profile->{name} = $logging_profile->{logprof_name};
            push (@additional_objects, $log_profile);

            if ($logging_profile->{target_system} eq 'http_full') {           
                my $fw_logger =  F5::ASM::COAPI::Object->new('fw_log_application');
                $fw_logger->{name}          = $logging_profile->{logprof_name};
                $fw_logger->{profile_name}  = $logging_profile->{logprof_name};
                $fw_logger->{local_storage} = $logging_profile->{local_storage_type} eq 'none' ? 0 : 1;
                $fw_logger->{guarantee_logging}          = $logging_profile->{guaranteed_logging};
                $fw_logger->{response_logging}           = ($logging_profile->{response_logging} && ($response_logging_map{$logging_profile->{response_logging}} > 0))
                                                           ? $response_logging_map{$logging_profile->{response_logging}} : $FW_LOG_RESPONSE_LOGGING_NONE;
                $fw_logger->{guarantee_response_logging} = $logging_profile->{response_guaranteed_logging} if exists $logging_profile->{response_guaranteed_logging};
                $fw_logger->{remote_storage}             = $remote_storage_map{$logging_profile->{remote_storage_type}}; 
                $fw_logger->{gui_config_level}           = $logging_profile->{user_config_level} eq 'basic' ? 0 : 1;

                #Dos profile migration
                my $dos_logging_profile = F5::ASM::COAPI::Object->new('fw_log_dos_application');
                $dos_logging_profile->{profile_name} = $logging_profile->{logprof_name};
                $dos_logging_profile->{name} = $logging_profile->{logprof_name};
                $dos_logging_profile->{local_publisher} = 'local-db-publisher';  #Preserve local logging for DoS attacks, which was always enabled, with the default local publisher.

                if ($logging_profile->{remote_loggers} && @{$logging_profile->{remote_loggers}} && $logging_profile->{remote_storage_type} ne 'none') {
                    #There are many fields that are common to all entries in the remote_loggers
                    my $common_remote_logger = $logging_profile->{remote_loggers}->[0];
                    $fw_logger->{protocol}        = $application_protocol_map{$common_remote_logger->{protocol}};
                    $fw_logger->{facility}        = ($common_remote_logger->{facility} - 16);  # Enum in MCP starts at 0, ASMconfig enum started at 16;
                    if ($logging_profile->{remote_storage_type} eq "remote") {
                        $fw_logger->{format}          = $common_remote_logger->{format_string} ? $FW_LOG_STORAGE_FORMAT{USER_DEFINED}
                                                                                               : $FW_LOG_STORAGE_FORMAT{PREDEFINED}; 

                        $common_remote_logger->{format_string} =~ s/\%web_application_name\%/\%http_class_name\%/g;
                        $fw_logger->{user_string}     = $common_remote_logger->{format_string};

                        if ($logging_profile->{storage_format} && @{$logging_profile->{storage_format}}) {
                            my @fields = map { $storage_pool_map{$_->{logprof_storage_field_id}} } @{$logging_profile->{storage_format}};
                            $fw_logger->{fields} = \@fields;
                        }
                    }
                    $fw_logger->{field_delimiter} = $common_remote_logger->{csv_field_delim} || ',';
                    $fw_logger->{maximum_request_size} = $common_remote_logger->{max_request_length};
                    $fw_logger->{maximum_header_size}  = $common_remote_logger->{max_header_length};
                    $fw_logger->{maximum_query_size}   = $common_remote_logger->{max_query_string_length};

                    # 'fw_logger->{maximum_entry_length}' can remain undef but cant be set to 0
                    if ($common_remote_logger->{max_log_length} && $common_remote_logger->{protocol}) {
                        # These lengths are set according to the validation 
                        #    in - FwLogApplication::validate_commit
                        #    at - ../tm_daemon/mcp/firewall/validation/FwLogApplication.cpp
                        # -------------------------------------------------------------------------
                        # FW_LOG_REMOTE_PROTOCOL_UDP
                        if ($common_remote_logger->{protocol} eq 'UDP') {
                            $fw_logger->{maximum_entry_length} = $FW_LOG_ENTRY_LENGTH_1K;
                        }
                        # FW_LOG_REMOTE_PROTOCOL_TCP_RFC3195
                        elsif ($common_remote_logger->{protocol} eq 'TCP-RFC3195') {
                            $fw_logger->{maximum_entry_length} = $FW_LOG_ENTRY_LENGTH_2K;
                        }
                        # FW_LOG_REMOTE_PROTOCOL_TCP
                        elsif ($common_remote_logger->{protocol} eq 'TCP') {
                            # maximum entry length cannot be 1k (or larger) while remote protocol is TCP
                            if ($common_remote_logger->{max_log_length} >= $FW_LOG_ENTRY_LENGTH_1K) {
                               $common_remote_logger->{max_log_length} = $FW_LOG_ENTRY_LENGTH_1K - 1;
                            }
                            $fw_logger->{maximum_entry_length} = $common_remote_logger->{max_log_length};
                        }
                        # -------------------------------------------------------------------------
                    }

                    $fw_logger->{report_anomalies}     = $common_remote_logger->{flg_log_anomaly_attacks} if exists $common_remote_logger->{flg_log_anomaly_attacks};

                    for my $remote_logger (@{$logging_profile->{remote_loggers}}) {
                        my $server_address = F5::ASM::COAPI::Object->new('fw_log_server_address');
                        $server_address->{parent_profile} = $logging_profile->{logprof_name};
                        $server_address->{parent_name} = $logging_profile->{logprof_name};
                        $server_address->{address} = $remote_logger->{server_ip};
                        $server_address->{port} = $remote_logger->{server_port};
                        push (@additional_objects, $server_address);
                    }

                    #Dos remote logging profiles are only created if the flg is on, and can only be for splunk or arcsight targets.
                    if ($common_remote_logger->{flg_log_anomaly_attacks}) {
                        if ($logging_profile->{remote_storage_type} eq 'remote') {
                            #Change syslog to splunk target, as syslog is not supported for DoS logging
                            $logging_profile->{remote_storage_type} = 'remote-logging';
                        }

                        my ($publisher, @other_objects) = create_publisher($logging_profile);
                        push (@additional_objects, $publisher, @other_objects);

                        $dos_logging_profile->{remote_publisher} = $publisher->{name};
                    }
                }
                push (@additional_objects, $dos_logging_profile);

                if ($logging_profile->{logging_profile_filter} && @{$logging_profile->{logging_profile_filter}}) {
                    my $logging_profile_filter = $logging_profile->{logging_profile_filter}->[0];
                    $fw_logger->{logic_operation}         = $logging_profile_filter->{logic_relation} eq 'OR' ? 1 : 0;
                    $fw_logger->{filter_gui_config_level} = $logging_profile_filter->{user_config_level} eq 'basic' ? 0 : 1;
                    for my $filter_field (qw(request_type protocol)) {
                        my $filter = F5::ASM::COAPI::Object->new('fw_log_request_filter');
                        $filter->{parent_profile} = $logging_profile->{logprof_name};
                        $filter->{parent_name}    = $logging_profile->{logprof_name};
                        $filter->{key}            = $filter_key_map{$filter_field};
                        $filter->{value_list}     = [ lc($logging_profile_filter->{$filter_field}) ];
                        push (@additional_objects, $filter);

                        if($fw_logger->{local_storage} == 1 && $fw_logger->{remote_storage} > 0)
                        {
                           my $local_filter = F5::ASM::COAPI::Object->new('fw_log_request_filter');
                           $local_filter->{parent_profile} = $logging_profile->{logprof_name} . "_local";
                           $local_filter->{parent_name}    = $logging_profile->{logprof_name} . "_local";
                           $local_filter->{key}            = $filter_key_map{$filter_field};
                           $local_filter->{value_list}     = [ lc($logging_profile_filter->{$filter_field}) ];
                           push (@additional_objects, $local_filter);
                        }
                    }
                }
                if ($logging_profile->{filter_method_codes} && @{$logging_profile->{filter_method_codes}}) {
                    my $filter = F5::ASM::COAPI::Object->new('fw_log_request_filter');
                    $filter->{parent_profile} = $logging_profile->{logprof_name};
                    $filter->{parent_name}    = $logging_profile->{logprof_name};
                    $filter->{key}            = $filter_key_map{filter_method_codes};
                    my @value_list = grep { defined $_ } # skip user-defined methods
                                     map { $http_method_map{$_->{method_code}} } @{$logging_profile->{filter_method_codes}};

                    if (@value_list) {
                        $filter->{value_list} = \@value_list;
                        push (@additional_objects, $filter);
                    }

                     if($fw_logger->{local_storage} == 1 && $fw_logger->{remote_storage} > 0)
                     {
                        my $local_filter = F5::ASM::COAPI::Object->new('fw_log_request_filter');
                        $local_filter->{parent_profile} = $logging_profile->{logprof_name} . "_local";
                        $local_filter->{parent_name}    = $logging_profile->{logprof_name} . "_local";
                        $local_filter->{key}            = $filter_key_map{filter_method_codes};

                        if (@value_list) {
                           $local_filter->{value_list}     = \@value_list;
                           push (@additional_objects, $local_filter);
                        }
                     }
                }
                if ($logging_profile->{filter_response_codes} && @{$logging_profile->{filter_response_codes}}) {
                    my $filter = F5::ASM::COAPI::Object->new('fw_log_request_filter');
                    $filter->{parent_profile} = $logging_profile->{logprof_name};
                    $filter->{parent_name}    = $logging_profile->{logprof_name};
                    $filter->{key}            = $filter_key_map{filter_response_codes};
                    my @value_list = map { $_->{response_code_id} } @{$logging_profile->{filter_response_codes}};
                    $filter->{value_list}     = \@value_list;
                    push (@additional_objects, $filter);

                    if($fw_logger->{local_storage} == 1 && $fw_logger->{remote_storage} > 0)
                     {
                        my $local_filter = F5::ASM::COAPI::Object->new('fw_log_request_filter');
                        $local_filter->{parent_profile} = $logging_profile->{logprof_name} . "_local";
                        $local_filter->{parent_name}    = $logging_profile->{logprof_name} . "_local";
                        $local_filter->{key}            = $filter_key_map{filter_response_codes};
                        $local_filter->{value_list}     = \@value_list;
                        push (@additional_objects, $local_filter);
                     }
                }
                if ($logging_profile->{filter_pattern} && @{$logging_profile->{filter_pattern}}) {
                    for my $filter_pattern (@{$logging_profile->{filter_pattern}}) {
                        my $filter = F5::ASM::COAPI::Object->new('fw_log_request_filter');
                        $filter->{parent_profile} = $logging_profile->{logprof_name};
                        $filter->{parent_name}    = $logging_profile->{logprof_name};
                        $filter->{key}            = $filter_pattern->{search_target};
                        $filter->{value_list}     = [ $filter_pattern->{pattern} ];
                        push (@additional_objects, $filter);

                        if($fw_logger->{local_storage} == 1 && $fw_logger->{remote_storage} > 0)
                        {
                           my $local_filter = F5::ASM::COAPI::Object->new('fw_log_request_filter');
                           $local_filter->{parent_profile} = $logging_profile->{logprof_name} . "_local";
                           $local_filter->{parent_name}    = $logging_profile->{logprof_name} . "_local";
                           $local_filter->{key}            = $filter_pattern->{search_target};
                           $local_filter->{value_list}     = [ $filter_pattern->{pattern} ];
                           push (@additional_objects, $local_filter);
                        }
                    }
                }
                #Needs to be created first
                  
                if($fw_logger->{local_storage} == 1 && $fw_logger->{remote_storage} > 0)
                {
                     # in order to create 2 profile from $fw_logger, 1 local & 1 remote 
                     # the $fw_logger will be init to be the remote and we will create another 
                     # profile $fw_logger_local that will hold the local properties. 

                     # save all local vars
                     my $guarantee_logging = $fw_logger->{guarantee_logging};
                     my $response_logging = $fw_logger->{response_logging};
                     my $guarantee_response_logging = $fw_logger->{guarantee_response_logging};

                     $fw_logger->{local_storage} = 0;
                     $fw_logger->{guarantee_logging} = 0;
                     $fw_logger->{response_logging} = 1;
                     $fw_logger->{guarantee_response_logging} = 0;

                     # save remote
                     push (@additional_objects, $fw_logger);

                     my $fw_logger_local =  F5::ASM::COAPI::Object->new('fw_log_application');

                     # create local
                     $fw_logger_local->{guarantee_logging} = $guarantee_logging;
                     $fw_logger_local->{response_logging} = $response_logging;
                     $fw_logger_local->{guarantee_response_logging} = $guarantee_response_logging;
                     $fw_logger_local->{local_storage} = 1;
                     $fw_logger_local->{logic_operation} = $fw_logger->{logic_operation};
                     $fw_logger_local->{filter_gui_config_level} = $fw_logger->{filter_gui_config_level};
                     
                     # create new log profile record
                     $log_profile->{name} = $logging_profile->{logprof_name} . "_local";
                     push (@additional_objects, $log_profile);

                     $fw_logger_local->{name} = $log_profile->{name};
                     $fw_logger_local->{profile_name} = $log_profile->{name};
                     
                     # save local
                     push (@additional_objects, $fw_logger_local);

                     # save local profile name for vs connection
                     push (@local_profiles, {remote => $logging_profile->{logprof_name}, local => $fw_logger_local->{name}});
                }
                else
                {    
                     push (@additional_objects, $fw_logger);
                }
            }
            elsif($logging_profile->{target_system} eq 'hsl') {
                #hsl profile
                #migrate to publisher (pool, remote logger, destinations, etc)
                my ($publisher, @other_objects) = create_publisher($logging_profile);

                push (@additional_objects, $publisher, @other_objects);

                my $protocol_logger = F5::ASM::COAPI::Object->new('fw_log_protocol_transfer');
                $protocol_logger->{name}          = $logging_profile->{logprof_name};
                $protocol_logger->{profile_name}  = $logging_profile->{logprof_name};
                $protocol_logger->{publisher}     = $publisher->{name};
                push (@additional_objects, $protocol_logger);
            }

            #Objects will be added in the order pushed into the array for dependencies.
            F5::ASM::COAPI::TransactionalSave(\@additional_objects);
        }
    }

    my %local_profiles_hash = map { ( $_->{remote} => $_ ) } @local_profiles;

    #Add default Logging Profiles for mapping
    push (@logging_profiles, {logprof_id => $LOG_ALL_LOGPROF_ID, logprof_name => 'Log all requests'});
    push (@logging_profiles, {logprof_id => $LOG_ILLEGAL_LOGPROF_ID, logprof_name => 'Log illegal requests'});
    my %logging_profile_hash = map { ( $_->{logprof_id} => $_ ) } @logging_profiles;

    #Match Logging Profiles to active accounts
    my %policy_logging_profiles;
    my $account_sth = $dbh->prepare(<<EndQuery);
SELECT old_accounts.*, accounts.policy_name 
  FROM $databases->{DCC}->{ucs_dbname}.ACCOUNTS old_accounts
    JOIN $databases->{DCC}->{staging_dbname}.ACCOUNTS accounts ON accounts.account_id = old_accounts.account_id
EndQuery

    $account_sth->execute();
    while (my $account = $account_sth->fetchrow_hashref) {
        if ((defined($idmap->{'__private_save'}->{hsl_accounts}) && $idmap->{'__private_save'}->{hsl_accounts}->{$account->{account_id}}) ||
            (!defined($idmap->{'__private_save'}->{hsl_accounts}) && F5::DbUtils::check_table_exists(dbh => $dbh,
                                                                                                     table_name => 'HSL_HTTP_PROFILES',
                                                                                                     dbname => $old_dbname))
                                                                                 && $self->was_hsl(dbh => $dbh,
                                                                                            account_id => $account->{account_id},
                                                                                            dbname => $old_dbname,)) {
            my ($do_remote_logging) = $dbh->selectrow_array(<<EndQuery,undef $account->{account_id});
SELECT flg_do_remote_logging
  FROM $old_dbname.HSL_PROFILES p JOIN $old_dbname.HSL_HTTP_PROFILES hp 
    ON p.hsl_profile_id = hp.hsl_profile_id
 WHERE account_id = ?
EndQuery
            next unless $do_remote_logging;
            $account->{logging_profile_id} = $HSL_LOGPROF_ID;
        }
        elsif (exists $account->{flg_all_requests}) {
            #If they have a flg_all_requests flag defined then they are old and just have All Requests or Only Illegal (F5::DbFixup::LoggingProfiles)
            $account->{logging_profile_id} = $account->{flg_all_requests} ? $LOG_ALL_LOGPROF_ID
                                                                          : $LOG_ILLEGAL_LOGPROF_ID;
        }
        
        next unless defined $account->{logging_profile_id};
        for my $policy_action (@{F5::ASM::COAPI::Query->new("policy_action", {target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM,
                                                                              action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE,
                                                                              policy => $account->{policy_name}})}) {
            $policy_logging_profiles{$policy_action->{policy_name}} = $logging_profile_hash{$account->{logging_profile_id}}->{logprof_name};
        }
    }

    # Assign profiles to VSs that had the policy assigned to them.
    my $virtual_servers = F5::ASM::COAPI::Query->new('virtual_server_policy');
    if ($virtual_servers && @$virtual_servers) {
        my %vs_log_profiles;
        for my $virtual_server (@$virtual_servers) {
            if (my $logging_profile = $policy_logging_profiles{$virtual_server->{policy_name}}) {
                $vs_log_profiles{$virtual_server->{vs_name}} = $logging_profile;
            }
        }
        for my $virtual_server (keys %vs_log_profiles) {
           my $fw_vs_log_profile = F5::ASM::COAPI::Object->new('fw_vs_log_profile');
           $fw_vs_log_profile->{virtual_server_name} = $virtual_server;
           $fw_vs_log_profile->{profile_name} = $vs_log_profiles{$virtual_server};
           $fw_vs_log_profile->save();

           if (exists $local_profiles_hash{$fw_vs_log_profile->{profile_name}})
           {
               $fw_vs_log_profile->{profile_name} = $local_profiles_hash{$fw_vs_log_profile->{profile_name}}->{local};
               $fw_vs_log_profile->save();
           }
       }
    }

    if (F5::DbUtils::check_table_exists(dbh => $dbh,
                                        table_name => 'HSL_HTTP_PROFILES',
                                        dbname => $old_dbname)) {

        ### Assign profiles to VSs that have PSM active
        my $logging_profiles = $dbh->selectall_hashref(<<EndQuery, "data_profile_name", undef, 1, 1);
SELECT data_profile_name
  FROM $old_dbname.HSL_PROFILES p JOIN $databases->{DCC}->{ucs_dbname}.HSL_DATA_PROFILES dp
    ON p.hsl_profile_id = dp.hsl_profile_id
 WHERE p.flg_do_remote_logging = ?
   AND dp.flg_is_enabled = ?
EndQuery
    
        if (%$logging_profiles) {
            my %PSM_TYPES = ( ftp  => F5::Mcp::ProfileType::PROFILE_FTP,
                              smtp => F5::Mcp::ProfileType::PROFILE_INSTANCE,);
                  
            my $vs_query = F5::ASM::COAPI::Query->new("virtual_server_profile" );  #vs_name, profile_name, profile_type
            if ($vs_query && @$vs_query) {
                my %psm_profiles;
                my $smtp_query = F5::ASM::COAPI::Query->new('ltcfg_instance_field', { class_name => 'profile_smtp', field_name => "security_enabled", value => 'true'});
                if ($smtp_query && @$smtp_query) {            
                    for my $instance (@$smtp_query) {
                        if ($logging_profiles->{$instance->{instance_name}}) {
                            $psm_profiles{$PSM_TYPES{smtp}}->{$instance->{instance_name}} = 1;
                        }
                    }
                }
                my $ftp_query = F5::ASM::COAPI::Query->new("profile_ftp", {security_enabled => 1}); # name, security_enabled
                if ($ftp_query && @$ftp_query) {
                    for my $profile (@$ftp_query) {
                        if ($logging_profiles->{$profile->{name}}) {
                            $psm_profiles{$PSM_TYPES{ftp}}->{$profile->{name}} = 1;
                        }
                    }
                }
                for my $virtual_server (@$vs_query) {
                    if ($psm_profiles{$virtual_server->{profile_type}} 
                        && $psm_profiles{$virtual_server->{profile_type}}->{$virtual_server->{profile_name}}) {
                        my $fw_vs_log_profile = F5::ASM::COAPI::Object->new('fw_vs_log_profile');
                        $fw_vs_log_profile->{virtual_server_name} = $virtual_server->{vs_name};
                        $fw_vs_log_profile->{profile_name} = $logging_profile_hash{$HSL_LOGPROF_ID}->{logprof_name};
                        $fw_vs_log_profile->save();
                    }
                }
            }
        }
    }
    $logger->debug("End Logging Profile fixup");
}

#----------------------------------------------------------------------
sub create_publisher {
    my ($logging_profile) = @_;
    my @objects;

    my $publisher = F5::ASM::COAPI::Object->new('log_publisher');
    #Currently there is no conflict with Dos + ASM, or PSM types.  
    #1. We do not create a PSM + DoS migration
    #2. ASM (application) and DoS logging use different object structures, so we are not limited by DoS' limitation of being only Splunk or Arcsight.
    $publisher->{name} = $logging_profile->{logprof_name} . "--publisher";  #Unique name with logging profile name.                                
    push (@objects, $publisher);

    my $pool = F5::ASM::COAPI::Object->new("pool");
    $pool->{name} = $logging_profile->{logprof_name};
    push (@objects, $pool);

    my $common_remote_logger = $logging_profile->{remote_loggers}->[0];
    for my $remote_logger (@{$logging_profile->{remote_loggers}}) {
        ## (pool -> pool_member) -> log_destination_remote_hsl -> log_destination_TYPE
        ## + (filters?) -> log_publisher
        ## -> log_publisher_destination
        my $pool_member = F5::ASM::COAPI::Object->new("pool_member");
        $pool_member->{node_name} = $remote_logger->{server_ip};
        $pool_member->{port} = $remote_logger->{server_port};
        $pool_member->{pool_name} = $logging_profile->{logprof_name};
        push (@objects, $pool_member);
    }

    my $remote_hsl = F5::ASM::COAPI::Object->new("log_destination_remote_hsl");
    $remote_hsl->{name} = $logging_profile->{logprof_name};
    $remote_hsl->{pool_name} = $logging_profile->{logprof_name};
    $remote_hsl->{protocol} = $destination_protocol_map{$common_remote_logger->{protocol}};
    push (@objects ,$remote_hsl);

    my $destination_type = $remote_storage_destination_map{$logging_profile->{remote_storage_type}};
    my $destination = F5::ASM::COAPI::Object->new($destination_type);
    $destination->{name} = $logging_profile->{logprof_name} . "--destination";

    # MCPDERR_LOG_DESTINATION_FORWARD_TO_EMPTY is thrown at tm_daemon/mcp/unit_test/validation/TestLogDestination.h
    #    in case forwarding destination is empty for any one of: syslog, arcsight, splunk
    # thus, we set 'forward_to' in all these cases, to avoid the error
    #    01071667:3: Log destination (/Common/<SOME_NAME>_remote_logging--destination) must specify a forwarding destination.
    if ($destination_type eq "log_destination_remote_syslog") {
        $destination->{default_facility} = ($common_remote_logger->{facility} - 16);  # Enum in MCP starts at 0, ASMconfig enum started at 16;
        $destination->{default_severity} = $SYSLOG_LEVEL_INFO;
        $destination->{remote_hsl_name} = $remote_hsl->{name};
        $destination->{forward_to} = $remote_hsl->{name};
    }
    elsif ($destination_type eq "log_destination_splunk"
            || $destination_type eq "log_destination_arcsight"
            || $destination_type eq "log_destination_bigiq") {
        $destination->{forward_to} = $remote_hsl->{name};
    }
    push (@objects, $destination);

    my $pub_destination = F5::ASM::COAPI::Object->new('log_publisher_destination');
    $pub_destination->{log_publisher_name} = $publisher->{name};
    $pub_destination->{log_destination_name} = $destination->{name};
    push (@objects, $pub_destination);

    return @objects;
}

#----------------------------------------------------------------------

1;
