package F5::DbFixup::NegsigSetsUserDefined;

use base 'F5::DbFixup::Default';

use F5::DbUtils;

# Package globals, handled by AUTOLOAD
our $priority = 98; # must run before SetNegsigSystemId
our $min_version = '0009.0004.0002.0000.0000';
our $fixup_db = 'PLC';
our @set_id_tables = (
                      'PLC.NEGSIG_SET_FILTERS',
                      'PLC.NEGSIG_SET_SIGNATURES',
                      'PLC.NEGSIG_SET_FILTER_SYSTEMS',
                     );
our $excluded_tables = [
                        'PLC.NEGSIG_SETS',
                        @set_id_tables,
                       ];

sub after_fixup {
  my($self,$dbh,$old_dbname,$new_dbname, $logger, $params) = @_;
  my $databases = $params->{databases};
  $logger->debug("Begin NegsigSetsUserDefined fixup");

  my $userdef_sigset_names = $dbh->selectall_hashref(<<EndQuery, 'set_id');
SELECT set_id, set_name
  FROM $old_dbname.NEGSIG_SETS
 WHERE flg_is_user_defined = 1
EndQuery

  my @user_defined_ids = keys %$userdef_sigset_names;

  return unless @user_defined_ids;

  check_name_collisions(dbh => $dbh,
                        old_dbname => $old_dbname,
                        new_dbname => $new_dbname,
                        logger => $logger,
                        userdef_sigset_names => $userdef_sigset_names,);

  my %copy_table_params = (
                           dbh => $dbh,
                           old_dbname => $old_dbname,
                           new_dbname => $new_dbname,
                           logger => $logger,
                           where_clause => 'set_id IN (' . join(',', @user_defined_ids) . ')',
                           idmap => $params->{idmap},
                          );
  my $idmap_out;
  if (grep($_ < 300000, @user_defined_ids)) {
    $idmap_out = {};
    $copy_table_params{idmap_out} = $idmap_out;
  }
  else {
    $copy_table_params{preserve_id} = 1;
  }

  $self->copy_table(
                    table_name => 'NEGSIG_SETS',
                    id_field_name => 'set_id',
                    %copy_table_params,
                   );

  if ($idmap_out) {
    $copy_table_params{idmap}->{set_id} = $idmap_out;
    delete $copy_table_params{idmap_out};

    # remap policy sets
    for my $old_id (keys %$idmap_out) {
      $dbh->do(<<EndQuery, undef, $idmap_out->{$old_id}, $old_id);
UPDATE $new_dbname.PL_POLICY_NEGSIG_SETS
   SET set_id = ?
 WHERE set_id = ?
EndQuery
    }

    $params->{external_idmap} = $idmap_out;
  }

  for my $excluded_table (@set_id_tables) {
    my($table_name) = ($excluded_table =~ /^.*?\.(.*)$/);

    if ($table_name eq 'NEGSIG_SET_FILTER_SYSTEMS' &&
        grep($_ eq 'filter_id', F5::DbUtils::get_db_fields($dbh, "$old_dbname.$table_name"))) {
      copy_filter_systems(%copy_table_params);
    }
    else {
      $self->copy_table(
                        table_name => $table_name,
                        id_field_name => '',
                        %copy_table_params,
                       );
    }
  }
  $logger->debug("End NegsigSetsUserDefined fixup");
}

#----------------------------------------------------------------------

sub check_name_collisions {
  my(%params) = @_;
  my $dbh = $params{dbh};
  my $old_dbname = $params{old_dbname};
  my $new_dbname = $params{new_dbname};
  my $logger = $params{logger};
  my $userdef_sigset_names = $params{userdef_sigset_names};

  my $exist_sigset_names = $dbh->selectall_hashref(<<EndQuery, 'set_name');
SELECT set_name
  FROM $new_dbname.NEGSIG_SETS
EndQuery

  my @rename_sigsets;
  for my $sigset ( values %$userdef_sigset_names ){
    if ($exist_sigset_names->{$sigset->{set_name}}) {
        push @rename_sigsets, {set_id => $sigset->{set_id}, set_name => $sigset->{set_name}};
    }
  }

  for my $sigset (@rename_sigsets) {
      my $set_name = $sigset->{set_name};
      my $suffix_count = 1;
      my $new_set_name = $set_name. '_'. $suffix_count;
      while($exist_sigset_names->{$new_set_name}) {
          ++$suffix_count;
          $new_set_name = $set_name. '_'. $suffix_count;
      }
      $logger->info("Update name for new user defined signature set: $new_set_name");
      $exist_sigset_names->{$new_set_name} = 1;
      $dbh->do(<<EndQuery, undef, $new_set_name, $sigset->{set_id});
UPDATE $old_dbname.NEGSIG_SETS
   SET set_name = ?
 WHERE set_id = ?
EndQuery
  }

}

#----------------------------------------------------------------------

sub copy_filter_systems {
  my(%params) = @_;
  my $dbh = $params{dbh};
  my $old_dbname = $params{old_dbname};
  my $new_dbname = $params{new_dbname};

  my $filter_id_sql = <<EndQuery;
SELECT sets.set_id, filter_systems.system_id
  FROM $old_dbname.NEGSIG_SETS sets,
       $old_dbname.NEGSIG_SET_FILTERS filters,
       $old_dbname.NEGSIG_SET_FILTER_SYSTEMS filter_systems
 WHERE sets.set_id = filters.set_id
   AND filters.filter_id = filter_systems.filter_id
   AND sets.flg_is_user_defined = 1
EndQuery

  my $filter_id_sth =$dbh->prepare($filter_id_sql);
  $filter_id_sth->execute;

  while (my $row = $filter_id_sth->fetchrow_hashref) {
    if ($params{idmap} && $params{idmap}->{set_id} &&
        $params{idmap}->{set_id}->{$row->{set_id}}) {
        $row->{set_id} = $params{idmap}->{set_id}->{$row->{set_id}};
    }
    F5::DbUtils::insert_row(dbh => $dbh,
                            table_name => "$new_dbname.NEGSIG_SET_FILTER_SYSTEMS",
                            rec => $row);
   }
}

1;

