package F5::DbFixup::PsmProfileMigration;
use warnings;
use strict;

use F5::DbUtils;
use F5::ASM::COAPI;
use F5::ASMConfig::Entity::AllowedMethod;

our @ISA = qw(F5::DbFixup::Default);

# Package globals, handled by AUTOLOAD
our $max_version = '0011.0005.0000.0000.0000';
our $fixup_db = 'PLC';
our $priority = 1000;

my %VIOLATION_TYPE_NAME = (
    method => 'ILLEGAL_METHOD',
    file_type => 'OBJ_TYPE',
    evasion_techniques => 'EVASION_DETECTED',
    mandatory_header => 'MISSING_MANDATORY_HEADER',
    http_rfc => 'HTTP_SANITY_CHECK_FAILED',
    maximum_length => 'QS_LEN',
);

my %RFC_NAME_CODE = (
    post_with_zero_length => 1,
    header_name_without_value => 2,
    several_content_length_headers => 8,
    chunked_with_content_length => 16,
    body_in_get_head => 64,
    host_header_is_ip => 2048,
    content_length_is_positive => 4096,
    bad_version => 8192,
    null_in_body => 16384,
    null_in_headers => 16384,
    high_ascii_in_headers => 32768,
    unparsable_content => 65536,
    maximum_headers => 131072,
    bad_host_header => 262144,
);

my %RESPONSE_TYPE_CODE = (
    default  => 0,
    custom   => 1,
    redirect => 2,
    'soap fault' => 3,
);

my %LEN_DB_FIELDS_REMAP = (
    request => 'total_len',
    query_string => 'qs_len',
    post_data => 'pd_len',
    uri => 'uri_len',
);

my %RESP_DB_FIELDS_REMAP = (
    alternate_response_header => 'response_header',
    alternate_response_content => 'response_body',
);

my %DB_FIELDS_TODELETE = (
    policy_id => 1,
    hsl_profile_id => 1,
    account_id => 1,
);

## PLC changes taken into consideration in fixup below:
###########################################################
# PL_POLICIES: 'flg_is_case_insensitive' added   in v11.0.0
# PL_BLOCKING: 'code' => 'viol_index'            in v11.3.0
# HSL_HTTP_PROFILES: 'account_id' => 'policy_id' in v11.4.0
# PL_MANDATORY_HEADERS => PL_HEADERS             in v11.4.0
# PL_ALTERNATE_RESPONSES: 'redirect_url' added   in v11.4.0
# VIOLATIONS added                               in v11.5.0
###########################################################

sub after_fixup {
    my ($self,$dbh,$old_dbname,$new_dbname,$logger,$params) = @_;
    $logger->debug("Begin Psm Profile fixup");

    my $databases = $params->{databases};
    my $old_version = $params->{old_version};

    # Psm was always in Common in versions before $max_version
    my $partition = 'Common'; 

    # the migrated profile names will get a suffix added to their original names
    my $suffix = '_migrated_httpsecurity_profile';

    if (     F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'HSL_HTTP_PROFILES') 
          && F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'HSL_PROFILES')       ) {

        # get the name of the httpsec system default profile
        my $default_httpsec_prof =
            F5::ASM::COAPI::Query->new(q{profile_httpsecurity} , { config_source => 1 })->[0]->{name}
            || $logger->warn("Could not figure out the system default httpsecurity profile name");

        # in v11.4.0 we switched from 'account_id' to 'policy_id' in PLC.HSL_HTTP_PROFILES.
        # can't use PLC_staging.HSL_HTTP_PROFILES cause PLC.HSL_HTTP_PROFILES is deleted in v11.5.0.
        # thus we are forced to choose the right field ('account_id'/'policy_id') based on ucs version.
        my $id_field_name = ($old_version lt '0011.0004.0000.0000.0000') ? 'account_id' : 'policy_id' ;

        # Find all http security profiles, their ids and their assosiated policy ids
        my $hsl_http_profiles = $dbh->selectall_arrayref(<<HslProfQuery, { Slice => {} });
SELECT *
  FROM $old_dbname.HSL_PROFILES a
  JOIN $old_dbname.HSL_HTTP_PROFILES b 
    ON a.hsl_profile_id = b.hsl_profile_id
HslProfQuery

        unless ($hsl_http_profiles && @$hsl_http_profiles) {
            $logger->debug("Could not find any Psm Profiles to fixup.");
            return;
        }

        ## for each of the http sec profs, we do:
        ## fetch info from db, migrate to mcp and delete info from db
        for my $hsl_http_profile (@$hsl_http_profiles) {

            # if we used account_id above - extract policy_id info based on it
            if ($id_field_name eq 'account_id') {
                my $pl_plc = $dbh->selectrow_hashref(<<PlcAcc, undef, $hsl_http_profile->{account_id});
SELECT id
  FROM $old_dbname.PL_POLICIES
 WHERE account_id = ?
PlcAcc
                $hsl_http_profile->{policy_id} = $pl_plc->{id};
            }

            # Start constructing the COAPI object
            my $coapi_profile = F5::ASM::COAPI::Object->new('profile_httpsecurity');

            # all migrated profiles are migrated as user defined profiles.
            # the former default httpsec profile is also migrated as a user defined profile
            # thus, negate default profile bit and set parent prof to the httpsec system default prof
            $coapi_profile->set_value('config_source',0);
            $coapi_profile->set_value('default_name', $default_httpsec_prof);

            ## These fields are depending only on partitions, and profile names 
            #-----------------------------------------------------------------------
            $coapi_profile->set_value('folder_name',"/$partition");
            $coapi_profile->set_value('leaf_name',$hsl_http_profile->{hsl_profile_name} . $suffix);
            $coapi_profile->set_value('name',"/$partition/$hsl_http_profile->{hsl_profile_name}" . $suffix);
            $coapi_profile->set_value('partition_id',$partition);
            $coapi_profile->set_value('description',"This profile was created by upgrade migration process from the original HTTP Security Profile named: [$hsl_http_profile->{hsl_profile_name}]");
            #-----------------------------------------------------------------------

            ## case sensitivity and four Length Checks
            # PLC.PL_POLICIES.flg_is_case_insensitive was added in v11.0.0.
            # Thus using PLC_staging.PL_POLICIES to extract info
            #-----------------------------------------------------------------------
            my $hsl_plc_acc = $dbh->selectrow_hashref(<<HslPolicyQuery, undef, $hsl_http_profile->{policy_id});
SELECT *
  FROM $databases->{PLC}->{ucs_dbname}.PL_POLICIES 
 WHERE id = ?
HslPolicyQuery

            # this is used later on
            my $http_rfc_maximum_headers_count = defined($hsl_plc_acc->{max_headers}) ? $hsl_plc_acc->{max_headers} : 20;
            $hsl_http_profile->{account_id} = $hsl_plc_acc->{account_id};

            my $hsl_plc_flg_is_case_insensitive = defined($hsl_plc_acc->{flg_is_case_insensitive}) ? $hsl_plc_acc->{flg_is_case_insensitive} : 0;
            $coapi_profile->set_value('case_sensitive',!$hsl_plc_flg_is_case_insensitive);
            for my $len_type (keys %LEN_DB_FIELDS_REMAP) {
                $coapi_profile->set_value("maximum_length_$len_type",$hsl_plc_acc->{$LEN_DB_FIELDS_REMAP{$len_type}});
            }
            #-----------------------------------------------------------------------

            # set alarm, block flags for the 6 viols listed in %VIOLATION_TYPE_NAME above
            # we take advantage of the fact that all four Length Checks in httpsec profile 
            #    have same alarm/block flag setting
            # thus, we only listed 'QS_LEN' in 'maximum_length' at %VIOLATION_TYPE_NAME above
            #
            # PLC.PL_BLOCKING.code field was replaced with viol_index field in v11.3.0.
            # Thus using PLC_staging.PL_BLOCKING to extract info.
            # PLC.VIOLATIONS was added in v11.5.0. Thus using it to extract info.
            #-----------------------------------------------------------------------
            for my $viol_type (keys %VIOLATION_TYPE_NAME) {
                my $viol = $dbh->selectrow_hashref(<<LenViol, undef, $VIOLATION_TYPE_NAME{$viol_type}, $hsl_http_profile->{policy_id});
SELECT b.flg_alarm, b.flg_reject
  FROM $databases->{PLC}->{bigip_dbname}.VIOLATIONS a 
  JOIN $databases->{PLC}->{staging_dbname}.PL_BLOCKING b 
    ON a.viol_index = b.viol_index 
 WHERE icontrol_name = ?
   AND policy_id = ?
LenViol

                $coapi_profile->set_value("${viol_type}_alarm",$viol->{flg_alarm});
                $coapi_profile->set_value("${viol_type}_block",$viol->{flg_reject});
            }
            #-----------------------------------------------------------------------

            ## HTTP Protocol Checks (RFC)
            #-----------------------------------------------------------------------
            my $http_checks = $dbh->selectall_arrayref(<<HttpCheck, { Slice => {} }, $hsl_http_profile->{policy_id});
SELECT flg_is_enabled, http_check_code 
  FROM $old_dbname.PL_HTTP_CHECKS 
 WHERE policy_id = ?
HttpCheck

            my %hsl_http_check = ();

            for my $http_check (@$http_checks) {
                $hsl_http_check{$http_check->{http_check_code}} = $http_check->{flg_is_enabled};
            }
            for my $check_name (keys %RFC_NAME_CODE) {
                $coapi_profile->set_value("http_rfc_$check_name", $hsl_http_check{$RFC_NAME_CODE{$check_name}});
            }

            # only if maximum_headers rfc check is turned ON - set the muneric maximu value to it
            if ($coapi_profile->{http_rfc_maximum_headers}) {
                $coapi_profile->set_value('http_rfc_maximum_headers',$http_rfc_maximum_headers_count);
            }
            #-----------------------------------------------------------------------

            ## [Dis]Allowed File Types
            ## we take advantage of the fact that all have same setting:
            ##     Allowed OR DisAllowed (so we look only into the flag of the first one)
            #-----------------------------------------------------------------------
            my $file_types = $dbh->selectall_arrayref(<<FileType, { Slice => {} }, $hsl_http_profile->{policy_id},'*');
SELECT name, flg_is_allowed
  FROM $old_dbname.PL_FILETYPES
 WHERE policy_id = ?
   AND name != ?
FileType

            if ($file_types && @$file_types) {
                $coapi_profile->set_value('file_type_allowed',$file_types->[0]->{flg_is_allowed});
                $coapi_profile->{file_types} = [ map { $_->{name} } @$file_types ];
            }
            #-----------------------------------------------------------------------

            ## Mandatory HEADERS
            # PLC.PL_MANDATORY_HEADERS replaced by PLC.PL_HEADERS in v11.4.0.
            # Using PLC_staging.PL_HEADERS to extract info did not work though.
            # Thus I'm forced to select a db name via version - this works fine
            #-----------------------------------------------------------------------
            my ($headers_and_clause, $headers_db_table);
            if ($old_version lt '0011.0004.0000.0000.0000') {
                $headers_and_clause = '';
                $headers_db_table = 'PL_MANDATORY_HEADERS';
            } else {
                $headers_and_clause = 'AND flg_is_mandatory = 1';
                $headers_db_table = 'PL_HEADERS';
            }

            my $headers = $dbh->selectall_arrayref(<<FileType, { Slice => {} }, $hsl_http_profile->{policy_id});
SELECT header_name
  FROM $old_dbname.$headers_db_table
 WHERE policy_id = ?
 $headers_and_clause
FileType

            $coapi_profile->{mandatory_headers} = [ map { $_->{header_name} } @$headers ]
                if $headers && @$headers;
            #-----------------------------------------------------------------------

            ## Allowed METHODS
            #-----------------------------------------------------------------------
            my $methods = $dbh->selectall_arrayref(<<Methods, { Slice => {} }, $hsl_http_profile->{policy_id});
SELECT method_code 
  FROM $old_dbname.PL_ALLOWED_METHODS 
 WHERE policy_id = ?
Methods

            my @methods_arr = ();
            for ($methods && @$methods) {
                my $allowed = F5::ASMConfig::Entity::AllowedMethod->get_method(code => $_->{method_code});
                push @methods_arr , $allowed->{leaf_name} if defined $allowed; # this can be undefined, apparently...
            }
            $coapi_profile->{methods} = \@methods_arr if @methods_arr;
            #-----------------------------------------------------------------------

            ## Custom blocking page
            # PLC.PL_ALTERNATE_RESPONSES.redirect_url was added in v11.4.0.
            # Using PLC_staging.PL_ALTERNATE_RESPONSES to extract info did not work though.
            # Thus I'm forced to construct the select clause via version - this works fine
            #-----------------------------------------------------------------------
            my @db_fields = qw(response_type alternate_response_header alternate_response_content);
            unless ($old_version lt '0011.0004.0000.0000.0000') {
                push (@db_fields, 'redirect_url');
            }
            my $select_fields = join(',', @db_fields);
            my $response = $dbh->selectrow_hashref(<<Response, undef, $hsl_http_profile->{policy_id});
SELECT $select_fields
  FROM $old_dbname.PL_ALTERNATE_RESPONSES
 WHERE cause = 'default'
   AND policy_id = ?
Response

            # set response type
            $coapi_profile->set_value('response_type',$RESPONSE_TYPE_CODE{$response->{response_type}});

            # If type is redirect - DO NOT set 'response_[header|body]'. Only set 'redirect_url'.
            if ($response->{response_type} eq 'redirect') {
                my ($url) = ($response->{redirect_url}) 
                    ? $response->{redirect_url}
                    : $response->{alternate_response_header} =~ m/Location: (.*?)(\r|\n)/ ;
                $coapi_profile->set_value('redirect_url',$url);
            } else {
                for my $brp_type (keys %RESP_DB_FIELDS_REMAP) {
                    next unless $response->{$brp_type};
                    # get rid of all \r. Why? See comments here - 
                    # ../tmos-security-asmdp/tm_daemon/mcp/firewall/validation/ProfileHTTPSecurity.cpp#38
                    $response->{$brp_type} =~ s/\r//g;
                    $coapi_profile->set_value($RESP_DB_FIELDS_REMAP{$brp_type},$response->{$brp_type});
                }
            }
            #-----------------------------------------------------------------------

            #All done. Save the MCP object.
            #-----------------------------------------------------------------------
            $coapi_profile->save();
            #-----------------------------------------------------------------------

            # Assign profiles to VSs
            #-----------------------------------------------------------------------
            my $virtuals = $dbh->selectall_arrayref(<<Virtuals, { Slice => {} }, $hsl_http_profile->{hsl_profile_id});
SELECT virtual_server_name
  FROM $databases->{DCC}->{ucs_dbname}.HSL_DATA_PROFILES a 
  JOIN $databases->{DCC}->{ucs_dbname}.HSL_VIRTUAL_SERVERS b 
    ON a.data_profile_name = b.data_profile_name 
 WHERE flg_is_enabled = 1 
   AND hsl_profile_id = ?
Virtuals

            if ($virtuals && @$virtuals) {
                for my $virtual (@$virtuals) {
                    # this VS along with the data profile were already created and linked by now
                    # just assign the httpsecurity profile that we created above to that VS

                    my $server_profile = F5::ASM::COAPI::Object->new('virtual_server_profile');
                    $server_profile->{vs_name}      = $virtual->{virtual_server_name};
                    $server_profile->{profile_name} = $coapi_profile->{name};
                    $server_profile->save();
                }
            }
            #-----------------------------------------------------------------------

            # Assign the predefined hsl remote logging profile to relevant VSs
            # This logging profile, along with the required destination(s) and publisher(s)
            #     should have already been created by F5::DbFixup::LoggingProfileMigration.
            #
            # This is only required if old_version is less then 11.3.0.
            # Starting 11.3.0 and on 'LOGPROF_LOGGING_PROFILES' does not exist because all
            #     logging profile config was moved to the 'High Speed Logging' infrastructure in MCP.
            #-----------------------------------------------------------------------
            if ( $hsl_http_profile->{flg_do_remote_logging} && 
                 F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'LOGPROF_LOGGING_PROFILES') ) {

                my $log_prof = $dbh->selectrow_hashref(<<LogProf);
SELECT logprof_name 
  FROM $old_dbname.LOGPROF_LOGGING_PROFILES 
 WHERE target_system = 'hsl' 
LogProf

                if ($log_prof->{logprof_name} && $virtuals && @$virtuals) {
                    $log_prof->{logprof_name} =~ s/ /_/g;
                    for my $virtual (@$virtuals) {
                        my $fw_vs_log_profile = F5::ASM::COAPI::Object->new('fw_vs_log_profile');
                        $fw_vs_log_profile->{virtual_server_name} = $virtual->{virtual_server_name};
                        $fw_vs_log_profile->{profile_name} = "/$partition/$log_prof->{logprof_name}";
                        $fw_vs_log_profile->save();
                    }
                }
            }
            #-----------------------------------------------------------------------

            ## set trust-xff and Custom Headers
            #-----------------------------------------------------------------------
            if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'PL_POLICY_CUSTOM_HEADERS') &&
                F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'PL_POLICY_XFF')
               ) {
                # Look for custom headers, can be many
                my $xff_headers = $dbh->selectall_arrayref(<<XffHead, { Slice => {} }, $hsl_http_profile->{policy_id});
SELECT header_name
  FROM $old_dbname.PL_POLICY_CUSTOM_HEADERS
 WHERE policy_id = ?
XffHead

                # Look for trust xff flag, can be only one
                my $xff_flag = $dbh->selectrow_hashref(<<XffFlag, undef, $hsl_http_profile->{policy_id});
SELECT flg_is_trust_xff
  FROM $old_dbname.PL_POLICY_XFF
 WHERE policy_id = ?
XffFlag

                # if found custom headers OR trust xff flag is TRUE on this policy - 
                #     we need to set these on all data profiles that assosiated with this policy.
                # Why? 
                #     We could have a single http_sec profile (and thus a single policy) 
                #     assosiated with multiple http data profiles.
                #     Each of these data profs assigned to other VS.
                #     That single policy has trust_xff flag and custom headers set in it.
                #     So, we need to migrate these into all the relevant data profiles
                if (($xff_headers && @$xff_headers) || $xff_flag->{flg_is_trust_xff}) {
                    my $data_profiles = $dbh->selectall_arrayref(<<DataProf, { Slice => {} }, $hsl_http_profile->{$id_field_name});
SELECT data_profile_name
  FROM $databases->{DCC}->{ucs_dbname}.HSL_DATA_PROFILES a
  JOIN $old_dbname.HSL_HTTP_PROFILES b
    ON a.hsl_profile_id = b.hsl_profile_id
 WHERE flg_is_enabled = 1
   AND $id_field_name = ?
DataProf

                    if ($data_profiles && @$data_profiles) {
                        for my $data_prof (@$data_profiles) {
                            # these data profiles are already created by COAPI, we just update them.
                            my $profile_http = F5::ASM::COAPI::Object->new('profile_http');
                            $profile_http->{name} = $data_prof->{data_profile_name};
                            $profile_http->{accept_xff} = $xff_flag->{flg_is_trust_xff};
                            $profile_http->{xff_alternative_names} = [ map { $_->{header_name} } @$xff_headers ] if @$xff_headers;
                            $profile_http->save();
                        }
                    }
                }
            }
            #-----------------------------------------------------------------------

            ## clean out HSL data from [PLC|DCC]_staging for currently handeled profile
            #-----------------------------------------------------------------------
            for my $db (keys %$databases) {
                my @tables = F5::DbUtils::get_config_sync_tables($db);
                for my $table (@tables) {
                    $table =~ s/^$db/$databases->{$db}->{staging_dbname}/;
                    my @all_fields = F5::DbUtils::get_db_fields($dbh,$table);
                    for my $field (@all_fields) {
                        next unless exists $DB_FIELDS_TODELETE{$field};
                        $dbh->do("DELETE FROM $table WHERE $field = ?",
                            undef, $hsl_http_profile->{$field});
                    }
                }
            }
            #-----------------------------------------------------------------------

        }
    }

    $logger->debug("End Psm Profile fixup");
}

1;

