package F5::DbFixup::ScheduledReportsMigration;

use warnings;
use strict;

use F5::DbUtils;
use F5::ASM::COAPI;
use PHP::Serialization qw(unserialize);
use MIME::Base64;

use base 'F5::DbFixup::Default';

# Package globals, handled by AUTOLOAD
our $fixup_db = 'DCC';
our $priority = 100;
our $max_version = '0011.0003.0000.0000.0000';

# GUI ASM entity names => monpd ASM entity names
my %entity_map = ( 'account_id'                => 'asm_repev_security_policy',
                   'attack_types_status'       => 'asm_repev_attack_type',
                   'status'                    => 'asm_repev_violation',
                   'uri'                       => 'asm_repev_url',
                   'src_ip'                    => 'asm_repev_ip',
                   'geo_location_country_code' => 'asm_repev_geo',
                   'severity_id'               => 'asm_repev_severity',
                   'response_code'             => 'asm_repev_response_code',
                   'flg_req_blocked'           => 'asm_repev_request_type',
                   'method_code'               => 'asm_repev_method',
                   'app_id'                    => 'asm_repev_protocol',
                   'virus_name'                => 'asm_repev_virus',
                   'username'                  => 'asm_repev_username',
                   'session_id'                => 'asm_repev_session',
                   'iprep_threat'              => 'asm_repev_ip_reputation',
                  );
                  

# GUI saved time range as an enumration istead of actual time range number
# See https://indexing.f5net.com/source/xref/bigip11.2.0/TrafficShield/gui/dms/policy/reporting_commons.php#getTimeOffset
my %filter_time_map = ( 0  => 31536000, # Last year
                        1  => 3600,     # Last hour
                        24 => 86400,    # Last day
                        7  => 604800,   # Last week
                        30 => 18144000, # Last month
                        -1 => 0,        # Custom time range
                      );

my $CHARTS_FILTER_TYPE_ID = 32;
my $MULTI_LEVEL_FILTER_TYPE_ID = 33;
my $USER_FILTER_ID = 0;                  

sub after_fixup {
    my ($self,$dbh,$old_dbname,$new_dbname, $logger, $params) = @_;
    my $databases = $params->{databases};

    $logger->debug("Begin Scheduled reports fixup");
 
    # If the REPORTING_SCHEDULE table exists the FILTERS table is guranteed to exist as well.
    if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  
        table_name => 'REPORTING_SCHEDULE')) {
  
        # --------------------------------------------------------------------
        # Migrate user defined filters
        # --------------------------------------------------------------------

        my $user_defined_filters = $dbh->selectall_arrayref(
            "SELECT filter_id,filter_name
            FROM $databases->{DCC}->{ucs_dbname}.FILTERS
            WHERE page_id=? AND filter_type=? ", { Slice => {} }, $CHARTS_FILTER_TYPE_ID, $USER_FILTER_ID);
            

        for my $user_defined_filter (@$user_defined_filters) {
            
            my $analytics_predefined_filter = F5::ASM::COAPI::Object->new('analytics_predefined_report');
           
            # Replace all '/' occurences in oid, since they represent folders in MCP
            $analytics_predefined_filter->{description} = $user_defined_filter->{filter_name};
            $analytics_predefined_filter->{name} = $analytics_predefined_filter->{description};
            $analytics_predefined_filter->{name} =~ s/\//\\/g;
            
            my @drilldown;
            my $policy_drilldown_added = 0;

            my $filter_custom_data = $dbh->selectall_arrayref(
                "SELECT filter_item,item_value
                FROM  $databases->{DCC}->{ucs_dbname}.FILTER_CUSTOM_DATA
                WHERE filter_id=?",
                { Slice => { } }, $user_defined_filter->{filter_id});
               
            #
            # Fill filter attributes
            #
            for my $filter_custom_data_row (@$filter_custom_data) {
                migrate_filter_data(filter_row => $filter_custom_data_row,
                                    stored_report_request => $analytics_predefined_filter,
                                    drilldown => \@drilldown,
                                    policy_drilldown_added => \$policy_drilldown_added, 
                                    db_params => $params, 
                                    dbh => $dbh);                
            }

            if (@drilldown) {
                $analytics_predefined_filter->{drilldown} = \@drilldown;
            }
            $analytics_predefined_filter->{measures} = ['events_count'];
            $analytics_predefined_filter->save();
        }

        # --------------------------------------------------------------------
        # Migrate scheduled reports
        # --------------------------------------------------------------------

        # Create a dummy SMTP server configuration object. It will be filled with correct values in the PLC DB fixup later on.
        my $smtp_conf_mcp = F5::ASM::COAPI::Object->new('smtp_config');
        $smtp_conf_mcp->{name} = 'asm_smtp_conf';
        $smtp_conf_mcp->save();

        my $scheduled_reports = $dbh->selectall_arrayref("SELECT * FROM $databases->{DCC}->{ucs_dbname}.REPORTING_SCHEDULE", { Slice => {} });

        for my $scheduled_report (@$scheduled_reports) {
            my $analytics_scheduled_report = F5::ASM::COAPI::Object->new('analytics_scheduled_report');
            $analytics_scheduled_report->{name} = "$scheduled_report->{title}";
            $analytics_scheduled_report->{name} =~ s/\//\\/g;
            $analytics_scheduled_report->{smtp_config_name} = $smtp_conf_mcp->{name};
            $analytics_scheduled_report->{frequency} = $scheduled_report->{schedule_interval};
            $analytics_scheduled_report->{last_sent} = $scheduled_report->{last_sent};
            $analytics_scheduled_report->{first_send} = $scheduled_report->{last_sent};
            $analytics_scheduled_report->{next_send} = $scheduled_report->{last_sent} + $scheduled_report->{schedule_interval};
            
            # Avoid scheduling in the past
            while($analytics_scheduled_report->{next_send} < time()) {
                $analytics_scheduled_report->{next_send} = $analytics_scheduled_report->{next_send} + $scheduled_report->{schedule_interval}; 
            }
            
            # Migrate email addresses
            my $sched_email_addrs = $dbh->selectall_arrayref(
                "SELECT email FROM $databases->{DCC}->{ucs_dbname}.REPORTING_SCHEDULE_EMAILS
                 WHERE reporting_schedule_id=?", { Slice => {} }, $scheduled_report->{reporting_schedule_id});
                           
            my @addrs;
            for my $sched_email_addr (@$sched_email_addrs) {
                push(@addrs,$sched_email_addr->{email});
            }
            $analytics_scheduled_report->{email_addresses} = \@addrs if (@addrs);
                        
            #            
            # filter details
            #
            
            my $scheduler_filter = $dbh->selectrow_hashref(
                "SELECT filter_id,page_id,filter_name
                FROM $databases->{DCC}->{ucs_dbname}.FILTERS
                WHERE filter_id=? ", undef, $scheduled_report->{filter_id});

            if($scheduler_filter) {
                if($scheduler_filter->{page_id} == $CHARTS_FILTER_TYPE_ID) { # Predefined / User filter 
                    my $filter_name = $scheduler_filter->{filter_name};
                    
                    # BZ 404455: predefined filter/report 'All' doesn't exist starting for 11.3.0 and has
                    # been replaced by the filter 'Top alerted and blocked policies'
                    # BZ 475816: The filter was again renamed in 11.5.0 from 'Top alerted and blocked policies' to
                    # 'Top alarmed and blocked policies', so we rename to it directly upon an upgrade.
                    if($filter_name eq 'All') {
                        $filter_name = 'Top alarmed and blocked policies';
                    #
                    # BZ 489682 filters renamed after 11.2.0
                    #
                    } elsif ($filter_name eq 'Top alerted and blocked policies') {
                        $filter_name = 'Top alarmed and blocked policies';
                    } elsif ($filter_name eq 'Top alerted URLs') {
                        $filter_name = 'Top alarmed URLs';
                    } elsif ($filter_name eq 'Top alerted applications') { 
                        $filter_name = 'Top alarmed policies';
                    } elsif ($filter_name eq 'Top blocked applications') {
                        $filter_name = 'Top blocked policies';
                    } elsif ($filter_name eq 'Top attackers in last hour' || 
                             $filter_name eq 'Top attackers in last day'  || 
                             $filter_name eq 'Top attackers in last week') {
                        $filter_name = 'Top attackers for alarmed requests';
                    } elsif ($filter_name eq 'Top applications with GET method') {
                        $filter_name = 'Top policies with GET method';
                    } elsif ($filter_name eq 'Top applications with POST method') {
                        $filter_name = 'Top policies with POST method';
                    } elsif ($filter_name eq 'Top applications with Response Code 200') {
                        $filter_name = 'Top policies with Response Code 200';
                    } elsif ($filter_name eq 'Top applications with Response Code 404') {
                        $filter_name = 'Top policies with Response Code 404';
                    }
                    $analytics_scheduled_report->{predefined_report_name} = $filter_name;
                }
                elsif($scheduler_filter->{page_id} == $MULTI_LEVEL_FILTER_TYPE_ID) { # Multileveled dynamic report
                
                    my $filter_custom_data = $dbh->selectall_arrayref(
                        "SELECT filter_item,item_value
                        FROM  $databases->{DCC}->{ucs_dbname}.FILTER_CUSTOM_DATA
                        WHERE filter_id=?", { Slice => {} }, $scheduler_filter->{filter_id});
                        
                    my @drilldown;
                    my $policy_drilldown_added = 0;
            
                    for my $filter_custom_data_row (@$filter_custom_data) {

                        migrate_filter_data(filter_row => $filter_custom_data_row,
                                            stored_report_request => $analytics_scheduled_report,
                                            drilldown => \@drilldown,
                                            policy_drilldown_added => \$policy_drilldown_added,
                                            db_params => $params,
                                            dbh => $dbh);
                    }
                    $analytics_scheduled_report->{drilldown} = \@drilldown if (@drilldown);

                    # Default measure events_count
                    $analytics_scheduled_report->{measures} = ['events_count'];
                }
            }
            
            $analytics_scheduled_report->save();
        }
        
        # --------------------------------------------------------------------
        # Delete migrated user defined reports
        # --------------------------------------------------------------------
        $dbh->do(<<EndQuery);
        DELETE
        FROM $new_dbname.FILTER_CUSTOM_DATA
        WHERE filter_id IN (SELECT filter_id 
                            FROM $new_dbname.FILTERS 
                            WHERE (page_id=$CHARTS_FILTER_TYPE_ID AND filter_type=$USER_FILTER_ID) OR page_id=$MULTI_LEVEL_FILTER_TYPE_ID)
EndQuery
 
        $dbh->do(<<EndQuery);
        DELETE
        FROM $new_dbname.FILTERS
        WHERE (page_id=$CHARTS_FILTER_TYPE_ID AND filter_type=$USER_FILTER_ID) OR page_id=$MULTI_LEVEL_FILTER_TYPE_ID
EndQuery
    }

    $logger->debug("End Scheduled reports fixup");   
}

#
# migrate_filter_data
# =============================================================================
# Migrates a single filter DB row to the MCP equivalent
#
sub migrate_filter_data {
    my (%params) = @_;

    my $filter_row = $params{filter_row};
    my $stored_report_request = $params{stored_report_request};
    my $drilldown = $params{drilldown};
    my $policy_drilldown_added = $params{policy_drilldown_added};
    my $db_params = $params{db_params};
    my $dbh = $params{dbh};    

    my $filter_item  = $filter_row->{filter_item};
    my $filter_value = $filter_row->{item_value};

    if($filter_item eq 'viewby') {
        $stored_report_request->{view_by} = $entity_map{$filter_value};
    } 
    elsif($filter_item eq 'drilldown' && $filter_value) {
        my $drilldown_hash_ref = unserialize(MIME::Base64::decode($filter_value));
        return unless (ref $drilldown_hash_ref eq 'HASH');                              
        for my $drilldown_key (keys %{$drilldown_hash_ref}) {

            my $drilldown_value = $drilldown_hash_ref->{$drilldown_key}->{value};

            if ($drilldown_key eq 'account_id') {
                next if ($$policy_drilldown_added || $drilldown_value eq '-1');       

                $$policy_drilldown_added = 1;
            }
                       
            my $drilldown_entity_name = $entity_map{$drilldown_key};
            my $drilldown_values = [];

            # Bitstring entities
            if ($drilldown_entity_name eq 'asm_repev_violation'   || 
                $drilldown_entity_name eq 'asm_repev_attack_type' ||
                $drilldown_entity_name eq 'asm_repev_ip_reputation') {

               my $bit_idx = 0;
               my $drilldown_bit_idx;
               while ($drilldown_value != 0) {
                   if ($drilldown_value & 1 == 1) {
                       $drilldown_bit_idx = $bit_idx;
                       if ($drilldown_entity_name eq 'asm_repev_violation') {

                           # Translate enum ID to viol index
                          my ($viol_index_result) = $dbh->selectrow_array(<<EndQuery, undef, $bit_idx);
SELECT viol_index
FROM $db_params->{databases}->{DCC}->{staging_dbname}.PARSER_STATUS_CODES
WHERE enum_id=?
EndQuery

                          $drilldown_bit_idx = $viol_index_result;
                       }
                       push(@$drilldown_values, $drilldown_bit_idx);
                   }
                   $drilldown_value >>= 1;
                   $bit_idx++;
               }
            } 
            else {
               push(@$drilldown_values, $drilldown_value);
            }
             
            my %drilldown_mcp = ( entity => $drilldown_entity_name,
                                  ids    => $drilldown_values, );
                                                
            push(@$drilldown, \%drilldown_mcp);
        }
    }
    elsif($filter_item eq 'filter_top') {
        $stored_report_request->{rows_num} = $filter_value;
    }
    elsif($filter_item eq 'filter_account') {
        return if ($$policy_drilldown_added || $filter_value eq '-1');
        $$policy_drilldown_added = 1;

        my %drilldown_mcp = ( entity => 'asm_repev_security_policy' ,
                              ids    => [$filter_value], );
        push(@$drilldown, \%drilldown_mcp);
    }
    elsif($filter_item eq 'filter_time') {
        $stored_report_request->{time_diff_from_now} = $filter_time_map{$filter_value};
    }
    elsif($filter_item eq 'filter_time_1') {
        $stored_report_request->{time_from} = $filter_value;
    }
    elsif($filter_item eq 'filter_time_2') {
        $stored_report_request->{time_to} = $filter_value;
    }
    elsif($filter_item eq 'multidrill' && $filter_value) {
        my $multidrill_array_ref = unserialize(MIME::Base64::decode($filter_value));
        
        my $drilldown_count = (scalar @$multidrill_array_ref)-1;
        if($drilldown_count > 0) {
            my @dynamic_report_entities;
            for(0..$drilldown_count-1) {
                push(@dynamic_report_entities, $entity_map{@$multidrill_array_ref[$_]});
            }
            $stored_report_request->{dynamic_report_entities} = \@dynamic_report_entities;
        }
        $stored_report_request->{view_by} = $entity_map{@$multidrill_array_ref[$drilldown_count]};
    }
}

1;
