package F5::DbFixup::SearchEngineMigration;
use warnings;
use strict;

use F5::DbUtils;
use F5::ASM::COAPI;

our @ISA = qw(F5::DbFixup::Default);

# Package globals, handled by AUTOLOAD
our $max_version = '0014.0001.0000.0000.0000';
our $fixup_db = 'DCC';
our $priority = 99;

# all user defined search engines have search_engine_id > 299999999
my $DEFAULT_SEARCH_ENGINE_MAX_ID = 299999999;

sub after_fixup {
    my ($self,$dbh,$old_dbname,$new_dbname, $logger, $params) = @_;

    my $idmap = $params->{idmap};
    my $databases = $params->{databases};

    $logger->debug("Begin Search Engine Migration fixup");

    if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $databases->{DCC}->{ucs_dbname},  table_name => 'SEARCH_ENGINES')) {

        # all but default search engines
        my $search_engines = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT user_agent, search_engine_name AS name, domain_name
  FROM $databases->{DCC}->{ucs_dbname}.SEARCH_ENGINES
 WHERE search_engine_id > $DEFAULT_SEARCH_ENGINE_MAX_ID
EndQuery

        my %mcp_signatures = F5::ASM::COAPI::Query->new('bot_defense_signature',
                                                         undef,
                                                         retrieve_as_hash => { key_field_name => 'name'},
                                                        );

        my @search_engines_to_add;
        if ($search_engines && @{$search_engines}) {
            for my $search_engine (@{$search_engines}) {
                $search_engine->{name} = '/Common/' . $search_engine->{name};
                if (! $mcp_signatures{$search_engine->{name}}) {
                    push (@search_engines_to_add, $search_engine);
                }
            }
        }

        my %signatures_to_add;
        for my $search_engine (@search_engines_to_add) {
            if ($search_engine->{user_agent} && ! $signatures_to_add{$search_engine->{name}}->{rule}) {
                $signatures_to_add{$search_engine->{name}}->{rule} = 'headercontent:"' . $search_engine->{user_agent}  . '"; useragentonly; nocase;';
            }
            if ($search_engine->{domain_name}) {
                $signatures_to_add{$search_engine->{name}}->{domain_names}->{$search_engine->{domain_name}} = 1;
            }
        }

        my @signature_objs = ();
        for my $signature_name (keys %signatures_to_add) {
            my $signature_obj = F5::ASM::COAPI::Object->new('bot_defense_signature');
            $signature_obj->set_value('name', $signature_name);
            $signature_obj->set_value('user_defined', 1);
            $signature_obj->set_value('category_name', '/Common/Search Engine');
            $signature_obj->set_value('class_name', '/Common/Trusted Bot');
            $signature_obj->set_value('rule', $signatures_to_add{$signature_name}->{rule}) if $signatures_to_add{$signature_name}->{rule};
            $signature_obj->set_value('domain_names', [ keys $signatures_to_add{$signature_name}->{domain_names} ]) if $signatures_to_add{$signature_name}->{domain_names};

            push (@signature_objs, $signature_obj);
        }

        if (@signature_objs) {
            eval {
                $logger->debug("About to save MCP transaction") if $logger;
                F5::ASM::COAPI::TransactionalSave(\@signature_objs);
                $logger->info("MCP transactional save complete") if $logger;
            };
            if ($@) {
                $logger->info("Cannot send objects to mcp: $@ $F5::COAPI::error") if $logger;
                return;
            }
        }
    }

    $logger->debug("End Search Engine Migration fixup");
}

#----------------------------------------------------------------------

1;
