package F5::DbFixup::SmtpConfiguration;
use warnings;
use strict;

use F5::DbUtils;
use F5::ASM::COAPI;
use PassCrypt;

use base 'F5::DbFixup::Default';

# Package globals, handled by AUTOLOAD
our $fixup_db = 'PLC';
our $priority = 99;
our $min_version = '0010.0001.0000.0000.0000';
our $max_version = '0011.0003.0000.0000.0000';

my %encryption_type_map = ( ''   => 'SMTP_AUTHENTICATION_ENCRYPTION_ALGORITHM_NONE',
                           'ssl' => 'SMTP_AUTHENTICATION_ENCRYPTION_ALGORITHM_SSL',
                           'tls' => 'SMTP_AUTHENTICATION_ENCRYPTION_ALGORITHM_TLS',
                         );
                       
sub after_fixup {
    my ($self,$dbh,$old_dbname,$new_dbname, $logger, $params) = @_;
    my $databases = $params->{databases};

    $logger->debug("Begin SmtpConfiguration fixup");
 
    # Gather old info. There should be only a single row in 
    # the SMTP server settings table.
    if (F5::DbUtils::check_table_exists(dbh => $dbh,  dbname => $old_dbname,  table_name => 'SMTP_SERVER_SETTINGS')) {

        my $smtp_conf_row = $dbh->selectrow_hashref("SELECT * FROM $old_dbname.SMTP_SERVER_SETTINGS");
        if($smtp_conf_row) {
            my $asm_smtp_conf_name = 'asm_smtp_conf';
        
            # Create new SMTP configuration object in MCP
            my $smtp_conf_mcp = F5::ASM::COAPI::Object->new('smtp_config');
           
            $smtp_conf_mcp->{name}     = $asm_smtp_conf_name;
            $smtp_conf_mcp->{username} = $smtp_conf_row->{username} if defined $smtp_conf_row->{username}; # can be NULL in database
            $smtp_conf_mcp->{from_address} = $smtp_conf_row->{from_address};
            $smtp_conf_mcp->{source_machine_address} = $smtp_conf_row->{source_machine_address};
            $smtp_conf_mcp->{smtp_server_address} = $smtp_conf_row->{smtp_server_address};
            $smtp_conf_mcp->{smtp_server_port} = $smtp_conf_row->{smtp_server_port};
            $smtp_conf_mcp->{is_auth} = $smtp_conf_row->{flg_is_auth};
            $smtp_conf_mcp->{encryption} = $encryption_type_map{$smtp_conf_row->{encryption}};
            
            # Password might be a plaintext in some versions (10.1.0 to <10.2.0).
            if (defined $smtp_conf_row->{password}) { # can be NULL in database
                if($params->{old_version} lt '0010.0002.0000.0000.0000') { 
                    $smtp_conf_mcp->{password} = PassCrypt::crypt_password($smtp_conf_row->{password});       
                } else {
                    $smtp_conf_mcp->{password} = $smtp_conf_row->{password};
                }
            }

            $smtp_conf_mcp->save();
            
            # Link security analytics global settings with new SMTP configuration
            # Since these settings have been introduced in 11.3.0 we don't worry
            # about overwriting them
            if($smtp_conf_row->{flg_is_enabled}) {
                my $analytics_settings = F5::ASM::COAPI::Query->new('fw_analytics_settings', {name=>'/Common/analytics-settings'});
                if($analytics_settings && @$analytics_settings) {
                    $analytics_settings->[0]->{smtp_config_name} = $asm_smtp_conf_name;
                    $analytics_settings->[0]->save();
                }
            }
        }
    }

    $logger->debug("End SmtpConfiguration fixup");   
}

1;
