package F5::DbSync;

use strict;
use warnings;

# logging related default values
use constant DEFAULT_DEBUG_LEVEL  => 1;

use constant ERROR_PREFIX      => "\nDbSync error:\n";
use constant BIGIP_TS_INTEGRATION_PROXY_USER => 'bigipTMUI';

use F5::ASMConfig::ASMConfig;
use F5::DbUtils;
use F5::SetActive;
use F5::Hsl qw(FTP SMTP); 
use F5::Logger qw(:LEGACY_ERROR_CONSTANTS);
use F5::ASM::COAPI;
use F5::ASMConfig::GuaranteedClient;
use F5::Mcp::ProfileType;

F5::ASM::COAPI::Login('root');

=pod 

=head1 NAME

F5::DbSync - a Perl module for synchronizing between BIGIP's database and
ASM's database. 

=head1 SYNOPSIS

	use F5::DbSync;

	my $dbsync = F5::DbSync->new;

        $dbsync->sync_policies or die $dbsync->error_message

=head1 DESCRIPTION

=head1 METHODS

=head2 new

The constructor is used for generating a new object,
initializing its attributes.

=cut


sub new {
    my ($class, %arguments) = @_;
    my $error_message = '';
    my $dbh = $arguments{dbh} || F5::DbUtils::get_dbh();
    my $asm_config_handler = $arguments{asm_config_handler};
    my $logger = $arguments{logger} || F5::Logger->get_logger(logger_name => 'dbsync');
    my $hsl = F5::Hsl->new(dbh => $dbh, logger => $logger);
    $error_message.= ERROR_PREFIX . $hsl->error_message unless $hsl->connected;
    bless {
        _dbh                 => $dbh,
        _hsl                 => $hsl,
        _asm_config_handler  => $asm_config_handler,
        _error_message       => $error_message, 
        _update_core_needed  => 0,
        _logger              => $logger
    } , $class;
}

sub set_asm_config_handler {
    my ($self, $asm_config_handler) = @_;
    $self->{_asm_config_handler} = $asm_config_handler;
    return $asm_config_handler;
}

sub get_asm_config_handler {
    my ($self) = @_;
    if (!$self->{_asm_config_handler}) {
        $self->{_asm_config_handler} = F5::ASMConfig::GuaranteedClient::get_client(username => 'dbsync');

    }
    return $self->{_asm_config_handler};
}

=head2 sync_policies

Synchronizes the ASM policies between MCP and ASM databases.

=cut

sub sync_policies {
    my($self) = @_;
    $self->{_error_message} = '';

    if (! F5::Cluster::Utils::mcp_config_loaded(logger => $self->{_logger}) ) {
        return $self->_error("MCP config is not ready");
    }

    # get mcp policies
    my %bigip_policies;
    my $class_query = F5::ASM::COAPI::Query->new("asm_policy");
    if ($class_query && @$class_query) {
        %bigip_policies = map { $_->{name} => $_ } @$class_query;
    }

    $self->_log_print($DBG, undef, "bigip_policies: ".join(', ', keys %bigip_policies));

    # get asm policies
    my $asm_config_client = $self->get_asm_config_handler();
    my $call_flags = F5::ASMConfig::CallFlags->new({do_not_relay => 1, skip_mcp => 1});
    my %ts_policies;
	
    eval {
        $asm_config_client->{session}->{partition} = F5::ASMConfig::Constants::PARTITION_ALL();
        my $search_result = $asm_config_client->search_policy($asm_config_client->{session},
                                                               $call_flags,
                                                               F5::ASMConfig::Policy->new());
        if ($search_result->{result_record_count}) {        
            %ts_policies = map { $_->{policy}->{policy_name} => $_->{policy}->{is_active_policy} }
                                       @{$search_result->{element_list}};
        }
    };
    if ($@) {
        return $self->_error($@);
    }

    $self->_log_print($DBG, undef, "ts_policies: ".join(', ',keys %ts_policies));

    # add policies missing from ASM and update active flag for existing policies
    # Parent Policies First
    foreach my $name (sort { $bigip_policies{$b}->{policy_type} <=> $bigip_policies{$a}->{policy_type} } keys %bigip_policies) {
        my $err;
        my $policy =  F5::ASMConfig::Policy->new({policy_name => $name,
                                                  parent_policy_name => $bigip_policies{$name}->{parent_policy},                                                  
                                                  policy_type => $bigip_policies{$name}->{policy_type}});
        my $set_active_flags = F5::ASMConfig::SetActiveFlags->new({do_update_core => 0});
        $call_flags->{set_active_flags} = $set_active_flags;
        $policy->{is_active_policy} = $bigip_policies{$name}->{is_active_policy};

        if ( ! exists($ts_policies{$name})) {
            $self->_log_print($INF, undef, "Adding Policy '$name' in ASM");

            $policy->{flg_encoding_selected} = 0;
            eval {
                if ($bigip_policies{$name}->{encoding}) {
                    $policy->{encoding} = $bigip_policies{$name}->{encoding};
                    $policy->{flg_encoding_selected} = 1;
                }
                if ($bigip_policies{$name}->{policy_template}) {
                    $policy->{policy_template} = F5::ASMConfig::PolicyTemplate->new({icontrol_name => $bigip_policies{$name}->{policy_template}});
                }
                $asm_config_client->add_policy($asm_config_client->{session}, $call_flags, $policy);
                $self->_set_update_core_needed(1);
            };
            $err = ref $@ ? $@->{error_message} : $@;      
        }
        elsif ($ts_policies{$name} != $policy->{is_active_policy}) {
            $self->_log_print($INF, undef, "Setting Policy '$name' with Active State '$policy->{is_active_policy}' in ASM");

            eval {
                $asm_config_client->update_policy($asm_config_client->{session}, $call_flags, $policy);
                $self->_set_update_core_needed(1);
            };
            $err = ref $@ ? $@->{error_message} : $@;      
        }
        
        if ($err) {
            $self->_log_print($DBG, undef,"Problems adding $name: $err") # don't quit
        }
    }

    # asm policies that don't exist in MCP
    foreach my $name ($self->_A_minus_B([keys %ts_policies], [keys %bigip_policies])) {
        $self->_log_print($INF, undef, "Deleting Policy '$name' from ASM");

        eval {
            $asm_config_client->delete_policy($asm_config_client->{session}, 
                                              $call_flags,
                                              F5::ASMConfig::Policy->new({policy_name => $name}),
                                             );
        };
        if ($@) {
            $self->_log_print($DBG, undef,"Problems deleting $name: $@"); # don't quit
        }
    }

    return $self->error_message if $self->error_message;

    return 1; # Success!
}


=head2 sync_ftp_profiles

Synchronizes the ProfileFTP  of BIGIP with ASM ftp security profiles.

Get FTP profiles list from BIGIP
Get FTP profiles list from ASM  security module
Perform add_profile foreach ftp profile that exists in BIGIP but does not
have a ASM(HSL) FTP profile associated with the same name.
Perform disable_profile foreach FTP profile that exists in ASM but does not
exist in BIGIP as an FTP security profile.

=cut

sub sync_ftp_profiles {
    my($self) = @_;

    return $self->sync_profiles(FTP,"ftp");
}


=head2 sync_profiles

Synchronizes the Profile  of BIGIP with ASM ftp security profiles.

Get profiles list from BIGIP
Get profiles list from ASM  security module
Perform add_profile foreach profile that exists in BIGIP but does not
have a ASM(HSL) profile associated with the same name.
Perform disable_profile foreach profile that exists in ASM but does not
exist in BIGIP as an security profile.

=cut

sub sync_profiles {
    my($self, $profile_type, $profile_name) = @_;

    if (! F5::Cluster::Utils::mcp_config_loaded(logger => $self->{_logger}) ) {
        return $self->_error("MCP config is not ready");
    }

    $self->{_error_message} = '';
    my %bigip_profiles;
    my $profile_class = "profile_$profile_name";

    # get bigip profiles
    if ($profile_name eq "smtp") { 
        my $profile_query = F5::ASM::COAPI::Query->new('ltcfg_instance_field', { class_name => $profile_class, field_name => "security_enabled" } );
        if ($profile_query && @$profile_query) {
            %bigip_profiles = map { ( $_->{instance_name} => ($_->{value} eq "true") ? 1 : 0) } @$profile_query;
        }
    }
    else {
         my $profile_query = F5::ASM::COAPI::Query->new($profile_class);
        if ($profile_query && @$profile_query) {
            %bigip_profiles = map { ( $_->{name} => $_->{security_enabled} ) } @$profile_query;
        }
    }
    $self->_log_print($INF, undef, "bigip_profiles: ".join(', ',keys %bigip_profiles));

    # get ASM VS list
    my $hsl = $self->get_Hsl;
    my @ts_profiles = $hsl->get_profile_names( profile_type => $profile_type);
    return $self->_error($hsl->error_message) if $hsl->error_message;
    $self->_log_print($INF, undef, "ts_profiles: ".join(', ',@ts_profiles));

    # only if profile security flag was changed
    foreach my $name (keys %bigip_profiles)
    {
        $self->_set_update_core_needed(1);
        my $flg_security = $bigip_profiles{$name};
        $self->_log_print($INF, undef, "Set $profile_name security flag '$flg_security' for profile '$name'");

        $hsl->add_profile( profile_name => $name, profile_type => $profile_type, flg_security => $flg_security ,do_update_core => 0 ) or
            $self->_log_print($INF, undef,"Problems adding $name: ".$hsl->error_message); # don't quit
    }

    # smtp profiles that don't exist in BIGIP
    foreach my $name ($self->_A_minus_B(\@ts_profiles,[keys %bigip_profiles])) {
        $self->_log_print($INF, undef, "Disabling $profile_name profile '$name' from ASM");
        # disable profile
        $hsl->disable_profile( profile_name => $name ) or
            $self->_log_print($INF, undef,"Problems disabling $name: ".$hsl->error_message); # don't quit
    }

    return 1; # Success!
}


=head2 sync_smtp_profiles

Synchronizes the ProfileSMTP  of BIGIP with ASM ftp security profiles.

Get SMTP profiles list from BIGIP
Get SMTP profiles list from ASM  security module
Perform add_profile foreach smtp profile that exists in BIGIP but does not
have a ASM(HSL) SMTP profile associated with the same name.
Perform disable_profile foreach SMTP profile that exists in ASM but does not
exist in BIGIP as an SMTP security profile.

=cut

sub sync_smtp_profiles {
    my($self) = @_;
    return $self->sync_profiles(SMTP, "smtp");
}

=pod

=head2 sync_virtual_servers

Synchronizes the VS of BIGIP with ASM VS pool.

Get VS list from BIGIP
Get VS list from ASM security module
Perform add_virtual_server foreach VS that exists in BIGIP but does not
have a ASM VS associated with the same name.
Perform delete_virtual_server foreach VS that exists in ASM but does not
exist in BIGIP as an VS with the same name.

=cut

sub sync_virtual_servers {
    my($self) = @_;

    if (! F5::Cluster::Utils::mcp_config_loaded(logger => $self->{_logger}) ) {
        return $self->_error("MCP config is not ready");
    }

    $self->{_error_message} = '';

    # get bigip VS
    my %PSM_TYPES = ( ftp  => F5::Mcp::ProfileType::PROFILE_FTP,
                      smtp => F5::Mcp::ProfileType::PROFILE_INSTANCE,);
    my %bigip_vs_names;  #(vs_name => profile_name)
    my $vs_query = F5::ASM::COAPI::Query->new("virtual_server_profile" );  #vs_name, profile_name, profile_type
    if ($vs_query && @$vs_query) {
        my %psm_profiles;
        my $smtp_query = F5::ASM::COAPI::Query->new('ltcfg_instance_field', { class_name => 'profile_smtp', field_name => "security_enabled", value => 'true'});
        if ($smtp_query && @$smtp_query) {
            for my $instance (@$smtp_query) {
                $psm_profiles{$PSM_TYPES{smtp}}->{$instance->{instance_name}} = 1;
            }
        }
        my $type_query = F5::ASM::COAPI::Query->new("profile_ftp", {security_enabled => 1});
        if ($type_query && @$type_query) {
            for my $profile (@$type_query) {
                $psm_profiles{$PSM_TYPES{ftp}}->{$profile->{name}} = 1;
            }
        }
        for my $virtual_server (@$vs_query) {
            if ($psm_profiles{$virtual_server->{profile_type}}
                && $psm_profiles{$virtual_server->{profile_type}}->{$virtual_server->{profile_name}}) {
                $bigip_vs_names{$virtual_server->{vs_name}} = $virtual_server->{profile_name};
            }
        }
    }
    
    $self->_log_print($INF, undef, "bigip_virtual_servers: ".join(', ',keys %bigip_vs_names));
       
    # get ASM VS list
    my $asm_config_client = $self->get_asm_config_handler;
    my $call_flags = F5::ASMConfig::CallFlags->new({do_not_relay => 1});
    my %ts_vs_names;

    eval {
        my $search_result = $asm_config_client->search_hsl_virtual_server($asm_config_client->{session},
                                                                          $call_flags,
                                                                          F5::ASMConfig::HslVirtualServer->new);
        if ($search_result->{result_record_count}) {        
            %ts_vs_names = map { $_->{hsl_virtual_server}->{virtual_server_name} => $_->{hsl_virtual_server}->{hsl_data_profile_name} }
                           @{$search_result->{element_list}};
        }
    };
    if ($@) {
        return $self->_error($@);
    }
        
    $self->_log_print($INF, undef, "ts_vs_names: ".join(', ',keys %ts_vs_names));
        
    # add vs names missing from ASM and
    # change data profile name for virtual server if user has changed security profile for it
    foreach my $name (keys %bigip_vs_names) {
        my $profile_name = $bigip_vs_names{$name};
        next if ($ts_vs_names{$name} eq $profile_name);
        
        $self->_log_print($INF, undef, "Updating vs '$name' profile: $profile_name in ASM");

        eval {
            $asm_config_client->update_hsl_virtual_server($asm_config_client->{session}, 
                                                          F5::ASMConfig::CallFlags->new({add_on_updating_nonexistent_record => 1, do_not_relay => 1}),
                                                          F5::ASMConfig::HslVirtualServer->new({
                                                              virtual_server_name => $name,
                                                              hsl_data_profile_name => $profile_name,
                                                          }),);
            $self->_set_update_core_needed(1);
        };
        if ($@) {
            $self->_log_print($DBG, undef,"Problems updating $name: $@"); # don't quit
        }
    }
        
    # vs that don't exist in BIGIP
    foreach my $name ($self->_A_minus_B([ keys %ts_vs_names],[keys %bigip_vs_names])) {
        $self->_log_print($INF, undef, "Deleting virtual server '$name' from ASM");

        eval {
            # deleting virtual server
            $asm_config_client->delete_hsl_virtual_server($asm_config_client->{session}, 
                                                          $call_flags,
                                                          F5::ASMConfig::HslVirtualServer->new({
                                                              virtual_server_name => $name,
                                                          }),);
        };
        if ($@) {
            $self->_log_print($DBG, undef,"Problems deleting $name: $@"); # don't quit
        }
    }

    return $self->error_message if $self->error_message;

    return 1; # Success!
}

=pod

=head2 get_dbh

Get the DBI database handle or the connection with ASM's database.

=cut

sub get_dbh {
    shift->{_dbh};
}

sub set_dbh {
    my ($self, $dbh) = @_;
    $self->{_dbh} = $dbh;
}


=pod

=head2 get_Hsl

Get the F5::Hsl object that communicates with ASM's database.

=cut


sub get_Hsl {
    shift->{_hsl};
}



sub _log_print {
    my ($self,$level, $action, $msg) = @_;
    $self->{_logger}->log_print($level, $action, $msg);
}

sub _A_minus_B {
    my ($self,$elements_in_A,$elements_in_B) = @_;
    my %B = map {$_ => 1} @$elements_in_B;
    return grep { not exists $B{$_} } @$elements_in_A;
}

=pod

=head2 error_message

Returns the last error message.
Can be used to probe for errors (empty message means no problem).

=cut

sub error_message {
    shift->{_error_message}
}

sub _error {
    my ($self,$error_string,$concat) = @_;
    if ($concat) {
        $self->{_error_message}.=ERROR_PREFIX.$error_string;
    } else {
        $self->{_error_message}=ERROR_PREFIX.$error_string;
    }
    $self->_log_print($ERR, undef, $self->error_message());
    return undef;
}

sub _set_update_core_needed {
    my ($self, $needed) = @_;
    $self->{_update_core_needed} = $needed;
    return $needed;
}

sub update_core_if_needed {
    my ($self) = @_;
    if($self->{_update_core_needed}) {
        $self->_log_print($INF, undef, "Enforcer configuration items added to database, doing update_core");
        F5::SetActive::set_active(do_update_core => 1, 
                                  dbh => $self->get_dbh, 
                                  logger => $self->{_logger},
                                  asm_config_handler => $self->get_asm_config_handler);
        $self->_set_update_core_needed(0);
    }
}

sub populate_asm_data_to_mcp {
    my ($self, %params) = @_;
    $self->_log_print($DBG, undef, "start"); 

    return $self->populate_policies(check_before_populate => 0,
                                    delete_unmatched_from_mcp => 1,
                                    %params);
}

sub is_policy_populate_needed {
    my ($self) = @_;
    my $sql = <<EndQuery;
  SELECT id,icontrol_name
    FROM PLC.POLICY_TEMPLATES
   WHERE flg_is_visible = 1
ORDER BY id DESC
   LIMIT 1
EndQuery

    my $mysql_row = $self->get_dbh->selectrow_hashref($sql);
    
    my $mcp_row = F5::ASM::COAPI::Query->new('asm_predefined_policy', {name => $mysql_row->{icontrol_name}});

    return 1 if(! defined $mcp_row);
    return (scalar @$mcp_row) ? 0 : 1;
}

sub populate_policies {
    my ($self, %params) = @_;

    if ($params{check_before_populate}) {
        if(! $self->is_policy_populate_needed()) {
            return 1;
        }
    }

    my $dbh = $self->get_dbh;
    my $db_recs = $dbh->selectall_hashref(<<EndQuery, 'icontrol_name');
SELECT id,icontrol_name
  FROM PLC.POLICY_TEMPLATES
 WHERE flg_is_visible = 1
EndQuery

    my $mcp_recs_query = F5::ASM::COAPI::Query->new('asm_predefined_policy');
    my %mcp_recs = map { ($_->{name} => $_) } @$mcp_recs_query;

    my %all_primary_keys = map { ($_ => 1) } (keys %mcp_recs, keys %$db_recs);

    for my $primary_key (keys %all_primary_keys) {
        my $mcp_rec = $mcp_recs{$primary_key};
        my $db_rec = $db_recs->{$primary_key};

        if ($db_rec) {
            if (! $mcp_rec || $mcp_rec->{predefined_asm_id} != $db_rec->{id}) {

                if (! $mcp_rec) {
                    $mcp_rec = F5::ASM::COAPI::Object->new('asm_predefined_policy');
                    $mcp_rec->{name} = $db_rec->{icontrol_name};
                }

                $mcp_rec->{predefined_asm_id} = $db_rec->{id};
                $mcp_rec->save;
            }
        }
        elsif ($params{delete_unmatched_from_mcp} && $mcp_rec) {
            # No longer exists in MySQL -- needs to be deleted from MCP
            eval {
                $mcp_rec->delete;
            };
            if ($@) {
                # delete may fail if template is in use. If this happens, we must log the failure, but this
                # must not fail this entire process.
                my $err = $@;
                chomp($err);
                $self->{_logger}->info("Policy template \"$mcp_rec->{name}\" could not be deleted -- skipping ($err)") if $self->{_logger};
            }
        }
    }

    return 1;
}

sub delete_mcp_policies {
    # Delete only policies that are not in use: inactive and without strict updates, security type
    my $mcp_policies = F5::ASM::COAPI::Query->new('asm_policy', {is_active_policy => 0,
                                                                 strict_app_updates => 0,
                                                                 policy_type => F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY(),
                                                                });

    for my $mcp_policy (@$mcp_policies) {
        $mcp_policy->delete();
    }
}

=pod

=head1 PREREQUISITES

F5::ASM::COAPI - another way to operate on BIGIP's database

=head1 BUGS

=over 1

=item Log pollutes namespace

The Log module which is used for logging capabilities is exporting a log of
subs and variables into the namespace even if those are not used.


=item intermediate failures

This may be a feature or a bug, depending on your view of granularity of
failures: Errors are accumulated instead of quitting on first failure.

=back

=head1 AUTHOR

Shlomo Yona <s.yona@f5.com>

=head1 SVN/CVS INFO

$Id$

$URL: file:///home/shlomo/MY_LOCAL_SVN_REPOSITORY/trunk/db_sync/DbSync.pm $

$Header$

=head1 SEE ALSO

=cut

1;
