package F5::DbUpgrade;

use strict;
use warnings;

use F5::LockFactory;
use F5::Cfg;
use File::Basename;
use F5::Logger;
use F5::DbFixup::Default;
use F5::CommonUpgrade;
use Getopt::Long;
use F5::DbUtils;
use F5::ASM::COAPI;

our $SQL_INIT_FILE_PATH;

my $logger;

init() unless $ENV{PUPPET_SERVER_PORT};

sub db_upgrade {
  my(%params) = @_;

  my $new_version = $params{new_version};
  my $old_version = $params{old_version};
  my $dbh = $params{dbh};
  my @databases = @{$params{databases}};
  my %databases = map(($_->{bigip_dbname} => $_), @databases);

  our @fixup_modules = load_fixup_modules($new_version, $old_version);

  #--------------------------------------------------
  # main
  #
  $logger = F5::Logger->get_logger(logger_name => 'db_upgrade') or die "Couldn't create logger!\n";
  $logger->info("Starting db_merge");

  # bypass_validation before creating Object->new() or Query->new()
  F5::ASM::COAPI::bypass_validation();

  my $common_upgrade = F5::CommonUpgrade->new(dbh => $dbh,
                                              new_version => $new_version,
                                              old_version => $old_version,
                                              logger => $logger,
                                             );

  # All cross-database access will be handled by explicit naming
  # to keep all database access in a single transaction.
  $dbh->{HandleError} = sub { $logger->error($_[0]); return undef; };
  $dbh->do('SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0');

  # db_upgrade should never log its operation to the binary log
  exit 1 if (! stop_binary_logging($dbh));

  my $idmap = {};
  # start working without autocommit
  for my $database (@databases) {
    my $new_dbname = $database->{staging_dbname};
    my $old_dbname = $database->{ucs_dbname};
    my $abstract_dbname = $database->{bigip_dbname};

    $logger->info("Obtaining list of tables to sync");
    my @sync_tables = get_config_sync_tables(abstract_dbname => $abstract_dbname,
                                             dbh => $dbh,
                                             old_dbname => $old_dbname,
                                             common_upgrade => $common_upgrade,
                                            );

    # Determine which table should not be automatically merged
    for my $fixup (@fixup_modules) {
      for my $excluded_table (@{$fixup->excluded_tables()}) {
	@sync_tables = grep($_ ne $excluded_table, @sync_tables);
	$logger->debug("Table $excluded_table excluded by fixup module " . ref($fixup));
      }
    }

    my $fixup_param_hashref = { 
                               databases => \%databases, 
                               new_version => $new_version, 
                               old_version => $old_version,
                               external_idmap => $params{external_idmap},
                               idmap => $idmap,
                              };

    # Call "before" function for each fixup
    for my $fixup (@fixup_modules) {
      if ($fixup->can('before_fixup') && $fixup->fixup_db eq $abstract_dbname) {
	$logger->debug("Calling 'before' fixup for module " . ref($fixup));
	$fixup->before_fixup($dbh, $old_dbname, $new_dbname, $logger, $fixup_param_hashref);
      }
    }

    for my $table (@sync_tables) {
      $logger->info("Merging data for table $table");
      F5::DbFixup::Default->automerge(dbh => $dbh,
                                      databases => \%databases,
                                      old_dbname => $old_dbname,
                                      new_dbname => $new_dbname,
                                      abstract_dbname => $abstract_dbname,
                                      table => $table,
                                      logger => $logger,
                                      common_upgrade => $common_upgrade,
                                      idmap => $fixup_param_hashref->{idmap},
                                     );
    }

    for my $fixup (@fixup_modules) {
      if ($fixup->can('after_fixup') && $fixup->fixup_db eq $abstract_dbname) {
	$logger->debug("Calling 'after' fixup for module " . ref($fixup));
	$fixup->after_fixup($dbh, $old_dbname, $new_dbname, $logger, $fixup_param_hashref);
      }
    }

    if ($abstract_dbname eq 'PLC') {
        # NOTE: add default rows only works for PLC (for now) -- that may be all that is ever needed
        $common_upgrade->add_default_rows(policy_id => '*',
                                          seen_tables => \@sync_tables,
                                          dbname => $new_dbname);

        $common_upgrade->after_all_tables_complete(policy_id => '*',
                                                   dbh => $dbh,
                                                   idmap => $fixup_param_hashref->{idmap},
                                                   dbname => $new_dbname);
    }
  }

  $params{external_idmap}->{idmap} = $idmap;

  $dbh->do('SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS');
  $dbh->{HandleError} = undef;

  $logger->info("Completed successfully");
}

#======================================================================

sub stop_binary_logging 
{
   my ($dbh) = @_;
   my $sql = "SET SQL_LOG_BIN=0";
   if(! $dbh->do($sql))
   {
      $logger->error("failed to execute query to stop binary logging ".DBI::errstr()) if $logger;
      return;
   }         
   return 1;   
}

#----------------------------------------------------------------------

sub get_config_sync_tables {
  my(%params) = @_;
  my($abstract_dbname) = $params{abstract_dbname};
  my($old_dbname) = $params{old_dbname};
  my($dbh) = $params{dbh};
  my($common_upgrade) = $params{common_upgrade};

  my $POLICY_TABLE = "PL_POLICIES";
  my @tables;  
  open(my $inf, get_sql_filename($abstract_dbname));
  while(my $line = <$inf>) {
    chomp($line);
    if ($line =~ /(?:CONFIG_SYNC|TSCONFIG|EVENT_BACKUP): \s* ([\w\.]+)/x) {
      next if $1 eq $POLICY_TABLE;
      push(@tables, $1);
    }
  }
  close $inf;

  # Include only CONFIG_SYNC tables that exist in the old DB
  my $old_db_tables = $dbh->selectcol_arrayref("SHOW TABLES FROM $old_dbname");
  my %old_db_tables = map(("$abstract_dbname.$_", 1), @$old_db_tables);

  my $renamed_tables = $common_upgrade->get_renamed_tables;
  my %reverse_renamed_tables = reverse %{$renamed_tables};
  my @filtered_tables;
  for my $table_name (@tables) {
      if ($reverse_renamed_tables{$table_name}) {
          push(@filtered_tables, $reverse_renamed_tables{$table_name});
      }
      if ($old_db_tables{$table_name}) {
          push(@filtered_tables, $table_name);
      }
  }
  for my $old_table_name (keys %old_db_tables) {
      if (defined ($renamed_tables->{$old_table_name}) &&
            $renamed_tables->{$old_table_name} eq "") {
          push(@filtered_tables, $old_table_name); #Added deleted tables (renamed to "") to list of tables to process.
      }
  }
  if ($abstract_dbname eq "PLC") {
      #Policy table must be first
      unshift (@filtered_tables, $POLICY_TABLE);
  }
  return @filtered_tables;
}

#----------------------------------------------------------------------

sub get_sql_filename {
  my($dbname) = @_;

  return $SQL_INIT_FILE_PATH . '/'. lc $dbname . '.sql';
}

#----------------------------------------------------------------------

sub load_fixup_modules {
  my($new_version, $old_version) = @_;
  my @fixup_modules;

  my $module_path = F5::DbFixup::Default::get_module_path();
  for my $module_file (glob($module_path . "/*.pm")) {
    next if ($module_file =~ m|DbFixup/Default.pm|);
    require $module_file;

    my $module_class = basename($module_file);
    $module_class =~ s/.pm$//;
    $module_class = "F5::DbFixup::$module_class";

    my $fixup = $module_class->new();
    if ($fixup->is_valid_version($old_version, $new_version)) {
      push(@fixup_modules, $fixup);
    }
  }

  @fixup_modules = sort { $a->priority <=> $b->priority } @fixup_modules;

  return @fixup_modules;
}

#----------------------------------------------------------------------

sub init {
  my $DB_UPGRADE_CFG_FILE = "/etc/ts/install/db_upgrade.cfg";

  if (! -e $DB_UPGRADE_CFG_FILE)
  {
     return;
  }

  $SQL_INIT_FILE_PATH = F5::Cfg::cfg_get_config_item($DB_UPGRADE_CFG_FILE, 'config', 'sql_init_file_path');
}

#----------------------------------------------------------------------

1;
