package F5::Deploy;

use strict;
use warnings;

use FindBin;
use Config;
use YAML::XS;
use File::Find;
use File::Basename;
use File::Path;
use File::Copy qw();

my $path_mapping;
my $local_lib_path;
my @potential_base_dirs;

BEGIN {
    $path_mapping = {};
    $local_lib_path = '.';
    my $path_mapping_file_within_base_dir = 'share/deploy_path_mapping.yaml';
    @potential_base_dirs = qw(/opt/f5waf
                              /opt/app_protect);
    my $path_mapping_file = $ENV{F5_WAF_DEPLOY} || '';

    my $current_file = __FILE__;
    if ($current_file =~ m|^(.*?)/lib/perl|) {
        push(@potential_base_dirs, $1);
    }

    if (! $path_mapping_file) {
        for my $base_dir (@potential_base_dirs) {
            if (-f "$base_dir/$path_mapping_file_within_base_dir") {
                $path_mapping_file = "$base_dir/$path_mapping_file_within_base_dir";
                last;
            }
        }
    }

    if (-f $path_mapping_file) {
        $path_mapping = YAML::XS::LoadFile($path_mapping_file);
        $local_lib_path = $path_mapping->{deploy_paths}->{perl_lib_cpan};
    }

    eval {
        require local::lib;
        local::lib->import("$local_lib_path");
    };
}


#----------------------------------------------------------------------

sub map_path {
    my($path_str) = @_;
    $path_str ||= '';

    my @path_elements;
    for my $path_element (split(/ +/, $path_str)) {
        if ($path_element =~ m|^/|) {
            $path_element = map_path_element($path_element);
        }
        push(@path_elements, $path_element);
    }
    return join(' ', @path_elements);
}

#----------------------------------------------------------------------

sub map_path_element {
    my($path_element) = @_;

    if (defined $path_mapping->{tmos_paths}) {
        $path_element =~ s|^/ts|/var/ts|;
        my @matching_path_prefixes = sort { length($b) <=> length($a) } grep { $path_element =~ m|^$_| } keys %{$path_mapping->{tmos_paths}};
        if (@matching_path_prefixes) {
            my $path_type_base = get_path_by_type($path_mapping->{tmos_paths}->{$matching_path_prefixes[0]});
            $path_element =~ s|^$matching_path_prefixes[0]|$path_type_base|;
        }
    }
    return $path_element;
}

#----------------------------------------------------------------------

sub get_path_by_type {
    my($path_type) = @_;
    my $rv = $path_mapping->{deploy_paths}->{$path_type} || '';
    return $rv;
}

#----------------------------------------------------------------------

my @perl_dirs = ($Config{installsitelib}, $Config{installvendorlib});

sub copy {
    my(%params) = @_;
    my $from = $params{from};
    my $to = $params{to};
    my $do_map_perl = defined($params{do_map_perl}) ? $params{do_map_perl} : 1;

    my $glob_match;
    if ($to) {
        $to =~ s/\{\{(.*?)\}\}/get_path_by_type($1)/ge;
    }
    if ($from =~ /\*/) {
        # Convert glob to regexp
        $glob_match = basename($from);
        $glob_match =~ s/\*/\.\*/g;
        $from = dirname($from);
        File::Path::make_path($to);
    }
    find({ wanted => sub {
               my $base_filename = basename($File::Find::name);
               return if ($base_filename eq '.' || $base_filename eq '..');
               if ($glob_match && $base_filename !~ /$glob_match/) {
                   return;
               }

               my $dest_filename = $to;
               if (-d $from && $to && ! $glob_match) {
                   $dest_filename =  $File::Find::name;
                   $dest_filename =~ s/$from/$to/;
               }
               if (! $dest_filename) {
                   $dest_filename = $File::Find::name;
                   $dest_filename =~ s/$from//;
                   if ($do_map_perl) {
                       for my $perl_dir (@perl_dirs) {
                           $dest_filename =~ s/$perl_dir/get_path_by_type('perl_lib_f5')/e;
                       }
                   }
                   $dest_filename = map_path($dest_filename);
               }
               if ($dest_filename && $dest_filename ne $File::Find::name) {
                   if (-f $File::Find::name) {
                       if (! -d dirname($dest_filename)) {
                           File::Path::make_path(dirname($dest_filename));
                       }
                       File::Copy::cp($File::Find::name, $dest_filename);
                   }
                   elsif (-d $File::Find::name) {
                       File::Path::make_path($dest_filename);
                   }
               }
           },
           no_chdir => 1,
           bydepth => 1,
         },
         $from);
}

#----------------------------------------------------------------------

sub bundle {
    my(%params) = @_;
    my $target = $params{target};
    my $ignore_file = $params{ignore_file};

    my $ignore_matcher;
    if ($ignore_file) {
       require File::Globstar::ListMatch;
       $ignore_matcher = File::Globstar::ListMatch->new($ignore_file,
                                                        ignoreCase => 1,
                                                        isExclude => 1);
    }
    
    find({ wanted => sub {
             my $file = $File::Find::name;

             $file =~ s|^/||; #remove leading slash
             if ($ignore_matcher && $ignore_matcher->match($file)) {
                 return;
             }
             my $dest_filename = "$target/$file";
             my $target_dir = (-d $File::Find::name) ? $dest_filename : dirname($dest_filename);
             File::Path::make_path($target_dir);
             if (-f $File::Find::name) {
                 File::Copy::cp($File::Find::name, $dest_filename);
             }
           },
           no_chdir => 1,
           bydepth => 1,
         },
         values %{$path_mapping->{deploy_paths}});
}

#----------------------------------------------------------------------

sub make_dirs {
    for my $dir (values %{$path_mapping->{deploy_paths}}) {
        if (! -e $dir) {
            File::Path::make_path($dir);
        }
    }
}

#----------------------------------------------------------------------

1;
