package F5::DeviceGroup;

use strict;
use warnings;

use F5::ASM::COAPI;
use F5::Mcp::DevicegroupType;
use F5::Mcp::HaState;

#----------------------------------------------------------------------

sub get_device_group_devices {
    my ($local_device_name) = @_;
    $local_device_name ||= get_local_device_name();
    my $device_group_devices = F5::ASM::COAPI::Query->new("devicegroup_device", {device => $local_device_name});
    if ($device_group_devices && @$device_group_devices) {
        return $device_group_devices;
    }
    return;
}

#----------------------------------------------------------------------

sub get_device_groups_for_device {
    my ($local_device_name) = @_;
    $local_device_name ||= get_local_device_name();
    my $device_group_devices = get_device_group_devices($local_device_name);
    if ($device_group_devices && @$device_group_devices) {
        my %matching_device_groups =  map { ($_->{devicegroup} => 1 )}  @$device_group_devices;
        if (%matching_device_groups) {
             my $device_groups = F5::ASM::COAPI::Query->new("devicegroup");
             if ($device_groups && @$device_groups) {                 
                 my @matching_asm_device_groups = grep { $matching_device_groups{$_->{name}} && !$_->{hidden} } @$device_groups;
                 if (@matching_asm_device_groups) {
                     return \@matching_asm_device_groups;
                 }
             }
        }
    }
    return;
}

#----------------------------------------------------------------------

sub get_device_group {
    my ($local_device_name) = @_;
    $local_device_name ||= get_local_device_name();
    my $device_groups = get_device_groups_for_device($local_device_name);
    if ($device_groups && @$device_groups) {
        my @matching_asm_device_groups = grep { $_->{asm_sync} == 1 } @$device_groups;
        if (@matching_asm_device_groups) {
            return $matching_asm_device_groups[0];
        }
    }
    my $default_res;
    $default_res->{'name'} = "";
    return $default_res;
}

#----------------------------------------------------------------------

sub get_failover_device_group {
    my ($local_device_name) = @_;
    $local_device_name ||= get_local_device_name();
    my $device_groups = get_device_groups_for_device($local_device_name);
    if ($device_groups && @$device_groups) {
        my @matching_asm_device_groups = grep { $_->{type} == F5::Mcp::DevicegroupType::dgt_failover() } @$device_groups;
        if (@matching_asm_device_groups) {
            return $matching_asm_device_groups[0];
        }
    }
    my $default_res;
    $default_res->{'name'} = "";
    return $default_res;
}

#----------------------------------------------------------------------

sub get_asm_sync_folder_name {
     my ($device_group) = @_;
     return '' unless defined($device_group);
     return $device_group . "_asm_sync";
}

#----------------------------------------------------------------------

sub get_or_create_asm_sync_folder {
     my ($device_group) = @_;
     return get_asm_sync_folder($device_group, 1);
}

#----------------------------------------------------------------------

sub get_asm_sync_folder {
    my ($device_group, $create) = @_;
    my $folders = F5::ASM::COAPI::Query->new("folder");

    my $asm_sync_folder =  get_asm_sync_folder_name($device_group);
    if ($folders && @$folders) {
        if (my ($sync_folder) = grep { $_->{name} eq $asm_sync_folder} @$folders) {
            if ($sync_folder->{devicegroup} ne $device_group) {
                if ($create) {
                    $sync_folder->{devicegroup} = $device_group;
                    $sync_folder->save();
                }
                else {
                    return;
                }
            }
            return $sync_folder->{name};
        }
    }

    if ($create) {
        my $object = F5::ASM::COAPI::Object->new("folder");
        $object->{'name'} = $asm_sync_folder;
        $object->{'devicegroup'} = $device_group;
        $object->{'traffic_group'} = "none";
        $object->save();
        return $object->{'name'};
    }
    return;
}
#----------------------------------------------------------------------

sub is_folder_synchronized {
    my (%params) = @_;
    my $device_group = $params{device_group};
    my $folder = $params{folder};
    my $create = $params{create};

    my %mcp_folders = map { ( $_->{name} => $_ ) } @{F5::ASM::COAPI::Query->new('folder')};
    if ($mcp_folders{$folder}) {
        return ($mcp_folders{$folder}->{devicegroup} eq $device_group);
    }

    if ($create) {
        #recursively create any missing folders
        my @path_parts = split('/', $folder);
        my $path = shift @path_parts; #get rid of first empty match;
        for my $part (@path_parts) {
            $path .= "/$part";
            if (!$mcp_folders{$path}) {
                my $obj = F5::ASM::COAPI::Object->new('folder');
                $obj->{name} = $path;
                $obj->save();
                $mcp_folders{$path} = $obj;
            }
        }
    }
    return 0;  
}

#----------------------------------------------------------------------

sub clean_asm_device_sync {
    my (%params) = @_;
    my $device_group = $params{device_group};
    my $logger = $params{logger};
    my $old_folder_name =  get_asm_sync_folder_name($device_group);
    my $device_syncs = F5::ASM::COAPI::Query->new("asm_device_sync");
    if ($device_syncs && @$device_syncs) {
       for my $device_sync (@$device_syncs) {
            if ($device_sync->{device_name} =~ /$old_folder_name\//) {
                $logger->debug("Removed asm_device_sync object $device_sync->{device_name}") if $logger;
                $device_sync->delete();
            }
        }
    }
}

#----------------------------------------------------------------------

sub get_local_device_name {
    my $devices = F5::ASM::COAPI::Query->new("sys_device");
    if ($devices && @$devices) {
        my($local_device) = grep { $_->{self_device} } @$devices;
        return $local_device->{name};
    }


    return;
}

#----------------------------------------------------------------------

sub get_devices_in_device_group {
    my($device_group) = @_;
    my $devices = F5::ASM::COAPI::Query->new("devicegroup_device", { devicegroup => $device_group});

    if ($devices && @$devices) {
        return map { $_->{device} } @$devices;
    }

    return;
}

#----------------------------------------------------------------------

sub get_device {
    my($device_name) = @_;
    my $devices = F5::ASM::COAPI::Query->new("sys_device", { name => $device_name});

    if ($devices && @$devices) {
        return $devices->[0];
    }

    return;
}

#----------------------------------------------------------------------

sub get_device_compare_string {
    my ($device) = @_;
    if ($device && $device->{configsync_ip} ne '::') {
        return "[$device->{name} $device->{configsync_ip}]";
    }
    return '';
}
#----------------------------------------------------------------------

sub is_device_active {
    my ($device_name) = @_;
    my $device = get_device($device_name);
    # Consider UNKNOWN as active as it indicates these devices are not in the same HA group.
    return ($device->{failover_state} == F5::Mcp::HaState::HA_ACTIVE || 
            $device->{failover_state} == F5::Mcp::HaState::HA_UNKNOWN);
}
    
#----------------------------------------------------------------------

1;
