package F5::Dos::BotSignatures;

use strict;
use warnings;
no if $] >= 5.017011, warnings => 'experimental::smartmatch';

use File::Find;

use XML::Hash::LX;
use F5::Deploy;
use F5::Mcp::CfgSrc;
use F5::CRC;
use F5::NegativeSignatures::Collection::Compiler;

use F5::Mcp::BotDefenseMitigationAction;
use F5::Mcp::BotDefenseSignatureMatchType;

use F5::ASM::COAPI;
use F5::Logger;
use F5::LockFactory;
use F5::Utils::File;

# F5::License uses SOAP::Lite
# there is no SOAP::Lite at build time
# we get here at build time while building the negsig target
eval {
    if (! F5::BigipUtils::is_install_mode()) {
        require F5::License;
    }
};

use List::Util qw(max first);
use Digest::MD5;

our $MALICIOUS_BOT_CLASS = '/Common/Malicious Bot';
our $UNTRUSTED_BOT_CLASS = '/Common/Untrusted Bot';
our $TRUSTED_BOT_CLASS = '/Common/Trusted Bot';

my $UNLIKELY_DELIMITER = '###UNLIKELY_DELIMITER###';
my $MOBILE_SIGNATURES_CATEGORY_NAME = '/Common/Mobile App without SDK';
my $MOBILE_SIGNATURES_CATEGORY_CLASS_NAME = '/Common/Mobile Application';

sub prepare {
    my(%params) = @_;

    # Needed to be able to pull the factory_rule correctly
    F5::ASM::COAPI::bypass_validation();

    my $logger = F5::Logger->get_logger(logger_name => 'dosl7_bot_signatures');

    if (!(F5::License::is_dos_licensed() || F5::License::is_asm_licensed())) {
        $logger->info("ASM and DOS are not licensed, skipping bot signatures prepare") if $logger;
        return;
    }

    my $locker = F5::LockFactory::get_locker(lock_filename => F5::Deploy::map_path('/ts/lock/dosl7_bot_signatures.lock'),
                                             logger => $logger);
    $locker->lock;

    my @updated_objs;

    my %compiled_enforcement_profile_by_parent_profile = F5::ASM::COAPI::Query->new('bot_defense_compiled_enforcement_profile',
                                                                                    undef,
                                                                                    retrieve_as_hash => { key_field_name => 'parent_profile'},
                                                                                   );

    my %categories = F5::ASM::COAPI::Query->new('bot_defense_signature_category',
                                                undef,
                                                retrieve_as_hash => { key_field_name => 'category'},
                                               );

    my $signatures_query = F5::ASM::COAPI::Query->new('bot_defense_signature');

    my $profiles = F5::ASM::COAPI::Query->new('profile_bot_defense');

    for my $profile (@$profiles) {
        $logger->info("Generating bot signatures for profile $profile->{name}") if $logger;

        my %profile_signature_overrides = F5::ASM::COAPI::Query->new('bot_defense_profile_signature_override',
                                                    {
                                                      'parent_profile' => $profile->{name},
                                                    },
                                                    retrieve_as_hash => { key_field_name => 'name'},
                                                   );

        my %profile_staged_signature = F5::ASM::COAPI::Query->new('bot_defense_profile_staged_signature',
                                                    {
                                                      'parent_profile' => $profile->{name},
                                                    },
                                                    retrieve_as_hash => { key_field_name => 'name'},
                                                   );
#
        my %profile_category_overrides = F5::ASM::COAPI::Query->new('bot_defense_profile_signature_category_override',
                                                    {
                                                      'parent_profile' => $profile->{name},
                                                    },
                                                    retrieve_as_hash => { key_field_name => 'category_name'},
                                                   );

        my %profile_class_overrides = F5::ASM::COAPI::Query->new('bot_defense_profile_class_override',
                                                    {
                                                      'parent_profile' => $profile->{name},
                                                    },
                                                    retrieve_as_hash => { key_field_name => 'class_name'},
                                                   );

        my %class_mitigation_mode = F5::ASM::COAPI::Query->new('bot_defense_class_mitigation_mode',
                                                    {
                                                      'mode' => $profile->{mode},
                                                    },
                                                    retrieve_as_hash => { key_field_name => 'parent_class'},
                                                   );

        my $compiled_sigs_query = F5::ASM::COAPI::Query->new('bot_defense_compiled_signatures',
                                                    {
                                                      'parent_profile' => $profile->{name},
                                                    });
        my %compiled_sigs = map { ($_->{name} => { sig_obj => $_ }) } @$compiled_sigs_query;

        my @sigs_to_compile = ();

        # We need to handle signatures in a consistent order, so that md5(s) are identical
        #     when compiling identical signature collections
        my $compiled_micro_service_sigs_by_service_name;
        my $compiled_micro_service_sigs_by_sig_id;
        my %profile_micro_services = F5::ASM::COAPI::Query->new('bot_defense_profile_micro_service',
                                                    {
                                                      'parent_profile' => $profile->{name},
                                                    },
                                                    retrieve_as_hash => { key_field_name => 'name'},
                                                   );

        for my $profile_micro_service (values %profile_micro_services) {
            my $compiled_micro_service_sigs_query = F5::ASM::COAPI::Query->new('bot_defense_compiled_signatures_micro_service',
                                                    {
                                                      'parent_profile' => $profile->{name},
                                                      'parent_micro_service' => $profile_micro_service->{name},
                                                    });

            $compiled_micro_service_sigs_by_service_name->{$profile_micro_service->{name}}->{by_signature_name} = { map { ($_->{name} => $_) } @$compiled_micro_service_sigs_query };
            $compiled_micro_service_sigs_by_service_name->{$profile_micro_service->{name}}->{profile_micro_service_type} = $profile_micro_service->{type};

            $compiled_micro_service_sigs_by_sig_id = { map { ($_->{id} => $_) } @$compiled_micro_service_sigs_query };
        }

        for my $signature (sort { $a->{name} cmp $b->{name} } @$signatures_query) {
            my $signature_name = $signature->{name};
            my $category_name = $signature->{category_name};
            my $category = $categories{$category_name};
            my $class_name = $category->{class_name};

            $logger->trace("Now handling signature $signature_name");

            my $global_action = List::Util::first { defined $_ } ($profile_category_overrides{$category_name},
                                                                  $profile_class_overrides{$class_name},
                                                                  $class_mitigation_mode{$class_name},
                                                                 );
            $global_action = $global_action->{action};

            if (! defined $global_action) {
                $logger->error("Signature $signature_name has no default class mitigation mode for class $class_name");
                next;
            }

            my $signature_action =
                (defined $profile_signature_overrides{$signature_name})
                    ? $profile_signature_overrides{$signature_name}->{action}
                    : undef ;

            my $action = (defined $signature_action) ? $signature_action : $global_action ;

            $logger->trace("Signature $signature_name category $category_name, class $class_name, profile $profile->{name}");
            $logger->trace("Signature $signature_name action " . (defined $signature_action ? $signature_action : "undef") .
                           ", global action " . (defined $global_action ? $global_action : "undef") .
                           ", final action $action");

            # if ! defined $signature_action - 
            #     then still go on, cause micro services might have their own action set

            if (defined $signature_action && $signature_action == $global_action &&
                $global_action == F5::Mcp::BotDefenseMitigationAction::BOT_DEFENSE_MITIGATION_ACTION_NONE &&
                # signatures of any mobile category (identified by the mobile class) should always be added to the compiled collection
                $class_name ne $MOBILE_SIGNATURES_CATEGORY_CLASS_NAME ) {

                $logger->debug("Signature $signature_name has action none for profile $profile->{name}, skipping");
                next;
            }

            my $staging = (defined $profile_staged_signature{$signature_name}) ? 1 : 0 ;
            $logger->trace("Profile staged signature $signature_name is " . ($staging ? "defined" : "not defined") . ", signature staging is set to $staging");

            my $compiled_sig;
            my $sig_enforcement_was_modified = 0;
            if ($compiled_sigs{$signature_name}) {
                $compiled_sig = $compiled_sigs{$signature_name}->{sig_obj};

                $logger->trace("Signature $signature_name is an existing profile signature");

                if ($compiled_sig->{signature_trans_id} != $signature->{transaction_id}) {
                    $logger->trace("Signature $signature_name enforcement was modified due to a new transaction id (was $compiled_sig->{signature_trans_id}, now $signature->{transaction_id})");
                    $compiled_sig->set_value('signature_trans_id', $signature->{transaction_id});
                    $sig_enforcement_was_modified = 1;
                }

                if ($compiled_sig->{action} != $action) {
                    $logger->trace("Signature $signature_name enforcement was modified due to a new action (was $compiled_sig->{action}, now $action)");
                    $compiled_sig->set_value('action', $action);
                    $sig_enforcement_was_modified = 1;
                }

                if ($compiled_sig->{staging} != $staging) {
                    $logger->trace("Signature $signature_name enforcement was modified due to a new staging (was $compiled_sig->{staging}, now $staging)");
                    $compiled_sig->set_value('staging', $staging);
                    $sig_enforcement_was_modified = 1;
                }

                $compiled_sigs{$signature_name}->{is_in_use} = 1;
            }
            else {
                $logger->trace("Signature $signature_name is a new profile signature");

                $compiled_sig = F5::ASM::COAPI::Object->new('bot_defense_compiled_signatures');
                $compiled_sig->set_value('id', F5::CRC::get_crc32($signature_name));
                $compiled_sig->set_value('parent_profile', $profile->{name});
                $compiled_sig->set_value('name', $signature_name);
                $compiled_sig->set_value('action', $action);
                $compiled_sig->set_value('staging', $staging);
                $compiled_sig->set_value('signature_trans_id', $signature->{transaction_id});
                $sig_enforcement_was_modified = 1;
            }

            if ($sig_enforcement_was_modified) {
                $logger->trace("Adding signature $signature_name to updated objects set");
                push (@updated_objs, $compiled_sig);
            }

            $logger->trace("Adding signature $signature_name to compile set");

            push(@sigs_to_compile,
                 { sig_id => $compiled_sig->{id},
                   rule => build_rule(sig => $signature),
                 });

            for my $profile_micro_service_name (keys %$compiled_micro_service_sigs_by_service_name) {
                my $micro_service_sig;
                # Use micro service actions, unless signature was overriden in profile on a signature level. In this case - use profile action for the signatures.
                if (defined $signature_action) {
                    if ($compiled_micro_service_sigs_by_service_name->{$profile_micro_service_name}->{by_signature_name}->{$signature_name}) {
                        $micro_service_sig = $compiled_micro_service_sigs_by_service_name->{$profile_micro_service_name}->{by_signature_name}->{$signature_name};
                        $logger->trace("Delete micro service signature $micro_service_sig->{name} on save");
                        $micro_service_sig->delete_on_save;
                        push (@updated_objs, $micro_service_sig);
                    }
                    next;
                }

                my $profile_micro_service_type = $compiled_micro_service_sigs_by_service_name->{$profile_micro_service_name}->{profile_micro_service_type};

                $logger->trace("Handle profile $profile->{name} micro service $profile_micro_service_name of type $profile_micro_service_type class $class_name");

                my $class_micro_service_type_action = F5::ASM::COAPI::Query->new('bot_defense_class_micro_service_type',
                                                                   {
                                                                    'name' => $profile_micro_service_type,
                                                                    'parent_class' => $class_name,
                                                                   },
                                                                  )->[0]->{action};

                $logger->trace("Class $class_name micro service type action is $class_micro_service_type_action");

                my $profile_micro_service_class_override_action = F5::ASM::COAPI::Query->new('bot_defense_profile_micro_service_class_override',
                                                                                   {
                                                                                    'parent_profile' => $profile->{name},
                                                                                    'parent_micro_service' => $profile_micro_service_name,
                                                                                    'class_name' => $class_name,
                                                                                   },
                                                                                  )->[0]->{action};

                $logger->trace("Profile $profile->{name} micro service class override action is $profile_micro_service_class_override_action");

                my $class_micro_service_action = (defined $profile_micro_service_class_override_action)
                                                     ? $profile_micro_service_class_override_action
                                                     : $class_micro_service_type_action ;

                $logger->trace("Signature $signature_name action is $action, selected micro service action is $class_micro_service_action");

                my $micro_service_enforcement_was_modified = 0;
                if ($compiled_micro_service_sigs_by_service_name->{$profile_micro_service_name}->{by_signature_name}->{$signature_name}) {

                    $micro_service_sig = $compiled_micro_service_sigs_by_service_name->{$profile_micro_service_name}->{by_signature_name}->{$signature_name};
                    $logger->trace("Profile micro service signature $micro_service_sig->{name} already exists w/ action $micro_service_sig->{action} and staging $micro_service_sig->{staging}");

                    if ($micro_service_sig->{action} != $class_micro_service_action) {
                        $logger->trace("Setting new action $class_micro_service_action for profile micro service signature $micro_service_sig->{name}");

                        $micro_service_sig->set_value('action', $class_micro_service_action);
                        $micro_service_enforcement_was_modified = 1;
                    }

                    # signature staging in profile takes precedence over signature staging on micro service
                    if ($micro_service_sig->{staging} != $compiled_sig->{staging}) {
                        $logger->trace("Setting new staging $compiled_sig->{staging}, for profile micro service signature $micro_service_sig->{name}, which was $micro_service_sig->{staging}");

                        $micro_service_sig->set_value('staging', $compiled_sig->{staging});
                        $micro_service_enforcement_was_modified = 1;
                    }
                }
                else {
                    $logger->trace("Profile micro service signature $compiled_sig->{name} is new, setting action $class_micro_service_action and staging $compiled_sig->{staging}");

                    $micro_service_sig = F5::ASM::COAPI::Object->new('bot_defense_compiled_signatures_micro_service');
                    for my $key (qw(staging parent_profile id name signature_trans_id)) {
                        $micro_service_sig->set_value($key, $compiled_sig->{$key});
                    }
                    $micro_service_sig->set_value('action', $class_micro_service_action);
                    $micro_service_sig->set_value('parent_micro_service', $profile_micro_service_name);
                    $micro_service_enforcement_was_modified = 1;
                }

                if ($micro_service_enforcement_was_modified) {
                    push (@updated_objs, $micro_service_sig);
                }
            }
        }

        for my $compiled_sig (values %compiled_sigs) {
            if (! $compiled_sig->{is_in_use}) {
                $logger->trace("Delete signature $compiled_sig->{sig_obj}->{name} on save");

                $compiled_sig->{sig_obj}->delete_on_save;
                push (@updated_objs, $compiled_sig->{sig_obj});

                if ($compiled_micro_service_sigs_by_sig_id->{$compiled_sig->{sig_obj}->{id}}) {
                    # surely compiled micro service sig IDs are identical to the compiled_sig IDs
                    my $compiled_micro_service_sig = $compiled_micro_service_sigs_by_sig_id->{$compiled_sig->{sig_obj}->{id}};
                    $logger->trace("Delete micro service signature $compiled_micro_service_sig->{name} on save");
                    $compiled_micro_service_sig->delete_on_save;
                    push (@updated_objs, $compiled_micro_service_sig);
                }
            }
        }

        my $sigs_to_compile_size = scalar @sigs_to_compile;
        $logger->debug("Compile set size is $sigs_to_compile_size");

        my $updated_objs_size = scalar @updated_objs;
        $logger->debug("Updated objects set size is $updated_objs_size");

        my $profile_has_previous_sigs = scalar keys %compiled_sigs;
        $logger->debug("Profile $profile->{name} already has $profile_has_previous_sigs signatures");

        # always compile a collection for a profile that has no sigs, even if there were no changes
        #     required for profiles that have no sisgs but have microservices
        if (@updated_objs || !$profile_has_previous_sigs) {
            $logger->debug("Building bot signature collection") if $logger;
            my $compiled_collection = F5::NegativeSignatures::Collection::Compiler::get_compiled_collection(sigs => [ sort { $a->{sig_id} cmp $b->{sig_id} } @sigs_to_compile ],
                                                                                                            sig_type => 'dos',
                                                                                                            logger => $logger);
            my $md5 = Digest::MD5->new;
            $md5->add($compiled_collection);
            my $compiled_collection_md5 = $md5->hexdigest;

            my $compiled_enforcement_query = F5::ASM::COAPI::Query->new('bot_defense_compiled_enforcement',
                                                                        {
                                                                         'compiled_collection_md5' => $compiled_collection_md5,
                                                                        });
            #if already have compiled enforcement use it as reference
            if ($compiled_enforcement_query && @$compiled_enforcement_query) {
                $logger->debug("Profile $profile->{name} compiled enforcement $compiled_collection_md5 already exists") if $logger;
            }
            #else create the new compiled enforcement row in table
            else {
                my $compiled_enforcement = F5::ASM::COAPI::Object->new('bot_defense_compiled_enforcement');
                $compiled_enforcement->set_value('compiled_collection', $compiled_collection);
                $compiled_enforcement->set_value('compiled_collection_md5', $compiled_collection_md5);
                push (@updated_objs, $compiled_enforcement);

                $logger->debug("Adding compiled enforcement $compiled_collection_md5 for profile $profile->{name}") if $logger;
            }

            #link it to the profile
            my $compiled_enforcement_profile;
            if ($compiled_enforcement_profile_by_parent_profile{$profile->{name}}) {
                $compiled_enforcement_profile = $compiled_enforcement_profile_by_parent_profile{$profile->{name}};
            }
            else {
                $compiled_enforcement_profile = F5::ASM::COAPI::Object->new('bot_defense_compiled_enforcement_profile');
                $compiled_enforcement_profile->set_value('parent_profile'   , $profile->{name}        );
            }

            $compiled_enforcement_profile->set_value('compiled_collection_md5', $compiled_collection_md5);
            push (@updated_objs, $compiled_enforcement_profile);

        }
        else {
            $logger->debug("No update necessary for profile $profile->{name}") if $logger;
        }
        $logger->debug("Completed bot signatures for profile $profile->{name}") if $logger;

        if (@updated_objs) {
            eval {
                $logger->info("Sending updated objects to MCP for profile $profile->{name}") if $logger;
                F5::ASM::COAPI::TransactionalSave(\@updated_objs);
                @updated_objs = ();
                $logger->info("MCP transaction complete") if $logger;
            };
            if ($@) {
                # an error may happen, for example when the license expired
                $logger->info("Cannot send updated objects to mcp: $@ $F5::COAPI::error") if $logger;
                return;
            }
        }
        else {
            $logger->info("Profile $profile->{name} does not need to be updated in MCP.") if $logger;
        }
    }

    my %compiled_enforcement_profile_by_md5 = F5::ASM::COAPI::Query->new('bot_defense_compiled_enforcement_profile',
                                                                         undef,
                                                                         retrieve_as_hash => { key_field_name => 'compiled_collection_md5'},
                                                                        );

    #delete old compiled enforcemnts that are not used anymore
    my $compiled_enforcements = F5::ASM::COAPI::Query->new('bot_defense_compiled_enforcement');
    for my $compiled_enforcement (@$compiled_enforcements) {
        my $compiled_collection_md5  = $compiled_enforcement->{'compiled_collection_md5'};

        $logger->debug("Trying to delete compiled enforcement $compiled_collection_md5");

        #first check if other profiles are using this compiled collection
        if ($compiled_enforcement_profile_by_md5{$compiled_collection_md5}) {
            $logger->debug("There is another reference to this compiled collection - avoid delete");
            next;
        }

        $logger->debug("No other reference - deleting $compiled_collection_md5");
        $compiled_enforcement->delete_on_save;
        push (@updated_objs, $compiled_enforcement);
    }

    if (@updated_objs) {
        eval {
            $logger->info("Sending updated objects to MCP (deleted bot_defense_compiled_enforcement objects)") if $logger;
            F5::ASM::COAPI::TransactionalSave(\@updated_objs);
            @updated_objs = ();
            $logger->info("MCP transaction complete") if $logger;
        };
        if ($@) {
            # an error may happen, for example when the license expired
            $logger->info("Cannot send updated objects to mcp: $@ $F5::COAPI::error") if $logger;
            return;
        }
    }
    else {
        $logger->info("None of the bot_defense_compiled_enforcement objects need to be deleted from MCP.") if $logger;
    }

    $locker->unlock;
}

#----------------------------------------------------------------------

sub build_rule {
    my(%params) = @_;

    my $sig = $params{sig};

    if ($sig->{rule}) {
        return $sig->{rule};
    }

    if ($sig->{factory_rule}) {
        return $sig->{factory_rule};
    }

    my $rule = '';
    if ($sig->{url_search_string}) {
        if ($sig->{url_match_type} == F5::Mcp::BotDefenseSignatureMatchType::BOT_DEFENSE_SIGNATURE_MATCH_TYPE_CONTAINS) {
            $rule .= qq{uricontent:"$sig->{url_search_string}"; };
        }
        else {
            $rule .= qq{ipp:"/$sig->{url_search_string}/U"; };
        }
    }

    if ($sig->{user_agent_search_string}) {
        if ($sig->{user_agent_match_type} == F5::Mcp::BotDefenseSignatureMatchType::BOT_DEFENSE_SIGNATURE_MATCH_TYPE_CONTAINS) {
            $rule .= qq{headercontent:"$sig->{user_agent_search_string}"; useragentonly; };
        }
        else {
            $rule .= qq{ipp:"/$sig->{user_agent_search_string}/H"; useragentonly; };
        }
    }
    return $rule;
}

#----------------------------------------------------------------------

sub load_from_xml {
    my(%params) = @_;

    my %field_map = (
                     'factory_rule' => 'rule',
                     'signature_references' => 'references',
                     'risk' => 'risk',
                     'category_name' => 'category',
                    );

    my $logger = $params{logger} || F5::Logger->new('/var/log/dosl7/dosl7_bot_signatures.log');
    my $dry_run = $params{dry_run};

    # bypass_validation before creating Object->new() or Query->new()
    F5::ASM::COAPI::bypass_validation();

    my @updated_objs;

    $logger->info("Loading state") if $logger;

    my $signature_query = F5::ASM::COAPI::Query->new('bot_defense_signature');
    my %mcp_signatures = map { ($_->{leaf_name} => { obj => $_ }) } @$signature_query;
    my $category_query = F5::ASM::COAPI::Query->new('bot_defense_signature_category');
    my %categories = map { ($_->{category} => { obj => $_ }) } @$category_query;

    # this category should always exist
    if (! $categories{$MOBILE_SIGNATURES_CATEGORY_NAME}) {
        my $mobile_signatures_category = F5::ASM::COAPI::Object->new('bot_defense_signature_category');
        $mobile_signatures_category->set_value('category', $MOBILE_SIGNATURES_CATEGORY_NAME);
        $mobile_signatures_category->set_value('class_name', $MOBILE_SIGNATURES_CATEGORY_CLASS_NAME);
        $mobile_signatures_category->set_value('user_defined', 0);
        $mobile_signatures_category->set_value('system_tool_update', 1);
        push(@updated_objs, $mobile_signatures_category);
    }

    if (%mcp_signatures && $params{initial_load_only}) {
        if (@updated_objs) {
            eval {
                $logger->debug("Adding category $MOBILE_SIGNATURES_CATEGORY_NAME to MCP on initial load") if $logger;
                F5::ASM::COAPI::TransactionalSave(\@updated_objs);
                $logger->info("MCP transactional save complete") if $logger;
            };
            if ($@) {
                # an error may happen the profile could not be created, for example when the license expired
                $logger->info("Cannot send updated objects to mcp: $@ $F5::COAPI::error") if $logger;
                return;
            }
        }

        $logger->info("Bot signatures are already configured -- skipping initial load") if $logger;
        return;
    }

    my $dir = $params{dir};
    find({ wanted => sub {
               my $filename = $File::Find::name;
               return unless (-f $filename);
               $logger->info("Parsing $filename") if $logger;

               eval {
                   my $file_contents = F5::Utils::File::file_slurp($filename);
                   return unless $file_contents;

                   my $xml_sig = xml2hash($file_contents, attr => '');
                   return unless ($xml_sig && $xml_sig->{sig} && $xml_sig->{sig}->{rev} &&
                                  $xml_sig->{sig}->{rev}->{category_type} &&
                                  $xml_sig->{sig}->{rev}->{bot_name} &&
                                  $xml_sig->{sig}->{rev}->{category} &&
                                  $xml_sig->{sig}->{rev}->{category} =~ /[a-z]/);

                   $xml_sig->{sig}->{rev}->{category} = '/Common/'.$xml_sig->{sig}->{rev}->{category};
                   if (! $categories{$xml_sig->{sig}->{rev}->{category}}) {
                       my $category = F5::ASM::COAPI::Object->new('bot_defense_signature_category');
                       $category->set_value('category', $xml_sig->{sig}->{rev}->{category});

                       # according to Client Classification in FS and bot_defense_class mcp table
                       if ($xml_sig->{sig}->{rev}->{class_name}) {
                           $category->set_value('class_name', $xml_sig->{sig}->{rev}->{class_name});
                       }
                       elsif ($xml_sig->{sig}->{rev}->{category_type} eq 'malicious') {
                           $category->set_value('class_name', $MALICIOUS_BOT_CLASS);
                       }
                       elsif ($xml_sig->{sig}->{rev}->{category_type} eq 'benign' && $xml_sig->{sig}->{rev}->{category} ne '/Common/Search Engine') {
                           $category->set_value('class_name', $UNTRUSTED_BOT_CLASS);
                       }
                       else {
                           $category->set_value('class_name', $TRUSTED_BOT_CLASS);
                       }

                       $category->set_value('user_defined', 0);
                       $category->set_value('system_tool_update', 1);
                       push(@updated_objs, $category);
                       $logger->debug("Creating Bot Signature Category $category->{category}") if $logger;
                       $categories{$category->{category}} = { obj => $category };
                   }
                   $categories{$xml_sig->{sig}->{rev}->{category}}->{was_seen} = 1;

                   my $bot_name = $xml_sig->{sig}->{rev}->{bot_name};
                   if (! $mcp_signatures{$bot_name} ||
                       ! ref($mcp_signatures{$bot_name}) ||
                       ! $mcp_signatures{$bot_name}->{obj} ||
                       ! ref($mcp_signatures{$bot_name}->{obj})) {

                       $logger->trace("Creating new Bot Signature record for $bot_name") if $logger;
                       my $new_signature = F5::ASM::COAPI::Object->new('bot_defense_signature');
                       $new_signature->set_value('name', $bot_name);
                       $new_signature->set_value('user_defined', 0);
                       $mcp_signatures{$bot_name} = { obj => $new_signature,
                                                                               was_added => 1,
                                                                             };
                   }

                   my $mcp_sig = $mcp_signatures{$bot_name};
                   $mcp_sig->{was_seen} = 1;
                   $logger->trace("Populating Bot Signature record for $bot_name") if $logger;

                   for my $field_name (keys %field_map) {
                       next unless defined $xml_sig->{sig}->{rev}->{$field_map{$field_name}};

                       if (! defined $mcp_sig->{obj}->{$field_name} ||
                           $mcp_sig->{obj}->{$field_name} ne $xml_sig->{sig}->{rev}->{$field_map{$field_name}}) {

                           $mcp_sig->{obj}->set_value($field_name, $xml_sig->{sig}->{rev}->{$field_map{$field_name}});
                           $mcp_sig->{was_modified} = 1;
                       }
                   }

                   #special handling for nested element
                   my $xml_domains = [];
                   if ((defined $xml_sig->{sig}->{rev}->{"hostnames"}) &&
                       (ref ($xml_sig->{sig}->{rev}->{"hostnames"})) &&
                       (defined $xml_sig->{sig}->{rev}->{"hostnames"}->{"hostname"})) {

                       $xml_domains = $xml_sig->{sig}->{rev}->{"hostnames"}->{"hostname"};
                       if (! ref ($xml_domains)) {
                           $xml_domains = [$xml_domains];
                       }
                   }
                   if (!($xml_domains ~~ $mcp_sig->{obj}->{"domain_names"}))  {

                       $mcp_sig->{obj}->set_value("domain_names", $xml_domains);
                       $mcp_sig->{was_modified} = 1;
                   }

                   if ($mcp_sig->{was_modified} || $mcp_sig->{was_added}) {
                       # this is our way to signal the MCP validation logic that this signature was
                       #    added/updated by the signature update
                       # it is also being set by MCP validation when a user defined signature is added
                       #
                       # the prepare process re-sets this back to defaul '0'
                       $mcp_sig->{obj}->set_value('newly_added_or_modified', 1);
                   }


                   if ($mcp_sig->{was_modified}) {
                       $mcp_sig->{obj}->set_value('system_tool_update', 1);
                       push(@updated_objs, $mcp_sig->{obj});
                       $logger->debug("Modifying Bot Signature $mcp_sig->{obj}->{name}") if $logger;
                   }
               };
               if ($@) {
                   my $err_msg = "Failed loading $filename ($@)";
                   $logger->error($err_msg) if $logger;
                   die "$err_msg\n";
               }
           },
           no_chdir => 1,
         },
         $dir);

    for my $deleted_sig (grep { ! $_->{was_seen} && ! $_->{obj}->{user_defined} } values %mcp_signatures) {
        if (! $dry_run) {
             # Checking there are no staged signatures that are set for deletion, if there are some then delete them
            my $staged_sigs = handle_staged_sigs(logger => $logger, signature => $deleted_sig->{obj});
            push(@updated_objs, @$staged_sigs) if (@$staged_sigs);
            # Factory signature to be deleted must be set to user_defined in order to enable deletion
            # (otherwise, it will be blocked by MCP validation).
            # As such, ""user_defined" in this sense means "unlocked" instead
            $deleted_sig->{obj}->set_value('user_defined', 1);
            # Rule must be overwritten with a valid temporary rule, to ensure that validation called from MCP will pass
            $deleted_sig->{obj}->set_value('rule', 'uricontent: "temporary legal rule";');
            $deleted_sig->{obj}->set_value('system_tool_update', 1);
            push(@updated_objs, $deleted_sig->{obj});
            my $obj_for_delete = F5::ASM::COAPI::Object->new('bot_defense_signature');
            $obj_for_delete->set_value('name', $deleted_sig->{obj}->{name});
            $obj_for_delete->delete_on_save;
            push(@updated_objs, $obj_for_delete);
            $logger->debug("Deleting Bot Signature \"$deleted_sig->{obj}->{name}\"") if $logger;
        }
        $deleted_sig->{was_deleted} = 1;
    }

    if (! $dry_run) {
        for my $deleted_category (grep { ! $_->{was_seen} && ! $_->{obj}->{user_defined} } values %categories) {
            # this category should always exist
            next if $deleted_category->{obj}->{category} eq $MOBILE_SIGNATURES_CATEGORY_NAME;

            # Deleted factory category associations in Profiles must be removed first
            my $deleted_category_in_profiles = F5::ASM::COAPI::Query->new('bot_defense_profile_signature_category_override',
                                                                          { 'category_name' => $deleted_category->{obj}->{category} });
            for my $deleted_category_in_profile (@$deleted_category_in_profiles) {
                $deleted_category_in_profile->delete_on_save;
                push(@updated_objs, $deleted_category_in_profile);
                $logger->debug("Deleting Profile Bot Signature Category " .
                        "\"$deleted_category_in_profile->{parent_profile}\" " .
                        "\"$deleted_category_in_profile->{category_name}\"") if $logger;
            }

            # Factory categories to be deleted must also be set to user_defined in order to enable deletion (see above)
            $deleted_category->{obj}->set_value('user_defined', 1);
            $deleted_category->{obj}->set_value('system_tool_update', 1);
            push(@updated_objs, $deleted_category->{obj});
            my $obj_for_delete = F5::ASM::COAPI::Object->new('bot_defense_signature_category');
            $obj_for_delete->set_value('category', $deleted_category->{obj}->{category});
            $obj_for_delete->delete_on_save;
            push(@updated_objs, $obj_for_delete);
            $logger->debug("Deleting Bot Signature Category \"$deleted_category->{obj}->{category}\"") if $logger;
        }
    }

    my %signature_count;
    my @added_names = grep { $_->{was_added} } values %mcp_signatures;
    my @updated_names = grep { $_->{was_modified} && ! $_->{was_added} } values %mcp_signatures;
    my @deleted_names = grep { $_->{was_deleted} } values %mcp_signatures;

    $signature_count{added_count} = scalar @added_names;
    $signature_count{updated_count} = scalar @updated_names;
    $signature_count{deleted_count} = scalar @deleted_names;

    $signature_count{added_names} = [];
    $signature_count{updated_names} = [];
    $signature_count{deleted_names} = [];
    push(@{$signature_count{added_names}},   map { $_->{obj}->{name} } @added_names);
    push(@{$signature_count{updated_names}}, map { $_->{obj}->{name} } @updated_names);
    push(@{$signature_count{deleted_names}}, map { $_->{obj}->{name} } @deleted_names);

    if (! $dry_run) {
        if (@updated_objs) {
            eval {
                $logger->debug("About to save MCP transaction") if $logger;
                F5::ASM::COAPI::TransactionalSave(\@updated_objs);
                $logger->info("MCP transactional save complete") if $logger;
            };
            if ($@) {
                # an error may happen the profile could not be created, for example when the license expired
                $logger->info("Cannot send updated objects to mcp: $@ $F5::COAPI::error") if $logger;
                return;
            }
        }

        if ($signature_count{added_count} || $signature_count{updated_count} || $signature_count{deleted_count}) {
            F5::ASM::COAPI::SaveAll();
        }
    }

    return \%signature_count;
}

#----------------------------------------------------------------------

sub handle_staged_sigs {
    my (%params) = @_;
    my $sig = $params{signature};
    my $logger = $params{logger};
    my $rs = F5::ASM::COAPI::Query->new('bot_defense_profile_staged_signature', { name => $sig->{name} } );
    $logger->debug("Signature $sig->{name} is staged in " . scalar @$rs . " profile, it will be removed due to signature deletion") if (@$rs);
    for my $staged_sig (@$rs) {
      $logger->debug("Signature $sig->{name} is removed from profile $staged_sig->{parent_profile}");
      $staged_sig->delete_on_save();
    }
    return $rs;
}

#----------------------------------------------------------------------

1;
