package F5::EmServices;

use strict;
use warnings;

use File::Temp qw/tempfile/;
use MIME::Base64;
use File::Copy;
use File::Basename;

use F5::DbUtils;
use F5::SetActive;
use F5::Syslog;
use F5::Policy;
use F5::PolicyHistory;
use F5::ASMConfig::EasyClient;

# NOTE: these constants should be equivalent to McpClient::COMMON_PARTITION / ALL_PARTITIONS in xconfig lib
use constant PARTITION_COMMON => '/Common';
use constant PARTITION_ALL => '/';

sub new {
    my ($class, %params) = @_;
    my $dbh = defined $params{dbh} ? $params{dbh} : F5::DbUtils::get_dbh();
    die 'Database connection error' unless defined $dbh;
    my $self = bless {dbh => $dbh, second_try => $params{second_try}}, $class;

    return $self;
}

#-------------------------------

sub add_to_account_transactions {
    my ($self, $policy_name_crc, $dcc_dbname) = @_;
    $dcc_dbname ||= "DCC";
    $self->{dbh}->do(<<EndQuery, undef, $policy_name_crc);
INSERT INTO $dcc_dbname.ACCOUNT_EM_TRANSACTIONS
(policy_name_crc)
VALUES
(?)
EndQuery
}

#-------------------------------

sub delete_from_account_transactions {
    my ($self, $policy_name_crc) = @_;

    my ($exist) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_name_crc);
SELECT COUNT(*)
  FROM DCC.ACCOUNT_EM_TRANSACTIONS
 WHERE policy_name_crc = ?
EndQuery

    $self->{dbh}->do(<<EndQuery) if $exist;
INSERT INTO DCC.ACCOUNT_EM_TRANSACTIONS
(policy_name_crc
VALUES
(NULL)
EndQuery
}

#-------------------------------

sub add_to_policy_transactions {
    my ($self, $policy_id) = @_;
    $self->{dbh}->do(<<EndQuery, undef, $policy_id);
INSERT INTO PLC.PL_POLICY_EM_TRANSACTIONS
(policy_id)
VALUES
(?)
EndQuery
}

#-------------------------------

sub delete_from_policy_transactions {
    my ($self, $policy_id) = @_;

    my ($exist) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT COUNT(*)
  FROM PLC.PL_POLICY_EM_TRANSACTIONS
 WHERE policy_id = ?
EndQuery

    $self->{dbh}->do(<<EndQuery) if $exist;
INSERT INTO PLC.PL_POLICY_EM_TRANSACTIONS
(policy_id)
VALUES
(NULL)
EndQuery
}

#-------------------------------

sub get_policy_class_delete_transaction_id  {
    my ($self) = @_;

    my ($delete_transaction_id) = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT MAX(em_transaction_id)
  FROM PLC.PL_POLICY_EM_TRANSACTIONS
 WHERE policy_id IS NULL
EndQuery

    return $delete_transaction_id;
}

#-------------------------------

sub get_account_class_delete_transaction_id  {
    my ($self) = @_;

    my ($delete_transaction_id) = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT MAX(em_transaction_id)
  FROM DCC.ACCOUNT_EM_TRANSACTIONS
 WHERE policy_name_crc IS NULL
EndQuery

    return $delete_transaction_id;
}

#-------------------------------

sub get_policy_class_transaction_id  {
    my ($self) = @_;

    my ($transaction_id) = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT MAX(em_transaction_id)
  FROM PLC.PL_POLICY_EM_TRANSACTIONS
 WHERE policy_id IS NOT NULL
EndQuery

    return $transaction_id;
}

#-------------------------------

sub get_account_class_transaction_id  {
    my ($self) = @_;

    my ($transaction_id) = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT MAX(em_transaction_id)
  FROM DCC.ACCOUNT_EM_TRANSACTIONS
 WHERE policy_name_crc IS NOT NULL
EndQuery

    return $transaction_id;
}

#-------------------------------

sub get_policy_list {
    my ($self, $partition_name) = @_;

    if (($partition_name ne PARTITION_COMMON) && ($partition_name ne PARTITION_ALL)){
        return;
    }

    local $" = ',';
    my $sth = $self->{dbh}->prepare(<<"EndQuery");
SELECT DISTINCT name, MAX(em_transaction_id) as em_transaction_id
  FROM PLC.PL_POLICIES, PLC.PL_POLICY_EM_TRANSACTIONS
 WHERE PL_POLICIES.id = PL_POLICY_EM_TRANSACTIONS.policy_id
   AND PL_POLICIES.flg_encoding_selected = 1 
 GROUP BY name
EndQuery

    $sth->execute();

    my @rows;
    while(my $row = $sth->fetchrow_hashref) {
        push @rows, $row;
    }

    return \@rows;
}

#-------------------------------

sub get_webapp_list {
    my ($self,$partition_name) = @_;

    if (($partition_name ne PARTITION_COMMON) && ($partition_name ne PARTITION_ALL)){
        return;
    }

    my $webapp_names =  $self->{dbh}->selectcol_arrayref(<<EndQuery);
SELECT policy_name
  FROM DCC.ACCOUNTS
 ORDER BY policy_name
EndQuery
   return $webapp_names;
}

#-------------------------------

sub throw {
    my ($self, $message) = @_;
    F5::Syslog::ts_syslog( 'err',$message) unless $self->{second_try};
    die $message;
}

#-------------------------------

sub import_policy {
    my ($self, $template, $partition_name, $action) = @_;

    if (($partition_name ne PARTITION_COMMON) && ($partition_name ne PARTITION_ALL)){
        return;
    }
    my $client = F5::ASMConfig::EasyClient->new(username => q{EmServices});

    my ($policy_name, $webapp_name, $policy_filename, $policy_description, $flg_is_policy_active) = parse_template(\$template);

    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new({ filename            => $policy_filename,
                                                                                policy_name         => $policy_name,
                                                                                force_policy_name_flg => 1, });


    my $result = $client->import_policy($client->{session}, undef, $import_export_properties);

    if (!$result->{completed_successfully}) {
        $self->throw("Enterprise Manager failed to import $policy_name policy ($webapp_name). " . 
                 join (" ", map {$_->{msg_text} } @{$result->{msgs}}));
    };
    my $new_policy = $result->{policy};
    $new_policy->{description} = $policy_description;
    $new_policy->{is_active_policy} ||= $flg_is_policy_active;  # Cannot import as inactive if the policy overwrote an already active policy.
    $client->update_policy($client->{session}, undef, $new_policy);
    if ($flg_is_policy_active && (basename($webapp_name) ne basename($policy_name))) {
         my $old_policy = F5::ASMConfig::Policy->new(policy_name => $webapp_name);
         $client->replace_policy($client->{session}, undef, $old_policy, $new_policy);
         $old_policy->{is_active_policy} = 0;
         $client->update_policy($client->{session}, undef, $old_policy);
    }
    my $log_text = "New policy %s was imported by Enterprise Manager and replaced the former policy %s.";
    $log_text .= " The policy was applied." if $flg_is_policy_active eq 'true';
            
        F5::Syslog::ts_syslog( 'info',
           $log_text,
           $policy_name, $webapp_name);
    return 1;
}

#-------------------------------

sub parse_template {
    my ($template_ref) = @_;

    require F5::XML::Twig;
    my $twig = F5::XML::Twig->new();
    $twig->parse($$template_ref);
    my $root = $twig->root;
    my $policy_name = $root->{'att'}->{'name'};
    my $policy_description = $root->{'att'}->{'description'};
    my $webapp_name = $root->{'att'}->{'asm:webapp'};
    my $flg_policy_active = $root->{'att'}->{'asm:activate-on-import'};

    my ($out, $policy_filename) = tempfile();
    print $out MIME::Base64::decode_base64($root->text_only);
    close $out;

    return ($policy_name, $webapp_name, $policy_filename, $policy_description, $flg_policy_active);
}

#-------------------------------

sub export_policy {
    my ($self, $policy_name) = @_;

    my ($policy_id, $policy_description, $account_id) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_name);
SELECT id, description, account_id
  FROM PLC.PL_POLICIES
 WHERE name = ?
EndQuery

    if (!$policy_id) {
        die "Policy not found";
    }

    my $flg_policy_active = $account_id > 0;
    $flg_policy_active = $flg_policy_active ? 'true' : 'false';

    my $webapp_name = $policy_name;

    my ($revision) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT MAX(revision)
  FROM PLC.PL_POLICY_HISTORY
 WHERE policy_id = ?
   AND to_date = 0
EndQuery

    my ($policy_activation_time)  = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT MAX(from_date)
  FROM PLC.PL_POLICY_HISTORY
 WHERE policy_id = ?
EndQuery

    my $filename;
    if (defined $revision) {
        $filename = F5::Policy->new(
            policy_id => $policy_id
        )->history->construct_filename(
            revision => $revision, 
            policy_id => $policy_id
        );
    }

    my $flg_file_existed = 1;
    if ((! defined $filename) || (! -f $filename)) {
        my ($fh, $export_filename) = tempfile();
        close $fh;

        my $client = F5::ASMConfig::EasyClient->new(username => q{EmServices});

        my $result = $client->export_policy($client->{session},
    				     undef,
    				     F5::ASMConfig::Policy->new(policy_id => $policy_id),
    				     F5::ASMConfig::ImportExportProperties->new(
    										filename => $export_filename,
    									       ));

        if ( !$result || !$result->{filename}) {
            $self->throw("Enterprise Manager failed to export $policy_name policy.");
        };


        $filename = $result->{filename};
        $flg_file_existed = 0;
    }

    my $xml =  build_xml(
        exported_policy_file => $filename,
        last_activated       => $policy_activation_time,
        policy_name          => $policy_name,
        policy_description   => $policy_description,
        flg_policy_active    => $flg_policy_active,
        webapp_name          => $webapp_name,
    );

    unlink $filename unless $flg_file_existed;
    return $xml;
}

#-------------------------------

sub build_xml {
    my (%params) = @_;
    my $exported_policy_file    = $params{exported_policy_file};
    my $last_activated    = $params{last_activated};
    my $policy_name             = $params{policy_name};
    my $policy_description      = $params{policy_description};
    my $webapp_name             = $params{webapp_name};
    my $flg_policy_active       = $params{flg_policy_active};

    require XML::Twig;
    open (my $in, "<", $exported_policy_file) or die "Cannot read exported policy";
    my $pol;
    while (<$in>) {
        $pol.=$_;
    }
    my $encoded_policy = MIME::Base64::encode_base64($pol);
    close $in;

    my $var1 = XML::Twig::Elt->new('variable');
    $var1->set_att(
        'mode' => "define",
        'name' => "policyName",
        'description' => "Policy name",
        'editable' => "true",
        'visible' => "true",
        'default' => $policy_name,);

    my $var2 = XML::Twig::Elt->new('variable');
    $var2->set_att(
        'mode' => "define",
        'name' => "policyDescription",
        'description' => "Policy description",
        'editable' => "true",
        'visible' => "true",
        'default' => $policy_description,);

    my $var3 = XML::Twig::Elt->new('variable');
    $var3->set_att(
        'mode' => "define",
        'name' => "policyActive",
        'description' => "Apply Policy?",
        'values' => "{'yes' 'no'}",
        'editable' => "true",
        'visible' => "true",
        'default' => "yes",);

    my $var4 = XML::Twig::Elt->new('variable');
    $var4->set_att(
        'mode' => "define",
        'name' => "webappName",
        'description' => "Webapp name",
        'editable' => "true",
        'visible' => "true",
        'dynamic' => "true",
        'values' => '@xconfig_query(\'$target\', \'ASM / Webapp\', \'.*\')',);

    my $var5 = XML::Twig::Elt->new('variable');
    $var5->set_att(
        'mode' => "define",
        'name' => "sourceWebApp",
        'description' => "Source Web Application",
        'editable' => "false",
        'visible' => "true",
        'default' => $webapp_name,);


    my $asm_policy = XML::Twig::Elt->new('asm:policy' => {'#CDATA' => 1}, $encoded_policy);
    $asm_policy->set_att(
        'class-path' => "ASM / Policy",
        'name' => '@policyName',
        'description' => '@policyDescription',
        'asm:last-activated' => $last_activated,
        'asm:webapp' => '@webappName',
        'asm:activate-on-import' => '@policyActive',
        'xconfig-visible' => "false",);

    my $config = XML::Twig::Elt->new('config', $asm_policy);
    my $configs = XML::Twig::Elt->new('configs', $config);

    my $template = XML::Twig::Elt->new('template', $var1, $var2, $var3, $var4, $var5, $configs);
    $template->set_att(
        'xmlns:xsi' => "http://www.w3.org/2001/XMLSchema-instance",
        'xmlns' => "http://www.f5.com/xsd/F5Configuration",
        'xmlns:asm' => "http://www.f5.com/xsd/ApplicationSecurityManager",
        'name' => $policy_name,
        'description' => $policy_description,
        'published' => "false",
        'require-verify' => "false",
        'allow-persist' => "false",
        'schema-version' => "1.0",);

    return $template->sprint;
}

1;

