package F5::ExportPolicy::XML;

use strict;
use warnings;

use base qw(F5::ImportExportPolicy::Base);
use Tie::File;
use Digest::MD5;
use File::Basename qw(basename dirname);
use YAML::XS;
use MIME::Base64;
use Encode qw(decode find_encoding);
use File::Slurp qw(write_file);
use JSON qw(to_json);

# xml <=> Perl hash
use XML::Hash::LX;
# Strictly keep the output order.
$XML::Hash::LX::X2H{order} = 1;
# do not trim leading and trailing whitespace from text nodes
$XML::Hash::LX::H2X{trim} = 0;

# this provides the required XML::Twig like methods
#    for Perl hash that represents an xml document
# TODO: this package needs to go OO
use F5::XMLHash;

# we use F5::Cluster::Sync::are_hashes_equivalent() for comparing a list of hashes
use F5::Cluster::Sync;

use F5::Deploy;
use F5::ObjUtils;
use F5::DbUtils;
use F5::NegativeSignatures::Set;
use F5::BigipVersionUtils;
use F5::BigipUtils;
use F5::ASMConfig::Handler;

# load all entities
use F5::ModuleLoader;
F5::ModuleLoader::load_from_base_class('F5::ASMConfig::Entity::Base');

use PassCrypt;

# globals
my $ASM_EXPORT_XML_CONFIG_FILE = F5::Deploy::map_path('/usr/share/ts/config/export_xml_policy.yaml');
my $ASM_EXPORT_CONF = F5::Deploy::map_path('/usr/share/ts/config/export_yaml');
my $INTEGRITY_CHECK_REPLACE_STRING = 'UNLIKELY_STRING_FOR_INTEGRITY_CHECK_REPLACE';
my $MAX_MEMORY_USAGE = 250_000_000;    # 200 MB approximately
my $MEMORY_CHECK_INTERVAL = 200;

#-----------------------------------------------------------
sub new {
    my ( $class, %params ) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#-----------------------------------------------------------
sub init {
    my ( $self, %params ) = @_;

    if ( exists $params{dbh} ) {
        $self->{dbh} = $params{dbh};
    }
    else {
        $self->{dbh} = F5::DbUtils::get_dbh();
    }
    $self->{logger} = $params{logger};
    eval {
        require Encode::HanExtra;
    };

    $self->SUPER::init();

    my $config_filename = $ASM_EXPORT_XML_CONFIG_FILE;
    $self->{config} = YAML::XS::LoadFile($config_filename);

    $self->{subnode__exists} = {};
    $self->{get_criteria_cache} = {};
    $self->{subnode_search_filter_cache} = {};

    $self->{entity_schema} = F5::ASMConfig::Utils::get_entity_schema();
    %{$self->{entity_schema}->{ElementTypeName}} = reverse %{$self->{entity_schema}->{ElementType}};

    $self->{loaded_files} = {};
    for my $f (glob($ASM_EXPORT_CONF . '/*')) {
        $self->{loaded_files}->{ File::Basename::basename($f) } = YAML::XS::LoadFile($f);
    }

    $self->{sections} = {};
    my @waf_enabled_features = ('', F5::BigipUtils::waf_enabled_features()); #Include empty string (no feature required) in the list of enabled features.
    $self->{enabled_features} = { map { $_ => 1 } @waf_enabled_features };

    $self->{initial_memory_usage} = 0;
    $self->{memory_usage_iterators} = 0;

    $self->{memory_check_count} = 0;
}

#-----------------------------------------------------------
sub check_dependent_field_values {
    my ( $self, %params ) = @_;
    my $dependent_field_values = $params{dependent_field_values} || {};
    my $node = $params{node};

    for my $dependent_field ( keys %$dependent_field_values ) {
        return unless exists $node->{temporary_store}->{$dependent_field};

        if ( ref( $dependent_field_values->{$dependent_field} ) ) {
            if ( !grep { $node->{temporary_store}->{$dependent_field} == $_ }
                @{ $dependent_field_values->{$dependent_field} } )
            {
                return;
            }
        }
        elsif ( ref( $node->{temporary_store}->{$dependent_field} ) ) {
            if ( !grep { $dependent_field_values->{$dependent_field} == $_ }
                @{ $node->{temporary_store}->{$dependent_field} } )
            {
                return;
            }
        }
        else {
            if ( $node->{temporary_store}->{$dependent_field} ne $dependent_field_values->{$dependent_field} )
            {
                return;
            }
        }
    }
    return 1;
}

#-----------------------------------------------------------

##################################################################
#  finding nodes and setting values for elements in policy hash  #
##################################################################
###############################################################################
#-----------------------------------------------------------
sub set_value_for_xpath {
    my ( $self, %params ) = @_;
    my $xpath = $params{xpath};
    my $hash_elt = $params{element};
    my $hash_value = $params{value};
    my $field = $params{field} || {};

    return unless defined($xpath);

    my $values = $self->convert_to_list($hash_value);
    return unless defined $values;

    my ( $xpath_only , $attribute) = split('@',$xpath);
    $xpath_only ||= '';
    $attribute ||= '';

    my $dir = '';
    my $base = $xpath_only;
    if ( $xpath_only =~ m|/| ) {
       $dir = File::Basename::dirname($xpath_only);
       $base = File::Basename::basename($xpath_only);
    }
    $dir = ( $dir eq '.' ) ? '' : $dir ;

    my $single_list = ( ( $attribute && $dir ) || ( $field->{layout} && $field->{layout} eq 'single_list' ) ) ? 1 : 0 ;

    my $id = 0;
    for my $val ( @{$values} ) {
        my $sub_hash_elt = $self->add_or_find_node(
            hash => $hash_elt,
            path => ( $dir ) ? $dir : $base ,
            list_element_id => ( $single_list ) ? undef : $id,
        );

        if ( $dir ) {
            $sub_hash_elt = $self->add_or_find_node(
                hash => $sub_hash_elt,
                path => $base,
                list_element_id => ( $attribute || $single_list ) ? $id : undef ,
            );
        }

        if ($self->{primary_encoding} ne 'utf-8') {
            if ( ref($val) ) {
                $self->recursively_decode(data_ref => $val);
            }
            else {
                $val = $self->{decoder}->decode($val);
            }
        }

        # attribut value has to be defined and not-empty but can be false i.e. '0'
        # parameter name is an exception when we do allow an empty value
        if ( $attribute && defined $val && ( ($field->{ent_field} && $field->{ent_field} eq 'param_name') || $val ne '') ) {
            F5::XMLHash::set_att(
                elt_ref => $sub_hash_elt,
                att_name => $attribute,
                att_val => $val,
            );
        }
        elsif ( ref($val) ) {
            F5::XMLHash::paste(
                parent => $sub_hash_elt,
                child => $val,
            );
        }
        else {
            F5::XMLHash::set_text(
                elt_ref => $sub_hash_elt,
                text => $val,
            );
        }

        $id++;
    }

}

#-----------------------------------------------------------
sub add_or_find_node {
    my ( $self, %params ) = @_;

    my $hash_elt = $params{hash};
    my $xpath = $params{path};
    my $add_to_existing_node = $params{add_to_existing_node};
    my $list_element_id = $params{list_element_id};

    my $rv_hash_elt;
    my $xpath_orig = $xpath;

    # this only happens when we are stepping into the first recursion step
    if ( defined $add_to_existing_node && ! $add_to_existing_node ) {
        $xpath = File::Basename::dirname($xpath);
    }

    if (
         my $found = F5::XMLHash::get_xpath(
            xpath => $xpath,
            elt_ref => $hash_elt,
            base => $self->{export_base},
            list_element_id => $list_element_id,
         )
       )
    {
        $rv_hash_elt = $found;
    }
    else {
        $rv_hash_elt = F5::XMLHash::new_elt( elt_name => File::Basename::basename($xpath) );

        my $rv = $self->add_or_find_node(
            hash => $hash_elt,
            path => File::Basename::dirname($xpath),
        );

        F5::XMLHash::paste(
            parent => $rv,
            child => $rv_hash_elt,
        );
    }

    # this only happens when we are stepping back out of the first recursion step
    if ( defined $add_to_existing_node && ! $add_to_existing_node ) {
        my $child_hash = F5::XMLHash::new_elt( elt_name => File::Basename::basename($xpath_orig) );

        F5::XMLHash::paste(
            parent => $rv_hash_elt,
            child => $child_hash,
        );

        return $child_hash;
    }

    return $rv_hash_elt;
}

#-----------------------------------------------------------
###############################################################################

#########################################
#  setting limit clauses and value maps #
#########################################
###############################################################################
#-----------------------------------------------------------
sub add_negsig_signature_limit_clause {
    my ( $self, %params ) = @_;
    $params{limit_clause} ||= [];
    my $limit_clause;
    if ( $self->{flg_is_whitehat} ) {
        $limit_clause = {
            field_name => 'flg_disabled',
            field_value => 0,
        };
    }
    else {
        $limit_clause = {
            field_name => 'flg_disabled',
            field_value => 1,
        };
    }

    push @{ $params{limit_clause} }, $limit_clause;
    $self->append_limit_clauses(
        node => $params{node},
        limit_clause => $params{limit_clause},
    );
}

#-----------------------------------------------------------
sub set_value_map {
    my ( $self, %params ) = @_;

    my $field_index = $params{field_index};
    my $node = $params{node};
    my $db_value_map = $params{db_value_map};
    return unless $db_value_map;

    my $f = $node->{fields}->[$field_index];
    return if $f->{export_value_map} || $f->{field_func};

    $f->{field_func} = 'convert_by_value_map';
    $f->{export_value_map} = $db_value_map;
}

#-----------------------------------------------------------
sub append_limit_clauses {
    my ( $self, %params ) = @_;

    return unless scalar @{ $params{limit_clause} };

    my $limit_clause = $params{limit_clause};
    my $node = $params{node};
    my $unique_clauses = [];
    my $seen_clauses = {};

    for my $clause ( @{$limit_clause}, @{ $node->{limit_clause} } ) {
        if ( !exists $seen_clauses->{ $clause->{field_name} } ) {
            $seen_clauses->{ $clause->{field_name} } = 1;
            push @{$unique_clauses}, $clause;
        }
    }

    $node->{limit_clause} = $unique_clauses;
}

#-----------------------------------------------------------
###############################################################################


##################################
#  value map population methods  #
##################################
###############################################################################
#-----------------------------------------------------------
sub populate_url_value_map {
    my ( $self, %params ) = @_;

    return if (exists $self->{url_value_map}->{url_name} &&
               exists $self->{url_value_map}->{is_wildcard} &&
               exists $self->{url_value_map}->{app_id} &&
               exists $self->{url_value_map}->{method_name}
              );

    $self->{url_value_map}->{url_name} = [];
    $self->{url_value_map}->{is_wildcard} = [];
    $self->{url_value_map}->{app_id} = [];
    $self->{url_value_map}->{method_name} = [];

    my $iter = F5::ASMConfig::Entity::URL->retrieve_all(
        dbh => $self->{dbh},
        policy => { policy_id => $self->{policy_id} },
    );
    while ( my $ent_rec = $iter->get_next ) {
        push @{$self->{url_value_map}->{url_name}}, ($ent_rec->{url_id} , $ent_rec->{url_name});
        push @{$self->{url_value_map}->{is_wildcard}}, ($ent_rec->{url_id} , $ent_rec->{is_wildcard} ? 'wildcard' : 'explicit');
        push @{$self->{url_value_map}->{app_id}}, ($ent_rec->{url_id} , $ent_rec->{app_id} ? 'HTTPS' : 'HTTP');
        push @{$self->{url_value_map}->{method_name}}, ($ent_rec->{url_id} , $ent_rec->{method_name});
    }
}
#-----------------------------------------------------------
sub populate_extraction_value_map {
    my ( $self, %params ) = @_;

    return if exists $self->{extraction_value_map}->{extract_name};

    $self->{extraction_value_map}->{extract_name} = [];

    my $ary_ref = $self->{dbh}->selectcol_arrayref(<<EndQuery, { Columns=>[1, 2] }, $self->{policy_id});
SELECT extract_attrib_id, param_attrib_id
  FROM PLC.PL_DYNAMIC_PARAM_EXTRACTS
 WHERE policy_id = ?
EndQuery

    my %extract_id_to_param_id = @$ary_ref;

    if (keys %extract_id_to_param_id) {
        my $iter = F5::ASMConfig::Entity::Extraction->retrieve_all(
            dbh => $self->{dbh},
            policy => { policy_id => $self->{policy_id} },
        );
        while ( my $ent_rec = $iter->get_next ) {
            if ( $ent_rec->{extract_name} && $ent_rec->{extract_id} && exists $extract_id_to_param_id{$ent_rec->{extract_id}} ) {
                push @{ $self->{extraction_value_map}->{extract_name} },
                      ( $extract_id_to_param_id{$ent_rec->{extract_id}} , $ent_rec->{extract_name} );
            }
        }
    }
}
###############################################################################


######################################################
#  field/node/policy/metachar export/fixup functions #
######################################################
###############################################################################
#-----------------------------------------------------------
sub export_policy {
    my ( $self, %params ) = @_;
  
    delete $self->{user_defined_sig_id_to_sig_name}; # Clear cache just in case

    my $policy_id = $params{policy_id};
    my $rv_filename = $params{rv_filename};
    my $minimal_export = $params{minimal_export};

    my $include_scanner_tables =
      ( defined $params{include_scanner_tables} )
      ? $params{include_scanner_tables}
      : 1;

    $self->{policy} = $params{policy};
    $self->{policy_id} = $policy_id;

    # this will croak if Encode does not recognoze the encoding
    $self->{decoder} = Encode::find_encoding( $self->get_primary_encoding($policy_id) );
    # for UTF-8 encoding, 'name()' returns 'utf-8-strict' , 'mime_name()' returns 'utf-8'. we need the later.
    $self->{primary_encoding} = lc $self->{decoder}->mime_name();

    $self->{initial_memory_usage} = F5::NwdUtils::get_process_stat()->{vmsize};

    # determine if flg_is_whitehat
    $self->{flg_is_whitehat} = 0;
    my $scanner_config = F5::ASMConfig::Entity::ScannerConfig->new(
        dbh => $self->{dbh},
        get_criteria => {},
    );
    if ($scanner_config->{scanner_type} && $scanner_config->{scanner_type} =~ /^sentinel/) {
         $self->{flg_is_whitehat} = 1;
    }

    # place main xml header
    # --------------------------------------------
    my $xml_document_header = <<XmlHeader;
<?xml version="1.0" encoding="utf-8"?>
XmlHeader

    my $outf;
    # the contents that will be written to the file should alredy be utf-8 encoded
    open( $outf, '>', $params{filename} ) or die "$!";
    print $outf $xml_document_header;
    # --------------------------------------------

    my $section_id = 0;
    my $fixup_response_page_on_next_pass = 0;
    my $mandatory_policy_child = '';
    eval {
        for my $node ( @{ $self->{config}->{nodes} } ) {
            my @sections = split('/',$node->{xpath});
            my $section = $sections[2] || 'main';

            if ( !$self->{sections}->{$section} ) {
                if ($section_id) {
                    $self->write_xml(
                        file_handle => $outf,
                        fixup_xml => $fixup_response_page_on_next_pass,
                        mandatory_policy_child => $mandatory_policy_child,
                        node => $self->{config}->{nodes}->[$section_id - 1],
                    );
                }

                if ( $fixup_response_page_on_next_pass ) {
                    $fixup_response_page_on_next_pass = 0;
                }

                if ( $mandatory_policy_child ) {
                    $mandatory_policy_child = '';
                }

                $self->{sections}->{$section} = $section_id++;
                $self->{export_base} = { 'policy' => [] };
            }

            if ( $node->{fixup_response_page} ) {
                $fixup_response_page_on_next_pass = 1;
            }

            if ( $node->{mandatory_policy_child} ) {
                $mandatory_policy_child = $section;
            }

            $self->export_node(
                node => $node,
                policy_id => $policy_id,
                hash => $self->{export_base},
                minimal_export => $minimal_export,
                include_scanner_tables => $include_scanner_tables,
            );
        }

        $self->write_xml(
            file_handle => $outf,
            node => $self->{config}->{nodes}->[$section_id - 1],
        );
    };

    if ($@) {
        my $error_msg = $@;
        if ( ref($@) ) {
            $error_msg = $@->{error_msg};
        }

        my $rv = $self->fatal_error( msg => $error_msg );

        if ($rv_filename) {
            File::Slurp::write_file($rv_filename, JSON::to_json($rv, { allow_blessed => 1, convert_blessed => 1 }));
        }

        return $rv;
    }

    # place main xml footer
    # --------------------------------------------
    my $xml_document_footer = <<XmlFooter;
</policy>
XmlFooter

    print $outf $xml_document_footer;
    close $outf;
    # --------------------------------------------

    # set integrity check string
    # --------------------------------------------
    # Digest::MD5 provides addfile method for more efficient reading of files.
    # the file contents are alredy utf-8 encoded - which is what we need 
    open (my $inf, "<", $params{filename}) or die "$!";
    my $hexdigest = Digest::MD5->new->addfile($inf)->hexdigest;
    close $inf;

    # This way the file is not loaded into memory
    #    and we need not traverse the whole file.
    tie my @lines, 'Tie::File', $params{filename} or die "$!";
    $lines[1] =~ s/$INTEGRITY_CHECK_REPLACE_STRING/$hexdigest/;
    untie @lines;
    # --------------------------------------------

    my $success_msg = "Policy '$self->{policy_name}' was successfully exported.";

    my $rv = $self->completed_successfully( msg => $success_msg );

    if ($rv_filename) {
        File::Slurp::write_file($rv_filename, JSON::to_json($rv, { allow_blessed => 1, convert_blessed => 1 }));
    }

    return $rv;
}

#-----------------------------------------------------------
sub write_xml {
    my ( $self, %params ) = @_;

    my $outf = $params{file_handle};
    my $fixup_xml = $params{fixup_xml};
    my $mandatory_policy_child = $params{mandatory_policy_child};
    my $node = $params{node};

    my $xml = XML::Hash::LX::hash2xml( $self->{export_base}, encoding => 'utf-8', doc => 1 )->toString(1);

    if ($xml) {
        substr($xml, 0, 39) = ''; # strip header - <?xml version="1.0" encoding="utf-8"?>
        substr($xml, -10) = ''; # strip footer - </policy> | <policy/>
    }

    if ($mandatory_policy_child && !$xml) {
        $xml= <<Xml;
<policy>
  <$mandatory_policy_child/>
Xml
    }

    if ($xml) {
        # strip the header <policy> but not the <policy integrity_check="..."> header
        substr($xml, 0, 9) = '' unless $node->{master_node};

        # control characters under 0x20 (with the exception of CR/LF/TAB) are not allowed.
        $xml =~ s/[\x00-\x08\x0B\x0C\x0E\x0F\x10-\x1F]//g;

        if ( $fixup_xml ) {
            $xml =~ s/&#13;/\r/g;
            $xml =~ s/&gt;/>/g;
        }

        print $outf $xml;
    }
}

#-----------------------------------------------------------
sub get_value_to_export {
    my ( $self, %params ) = @_;

    my $rec = $params{record};
    my $fields = $params{fields};
    my $ent_config = $params{ent_config};

    my $field = shift @{$fields};

    if ( my $ref_type = ref($rec->{$field}) ) {
        if ( $ref_type eq 'ARRAY' ) {
            if ( ref($rec->{$field}->[0]) ) {
                my $rv;
                my $db_map;
                $ent_config = $ent_config->{$field}->{fields};
                for my $sub_rec (@{ $rec->{$field} }) {
                    my @fields_cpy = @{$fields};
                    my $sub_rv = $self->get_value_to_export(record => $sub_rec,
                                                            fields => \@fields_cpy,
                                                            ent_config => $ent_config,
                                                           );
                    $db_map = $sub_rv->{db_map} unless $db_map;
                    for my $val (@{$sub_rv->{values_to_export}}) {
                        push @{$rv} , $val ;
                    }
                }
                return { values_to_export => $rv , db_map => $db_map };
            }
            else {
                return { values_to_export => $rec->{$field} , db_map => $ent_config->{$field}->{db_value_map} };
            }
        }
        else {
            return $self->get_value_to_export(record => $rec->{$field},
                                              fields => $fields,
                                              ent_config => $ent_config->{$field}->{fields},
                                             );
        }
    }
    else {
        return { values_to_export => [ $rec->{$field} ] , db_map => $ent_config->{$field}->{db_value_map} };
    }
}

#-----------------------------------------------------------
sub export_node {
    my ( $self, %params ) = @_;
    my $policy_id = $params{policy_id};
    my $node = $params{node};
    my $dbh = $self->{dbh};

    my $base_xpath =
      ( $params{base_xpath} )
      ? $params{base_xpath} . '/' . $params{node}->{xpath}
      : $params{node}->{xpath};

    my $parent_entity = $params{parent_entity};
    my $parent_entity_name = $params{parent_entity_name};

    my $is_subnode = 0;
    if ($parent_entity_name) {
        $is_subnode = 1;
    }

    # TODO: move limit clauses back to export_xml_policy (and load defaults, maybe?)
    #       this is necessary to eventually migrate xpath for fields into TrafficShield/asm-config/yaml/elements/*
    #       and get rid of the yamls files
    # ------------------------------------------
    my $ent = $node->{entity};
    my $xpth = $node->{xpath};
    $xpth =~ s/\//_s_/g;
    my $node_file = $ent . '_x_' . $xpth . '.yaml';

    if ( exists $self->{loaded_files}->{$node_file} ) {
        $node = $params{node} = $self->{loaded_files}->{$node_file};
    }
    # ------------------------------------------
    my $short_entity_name = $node->{entity};
    my $ent_config = $self->{entity_schema}->{$short_entity_name};
    my $entity = $ent_config->{entity_type};

    if ( $node->{policy_export_flag} && ! $params{ $node->{policy_export_flag} } )
    {
        $self->{logger}->debug( "Export $node->{policy_export_flag} not selected -- $node->{xpath} skipped" ) if $self->{logger};
        return;
    }
    if ($entity->feature_name && !$self->{enabled_features}{$entity->feature_name} && !$ENV{F5WAF_ENABLE_ALL_FEATURES}) {
        return;
    }

    $self->fixup_node(%params);

    $node->{limit_clause} ||= [];

    my $subnode_search_filter = {};

    # TODO: to Use YAML constant and rest_filter_comparison_map rather than eval
    my $search_filters = [];
    if ( $node->{limit_clause} ) {
        # all subnode records (say BF) of a specific parent node record (say URL) will have same search filters
        # thus we only create those once and re-use instead of creating them once per every parent URL record handled
        if ( $is_subnode && exists $self->{subnode_search_filter_cache}->{$short_entity_name}->{$parent_entity_name} ) {
            $subnode_search_filter = $self->{subnode_search_filter_cache}->{$short_entity_name}->{$parent_entity_name};
        }
        else {
            for my $lim ( @{ $node->{limit_clause} } ) {

                my $f_name = $lim->{field_name};
                my $f_comparison = ( $lim->{field_comparison} )
                    ? eval "$lim->{field_comparison}"
                    : F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ();

                my $field_value = $lim->{field_value};
                if ( $is_subnode && $parent_entity_name && ref $field_value ) {
                    $field_value = $field_value->{$parent_entity_name};
                }

                my $f_value = ( $lim->{field_value_type} && $lim->{field_value_type} eq 'asmconfig' )
                    ? eval "$field_value"
                    : $field_value;

                if ( $is_subnode && $f_comparison == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() ) {
                    $subnode_search_filter = $self->{subnode_search_filter_cache}->{$short_entity_name}->{$parent_entity_name} = { $f_name => $f_value };
                } elsif ( $is_subnode ) {
                    $subnode_search_filter = $self->{subnode_search_filter_cache}->{$short_entity_name}->{$parent_entity_name} = { };
                }

                if ( !$is_subnode ) {
                    push @{$search_filters},
                      F5::ASMConfig::SearchFilter->new(
                        {
                            field_name => $f_name,
                            field_comparison => $f_comparison,
                            field_value => $f_value,
                        }
                      );
                }
            }
        }
    }
    if ($entity->fields->{feature_name} && !$ENV{F5WAF_ENABLE_ALL_FEATURES}) {
        my @enabled_features = keys %{$self->{enabled_features}};
        push @{$search_filters},
          F5::ASMConfig::SearchFilter->new( field_name => 'feature_name',
                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                            field_value_list => \@enabled_features,
                                          );
    }

    # TODO: to Use YAML constant and rest_filter_comparison_map rather than eval
    my $sort_orders = [];
    if ( !$is_subnode && $node->{order_by} ) {
        for my $ord ( @{ $node->{order_by} } ) {
            push @{$sort_orders},
              F5::ASMConfig::SortOrder->new(
                {
                    field_name => $ord->{field_name},
                    sort_direction => ( $ord->{sort_direction} )
                        ? eval "$ord->{sort_direction}"
                        : F5::ASMConfig::SortDirection::SORT_DIRECTION_ASC(),
                }
              );
        }
    }

    my $iter;
    if ( $is_subnode ) {
        my $parent_id_field = $node->{parent_id_field};
        if (ref $parent_id_field) {
            $parent_id_field = $parent_id_field->{$parent_entity_name};
        }

        my $get_criteria;
        if ( exists $self->{get_criteria_cache}->{$short_entity_name}->{$parent_entity_name} ) {
            $get_criteria = $self->{get_criteria_cache}->{$short_entity_name}->{$parent_entity_name};
            $get_criteria->{$parent_id_field} = $parent_entity;
        }
        else {
            my %search_params = (%{$subnode_search_filter}, $parent_id_field => $parent_entity);
            my $current_entity_asm_config = 'F5::ASMConfig::' . $short_entity_name;
            $get_criteria = $self->{get_criteria_cache}->{$short_entity_name}->{$parent_entity_name} = $current_entity_asm_config->new(%search_params);
        }

        # compensate for Bug 480323
        # --------------------------------------
        my $d1 = F5::NwdUtils::get_process_stat()->{vmsize};
        my %ret_params = (
            dbh => $dbh,
            get_criteria => $get_criteria,
            policy => { policy_id => $self->{policy_id} },
        );
        $iter = $entity->retrieve_all(%ret_params);
        my $d2 = F5::NwdUtils::get_process_stat()->{vmsize};
        $self->{memory_usage_iterators} += ($d2 - $d1);
        # --------------------------------------
    }
    else {
        # compensate for Bug 480323
        # --------------------------------------
        my $d1 = F5::NwdUtils::get_process_stat()->{vmsize};
        $iter = $entity->retrieve_all(
            dbh => $dbh,
            policy => { policy_id => $self->{policy_id} },
            search_filters => $search_filters,
            sort_orders => $sort_orders,
            page_size => $node->{limit} || 0,
        );
        my $d2 = F5::NwdUtils::get_process_stat()->{vmsize};
        $self->{memory_usage_iterators} += ($d2 - $d1);
        # --------------------------------------
    }

    if ($node->{subnodes}) {
        for my $subnode ( @{ $node->{subnodes} } ) {
            if ( ! exists $self->{subnode__exists}->{$subnode->{entity}} ) {
                my $sub_entity = $self->{entity_schema}->{$subnode->{entity}}->{entity_type};

                # compensate for Bug 480323
                # --------------------------------------
                my $d1 = F5::NwdUtils::get_process_stat()->{vmsize};
                my $sub_iter = $sub_entity->retrieve_all(
                    dbh => $dbh,
                    policy => { policy_id => $self->{policy_id} },
                );
                my $d2 = F5::NwdUtils::get_process_stat()->{vmsize};
                $self->{memory_usage_iterators} += ($d2 - $d1);
                # --------------------------------------

                $self->{subnode__exists}->{$subnode->{entity}} = ( $sub_iter->{total_record_count} ) ? 1 : 0 ;
            }
        }
    }

    my $load_defaults = 0;
    if ( $node->{load_defaults} && ref( $node->{load_defaults} ) ) {
        $load_defaults = 1;
    }

    # traverse the asm-config search result records
    my $found_nodes = 0;
    while ( my $ent_rec = $iter->get_next ) {
        $found_nodes = 1;

        if ( $short_entity_name eq 'Policy' ) {
            $self->{policy_name} = $ent_rec->{policy_name};
        }
        
        # sorting nested lists by spesific field
        if ( $node->{nested_list_sort_by} ) {
            for my $ord ( @{ $node->{nested_list_sort_by} } ) {
                my $nested_list_field = $ord->{nested_list_field};
                my $sort_by_field = $ord->{sort_by_field};

                my @sorted = map  { $_->[0] }
                             sort { $a->[1] <=> $b->[1] }
                             map  { [$_, $_->{$sort_by_field} ] }
                                 @{$ent_rec->{$nested_list_field}};

                $ent_rec->{$nested_list_field} = \@sorted;
            }
        }

        # not everything can be filtered out via search_filters
        # so we have alternative filtering methods
        if ( $node->{skip_record_if} ) {
            my $skip_record = 0;
            for my $k ( keys %{ $node->{skip_record_if} } ) {
                if ( defined $ent_rec->{$k} && $ent_rec->{$k} eq $node->{skip_record_if}->{$k} ) {
                    $skip_record = 1;
                    last;
                }
            }
            next if $skip_record;
        }

        # for each field that we need to export in the current node
        #   find the related asm-config entity field in the current search result record
        #   and use it's value for export
        my $field_index = 0;

        # node polution by previous records
        delete $node->{temporary_store}; # cleared per record

        for my $field ( @{ $node->{fields} } ) {
            delete $field->{values_to_export};
            my @drill_down;

            # some fields are there only to activate a field_func
            next unless $field->{entity_field} || $field->{ent_field};

            # TODO refactor the yamls and get rid of ent_field all together
            if    (  $field->{ent_field} && !$field->{entity_field} ) { @drill_down = ( $field->{ent_field} ); }
            elsif ( !$field->{ent_field} &&  $field->{entity_field} ) { @drill_down = @{$field->{entity_field}} ; }
            elsif (  $field->{ent_field} &&  $field->{entity_field} ) { @drill_down = ( $field->{entity_field} , $field->{ent_field} ); }

            my $rec_to_process = $ent_rec;
            my $cfg_to_process = $ent_config;

            # this node field contains a hint about the fact that we need to query another asm-config entity
            # where we should look for the entity field of the current search result record
            # where we should look for the values for export
            if ( my $ext_ent = $field->{entity} ) {
                my $search_filters = [];
                if ( my $filter_by = $ext_ent->{filter_by} ) {
                    push @{$search_filters},
                      F5::ASMConfig::SearchFilter->new(
                        {
                            field_name => $filter_by,
                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                            field_value => $ent_rec->{$filter_by},
                        }
                      );
                }

                my $sub_ent_config = $self->{entity_schema}->{ $ext_ent->{name} };
                my $sub_entity = $sub_ent_config->{entity_type};

                # compensate for Bug 480323
                # --------------------------------------
                my $d3 = F5::NwdUtils::get_process_stat()->{vmsize};
                my $sub_ent_rec = $sub_entity->retrieve_all(
                    dbh => $dbh,
                    policy => { policy_id => $self->{policy_id} },
                    search_filters => $search_filters,
                )->get_next;
                my $d4 = F5::NwdUtils::get_process_stat()->{vmsize};
                $self->{memory_usage_iterators} += ($d4 - $d3);
                # --------------------------------------

                $rec_to_process = $sub_ent_rec;
                $cfg_to_process = $sub_ent_config;
            }

            my $to_export = $self->get_value_to_export(record => $rec_to_process,
                                                       fields => \@drill_down,
                                                       ent_config => $cfg_to_process->{fields},
                                                      );

            $field->{values_to_export} = (defined $to_export->{values_to_export}->[1])
                                          ? $to_export->{values_to_export}
                                          : $to_export->{values_to_export}->[0];

            if ( ( $field->{has_dependencies} || $field->{implied_mask} || $load_defaults )
                 && defined $field->{values_to_export}
                 && ! exists $node->{temporary_store}->{ $field->{ent_field} }
               ) {

                $node->{temporary_store}->{ $field->{ent_field} } = $field->{values_to_export};
            }

            if ( $node->{copy_policy_attributes}
                 && exists $self->{policy}->{ $field->{ent_field} }
                 && !ref($self->{policy}->{ $field->{ent_field} })
               ) {

                $field->{values_to_export} = $self->{policy}->{ $field->{ent_field} };
            }

            $self->set_value_map(
                db_value_map => $to_export->{db_map},
                node => $node,
                field_index => $field_index++,
            );
        }

        #  check if memory usage has grown too much
        if ( $self->{memory_check_count}++ % $MEMORY_CHECK_INTERVAL == 0 ) {
            my $current_memory = F5::NwdUtils::get_process_stat()->{vmsize};

            my $base_line_mem_usage = $self->{initial_memory_usage}   +
                                      $self->{memory_usage_iterators} ;

            if ( $current_memory && $current_memory - $base_line_mem_usage > $MAX_MEMORY_USAGE )
            {
                die "The security policy is too large and cannot be exported as XML. Please use regular export instead.\n";
            }
        }

        if ( $load_defaults && $params{minimal_export} ) {
            my $default_table = $node->{load_defaults}->{defaults_table_name};
            my $join_field = $node->{load_defaults}->{join_field};

            my @where_fields = ($join_field);
            my @where_values = ($node->{temporary_store}->{$join_field});

            if ( $node->{load_defaults}->{where_values} ) {
                for my $field ( keys %{ $node->{load_defaults}->{where_values} } )
                {
                    push( @where_fields, $field );
                    push( @where_values,
                        $node->{load_defaults}->{where_values}->{$field} );
                }
            }

            # this still uses direct SQL due to bz#476237
            my $where_clause = join( " AND ", map { "$_ = ?" } @where_fields );
            my $default_row = $dbh->selectrow_hashref( <<EndQuery, undef, @where_values );
SELECT *
FROM $default_table
WHERE $where_clause
EndQuery

            if ($default_row) {
                my $matches_defaults = 1;
                for my $default_field ( keys %$default_row ) {
                    if ( exists $node->{temporary_store}->{$default_field} &&
                        (
                            ( defined( $default_row->{$default_field} ) != defined( $node->{temporary_store}->{$default_field} ) )
                            ||
                            ( defined( $default_row->{$default_field} )
                              && $default_row->{$default_field} ne $node->{temporary_store}->{$default_field} )
                        )
                       )
                    {
                        $matches_defaults = 0;
                        last;
                    }
                }
                next if $matches_defaults; # skip this record
            }
        }

        my $hash_elt = $self->add_or_find_node(
            hash => $params{hash},
            path => $params{node}->{xpath},
            add_to_existing_node => ( $params{node}->{add_to_existing_node} ) ? 1 : 0,
        );

        if ( $node->{convert_metachar} ) {
            $self->export_metachar(
                node => $node,
                node_hash => $hash_elt,
                minimal_export => $params{minimal_export},
            );
        }
        else {
            my $record_state = 0;

            for my $field ( @{ $node->{fields} } ) {

                # some fields are postponed till after the subnodes are done with
                # this is required to maintain the right order in the resulting xml file
                #     so that it complies with the xsd schema that both import and export are bound to
                # the reason why we need to postone some fields is because they used to be subnodes
                #     but now they are part of the fields of the main node
                # however, they still need to be exported into the xml in th eright order - after the subnodes

                next if $field->{postpone};

                $self->export_field(
                    field => $field,
                    node => $node,
                    element => $hash_elt,
                    field_value => $field->{values_to_export},
                    record_state => \$record_state,
                    base_xpath => $base_xpath,
                );
                last if $record_state eq 'skip';
            }

            if ( $record_state eq 'skip' ) {
                F5::XMLHash::delete_elt(
                    xpath => $base_xpath,
                    base => $self->{export_base},
                );
                next;
            }

            if ($node->{subnodes}) {
                # Add subnodes
                my $parent_entity_name_to_subnode = ref $ent_rec;
                $parent_entity_name_to_subnode =~ s/.*://; # parent short name: F5::ASMConfig::Parameter => Parameter

                for my $subnode ( @{ $node->{subnodes} } ) {
                    next unless $self->{subnode__exists}->{$subnode->{entity}};
                    $self->export_node(
                        node => $subnode,
                        policy_id => $policy_id,
                        hash => $hash_elt,
                        parent_entity => $ent_rec,
                        parent_entity_name => $parent_entity_name_to_subnode,
                        minimal_export => $params{minimal_export},
                        base_xpath => $base_xpath,
                    );
                }
            }

            # now when we are back from handling the subnodes
            #    finish handling the posponed fields of the main node

            for my $field ( @{ $node->{fields} } ) {
                next unless $field->{postpone};
                $self->export_field(
                    field => $field,
                    node => $node,
                    element => $hash_elt,
                    field_value => $field->{values_to_export},
                    record_state => \$record_state,
                    base_xpath => $base_xpath,
                );
            }

        }
    }

    if ( ! $found_nodes && $node->{load_defaults} ) {
        #Export an empty node so the xpath will be found on import, and full defaults will not be created.
        $self->add_or_find_node(
            hash => $params{hash},
            path => $params{node}->{xpath},
            add_to_existing_node => ( $params{node}->{add_to_existing_node} ) ? 1 : 0,
        );
    }
}

#-----------------------------------------------------------
sub fixup_node {
    my ( $self, %params ) = @_;
    my $node = $params{node};
    my $dbh = $self->{dbh};

    return if $node->{fixup_done};

    if ( $params{minimal_export} ) {
        if ( $node->{minimal_export_func} ) {
            my $func_name = $node->{minimal_export_func};
            $self->$func_name(%params);
        }
    }

    for my $field ( @{ $node->{fields} } ) {
        my $need_reverse = 1;
        # Set field convertion functions
        if ( $field->{value_map_lookup} ) {
            my $short_lookup_entity_name = $field->{value_map_lookup}->{entity};
            my $entity = $self->{entity_schema}->{ $short_lookup_entity_name }->{entity_type};
            my $lc_short_lookup_entity_name = lc($short_lookup_entity_name);
            my $value_map_type = $lc_short_lookup_entity_name . '_value_map';

            $need_reverse = 0;

            if ( !$self->{ $value_map_type } || !$self->{ $value_map_type }->{ $field->{value_map_lookup}->{value_field} } ) {
                # compensate for Bug 480323
                # --------------------------------------
                my $d1 = F5::NwdUtils::get_process_stat()->{vmsize};
                # --------------------------------------

                if ( $field->{value_map_lookup}->{customized} ) {
                    my $populate_method = 'populate_' . $value_map_type;
                    $self->$populate_method();
                }
                else {
                    $self->{ $value_map_type }->{ $field->{value_map_lookup}->{value_field} } = [];
                    my $iter = $entity->retrieve_all(
                        dbh => $dbh,
                        policy => { policy_id => $self->{policy_id} },
                    );
                    while ( my $ent_rec = $iter->get_next ) {
                        push @{ $self->{ $value_map_type }->{ $field->{value_map_lookup}->{value_field} } },
                             ( $ent_rec->{ $field->{value_map_lookup}->{join_field} } , $ent_rec->{ $field->{value_map_lookup}->{value_field} } );
                    }
                }

                # compensate for Bug 480323
                # --------------------------------------
                my $d2 = F5::NwdUtils::get_process_stat()->{vmsize};
                $self->{memory_usage_iterators} += ($d2 - $d1);
                # --------------------------------------
            }

            $field->{value_map} = { @{ $self->{ $value_map_type }->{ $field->{value_map_lookup}->{value_field} } } };
            delete $field->{value_map_lookup};
        }

        # Fixup value maps
        if ( !$field->{export_value_map} && $field->{value_map} ) {
            if ($need_reverse) {
                $field->{export_value_map} = { reverse %{ $field->{value_map} } };
            }
            else {
                $field->{export_value_map} = $field->{value_map};
            }
            delete $field->{value_map};
        }

        # Set field func
        if ( ! $field->{field_func} ) {
            if ( $field->{export_value_map} ) {
                $field->{field_func} = 'convert_by_value_map';
            }
            elsif ( $field->{value_type} ) {
                $field->{field_func} = 'convert_by_value_type_' . $field->{value_type};
            }
            elsif ( $field->{value_mask} ) {
                $field->{field_func} = 'convert_to_value_mask';
            }
        }

    }

    $node->{fixup_done} = 1;
}

#-----------------------------------------------------------
sub export_field {
    my ( $self, %params ) = @_;
    my $field = $params{field};
    my $node = $params{node};
    my $hash_elt = $params{element};
    my $field_value = $params{field_value};
    my $record_state = $params{record_state};
    my $base_xpath = $params{base_xpath};

    if ( $field->{included_only_if} &&
        ! $self->check_dependent_field_values(
            dependent_field_values => $field->{included_only_if},
            node => $node,
          )
       )
    {
        return; # skip to the next field
    }

    if (my $func = $field->{skip_field_func}) {
        return if $self->$func(field_value => $field_value,
                               field => $field,
                               element => $hash_elt,
                               base_xpath => $base_xpath,
                               mode => 'export',);
    }

    if ($field->{skip_if_empty} && (! defined $field_value || $field_value eq '')) {
        return;
    }

    my $do_skip_record = 0;
    if ( $field->{if_missing} && $field->{if_missing} eq 'skip_record') {
        $do_skip_record = 1;
    }

    my $do_skip_field = 1;
    if ( $field->{if_missing} && $field->{if_missing} eq 'keep_field') {
        $do_skip_field = 0;
    }

    if ( my $field_func = $field->{field_func} ) {
        $field_value = $self->$field_func(
            field_value => $field_value,
            field => $field,
            element => $hash_elt,
            base_xpath => $base_xpath,
            mode => 'export',
        );
    }

    # these field_values are considered undefined:
    #   undef
    #   []
    #   [ undef ]
    my $field_value_defined = 1;
    $field_value_defined = 0 unless defined($field_value);
    if (ref($field_value) && ref($field_value) eq 'ARRAY') {
        if ( ! scalar @{$field_value} ||
             ( scalar @{$field_value} == 1 && !defined($field_value->[0]) )
           ) {
            $field_value_defined = 0;
        }
    }

    if ( !$field_value_defined ) {
        if ( $do_skip_record ) {
            $$record_state = 'skip';
            return; # skip to the next record
        }
        if ( $do_skip_field ) {
            return; # skip to the next field
        }
    }

    if ( $field->{implied_mask} ) {
        for my $mask_field ( keys %{ $field->{implied_mask} } ) {
            if ( ! (defined $node->{temporary_store}->{$mask_field}) ||
                ! (
                    $node->{temporary_store}->{$mask_field} &
                    $field->{implied_mask}->{$mask_field}
                  )
               )
            {
                return; # skip to the next field
            }
        }
    }

    if ( defined( $field->{xpath} ) ) {
        $self->set_value_for_xpath(
            field => $field,
            value => $field_value,
            element => $hash_elt,
            xpath => $field->{xpath},
        );
    }

}

#-----------------------------------------------------------
sub export_metachar {
    my ( $self, %params ) = @_;
    my $node = $params{node};
    my $hash_elt = $params{node_hash};
    my $minimal_export = $params{minimal_export};

    my $charset_type;
    my $charset;
    for my $field ( @{ $node->{fields} } ) {
        if ( $field->{ent_field} eq 'charset_type' ) {
            $charset_type = $field->{values_to_export};
        } elsif ( $field->{ent_field} eq 'charset' ) {
            $charset = $field->{values_to_export};
        }
    }

    #TODO - need to retrieve the default charsets via asm-config.

    my ($default_charset) = $self->{dbh}->selectrow_array(
        <<EndQuery, undef, $self->{primary_encoding}, $charset_type );
SELECT DC.charset
  FROM DCC.DEFAULT_CHARSETS DC
       JOIN DCC.ENCODINGS E
    ON DC.encoding_id = E.encoding_id
 WHERE E.encoding = ?
   AND DC.charset_type = ?
EndQuery

    my @chars = split( //, $charset );
    my @default_chars = split( //, $default_charset );
    for my $i ( 0 .. $#chars ) {
        if ( !$minimal_export || $chars[$i] != $default_chars[$i] ) {
            my $metachar_hash_elt = F5::XMLHash::new_elt( elt_name => 'metachar' );

            if ($self->{primary_encoding} ne 'utf-8') {
                $i = $self->{decoder}->decode($i);
            }

            F5::XMLHash::set_att(
                elt_ref => $metachar_hash_elt,
                att_name => 'character',
                att_val => $self->convert_to_hex( field_value => $i )->[0],
            );

            F5::XMLHash::set_text(
                elt_ref => $metachar_hash_elt,
                text => $chars[$i] ? 'allow' : 'disallow',
            );

            F5::XMLHash::paste(
                parent => $hash_elt,
                child => $metachar_hash_elt,
            );

        }
    }
}

#-----------------------------------------------------------
###############################################################################

##################
#  field_func(s) #
##################
###############################################################################
sub convert_to_list {
    my ( $self, $value ) = @_;

    return unless defined $value;

    return ( ref($value) && ref($value) eq 'ARRAY' ) ? $value : [ $value ] ;
}

#-----------------------------------------------------------
sub convert_to_value_mask {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};
    my $field = $params{field};

    my $values = $self->convert_to_list($field_value);
    return unless defined $values;

    my @rv = map { $_ & $field->{value_mask} ? 'true' : 'false' } @{$values};
    return \@rv;
}

#-----------------------------------------------------------
sub convert_by_value_map {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};
    my $field = $params{field};

    my $values = $self->convert_to_list($field_value);
    return unless defined $values;

    my @rv = map { $field->{export_value_map}->{$_} } @{$values};
    return \@rv;
}

#-----------------------------------------------------------
sub convert_by_value_type_json_string {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};

    my $values = $self->convert_to_list($field_value);
    return unless defined $values;

    my $rv = '["' . join('","', @{$values}) . '"]';

    return $rv;
}

#-----------------------------------------------------------
sub convert_by_value_type_boolean {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};

    my $values = $self->convert_to_list($field_value);
    return unless defined $values;

    my @rv = map { $_ ? 'true' : 'false' } @{$values};
    return \@rv;
}

#-----------------------------------------------------------
sub convert_by_value_type_date {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};

    my $values = $self->convert_to_list($field_value);
    return unless defined $values;

    my @rv = map { F5::DateTimeUtils::convert_epoch_to_rfc3339($_) } @{$values};
    return \@rv;
}

#-----------------------------------------------------------
sub negsig_set_definition {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};

    return unless defined $field_value;

    my $sets_xml = F5::NegativeSignatures::Set->export_set(
        set_ids => [$field_value],
        dbh => $self->{dbh},
    )->{sets_xml};

    my $sets_hash = XML::Hash::LX::xml2hash($sets_xml);
    my $hash_elt = F5::XMLHash::first_child( elt_ref => $sets_hash );

    if (
         F5::NegativeSignatures::Set->new(
            set_id => $field_value,
            dbh => $self->{dbh},
         )->set_flg_is_user_defined
       )
    {
        return $hash_elt;
    }

    F5::XMLHash::purge_children( elt_ref => $hash_elt );
    return $hash_elt;
}

#-----------------------------------------------------------
sub user_defined_violation_definition {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};

    return unless defined $field_value;

    my $violation = F5::ASMConfig::Entity::Violation->new(
        dbh => $self->{dbh},
        get_criteria => { viol_index => $field_value },
    );

    if ( $violation->{flg_is_user_defined} ) {
        my $viol_hash_elt = XML::Hash::LX::xml2hash( $violation->to_xml->sprint );

        F5::XMLHash::set_tag(
            elt_ref => $viol_hash_elt,
            new_tag => 'properties',
        );

        # <properties> node in XML policy differs from <violation> in XML violation export
        # in the following ways:
        # - viol_index is stored in properties/internal_id, instead of violation/@id attribute
        # - viol_name is taken from properties/icontrol_name (forced already to be identical),
        #     instead of violation/@name
        $self->set_value_for_xpath(
            element => $viol_hash_elt,
            xpath => 'internal_id',
            value => $viol_hash_elt->{properties}->[0]->{'-id'},
        );

        F5::XMLHash::del_att(
            elt_ref => $viol_hash_elt,
            att_name => 'name',
        );
        F5::XMLHash::del_att(
            elt_ref => $viol_hash_elt,
            att_name => 'id',
        );

        return $viol_hash_elt;
    }

    return;
}

#-----------------------------------------------------------

sub handle_sig_name {
    my ( $self, %params ) = @_;
    my $sig_id = $params{field_value};
    return unless $sig_id;

    if(! exists $self->{user_defined_sig_id_to_sig_name}) {
        $self->{user_defined_sig_id_to_sig_name} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_id', undef, 1);
SELECT sig_id, sig_name
  FROM PLC.NEGSIG_SIGNATURES
  WHERE flg_is_user_defined = ?;
EndQuery
    }

    return exists $self->{user_defined_sig_id_to_sig_name}->{$sig_id} ? 
                  $self->{user_defined_sig_id_to_sig_name}->{$sig_id}->{sig_name} : undef;
}

#-----------------------------------------------------------
sub handle_sigfile_timestamp {
    my ( $self, %params ) = @_;

    # Get the timestamp from the last row      
    my $sigfile_timestamp = $self->{dbh}->selectrow_array(<<EndQuery);
SELECT sigfile_timestamp
  FROM PLC.SIGFILE_STATUS
  ORDER BY sigfile_status_id DESC LIMIT 1;
EndQuery
  
    return $sigfile_timestamp;
}

#-----------------------------------------------------------
sub obfuscate {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};

    my $chunk_size = 64;
    my $field_value_length = length($field_value);
    my @string_chunks = ();

    for(my $i = 0; $i < $field_value_length; $i += $chunk_size) {
        push @string_chunks, MIME::Base64::encode_base64(PassCrypt::crypt_password(substr($field_value, $i, $chunk_size)));
    }

    my $arbitrary_delimiter = ':'; # This delimiter is not part of 'base64'

    return join ($arbitrary_delimiter, @string_chunks);
}

sub obfuscate_previous_enforced_rule {
    my $self = shift;
    $self->obfuscate(@_);
}

#-----------------------------------------------------------
sub convert_to_hex {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};

    my $values = $self->convert_to_list($field_value);
    return unless defined $values;

    my @rv = map { sprintf( '0x%x', $_ )  } @{$values};
    return \@rv;
}

#-----------------------------------------------------------
sub convert_in_chunks {
    my ( $self, %params ) = @_;
    my $field_value = $params{field_value};
    my $field = $params{field};
    my $max_chunk_size = $field->{max_chunk_size};
    my $hash_elt = $params{element};

    return unless defined $field_value;

    if ( $field_value =~ /\000/ ) {
        $field_value = MIME::Base64::encode_base64($field_value);

        F5::XMLHash::set_att(
            elt_ref => $hash_elt,
            att_name => 'base64_encoded',
            att_val => '1',
        );
    }
    elsif ($self->{primary_encoding} ne 'utf-8') {
        $field_value = $self->{decoder}->decode($field_value);
    }

    if ( length($field_value) > $max_chunk_size ) {
        while ($field_value) {
            my $chunk_hash_elt = F5::XMLHash::new_elt( elt_name => 'chunk' );

            F5::XMLHash::set_text(
                elt_ref => $chunk_hash_elt,
                text => substr( $field_value, 0, $max_chunk_size, '' ),
            );
            F5::XMLHash::paste(
                parent => $hash_elt,
                child => $chunk_hash_elt,
            );
        }
    }
    else {
        F5::XMLHash::set_text(
            elt_ref => $hash_elt,
            text => $field_value,
        );
    }
}

#-----------------------------------------------------------
sub get_integrity_check {
    return $INTEGRITY_CHECK_REPLACE_STRING;
}

#-----------------------------------------------------------

#Export only
sub get_parent_policy_name {
    my($self, %params) = @_;

    my ($parent_policy_name) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $self->{policy_id});
SELECT name
  FROM PLC.PL_POLICIES p
   JOIN PLC.PL_POLICY_PARENT pa
  ON p.id = pa.parent_policy_id
 WHERE pa.policy_id = ?
EndQuery

    return $parent_policy_name;
}

#-----------------------------------------------------------
sub get_bigip_version {
    return F5::BigipVersionUtils::bigip_version();
}


#----------------------------------------------------------------------

sub is_feature_enabled {
    my ($self, %params) = @_;
    my $feature_name = $params{feature_name} || $params{field}->{feature_name};
    return F5::BigipUtils::is_feature_enabled($feature_name);
}

#----------------------------------------------------------------------

sub is_feature_disabled {
    my ($self, %params) = @_;
    return ! $self->is_feature_enabled(%params);;
}

###############################################################################

###################
# decoding values #
###################
################################################################################
sub recursively_decode {
    my ($self, %params) = @_;
    my $data_ref = $params{data_ref};

    F5::ObjUtils::recursively_callback(data_ref => $data_ref,
                                       callback => sub {my (%prms) = @_;
                                                        ${$prms{val_ref}} = $self->{decoder}->decode( ${$prms{val_ref}} );
                                                       },
                                      );
}
################################################################################

1;
