package F5::FPS::Signatures::Update;

use strict;
use warnings;

use base qw(F5::Sigfile::Update);

use Time::Local;
use Date::Format;
use File::Basename;
use File::Path;
use File::Copy;
use File::Find;
use Data::Dumper; # for dumping hashes
use F5::Logger;
use F5::FPS::Signatures::AutoDownload;
use F5::ASM::COAPI;
use F5::ASM::Tmsh;
use F5::DateTimeUtils;
use File::Temp qw/tempfile tempdir/;
use Digest::MD5;
use XML::Simple;
use F5::Syslog;

# Used to call gpg externally
use IPC::Open3;
use Symbol;
use File::Spec;


our $SIGFILE_TYPE = "FPS Signatures";
our $SAVE_DIR = "/ts/var/fps";

my %STATUS_MAP = ( NONE => 0,
                   SUCCESS => 1,
                   FAILURE => 2,
                   IN_PROGRESS => 3,);

my $GPG_CMD = '/usr/bin/gpg';

my $IM_FILE_OUTPUT = 'ISO 9660 CD-ROM';


#----------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;
    $self->SUPER::init(%params);
    $self->{sigfile_type} = $SIGFILE_TYPE;
    $self->{save_dir} = $SAVE_DIR;
    $self->{mcp_status_object} = "antifraud_signatures_update";
    $self->{lock_file} = '/ts/lock/fps_signatures.lock';
    $self->{metadata_validation} = 'fpsAttackSignaturesImage';
    $self->{automatic_mode} = $params{automatic_mode};
}

#----------------------------------------------------------------------
sub get_schedule_info {
    my ($self, %params) = @_;
    my @schedules;
    my $query = F5::ASM::COAPI::Query->new($self->{mcp_status_object});
    if ($query && @$query) {
        for my $schedule (@$query) {
            my $rv;
            $rv->{auto_update_interval} = "daily";
            $rv->{update_mode} = $schedule->{update_automatically} ? "auto-interval"
                                                                 : "check-for-updates";
            $rv->{last_update_attempt} = $schedule->{last_attempt_datetime};
            push (@schedules, $rv);
        }
    }
    return @schedules;
}

#----------------------------------------------------------------------

sub process_im {
  my ($self, %params) = @_;
  my %rv = $self->SUPER::process_im(%params);
  my $invalid_err = "Invalid update file for $self->{sigfile_type}.\n";
  if ($rv{metadata} && $self->{metadata_validation} && $rv{metadata} !~ /$self->{metadata_validation}/) {
      $self->{logger}->info("Metadata does not match expected validation ($self->{metadata_validation}).");
      die $invalid_err;
  }
  #parse metadata xml
  my $xml = eval { XMLin($rv{metadata}); };
  #catch xml errors
  if($@) { die $invalid_err; }
  #extract version from xml
  if($xml && $xml->{supportedProducts} && $xml->{supportedProducts}->{product}){
      my $product = $xml->{supportedProducts}->{product};
      my $bigip_ver = F5::BigipVersionUtils::bigip_version(max_elements => 2);
      my $update_ver = F5::BigipVersionUtils::max_elements(version => $product->{version}, max_elements => 2);
      #validate metadata version against big-ip
      if($product->{name} eq 'BIG-IP' && $update_ver ne $bigip_ver) {
          my $err = "$self->{sigfile_type} package version is incompatible with this release, expected $bigip_ver, got $update_ver, aborting upgrade.";
          $self->{logger}->info($err);
          die "$err\n";
    }
  }
  #<version>2014/11/14 14:55:00</version>
  if ($rv{metadata} && $rv{metadata} =~ /version>(.+)<\/version/) {
      my $update_datetime = $1;
      $rv{update_time} = F5::DateTimeUtils::convert_datetime_to_epoch($update_datetime);
  }
  elsif ($params{filename} =~ /(\d\d\d\d)(\d\d)(\d\d)_(\d\d)(\d\d)(\d\d)/) {
      my ($year, $month, $day, $hour, $minute, $second) = ($1, $2, $3, $4, $5, $6);
      my $update_datetime = "$year-$month-$day $hour:$minute:$second";
      $rv{update_time} = F5::DateTimeUtils::convert_datetime_to_epoch($update_datetime);
  }
  else {
      $self->{logger}->info("Could not determine update_time from file name.  Using current time.");
      $rv{update_time} = time();
      $rv{using_current_time} = 1;
  }
    return %rv;
}

#----------------------------------------------------------------------

sub apply_changes {
  my($self, %params) = @_;

  my $automatic_mode = $params{automatic_mode};
  my $filename = $params{filename};
  my $username = $params{username};
  my $update_time = $params{update_time};
  my $check_interval = $params{check_interval};
  my $update_mode = $params{update_mode};
  my $schedule_info = $params{schedule_info};
  my $logger = $self->{logger};
  my $timestamp = F5::DateTimeUtils::convert_epoch_to_datetime($update_time);
  my $update_datetime = $update_time;


  my $base_dir = '/ts/var/tmp';
  my $decrypted_filename;
  my $decrypted_is_temporary = 0;

  if ($filename =~ /\.enc$/) {
    #Placeholder code... may go away
    $decrypted_is_temporary = 1;
    my $keyring_path = "/ts/share/negsig/gpg_asm_sigfile_installer";
    $decrypted_filename = "$base_dir/sigfile.tmp.$$.decrypted_sigfile.tgz";

    open(my $null1, ">", File::Spec->devnull);
    open(my $null2, File::Spec->devnull);
    my $gpg_err_fh = gensym;
    my $gpg_pid = open3($null1, $null2, $gpg_err_fh,
                        $GPG_CMD, qw(--always-trust --yes --homedir), $keyring_path,
                        '--output', $decrypted_filename,
                        '--decrypt', $filename);
    my $gpg_err = join('', <$gpg_err_fh>);
    waitpid($gpg_pid, 0);
    if ($?) {
      die "Signature file is not valid\n";
    }
  }
  else {
    $decrypted_filename = $filename;
  }

  my $tmp_base_dir = "$base_dir/sigfile.tmp.$$";
  mkdir $tmp_base_dir;
  system('tar','-zmxf', $decrypted_filename, '-C', $tmp_base_dir);
  unlink $decrypted_filename;
 
  my @files = glob("$tmp_base_dir/*");
  for my $file (@files) {
      $self->load_file_for_datasyncd(orig_filename => $file,
                                     version_create_time => $update_time);
  }

  my $rv = {completed_successfully => 1,
            update_installed => 1,};

  my $sig_version_query = F5::ASM::COAPI::Query->new($self->{mcp_status_object});
  if ($sig_version_query && @$sig_version_query) {
      my ($sig_version) = @$sig_version_query;
      $sig_version->{current_version_create_datetime} = $update_time;
      $sig_version->{title} = $self->generate_lu_title(%params);
      $sig_version->save;
  }
  $self->check_for_updates(%params); #Update available flag

  return $rv;
}


#----------------------------------------------------------------------

sub check_for_updates {
    my ($self, %params) = @_;
    if ($params{check_for_updates}) {
        #Running in special check-for-updates mode
        $self->report_status();
    }
    eval {
        my $autodownload = $self->get_autodownload();
        my $available = $autodownload->is_newer_sigfile_available || 0;
        my $status_query = F5::ASM::COAPI::Query->new($self->{mcp_status_object});

        if($available) {
            F5::Syslog::ltm_syslog('info', "$self->{sigfile_type} LU Available");
        }

        if ($status_query && @$status_query) {
            my ($status) = @$status_query;
            if ($status->{download_available} != $available) {
                $self->{logger}->info("New $self->{sigfile_type} " . ($available ? "is" : "is not") . " available for download");
                $status->{download_available} = $available;
            }
            #always save the time checked for updates
            #whether new update is available or not
            $status->{last_update_check_datetime} = time();
            $status->save();
        }
    };
    my $err = $@;
    if ($params{check_for_updates}) {
        #Running in special check-for-updates mode
        $self->report_status(error_msg => $err,
                             completed_successfully => ($err ? 0 : 1));
    }
}

#----------------------------------------------------------------------

sub start_update {
    my ($self, %params) = @_;
    $self->report_status();
}

#----------------------------------------------------------------------
sub report_status {
  my($self, %params) = @_;
  my $now = time();
  my $status;
  my $query = F5::ASM::COAPI::Query->new($self->{mcp_status_object});
  if ($query && @$query) {
      ($status) = @$query;
  }
  else {
      #TODO: Remove
      print "Can't Update MCP side\n";
      return;
  }
  $status->{progress_status_datetime} = $now;
  $status->{progress_status} = $STATUS_MAP{IN_PROGRESS};
  $status->{message} = $params{msg} || $params{error_msg} || "";
  $status->{install_user} = $self->{username};

  if (defined $params{completed_successfully}) {
      if ($params{completed_successfully}) {
          $status->{progress_status} = $STATUS_MAP{SUCCESS};
          if ($params{update_installed}) {
              $status->{install_user} = $self->{username};
              $status->{install_datetime} = $now;
          }
      }
      else {
           $status->{progress_status} = $STATUS_MAP{FAILURE};
           F5::Syslog::ltm_syslog('err', "$self->{sigfile_type} LU Error: $status->{message}");
      }
  }
  $status->save();

}

#----------------------------------------------------------------------

sub get_autodownload {
    my ($self, %params) = @_;
    return F5::FPS::Signatures::AutoDownload->new(logger => $self->{logger},
            outbound_call_timeout => $self->{outbound_call_timeout},
            );
}

#----------------------------------------------------------------------

sub check_filetype {
    my ($self, %params) = @_;
    return $params{filetype} =~ /$IM_FILE_OUTPUT/;
}

#----------------------------------------------------------------------

sub automatic_download {
    my ($self) = @_;
    my $filename = $self->SUPER::automatic_download();
    if($filename) {
        #a|automatic flag by cron TrafficShield/mng/scripts/tools/update_fps_signatures
        F5::Syslog::ltm_syslog('info', "$self->{sigfile_type} LU ".($self->{automatic_mode} ? "Automatically Downloaded" : "Downloaded"));
    }
    return $filename;
}

#----------------------------------------------------------------------

sub generate_lu_title {
    my ($self, %params) = @_;
    my $base_title = "$self->{sigfile_type} Update File";
    if($params{using_current_time}) {return $base_title;} #return $base_title if update time could not be determined
    my $major = time2str("%M", $params{update_time}) + 0; #minutes without leading zero
    my $minor = time2str("%S", $params{update_time}); #seconds
    return "$base_title - LU$major.$minor"; #FPS Engine Update File - LU7.00
}

#----------------------------------------------------------------------

1;
