package F5::GRPC::Schema::Validate;

use strict;
use warnings;

use JSON::XS;
use F5::Deploy;
use F5::SystemUtils;
use F5::BigipUtils;

my $GRPC_VALIDATION_TIMEOUT = 3; #sec

# CLI utility -
#   /usr/share/ts/bin/is_grpc_proto_valid -f <filename> -u
# returns a list of errors, optionally a list of urls, in JSON format
my $GRPC_PROTO_VALIDATE_TOOL = F5::Deploy::get_path_by_type('executables') . '/is_grpc_proto_valid';

#-------------------------------------------

sub is_grpc_proto_valid {
    my(%params) = @_;

    my $idl_file = $params{idl_file};
    my $proto_dir = $params{proto_dir};
    my $logger = $params{logger};
    my $get_urls = $params{get_urls} || 0;

    my $rv;

    if (! -f $GRPC_PROTO_VALIDATE_TOOL) {
        my $error_msg = 'The gRPC validation utility could not be found on the device';
        $logger->error($error_msg) if $logger;
        $rv->{completed_successfully} = 0;
        $rv->{messages} = [ $error_msg ];
        return $rv;
    }

    # to avoid malformed JSON string error by JSON::XS decode
    my $json_string = "{}";
    
    my @system_tool_params;
    push (@system_tool_params, $GRPC_PROTO_VALIDATE_TOOL, '-f', "$proto_dir/$idl_file");
    if ($get_urls) {
        push (@system_tool_params, '-u');
    }
   
    my $system_tool_res = F5::SystemUtils::run_system_tool(command_line_params => \@system_tool_params,
                                                           timeout => $GRPC_VALIDATION_TIMEOUT,
                                                          );
    if ($system_tool_res && ($system_tool_res->{standard_error} || ! $system_tool_res->{completed_successfully})) {
        if ($logger) {
            my $log_msg =  join (" ", @system_tool_params);
            $logger->error("Error in gRPC file validation - '$log_msg'.");
            $logger->trace($system_tool_res);
        }

        $rv->{completed_successfully} = 0;
        $rv->{messages} = [ 'Error occurred while validating gRPC schema.' ];
    }

    if ($system_tool_res && $system_tool_res->{standard_output}) {
        $json_string = $system_tool_res->{standard_output};
        chomp($json_string);
    }

    my $validation_rv;
    eval {
        $validation_rv = JSON::XS->new->utf8->decode($json_string);
    };

    $rv->{completed_successfully} = $validation_rv->{completed_successfully} || 0;

    if ($validation_rv->{errors}) {
        for my $error (@{$validation_rv->{errors}}) {
            my $msg = "File: $error->{filename}";
            $msg =~ s/$proto_dir\///;
            if ($error->{line} && $error->{line} > 0) {
                $msg .= " line $error->{line}, column $error->{column}";
            }
            $msg .= ": $error->{message}";
            push(@{$rv->{messages}}, $msg );
        }
    }
    if ($get_urls) {
        $rv->{urls} = $validation_rv->{urls} || [];
        $rv->{services} = $validation_rv->{services} || [];
    }

    return $rv;
}

#-------------------------------------------
1;
