package F5::GRPCProfile;

use strict;
use warnings;

use File::Path;
use File::Copy;
use File::stat;
use File::Basename;
use File::Slurp;
use Text::Iconv;
use Digest::MD5;
use YAML::XS;
use F5::Deploy;
use F5::ParseUri;
use F5::DbUtils;
use F5::ContentProfile;
use F5::GRPC::Schema::Validate;
use F5::ASMConfig::Entity::GrpcProfile;
use F5::ASMConfig::Entity::IdlFile;
use F5::SystemUtils;

my $GRPC_SCHEMA_BASE_DIR = F5::Deploy::get_path_by_type('var') . '/grpc/schemas';
my $GRPC_SCHEMA_TMP_BASE_DIR = F5::Deploy::get_path_by_type('tmp') . '/grpc/schemas';
my $PROTOBUF_WELL_KNOWN_TYPES_DIR = F5::Deploy::get_path_by_type('var') . '/protobuf_imports/google';

#----------------------------------------------------------------------
sub clean_unreferenced_profile_files {
  my(%params) = @_;

  for my $dir ($GRPC_SCHEMA_BASE_DIR) {
    F5::DbUtils::delete_unreferenced_files(dbh => $params{dbh},
                                       logger => $params{logger},
                                       db_file_refs => [
                                                        { table_name => 'PLC.PL_CONTENT_PROFILES',
                                                          field => 'id' },
                                                       ],
                                       db_value_to_filename_conversion_func => sub {
                                         my $id = shift;
                                         my $hash = F5::ASMConfig::Cache::profile_id_to_hash(dbh => $params{dbh}, profile_id => $id);
                                         my $policy_name_crc = $hash->{policy_name_crc};
                                         my $profile_rest_uuid = $hash->{profile_rest_uuid};
                                         return "$dir/grpc_profile_${policy_name_crc}__" . $profile_rest_uuid;
                                       },
                                       dir => $dir);
  }
}
#----------------------------------------------------------------------
sub handle_idl_files {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id} || 0;
  my $grpc_profile = $params{grpc_profile};
  my $policy = $params{policy};
  my $policy_name_crc = $params{policy_name_crc};
  my $logger = $params{logger};
  my $get_urls = $params{get_urls};
  my $base_dir = $params{base_dir} || $GRPC_SCHEMA_BASE_DIR;

  if (!$grpc_profile) {
      $grpc_profile = F5::ASMConfig::Entity::GrpcProfile->new(dbh => $dbh,
                                                              get_criteria => {profile_id => $profile_id});
      if (!$grpc_profile) {
           return {
            completed_successfully => 0,
            messages => [ 'No profile found' ],
           };
      }
      $policy = $grpc_profile->policy;
  }
  my $profile_rest_uuid = $params{profile_rest_uuid} || $grpc_profile->{rest_uuid};

  my $profile_idl_files = $grpc_profile->{idl_files} || [];

  my $grpc_schema_dir = "$base_dir/grpc_profile_${policy_name_crc}__${profile_rest_uuid}";
  my $validation_cache_file_name = "$GRPC_SCHEMA_BASE_DIR/grpc_profile_${policy_name_crc}__${profile_rest_uuid}/__grpc_profile.yaml"; #validation cache is always in the CP location, and not written to the BD location
  my $old_validation_cache;
  my $validation_cache;

  if (-f $validation_cache_file_name) {
      $old_validation_cache = YAML::XS::LoadFile($validation_cache_file_name);
  }
  if (-d $grpc_schema_dir) {
      rmtree($grpc_schema_dir);
  }

  my $idl_file_count = scalar @$profile_idl_files;
  if ($idl_file_count == 0) {
    return {
            completed_successfully => 0,
            messages => [ 'No IDL files for this profile' ],
           };
  }

  mkpath($grpc_schema_dir);
  symlink($PROTOBUF_WELL_KNOWN_TYPES_DIR, "$grpc_schema_dir/google");

  $validation_cache->{latest_verification_file_time} = List::Util::max( map { $_->{idl_file}->{last_updated} } @$profile_idl_files );
  $validation_cache->{file_count} = $idl_file_count;

  my $tarfile;
  my $primary_file;
  my $md5_calc = Digest::MD5->new;
  for my $profile_idl_file (@$profile_idl_files) {
    if (!$profile_idl_file->{idl_file}->{file_contents}) {
        $profile_idl_file->{idl_file} = F5::ASMConfig::Entity::IdlFile->new(dbh => $dbh,
                                                                                     get_criteria => $profile_idl_file->{idl_file});
    }

    my $file_name = $profile_idl_file->{idl_file}->{file_name};

    $md5_calc->add($file_name);
    $md5_calc->add($profile_idl_file->{idl_file}->{file_contents});

    if (my $primary_idl_file_name = $profile_idl_file->{primary_idl_file_name}) {
        my $error_message;
        if ($idl_file_count > 1) {
           $error_message = 'Cannot specify more than one IDL file when utilizing an archive';
        }
        else {
            my $grpc_schema_tmp_dir = "$GRPC_SCHEMA_TMP_BASE_DIR/grpc_profile_${policy_name_crc}__${profile_rest_uuid}";
            if (-d $grpc_schema_tmp_dir) {
                rmtree($grpc_schema_tmp_dir);
            }
            mkpath ($grpc_schema_tmp_dir);
            File::Slurp::write_file("$grpc_schema_tmp_dir/$file_name", $profile_idl_file->{idl_file}->{file_contents});
            my $res = F5::Utils::File::extract_archive(filename => $file_name, source_dir => $grpc_schema_tmp_dir, target_dir => $grpc_schema_dir);
            if (!$res->{completed_successfully}) {
                $error_message = $res->{standard_error} || $res->{error_message} || $res->{standard_output};
                $error_message =~ s/$grpc_schema_tmp_dir\///;
            }
            if (-d $grpc_schema_tmp_dir) {
                rmtree($grpc_schema_tmp_dir);
            }
        }
        if ( ! $error_message) {
            $primary_file = $primary_idl_file_name;
            if (! -f "$grpc_schema_dir/$primary_idl_file_name") {
                $error_message = "Primary IDL file '$primary_idl_file_name', does not exist in archive '$profile_idl_file->{idl_file}->{file_name}'";
            }
        }

        if ($error_message) {
            $error_message =~ s/$grpc_schema_dir\///;
            return { completed_successfully => 0,
                     messages => [ $error_message ],};
        }
    }
    else {
        if (defined ($profile_idl_file->{import_url}) ) {
            my $local_dir = $grpc_schema_dir . "/$profile_idl_file->{import_url}";
            mkpath ($local_dir);
            $file_name = "$profile_idl_file->{import_url}/$file_name";
        }
        my $local_filename = "$grpc_schema_dir/$file_name";
        # these files remain on the disk, for later use by BD
        File::Slurp::write_file($local_filename, $profile_idl_file->{idl_file}->{file_contents});
        if (!$primary_file || $profile_idl_file->{flg_is_primary_enforced}) {
            $primary_file = $file_name;
        }
    }
  }

  $validation_cache->{md5} = $md5_calc->hexdigest;

  my $rv;
  if (!$old_validation_cache ||
      ($validation_cache->{md5} ne $old_validation_cache->{md5} ||
       $validation_cache->{file_count} != $old_validation_cache->{file_count} ||
       $validation_cache->{latest_verification_file_time} != $old_validation_cache->{latest_verification_file_time})) {

      $rv = F5::GRPC::Schema::Validate::is_grpc_proto_valid(logger => $logger,
                                                            get_urls => $get_urls,
                                                            idl_file => $primary_file,
                                                            proto_dir => $grpc_schema_dir);
  }
  else {
      $rv->{completed_successfully} = 1;
  }

  if ($rv->{completed_successfully}) {
     YAML::XS::DumpFile($validation_cache_file_name, $validation_cache);
     #Move primary file to unique named location to ensure push to bd on changes
     my $unique_filename = "__$validation_cache->{md5}__$primary_file";
     move("$grpc_schema_dir/$primary_file", "$grpc_schema_dir/$unique_filename");
     $primary_file = $unique_filename;
  }

  return {
          completed_successfully => $rv->{completed_successfully},
          messages => $rv->{messages},
          urls => $rv->{urls},
          services => $rv->{services},
          idl_file => $primary_file,
          proto_dir => $grpc_schema_dir,
         };
}
#----------------------------------------------------------------------
1;
