package F5::GlobalParametersUpgrade;

use strict;
use warnings;

use F5::DbUtils qw/insert_data_to_table/;
use F5::CommonUpgrade::FixParamsNullFields;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(params_upgrade_import set_idmap get_idmap);

use constant DYNAMIC_CONTENT_VALUE    =>   2;
use constant DYNAMIC_PARAMETER_NAME   =>  16;
use constant REGEXP_DYNAMIC_DELIMITER => 255;

use constant URL_DYNAMIC_PARAM => 2;
use constant LINK_DYNAMIC_PARAM => 128;

# key_ui_type
use constant  KEY_UI_TYPE_INTEGER           => 0x2;
use constant  KEY_UI_TYPE_INTEGER_CHECK_RANGE => 0x8;
use constant  KEY_UI_TYPE_DECIMAL           => 0x4;
use constant  KEY_UI_TYPE_DECIMAL_CHECK_RANGE => 0x10;

my $exported_idmap = {};
=pod 

=head1 NAME

F5::GlobalParametersUpgrade - module containing functions necessary for
database changes made for adding global and object parameter/extract
support.

=head1 SYNOPSIS

        use F5::GlobalParametersUpgrade qw(params_upgrade_import);

        my $rv = params_upgrade_import($dbh, $logger, $table_data, undef, undef, $idmap_hash);

=head1 DESCRIPTION


=head1 FUNCTIONS

=head2 params_upgrade_import

Description goes here. #todo

=cut

#--------------------------------------------------------------------
# For importing pre-v4 parameters into new table format
sub params_upgrade_import {
  my($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash, $params) = @_;

  my $new_dbname = 'PLC';
  if ($params) {
    $new_dbname = $params->{databases}->{PLC}->{staging_dbname};
  }

  for my $idmap_field (qw(param_attrib_id extract_attrib_id)) {
    $subst_hash->{$idmap_field} = {} unless $subst_hash->{$idmap_field};
  }

  my @table_data = values %$import_table;

  for my $row (@table_data) {
    $row->{param_attrib_id} = $row->{id};
    $row->{extract_attrib_id} = $row->{id};
    $row->{obj_id} = $row->{extract_from_object_id};
    if ($row->{dynamic_param_type} & URL_DYNAMIC_PARAM) {
      $row->{dynamic_param_type} = ($row->{dynamic_param_type} |  LINK_DYNAMIC_PARAM) ^ URL_DYNAMIC_PARAM;
    }

    if ($row->{dynamic_regexp}) {
      if (index($row->{dynamic_regexp},chr(REGEXP_DYNAMIC_DELIMITER)) >= 0) {
	($row->{regexp_prefix},$row->{regexp_body},$row->{regexp_suffix}) = split(chr(REGEXP_DYNAMIC_DELIMITER), $row->{dynamic_regexp});
      } else {
	$row->{regexp_body} = $row->{dynamic_regexp};
      }
    }
    # Set not null columns to blank
    for my $k (qw(regexp_prefix regexp_body regexp_suffix)) {
      $row->{$k} = '' unless $row->{$k};
    }

    # Remove integer check range or decimal check range flag via mask
    if ($row->{key_ui_type} == KEY_UI_TYPE_INTEGER_CHECK_RANGE) {
      $row->{key_ui_type} = KEY_UI_TYPE_INTEGER;
    }
    elsif ($row->{key_ui_type} == KEY_UI_TYPE_DECIMAL_CHECK_RANGE) {
      $row->{key_ui_type} = KEY_UI_TYPE_DECIMAL;
    }
    else {
      $row->{key_ui_type} &= ~( KEY_UI_TYPE_INTEGER_CHECK_RANGE | KEY_UI_TYPE_DECIMAL_CHECK_RANGE);
    }
  }

  my @parameters_data;
  my %seen_unique_flow_parameters;
  my %merged_parameters;
  for my $row (@table_data) {
    next unless $row->{flow_id};

    my $row_unique_key = "$row->{flow_id},$row->{name}";

    unless ($seen_unique_flow_parameters{$row_unique_key}) {
      push(@parameters_data, $row);
      $seen_unique_flow_parameters{$row_unique_key} = $row->{id};
    }
    else {
      $logger->info("Parameter $row->{id} conflicts with an earlier definition (flow_id = $row->{flow_id}, param name = $row->{name}, conflicts with param id = $seen_unique_flow_parameters{$row_unique_key})");

      # All associated records should now associate with the previous definition
      $merged_parameters{$row->{id}} = $seen_unique_flow_parameters{$row_unique_key};
    }
  }

  if (@parameters_data) {
    my $parameters_table = { map(($_->{param_attrib_id},$_), @parameters_data) };

    my $common_upgrade = F5::CommonUpgrade->new(dbh => $dbh, logger => $logger,);
    $common_upgrade->table_row_fixup(table_name => "PLC.PL_PARAM_ATTRIBUTES",
                                     idmap => $subst_hash,
                                     table_data => $parameters_table,);

    my $param_attrib_idmap = insert_data_to_table(dbh => $dbh,
						  logger => $logger,
						  table_name => "$new_dbname.PL_PARAM_ATTRIBUTES",
						  data => $parameters_table,
						  idmap => $subst_hash,
						  idmap_save => 1,
						  idmap_strictness => 1) or return;

    $subst_hash->{param_attrib_id} = { %{$subst_hash->{param_attrib_id}},
				       %$param_attrib_idmap };

    for my $merged_param_id (keys %merged_parameters) {
      $subst_hash->{param_attrib_id}->{$merged_param_id} = $subst_hash->{param_attrib_id}->{$merged_parameters{$merged_param_id}};
    }

    insert_data_to_table(dbh => $dbh,
			 logger => $logger,
			 table_name => "$new_dbname.PL_FLOW_PARAMS",
			 data => $parameters_table,
			 idmap => $subst_hash,
			 idmap_strictness => 1) or return;

    my $dynamic_param_rows = { map(($_->{param_attrib_id},$_), 
				   grep($_->{extract_from_object_id},
					@parameters_data)) };

    if (%$dynamic_param_rows) {
      my $extract_attrib_idmap = insert_data_to_table(dbh => $dbh,
						      logger => $logger,
						      table_name => "$new_dbname.PL_EXTRACT_ATTRIBUTES",
						      data => $dynamic_param_rows,
						      idmap => $subst_hash,
						      idmap_save => 1,
						      idmap_strictness => 1) or return;
      $subst_hash->{extract_attrib_id} = { %{$subst_hash->{extract_attrib_id}},
					   %$extract_attrib_idmap };
      insert_data_to_table(dbh => $dbh,
			   logger => $logger,
			   table_name => "$new_dbname.PL_OBJECT_EXTRACTS",
			   data => $dynamic_param_rows,
			   idmap => $subst_hash,
			   idmap_strictness => 1,
                           ignore => 1) or return;

      my $dynamic_param_name_rows = { map(($_->{param_attrib_id},$_),
					  grep($_->{key_input_type} == DYNAMIC_PARAMETER_NAME,
					       @parameters_data)) };
      if (%$dynamic_param_name_rows) {
	insert_data_to_table(dbh => $dbh,
			     logger => $logger,
			     table_name => "$new_dbname.PL_DYNAMIC_PARAM_EXTRACTS",
			     data => $dynamic_param_name_rows,
			     idmap => $subst_hash,
			     idmap_strictness => 1) or return;
      }
    }
  }

  my @extract_data = grep(!$_->{flow_id} && $_->{extract_from_object_id}, @table_data);
  if (@extract_data) {
    my $extract_table = { map(($_->{extract_attrib_id},$_), @extract_data) };
    my $extract_attrib_idmap = insert_data_to_table(dbh => $dbh,
						    logger => $logger,
						    table_name => "$new_dbname.PL_EXTRACT_ATTRIBUTES",
						    data => $extract_table,
						    idmap => $subst_hash,
						    idmap_save => 1,
						    idmap_strictness => 1) or return;
    $subst_hash->{extract_attrib_id} = { %{$subst_hash->{extract_attrib_id}},
					 %$extract_attrib_idmap };
    insert_data_to_table(dbh => $dbh,
			 logger => $logger,
			 table_name => "$new_dbname.PL_OBJECT_EXTRACTS",
			 data => $extract_table,
			 idmap => $subst_hash,
			 idmap_strictness => 1,
                         ignore => 1) or return;
  }

  return 1;
}

sub set_idmap
{
   my ($idmap) = @_;
   $exported_idmap = $idmap;
}

sub get_idmap
{
   return $exported_idmap;
}
#----------------------------------------------------------------------

1;

