package F5::GlobalSettings;

use strict;
use warnings;
use JSON;
use File::Slurp qw(read_file);

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless({ %params }, $class);
    $self->init();
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my $self = shift;
    if ($self->{filename}) {
        $self->load_from_file();
    }
    $self->set_values();
}

#----------------------------------------------------------------------

sub apply {
    my($self, %params) = @_;
    my $rv = {};
    my $call_flags = $params{call_flags} || F5::ASMConfig::CallFlags->new();
    if ($self->{cookieProtection}->{seed} && $self->{cookieProtection}->{digest}) {
        my $msg = "Both seed and digest are configured";
        $self->{logger}->error($msg) if $self->{logger};
        $rv->{completed_successfully} = 0;
        $rv->{failure_msg} = $msg;
        return $rv;
    }
    elsif ($self->{cookieProtection}->{seed}) {
        $self->{logger}->info("Data protection cookie configuration=$self->{cookieProtection}->{seed}")  if $self->{logger};
        $self->{logger}->info("Generating new data protection cookie") if $self->{logger};
        $rv = $self->{client}->update_data_protection($self->{client}->{session},
                                                      $call_flags,
                                                      undef,
                                                      F5::ASMConfig::DataProtection->new(secure_key => F5::DataProtection::generate_key($self->{cookieProtection}->{seed}))
                                                     );
        $rv->{configured_settings} = $self->{cookieProtection}->{seed};
    }
    elsif ($self->{cookieProtection}->{digest}) {
        $self->{logger}->info("Data protection cookie configuration=$self->{cookieProtection}->{digest}")  if $self->{logger};
        $self->{logger}->info("Generating new data protection cookie") if $self->{logger};
        eval {
            $rv = $self->{client}->import_data_protection($self->{client}->{session},
                                                          $call_flags,
                                                          undef,
                                                          F5::ASMConfig::DataProtectionImportExport->new(import_text => $self->{cookieProtection}->{digest}),
                                                         );
        };
        my $ex = $@;
        if ($ex) {
            $rv->{completed_successfully} = 0;
            $rv->{failure_msg} = $ex->{error_message};
            return $rv;
        }
        if (!$rv->{completed_successfully}) {
            return $rv;
        }
        $rv->{configured_settings} = $self->{cookieProtection}->{digest};
    }
    if ($self->{attackSignatures}->{securityUpdateFile}) {
        $self->{logger}->info("Updating sigfile $self->{attackSignatures}->{securityUpdateFile}") if $self->{logger};
        eval {
            my $get_file_from_link_rv = F5::Utils::File::get_file_from_link(file_path_map => $self->{file_path_map},
                                                                            link => $self->{attackSignatures}->{securityUpdateFile},
                                                                            allow_local_filename => 1);
            my $sigfile = F5::ASMConfig::Sigfile->new(filename => $get_file_from_link_rv->{filepath});
            my $settings = F5::ASMConfig::SigfileUpdateSettings->new(update_type => F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES(),
                                                                     allow_older_timestamp => 1,
                                                                     im_file => 1);
            $rv = $self->{client}->update_sigfile($self->{client}->{session}, $call_flags, undef, $sigfile, $settings);

            $self->{sigfile_status_id} = $rv->{sigfile_status_id};
        };
        my $ex = $@;
        if ($ex) {
            $rv->{completed_successfully} = 0;
            $rv->{failure_msg} = $ex->{error_message};
            return $rv;
        }
        $self->{logger}->info("Sigfile status id=$self->{sigfile_status_id}") if $self->{logger};
        if (!$rv->{completed_successfully}) {
            return $rv;
        }
    }
    if ($self->{threatCampaigns}->{securityUpdateFile}) {
        $self->{logger}->info("Updating threat campaigns file $self->{threatCampaigns}->{securityUpdateFile}") if $self->{logger};
        eval {
            my $get_file_from_link_rv = F5::Utils::File::get_file_from_link(file_path_map => $self->{file_path_map},
                                                                            link => $self->{threatCampaigns}->{securityUpdateFile},
                                                                            allow_local_filename => 1);
            my $threat_campaigns_file = F5::ASMConfig::ThreatCampaignFile->new(filename => $get_file_from_link_rv->{filepath});
            my $settings = F5::ASMConfig::ThreatCampaignFileUpdateSettings->new(allow_older_timestamp => 1);
            $rv = $self->{client}->update_threat_campaign_file($self->{client}->{session}, $call_flags, undef, $threat_campaigns_file, $settings);
            $self->{threat_campaign_file_status_id} = $rv->{threat_campaign_file_status_id};
        };
        my $ex = $@;
        if ($ex) {
            $rv->{completed_successfully} = 0;
            $rv->{failure_msg} = $ex->{error_message};
            return $rv;
        }
        $self->{logger}->info("Threat campaign file status id=$self->{threat_campaign_file_status_id}") if $self->{logger};
        if (!$rv->{completed_successfully}) {
            return $rv;
        }
    }
    if ($self->{user_defined_signatures_refs}) {
        $rv->{user_defined_signatures} = $self->load_user_defined_signatures();
    }
    return $rv;
}

#----------------------------------------------------------------------

sub load_from_file {
    my($self, %params) = @_;
    my $conf_file_text;
    eval {
        my $get_file_rv = F5::Utils::File::get_data_from_link(file_path_map => $self->{file_path_map},
                                                              link => $self->{filename},
                                                              parse_json => 1,
                                                              allow_local_filename => 1);
        if ($get_file_rv->{completed_successfully}) {
            $self->{conf_file_content} = $get_file_rv->{data};
        }
    };
    my $ex = $@;
    if ($ex) {
        $ex=~ s/ at \S*? line \d+.$//;
        $self->{load_from_file_failure_msg} = "Cannot read configuration file '$self->{filename}': $ex->{error_message}";
        $self->{logger}->error($self->{load_from_file_failure_msg}) if $self->{logger};
        $self->{load_from_file_successfully} = 0;
        return;
    }
}

#----------------------------------------------------------------------

sub set_values {
    my($self, %params) = @_;

    $self->{cookieProtection}->{seed} ||= $self->{cookie_protection__seed} || $self->{conf_file_content}->{wafSettings}->{cookieProtection}->{seed} || $self->{conf_file_content}->{'waf-settings'}->{'cookie-protection'}->{seed};
    $self->{cookieProtection}->{digest} ||= $self->{cookie_protection__digest} || $self->{conf_file_content}->{wafSettings}->{cookieProtection}->{digest};
    $self->{attackSignatures}->{securityUpdateFile} ||= $self->{attack_signatures__security_update_file} || $self->{conf_file_content}->{wafSettings}->{attackSignatures}->{securityUpdateFile};
    $self->{threatCampaigns}->{securityUpdateFile} ||= $self->{threat_campaigns__security_update_file} || $self->{conf_file_content}->{wafSettings}->{threatCampaigns}->{securityUpdateFile};
    $self->{user_defined_signatures_refs} ||= $self->{conf_file_content}->{'waf-settings'}->{'user-defined-signatures'};
}

#----------------------------------------------------------------------

sub load_user_defined_signatures {
    my($self, %params) = @_;
    my @user_defined_signatures;
    my $sigfile;
    my $seen_tags;
    for my $ref ( @{ $self->{user_defined_signatures_refs} } ) {
        my $signature_file = F5::Utils::File::get_file_from_link( link => $ref->{'$ref'}, calc_md5 => 1 );
        my $signature_filepath = $signature_file->{filepath};
        my $signature_md5 = $signature_file->{md5};
        $sigfile = F5::Utils::File::load_file( filename => $signature_filepath, json_only => 1 );
        my $error_message;
        if (!$sigfile->{tag}) {
            $error_message = "Missing tag in file '$signature_filepath'";
        }
        elsif ($seen_tags->{$sigfile->{tag}}++) {
            $error_message = "Duplicate tag '$sigfile->{tag}' in file '$signature_filepath'";
        }
        if ($error_message) {
            $self->{logger}->error($error_message) if $self->{logger};
            my $error = {"error_message" => $error_message};
            die $error;
        }
        $self->{logger}->debug("Importing signatures with tag: '$sigfile->{tag}' from file '$signature_filepath'") if $self->{logger};
        my $update_rv      = $self->{client}->update_sigfile(
            $self->{client}->{session},
            F5::ASMConfig::CallFlags->new(),
            F5::ASMConfig::Sigfile->new(
                filename            => $signature_filepath,
                flg_is_user_defined => 1,
            ),
            F5::ASMConfig::SigfileUpdateSettings->new(
                allow_older_timestamp => 1,
                im_file               => 0,
                update_type           => F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES(),
                validate_requirements => 1,
            )
        );
        my %signature_info = (
            originalFilename => $signature_filepath,
            fileMD5          => $signature_md5,
            tag              => $sigfile->{tag},
        );
        $signature_info{revisionDatetime} = $sigfile->{revisionDatetime} if $sigfile->{revisionDatetime};
        push( @user_defined_signatures, \%signature_info );
    }
    return \@user_defined_signatures;
}

#----------------------------------------------------------------------

sub generate_cookie_seed_string {
    my($self, %params) = @_;
    my @chars       = ( 'a' .. 'z', 'A' .. 'Z', '0' .. '9' );
    my $cookie_seed = join '', map { @chars[ rand @chars ] } 1 .. 16;
    return $cookie_seed;
}

#----------------------------------------------------------------------

sub regenerate_data_protection_configuration {
    my($self, %params) = @_;
    my $data_protection = F5::ASMConfig::DataProtection->new(secure_key => F5::DataProtection::generate_key());
    $data_protection->{grace_accepting_interval} = 1;
    $self->{client}->update_data_protection($self->{client}->{session},
                                            F5::ASMConfig::CallFlags->new(),
                                            undef,
                                            $data_protection);
}

#----------------------------------------------------------------------

1;
