package F5::Graphql::Schema::Parse;

use strict;
use warnings;

use JSON::XS;
use F5::SystemUtils;

my $GRAPHQL_SCHEMA_PARSING_TIMEOUT = 10; #sec

# CLI utility -
#   /usr/share/ts/bin/graphql_schema_parser <schema_file_path> <result_file_path>
# returns errors and warnings in JSON format
my $GRAPHQL_SCHEMA_PARSING_UTILITY = '/usr/share/ts/bin/graphql_schema_parser';

#-------------------------------------------

sub parse_graphql_schema {
    my(%params) = @_;

    my $profile_validation_file = $params{profile_validation_file};
    my $result_file = $params{result_file};
    my $logger = $params{logger};

    my $rv;

    if (! -f $GRAPHQL_SCHEMA_PARSING_UTILITY) {
        my $error_msg = 'The GraphQL schema parsing utility could not be found on the device';
        $logger->error($error_msg) if $logger;
        $rv->{completed_successfully} = 0;
        $rv->{messages} = [ $error_msg ];
        return $rv;
    }

    my @system_tool_params;
    push (@system_tool_params, $GRAPHQL_SCHEMA_PARSING_UTILITY, $profile_validation_file->{local_filename}, $result_file);
    my $system_tool_res = F5::SystemUtils::run_system_tool(command_line_params => \@system_tool_params,
                                                           timeout => $GRAPHQL_SCHEMA_PARSING_TIMEOUT,
                                                          );
    if ($system_tool_res && $system_tool_res->{standard_error}) {
        if ($logger) {
            my $log_msg =  join (" ", @system_tool_params);
            $logger->error("Error running - '$log_msg'");
            $logger->error($system_tool_res);
        }

        $rv->{completed_successfully} = 0;
        $rv->{messages} = [ 'Error occurred while parsing GrapQL schema' ];

        return $rv;
    }

    my $json_string = "{}";
    if ($system_tool_res && $system_tool_res->{completed_successfully} && $system_tool_res->{standard_output}) {
        $json_string = $system_tool_res->{standard_output};
        chomp($json_string);
    }

    my $parsing_rv = JSON::XS->new->utf8->decode($json_string);

    if (ref($parsing_rv->{errors}) eq 'ARRAY' && scalar @{$parsing_rv->{errors}}) {
        $rv->{completed_successfully} = 0;
        $rv->{messages} = [ 'GraphQL schema parsing failed' ];
        push(@{$rv->{messages}}, @{$parsing_rv->{errors}});
    }
    else {
        if (ref($parsing_rv->{warnings}) eq 'ARRAY' && scalar @{$parsing_rv->{warnings}}) {
            $logger->trace("Warnings found while parsing $profile_validation_file->{file_name}") if $logger;
            $logger->trace($parsing_rv->{warnings}) if $logger;
        }

        $rv->{completed_successfully} = 1;
        $rv->{result_file} = $result_file;
        $rv->{messages} = [ 'Successfully parsed GraphQL schema' ];
    }

    return $rv;
}

#-------------------------------------------

1;
