package F5::GraphqlProfile;

use strict;
use warnings;

use File::Path;
use File::Copy;
use File::stat;
use File::Basename;
use Text::Iconv;
use Digest::MD5;
use F5::ParseUri;
use F5::DbUtils;
use F5::ContentProfile;
use F5::ASMConfig::Entity::GraphqlProfile;
use F5::ASMConfig::Entity::IdlFile;
use F5::Graphql::Schema::Parse;
use F5::Deploy;

my $ARBITRARY_DIVIDER = '__//ARBITRARY_DIVIDER//__';
my $GRAPHQL_SCHEMA_BASE_DIR = F5::Deploy::map_path("/ts/var/graphql/schemas");

#----------------------------------------------------------------------
sub clean_unreferenced_profile_files {
  my(%params) = @_;

    F5::DbUtils::delete_unreferenced_files(dbh => $params{dbh},
                                       logger => $params{logger},
                                       db_file_refs => [
                                                        { table_name => 'PLC.PL_CONTENT_PROFILES',
                                                          field => 'id' },
                                                       ],
                                       db_value_to_filename_conversion_func => sub {
                                         return "$GRAPHQL_SCHEMA_BASE_DIR/graphql_profile_" . shift;
                                       },
                                       dir => $GRAPHQL_SCHEMA_BASE_DIR);
}
#----------------------------------------------------------------------
sub handle_idl_files {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $graphql_profile = $params{graphql_profile};
  my $profile_id = $params{profile_id} || ( defined $graphql_profile ? $graphql_profile->{profile_id} : 0 ) ;
  my $policy = $params{policy};
  my $logger = $params{logger};
  my $base_dir = $params{base_dir} || $GRAPHQL_SCHEMA_BASE_DIR;

  if (!$graphql_profile) {
      $graphql_profile = F5::ASMConfig::Entity::GraphqlProfile->new(dbh => $dbh,
                                                                    get_criteria => {profile_id => $profile_id});
      if (!$graphql_profile) {
           return {
            completed_successfully => 0,
            messages => [ 'No profile found' ],
           };
      }
      $policy = $graphql_profile->policy;
  }

  my $profile_idl_files = $graphql_profile->{idl_files} || [];

  my $idl_file_count = scalar @$profile_idl_files;
  if ($idl_file_count == 0) {
    return {
            completed_successfully => 1,
            messages => [ 'No idl files for this profile' ],
           };
  }

  my $graphql_schema_dir = "$base_dir/graphql_profile_${profile_id}";
  if (-d $graphql_schema_dir) {
    rmtree($graphql_schema_dir);
  }
  mkdir ($graphql_schema_dir, 0777);

  my $md5_calc = Digest::MD5->new;
  my $file_count = 0;
  for my $profile_idl_file (@$profile_idl_files) {
    if (!$profile_idl_file->{idl_file}->{file_contents}) {
        $profile_idl_file->{idl_file} = F5::ASMConfig::Entity::IdlFile->new(dbh => $dbh,
                                                                            get_criteria => $profile_idl_file->{idl_file});
    }

    $file_count++;
    if ( $file_count > 1 ) {
        die "GraphQL Profile can not have more than one IDL File connected to it.\n";
    }

    $profile_idl_file->{local_filename} = "$graphql_schema_dir/$profile_idl_file->{idl_file}->{file_name}";

    # these files remain on the disk, for later use by BD
    open(my $outf, '>', $profile_idl_file->{local_filename});
    print $outf $profile_idl_file->{idl_file}->{file_contents};
    close $outf;

    $md5_calc->add($profile_idl_file->{idl_file}->{file_contents});

  }

  my $graphql_md5 = $md5_calc->hexdigest;

  my $rv;

  for my $idl_file (@$profile_idl_files) {
    my $parsed_filename = "$idl_file->{local_filename}.parsed";
    $rv = F5::Graphql::Schema::Parse::parse_graphql_schema(profile_validation_file => $idl_file,
                                                           result_file => $parsed_filename,
                                                           logger => $logger,
                                                          );
    last if !$rv->{completed_successfully};
    $idl_file->{parsed_filename} = $parsed_filename;
  }

  return {
          completed_successfully => $rv->{completed_successfully},
          messages => $rv->{messages},
          profile_idl_files => $profile_idl_files,
          graphql_md5 => $graphql_md5,
         };
}
#----------------------------------------------------------------------
1;
