package F5::Hsl;

use warnings;
use strict;

use constant ERROR_PREFIX      => "F5::Hsl: ";

# hard coded "magic values":
use constant TIME_FORMAT => '%m/%d/%y %H:%M:%S';
use constant FALSE => 0;
use constant TRUE => 1;

# NOTE: it would be an overhead to import entire F5::ASMConfig::Types for corresponding constants
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(FTP SMTP);

#protocol type enum
use constant FTP => 1;
use constant SMTP => 2;


# Carp provides replacements for warn and for die that
# work from the perspective of the caller and provide stack
# trace (useful for debugging and for development).
#           warn  die
use Carp qw/cluck confess/;

# packages

use F5::LockFactory;
use F5::Logger;
use F5::DbUtils;

our $VERSION = 0.01;

=pod 

=head1 NAME

F5::Hsl - an object oriented interface to some ASM
database functionality

=head1 SYNOPSIS

	use F5::Hsl;

	my $db = F5::Hsl->new;

	die $db->error_message unless $db->connected;

	$db->add_profile( 
		profile_name => $profile_name,
		profile_type => $Profile_type,
	) or die $db->error_message;


	$db->delete_profile( profile_name => $profile_name ) 
		or die $db->error_message;

	$db->disable_profile( profile_name => $profile_name ) 
		or die $db->error_message;

        my $default_profile_id = $db->_get_default_profile( profile_type => $profile_type); 
	die $db->error_message if $db->error_message;


=head1 DESCRIPTION

F5::Hsl provides an Object Oriented Perl interface to common
management functionalities of ASM that are
database dependent, while hiding implementation details from
the user.

=cut

=head1 METHODS

=cut

=head2 new

The constructor is used for generating a new object,
initializing its attributes and connecting to the databases
involved.

Failure is expected upon unsuccessful connection to one of
the underlying databases that are being used. Failure in
connection can be checked using the $db->connected
predicate.

Supported optional arguments:

=over 1

=item databse_username 

=over 1

=item a username for accessing the databases (default is 'root')

=back

=item database_password

=over 1

=item a password for accessing the databases (no default)

=back

=item database_type

=over 1

=item database type (default is 'mysql')

=back

=item debug

=over 1

=item debug level (flag).

=back

=item log_file

=over 1

=item log filename. You'd normally want to use the logfile of the script that
uses F5::Hsl.

=back

=item print_target

=over 1

=item print target used for debugging.

=back

=back

A default will be used instead of any omitted argument.
		
An object is returned upon success or failure. Failure is
expected in case a database connection was not established.
This can be checked using the $db->connect predicate. In
case of errors, an apropriate error message can be obtained
by calling $db->error_message.

=cut

sub new {
	my ($class, %arguments) = @_;

	my $error_message = '';
	my $atts = { 
		PrintError=>1,
		RaiseError=>0,
		Taint => 1,
		FetchHashKeyName   => 'NAME_lc',
	};

	my $dbh = $arguments{dbh} || F5::DbUtils::get_dbh(db_opts => $atts);
	$error_message.=ERROR_PREFIX.$DBI::errstr unless $dbh;
	my $logger = $arguments{logger} || F5::Logger->get_logger(logger_name => 'hsl');
	
	my $locker = F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
	                                         logger => $logger,);

	bless {
		_dbh               => $dbh,
		_error_message     => $error_message, 
        _locker            => $locker,
        logger             => $logger,
	} , $class;
}

=head2 connected
 
This method serves as a predicate and returns a true value
if a successful connection to the database was made during
instanciation and a false value if no successful connection
to the database was made during instanciation.

=cut

sub connected {
	defined shift->{_dbh}
}

=head2 get_dbh

This method returns the DBI database handle for the
connection with the DCC database.

=cut

sub get_dbh {
	shift->{_dbh};
}


=head2 error_message

This method returns the last error message reported

=cut

sub error_message {
	shift->{_error_message};
}

=head2 add_profile

This method is responsible for creating new data
profile with according default hsl profile id

Mandatory arguments:

=over 1

=item B<profile_name>

=over 1

=item data profile name

=back

=item B<profile type>

=over 1 

=item one of supported profile types (FTP, ...)

=back
=back

=cut

sub add_profile {
        
        my $self = shift @_;
        my %args  = (@_);
        my $profile_name = $args{profile_name};
        my $profile_type = $args{profile_type};
        my $flg_security = $args{flg_security} || FALSE;

    $self->{_locker}->lock();

	$self->get_dbh->begin_work or return $self->_error($self->get_dbh->errstr)
		if $self->get_dbh->{AutoCommit};

	# check if data profile name already exists in
	# DCC.HSL_DATA_PROFILES
	# if so, then operation must fail.
	# otherwise, proceed.
	my $hsl_profile_id; ## the $id scalar will be repeatedly used
		## and abused, so
	eval {
		$hsl_profile_id=$self->_is_data_profile_exists($profile_name);                
	};
        return $self->_error($@) if $@;
	if ($hsl_profile_id) {
                
                $self->{logger}->debug("F5::Hsl Profile id  $hsl_profile_id is exists for data-profile-name $profile_name in DCC.HSL_DATA_PROFILES");

                # code for enable existed account
                eval {

                   $hsl_profile_id = $self->_enable_profile($profile_name, $flg_security);
                };
                return $self->_error($@) if $@;
                return $self->_error("Problems set profile to enable into DCC.HSL_DATA_PROFILES",1) 
                        unless $hsl_profile_id;

                # code for checking if profile type was not changed for recieved profile name
                # if it was changed set other default hsl_profile_id

                eval {
                    $hsl_profile_id = $self->_update_hsl_profile_id($profile_name,$profile_type)
                };
                return $self->_error($@) if $@;
                return $self->_error("Problems set new hsl_profile_id into DCC.HSL_DATA_PROFILES",1) 
                        unless $hsl_profile_id;

		$self->get_dbh->commit or eval {$self->get_dbh->rollback}
			unless $self->get_dbh->{AutoCommit};
        $self->_do_unlock();
		return 1; #success
	}

        $self->{logger}->info("Add new profile $profile_name");

	# get default hsl profile for current type data profile type 
	$hsl_profile_id = $self->_get_default_profile($profile_type);
	return $self->_error("Problems computing default hsl profile id")
		unless $hsl_profile_id;

         my $dbh = $self->get_dbh();
         local $" = ','; # so that lists inside double-quotes will be comma-separated
         my @fields_tbl_hsl_data_profiles = qw (hsl_profile_id data_profile_name flg_is_enabled);
         my @placeholders = ('?') x @fields_tbl_hsl_data_profiles;

         my $sql_statement = qq{
                INSERT INTO DCC.HSL_DATA_PROFILES (@fields_tbl_hsl_data_profiles) VALUES (@placeholders)
         };

         my $sth = $dbh->prepare($sql_statement);
         confess("Problems preparing statement:\n$sql_statement")
                unless $sth;

         $sth->execute($hsl_profile_id,$profile_name,$flg_security) or confess("Problems executing:\n$sql_statement");  

         $dbh->commit or eval {$dbh->rollback}
                       unless $dbh->{AutoCommit};

    $self->_do_unlock();
	return 1; # success
}

sub  _update_hsl_profile_id
{

	my ($self, $profile_name,$profile_type) = @_;

        my $dbh = $self->get_dbh();

        # get hsl_profile_id 
        my $sql_statement =
                qq{SELECT hsl_profile_id FROM DCC.HSL_DATA_PROFILES WHERE data_profile_name = ?};

        my ($hsl_profile_id) = $dbh->selectrow_array($sql_statement, undef, $profile_name);
        if ( $dbh->err )
        {
                confess("Problems with statement:\n" .
                    "$sql_statement ($profile_type)\n" .
                    "Error: $DBI::errstr");
                return; 
        }
         
        # get hsl_profile_type_id according to hsl_profile_id
        $sql_statement =
                qq{SELECT hsl_profile_type_id FROM DCC.HSL_PROFILES WHERE hsl_profile_id = ?};

        my ($hsl_profile_type_id) = $dbh->selectrow_array($sql_statement, undef, $hsl_profile_id);
        if ( $dbh->err )
        {
                confess("Problems with statement:\n" .
                    "$sql_statement ($profile_type)\n" .
                    "Error: $DBI::errstr");
                return;
        }

        # check if profile type was not changed for recieved profile name
        if ( $hsl_profile_type_id != $profile_type)
        {
                 # the next situation was occured: the profile from one type has been deleted
                 # and new profile from other type has been added, the problem is that in this
                 # case both profiles have been named with the same name 
             
                 # get default profile for profile type
		 $hsl_profile_id = $self->_get_default_profile($profile_type);
		 return $self->_error("Problems computing default hsl profile id")
                	unless $hsl_profile_id;		

                 # update new hsl_profile_id with default value according to profile type
                 $self->_set_default_hsl_profile_id($profile_name, $hsl_profile_id);
           
        }       

        return $hsl_profile_id;
}

sub _set_default_hsl_profile_id
{
	my ($self, $profile_name, $hsl_profile_id) = @_;

        $self->{logger}->info("Trying to update hsl_profile id : $hsl_profile_id for profile: $profile_name");

        my $dbh = $self->get_dbh;
        my $sql_statement = qq{
                UPDATE DCC.HSL_DATA_PROFILES SET
                        hsl_profile_id=?
                WHERE
                        data_profile_name=?
        };

        my $sth = $dbh->prepare($sql_statement);
        confess("Problems preparing statement:\n$sql_statement")
                unless $sth;

        $sth->execute($hsl_profile_id,$profile_name) or confess("Problems executing:\n$sql_statement");

        return 1; # success        
}


sub _get_default_profile
{
	my ( $self, $profile_type ) = @_;


	 my $dbh = $self->get_dbh();

        my $sql_statement =
                qq{SELECT hsl_profile_id FROM PLC.HSL_PROFILES WHERE hsl_profile_type_id = ? and flg_default = ?};

        my ($hsl_profile_id) = $dbh->selectrow_array($sql_statement, undef, $profile_type,TRUE);

	if ( $dbh->err )
    	{
        	confess("Problems with statement:\n" .
	            "$sql_statement ($profile_type)\n" .
        	    "Error: $DBI::errstr");
	}

        return $hsl_profile_id;

}

sub _is_data_profile_exists {
        my ($self,$profile_name) = @_;
        my $dbh = $self->get_dbh;
        my $sql_statement = qq{
                SELECT hsl_profile_id FROM DCC.HSL_DATA_PROFILES WHERE
                        data_profile_name = ?
        };
        my $sth = $dbh->prepare($sql_statement);
        confess("Problems preparing statement:\n$sql_statement")
                unless $sth;

        $sth->execute($profile_name) or confess("Problems executing:\n$sql_statement");

        my $row = $sth->fetchrow_hashref;
        return $row unless $row;
        return $row->{hsl_profile_id};
}
	
	
# only set the internal error message - don't print ERR message to log
sub _set_error {
	my ($self,$error_string,$concat) = @_;
	if ($concat) {
		$self->{_error_message}.=ERROR_PREFIX.$error_string;
	} else {
		$self->{_error_message}=ERROR_PREFIX.$error_string;
	}
    return undef;
}

sub _error {
	my ($self,$error_string,$concat) = @_;
	eval {$self->get_dbh->rollback unless $self->get_dbh->{AutoCommit}};
    $self->_set_error( $error_string, $concat );
    $self->{logger}->error($self->error_message());
    $self->_do_unlock();
	return undef;
}

sub _do_unlock {
    my ($self) = @_;
    $self->{_locker}->unlock();
}



=head2 disable_profile

This method is responsible for disable profile. To do it
set flag DCC.HSL_DATA_PROFILES.flg_is_enabled to false. 

Mandatory arguments:

=over 1

=item B<profile_name>

=over 1

=item data-profile-name 

=back

=back

=cut

sub disable_profile {
	my ($self,%args) = @_;
	return $self->_enable_profile($args{profile_name}, FALSE);
}


=head2 _enable_profile

This method is responsible for enable or disable profile. To do it
set flag DCC.HSL_DATA_PROFILE.flg_is_enabled to flg_security value. 

Mandatory arguments:

=over 1

=item B<profile_name>

=over 1

=item data-profile-name

=back

=back

=cut

sub _enable_profile {
	my ($self,$profile_name,$flg_security) = @_;
	my $action = ($flg_security == TRUE ? 'enable' : 'disable');
	return $self->_error("No profile name given to $action profile")
		unless $profile_name;
    
	$self->{logger}->debug("Trying to $action profile for profile name |$profile_name|");

        my $dbh = $self->get_dbh;
	my $sql_statement = qq{
		UPDATE DCC.HSL_DATA_PROFILES SET
			flg_is_enabled=?
		WHERE                 
			data_profile_name=? 
	};

	my $sth = $dbh->prepare($sql_statement);
	confess("Problems preparing statement:\n$sql_statement")
		unless $sth;

	$sth->execute($flg_security,$profile_name) or confess("Problems executing:\n$sql_statement");
        
	return 1; # success
}


=This method is responsible for creating new virtual server
=with assigned data profile name

=Mandatory arguments:

=over 1

=item B<vs_name>

=over 1

=item virtual server name

=back

=over 1

=item B<profile_name>

=over 1

=item data profile name

=back

=item B<profile type>

=over 1

=item one of supported profile types (FTP, ...)

=back
=back

=cut


=head2 delete_virtual_server

This method is responsible for delete virtual server. 

Mandatory arguments:

=over 1

=item B<vs_name>

=over 1

=item virtual-server-name

=back

=back

=cut

=pod

=head2 get_profile_names

This method returns a list of profile names as appear in DCC.HSL_DATA_PROFILES.
Optional parameter is a flag of enabled, send it to get only profile names which are enabled/disabled

=cut

sub get_profile_names {
        my ($self, %args) = @_;

        my $profile_type = $args{profile_type};

        my $dbh = $self->get_dbh;
        my $sql_statement = qq{
                SELECT 
                     A.data_profile_name 
                FROM 
                     DCC.HSL_DATA_PROFILES AS A, DCC.HSL_PROFILES AS B 
                WHERE 
                     A.hsl_profile_id=B.hsl_profile_id AND B.hsl_profile_type_id=$profile_type  		 
        };

        if($args{get_enabled_only})
        {
                $sql_statement.= qq{ AND flg_is_enabled=1};
        }
        my $sth = $dbh->prepare($sql_statement);
        return $self->_error("Problems preparing statement:\n$sql_statement")
                unless $sth;
        $sth->execute() or return $self->_error("Problems executing:\n$sql_statement");

        my @data_profile_names;
        while ( my $row = $sth->fetchrow_hashref ) {
                push @data_profile_names, $row->{data_profile_name};
        }
        return @data_profile_names;
}

1;
