package F5::IO::Handle;

use strict;
use warnings;

use IO::Socket::INET6;
use Socket;
use Scalar::Util;
use POSIX qw(:errno_h);
use Time::HiRes;

my $DEFAULT_WRITE_BUFFER_SIZE_LIMIT = 10_000_000; # 10 MB (approx)

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    if (ref($class)) {
        %params = (%$class,
                   %params);
        $class = ref($class);
    }
    my $self = bless {}, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    %params = (in_buf => '',
               out_buf => '',
               new_connection_handler => sub {
                   my($handle, $new_connection) = @_;
                   $handle->{selector}->add($new_connection) if $handle->{selector};
               },
               read_message_handler => sub {},
               before_close_handler => undef,
               socket_close_handler => sub {
                   my ($handle, $errno) = @_;
                   if ($handle->{before_close_handler}) {
                       $handle->{before_close_handler}->($handle, $errno);
                   }

                   $handle->{logger}->debug("Socket close received - normal shutdown") if $handle->{logger};
                   $handle->{selector}->remove($handle) if $handle->{selector};
                   $handle->reconnect() if $handle->{reconnect_on_failure};
               },
               socket_full_initial_handler => sub {},
               other_error_handler => sub {
                   my ($handle, $errno) = @_;
                   $errno ||= '';
                   if ($handle->{before_close_handler}) {
                       $handle->{before_close_handler}->($handle, $errno);
                   }

                   $handle->{logger}->debug("Unknown error (errno: $errno) -- aborting and closing socket") if $handle->{logger};
                   $handle->{selector}->remove($handle) if $handle->{selector};
                   $handle->reconnect() if $handle->{reconnect_on_failure};
               },
               blocking_socket_level_writes => 0,
               reconnect_on_failure => 0,
               flush_write_attempt_limit => 10,
               flush_attempt_delay => 0.5,
               write_buffer_size_limit => $DEFAULT_WRITE_BUFFER_SIZE_LIMIT,
               reconnect_attempt_limit => 0,
               %params);

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }
    Scalar::Util::weaken($self->{selector});

    if (! defined $self->{socket}) {
        if ($self->{server}) {
            $self->{socket} = IO::Socket::INET6->new(LocalAddr => $self->{local_addr} || '*',
                                                     LocalPort => $self->{port},
                                                     Listen    => 10,
                                                     Reuse     => 1,
                                                    );
        }
        else {
            if ($self->{port}) {
                $self->{socket} = IO::Socket::INET6->new(PeerAddr => '127.0.0.1',
                                                         PeerPort => $self->{port},
                                                         Proto    => 'tcp',
                                                        );
            }
        }
        $self->{socket}->blocking($self->{blocking_socket_level_writes}) if $self->{socket};
    }

    if ($self->{selector}) {
        $self->{selector}->add($self);
    }

    $self->{initial_write_succeeded} = 0;
    $self->{partial_message_sent_since_flush} = 0;
    $self->{reconnect_attempt_count} = 0;
    return $self;
}

#----------------------------------------------------------------------

sub reconnect {
    my($self, %params) = @_;

    $self->{reconnect_attempt_count}++;

    if ($self->{port}) {
        $self->{logger}->trace("Attempting to reconnect socket") if $self->{logger};
        $self->{socket} = IO::Socket::INET6->new(PeerAddr => '127.0.0.1',
                                                 PeerPort => $self->{port},
                                                 Proto    => 'tcp',
                                                );

        if ($self->{socket}) {
            $self->{socket}->blocking($self->{blocking_socket_level_writes});
            $self->{logger}->debug("Socket was reconnected") if $self->{logger};
            if ($self->{partial_message_sent_since_flush}) {
                $self->{out_buf} = '';
                $self->{partial_message_sent_since_flush} = 0;
            }
        }
    }
}

#----------------------------------------------------------------------

sub handle_read {
    my($self, %params) = @_;
    if ($self->{server}) {
        my $new_connection = $self->{socket}->accept;
        $new_connection->blocking($self->{blocking_socket_level_writes});
        $self->{new_connection_handler}->($self,
                                          $self->new(server => 0,
                                                     socket => $new_connection),
                                         );
        return 1;
    }
    else {
        my($message, $message_size) = $self->read_message(%params);
        if ($message) {
            $self->{read_message_handler}->($self, $message, $message_size);
            return 1;
        }
    }

    return;
}

#----------------------------------------------------------------------

sub read_message {
    die "Abstract call\n";
}

#----------------------------------------------------------------------

sub write_message {
    die "Abstract call\n";
}

#----------------------------------------------------------------------

sub unshift_to_read_buffer {
    my($self, $buffer) = @_;
    $self->{in_buf} = $buffer . $self->{in_buf};
}

#----------------------------------------------------------------------

sub read {
    my($self, $read_length) = @_;

    my $current_buffer_length = length($self->{in_buf});
    if ($read_length > 0 && $current_buffer_length < $read_length) {
        my $bytes_read = sysread($self->{socket},
                                 $self->{in_buf},
                                 $read_length - $current_buffer_length,
                                 $current_buffer_length);
        if (!defined($bytes_read)) {
            my $file_handle_error = $!;
            if ($file_handle_error == EAGAIN) {
                # No data available at the moment. Try again later.
                $self->{logger}->trace("No data available on read -- will be retried") if $self->{logger};
                return;
            }
            elsif ($file_handle_error == ECONNRESET) {
                # Socket closed
                $self->close(file_handle_error => $file_handle_error);
                return;
            }
            else {
                $self->close(file_handle_error => $file_handle_error);
                return;
            }
        }
        elsif (!$bytes_read) {
            # RPC channel was closed normally
            $self->close();
            return;
        }
    }

    my $retval = substr($self->{in_buf}, 0, $read_length, '');
    return $retval;
}


#----------------------------------------------------------------------

sub write {
    my($self, $buffer) = @_;

    if ($buffer && length($self->{out_buf}) >= $self->{write_buffer_size_limit}) {

        # Attempt to flush, and recheck remaining buffer after flush
        $self->flush();
        if (length($self->{out_buf}) >= $self->{write_buffer_size_limit}) {
            return;
        }
    }

    $self->{out_buf} .= $buffer;

    my $bytes_written;
    eval {
        $bytes_written = syswrite($self->{socket}, $self->{out_buf});
    };
    if ($@) {
        # unknown error
        $self->{other_error_handler}->($self, $!);
    }
    my $file_handle_error = $!;

    if (defined($bytes_written)) {
        $self->{initial_write_succeeded} = 1;
        $self->{partial_message_sent_since_flush} = 1;
        $self->{socket_full} = 0;
        $self->{reconnect_attempt_count} = 0;

        substr($self->{out_buf}, 0, $bytes_written, '');

        if (! $self->{out_buf}) {
            # Write completed successfully
            $self->{partial_message_sent_since_flush} = 0;
        }
    }
    else {
        if ($file_handle_error == EAGAIN) {
            # Socket full reached
            if (!$self->{socket_full}) {
                $self->{socket_full_initial_handler}->($self, $file_handle_error);
            }

            $self->{socket_full} = 1;
        }
        elsif ($file_handle_error == ECONNRESET) {
            # event socket closed
            $self->{logger}->trace("Socket close triggered") if $self->{logger};
            $self->{socket_close_handler}->($self, $file_handle_error);
            return;
        }
        else {
            # unknown error
            $self->{other_error_handler}->($self, $file_handle_error);
            return;
        }
    }

    # Any write that reaches the end is considered a success, even if it was only written to out_buf
    return 1;
}

#----------------------------------------------------------------------

sub is_socket_full {
    my($self) = @_;
    return $self->{socket_full};
}

#----------------------------------------------------------------------

sub rewrite {
    my($self) = @_;

    if (! $self->{last_rewrite_time} || $self->{last_rewrite_time} > Time::HiRes::time() - 0.01) {
        Time::HiRes::usleep(10_000);
    }
    $self->{last_rewrite_time} = Time::HiRes::time();
    $self->write('');
}

#----------------------------------------------------------------------

sub flush {
    my($self, %params) = @_;

    my $write_attempt_limit = defined $params{write_attempt_limit} ? $params{write_attempt_limit} : $self->{flush_write_attempt_limit};
    my $attempt_delay = defined $params{attempt_delay} ? $params{attempt_delay} : $self->{flush_attempt_delay};

    my $write_attempt_count = 0;
    while ($self->{out_buf}) {
        $self->{logger}->trace("Attempting to flush (buffer size = " . length($self->{out_buf}) . ")") if $self->{logger};
        my $preflush_out_buf_size = length($self->{out_buf});
        $self->write('');

        if ($preflush_out_buf_size == length($self->{out_buf})) {
            if ($write_attempt_limit && ++$write_attempt_count >= $write_attempt_limit) {

                if ($self->{reconnect_on_failure} && 
                    (! $self->{reconnect_attempt_limit} || $self->{reconnect_attempt_limit} < $self->{reconnect_attempt_count})) {

                    $write_attempt_count = 0;
                    $self->reconnect();
                    next;
                }

                $self->{logger}->trace("Unable to write after $write_attempt_count attempts -- giving up") if $self->{logger};
                last;
            }

            if ($self->{out_buf}) {
                $self->{logger}->trace("Flush unsuccessful, will retry in $attempt_delay seconds (buffer size = " . length($self->{out_buf}) . ")") if $self->{logger};
                Time::HiRes::sleep($attempt_delay);
            }
        }
        else {
            # out_buf is progressing -- reset attempt count
            $write_attempt_count = 0;
        }
    }

    if ($self->{out_buf}) {
        if (!$self->{socket_full}) {
            $self->{socket_full_initial_handler}->($self, 'could not flush');
        }
        $self->{socket_full} = 1;
        return 0;
    }

    return 1;
}

#----------------------------------------------------------------------

sub close {
    my($self, %params) = @_;

    $self->{socket_close_handler}->($self, $params{file_handle_error});
    if ($self->{socket}) {
        $self->{socket}->close();
    }
}

#----------------------------------------------------------------------

1;
