package F5::IO::LrnPipe;

use strict;
use warnings;

use base qw(F5::IO::Handle);

use Scalar::Util;
use IO::Handle;
use F5::LrnPipe;
use F5::PbRqlPipe;
use F5::BotDefenseRqlPipe;

=head1 CAVEATS

Despite the name, this module wraps shmem pipe reading for both the original F5::LrnPipe and its near-exact copy
F5::PbRqlPipe. This was done because both PBRQL shmem pipe and the RQL pipe are needed simultaneously for asmlogd, and
the same static asm_pipe g_handle (within LrnPipe.xs) can't be re-used within the same Perl VM.

Only a single instance of each pipe type can be used at any one time within the calling code. This is not enforced
by this module internally, and therefore must be enforced by the caller.

Same goes for BOTDEFENSERQL pipe

=cut

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    if (ref($class)) {
        %params = (%$class,
                   %params);
        $class = ref($class);
    }
    my $self = bless {}, $class;
    my $rv = $self->init(%params);
    if (! $rv) {
        return;
    }
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{pipe_open_opts} = {};
    if ($params{pipe_type}) {
        $self->{pipe_open_opts}->{pipe_type} = $params{pipe_type};
    }
    my $pipe_type = $self->{pipe_open_opts}->{pipe_type};

    # Logger intentionally omitted (passed as 'undef')
    #    to avoid ugly ERROR level message from 'F5::PbRqlPipe::open'/'F5::LrnPipe::open'
    my $rv;
    if ($pipe_type eq 'PBRQL') {
        $rv = F5::PbRqlPipe::open(undef, $self->{pipe_open_opts});
    }
    elsif ($pipe_type eq 'BOTDEFENSERQL') {
        $rv = F5::BotDefenseRqlPipe::open(undef, $self->{pipe_open_opts});
    }
    else {
        $rv = F5::LrnPipe::open(undef, $self->{pipe_open_opts});
    }
    if ($rv < 0) {
        $params{logger}->info("Pipe (type: $pipe_type) could not be opened (error code = $rv)") if $params{logger};
        return;
    }

    if ($pipe_type eq 'PBRQL') {
        $self->{socket} = IO::Handle->new_from_fd(F5::PbRqlPipe::get_pipe_fd(), 'r');
    }
    elsif ($pipe_type eq 'BOTDEFENSERQL') {
        $self->{socket} = IO::Handle->new_from_fd(F5::BotDefenseRqlPipe::get_pipe_fd(), 'r');
    }
    else {
        $self->{socket} = IO::Handle->new_from_fd(F5::LrnPipe::get_pipe_fd(), 'r');
    }

    return $self->SUPER::init(%params);
}

#----------------------------------------------------------------------

sub handle_read {
    my($self, %params) = @_;

    # Logger intentionally omitted to avoid ugly ERROR level message from LrnPipe
    my $read_result;
    if ($self->{pipe_open_opts}->{pipe_type} eq 'PBRQL') {
        $read_result = F5::PbRqlPipe::read(undef);
    }
    elsif ($self->{pipe_open_opts}->{pipe_type} eq 'BOTDEFENSERQL') {
        $read_result = F5::BotDefenseRqlPipe::read(undef);
    }
    else {
        $read_result = F5::LrnPipe::read(undef);
    }
    if ($read_result->{rv} < 0) {
        # no message
    }
    elsif ($read_result->{rv} == 0) {
        $self->{socket_close_handler}->($self, $read_result->{rv});
    }
    else {
        $self->{read_message_handler}->($self, $read_result->{msg});
    }

    return;
}

#----------------------------------------------------------------------

sub close {
    my($self) = @_;

    $self->{logger}->info("Closing pipe") if $self->{logger};
    # Logger intentionally omitted to avoid ugly ERROR level message from LrnPipe
    if ($self->{pipe_open_opts}->{pipe_type} eq 'PBRQL') {
        F5::PbRqlPipe::close(undef);
    }
    elsif ($self->{pipe_open_opts}->{pipe_type} eq 'BOTDEFENSERQL') {
        F5::BotDefenseRqlPipe::close(undef);
    }
    else {
        F5::LrnPipe::close(undef);
    }
}

#----------------------------------------------------------------------

1;
