package F5::IO::Selector;

use strict;
use warnings;

use IO::Select;
use Tie::RefHash;
use ZMQ::LibZMQ4;
use ZMQ::Constants qw(ZMQ_PUSH ZMQ_PULL ZMQ_FD ZMQ_EVENTS ZMQ_POLLIN);

sub new {
	my($class, %params) = @_;
	my $self = bless {}, $class;
	$self->init(%params);
	return $self;
}

#----------------------------------------------------------------------

sub init {
	my($self, %params) = @_;

        my $connections = $params{connections};
        $self->{sleep_on_poll_when_no_connections} = $params{sleep_on_poll_when_no_connections};

	$self->{select} = IO::Select->new;
	$self->{socket_map} = {};
	$self->{zmq_map} = {};
	tie %{$self->{socket_map}}, 'Tie::RefHash';
	tie %{$self->{zmq_map}}, 'Tie::RefHash';
	if ($connections && ref($connections) eq 'ARRAY' && scalar @$connections) {
		$self->add(@$connections);
	}
}

#----------------------------------------------------------------------

sub add {
	my($self, @connections) = @_;
	for my $connection (@connections) {
	   my $socket = $connection->{socket};
           if(exists $connection->{zmq_connection}) {
	      if(! exists $self->{zmq_map}->{$socket}) {		
		$self->{zmq_map}->{$socket} = { socket => $connection->{zmq_connection},
                                                events => ZMQ_POLLIN,
                                                callback => sub {},
                                                connection => $connection,
                                              };
 	      }
	   }
	   elsif(! exists $self->{socket_map}->{$socket}) {
		  $self->{select}->add($socket);

		  $self->{socket_map}->{$socket} = { fd => $socket->fileno,
                                                     events => ZMQ_POLLIN,
                                                     callback => sub {},
                                                     connection => $connection,
                                                   };
	   }
	}
}

#----------------------------------------------------------------------

sub remove {
	my($self, @connections) = @_;
	for my $connection (@connections) {
	   my $socket = $connection->{socket};
	   if(exists $connection->{zmq_connection}) {
              if(exists $self->{zmq_map}->{$socket}) {
	         delete $self->{zmq_map}->{$socket};
	      }
	   }
	   elsif(exists $self->{socket_map}->{$socket}) {
	         $self->{select}->remove($socket);

                 delete $self->{socket_map}->{$socket};
	   }
	}
}

#----------------------------------------------------------------------

sub can_read {
    my($self, $timeout) = @_;

    return %{ $self->{zmq_map} } ? $self->can_read_zmq($timeout) :
			           $self->can_read_socket($timeout);
}

#----------------------------------------------------------------------

# zmq_poll can handle both regular sockets and zmq connections
sub can_read_zmq {
    my($self, $timeout) = @_;
    $timeout ||= 0;

    # todo: sleep_on_poll_when_no_connections, like in can_read_socket

    my @sockets;
    push(@sockets, values % {$self->{zmq_map}});
    push(@sockets, values % {$self->{socket_map}});

    my @reader_indexes = zmq_poll(\@sockets, $timeout * 1_000);
    my @readers;
    for my $i (0..$#reader_indexes) {
        if ($reader_indexes[$i]) {
            push(@readers, $sockets[$i]->{connection});
        }
    }

    return @readers;
}

#----------------------------------------------------------------------

sub can_read_socket {
    my($self, $timeout) = @_;

    my $connections = $self->has_handles();

    if (!$connections && $self->{sleep_on_poll_when_no_connections}) {
        sleep($timeout);
    }
    else {
        my @readers = $self->{select}->can_read($timeout);
        if (@readers) {
            return map { $self->{socket_map}->{$_}->{connection} } @readers;
        }
    }

    return;
}

#----------------------------------------------------------------------

sub has_handles {
    my($self) = @_;
    return %{$self->{socket_map}} || %{$self->{zmq_map}};
}

#----------------------------------------------------------------------

sub handles {
    my($self) = @_;

    my @connections = map { $_->{connection} } values % {$self->{socket_map}};
    my @zmq_connections = map { $_->{connection} } values % {$self->{zmq_map}};

    push(@connections, @zmq_connections);

    return @connections;
}

#----------------------------------------------------------------------

1;
