package F5::IO::Storable;

use strict;
use warnings;

use base 'F5::IO::Handle';
use F5::ObjUtils qw(freeze);
use Storable qw(thaw);

my $DEFAULT_MSG_SIZE_LIMIT = 10_000_000; # 10 MB
my $counter = 0;
#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->SUPER::init(%params);

    $self->{storable_msg_size_limit} = defined($params{storable_msg_size_limit}) ? $params{storable_msg_size_limit} : $DEFAULT_MSG_SIZE_LIMIT;
}

#----------------------------------------------------------------------

sub read_message {
    my($self) = @_;
    my $raw_length = $self->read(4);
    if (! defined($raw_length)) {
        return;
    }
    my $message_length = unpack('N', $raw_length);

    if ($message_length < 0) {
        $self->{logger}->info("Negative length message received (received as $raw_length) -- skipping") if $self->{logger};
        $self->close();
        return;
    }
    elsif ($self->{storable_msg_size_limit} && $message_length > $self->{storable_msg_size_limit}) {
        # Message would exceed maximum expected size. This implies some sort of corruption, such as
        # a partial message received and some part of message contents incorrectly read as packed length.
        # Close this handle, which will force the sender to re-send.
        $self->{logger}->info("Message length exceeds limit (expected message length = $message_length, raw length = $raw_length) -- skipping") if $self->{logger};
        $self->close();
        return;
    }

    my $frozen_message = $self->read($message_length);
    if (! defined($frozen_message) || length($frozen_message) < $message_length) {
        if (defined $frozen_message) {
            $self->unshift_to_read_buffer($raw_length . $frozen_message);
        }
        else {
            $self->unshift_to_read_buffer($raw_length);
        }
        return;
    }
    my $message;
    eval {
        $message = thaw($frozen_message);
    };
    if ($@) {
        $self->{logger}->info("Message could not be thawed (expected message length = $message_length, raw length = $raw_length, message = $frozen_message)") if $self->{logger};
        $self->close();
        return;
    }

    return ($message, $message_length);
}

#----------------------------------------------------------------------

sub write_message {
    my($self, $message) = @_;

    my $frozen_message = freeze($message);
    my $frozen_message_length = length($frozen_message);
    if ($frozen_message_length < 0) {
        $self->{logger}->info("Attempted to send negative length message (message length = $frozen_message_length) -- skipping") if $self->{logger};
        return;
    }
    if ($frozen_message_length > $self->{storable_msg_size_limit}) {
        $self->{logger}->info("Attempted to send message larger than maximum allowed size (length = $frozen_message_length) -- skipping") if $self->{logger};
        return;
    }

    return $self->write(pack('N', $frozen_message_length) . $frozen_message);
}

#----------------------------------------------------------------------

1;
