package F5::IO::ZMQ;

use strict;
use warnings;

use ZMQ::LibZMQ4;
use ZMQ::Constants qw(ZMQ_PUSH ZMQ_PULL ZMQ_FD ZMQ_EVENTS ZMQ_POLLIN ZMQ_LINGER ZMQ_SNDHWM ZMQ_RCVHWM);
use Scalar::Util;

use constant DEFAULT_TIMEOUT => 60;
use constant DEFAULT_OUTBOUND_HIGH_WATER_MARK => 10000;
use constant DEFAULT_INBOUND_HIGH_WATER_MARK => 20000;

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    if (ref($class)) {
        %params = (%$class,
                   %params);
        $class = ref($class);
    }
    my $self = bless {}, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    %params = (
	       read_message_handler => sub {},
               before_close_handler => undef,
               socket_close_handler => sub {
                   my ($handle, $errno) = @_;
                   if ($handle->{before_close_handler}) {
                       $handle->{before_close_handler}->($handle, $errno);
                   }

                   $handle->{logger}->debug("Socket close received - normal shutdown") if $handle->{logger};
                   $handle->{selector}->remove($handle) if $handle->{selector};
               },
	       timeout => DEFAULT_TIMEOUT,
	       high_water_mark => $params{server} ? DEFAULT_INBOUND_HIGH_WATER_MARK : DEFAULT_OUTBOUND_HIGH_WATER_MARK,
               %params);

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }
    Scalar::Util::weaken($self->{selector});

    my $context = zmq_init();
    if ($self->{server}) {
        $self->{zmq_connection} = zmq_socket($context, ZMQ_PULL);
	if(defined $self->{high_water_mark}) {
           zmq_setsockopt($self->{zmq_connection}, ZMQ_RCVHWM, $self->{high_water_mark});
       }
        zmq_bind($self->{zmq_connection}, 'tcp://*:' . $self->{port});
    }
    else {
       $self->{zmq_connection} = zmq_socket($context, ZMQ_PUSH);
       if(defined $self->{timeout}) {
           zmq_setsockopt($self->{zmq_connection}, ZMQ_LINGER, $self->{timeout} * 1_000);
       }
       if(defined $self->{high_water_mark}) {
	   zmq_setsockopt($self->{zmq_connection}, ZMQ_SNDHWM, $self->{high_water_mark});
       }
       zmq_connect($self->{zmq_connection}, 'tcp://localhost:' . $self->{port});
    }
    $self->{socket} = zmq_getsockopt( $self->{zmq_connection}, ZMQ_FD );

    if ($self->{selector}) {
        $self->{selector}->add($self);
    }

    return $self;
}

#----------------------------------------------------------------------

sub handle_read {
    my($self, %params) = @_;
    my($message, $message_size) = $self->read_message(%params);
    if ($message) {
        $self->{read_message_handler}->($self, $message, $message_size);
    }
}

#----------------------------------------------------------------------

sub read {
    my($self) = @_;
    my $msg = zmq_msg_init();
    my $bytes = zmq_msg_recv($msg, $self->{zmq_connection});
    my $buf = zmq_msg_data($msg);
    return $buf;
}

#----------------------------------------------------------------------

sub read_message {
    die "Abstract call\n";
}

#----------------------------------------------------------------------

sub write_message {
    die "Abstract call\n";
}

#----------------------------------------------------------------------

sub write {
    my($self, $buffer) = @_;
    my $msg = zmq_msg_init_data($buffer); 
    my $bytes = zmq_msg_send($msg, $self->{zmq_connection});
    return $bytes;
}

#----------------------------------------------------------------------

sub flush {
    my($self, %params) = @_;

    return 1;
}

#----------------------------------------------------------------------

sub close {
    my($self, %params) = @_;

    $self->{socket_close_handler}->($self, $params{file_handle_error});

    zmq_close($self->{zmq_connection});
}

#----------------------------------------------------------------------

1;
