package F5::Iface;

use strict;
use warnings;


use F5::Cfg;
use F5::Cluster::Utils;
use F5::BigipUtils;

# the following should stay consistent with the c "iface" library
use constant IFACE_ERROR_IFCONFIG    => -1; # ifconfig system call failed
use constant IFACE_ERROR_TS_ROOT     => -2; # ts root not found in the /ts.cfg file
use constant IFACE_ERROR_IMAGE_TYPE  => -3; # image type not found in the /image.type file 
use constant IFACE_ERROR_IFACE_ITEM  => -4; # interface item not found in the /etc/ts/common/image.cfg file

use constant COMMON_PATH_IMAGE_CFG_FILE => 'image.cfg';

use constant APP_ID_HTTP             => 0;
use constant APP_ID_HTTPS            => 1;

my $TS_ROOT = '/ts';

my $IMAGE_TYPE = 4000;
my $DEFAULT_INTERFACE = "eth0";
#--------------------------------------------------------------------------------
# function: iface_set_ip
# purposes: wraps the ifconfig ommand
# arguments: index in the interfaces array, the alias number (use -1 to config on the interface iself), ip, mask, bcast
# return: 0 - added successfuly, or an appropriate error from IFACE_ERRORS (constants.pm)

sub iface_set_ip
{

   # check for the ts_root
   return IFACE_ERROR_TS_ROOT if(not defined $TS_ROOT);
   # check for the image type
   return IFACE_ERROR_IMAGE_TYPE if(not defined $IMAGE_TYPE);

   my ($ind, $alias, $ip, $mask, $bcast) = @_;
   # get the image type from the image_type file
   my $interface = get_interface_by_index($ind);
   return IFACE_ERROR_IFACE_ITEM if(not defined $interface);
   
   $interface.= ":$alias" if($alias >= 0);
   my $cmd = "/sbin/ifconfig $interface $ip";
   $cmd.= " netmask $mask" if defined $mask;
   $cmd.= " broadcast $bcast" if defined $bcast;

   my $ret = system($cmd);
   return ($ret == 0) ? 0 : IFACE_ERROR_IFCONFIG;
}
#--------------------------------------------------------------------------------
# function: iface_remove_ip
# purposes: wraps the ifconfig command
# arguments: index in the interfaces array, the alias number (use -1 to config on the interface iself)
# return: 0 - removed successfuly, or an appropriate error from IFACE_ERRORS (constants.pm)

sub iface_remove_ip
{
   # check for the ts_root
   return IFACE_ERROR_TS_ROOT if(not defined $TS_ROOT);
   # check for the image type
   return IFACE_ERROR_IMAGE_TYPE if(not defined $IMAGE_TYPE);
   
   my ($ind, $alias) = @_;
   # get the image type from the image_type file
   my $interface = get_interface_by_index($ind);
   return IFACE_ERROR_IFACE_ITEM if(not defined $interface);

   my $groups = iface_get_iface_sets();
   my $removed_addr;
   my (@removed_interfaces, @removed_routes);
   # first search for the data of the removed interface
   for(my $i = 0; $i < scalar @{$groups->{$interface}->{addresses}}; $i++)
   {
      my $addr =  $groups->{$interface}->{addresses}->[$i];
      if((($addr->{alias} < 0) && (not defined $alias)) ||
         ($addr->{alias} == $alias))
      {
         $removed_addr = $addr;
         delete $groups->{$interface}->{addresses}->[$i];
         last;
      }
      
   }
   if(! $removed_addr->{flags}->{secondary})
   {
      my $removed_ip = $removed_addr->{addr};
      my $removed_mask = convert_mask($removed_addr->{mask});
      my $removed_network = inet_ntoa(inet_aton($removed_ip) & inet_aton($removed_mask));
      #print "removed network : $removed_network\n";
      foreach my $addr (@{$groups->{$interface}->{addresses}}) 
      {
         # check if this is a secondary ip with the same ip and mask of the reomoved address
         next if ($removed_addr == $addr) || (! $addr->{addr}) || (! $addr->{mask});
         my $candidate_ip = $addr->{addr};
         my $candidate_mask = convert_mask($addr->{mask});
         my $candidate_network = inet_ntoa(inet_aton($candidate_ip) & inet_aton($candidate_mask));
         #print "candidate network : $candidate_network\n";
         if($removed_network eq $candidate_network)
         {
           #print "will be removed\n";
            push @removed_interfaces, $addr;
         }
         # mark routes for recovery - only if ips are removed

         if(scalar @removed_interfaces)
         {
            my $routes = iface_get_current_routes($interface);
            foreach my $route (@$routes)
            {
               my $route_gateway_network = inet_ntoa(inet_aton($route->{gw}) &  inet_aton($removed_mask));
               if($route_gateway_network eq $removed_network)
               {
                  push @removed_routes, $route;
               }
            }
         }
      }
   }
   
   
   $interface.= ":$alias" if($alias > 0);
   my $cmd = "/sbin/ifconfig $interface down";

   my $ret = system($cmd);
   if($ret == 0)
   {
      foreach my $addr(@removed_interfaces)
      {
         #print "set_ip ".join(',',$ind, $alias, $addr->{addr}, convert_mask($addr->{mask}), $addr->{bcast})."\n" ;
         iface_set_ip($ind, $addr->{alias}, $addr->{addr}, convert_mask($addr->{mask}), $addr->{bcast});
      }
      foreach my $route(@removed_routes)
      {
         recover_route($route);
      }
      return 0;
   }
   else
   {
      return undef;
   }
}
#--------------------------------------------------------------------------------
# function: iface_config_up
# purposes: wraps the ifconfig command
# arguments: index in the interfaces array
# return: 0 - configured successfuly, or an appropriate error from IFACE_ERRORS (constants.pm)

sub iface_config_up
{
   # check for the ts_root
   return IFACE_ERROR_TS_ROOT if(not defined $TS_ROOT);
   # check for the image type
   return IFACE_ERROR_IMAGE_TYPE if(not defined $IMAGE_TYPE);
   
   my ($ind) = @_;
   my $image_type;
   # get the image type from the image_type file
   my $interface = get_interface_by_index($ind);
   return IFACE_ERROR_IFACE_ITEM if(not defined $interface);

   my $cmd = "/sbin/ifconfig $interface up";
   my $ret = system($cmd);

   return ($ret == 0) ? 0 : IFACE_ERROR_IFCONFIG;
}
#--------------------------------------------------------------------------------
# function: iface_config_down
# purposes: wraps the ifconfig command
# arguments: index in the interfaces array
# return: 0 - removed successfuly, or an appropriate error from IFACE_ERRORS (constants.pm)

sub iface_config_down
{  
   # check for the ts_root
   return IFACE_ERROR_TS_ROOT if(not defined $TS_ROOT);
   # check for the image type
   return IFACE_ERROR_IMAGE_TYPE if(not defined $IMAGE_TYPE);

   my ($ind) = @_;
   
   # get the image type from the image_type file
   my $interface = get_interface_by_index($ind);
   return IFACE_ERROR_IFACE_ITEM if(not defined $interface);
   
   my $cmd = "/sbin/ifconfig $interface down";
   my $ret = system($cmd);

   return ($ret == 0) ? 0 : IFACE_ERROR_IFCONFIG;
}
#--------------------------------------------------------------------------------
# function: iface_get_unit_id
# purposes: return mac address of current machine as stored (by tsconfig.pl) in DCC.SITE_UNITS
# arguments: -
# return: mac addr string on success, undef - error

sub iface_get_unit_id {
    my (%params) = @_;
    my $mac;
    my $dbh = $params{dbh};

    if ($dbh)
    {
        $mac = $dbh->selectrow_array('SELECT unit_id from DCC.SITE_UNITS');
    }

    return ($mac) ? $mac : get_mac_addr($DEFAULT_INTERFACE) ;
}
#--------------------------------------------------------------------------------
# function: iface_get_unit_id_from_ifconfig
# purposes: wraps the ifconfig command - needed for tsconfig.pl
# arguments: -
# return: mac addr string on success, undef - error

sub iface_get_unit_id_from_ifconfig
{

   # check for the ts_root
   return if(not defined $TS_ROOT);
   # check for the image type
   return if(not defined $IMAGE_TYPE);
                                            
   my $mac_addr;
   for my $index (0,1) {
       # get the image type from the image_type file
       my $interface = get_interface_by_index($index);
       next if (not defined $interface);
       $mac_addr = get_mac_addr($interface);
       last if $mac_addr;
   }
   if (!$mac_addr) {
       $mac_addr = get_mac_addr('');
   }
   return $mac_addr;
}
#--------------------------------------------------------------------------------
#gets
#   intrfc - the interface name
#returns
#   mac addr of received interface
sub get_mac_addr
{
    my $intrfc = shift;

    # STDERR redirected to /dev/null to ignore harmless warning:
    # "info: Warning: cannot open /proc/net/dev (No such file or directory). Limited output."
    if(not open(IFCNG, "/sbin/ifconfig $intrfc 2>/dev/null |"))
    {
	print "get_mac_addr - Can\"t run /sbin/ifconfig: $!\n";
	return undef;
    }
    
    while(<IFCNG>)
    {
	if(/HWaddr\s+(\S+)/i)
	{
	    close(IFCNG);
	    return $1;
	}
	if(/ether\s+(\S+)/i)
	{
	    close(IFCNG);
	    return $1;
	}
    }
    
    close(IFCNG);
	print "get_mac_addr - no HWaddr for interface $intrfc was found\n";
    return undef;
}
#--------------------------------------------------------------------------------
# function: iface_get_ip_addr
# purposes: wraps the ifconfig command
# arguments: index in the interfaces array, alias index
# return: ip addr string on success, undef - error

sub iface_get_ip_addr
{

   # check for the ts_root
   return IFACE_ERROR_TS_ROOT if(not defined $TS_ROOT);
   # check for the image type
   return IFACE_ERROR_IMAGE_TYPE if(not defined $IMAGE_TYPE);

   my ($ind, $alias) = @_;
   
   my $interfaces = iface_get_ifaces();
   foreach my $interface (@$interfaces)
   {
      if(($interface->{ind} == $ind) &&
         (defined $interface->{ind}) &&
         (($interface->{alias} == $alias) || (not defined $alias)))
      {
         return $interface->{inet_addr};
      }
   }
   return;
}
#--------------------------------------------------------------------------------
# function: iface_get_iface_sets
# purposes: wraps the ip addr command and parses the result
# arguments: none
# return: hash reference of interfaces. Hash keys are the device names, and values are hash reference with the following format:
# synopsys:
#foreach my $device (keys %$groups)
#{
#        print "device $device (mac address ".$groups->{$device}->{mac}.")\n";
#        foreach my $addr (@{$groups->{$device}->{addresses}})
#        {
#                print " ip address ".$addr->{addr}." mask ".$addr->{mask}." bcast ".$addr->{bcast}." alias ".$addr->{alias}.(($addr->{flags}->{secondary}) ? " secondary": '' )."\n";
#        }
#}

sub iface_get_iface_sets
{
   my $cmd = "ip addr";
   if(not open(IP_ADDR, "$cmd|"))
   {
       return undef;
   }
   my ($result, $iface, $mac);
   my $i = 0;
   while(<IP_ADDR>)
   {
      chomp;
      if(/\d+:\s*(.*):\s*(<.*>)/)
      {
         # a new interface
         $i = 0;
         $iface = $1;
         my  $iface_flags = $2;
         $result->{$iface}->{flags} = $iface_flags;
      }
      elsif(/\s+link\/\w+\s+((?:[a-f0-9]{2}:){5}[a-f0-9]{2})/i)
      {
         $mac = $1;
         $result->{$iface}->{mac} = $mac;
      }
      elsif(/\s+inet\s+([\d\.]+)\/(\d+)\s+brd\s+([\d\.]+)\s+scope\s+([\s\w]+)\s+(.*)/)
      {
         my ($addr, $mask, $bcast, $flags, $full_iface_name) = ($1, $2, $3, $4, $5);
         $result->{$iface}->{addresses}->[$i]->{addr} = $addr;
         $result->{$iface}->{addresses}->[$i]->{mask} = $mask;
         $result->{$iface}->{addresses}->[$i]->{bcast} = $bcast;
         # flags
         foreach my $flag(split(/\s/, $flags))
         {
            $result->{$iface}->{addresses}->[$i]->{flags}->{$flag} = 1;
         }
         # alias parsing
         my (undef, $alias) = split(/:/, $full_iface_name);
         
         $result->{$iface}->{addresses}->[$i]->{alias} = (defined $alias) ? $alias : -1;
         $i++;

      }
      else
      {
         #print;
         #print "\n";
      }
      
   }
   close IP_ADDR;
   return $result;
}
#--------------------------------------------------------------------------------
# function: iface_get_ifaces
# purposes: wraps the ifconfig command and parsing it
# arguments: none
# return: list of interfaces. each one is a hash reference with the following members: inet_addr, mac, type, interface, bcast, alias, undef - error
# synopsys:
#$interfaces = iface_get_ifaces();
#foreach $interface(@$interfaces)
#{
#         foreach my $interface(@$interfaces)
#        {
#                foreach my $key (keys %$interface)
#                {
#                        print "$key:".$interface->{$key}."\n";
#                }
#        }
#}


sub iface_get_ifaces
{
   # get the name of the interfaces from the image cfg file
   my $iface0_name = F5::Cfg::cfg_get_config_item("/etc/ts/common/".COMMON_PATH_IMAGE_CFG_FILE, $IMAGE_TYPE, "iface0");
   my $iface1_name = F5::Cfg::cfg_get_config_item("/etc/ts/common/".COMMON_PATH_IMAGE_CFG_FILE, $IMAGE_TYPE, "iface1");
   if(not open(IFCNG, "/sbin/ifconfig|"))
   {
       return undef;
   }
   my $result;
   my $i = 0;
   while(<IFCNG>)
   {
      chomp;
      if(! $_)
      {
         # a newline encountered
         $i++;
         next;
      }
      elsif(/^([\w\.\:]+)\s+Link\s*encap\:(\w+)\s*(?:[\w]+ )*([\w\:]*)/)
      {
         my ($interface, $alias_index) = split(/\:/, $1);
         $result->[$i]->{type} = $2;
         $result->[$i]->{mac} = $3;
         $result->[$i]->{interface} = $interface;
         $result->[$i]->{alias} = (defined $alias_index) ? $alias_index : -1;
         if($interface eq $iface0_name)
         {
            $result->[$i]->{ind} = 0;
         }
         if($interface eq $iface1_name)
         {
            $result->[$i]->{ind} = 1;
         }
      }
      else
      {
         if(/inet addr:\s*([\d\.]+)/)
         {
            $result->[$i]->{inet_addr} = $1;
         }
         if(/Bcast:\s*([\d\.]+)+/)
         {
            $result->[$i]->{bcast} = $1;
         }
         if(/Mask:\s*([\d\.]+)/)
         {
            $result->[$i]->{mask} = $1;
         }
      }
   }
   close(IFCNG);
   return $result;
}
#--------------------------------------------------------------------------------
# function: iface_get_iface
# purposes: internal function, used to get the configuration on all the wrappers
# arguments: index
# return: the interface name, or undef on error
sub iface_get_iface
{
   # check for the ts_root
   return undef if(not defined $TS_ROOT);
   # check for the image type
   return undef if(not defined $IMAGE_TYPE);

   my ($iface_ind) = @_;
   
   my $interface_name = get_interface_by_index($iface_ind);

   return undef if(not defined $interface_name);
   
   return $interface_name;
}
#--------------------------------------------------------------------------------
# function: iface_get_current_routes
# purposes: parse route -n 
# arguments: interface name (optional)
# return: array reference of route entries (hash references) / undef (error), 

sub iface_get_current_routes
{
   my ($required_interface) = @_;
   my $cmd = "route -n";
   my ($res, $count);
   $count = 0;

   if(not open(ROUTE, "$cmd|"))
   {
       return undef;
   }
   # skip first line
   <ROUTE>;
   foreach(<ROUTE>)
   {
      chomp;
      my ($network,   $gw, $mask, undef, undef, undef, undef, $interface) = split(/\s+/);
      if(defined $network && defined $gw && defined $mask && defined $interface)
      {
         if($interface eq $required_interface || (not defined $required_interface))
         {
            $res->[$count]->{network} = $network;
            $res->[$count]->{gw} = $gw;
            $res->[$count]->{mask} = $mask;
            $res->[$count]->{interface} = $interface;
            $count++;
         }
      }

   }
   close (ROUTE);
   return $res;


}
#--------------------------------------------------------------------------------
# function: get_interface_by_index
# purposes: internal function, used to get the configuration on all the wrappers
# arguments: index
# return: the interface name, or undef on error
sub get_interface_by_index
{
   my ($ind) = @_;
   if ($ind == 0 && !F5::BigipUtils::is_docker()) {
       return F5::Cluster::Utils::get_mcp_val("provision.managementeth");
   }
   else {
       return F5::Cfg::cfg_get_config_item("/etc/ts/common/".COMMON_PATH_IMAGE_CFG_FILE, $IMAGE_TYPE, "iface$ind");
   }
}
#-------------------------------------------------------------------------------
# internal function
# gets a mask as returned from the ip command, and turns it to ip
# I am sure that there's a better way to do this
# Time was an issue..
sub convert_mask
{
        my ($mask) = @_;
        my $string = 1 x $mask . 0 x (32 - $mask);
        my $first = sprintf("%d", oct('0b'.substr($string, 0, 8)));
        my $second = sprintf("%d", oct('0b'.substr($string, 8, 8)));
        my $third = sprintf("%d", oct('0b'.substr($string, 16, 8)));
        my $fourth = sprintf("%d", oct('0b'.substr($string, 24, 8)));
        return join('.', $first, $second, $third, $fourth);
}
#--------------------------------------------------------------------
# internal function
sub recover_route
{
   my ($route) = @_;
   my $cmd = sprintf("route add -net %s gw %s netmask %s dev %s",
                     $route->{network},
                     $route->{gw},
                     $route->{mask},
                     $route->{interface});
   return system($cmd);
}
1;
