package F5::ImUtils;


use strict;
use warnings;

use FindBin;
use File::Path;
use File::Basename;
use File::Copy;
use File::Find;
use POSIX qw(strftime);
use File::Temp qw/tempfile tempdir/;
use F5::Utils::File;
use F5::BigipUtils;

my $IMAGE_ROOT = "image_root";
sub create {
    my (%arguments) = @_;
    die $arguments{signature_filename}." doesn't exist" 
       unless -f $arguments{signature_filename};

    my $enc_filename = basename($arguments{signature_filename});
    my $im_filename = $arguments{im_filename};
    my $force_timestamp = $arguments{force_timestamp};

    my $image_root = (defined $arguments{image_root}) ? $arguments{image_root} : $IMAGE_ROOT;

    init_image_root(%arguments, 
                    image_root => $image_root);
    add_pre_upgrade(%arguments,
                    image_root => $image_root);
    add_post_upgrade(%arguments,
                     image_root => $image_root, 
                     enc_filename => $enc_filename);

    copy($arguments{signature_filename}, "$image_root/RPMS/packages");
    add_metadata(%arguments,
                 image_root => $image_root);
 
    if ($force_timestamp) {
        find({
              wanted => sub {
                  utime($force_timestamp, $force_timestamp, $File::Find::name);
              },
              no_chdir => 1,
             },
             $image_root);
    }

    {
        local $ENV{LD_PRELOAD};
        local $ENV{FAKETIME};

        if ($force_timestamp && $ENV{USE_FAKETIME_LIBRARY}) {
            $ENV{LD_PRELOAD} = $ENV{USE_FAKETIME_LIBRARY};
            $ENV{FAKETIME} = strftime("%Y-%m-%d %H:%M:%S", localtime($force_timestamp));
        }

        system "mkisofs --quiet -r -o $im_filename $image_root";
    }

    return;
}

###########################################################################

sub mount {
    my (%arguments) = @_;
    my $filename = $arguments{filename};
    my $metadata_file = $arguments{metadata_file} || 'metadata.xml';
    my $tempdir_for_mount = tempdir();
    my @cmd;
    if (F5::BigipUtils::is_docker()) {
        @cmd = ('bsdtar', '-xf', $filename, '-C', $tempdir_for_mount);
    }
    else {
        @cmd = ('/bin/mount', '-o', 'loop', $filename, $tempdir_for_mount);
    }

    system (@cmd) == 0 or die( qq/Failed to mount: '$cmd[0]': $?/ );

    my $mount_success = 0;
    for (1..10) {
        if (! -e "$tempdir_for_mount/$metadata_file") {
            sleep 1;
        }
        else {
            $mount_success = 1;
            last;
        }
    }

    if (!$mount_success) {
        die( qq/Failed to mount: '$cmd[0]'/);
    }

    return {tempdir => $tempdir_for_mount,
            metadata => F5::Utils::File::file_slurp("$tempdir_for_mount/$metadata_file")};
}

###########################################################################

sub mount_and_unpack {
    my (%arguments) = @_;

    my $filename = $arguments{filename};
    my $mount_rv = mount(%arguments);
    my $tempdir_for_mount = $mount_rv->{tempdir};
    my $metadata = $mount_rv->{metadata};

    opendir (my $enc_file_dir_fh, "$tempdir_for_mount/RPMS/packages");
    my $tempdir_for_save = tempdir();

    my @encfile = grep { /^[^\.]/ } readdir $enc_file_dir_fh;
    my $encfile = $encfile[0];
    copy("$tempdir_for_mount/RPMS/packages/$encfile", "$tempdir_for_save/");

    closedir ($enc_file_dir_fh);
    unmount(dir => $tempdir_for_mount);

    $filename = "$tempdir_for_save/$encfile";
    return {filename => $filename,
            metadata => $metadata};
}

###########################################################################

sub unmount {
    my (%arguments) = @_;
    my @cmd = ((F5::BigipUtils::is_docker() ? ('/bin/rm', '-rf') : '/bin/umount'), $arguments{dir});
    system (@cmd) == 0 or die( qq/Failed to umount: '$cmd[0]': $?/ );
    return;
}


############################################################################

sub init_image_root {
    my (%arguments) = @_;
    my $image_root = $arguments{image_root};
    rmtree ($image_root);
    mkpath ("$image_root/bin");
    mkpath ("$image_root/RPMS/packages");
}

sub add_pre_upgrade {
    my (%arguments) = @_;
    my $image_root = $arguments{image_root};
    open my $pre, ">", "$image_root/bin/pre-upgrade";
    print $pre "#!/bin/bash";
    close $pre;
    chmod 0755, "$image_root/bin/pre-upgrade";
}

sub add_post_upgrade {
    my (%arguments) = @_;
    my $image_root = $arguments{image_root};
    my $enc_filename = $arguments{enc_filename};
    my $post_upgrade_script = $arguments{post_upgrade_script} || 'im_post_upgrade.pl';

    open my $post, ">", "$image_root/bin/post-upgrade";
    my $program_text = <<EndProgram;
if [ -e /usr/share/ts/bin/$post_upgrade_script ]; then
  /usr/bin/perl /usr/share/ts/bin/$post_upgrade_script "RPMS/packages/$enc_filename"
else
  /usr/bin/perl /ts/tools/$post_upgrade_script "RPMS/packages/$enc_filename"
fi

EndProgram

    print $post $program_text;
    close $post;
    chmod 0755, "$image_root/bin/post-upgrade";
}

sub add_metadata {
    my (%arguments) = @_;
    my $image_root = $arguments{image_root};
    my $header = $arguments{header} || 'asmAttackSignatureImage';

    my $metadata = qq#<$header version="1.0">
<version>$arguments{file_version}</version>
<description>$arguments{description}</description>
<supportedProducts>
<product><name>BIG-IP</name>
<version>$arguments{bigip_version}</version>
</product>
</supportedProducts>
</$header>#;

    open my $md_out, ">", "$image_root/metadata.xml";
    print $md_out $metadata;
    close $md_out;
}


1;

