package F5::ImportExportPolicy;

use strict;
use warnings;

use F5::LockFactory;
use F5::Utils::Rest;
use F5::ASMConfig::Entity::Policy;
use F5::YAML::Schema::Utils;
use F5::BigipUtils;

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;
  $self->{initial_params} = \%params;
}

#----------------------------------------------------------------------

sub import_policy {
  my($self, %params) = @_;
  my $bypass_lock = defined $params{is_within_ucs_upgrade} ? $params{is_within_ucs_upgrade} : 0;
  my $first_part = $params{content};
  
  if (!$first_part) {
      # Auto-detect whether the input file is XML
      open(my $inf, '<', $params{filename});
      read($inf, $first_part, 100);
      close $inf;
  }

  if ($first_part =~ /^\s*(<!--.*?-->\s*)*(<\?xml |<policy[ >])/s) {
    $params{file_format} = 'xml';
  }
  else {
    my $rv;
    if ( $first_part =~ /^\s*\{\s*\"/s) {
        $rv = F5::YAML::Schema::Utils::parse_json(filename => $params{filename},
                                                  content  => $params{content},);
        if (!$rv->{completed_successfully}) {
            die {error_message => "Failed to parse JSON Policy\n$rv->{error_message}"};
        }
    } elsif ($params{filename} =~ /\.yaml|\.yml/ || $first_part =~ /^\s*(\#.*?\n)*(---\s*)?[\w-]+: /s) {
        if (! $ENV{F5WAF_ALLOW_YAML_POLICY}) {
            # Icicle will not support yaml policies, currently only json policies are fully supported
            die {error_message => "YAML Policies are not supported."};
        }

        local $YAML::XS::Boolean = "JSON::PP";
        $rv = F5::YAML::Schema::Utils::parse_yaml(filename => $params{filename},
                                                  content  => $params{content},);
        if (!$rv->{completed_successfully}) {
            die {error_message => "Fail parse YAML Policy: $rv->{error_message}."};
        }
    }

    if ($rv && $rv->{completed_successfully}) {
        $params{file_format} = 'yaml';
        $params{file_data} = $rv->{data};
    }
  }

  my $import_obj = $self->get_format_obj(%params , direction => 'import');
  my $dbh = $import_obj->{dbh};

  my $locker;
  if (! $bypass_lock) {
      $locker = F5::LockFactory::get_locker(lock_filename => $F5::DbUtils::LOCK_FILE,
                                            logger => $self->{initial_params}->{logger},
                                           );
      $locker->lock();
  }

  my $rv;
  eval {
      $rv = $import_obj->import_policy(%params);
      my $populate_rest_uuid = defined $params{populate_rest_uuid} ? $params{populate_rest_uuid} : F5::BigipUtils::is_feature_enabled('populate_rest_uuids');
      if ($rv->{policy_id} && $rv->{completed_successfully} && ! $params{get_policy_attributes_only} && $populate_rest_uuid) {
          F5::Utils::Rest::populate_uuids(dbh => $dbh,
                                          policy_id => $rv->{policy_id},
                                          skip_mcp => $params{skip_mcp},
                                         );
      }
  };
  if ($@) {
      if (! $bypass_lock) {
          $locker->unlock();
      }
      die $@;
  }

  if ($params{is_relayed_import}) {
      my $logger = $self->{initial_params}->{logger};
      $logger->debug("Turning off policy builder on relayed policy import") if $logger;
      $dbh->do(<<EndQuery, undef, $rv->{policy_id});
UPDATE PLC.PL_POLICY_BUILDER_CONFIG 
  SET flg_enabled = 0 
  WHERE policy_id = ?
EndQuery
  }

  if (! $bypass_lock) {
      $locker->unlock();
  }
  return $rv;
}

#----------------------------------------------------------------------

sub export_policy {
  my($self, %params) = @_;

  if ($params{policy_id} && ! $params{policy}) {
      $params{policy} = F5::ASMConfig::Entity::Policy->new(dbh => $self->{initial_params}->{dbh}, get_criteria => {policy_id => $params{policy_id}});
  }

  if (! ($params{file_format} && $params{file_format} eq 'yaml')) {
      $self->copy_global_scanner_config_to_policy(%params);
  }

  my $export_obj = $self->get_format_obj(%params , direction => 'export');
  return $export_obj->export_policy(%params);
}

#----------------------------------------------------------------------

sub get_available_policy_name {
    my($self, %params) = @_;
    require F5::ImportExportPolicy::Binary;
    return F5::ImportExportPolicy::Binary->new(%{$self->{initial_params}})->get_available_policy_name(%params);
}

#----------------------------------------------------------------------

sub get_format_obj {
  my($self, %params) = @_;

 if ($params{file_format} && $params{file_format} eq 'xml' && $params{direction} eq 'export') {
   require F5::ExportPolicy::XML;
   return F5::ExportPolicy::XML->new(%{$self->{initial_params}});
 }
 elsif ( ($params{file_format} && $params{file_format} eq 'old_xml' && $params{direction} eq 'export') ||
         ($params{file_format} && $params{file_format} eq 'xml'     && $params{direction} eq 'import') ) {
   require F5::ImportExportPolicy::XML;
   return F5::ImportExportPolicy::XML->new(%{$self->{initial_params}});
 }
 elsif ($params{file_format} && ($params{file_format} eq 'yaml' || $params{file_format} eq 'json'))  {
   require F5::ImportExportPolicy::YAML;
   return F5::ImportExportPolicy::YAML->new(%{$self->{initial_params}});
 }
 else {
    require F5::ImportExportPolicy::Binary;
    return F5::ImportExportPolicy::Binary->new(%{$self->{initial_params}});
  }
}

#----------------------------------------------------------------------

sub copy_global_scanner_config_to_policy {
    my($self, %params) = @_;

    my $policy_id = $params{policy}->{policy_id};
    my $dbh = $self->{initial_params}->{dbh};

    my $global_scanner_config = $dbh->selectrow_hashref(<<EndQuery, undef, $policy_id);
SELECT global.api_key, global.scanner_site_name
  FROM PLC.GLOBAL_SCANNER_CONFIG global,
       PLC.PL_SCANNER_CONFIG policy
 WHERE global.scanner_type = policy.scanner_type
   AND policy.policy_id = ?
EndQuery

    if ($global_scanner_config && $global_scanner_config->{api_key}) {
        # Copy values to PL_SCANNER_CONFIG -- these are overriden in ScannerConfig entity retrieve
        # with values from global, but are needed for policy export

        F5::DbUtils::update_row(dbh => $dbh,
                                table_name => 'PLC.PL_SCANNER_CONFIG',
                                rec => $global_scanner_config,
                                where_fields => { policy_id => $policy_id });
    }
}

#----------------------------------------------------------------------

1;
