package F5::ImportExportPolicy::Base;

use strict;
use warnings;
use Config::IniFiles;
use F5::Deploy;
use F5::Logger;
use File::Basename;
use MIME::Base64;
use F5::DbUtils;
use F5::BigipUtils;
use F5::ASMConfig::Entity::Policy;
use F5::PolicyValidator;

sub init
{
    my($self, %params) = @_;

    my $cfg_file = F5::Deploy::map_path("/etc/ts/tools/import_export_policy.cfg");
    if (-e $cfg_file) {
        $self->{cfg} = new Config::IniFiles(-file => $cfg_file);
    }
    else {
        # Build-time load -- no need for config, so use empty hashref instead
        $self->{cfg} = {};
    }

    unless ($self->{logger}) {
      $self->{logger} = F5::Logger->get_logger(logger_name => 'import_export_policy');
    }
    $self->{seen_warnings} = {};
    $self->{do_validation} = $params{do_validation};
}

#----------------------------------------------------------------------

sub fatal_error {
  my($self, %params) = @_;

  my @msgs;
  if ($params{msgs}) {
    for my $param_msg (@{$params{msgs}}) {
      if (ref($param_msg)) {
        push(@msgs, $param_msg);
      }
      else {
        push(@msgs,
             { user_msg => $param_msg,
               log_msg => $param_msg });
      }
    }
  }
  else {
    push(@msgs,
         { user_msg => $params{user_msg} || $params{msg},
           log_msg => $params{log_msg} || $params{msg} });
  }

  for my $msg (@msgs) {
    push(@{$self->{msgs}}, { 
                            msg_level => 'fatal',
                            msg_text => $msg->{user_msg},
                           }
        );
    $self->{error_msg} = $msg->{user_msg};

    $self->{logger}->fatal($msg->{log_msg}) if $self->{logger};
  }

  $self->{completed_successfully} = 0;

  return $self->build_return_hash;
}

#----------------------------------------------------------------------

sub add_warning {
  my($self, %params) = @_;

  my $user_msg = $params{user_msg} || $params{msg};
  my $log_msg = $params{log_msg} || $params{msg};

  my $log_level = (defined $params{log_level}) ? $params{log_level} : "info";

  #Only add unique warnings.
  return if $self->{seen_warnings}->{$user_msg}++;

  push(@{$self->{msgs}}, { 
                          msg_level => 'warning',
                          msg_text => $user_msg,
                          }
      );

  $self->{logger}->$log_level($log_msg) if $self->{logger};
}


#----------------------------------------------------------------------

sub add_info {
  my($self, %params) = @_;

  my $user_msg = $params{user_msg} || $params{msg};
  my $log_msg = $params{log_msg} || $params{msg};

  push(@{$self->{msgs}}, { 
                          msg_level => 'info',
                          msg_text => $user_msg,
                          }
      );

  $self->{logger}->info($log_msg) if $self->{logger};
}

#----------------------------------------------------------------------

sub completed_successfully {
  my($self, %params) = @_;

  my $user_msg = $params{user_msg} || $params{msg};
  my $log_msg = $params{log_msg} || $params{msg};

  push(@{$self->{msgs}}, { 
                          msg_level => 'success',
                          msg_text => $user_msg,
                          }
      );
  $self->{success_msg} = $user_msg;

  $self->{logger}->info($log_msg) if $self->{logger};
  $self->{completed_successfully} = 1;
  if ($self->{do_validation}) {
       my $validator = F5::PolicyValidator->new(dbh => $self->{dbh},
                                                logger => $self->{initial_params}->{logger} );
       my $validation_report = $validator->validate(policy => $self->{policy},
                                                     action => 'import_policy',);
       if ($validation_report) {
           push(@{$self->{msgs}}, {
                          msg_level => 'warning',
                          msg_text => $validation_report,
                         }
           );
       }
  }

  return $self->build_return_hash;
}

#----------------------------------------------------------------------

sub build_return_hash {
  my($self) = @_;

  if ($self->{overlay}) {
      push (@{$self->{msgs}}, @{$self->{overlay}->get_msgs()})
  }
  my $return_hash = {};
  my @copied_keys = qw(completed_successfully
                       msgs
                       policy_id
                       policy_name
                       displaced_policy_id
                       policy
                       policy_version_time
                       policy_version_device_name
                       policy_version_policy_name
                       policy_version_last_change
                       encoding_changed
                       flg_allow_parent_policy
                       sync_files
                       external_reference_count
                       external_references
                       content
                     );

  for my $copied_key (@copied_keys) {
    $return_hash->{$copied_key} = $self->{$copied_key};
  }
  return $return_hash;
}

#----------------------------------------------------------------------

sub get_primary_encoding 
{
   my ($self, $policy_id) = @_;
   
   my $sql = "SELECT encoding FROM PLC.PL_POLICIES 
               WHERE PLC.PL_POLICIES.id=?";
   my @row = $self->{dbh}->selectrow_array($sql, undef, $policy_id);
   if(! scalar @row or $DBI::err)
   {
      my $dbi_err = $DBI::err ? $DBI::errstr : 'unknown';
      $self->{logger}->fatal("could not fetch encoding of policy_id $policy_id ($dbi_err)") if $self->{logger};
      return;
   }
   return $row[0];
}

#--------------------------------------------------------------------

sub get_unique_policy_name {
  my($self, %params) = @_;

  my $force_policy_name_flg =  $params{force_policy_name_flg};
  my $force_policy_id = $params{force_policy_id} || 0;
  my $overwrite_flag = $params{overwrite_flag} || $force_policy_name_flg;
  my $dbh = $self->{dbh};
  my $policy_name = $params{policy_name};
  my $add_info = defined $params{add_info} ? $params{add_info} : 1;
  my $force_policy_name;

  my $sql = 'SELECT id, name FROM PLC.PL_POLICIES';

  my $existing_policies = $dbh->selectall_hashref($sql, 'name');
  $self->{logger}->debug("Checking for conflict with existing policy names (desired policy name = $policy_name)") if $self->{logger};

  if (exists $existing_policies->{$policy_name}) {
    if (!($overwrite_flag && $force_policy_id == $existing_policies->{$policy_name}->{id})) {
      # Policy name conflicts with another existing policy, and force_policy_name is not set and we are not overwriting the conflicting policy

      my $new_name = get_new_name($policy_name, $existing_policies);
      $self->{logger}->debug("New policy name is $new_name") if $self->{logger};

      if (not defined $new_name) {
        return $self->fatal_error(user_msg => "A security policy named '$policy_name' already exists in the security policy database.",
                                  log_msg => "could not generate a new name for security policy '$policy_name'");
      }

      $self->add_info(user_msg => "A security policy named '$policy_name' already exists in the security policy database; the new name is '$new_name'.",
                      log_msg => "security policy '$policy_name' already exists, new name is '$new_name'") if $add_info;

      $policy_name = $new_name;
    } 
  }
  return $policy_name;
}

#----------------------------------------------------------------------

sub policy_name_fixup {
    my ($self, %params) = @_;
    my $policy_name = $params{policy_name};

    my $is_standalone = F5::BigipUtils::is_feature_enabled('standalone');
    if (($params{force_policy_name_flg} || $params{skip_mcp}) && ($params{policy_name} =~ m|^/| || $is_standalone ) ) {
        #if replacing a policy do not change the partition
        #Leave name as given in containerized environment
    }
    else {
        my $partition = $params{partition} || "Common";
        my $fix_path = 1;
        if ($policy_name  =~ m|^/$partition/|) {
            my $dirname = dirname($policy_name);
            my ($folder) = @{F5::ASM::COAPI::Query->new('folder', { name => $dirname } )};
            $fix_path = !$folder || $folder->{strict_app_updates};
        }
        if ($fix_path) {
            my $policy_name_basename = basename($policy_name);
            $policy_name = "/$partition/$policy_name_basename";
        }
    }

    #Validate name of the policy to import, but allow duplicate names as import will either rename the policy or overwrite the existing one.
    my %import_policy_field_def = %{F5::ASMConfig::Entity::Policy->fields->{policy_name}};
    delete $import_policy_field_def{validate_unique_value};
    if ($is_standalone || $params{skip_mcp})  {
        delete $import_policy_field_def{validate_mcp_path};
    }
    F5::ASMConfig::Entity::Policy->validate_field(%params,
                        field_value => $policy_name,
                        field_def => \%import_policy_field_def,
                        field_name => 'policy_name',
                        dbh => $self->{dbh},
                       );

    return $policy_name;
}

#----------------------------------------------------------------------

sub get_new_name
{
    my ($old_name, $names) = @_;

    my ($base, $num, $new_name);

    # TODO: should get this from size of name column in PL_POLICIES, or cfg file?
    my $MAX_NAME_LEN = 255;
    
    # There are some templates with the suffix of the year like: owa_2007. If it exists we don't want to get a new name: owa_2008, but rather owa_2007_2.
    if ($old_name =~ /^(.*?)_(?!20\d\d)(\d+)$/)
    {
        ($base, $num) = ($1, $2);
    }
    elsif (not exists $names->{$old_name})
    {
        return $old_name;
    }
    elsif (not exists $names->{"${old_name}_2"})
    {
        $new_name = "${old_name}_2";
        return if (length $new_name > $MAX_NAME_LEN);
        return $new_name;
    }
    else
    {
        $base = $old_name;
        $num = 2;
    }

    # FIND THE NEXT AVAILABLE NAME

    $num++ while (exists $names->{"${base}_$num"});
    $new_name = "${base}_$num";

    return if (length $new_name > $MAX_NAME_LEN);
    return $new_name;
}

#----------------------------------------------------------------------

sub insert_charset_defaults {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $ignore = F5::DbUtils::get_ignore_string($dbh);

  # insert charset for each type
    $dbh->do(<<EndQuery, undef, $params{policy_id}, $params{policy_id});
INSERT $ignore INTO PLC.PL_CHARSETS (policy_id, charset_type, charset, section_type, rest_uuid)
     SELECT ?, default_charsets.charset_type, default_charsets.charset, default_charsets.section_type, default_charsets.rest_uuid
       FROM DCC.DEFAULT_CHARSETS default_charsets, DCC.ENCODINGS encodings, PLC.PL_POLICIES policies
      WHERE encodings.encoding = policies.encoding AND policies.id = ?
        AND default_charsets.encoding_id = encodings.encoding_id 
EndQuery
}

#----------------------------------------------------------------------

sub add_negsig_set_warning {
    my($self, %params) = @_;

    my $idmap = $params{idmap};

    for my $rv (@{$params{rvs}}) {
        if ($rv->{update_status} eq $F5::NegativeSignatures::Set::SET_UPDATE_STATUS_EMPTY) {
            $self->add_warning(msg => "Signature Set $rv->{requested_set_name} is empty and will not be assigned to this security policy.");
        } 

	$$idmap->{$rv->{old_set_id}} = defined $rv->{new_set_id} ? $rv->{new_set_id} : $rv->{old_set_id};

        if ($rv->{update_status} eq $F5::NegativeSignatures::Set::SET_UPDATE_STATUS_ADDED) {
            if ($rv->{set_name} ne $rv->{requested_set_name}) {
                $self->add_warning(msg => <<EndMessage);
The signature set that was imported, $rv->{requested_set_name}, has a different configuration than $rv->{requested_set_name}
currently on this system. Therefore, the imported signature set $rv->{requested_set_name} was renamed to $rv->{set_name},
and will be assigned to this security policy.
EndMessage
            }
            else {
                $self->add_warning(msg => "Signature Set $rv->{requested_set_name} (previously used in this security policy) was added to this system.");
            }
        }

        if ($rv->{missing_manual_sigs}) {
            my $sig_ids = join(', ', @{$rv->{missing_manual_sigs}});
            my $sig_count = scalar @{$rv->{missing_manual_sigs}};

            if ($sig_count > 1) {
                $self->add_warning(msg => "$sig_count signatures defined in Signature Set $rv->{set_name} were not found on this system. (Signature IDs = $sig_ids)");
            }
            else {
                $self->add_warning(msg => "$sig_count signature defined in Signature Set $rv->{set_name} was not found on this system. (Signature ID = $sig_ids)");
            }
        }

        if ($rv->{missing_systems}) {
            my $systems = join(', ', @{$rv->{missing_systems}});
            my $system_count = scalar @{$rv->{missing_systems}};
            if ($system_count > 1) {
                $self->add_warning(msg => "$system_count systems defined in Signature Set $rv->{set_name} were not found on this system. (System Names = $systems)");
            } else {
                $self->add_warning(msg => "$system_count system defined in Signature Set $rv->{set_name} was not found on this system. (System Name = $systems)");
            }
        }
    }

}

#----------------------------------------------------------------------

sub add_unused_default_negsig_set_warning {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my $unused_default_sets = $dbh->selectall_hashref(<<EndQuery, 'set_id', undef, $self->{policy_id});
SELECT sets.set_id, sets.set_name
  FROM PLC.NEGSIG_SETS sets
       LEFT JOIN PLC.PL_POLICY_NEGSIG_SETS pol_sets
              ON (    pol_sets.set_id = sets.set_id
                  AND pol_sets.policy_id = ?)
 WHERE sets.flg_assign_to_policy_default = 1
   AND pol_sets.set_id IS NULL
EndQuery

    for my $set (values %$unused_default_sets) {
        $self->add_info(msg => qq{Default Signature Set "$set->{set_name}" is not used in this security policy.});
    }
}

#----------------------------------------------------------------------

sub overwrite_pl_policies {
  my($self, %params) = @_;

  local $" = ',';
  my $dbh = $self->{dbh};
  my $policy_id = $params{policy_id};
  my $data = $params{data};
  my $row_data = $params{row_data};

  F5::ASMConfig::Entity::Policy::empty_policy_tables({dbh => $dbh,
                                                      policy_id => $policy_id,
                                                      plc_dbname => "PLC",
                                                     },
                                                     import_tables_only => 1,
                                                    );

  my $row = $row_data || (values %$data)[0];
  $row->{has_been_applied} = 0;
  my %excluded_fields = ("id" => 1,
                         "account_id" => 1);
  my @db_fields = F5::DbUtils::get_db_fields($dbh, "PLC.PL_POLICIES");
  my @field_list = grep(exists $row->{$_} && !$excluded_fields{$_} , @db_fields);

  my @update_values = map { $row->{$_} } @field_list;
  my @fields_with_placeholders  = map { "$_ = ?" } @field_list;

  $dbh->do(<<EndQuery, undef, @{$row}{@field_list}, $policy_id);
UPDATE PLC.PL_POLICIES
   SET @fields_with_placeholders
 WHERE id = ?
EndQuery

}

#----------------------------------------------------------------------

sub update_pl_policy_changes_log {
  my($self, %params) = @_;

  local $" = ',';
  my $dbh = $self->{dbh};
  my $policy_id = $params{policy_id};

  $dbh->do('DELETE FROM PLC.PL_POLICY_CHANGES_LOG WHERE policy_id = ?', undef, $policy_id);
  $dbh->do('DELETE FROM PLC.PL_POLICY_ACTION_ITEMS WHERE policy_id = ?', undef, $policy_id);

  $dbh->do(<<EndQuery, undef, $policy_id, 'PL_POLICIES', 'add', '', time);
INSERT INTO PLC.PL_POLICY_CHANGES_LOG (policy_id, modified_table, modify_action, user_name, update_time)
   VALUES (?,?,?,?,?)
EndQuery
}

#----------------------------------------------------------------------

sub complete_user_defined_violation_handling {
    my($self, %params) = @_;

    my $idmap = $params{idmap} || {};
    my $policy_id = $params{policy_id};
    my $import_rvs = $params{import_rvs};

    for my $import_rv (@$import_rvs) {
        # Delete from current policy, to prevent clash with PL_BLOCKING import
        $self->{dbh}->do('DELETE FROM PLC.PL_BLOCKING WHERE policy_id = ? AND viol_index = ?',
                         undef, $params{policy_id}, $import_rv->{entity}->{viol_index});

        if ($import_rv->{import_result} eq $F5::ASMConfig::Entity::Base::IMPORT_RESULT_ADDED) {
            $self->add_warning(msg => "Violation $import_rv->{entity}->{viol_name} (previously used in this security policy) was added to this system.");
        }

        $idmap->{$import_rv->{old_id}} = $import_rv->{entity}->get_id;
    }
}

#----------------------------------------------------------------------
sub map_element_type {

    my($self, %params) = @_;
    my $field_value = $params{field_value};
    my $field = $params{field};

    if ($params{mode} eq 'export'){
        return $self->{entity_schema}->{ElementTypeName}->{$field_value};
    }
    else{
        return $self->{entity_schema}->{ElementType}->{$field_value};
    }
}
#-----------------------------------------------------------
sub convert_to_base64 {
  my($self, %params) = @_;
  my $field_value = $params{field_value};

  if ($params{mode} eq 'export') {
    return MIME::Base64::encode_base64($field_value);
  }
  else {
    return MIME::Base64::decode_base64($field_value);
  }
}

#----------------------------------------------------------------------

1;
