package F5::ImportExportPolicy::YAML;

use strict;
use warnings;

use F5::YAML::Schema::Utils;
use Scalar::Util qw(reftype);
use Encode;
use MIME::Base64;
use JSON::PP;
use F5::ASMConfig::Handler::TaskBase;
use File::Basename;
use F5::BigipUtils;
use F5::Deploy;
use F5::BigipVersionUtils;
use F5::LWP::UserAgent;
use F5::ASMConfig::Entity::SignatureSet;
use F5::DateTimeUtils;
use F5::DbUtils;
use F5::ObjUtils;
use F5::Policy::Overlay;

use base qw(F5::ImportExportPolicy::Base);

my $DEFAULT_TEMPLATE;
my $FULL_EXPORT_DEFAULT_TEMPLATE = 'POLICY_TEMPLATE_BLANK';
my $BASE_URI = '/mgmt/tm/asm';
my $SYNC_DIR = $F5::ASMConfig::Handler::SYNC_DIR;
my $user_defined_sig_id_to_sig_name_and_tag;

sub new {
    my($class, %params) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->{dbh} = $params{dbh} || F5::DbUtils::get_dbh();
    $self->{logger} = $params{logger} || F5::Logger->get_logger(logger_name => 'import_export_policy');
    $self->{_cache} = {};
    $self->{enforce_strictness} = defined($params{enforce_strictness}) ? $params{enforce_strictness} : 1;
    $self->{overlay} = F5::Policy::Overlay->new(logger => $self->{logger});
    $self->{do_validation} = $params{do_validation};
    $self->{external_reference_count} = 0;
    $self->{external_references} = [];
    $DEFAULT_TEMPLATE = 'POLICY_TEMPLATE_MINIMAL_SECURITY';
    $user_defined_sig_id_to_sig_name_and_tag = undef;
}

#----------------------------------------------------------------------

sub import_policy {
    my($self, %params) = @_;

    $self->{asm_config_handler} = $params{asm_config_handler};
    $self->{file_path_map} = $params{file_path_map};

    my $get_policy_attributes_only = $params{get_policy_attributes_only};
    my $declarative_policy = $params{file_data}->{policy} || $params{file_data};
    #Force population of required template field
    $declarative_policy->{template}->{name} ||= '';

    if (! $get_policy_attributes_only) {
        $self->{overlay}->apply_overlay(declarative_policy => $declarative_policy);
    }

    my $policy_modifications = $params{file_data}->{modifications};

    $self->{softwareVersion} = $declarative_policy->{softwareVersion};

    if (F5::BigipUtils::get_feature_value('default_template')) {
        $DEFAULT_TEMPLATE = F5::BigipUtils::get_feature_value('default_template')
    }
    elsif ( !defined $self->{softwareVersion} ||
          ( F5::BigipVersionUtils::compare_versions(version1 => $self->{softwareVersion}, version2 => '16.1.0') < 0 ) ) {
             $DEFAULT_TEMPLATE = 'POLICY_TEMPLATE_FUNDAMENTAL';
    }
    $self->{policy_size} = -s $params{filename};
    $self->check_policy_size();

    if ($policy_modifications && exists $params{file_data}->{modificationsReference} ){
        die {error_message => "Modifications Reference should not be defined in case Modifications are defined."};
    }

    if ( exists($params{file_data}->{modificationsReference}) && exists($params{file_data}->{modificationsReference}->{link}) ){
        $self->{external_reference_count}++;
        my $rv = F5::Utils::File::get_data_from_link(user => $self->{asm_config_handler}->{session}->{gui_user_name},
                                                     link => $params{file_data}->{modificationsReference}->{link},
                                                     parse_json => 1,
                                                     basic_auth => $params{file_data}->{modificationsReference}->{basicAuth},
                                                     file_path_map => $self->{file_path_map},
                                                     calc_md5 => 1,
                                                     );

        if ($rv->{completed_successfully}) {
            $policy_modifications = $rv->{data}->{modifications} || $rv->{data};
            $self->{policy_size} += $rv->{size};
            $self->check_policy_size();
            $self->add_external_reference(file_info => $rv);
        }
        else {
            die {error_message => "Failed parsing JSON Policy modifications file : $rv->{error_message}."};
        }
    }

    delete $params{file_data};

    my $entity_field_map = F5::ASMConfig::Entity::Policy->rest_field_map();
    my %policy_obj;

    my @policy_fields = qw( policy_name
                            leaf_name
                            policy_type
                            encoding
                            parent_policy_name
                          );

    $policy_obj{$entity_field_map->{policy_name}} = $params{policy_name};

    for my $field (@policy_fields) {
        if (defined $declarative_policy->{$entity_field_map->{$field}} && !defined $policy_obj{$entity_field_map->{$field}}) {
            $policy_obj{$entity_field_map->{$field}} = $declarative_policy->{$entity_field_map->{$field}};
        }
    }

    $self->{policy} ||= F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh}, object => {})->from_rest(object => \%policy_obj);
    $self->{policy}->{policy_name} ||= $self->{policy}->{leaf_name};
    if (!$self->{policy}->{policy_name}) {
        return $self->fatal_error(msg => "Policy name must be defined.");
    }

    if ($params{policy_type}) {
        $self->{policy}->{policy_type} = $params{policy_type};
    }

    $self->{orig_policy_name} = $self->policy_name_fixup(%params,
                                                         policy_name => $self->{policy}->{policy_name},);

    $self->{policy}->{policy_name} = $self->{orig_policy_name};
    $self->{policy_name} = $self->{orig_policy_name};

    $self->{policy}->{policy_type} ||= F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY();

    if (defined $declarative_policy->{applicationLanguage} && $declarative_policy->{applicationLanguage} eq 'auto-detect'){
        $self->{policy}->{encoding} = '';
    }

    $self->{policy}->{flg_encoding_selected} = $self->{policy}->{encoding} ? 1 : 0;

    if ($get_policy_attributes_only) {
        if ($params{resolve_external_references}) {
            $self->add_openapi_file(%params,
                                    declarative_policy => $declarative_policy);
            $self->add_references(%params,
                                  references => $declarative_policy);
            $self->add_user_defined_signatures(%params,
                                  declarative_policy => $declarative_policy);
            $self->resolve_links(%params,
                                 declarative_policy => $declarative_policy);

        }
        if ($self->{external_reference_count}) {
            my $json = JSON->new->convert_blessed->allow_blessed->canonical;
            my $file_data = { policy => $declarative_policy,
                              modifications => $policy_modifications };
            $self->{content} = decode('utf8', $json->encode($file_data));
        }
        $self->{flg_allow_parent_policy} = 0;
        return $self->completed_successfully(msg => '');
    }

    my $msgs = '';
    if (! ($declarative_policy->{template} && $declarative_policy->{template}->{name} )) {
        $declarative_policy->{template}->{name} = $DEFAULT_TEMPLATE;
    }

    my $template = F5::ASMConfig::Entity::PolicyTemplate->new(dbh => $self->{dbh},
                                                              get_criteria => {icontrol_name => $declarative_policy->{template}->{name}});

    if (!$template) {
        return $self->fatal_error(msg => "Policy Template '$declarative_policy->{template}->{name}' not found");
    }

    $self->handle_parent_policy(%params);

    if (!defined $self->{policy}->{encoding} ){
        $self->{policy}->{encoding} = 'utf-8';
    }

    if ($params{encoding} && ($params{encoding} ne $self->{policy}->{encoding})) {
        $self->{encoding_changed} = 1;
    }

    eval {
        my $rv = $self->create_base_policy(%params,
                                           policy_template => $template,
                                           template_content => $declarative_policy->{template});

        my $update_call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                              is_in_add_policy => 1,
                                                             );


        $self->{policy} = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                             get_criteria => {policy_id => $rv->{policy_id}});
        $self->{policy}->{policy_template} = $self->{policy}->{policy_template} || $template;
        $self->{policy}->associate_policy_template(%params,
                                                   template => $template,
                                                   defined_signature_sets => exists($declarative_policy->{'signature-sets'}),
                                                  );

        if ($rv->{encoding_changed}) {
            F5::ASMConfig::Entity::Policy::charset_config($self->{policy}->{policy_id},
                                                          {PLC => 'PLC', DCC => 'DCC'},
                                                          $self->{dbh});
        }

        my %update_policy_obj;
        my @update_policy_fields = qw( flg_is_case_insensitive
                                       flg_passive_mode
                                       flg_is_protocol_free
                                       flg_transparent_mode
                                     );

        for my $field (@update_policy_fields) {
            if (defined $declarative_policy->{$entity_field_map->{$field}}) {
                $update_policy_obj{$entity_field_map->{$field}} = $declarative_policy->{$entity_field_map->{$field}};
            }
        }

        if (defined $params{policy_description}) {
            $update_policy_obj{description} = $params{policy_description};
        }
        elsif ( defined $declarative_policy->{description}) {
            $update_policy_obj{description} = $declarative_policy->{description};
        }

        $self->{asm_config_handler}->update_policy($self->{asm_config_handler}->{session},
                                                   $update_call_flags,
                                                   F5::ASMConfig::Policy->new(policy_id => $rv->{policy_id}),
                                                   F5::ASMConfig::Entity::Policy->new(object => {})->from_rest(object => \%update_policy_obj),
                                                  );

        my $template_name = $self->{policy}->{templ__policy_template_name} || $self->{policy}->{policy_template}->{policy_template_name};
        $msgs = "Policy Template set to $template_name.";

        $self->add_openapi_file(%params,
                                declarative_policy => $declarative_policy);

        $self->{policy_id} = $self->{policy}->{policy_id};

        F5::ASMConfig::Cache::clear_cached_policy(policy => {policy_id => $self->{policy_id}});
        $self->delete_old_signatures_with_tag_policy(%params);
        $self->add_user_defined_signatures(%params,
                                           declarative_policy => $declarative_policy);

        $rv = $self->add_references(%params,
                                    references => $declarative_policy);

        if ($rv->{deleted_items}) {
            my $do_delete = 1;
            my $items_to_delete = $rv->{deleted_items};
            while ($do_delete) {
                my $delete_rv = $self->delete_items(%params,
                                                    items => $items_to_delete);
                $do_delete = $delete_rv->{delete_count} && %{$delete_rv->{deferred_items}};
                $items_to_delete = $delete_rv->{deferred_items};
            }
        }

        $self->validate_tag(%params,
                            references => $declarative_policy);

        $self->modify_policy(%params,
                             modifications => $policy_modifications);
    };
    if (my $err = $@) {
        my $error_message = ref $@ ? $@->{error_message} : $@;
        return $self->fatal_error(msg => $error_message);
    }

    if ($params{is_in_nested_import}) {
        $self->{completed_successfully} = 1;
        return $self->build_return_hash;
    }

    return $self->completed_successfully(msg => "The operation was completed successfully. The security policy name is '$self->{policy}->{policy_name}'. $msgs");
}

#----------------------------------------------------------------------

sub add_openapi_file {
    my($self, %params) = @_;

    my $declarative_policy = $params{declarative_policy};
    my $filepath;

    if ($declarative_policy->{'open-api-files'} && $declarative_policy->{openApiFileReference}) {
       die {error_message => "'open-api-files' and 'openApiFileReference' are mutually exclusive."};
    }

    if (exists $declarative_policy->{openApiFileReference}) {
        $self->handle_reference_link(references => $declarative_policy,
                                     reference_link => 'openApiFileReference',
                                     get_policy_attributes_only => $params{get_policy_attributes_only},);
    }
    my $open_api_files = $declarative_policy->{'open-api-files'};
    delete $declarative_policy->{'open-api-files'};

    if (!$open_api_files) {
       return;
    }

    if (reftype $open_api_files ne 'ARRAY') {
       die {error_message => "'open-api-files' should contain array element."};
    }

    my $open_api_files_size = scalar @$open_api_files;

    if ($open_api_files_size > 1) {
        die {error_message => "'open-api-files' should not contain more than one element."};
    }

    my ($open_api_file) = @$open_api_files;
    if (my $open_api_file_path = $open_api_file->{link} || $open_api_file->{filename} || $open_api_file->{'$ref'}) {
        $self->{external_reference_count}++;
        my $rvfile =  F5::Utils::File::get_file_from_link(user => $self->{policy}->{creator_name},
                                                          link => $open_api_file_path,
                                                          basic_auth => $open_api_file->{basicAuth},
                                                          file_path_map => $self->{file_path_map},
                                                          logger => $self->{logger},
                                                          calc_md5 => 1,
                                                          );
        $filepath = $rvfile->{filepath};
        $self->{policy_size} += -s $filepath;

        $self->add_external_reference(file_info => $rvfile);
        $self->check_policy_size();
    }

    if (!$filepath) {
        $self->add_warning(msg => "'open-api-files' does not contain 'link'.");
        return;
    }

    if ($params{get_policy_attributes_only}) {
        my $md5 = F5::Utils::File::get_file_md5($filepath);
        $declarative_policy->{'open-api-files'} = [{filename => "file://$filepath", md5 => $md5}];
        return;
    }


    my $import_open_api_actions_fields = F5::ASMConfig::ImportOpenApiActionsFields->new(filename => $filepath,
                                                                                        is_base64 => 0,
                                                                                        username => $self->{policy}->{creator_name},
                                                                                        api_type => F5::ASMConfig::OpenApiFileType::OPEN_API_FILE_TYPE_SWAGGER());

    my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                   apply_side_effects => 1,
                                                   include_implied_server_technologies => 0,
                                                   do_not_relay => 1);

    my $rv = $self->{asm_config_handler}->import_open_api_update_policy($self->{session},
                                                                        $call_flags,
                                                                        $self->{policy},
                                                                        $import_open_api_actions_fields);

    $self->{policy} = $rv->{policy};
}

#----------------------------------------------------------------------

sub delete_old_signatures_with_tag_policy {
    my($self, %params) = @_;

    my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                   do_not_relay => 1,
                                                   skip_set_sigs_active => 1,
                                                  );

    my $rv = $self->{asm_config_handler}->delete_many_signature_sets($self->{session},
                                                                     $call_flags,
                                                                     undef,
                                                                     F5::ASMConfig::SignatureSet->new(signature_set_filter => { sig_tag_policy_val => $self->{policy}->{policy_name} }),
                                                                    );

    $rv = $self->{asm_config_handler}->delete_many_signatures($self->{session},
                                                              $call_flags,
                                                              undef,
                                                              F5::ASMConfig::Signature->new(sig_tag_policy => $self->{policy}->{policy_name}),
                                                             );
}

#----------------------------------------------------------------------

sub add_user_defined_signatures {
    my($self, %params) = @_;

    my $declarative_policy = $params{declarative_policy};
    my $filepath;

    my $user_defined_signatures_contents = $declarative_policy->{'user-defined-signatures'} || [];

    if (!$params{get_policy_attributes_only}) {
        delete $declarative_policy->{'user-defined-signatures'};
    }

    for my $user_defined_signatures_struct (@$user_defined_signatures_contents) {
        $self->expand_node_references(node => $user_defined_signatures_struct);
        if (!$params{get_policy_attributes_only}) {
            my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                           do_not_relay => 1,
                                                           skip_set_sigs_active => 1,
                                                          );

            my $rv = $self->{asm_config_handler}->update_sigfile($self->{session},
                                                                 $call_flags,
                                                                 F5::ASMConfig::Sigfile->new(file_format => 'json',
                                                                                             file_contents_struct => $user_defined_signatures_struct,
                                                                                             flg_is_user_defined => 1,
                                                                                            ),
                                                                 F5::ASMConfig::SigfileUpdateSettings->new(allow_older_timestamp => 1,
                                                                                                           im_file => 0,
                                                                                                           update_type => F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES(),
                                                                                                           sig_tag_policy => $self->{policy}->{policy_name},
                                                                                                          ),
                                                                );
        }
    }
}

#----------------------------------------------------------------------

sub validate_tag {
    my($self, %params) = @_;

    my $references = $params{references};

    if (!exists $references->{'signature-requirements'}) {
       return;
    }

    my $sig_tags = $self->get_sig_tags();

    my $req_tag_sig = $references->{'signature-requirements'};
    my $sig_sets_tag_map = {};
    my $sig_sets;

    if ($references->{'signature-sets'}) {
       $sig_sets = $references->{'signature-sets'};
       for my $sig_sets_value (@$sig_sets) {
          if (exists $sig_sets_value->{signatureSet} &&
              exists $sig_sets_value->{signatureSet}->{filter} &&
              exists $sig_sets_value->{signatureSet}->{filter}->{tagValue}) {
             $sig_sets_tag_map->{$sig_sets_value->{signatureSet}->{filter}->{tagValue}} = $sig_sets_value->{signatureSet}->{filter}->{tagValue};
          }
          elsif (!exists $sig_sets_value->{signatureSet} && exists $sig_sets_value->{name}) {
               my $set_id = $self->{dbh}->selectrow_array(<<EndQuery, undef, $sig_sets_value->{name});
SELECT set_id
FROM NEGSIG_SETS WHERE set_name = ?
EndQuery
            if ($set_id) {
               my $sig_tag_val = $self->{dbh}->selectrow_array(<<EndQuery, undef, $set_id);
SELECT sig_tag_val
FROM NEGSIG_SET_FILTERS WHERE set_id = ?
EndQuery
               if ($sig_tag_val) {
                  $sig_sets_tag_map->{$sig_tag_val} = $sig_tag_val;
               }
            }
          }
          elsif (exists $sig_sets_value->{signatureSet} &&
                 exists $sig_sets_value->{signatureSet}->{type} &&
                 ($sig_sets_value->{signatureSet}->{type} eq 'manual')) {
                  $sig_sets_tag_map->{$sig_sets_value->{signatureSet}->{type}} = $sig_sets_value->{signatureSet}->{type};
          }
       }
    }

    for my $value (@$req_tag_sig) {
       my $tag = $value->{tag} || '';
       if ($sig_sets && length($tag)) {
          if (!$sig_sets_tag_map->{$tag} && !$sig_sets_tag_map->{'any'} && !$sig_sets_tag_map->{'manual'}) {
             $self->add_warning(msg => "Policy requires signature tag '$tag', but does not include any signature of that tag.");
          }
       }
    } #for

    for my $value (keys %$sig_sets_tag_map) {
       if (length($value) && !$sig_tags->{$value} && $value ne 'any') {
          $self->add_warning(msg => "Tag '$value' is not required in policy but exists in 'signature-sets'.");
       }
    }
} #validate_tag

#----------------------------------------------------------------------

sub handle_reference_link {
    my($self, %params) = @_;

    my $references = $params{references};
    my $reference_link = $params{reference_link};
    my $reference_name = F5::YAML::Schema::Utils::get_reference_name(reference_name => $reference_link);
    my $refs_map = $params{refs_map};
    if (!$reference_name) {
       if ( $references->{$reference_link} && reftype $references->{$reference_link} &&
            reftype $references->{$reference_link} eq 'HASH' &&
            exists $references->{$reference_link}->{'$ref'} ){

           $references->{$reference_link}  =  $self->handle_generic_link(link => $references->{$reference_link}->{'$ref'},
                                                                         basic_auth => $references->{$reference_link}->{basicAuth});
       }
       elsif ($refs_map->{$reference_link} && !$refs_map->{$reference_link}->{entity_type} && $refs_map->{$reference_link}->{references}) {
           for my $sub_reference_link (sort keys %{$references->{$reference_link}}) {
               $self->handle_reference_link(references => $references->{$reference_link},
                                            reference_link => $sub_reference_link,
                                            refs_map => $refs_map->{$reference_link},
                                            get_policy_attributes_only => $params{get_policy_attributes_only},);
           }
       }
       return $reference_link;
    }

    if (exists $references->{$reference_link} && exists $references->{$reference_name}) {
       die {error_message => "'$reference_name' and '$reference_link' are mutually exclusive."};
    }

    my $item = $references->{$reference_link};

    if (reftype $item eq 'HASH' && exists $item->{link}) {
        $references->{$reference_name} = $self->handle_generic_link(link => $item->{link}, basic_auth => $item->{basicAuth});
        delete($references->{$reference_link});
        return $reference_name;
    }
    elsif (reftype $item eq 'ARRAY') {
        die {error_message => "'$reference_link' reference type has to be '{HASH}'."};
    }

    die {error_message => "No 'link' found in '$reference_link'."};
} #handle_referenced_link
#----------------------------------------------------------------------

sub handle_generic_link {
   my($self, %params) = @_;

   my $link = $params{link};
   my $basic_auth = $params{basic_auth};

   if ($link) {
       $self->{external_reference_count}++;
       my $rv = F5::Utils::File::get_data_from_link(user => $self->{asm_config_handler}->{session}->{gui_user_name},
                                                    link => $link,
                                                    basic_auth => $basic_auth,
                                                    parse_json => 1,
                                                    file_path_map => $self->{file_path_map},
                                                    calc_md5 => 1,
                                                    );

       if (!$rv->{completed_successfully}) {
          die {error_message => "Failed parsing validation file : $rv->{error_message}."};
       }
       else {
          $self->{policy_size} += $rv->{size};
          $self->check_policy_size();
          $self->add_external_reference(file_info => $rv);
          return $rv->{data};
       }
   }
} #handle_generic_link
#----------------------------------------------------------------------

my @reference_keywords = ('$ref', 'link');

sub expand_node_references {
    my($self, %params) = @_;
    my $node = $params{node};

    if (reftype($node) eq 'HASH') {
        for my $reference_keyword (@reference_keywords) {
            if ($node->{$reference_keyword}) {
                my $expanded_node = $self->handle_generic_link(link => $node->{$reference_keyword}) || {};
                for my $k (keys %$expanded_node) {
                    if (! exists $node->{$k}) {
                        $node->{$k} = $expanded_node->{$k};
                    }
                }
                delete $node->{$reference_keyword};
                last;
            }
        }
    }
}

#----------------------------------------------------------------------

sub handle_parent_policy {
    my($self, %params) = @_;
    if ($self->{policy}->{policy_type} eq F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        die "Import declarative Parent Policy is not allowed.\n";
    }
    if ($self->{policy}->{parent_policy_name} || $params{parent_policy}) {
        die "Import declarative Security Policy with Parent Policy is not allowed.\n";
    }
    else {
        my $parent_policy_settings = $self->{asm_config_handler}->get_parent_policy_settings($self->{asm_config_handler}->{session},
                                                                                             F5::ASMConfig::CallFlags->new());
        if ( $parent_policy_settings && $parent_policy_settings->{flg_is_parent_policy_mandatory} ) {
            die "Import declarative Policy is not allowed while Parent Policy is mandatory.\n";
        }
    }
}

#----------------------------------------------------------------------

sub create_base_policy {
    my($self, %params) = @_;

    my $template_content = $params{template_content};

    if ($params{force_policy_name_flg}) {
        my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                        get_criteria => { policy_name => $params{policy_name}});
        if ($policy) {
            if ($self->{encoding_changed} && !$params{allow_encoding_change}) {
                die "Failed action: Imported and replaced policies have different encodings.\n";
            }
            $params{force_policy_id} = $policy->{policy_id};
        }
    }

    my $encoding = $params{encoding} || $self->{policy}->{encoding};

    if (exists $template_content->{link}) {
       $self->{external_reference_count}++;
       my $file = F5::Utils::File::get_file_from_link(link => $template_content->{link},
                                                      user => $self->{asm_config_handler}->{session}->{gui_user_name},
                                                      file_path_map => $self->{file_path_map},
                                                      logger => $self->{logger},
                                                      calc_md5 => 1,
                                                      );
       $self->add_external_reference(file_info => $file);
       my $rv = F5::YAML::Schema::Utils::parse_json(filename => $file->{filepath});

       # In case that parse_json fails, the referenced template is not in JSON format (either XML or Binary).
       # Meaning, the import process can continue with the referenced template file.
       # If parse_json succeeds, the template will be built as "JSON policy".
       if ($rv->{completed_successfully}) {
          # Check if the referenced template is referencing to another template, return error if it does to avoid loops.
          my $is_template_referencing = $rv->{data}->{policy} || $rv->{data};
          if (exists $is_template_referencing->{template}->{link}) {
             die {error_message => "Referenced JSON template contains link to another file: '$is_template_referencing->{template}->{link}'."};
          }

          $params{file_data} = $rv->{data};
          return $self->import_policy(%params, is_in_nested_import => 1);
       }
       else {
          $params{policy_template}{template_file} = $file->{filepath};
       }
    }

    #This needs to populate rest_uuid for at least Policy Signatures which would have have existing in the binary default policy, so it is not passing in the value for populate_rest_uuid
    return $self->{policy}->from_template(policy_template => $params{policy_template},
                                          force_policy_name_flg => $params{force_policy_name_flg},
                                          force_policy_id => $params{force_policy_id},
                                          overwrite => $params{force_policy_id} ? 1 : 0,
                                          encoding => $encoding ? $encoding : '',
                                          allow_encoding_change => $params{allow_encoding_change},
                                          flg_encoding_selected => $encoding ? 1 : 0,
                                          policy_name => $self->{policy}->{policy_name},
                                          policy_type => $self->{policy}->{policy_type},
                                          partition => $params{partition},
                                          creator_name => $params{creator_name},
                                          skip_mcp => $params{skip_mcp},
                                          policy_description => $self->{policy}->{description},
                                          retain_inheritance_settings => $params{retain_inheritance_settings},
                                         );
}

#----------------------------------------------------------------------

sub add_references {
    my($self, %params) = @_;

    my $references = $params{references};
    my $refs_map = $F5::ASMConfig::Handler::Rest::path_mapper->{references}->{policies}->{references};
    my $policy_fields = F5::ASMConfig::Entity::Policy->reverse_rest_field_map();

    my $policy_uri;
    if (!$params{get_policy_attributes_only}) {
        #rest_uuid won't be set in get_policy_attributes_only, and the policy_uri is not needed in that case
        $policy_uri  = "$BASE_URI/policies/$self->{policy}->{rest_uuid}";
    }
    my %deferred_items = ();
    my %deleted_items = ();
    my $resolved_external_reference_count = $self->{external_reference_count};

    my @policy_builder_refs;
    my @non_pb_refs;

    for my $ref ( sort keys %$references ) {
        $ref =~ 'policy-builder' ? push @policy_builder_refs, $ref : push @non_pb_refs, $ref ;
    }

    for my $reference (@policy_builder_refs,@non_pb_refs) {
        $reference = $self->handle_reference_link(references => $references,
                                                  reference_link => $reference,
                                                  refs_map => $refs_map,
                                                  get_policy_attributes_only => $params{get_policy_attributes_only},);
        $reference = $self->handle_reference_link(references => $references, reference_link => $reference);
        next if $params{get_policy_attributes_only};
        if (!$refs_map->{$reference}) {
            if (!$policy_fields->{$reference}) {
                $self->{logger}->debug("Invalid reference path '$reference'");
            }
            next;
        }
        if ($resolved_external_reference_count != $self->{external_reference_count}) {
            $resolved_external_reference_count = $self->{external_reference_count};
            my $resolved_reference = { $reference => $references->{$reference} };
            $self->{overlay}->apply_overlay(declarative_policy => $resolved_reference);
            $references->{$reference} = $resolved_reference->{$reference};
        }

        my $reference_def = $refs_map->{$reference};

        my $uri = "$policy_uri/$reference";
        my $entity_type = $reference_def->{entity_type};

        # Init sub_refs hash with sub endpoints (and their definitions) of the current policy reference if exists (for example, blocking-settings/general)
        # If no sub references, init the hash with the policy reference itself.

        my $sub_refs = { $reference => $reference_def };
        if (!$entity_type) {
            next if !$refs_map->{$reference}->{references};
            $sub_refs = {};
            # This was added to accommodate REST grouping URI bot-defense/mitigations/signatures -- i.e. with
            # 3 levels of sub-refs. This should be sufficient for any reason declarative policy structure, but
            # if not, this should be refactored for true recursion.
            for my $sub_ref_name (keys %{$refs_map->{$reference}->{references}}) {
                if (exists $refs_map->{$reference}->{references}->{$sub_ref_name}->{references}) {
                    my $sub_sub_refs = $refs_map->{$reference}->{references}->{$sub_ref_name}->{references};
                    for my $sub_sub_ref (keys %$sub_sub_refs) {
                        $sub_refs->{"$sub_ref_name/$sub_sub_ref"} = $sub_sub_refs->{$sub_sub_ref};
                    }
                }
                else {
                    $sub_refs->{$sub_ref_name} = $refs_map->{$reference}->{references}->{$sub_ref_name};
                }
            }
        }

        # For each policy reference:
        #    Retrieve items (collection)/item(singleton) for each policy reference.
        #    Calculate the appropriate rest method.
        #    Send items for further handling.
        #    Save deferred items for later iteration of add_references.

        for my $sub_ref (keys %$sub_refs) {
            $reference_def = $sub_refs->{$sub_ref};

            my $items = $self->extract_items(references => $references,
                                             reference => $reference,
                                             sub_reference => $sub_ref,
                                             reference_def => $reference_def
                                            );

            next if !@$items;

            $entity_type = $reference_def->{entity_type};
            my $singleton = $reference_def->{singleton};
            my $method = $singleton ? 'PATCH' : 'POST';
            my $allowed_methods = $reference_def->{allowed_methods};
            my $collection_allowed_methods = $reference_def->{collection_allowed_methods};

            if ($singleton) {
                next if $allowed_methods && !$allowed_methods->{$method};
            }
            elsif ($collection_allowed_methods || $allowed_methods) {
                my $collection_allowed_method;
                for my $collection_method ( qw(POST PATCH) ) {
                    if ($collection_allowed_methods && $collection_allowed_methods->{$collection_method}) {
                        $collection_allowed_method = $collection_method;
                        last;
                    }
                    if ($allowed_methods && $allowed_methods->{$collection_method}) {
                        $collection_allowed_method = $collection_method;
                        last;
                    }
                }
                next if !$collection_allowed_method;
                $method = $collection_allowed_method;
            }

            my $rv = $self->handle_items(references => $references,
                                         items => $items,
                                         entity_type => $entity_type,
                                         uri => ($sub_ref eq $reference) ? $uri : "$uri/$sub_ref",
                                         singleton => $singleton,
                                         method => $method);

            if (@{$rv->{deferred_items}}) {
                my $path = ($sub_ref eq $reference) ? $reference : "$reference/$sub_ref";
                my @parts = split '/', $path;
                my $last = pop @parts;
                my $deferred_struct_pointer = \%deferred_items;
                for my $part (@parts) {
                    $deferred_struct_pointer = ($deferred_struct_pointer->{$part} ||= {});
                }
                if ($singleton) {
                    $deferred_struct_pointer->{$last} = $rv->{deferred_items}->[0];
                }
                else {
                    push (@{$deferred_struct_pointer->{$last}}, @{$rv->{deferred_items}});
                }
            }

            if (defined $rv->{deleted_items} && @{$rv->{deleted_items}}) {
                push (@{$deleted_items{$entity_type}}, @{$rv->{deleted_items}});
            }
        }
    }

    # After all dependencies has been added, add the current deferred items to policy
    if (%deferred_items) {
        $self->validate_cyclic_reference(references => $references,
                                         deferred_items => \%deferred_items);
        $self->add_references(%params, references => \%deferred_items);
    }

    return { deleted_items => \%deleted_items };
}

#----------------------------------------------------------------------

sub extract_items {
    my($self, %params) = @_;

    my $reference = $params{reference};
    my $sub_reference = $params{sub_reference};
    my $reference_def = $params{reference_def};

    my $reference_content;
    my $reference_display_string;
    if ($sub_reference eq $reference) {
        $reference_content = $params{references}->{$reference};
        $reference_display_string = "$reference";
    }
    else {
        $reference_content = $params{references}->{$reference};
        for my $sub_ref_part (split(m|/|, $sub_reference)) {
            if ($reference_content && reftype($reference_content) eq 'HASH' && exists $reference_content->{$sub_ref_part}) {
                $reference_content = $reference_content->{$sub_ref_part};
            }
            else {
                $reference_content = undef;
            }
        }
        if (!defined $reference_content) {
            return [];
        }
        $reference_display_string = "$reference/$sub_reference";
    }

    my $items;
    my $reference_type = reftype($reference_content);
    if ($reference_def->{singleton}) {
        if ($reference_type ne 'HASH') {
            my $content_display_string = substr(encode_json($reference_content), 0, 50);
            die "Invalid reference '$reference_display_string' -- singleton reference must be a dictionary. Found $content_display_string\n";
        }
        $items = [ $reference_content ];
    }
    else {
        if ($reference_type ne 'ARRAY') {
            my $content_display_string = substr(encode_json($reference_content), 0, 50);
            die "Invalid reference '$reference_display_string' -- collection reference must be a list. Found $content_display_string\n";
        }
        $items = $reference_content;
    }

    return $items;
}

#----------------------------------------------------------------------

sub handle_items {
    my($self, %params) = @_;

    my $references = $params{references};
    my $items = $params{items};
    my $entity_type = $params{entity_type};
    my $uri = $params{uri};
    my $singleton = $params{singleton};

    my @deferred_items;
    my @deleted_items;
    my @policy_references;

    $entity_type->before_declarative_import(dbh => $self->{dbh},
                                            policy => $self->{policy},
                                            entities => $items,
                                           );

    if ( $entity_type && $entity_type->entity_config->{declarative_items_handler_func} ) {
        my $func = $entity_type->entity_config->{declarative_items_handler_func};
        $self->$func(items => $items);

        return { deferred_items => \@deferred_items,
                 policy_references => \@policy_references };
    }

    # check if items (collection)/$item (singleton) contains policy references and fix references to declarative format.
    # If item has reference to another policy entity, add it to deferred items list.

    for my $item (@$items) {
        my $method = $params{method};
        my $rv = F5::YAML::Schema::Utils::item_references_fixup(item => $item,
                                                                entity_type => $entity_type,
                                                                direction => 'from_declarative',
                                                               );
        my $item_policy_references = $rv->{policy_references};
        my $deferred = 0;
        for my $policy_reference (@$item_policy_references) {
            if ($references->{$policy_reference}) {
                $deferred = 1;
                push(@policy_references, $policy_reference);
            }
        }

        if ($deferred) {
            push (@deferred_items, $item);
            next;
        }

        $self->handle_negsig_overrides(item => $item);

        if ( !$singleton && $item && ref($item) eq 'HASH' && $item->{'$action'} ){
            my $action = lc($item->{'$action'});
            delete $item->{'$action'};
            if ($action eq 'delete'){
                push (@deleted_items, $item);
                next;
            }
        }

        if (my $func =((F5::BigipUtils::is_docker() && $entity_type->entity_config->{standalone_declarative_handler_func}) || $entity_type->entity_config->{declarative_handler_func})) {
            $self->$func(item => $item,
                         entity_type => $entity_type,
                         uri => $uri,
                         method => $method,
                        );
            next;
         }


        # If collection is a closed set, retrieve the exists item and update it.

        my $item_uri;
        if (!$singleton && $method eq 'PATCH') {
            $item_uri = $self->get_item_uri(entity_type => $entity_type,
                                            uri => $uri,
                                            item => $item,
                                           );

            if (!$item_uri->{uri}) {
                my $element_type_display_name = $item_uri->{entity_criteria}->get_element_type_display_name;
                my $entity_display_name = $item_uri->{entity_criteria}->get_display_name();
                $self->add_warning(msg => "$element_type_display_name $entity_display_name was not found on the system -- modification aborted.");
                next;
            }
        }

        $self->send_rest_call(uri => (!$singleton && $method eq 'PATCH') ? $item_uri->{uri} : $uri,
                              item => $item,
                              method => $method,
                             );
    }

    return { deferred_items => \@deferred_items,
             deleted_items =>  \@deleted_items,
             policy_references => \@policy_references };
}

#----------------------------------------------------------------------

sub delete_items {
    my($self, %params) = @_;
    my $deleted_items = $params{items};
    my %deferred_items = ();

    my $delete_count = 0;
    my $deferred_item_count = 0;
    my $default_item_count = 0;
    for my $entity_type (keys %$deleted_items) {
        my $is_referenced_entity = exists $entity_type->fields->{is_referenced};
        my @items = @{$deleted_items->{$entity_type}};
        while (my $item = shift @items) {
            my $rv = $self->get_entity(item => $item,
                                       entity_type => $entity_type,
                                       selected_fields => [ 'rest_uuid', 'is_referenced', 'is_default' ],);
            next if !$rv;
            my $entity = $rv->{entity};
            if (!$entity->{rest_uuid}) {
                my $element_type_display_name = $rv->{entity_criteria}->get_element_type_display_name;
                my $entity_display_name = $rv->{entity_criteria}->get_display_name();
                $self->add_warning(msg => "$element_type_display_name $entity_display_name was not found on the system -- modification aborted.");
                next;
            }

            if ($is_referenced_entity && $entity->{is_referenced}) {
                push (@{$deferred_items{$entity_type}}, $item);
                $deferred_item_count++;
                if ($entity->{is_default}) {
                    $default_item_count++;
                }
                next;
            }

            my $element_type_name = $entity_type->element_type_name;
            my $rpc_method = "delete_$element_type_name";
            my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                           do_not_relay => 1, );
            $self->{asm_config_handler}->$rpc_method($self->{asm_config_handler}->{session},
                                                     $call_flags,
                                                     $self->{policy},
                                                     $entity);
            $delete_count++;
        }
    }

    if ($deferred_item_count && !$delete_count && $deferred_item_count != $default_item_count) {
        die { error_message => "Invalid Policy -- referenced objects cannot be deleted.\n" };
    }

    return { delete_count => $delete_count,
             deferred_items => \%deferred_items,
           }

}

#----------------------------------------------------------------------

sub send_rest_call {
    my($self, %params) = @_;
    my $item = $params{item};
    my $uri = $params{uri};
    my $method = $params{method};

    my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                   apply_side_effects => 1,
                                                   include_implied_server_technologies => 0,
                                                   do_not_relay => 1,
                                                   display_rest_names => 1,
                                                  );

    my $rest_request = F5::ASMConfig::RestRequest->new(rest_path => $uri,
                                                       rest_method => $method,
                                                       rest_body => encode_json($item),
                                                       rest_parameters => { updateOnDuplicateRecord => 1, ver => $self->{softwareVersion} },
                                                      );

    $self->{asm_config_handler}->rest_call($self->{asm_config_handler}->{session}, $call_flags, $rest_request);
}

#----------------------------------------------------------------------

sub get_item_uri {
    my($self, %params) = @_;

    my $entity_type = $params{entity_type};
    my $uri = $params{uri};
    my $item = $params{item};

    my $entity_rv = $self->get_entity(%params,
                                      selected_fields => [ 'rest_uuid' ],);
    my $rv;
    if ($entity_rv) {
        $rv->{entity_criteria} = $entity_rv->{entity_criteria};
        if ($entity_rv->{entity}->{rest_uuid}) {
            $rv->{uri} = "$uri/$entity_rv->{entity}->{rest_uuid}";
        }
    }
    return $rv;
}

#----------------------------------------------------------------------

sub get_entity {
    my($self, %params) = @_;
    my $entity_type = $params{entity_type};
    my $item = $params{item};
    my $selected_fields = $params{selected_fields};

    if ( !$item || ref($item) ne 'HASH'){
        return;
    }
    my @unique_record_parts = grep {defined($_)}  map { $entity_type->rest_field_map->{$_} } @{$entity_type->unique_record_parts};
    my %obj = map { $_ => $item->{$_} } @unique_record_parts;

    my $entity_criteria = $entity_type->new(dbh => $self->{dbh},
                                            object => {})->from_rest(object => \%obj);

    my $rv = { entity_criteria => $entity_criteria };
    $rv->{entity} = $entity_type->new(dbh => $self->{dbh},
                                      policy => $self->{policy},
                                      get_criteria => $entity_criteria,
                                      selected_fields => $selected_fields,
                                     );
    return $rv;
}

#----------------------------------------------------------------------

sub is_deferred_infinitely {
    my($self, %params) = @_;

    my $references = $params{references};
    my $deferred_items = $params{deferred_items};

    # Check if the current and the next iterations of references are the same.
    # If items are being deferred infinitely, the next iteration of items has not changed.

    if ( scalar(keys %$references) != scalar(keys %$deferred_items) ) {
        return 0;
    }

    for my $reference (keys %$references) {
        return if !exists $deferred_items->{$reference};
        if (reftype($references->{$reference}) eq 'ARRAY') {
            return 0 if @{$references->{$reference}} != @{$deferred_items->{$reference}};
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub validate_cyclic_reference {
    my($self, %params) = @_;

    my $references = $params{references};
    my $deferred_items = $params{deferred_items};

    return if !$self->is_deferred_infinitely(%params);

    # Attempt to fix cycle by handling remaining urls.
    # At this point, the only dependency for urls is positional parameters.
    # Since they must be global params, the dependency has already satisfied on a previous iterations.

    if ($references->{urls}) {
        delete $deferred_items->{urls};
        $self->add_references(references => { urls => $references->{urls} } );
    }
    else {
        die { error_message => "Cyclic reference detected." };
    }
}
#----------------------------------------------------------------------

sub handle_excluded_entity {
    my($self, %params) = @_;

    my $entity_type = $params{entity_type};
    my $element_type_display_name = $entity_type->get_element_type_display_name;

    $self->add_warning(msg => "$element_type_display_name is not supported in Declarative Policy -- skipped.");

    return { completed_successfully => 1 };
}
#----------------------------------------------------------------------

my %validation_file_map = (
    'F5::ASMConfig::Entity::JsonProfile' => { files => 'validationFiles',
                                              file_key => 'jsonValidationFile',
                                              file_entity_type => 'F5::ASMConfig::Entity::JsonValidationFile',},
    'F5::ASMConfig::Entity::XmlProfile'  => { files => 'validationFiles',
                                              file_key => 'xmlValidationFile',
                                              file_entity_type => 'F5::ASMConfig::Entity::XmlValidationFile',},
    'F5::ASMConfig::Entity::GrpcProfile' => { files => 'idlFiles',
                                              file_key => 'idlFile',
                                              file_entity_type => 'F5::ASMConfig::Entity::IdlFile',},
 );

sub handle_profile {
    my($self, %params) = @_;

    my $entity_type = $params{entity_type};
    my $item        = $params{item};

    my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                   update_on_adding_duplicate_record => 1);

    my $validation_file_mapping = $validation_file_map{$entity_type};
    my $validation_files = $item->{$validation_file_mapping->{files}};

    for my $validation_file_item (@$validation_files) {
       if (my $link_item = $validation_file_item->{$validation_file_mapping->{file_key}} ) {
           $self->handle_profile_validation_file(item => $link_item,
                                                 entity_type => $validation_file_mapping->{file_entity_type},);
       }
    }

    my $profile = $entity_type->new(dbh => $self->{dbh},
                                    object => {})->from_rest(object => $item, policy => $self->{policy});

    my $add_func = 'add_' . $entity_type->element_type_name;
    $self->{asm_config_handler}->$add_func($self->{asm_config_handler}->{session},
                                           $call_flags,
                                           $self->{policy},
                                           $profile);

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

sub resolve_links {
    my ($self, %params) = @_;
    my $declarative_policy = $params{declarative_policy};
    my $file_path_map = $params{file_path_map};
    my @items = F5::ObjUtils::get_deeply(ref => $declarative_policy, search_keys => \@reference_keywords, return_objects => 1);
    for my $item (@items) {
        $self->{external_reference_count}++;
        my $link = $item->{link} || $item->{'$ref'};
        my $rvfile =  F5::Utils::File::get_file_from_link(user => $self->{policy}->{creator_name},
                                                          link => $link,
                                                          basic_auth => $item->{basicAuth},
                                                          file_path_map => $self->{file_path_map},
                                                          logger => $self->{logger},
                                                          calc_md5 => 1,
                                                          );
        $self->add_external_reference(file_info => $rvfile);
        $item->{md5} = $rvfile->{md5};
    }
}

#----------------------------------------------------------------------

sub item_link_to_contents {
   my($self, %params) = @_;
   my $name_field = $params{name_field};

   my $link = $params{item}->{link} || $params{item}->{'$ref'};
   if ($link &&
       $params{item}->{contents}){
       die {error_message => "Link and contents are mutually exclusive."};
   }

   if ($link) {
       $self->{external_reference_count}++;
       my $rv = F5::Utils::File::get_data_from_link(user => $self->{asm_config_handler}->{session}->{gui_user_name},
                                                    link => $link,
                                                    basic_auth => $params{item}->{basicAuth},
                                                    file_path_map => $self->{file_path_map},
                                                    calc_md5 => 1,
                                                    );

       if (!$rv->{completed_successfully}) {
          die {error_message => "Failed parsing validation/IDL file : $rv->{error_message}."};
       }
       else {
          $params{item}->{contents} = $rv->{data};
          if ($name_field && !$params{item}->{$name_field} && $rv->{filename}) {
              $params{item}->{$name_field} = $rv->{filename};
          }
       }

       $self->add_external_reference(file_info => $rv);
       delete($params{item}->{link});
       delete($params{item}->{'$ref'});
   }
}

#----------------------------------------------------------------------

sub handle_profile_validation_file {
    my($self, %params) = @_;

    $self->item_link_to_contents(item => $params{item}, name_field => 'fileName');

    if ($params{item} && defined($params{item}->{contents})) {
        if ($params{item}->{isBase64}) {
            delete $params{item}->{isBase64};
            $params{item}->{contents} = MIME::Base64::decode_base64($params{item}->{contents})
        }

        my $entity_type = $params{entity_type};
        my $validation_file = $entity_type->new(dbh => $self->{dbh},
                                                object => {})->from_rest(object => $params{item});

        my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                       update_on_adding_duplicate_record => 1);

        my $add_func = 'add_' . $entity_type->element_type_name;

        $self->{asm_config_handler}->$add_func($self->{asm_config_handler}->{session},
                                               $call_flags,
                                               $self->{policy},
                                               $validation_file);
    }

    return { completed_successfully => 1 };
}
#----------------------------------------------------------------------

sub handle_login_enforcement {
    my($self, %params) = @_;

    my $item = $params{item};
    my $uri = $params{uri};

    my $logout_urls = $item->{logoutUrls};
    if (defined $logout_urls && @{$logout_urls}) {
        for my $logout_url (@{$logout_urls}) {
            my $url_uri = $self->get_item_uri(entity_type => 'F5::ASMConfig::Entity::URL',
                                              uri => "$BASE_URI/policies/$self->{policy}->{rest_uuid}/urls",
                                              item => $logout_url->{url},
                                             );
            $logout_url->{urlReference}->{link} = $url_uri->{uri};
            delete $logout_url->{url};
        }
    }

    $self->send_rest_call(uri => $uri,
                          item => $item,
                          method => 'PATCH',
                         );

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

sub handle_brute_force {
    my($self, %params) = @_;

    my $item = $params{item};
    my $uri = $params{uri};
    my $method = 'POST';

    my $url = $item->{url};
    if (!$url) {
        my $default_brute_force = F5::ASMConfig::Entity::BruteForce->new(dbh => $self->{dbh},
                                                                         policy => $self->{policy},
                                                                         get_criteria => { flg_is_default => 1 } );
        if ($default_brute_force) {
            $method = 'PATCH';
            $uri = "$uri/$default_brute_force->{rest_uuid}"
        }
    }
    else {
        my $url_uri = $self->get_item_uri(entity_type => 'F5::ASMConfig::Entity::URL',
                                          uri => "$BASE_URI/policies/$self->{policy}->{rest_uuid}/urls",
                                          item => $url,
                                         );
        $item->{urlReference}->{link} = $url_uri->{uri};
        delete $item->{url};
    }

    $self->send_rest_call(uri => $uri,
                          item => $item,
                          method => $method,
                         );

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

my %types = (
             'json' => 'F5::ASMConfig::Entity::JsonProfile',
             'object' => 'F5::ASMConfig::Entity::JsonProfile',
             'openapi-array' => 'F5::ASMConfig::Entity::JsonProfile',
             'xml' => 'F5::ASMConfig::Entity::XmlProfile',
            );

sub handle_parameter_content_profile {
    my($self, %params) = @_;
    my $item = $params{item};
    my $uri = $params{uri};
    my $method = $params{method};

    my $param_profile = $item->{contentProfile};
    my $value_type = $item->{valueType} || '';
    my $entity_type = $types{$value_type};
    if (defined $param_profile && (reftype $param_profile eq 'HASH')
         && defined $param_profile->{contentProfile} && defined $entity_type) {
        my $rest_path = $entity_type->entity_config->{rest_paths}->[0]->{rest_path_properties}->{rest_path};
        my $profile_uri = $self->get_item_uri(entity_type => $entity_type,
                                              uri => "$BASE_URI/policies/$self->{policy}->{rest_uuid}/$rest_path",
                                              item => $param_profile->{contentProfile},
                                             );
        $param_profile->{contentProfileReference}->{link} = $profile_uri->{uri};
        delete $param_profile->{contentProfile};
    }

    $self->send_rest_call(uri => $uri,
                          item => $item,
                          method => $method,
                         );

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

sub handle_websocket_url_content_profile {
    my($self, %params) = @_;
    my $item = $params{item};
    my $uri = $params{uri};
    my $method = $params{method};

    for my $type ( qw(jsonProfile plainTextProfile) ) {
        next if !defined $item->{$type};
        my $entity_type = 'F5::ASMConfig::Entity::' . ucfirst($type);
        my $rest_path = $entity_type->entity_config->{rest_paths}->[0]->{rest_path_properties}->{rest_path};
        my $profile_uri = $self->get_item_uri(entity_type => $entity_type,
                                              uri => "$BASE_URI/policies/$self->{policy}->{rest_uuid}/$rest_path",
                                              item => $item->{$type},
                                             );
        $item->{$type . 'Reference'}->{link} = $profile_uri->{uri};
        delete $item->{$type};
    }

    $self->send_rest_call(uri => $uri,
                          item => $item,
                          method => $method,
                         );

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

sub handle_negsig_overrides {
    my($self, %params) = @_;
    my $item = $params{item};

    return if !$item->{signatureOverrides};

    my $negsig_overrides = $item->{signatureOverrides};
    if (!$self->{sig_ids}) {
        $self->{sig_ids} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_id');
SELECT sig_id
FROM PLC.NEGSIG_SIGNATURES
EndQuery
    }

    my @non_exists_sigs = grep { $_->{signature}->{signatureId} && !$self->{sig_ids}->{$_->{signature}->{signatureId}} } @$negsig_overrides;
    if (@non_exists_sigs) {
        @{$item->{signatureOverrides}} = grep { !$_->{signature}->{signatureId} || $self->{sig_ids}->{$_->{signature}->{signatureId}} } @$negsig_overrides;
    }

    for my $sig (@non_exists_sigs) {
        $self->add_warning(msg => "Signature with id $sig->{signature}->{signatureId} (previously used in this policy) does not exist on this system.");
    }

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

sub get_sig_tags {
    my($self) = @_;

    if (! $self->{_cache}->{sig_tags}) {
        $self->{_cache}->{sig_tags} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_tag');
SELECT sig_tag, MAX(sig_tag_policy) as sig_tag_policy_exists
FROM PLC.NEGSIG_SIGNATURES
GROUP BY sig_tag
EndQuery
    }
    return $self->{_cache}->{sig_tags};
}

#----------------------------------------------------------------------

sub handle_signature_set {
    my($self, %params) = @_;
    my $item = $params{item};
    my $uri = $params{uri};

    my $set_name = '';
    my $add_new_set = 1;

    if ($item->{signatureSet} && $item->{signatureSet}->{name}) {
        $set_name = $item->{signatureSet}->{name};
    }

    if (!$self->{sig_sets}) {
        $self->{sig_sets} = $self->{dbh}->selectall_hashref(<<EndQuery, 'set_name');
SELECT set_name
  FROM PLC.NEGSIG_SETS
EndQuery
    }

    if( defined $item->{signatureSet} && defined $item->{signatureSet}->{systems} ){
        for my $system (@{$item->{signatureSet}->{systems}} ) {
             $system->{system}->{name} = $system->{name};
              delete $system->{name};
        }
    }

    my $sig_tags = $self->get_sig_tags();
    if ($item->{tag}) {
        if (! $set_name) {
            if ($sig_tags->{$item->{tag}}->{sig_tag_policy_exists}) {
                $set_name = "$self->{policy}->{policy_name} $item->{tag} tagged signatures";
            }
            else {
                $set_name = "$item->{tag} tagged signatures";
            }
        }
        $item->{signatureSet}->{filter}->{tagFilter} = 'eq';
        $item->{signatureSet}->{filter}->{tagValue} = $item->{tag};
        if ($sig_tags->{$item->{tag}}->{sig_tag_policy_exists}) {
            $item->{signatureSet}->{filter}->{tagPolicyValue} = $self->{policy}->{policy_name};
        }
        delete $item->{tag};
    }

    my $next_set_name_suffix = 2;
    my $orig_set_name = $set_name;
    while ($self->{sig_sets}->{$set_name}) {
         if ($self->is_policy_set_identical(set_name => $set_name, item => $item)) {
             $add_new_set = 0;
             last;
         }
         else {
             $set_name = $orig_set_name . "_" . $next_set_name_suffix;
             $next_set_name_suffix++;
         }
    }

    $item->{signatureSet}->{name} = $set_name;
    if ( $add_new_set ) {

        $self->add_warning(msg => qq{Creating Signature Set "$set_name"});
        $item->{signatureSet}->{defaultAlarm} = JSON::false;
        $item->{signatureSet}->{defaultBlock} = JSON::false;
        $item->{signatureSet}->{defaultLearn} = JSON::false;
        $item->{signatureSet}->{isUserDefined} = JSON::true;
        $item->{signatureSet}->{assignToPolicyByDefault} = JSON::false;

        $self->send_rest_call(uri => $BASE_URI.'/signature-sets',
                              item => $item->{signatureSet},
                              method => 'POST',
                             );
    }

    $self->send_rest_call(uri => $uri,
                          item => $item,
                          method => 'POST',
                         );

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------
sub handle_policy_signatures_items {
    my($self, %params) = @_;
    my $items = $params{items};

    my $update_sig_map = {
                            enabled_sigs  => { update_field => 'flg_disabled',    update_value => 0 },
                            disabled_sigs => { update_field => 'flg_disabled',    update_value => 1 },
                            staged_sigs   => { update_field => 'flg_quarantined', update_value => 1 },
                            unstaged_sigs => { update_field => 'flg_quarantined', update_value => 0 },
                         };

    my $policy_id = $self->{policy_id};
    my $dbh = $self->{dbh};

    if ( !$items || !@{$items} ){
        return;
    }

    for my $sig ( @{$items} ){

        if (!$sig->{signatureId}) {
            if ($sig->{name}){
                $sig->{signatureId} = $self->get_sig_id_from_name_and_tag( sig_name => $sig->{name}, sig_tag => $sig->{tag});
                if (!$sig->{signatureId}){
                    $self->add_warning(msg => "Signature with name $sig->{name} (previously used in this policy) does not exist on this system.");
                }
            }

            next if !$sig->{signatureId};
        }

        if ( defined $sig->{enabled} ){
            if ( $sig->{enabled} eq JSON::true) {
                push @{$update_sig_map->{enabled_sigs}->{signatures}}, $sig->{signatureId};
            }
            else{
                push @{$update_sig_map->{disabled_sigs}->{signatures}}, $sig->{signatureId};
            }
        }

        if ( defined $sig->{performStaging} ){
            if ( $sig->{performStaging} eq JSON::true) {
                push @{$update_sig_map->{staged_sigs}->{signatures}}, $sig->{signatureId};
            }
            else {
                push @{$update_sig_map->{unstaged_sigs}->{signatures}}, $sig->{signatureId};
            }
        }
    }

    for my $sig_list ( keys %$update_sig_map) {
        if ( $update_sig_map->{$sig_list}->{signatures} ){
            my $field = $update_sig_map->{$sig_list}->{update_field};
            my $update_value = $update_sig_map->{$sig_list}->{update_value};

            my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => { policy_id => $policy_id,
                                                                                                 sig_id => $update_sig_map->{$sig_list}->{signatures} });
            my $where_clause = $constructed_where_clause->{where_clause};
            my @where_values = @{$constructed_where_clause->{where_values}};

            eval{
                $dbh->do(<<EndQuery, undef, $update_value,  @where_values );
UPDATE PLC.PL_POLICY_NEGSIG_SIGNATURES
   SET $field = ?
 WHERE $where_clause
EndQuery

            };
            if (my $err = $@) {
               $self->{logger}->warn($err);
               $self->add_warning(msg => "Failed handling policy signatures: $err");
            }
        }
    }

}
#----------------------------------------------------------------------
sub modify_policy {
    my($self, %params) = @_;

    my $asm_config_handler = $self->{asm_config_handler};
    my $modifications = $params{modifications};

    for my $modification (@$modifications) {
        my $entity_type;
        my $mapping_key = $modification->{entityKind} || $modification->{entityType};
        $entity_type = F5::YAML::Schema::Utils::get_entity_type(entity_kind => $mapping_key);

        my $action = $modification->{action};
        my $entity;
        my $store_entity;

        F5::YAML::Schema::Utils::item_references_fixup(item => $modification->{entity},
                                                       entity_type => $entity_type,
                                                       direction => 'from_declarative',
                                                      );

        $entity = $entity_type->new(dbh => $self->{dbh},
                                    policy => $self->{policy},
                                    object => {})->from_rest(object => $modification->{entity});

        if ($modification->{entityChanges}) {
            F5::YAML::Schema::Utils::item_references_fixup(item => $modification->{entityChanges},
                                                           entity_type => $entity_type,
                                                           direction => 'from_declarative',
                                                          );

            $store_entity = $entity_type->new(dbh => $self->{dbh},
                                              object => {})->from_rest(object => $modification->{entityChanges});
        }

        my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                       append_nested_struct_on_update => ($modification->{action} eq 'update-append'),
                                                       remove_nested_struct_on_update => ($modification->{action} eq 'update-remove'),
                                                       add_on_updating_nonexistent_record => 1,
                                                       cascade_disable_flags => 1,
                                                       apply_side_effects => 1,
                                                      );
        my @missing_entities;
        my @missing_dependent_external_entities;

        my $event_type = ($action eq 'delete' ) ? F5::ASMConfig::EventType::DELETE() : F5::ASMConfig::EventType::UPDATE();
        my $function = $asm_config_handler->get_function_name(element_type => $entity_type->element_type,
                                                              event_type => $event_type,);

        if ($event_type == F5::ASMConfig::EventType::UPDATE()) {
            @missing_entities = @{$entity->get_missing_external_entities(existing_entity => $entity, call_flags => $call_flags)};

            for my $missing_entity (@missing_entities) {
                my $add_missing_entity_function_name = $asm_config_handler->get_function_name(element_type => $missing_entity->element_type(),
                                                                                              event_type => F5::ASMConfig::EventType::UPDATE(),);
                $missing_entity->{in_staging} = $entity->{in_staging};

                $asm_config_handler->$add_missing_entity_function_name($asm_config_handler->{session}, $call_flags, $self->{policy}, $missing_entity);
            }

            @missing_dependent_external_entities = @{$entity->get_missing_dependent_external_entities()};
        }

        $asm_config_handler->$function($asm_config_handler->{session}, $call_flags, $self->{policy}, $entity, $store_entity);

        for my $missing_entity (@missing_dependent_external_entities) {
            my $add_missing_entity_function_name = $asm_config_handler->get_function_name(element_type => $missing_entity->element_type(),
                                                                                          event_type => F5::ASMConfig::EventType::ADD(),);
            $missing_entity->{in_staging} = $entity->{in_staging};
            $asm_config_handler->$add_missing_entity_function_name($asm_config_handler->{session}, $call_flags, $self->{policy}, $missing_entity);
        }
    }
}

#----------------------------------------------------------------------

sub is_policy_set_identical{
    my($self, %params) = @_;

    my $set_name = $params{set_name};
    my $item = $params{item};

    my $new_sig_set = F5::ASMConfig::Entity::SignatureSet->new(preload_default => 1, dbh => $self->{dbh}, object => {} )->from_rest(object => $item->{signatureSet} );

    my $current_sig_set = F5::ASMConfig::Entity::SignatureSet->new(dbh => $self->{dbh},
                                                                  get_criteria => {set_name => $set_name} );

    my @check_fields = qw(flg_is_manual signature_set_filter systems );
    if ($new_sig_set->{flg_is_manual}) {
        push (@check_fields, 'sigs');
    }
    my $has_changed = $current_sig_set->has_value_changed(existing_obj => $current_sig_set, store_obj => $new_sig_set,
                                                          check_fields => \@check_fields);

    return !$has_changed;
}

#----------------------------------------------------------------------

sub export_policy {
    my($self, %params) = @_;
    my $dbh = $self->{dbh};
    my $policy_id = $params{policy_id};
    my $minimal_export = $params{minimal_export};
    my $export_suggestions = $params{export_suggestions};

    $self->{policy_id} = $policy_id;
    $self->{policy} = $params{policy};
    $self->{asm_config_handler} = $params{asm_config_handler};

    my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                    get_criteria => { policy_id => $policy_id },);

    if ($policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        return $self->fatal_error(msg => "Parent Policies cannot be exported in JSON format.");
    }

    $self->{policy_name} = $policy->{policy_name};

    my $template_name;
    my $original_template = exists $policy->{policy_template} ? $policy->{policy_template}->{icontrol_name} : undef;
    if ($minimal_export) {
        $template_name = defined $params{policy_template} ? $params{policy_template}->{icontrol_name} : $original_template;
    }
    else {
        $template_name = $FULL_EXPORT_DEFAULT_TEMPLATE;
    }

    $policy = $policy->to_rest();
    delete $policy->{templateReference};
    $policy->{softwareVersion} = F5::BigipVersionUtils::bigip_version();
    F5::YAML::Schema::Utils::item_fixup(item => $policy,
                                        entity_type => 'F5::ASMConfig::Entity::Policy',
                                        direction => 'to_declarative',
                                       );
    if ($template_name) {
        $policy->{template}->{name} = $template_name;
    }

    if ($minimal_export && !$template_name) {
        $self->add_warning(msg => "Policy has no associated Policy Template -- Full export initiated. In order to accomplish minimal export, please specify Policy Template");
        $minimal_export = 0;
    }

    my $minimal_export_items = {};
    if ($minimal_export) {

        my $description_query = 'SELECT description, template_type FROM PLC.POLICY_TEMPLATES WHERE icontrol_name = ?';
        my ($template_description, $type) = $self->{dbh}->selectrow_array($description_query, undef, $template_name);

        if ($original_template && ($template_name ne $original_template)) {
            my ($original_template_description) = $self->{dbh}->selectrow_array($description_query, undef, $original_template);
            if (defined $policy->{description} && ($original_template_description eq $policy->{description})) {
                $policy->{description} = $template_description;
            }
        }

        if ($type == F5::ASMConfig::TemplateType::TEMPLATE_TYPE_USER_DEFINED()) {
            $self->add_warning(msg => "Selected Policy Template is a custom template and may not exist on other machines.");
        }

        eval {
            $minimal_export_items = $self->get_minimal_export_items(%params,
                                                                    template_name => $template_name);
        };
        if (my $err = $@) {
            my $error_message = ref $@ ? $@->{error_message} : $@;
            return $self->fatal_error(msg => $error_message);
        }
    }

    $self->export_policy_references(%params,
                                    minimal_export => $minimal_export,
                                    minimal_export_items => $minimal_export_items,
                                    policy_pointer => $policy);

    if ($self->{enforce_strictness}) {
        $self->{overlay}->apply_overlay(declarative_policy => $policy,
                                        is_in_declarative_export => 1,);
        if (!$minimal_export) {
            #In full export we do not want to show warnings for all the elements removed by the overlay
            $self->{overlay}->clear_msgs();
        }
    }

    my $result_policy = JSON->new->ascii->canonical->pretty->encode({policy => $policy});
    my $modifications_section;
    if ($export_suggestions) {
        my $export_rv = F5::ASMConfig::Entity::Suggestion->export_declarative(dbh => $self->{dbh},
                                                                              policy => $self->{policy},
                                                                             );
        if ($export_rv->{warnings}) {
           for my $msg (${$export_rv->{warnings}}) {
               $self->add_warning(msg => $msg);
           }
        }

        my $suggestions = $export_rv->{items};
        if (@$suggestions) {
            # modify result json -- to keep modifications section after policy section for readability
            substr($result_policy, -3) = "";
            $modifications_section = JSON->new->ascii->canonical->pretty->encode({modifications => $suggestions});
            substr $modifications_section, 0, 1, "";
            $result_policy = "$result_policy,$modifications_section";
        }
    }

    File::Slurp::write_file($params{filename}, $result_policy);

    return $self->completed_successfully(msg => "Policy '$self->{policy_name}' was successfully exported.");
}

#----------------------------------------------------------------------

my %minimal_export_excluded = (
                                'F5::ASMConfig::Entity::IdlFile' => 1,
                              );

sub export_policy_references {
    my($self, %params) = @_;
    my $path_pointer = $params{path_pointer} || $F5::ASMConfig::Handler::Rest::path_mapper->{references}->{policies}->{references};
    my $policy_pointer = $params{policy_pointer};

    for my $uri (sort keys %$path_pointer) {
        my $reference = $path_pointer->{$uri};
        my $is_collection = !$reference->{singleton};
        my $path = ($params{path} || '') . "/$uri";
        if (my $entity_type = $reference->{entity_type}) {
            if ( !F5::YAML::Schema::Utils::is_supported_entity(entity_type => $entity_type)
                   || F5::YAML::Schema::Utils::is_read_only_entity(entity_type => $entity_type)) {
                next;
            }

            my $minimal_export = $minimal_export_excluded{$entity_type} ? 0 : $params{minimal_export};
            next if $minimal_export && !defined $params{minimal_export_items}->{$entity_type};
            my $export_rv = $entity_type->export_declarative(dbh => $self->{dbh},
                                                           logger => $self->{logger},
                                                           policy => $self->{policy},
                                                           minimal_export => $params{minimal_export},
                                                           diff_items => $params{minimal_export} ? $params{minimal_export_items}->{$entity_type} : undef, );
            if ($export_rv->{warnings}) {
                for my $msg (@{$export_rv->{warnings}}) {
                    $self->add_warning(msg => $msg);
                }
            }

            my @items = @{$export_rv->{items}};
            next if !@items;
            if ($is_collection && !%{$items[0]}) {
                next;
            }

            $policy_pointer->{$uri} = $is_collection? \@items : $items[0];
        }
        elsif ($reference->{references}) {
            $policy_pointer->{$uri} = {};
            $self->export_policy_references(%params,
                                            path => $path,
                                            path_pointer => $reference->{references},
                                            policy_pointer => $policy_pointer->{$uri});
            if ($params{minimal_export} && !%{$policy_pointer->{$uri}}) {
                delete $policy_pointer->{$uri};
            }
        }
        if ($policy_pointer->{$uri} && F5::ASMConfig::Handler::Rest::is_path_excluded(path_pointer => $reference,
                                                                                      dbh => $self->{dbh},
                                                                                      policy => $self->{policy}) ) {
            delete $policy_pointer->{$uri};
            if ($self->{enforce_strictness} && $params{minimal_export}) {
                $self->add_warning(msg => "Element '$path' is unsupported.");
            }
        }
    }
}

#----------------------------------------------------------------------

sub get_minimal_export_items {
    my($self, %params) = @_;
    my $template_name = $params{template_name};

    my $timestamp = time();
    my $template_policy = F5::ASMConfig::Policy->new(policy_name => "$timestamp",
                                                     encoding => $self->{policy}->{encoding},
                                                     is_active_policy => 0,
                                                     flg_encoding_selected => $self->{policy}->{flg_encoding_selected},
                                                     flg_is_case_insensitive => $self->{policy}->{flg_is_case_insensitive},
                                                     flg_is_protocol_free => $self->{policy}->{flg_is_protocol_free},
                                                     policy_template => {
                                                         icontrol_name => $template_name,
                                                     },
                                                    );

    my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1,
                                                   skip_mcp => 1,
                                                   is_in_declarative_export => 1,);

    $template_policy = $self->{asm_config_handler}->add_policy($self->{asm_config_handler}->{session},
                                                               $call_flags,
                                                               $template_policy,
                                                              );

    if (!F5::BigipUtils::is_feature_enabled('allow_staging')) {
        $F5::ASMConfig::Entity::Base::common_diff_excludes->{flg_learn} = 1;
        $F5::ASMConfig::Entity::Base::common_diff_excludes->{in_staging} = 1;
        $F5::ASMConfig::Entity::Base::common_diff_excludes->{flg_quarantined} = 1;
        $F5::ASMConfig::Entity::Base::common_diff_excludes->{flg_pb_enabled} = 1;
    }
    my $rv = $self->{asm_config_handler}->add_diff($self->{asm_config_handler}->{session},
                                                   $call_flags,
                                                   undef,
                                                   F5::ASMConfig::Diff->new(first_policy => $self->{policy},
                                                                            second_policy => $template_policy),
                                                  );
    my @diff_items = @{$rv->{diff_items}};
    my $minimal_export_items;
    for my $diff_item (@diff_items) {
        my $entity_type = F5::YAML::Schema::Utils::get_entity_type(entity_kind => $diff_item->{element_kind});
        if ($diff_item->{second_element_id}) {
            $diff_item->{second_element} = $entity_type->new(dbh => $self->{dbh},
                                                             policy=> $template_policy,
                                                             get_criteria => { $entity_type->unique_id_field => $diff_item->{second_element_id} });
        }
        push (@{$minimal_export_items->{$entity_type}}, $diff_item);
    }

    $self->{asm_config_handler}->delete_policy($self->{asm_config_handler}->{session},
                                               $call_flags,
                                               $template_policy,
                                              );
    return $minimal_export_items;
}

#----------------------------------------------------------------------
sub get_sig_id_from_name_and_tag {
    my ( $self, %params ) = @_;

    my $sig_name = $params{sig_name};
    my $sig_tag = $params{sig_tag} || '';

    return unless $sig_name;

    if(! defined $user_defined_sig_id_to_sig_name_and_tag) {
        $user_defined_sig_id_to_sig_name_and_tag = $self->{dbh}->selectall_hashref(<<EndQuery, ['sig_name', 'sig_tag'], undef, 1);
SELECT sig_id, sig_name, sig_tag
  FROM PLC.NEGSIG_SIGNATURES
  WHERE flg_is_user_defined = ?;
EndQuery
    }

    if ( $user_defined_sig_id_to_sig_name_and_tag &&
         $user_defined_sig_id_to_sig_name_and_tag->{$sig_name} &&
           $user_defined_sig_id_to_sig_name_and_tag->{$sig_name}->{$sig_tag} ){

       return $user_defined_sig_id_to_sig_name_and_tag->{$sig_name}->{$sig_tag}->{sig_id};
   }
   return;
}

#----------------------------------------------------------------------
sub check_policy_size {
    my $self = shift;

    return if $self->{skip_check_policy_size};

    if (!defined $self->{max_policy_size} ) {
        my $max_json_policy_size = $self->{asm_config_handler}->get_internal_config_items(config_item_name => 'max_json_policy_size');
        if ($max_json_policy_size->{current_value}) { # 0 = disabled
            $self->{max_policy_size} = $max_json_policy_size->{current_value} * 1024;
            $self->{error_message_policy_size} = "Policy size exceeded maximum allowed advanced system variable $max_json_policy_size->{config_item_name} limit ($max_json_policy_size->{current_value} $max_json_policy_size->{units})";
        }
        else {
            $self->{skip_check_policy_size} = 1;
            return;
        }
    }

    if ($self->{policy_size} && ($self->{policy_size} > $self->{max_policy_size})) {
        die ({error_message => $self->{error_message_policy_size}});
    }
}

#----------------------------------------------------------------------

sub add_external_reference {
    my($self, %params) = @_;
    my $file_info = $params{file_info};

    my %external_reference = %$file_info;
    # Remove full file contents if present
    delete $external_reference{data};

    push (@{$self->{external_references}}, \%external_reference);
}

#----------------------------------------------------------------------


1;
