package F5::ImportOpenApi;

use strict;
use warnings;

use JSON;
use File::Slurp;
use MIME::Base64;
use Encode;
eval {
    require Encode::HanExtra;
};
use F5::YAML::Schema::Utils;

my %apiTypeModuleMap = (
    F5::ASMConfig::OpenApiFileType::OPEN_API_FILE_TYPE_SWAGGER() => 'F5::ImportOpenApi::Swagger',
);

#-------------------------------------------------------------------------------

sub new {
    my ( $class, %params ) = @_;

    my $api_type = $params{api_type} || F5::ASMConfig::OpenApiFileType::OPEN_API_FILE_TYPE_SWAGGER();

    my $self = {};

    #TODO: handle wrong api_type

    eval "require $apiTypeModuleMap{ $api_type }";

    bless $self, $apiTypeModuleMap{ $api_type };
    $self->init(%params);
    return $self;
}

#-------------------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;

    $self->{dbh} = $params{dbh};
    $self->{username} = $params{username};
    $self->{encoding} = $params{encoding} || 'utf-8';
    $self->{logger} = $params{logger};
    $self->{session} = $params{session};
    $self->{call_flags} = $params{call_flags};
    $self->{handler} = $params{handler};
}

#-------------------------------------------------------------------------------

sub get_openapi_content {
    my ($self, %params) = @_;

    my $filename = $params{filename};
    my $api_content_format = $params{api_content_format};
    my $import_encoding = $params{encoding} || 'utf-8';
    my $is_base64 = $params{is_base64};

    my $encoding = $self->get_input_encoding(import_encoding => $import_encoding);

    my $api_content;
    eval {
        $api_content = read_file($filename, { binmode => ":encoding($encoding)" });
        if ($is_base64) {
            $api_content = MIME::Base64::decode_base64($api_content);
        }
    };
    if ($@) {
        die {error_message => "Failed to read swagger file content"};
    }

    if (! $api_content_format) {
        eval {
            if ($filename =~ /\.json/) {
                $api_content_format = F5::ASMConfig::OpenApiContentFormat::OPEN_API_CONTENT_FORMAT_JSON();
            }
            elsif ($filename =~ /\.yaml/) {
                $api_content_format = F5::ASMConfig::OpenApiContentFormat::OPEN_API_CONTENT_FORMAT_YAML();
            }

            if (! $api_content_format) {
                if ($api_content =~ /^\s*?[\[\{]/m) {
                    $api_content_format = F5::ASMConfig::OpenApiContentFormat::OPEN_API_CONTENT_FORMAT_JSON();
                }
                if ($api_content =~ /^\s*?\w+:\s/m) {
                    $api_content_format = F5::ASMConfig::OpenApiContentFormat::OPEN_API_CONTENT_FORMAT_YAML();
                }
            }
        };
        if ($@) {
            my $error_msg = $@;
            $error_msg =~ s|\s+?at \/.*$||;
            die {error_message => "Failed to detect swagger file format: $error_msg"};
        }
    }

    my $openapi_content;
    if ($api_content_format == F5::ASMConfig::OpenApiContentFormat::OPEN_API_CONTENT_FORMAT_YAML()) {
        eval {
            if (! $encoding || $encoding =~ /^utf.?8/i) {
                $encoding = 'utf8';
            }

            # yaml cont first converted into json content, this because YAML module convert integers to strings
            my $rv = F5::YAML::Schema::Utils::yaml_to_json(encoding => $encoding, filename => $filename,);
            if (! $rv->{completed_successfully}) {
                die $rv->{error_message};
            }

            $api_content = $rv->{data};
            $api_content_format = F5::ASMConfig::OpenApiContentFormat::OPEN_API_CONTENT_FORMAT_JSON();
        };
        if ($@) {
            # Clean up YAML exception for readability
            my $yaml_parse_error = $@;
            $yaml_parse_error =~ s/\s+?File\s[^\n]*?\n[^\n]*?\n//gm;
            die {error_message => "Failed to parse YAML Open API Content: $yaml_parse_error"};
        }
    }

    if ($api_content_format == F5::ASMConfig::OpenApiContentFormat::OPEN_API_CONTENT_FORMAT_JSON()) {
        eval {
            $openapi_content = JSON::XS->new->decode($api_content);
        };
        if ($@) {
            my $error_msg = $@;
            $error_msg =~ s/ at \/usr\/local.*$//;
            die {error_message => "Fail parse JSON Open API Content: $error_msg."};
        }
    }
    else {
        die {error_message => "Unknown Open API Content format."};
    }

    $self->{openapi_content} = $openapi_content;
}

#-------------------------------------------------------------------------------

sub log_error {
    my ($self, %params) = @_;

    my $error = $params{error};
    my $error_title = $params{error_title};

    my $error_message;
    if (ref($error)) {
        $error_message = $error->{error_message};
    }
    else {
        $error_message = $error;
    }

    $self->{logger}->error( join(': ', grep {$_} $error_title, $error_message));
}

#-------------------------------------------------------------------------------

sub get_input_encoding {
    my ($self, %params) = @_;

    my $import_encoding = $params{import_encoding};

    my $encoding_obj = find_encoding($import_encoding);
    if (! ref($encoding_obj)) {
        die "Encoding $import_encoding is not supported.";
    }

    my $encoding = $encoding_obj->mime_name();
    if (! $encoding) {
        $encoding = $import_encoding;
    }

    return $encoding;
}

#-------------------------------------------------------------------------------


1;

