package F5::ImportOpenApi::Swagger;

use strict;
use warnings;

use F5::Deploy;
use F5::ASMConfig::Entity::AllowedMethod;
use F5::ASMConfig::Entity::URL;
use F5::ASMConfig::Entity::Parameter;
use F5::ASMConfig::Entity::ResponsePage;
use F5::ASMConfig::Entity::JsonValidationFile;
use F5::ParseUri;
use F5::Logger;
use F5::ImportOpenApi;
use F5::ObjUtils qw(dclone);

use base qw/F5::ImportOpenApi/;


my %param_type_map = (
    integer => F5::ASMConfig::ParamDataType::PARAM_DATA_INTEGER(),
    number => F5::ASMConfig::ParamDataType::PARAM_DATA_DECIMAL(),
    decimal => F5::ASMConfig::ParamDataType::PARAM_DATA_DECIMAL(),
    string => F5::ASMConfig::ParamDataType::PARAM_DATA_ALPHA_NUMERIC(),
    boolean => F5::ASMConfig::ParamDataType::PARAM_DATA_BOOLEAN(),
    file => F5::ASMConfig::ParamDataType::PARAM_DATA_BINARY(),
);

my %param_location_map = (
    any => F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_ANY(),
    query => F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_QUERY(),
    formData => F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_FORM_DATA(),
    path => F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_PATH(),
);

my %param_serialization_map = (
    csv => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_CSV(),
    ssv => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_SSV(),
    tsv => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_TSV(),
    pipes => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_PIPE(),
    multi => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MULTI(),
    multipart => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MULTIPART(),
);

my %serialization_style_map = (
    simple => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_SIMPLE(),
    label => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_LABEL(),
    matrix => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_MATRIX(),
    form => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_FORM(),
    deepObject => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_DEEP_OBJECT(),
    spaceDelimited => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_SPACE_DELIMITED(),
    pipeDelimited => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_PIPE_DELIMITED(),
);

my %default_serialization_style = (
    path => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_SIMPLE(),
    query => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_FORM(),
    header => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_SIMPLE(),
    cookie => F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_FORM(),
);

my %serialization_array_style_map = (
    simple => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_CSV(),
    form => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_FORM(),
    label => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_LABEL(),
    matrix => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MATRIX(),
    spaceDelimited => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_SSV(),
    pipeDelimited => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_PIPE(),
);

my %default_array_serialization_style = (
    path => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_CSV(),
    query => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MULTI(),
    header => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_CSV(),
    cookie => F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MULTI(),
);

my %default_explode = (
    path => 0,
    query => 1,
    header => 0,
    cookie => 1,
);

my %default_open_api_array_style = (
    path => 'simple',
    query => 'form',
    header => 'simple',
    cookie => 'form',
);

my $DEFAULT_RESPONSES_FILE = F5::Deploy::map_path('/etc/ts/tools/open_api_default_responses.yaml');
my $MIN_SWAGGER_VERSION = 2;
my $MAX_SWAGGER_VERSION = 3;
my $DEFAULT_JSON_SCHEMA = 'http://json-schema.org/draft-04/schema#';
my $JSON_SCHEMA_ENTRY_NAME = '$schema';
my $ARRAY_ANY_TYPE_JSON_SCHEMA = '{"type":"array","'.$JSON_SCHEMA_ENTRY_NAME.'":"'.$DEFAULT_JSON_SCHEMA.'"}';


#-------------------------------------------------------------------------------

sub process_openapi_content {
    my ($self, %params) = @_;

    $self->{policy} = $params{policy};
    $self->{open_api_profile_name} = $params{open_api_profile_name};

    my $version_error = $self->verify_swagger_version();
    if ($version_error) {
        die "$version_error\n";
    }

    $self->{response_codes} = {};
    $self->{produces} = {};
    $self->{processed_url_ids} = {};

    $self->{ref_paths_seen} = {};
    $self->replace_refs();
    $self->check_produces(process_elem => $self->{openapi_content}, scope => 'global',);
    $self->create_urls();
    $self->remove_old_urls();
    $self->remove_unrelated_path_params();
    # swagger never generates simple global paramters.
    # any global paramters that are not path paramters are old ones and need to be removed.
    # (the old path paramters where already removed in remove_unrelated_path_params)
    $self->remove_unrelated_global_params();
    $self->remove_unrelated_json_profiles();
    $self->create_default_response();
    $self->create_response_codes();
    $self->update_path_parameter_handling_to_url();
}

#-------------------------------------------------------------------------------

sub create_urls {
    my ($self, %params) = @_;

    my $openapi_content = $self->{openapi_content};

    my $base_path_list = [];
    if ($self->support_version(3)) {
        $base_path_list = $self->servers_base_path(openapi_content => $openapi_content,);
    }
    else {
        push @$base_path_list, $openapi_content->{basePath};
    }

    my $paths = $openapi_content->{paths};
    return unless ref($paths);

    for my $base_path (@$base_path_list) {
        for my $relative_path (keys %$paths) {
            $self->create_url(base_path => $base_path, relative_path => $relative_path, path_data => dclone($paths->{$relative_path}),);
        }
    }
}

#-------------------------------------------------------------------------------

sub servers_base_path {
    my ($self, %params) = @_;

    my $openapi_content = $params{openapi_content};

    my @all_urls;
    my $servers = $openapi_content->{servers};
    if (ref($servers) eq 'ARRAY') {
        for my $server (@$servers) {
            my $urls_list = $self->server_template(server => $server,);
            if (ref($urls_list)) {
                push @all_urls, @$urls_list;
            }
        }
    }

    my $base_path_list = $self->base_urls(all_urls => \@all_urls,);
    if (! scalar @$base_path_list) {
        push @$base_path_list, '/';
    }

    return $base_path_list;
}

#-------------------------------------------------------------------------------

sub base_urls {
    my ($self, %params) = @_;

    my $all_urls = $params{all_urls};

    my %base_path_list;
    for my $url (@$all_urls) {
        next if $url =~ m|^wss?://|;
        my $temp_url = $url;
        $temp_url =~ s|^https?://[^/]*/?|/|;
        $base_path_list{$temp_url} = 1;
    }

    return [keys %base_path_list];
}

#-------------------------------------------------------------------------------

sub server_template {
    my ($self, %params) = @_;

    my $server = $params{server};

    my $url_name = $server->{url};
    return unless $url_name;

    my @subst_vars = $url_name =~ /\{(.+?)\}/g;
    my $url_variables = $server->{variables};
    my $urls_list = [];
    $self->replace_variables(url_name => $url_name, subst_vars => \@subst_vars, urls_list => $urls_list, url_variables => $url_variables,);

    return $urls_list;
}

#-------------------------------------------------------------------------------

sub replace_variables {
    my ($self, %params) = @_;

    my $url_name = $params{url_name};
    my $subst_vars = dclone($params{subst_vars});
    my $urls_list = $params{urls_list};
    my $url_variables = $params{url_variables};

    if (@$subst_vars > 0) {
        my $var_name = shift @$subst_vars;
        my $var_defs = $url_variables->{$var_name};
        return unless $var_defs;

        my %var_values;
        if (ref($var_defs->{enum})) {
            map {$var_values{$_} = 1} @{$var_defs->{enum}};
        }
        else {
            $var_values{'*'} = 1;
        }

        for my $val (keys %var_values) {
            my $temp_url_name = $url_name;
            $temp_url_name =~ s/\{$var_name\}/$val/;
            $self->replace_variables(url_name => $temp_url_name, subst_vars => $subst_vars, urls_list => $urls_list, url_variables => $url_variables,);
        }
    }
    else {
        push @$urls_list, $url_name;
    }

}

#-------------------------------------------------------------------------------

sub normalize_path {
    my ($self, %params) = @_;

    my $path = $params{path};

    $path =~ s|\/\/+|/|g;

    return $path;
}

#-------------------------------------------------------------------------------

sub remove_old_urls {
    my ($self) = @_;

    my $iterator = F5::ASMConfig::Entity::URL->retrieve_all( dbh => $self->{dbh},
                                                             policy => $self->{policy},
                                                             minimal_load => 1, );
    while (my $url = $iterator->get_next) {
        if ($self->{open_api_profile_name} && $url->{url_name} eq '*') {
            next;
        }
        if (! $self->{processed_url_ids}->{$url->{url_id}}) {
            eval {
                $self->{handler}->delete_url($self->{session}, $self->{call_flags}, $self->{policy}, $url,);
            };
            if ($@) {
                $self->log_error(error_title => "Fail to delete removed swagger URL: ". $url->get_full_display_name, error => $@);
            }
        }
    }
}

#-------------------------------------------------------------------------------

sub remove_unrelated_json_profiles {
    my ($self) = @_;

    eval {
        my $search = F5::ASMConfig::JsonProfile->new(is_referenced => 0);
        $self->{handler}->delete_many_json_profiles($self->{session}, $self->{call_flags}, $self->{policy}, $search,);
    };
    if ($@) {
        $self->log_error(error_title => "Fail to delete unrelated JSON profiles: ", error => $@);
    }

    eval {
        my $search = F5::ASMConfig::JsonValidationFile->new(is_referenced => 0);
        $self->{handler}->delete_many_json_validation_files($self->{session}, $self->{call_flags}, $self->{policy}, $search,);
    };
    if ($@) {
        $self->log_error(error_title => "Fail to delete unrelated JSON Validation files: ", error => $@);
    }

}

#-------------------------------------------------------------------------------

sub remove_unrelated_path_params {
    my ($self) = @_;

    my $param_ids = $self->{dbh}->selectcol_arrayref(<<EndQuery, undef, $self->{policy}->{policy_id}, F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_PATH, $self->{policy}->{policy_id},);
SELECT id
  FROM PLC.PL_PARAM_ATTRIBUTES
 WHERE policy_id = ?
   and parameter_location = ?
   and id not in
   (SELECT param_attrib_id
      FROM PLC.PL_OBJECT_POSITIONAL_PARAMETERS
     WHERE policy_id = ?)
EndQuery

    if (! scalar @$param_ids) {
        return;
    }

    eval {
        my $local_call_flags = F5::ASMConfig::CallFlags->new($self->{call_flags});
        $local_call_flags->{search_filters} = [
                F5::ASMConfig::SearchFilter->new({ field_name => 'param_id',
                                                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                   field_value_list => $param_ids }),
        ];
        $self->{handler}->delete_many_parameters($self->{session}, $local_call_flags, $self->{policy},);
    };
    if ($@) {
        $self->log_error(error_title => "Fail to delete unrelated path parameters", error => $@);
    }
}

#-------------------------------------------------------------------------------

sub remove_unrelated_global_params {
   my ($self) = @_;

   eval {
	  my $local_call_flags = F5::ASMConfig::CallFlags->new($self->{call_flags});
	  $local_call_flags->{search_filters} = [
              F5::ASMConfig::SearchFilter->new({ field_name => 'param_level',
                                                 field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                 field_value => F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL }),
              F5::ASMConfig::SearchFilter->new({ field_name => 'is_referenced',
                                                 field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                 field_value => 0 }),
              F5::ASMConfig::SearchFilter->new({ field_name => 'param_name',
                                                 field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                                                 field_value => '*' }),

	  ];
	  $self->{handler}->delete_many_parameters($self->{session}, $local_call_flags, $self->{policy},);
   };
   if ($@) {
       $self->log_error(error_title => "Fail to delete unrelated global parameters", error => $@);
   }
}

#-------------------------------------------------------------------------------

sub remove_old_params {
    my ($self, %params) = @_;

    my $url = $params{url};
    my $processed_param_ids = $params{processed_param_ids};

    eval {
        my $local_call_flags = F5::ASMConfig::CallFlags->new($self->{call_flags});
        if (scalar keys %$processed_param_ids) {
            $local_call_flags->{search_filters} = [
                F5::ASMConfig::SearchFilter->new({ field_name => 'param_id',
                                                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN(),
                                                   field_value_list => [keys %$processed_param_ids] }),
        ];
        }
        $self->{handler}->delete_many_parameters($self->{session}, $local_call_flags, $self->{policy}, {url => $url},);
    };
    if ($@) {
        $self->log_error(error_title => "Fail to delete removed swagger Parameters for URL: ". $url->get_full_display_name, error => $@);
    }
}

#-------------------------------------------------------------------------------

sub create_url {
    my ($self, %params) = @_;

    my $base_path = $params{base_path} || '';
    my $relative_path = $params{relative_path};
    my $path_data = $params{path_data};

    my $all_methods_params = $self->get_path_attribute(path_data => $path_data, attribute_name => 'parameters');
    my $description = $self->get_path_attribute(path_data => $path_data, attribute_name => 'description');
    my $summary = $self->get_path_attribute(path_data => $path_data, attribute_name => 'summary');
    my $url_desc = join("\n", grep {$_} $summary, $description);

    my $is_url_wildcard = 0;
    if ($base_path =~ /\*/) {
        $is_url_wildcard = 1;
    }
    my $absolute_path = $base_path. $relative_path;
    $absolute_path = $self->normalize_path(path => $absolute_path,);
    for my $method_name (keys %$path_data) {
        my $method_data = $path_data->{$method_name};

        next unless ref($method_data) eq 'HASH';

        $self->create_url_with_method(absolute_path => $absolute_path,
                                      method_name => uc($method_name),
                                      description => $url_desc,
                                      method_data => $method_data,
                                      all_methods_params => $all_methods_params,
                                      is_url_wildcard => $is_url_wildcard,);
    }
}

#-------------------------------------------------------------------------------

sub get_path_attribute {
    my ($self, %params) = @_;

    my $attribute_name = $params{attribute_name};
    my $path_data = $params{path_data};

    my $attribute = $path_data->{$attribute_name};
    if ($attribute){
        delete($path_data->{$attribute_name});
        if (! ref($attribute)) {
            $attribute = $self->check_convert_encoding(content_value => $attribute);
        }
    }

    return $attribute;
}

#-------------------------------------------------------------------------------

sub create_url_with_method {
    my ($self, %params) = @_;

    my $absolute_path = $params{absolute_path};
    my $method_name = $params{method_name};
    my $method_data = $params{method_data};
    my $description = $params{description};
    my $all_methods_params = $params{all_methods_params} || [];
    my $method_params = $method_data->{parameters} || [];
    my $is_url_wildcard = $params{is_url_wildcard};

    $self->check_produces(process_elem => $method_data, scope => 'url');

    my $method = F5::ASMConfig::Entity::AllowedMethod->get_method(name => $method_name);
    return unless $method;

    my %uniq_parameters;
    for my $param (@$all_methods_params, @$method_params) {
        if ($param->{name}) {
            $uniq_parameters{$param->{name}} = dclone($param);
        }
    }

    my $url_object = {
        method_name => $method_name,
        url_name => $absolute_path,
        app_id => F5::ParseUri::APP_ID_HTTP(),
        flg_check_signatures => 1,
    };

    if ($is_url_wildcard) {
        $url_object->{is_wildcard} = 1;
        $url_object->{flg_wildcard_includes_slash} = 1;
    }

    $description = join("\n", grep {$_} $description, $method_data->{summary}, $method_data->{description});
    $url_object->{description} = $description if $description;

    if ($method_data->{operationId}) {
        $url_object->{operation_id} = $method_data->{operationId};
    }

    my ($path_params);
    if ($absolute_path =~ m|\{.+?\}|) {
        my $res = $self->parse_path_parameters(%params);
        $absolute_path = $res->{url_name};
        $path_params = $res->{path_params};
        $url_object->{is_wildcard} = 1;
        $url_object->{url_name} = $absolute_path;
    }

    $url_object->{url_name} = $self->check_convert_encoding(content_value => $url_object->{url_name});

    $url_object = F5::ASMConfig::Entity::URL->new(policy => $self->{policy},
                                                  dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  created_by => $self->{username},
                                                  preload_default => 1,
                                                  object => $url_object,);

    if ($self->support_version(3)) {
        eval {
            $self->handle_request_body(method_data => $method_data,
                                       url_object => $url_object,
                                       uniq_parameters => \%uniq_parameters,
                                       method_name => $method_name,);
        };
        if ($@) {
            $self->log_error(error_title => "Fail handle requestBody on swagger URL: $method_name, $absolute_path", error => $@);
        }
    }
    else {
        eval {
            $self->handle_consumes(method_data => $method_data,
                                   url_object => $url_object,
                                   parameters => \%uniq_parameters,
                                   method_name => $method_name,);
        };
        if ($@) {
            $self->log_error(error_title => "Fail handle consumes on swagger URL: $method_name, $absolute_path", error => $@);
        }
    }

    my $url = F5::ASMConfig::Entity::URL->new(dbh => $self->{dbh},
                                              policy => $self->{policy},
                                              logger => $self->{logger},
                                              get_criteria => $url_object,);
    if ($url && $path_params) {
        if ($self->path_param_name_changed(url => $url, path_params => $path_params,)) {
            eval {
                $self->{handler}->delete_url($self->{session}, $self->{call_flags}, $self->{policy}, $url,);
                $url = undef;
            };
            if ($@) {
                $self->log_error(error_title => "Fail to delete swagger URL with changed path parameter: ". $url->get_full_display_name, error => $@);
            }
        }
    }

    my $action_name = $url ? 'Update' : 'Create';
    $self->{logger}->debug($action_name. " swagger URL: $method_name, $absolute_path") if $self->{logger};
    eval {
        if ($url) {
            my $url_entity = $self->{handler}->update_url($self->{session}, $self->{call_flags}, $self->{policy}, $url, $url_object);
            $self->{processed_url_ids}->{$url_entity->{url_id}} = 1;

            my $url_level_params = \%uniq_parameters;
            if ($path_params) {

                #TODO: handle case when path parameter name changed in swagger file

                $url_level_params = $self->create_parameters(url => $url, parameters => \%uniq_parameters, path_params => $path_params);
            }

            my $processed_param_ids = {};
            $self->create_parameters(url => $url, parameters => $url_level_params, processed_param_ids => $processed_param_ids,);
            $self->remove_old_params(url => $url, processed_param_ids => $processed_param_ids,);
        }
        else {
            my $call_flags = $self->{call_flags} || {};
            $call_flags = dclone($call_flags);
            $call_flags->{internal_flags_hash} ||= {};
            $call_flags->{internal_flags_hash}->{skip_populate_defaults_on_add} = 1;

            # posisitional params are read only, "add_url" postponed
            my $url_level_params = \%uniq_parameters;
            if ($path_params) {
                $url_level_params = $self->create_parameters(url => $url_object, parameters => \%uniq_parameters, path_params => $path_params);
            }

            my $url_entity = $self->{handler}->add_url($self->{session}, $call_flags, $self->{policy}, $url_object);
            $self->{processed_url_ids}->{$url_entity->{url_id}} = 1;
            $self->create_parameters(url => $url_entity, parameters => $url_level_params);
        }
    };
    if ($@) {
        $self->log_error(error_title => "Fail ". $action_name. " swagger URL: $method_name, $absolute_path", error => $@);
    }

    $self->check_response_codes(responses => $method_data->{responses});
}

#-------------------------------------------------------------------------------

sub path_param_name_changed {
    my ($self, %params) = @_;

    my $url = $params{url};
    my $path_params = $params{path_params};

    my %params_indexes;
    for my $param_name (keys %$path_params) {
        $params_indexes{ $path_params->{$param_name}->{url_segment_index} } = $path_params->{$param_name}->{unique_param_name};
    }

    for my $exist_path_param (@{$url->{positional_parameters}}) {
        if ($params_indexes{ $exist_path_param->{url_segment_index} } ne $exist_path_param->{param_name}) {
            return 1;
        }
    }

    return;
}

#-------------------------------------------------------------------------------

sub handle_request_body {
    my ($self, %params) = @_;

    my $method_data = $params{method_data};
    my $url_object = $params{url_object};
    my $uniq_parameters = $params{uniq_parameters};
    my $method_name = $params{method_name};

    $self->set_default_header_profile(url_object => $url_object,);

    return if ! $self->is_method_with_body(method_name => $method_name,);

    # note - got here only if the method allows body.
    # (so ok to set flg_mandatory_body based on 'required' field)
    my $content;
    if (ref($method_data->{requestBody}) eq 'HASH') {
        if (defined($method_data->{requestBody}->{required})) {
            $url_object->{flg_mandatory_body} = $method_data->{requestBody}->{required} ? 1 : 0;
        }
        if (ref($method_data->{requestBody}->{content}) eq 'HASH') {
            $content = $method_data->{requestBody}->{content};
        }
    }

    if (ref($self->{openapi_content}->{requestBody}) eq 'HASH') {
        if (!defined($url_object->{flg_mandatory_body}) && defined($self->{openapi_content}->{requestBody}->{required})) {
            $url_object->{flg_mandatory_body} = $self->{openapi_content}->{requestBody}->{required} ? 1 : 0;
        }
        if (!defined($content) && ref($self->{openapi_content}->{requestBody}->{content}) eq 'HASH') {
            $content = $self->{openapi_content}->{requestBody}->{content};
        }
    }

    return if ! $content;

    my @profiles_list;
    my $content_type_wc_str;
    for my $content_type (keys %$content) {
        my ($content_profile_type, $enforcement_type, $content_profile_entity, $header_priority,);
        if ($content_type =~ /\bjson\b/i) {
            $content_profile_type = F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON();
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON();
            if (ref($content->{$content_type}) eq 'HASH' && ref($content->{$content_type}->{schema}) eq 'HASH') {
                my $json_schema_struct = $content->{$content_type}->{schema};
                if ($self->schema_is_type_object(schema => $json_schema_struct,) || $self->schema_is_type_array(schema => $json_schema_struct,)) {
                    $content_profile_entity = $self->json_schema_profile(json_schema_struct => $json_schema_struct,
                                                                         url_name => $url_object->{url_name},
                                                                         method_name => $url_object->{method_name},);
                }
            }
            $header_priority = 3;
        }
        elsif ($content_type =~ /\bxml\b/i) {
            $content_profile_type = F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML();
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_XML();
            $header_priority = 4;
        }
        elsif ($content_type =~ m|multipart\/.*?form\b| || $content_type eq 'multipart/mixed') {
            $content_type_wc_str = $content_type . ';*boundary*';
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_HTTP();
            $header_priority = 1;

            $self->multipart_form_parameters(schema => $content->{$content_type}->{schema},
                                             encoding => $content->{$content_type}->{encoding},
                                             uniq_parameters => $uniq_parameters,);
            # in addition for URL that accepts multipart need to add hard coded an alphanum param called 'filename'.
            # bd uses it to enforce the name of each attached file
            # allow repeated occurences - set to 1 to handle cases of requests with multiple file uploads
            # (note that allow-repeated-occurences is not an attribute of swagger param.
            # we add this dummy field to indicate allow-repeated-occurences = 1 for 'filename' paramter)
            my $filename_param_name = 'filename';
            my $filename_param = {name => $filename_param_name,
                                  in => 'formData',
                                  flg_allow_repeated_parameter_name => 1,
                                  allowEmptyValue => 1};
            $uniq_parameters->{$filename_param_name} = $filename_param;
        }
        elsif ($content_type =~ m|\/.*?\bform\b|) {
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_HTTP();
            $header_priority = 2;

            if (ref($content->{$content_type}) eq 'HASH' && ref($content->{$content_type}->{schema}) eq 'HASH') {
                if ($self->schema_is_type_object(schema => $content->{$content_type}->{schema},)) {
                    $self->form_content_parameters(uniq_parameters => $uniq_parameters, form_content => $content->{$content_type},);
                }
            }
        }
        elsif ($content_type =~ m|^image\/| || $content_type =~ m|^video\/| || $content_type =~ m|^audio\/|) {
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_DO_NOTHING();
            $url_object->{flg_disallow_file_upload_of_executables} = 1;
        }
        else {
            if (ref($content->{$content_type}) eq 'HASH' && ref($content->{$content_type}->{schema}) eq 'HASH') {
                my $type = $content->{$content_type}->{schema}->{type};
                my $format = $content->{$content_type}->{schema}->{format};
                if ($self->is_param_binary(type => $type, format => $format,)) {
                    $url_object->{flg_disallow_file_upload_of_executables} = 1;
                }
            }

            if (! $enforcement_type) {
                $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_APPLY_VALUE_SIGNATURES();
            }
        }

        if ($enforcement_type) {
            push @profiles_list, {
                content_profile_type => $content_profile_type,
                enforcement_type => $enforcement_type,
                content_profile_entity => $content_profile_entity,
                header_priority => $header_priority,
                content_type =>$content_type_wc_str ? $content_type_wc_str : $content_type,
            };
        }
    }

    my @ordered_profile_list = sort {$a->{header_priority} <=> $b->{header_priority}} @profiles_list;

    my $header_order = 0;
    for my $profile (@ordered_profile_list) {
        ++$header_order;
        my $content_profile_obj = {
                                    enforcement_type => $profile->{enforcement_type},
                                    header_name => 'Content-Type',
                                    header_value => $profile->{content_type},
                                    header_order => $header_order,
                                  };

        my $content_profile_type = $profile->{content_profile_type};
        if ($content_profile_type) {
            my $content_profile_entity = $profile->{content_profile_entity};
            if ($content_profile_entity) {
                $content_profile_obj->{content_profile} = $content_profile_entity;
            }
            else {
                $content_profile_obj->{content_profile} = $self->default_content_profile(content_profile_type => $content_profile_type,);
            }
        }

        my $url_content_profile = F5::ASMConfig::Entity::UrlContentProfile->new(dbh => $self->{dbh},
                                                                                policy => $self->{policy},
                                                                                logger => $self->{logger},
                                                                                created_by => $self->{username},
                                                                                object => $content_profile_obj,);

        push @{$url_object->{url_content_profiles}}, $url_content_profile;
        $url_object->{url_content_profiles}->[0]->{enforcement_type} = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_DISALLOW();
    }

}

#-------------------------------------------------------------------------------

sub multipart_form_parameters {
    my ($self, %params) = @_;

    my $schema = $params{schema};
    my $uniq_parameters = $params{uniq_parameters};
    my $encoding = $params{encoding};

    if (ref($schema->{properties}) eq 'HASH') {
        my $properties = $schema->{properties};
        for my $param_name (keys %$properties) {
            # do not override existing parameter
            next if $uniq_parameters->{$param_name};

            my $swagger_param = {name => $param_name, in => 'formData'};
            if (ref($properties->{$param_name}) eq 'HASH') {
                $swagger_param->{type} = $properties->{$param_name}->{type};
                $swagger_param->{format} = $properties->{$param_name}->{format};
                if ($self->is_param_binary(type => $swagger_param->{type}, format => $swagger_param->{format},)) {
                    $swagger_param->{type} = 'file';
                }
                elsif ($swagger_param->{type} eq 'object' && $properties->{$param_name}->{properties}) {
                    $swagger_param->{content}->{'application/json'}->{schema} = $properties->{$param_name};
                }
                elsif ($swagger_param->{type} eq 'array') {
                    $swagger_param->{collectionFormat} = 'multipart';
                    $swagger_param->{items} = $properties->{$param_name}->{items} if $properties->{$param_name}->{items};
                }
                else {
                    $swagger_param = $properties->{$param_name};
                }

                if (defined($properties->{$param_name}->{required})) {
                    $swagger_param->{required} = $properties->{$param_name}->{required};
                }
            }

            if (ref($encoding) && $swagger_param->{type} ne 'file') {
                if (ref($encoding->{$param_name}) && $encoding->{$param_name}->{contentType}) {
                    my $content_type = $encoding->{$param_name}->{contentType};
                    if ($content_type =~ /\bjson\b/i) {
                        $swagger_param->{content} = {};
                        $swagger_param->{content}->{$content_type}->{schema} = $properties->{$param_name};
                    }
                    elsif ($content_type =~ /\bxml\b/i) {
                        $swagger_param->{content} = {};
                        $swagger_param->{content}->{$content_type} = {};
                    }
                }
            }

            $swagger_param->{type} ||= 'string';

            $uniq_parameters->{$param_name} = $swagger_param;
        }
    }
}

#-------------------------------------------------------------------------------

sub form_content_parameters {
    my ($self, %params) = @_;

    my $uniq_parameters = $params{uniq_parameters};
    my $form_content = $params{form_content};

    my $schema_struct = $form_content->{schema};
    my $encoding = $form_content->{encoding};
    if (!$encoding && $schema_struct->{encoding}) {
        $encoding = $schema_struct->{encoding};
    }

    my $required_properties = {};
    if (ref($schema_struct->{required}) eq 'ARRAY') {
        for my $property_name (@{$schema_struct->{required}}) {
            $required_properties->{$property_name} = 1;
        }
    }

    if (ref($schema_struct->{properties}) eq 'HASH') {
        my $properties = $schema_struct->{properties};
        for my $property_name (keys %$properties) {
            # do not override existing parameter
            next if $uniq_parameters->{$property_name};

            my $form_param = {};
            if (ref($properties->{$property_name}) eq 'HASH') {
                my $property_data = dclone($properties->{$property_name});
                if ($property_data->{type} eq 'object' || $property_data->{type} eq 'array') {
                    $form_param->{schema} = $property_data;
                }
                else {
                    $form_param = $property_data;
                }
            }
            $form_param->{name} = $property_name;
            $form_param->{in} = 'formData';
            $form_param->{required} = 1 if $required_properties->{$property_name};

            if (ref($encoding) eq 'HASH' && ref($encoding->{$property_name}) eq 'HASH') {
                for my $encoding_entry (keys %{$encoding->{$property_name}}) {
                    my $encoding_value = $encoding->{$property_name}->{$encoding_entry};
                    $encoding_value = dclone($encoding_value) if (ref($encoding_value));
                    $form_param->{$encoding_entry} = $encoding_value;
                }
            }

            $uniq_parameters->{$property_name} = $form_param;
        }
    }
}

#-------------------------------------------------------------------------------

sub is_method_with_body {
    my ($self, %params) = @_;

    my $method_name = uc($params{method_name});

    if ($method_name eq 'POST' || $method_name eq 'PUT' || $method_name eq 'PATCH' ) {
        return 1;
    }
    return;
}

#-------------------------------------------------------------------------------

sub is_param_binary {
    my ($self, %params) = @_;

    my $type = $params{type};
    my $format = $params{format};

    if ($type eq 'string' && $format && ($format eq 'binary' || $format eq 'base64')) {
        return 1;
    }

    return 0;
}

#-------------------------------------------------------------------------------

sub set_default_header_profile {
    my ($self, %params) = @_;

    my $url_object = $params{url_object};

    $url_object->{url_content_profiles} = [];
    my $default_header_profile = {
        header_name => '*',
        header_value => '*',
        header_order => 0,
        enforcement_type => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_APPLY_VALUE_SIGNATURES(),
    };
    push @{$url_object->{url_content_profiles}}, $default_header_profile;
}

#-------------------------------------------------------------------------------

sub default_content_profile {
    my ($self, %params) = @_;

    my $content_profile_type = $params{content_profile_type};

    my $default_profile = F5::ASMConfig::Entity::ContentProfile->get_default_content_profile(
                                                                   dbh => $self->{dbh},
                                                                   policy => $self->{policy},
                                                                   content_profile_type => $content_profile_type,
                                                             );
    if ($default_profile->{__new_entity}) {
        $default_profile->store_add();
    }

    return $default_profile;
}

#-------------------------------------------------------------------------------

sub handle_consumes {
    my ($self, %params) = @_;

    my $method_data = $params{method_data};
    my $url_object = $params{url_object};
    my $parameters = $params{parameters};
    my $method_name = $params{method_name};

    $self->set_default_header_profile(url_object => $url_object,);

    return if ! $self->is_method_with_body(method_name => $method_name,);

    my $consumes = [];
    if ($method_data->{consumes} && ref($method_data->{consumes}) eq 'ARRAY') {
        $consumes = $method_data->{consumes};
    }
    elsif ($self->{openapi_content}->{consumes} && ref($self->{openapi_content}->{consumes}) eq 'ARRAY') {
        $consumes = $self->{openapi_content}->{consumes};
    }

    my $header_order = 1;
    for my $consume (@$consumes) {
        my ($content_profile_type, $enforcement_type, $header_value, $content_profile_entity,);
        if ($consume =~ /json$/ || $consume =~ /json;/) {
            $content_profile_type = F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON();
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON();
            $content_profile_entity = $self->check_body_schema_parameters(parameters => $parameters,
                                                                           method_name => $method_name,
                                                                           url_name => $url_object->{url_name},);
        }
        elsif ($consume =~ /xml$/ || $consume =~ /xml;/) {
            $content_profile_type = F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML();
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_XML();
        }
        elsif ($consume =~ m|multipart\/.*?form\b| || $consume eq 'multipart/mixed') {
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_HTTP();

            my $method_params = $method_data->{parameters};
            if (ref($method_params) eq 'ARRAY') {
                my $new_schema = {properties => {}};
                for my $param (@$method_params) {
                    if (ref($param) eq 'HASH' && $param->{in} eq 'formData') {
                        $new_schema->{properties}->{$param->{name}} = $param;
                        delete $parameters->{$param->{name}};
                    }
                }

                if (keys %{$new_schema->{properties}}) {
                    $self->multipart_form_parameters(schema => $new_schema, uniq_parameters => $parameters,);
                }
            }
        }
        elsif ($consume =~ m|\/.*?\bform\b|) {
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_HTTP();
        }
        elsif ($consume =~ m|^image\/| || $consume =~ m|^video\/|) {
            $enforcement_type = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_DO_NOTHING();
            $url_object->{flg_disallow_file_upload_of_executables} = 1;
        }

        if ($enforcement_type) {
            my $content_profile_obj = {
                                        enforcement_type => $enforcement_type,
                                        header_name => 'Content-Type',
                                        header_value => $consume,
                                        header_order => $header_order,
                                      };
            ++$header_order;

            if ($content_profile_type) {
                if ($content_profile_entity) {
                    $content_profile_obj->{content_profile} = $content_profile_entity;
                }
                else {
                    $content_profile_obj->{content_profile} = $self->default_content_profile(content_profile_type => $content_profile_type,);
                }
            }

            my $url_content_profile = F5::ASMConfig::Entity::UrlContentProfile->new(dbh => $self->{dbh},
                                                                                policy => $self->{policy},
                                                                                logger => $self->{logger},
                                                                                created_by => $self->{username},
                                                                                object => $content_profile_obj,);

            push @{$url_object->{url_content_profiles}}, $url_content_profile;
            $url_object->{url_content_profiles}->[0]->{enforcement_type} = F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_DISALLOW();
        }
    }

}

#-------------------------------------------------------------------------------

sub check_body_schema_parameters {
    my ($self, %params) = @_;

    my $parameters = $params{parameters};
    my $method_name = $params{method_name};
    my $url_name = $params{url_name};

    my $json_profile;
    for my $param_name (keys %$parameters) {
        my $swagger_param = $parameters->{$param_name};
        my $json_schema_struct = $swagger_param->{schema};
        if ($swagger_param->{in} eq 'body' && $self->schema_is_type_object(schema => $json_schema_struct,)) {
            $json_profile = $self->json_schema_profile(json_schema_struct => $json_schema_struct,
                                                       url_name => $url_name,
                                                       method_name => $method_name,
                                                       param_name => $param_name,);
            last;
        }
    }

    return $json_profile;
}

#-------------------------------------------------------------------------------

sub json_schema_profile {
    my ($self, %params) = @_;

    my $url_name = $params{url_name};
    my $json_schema_struct = $params{json_schema_struct};
    my $method_name = uc($params{method_name});
    my $param_name = $params{param_name};
    my $param_obj = $params{param_obj};
    my $set_defence_values_to_any = $params{set_defence_values_to_any};
    my $disable_check_metachar = $params{disable_check_metachar};
    my $normalized_url_name = $url_name;
    $normalized_url_name =~ s|\/|~|g;
    $normalized_url_name =~ s|\*|\#|g;
    my $profile_name = "json_". $method_name. '_'. $normalized_url_name;
    $profile_name .= '_'. $param_name if $param_name;
    my $json_profile = F5::ASMConfig::Entity::JsonProfile->new( dbh => $self->{dbh},
                                                                policy => $self->{policy},
                                                                get_criteria => {
                                                                    profile_name => $profile_name,
                                                                    content_profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),});

    $self->check_type_nullable(json_schema_struct => $json_schema_struct,);

    if (! exists $json_schema_struct->{$JSON_SCHEMA_ENTRY_NAME}) {
        $json_schema_struct->{$JSON_SCHEMA_ENTRY_NAME} = $DEFAULT_JSON_SCHEMA;
    }
    # in case the param schema specifies 'items: {}' - attach "any-type" schema for json validation
    my $json_validation_content = (ref($json_schema_struct->{items}) && (! scalar keys %{$json_schema_struct->{items}})) ?
                                   $ARRAY_ANY_TYPE_JSON_SCHEMA : JSON::XS->new->encode($json_schema_struct);

    # UTF8 decode required to compensate UTF8 encoding in JSON validation process
    if (utf8::is_utf8($json_validation_content)) {
        utf8::decode($json_validation_content);
    }

    if ($json_profile) {
        my $json_profile_validation_file = $json_profile->{validation_files}->[0];
        if ($json_profile_validation_file) {
            my $json_validation_file = $json_profile_validation_file->{json_validation_file};
            my $exist_file_contents = $json_validation_file->{file_contents};
            if ($json_validation_content ne $exist_file_contents) {
                my $store_validation_file = F5::ASMConfig::Entity::JsonValidationFile->new(
                                                                  policy => $self->{policy},
                                                                  dbh => $self->{dbh},
                                                                  logger => $self->{logger},
                                                                  object => {
                                                                      file_contents => $json_validation_content,
                                                                      file_name => $json_validation_file->{file_name},
                                                                  },);
                $store_validation_file->{rest_uuid} = $store_validation_file->get_rest_uuid();
                my $local_call_flags = F5::ASMConfig::CallFlags->new($self->{call_flags});
                $local_call_flags->{force_update_json_validation_file} = 1;
                $local_call_flags->{force_update_read_only} = 1;
                $self->{handler}->update_json_validation_file($self->{session}, $local_call_flags, $self->{policy},
                                                              $json_validation_file, $store_validation_file,);
                $self->{handler}->handle_json_validation_files($self->{session}, $local_call_flags, $self->{policy}, $json_profile,);
            }

            return $json_profile;
        }
    }

    my $json_validation_entity;
    eval {
        $json_validation_entity = $self->{handler}->add_json_validation_file($self->{session}, $self->{call_flags}, $self->{policy},
                                                                             {file_contents => $json_validation_content,
                                                                              file_name => $profile_name,});
    };
    if ($@) {
        if ($param_obj) {
            $self->log_error(error => $@,);
            $param_obj->{_convert_to_basic_parameter} = 'Fail to create "JSON Schema Validation File"';
            $param_obj->{_check_value_metachar} = 0;
            return;
        }
        die $@;
    }

    if ($json_profile) {
        $self->{handler}->update_json_profile($self->{session},
                                              $self->{call_flags},
                                              $self->{policy},
                                              $json_profile,
                                              {validation_files => [F5::ASMConfig::JsonProfileValidationFile->new({flg_is_primary_enforced => 1,
                                                                                                                   json_validation_file => $json_validation_entity})]},);
    }
    else {
        my $profile_obj = { profile_name => $profile_name,
                            flg_check_signatures => 1,
                            flg_handle_json_parameters => 0,
                            validation_files => [
                                F5::ASMConfig::JsonProfileValidationFile->new({
                                    flg_is_primary_enforced => 1,
                                    json_validation_file => $json_validation_entity})],
                          };

        if ($disable_check_metachar) {
           $profile_obj->{flg_check_metachar} = 0;
        }

        if ($set_defence_values_to_any) {
            my $defense_attributes = F5::ASMConfig::Entity::JsonProfile->get_defense_attributes(dbh => $self->{dbh},);
            my @defences;
            for my $def_attrib_id (keys %$defense_attributes) {
                my $def_attribute = $defense_attributes->{$def_attrib_id};
                if ($def_attribute->{value_type} == F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER()) {
                    my $defense_obj = {def_attrib_id => $def_attrib_id, flg_check_attrib => 0,};
                    push @defences, $defense_obj;
                }
            }
            $profile_obj->{json_defense} = \@defences;

        }

        $json_profile = $self->{handler}->add_json_profile($self->{session}, $self->{call_flags}, $self->{policy}, $profile_obj);
    }

    return $json_profile;
}

#-------------------------------------------------------------------------------

sub check_type_nullable {
    my ($self, %params) = @_;

    my $json_schema_struct = $params{json_schema_struct};

    if (ref($json_schema_struct) eq 'HASH') {
        for my $entry_key (keys %$json_schema_struct) {
            my $value = $json_schema_struct->{$entry_key};
            if (ref($value) eq 'HASH') {
                $self->check_type_nullable(json_schema_struct => $value,);
            }
            elsif ($entry_key eq 'nullable' && $json_schema_struct->{nullable} && $json_schema_struct->{type}) {
                delete $json_schema_struct->{nullable};
                if (ref($json_schema_struct->{type}) eq 'ARRAY') {
                    push @{$json_schema_struct->{type}}, 'null';
                }
                elsif ($json_schema_struct->{type} && ! ref($json_schema_struct->{type})) {
                    $json_schema_struct->{type} = [$json_schema_struct->{type}, 'null'];
                }
            }
        }
    }
}

#-------------------------------------------------------------------------------

sub check_response_codes {
    my ($self, %params) = @_;

    my $responses = $params{responses};
    return unless (ref($responses) eq 'HASH');

    for my $code (keys %$responses) {
        if ($code =~ /^[45]\d\d/) {
            $self->{response_codes}->{$code} = 1;
        }
    }
}

#-------------------------------------------------------------------------------

sub create_response_codes {
    my ($self) = @_;

    F5::ASMConfig::Entity::AllowedResponseCode->delete_many(dbh => $self->{dbh},
                                                            policy => $self->{policy},
                                                            logger => $self->{logger},);

    for my $code (sort keys %{$self->{response_codes}}) {
        my $code_entity = F5::ASMConfig::Entity::AllowedResponseCode->new(policy => $self->{policy},
                                                                          dbh => $self->{dbh},
                                                                          logger => $self->{logger},
                                                                          created_by => $self->{username},
                                                                          object => {response_code => $code},);
        $self->{handler}->add_allowed_response_code($self->{session}, $self->{call_flags}, $self->{policy}, $code_entity);
    }
}

#-------------------------------------------------------------------------------

sub is_cookie_obj_form_true {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};

    my $is_type = 0;
    if (($swagger_param->{in} && $swagger_param->{in} eq 'cookie') &&
	    ($self->schema_is_type_object(schema => $swagger_param->{schema},)) &&     
        (!defined($swagger_param->{style}) || $swagger_param->{style} eq 'form') &&
        (!defined($swagger_param->{explode}) || $swagger_param->{explode})) {
       $is_type = 1;
    }

    return $is_type;
}

#-------------------------------------------------------------------------------

sub create_parameters {
    my ($self, %params) = @_;

    my $url = $params{url};
    my $parameters = $params{parameters};
    my $path_params = $params{path_params};
    my $processed_param_ids = $params{processed_param_ids};

    if (! $path_params) {
        $parameters = $self->query_as_value_parameters(parameters => $parameters, url => $url,);
    }

    my %url_level_params;
    for my $param_name (keys %$parameters) {
        my $swagger_param = $parameters->{$param_name};
        $self->param_schema_entry_data(swagger_param => $swagger_param,);

        if (! $self->support_version(3) && $swagger_param->{in} eq 'body') {
            # "body" parameter will be processed with "consume" handler
            next;
        }

        if ($self->is_cookie_obj_form_true(swagger_param => $swagger_param,)) {
            # skip param creation in case of object paramter in cookie with style:form and explode true.
            # '*' cookie will be used for enforcing in this case
            # note - that deafult serialization for cookie param is style: form explode: true
            my $url_name;
            if ($url) {
                $url_name = $url->get_full_display_name;
            }
            my $log_entity_name = join(', ', grep { $_} $param_name, $url_name);
            $self->{logger}->info("Skipping creation of cookie object param style:form explode:true. $log_entity_name") if $self->{logger};
            next;
        }

        if ($swagger_param->{in} && $swagger_param->{in} eq 'header') {
            $param_name = lc($param_name);
        }

        $param_name = $self->check_convert_encoding(content_value => $param_name);

        my $param_obj = { param_name => $param_name, flg_check_signatures => 1, };
        my $url_segment_index;
        if ($path_params) {
            if ($path_params->{$param_name}) {
                $param_obj->{param_level} = F5::ASMConfig::ParamLevel::PARAM_LEVEL_GLOBAL();
                $param_obj->{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_PATH();
                $param_obj->{param_name} = $path_params->{$param_name}->{unique_param_name};
                $url_segment_index = $path_params->{$param_name}->{url_segment_index};
                if ($swagger_param->{in} ne 'path') {
                    $self->log_error(error_title => "parameter '$param_name' in URL '". $url->{url_name}. "' will be defined as 'path' parameter.");
                    # parameter should be defined in 'path', because URL already has '/*/' for this parameter
                    $swagger_param->{in} = 'path';
                }
            }
            else {
                # regular params are not processed in "path params" call
                # because url obj is not stored yet
                $url_level_params{$swagger_param->{name}} = $swagger_param;
                next;
            }
        }
        else {
            $param_obj->{param_level} = F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL(),
            $param_obj->{url} = $url;
        }

        if ($swagger_param->{in} && $swagger_param->{in} eq 'path' && !$url_segment_index) {
            $self->log_error(error_title => "Position for path parameter '$param_name' in URL '". $url->{url_name}. "' is not defined.");
            next;
        }

        $param_obj = F5::ASMConfig::Entity::Parameter->new(policy => $self->{policy},
                                                           dbh => $self->{dbh},
                                                           logger => $self->{logger},
                                                           created_by => $self->{username},
                                                           preload_default => 1,
                                                           object => $param_obj,);

        if (defined($url_segment_index)) {
            $param_obj->{flg_mandatory} = 0;
        }
        elsif (defined($swagger_param->{required}) && ($swagger_param->{in} || '') ne 'path') {
            $param_obj->{flg_mandatory} = $swagger_param->{required} ? 1 : 0;
        }

        $param_obj->{flg_null_value} = $swagger_param->{allowEmptyValue} ? 1 : 0;

        if ($swagger_param->{pattern}) {
            $param_obj->{flg_regexpr} = 1;
            $param_obj->{regexpr_values} = $swagger_param->{pattern};
        }

        if (ref($swagger_param->{content}) eq 'HASH' ) {
            $self->param_content(content => $swagger_param->{content}, param_obj => $param_obj, url => $url, param_name => $param_name,);
        }

        if (! $param_obj->{param_content_profile}) {
            if ($self->schema_is_type_object(schema => $swagger_param->{schema},)) {
                $self->param_value_type_object(swagger_param => $swagger_param, param_obj => $param_obj, url => $url, param_name => $param_name,);
            }
            else {
                $self->param_schema_array(param_obj => $param_obj, swagger_param => $swagger_param, url => $url, param_name => $param_name,);
                if ($param_obj->{param_value_type} != F5::ASMConfig::ParamValueType::PARAM_VALUE_OPENAPI_ARRAY_VALUE()) {
                   $self->apply_parameter_type(swagger_param => $swagger_param, param_obj => $param_obj);
                   $self->apply_enum(enum_obj => $swagger_param, param_obj => $param_obj);
                }
            }
        }

        $self->apply_parameter_location(swagger_param => $swagger_param, param_obj => $param_obj);

        my $parameter = F5::ASMConfig::Entity::Parameter->new(policy => $self->{policy},
                                                              dbh => $self->{dbh},
                                                              logger => $self->{logger},
                                                              get_criteria => $param_obj,);
        my $url_name = $self->get_url_name(url => $url,);

        if ($param_obj->{_convert_to_basic_parameter}) {
            $self->log_error(error_title => "$url_name, Parameter: $param_name. ". $param_obj->{_convert_to_basic_parameter},);
            $param_obj = $self->simplify_parameter(param_obj => $param_obj,);
        }
        elsif (! $parameter) {
            # in case parameter has invalid configuration, we should try to create it as most simple parameter
            my $new_param_error_msg = $self->validate_parameter(param_obj => $param_obj,);
            if ($new_param_error_msg) {
                my $detail_msg_string = ref($new_param_error_msg) ? $new_param_error_msg->{error_message} : $new_param_error_msg;
                $self->log_error(error_title => "Fail validation for new parameter: $url_name, Parameter: $param_name. $detail_msg_string Parameter of type Alpha-Numeric will be created.");
                $param_obj = $self->simplify_parameter(param_obj => $param_obj,);
            }
        }

        # prevalidation required to prevent "rollback" in case invalid parameter
        my $error_msg = $self->validate_parameter(parameter => $parameter, param_obj => $param_obj,);
        if ($error_msg) {
            $self->log_error(error_title => "Fail validation for new parameter: $param_name, $url_name",error => $error_msg);
            $processed_param_ids->{$parameter->{param_id}} = 1;
            next; # skip param creation alltogether
        }

        if ($parameter) {
            $self->{logger}->debug("Update swagger Parameter: $param_name, $url_name") if $self->{logger};
            my $local_call_flags = F5::ASMConfig::CallFlags->new($self->{call_flags});
            $local_call_flags->{force_update_read_only} = 1;
            eval {
                $self->{handler}->update_parameter($self->{session}, $local_call_flags, $self->{policy}, $parameter, $param_obj);
            };
            $self->log_error(error_title => "Fail update parameter: $param_name, $url_name",error => $@) if $@;
        }
        else {
            $self->{logger}->debug("Create swagger Parameter: $param_name, $url_name") if $self->{logger};
            eval {
                $parameter = $self->{handler}->add_parameter($self->{session}, $self->{call_flags}, $self->{policy}, $param_obj);
            };
            $self->log_error(error_title => "Fail create new parameter: $param_name, $url_name",error => $@) if $@;
        }

        if (ref( $processed_param_ids )) {
            $processed_param_ids->{$parameter->{param_id}} = 1;
        }

        if ($path_params && defined($url_segment_index)) {
            my $pos_param = F5::ASMConfig::PositionalParameter->new(param => $parameter,
                                                                    url_segment_index => $url_segment_index,);

            $url->{positional_parameters} ||= [];
            push @{$url->{positional_parameters}}, $pos_param;
        }
    }

    return \%url_level_params;
}

#-------------------------------------------------------------------------------

sub simplify_parameter {
    my ($self, %params) = @_;

    my $param_obj = $params{param_obj};

    my $new_param_obj = F5::ASMConfig::Entity::Parameter->new(policy => $self->{policy},
                                                              dbh => $self->{dbh},
                                                              logger => $self->{logger},
                                                              created_by => $self->{username},
                                                              preload_default => 1,
                                                              object => {
                                                                  param_name => $param_obj->{param_name},
                                                                  url => $param_obj->{url},
                                                                  param_level => $param_obj->{param_level},
                                                                  flg_check_signatures => $param_obj->{flg_check_signatures},
                                                                  parameter_location => $param_obj->{parameter_location},
                                                              },);
    if (exists $param_obj->{_check_value_metachar}) {
        $new_param_obj->{flg_check_param_value_metachar} = $param_obj->{_check_value_metachar};
    }

    return $new_param_obj;
}

#-------------------------------------------------------------------------------

sub validate_parameter {
    my ($self, %params) = @_;

    my $parameter = $params{parameter};
    my $param_obj = $params{param_obj};

    my $error_msg;
    eval {
        my $param = $parameter ? $parameter : $param_obj;
        $param->before_store(store_obj => $param_obj,
                             validate_value => 1,
                             store_mode => 'add',
                             call_flags => F5::ASMConfig::CallFlags->new(),
                             policy => $self->{policy},);
    };
    if ($@) {
        $error_msg = $@;
        if (ref($error_msg) eq 'HASH' && $error_msg->{error_code} == F5::ASMConfig::ErrorCode::SWAGGER_SERIALIZATION_ERRORS()) {
            $param_obj->{_check_value_metachar} = 0;
        }
    }

    return $error_msg;
}

#-------------------------------------------------------------------------------

sub check_convert_encoding {
    my ($self, %params) = @_;

    my $content_value = $params{content_value};

    # utf8 decode required because JSON parser that convert content into utf8
    if($self->{encoding} && $self->{encoding} !~ /utf.?8/i) {
        eval {
            utf8::decode $content_value;
        };
        if ($@) {
            $self->log_error(error_title => "Fail decode content - $content_value", error => $@);
        }
    }

    return $content_value;
}

#-------------------------------------------------------------------------------

sub query_as_value_parameters {
    my ($self, %params) = @_;

    my $parameters = $params{parameters};
    my $url = $params{url};

    my $found_param = 0;
    for my $param_name (keys %$parameters) {
        my $swagger_param = $parameters->{$param_name};
        my $param_obj = {};
        if ($self->schema_is_type_object(schema => $swagger_param->{schema},)) {
            $self->object_serialization(swagger_param => $swagger_param, param_obj => $param_obj,);
        }

        $self->apply_parameter_location(swagger_param => $swagger_param, param_obj => $param_obj,);

        if (($param_obj->{parameter_location} || 0) == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_QUERY()) {
            $swagger_param->{_is_query_location} = 1;

            if ($param_obj->{object_serialization_style} &&
                $param_obj->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_FORM() &&
                $param_obj->{flg_object_serialization_explode}) {
                    $found_param = 1;
                    $swagger_param->{_is_query_as_value} = 1;
            }
        }
    }

    my $result_params = $parameters;
    if ($found_param) {
        $result_params = {};
        my $query_param_taken = 0;

        my $preserved_parameter;
        my @discard_parameters;
        for my $param_name (keys %$parameters) {
            my $swagger_param = $parameters->{$param_name};
            if ($swagger_param->{_is_query_as_value} && ! $query_param_taken) {
                $result_params->{$param_name} = $swagger_param;
                $query_param_taken = 1;
                $preserved_parameter = $swagger_param;
            }
            elsif($swagger_param->{_is_query_location}) {
                push @discard_parameters, $swagger_param;
            }

            next if $swagger_param->{_is_query_location};

            $result_params->{$param_name} = $swagger_param;
        }

        if ($preserved_parameter && scalar @discard_parameters) {
            my $error_msg = $self->get_url_name(url => $url,);
            $error_msg .= ": Found query location object parameter with serialization style='form' and explode='true' - \"$preserved_parameter->{name}\".";
            $error_msg .= " All other query location parameters will be discarded: ". join(', ', map {"\"$_->{name}\""} @discard_parameters);

            $self->log_error(error_title => $error_msg);
        }
    }

    return $result_params;
}

#-------------------------------------------------------------------------------

sub get_url_name {
    my ($self, %params) = @_;

    my $url = $params{url};

    my $url_name = 'Path: '. $url->get_full_display_name;
    $url_name = 'OperationId: '. $url->{operation_id}. ', '. $url_name if $url->{operation_id};

    return $url_name;
}

#-------------------------------------------------------------------------------

sub param_schema_array {
    my ($self, %params) = @_;

    my $param_obj = $params{param_obj};
    my $swagger_param = $params{swagger_param};
    my $url = $params{url};
    my $param_name = $params{param_name};

    my $schema = $swagger_param->{schema};

    return if (! $self->schema_is_type_array(schema => $schema,));

    if ($self->support_version(3)) {
        if ($swagger_param->{style} && $swagger_param->{in} eq 'cookie' && $swagger_param->{style} ne 'form') {
            $param_obj->{_convert_to_basic_parameter} = 'Cookie array parameter supports only serialization style "form".';
            $param_obj->{_check_value_metachar} = 0;
            return;
        }

        if ($swagger_param->{style} && $swagger_param->{in} eq 'header' && $swagger_param->{style} ne 'simple') {
            $param_obj->{_convert_to_basic_parameter} = 'Header array parameter supports only serialization style "simple".';
            $param_obj->{_check_value_metachar} = 0;
            return;
        }

        if ($swagger_param->{style} && $swagger_param->{in} eq 'query' &&
            ! ($swagger_param->{style} eq 'form' ||
               $swagger_param->{style} eq 'spaceDelimited' ||
               $swagger_param->{style} eq 'pipeDelimited')) {
            $param_obj->{_convert_to_basic_parameter} = 'Query array parameter supports only serialization style "form", "spaceDelimited" or "pipeDelimited".';
            $param_obj->{_check_value_metachar} = 0;
            return;
        }
    }

    my $local_schema = dclone($schema);
    if (ref($schema->{items}) &&
        $schema->{items}->{type} &&
        ! ref($schema->{items}->{type}) &&
        $schema->{items}->{type} ne 'object' &&
        $schema->{items}->{type} ne 'array') {
        for my $schema_entry (keys %$local_schema) {
            $swagger_param->{$schema_entry} = $local_schema->{$schema_entry};
        }
        delete $swagger_param->{schema};

        if ($self->support_version(3)) {
            # in v3 set default style/explode if not defined in the swagger file (needed to determine serialization format of the array paramter)
            # (do not set it if v2 since in v2 array serialization is determined based on the tag collectionFormat. Not based on style & explode)
            if (!defined($swagger_param->{style}) && $default_serialization_style{$swagger_param->{in}}) {
                $swagger_param->{style} = $default_open_api_array_style{$swagger_param->{in}};
            }

            if (!defined($swagger_param->{explode}) && $default_explode{$swagger_param->{in}}) {
                $swagger_param->{explode} = $default_explode{$swagger_param->{in}};
            }
        }
    }
    else {
        # type is openapi array (e.g mixed-type array or array with schema items: {})
        $self->param_value_type_object(swagger_param => $swagger_param,
                                       schema => $local_schema,
                                       param_obj => $param_obj,
                                       url => $url,
                                       param_name => $param_name,
                                       param_value_type => F5::ASMConfig::ParamValueType::PARAM_VALUE_OPENAPI_ARRAY_VALUE(),);
    }
}

#-------------------------------------------------------------------------------

sub param_value_type_object {
    my ($self, %params) = @_;

    my $param_obj = $params{param_obj};
    my $url = $params{url};
    my $param_name = $params{param_name};
    my $swagger_param = $params{swagger_param};
    my $schema = $params{schema} || $swagger_param->{schema};
    my $param_value_type = $params{param_value_type};

    $param_obj->{param_value_type} = $param_value_type || F5::ASMConfig::ParamValueType::PARAM_VALUE_OBJECT_VALUE();
    my $json_profile = $self->json_schema_profile(json_schema_struct => $schema,
                                                  url_name => $url->{url_name},
                                                  method_name => $url->{method_name},
                                                  param_name => $param_name,
                                                  param_obj => $param_obj,
                                                  set_defence_values_to_any => 1,
                                                  disable_check_metachar => 1,
                                                  );

    my $param_content_profile = F5::ASMConfig::Entity::ParameterContentProfile->new(policy => $self->{policy},
                                                                                    dbh => $self->{dbh},
                                                                                    logger => $self->{logger},
                                                                                    object => {
                                                                                        content_profile => $json_profile,
                                                                                    });
    $param_obj->{param_content_profile} = $param_content_profile;

    if (($param_obj->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_FORM ||
         $param_obj->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_SPACE_DELIMITED ||
         $param_obj->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_PIPE_DELIMITED ||
         $param_obj->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_MATRIX ) &&
         $param_obj->{flg_object_serialization_explode} == 1) {
        $param_obj->{flg_allow_repeated_parameter_name} = 1;
    }
    else {
        $param_obj->{flg_allow_repeated_parameter_name} = 0;
    }

    $self->object_serialization(swagger_param => $swagger_param, param_obj => $param_obj,);
}

#-------------------------------------------------------------------------------

sub object_serialization {
    my ($self, %params) = @_;

    my $param_obj = $params{param_obj};
    my $swagger_param = $params{swagger_param};

    if (defined($swagger_param->{style})) {
        $param_obj->{object_serialization_style} = $serialization_style_map{$swagger_param->{style}};
    }
    else {
        $param_obj->{object_serialization_style} = $default_serialization_style{$swagger_param->{in}} if $default_serialization_style{$swagger_param->{in}};
    }

    if (defined($swagger_param->{explode})) {
        $param_obj->{flg_object_serialization_explode} = $swagger_param->{explode} ? 1 : 0;
    }
    else {
        $param_obj->{flg_object_serialization_explode} = $default_explode{$swagger_param->{in}} if $default_explode{$swagger_param->{in}};
    }
}

#-------------------------------------------------------------------------------

sub param_content {
    my ($self, %params) = @_;

    my $content = $params{content};
    my $param_obj = $params{param_obj};
    my $url = $params{url};
    my $param_name = $params{param_name};

    my ($param_value_type, $content_profile, $json_schema);
    for my $content_type (keys %$content) {
        if ($content_type =~ /\bjson\b/ &&
            ref($content->{$content_type}) eq 'HASH' &&
            ref($content->{$content_type}->{schema}) eq 'HASH') {
                $json_schema = $content->{$content_type}->{schema};
                last;
            }
    }

    if ($json_schema) {
        $param_value_type = F5::ASMConfig::ParamValueType::PARAM_VALUE_JSON_VALUE();
        $param_obj->{flg_allow_repeated_parameter_name} = 0;

        if ($self->schema_is_type_object(schema => $json_schema,) || $self->schema_is_type_array(schema => $json_schema,)) {
            $content_profile = $self->json_schema_profile(json_schema_struct => $json_schema,
                                                          url_name => $url->{url_name},
                                                          method_name => $url->{method_name},
                                                          param_obj => $param_obj,
                                                          param_name => $param_name,);
        }
        else {
            $content_profile = $self->default_content_profile(content_profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),);
        }
    }
    else {
        my @xml_contents = grep {$_ =~ /\bxml\b/} keys %$content;
        if (scalar @xml_contents) {
            $param_value_type = F5::ASMConfig::ParamValueType::PARAM_VALUE_XML_VALUE();
            $param_obj->{flg_allow_repeated_parameter_name} = 0;
            $content_profile = $self->default_content_profile(content_profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(),);
        }
        else {
            my @form_contents = grep {$_ =~ /\bform\b/} keys %$content;
            if (scalar @form_contents) {
                $param_obj->{parameter_location} = $param_location_map{formData};
            }
        }
    }

    if ($param_value_type) {
        $param_obj->{param_value_type} = $param_value_type;
        my $param_content_profile = F5::ASMConfig::Entity::ParameterContentProfile->new(policy => $self->{policy},
                                                                                        dbh => $self->{dbh},
                                                                                        logger => $self->{logger},
                                                                                        object => {
                                                                                            content_profile => $content_profile,
                                                                                        });
        $param_obj->{param_content_profile} = $param_content_profile;
    }
}

#-------------------------------------------------------------------------------

sub param_schema_entry_data {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    return if ! ref($swagger_param->{schema});
    return if $self->schema_is_type_object(schema => $swagger_param->{schema},);
    return if $self->schema_is_type_array(schema => $swagger_param->{schema},);

    my %schema_reference_entries = (enum => 1,);
    for my $schema_entry (keys %{$swagger_param->{schema}}) {
        if (! exists($swagger_param->{$schema_entry})) {
            if ((! ref($swagger_param->{schema}->{$schema_entry})) ||
                $schema_reference_entries{ $schema_entry } ||
                F5::ObjUtils::is_ref_boolean_type($swagger_param->{schema}->{$schema_entry})) {
                $swagger_param->{$schema_entry} = $swagger_param->{schema}->{$schema_entry};
            }
        }
    }
}

#-------------------------------------------------------------------------------

sub schema_is_type_object {
    my ($self, %params) = @_;

    my $schema = $params{schema};

    my $is_type_object = 0;
    if (ref($schema) eq 'HASH') {
        if (exists $schema->{type}) {
            if ($schema->{type} eq 'object') {
                $is_type_object = 1;
            }
        }

        if (exists $schema->{oneOf} || exists $schema->{anyOf} || exists $schema->{allOf}) {
            $is_type_object = 1;
        }

        if (! $is_type_object && exists $schema->{properties} && ! exists $schema->{type}) {
            $schema->{type} = 'object';
            $is_type_object = 1;
        }
    }

    return $is_type_object;
}

#-------------------------------------------------------------------------------

sub schema_is_type_array {
    my ($self, %params) = @_;

    my $schema = $params{schema};

    my $is_type_array = 0;
    if (ref($schema) eq 'HASH' && $schema->{type} eq 'array') {
        $is_type_array = 1;
    }

    return $is_type_array;
}

#-------------------------------------------------------------------------------

sub apply_parameter_location {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    my $location = $swagger_param->{in} || '';
    if (defined($param_location_map{$location})) {
        $param_obj->{parameter_location} = $param_location_map{$location};
    }
    elsif ($location eq 'header') {
        $param_obj->{flg_is_header} = 1;
        $param_obj->{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_HEADER();
    }
    elsif ($location eq 'cookie') {
        $param_obj->{flg_is_cookie} = 1;
        $param_obj->{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE();
    }
    else {
        $param_obj->{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_ANY();
    }
}

#-------------------------------------------------------------------------------

sub apply_parameter_type {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    my $type = $swagger_param->{type} || '';
    if (defined($param_type_map{$type})) {
        $param_obj->{param_value_type} = F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT();
        $param_obj->{param_data_type} = $param_type_map{$type};
    }
    elsif ($type eq 'array') {
        $param_obj->{param_value_type} = F5::ASMConfig::ParamValueType::PARAM_VALUE_ARRAY_VALUE();
        if (ref($swagger_param->{items})) {
            my $itemType = $swagger_param->{items}->{type};
            my $itemFormat = $swagger_param->{items}->{format};
            if ($self->is_param_binary(type => $itemType, format => $itemFormat,)) {
                $param_obj->{param_data_type} = $param_type_map{file};
            }
            elsif ($itemType && $param_type_map{$itemType}) {
                $param_obj->{param_data_type} = $param_type_map{$swagger_param->{items}->{type}};
            }
        }
        $self->apply_parameter_array(swagger_param => $swagger_param, param_obj => $param_obj);
    }
    else {
        $param_obj->{param_value_type} = F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT();
        $param_obj->{param_data_type} = F5::ASMConfig::ParamDataType::PARAM_DATA_ALPHA_NUMERIC();
        $swagger_param->{type} ||= 'string';
    }

    $self->apply_parameter_type_numbers(swagger_param => $swagger_param, param_obj => $param_obj);
    $self->apply_parameter_type_string(swagger_param => $swagger_param, param_obj => $param_obj);
    $self->apply_parameter_type_boolean(swagger_param => $swagger_param, param_obj => $param_obj);
    $self->apply_parameter_type_file(swagger_param => $swagger_param, param_obj => $param_obj);
}

#-------------------------------------------------------------------------------

sub apply_parameter_type_file {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    my $param_type = $swagger_param->{type};
    return unless $param_type eq 'file';

    $param_obj->{in_staging} = 0;
    $param_obj->{param_value_type} = F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT();
    $param_obj->{flg_check_max_value_len} = 0;
    $param_obj->{flg_disallow_upload_of_exe} = 1;
    if ($swagger_param->{format} eq 'base64') {
        $param_obj->{flg_is_base64} = 1;
        $param_obj->{flg_detect_base64} = 0;
    }
    else {
        $param_obj->{flg_is_base64} = 0;
    }
    $param_obj->{flg_allow_repeated_parameter_name} = 0;
    $param_obj->{flg_is_sensitive} = 0;

    $self->apply_parameter_value_length(%params);
}

#-------------------------------------------------------------------------------

sub apply_parameter_type_boolean {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    my $param_type = $swagger_param->{type};
    return unless $param_type eq 'boolean';

    $param_obj->{in_staging} = 0;
    $param_obj->{flg_allow_repeated_parameter_name} = 0;
}

#-------------------------------------------------------------------------------

sub apply_parameter_type_string {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    my $param_type = $swagger_param->{type};
    return unless $param_type eq 'string';

    $param_obj->{in_staging} = 0;
    if ($swagger_param->{flg_allow_repeated_parameter_name}) {
         $param_obj->{flg_allow_repeated_parameter_name} = $swagger_param->{flg_allow_repeated_parameter_name};
    }
    else {
         $param_obj->{flg_allow_repeated_parameter_name} = 0;
    }

    $self->apply_parameter_value_length(%params);
}

#-------------------------------------------------------------------------------

sub apply_parameter_value_length {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    if (defined($swagger_param->{minLength})) {
        $param_obj->{min_value_len} = $swagger_param->{minLength};
        $param_obj->{flg_check_min_value_len} = 1;
    }

    if (defined($swagger_param->{maxLength})) {
        $param_obj->{max_value_len} = $swagger_param->{maxLength};
        $param_obj->{flg_check_max_value_len} = 1;
    }

    if (defined($swagger_param->{pattern})) {
        $param_obj->{flg_regexpr} = 1;
        $param_obj->{regexpr_values} = $swagger_param->{pattern};
    }

    my $swagger_param_format = $swagger_param->{format} || '';
    if ($swagger_param_format eq 'byte') {
        $param_obj->{flg_is_base64} = 1;
        $param_obj->{flg_detect_base64} = 0;
    }
    elsif ($swagger_param_format eq 'password') {
        $param_obj->{flg_is_sensitive} = 1;
    }
    elsif ($swagger_param_format eq 'date') {
        $param_obj->{flg_regexpr} = 1;
        $param_obj->{regexpr_values} = '^([12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01]))$';
    }
    elsif ($swagger_param_format eq 'date-time') {
        $param_obj->{flg_regexpr} = 1;
        $param_obj->{regexpr_values} = '^([12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01]))T(0\d|2[0-3]):([0-5]\d):([0-5]\d)(\.\d+)?(Z|((\+|-)(0\d|2[0-3]):([0-5]\d)))$';
    }
}

#-------------------------------------------------------------------------------

sub apply_parameter_type_numbers {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    my $param_type = $swagger_param->{type};
    return unless ($param_type eq 'integer' || $param_type eq 'number');

    $param_obj->{in_staging} = 0;
    $param_obj->{flg_allow_repeated_parameter_name} = 0;
    $param_obj->{flg_check_max_value_len} = 0;

    if (defined($swagger_param->{minimum})) {
        $param_obj->{min_value} = $swagger_param->{minimum};
        $param_obj->{flg_check_min_value} = 1;
        if (defined($swagger_param->{exclusiveMinimum})) {
            $param_obj->{flg_exclusive_min} = $swagger_param->{exclusiveMinimum} ? 1 : 0;
        }
    }

    if (defined($swagger_param->{maximum})) {
        $param_obj->{max_value} = $swagger_param->{maximum};
        $param_obj->{flg_check_max_value} = 1;
        if (defined($swagger_param->{exclusiveMaximum})) {
            $param_obj->{flg_exclusive_max} = $swagger_param->{exclusiveMaximum} ? 1 : 0;
        }
    }

    if (defined($swagger_param->{multipleOf})) {
        $param_obj->{multiple_of_value} = $swagger_param->{multipleOf};
        $param_obj->{flg_check_multiple_of_value} = 1;
    }
    else {
        $param_obj->{flg_check_multiple_of_value} = 0;
    }

    if ($swagger_param->{format} && $swagger_param->{format} eq 'int32') {
        # length of min value 'int32' "-2147483647" longer than 10, will not be defined
        if (! defined($param_obj->{max_value})) {
            $param_obj->{max_value} = 2147483647;
            $param_obj->{flg_check_max_value} = 1;
        }
    }

    $self->apply_parameter_value_length(%params);
}

#-------------------------------------------------------------------------------

sub apply_parameter_array {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    if (defined($swagger_param->{minItems})) {
        $param_obj->{flg_check_min_items_in_array} = 1;
        $param_obj->{min_items_in_array} = $swagger_param->{minItems};
    }
    if (defined($swagger_param->{maxItems})) {
        $param_obj->{flg_check_max_items_in_array} = 1;
        $param_obj->{max_items_in_array} = $swagger_param->{maxItems};
    }

    $param_obj->{flg_check_array_unique_items} = $swagger_param->{uniqueItems} ? 1 : 0;
    if ($swagger_param->{collectionFormat} && $param_serialization_map{$swagger_param->{collectionFormat}}) {
        $param_obj->{array_serialization_format} = $param_serialization_map{$swagger_param->{collectionFormat}}
    }
    else {
        $self->apply_array_serialization_style(swagger_param => $swagger_param, param_obj => $param_obj,);
    }

    $self->apply_enum(enum_obj => $swagger_param->{items}, param_obj => $param_obj);
    $self->apply_parameter_type_numbers(swagger_param => $swagger_param->{items}, param_obj => $param_obj);
    $self->apply_parameter_type_string(swagger_param => $swagger_param->{items}, param_obj => $param_obj);
    $self->apply_parameter_type_boolean(swagger_param => $swagger_param->{items}, param_obj => $param_obj);
    $self->apply_parameter_type_file(swagger_param => $swagger_param->{items}, param_obj => $param_obj);

}

#-------------------------------------------------------------------------------

sub apply_array_serialization_style {
    my ($self, %params) = @_;

    my $swagger_param = $params{swagger_param};
    my $param_obj = $params{param_obj};

    if ($swagger_param->{style}) {
        my $style = $swagger_param->{style};
        if (! $serialization_array_style_map{$style}) {
            $param_obj->{_convert_to_basic_parameter} = "Serialization style \"$style\" is not supported for parameter type \"array\".";
            $param_obj->{_check_value_metachar} = 0;
            return;
        }

        my $explode = $swagger_param->{explode};
        if ($style eq 'simple') {
            $param_obj->{array_serialization_format} = F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_CSV();
        }
        elsif($explode) {
            if ($style eq 'form' || $style eq 'spaceDelimited' || $style eq 'pipeDelimited') {
                $param_obj->{array_serialization_format} = F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MULTI();
            }
            elsif ($style eq 'label') {
                $param_obj->{flg_object_serialization_explode} = 1;
                $param_obj->{array_serialization_format} = F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_LABEL();
            }
            elsif ($style eq 'matrix') {
                $param_obj->{flg_object_serialization_explode} = 1;
                $param_obj->{array_serialization_format} = F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MATRIX();
            }
        }
        else {
            if ($serialization_array_style_map{$style}) {
                $param_obj->{array_serialization_format} = $serialization_array_style_map{$style};
                $param_obj->{flg_object_serialization_explode} = 0;
            }
        }
    }

    if ((!$param_obj->{array_serialization_format}) && $swagger_param->{in} && $default_array_serialization_style{$swagger_param->{in}}) {
        $param_obj->{array_serialization_format} = $default_array_serialization_style{$swagger_param->{in}};
    }
}

#-------------------------------------------------------------------------------

sub apply_enum {
    my ($self, %params) = @_;

    my $enum_obj = $params{enum_obj};
    my $param_obj = $params{param_obj};

    my $enums_list = $enum_obj->{enum};
    return unless (ref($enums_list) || '') eq 'ARRAY';

    for my $i (0..$#$enums_list) {
        my $enum_elem = $enums_list->[$i];
        if (F5::ObjUtils::is_ref_boolean_type($enum_elem)) {
            # in string context 'Boolean' objects give proper string representation
            $enums_list->[$i] = "$enum_elem";
        }
    }

    $param_obj->{param_enum_values} = $enums_list;
}

#-------------------------------------------------------------------------------

sub get_policy_description {
    my ($self) = @_;

    my $policy_description;
    if ($self->{openapi_content} && $self->{openapi_content}->{title}) {
        $policy_description = $self->{openapi_content}->{title};
    }

    return $policy_description;
}

#-------------------------------------------------------------------------------

sub check_produces {
    my ($self, %params) = @_;

    my $process_elem = $params{process_elem};
    my $scope = $params{scope};

    return unless (ref($process_elem) eq 'HASH');
    $self->{produces}->{$scope} ||= {};

    if ($self->support_version(3)) {
        my $all_responses = $process_elem->{responses};
        if (ref($all_responses) eq 'HASH') {
            for my $response (keys %$all_responses) {
                if (ref($all_responses->{$response}) eq 'HASH') {
                    my $content = $all_responses->{$response}->{content};
                    if (ref($content) eq 'HASH') {
                        for my $produce (keys %$content) {
                            $self->{produces}->{$scope}->{$produce} = 1;
                        }
                    }
                }
            }
        }
    }
    else {
        my $produces = $process_elem->{produces};
        return unless (ref($produces) eq 'ARRAY');

        for my $produce (@$produces) {
            $self->{produces}->{$scope}->{$produce} = 1;
        }
    }
}

#-------------------------------------------------------------------------------

sub create_default_response {
    my ($self) = @_;

    my $produces;
    if (ref($self->{produces}->{global}) eq 'HASH' && scalar keys %{$self->{produces}->{global}}) {
        $produces = $self->{produces}->{global};
    }
    elsif (ref($self->{produces}->{url}) eq 'HASH' && scalar keys %{$self->{produces}->{url}}) {
        $produces = $self->{produces}->{url};
    }

    if (ref($produces) eq 'HASH') {
        if ($self->produces_contain_type(produces => $produces, produce_type => 'json',)) {
            $self->set_default_response(response_type => 'application/json');
        }
        elsif ($self->produces_contain_type(produces => $produces, produce_type => 'xml',)) {
            $self->set_default_response(response_type => 'application/xml');
        }
        else {
            $self->set_default_response(response_type => 'text/html');
        }
    }
    else {
        $self->set_default_response(response_type => 'application/json');
    }
}

#-------------------------------------------------------------------------------

sub produces_contain_type {
    my ($self, %params) = @_;

    my $produce_type = $params{produce_type};
    return unless $produce_type;

    my $produces = $params{produces};

    my @type_elements = grep {$_ =~ /\b$produce_type\b/} keys %$produces;
    return scalar @type_elements;
}

#-------------------------------------------------------------------------------

sub set_default_response {
    my ($self, %params) = @_;

    my $response_type = $params{response_type};
    return unless $response_type;

    my $responses_defs = YAML::XS::LoadFile($DEFAULT_RESPONSES_FILE);

    my $response_obj = { response_type => F5::ASMConfig::ResponseType::RESPONSE_TYPE_DEFAULT(), };
    my $default_response = F5::ASMConfig::Entity::ResponsePage->new(dbh => $self->{dbh},
                                                                    policy => $self->{policy},
                                                                    logger => $self->{logger},
                                                                    get_criteria => $response_obj,);

    $response_obj->{response_action_type} = F5::ASMConfig::ResponseActionType::RESPONSE_ACTION_TYPE_CUSTOM();
    $response_obj->{response_header} = $responses_defs->{$response_type}->{response_headers};
    $response_obj->{response_content} = $responses_defs->{$response_type}->{response_body};

    eval {
        $default_response->store_update(store_obj => $response_obj,);
    };
    if ($@) {
        $self->log_error(error_title => "Fail update default response", error => $@);
    }
}

#-------------------------------------------------------------------------------

sub replace_refs {
    my ($self, %params) = @_;

    my $openapi_content = $self->{openapi_content};
    return unless $openapi_content;

    # dereference will be applied only on parts of swagger relevant for security policy
    # all other parts of swagger will be dereferenced only if referenced by relevat parts
    my @dereference_entries = qw(basePath paths servers requestBody consumes);
    for my $ref_entry (@dereference_entries) {
        $self->replace_element_refs(process_elem => $openapi_content->{$ref_entry},);
    }
}

#-------------------------------------------------------------------------------

sub replace_element_refs {
    my ($self, %params) = @_;

    my $process_elem = $params{process_elem};
    my $schema_entry = $params{schema_entry};
    return unless $process_elem;

    if (ref($process_elem) eq 'HASH') {
        for my $elem_name (keys %$process_elem) {
            my $schema_struct_level = 0;

            if ($elem_name eq 'schema' && ref($process_elem->{$elem_name}) eq 'HASH' && ! $schema_entry) {
                $schema_entry = dclone($process_elem->{$elem_name});
                $schema_struct_level++;
            }

            my $ref_elem = $self->get_ref_elem(elem_value => $process_elem->{$elem_name}, schema_entry => $schema_entry, schema_struct_level => $schema_struct_level,);
            if ($ref_elem) {
                $process_elem->{$elem_name} = dclone($ref_elem);
            }
            else {
                $self->replace_element_refs(process_elem => $process_elem->{$elem_name}, schema_entry => $schema_entry,);
            }

            if ($schema_struct_level) {
                $process_elem->{$elem_name} = {%{$process_elem->{$elem_name}}, %$schema_entry};
                delete $process_elem->{$elem_name}->{'$ref'};
            }
        }
    }
    elsif (ref($process_elem) eq 'ARRAY') {
        my $array_size = scalar @$process_elem;
        for (my $i=0; $i<$array_size; ++$i) {
            my $ref_elem = $self->get_ref_elem(elem_value => $process_elem->[$i], schema_entry => $schema_entry,);
            if ($ref_elem) {
                $process_elem->[$i] = dclone($ref_elem);
            }
            else {
                $self->replace_element_refs(process_elem => $process_elem->[$i], schema_entry => $schema_entry,);
            }
        }
    }
}

#-------------------------------------------------------------------------------

sub get_ref_elem {
    my ($self, %params) = @_;

    my $elem_value = $params{elem_value};
    my $schema_entry = $params{schema_entry};
    my $schema_struct_level = $params{schema_struct_level};

    if (ref($elem_value) eq 'HASH' ) {
        my $path_string = $elem_value->{'$ref'};
        if ($path_string && $path_string =~ s|^\#\/(?=.)||) {
            if ($self->{ref_paths_seen}->{$path_string}) {
                if (! $schema_entry) {
                    $self->log_error(error_title => "Recursive reference is not allowed: reference - $path_string");
                    delete $elem_value->{'$ref'};
                }
                return undef;
            }

            my @ref_path = split(/\//, $path_string);
            my $current_element = $self->{openapi_content};
            my $current_schema_entry = $schema_entry;
            for my $current_name (@ref_path) {
                unless (ref($current_element) && $current_element->{$current_name}) {
                    return undef;
                }
                $current_element = $current_element->{$current_name};
                if (ref($current_schema_entry)) {
                    $current_schema_entry = $current_schema_entry->{$current_name};
                }
                else {
                    $current_schema_entry = undef;
                }
            }

            if ($schema_entry && ref($current_element) && ! $schema_struct_level) {
                if (! $current_schema_entry) {
                    my $new_schema_entry = $schema_entry;
                    my $path_idx = 0;
                    while ( 1 ) {
                        my $current_name = $ref_path[$path_idx];
                        ++$path_idx;
                        if ($path_idx <= $#ref_path) {
                            $new_schema_entry->{$current_name} ||= {};
                            $new_schema_entry = $new_schema_entry->{$current_name};
                        }
                        else {
                            $new_schema_entry->{$current_name} = $current_element;
                            $self->replace_element_refs(process_elem => $current_element, schema_entry => $schema_entry,);
                            last;
                        }
                    }
                }

                $current_element = $elem_value;
            }
            else {
                $self->{ref_paths_seen}->{$path_string} += 1 if (! $schema_struct_level);
                $self->replace_element_refs(process_elem => $current_element, schema_entry => $schema_entry,);
                $self->{ref_paths_seen}->{$path_string} -= 1 if (! $schema_struct_level);
            }

            return $current_element;
        }
    }

    return undef;
}

#-------------------------------------------------------------------------------

sub parse_path_parameters {
    my ($self, %params) = @_;

    my $absolute_path = $params{absolute_path};
    my $normalized_url_name = $absolute_path;
    my $method_name = $params{method_name} || '';

    $absolute_path =~ s|^\/||;

    $normalized_url_name =~ s|\/|~|g;
    $normalized_url_name =~ s|\{(.+)\}|\*|g;
    $normalized_url_name =~ s|\*|\#|g;

    my @path_list = split(/\/+/, $absolute_path);
    my $url_segment_index = 0;
    my %path_params;
    my $url_name;
    for my $path_element (@path_list) {
        ++$url_segment_index;
        if ($path_element =~ s|\{(.+)\}|\*|) {
            my $param_name = $1;
            $path_params{$param_name} = {
                url_segment_index => $url_segment_index,
                unique_param_name => "${method_name}_${normalized_url_name}_${param_name}",
            };
        }
        $url_name .= '/'. $path_element;
    }

    return {url_name => $url_name, path_params => \%path_params,};
}

#-------------------------------------------------------------------------------

sub verify_swagger_version {
    my ($self) = @_;

    my $content = $self->{openapi_content};
    my $swagger_version;
    if ($content->{openapi}) {
        $swagger_version = $content->{openapi};
    }
    elsif ($content->{swagger}) {
        $swagger_version = $content->{swagger};
    }
    elsif ($content->{swaggerVersion}) {
        $swagger_version = $content->{swaggerVersion};
    }

    my $version_error;
    if ($swagger_version) {
        my ($main_version) = split(/\./, $swagger_version);
        $self->{swagger_main_version} = $main_version;
        if ($main_version < $MIN_SWAGGER_VERSION || $main_version > $MAX_SWAGGER_VERSION) {
            $version_error = "Swagger version $swagger_version is not supported.";
        }
    }
    else {
        $version_error = "Unknown Swagger version.";
    }

    return $version_error;
}

#-------------------------------------------------------------------------------

sub support_version {
    my ($self, $check_version) = @_;

    my $version_supported;
    if ($check_version && $check_version <= $self->{swagger_main_version}) {
        $version_supported = 1;
    }

    return $version_supported;
}

#-------------------------------------------------------------------------------

sub update_path_parameter_handling_to_url {
    my $self = shift;

    my $dbh = $self->{dbh};
    my $session = $self->{session};
    my $call_flags = $self->{call_flags};
    my $policy = $self->{policy};

    my $handle_path_parameter_attribute =  F5::ASMConfig::PolicyAttributes->new({
                                                                                  path_parameter_handling => F5::ASMConfig::PathParameterHandling::PATH_PARAMETER_HANDLING_AS_URL()
                                                                      });
    $self->{handler}->update_policy_attributes($session, $call_flags, $policy, $handle_path_parameter_attribute);
}

#-------------------------------------------------------------------------

1;

