package F5::IpUtils;

use strict;
use warnings;

use F5::IpAddr;
use Socket qw(:DEFAULT inet_pton inet_ntop);

my $DEFAULT_IPV4_MASK = '255.255.255.255';
my $DEFAULT_IPV6_MASK = 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff';

#----------------------------------------------------------------------
sub split_ip_route {
  my  %params = (@_);
  my  $ip_add = $params{ip_address};
  my  $die_on_error = defined($params{die_on_error}) ? $params{die_on_error} : 1;
  my  $norm_without_route = $ip_add;
  my  $route  = 0;

  if (! defined $ip_add) {
      if ($die_on_error) {
          die "IP Address was not specified\n";
      }
      else {
          return {completed_successfully => 0,
                 };
      }
  }

  if ( $ip_add =~ /^(.*)\%([0-9]+)$/g ){
    $norm_without_route = $1;
    $route  = $2;
  }  

  if (! is_ip_address($norm_without_route) ) {
      if ($die_on_error) {
          die "The ip '$ip_add' is not valid. \n";
      }
      else {
          return {completed_successfully => 0,
                 };
      }
  }

  return { completed_successfully => 1,
           ip_address => $norm_without_route,
           route_domain => $route,
         }
}

sub cidr_to_netmask_ipv4 {
    my ($netbit) = @_;

    if ($netbit == 0) {
        return $DEFAULT_IPV4_MASK;
    }

    my $mask =  (1 << (32 - $netbit)) -  1;
    my $full_mask = unpack ("N", pack( "C4", (255) x 4 ));

    return join( '.', unpack( "C4", pack( "N", ( $full_mask ^ $mask ) ) ) );
}

sub cidr_to_netmask_ipv6 {
    my ($netbit) = @_;
    my @hextets;
    my $num_hextets = int($netbit / 16);
    my $end_hextet = $netbit % 16;
    my $end_mask = $end_hextet ? (0xffff ^ ((1 << (16 - $end_hextet)) - 1)) : 0;
    my $ipv6_netmask = '';

    if ($netbit == 0) {
        return $DEFAULT_IPV6_MASK;
    }

    for (my $index = 0; $index < $num_hextets; $index++) {
        push @hextets, 'ffff';
    }

    if ($end_mask) {
        push @hextets, unpack( "H*", pack( "n", ($end_mask) ));
    }

    $ipv6_netmask = join ':', @hextets;

    if ($num_hextets != 8) {
        $ipv6_netmask .= "::";
    }

    return $ipv6_netmask;
}

sub is_netmask_valid_for_address {
    my (%params) = @_;

    my $ip_addr = $params{ip_address};
    my $mask = $params{mask};

    my $ip_bin = F5::IpAddr::ip_pton($ip_addr);
    my $mask_bin = F5::IpAddr::ip_pton($mask);

    my $pref_up = $ip_bin | $mask_bin;
    my $mask_inv = ~$mask_bin;
    my $res_bin = $pref_up & $mask_inv;

    my @res_set = unpack("N*", $res_bin);
    my $is_valid = 1;

    for my $res_entry (@res_set) {
        if ($res_entry > 0) {
            $is_valid = 0;
        }
    }

    return $is_valid;
}

sub is_ip_address {
    my $value = shift;

    return undef unless defined $value;
    return undef if $value =~ /\0/;
    return (defined inet_pton(AF_INET, $value) || defined inet_pton(AF_INET6, $value)) ? 1 : 0;
}

#----------------------------------------------------------------------
sub is_valid_ip {
    my (%params) = @_;

    my $host = $params{field_value};
    if ($host =~ /(.*)\/(?:\d+)/g) {
        $host = $1;
    }

    my $is_valid_ipv4 = (defined inet_pton(AF_INET, $host)) ? 1 : 0;
    my $is_valid_ipv6 = (defined inet_pton(AF_INET6, $host)) ? 1 : 0;

    my $is_valid_ip   = ($is_valid_ipv4 || $is_valid_ipv6);
    my $error_message = $is_valid_ip ? '' : 'The value must be a valid IP address';

    return {
             completed_successfully => $is_valid_ip,
             error_message => $error_message,
           }
}

#----------------------------------------------------------------------
1;

