package F5::JSON::Schema::Validate;

use strict;
use warnings;

use JSON::XS;
use F5::Deploy;
use F5::SystemUtils;
use F5::BigipUtils;

my $JSON_SCHEMA_VALIDATION_TIMEOUT = 10; #sec
my $REFERENCE_ERROR = 2;

# CLI utility -
#   /usr/share/ts/bin/is_json_schema_valid <comma_delimited_schema_pathnames_list>
# returns a list of errors, perl file, in JSON format
my $JSON_SCHEMA_VALIDATION_UTILITY = F5::Deploy::map_path('/usr/share/ts/bin/is_json_schema_valid');

#-------------------------------------------

sub is_json_schema_valid {
    my(%params) = @_;

    my $profile_validation_files = $params{profile_validation_files};
    my $logger = $params{logger};
    my $tolerate_reference_errors = $params{tolerate_reference_errors} || 0;

    my $rv;

    # MBIPWAF-749
    if (! F5::BigipUtils::is_feature_enabled('perform_json_schema_validation')) {
        $rv->{completed_successfully} = 1;
        return $rv;
    }

    if (! -f $JSON_SCHEMA_VALIDATION_UTILITY) {
        my $error_msg = 'The json schema validation utility could not be found on the device';
        $logger->error($error_msg) if $logger;
        $rv->{completed_successfully} = 0;
        $rv->{messages} = [ $error_msg ];
        return $rv;
    }

    # to avoid malformed JSON string error by JSON::XS decode
    my $json_string = "{}";
    my $args = join(',', map { $_->{local_filename} } @$profile_validation_files);

    my @system_tool_params;
    push (@system_tool_params, $JSON_SCHEMA_VALIDATION_UTILITY, $args);
    my $system_tool_res = F5::SystemUtils::run_system_tool(command_line_params => \@system_tool_params,
                                                           timeout => $JSON_SCHEMA_VALIDATION_TIMEOUT,
                                                          );
    if ($system_tool_res && ($system_tool_res->{standard_error} || ! $system_tool_res->{completed_successfully})) {
        if ($logger) {
            my $log_msg =  join (" ", @system_tool_params);
            $logger->error("Error running - '$log_msg'.");
            $logger->trace($system_tool_res);
        }

        $rv->{completed_successfully} = 0;
        $rv->{messages} = [ 'Error occurred while validating JSON schema.' ];

    }

    if ($system_tool_res && $system_tool_res->{completed_successfully} && $system_tool_res->{standard_output}) {
        $json_string = $system_tool_res->{standard_output};
        chomp($json_string);
    }

    my $wjelement_rv;
    eval {
        $wjelement_rv = JSON::XS->new->utf8->decode($json_string);
    };

    if (ref($wjelement_rv->{errors}) eq 'ARRAY' && scalar @{$wjelement_rv->{errors}}) {
        if ((grep { $_->{error} != $REFERENCE_ERROR } @{$wjelement_rv->{errors}}) || !$tolerate_reference_errors) {
            $rv->{completed_successfully} = 0;
            $rv->{messages} = [ 'JSON schema validation failed' ];
        }
        elsif ($system_tool_res->{completed_successfully}) {
            $rv->{completed_successfully} = 1;
            $rv->{messages} = [ 'JSON schema validation found missing references' ];
        }

        push(@{$rv->{messages}}, map { "$_->{error_str}: $_->{details}" } @{$wjelement_rv->{errors}});
        push(@{$rv->{missing_files}}, map { $_->{details} } grep { $_->{error} == $REFERENCE_ERROR } @{$wjelement_rv->{errors}});
    }
    elsif ($system_tool_res->{completed_successfully}) {
        $rv->{completed_successfully} = 1;
        $rv->{messages} = [ 'Successfully validated JSON schema' ];
    }

    return $rv;
}

#-------------------------------------------
1;
