package F5::JSONProfile;

use strict;
use warnings;

use F5::XML::Twig;

use File::Path;
use File::Copy;
use File::stat;
use File::Slurp;
use File::Basename;
use Text::Iconv;
use Digest::MD5;
use F5::Deploy;
use F5::ParseUri;
use F5::DbUtils;
use F5::ContentProfile;
use F5::ASMConfig::Entity::JsonProfile;
use F5::ASMConfig::Entity::JsonValidationFile;
use F5::JSON::Schema::Validate;
use YAML::XS;

my $ARBITRARY_DIVIDER = '__//ARBITRARY_DIVIDER//__';
my $DEFENSE_CONFIG_BASE_DIR = F5::Deploy::map_path("/ts/var/json/defense_config");
my $JSON_SCHEMA_BASE_DIR = F5::Deploy::map_path("/ts/var/json/schemas");

my $MAX_TOTAL_FILESIZE_DEFENSE_PROFILES = 5_000_000;

my $SENSITIVE_JSON_PARAMS_XPATH = '/policy/json/sensitive';

#----------------------------------------------------------------------
sub create_defense_config {
  my(%params) = @_;
  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id};
  my $policy_name_crc = $params{policy_name_crc};
  my $profile_rest_uuid = $params{profile_rest_uuid};


  F5::ContentProfile::clean_old_files(dir => $DEFENSE_CONFIG_BASE_DIR,
                  max_size => $MAX_TOTAL_FILESIZE_DEFENSE_PROFILES);

  my $defense_config_template = <<EndDoc;
<?xml version="1.0" encoding="utf-8"?>
<policy xmlns="http://www.f5.com/ASM/xmlparser.conf"/>
EndDoc

  my $defense_twig = F5::XML::Twig->new(pretty_print => 'indented');
  $defense_twig->parse($defense_config_template);

  my $defense_attribs_sth = $dbh->prepare(<<EndQuery);
  SELECT def_attrib_xpath, policy_def_attrib.def_attrib_value
    FROM PLC.JSON_DEFENSE_ATTRIBUTES def_attrib,
         PLC.PL_JSON_PROFILE_DEFENSE_ATTRIBUTES policy_def_attrib
   WHERE def_attrib.def_attrib_id = policy_def_attrib.def_attrib_id
     AND policy_def_attrib.profile_id = ? AND policy_def_attrib.flg_check_attrib = ? 
ORDER BY def_attrib.def_attrib_id
EndQuery

  $defense_attribs_sth->execute($profile_id, 1);
  while(my $def_attrib = $defense_attribs_sth->fetchrow_hashref) {
    F5::ContentProfile::add_node_with_value( $defense_twig,
                         $def_attrib->{def_attrib_xpath},
                         $def_attrib->{def_attrib_value});
  }

  add_sensitive_params_nodes( defense_twig => $defense_twig, 
                              profile_id => $profile_id,
                              dbh => $dbh);

  my $defense_config_filename = "$DEFENSE_CONFIG_BASE_DIR/json_profile_${policy_name_crc}__${profile_rest_uuid}";

  open(my $defense_config_fh, '>', $defense_config_filename);
  $defense_twig->print($defense_config_fh);
  close $defense_config_fh;

  chmod(0664, $defense_config_filename);

  return {
          completed_successfully => 1,
          messages => [ 'Defense configuration created successfully.' ],
          defense_config_filename => $defense_config_filename,
         };
}
#---------------------------------------------------------------------
sub add_sensitive_params_nodes{
  my (%params) = @_ ;

  my $defense_twig = $params{defense_twig};
  my $profile_id = $params{profile_id};
  my $dbh = $params{dbh};

  my @sensitve_data_elements;


  my $sensitive_params = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} } , $profile_id);
SELECT sensitive_param_name
  FROM PLC.PL_CONTENT_SENSITIVE_PARAMS
 WHERE profile_id = ?
EndQuery

 for my $sense_param (@$sensitive_params){
      push(@sensitve_data_elements, create_sensitive_data(sense_param => $sense_param,
                                                          profile_id => $profile_id, 
                                                          dbh => $dbh))
  }

  return unless @sensitve_data_elements;

  my $sensitive = XML::Twig::Elt->new('sensitive',
                                   { items => scalar @sensitve_data_elements },
                                   @sensitve_data_elements);

  $sensitive->paste(last_child => F5::ContentProfile::add_or_find_node($defense_twig, dirname($SENSITIVE_JSON_PARAMS_XPATH)));
}
#---------------------------------------------------------------------
sub create_sensitive_data {

  my (%params) = @_ ;

  my $sense_param = $params{sense_param};
  my $profile_id = $params{profile_id};
  my $dbh = $params{dbh};
  
  my $name = F5::ContentProfile::convert_to_utf8(string => $sense_param->{sensitive_param_name}, 
                             profile_id => $profile_id,  
                             dbh => $dbh);

   my $new_node = XML::Twig::Elt->new('name' => $name,
                                     undef,
                                    );
}
#----------------------------------------------------------------------
sub clean_unreferenced_profile_files {
  my(%params) = @_;

  for my $dir (($JSON_SCHEMA_BASE_DIR, $DEFENSE_CONFIG_BASE_DIR)) {
    F5::DbUtils::delete_unreferenced_files(dbh => $params{dbh},
                                       logger => $params{logger},
                                       db_file_refs => [
                                                        { table_name => 'PLC.PL_CONTENT_PROFILES',
                                                          field => 'id' },
                                                       ],
                                       db_value_to_filename_conversion_func => sub {
                                         my $id = shift;
                                         my $hash = F5::ASMConfig::Cache::profile_id_to_hash(dbh => $params{dbh}, profile_id => $id);
                                         my $policy_name_crc = $hash->{policy_name_crc};
                                         my $profile_rest_uuid = $hash->{profile_rest_uuid};
                                         return "$dir/json_profile_${policy_name_crc}__" . $profile_rest_uuid;
                                       },
                                       dir => $dir);
  }
}
#----------------------------------------------------------------------
sub handle_validation_files {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $profile_id = $params{profile_id} || 0;
  my $json_profile = $params{json_profile};
  my $policy = $params{policy};
  my $logger = $params{logger};
  my $policy_name_crc = $params{policy_name_crc};

  if (!$json_profile) {
      $json_profile = F5::ASMConfig::Entity::JsonProfile->new(dbh => $dbh,
                                                              get_criteria => {profile_id => $profile_id});
      if (!$json_profile) {
           return {
            completed_successfully => 0,
            messages => [ 'No profile found' ],
           };
      }
      $policy = $json_profile->policy;
  }
  my $profile_rest_uuid = $params{profile_rest_uuid} || $json_profile->{rest_uuid};

  my $json_schema_dir = "$JSON_SCHEMA_BASE_DIR/json_profile_${policy_name_crc}__${profile_rest_uuid}";
  my $validation_cache_file_name = "$json_schema_dir/__json_profile.yaml";
  my $old_validation_cache;
  my $validation_cache;

  if (-f $validation_cache_file_name) {
      $old_validation_cache = YAML::XS::LoadFile($validation_cache_file_name);
  }
  if (-d $json_schema_dir) {
      rmtree($json_schema_dir);
  }

  my $profile_validation_files = $json_profile->{validation_files} || [];

  my $validation_file_count = scalar @$profile_validation_files;
  if ($validation_file_count == 0) {
    return {
            completed_successfully => 1,
            messages => [ 'No validation files for this profile' ],
           };
  }

  mkdir ($json_schema_dir, 0777);

  # move the primary file to the front of the list
  my ($primary_pos) = grep { $profile_validation_files->[$_]->{flg_is_primary_enforced} } 0..$#{ $profile_validation_files };
  unshift (@$profile_validation_files, splice(@$profile_validation_files, $primary_pos, 1));

  $validation_cache->{latest_verification_file_time} = List::Util::max( map { $_->{json_validation_file}->{last_updated} } @$profile_validation_files );
  $validation_cache->{file_count} = $validation_file_count;

  my $md5_calc = Digest::MD5->new;
  my $file_count = 0;
  for my $profile_validation_file (@$profile_validation_files) {
    if (!$profile_validation_file->{json_validation_file}->{file_contents}) {
        $profile_validation_file->{json_validation_file} = F5::ASMConfig::Entity::JsonValidationFile->new(dbh => $dbh,
                                                                                                          get_criteria => $profile_validation_file->{json_validation_file});
    }

    $file_count++;
    $profile_validation_file->{local_filename} = "$json_schema_dir/$file_count.json_schema"; # so that primary schema file is called '1.json_schema'

    # these files remain on the disk, for later use by BD
    File::Slurp::write_file($profile_validation_file->{local_filename}, $profile_validation_file->{json_validation_file}->{file_contents});

    $md5_calc->add($profile_validation_file->{json_validation_file}->{file_contents});

    delete $profile_validation_file->{json_validation_file};
  }

  $validation_cache->{json_md5} = $md5_calc->hexdigest;

  my $rv;
  if (!$old_validation_cache ||
      ($validation_cache->{json_md5} ne $old_validation_cache->{json_md5} ||
       $validation_cache->{file_count} != $old_validation_cache->{file_count} || 
       $validation_cache->{latest_verification_file_time} != $old_validation_cache->{latest_verification_file_time})) {

      $rv = F5::JSON::Schema::Validate::is_json_schema_valid(profile_validation_files => $profile_validation_files,
                                                            logger => $logger,
                                                           );
  }
  else {
      $rv->{completed_successfully} = 1;
  }

  if ($rv->{completed_successfully}) {
     YAML::XS::DumpFile($validation_cache_file_name, $validation_cache);
  }

  return {
          completed_successfully => $rv->{completed_successfully},
          messages => $rv->{messages},
          missing_files => $rv->{missing_files},
          profile_validation_files => $profile_validation_files,
          json_md5 => $validation_cache->{json_md5},
         };
}
#----------------------------------------------------------------------
1;
