package F5::LWP::UserAgent;

use strict;
use warnings;

use LWP::Protocol::http;

use Memoize;

use F5::ServerLookup;
use F5::ParseUri;
use F5::Utils::Net;
use F5::BigipUtils;
use F5::BigipVersionUtils;
use F5::SystemUtils;
use F5::Cluster::Utils;
use File::Basename;

memoize('f5_ssl_opts');

my $local_device_name;
my $is_docker = F5::BigipUtils::is_docker();

use base 'LWP::UserAgent';

sub new {
    my ($class, %params) = @_;
    if (!$params{ssl_opts}{SSL_ca_file} && !$params{ssl_opts}{SSL_ca_path}) {
        my $ssl_opts = f5_ssl_opts();
        $params{ssl_opts} = {%{$params{ssl_opts}}, %$ssl_opts};
    }
    return $class->SUPER::new(%params);
}

my $OPENSSL_PATH_VERSION = '1.1.1';

sub f5_ssl_opts {
    my $rv = {};
    if (F5::BigipUtils::is_feature_enabled('detect_ssl_ca_bundle')) {
        my $output = F5::SystemUtils::run_system_tool(command_line_params => ['openssl', 'version', '-a'])->{standard_output} || '';
        my $ssl_version;
        my $ssl_ca_path;
        my $ssl_ca_default_file;
        if ($output =~ /OPENSSLDIR: "([^"]*)"/) {
            $ssl_ca_path = "$1/certs";
            $ssl_ca_default_file = "$ssl_ca_path/ca-bundle.crt";
        }
        if ($output =~ /OpenSSL ([\d\.]*)/) {
            $ssl_version = $1;
        }
        my $mozilla_ca_path;
        eval {
            require Mozilla::CA;
            $mozilla_ca_path = dirname(Mozilla::CA::SSL_ca_file());
        };
        if ($mozilla_ca_path && $ssl_ca_path && $mozilla_ca_path eq $ssl_ca_path) {
            #Do nothing if the paths are the same.
        }
        elsif ($ssl_version && F5::BigipVersionUtils::compare_versions(version1 => $ssl_version, version2 => $OPENSSL_PATH_VERSION) >= 0 ) {
            $rv->{SSL_ca_path} = $ssl_ca_path;
        }
        elsif ($ssl_ca_default_file && -e $ssl_ca_default_file) {
            $rv->{SSL_ca_file} = $ssl_ca_default_file;
        }
    }
    return $rv;
}

#----------------------------------------------------------------------

sub request {
    my ($self, $request, @params) = @_;

    my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($request->uri);

    my $proxy_params = F5::Utils::Net::get_proxy_settings();

    unless ($proxy_params->{proxy_uri}) {
        my $ip_address = F5::ServerLookup::get_ip_address($domain, $self->timeout);
        if (!$is_docker) {
            if(!defined $local_device_name){
                $local_device_name = F5::Cluster::Utils::get_mcp_val('hostname');
            }
            if ($domain eq $local_device_name) {
                die "Invalid localhost access in external link\n";
            }
            if ($ip_address =~ m/^127\./ || $ip_address =~ m/^0\.0\.0\.0/) {
                die "Invalid localhost access in external link\n";
            }
        }
        local @LWP::Protocol::http::EXTRA_SOCK_OPTS;
        @LWP::Protocol::http::EXTRA_SOCK_OPTS = ( PeerAddr => $ip_address,
                                              SSL_verifycn_name => $domain,
                                              SSL_hostname => $domain,
                                              Host => $domain);
    }

    local $ENV{PERL_LWP_SSL_VERIFY_HOSTNAME} = ($proxy_params->{proxy_uri}) ? 0 : 1;
    local $ENV{HTTP_PROXY} = $proxy_params->{proxy_uri};
    local $ENV{HTTP_PROXY_USERNAME} = $proxy_params->{proxy_username};
    local $ENV{HTTP_PROXY_PASSWORD} = $proxy_params->{proxy_password};
    local $ENV{HTTPS_PROXY} = $proxy_params->{proxy_uri};
    local $ENV{HTTPS_PROXY_USERNAME} = $proxy_params->{proxy_username};
    local $ENV{HTTPS_PROXY_PASSWORD} = $proxy_params->{proxy_password};
    $self->env_proxy();

    return $self->SUPER::request($request, @params);
}

1;
