package F5::LazyInsert;
use strict;
use warnings;

use F5::DbUtils;
use F5::Db::Single;

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{dbh} = $params{dbh};
    $self->{logger} = $params{logger};
    $self->{batch_insert} = $params{batch_insert};
    $self->{default_limit} = $params{default_limit} || {};

    $self->{tables} = {};
}

#----------------------------------------------------------------------

sub handle_table {
    my($self, %params) = @_;

    my $table_name = $params{table_name};
    my $unique_key = $params{unique_key};

    my %unique_key_fields;
    if (ref $unique_key) {
        %unique_key_fields = map { $_ => 1 } @{$params{unique_key}};
        $unique_key = 'record_unique_key';
    }

    if (! $unique_key) {
        ($unique_key) = F5::DbUtils::get_db_fields($self->{dbh}, $params{table_name}, F5::DbUtils::FIND_ONLY_INDEX_FIELDS);
    }
    my %all_fields = map { ($_ => 1) } F5::DbUtils::get_db_fields($self->{dbh}, $params{table_name});
    my %limit;
    if ($params{limit}) {
        %limit = %{$params{limit}};
    }
    elsif ($self->{default_limit}) {
        for my $k (keys %{$self->{default_limit}}) {
            if ($all_fields{$k}) {
                $limit{$k} = $self->{default_limit}->{$k};
            }
        }
    }

    my $fields = ['*'];
    if ($params{ignore_fields}) {
        for my $field (@{$params{ignore_fields}}) {
            delete $all_fields{$field};
        }
        $fields = [ keys %all_fields ];
    }

    $self->{tables}->{$table_name} = { table_name => $table_name,
                                       limit => \%limit,
                                       unique_key_fields => \%unique_key_fields,
                                       unique_key => $unique_key,
                                       fields => $fields,
                                       _was_table_load_done => 0,
                                       data => undef,
                                       delete_missing_on_flush => defined($params{delete_missing_on_flush}) ? $params{delete_missing_on_flush} : 1,
                                     };
    $self->load_table(table => $self->{tables}->{$table_name});
}

#----------------------------------------------------------------------

sub load_table {
    my($self, %params) = @_;

    my $table = $params{table};

    my @where_clauses;
    my @bind_values;

    for my $field (keys %{$table->{limit}}) {
        push(@where_clauses, "$field = ?");
        push(@bind_values, $table->{limit}->{$field});
    }

    my $where_clauses = '';
    if (@where_clauses) {
        $where_clauses = 'WHERE ' . join(' AND ', @where_clauses);
    }

    my $fields = join(',', @{$table->{fields}});

    if (keys %{$table->{unique_key_fields}}) {
        $table->{data} = $self->{dbh}->selectall_arrayref("SELECT $fields FROM $table->{table_name} $where_clauses", { Slice => {} }, @bind_values);
        for my $row (@{$table->{data}}) {
            $row->{$table->{unique_key}} = join(',', map { $row->{$_}}  grep { defined $table->{unique_key_fields}->{$_} } sort { $a <=> $b} keys %$row);
        }
        $table->{data} = { map { $_->{$table->{unique_key}} => $_ } @{$table->{data}} };
    }
    else {
        $table->{data} = $self->{dbh}->selectall_hashref("SELECT $fields FROM $table->{table_name} $where_clauses", $table->{unique_key}, undef, @bind_values);
    }

    $table->{_was_table_load_done} = 1;
}

#----------------------------------------------------------------------

sub insert_row {
    my($self, %params) = @_;

    my $table_name = $params{table_name};
    my $rec = $params{rec};

    my $table = $self->{tables}->{$table_name};
    $table->{_was_insert_done} = 1;

    if (! $table->{_was_table_load_done}) {
        $self->load_table(table => $table);
    }

    if (keys %{$table->{unique_key_fields}}) {
        my $rec_unique_key = join(',', map { $rec->{$_}}  grep { defined $table->{unique_key_fields}->{$_} } sort { $a cmp $b} keys %$rec);
        $rec->{$table->{unique_key}} = $rec_unique_key;
    }

    if (my $table_rec = $table->{data}->{$rec->{$table->{unique_key}}}) {
        # Key exists, compare to see if the values have changed
        $table_rec->{_was_seen} = 1;
        if ($self->are_hashes_equal($table_rec, $rec)) {
            # No change
            return;
        }
        else {
            $self->{batch_insert}->insert_row(%params, do_replace => 1);
        }
    }
    else {
        $self->{batch_insert}->insert_row(%params);
    }
}

#----------------------------------------------------------------------

sub are_hashes_equal {
    my($self, $hash_a, $hash_b, %params) = @_;
    for my $k (keys %$hash_a) {
        next if (substr($k,0,1) eq '_');
        if (defined $hash_a->{$k} != defined $hash_b->{$k} ||
            (defined $hash_a->{$k} && $hash_a->{$k} ne $hash_b->{$k})) {

            return;
        }
    }
    return 1;
}

#----------------------------------------------------------------------

sub flush {
    my($self) = @_;

    $self->{batch_insert}->flush();

    for my $table_name (keys %{$self->{tables}}) {
        my $table = $self->{tables}->{$table_name};
        if (! $table->{_was_table_load_done}) {
            # Short-circuit to force table reload
            next;
        }

        if ($table->{delete_missing_on_flush}) {
            # Delete all non-existent rows
            if (keys %{$table->{unique_key_fields}}) {
                my @deleted_rows = map { $_ } grep { ! $_->{_was_seen} } values %{$table->{data}};
                for my $row (@deleted_rows) {
                    my %where_fields = map { $_ => $row->{$_} } keys %{$table->{unique_key_fields}};
                    F5::DbUtils::delete_rows(dbh => $self->{dbh},
                                             table_name => $table_name,
                                             where_fields => \%where_fields);
                }
            }
            else {
                my @all_delete_keys = map { $_->{$table->{unique_key}} } grep { ! $_->{_was_seen} } values %{$table->{data}};
                my $max_placeholders = F5::DbUtils::get_max_placeholders(dbh => $self->{dbh});
                while (my @delete_keys = splice(@all_delete_keys, 0, $max_placeholders)) {
                    my @where_clauses;
                    my @bind_values;

                    for my $field (keys %{$table->{limit}}) {
                        push(@where_clauses, "$field = ?");
                        push(@bind_values, $table->{limit}->{$field});
                    }

                    local $" = ',';
                    my @placeholders = ('?') x scalar @delete_keys;
                    push(@where_clauses, "$table->{unique_key} IN (@placeholders)");
                    push(@bind_values, @delete_keys);

                    my $where_clauses = join(' AND ', @where_clauses);
                    $self->{dbh}->do("DELETE FROM $table_name WHERE $where_clauses", undef, @bind_values);
                    $table->{_was_delete_done} = 1;
                }
            }
        }

        $self->{_was_table_load_done} = 0;
    }
}

#----------------------------------------------------------------------

sub get_modified_tables {
    my($self) = @_;

    my @modified_tables;
    for my $table_name (keys %{$self->{tables}}) {
        my $table = $self->{tables}->{$table_name};
        if ($table->{_was_insert_done} || $table->{_was_delete_done}) {
            push(@modified_tables, $table_name);
        }
    }

    return @modified_tables;
}

#----------------------------------------------------------------------

1;
