package F5::License;

use strict;
use warnings;

use F5::Utils::Net;
use SOAP::Lite;
use F5::Deploy;
use F5::ASM::COAPI;
use F5::BigipUtils;
#use SOAP::Lite + trace => qw(debug);
use F5::BigipVersionUtils;
use F5::Cluster::Utils;

my $CONF_INI_FILE = F5::Deploy::map_path('/ts/etc/services.ini');
my $LICENSE_FILE    = '/config/bigip.license';
my $ACTIVATION_URN = 'urn:com.f5.license.v5b.ActivationService';
my $ACTIVATION_URL = "/license/services/$ACTIVATION_URN";
my $LICENSE_ASM_STATE_FILE = F5::Deploy::map_path("/ts/var/license/current");

my $ASM_LICENSE_STATE_ASM_ENABLED = "ASM_ENABLED";
my $ASM_LICENSE_STATE_NOT_LICENSED = "ASM_DISABLED";
my $ASM_MODE_ENABLED = 0;
my $ASM_MODE_BYPASS = 2;
my $AWS_ICM_PLATFORM = "Z100a_icm";

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;
  $self->{outbound_call_timeout} = $params{outbound_call_timeout};
  $self->get_reg_key;
  $self->get_platform_id;
}

#----------------------------------------------------------------------

sub is_aws_icm_platform {
  my($self) = @_;

  return ($self->{platform_id} && $self->{platform_id} eq $AWS_ICM_PLATFORM);
}

#----------------------------------------------------------------------

sub get_platform_id {
  my($self) = @_;

  my $ret = open(my $license_fh, '<', $LICENSE_FILE);
  return unless ($ret);
  while(my $line = <$license_fh>) {
    if ($line =~ /^Platform ID :\s+([-\w]+)/) {
      $self->{platform_id} = $1;
    }
  }
}

#----------------------------------------------------------------------

sub get_reg_key {
  my($self) = @_;

  my $ret = open(my $license_fh, '<', $LICENSE_FILE);
  return unless ($ret);
  while(my $line = <$license_fh>) {
    if ($line =~ /^Registration Key :\s+([-\w]+)/) {
      $self->{reg_key} = $1;
    }
  }
}

#----------------------------------------------------------------------

sub get_from_server {
  my($self, %params) = @_;

  eval "use SOAP::Lite;";
  die "Can't load SOAP::Lite" if ($@);
  eval "use Config::IniFiles;";
  die "Can't load Config::IniFiles" if ($@);

  my $config = Config::IniFiles->new(-file => $CONF_INI_FILE);

  $self->get_dossier;
  my $proxy_params = F5::Utils::Net->get_proxy_settings();
  my $proxy_uri = $proxy_params->{proxy_uri};
  my $proxy_username = $proxy_params->{proxy_username};
  my $proxy_password = $proxy_params->{proxy_password};

  local @LWP::Protocol::http::EXTRA_SOCK_OPTS;

  my $host = $self->get_host_from_reg_key(do_dns_lookup => ! $proxy_uri);
  my $host_name = $host;
  if (! $proxy_uri) {
     my $host_ip = F5::ServerLookup::get_ip_address($host, $self->{outbound_call_timeout});
     @LWP::Protocol::http::EXTRA_SOCK_OPTS = ( PeerAddr => $host_ip,
                                                SSL_verifycn_name => $host,
                                                SSL_hostname      => $host,
                                                Host     => $host);
  }
  local $ENV{HTTPS_CA_FILE};
  if ($self->is_production()) {
      $ENV{HTTPS_CA_FILE} = '/config/ssl/ssl.crt/ca-bundle.crt';
  }


  my %proxy_opts;
  $proxy_opts{proxy} = [ 'https' => undef ];
  local $ENV{PERL_LWP_SSL_VERIFY_HOSTNAME} = $proxy_uri ? 0 : 1;
  local $ENV{HTTPS_PROXY} = $proxy_uri;
  local $ENV{HTTPS_PROXY_USERNAME} = $proxy_username;
  local $ENV{HTTPS_PROXY_PASSWORD} = $proxy_password;

  if ($self->{outbound_call_timeout}) {
      $proxy_opts{timeout} = $self->{outbound_call_timeout};
  }

  my $soap_service = SOAP::Lite
    -> readable(1)
    -> deserializer(F5::License::LocalDeserializer->new)
      -> uri ($ACTIVATION_URN)
        -> proxy ("https://$host$ACTIVATION_URL", %proxy_opts)
          -> on_fault(sub { 
                        my($soap, $res) = @_; 
                        die ref $res ? $res->faultdetail : $soap->transport->status, "\n";
                      });
  $soap_service->transport->http_request->header('Host' => $host_name);
  my $bigip_version = F5::BigipVersionUtils::bigip_version();
  $soap_service->transport->http_request->header('User-Agent' => "bigip/$bigip_version");

  my $eula = $params{eula} || '';

  my $request_data = SOAP::Data->value(
                                       SOAP::Data->name("dossier" => $self->{dossier}), 
                                       SOAP::Data->name("eula" => SOAP::Data->type(string => $eula)),
                                       SOAP::Data->name("email" => 'email'),
                                       SOAP::Data->name("firstName" => 'first'),
                                       SOAP::Data->name("lastName" => 'last'),
                                       SOAP::Data->name("companyName" => 'company'),
                                       SOAP::Data->name("phone" => 'phone'),
                                       SOAP::Data->name("jobTitle" => 'job'),
                                       SOAP::Data->name("address" => 'asm'),
                                       SOAP::Data->name("city" => 'city'),
                                       SOAP::Data->name("stateProvince" => 'state'),
                                       SOAP::Data->name("postalCode" => 'post'),
                                       SOAP::Data->name("country" => 'country'),
                                      );

  my $soap_response;
  eval {
      local $SIG{ALRM} = sub { die "Timeout triggered\n" };
      if ($self->{outbound_call_timeout}) {
          alarm $self->{outbound_call_timeout};
      }

      $soap_response = $soap_service->getLicense($request_data);
  };
  my $ex = $@;
  if ($self->{outbound_call_timeout}) {
      alarm 0;
  }

  if ($ex) {
      die "No response from license server\n";
  }

  if ($soap_response->fault) {
    die "SOAP call failed: " . $soap_response->faultcode .  " " . $soap_response->faultstring . "\n";
  }
  else {
    my $result = $soap_response->result;

    if ($result->{state} eq 'EULA_REQUIRED') {
      if ($params{force_eula}) {
        $self->get_from_server(eula => $result->{eula});
      }
    }
    elsif ($result->{state} eq 'LICENSE_RETURNED') {
      $self->{license} = $result->{license};
    }
    else {
      die "Cannot handle state $result->{state}\n";
    }
  }
}

#----------------------------------------------------------------------

sub as_string {
  my($self) = @_;
  return $self->{license};
}

#----------------------------------------------------------------------

sub get_host_from_reg_key {
  my($self, %params) = @_;

  my $do_dns_lookup = (exists $params{do_dns_lookup} ? $params{do_dns_lookup} : 1);
  eval "use Config::IniFiles;";
  die "Can't load Config::IniFiles" if ($@);
  eval "use F5::ServerLookup;";
  die "Can't load F5::ServerLookup" if ($@);

  my $config = Config::IniFiles->new(-file => $CONF_INI_FILE);
  my $server_name;
  if ($self->is_production) {
    # Production key
    $server_name = $config->val('license_soap_service', 'prod_server_name');
  }
  else {
    # Development key
    $server_name = $config->val('license_soap_service', 'dev_server_name');
  }

  return $server_name;

}

#----------------------------------------------------------------------

sub is_production {
  my($self) = @_;

  return ($self->{reg_key} && $self->{reg_key} =~ /^.[A-Z]/);
}

#----------------------------------------------------------------------

sub get_dossier {
  my($self) = @_;

  $self->{dossier} = "";

  if (open (CMD, "-|", "/usr/bin/get_dossier", "-b", $self->{reg_key})) {
    $self->{dossier} = <CMD>;
    close CMD;
  }
}

#----------------------------------------------------------------------

sub get_currently_installed {
  my($self) = @_;

  my $ret = open(my $license_fh, $LICENSE_FILE);
  return unless ($ret);
  my $contents = join('', <$license_fh>);
  close $license_fh;

  return $contents;
}
#----------------------------------------------------------------------

sub get_asm_licensed_state_str {

    my ($self) = @_;


    return (is_asm_licensed()) ? $ASM_LICENSE_STATE_ASM_ENABLED : $ASM_LICENSE_STATE_NOT_LICENSED ;
}
#----------------------------------------------------------------------

sub get_asm_license_mode {
   return (is_asm_licensed()) ? $ASM_MODE_ENABLED : $ASM_MODE_BYPASS ;
}

#----------------------------------------------------------------------

sub is_asm_licensed {
    return F5::BigipUtils::is_docker() || F5::ASM::COAPI::is_fflag_enabled('mod_asm') || is_bewaf_licensed();
}
#----------------------------------------------------------------------
sub is_ts_licensed {
    my ($self) = (@_);
    return (is_asm_licensed());

}
#----------------------------------------------------------------------
sub is_ltm_licensed {
    return F5::BigipUtils::is_docker() || F5::ASM::COAPI::is_fflag_enabled('mod_ltm');
}
#----------------------------------------------------------------------
sub is_iprep_licensed {
    return F5::ASM::COAPI::is_fflag_enabled('subscr_trusted_ip');
}
#----------------------------------------------------------------------
sub is_threat_campaigns_licensed {
    return F5::ASM::COAPI::is_fflag_enabled('subscr_threat_campaigns') || is_bewaf_licensed();
}
#----------------------------------------------------------------------
sub is_bot_signatures_licensed {
    return F5::ASM::COAPI::is_fflag_enabled('subscr_bot_signatures');
}
#----------------------------------------------------------------------
sub is_dos_licensed {
    return F5::ASM::COAPI::is_fflag_enabled('mod_dos');
}
#----------------------------------------------------------------------
sub is_waf_licensed {
    return F5::ASM::COAPI::is_fflag_enabled('mod_waf') || is_bewaf_licensed();
}
#----------------------------------------------------------------------
sub is_credential_stuffing_licensed {
    return F5::ASM::COAPI::is_fflag_enabled('subscr_credential_stuffing') && F5::Cluster::Utils::get_mcp_val('ASM.credential_stuffing_service') eq "enable";
}

#----------------------------------------------------------------------
sub is_bewaf_licensed {
    return F5::Cluster::Utils::get_mcp_val('cloud.licensed.bewaf') eq "enable";
}
#======================================================================

package F5::License::LocalDeserializer;
use strict;
use warnings;

use base qw(SOAP::Deserializer);

my %LICENSE_TYPECAST_TYPES = (
                               '{urn:com.f5.license.v5b}TransactionState' => 1,
                              );

#Implement Typecast for license enumeration Elements
sub typecast {
    my ($self, $value, $name, $attrs, $children, $type) = @_;
    return undef unless $type;
    return unless ( $type && $LICENSE_TYPECAST_TYPES{$type} );
    return $value;
}

#----------------------------------------------------------------------

1;
