package F5::LockFactory::Base;

use File::Basename;
use Fcntl qw/:flock/; # import LOCK_* constants
use F5::Deploy;

our $SHARED_LOCK = LOCK_SH;
our $EXCLUSIVE_LOCK = LOCK_EX;

sub _caller_info {
    my (%params) = @_;
    my $call_depth = $params{call_depth} || 1;
    my ($package, $filename, $line) = caller($call_depth);
    if ($package eq "main") {
        $package = basename($0);
    }
    my $rv;
    if ($params{get_lock_name}) {
        $package =~ s/::/_/g;
        $rv = F5::Deploy::map_path("/ts/lock/${package}_${line}.lock");
    }
    else {
        $rv = "${package} Line ${line}";
    }
    return $rv;
}

sub new {
  my($class, %params) = @_;
  my $lock_filename = _caller_info(get_lock_name => 1);
  my $self = {lock_filename => $lock_filename,
              locks => $params{locks},
              lock_type => $params{lock_type} || $EXCLUSIVE_LOCK,
              logger => $params{logger},
              unlock_only_by_owner => $params{unlock_only_by_owner},
              lock_owner_pid => $$,
              lock_log_level => "info",
              count => 0,};
  bless $self, $class;
  $self->init(%params);
  return $self;
}

sub init_lock {
    my ($self) = @_;
    return;
}

sub add_lock {
    my ($self, $flags) = @_;
    return;
}

sub lock {
    my ($self, %params) = @_;
    my $dont_block = $params{dont_block} || 0;
    my $log_level = $params{log_level} || $self->{lock_log_level};
    my $log_attempt_level = $params{log_attempt_level} || $log_level;
    my $single_lock = $params{single_lock} || 0;

    if (!$self->{locks}->{$self->{lock_filename}}->{count}) {
        eval {
            $self->init_lock();
        };
        if ($@) {
            $self->{logger}->error(_caller_info() .": $@");
            die $@;
        }
    }

    my $flags = $self->{lock_type};
    my $caller_info = _caller_info();
    $self->{logger}->$log_attempt_level("$caller_info is going to lock '$self->{lock_filename}'. (flock flags: '$flags', dont_block: '$dont_block').");

    if (!$self->{count}) {

        if ($dont_block) {
            $flags |= LOCK_NB;
        }

        if (!$self->add_lock($flags)) {
            if ($dont_block) {
                $self->{logger}->$log_attempt_level("$caller_info could not lock '$self->{lock_filename}'.");
            }
            return;
        }
    }
    elsif ($single_lock) {
        return $self->{count};
    }

    $self->{count}++;
    $self->{locks}->{$self->{lock_filename}}->{count}++;

    $self->{logger}->$log_level("$caller_info has locked '$self->{lock_filename}'. (Lock count: '$self->{locks}->{$self->{lock_filename}}->{count}', Self count: '$self->{count}')");
    return $self->{count};
}

sub unlock {
    my ($self, %params) = @_;
    if ($self->{count}) {
        my $log_level = $self->{lock_log_level};
        my $caller_info = _caller_info(%params);
        $self->{logger}->$log_level("$caller_info is going to unlock '$self->{lock_filename}', which is owned by PID '$self->{lock_owner_pid}'.");

        $self->{count}--;
        $self->{locks}->{$self->{lock_filename}}->{count}--;
        if (!$self->{locks}->{$self->{lock_filename}}->{count}) {
           $self->_cleanup();
        }

        $self->{logger}->$log_level("$caller_info has unlocked '$self->{lock_filename}'. (Lock count: '$self->{locks}->{$self->{lock_filename}}->{count}', Self count: '$self->{count}')");
    }
    return $self->{count};
}

sub _cleanup {
    my ($self, %params) = @_;
    return;
}

sub unlock_all {
    my ($self) = @_;
    while ($self->{count}) {
      $self->unlock(call_depth => 2);
    }
}

sub DESTROY {
    my ($self) = @_;
    if (!$self->{unlock_only_by_owner} || $self->{lock_owner_pid} == $$) {
        $self->unlock_all();
    }
}

1;

