package F5::Log::Job;

use strict;
use warnings;

use JSON;
use Digest::MD5;
use Time::HiRes;
use F5::DateTimeUtils;

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }
    if (! $self->{jobCreateDatetime}) {
        $self->add_create_datetime();
    }
    if (! $self->{jobId}) {
        $self->generate_job_id();
    }
    if (! defined $self->{_log_when_status_unchanged}) {
        $self->{_log_when_status_unchanged} = 0;
    }
    # todo: determine default component based on caller
}


#----------------------------------------------------------------------

sub copy {
    my($self, %params) = @_;
    my %copy_params = (%$self, %params);
    delete $copy_params{jobStatus};
    return ref($self)->new(%copy_params);
}

#----------------------------------------------------------------------

sub generate_job_id {
    my($self) = @_;
    my $md5 = Digest::MD5->new;
    $md5->add($self->{jobCreateDatetime});
    $md5->add((Time::HiRes::gettimeofday())[1]);
    $self->{jobId} = $md5->hexdigest;
}

#----------------------------------------------------------------------

sub add_create_datetime {
    my($self) = @_;
    $self->{jobCreateDatetime} = F5::DateTimeUtils::convert_epoch_to_rfc3339(F5::DateTimeUtils::get_time_with_milliseconds());
}

#----------------------------------------------------------------------


sub log_job_status {
    my($self) = @_;

    my $log_rec = $self->to_hash();
    $log_rec->{datetime} = F5::DateTimeUtils::convert_epoch_to_rfc3339(F5::DateTimeUtils::get_time_with_milliseconds());
    my $json = JSON->new->allow_nonref->canonical(1);
    $self->{logger}->info($json->encode($log_rec)) if $self->{logger};
}

#----------------------------------------------------------------------

sub to_hash {
    my($self) = @_;

    my $hash_val = {};
    for my $k (keys %$self) {
        if ($k !~ /^_/ && ! ref($self->{$k})) {
            $hash_val->{$k} = $self->{$k};
        }
    }
    return $hash_val;
}

#----------------------------------------------------------------------

sub set_pending {
    my($self) = @_;

    if ($self->{_log_when_status_unchanged} || ! $self->{jobStatus} || $self->{jobStatus} ne 'pending') {
        $self->{jobStatus} = 'pending';
        $self->log_job_status();
    }
}

#----------------------------------------------------------------------

sub set_started {
    my($self) = @_;

    if ($self->{_log_when_status_unchanged} || ! $self->{jobStatus} || $self->{jobStatus} ne 'started') {
        $self->{jobStatus} = 'started';
        $self->log_job_status();
    }
}

#----------------------------------------------------------------------

sub set_completed {
    my($self) = @_;

    if ($self->{_log_when_status_unchanged} || ! $self->{jobStatus} || $self->{jobStatus} ne 'completed') {
        $self->{jobStatus} = 'completed';
        $self->log_job_status();
    }
}

#----------------------------------------------------------------------

sub set_failed {
    my($self) = @_;

    if ($self->{_log_when_status_unchanged} || ! $self->{jobStatus} || $self->{jobStatus} ne 'failed') {
        $self->{jobStatus} = 'failed';
        $self->log_job_status();
    }
}

#----------------------------------------------------------------------

1;
