package F5::Logger::BD;

use strict;
use warnings;

use base 'F5::Logger';

use F5::Deploy;
use F5::Logger;
use F5::ProcessTable;
use F5::ASMConfig::GuaranteedClient;
use File::Slurp;

our $BD_LOGGER_CFG_FILENAME = F5::Deploy::map_path("/etc/ts/bd/logger.cfg");

# Levels taken from read_logger_cfg.c
my %BD_LOG_LEVELS = (
                     'trace' => 'TS_DEBUG',
                     'debug' => 'TS_INFO',
                     'info' => 'TS_NOTICE',
                     'warn' => 'TS_WARNING',
                     'error' => 'TS_ERR',
                     'fatal' => 'TS_CRIT',
                    );

my @BD_COMMON_MODULES = qw(IO_PLUGIN
                           ECARD_POLICY
                           BD_MISC
                         );

#----------------------------------------------------------------------

sub apply {
    my($self) = @_;

    $self->generate_bd_logger_cfg();
    if (F5::ProcessTable::is_proc_up('bd_agent')) {
        $self->send_log_cfg_to_bd();
    }
}

#----------------------------------------------------------------------

sub generate_bd_logger_cfg {
    my($self) = @_;
    my %env_log_levels = $self->get_env_log_levels();

    if (exists $env_log_levels{$F5::Logger::ALL_MODULES}) {
        $env_log_levels{"bd"} ||= $env_log_levels{$F5::Logger::ALL_MODULES};
    }
    if (exists $env_log_levels{"bd"}) {
        $env_log_levels{"bd.common"} ||= $env_log_levels{"bd"};
        delete $env_log_levels{"bd"};
    }
    if (exists $env_log_levels{"bd.common"}) {
        for my $bd_common_module (@BD_COMMON_MODULES) {
            $env_log_levels{"bd.$bd_common_module"} ||= $env_log_levels{"bd.common"};
        }
        delete $env_log_levels{"bd.common"};
    }

    my $logger_cfg = '';
    for my $k (keys %env_log_levels) {
        if ($k =~ /^bd\.(.*)/) {
            $logger_cfg .= $self->get_bd_log_file_section(module => $1,
                                                          log_level => $env_log_levels{$k},
                                                         );
        }
    }
    $logger_cfg .= $self->get_bd_log_file_section(module => 'ALL',
                                                  log_level => 'info',
                                                 );
    File::Slurp::write_file($BD_LOGGER_CFG_FILENAME, $logger_cfg);
}

#----------------------------------------------------------------------

sub get_bd_log_file_section {
    my($self, %params) = @_;
    my $module = $params{module};
    my $log_level_num = $F5::Logger::LOG_LEVEL_NAME_MAP{$params{log_level}};

    my @log_level_strings;
    for my $log_level (keys %BD_LOG_LEVELS) {
        if ($log_level_num >= $F5::Logger::LOG_LEVEL_NAME_MAP{$log_level}) {
            push(@log_level_strings, $BD_LOG_LEVELS{$log_level});
        }
    }
    my $log_level_string = join(' | ', @log_level_strings);

    return <<EndText;
MODULE = $module;
LOG_LEVEL = $log_level_string;
FILE=2;
EndText
}

#----------------------------------------------------------------------

sub send_log_cfg_to_bd {
    my($self, %params) = @_;

    my $client = F5::ASMConfig::GuaranteedClient::get_client(username => '');
    my $session = F5::ASMConfig::Session->new(user_name => 'logger_bd', session_id => $$);
    my $call_flags = F5::ASMConfig::CallFlags->new(suppress_events => 1);
    my $set_active_flags = F5::ASMConfig::SetActiveFlags->new(update_logger_cfg => 1);

    $client->set_active($session, $call_flags, $set_active_flags);
}

#----------------------------------------------------------------------

1;
