package F5::Mtcl;

use strict;
use warnings;

use C::Include;
use F5::Deploy;
use F5::IpAddr;

# This is a module for transparent BigInteger support for Perl
# uses Math::BigInt under the hood and much simpler to work with
use bigint;
# tbd:sig: because of 'bigint': 'perl -c Mtcl.pm' throws a warning: 
#    'Useless use of a constant in void context at Mtcl.pm line XXX.'
# where line XXX is the line number of the '1;' at the bottom of this package.
# seems harmless but need to look into it...

use constant DEBUG => 0;

my $SHL_32 = 2 ** 32;
my $SHL_64 = 2 ** 64;
my $SHL_96 = 2 ** 96;

my $MSG_SIGNATURE_DEFAULT = 98716425;
my $CONFIG_HEADER_SIZE = 12;
my $COMMON_EVENT_TYPE_STRUCT_SIZE = 38;

# there are 9 variables extracted from the message header, each 32bit - pack string:'L','l','i' (tot: 288bit)
# there are 2 ignored message header chunks, each 80bit (10 bytes) - pack string:'x10' (tot: 160bit)
# overall message header byte size is 56: (288 + 160)/8
my $MSG_HEADER_PACK_STRING = 'L x10 i l x10 l l l l l l'; # 32 + 80 + 64 + 80 + 192 = 448 bit
my $MSG_HEADER_SIZE = 56; # 56 bytes = 448 bit
# these are the 9 variables extracted from the message header
my @MSG_HEADER_UNPACK_ORDER = qw(
                                 msg_signature
                                 msg_type
                                 msg_code
                                 ack_id
                                 msg_data_size
                                 records_received_by_bd
                                 records_deleted_by_bd
                                 records_added_by_bd
                                 records_updated_by_bd
                              );

my $C_INCLUDE_EVENTS_STORABLE = F5::Deploy::map_path('/usr/share/ts/config/t_ev_cache.ph');

# like in typedef enum _COMMON_EVENT_TYPE
#     at TrafficShield/common/include/common_events.h
my %COMMON_EVENT_TYPE = (
    0 => 'EV_TYPE_CORE_SYSTEM', # t_ev_core_system_event
    1 => 'EV_TYPE_CONTENT',
    2 => 'EV_TYPE_CMD',
    3 => 'EV_TYPE_INFO',
    4 => 'EV_TYPE_USER', # NOTE: defined but not in use
    5 => 'EV_TYPE_STAT', # NOTE: defined but not in use
    6 => 'EV_TYPE_SYSLOG',
    8 => 'EV_TYPE_CLIENT_SIDE_DROPPED', # CLIENT_SIDE
);

# same as the last two above but constant for numerical '==' comparison 
use constant {
    EV_TYPE_CLIENT_SIDE_DROPPED => 8,
};

# like in typedef enum client_side_dropped_data_event_code_t
#     at TrafficShield/common/include/core_events.h
my %CLIENT_SIDE_CODE = (
    0 => 'CLIENT_SIDE_DROPPED_DATA_CODE_STATE_CHANGE',
    1 => 'CLIENT_SIDE_DROPPED_DATA_CODE_REPORTING',
);

my %EVENT_TYPE_TO_C_STRUCT = (
    'COMMON_EVENT_TYPE' => 't_ev_common',

    'EV_TYPE_CORE_SYSTEM' => 't_ev_core_system_event',
    'EV_TYPE_CONTENT' => 't_ev_core_content_event',
    'EV_TYPE_CMD' => 't_ev_mng_cmd_event',
    'EV_TYPE_INFO' => 't_ev_mng_info_event',
    'EV_TYPE_SYSLOG' => 't_ev_mng_syslog_event',

    'DOS_CODE_ACCOUNT_STATE_CHANGE' => 't_ev_dos_state_account_event',
    'DOS_CODE_IP_STATE_CHANGE' => 't_ev_dos_state_ip_event',
    'DOS_CODE_ACCOUNT_MONITORING' => 't_ev_dos_entity_monitoring_data',
    'DOS_CODE_IP_MONITORING' => 't_ev_dos_entity_monitoring_data',
    'DOS_CODE_MITIGATION_STATE_CHANGE' => 't_ev_dos_state_mitigation_event',

    'CLIENT_SIDE_DROPPED_DATA_CODE_STATE_CHANGE' => 't_ev_client_side_dropped_data_state_change',
    'CLIENT_SIDE_DROPPED_DATA_CODE_REPORTING' => 't_ev_client_side_dropped_data_reporting',
);

my $event_include;
my %EVENT_STRUCTS = ();

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;
    my $rv = $self->init(%params);
    if ($rv) {
        return $self;
    }
    else {
        return;                 # undef
    }
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    if (! $event_include) {
        $event_include = new C::Include(F5::Deploy::map_path('/usr/local/include/bd_agent_c_include.h'), -cache, $C_INCLUDE_EVENTS_STORABLE );
        for my $c_st_name (values %EVENT_TYPE_TO_C_STRUCT) {
            next if exists $EVENT_STRUCTS{$c_st_name}; # because some events translate to same structs
            $EVENT_STRUCTS{$c_st_name} = $event_include->make_struct($c_st_name);
        }
    }

    if ($params{buf}) {
        my $buf = $params{buf};
        if (length($buf) >= $MSG_HEADER_SIZE) {

            my $header = substr($buf, 0, $MSG_HEADER_SIZE, '');
            my(@elements) = unpack($MSG_HEADER_PACK_STRING, $header);

            for my $i (0..$#elements) {
                $self->{$MSG_HEADER_UNPACK_ORDER[$i]} = $elements[$i];
            }

            # validate msg signature after unpackind
            return unless $self->{msg_signature} == $MSG_SIGNATURE_DEFAULT;

            if ($self->{msg_code} == 1) { # This is a configuration table
                if (length($buf) >= $CONFIG_HEADER_SIZE) {
                    my(@elements) = unpack('i i i', substr($buf, 0, $CONFIG_HEADER_SIZE, ''));
                    $self->{config_table_type} = $elements[0];
                    $self->{config_table_block_size} = $elements[1];
                    $self->{config_table_num_blocks} = $elements[2];
                    $self->{msg_data_size} = $self->{config_table_block_size} * $self->{config_table_num_blocks};
                }
                else {
                    $self->{msg_data_size} = $CONFIG_HEADER_SIZE;
                }
            }

            if (length($buf) >= $self->{msg_data_size}) {
                $self->{data} = substr($buf, 0, $self->{msg_data_size}, '');
            }
            else {
                $self->{buf_needed_length} = $MSG_HEADER_SIZE + $self->{msg_data_size};
                if ($self->{msg_code} == 1) {
                    $self->{buf_needed_length} += $CONFIG_HEADER_SIZE;
                }
            }

            $self->{data} ||= '';
            $self->{buf_remainder} = $buf;

            return $self;
        }
        return;
    }

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }

    return $self;
}

#----------------------------------------------------------------------

sub as_string {
    my($self) = @_;

    $self->{msg_data_size} = length($self->{data});
    $self->{msg_signature} ||= $MSG_SIGNATURE_DEFAULT;

    $self->{records_received_by_bd} ||= 0;
    $self->{records_deleted_by_bd} ||= 0;
    $self->{records_added_by_bd} ||= 0;
    $self->{records_updated_by_bd} ||= 0;

    my $retval = pack($MSG_HEADER_PACK_STRING,
                       @{$self}{@MSG_HEADER_UNPACK_ORDER[0..$#MSG_HEADER_UNPACK_ORDER]});

    $retval .= $self->{data};

    return $retval;
}

#----------------------------------------------------------------------
# this is used both locally by F5::Mtcl::get_event 
# and externally by F5::BdAgent::handle_bd_pipe_message

sub get_event_code {
    my ($self, $buffer) = @_;
    $buffer = $self->{data} unless defined $buffer;

    return unpack('I',substr($buffer, 0, 4));
}

#----------------------------------------------------------------------

sub get_event {
    my ($self) = @_;
    my ($c_struct_name,$event_name);
    my $retval = {};
    my $buffer = $self->{data};

    $event_name = 'COMMON_EVENT_TYPE';
    $c_struct_name = $EVENT_TYPE_TO_C_STRUCT{$event_name};
    $retval->{$event_name} = $self->unpack_event( buffer_ref => \$buffer, c_struct_name => $c_struct_name );

    if ($retval->{$event_name}->{event_type} == EV_TYPE_CLIENT_SIDE_DROPPED) {
        my $ev_code = $self->get_event_code($buffer);
        $event_name = $CLIENT_SIDE_CODE{$ev_code};
    }
    else {
        $event_name = $COMMON_EVENT_TYPE{$retval->{$event_name}->{event_type}};
    }
    $c_struct_name = $EVENT_TYPE_TO_C_STRUCT{$event_name};
    $retval->{$event_name} = $self->unpack_event( buffer_ref => \$buffer, c_struct_name => $c_struct_name);

    $retval->{bd_event} = $event_name;

    return $retval;
}

#----------------------------------------------------------------------

sub unpack_event {
    my ($self, %params) = @_;

    my $c_struct_name = $params{c_struct_name};
    my $buffer_ref = $params{buffer_ref};

    my $struct = $EVENT_STRUCTS{$c_struct_name};
    my $struct_size = $struct->size();
    $struct->unpack(substr($$buffer_ref, 0, $struct_size, ''));

    my $ret_hash_ref = {};
    extract_values($ret_hash_ref, $struct);

    if ( DEBUG ) {
        $ret_hash_ref->{$c_struct_name} = $struct;
    }

    return $ret_hash_ref;
}

#----------------------------------------------------------------------

sub extract_values {
    my ($ret_hash_ref, $struct) = @_;

    for my $entry (keys %{$struct}) {
       next unless $entry;

       # h_ref
       if (ref($struct->{$entry}) eq 'HASH') {
           if (exists $struct->{$entry}->{perl_c_include_uint128}) {
               $ret_hash_ref->{$entry} =
                   $struct->{$entry}->{perl_c_include_uint128}->[0]->{perl_c_include_uint64}->[0]           +
                   $struct->{$entry}->{perl_c_include_uint128}->[0]->{perl_c_include_uint64}->[1] * $SHL_32 +
                   $struct->{$entry}->{perl_c_include_uint128}->[1]->{perl_c_include_uint64}->[0] * $SHL_64 +
                   $struct->{$entry}->{perl_c_include_uint128}->[1]->{perl_c_include_uint64}->[1] * $SHL_96 ;

               $ret_hash_ref->{$entry . '_raw'} = [
                   $struct->{$entry}->{perl_c_include_uint128}->[0]->{perl_c_include_uint64}->[0],
                   $struct->{$entry}->{perl_c_include_uint128}->[0]->{perl_c_include_uint64}->[1],
                   $struct->{$entry}->{perl_c_include_uint128}->[1]->{perl_c_include_uint64}->[0],
                   $struct->{$entry}->{perl_c_include_uint128}->[1]->{perl_c_include_uint64}->[1]
                                                  ];
           } elsif (exists $struct->{$entry}->{perl_c_include_uint64}) {
               $ret_hash_ref->{$entry} =
                   $struct->{$entry}->{perl_c_include_uint64}->[0]           +
                   $struct->{$entry}->{perl_c_include_uint64}->[1] * $SHL_32 ;

               $ret_hash_ref->{$entry . '_raw'} = [
                   $struct->{$entry}->{perl_c_include_uint64}->[0],
                   $struct->{$entry}->{perl_c_include_uint64}->[1]
                                                  ];
           } else {
               my @vals = values %{$struct->{$entry}};
               $ret_hash_ref->{$entry} = $vals[0];
           }
       } 
       # array of h_refs that contain values that are either h_ref (like the one above) OR plain values (like the one below)
       elsif (ref($struct->{$entry}) eq 'ARRAY') {
            my @arr = (); 
            for my $href (@{$struct->{$entry}}) {
                my $t_href = {};
                extract_values($t_href, $href);
                push @arr,$t_href;
            }
            $ret_hash_ref->{$entry} = \@arr;
       } 
       # plain value
       else {
           $ret_hash_ref->{$entry} = $struct->{$entry};
       }

    }

}

1;

