package F5::NATSClient;

use strict;
use warnings;

use Net::NATS::Client;
use File::Slurp;
use MIME::Base64;
use JSON;
use Data::Dumper;


#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }

    my $username_prefix = '';
    if ($self->{username}) {
        $username_prefix = $self->{username};
        if ($self->{password}) {
            $username_prefix .= ":$self->{password}";
        }
        $username_prefix .= '@';
    }

    my $connect_rv;
    eval {
        # CPAN module Net::NATS::Client uses module URI under the hood in order to parse the URI. The
        # username and password cannot be extracted if an unknown protocol is used. Thus, although the
        # actual protocol is NATS internal protocol, this is listed as "ftp" in order for URI parsing to
        # work correctly.
        my %socket_args;
        if ($self->{nats_connect_timeout}) {
            $socket_args{Timeout} = $self->{nats_connect_timeout};
        }
        if ($self->{ssl_ca_file}) {
            $socket_args{SSL_ca_file} = $self->{ssl_ca_file};
        }
        if ($self->{ssl_cert_file}) {
            $socket_args{SSL_cert_file} = $self->{ssl_cert_file};
        }
        if ($self->{ssl_key_file}) {
            $socket_args{SSL_key_file} = $self->{ssl_key_file};
        }
        if ($self->{ssl_version}) {
            $socket_args{SSL_version} = $self->{ssl_version};
        }
        if ($self->{ssl_cipher_list}) {
            $socket_args{SSL_cipher_list} = $self->{ssl_cipher_list};
        }
        $self->{client} = Net::NATS::Client->new(uri => 'ftp://' . "$username_prefix$self->{host}:$self->{port}",
                                                      socket_args => \%socket_args,
                                                         );

        $connect_rv = $self->{client}->connect();
    };

    if ($@ || ! $connect_rv) {
        my $error_msg = $@ || '';
        $error_msg =~ s| at /\S+ line \d+$||;
        if ($error_msg) {
            $error_msg = "($error_msg)";
        }
        print "nats-client|ERR| Couldn't connect to NATS $error_msg\n";
        die $!;
    }
    print "nats-client|INFO| Connected to NATS\n";

    my $subscription = $self->{client}->subscribe($self->{subject}, sub {
                                                      my ($message) = @_;
                                                      $self->handle_message(message => $message->data);
                                                  });
}

#----------------------------------------------------------------------

sub run {
    my($self) = @_;
    while (1) {
        $self->{client}->wait_for_op();
    }
}

#----------------------------------------------------------------------

sub handle_message {
    my($self, %params) = @_;
    my $message = $params{message};

    my $revision = $params{revision};

    my $proto_type = $params{proto_type} || $self->{proto_base_struct};
    print "Handling $proto_type message:" . Dumper($message) . "\n" if $self->{debug};

    my $json = $params{json} || JSON->new->allow_nonref->convert_blessed;;

    my $message_rec = $proto_type->decode($message);
    print "Decoded message:" . Dumper($message_rec) . "\n" if $self->{debug};

    if ($message_rec->{embedded} && $message_rec->{embedded}->{type_url} =~ /declTmm\.(.*)/) {
        my $embedded_proto_type = 'declTmm::' . $1;
        print "Handling embedded $message_rec->{embedded}->{type_url} message\n" if $self->{debug};
        $self->handle_message(revision => $message_rec->{revision}, message => $message_rec->{embedded}->{value}, proto_type => $embedded_proto_type, json => $json);
    }
    else {
        my $pretty_json = $json->pretty->encode($message_rec);
        if ($self->{debug}) {
            my $msg = "$proto_type message";
            $msg = $msg . " (revision $revision)" if defined $revision;
            print $msg . ":\n" . $pretty_json . "\n";
        }
        else {
            print $pretty_json . "\n";
        }
    }
}

#----------------------------------------------------------------------

sub UNIVERSAL::TO_JSON { return { %{ shift() } }; }

1;
