package F5::NegativeSignatures;

use strict;
use warnings;

use F5::BigipVersionUtils;
use JSON;
use F5::DbUtils;
use F5::BigipUtils;

my $NEGSIG_SIGNATURE_ACCURACY_HIGH = 3;

our $PARAM_VALUE_USER_INPUT = 4;

my $GENERIC_SIGSET_ID = 1;
my $GENERIC_HIGH_MEDIUM_ACCURACY_SIGSET_ID = "299999979";
my $GENERIC_HIGH_ACCURACY_SIGSET_ID = "299999978";


sub new {
  my($class, @params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(@params);
  return $self;
}

sub init {
  my($self, @params) = @_;

  my %params;
  if (scalar @params == 1) {
    # Backwards compatible call -- just take dbh
    $params{dbh} = $params[0];
  }
  else {
    %params = @params;
  }

  $self->{dbh} = $params{dbh};
  $self->{dbname} = $params{dbname} || 'PLC';
  $self->{set_signatures_cache} = {};
  $self->{use_memory_cache} = $params{use_memory_cache} || 0;
}

#----------------------------------------------------------------------

sub delete_signatures {
    my($self, $request) = @_;
    require F5::NegativeSignatures::Signature;

    my $results = $self->get_signatures({ 
        filter => $request->{filter}
    });

    for my $sig_id (map { $_->{sig_id} } @{$results->{results}}) {
        F5::NegativeSignatures::Signature->new(
            dbh     => $self->{dbh},
            sig_id  => $sig_id
        )->delete;
    }
}

#----------------------------------------------------------------------
# Multiplexing entry point for all styles of call 
# (all signatures, signatures in a set, signatures in a policy)

sub get_signatures {
  my($self, $request) = @_;

  my @signatures;
  my($order_by_fields, $order_by_order);

  # Apply order-by before limiting selection
  if (exists $request->{order_by}) {
    ($order_by_fields, $order_by_order) = $self->get_order_by_fields($request->{order_by});
  }

  my %calculated_order_by_fields = map { ($_ => 1) } qw(recent_incidents);
  my $noncalculated_order_by_fields = [ grep { !$calculated_order_by_fields{$_} } @$order_by_fields ];

  if($request->{extra_fields}){
   $noncalculated_order_by_fields = [ @$noncalculated_order_by_fields, @{$request->{extra_fields}} ] ;
  }

  if ($request->{policy_id}) {
    @signatures = $self->get_signatures_for_policy($request->{policy_id},
                                                   $request->{filter},
                                                   $noncalculated_order_by_fields);
  }
  elsif ($request->{set_id}) {
    @signatures = $self->get_signatures_for_set(set_id => $request->{set_id},
                                                ext_filter => $request->{filter},
                                                extra_fields => $noncalculated_order_by_fields);
  }
  else {
    if (!exists $request->{filter} || !exists $request->{filter}->{flg_is_hidden}) {
      $request->{filter}->{flg_is_hidden} = 0;
    }
    @signatures = $self->get_all_signatures($request->{filter},
                                            $noncalculated_order_by_fields);
  }

  if (exists($request->{filter}->{recent_incidents}) || grep { $_ eq 'recent_incidents' } @$order_by_fields) {
    $self->add_recent_incidents(\@signatures, $request->{policy_id});
    if (exists($request->{filter}->{recent_incidents})) {
      @signatures = grep {$_->{recent_incidents} == $request->{filter}->{recent_incidents}} @signatures;
    }
  }

  my $total_signatures = scalar @signatures;
  my $start = 0;
  if (@signatures) {
    # Apply order-by before limiting selection
    if (exists $request->{order_by}) {
      @signatures = $self->sort_signatures(\@signatures, $order_by_fields, $order_by_order);
    }

    if ($request->{limit}) {
      if ($request->{start}) {
        $start = $request->{start} - 1;
      } elsif ($request->{start_page}) {
        $start = ($request->{start_page} - 1) * $request->{limit};
      }
      my $end = $start + $request->{limit} - 1;
      $end = $#signatures if ($end > $#signatures);
      @signatures = @signatures[$start..$end];
    }

    # Add all other needed fields to results
    if (exists $request->{fields}) {
      $self->add_signature_fields(\@signatures, $request->{fields}, $request->{policy_id});
      if (@signatures && !$signatures[0]->{sig_references} && grep($_ eq 'sig_references', @{$request->{fields}}) ) {
        $self->map_sig_references(\@signatures);
      }
    }
  }

  my $result = { start => $start + 1,
                 total_found => $total_signatures,
                 results => \@signatures};
  return $result;
}

#----------------------------------------------------------------------

sub get_signatures_for_policy {
  my($self, $policy_id, $ext_filter, $extra_fields, %params) = @_;

  my $dbh = $self->{dbh};
  my $single_sig_id;
  if ($ext_filter && $ext_filter->{sig_id} && ! ref($ext_filter->{sig_id})) {
      $single_sig_id = $ext_filter->{sig_id};
  }

  my $include_previous_enforced = $params{include_previous_enforced} ? 1 : 0;
  my $learning_disabled = $params{learning_disabled};

  my $previous_enforced_rules = $dbh->selectall_hashref(<<EndQuery, 'sig_id', undef, $policy_id);
SELECT sig_id, previous_enforced_rule
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES	
 WHERE previous_enforced_rule_md5 != ''
   AND policy_id = ?
EndQuery
  
  my %policy_signatures;  
  my $policy_sets = $dbh->selectall_hashref(<<EndQuery, 'set_id', undef, $policy_id);
SELECT policy_set.set_id, policy_set.flg_learn, policy_set.flg_alarm, policy_set.flg_reject,
       policy_set.is_inherited,
       negsig_set.set_name, policy_set.rest_uuid
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SETS AS policy_set,
       $self->{dbname}.NEGSIG_SETS AS negsig_set
 WHERE policy_set.policy_id = ?
   AND negsig_set.set_id = policy_set.set_id
EndQuery

  for my $policy_set (values %$policy_sets) {
    my @set_signatures = $self->get_signatures_for_set(set_id => $policy_set->{set_id}, 
                                                       ext_filter => $ext_filter, 
                                                       extra_fields => $extra_fields);

    my $set_rec = { set_id => $policy_set->{set_id},
                    set_name => $policy_set->{set_name} };
    if ($learning_disabled) {
        $policy_set->{flg_learn} = 0;
    }

    my @add_set_for_learn = ();
    if ($policy_set->{flg_learn}) {
        @add_set_for_learn = ($set_rec);
    }
    my @add_set_for_alarm = ();
    if ($policy_set->{flg_alarm}) {
        @add_set_for_alarm = ($set_rec);
    }
    my @add_set_for_reject = ();
    if ($policy_set->{flg_reject}) {
        @add_set_for_reject = ($set_rec);
    }

    for my $sig (@set_signatures) {
      if (! $policy_signatures{$sig->{sig_id}}) {
        $policy_signatures{$sig->{sig_id}} = { %$sig,
                                               reject_sets => [ @add_set_for_reject ],
                                               learn_sets => [ @add_set_for_learn ],
                                               alarm_sets => [ @add_set_for_alarm ],
                                               all_sets => [ $set_rec ],
                                               flg_alarm => $policy_set->{flg_alarm},
                                               flg_reject => $policy_set->{flg_reject},
                                               flg_learn => $policy_set->{flg_learn},
                                               flg_disabled => 0,
                                               flg_quarantined => 0,
                                               extra_has_suggestions => 0,
                                               extra_updated_in_enforcement_readiness_period => 1,
                                               sig_rest_uuid => $sig->{rest_uuid},
                                               is_inherited => $policy_set->{is_inherited},
                                             };
        next;
      }

      push(@{$policy_signatures{$sig->{sig_id}}->{all_sets}}, $set_rec);

      if ($policy_set->{flg_learn}) {
          push(@{$policy_signatures{$sig->{sig_id}}->{learn_sets}}, $set_rec);
          $policy_signatures{$sig->{sig_id}}->{flg_learn} = 1;
      }
      if ($policy_set->{flg_alarm}) {
          push(@{$policy_signatures{$sig->{sig_id}}->{alarm_sets}}, $set_rec);
          $policy_signatures{$sig->{sig_id}}->{flg_alarm} = 1;
      }
      if ($policy_set->{flg_reject}) {
          push(@{$policy_signatures{$sig->{sig_id}}->{reject_sets}}, $set_rec);
          $policy_signatures{$sig->{sig_id}}->{flg_reject} = 1;
      }

      if ($policy_set->{is_inherited}) {
          $policy_signatures{$sig->{sig_id}}->{is_inherited} = 1;
      }
    }
  }

  my %where_fields = ( policy_id => $policy_id );
  if ($single_sig_id) {
      $where_fields{sig_id} = $single_sig_id;
  }
  my $constructed_where_clause;

  my ($flg_quarantine) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id, F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY());
SELECT flg_quarantine_signatures 
  FROM PLC.PL_POLICY_NEGSIG_ATTRIBUTES na 
  JOIN PLC.PL_POLICIES policy ON na.policy_id = policy.id
 WHERE policy_id = ? AND policy_type = ?
EndQuery

  if ($flg_quarantine) {
      my($are_any_sigs_unquarantined) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT sig_id
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES policy_sig
 WHERE policy_id = ?
   AND flg_quarantined = 0
 LIMIT 1
EndQuery

      my $quarantined_sigs;
      $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => { %where_fields,
                                                                                        flg_quarantined => 1 });
      if ($are_any_sigs_unquarantined) {
          $quarantined_sigs = $dbh->selectcol_arrayref(<<EndQuery, undef, @{$constructed_where_clause->{where_values}});
SELECT sig_id
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES policy_sig
 WHERE $constructed_where_clause->{where_clause}
EndQuery
      }
      else {
          $quarantined_sigs = [ keys %policy_signatures ];
      }

      for my $quarantined_sig (@$quarantined_sigs) {
          my $policy_sig = $policy_signatures{$quarantined_sig};
          next unless $policy_sig;
          
          if(exists $previous_enforced_rules->{$quarantined_sig}) {          	 
              if($params{include_previous_enforced}) { 
              	  # save fields involved in claculating the 'blocking_mask' so 'bd' would still enforce the previous enforced rule  
                  $policy_sig->{saved_flg_learn} = $policy_sig->{flg_learn};
                  $policy_sig->{saved_flg_alarm} = $policy_sig->{flg_alarm};
                  $policy_sig->{saved_flg_reject} = $policy_sig->{flg_reject};

                  $policy_sig->{saved_alarm_sets} = $policy_sig->{alarm_sets};
                  $policy_sig->{saved_reject_sets} = $policy_sig->{reject_sets};          	          	          	
              }
              # GUI (in 'search_policy_signature') has to know the prior alarm/reject flags
              $policy_sig->{prior_enforced_alarm} = $policy_sig->{flg_alarm};
              $policy_sig->{prior_enforced_reject} = $policy_sig->{flg_reject};
          }
                              
          $policy_sig->{flg_quarantined} = 1;
          $policy_sig->{flg_learn} = 1;
          $policy_sig->{flg_alarm} = 0;
          $policy_sig->{flg_reject} = 0;

          $policy_sig->{alarm_sets} = [];
          $policy_sig->{reject_sets} = [];
      }
  }

  $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => { %where_fields,
                                                                                    flg_disabled => 1,
                                                                                  });
  my $disabled_sigs = $dbh->selectcol_arrayref(<<EndQuery, undef, @{$constructed_where_clause->{where_values}});
SELECT sig_id
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES policy_sig
 WHERE $constructed_where_clause->{where_clause}
EndQuery

  for my $disabled_sig (@$disabled_sigs) {
      my $policy_sig = $policy_signatures{$disabled_sig};
      next unless $policy_sig;

      $policy_sig->{flg_disabled} = 1;
      $policy_sig->{predisabled_flg_learn} = $policy_sig->{flg_learn};
      $policy_sig->{predisabled_flg_alarm} = $policy_sig->{flg_alarm};
      $policy_sig->{predisabled_flg_reject} = $policy_sig->{flg_reject};

      $policy_sig->{flg_learn} = 0;
      $policy_sig->{flg_alarm} = 0;
      $policy_sig->{flg_reject} = 0;

      $policy_sig->{learn_sets} = [];
      $policy_sig->{alarm_sets} = [];
      $policy_sig->{reject_sets} = [];
  }

  $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => \%where_fields);
  my $rest_uuids = $dbh->selectall_arrayref(<<EndQuery, undef, @{$constructed_where_clause->{where_values}});
SELECT sig_id, rest_uuid
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES policy_sig
 WHERE $constructed_where_clause->{where_clause}
EndQuery

  for my $rest_uuid_rec (@$rest_uuids) {
      if ($policy_signatures{$rest_uuid_rec->[0]}) {
          $policy_signatures{$rest_uuid_rec->[0]}->{rest_uuid} = $rest_uuid_rec->[1];
      }
  }

  if ($flg_quarantine && $extra_fields) { # non-optimized calls will have at least an empty arrayref for $extra_fields
      my $get_excluded_sigs_sql = <<"EndQuery";
SELECT policy_sig.sig_id, FROM_UNIXTIME(policy_sig.last_updated) AS quarantined_since,
       DATE_SUB(NOW(), INTERVAL general.quarantined_period DAY) > FROM_UNIXTIME(policy_sig.last_updated) AS quarantine_period_passed
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES policy_sig
    INNER JOIN $self->{dbname}.PL_POLICIES policy
            ON policy_sig.policy_id = policy.id
    INNER JOIN $self->{dbname}.PL_POLICY_GENERAL general
            ON policy_sig.policy_id = general.policy_id
 WHERE policy_sig.policy_id = ?
   AND policy_sig.flg_quarantined = 1
EndQuery

      my $get_excluded_sigs_sth = $dbh->prepare($get_excluded_sigs_sql);
      $get_excluded_sigs_sth->execute($policy_id);

      while (my $excluded_sig = $get_excluded_sigs_sth->fetchrow_hashref) {
          my $policy_sig = $policy_signatures{$excluded_sig->{sig_id}};
          if ($excluded_sig->{quarantined_since}) {
              $policy_sig->{quarantined_since} = $excluded_sig->{quarantined_since};
              $policy_sig->{quarantine_period_passed} = $excluded_sig->{quarantine_period_passed};
          }

      }
  }

  my $get_param_override_sth = $dbh->prepare(<<EndQuery);
SELECT DISTINCT sig_id, flg_is_enabled
  FROM $self->{dbname}.PL_PARAM_NEGSIG_SIGNATURES
 WHERE policy_id = ?
EndQuery
  $get_param_override_sth->execute($policy_id);

  while (my $param_override = $get_param_override_sth->fetchrow_hashref) {
    if (exists $policy_signatures{$param_override->{sig_id}}) {
      $policy_signatures{$param_override->{sig_id}}->{flg_has_param_override} = 1;
      $policy_signatures{$param_override->{sig_id}}->{flg_has_value_override} = 1;
      if ($param_override->{flg_is_enabled}) {
        $policy_signatures{$param_override->{sig_id}}->{flg_has_value_override_enable} = 1;
        $policy_signatures{$param_override->{sig_id}}->{flg_has_param_override_enable} = 1;
      }
      else {
        $policy_signatures{$param_override->{sig_id}}->{flg_has_param_override_disable} = 1;
      }
    }
  }

  my $get_url_override_sth = $dbh->prepare(<<EndQuery);
SELECT DISTINCT sig_id, flg_is_enabled
  FROM $self->{dbname}.PL_OBJECT_NEGSIG_SIGNATURES
 WHERE policy_id = ?
EndQuery
  $get_url_override_sth->execute($policy_id);

  while (my $url_override = $get_url_override_sth->fetchrow_hashref) {
    if (exists $policy_signatures{$url_override->{sig_id}}) {
      $policy_signatures{$url_override->{sig_id}}->{flg_has_url_override} = 1;
      $policy_signatures{$url_override->{sig_id}}->{flg_has_value_override} = 1;
      if ($url_override->{flg_is_enabled}) {
        $policy_signatures{$url_override->{sig_id}}->{flg_has_value_override_enable} = 1;
        $policy_signatures{$url_override->{sig_id}}->{flg_has_url_override_enable} = 1;
      }
      else {
        $policy_signatures{$url_override->{sig_id}}->{flg_has_url_override_disable} = 1;
      }
    }
  }

  my $get_header_override_sth = $dbh->prepare(<<EndQuery);
SELECT DISTINCT sig_id, flg_is_enabled
  FROM $self->{dbname}.PL_HEADER_NEGSIG_SIGNATURES
 WHERE policy_id = ?
EndQuery
  $get_header_override_sth->execute($policy_id);

  while (my $header_override = $get_header_override_sth->fetchrow_hashref) {
    if (exists $policy_signatures{$header_override->{sig_id}}) {
      $policy_signatures{$header_override->{sig_id}}->{flg_has_header_override} = 1;
      $policy_signatures{$header_override->{sig_id}}->{flg_has_value_override} = 1;
      if ($header_override->{flg_is_enabled}) {
        $policy_signatures{$header_override->{sig_id}}->{flg_has_value_override_enable} = 1;
        $policy_signatures{$header_override->{sig_id}}->{flg_has_header_override_enable} = 1;
      }
      else {
        $policy_signatures{$header_override->{sig_id}}->{flg_has_header_override_disable} = 1;
      }
    }
  }

  my $get_cookie_override_sth = $dbh->prepare(<<EndQuery);
SELECT DISTINCT sig_id, flg_is_enabled
  FROM $self->{dbname}.PL_COOKIE_NEGSIG_SIGNATURES cookie_sig
 WHERE policy_id = ?
EndQuery
  $get_cookie_override_sth->execute($policy_id);

  while (my $cookie_override = $get_cookie_override_sth->fetchrow_hashref) {
    if (exists $policy_signatures{$cookie_override->{sig_id}}) {
      $policy_signatures{$cookie_override->{sig_id}}->{flg_has_cookie_override} = 1;
      if ($cookie_override->{flg_is_enabled}) {
        $policy_signatures{$cookie_override->{sig_id}}->{flg_has_cookie_override_enable} = 1;
      }
      else {
        $policy_signatures{$cookie_override->{sig_id}}->{flg_has_cookie_override_disable} = 1;
      }
    }
  }
  
  # GUI (in 'search_policy_signature') has to know that a prior rule is enforced
  for my $sig_id (keys %{$previous_enforced_rules}) {
      if (exists $policy_signatures{$sig_id}) {
          $policy_signatures{$sig_id}->{is_prior_rule_enforced} = 1;      	
      }
  }
 
  my $get_profile_override_sth = $dbh->prepare(<<EndQuery);
SELECT DISTINCT profile.content_profile_type, 
                profile_sig.sig_id AS sig_id, 
                profile_sig.flg_is_enabled AS flg_is_enabled
  FROM $self->{dbname}.PL_CONTENT_PROFILES profile,
       $self->{dbname}.PL_CONTENT_PROFILE_NEGSIG_SIGNATURES profile_sig
 WHERE profile.policy_id = ?
   AND profile.id = profile_sig.profile_id
EndQuery
  $get_profile_override_sth->execute($policy_id);

  my %content_profile_types = (
                                F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML() => 'xml',
                                F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON() => 'json',
                                F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GWT() => 'gwt',
                                F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_PLAIN_TEXT() => 'plain_text',
                                F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRAPHQL() => 'graphql',
                                F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRPC() => 'grpc',
                              );

  while (my $profile_override = $get_profile_override_sth->fetchrow_hashref) {
      if (exists $policy_signatures{$profile_override->{sig_id}}) {
          my $content_profile_type = $content_profile_types{$profile_override->{content_profile_type}};
          $policy_signatures{$profile_override->{sig_id}}->{"flg_has_${content_profile_type}_override"} = 1;
          $policy_signatures{$profile_override->{sig_id}}->{flg_has_value_override} = 1;
          $policy_signatures{$profile_override->{sig_id}}->{flg_has_content_profile_override} = 1;
          if ($profile_override->{flg_is_enabled}) {
              $policy_signatures{$profile_override->{sig_id}}->{flg_has_value_override_enable} = 1;
              $policy_signatures{$profile_override->{sig_id}}->{"flg_has_${content_profile_type}_override_enable"} = 1;
          }
          else {
              $policy_signatures{$profile_override->{sig_id}}->{"flg_has_${content_profile_type}_override_disable"} = 1;
          }
      }
  }

  if ((grep($_ eq 'extra_has_suggestions', @$extra_fields)) || exists($ext_filter->{has_suggestions})) {
    my $get_suggestions_sth = $dbh->prepare(<<EndQuery);
SELECT DISTINCT suggestions.sig_id as sig_id
  FROM $self->{dbname}.PL_SUGGESTIONS suggestions
 WHERE suggestions.policy_id = ?
   AND suggestions.suggestion_status = 1
EndQuery
    $get_suggestions_sth->execute($policy_id);
    while (my $suggestion_sig = $get_suggestions_sth->fetchrow_hashref) {
        if ($suggestion_sig->{sig_id} && exists $policy_signatures{$suggestion_sig->{sig_id}}) {
            $policy_signatures{$suggestion_sig->{sig_id}}->{extra_has_suggestions} = 1;
        }
    }
  }

  if ((grep($_ eq 'extra_updated_in_enforcement_readiness_period', @$extra_fields)) || exists($ext_filter->{updated_in_enforcement_readiness_period})) {
    my $staging_period = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT quarantined_period
  FROM PLC.PL_POLICY_GENERAL
  WHERE policy_id = ?
EndQuery

    my $now = time();
    my $staging_start_time = $now - ($staging_period * 60*60*24);

    my $sig;
    my $is_updated = 0;

    for my $sig (values %policy_signatures) {
        my $quarantined_since_epoch = F5::DateTimeUtils::convert_datetime_to_epoch($sig->{quarantined_since});
        if ($sig->{flg_quarantined}) {
            $is_updated =  ($quarantined_since_epoch > $staging_start_time) ? 1 : 0;
        }

        $policy_signatures{$sig->{sig_id}}->{extra_updated_in_enforcement_readiness_period} = $is_updated;
    }
  }

 # Filter for blocking masks and excluded signatures
  my @override_flags;
  for my $override_type (qw(param xml json gwt cookie plain_text url header graphql grpc)) {
      push(@override_flags,
           "has_${override_type}_override",
           "has_${override_type}_override_enable",
           "has_${override_type}_override_disable",
          );

  }

  if ($ext_filter && %$ext_filter) {
  SIG_LOOP:
    for my $sig_id (keys %policy_signatures) {
      for my $blocking_filter (('disabled', 'quarantined', @override_flags)) {
        if (exists $ext_filter->{'flg_' . $blocking_filter} &&
            $policy_signatures{$sig_id}->{'flg_' . $blocking_filter} != $ext_filter->{'flg_' . $blocking_filter}) {
          delete $policy_signatures{$sig_id};
          next SIG_LOOP;
        }
      }

      for my $blocking_filter (qw(reject alarm learn)) {
        if (exists $ext_filter->{'flg_' . $blocking_filter}) {

          # For filter purposes, flg_quarantined disables all blocking flags.
          my $sig_flag = ($policy_signatures{$sig_id}->{'flg_' . $blocking_filter} && 
                          ! $policy_signatures{$sig_id}->{'flg_quarantined'});

          if ($sig_flag != $ext_filter->{'flg_' . $blocking_filter}) {
            delete $policy_signatures{$sig_id};
            next SIG_LOOP;
          }
        }
      }

      # Filter for quarantined_since if set
      if (exists $ext_filter->{quarantined_since_val}) {
        unless ($policy_signatures{$sig_id}->{flg_quarantined} &&
                $policy_signatures{$sig_id}->{quarantined_since} &&
                (($ext_filter->{quarantined_since_op} eq 'ge' &&
                  $policy_signatures{$sig_id}->{quarantined_since} ge $ext_filter->{quarantined_since_val}) ||
                 ($ext_filter->{quarantined_since_op} eq 'le' &&
                  $policy_signatures{$sig_id}->{quarantined_since} le $ext_filter->{quarantined_since_val}) ||
                  ($ext_filter->{quarantined_since_op} eq 'eq' &&
                  $policy_signatures{$sig_id}->{quarantined_since} eq $ext_filter->{quarantined_since_val}))) {

          delete $policy_signatures{$sig_id};
          next SIG_LOOP;
        }
      }

      if (exists $ext_filter->{quarantine_period_passed}) {
        unless ($policy_signatures{$sig_id}->{flg_quarantined} &&
                exists $policy_signatures{$sig_id}->{quarantine_period_passed} &&
                $policy_signatures{$sig_id}->{quarantine_period_passed} == $ext_filter->{quarantine_period_passed}) {

          delete $policy_signatures{$sig_id};
          next SIG_LOOP;
        }
      }
      if (exists $ext_filter->{staging_or_tightening_past_period}) {
        unless ($policy_signatures{$sig_id}->{flg_quarantined} &&
                exists $policy_signatures{$sig_id}->{quarantine_period_passed} &&
                exists $policy_signatures{$sig_id}->{extra_has_suggestions} &&
                $policy_signatures{$sig_id}->{quarantine_period_passed} == 1 &&
                $policy_signatures{$sig_id}->{extra_has_suggestions} == 0) {

          delete $policy_signatures{$sig_id};
          next SIG_LOOP;
        }
      }
      if (exists $ext_filter->{has_suggestions}) {
        unless ($policy_signatures{$sig_id}->{extra_has_suggestions} == $ext_filter->{has_suggestions}) {
          delete $policy_signatures{$sig_id};
          next SIG_LOOP;
        }
      }
      if (exists $ext_filter->{updated_in_enforcement_readiness_period}) {
        unless ($policy_signatures{$sig_id}->{extra_updated_in_enforcement_readiness_period} == $ext_filter->{updated_in_enforcement_readiness_period}) {
          delete $policy_signatures{$sig_id};
          next SIG_LOOP;
        }
      }
      if (exists $ext_filter->{rest_uuid}) {
        if ($policy_signatures{$sig_id}->{rest_uuid} ne $ext_filter->{rest_uuid}) {
          delete $policy_signatures{$sig_id};
          next SIG_LOOP;
        }
      }
      if (exists $ext_filter->{is_inherited}) {
          if ($policy_signatures{$sig_id}->{is_inherited} ne $ext_filter->{is_inherited}) {
              delete $policy_signatures{$sig_id};
              next SIG_LOOP;
          }
      }
  }
  }
   
  if ($include_previous_enforced) {
     for my $sig_id (keys %{$previous_enforced_rules}) {
        if (exists $policy_signatures{$sig_id}) {
      	   my $policy_signature = $policy_signatures{$sig_id};      	         	         	         	         	                    	  
      	   my $duplicated_policy_signature = $self->get_previous_enforced_signature(signature => $policy_signature, 
      	                                                                            previous_enforced_rule => $previous_enforced_rules->{$sig_id}->{previous_enforced_rule});
           $policy_signatures{$duplicated_policy_signature->{sig_id}} = $duplicated_policy_signature;   	         	             
        }
     } 
  }
  
  if ($params{get_as_composite_result}) {
     return { signatures => \%policy_signatures,
              sets => $policy_sets,
            };
  }

  return values %policy_signatures;
}

sub get_previous_enforced_signature {
   my($self, %params) = @_;
   
   my $signature = $params{signature};
   my $previous_enforced_rule = $params{previous_enforced_rule};      
                     
   my $duplicated_signature = {%$signature};      	  
   $duplicated_signature->{rule} = $previous_enforced_rule;   
   $duplicated_signature->{sig_id_orig} = $duplicated_signature->{sig_id};

   # Temporary sig_id, used by prepare_negsig_collections within the context of a single policy only
   # Actual ACY rule_id in Collection::generate is overriden by Collection::get_rule_id
   $duplicated_signature->{sig_id} = $duplicated_signature->{sig_id} + 400000000;

   $duplicated_signature->{flg_quarantined} = 0;
    
   # We need to restore the fields involved in calculating the 'blocking_mask' so 'bd' would still enforce the previous enforced rule           
   if(exists $signature->{saved_flg_learn}) {
      $duplicated_signature->{flg_learn} = $signature->{saved_flg_learn};
      $duplicated_signature->{flg_alarm} = $signature->{saved_flg_alarm};
      $duplicated_signature->{flg_reject} = $signature->{saved_flg_reject};
      $duplicated_signature->{alarm_sets} = $signature->{saved_alarm_sets};
      $duplicated_signature->{reject_sets} = $signature->{saved_reject_sets};
   }
             
   return $duplicated_signature;
}

#----------------------------------------------------------------------

sub get_signatures_for_set {
  my ($self,%params) = @_;  
  my $set_id = $params{set_id};
  my $ext_filter = $params{ext_filter};
  my $extra_fields = $params{extra_fields};
  my $dbh = $self->{dbh};
  my $include_previous_enforced = $params{include_previous_enforced} ? 1 : 0;

  if ($self->{use_memory_cache} && ! $include_previous_enforced &&
      (! $ext_filter || ! %$ext_filter) && 
      (! $extra_fields || ! @$extra_fields) && 
      $self->has_cached_set_signatures(set_id => $set_id)) {

      return @{$self->get_cached_set_signatures(set_id => $set_id)};
  }

  my @tables = ("$self->{dbname}.NEGSIG_SIGNATURES AS sig");
  my @where_clauses;
  my @bind_values;

  # Try to get a filter based set
  my $set_filter = $dbh->selectrow_hashref("SELECT * FROM $self->{dbname}.NEGSIG_SET_FILTERS WHERE set_id = ?", undef, $set_id);
  if ($set_filter && $params{reload_cache}) {

    # If there is at least one system for the set filter, include SET_FILTER_SYSTEMS in the join
    if ($dbh->selectrow_array("SELECT set_id FROM $self->{dbname}.NEGSIG_SET_FILTER_SYSTEMS where set_id = ?", undef, 
                              $set_id)) {
      push(@tables,
           "$self->{dbname}.NEGSIG_SIGNATURE_SYSTEMS AS sig_sys",
           "$self->{dbname}.NEGSIG_SET_FILTER_SYSTEMS AS set_sys");
      push(@where_clauses,
           'set_sys.set_id = ?',
           'sig_sys.system_id = set_sys.system_id',
           'sig.sig_id = sig_sys.sig_id');
      push(@bind_values,
           $set_id);
    }

    unless (exists $set_filter->{flg_is_hidden}) {
      $set_filter->{flg_is_hidden} = 0;
    }

    add_filter_to_where_clauses(filter => $set_filter,
                                where_clauses => \@where_clauses,
                                tables => \@tables,
                                bind_values => \@bind_values,
                                dbname => $self->{dbname},
                               );
  }
  else {
    push(@where_clauses,
         "sig.sig_id IN (SELECT sig_id FROM $self->{dbname}.NEGSIG_SET_SIGNATURES WHERE set_id = ?)");
    push(@bind_values, $set_id);
  }
  my @sigs = $self->get_all_signatures($ext_filter,
                                        $extra_fields,                                        
                                        { where_clauses => \@where_clauses,
                                          tables => \@tables,
                                          bind_values => \@bind_values,
                                          include_previous_enforced => $include_previous_enforced,
                                        });
  if ($self->{use_memory_cache} && ! $include_previous_enforced &&
      (! $ext_filter || ! %$ext_filter) && 
      (! $extra_fields || ! @$extra_fields)) {

      $self->set_cached_set_signatures(set_id => $set_id,
                                       sigs => \@sigs);
  }

  return @sigs;
}

#----------------------------------------------------------------------

sub get_signatures_for_sets {
  my ($self,%params) = @_;
  my $set_ids = $params{set_ids};
  my $ext_filter = $params{ext_filter};
  my $extra_fields = $params{extra_fields};
  my $dbh = $self->{dbh};
  my $include_previous_enforced = $params{include_previous_enforced} ? 1 : 0;
  my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => { set_id => $set_ids });
  my @where_clauses = ("sig.sig_id IN (SELECT sig_id FROM $self->{dbname}.NEGSIG_SET_SIGNATURES WHERE $constructed_where_clause->{where_clause})");
  my @sigs = $self->get_all_signatures($ext_filter,
                                       $extra_fields,
                                        { where_clauses => \@where_clauses,
                                          bind_values => $constructed_where_clause->{where_values},
                                          include_previous_enforced => $include_previous_enforced,
                                        });
  return @sigs;
}

#----------------------------------------------------------------------

sub get_all_signatures {
  my($self, $ext_filter, $extra_fields, $params) = @_;

  my $dbh = $self->{dbh};
  my @where_clauses;
  my @tables = ("$self->{dbname}.NEGSIG_SIGNATURES AS sig");
  my @bind_values;
  my @left_join_tables;
  my @left_join_bind_values;

  @where_clauses = @{$params->{where_clauses}} if ($params && exists $params->{where_clauses});
  @tables = @{$params->{tables}} if ($params && exists $params->{tables});
  @bind_values = @{$params->{bind_values}} if ($params && exists $params->{bind_values});

  add_filter_to_where_clauses(filter => $ext_filter,
                              where_clauses => \@where_clauses,
                              tables => \@tables,
                              bind_values => \@bind_values,
                              left_join_tables => \@left_join_tables,
                              left_join_bind_values => \@left_join_bind_values,
                              dbname => $self->{dbname},
                             );

  local $" = ',';
  my @fields = qw(sig.sig_id sig.sig_name
                  sig.sig_tag
                  sig.sig_tag_policy
                  sig.flg_has_valuecontent_http
                  sig.flg_has_valuecontent_xml
                  sig.flg_has_valuecontent_json
                  sig.flg_has_valuecontent_gwt
                  sig.flg_has_headercontent
                  sig.flg_has_uricontent
                  sig.flg_has_requestcontent
                  sig.flg_has_responsecontent
                  sig.rest_uuid
                  sig.apply_to
                  sig.flg_has_cve
                  sig.flg_has_valuecontent_plain_text
                  sig.rule
                  sig.last_enforcement_change
                  sig.threshold_in_seconds
                  sig.buffer_type_mask
                  );
  my %calculated_fields = map(($_ => 1), qw(recent_incidents quarantined_since extra_has_suggestions extra_updated_in_enforcement_readiness_period sig_references systems));
  push(@fields, map { "sig.$_" } grep { !$calculated_fields{$_} } @$extra_fields);
  
  my $where_clause = join(' AND ', @where_clauses);
  $where_clause = '1 = 1' unless $where_clause;
  my $left_join_tables = '';
  if (@left_join_tables) {
    $left_join_tables = join(' ', @left_join_tables);
  }

  my $sql = <<EndQuery;
SELECT DISTINCT @fields
  FROM (@tables)
       $left_join_tables
 WHERE $where_clause
EndQuery

  my $sth = $dbh->prepare($sql);
  $sth->execute(@left_join_bind_values, @bind_values);
  my $signatures = $sth->fetchall_arrayref({});
  if (grep($_ eq 'sig_references', @$extra_fields)) {
    $self->map_sig_references($signatures);
  }

  if($params->{include_previous_enforced}) {
     my $previous_enforced_rules = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT DISTINCT sig_id, previous_enforced_rule
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES
 WHERE previous_enforced_rule_md5 != ''
EndQuery
   
     my %sig_id_to_previous_rules = ();
     for my $sig_id_previous_rule_rec (@$previous_enforced_rules) {
        my $sig_id = $sig_id_previous_rule_rec->{sig_id};
        if(!exists $sig_id_to_previous_rules{$sig_id}) {
           $sig_id_to_previous_rules{$sig_id} = [];
        }         
        push @{$sig_id_to_previous_rules{$sig_id}}, $sig_id_previous_rule_rec->{previous_enforced_rule};
     }
        
     my @duplicated_sigs;          
     for my $sig (@$signatures) {
     	my $sig_id = $sig->{sig_id};
        if (exists $sig_id_to_previous_rules{$sig_id}) {
           for my $previous_enforced_rule (@{$sig_id_to_previous_rules{$sig_id}}) {
              my $duplicated_signature = $self->get_previous_enforced_signature(signature => $sig,
                                                                                previous_enforced_rule => $previous_enforced_rule);
              push @duplicated_sigs, $duplicated_signature;
           }           
        }
     }
     push @$signatures, @duplicated_sigs;
  }
      
  return @$signatures;
}

#----------------------------------------------------------------------

sub add_filter_to_where_clauses {
  my(%params) = @_;
  my $filter = $params{filter};
  my $where_clauses = $params{where_clauses};
  my $bind_values = $params{bind_values};
  my $left_join_tables = $params{left_join_tables};
  my $left_join_bind_values = $params{left_join_bind_values};
  my $tables = $params{tables};
  my $dbname = $params{dbname} || 'PLC';
  my $filter_key;
  local $" = ',';

  my @filterable_keys = qw(
                            apply_to 
                            attack_type_id
                            flg_is_user_defined
                            flg_has_cve
                            risk
                            accuracy
                            sig_id
                            flg_has_valuecontent_http
                            flg_has_valuecontent_xml
                            flg_has_valuecontent_json
                            flg_has_valuecontent_gwt
                            flg_is_hidden
                            system
                            flg_has_headercontent
                            flg_has_uricontent
                            flg_has_requestcontent
                            flg_has_responsecontent
                            extra_has_suggestions
                            flg_has_valuecontent_plain_text
                            extra_updated_in_enforcement_readiness_period
                            sig_tag
                            sig_tag_policy
                         );

  for $filter_key (@filterable_keys) {
    if (defined $filter->{$filter_key}) {
      if (ref($filter->{$filter_key}) eq 'ARRAY') {
        if (@{$filter->{$filter_key}}) {
          my @placeholders = ('?') x scalar @{$filter->{$filter_key}};
          push(@$where_clauses, "sig.$filter_key IN (@placeholders)");
          push(@$bind_values, @{$filter->{$filter_key}});
        }
      }
      else {
        push(@$where_clauses, "sig.$filter_key = ?");
        push(@$bind_values, $filter->{$filter_key});
      }
    }
  }

  for $filter_key (qw(sig_name)) {
    if (defined $filter->{$filter_key}) {
      push(@$where_clauses, "sig.$filter_key LIKE ?");
      push(@$bind_values, "$filter->{$filter_key}");
    }
  }

  my %op_mapping = ('le' => '<=',
                    'eq' => '=',
                    'ge' => '>=');

  for $filter_key (qw(risk accuracy last_update)) {
    if (defined $filter->{$filter_key . '_val'}) {
      my $operator = $op_mapping{$filter->{$filter_key . '_op'}};
      push(@$where_clauses, "sig.$filter_key $operator ?");
      push(@$bind_values, $filter->{$filter_key . '_val'});
    }
  }

  if  (defined $filter->{sig_tag_op}){
      if (  $filter->{sig_tag_op} eq 'untagged' ){
          push(@$where_clauses, "sig.sig_tag = ?");
          push(@$bind_values, "");
      }
      elsif ( $filter->{sig_tag_op} eq 'eq' &&  defined $filter->{sig_tag_val}){
          push(@$where_clauses, "sig.sig_tag = ?");
          push(@$bind_values,  $filter->{sig_tag_val});
      }
      if ($filter->{sig_tag_policy_val}) {
          push(@$where_clauses, "sig.sig_tag_policy = ?");
          push(@$bind_values,  $filter->{sig_tag_policy_val});
      }
  }

  # if sig_references filters exists in search filters, get the relevant field names
  if ( my %sig_references_filters = map { $_ =~ /^sig_references__(.*)$/ ? ($1, $filter->{$_}) : () } keys %$filter ) {
      my %sig_references_fields = ( reference_type => 1,
                                    reference_val => 1,
                                  );

      push(@$left_join_tables, <<EndQuery);
LEFT JOIN $dbname.NEGSIG_SIGNATURE_REFERENCES AS ext_filter_sig_refs
 ON sig.sig_id = ext_filter_sig_refs.sig_id
EndQuery

    # filtering references will only support LIKE comparison for reference_val and 'equals' for reference_type
    for my $sig_references_filter (keys %sig_references_filters) {
      if ( !$sig_references_fields{$sig_references_filter} ) {
          next;
      }
      my $sig_references_field = "ext_filter_sig_refs.$sig_references_filter";
      my $sig_references_field_value = $sig_references_filters{$sig_references_filter};
      if ($sig_references_filter eq 'reference_val') {
          $sig_references_field .= " LIKE ?";
      }
      else {
          $sig_references_field .= " = ?";
      }
      push(@$where_clauses, $sig_references_field);
      push(@$bind_values, $sig_references_field_value);
    }
  }

  if ($filter->{systems}) {
    $filter->{systems} = [ $filter->{systems} ] unless (ref($filter->{systems}));

    if (@{$filter->{systems}}) {
      push(@$tables, "$dbname.NEGSIG_SIGNATURE_SYSTEMS AS ext_filter_sig_sys");
      push(@$where_clauses, 'sig.sig_id = ext_filter_sig_sys.sig_id');

      my @placeholders = ('?') x scalar @{$filter->{systems}};
      push(@$where_clauses,"ext_filter_sig_sys.system_id IN (@placeholders)");
      push(@$bind_values, @{$filter->{systems}});
    }
  }

  if ($filter->{systems_excluded}) {
    $filter->{systems_excluded} = [ $filter->{systems_excluded} ] unless (ref($filter->{systems_excluded}));

    if (@{$filter->{systems_excluded}}) {
      my @placeholders = ('?') x scalar @{$filter->{systems_excluded}};
      push(@$left_join_tables, <<EndQuery);
LEFT JOIN $dbname.NEGSIG_SIGNATURE_SYSTEMS AS ext_filter_exclude_sys
 ON (sig.sig_id = ext_filter_exclude_sys.sig_id AND ext_filter_exclude_sys.system_id IN (@placeholders))
EndQuery

      push(@$left_join_bind_values, @{$filter->{systems_excluded}});
      push(@$where_clauses,"ext_filter_exclude_sys.sig_id IS NULL");
    }
  }

  if ($filter->{support_id}) {
    push(@$tables, "$dbname.PL_SUGGESTIONS AS suggestions");
    push(@$tables, "$dbname.PL_SUGGESTION_SUPPORT_IDS AS support_ids");
    push(@$where_clauses, 'sig.sig_id = suggestions.sig_id');
    push(@$where_clauses, 'suggestions.suggestion_id = support_ids.suggestion_id');
    push(@$where_clauses, 'support_ids.support_id = ?');
    push(@$bind_values, $filter->{support_id});
  }

}

#-----------------------------------------------------------------------

sub add_signature_fields {
  my($self, $signatures, $fields, $policy_id) = @_;
  my $dbh = $self->{dbh};
  local $" = ',';
  my %signatures = map { ($_->{sig_id}, $_) } @$signatures;

  return unless @$signatures;

  my %calculated_fields = map(($_ => 1), qw(recent_incidents extra_has_suggestions extra_updated_in_enforcement_readiness_period sig_references systems));
  my @extra_fields = grep { !$calculated_fields{$_} } @$fields;

  if (@extra_fields) {
    my $where_clause = '';
    my @bind_values;
    if (scalar @$signatures < 25) {
      my @placeholders = ('?') x scalar @$signatures;
      $where_clause = "WHERE sig_id in (@placeholders)";
      @bind_values = keys %signatures;
    }

    my $additional_fields_sth = $dbh->prepare("SELECT sig_id,@extra_fields FROM $self->{dbname}.NEGSIG_SIGNATURES $where_clause");

    $additional_fields_sth->execute(@bind_values);
    while (my $sig = $additional_fields_sth->fetchrow_hashref) {
      if (exists $signatures{$sig->{sig_id}}) {
        @{$signatures{$sig->{sig_id}}}{@extra_fields} = @{$sig}{@extra_fields};
      }
    }
  }

  if (grep { $_ eq 'recent_incidents' } @$fields) {
    $self->add_recent_incidents($signatures, $policy_id);
  }
}

#----------------------------------------------------------------------

sub get_order_by_fields {
  my ($self, $order_by) = @_;

  my @order_by_fields;
  my @order_by_order;
  for my $order_by_input (@$order_by) {
    my($field, $order) = split(/\s+/, lc $order_by_input);
    $order ||= 'asc';
    push(@order_by_fields, $field);
    push(@order_by_order, $order);
  }
  return (\@order_by_fields, \@order_by_order);
}

#----------------------------------------------------------------------

sub sort_signatures {
  my($self, $signatures, $order_by_fields, $order_by_order) = @_;

  my $sort_function = sub {
    for my $i (0..$#$order_by_fields) {
      my $comparison;
      my $order_by_field = $order_by_fields->[$i];
      my $order_by_order_multiplier = ($order_by_order->[$i] eq 'desc' ? -1 : 1);

      if (!defined($a->{$order_by_field})) {
        if (!defined($b->{$order_by_field})) {
          return 0;
        }
        else {
          return -1 * $order_by_order_multiplier;
        }
      }
      elsif (!defined($b->{$order_by_field})) {
        return 1 * $order_by_order_multiplier;
      }


      if (grep($_ eq $order_by_field, qw(sig_id recent_incidents))) {
        # Sort numerically
        $comparison = ($a->{$order_by_field} <=> $b->{$order_by_field}) * $order_by_order_multiplier;
      }
      else {
        # Sort string-wise
        $comparison = (lc($a->{$order_by_field}) cmp lc($b->{$order_by_field})) * $order_by_order_multiplier;
      }
      return $comparison if $comparison;
    }
    return 0;
  };

  return sort $sort_function @$signatures;
}

#----------------------------------------------------------------------

sub get_sets_for_signature {
  my($self, $sig_id,$reload_cache) = @_;

  my $dbh = $self->{dbh};

  my @sig_ids = (ref($sig_id) eq 'ARRAY' ? @$sig_id : $sig_id);

  my @signature_sets;
  # Get all sets to search
  my $all_sets = $dbh->selectall_hashref("SELECT set_id, set_name FROM $self->{dbname}.NEGSIG_SETS", 'set_id');
  for my $set (values %$all_sets) {
    my @found_signatures = $self->get_signatures_for_set(set_id => $set->{set_id},
                                                         ext_filter => { sig_id => \@sig_ids},
                                                         reload_cache => $reload_cache);
    if (@found_signatures) {
      push(@signature_sets, $set);
    }
  }

  return @signature_sets;
}

#----------------------------------------------------------------------
sub link_sigs_to_set{
  my($self, $set, %params) = @_;
  require F5::NegativeSignatures::Set;

  my $dbh = $self->{dbh};
  my @sets;

  if ($set && F5::NegativeSignatures::Set->new(dbh => $dbh, set_id => $set)->flg_is_manual){
    return;
  }

  if ($set) {
      @sets = qq($set);
  }
  else {
      my $filter_based_sets = $dbh->selectcol_arrayref(<<EndQuery);
SELECT set_id
  FROM $self->{dbname}.NEGSIG_SETS
 WHERE flg_is_manual = 0
EndQuery

      @sets = @$filter_based_sets;
  }

  for my $set_id (@sets) {
      my @found_signatures = $self->get_signatures_for_set(set_id => $set_id, reload_cache => 1);

      if (! $set || ! $params{store_type} || $params{store_type} eq 'update') {
          # Check whether records exist before attempting to delete; otherwise, deleting zero rows may cause DB deadlock
          my($existing_row_in_db) = $dbh->selectrow_array("SELECT distinct set_id FROM $self->{dbname}.NEGSIG_SET_SIGNATURES WHERE set_id = ?", undef, $set_id);

          if ($existing_row_in_db) {
              $dbh->do(<<EndQuery,undef,$set_id);
DELETE FROM $self->{dbname}.NEGSIG_SET_SIGNATURES
 WHERE set_id = ?
EndQuery
          }
      }

      my $batch_insert = F5::BatchInsert->new(dbh => $dbh);

      for my $sig (@found_signatures) {
          $batch_insert->insert_row(
                                    table_name => "$self->{dbname}.NEGSIG_SET_SIGNATURES",
                                    rec => {
                                            set_id => $set_id,
                                            sig_id => $sig->{sig_id},
                                           });
      }

      $batch_insert->flush;

  }

}

#----------------------------------------------------------------------

sub link_sets_to_sig{
  my($self, $sig_id, %params) = @_;

  my $reload_cache = 1;
  my @found_sets = $self->get_sets_for_signature($sig_id,$reload_cache);
  my $dbh = $self->{dbh};

  if (! $params{store_type} || $params{store_type} eq 'update') {
      # Check whether records exist before attempting to delete; otherwise, deleting zero rows may cause DB deadlock
      my($existing_row_in_db) = $dbh->selectrow_array("SELECT distinct sig_id FROM $self->{dbname}.NEGSIG_SET_SIGNATURES WHERE sig_id = ?", undef, $sig_id);

      if ($existing_row_in_db) {
          $dbh->do(<<EndQuery,undef,$sig_id);
DELETE FROM $self->{dbname}.NEGSIG_SET_SIGNATURES
 WHERE sig_id = ?
EndQuery
      }
  }

  for my $set (@found_sets){
       F5::DbUtils::insert_row(dbh => $dbh,
                               table_name => "$self->{dbname}.NEGSIG_SET_SIGNATURES",
                               rec => {
                                     set_id => $set->{set_id},
                                     sig_id => $sig_id,
                                });

  }
  
}
#----------------------------------------------------------------------
sub set_flg_quarantined {
  my($self, %params) = @_;

  my $policy_id = $params{policy_id};
  my $sig_id = $params{sig_id};
  my $flg_quarantined = $params{flg_quarantined};

  if (!$sig_id || (ref($sig_id) && !@$sig_id)) {
    return;
  }

  $sig_id = [ $sig_id ] unless (ref($sig_id));
  my $dbh = $self->{dbh};

  local $" = ',';
  
  my $policy_ids = $dbh->selectall_hashref(<<EndQuery, 'policy_id', undef, 1);
SELECT policy_id
  FROM $self->{dbname}.PL_POLICY_NEGSIG_ATTRIBUTES
 WHERE place_sigs_in_staging = ?
EndQuery
  
  if ( !(keys %$policy_ids) ){
    return;
  }

  my %where_fields = (sig_id => $sig_id, flg_disabled => 0);
  
  if ($policy_id && $policy_id ne "*") {
     if ( $policy_ids->{$policy_id} ) {
        $where_fields{policy_id} = $policy_id;
     }
     else {
        return;
     }
  }
  else {
     my @ids = keys %$policy_ids;
     $where_fields{policy_id} = \@ids;
  }
      
  my %rec = (flg_quarantined => $flg_quarantined);
              
  if($flg_quarantined == 0) {
  	$rec{previous_enforced_rule} = '';     
  	$rec{previous_enforced_rule_md5} = '';     
  }
                   
  F5::DbUtils::update_row(dbh => $dbh,
                                  table_name => "$self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES",
                                  rec => \%rec,
                                  where_fields => \%where_fields,
                                 );
}

#----------------------------------------------------------------------
sub add_all_signatures_to_policy {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $policy_id = $params{policy_id};
  my $dbname = $params{dbname} || $self->{dbname};
  my $flg_quarantine_signatures = $params{flg_quarantine_signatures};

  if (!defined($flg_quarantine_signatures)) {
      $flg_quarantine_signatures = 1;
  }

  my $enforce_signatures = $flg_quarantine_signatures ? 0 : 1;
  my $policy_where_clause = '';
  if ($policy_id ne '*') {
    if (ref $policy_id) {
        $policy_id = join(',',@$policy_id);
    }
    $policy_where_clause = "AND pol.id IN ($policy_id)";
  }

  my $time = time();

  $dbh->do(<<EndQuery, undef, $time);
INSERT INTO $dbname.PL_POLICY_NEGSIG_SIGNATURES 
      (policy_id, sig_id, flg_disabled, flg_quarantined, last_updated)
  SELECT pol.id, sig.sig_id, 
         CASE WHEN (sc.scanner_type != '') THEN 1 ELSE 0 END,
         CASE WHEN ($enforce_signatures OR pol.flg_managed_by_bewaf OR sc.scanner_type != '') THEN 0 ELSE 1 END,
         ?
    FROM ($dbname.PL_POLICIES pol,
          $dbname.NEGSIG_SIGNATURES sig)
         LEFT JOIN $dbname.PL_POLICY_NEGSIG_SIGNATURES policy_sig
             ON policy_sig.sig_id = sig.sig_id
            AND policy_sig.policy_id = pol.id
         LEFT JOIN $dbname.PL_SCANNER_CONFIG sc
             ON sc.policy_id = pol.id
   WHERE policy_sig.sig_id IS NULL
$policy_where_clause
EndQuery

  $self->enable_evasion_sigs_on_policies(dbname => $dbname, policy_id_where_clause => $policy_where_clause);

}

#----------------------------------------------------------------------

sub enable_evasion_sigs_on_policies {
    my ($self, %params) = @_;

    return if F5::BigipUtils::is_docker();

    my $dbname = $params{dbname} || 'PLC';
    my $policy_id_where_clause = $params{policy_id_where_clause} || '';

    $self->{dbh}->do(<<EndQuery, undef, $NEGSIG_SIGNATURE_ACCURACY_HIGH, 'Detection Evasion' );
UPDATE $dbname.PL_POLICY_NEGSIG_SIGNATURES pol_sig,
       $dbname.NEGSIG_SIGNATURES sig,
       $dbname.PL_POLICIES pol,
       $dbname.NEGSIG_ATTACK_TYPES nat,
       $dbname.PL_SCANNER_CONFIG sc
   SET pol_sig.flg_disabled = 0, pol_sig.previous_enforced_rule='', pol_sig.previous_enforced_rule_md5=''
 WHERE sc.scanner_type != ''
   AND sc.policy_id = pol.id
   AND accuracy = ?
   AND pol_sig.policy_id = pol.id
   $policy_id_where_clause
   AND nat.attack_type_name = ?
   AND nat.attack_type_id = sig.attack_type_id
   AND pol_sig.sig_id = sig.sig_id
EndQuery
}

#----------------------------------------------------------------------

sub get_policies_using_sets {
  my($self, $set_ids) = @_;

  local $" = ',';
  my $dbh = $self->{dbh};
  $set_ids = [ $set_ids ] unless (ref($set_ids));

  return () unless @$set_ids;
  my @placeholders = ('?') x scalar @$set_ids;
  my $policy_ids = $dbh->selectcol_arrayref(<<EndQuery, undef, @$set_ids);
SELECT DISTINCT policy_id
  FROM $self->{dbname}.PL_POLICY_NEGSIG_SETS
 WHERE set_id IN (@placeholders)
EndQuery

  return @$policy_ids;
}

#----------------------------------------------------------------------

sub get_policies_using_signatures {
  my($self, $sig_ids) = @_;

  return $self->get_policies_using_sets( [ map { $_->{set_id} } $self->get_sets_for_signature($sig_ids) ] );
}

#----------------------------------------------------------------------

sub add_default_sets_to_policy {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $policy_id = $params{policy_id};
  my $policy_negsig_attributes = $params{policy_negsig_attributes};
  my $dbname = $params{dbname} || $self->{dbname};
  my $flg_reject = defined $params{flg_reject} ? $params{flg_reject} : 'sets.flg_reject_default';
  my $flg_learn = defined $params{flg_learn} ? $params{flg_learn} : 'sets.flg_learn_default';

  my $negsig_accuracy_high = F5::ASMConfig::SignatureAccuracy::SIGNATURE_ACCURACY_HIGH();
  my $negsig_accuracy_medium = F5::ASMConfig::SignatureAccuracy::SIGNATURE_ACCURACY_MEDIUM();

  my $generic_sigsets_list_db;
  if (ref($policy_negsig_attributes) && $policy_negsig_attributes->{min_accuracy_for_auto_added_signatures} > 0) {
    my $min_accuracy_for_auto_added_signatures = $policy_negsig_attributes->{min_accuracy_for_auto_added_signatures};
    my $generic_sigset_id = $GENERIC_SIGSET_ID;
    if ($min_accuracy_for_auto_added_signatures >= $negsig_accuracy_high) {
        $generic_sigset_id = $GENERIC_HIGH_ACCURACY_SIGSET_ID;
    }
    elsif ($min_accuracy_for_auto_added_signatures == $negsig_accuracy_medium) {
        $generic_sigset_id = $GENERIC_HIGH_MEDIUM_ACCURACY_SIGSET_ID;
    }

    $generic_sigsets_list_db = $generic_sigset_id;
  }
  else {
    $generic_sigsets_list_db =
                "SELECT IF(pna.min_accuracy_for_auto_added_signatures=$negsig_accuracy_high, $GENERIC_HIGH_ACCURACY_SIGSET_ID,
                        IF(pna.min_accuracy_for_auto_added_signatures=$negsig_accuracy_medium, $GENERIC_HIGH_MEDIUM_ACCURACY_SIGSET_ID, $GENERIC_SIGSET_ID))
                   FROM $dbname.PL_POLICY_NEGSIG_ATTRIBUTES pna
                   WHERE pna.policy_id = pol.id";
  }

  my $policy_where_clause = '';
  if ($policy_id ne '*') {
    if (ref $policy_id) {
        $policy_id = join(',',@$policy_id);
    }
    $policy_where_clause = "AND pol.id IN ($policy_id)";
  }

  my $set_where_clause = "sets.set_id in ($generic_sigsets_list_db)";
  if (!$params{is_in_declarative_export}) {
    $set_where_clause = "$set_where_clause OR sets.flg_assign_to_policy_default = 1";
  }


  my $ignore = F5::DbUtils::get_ignore_string($dbh);
    $dbh->do(<<EndQuery);
INSERT $ignore INTO $dbname.PL_POLICY_NEGSIG_SETS
       (policy_id, set_id, flg_learn, flg_alarm, flg_reject)
SELECT pol.id as policy_id, sets.set_id, $flg_learn, sets.flg_alarm_default, $flg_reject
  FROM $dbname.PL_POLICIES pol,
       $dbname.NEGSIG_SETS sets
 WHERE ($set_where_clause)
$policy_where_clause
EndQuery

}

#----------------------------------------------------------------------

sub hashify {
  return map { ($_, 1) } @_;
}

#----------------------------------------------------------------------

sub map_sig_references {
  my($self, $signatures) = @_;

  my $dbh = $self->{dbh};
  my $reference_url = $dbh->selectcol_arrayref("SELECT reference_type, url_template FROM $self->{dbname}.NEGSIG_REFERENCES", {Columns => [1,2]});
  my %reference_url = @$reference_url;

  for my $sig (@$signatures) {
    my $sig_references;
    if ($sig->{sig_references}) {
        $sig_references = $sig->{sig_references};
    }
    else {
        $sig_references = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $sig->{sig_id});
SELECT reference_type, reference_val
 FROM $self->{dbname}.NEGSIG_SIGNATURE_REFERENCES
 WHERE sig_id = ?
EndQuery
    }

    my @linked_sig_references;
    for my $sig_reference (@$sig_references) {
      unless ($reference_url{$sig_reference->{reference_type}}) {
        # todo: deal with arachnids
#        warn "Unknown reference type: $reference_type\n";
        next;
      }

      $sig_reference->{url} = sprintf($reference_url{$sig_reference->{reference_type}}, $sig_reference->{reference_val});
      push(@linked_sig_references, $sig_reference);
    }
    $sig->{sig_references} = \@linked_sig_references;
  }
}

#----------------------------------------------------------------------

sub add_recent_incidents {
  my($self, $signatures, $policy_id) = @_;

  return if (!@$signatures || exists $signatures->[0]->{recent_incidents});

  my $dbh = $self->{dbh};
  local $" = ',';
  my @sig_ids = map { $_->{sig_id} } @$signatures;
  my @placeholders = ('?') x scalar @sig_ids;

  my $recent_incidents = $dbh->selectcol_arrayref(<<EndQuery, { Columns => [1,2] }, $policy_id, @sig_ids);
  SELECT suggestions.sig_id, COUNT(DISTINCT support_ids.support_id) as recent_incidents
    FROM PLC.PL_SUGGESTIONS suggestions,
         PLC.PL_SUGGESTION_SUPPORT_IDS support_ids
   WHERE suggestions.policy_id = ?
     AND suggestions.suggestion_id = support_ids.suggestion_id
     AND suggestions.sig_id IN (@placeholders)
GROUP BY suggestions.sig_id
EndQuery
  my %recent_incidents = @$recent_incidents;

  for my $sig (@$signatures) {
    $sig->{recent_incidents} = $recent_incidents{$sig->{sig_id}} || 0;
  }
}

#----------------------------------------------------------------------

sub get_update_file_date {
  my($self, %params) = @_;
  my $flg_any_version = defined $params{flg_any_version} ? $params{flg_any_version} : 0;
  my @where_clauses = ('flg_is_user_defined = ?');
  my @bind_values = (0);

  if (!$flg_any_version) {
    push (@where_clauses, 'bigip_version = ?');
    push(@bind_values, F5::BigipVersionUtils::bigip_version());
  }

  my $where_clause = 'WHERE ' . join (' AND ', @where_clauses);
  my($last_update) = $self->{dbh}->selectrow_array(<<EndQuery, undef, @bind_values);
  SELECT sigfile_timestamp 
    FROM $self->{dbname}.SIGFILE_STATUS 
$where_clause
ORDER BY sigfile_status_id DESC
   LIMIT 1
EndQuery
  return $last_update || '';
}

#----------------------------------------------------------------------

sub get_cached_set_signatures {
    my($self, %params) = @_;
    return $self->{set_signatures_cache}->{$params{set_id}};
}

#----------------------------------------------------------------------

sub has_cached_set_signatures {
    my($self, %params) = @_;
    return exists $self->{set_signatures_cache}->{$params{set_id}};
}

#----------------------------------------------------------------------

sub set_cached_set_signatures {
    my($self, %params) = @_;
    $self->{set_signatures_cache}->{$params{set_id}} = $params{sigs};
}

#----------------------------------------------------------------------


1;
