package F5::NegativeSignatures::Collection;

use strict;
use warnings;

use List::Util;
use Config::IniFiles;
use Digest::MD5 qw(md5_hex);
use JSON;

use F5::Deploy;
use F5::NegativeSignatures;
use F5::NegativeSignatures::Collection::Compiler;
use F5::DbUtils;
use F5::CRC;
use F5::Utils::File;
use F5::ThreatCampaigns;
use F5::BotSignatures;
use F5::Browsers;
use F5::License;
use F5::BigipUtils;

my $CFG_FILE = F5::Deploy::map_path('/etc/ts/tools/negsig.cfg');
my $cfg = (-e $CFG_FILE && Config::IniFiles->new(-file => $CFG_FILE));
our $ACCOUNT_DATA_DIR = F5::Deploy::map_path(($cfg && $cfg->val('collection', 'account_data_dir')) || '/ts/var/account');
my $ALL_SIGNATURES_SET_ID = 299999999;

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->{dbh} = $params{dbh};
    $self->{logger} = $params{logger};
    $self->{negative_signatures} = $params{negative_signatures} || F5::NegativeSignatures->new(%params);
    $self->{previous_rule_md5_to_rule_id} = {};
    $self->{browser_rule_id_map} = {};
    $self->{modified_tables} = {};
}

#----------------------------------------------------------------------


sub add_for_account {
    my($self, %params) = @_;
    my $buffer_type_mask = $params{buffer_type_mask};
    my $single_sig_id = $params{single_sig_id};
    my $max_last_enforcement_change = $params{max_last_enforcement_change};

    # All accounts should use a single ACY collection consisting of all signatures currently in use on any policy.
    # To accomplish this, we override the sets with the list of all sets associated with any active policy (or the
    # current policy, which may not be marked active yet).
    # This renders parts of generate() and cleanup() superfluous, since there will only be one active ACY collection
    # at any time. However, this does not impact performance, and therefore, to ease backporting, these sections have
    # been left in.
    my $all_sets_in_use = $self->{dbh}->selectcol_arrayref(<<EndQuery);
SELECT DISTINCT sets.set_id
  FROM PLC.PL_POLICY_NEGSIG_SETS sets,
       PLC.PL_POLICIES pol
 WHERE sets.policy_id = pol.id
   AND pol.account_id > 0
ORDER BY set_id
EndQuery
    my %sets = map { ($_ => 1) } (@{$params{sets}}, @$all_sets_in_use);
    my $sets_json = encode_json([ sort keys %sets ]);

    my $always_disabled_sigs = encode_json($params{always_disabled_sigs});

    if ($single_sig_id) {
        # All signatures other than single signature added/modified are missing from buffer_type_mask
        # Thus, the existing buffer type mask must be retained, with newly added buffer type bits added by OR
        my $existing_account_negsig_collection = $self->{dbh}->selectrow_hashref(<<EndQuery, undef, $params{policy_name_crc});
SELECT *
  FROM DCC.ACCOUNT_NEGSIG_COLLECTIONS
 WHERE policy_name_crc = ?
EndQuery

        $buffer_type_mask->{Request} |= $existing_account_negsig_collection->{request_buffer_type_mask};
        $buffer_type_mask->{Response} |= $existing_account_negsig_collection->{response_buffer_type_mask};
        $max_last_enforcement_change = List::Util::max($existing_account_negsig_collection->{max_last_enforcement_change},
                                                       $max_last_enforcement_change || 0);
    }

    F5::DbUtils::insert_row(dbh => $self->{dbh},
                            table_name => 'DCC.ACCOUNT_NEGSIG_COLLECTIONS',
                            do_replace => 1,
                            rec => { policy_name_crc => $params{policy_name_crc},
                                     request_buffer_type_mask => $params{buffer_type_mask}->{Request},
                                     response_buffer_type_mask => $params{buffer_type_mask}->{Response},
                                     sets_json => $sets_json,
                                     sets_json_md5 => md5_hex($sets_json),
                                     always_disabled_sigs => $always_disabled_sigs,
                                     max_last_enforcement_change => $max_last_enforcement_change,
                                   }
                           );

    # Update all other accounts to use the same composite list of all sets in use.
    # This will ensure a single ACY collection.
    F5::DbUtils::update_row(dbh => $self->{dbh},
                            table_name => 'DCC.ACCOUNT_NEGSIG_COLLECTIONS',
                            rec => {
                                    sets_json => $sets_json,
                                    sets_json_md5 => md5_hex($sets_json),
                                   },
                           );
    $self->{modified_tables}->{"DCC.ACCOUNT_NEGSIG_COLLECTIONS"} = 1;
}

#----------------------------------------------------------------------

sub generate {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    $self->{logger}->debug("Generating signature collections") if $self->{logger};
    $self->pre_cleanup();

    my $threat_campaigns_enabled = F5::License::is_threat_campaigns_licensed();
    my $threat_campaign_rules = {};
    my $threat_campaign_last_updated = 0;
    if ($threat_campaigns_enabled) {
        my $active_campaign_rules_rv = F5::ThreatCampaigns::get_active_campaign_rules(dbh => $dbh, logger => $logger,);
        $threat_campaign_rules = $active_campaign_rules_rv->{campaign_rules};
        $threat_campaign_last_updated = $active_campaign_rules_rv->{last_updated};
    }
    my $bot_signatures_enabled = F5::BigipUtils::is_feature_enabled('bot_defense');
    my $bot_signatures_last_updated = 0;
    my $bot_signatures = [];
    my $browsers_last_updated = 0;
    my $browsers = [];
    if ($bot_signatures_enabled) {
        $bot_signatures = F5::BotSignatures::get_all_bot_signatures(dbh => $dbh, logger => $logger);
        $bot_signatures_last_updated = List::Util::max(map { $_->{last_updated} } @$bot_signatures) || 0;
        $browsers = F5::Browsers::get_user_defined_browsers(dbh => $dbh);
        $browsers_last_updated = List::Util::max(map { $_->{last_updated} } @$browsers) || 0;
    }

    my $max_last_enforcement_change = List::Util::max($threat_campaign_last_updated,
                                                      $bot_signatures_last_updated,
                                                      $browsers_last_updated
                                                     );

    my $set_combinations_to_generate = $self->{dbh}->selectall_hashref(<<EndQuery, 'sets_json_md5', undef, $max_last_enforcement_change, $threat_campaigns_enabled, $bot_signatures_enabled);
SELECT DISTINCT account_collections.sets_json, account_collections.sets_json_md5
  FROM DCC.ACCOUNT_NEGSIG_COLLECTIONS account_collections
LEFT JOIN DCC.NEGSIG_SET_COMBINATIONS set_combos
       ON account_collections.sets_json_md5 = set_combos.sets_json_md5
 WHERE set_combos.negsig_set_combination_id IS NULL
    OR set_combos.max_last_enforcement_change < account_collections.max_last_enforcement_change
    OR set_combos.max_last_enforcement_change < ?
    OR set_combos.threat_campaigns_enabled != ?
    OR set_combos.bot_signatures_enabled != ?
EndQuery

    for my $set_combination (values %$set_combinations_to_generate) {
        my $max_last_enforcement_change = $threat_campaign_last_updated;
        my %sigs_by_apply_to;
        my $sets = decode_json($set_combination->{sets_json});
        for my $set_sig ($self->{negative_signatures}->get_signatures_for_sets(set_ids => $sets, include_previous_enforced => 1, extra_fields => ['packed_rule_obj'])) {
            # Is this the signature with previous enforced rule ?
            if(exists $set_sig->{sig_id_orig}) {
            	$set_sig->{sig_id} = $self->get_rule_id($set_sig->{rule});
            }
            $sigs_by_apply_to{$set_sig->{apply_to}}->{$set_sig->{sig_id}} = $set_sig;
            if ($set_sig->{last_enforcement_change} > $max_last_enforcement_change) {
                $max_last_enforcement_change = $set_sig->{last_enforcement_change};
            }
        }

        for my $threat_campaign_id (keys %$threat_campaign_rules) {
            my $campaign_rules = $threat_campaign_rules->{$threat_campaign_id};
            for my $match_rule (@$campaign_rules) {
                my $sig = {};
                $sig->{rule} = $match_rule->{match_rule};
                $sig->{packed_rule_obj} = $match_rule->{packed_rule_obj};
                $sig->{sig_id} = $match_rule->{sig_id};

                if ($match_rule->{apply_to} == F5::ASMConfig::ApplyTo::APPLY_TO_REQUEST()) {
                    $sig->{apply_to} = "Request";
                }
                elsif ($match_rule->{apply_to} == F5::ASMConfig::ApplyTo::APPLY_TO_RESPONSE()) {
                    $sig->{apply_to} = "Response";
                }

                if ($sig->{apply_to}) {
                    $sigs_by_apply_to{$sig->{apply_to}}->{$sig->{sig_id}} = $sig;
                }
            }
        }

        if ($bot_signatures_enabled) {
            for my $bot_signature (@$bot_signatures) {
                my $sig = $bot_signature;
                $sig->{apply_to} = "Request";
                $sigs_by_apply_to{$sig->{apply_to}}->{$sig->{sig_id}} = $sig;
            }

            for my $browser (@$browsers) {
                my $sig = { apply_to => "Request",
                            rule => $browser->{rule},
                            sig_id => F5::Browsers::get_browser_sig_id(dbh => $dbh, browser => $browser) };
                $sigs_by_apply_to{$sig->{apply_to}}->{$sig->{sig_id}} = $sig;
            }
        }

        $set_combination->{max_last_enforcement_change} = $max_last_enforcement_change;
        $set_combination->{threat_campaigns_enabled} = $threat_campaigns_enabled;
        $set_combination->{bot_signatures_enabled} = $bot_signatures_enabled;

        my $request_sigs = [ map { $sigs_by_apply_to{Request}->{$_} } sort keys %{$sigs_by_apply_to{Request}} ];
        my $response_sigs = [ map { $sigs_by_apply_to{Response}->{$_} } sort keys %{$sigs_by_apply_to{Response}} ];
        $set_combination->{request_sig_collection_filename} = $self->get_compiled_collection_filename(sigs => $request_sigs);
        $set_combination->{response_sig_collection_filename} = $self->get_compiled_collection_filename(sigs => $response_sigs);
	$set_combination->{request_sig_collection_crc} = F5::CRC::get_crc($set_combination->{request_sig_collection_filename});
	$set_combination->{response_sig_collection_crc} = F5::CRC::get_crc($set_combination->{response_sig_collection_filename});
        F5::DbUtils::insert_row(dbh => $self->{dbh},
                                table_name => 'DCC.NEGSIG_SET_COMBINATIONS',
                                rec => $set_combination,
                                do_replace => 1,
                               );
    }

    $self->{dbh}->do(<<EndQuery);
UPDATE DCC.ACCOUNT_NEGSIG_COLLECTIONS
    SET (request_sig_collection_filename, request_sig_collection_crc) = 
                             (SELECT set_combos.request_sig_collection_filename, set_combos.request_sig_collection_crc
                                   FROM DCC.NEGSIG_SET_COMBINATIONS set_combos
                                   WHERE DCC.ACCOUNT_NEGSIG_COLLECTIONS.sets_json_md5 = set_combos.sets_json_md5),
        (response_sig_collection_filename, response_sig_collection_crc) = 
                             (SELECT set_combos.response_sig_collection_filename, set_combos.response_sig_collection_crc
                                   FROM DCC.NEGSIG_SET_COMBINATIONS set_combos
                                   WHERE DCC.ACCOUNT_NEGSIG_COLLECTIONS.sets_json_md5 = set_combos.sets_json_md5)
EndQuery
    $self->{modified_tables}->{"DCC.ACCOUNT_NEGSIG_COLLECTIONS"} = 1;

    $self->cleanup();
    $self->{logger}->debug("Finished generating signature collections") if $self->{logger};
}

#----------------------------------------------------------------------

sub get_compiled_collection_filename {
    my($self, %params) = @_;
    my $sigs = $params{sigs};

    $self->{logger}->trace("Compiling signature collection") if $self->{logger};
    my $compiled_collection = F5::NegativeSignatures::Collection::Compiler::get_compiled_collection(sigs => $sigs);

    my $md5_calc = Digest::MD5->new;
    $md5_calc->add($compiled_collection);
    my $md5_sum = $md5_calc->hexdigest;

    my $collection_filename = "$ACCOUNT_DATA_DIR/negsig/collection.$md5_sum";
    if ( ! -s $collection_filename) {
        F5::Utils::File::write_file(filename => $collection_filename,
                                   write_options => ':bytes',
                                   contents => $compiled_collection,
                                  );
    }

    return $collection_filename;
}

#----------------------------------------------------------------------

sub pre_cleanup {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    # Delete all missing combinations
    my $find_missing_combinations_sql = <<EndQuery;
SELECT negsig_set_combination_id, request_sig_collection_filename, response_sig_collection_filename
  FROM DCC.NEGSIG_SET_COMBINATIONS
EndQuery

    my $combinations = $dbh->selectall_arrayref($find_missing_combinations_sql, { Slice => {} });
    for my $combination (@$combinations) {
        $self->{logger}->trace("Checking signature collections for combination $combination->{negsig_set_combination_id} ($combination->{request_sig_collection_filename}, $combination->{response_sig_collection_filename})") if $self->{logger};
        if (! -s $combination->{request_sig_collection_filename} || ! -s $combination->{response_sig_collection_filename} ) {
            $self->{logger}->info("Signature combination $combination->{negsig_set_combination_id}'s collection ($combination->{request_sig_collection_filename} or $combination->{response_sig_collection_filename}) is missing -- removing") if $self->{logger};
            F5::DbUtils::delete_row(dbh => $dbh,
                                    table_name => 'DCC.NEGSIG_SET_COMBINATIONS',
                                    rec => { negsig_set_combination_id => $combination->{negsig_set_combination_id} },
                                   );
        }
    }
}
#----------------------------------------------------------------------

sub cleanup {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};

    # Delete all unused set combinations
    my $find_unused_set_combos_sql = <<EndQuery;
SELECT DISTINCT negsig_set_combination_id
  FROM DCC.NEGSIG_SET_COMBINATIONS set_combos
    LEFT JOIN DCC.ACCOUNT_NEGSIG_COLLECTIONS account_collections
      ON set_combos.sets_json_md5 = account_collections.sets_json_md5
 WHERE account_collections.policy_name_crc IS NULL
EndQuery

    my $set_combos_to_delete = $dbh->selectcol_arrayref($find_unused_set_combos_sql);

    if (@$set_combos_to_delete) {
        local $" = ',';
        my @placeholders = ('?') x scalar @$set_combos_to_delete;
        $dbh->do("DELETE FROM DCC.NEGSIG_SET_COMBINATIONS WHERE negsig_set_combination_id IN (@placeholders)", undef, @$set_combos_to_delete);
    }

    F5::DbUtils::delete_unreferenced_files(dbh          => $dbh,
                                           logger       => $self->{logger},
                                           db_file_refs => [
                                                            { table_name => 'DCC.ACCOUNT_NEGSIG_COLLECTIONS',
                                                              field      => 'request_sig_collection_filename',
                                                            },
                                                            { table_name => 'DCC.ACCOUNT_NEGSIG_COLLECTIONS',
                                                              field      => 'response_sig_collection_filename',
                                                            },
                                                           ],
                                           dir          => "$F5::NegativeSignatures::Collection::ACCOUNT_DATA_DIR/negsig" );
                                                               
                    
    $dbh->do(<<EndQuery);
DELETE FROM PLC.NEGSIG_PREVIOUS_ENFORCED_RULES
WHERE rule_md5 NOT IN (
   SELECT DISTINCT previous_enforced_rule_md5 
   FROM PLC.PL_POLICY_NEGSIG_SIGNATURES
   )
EndQuery

    if ( F5::BigipUtils::is_feature_enabled('bot_defense') ) {
        F5::Browsers::cleanup(dbh => $dbh);
    }
}

#----------------------------------------------------------------------

sub get_rule_id {
   my($self, $rule) = @_;
  
   my $md5_calc = Digest::MD5->new;
   $md5_calc->add($rule);
   my $rule_md5 = $md5_calc->hexdigest;
   
   if(exists $self->{previous_rule_md5_to_rule_id}->{$rule_md5}) {
      return $self->{previous_rule_md5_to_rule_id}->{$rule_md5};
   }
     
   my $rule_id = $self->{dbh}->selectrow_array(<<EndQuery, undef, $rule_md5);
SELECT rule_id
  FROM PLC.NEGSIG_PREVIOUS_ENFORCED_RULES
 WHERE rule_md5 = ?
EndQuery
    
   if(! $rule_id) {            
      F5::DbUtils::insert_row(dbh => $self->{dbh},
                           table_name => "PLC.NEGSIG_PREVIOUS_ENFORCED_RULES",
                           rec => {
                                     rule => $rule,
                                     rule_md5 => $rule_md5,
                                  },
                           last_insert_id => \$rule_id,
                             );
   }
                                           
   $self->{previous_rule_md5_to_rule_id}->{$rule_md5} = $rule_id + 600000000;
                            
   return $self->{previous_rule_md5_to_rule_id}->{$rule_md5};
}

#----------------------------------------------------------------------

sub get_modified_tables {
    my($self) = @_;

    return keys %{$self->{modified_tables}};
}

#----------------------------------------------------------------------

1;
