package F5::NegativeSignatures::Collection::Compiler;

use strict;
use warnings;

use F5::NegativeSignatures::Rule ':constants';
use F5::PatternMatching;

use Digest::MD5 qw(md5_hex);
use F5::Utils::Encode;
use List::Util;

use constant LARGE_NUM_STATES => 104000;
use constant MAX_UNSIGNED_INT => 4294967295;
use constant MAX_INT => 2147483647;

my $arbitrary_delimiter = chr(1) . chr(2) . chr(3);

#----------------------------------------------------------------------

sub get_compiled_collection {
    my(%params) = @_;
  
    my $sigs = $params{sigs};
    my $sig_type = $params{sig_type} || 'asm';
    my $logger = $params{logger};

    my @keywords;
    my @rules;

    for my $sig (@$sigs) {
        if ($sig->{packed_rule_obj}) {
            eval {
                my $packed_rule_obj = delete $sig->{packed_rule_obj};
                $sig->{rule_obj} = F5::NegativeSignatures::Rule->new(packed_rule_obj => $packed_rule_obj, 
                                                                     sig_id => $sig->{sig_id} );
            };
            if ($@) {
                 $logger->error("Failed to unpack rule for signature id $sig->{sig_id}") if $logger;
            }
        }

        if (! $sig->{rule_obj}) {
            eval {
                $sig->{rule_obj} = F5::NegativeSignatures::Rule->new(sig_id => $sig->{sig_id}, 
                                                                     rule_string => $sig->{rule},
                                                                     sig_type => $sig_type,
                                                                    );
            };
            if ($@) {
                $logger->error("Failed to compile rule \"$sig->{rule}\" for signature id $sig->{sig_id} -- skipping") if $logger;
                next;
            }
        }

        my $rule = { rule_obj => $sig->{rule_obj},
                     regexp_keyword_indexes => [] };
        push(@rules, $rule);
    }

    for my $rule_index (0..$#rules) {
        my $rule = $rules[$rule_index];

        for my $keyword (@{$rule->{rule_obj}->keywords}) {
            $keyword->{rule_id} = [ $rule_index ];
            if ($keyword->{regexp_type} && $keyword->{nocase}) {
                $keyword->{text} = '(?i)' . $keyword->{text};
            }
        }

        my $start_relative_index_assign = $#keywords + 1;
        push(@keywords, @{$rule->{rule_obj}->keywords});

        for my $i ($start_relative_index_assign..$#keywords) {
            if ($keywords[$i]->{is_relative}) {
                $keywords[$i]->{relative_keyword_ref} = $keywords[$i - 1];
                $keywords[$i-1]->{prerequisite_keyword} = 1;
            }

            if ($keywords[$i]->{regexp_type}) {
                if ($keywords[$i]->{is_relative}) {
                    $keywords[$i-1]->{regexp_following_keyword_ref} = $keywords[$i];
                }
            }
        }
    }

    # Try to consolidate identical keywords
    my @keyword_compare_fields = qw(
                                       text
                                       buffer_type
                                       regexp_type
                                       nocase
                                       negation
                                       offset
                                       depth
                                       distance
                                       within
                                       isdataat_relative
                                       prerequisite_keyword
                                       relative_keyword_ref
                                       regexp_following_keyword_ref
                                  );
    my %keyword_ref_fields = ( relative_keyword_ref => 'relative_keyword_index',
                               regexp_following_keyword_ref => 'regexp_following_keyword_index');

    my $keyword_index = 0;
    my %keyword_checksums;
    for my $keyword (@keywords) {
        $keyword->{md5} = md5_hex(join($arbitrary_delimiter, map { $_ || '' } @{$keyword}{@keyword_compare_fields}));
    }
    for my $keyword (@keywords) {
        for my $keyword_ref_field (keys %keyword_ref_fields) {
            if ($keyword->{$keyword_ref_field}) {
                $keyword->{$keyword_ref_field} = $keyword->{$keyword_ref_field}->{md5};
            }
        }
    }

    while ($keyword_index <= $#keywords) {
        my $keyword_md5 = $keywords[$keyword_index]->{md5};
        if (exists $keyword_checksums{$keyword_md5}) {
            # Matching keyword found
            my $source_keyword = $keywords[$keyword_index];
            my $target_keyword = $keywords[$keyword_checksums{$keyword_md5}];

            push(@{$target_keyword->{rule_id}}, @{$source_keyword->{rule_id}});
            splice(@keywords, $keyword_index, 1);
        }
        else {
            $keyword_checksums{$keyword_md5} = $keyword_index;
            $keyword_index++;
        }
    }

    for my $i (0..$#keywords) {
        $keywords[$i]->{keyword_index} = $i;
    }
    for my $keyword (@keywords) {
        for my $keyword_ref_field (keys %keyword_ref_fields) {
            if ($keyword->{$keyword_ref_field}) {
                $keyword->{$keyword_ref_fields{$keyword_ref_field}} = $keyword_checksums{$keyword->{$keyword_ref_field}};
            }
        }

        # Non-relative REGEXP keyword indexes are listed as part of the rule
        if ($keyword->{regexp_type} && !$keyword->{is_relative}) {
            for my $rule_index (@{$keyword->{rule_id}}) {
                push(@{$rules[$rule_index]->{regexp_keyword_indexes}}, $keyword->{keyword_index});
            }
        }
    }

    my $max_suffix_len = List::Util::max( map(length($_->{text}),
                                              grep (!$_->{nocase}, @keywords))) || 0;

    my $max_keyword_len = List::Util::max( map(length($_->{text}), @keywords)) || 0;

    my $compiled_collection = pack('I3',
                                   $max_suffix_len,
                                   $max_keyword_len,
                                   scalar @keywords);

    for my $keyword (@keywords) {
        $compiled_collection .= pack('I/a* I C3 i5 c1 i2 i*',
                                     $keyword->{text},
                                     $keyword->{buffer_type},
                                     $keyword->{regexp_type} || 0,
                                     ($keyword->{nocase} ? 0 : 1),
                                     $keyword->{negation} || 0,
                                     (defined $keyword->{offset} ? $keyword->{offset} : 0), # aka abs_min_index
                                     (defined $keyword->{depth} ? $keyword->{depth} : MAX_INT), # aka abs_max_index
                                     (defined $keyword->{distance} ? $keyword->{distance} : 0), # aka rel_min_index
                                     (defined $keyword->{within} ? $keyword->{within} :  MAX_INT), # aka rel_max_index
                                     (defined $keyword->{isdataat_relative} ? $keyword->{isdataat_relative} : -1), # aka isdataat_relative_to_this
                                     $keyword->{prerequisite_keyword} || 0,
                                     (defined $keyword->{relative_keyword_index} ? $keyword->{relative_keyword_index} : -1),
                                     (defined $keyword->{regexp_following_keyword_index} ? $keyword->{regexp_following_keyword_index} : -1),
                                     scalar @{$keyword->{rule_id}},
                                     @{$keyword->{rule_id}});
    }


    $compiled_collection .= pack('i*',
                                 scalar @rules,
                                );

    for my $rule (@rules) {
        $compiled_collection .= pack('i*',
                                     $rule->{rule_obj}->sig_id,
                                     $rule->{rule_obj}->non_regexp_keyword_count,
                                     (defined $rule->{rule_obj}->isdataat ? $rule->{rule_obj}->isdataat : -1), # aka abs_isdataat
                                     BUFFER_TYPE_ALL_CONTENT, # hardcoded currently to entire request/response, but may be enhanced in the future
                                     scalar @{$rule->{regexp_keyword_indexes}},
                                     @{$rule->{regexp_keyword_indexes}});
    }

    my $num_states = 0;
    if (@keywords) {
        my $compiled_collection_with_max_states = pack('i1',LARGE_NUM_STATES) . $compiled_collection;
        $num_states = F5::PatternMatching::get_num_states($compiled_collection_with_max_states,
                                                             length($compiled_collection_with_max_states));
        if (! defined $num_states) {
            die "Signature collection could not be created\n";
        }
    }

    $compiled_collection = pack('i1', $num_states) . $compiled_collection;

    return $compiled_collection;
}

#----------------------------------------------------------------------

1;
