package F5::NegativeSignatures::Set;

use strict;
use warnings;

use F5::NegativeSignatures;
use F5::NegativeSignatures::Rule;
use F5::NegativeSignatures::Utils qw(find_or_add_attack_type find_or_add_system);
use F5::BigipVersionUtils;
use F5::XML::Twig;
use F5::DbUtils;

our $AUTOLOAD;

require Exporter;
our @ISA = qw(Exporter);
our @export_constants = qw(
                            $SET_UPDATE_STATUS_NOT_FOUND
                            $SET_UPDATE_STATUS_UPDATE
                            $SET_UPDATE_STATUS_ADDED
                            $SET_UPDATE_STATUS_DELETED
                            $SET_UPDATE_STATUS_NO_CHANGE
			    $SET_UPDATE_STATUS_EMPTY
                         );
our %EXPORT_TAGS = (constants => \@export_constants);
our @EXPORT_OK = (@export_constants);

our $SET_UPDATE_STATUS_NOT_FOUND = 'not found';
our $SET_UPDATE_STATUS_UPDATE = 'update';
our $SET_UPDATE_STATUS_ADDED = 'added';
our $SET_UPDATE_STATUS_DELETED = 'deleted';
our $SET_UPDATE_STATUS_NO_CHANGE = 'no change';
our $SET_UPDATE_STATUS_EMPTY = 'empty';

my $attribs = {
               set_name => { xml => 1,
                             table => 'NEGSIG_SETS',
                           },
               flg_assign_to_policy_default => { xml => 1,
                                                 table => 'NEGSIG_SETS',
                                               },
               flg_learn_default => { xml => 1,
                                      table => 'NEGSIG_SETS',
                                    },
               flg_alarm_default => { xml => 1,
                                      table => 'NEGSIG_SETS',
                                    },
               flg_reject_default => { xml => 1,
                                       table => 'NEGSIG_SETS',
                                     },
               apply_to => { xml => 1,
                             table => 'NEGSIG_SET_FILTERS',
                           },
               risk_op => { xml => 1,
                            table => 'NEGSIG_SET_FILTERS',
                            dependent_field => 'risk_val',
                          },
               risk_val => { xml => 1,
                             table => 'NEGSIG_SET_FILTERS',
                           },
               sig_tag_op => { xml => 1,
                            table => 'NEGSIG_SET_FILTERS',
                            dependent_field => 'sig_tag_val',
                          },
               sig_tag_val => { xml => 1,
                             table => 'NEGSIG_SET_FILTERS',
                           },
               sig_tag_policy_val => { 
                                      table => 'NEGSIG_SET_FILTERS',
                                     },
               accuracy_op => { xml => 1,
                                table => 'NEGSIG_SET_FILTERS',
                                dependent_field => 'accuracy_val',
                              },
               accuracy_val => { xml => 1,
                                 table => 'NEGSIG_SET_FILTERS',
                               },
               last_update_op => { xml => 1,
                                   table => 'NEGSIG_SET_FILTERS',
                                   dependent_field => 'last_update_val',
                                 },
               last_update_val => { xml => 1,
                                    table => 'NEGSIG_SET_FILTERS',
                                  },
               flg_is_user_defined => { xml => 1,
                                        table => 'NEGSIG_SET_FILTERS',
                                      },
               flg_has_cve => { xml => 1,
                                table => 'NEGSIG_SET_FILTERS',
                              },
               set_flg_is_user_defined => { xml => 1,
                                           table => 'NEGSIG_SETS',
                                           column => 'flg_is_user_defined',
                                          },
               set_group_name => { xml => 1,
                                   table => 'NEGSIG_SETS',
                                 },
               flg_is_hidden => { xml => 1,
                                        table => 'NEGSIG_SET_FILTERS',
                                      },
               set_flg_is_hidden => { xml => 1,
                                           table => 'NEGSIG_SETS',
                                           column => 'flg_is_hidden',
                                          },

               flg_is_manual  => { xml => 1,
                                   table => 'NEGSIG_SETS',
                                 },
               sig_id  => { xml => 1,
                            table => 'NEGSIG_SET_SIGNATURES',
                          },
               attack_type_id  => { xml => 1,
                            table => 'NEGSIG_SET_FILTERS',
                          },
               rest_uuid => {
                              table => 'NEGSIG_SETS',
                            },
               set_id => {},
               attack_type => {},
               set_systems => {},
               set_sigs => {},
              };

my @xml_fields = grep { $attribs->{$_}->{xml} } keys %$attribs;
my @negsig_sets_fields = grep { (defined $attribs->{$_}->{table}) && ($attribs->{$_}->{table} eq 'NEGSIG_SETS') } keys %$attribs;
my @negsig_set_filters_fields = grep { (defined $attribs->{$_}->{table}) && ($attribs->{$_}->{table} eq 'NEGSIG_SET_FILTERS') } keys %$attribs;
my @dependent_set_filter_fields = grep { $attribs->{$_}->{dependent_field} } keys %$attribs;

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;

  return $self->init(%params);
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;
  $self->{dbh} = $params{dbh};
  $self->{dbname} = $params{dbname} || 'PLC';
  $self->{set_flg_is_user_defined} = $params{set_flg_is_user_defined} || 0;
  $self->{set_flg_is_hidden} = $params{set_flg_is_hidden} || 0;
  $self->{flg_is_manual} = $params{flg_is_manual} || 0;

  if ($params{set_id} && $self->{dbh}) {
    return $self->retrieve(set_id => $params{set_id});
  }
  if ($params{set_name} && $self->{dbh}) {
    return $self->retrieve(set_name => $params{set_name});
  }

  if ($params{xml_element}) {
    $self->parse_xml($params{xml_element});
  }

#  $self->attack_type($self->get_attack_type_from_id($self->attack_type_id));

  if ($params{set_fields}) {
    # Convert attack_type and set_systems to ids
    if ($params{set_fields}->{attack_type} && $params{set_fields}->{attack_type} !~ /\D/) {
      $params{set_fields}->{attack_type} = $self->get_attack_type_from_id( $params{set_fields}->{attack_type} );
    }
    if ($params{set_fields}->{set_systems} && ! grep(/\D/, @{$params{set_fields}->{set_systems}})) {
      $params{set_fields}->{set_systems} = $self->get_system_from_id( $params{set_fields}->{set_systems} );
    }

    for my $field (keys %{$params{set_fields}}) {
        $self->{$field} = $params{set_fields}->{$field};
    }
    for my $field (@dependent_set_filter_fields) {
        if (! defined $self->{$field}) {
            $self->{$attribs->{$field}->{dependent_field}} = undef;
        }
    }
  }

  return $self;
}

#----------------------------------------------------------------------

sub retrieve {
  my($self, %params) = @_;

  my ($search_field, $search_val);
  if ($params{set_id}) {
    $search_field = 'set_id';
    $search_val = $params{set_id};
  }
  if ($params{set_name}) {
    $search_field = 'set_name';
    $search_val = $params{set_name};
  }

  local $" = ',';
  my $dbh = $self->{dbh};

  my %set_db_fields = map { (($attribs->{$_}->{column} || $_) => $_) }
    grep { $attribs->{$_}->{table} && $attribs->{$_}->{table} eq 'NEGSIG_SETS' } keys %$attribs;
  my @set_db_fields = keys %set_db_fields;
  my $set_rec = $dbh->selectrow_hashref(<<EndQuery, undef, $search_val);
SELECT @set_db_fields, set_id
  FROM $self->{dbname}.NEGSIG_SETS
 WHERE $search_field = ?
EndQuery

  return unless $set_rec;

  $self->set_id( $set_rec->{set_id} );
  while(my($db_field, $obj_attrib) = each %set_db_fields) {
    $self->$obj_attrib( $set_rec->{$db_field} );
  }

  # todo: handle manual based sets
  my @set_filter_db_fields = grep { $attribs->{$_}->{table} && $attribs->{$_}->{table} eq 'NEGSIG_SET_FILTERS' } keys %$attribs;
  my $filter_rec = $dbh->selectrow_hashref(<<EndQuery, undef, $self->set_id);
SELECT @set_filter_db_fields, attack_type_id
  FROM $self->{dbname}.NEGSIG_SET_FILTERS
 WHERE set_id = ?
EndQuery

  for my $field (@set_filter_db_fields) {
    $self->$field( $filter_rec->{$field} );
  }

  $self->attack_type( $self->get_attack_type_from_id($filter_rec->{attack_type_id}) );

  my $set_systems = $dbh->selectcol_arrayref(<<EndQuery, undef, $self->set_id);
SELECT system_name
  FROM $self->{dbname}.NEGSIG_SYSTEMS sys,
       $self->{dbname}.NEGSIG_SET_FILTER_SYSTEMS filter_sys,
       $self->{dbname}.NEGSIG_SET_FILTERS set_filter
 WHERE set_filter.set_id = filter_sys.set_id
   AND filter_sys.system_id = sys.system_id
   AND set_filter.set_id = ?
EndQuery

  $self->set_systems( $set_systems);

  my $set_sigs = $dbh->selectcol_arrayref(<<EndQuery, undef, $self->set_id);
SELECT sig_id
  FROM $self->{dbname}.NEGSIG_SET_SIGNATURES
 WHERE set_id = ?
EndQuery

  if ($set_sigs) {
      $self->set_sigs ($set_sigs);
  }


  $self->{_loaded_from_db} = 1;

  return $self;
}

#----------------------------------------------------------------------

sub update {
  my($self, %params) = @_;

  my $rv = {};
  local $" = ',';
  my $dbh = $self->{dbh};

  # todo: handle manual sets

  # Check for existence of all systems first, in order to preemptively fail creation of
  # Set with only non-existent systems in filter
  my @set_system_ids;
  if ($self->{set_systems} && @{$self->{set_systems}}) {
      for my $set_system_name (@{$self->{set_systems}}) {
          my %additional_opts;
          if (! $ENV{ASM_GENERATE_NEGSIG_CSV}) {
              $additional_opts{add_if_missing} = 0;
          }
          my $set_system_id = find_or_add_system(dbh => $self->{dbh},
                                                 dbname => $self->{dbname},
                                                 system_name => $set_system_name,
                                                 %additional_opts,
                                                );
          if ($set_system_id) {
              push(@set_system_ids, $set_system_id);
          }
          else {
              push(@{$rv->{missing_systems}}, $set_system_name);
          }
      }
      if (! @set_system_ids) {
          $rv->{update_status} = $SET_UPDATE_STATUS_EMPTY;
          return $rv;
      }
  }


  my $is_set_in_db = $self->set_id && $dbh->selectcol_arrayref(<<EndQuery, undef, $self->set_id)->[0];
SELECT 1
  FROM $self->{dbname}.NEGSIG_SETS
 WHERE set_id = ?
EndQuery

  if ($self->set_name) {
      $self->rest_uuid(F5::Utils::Rest::generate_uuid($self->set_name));
  }

  my @set_db_fields = grep { $attribs->{$_}->{table} && $attribs->{$_}->{table} eq 'NEGSIG_SETS' && defined $self->$_ } keys %$attribs;

  my @set_db_vals = map { $self->$_ } @set_db_fields;
  for my $i (0..$#set_db_fields) {
    if (exists $attribs->{$set_db_fields[$i]}->{column}) {
      $set_db_fields[$i] = $attribs->{$set_db_fields[$i]}->{column};
    }
  }

  if ($is_set_in_db) {
    my @update_assigns = map("$_ = ?", @set_db_fields);
    $dbh->do("UPDATE $self->{dbname}.NEGSIG_SETS SET @update_assigns WHERE set_id = ?",
             undef,
             @set_db_vals, $self->set_id);
  }
  else {
    if ($self->set_id) {
      push(@set_db_fields, 'set_id');
      push(@set_db_vals, $self->set_id);

    }

    my @placeholders = ('?') x scalar @set_db_fields;
    my $rv = $dbh->do("INSERT INTO $self->{dbname}.NEGSIG_SETS (@set_db_fields) VALUES (@placeholders)", undef,
                      @set_db_vals);
    unless ($rv) {
      die "Set '". $self->set_name . "' cannot be loaded (possible name conflict)\n";
    }

    if (!$self->set_id) {
      my $new_set_id = $dbh->last_insert_id(undef, undef, undef, undef);
      $self->set_id( $new_set_id );
    }
  }

  if ($is_set_in_db) {
      # Check whether records exist before attempting to delete; otherwise, deleting zero rows may cause DB deadlock
      my($existing_row_in_db) = $dbh->selectrow_array("SELECT distinct set_id FROM $self->{dbname}.NEGSIG_SET_FILTERS WHERE set_id = ?", undef, $self->set_id);

      if ($existing_row_in_db) {
          $dbh->do("DELETE FROM $self->{dbname}.NEGSIG_SET_FILTERS WHERE set_id = ?", undef, $self->set_id);
      }
  }

  my $filter_db_rec = {};
  for my $db_field (grep { $attribs->{$_}->{table} && $attribs->{$_}->{table} eq 'NEGSIG_SET_FILTERS' } keys %$attribs) {
    $filter_db_rec->{$db_field} = $self->$db_field;
  }
  if ($self->attack_type) {
    $filter_db_rec->{attack_type_id} = find_or_add_attack_type(dbh => $self->{dbh},
                                                               dbname => $self->{dbname},
                                                               attack_type_name => $self->attack_type);
  }

  if (! $self->flg_is_manual) {
      my @filter_db_fields = keys %$filter_db_rec;
      my @placeholders = ('?') x scalar @filter_db_fields;
      $dbh->do("INSERT INTO $self->{dbname}.NEGSIG_SET_FILTERS (@filter_db_fields, set_id) VALUES (@placeholders,?)",
               undef, @{$filter_db_rec}{@filter_db_fields}, $self->set_id);
  }

  if ($is_set_in_db) {
      # Check whether records exist before attempting to delete; otherwise, deleting zero rows may cause DB deadlock
      my($existing_row_in_db) = $dbh->selectrow_array("SELECT distinct set_id FROM $self->{dbname}.NEGSIG_SET_FILTER_SYSTEMS WHERE set_id = ?", undef, $self->set_id);

      if ($existing_row_in_db) {
          $dbh->do("DELETE FROM $self->{dbname}.NEGSIG_SET_FILTER_SYSTEMS WHERE set_id = ?", undef, $self->set_id);
      }
  }

  if (@set_system_ids) {
    for my $set_system_id (@set_system_ids) {
      $dbh->do("INSERT INTO $self->{dbname}.NEGSIG_SET_FILTER_SYSTEMS (set_id, system_id) VALUES (?,?)",
               undef, $self->set_id, $set_system_id);
    }
  }

  if ($self->flg_is_manual) {
    $dbh->do("DELETE FROM $self->{dbname}.NEGSIG_SET_SIGNATURES WHERE set_id = ?", undef, $self->set_id);
    my %save_dbh_attribs;
    $save_dbh_attribs{RaiseError} = $dbh->{RaiseError};
    $save_dbh_attribs{PrintError} = $dbh->{PrintError};
    $dbh->{RaiseError} = 1;
    $dbh->{PrintError} = 0;

    if ($self->set_sigs) {
      for my $sig_id (@{$self->set_sigs}) {
        next unless $sig_id;

        eval {
            F5::DbUtils::insert_row(dbh => $dbh,
                                    table_name => 'PLC.NEGSIG_SET_SIGNATURES',
                                    rec => {set_id => $self->set_id, sig_id => $sig_id},
                                   );
        };
        if ($@) {
            if ($params{warn_on_failed_manual_sigs}) {
                push(@{$rv->{missing_manual_sigs}}, $sig_id);
            }
            else {
                $dbh->{RaiseError} = $save_dbh_attribs{RaiseError};
                $dbh->{PrintError} = $save_dbh_attribs{PrintError};
                die $@;
            }
        }
      }
    }
    $dbh->{RaiseError} = $save_dbh_attribs{RaiseError};
    $dbh->{PrintError} = $save_dbh_attribs{PrintError};
  }

  if ( ! $ENV{ASM_GENERATE_NEGSIG_CSV} ) {
    F5::NegativeSignatures->new($self->{dbh})->link_sigs_to_set($self->{set_id},
                                                                store_type => ($is_set_in_db ? 'update' : 'add'),
                                                               );
  }

  return $rv;
}

#----------------------------------------------------------------------

sub delete {
  my($self) = @_;

  my @affected_policies = F5::NegativeSignatures->new($self->{dbh})->get_policies_using_sets($self->set_id);
  $self->{dbh}->do("DELETE FROM $self->{dbname}.NEGSIG_SETS WHERE set_id = ?", undef, $self->set_id);

  return \@affected_policies;
}


#----------------------------------------------------------------------

sub import_set {
    my ($class, %params) = @_;

    my $dbh = defined $params{dbh} ? $params{dbh} : F5::DbUtils::get_dbh();

    my @rvs;
    my $twig;
    if (defined $params{twig}) {
        $twig = $params{twig};
        for my $elm ($twig->findnodes("set")) {
            if ($elm->first_child_text('set_flg_is_user_defined') == 1) {
                my $rv = F5::NegativeSignatures::Set->update_if_changed(dbh => $dbh,
                                                               xml_element => $elm,
                                                               mode => 'import',
                                                               lookup_by => 'set_name');
                push(@rvs, $rv);
            }
            else {
                push(@rvs, {new_set_id => $elm->att('id'), old_set_id => $elm->att('id'), update_status => $SET_UPDATE_STATUS_NO_CHANGE});
            }
        }
    }
    elsif (defined $params{xml_text}) {
        $twig = F5::XML::Twig->new(twig_handlers => { 
                                                 set => sub {
                                                     my($twig, $elm) = @_;
                                                     if ($elm->first_child_text('set_flg_is_user_defined') == 1) {
                                                         my $rv = F5::NegativeSignatures::Set->update_if_changed(dbh => $dbh,
                                                                                                                 xml_element => $elm, 
                                                                                                                 mode => 'import', 
                                                                                                                 lookup_by => 'set_name');
 	                                                 push @rvs, $rv;
                                                     }
                                                     else {
                                                         push(@rvs, {new_set_id => $elm->att('id'), old_set_id => $elm->att('id'), update_status => $SET_UPDATE_STATUS_NO_CHANGE});
                                                     }
                                                 }
                                                }
                              );
        $twig->parse($params{xml_text});
    }
    else {
        return;
    }


    return \@rvs;
}

#----------------------------------------------------------------------

sub export_set {
    my ($self_or_class, %params) = @_;

    my @sets;
    my $sets_twig = XML::Twig::Elt->new('sets');

    if (ref($self_or_class)) {
        push @sets, $self_or_class;
    }  
    elsif (defined $params{set_names}) {
        for my $set_name (@{$params{set_names}}) {
            my $set = $self_or_class->new(dbh => $params{dbh},
                                          set_name => $set_name);
            push @sets, $set;
        }
    }
    elsif (defined $params{set_ids}) {
        for my $set_id (@{$params{set_ids}}) {
            my $set = $self_or_class->new(dbh => $params{dbh},
                                          set_id => $set_id);
            push @sets, $set;
        }
    }
    else {
        return;
    }

    for my $set (@sets) {
        $set->build_xml->paste(last_child => $sets_twig);
    }

    $sets_twig->set_pretty_print('indented');

    return { sets_twig => $sets_twig, sets_xml => $sets_twig->sprint };
}

#----------------------------------------------------------------------

sub build_xml {
    my ($self) = @_;

    my $set = XML::Twig::Elt->new('set');
    $set->set_att(
                  id => $self->set_id,
                  name => $self->set_name,
                  'xmlns:xsi' => "http://www.w3.org/2001/XMLSchema-instance",
                 );

    for my $negsig_sets_field_name (@negsig_sets_fields) {
        next unless $attribs->{$negsig_sets_field_name}->{xml};
        my $elt = XML::Twig::Elt->new($negsig_sets_field_name);
        if (defined $self->$negsig_sets_field_name) {
            $elt->set_text($self->$negsig_sets_field_name);
            $elt->paste(first_child => $set);
        }
    }

    if ( $self->flg_is_manual) {
        my $sigs = XML::Twig::Elt->new('sigs');
    
        for my $sig_id (@{$self->set_sigs}) {
            my $elt = XML::Twig::Elt->new('sig_id');
            if (defined $sig_id) {
                if(! exists $self->{user_defined_sig_id_to_sig_name}) {
                    $self->{user_defined_sig_id_to_sig_name} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_id', undef, 1);
SELECT sig_id, sig_name
  FROM PLC.NEGSIG_SIGNATURES
  WHERE flg_is_user_defined = ?;
EndQuery
                }
                if(exists $self->{user_defined_sig_id_to_sig_name}->{$sig_id}) {
		    $elt->set_att(name => $self->{user_defined_sig_id_to_sig_name}->{$sig_id}->{sig_name});
                }
                $elt->set_text($sig_id);
                $elt->paste(first_child => $sigs);
            }
        }
    
        $sigs->paste(last_child => $set);
    }
    else {
        my $filter = XML::Twig::Elt->new('filter');
    
        for my $negsig_set_filters_field (@negsig_set_filters_fields) {
            my $elt = XML::Twig::Elt->new($negsig_set_filters_field);
            if (defined $self->$negsig_set_filters_field) {
                $elt->set_text($self->$negsig_set_filters_field);
                $elt->paste(first_child => $filter);

                # Manually export attack_type_name after attack_type_id
                if ($negsig_set_filters_field eq 'attack_type_id') {
                    my $elt = XML::Twig::Elt->new('attack_type_name');
                    $elt->set_text(F5::NegativeSignatures::Utils->get_attack_type_name(attack_type_id => $self->attack_type_id,
                                                                                       dbh => $self->{dbh}));
                    $elt->paste(first_child => $filter);
                }
            }
        }
    
        $filter->paste(last_child => $set);
    }

    my $systems = XML::Twig::Elt->new('systems');

    for my $system_name (@{$self->set_systems}) {
        my $elt = XML::Twig::Elt->new('system_name');
        if (defined $system_name) {
            $elt->set_text($system_name);
            $elt->paste(first_child => $systems);
        }
    }

    $systems->paste(last_child => $set);

    $set->set_pretty_print('indented');
    return $set;
}
#----------------------------------------------------------------------

sub parse_xml {
  my($self, $elm) = @_;

  my $set = $elm->simplify( keyattr => {id => 'set_id'},);
  $set->{set_name} =~ s/^\s*//;
  $set->{set_name} =~ s/\s*$//;

  $self->set_id ($set->{id});

  if (!$self->set_flg_is_user_defined) {
    $self->set_id( $set->{att}->{id} );
  }
  $self->attack_type( $set->{attack_type} );


  if (defined $set->{systems} && defined $set->{systems}->{system_name}) {
      if (ref($set->{systems}->{system_name}) eq 'ARRAY') {
          push @{$self->{set_systems}}, @{$set->{systems}->{system_name}};
      }
      else {
          push @{$self->{set_systems}}, $set->{systems}->{system_name};
      }
  }

  if (defined $set->{sigs} && defined $set->{sigs}->{sig_id}) {
      if (! defined $set->{flg_is_manual}) {
          $set->{flg_is_manual} = 1;
      }

      my @set_sigs;
      if (ref($set->{sigs}->{sig_id}) eq 'ARRAY') {
          @set_sigs = @{$set->{sigs}->{sig_id}};
      }
      else {
          @set_sigs =  ($set->{sigs}->{sig_id});
      }
     
      for my $set_sig (@set_sigs) {
          if (ref($set_sig) eq 'HASH' && defined $set_sig->{name}) { # Calculate user defined signature id based on name
              if(! exists $self->{user_defined_sig_name_to_sig_id}) {
                  $self->{user_defined_sig_name_to_sig_id} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_name', undef, 1);
SELECT sig_name, sig_id
  FROM PLC.NEGSIG_SIGNATURES
  WHERE flg_is_user_defined = ?;
EndQuery
              }
              if(exists $self->{user_defined_sig_name_to_sig_id}->{$set_sig->{name}}) {
                  my $sig_id = $self->{user_defined_sig_name_to_sig_id}->{$set_sig->{name}}->{sig_id};
                  push(@{$self->{set_sigs}}, $sig_id);
              }
          }
          else { # Factory signature - make sure it exists
              if(! exists $self->{sig_ids}) {
                  $self->{sig_ids} = $self->{dbh}->selectall_hashref(<<EndQuery, 'sig_id');
SELECT sig_id
  FROM PLC.NEGSIG_SIGNATURES
EndQuery
              }
              if (exists $self->{sig_ids}->{$set_sig}) {
                  push(@{$self->{set_sigs}}, $set_sig);
              }
          }
      }
  }

  if ($set->{filter}->{attack_type_name}) {
      $set->{filter}->{attack_type_id} = F5::NegativeSignatures::Utils->get_attack_type_id(dbh => $self->{dbh},
                                                                                           attack_type_name => $set->{filter}->{attack_type_name});
      delete $set->{filter}->{attack_type_name};
  }

  for my $filter_field (keys %{$set->{filter}}) {
    $self->{$filter_field} = $set->{filter}->{$filter_field};
  }

  for my $copy_field (@xml_fields) {
    if (defined $set->{$copy_field}) {
        $self->$copy_field( $set->{$copy_field} );
    }
  }
}

#----------------------------------------------------------------------

sub differs_from {
  my($self, $other_set) = @_;

  if ($self->compare($other_set)) {
    return 1;
  }
  return;
}

#----------------------------------------------------------------------

sub compare {
  my($self, $other_set) = @_;

  my @differences;
  my @attack_type_fields = qw(attack_type_id
                              attack_type
                              attack_type_name);

  my %noncompare_fields = map { ($_ => 1) } (qw(set_id
                                               rest_uuid
                                             ),
                                            @attack_type_fields);

  my %compare_fields = map { ($_ => 1) } grep { ! $noncompare_fields{$_} } keys %$attribs;

  # Only compare 1 attack type field, using the field that is present in both.
  # If none are present in both, use the first that is present in either,
  # which will inherently (and correctly) be found as a difference.
  # If none are present in either, ignore attack type fields entirely.
  my $compare_attack_type_field;
  for my $attack_type_field (@attack_type_fields) {
      if (defined $self->$attack_type_field && defined $other_set->$attack_type_field) {
          $compare_attack_type_field = $attack_type_field;
          last;
      }
  }
  if (! $compare_attack_type_field) {
      for my $attack_type_field (@attack_type_fields) {
          if (defined $self->$attack_type_field || defined $other_set->$attack_type_field) {
              $compare_attack_type_field = $attack_type_field;
              last;
          }
      }
  }
  if ($compare_attack_type_field) {
      $compare_fields{$compare_attack_type_field} = 1;
  }

  for my $field (keys %compare_fields) {
    if ($field eq 'set_systems') {
      # No sig system would contain ///, so just join for the comparison
      if ($self->set_systems && $other_set->set_systems &&
          join('///', sort @{$self->set_systems}) ne join('///', sort @{$other_set->set_systems})) {
        push(@differences,
             { diff_type => 'modify',
               diff_field => $field,
               old_val => $self->set_systems,
               new_val => $other_set->set_systems,
             });
      }
    }
    elsif ($field eq 'set_sigs') {
        if ($self->set_sigs && $other_set->set_sigs && $self->flg_is_manual &&
          join('///', sort @{$self->set_sigs}) ne join('///', sort @{$other_set->set_sigs})) {
        push(@differences,
             { diff_type => 'modify',
               diff_field => $field,
               old_val => $self->set_sigs,
               new_val => $other_set->set_sigs,
             });
      }
    }
    else {
      if ((!defined $self->$field && defined $other_set->$field) ||
          (defined $self->$field && !defined $other_set->$field) ||
          (defined $self->$field && defined $other_set->$field &&
           $self->$field ne $other_set->$field)) {
        push(@differences,
             { diff_type => 'modify',
               diff_field => $field,
               old_val => $self->$field,
               new_val => $other_set->$field
             });
      }
    }
  }

  return @differences;
}

#----------------------------------------------------------------------

sub get_attack_type_from_id {
  my($self, $attack_type_id) = @_;

  return $self->{dbh}->selectcol_arrayref(<<EndQuery, undef, $attack_type_id)->[0];
SELECT attack_type_name
  FROM $self->{dbname}.NEGSIG_ATTACK_TYPES 
 WHERE attack_type_id = ?
EndQuery
}

#----------------------------------------------------------------------

sub get_system_from_id {
  my($self, $system_ids) = @_;

  my @system_names;
  for my $system_id (@$system_ids) {
    push(@system_names, $self->{dbh}->selectcol_arrayref(<<EndQuery, undef, $system_id)->[0]);
SELECT system_name
  FROM $self->{dbname}.NEGSIG_SYSTEMS 
 WHERE system_id = ?
EndQuery
  }

  return \@system_names;
}


#======================================================================
# Accessors

sub AUTOLOAD {
  my($self, $set_val) = @_;

  my $name = $AUTOLOAD;
  $name =~ s/.*://;   # strip fully-qualified portion
  return if ($name eq 'DESTROY');

  if (exists $attribs->{$name}) {
    if (defined $set_val) {
      $self->{$name} = $set_val;
    }
    return $self->{$name};
  }

  return; # undef if no such value found
}

#======================================================================
# Class method

sub update_if_changed {
    my($class, %params) = @_;

    $params{dbname} ||= 'PLC';
    my $set = F5::NegativeSignatures::Set->new(%params);

    my $update_status;
    my %lookup_key;
    my $original_set_id = $set->set_id;
    my $original_set_name = $set->set_name;
    my $update_rv;
    my @affected_policies;

    if ($set->{flg_is_manual} && (!exists $set->{set_sigs} || !@{$set->{set_sigs}})) { # Empty set
       $update_status = $SET_UPDATE_STATUS_EMPTY;
    }
    else {
       if ($params{lookup_by} eq 'set_name') {
           $lookup_key{set_name} = $set->set_name;
       }   
       else {
           $lookup_key{set_id} = $set->set_id;
       }

       my $existing_set;
       if ((values %lookup_key)[0]) {
           $existing_set = F5::NegativeSignatures::Set->new(dbh => $params{dbh},
                                                            dbname => $params{dbname},
                                                            %lookup_key);

           my $rename_number_suffix = 0;
           while ($existing_set && $params{mode} && $params{mode} eq 'import' && $existing_set->differs_from($set)) {
               $rename_number_suffix++;
               my $suffix = "_$rename_number_suffix";
	       if(length($original_set_name) + length($suffix) > 255) {
                   my $original_set_name_cut_off = substr $original_set_name, 0, 255 - length($suffix);
                   $set->set_name($original_set_name_cut_off . $suffix);
 	       }
               else {
                   $set->set_name($original_set_name . $suffix);
	       }
               $existing_set = F5::NegativeSignatures::Set->new(dbh => $params{dbh},
                                                                dbname => $params{dbname},
                                                                set_name => $set->set_name);
           }
       }

       if (! $existing_set  || ($existing_set->differs_from($set) && $existing_set->set_flg_is_user_defined)) {
           # Add new set, or update user-defined set if differs

           if ($existing_set) {
               $update_status = $SET_UPDATE_STATUS_UPDATE;
               @affected_policies = F5::NegativeSignatures->new(dbh => $params{dbh},
                                                                dbname => $params{dbname},
                                                               )->get_policies_using_sets($existing_set->set_id);
               $set->set_id( $existing_set->set_id);
           }
           else {
               $update_status = $SET_UPDATE_STATUS_ADDED;
               if ($set->set_flg_is_user_defined) {
                   $set->{set_id} = undef;
               }
           }

           $update_rv = $set->update(warn_on_failed_manual_sigs => 1);
       }
       else {
           $set->set_id( $existing_set->set_id);
           $update_status = $SET_UPDATE_STATUS_NO_CHANGE;
       }
    }

    if (! $update_rv) {
        $update_rv = {};
    }

    return { new_set_id => $set->{set_id},
             old_set_id => $original_set_id,
             set_name => $set->set_name,
             requested_set_name => $original_set_name,
             update_status => $update_status,
             affected_policies => \@affected_policies,
             %$update_rv,
           };
}



1;
