
package F5::NegativeSignatures::SigFile;

use strict;
use warnings;

use File::Find;
use File::Path;
use File::Copy;
use File::Basename;
use XML::Simple;
use F5::XML::Twig;
use YAML::XS;
use JSON -convert_blessed_universally;
use Data::MessagePack;
use Digest::MD5;
use Encode;
use File::Slurp;

# Used to call gpg externally
use IPC::Open3;
use Symbol;
use File::Spec;

use F5::Deploy;
use F5::Syslog;
use F5::ObjUtils;
use F5::BatchInsert;
use F5::NegativeSignatures;
use F5::NegativeSignatures::Signature qw(:constants);
use F5::NegativeSignatures::Set qw(:constants);
use F5::NegativeSignatures::Utils qw(find_or_add_attack_type find_or_add_system);
use F5::BigipUtils;
use F5::BigipVersionUtils;
use F5::DateTimeUtils;
use F5::ASMConfig::ASMConfig;
use F5::Dos::BotSignatures;
use F5::ServerTechnologies;
use F5::YAML::Schema::Utils;
use F5::SystemUtils;
use F5::Utils::File;
use F5::Utils::Policy;

$XML::Simple::PREFERRED_PARSER = 'XML::Parser';

my $DEFAULT_TEMP_BASE_DIR = F5::Deploy::map_path("/ts/var/tmp");

my $XMLLINT_CMD = '/usr/bin/xmllint';
my $GPG_CMD = '/usr/bin/gpg';
our $MAX_NEGSIG_VERSION_SUPPORTED = 2;
our $SIGFILE_FORMAT_BINARY = 'BINARY';
our @SIGFILE_FIELDS = qw(name rule signatureType lastUpdateMicros attackType revision risk accuracy description systems);
our @DEFAULT_SIG_REPORT_FIELDS = qw(name signatureId signatureType attackType accuracy tag risk systems hasCve references isUserDefined);
my $IM_FILE_OUTPUT = 'ISO 9660 CD-ROM';

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;

  my $self = {};
  bless $self, $class;
  $self->init(%params);

  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{filename} = $params{filename};
  $self->{file_contents_struct} = $params{file_contents_struct};
  $self->{sets_dir} = $params{sets_dir};
  $self->{sigs_dir} = $params{sigs_dir};
  $self->{base_dir} = $params{base_dir};
  $self->{dbh} = $params{dbh};
  $self->{dbname} = $params{dbname} || 'PLC';
  $self->{flg_is_user_defined} = $params{flg_is_user_defined};
  $self->{output_filename} = $params{output_filename};
  $self->{sigfile_format} = $params{sigfile_format} || 'XML';
  $self->{sig_tag} = $params{sig_tag};
  $self->{sig_tag_policy} = $params{sig_tag_policy};
  $self->{gpg_sign} = (exists $params{gpg_sign} ? $params{gpg_sign} : !$params{sigfile_format});
  $self->{temp_base_dir} = $params{temp_base_dir} || $DEFAULT_TEMP_BASE_DIR;
  $self->{sig_report_fields} = $params{sig_report_fields};
  $self->{validate_requirements} = defined $params{validate_requirements} ? $params{validate_requirements} : 1;

  if ($self->{filename}) {
    if (! -e $self->{filename}) {
      F5::Syslog::ts_syslog( 'warning',
                             "Attack Signature file not found (%s)",
                             $self->{filename}
                           );

      $self->{filename} = '';
    }
    else {
      if ($self->{flg_is_user_defined}) {
        delete $self->{sigfile_format};
        $self->{sigfile_format} = $params{sigfile_format} || $self->get_sigfile_format() || 'XML';
        $self->check_user_sigfile_schema;
      }
      else {
        $self->unzip;
      }
    }
  }
}

#----------------------------------------------------------------------

sub get_sigfile_format {
     my ($self) = @_;

     if (defined $self->{sigfile_format} ){
         return $self->{sigfile_format};
     }

     if ( !$self->{filename} || ! -f $self->{filename} ){
         return;
     }

     my $filetype_check_output;

     my @params;
     push ( @params, '/usr/bin/file',$self->{filename});
     my $res = F5::SystemUtils::run_system_tool( command_line_params => \@params,
                                                  timeout => 60,
                                                 );
     if ($res && $res->{completed_successfully}) {
         $filetype_check_output = $res->{standard_output};
         chomp($filetype_check_output);
     }

     if ( $filetype_check_output =~ /$IM_FILE_OUTPUT/ ){
         return $IM_FILE_OUTPUT;
     }
     elsif ( $filetype_check_output =~ /XML/ ){
         return 'XML';
     }
     elsif ( $filetype_check_output =~ /JSON/ ){
         return 'JSON';
     }
     elsif ( $filetype_check_output =~ /data/ ){
         return $SIGFILE_FORMAT_BINARY;
     }
     elsif (  $filetype_check_output =~ /text/ ){
         my $inf;
         if (! open($inf, '<', $self->{filename}) ) {
             return;
         }

         my $first_part;
         read($inf, $first_part, 50);
         close $inf;

         if ($self->{filename} =~ /\.json/ || $first_part =~ /^\s*\{\s*\"/m) {
             return 'JSON';
         }
     }

     return;
}


#----------------------------------------------------------------------
sub unzip {
  my($self) = @_;

  my $decrypted_filename;
  my $decrypted_is_temporary = 0;

  if ($self->{gpg_sign}) {
    $decrypted_is_temporary = 1;
    $decrypted_filename = "$self->{temp_base_dir}/sigfile.tmp.$$.decrypted_sigfile.tgz";

    gpg_decrypt($decrypted_filename, $self->{filename});
  }
  else {
    $decrypted_filename = $self->{filename};
  }

  $self->{base_dir} = "$self->{temp_base_dir}/sigfile.tmp.$$";
  $self->{base_dir_is_temporary} = 1;
  mkpath $self->{base_dir};
  my $sys_rv = F5::SystemUtils::run_system_tool(command_line_params => ['tar','-zmxf', $decrypted_filename, '-C', $self->{base_dir}],
                                                find_command_path => 1);
  unlink $decrypted_filename if $decrypted_is_temporary;
  if (!$sys_rv->{completed_successfully}) {
       die "Failed to unpack $self->{filename}: '$sys_rv->{standard_error}'\n";
  }

  $self->{sets_dir} = "$self->{base_dir}/sets";
  $self->{sigs_dir} = "$self->{base_dir}/sigs";
  if ($self->{sigfile_format} eq $SIGFILE_FORMAT_BINARY) {
      $self->{filename} = glob("$self->{base_dir}/*");
  }
  for my $version_dir (reverse glob("$self->{base_dir}/negsig_version_*")) {
    my $dir_negsig_version;
    if ( $version_dir =~ /negsig_version_(\d+)/ ){
        $dir_negsig_version = $1;
    }
    if (defined $dir_negsig_version && $dir_negsig_version <= $MAX_NEGSIG_VERSION_SUPPORTED) {
       $self->{negsig_version_sigs_specific_dir} = "$version_dir/sigs";
       $self->{negsig_version_readmes_specific_dir} = "$version_dir/readmes";
       last;
    }
  }
  $self->{negsig_version} = $MAX_NEGSIG_VERSION_SUPPORTED;
}

#----------------------------------------------------------------------

sub gpg_decrypt {
    my ($decrypted_filename, $input_file_name) = @_;

    my $keyring_path = F5::Deploy::map_path("/ts/share/negsig/gpg_asm_sigfile_installer");

    open(my $null1, ">", File::Spec->devnull);
    open(my $null2, File::Spec->devnull);
    my $gpg_err_fh = gensym;
    my $gpg_pid = open3($null1, $null2, $gpg_err_fh, 
                        $GPG_CMD, qw(--always-trust --yes --homedir), $keyring_path,
                        '--output', $decrypted_filename,
                        '--decrypt', $input_file_name);
    my $gpg_err = join('', <$gpg_err_fh>);
    waitpid($gpg_pid, 0);
    if ($?) {
      die "Signature file is not valid\n";
    }
}

#----------------------------------------------------------------------

sub finish {
  my($self) = @_;

  my $dbh = $self->{dbh};

  my %sigfile_status = (
                         bigip_version => F5::BigipVersionUtils::bigip_version(),
                         sigs_updated_count => 0,
                         sigs_updated_minor_count => 0,
                         sigs_updated_major_count => 0,
                         sigs_added_count => 0,
                         sigs_deleted_count => 0,
                         flg_is_user_defined => $self->{flg_is_user_defined},
                         sig_tag => $self->{sig_tag} || '',
                         sig_tag_policy => $self->{sig_tag_policy} || '',
                         load_datetime => F5::DateTimeUtils::convert_epoch_to_datetime(time()),
                         negsig_version => $self->{negsig_version} || $MAX_NEGSIG_VERSION_SUPPORTED,
                        );

  if ($sigfile_status{flg_is_user_defined}) {
      $sigfile_status{sigfile_timestamp} = $self->{timestamp} || $sigfile_status{load_datetime};
  }
  else {
     $sigfile_status{sigfile_timestamp} = $self->timestamp || $sigfile_status{load_datetime};
     $sigfile_status{readme} = $self->readme;
     $sigfile_status{last_added_readme_file} = $self->last_added_readme_file;
  }

  F5::DbUtils::insert_row(dbh => $dbh,
                          table_name => "$self->{dbname}.SIGFILE_STATUS",
                          rec => \%sigfile_status
                          );

  if ($self->{base_dir_is_temporary}) {
    rmtree($self->{base_dir});
  }
}

#----------------------------------------------------------------------

sub parse_set {
  my($self, %params) = @_;

  my $elm = $params{elm};
  my $twig = $params{twig};
  my $set_update_status = $params{set_update_status};

  my $lookup_by = ($self->{flg_is_user_defined} ? 'set_name' : 'set_id');
  my $set_update = F5::NegativeSignatures::Set->update_if_changed(dbh => $self->{dbh},
                                                                  dbname => $self->{dbname},
                                                                  flg_is_user_defined => $self->{flg_is_user_defined},
                                                                  xml_element => $elm,
                                                                  lookup_by => $lookup_by);

  $set_update_status->{$set_update->{new_set_id}} = $set_update;

  $twig->purge;
  return 1;
}

#----------------------------------------------------------------------

my %load_signatures_dispatch = ('XML'    => 'load_xml_signatures',
                                $SIGFILE_FORMAT_BINARY => 'load_binary_signatures',
                                'JSON'   => 'load_json_signatures',
                               );

sub load_sigs {
  my($self, %params) = @_;
  
  my $policy_to_sigs = $params{policy_to_sigs};

  my $dbh = $self->{dbh};
  
  # If flg_is_user_defined wasn't defined, default to factory sigs
  $self->{flg_is_user_defined} = 0 unless defined($self->{flg_is_user_defined});
  
  my $sig_update_status = $dbh->selectall_hashref(<<EndQuery, 'sig_id', undef, $self->{flg_is_user_defined});
SELECT sig_id, sig_name, sig_tag, sig_tag_policy, rev 
  FROM $self->{dbname}.NEGSIG_SIGNATURES 
WHERE flg_is_user_defined = ?
EndQuery
  
  for my $update_status_rec (values %$sig_update_status) {
    $update_status_rec->{update_status} = $SIG_UPDATE_STATUS_NOT_FOUND;
  }
    
  my @sig_files;
  if ($self->{negsig_version_sigs_specific_dir} && -d $self->{negsig_version_sigs_specific_dir}) {
    @sig_files = grep { /\.xml$/ } get_recursive_files($self->{negsig_version_sigs_specific_dir});
  }
  elsif ($self->{sigs_dir} && -d $self->{sigs_dir}) {
    @sig_files = grep { /\.xml$/ } get_recursive_files($self->{sigs_dir});
  }
  elsif ($self->{filename}) {
    @sig_files = ($self->{filename});
  }
  elsif ($self->{file_contents_struct}) {
    @sig_files = ($self->{file_contents_struct});
  }
  return {} unless @sig_files;
  
  my $load_signatures_func = $load_signatures_dispatch{$self->{sigfile_format}} || 'load_xml_signatures';
  $self->$load_signatures_func(sig_files => \@sig_files,
                               sig_update_status => $sig_update_status,
                               policy_to_sigs => $policy_to_sigs);

  if (!$self->{flg_is_user_defined}) {
    for my $deleted_sig_update_rec (grep($_->{update_status} eq $SIG_UPDATE_STATUS_NOT_FOUND, values %$sig_update_status)) {
      my $deleted_sig = F5::NegativeSignatures::Signature->new(dbh => $dbh,
                                                               sig_id => $deleted_sig_update_rec->{sig_id});
      $deleted_sig->delete();
      $sig_update_status->{$deleted_sig->sig_id}->{update_status} = $SIG_UPDATE_STATUS_DELETED;
    }
  }

  if ($self->{sigfile_format} eq $SIGFILE_FORMAT_BINARY) {
      my $negative_signatures = F5::NegativeSignatures->new(dbh => $dbh, dbname => 'PLC');
      $negative_signatures->add_all_signatures_to_policy(policy_id => '*',
                                                         flg_quarantine_signatures => F5::BigipUtils::is_feature_enabled('allow_staging'));

  }

  return $sig_update_status;
}

#----------------------------------------------------------------------

sub load_xml_signatures {
    my ($self, %params) = @_;
    my $sig_files         = $params{sig_files};
    my $sig_update_status = $params{sig_update_status};
    my $policy_to_sigs    = $params{policy_to_sigs};

    my $twig = F5::XML::Twig->new(twig_handlers => {
                                              sig => sub {
                                                my($twig, $elm) = @_;
                                                $self->parse_sig( elm => $elm,
                                                                  twig => $twig,
                                                                  sig_update_status => $sig_update_status,
                                                                  policy_to_sigs => $policy_to_sigs,
                                                                );
                                              }
                                             }
                           );

    for my $filename (@$sig_files) {
      $twig->parsefile($filename);
    }
}

#----------------------------------------------------------------------

sub load_json_signatures {
    my ($self, %params) = @_;
    my $sig_files = $params{sig_files};
    my $dbh  = $params{dbh} || $self->{dbh};

    my ($filename) = @$sig_files;
    my $sigfile;
    if (ref($filename)) {
        $sigfile = $sig_files->[0];
    }
    else {
        my $rv = F5::YAML::Schema::Utils::parse_yaml( filename => $filename );
        if (!$rv->{completed_successfully}) {
            die {error_message => "Failed parse JSON signatures file: $rv->{error_message}."};
        }

        $sigfile = $rv->{data};
    }

    my %json_params;
    $json_params{file_format} = 'yaml';
    $json_params{file_data} = $sigfile;

    my $object_constructor = sub {
              my (%params) = @_;
              return F5::ASMConfig::Entity::Signature->new(%json_params , dbh => $dbh, object => {})->from_rest(object => $params{object});
          };

    $self->load_serialized_signatures(%params, 
                                      object_constructor => $object_constructor, 
                                      sigfile => $sigfile);
}

#----------------------------------------------------------------------

sub load_binary_signatures {
    my ($self, %params) = @_;
    my $sig_files = $params{sig_files};
    my $dbh  = $params{dbh} || $self->{dbh};

    my ($filename) = @$sig_files;
    my $mp = Data::MessagePack->new();
    my $packed_data = F5::Utils::File::file_slurp($filename);
    my $sigfile = $mp->unpack($packed_data);

    my $object_constructor = sub {
              my (%params) = @_;
              return F5::ASMConfig::Entity::Signature->new(dbh => $dbh, object => $params{object});
          };

    $self->load_serialized_signatures(%params,
                                      object_constructor => $object_constructor,
                                      sigfile => $sigfile);
}


#----------------------------------------------------------------------

sub load_serialized_signatures {
    my ($self, %params) = @_;
    my $sig_update_status  = $params{sig_update_status};
    my $policy_to_sigs     = $params{policy_to_sigs};
    my $object_constructor = $params{object_constructor};
    my $sigfile            = $params{sigfile};
    my $dbh  = $params{dbh} || $self->{dbh};

    my $sig_tag = defined($sigfile->{tag}) ? $sigfile->{tag} : '';
    if (length($sig_tag) > 50) {
        die "Invalid Tag '$sig_tag': Tag is limited to 50 characters.\n";
    }
    my $sig_tag_policy = $self->{sig_tag_policy} || '';

    if ($sig_tag) {
        my $existing_sig_tag = $dbh->selectrow_hashref(<<EndQuery, undef, $sig_tag);
SELECT DISTINCT(sig_tag), MAX(sig_tag_policy) as sig_tag_policy_exists
  FROM $self->{dbname}.NEGSIG_SIGNATURES
 WHERE sig_tag = ?
EndQuery

        if ($existing_sig_tag) {
            if (! $sig_tag_policy && $existing_sig_tag->{sig_tag_policy_exists}) {
                die "Invalid Tag '$sig_tag': Tag is already in use for policy-specific signatures\n";
            }
            elsif ($sig_tag_policy && $existing_sig_tag->{sig_tag} && ! $existing_sig_tag->{sig_tag_policy_exists}) {
                die "Invalid Tag '$sig_tag': Tag is already in use for global signatures\n";
            }
        }
    }

    my $existing_sigs = $dbh->selectall_hashref(<<EndQuery, 'sig_name',undef, $sig_tag, $sig_tag_policy, $self->{flg_is_user_defined});
SELECT sig_name, sig_id, last_update
  FROM $self->{dbname}.NEGSIG_SIGNATURES
WHERE sig_tag = ?
  AND sig_tag_policy = ?
  AND flg_is_user_defined = ?
EndQuery

    my %seen_sig_names;

    my $sigs_last_updated;
    if ( defined $sigfile->{revisionDatetime} ) {
        $sigs_last_updated = F5::DateTimeUtils::convert_epoch_to_datetime( F5::DateTimeUtils::convert_rfc3339_to_epoch ($sigfile->{revisionDatetime}) );
        $self->validate_signature_requirements_for_policies(
                                                            sig_tag => $sigfile->{tag},
                                                            timestamp => $sigfile->{revisionDatetime},
                                                            flg_is_user_defined => $self->{flg_is_user_defined},
                                                           );
    }

    if (F5::BigipUtils::is_feature_enabled('validate_software_version')) {
        my $current_bigip_ver = F5::BigipVersionUtils::bigip_version();

        if ( defined $sigfile->{softwareVersion} ) {
            if (($sigfile->{softwareVersion} !~ /^\d{1,4}\.\d{1,4}\.\d{1,4}(\.|$)/) ){
                die {error_message => "'softwareVersion' format is incorrect."};
            }
            elsif ( F5::BigipVersionUtils::compare_versions(version1 => $sigfile->{softwareVersion},
                                                            version2 => $current_bigip_ver) > 0 ){
                die {error_message => "'softwareVersion' should be not higher than $current_bigip_ver"};
            }
        }
    }

    if (length($sig_tag)) {
        $self->{sig_tag} = $sig_tag;
    }
    if (defined $sigs_last_updated ) {
        $self->{timestamp} = $sigs_last_updated;
    }

    if (defined($sigfile->{readme})) {
        $self->{readme} = $sigfile->{readme};
    }

    my $batched_insertor = F5::BatchInsert->new(dbh => $dbh);

    if (!$self->{flg_is_user_defined} && $sigfile->{signatureSystems}) {
        my $existing_systems = $dbh->selectall_hashref(<<EndQuery, 'system_name');
SELECT system_name
  FROM $self->{dbname}.NEGSIG_SYSTEMS
EndQuery

        for my $signature_system ( @{$sigfile->{signatureSystems}} ) {
            if (!$existing_systems->{$signature_system->{system_name}}) {
                my $system = F5::ASMConfig::Entity::SignatureSystem->new(dbh => $dbh, object => $signature_system);
                $system->store_add(batched_insertor => $batched_insertor);
            }
        }
        $batched_insertor->flush;
    }

    my $lookup_by = ($self->{flg_is_user_defined} ? 'sig_name' : 'sig_id');
    for my $data_sig ( @{$sigfile->{signatures}} ) {
        if ( defined $data_sig->{tag}){
            die {error_message => "Tag should not be specified per signature."};
        }

        if ( length $sig_tag ) {
            $data_sig->{tag} = $sig_tag;
        }

        my $sig = $data_sig;
        if ($object_constructor) {
            $sig = $object_constructor->(object => $data_sig);
        }
        $seen_sig_names{$sig->{sig_name}} = 1;

        if ( !defined $sig->{attack_type_id}) {
            die { error_message => "Attack type name must be specified."};
        }

        if ($sig_tag_policy) {
            $sig->{sig_tag_policy} = $sig_tag_policy;
        }

        if ( $self->{flg_is_user_defined} && defined $sigs_last_updated ){
            $sig->{last_update} = $sigs_last_updated;
        }

        my $existing_sig = $existing_sigs->{$sig->{sig_name}};
        if ($existing_sig && $sig->{last_update} && $sig->{last_update} eq $existing_sig->{last_update}) {
            $sig_update_status->{$existing_sig->{sig_id}}->{update_status} = $SIG_UPDATE_STATUS_NO_CHANGE;
            next;
        }

        $sig->{sig_systems} = ();

        if( defined $sig->{systems} ){
            for my $system ( @{$sig->{systems}} ) {
                push(@{$sig->{sig_systems}}, $system->{system_name});
            }
        }

        if ($sig->{sig_id} && !$self->{flg_is_user_defined}) {
            my $update_status = ($existing_sig || $sig_update_status->{$sig->{sig_id}}) ? $SIG_UPDATE_STATUS_UPDATE 
                                                                                        : $SIG_UPDATE_STATUS_ADDED;
            $sig_update_status->{$sig->{sig_id}} = { update_status => $update_status,
                                                     sig_id        => $sig->{sig_id},
                                                     sig_name      => $sig->sig_name,
                                                     sig_tag       => $sig_tag,
                                                     sig_tag_policy => $sig_tag_policy,
                                                     rev           => $sig->{rev}, 
                                                   };
            if ($update_status eq $SIG_UPDATE_STATUS_UPDATE) {
                my $old_sig = F5::ASMConfig::Entity::Signature->new(dbh => $self->{dbh}, object => { sig_id => $sig->{sig_id} });
                $old_sig->store_update(store_obj => $sig);
            }
            else {
               $sig->store_add();
               $sig_update_status->{$sig->{sig_id}}->{update_status} = $SIG_UPDATE_STATUS_ADDED;
            }
            next;
        }

        my $sig_update = F5::NegativeSignatures::Signature->update_if_changed(dbh => $self->{dbh},
                                                                              dbname => $self->{dbname},
                                                                              flg_is_user_defined => $self->{flg_is_user_defined},
                                                                              object => $sig,
                                                                              sigs_last_updated => $sigs_last_updated,
                                                                              lookup_by => $lookup_by,
                                                                              policy_to_sigs => $policy_to_sigs,
                                                                              copy_all_fields => 1,
                                                                              skip_lookup => !$existing_sig,
                                                                           );

        $sig_update_status->{$sig_update->{sig_id}} = $sig_update if($sig_update);
    }

    if ($self->{flg_is_user_defined} && length $sig_tag) {
        if (!%seen_sig_names) {
            # 0 signatures are imported. Check if any policy requires the given tag
            my @violating_sig_reqs = F5::Utils::Policy::get_policy_signature_requirements( dbh => $self->{dbh},
                                                                                           sig_tags => [$sig_tag] );
             if (@violating_sig_reqs) {
                 my $plc_string = (scalar @violating_sig_reqs) > 1 ? "policies" : "policy";
                 my $error_msg = "Can't delete tag '$sig_tag'. Following $plc_string require it:\n";
                 foreach my $sig_req (@violating_sig_reqs) {
                     $error_msg .= $sig_req->{policy}->{policy_name} . "\n";
                 }
                 die {error_message => $error_msg};
             }
        }

        for my $existing_sig_name (keys %$existing_sigs ) {
            next if $seen_sig_names{$existing_sig_name};
            my $deleted_sig = F5::NegativeSignatures::Signature->new(dbh => $dbh,
                                                                     sig_name => $existing_sig_name,
                                                                     sig_tag => $sig_tag,
                                                                     sig_tag_policy => $existing_sigs->{$existing_sig_name}->{sig_tag_policy},
                                                                  );
            $deleted_sig->delete();
            $sig_update_status->{$deleted_sig->sig_id}->{update_status} = $SIG_UPDATE_STATUS_DELETED;
        }
    }

}

#----------------------------------------------------------------------

sub parse_sig {
  my($self, %params) = @_;

  my $elm = $params{elm};
  my $twig = $params{twig};
  my $sig_update_status = $params{sig_update_status};
  my $policy_to_sigs = $params{policy_to_sigs};

  my $lookup_by = ($self->{flg_is_user_defined} ? 'sig_name' : 'sig_id');
  my $sig_update = F5::NegativeSignatures::Signature->update_if_changed(dbh => $self->{dbh},
                                                                        dbname => $self->{dbname},
                                                                        flg_is_user_defined => $self->{flg_is_user_defined},
                                                                        xml_element => $elm,
                                                                        lookup_by => $lookup_by,
									policy_to_sigs => $policy_to_sigs,
                                                                       );

  $sig_update_status->{$sig_update->{sig_id}} = $sig_update if($sig_update);

  $twig->purge;
  return 1;
}

#----------------------------------------------------------------------

sub get_recursive_files {
  my($dir) = @_;

  my @filelist;
  my $find_func = sub {
    if (! -d $File::Find::name) {
      push(@filelist, $File::Find::name);
    }
  };

  find({ wanted => $find_func,
         no_chdir => 1 },
       $dir);

  return @filelist;
}

#----------------------------------------------------------------------

sub validate_version {
    my($self) = @_;

    if (! -e $self->get_timestamp_file) {
        my $error_msg = "The uploaded Attack signature update file does not match the current version of BIG-IP.";
        die {
              error_code   => F5::ASMConfig::ErrorCode::INVALID_SIGFILE_VERSION(),
              display_message => $error_msg,
              error_message => $error_msg,
              error_level  => 'info',
            };
    }
}

#----------------------------------------------------------------------

sub validate_timestamp {
    my($self, %params) = @_;

    my $current_timestamp = F5::NegativeSignatures->new(dbh => $self->{dbh})->get_update_file_date;
    if ($self->timestamp lt $current_timestamp && ! $params{allow_older_timestamp}) {
        my $error_msg = "Attack signature update file is older than currently installed version";
        die {
              error_code   => F5::ASMConfig::ErrorCode::OLDER_SIGFILE_TIMESTAMP(),
              display_message => $error_msg,
              error_message => $error_msg,
              error_level  => 'info',
            };
    }
}

#----------------------------------------------------------------------

sub update_common_schema {
    my($class_or_self, %params) = @_;

    my $new_schema_filename;
    if (ref($class_or_self)) {
        $new_schema_filename = "$class_or_self->{base_dir}/schema/common.xsd";
    }
    else {
        $new_schema_filename = $params{new_schema_filename};
    }
    return unless ($new_schema_filename && -f $new_schema_filename);

    my $target_schema_filename = $F5::NegativeSignatures::Utils::NEGSIG_COMMON_SCHEMA_FILENAME;

    my $target_schema_highest_system_id = 0;
    if (-f $target_schema_filename) {
        my $target_schema_attributes = F5::NegativeSignatures::Utils::get_schema_attributes($target_schema_filename);
        $target_schema_highest_system_id = List::Util::max(values %{$target_schema_attributes->{internal_ids}->{system_name}});
    }

    my $new_schema_attributes = F5::NegativeSignatures::Utils::get_schema_attributes($new_schema_filename);
    my $new_schema_highest_system_id = List::Util::max(values %{$new_schema_attributes->{internal_ids}->{system_name}});

    if ($new_schema_highest_system_id > $target_schema_highest_system_id) {
        $class_or_self->{logger}->debug('Common schema for signatures has been updated') if (ref($class_or_self) && $class_or_self->{logger});
        copy($new_schema_filename, $target_schema_filename);
    }
}

#----------------------------------------------------------------------

sub get_timestamp_file {
    my($self) = @_;
    return "$self->{base_dir}/timestamps/" . F5::BigipVersionUtils::bigip_version() . ".timestamp";
}

#----------------------------------------------------------------------

sub timestamp {
  my($self) = @_;

  if ($self->{timestamp}) {
      return $self->{timestamp};
  }
  my $timestamp;
  my $timestamp_filename = $self->get_timestamp_file;
  if (-e $timestamp_filename) {
    $timestamp = read_file_contents($timestamp_filename);
    $timestamp =~ s|/|-|g;
  }
  else {
    my($last_update) = $self->{dbh}->selectrow_array("SELECT MAX(last_update) FROM $self->{dbname}.NEGSIG_SIGNATURES WHERE flg_is_user_defined = 0");
    $timestamp = $last_update;
  }
  $self->{timestamp} = $timestamp;
  return $timestamp;
}

#----------------------------------------------------------------------

sub readme {
  my($self) = @_;

  if ($self->{readme}) {
      return $self->{readme};
  }

  my $readme_dir = $self->readme_dir;
  return unless $readme_dir;
  my $readme_header_dir = $self->top_readme_dir;
  my $readme_txt = read_file_contents("$readme_header_dir/readme_header");

  my $last_sigfile = $self->{dbh}->selectrow_hashref(<<EndQuery);
  SELECT * 
    FROM $self->{dbname}.SIGFILE_STATUS
   WHERE flg_is_user_defined = 0
ORDER BY sigfile_status_id DESC
   LIMIT 1
EndQuery

  if ($last_sigfile && $last_sigfile->{sigfile_timestamp} && $self->timestamp lt $last_sigfile->{sigfile_timestamp}) {
    $readme_txt .= "\n\nThis signature update file is older than the previously loaded file.\n";
  }
  elsif ($last_sigfile && $last_sigfile->{sigfile_timestamp} && $self->timestamp eq $last_sigfile->{sigfile_timestamp}) {
    $readme_txt .= "\n\nThis signature update file is identical to the previously loaded file.\n";
  }
  else {
    $readme_txt .= "\n\n";

    my @readme_files = reverse sort grep { ! /readme_header/ } glob("$readme_dir/*");
    for my $readme_file (@readme_files) {
      last if ($last_sigfile->{last_added_readme_file} &&
               basename($readme_file) le $last_sigfile->{last_added_readme_file});

      $readme_txt .= read_file_contents($readme_file) . "\n";
    }
  }
  return $readme_txt;
}
#---------------------------------------------------------------------

sub top_readme_dir {
  my($self) = @_;

  my @readme_dirs;
  if ($self->{negsig_version_readmes_specific_dir} && -d $self->{negsig_version_readmes_specific_dir}) {
      @readme_dirs = grep { m|/[\d.]+$| && -d $_ } glob("$self->{negsig_version_readmes_specific_dir}/*");
  }
  else {
      @readme_dirs = grep { m|/[\d.]+$| && -d $_ } glob("$self->{base_dir}/readmes/*");
  }

    my $readme_dir = F5::BigipVersionUtils::find_by_version(elements => \@readme_dirs,
                                                           version_extract => sub {
                                                            my($version) = @_;
                                                            $version =~ m|/([\d.]+)|;
                                                            return $1;
                                                          },
                                                         );
  return $readme_dir;
}

#----------------------------------------------------------------------
sub readme_dir {
  my($self) = @_;

  my @readme_dirs;
  my @all_readme_dirs;

  if ($self->{negsig_version_readmes_specific_dir} && -d $self->{negsig_version_readmes_specific_dir}) {
     @all_readme_dirs = grep { m|/[\d.]+$| && -d $_ } glob("$self->{negsig_version_readmes_specific_dir}/*");
  }
  else {
     @all_readme_dirs = grep { m|/[\d.]+$| && -d $_ } glob("$self->{base_dir}/readmes/*");
  }

  for my $dir (@all_readme_dirs){
      my @version_readme_files = glob("$dir/*");
      if (@version_readme_files) {
          push(@readme_dirs, $dir);
      }
  }

  my $readme_dir = F5::BigipVersionUtils::find_by_version(elements => \@readme_dirs,
                                                          version_extract => sub {
                                                            my($version) = @_;
                                                            $version =~ m|/([\d.]+)|;
                                                            return $1;
                                                          },
                                                         );
  return $readme_dir;
}

#----------------------------------------------------------------------

sub last_added_readme_file {
  my($self) = @_;

  my $readme_dir = $self->readme_dir;
  return unless $readme_dir;
  my @readme_files = reverse sort grep { ! /readme_header/ } glob("$readme_dir/*");
  if (@readme_files) {
    return basename($readme_files[0]);
  }
  return undef;
}

#----------------------------------------------------------------------

sub read_file_contents {
  my($filename) = @_;

  return '' unless ($filename && -e $filename);
  open(my $fh, $filename);
  my $contents = join('', <$fh>);
  close $fh;

  return $contents;
}

#----------------------------------------------------------------------

sub get_latest_update_per_version {
  my($sigs_dir) = @_;

  my %version_last_update;
  my $xml_simple = XML::Simple->new( GroupTags => { 'systems' => 'system_name' },
                                     ForceArray => [ 'system_name', 'rev' ],
                                     SuppressEmpty => '',
                                   );
  for my $filename (get_recursive_files($sigs_dir)) {
    next unless $filename =~ /\.xml$/;

    my $sig_all_revs;
    eval {
      $sig_all_revs = $xml_simple->XMLin($filename);
    };

    if ($@) {
      print "Error in parsing attack signature $filename: $@\n";
      next;
    }

    for my $sig_rev (@{$sig_all_revs->{rev}}) {
      my $last_update = $sig_rev->{last_update};
      $last_update =~ s/\$DateTime: //;
      $last_update =~ s/ \$//;
      if (!exists $version_last_update{$sig_rev->{bigip_version_min}} || 
          $version_last_update{$sig_rev->{bigip_version_min}} lt $last_update) {
        $version_last_update{$sig_rev->{bigip_version_min}} = $last_update;
      }
    }
  }
  return %version_last_update;
}

#----------------------------------------------------------------------


my %create_sigfile_dispatch = ( 'XML' => 'create_xml_sigfile',
                                $SIGFILE_FORMAT_BINARY => 'create_binary_sigfile',
                                'JSON'  => 'create_json_sigfile',
                              );

sub create_sigfile {
  my ($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $sig_iterator = $params{sig_iterator};

  if (my $func = $create_sigfile_dispatch{$self->{sigfile_format}}) {
      my $bigip_version = F5::BigipVersionUtils::bigip_version();
      $self->$func(bigip_version => $bigip_version,
                   sig_iterator => $sig_iterator);
  }
  else {
      die "Cannot create sigfile in $self->{sigfile_format} format.\n";
  }
}

#----------------------------------------------------------------------

sub create_xml_sigfile {
  my ($self, %params) = @_;

  my $bigip_version = $params{bigip_version};
  my $sig_iterator  = $params{sig_iterator};

  my $sig_tag = $self->{sig_tag} ? "tag=\"$self->{sig_tag}\"" : "";
  open(my $sigfile_fh, ">", $self->{output_filename});

  print $sigfile_fh <<EndXML;
<?xml version="1.0" encoding="utf-8"?>
<signatures export_version="$bigip_version" $sig_tag>
EndXML

  local $" = ',';
  my @sig_xml_fields = qw(sig_name rule last_update apply_to risk accuracy );
  my @other_sig_fields = qw(sig_desc rev sig_id);

  while(my $sig = $sig_iterator->get_next) {
    $sig->{doc} = $sig->{sig_desc};
    $sig->{attack_type} = $sig->{attack_type_name};

    my @sig_xml = map { XML::Twig::Elt->new( $_ => $sig->{$_}) } (@sig_xml_fields, 'doc', 'attack_type');

    if ($sig->{systems} && @{$sig->{systems}}) {
      push(@sig_xml,
           XML::Twig::Elt->new( systems => 
                                map { XML::Twig::Elt->new( system_name => $_) } @{$sig->{systems}})
          );
    }

    my $full_sig_xml = XML::Twig::Elt->new( 'sig', 
                                            { id => $sig->{sig_id} },
                                            XML::Twig::Elt->new( 'rev', 
                                                                 { num => $sig->{rev} },
                                                                 @sig_xml));
    $full_sig_xml->print($sigfile_fh, 'indented');
  }

  print $sigfile_fh "</signatures>\n";
  close $sigfile_fh;
}

#----------------------------------------------------------------------
sub create_binary_sigfile {
  my ($self, %params) = @_;

  require Data::Structure::Util;

  my $bigip_version = $params{bigip_version};
  my $sig_iterator  = $params{sig_iterator};

  my @signatures = $sig_iterator->get_all();

  F5::ObjUtils::remove_unfreezable_references(\@signatures);
  Data::Structure::Util::unbless(\@signatures);
 
  my $sigfile = {
                  signatures => \@signatures,
                };

  if ($self->{flg_is_user_defined}) {
      $sigfile->{revisionDatetime} =  F5::DateTimeUtils::convert_epoch_to_rfc3339(time());
      $sigfile->{softwareVersion}  = $bigip_version;
      $sigfile->{tag} = $self->{sig_tag} if $self->{sig_tag};
  }
  else {
      my $last_sigfile = $self->{dbh}->selectrow_hashref(<<EndQuery);
SELECT sigfile_timestamp, readme
    FROM $self->{dbname}.SIGFILE_STATUS
   WHERE flg_is_user_defined = 0
ORDER BY sigfile_status_id DESC
   LIMIT 1
EndQuery

      $sigfile->{revisionDatetime} = F5::DateTimeUtils::convert_epoch_to_rfc3339(F5::DateTimeUtils::convert_datetime_to_epoch($last_sigfile->{sigfile_timestamp}));
      $sigfile->{readme} = $last_sigfile->{readme};
      my @signature_systems = F5::ASMConfig::Entity::SignatureSystem->retrieve_all(dbh => $self->{dbh}, get_criteria => {});
      F5::ObjUtils::remove_unfreezable_references(\@signature_systems);
      Data::Structure::Util::unbless(\@signature_systems);
      $sigfile->{signatureSystems} = \@signature_systems;
  }

  my $dir = "$self->{temp_base_dir}/sigfile.tmp.$$";
  my $tmp_filename = "signatures.bin";
  mkpath $dir;

  open(my $sigfile_fh, ">", "$dir/$tmp_filename");
  my $mp = Data::MessagePack->new();
  my $packed_data = $mp->pack($sigfile);
  print $sigfile_fh $packed_data;
  close $sigfile_fh;

  if ($self->{flg_is_user_defined}) {
      mv("$dir/$tmp_filename", $self->{output_filename});
  }
  else {
      delete $sigfile->{signatures};
      delete $sigfile->{signatureSystems};
      delete $sigfile->{readme};

      my $zipped_filename = $tmp_filename . '.tgz';
      my $metadata_filename = "signature_update.yaml";

      $sigfile->{checksum} = Digest::MD5::md5_base64($packed_data);
      $sigfile->{filename} = $zipped_filename;
      YAML::XS::DumpFile("$dir/$metadata_filename", $sigfile);

      my @cmd = ('tar', '-zcf', "$dir/$zipped_filename", '-C', $dir,  $tmp_filename);
      my $res =  F5::SystemUtils::run_system_tool( command_line_params => \@cmd );

      @cmd = ('tar', '-zcf', $self->{output_filename}, '-C', $dir, $metadata_filename, $zipped_filename);
      $res =  F5::SystemUtils::run_system_tool( command_line_params => \@cmd );
  }
  rmtree($dir);
}

#----------------------------------------------------------------------

sub create_json_sigfile {
  my ($self, %params) = @_;

  require Data::Structure::Util;

  my $bigip_version = $params{bigip_version};
  my $sig_iterator  = $params{sig_iterator};
  my $sigfile;
  my @rest_fields = @{$self->{sig_report_fields}} if exists $self->{sig_report_fields};
  my %rest_selected_fields;
  my $filename = $self->{output_filename};
  my $rv;

  $sigfile = { signatures => [],};

if ($self->{flg_is_user_defined}) {
      $sigfile->{revisionDatetime} =  F5::DateTimeUtils::convert_epoch_to_rfc3339(time());
  }
  else {
      my $last_sigfile = $self->{dbh}->selectrow_hashref(<<EndQuery);
SELECT sigfile_timestamp
    FROM $self->{dbname}.SIGFILE_STATUS
   WHERE flg_is_user_defined = 0
ORDER BY sigfile_status_id DESC
   LIMIT 1
EndQuery

  $sigfile->{revisionDatetime} = F5::DateTimeUtils::convert_epoch_to_rfc3339(F5::DateTimeUtils::convert_datetime_to_epoch($last_sigfile->{sigfile_timestamp}));
  }
  $sigfile->{tag} = $self->{sig_tag} if $self->{sig_tag};
  $sigfile->{softwareVersion} = $bigip_version unless $self->{sig_report_fields}; # no need of bigip_version in NAP for now

  @rest_fields = @SIGFILE_FIELDS unless @rest_fields;

  %rest_selected_fields = map { ( $_ => {} ) } @rest_fields;

  $rest_selected_fields{systems} = {name => {} } if $rest_selected_fields{systems};

  while (my $sig = $sig_iterator->get_next()) {
      my $rest_sig = $sig->to_rest(rest_selected_fields => \%rest_selected_fields,
                                   dbh => $self->{dbh},
                                   logger => $self->{logger},);
      # TODO: use to_declarative
      if ($rest_sig->{attackTypeReference}) {
          $rest_sig->{attackType} = delete $rest_sig->{attackTypeReference};
          delete $rest_sig->{attackType}->{link};
      }
      delete $rest_sig->{selfLink};
      delete $rest_sig->{references} unless exists($rest_sig->{references}[0]);
      push (@{$sigfile->{signatures}}, $rest_sig);
  }

  my $json = JSON->new->convert_blessed->allow_blessed;
  my $content = $json->encode($sigfile);
  decode ('utf8', $content);
  File::Slurp::write_file($filename, $content);
  $rv->{file_size} = -s $filename;
  $rv->{filename} = $filename;

  return $rv;
}


#----------------------------------------------------------------------
sub check_user_sigfile_schema {
  my($self) = @_;

  if ( $self->{sigfile_format} ne 'XML' ) {
      return;
  }

  my @params = ( $XMLLINT_CMD, '--schema', F5::Deploy::map_path('/ts/var/schema/negsig/sigfile.xsd'), '--noout', $self->{filename});
  my $res =  F5::SystemUtils::run_system_tool( command_line_params => \@params );
  if ($res && !$res->{completed_successfully}) {
     die "User signature file does not conform to schema ($res->{standard_error})\n";
  }
}

#----------------------------------------------------------------------

sub compute_md5_for_file {
  my($filename) = @_;

  open(my $fh, $filename);
  my $md5 = Digest::MD5->new->addfile($fh)->digest;
  close $fh;

  return $md5;
}

#----------------------------------------------------------------------

sub load_botsigs {
    my($self, %params) = @_;

    return F5::Dos::BotSignatures::load_from_xml(dir => "$self->{base_dir}/botsigs",
                                                 logger => $self->{logger},
                                                );
}

#----------------------------------------------------------------------
sub load_server_technologies {
    my($self, %params) = @_;

    my $dir = "$self->{base_dir}/servtech";
    return F5::ServerTechnologies::load_from_xml(dir => $dir,
                                                 dbh => $self->{dbh},
                                                );

}

#----------------------------------------------------------------------
sub validate_signature_requirements_for_policies {
    my($self, %params) = @_;

    return 1 unless $self->{validate_requirements};   

    my $timestamp = $params{timestamp} || $self->timestamp;
    my $sig_tag = $params{sig_tag} || '';
    my $flg_is_user_defined = $params{flg_is_user_defined};
    my $dbh = $self->{dbh};
    my $error_msg;

    if ($flg_is_user_defined) {
        $error_msg = "revisionDatetime [$timestamp] for tag '$sig_tag' fails Policy Signature Requirements. Failed for:";
    }
    else {
        $error_msg = "Signature file timestamp [$timestamp] fails Policy Signature Requirements. Failed for:";
    }

    my @violating_sig_reqs = F5::Utils::Policy::get_policy_signature_requirements( dbh => $dbh,
                                                                                   datetime => $timestamp,
                                                                                   sig_tags => [$sig_tag] );

    if (@violating_sig_reqs) {
        foreach my $sig_req (@violating_sig_reqs) {
            $error_msg .= "\nPolicy: $sig_req->{policy}->{policy_name} -";
            if ( defined ($sig_req->{min_revision_datetime}) ) {
                my $min_revision_datetime = F5::DateTimeUtils::convert_epoch_to_datetime($sig_req->{'min_revision_datetime'});
                $error_msg .= " minRevisionDatetime [" . $min_revision_datetime . "]";
            }
            if ( defined ($sig_req->{max_revision_datetime}) ) {
                my $max_revision_datetime = F5::DateTimeUtils::convert_epoch_to_datetime($sig_req->{'max_revision_datetime'});
                $error_msg .= " maxRevisionDatetime [" . $max_revision_datetime . "]";
            }
        }
        die {error_msg => $error_msg};
    }
    return 1;
}

1;
