package F5::NegativeSignatures::SigFile::Converter;

use warnings;
use strict;

use Encode;
use JSON;
use XML::Simple;
use File::Slurp;

use F5::DateTimeUtils;
use F5::SystemUtils;
use F5::Deploy;

## Note: This function is purposefully being designed with as few external framework dependencies as possible 
##       with the intention of allowing this functionality to be easily ported to a standalone tool.

my %LEVEL_MAP = ( 1 => 'low', 2 => 'medium', 3 => 'high' );
my $XMLLINT_CMD = '/usr/bin/xmllint';
my $schema_file = F5::Deploy::get_path_by_type('var') . '/schema/negsig/sigfile.xsd';
my $default_tag = 'user-defined-signatures';

#Currently only XML -> JSON supported
sub convert_signatures {
    my (%params) = @_;
    my $xml_file = $params{import_file};
    my $json_file = $params{export_file};
    my $tag = $params{tag};
    my $rv = { completed_successfully => 0 };

    my @cmd = ( $XMLLINT_CMD, '--schema', $schema_file, '--noout', $xml_file);
    my $res =  F5::SystemUtils::run_system_tool( command_line_params => \@cmd );
    if ($res && !$res->{completed_successfully}) {
       $rv->{error_message} = "User signature file does not conform to schema ($res->{standard_error})";
       return $rv;
    }

    my $xml_sigs;
    eval {
        $xml_sigs = XMLin($xml_file,
                          ForceArray => [ 'system_name', 'rev', 'reference' ]
                         );
    };

    if (!$xml_sigs) {
        $rv->{error_message} = "Invalid XML signature file $xml_file";
        return $rv;
    }

    # If relevant, BIG-IP Version would be found in: $xml_sigs->{export_version}
    my $json_doc;
    $json_doc->{tag} = defined $tag ? $tag :
                       defined $xml_sigs->{tag} ? $xml_sigs->{tag} :
                       $default_tag;
    $json_doc->{signatures} = [];

    #Fixup for file with one signature
    my $signatures = $xml_sigs->{sig};
    if (exists $signatures->{id}) {
        my $id = delete $signatures->{id};
        my $rev = delete $signatures->{rev};
        $signatures->{$id}->{rev} = $rev;
    }

    for my $signature_id (keys %$signatures) {
        my $xml_signature = $signatures->{$signature_id}->{rev}->[0];
        my $last_update = F5::DateTimeUtils::convert_datetime_to_epoch($xml_signature->{last_update}) * 1000000;
        my $description = ref $xml_signature->{doc} ? '' : $xml_signature->{doc};
        my @systems;
        if ($xml_signature->{systems} && $xml_signature->{systems}->{system_name}) {
            @systems = map { { "name" => $_ } }  @{$xml_signature->{systems}->{system_name}} ;
        }
        my $json_signature = { signatureId      => $signature_id,
                               signatureType    => lc($xml_signature->{apply_to}),
                               name             => $xml_signature->{sig_name},
                               rule             => $xml_signature->{rule},
                               lastUpdateMicros => $last_update,
                               attackType       => { name => $xml_signature->{attack_type} },
                               revision         => $xml_signature->{num},
                               risk             => $LEVEL_MAP{$xml_signature->{risk}},
                               accuracy         => $LEVEL_MAP{$xml_signature->{accuracy}},
                               description      => $description,
                               systems          => \@systems,
                             };
       push @{$json_doc->{signatures}}, $json_signature;
    }

    $rv->{completed_successfully} = 1;

    my $json = JSON->new->convert_blessed->allow_blessed;
    $rv->{content} = $json->encode($json_doc);
    decode ('utf8', $rv->{content});
    if ($json_file) {
       File::Slurp::write_file($json_file, $rv->{content});
       $rv->{file_size} = -s $json_file;
       $rv->{filename} = $json_file;
    }

    return $rv;
}

1;
