package F5::NegativeSignatures::SigFile::Update;

use strict;
use warnings;

use base qw(F5::Sigfile::Update);

use Time::Local;
use File::Copy;
use F5::Deploy;
use F5::ASMConfig::Entity::SigfileUpdateRes;
use F5::NegativeSignatures::Signature qw(:constants);
use F5::NegativeSignatures::SigFile;
use F5::NegativeSignatures::SigFile::AutoDownload;
use F5::NegativeSignatures;
use F5::NegativeSignatures::Utils  qw(find_or_add_attack_type);
use F5::PrepareConf::Policy;
use F5::SetActive;
use F5::Provisioning;
use F5::Syslog;
use F5::Cfg;
use F5::BatchInsert;
use F5::DbUtils;
use F5::BigipUtils;
use F5::LockFactory;
use JSON;

use constant  PRPLC_CFG_FILE  =>  'prepare_policy.cfg';
use constant HTTP_RESPONSE_SPLITTING_ATTACK_TYPE_ID => 32;


my %wh_third_party_attacks = (
    '(WH) XSS'     => 'Cross Site Scripting (XSS)',
    '(WH) SQLI'    => 'SQL-Injection',
    '(WH) COMMAND INJECTION' => 'Server Side Code Injection',
    '(WH) XPATH INJECTION' => 'XPath Injection',
    '(WH) HTTP RESPONSE SPLITTING' => 'HTTP Response Splitting',
);

my $IM_FILE_OUTPUT = 'ISO 9660 CD-ROM';

our $SIGFILE_TYPE = "Attack Signatures";
our $SAVE_DIR = F5::Deploy::map_path("/ts/var/negsig");

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;
  $self->SUPER::init(%params);
  $self->{default_rv} = { sig_count => 0,
                          sigs_updated_count => 0,
                        };
  $self->{sigfile_type} = $SIGFILE_TYPE;
  $self->{save_dir} = $SAVE_DIR;
}

#----------------------------------------------------------------------

sub get_schedule_info {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
#Gets auto_update_interval, update_mode
    my $rv = $dbh->selectrow_hashref(<<EndQuery);
SELECT * 
    FROM PLC.SIGFILE_FLAGS
EndQuery

    my $unix_timestamp_string = F5::DbUtils::get_unix_timestamp_string(dbh => $dbh, field => 'attempt_datetime');  
    ($rv->{last_update_attempt}) =  $dbh->selectrow_array(<<EndQuery);
SELECT $unix_timestamp_string
    FROM PLC.SIGFILE_AUTO_UPDATE_HISTORY
ORDER BY sigfile_auto_update_id DESC
   LIMIT 1
EndQuery

    return ($rv);
          
}

#----------------------------------------------------------------------

sub apply_changes {
  my($self, %params) = @_;

  my $automatic_mode = $params{automatic_mode};
  my $filename = $params{filename};
  my $file_contents_struct = $params{file_contents_struct};
  my $flg_is_user_defined = $params{flg_is_user_defined} || 0;
  my $flg_is_genesis_file = $params{flg_is_genesis_file};
  my $asm_config_handler = $params{asm_config_handler};
  my $username = $params{username};
  my $check_interval = $params{check_interval};
  my $update_mode = $params{update_mode};
  my $sigfile_status_update_mode = $params{sigfile_status_update_mode};
  my $sigfile_format = $params{sigfile_format};
  my $skip_set_active = $params{skip_set_active};
  my $skip_set_sigs_active    = $params{skip_set_sigs_active};
  my $sig_tag_policy = $params{sig_tag_policy};
  my $validate_requirements = $params{validate_requirements};

  if ($params{schedules} && @{$params{schedules}}) {
      my ($schedule_info) = @{$params{schedules}};
  }

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  my $snapshot_before_update = $self->wh_signatures_snapshot();
  my $whitehat_policies_without_response_splitting = $dbh->selectcol_arrayref(<<EndQuery, undef, HTTP_RESPONSE_SPLITTING_ATTACK_TYPE_ID);
SELECT DISTINCT p.id
  FROM PLC.PL_POLICIES AS p
 INNER JOIN PLC.PL_POLICY_NEGSIG_SIGNATURES AS pn
    ON p.id = pn.policy_id
 INNER JOIN PLC.NEGSIG_SIGNATURES AS ns
    ON pn.sig_id = ns.sig_id
 INNER JOIN PLC.PL_SCANNER_CONFIG sc
    ON p.id = sc.policy_id
 WHERE ns.attack_type_id = ?
   AND sc.scanner_type != ''
   AND pn.flg_disabled = 1
EndQuery

  my $sigfile;
  my $sig_update_status = {};
  my $botsigs_updated_result = {};
  my $server_technologies_updated_result = {};
  my $policy_to_sigs = {};
  eval {
    $sigfile = F5::NegativeSignatures::SigFile->new(dbh => $dbh,
                                                    filename => $filename,
                                                    file_contents_struct => $file_contents_struct,
                                                    sigfile_format => $sigfile_format,
                                                    flg_is_user_defined => $flg_is_user_defined,
                                                    sig_tag_policy => $sig_tag_policy,
                                                    validate_requirements => $validate_requirements,
                                                   );

    if (! $flg_is_user_defined && !$sigfile_format ) {
        $sigfile->validate_version(%params) unless ($flg_is_genesis_file);
        $sigfile->validate_timestamp(%params);
        $sigfile->validate_signature_requirements_for_policies(%params);
        $sigfile->update_common_schema(%params);
    }
 
    if(!defined ($params{update_type}) || $params{update_type} eq F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES()) {
        $sig_update_status = $sigfile->load_sigs(policy_to_sigs => $policy_to_sigs);
        $logger->debug("Signatures loaded to database") if $logger;
    }

    if (! $flg_is_user_defined) {
        if(!defined $params{update_type}) {
            $botsigs_updated_result = $sigfile->load_botsigs();
        }
        if(!defined ($params{update_type}) || $params{update_type} eq F5::ASMConfig::UpdateType::UPDATE_TYPE_SERVER_TECHNOLOGIES()) {
            $server_technologies_updated_result = $sigfile->load_server_technologies();
            $logger->debug("Server technologies loaded to database") if $logger;
        }
        if(!defined $params{update_type}) {
            my @clntcap_files = glob("$sigfile->{base_dir}/clntcap/*");     
            # There should be only 0 or 1 -- nonetheless, accept generously whatever clntcap files are provided
            for my $clntcap_file (@clntcap_files) {
                $self->load_file_for_datasyncd(orig_filename => $clntcap_file,
                                               version_create_time => F5::DateTimeUtils::convert_datetime_to_epoch($sigfile->timestamp()));
            }
        }
    }

    $sigfile->finish;
  };

  if ($@) {
      my $error = $@;
      if (ref($error)) {
          die $error;
      }
      else {
          $error =~ s|\s+ at \s+ /\S+ \s+ line \s+ \d .*||sx;
          $self->store_error_and_die($error, \%params);
      }
  }

  #Get last sigfile status
  my ($sigfile_status) = F5::ASMConfig::Entity::SigfileUpdateRes->retrieve_all(dbh => $dbh,
                                                                               sort_orders => [ F5::ASMConfig::SortOrder->new({field_name => "sigfile_status_id",
                                                                                                                                sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC(),})
                                                                                              ],
                                                                               page_size => 1,
                                                                              );

  my $updated_by;
  if ($self->{license_text}) {
      my ($update_mode) = $dbh->selectrow_array('SELECT update_mode FROM PLC.SIGFILE_FLAGS');
      if ($update_mode ne 'manual') {
          $dbh->do('UPDATE PLC.SIGFILE_FLAGS SET update_mode = ?', undef, 'manual');
          F5::Syslog::ts_syslog( 'info', 'Signature file update mode is now set to manual' );
      }
      $updated_by = 'Enterprise Manager';
  }
  else {
      $updated_by = 'ASM User Interface';
      if ($params{originator}) {
          $updated_by .= " (updated from $params{originator})";
      }
  }

  # This has already been inserted by F5::NegativeSignatures::SigFile::complete, but without the rest_uuid 
  # since it cannot use the Entity there due to a build time dependency of using F5::DBD::CSV
  # So we forcefully update the rest_uuid of the object here during live ASU

  $sigfile_status->store_update(store_obj => {updated_by => $updated_by,
                                              rest_uuid => $sigfile_status->get_rest_uuid()});

  my $sigs_loaded_count = scalar grep($_->{update_status} ne $SIG_UPDATE_STATUS_DELETED, values %$sig_update_status);
  my $sigs_updated_count = scalar grep(($_->{update_status} eq $SIG_UPDATE_STATUS_ADDED ||
                                        $_->{update_status} eq $SIG_UPDATE_STATUS_UPDATE_MINOR ||
                                        $_->{update_status} eq $SIG_UPDATE_STATUS_UPDATE),
                                       values %$sig_update_status);
  my $sigs_updated_minor_count = scalar grep($_->{update_status} eq $SIG_UPDATE_STATUS_UPDATE_MINOR, values %$sig_update_status);
  my $sigs_updated_major_count = scalar grep($_->{update_status} eq $SIG_UPDATE_STATUS_UPDATE, values %$sig_update_status);
  my $sigs_added_count = scalar grep($_->{update_status} eq $SIG_UPDATE_STATUS_ADDED, values %$sig_update_status);
  my $sigs_deleted_count = scalar grep($_->{update_status} eq $SIG_UPDATE_STATUS_DELETED, values %$sig_update_status);

  $sigfile_status->store_update(store_obj => { sigs_updated_count => $sigs_updated_count,
                                               sigs_updated_minor_count => $sigs_updated_minor_count,
                                               sigs_updated_major_count => $sigs_updated_major_count,
                                               sigs_added_count => $sigs_added_count,
                                               sigs_deleted_count => $sigs_deleted_count,
                                               update_mode => $sigfile_status_update_mode,
                                               sig_count => $sigs_loaded_count, 
                                               botsigs_updated_count => $botsigs_updated_result->{updated},
                                               botsigs_added_count => $botsigs_updated_result->{added},
                                               botsigs_deleted_count => $botsigs_updated_result->{deleted},
                                               server_technologies_updated_count => $server_technologies_updated_result->{server_technologies_updated_count},
                                               server_technologies_added_count => $server_technologies_updated_result->{server_technologies_added_count},
                                               server_technologies_deleted_count => $server_technologies_updated_result->{server_technologies_deleted_count},
                                               server_technologies_modified => $server_technologies_updated_result->{server_technologies_modified},
                                             });
  my $change_made = $sigs_updated_count || $sigs_deleted_count || $botsigs_updated_result->{added} || $botsigs_updated_result->{updated} || 
                    $botsigs_updated_result->{deleted} || $server_technologies_updated_result->{server_technologies_updated_count} || 
                    $server_technologies_updated_result->{server_technologies_added_count} || $server_technologies_updated_result->{server_technologies_deleted_count};

  my $batch_insert = F5::BatchInsert->new(dbh => $dbh);

  my %change_levels = (
                       $SIG_UPDATE_STATUS_UPDATE_MINOR => 'minor',
                       $SIG_UPDATE_STATUS_UPDATE => 'major',
                       $SIG_UPDATE_STATUS_ADDED => 'added',
		       $SIG_UPDATE_STATUS_DELETED => 'deleted',
                      );

  my %sig_change_levels = (
		      $SIG_UPDATE_STATUS_UPDATE_MINOR => F5::ASMConfig::SigChangeLevel::SIG_CHANGE_LEVEL_MINOR(),
                      $SIG_UPDATE_STATUS_UPDATE => F5::ASMConfig::SigChangeLevel::SIG_CHANGE_LEVEL_MAJOR(),
                      $SIG_UPDATE_STATUS_ADDED => F5::ASMConfig::SigChangeLevel::SIG_CHANGE_LEVEL_ADDED(),
		      $SIG_UPDATE_STATUS_DELETED => F5::ASMConfig::SigChangeLevel::SIG_CHANGE_LEVEL_DELETED(),
		      );

  my @modified_signatures=();

  for my $sig_id (keys %$sig_update_status){
      my $change_level = $change_levels{$sig_update_status->{$sig_id}->{update_status}};
      if (! $change_level) {
          next;
      }
      my $rec = {
                 sig_id => $sig_id,
                 sig_name => $sig_update_status->{$sig_id}->{sig_name},
                 sig_tag => $sig_update_status->{$sig_id}->{sig_tag} || '',
                 sig_tag_policy => $sig_update_status->{$sig_id}->{sig_tag_policy} || '',
                 change_level => $change_level,
                 sigfile_status_id => $sigfile_status->{sigfile_status_id}
	        };

      if(!$params{dry_run} && $change_level ne 'deleted') { # Avoid a missing foreign key
          $batch_insert->insert_row(table_name => 'PLC.SIGFILE_MODIFIED_SIGNATURES',
                                    rec => $rec);
      }
	
      $rec->{change_level} = $sig_change_levels{$sig_update_status->{$sig_id}->{update_status}};
      push @modified_signatures, $rec;
  }

  if(!defined ($params{update_type}) || $params{update_type} eq F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES()) {
      $sigfile_status->{sigs_modified} = [ map { F5::ASMConfig::SigfileModifiedSignature->new(%$_) } @modified_signatures ];
  }
  
  $batch_insert->flush;
                   
  $dbh->do(<<EndQuery);
DELETE FROM  PLC.PL_PARAM_NEGSIG_SIGNATURES 
 WHERE sig_id NOT IN ( SELECT sig_id FROM  PLC.NEGSIG_SIGNATURES WHERE flg_has_valuecontent_http = 1 )
EndQuery

  $dbh->do(<<EndQuery);
DELETE FROM  PLC.PL_COOKIE_NEGSIG_SIGNATURES
 WHERE sig_id NOT IN ( SELECT sig_id FROM  PLC.NEGSIG_SIGNATURES WHERE flg_has_valuecontent_http = 1 )
EndQuery

  my $success_msg = "$SIGFILE_TYPE updated successfully - $sigs_loaded_count signatures loaded ($sigs_updated_count updated)";
  $self->report_status(completed_successfully => 1,
                       msg => $success_msg,
                      );

  my $snapshot_after_update = $self->wh_signatures_snapshot();
  $self->whitehat_settings($sig_update_status, $snapshot_before_update, $snapshot_after_update, $whitehat_policies_without_response_splitting);

  if (!$params{dry_run} && $change_made) {
      my $negsig = F5::NegativeSignatures->new($dbh);
      $negsig->link_sigs_to_set();
      $negsig->enable_evasion_sigs_on_policies();
      F5::Utils::Rest::populate_uuids(dbh => $dbh,
                                      rest_entities => ['F5::ASMConfig::Entity::SignatureSystem',
                                                        'F5::ASMConfig::Entity::SignatureSystemGroup'],
                                     );
      if (!$skip_set_sigs_active) {     
          $logger->debug("Applying active policy for all accounts") if $logger;
          $self->set_sigs_active(asm_config_handler => $asm_config_handler, policy_to_sigs => $policy_to_sigs, skip_set_active => $skip_set_active);
      }
  }

 
  if($params{dry_run} && !$params{in_wrapping_transaction}) {
      $dbh->rollback;
  }

  $sigfile_status->{completed_successfully} = 1;
  $sigfile_status->{msg} = $success_msg;
  $sigfile_status->{error_msg} = '';
  return ($sigfile_status);
}


#----------------------------------------------------------------------

sub wh_signatures_snapshot {
    my ($self) = @_;

    my $dbh = $self->{dbh};
    my $negsig = F5::NegativeSignatures->new($dbh);
    my $sigs;

    for my $key(keys %wh_third_party_attacks) {
         $sigs->{$key} = [@{$negsig->get_signatures({
            filter => {
                attack_type_id => find_or_add_attack_type (dbh => $dbh, dbname => 'PLC', attack_type_name => $wh_third_party_attacks{$key}),
            }
         })->{results}}];
    }

    return $sigs;
}

#----------------------------------------------------------------------

sub whitehat_settings {
    my ($self, $sig_update_status, $sigs_before, $sigs_after, $whitehat_policies_without_response_splitting) = @_;
    my $dbh = $self->{dbh};

    return if F5::BigipUtils::is_docker(); #Used at build time when there are no policies, and SQL statements are not compatible with SQLite

    for my $type (keys %wh_third_party_attacks) {

        my @new_sigs;
	my $params = $self->get_params_by_attack_name($wh_third_party_attacks{$type});
        
        for my $sig (@{$sigs_after->{$type}}) {
            unless (grep $sig->{sig_id} eq $_->{sig_id}, @{$sigs_before->{$type}}) {
                push @new_sigs, $sig;
            }
        }

        for my $param_id (keys %$params) {

            for my $sig (@new_sigs) {
                $dbh->do(<<EndQuery, undef, $params->{$param_id}->{policy_id}, $param_id, $sig->{sig_id}, 1);
INSERT INTO PLC.PL_PARAM_NEGSIG_SIGNATURES
(policy_id, param_attrib_id, sig_id, flg_is_enabled)
VALUES
(?,?,?,?)
EndQuery
           }
        }
    }

    $dbh->do(<<EndQuery, undef, 1, 0, HTTP_RESPONSE_SPLITTING_ATTACK_TYPE_ID);
UPDATE PLC.PL_POLICY_NEGSIG_SIGNATURES AS pn,
       PLC.NEGSIG_SIGNATURES AS ns,
       PLC.PL_POLICIES as p,
       PLC.PL_SCANNER_CONFIG sc
   SET pn.flg_disabled = ?, pn.flg_quarantined = ?
 WHERE pn.sig_id = ns.sig_id
   AND ns.attack_type_id != ?
   AND p.id = pn.policy_id
   AND p.id = sc.policy_id
   AND sc.scanner_type != ''
   AND pn.flg_disabled = 1
EndQuery

    for my $whitehat_policy_id (@{$whitehat_policies_without_response_splitting}) {
       $dbh->do(<<EndQuery, undef, 1, 0, HTTP_RESPONSE_SPLITTING_ATTACK_TYPE_ID, $whitehat_policy_id);
UPDATE PLC.PL_POLICY_NEGSIG_SIGNATURES AS pn,
       PLC.NEGSIG_SIGNATURES AS ns
   SET pn.flg_disabled = ?, pn.flg_quarantined = ?
 WHERE pn.sig_id = ns.sig_id
   AND ns.attack_type_id = ?
   AND pn.policy_id = ?
   AND pn.flg_disabled = 1
EndQuery
    }
}

#----------------------------------------------------------------------

sub get_params_by_attack_name {
    my ($self, $attack_name) = @_;
    my $dbh = $self->{dbh};

    my $params = $dbh->selectall_hashref(<<EndQuery, 'id', undef, $attack_name);
SELECT params.id AS id, params.policy_id
  FROM PLC.PL_PARAM_ATTRIBUTES params,
       PLC.PL_SCANNER_VULNS vulns,
       PLC.NEGSIG_ATTACK_TYPES nat
 WHERE params.id = vulns.param_attrib_id
   AND nat.attack_type_id = vulns.attack_type_id
   AND nat.attack_type_name = ?
EndQuery

    return $params;
}

#----------------------------------------------------------------------

sub set_sigs_active {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};
  my $asm_config_handler = $params{asm_config_handler};
  my $policy_to_sigs = $params{policy_to_sigs};
  my $single_sig_id = $params{single_sig_id};
  my $affected_policies = $params{affected_policies} || [];
  my $skip_set_active = $params{skip_set_active};

  my $lock_file =  $F5::DbUtils::LOCK_FILE;
  my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                           logger => $logger);
  $locker->lock();

  my $active_policies = $dbh->selectall_hashref('SELECT account_id, id as active_policy_id FROM PLC.PL_POLICIES WHERE account_id > 0', 'account_id');

  my $negative_signatures = F5::NegativeSignatures->new(dbh => $dbh,
                                                        logger => $logger,
                                                        use_memory_cache => 1);

  my $prepareconf_policy = F5::PrepareConf::Policy->new(
                                                        dbh     =>  $dbh,
                                                        logger  =>  $logger,
                                                        unescape    => F5::Cfg::cfg_get_config_item("/etc/ts/dcc/".PRPLC_CFG_FILE, 'GENERAL', 'Unescape'),
                                                        prepare_negsig_exclusions => 0,
                                                        refresh_negsig_exclusions => $policy_to_sigs,
                                                        generate_negsig_collections => 0,
                                                        negative_signatures => $negative_signatures,
                                                       );
  my %affected_policy_hash = map { ($_ => 1) } @$affected_policies;
  for my $active_policy (values %$active_policies) {
      next if (%affected_policy_hash && !$affected_policy_hash{$active_policy->{active_policy_id}});
      if ($single_sig_id) {
          $prepareconf_policy->prepare(account_id => $active_policy->{account_id},
                                       policy_id => $active_policy->{active_policy_id},
                                       policy_parts_to_prepare => [ 'negsig_collections',
                                                                  ],
                                       extra_prepare_params => { single_sig_id => $single_sig_id },
                                      );
      }
      else {
          $prepareconf_policy->prepare(account_id => $active_policy->{account_id},
                                       policy_id => $active_policy->{active_policy_id},
                                       policy_parts_to_prepare => [ 'negsig_collections',
                                                                    'server_technologies_fingerprints',
                                                                    'threat_campaigns' ],
                                      );
      }
  }

  $prepareconf_policy->prepare_negsig_enforcer_names();
  my @modified_dcc_tables = $prepareconf_policy->get_modified_dcc_tables();

  if (@modified_dcc_tables) {
      my $collection = F5::NegativeSignatures::Collection->new(dbh => $dbh,
                                                               logger => $logger,
                                                               negative_signatures => $negative_signatures,
                                                              );
      $collection->generate();
      if (!$skip_set_active) {
          push(@modified_dcc_tables, $collection->get_modified_tables());

          $dbh->commit unless $dbh->{AutoCommit};

          my %additional_set_active_params;
          if ($single_sig_id) {
              $additional_set_active_params{modified_dcc_tables} = \@modified_dcc_tables;
          }

          F5::SetActive::set_active(dbh => $dbh, 
                                    logger => $logger, 
                                    do_update_core => 1,
                                    send_sync => 1,
                                    asm_config_handler => $asm_config_handler,
                                    %additional_set_active_params,
                                   );
      }
  }

  $locker->unlock();
}

#----------------------------------------------------------------------

sub check_service_agreement {
    my($self, %params) = @_;

    if ($self->{license_text}) {
        if (! F5::Provisioning::is_asm_on()) {
            $self->store_error_and_die("ASM is not enabled on the device");
        }
    }
    return $self->SUPER::check_service_agreement(%params);
}

#----------------------------------------------------------------------

sub check_filetype {
    my ($self, %params) = @_;
    return $params{filetype} =~ /$IM_FILE_OUTPUT|XML|JSON|data/; 
}

#----------------------------------------------------------------------

sub report_status {
  my($self, %params) = @_;
  if ($self->{check_interval}) {
      $self->{dbh}->do(<<EndQuery, undef, $params{completed_successfully}, $params{msg});
INSERT INTO PLC.SIGFILE_AUTO_UPDATE_HISTORY 
    (completed_successfully, msg) VALUES (?,?)
EndQuery
  }
}

#----------------------------------------------------------------------

sub get_autodownload {
    my ($self, %params) = @_;
    return F5::NegativeSignatures::SigFile::AutoDownload->new(dbh => $self->{dbh},
                                                              outbound_call_timeout => $self->{outbound_call_timeout},
                                                              logger => $self->{logger},
                                                             );

}

#----------------------------------------------------------------------

1;
