package F5::NegativeSignatures::Signature;

use strict;
use warnings;

use F5::Deploy;
use F5::NegativeSignatures;
use F5::NegativeSignatures::Rule;
use F5::BigipVersionUtils;
use F5::BigipUtils;
use F5::Utils::Rest;
use MIME::Base64;
use Digest::MD5;
use Time::Local;
use POSIX;
my $NEGSIG_CURRENT_VERSION_CODENAME_FILENAME = $ENV{ASM_NEGSIG_CURRENT_VERSION_CODENAME_FILENAME} || F5::Deploy::map_path('/ts/share/negsig/negsig_current_version_codename.txt');
our $UNLIKELY_DELIMITER = '###UNLIKELY_DELIMITER###';

our $AUTOLOAD;

require Exporter;
our @ISA = qw(Exporter);
our @export_constants = qw(
                            $SIG_UPDATE_STATUS_NOT_FOUND
                            $SIG_UPDATE_STATUS_UPDATE_MINOR
                            $SIG_UPDATE_STATUS_UPDATE
                            $SIG_UPDATE_STATUS_ADDED
                            $SIG_UPDATE_STATUS_DELETED
                            $SIG_UPDATE_STATUS_NO_CHANGE
                         );
our %EXPORT_TAGS = (constants => \@export_constants);
our @EXPORT_OK = (@export_constants);

our $SIG_UPDATE_STATUS_NOT_FOUND = 'not found';
our $SIG_UPDATE_STATUS_UPDATE_MINOR = 'update minor';
our $SIG_UPDATE_STATUS_UPDATE = 'update';
our $SIG_UPDATE_STATUS_ADDED = 'added';
our $SIG_UPDATE_STATUS_DELETED = 'deleted';
our $SIG_UPDATE_STATUS_NO_CHANGE = 'no change';

my @xml_fields = qw(
                     sig_name
                     sig_tag
                     rule
                     last_update
                     apply_to
                     risk
                     accuracy
                     flg_is_hidden
                  );


my @db_fields = (qw(
                     rev
                     flg_is_user_defined
                     sig_desc
                     flg_has_valuecontent_http
                     flg_has_valuecontent_xml
                     flg_has_valuecontent_json
                     flg_has_valuecontent_gwt
                     flg_has_headercontent
                     flg_has_uricontent
                     flg_has_requestcontent
                     flg_has_responsecontent
                     threshold_in_seconds
                     buffer_type_mask
                     flg_has_valuecontent_plain_text
                     sig_tag_policy
                  ),
                 @xml_fields);

my @all_fields = (qw(
                      sig_id
                      attack_type
                      sig_systems
                      rest_uuid
                   ),
                  @db_fields);

my %non_compare_fields = (
                          flg_has_valuecontent_http => 1,
                          flg_has_valuecontent_xml => 1,
                          flg_has_valuecontent_json => 1,
                          flg_has_valuecontent_gwt => 1,
                          flg_has_valuecontent_plain_text => 1,
                          flg_has_headercontent => 1,
                          flg_has_uricontent => 1,
                          flg_has_requestcontent => 1,
                          flg_has_responsecontent => 1,
                          threshold_in_seconds => 1,
                          sig_references => 1,
                          buffer_type_mask => 1,
                          last_update => 1,
                          rest_uuid => 1,
                          rev => 1,
                         );

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;

  return $self->init(%params);
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;
  $self->{dbh} = $params{dbh};
  $self->{dbname} = $params{dbname} || 'PLC';
  $self->{flg_is_user_defined} = $params{flg_is_user_defined} || 0;
  $self->{flg_is_hidden} = $params{flg_is_hidden} || 0;
  $self->{sig_tag} = $params{sig_tag} || '';
  $self->{sig_tag_policy} = $params{sig_tag_policy} || '';

  if ($params{sig_id} && $self->{dbh}) {
    return $self->retrieve(sig_id => $params{sig_id}, sig_tag => $self->{sig_tag}, sig_tag_policy => $self->{sig_tag_policy});
  }
  if ($params{sig_name} && $self->{dbh}) {
    return $self->retrieve(sig_name => $params{sig_name}, sig_tag => $self->{sig_tag}, sig_tag_policy => $self->{sig_tag_policy});
  }

  if ($params{object}) {
      return unless $self->from_object(sig => $params{object}, copy_all_fields => $params{copy_all_fields});
  }

  if ($params{xml_element}) {
      return unless $self->parse_xml($params{xml_element});
  }

  if ($params{sig_fields}) {
    # Convert attack_type and sig_systems to ids
    if ($params{sig_fields}->{attack_type} !~ /\D/) {
      $params{sig_fields}->{attack_type} = $self->get_attack_type_from_id( $params{sig_fields}->{attack_type} );
    }
    if (! grep(/\D/, @{$params{sig_fields}->{sig_systems}})) {
      $params{sig_fields}->{sig_systems} = $self->get_system_from_id( $params{sig_fields}->{sig_systems} );
    }

    for my $field (keys %{$params{sig_fields}}) {
        $self->$field( $params{sig_fields}->{$field} );
    }

    if ($params{sig_fields}->{keywords} && @{$params{sig_fields}->{keywords}}) {
        $self->{_orig_rule} = $self->{rule};
        $self->{rule} = F5::NegativeSignatures::Rule->from_simple_keywords(keywords => $params{sig_fields}->{keywords},
                                                                           apply_to => $params{sig_fields}->{apply_to},
                                                                          );
    }
  }

  return $self;
}

#----------------------------------------------------------------------

sub retrieve {
  my($self, %params) = @_;

  my $sig_tag_val = $params{sig_tag} ? $params{sig_tag} : '';
  my $sig_tag_policy_val = $params{sig_tag_policy} ? $params{sig_tag_policy} : '';

  my ($search_field, $search_val);
  if ($params{sig_id}) {
    $search_field = 'sig_id';
    $search_val = $params{sig_id};
  }
  if ($params{sig_name}) {
    $search_field = 'sig_name';
    $search_val = $params{sig_name};
  }

  local $" = ',';
  my $dbh = $self->{dbh};

  my $rec = $dbh->selectrow_hashref(<<EndQuery, undef, $search_val, $sig_tag_val, $sig_tag_policy_val);
SELECT @db_fields, sigs.rest_uuid as rest_uuid, sig_id, attack_type_name as attack_type
  FROM $self->{dbname}.NEGSIG_SIGNATURES sigs,
       $self->{dbname}.NEGSIG_ATTACK_TYPES attack_types
 WHERE $search_field = ?
   AND sigs.attack_type_id = attack_types.attack_type_id
   AND sigs.sig_tag = ?
   AND sigs.sig_tag_policy = ?
EndQuery

  return unless $rec;

  $self->sig_id( $rec->{sig_id} );
  $self->rest_uuid( $rec->{rest_uuid} );
  for my $field (@db_fields) {
    $self->$field( $rec->{$field} );
  }

  $self->attack_type( $rec->{attack_type} );

  my $sig_systems = $dbh->selectcol_arrayref(<<EndQuery, undef, $self->sig_id);
SELECT system_name
  FROM $self->{dbname}.NEGSIG_SYSTEMS sys,
       $self->{dbname}.NEGSIG_SIGNATURE_SYSTEMS sig_sys
 WHERE sig_sys.sig_id = ?
   AND sig_sys.system_id = sys.system_id
EndQuery

  $self->sig_systems( $sig_systems);

  $self->{_loaded_from_db} = 1;

  return $self;
}

#----------------------------------------------------------------------

sub update {
  my($self, %params) = @_;
  require F5::NegativeSignatures::Utils;
  require F5::ASMConfig::ASMConfig;
  local $" = ',';
  my $dbh = $self->{dbh};

  my %sig_references_type_map = (
                               'url' => F5::ASMConfig::SigReferenceType::SIGNATURE_REFERENCE_TYPE_URL(),
                               'cve' => F5::ASMConfig::SigReferenceType::SIGNATURE_REFERENCE_TYPE_CVE(),
                               'nessus' => F5::ASMConfig::SigReferenceType::SIGNATURE_REFERENCE_TYPE_NESSUS(),
                               'bugtraq' => F5::ASMConfig::SigReferenceType::SIGNATURE_REFERENCE_TYPE_BUGTRAQ(),
                             );

  # Set flg_has_valuecontent_* and sig_references just before update
  my $rule_obj = $self->rule_obj();
  $self->flg_has_valuecontent_http( $rule_obj->flg_has_valuecontent_http );
  $self->flg_has_valuecontent_xml( $rule_obj->flg_has_valuecontent_xml );
  $self->flg_has_valuecontent_json( $rule_obj->flg_has_valuecontent_json );
  $self->flg_has_valuecontent_gwt( $rule_obj->flg_has_valuecontent_gwt );
  $self->flg_has_valuecontent_plain_text( $rule_obj->flg_has_valuecontent_plain_text );
  $self->flg_has_headercontent( $rule_obj->flg_has_headercontent ||
                                $rule_obj->flg_has_requestcontent );
  $self->flg_has_uricontent( $rule_obj->flg_has_uricontent );
  $self->flg_has_requestcontent( $rule_obj->flg_has_requestcontent );
  $self->flg_has_responsecontent( $rule_obj->flg_has_responsecontent );
  $self->sig_references( $rule_obj->sig_references );
  $self->threshold_in_seconds( $rule_obj->threshold('seconds') || 0 );
  $self->buffer_type_mask( $rule_obj->buffer_type_mask );

  if ( (! $self->last_update) ||  ( $self->flg_is_user_defined && !$params{sigs_last_updated} ) ) {
    $self->last_update( strftime('%Y-%m-%d %H:%M:%S', localtime) );
  }

  if ($self->sig_id && F5::BigipUtils::is_feature_enabled('populate_rest_uuids')) {
      $self->rest_uuid(F5::Utils::Rest::generate_uuid(join($UNLIKELY_DELIMITER, ($self->sig_id, $self->sig_name))));
  }

  my $db_rec = {};
  
  for my $db_field (@db_fields, 'rest_uuid') {
    $db_rec->{$db_field} = $self->$db_field;
  }

  $db_rec->{packed_rule_obj} = $rule_obj->packed_rule_obj;

  $db_rec->{attack_type_id} = $self->attack_type_id || F5::NegativeSignatures::Utils::find_or_add_attack_type(dbh => $self->{dbh},
                                                                               dbname => $self->{dbname},
                                                                               attack_type_name => $self->attack_type);

  my $references = {};
  if ($self->sig_references) {
    for my $raw_sig_reference (@{$self->sig_references}) {
      my ($reference_type, $reference_val) = split(/,/, $raw_sig_reference, 2);
      push (@{$references->{$sig_references_type_map{$reference_type}}}, $reference_val);
    }
  }

  $db_rec->{flg_has_cve} = defined($references->{F5::ASMConfig::SigReferenceType::SIGNATURE_REFERENCE_TYPE_CVE()});

  my $is_sig_in_db = $self->sig_id && $dbh->selectcol_arrayref(<<EndQuery, undef, $self->sig_id)->[0];
SELECT rev
  FROM $self->{dbname}.NEGSIG_SIGNATURES
 WHERE sig_id = ?
EndQuery

  if ($is_sig_in_db) {
    my @update_fields = (@db_fields, 'attack_type_id', 'rest_uuid', 'packed_rule_obj');
    if ($self->flg_is_user_defined) {
        $db_rec->{rev} = $is_sig_in_db + 1;
    }

    my @update_assigns = map("$_ = ?", @update_fields);
    $dbh->do("UPDATE $self->{dbname}.NEGSIG_SIGNATURES SET @update_assigns WHERE sig_id = ?",
             undef,
             @{$db_rec}{@update_fields}, $self->sig_id);
  }
  else {
    if ($self->flg_is_user_defined) {
        $db_rec->{rev} ||= 1;

        # If imported sig_id is not currently in use, prefer that ID to the next auto-increment
        if ($self->{sig_id_from_xml}) {
            my($is_desired_sig_id_in_use) = $dbh->selectrow_array("SELECT COUNT(*) FROM $self->{dbname}.NEGSIG_SIGNATURES WHERE sig_id = ?",
                                                                  undef, $self->{sig_id_from_xml});

            if (! $is_desired_sig_id_in_use) {
                $self->sig_id( $self->{sig_id_from_xml});
                $self->rest_uuid(F5::Utils::Rest::generate_uuid(join($UNLIKELY_DELIMITER, ($self->sig_id, $self->sig_name))));
                $db_rec->{rest_uuid} = $self->rest_uuid;
            }
        }
    }

    if ($self->sig_id) {
      $db_rec->{sig_id} = $self->sig_id;
    }

    $db_rec->{last_enforcement_change} = time;
    my $rv = F5::DbUtils::insert_row(dbh => $dbh,
                                     table_name => "$self->{dbname}.NEGSIG_SIGNATURES",
                                     rec => $db_rec);
    unless ($rv) {
      die "Signature '". $self->sig_name . "' cannot be loaded (possible name conflict)\n";
    }

    if (!$self->sig_id) {
      my $new_sig_id = $dbh->last_insert_id(undef, undef, undef, undef);
      $self->sig_id( $new_sig_id );
      $self->rest_uuid(F5::Utils::Rest::generate_uuid(join($UNLIKELY_DELIMITER, ($self->sig_id, $self->sig_name))));
      F5::DbUtils::update_row(dbh => $dbh,
                              table_name => "$self->{dbname}.NEGSIG_SIGNATURES",
                              rec => { rest_uuid => $self->rest_uuid },
                              where_fields => { sig_id => $self->sig_id },
                             );
    }

    my $rest_uuid_for_policies = $self->rest_uuid(F5::Utils::Rest::generate_uuid($self->sig_id));
    my $existing_policies = $dbh->selectall_hashref(<<EndQuery, 'id');
SELECT id, flg_managed_by_bewaf
  FROM $self->{dbname}.PL_POLICIES
EndQuery

    # Add to all existing policies
    my $time = time();
 
    for my $policy_id (keys %$existing_policies) {
        my $flg_quarantine_signature = (!F5::BigipUtils::is_feature_enabled('allow_staging') || $existing_policies->{$policy_id}->{flg_managed_by_bewaf}) ? 0 : 1;
        F5::DbUtils::insert_row(dbh => $dbh,
                                table_name => "$self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES",
                                rec => { policy_id => $policy_id,
                                         sig_id => $self->sig_id,
                                         last_updated => $time,
                                         rest_uuid => $rest_uuid_for_policies,
                                         flg_quarantined => $flg_quarantine_signature },
                                ignore => 1,
                               );
    }
  }

  if ($is_sig_in_db) {
      # Check whether records exist before attempting to delete; otherwise, deleting zero rows may cause DB deadlock
      my($existing_row_in_db) = $dbh->selectrow_array("SELECT distinct sig_id FROM $self->{dbname}.NEGSIG_SIGNATURE_SYSTEMS WHERE sig_id = ?", undef, $self->sig_id);

      if ($existing_row_in_db) {
          $dbh->do("DELETE FROM $self->{dbname}.NEGSIG_SIGNATURE_SYSTEMS WHERE sig_id = ?", undef, $self->sig_id);
      }

      ($existing_row_in_db) = $dbh->selectrow_array("SELECT distinct sig_id FROM $self->{dbname}.NEGSIG_SIGNATURE_REFERENCES WHERE sig_id = ?", undef, $self->sig_id);

      if ($existing_row_in_db) {
          $dbh->do("DELETE FROM $self->{dbname}.NEGSIG_SIGNATURE_REFERENCES WHERE sig_id = ?", undef, $self->sig_id);
      }
  }

  if ($self->sig_systems) {
    for my $sig_system_name (@{$self->sig_systems}) {
      my $sig_system_id = F5::NegativeSignatures::Utils::find_or_add_system(dbh => $self->{dbh},
                                             dbname => $self->{dbname},
                                             system_name => $sig_system_name);
      next unless $sig_system_id;

      $dbh->do("INSERT INTO $self->{dbname}.NEGSIG_SIGNATURE_SYSTEMS (sig_id, system_id) VALUES (?,?)",
               undef, $self->sig_id, $sig_system_id);
    }
  }

  for my $reference_type (keys %$references) {
      for my $reference_val (@{$references->{$reference_type}}) {
          $dbh->do("INSERT INTO $self->{dbname}.NEGSIG_SIGNATURE_REFERENCES (sig_id, reference_type, reference_val) VALUES (?,?,?)",
                   undef, $self->sig_id, $reference_type, $reference_val);
      }
  }

  if ($self->flg_is_user_defined) {
      #all factory signatures will be linked to sets after they are installed from a binary sigfile
      F5::NegativeSignatures->new($self->{dbh})->link_sets_to_sig($self->sig_id,
                                                                  store_type => ($is_sig_in_db ? 'update' : 'add'),
                                                                 );
  }
}

#----------------------------------------------------------------------

sub delete {
  my($self, %params) = @_;

  my @affected_policies;
  if ($params{find_affected_policies}) {
      @affected_policies = F5::NegativeSignatures->new($self->{dbh})->get_policies_using_signatures($self->sig_id);
  }
  $self->{dbh}->do("DELETE FROM $self->{dbname}.NEGSIG_SIGNATURES WHERE sig_id = ?", undef, $self->sig_id);

  return \@affected_policies;
}

#----------------------------------------------------------------------

sub parse_xml {
  my($self, $elm) = @_;

  my $sig_all_revs = $elm->simplify(forcearray => [ 'system_name', 'rev', 'reference' ]);

  my $sig;
  if ($self->flg_is_user_defined) {
    $sig = $sig_all_revs->{rev}->[0];
  }
  else {
      if (exists $sig_all_revs->{rev}->[0]->{bigip_version_codename} || $sig_all_revs->{rev}->[0]->{bigip_version_min} !~ /^\d/) {
          my $codename_field = 'bigip_version_codename';
          if (! exists $sig_all_revs->{rev}->[0]->{bigip_version_codename}) {
              $codename_field = 'bigip_version_min';
          }
          $sig = F5::BigipVersionUtils::find_by_codename(elements => $sig_all_revs->{rev},
                                                         current_version_codename_filename => $NEGSIG_CURRENT_VERSION_CODENAME_FILENAME,
                                                         codename_extract => sub {
                                                             return $_[0]->{$codename_field};
                                                         },
                                                        );
    }
    else {
      $sig = F5::BigipVersionUtils::find_by_version(elements => $sig_all_revs->{rev},
                                                    version_extract => sub {
                                                      return $_[0]->{bigip_version_min};
                                                    },
                                                   );
    }

    unless ($sig) {
      ($sig) = grep { ! $_->{bigip_version_min} } @{$sig_all_revs->{rev}};
    }

    return unless $sig;
    return if ($sig->{deleted});
  }

  #remove empty hashes entered by XML::Twig
  if (ref($sig) eq 'HASH') {
    for my $entry (keys %$sig) {
      if (ref($sig->{$entry}) eq 'HASH' && scalar keys %{$sig->{$entry}} == 0) {
        $sig->{$entry} = undef;
      }
    }
  }

  if (!$self->flg_is_user_defined) {
      $self->sig_id( $elm->id );
  }
  $self->{sig_id_from_xml} = $elm->id;

  $self->rev( $sig->{num} || 1 );
  $self->sig_desc( $sig->{doc} );
  $self->attack_type( $sig->{attack_type} );

  if (! $sig->{systems} || ! $sig->{systems}->{system_name}) {
      $sig->{systems}->{system_name} = [];
  }
  $self->sig_systems( $sig->{systems}->{system_name} );

  return $self->from_object(sig => $sig);
}

#----------------------------------------------------------------------

my $trim_regex = qr/^\s*(.*?)\s*$/;


sub from_object {

  my($self, %params) = @_;
  my $sig = $params{sig};
  my $copy_all_fields = $params{copy_all_fields};

  if (exists $sig->{references} && exists $sig->{references}->{reference}) {
    for my $reference (@{$sig->{references}->{reference}}) {
      if (defined $reference->{ref_id} && ! ref($reference->{ref_id}) && $reference->{ref_id} =~ /\S/) {
        $sig->{rule} .= " reference:$reference->{ref_type},$reference->{ref_id};";
      }
    }
  }

  if ($sig->{last_update}) {
    $sig->{last_update} =~ s|\$DateTime: (\d+)/(\d+)/(\d+) ([\d:]+) \$|$1-$2-$3 $4|;
    # Validate date
    my @date_parts;
    unless (@date_parts = ($sig->{last_update} =~ /^(\d{4})[-\/](\d\d)[-\/](\d\d) (\d\d):(\d\d):(\d\d)$/)) {
      die "Invalid date: $sig->{last_update}\n";
    }
    eval {
      $date_parts[0] -= 1900; # year
      $date_parts[1]--; # month
      timelocal(reverse @date_parts); # dies in case of bad date
    };
    if ($@) {
      my $err_msg = $@;
      $err_msg =~ s/ at \S+ line \d+//;
      die "Invalid date: $sig->{last_update} - $err_msg\n";
    }
  }

  $sig->{sig_name} =~ s/$trim_regex/$1/;

  if ( $sig->{sig_systems} ) {
      $self->sig_systems(delete($sig->{sig_systems}));
  }
  else {
      if (! $sig->{systems} || ! $sig->{systems}->{system_name}) {
          $sig->{systems}->{system_name} = [];
      }
      $self->sig_systems( $sig->{systems}->{system_name} );
  }

  if ($sig->{attack_type_name}) {
      $self->{attack_type} = $sig->{attack_type_name};
  }
  elsif ($sig->{attack_type_id} ) {
       $self->{attack_type} = $self->get_attack_type_from_id( $sig->{attack_type_id} );
  }

  my @copy_fields = $copy_all_fields ? @all_fields : @xml_fields;
  for my $copy_field (@copy_fields) {
    if (defined $sig->{$copy_field}) {
        utf8::decode($sig->{$copy_field});
        $self->$copy_field( $sig->{$copy_field} );
    }
  }

  return 1;
}

#----------------------------------------------------------------------

sub differs_significantly_from {
  my($self, $other_sig) = @_;

  my @differences = grep($_->{diff_field} eq 'rule' || $_->{diff_field} eq 'apply_to',
                         $self->compare($other_sig));
  if (scalar @differences == 1 && $differences[0]->{diff_field} eq 'rule') {
    my $rule_without_ref = $self->rule;
    $rule_without_ref =~ s/reference:.*?;//g;
    $rule_without_ref =~ s/\s*$//;

    my $other_sig_rule_without_ref = $other_sig->rule;
    $other_sig_rule_without_ref =~ s/reference:.*?;//g;
    $other_sig_rule_without_ref =~ s/\s*$//;

    if ($rule_without_ref eq $other_sig_rule_without_ref) {
      # Only difference is reference within rule
      return;
    }
  }

  if (@differences) {
    return 1;
  }
  else {
    return;
  }
}

#----------------------------------------------------------------------

sub differs_from {
  my($self, $other_sig) = @_;

  if ($self->compare($other_sig)) {
    return 1;
  }
  return;
}

#----------------------------------------------------------------------

sub compare {
  my($self, $other_sig) = @_;

  my @differences;
  for my $field (@all_fields) {
    next if (exists $non_compare_fields{$field});

    if ($field eq 'sig_systems') {
      # No sig system would contain ///, so just join for the comparison
      if ($self->sig_systems && $other_sig->sig_systems &&
          join('///', sort @{$self->sig_systems}) ne join('///', sort @{$other_sig->sig_systems})) {
        push(@differences,
             { diff_type => 'modify',
               diff_field => $field,
               old_val => $self->sig_systems,
               new_val => $other_sig->sig_systems,
             });
      }
    }
    else {
      if ((!defined $self->$field &&  defined $other_sig->$field) ||
          (defined $self->$field && !defined $other_sig->$field) ||
          (defined $self->$field && defined $other_sig->$field &&
           $self->$field ne $other_sig->$field)) {
        push(@differences,
             { diff_type => 'modify',
               diff_field => $field,
               old_val => $self->$field,
               new_val => $other_sig->$field
             });
      }
    }
  }

  return @differences;
}

#----------------------------------------------------------------------

sub reset_quarantine_status {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};

  F5::NegativeSignatures->new($dbh)->set_flg_quarantined(policy_id => '*',
                                                         sig_id => $self->sig_id,
                                                         flg_quarantined => 1);

  $dbh->do(<<EndQuery, undef, time(), $self->sig_id);
UPDATE $self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES  
   SET last_updated = ? 
   WHERE sig_id = ?
EndQuery

}

#----------------------------------------------------------------------

sub update_last_enforcement_change {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};

  $dbh->do("UPDATE $self->{dbname}.NEGSIG_SIGNATURES SET last_enforcement_change = ? WHERE sig_id = ?",
           undef, time, $self->sig_id);

}

#----------------------------------------------------------------------

sub rule_obj {
  my($self) = @_;

  unless ($self->{rule_obj}) {
    eval {
      $self->{rule_obj} = F5::NegativeSignatures::Rule->new(rule_string => $self->rule,
                                                            apply_to => $self->apply_to,
                                                           );
    };

    if ($@) {
      die "Signature '" . $self->sig_name . "' is invalid: $@\n";
    }
  }

  return $self->{rule_obj};
}

#----------------------------------------------------------------------

sub get_attack_type_from_id {
  my($self, $attack_type_id) = @_;

  return $self->{dbh}->selectcol_arrayref(<<EndQuery, undef, $attack_type_id)->[0];
SELECT attack_type_name
  FROM $self->{dbname}.NEGSIG_ATTACK_TYPES 
 WHERE attack_type_id = ?
EndQuery
}

#----------------------------------------------------------------------

sub get_system_from_id {
  my($self, $system_ids) = @_;

  my @system_names;
  for my $system_id (@$system_ids) {
    push(@system_names, $self->{dbh}->selectcol_arrayref(<<EndQuery, undef, $system_id)->[0]);
SELECT system_name
  FROM $self->{dbname}.NEGSIG_SYSTEMS 
 WHERE system_id = ?
EndQuery
  }

  return \@system_names;
}

#----------------------------------------------------------------------

sub clean_irrelevant_data{
   my($self, %params) = @_;

   return if $self->{flg_has_valuecontent_http};
   return if F5::BigipUtils::is_docker(); #Used at build time when there are no policies, and SQL statements are not compatible with SQLite

   $self->{dbh}->do(<<EndQuery,undef,$self->{sig_id});
DELETE FROM  $self->{dbname}.PL_PARAM_NEGSIG_SIGNATURES 
 WHERE sig_id = ?
EndQuery

   $self->{dbh}->do(<<EndQuery,undef,$self->{sig_id});
DELETE FROM  $self->{dbname}.PL_COOKIE_NEGSIG_SIGNATURES
 WHERE sig_id = ?
EndQuery

   for my $content_type ( qw(xml json gwt graphql grpc) ){
       if ( !$self->{"flg_has_valuecontent_$content_type"} ) {
   $self->{dbh}->do(<<EndQuery,undef,$self->{sig_id},$content_type);
DELETE A
  FROM PLC.PL_CONTENT_PROFILE_NEGSIG_SIGNATURES A
   LEFT JOIN PLC.PL_CONTENT_PROFILES B ON (  A.profile_id=B.id )
   WHERE A.sig_id = ? 
  AND B.content_profile_type = ?
EndQuery
        }
  }

}
#======================================================================
# Accessors

sub AUTOLOAD {
  my($self, $set_val) = @_;

  my $name = $AUTOLOAD;
  $name =~ s/.*://;   # strip fully-qualified portion

  if (grep($name, @all_fields)) {
    if (defined $set_val) {
      $self->{$name} = $set_val;
    }
    return $self->{$name};
  }

  return; # undef if no such value found
}

#----------------------------------------------------------------------

sub save_previous_enforced_rule {
    my($self, %params) = @_;
    my $sig_id = $params{sig_id};
    my $rule = $params{rule};
    my @policy_ids = @{$params{policy_ids}};
    my $policy_to_sigs = $params{policy_to_sigs};

    my $md5_calc = Digest::MD5->new;
    $md5_calc->add($rule);
    my $rule_md5 = $md5_calc->hexdigest;
	
    my $dbh = $self->{dbh};
    my $policies_with_retain_previous = $dbh->selectall_hashref(<<EndQuery, 'policy_id', undef, 1);
SELECT policy_id, place_sigs_in_staging
  FROM $self->{dbname}.PL_POLICY_NEGSIG_ATTRIBUTES
  WHERE place_sigs_in_staging = ?
EndQuery
                            
    for my $policy_id (@policy_ids) {
        if (exists $policies_with_retain_previous->{$policy_id}) {	   	  
            my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                             table_name => "$self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES",
                                             rec => { previous_enforced_rule => $rule,
                                                      previous_enforced_rule_md5 => $rule_md5 },
                                             where_fields => { policy_id => $policy_id, 
                                                               sig_id => $sig_id, 
                                                               flg_quarantined => 0,
                                                               flg_disabled => 0 });
            next unless $rv;
            $policy_to_sigs->{$policy_id}->{$sig_id} = 1;
        }
    }
}

#======================================================================

sub reset_previous_enforced_rules_on_rollback {
    my($self, %params) = @_;
    my @policy_ids = @{$params{policy_ids}};

    my $md5_calc = Digest::MD5->new;
    $md5_calc->add($self->{rule});
    my $rule_md5 = $md5_calc->hexdigest;  

    my $dbh = $self->{dbh};
    my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                     table_name => "$self->{dbname}.PL_POLICY_NEGSIG_SIGNATURES",
                                     rec => { previous_enforced_rule => '',
                                              previous_enforced_rule_md5 => '',
                                 	      flg_quarantined => 0,
                                            },
                                     where_fields => { sig_id => $self->{sig_id},
                                     		       policy_id => \@policy_ids,
                                	               previous_enforced_rule_md5 => $rule_md5 });
}

#======================================================================
# Class method

sub update_if_changed {
    my($class, %params) = @_;

    my $policy_to_sigs = $params{policy_to_sigs};
    my $find_affected_policies = $params{find_affected_policies};

    my $sig = F5::NegativeSignatures::Signature->new(%params);

    return unless $sig;

    my $allow_staging = F5::BigipUtils::is_feature_enabled('allow_staging');
    my $all_affected_policies;
    my $update_status;
    my %lookup_key;
    if ($params{lookup_by} eq 'sig_name') {
        $lookup_key{sig_name} = $sig->sig_name;
        $lookup_key{sig_tag} = $sig->{sig_tag} || '';
    }
    else {
        $lookup_key{sig_id} = $sig->sig_id;
    }

    my $negsig;
    if ($find_affected_policies) {
        $negsig = F5::NegativeSignatures->new(dbh => $params{dbh},
                                              dbname => $params{dbname} || 'PLC',
                                            );
    }

    my $existing_sig;
    if (!$params{skip_lookup} && !($params{lookup_by} eq 'sig_id' && !$sig->sig_id)) {
        $existing_sig = F5::NegativeSignatures::Signature->new(dbh => $params{dbh},
                                                               dbname => $params{dbname} || 'PLC',
                                                               %lookup_key);
    }

    if (!$existing_sig) {
        # New signature -- add it
        $sig->update(sigs_last_updated => $params{sigs_last_updated});
        if ($find_affected_policies) {
            $all_affected_policies = [ $negsig->get_policies_using_signatures($sig->sig_id)];
        }
        $update_status = $SIG_UPDATE_STATUS_ADDED;
    }
    else {    	    	    	
        if (! $existing_sig->flg_is_user_defined && $sig->flg_is_user_defined) {
            die "Factory signature cannot be overwritten with a user-defined signature (" . $sig->sig_name . ")\n";
        }

        # If the updated signature included both keywords and rule in the update call, and keywords are identical to the
        # previously saved signature (existing), and the update call's rule differs, then the actual rule should be used
        # (ignoring the keywords).
        # Note that $sig->{rule} has already been converted from keywords on instantiation, and is used here because string
        # comparison is easier than comparing the keywords lists of hashes.
        if ($sig->{_orig_rule} && $sig->{rule} eq $existing_sig->{rule} && $sig->{_orig_rule} ne $sig->{rule}) {
            $sig->{rule} = $sig->{_orig_rule};
        }

        # Copy sig_id from existing_sig for user-defined
        $sig->sig_id( $existing_sig->sig_id ) if ($sig->flg_is_user_defined);

        if ($sig->differs_from($existing_sig)) {
            my @policy_ids;
            if ($allow_staging && $find_affected_policies) {
                @policy_ids = $negsig->get_policies_using_signatures($existing_sig->sig_id);
            }
            my %policies_before_update = map(($_ => 1), @policy_ids);

            my $differed_significantly =  $sig->differs_significantly_from($existing_sig);
                                                                                 
            if($allow_staging && $differed_significantly && !$sig->{flg_is_user_defined}) {
               $sig->save_previous_enforced_rule(sig_id => $existing_sig->sig_id, 
                                                 rule=> $existing_sig->rule, 
                                                 policy_ids => \@policy_ids,
						 policy_to_sigs => $policy_to_sigs,);
            }
                                    
            $sig->update(sigs_last_updated => $params{sigs_last_updated});
            if ($allow_staging && $find_affected_policies) {
                my %policies_after_update = map(($_ => 1), $negsig->get_policies_using_signatures($sig->sig_id));
                my %all_policies = (%policies_before_update, %policies_after_update);
                $all_affected_policies = [keys %all_policies];
            }

            if ($differed_significantly) {
                if ($allow_staging) {
                    $sig->reset_quarantine_status;
                    $sig->reset_previous_enforced_rules_on_rollback(policy_ids => \@policy_ids);
                }
                $sig->clean_irrelevant_data;
                $sig->update_last_enforcement_change;                	                
                $update_status = $SIG_UPDATE_STATUS_UPDATE;
            }
            else {
                $update_status = $SIG_UPDATE_STATUS_UPDATE_MINOR;
            }
        }
        else {
            $update_status = $SIG_UPDATE_STATUS_NO_CHANGE;
        }
    }
  
    return { sig_id => $sig->sig_id,
             sig_name => $sig->sig_name,
             sig_tag => $sig->sig_tag,
             rev => $sig->rev,
             update_status => $update_status,
             affected_policies => $all_affected_policies };
}
