package F5::NwdUtils;

use strict;
use List::Util qw/first/;
use F5::SystemUtils;
use F5::Cfg;

use constant  FALSE  =>  0;
use constant  TRUE   =>  1;

my $MINIMUM_BD_MEMORY = 750 * 1024; # 750 MB
my $ASM_HOST_NON_BD_MEMORY = 1024 * 1024; # 1 GB

my $ASM_HOST_START_CONTRIBUTE_TO_AVR = 10 * 1024 * 1024; # 10  GB
my $ASM_HOST_MAX_CONTRIBUTE_TO_AVR =         512 * 1024; # 512 MB

sub is_wa_enabled {
    open (my $in, "<", "/config/bigip.conf") or return undef;
    my @lines = <$in>;
    close $in;

    my ($wa_enabled) = grep /wa enable/, @lines;
    return 1 if $wa_enabled;
    return 0;
}

sub get_relevant_cpu_count {
    my (%arguments) = @_;
    my $consider_wa = (defined $arguments{'consider_wa'}) ? 1 : 0;

    my $number_of_cpus = _get_number_of_cpus();   

    if(is_vadc(logger => $arguments{logger})) {
      $number_of_cpus = 2;
   }
   elsif(_get_number_of_tmms() eq '1') {
        $number_of_cpus--;
        if ($consider_wa && is_wa_enabled()) {
            $number_of_cpus--;
        }
        if($number_of_cpus < 1){
            $number_of_cpus = 1;
        }
    }

    return $number_of_cpus;
}

sub _get_number_of_cpus {
    open (CMD, "-|", ". /usr/lib/bigstart/functions; get_number_cpu") or return ;
    my $number_of_cpus = <CMD>;
    chomp $number_of_cpus;
    close CMD;
    return $number_of_cpus;
}

sub _get_number_of_tmms {
    open (CMD, "-|", ". /usr/lib/bigstart/functions; get_tmm_count") or return 1;
    my $number_of_tmms = <CMD>;
    chomp $number_of_tmms;
    close CMD;
    return $number_of_tmms;
}

#----------------------------------------------------------------------
# IMPORTANT : 
#    the same calculation is done in 'tm_daemon/mcp/asm/validation/AsmPolicy.cpp' at 'AsmPolicyValidator::run'
#    the validation logic has to be the same in both places
sub get_relevant_bd_memory {
    my(%params) = @_;
    my $logger = $params{logger};

    my $mcp_var = 'provision.memory.asm.host';

    # this value was empirically derived from a 4GB VADC (Z100) platform with ltm/asm provisioned to nominal
    # this is the scenario that generates the smallest possible value for asm host memory provisioning
    # we take this value as a default in case we fail to extract a value from 'provision.memory.asm.host' (not 'provisioned.memory.asm.host')
    my $total_asm_host_memory = 850 * 1024;

    # Load dynamically to prevent F5::COAPI usage at build time
    eval "use F5::Cluster::Utils;";
    if (! $@) {
        eval {
            $total_asm_host_memory = F5::Cluster::Utils::get_mcp_val($mcp_var) || 0;
        };
        if ($@ && $logger) {
            my $perl_err = $@; chomp($perl_err);
            my $coapi_err = ($F5::COAPI::error) ? $F5::COAPI::error : 'F5::COAPI::error was not set' ;
            $logger->error('Error calling:[F5::Cluster::Utils::get_mcp_val('.$mcp_var.')]');
            $logger->error("Perl error:[$perl_err].");
            $logger->error("COAPI error:[$coapi_err].");
        }
        $total_asm_host_memory *= 1024; # convert MB to kB
    } else {
        $logger->error("Could not retrieve a value for '$mcp_var' from mcp. Failed to load 'F5::Cluster::Utils'.") if $logger;
    }

    my $platform_specific_additional_memory = 0;
    my $platform = F5::Cfg::cfg_get_config_item('/PLATFORM', undef, 'platform') || '';
    if (lc $platform eq 'c112') {
        # Use an additional 400MB on stratos lo platform specifically
        $platform_specific_additional_memory = 400 * 1024;
    }

    # AVRD takes additional memory when ASM gets above 10G.
    # The formula is that AVR gets 10% of the memory that is above 10G, but no more than 512 MB.
    my $contribute_to_avrd = 0;
    if ($total_asm_host_memory > $ASM_HOST_START_CONTRIBUTE_TO_AVR) {
        $contribute_to_avrd = List::Util::min( $ASM_HOST_MAX_CONTRIBUTE_TO_AVR,
                                               int(($total_asm_host_memory - $ASM_HOST_START_CONTRIBUTE_TO_AVR) / 10) );

        $logger->info("Additional memory given to avrd: $contribute_to_avrd") if $logger;
    }

    my $bd_memory = List::Util::max( $MINIMUM_BD_MEMORY,
                                  $total_asm_host_memory - $ASM_HOST_NON_BD_MEMORY + $platform_specific_additional_memory - $contribute_to_avrd);

    $logger->info("bd will start with ${bd_memory}k allocated (total asm host provisioning = ${total_asm_host_memory}k)") if $logger;
    return $bd_memory;
}

#----------------------------------------------------------------------

sub get_proc_cpuinfo_cpu_mhz {
    my (%arguments) = @_;

    my $cpu_info_path = (defined $arguments{'cpu_info_path'}) ? $arguments{'cpu_info_path'} : '/proc/cpuinfo';

    open my $fh, "<", $cpu_info_path or return undef;

    my $cpu_mhz = 0;
    while (<$fh>) {
        chomp;
        if(/^\s*cpu MHz\s*:\s(\d+\.\d+)/)
        {
	    $cpu_mhz = $1;	
            last; 
        }

    }
    close $fh;
    return $cpu_mhz * 1000;

}


#--------------------------------------------------------------------------------
# match all patterns as WHOLE WORDS in some line in "cmd" output
# receive - (cmd, pattern1, pattern2,..., patternN)
# return TRUE - match all patterns in one line
#        FALSE - didn't match!
sub cmd_out_grep_whole
{
    my $cmd = shift;
    my @patterns = @_;

    if(! open(CMD, "$cmd|"))
    {
	print "Failed to Exec \"$cmd\" $!";
	return FALSE;
    }
    
    while(<CMD>)
    {
	my $found = 0;
	foreach my $ptrn (@patterns)
	{
	    $found++ if((/\s+$ptrn\s+/i) or (/^$ptrn\s+/i));
	}
	if($found == scalar(@patterns))
	{
	    close(CMD);
	    return TRUE;
	}
    }

    close(CMD);
    return FALSE;
}

#----------------------------------------------------------------------

my %process_stat_fields = (
                           0 => 'pid',
                           1 => 'process_name',
                           22 => 'vmsize',
                           23 => 'rss_pages',
                           13 => 'cpu_user_time',
                           14 => 'cpu_system_time',
                           19 => 'thread_count', # number of threads
                          );

my $BYTES_PER_PAGE = 4 * 1024;

sub get_process_stat {
    my($pid) = @_;
    if (! $pid) {
        $pid = $$;
    }

    open(my $stat_fh, '<', "/proc/$pid/stat") or return {};
    my $stat_line = <$stat_fh>;
    close $stat_fh;

    chomp($stat_line);
    my @stat_fields = split(/ /, $stat_line);
    my %retval = map { ($process_stat_fields{$_} => $stat_fields[$_]) } keys %process_stat_fields;

    # Cleanup process name
    $retval{process_name} =~ s/^\((.*)\)$/$1/;
    # Add RSS size
    $retval{rss} = $retval{rss_pages} * $BYTES_PER_PAGE;

    return \%retval;
}

#----------------------------------------------------------------------
sub is_vadc
{
    my(%params) = @_;
    my $logger = $params{logger};

    my $rv = F5::SystemUtils::run_system_tool(command_line_params => [ "qp" ] );
    if($rv->{completed_successfully}) {
        chomp($rv->{standard_output});
        return ($rv->{standard_output} eq "Z100") ? 1 : 0;

    } else {
        $logger->error("failed to run qp to determine if platform is VADC - bd RWThreads will be set based on CPU count") if $logger;
        return 0;
    }
}

#----------------------------------------------------------------------
1;
