package F5::NwdUtils::Nwd;

use strict;
use Cwd;
use File::Basename;

# tsdev includes
use F5::Provisioning;
use F5::DbUtils;
use F5::NwdUtils;
use F5::NwdUtils::UpdateGTMScore;
use File::Copy;
use F5::Cluster::Utils;
use F5::Cfg;
use F5::Logger qw(:LEGACY_ERROR_CONSTANTS);
use F5::ProcessHandler;
use F5::Db::Single;

use constant  WD_CFG_FILE     =>  '/etc/ts/tools/nwd.cfg';

use constant  FALSE       => 0;
use constant  TRUE        => 1;

#defaults
use constant  DEFAULT_LOG_FILE  => 'nwd.log';

use constant  FAILURE_PROC_RESTARTS_EXCEEDED => 'Consecutive process restarts reached a threshold';
use constant  FAILURE_NOT_ENOUGH_THREADS   => 'Insufficient number of threads';
use constant  FAILURE_NOT_ENOUGH_PROCS   => 'Insufficient number of processes';
use constant  FAILURE_THREAD_NUM_DECREASED => 'Number of threads decreased';
use constant  FAILURE_CPU_USAGE_FAILURE    => 'CPU usage reached a threshold';
use constant  FAILURE_MEMORY_SIZE_USAGE_FAILURE    => 'Memory Size reached a threshold';

############################################################################
#  Create new object
############################################################################
sub new{
    my ($class, %params) = @_;
    my $objref = {};
    bless $objref, $class;
    $objref->init(%params);
    return $objref;
}



############################################################################
#  Returns the time interval for running
############################################################################
sub get_exe_time{

    my ($self) = @_;
    return $self->{LOOP_RESOLUTION};

}

############################################################################
#  sub: verify
############################################################################
sub verify
{
    my ($self, %params) = @_;

    # chdir to work directory
    chdir(dirname($0));

    my %h_failure_info;
    my $verify_rv;
    my $single_proc_to_verify = $params{proc};

    # ps() only calculates the CPU on an interval, so only update the $h_cpu if ps() updates the info
    my $ps = $self->ps();
    $self->{h_cpu} = $ps if defined $ps;

    my @procs_to_verify = ();
    if ($single_proc_to_verify) {
        push @procs_to_verify , $single_proc_to_verify ;
    } else {
        @procs_to_verify = keys %{$self->{h_procs}};
    }

    foreach my $proc (@procs_to_verify)
    {
        $self->{next_watch_time} = $self->{h_procs}->{$proc}{'NextWatchTime'};
        my $now = time();

        if ( !$single_proc_to_verify && $self->{next_watch_time} > $now ) {
            next;
        }

        $self->{next_watch_time} = $now + $self->{FREQUENCY};
        $self->{h_procs}->{$proc}{'NextWatchTime'} = $self->{next_watch_time};
        $self->{logger}->trace("$proc: performing watch now ($now), next watch: $self->{next_watch_time}") if $self->{logger};

        #$self->{actual_threads} = keys %{$self->{h_cpu}->{$proc}};
        $self->{actual_threads} = 0;
        $self->{actual_procs} = 0;
        $self->{memory_size} = 0;
        my $proc_found_in_list = 0;
        for my $pid (keys %{$self->{h_cpu}->{$proc}}) {
            $proc_found_in_list = 1;
            $self->{actual_threads}+= $self->{h_cpu}->{$proc}->{$pid}->{number_of_threads};                 
            $self->{h_pids}->{$proc} = $pid;
            $self->{actual_procs}++;
            $self->{memory_size} = $self->{memory_size} > $self->{h_cpu}->{$proc}->{$pid}->{memory_size} ? $self->{memory_size}
                                                                                 : $self->{h_cpu}->{$proc}->{$pid}->{memory_size};
        }
        $self->{max_memory_size} = $self->{h_procs}->{$proc}{'MaxMemorySize'};
        $self->{required_procs} = $self->{h_procs}->{$proc}{'MinimalNumProcs'};
        $self->{required_threads} = $self->{h_procs}->{$proc}{'MinimalNumThreads'};
        $self->{max_expected_threads} = $self->{h_procs}->{$proc}{'MaxExpectedThreads'};
        $self->wd_logprint($DBG, undef,"$proc: $self->{actual_threads} running threads");


        if ($self->{max_memory_size}) {
           $self->wd_logprint($DBG, undef,"$proc: $self->{memory_size} RSS Memory");
           if ($self->{memory_size} > $self->{max_memory_size}) {
               $self->wd_logprint($INF, undef, "$proc: RSS Memory ($self->{memory_size}) exceeded MaxMemorySize (".$self->{h_procs}->{$proc}->{'MaxMemorySize'}.")!!");

                 %h_failure_info = ('failure' =>  FAILURE_MEMORY_SIZE_USAGE_FAILURE,
                                    'max_memory_size' => $self->{max_memory_size},
                                    'memory_size' => $self->{memory_size},
                                    'proc_name'       => $proc);
                 $self->log_failure(\%h_failure_info);

                 $self->run_cmds($self->{h_procs}->{$proc}{'OnProcFailCommands'});
                 # restart watch
                 $self->proc_failure_occurs(\%h_failure_info);
                 return;
            }
        }

        # check minimal number of threads:
        if($self->{actual_threads} < $self->{required_threads})
        {
            last if $single_proc_to_verify;

            $self->{h_times_threads_exceed}->{$proc} = 0 unless defined $self->{h_times_threads_exceed}->{$proc};
            $self->{h_times_threads_exceed}->{$proc}++;

            $self->wd_logprint($INF, undef, "$proc: not enough threads (found: $self->{actual_threads}, required: $self->{required_threads}, proc_found_in_list=$proc_found_in_list)!");
            # if passed TimesThreadsFail times - exec the OnProcFailCommands
            if($self->{h_times_threads_exceed}->{$proc} >= $self->{h_procs}->{$proc}->{'TimesThreadsFail'})
            {
                $self->wd_logprint($INF, undef, "$proc: not enough threads, exceeded TimesThreadsFail (".$self->{h_procs}->{$proc}->{'TimesThreadsFail'}.")!!");

                 %h_failure_info = ('failure' =>  FAILURE_NOT_ENOUGH_THREADS,
                                    'required_threads' => $self->{required_threads},
                                     'present_threads' => $self->{actual_threads},
                                     'proc_name'       => $proc);
                 $self->log_failure(\%h_failure_info);

                 $self->{h_times_recent_proc_failed}->{$proc} = 0 unless defined $self->{h_times_recent_proc_failed}->{$proc};
                 $self->{h_times_recent_proc_failed}->{$proc}++;

                 if($self->{h_procs}->{$proc}->{'TimesProcRestart'} &&
                    $self->{h_times_recent_proc_failed}->{$proc} >= $self->{h_procs}->{$proc}->{'TimesProcRestart'})
                 {
                     $self->wd_logprint($INF, undef, "$proc: restarted too many times, exceeded TimesProcRestart (".$self->{h_procs}->{$proc}->{'TimesProcRestart'}.")!!");

                     %h_failure_info = ('failure' => FAILURE_PROC_RESTARTS_EXCEEDED,
                                        'detected_fails' => $self->{h_times_recent_proc_failed}->{$proc},
                                        'threshold_fails' => $self->{h_procs}->{$proc}->{'TimesProcRestart'},
                                        'proc_name' => $proc,
                                       );

                     $self->log_failure(\%h_failure_info);

                     exit;
                 }

                 $self->run_cmds($self->{h_procs}->{$proc}{'OnProcFailCommands'});
                        # restart watch
                        $self->proc_failure_occurs(\%h_failure_info);
                        return;
            }

        }
        else
        {
            $self->{h_times_recent_proc_failed}->{$proc} = 0;
            $self->{h_times_threads_exceed}->{$proc} = 0;
            $verify_rv = 1;
        }

        if($self->{actual_procs} < $self->{required_procs})
        {
            last if $single_proc_to_verify;

            $self->{h_times_procs_exceed}->{$proc} = 0 unless defined $self->{h_times_procs_exceed}->{$proc};
            $self->{h_times_procs_exceed}->{$proc}++;


            $self->wd_logprint($INF, undef, "$proc: not enough processes (found: $self->{actual_procs}, required: $self->{required_procs}, proc_found_in_list=$proc_found_in_list)!");
            # if passed TimesThreadsFail times - exec the OnProcFailCommands
            if ($self->{h_times_procs_exceed}->{$proc} >= $self->{h_procs}->{$proc}->{'TimesThreadsFail'})
            {

                 $self->wd_logprint($INF, undef, "$proc: not enough procs, exceeded TimesThreadsFail (".$self->{h_procs}->{$proc}->{'TimesThreadsFail'}.")!!");

                 %h_failure_info = ('failure' =>  FAILURE_NOT_ENOUGH_PROCS,
                                    'required_procs' => $self->{required_procs},
                                     'present_procs' => $self->{actual_procs},
                                     'proc_name'       => $proc);
                 $self->log_failure(\%h_failure_info);

                 $self->run_cmds($self->{h_procs}->{$proc}{'OnProcFailCommands'});
                        # restart watch
                        $self->proc_failure_occurs(\%h_failure_info);
                        return;
            }
        }
        else
        {
            $self->{h_times_procs_exceed}->{$proc} = 0;
            $verify_rv = 1;
        }

        # check for decreased number of threads
        if ($self->{h_procs}->{$proc}{'WatchDecreasedThreads'} && !($self->{actual_threads} < $self->{required_threads}))
        {
            $self->{h_max_thread_count}->{$proc} = 0 unless defined $self->{h_max_thread_count}->{$proc};

            if ($self->{actual_threads} < $self->{h_max_thread_count}->{$proc})
            {
                last if $single_proc_to_verify;

                $self->{h_times_threads_decreased}->{$proc} = 0 unless defined $self->{h_times_threads_decreased}->{$proc};
                $self->{h_times_threads_decreased}->{$proc}++;

                $self->wd_logprint($INF, undef, "$proc: decreased thread count (max: $self->{h_max_thread_count}->{$proc}, now: $self->{actual_threads}) !");

                if ($self->{h_times_threads_decreased}->{$proc} >= $self->{h_procs}->{$proc}->{'TimesThreadsFail'})
                {

                    $self->wd_logprint($INF, undef, "$proc: decreased number of threads, exceeded TimesThreadsFail (".$self->{h_procs}->{$proc}->{'TimesThreadsFail'}.")!!");

                    %h_failure_info = ('failure' =>  FAILURE_THREAD_NUM_DECREASED,
                                       'previous_threads' => $self->{h_max_thread_count}->{$proc},
                                       'current_threads' => $self->{actual_threads},
                                       'proc_name'       => $proc);
                    $self->log_failure(\%h_failure_info);

                    $self->run_cmds($self->{h_procs}->{$proc}{'OnProcFailCommands'});
                    # restart watch
                    $self->proc_failure_occurs(\%h_failure_info);
                    return;

                }
            }
            else
            {
                $self->{h_times_threads_decreased}->{$proc} = 0;
                $verify_rv = 1;
            }

            if($self->{actual_threads} > $self->{h_max_thread_count}->{$proc})
            {
                if(defined $self->{max_expected_threads})
                {
                    $self->{h_max_thread_count}->{$proc} = min($self->{max_expected_threads}, $self->{actual_threads});
                }
                else
                {
                    $self->{h_max_thread_count}->{$proc} = $self->{actual_threads};
                }
            }

        }

        # check CPU for all threads
        foreach my $pid (keys %{$self->{h_cpu}->{$proc}})
        {

            $self->wd_logprint($DBG, undef, "$proc: CPU for $pid is: $self->{h_cpu}->{$proc}->{$pid}->{cpu_usage}");

            # Skip if CpuThreshold is not defined for this process
            next unless $self->{h_procs}->{$proc}->{'CpuThreshold'};

            # if passed the CpuThreshold
            if($self->{h_cpu}->{$proc}->{$pid}->{cpu_usage} >= $self->{h_procs}->{$proc}->{'CpuThreshold'})
            {
                last if $single_proc_to_verify;

                $self->{h_times_cpu_exceed}->{$pid} = 0 unless defined $self->{h_times_cpu_exceed}->{$pid};
                $self->{h_times_cpu_exceed}->{$pid}++;

                $self->wd_logprint($INF, undef, "$proc: pid $pid has CPU ".$self->{h_cpu}->{$proc}->{$pid}->{cpu_usage}.", exceeding cpu threshold(".$self->{h_procs}->{$proc}->{'CpuThreshold'}.") ! ");

                # if passed CpuThreshold TimesCpuExceed times - exec the OnProcFailCommands
                if($self->{h_times_cpu_exceed}->{$pid} >= $self->{h_procs}->{$proc}->{'TimesCpuExceed'})
                {

                    $self->wd_logprint($INF, undef, "$proc: pid $pid passed TimesCpuExceed (".$self->{h_procs}->{$proc}->{'TimesCpuExceed'}.")! ");

                    %h_failure_info  = ('failure' =>  FAILURE_CPU_USAGE_FAILURE,
                                        'detected_cpu' => $self->{h_cpu}->{$proc}->{$pid}->{cpu_usage},
                                        'cpu_threshold' => $self->{h_procs}->{$proc}->{'CpuThreshold'},
                                        'proc_name'       => $proc);
                    $self->log_failure(\%h_failure_info);

                    $self->run_cmds($self->{h_procs}->{$proc}{'OnProcFailCommands'});
                    # restart watch
                    $self->proc_failure_occurs(\%h_failure_info);
                    return;

                }
            }
            else
            {
                $self->{h_times_cpu_exceed}->{$pid} = 0;
                $verify_rv = 1;
            }

        }
    }
    my $now = time();
    if($self->{is_provisioned_asm}
       && ($now - $self->{last_db_report_time} >=  $self->{ENFORCER_CPU_REPORT_TIME_INTERVAL})
       && F5::Provisioning::is_module_enabled('gtm')) {

        $self->store_enforcer_cpu_reading();
        $self->{last_db_report_time} = $now;

    }

    return $verify_rv;
}

############################################################################
#  sub: proc_failure_occurs
############################################################################
sub proc_failure_occurs{

    my ($self,$h_failure_info) = @_;

    my $now = time();

    #
    # do not log here, because by now the proc already got re-started by '$self->run_cmds'
    #     logging now creates missleading reports about fails
    # instead, this log was added before each one of the '$self->run_cmds' above - where it belongs
    #
    # $self->log_failure($h_failure_info);
    #

    my $proc_name = $h_failure_info->{proc_name};

    if ( $self->{h_procs}->{$proc_name}{'SaveCoreFileWithPid'} ) {
        my $glob_pattern = "/shared/core/${proc_name}.*.core{.gz,}";
        my ($corefile) = glob($glob_pattern);  #core file with no pid
        if ($corefile) {
            $self->wd_logprint($INF, undef, "$proc_name: SaveCoreFileWithPid found core file $corefile");
            my $pid = $self->{h_pids}->{$proc_name};
            if ( !$pid
                && open (my $PIDFILE, '<', "/var/run/${proc_name}.pid")) {

                $pid = <$PIDFILE>;
                close ($PIDFILE);
                chomp($pid);            
            }
            
            if ( !$pid ) {
                $self->wd_logprint($INF, undef, "$proc_name: SaveCoreFileWithPid is on, but could not determine last active PID.");
                $pid = time();
            } 

            my $new_filename = $corefile;
            $new_filename =~ s/\.core/\.core\.$pid/;
            move($corefile, $new_filename);
        }
    }

    # do not reset this counter here - 
    #     $self->{h_times_recent_proc_failed}->{$proc_name}; # count consecutive times proc failed
    # it is a global counter that is being reset upon start-up only

    delete $self->{h_max_thread_count}->{$proc_name}; # count max threads seen
    delete $self->{h_times_cpu_exceed}->{$proc_name}; # count times cpu exceed for each thread
    delete $self->{h_times_threads_exceed}->{$proc_name}; # count times threads not reached for each proc
    delete $self->{h_times_threads_decreased}->{$proc_name}; # count times threads decreased
    delete $self->{h_pids}->{$proc_name}; # Last seen pid per proc    
    my $ps = $self->ps();
    $self->{h_cpu} = $ps if defined $ps;


    return

}

############################################################################
#  sub: init
############################################################################
sub init{
    my ($self, %params) = @_;

    F5::Db::Single->init();
    $self->{script_dir} = dirname($0);

    $self->{time_counter} = 0;
    $self->{last_call_time} = 0;

    $self->{last_cpu_time} = time();
    $self->{relevant_cpu_count} = 0;
    $self->{last_db_report_time} = time();
    $self->{last_reading_time} = undef;

    # /proc directory
    $self->{PROC_DIR}  = '/proc';

    # max process name in /proc/stat file
    $self->{MAX_PROC_NAME} = 15;

    my $h_init = {};
    my $cfg_file = WD_CFG_FILE;

    $self->{INITIAL_SLEEP} = F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'InitialSleepTime');
    $self->{FREQUENCY} = F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'Frequency');
    $self->{LOOP_RESOLUTION} = F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'LoopResolution');
    $self->{CPU_CALC_INTERVAL} = F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'CpuCalcInterval');

   if( F5::Cluster::Utils::is_clustering_enabled() ){
      $self->{slot_number} = F5::Cluster::Utils::get_slot_number();
    }else{
      $self->{slot_number} = 0;
    }
    $self->{curr_master} = 0;

    $self->{TS_LOG_FILE} = 'ts.log'; # currently not read from configuration

    $self->{logger} = $params{logger} || F5::Logger->get_logger();
    $self->{ts_logger} = F5::Logger->new('/ts/log/'.$self->{TS_LOG_FILE});

    $self->{ENFORCER_CPU_REPORT_TIME_INTERVAL} = F5::Cfg::cfg_get_config_item($cfg_file, 'ENFORCER_CPU_REPORT', 'EnforcerCpuReportTimeInterval');
    $self->{relevant_cpu_count} = F5::NwdUtils::get_relevant_cpu_count();
    $self->wd_logprint($DBG, undef, "relevant_cpu_count=$self->{relevant_cpu_count}");
    $self->wd_logprint($DBG, undef, "-------------------------------------------------");
    my $l_sections = F5::Cfg::cfg_get_config_sections($cfg_file);

    my $now = time();

    foreach my $section (@$l_sections)
    {
        if($section =~ /WatchProc=(\S+)/)
        {

        my $proc_name = $1;

        my $should_watch_proc = 0;
        my $watch_for_provisioned_modules = F5::Cfg::cfg_get_config_item($cfg_file, $section, 'WatchForProvisionedModules') || 'asm';
        for my $module (split(/,/,$watch_for_provisioned_modules)) {
            if (! exists $self->{'is_provisioned_' . $module}) {
                $self->{'is_provisioned_' . $module} = F5::Provisioning::is_module_enabled($module);
            }

            if ($self->{'is_provisioned_' . $module}) {
                $self->wd_logprint($INF, 'init', "Watching '$proc_name' for '$module'");
                $should_watch_proc = 1;
            }
        }
        if (!$should_watch_proc) {
            $self->wd_logprint($INF, 'init', "Not watching '$proc_name'");
            next;
        }

        $h_init->{$proc_name}->{'TimesCpuExceed'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'TimesCpuExceed') || 0;
        $h_init->{$proc_name}->{'CpuThreshold'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'CpuThreshold') || 0;
        $h_init->{$proc_name}->{'MinimalNumThreads'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'MinimalNumThreads');
        $h_init->{$proc_name}->{'MinimalNumProcs'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'MinimalNumProcs') || 0;
        $h_init->{$proc_name}->{'MaxMemorySize'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'MaxMemorySize') || 0;
        $h_init->{$proc_name}->{'TimesThreadsFail'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'TimesThreadsFail');
        $h_init->{$proc_name}->{'TimesProcRestart'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'TimesProcRestart') || 0;
        $h_init->{$proc_name}->{'OnProcFailCommands'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'OnProcFailCommands');
        $h_init->{$proc_name}->{'SaveCoreFileWithPid'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'SaveCoreFileWithPid');
        $h_init->{$proc_name}->{'MaxExpectedThreads'} = F5::Cfg::cfg_get_config_item($cfg_file, $section, 'MaxExpectedThreads');
        $h_init->{$proc_name}->{'WatchDecreasedThreads'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'WatchDecreasedThreads') || 0;
        $h_init->{$proc_name}->{'CheckFullProcessName'}= F5::Cfg::cfg_get_config_item($cfg_file, $section, 'CheckFullProcessName');
        $h_init->{$proc_name}->{'NextWatchTime'}= $now + $self->{INITIAL_SLEEP};

        my $msg = " \"$proc_name\": TimesCpuExceed = $h_init->{$proc_name}->{'TimesCpuExceed'}, ".
                        "CpuThreshold = $h_init->{$proc_name}->{'CpuThreshold'}, ".
                        "MaxExpectedThreads = $h_init->{$proc_name}->{'MaxExpectedThreads'}, ".
                        "MinimalNumThreads = $h_init->{$proc_name}->{'MinimalNumThreads'}, ".
                        "MinimalNumProcs = $h_init->{$proc_name}->{'MinimalNumProcs'}, ".
                        "TimesThreadsFail = $h_init->{$proc_name}->{'TimesThreadsFail'}, ".
                        "TimesProcRestart = $h_init->{$proc_name}->{'TimesProcRestart'}, ".
                        "OnProcFailCommands = $h_init->{$proc_name}->{'OnProcFailCommands'}, ".
                        "WatchDecreasedThreads = $h_init->{$proc_name}->{'WatchDecreasedThreads'}, ".
                        "SaveCoreFileWithPid = $h_init->{$proc_name}->{'SaveCoreFileWithPid'}, ".
                        "CheckFullProcessName = $h_init->{$proc_name}->{'CheckFullProcessName'}, ".
                        "MaxMemorySize = $h_init->{$proc_name}->{'MaxMemorySize'}, ".
                        "NextWatchTime = $h_init->{$proc_name}->{'NextWatchTime'}" ;

        $self->wd_logprint($DBG, 'init', $msg);

        }
    }

        $self->{h_max_thread_count} = {}; # count max threads seen
        $self->{h_times_cpu_exceed} = {}; # count times cpu exceed for each thread
        $self->{h_times_threads_exceed} = {}; # count times threads not reached for each proc
        $self->{h_times_recent_proc_failed} = {}; # count consecutive times proc failed
        $self->{h_times_threads_decreased} = {}; # count times threads decreased
        $self->{h_pids} = {}; # Last seen pid per proc
        my $ps = $self->ps();
        $self->{h_cpu} = $ps if defined $ps;

   my $dbh = F5::Db::Single->instance;
   eval {
       my $ary_ref = $dbh->selectcol_arrayref("SELECT field_name, current_value FROM DCC.INTERNAL_CONFIG_ITEMS", { Columns=>[1, 2] });
       $self->{internals} = {@{$ary_ref}};
    };
    if ($@) {
        $self->{logger}->info("Could not retrieve from INTERNAL_CONFIG_ITEMS ($@)") if $self->{logger};
        return;
    }

   # when bd send security events messages to bd_agent (internal 'send_content_events' is set to 1)
   #   the latter should not be limited by CPU threshold as it will be working very hard
   # each time internals change asm should be restarted - we'll go through this init
   #   and these values will be re-adjusted accordingly
   if ($self->{internals}->{send_content_events} == 1) {
      $self->wd_logprint($INF, 'init', "Detected 'send_content_events=1' - disabling 'CpuThreshold' and 'TimesCpuExceed' for 'bd_agent'.");
      $h_init->{bd_agent}->{CpuThreshold} = 0;
      $h_init->{bd_agent}->{TimesCpuExceed} = 0;

      my $msg = " 'bd_agent': TimesCpuExceed = $h_init->{'bd_agent'}->{'TimesCpuExceed'}, ".
                        "CpuThreshold = $h_init->{'bd_agent'}->{'CpuThreshold'}, ".
                        "MaxExpectedThreads = $h_init->{'bd_agent'}->{'MaxExpectedThreads'}, ".
                        "MinimalNumThreads = $h_init->{'bd_agent'}->{'MinimalNumThreads'}, ".
                        "TimesThreadsFail = $h_init->{'bd_agent'}->{'TimesThreadsFail'}, ".
                        "TimesProcRestart = $h_init->{'bd_agent'}->{'TimesProcRestart'}, ".
                        "OnProcFailCommands = $h_init->{'bd_agent'}->{'OnProcFailCommands'}, ".
                        "WatchDecreasedThreads = $h_init->{'bd_agent'}->{'WatchDecreasedThreads'}, ".
                        "SaveCoreFileWithPid = $h_init->{'bd_agent'}->{'SaveCoreFileWithPid'}, ".
                        "CheckFullProcessName = $h_init->{'bd_agent'}->{'CheckFullProcessName'}, ".
                        "NextWatchTime = $h_init->{'bd_agent'}->{'NextWatchTime'}" ;

      $self->wd_logprint($DBG, 'init', $msg);
   }

   $self->{h_procs} = $h_init;
   $self->{process_handler} = $params{process_handler} || F5::ProcessHandler->new(logger => $self->{logger});
}

############################################################################
#  sub: get_process_table
#  Description: - receives: list of process ids for process
#               - returns: return reference hash to the process table
############################################################################
sub get_process_table
{

    my ($self) = @_;

    my @proc_arr;

    chdir $self->{PROC_DIR};

    # look for process (pid)s under /proc directory
    # open /proc directory
    if ( ! opendir(DIR, $self->{PROC_DIR}  )  )
    {
      $self->wd_logprint($F_ERR, undef, "Failed to open  $self->{PROC_DIR} directory");
      exit(-1);
    }

    while (defined (my $entry = readdir DIR))
    {
       # look only for pid directories

       next if ( not -d $entry);
       next if ($entry !~ /^\.?\d+$/);

       my $process_stat = F5::NwdUtils::get_process_stat($entry);
       next unless %$process_stat;

       my $name = $process_stat->{process_name};

       # Get full process name
       my $cmdline_fh;
       if (! open($cmdline_fh, '<', "$self->{PROC_DIR}/$entry/cmdline")) {
           next;
       }
       my $full_cmd_name = <$cmdline_fh>;
       close($cmdline_fh);

       if ($full_cmd_name) {
           chomp($full_cmd_name);
           my @cmd_parts = split(/\000/, $full_cmd_name);
           if ($cmd_parts[0] eq '/usr/bin/perl' || $cmd_parts[0] eq '/bin/perl') {
               shift(@cmd_parts);
           }
           $full_cmd_name = basename($cmd_parts[0]);
       }

       # do we want this process?
       my $full_name;
       for my $wanted_proc ( keys %{$self->{h_procs}} ) {
           if ($self->{h_procs}->{$wanted_proc}->{CheckFullProcessName}) {
               if ($full_cmd_name && $full_cmd_name eq $wanted_proc) {
                   $full_name = $full_cmd_name;
                   last;
               }
           }
           else {
               if ( $name eq substr($wanted_proc, 0, $self->{MAX_PROC_NAME}) ) {
                   $full_name = $wanted_proc;
                   last;
               }
           }
       }
       next unless defined $full_name;

       my $pid = $process_stat->{pid};
       my $number_of_threads = $process_stat->{thread_count};

       if($number_of_threads == 0) {
           # sanity check - this does not make sense, the process is listed but has 0 threads. Giving a default of the minimal required
           $number_of_threads = $self->{h_procs}->{$full_name}->{MinimalNumThreads};
           $self->wd_logprint($INF, undef, "setting number of threads to $number_of_threads since kernel set 0 in thread count");
       }
       # init process data

       my %proc = (dname  => $entry, fname  => $full_name, pid => $pid,  number_of_threads => $number_of_threads );

       push @proc_arr, \%proc;
    }

    chdir $self->{script_dir};

    # return array reference to process table
    return \@proc_arr;
}

############################################################################
# sub: ps
############################################################################
sub ps
{
    my ($self) = @_;

    my $h_cpu = {};

    return unless time() - $self->{last_cpu_time} >= $self->{CPU_CALC_INTERVAL};

    $self->{last_cpu_time} = time();

    my $proc_table = $self->get_process_table();

    foreach my $proc_entry (@$proc_table) {

        my ($pid, $fname, $dname, $number_of_threads) = ($proc_entry->{pid}, $proc_entry->{fname}, $proc_entry->{dname}, $proc_entry->{number_of_threads});

        if ( exists $self->{h_procs}->{$fname} ) {
            my $process_stat = F5::NwdUtils::get_process_stat($dname);
            if (! %$process_stat)  {
                $self->wd_logprint($DBG, undef, "Failed to open process stat file - process $pid has probably exited($proc_entry->{fname})");
                next;
            }

            my $now = time();
            my $curr_stat = $process_stat->{cpu_user_time} + $process_stat->{cpu_system_time}; # in jiffies, 1/100 of a second

            if (! defined $self->{h_prev_stat}->{$pid}) {
                $self->{h_prev_stat}->{$pid}->{cputime} = $self->{h_prev_stat}->{$pid}->{realtime} = 0;
            }

            my $time_passed_since_prev_check = $now - $self->{h_prev_stat}->{$pid}->{realtime};
            if(! $time_passed_since_prev_check) {
                $time_passed_since_prev_check = 1; 
            }
            $h_cpu->{$fname}->{$pid}->{cpu_usage} = ($curr_stat - $self->{h_prev_stat}->{$pid}->{cputime}) / $time_passed_since_prev_check;
            $h_cpu->{$fname}->{$pid}->{number_of_threads} = $number_of_threads;
            $h_cpu->{$fname}->{$pid}->{memory_size} = $process_stat->{rss};
            $self->{h_prev_stat}->{$pid}->{cputime} = $curr_stat;
            $self->{h_prev_stat}->{$pid}->{realtime} = $now;

        }
    }

    return $h_cpu;
}

############################################################################
# sub: store_enforcer_cpu_reading
############################################################################
sub store_enforcer_cpu_reading {
    my ($self) = @_;

    my $h_cpu = $self->{h_cpu};

    my $reading = 0;
    for my $pid (keys %{$h_cpu->{'bd'}}) {
        $reading += ($h_cpu->{'bd'}->{$pid}->{cpu_usage});
    }

         
    $reading /= $self->{relevant_cpu_count} if ($self->{relevant_cpu_count});
        
    if ($reading > 100) {
        $reading  = 100;
    }

    my $now = time();

    $self->wd_logprint($DBG, undef, "enforcer cpu reading is $reading");

    if( !F5::Cluster::Utils::is_failover_standby() && !(F5::Cluster::Utils::is_clustering_enabled() ) || (F5::Cluster::Utils::is_master_blade()) ){
        my $score = ( 100 - int($reading) ) * $self->{relevant_cpu_count}; #Theoretically should be multiplied by the number of active slots

        $self->wd_logprint($DBG, undef, "Calculated score is $score");

        eval {
            F5::NwdUtils::UpdateGTMScore::update_GTM_score(score => $score,
                                                           logger => $self->{logger});
        };
        if ($@) {
            $self->wd_logprint($INF, undef, "Could not update GTM score.");
        }
    }

}

############################################################################
# sub: log_failure
############################################################################
sub log_failure
{
   my ($self, $failure_info) = @_;

   # build the log line
   $self->wd_logprint($DBG, undef, "Adding failure to ts.log and events");
   my $line = 'Watchdog detected failure for process. Process name: '.$failure_info->{proc_name}. ", Failure: ".$failure_info->{failure};
   if($failure_info->{failure} eq FAILURE_NOT_ENOUGH_THREADS)
   {
      $line .= " (required: ".$failure_info->{required_threads}.", found: ".$failure_info->{present_threads}.").";
   }
   elsif($failure_info->{failure} eq FAILURE_PROC_RESTARTS_EXCEEDED)
   {
      $line .= " (detected: ".$failure_info->{detected_fails}.", threshold: ".$failure_info->{threshold_fails}.").";
   }
   elsif($failure_info->{failure} eq FAILURE_NOT_ENOUGH_PROCS)
   {
      $line .= " (required: ".$failure_info->{required_procs}.", found: ".$failure_info->{present_procs}.").";
   }
   elsif($failure_info->{failure} eq FAILURE_THREAD_NUM_DECREASED)
   {
      $line .= " (previous number: ".$failure_info->{previous_threads}.", current: ".$failure_info->{current_threads}.").";
   }
   elsif($failure_info->{failure} eq  FAILURE_CPU_USAGE_FAILURE)
   {
      $line .= " (detected cpu usage: ".$failure_info->{detected_cpu}.", threshold: ".$failure_info->{cpu_threshold}.").";
   }
   elsif($failure_info->{failure} eq  FAILURE_MEMORY_SIZE_USAGE_FAILURE)
   {
      $line .= " (detected memory usage: ".$failure_info->{memory_size}.", threshold: ".$failure_info->{max_memory_size}.").";
   }
   else
   {
      $self->wd_logprint($ERR, undef, "Internal error: bad failure_info structure, can not export report on this failure");
      return FALSE;
   }
   # add a ts.log line
   $self->ts_log_logprint($ERR, undef, $line);
}

############################################################################
# sub: ts_log_logprint
############################################################################
sub ts_log_logprint
{
    my ($self,$level,$action, $msg) = @_;
    # no debug prints are allowed to ts.log!
    return 0 if($level eq $DBG);
    $self->{ts_logger}->log_print($level, $action, $msg);
}


############################################################################
# sub: wd_logprint
############################################################################
sub wd_logprint
{
    my ($self,$level,$msg1,$msg2) = @_;

    $self->{logger}->log_print($level, $msg1, $msg2);
}

############################################################################
# sub: min
############################################################################
sub min
{
   my ($a, $b) = @_;
   return ($a < $b) ? $a : $b;
}

############################################################################
# sub: run_cmds
############################################################################
sub run_cmds {
    my ($self,$cmd_string) = @_;

    $self->{process_handler}->handle_actions(split(/,/, $cmd_string));
}

1;
