package F5::NwdUtils::UpdateGTMScore;

=head1 GTM/ASM integrations
***********************************************
Description : Give the ability for an ASM enabled LTM to maintain a gtm_vs_score for each LTM virtual associated with the ASM module. The score reflects in a proportional manner the ability of that virtual to handle ASM traffic.

Created by  : Avivit Polak

=cut
use strict;
use warnings;

use F5::ASM::COAPI;
use F5::Logger;
use F5::Mcp::PolicyTarget;
use F5::Mcp::PolicyAction;


use constant GTM_VS_SCORE_MODULE_ASM        => 0;


########################################################################################
# Updating the score:
# For each ASM virtual, add/create a gtm_vs_score object which has its vs_name
# set to the the name of the virtual and the module_type set to GTM_VS_SCORE_MODULE_ASM
########################################################################################
sub update_GTM_score{

    my(%params)    = @_ ;
    my $score      = $params{score};
    my $logger     = $params{logger};


    $logger->trace("Starting ...") if ($logger);

        if ( not defined $score ) {
             $logger->debug("Score is not defined") if ($logger);
             return;
        }

        if ( (int($score)) != ($score)) {
            $logger->error("Score should be an integer, got score value: $score") if ($logger);
            return;
        }

        if (!(F5::ASM::COAPI::Login("%asm_gtm_vs_score"))) {
            $logger->error("could not log in") if ($logger);
            return;
        }

        my %ASM_vs = get_ASM_virtuals(logger=>$logger);

        if ( !( clean_up(ASM_vs=>\%ASM_vs,logger=>$logger) ) ) {
            $logger->error("Failed to clean up") if ($logger);
            return;
        }

        my @updated_objs;

        for my $vs (keys %ASM_vs) {
            my $result = F5::ASM::COAPI::Object->new("gtm_vs_score");
            if (!($result)) {
                $logger->error("could not create gtm_vs_score object") if ($logger);
                return;
            }

            $result->{"vs_name"} = $vs;
            $result->{"module_type"} = GTM_VS_SCORE_MODULE_ASM ;
            $result->{"score"} = $score;

            $logger->debug("updated $vs with score $score") if ($logger);

            push (@updated_objs, $result);
        }

        eval {
            F5::ASM::COAPI::TransactionalSave(\@updated_objs);
        };
        if (my $err = $@) {
            $logger->error("could not save gtm_vs_score objects: $err $F5::COAPI::error") if ($logger);
            return;
        }

        $logger->info("Successfully finished updating gtm_vs_score with value $score") if ($logger);

    return;
}

########################################################################################
# clean_up:
#    * Cleaning up: erasing virtual servers which are no longer connected to ASM
########################################################################################
sub clean_up{

    my(%params) = @_;

    my $logger     = $params{logger};
    my $new_ASM_vs = $params{ASM_vs} ;
    my $log_in     = $params{log_in};

    if($log_in){
        if(!(F5::ASM::COAPI::Login("%asm_gtm_vs_score"))){
            $logger->error("could not log in") if ($logger);
            return;
         }

        $logger->trace("Successfully logged in for cleaning up") if ($logger);
       
    }

    my $prev_gtm = F5::ASM::COAPI::Query->new('gtm_vs_score');
    if(!($prev_gtm)){
        $logger->error("could not query on gtm_vs_score object") if ($logger);
        return 0;
    }

    my @objects;
    for my $prev_vs(@{$prev_gtm}) {
        if ( $prev_vs->{module_type} == GTM_VS_SCORE_MODULE_ASM && !(exists $new_ASM_vs->{$prev_vs->{vs_name}}) ) {
            $prev_vs->delete_on_save();
            push (@objects, $prev_vs);  
        }
    }

    eval {
        F5::ASM::COAPI::TransactionalSave(\@objects);
    };
    if (my $err = $@) {
        $logger->error("Could not delete gtm_vs_score objects: $err $F5::COAPI::error") if ($logger);
        return 0;
    }

    return 1; #Success
}

########################################################################################
# get_ASM_profiles:
#    * Create hash of all profiles with ‘security_enabled’ flag on
########################################################################################
sub get_ASM_profiles{

    my(%params)  = @_ ;
    my $logger   = $params{logger};

    my %ASM_profiles = ();

    my $ftp_profiles = F5::ASM::COAPI::Query->new('profile_ftp');
    my $websecurity_profiles = F5::ASM::COAPI::Query->new('profile_websecurity');
    my $smtp_profiles = F5::ASM::COAPI::Query->new('ltcfg_instance_field',{class_name => 'profile_smtp'});

    if(!($ftp_profiles && $smtp_profiles)){
        $logger->error("could not query profiles") if ($logger);
        return;
     }

    #Getting all FTP data profiles that have security turned on.
    for my $prof (@{$ftp_profiles}) {
       if($prof->{security_enabled} == 1){
           $ASM_profiles{($prof->{name})} = 1;
       }
    }

    #Websecurity profiles signify ASM usage
    for my $prof (@{$websecurity_profiles}) {
        $ASM_profiles{($prof->{name})} = 1;
    }

    #Getting all SMTP data profiles that have security turned on.
    for my $prof (@{$smtp_profiles}) {
       if(($prof->{field_name} eq 'security_enabled') and ($prof->{value} eq 'true')){
           $ASM_profiles{($prof->{instance_name})} = 1;
       }
    }
   
    $logger->trace("-----ASM_profiles------") if ($logger);
    $logger->trace(\%ASM_profiles) if ($logger);

    return %ASM_profiles;
}

########################################################################################
#Getting the list of virtuals is by unifying the following lists of virtuals:
#    * All virtuals with assigned class that has ASM enabled.
#    * All virtuals with assigned data profile (SMTP/FTP/HTTP) that has security turned on.
########################################################################################
sub get_ASM_virtuals{

    my(%params) = @_;
    my $logger     = $params{logger};

    my %ASM_vs = ();

    my %ASM_profiles = get_ASM_profiles(logger => $logger);

    my $vs_profiles = F5::ASM::COAPI::Query->new('virtual_server_profile');

    if(!($vs_profiles)){
        $logger->error("could not query profiles") if ($logger);
        return;
    }

    #Getting all virtuals with assigned data profile HTTP/FTP/SMTP that has security turned on.
    for my $vs (@{$vs_profiles}) {
       if(exists $ASM_profiles{ $vs->{profile_name} }) {
            $ASM_vs{ $vs->{vs_name} } = 1;
       }
    }

    $logger->trace("-----------ASM_vs-----------------") if ($logger);
    $logger->trace(\%ASM_vs) if ($logger);

    return %ASM_vs;

}


#package return:
1;
