package F5::PEM::Tacdb::AutoDownload;

use strict;
use warnings;

use base qw(F5::Sigfile::AutoDownload);

use F5::Iface;
#use SOAP::Lite +trace => qw(debug);
use SOAP::Lite;
use MIME::Base64;
use Config::IniFiles;
use F5::BigipVersionUtils;
use F5::License;
use F5::ServerLookup;
use F5::DateTimeUtils;
use F5::ASM::COAPI;

my $save_name = 'gsmadb.bin';
my $CONF_INI_FILE = '/ts/etc/services.ini';
#----------------------------------------------------------------------
sub init {
  my($self, %params) = @_;
  $self->SUPER::init(%params);
  my $timestamp;
  my $query = F5::ASM::COAPI::Query->new("tacdb_licenseddb");
  if ($query && @$query) {
      my ($version) = @$query;
      $timestamp = F5::DateTimeUtils::convert_epoch_to_datetime($version->{update_time});
  }
  $self->{sigfile_timestamp} = $timestamp || '2001-01-01 00:00:00'; # Arbitrarily old date to force update
  #Take the oldest timestamp as the default for downloading a new file
  if (!$self->{sigfile_timestamp} ||
      $self->{sigfile_timestamp} ge $self->{sigfile_timestamp}) {
      $self->{sigfile_timestamp} = $self->{sigfile_timestamp};
  }
}

#----------------------------------------------------------------------
sub is_newer_sigfile_available {
  my($self) = @_;

  if (F5::License->new->is_aws_icm_platform()) {
      $self->{logger}->info("Updates are disallowed on an IC platform") if $self->{logger};
      return;
  }

  return 1;
}

#----------------------------------------------------------------------
sub get_latest_sigfile {
  my($self, $save_dir) = @_;

  my $soap_resp = $self->call_soap_server(check_update_only => 0);
  my $filename = "$save_dir/$save_name";
  my $url = $soap_resp->valueof('//PemUpdateResponse/PemDbFile/DownloadUrl');
  system("curl -o $filename '$url'");
  $self->{logger}->debug("curler result is $?");
  if ($? != 0) {
    $self->{logger}->debug("Downloading failed.");
    $self->store_error_and_die("Cannot download file");
  } else {
    $self->{logger}->debug("Downloading successful.");
  }

  $self->{logger}->debug("Received file: '$filename'");
  return $filename;
}

#----------------------------------------------------------------------
sub call_soap_server {
  my($self, %params) = @_;

  my $config = Config::IniFiles->new(-file => $CONF_INI_FILE);

  my $last_updated_timestamp = $self->{sigfile_timestamp};

  my $host;
  my $url = '/soa/softwareupdate/';
  my $soap_action = 'PemUpdateRequest';

  my $proxy_params = $self->get_https_proxy_params();
  my $proxy_uri = $proxy_params->{proxy_uri};
  my $proxy_username = $proxy_params->{proxy_username};
  my $proxy_password = $proxy_params->{proxy_password};

  local $ENV{HTTPS_CA_FILE};
  if (F5::License->new->is_production) {
    $host = $config->val('sigfile_soap_service', 'prod_server_name');
    $ENV{HTTPS_CA_FILE} = '/config/ssl/ssl.crt/ca-bundle.crt';
  }
  else {
    $host = $config->val('sigfile_soap_service', 'tacdb_dev_server_name');
  }

  local @LWP::Protocol::http::EXTRA_SOCK_OPTS;

  my $host_name = $host;
  unless ($proxy_uri) {
    my $host_ip = F5::ServerLookup::get_ip_address($host, $self->{outbound_call_timeout});
    @LWP::Protocol::http::EXTRA_SOCK_OPTS = ( PeerAddr => $host_ip,
                                               SSL_verifycn_name => $host,
                                               SSL_hostname      => $host,
                                               Host     => $host);
  }

  my %proxy_opts;
  $proxy_opts{proxy} = [ 'https' => undef ];
  local $ENV{PERL_LWP_SSL_VERIFY_HOSTNAME} = $proxy_uri ? 0 : 1;
  local $ENV{HTTPS_PROXY} = $proxy_uri;
  local $ENV{HTTPS_PROXY_USERNAME} = $proxy_username;
  local $ENV{HTTPS_PROXY_PASSWORD} = $proxy_password;

  if ($self->{outbound_call_timeout}) {
      $proxy_opts{timeout} = $self->{outbound_call_timeout};
  }

  my $soap_service = SOAP::Lite
    -> readable(1)
      -> uri ("https://$host$url")
        -> on_action( sub { return $soap_action } )
          -> proxy ("https://$host$url", %proxy_opts)
            -> on_fault(sub {
                          my($soap, $res) = @_;
                          die ref $res ? $res->faultdetail : $soap->transport->status, "\n";
                        });
  $soap_service->transport->http_request->header('Host' => $host_name);
  my $authoriz = 'Basic ZHBpQGY1LmNvbTpEUElhdXRvU0lHdXBkYXRl';
  $soap_service->transport->http_request->headers->push_header('Authorization' => $authoriz);

  if (F5::License->new->is_production) {
    $soap_service->transport->http_request->header('If-SSL-Cert-Subject' => '/CN='.quotemeta($host_name).'$');
  }

  my $bigip_version = F5::BigipVersionUtils::bigip_version();
  my $bigip_version_base = (split(/\./, $bigip_version))[0];
  $soap_service->transport->http_request->header('User-Agent' => "bigip/$bigip_version");

  $self->{logger}->debug("Timestamp: $last_updated_timestamp, BigIP: $bigip_version");
  my $data = SOAP::Data->value(
                               SOAP::Data->name("su:ApiVersion" => '2015-10-01'),
                               SOAP::Data->name("su:UnitIdentifier" => F5::Iface::iface_get_unit_id(dbh => $self->{dbh})),
                               SOAP::Data->name("su:Product" => 'big-ip'),
                               SOAP::Data->name("su:Version" => '12.1.0'),
                               SOAP::Data->name("su:PemTacDbVersion" => "1.0"),
                               SOAP::Data->name("su:Build" => "2806.0"),
                               SOAP::Data->name("su:LastUpdatedFilter" => $last_updated_timestamp . ' GMT'),
                              );

  my $serviceNs = 'http://callhome.f5.com/schemas/su';

  my $soap_response;
  eval {
      local $SIG{ALRM} = sub { die "Timeout triggered\n" };
      if ($self->{outbound_call_timeout}) {
          alarm $self->{outbound_call_timeout};
      }

      $soap_response = $soap_service->call(
                                           SOAP::Data->name('su:PemUpdateRequest') ->attr({'xmlns:su' => $serviceNs})
                                           => $data
                                          );
  };

  my $ex = $@;
  if ($self->{outbound_call_timeout}) {
      alarm 0;
  }

  if ($ex) {
      $self->{logger}->info("SOAP request failed:") if $self->{logger};
      $self->{logger}->info($ex) if $self->{logger};
      die "No response from update server\n";
  }

  if ( $soap_response->fault ) {
     die $soap_response->faultcode . " aha " . $soap_response->faultstring . "\n";
  }
  return $soap_response;
}

#----------------------------------------------------------------------

1;

