package F5::PEM::Tacdb::Update;

use strict;
use warnings;

use base qw(F5::Sigfile::Update);

use Time::Local;
use File::Basename;
use File::Path;
use File::Copy;
use File::Find;
use Data::Dumper; # for dumping hashes
use F5::Logger;
use F5::PEM::Tacdb::AutoDownload;
use F5::BigipVersionUtils;
use F5::ASM::COAPI;
use F5::ASM::Tmsh;
use F5::DateTimeUtils;
use File::Temp qw/tempfile tempdir/;

my %AUTO_UPDATE_INTERVAL_MAP = ( 1 => "daily",
                                 2 => "weekly",
                                 3 => "monthly",);

my %STATUS_MAP = ( DEFAULT => 0,
                   SUCCESS => 1,
                   FAILURE => 2,
                   IN_PROGRESS => 3,);

my $PEM = "/var/local/pem/dtos";
my $CONF = "$PEM/conf/tacdb_update.conf";
my $LICENSE_FILE = '/config/bigip.license';

#----------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;
    $self->SUPER::init(%params);
    $self->{sigfile_type} = "Licensed Tacdb";
    $self->{tacdb_type} = "Licensed Tacdb";
    $self->{save_dir} = $PEM;
    $self->{tempdir} = "";
    $self->{update_installed} = 0;

    system("mkdir -p $PEM/{im,conf,reset}");
    system("touch $CONF");
    $self->{lock_file} = '/ts/lock/pem_tacdb.lock';
}

#----------------------------------------------------------------------
sub get_schedule_info {
    my ($self, %params) = @_;
    my @schedules;
    my $query = F5::ASM::COAPI::Query->new("tacdb_licenseddb");
    if ($query && @$query) {
        for my $schedule (@$query) {
            my $rv;
            $rv->{auto_update_interval} = $AUTO_UPDATE_INTERVAL_MAP{2};
            $rv->{update_mode} = ($schedule->{poll_interval} != 0) ? "auto-interval"
                                                                    : "auto-user";
            $rv->{interval_count} = $schedule->{poll_interval};
            $rv->{last_update_attempt} = 0;
            my ($update_status) = @$query;
            if ($update_status) {
                $rv->{last_update_attempt} = $update_status->{last_sync_datetime};
            }
            push (@schedules, $rv);
        }
    }

    return @schedules;
}

sub retval {
    my($result, %params) = @_;
    my $rv = {completed_successfully => $result->{completed_successfully},
        msg => $result->{standard_error}, };
    return $rv;
}

#----------------------------------------------------------------------
sub apply_changes {
  my($self, %params) = @_;

  $self->{update_installed} = 1;
  return { completed_successfully => 1 };
}


#----------------------------------------------------------------------

sub start_update {
    my ($self, %params) = @_;
    $self->report_status();
}

#----------------------------------------------------------------------
sub report_status {
  my($self, %params) = @_;
  my $now = time();
  my $status;
  my $query = F5::ASM::COAPI::Query->new("tacdb_licenseddb");
  if ($query && @$query) {
      ($status) = @$query;
  }
  else {
      $status = F5::ASM::COAPI::Object->new("tacdb_licenseddb");
      $status->{name} = "/Common/status";
  }

  if ($status->{progress_status} != $STATUS_MAP{IN_PROGRESS}) {
      $status->{progress_status} = $STATUS_MAP{IN_PROGRESS};
  } else {
      if (defined $params{completed_successfully} && $params{completed_successfully}) {
          $status->{progress_status} = $STATUS_MAP{SUCCESS};
          if ($self->{update_installed}) {
              $status->{last_sync_datetime} = $now;
          }
      } else {
          $status->{progress_status} = $STATUS_MAP{FAILURE};
      }
  }
  $status->save();

  if ($self->{update_installed}) {
      my @tmsh_cmd = ("tmsh", "save", "sys", "config");
      F5::ASM::Tmsh::Query->new(cmd => \@tmsh_cmd, logger => $self->{logger});
  }

}

#----------------------------------------------------------------------
sub check_service_agreement {
  my($self, %params) = @_;

  if(!$self->is_tacdb_license_valid()) {
      $self->store_error_and_die("GSMA License not valid");
  }

  return $self->SUPER::check_service_agreement(%params);
}

#----------------------------------------------------------------------

sub is_tacdb_license_valid {
    my($self) = @_;
    my $logger = $self->{logger};
    my $ret = open (my $license_fh, '<', $LICENSE_FILE);
    return unless ($ret);

    while(my $line = <$license_fh>) {
        if ($line =~ /^(\d{4})(\d{2})(\d{2})_pem_dtos_gsma_db\s+:\s+([-\w]+)/) {
            my ($year, $month, $day, $hour, $minute, $second) = ($1, $2, $3, "00", "00", "00");
            my $license_status = $4;
            my $expire_time_str = "$year-$month-$day $hour:$minute:$second";
            my $expire_time = F5::DateTimeUtils::convert_datetime_to_epoch($expire_time_str);
            my $now = time();

            if (($now < $expire_time) && (lc $license_status eq "enabled")) {
                $logger->debug("GSMA license is valid") if $logger;
                return 1;
            }
        }
    }

    return 0;
}

#----------------------------------------------------------------------
sub get_autodownload {
    my ($self, %params) = @_;
    return F5::PEM::Tacdb::AutoDownload->new(logger => $self->{logger},
            outbound_call_timeout => $self->{outbound_call_timeout},
            );
}

sub process_im {
    my ($self, %params) = @_;

    return %params;
}

sub DESTROY {
    my $self = shift;

    $self->{logger}->info("PEM.TACDB.DESTROY: Update installed:$self->{update_installed}");
}

1;
