package F5::PeriodicTask::ActivePolicySweep;

use strict;
use warnings;
use base qw(F5::PeriodicTask::Base);
use F5::BigIpLog;
use F5::ASM::COAPI;
use F5::Db::Single;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::Policy;
use F5::ASMConfig::Entity::L7Policy;
use F5::Cfg;
use F5::Mcp::PolicyTarget;
use F5::Mcp::PolicyAction;
use F5::DeviceGroup;

use constant ASMCROND_CFG_FILE  =>  '/etc/ts/tools/asmcrond.cfg';

my $WEEK_IN_SECONDS = 60*60*24*7;

my $GRACE_PERIOD = F5::Cfg::cfg_get_config_item(ASMCROND_CFG_FILE, 'ActivePolicySweep', 'GracePeriod') || $WEEK_IN_SECONDS;

#----------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;
    $self->SUPER::init(%params);
}

#----------------------------------------------------------------------

sub module_run {
    my ($self, %params) = @_;

    my $fo_dg = F5::DeviceGroup::get_failover_device_group();
    my $mcp_device_group = F5::DeviceGroup::get_device_group();

    if ($fo_dg->{name} && $mcp_device_group->{name} && 
        $fo_dg->{name} ne $mcp_device_group->{name}) {

        my $skip_reason = "this device is a member of a sync-failover device group '$fo_dg->{name}' and a sync-only device group '$mcp_device_group->{name}' where asm-sync is enabled.";

        if ($self->{logger}) {
            $self->{logger}->info("Skipping active policy sweeper. Reason: $skip_reason");
        }
        F5::BigIpLog::bigip_ts_ts_info_msg_log("Skipping active policy sweeper. Reason: $skip_reason");

        return;
    }

    my $dbh = F5::Db::Single->instance;   

    my @search_filters = (
            F5::ASMConfig::SearchFilter->new({field_name => 'flg_is_candidate_for_sweep',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                              field_value => 1,
                                             }),
            F5::ASMConfig::SearchFilter->new({field_name => 'is_active_policy',
                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                              field_value => 1,
                                             }),
        );

    my $policy_iter = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $dbh,
                                                                  search_filters => \@search_filters,
                                                                 );

    while (my $policy = $policy_iter->get_next()) {
        my $policy_name = $policy->{policy_name};

        # has L7 Policy associated
        my $assigned_to_l7_policy = 0;
        my %l7_policy_names = map { ( $_->{policy_name} => 1) } (@{F5::ASM::COAPI::Query->new("policy_action",
                                                                                           { target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM,
                                                                                             action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE,
                                                                                             policy => $policy_name},
                                                                                           )});

        #Sometimes policy_actions are left in MCP without corresponding l7_policy entities.
        for my $l7_policy_name (keys %l7_policy_names) {
            my $l7_policy = F5::ASMConfig::Entity::L7Policy->new(get_criteria => {l7_policy_name => $l7_policy_name}, dbh => $dbh);
            if ($l7_policy) {
                $assigned_to_l7_policy = 1;
                last;
            }
        }
        next if $assigned_to_l7_policy;

        if (!F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $GRACE_PERIOD,
                                                                start_time => $policy->{created_time})
            &&
            !F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $GRACE_PERIOD,
                                                                start_time => $policy->{last_vs_assoc_update_time})
           ) {

            my $mcp_asm_policies = F5::ASM::COAPI::Query->new('asm_policy',{name => $policy_name,
                                                                            strict_app_updates => 0,
                                                                           });
            if ( !($mcp_asm_policies && ref($mcp_asm_policies) eq 'ARRAY' && scalar(@$mcp_asm_policies)) ) {
                # iApp policies are expected not to be found
                next;
            }

            my $mcp_asm_policy = $mcp_asm_policies->[0];
            if (!$mcp_asm_policy->{is_active_policy}) { # already inactive in mcp, but should not be
                F5::BigIpLog::bigip_ts_ts_error_msg_log("Active Policy Sweeper: ASM security policy named '$policy_name' is active in ASMConfig but inactive in MCP");
                next;
            }

            # async de-activate by tsconfd
            $mcp_asm_policy->{is_active_policy} = 0;
            # we could be trying to deactivate a policy that is assigned to a L7 policy but not to a VS
            #     in which case it will fail, we should visualize this clearly in a log
            eval {
                $mcp_asm_policy->save();
            };
            if (my $err = $@) {
                if ($self->{logger}) {
                    $self->{logger}->info("Could not asynchronously deactivate ASM security policy named '$policy_name' (it's probably referenced by an LTM (L7) policy or an iRule)");
                    $self->{logger}->trace($err);
                }
            }
            else {
                F5::BigIpLog::bigip_ts_ts_info_msg_log("Active Policy Sweeper: asynchronously deactivated ASM security policy named '$policy_name'");
            }
        }
    }

    return;
}

#----------------------------------------------------------------------

1;
