package F5::PeriodicTask::AutoApplyPolicy;

use strict;
use warnings;
use base qw(F5::PeriodicTask::Base);

use F5::Db::Single;
use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::PBConfig;
use F5::ASMConfig::Cache;
use F5::Cluster::Utils;

use POSIX qw(strftime);
use Time::Local;

#----------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;
    $self->SUPER::init(%params);
    $self->get_policies(%params);

    my $pbconfig_fields = F5::ASMConfig::Entity::PBConfig->fields;

    my @auto_apply_fields = ();
    for my $pbconfig_field (keys %{$pbconfig_fields}) {
        if ($pbconfig_fields->{$pbconfig_field}->{rest_group} &&
            $pbconfig_fields->{$pbconfig_field}->{rest_group} eq 'autoApply') {

            push(@auto_apply_fields,$pbconfig_field);
        }
    }

    $self->{auto_apply_fields} = \@auto_apply_fields;
}

#----------------------------------------------------------------------

my %WEEK_DAYS = (
                 0 => 'sundays',
                 1 => 'mondays',
                 2 => 'tuesdays',
                 3 => 'wednesdays',
                 4 => 'thursdays',
                 5 => 'fridays',
                 6 => 'saturdays',
                );

sub module_run {
    my ($self, %params) = @_;
    my $client = $params{client};
    my $session = $client->{session};

    if (F5::Cluster::Utils::is_failover_standby()) {
        return;
    }

    my $dbh = F5::Db::Single->instance();

    my $now = strftime "%H:%M", localtime;
    for my $policy_id (keys %{$self->{policies}}) {

        my $policy = $self->{policies}->{$policy_id};
        my $pbconfig = $self->{policies_pbconfig}->{$policy_id};

        if ( $pbconfig && $pbconfig->{auto_apply_frequency} != F5::ASMConfig::AutoApplyFrequency::AUTO_APPLY_DISABLED() && $policy->is_modified_by_pb() ) {

            my $policy_display_name = $policy->get_display_name();
            $self->{logger}->trace("$policy_display_name is modified by Policy Builder.") if $self->{logger};

            if ( $pbconfig->{auto_apply_frequency} == F5::ASMConfig::AutoApplyFrequency::AUTO_APPLY_SCHEDULED() ) {
                my $start_time = $pbconfig->{auto_apply_start_time};
                my $end_time = $pbconfig->{auto_apply_end_time};
                my $today = (localtime)[6];
                my $today_flg = 'flg_auto_apply_on_' . $WEEK_DAYS{$today};
                my $out_of_range_log = "Current time is out of range according to scheduled apply policy settings -- policy $policy_display_name will not be applied";
                if ( !$pbconfig->{flg_auto_apply_at_all_times} ) {
                    if ( ($start_time lt $end_time) && (($start_time gt $now) || ($now gt $end_time)) ) {
                        $self->{logger}->debug($out_of_range_log) if $self->{logger};
                        next;
                    }
                    elsif ( ($start_time gt $end_time) && (($start_time gt $now) && ($now gt $end_time)) ) {
                        $self->{logger}->debug($out_of_range_log) if $self->{logger};
                        next;
                    }
                }
                if ( !$pbconfig->{flg_auto_apply_on_all_days} ) {
                    # start > end is an overnight case - we need to check that yesterday flag is on
                    if ( !$pbconfig->{flg_auto_apply_at_all_times} && ($start_time gt $end_time) && ($now lt $end_time) ) {
                        my $yesterday = ($today == 0) ? 6 : $today - 1;
                        my $yesterday_flg = 'flg_auto_apply_on_' . $WEEK_DAYS{$yesterday};
                        if ( !$pbconfig->{$yesterday_flg} ) {
                            $self->{logger}->debug($out_of_range_log) if $self->{logger};
                            next;
                        }
                    }
                    elsif ( !$pbconfig->{$today_flg} ) {
                        $self->{logger}->debug($out_of_range_log) if $self->{logger};
                        next;
                    }
                }
            }
            $self->{logger}->info("Apply policy $policy_display_name");
            eval {
                $client->set_active($session, F5::ASMConfig::CallFlags->new(), $policy, F5::ASMConfig::SetActiveFlags->new());
            };
            if (my $err = $@) {
                if (ref($err) && (ref($err) eq 'F5::ASMConfig::ASMConfigException')) {
                       $err = $err->{display_message} || $err->{error_message};
                }
                $self->{logger}->info("Could not apply policy $policy_display_name ($err)") if $self->{logger};
            }
        }
    }

    return;
}
#----------------------------------------------------------------------

sub get_subscriptions {
    return [F5::ASMConfig::EventSubscription->new({element_event_types => [ F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                  element_type => F5::ASMConfig::ElementType::POLICY}),
                                                                            F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                  element_type => F5::ASMConfig::ElementType::PBCONFIG}),
                                                                            F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                  element_type => F5::ASMConfig::ElementType::UCS_CONFIG_LOAD}),
                                                                            F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                  element_type => F5::ASMConfig::ElementType::SET_ACTIVE}),
                                                                          ],
                                                   send_topology_suppressed_events => 1,
                                                  })
           ];
}

#----------------------------------------------------------------------

sub handle_event {
    my ($self, %params) = @_;
    my $event = $params{event};
    my $element = $event->{element};
    my $dbh = F5::Db::Single->instance();
    my $client = $params{client};

    if ($element) {
        if ( $element->{element_type} == F5::ASMConfig::ElementType::POLICY ) {
            my $event_policy = $element->{policy};
            if ( $event->{event_type} == F5::ASMConfig::EventType::ADD() && $event_policy->{is_learning_enabled}
                 && $event_policy->{is_pb_enabled} && $event_policy->{is_active_policy} ) {
                my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $dbh, get_criteria => { policy_id => $event_policy->{policy_id} });
                my $pbconfig = F5::ASMConfig::Entity::PBConfig->new(dbh => $dbh, policy => $event_policy, get_criteria => {} );
                if ( $policy && $pbconfig ) {
                    $self->{policies}->{$event_policy->{policy_id}} = $policy;
                    $self->{policies_pbconfig}->{$event_policy->{policy_id}} = $pbconfig;
                }
            }
            elsif ( $event->{event_type} == F5::ASMConfig::EventType::DELETE() && $self->{policies}->{$event_policy->{policy_id}} ) {
                delete $self->{policies}->{$event_policy->{policy_id}};
                delete $self->{policies_pbconfig}->{$event_policy->{policy_id}};
            }
        }
        elsif ( $element->{element_type} == F5::ASMConfig::ElementType::SET_ACTIVE ) {
            my $event_policy = $event->{policy};
            my $cache_policy = $self->{policies}->{$event_policy->{policy_id}};
            if ( $event_policy->{is_active_policy} && $event_policy->{is_learning_enabled} && $event_policy->{is_pb_enabled} ) {
                my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $dbh, get_criteria => { policy_id => $event_policy->{policy_id} });
                my $pbconfig = F5::ASMConfig::Entity::PBConfig->new(dbh => $dbh, policy => $event_policy, get_criteria => {} );
                if ( $policy && $pbconfig ) {
                    $self->{policies}->{$event_policy->{policy_id}} = $policy;
                    $self->{policies_pbconfig}->{$event_policy->{policy_id}} = $pbconfig;
                }
            }
            elsif ( !$event_policy->{is_active_policy} && $cache_policy ) {
                delete $self->{policies}->{$event_policy->{policy_id}};
                delete $self->{policies_pbconfig}->{$event_policy->{policy_id}};
            }
        }
        elsif ( $element->{element_type} == F5::ASMConfig::ElementType::PBCONFIG ) {
            my $event_pbconfig = $element->{pbconfig};
            my $cache_pbconfig = $self->{policies_pbconfig}->{$event_pbconfig->{policy_id}};

            if ( $cache_pbconfig ) {
                if ( F5::ASMConfig::Entity::PBConfig->has_value_changed( existing_obj => $cache_pbconfig,
                                                                         store_obj => $event_pbconfig,
                                                                         check_fields => ['flg_enabled'], )) {

                    delete $self->{policies}->{$event_pbconfig->{policy_id}};
                    delete $self->{policies_pbconfig}->{$event_pbconfig->{policy_id}};
                } elsif ( F5::ASMConfig::Entity::PBConfig->has_value_changed( existing_obj => $cache_pbconfig,
                                                                              store_obj => $event_pbconfig,
                                                                              check_fields => $self->{auto_apply_fields}, )) {
                    $self->{policies_pbconfig}->{$event_pbconfig->{policy_id}} = $event_pbconfig;
                }
            }
            elsif ( $event_pbconfig->{flg_enabled} && $event->{policy}->{is_active_policy} ) {
                my $policy = F5::ASMConfig::Entity::Policy->new( dbh => $dbh, get_criteria => {policy_id => $event_pbconfig->{policy_id}} );
                if ( $policy ) {
                    $self->{policies}->{$policy->{policy_id}} = $policy;
                    $self->{policies_pbconfig}->{$policy->{policy_id}} = $event_pbconfig;
                }
            }
        }
        elsif ( $element->{element_type} == F5::ASMConfig::ElementType::UCS_CONFIG_LOAD ) {
            $self->get_policies(%params);
        }
    }
}

#----------------------------------------------------------------------

sub get_policies {
    my ($self, %params) = @_;
    my $dbh = F5::Db::Single->instance();
    my @policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $dbh,
                                                               get_criteria => { 'is_learning_enabled' => 1,
                                                                                 'is_pb_enabled' => 1,
                                                                                 'is_active_policy' => 1,
                                                                               },
                                                               selected_fields => [],
                                                              );
    my %policies_by_id = map { $_->{policy_id} => $_ } @policies;
    $self->{policies} = \%policies_by_id;
    for my $policy (values %{$self->{policies}}) {
        $self->{policies_pbconfig}->{$policy->{policy_id}} = F5::ASMConfig::Entity::PBConfig->new(dbh => $dbh,
                                                                                                  policy => $policy,
                                                                                                  get_criteria => {},
                                                                                                 );
    }
}

#----------------------------------------------------------------------

1;
