package F5::PeriodicTask::Webhook;

use strict;
use warnings;
use base qw(F5::PeriodicTask::Base);
use F5::Db::Single;
use F5::ASMConfig::Entity::Webhook;
use F5::ASMConfig::Entity::RequestLog;
use F5::ASMConfig::Entity::Base;
use constant WEBHOOKS_CFG_FILE  =>  '/etc/ts/tools/webhooks.cfg';

our $MAX_MESSAGE_PER_URL_PER_WEBHOOK_EVENT = F5::Cfg::cfg_get_config_item(WEBHOOKS_CFG_FILE, 'webhooks', 'max_message_per_url_per_webhook_event') || 10;

my %policies_webhook_event_hash = ();
my %policies_webhook_values = ();

my $webhook_url_events = {};

my $webhook_events_order = [F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_APPLY_POLICY_FAILED,
                            F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_APPLY_POLICY,
                            F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_INCIDENT_START,
                            F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_INCIDENT_END,
                            F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_LIKELY_MALICIOUS_HTTP_REQUEST,
                            F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_HTTP_REQUEST_NEEDING_EXAMINATION,
                            F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_ILLEGAL_HTTP_REQUEST];


my $webhook_event_viol_rating = {F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_ILLEGAL_HTTP_REQUEST              => 1,
                                 F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_HTTP_REQUEST_NEEDING_EXAMINATION  => 3,
                                 F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_LIKELY_MALICIOUS_HTTP_REQUEST     => 4 };

sub init {
    my ($self, %params) = @_;
    $self->{dbh} = F5::Db::Single->instance;
    $self->SUPER::init(%params);
}

#----------------------------------------------------------------------

sub load_webhooks {
    my ($self, %params) = @_;
    my $policy = $params{policy};

    my @webhooks = F5::ASMConfig::Entity::Webhook->retrieve_all(dbh => $self->{dbh}, get_criteria => { policy_id => $policy->{policy_id}});
    for my $webhook (@webhooks) {
        $self->{logger}->trace($webhook);
        $policies_webhook_event_hash{$webhook->{policy_id}}->{$webhook->{trigger_event}}->{$webhook->{webhook_id}} = $webhook;
    }
    $self->{logger}->debug(scalar @webhooks . " webhooks were loaded for policy id " . $policy->{policy_id}) if($self->{logger});
}

#----------------------------------------------------------------------

sub module_run {
    my ($self, %params) = @_;

    for my $webhook_url (keys %{$self->{webhook_url_events}}) {
        for my $webhook_event (@{$webhook_events_order}) {
            if (!$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}) {
                next;
            }
            my $item = shift(@{$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}});
            my $webhook_id = $item->{webhook_id};
            my $policy_id = $item->{policy}->{policy_id};
            my $webhook = $policies_webhook_event_hash{$policy_id}->{$webhook_event}->{$webhook_id};
            eval {
                 my $rv = $webhook->send(element => $item->{element},
                                         policy => $item->{policy});
                 if (!$rv->is_success) {
                    $self->{logger}->error("Failed to send webhook: ");
                    $self->{logger}->error($rv->status_line);
                 }
            };
            if (my $err = $@) {
                $self->{logger}->error("Send method for webhook failed: ");
                $self->{logger}->error($err);
            }

            if (!@{$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}}) {
                delete $self->{webhook_url_events}->{$webhook_url}->{$webhook_event};
            }
            last;
        }

        if (!keys %{$self->{webhook_url_events}->{$webhook_url}}) {
            delete $self->{webhook_url_events}->{$webhook_url};
        }
    }
}

#----------------------------------------------------------------------

sub get_subscriptions {
    my ($self, %params) = @_;
    return [F5::ASMConfig::EventSubscription->new({element_event_types => [F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                 element_type => F5::ASMConfig::ElementType::SET_ACTIVE}),
                                                                           F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                 element_type => F5::ASMConfig::ElementType::WEBHOOK}),
                                                                           F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                 element_type => F5::ASMConfig::ElementType::INCIDENT}),
                                                                           F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                 element_type => F5::ASMConfig::ElementType::POLICY}),
                                                                           F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                 element_type => F5::ASMConfig::ElementType::REQUEST_LOG}),
                                                                           F5::ASMConfig::ElementEventType->new({event_type => F5::ASMConfig::EventType::ALL_TYPES,
                                                                                                                 element_type => F5::ASMConfig::ElementType::UCS_CONFIG_LOAD})],
                                                   send_topology_suppressed_events => 1,
                                                   send_catchup_events => 0,
                                                  })
           ];
}

#----------------------------------------------------------------------

sub handle_request_log_event {
    my ($self, %params) = @_;
    my $event = $params{event};
    my $search_result = $event->{search_result};
    my $element_list = $search_result ? $search_result->{element_list} : ($event->{element});

    my $url_needed_sort_list;
    for my $element (@{$element_list}) {
        my $request_log = $element->{request_log};
        my $policy = $request_log->{request_policy};
        my $policy_id = $policy->{policy_id};
        my $rating = $request_log->{enforcement_state_rating};

        my @events = grep { $rating >= $webhook_event_viol_rating->{$_} } keys %{$webhook_event_viol_rating};
        for my $webhook_event (@events) {
            next if !$policies_webhook_event_hash{$policy_id}->{$webhook_event};
            for my $webhook_id (keys %{$policies_webhook_event_hash{$policy_id}->{$webhook_event}}) {
                my $webhook = $policies_webhook_event_hash{$policy_id}->{$webhook_event}->{$webhook_id};
                my $webhook_url = $webhook->{webhook_url};
                if (!$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}) {
                    push (@{$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}},{element => $element, webhook_id => $webhook_id, policy => $policy});
                }
                elsif ((scalar @{$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}}) < $MAX_MESSAGE_PER_URL_PER_WEBHOOK_EVENT) {
                    push (@{$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}},{element => $element, webhook_id => $webhook_id, policy => $policy});
                    if (!$url_needed_sort_list->{$webhook_url}) {
                        $url_needed_sort_list->{$webhook_url} = [];
                    }
                    push (@{$url_needed_sort_list->{$webhook_url}}, $webhook_event);
                }
            }
        }
    }

    for my $webhook_url (keys %{$url_needed_sort_list}) {
        for my $webhook_event (@{$url_needed_sort_list->{$webhook_url}}) {
            my @sorted_logs = sort { $a->{element}->{request_log}->{log_time} <=> $b->{element}->{request_log}->{log_time} } @{$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}};
            $self->{webhook_url_events}->{$webhook_url}->{$webhook_event} = \@sorted_logs;
        }
    }
}

#----------------------------------------------------------------------
sub handle_event {
    my ($self, %params) = @_;
    my $event = $params{event};
    my $element = $event->{element};
    my $event_type = $event->{event_type};
    my $policy = $event->{policy};

    if (!$element) {
        return;
    }

    $self->{logger}->debug("Webhook handle_event Element:");
    $self->{logger}->debug($element);

    if ($element->{element_type} == F5::ASMConfig::ElementType::POLICY) {
        if ($event->{event_type} == F5::ASMConfig::EventType::ADD) {
            $self->load_webhooks(policy => $policy);
        }
        elsif ($event->{event_type} == F5::ASMConfig::EventType::DELETE) {
            delete $policies_webhook_event_hash{$event->{policy}->{policy_id}};
        }
    }
    elsif ($element->{element_type} == F5::ASMConfig::ElementType::WEBHOOK) {
        my $webhook = $element->{webhook};
        if ($event_type == F5::ASMConfig::EventType::ADD || $event_type == F5::ASMConfig::EventType::UPDATE) {
            $self->{logger}->debug("Add/update webhook:");
            $self->{logger}->debug($webhook);

            if ($event_type == F5::ASMConfig::EventType::UPDATE) {
                for my $trigger_event_id (keys %{$policies_webhook_event_hash{$policy->{policy_id}}}) {
                    if (exists $policies_webhook_event_hash{$policy->{policy_id}}->{$trigger_event_id}->{$webhook->{webhook_id}}) {
                        delete $policies_webhook_event_hash{$policy->{policy_id}}->{$trigger_event_id}->{$webhook->{webhook_id}};
                        last;
                    }
                }
            }

            my $new_webhook = F5::ASMConfig::Entity::Webhook->new(dbh => $self->{dbh}, get_criteria => {webhook_id => $webhook->{webhook_id}} );
            $policies_webhook_event_hash{$policy->{policy_id}}->{$webhook->{trigger_event}}->{$webhook->{webhook_id}} = $new_webhook;
        }
        elsif ($event_type == F5::ASMConfig::EventType::DELETE) {
            $self->{logger}->debug("Delete webhook:");
            $self->{logger}->debug($webhook);
            delete $policies_webhook_event_hash{$policy->{policy_id}}->{$webhook->{trigger_event}}->{$webhook->{webhook_id}};
        }
    }
    elsif (($element->{element_type} == F5::ASMConfig::ElementType::REQUEST_LOG) && ($event->{event_type} == F5::ASMConfig::EventType::ADD)) {
        $self->handle_request_log_event(event => $event);
    }
    elsif ($element->{element_type} == F5::ASMConfig::ElementType::UCS_CONFIG_LOAD) {
        $self->handle_ucs_config_load_event(event => $event);
    }

    my $webhook_event;
    my $policy_id;

    if ($element->{element_type} == F5::ASMConfig::ElementType::INCIDENT) {
        my $incident = $element->{incident};
        my $incident_status = $incident->{incident_status};
        my $incident_policy = $incident->{incident_policy};
        $policy_id = $incident_policy->{policy_id};

        if ($event_type == F5::ASMConfig::EventType::ADD) {
            $webhook_event = F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_INCIDENT_START;
        }
        elsif (($event_type == F5::ASMConfig::EventType::UPDATE) && ($incident_status == F5::ASMConfig::IncidentStatus::INCIDENT_CLOSED)) {
            $webhook_event = F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_INCIDENT_END;
        }

    }
    elsif ($element->{element_type} == F5::ASMConfig::ElementType::SET_ACTIVE) {
        if (($event_type == F5::ASMConfig::EventType::ADD) || ($event_type == F5::ASMConfig::EventType::UPDATE)) {
            if (! $element->{set_active}->{deactivate_policy}) {
                $webhook_event = F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_APPLY_POLICY;
                my %policy_webhook_values = map { $_ => $policy->{$_} } keys %{F5::ASMConfig::Entity::Policy->webhook_fields};
                $policies_webhook_values{$policy->{policy_id}} = \%policy_webhook_values;
            }
        }
        elsif ($event_type == F5::ASMConfig::EventType::ERROR) {
            $webhook_event = F5::ASMConfig::WebhookEvent::WEBHOOK_EVENT_APPLY_POLICY_FAILED;
        }
        $policy_id = $policy->{policy_id};
    }

    if (defined $webhook_event && defined $policy_id) {
        $self->{logger}->debug("Get webhooks with trigger event $webhook_event, policy_id = $policy_id");
        $self->handle_webhooks_event(policy_id => $policy_id, webhook_event => $webhook_event, element => $element);
    }
}

#----------------------------------------------------------------------
sub handle_webhooks_event {
    my ($self, %params) = @_;
    my $policy_id = $params{policy_id};
    my $webhook_event = $params{webhook_event};
    my $element = $params{element};

    if ($policies_webhook_event_hash{$policy_id} && $policies_webhook_event_hash{$policy_id}->{$webhook_event}) {
        for my $webhook (values %{$policies_webhook_event_hash{$policy_id}->{$webhook_event}}) {
            my $webhook_url = $webhook->{webhook_url};
            my $webhook_id = $webhook->{webhook_id};

            if ( $self->{webhook_url_events}->{$webhook_url}->{$webhook_event}
                  && (scalar @{$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}}) >= $MAX_MESSAGE_PER_URL_PER_WEBHOOK_EVENT ) {
                next;
            }

            my $policy = F5::ASMConfig::Policy->new(policy_id => $policy_id,
                                                    %{$policies_webhook_values{$policy_id}});

            push (@{$self->{webhook_url_events}->{$webhook_url}->{$webhook_event}}, { element => $element,
                                                                                      webhook_id => $webhook_id,
                                                                                      policy => $policy });
        }
    }
}

#----------------------------------------------------------------------
sub handle_ucs_config_load_event {
    my ($self, %params) = @_;

    %policies_webhook_event_hash = ();

    my @policies = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $self->{dbh});
    for my $policy (@policies) {
        $self->{logger}->debug("Loading Webhooks for policy id " . $policy->{policy_id}) if ($self->{logger});
        $self->load_webhooks(policy => $policy);
    }
}

#----------------------------------------------------------------------
1;

