package F5::PeriodicTaskManager;

use strict;
use warnings;
use F5::ModuleLoader;
use F5::PeriodicTask::Base;
use F5::DbUtils;
use F5::DateTimeUtils;
use List::Util;
use F5::Db::Single;
use F5::ASMConfig::ASMConfig;
use F5::IO::Selector;
use F5::IO::ASMConfig::EventListener;
use F5::Cfg;

our @MODULES = F5::ModuleLoader::load_from_base_class('F5::PeriodicTask::Base');
use constant ASMCROND_CFG_FILE  =>  '/etc/ts/tools/asmcrond.cfg';

our $CONNECT_ATTEMPT_MINIMUM_INTERVAL = F5::Cfg::cfg_get_config_item(ASMCROND_CFG_FILE, 'asmcrond', 'connect_attempt_minimum_interval') || 5;


#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;

    my $self = bless({}, $class);
    $self->init(config_file => ASMCROND_CFG_FILE,
                %params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->{dbh} = $params{dbh};
    $self->{logger} = $params{logger};
    $self->{config_file} = $params{config_file};
    $self->{selector} = F5::IO::Selector->new;
    $self->{last_connect_time} = 0;
    $self->{last_run_time} = 0;
    for my $module (@MODULES) {
        $self->{instances}->{$module} = $module->new(%params);
    }

    $self->{pulse_msg_interval} = F5::Cfg::cfg_get_config_item($self->{config_file}, 'asmcrond', 'pulse_msg_interval') || 30;
    $self->{modules_run_count} = 0;
    $self->{events_received_count} = 0;
}

#----------------------------------------------------------------------

sub run {
    my ($self) = @_;
    my $last_pulse_time = time;

    $self->open_asmconfig_client();
    for my $module (@MODULES) {
        $self->{instances}->{$module}->run(client => $self->{asmconfig_event_listener}->get_client());
        $self->{modules_run_count}++;
        $self->{next_run_intervals}->{$module} = $self->{instances}->{$module}->next_run_interval();
    }

    while (1) {
        my $poll_time = undef;
        my @potential_poll_times;

        $self->open_asmconfig_client();
        next unless $self->{asmconfig_event_listener};

        my @next_run_intervals = values(%{$self->{next_run_intervals}});
        my $next_run_interval = List::Util::min(@next_run_intervals);
        push(@potential_poll_times, F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $next_run_interval,
                                                                                       minimum_interval => 1,
                                                                                       start_time => $self->{last_run_time}));

        if ($self->{pulse_msg_interval}) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                                    start_time => $last_pulse_time));
        }

        if (@potential_poll_times) {
            $poll_time = List::Util::min(@potential_poll_times);
        }

        my @readers = $self->{selector}->can_read($poll_time);
        for my $reader (@readers) {
            $reader->handle_read;
        }

        $self->open_asmconfig_client();
        for my $module_name (@MODULES) {
            my $module = $self->{instances}->{$module_name};
            $self->{next_run_intervals}->{$module_name} = $module->next_run_interval();
            if ($self->{next_run_intervals}->{$module_name} == 0) {
                $self->{instances}->{$module_name}->run( client => $self->{asmconfig_event_listener}->get_client(), );
                $self->{next_run_intervals}->{$module_name} = $module->next_run_interval();
                $self->{modules_run_count}++;
            }
        }

        if ($self->{pulse_msg_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                               start_time => $last_pulse_time) == 0) {

            $self->log_pulse_message;
            $last_pulse_time = time;
        }
        $self->{last_run_time} = time();
    }
}

#----------------------------------------------------------------------

sub open_asmconfig_client {
    my ($self) = @_;

    if (!$self->{asmconfig_event_listener}) {
        my @subscriptions;
        for my $module (@MODULES) {
            push (@subscriptions, @{$self->{instances}->{$module}->get_subscriptions()});
        }

        while (!$self->{asmconfig_event_listener}) {
            $self->{logger}->debug("Connecting to ASMConfig");
            my $time_remaining = F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $CONNECT_ATTEMPT_MINIMUM_INTERVAL,
                                                                                    start_time => $self->{last_connect_time});
            if ($time_remaining) {
                sleep ($time_remaining);
            }

            $self->{asmconfig_event_listener} = F5::IO::ASMConfig::EventListener->new(event_subscriptions => \@subscriptions,
                                                                                      logger => $self->{logger},
                                                                                      read_message_handler => sub {
                                                                                          $self->handle_asmconfig_event(event => $_[1]);
                                                                                      },
                                                                                      socket_close_handler => sub {
                                                                                          my ($handle, $errno) = @_;
                                                                                          $self->{logger}->info("Event listener socket close received") if $handle->{logger};
                                                                                          $handle->{selector}->remove(shift) if $handle->{selector};
                                                                                          $handle->close if $handle;
                                                                                          delete $self->{asmconfig_event_listener};
                                                                                      },
                                                                                      selector => $self->{selector},
                                                                                     );
            $self->{last_connect_time} = time();
        }
    }
}


#----------------------------------------------------------------------

sub handle_asmconfig_event {
    my($self, %params) = @_;

    my $event = $params{event};

    $self->{logger}->trace("asmconfig event received") if $self->{logger};
    $self->{logger}->trace($event) if $self->{logger};
    $self->{events_received_count}++;

    for my $module (@MODULES) {
        $self->{instances}->{$module}->handle_event(event => $event,
                                                  client => $self->{asmconfig_event_listener}->get_client(),
                                                  );
        $self->{next_run_intervals}->{$module} = $self->{instances}->{$module}->next_run_interval();
    }
    my $element = $event->{element};
    if ($element && 
        ($element->{element_type} == F5::ASMConfig::ElementType::UCS_CONFIG_LOAD ||
         ($element->{element_type} == F5::ASMConfig::ElementType::CLUSTER_STATE && $element->{cluster_state}->{is_master_blade}))
       ) {
       $self->{asmconfig_event_listener}->reinit_subscriptions();
    }
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;

    $self->{logger}->info("$self->{modules_run_count} modules run, $self->{events_received_count} events received since startup") if $self->{logger};
}

#----------------------------------------------------------------------

1;

