package F5::Policy;

use strict;
use warnings;
use Carp;

use F5::ASMConfig::ASMConfig;
use F5::ASMConfig::Entity::Policy;
use F5::PolicyHistory;
use F5::SetActive;
use F5::Syslog;
use List::Util;

our $AUTOLOAD;
my $UCS_V2_BIGIP_VERSION = '0009.0004.0000.0000.0000';

my %other_table_fields = ( 'flg_quarantine_signatures' => 'PL_POLICY_NEGSIG_ATTRIBUTES',
                           'quarantined_period' => 'PL_POLICY_GENERAL',
                         );

sub new {
  my($class, %params) = @_;
  my $self = bless {}, $class;

  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{policy_id} = $params{policy_id};
  $self->{_dbh} = $params{dbh};
  $self->{policy_name} = $params{policy_name};
  $self->{dcc_dbname} = $params{dcc_dbname} || 'DCC';
  $self->{plc_dbname} = $params{plc_dbname} || 'PLC';

  if (($self->{policy_id} || $self->{policy_name}) && $self->{_dbh}) {

    my($search_field, $search_value);
    if ($self->{policy_id}) {
      $search_field = 'id';
      $search_value = $self->{policy_id};
    }
    if ($self->{policy_name}) {
      $search_field = 'name';
      $search_value = $self->{policy_name};
    }

    $self->{_db_rec} = $self->{_dbh}->selectrow_hashref(<<EndQuery, undef, $search_value);
SELECT pol.*, negsig_attrib.*
  FROM $self->{plc_dbname}.PL_POLICIES pol,
       $self->{plc_dbname}.PL_POLICY_NEGSIG_ATTRIBUTES negsig_attrib
 WHERE $search_field = ?
   AND pol.id = negsig_attrib.policy_id
EndQuery

    $self->{_orig_db_rec} = {};
    if ($self->{_db_rec} && %{$self->{_db_rec}}) {
        $self->{_orig_db_rec} = { %{$self->{_db_rec}} };
    }
  }
}

#----------------------------------------------------------------------

sub detach_all_negsig_sets {
    my($self) = @_;
    
    $self->{_dbh}->do("DELETE FROM PLC.PL_POLICY_NEGSIG_SETS WHERE policy_id = ?", undef, $self->policy_id);
}

#----------------------------------------------------------------------

sub clear_modification_log {
  my($self, %params) = @_;
  my @where_clauses = ('policy_id = ?');
  my @bind_values = ($self->policy_id);

  if ($params{max_modification_log_id}) {
    push(@where_clauses, 'policy_changes_log_id <= ?');
    push(@bind_values, $params{max_modification_log_id});
  }
  my $where_clauses = 'WHERE ' . join(' AND ', @where_clauses);

  $self->{_dbh}->do(<<EndQuery, undef, @bind_values);
DELETE FROM $self->{plc_dbname}.PL_POLICY_CHANGES_LOG 
$where_clauses
EndQuery

}

#----------------------------------------------------------------------

sub get_max_modification_log_id {
  my($self) = @_;
  my($max_modification_log_id) = $self->{_dbh}->selectrow_array(<<EndQuery, undef, $self->policy_id);
SELECT MAX(policy_changes_log_id)
  FROM $self->{plc_dbname}.PL_POLICY_CHANGES_LOG
 WHERE policy_id = ?
EndQuery

  return $max_modification_log_id;
}

#----------------------------------------------------------------------

sub is_modified {
  my($self) = @_;
  return $self->{_dbh}->selectrow_arrayref("SELECT MAX(update_time) FROM $self->{plc_dbname}.PL_POLICY_CHANGES_LOG WHERE policy_id = ?", 
					   undef, $self->policy_id)->[0] || 0;
}

#----------------------------------------------------------------------

sub update_db {
  my($self) = @_;

  local $" = ',';

  my @modified_fields;
  for my $field (keys %{$self->{_orig_db_rec}}) {
    if (exists $self->{_orig_db_rec}->{$field} && 
	exists $self->{_db_rec}->{$field} &&
        ((defined $self->{_orig_db_rec}->{$field} && !defined $self->{_db_rec}) ||
         (!defined $self->{_orig_db_rec}->{$field} && defined $self->{_db_rec}) ||
         (defined $self->{_orig_db_rec}->{$field} && defined $self->{_db_rec} &&
          $self->{_orig_db_rec}->{$field} ne $self->{_db_rec}->{$field}))) {
      push(@modified_fields, $field);
    }
  }

  if (@modified_fields) {
    my %table_updates = ( 'PL_POLICIES' => [ grep { ! exists $other_table_fields{$_} } @modified_fields ]);

    for my $modified_field (@modified_fields) {
      if (exists $other_table_fields{$modified_field}) {
        my $other_table_name = $other_table_fields{$modified_field};
        unless (exists $table_updates{$other_table_name}) {
          $table_updates{$other_table_name} = [];
        }
        push(@{$table_updates{$other_table_name}}, $modified_field);
      }
    }

    while (my($update_table_name, $update_fields) = each %table_updates) {
      next unless (@$update_fields);
      my $policy_id_field = ($update_table_name eq 'PL_POLICIES' ? 'id' : 'policy_id');

      my @update_field_assignments = map("$_ = ?", @$update_fields);
      $self->{_dbh}->do(<<EndQuery, undef, @{$self->{_db_rec}}{@$update_fields}, $self->policy_id);
UPDATE $self->{plc_dbname}.$update_table_name 
   SET @update_field_assignments
 WHERE $policy_id_field = ?
EndQuery
    }

    $self->{_orig_db_rec} = { %{$self->{_db_rec}} };
  }
}

#----------------------------------------------------------------------

sub history {
  my($self) = @_;
  return F5::PolicyHistory->new($self);
}

#----------------------------------------------------------------------

sub attach_or_update_negsig_set {
    my ($self, $params) = @_;
    
    my $sql = <<EndQuery;
REPLACE INTO PLC.PL_POLICY_NEGSIG_SETS 
        (policy_id, set_id, flg_learn, flg_alarm, flg_reject)
 VALUES (?,?,?,?,?)
EndQuery

    $self->{_dbh}->do($sql, undef, $self->policy_id, $params->{set_id}, $params->{flg_learn}, $params->{flg_alarm}, $params->{flg_reject});
}

#----------------------------------------------------------------------

sub detach_negsig_set {
    my ($self, $params) = @_;

    my $sql = <<EndQuery;
DELETE FROM PLC.PL_POLICY_NEGSIG_SETS
 WHERE policy_id = ? 
   AND set_id = ?
EndQuery

    $self->{_dbh}->do($sql, undef, $self->policy_id, $params->{set_id});
}

#----------------------------------------------------------------------

sub get_negsig_sets {
    my ($self) = @_;

    return $self->{_dbh}->selectall_hashref(<<EndQuery, 'set_id', undef, $self->policy_id);
SELECT *
  FROM PLC.PL_POLICY_NEGSIG_SETS
 WHERE policy_id = ?
EndQuery
}

#----------------------------------------------------------------------

sub set_active_from_history {
  my($self, %params) = @_;
  my $dbh = $self->get_dbh;

  my $restore_policy_name = $self->policy_name . "_restore_for_set_active_$$";
  my $started_local_transaction = 0;
  my $last_active_version;
  my $is_modified = $self->is_modified;

  if ($is_modified && !$params{has_parent} &&
      (!$params{ucs_version} || $params{ucs_version} ge $UCS_V2_BIGIP_VERSION)) {
    if ($dbh->{AutoCommit}) {
      $dbh->begin_work;
      $started_local_transaction = 1;
    }
    # look for policy history
    $last_active_version = $self->history->restore(revision_to_restore => -1, 
                                                   new_policy_name => $restore_policy_name,
                                                   is_within_ucs_upgrade => $params{ucs_version},
                                                   external_idmap => $params{external_idmap},
                                                   base_dir       => $params{base_dir},
                                                  );
    # Policy history files only appear from Sonoma onward 
    #    - but they do not have to be present. a ucs could have been created with only one active policy on the box
    if (!$last_active_version) {
      F5::Syslog::ts_syslog( 'info',
                           "Policy history (for policy %s) was not found while installing ucs file. Activating current version of the policy.",
                           $self->policy_name
                         );
    }

    if ($started_local_transaction) {
      $dbh->commit;
    }
  }
  else {
      $params{logger}->info("Activating current version of policy " . $self->policy_name) if $params{logger};
  }

  my $username = $self->history->get_last_username();
  my $rv = 1;

  # policy history was found
  if ($last_active_version) {
    $rv = F5::SetActive::set_active(
                                dbh => $dbh,
                                logger => $params{logger},
                                policy => $last_active_version,
                                account_id => $self->account_id,
                                policy_name => $self->policy_name,
                                force_policy_name_flg => 1,
                                transaction_start_time => $params{transaction_start_time},
                                username => $username,
                                send_sync => 0,
                                use_daemon => $params{use_daemon},
                                do_update_core => 0, # Only perform a single update core at the very end
                                do_hbcl => 0,
                                do_transaction => 0,
                                ucs_upgrade_real_policy_id => $self->policy_id,  #For id lookups
                                asm_config_handler => $params{asm_config_handler},
                                do_save_policy_history => 0,
                               );
    # Clear current version's history record
    $self->history->clear_last_revision_end_time;
    $last_active_version->delete(skip_recycle_bin => 1);
  }
  # policy history was not found
  else {
    $rv = F5::SetActive::set_active(
                                dbh => $dbh,
                                logger => $params{logger},
                                policy => $self,
                                transaction_start_time => $params{transaction_start_time},
                                username => $username,
                                do_update_core => 0, # Only perform a single update core at the very end
                                use_daemon => $params{use_daemon},
                                do_hbcl => 0,
                                send_sync => 0,
                                do_transaction => 0,
                                asm_config_handler => $params{asm_config_handler},
                                do_save_policy_history => 0,
                               );

  }

  return $rv;
}

#----------------------------------------------------------------------

sub export_to_file {
    my($self, $filename) = @_;
    require F5::ImportExportPolicy;
    my $import_export = F5::ImportExportPolicy->new(dbh => $self->{_dbh});
    $import_export->export_policy (
        policy_id       => $self->policy_id,
        filename        => $filename,
        include_scanner_tables => 1,
    );
}

#----------------------------------------------------------------------

sub import_from_file {
  my($self_or_class, @params) = @_;
  my($dbh, $filename, $account_id, %params);
  my $is_within_ucs_upgrade;

  if (scalar @params < 4) {
    ($dbh, $filename, $account_id) = @params;
  }
  else {
    %params = @params;
    $dbh = $params{dbh};
    $filename = $params{filename};
    $account_id = $params{account_id};
    $is_within_ucs_upgrade = $params{is_within_ucs_upgrade};
  }

  if (ref($self_or_class) && !$account_id) {
    $account_id = $self_or_class->account_id;
  }

  require F5::ImportExportPolicy;
  my $import_export = F5::ImportExportPolicy->new(dbh => $dbh);
  my $result = $import_export->import_policy (
                                              target_account_id       => $account_id,
                                              filename                => $filename,
                                              is_within_ucs_upgrade   => $is_within_ucs_upgrade,
                                              external_idmap => $params{external_idmap},
                                             );

  if ($result->{completed_successfully}) {
    return F5::Policy->new(dbh => $dbh,
		       policy_name => $result->{policy_name});
  }
  else {
    # Policy could not be created.
    return;
  }
}

#----------------------------------------------------------------------

sub delete {
  my($self, %params) = @_;
  my $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{_dbh},
                                                  get_criteria => { policy_id => $self->policy_id });

  $policy->delete(%params) if $policy;
}

#----------------------------------------------------------------------

sub get_dbh {
  my($self) = @_;
  return $self->{_dbh};
}

#----------------------------------------------------------------------

sub account_name {
  my($self) = @_;
  my($account_name) = $self->{_dbh}->selectrow_array("SELECT policy_name FROM $self->{dcc_dbname}.ACCOUNTS WHERE account_id = ?", 
						     undef, $self->account_id);
  return $account_name;
}

#======================================================================
# Accessor methods - AUTOLOAD

sub AUTOLOAD {
    my($self, $newval) = @_;

    my %field_map = (
        policy_id => "id",
        policy_name => "name",
    );
    my @readonly_fields = (qw/id/);

    my $name = $AUTOLOAD;
    $name =~ s/.*://;   # strip fully-qualified portion

    if (exists $field_map{$name}) {
        $name = $field_map{$name};
    }

    return if ($name eq 'DESTROY');

    if (exists $self->{_db_rec}->{$name}) {
        if (defined $newval) {
            if (grep {$_ eq $name} @readonly_fields) {
                die "Readonly field: $name\n";
            } else {
                $self->{_db_rec}->{$name} = $newval;
            }
        }
        return $self->{_db_rec}->{$name};
    }

    return; # undef if no such value found
}

#------------------------------------------------------------------
sub create_wildcard_objects {
  my($class_or_self,%params) = @_;
  my $dbh = $params{dbh};
  my $databases = $params{databases};
  my $policy_id = $params{policy_id};
  my $account_id = $params{account_id};
  my @object_ids;


  if(ref $class_or_self){
   $policy_id = defined $policy_id ? $policy_id : $class_or_self->policy_id;
  }

  for my $is_https ((0,1)) {
    my $object_id;

    F5::DbUtils::insert_row(dbh => $dbh,
                            table_name => "$databases->{PLC}.PL_OBJECTS",
                            rec => {
                                    policy_id => $policy_id,
                                    app_id => $is_https,
                                    uri => '*',
                                    flg_is_pattern => 1,
                                    flg_tightening => 0,
                                    pattern_order => 1
                                   },
                            last_insert_id => \$object_id,
                           );
    push(@object_ids, $object_id);
  }
  return @object_ids;
}



#------------------------------------------------------------------

1;

