package F5::PolicyHistory;

use strict;
use warnings;
use Carp;

use English;

use F5::Deploy;
use F5::Policy;
use F5::DbUtils;
use F5::ASMConfig::Entity::Policy;
use File::Path;
use File::Find;
use File::Copy;
use File::Basename;
use F5::Utils::File;
use F5::Cfg;

use constant DEFAULT_BASE_DIRECTORY => F5::Deploy::map_path('/ts/dms/policy/policy_versions');
use constant DEFAULT_FILENAME_TEMPLATE => '$POLICY_ID$/$VERSION$.plc';
use constant MAX_NUMBER_OF_SAVED_VERSIONS => 50;
use constant POLICY_HISTORY_CFG_FILE  =>  '/etc/ts/tools/policy_history.cfg';

my $MAX_SIZE_OF_SAVED_VERSIONS = F5::Cfg::cfg_get_config_item(POLICY_HISTORY_CFG_FILE, 'CleanPolicyHistory', 'maxSizeOfSavedVersions');
my $MIN_RETAINED_FILES_IN_DIR =  F5::Cfg::cfg_get_config_item(POLICY_HISTORY_CFG_FILE, 'CleanPolicyHistory', 'minRetainedFilesInDir') || 2;

sub new {
  my($class, $policy) = @_;
  my $self = bless {}, $class;

  $self->init($policy);
  return $self;
}


sub init {
  my($self, $policy) = @_;
  $self->{_policy} = $policy;
}

sub save {
  my ($self, @params) = @_;
  my $dbh = $self->{_policy}->get_dbh;

  my ($validation_report, $username, $save_policy_history_frequency_limit);
  if (scalar @params == 1) {
      $validation_report = $params[0];
  }
  else {
      my %params = @params;
      $validation_report = $params{validation_report};
      $username = $params{username};
      $save_policy_history_frequency_limit = $params{save_policy_history_frequency_limit};
  }

  my $changeover_time = time;
  if ($save_policy_history_frequency_limit && $username) {
      # Check if last save by this user was within save_policy_history_frequency_limit seconds
      my($previous_revision) = $dbh->selectrow_array(<<EndQuery, undef, $self->{_policy}->policy_id, $username, $changeover_time, $changeover_time - $save_policy_history_frequency_limit);
SELECT id
  FROM PLC.PL_POLICY_HISTORY
 WHERE policy_id = ?
   AND username = ?
   AND from_date < ?
   AND from_date > ?
EndQuery

      if ($previous_revision) {
          return;
      }
  }

  my $current_revision = $self->get_last_revision_number + 1;

  my $filename = $self->construct_filename(revision => $current_revision);

  mkpath(dirname($filename));
  $self->{_policy}->export_to_file($filename);

  #if an empty file has been created, delete it and return.
  if( -e $filename && -z $filename ){
      $self->{logger}->error("Failed saving policy history file.") if $self->{logger};
      unlink $filename;
      return;
  }

  # Set end time on previously active policy to now;
  my $exists_prev_active_revision = F5::DbUtils::exists_in_table(dbh => $dbh,
                                                                 field_name => 'to_date',
                                                                 field_value => 0,
                                                                 where_fields => { policy_id => $self->{_policy}->policy_id },
                                                                 table_name => 'PLC.PL_POLICY_HISTORY');
  if ($exists_prev_active_revision) {
      F5::DbUtils::update_row(dbh => $dbh,
                              table_name => 'PLC.PL_POLICY_HISTORY',
                              where_fields => { policy_id => $self->{_policy}->policy_id,
                                                to_date => 0, },
                              rec => { to_date => $changeover_time });
  }

  my $rec = {
            policy_id => $self->{_policy}->policy_id,
            revision => $current_revision,
            from_date => $changeover_time,
            to_date => 0,
            validation_report => $validation_report || '',
            username => $username || '',
            };

  F5::ASMConfig::Entity::Policy->populate_version_fields(policy => $rec, dbh => $dbh);

  F5::DbUtils::insert_row(dbh => $dbh,
                          table_name => 'PLC.PL_POLICY_HISTORY',
                          rec => $rec,
                         );

  $self->clean;
  return $filename;
}

sub restore {
  my ($self, @params) = @_;
  my ($revision_to_restore, $new_policy_name, $base_dir, %params);
  my $is_within_ucs_upgrade;

  if (scalar @params < 3) {
    ($revision_to_restore, $new_policy_name) = @params;
  }
  else {
    %params = @params;
    $revision_to_restore = $params{revision_to_restore};
    $new_policy_name = $params{new_policy_name};
    $is_within_ucs_upgrade = $params{is_within_ucs_upgrade};
    $base_dir = $params{base_dir};
  }

  my $dbh = $self->{_policy}->get_dbh;
  my $requested_find_last_revision = 0;

  if ($revision_to_restore == -1) {
    $requested_find_last_revision = 1;
    $revision_to_restore = $self->get_last_revision_number;
    return unless ($revision_to_restore);
  }

  my $revision_filename;
  eval{
       $revision_filename = $self->construct_filename(revision => $revision_to_restore, base_dir => $base_dir);
  };

  my $err = $@;

  # If revision requested was last revision (-1) and the file doesn't exist or is empty/corrupted,
  # return undef. F5::Policy->set_active_from_history will set the current policy
  # active instead.
  return if ($requested_find_last_revision && (! -e $revision_filename || -z $revision_filename) );
  die $err if $err;

  my $new_policy = F5::Policy->import_from_file(dbh => $dbh, 
                                                filename => $revision_filename, 
                                                account_id => $self->{_policy}->account_id,
                                                is_within_ucs_upgrade => $is_within_ucs_upgrade,
                                                external_idmap => $params{external_idmap},
                                               );

  return if !$new_policy;
  $new_policy->policy_name($new_policy_name);
  $new_policy->update_db;
  return $new_policy;
}

sub clean {
  my ($self) = @_;
  if (! F5::BigipUtils::is_docker() && $EFFECTIVE_USER_ID == 0) {
    my ($login, $pass,$apache_uid,$apache_gid);
    ($login,$pass,$apache_uid,$apache_gid) = getpwnam("apache")
      or die "Cannot getpwnam user apache";

# We must not change the owner of DEFAULT_BASE_DIRECTORY itself, because it belongs to tsui rpm.
    find(sub {chown $apache_uid,$apache_gid,$_}, DEFAULT_BASE_DIRECTORY);
  }

  my $dbh = $self->{_policy}->get_dbh;
  my $revisions_to_delete = $dbh->selectall_hashref('SELECT policy_history_id, revision FROM PLC.PL_POLICY_HISTORY WHERE policy_id = ? ' . 
						   'ORDER BY revision DESC LIMIT ' . MAX_NUMBER_OF_SAVED_VERSIONS .', 100000', 
						   'policy_history_id', undef, $self->{_policy}->policy_id);
  for my $revision_to_delete (values %$revisions_to_delete) {
    unlink($self->construct_filename(revision => $revision_to_delete->{revision}));
    $dbh->do('DELETE FROM PLC.PL_POLICY_HISTORY WHERE policy_history_id = ?', undef, $revision_to_delete->{policy_history_id});
  }

  if ( defined $MAX_SIZE_OF_SAVED_VERSIONS ) {
      F5::Utils::File::cleanup_files(dir => DEFAULT_BASE_DIRECTORY,
                                     max_size => $MAX_SIZE_OF_SAVED_VERSIONS,
                                     min_retained_files_in_dir => $MIN_RETAINED_FILES_IN_DIR);
  }
}

sub construct_filename {
  my($self, %params) = @_;

  my $revision = $params{revision};
  my $policy_id = $params{policy_id} || 0;

  if (ref($self)) {
    $revision ||= $self->get_last_revision_number;
    $policy_id ||= $self->{_policy}->policy_id;
  }
  my $base_dir = $params{base_dir} || "";
  $base_dir =~ s#(/ts/var/install|tmp)[/]?\Z#/#xms;  #Unzipped UCS passes $UCS/var/tmp as base_dir -- policy history files are under $UCS/var/ts/dms/policy/policy_versions
                                                     #Delayed load passes in /ts/var/install as the base dir, but policy history files are in the standard location
  my $filename = $base_dir . DEFAULT_BASE_DIRECTORY . '/' . DEFAULT_FILENAME_TEMPLATE;
  $filename =~ s/\$POLICY_ID\$/$policy_id/e;
  $filename =~ s/\$VERSION\$/$revision/;
  $filename = F5::Deploy::map_path($filename);
  return $filename;
}

sub clear_last_revision_end_time {
  my($self) = @_;
  my $dbh = $self->{_policy}->get_dbh;

  $dbh->do('UPDATE PLC.PL_POLICY_HISTORY SET to_date = 0 WHERE policy_id = ? AND revision = ?',
	   undef, $self->{_policy}->policy_id, $self->get_last_revision_number);
}

sub get_last_revision_number {
  my($self) = @_;
  my $dbh = $self->{_policy}->get_dbh;

  my($last_revision_number) = $dbh->selectrow_array('SELECT MAX(revision) FROM PLC.PL_POLICY_HISTORY WHERE policy_id = ?',
						   undef, $self->{_policy}->policy_id);
  return $last_revision_number || 0;
}

sub get_last_revision_file {
  my($self) = @_;
  return $self->construct_filename(revision => $self->get_last_revision_number);
}

sub get_last_username {
     my($self) = @_;
     my $dbh = $self->{_policy}->get_dbh;
     my ($last_username) = $dbh->selectrow_array('SELECT username FROM PLC.PL_POLICY_HISTORY WHERE policy_id = ? AND revision = ?',
                                                   undef, $self->{_policy}->policy_id, $self->get_last_revision_number);
     return $last_username;
}

#----------------------------------------------------------------------
# Class methods

sub delete_all_for_policy {
  my($dbh, $policy_id) = @_;
  # Todo: handle generic file templates
  rmtree(DEFAULT_BASE_DIRECTORY . "/$policy_id");
}



1;
