package F5::PolicyStabilityGraphs;

use strict;
use warnings;

use F5::ASMConfig::Types;
use F5::Logger;
use F5::Db::Single;
use F5::DbUtils;

#----------------------------------------------------------------------

sub new {
    my ( $class, %params ) = @_;

    my $self = bless( {}, $class );
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my ( $self, %params ) = @_;
    $self->{dbh}    = $params{dbh};
    $self->{logger} = $params{logger};
}

#----------------------------------------------------------------------

sub update {
    my ( $self, %params ) = @_;

    my $seconds_since_last_run = $params{seconds_since_last_run} || 3600;

    my $all_policies = $self->load_policy_table();
    $self->update_policy_changes( all_policies => $all_policies, seconds_since_last_run => $seconds_since_last_run );
    $self->update_learning_suggestions( all_policies => $all_policies );
    $self->update_enforced_entities( all_policies => $all_policies );

}

#----------------------------------------------------------------------

sub update_enforced_entities {
    my ( $self, %params ) = @_;

    my $all_policies = $params{all_policies};

    my $tmctl_table_name = 'asm_enforced_entities_stats';
    my @tmctl_fields     = qw(
      policy_name
      enforced_entities_cnt
      not_enforced_entities_cnt
    );

    my @entity_types = (
        {
            entity_type  => 'URL',
            entity_table => 'PLC.PL_OBJECTS',
            allowed_only => 1,
        },
        {
            entity_type  => 'WebSocket',
            entity_table => 'PLC.PL_WEBSOCKET_URLS',
            allowed_only => 1,
        },
        {
            entity_type  => 'Filetype',
            entity_table => 'PLC.PL_FILETYPES',
            allowed_only => 1,
        },
        {
            entity_type  => 'Parameter',
            entity_table => 'PLC.PL_PARAM_ATTRIBUTES',
            allowed_only => 0,
        },
        {
            entity_type  => 'Cookie',
            entity_table => 'PLC.PL_ALLOWED_COOKIES',
            allowed_only => 0,
        },
        {
            entity_type  => 'Signature',
        },
        {
            entity_type  => 'Threat Campaign',
        },
        {
            entity_type  => 'HTTP Protocol Compliance',
            entity_table => 'PLC.PL_HTTP_CHECKS',
            not_entity   => 1,
        },
        {
            entity_type  => 'Evasion Techniques',
            entity_table => 'PLC.PL_NORMALIZATION',
            not_entity   => 1,
        },
        {
            entity_type  => 'Web Services Security',
            entity_table => 'PLC.PL_XML_WS_SUBVIOLATIONS',
            not_entity   => 1,
        },
    );

    my $total_type_counts = {};
    for my $type (@entity_types) {
        my $type_counts;
        if ($type->{entity_type} eq 'Signature') {
            $type_counts = $self->get_signature_counts(all_policies => $all_policies);
        }
        elsif ($type->{entity_type} eq 'Threat Campaign') {
            $type_counts = $self->get_threat_campaign_counts(all_policies => $all_policies);
        }
        elsif($type->{not_entity}) {
            $type_counts = $self->get_not_entity_counts(%$type);
        }
        else {
            $type_counts = $self->get_type_counts(%$type);
        }

        for my $policy_id ( keys %$type_counts ) {
            if ( $type_counts->{$policy_id} && $type_counts->{$policy_id}->{total} ) {
                $total_type_counts->{$policy_id}->{total} += $type_counts->{$policy_id}->{total};
                if ( $type_counts->{$policy_id}->{not_enforced} ) {
                    $total_type_counts->{$policy_id}->{not_enforced} += $type_counts->{$policy_id}->{not_enforced};
                }
            }
        }
    }

    my @update_data;
    for my $policy_id ( keys %$total_type_counts ) {
        my $policy_name = $all_policies->{$policy_id}->{name};

        if ($policy_name) {
            my $total_entities        = $total_type_counts->{$policy_id}->{total}        || 0;
            my $not_enforced_entities = $total_type_counts->{$policy_id}->{not_enforced} || 0;
            my $enforced_entities     = $total_entities - $not_enforced_entities;

            #fields should be populated with same order as defined in '@tmctl_fields'
            push @update_data, [ $policy_name, $enforced_entities, $not_enforced_entities ];
        }
    }

    $self->update_tmstat(
        stat_values => \@update_data,
        tmctl_table_name => $tmctl_table_name,
        tmctl_fields => \@tmctl_fields,
    );
}

#----------------------------------------------------------------------

sub get_threat_campaign_counts {
    my ( $self, %params ) = @_;

    my $all_policies = $params{all_policies};
    my $dbh          = $self->{dbh} || F5::Db::Single->instance;

    $self->{logger}->debug("Load counts for entity: Threat Campaign");

    my $type_counts = {};
    my $sql = <<EndQuery;
SELECT ptc.policy_id policy_id,
       IF( ptc.flg_is_staging > 0 AND tca.flg_quarantine_threat_campaigns > 0, 1, 0) flg_staging,
       count(distinct ptc.policy_threat_campaign_id) count
  FROM PLC.PL_THREAT_CAMPAIGNS ptc, PLC.PL_POLICY_THREAT_CAMPAIGN_ATTRIBUTES tca
  WHERE ptc.policy_id = tca.policy_id
  GROUP BY policy_id, flg_staging
EndQuery

    $self->run_sql_iterator(
        sql              => $sql,
        iterator_handler => sub {
            my ($row_ref) = @_;
            my ($policy_id, $flg_staging, $count) = ($row_ref->{policy_id}, $row_ref->{flg_staging}, $row_ref->{count});
            if ($flg_staging) {
                $type_counts->{$policy_id}->{not_enforced} = $count;
            }
            $type_counts->{$policy_id}->{total} += $count;
        }
    );

    return $type_counts;
}

#----------------------------------------------------------------------

sub get_signature_counts {
    my ( $self, %params ) = @_;

    my $all_policies = $params{all_policies};
    my $dbh          = $self->{dbh} || F5::Db::Single->instance;

    $self->{logger}->debug("Load counts for entity: Signature");

    my($policy_signatures_staged) = $dbh->selectall_hashref(<<EndQuery, 'policy_id');
SELECT policy_id, flg_quarantine_signatures
    FROM PLC.PL_POLICY_NEGSIG_ATTRIBUTES
EndQuery

    my $type_counts = {};
    my $sql = <<EndQuery;
SELECT pns.policy_id as policy_id, pns.flg_quarantined as flg_quarantined, count(distinct nss.sig_id) as count
    FROM PLC.NEGSIG_SET_SIGNATURES nss, PLC.PL_POLICY_NEGSIG_SIGNATURES pns, PLC.PL_POLICY_NEGSIG_SETS pst
    WHERE pst.policy_id = pns.policy_id
    AND pst.set_id = nss.set_id
    AND nss.sig_id = pns.sig_id
    GROUP BY policy_id, flg_quarantined
EndQuery

    $self->run_sql_iterator(
        sql              => $sql,
        iterator_handler => sub {
            my ($row_ref) = @_;
            my ($policy_id, $flg_quarantined, $count) = ($row_ref->{policy_id}, $row_ref->{flg_quarantined}, $row_ref->{count});
            if ($policy_signatures_staged->{$policy_id}->{flg_quarantine_signatures} && $flg_quarantined > 0) {
                $type_counts->{$policy_id}->{not_enforced} = $count;
            }
            $type_counts->{$policy_id}->{total} += $count;
        }
    );

    return $type_counts;
}

#----------------------------------------------------------------------

sub get_not_entity_counts {
    my ( $self, %params ) = @_;

    my $entity_type  = $params{entity_type};
    my $entity_table = $params{entity_table};

    $self->{logger}->debug("Load counts for entity: $entity_type");

    my $type_counts = {};

    my $sql_total = <<EndQuery;
SELECT distinct policy_id, count(1) as count
    FROM $entity_table
    GROUP BY (policy_id)
EndQuery

    $self->run_sql_iterator(
        sql              => $sql_total,
        iterator_handler => sub {
            my ($row_ref) = @_;
            my ( $policy_id, $count ) = ( $row_ref->{policy_id}, $row_ref->{count} );
            $type_counts->{$policy_id}->{total} = $count;
        },
    );

    my $non_enforced_where_fields = {'flg_is_enabled' => 0};
    my $non_enforced_constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $non_enforced_where_fields);
    my $non_enforced_where_clause = $non_enforced_constructed_where_clause->{where_clause};
    my @non_enforced_where_values = @{$non_enforced_constructed_where_clause->{where_values}};

    my $sql = <<EndQuery;
SELECT distinct policy_id, count(1) as count
    FROM $entity_table
    WHERE $non_enforced_where_clause
    GROUP BY (policy_id)
EndQuery

    $self->run_sql_iterator(
        sql              => $sql,
        iterator_handler => sub {
            my ($row_ref) = @_;
            my ( $policy_id, $count ) = ( $row_ref->{policy_id}, $row_ref->{count} );
            $type_counts->{$policy_id}->{not_enforced} = $count;
        },
        bind_values => \@non_enforced_where_values,
    );

    return $type_counts;
}

#----------------------------------------------------------------------

sub get_type_counts {
    my ( $self, %params ) = @_;

    my $entity_type  = $params{entity_type};
    my $entity_table = $params{entity_table};
    my $allowed_only = $params{allowed_only};
    my $staging_field_name = $params{staging_field_name};

    $self->{logger}->debug("Load counts for entity: $entity_type");

    my $type_counts = {};

    my $where_fields = {};
    if ($allowed_only) {
        $where_fields->{'flg_is_allowed'} = 1;
    }

    my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    my $sql_total = <<EndQuery;
SELECT distinct policy_id, count(1) as count
    FROM $entity_table
    WHERE $where_clause
    GROUP BY (policy_id)
EndQuery

    $self->run_sql_iterator(
        sql              => $sql_total,
        iterator_handler => sub {
            my ($row_ref) = @_;
            my ( $policy_id, $count ) = ( $row_ref->{policy_id}, $row_ref->{count} );
            $type_counts->{$policy_id}->{total} = $count;
        },
        bind_values => \@where_values,
    );

    $staging_field_name ||= 'flg_is_staging';
    my $staging_where_fields = {$staging_field_name => 1};
    if ($allowed_only) {
        $staging_where_fields->{'flg_is_allowed'} = 1;
    }

    my $staging_constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $staging_where_fields);
    my $staging_where_clause = $staging_constructed_where_clause->{where_clause};
    my @staging_where_values = @{$staging_constructed_where_clause->{where_values}};

    my $sql = <<EndQuery;
SELECT distinct policy_id, count(1) as count
    FROM $entity_table
    WHERE $staging_where_clause
    GROUP BY (policy_id)
EndQuery

    $self->run_sql_iterator(
        sql              => $sql,
        iterator_handler => sub {
            my ($row_ref) = @_;
            my ( $policy_id, $count ) = ( $row_ref->{policy_id}, $row_ref->{count} );
            $type_counts->{$policy_id}->{not_enforced} = $count;
        },
        bind_values => \@staging_where_values,
    );

    return $type_counts;
}

#----------------------------------------------------------------------

sub update_policy_changes {
    my ( $self, %params ) = @_;

    my $all_policies = $params{all_policies};
    my $seconds_since_last_run = $params{seconds_since_last_run};

    my $tmctl_table_name = 'asm_policy_changes_stats';
    my @tmctl_fields     = qw(
      policy_name
      user_name
      policy_changes_cnt
    );

    my $sql = <<EndQuery;
SELECT policy_id, user_name, gui_user_name, count(1) as count
    FROM PLC.PL_CONFIG_LOG
    WHERE event_time > UNIX_TIMESTAMP(NOW())-?
    GROUP BY policy_id, user_name, gui_user_name
EndQuery

    my $builder_user_name = 'Policy Builder';
    my $policy_changes = {};
    $self->run_sql_iterator(
        sql              => $sql,
        iterator_handler => sub {
            my ($row_ref) = @_;
            my ( $policy_id, $user_name, $gui_user_name, $count ) =
              ( $row_ref->{policy_id}, $row_ref->{user_name}, $row_ref->{gui_user_name}, $row_ref->{count} );

            my $changes_user_name = $gui_user_name || $user_name || $builder_user_name;
            $policy_changes->{$policy_id}->{$changes_user_name} += $count;
        },
        bind_values => [$seconds_since_last_run]
    );

    my @update_data;
    for my $policy_id (keys %$all_policies) {
        my $policy_name = $all_policies->{$policy_id}->{name};
        if ($policy_changes->{$policy_id}) {
            for my $user_name (keys %{$policy_changes->{$policy_id}}) {
                push @update_data, [ $policy_name, $user_name, $policy_changes->{$policy_id}->{$user_name}];
            }
        }
        else {
            push @update_data, [ $policy_name, $builder_user_name, 0];
        }
    }

    $self->update_tmstat(
        stat_values => \@update_data,
        tmctl_table_name => $tmctl_table_name,
        tmctl_fields => \@tmctl_fields,
    );
}

#----------------------------------------------------------------------

sub update_learning_suggestions {
    my ( $self, %params ) = @_;

    my $all_policies = $params{all_policies};

    my $tmctl_table_name = 'asm_learning_suggestions_stats';
    my @tmctl_fields     = qw(
      policy_name accepted_cnt
      pending_lowscore_cnt
      pending_highscore_cnt
      ignored_cnt
    );

    my $sql = <<EndQuery;
SELECT policy_id,
       case when suggestion_score < 50 then 0 else 1 end as score_high,
       suggestion_status,
       count(1) as count
    FROM PLC.PL_SUGGESTIONS
    GROUP BY policy_id, score_high, suggestion_status
EndQuery

    my $policy_suggestion = {};
    $self->run_sql_iterator(
        sql              => $sql,
        iterator_handler => sub {
            my ($row_ref) = @_;
            my ( $policy_id, $score_high, $suggestion_status, $count ) =
              ( $row_ref->{policy_id}, $row_ref->{score_high}, $row_ref->{suggestion_status}, $row_ref->{count} );
            if ( $suggestion_status == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_IGNORED ) {
                $policy_suggestion->{$policy_id}->{ignored} += $count;
            }
            elsif ( $suggestion_status == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING ) {
                if ( $score_high == 0 ) {
                    $policy_suggestion->{$policy_id}->{pending_lowscore} += $count;
                }
                else {
                    $policy_suggestion->{$policy_id}->{pending_highscore} += $count;
                }
            }
            elsif ($suggestion_status == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT
                || $suggestion_status == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT_GLOBALLY()
                || $suggestion_status == F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_ACCEPT_AND_STAGE )
            {
                $policy_suggestion->{$policy_id}->{accepted} += $count;
            }
        }
    );

    my @update_data;
    for my $policy_id (keys %$all_policies) {
        my $policy_name = $all_policies->{$policy_id}->{name};
        if ($policy_name) {
            if ($policy_suggestion->{$policy_id}) {
                #fields should be populated with same order as defined in '@tmctl_fields'
                push @update_data,
                  [
                    $policy_name,
                    $policy_suggestion->{$policy_id}->{accepted}          || 0,
                    $policy_suggestion->{$policy_id}->{pending_lowscore}  || 0,
                    $policy_suggestion->{$policy_id}->{pending_highscore} || 0,
                    $policy_suggestion->{$policy_id}->{ignored}           || 0
                  ];
            }
            else {
                push @update_data, [ $policy_name, 0, 0, 0, 0];
            }
        }
    }

    $self->update_tmstat(
        stat_values => \@update_data,
        tmctl_table_name => $tmctl_table_name,
        tmctl_fields => \@tmctl_fields,
    );
}

#----------------------------------------------------------------------

sub load_policy_table {
    my ($self) = @_;

    my $sql = <<EndQuery;
SELECT id, name from PLC.PL_POLICIES
EndQuery

    my $all_policies = {};
    $self->run_sql_iterator(
        sql              => $sql,
        iterator_handler => sub {
            my ($row_ref) = @_;
            $all_policies->{ $row_ref->{id} } = $row_ref;
        }
    );

    return $all_policies;
}

#----------------------------------------------------------------------

sub update_tmstat {
    my ( $self, %params ) = @_;

    my $stat_values      = $params{stat_values};
    my $tmctl_table_name = $params{tmctl_table_name};
    my $tmctl_fields     = $params{tmctl_fields};

    #clear records from prvious run
    system('tmctl', '-Z', $tmctl_table_name);

    for my $stat_row (@$stat_values) {
        my @tmctl_args = ('-d', '/var/tmstat/blade/', '-W', $tmctl_table_name);
        for my $field_name (@$tmctl_fields) {
            my $field_value = shift @$stat_row;
            my $tmctl_argument = "$field_name=$field_value";
            push @tmctl_args, $tmctl_argument;
        }

        $self->{logger}->debug("Run tmctl command: ". join(' ', 'tmctl', @tmctl_args));
        system('tmctl', @tmctl_args);
    }
}

#----------------------------------------------------------------------

sub run_sql_iterator {
    my ( $self, %params ) = @_;

    my $sql              = $params{sql};
    my $bind_values      = $params{bind_values} || [];
    my $iterator_handler = $params{iterator_handler} || sub { };
    my $dbh              = $self->{dbh} || F5::Db::Single->instance;

    my $sth = $dbh->prepare($sql)
      || do {
        my $error_msg = "Fail prepare SQL statement: $sql - $DBI::errstr";
        $self->{logger}->error($error_msg);
        die $error_msg;
      };

    $sth->execute(@$bind_values)
      || do {
        my $error_msg = "Fail execute SQL statement: $sql, for values " . join( ':', @$bind_values ) . " - $DBI::errstr";
        $self->{logger}->error($error_msg);
        die $error_msg;
      };

    while ( my $row_ref = $sth->fetchrow_hashref ) {
        $iterator_handler->($row_ref);
    }

}

#----------------------------------------------------------------------

1;

