package F5::PolicyTemplateDefaults;

use strict;
use warnings;

use XML::LibXML;
use JSON::XS;


my $entry_lookup_path = [
                         ['type', q(/policy/type), 0,  sub {lc $_[0]}, 'security'],
                         ['applicationLanguage', q(/policy/encoding), 0, undef, 'auto-detect'],
                         ['serverTechnologies', q(/policy/server_technologies/server_technology/server_technology_name), 1, undef, []],
                         ['enforcementMode', q(/policy/blocking/enforcement_mode), 0],
                         ['enablePassiveMode', q(/policy/blocking/passive_mode), 0, \&map_enabled_to_boolean, JSON::XS::false],
                         ['enableBehavioralEnforcement', q(/policy/behavioral_enforcement/enabled), 0, \&set_boolean, JSON::XS::false],
                         ['learningMode', q(/policy/policy_builder/learning_mode), 0, sub {lc $_[0]}],
                         ['fullyAutomatic', q(/policy/policy_builder/fully_automatic), 0, \&set_boolean],
                         ['minimumAccuracyForAutoAddedSignatures', q(/policy/attack_signatures/min_accuracy_for_auto_added_signatures), 0, sub {lc $_[0]}, "medium"],
                         ['trustedIps', q(/policy/whitelist[policy_builder_trusted='true']),
                                                                {ipAddress => q(./ip_address),
                                                                 subnetMask => q(./subnet_mask)}],
                         ['enforcementReadinessPeriod', q(/policy/attack_signatures/staging_period_in_days), 0, \&set_int],
                         ['signatureStaging', q(/policy/attack_signatures/enable_staging), 0, \&set_boolean],
                         ['caseInsensitive', q(/policy/case_insensitive), 0, \&set_boolean],
                         ['protocolIndependent', q(/policy/protocol_free), 0, \&set_boolean],

                         [['learningSpeed', 'trustedTrafficLoosen', 'differentSources'], q(/policy/policy_builder/loosen_rule/trusted/distinct_sources), 0, \&set_int],
                         [['learningSpeed', 'trustedTrafficLoosen', 'minHoursBetweenSamples'], q(/policy/policy_builder/loosen_rule/trusted/minimum_interval), 0,
                             sub {return $_[0] / 3600}],
                         [['learningSpeed', 'trustedTrafficLoosen', 'maxDaysBetweenSamples'], q(/policy/policy_builder/loosen_rule/trusted/maximum_interval), 0,
                             sub {return $_[0] / 86400}],

                         [['learningSpeed', 'untrustedTrafficLoosen', 'differentSources'], q(/policy/policy_builder/loosen_rule/untrusted/distinct_sources), 0, \&set_int],
                         [['learningSpeed', 'untrustedTrafficLoosen', 'minHoursBetweenSamples'], q(/policy/policy_builder/loosen_rule/untrusted/minimum_interval), 0,
                             sub {return $_[0] / 3600}],
                         [['learningSpeed', 'untrustedTrafficLoosen', 'maxDaysBetweenSamples'], q(/policy/policy_builder/loosen_rule/untrusted/maximum_interval), 0,
                             sub {return $_[0] / 86400}],

                         [['learningSpeed', 'trustedTrafficSiteChangeTracking', 'differentSources'], q(/policy/policy_builder/track_site_changes/trusted/distinct_sources), 0, \&set_int],
                         [['learningSpeed', 'trustedTrafficSiteChangeTracking', 'minMinutesBetweenSamples'], q(/policy/policy_builder/track_site_changes/trusted/minimum_interval), 0,
                             sub {return $_[0] / 60}],
                         [['learningSpeed', 'trustedTrafficSiteChangeTracking', 'maxDaysBetweenSamples'], q(/policy/policy_builder/track_site_changes/trusted/maximum_interval), 0,
                             sub {return $_[0] / 86400}],

                         [['learningSpeed', 'untrustedTrafficSiteChangeTracking', 'differentSources'], q(/policy/policy_builder/track_site_changes/untrusted/distinct_sources), 0, \&set_int],
                         [['learningSpeed', 'untrustedTrafficSiteChangeTracking', 'minMinutesBetweenSamples'], q(/policy/policy_builder/track_site_changes/untrusted/minimum_interval), 0,
                             sub {return $_[0] / 60}],
                         [['learningSpeed', 'untrustedTrafficSiteChangeTracking', 'maxDaysBetweenSamples'], q(/policy/policy_builder/track_site_changes/untrusted/maximum_interval), 0,
                             sub {return $_[0] / 86400}],

                         [['learningSpeed', 'trafficTighten', 'maxModificationSuggestionScore'], q(/policy/policy_builder/tighten_rule/max_modification_suggestion_score), 0, \&set_int],
                         [['learningSpeed', 'trafficTighten', 'totalRequests'], q(/policy/policy_builder/tighten_rule/total_requests), 0, \&set_int],
                         [['learningSpeed', 'trafficTighten', 'minDaysBetweenSamples'], q(/policy/policy_builder/tighten_rule/minimum_interval), 0,
                             sub {return $_[0] / 86400}],

                        ];

sub set_int {
    return int($_[0]);
}

sub set_boolean {
    no strict 'refs';
    my $bool_func_name='JSON::XS::'.$_[0];
    return $$bool_func_name;
}

sub map_enabled_to_boolean {
    no strict 'refs';
    my $bool_func_name= $_[0] eq 'enabled' ? 'JSON::XS::true' : 'JSON::XS::false' ;
    return $$bool_func_name;
}

#----------------------------------------------------------------------

sub xml_template_defaults {
    my ($template_file) = @_;

    my $dom;
    eval {
        $dom = XML::LibXML->load_xml(location => $template_file);
    };

    if ($@) {
        die "Fail to parse policy template XML file: $template_file - $@\n";
    }

    my $template_defaults = {};
    for my $template_entry (@$entry_lookup_path) {
        my ($defaults_entry_name, $defaults_xpath, $is_list, $convert_handler, $default_value) = @$template_entry;
        my @nodes = $dom->findnodes($defaults_xpath);
        if (! scalar @nodes) {
            if (defined($default_value)) {
                $template_defaults->{$defaults_entry_name} = $default_value;
            }
            next;
        }

        my $template_data;
        if (ref($is_list) eq 'HASH') {
            my @structs;
            for my $node (@nodes) {
                my $struct = {};
                for my $struct_entry (keys %$is_list) {
                    my @curr_nodes = $node->findnodes($is_list->{$struct_entry});
                    $struct->{$struct_entry} = $curr_nodes[0]->textContent;
                }
                push @structs, $struct;
            }
            $template_data = \@structs;
        }
        elsif ($is_list) {
            my @values;
            for my $node (@nodes) {
                push @values, $node->textContent;
            }
            $template_data = \@values;
        }
        else {
            $template_data = $nodes[0]->textContent;
        }

        if ($convert_handler) {
            $template_data = $convert_handler->($template_data);
        }

        if (ref($defaults_entry_name) eq 'ARRAY') {
            my (@path_names) = @$defaults_entry_name;
            my $current_entry = $template_defaults;
            while ( scalar @path_names > 1 ) {
                my $path_entry = shift @path_names;
                $current_entry->{$path_entry} ||= {};
                $current_entry = $current_entry->{$path_entry};
            }
            my $last_entry = shift @path_names;
            $current_entry->{$last_entry} = $template_data;
        }
        else {
            $template_defaults->{$defaults_entry_name} = $template_data;
        }
    }

    my $template_defaults_json;
    if (scalar keys %$template_defaults) {
        my $coder = JSON::XS->new->ascii->allow_nonref;
        $template_defaults_json = $coder->encode($template_defaults);
    }

    return $template_defaults_json;
}

#----------------------------------------------------------------------

1;
