package F5::PolicyValidator;

use strict;
use warnings;

use File::Basename;
use F5::ASMConfig::Entity::Policy;
use F5::Policy::Overlay;

our $ERROR_TYPE_SYNTAX = 1;
our $ERROR_TYPE_LOGIC = 2;

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{dbh} = $params{dbh};
  $self->{logger} = $params{logger};
  $self->{errors} = { $ERROR_TYPE_LOGIC => {},
		      $ERROR_TYPE_SYNTAX => {} };
}

#----------------------------------------------------------------------

sub validate {
    my($self, %params) = @_;
    my $action = $params{action} || 'set_active';
    my $policy = $params{policy};
    if (!$policy && $params{policy_id}) {
        $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh},
                                                     logger => $self->{logger},
                                                     get_criteria => {policy_id => $params{policy_id}});
    }
    my ($is_policy_transparent) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy->{policy_id});
SELECT flg_disable_rejects
  FROM PLC.PL_POLICIES
 WHERE id = ?
EndQuery

    my $violations = $self->{dbh}->selectall_hashref(<<EndQuery, 'viol_index' , undef, $policy->{policy_id} );
SELECT flg_learn,flg_alarm,flg_reject, viol_index
  FROM PLC.PL_BLOCKING
 WHERE policy_id = ?
EndQuery

    my $overlay = F5::Policy::Overlay->new();
    my @modules = load_modules(
                               search_path => search_include_path('F5/PolicyValidator'),
                               basename => 'F5::PolicyValidator::',
                               base_class => 'F5::PolicyValidator::Default',
                              );
 
    for my $module (@modules) {
      $module->validate(dbh => $self->{dbh}, 
                        logger => $self->{logger}, 
                        validator => $self,
                        action => $action,
                        overlay => $overlay,
                        is_policy_transparent  => $is_policy_transparent,
                        violations => $violations,
                        policy => $policy);
    }

    my $validate_report = '';

    if (values %{$self->{errors}->{$ERROR_TYPE_SYNTAX}}) {
      $validate_report .= "Minor key referential errors were found in this policy. They can be fixed by exporting and re-importing the policy.\n\n"
    }

    for my $error_type (($ERROR_TYPE_LOGIC)) {
        for my $error_msg (keys %{$self->{errors}->{$error_type}}) {
            $validate_report .= "$error_msg\n";

            for my $error_row (@{$self->{errors}->{$error_type}->{$error_msg}}) {
                if ($error_row->{entity}) {
                    my $display = $error_row->{entity}->get_full_display_name();
                    $validate_report .= "\t$display\n";
                }
                elsif ($error_row->{table}) {
                     $validate_report .= "\tTable: $error_row->{table}, row id: $error_row->{row_id}\n";
                }
            }
            $validate_report .= "\n";
        }
    }
    return $validate_report;
}

#----------------------------------------------------------------------

sub search_include_path {
  my($package) = @_;
  for my $path (@INC) {
    return "$path/$package" if (-d "$path/$package");
  }
  return;
}

#----------------------------------------------------------------------

sub load_modules {
  my(%params) = @_;
  my @modules;

  my $search_path = $params{search_path};
  my $basename = $params{basename};

  while (my $module_file = glob($search_path . "/*.pm")) {
    my $module = $basename . (fileparse($module_file, '\..*'))[0];
    next if ($params{base_class} && $module eq $params{base_class});
    require $module_file;

    push(@modules, $module->new());
  }
  
  return @modules;
}

#----------------------------------------------------------------------

sub report_error {
  my($self, %params) = @_;

  unless ($self->{errors}->{$params{error_type}}->{$params{error_msg}}) {
    $self->{errors}->{$params{error_type}}->{$params{error_msg}} = [];
  }
  push(@{$self->{errors}->{$params{error_type}}->{$params{error_msg}}}, 
       { table => $params{table},
	 row_id => $params{row_id},
         entity => $params{entity}, });
}

1;
