package F5::PolicyValidator::CrossPolicyEntities;

use strict;
use warnings;

use base 'F5::PolicyValidator::Default';
use F5::ASMConfig::Entity::Parameter;
use F5::ASMConfig::Entity::Flow;
use F5::ASMConfig::Entity::DynamicFlow;
use F5::ASMConfig::Entity::Filetype;
use F5::ASMConfig::Entity::URL;
use F5::ASMConfig::Entity::Extraction;

sub validate {
  my($self, %params) = @_;
  my $dbh = $params{dbh};
  my $logger = $params{logger};
  my $validator = $params{validator};
  my $policy = $params{policy};
  my $action = $params{action};

  return if ($action && $action eq 'import_policy');

  my @foreign_keys = (
                      { 
                         dependent_rec => 'PLC.PL_FLOWS.from_id',
                         foreign_rec   => 'PLC.PL_OBJECTS.id',
                         id_field      => 'mydep.id',
                         entity        => 'F5::ASMConfig::Entity::Flow',
                      },
                      {
                          dependent_rec => 'PLC.PL_FLOWS.to_id',
                          foreign_rec   => 'PLC.PL_OBJECTS.id',
                          id_field      => 'mydep.id',
                          entity        => 'F5::ASMConfig::Entity::Flow',
                      },
                      {
                          dependent_rec => 'PLC.PL_FLOW_PARAMS.flow_id',
                          foreign_rec   => 'PLC.PL_FLOWS.id',
                          id_field      => 'mydep.param_attrib_id',
                          entity        => 'F5::ASMConfig::Entity::Parameter',
                      },
                      {
                          dependent_rec => 'PLC.PL_OBJECT_PARAMS.obj_id',
                          foreign_rec   => 'PLC.PL_OBJECTS.id',
                          id_field      => 'mydep.param_attrib_id',
                          entity        => 'F5::ASMConfig::Entity::Parameter',
                      },
                      {
                          dependent_rec => 'PLC.PL_OBJECT_EXTRACTS.obj_id',
                          foreign_rec   => 'PLC.PL_OBJECTS.id',
                          id_field      => 'mydep.extract_attrib_id',
                          entity        => 'F5::ASMConfig::Entity::Extraction',
                      },
                      {
                          dependent_rec => 'PLC.PL_GLOBAL_EXTRACT_OBJECTS.global_extract_id',
                          foreign_rec   => 'PLC.PL_GLOBAL_EXTRACTS.id',
                          id_field      => 'myfor.extract_attrib_id',
                          entity        => 'F5::ASMConfig::Entity::Extraction',
                      },
                      {
                          dependent_rec => 'PLC.PL_GLOBAL_EXTRACT_OBJECTS.obj_id',
                          foreign_rec   => 'PLC.PL_OBJECTS.id',
                          id_field      => 'myfor.id',
                          entity        => 'F5::ASMConfig::Entity::URL',
                      },
                      {
                           dependent_rec => 'PLC.PL_GLOBAL_EXTRACT_OBJECT_TYPES.global_extract_id',
                           foreign_rec   => 'PLC.PL_GLOBAL_EXTRACTS.id',
                           id_field      => 'myfor.extract_attrib_id',
                           entity        => 'F5::ASMConfig::Entity::Extraction',
                      },
                      {
                           dependent_rec => 'PLC.PL_GLOBAL_EXTRACT_OBJECT_TYPES.obj_type_id',
                           foreign_rec   => 'PLC.PL_FILETYPES.id',
                           id_field      => 'myfor.id',
                           entity        => 'F5::ASMConfig::Entity::Filetype',
                      },
                      {
                           dependent_rec => 'PLC.PL_DYNAMIC_PARAM_EXTRACTS.param_attrib_id',
                           foreign_rec   => 'PLC.PL_PARAM_ATTRIBUTES.id',
                           id_field      => 'mydep.extract_attrib_id',
                           entity        => 'F5::ASMConfig::Entity::Extraction',
                      },
                      {
                           dependent_rec => 'PLC.PL_DYNAMIC_PARAM_EXTRACTS.extract_attrib_id',
                           foreign_rec   => 'PLC.PL_EXTRACT_ATTRIBUTES.id',
                           id_field      => 'mydep.extract_attrib_id',
                           entity        => 'F5::ASMConfig::Entity::Extraction',
                      },
                      {
                           dependent_rec => 'PLC.PL_DYNAMIC_FLOWS.obj_id',
                           foreign_rec   => 'PLC.PL_OBJECTS.id',
                           id_field      => 'mydep.id',
                           entity        => 'F5::ASMConfig::Entity::DynamicFlow',
                      },
                      {
                           dependent_rec => 'PLC.PL_STATIC_PARAMS.param_attrib_id',
                           foreign_rec   => 'PLC.PL_PARAM_ATTRIBUTES.id',
                           id_field      => 'mydep.param_attrib_id',
                           entity        => 'F5::ASMConfig::Entity::Parameter',
                      },
                      {
                          dependent_rec => 'PLC.PL_PARAM_VALUE_METACHARS.param_attrib_id',
                          foreign_rec   => 'PLC.PL_PARAM_ATTRIBUTES.id',
                          id_field      => 'mydep.param_attrib_id',
                          entity        => 'F5::ASMConfig::Entity::Parameter',
                      }
    );


  for my $foreign_key_check (@foreign_keys) {
    my($dependent_db, $dependent_table, $dependent_field) = split(/\./, $foreign_key_check->{dependent_rec});
    my($foreign_db, $foreign_table, $foreign_field) = split(/\./, $foreign_key_check->{foreign_rec});
    my $allow_entry_point_clause = '';

    if ($foreign_key_check->{dependent_rec} eq 'PLC.PL_FLOWS.from_id') {
      $allow_entry_point_clause = 'AND mydep.from_id != 0';
    }

    my $sql = <<"EndQuery";
SELECT $foreign_key_check->{id_field} AS id_field, mydep.$dependent_field FROM $dependent_db.$dependent_table AS mydep,
$foreign_db.$foreign_table AS myfor
WHERE mydep.$dependent_field = myfor.$foreign_field
AND mydep.policy_id = ?
AND mydep.policy_id != myfor.policy_id
$allow_entry_point_clause
EndQuery

    my $sth = $dbh->prepare($sql);
    $sth->execute($policy->{policy_id});
    while(my $row = $sth->fetchrow_hashref) {
      $validator->report_error(error_type => $F5::PolicyValidator::ERROR_TYPE_SYNTAX,
			       error_msg => 'The following entities have a reference to, or are referenced in, another policy',
                               entity => $foreign_key_check->{entity}->new(dbh => $dbh,
                                                                           id => $row->{id_field}),
                               table  => $dependent_table,
                               row_id =>  $row->{id},
                              );
    }
  }
}


1;
