package F5::PopulateTemplate;

use strict;
use warnings;

use File::Basename;
use File::Slurp;
use F5::Deploy;
use F5::ASMConfig::Client;
use File::Path;
use YAML::XS;
use JSON -convert_blessed_universally;
use F5::ASM::COAPI;
use F5::Cluster::Utils;
use F5::Utils::Rest;
use POSIX 'strftime';
use MIME::Base64;
use Encode qw(encode);

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;

    my $self = bless({}, $class);
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->{dbh} = $params{dbh};
    $self->{asm_config_server} = $params{asm_config_server};
    $self->{logger} = $params{logger};
    $self->{config} = YAML::XS::LoadFile(F5::Deploy::map_path("/usr/share/ts/config/populate_template.yaml"));
}

#----------------------------------------------------------------------

sub populate_template {
    my ($self, %params) = @_;

    if ($params{rest_path} =~ s|/policies/(.*?)/|/policies/POLICY-HASH/|) {
        $params{policy_rest_uuid} = $1;
    }
    my $base_rest_path = $params{rest_path};
    my $template_filename = $self->{config}->{templates}->{$base_rest_path};
    if (! $template_filename) {
        die "Export template does not exist for $base_rest_path\n";
    }
    my $template_content = File::Slurp::read_file($template_filename);
    my $placeholders = $self->{config}->{placeholders};

    for my $placeholder (keys %$placeholders) {
        eval {
            $template_content =~ s|\'?\%($placeholder)(.*?)\%\'?|$self->replace_field(%params, placeholder => $1, placeholder_options => $2)|ge;
        };

        if ($@) {
            die "Failed while replacing placeholder $placeholder ($@)\n";
        }
    }

    return $template_content;
}

#----------------------------------------------------------------------

sub replace_field {
    my ($self, %params) = @_;
    my $placeholder = $params{placeholder};
    my $placeholder_options = $params{placeholder_options};

    my $field_def = $self->{config}->{placeholders}->{$placeholder};

    my $replace_func = $field_def->{replace_func};
    my $replace_value = $self->$replace_func(orig_params => \%params, field_def => $field_def);

    if ($field_def->{encode_json}) {
        my $json = JSON->new->convert_blessed->allow_blessed;
        $replace_value = $json->encode($replace_value);
        eval { $replace_value = decode('utf8', $replace_value) };
    }

    if ($field_def->{encode_base64}) {
        my @content_b64_lines = split(/\n/m, encode_base64($replace_value));
        my $rest_content_b64 = "";
        my $lines_counter = 0;
        for my $line (@content_b64_lines) {
            $lines_counter++;
            if ($lines_counter == 100) {
                $rest_content_b64 .= "\"+\n\"";
                $lines_counter = 0;
            }
            $rest_content_b64 .= $line;
        }

        # utfBytesToUnicodeChars is js function that convert UTF8 encoded string
        # into multi-byte Unicode characters and should be included in template

        my $convert_func_name = "conv_value_$placeholder";
        $replace_value = "$convert_func_name (\"$rest_content_b64\");\n";
        $replace_value .= "function $convert_func_name (convCont) {
    var returnContVal;
    var nonUtfCont=window.atob(convCont);
    var utfCont = utfBytesToUnicodeChars(nonUtfCont);
    try {
        returnContVal = JSON.parse(utfCont);
    }
    catch(er) {
        returnContVal = JSON.parse(nonUtfCont);
    }
    return returnContVal;
}";

    }

    return $replace_value;
}

#----------------------------------------------------------------------

sub get_rest {
    my ($self, %params) = @_;

    my $client = $self->{asm_config_server};
    my $filter = $params{filter};
    my $order_by = $params{order_by};
    my $field_def = $params{field_def};
    if ($field_def->{rest_path} && $params{orig_params} && $field_def->{rest_path} eq $params{orig_params}->{rest_path}) {
        $order_by = $params{orig_params}->{order_by};
        $filter = $params{orig_params}->{filter};
    }
    if ($params{orig_params} && $params{orig_params}->{policy_rest_uuid}) {
        $field_def->{rest_path} =~ s|/policies/POLICY-HASH|/policies/$params{orig_params}->{policy_rest_uuid}|;
    }

    my $call_flags = F5::ASMConfig::CallFlags->new();
    my $session = $client->{session};
    my $rest_path = $field_def->{rest_path};

    my $rest_call = F5::ASMConfig::RestRequest->new(rest_path => $rest_path,
                                                    rest_method => 'GET',
                                                    want_object => 1,
                                                   );
    my $rest_parameters = {};
    if ($field_def->{rest_parameters}) {
        for my $k (keys %{$field_def->{rest_parameters}}) {
            if (ref($field_def->{rest_parameters}->{$k}) && ref($field_def->{rest_parameters}->{$k}) eq 'ARRAY') {
                $rest_parameters->{$k} = join(',', @{$field_def->{rest_parameters}->{$k}});
            }
            else {
                $rest_parameters->{$k} = $field_def->{rest_parameters}->{$k};
            }
        }
    }
    $rest_parameters->{'$filter'} = $filter if $filter;
    $rest_parameters->{'$orderBy'} = $order_by if $order_by;

    $rest_call->{rest_parameters} = $rest_parameters;

    my $rest_content;
    eval {
        $rest_content = $client->rest_call($session, $call_flags, $rest_call);
    };
    if ($@) {
        die "Failed to retrieve from $field_def->{rest_path} ($@)";
    }

    if ($field_def->{item_fixup}) {
        my $item_fixup_func = $field_def->{item_fixup};
        for my $item (@{$rest_content->{items}}) {
            $self->$item_fixup_func(item => $item);
        }
    }

    if ($field_def->{rest_expand}) {
        # YAML::XS does not handle << key merge -- see http://www.perlmonks.org/?node_id=813443
        # Therefore, expanding it inline here (only one level of copying is needed)
        if ($field_def->{rest_expand}->{'<<'}) {
            for my $k (keys %{$field_def->{rest_expand}->{'<<'}}) {
                if (! exists $field_def->{rest_expand}->{$k}) {
                    $field_def->{rest_expand}->{$k} = $field_def->{rest_expand}->{'<<'}->{$k};
                }
            }
            delete $field_def->{rest_expand}->{'<<'};
        }

        $self->rest_expand_in_client(rest_content => $rest_content,
                                     rest_expand_def => $field_def->{rest_expand});
    }

    if ($field_def->{rest_map}) {
        my $key_field = $field_def->{rest_map_key} || 'id';
        $rest_content = { map { ($_->{$key_field} => $_) } @{$rest_content->{items}} };
    }

    if ($field_def->{output_field}) {
        $rest_content = $rest_content->{$field_def->{output_field}};
    }

    return $rest_content;
}

#----------------------------------------------------------------------

sub rest_expand_in_client {
    my ($self, %params) = @_;
    my $rest_content = $params{rest_content};
    my $rest_expand_def = $params{rest_expand_def};

    my $pre_expand_key = $rest_expand_def->{pre_expand_key};
    my $post_expand_key = $rest_expand_def->{post_expand_key};
    my $per_record_limit = $rest_expand_def->{per_record_limit};
    my $total_limit = $rest_expand_def->{total_limit};

    my %records_to_retrieve;
    my $records_counter = 0;
    for my $item (@{$rest_content->{items}}) {
        $item->{$post_expand_key} = [];
        next unless $item->{$pre_expand_key};
        my @pre_expand_links = @{$item->{$pre_expand_key}};
        if (scalar $#pre_expand_links > $per_record_limit) {
            splice(@pre_expand_links, $per_record_limit);
        }

        for my $pre_expand_link (@pre_expand_links) {
            my $link_id = F5::Utils::Rest::extract_rest_uuid_from_link('link' => $pre_expand_link->{'link'});

            if ($records_counter < $total_limit) {
                push(@{$item->{$post_expand_key}}, $link_id);
                $records_to_retrieve{$link_id} = 1;
                $records_counter++;
            }
        }
        delete $item->{$pre_expand_key};
    }

    my %retrieved_records;
    if (%records_to_retrieve) {
        my $expand_rest_result = $self->get_rest(field_def => $rest_expand_def,
                                                 filter => 'id IN (' . join(',', keys %records_to_retrieve) . ')',
                                                );
        for my $item (@{$expand_rest_result->{items}}) {
            $retrieved_records{$item->{id}} = $item;
        }
    }

    for my $item (@{$rest_content->{items}}) {
        for my $i (reverse 0..$#{$item->{$post_expand_key}}) {
            if ($retrieved_records{$item->{$post_expand_key}->[$i]}) {
                $item->{$post_expand_key}->[$i] = $retrieved_records{$item->{$post_expand_key}->[$i]};
            }
            else {
                splice(@{$item->{$post_expand_key}}, $i, 1);
            }
        }
    }
}

#----------------------------------------------------------------------

sub violations_item_fixup {
    my ($self, %params) = @_;

    my $violation = $params{item};

    $violation->{severity} = ucfirst( $violation->{severity} );
    my $viol_type = $violation->{type};
    if ($viol_type eq 'unspecified') {
        $viol_type = 'Other Violation';
    } else {
        my @chunks = split( /-/, $viol_type);
        for (my $i=0; $i<@chunks; $i++) {
            if ($chunks[$i] eq 'rfc') {
                $chunks[$i] = 'RFC';
            } else {
                $chunks[$i] = ucfirst( $chunks[$i] );
            }
        }
        $viol_type = join(' ', @chunks);
    }

    $violation->{type} = $viol_type;
    $violation->{userDefined} = $violation->{isUserDefined};
}

#----------------------------------------------------------------------

sub get_ip_reputation_pool {
    my ($self, %params) = @_;

    my $client = $self->{asm_config_server};
    my $session = $client->{session};

    my $call_flags = F5::ASMConfig::CallFlags->new();

    my $object = F5::ASMConfig::Entity::IpReputationCategoryAttributes->new(flg_is_visible => 1);
    my $res = $client->search_ip_reputation_category_attributes($session, $call_flags, $object);

    my $ip_reputation_pool = [];
    for my $elem (@{$res->{element_list}}) {
        my $entry = $elem->{ip_reputation_category_attributes};
        my $icontrol_name = lc $entry->{category_icontrol_name};
        $icontrol_name =~ s/_/-/g;
        push @$ip_reputation_pool, {
            value => $icontrol_name,
            title => $entry->{category_name},
            description => $entry->{description},
        };
    }

    return $ip_reputation_pool;
}

#----------------------------------------------------------------------

sub get_mcp {
    my ($self, %params) = @_;
    my $field_def = $params{field_def};
    my $res = F5::ASM::COAPI::Query->new($field_def->{mcp_type}, $field_def->{mcp_filter});

    return $res->[0] ? $res->[0]->{$field_def->{mcp_return_field}} : '';
}

#----------------------------------------------------------------------

sub get_country_names {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $country_names = { @{$dbh->selectcol_arrayref(<<EndQuery, { Columns=>[1, 2] })} };
SELECT country_code, country_name FROM PLC.COUNTRY_CODES
EndQuery

    $country_names->{'0'} = 'All';
    $country_names->{'N/A'} = 'N/A';

    return $country_names;
}

#----------------------------------------------------------------------

sub get_defense_attributes {
    my ($self, %params) = @_;

    my $client = $self->{asm_config_server};

    my $call_flags = F5::ASMConfig::CallFlags->new();
    my $session = $client->{session};

    my @defense_attribute_types = (
        ['xml', 'XmlDefenseAttribute'],
        ['json', 'JsonDefenseAttribute'],
        ['plain_text', 'PlainTextDefenseAttribute'],
        ['graphql', 'GraphqlDefenseAttribute'],
    );

    my $defense_attributes = {};
    for my $type (@defense_attribute_types) {
        my ($type_name, $type_class) = @$type;
        my $search_name = "search_${type_name}_defense_attribute";
        my $class_name = "F5::ASMConfig::Entity::$type_class";
        my $type_defense_attributes = $client->$search_name($session, $call_flags, $class_name->new);

        for my $attribute (@{$type_defense_attributes->{element_list}}) {
            my $attribute_data = $attribute->{"${type_name}_defense_attribute"};

            my $result = {
                name => $attribute_data->{def_attrib_name},
                type => $attribute_data->{value_type} == 3 ? 'boolean' : 'integer',
                units => $attribute_data->{units},
            };

            $defense_attributes->{$type_name}->{$attribute_data->{def_attrib_xpath}} = $result;
        }
    }

    return $defense_attributes;
}

#----------------------------------------------------------------------

sub get_num_slots {
    return F5::Cluster::Utils::get_num_slots();
}

#----------------------------------------------------------------------

sub is_clustering_enabled {
    return F5::Cluster::Utils::is_clustering_enabled() ? 'true' : 'false';
}

#----------------------------------------------------------------------

sub get_printable_time {
    return strftime("%F %T",localtime);
}

#----------------------------------------------------------------------

sub from_params {
    my($self, %params) = @_;
    return $params{orig_params}->{$params{field_def}->{param_name}};
}

#----------------------------------------------------------------------

1;
