package F5::PrepareConf::Base;

use strict;
use warnings;
no if $] >= 5.017011, warnings => 'experimental::smartmatch';

use File::Basename;
use Bit::Vector;
use F5::BigipUtils;
use F5::DbUtils;
use F5::IpUtils;
use YAML::XS;
use F5::Deploy;
use F5::Enum::ViolRatingCriteria;
use F5::Provisioning;
use F5::ASM::COAPI;
use F5::License;
use F5::CRC;
use F5::Utils::File;
use F5::Captcha;
use Digest::MD5;
use Socket qw(:DEFAULT inet_pton inet_ntop);
use HTTP::Response;
use File::Copy;
use File::Slurp;
use JSON::XS;

use F5::DateTimeUtils;
use F5::Logger;
use F5::PrepareConf::Constants qw(ERRMSG_DBH_LOST RET_OK RET_NOT_OK FALSE TRUE);
use F5::PrepareConf::Checksums;

our $ACCOUNT_DATA_DIR = F5::Deploy::map_path("/ts/var/account");
my $VIOL_RATING_CONFIG_FILE = F5::Deploy::map_path('/usr/share/ts/config/viol_rating.yaml');
my $SPAN_PORT_CONFIG_FILE = F5::Deploy::map_path('/usr/share/ts/config/passive_mode_affected_features.yaml');
my $FACTORY_CREDENTIALS_BLOB_FILE = F5::Deploy::map_path('/var/ts/bloom/bloom_credentials_blob.dat');
our $CLIENT_SIDE_JS_SRC_DIR = F5::Deploy::get_path_by_type('share') . '/client-side-js';
our $CLIENT_SIDE_JS_DEST_DIR = F5::Deploy::get_path_by_type('bd_config_files') . '/client-side-js';

my %risk_display = ( 1 => 'low',
                     2 => 'medium',
                     3 => 'high',
                   );

my %accuracy_display = ( 1 => 'low',
                         2 => 'medium',
                         3 => 'high',
                       );

#--------------------------------------------------------------------------------

sub new {
    my ($class, %params) = @_;

    my $self = bless {}, $class;
    $self->init(%params);
    return $self;
}

#--------------------------------------------------------------------------------

sub init
{
    my ($self, %params) = @_;
    $self->{dbh} = $params{dbh} || F5::DbUtils::get_dbh();
    $self->{logger} = $params{logger} || F5::Logger->get_logger(logger_name => 'set_active');
    $self->{checksums} = F5::PrepareConf::Checksums->new(dbh => $self->{dbh},
                                                         logger => $self->{logger},
                                                         checksums_policy_id => $params{checksums_policy_id});
    $self->{modified_dcc_tables} = {};
    $self->{passive_mode_config} =  YAML::XS::LoadFile($SPAN_PORT_CONFIG_FILE);
    $self->{batch_insert} = $params{batch_insert};

    return $self;
}

#--------------------------------------------------------------------------------

sub get_modified_dcc_tables {
    my($self) = @_;
    return keys %{$self->{modified_dcc_tables}};
}

#--------------------------------------------------------------------------------

sub add_modified_dcc_tables {
    my($self, @modified_tables) = @_;

    for my $table (@modified_tables) {
        $self->{modified_dcc_tables}->{$table} = 1;
    }
}

#--------------------------------------------------------------------------------

sub empty_table {
    my ($self, $tables, $policy_name_crcs, $where_fields) = @_;
    local $"=',';
    my $change_made = 0;
    my $logger = $self->{logger};

    my $db_fields = F5::DbUtils::get_db_fields($self->{dbh}, $tables, undef, 1);
 
    if(defined $policy_name_crcs && defined $db_fields->{policy_name_crc})
    {
       if (!ref($policy_name_crcs)) {
           $policy_name_crcs = [$policy_name_crcs];
       }
       if (ref($policy_name_crcs) eq 'HASH') {
           $policy_name_crcs = [$policy_name_crcs->{policy_name_crc}];
       }
       $where_fields->{policy_name_crc} = $policy_name_crcs;
    }

    my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    if (! ref($tables)) {
        $tables = [ $tables ];
    }

    for my $table (@$tables) {
        my($exists) = $self->{dbh}->selectrow_array(<<EndQuery, undef, @where_values);
SELECT count(*)
  FROM $table
 WHERE $where_clause
EndQuery

        next unless $exists; #Protect against bad table locking

        my $rv = $self->{dbh}->do(<<EndQuery, undef, @where_values);
DELETE FROM $table 
  WHERE $where_clause
EndQuery

        if (! $rv) {
            # failure
            #error($DBI::errstr);
            $logger->error("empty_table failed: $table") if $logger;
            next;
        }

        $self->{modified_dcc_tables}->{$table} = 1;
        # Cast E0E to 0.
        $change_made += $rv;
    }

    return $change_made; # success
}

#----------------------------------------------------------------------

sub insert_row {
    my($self, %params) = @_;
    my $is_policy_passive = defined $params{is_policy_passive} ? $params{is_policy_passive} : $self->{is_policy_passive};
    if ( $is_policy_passive &&  exists $self->{passive_mode_config}->{$params{table_name}}) {
        if ( $self->{passive_mode_config}->{$params{table_name}}->{skip} ){
            $self->{logger}->debug("Passive mode: Skipping table $params{table_name}.") if $self->{logger};
            return;
        }

        if ( $self->{passive_mode_config}->{$params{table_name}}->{field_values} ){
            for my $field_val_hash ( @{$self->{passive_mode_config}->{$params{table_name}}->{field_values}} ) {
                for my $field_name ( keys %$field_val_hash ) {
                   $self->{logger}->debug("Passive mode: Changing table $params{table_name}. Modified field '$field_name' value to '$field_val_hash->{$field_name}'.") if $self->{logger};
                   $params{rec}->{$field_name} = $field_val_hash->{$field_name};
                }
            }
        }
    }
    if ($self->{batch_insert}) {
        $self->{batch_insert}->insert_row(%params);
    }
    else {
        F5::DbUtils::insert_row(dbh => $self->{dbh},
                                logger => $self->{logger},
                                %params);
    }
    $self->{modified_dcc_tables}->{$params{table_name}} = 1;
}

#----------------------------------------------------------------------

sub update_row {
    my($self, %params) = @_;
    F5::DbUtils::update_row(dbh => $self->{dbh},
                            %params);
    $self->{modified_dcc_tables}->{$params{table_name}} = 1;
}

#----------------------------------------------------------------------

sub prepare {
    die;
}

#----------------------------------------------------------------------

sub prepare_generic {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    my $dbh = $self->{dbh};
    my $do_checksums = $params->{do_checksums};

    my $policy_part = $params->{policy_part};
    my @sources = @{$policy_part->{tables}->{source_tables}};
    my $required_features = $policy_part->{required_features} || [];

    if ($do_checksums) {
        my $have_tables_changed = $self->{checksums}->have_tables_changed(tables => \@sources );
        return RET_OK unless $have_tables_changed;
    }

    my $old_config;
    for my $dest_table (@{$policy_part->{tables}->{dest_tables}}) {
        my $db_fields = F5::DbUtils::get_db_fields($self->{dbh}, $dest_table, undef, 1);

        if (not defined $db_fields->{policy_name_crc}) {
            if ($policy_part->{pass_old_config}) {
                $old_config->{$dest_table} = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT *
  FROM $dest_table
EndQuery
            }
        }
        else {
            if ($policy_part->{pass_old_config}) {
                $old_config->{$dest_table} = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc);
SELECT *
  FROM $dest_table
  WHERE policy_name_crc = ?
EndQuery
            }
        }

        $self->empty_table($dest_table, $policy_name_crc);       
    }

    if ( !$policy_id && !$policy_part->{policy_independent} ){
        return RET_OK;
    }

    for my $feature (@$required_features) {
        if (!F5::BigipUtils::is_feature_enabled($feature)) {
            return RET_OK;
        }
    }

    if ($policy_part->{multiple_source_table_handling} &&
        $policy_part->{multiple_source_table_handling} eq 'join') {
        # Combine all tables in a single query
        @sources = ( [ @sources ] );
    }

    for my $source (@sources) {
        my(%where_fields, %literal_where_fields);
        my $db_fields;
        if (ref($source)) {
            my @joined_tables = @$source;
            $source = join(', ', @$source);
            for my $i (1..$#joined_tables) {
                $literal_where_fields{$joined_tables[$i-1] . ".$policy_part->{join_field_name}"} = $joined_tables[$i] . ".$policy_part->{join_field_name}";
            }

            $db_fields = F5::DbUtils::get_db_fields($self->{dbh}, $joined_tables[0], undef, 1);
            if (defined $db_fields->{policy_id}) {
                # If multiple source tables are joined and use policy_id, then policy_id must exist in the first source table
                $where_fields{$joined_tables[0] . ".policy_id"} = $policy_id;
            }
        }
        else {
            $db_fields = F5::DbUtils::get_db_fields($self->{dbh}, $source, undef, 1);
            if (defined $db_fields->{policy_id}) {
                $where_fields{policy_id} = $policy_id;
            }
        }

        my $where_clause = F5::DbUtils::construct_where_clause(where_fields => \%where_fields,
                                                               literal_where_fields => \%literal_where_fields);

        my $order_by_clause = '';
        if($policy_part->{order_by}){
            $order_by_clause = "ORDER BY $policy_part->{order_by}";
        }
        my $plc_sth = $dbh->prepare("SELECT * from $source WHERE $where_clause->{where_clause} $order_by_clause");
        $plc_sth->execute(@{$where_clause->{where_values}});
        while (my $row = $plc_sth->fetchrow_hashref) {
            $row->{policy_name_crc} = $policy_name_crc;
    
            if ($policy_part->{row_handler}) {
                my $function_name = $policy_part->{row_handler};
                $self->$function_name(%$params,
                                      policy_name_crc => $policy_name_crc,
                                      policy_id => $policy_id,
                                      row => $row,
                                      old_config => $old_config,
                                     );
            }
    
            for my $dest_table (@{$policy_part->{tables}->{dest_tables}}) {
                if (!$row->{__skip_row}){
                    $self->insert_row(
                                      table_name => $dest_table,
                                      rec => $row );
                }
            }
        }
    }

    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_server_technologies_fingerprints {
    my($self) = @_;
    require F5::NegativeSignatures::Collection::Compiler;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $logger->debug("prepare_server_technologies_fingerprints, start") if $logger;

    my $have_tables_changed = $self->{checksums}->have_tables_changed(tables => [ qw(
                                                                                     PLC.SERVER_TECHNOLOGIES
                                                                                 ) ] );

    if ($have_tables_changed) {
        $logger->debug("Preparing DCC.SERVER_TECHNOLOGIES_FINGERPRINTS") if $logger;
        $self->empty_table('DCC.SERVER_TECHNOLOGIES_FINGERPRINTS');
        $self->empty_table('DCC.SERVER_TECHNOLOGIES_FINGERPRINTS_COLLECTIONS');


        my $server_technologies_sth = $dbh->prepare(<<EndQuery);
SELECT ST.* , FP.*
  FROM PLC.SERVER_TECHNOLOGIES as ST, PLC.SERVER_TECHNOLOGIES_FINGERPRINTS as FP
WHERE ST.server_technology_id = FP.server_technology_id
ORDER BY ST.server_technology_id, FP.server_technology_fingerprint_id
EndQuery

      $server_technologies_sth->execute;
      my @request_fingerprints;
      my @response_fingerprints;

      while(my $fingerprint = $server_technologies_sth->fetchrow_hashref) {

          $self->insert_row(dbh => $dbh,
                            table_name => 'DCC.SERVER_TECHNOLOGIES_FINGERPRINTS',
                            rec => {
                                       server_technology_fingerprint_id => $fingerprint->{server_technology_fingerprint_id},
                                       server_technology_id => $fingerprint->{server_technology_id},
                                       server_technology_name => $fingerprint->{server_technology_name},
                                   }
                           );

          $fingerprint->{sig_id} = $fingerprint->{server_technology_fingerprint_id};
          if ($fingerprint->{apply_to} ==  F5::ASMConfig::ApplyTo::APPLY_TO_REQUEST() ){
              push @request_fingerprints, $fingerprint;
          }
          else{
              push @response_fingerprints, $fingerprint;
          }
      }

      my $request_mask = 0;
      my $response_mask = 0;
      my $request_md5sum = Digest::MD5->new;
      my $response_md5sum = Digest::MD5->new;

      my $request_server_technology_acy_collection  = F5::NegativeSignatures::Collection::Compiler::get_compiled_collection(sigs => \@request_fingerprints );

      my $response_server_technology_acy_collection  = F5::NegativeSignatures::Collection::Compiler::get_compiled_collection(sigs => \@response_fingerprints );

      $request_md5sum->add($request_server_technology_acy_collection);
      $response_md5sum->add($response_server_technology_acy_collection);

      $request_md5sum = $request_md5sum->hexdigest;
      $response_md5sum = $response_md5sum->hexdigest;

      for my $request_fingerprint ( @request_fingerprints ) {
          if (defined  $request_fingerprint->{rule_obj}){
              $request_mask |= $request_fingerprint->{rule_obj}->buffer_type_mask;
          }
      }

      for my $response_fingerprint ( @response_fingerprints ) {
          if (defined  $response_fingerprint->{rule_obj}){
              $response_mask |= $response_fingerprint->{rule_obj}->buffer_type_mask;
          }
      }

      $self->insert_row(dbh => $dbh,
                        table_name => 'DCC.SERVER_TECHNOLOGIES_FINGERPRINTS_COLLECTIONS',
                              rec => {
                                       request_server_technology_acy_collection => $request_server_technology_acy_collection,
                                       response_server_technology_acy_collection => $response_server_technology_acy_collection,
                                       request_mask => $request_mask,
                                       response_mask => $response_mask,
                                       request_md5sum => $request_md5sum,
                                       response_md5sum => $response_md5sum,
                                     },
                             );

      $self->{checksums}->store;
    }

    $logger->debug("prepare_server_technologies_fingerprints, end") if $logger;
    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_icap_server {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $logger->debug("prepare_icap_server, start") if $logger;

    $self->prepare_generic(0, 0,
                               {
                                policy_part => {
                                                policy_independent => 1,
                                                tables => {
                                                           source_tables => [ qw( PLC.VIRUS_DETECTION_SERVER ) ],
                                                           dest_tables => [ qw( DCC.VIRUS_DETECTION_SERVER ) ],
                                                          },
                                                row_handler => sub {
                                                                   my ($self, %params) = @_;

                                                                   my $row = $params{row};
                                                                   my $hostname = $row->{hostname};
                                                                   if ($hostname) {
                                                                       my $split_ip_route = F5::IpUtils::split_ip_route( ip_address => $hostname,
                                                                                                                         die_on_error => 0 );
                                                                       if ($split_ip_route->{ip_address} && defined inet_pton(AF_INET, $split_ip_route->{ip_address})) {
                                                                           $row->{icap_server_ip_addr} = $hostname;
                                                                       }
                                                                       if (!$row->{icap_server_ip_addr}) {
                                                                           $row->{icap_server_host_name} = $hostname;
                                                                       }
                                                                   }
                                                                   else {
                                                                       $row->{__skip_row} = 1;
                                                                   }
                                                               },
                                               },
                                do_checksums => 1,
                               });

    $logger->debug("prepare_icap_server, end") if $logger;
    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_database_protection_server {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $logger->debug("prepare_database_protection_server, start") if $logger;

    $self->prepare_generic(0, 0,
                               {
                                policy_part => {
                                                policy_independent => 1,
                                                tables => {
                                                           source_tables => [ qw( PLC.DATABASE_PROTECTION_SETTINGS ) ],
                                                           dest_tables => [ qw( DCC.DATABASE_PROTECTION_SETTINGS ) ],
                                                          },
                                                row_handler => sub {
                                                                   my ($self, %params) = @_;

                                                                   my $row = $params{row};
                                                                   my $hostname = $row->{hostname};
                                                                   if ($hostname) {
                                                                       my $split_ip_route = F5::IpUtils::split_ip_route( ip_address => $hostname,
                                                                                                                         die_on_error => 0 );
                                                                       if ($split_ip_route->{ip_address} && defined inet_pton(AF_INET, $split_ip_route->{ip_address})) {
                                                                           $row->{server_ip_addr} = $hostname;
                                                                       }
                                                                       if (!$row->{server_ip_addr}) {
                                                                           $row->{server_host_name} = $hostname;
                                                                       }
                                                                   }
                                                                   else {
                                                                       $row->{__skip_row} = 1;
                                                                   }
                                                               },
                                               },
                                do_checksums => 1,
                               });


    $logger->debug("prepare_database_protection_server, end") if $logger;
    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_bot_signature_names {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $logger->debug("prepare_bot_signature_names, start") if $logger;

    my $have_tables_changed = $self->{checksums}->have_tables_changed(tables => [ qw(
                                                                                    PLC.BOT_SIGNATURES
                                                                                ) ] );
    if ($have_tables_changed) {
        require F5::BotSignatures;
        $self->empty_table('DCC.BOT_SIGNATURES');
        $self->empty_table('DCC.BOT_SIGNATURE_HOSTNAMES');

        my $batch_insert = F5::BatchInsert->new(dbh => $dbh);

        my $bot_signatures = F5::BotSignatures::get_all_bot_signatures(dbh => $dbh,
                                                                       logger => $logger,
                                                                       include_hostnames => 1, 
                                                                      );

        my @hostnames;
        for my $bot_signature (values @$bot_signatures) {
            $bot_signature->{bot_sig_id} = $bot_signature->{bot_signature_id};
            $bot_signature->{bot_sig_name} = $bot_signature->{name};
            $bot_signature->{bot_sig_class} = $bot_signature->{bot_class};
            $bot_signature->{bot_sig_category} = $bot_signature->{category};

            $batch_insert->insert_row(table_name => 'DCC.BOT_SIGNATURES',
                                      do_insert_ignore => 1,
                                      rec => {
                                              %$bot_signature,
                                             }
                                     );

            next if !defined $bot_signature->{hosts};
            for my $host (@{$bot_signature->{hosts}}) {
                my $bot_signature_hostname = {bot_sig_id => $bot_signature->{bot_signature_id}, hostname => $host};
                push (@hostnames, $bot_signature_hostname);
            }
        }
        $batch_insert->flush;

        for my $hostname (@hostnames) {
            $batch_insert->insert_row(table_name => 'DCC.BOT_SIGNATURE_HOSTNAMES',
                                      do_insert_ignore => 1,
                                      rec => $hostname,
                                     );
        }
        $batch_insert->flush;

        $self->{checksums}->store;
    }

    $logger->debug("prepare_bot_signature_names and hostnames, end") if $logger;
    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_negsig_enforcer_names {
    my($self, $account_id, $policy_id, $params) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $logger->debug("prepare_negsig_enforcer_names, start") if $logger;

    my $have_tables_changed = $self->{checksums}->have_tables_changed(tables => [ qw(
                                                                                     PLC.NEGSIG_SIGNATURES
                                                                                     DCC.INTERNAL_CONFIG_ITEMS
                                                                                     DCC.DECEPTION_TRAP_CONDITIONS
                                                                                ) ] );
    if ($have_tables_changed) {
        $logger->debug("Preparing DCC.NEGSIG_SIGNATURES_NAMES") if $logger;
        $self->empty_table('DCC.NEGSIG_SIGNATURES_NAMES');

        my $conviction_sigs = $dbh->selectall_hashref(<<EndQuery, 'condition_value', undef, F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_SIGNATURE());
SELECT *
  FROM DCC.DECEPTION_TRAP_CONDITIONS
 WHERE condition_source = ?
EndQuery

        my($include_sig_risk_accuracy) = $dbh->selectrow_array(<<EndQuery, undef, 'remote_logger_include_sig_risk_accuracy');
SELECT current_value
  FROM DCC.INTERNAL_CONFIG_ITEMS
 WHERE field_name = ?
EndQuery


        my $batch_insert = F5::BatchInsert->new(dbh => $dbh,
                                                batch_size => 500,);

        my $negsig_sth = $dbh->prepare(<<EndQuery);
SELECT sig_id, sig_name, sig_tag, attack_type_id, risk, accuracy, flg_has_cve
  FROM PLC.NEGSIG_SIGNATURES
EndQuery

        my $sig_cves = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, F5::ASMConfig::SigReferenceType::SIGNATURE_REFERENCE_TYPE_CVE());
SELECT sig_id, reference_val
  FROM PLC.NEGSIG_SIGNATURE_REFERENCES
 WHERE reference_type = ?
EndQuery

        $negsig_sth->execute;
        while (my $negsig_rec = $negsig_sth->fetchrow_hashref) {
            my $sig_name = $negsig_rec->{sig_name};

            if ( defined $negsig_rec->{sig_tag} && $negsig_rec->{sig_tag} ne '' ){
                $sig_name .= " [$negsig_rec->{sig_tag}]";
            }

            if ($include_sig_risk_accuracy) {
                $sig_name .= '!@#${' . "sig_risk=$risk_display{$negsig_rec->{risk}};sig_accuracy=$accuracy_display{$negsig_rec->{accuracy}}" . '}';
            }

            my $attack_type_code = 0;
            if ($negsig_rec->{attack_type_id}) {
                $attack_type_code = 2 ** ($negsig_rec->{attack_type_id} - 1);
            }

            my @cves = map { $_->{reference_val} } grep { $_->{sig_id} == $negsig_rec->{sig_id} } @$sig_cves;
            my $cves_list = @cves ? uc(join(',', @cves)) : '';

            $batch_insert->insert_row(table_name => 'DCC.NEGSIG_SIGNATURES_NAMES',
                                      do_insert_ignore => 1,
                                      rec => {
                                              %$negsig_rec,
                                              sig_name => $sig_name,
                                              sig_name_len => length($sig_name),
                                              attack_type_code => $attack_type_code,
                                              sig_cves => $cves_list,
                                              flg_conviction => defined $conviction_sigs->{$negsig_rec->{sig_id}} || 0,
                                             }
                                     );
        }

        $batch_insert->flush;
    }

    if ($self->{checksums}->have_tables_changed(tables => [ qw(
                                                               PLC.NEGSIG_ATTACK_TYPES
                                                          ) ] )) {
        $logger->debug("Preparing PLC.NEGSIG_ATTACK_TYPES") if $logger;
        $self->empty_table('DCC.ATTACK_TYPE_CODES');

        my $attacks_sth = $dbh->prepare(<<EndQuery);
SELECT attack_type_id, 
       attack_type_display_name, 
       attack_type_name
  FROM PLC.NEGSIG_ATTACK_TYPES
EndQuery

        $attacks_sth->execute();
        while (my $attack_rec = $attacks_sth->fetchrow_hashref) {
            my $attack_type_backward_compatible_name = 'ATTACK_TYPE_' . uc($attack_rec->{attack_type_name});
            $attack_type_backward_compatible_name =~ s/\s*\(.*\)//g;
            $attack_type_backward_compatible_name =~ s/[\s\-\/]+/_/g;

            my $attack_type_code = 0;
            if ($attack_rec->{attack_type_id}) {
                $attack_type_code = Bit::Vector->new_Enum(64, $attack_rec->{attack_type_id} - 1)->to_Dec;
            }
            $self->insert_row(table_name => 'DCC.ATTACK_TYPE_CODES',
                              rec => {
                                      attack_type_id => $attack_rec->{attack_type_id},
                                      attack_type_name => $attack_rec->{attack_type_display_name},
                                      attack_type_backward_compatible_name => $attack_type_backward_compatible_name,
                                      attack_type_code => $attack_type_code,
                                      attack_type_bit_index => $attack_rec->{attack_type_id} - 1,
                                     }
                             );
        }

        # Add hard-coded value for "Aggregated" attack type (requested by AVR - bug 432504)
        my $AGGREGATED_ATTACK_CODE = 63;
        $self->insert_row(table_name => 'DCC.ATTACK_TYPE_CODES',
                          rec => {
                                  attack_type_id => $AGGREGATED_ATTACK_CODE,
                                  attack_type_name => 'Aggregated',
                                  attack_type_backward_compatible_name => 'Aggregated',
                                  attack_type_code => Bit::Vector->new_Enum(64, $AGGREGATED_ATTACK_CODE - 1)->to_Dec,
                                  attack_type_bit_index => $AGGREGATED_ATTACK_CODE - 1,
                                 }
                         );
    }
    $self->{checksums}->store;

    $logger->debug("prepare_negsig_enforcer_names, end") if $logger;
    return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_method_pool {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $logger->debug("Preparing DCC.METHOD_POOL") if $logger;

    my $http_methods = F5::ASM::COAPI::Query->new('asm_http_method');

    #Since method code is the primary key, first delete any other instances of methods by name. 
    #This will prevent multiple instances of the same method name with different method codes
    if ($http_methods && @$http_methods) {
        my @method_names = map { $_->{leaf_name} } @$http_methods;
        F5::DbUtils::delete_rows(dbh => $dbh,
                                 table_name => 'DCC.METHOD_POOL',
                                 where_fields => { method_name => \@method_names },
                                );

        for my $method (@$http_methods) {
            $self->insert_row(table_name => 'DCC.METHOD_POOL',
                              rec => {
                                      method_code => $method->{code},
                                      method_name => $method->{leaf_name},
                                     },
                              do_replace => 1,
                             );
        }
    }
    else {
        $logger->warn("asm http-method table is missing from MCP") if $logger;
    }

    return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_violations {
    my($self) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $logger->debug("prepare_violations, start") if $logger;

    $self->prepare_generic(0, 0,
                               {
                                policy_part => {
                                                policy_independent => 1,
                                                tables => {
                                                           source_tables => [ qw( PLC.VIOLATIONS ) ],
                                                           dest_tables => [ qw( DCC.PARSER_STATUS_CODES ) ],
                                                          },
                                                row_handler => sub {
                                                                   my ($self, %params) = @_;

                                                                   my $row = $params{row};

                                                                   if ($row->{flg_is_deleted}) {
                                                                       $row->{__skip_row} = 1;
                                                                   }
                                                               },
                                               },
                                do_checksums => 1,
                               });

    $self->prepare_generic(0, 0,
                               {
                                policy_part => {
                                                policy_independent => 1,
                                                tables => {
                                                           source_tables => [ qw( PLC.VIOLATIONS_PROPERTIES ) ],
                                                           dest_tables => [ qw( DCC.PARSER_STATUS_CODE_SEVERITIES ) ],},
                                               },
                                do_checksums => 1,
                               });
}

#--------------------------------------------------------------------------------

sub prepare_viol_rating {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;
    my $dbh = $self->{dbh};

    $self->empty_table('DCC.VIOL_RATING');
    my $viol_rating_config = YAML::XS::LoadFile($VIOL_RATING_CONFIG_FILE);
    my %viol_rating_criteria_by_name = reverse %F5::Enum::ViolRatingCriteria::as_string;

    for my $criteria (keys %$viol_rating_config) {
        $self->insert_row(
                          table_name => 'DCC.VIOL_RATING',
                          rec => { %{$viol_rating_config->{$criteria}},
                                   viol_rating_criteria => $viol_rating_criteria_by_name{$criteria},
                                 });
    }
}

#----------------------------------------------------------------------
# prepare_enforcer_accounts can be called for all accounts (usually on ASM startup)
# or for a specific account (upon Set Active)
#
sub prepare_enforcer_accounts {
    my($self, $policy_name_crc, $policy_id) = @_;

    my $dbh = $self->{dbh};

    $self->empty_table('DCC.ENFORCER_ACCOUNTS', $policy_name_crc);

    if (! F5::Provisioning::is_module_enabled('asm')) {
        # If ASM is not provisioned, do not populate this table at all
        return;
    }

    my @where_clauses;
    my @bind_values;
    if ($policy_name_crc) {
        push(@where_clauses, 'A.policy_name_crc = ?');
        push(@bind_values, $policy_name_crc);
    }

    my $where_clause = '';
    if (@where_clauses) {
        $where_clause = 'WHERE ' . join(' AND ', @where_clauses);
    }

    # specify the actual fields to prevent values from being overridden
    #
    # because this is a LEFT JOIN, unless we specify which fields we are interested in,
    # a field like say 'policy_name_crc' will appear with a value of 'NULL' for a table that has no 'policy_name_crc' field
    # in our case for tables DCC.LANGUAGE_ENCODINGS and DCC.ENCODINGS
    #
    # that 'NULL' will override the actual value of 'policy_name_crc' that was previously extracted from DCC.ACCOUNTS
    my $sth = $dbh->prepare(<<EndQuery);
SELECT A.*,
       X.flg_is_trust_xff,
       L.lang_id, L.convert_order, L.flg_visible,
       E.encoding_id, E.encoding, E.encoding_type, E.display_priority, E.icontrol_name,
       R.flg_ajax_enabled,
       EXISTS
           (SELECT flg_identify_device FROM DCC.ACCOUNT_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES B
            WHERE B.policy_name_crc = A.policy_name_crc AND flg_identify_device > 0)
       flg_persistent_bf_obj
FROM DCC.ACCOUNTS AS A
LEFT JOIN DCC.ACCOUNT_XFF AS X
   ON (A.policy_name_crc = X.policy_name_crc)
LEFT JOIN DCC.LANGUAGE_ENCODINGS AS L
   ON (A.lang_set_id = L.lang_set_id)
LEFT JOIN DCC.ENCODINGS AS E
   ON (L.encoding_id = E.encoding_id)
LEFT JOIN DCC.ACCOUNT_ALTERNATE_RESPONSE_SETTINGS AS R
   ON (A.policy_name_crc = R.policy_name_crc)
$where_clause
EndQuery

    my ($deep_object_params_count) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id, F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_DEEP_OBJECT());
SELECT count(*)
  FROM PLC.PL_PARAM_ATTRIBUTES
  WHERE policy_id = ?
  AND object_serialization_style  = ?
EndQuery

    my ($mixed_type_array_params_count) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id, F5::ASMConfig::ParamValueType::PARAM_VALUE_OPENAPI_ARRAY_VALUE());
SELECT count(*)
  FROM PLC.PL_PARAM_ATTRIBUTES
  WHERE policy_id = ?
  AND key_input_type  = ?
EndQuery

    $sth->execute(@bind_values);
    while (my $rec = $sth->fetchrow_hashref) {
        $rec->{flg_is_trust_xff} ||= 0;
        $rec->{flg_ajax_enabled} ||= 0;
        $rec->{encoding_type} ||= 1;
        $rec->{flg_mask_sensitive} ||= 0;
        $rec->{flg_has_deep_object_params} = $deep_object_params_count ? 1 : 0;
        $rec->{flg_has_openapi_array_params} = $mixed_type_array_params_count ? 1 : 0;

        my $is_policy_passive = $rec->{flg_passive_mode} || 0;

        $self->insert_row(dbh => $dbh,
                          table_name => 'DCC.ENFORCER_ACCOUNTS',
                          is_policy_passive => $is_policy_passive,
                          rec => $rec,
                          do_replace => 1,
                          );
    }
}

#--------------------------------------------------------------------------------

sub prepare_data_protection {
    my($self) = @_;
    require F5::PrepareConf::DataProtection;
    return F5::PrepareConf::DataProtection->new(dbh => $self->{dbh},
                                                logger => $self->{logger})->prepare();
}

#--------------------------------------------------------------------------------

sub prepare_credential_stuffing_file {
    my($self) = @_;

    my $flg_credential_stuffing_subscription = F5::License::is_credential_stuffing_licensed();
    my $credential_filename = $FACTORY_CREDENTIALS_BLOB_FILE;
    my $table_name = 'DCC.CREDENTIAL_STUFFING_UPDATE_FILE';

    $self->empty_table($table_name);

    my $rec = { credential_filename => $credential_filename, flg_credential_stuffing_subscription => $flg_credential_stuffing_subscription };
    $self->insert_row(table_name => $table_name, rec => $rec);

    return RET_OK;
}

#--------------------------------------------------------------------------------

sub get_encoding {
    return 'utf-8'
}

#--------------------------------------------------------------------------------

sub parse_response_page_headers {
    my ($self, %params) = @_;

    local $HTTP::Headers::TRANSLATE_UNDERSCORE;

    $params{response_headers} =~ s/(?:\r\n)*$//g;
    $params{response_headers}.= "\r\n\r\n";

    my $parsed_response;
    eval {$parsed_response = HTTP::Response->parse($params{response_headers})};
    if (($@) ||
        (! $parsed_response)||
        (! defined $parsed_response->code) ) {
        die { error_message => "failed to parse response headers - please check response page"};
    }

    my %headers;
    for my $header_name ($parsed_response->header_field_names) {
        $headers{$header_name} = $parsed_response->header($header_name);
    }

    if (! $headers{'Content-Type'})  {
        $headers{'Content-Type'} = "text/html";
    }

    my $encoding = $self->get_encoding(%params);
    if ($headers{'Content-Type'} !~ /;\s*charset=/i && $encoding) {
        $headers{'Content-Type'} .= "; charset=$encoding";
    }

    return { response_headers => \%headers, response_code => $parsed_response->code };
}

#----------------------------------------------------------------------

sub prepare_deception_response_pages {
    my ($self) = @_;

    $self->empty_table('DCC.DECEPTION_RESPONSES');
    $self->empty_table('DCC.DECEPTION_RESPONSE_HEADERS');

    return unless F5::License::is_waf_licensed();

    my $logger = $self->{logger};
    $logger->debug("prepare_deception_response_pages, start") if $logger;

    my $plc_sth = $self->{dbh}->prepare("SELECT * FROM PLC.DECEPTION_RESPONSES");
    $plc_sth->execute();
    my $responses = $plc_sth->fetchall_arrayref({});

    for my $response (@$responses) {
        $self->insert_row(table_name => 'DCC.DECEPTION_RESPONSES',
                          rec => $response);

        my $headers = { Connection => 'close'};

        if ($response->{response_headers}) {
                my $parsed_response;
                eval {
                    $parsed_response = $self->parse_response_page_headers(response_headers => $response->{response_headers});

                };
                if (my $err = $@) {
                    $logger->error($err->{error_message});
                    next;
                }
                $headers = { %$headers, %{$parsed_response->{response_headers}} };
        }

        for my $header_name (keys %$headers) {
            $self->insert_row(table_name => 'DCC.DECEPTION_RESPONSE_HEADERS',
                              rec => { deception_response_id => $response->{deception_response_id},
                                       header_name => $header_name,
                                       header_value => $headers->{$header_name}});
        }
    }

    $logger->debug("prepare_deception_response_pages, end") if $logger;
    return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_deception_traps {
    my ($self, %params) = @_;

    # empty tables in case awaf is not licensed
    my @tables = (
                   qw(
                      DECEPTION_TRAPS
                      DECEPTION_TRAP_CONDITIONS
                      DECEPTION_TRAP_INJECTIONS
                      DECEPTION_TRAP_INJECTION_HEADERS
                     )
                 );

    for my $table (@tables) {
        $self->empty_table("DCC.$table");
    }
    $self->empty_table('DCC.DECEPTION_TRAPS_ENTITIES');

    return unless F5::License::is_waf_licensed();

    my $logger = $self->{logger};
    $logger->debug("prepare_deception_traps, start") if $logger;

    my %source_to_entity_type = (
          F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_COOKIE_NAME() => F5::ASMConfig::DeceptionTrapEntityType::DECEPTION_TRAP_ENTITY_TYPE_COOKIE(),
          F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_PARAM_NAME() => F5::ASMConfig::DeceptionTrapEntityType::DECEPTION_TRAP_ENTITY_TYPE_PARAM(),
          F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_HEADER_NAME() => F5::ASMConfig::DeceptionTrapEntityType::DECEPTION_TRAP_ENTITY_TYPE_HEADER(),
          F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_URL() => F5::ASMConfig::DeceptionTrapEntityType::DECEPTION_TRAP_ENTITY_TYPE_URL(),
          F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_USERNAME() => F5::ASMConfig::DeceptionTrapEntityType::DECEPTION_TRAP_ENTITY_TYPE_LOGIN_REQUEST(),
          F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_SIGNATURE() => F5::ASMConfig::DeceptionTrapEntityType::DECEPTION_TRAP_ENTITY_TYPE_SIGNATURE(),
    );

    my $condition_handler = sub {
        my ($self, %params) = @_;

        my $row = $params{row};

        my $entity_key;
        my $case_insensitive_entity_key;
        if ($row->{condition_source} eq F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_SIGNATURE()) {
            $entity_key = $row->{condition_value};
        }
        elsif ($row->{condition_source} eq F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_USERNAME()) {
            $entity_key = 0;
        }
        elsif (my $source = $source_to_entity_type{$row->{condition_source}}) {
            if ( ($source eq F5::ASMConfig::DeceptionTrapEntityType::DECEPTION_TRAP_ENTITY_TYPE_PARAM()
                   || $source eq F5::ASMConfig::DeceptionTrapEntityType::DECEPTION_TRAP_ENTITY_TYPE_URL())
                  && $row->{condition_value} =~ /[A-Z]/) {
                $case_insensitive_entity_key = F5::CRC::get_crc32(lc($row->{condition_value}));
            }
            $entity_key = F5::CRC::get_crc32($row->{condition_value});
        }

        if (defined $entity_key) {
            $self->insert_row(table_name => 'DCC.DECEPTION_TRAPS_ENTITIES',
                              rec => {
                                      entity_key => $entity_key,
                                      entity_type => $source_to_entity_type{$row->{condition_source}},
                                      deception_trap_id => $row->{deception_trap_id},
                                     },
                             );
        }

        if (defined $case_insensitive_entity_key) {
            $self->insert_row(table_name => 'DCC.DECEPTION_TRAPS_ENTITIES',
                              rec => {
                                      entity_key => $case_insensitive_entity_key,
                                      entity_type => $source_to_entity_type{$row->{condition_source}},
                                      deception_trap_id => $row->{deception_trap_id},
                                     },
                             );
        }
    };

    for my $table (@tables) {
        $self->prepare_generic(0, 0,
                                   {
                                    policy_part => {
                                                    policy_independent => 1,
                                                    tables => {
                                                               source_tables => [ "PLC.$table" ],
                                                               dest_tables => [ "DCC.$table" ],
                                                              },
                                                    row_handler => $table eq 'DECEPTION_TRAP_CONDITIONS' ? $condition_handler : undef,
                                                   },
                                   });
   }

    $logger->debug("prepare_deception_traps, end") if $logger;
    return RET_OK;
}

#----------------------------------------------------------------------
#
# This is needed to verify that an MCP object will be inserted/updated only when necessary.
# The super-paranoid single record query is used to determine if a hidden record (under
# /Common/asm-hidden, for example) already exists with identical values.

sub update_mcp_object_if_changed {
    my($self, %params) = @_;
    my $mcp_object_type = $params{mcp_object_type};
    my $key_fields = $params{key_fields} || [ 'name' ];
    my $rec = $params{rec};

    my %search_criteria;
    for my $key_field (@$key_fields) {
        $search_criteria{$key_field} = $rec->{$key_field};
    }

    my $query_result = F5::ASM::COAPI::Query->new($mcp_object_type, \%search_criteria,
                                                  direct_coapi_filter_key => $key_fields->[0]);

    my $existing_mcp_rec;
    if ($query_result && @$query_result) {
        $existing_mcp_rec = $query_result->[0];

        my $is_update_needed = 0;
        for my $k (keys %$rec) {
            # Smart match is needed for nested arrays (such as domain_names)
            if (! ($rec->{$k} ~~ $existing_mcp_rec->{$k})) {
                $is_update_needed = 1;
                last;
            }
        }
        if (! $is_update_needed) {
            return;
        }
    }

    if (! $existing_mcp_rec) {
        $existing_mcp_rec = F5::ASM::COAPI::Object->new($mcp_object_type);
    }
    for my $k (keys %$rec) {
        $existing_mcp_rec->set_value($k => $rec->{$k});
    }
    $existing_mcp_rec->save();
    $self->{persist_to_mcp} = 1;
    return $existing_mcp_rec;
}

#----------------------------------------------------------------------

sub persist_to_mcp {
    my ($self) = @_;
    if ($self->{persist_to_mcp}) {
        eval {
            F5::ASM::COAPI::SaveAll(logger => $self->{logger});
            delete $self->{persist_to_mcp};
        };
        if (my $err = $@) {
            $self->{logger}->warn("Could not persist MCP state.  Error: $err") if $self->{logger};
        }
    }
    return;
}

#----------------------------------------------------------------------

sub prepare_client_side_js {
    my ($self) = @_;
    my $logger = $self->{logger};
    my $obfuscator_base_filename = "obfuscator.conf.all.js.test.json";
    my $obfuscated_dir_name = "obfuscated";

    if (!F5::BigipUtils::get_feature_value('prepare_client_side_js')) {
        return;
    }

    if (! -d $CLIENT_SIDE_JS_SRC_DIR) {
        $logger->info("Client-side JS source dir ($CLIENT_SIDE_JS_SRC_DIR) doesn't exist -- skipping");
        return;
    }

    # get name of the timestamped directory
    my @timestamped_client_side_js_dirs = sort glob("$CLIENT_SIDE_JS_SRC_DIR/*");
     if (!@timestamped_client_side_js_dirs) {
        $logger->info("Client-side JS source dir ($CLIENT_SIDE_JS_SRC_DIR) is empty -- skipping") if $logger;
        return;
    }
    my $timestamp_dir_name = basename($timestamped_client_side_js_dirs[-1]);

    # check whether source and destination dirs differ (comparing obfuscator json md5)
    my $should_write_to_db = 0;
    my $src_obfuscator_file = "$CLIENT_SIDE_JS_SRC_DIR/$timestamp_dir_name/$obfuscator_base_filename";
    my $dest_obfuscator_file = "$CLIENT_SIDE_JS_DEST_DIR/$timestamp_dir_name/$obfuscator_base_filename";
    if (! -f $dest_obfuscator_file) {
        $logger->debug("Prepare client-side JS: No previous destination Client-side JS files") if $logger;
        $should_write_to_db = 1;
    }
    else {
        my $src_obfuscator_md5 = F5::Utils::File::get_file_md5("$src_obfuscator_file");
        my $dest_obfuscator_md5 = F5::Utils::File::get_file_md5("$dest_obfuscator_file");
        if ($src_obfuscator_md5 ne $dest_obfuscator_md5) {
            $logger->debug("Prepare client-side JS: source dir ($CLIENT_SIDE_JS_SRC_DIR) and destination dir ($CLIENT_SIDE_JS_DEST_DIR) obfuscators differ ($CLIENT_SIDE_JS_SRC_DIR/$timestamp_dir_name/$obfuscator_base_filename).");
            $should_write_to_db = 1;
        }
    }

    if (! $should_write_to_db) {
        $logger->debug("Prepare client-side JS: source dir ($CLIENT_SIDE_JS_SRC_DIR) "
            . "and destination dir ($CLIENT_SIDE_JS_DEST_DIR) obfuscators are identical. No write is required.");
        return;
    }

    # copy client-side json files to destination folder shared with bd
    my $copied_files = F5::Utils::File::copy_files_recursively(
        from_dir => $CLIENT_SIDE_JS_SRC_DIR,
        to_dir => $CLIENT_SIDE_JS_DEST_DIR);

    # empty DCC.CLIENT_SIDE_OBFUSCATED_JS
    $self->empty_table('DCC.CLIENT_SIDE_OBFUSCATED_JS');
    
    # fill DCC.CLIENT_SIDE_OBFUSCATED_JS with rows for each client-side js dir
    my $obfuscator_json_filename = "$CLIENT_SIDE_JS_DEST_DIR/$timestamp_dir_name/$obfuscator_base_filename";
    my $obfuscator_json = File::Slurp::read_file($obfuscator_json_filename);
    my $obfuscator_rec;
    eval {
        $obfuscator_rec = JSON::XS->new->utf8->decode($obfuscator_json);
    };
    if (my $err = $@) {
        $logger->error("Prepare client-side JS: json parsing error on file $obfuscator_json_filename: $err") if $logger;
    }

    my $generation_time = time();
    if ($obfuscator_rec && @$obfuscator_rec && $obfuscator_rec->[0] && %{$obfuscator_rec->[0]}) {
        $generation_time = $obfuscator_rec->[0]->{activationEpoch};
    }


    for my $version ( @{$obfuscator_rec->[0]->{versions}} ) {
        my $key = $version->{key};
        my $dirname = $version->{id};
        my $full_dirname = "$CLIENT_SIDE_JS_DEST_DIR/$timestamp_dir_name/$obfuscated_dir_name/$dirname";

        $logger->trace("Prepare client-side JS: Adding DCC.CLIENT_SIDE_OBFUSCATED_JS row: dirname=$full_dirname, datacrypt_key=$key, generation_time=$generation_time, is_latest_update=1") if $logger;

        $self->insert_row(dbh => $self->{dbh},
                    table_name => 'DCC.CLIENT_SIDE_OBFUSCATED_JS',
                            rec => {
                                    dirname => $full_dirname,
                                    datacrypt_key => $key,
                                    generation_time => $generation_time,
                                    is_latest_update => 1,
                                },
                        );
    }
    return 1;
}

#----------------------------------------------------------------------

sub prepare_client_side_captcha {
    my ($self) = @_;

    if (!F5::BigipUtils::get_feature_value('client_side_captcha')) {
        return;
    }

    if (-e $F5::Captcha::CAPTCHA_OUT_DIR) {
        # empty DCC.CLIENT_SIDE_CAPTCHA
        $self->empty_table('DCC.CLIENT_SIDE_CAPTCHA');

        opendir(my $dirh, $F5::Captcha::CAPTCHA_OUT_DIR);
        my @sorted_captcha_dirs = reverse sort readdir($dirh);
        my $active_captcha_subdir = $sorted_captcha_dirs[0];
        my $inactive_captcha_subdir = $sorted_captcha_dirs[1];
        $self->{logger}->debug("Active captcha subdir: $active_captcha_subdir, inactive captcha subdir: $inactive_captcha_subdir.") if $self->{logger};
        closedir $dirh;

        F5::DbUtils::insert_row(dbh => $self->{dbh},
                                logger => $self->{logger},
                                table_name => 'DCC.CLIENT_SIDE_CAPTCHA',
                                rec => {
                                         dirname => "$F5::Captcha::CAPTCHA_OUT_DIR/$active_captcha_subdir",
                                         is_active => 1,
                                       }
                               );

        if (index($inactive_captcha_subdir, '.') == -1) {
            F5::DbUtils::insert_row(dbh => $self->{dbh},
                                    logger => $self->{logger},
                                    table_name => 'DCC.CLIENT_SIDE_CAPTCHA',
                                    rec => {
                                             dirname => "$F5::Captcha::CAPTCHA_OUT_DIR/$inactive_captcha_subdir",
                                             is_active => 0,
                                           }
                                   );
        }

        F5::DbUtils::delete_unreferenced_files(dbh          => $self->{dbh},
                                               logger       => $self->{logger},
                                               db_file_refs => [
                                                                { table_name => 'DCC.CLIENT_SIDE_CAPTCHA',
                                                                  field      => 'dirname',
                                                                },
                                                               ],
                                               dir          => "$F5::Captcha::CAPTCHA_OUT_DIR");
    }
}

#----------------------------------------------------------------------

sub prepare_display_vs_name {
    my ($self, %params) = @_;

    my $display_vs_names = $params{display_vs_names};

    my $dbh = $self->{dbh};
    my $table_name = 'DCC.VS_ATTRIBUTES';

    $self->empty_table($table_name);

    my $batch_insert = F5::BatchInsert->new(dbh => $dbh,
                                            logger => $self->{logger},
                                           );

    for my $vs_uuid (keys %$display_vs_names) {
        $batch_insert->insert_row(table_name => $table_name,
                                  rec => { display_vs_name => "$display_vs_names->{$vs_uuid}->{display_vs_name}",
                                           vs_name_crc => F5::CRC::get_crc32($vs_uuid),
                                         },
                         );
    }

    $batch_insert->flush;
}

#======================================================================

1;