package F5::PrepareConf::BotDefense;

use strict;
use warnings;
use File::Basename;

use base 'F5::PrepareConf::Base';
use F5::BigipUtils;

use F5::ASM::COAPI;

my $PROFILE_BOTDEFENSE_ASM_PREFIX = "ASM_";

#----------------------------------------------------------------------
# Update single profile of Bot Profile in MCP

my %DEFAULT_BOT_DEFENSE_ATTRIBS = (
                                    send_javascript_fingerprint => 0,
                                    send_javascript_efoxy => 0,
                                    send_javascript_challenge => 0,
                                    send_brute_force_challenge => 0,
                                    inject_javascript => 0,
                                    clientside_in_use => 0,
                                    persistent_data_validity_period => 0,
                                   );

our $AGGREGATED_PROFILE_NAME = '/Common/bot_defense_asm_aggregated';

sub update_mcp_bot_defense_profiles {
    my($self, %params) = @_;
    my $policies_by_policy_name_crc = $params{policies_by_policy_name_crc};
    my @updated_objs;

    # bypass_validation before creating Object->new() or Query->new()
    F5::ASM::COAPI::bypass_validation();

    my $bot_defense_attribs_by_policy_name_crc = $self->get_bot_defense_attribs_for_policies(policies_by_policy_name_crc => $policies_by_policy_name_crc);
    for my $policy_name_crc (keys %$policies_by_policy_name_crc) {
        my $bot_defense_profile_name = $policies_by_policy_name_crc->{$policy_name_crc}->{bot_defense_profile_name};
        my $bot_defense_attribs = $bot_defense_attribs_by_policy_name_crc->{$policy_name_crc};

        $self->{logger}->debug("Saving bot defense profile $bot_defense_profile_name") if $self->{logger};
        $self->{logger}->trace($bot_defense_attribs) if $self->{logger};

        # create or modify the mcp object
        my $obj = F5::ASM::COAPI::Object->new('profile_bot_defense_asm');

        $obj->set_value('name', $bot_defense_profile_name);
        for my $k (keys %DEFAULT_BOT_DEFENSE_ATTRIBS) {
            my $value = defined($bot_defense_attribs->{$k}) ? $bot_defense_attribs->{$k} : $DEFAULT_BOT_DEFENSE_ATTRIBS{$k};
            $obj->set_value($k, $value);
        }
        push (@updated_objs, $obj);
    }

    # TBD: need to change aggregated attribs to use ref count
    # updated by each asm policy.
    my $obj = F5::ASM::COAPI::Object->new('profile_bot_defense_asm');
    $obj->set_value('name', $AGGREGATED_PROFILE_NAME);
    $obj->set_value('send_javascript_fingerprint', 1);
    $obj->set_value('send_javascript_efoxy', 1);
    $obj->set_value('send_javascript_challenge', 1);
    $obj->set_value('send_brute_force_challenge', 1);
    $obj->set_value('inject_javascript', 1);
    $obj->set_value('clientside_in_use', 0);
    $obj->set_value('persistent_data_validity_period', 0);

    push (@updated_objs, $obj);

    eval {
        F5::ASM::COAPI::TransactionalSave(\@updated_objs);
    };
    if ($@) {
        # an error may happen the profile could not be created, for example when the license expired
        $self->{logger}->info("cannot send updated objects to mcp: $@ $F5::COAPI::error") if $self->{logger};
        return;
    }
}

#----------------------------------------------------------------------

sub get_bot_defense_attribs_for_policies {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $policies_by_policy_name_crc = $params{policies_by_policy_name_crc};
    my $cs_ajax_enabled = 0;

    #Ajax request hook
    my ($single_page_application) = $dbh->selectrow_array(<<EndQuery, undef, "single_page_application");
SELECT current_value
    FROM DCC.INTERNAL_CONFIG_ITEMS
WHERE field_name = ?
EndQuery

    if ($single_page_application) {
        $cs_ajax_enabled = 1;
    }

    my @policy_name_crcs = keys %$policies_by_policy_name_crc;
    # Add all asm white ips that has Ignore Anomaly Detection
    my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => {'a.policy_name_crc' => \@policy_name_crcs});
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    #Brute Force
    #CSRF
    #AJAX Blocking Page
    my $bot_defense_attribs_by_policy_name_crc = $dbh->selectall_hashref(<<EndQuery, 'policy_name_crc', undef, @where_values);
SELECT ($cs_ajax_enabled && bf.policy_name_crc) AS send_brute_force_challenge,

        (bf.policy_name_crc || csrf.flg_is_enabled || a.flg_ajax_enabled) AS clientside_in_use,

        a.policy_name_crc AS policy_name_crc

  FROM DCC.ENFORCER_ACCOUNTS a
  JOIN DCC.ACCOUNT_CSRF csrf ON csrf.policy_name_crc = a.policy_name_crc
  LEFT JOIN DCC.ACCOUNT_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES bf ON bf.policy_name_crc = a.policy_name_crc
 WHERE $where_clause
EndQuery

    if ($cs_ajax_enabled) {
        for my $policy_name_crc (@policy_name_crcs) {
            $bot_defense_attribs_by_policy_name_crc->{$policy_name_crc}->{inject_javascript} ||= $bot_defense_attribs_by_policy_name_crc->{$policy_name_crc}->{send_brute_force_challenge};
        }
    }

    return $bot_defense_attribs_by_policy_name_crc;
}

#----------------------------------------------------------------------

sub refresh_whitelist_ips {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $policies_by_policy_name_crc = $params{policies_by_policy_name_crc};

    # bypass_validation before creating Object->new() or Query->new()
    F5::ASM::COAPI::bypass_validation();

    my $obj;
    my @updated_objs;

    for my $policy (values %$policies_by_policy_name_crc) {
        $obj = F5::ASM::COAPI::Object->new("bot_defense_asm_white_ip");
        # delete all existing ips for the specified policies
        $obj->{profile_name} = $policy->{bot_defense_profile_name};
        $obj->delete_on_save();
        push(@updated_objs, $obj);
    }

    my @policy_name_crcs = keys %$policies_by_policy_name_crc;
    # Add all asm white ips that has Ignore Anomaly Detection
    my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => {policy_name_crc => \@policy_name_crcs});
    my $where_clause = $constructed_where_clause->{where_clause};
    my @where_values = @{$constructed_where_clause->{where_values}};

    my($white_ips_query) = $dbh->prepare(<<EndQuery);
SELECT ip_address, ip_mask, route_domain, ignore_flags_mask, policy_name_crc
FROM DCC.ACCOUNT_IP_WHITELIST
WHERE $where_clause
EndQuery

    $white_ips_query->execute(@where_values);

    while (my $row = $white_ips_query->fetchrow_hashref ) {

        # check that Anomaly Detection is set to ignore on this white ip
        if ($row->{ignore_flags_mask} & 0x1) {

            $obj = F5::ASM::COAPI::Object->new("bot_defense_asm_white_ip");
            $obj->{address} = $row->{ip_address};
            if ($row->{route_domain} != 0) {
                $obj->{address} .= "%".$row->{route_domain};
            }
            $obj->{netmask} = $row->{ip_mask};
            $obj->{profile_name} = $policies_by_policy_name_crc->{$row->{policy_name_crc}}->{bot_defense_profile_name};

            # add new white list ip
            push(@updated_objs, $obj);
        }
    }

    if (!@updated_objs) {
        $self->{logger}->error("no mcp bot_defense_asm_white_ip objects were updated") if $self->{logger};
        return;
    }

    eval {
        F5::ASM::COAPI::TransactionalSave(\@updated_objs);
    };
    if ($@) {
        # an error may happen the profile could not be created, for example when the license expired
        $self->{logger}->info("cannot send updated objects to mcp: $@ $F5::COAPI::error") if $self->{logger};
        return;
    }
}

#----------------------------------------------------------------------

# Iterate through all the ASM policies and update the Bot Profiles in MCP
our $BASE_PROFILE_NAME = '/Common/bot_defense_asm';

sub prepare_bot_defense_profiles {
    my($self, %params) = @_;

    return if F5::BigipUtils::is_docker();

    if (F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade()) {
        # Secondary blades will receive bot defense profile MCP updates from the primary via MCP itself
        $self->{logger}->debug("Skipping bot defense profile update on secondary blade") if $self->{logger};
        return;
    }

    $self->{logger}->info("Updating bot defense profiles") if $self->{logger};

    # bypass_validation before creating Object->new() or Query->new()
    F5::ASM::COAPI::bypass_validation();

    # Create the base profile if it does not exist.
    # This may happen when renewing an expired license, because we try to update mcp
    # after the base config was loaded, but before the full config load.
    my $check_base_bot_defense_profile_existence = F5::ASM::COAPI::Query->new('profile_bot_defense_asm', {name => $BASE_PROFILE_NAME});

    if (!$check_base_bot_defense_profile_existence || !@$check_base_bot_defense_profile_existence) {
        $self->{logger}->info("Base bot defense profile \"$BASE_PROFILE_NAME\" not found -- creating it") if $self->{logger};
        my $obj = F5::ASM::COAPI::Object->new('profile_bot_defense_asm');
        $obj->set_value('name', $BASE_PROFILE_NAME);
        eval {
            $obj->save;
        };
        if ($@) {
            $self->{logger}->info("Cannot save base bot defense profile: $@ $F5::COAPI::error") if $self->{logger};
            return;
        }
    }

    my $policies_by_policy_name_crc = $self->get_policies_by_policy_name_crc(%params);
    return unless %$policies_by_policy_name_crc;

    #TODO: Only update these items if a relevant table changed.
    $self->update_mcp_bot_defense_profiles(policies_by_policy_name_crc => $policies_by_policy_name_crc);
    $self->refresh_whitelist_ips( policies_by_policy_name_crc => $policies_by_policy_name_crc);

}

#----------------------------------------------------------------------

sub get_policies_by_policy_name_crc {
    my ($self, %params) = @_;

    my $policies;
    my $dbh = $self->{dbh};

    if ($params{policy} && $params{policy}->{policy_name_crc} && $params{policy}->{policy_name}) {
        $policies = [ $params{policy} ];
    } else {
        $policies = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT policy_name_crc, policy_name
FROM DCC.ENFORCER_ACCOUNTS
EndQuery
    }

    # if no policies were found then return an empty hashref 
    return {} unless @$policies;
    my %mcp_folders = map { ( $_->{name} => $_ ) } @{F5::ASM::COAPI::Query->new('folder')};
    
    for my $policy (@$policies) {
        my $policy_dir = dirname($policy->{policy_name});
        if(exists $mcp_folders{$policy_dir}){
            $policy->{bot_defense_profile_name} = get_bot_defense_profile_name(policy_name => $policy->{policy_name});
        }
    }

    my %policies_by_policy_name_crc = map { $_->{policy_name_crc} => $_ } grep { exists $_->{bot_defense_profile_name} } @$policies;

    return \%policies_by_policy_name_crc;
}

#----------------------------------------------------------------------

sub get_bot_defense_profile_name {
    my(%params) = @_;

    my $bot_defense_profile_name = $params{policy_name};
    # add PROFILE_BOTDEFENSE_ASM_PREFIX between the folder name & the policy name
    $bot_defense_profile_name =~ s|^(.*)/(.*?)$|$1/$PROFILE_BOTDEFENSE_ASM_PREFIX$2|;

    return $bot_defense_profile_name;
}

#----------------------------------------------------------------------

1;

