package F5::PrepareConf::Checksums;

use strict;
use warnings;

use F5::DbUtils;
use Memoize;

memoize('get_last_update_field_name');

# This class checks that database table contents have not changed since the last PrepareConf run
# to skip unchanged tables.
# This was originally done by using MySQL checksums, but those were not performant in more recent
# versions of MySQL (see https://jira.pdsea.f5net.com/browse/ASMMNG-1115), and do not exist in sqlite.
# Instead, this class checks that the maximum "last update" has not changed and the total row count
# remains the same (to check for deletions). This presumes that last update value always ascends.
# The class name "Checksums" has been retained for compatibility even though checksums
# are not used.
#----------------------------------------------------------------------

our %last_update_fields_by_table = ( 'PLC.NEGSIG_SIGNATURES' => 'last_enforcement_change',
                                   );
our @default_last_update_field_names = ('last_updated',
                                        'last_update',
                                       );

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;

  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{dbh} = $params{dbh};
  $self->{logger} = $params{logger};
  $self->{checksums_policy_id} = $params{checksums_policy_id};
  $self->{changed_tables} = {};
}

#----------------------------------------------------------------------

sub have_tables_changed {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};

  my $policy_id = $params{policy_id} || 0;

  my @all_tables;
  if (ref($params{tables}) eq 'ARRAY') {
    push(@all_tables, @{$params{tables}});
  }
  elsif (ref($params{tables}) eq 'HASH') {
    if (defined $params{tables}->{policy_tables}) {
      push(@all_tables, @{$params{tables}->{policy_tables}});
    }
    if (defined $params{tables}->{tables}) {
      push(@all_tables, @{$params{tables}->{tables}});
    }
  }

  my $have_tables_changed = 0;
  for my $table_name (@all_tables) {
      $logger->debug("Checking checksum for table $table_name") if $logger;
      my $has_current_table_changed = 0;
      my $last_update_field = $self->get_last_update_field_name(table_name => $table_name);
      if (! $last_update_field) {
          # Even without a last_update_field, if the table was empty before and now, then consider the table unchanged.
          $last_update_field = '1';
      }

      my $saved_checksum = $dbh->selectrow_hashref(<<EndQuery, undef, $table_name, $policy_id);
SELECT *
  FROM DCC.TABLE_CHECKSUMS_LAST_UPDATE
 WHERE table_name = ?
   AND policy_id = ?
EndQuery

      my $source_policy_id = $policy_id ? $self->{checksums_policy_id} : 0; #global checksums are still stored on policy_id 0
      my $where_fields = {};
      if ($source_policy_id) {
          $where_fields->{policy_id} = $source_policy_id;
      }
      my $where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);

      my $current_checksum = $dbh->selectrow_hashref(<<EndQuery, undef, @{$where_clause->{where_values}});
SELECT MAX($last_update_field) as max_last_update, COUNT(*) AS row_count
  FROM $table_name
 WHERE $where_clause->{where_clause}
EndQuery
      $current_checksum->{max_last_update} ||= 1;

      if (!defined $saved_checksum || $saved_checksum->{max_last_update} ne $current_checksum->{max_last_update} ||
          $saved_checksum->{row_count} != $current_checksum->{row_count}) {

          $has_current_table_changed = 1;
      }
      elsif ($last_update_field eq '1' && $current_checksum->{row_count} != 0) {
          $has_current_table_changed = 1;
      }

      if ($has_current_table_changed) {
          $logger->debug("Table $table_name has differing checksum") if $logger;
          $have_tables_changed = 1;
          $self->{changed_tables}->{$policy_id}->{$table_name} = $current_checksum;
      }
  }

  return $have_tables_changed;
}

#----------------------------------------------------------------------

sub get_last_update_field_name {
    my($self, %params) = @_;
    my $table_name = $params{table_name};

    if ($last_update_fields_by_table{$table_name}) {
        return $last_update_fields_by_table{$table_name};
    }
    else {
        my %db_fields = map { ($_ => 1) } F5::DbUtils::get_db_fields($self->{dbh}, $table_name);
        for my $field_name (@default_last_update_field_names) {
            if ($db_fields{$field_name}) {
                return $field_name;
            }
        }
    }
    return;
}

#----------------------------------------------------------------------

sub store {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $logger = $self->{logger};
  my $policy_id = $params{policy_id} || 0;

  if ($self->{changed_tables}->{$policy_id}) {
      for my $table_name (keys %{$self->{changed_tables}->{$policy_id}}) {
          $logger->debug("Updating checksum for table $table_name") if $logger;

          F5::DbUtils::insert_row(dbh => $dbh,
                                  table_name => 'DCC.TABLE_CHECKSUMS_LAST_UPDATE',
                                  do_replace => 1,
                                  rec => {
                                          table_name => $table_name,
                                          policy_id => $policy_id,
                                          %{$self->{changed_tables}->{$policy_id}->{$table_name}},
                                         },
                                 );

      }

      delete $self->{changed_tables}->{$policy_id};
  }
}

#----------------------------------------------------------------------

sub empty_checksums {
    my($self) = @_;

    $self->{dbh}->do("DELETE FROM DCC.TABLE_CHECKSUMS_LAST_UPDATE");
}

#----------------------------------------------------------------------

1;
