package F5::PrepareConf::DataProtection;

use strict;
use warnings;

use base 'F5::PrepareConf::Base';
use F5::DbUtils;
use F5::PrepareConf::Constants qw(ERRMSG_DBH_LOST RET_OK RET_NOT_OK FALSE TRUE);
use F5::DataProtection;
use F5::ASMConfig::ASMConfig;
use Digest::MD5;
use F5::DbUtils;
use F5::Utils::File;

my $DATA_PROTECTION_VERSION = 1;

my $MAPPING = { cookie_type => {
                                F5::ASMConfig::DataProtectionCookieType::DATA_PROTECTION_COOKIE_TYPE_MAIN() => 'main',
                                F5::ASMConfig::DataProtectionCookieType::DATA_PROTECTION_COOKIE_TYPE_FRAME() => 'frame',                                
                               },
                scramble_algorithm => {
                                F5::ASMConfig::DataProtectionCookieScrambleType::DATA_PROTECTION_COOKIE_SCRAMBLE_TYPE_NONE() => 'none', 
                                F5::ASMConfig::DataProtectionCookieScrambleType::DATA_PROTECTION_COOKIE_SCRAMBLE_TYPE_AES()  => 'aes', 
                                F5::ASMConfig::DataProtectionCookieScrambleType::DATA_PROTECTION_COOKIE_SCRAMBLE_TYPE_SMK3() => 'smk3', 
                                F5::ASMConfig::DataProtectionCookieScrambleType::DATA_PROTECTION_COOKIE_SCRAMBLE_TYPE_BC5()  => 'bc5', 
                               },
                hash_algorithm => {
                                   F5::ASMConfig::DataProtectionCookieHashType::DATA_PROTECTION_COOKIE_HASH_TYPE_NONE()      => 'none',
                                   F5::ASMConfig::DataProtectionCookieHashType::DATA_PROTECTION_COOKIE_HASH_TYPE_CRC32()     => 'crc32',
                                   F5::ASMConfig::DataProtectionCookieHashType::DATA_PROTECTION_COOKIE_HASH_TYPE_CRC64()     => 'crc64',
                                   F5::ASMConfig::DataProtectionCookieHashType::DATA_PROTECTION_COOKIE_HASH_TYPE_SHA256()    => 'sha256',
                                   F5::ASMConfig::DataProtectionCookieHashType::DATA_PROTECTION_COOKIE_HASH_TYPE_SHA256_8()  => 'sha256_8',
                                   F5::ASMConfig::DataProtectionCookieHashType::DATA_PROTECTION_COOKIE_HASH_TYPE_SHA256_16() => 'sha256_16',
                                   F5::ASMConfig::DataProtectionCookieHashType::DATA_PROTECTION_COOKIE_HASH_TYPE_MCY3()      => 'mcy3',
                                   F5::ASMConfig::DataProtectionCookieHashType::DATA_PROTECTION_COOKIE_HASH_TYPE_MCY2()      => 'mcy2',

                               },
                mac_algorithm => {
                                   F5::ASMConfig::DataProtectionCookieMacType::DATA_PROTECTION_COOKIE_MAC_TYPE_NONE()         => 'none',
                                   F5::ASMConfig::DataProtectionCookieMacType::DATA_PROTECTION_COOKIE_MAC_TYPE_HMAC_SHA256()  => 'hmac-sha256',

                               },
};

#----------------------------------------------------------------------

sub prepare {
    my ($self, %params) = @_;

    my $security_context_changed = $params{security_context_changed};
    my $dbh = $self->{dbh};
    my $table_name = 'DCC.DATA_PROTECTION';     
    local $" = ',';

    $self->create_if_needed();

    my $current_data_protection = $dbh->selectrow_hashref(<<EndQuery);
SELECT * FROM $table_name LIMIT 1
EndQuery

    my @plc_fields = qw(key_gen_time grace_signing_end_time grace_accepting_end_time);
    my @dcc_fields = qw(conf_file key_gen_time grace_conf_file grace_signing_end_time grace_accepting_end_time);

    my $plc_row = $dbh->selectrow_hashref(<<EndQuery);
SELECT @plc_fields
  FROM PLC.DATA_PROTECTION
EndQuery

    my $data_dir       = "$F5::PrepareConf::Base::ACCOUNT_DATA_DIR/data_protection";
    my $conf_file_string = $self->build_conf_file();
    my $md5_sum = Digest::MD5->new->add($conf_file_string)->hexdigest;
    my $filename = "$data_dir/data_protection_$md5_sum";

    if ($security_context_changed || !$current_data_protection->{conf_file} || !(-e $current_data_protection->{conf_file}) || !(-e $filename)) {
        #Create new config file
        F5::DbUtils::delete_unreferenced_files(dbh          => $dbh,
                                               logger       => $self->{logger},
                                               db_file_refs => [
                                                                { table_name => $table_name,
                                                                  field     => 'conf_file',
                                                                },
                                                                { table_name => $table_name,
                                                                  field     => 'grace_conf_file',
                                                                },
                                                               ],
                                               dir          => "$data_dir");

        F5::Utils::File::write_file(filename => $filename,
                                    contents => $conf_file_string,
                                    overwrite => 0,
                                    permissions_mode => 0664,
                                   );
        $plc_row->{conf_file} = $filename;
        if (-e $current_data_protection->{conf_file}) {
            $plc_row->{grace_conf_file} = $current_data_protection->{conf_file};
        }
    }
    else {
        $plc_row->{conf_file} = $current_data_protection->{conf_file};
        $plc_row->{grace_conf_file} = $current_data_protection->{grace_conf_file};
    }

    $self->empty_table($table_name);
    $self->insert_row(table_name => $table_name,
                      fields => \@dcc_fields,
                      rec => $plc_row,);
  
    return RET_OK;
}

#----------------------------------------------------------------------

sub build_conf_file {
    my ($self, %params) = @_;
    my $dbh = $self->{dbh};
    my ($key) = $dbh->selectrow_array(<<EndQuery);
SELECT secure_key
  FROM PLC.DATA_PROTECTION LIMIT 1
EndQuery

    my $conf =<<EndHeader;
[general]
key=$key
version=$DATA_PROTECTION_VERSION
EndHeader

    my $cookie_type_sth = $dbh->prepare(<<EndQuery);
SELECT *
    FROM PLC.DATA_PROTECTION_CONFIGURATION
EndQuery

    $cookie_type_sth->execute();
    while (my $cookie_type_row = $cookie_type_sth->fetchrow_hashref()) {
        my @conf_text;
        push (@conf_text, "[" . $MAPPING->{cookie_type}->{$cookie_type_row->{cookie_type}} ."]");
        my @fields = qw(scramble_algorithm hash_algorithm mac_algorithm);
        for my $field (@fields) {
            push (@conf_text, "$field=" . $MAPPING->{$field}->{$cookie_type_row->{$field}});
            if ($field eq "scramble_algorithm") {
                my $operation_mode = "none";
                if ($MAPPING->{$field}->{$cookie_type_row->{$field}} ne "none") {
                    $operation_mode = "cbc-pdd";
                }
                push (@conf_text, "operation_mode=$operation_mode");
            }
        }
        $conf .= "\n" . join("\n", @conf_text) . "\n";
    }
    
    return $conf;
}

#----------------------------------------------------------------------

sub create_if_needed {
    my($self, %params) = @_;
    my $dbh = $self->{dbh};

    #create default Data Protection data if not-already existant;
    my ($data_protection_id) = $dbh->selectrow_array(<<EndQuery);
SELECT data_protection_id from PLC.DATA_PROTECTION LIMIT 1
EndQuery

    if (!$data_protection_id) {
        $self->{logger}->info("Creating new Data Protection context") if $self->{logger};

        my $key = F5::DataProtection::generate_key();
        my $now = time();
        F5::DbUtils::insert_row(dbh => $dbh,
                                table_name => 'PLC.DATA_PROTECTION',
                                rec => {secure_key => $key,
                                        key_gen_time => $now,
                                        grace_signing_end_time => 0,
                                        grace_accepting_end_time => 0,},
                                last_insert_id => \$data_protection_id,
                               );

    }

    F5::DbUtils::copy_defaults_table(dbh => $dbh,
                                     table_name => "DATA_PROTECTION_CONFIGURATION",
                                     ignore => 1,
                                     insert_values => { data_protection_id => $data_protection_id },
                                     where_values => {  security_level => F5::ASMConfig::DataProtectionSecurityLevel::DATA_PROTECTION_SECURITY_LEVEL_SECURE() },
                                    );

}

#----------------------------------------------------------------------

1;
