package F5::ProtobufDataFile;

use strict;
use warnings;
no warnings qw(redefine);

use File::Basename;
use File::Copy;
use MIME::Base64;

use File::Slurp;
use File::Find;
use F5::BdCfgTables;
use F5::Utils::File;
use Data::MessagePack;
use F5::MessagePack;


my $use_protobuf_dynamic_lib = 0;
eval {
    require Google::ProtocolBuffers::Dynamic;
    Google::ProtocolBuffers::Dynamic->import();
    $use_protobuf_dynamic_lib = 1;
    1;
} or do {
    require Google::ProtocolBuffers;
    Google::ProtocolBuffers->import();
};

our $BD_PROTO_PATH = F5::Deploy::map_path('/usr/share/ts/share/proto/');
our $BD_CONFIG_PROTO_FILENAME = 'bd_config.proto';
my $DATA_FILE_SUFFIX = 'msgp';

my %types_fixup = (
                   'bytes' => {
                               fixup_func => sub {
                                   my($val) = @_;
                                   if (utf8::is_utf8($val)) {
                                       # Treat scalar as bytes, not utf8
                                       # See https://jira.pdsea.f5net.com/browse/MBIPWAF-1980
                                       utf8::encode($val);
                                   }
                                   return $val;
                               },
                              },
                   'string' => {
                                # Several "string" values are actually binary bytes
                                # See https://jira.pdsea.f5net.com/browse/MBIPWAF-1980
                                fixup_func => sub {
                                    my($val) = @_;
                                    if (utf8::is_utf8($val)) {
                                        # Treat scalar as bytes, not utf8
                                        # See https://jira.pdsea.f5net.com/browse/MBIPWAF-1980
                                        utf8::encode($val);
                                    }
                                    return $val;
                                },
                               },
                  );

# Bd::Config::ConfigAction::* constants are loaded by
#     'Google::ProtocolBuffers->parse', below, in 'sub prepare_protobuf_modules',
prepare_protobuf_modules();

#----------------------------------------------------------------------

sub prepare_protobuf_modules {
    if ($use_protobuf_dynamic_lib) {
        my $protobuf = Google::ProtocolBuffers::Dynamic->new($BD_PROTO_PATH);
        $protobuf->load_file('bd_config.proto');
        $protobuf->load_file('bd_config_auto.proto');
        $protobuf->load_file('bd_config_fixed.proto');
        $protobuf->load_file('common.proto');
        $protobuf->map({package => 'bd.config', prefix => "Bd::Config",
                        options => { encode_defaults => 1 }
                       });
    }
    else {
        my $protobuf_cont = File::Slurp::read_file($BD_PROTO_PATH.$BD_CONFIG_PROTO_FILENAME);
        Google::ProtocolBuffers->parse($protobuf_cont, {create_accessors => 1, include_dir => $BD_PROTO_PATH});
    }
}

#----------------------------------------------------------------------

sub convert_single_protobuf_data_file{
    my (%params) = @_;
    my $from_file = $params{from_file};
    my $logger = $params{logger};
    my $protobuf_out_file_name = $params{to_file};
    my $table_def = $params{table_def};

    if (! $protobuf_out_file_name) {
        $protobuf_out_file_name = $from_file;
        $protobuf_out_file_name =~ s/\.$DATA_FILE_SUFFIX/\.dat/g;
    }

    my $protobuf_class_enum = uc $table_def->{name};
    my $protobuf_entry_name = $table_def->{name};
    my $protobuf_class = snake_case_to_camel_case($protobuf_entry_name);
    my $message_protobuf_class_name = 'Bd::Config::'. $protobuf_class;

    my %field_name_to_proto_type;
    for my $field_def (@{$table_def->{message_fields}}) {
        $field_name_to_proto_type{$field_def->{name}} = $field_def->{proto_type};
    }

    $logger->debug("Converting BD config from $from_file") if $logger;
    my $config_data_iterator = F5::MessagePack::Iterator->new(filename => $from_file);
    open(my $outf, '>', $protobuf_out_file_name);
    while (my $config_row = $config_data_iterator->get_next()) {
        my $operation = Bd::Config::ConfigAction::ADD();
        if ($config_row->{action}) {
            my $operation_name =  uc $config_row->{action};
            $operation = Bd::Config::ConfigAction->$operation_name();
        }

        my ($skip_row, $element) = fixup_element(
                                                 element => $config_row->{element},
                                                 field_name_to_proto_type => \%field_name_to_proto_type,
                                                 logger => $logger
                                                );
        next if $skip_row;
        my $message = $message_protobuf_class_name->new($element);
        my $bd_config_element = Bd::Config::BdConfigElement->new({
            config_action => $operation,
            bd_table => Bd::Config::BdTable->$protobuf_class_enum(),
            $protobuf_entry_name => $message,
            key_hash => $config_row->{key_hash},
            full_hash => $config_row->{full_hash},
        });

        my @config_element_array = ( $bd_config_element );
        my $bd_config = Bd::Config::BdConfig->new({ conf_element => \@config_element_array });
        my $encoded_protobuf = $bd_config->encode;
        print $outf pack("N", length($encoded_protobuf));
        print $outf $encoded_protobuf;
    }
    close $outf;
    $logger->debug("Done converting BD config from $from_file") if $logger;

    return $protobuf_out_file_name;
}

#----------------------------------------------------------------------

sub convert_protobuf_data_files{
    my (%params) = @_;
    my $input_dirs = $params{input_dirs};
    my $previous_dirs = $params{previous_dirs} || [];
    my $output_dir_map = $params{output_dir_map};
    my $logger = $params{logger};

    my $dcc_tables = F5::BdCfgTables::get_dcc_table_config();
    my %dcc_tables_by_name = map { ($_->{output_file} => $_) } @$dcc_tables;

    my @existing_files;

    if (@$previous_dirs) {
        find({wanted => sub {
                  my $filename = $File::Find::name;
                  return unless (-f $filename);

                  if ($filename =~ /\.$DATA_FILE_SUFFIX$/) {
                      my $proto_filename = $filename;
                      $proto_filename =~ s/\.$DATA_FILE_SUFFIX$/\.dat/;
                      if (-e $proto_filename) {
                          push(@existing_files,
                               {
                                data_filename => $filename,
                                proto_filename => $proto_filename,
                                data_md5 => F5::Utils::File::get_file_md5($filename),
                               }
                              );
                      }
                  }

              },
              no_chdir => 1,
             },
             @$previous_dirs);
    }

    my %existing_files_by_md5 = map { $_->{data_md5} => $_ } @existing_files;

    my @protobuf_data_filenames;
    find({wanted => sub {
              my $config_file = $File::Find::name;
              return unless (-f $config_file);
              if ($config_file !~ /\.$DATA_FILE_SUFFIX$/) {
                  return;
              }

              my $config_file_base = basename($config_file);
              $config_file_base=~ s/\.$DATA_FILE_SUFFIX$//;
              $config_file_base =~ s/__account_(\d+)$//;
              $config_file_base =~ s/__logprofile_(\d+)$//;
              $config_file_base =~ s/^.*__//;
              my $table_def = $dcc_tables_by_name{$config_file_base};

              my $protobuf_data_filename = $config_file;
              $protobuf_data_filename =~ s/\.$DATA_FILE_SUFFIX/\.dat/g;

              if ($table_def->{proto_index}) {
                  my $config_file_md5 = F5::Utils::File::get_file_md5($config_file);
                  if ($existing_files_by_md5{$config_file_md5}) {
                      $logger->debug("Existing data file matches -- re-using previous file for $config_file") if $logger;
                      File::Copy::copy($existing_files_by_md5{$config_file_md5}->{proto_filename}, $protobuf_data_filename);
                  }
                  else {
                      $protobuf_data_filename = convert_single_protobuf_data_file(
                                                                                  from_file => $config_file,
                                                                                  to_file => $protobuf_data_filename,
                                                                                  table_def => $table_def,
                                                                                  logger => $logger,
                                                                                 );
                      $logger->debug("Successfully created protobuf data file: $protobuf_data_filename") if $logger;
                  }

                  if (-s $protobuf_data_filename) {
                      push(@protobuf_data_filenames, $protobuf_data_filename);
                  }
                  else {
                      $logger->debug("Created empty protobuf data file: $protobuf_data_filename, removing.") if $logger;
                      unlink $protobuf_data_filename;
                  }
              }
              no_chdir => 1,
          }
         },
         @$input_dirs);

    return @protobuf_data_filenames;
}

#----------------------------------------------------------------------

sub fixup_element {
    my (%params) = @_;
    my $element = $params{element};
    my $field_name_to_proto_type = $params{field_name_to_proto_type};
    my $logger = $params{logger};
    my $skip_item = 0;
    my $proto_type;
    for my $field_name (keys %$element) {
        $proto_type = $field_name_to_proto_type->{$field_name};
        my $field_value = $element->{$field_name};
        if ($proto_type && exists($types_fixup{$proto_type})) {
            eval {
                $element->{$field_name} = $types_fixup{$proto_type}->{fixup_func}->($field_value);
            };
            if (my $err = $@) {
                $logger->error("Error encountered while handling config element = $element ($err)") if $logger;
                $skip_item = 1;
                last;
            }
        }
    }
    return ($skip_item, $element);
}


#----------------------------------------------------------------------

sub snake_case_to_camel_case {
    my $rv = shift;
    $rv =~ s/^(\w)/\u$1/;
    $rv =~ s/_(\w)/\u$1/g;
    return $rv;
}

#----------------------------------------------------------------------

sub convert_bd_cfg_manifest {
    my(%params) = @_;
    my $manifest_file = $params{manifest_file};
    my $logger = $params{logger};

    my $source_manifest = YAML::XS::LoadFile($manifest_file);


    my @protobuf_generations;
    for my $generation (@{$source_manifest->{generations}}) {
        # Convert filenames into protobuf filename messages
        for my $file_list_field (qw(fullFiles deltaFiles)) {
            my @protobuf_file_list = ();
            for my $file (@{$generation->{$file_list_field}}) {
                my $protobuf_file = Bd::Config::BdCfgManifestFile->new({ filename => $file->{filename}});
                if ($file->{dependentFiles} && @{$file->{dependentFiles}}) {
                    $protobuf_file->{dependentFiles} = [ map { Bd::Config::BdCfgDependentFile->new({ filename => $_->{filename} }) } @{$file->{dependentFiles}} ];
                }

                push(@protobuf_file_list, $protobuf_file);
            }
            $generation->{$file_list_field} = \@protobuf_file_list;
        }
        push(@protobuf_generations, Bd::Config::BdCfgManifestGeneration->new($generation));
    }
    my $protobuf_manifest = Bd::Config::BdCfgManifest->new({ baseConfigId => $source_manifest->{baseConfigId},
                                                             generations => \@protobuf_generations,
                                                           });

    my $bd_config_element = Bd::Config::BdConfigElement->new({
                                                              bd_cfg_manifest => $protobuf_manifest,
                                                             });
    my @config_element_array = ( $bd_config_element );
    my $bd_config = Bd::Config::BdConfig->new({ conf_element => \@config_element_array });
    my $encoded_protobuf = $bd_config->encode;

    my $protobuf_manifest_file = $manifest_file;
    $protobuf_manifest_file =~ s/\.yaml$/.dat/;
    open(my $outf, '>', $protobuf_manifest_file);
    print $outf pack("N", length($encoded_protobuf));
    print $outf $encoded_protobuf;
    close $outf;

    return $protobuf_manifest_file;
}

#----------------------------------------------------------------------

1;
