package F5::RamCache;

use strict;
use warnings;

use F5::CRC;
use F5::ASM::COAPI;
use F5::Mcp::ProfileType;
use F5::Mcp::PolicyTarget;
use F5::Mcp::PolicyAction;
use F5::Cluster::Utils;

# NOTE: these constants should be equivalent to BD enum ramcache_type_configured_t
use constant {
    NO_CACHE_CONFIGURED => 0,
    RAMCACHE_CONFIGURED => 1,
    WA_CACHE_CONFIGURED => 2,
};

# Corresponding structs in MCP (to reset stats)
use constant {
    CACHE_TYPE_RAMCACHE => 'ramcache_info',
    CACHE_TYPE_WA_CACHE => 'wa_cache_info',
};

# "public"
sub invalidate_cache_by_policy_name {
    my %params = @_;

    invalidate_cache_profiles(vs_cache_profiles => get_vs_cache_profiles(policy_name => $params{policy_name}),
                              logger => $params{logger});
}

sub invalidate_cache_by_profile_name {
    my %params = @_;

    invalidate_cache_profiles(vs_cache_profiles => get_vs_cache_profiles(profile_name => $params{profile_name}),
                              logger => $params{logger});
}

sub invalidate_cache_by_vs_name {
    my %params = @_;

    invalidate_cache_profiles(vs_cache_profiles => get_vs_cache_profiles(vs_name => $params{vs_name}),
                              logger => $params{logger});
}

sub prepare_conf {
    my(%params) = @_;
    
    my $dbh = $params{dbh};
    my $logger = $params{logger};

    my $vs_cache_profiles = get_vs_cache_profiles();

    $logger->debug("Preparing DCC.VS_RAMCACHE") if $logger;
    $dbh->do('DELETE FROM DCC.VS_RAMCACHE');

    for my $vs_name (keys %$vs_cache_profiles) {
        my $ramcache_state = NO_CACHE_CONFIGURED;

        if ($vs_cache_profiles->{$vs_name}->{cache_type}) {
            $ramcache_state |= RAMCACHE_CONFIGURED;

            if ($vs_cache_profiles->{$vs_name}->{cache_type} eq CACHE_TYPE_WA_CACHE) {
                $ramcache_state |= WA_CACHE_CONFIGURED;
            }
        }

        F5::DbUtils::insert_row(dbh => $dbh,
                                table_name => 'DCC.VS_RAMCACHE',
                                rec => {
                                        vs_name => $vs_name,
                                        vs_name_crc => F5::CRC::get_crc32($vs_name),
                                        ramcache_state => $ramcache_state,
                                       });
    }

    # BD enum plugin_type_configured_t
    my %plugin_mask = (
                       F5::Mcp::ProfileType::PROFILE_WEBSECURITY()  => 1, # ASM_HTTP_CONFIGURED
                       F5::Mcp::ProfileType::PROFILE_FTP()          => 4, # FTP_CONFIGURED
                       F5::Mcp::ProfileType::PROFILE_INSTANCE()     => 8, # SMTP_CONFIGURED
                      );

    $logger->debug("Preparing DCC.PLUGIN_USAGE") if $logger;

    my $plugin_state = 0; # NONE_CONFIGURED

    if (F5::BigipUtils::get_traffic_engine() eq 'tmm' && F5::BigipUtils::is_feature_enabled('standalone')) {
        my($policy_count) = $dbh->selectrow_array('SELECT count(*) from PLC.PL_POLICIES');
        $plugin_state  = $policy_count ? 1 : 0;
    }

    for my $vs_info (values %$vs_cache_profiles) {
        $plugin_state |= $plugin_mask{$vs_info->{security_profile_type}}; # regardless of cache state
    }

    return if ( F5::DbUtils::exists_in_table(dbh => $dbh,
                                             field_name => 'plugin_state',
                                             field_value => $plugin_state,
                                             table_name => 'DCC.PLUGIN_USAGE') );


    $dbh->do('DELETE FROM DCC.PLUGIN_USAGE');

    F5::DbUtils::insert_row(dbh => $dbh,
                            table_name => 'DCC.PLUGIN_USAGE',
                            rec => {
                                    plugin_state => $plugin_state
                                   },
                           );
}

# "private"
sub get_vs_cache_profiles {
    my(%params) = @_;
    
    my $policy_name = $params{policy_name};
    my $profile_name = $params{profile_name};
    my $vs_name = $params{vs_name};

    my %vs_profile_filter;

    if ($vs_name) {
        $vs_profile_filter{vs_name} = $vs_name;
    }

    if ($profile_name) {
        $vs_profile_filter{profile_name} = $profile_name;
    }

    my @policy_vs_names;

    if ($policy_name) {
        for my $l7_action (@{F5::ASM::COAPI::Query->new('policy_action', {target => F5::Mcp::PolicyTarget::L7_POLICY_TARGET_ASM,
                                                                          action => F5::Mcp::PolicyAction::L7_POLICY_ACTION_ENABLE,
                                                                          policy => $policy_name})}) {
            for my $vs_policy (@{F5::ASM::COAPI::Query->new('virtual_server_policy', {policy_name => $l7_action->{policy_name}})}) {
                push(@policy_vs_names, $vs_policy->{vs_name});
            }
        }

        if (!@policy_vs_names) {
            return {};
        }

        if (scalar @policy_vs_names == 1 && !$vs_name) {
            $vs_profile_filter{vs_name} = shift(@policy_vs_names); # more effective
        }
    }

    # PROFILE_WEBSECURITY is unnecessary, since definitely secure
    my %security_profiles = (F5::Mcp::ProfileType::PROFILE_FTP()      => 'profile_ftp',
                             F5::Mcp::ProfileType::PROFILE_INSTANCE() => 'profile_smtp');

    my $vs_cache_profiles = {};

    for my $vs_profile (@{F5::ASM::COAPI::Query->new('virtual_server_profile', {%vs_profile_filter})}) {
        next if ($vs_cache_profiles->{$vs_profile->{vs_name}} || # security profiles are mutually exclusive
                 @policy_vs_names && !grep { $_ eq $vs_profile->{vs_name} } @policy_vs_names); # extra filter by array

        my $profile_class = $security_profiles{$vs_profile->{profile_type}};

        if ($vs_profile->{profile_type} == F5::Mcp::ProfileType::PROFILE_WEBSECURITY ||
              (($vs_profile->{profile_type} == F5::Mcp::ProfileType::PROFILE_FTP) &&
               scalar @{F5::ASM::COAPI::Query->new($profile_class, {name => $vs_profile->{profile_name},
                                                                    security_enabled => 1})}) ||
              ($vs_profile->{profile_type} == F5::Mcp::ProfileType::PROFILE_INSTANCE &&
               scalar @{F5::ASM::COAPI::Query->new('ltcfg_instance_field', {class_name => $profile_class,
                                                                            field_name => 'security_enabled',
                                                                            value => 'true',
                                                                            instance_name => $vs_profile->{profile_name}})})) {
            # Include also secured virtual servers without any cache
            $vs_cache_profiles->{$vs_profile->{vs_name}}->{security_profile_type} = $vs_profile->{profile_type};

            # NOTE: Web Acceleration is relevant only for HTTP profile (maybe insecure under Web Security profile)
            my ($vs_wa_profile) = @{F5::ASM::COAPI::Query->new('virtual_server_profile', {vs_name => $vs_profile->{vs_name},
                                                                                          profile_type => F5::Mcp::ProfileType::PROFILE_WEBACCELERATION})};

            if ($vs_wa_profile) { # can be at most one
                my ($wa_profile) = @{F5::ASM::COAPI::Query->new('profile_webacceleration', {name => $vs_wa_profile->{profile_name}})};

                if ($wa_profile) { # surely exists
                    $vs_cache_profiles->{$vs_profile->{vs_name}}->{cache_profile_name} = $wa_profile->{name};

                    if (@{$wa_profile->{applications}}) {
                        $vs_cache_profiles->{$vs_profile->{vs_name}}->{cache_type} = CACHE_TYPE_WA_CACHE;
                    }
                    else {
                        $vs_cache_profiles->{$vs_profile->{vs_name}}->{cache_type} = CACHE_TYPE_RAMCACHE;
                    }
                }
            }
        }
    }
       
    return $vs_cache_profiles;
}

sub invalidate_cache_profiles {
    my %params = @_;
    if (F5::Cluster::Utils::is_clustering_enabled() && ! F5::Cluster::Utils::is_master_blade()) {
        #Skip reset if we are not master blade
        return;
    }

    my $vs_cache_profiles = $params{vs_cache_profiles};
    my $logger = $params{logger};
    my $partition = "Common";  # RamCache doesn't care

    for my $cache_profile (values %$vs_cache_profiles) {
        if ($cache_profile->{cache_profile_name}) {
            $logger->debug("Invalidate $cache_profile->{cache_type} of profile $cache_profile->{cache_profile_name}") if $logger;

            # Not overridden in F5::ASM::COAPI
            F5::COAPI::Stats::reset($cache_profile->{cache_type}, $partition, {"profile" => $cache_profile->{cache_profile_name}});
        }
    }
}

1;
