package F5::Scanner;

use strict;
use warnings;

use YAML::XS;

use POSIX;
use F5::XML::Twig;
use Config::IniFiles;
use LWP::UserAgent;
use Time::Local;
use Memoize;
use Digest::MD5;
use File::Basename;
use List::Util 'max';
use JSON;
use Encode qw(encode_utf8);
use File::Copy;
use URI::Escape;
use MIME::Base64;

use F5::Deploy;
use F5::XMLUtils;
use F5::ASMConfig::ASMConfig;
use F5::NegativeSignatures::Utils;
use F5::ASMConfig::Entity::ScannerVuln;
use F5::Scanner::Cenzic;
use F5::Scanner::Sentinel;
use F5::Scanner::WebInspect;
use F5::LWP::UserAgent;
use F5::Utils::File;
use F5::ObjUtils qw(dclone);
use F5::DateTimeUtils;
use F5::WhiteHatKey;
use F5::ASM::COAPI;
use F5::ASMConfig::Handler::SignatureSet;
use F5::ASMConfig::Handler::PolicySignatureSet;
use F5::ASMConfig::Entity::SignatureSet;

memoize('get_attack_type_mapping');
memoize('get_parser_status_codes');

my $SCANNERS_CONFIG_FILENAME = F5::Deploy::map_path('/usr/share/ts/config/scanners_config.yaml');
my $SITEMAP_BASE_DIR = F5::Deploy::map_path('/var/ts/var/sitemap');
our $scanners_config;

our $OBJECT_PARAM_TYPE_USER_INPUT = 4;
our $OBJECT_PARAM_USER_INPUT_TYPE_ALPHA_NUMERIC = 32;
our $OBJECT_PARAM_USER_INPUT_TYPE_FILE_UPLOAD = 1;
our $CANT_DOWNLOAD_VULNS_FROM_SCANNER = 1;

our %QUALYS_COOKIE_VULN_NAMES = map { ($_ => 1) } qw(150120 150121 150122 150123);
our %APPSCAN_COOKIE_VULN_NAMES = map { ($_ => 1) } qw(attRespCookieNotHttpOnlySessionCookie attRespCookieNotSecureSSL);

my $URL_FIELD_NAME = 'request_url';

my @retest_status_hierarchy = (
    {
     order => 1,
     xml_status => 'retest_refused',
     db_status => F5::ASMConfig::ScannerRetestStatus::SCANNER_RETEST_STATUS_REFUSED,
    },
    {
     order => 2,
     xml_status => 'retest_temporarily_unavailable',
     db_status => F5::ASMConfig::ScannerRetestStatus::SCANNER_RETEST_STATUS_UNAVAILABLE,
    },
    {
     order => 3,
     xml_status => 'manual_retest_request',
     db_status => F5::ASMConfig::ScannerRetestStatus::SCANNER_RETEST_STATUS_MANUAL,
    },
    {
     order => 4,
     xml_status => 'manual_retest_pending',
     db_status => F5::ASMConfig::ScannerRetestStatus::SCANNER_RETEST_STATUS_MANUAL,
    },
    {
     order => 5,
     xml_status => 'automatic_retest_running',
     db_status => F5::ASMConfig::ScannerRetestStatus::SCANNER_RETEST_STATUS_REQUESTED,
    },
    {
     order => 6,
     xml_status => 'automatic_retest_available',
     db_status => F5::ASMConfig::ScannerRetestStatus::SCANNER_RETEST_STATUS_NONE,
    },
);

our %status_hierarchy_by_xml_status = map { ($_->{xml_status} => $_) } @retest_status_hierarchy;
our %status_hierarchy_by_order = map { ($_->{order} => $_) } @retest_status_hierarchy;

#----------------------------------------------------------------------

sub new {

    my($class, %params) = @_;

    my $self = {};
    bless $self, $class;
    $self->init(%params);

    if (my $subclass = $self->get_subclass) {
      bless $self, $subclass;
    }

    return $self;
}

#----------------------------------------------------------------------

sub get_subclass {
    my($self) = @_;

    if ($self->{scanner_config}->{scanner_type} eq 'hailstorm') {
        return "F5::Scanner::Cenzic";
    }
    elsif ($self->{scanner_config}->{scanner_type} =~ /^sentinel/) {
        return "F5::Scanner::Sentinel";
    }
    elsif ($self->{scanner_config}->{scanner_type} eq 'webinspect') {
        return "F5::Scanner::WebInspect";
    }

    return "F5::Scanner";
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    if (! $scanners_config) {
        $scanners_config = YAML::XS::LoadFile($SCANNERS_CONFIG_FILENAME);

        for my $vuln_handling (map { $_->{vuln_handling} } @{$scanners_config->{nodes}}) {

            # Map same_as scanner attack types
            for my $scanner_attack_type (keys %$vuln_handling) {
                if ($vuln_handling->{$scanner_attack_type}->{same_as}) {
                    $vuln_handling->{$scanner_attack_type} = $vuln_handling->{$vuln_handling->{$scanner_attack_type}->{same_as}};
                }
            }
        }
    }

    $self->{policy} = $params{policy};
    $self->reinit(%params);

    my $scanner_type;

    if (defined $self->{scanner_config}) {
        $scanner_type = $self->{scanner_config}->{scanner_type};
        $self->{scanner_name} = $self->{scanner_config}->{scanner_type};
    }
    else {
        $scanner_type = $params{scanner_type};
        $self->{scanner_name} = $params{scanner_type};
    }

    for my $node (@{$scanners_config->{nodes}}) {
        if ( grep { $_ eq $scanner_type } @{$node->{scanners}} ) {
            ($self->{config}) = $node;
        }
    }
    $self->{scanner_display_name} = $self->{config}->{scanner_display_name} || $self->{scanner_name};

    $self->{pre_resolve_templates} = $scanners_config->{general}->{pre_resolve_messages};
    $self->{save_vuln_files} = $scanners_config->{general}->{save_vuln_files};


    $self->{vuln_handling_attack_ids} = $self->get_attack_type_mapping();
    $self->{import_value_maps} = {};
}

#----------------------------------------------------------------------

sub reinit {
    my($self, %params) = @_;

    $self->{dbh} = $params{dbh} if defined $params{dbh};
    $self->{client} = $params{client} if defined $params{client};
    $self->{last_vuln_id} = 0;
    my $cfg_file = '/etc/ts/tools/vulnerability_assessment.cfg';
    $self->{page_size} = F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'RetestStatusPageSize') || '10';
    $self->{outbound_call_timeout} = defined $params{call_flags}->{outbound_call_timeout}
                                     ? $params{call_flags}->{outbound_call_timeout}
                                     : defined F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'OutbandCallTimeout')
                                     ? F5::Cfg::cfg_get_config_item($cfg_file, 'General', 'OutbandCallTimeout')
                                     : 300;

    if (defined $self->{policy}) {
        $self->{scanner_config} = $self->{client}->get_scanner_config($self->{client}->{session}, undef, $self->{policy});
    }
    elsif ($self->{client}) {
        my $search_results = $self->{client}->search_global_scanner_config(
                                      $self->{client}->{session},
                                      undef,
                                      F5::ASMConfig::GlobalScannerConfig->new(scanner_type => $params{scanner_type}));

        if ($search_results->{result_record_count}) {
            $self->{scanner_config} = $search_results->{element_list}->[0]->{global_scanner_config};
        }
        else {
            $self->{scanner_config} = F5::ASMConfig::GlobalScannerConfig->new(scanner_type => $params{scanner_type}, flg_is_setting_global => 1);
        }
    }
    else {
        # Create empty scanner config for ad-hoc URL conversion
        $self->{scanner_config} = {
                                   scanner_type => $params{scanner_type},
                                   api_key => $params{api_key},
                                  };
    }
}

#----------------------------------------------------------------------

sub policy_check_retest_requests {}

#----------------------------------------------------------------------

sub get_full_url {
    my($self, %params) = @_;

    my $url_type = $params{url_type};
    my $vuln_ids = defined $params{vuln_ids} ? join (',', @{$params{vuln_ids}}) : '';

    my $config = $self->{config}->{api};

    my $url = $self->replace_url_parts($config->{$url_type}, {
                                                              VULN_IDS => $vuln_ids,
                                                              API_KEY => $self->{scanner_config}->{api_key},
                                                              SITE_NAME => $self->{scanner_config}->{scanner_site_name},
                                                              SCANNER_URL => $config->{url},
                                                              PORT => $config->{port},});

    return $url;
}

#----------------------------------------------------------------------

sub get_ua {

    my (%params) = @_;

    my $timeout = defined $params{timeout} ? $params{timeout} : 30;
    my $ua = F5::LWP::UserAgent->new(timeout => $timeout);
    return $ua;
}

#----------------------------------------------------------------------

sub get_logger {
    my($class_or_self) = @_;
    if (ref($class_or_self) and defined $class_or_self->{logger}) {
        return $class_or_self->{logger};
    }
    else {
        my $logger = F5::Logger->get_logger(logger_name => "scanner");
        return $logger;
    }
}

#----------------------------------------------------------------------

sub get_attack_type_mapping {
    my ($self) = @_;

    my $dbh = $self->{dbh};
    my $config = $self->{config};

    my $asm_attack_types = $dbh->selectall_hashref(<<EndQuery, 'attack_type_name');
SELECT attack_type_id, attack_type_name
  FROM PLC.NEGSIG_ATTACK_TYPES
EndQuery

    my %retval;

    for my $scanner_attack_type_name (keys %{$config->{vuln_handling}}) {
        if ($config->{vuln_handling}->{$scanner_attack_type_name}) {
            my $asm_attack_type_name = $config->{vuln_handling}->{$scanner_attack_type_name}->{asm_attack_type};
            $retval{$scanner_attack_type_name} = $asm_attack_types->{$asm_attack_type_name}->{attack_type_id};
        }
    }

    return \%retval;
}

#----------------------------------------------------------------------

sub get_xml_data { }

#----------------------------------------------------------------------

sub get_host_name {
    my ($self, %params) = @_;

    my $scanner_vuln = $params{scanner_vuln};

    my $url = $scanner_vuln->{$URL_FIELD_NAME};

    if ($url !~ /^http/) {
        $url = 'http://' . $url;
    }
    my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($url);

    return $domain;
}

#----------------------------------------------------------------------

sub parse_xml {
    my ($self, %params) = @_;

    my $events = $params{events};
    my $scanner_actions_fields = $params{scanner_actions_fields};
    my $xml_data = $scanner_actions_fields->{scanner_vuln_data_xml};
    my $xml_file_path = $scanner_actions_fields->{xml_file_path};
    my $get_host_names_only = $scanner_actions_fields->{get_host_names_only};
    my $import_all_host_names = $scanner_actions_fields->{import_all_host_names};
    my $copy_file = $scanner_actions_fields->{copy_file};
    my %requested_host_names = map {$_ => 1} @{$scanner_actions_fields->{host_names}} if defined $scanner_actions_fields->{host_names};

    my $client = $self->{client};
    my $policy = $self->{policy};

    if ((!$xml_data) && (!$xml_file_path)) {
        $self->update_scanner_config(
            scanner_config => F5::ASMConfig::ScannerConfig->new({scanner_auth_data_1 => $scanner_actions_fields->{subscription_id},}));

        my $scanner_url = $client->get_scanner_vuln_url($client->{session}, undef, $policy, $scanner_actions_fields);
        my $retval_get_xml_data = $self->get_xml_data(url => $scanner_url,
                                                      scan_id => $scanner_actions_fields->{scan_id});

        if ($retval_get_xml_data->{is_success}){
            $xml_data = $retval_get_xml_data->{xml_data};
            $scanner_actions_fields->{scanner_vuln_data_xml} = $xml_data; # For CMI relay if required
        }
        else {
            die {error_message => "Cannot get vuln data from Scanner server",
                 error_code => $CANT_DOWNLOAD_VULNS_FROM_SCANNER};
        }
    }

    my $permanent_filename = $xml_file_path;
    if ($xml_file_path && $get_host_names_only) {
        $permanent_filename = $self->get_permanent_filename();
        if ($copy_file) {
            copy ($xml_file_path, $permanent_filename);
        }
        else {
            move ($xml_file_path, $permanent_filename);
        }
    }
    elsif ($xml_data) {
        $permanent_filename = $self->get_permanent_filename();
        open (my $out, ">", $permanent_filename);
        print $out $xml_data;
        close $out;
    }

    my $vuln_handler = {};
    my $twig_rescan_needed = 0;
    my %vuln_count_by_host_name;
    my %updated_vuln_count_by_host_name;
    my $parsing_rules = $self->{config}->{parsing_rules};
    $parsing_rules->{value_maps} ||= {};

    my @parsing_rules_vulns;
    for my $vuln_header (keys %$parsing_rules) {
        if ($vuln_header eq 'vuln' || $vuln_header eq 'sensitive_vuln' ||
            $vuln_header =~ /^vuln\(/) {

            push(@parsing_rules_vulns, $parsing_rules->{$vuln_header});
        }
    }

    for my $parsing_rules_vuln (@parsing_rules_vulns) {
        my %xpath_depth = map { $_->{xpath} => 0 }
          grep { defined $_->{xpath} }
            @{$parsing_rules_vuln->{fields}};

        for my $xpath (keys %xpath_depth) {
            while ($xpath =~ /\.\.\//g) {
                $xpath_depth{$xpath}++;
            }
        }

        my $max_depth = max (values %xpath_depth);
        my $purge_xpath = $parsing_rules_vuln->{xpath};

        for (1..$max_depth) {
            $purge_xpath = dirname($purge_xpath);
        }

        my $load_time = time();

        my $process_handler_sub = sub {
            my ($twig, $vuln) = @_;

            my $retval = $self->parse_and_update_scanner_vuln(
                                                              twig => $twig,
                                                              vuln_elm => $vuln,
                                                              load_time => $load_time,
                                                              parsing_rules_vuln => $parsing_rules_vuln,
                                                              import_all_host_names => $import_all_host_names,
                                                              requested_host_names => \%requested_host_names,);
            return unless $retval;

            $updated_vuln_count_by_host_name{$retval->{host_name}} += $retval->{vuln_update_count};
            $vuln_count_by_host_name{$retval->{host_name}} += $retval->{vuln_count};
        };

        my $twig_handler_sub = sub {
            my ($twig, $vuln) = @_;

            # Value map was needed, but hasn't been read yet -- the import file needs
            # to be re-scanned after the value_maps have been loaded
            if (%{$parsing_rules->{value_maps}} && ! %{$self->{import_value_maps}} &&
               grep { exists $_->{value_map} } @{$parsing_rules_vuln->{fields}}) {

                $twig_rescan_needed = 1;
                return;
            }

            if (! $parsing_rules_vuln->{non_sax_xml_process}) {
                $process_handler_sub->($twig, $vuln);
            }

            if ($parsing_rules_vuln->{xpath} eq $purge_xpath) {
                $twig->purge;
            }
        };

        $vuln_handler->{$parsing_rules_vuln->{xpath}} = $twig_handler_sub;

        if ($purge_xpath ne $parsing_rules_vuln->{xpath}) {
            $vuln_handler->{$purge_xpath} = sub {
                my ($twig, $vuln_type) = @_;
                if ($parsing_rules_vuln->{non_sax_xml_process}) {
                    my @vulns = $twig->get_xpath($parsing_rules_vuln->{xpath});
                    for my $vuln (@vulns) {
                        $process_handler_sub->($twig, $vuln);
                    }
                }
                $twig->purge;
            };
        }
    }

          for my $import_value_map_name (keys %{$parsing_rules->{value_maps}}) {
              my $map_def = $parsing_rules->{value_maps}->{$import_value_map_name};
              $vuln_handler->{$map_def->{xpath}} = sub {
                  my ($twig, $element) = @_;
                  $self->store_import_value_map(
                                                element => $element,
                                                map_def => $map_def,
                                               );
              }
          }

    my $twig = F5::XML::Twig->new(twig_roots => $vuln_handler);
    eval {
        if (defined $xml_data) {
            $twig->parse($xml_data);
            if ($twig_rescan_needed) {
                $twig->parse($xml_data);
            }
            $xml_file_path = $self->save_vuln_file(\$xml_data);
        }
        elsif (defined $permanent_filename) {
            $twig->parsefile($permanent_filename);
            if ($twig_rescan_needed) {
                $twig->parsefile($permanent_filename);
            }
            $xml_file_path = $permanent_filename;
        }
    };

    if ($@) {
        return { error => $@};
    }
    else {
        return { vuln_hosts => \%vuln_count_by_host_name,
                 updated_vuln_hosts => \%updated_vuln_count_by_host_name,
                 vuln_file_path => $xml_file_path };
    }
}

#----------------------------------------------------------------------

sub save_vuln_file {
    my ($self, $xml_data_ref) = @_;

    my $filename = $self->get_permanent_filename();

    if (open (my $out, ">", $filename)) {
        # if cannot open the file then skip silently.
        print $out $$xml_data_ref;
        close ($out);
        $self->get_logger->info("Vulnerability file was saved at $filename");
    }

    return $filename;
}

#----------------------------------------------------------------------

sub get_permanent_filename {
    my ($self) = @_;

    my $filename;
    F5::Utils::File::cleanup_files(dir => $self->{save_vuln_files}->{location},
                                   max_retained_files => $self->{save_vuln_files}->{max_retained_files},);

    my $datetime = F5::DateTimeUtils::convert_epoch_to_datetime(time());
    $datetime =~ s/ /-/g;
    $datetime =~ s/:/_/g;
    $filename = $self->{save_vuln_files}->{location} . '/' . $self->{policy}->{policy_id} . '-' . $datetime . '.xml';

    return $filename;
}

#----------------------------------------------------------------------

sub resolve_vuln {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $flg_staging = $params{flg_staging};
    my $get_messages_only = $params{get_messages_only};

    my $policy = $self->{policy};
    my $config = $self->{config};
    my $client = $self->{client};
    my $call_flags = F5::ASMConfig::CallFlags->new;

    return unless $config->{vuln_handling}->{$vuln->{scanner_vuln_name}};

    my %retval;

    my $actions = $config->{vuln_handling}->{$vuln->{scanner_vuln_name}}->{resolve_actions} || [];

    my @vuln_group = F5::ASMConfig::Entity::ScannerVuln->retrieve_all(
                         dbh => $self->{dbh},
                         policy => $policy,
                         get_criteria => { scanner_vuln_id => $vuln->{scanner_vuln_id} });

    for my $action (@$actions) {
        my $method = $action;
        my @internal_params;

        if ($method =~ s/\((.*)\)//) {
            @internal_params = split (/,/, $1);
        }

        $self->$method( retval => \%retval,
                        staging => $flg_staging,
                        vuln_group => \@vuln_group,
                        internal_params => \@internal_params,
                        get_messages_only => $get_messages_only,
                        vuln => $vuln);
        if ($retval{stop_resolving_vuln}) {
            return $retval{messages}; #it's good even if we're in get_messages_only, because we won't perform those anyway
        }
    }

    if ($get_messages_only) {
        return $retval{pre_resolve_messages};
    }

    if (defined $config->{vuln_handling}->{$vuln->{scanner_vuln_name}}->{messages}) {
        $retval{messages} = $config->{vuln_handling}->{$vuln->{scanner_vuln_name}}->{messages};
    }

    #update vulnerabilities status only if there are resolve_actions == automatically resolvable 
    if (defined $config->{vuln_handling}->{$vuln->{scanner_vuln_name}}->{resolve_actions}) {
        my $vuln_criteria;
        if ($config->{vuln_handling}->{$vuln->{scanner_vuln_name}}->{resolve_all}) {
            my @relevant_config_vulns =
                grep {(defined $config->{vuln_handling}->{$_}->{resolve_all}) &&
                      ($config->{vuln_handling}->{$_}->{resolve_all} == 1) &&
                      ($self->{vuln_handling_attack_ids}->{$_} eq $vuln->{attack_type_id}) }
                keys %{$config->{vuln_handling}};
            push @{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new({
                                                       field_name => 'scanner_vuln_name',
                                                       field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN,
                                                       field_value_list => \@relevant_config_vulns,
                                                   });

            $vuln_criteria = undef;
        }
        else {
            $vuln_criteria = $vuln;
        }

        $client->update_many_scanner_vulns(
            $client->{session}, $call_flags, $policy, F5::ASMConfig::ScannerVuln->new($vuln_criteria),
            F5::ASMConfig::ScannerVuln->new({
                asm_vuln_status => $flg_staging
                                   ? F5::ASMConfig::ScannerAsmVulnStatus::SCANNER_ASM_VULN_STATUS_HANDLED_STAGING()
                                   : F5::ASMConfig::ScannerAsmVulnStatus::SCANNER_ASM_VULN_STATUS_HANDLED(),
                resolve_time => time(),}),
        );
    }

    return $retval{messages};
}

#----------------------------------------------------------------------

sub enable_blocking {
    my ($self, %params) = @_;

    my $client = $self->{client};
    my $policy = $self->{policy};
    my $retval = $params{retval};

    my $violation_codes = $params{internal_params};

    my $parser_status_codes = $self->get_parser_status_codes();
    for my $violation (@$violation_codes) {
        if ( $params{get_messages_only} ) {
            if ( $policy->{has_parent} ) {
                my $blocking = $client->get_blocking($client->{session},
                                                     undef,
                                                     $policy,
                                                     F5::ASMConfig::Blocking->new({ violation => F5::ASMConfig::Violation->new({viol_index => $parser_status_codes->{$violation}->{viol_index}})}),
                                                    );
                if ( $blocking->{is_inherited} ) {
                    $self->add_inheritance_warning(%params, inherited_entity => $blocking);
                }
            }
        }
        else {
            $client->update_blocking($client->{session},
                                     undef,
                                     $policy,
                                     F5::ASMConfig::Blocking->new({
                                         violation => F5::ASMConfig::Violation->new({viol_index => $parser_status_codes->{$violation}->{viol_index}}),
                                         flg_alarm => 1,
                                         flg_reject => 1,
                                         flg_learn => 1,}));
        }
    }
    return $retval;
}

#----------------------------------------------------------------------

sub add_csrf_url {
    my ($self, %params) = @_;

    my $vuln_group = $params{vuln_group};

    my $client = $self->{client};
    my $policy = $self->{policy};

    my @objects;
    for my $vuln (@$vuln_group) {
        my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($vuln->{request_url});
        push (@objects, $object);
    }

    return unless scalar @objects;

    $client->update_csrf($client->{session}, undef, $policy, F5::ASMConfig::Csrf->new({flg_is_enabled=>1}));
    my %csrf_urls = map { ($_ => 1) } @objects;

    for my $target_pattern (keys %csrf_urls) {
        $client->add_csrf_url($client->{session},
                              F5::ASMConfig::CallFlags->new(ignore_duplicate_record_on_add => 1),
                              $policy,
                              F5::ASMConfig::CsrfURL->new({ target_pattern => $target_pattern }));
    }

}

#----------------------------------------------------------------------

sub add_valid_hostname {
    my ($self, %params) = @_;

    my $vuln_group = $params{vuln_group};

    my $client = $self->{client};
    my $policy = $self->{policy};

    return if $params{get_messages_only};

    my %domains;
    for my $vuln (@$vuln_group) {
        my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($vuln->{request_url});
        if ( $domain && !$domains{$domain} ) {
            $domains{$domain} = 1;
            my $valid_host_name=  $client->update_valid_host_name(
                                            $client->{session},
                                            F5::ASMConfig::CallFlags->new(add_on_updating_nonexistent_record => 1),
                                            $policy,
                                            F5::ASMConfig::ValidHostName->new(host_name => $domain,
                                                                              created_by => $self->{scanner_display_name},
                                                                              flg_include_subdomains => 1,));
            if( defined $valid_host_name ) {
                  $vuln->{asm_valid_host_name} = $valid_host_name;
                  my $update_vuln = $client->update_scanner_vuln( $client->{session}, undef, $policy, $vuln);
            }
       }

    }

    return;
}

#----------------------------------------------------------------------

sub enable_data_guard {
    my ($self, %params) = @_;

    my $data_guard_types = $params{internal_params};

    my $client = $self->{client};
    my $policy = $self->{policy};
    my $retval = $params{retval};

    if ( $params{get_messages_only} ) {
        if ( $policy->{has_parent} ) {
            my $response_scrubbing_rv = $client->get_response_scrubbing( $client->{session},
                                                                         undef,
                                                                         $policy,
                                                                       );

            if ( $response_scrubbing_rv->{is_inherited} ) {
                $self->add_inheritance_warning(%params, inherited_entity => $response_scrubbing_rv);
            }
        }
    }
    else {
        my %data_guard_flags = map {$_ => 1} @$data_guard_types;
        my $response_scrubbing = F5::ASMConfig::ResponseScrubbing->new({
                                     flg_is_enabled => 1,
                                     %data_guard_flags });

        $client->update_response_scrubbing($client->{session},
                                           undef,
                                           $policy,
                                           $response_scrubbing);
    }
    return $retval;
}

#----------------------------------------------------------------------

sub add_cookie {
    my ($self, %params) = @_;

    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $vuln_group = $params{vuln_group};
    my $get_messages_only = $params{get_messages_only};

    my $client = $self->{client};
    my $policy = $self->{policy};

    my @cookies;
    for my $vuln (@$vuln_group) {
        my $cookie_rv = $client->search_cookie(
                         $client->{session},
                         undef,
                         $policy,
                         F5::ASMConfig::Cookie->new({cookie_name => $vuln->{cookie}, is_wildcard => 0}));

        my $cookie;
        if ($cookie_rv->{result_record_count}) {
            $cookie = $cookie_rv->{element_list}->[0]->{cookie};
        }

        if ($get_messages_only) {
            if (! $cookie) {
                $cookie = { cookie_name => $vuln->{cookie} };
            }
        }
        else {
            if (! defined $cookie) {
                $cookie = $client->add_cookie(
                                              $client->{session},
                                              F5::ASMConfig::CallFlags->new(ignore_duplicate_record_on_add => 1),
                                              $policy,
                                              F5::ASMConfig::Cookie->new({ cookie_name => $vuln->{cookie}, 
                                                                           cookie_enforcement_type => F5::ASMConfig::CookieEnforcementType::COOKIE_ENFORCEMENT_TYPE_ALLOWED(),
                                                                           created_by => $self->{scanner_display_name},
                                                                           in_staging => $staging, }));
            }

        }
        
        if( (! $get_messages_only) && (defined $cookie) ) {
            $vuln->{asm_cookie} = $cookie;
            my $update_vuln = $client->update_scanner_vuln( $client->{session}, undef, $policy, $vuln);
        }

        push(@cookies, $cookie);
    }

    $retval->{cookie} = $cookies[0];
    $retval->{cookies} = \@cookies;
    return $retval;
}

#----------------------------------------------------------------------

sub enable_cookie_flag {
    my ($self, %params) = @_;

    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $vuln_group = $params{vuln_group};
    my $get_messages_only = $params{get_messages_only};

    my $client = $self->{client};
    my $policy = $self->{policy};
    
    my $flag_settings = { is_http_only => { flag => "is_http_only", pre_resolve_msg => "enable_httponly_on_cookie" },
                          is_secure => { flag => "is_secure", pre_resolve_msg => "enable_secure_on_cookie" }};
    
    if ($params{internal_params}->[0] && defined $flag_settings->{$params{internal_params}->[0]}) {
        my $cookie_flag = $flag_settings->{$params{internal_params}->[0]}->{flag};
        my $pre_resolve_msg = $flag_settings->{$params{internal_params}->[0]}->{pre_resolve_msg};
        
        for my $i (0..$#{$vuln_group}) {
            my $vuln = $vuln_group->[$i];
            my $cookie = $retval->{cookies}->[$i] || $retval->{cookie};
            
            if ($get_messages_only) {
                if(!$cookie->{$cookie_flag}) {
                    if ( $policy->{has_parent} ) {
                        my $rv = $client->search_cookie($client->{session},
                                                        undef,
                                                        $policy,
                                                        $cookie,
                                                       );
                        if ( $rv->{result_record_count} ) {
                            my $cookie_rv = $rv->{element_list}->[0]->{cookie};
                            if ( $cookie_rv->{is_inherited} ) {
                                $self->add_inheritance_warning(%params, inherited_entity => $cookie_rv);
                                next;
                            }
                        }
                    }
                    push (@{$retval->{pre_resolve_messages}}, {
                        message_text => $self->{pre_resolve_templates}->{$pre_resolve_msg},
                        bind_values => [ $cookie->{cookie_name} ],
                    });
                }
           }
           else {
                $cookie->{$cookie_flag} = 1;
                $cookie->{in_staging} = $staging;
                $client->update_cookie(
                    $client->{session},
                    F5::ASMConfig::CallFlags->new(),
                    $policy,
                    $cookie);
            }
        }
    }

    return $retval;
}

#----------------------------------------------------------------------

sub add_url {
    my ($self, %params) = @_;

    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $vuln_group = $params{vuln_group};
    my $get_messages_only = $params{get_messages_only};

    my $client = $self->{client};
    my $policy = $self->{policy};

    my @urls;
    for my $vuln (@$vuln_group) {
        my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($vuln->{request_url});

        my $search_url = F5::ASMConfig::URL->new({url_name => $object, app_id => $app_id, is_wildcard => 0});
        my $url_rv = $client->search_url(
                                         $client->{session},
                                         undef,
                                         $policy,
                                         $search_url,
                                        );

        my $url;
        if ($url_rv->{result_record_count}) {
            $url = $url_rv->{element_list}->[0]->{url};
        }

        if ((! $get_messages_only) && (! defined $url)) {
            $url = $client->add_url(
                       $client->{session},
                       F5::ASMConfig::CallFlags->new(ignore_duplicate_record_on_add => 1),
                       $policy,
                       F5::ASMConfig::URL->new({url_name => $object, app_id => $app_id, created_by => $self->{scanner_display_name}}));
        }
        
        if( (! $get_messages_only) && (defined $url) ) {
            $vuln->{asm_url} = $url;
            my $update_vuln = $client->update_scanner_vuln( $client->{session}, undef, $policy, $vuln);
        }

        push(@urls, $url || $search_url);
    }


    $retval->{url} = $urls[0];
    $retval->{urls} = \@urls;
    return $retval;
}

#----------------------------------------------------------------------

sub add_filetype {
    my ( $self, %params ) = @_;

    my $retval            = $params{retval};
    my $vuln_group        = $params{vuln_group};
    my $get_messages_only = $params{get_messages_only};

    my $client = $self->{client};
    my $policy = $self->{policy};

    my @filetypes;
    for my $vuln (@$vuln_group) {
        my ( $app_id, $domain, $object, $query ) = F5::ParseUri::parse_uri( $vuln->{request_url} );

        my ( $filename, $directories, $filetype_name ) = fileparse( $object, qr/(?<=\.)[^.\/]+?$/ );
        if (! $filetype_name) {
            $filetype_name = 'no_ext';
        }

        my $filetype_rv = $client->search_filetype(
            $client->{session},
            undef, $policy,
            F5::ASMConfig::Filetype->new({ filetype_name => $filetype_name, is_wildcard => 0 })
        );
        my $filetype;
        if ( $filetype_rv->{result_record_count} ) {
            $filetype      = $filetype_rv->{element_list}->[0]->{filetype};
        }

        if ( ( !$get_messages_only ) && ( !defined $filetype ) ) {
            $filetype = $client->add_filetype(
                $client->{session},
                F5::ASMConfig::CallFlags->new(ignore_duplicate_record_on_add => 1),
                $policy,
                F5::ASMConfig::Filetype->new({
                        filetype_name        => $filetype_name,
                        is_wildcard          => 0
                    })
            );
        }
        
        push @filetypes, $filetype;
    }

    $retval->{filetypes} = \@filetypes;
    return $retval;
}

#----------------------------------------------------------------------

sub apply_response_signatures_on_filetype {
    my ( $self, %params ) = @_;

    my $retval            = $params{retval};
    my $get_messages_only = $params{get_messages_only};
    my @filetypes = $retval->{filetypes} ? @{$retval->{filetypes}} : ();

    my $client = $self->{client};
    my $policy = $self->{policy};
    
    for my $filetype (@filetypes) {
        if ( defined $filetype ) {
            if ( $get_messages_only ) {
               if ( $policy->{has_parent} ) {
                    my $rv = $client->search_filetype( $client->{session},
                                                       undef,
                                                       $policy,
                                                       $filetype,
                                                     );
                    if ( $rv->{result_record_count} ) {
                        my $filetype_rv = $rv->{element_list}->[0]->{filetype};
                        if ( $filetype_rv->{is_inherited} ) {
                            $self->add_inheritance_warning(%params, inherited_entity => $filetype_rv);
                        }
                    }
                }
            }
            else {
                $filetype->{flg_filter_app_error} = 1;
                $client->update_filetype( $client->{session}, undef, $policy, $filetype );
            }
        }
    }

    return $retval;
}

#----------------------------------------------------------------------

sub add_disallowed_url {
    my ($self, %params) = @_;

    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $vuln_group = $params{vuln_group};
    my $get_messages_only = $params{get_messages_only};

    my $client = $self->{client};
    my $policy = $self->{policy};

    my @urls;
    for my $vuln (@$vuln_group) {
        my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($vuln->{request_url});
        my $url;

        # checking pre-requisites
        my $call_flags = F5::ASMConfig::CallFlags->new;
        push @{$call_flags->{search_filters}},
            F5::ASMConfig::SearchFilter->new({
                field_name => 'flg_is_allowed',
                field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                field_value => 1 });
        my $allowed_url_search_rv = $client->search_url(
                                        $client->{session},
                                        $call_flags,
                                        $policy,
                                        F5::ASMConfig::URL->new({url_name => $object,
                                                                 app_id => $app_id}));

        if ($allowed_url_search_rv->{result_record_count}) {
            $url = $allowed_url_search_rv->{element_list}->[0]->{url};
        }

        $self->check_blocking(%params, violation => 'OBJ_DOESNT_EXIST');

        if ($url) {
            if ($get_messages_only) {
                push (@{$retval->{pre_resolve_messages}}, {
                    message_text => $self->{pre_resolve_templates}->{disallowed_url_is_allowed},
                    bind_values => [ $url->{url_name} ],
                });
            }
            else {
                $client->delete_url($client->{session}, undef, $policy, $url);
            }
        }

        next if $get_messages_only;

        # resolve actions

        my $call_flags_disallowed = F5::ASMConfig::CallFlags->new;
        push @{$call_flags_disallowed->{search_filters}},
            F5::ASMConfig::SearchFilter->new({
                field_name => 'flg_is_allowed',
                field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                field_value => 0 });
        my $disallowed_url_search_rv = $client->search_url(
                                           $client->{session},
                                           $call_flags_disallowed,
                                           $policy,
                                           F5::ASMConfig::URL->new({url_name => $object,
                                                                    app_id => $app_id}));
        if (!$disallowed_url_search_rv->{result_record_count}) {
            $url = $client->add_url(
                       $client->{session},
                       F5::ASMConfig::CallFlags->new(ignore_duplicate_record_on_add => 1),
                       $policy,
                       F5::ASMConfig::URL->new({url_name => $object,
                                                app_id => $app_id,
                                                created_by => $self->{scanner_display_name},
                                                flg_is_allowed => 0}));
        }
        else {
            $url = $disallowed_url_search_rv->{element_list}->[0]->{url};
        }

        if( (! $get_messages_only) && (defined $url) ) {
            $vuln->{asm_url} = $url;
            my $update_vuln = $client->update_scanner_vuln( $client->{session}, undef, $policy, $vuln);
        }
        
        push @urls, $url;
    }


    $retval->{url} = $urls[0];
    $retval->{urls} = \@urls;
    return $retval;
}
#----------------------------------------------------------------------

sub add_url_param {
    my ($self, %params) = @_;

    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $vuln_group = $params{vuln_group};
    my $get_messages_only = $params{get_messages_only};

    my $client = $self->{client};
    my $policy = $self->{policy};

    my $data_type;

    if (defined $params{internal_params}->[0]
             && $params{internal_params}->[0] eq 'FILE_UPLOAD') {
        $data_type = $OBJECT_PARAM_USER_INPUT_TYPE_FILE_UPLOAD;
    }
    else {
        $data_type = $OBJECT_PARAM_USER_INPUT_TYPE_ALPHA_NUMERIC; #default
    }

    my @params;
    for my $i (0..$#{$vuln_group}) {
        my $vuln = $vuln_group->[$i];
        my $url = $retval->{urls}->[$i] || $retval->{url};

        my $param;
        my $param_name;
        my $is_wildcard;

        if ($vuln->{parameter}) {

            if ($vuln->{parameter} =~ /[ _+.]/ && $scanners_config->{general}->{php_parameter_underscore_mapping} == 1) {
                $param_name = $vuln->{parameter};
                # PHP automatically considers all of these characters as equivalent;
                # thus, a wildcard parameter needs to be created considering all of them
                $param_name =~ s/[ _+.]/[ _+.]/g;
                $is_wildcard = 1;
            }
            else {
                $param_name = $vuln->{parameter};
                $is_wildcard = 0;
            }
        }
        else {
            return $retval;
        }

        my $param_rv = $client->search_parameter($client->{session},
                           undef,
                           $policy,
                           F5::ASMConfig::Parameter->new({
                               param_name => $param_name,
                               is_wildcard => $is_wildcard,
                               param_level => F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL,
                               url => $url, }));

        if ($param_rv->{result_record_count}) {
            $param = $param_rv->{element_list}->[0]->{parameter};
            if (($param->{param_data_type} != $data_type) && !$get_messages_only) {
                my $message = $self->{pre_resolve_templates}->{param_type_mismatch};
                $message =~ s/PARAM/$param_name/;
                $retval->{messages} = [ $message ];
                $retval->{stop_resolving_vuln} = 1;
                return $retval;
            }
        }

        if (defined $param && $param->{in_staging} && !$staging) {
            # pre-resolve messages
            if ($get_messages_only) {
                if ( $policy->{has_parent} && $param->{is_inherited} ) {
                    $self->add_inheritance_warning(%params, inherited_entity => $param);
                }
                else {
                    push (@{$retval->{pre_resolve_messages}}, {
                        message_text => $self->{pre_resolve_templates}->{staging_enabled_on_existing_param},
                        bind_values => [ $param->get_full_display_name() ],
                    });
                }
                push (@params, $param);
                next;
            }
            else {
                # pre-resolve actions
                $param->{in_staging} = $staging;
                $param = $client->update_parameter($client->{session}, undef, $policy, $param);
            }
        }

        if ($get_messages_only) {
            if (!$param) {
              $param = {
                param_name => $param_name
              };
              push (@params, $param);
              $retval->{params} = \@params;
            }
            return $retval;
        }

        #resolve actions

        if (!$param) {
            $param = $client->add_parameter($client->{session},
                                            F5::ASMConfig::CallFlags->new(ignore_duplicate_record_on_add => 1),
                                            $policy,
                                            F5::ASMConfig::Parameter->new({
                                                param_name => $param_name,
                                                is_wildcard => $is_wildcard,
                                                param_level => F5::ASMConfig::ParamLevel::PARAM_LEVEL_URL,
                                                param_value_type => $OBJECT_PARAM_TYPE_USER_INPUT,
                                                param_data_type => $data_type,
                                                max_value_len => 0,
                                                min_value => '',
                                                max_value => '',
                                                url => $url,
                                                created_by => $self->{scanner_display_name},
                                                in_staging => $staging, }));
        }

        
        if( (! $get_messages_only) && (defined $param) ) {
            $vuln->{asm_parameter} = $param;
            my $update_vuln = $client->update_scanner_vuln( $client->{session}, undef, $policy, $vuln);
        }
        push (@params, $param);

    }

    $retval->{params} = \@params;
    return $retval;

}

#----------------------------------------------------------------------

sub enable_evasion_by_attack_type {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $vuln_group = $params{vuln_group};
    my $get_messages_only = $params{get_messages_only};
    my $retval = $params{retval};

    my $client = $self->{client};
    my $policy = $self->{policy};
    my $dbh = $self->{dbh};

    $self->check_blocking(%params, violation => 'EVASION_DETECTED');

    my ($normalization_code) = $dbh->selectrow_array(<<EndQuery, undef, $vuln->{attack_type_id});
SELECT normalization_code
  FROM PLC.NORMALIZATION_CODES
 WHERE attack_type_id = ?
EndQuery

    return unless $normalization_code;

    if ( $get_messages_only ) {
        if ( $policy->{has_parent} ) {
            my $normalization = $client->get_normalization( $client->{session},
                                                            undef,
                                                            $policy,
                                                            F5::ASMConfig::Normalization->new({normalization_code => $normalization_code})
                                                           );

            if ( $normalization->{is_inherited} ) {
                $self->add_inheritance_warning(%params, inherited_entity => $normalization);
            }
        }
    }
    else {
        $client->update_normalization($client->{session},
                                      undef,
                                      $policy,
                                      F5::ASMConfig::Normalization->new({normalization_code => $normalization_code, flg_is_enabled => 1}));
    }

    return $retval;
}

#----------------------------------------------------------------------

sub enable_evasion_by_name {
    my ($self, %params) = @_;

    my $evasion_names = $params{internal_params};
    my $get_messages_only = $params{get_messages_only};
    my $retval = $params{retval};

    my $client = $self->{client};
    my $policy = $self->{policy};
    my $dbh = $self->{dbh};

    $self->check_blocking(%params, violation => 'EVASION_DETECTED');

    for my $evasion_name (@$evasion_names) {

        my ($normalization_code) = $dbh->selectrow_array(<<EndQuery, undef, $evasion_name);
SELECT normalization_code
  FROM PLC.NORMALIZATION_CODES
 WHERE api_name = ?
EndQuery

        return unless $normalization_code;

        if ( $get_messages_only ) {
            if ( $policy->{has_parent} ) {
                my $normalization = $client->get_normalization($client->{session},
                                                               undef,
                                                               $policy,
                                                               F5::ASMConfig::Normalization->new({normalization_code => $normalization_code})
                                                              );

                if ( $normalization->{is_inherited} ) {
                    $self->add_inheritance_warning(%params, inherited_entity => $normalization);
                }
            }
        }
        else {
            $client->update_normalization(
                $client->{session},
                undef,
                $policy,
                F5::ASMConfig::Normalization->new({normalization_code => $normalization_code, flg_is_enabled => 1}));
        }
    }

    return $retval;
}

#----------------------------------------------------------------------

sub check_blocking {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $vuln_group = $params{vuln_group};
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};
    my $violation = $params{violation};

    my $client = $self->{client};
    my $policy = $self->{policy};
    my $dbh = $self->{dbh};

    my $parser_status_codes = $self->get_parser_status_codes();
    my $blocking = $client->get_blocking($client->{session},
                                         undef,
                                         $policy,
                                         F5::ASMConfig::Blocking->new({ violation => F5::ASMConfig::Violation->new({viol_index => $parser_status_codes->{$violation}->{viol_index}})}),
                                         );

    if (!$blocking->{flg_alarm} ||
        !$blocking->{flg_reject} ||
        !$blocking->{flg_learn}) {
        if ($get_messages_only) {
            push (@{$retval->{pre_resolve_messages}}, {
                message_text => $self->{pre_resolve_templates}->{violation_is_not_blocking},
                bind_values => [ $parser_status_codes->{$violation}->{description} ],
            });
        }
        else {
            $client->update_blocking($client->{session},
                                     undef,
                                     $policy,
                                     F5::ASMConfig::Blocking->new({
                                         violation => F5::ASMConfig::Violation->new({viol_index => $parser_status_codes->{$violation}->{viol_index}}),
                                         flg_alarm => 1,
                                         flg_reject => 1,
                                         flg_learn => 1,}));
        }
    }
}

#----------------------------------------------------------------------

sub enable_sigs_on_policy_by_attack_type {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};

    my $client = $self->{client};
    my $policy = $self->{policy};

    $self->sig_set_on_policy(%params);

    if ($params{get_messages_only}) {
        return;
    }

    my @sigs = F5::ASMConfig::Entity::Signature->retrieve_all(
                   dbh => $self->{dbh},
                   get_criteria => {attack_type_id => $vuln->{attack_type_id}});

    my @sigs_in_policy = F5::ASMConfig::Entity::PolicySignature->retrieve_all(
                   dbh => $self->{dbh}, policy => $policy,);


    my %lookupSigs = map { $_->{sig_id} => 1 } @sigs;
    my @sig_ids = ();
    for my $sig_in_policy (@sigs_in_policy) {
        if ($lookupSigs{ $sig_in_policy->{sig_id} }) {
            push @sig_ids, $sig_in_policy->{sig_id};
        }
    }

    my $call_flags = F5::ASMConfig::CallFlags->new;
    push @{$call_flags->{search_filters}},
         F5::ASMConfig::SearchFilter->new({
             field_name => 'sig_id',
             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN,
             field_value_list => \@sig_ids, });

    $client->update_many_policy_signatures(
        $client->{session},
        $call_flags,
        $policy,
        F5::ASMConfig::PolicySignature->new(),
        F5::ASMConfig::PolicySignature->new({flg_disabled => 0}));
    
}

#----------------------------------------------------------------------

sub add_data_guard_url {
    my ($self, %params) = @_;

    my $vuln_group = $params{vuln_group};

    my $client = $self->{client};
    my $policy = $self->{policy};

    my $retval = $params{retval};

    if ( $params{get_messages_only} ) {
        if ( $policy->{has_parent} ) {
            my $response_scrubbing = $client->get_response_scrubbing( $client->{session},
                                                                      undef,
                                                                      $policy, );
            if ( $response_scrubbing->{is_inherited} ) {
                $self->add_inheritance_warning(%params, inherited_entity => $response_scrubbing);
            }
        }
    }
    else {
        my @objects;
        for my $vuln (@$vuln_group) {
            my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($vuln->{request_url});
            push (@objects, $object);
        }

        return unless scalar @objects;

        my $call_flags = F5::ASMConfig::CallFlags->new(append_nested_struct_on_update => 1);
        $client->update_response_scrubbing($client->{session},
                              $call_flags,
                              $policy,
                              F5::ASMConfig::ResponseScrubbing->new({
                                  enforcement_mode => F5::ASMConfig::EnforcementMode::ENFORCE_ONLY_URL_LIST(),
                                  target_patterns => [@objects]
                              }));
    }

    return $retval;
}

#----------------------------------------------------------------------

sub enable_sigs_on_param_by_attack_name {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $attack_names = $params{internal_params};

    for my $attack_name (@$attack_names) {
        ($vuln->{attack_type_id}) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $attack_name);
SELECT attack_type_id
  FROM PLC.NEGSIG_ATTACK_TYPES
 WHERE attack_type_name = ?
EndQuery
        $self->enable_sigs_on_param_by_attack_type(%params);
    }
}

#----------------------------------------------------------------------

sub sig_set_on_policy {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};

    my $call_flags = F5::ASMConfig::CallFlags->new;
    my $client = $self->{client};
    my $policy = $self->{policy};

    my $set_id_name = F5::NegativeSignatures::Utils::get_set_id_name(
                          dbh=>$self->{dbh},
                          attack_type_id => $vuln->{attack_type_id});

    my ($new_sigset, $set_name) = F5::NegativeSignatures::Utils::get_signature_set(asm_config_handler => $client,
                                                                                   call_flags => $call_flags,
                                                                                   policy => $policy,
                                                                                   get_messages_only => 0,
                                                                                   base_set_name => $set_id_name->{set_name}, );
    my $set_id = $new_sigset->{set_id};
    my $policy_sig_set_search_result = $client->search_policy_signature_set(
                                       $client->{session},
                                       $call_flags,
                                       $policy,
                                       F5::ASMConfig::PolicySignatureSet->new({set_id => $set_id}));

    my $policy_sig_set;
    if ($policy_sig_set_search_result->{result_record_count}) {
        $policy_sig_set = $policy_sig_set_search_result->{element_list}->[0]->{policy_signature_set};
    }

    if (!$policy_sig_set) {
        if ($get_messages_only) {
            push (@{$retval->{pre_resolve_messages}}, {
                message_text => $self->{pre_resolve_templates}->{unassociated_set},
                bind_values => [ $set_name ],
            });
        }
        else {
            $client->add_policy_signature_set(
                $client->{session},
                $call_flags,
                $policy,
                F5::ASMConfig::PolicySignatureSet->new({set_id => $set_id, flg_alarm => 1, flg_reject => 1, flg_learn => 1}));

        }
    }
    elsif (!$policy_sig_set->{flg_alarm} || !$policy_sig_set->{flg_reject} || !$policy_sig_set->{flg_learn}) {
        if ($get_messages_only) {
            if ( $policy->{has_parent} && $policy_sig_set->{is_inherited} ) {
                $self->add_inheritance_warning(%params, inherited_entity => $policy_sig_set);
            }
            else {
                push (@{$retval->{pre_resolve_messages}}, {
                    message_text => $self->{pre_resolve_templates}->{set_on_policy_is_not_blocking},
                    bind_values => [ $set_name ],
                });
            }
        }
        else {
            $client->update_policy_signature_set(
                $client->{session},
                $call_flags,
                $policy,
                F5::ASMConfig::PolicySignatureSet->new({set_id => $set_id, flg_alarm => 1, flg_reject => 1, flg_learn => 1}));

        }
    }

    my $sig_set_search_result = $client->search_signature_set(
                                    $client->{session},
                                    $call_flags,
                                    F5::ASMConfig::PolicySignatureSet->new({set_id => $set_id}));

    my $sig_set;

    if ($sig_set_search_result->{result_record_count}) {
        $sig_set = $sig_set_search_result->{element_list}->[0]->{signature_set};
    }

    push @{$call_flags->{search_filters}},
         F5::ASMConfig::SearchFilter->new({
             field_name => 'sig_id',
             field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN,
             field_value_list => $sig_set->{sigs}, });

    my $sigs_result = $client->search_policy_signature(
                $client->{session},
                $call_flags,
                $policy,
                F5::ASMConfig::PolicySignature->new({flg_quarantined => 1}));

    my $policy_negsig_attributes = $client->get_policy_negsig_attributes(
        $client->{session},
        undef,
        $policy,
        F5::ASMConfig::PolicyNegsigAttributes->new({}));

    if ($policy_negsig_attributes->{flg_quarantine_signatures} && !$staging && $sigs_result->{result_record_count}) {
        if ($get_messages_only) {
            push (@{$retval->{pre_resolve_messages}}, {
                message_text => $self->{pre_resolve_templates}->{policy_sig_staging_is_on},
                bind_values => [ $set_name ],
            });
        }
        else {
            $client->update_many_policy_signatures(
                $client->{session},
                $call_flags,
                $policy,
                F5::ASMConfig::PolicySignature->new(),
                F5::ASMConfig::PolicySignature->new({flg_quarantined => 0}));
        }
    }

    return {set_id => $set_id, set_name => $set_name};
}

#----------------------------------------------------------------------
sub get_or_create_sig_set_on_policy{
    my ($self, %params) = @_;

    my $get_messages_only = $params{get_messages_only};
    my $call_flags = F5::ASMConfig::CallFlags->new;
    my $client = $self->{client};
    my $policy = $self->{policy};
    my $system_name = $params{system_name};

    return if !$system_name;

    my $rv = F5::NegativeSignatures::Utils::create_and_assign_system_signatures_set( asm_config_handler => $client,
                                                                                     session => $self->{session},
                                                                                     call_flags => $call_flags,
                                                                                     policy => $policy,
                                                                                     get_messages_only => $get_messages_only,
                                                                                     system_name => $system_name );
   return $rv;
}
#----------------------------------------------------------------------

sub enable_clickjacking_protection {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};

    my @urls;
    if (ref $retval->{urls}) {
        @urls = @{$retval->{urls}};
    }
    else {
        @urls = ($retval->{url});
    }

    for my $url (@urls) {
        if ($get_messages_only) {
            push (@{$retval->{pre_resolve_messages}}, 
                  {
                   message_text => $self->{pre_resolve_templates}->{enable_clickjacking},
                   bind_values => [ $url->{url_name} ],
                  });
        }
        else {
            $url->{flg_clickjacking_protection} = 1;
            $url->{clickjacking_type} = F5::ASMConfig::ClickjackingType::CLICKJACKING_ONLY_SAME;
            if (defined $params{staging}) {
                $url->{in_staging} = $params{staging};
            }
        }
    }

    return $retval if ($get_messages_only);

    my $call_flags = F5::ASMConfig::CallFlags->new;
    my $client = $self->{client};
    my $policy = $self->{policy};

    for my $url (@urls) {
        $client->update_url($client->{session}, $call_flags, $policy, $url);
    }
}

#----------------------------------------------------------------------

sub disallow_repeated_occurrences_on_param {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};
    my $policy = $params{policy};
    my $client = $self->{client};
    my $call_flags = F5::ASMConfig::CallFlags->new;

    for my $param (@{$retval->{params}}) {
        if (defined $params{staging}) {
            $param->{in_staging} = $params{staging};
        }

        if ( $get_messages_only ) {
            if ( $policy->{has_parent} ) {
                my $rv = $client->search_parameter($client->{session}, undef, $policy, $param);
                if ( $rv->{result_record_count} ) {
                my $param_rv = $rv->{element_list}->[0]->{parameter};
                    if ( $param_rv->{is_inherited} ) {
                        $self->add_inheritance_warning(%params, inherited_entity => $param_rv);
                    }
                }
            }
        }
        else {
            $param->{flg_allow_repeated_parameter_name} = 0;
            $param = $client->update_parameter($client->{session}, $call_flags, $policy, $param);
        }
    }

    return $retval;
}

#----------------------------------------------------------------------

sub enable_sigs_on_param_by_attack_type {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};

    my $call_flags = F5::ASMConfig::CallFlags->new;
    my $client = $self->{client};
    my $policy = $self->{policy};

    my @sigs = F5::ASMConfig::Entity::Signature->retrieve_all(
                   dbh => $self->{dbh},
                   get_criteria => {
                     attack_type_id => $vuln->{attack_type_id},
                     flg_has_valuecontent_http => 1 });

    my @sig_overrides = map { F5::ASMConfig::SignatureOverride->new({
                                 sig_id => $_->{sig_id},
                                 is_enabled => 1})
                           } @sigs;

    $self->sig_set_on_policy(%params);
    if ( !$retval->{params} ) {
        return $self->enable_sigs_on_url_by_attack_type(%params);
    }
    for my $param (@{$retval->{params}}) {
        if (!$param->{flg_check_signatures}) {
            if ($get_messages_only) {
                if ($retval->{url}->{url_name}) {
                    if ( $policy->{has_parent} ) {
                        my $rv = $client->search_parameter($client->{session}, undef, $policy, $param);
                        if ( $rv-> {result_record_count} ) {
                            my $param_rv = $rv->{element_list}->[0]->{parameter};
                            if ($param_rv->{is_inherited} ) {
                                $self->add_inheritance_warning(%params, inherited_entity => $param_rv);
                                next;
                            }
                        }
                    }
                    push (@{$retval->{pre_resolve_messages}}, {
                                                           message_text => $self->{pre_resolve_templates}->{no_attack_sig_check_on_param},
                                                           bind_values => [
                                                                           $param->{param_name} .
                                                                           " on [" .
                                                                           ($retval->{url}->{app_id} ? 'HTTPS]' : 'HTTP]') .
                                                                           $retval->{url}->{url_name} ],
                                                          });
                }
            }
            else {
                $param->{flg_check_signatures} = 1;
                $param->{in_staging} = $staging;
                $client->update_parameter(
                    $client->{session},
                    $call_flags,
                    $policy,
                    $param);
            }
        }
    }

    return $retval if $get_messages_only;

    # resolve actions

    for my $param (@{$retval->{params}}) {
        $call_flags->{append_nested_struct_on_update} = 1;
        $param->{param_negsig_overrides} = \@sig_overrides;
        $param->{in_staging} = $staging;
        $param = $client->update_parameter($client->{session}, $call_flags, $policy, $param);
    }

}

#----------------------------------------------------------------------
sub enable_sigs_on_url_by_attack_type {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};

    my $call_flags = F5::ASMConfig::CallFlags->new;
    my $client = $self->{client};
    my $policy = $self->{policy};

    my @sigs = F5::ASMConfig::Entity::Signature->retrieve_all(
                   dbh => $self->{dbh},
                   get_criteria => {
                     attack_type_id => $vuln->{attack_type_id},
                      });

    my @sig_overrides = map { F5::ASMConfig::SignatureOverride->new({
                                 sig_id => $_->{sig_id},
                                 is_enabled => 1})
                           } @sigs;

    $self->sig_set_on_policy(%params);
    for my $url (@{$retval->{urls}}) {
        if (!$url->{flg_check_signatures}) {
            if ($get_messages_only) {
                if ($retval->{url}->{url_name}) {
                    if ( $policy->{has_parent} ) {
                        my $rv = $client->search_url($client->{session}, undef, $policy, $url);
                        if ( $rv-> {result_record_count} ) {
                            my $url_rv = $rv->{element_list}->[0]->{url};
                            if ($url_rv->{is_inherited} ) {
                                $self->add_inheritance_warning(%params, inherited_entity => $url_rv);
                                next;
                            }
                        }
                    }
                    push (@{$retval->{pre_resolve_messages}}, {
                                                           message_text => $self->{pre_resolve_templates}->{no_attack_sig_check_on_url},
                                                           bind_values => [
                                                                           $url->{url_name}
                                                                          ]
                                                          });
                }
            }
            else {
                $url->{flg_check_signatures} = 1;
                $url->{in_staging} = $staging;
                $client->update_url(
                    $client->{session},
                    $call_flags,
                    $policy,
                    $url);
            }
        }
    }

    return $retval if $get_messages_only;

    # resolve actions

    for my $url (@{$retval->{urls}}) {
        $call_flags->{append_nested_struct_on_update} = 1;
        $url->{url_negsig_overrides} = \@sig_overrides;
        $url->{in_staging} = $staging;
        $url = $client->update_url($client->{session}, $call_flags, $policy, $url);
    }
}


#----------------------------------------------------------------------

sub disallow_upload_of_exe_on_param {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};

    my $call_flags = F5::ASMConfig::CallFlags->new;
    my $client = $self->{client};
    my $policy = $self->{policy};
    my $get_messages_only = $params{get_messages_only};

    for my $param (@{$retval->{params}}) {
        if ( $get_messages_only ) {
            if ( $policy->{has_parent} ) {
                my $rv = $client->search_parameter($client->{session}, undef, $policy, $param);
                if ( $rv->{result_record_count} ) {
                    my $param_rv = $rv->{element_list}->[0]->{parameter};
                    if ( $param_rv->{is_inherited} ) {
                        $self->add_inheritance_warning(%params, inherited_entity => $param_rv);
                    }
                }
            }
        }
        else {
            $param->{flg_disallow_upload_of_exe} = 1;
            $param->{in_staging} = $staging;
            $client->update_parameter($client->{session}, $call_flags, $policy, $param);
        }
    }
    return $retval;
}

#----------------------------------------------------------------------

sub disallow_method {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $staging = defined $params{staging} ? $params{staging} : 0;

    my $call_flags = F5::ASMConfig::CallFlags->new(ignored_errors => [F5::ASMConfig::IgnoredError->new({error_code => F5::ASMConfig::ErrorCode::NONEXISTENT_RECORD()})]);
    my $client = $self->{client};
    my $policy = $self->{policy};
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};

    for my $method (@{$params{internal_params}}) {
       if ( $get_messages_only ) {
            if ( $policy->{has_parent} ) {
                my $rv = $client->search_allowed_method($client->{session}, undef, $policy, F5::ASMConfig::AllowedMethod->new({method_name => $method}));
                if ( $rv->{result_record_count} ) {
                    my $allowed_method_rv = $rv->{element_list}->[0]->{allowed_method};
                    if ( $allowed_method_rv->{is_inherited} ) {
                        $self->add_inheritance_warning(%params, inherited_entity => $allowed_method_rv);
                    }
                }
            }
        }
        else {
            $client->delete_allowed_method($client->{session}, $call_flags, $policy, F5::ASMConfig::AllowedMethod->new({method_name => $method}));
        }
    }

    return $retval;
}

#----------------------------------------------------------------------

sub store_import_value_map {
    my ($self, %params) = @_;

    my $element = $params{element};
    my $config = $self->{config};
    my $map_def = $params{map_def};

    my $field_to_push;
    my ($key_field) = grep { $_->{primary_key} } @{$map_def->{fields}};

    for my $field (@{$map_def->{fields}}) {
        if ($field->{conditionXPath}) {
            my $condition_found = $element->findnodes($field->{conditionXPath});
            if (! $condition_found) {
                next;
            }
        }
        $field_to_push->{$field->{field_name}} = F5::XMLUtils->get_value_from_xpath(
                                                                                    xpath => $field->{xpath},
                                                                                    element => $element);
    }

    $self->{import_value_maps}->{$key_field->{field_name}}->{$field_to_push->{$key_field->{field_name}}} = $field_to_push;
}

#----------------------------------------------------------------------

sub parse_and_update_scanner_vuln {
    my ($self, %params) = @_;

    my $twig = $params{twig};
    my $vuln_elm = $params{vuln_elm};
    my $load_time = $params{load_time};
    my $requested_host_names = $params{requested_host_names};
    my $import_all_host_names = $params{import_all_host_names};
    my $parsing_rules_vuln = $params{parsing_rules_vuln};

    my $policy = $self->{policy};
    my $config = $self->{config};
    my $client = $self->{client};
    my $call_flags = F5::ASMConfig::CallFlags->new;

    my @scanner_vulns = $self->retrieve_scanner_entities(
                                                         twig => $twig,
                                                         fields => $parsing_rules_vuln->{fields},
                                                         element => $vuln_elm,
                                                         import_all_host_names => $import_all_host_names,
                                                         requested_host_names => $requested_host_names,);

    return unless @scanner_vulns;

    my $host_name;
    my $vuln_update_count = 0;
    my $has_add_cookie_action;

    for my $scanner_vuln (@scanner_vulns) {
        my $condition = $parsing_rules_vuln->{condition};

        my $vuln_has_add_cookie_action = 0;
        if (($config->{vuln_handling}->{$scanner_vuln->{scanner_vuln_name}}) &&
            ($config->{vuln_handling}->{$scanner_vuln->{scanner_vuln_name}}->{resolve_actions})) {
            my $resolve_actions = $config->{vuln_handling}->{$scanner_vuln->{scanner_vuln_name}}->{resolve_actions};
            for my $action (@$resolve_actions) {
                if ($action eq 'add_cookie') {
                    $has_add_cookie_action = 1;
                    $vuln_has_add_cookie_action = 1;
                }
            }
        }

        if (($config->{vuln_handling}->{$scanner_vuln->{scanner_vuln_name}}) &&
            ($config->{vuln_handling}->{$scanner_vuln->{scanner_vuln_name}}->{condition})) {
            $condition = $config->{vuln_handling}->{$scanner_vuln->{scanner_vuln_name}}->{condition};
        }

        if (defined $condition) {
            my $flg_continue = $self->check_insert_condition(
                                                             condition => $condition,
                                                             element => $vuln_elm,);

            next unless $flg_continue;
        }

        $host_name = $self->get_host_name( scanner_vuln => $scanner_vuln );

        if (! ($host_name || ($self->{scanner_name} eq 'hailstorm' && $vuln_has_add_cookie_action)) ) {
            next;
        }

        if (!$import_all_host_names &&((!defined $requested_host_names) || (!exists $requested_host_names->{$host_name}))) {
            next;
        }

        $call_flags->{add_on_updating_nonexistent_record} = 1;
        $call_flags->{suppress_events} = 1;


        my $attack_vector = F5::ASMConfig::Entity::ScannerVuln->new(
                                                                    dbh => $self->{dbh},
                                                                    policy => $policy,
                                                                    get_criteria =>  $scanner_vuln );

        if ( ! defined $vuln_elm->{att}->{tested}) {
            $scanner_vuln->{scanner_attack_vector_id} = 0;
        }

        if (scalar $attack_vector && defined $vuln_elm->{att}->{tested}) {
            # need to check whether to update
            my @tested_time_in_xml = $vuln_elm->{att}->{tested};
            my @tested_time_in_db = $attack_vector->{tested_time};

            # if the tested time in the xml didn't change since the last db update, then we assume that the outcome id old
            # meaning that retest is not completed yet, so we'll not update anything.
            if ((scalar @tested_time_in_xml) &&
                (scalar @tested_time_in_db) &&
                ($tested_time_in_xml[0] le $tested_time_in_db[0])) {

                next;
            }
        }

        $scanner_vuln->{load_time} = $load_time;
        $client->update_scanner_vuln($client->{session}, $call_flags, $policy, $scanner_vuln);
        $vuln_update_count++;
    }

    if (! $host_name) {
        if ($self->{scanner_name} eq 'hailstorm') {
            if (! $has_add_cookie_action) {
                return;
            }
        }
        else {
            return;
        }
    }

    return { host_name => $host_name,
             vuln_update_count => $vuln_update_count,
             vuln_count => scalar @scanner_vulns,
           };
}

#----------------------------------------------------------------------

sub check_insert_condition {
    my ($self, %params) = @_;

    my $condition = $params{condition};
    my $element = $params{element};

    my $field_value = F5::XMLUtils->get_value_from_xpath(
                          xpath => $condition->{xpath},
                          element => $element);

    if (grep {$field_value eq $_} @{$condition->{values}}) {
        return 1;
    }

    return;
}
#----------------------------------------------------------------------

sub retrieve_scanner_entities {
    my ($self, %params) = @_;

    my $twig = $params{twig};
    my @fields = @{$params{fields}};
    my $element = $params{element};
    my $requested_host_names = $params{requested_host_names};
    my $import_all_host_names = $params{import_all_host_names};

    my $config = $self->{config};

    my @vulns = (F5::ASMConfig::ScannerVuln->new);
    my $asm_cookie_prefix_object = F5::ASM::COAPI::Query->new("db_variable", { name => 'asm.cookie_prefix'})->[0];
    my $ASM_COOKIE_PREFIX = $asm_cookie_prefix_object->{value} ? $asm_cookie_prefix_object->{value} : $asm_cookie_prefix_object->{default};

    for my $field (@fields) {

        my $field_value;
        if ($field->{xpath}) {
            my $get_value = 1;
            if (defined $field->{condition}) {
                $get_value = $self->check_insert_condition(
                                                         condition => $field->{condition},
                                                         element => $element,);
            }

            if ( $get_value ) {
                $field_value = F5::XMLUtils->get_value_from_xpath(
                                                              xpath => $field->{xpath},
                                                              element => $element);
            }
        }

        if ($requested_host_names || $import_all_host_names) {
            if ($field->{value_map}) {
                my $map_name = $field->{value_map}->{map_name};
                my $map_field = $field->{value_map}->{map_field};
                $field_value = $self->{import_value_maps}->{$map_name}->{$field_value}->{$map_field};
            }

            if (defined $field->{regexp}) {
                eval {
                    if ($field_value) {
                        $field_value =~ /$field->{regexp}/s;
                        $field_value = $1;
                    }
                };
            }

            if (defined $field->{value_func}) {
                my $method_name = $field->{value_func};
                next if $method_name =~ /md5sum/;
                $field_value = $self->$method_name(
                                                   twig => $twig,
                                                   field_value => $field_value,
                                                   elm => $element,
                                                   field_def => $field,
                                                  );
            }
        }

        if (ref($field_value) && ref($field_value) eq 'ARRAY') {
            # Multiple values

            # Ensure that there are enough vulns slots to accommodate all values
            while($#vulns < $#{$field_value}) {
                push(@vulns, dclone($vulns[0]));
            }

            for my $i (0..$#{$field_value}) {
                $field_value->[$i] =~ s/\s*$//s if $field_value->[$i];
                $vulns[$i]->{$field->{field_name}} = $field_value->[$i];
            }
        }
        else {
            $field_value =~ s/\s*$//s if $field_value;
            for my $vuln (@vulns) {
                $vuln->{$field->{field_name}} = $field_value;
            }
        }
    }

    for my $vuln (@vulns) {
        # Vulnerability suggestions for TS cookies are skipped
        if (defined $vuln->{'cookie'} && $vuln->{'cookie'} =~ /^$ASM_COOKIE_PREFIX[A-Fa-f0-9]{6}/) {
            return;
        }

        if (defined $vuln->{attack_type_id} && ( scalar grep {$vuln->{attack_type_id} == $_}
                                                 values %{$self->{vuln_handling_attack_ids}} )
            && ( scalar grep {$vuln->{scanner_vuln_name} eq $_}
                 keys %{$self->{vuln_handling_attack_ids}} ) ) {
            if (defined $config->{vuln_handling}->{$vuln->{scanner_vuln_name}}->{resolve_actions}) {
                $vuln->{'flg_is_resolvable'} = F5::ASMConfig::ScannerVulnType::SCANNER_VULN_TYPE_RESOLVABLE();
            }
            else {
                $vuln->{'flg_is_resolvable'} = F5::ASMConfig::ScannerVulnType::SCANNER_VULN_TYPE_RESOLVABLE_MANUALLY();
            }
        }
        else {
            $vuln->{'flg_is_resolvable'} = F5::ASMConfig::ScannerVulnType::SCANNER_VULN_TYPE_NOT_RESOLVABLE();
        }

        for my $md5field (grep{ defined $_->{value_func} and
                                  $_->{value_func} =~ /md5sum\((.*)\)/
                              } @fields) {

            my $md5 = Digest::MD5->new;
            my ($md5field_list) = $md5field->{value_func} =~ /md5sum\((.*)\)/;
            my @md5_field_names = split(/,/, $md5field_list);

            for my $md5_field_name (@md5_field_names) {
                my $md5_field_value;
                if ($md5_field_name eq 'request_url') {
                    my ($app_id, $domain, $object, $query) = F5::ParseUri::parse_uri($vuln->{$md5_field_name});
                    $md5_field_value = $app_id . $object;
                }
                else {
                    $md5_field_value = $vuln->{$md5_field_name};
                }

                if (defined $md5_field_value) {
                    $md5->add(encode_utf8($md5_field_value));
                }
            }

            $vuln->{$md5field->{field_name}} = $md5->hexdigest;
        }
    }

    return @vulns;
}

#----------------------------------------------------------------------

sub threat_and_severity {
    my ($self, %params) = @_;

    my $original_value = lc($params{field_value});
    my %threat_and_severity = (
        urgent => F5::ASMConfig::SeverityLevel::SEVERITY_URGENT(),
        critical => F5::ASMConfig::SeverityLevel::SEVERITY_CRITICAL(),
        high => F5::ASMConfig::SeverityLevel::SEVERITY_HIGH(),
        medium => F5::ASMConfig::SeverityLevel::SEVERITY_MEDIUM(),
        low => F5::ASMConfig::SeverityLevel::SEVERITY_LOW(),
        info => F5::ASMConfig::SeverityLevel::SEVERITY_INFO(),
    );

    return $threat_and_severity{$original_value};
}

#----------------------------------------------------------------------

sub hailstorm_get_cookie_specific_vuln {
    my ($self, %params) = @_;

    my $elm = $params{elm};
    my $field_value = $params{field_value};

    if ($field_value eq 'Cookie Vulnerabilities') {
        my $message = F5::XMLUtils->get_value_from_xpath(xpath => 'Message', element => $elm);
        my @cookie_vulns;

        if ($message =~ /HTTPOnly attribute/) {
            push(@cookie_vulns, 'Set-Cookie does not use HTTPOnly Attribute');
        }
        if ($message =~ /set insecurely|secure attribute/) {
            push(@cookie_vulns, 'Set-Cookie does not use Secure Attribute');
        }
        if (@cookie_vulns) {
            $field_value = \@cookie_vulns;
        }
    }

    return $field_value;
}

#----------------------------------------------------------------------

sub hailstorm_lookup_by_vuln_handling {
    my ($self, %params) = @_;

    my $field_value = $self->hailstorm_get_cookie_specific_vuln(
                                                             field_value => $params{field_value},
                                                             elm => $params{elm},
                                                             field_def => $params{field_def},
                                                            );

    if (! ref($field_value)) {
        $field_value = [ $field_value ];
    }

    return [ map { $self->{vuln_handling_attack_ids}->{$_} } @$field_value ];
}

#----------------------------------------------------------------------

sub hailstorm_get_cookie {
    my ($self, %params) = @_;

    my $elm = $params{elm};
    my $field_value = undef;

    my $vuln_name = F5::XMLUtils->get_value_from_xpath(xpath => '../../SmartAttackInfo/SmartAttackName', element => $elm);
    if ($vuln_name) {
        $vuln_name =~ /(.*?)(:? - \[.*)?$/s;
        $vuln_name = $1;
        
        if ($vuln_name eq 'Cookie Vulnerabilities') {
            $field_value = $params{field_value};
        }
    }

    return $field_value;
}

sub appscan_get_cookie {
    my ($self, %params) = @_;

    my $elm = $params{elm};

    my $vuln_name = F5::XMLUtils->get_value_from_xpath(xpath => '../@IssueTypeID', element => $elm);
    if ($APPSCAN_COOKIE_VULN_NAMES{$vuln_name}) {
        return $params{field_value};
    }

    return undef;
}

#----------------------------------------------------------------------

sub qualys_get_cookie {
    my ($self, %params) = @_;

    my $elm = $params{elm};
    my $field_value = undef;

    my $vuln_name = F5::XMLUtils->get_value_from_xpath(xpath => 'QID', element => $elm);
    if ($QUALYS_COOKIE_VULN_NAMES{$vuln_name}) {
        my $response_contents = $params{field_value};
        my($response_elm) =  $elm->findnodes('PAYLOADS/PAYLOAD/RESPONSE/CONTENTS');
        if ($response_elm->{att}->{base64} && $response_elm->{att}->{base64} eq 'true') {
            $response_contents = MIME::Base64::decode_base64($response_contents);
        }

        $response_contents =~ /^(.*?)\s*=/;
        $field_value = $1;
    }

    return $field_value;
}

#----------------------------------------------------------------------

sub vuln_status {
    my ($self, %params) = @_;
    my $elm = $params{elm};
    my $field_def = $params{field_def};
    my $field_value = $params{field_value};

    if (defined $field_value and $field_value == 1) {
        $field_value = 'mitigated';
    }
    else {
        $field_value = F5::XMLUtils->get_value_from_xpath(xpath => $field_def->{xpath2}, element => $elm);
    }

    my %attack_vector_status = (
        open => F5::ASMConfig::ScannerVulnStatus::SCANNER_VULN_STATUS_OPEN(),
        closed => F5::ASMConfig::ScannerVulnStatus::SCANNER_VULN_STATUS_CLOSED(),
        mitigated => F5::ASMConfig::ScannerVulnStatus::SCANNER_VULN_STATUS_MITIGATED(),
    );

    return $attack_vector_status{$field_value};
}

#----------------------------------------------------------------------

sub convert_zulu_to_epoch {
    my ($self, %params) = @_;
    my ($time) = $params{field_value};

    return 0 unless $time;
    $time =~ s/T/ /g;
    $time =~ s/Z//g;
    my @time_parts = reverse split(/[- :]/, $time);
    $time_parts[4]--;

    return timelocal(@time_parts);
}

#----------------------------------------------------------------------

sub lookup_by_vuln_handling {
    my ($self, %params) = @_;

    return $self->{vuln_handling_attack_ids}->{$params{field_value}};
}

#----------------------------------------------------------------------

sub get_attack_type_by_vuln_handling {
    my ($self, %params) = @_;

    return ( $self->{vuln_handling_attack_ids}->{$params{field_value}}
                ||  F5::NegativeSignatures::Utils::find_or_add_attack_type ( dbh => $self->{dbh},
                                                                             dbname => 'PLC',
                                                                             attack_type_name => $params{field_value},
                                                                             find_only => '1', )
           );

}

#----------------------------------------------------------------------

sub get_webinspect_cookie {
    my ($self, %params) = @_;

    my $elm = $params{elm};
    my $twig = $params{twig};
    my $field_def = $params{field_def};

    my $vuln_handling = $self->{config}->{vuln_handling};

    my $vuln_name = $elm->first_child('VulnerabilityID');
    return unless $vuln_name;
    my $vuln_def = $vuln_handling->{$vuln_name->text};
    return unless $vuln_def;
    my $resolve_actions = $vuln_def->{resolve_actions};
    return unless $resolve_actions;

    my @cookie_actions = grep { $_ eq 'add_cookie'} @$resolve_actions;

    if (scalar @cookie_actions) {
        my $session_elm = $elm->parent->parent;

        my $detection = $elm->first_child('DetectionSelection');
        if ($detection) {
            my $location = $detection->first_child('Location');
            return undef unless $location;
            
            my $offset = $location->{att}->{offset};
            my $length = $location->{att}->{'length'};
    
            my $raw_response_elm = $session_elm->first_child('RawResponse');
            my $raw_response = $raw_response_elm->text;
    
            $raw_response =~ s/\n/\r\n/gm; # offset is wrong for original raw response
            my $cookie = substr($raw_response, $offset, $length);
    
            if ($cookie =~ /^ *([^;, ]+)=/) {
                my $cookie_name = $1;
                return $cookie_name;
            }
            else {
                return undef;
            }
        }
        else {
            my $request = $session_elm->first_child('Request');
            return undef unless $request;
            my $cookies = $request->first_child('Cookies');
            return undef unless $cookies;
            my $cookie_name = $cookies->first_child('Cookie')->first_child('Name')->text;
            
            return $cookie_name;
        }
    }

    return undef;
}

#----------------------------------------------------------------------

sub get_webinspect_param {
    my ($self, %params) = @_;

    my ($attack_type) = ($params{field_value} =~ /AttackType="/)
                        ? $params{field_value} =~ /AttackType="([^"]+)"/
                        : $params{field_value} =~ /AT:([^,]+)/;
    if (($attack_type) && ($attack_type eq 'Search' || $attack_type eq "None")) {
        return;
    }

    my ($param_name) = ($params{field_value} =~ /AttackParamDesc="/)
                       ? $params{field_value} =~ /AttackParamDesc="([^"]+)"/
                       : $params{field_value} =~ /APD:([^,]+)/;

    return $param_name;
}

#----------------------------------------------------------------------

sub get_parser_status_codes {
    my ( $self) = @_;

    my $dbh = $self->{dbh};
    my $parser_status_codes = $dbh->selectall_hashref(<<EndQuery, 'icontrol_name', undef, 0);
SELECT viol_index, description, icontrol_name
  FROM PLC.VIOLATIONS
WHERE flg_is_user_defined = ?
EndQuery

}

#----------------------------------------------------------------------

sub get_scanner_url {
    my($self, %params) = @_;

    my $bigip_redirect_url = $params{bigip_redirect_url};
    my $url = $params{url};
    my $config = $self->{config}->{api};

    my $response = $self->do_scanner_service_request(
                              url => $config->{auth_redirect_request_url},
                              client_id => $config->{client_id},);

    my $content = from_json($response->content);
    my $policy = $self->{policy};
    my $client = $self->{client};
    $self->{scanner_config}->{api_key} = $content->{AccessInfo}->{AccessToken};

    $self->update_scanner_config( scanner_config => $self->{scanner_config});

    return $self->do_scanner_service_request(
                      url => $url,
                      client_id => $config->{client_id},
                      bigip_redirect_url => $bigip_redirect_url,
                      token_request_id => $content->{AccessInfo}->{TokenRequestId},
                      get_url_only => 1,);
}

#----------------------------------------------------------------------

sub get_scanner_register_url {
    my($self, %params) = @_;

    my $config = $self->{config}->{api};
    $self->get_scanner_url(%params, url => $config->{register_redirect_url_to_return});
}

#----------------------------------------------------------------------

sub get_scanner_auth_url {
    my($self, %params) = @_;

    my $config = $self->{config}->{api};
    $self->get_scanner_url(%params, url => $config->{auth_redirect_url_to_return});
}

#----------------------------------------------------------------------

sub get_scanner_site_names_url {
    my($self, %params) = @_;

    return $self->get_full_url(url_type => 'site_names_url');
}

#----------------------------------------------------------------------

sub get_scanner_subscriptions {
    my($self, %params) = @_;

    my $config = $self->{config}->{api};

    my $response = $self->do_scanner_service_request(url => $config->{get_scanner_subscriptions_url},
                                                     scanner_config => $params{scanner_config},
                                                     client_id => $config->{client_id},);

    my $subs = from_json($response->content);
    die "Cannot get subscription data" if $subs->{ReturnCode};

    my @subscriptions_to_return;

    for my $subscription (@{$subs->{Subscriptions}}) {
        my $subscription_to_return;
        next unless $subscription->{Url};
        $subscription_to_return->{subscription_id} = $subscription->{SubscriptionId};
        $subscription_to_return->{url} = $subscription->{Url};
        $subscription_to_return->{product_id} = $subscription->{ProductId};

        my @scans;
        $response = $self->do_scanner_service_request(url => $config->{get_scanner_subscription_scans_url},
                                                      scanner_config => $params{scanner_config},
                                                      subscription_id => $subscription_to_return->{subscription_id},);

        next unless $response;

        my $scans = from_json($response->content);

        for my $scan (@{$scans->{Scans}}) {
            next if $scan->{Status} ne 'Complete';
            $scan->{CompleteDate} =~ s/\/Date\(//;
            $scan->{CompleteDate} =~ s/\d\d\d\-\d\d\d\d\)\/$//;
            push (@scans, {
                scan_id => $scan->{ScanId},
                status => $scan->{Status},
                type => $scan->{ScanType},
                completed_date => $scan->{CompleteDate},
            });
        }
        $subscription_to_return->{subscription_scans} = \@scans;
        push (@subscriptions_to_return, $subscription_to_return);
    }

    return \@subscriptions_to_return;
}

#----------------------------------------------------------------------

sub complete_scanner_auth {
    my($self, %params) = @_;

    my $config = $self->{config}->{api};
    my $policy = $self->{policy};
    my $client = $self->{client};
    my $call_flags = F5::ASMConfig::CallFlags->new;

    die "Cannot finalize AccessToken" if $params{qs_map}->{return_code};

    return $self->update_scanner_config(
               scanner_config => F5::ASMConfig::ScannerConfig->new({
                                                                    flg_is_api_key_final => 1,
                                                                    scanner_site_name => $params{qs_map}->{customer_id},
                                                                    scan_requested_time => 0,
                                                                    scanner_auth_data_2 => ''}));
}

#----------------------------------------------------------------------

sub complete_global_scanner_auth {
    my($self, %params) = @_;

    my $client = $params{client};
    my $qs_map = $params{qs_map};

    die "Cannot finalize AccessToken" if $qs_map->{return_code};

    return $self->update_scanner_config(
               scanner_config => F5::ASMConfig::GlobalScannerConfig->new(scanner_type => 'hailstorm',
                                                                         flg_is_setting_global => 1,
                                                                         scanner_site_name => $qs_map->{customer_id},));
}


#----------------------------------------------------------------------

sub enqueue_scanner_scan {
    my($self, %params) = @_;

    my $config = $self->{config}->{api};

    my $response = $self->do_scanner_service_request(url => $config->{enqueue_scanner_scan_url},
                                                     subscription_id => $params{subscription_id},
                                                     method => 'POST');

    die "Cannot connect to scanner server\n" unless $response;

    my $scan = from_json($response->content);

    if ($scan->{ReturnCode} == 404 || ($scan->{ReturnCode} >= 300 && $scan->{ErrorInfo} && $scan->{ErrorInfo}->{ErrorMessage})) {
        my $errorMessage = $scan->{ErrorInfo} ? $scan->{ErrorInfo}{ErrorMessage} : "Unknown scanner server failure\n";
        my $logger = $self->get_logger;
        $logger->info($errorMessage) if $logger;
        die $errorMessage;
    }

    my $scan_id = $scan->{ScanId};
    my $policy = $self->{policy};
    my $client = $self->{client};

    $self->update_scanner_config(
        scanner_config => F5::ASMConfig::ScannerConfig->new({ scanner_auth_data_2 => $scan_id,
                                                              scanner_auth_data_1 => $params{subscription_id},
                                                              scan_requested_time => time(),}));

     return $scan_id;
}

#----------------------------------------------------------------------

sub terminate_scanner_scan {
    my($self, %params) = @_;

    my $config = $self->{config}->{api};

    my $response = $self->do_scanner_service_request(url => $config->{terminate_scanner_scan_url},
                                                     subscription_id => $params{subscription_id},
                                                     content => $config->{terminate_scanner_scan_url_content},
                                                     method => 'PUT');

    die "Cannot connect to scanner server\n." unless $response;

    my $scan = from_json($response->content);
    my $policy = $self->{policy};
    my $client = $self->{client};

    $self->update_scanner_config(
        scanner_config => F5::ASMConfig::ScannerConfig->new({ scanner_auth_data_2 => '', scan_requested_time => 0,}));
}

#----------------------------------------------------------------------

sub policy_check_scan_results { }

#----------------------------------------------------------------------

sub get_secret { return ''; }

#----------------------------------------------------------------------

sub do_scanner_service_request {
    my ($self, %params) = @_;

    my $code = $params{code};
    my $bigip_redirect_url = $params{bigip_redirect_url};
    my $method = $params{method} ? $params{method} : 'GET';
    my $get_url_only = $params{get_url_only};
    my $token_request_id = $params{token_request_id};
    my $content = $params{content};
    my $content_type = $params{content_type};
    my $allow_non_successful_response_code = $params{allow_non_successful_response_code};

    my $policy = $self->{policy};
    my $config = $self->{config}->{api};
    my $client = $self->{client};
    my $call_flags = F5::ASMConfig::CallFlags->new;

    my $logger = $self->get_logger;

    my $scanner_config = defined $params{scanner_config}
                         ? $params{scanner_config}
                         : $self->{scanner_config};
    my $scan_id = defined $params{scan_id} ? $params{scan_id} : $scanner_config->{scanner_auth_data_2};
    my $subscription_id = defined $params{subscription_id} ? $params{subscription_id} : $scanner_config->{scanner_auth_data_1};

    my $f5secret = $self->get_secret;

    my $url;
    eval {
        $url = $self->replace_url_parts($params{url}, {
                                                       API_SERVER => $self->get_scan_server_address(server_type => 'api'),
                                                       KEY_SERVER => $self->get_scan_server_address(server_type => 'key'),
                                                       CUSTOMER_ID => $scanner_config->{scanner_site_name},
                                                       SCAN_ID => $scan_id,
                                                       TOKEN_REQUEST_ID => $token_request_id,
                                                       CLIENT_ID => $config->{client_id}, });
        if (defined $subscription_id) {
            $url = $self->replace_url_parts($url, {SUBSCRIPTION_ID => $subscription_id});
        }
        if (defined $bigip_redirect_url) {
            $url = $self->replace_url_parts($url, {REDIRECT_URL => $bigip_redirect_url});
        }
    };

    return $url if $get_url_only;

    my $ua = get_ua(timeout => $self->{outbound_call_timeout});
    my $request = HTTP::Request->new( $method => encode_utf8($url) );

    # setup headers and content
    $request->header('CTSAuth' => "client_secret=$f5secret, access_token=$scanner_config->{api_key}");
    if ($content_type) {
        $request->header('Content-Type' => $content_type);
    }
    if ($content) {
        $request->header('Content-Length' => length($content));
        $request->content($content);
        if ($config->{expect_100_continue}) {
            $request->header('Expect' => '100-continue');
        }
    } elsif ($method eq 'POST') {
        $request->header('Content-Length' => 0);
    }

    $logger->debug("Sending request to $url with CTSAuth header: client_secret=$f5secret, access_token=$scanner_config->{api_key}");

    my $response = $ua->request($request);

    if (! $response->is_success) {
        $logger->info("Could not get response to $url ... retrying");
        $response = $ua->request($request);
        if (! $response->is_success) {
            if ($allow_non_successful_response_code) {
                return $response;
            }
            $logger->info("Could not get response to $url");
            die ("Could not get response to $url\n");
        }
    }

    $logger->debug("Request successful. Response content is:\n" . $response->content);

    return $response;
}

#----------------------------------------------------------------------

sub status_hierarchy_by_xml_status {
  return \%status_hierarchy_by_xml_status;
}

#----------------------------------------------------------------------

sub status_hierarchy_by_order {
  return \%status_hierarchy_by_order;
}

#----------------------------------------------------------------------

sub get_scan_server_address {
    return '';
}

#----------------------------------------------------------------------

sub get_register_url {}

#----------------------------------------------------------------------

sub get_and_store_api_key {}

#----------------------------------------------------------------------

sub get_scanner_more_info_url {}

#----------------------------------------------------------------------

sub get_global_scanner_more_info_url {}

#----------------------------------------------------------------------

sub get_scanner_more_info_url_with_subscription {}

#----------------------------------------------------------------------

sub get_free_scan_results {}

#----------------------------------------------------------------------

sub replace_url_parts {
    my ($self, $url, $replace_values) = @_;

    for my $key (keys %$replace_values) {
        my $replace_value = $replace_values->{$key};
        if (defined $replace_value) {
            if (ref($replace_value)) {
                $replace_value = $replace_value->{$self->{scanner_config}->{scanner_type}};
            }
        }
        else {
            $replace_value = '';
        }

        $url =~ s/%$key%/$replace_value/g;
    }

    return $url;
}

#------------------------------------------------------------------

sub update_scanner_config {
    my ($self, %params) = @_;

    if ($params{scanner_config}->{flg_is_setting_global}) {
        return $self->{client}->update_global_scanner_config(
                   $self->{client}->{session},
                   F5::ASMConfig::CallFlags->new(add_on_updating_nonexistent_record => 1),
                   $params{scanner_config});
    }
    else {
        return $self->{client}->update_scanner_config(
                   $self->{client}->{session},
                   undef,
                   $self->{policy},
                   $params{scanner_config});
    }
}

#----------------------------------------------------------------------

sub get_scanner_upload_status {
    my($self, %params) = @_;
    my $logger = $self->get_logger;
    my $config = $self->{config}->{api};
    my $url = $self->replace_url_parts($config->{upload_status_url}, {
            API_KEY => $self->{scanner_config}->{api_key},
            SITE_NAME => $self->{scanner_config}->{scanner_site_name},
            SCANNER_URL => $config->{url},
            PORT => $config->{port},});
    my $response = $self->do_scanner_service_request(url => $url,
        allow_non_successful_response_code => 1);
    $logger->trace($response->decoded_content);
    return $response;
}

#----------------------------------------------------------------------

sub get_scanner_site_map_filename {
    my($self, %params) = @_;
    my $stamp = $params{stamp} || '';
    return "$SITEMAP_BASE_DIR/$params{stage}/sitemap$self->{policy}->{policy_id}$stamp.json";
}

#----------------------------------------------------------------------

sub add_inheritance_warning {
    my($self, %params) = @_;
    my $inherited_entity = $params{inherited_entity};
    my $retval = $params{retval};

    my $element_type_display_name = $inherited_entity->get_element_type_display_name;
    push (@{$retval->{pre_resolve_messages}}, {
            message_text => $element_type_display_name,
            bind_values => [ $inherited_entity->get_full_display_name ],
            vuln_id => $params{vuln}{scanner_vuln_id},
            flg_inheritance_warnings => 1,
    });
}

#----------------------------------------------------------------------

sub post_scanner_site_map_data {
    my($self, %params) = @_;
    my $config = $self->{config}->{api};
    my $logger = $self->get_logger;

    my $url = $self->replace_url_parts($config->{post_site_map_url}, {
            API_KEY => $self->{scanner_config}->{api_key},
            SITE_NAME => $self->{scanner_config}->{scanner_site_name},
            SCANNER_URL => $config->{url},
            PORT => $config->{port},});
    my $sitemap_filename = $self->get_scanner_site_map_filename(stage => 'new');

    # read site map information into memory and post to the scanner website
    $logger->info("Post data from $sitemap_filename");
    open(my $sitemap_fh, '<:encoding(UTF-8)', $sitemap_filename) or die "cannot open file $sitemap_filename";
    my $sitemap_data = join('', <$sitemap_fh>);
    close($sitemap_fh);
    my $response = $self->do_scanner_service_request(url => $url,
        method => 'POST', content => encode_utf8($sitemap_data),
        content_type => 'application/json; charset=utf-8', allow_non_successful_response_code => 1);
    if ($response->is_success) {
        # move data file into sent folder
        my $stamp = strftime("-%Y%m%d%H%M%S", localtime);
        my $sent_filename = $self->get_scanner_site_map_filename(stage => 'sent',
            stamp => $stamp);
        move($sitemap_filename, $sent_filename);

        # delete old site maps - keep 4 files
        my $policy_id = ${self}->{policy}->{policy_id};
        my @files = sort glob("$SITEMAP_BASE_DIR/sent/sitemap${policy_id}-*.json");
        if (scalar @files > 4) {
            my $last_file_idx = scalar @files - 5;
            foreach my $filename (@files[0..$last_file_idx]) {
                unlink $filename;
            }
        }
        $logger->info("Scanner site map information was posted - $sent_filename");
    } else {
        $logger->error("Posting site map information failed (code " . $response->code . ")");
    }

    # log response information
    if ($logger->is_debug()) {
        eval {
            my $response_json = from_json($response->content);
            if (exists $response_json->{HttpRequestId}) {
                $logger->debug("Posted sitemap data associated with request ID $response_json->{HttpRequestId}");
            }
        };
        $logger->trace($response->content);
    }
    return $response;
}

#--------------------------------------------------------------------------

sub add_header {
    my ($self, %params) = @_;

    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};

    my $client = $self->{client};
    my $policy = $self->{policy};

    my $header_names = $params{internal_params};
    my @headers;
    my $call_flags = F5::ASMConfig::CallFlags->new(ignore_duplicate_record_on_add => 1);

    for my $header_name (@$header_names) {

        my $search_header = F5::ASMConfig::Header->new({header_name => $header_name, is_wildcard => 0});
        my $header_rv = $client->search_header(
                                         $client->{session},
                                         undef,
                                         $policy,
                                         $search_header,
                                        );

        my $header;
        if ($header_rv->{result_record_count}) {
            $header = $header_rv->{element_list}->[0]->{header};
        }

        if ((! $get_messages_only) && (! defined $header)) {
            $header = $client->add_header(
                       $client->{session},
                       $call_flags,
                       $policy,
                       F5::ASMConfig::Header->new({header_name => $header_name,is_wildcard => 0 } ));
        }

        push(@headers, $header || $search_header);
    }


    $retval->{header} = $headers[0];
    $retval->{headers} = \@headers;
    return $retval;
}
#----------------------------------------------------------------------
sub enable_sigs_on_header_by_attack_type {
    my ($self, %params) = @_;

    my $vuln = $params{vuln};
    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};

    my $call_flags = F5::ASMConfig::CallFlags->new;
    my $client = $self->{client};
    my $policy = $self->{policy};

    my @sigs = F5::ASMConfig::Entity::Signature->retrieve_all(
                   dbh => $self->{dbh},
                   get_criteria => {
                     attack_type_id => $vuln->{attack_type_id},
                     flg_has_headercontent => 1 });

    my @sig_overrides = map { F5::ASMConfig::SignatureOverride->new({
                                 sig_id => $_->{sig_id},
                                 is_enabled => 1})
                           } @sigs;

    $self->sig_set_on_policy(%params);
    for my $header (@{$retval->{headers}}) {
        if (!$header->{flg_check_signatures}) {
            if ($get_messages_only) {
                if ($retval->{header}->{header_name}) {
                    if ( $policy->{has_parent} ) {
                        my $rv = $client->search_header($client->{session}, undef, $policy, $header);
                        if ( $rv->{result_record_count} ) {
                            my $header_rv = $rv->{element_list}->[0]->{header};
                            if ($header_rv->{is_inherited} ) {
                                $self->add_inheritance_warning(%params, inherited_entity => $header_rv);
                                next;
                            }
                        }
                    }
                    push (@{$retval->{pre_resolve_messages}}, {
                                                           message_text => $self->{pre_resolve_templates}->{no_attack_sig_check_on_header},
                                                           bind_values => [
                                                                           $header->{header_name}
                                                                          ]
                                                          });
                }
            }
            else {
                $header->{flg_check_signatures} = 1;
                $header->{in_staging} = $staging;
                $client->update_header(
                    $client->{session},
                    $call_flags,
                    $policy,
                    $header);
            }
        }
    }

    return $retval if $get_messages_only;

    # resolve actions
    $call_flags->{append_nested_struct_on_update} = 1;
    for my $header (@{$retval->{headers}}) {
        $header->{header_negsig_overrides} = \@sig_overrides;
        $header->{in_staging} = $staging;
        $header = $client->update_header($client->{session}, $call_flags, $policy, $header);
    }
}
#----------------------------------------------------------------------
sub enable_sigs_on_param_by_system {
    my ($self, %params) = @_;

    my $staging = defined $params{staging} ? $params{staging} : 0;
    my $retval = $params{retval};
    my $get_messages_only = $params{get_messages_only};

    my $call_flags = F5::ASMConfig::CallFlags->new;
    my $client = $self->{client};
    my $policy = $self->{policy};
    my $system_name = $params{internal_params} ? $params{internal_params}->[0] : undef;

    return if !$system_name;

    my @sigs = F5::ASMConfig::Entity::Signature->retrieve_all(
                   dbh => $self->{dbh},
                   get_criteria => {
                     systems => [ F5::ASMConfig::SignatureSetSystem->new(system_name => $system_name )],
                     flg_has_valuecontent_http => 1 });

    my @sig_overrides = map { F5::ASMConfig::SignatureOverride->new({
                                 sig_id => $_->{sig_id},
                                 is_enabled => 1})
                           } @sigs;


   my $rv = $self->get_or_create_sig_set_on_policy(%params, system_name => $system_name);
   if ( $rv  && $rv->{return_value} && @{$rv->{return_value}} ){
        push (@{$retval->{pre_resolve_messages}}, { message_text => @{$rv->{return_value}}});
   }

   for my $param (@{$retval->{params}}) {
        if (!$param->{flg_check_signatures}) {
            if ($get_messages_only) {
                if ($retval->{url}->{url_name}) {
                    if ( $policy->{has_parent} ) {
                        my $rv = $client->search_parameter($client->{session}, undef, $policy, $param);
                        if ( $rv-> {result_record_count} ) {
                            my $param_rv = $rv->{element_list}->[0]->{parameter};
                            if ($param_rv->{is_inherited} ) {
                                $self->add_inheritance_warning(%params, inherited_entity => $param_rv);
                                next;
                            }
                        }
                    }
                    push (@{$retval->{pre_resolve_messages}}, {
                                                           message_text => $self->{pre_resolve_templates}->{no_attack_sig_check_on_param},
                                                           bind_values => [
                                                                           $param->{param_name} .
                                                                           " on [" .
                                                                           ($retval->{url}->{app_id} ? 'HTTPS]' : 'HTTP]') .
                                                                           $retval->{url}->{url_name} ],
                                                          });
                }
            }
            else {
                $param->{flg_check_signatures} = 1;
                $param->{in_staging} = $staging;
                $client->update_parameter(
                    $client->{session},
                    $call_flags,
                    $policy,
                    $param);
            }
        }
    }

    return $retval if $get_messages_only;

    # resolve actions
    for my $param (@{$retval->{params}}) {
        $call_flags->{append_nested_struct_on_update} = 1;
        $param->{param_negsig_overrides} = \@sig_overrides;
        $param->{in_staging} = $staging;
        $param = $client->update_parameter($client->{session}, $call_flags, $policy, $param);
    }
}
#----------------------------------------------------------------------



1;
