package F5::Scanner::GenericScannerSchema;

use strict;
use warnings;

use F5::Deploy;
use F5::XML::Twig;
use YAML::XS;
use F5::BatchInsert;
use F5::DbUtils;
use F5::NwdUtils;
use F5::XML::Schema::Builder;
use F5::NegativeSignatures::Utils;
use Memoize;

my $UNLIKELY_DELIMITER = '###UNLIKELY_DELIMITER###';
my $ASM_GENERIC_SCANNER_CONFIG_FILENAME  = F5::Deploy::map_path('/usr/share/ts/config/scanners_config.yaml');
my $XMLLINT_CMD = '/usr/bin/xmllint';
my $MAX_MEMORY_USAGE = 200_000_000; # 200 MB approximately
my $MEMORY_CHECK_INTERVAL = 200;

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  if (exists $params{dbh}) {
    $self->{dbh} = $params{dbh};
  }
  else {
    $self->{dbh} = F5::DbUtils::get_dbh();
  }
  $self->{logger} = $params{logger};

  $self->{batch_insert} = F5::BatchInsert->new(dbh => $self->{dbh});
  my $config_filename = $ENV{ASM_GENERIC_SCANNER_CONFIG_FILENAME} || $ASM_GENERIC_SCANNER_CONFIG_FILENAME;
  my $config = YAML::XS::LoadFile($config_filename);

  for my $node (@{$config->{nodes}}) {
       if ( grep { $_ eq 'generic_scanner' } @{$node->{scanners}} ) {
           $self->{config}->{parsing_rules} = $node->{parsing_rules};
           $self->{config}->{vuln_handling} = $node->{vuln_handling};
           last;
       }
  }

  $self->{initial_memory_usage} = F5::NwdUtils::get_process_stat()->{vmsize};
  $self->{memory_check_count} = 0;
}
#----------------------------------------------------------------------

sub build_schema {
  my($self, %params) = @_;

  my $output_filename = $params{output_filename};
  my $xml_document_template = <<EndDoc;
<?xml version="1.0" encoding="utf-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
</xs:schema>
EndDoc

  my $twig = F5::XML::Twig->new(pretty_print => 'indented');
  $twig->parse($xml_document_template);

  my $subnode_base_xpath = $self->{config}->{parsing_rules}->{vuln}->{xpath};

  if ( $subnode_base_xpath =~ /^\/(.*)/){
       $subnode_base_xpath = $1;
  }


  for my $node (@{$self->{config}->{parsing_rules}->{vuln}->{fields}}) {
            F5::XML::Schema::Builder::create_node_schema(node => $node,
                                                         base_twig => $twig,
                                                         node_handler => $self,
                                                         elements_as_choice => {'vulnerability' => 1,},
                                                         subnode_base_xpath => $subnode_base_xpath,
                                        );
  }

  for my $element ($twig->findnodes('//xs:element')) {
        if (!$element->{att}->{minOccurs} && $element->{att}->{name} ne 'vulnerability' && $element->{att}->{name} ne 'scanner_vulnerabilities') {
              $element->set_att(minOccurs => 0, maxOccurs => 1);
        }
  }

  for my $choice_path ( $twig->findnodes('//xs:choice') ){
      $choice_path->set_att(maxOccurs => "unbounded");
      last;
  }

  my $attribute_element = XML::Twig::Elt->new('xs:attribute' => { name => 'version',
                                                                  type => 'xs:string', 
                                                                });
 

  $attribute_element->paste(last_child => $twig->findnodes('//xs:complexType'));

  my $comment_content = "\n Attack types values:\n";
  for my $attack (keys %{$self->{config}->{vuln_handling}} ){
    $comment_content .= "* $attack\n";
  }

  my $schema_attributes = F5::NegativeSignatures::Utils::get_schema_attributes(); 
  my %all_attacks = %{$schema_attributes->{internal_ids}->{attack_type}};

  for my $attack ( keys %all_attacks ){
    if ( not exists $self->{config}->{vuln_handling}->{$attack} ){
       $comment_content .= "* $attack\n";
    }
  }

  my $comment = XML::Twig::Elt->new( '#COMMENT', $comment_content);
  $comment->paste(first_child => $twig->findnodes('//xs:element'));    

  open(my $outf, '>', $output_filename);
  $twig->print($outf);
  close $outf;

}

#----------------------------------------------------------------------

sub check_schema {
  my($filename) = @_;

  my $error_msg = "The generic scanner vulnerabilities file does not conform to the schema and cannot be imported.";

  F5::XML::Schema::Builder::check_schema( filename => $filename,
                                          schema => F5::Deploy::map_path('/ts/share/generic_scanner.xsd'),
                                          error_msg => $error_msg);

}

#----------------------------------------------------------------------

sub finalize_node {
    my($self, %params) = @_;

    my $node = $params{node};
    my $node_element = $params{node_element};

    add_restriction($node, $node_element);

    if ($node->{field_description}) {
        my $comment = XML::Twig::Elt->new( '#COMMENT' => $node->{field_description});
        $comment->paste($node_element);
    }
}

#----------------------------------------------------------------------

sub add_restriction {
    my ($node, $node_element) = @_;

    if ($node->{field_type} || $node->{valid_values} || $node->{field_max_length}) {
        my $field_type = 'string';
        if ($node->{field_type}) {
            $field_type = $node->{field_type};
        }

        my $restriction = XML::Twig::Elt->new('xs:restriction' => {base => 'xs:'. $field_type});
        $restriction->paste(XML::Twig::Elt->new('xs:simpleType' => {})->paste($node_element));

        if (defined $node->{field_max_length}) {
            my $max_length = XML::Twig::Elt->new('xs:maxLength' => {value => $node->{field_max_length}});
            $max_length->paste($restriction);
        }

        if (ref $node->{valid_values}) {
            for my $enum_value (@{$node->{valid_values}}) {
                my $enum_node = XML::Twig::Elt->new('xs:enumeration' => {value => $enum_value});
                $enum_node->paste($restriction);
            }
        }
    }
}

#----------------------------------------------------------------------
sub fixup_node {
  my($self, %params) = @_;
  return;
}
#----------------------------------------------------------------------
1;
