package F5::Scanner::Sentinel;

use strict;
use warnings;

use MIME::Base64;
use JSON;
use HTTP::Request;
use F5::XML::Twig;
use URI::Escape;

use F5::WhiteHatKey;
use F5::License;

use base qw/F5::Scanner/;

sub policy_check_retest_requests {
    my ($self, %params) = @_;

    my $policy = $self->{policy};
    my $config = $self->{config};
    my $client = $self->{client};
    my $call_flags = F5::ASMConfig::CallFlags->new;

    my $scanner_config = $self->{client}->get_scanner_config(
                             $self->{client}->{session}, undef, $self->{policy},);

    push @{$call_flags->{sort_orders}},
         F5::ASMConfig::SortOrder->new({field_name => 'vuln_id', sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_ASC});

    $call_flags->{page_size} = $self->{page_size} + 1;
    $call_flags->{start_page} = 1;
    $call_flags->{grouping_fields} = ['scanner_vuln_id'];

    if ($self->{last_vuln_id}) {
        push @{$call_flags->{search_filters}}, F5::ASMConfig::SearchFilter->new({
                                                   field_name => 'vuln_id',
                                                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GE(),
                                                   field_value => $self->{last_vuln_id},
                                               });
    }

    my $vulns_search_result = $client->search_scanner_vuln (
                                  $client->{session},
                                  $call_flags,
                                  $policy,
                                  F5::ASMConfig::ScannerVuln->new({
                                      retest_status => F5::ASMConfig::ScannerRetestStatus::SCANNER_RETEST_STATUS_REQUESTED,
                                      flg_retest_local => 1,
                                  }),
                              );

    if (! $vulns_search_result->{result_record_count}) {
        $self->{last_vuln_id} = 0;
        return;
    }
    elsif (scalar @{$vulns_search_result->{element_list}} > $self->{page_size}) {
        $self->{last_vuln_id} = $vulns_search_result->{element_list}->[-1]->{scanner_vuln}->{vuln_id};
        pop(@{$vulns_search_result->{element_list}});
    }
    else {
        $self->{last_vuln_id} = 0;
    }
    my $url = $self->get_full_url(
        vuln_ids => [map {$_->{scanner_vuln}->{scanner_vuln_id}} @{$vulns_search_result->{element_list}}],
        url_type => 'retest_status_url'
    );

    $self->get_logger->debug("Sending GET request to Sentinel: $url");

    my $ua = F5::Scanner::get_ua(timeout => $self->{outbound_call_timeout});
    my $request = HTTP::Request->new( GET => $url );

    my $response = $ua->request($request);
    my $xml_data;

    if (! $response->is_success) {
        $self->get_logger->error("Could get response from Sentinel to $url");
        return;
    }

    $xml_data = $response->content;
    $self->get_logger->debug("Success. XML content is: $xml_data");

    my %retest_status;
    my $twig = F5::XML::Twig->new(twig_handlers => { vulnerability => sub {
        my ($twig, $vuln) = @_;

        my @attack_vectors = $vuln->first_child('attack_vectors')->children("attack_vector");
        # we need the first (by order) status wich matches the retest_state
        ($retest_status{$vuln->{att}->{id}}) = map {$self->SUPER::status_hierarchy_by_order()->{$_}->{xml_status}}
                                               sort
                                               map {$self->SUPER::status_hierarchy_by_xml_status()->{$_->{'att'}->{'retest_state'}}->{order} }
                                               @attack_vectors;

        my $store_scanner_vuln = F5::ASMConfig::ScannerVuln->new({
            retest_status => $self->SUPER::status_hierarchy_by_xml_status()->{$retest_status{$vuln->{'att'}->{'id'}}}->{db_status},
        });

        my $scanner_vuln_criteria = F5::ASMConfig::ScannerVuln->new({
            scanner_vuln_id => $vuln->{'att'}->{'id'},
        });

        $client->update_many_scanner_vulns(
            $client->{session},
            undef,
            $policy,
            $scanner_vuln_criteria,
            $store_scanner_vuln,
        );
    }});

    $twig->parse($xml_data);
    $self->save_vuln_file(\$xml_data);

    my @retest_available_vuln_ids = grep {$retest_status{$_} eq 'automatic_retest_available'} keys %retest_status;
    if (@retest_available_vuln_ids) {
        $client->update_scanner_vuln_data(
            $client->{session},
            undef,
            $policy,
            F5::ASMConfig::ScannerActionsFields->new({scanner_vuln_ids => \@retest_available_vuln_ids}),
        );
    }

}

#------------------------------------------------------------------

sub get_xml_data {
    my ($self, %params) = @_;

    my $scanner_url = $params{url};

    my $logger = $self->get_logger();

    my $ua = F5::Scanner::get_ua(timeout => $self->{outbound_call_timeout});

    $logger->debug("Sending GET request $scanner_url to the Scanner");
    my $request = HTTP::Request->new( GET => $scanner_url);

    my $response = $ua->request($request);

    my $retval;
    $retval->{is_success} = $response->is_success;

    if ($response->is_success) {
        my $xml_data = $response->content;
        $logger->debug("Success. Response content: $xml_data");
        $retval->{xml_data} = $xml_data;
    }

    return $retval;
}

#------------------------------------------------------------------

sub get_and_store_api_key {
    my ($self) = @_;

    my $config = $self->{config}->{api};
    my $client = $self->{client};

    my $scanner_config = $self->{client}->get_scanner_config($self->{client}->{session}, undef, $self->{policy},);
    if ($scanner_config->{flg_is_api_key_final} || !$scanner_config->{api_key}) {
        return;
    }

    my $token = F5::WhiteHatKey::decrypt_with_whitehat_key(decode_base64(uri_unescape($scanner_config->{api_key})));

    my $api_and_site_query_url = $self->replace_url_parts($config->{get_api_key_url}, {
                                                                                       SCANNER_URL => $config->{url},
                                                                                       PORT => $config->{port},
                                                                                       TOKEN => $token,});
    my $response = $self->get_xml_data(url => $api_and_site_query_url);
    my $json;
    if ($response->{is_success}) {
        $json = from_json($response->{xml_data});

        $client->update_scanner_config($self->{client}->{session}, 
                                       undef, 
                                       $self->{policy},
                                       F5::ASMConfig::ScannerConfig->new(
                                           flg_is_api_key_final => 1,
                                           api_key => $json->{key},
                                           flg_is_api_key_encrypted => 0,
                                           scan_requested_time => time(),
                                           scanner_site_name => $json->{site_id},));
    }
}

#------------------------------------------------------------------

sub get_free_scan_results {
    my ($self) = @_;
  
    my $config = $self->{config}->{api};
    my $client = $self->{client};

    my $scanner_config = $self->{client}->get_scanner_config($self->{client}->{session}, undef, $self->{policy},);
    if ((!$scanner_config->{api_key}) || (!$scanner_config->{flg_is_api_key_final}) || ($scanner_config->{scan_status} == F5::ASMConfig::ScanStatus::SCAN_STATUS_SCAN_COMPLETED())) {
        return;
    }

    my $check_free_scan_complete_url = $self->replace_url_parts($config->{check_free_scan_complete_url},
                                                                {
                                                                 SCANNER_URL => $config->{url},
                                                                 PORT => $config->{port},
                                                                 API_KEY => $scanner_config->{api_key},
                                                                 SITE_NAME => $scanner_config->{scanner_site_name},
                                                                });

    my $response = $self->get_xml_data(url => $check_free_scan_complete_url);
    return unless $response->{is_success};
    

    my $assessment_complete;
    my $twig = F5::XML::Twig->new(twig_handlers => { '/site/assessment_complete' => sub {
        my ($twig, $elt) = @_;
        $assessment_complete = $elt->text;
    }});

    $twig->parse($response->{xml_data});
    return unless $assessment_complete;

    my $retval = $self->parse_xml(scanner_actions_fields => F5::ASMConfig::ScannerActionsFields->new(get_host_names_only => 1));
    $self->parse_xml(scanner_actions_fields => F5::ASMConfig::ScannerActionsFields->new(host_names => [keys %{$retval->{vuln_hosts}}], xml_file_path => $retval->{vuln_file_path}));
}

#------------------------------------------------------------------

sub get_register_url {
    my($self, %params) = @_;

    my $config = $self->{config}->{api};
    my $redirect_url = URI::Escape::uri_escape($params{redirect_url});

    my $callback_sig_encoded = URI::Escape::uri_escape(encode_base64(F5::WhiteHatKey::sign_with_whitehat_key($params{redirect_url})));
    my $free_scan_url = F5::License->new->is_production() ? $config->{production_free_scan_url}
                      :                                     $config->{test_free_scan_url}
                      ;

    my $retval = $self->replace_url_parts($free_scan_url, {
                                                           SCANNER_URL => $config->{url},
                                                           PORT => $config->{port},
                                                           CALLBACK_URL => $redirect_url,
                                                           CALLBACK_URL_SIG => $callback_sig_encoded,});

    return $retval;
}

#------------------------------------------------------------------

1;
