package F5::ServerTechnologies;

use strict;
use warnings;

use File::Basename;
use F5::Logger;
use File::Find;
use F5::Deploy;
use F5::Utils::File;
use F5::XML::Twig;
use F5::ASMConfig::EasyClient;
use F5::DbUtils;
use F5::ASMConfig::Entity::ServerTechnology;
use File::Copy;
use File::Path;


my $SERVER_TECHNOLOGIES_LOGOS_DIR = F5::Deploy::map_path('/var/ts/dms/server_technologies_logos');

sub load_from_xml {

    my (%params) = @_;

    my $dir = $params{dir};
    my $dbh = $params{dbh};


    if (! -d $SERVER_TECHNOLOGIES_LOGOS_DIR ) {
        mkpath($SERVER_TECHNOLOGIES_LOGOS_DIR);
    }

    my $logger = $params{logger} || F5::Logger->new('/ts/log/server_technologies.log');
    my $handled_files = 0;
    my $call_flags =  F5::ASMConfig::CallFlags->new();
    my $entity_class = 'F5::ASMConfig::Entity::ServerTechnology';
    my @server_technologies = $entity_class->retrieve_all(dbh => $dbh,
                                                          call_flags => $call_flags,
                                                         );
    my %existing_server_technologies = map { ( $_->{server_technology_name} => $_ ) } @server_technologies;
    my %seen_server_technologies;
    my $modified_entity = {};
    $modified_entity->{updated_count} = 0;
    $modified_entity->{added_count} = 0;
    $modified_entity->{deleted_count} = 0;
    $modified_entity->{modify_entities} = [];

    my $deferred_server_technologies = [];

    if ( ! -d $dir ) {
        my $error_msg = "Cannot load Server Technologies from unknown directory ($dir)";
        $logger->error($error_msg) if $logger;
        die "$error_msg\n";
    }

    $logger->info("Loading Server technologies") if $logger;

    $call_flags->{add_on_updating_nonexistent_record} = 1;

    find({ wanted => sub {
           if (-f $File::Find::name ) {
              copy($File::Find::name,$SERVER_TECHNOLOGIES_LOGOS_DIR);
           }
         },
         no_chdir => 1 },
       $dir."/logos/");

    find({ wanted => sub {
            my $filename = $File::Find::name;

           return unless (-f $filename);

           $logger->debug("Parsing $filename") if $logger;

           $handled_files++;

           my $twig = F5::XML::Twig->new(twig_handlers => {
                                                          'server_technology' => sub {
                                                              my($twig, $elm) = @_;
                                                              my $server_technology = F5::ASMConfig::Entity::ServerTechnology->new(dbh => $dbh, xml => $elm);
                                                              $seen_server_technologies{$server_technology->{server_technology_name}} = 1;
                                                              if ($server_technology->{implied_server_technologies}) {
                                                                  $server_technology->{implied_server_technologies} = $server_technology->{implied_server_technologies}->[0]->{server_technology_name};

                                                                  if ( ! ref($server_technology->{implied_server_technologies})) {
                                                                      $server_technology->{implied_server_technologies} = $server_technology->{implied_server_technologies} ?
                                                                                                                          [ $server_technology->{implied_server_technologies} ] :
                                                                                                                          [];
                                                                  }
                                                              }

                                                              if ($server_technology->{policy_actions}) {
                                                                  $server_technology->{policy_actions} = $server_technology->{policy_actions}->[0]->{policy_action};

                                                                  if ( ! ref($server_technology->{policy_actions})) {
                                                                      $server_technology->{policy_actions} = [ $server_technology->{policy_actions} ];
                                                                  }
                                                              }

                                                              if ($server_technology->{fingerprints}) {
                                                                  $server_technology->{flg_is_learnable} = 1;
                                                              }
 
                                                              handle_add_server_technology ( server_technology => $server_technology,
                                                                                             existing_server_technologies => \%existing_server_technologies,
                                                                                             deferred_server_technologies => $deferred_server_technologies,
                                                                                             dbh => $dbh,
                                                                                             modified_entity => $modified_entity,
                                                                                           );
                                                          }
                                                      }
                                    );
           my $rc = $twig->safe_parsefile($filename);
           if (!$twig) {
               my $error_msg = "Cannot load Server Technology from empty file ($filename).";
               $logger->error($error_msg) if $logger;
               die "$error_msg\n";
           }

        },
        preprocess => sub { return sort @_ },
    },$dir);

    while ( @$deferred_server_technologies ){
        my $current_deferred_server_tech = shift @$deferred_server_technologies;
        @{$current_deferred_server_tech->{implied_server_technologies}}  = grep { $seen_server_technologies{$_} } @{$current_deferred_server_tech->{implied_server_technologies}};
        handle_add_server_technology ( server_technology => $current_deferred_server_tech,
                                       existing_server_technologies => \%existing_server_technologies,
                                       deferred_server_technologies => $deferred_server_technologies,
                                       dbh => $dbh,
                                       modified_entity => $modified_entity,
                                     );
   }

   for my $server_technology ( keys %existing_server_technologies ){
       if(!$existing_server_technologies{$server_technology}->{was_seen}){
           $existing_server_technologies{$server_technology}->delete();
           $modified_entity->{deleted_count}++;
           my $modified_server_technology = F5::ASMConfig::ServerTechnologyModified->new({
                                               server_technology_name => $server_technology,
                                               change_level => F5::ASMConfig::LiveUpdateChangeLevel::LIVE_UPDATE_CHANGE_LEVEL_DELETED(),
                                            });
           push @{$modified_entity->{modify_entities}}, $modified_server_technology;
       }
   }

   return {
            completed_successfully => 1,
            server_technologies_count => $handled_files,
            server_technologies_updated_count => $modified_entity->{updated_count},
            server_technologies_added_count => $modified_entity->{added_count},
            server_technologies_deleted_count => $modified_entity->{deleted_count},
            server_technologies_modified => $modified_entity->{modify_entities},
          };

}

#----------------------------------------------------------------------------------------

sub add_or_update_server_technology{
    my (%params) = @_;
    my $server_technology = $params{server_technology};
    my $existing_server_technologies = $params{existing_server_technologies};
    my $dbh = $params{dbh};
    my $modified_entity = $params{modified_entity};

    if ($existing_server_technologies->{$server_technology->{server_technology_name}} ) {
        $existing_server_technologies->{$server_technology->{server_technology_name}}->{was_seen} = 1;
        my $existing_server_technology = F5::ASMConfig::Entity::ServerTechnology->new(dbh => $dbh, get_criteria => $existing_server_technologies->{$server_technology->{server_technology_name}});
        my $rv = $existing_server_technology->store_update(store_obj => $server_technology);
        my $db_change_made = $rv->{db_change_made};
        if ($db_change_made ) {
            $modified_entity->{updated_count}++;
            my $modified_server_technology = F5::ASMConfig::ServerTechnologyModified->new({
                                                server_technology_id => $existing_server_technology->{server_technology_id},
                                                server_technology_name => $existing_server_technology->{server_technology_name},
                                                change_level => F5::ASMConfig::LiveUpdateChangeLevel::LIVE_UPDATE_CHANGE_LEVEL_UPDATED(),
                                             });
            push @{$modified_entity->{modify_entities}}, $modified_server_technology;
        }
    }
    else {
        $server_technology->store_add(store_obj => $server_technology);
        $existing_server_technologies->{$server_technology->{server_technology_name}} = $server_technology;
        $existing_server_technologies->{$server_technology->{server_technology_name}}->{was_seen} = 1;
        $modified_entity->{added_count}++;
        my $modified_server_technology = F5::ASMConfig::ServerTechnologyModified->new({
                                            server_technology_id => $server_technology->{server_technology_id},
                                            server_technology_name => $server_technology->{server_technology_name},
                                            change_level => F5::ASMConfig::LiveUpdateChangeLevel::LIVE_UPDATE_CHANGE_LEVEL_ADDED(),
                                         });
        push @{$modified_entity->{modify_entities}}, $modified_server_technology;
    }
}

#----------------------------------------------------------------------------------------

sub handle_add_server_technology {
    my (%params) = @_;
    my $server_technology = $params{server_technology};
    my $existing_server_technologies = $params{existing_server_technologies};
    my $deferred_server_technologies = $params{deferred_server_technologies};
    my $modified_entity = $params{modified_entity};

    my $is_implied_server_tech_missing = 0;

    if ($server_technology->{implied_server_technologies} ){
        for my $implied_server_technology (@{$server_technology->{implied_server_technologies}}) {
            if (! $existing_server_technologies->{$implied_server_technology} ) {
                $is_implied_server_tech_missing = 1;
                last;
            }
        }
    }

    if ( $is_implied_server_tech_missing) {
        push (@$deferred_server_technologies, $server_technology);
    }
    else {
        add_or_update_server_technology( dbh => $params{dbh},
                                         server_technology => $server_technology,
                                         existing_server_technologies => $existing_server_technologies,
                                         modified_entity => $modified_entity, );
    }

}

#----------------------------------------------------------------------------------------

1;
