package F5::SessionDB;

use strict;
use warnings;

use Fcntl;
use IO::Socket::INET6;
use Socket;
use POSIX qw(:errno_h);
use Time::HiRes;

use F5::ASM::COAPI;

our $DEFAULT_ADDRESS = '127.1.1.1'; # tmm0
our $DEFAULT_PORT = '11211';
our $END_KEYWORD = 2;
our $DEFAULT_UNHANDLED_KEYWORD_LIMIT = 20;
our $DEFAULT_NO_RESPONSE_LIMIT = 200; # 200 * 0.05 sec = 10 seconds
our $READ_BUFFER_SIZE = 100_000;
our $NO_RESPONSE_RETRY_INTERVAL_NANOS = 100_000;

our %bin_name_cache;

my $singleton;
my $singleton_pid;

#----------------------------------------------------------------------

sub get_instance {
    my($class, %params) = @_;

    if ($singleton) {
        if ($singleton_pid != $$) {
            $singleton = undef;
        }
        else {
            # Call ping (even though no such command exists) to check if socket is up -
            # session db will respond "CLIENT_ERROR bogus command"
            my $ping_responded = 0;
            eval {
                $singleton->send_command(command => "ping\r\n",
                                         no_response_limit => 2, # ping must return within 0.1 sec
                                         keyword_handlers => {
                                                              CLIENT_ERROR => sub {
                                                                  $ping_responded = 1;
                                                                  return $END_KEYWORD;
                                                              }
                                                             }
                                        );
            };

            if (! $ping_responded) {
                $singleton = undef;
            }
        }
    }

    if (! $singleton) {
        $singleton = $class->new(%params);
        $singleton_pid = $$;
    }

    $singleton->init(%params);
    return $singleton;
}

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->{address} = $params{address} || $DEFAULT_ADDRESS;
    $self->{port} = $params{port} || $DEFAULT_PORT;
    $self->{batch_updates} = $params{batch_updates} || 0;
    $self->{batch_size} = $params{batch_size};
    $self->{batches} = {};
    $self->{batch_opts} = {};
    $self->{default_bin_name} = $params{default_bin_name};
}

#----------------------------------------------------------------------

sub get {
    my($self, %params) = @_;
    my $key = $params{key};
    return unless $key;

    my @retval;
    my $get_command = 'get';
    if ($params{extended}) {
        $get_command = 'get/t';
    }
    if (! ref($key)) {
        $key = [ $key ];
    }
    return unless @$key;

    my $bin_prefix = $self->get_bin_prefix(sessiondb_action => 'get',
                                           %params
                                          );
    my @commands = map { "$get_command $bin_prefix$_\r\n" } @$key;

    $self->send_command(command => \@commands,
                        keyword_handlers => {
                                             CLIENT_ERROR => sub {
                                                 my(%handler_params) = @_;
                                                 die "SessionDB failed to get $key - $handler_params{data}\n";
                                             },
                                             'END' => sub {
                                                 return $END_KEYWORD;
                                             },
                                             'VALUE' => sub {
                                                 my(%handler_params) = @_;
                                                 my(@data_parts) = split(/ +/, $handler_params{data});
                                                 my $value_rec = {
                                                                  value => $handler_params{value},
                                                                  key => $data_parts[0],
                                                                  flag => $data_parts[1],
                                                                  value_length => $data_parts[2],
                                                                  inactivity_time => $data_parts[3],
                                                                  age => $data_parts[4],
                                                                  timeout => $data_parts[5],
                                                                  lifetime => $data_parts[6],
                                                                 };
                                                 push(@retval, $value_rec);
                                                 return 1;
                                             },
                                            }
                       );
    return @retval;
}

#----------------------------------------------------------------------

sub binlookup {
    my($self, %params) = @_;
    my $bin_name = $params{bin_name};
    return unless $bin_name;

    my $retval = '';
    $self->send_command(command => "binlookup $bin_name\r\n",
                        keyword_handlers => {
                                             CLIENT_ERROR => sub {
                                                 my(%handler_params) = @_;
                                                 die "SessionDB failed to get binlookup for $bin_name\n";
                                             },
                                             'END' => sub {
                                                 return $END_KEYWORD;
                                             },
                                            },
                        nonkeyword_handler => sub {
                            my(%handler_params) = @_;
                            $retval = $handler_params{keyword}; # first word of response line
                        },
                       );

    if ($retval) {
        return $retval;
    }
    return;
}

#----------------------------------------------------------------------

sub set {
    my($self, %params) = @_;
    return $self->set_or_add(%params,
                             command => 'set');
}

#----------------------------------------------------------------------

sub add {
    my($self, %params) = @_;
    return $self->set_or_add(%params,
                             command => 'add');
}

#----------------------------------------------------------------------

sub set_or_add {
    my($self, %params) = @_;
    my $command = $params{command};
    my $key = $params{key};
    my $flag = $params{flag} || 0;
    my $expiration = $params{expiration} || 0;
    my $value = $params{value} || '';
    my $length = length($value);
    my $skip_if_exists = $params{skip_if_exists};

    return unless $key;

    my $bin_prefix = $self->get_bin_prefix(sessiondb_action => 'set_or_add',
                                           %params
                                          );

    if ($params{skip_if_exists} && ! $self->{batch_updates}) {
        if ($self->get(key => $key)) {
            return;
        }
    }

    push(@{$self->{batches}->{add}}, "$command $bin_prefix$key $flag $expiration $length\r\n$value\r\n");
    if ($params{skip_if_exists}) {
        $self->{batch_opts}->{skip_if_exists} = 1;
    }

    if (! $self->{batch_updates} ||
        scalar @{$self->{batches}->{add}} >= $self->{batch_size}) {

        $self->add_batch(commands => $self->{batches}->{add});
    }
}

#----------------------------------------------------------------------

sub add_batch {
    my($self, %params) = @_;
    my $commands = $params{commands};
    my $skip_if_exists = $self->{batch_opts}->{skip_if_exists} || 0;
    if (! ref($commands)) {
        $commands = [ $commands ];
    }

    my $key = '';
    if (scalar @$commands == 1) {
        my(@fields) = split(/ +/, $commands->[0]);
        $key = $fields[1];
    }

    eval {
        $self->send_command(command => $commands,
                            keyword_handlers => {
                                                 CLIENT_ERROR => sub {
                                                     my(%handler_params) = @_;
                                                     die "SessionDB failed to set $key - $handler_params{data}\n";
                                                 },
                                                 INVALID_KEY => sub {
                                                     my(%handler_params) = @_;
                                                     die "SessionDB failed to set $key (invalid key) - $handler_params{data}\n";
                                                 },
                                                 'STORED' => sub {
                                                     return $END_KEYWORD;
                                                 },
                                                 'NOT_STORED' => sub {
                                                     my(%handler_params) = @_;
                                                     if ($skip_if_exists) {
                                                         return $END_KEYWORD;
                                                     }
                                                     else {
                                                         die "SessionDB failed to set $key (not stored) - $handler_params{data}\n";
                                                     }
                                                 },
                                                }
                           );
    };
    if ($@) {
        @$commands = ();
        die $@;
    }
    @$commands = ();
    $self->{batch_opts} = {};
}

#----------------------------------------------------------------------

sub delete {
    my($self, %params) = @_;
    my $key = $params{key};

    return unless $key;

    my $bin_prefix = $self->get_bin_prefix(sessiondb_action => 'delete',
                                           %params
                                          );

    push(@{$self->{batches}->{delete}}, "delete $bin_prefix$key\r\n");

    if (! $self->{batch_updates} ||
        scalar @{$self->{batches}->{delete}} >= $self->{batch_size}) {

        $self->delete_batch(commands => $self->{batches}->{delete},
                           );
    }
}

#----------------------------------------------------------------------

sub delete_batch {
    my($self, %params) = @_;
    my $commands = $params{commands};
    if (! ref($commands)) {
        $commands = [ $commands ];
    }

    my $key = '';
    if (scalar @$commands == 1) {
        my(@fields) = split(/ +/, $commands->[0]);
        $key = $fields[1];
    }

    $self->send_command(command => $commands,
                        keyword_handlers => {
                                             CLIENT_ERROR => sub {
                                                 my(%handler_params) = @_;
                                                 die "SessionDB failed to delete $key - $handler_params{data}\n";
                                             },
                                             'DELETED' => sub {
                                                 return $END_KEYWORD;
                                             },
                                             'NOT_DELETED' => sub {
                                                 return $END_KEYWORD;
                                             },
                                            }
                       );
    if ($@) {
        @$commands = ();
        die $@;
    }
    @$commands = ();
}

#----------------------------------------------------------------------

sub send_command {
    my($self, %params) = @_;
    my $command = $params{command};
    my $keyword_handlers = $params{keyword_handlers};
    my $nonkeyword_handler = $params{nonkeyword_handler};

    my $command_count = 1;
    if (ref($command)) {
        $command_count = scalar @$command;
        $command = join('', @$command);
    }

    unless ($self->{socket}) {
        $self->open_socket;
    }

    send($self->{socket}, $command, 0);
    my $buf = '';
    my $reached_end_keyword = 0;
    my $unhandled_keyword_attempt = 0;
    my $end_keyword_count = 0;
    my $no_response_attempt = 0;
    my $no_response_limit = defined $params{no_response_limit} ? $params{no_response_limit} : $DEFAULT_NO_RESPONSE_LIMIT;

    while (! $reached_end_keyword && ! $params{ignore_response}) {
        my $read_bytes = sysread($self->{socket}, $buf, $READ_BUFFER_SIZE, length($buf));
        if ($buf) {
            my $orig_buf = $buf;
            $no_response_attempt = 0;

            my $keyword = '';
            while ($buf =~ s/^(\S+) *(.*?)\r\n//) {
                $keyword = $1;
                my $data = $2;

                # Special handling for VALUE
                my $value;
                if ($keyword eq 'VALUE') {
                    $data =~ /^\S+\s+\S+\s+(\d+)/;
                    my $value_len = $1;
                    if (length($buf) < $value_len + 2) {
                        $buf = $orig_buf;
                        last;
                    }
                    $buf =~ s/^(.{$value_len})\r\n//;
                    $value = $1;
                }

                $orig_buf = $buf;
                if ($keyword_handlers->{$keyword}) {
                    $unhandled_keyword_attempt = 0;
                    my $rv = $keyword_handlers->{$keyword}->(keyword => $keyword,
                                                             'data' => $data,
                                                             'value' => $value,
                                                            );
                    if ($rv == $END_KEYWORD) {
                        $end_keyword_count++;
                        if ($end_keyword_count >= $command_count) {
                            $reached_end_keyword = 1;
                            last;
                        }
                    }
                }
                elsif ($nonkeyword_handler) {
                    $unhandled_keyword_attempt = 0;
                    $nonkeyword_handler->(keyword => $keyword,
                                          'data' => $data,
                                         );
                }
                else {
                    last;
                }
            }

            if ($buf) {
                $unhandled_keyword_attempt++;
                if ($unhandled_keyword_attempt > $DEFAULT_UNHANDLED_KEYWORD_LIMIT) {
                    die "Unhandled keyword ($keyword)\n";
                }
                elsif ($unhandled_keyword_attempt > 1) {
                    Time::HiRes::usleep(50_000);
                }
            }
        }

        if (! $buf && $no_response_attempt++) {
            if ($no_response_attempt > $no_response_limit) {
                die "No response from persistence layer\n";
            }
            elsif ($no_response_attempt > 3) {
                Time::HiRes::usleep($NO_RESPONSE_RETRY_INTERVAL_NANOS);
            }
        }
    }
}

#----------------------------------------------------------------------

sub open_socket {
    my($self) = @_;
    my $socket = IO::Socket::INET6->new(PeerAddr => $self->{address},
                                        PeerPort => $self->{port},
                                        Proto    => 'tcp',
                                        Timeout  => 5,
                                       );

    if (! $socket) {
        my $error_msg = "Could not connect to persistence layer";
		my $vlans = F5::ASM::COAPI::Query->new('vlan');
		if (! $vlans || ! @$vlans) {
			$error_msg .= ' (no VLANs exist)';
		}
		die "$error_msg\n";
    }
    nonblock($socket);
    $socket->autoflush(1);
    $self->{socket} = $socket;
}

#----------------------------------------------------------------------
# nonblock($socket) puts socket into nonblocking mode
sub nonblock {
  my $socket = shift;
  my $flags;

  $flags = fcntl($socket, F_GETFL, 0);
  if ($socket && $flags) {
    fcntl($socket, F_SETFL, $flags | O_NONBLOCK);
  }
}

#----------------------------------------------------------------------

sub flush {
    my($self) = @_;

    for my $batch_action (keys %{$self->{batches}}) {
        if (@{$self->{batches}->{$batch_action}}) {
            my $batch_action_func = $batch_action . '_batch';
            $self->$batch_action_func(commands => $self->{batches}->{$batch_action});
        }
    }
}

#----------------------------------------------------------------------

sub close {
    my($self) = @_;
    $self->flush;
    $self->send_command(command => "quit\r\n",
                        ignore_response => 1,
                       );
}

#----------------------------------------------------------------------

sub DESTROY {
    my($self) = @_;
    eval {
        $self->close;
    };
}

#----------------------------------------------------------------------

sub get_bin_prefix {
    my($self, %params) = @_;
    my $action = $params{sessiondb_action} || '';

    # HA SessionDB weirdness:
    #     bin_prefix formart: prefix[.<ha_unit>]
    # We will default to always use ha_unit .1 (HA_ACCESS_UNIT) as BD is doing the same
    my $ha_unit = '.1' ;

    my $bin_name = $self->{default_bin_name};
    if (exists $params{bin_name}) {
        $bin_name = $params{bin_name};
    }

    if ($bin_name) {
        if (! $bin_name_cache{$bin_name}) {
            $bin_name_cache{$bin_name} = $self->binlookup(bin_name => $bin_name);
        }
	return "/$bin_name_cache{$bin_name}$ha_unit/";
    }

    return '';
}

#----------------------------------------------------------------------

1;
