package F5::SetActive::Impl;

use strict;
use warnings;
use Carp;

use File::Basename;
use File::Path;
use File::stat;
use Time::HiRes;
use English;

use F5::Provisioning;
use F5::PrepareFiles;
use F5::BigipUtils;
use F5::Cfg;
use F5::LockFactory;
use F5::DbUtils;
use F5::UserActivityLog;
use F5::PrepareConf::Base;
use F5::PrepareConf::Policy;
use F5::PrepareConf::Hsl;
use F5::PrepareConf::BotDefense;
use F5::PrepareConf::LoggingProfile;
use F5::RamCache;
use F5::PolicyValidator;
use F5::Cluster::Sync;
use F5::Cluster::Utils;
use F5::EmServices;
use F5::Logger;
use F5::Hbcl;
use File::Copy;
use F5::Policy;
use F5::ASMReady;
use F5::ASMConfig::EasyClient;
use F5::ASMConfig::Entity::PolicyAttributes;
use F5::ASMConfig::Entity::PolicyHistory;
use F5::ASMConfig::Cache;
use F5::Utils::Account;
use F5::Utils::File;
use F5::IO::Storable;
use F5::IO::Selector;
use F5::DateTimeUtils;
use F5::Log::Job;
use YAML::XS;

use constant CMD_WEB_UPDATE_CORE => 2;
use constant BUSY_RESULT_CODE => -24;

my $UPDATE_CORE_PULSE_MSG_INTERVAL = 30;
my $UPDATE_CORE_MAX_INTERVAL = 5 * 60; # 5 minutes
my $TSCONFD_USER_NAME = "tsconfd";
use constant  PRPLC_CFG_FILE  =>  'prepare_policy.cfg';

sub set_active {
    my(%params) = @_;
    my $dbh                 = $params{dbh};
    my $logger              = $params{logger};
    my $policy              = $params{policy};
    my $policy_name         = $params{policy_name};
    my $force_policy_name_flg = $params{force_policy_name_flg};
    my $account_id          = $params{account_id};
    my $hsl_profile_id      = $params{hsl_profile_id};
    my $invalidate_hsl_data_profiles   = $params{invalidate_hsl_data_profiles};
    my $hsl_profile_options  = $params{hsl_profile_options};
    my $username            = $params{username} || 'SYSTEM';
    my $dcc_dbname          = $params{dcc_dbname} || 'DCC';
    my $do_update_core      = defined($params{do_update_core})  ? $params{do_update_core}   : 1;
    my $skip_mcp            = defined($params{skip_mcp})  ? $params{skip_mcp}  : 0;
    my $update_core_retries = defined($params{update_core_retries})   ? $params{update_core_retries}    : 5;
    my $do_transaction       = defined($params{do_transaction})  ? $params{do_transaction}   : 1;
    my $delayed_update_core  = defined($params{delayed_update_core})  ? $params{delayed_update_core} : 0;
    my $update_ramcache      = defined($params{update_ramcache}) ? $params{update_ramcache} : 1 ;
    my $do_validation       = defined($params{do_validation})   ? $params{do_validation}    : 1;
    my $do_hbcl             = defined($params{do_hbcl})   ? $params{do_hbcl}    : 1;
    my $do_save_policy_history = $params{do_save_policy_history};
    my $save_policy_history_frequency_limit = $params{save_policy_history_frequency_limit};
    my $force_full_sync    = defined($params{force_full_sync})   ? $params{force_full_sync}    : 0;
    my $wait_for_enforcer   = $params{wait_for_enforcer} || $params{update_logger_cfg};

    my $send_sync       = defined($params{send_sync})   ? $params{send_sync}    : 1;
    my $ignore_if_policy_unmodified = defined($params{ignore_if_policy_unmodified}) ? $params{ignore_if_policy_unmodified} : 0;
    my $ignore_if_policy_not_active = defined($params{ignore_if_policy_not_active}) ? $params{ignore_if_policy_not_active} : 0;
    my $prepare_account_tables = defined($params{prepare_account_tables}) ? $params{prepare_account_tables} : 1;
    my $validation_report;
    my $asm_config_handler =  $params{asm_config_handler};
    my $job = $params{job};
    if ($job) {
        $job = $job->copy(component => 'set_active');
    }
    else {
        $job = F5::Log::Job->new(logger => $logger,
                                 component => 'set_active',
                                 operationId => 'set_active',
                                );
    }
    $job->set_started();

    if ($params{update_core_only}) {
      $do_update_core = 1;
    }

    if ($delayed_update_core) {
        $do_transaction = 0;
        $send_sync = 0;
        $do_update_core = 0;
    }

    if (!was_valid_command_issued(%params)) {
      return { completed_successfully => 0,
               error_message => 'No commands issued' };
    }

    if ($params{update_logger_cfg}) {
      $dbh->do(<<EndQuery);
DELETE FROM DCC.CONFIG_TYPE_UPDATE_LOGGER_CFG
EndQuery

      $dbh->do(<<EndQuery, undef, time());
INSERT INTO DCC.CONFIG_TYPE_UPDATE_LOGGER_CFG
   (last_update_time)  VALUES (?)
EndQuery

      $do_update_core = 1;
    }

    if ($params{dump_bd_config_tables}) {
		F5::DbUtils::insert_row(dbh => $dbh,
		                        table_name => 'DCC.DUMP_BD_CONFIG_TABLES',
		                        rec => { 1 },
		                        do_replace => 1,
		                       );
      $do_update_core = 1;
    }

    if ($dcc_dbname ne 'DCC') {
      return { completed_successfully => 0,
               error_message => 'Set active cannot be called with a database other than DCC' };
    }



    my $lock_file =  $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                             logger => $logger);

    $locker->lock();

    my $policy_id = ($policy) ? $policy->policy_id : $params{policy_id};
    if ($policy_id) {
        $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh ,
                                                          policy => {policy_id  => $policy_id},
                                                          force_reload => 1);
    }
    elsif ($account_id) {
        $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh ,
                                                          force_reload => 1,
                                                          policy => {account_id => $account_id});
    }

    if ($policy && $policy->{_failed_to_load}) {
          return { completed_successfully => 0,
                   error_message => 'Set active failed: Policy no longer exists or no longer Active',
                   error_level   => 'info',};
    }

    if ((!defined $account_id ) && $policy) {
        $account_id = $policy->{account_id};
    }
    elsif ($account_id && !$policy) {
        $policy = F5::ASMConfig::Entity::Policy->retrieve_all(dbh => $dbh, get_criteria => {account_id => $account_id});
        if (!$policy) {
            return { completed_successfully => 0,
                    error_message => 'Set active failed: Policy no longer active',
                    error_level   => 'info',};
        }
    }

    my $cfg_do_table_checksums = F5::Cfg::cfg_get_config_item("/etc/ts/dcc/prepare_policy.cfg", 'GENERAL', 'DoTableChecksums');
    my $do_table_checksums = defined($cfg_do_table_checksums) ? $cfg_do_table_checksums : 1;

    my $logger_for_prepareconf = $logger || F5::Logger->get_logger(logger_name => 'prepare_conf');

    my $active_policy_was_changed = 0;
    my $do_add_to_policy_transactions = 0;

    if ($policy) {
        my $exit_rv;
        if (!$params{update_core_only} && $ignore_if_policy_unmodified && ! $policy->is_modified && !$params{activate_policy} && !$params{deactivate_policy}) {
            $exit_rv = { completed_successfully => 0,
                         error_message => 'Policy is unmodified - aborting',
                       };
        }
        elsif (!$policy->{is_active_policy} && $ignore_if_policy_not_active) {
            $exit_rv = { completed_successfully => 0,
                         error_message => 'Policy is not active - aborting',
                       };
        }
        elsif (($policy->{account_id} && $params{activate_policy}) ||
            (!$policy->{account_id} && $params{deactivate_policy}) ) {
            $exit_rv = { completed_successfully => 0,
                         error_message => 'Policy is already Active/Inactive - aborting',
                         error_level => 'info',
                       };
        }
        elsif (!($account_id || $params{activate_policy})) {
            $params{activate_policy} = 1;
        }

        if ($exit_rv) {
            $logger->info($exit_rv->{error_message}) if $logger;
            $locker->unlock();
            return $exit_rv;
        }

        if (  !$policy->{has_been_applied} ) {
            $active_policy_was_changed = 1;
            $do_add_to_policy_transactions = 1;
        }
    }

    unless (defined($do_save_policy_history)) {
      my $cfg_do_save_policy_history = F5::Cfg::cfg_get_config_item("/etc/ts/dcc/".PRPLC_CFG_FILE, 'GENERAL', 'SavePolicyHistory');
      $do_save_policy_history = defined($cfg_do_save_policy_history) ? $cfg_do_save_policy_history : 1;
    }

    my $original_account_id;
    my $original_policy_name_crc;
    if (!$params{update_core_only} && (defined $policy  || $hsl_profile_id)) {
        croak 'SetActive did not receive a database handle' unless (defined $dbh);

        my(%saved_dbh_attribs) = set_dbh_raise_error(dbh => $dbh);
        my $modification_log_id;

        $dbh->begin_work if $do_transaction;

        if (defined $policy) {
            if ($params{deactivate_policy}) {
                $logger->info("Deactivating policy " . $policy->{policy_id}) if $logger;
                F5::Utils::Account::delete_account(dbh => $dbh, policy => $policy, logger => $logger);
                $active_policy_was_changed = 1;
                $account_id = 0;
                $original_account_id = $policy->{account_id};
                $original_policy_name_crc = $policy->{policy_name_crc};

                my $prepare_conf = F5::PrepareConf::Base->new(dbh => $dbh);
                $prepare_conf->prepare_enforcer_accounts($policy->{account_id});
                $policy->{is_active_policy} = 0;
            }
            else {
                $logger->info("Activating policy " . $policy->{policy_id}) if $logger;

                if ($do_validation) {
                  my $validator = F5::PolicyValidator->new(dbh => $dbh,
                                     logger => $logger);
                  $validation_report = $validator->validate(policy => $policy,
                                                            action => 'set_active',
                                                           );
                }

                eval {
                  if ($params{activate_policy}) {
                      $logger->info("Marking active for policy " . $policy->{policy_id}) if $logger;
                      $account_id = F5::Utils::Account::create_account(dbh => $dbh, policy => $policy, logger => $logger, dcc_dbname => $dcc_dbname);
                      $active_policy_was_changed = 1;
                      $policy->{is_active_policy} = 1;
                  }
                  if (!$delayed_update_core) {
                      $modification_log_id = $policy->get_max_modification_log_id;
                  }

                  if ($params{clear_traffic_data}) {
                      #Changing Language set.  Clear out Proxy Log and Learning.
                      $policy->clear_traffic_data();
                  }

                  # F5::PrepareConf::Policy accepts undef as an override value for arguments. It's not clear whether
                  # this has an existing designed use, but to be on the safe side, arguments such as generate_negsig_collections
                  # should only be included in instantiation when defined.
                  my %additional_params;
                  if (defined $params{generate_negsig_collections}) {
                      $additional_params{generate_negsig_collections} = $params{generate_negsig_collections};
                  }
                  my $prepareconf_policy = F5::PrepareConf::Policy->new(
                                                                        dbh     =>  $dbh,
                                                                        logger  =>  $logger_for_prepareconf,
                                                                        unescape    => F5::Cfg::cfg_get_config_item("/etc/ts/dcc/".PRPLC_CFG_FILE, 'GENERAL', 'Unescape'),
                                                                        policy_name => $policy_name,
                                                                        force_policy_name_flg => $force_policy_name_flg,
                                                                        ucs_upgrade_real_policy_id =>  $params{ucs_upgrade_real_policy_id},
                                                                        checksums_policy_id => $params{checksums_policy_id},
                                                                        %additional_params,
                                                                       );
                  $prepareconf_policy->prepare(
                                               policy                  =>  $policy,
                                               policy_id               =>  $policy->{policy_id},
                                               account_id              =>  $account_id,
                                               transaction_start_time  =>  $params{transaction_start_time},
                                               active_policy_was_changed => $active_policy_was_changed,
                                               do_table_checksums      => $do_table_checksums,
                                               asm_config_handler      => $asm_config_handler,
                                              );

                  my $prepareconf_logging = F5::PrepareConf::LoggingProfile->new(dbh => $dbh,
                                                                                 logger => $logger_for_prepareconf,
                                                                                );
                  $prepareconf_logging->prepare(policy => $policy,
                                                do_cleanup => 0,
                                               );

                  if ($params{force_failover}) {
                      die "Forced Failover\n";
                  }

                };

                if (my $err = $@) {
                  if (ref($err)) {
                      $err = $err->{error_message};
                  }
                  $logger->error("Setting policy active failed: $err") if $logger;
                  if (F5::Cluster::Utils::is_clustering_enabled() && $do_hbcl) {
                      F5::Hbcl::set_state_unavailable($logger);
                  }
                  $dbh->rollback if $do_transaction;
                  restore_dbh_attribs(dbh => $dbh,
                                      saved_dbh_attribs => \%saved_dbh_attribs);
                  $locker->unlock();

                  return { completed_successfully => 0,
                           error_message => 'Set active failed' };
                }
            }

            if (($params{deactivate_policy} || $params{clear_traffic_data}) && !F5::BigipUtils::is_docker() ) {
                $asm_config_handler->add_post_commit_hook(name => 'logical_delete_from_incidents',
                                                          func => sub {
                                                              $policy->logical_delete_from_incidents;
                                                          });
            }

        }

        if (defined $hsl_profile_id) {
            my $prepare_hsl = F5::PrepareConf::Hsl->new(
                dbh => $dbh,
                logger  =>  $logger_for_prepareconf,
            );

            $prepare_hsl->prepare(
                hsl_profile_id => $hsl_profile_id,
                hsl_profile_options => $hsl_profile_options,
                asm_config_handler => $asm_config_handler,
            );
        }

        if (F5::Cluster::Utils::is_clustering_enabled() && $do_hbcl) {
            F5::Hbcl::set_state_available($logger);
        }

        $dbh->commit if $do_transaction;

        restore_dbh_attribs(dbh => $dbh,
                            saved_dbh_attribs => \%saved_dbh_attribs);

        if ($active_policy_was_changed) {
            if( ! $params{ucs_upgrade_real_policy_id}  ){
                $policy->store_update(store_obj => {account_id => $account_id,
                                                    has_been_applied => 1,
                                                    is_active_policy => $account_id ? 1 : 0},
                                      skip_mcp  => 1,
                                      suppress_events => 1,);
            }

            $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh ,
                                                              policy => {policy_id  => $policy_id},
                                                              force_reload => 1);
            if ($account_id) {
               my $objects_iter = F5::ASMConfig::Entity::URL->retrieve_all(dbh => $dbh, policy => $policy, minimal_load => 1);
               while (my $object = $objects_iter->get_next) {
                   $object->store_update(store_obj => {},
                                force_crc_update => 1,
                                suppress_events => 1,
                                validate_value => 0,),
               }
            }
            else {
                if (my $old_version = F5::ASMConfig::Entity::PolicyHistory->new(dbh => $dbh, policy => $policy, get_criteria => {to_date => 0})) {
                    $old_version->store_update(store_obj => {to_date => time}, suppress_events => 1);
                }
            }
        }

        if ($modification_log_id) {
            $policy->clear_modification_log(max_modification_log_id => $modification_log_id);
        }
    }
    else {
        # The following are updated by GUI with a direct call to update core
        my $prepare_conf = F5::PrepareConf::Base->new(dbh => $dbh);
        $prepare_conf->prepare_icap_server();
        $prepare_conf->prepare_database_protection_server();
        $prepare_conf->prepare_client_side_captcha();
        $prepare_conf->persist_to_mcp() unless ($username eq $TSCONFD_USER_NAME || $username eq 'SYSTEM');
    }


    my $update_core_rv = 0;
    if ($do_update_core ) {

      unless ($dbh) {
        $dbh = F5::DbUtils::get_dbh();
      }

      if ($update_ramcache) {
          F5::RamCache::prepare_conf(dbh => $dbh);
      }

      $update_core_rv = send_update_core(dbh => $dbh,
                                         logger => $logger,
                                         update_core_retries => $update_core_retries,
                                         do_table_checksums => $do_table_checksums,
                                         policy => $policy,
                                         original_account_id => $original_account_id,
                                         original_policy_name_crc => $original_policy_name_crc,
                                         prepare_account_tables => $prepare_account_tables,
                                         username => $username,
                                         modified_dcc_tables => $params{modified_dcc_tables},
                                         wait_for_enforcer => $wait_for_enforcer,
                                         job => $job,
                                        );

      log_update_core_rv(dbh => $dbh,
                         logger => $logger,
                         policy => $policy,
                         username => $username,
                         update_core_rv => $update_core_rv,);
    }

    if ($invalidate_hsl_data_profiles) {
        foreach my $hsl_data_profile (@{$invalidate_hsl_data_profiles}) {
            $logger->info("Invalidate cache for profile $hsl_data_profile") if $logger;
            F5::RamCache::invalidate_cache_by_profile_name(profile_name => $hsl_data_profile, logger => $logger);
        }

    }

    if (defined $policy && !$delayed_update_core) {
       invalidate_ramcache_cache(dbh => $dbh,
                                 policy    => $policy,
                                 logger => $logger);
    }


    if (defined $policy && $do_save_policy_history && $policy->{account_id}) {
        my $policy_history;
        eval {
            my %policy_history_params = (validation_report => $validation_report,
                                         save_policy_history_frequency_limit => $save_policy_history_frequency_limit,
                                         user_name => $username,
                                         username  => $username,);
            if ($asm_config_handler) {
                $policy_history = $asm_config_handler->save_policy_history($asm_config_handler->{session},
                                                                           F5::ASMConfig::CallFlags->new({do_not_relay => 1}),
                                                                           $policy,
                                                                           F5::ASMConfig::PolicyHistory->new(\%policy_history_params));
            }
            else {
                $policy_history = F5::Policy->new(policy_id  => $policy->{policy_id},
                                                  dbh         => $dbh,)->history->save(%policy_history_params);
            }
        };

        if (!$EVAL_ERROR) {
            if ($policy_history) {
                F5::UserActivityLog::log($dbh, $logger, F5::UserActivityLog::LOG_MSG_CREATE_HISTORY_VERSION,
                                         $username, $policy->{policy_name}, $policy->{policy_name},
                                         'Operation completed successfully.');
            }
        }
        else {
            F5::UserActivityLog::log($dbh, $logger, F5::UserActivityLog::LOG_MSG_CREATE_HISTORY_VERSION,
                                     $username, $policy->{policy_name}, $policy->{policy_name},
                                     "Operation failed. Error: $EVAL_ERROR");
        }
    }

    if ($send_sync) {
      my %send_sync_opts;
      if (defined $policy) {
        %send_sync_opts = ( policy_id => $policy->{policy_id},
                            policy_name_crc => $policy->{policy_name_crc},
                            account_id => $policy->{account_id},
                            force_full_sync => $force_full_sync || $active_policy_was_changed);
      }
      my $cluster_sync = F5::Cluster::Sync->new(dbh => $dbh);
      if(!$cluster_sync){
         $logger->error("Can't sync - ASM is not ready") if $logger;
         return { completed_successfully => 0,
                  update_core_rv => $update_core_rv };
      }
      $cluster_sync->send_sync(%send_sync_opts);
    }

    my $do_mcp_store = 0;
    if ($active_policy_was_changed && ! $params{ucs_upgrade_real_policy_id} && $policy) {
        $do_mcp_store = 1;
        $policy->mcp_store(skip_mcp => $skip_mcp);
    }

    # this should be called, after 'mcp_store', in case 'send_update_core' was called
    if ($do_update_core && $policy && $policy->{policy_name} && $policy->{account_id}) {
        my $bot_defense = F5::PrepareConf::BotDefense->new(dbh => $dbh,
                                                            logger => $logger);
        $bot_defense->prepare_bot_defense_profiles(policy => $policy);
    }

    $locker->unlock();
    if ($job) {
        if ($update_core_rv eq '0') {
            $job->set_completed();
        }
        else {
            $job->set_failed();
        }
    }

    return { completed_successfully => $update_core_rv eq '0',
             do_mcp_store => $do_mcp_store,
             active_policy_was_changed => $active_policy_was_changed,
             validation_report => $validation_report,
             update_core_rv => $update_core_rv,
             delayed_update_core => $delayed_update_core,
             do_table_checksums => $do_table_checksums,
             deactivate_policy => $params{deactivate_policy},
           };
}

#----------------------------------------------------------------------

sub log_update_core_rv {
    my (%params) = @_;
    my $policy = $params{policy};
    my $update_core_rv = $params{update_core_rv};
    my $username = $params{username};
    my $dbh = $params{dbh};
    my $logger = $params{logger};

    if ($params{log_result} && defined $policy) {
        if ($update_core_rv eq '0') {
            F5::UserActivityLog::log($dbh, $logger, F5::UserActivityLog::LOG_MSG_APPLY_POLICY,
                            $username, $policy->{policy_name}, $policy->{policy_name},
                            'Operation completed successfully.');
        }
        else {
            F5::UserActivityLog::log($dbh, $logger, F5::UserActivityLog::LOG_MSG_APPLY_POLICY,
                            $username, $policy->{policy_name}, $policy->{policy_name},
                            "Apply Policy operation failed. Error: $update_core_rv");
        }
    }

}

#----------------------------------------------------------------------

sub send_update_core {
    my(%args) = @_;
    my $logger = $args{logger};
    my $update_core_retries = (exists $args{update_core_retries} ? $args{update_core_retries} : 5);
    my $do_table_checksums = (defined $args{do_table_checksums} ? $args{do_table_checksums} : 1);

    if (! F5::ASMReady::is_asm_ready()) {
        $logger->info("Not sending update core - ASM not ready, core will be updated when core is ready") if ($logger);
        return 0;
    }

    if (!F5::Provisioning::is_module_enabled('asm')) {
        $logger->info("Not sending update core - ASM is running in non-provisioned mode") if ($logger);
        return 0;
    }

    my $lock_file = $F5::DbUtils::LOCK_FILE;
    my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                             logger => $logger);

    $locker->lock();

    $logger->debug("ASM is ready for update core") if ($logger);

    F5::PrepareFiles::init(dbh => $args{dbh},
                           logger => $logger,
                          );
    my $core_file_update_rv = F5::PrepareFiles::PrepareCoreFilesUpdate(site_id => 1,
                                                                       dbh => $args{dbh},
                                                                       do_table_checksums => $do_table_checksums,
                                                                       job => $args{job},
                                                                       policy => $args{policy},
                                                                       original_account_id => $args{original_account_id},
                                                                       prepare_account_tables => $args{prepare_account_tables},
                                                                       modified_dcc_tables => $args{modified_dcc_tables},
                                                                      );
    $logger->debug("Publishing '$core_file_update_rv->{protobuf_manifest_file}' (generationIndex = '$core_file_update_rv->{generationIndex}', datetimeMicros = '$core_file_update_rv->{datetimeMicros}', jobId = '$core_file_update_rv->{jobId}')") if ($logger);

    my $expected_generation_index = $core_file_update_rv->{generationIndex};
    $logger->info("Writing update generation '$expected_generation_index'")  if ($logger);
    my $last_pulse_time = time;
    my $start_time = $last_pulse_time;

    my $load_state_file = $F5::PrepareFiles::BD_CONFIG_DIR . "/bd_cfg_manifest_load_state.json";
    my $load_state_file_mtime;

    F5::PrepareFiles::publish_manifest(logger => $logger,
                                       %$core_file_update_rv);

    my $response;
    my $error_msg = '';
    if (F5::BigipUtils::is_docker() && !$args{wait_for_enforcer}) {
        $response->{completedSuccessfully} = 1;
        $response->{generationIndex} = $expected_generation_index;
    }
    else {
        while (1) {
            if (-e $load_state_file) {
                my $mtime = stat($load_state_file)->mtime;
                if (!$load_state_file_mtime || $load_state_file_mtime != $mtime) {
                    my $parse_rv = F5::YAML::Schema::Utils::parse_json(filename => $load_state_file);
                    if ($parse_rv->{completed_successfully}) {
                        if ( (!$parse_rv->{data}->{completedSuccessfully} && $parse_rv->{data}->{loadDatetime} && F5::DateTimeUtils::convert_rfc3339_to_epoch($parse_rv->{data}->{loadDatetime}) > $start_time)
                             || ($parse_rv->{data}->{completedSuccessfully} && $parse_rv->{data}->{generationIndex} >= $core_file_update_rv->{generationIndex}) ) {
                            $response = $parse_rv->{data};
                            last;
                        }
                        else {
                            $logger->info("Bd has loaded up to generation '" . ($parse_rv->{data}->{generationIndex} || 0) . "'") if ($logger);
                            $load_state_file_mtime = $mtime;
                        }
                    }
                    else {
                        $error_msg = "Could not parse BD update response";
                        last;
                    }
                }
            }

            if ($UPDATE_CORE_PULSE_MSG_INTERVAL &&
                F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $UPDATE_CORE_PULSE_MSG_INTERVAL,
                                                                   start_time => $last_pulse_time) == 0) {

                $logger->info("Still waiting for bd to update $load_state_file (" . (time - $start_time) . " seconds)") if $logger;
                $last_pulse_time = time;
            }

            if (F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $UPDATE_CORE_MAX_INTERVAL,
                                                                   start_time => $start_time) == 0) {
                $logger->info("Configuration update exceeded maximum interval, aborting") if $logger;
                last;
            }
            Time::HiRes::sleep(0.2);
        }
        if ($start_time != $last_pulse_time) {
            $logger->info("Found update from bd in $load_state_file after " . (time - $start_time) . " seconds") if $logger;
        }
    }

    F5::PrepareFiles::cleanup_update_files();
    my $rv;
    if ($response && $response->{messages}) {
        $error_msg = "(" . join('\n', @{$response->{messages}}) . ")";
    }

    if ($response->{completedSuccessfully}) {
        $logger->debug("Update core completed to $response->{generationIndex} $error_msg") if ($logger);
        $rv = '0';
    }
    else {
        $error_msg ||= 'No response from BD';
        $logger->info("Update core failed $error_msg") if ($logger);
        $rv = $error_msg;
    }

    $locker->unlock();
    return $rv;
}

#----------------------------------------------------------------------

#TODO: Evaluation the best value for this
my $FULL_UPDATE_CORE_PERCENT = 0.05;

sub delayed_update_core {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $username = $params{username};
    my $full_sync = $params{full_sync};
    my $skip_cluster_sync = $params{skip_cluster_sync};
    my $do_table_checksums = $params{do_table_checksums};
    my @update_core_queue = @{$params{update_core_queue} || []};

    if (@update_core_queue) {
        my $lock_file = $F5::DbUtils::LOCK_FILE;
        my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                                 logger => $logger);

        $locker->lock();

        my $update_core_rv;
        my $cluster_sync;
        if (!$skip_cluster_sync) {
            $cluster_sync = F5::Cluster::Sync->new(dbh => $dbh, logger => $logger);
        }

        my $bot_defense = F5::PrepareConf::BotDefense->new(dbh => $dbh, logger => $logger);
        F5::RamCache::prepare_conf(dbh => $dbh);

        my $num_applied_policies = scalar @update_core_queue;
        if ($full_sync || ($num_applied_policies > 1 &&
            ($num_applied_policies / F5::DbUtils::exists_in_table(dbh => $dbh, table_name => 'DCC.ACCOUNTS') >= $FULL_UPDATE_CORE_PERCENT))) {

            $full_sync = 1;
            $update_core_rv = F5::SetActive::Impl::send_update_core(dbh => $dbh,
                                                                    logger => $logger,
                                                                    do_table_checksums => $do_table_checksums,
                                                                    prepare_account_tables => 1,
                                                                    wait_for_enforcer => $params{wait_for_enforcer},
                                                                    job => $update_core_queue[0]->{job},
                                                                   );
        }

        for my $update_core (@update_core_queue) {
            my $policy = $update_core->{policy};
            if (!$full_sync) {
                my $update_core_rv = F5::SetActive::Impl::send_update_core(dbh => $dbh,
                                                                           logger => $logger,
                                                                           do_table_checksums => $update_core->{do_table_checksums},
                                                                           policy => $policy,
                                                                           original_account_id => $update_core->{original_account_id},
                                                                           prepare_account_tables => 1,
                                                                           wait_for_enforcer => $params{wait_for_enforcer},
                                                                           job => $update_core->{job},
                                                                          );
                F5::SetActive::Impl::invalidate_ramcache_cache(dbh => $dbh,
                                                               logger => $logger,
                                                               policy => $policy,);

                $bot_defense->prepare_bot_defense_profiles(policy => $policy);
            }
            F5::SetActive::Impl::log_update_core_rv(dbh => $dbh,
                                                    logger => $logger,
                                                    policy => $policy,
                                                    username => $username,
                                                    update_core_rv => $update_core_rv,);

            if ($cluster_sync && !$full_sync) {
                $cluster_sync->send_sync(policy_id => $policy->{policy_id}, account_id => $policy->{account_id}, policy_name_crc => $policy->{policy_name_crc});
            }
        }
        if ($cluster_sync && $full_sync) {
            $cluster_sync->send_sync(force_full_sync => 1);
            $bot_defense->prepare_bot_defense_profiles();
        }
        $locker->unlock();
    }

}

#----------------------------------------------------------------------

sub set_dbh_attribs {
  my(%params) = @_;
  my $dbh = $params{dbh};
  my $set_attribs = $params{set_attribs};

  my %saved_attribs;
  for my $k (keys %$set_attribs) {
    $saved_attribs{$k} = $dbh->{$k};
    $dbh->{$k} = $set_attribs->{$k};
  }

  return %saved_attribs;
}

#----------------------------------------------------------------------

sub set_dbh_raise_error {
  my(%params) = @_;

  return set_dbh_attribs(dbh => $params{dbh},
                         set_attribs => { RaiseError => 1,
                                          PrintError => 0 });
}

#----------------------------------------------------------------------

sub restore_dbh_attribs {
  my(%params) = @_;

  set_dbh_attribs(dbh => $params{dbh},
                  set_attribs => $params{saved_dbh_attribs});
}

#----------------------------------------------------------------------
sub invalidate_ramcache_cache {
    my (%args) = @_;
    my $dbh = $args{dbh};
    my $logger = $args{logger};
    my $policy_name = $args{policy}->{policy_name};

    return unless defined $policy_name;

    $logger->info("Invalidate cache for policy $policy_name") if $logger;
    F5::RamCache::invalidate_cache_by_policy_name(policy_name => $policy_name, logger => $logger);
}
#----------------------------------------------------------------------
sub was_valid_command_issued {
  my(%params) = @_;
  if (defined $params{policy_id} ||
      $params{hsl_profile_id} ||
      defined $params{update_core_only} ||
      defined $params{do_update_core} ||
      $params{policy} ||
      $params{dump_bd_config_tables} ||
      $params{update_logger_cfg}) {
    return 1;
  }
  return;
}

#----------------------------------------------------------------------


1;
