package F5::Sigfile::AutoDownload;

use strict;
use warnings;

use F5::Deploy;
use F5::Iface;
#use SOAP::Lite +trace => qw(debug);
use SOAP::Lite;
use MIME::Base64;
use Config::IniFiles;
use F5::BigipVersionUtils;
use F5::License;
use F5::ServerLookup;
use F5::Utils::Net;
use PassCrypt;
use MIME::Base64;

my $CONF_INI_FILE = F5::Deploy::map_path('/ts/etc/services.ini');
my $DEFAULT_MAX_OUTBOUND_CALL_ATTEMPTS = 3;
my $DEFAULT_OUTBOUND_CALL_ATTEMPT_DELAY = 15;

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);
  return $self;
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;
  $self->{dbh} = $params{dbh};
  $self->{outbound_call_timeout} = $params{outbound_call_timeout};
  $self->{max_outbound_call_attempts} = $params{max_outbound_call_attempts} || $DEFAULT_MAX_OUTBOUND_CALL_ATTEMPTS;
  $self->{outbound_call_attempt_delay} = defined $params{outbound_call_attempt_delay} ? $params{outbound_call_attempt_delay} :
    $DEFAULT_OUTBOUND_CALL_ATTEMPT_DELAY;
  $self->{logger} = $params{logger};
  $self->{config} = Config::IniFiles->new(-file => $CONF_INI_FILE);

  $self->{user} = $ENV{F5WAF_SIGFILE_SOAP_USER} || 
                  PassCrypt::decrypt_password(MIME::Base64::decode_base64($self->{config}->val('sigfile_soap_service', 'user')));
  $self->{password} = $ENV{F5WAF_SIGFILE_SOAP_PASSWORD} || 
                      PassCrypt::decrypt_password(MIME::Base64::decode_base64($self->{config}->val('sigfile_soap_service', 'password')));

}

#----------------------------------------------------------------------

sub is_newer_sigfile_available {
  my($self) = @_;

  if (F5::License->new->is_aws_icm_platform()) {
      $self->{logger}->info("Updates are disallowed on an IC platform") if $self->{logger};
      return;
  }

  if ($self->call_soap_server(check_update_only => 1, 
                              )->valueof('//SoftwareUpdateResponse/UpdateExists') eq "true") {
      return 1;
  }
  return;
}

#----------------------------------------------------------------------

sub get_latest_sigfile {
  my($self, $save_dir) = @_;

  my $soap_resp = $self->call_soap_server(check_update_only => 0);
  my $filename = "$save_dir/" . $soap_resp->valueof('//SoftwareUpdateResponse/FileName');
  open(my $out_fh, ">", $filename) 
      or $self->store_error_and_die("Cannot open file: $filename");
  print $out_fh decode_base64($soap_resp->valueof('//SoftwareUpdateResponse/BinaryData')); 
  close($out_fh);

  $self->{logger}->debug("Received file: '$filename'");
  return $filename;
}

#----------------------------------------------------------------------

sub call_soap_server {
  my($self, %params) = @_;

  my $check_update_only = $params{check_update_only} || 0;

  my $last_updated_timestamp = $self->{sigfile_timestamp};

  my $user = $self->{user};
  my $password = $self->{password};
  my $host;
  my $url = '/soa/softwareupdate/'; 
  my $soap_action = 'SoftwareUpdate';

  my $proxy_params = $self->get_https_proxy_params();
  my $proxy_uri = $proxy_params->{proxy_uri};
  my $proxy_username = $proxy_params->{proxy_username};
  my $proxy_password = $proxy_params->{proxy_password};

  local $ENV{HTTPS_CA_FILE};

  if (F5::License->new->is_production) {
    $host = $self->{config}->val('sigfile_soap_service', 'prod_server_name');
    $ENV{HTTPS_CA_FILE} = '/config/ssl/ssl.crt/ca-bundle.crt';
  }
  else {
    $host = $self->{config}->val('sigfile_soap_service', 'dpi_dev_server_name');
  }

  local @LWP::Protocol::http::EXTRA_SOCK_OPTS;

  my $host_name = $host;
  unless ($proxy_uri) {
    my $host_ip = F5::ServerLookup::get_ip_address($host, $self->{outbound_call_timeout});
    @LWP::Protocol::http::EXTRA_SOCK_OPTS = ( PeerAddr => $host_ip,
                                              SSL_verifycn_name => $host,
                                              SSL_hostname      => $host,
                                              Host     => $host);
  }

  my %proxy_opts;
  $proxy_opts{proxy} = [ 'https' => undef ];
  local $ENV{PERL_LWP_SSL_VERIFY_HOSTNAME} = $proxy_uri ? 0 : 1;
  local $ENV{HTTPS_PROXY} = $proxy_uri;
  local $ENV{HTTPS_PROXY_USERNAME} = $proxy_username;
  local $ENV{HTTPS_PROXY_PASSWORD} = $proxy_password;

  if ($self->{outbound_call_timeout}) {
      $proxy_opts{timeout} = $self->{outbound_call_timeout};
  }

  my $soap_service = SOAP::Lite
    -> readable(1)
      -> uri ("https://$host$url")
        -> on_action( sub { return $soap_action } )
          -> proxy ("https://$user:$password\@$host$url", %proxy_opts)
            -> on_fault(sub { 
                          my($soap, $res) = @_; 
                          die ref $res ? $res->faultdetail : $soap->transport->status, "\n";
                        });
  $soap_service->transport->http_request->header('Host' => $host_name);
  if (F5::License->new->is_production) {
    $soap_service->transport->http_request->header('If-SSL-Cert-Subject' => '/CN='.quotemeta($host_name).'$');
  }

  my $bigip_version = F5::BigipVersionUtils::bigip_version();
  my $bigip_version_base = (split(/\./, $bigip_version))[0];
  $soap_service->transport->http_request->header('User-Agent' => "bigip/$bigip_version");

  $self->{logger}->debug("Timestamp: $last_updated_timestamp, BigIP: $bigip_version");
  my $data = SOAP::Data->value(
                               SOAP::Data->name("x:UnitIDFilter" => F5::Iface::iface_get_unit_id(dbh => $self->{dbh})), 
                               SOAP::Data->name("x:LastUpdatedFilter" => "$last_updated_timestamp GMT"),
                               SOAP::Data->name("x:FamilyFilter" => 'big-ip'),
                               SOAP::Data->name("x:ContextFilter" => "big-ip_v${bigip_version_base}.x"),
                               SOAP::Data->name("x:VersionFilter" => $bigip_version),
                               SOAP::Data->name("x:ContainerFilter" => $self->{container_filter}),
                               SOAP::Data->name("x:FileFilter" => $self->{file_filter}),
                               SOAP::Data->name("x:CheckForUpdateFilter" => $check_update_only),
                              );

  my $serviceNs = 'http://callhome.f5.com/schemas/su'; 

  my $soap_response;
  eval {
      local $SIG{ALRM} = sub { die "Timeout triggered\n" };
      if ($self->{outbound_call_timeout}) {
          alarm $self->{outbound_call_timeout};
      }

      $soap_response = $soap_service->call(
                                           SOAP::Data->name('x:SoftwareUpdateRequest') ->attr({'xmlns:x' => $serviceNs})
                                           => $data 
                                          );
  };

  my $ex = $@;
  if ($self->{outbound_call_timeout}) {
      alarm 0;
  }

  if ($ex) {
      $self->{logger}->info("SOAP request failed: ") if $self->{logger};
      $self->{logger}->info($ex) if $self->{logger};
      die "No response from update server\n";
  }

  if ( $soap_response->fault ) {
     die $soap_response->faultcode . " aha " . $soap_response->faultstring . "\n";
  }
  return $soap_response;
}

#----------------------------------------------------------------------

sub get_https_proxy_params {
    return F5::Utils::Net::get_proxy_settings();
}

#----------------------------------------------------------------------


1;

