package F5::Sigfile::Update;

use strict;
use warnings;

use File::Copy;
use File::Basename;
use Time::Local;
use F5::License;
use F5::Logger;
use IPC::Open3;
use Digest::MD5;

use F5::Deploy;
use F5::DateTimeUtils;
use F5::Syslog;
use F5::ImUtils;
use F5::LockFactory;
use F5::DbUtils;
use F5::BigipUtils;
use POSIX qw(strftime);

use F5::ASM::COAPI;

my $DAY_IN_SECONDS = 24 * 60 * 60;

my $FILE_CMD = '/usr/bin/file';
my $IM_FILE_OUTPUT = 'ISO 9660 CD-ROM';
my $LOCK_FILE = F5::Deploy::map_path("/ts/lock/sigfile_update.lock");

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  $self->init(%params);

  $self->{locker} = F5::LockFactory::get_locker(lock_filename => $LOCK_FILE,
                                                logger => $params{logger});
  $self->{locker}->lock();

  return $self;
}

#----------------------------------------------------------------------

sub DESTROY {
  my ($self) = @_;
  $self->{locker}->unlock();
}

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;

  $self->{dbh} = $params{dbh};
  $self->{logger} = $params{logger};
  $self->{license_text} = $params{license_text};
  $self->{outbound_call_timeout} = $params{outbound_call_timeout};
  $self->{lock_file} = $F5::DbUtils::LOCK_FILE;
  $self->{default_rv} = {};
}

#----------------------------------------------------------------------

sub update {
  my($self, %params) = @_;

  my $automatic_mode = $params{automatic_mode};
  my $filename = $params{filename};
  my $username = $params{username} || $params{user_name} || "";
  my $check_interval = $params{check_interval};
  my $check_for_updates = $params{check_for_updates};
  my $update_mode = $params{update_mode};

  $self->{check_interval} = $check_interval; # check_interval is needed for store_error_and_die
  $self->{username} = $username;

  if ( !$filename ) {
      $automatic_mode = 1 ;
  }

  my $logger = $self->{logger};
  my $sigfile_type = $self->{sigfile_type};

  my $lock_file = $self->{lock_file};
  my $locker = F5::LockFactory::get_locker(lock_filename => $lock_file,
                                           logger => $logger);
  $locker->lock();

  my @schedules;
  my $sigfile_status_update_mode = "Manual";
  if ($check_for_updates) {
      $self->check_for_updates(%params);
      $locker->unlock();
      return ({ %{$self->{default_rv}},
                completed_successfully => 1,
              });
  }

  if ($check_interval) {
      $sigfile_status_update_mode = "Scheduled";
      my %interval_in_days = ( 'daily' => 1,
                               'weekly' => 7,
                               'monthly' => 30 );

      for my $schedule_info ( $self->get_schedule_info(%params) ) {
          $update_mode ||= $schedule_info->{update_mode};
          if ($update_mode eq 'check-for-updates') {
              $self->check_for_updates(%params);
          }
          next if ($update_mode ne 'auto-interval');

          my $days_since_last_run = 9999; # Force initial run
          if ($schedule_info->{last_update_attempt}) {
              # Allow for long-running update process on previous attempt by adding 30 minutes to current time
              $days_since_last_run = (time + (30 * 60) - $schedule_info->{last_update_attempt}) / $DAY_IN_SECONDS;
          }

          my $interval_count = $schedule_info->{interval_count} || 1;
          if ($days_since_last_run < 0) {
              $logger->debug("$sigfile_type update was last run $days_since_last_run days into the future (system clock change?) -- running update") if $logger;
          }
          elsif ($days_since_last_run < ($interval_in_days{$schedule_info->{auto_update_interval}} * $interval_count)) {
              my $msg = "No need to update $sigfile_type -- only $days_since_last_run days since last run";
              $logger->debug($msg) if $logger;
              next;
          }
          push (@schedules, $schedule_info);
      }

      if ( !scalar @schedules) {
          $locker->unlock();
          return ({ %{$self->{default_rv}},
                    completed_successfully => 0,
                    error_msg => 'No update is scheduled.',
                    unscheduled => 1,
                  });
      }
  }

  $logger->info("Start $sigfile_type update") if $logger;
  $self->start_update(%params,
                      schedules => \@schedules);

  $self->check_service_agreement;

  if ($automatic_mode) {
      $params{filename} = $self->automatic_download();
      if ( !$params{filename}) {
          my $non_download_message = "The most recent $self->{sigfile_type} file is already installed.";
          $logger->info($non_download_message) if $logger;
          F5::Syslog::ts_syslog( 'info', "$non_download_message Signature file update skipped." );

          $locker->unlock();
          return ({ %{$self->{default_rv}},
                   completed_successfully => 1,
                   error_msg => $non_download_message,
                 });
      }
  }
  else {
    copy($params{filename}, $self->{save_dir} . "/" . basename($params{filename}));
    $params{filename} = $self->{save_dir} . "/" . basename($params{filename});
  }

  my $rv;
  my $filetype_check_output;
  if (F5::BigipUtils::is_docker()) {
      #Used at build time to update the signature DB
      $filetype_check_output = $IM_FILE_OUTPUT;
  }
  else {
  open(my $filetype_check, '-|', $FILE_CMD, $params{filename});
      $filetype_check_output = <$filetype_check>;
  close $filetype_check;

  if (!$self->check_filetype(filetype => $filetype_check_output)) {
      $locker->unlock();
      $self->store_error_and_die("Invalid $sigfile_type file format");
  }
  }
  
  if ($filetype_check_output =~ /$IM_FILE_OUTPUT/) {
      eval {
          %params = $self->process_im(%params);
      };
      if (my $err = $@) {
          $self->store_error_and_die($err);
      }
  }

  eval {
      $rv = $self->apply_changes(%params,
                                 sigfile_status_update_mode => $sigfile_status_update_mode,
                                 schedules => \@schedules,
                                );
  };
  if (my $err = $@) {
      $self->store_error_and_die($err);
  }


#$rv needs completed_successfully and msg
  $logger->info($rv->{msg}) if $logger;

  $self->report_status(%$rv);

  $self->post_process(%params);

  $locker->unlock();

  return $rv;
}


#----------------------------------------------------------------------

sub process_im {
    my ($self, %params) = @_;
    my $unpack_rv = F5::ImUtils::mount_and_unpack(%params);
    $params{filename} = $unpack_rv->{filename};
    $params{metadata} = $unpack_rv->{metadata};
    return %params;
}

#----------------------------------------------------------------------

sub automatic_download {
  my($self) = @_;

  my $logger = $self->{logger};

  my $filename;
  eval {
      my $download = $self->get_autodownload();
      my $available = $download->is_newer_sigfile_available;
      if ( $available ) {
          $filename = $download->get_latest_sigfile($self->{save_dir});
          $logger->info("New $self->{sigfile_type} file downloaded ($filename)") if $logger;
      }
      else {
          $self->report_status(completed_successfully => 1,
                               msg => "The most recent $self->{sigfile_type} file is already installed.",
                              );
      }
  };

  if ($@) {
      my $err_msg = $@;
      chomp($err_msg);
      $self->store_error_and_die("Download server cannot be reached ($err_msg). Please download the $self->{sigfile_type} file and install manually.");
  }

  return $filename;
}

#----------------------------------------------------------------------

sub store_error_and_die {
  my($self, $err_msg, $params) = @_;
  $params ||= {};

  my $result = { completed_successfully => 0,
                 error_msg => $err_msg,
               };

  $self->report_status(completed_successfully => 0,
                       msg => $err_msg);
  $self->{logger}->fatal($err_msg) if $self->{logger};
  die $result;
}

#----------------------------------------------------------------------

sub check_service_agreement {
    # Verification of "Service Check Date" was removed for BZ 709500. However, check_service_agreement must still be
    # called to handle license/provisioning check performed by subclasses.
}
#----------------------------------------------------------------------

sub report_status {
}

#----------------------------------------------------------------------

sub check_filetype {
    return 1;
}

#----------------------------------------------------------------------

sub start_update {
}

#----------------------------------------------------------------------

sub post_process {
}

#----------------------------------------------------------------------

sub check_for_updates {
}

#----------------------------------------------------------------------

sub load_file_for_datasyncd {
    my($self, %params) = @_;

    my $orig_filename = $params{orig_filename};
    my $version_create_time = $params{version_create_time};
    open(my $fh, $orig_filename);
    binmode($fh);
    my $md5 = Digest::MD5->new->addfile($fh)->hexdigest;
    close $fh;

    my $install_timestamp = F5::DateTimeUtils::convert_epoch_to_datetime(time(), 0);
    $install_timestamp =~ tr| |_|;
    $install_timestamp =~ s|[-/:]||g;
    my $create_timestamp = F5::DateTimeUtils::convert_epoch_to_datetime($version_create_time, 0);
    $create_timestamp =~ tr| |_|;
    $create_timestamp =~ s|[-/:]||g;
    my $filename = basename($orig_filename) . "__${install_timestamp}__${create_timestamp}__${md5}";
    my $tmp_filename = ".copying.$filename"; #intermediate step to move to same partition so the move is atomic.
    # The files in /var/datasync/updates/ are being read by datasyncd and by mcpd.
    # Because mcpd has SElinux enforcement, we need to make sure that files are moved with
    # the context of the destination directory. This is done with the -Z option.
    system("mv", "-Z", $orig_filename, "/var/datasync/updates/$tmp_filename");
    move("/var/datasync/updates/$tmp_filename", "/var/datasync/updates/$filename");
}


#----------------------------------------------------------------------

1;
