package F5::SystemUtils;

use strict;
use warnings;

use Symbol;
use IPC::Open3;
use File::Spec;
use File::Which;
use Memoize;

memoize('find_command_path');

#----------------------------------------------------------------------

sub run_system_tool {
 my(%params) = @_;

 my $timeout = $params{timeout};
 my $find_command_path = $params{find_command_path};
 my $command_line_params = $params{command_line_params};
 my $err_fh = gensym;
 my $out_fh = gensym;
 my $in_fh;

 if ($find_command_path) {
     my $command_path = find_command_path($command_line_params->[0]);
     if ($command_path) {
         $command_line_params->[0] = $command_path;
     }
     else {
         return {
            completed_successfully => 0,
            standard_error => "Could not find command: $command_line_params->[0]",
         };
    }
 }

 if ( ! $params{STDIN}) {
     open($in_fh, ">", File::Spec->devnull);
 }

 my $pid;
 my $err;
 my $output;
 my $child_exit_status;
#Deal with open3 bug in perl 5.8.8 See: http://www.perlmonks.org/?node_id=720878
 eval {
     local $SIG{ALRM} = sub { die "Timeout triggered\n" };
     if ($timeout) {
         alarm $timeout;
     }

     $pid = open3($in_fh, $out_fh, $err_fh,@{$command_line_params});
     if ($params{STDIN}) {
         print $in_fh $params{STDIN};
         close ($in_fh);
     }
     waitpid($pid, 0);
     $child_exit_status = $? >> 8;

     if ($timeout) {
         alarm 0;
     }

     $err = join('', <$err_fh>);
     $output =join('',<$out_fh>);
 };
 my $ex = $@;
 if ($ex) {
     if ($ex =~ /^open3:/) {
         #This is the failed forked child that should never have gotten here
         print STDERR "$ex\n";
         eval { require POSIX; POSIX::_exit(255); };
         exit 255;
     }
     elsif ($ex =~ /Timeout triggered/) {
         kill 9, $pid;
         waitpid($pid, 0);
         $child_exit_status = 1;
         $err = $ex;
     }
     else {
         die $ex; 
     }
 }

 return {
          completed_successfully => $child_exit_status ? 0 : 1,
          standard_error => $err,
          standard_output => $output,
        };
}

#----------------------------------------------------------------------

sub daemonize {
    # Double fork to create a grandchild

    if (my $child_pid = fork) {
        # parent
        waitpid($child_pid,0);
        # child exited, parent continues
    }
    else {
        # child
        if (my $pid = fork) {
            # child has a grandchild - can exit now
            exit;
        }
        # grandchild continues
    }

    #parent and grandchild continue
    return $$;
}

#----------------------------------------------------------------------

sub get_physical_memory_in_MB {
    open (my $in, "<", '/proc/meminfo');
    my $mem = 0;

    for my $line (<$in>) {
        if ($line =~ /MemTotal:\s+(\d+) kB$/) {
            $mem = int($1 / 1024);
        }
    }

    close $in;

    return $mem;
}

#----------------------------------------------------------------------

sub find_command_path {
    my($cmd) = @_;

    my @search_paths;
    if ($ENV{PATH}) {
        push(@search_paths, split(/:/, $ENV{PATH}));
    }
    push(@search_paths, qw(/bin
                           /usr/bin
                           /usr/local/bin
                         ));

    for my $search_path (@search_paths) {
        my $search_path_cmd = "$search_path/$cmd";
        if (-f $search_path_cmd) {
            return $search_path_cmd;
        }
    }
    return;
}

#----------------------------------------------------------------------

my $PYTHON_EXE = 'python';
my $PYTHON3_EXE = 'python3';

sub get_python_exe {
    my(%params) = @_;

    my @required_modules = @{$params{required_modules}};
    my $rv = $PYTHON_EXE;
    if (which($PYTHON3_EXE)) {
        if (@required_modules) {
            my $combined_required_modules = join(" ", @required_modules);            
            my $arg = "import pkgutil; print( (1,0)[False in list(map(lambda x: bool(pkgutil.find_loader(x)), \"$combined_required_modules\".split()))] )";
            my $res = run_system_tool( command_line_params => [ $PYTHON3_EXE, '-c', $arg] );
            if (1 == $res->{standard_output}) {
                $rv = $PYTHON3_EXE;
            }
        } else {
            $rv = $PYTHON3_EXE;
        }
    }
    return $rv;
}

#----------------------------------------------------------------------

1;
