package F5::ThreatCampaignFile::Update;

use strict;
use warnings;

use base qw(F5::Sigfile::Update);

use Time::Local;
use File::Copy;
use F5::Deploy;
use F5::ASMConfig::Entity::ThreatCampaignFileUpdateRes;
use F5::PrepareConf::Policy;
use F5::SetActive;
use F5::Provisioning;
use F5::Syslog;
use F5::Cfg;
use F5::BatchInsert;
use F5::DbUtils;
use F5::LockFactory;
use F5::ThreatCampaigns;
use F5::ThreatCampaignFile;
use F5::NegativeSignatures::SigFile::Update;
use F5::License;
use JSON;

use constant PRPLC_CFG_FILE  =>  'prepare_policy.cfg';

my $IM_FILE_OUTPUT = 'ISO 9660 CD-ROM';

our $SIGFILE_TYPE = "Threat Campaigns";
our $SAVE_DIR = F5::Deploy::map_path("/ts/var/threat_campaigns");

#----------------------------------------------------------------------

sub init {
  my($self, %params) = @_;
  $self->SUPER::init(%params);
  $self->{default_rv} = { threat_campaigns_count => 0,
                          threat_campaigns_updated_count => 0,
                        };
  $self->{sigfile_type} = $SIGFILE_TYPE;
  $self->{save_dir} = $SAVE_DIR;
}

#----------------------------------------------------------------------

sub get_schedule_info {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
#Gets auto_update_interval, update_mode
    my $rv = $dbh->selectrow_hashref(<<EndQuery);
SELECT * 
    FROM PLC.THREAT_CAMPAIGN_FILE_FLAGS
EndQuery

    my $unix_timestamp_string =  F5::DbUtils::get_unix_timestamp_string(dbh => $dbh, field => 'attempt_datetime');
    ($rv->{last_update_attempt}) =  $dbh->selectrow_array(<<EndQuery);
SELECT $unix_timestamp_string
    FROM PLC.THREAT_CAMPAIGN_FILE_AUTO_UPDATE_HISTORY
ORDER BY threat_campaign_file_auto_update_id DESC
   LIMIT 1
EndQuery

    return ($rv);
}

#----------------------------------------------------------------------

sub apply_changes {
    my($self, %params) = @_;

    my $automatic_mode = $params{automatic_mode};
    my $filename = $params{filename};
    my $asm_config_handler = $params{asm_config_handler};
    my $username = $params{username};
    my $check_interval = $params{check_interval};
    my $update_mode = $params{update_mode};
    my $sigfile_status_update_mode = $params{sigfile_status_update_mode};
    my $flg_is_genesis_file = $params{flg_is_genesis_file};
    my $threat_campaign_file_format = $params{threat_campaign_file_format};
    my $skip_set_active = $params{skip_set_active};
    my $skip_set_sigs_active = $params{skip_set_sigs_active};
    my $gpg_sign = $params{gpg_sign};

    if ($params{schedules} && @{$params{schedules}}) {
        my ($schedule_info) = @{$params{schedules}};
    }

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    my $threat_campaign_file;
    my $threat_campaign_file_update_status;
    eval {
        $threat_campaign_file = F5::ThreatCampaignFile->new(dbh => $dbh,
                                                            filename => $filename,
                                                            threat_campaign_file_format => $threat_campaign_file_format,
                                                            logger => $logger,
                                                           );

        if ($threat_campaign_file_format ne 'BINARY') {
            $threat_campaign_file->validate_version(%params) unless ($flg_is_genesis_file);
            $threat_campaign_file->validate_timestamp(%params);
        }

        $threat_campaign_file_update_status = $threat_campaign_file->load_threat_campaigns();

        $logger->debug("ThreatCampaigns loaded to database") if $logger;

        $threat_campaign_file->finish;
    };

    if ($@) {
        my $error = $@;

        if (ref($error)) {
            die $error;
        }
        else {
            $error =~ s|\s+ at \s+ /\S+ \s+ line \s+ \d .*||sx;
            $self->store_error_and_die($error, \%params);
        }
    }

    #Get last threat_campaign_file status
    my ($threat_campaign_file_status) = F5::ASMConfig::Entity::ThreatCampaignFileUpdateRes->retrieve_all(dbh => $dbh,
                                                                                                         sort_orders => [ F5::ASMConfig::SortOrder->new({field_name => "threat_campaign_file_status_id",
                                                                                                                                                         sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC(),})
                                                                                                                        ],
                                                                                                         page_size => 1,
                                                                                                        );

    my $updated_by = 'ASM User Interface';
    if ($params{originator}) {
        $updated_by .= " (updated from $params{originator})";
    }

    if(!$params{dry_run} && !$skip_set_sigs_active) {
	$logger->debug("Applying active policy for all accounts") if $logger;
        # Since threat campaigns are included in negsig ACY and prepare_negsig_collections,
        # they are applied via signatures
        my $negsig_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
        $negsig_update->set_sigs_active(asm_config_handler => $asm_config_handler, skip_set_active => $skip_set_active);
    }

    $threat_campaign_file_update_status->{rest_uuid} = $threat_campaign_file_status->get_rest_uuid();
    $threat_campaign_file_update_status->{updated_by} = $updated_by;
    $threat_campaign_file_update_status->{update_mode} = $sigfile_status_update_mode;
    $threat_campaign_file_status->store_update(store_obj => $threat_campaign_file_update_status);

    # We don't store the deleted list in the DB since it has a foreign key to the signatures table (and deleted items won't have that foreign key present)
    $threat_campaign_file_status->{threat_campaigns_modified} = $threat_campaign_file_update_status->{threat_campaigns_modified};

    my $success_msg = "$SIGFILE_TYPE updated successfully - $threat_campaign_file_status->{threat_campaigns_count} threat_campaigns loaded ($threat_campaign_file_status->{threat_campaigns_updated_count} updated)";

    if($params{dry_run} && !$params{in_wrapping_transaction}) {
        $dbh->rollback;
    }
    else {
        F5::ThreatCampaigns::create_policy_threat_campaigns(dbh => $dbh,
                                                            logger => $logger,
                                                           );

        F5::Utils::Rest::populate_uuids(dbh => $dbh,
                                        rest_entities => ['F5::ASMConfig::Entity::ThreatCampaign',
                                                          'F5::ASMConfig::Entity::PolicyThreatCampaign'],
                                       );
    }

    $threat_campaign_file_status->{completed_successfully} = 1;
    $threat_campaign_file_status->{msg} = $success_msg;
    $threat_campaign_file_status->{error_msg} = '';
    return ($threat_campaign_file_status);
}

#----------------------------------------------------------------------

sub check_service_agreement {
    my($self, %params) = @_;

    if (! F5::Provisioning::is_asm_on()) {
        $self->store_error_and_die("ASM is not enabled on the device");
    }
    if (! F5::License::is_threat_campaigns_licensed() ) {
        $self->store_error_and_die("Threat Campaigns are not licensed");
    }
        
    return $self->SUPER::check_service_agreement(%params);
}

#----------------------------------------------------------------------

sub check_filetype {
    my ($self, %params) = @_;
    return $params{filetype} =~ /$IM_FILE_OUTPUT|XML|data/; 
}

#----------------------------------------------------------------------

sub report_status {
  my($self, %params) = @_;
  my $completed_successfully = defined $params{completed_successfully} ? $params{completed_successfully} : 'undef';
  my $msg = $params{msg} || '';

  if ($self->{check_interval}) {
      $self->{logger}->debug("Automatic Threat Campaign update completed with rc = $completed_successfully ($msg)") if $self->{logger};
      $self->{dbh}->do(<<EndQuery, undef, $completed_successfully, $msg);
INSERT INTO PLC.THREAT_CAMPAIGN_FILE_AUTO_UPDATE_HISTORY 
    (completed_successfully, msg) VALUES (?,?)
EndQuery
  }
}

#----------------------------------------------------------------------

sub get_autodownload {
    my ($self, %params) = @_;
    return F5::ThreatCampaignFile::AutoDownload->new(dbh => $self->{dbh},
                                               outbound_call_timeout => $self->{outbound_call_timeout},
                                               logger => $self->{logger},
                                              );

}

#----------------------------------------------------------------------

1;
