package F5::Thrift;

use strict;
use warnings;

use Thrift;
use Time::HiRes;

use constant THRIFT_STRICT_MODE => 0;
our %thrift_types = (
                     string => {
                                ttype => TType::STRING(),
                                write_func => 'writeString',
                                read_func => 'readString',
                               },
                     i32 => {
                                ttype => TType::I32(),
                                write_func => 'writeI32',
                                read_func => 'readI32',
                               },
                     i64 => {
                                ttype => TType::I64(),
                                write_func => 'writeI64',
                                read_func => 'readI64',
                               },
                     bool => {
                                ttype => TType::BOOL(),
                                write_func => 'writeBool',
                                read_func => 'readBool',
                               },
                     double => {
                                ttype => TType::DOUBLE(),
                                write_func => 'writeDouble',
                                read_func => 'readDouble',
                               },
                    );


#======================================================================

package F5::Thrift::Object;

our $AUTOLOAD;

sub new {
    my($class, @vals) = @_;
    my $vals = {};
    if (scalar @vals == 1) {
        if ($vals[0]) {
            $vals = $vals[0];
        }
    }
    else {
        $vals = { @vals };
    }

    my $self = { %$vals };
    bless $self, $class;
    for my $field (values %{$self->get_struct_config()}) {
        if (! defined $self->{$field->{name}} && defined $field->{default}) {
            $self->{$field->{name}} = $field->{default};
        }
    }
    return $self;
}

#----------------------------------------------------------------------

sub getName {
    my($self) = @_;
    return $self->short_class;
}

#----------------------------------------------------------------------

sub short_class {
    my($self) = @_;
    if (! $self->{__short_class}) {
        my $short_class = ref($self);
        $short_class =~ s/^.*:://;
        $self->{__short_class} = $short_class;
    }
    return $self->{__short_class};
}

#----------------------------------------------------------------------

sub read {
    my ($self, $input) = @_;
    my $xfer  = 0;
    my $fname;
    my $ftype = 0;
    my $ktype = 0; # key type, if this is a map
    my $vtype = 0; # value type, if this is a map
    my $fid   = 0;

    $xfer += $input->readStructBegin(\$fname) if F5::Thrift::THRIFT_STRICT_MODE;
    while (1) {
        $xfer += $input->readFieldBegin(\$fname, \$ftype, \$fid);
        if ($ftype == TType::STOP) {
            last;
        }

        my $field = $self->get_struct_config()->{$fid};
        if (! $field) {
            $xfer += $input->skip($ftype);
            $xfer += $input->readFieldEnd() if F5::Thrift::THRIFT_STRICT_MODE;
            next;
        }

        my $field_type = $field->{type};
        my $list_count = 0;
        my $iteration_count = 1;
        if ($field->{list}) {
            $xfer += $input->readListBegin(\$ftype, \$list_count);
            $iteration_count = $list_count;
            $self->{$field->{name}} = [];
        }

        if ($field->{"map"}) {
            $xfer += $input->readMapBegin(\$ktype, \$vtype, \$list_count);
            $iteration_count = $list_count * 2;
            $self->{$field->{name}} = {};
        }

        my $map_key;
        my $type_category = $field->{type_category};
        for my $i (1.. $iteration_count) {
            if ($field->{"map"}) {
                $ftype = ($i % 2 ? $ktype : $vtype);
                $type_category = ($i % 2 ? $field->{type_category} : $field->{map_value_type_category});
            }

            my $read_value;
            if ($type_category eq 'enum') {
                $field_type = 'i32';
            }
            if ($thrift_types{$field_type}->{ttype}) {
                if ($thrift_types{$field_type}->{ttype} == $ftype) {
                    my $read_func = $thrift_types{$field_type}->{read_func};
                    $xfer += $input->$read_func(\$read_value);
                }
                else {
                    $xfer += $input->skip($ftype);
                }
            }
            elsif ($type_category eq 'struct') {
                if ($ftype == TType::STRUCT) {
                    my $class = $self->get_prefix() . '::' . $field_type;
                    $read_value = $class->new();
                    $xfer += $read_value->read($input);
                }
                else {
                    $xfer += $input->skip($ftype);
                }
            }

            if ($field->{list}) {
                push(@{$self->{$field->{name}}}, $read_value);
            }
            elsif ($field->{"map"}) {
                if ($i % 2) {
                    $map_key = $read_value;
                }
                else {
                    $self->{$field->{name}}->{$map_key} = $read_value;
                }
            }
            else {
                $self->{$field->{name}} = $read_value;
            }
        }

        if ($field->{list}) {
            $xfer += $input->readListEnd() if F5::Thrift::THRIFT_STRICT_MODE;
        }
        if ($field->{"map"}) {
            $xfer += $input->readMapEnd() if F5::Thrift::THRIFT_STRICT_MODE;
        }

        $xfer += $input->readFieldEnd() if F5::Thrift::THRIFT_STRICT_MODE;
    }
    $xfer += $input->readStructEnd() if F5::Thrift::THRIFT_STRICT_MODE;
    return $xfer;
}

#----------------------------------------------------------------------

sub write {
  my ($self, $output) = @_;
  my $xfer   = 0;
  my @map_as_list;
  $xfer += $output->writeStructBegin($self->short_class()) if F5::Thrift::THRIFT_STRICT_MODE;

  my $struct_config = $self->get_struct_config;
  for my $field (sort { $a->{index} <=> $b->{index} } values %$struct_config) {
      if (defined $self->{$field->{name}}) {
          my $field_type = $field->{type};
          my $list_count = 0;
          if ($field->{list}) {
              $list_count = scalar(@{$self->{$field->{name}}});
          }

          if ($field->{type_category} && $field->{type_category} eq 'enum') {
              $field_type = 'i32';
          }
          my $ttype;
          if ($thrift_types{$field_type}->{ttype}) {
              $ttype = $thrift_types{$field_type}->{ttype};
          }
          elsif ($field->{type_category} && $field->{type_category} eq 'struct') {
              $ttype = TType::STRUCT;
          }

          my ($map_key_field_type, $map_value_field_type);
          my ($map_key_ttype, $map_value_ttype);
          if ($field->{"map"}) {
              $map_value_field_type = $field->{map_value_type};
              @map_as_list = %{$self->{$field->{name}}};
              $list_count = scalar(@map_as_list);

              if ($field->{map_value_type_category} && $field->{map_value_type_category} eq 'enum') {
                  $map_value_field_type = 'i32';
              }
              if ($thrift_types{$map_value_field_type}->{ttype}) {
                  $map_value_ttype = $thrift_types{$field_type}->{ttype};
              }
              elsif ($field->{map_value_type_category} && $field->{map_value_type_category} eq 'struct') {
                  $map_value_ttype = TType::STRUCT;
              }
              $map_key_ttype = $ttype;
              $map_key_field_type = $field_type;
          }

          my $iteration_count = 1;
          if ($field->{list}) {
              $xfer += $output->writeFieldBegin($field->{name}, TType::LIST, $field->{index});
              $output->writeListBegin($ttype, $list_count);
              $iteration_count = $list_count;
          }
          elsif ($field->{"map"}) {
              $xfer += $output->writeFieldBegin($field->{name}, TType::MAP, $field->{index});
              $output->writeMapBegin($map_key_ttype, $map_value_ttype, $list_count / 2);
              $iteration_count = $list_count;
          }
          else {
              if ($ttype) {
                  $xfer += $output->writeFieldBegin($field->{name}, $ttype, $field->{index});
              }
          }

          for my $i (1.. $iteration_count) {
              my $write_value = $self->{$field->{name}};
              if ($field->{list}) {
                  $write_value = $write_value->[$i - 1];
              }
              elsif ($field->{"map"}) {
                  $write_value = $map_as_list[$i - 1];
                  $field_type = ($i % 2 ? $map_key_field_type : $map_value_field_type);
                  $ttype = ($i % 2 ? $map_key_ttype : $map_value_ttype);
              }

              if ($thrift_types{$field_type}->{ttype}) {
                  my $write_func = $thrift_types{$field_type}->{write_func};
                  $xfer += $output->$write_func($write_value);
              }
              elsif ($field->{type_category} && $field->{type_category} eq 'struct') {
                  $xfer += $write_value->write($output);
              }
          }
          if ($field->{list}) {
              $output->writeListEnd() if F5::Thrift::THRIFT_STRICT_MODE;
          }
          elsif ($field->{"map"}) {
              $output->writeMapEnd() if F5::Thrift::THRIFT_STRICT_MODE;
          }
          $xfer += $output->writeFieldEnd() if F5::Thrift::THRIFT_STRICT_MODE;
      }
  }

  $xfer += $output->writeFieldStop();
  $xfer += $output->writeStructEnd() if F5::Thrift::THRIFT_STRICT_MODE;
  return $xfer;
}

#----------------------------------------------------------------------

sub AUTOLOAD {
    my($self, $val) = @_;
    $AUTOLOAD =~ /::([^:]+)$/;
    my $key = $1;
    if (defined $val) {
        $self->{$key} = $val;
    }
    return $self->{$key};
}

#----------------------------------------------------------------------

sub get_struct_config {
    my($self) = @_;
    if ($self->{__struct_config}) {
        return $self->{__struct_config};
    }
    else {
        return $self->get_thrift_config(structs => $self->short_class());
    }
}

#----------------------------------------------------------------------

sub remove_undefined_fields {
    my ($self) = @_;
    my %fields = map { ( $_->{name} => $_->{type_category} ) } values %{$self->get_struct_config()};
    for my $field (keys %$self) {
        if (! exists $fields{$field}) {
          delete $self->{$field};
        }
        elsif ($fields{$field} eq "struct" && ref $self->{$field} && ref $self->{$field} =~ /ASMConfig/) {
            $self->{$field}->remove_undefined_fields;
        }
    }
}

#----------------------------------------------------------------------

sub get_thrift_config {
    die "Abstract method -- must be implemented by subclass";
}

#----------------------------------------------------------------------

sub get_prefix {
    die "Abstract method -- must be implemented by subclass";
}

#======================================================================

package F5::Thrift::Client;

our $AUTOLOAD;

sub new {
  my ($classname, $input, $output) = @_;
  my $self      = {};
  $self->{input}  = $input;
  $self->{output} = defined $output ? $output : $input;
  $self->{seqid}  = 0;
  return bless($self,$classname);
}

sub DESTROY {
}

sub AUTOLOAD {
    my ($self, @params) = @_;
    $AUTOLOAD =~ /::([^:]+)$/;
    my $method_call = $1;

    $self->initiate_call(method_call => $method_call,
                         rpc_params => \@params,
                        );

    return $self->get_call_response(method_call => $method_call,
                                   );
}

sub initiate_call {
    my($self, %params) = @_;
    my $method_call = $params{method_call};
    my $rpc_params = $params{rpc_params};

    my $this_config = $self->get_thrift_config(rpc => $method_call);
    my $method_prefix = $self->get_prefix();
    $method_prefix =~ s/^.*:://;

    # Send request
    my %args;
    for my $i (0..$#{$this_config->{arguments}}) {
        $args{$this_config->{arguments}->[$i]->{name}} = $rpc_params->[$i];
    }

    my $generic_object_class = $self->get_generic_object_class();
    my $args_obj = $generic_object_class->new({
                                               __short_class => "${method_prefix}_${method_call}_args",
                                               __struct_config => { map { ($_->{index} => $_) } @{$this_config->{arguments}} },
                                               %args,
                                              });
    $self->{output}->writeMessageBegin($method_call, TMessageType::CALL, ++$self->{seqid});
    $args_obj->write($self->{output});
    $self->{output}->writeMessageEnd() if F5::Thrift::THRIFT_STRICT_MODE;
    $self->{output}->getTransport()->flush();
}

sub get_call_response {
    my($self, %params) = @_;
    my $method_call = $params{method_call};

    my $this_config = $self->get_thrift_config(rpc => $method_call);
    my $method_prefix = $self->get_prefix();
    $method_prefix =~ s/^.*:://;

    # Receive result
    my $rseqid = 0;
    my $fname;
    my $mtype = 0;

    $self->{input}->readMessageBegin(\$fname, \$mtype, \$rseqid);
    if ($mtype == TMessageType::EXCEPTION) {
        my $x = new TApplicationException();
        $x->read($self->{input});
        $self->{input}->readMessageEnd() if F5::Thrift::THRIFT_STRICT_MODE;
        die $x;
    }

    my $generic_object_class = $self->get_generic_object_class();
    my $result = $generic_object_class->new({
                                             __short_class => "${method_prefix}_${method_call}_result",
                                             __struct_config => { map { ($_->{index} => $_) } @{$this_config->{return_value}} },
                                            });
    $result->read($self->{input});
    $self->{input}->readMessageEnd() if F5::Thrift::THRIFT_STRICT_MODE;

    if ($this_config->{return_value}->[0]->{type} eq 'void') {
        if ($params{raw_return_value}) {
            return { void_return_value => 1 };
        }
        else {
            return undef;
        }
    }

    if ($params{raw_return_value} && defined $result && 
        (defined $result->{success} || defined $result->{ex})) {

        return $result;
    }

    if (defined $result->{success} ) {
        return $result->{success};
    }
    if (defined $result->{ex}) {
        die $result->{ex};
    }
    die "$method_call failed: unknown result";
}

sub set_transport {
    my($self, $input_transport, $output_transport) = @_;
    if (! $output_transport) {
        $output_transport = $input_transport;
    }

    $self->{input} = $input_transport;
    $self->{output} = $output_transport;
}

#----------------------------------------------------------------------

sub get_thrift_config {
    die "Abstract method -- must be implemented by subclass";
}

#----------------------------------------------------------------------

sub get_prefix {
    die "Abstract method -- must be implemented by subclass";
}

#======================================================================

package F5::Thrift::Processor;

sub new {
    my($class, $handler) = @_;
    my $self = {
                handler => $handler,
               };
    bless $self, $class;
    return $self;
}

sub process {
    my ($self, $input, $output) = @_;
    my $rseqid = 0;
    my $fname  = undef;
    my $mtype  = 0;

    my $method_prefix = $self->get_prefix();
    $method_prefix =~ s/^.*:://;
    $input->readMessageBegin(\$fname, \$mtype, \$rseqid);

    my $this_config = $self->get_thrift_config(rpc => $fname);
    if (! $this_config) {
      $input->skip(TType::STRUCT);
      $input->readMessageEnd() if F5::Thrift::THRIFT_STRICT_MODE;
      my $x = new TApplicationException('Function '.$fname.' not implemented.', TApplicationException::UNKNOWN_METHOD);
      $output->writeMessageBegin($fname, TMessageType::EXCEPTION, $rseqid);
      $x->write($output);
      $output->writeMessageEnd() if F5::Thrift::THRIFT_STRICT_MODE;
      $output->getTransport()->flush();
      return;
    }

    my $generic_object_class = $self->get_generic_object_class();
    my $args = $generic_object_class->new({
                                           __short_class => "${method_prefix}_${fname}_args",
                                           __struct_config => { map { ($_->{index} => $_) } @{$this_config->{arguments}} },
                                          });
    $args->read($input);
    $input->readMessageEnd() if F5::Thrift::THRIFT_STRICT_MODE;
    my @field_names = map { $_->{name} } @{$this_config->{arguments}};
    my $result = $generic_object_class->new({
                                             __short_class => "${method_prefix}_${fname}_result",
                                             __struct_config => { map { ($_->{index} => $_) } @{$this_config->{return_value}} },
                                            });

    # Special handling for full_buffering - update the BufferedTransport directly
    if ($fname eq 'full_buffering') {
        $input->getTransport()->{rBufSize} = 10_000_000; # 10 MB
        $input->getTransport()->{wBufSize} = 10_000_000; # 10 MB
        $output->getTransport()->{rBufSize} = 10_000_000; # 10 MB
        $output->getTransport()->{wBufSize} = 10_000_000; # 10 MB
        $result->{success} = 1;
    }
    else {
        eval {
            $result->{success} = $self->{handler}->$fname(@{$args}{@field_names});
        };
        if ($@) {
            $result->{ex} = $@;
        }
    }

    $output->writeMessageBegin($fname, TMessageType::REPLY, $rseqid);
    $result->write($output);
    $output->writeMessageEnd() if F5::Thrift::THRIFT_STRICT_MODE;
    $output->getTransport()->flush();

    return 1;
}

sub get_thrift_config {
    die "Abstract method -- must be implemented by subclass";
}

#----------------------------------------------------------------------

sub get_prefix {
    die "Abstract method -- must be implemented by subclass";
}

#======================================================================

package F5::Thrift::SSLSocket;

use base qw/Thrift::Transport/;

sub new {
    my($class, %params) = @_;
    my $self = {};
    bless $self, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    if (! $IO::Socket::SSL::VERSION) {
        eval {
            require IO::Socket::SSL;
        };
        if ($@) {
            die "Could not load IO::Socket::SSL\n";
        }
    }

    $self->{socket} = IO::Socket::SSL->new(%params);
    $self->{initial_write_succeeded} = 0;
}

#----------------------------------------------------------------------

sub isOpen {
    return 1;
}

sub open {
    # nop
}

sub close {
    # nop
}

sub flush {
    # nop
}

sub blocking {
    my($self, $val) = @_;
    $self->{socket}->blocking($val);
}

#----------------------------------------------------------------------

sub read {
    my($self, $bufsize) = @_;
    my $buf;
    sysread($self->{socket}, $buf, $bufsize);
    return $buf || '';
}

#----------------------------------------------------------------------

sub write {
    my($self, $buf) = @_;
    my $pos = 0;

    eval {
        while ($pos < length($buf)) {
            my $last_write_count = syswrite($self->{socket}, $buf, length($buf) - $pos, $pos);
            if (! $last_write_count) {
                Time::HiRes::usleep(50_000);
            }
            else {
                $self->{initial_write_succeeded} = 1;
            }
            $pos += $last_write_count;
        }
    };
    if ($@) {
        if (! $self->{initial_write_succeeded}) {
            die "Could not connect to device\n";
        }
        die $@;
    }
}

#======================================================================

1;
