package F5::Utils::Account;

use F5::DbUtils;
use F5::Cfg;
use F5::CRC;
use F5::EmServices;
use F5::ASMConfig::Entity::SessionAwarenessDataPoint;

#----------------------------------------------------------------------

sub validate_create_account {
    my (%params) = @_;

    my $policy = $params{policy};
    my $dbh = $params{dbh} || $policy->{dbh};
    my $logger = $params{logger};
    my $dcc_dbname = $params{dcc_dbname} || 'DCC';

    if ($policy->{account_id}) {
        $logger->info("Cannot create account for Policy $policy->{policy_name}. It already exists.") if $logger;
        return $policy->{account_id};
    }

    my ($account_id) =  $dbh->selectrow_array(<<EndQuery, undef, $policy->{policy_name});
SELECT account_id
  FROM $dcc_dbname.ACCOUNTS
 WHERE policy_name = ?
EndQuery

    if ($account_id) {
        $logger->info("Found account id $account_id for Policy $policy->{policy_name}") if $logger;
        return $account_id;
    }

    $account_id = F5::DbUtils::find_lowest_unused_id(dbh => $dbh,
                                                     table_name => "$dcc_dbname.ACCOUNTS",
                                                     id_name => 'account_id',
                                                     min_id_value => 1,
                                                    );

    # NOTE: 'MaxAccounts' value in 'dcc.cfg' is updated by F5::ProcessHandler::start_bd on startup.
    my $max_accounts = F5::Cfg::cfg_get_config_item('/etc/ts/dcc/dcc.cfg', 'GENERAL', 'MaxAccounts');

    # the same validation is done in 'tm_daemon/mcp/asm/validation/AsmPolicy.cpp' at 'AsmPolicyValidator::run'
    # the validation logic has to be the same in both places
    if ($account_id > $max_accounts) {
        die "Cannot activate Policy $policy->{policy_name}. The system has reached its maximum number of $max_accounts active policies. You cannot create one more active policy on this BIG-IP with the current configuration.\n";
    }

    return $account_id;
}

#----------------------------------------------------------------------

sub create_account {
    my (%params) = @_;
    my $policy = $params{policy};
    my $dbh = $params{dbh} || $policy->{dbh};
    my $logger = $params{logger};
    my $dcc_dbname = $params{dcc_dbname} || 'DCC';


    my $account_id = validate_create_account(%params);

    my ($account_exists) = $dbh->selectrow_array(<<EndQuery, undef, $account_id);
SELECT account_id
  FROM $dcc_dbname.ACCOUNTS
 WHERE account_id = ?
EndQuery

    if ($account_exists) {
        return $account_id;
    }

    my $policy_name_crc = F5::CRC::get_crc($policy->{policy_name});
    F5::DbUtils::insert_row(dbh => $dbh,
                            table_name => "$dcc_dbname.ACCOUNTS",
                            rec => { account_id => $account_id,
                                     policy_name => $policy->{policy_name},
                                     policy_name_crc => $policy_name_crc,
                                   });

    F5::DbUtils::insert_row(dbh => $dbh,
                            table_name => "$dcc_dbname.ACCOUNT_APPLICATIONS",
                            rec => {policy_name_crc => $policy_name_crc,
                                    app_id => 0,
                                    max_sessions => 0,
                                    strict_mode => 0,
                                    verification_object =>, ''});

    F5::DbUtils::insert_row(dbh => $dbh,
                            table_name => "$dcc_dbname.ACCOUNT_APPLICATIONS",
                            rec => {policy_name_crc => $policy_name_crc,
                                    app_id => 1,
                                    max_sessions => 0,});

    F5::EmServices->new(dbh => $dbh)->add_to_account_transactions($policy_name_crc);
    $logger->info("Created account $account_id for Policy $policy->{policy_name}") if $logger;

    return $account_id;

}

#----------------------------------------------------------------------

sub delete_account {
    my (%params) = @_;
    my $policy = $params{policy};
    my $dbh = $params{dbh} || $policy->{dbh};
    my $logger = $params{logger};
    my $account_id = $params{account_id};
    my $policy_name_crc = $params{policy_name_crc};

    if ($policy) {
        if (!defined $account_id) {
            $account_id = $policy->{account_id};
        }
        $policy->clear_traffic_data(%params);
        $policy_name_crc ||= $policy->{policy_name_crc};
    }

    return unless ( $account_id || $policy_name_crc ) ;

    if ( !$policy_name_crc ) {
        ($policy_name_crc) = $dbh->selectrow_array(<<EndQuery, undef, $account_id);
SELECT policy_name_crc
    FROM DCC.ACCOUNTS
 WHERE account_id = ?
EndQuery
    }


    $dbh->do(<<EndQuery, undef, $policy_name_crc);
DELETE
    FROM DCC.ACCOUNTS
 WHERE policy_name_crc = ?
EndQuery

    #Cleanup tables that do not have a foreign key relation by account_id
    for my $table (F5::DbUtils::get_account_cleanup_tables("DCC")) {
        $dbh->do(<<EndQuery, undef, $policy_name_crc);
DELETE
  FROM $table
 WHERE policy_name_crc = ?
EndQuery
    }

    eval {
        F5::ASMConfig::Entity::SessionAwarenessDataPoint->delete_many(dbh => $dbh,
                                                                      policy => { policy_name_crc => $policy_name_crc },
                                                                      get_criteria => {},
                                                                      master_keys => 1,
                                                                     );
    };
    if ($@) {
        $logger->error("Session Awarenesss data could not be cleared ($@)") if $policy->{logger};
    }
}

#----------------------------------------------------------------------


1;
