package F5::Utils::FactorySettings;

use strict;
use warnings;

use F5::Deploy;
use F5::DbUtils;
use F5::Logger;
use YAML::XS;

#----------------------------------------------------------

sub compile_templates {
    my (%params) = @_;
    require F5::ASMConfig::Handler;
    require F5::ASMConfig::Entity::PolicyTemplate;
    my $logger = $params{logger} || F5::Logger->new('/ts/log/compile_templates.log');
    my $dbh = $params{dbh} || F5::DbUtils::get_dbh();
    my $client = $params{asm_config_client} ||  F5::ASMConfig::Handler->new(dbh => $dbh, username => "compile_templates", logger => $logger);

    my $allowed_templates = get_allowed_templates();

    my $restore_max_memory_size = $client->{max_memory_size};
    $client->{max_memory_size} = 1024 * 1024 * 1024;

    $client->begin_work();
    my $templates_iter = F5::ASMConfig::Entity::PolicyTemplate->retrieve_all(dbh => $dbh, get_criteria => {});

    while (my $template = $templates_iter->get_next) {
        if ($allowed_templates && ! $allowed_templates->{$template->{icontrol_name}}) {
            $logger->debug("Skipping template $template->{icontrol_name}") if $logger;
            next;
        }

        $logger->info("Compile $template->{icontrol_name}");
        my $policy = $client->add_policy($client->{session}, F5::ASMConfig::CallFlags->new(skip_mcp => 1, force_populate_rest_uuid => 1), 
                                         F5::ASMConfig::Policy->new(policy_template => $template,
                                                                    policy_name => $template->{icontrol_name},
                                                                    encoding => "utf-8",
                                                                    flg_encoding_selected => 1));
    }

    $client->rollback();

    $client->{max_memory_size} = $restore_max_memory_size;
}

#----------------------------------------------------------

sub delete_uneeded_templates {
    my (%params) = @_;
    require F5::ASMConfig::Handler;
    require F5::ASMConfig::Entity::PolicyTemplate;
    my $logger = $params{logger} || F5::Logger->new('/ts/log/delete_templates.log');
    my $dbh = $params{dbh} || F5::DbUtils::get_dbh();
    my $client = $params{asm_config_client} ||  F5::ASMConfig::Handler->new(dbh => $dbh, username => "delete_uneeded_templates", logger => $logger);

    my $allowed_templates = get_allowed_templates();
    if (! $allowed_templates) {
        $logger->info("All templates allowed.");
        return;
    }

    my $templates_iter = F5::ASMConfig::Entity::PolicyTemplate->retrieve_all(dbh => $dbh, get_criteria => {});

    while (my $template = $templates_iter->get_next) {
        if ($allowed_templates->{$template->{icontrol_name}}) {
            $logger->info("Keep $template->{icontrol_name}");
            next;
        }

        my $template_file = $template->{template_file};
        unlink($template_file);
        $template_file =~ s/\.xml/.bin/;
        if (-e $template_file) {
            unlink($template_file);
        }
        $template->{flg_is_user_defined} = 1;
        $template->delete();
        $logger->info("Delete $template->{icontrol_name}");
    }
}

#----------------------------------------------------------

sub set_internal_config_items {
    my (%params) = @_;
    require F5::ASMConfig::Handler;
    my $settings = get_factory_settings();
    return unless $settings->{internal_config_items};

    my $logger = $params{logger} || F5::Logger->new('/ts/log/set_internal_config_items.log');
    my $dbh = $params{dbh} || F5::DbUtils::get_dbh();
    my $client = $params{asm_config_client} ||  F5::ASMConfig::Handler->new(dbh => $dbh, username => "set_internal_config_items", logger => $logger);
    my $call_flags = F5::ASMConfig::CallFlags->new(skip_mcp => 1);

    for my $internal_config_item (keys %{$settings->{internal_config_items}}) {
        my $internal_config_item = F5::ASMConfig::InternalConfigItem->new({ config_item_name => $internal_config_item,
                                                                            current_value => $settings->{internal_config_items}->{$internal_config_item} });

        $client->update_internal_config_item($client->{session}, $call_flags, $internal_config_item);
    }
}

#----------------------------------------------------------

my $factory_settings;

sub get_factory_settings { 
    if (!$factory_settings) {
        my $settings_file = F5::Deploy::map_path('/usr/share/ts/config/factory_settings.yaml');
        if (-f $settings_file) {
            $factory_settings = YAML::XS::LoadFile($settings_file);
        }
        else {
            $factory_settings = {};
        }
    }
    return $factory_settings;
}

#----------------------------------------------------------

sub get_allowed_templates {
    # Get list of templates
    my $allowed_templates;
    my $settings = get_factory_settings();
    if ($settings->{templates}) {
        my @template_list = @{$settings->{templates}};
        $allowed_templates = { map {$_ => 1} @template_list };
    }
    return $allowed_templates;
}

#----------------------------------------------------------

1;

