package F5::WildcardEntityUpgrade;

use strict;
use warnings;

use List::Util qw(max);
use F5::Syslog;
use F5::CRC;
use F5::Violation;

use constant KEY_INPUT_TYPE_USER_INPUT_VALUE => 0x4;
use constant KEY_UI_TYPE_ALPHA_NUMERIC => 0x20;

sub globalize_charset_definitions {
  my(%params) = @_;

  my $dbh = $params{dbh};
  my $logger = $params{logger};
  my $policy_id = $params{policy_id};
  my $dbname = $params{dbname};

  my $where_clause = '';
  if ($policy_id ne '*') {
    $where_clause = "WHERE policy_id in ($policy_id)";
  }

  my $insert_defaults_sql = <<EndQuery;
UPDATE $dbname.PL_CHARSETS
   SET charset = replace(replace(charset, '3', '1') , '2', '0')
$where_clause
EndQuery

  $dbh->do($insert_defaults_sql);
}

#----------------------------------------------------------------------

sub add_wildcard_entities {
  my(%params) = @_;

  local $" = ',';
  my $dbh = $params{dbh};
  my $logger = $params{logger};
  my $policy_id = $params{policy_id};
  my $dbname = $params{dbname};

  my @entities = (
		  {
		   table => 'PL_OBJECTS',
		   wildcard_field => 'uri',
		   extra_fields => [ qw(app_id flg_tightening ) ],
		   violation => F5::Violation::VIOL_URL(),
		  },
		  {
		   table => 'PL_FILETYPES',
		   wildcard_field => 'name',
		   extra_fields => [ qw(flg_tightening) ],
		   violation => F5::Violation::VIOL_FILETYPE(),
		  },
		  {
		   table => 'PL_GLOBAL_PARAMS',
		   wildcard_field => 'name',
		   extra_fields => [ qw(param_attrib_id) ],
		   violation => F5::Violation::VIOL_PARAMETER(),
		  },
		 );

  my $policy_clause = '';
  my $policy_where_clause = '';
  if ($policy_id ne '*') {
    $policy_clause = "AND policy_id in ($policy_id)";
    $policy_where_clause = "WHERE id in ($policy_id)";
  }

  my $policies = $dbh->selectall_hashref("SELECT id, account_id, flg_disable_rejects FROM $dbname.PL_POLICIES $policy_where_clause",
                                         'id');

  for my $entity (@entities) {
    my $blocking_row_sth = $dbh->prepare(<<EndQuery);
SELECT policy_id, viol_index, flg_alarm, flg_reject, flg_learn 
  FROM $dbname.PL_BLOCKING
WHERE viol_index = ?
$policy_clause
EndQuery

    $blocking_row_sth->execute($entity->{violation});
    while (my $blocking_row = $blocking_row_sth->fetchrow_hashref) {
      if (!$blocking_row->{flg_reject} || $policies->{$blocking_row->{policy_id}}->{flg_disable_rejects}) {
	my %wildcard_entity;
	$wildcard_entity{$entity->{wildcard_field}} = '*';
	$wildcard_entity{policy_id} = $blocking_row->{policy_id};
	$wildcard_entity{app_id} = 0;
	$wildcard_entity{flg_is_pattern} = 1;
	my($max_pattern_order) = $dbh->selectrow_array(<<EndQuery, undef, $blocking_row->{policy_id});
SELECT MAX(pattern_order)
  FROM $dbname.$entity->{table}
WHERE policy_id = ?
EndQuery
	$max_pattern_order ||= 0;
	$wildcard_entity{pattern_order} = $max_pattern_order + 1;

	$wildcard_entity{flg_tightening} = 0;
	if ($blocking_row->{flg_alarm} || $blocking_row->{flg_learn}) {
	  $wildcard_entity{flg_tightening} = 1;
	}

	if ($entity->{table} eq 'PL_GLOBAL_PARAMS') {
	  # Add PL_PARAM_ATTRIBUTES record for this parameter
	  my $param_attrib_row = {};
	  $param_attrib_row->{policy_id} = $wildcard_entity{policy_id};
	  $param_attrib_row->{flg_tightening} = $wildcard_entity{flg_tightening};
	  $param_attrib_row->{flg_check_param_name_metachar} = 1;
	  $param_attrib_row->{key_input_type} = KEY_INPUT_TYPE_USER_INPUT_VALUE;
	  $param_attrib_row->{key_ui_type} = KEY_UI_TYPE_ALPHA_NUMERIC;

	  my @param_attrib_fields = keys %$param_attrib_row;
	  my @placeholders = ('?') x scalar @param_attrib_fields;

	  $dbh->do("INSERT INTO $dbname.PL_PARAM_ATTRIBUTES (@param_attrib_fields) VALUES (@placeholders)",
		   undef, @{$param_attrib_row}{@param_attrib_fields});

	  ($wildcard_entity{param_attrib_id}) = $dbh->last_insert_id(undef, undef, undef, undef);
	}

	my @entity_fields = qw(policy_id flg_is_pattern pattern_order);
	push(@entity_fields, $entity->{wildcard_field});
	push(@entity_fields, @{$entity->{extra_fields}}) if ($entity->{extra_fields});

	my @placeholders = ('?') x scalar @entity_fields;

	$dbh->do("REPLACE INTO $dbname.$entity->{table} (@entity_fields) VALUES (@placeholders)",
		 undef, @wildcard_entity{@entity_fields});

	if ($entity->{table} eq 'PL_OBJECTS') {
	  $wildcard_entity{app_id} = 1;
	  $wildcard_entity{pattern_order}++;
	  $dbh->do("REPLACE INTO $dbname.$entity->{table} (@entity_fields) VALUES (@placeholders)",
		   undef, @wildcard_entity{@entity_fields});
	}
      }
    }
  }
}

#----------------------------------------------------------------------

sub convert_allowed_objects_regexp_to_wildcard {
  my($dbh, $logger, $import_table, $import_table_name, $import_config, $new_dbname, $param_account_id) = @_;

  $new_dbname ||= 'PLC';
  my $pattern_order = 1;
  my $next_avail_key = (defined max keys %$import_table) ? (max keys %$import_table) + 1 : 1;

  my $policy_to_account = $dbh->selectall_hashref("SELECT id, account_id FROM $new_dbname.PL_POLICIES", 'id');


  for my $key (keys %$import_table) {
    my $row = { %{$import_table->{$key}} };
    my $account_id = defined $param_account_id ? $param_account_id : $policy_to_account->{$row->{policy_id}}->{account_id};
    my @wildcard_uris = convert_regexp_to_wildcard($row->{reg_exp});
    unless (@wildcard_uris) {
      $logger->info("Can't convert $row->{reg_exp}") if $logger;
      delete $import_table->{$key};
      next;
    }

    $row->{flg_is_pattern} = 1;
    # Remove the current row -- it will re-inserted in the loop below.
    delete $import_table->{$key};

    for my $wildcard_uri (@wildcard_uris) {

      for my $app_id ((0,1)) {
        $import_table->{$next_avail_key++} = { %$row,
                                               uri => $wildcard_uri,
                                               pattern_order => $pattern_order++,
                                               app_id => $app_id,
                                             };
      }
    }
  }

}

#----------------------------------------------------------------------

sub convert_param_name_regex_to_wildcard {
    my($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash) = @_;

    my %next_pattern_order;

    for my $key (keys %$import_table) {
        my $row = $import_table->{$key};
        $row->{flg_is_pattern} = 0 unless exists $row->{flg_is_pattern};
        $row->{pattern_order} = undef unless exists $row->{pattern_order};
        next unless $row->{param_name_regexp};

        $next_pattern_order{$row->{policy_id}} = 1
            if not defined $next_pattern_order{$row->{policy_id}};
        $row->{flg_is_pattern} = 1;

        my @wildcard_names = convert_regexp_to_wildcard( $row->{param_name_regexp});
        unless (@wildcard_names) {
            delete $import_table->{$key};
            F5::Syslog::ts_syslog( 'warning', "%s", "Failed to translate regex $row->{param_name_regexp} to wildcard, global parameter record is omitted." );
# There might be certain regexes which we are not able to transfer to wildcard.
# In these cases we ommit the global parameter, and we give a warning in the syslog.
            next;
        }

        # Only insert the first; all others (from alternation) will be ignored
        $row->{name} = $wildcard_names[0];
	if(not defined $row->{pattern_order}) {
            $row->{pattern_order} = $next_pattern_order{$row->{policy_id}}++;
        }
    }
}

#----------------------------------------------------------------------

sub convert_regexp_to_wildcard {
  my($regexp) = @_;

  # Convert [^ to [!
  $regexp =~ s/(?<!\\)\Q[^/[!/g;
  # Convert perl \d digit to [0-9] character class
  $regexp =~ s/\\d/[0-9]/g;
  # Convert perl \w digit to [0-9] character class
  $regexp =~ s/\\w/[a-zA-Z0-9_]/g;
  # Convert .* to *
  $regexp =~ s/(?<!\\)\Q.*/*/g;
  # Convert .+ to *
  $regexp =~ s/(?<!\\)\Q.+/*/g;
  # Convert lone . (period) to ?
  $regexp =~ s/(?<!\\)\Q./?/g;

  # Convert simple pipe-delimited parenthesized alternates
  my @regexps;
  # Cannot convert if there are nested parens
  return if ($regexp =~ / (?<!\\) \( .* (?<!\\) \( /x);

  if ($regexp =~ / (?<!\\) \( (.*) (?<!\\) \) /x) {
    my @alternates = split(/ (?<!\\) \| /x, $1);
    @regexps = ($regexp) x scalar @alternates;
    for my $i (0..$#regexps) {
      $regexps[$i] =~ s/ (?<!\\) \( (.*) (?<!\\)\) /$alternates[$i]/x;
    }
  }
  else {
    $regexps[0] = $regexp;
  }

  for my $i (reverse 0..$#regexps) {
    if ($regexps[$i] =~ / (?<!\\) [{}()] /x) {
      splice(@regexps, $i, 0);
      next;
    }

    # Drop backslash escaping on regexp special characters
    $regexps[$i] =~ s/\\([.+()|^\$])/$1/g;

    if ($regexp =~ /\\/) {
      splice(@regexps, $i, 0);
    }
  }

  if (wantarray) {
    return @regexps;
  }
  elsif (@regexps) {
    return $regexps[0];
  }
  else {
    return;
  }
}

#----------------------------------------------------------------------

sub get_metachar_ids {
  my($dbh, $logger, $import_table, $import_table_name, $import_config, $subst_hash) = @_;

  # Preload defaults from database
  my %metachar_id_to_char = @{$dbh->selectcol_arrayref("SELECT id,hex_value FROM PLC.PL_ILLEGAL_META_CHARS_POOL", { Columns => [1,2] })};

  # Load whatever was passed in the policy file
  for my $row (values %$import_table) {
    if ($row->{hex_value} =~ /^[0-9a-f]{1,2}$/i) {
      $metachar_id_to_char{$row->{id}} = $row->{hex_value};
    }
    elsif ($row->{value} =~ /^[0-9a-f]{1,2}$/i) {
      $metachar_id_to_char{$row->{id}} = $row->{value};
    }
  }

  # Convert values from hex to decimal
  for my $k (keys %metachar_id_to_char) {
    $metachar_id_to_char{$k} = hex($metachar_id_to_char{$k});
  }

  return \%metachar_id_to_char;
}

1;
