package F5::XML::Schema::Builder;

use strict;
use warnings;

use XML::Twig;
use File::Basename;
use YAML::XS;
use MIME::Base64;
use Time::Local;
use Encode qw(:all);
eval {
    require Encode::HanExtra;
};

use IPC::Open3;
use Symbol;
use File::Spec;

use F5::Logger;

my $XMLLINT_CMD = '/usr/bin/xmllint';

#----------------------------------------------------------------------

sub create_node_schema {
  my(%params) = @_;
  
  my $node_handler = $params{node_handler};
  my $node = $params{node};
  my $base_twig = $params{base_twig};
 
  my $node_xpath;

  $node_handler->fixup_node(node => $node);

  if ($params{subnode_base_xpath} && $node->{xpath}) {
    $node_xpath = $params{subnode_base_xpath} . '/' . $node->{xpath};
  }
  else {
    $node_xpath = $node->{xpath};
  }
  return unless ($node_xpath);

  for my $field (@{$node->{fields}}) {
    my $add_point;
    my @field_xpaths;
    if (defined($field->{xpath})) {
      push(@field_xpaths, $field->{xpath});
    }
    if (defined($field->{foreign_links})) {
        for my $foreign_link (@{$field->{foreign_links}}) {
            if ($foreign_link->{xpath} && ! grep { $_->{xpath} && $foreign_link->{xpath} eq $_->{xpath} } @{$node->{fields}}) {
                push(@field_xpaths, $foreign_link->{xpath});
            }
        }
    }

    for my $field_xpath (@field_xpaths) {
      next if ($field_xpath =~ /\.\.\//);

      if ($field_xpath eq '') {
        $add_point = get_complex_type_node(   node_handler => $node_handler,
                                              start_schema_xpath => '/xs:schema',
                                              node_xpath => $node_xpath,
                                              base_twig => $base_twig,
                                              as_simple_content => 1,
                                          );
        $add_point->set_att(base => 'xs:string');
      }
      else {
        my($xpath) = $field_xpath;
        if ($xpath =~ /^(.*)\@(.*)$/) {
          my($field_name, $attribute) = ($1, $2);
          $add_point = get_complex_type_node( node_handler => $node_handler,
                                              start_schema_xpath => '/xs:schema',
                                              node_xpath => $node_xpath . '/' . $field_name,
                                              base_twig => $base_twig,
                                              get_attribute_insert_point => 1,
                                            );

          my @existing_attributes = $add_point->findnodes('xs:attribute');
          if (!grep { $_->{att}->{name} eq $attribute } @existing_attributes) {
              my $attribute_element = XML::Twig::Elt->new('xs:attribute' => { name => $attribute,
                                                                              type => 'xs:string',
                                                                            });
              $attribute_element->paste(last_child => $add_point);
          }
        }
        else {
          if ($xpath =~ s|^(.*)/||) {
            my $pre_xpath = $1;
            $add_point = get_complex_type_node( node_handler => $node_handler,
                                                start_schema_xpath => '/xs:schema',
                                                node_xpath => $node_xpath . '/' . $pre_xpath,
                                                base_twig => $base_twig);
          }
          else {
            $add_point = get_complex_type_node(  node_handler => $node_handler,
                                                 start_schema_xpath => '/xs:schema',
                                                 node_xpath => $node_xpath,
                                                 base_twig => $base_twig);
          }
          my $type = 'xs:string';

          my $field_element;
          if ($field->{field_func} && $field->{use_field_func_for_schema}) {
              my $field_func = $field->{field_func};
              $field_element = $node_handler->$field_func(mode => 'build_schema');
          }
          elsif ($field->{allow_any_in_schema}) {
              $add_point = get_complex_type_node( node_handler => $node_handler,
                                                  start_schema_xpath => '/xs:schema',
                                                  node_xpath => $node_xpath . '/' . $xpath,
                                                  base_twig => $base_twig);

              $field_element = XML::Twig::Elt->new('xs:any' => {
                                                                minOccurs => 0,
                                                                maxOccurs => 'unbounded',
                                                                processContents => 'lax'
                                                               });
          }

          if (!$field_element) {
              my $type = 'xs:string';
              if ($field->{value_type} && $field->{value_type} eq 'boolean') {
                  $type = 'xs:boolean';
              }
              $field_element = XML::Twig::Elt->new('xs:element' => { name => $xpath,
                                                                     minOccurs => 0,
                                                                     maxOccurs => 1,
                                                                     type => $type,
                                                                   });
          }

          $field_element->paste(last_child => $add_point);

        }
      }
    }

    if ($field->{schema_mixed_content}) {
      my $node_element = get_complex_type_node(  node_handler => $node_handler,
                                                 start_schema_xpath => '/xs:schema',
                                                 node_xpath => $node_xpath,
                                                 base_twig => $base_twig,
                                                 elements_as_choice => $params{elements_as_choice},
                                                     );
      my $any_element = XML::Twig::Elt->new('xs:any' => {
                                                         minOccurs => 0,
                                                         maxOccurs => 'unbounded',
                                                         processContents => 'lax'
                                                        });
      $any_element->paste(last_child => $node_element);
      $node_element->parent->set_att('mixed' => 'true');

      my $attribute_element = XML::Twig::Elt->new('xs:attribute' => { name => 'base64_encoded',
                                                                      type => 'xs:string',
                                                                    });
      $attribute_element->paste(last_child => $node_element->parent);
    }
  }

  my $node_element = get_complex_type_node( node_handler => $node_handler,
                                            start_schema_xpath => '/xs:schema',
                                            node_xpath => $node_xpath,
                                            base_twig => $base_twig,
                                            get_node_element => 1,
                                            elements_as_choice => $params{elements_as_choice},
                                                 );
  if ($node_xpath ne '/policy') {
    $node_element->set_att(minOccurs => 0,
                           maxOccurs => 'unbounded');
  }

  if ($node_handler->can('finalize_node')) {
    $node_handler->finalize_node(node_element => $node_element, node => $node);
  }

  for my $subnode (@{$node->{subnodes}}) {
    create_node_schema(node_handler => $node_handler,
                       node => $subnode,
                       base_twig => $base_twig,
                       subnode_base_xpath => $node_xpath);
  }
}

#----------------------------------------------------------------------
sub get_complex_type_node {
  my(%params) = @_;

  my $node_handler = $params{node_handler};
  my $schema_xpath = $params{start_schema_xpath} || '';
  my $node_xpath = $params{node_xpath};
  my $base_twig = $params{base_twig};

  $node_xpath =~ s|/./|/|g;
  my @elements = split(m|/|, $node_xpath);
  shift(@elements);
  my $last_element = pop(@elements);
  for my $element (@elements) {
    if ( $params{elements_as_choice} && $params{elements_as_choice}->{$element} ) {
        $schema_xpath .= '/xs:element[@name="' . $element . '"]/xs:complexType/xs:choice';
    }
    else{
        $schema_xpath .= '/xs:element[@name="' . $element . '"]/xs:complexType/xs:sequence';
    }
  }

  if ($params{get_attribute_insert_point}) {
    $schema_xpath .= '/xs:element[@name="' . $last_element . '"]/xs:complexType';
  }
  elsif ($params{get_node_element}) {
    $schema_xpath .= '/xs:element[@name="' . $last_element . '"]';
  }
  elsif ($params{as_simple_content}) {
    $schema_xpath .= '/xs:element[@name="' . $last_element . '"]/xs:complexType/xs:simpleContent/xs:extension';
  }
  else {
    $schema_xpath .= '/xs:element[@name="' . $last_element . '"]/xs:complexType/xs:sequence';
  }

  my $retval = add_or_find_node($base_twig, $schema_xpath);
  if ($params{get_attribute_insert_point}) {
    my @found;
    if (@found = $retval->get_xpath('xs:simpleContent/xs:extension')) {
      $retval = $found[0];
    }
    else {
      $retval = add_or_find_node($base_twig, $schema_xpath . '/xs:sequence')->parent;
    }
  }
  return $retval;
}

#----------------------------------------------------------------------

sub add_node_with_value {
  my($twig, $path, $value) = @_;
  if (ref($value)) {
      $value->paste(last_child => $twig);
  }
  else {
      add_or_find_node($twig, $path)->set_text($value);
  }
}

#----------------------------------------------------------------------

sub add_or_find_node {
  my($root, $path) = @_;

  if (my @found = $root->get_xpath($path)) {
    return $found[0];
  }
  else {
    my $new_node_xpath = basename($path);
    my ($attribute_name, $attribute_value);
    if ($new_node_xpath =~ s/\[\@(.*?)="(.*?)"\]//) {
      ($attribute_name, $attribute_value) = ($1, $2);
    }

    my $new_node = XML::Twig::Elt->new($new_node_xpath);
    if ($attribute_name) {
      $new_node->set_att($attribute_name => $attribute_value);
    }
    $new_node->paste(last_child => add_or_find_node($root, dirname($path)));
    return $new_node;
  }
}

#----------------------------------------------------------------------

sub check_schema {
  my(%params) = @_;

  my $filename = $params{filename};
  my $schema = $params{schema};
  my $error_msg = $params{error_msg};
  my $network_opt = ($params{flg_no_network}) ? '--nonet' : '';
  my $logger = $params{logger} || F5::Logger->new("/ts/log/xml_schema_validation.log");

  open(my $null1, ">", File::Spec->devnull);
  open(my $null2, File::Spec->devnull);
  my $xmllint_err_fh = gensym;
  my $xmllint_pid = open3($null1, $null2, $xmllint_err_fh,
                      $XMLLINT_CMD, '--schema', $schema, '--noout', $network_opt,
                      $filename);
  my $xmllint_err = join('', <$xmllint_err_fh>);
  waitpid($xmllint_pid, 0);
  if ($?) {
      # all error details in log
      $logger->error("$error_msg ($xmllint_err). \n caller was:".caller());

      # hide internal paths in error that is thrown to caller
      $xmllint_err =~ s/^\Q$filename\E:/line /mg;
      $xmllint_err =~ s/^\Q$filename\E fails to validate\s*\n//m;
      $xmllint_err =~ s/\Q$filename\E/ file/g;
      die "$error_msg ($xmllint_err)\n";
  }
}

#----------------------------------------------------------------------

1;

