package F5::XML::Schema::Processor;

use strict;
use warnings;

#This packiage builds a binary blob representing an xml schema that is later consumed by the xml parser.
#Mixed binary format causes a confused interal perl state with mixed utf8/non-utf8 data in the blob.
#This purifies it to binary.
use bytes;

use XML::Twig;
use Data::Dumper;
use Clone;
use File::Basename;
use Cwd;
use URI;
use Math::BigInt;
use Digest::MD5;
use Scalar::Util qw(weaken);
use F5::Deploy;
use F5::XML::Schema::SchemaDefinitions;

use constant COMPLEXTYPE_RECURSION_DEPTH => 3;


use constant XMLSCHEMAURL            => 'http://www.w3.org/2001/XMLSchema';
use constant XMLBASEURL              => 'http://www.w3.org/XML/1998/namespace';
use constant XLINKURL                => 'http://www.w3.org/1999/xlink';
use constant XSDPREFIX               => 'xsd';
use constant XSDELEMENTFORMDEFAULT   => 'elementFormDefault';
use constant XSDATTRIBUTEFORMDEFAULT => 'attributeFormDefault';
use constant XSDREQUIRED             => 'required';
use constant XSDOPTIONAL             => 'optional';
use constant XSDPROHIBITED           => 'prohibited';
use constant XSDQUALIFIED            => 'qualified';
use constant XSDUNQUALIFIED          => 'unqualified';
use constant XSDFORMILLEGAL          => 'illegal form';
use constant XSDMINOCCURS            => 'minOccurs';
use constant XSDMAXOCCURS            => 'maxOccurs';
use constant XSDWHITESPACE           => 'whiteSpace';
use constant XSDLENGTH               => 'length';
use constant XSDENUMERATION          => 'enumeration';
use constant XSDTARGETNAMESPACE      => 'targetNamespace';
use constant XSDXMLNS                => 'xmlns';
use constant XSDDEFAULT              => 'default';
use constant XSDFIXED                => 'fixed';
use constant XSDPROCESSCONTENTS      => 'processContents';

use constant XSD__STRICT             => 'strict';
use constant XSD__LAX                => 'lax';
use constant XSD__SKIP               => 'skip';

use constant XSDTOKEN                => 'token';
use constant XSDSTRING               => 'string';

use constant UNBOUNDED2LONG => '-1';
use constant UNBOUNDED => 'unbounded';

our $AUTOLOAD; ## to allow fallback autoloading of subs

my $DUMMY_XML_LOCATION = F5::Deploy::map_path('/var/ts/xml-processor/dummy.xml');

my $twig_params = {
	map_xmlns => {
		'http://www.w3.org/2001/XMLSchema' => XSDPREFIX,
	},
	keep_original_prefix => 1,
	twig_handlers => {
		'xsd:annotation' => sub { $_->delete },
	},
	comments => 'drop',

};

my $xsdl_vocabulary = {
	all => 1,
	annotation => 1,
	any => 1,
	anyAttribute => 1,
	attribute => 1,
	attributeGroup => 1,
	choice => 1,
	complexContent => 1,
	complexType => 1,
	element => 1,
	enumeration => 1,
	extension => 1,
	fractionDigits => 1,
	group => 1,
	'import' => 1,
	include => 1,
	length => 1,
	list => 1,
	maxExclusive => 1,
	maxInclusive => 1,
	maxLength => 1,
	minExclusive => 1,
	minInclusive => 1,
	minLength => 1,
	pattern => 1,
	redefine => 1,
	restriction => 1,
	schema => 1,
	sequence => 1,
	simpleContent => 1,
	simpleType => 1,
	totalDigits => 1,
	union => 1,
	whiteSpace => 1,
	%{$F5::XML::Schema::SchemaDefinitions::builtin_simple_types},
};

sub handle_all_children_have_minOccurs_0 {
	my ($xsd) = @_;
	foreach my $namespace (keys %{$xsd->{element}}){
		foreach my $element_name (keys %{$xsd->{element}->{$namespace}}) {
			fix_all_children_have_minOccurs_0($xsd,$xsd->{element}->{$namespace}->{$element_name}->{id});
		}
	}
}

sub fix_all_children_have_minOccurs_0 {
	my($xsd,$id) = @_;
	if(0==$xsd->{container}->[$id]->{num_children}){
		return;
	}
	my $sum_minOccurs_of_children=0;
	foreach my $child_id (@{$xsd->{container}->[$id]->{children}}){
		fix_all_children_have_minOccurs_0($xsd,$child_id);
		$sum_minOccurs_of_children+=$xsd->{container}->[$child_id]->{min_self_count};
	}
	if (
		0==$sum_minOccurs_of_children and 
		0<$xsd->{container}->[$id]->{min_self_count} and
		## TODO: think why it is a good idea to stop propagating
		## zeroes upwards when reaching element
		($xsd->{container}->[$id]->{container_type} ne 'element') 
	){
#		warn "Container with nonzero minOccurs updated to zero as all its children have a zero minOccurs\n";
		$xsd->{container}->[$id]->{min_self_count}=0;
	}
}

sub set_min_max_self_count {
	my ($twig, $xsd, $id)=@_;
	$xsd->{container}->[$id]->{min_self_count}=defined $twig->att(XSDMINOCCURS)?$twig->att(XSDMINOCCURS):1;
	$xsd->{container}->[$id]->{max_self_count}=defined $twig->att(XSDMAXOCCURS)?$twig->att(XSDMAXOCCURS):1;
	if ($xsd->{container}->[$id]->{max_self_count} eq UNBOUNDED) {
		$xsd->{container}->[$id]->{max_self_count}=UNBOUNDED2LONG;
	} else {
		check_min_is_le_max(
			\$xsd->{container}->[$id]->{min_self_count},
			\$xsd->{container}->[$id]->{max_self_count}
		);
	}
}

################################################################################
# process_element_child - parse single <xsd:element ...>
# input: $twig - twig which childrens attribute requested to parse
#        $xsd  - main data structure of processed schema
#        $parser mode - by default 'strict'
#        $parent_id - id of parent element to which assigned attributes
# return: element container id
################################################################################
sub process_element {
	my ($das, $xsd, $twig, $parser_mode, $parent_id)=@_;

	return unless (is_element($twig));
	
	my $element_name='';
	my $form=XSDFORMILLEGAL;
	my $explicit_form;
	my $mixed = 'false';
	my $nillable = 'false';
	my $default;
	my $simpleType;
	my $ns_info = get_namespace_information($twig);
	my $namespace_uri = $ns_info->{targetNamespace};
	my $ns_uri_id = $das->{namespace_uris2id}->{$namespace_uri};
	my $container_type='element';
	my $id = $xsd->{next_container_i};
	
	if(!$twig->parent or $twig->parent->local_name eq 'schema') {
		$id = $parent_id;
	}
	elsif($xsd->{container}->[$xsd->{next_container_i}]) {
		$id++;
	}
	
	$nillable = $twig->att('nillable') if ($twig->att('nillable'));

	my $element_type_info = get_element_type_info($das, $xsd, $twig, $id);
	if($element_type_info->{type} eq "anyType" and not $twig->att('ref')) {
		my $anyType_twig= parse XML::Twig::Elt(
			'<xsd:complexType mixed="true">
				<xsd:choice minOccurs="0" maxOccurs="unbounded">
					<xsd:any processContents="lax" namespace="##any"/>
				</xsd:choice>
				<xsd:anyAttribute namespace="##any" processContents="lax"/>
			</xsd:complexType>');
		$anyType_twig->paste($twig);
		$twig->del_att('type') if $twig->{att}->{'type'};
		my $element_id = process_element($das, $xsd, $twig, $parser_mode, $parent_id);
		$xsd->{container}->[$element_id]->{anyType} = 1;
		return $element_id;
	}
	my $metatype_id_list = [];
	if(defined $element_type_info->{simpleType}) {
		$simpleType = $element_type_info->{simpleType};
		$metatype_id_list = Clone::clone($element_type_info->{simpleType}->{metatype_id_list});
	}
	
	# default works only if no nillable appears
	if(defined $simpleType and $twig->att('default')) {
		$simpleType = {
			builtin_type     => 'union',
			name             => $simpleType->{name}."\@default",
			namespace_uri    => "\@local[$id]",
			metatype_id_list => process_default_value_attribute($das, $metatype_id_list),
			parent           => $simpleType->{name},
			parent_namespace => $simpleType->{namespace_uri},
		};
		$xsd->{simpleType}->{$simpleType->{namespace_uri}}->{$simpleType->{name}} = $simpleType;
		$metatype_id_list = Clone::clone($simpleType->{metatype_id_list});
		$default = 1;
	}
	
	if(defined $simpleType and $twig->att('fixed')) {
		$metatype_id_list = process_fixed_value_attribute($das, $xsd, $metatype_id_list, $twig->att('fixed'));
		$simpleType = {
			builtin_type     => 'union',
			name             => $simpleType->{name}."\@fixed",
			namespace_uri    => "\@local[$id]",
			metatype_id_list => $metatype_id_list,
			parent           => $simpleType->{name},
			parent_namespace => $simpleType->{namespace_uri},
		};
		$xsd->{simpleType}->{$simpleType->{namespace_uri}}->{$simpleType->{name}} = $simpleType;
		$default = 1;
	}

	my $element_parent_id =
		$xsd->{container}->[$parent_id]->{container_type} eq 'element' ?
			$parent_id :
			$xsd->{container}->[$parent_id]->{element_parent_id}
	;
	if ($twig->att('name')) {
		$element_name = $twig->att('name');
	} elsif ($twig->att('ref')) {
		my $type = extract_type_info($xsd, $das, $twig, 'ref');
		die "Cannot find ref element ",$twig->att('ref'),"\n"
			unless exists $xsd->{element}->{$type->{namespace}}->{$type->{local_name}};
		$element_name   = $type->{local_name};
		$container_type = 'ref';
		$namespace_uri  = $type->{namespace};
		$ns_uri_id      = $das->{namespace_uris2id}->{$namespace_uri};
	} else {
		die "no name for element definition\n";
	}
	if ($twig->att('form')) {
		$explicit_form=$form=$twig->att('form');
	} else {
		$form=$ns_info->{+XSDELEMENTFORMDEFAULT};
	}

	# if the element is a ref, use its form property
	if ($twig->att('ref')) {
	    my $info = extract_type_info($xsd, $das, $twig, 'ref');
	    my $elem_twig = ${$xsd->{element}->{$info->{namespace}}->{$info->{local_name}}->{twig}};
	    if ($elem_twig->root->att('targetNamespace')) {
		$form=XSDQUALIFIED;
	    } else {
		$form=XSDUNQUALIFIED;
	    }
	}
	
	# skip processing of global level element
	if(!$twig->parent or $twig->parent->local_name eq 'schema') {
		$xsd->{container}->[$parent_id]->{element_type_info} = $element_type_info;
		if(defined $simpleType) {
			$xsd->{container}->[$parent_id]->{simpleType}->{name}          = $simpleType->{name};
			$xsd->{container}->[$parent_id]->{simpleType}->{namespace_uri} = $simpleType->{namespace_uri};
		}
		set_min_max_self_count($twig, $xsd, $parent_id);
		$xsd->{container}->[$parent_id]->{default} = $default if($default);
		return $parent_id; 
	}


	# create container for element
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	my $id_in_parent = $xsd->{container}->[$parent_id]->{num_children};
	$xsd->{container}->[$parent_id]->{num_children}++;
	push @{$xsd->{container}->[$parent_id]->{children}}, $xsd->{next_container_i};

	my $new_container = {
		id                => $xsd->{next_container_i},
		name              => $element_name,
		container_type    => $container_type,
		parent_id         => $parent_id,
		element_parent_id => $element_parent_id,
		id_in_parent      => $id_in_parent,
		num_children      => 0,
		children          => [],
		num_attributes    => 0,
		attributes        => [],
		max_self_count    => 1,
		min_self_count    => 1,
		parser_mode       => $parser_mode,
		explicit_form     => $explicit_form,
		form              => $form,
		mixed             => $mixed,
		nillable          => $nillable,
		namespace_uri     => $namespace_uri,
		ns_uri_id         => $ns_uri_id,
		element_type_info => $element_type_info,
	};
	if(defined $simpleType) {
		$new_container->{simpleType}->{name}          = $simpleType->{name};
		$new_container->{simpleType}->{namespace_uri} = $simpleType->{namespace_uri};
	}
	$new_container->{default} = $default if($default);
	
	##$xsd->{container}->[$xsd->{next_container_i}] = $new_container;
	push @{$xsd->{container}}, $new_container;
	set_min_max_self_count($twig, $xsd, $xsd->{next_container_i});
	
	return $xsd->{next_container_i};
}

sub get_substitution_group_container {
	my ($das, $xsd, $twig, $parent_id, $element_parent_id) = @_;
	my $target_namespace = get_target_namespace($twig);
	
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	my $id_in_parent = $xsd->{container}->[$parent_id]->{num_children};
	$xsd->{container}->[$parent_id]->{num_children}++;
	push @{$xsd->{container}->[$parent_id]->{children}}, $xsd->{next_container_i};
	
	my $name = $twig->att('ref') if($twig->att('ref'));
	$name = $twig->att('name') if($twig->att('name'));
	my $new_container = {
		id                => $xsd->{next_container_i},
		name              => 'substitution_of['.($name).']',
		container_type    => 'choice',
		parent_id         => $parent_id,
		element_parent_id => $element_parent_id,
		id_in_parent      => $id_in_parent,
		num_children      => 0,
		children          => [],
		num_attributes    => 0,
		attributes        => [],
		max_self_count    => 1,
		min_self_count    => 1,
		parser_mode       => XSD__STRICT,
		explicit_form     => undef,
		form              => XSDFORMILLEGAL,
		namespace_uri     => $target_namespace,
		ns_uri_id         => $das->{namespace_uris2id}->{$target_namespace},
	};
	push @{$xsd->{container}}, $new_container;
	set_min_max_self_count($twig, $xsd, $xsd->{next_container_i});
	
	return $xsd->{next_container_i};

}

sub add_element_ref_container {
	my ($das, $xsd, $twig, $element_twig, $namespace, $parser_mode, $parent_id, $element_parent_id) = @_;
	my $form=XSDFORMILLEGAL;
	my $explicit_form;
	my $ns_info = get_namespace_information($twig);
	
	if($xsd->{substitutionGroup}->{$namespace}->{$element_twig->att('name')}) {
		my $substitution_group_container =
			get_substitution_group_container($das, $xsd, $element_twig, $parent_id, $element_parent_id);
			foreach my $element_name (keys %{$xsd->{substitutionGroup}->{$namespace}->{$element_twig->att('name')}}) {
				my $element_ns = $xsd->{substitutionGroup}->{$namespace}->{$element_twig->att('name')}->{$element_name};
				my $sub_element_twig = ${$xsd->{element}->{$element_ns}->{$element_name}->{twig}};
				next if($sub_element_twig->att('abstract') and $sub_element_twig->att('abstract') eq 'true');
				add_element_ref_container($das, $xsd, $element_twig, $sub_element_twig, $element_ns, $parser_mode, $substitution_group_container, $element_parent_id); 
			}
	}

	
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	my $id_in_parent = $xsd->{container}->[$parent_id]->{num_children};
	$xsd->{container}->[$parent_id]->{num_children}++;
	push @{$xsd->{container}->[$parent_id]->{children}}, $xsd->{next_container_i};
	
	if ($namespace eq '') {
		$form = $ns_info->{+XSDELEMENTFORMDEFAULT};
	} else {
		$form = XSDQUALIFIED;
	}
	
	my $new_container = {
		id                => $xsd->{next_container_i},
		name              => $element_twig->att('name'),
		container_type    => 'ref',
		parent_id         => $parent_id,
		element_parent_id => $element_parent_id,
		id_in_parent      => $id_in_parent,
		num_children      => 0,
		children          => [],
		num_attributes    => 0,
		attributes        => [],
		max_self_count    => 1,
		min_self_count    => 1,
		parser_mode       => $parser_mode,
		explicit_form     => $form,
		form              => $form,
		namespace_uri     => $namespace,
		ns_uri_id         => $das->{namespace_uris2id}->{$namespace},
	};
	push @{$xsd->{container}}, $new_container;
	
	return $xsd->{next_container_i};
}


sub add_element_any_ref_container {
	my ($das, $xsd, $twig, $namespace, $parent_id, $element_parent_id) = @_;
	my $namespace_id;
	my $form = XSDFORMILLEGAL;
	my $ns_info = get_namespace_information($twig);
	
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	my $id_in_parent = $xsd->{container}->[$parent_id]->{num_children};
	$xsd->{container}->[$parent_id]->{num_children}++;
	push @{$xsd->{container}->[$parent_id]->{children}}, $xsd->{next_container_i};
	
	my $ns_wildcard = "";
	my $ns;
	if($namespace =~ /(##other)\{(.*)\}/) {
		($ns_wildcard, $ns) = ($1, $2);
	}
	elsif($namespace =~ /(##any)/) {
		($ns_wildcard, $ns) = ($1, "");
	}
	elsif($namespace =~ /(##local)/) {
		($ns_wildcard, $ns) = ($1, "");
	}
	else {
		$ns = $namespace;
	}

	
	if ($ns eq '') {
		$form = $ns_info->{+XSDELEMENTFORMDEFAULT};
	} else {
		$form = XSDQUALIFIED;
	}
	$namespace_id = add_namespace_unless_in_cache_and_get_namespace_id($das, $ns);
	my $new_container = {
		id                => $xsd->{next_container_i},
		name              => 'any_ref['.$namespace.']',
		container_type    => 'any_ref',
		parent_id         => $parent_id,
		element_parent_id => $element_parent_id,
		id_in_parent      => $id_in_parent,
		num_children      => 0,
		children          => [],
		num_attributes    => 0,
		attributes        => [],
		max_self_count    => 1,
		min_self_count    => 1,
		parser_mode       => XSD__STRICT,
		explicit_form     => $form,
		form              => $form,
		ns_wildcard       => $ns_wildcard,
		namespace_uri     => $ns,
		ns_uri_id         => $namespace_id,
	};
	push @{$xsd->{container}}, $new_container;
	
	return $xsd->{next_container_i};
}

sub resolve_namespace_in_attribute_wildcard_entity {
	my ($xsd, $element_id, $namespace) = @_;
	
	my $any_attribute = $xsd->{container}->[$element_id]->{any_attribute};
	return 0 if (not $any_attribute);
	
	my $anyatt_namespace = $any_attribute->{namespace};
	
	return 1 if (exists $anyatt_namespace->{any});
	
	return 1 if (exists $anyatt_namespace->{list}->{$namespace});
	
	if (exists $anyatt_namespace->{other}) {
		if($anyatt_namespace->{other}->{count} and
		   $anyatt_namespace->{other}->{namespace} ne $namespace and
		   $namespace ne '') {
			return 1;
		}
	}
	
	return 1 if (exists $anyatt_namespace->{local} and $namespace eq '');
	
	return 0;
}
	
sub process_any_element_entity {
	my ($das, $xsd, $twig, $parser_mode, $element_parent_id, $id)=@_;
	my $any_namespaces;
	my $ns_info;
	my $namespace_uri;
	my $parent_id = $id;

	# create choice container that will contain all 'any' substructure
	$ns_info = get_namespace_information($twig);
	$namespace_uri=$ns_info->{targetNamespace};
	$any_namespaces = get_wildcard_namespaces_list($twig, $ns_info);
	
	my $any_parser_mode = XSD__STRICT;
	$any_parser_mode = $twig->att(XSDPROCESSCONTENTS) if $twig->att(XSDPROCESSCONTENTS);
	
	add_choice_container($das, $xsd, $twig, $parser_mode, $element_parent_id, $parent_id);
	$parent_id = $xsd->{next_container_i};

	# add all strict elements containers that match namespace wildcard
	foreach my $namespace (@{$any_namespaces}) {
		add_namespace_unless_in_cache_and_get_namespace_id($das, $namespace);
		if($any_parser_mode eq XSD__STRICT or $any_parser_mode eq XSD__LAX) {
			add_strict_element_refs($das, $xsd, $twig, $namespace, $element_parent_id, $parent_id);
		}
	}
		
	# for each namespace add 'any' container with 'skip' processContents
	foreach my $namespace (@{$any_namespaces}) {
		add_namespace_unless_in_cache_and_get_namespace_id($das, $namespace);
		if($any_parser_mode eq XSD__SKIP or $any_parser_mode eq XSD__LAX) {
			add_any_container($das, $xsd, $twig, $any_parser_mode, $namespace, $element_parent_id, $parent_id);
		}
	}
}

sub add_choice_container {
	my ($das, $xsd, $twig, $parser_mode, $element_parent_id, $id)=@_;
	
	my $ns_info = get_namespace_information($twig);
	my $namespace_uri=$ns_info->{targetNamespace};
	my $id_in_parent = $xsd->{container}->[$id]->{num_children};

	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	$xsd->{container}->[$id]->{num_children}++;
	push @{$xsd->{container}->[$id]->{children}}, $xsd->{next_container_i};

	# create virtual model group - sequence
	my $new_container = {
		id                => $xsd->{next_container_i},
		name              => '[choice of any]',
		container_type    => 'choice',
		parent_id         => $id, 
		element_parent_id => $element_parent_id,
		id_in_parent      => $id_in_parent,
		num_children      => 0,
		children          => [],
		num_attributes    => 0,
		attributes        => [],
		max_self_count    => 1,
		min_self_count    => 1,
		parser_mode       => $parser_mode,
		explicit_form     => undef,
		form              => XSDFORMILLEGAL,
		namespace_uri     => $namespace_uri,
		ns_uri_id         => $das->{namespace_uris2id}->{$namespace_uri},
	};
	
	$xsd->{container}->[$xsd->{next_container_i}] = $new_container;
	set_min_max_self_count($twig, $xsd, $xsd->{next_container_i});
}

sub add_any_container {
	my ($das, $xsd, $twig, $parser_mode, $namespace, $element_parent_id, $id)=@_;
	
	my $namespace_id;
	my $schema_h = $F5::XML::Schema::SchemaDefinitions::schema_h;
	my $ns_info = get_namespace_information($twig);
	my $id_in_parent = $xsd->{container}->[$id]->{num_children};
	
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	$xsd->{container}->[$id]->{num_children}++;
	push @{$xsd->{container}->[$id]->{children}}, $xsd->{next_container_i};
	
	my $ns_wildcard = "";
	my $ns;
	if($namespace =~ /(##other)\{(.*)\}/) {
		($ns_wildcard, $ns) = ($1, $2);
	}
	elsif($namespace =~ /(##any)/) {
		($ns_wildcard, $ns) = ($1, "");
	}
	elsif($namespace =~ /(##local)/) {
		($ns_wildcard, $ns) = ($1, "");
	}
	else {
		$ns = $namespace;
	}

	$namespace_id = add_namespace_unless_in_cache_and_get_namespace_id($das, $ns);

	my $form=XSDFORMILLEGAL;
	my $explicit_form;
	if ($twig->att('form')) {
		$explicit_form=$form=$twig->att('form');
	} else {
		$form=$ns_info->{+XSDELEMENTFORMDEFAULT};
	}
	$form=XSDQUALIFIED if $ns_info->{targetNamespace} ne '';
	
	my $new_container = {
		id                => $xsd->{next_container_i},
		name              => 'any['.$namespace.']',
		container_type    => 'any',
		parent_id         => $id,
		element_parent_id => $element_parent_id,
		id_in_parent      => $id_in_parent,
		num_children      => 0,
		children          => [],
		num_attributes    => 0,
		attributes        => [],
		max_self_count    => 1,
		min_self_count    => 1,
		parser_mode       => ($parser_mode eq XSD__STRICT) ? $parser_mode : XSD__SKIP,
		explicit_form     => $explicit_form,
		form              => $form,
		mixed             => 'false',
		ns_wildcard       => $ns_wildcard,
		namespace_uri     => $ns,
		ns_uri_id         => $namespace_id,
	};
	
	$xsd->{container}->[$xsd->{next_container_i}] = $new_container;
}

sub add_strict_element_refs {
	my ($das, $xsd, $twig, $namespace, $element_parent_id, $id) = @_;
	
	add_element_any_ref_container($das, $xsd, $twig, $namespace, $id, $element_parent_id);
	$xsd->{container}->[$id]->{num_children} =
		scalar @{$xsd->{container}->[$id]->{children}};
}

sub process_any_attribute_entity {
	my ($das, $xsd, $twig, $element_parent_id, $id)=@_;

	my $any_att_process_contents;
	my $any_att_namespaces_list;
	
	my $ns_info = get_namespace_information($twig);

	if(exists $xsd->{container}->[$element_parent_id]->{any_attribute}) {
		$any_att_namespaces_list = $xsd->{container}->[$element_parent_id]->{any_attribute}->{namespace};
	}
	$any_att_process_contents = XSD__STRICT; # default value
	$any_att_process_contents = $twig->att(+XSDPROCESSCONTENTS) if $twig->att(+XSDPROCESSCONTENTS);	
	
	foreach my $namespace (@{get_wildcard_namespaces_list($twig, $ns_info)}) {
		if($namespace =~ /(##other)\{(.*)\}/) {
			if(exists $any_att_namespaces_list->{other} and
			   $any_att_namespaces_list->{other}->{namespace} ne $2) {
				$any_att_namespaces_list->{other}->{count} = 0;
				$any_att_namespaces_list->{any}++;
			}
			else {
				$any_att_namespaces_list->{other}->{count}++;
				$any_att_namespaces_list->{other}->{namespace} = $2;
			}
		}
		elsif($namespace =~ /(##any)/) {
			$any_att_namespaces_list->{any}++;
		}
		elsif($namespace =~ /(##local)/) {
			$any_att_namespaces_list->{local}++;
		}
		else {
			$any_att_namespaces_list->{list}->{$namespace}++;
		}
	}
	
	$xsd->{container}->[$element_parent_id]->{any_attribute}->{namespace} = $any_att_namespaces_list;
	$xsd->{container}->[$element_parent_id]->{any_attribute}->{process_contents} = $any_att_process_contents;

}

sub process_any_attribute_for_element {
	my ($das, $xsd, $twig, $element_parent_id)=@_;
	my $ns_info = get_namespace_information($twig);

	return if(not exists $xsd->{container}->[$element_parent_id]->{any_attribute});
	
	my $parser_mode = $xsd->{container}->[$element_parent_id]->{any_attribute}->{process_contents};

	# if processContent is 'strict' or 'lax' add all known named attributes
	# which resign in correct namespace (from 'namespace' attribute) 
	if($parser_mode eq XSD__STRICT or $parser_mode eq XSD__LAX) {
		add_strict_attributes_containers($das, $xsd, $twig,  $element_parent_id, $ns_info, $parser_mode);
	}
	
	# if processContent is 'lax' or 'skip' add spacial 'anyAttribute' container
	# which have 'skip' parser_mode and namespace which we want to skip
	# also namespace field may have 2 'special' values
	if($parser_mode eq XSD__LAX or $parser_mode eq XSD__SKIP) {
		add_skip_any_attribute_containers($das, $xsd, $twig, $element_parent_id, $ns_info, $parser_mode);
	}
	
	
}

sub add_strict_attributes_containers {
	my ($das, $xsd, $twig,  $id, $ns_info, $parser_mode) = @_;

	my $form;
	my $explicit_form;
	
	my $id_in_parent    = $xsd->{container}->[$id]->{num_attributes};
	my $anyAtt_ns       = $xsd->{container}->[$id]->{any_attribute}->{namespace};
	my $process_Content = $xsd->{container}->[$id]->{any_attribute}->{process_contents};
	
	foreach my $namespace (keys %{$xsd->{attribute}}) {
		next unless resolve_namespace_in_attribute_wildcard_entity($xsd, $id, $namespace);
		foreach my $global_attribute_name (keys %{$xsd->{attribute}->{$namespace}}) {
			# if no twig exists then the attribute is irrelevant
			next unless exists $xsd->{attribute}->{$namespace}->{$global_attribute_name}->{twig};

			if($xsd->{attribute}->{$namespace}->{$global_attribute_name}->{id}) {
				push @{$xsd->{container}->[$id]->{attributes}}, $xsd->{attribute}->{$namespace}->{$global_attribute_name}->{id};
				next;
			}
			
			# create new attribute as copy of global attribute
			$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
			my $att_twig = ${$xsd->{attribute}->{$namespace}->{$global_attribute_name}->{twig}};
			
			# detect attribute form
			my $att_ns_info = get_namespace_information($att_twig);
			if($namespace ne "" and $namespace ne $ns_info->{+XSDTARGETNAMESPACE}) {
				$form = $explicit_form = XSDQUALIFIED;
			}
			elsif( $ns_info->{+XSDTARGETNAMESPACE} eq "" or
					exists $anyAtt_ns->{local}) {
				$form = $att_ns_info->{+XSDATTRIBUTEFORMDEFAULT};
				$explicit_form = $form;
			}
			else {
				$form = $explicit_form = XSDQUALIFIED;
			}
			$att_twig = fake_twig_for_special_atts($das,$xsd,$att_twig);
			my $attribute_type = get_attribute_simple_type($das, $xsd, $att_twig, $id);
			my $attribute_cont = {
				id                => $xsd->{next_container_i},
				name              => $global_attribute_name,
				container_type    => 'attribute',
				min_self_count    => 0,
				max_self_count    => 1,
				parent_id         => $id,
				element_parent_id => $id,
				explicit_form     => $explicit_form,
				form              => $form,
				parser_mode       => XSD__STRICT,
				id_in_parent      => $id_in_parent,
				namespace_uri     => $namespace,
				ns_uri_id         => $das->{namespace_uris2id}->{$namespace},
			};
			++$id_in_parent;
			$attribute_cont->{simpleType}->{name} = $attribute_type->{name};
			$attribute_cont->{simpleType}->{namespace_uri} = $attribute_type->{namespace_uri};
			# push it into parent container
			push @{$xsd->{container}}, $attribute_cont;
			push @{$xsd->{container}->[$id]->{attributes}}, $xsd->{next_container_i};
			$xsd->{attribute}->{$namespace}->{$global_attribute_name}->{id} = $xsd->{next_container_i};
		}
	}
	
	$xsd->{container}->[$id]->{num_attributes} =
		scalar @{$xsd->{container}->[$id]->{attributes} || []};
}

sub add_skip_any_attribute_containers {
	my ($das, $xsd, $twig, $id, $ns_info, $parser_mode) = @_;
	
	my $schema_h = $F5::XML::Schema::SchemaDefinitions::schema_h;
	my $form = $ns_info->{+XSDATTRIBUTEFORMDEFAULT};
	my $id_in_parent = $xsd->{container}->[$id]->{num_attributes};

	my $anyAtt_ns   = $xsd->{container}->[$id]->{any_attribute}->{namespace};
	
	my $ns_wildcard = "";
	my $ns;
	foreach my $namespace (@{make_wildcard_namespaces_list($anyAtt_ns)}) {
		$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
		if($namespace =~ /(##other)\{(.*)\}/) {
			($ns_wildcard, $ns) = ($1, $2);
		}
		elsif($namespace =~ /(##any)/) {
			($ns_wildcard, $ns) = ($1, "");
		}
		elsif($namespace =~ /(##local)/) {
			($ns_wildcard, $ns) = ($1, "");
		}
		else {
			$ns = $namespace;
			$ns_wildcard = "";
		}

		my $namespace_id = add_namespace_unless_in_cache_and_get_namespace_id($das, $ns);
		
		my $any_attribute_cont = {
			id                => $xsd->{next_container_i},
			name              => 'anyAttribute['.$namespace.']',
			container_type    => 'anyAttribute',
			parent_id         => $id,
			element_parent_id => $id,
			id_in_parent      => $id_in_parent,
			num_children      => 0,
			children          => [],
			num_attributes    => 0,
			attributes        => [],
			max_self_count    => -1,
			min_self_count    => 0,
			parser_mode       => XSD__SKIP,
			explicit_form     => $form,
			form              => $form,
			mixed             => 'false',
			ns_wildcard       => $ns_wildcard,
			namespace_uri     => $ns,
			ns_uri_id         => $namespace_id,
		};
		++$id_in_parent;
		# push it into parent container
		push @{$xsd->{container}}, $any_attribute_cont;
		push @{$xsd->{container}->[$id]->{attributes}}, $xsd->{next_container_i};
	}
	# end loop
	$xsd->{container}->[$id]->{num_attributes} =
		scalar @{$xsd->{container}->[$id]->{attributes}};
}

sub make_wildcard_namespaces_list {
	my ($anyAtt_ns) = @_;
	my @list;
	
	return ["##any"] if(exists $anyAtt_ns->{any});
	
	push( @list, "##local") if(exists $anyAtt_ns->{local});
	
	push( @list, "##other{".$anyAtt_ns->{other}->{namespace}."}") if(exists $anyAtt_ns->{other} and $anyAtt_ns->{other}->{count});
	
	if(exists $anyAtt_ns->{list}) {
		foreach my $ns (keys %{$anyAtt_ns->{list}}) {
			push(@list, $ns);
		}
	}
	
	return [@list];

}

sub get_wildcard_namespaces_list {
	my ($twig, $ns_info) = @_;
	my $namespace_att = $twig->att('namespace');

	# default value is ##any
	return ["##any"] if !$namespace_att;
	
	return ["##other{".$ns_info->{targetNamespace}."}"] if $namespace_att eq "##other";
	# return ["##any"] if $namespace_att eq "##any";
	# return [$ns_info->{targetNamespace}] if $namespace_att eq "##targetNamespace";
	# return ["##local"] if $namespace_att eq "##local";
	
	# at this point we have list of uri's
	my @ns_list = split ' ', $namespace_att;
	for(my $i=0; $i < scalar(@ns_list); $i++) {
		$ns_list[$i] = $ns_info->{targetNamespace} if($ns_list[$i] eq "##targetNamespace"); 
	}
	return [@ns_list]; #[split ' ', $namespace_att];
}

################################################################################
# process one of (all|choice|sequence) model groups
# create container for content model
# input: $das  - global parser information
#        $xsd  - main data structure of processed schema
#        $twig - twig of content model
#        $parser mode - by default 'strict'
#        $element_parent_id - id of parent element to which assigned attributes
#	     $id - direct parent id
################################################################################
sub process_content_model {
	my ($das, $xsd, $twig, $parser_mode, $element_parent_id, $id)=@_;
	my $ns_info = get_namespace_information($twig);
	
	my $id_in_parent = $xsd->{container}->[$id]->{num_children};
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	push @{$xsd->{container}->[$id]->{children}}, $xsd->{next_container_i};
	$xsd->{container}->[$id]->{num_children}= @{$xsd->{container}->[$id]->{children}};
	
	my $new_container;

	# create model group child
	$new_container = {
		id => $xsd->{next_container_i},
		name              => '',
		container_type    => get_local_name($twig->gi),
		parent_id         => $id, 
		element_parent_id => $element_parent_id,
		id_in_parent      => $id_in_parent,
		num_children      => 0,
		children          => [],
		max_self_count    => 1,
		min_self_count    => 1,
		parser_mode       => $parser_mode,
		explicit_form     => undef,
		form              => XSDFORMILLEGAL,
		namespace_uri     => $ns_info->{targetNamespace},
		ns_uri_id         => $das->{namespace_uris2id}->{$ns_info->{targetNamespace}},
	};
	
	# update named complex type name in container of its content model
	$new_container->{name} = $new_container->{container_type}.'['.$twig->parent->att('name').']' if $twig->parent->att('name');
	$new_container->{name} = $new_container->{container_type}.'['.$twig->parent->att('ref').']' if $twig->parent->att('ref');
	
	if($xsd->{container}->[$element_parent_id]->{element_type_info} and
	   $xsd->{container}->[$element_parent_id]->{element_type_info}->{local_name} and
	   !$new_container->{name}) {
		$new_container->{name} = $new_container->{container_type}.'['.$xsd->{container}->[$element_parent_id]->{element_type_info}->{local_name}.']';
	}	
	$xsd->{container}->[$xsd->{next_container_i}] = $new_container;
	$xsd->{container}->[$xsd->{next_container_i}]->{id} = $xsd->{next_container_i};
	set_min_max_self_count($twig, $xsd, $xsd->{next_container_i});
}

################################################################################
# process_attribute_child - parse single <xsd:attribute ...>
# input: $attribute - twig of attribute requested to parse
#        $xsd  - main data structure of processed schema
#        $das  - global parser information
#        $parser mode - by default 'strict'
#        $element_parent_id - id of parent element to which assigned attributes
#        $exist - true|false - if true return container structure
#                              (do not add to $xsd->{container} )
################################################################################
sub process_attribute_child {
	my ($attribute, $xsd, $das, $parser_mode, $element_parent_id, $exist) = @_;
	my $attribute_ns = '';
	my $ns_info = get_namespace_information($attribute);
	$attribute_ns = $ns_info->{targetNamespace} if $ns_info->{attributeFormDefault} eq XSDQUALIFIED;
	
	$attribute = fake_twig_for_special_atts($das,$xsd,$attribute);

	my $attribute_type = get_attribute_simple_type($das, $xsd, $attribute, $element_parent_id);
	#$attribute_type = process_default_value_attribute($das, $attribute_type) if($attribute->att('default'));
	if($attribute->att('fixed')) {
		
		my $fixed_attribute_type = {
			builtin_type     => 'union',
			name             => $attribute->{att}->{name}. ':'. $attribute_type->{name}."\@fixed",
			namespace_uri    => "\@local[$element_parent_id]",
			metatype_id_list => process_attribute_fixed_value_attribute($das, $xsd, $attribute_type->{metatype_id_list}, $attribute->att('fixed')),
			parent           => $attribute_type->{name},
			parent_namespace => $attribute_type->{namespace_uri},
		};
		$xsd->{simpleType}->{$fixed_attribute_type->{namespace_uri}}->{$fixed_attribute_type->{name}} = $fixed_attribute_type;
		$attribute_type = $fixed_attribute_type;
	}

	my $min=0;
	my $max=0;
	if ($attribute->att('use')) {
		if ($attribute->att('use') eq XSDREQUIRED) {
			$min=$max=1;
		} elsif ($attribute->att('use') eq XSDOPTIONAL) {
			$min=0;
			$max=1;
		} elsif ($attribute->att('use') eq XSDPROHIBITED) {
			$min=$max=0;
		}
	} else {
		$min=0;
		$max=1;
	}
	
	my $name;
	if ($attribute->att('name')) {
		$name = $attribute->att('name');
	} elsif ($attribute->att('ref')) {
		my $prefix = get_prefix($attribute->att('ref'));
		$attribute_ns  = get_namespace_of_prefix($attribute, $prefix);
		#$attribute_ns = $attribute->namespace($prefix);
		$name = get_local_name($attribute->att('ref'));
	} else {
		die "Attribute declaration must have a name or a ref\n";
	}
	
	my $form=$ns_info->{+XSDATTRIBUTEFORMDEFAULT};
	$form=XSDQUALIFIED if($attribute->att('ref') and $attribute_ns);
	my $explicit_form;
	if ($attribute->att('form')) {
		$explicit_form=$form=$attribute->att('form');
		
	}

	$attribute_ns = $ns_info->{targetNamespace} if ($form eq XSDQUALIFIED and not $attribute_ns);
	$attribute_ns = '' if $form eq XSDUNQUALIFIED;
	
	my $attribute_cont = {
		name              => $name,
		container_type    => 'attribute',
		min_self_count    => $min,
		max_self_count    => $max,
		parent_id         => $element_parent_id,
		element_parent_id => $element_parent_id,
		id                => -1,
		explicit_form     => $explicit_form,
		form              => $form,
		parser_mode       => $parser_mode,
		id_in_parent      => -1,
		children          => [],
		num_children      => 0,
		mixed             => 'false',
		namespace_uri     => $attribute_ns,
		ns_uri_id         => $das->{namespace_uris2id}->{$attribute_ns},
	};
	
	$attribute_cont->{simpleType}->{name} = $attribute_type->{name};
	$attribute_cont->{simpleType}->{namespace_uri} = $attribute_type->{namespace_uri};
	
	return $attribute_cont if($exist);
	
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	my $id_of_attribute = $xsd->{next_container_i};
	my $id_in_parent = $xsd->{container}->[$element_parent_id]->{num_attributes};
	
	$xsd->{container}->[$id_of_attribute] = $attribute_cont;
	$xsd->{container}->[$id_of_attribute]->{id} = $id_of_attribute;
	$xsd->{container}->[$id_of_attribute]->{id_in_parent} = $id_in_parent;
	
	push @{$xsd->{container}->[$element_parent_id]->{attributes}}, $id_of_attribute; # parent element maintains list of its atts
	$xsd->{container}->[$element_parent_id]->{num_attributes} = @{$xsd->{container}->[$element_parent_id]->{attributes}};
}

# $twig points to <xsd:simpleContent...> tag
# TODO: it is possible to derive one complexType->simpleContent from another complexType->simpleContent
sub prosess_simple_content {
	my ($twig, $xsd, $das,$parser_mode, $id) = @_;
	my $simpleType = undef;
	my $complexType_twig;
	my $builtin_type;
	my $metatype_id_list = [];
	my $element_parent_id = get_element_parent_id($xsd, $id);

	my $element_cont = $xsd->{container}->[$element_parent_id];	
	my $element_name = $element_cont->{name};
	my $element_ns   = $element_cont->{namespace_uri};
	
	my $child = $twig->first_child or return;
	if(!(is_extension($child) or is_restriction($child))) {
		# do not contain extension or restriction - illegal
			die "'simpleContent' must contain either 'extention' or 'restriction' child element";
	}
	my $content_type = extract_type_info($xsd, $das, $child, 'base');
	
	## process simpleContent extension/restriction base
	if($xsd->{complexType}->{$content_type->{namespace}}
	   and
	   $xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}) {
		if($element_cont->{complexType}) {
			my $complexType_name      = $element_cont->{complexType}->{name};
			my $complexType_namespace = $element_cont->{complexType}->{namespace_uri};
			$xsd->{complexType}->{$complexType_namespace}->{$complexType_name}->{base_type} = $content_type->{local_name};
			$xsd->{complexType}->{$complexType_namespace}->{$complexType_name}->{base_namespace} = $content_type->{namespace};
		}
		$complexType_twig = ${$xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}->{twig}};
	}
	
	if($complexType_twig) {
		traverse($das, $complexType_twig, $xsd, $id);
		## process simpleType of simpleContent
		if(is_restriction($child)) {
			my $old_simpleType = $xsd->{container}->[$element_parent_id]->{simpleType};
			$old_simpleType    = $xsd->{simpleType}->{$old_simpleType->{namespace_uri}}->{$old_simpleType->{name}};
			$metatype_id_list  = $old_simpleType->{metatype_id_list};
			$builtin_type      = $old_simpleType->{builtin_type};
			
			$simpleType = {
				name             => get_local_simpleType_name($das, $xsd, $element_name, $element_ns, $element_parent_id),
				builtin_type     => $builtin_type,
				namespace_uri    => "\@local",
				parent           => '', # only if simpleContent derived from simple type
				parent_namespace => '', 
				metatype_id_list => get_consolidated_metatype_id_list($das, $xsd, $child, $metatype_id_list, $builtin_type),
			};
			$xsd->{simpleType}->{$simpleType->{namespace_uri}}->{$simpleType->{name}} = $simpleType;
			
			$xsd->{container}->[$element_parent_id]->{simpleType}->{name}          = $simpleType->{name};
			$xsd->{container}->[$element_parent_id]->{simpleType}->{namespace_uri} = $simpleType->{namespace_uri};
		}
	}
	else {
		$simpleType = get_named_simple_type($das, $xsd, $twig, $content_type);
		
		if(is_extension($child)) {
			die schema_parsing_error_at($twig) .  " unknown base simple type in simpleContent\n"
				unless $simpleType;
		} elsif(is_restriction($child)) {
			if(not defined $simpleType) {
				my @simpleType_children = get_xsd_children($xsd,$child,'simpleType');
				die schema_parsing_error_at($child) . " missing simple type definition\n"
					unless @simpleType_children>0;
				my $stype_name = get_local_simpleType_name($das, $xsd, $element_name, $element_ns, $element_parent_id);
				$simpleType = get_simpleType_from_twig_simple_type($das, $xsd, $simpleType_children[0], $stype_name, "\@local{$element_ns}");
				$simpleType->{metatype_id_list} = get_consolidated_metatype_id_list($das, $xsd, $child, $simpleType->{metatype_id_list}, $simpleType->{builtin_type});
				$xsd->{simpleType}->{$simpleType->{namespace_uri}}->{$simpleType->{name}} = $simpleType;
			}
		}

		$xsd->{container}->[$element_parent_id]->{simpleType}->{name}          = $simpleType->{name};
		$xsd->{container}->[$element_parent_id]->{simpleType}->{namespace_uri} = $simpleType->{namespace_uri};
	}
	
	## process simpleContenet extension/restriction attributes
	if(is_extension($child)) {
		foreach my $item ($child->children()) {
			traverse($das, $item, $xsd, $element_parent_id)
				if(is_attribute($item) or is_attributeGroup($item) or is_anyAttribute($item));
		}
	}
	else {
		process_restriction_attributes($child, $xsd, $das, $parser_mode, $element_parent_id);
	}
}

# process attributes in complec type derivation restriction
# we need to parse new attributes and update corresponding
# attributes from base type
sub process_restriction_attributes {
	my ($twig, $xsd, $das, $parser_mode, $element_parent_id) = @_;
	foreach my $item ($twig->children()) {
		process_restriction_attribute($item, $xsd, $das, $parser_mode, $element_parent_id) if(is_attribute($item));
		process_restriction_attribute_group($item, $xsd, $das, $parser_mode, $element_parent_id) if(is_attributeGroup($item));
		process_restriction_any_attribute($das, $xsd, $item, $element_parent_id) if(is_anyAttribute($item));
	}
}

sub process_restriction_attribute {
	my ($attribute, $xsd, $das, $parser_mode, $element_parent_id) = @_;
	my $found = 0;
	my $container =	process_attribute_child($attribute, $xsd, $das, $parser_mode, $element_parent_id, "1");
	foreach my $attr_id (@{$xsd->{container}->[$element_parent_id]->{attributes}}) {
		if($xsd->{container}->[$attr_id]->{name} eq $container->{name} and
		   $xsd->{container}->[$attr_id]->{namespace_uri} eq $container->{namespace_uri} and
		   not $found) {
			# TODO copy with check of restriction
				$xsd->{container}->[$attr_id]->{min_self_count}   = $container->{min_self_count};
				$xsd->{container}->[$attr_id]->{max_self_count}   = $container->{max_self_count};
				$xsd->{container}->[$attr_id]->{explicit_form}    = $container->{explicit_form};
				$xsd->{container}->[$attr_id]->{form}             = $container->{form};
				$xsd->{container}->[$attr_id]->{parser_mode}      = $container->{parser_mode};
				$xsd->{container}->[$attr_id]->{simpleType}       = $container->{simpleType};
				$found = 1;
		}
	}
	if(not $found) {
#		warn "Illegal attribute ".$container->{name}." in derived complexType -- skipping";  
	}
}

sub process_restriction_attribute_group {
	my ($attribute_group, $xsd, $das, $parser_mode, $element_parent_id) = @_;

	my $group_type_info = extract_type_info($xsd, $das, $attribute_group, 'ref');
	my $group_namespace = $group_type_info->{namespace};
	my $group_name = $group_type_info->{local_name};
	my $attGroup = ${$xsd->{attributeGroup}->{$group_namespace}->{$group_name}->{twig}};
	process_restriction_attributes($attGroup, $xsd, $das, $parser_mode, $element_parent_id);
}

sub process_restriction_any_attribute {
	my ($das, $xsd, $twig, $element_parent_id) = @_;
	
	my $any_att_process_contents;
	my $any_att_namespaces_list;

	my $ns_info = get_namespace_information($twig);

	$any_att_process_contents = XSD__STRICT; # default value
	$any_att_process_contents = $twig->att(+XSDPROCESSCONTENTS) if $twig->att(+XSDPROCESSCONTENTS);	
	
	foreach my $namespace (@{get_wildcard_namespaces_list($twig, $ns_info)}) {
		if($namespace =~ /(##other)\{(.*)\}/) {
			$any_att_namespaces_list->{other}->{count}++;
			$any_att_namespaces_list->{other}->{namespace} = $2;
		}
		elsif($namespace =~ /(##any)/) {
			$any_att_namespaces_list->{any}++;
		}
		elsif($namespace =~ /(##local)/) {
			$any_att_namespaces_list->{local}++;
		}
		else {
			$any_att_namespaces_list->{list}->{$namespace}++;
		}
	}
	
	$xsd->{container}->[$element_parent_id]->{any_attribute}->{namespace}        = $any_att_namespaces_list;
	$xsd->{container}->[$element_parent_id]->{any_attribute}->{process_contents} = $any_att_process_contents;
}

sub process_named_complexType {
	my ($das, $xsd, $complexType, $namespace) = @_;
	my $complexType_twig;
	
	$complexType_twig = ${$xsd->{complexType}->{$namespace}->{$complexType}->{twig}};
	return if not $complexType_twig;
	
	# create complex type container
	my $new_container = {
		id                => -1,
		name              => $complexType,
		container_type    => 'complex_type',
		min_self_count    => 1,
		max_self_count    => 1,
		parent_id         => -1,
		element_parent_id => -1,
		explicit_form     => XSDFORMILLEGAL,
		form              => XSDFORMILLEGAL,
		parser_mode       => XSD__STRICT,
		id_in_parent      => 0,
		num_children      => 0,
		children          => [],
		num_attributes    => 0,
		attributes        => [],
		mixed             => 'false',
		namespace_uri     => $namespace,
		ns_uri_id         => $das->{namespace_uris2id}->{$namespace},
		complexType       => {
			name          => $complexType,
			namespace_uri => $namespace,
		},
	};
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	$xsd->{container}->[$xsd->{next_container_i}] = $new_container;
	$xsd->{container}->[$xsd->{next_container_i}]->{id} = $xsd->{next_container_i};
	
	$xsd->{complexType}->{$namespace}->{$complexType}->{id} = $xsd->{next_container_i};
	
	if(exists $xsd->{complexType}->{$namespace}->{$complexType}->{content_id}) {
		$xsd->{container}->[$xsd->{next_container_i}]->{complexType_id} = $xsd->{complexType}->{$namespace}->{$complexType}->{content_id};
	}
	else {
		# complexType not used explicitly in schema, but can be used with xsi:type
		# parse complexType twig as content model of complexType container 
		$xsd->{complexType}->{$namespace}->{$complexType}->{content_id} = $xsd->{next_container_i};
		$xsd->{container}->[$xsd->{next_container_i}]->{complexType_id} = $xsd->{next_container_i};
		traverse($das, $complexType_twig, $xsd, $xsd->{next_container_i});
		# process anyAttribute attributes...
		process_any_attribute_for_element($das, $xsd, $complexType_twig, $xsd->{complexType}->{$namespace}->{$complexType}->{content_id});
	}

	
}

sub process_named_complexType_derivation {
	my ($das, $xsd, $complexType, $namespace) = @_;
	
	my $bname      = $xsd->{complexType}->{$namespace}->{$complexType}->{base_type};
	my $bnamespace = $xsd->{complexType}->{$namespace}->{$complexType}->{base_namespace};
	
	if($bname) {
		$xsd->{complexType}->{$namespace}->{$complexType}->{base_id} = $xsd->{complexType}->{$bnamespace}->{$bname}->{id};
	}
}

sub prosess_complex_content {
	my ($twig, $xsd, $das,$parser_mode, $id) = @_;
	my $complexType_twig;
	my $element_parent_id = get_element_parent_id($xsd, $id);

	my $element_cont = $xsd->{container}->[$element_parent_id];	
	my $child = $twig->first_child or return;
	my $ns_info = get_namespace_information($twig);
	
	my $content_type = extract_type_info($xsd, $das, $child, 'base');
	if($content_type
	   and
	   $xsd->{complexType}->{$content_type->{namespace}}
	   and
	   $xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}) {
		if($element_cont->{complexType}) {
			my $complexType_name      = $element_cont->{complexType}->{name};
			my $complexType_namespace = $element_cont->{complexType}->{namespace_uri};
                        if(not $xsd->{complexType}->{$complexType_namespace}->{$complexType_name}->{base_type}) {
                           $xsd->{complexType}->{$complexType_namespace}->{$complexType_name}->{base_type} = $content_type->{local_name};
                           $xsd->{complexType}->{$complexType_namespace}->{$complexType_name}->{base_namespace} = $content_type->{namespace};
                        }
		}
	    $complexType_twig = ${$xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}->{twig}};
		if($xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}->{recursion_count}
	       and
		   $xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}->{recursion_count} > 0 ) {
            return;
		}
		else {
			$xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}->{recursion_count}++;
		}
	}
	
	if($content_type->{local_name} eq 'anyType' and $content_type->{namespace} eq +XMLSCHEMAURL) {
		# base type is anytype
		$complexType_twig= parse XML::Twig::Elt(
		'<xsd:complexType mixed="true">
			<xsd:choice minOccurs="0" maxOccurs="unbounded">
				<xsd:any processContents="lax" namespace="##any"/>
			</xsd:choice>
			<xsd:anyAttribute namespace="##any" processContents="lax"/>
		</xsd:complexType>');
		$complexType_twig->paste('last_child' => ($twig->root));
	}

	if(is_extension($child)){
		my $new_container = undef;
		my $current_att_num = $xsd->{container}->[$element_parent_id]->{num_attributes};
		# in extension we place base type and inline in sequence
		# if one of 'base type' or 'inline type' have empty content then skip sequence
		if(!is_empty_content_complexType($child) and !is_empty_content_complexType($complexType_twig)) {
			$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
			push @{$xsd->{container}->[$id]->{children}}, $xsd->{next_container_i};
			$xsd->{container}->[$id]->{num_children}= @{$xsd->{container}->[$id]->{children}};
			
			# create virtual model group - sequence
			$new_container = {
				id                => $xsd->{next_container_i},
				name              => '',
				container_type    => 'sequence',
				parent_id         => $id, 
				element_parent_id => $element_parent_id,
				id_in_parent      => 0,
				num_children      => 0,
				children          => [],
				max_self_count    => 1,
				min_self_count    => 1,
				parser_mode       => $parser_mode,
				explicit_form     => undef,
				form              => XSDFORMILLEGAL,
				namespace_uri     => $ns_info->{targetNamespace},
				ns_uri_id         => $das->{namespace_uris2id}->{$ns_info->{targetNamespace}},
			};
			
			$id = $xsd->{next_container_i};
		
			# update named complex type name in container of its content model
			$new_container->{name} = '['.$twig->parent->att('name').']' if $twig->parent->att('name');
	
			$xsd->{container}->[$id] = $new_container;
		}
		if($complexType_twig) {
			traverse($das, $complexType_twig, $xsd, $id);
		}
		# traverse inline complexType
		foreach my $item ($child->children()) {
		    traverse($das, $item, $xsd, $id);
		}

		# if no content elements was added by the extension (elements or attributes) undo extension.
		if((defined $new_container) and 
			($id eq $xsd->{next_container_i}) and
			($xsd->{container}->[$element_parent_id]->{num_attributes} eq $current_att_num)) {
			# bug 403702 fix empty extension - backout empty sequence container	
			$id--;
			my $temp = pop @{$xsd->{container}->[$id]->{children}};
			$xsd->{container}->[$id]->{num_children}= @{$xsd->{container}->[$id]->{children}};
			$xsd->{next_container_i} = $id;
		}	
	}
	elsif (is_restriction($child)) {
		
		my $old_next_container_i = $xsd->{next_container_i};

		if($complexType_twig) {
		    # parsing of base type
		    foreach my $item ($complexType_twig->children()) {
				traverse($das, $item, $xsd, $id);
		    }
		}

		my $new_next_container_i = $xsd->{next_container_i};

		# first 'remove' all the new containers ...
		my $parent = $xsd->{container}->[$id];
		my $element = $xsd->{container}->[$element_parent_id];
		$parent->{num_children} = 0;
 		$parent->{children} = ();
		$xsd->{next_container_i} = $old_next_container_i;

		# ... now slide the attributes backwards and keep only them
		my @old_atts = @{$element->{attributes}};
 		$element->{attributes} = ();
		for my $att (@old_atts) {
		    # don't move (or increase counter for) previously existing attributes
		    if ($att >= $xsd->{next_container_i}+1) {
			$xsd->{next_container_i}++;
			# only move attributes that are not in the first free position
			if ($att > $xsd->{next_container_i}) {
			    my $old_att = $att;
			    $att = $xsd->{next_container_i};
			    $xsd->{container}->[$att] = Clone::clone($xsd->{container}->[$old_att]);
			    $xsd->{container}->[$att]->{id} = $att;
			}
		    }
		    push @{$element->{attributes}}, $att;
		}

		# delete the unwanted containers
		delete @{$xsd->{container}}[$xsd->{next_container_i}+1 .. $new_next_container_i];
		#delete @{$xsd->{container}}[$xsd->{next_container_i} .. $new_next_container_i];

		# traverse inline complexType
		foreach my $item ($child->children()) {
		    next if is_attribute($item) || is_attributeGroup($item) || is_anyAttribute($item);
		    traverse($das, $item, $xsd, $id);
		}

		process_restriction_attributes($child, $xsd, $das, $parser_mode, $element_parent_id);
	}
	else {
		die "Illegal complexType definition - must have extension | restriction child\n";
	}
	
	if($content_type
	   and
	   $xsd->{complexType}->{$content_type->{namespace}}
	   and
	   $xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}
	   and
	   $xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}->{recursion_count}
	   and
	   $xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}->{recursion_count} > 0) {
		$xsd->{complexType}->{$content_type->{namespace}}->{$content_type->{local_name}}->{recursion_count}--;
	}

}

sub get_empty_value_simpletype_id {
	my ($das, $builtin) = @_;
	
	my $facet_id = add_facet_unless_in_cache_and_get_facet_id($das, XSDLENGTH, 0);
	my $whitespace_facet_id;
	if($builtin =~ /^(string|normalizedString)$/) {
		$whitespace_facet_id = add_facet_unless_in_cache_and_get_facet_id($das, XSDWHITESPACE, 'preserve');
	}
	else {
		$whitespace_facet_id = add_facet_unless_in_cache_and_get_facet_id($das, XSDWHITESPACE, 'collapse');
	}
	
	return add_simpletype_unless_in_cache_and_get_simpletype_id($das, $builtin, [($whitespace_facet_id, $facet_id)]);
}

sub get_empty_value_metatype_id {
	my ($das, $base_type) = @_;
	my $simpletype_id = get_empty_value_simpletype_id($das, $base_type);
	return add_metatype_unless_in_cache_and_get_metatype_id($das, $base_type, [$simpletype_id]);
}

sub get_fixed_value_simpletype_id {
	my ($das, $xsd, $fixed, $builtin) = @_;
	
	my $fixed_facet_id = add_facet_unless_in_cache_and_get_facet_id($das, XSDENUMERATION, $fixed);
	##my $empty_facet_id = add_facet_unless_in_cache_and_get_facet_id($das, XSDENUMERATION, "");
	my $whitespace = $xsd->{simpleType}->{+XMLSCHEMAURL}->{$builtin}->{facet}->{whiteSpace};
	my $whitespace_facet_id;
	if($whitespace) {
		$whitespace_facet_id = add_facet_unless_in_cache_and_get_facet_id($das, XSDWHITESPACE, $whitespace);
		return add_simpletype_unless_in_cache_and_get_simpletype_id($das, $builtin, [($whitespace_facet_id, $fixed_facet_id)]);
	}
	return add_simpletype_unless_in_cache_and_get_simpletype_id($das, $builtin, [$fixed_facet_id]);
}

sub process_attribute_fixed_value_attribute {
	my ($das, $xsd, $metatype_ids, $fixed) = @_;

	my $new_simpletype_ids = [()];
	my $fixed_facet_id = add_facet_unless_in_cache_and_get_facet_id($das, XSDENUMERATION, $fixed);

	if(scalar @{$metatype_ids} > 1) {
		# this is the case of union of types when one or more of them are lists
		push @{$new_simpletype_ids}, get_fixed_value_simpletype_id($das, $xsd, $fixed, XSDTOKEN);
		my $empty_string = 0;
		foreach my $metatype_id ($metatype_ids) {
			if(not $das->{metatype}->[$metatype_id]->{is_list}) {
				my @simpletype_ids = @{$das->{metatype}->[$metatype_id]->{simpletype_id_list}};
				foreach my $simpletype_id (@simpletype_ids) {
					push @{$new_simpletype_ids}, get_fixed_value_simpletype_id($das, $fixed, $das->{simpletype}->[$simpletype_id]->{type});
				}
			}
		}
		#push @{$new_simpletype_ids}, get_empty_value_simpletype_id($das, XSDSTRING) if($empty_string);
	}
	else {
		my $metatype_id = $metatype_ids->[0];
		if($das->{metatype}->[$metatype_id]->{is_list}) {
			# list
			push @{$new_simpletype_ids}, get_fixed_value_simpletype_id($das, $xsd, $fixed, XSDTOKEN);
			return [add_metatype_unless_in_cache_and_get_metatype_id($das, XSDTOKEN, $new_simpletype_ids)];
		}
		else {
			# union or atom type
			my @simpletype_ids = @{$das->{metatype}->[$metatype_id]->{simpletype_id_list}};
			my $empty_string = 0;
			my $empty_token = 0;
			foreach my $simpletype_id (@simpletype_ids) {
				push @{$new_simpletype_ids}, get_fixed_value_simpletype_id($das, $xsd, $fixed, $das->{simpletype}->[$simpletype_id]->{type});
			}
		}
	}
	if(scalar @{$new_simpletype_ids} > 1) {
		return [add_metatype_unless_in_cache_and_get_metatype_id($das, 'union', $new_simpletype_ids)];
	}
	else {
		return [add_metatype_unless_in_cache_and_get_metatype_id($das, $das->{simpletype}->[$new_simpletype_ids->[0]]->{type}, $new_simpletype_ids)];
	}
}

sub process_fixed_value_attribute {
	my ($das, $xsd, $metatype_ids, $fixed) = @_;

	my $new_simpletype_ids = [()];
	my $fixed_facet_id = add_facet_unless_in_cache_and_get_facet_id($das, XSDENUMERATION, $fixed);

	if(scalar @{$metatype_ids} > 1) {
		# this is the case of union of types when one or more of them are lists
		push @{$new_simpletype_ids}, get_fixed_value_simpletype_id($das, $xsd, $fixed, XSDTOKEN);
		push @{$new_simpletype_ids}, get_empty_value_simpletype_id($das, XSDTOKEN);
		my $empty_string = 0;
		foreach my $metatype_id ($metatype_ids) {
			if(not $das->{metatype}->[$metatype_id]->{is_list}) {
				my @simpletype_ids = @{$das->{metatype}->[$metatype_id]->{simpletype_id_list}};
				foreach my $simpletype_id (@simpletype_ids) {
					push @{$new_simpletype_ids}, get_fixed_value_simpletype_id($das, $fixed, $das->{simpletype}->[$simpletype_id]->{type});
					$empty_string++ if ($das->{simpletype}->[$simpletype_id]->{type} =~ /^(string|normalizedString)$/);
				}
			}
		}
		push @{$new_simpletype_ids}, get_empty_value_simpletype_id($das, XSDSTRING) if($empty_string);
	}
	else {
		my $metatype_id = $metatype_ids->[0];
		if($das->{metatype}->[$metatype_id]->{is_list}) {
			# list
			push @{$new_simpletype_ids}, get_fixed_value_simpletype_id($das, $xsd, $fixed, XSDTOKEN);
			push @{$new_simpletype_ids}, get_empty_value_simpletype_id($das, XSDTOKEN);
		}
		else {
			# union or atom type
			my @simpletype_ids = @{$das->{metatype}->[$metatype_id]->{simpletype_id_list}};
			my $empty_string = 0;
			my $empty_token = 0;
			foreach my $simpletype_id (@simpletype_ids) {
				push @{$new_simpletype_ids}, get_fixed_value_simpletype_id($das, $xsd, $fixed, $das->{simpletype}->[$simpletype_id]->{type});
				($das->{simpletype}->[$simpletype_id]->{type} =~ /^(string|normalizedString)$/) ? $empty_string++ : $empty_token++ ;
			}

			if($empty_token) {
				push @{$new_simpletype_ids}, get_empty_value_simpletype_id($das, XSDTOKEN);
			}
			if($empty_string) {
				push @{$new_simpletype_ids}, get_empty_value_simpletype_id($das, XSDSTRING);
			}
		}
	}
	return [add_metatype_unless_in_cache_and_get_metatype_id($das, 'union', $new_simpletype_ids)];
}

sub process_default_value_attribute {
	my ($das, $metatype_ids) = @_;
	
	if(scalar @{$metatype_ids} > 1) {
		# this is the case of union of types when one or more of them are lists
		push @{$metatype_ids}, get_empty_value_metatype_id($das, XSDTOKEN);
	}
	else {
		my $metatype_id = $metatype_ids->[0];
		if($das->{metatype}->[$metatype_id]->{is_list}) {
			# list
			push @{$metatype_ids}, get_empty_value_metatype_id($das, XSDTOKEN);
		}
		else {
			# union or atom type
			my @simpletype_ids = @{$das->{metatype}->[$metatype_id]->{simpletype_id_list}};
			my $empty_string = 0;
			my $empty_token = 0;
			foreach my $simpletype_id (@simpletype_ids) {
				($das->{simpletype}->[$simpletype_id]->{type} =~ /^(string|normalizedString)$/) ? $empty_string++ : $empty_token++ ;
			}
			
			if($empty_token) {
				push @simpletype_ids, get_empty_value_simpletype_id($das, XSDTOKEN);
			}
			else {
				push @simpletype_ids, get_empty_value_simpletype_id($das, XSDSTRING);
			}
			$metatype_ids = [add_metatype_unless_in_cache_and_get_metatype_id($das, 'union', [@simpletype_ids])];
		}
	}
	return [@{$metatype_ids}];
}

sub traverse {
	my ($das,$twig,$xsd,$id)=@_;
	return unless $twig;
	my $simpleType;
	my $parser_mode = XSD__STRICT;
	my $metatype_id_list = [];
	
	my $element_parent_id = get_element_parent_id($xsd, $id);
	
	if (is_all($twig) or
		is_choice($twig) or
		is_sequence($twig)) {
		
		return if (not $twig->att('ref') and defined $twig->{empty});
		process_content_model($das, $xsd, $twig, $parser_mode, $element_parent_id, $id);

		# now traverse children of content model group (elements | attributes | any )
		my $current_id = $xsd->{next_container_i};
		foreach my $item ($twig->children()) {
			traverse($das, $item, $xsd, $current_id);
		}
	}
	elsif (is_any($twig)) {
		process_any_element_entity($das, $xsd, $twig, $parser_mode, $element_parent_id, $id);
	}
	elsif (is_anyAttribute($twig)) {
		process_any_attribute_entity($das, $xsd, $twig, $element_parent_id, $id);
	}
	elsif (is_attribute($twig)) {
		process_attribute_child($twig, $xsd, $das, $parser_mode, $element_parent_id);
	}
	elsif (is_attributeGroup($twig)) {
		
		my $group_type_info = extract_type_info($xsd, $das, $twig, 'ref');
		return unless defined $xsd->{attributeGroup}->{$group_type_info->{namespace}}
				and
				$xsd->{attributeGroup}->{$group_type_info->{namespace}}->{$group_type_info->{local_name}};
		my $attGroup = ${$xsd->{attributeGroup}->{$group_type_info->{namespace}}->{$group_type_info->{local_name}}->{twig}};
		foreach my $item ($attGroup->children()) {
			traverse($das, $item, $xsd, $element_parent_id)
				if(is_attribute($item) or is_attributeGroup($item) or is_anyAttribute($item));
		}
	}
	elsif (is_group($twig)) {
		return if (not $twig->att('ref') and defined $twig->{empty});
		my $group_type_info = extract_type_info($xsd, $das, $twig, 'ref');
		return unless defined $xsd->{group}->{$group_type_info->{namespace}}
				and
				$xsd->{group}->{$group_type_info->{namespace}}->{$group_type_info->{local_name}};
		
		my $group = $xsd->{group}->{$group_type_info->{namespace}}->{$group_type_info->{local_name}};
		my $group_twig = ${$group->{twig}};
		# now traverse children of model group (elements | any )
		foreach my $item ($group_twig->children()) {
		    # have group [min|max]Occur apply to its children 
		    $item->set_att(minOccurs => $twig->att(XSDMINOCCURS)) if (defined $twig->att(XSDMINOCCURS));
		    $item->set_att(maxOccurs => $twig->att(XSDMAXOCCURS)) if (defined $twig->att(XSDMAXOCCURS));
		    traverse($das, $item, $xsd, $id);
		    $item->del_att(XSDMINOCCURS);
		    $item->del_att(XSDMAXOCCURS);
		}
	}
	elsif(is_simpleContent($twig)) {
		prosess_simple_content($twig, $xsd, $das,$parser_mode, $id);
	}
	elsif(is_complexContent($twig)) {
		prosess_complex_content($twig, $xsd, $das, $parser_mode, $id);
	}
	elsif (is_complexType($twig)) {
		my $mixed="false";
		$mixed = $twig->att('mixed') if $twig->att('mixed');
		$xsd->{container}->[$element_parent_id]->{mixed}=$mixed;
		# process contents of complex type
		foreach my $item ($twig->children()) {
			traverse($das, $item, $xsd, $id); 
		}
	}
	elsif (is_element($twig)) {
		if ($twig->att('ref')) {
			my $info = extract_type_info($xsd, $das, $twig, 'ref');
	
			if($info->{local_name} eq "schema" and $info->{namespace} eq XMLSCHEMAURL) {
				my $anySchemaType_twig= parse XML::Twig::Elt(
					'<xsd:any processContents="skip" namespace="'.XMLSCHEMAURL.'"/>');
				$anySchemaType_twig->paste($twig);
				traverse($das, $anySchemaType_twig, $xsd, $id);
				return;
			}
		}
		# check if element is refference to substitution group head
		my $substitution_group_container;
		if($twig->att('ref')) {
			my $info = extract_type_info($xsd, $das, $twig, 'ref');
			
			if($xsd->{substitutionGroup}->{$info->{namespace}}->{$info->{local_name}}) {
				$substitution_group_container =
					get_substitution_group_container($das, $xsd, $twig, $id, $element_parent_id);
					foreach my $element_name (keys %{$xsd->{substitutionGroup}->{$info->{namespace}}->{$info->{local_name}}}) {
						my $element_ns = $xsd->{substitutionGroup}->{$info->{namespace}}->{$info->{local_name}}->{$element_name};
						my $element_twig = ${$xsd->{element}->{$element_ns}->{$element_name}->{twig}};
						next if($element_twig->att('abstract') and $element_twig->att('abstract') eq 'true');
						add_element_ref_container($das, $xsd, $twig, $element_twig, $element_ns, $parser_mode, $substitution_group_container, $element_parent_id); 
					}
			}
		}
		my $new_element_parent_id = $id;
		## if element defined as abstract or ref to abstract do not process it.
		return if($twig->att('abstract') and $twig->att('abstract') eq 'true');
		if($twig->att('ref')) {
			my $info = extract_type_info($xsd, $das, $twig, 'ref');
			die schema_parsing_error_at($twig)."\n"."Namespace ".$info->{namespace}." not found" 
				unless defined $xsd->{element}->{$info->{namespace}};
            die schema_parsing_error_at($twig)."\n"."Element ".$info->{local_name}." is not declared" 
                unless defined $xsd->{element}->{$info->{namespace}}->{$info->{local_name}};
			my $elem_twig = ${$xsd->{element}->{$info->{namespace}}->{$info->{local_name}}->{twig}};
			return if($elem_twig->att('abstract') and $elem_twig->att('abstract') eq 'true');
		}
		if($substitution_group_container) {
			$new_element_parent_id = process_element($das, $xsd, $twig, $parser_mode, $substitution_group_container);
		}
		else {
			$new_element_parent_id = process_element($das, $xsd, $twig, $parser_mode, $id);
		}
		# in case no new container was created parrent keep previous parent id
		$new_element_parent_id = $id if(not defined $new_element_parent_id);
		
		my $element_type_info = $xsd->{container}->[$new_element_parent_id]->{element_type_info};
		if (($element_type_info->{type} eq 'complexType') and $element_type_info->{'att_type'}) {
			# enter here only on named complexTypes
			my $complexType;
			if ($element_type_info->{complexType}) {
				$xsd->{container}->[$new_element_parent_id]->{complexType}->{name}          = $element_type_info->{local_name};
				$xsd->{container}->[$new_element_parent_id]->{complexType}->{namespace_uri} = $element_type_info->{namespace};
				#$complexType = $xsd->{complexType}->{$element_type_info->{namespace}}->{$element_type_info->{local_name}};
				#$xsd->{complexType}->{$element_type_info->{namespace}}->{$element_type_info->{local_name}}->{content_id} = $new_element_parent_id;
			}
			#my $complexType_twig = ${$complexType->{twig}};
			#if ($complexType and !is_empty_complexType($xsd,$complexType_twig)) {
			#	$complexType->{recursion_count}++;
			#	if($complexType->{recursion_count} > COMPLEXTYPE_RECURSION_DEPTH ) {
#					warn "Detected deep complexType '".$complexType_twig->att('name')."' recursion from element '".$xsd->{container}->[$new_element_parent_id]->{name}."' - skip\n";
			#	}
			#	else {
			#		traverse($das, $complexType_twig, $xsd, $new_element_parent_id);
			#	}
			#	$complexType->{recursion_count}--;
			#}
			#$xsd->{container}->[$new_element_parent_id]->{mixed} = $complexType_twig->att('mixed') if $complexType_twig->att('mixed');
		}
		foreach my $child ($twig->children()) {

					traverse($das, $child, $xsd, $new_element_parent_id);
		}
		# process anyAttribute attributes...
		process_any_attribute_for_element($das, $xsd, $twig, $new_element_parent_id);
	}

	return;
}

sub get_element_parent_id {
	my ($xsd, $id)  = @_;
	return (exists $xsd->{container}->[$id]
		and
		($xsd->{container}->[$id]->{container_type} eq 'element' or
		 $xsd->{container}->[$id]->{container_type} eq 'complex_type')) ?
		$id :
		$xsd->{container}->[$id]->{element_parent_id}
	;
}

sub get_local_name {
	my ($s) = @_;
	$s=~s/.*://;
	return $s;
}

sub get_prefix {
	my ($s) = @_;
	if ($s=~/:/) {
		$s=~s/:.*$//;
		return $s;
	}
	return '';
}

sub get_target_namespace {
	my ($twig) = @_;

	if(not $twig->root) {
		warn "Twig root disappeared...";
		return "";
	}

	return $twig->root->att('targetNamespace') || "";
}

sub get_namespace_information {
	my ($twig) = @_;
	
	return unless $twig->root;

	my $namespace = {
		elementFormDefault   => XSDUNQUALIFIED,
		attributeFormDefault => XSDUNQUALIFIED,
	};

	foreach my $att (keys %{$twig->root->atts}) {
		if ($att eq 'elementFormDefault') {
			if (!$twig->root->att('targetNamespace') and $twig->root->att('elementFormDefault') eq XSDQUALIFIED) {
				$namespace->{elementFormDefault}  = XSDUNQUALIFIED;
			} else {
				$namespace->{elementFormDefault}  = $twig->root->att('elementFormDefault');
			}
		} elsif ($att eq 'attributeFormDefault') {
			$namespace->{attributeFormDefault}= $twig->root->att('attributeFormDefault');
		} elsif ($att=~/^xmlns:(\S+)/) {
			$namespace->{prefix2uri}->{$1}    = $twig->root->att($att);
		} elsif ($att eq 'xmlns') {
			$namespace->{prefix2uri}->{''}    = $twig->root->att($att);
			$namespace->{defaultNamespaceURI} = $twig->root->att($att);
		} else {
	#		warn("unsupported attribute |$att| of <schema> element\n");
		}
	}

	if ('XML::Twig::Elt' eq ref $twig) {
		foreach my $att (keys %{$twig->atts}) {
			if ($att=~/^xmlns:(\S+)/) {
				$namespace->{prefix2uri}->{$1}    = $twig->att($att);
			} elsif ($att eq 'xmlns') {
				$namespace->{prefix2uri}->{''}    = $twig->att($att);
				$namespace->{defaultNamespaceURI} = $twig->att($att);
			} else {
		#		warn("unsupported attribute |$att| of <schema> element\n");
			}
		}
	}

	$namespace->{targetNamespace} = get_target_namespace($twig); 

	$namespace->{xsdprefix}=XSDPREFIX;

	return $namespace;
}

sub get_local_simpleType_name {
	my ($das, $xsd, $name, $namespace, $id) = @_;
	return (($namespace) ? $namespace.":" : "").$name."[$id]\@simpleType";
}

sub blobber {
	my ($xsd) = @_;
    my $das = $xsd->{das};
	my $blob='';
	my $schema_h = $F5::XML::Schema::SchemaDefinitions::schema_h;
	
	my $num_children_indexes = 0;
	my $num_attribute_indexes = 0;
	foreach my $item (@{$xsd->{container}}) { ## BEGIN blob_container_t
		$num_children_indexes += (exists $item->{num_children}) ? $item->{num_children} : 0;
		$num_attribute_indexes += (exists $item->{num_attributes}) ? $item->{num_attributes} : 0
	}

	my $metatype_facet_indexes = 0;
	for (my $i=0; $i<@{$das->{metatype}}; ++$i) { # BEGIN blob_metatype_t
		if(exists $das->{metatype}->[$i]->{facet_id_list}) {
			$metatype_facet_indexes += @{$das->{metatype}->[$i]->{facet_id_list}};
		} 
	}
	my $simpletype_facet_indexes = 0;
	for (my $i=0; $i<@{$das->{simpletype}}; ++$i) { # BEGIN blob_simpletype_t
		$simpletype_facet_indexes += @{$das->{simpletype}->[$i]->{facet_id_list}};
	} 

	$blob.= pack('I',eval($schema_h->XML_SCHEMA_FILE_SIGNATURE)); # signature
	$blob.= pack('I',$schema_h->XML_SCHEMA_DAS_VERSION);          # version
	$blob.= pack('I',scalar @{$xsd->{container}});                # container_count
	$blob.= pack('I',$num_children_indexes);                      # children_indexes
	$blob.= pack('I',$num_attribute_indexes);                     # attribute_indexes
	$blob.= pack('I',scalar(@{$das->{namespace_uris}}));          # ns_uris_count
	$blob.= pack('I',scalar(@{$das->{simpleType_ids}}));          # simpleTypes_count
	$blob.= pack('I',scalar(@{$das->{metatype}}));                # metatype_count
	$blob.= pack('I',scalar(@{$das->{simpletype}}));              # simpletype_count
	$blob.= pack('I',scalar(@{$das->{facet}}));                   # facet_count
	$blob.= pack('I',$metatype_facet_indexes);                    # metatype facet indexes count
	$blob.= pack('I',$simpletype_facet_indexes);                  # simpletype facet indexes count
	$blob.= pack('I20',0); # place holder pointers before starting container section
	
	my $max_element_name = $schema_h->MAX_ELEMENT_NAME;
	foreach my $item (@{$xsd->{container}}) { ## BEGIN blob_container_t
                $blob.= pack('I',$item->{id});    # id
                $blob.= pack('I',0);              # id
		$item->{name} = $item->{container_type}."_".$item->{id} if($item->{name} eq "");
		$blob.= pack("Z$max_element_name",$item->{name}); # name
		# container_type
		if ($item->{container_type} && exists $F5::XML::Schema::SchemaDefinitions::container_type2id->{$item->{container_type}})  {
			$blob.= pack('I',$F5::XML::Schema::SchemaDefinitions::container_type2id->{$item->{container_type}});
		} else {
			$blob.= pack('I', $schema_h->XSD_CONT_TYPE__ILLEGAL);
		}
		$blob.= pack('I',$item->{parent_id});         # parent_id
		$blob.= pack('I',$item->{element_parent_id}); # element_parent_id
		$blob.= pack('I',$item->{id_in_parent});      # id_in_parent
		my $num_children = (exists $item->{num_children}) ? $item->{num_children} : 0;
		$blob.= pack('I',$num_children); # children_count
		my $num_attributes = (exists $item->{num_attributes}) ? $item->{num_attributes} : 0;
		$blob.= pack('I',$num_attributes); # attribute_count
		$blob.= pack('I',$item->{max_self_count}); # max_self_count
		$blob.= pack('I',$item->{min_self_count}); # min_self_count
		$blob.= pack('I',(defined $item->{simpleType_id}) ? $item->{simpleType_id} : -1); # simpleType index
		$blob.= pack('I',(defined $item->{complexType_id}) ? $item->{complexType_id} : -1); # complexType index
		$blob.= pack('I',$F5::XML::Schema::SchemaDefinitions::parser_mode2int->{$item->{parser_mode}}); # parser_mode
		if($item->{ns_wildcard} and $item->{ns_wildcard} eq "##any") {
			$blob.= pack('I', $schema_h->ANYNAMESPACE_INDEX);
		}
		elsif($item->{ns_wildcard} and $item->{ns_wildcard} eq "##other") {
			$blob.= pack('I', $schema_h->OTHERNAMESPACE_INDEX);
		}
		elsif($item->{ns_wildcard} and $item->{ns_wildcard} eq "##local") {
			$blob.= pack('I', $schema_h->LOCALNAMESPACE_INDEX);
		}
		else {
			$blob.= pack('I',0); # no namespace wildcard
		}
		$blob.= pack('I',$item->{ns_uri_id}); # namespace_uri_id
		$blob.= pack('I',$F5::XML::Schema::SchemaDefinitions::form2int->{$item->{form}}); # form
		$blob.= pack('I',(defined $item->{anyType}) ? $item->{anyType} : 0); # anyType flag
		$blob.= pack('I',(exists $item->{mixed}) ? ( ($item->{mixed} eq 'true' or $item->{mixed} eq "1") ? 1 : 0) : -1); # mixed
		$blob.= pack('I',(exists $item->{nillable}) ? ( ($item->{nillable} eq 'true' or $item->{nillable} eq "1") ? 1 : 0) : -1); # nillable
		$blob.= pack('I',(exists $item->{default}) ? 1 : 0); # default
		$blob.= pack('I',$item->{is_optional}); # is_optional
	} 
        # END blob_container_t

        # BEGIN children indexes
	foreach my $item (@{$xsd->{container}}) { 
		my $num_children = (exists $item->{num_children}) ? $item->{num_children} : 0;
		$blob.= pack("I$num_children",@{$item->{children}}) if $num_children > 0;
	}

        # BEGIN attribute indexes
	foreach my $item (@{$xsd->{container}}) { 
		my $num_attributes = (exists $item->{num_attributes}) ? $item->{num_attributes} : 0;
		$blob.= pack("I$num_attributes",@{$item->{attributes}}) if $num_attributes > 0;
	}

        # BEGIN blob_namespace_uri_t
	my $max_namespace_uri = $schema_h->MAX_NAMESPACE_URI;
	for (my $i=0; $i<@{$das->{namespace_uris}}; ++$i) { 
		$blob.= pack('I',$i);
		$blob.= pack("Z$max_namespace_uri",$das->{namespace_uris}->[$i]);
	} 
        # END blob_namespace_uri_t

        # BEGIN blob_stype_header_t
	for (my $i=0; $i<@{$das->{simpleType_ids}}; ++$i) { 
                $blob.= pack('I',$i);  # id
                $blob.= pack('I',0); # id
		$blob.= pack("Z$max_element_name",$das->{simpleType_ids}->[$i]->{name}); # name
		$blob.= pack('I',((defined $das->{simpleType_ids}->[$i]->{namespace_id}) ? $das->{simpleType_ids}->[$i]->{namespace_id} : -1)); # namespace_uri_id
		$blob.= pack('I',$das->{simpleType_ids}->[$i]->{parent_id}); # parent_id
		my $metatype_count = 0;
		$das->{simpleType_ids}->[$i]->{metatype_id_list} = [] unless $das->{simpleType_ids}->[$i]->{metatype_id_list};
		$metatype_count = @{$das->{simpleType_ids}->[$i]->{metatype_id_list}};
		if ($metatype_count>$schema_h->MAX_META_TYPE_COUNT) {
			die "\nThe number of metatype ids in a container is (".
				scalar(@{$das->{simpleType_ids}->[$i]->{metatype_id_list}}).
				") larger then pre-defined value ($schema_h->MAX_META_TYPE_COUNT).\n";
		}
		$blob.= pack('I',$metatype_count); # metatype_count
		for (my $j=0;$j<$schema_h->MAX_META_TYPE_COUNT;++$j) { # metatype_id[]
			if (not exists $das->{simpleType_ids}->[$i]->{metatype_id_list}) {
				$blob.= pack('I',-1) x $schema_h->MAX_META_TYPE_COUNT;
				last;
			}
			if($j < @{$das->{simpleType_ids}->[$i]->{metatype_id_list}}){
				$blob.= pack('I',$das->{simpleType_ids}->[$i]->{metatype_id_list}->[$j]);
			} else {
				$blob.= pack('I',-1);
			}
		}
	} 
        # END blob_stype_header_t

        # BEGIN blob_metatype_t
	for (my $i=0; $i<@{$das->{metatype}}; ++$i) { 
		$blob.= pack('I2',0); # id pointer placeholder 8 bytes
		$blob.= pack('I',$das->{metatype}->[$i]->{is_list}); # is list
		
		my $facet_count = 0;
		if(exists $das->{metatype}->[$i]->{facet_id_list}) {
			$facet_count = @{$das->{metatype}->[$i]->{facet_id_list}};
		} 
		$blob.= pack('I',int $facet_count); # facet_count
	
		my $simpletype_count = @{$das->{metatype}->[$i]->{simpletype_id_list}};
                if ($simpletype_count >= $schema_h->MAX_SIMPLE_TYPE_COUNT) {
                   die "Error: amount of simple-type elements inside a metatype exceeded, allowed: $schema_h->{MAX_SIMPLE_TYPE_COUNT}, actual: $simpletype_count";
                }
		$blob.= pack('I',$simpletype_count); # simpletype_count
		for (my $j=0; $j<$schema_h->MAX_SIMPLE_TYPE_COUNT; ++$j) { # simpletype_id[]
			if ($j<@{$das->{metatype}->[$i]->{simpletype_id_list}}) {
				$blob.= pack('I',$das->{metatype}->[$i]->{simpletype_id_list}->[$j]);
			}else{
				$blob.= pack('I',-1);
			}
		}
	} 
        # END   blob_metatype_t

        # BEGIN blob_simpletype_t
	for (my $i=0; $i<@{$das->{simpletype}}; ++$i) { 
		$blob.= pack('I2',0); # id pointer placeholder 8 bytes
		if (defined $F5::XML::Schema::SchemaDefinitions::simpleType2int->{$das->{simpletype}->[$i]->{type}}) {
			$blob.= pack('I',eval($F5::XML::Schema::SchemaDefinitions::simpleType2int->{$das->{simpletype}->[$i]->{type}}));
		}else{
			$blob.= pack('I',eval($F5::XML::Schema::SchemaDefinitions::simpleType2int->{anySimpleType}));
		}
		my $facet_count = @{$das->{simpletype}->[$i]->{facet_id_list}};
		$blob.= pack('I',$facet_count); # facet_count
	} 
        # END blob_simpletype_t

        # BEGIN blob_facet_t
	my $max_facet_value_length = $schema_h->MAX_FACET_VALUE_LENGTH;
	my $num_facets = @{$das->{facet}} if defined $das->{facet};
	for (my $i=0; $i<$num_facets; ++$i) { 
		$blob.=pack('I',eval($F5::XML::Schema::SchemaDefinitions::facet2int->{$das->{facet}->[$i]->{type}}));
		$blob.=pack("Z$max_facet_value_length",$das->{facet}->[$i]->{value});
	} 
        # END blob_facet_t

        # BEGIN metatype facet indexes
	for (my $i=0; $i<@{$das->{metatype}}; ++$i) { 
		my $facet_count = 0;
		if(exists $das->{metatype}->[$i]->{facet_id_list}) {
			$facet_count = scalar @{$das->{metatype}->[$i]->{facet_id_list}};
		} 
		$blob.= pack("I$facet_count",@{$das->{metatype}->[$i]->{facet_id_list}}) if $facet_count > 0;
	}
        # END metatype facet indexes

        # BEGIN simpletype facet indexes
	for (my $i=0; $i<@{$das->{simpletype}}; ++$i) { 
		my $facet_count = scalar @{$das->{simpletype}->[$i]->{facet_id_list}};
		$blob.= pack("I$facet_count",@{$das->{simpletype}->[$i]->{facet_id_list}}) if $facet_count > 0;
	}
        # END simpletype facet indexes
	
	return $blob;
}

sub check_min_is_le_max {
	my($min_ref,$max_ref)=@_;
	if ($$min_ref>$$max_ref) {
		warn XSDMINOCCURS." is greater than ".XSDMAXOCCURS. "($$min_ref > $$max_ref)\n";
		warn "Overwriting ".XSDMAXOCCURS." to be ".XSDMINOCCURS."($$max_ref := $$min_ref)\n";
		$$min_ref=$$max_ref;
		return 0;
	}
	return 1;
}

sub is_empty_content_complexType {
	my ($twig) = @_;
	foreach my $child ($twig->children) {
		return 0 if (
			is_group($child) or
			is_choice($child) or
			is_sequence($child) or
			is_simpleContent($child) or
			is_complexContent($child) or
			is_all($child)
		);
	}
	return 1;
}

sub is_empty_model_group {
	my ($twig) = @_;
	if (is_sequence($twig) or is_choice($twig)) {
		foreach my $child ($twig->children) {
			if (
				is_element($child) or
				is_group($child) or
				is_choice($child) or
				is_sequence($child) or
				is_any($child)
			) {
				return 0;
			}
		}
		return 1;
	} elsif (is_all($twig)) {
		foreach my $child ($twig->children) {
			return 0 if is_element($child);
		}
		return 1;
	}
	return 0;
}

sub get_twig_of_model_group {
	my ($xsd,$complexType_twig)=@_;
	my $ns_info = get_namespace_information($complexType_twig);
	foreach my $mg ( (xsd_group($ns_info),
			  xsd_sequence($ns_info),
			  xsd_choice($ns_info),
			  xsd_all($ns_info),
			  xsd_simpleContent($ns_info),
			  xsd_complexContent($ns_info)) ) {
		if ($complexType_twig->first_child($mg) and
		    (not is_empty_model_group($complexType_twig->first_child($mg)))) {
			return $complexType_twig->first_child($mg);
		}
	}
	return undef;
}

sub is_empty_complexType {
	my ($xsd,$twig) = @_;
	return 0 if get_twig_of_model_group($xsd,$twig);
	return 0 if $twig->children(xsd_attribute(get_namespace_information($twig)));
	return 0 if $twig->children(xsd_attributeGroup(get_namespace_information($twig)));
	return 0 if $twig->children(xsd_anyAttribute(get_namespace_information($twig)));
	return 1;
}


sub new_xsd {
	return bless {
		schemaLocation  => '',
		container => [],
		next_container_i => 0,
		attribute => {},
		element => {},
		complexType => {},
		simpleType => { %$F5::XML::Schema::SchemaDefinitions::builtin_simple_types, },
		group => {},
		substitutionGroup => {},
	}, 'XSD';
}

sub DESTROY { } # so AUTOLOAD won't catch it in bless...

sub AUTOLOAD {
	my $subname = $AUTOLOAD;
	$subname=~s/.*:://;
	if ($subname=~/^xsd_(\S+)$/) {
		die "bad xsd_ sub pattern\n" unless $1;
		if ($_[0]) { # and 'XSD' eq ref $_[0]) {
			die "missing or unknown XSDL vocabulary term in AUTOLOADing sub: |$subname|\n" unless exists $xsdl_vocabulary->{$1};
			return $_[0]->{xsdprefix}.':'.$1;
		} else {
			die "unexpected argument to XSDL vocabulary AUTOLOADing sub: expected an XSD blessed data structure and got something else: |$subname|\n";
		}
	}elsif($subname=~/^is_(\S+)$/){
		die "missing or unknown XSDL vocabulary term in AUTOLOADing sub: |$subname|\n"
			unless exists $xsdl_vocabulary->{$1};
		die "sub $subname requires a twig as an argument\n" unless @_==1;
		if ($_[0]->local_name eq $1){ #&& ($_[0]->namespace eq XMLSCHEMAURL)) {
			return 1;
		} else {
			return 0;
		}
	}else{
		die "undefined sub >>>$subname<<< this is a bug!\n";
	}
}

sub handle_top_level {
	my ($das, $xsd, $root)=@_;
	foreach my $child ($root->children()) {
		handle_global_element_declaration($das, $xsd, $child) 
			if is_element($child);
		handle_global_attribute_declaration($das, $xsd, $child) 
			if is_attribute($child);
		handle_global_definition($das, $xsd, $child) 
			if (is_complexType($child) or
				is_group($child) or
				is_attributeGroup($child) or
				is_simpleType($child));
	}
}

sub handle_redefine_elements {
	my ($das, $xsd, $twig) = @_;
	foreach my $child ($twig->children()) {
		handle_redefine_simpleType($das, $xsd, $child) 
			if is_simpleType($child);
		handle_redefine_complexType($das, $xsd, $child) 
			if is_complexType($child);
		handle_redefine_group($das, $xsd, $child) 
			if is_group($child);
		handle_redefine_attributeGroup($das, $xsd, $child) 
			if (is_attributeGroup($child));
	}

}

sub handle_redefine_simpleType {
	my ($das, $xsd, $child) = @_;
	
	my $namespace = "";
	if(exists $child->root->{'att'}->{'targetNamespace'}) {
		$namespace = $child->root->att('targetNamespace');		
	}
	my $name = get_local_name($child->att('name')) ;
	
	if(not (exists $xsd->{simpleType}->{$namespace} and exists $xsd->{simpleType}->{$namespace}->{$name})) {
		#print "'$name' in '$namespace' not found \n";
		return;
	}
	
	my $redefined_name = $name.'@redefined';
	$xsd->{simpleType}->{$namespace}->{$redefined_name} = $xsd->{simpleType}->{$namespace}->{$name};
	undef $xsd->{simpleType}->{$namespace}->{$name} ;
	
	#change redefined type name
	if($xsd->{simpleType}->{$namespace}->{$redefined_name}->{twig}) {
		my $twig = ${$xsd->{simpleType}->{$namespace}->{$redefined_name}->{twig}};
		$twig->set_att('name' => $redefined_name);
	}
	else {
		$xsd->{simpleType}->{$namespace}->{$redefined_name}->{name} = $redefined_name;
	}
	
	if(is_restriction($child->first_child)) {
		if(get_local_name($child->first_child->att('base')) eq $name) {
			$child->first_child->set_att( 'xmlns' => $namespace );
			$child->first_child->set_att( 'base' => $child->first_child->att('base').'@redefined');
		}
	}
	
	$xsd->{simpleType}->{$namespace}->{$name}->{twig}=\$child;
	#handle_named_simple_type($das, $xsd, $child);
	
	
}

sub handle_redefine_complexType {
	my ($das, $xsd, $child) = @_;

	my $namespace = "";
	if(exists $child->root->{'att'}->{'targetNamespace'}) {
		$namespace = $child->root->att('targetNamespace');		
	}
	my $name = get_local_name($child->att('name')) ;
	
	if(not (exists $xsd->{complexType}->{"$namespace"} and exists $xsd->{complexType}->{"$namespace"}->{"$name"})) {
		return;
	}
		
	my $redefined_name = $name.'@redefined';
	$xsd->{complexType}->{$namespace}->{$redefined_name} = $xsd->{complexType}->{$namespace}->{$name};
	undef $xsd->{complexType}->{$namespace}->{$name} ;
	
		#change redefined type name
	if($xsd->{complexType}->{$namespace}->{$redefined_name}->{twig}) {
		my $twig = ${$xsd->{complexType}->{$namespace}->{$redefined_name}->{twig}};
		$twig->set_att('name' => $redefined_name);
	}
	
	my $content = $child->first_child;
	if (is_complexContent($content) or is_simpleContent($content)) {
		if(is_restriction($content->first_child) or is_extension($content->first_child)) {
			if(get_local_name($content->first_child->att('base')) eq $name) {
				$content->first_child->set_att( 'xmlns' => $namespace );
				$content->first_child->set_att( 'base' => $content->first_child->att('base').'@redefined');
			}
			
		}
	}

	$xsd->{complexType}->{$namespace}->{$name}->{twig}=\$child;
}

sub handle_redefine_group {
	my ($das, $xsd, $child) = @_;
	
	my $namespace = "";
	if(exists $child->root->{'att'}->{'targetNamespace'}) {
		$namespace = $child->root->att('targetNamespace');		
	}
	my $name = get_local_name($child->att('name')) ;
	

	if(not (exists $xsd->{group}->{"$namespace"} and exists $xsd->{group}->{"$namespace"}->{"$name"})) {
		return;
	}
		
	my $redefined_name = $name.'@redefined';
	$xsd->{group}->{$namespace}->{$redefined_name} = $xsd->{group}->{$namespace}->{$name};
	undef $xsd->{group}->{$namespace}->{$name} ;
	
	my $elt = $child;
	while( $elt= $elt->next_elt( $child ) )
    { # insert processing code here
		if(exists $elt->{'att'}->{'ref'} and get_local_name($elt->att('ref')) eq $name) {
			$elt->set_att('xmlns' => $namespace );
			$elt->set_att('ref' => $elt->att('ref').'@redefined' );
			last;
		}
    }
	$xsd->{group}->{$namespace}->{$name}->{twig}=\$child;
	
	
}

sub handle_redefine_attributeGroup {
	my ($das, $xsd, $child) = @_;
	
	my $namespace = "";
	if(exists $child->root->{'att'}->{'targetNamespace'}) {
		$namespace = $child->root->att('targetNamespace');		
	}
	my $name = get_local_name($child->att('name')) ;
	

	if(not (exists $xsd->{attributeGroup}->{"$namespace"} and exists $xsd->{attributeGroup}->{"$namespace"}->{"$name"})) {
		return;
	}
		
	my $redefined_name = $name.'@redefined';
	$xsd->{attributeGroup}->{$namespace}->{$redefined_name} = $xsd->{attributeGroup}->{$namespace}->{$name};
	undef $xsd->{attributeGroup}->{$namespace}->{$name};
	
	my $elt = $child;
	while( $elt= $elt->next_elt( $child ) )
    { # insert processing code here
		if(exists $elt->{'att'}->{'ref'} and get_local_name($elt->att('ref')) eq $name) {
			$elt->set_att('xmlns' => $namespace );
			$elt->set_att('ref' => $elt->att('ref').'@redefined' );
			last;
		}
    }

	$xsd->{attributeGroup}->{$namespace}->{$name}->{twig}=\$child;
	

} 

sub handle_global_element_declaration {
	my ($das, $xsd, $twig) = @_;
	my $ns_info = get_namespace_information($twig);
	my $nillable = "false";
	my $min = 1;
	my $max = 1;
	
	my $element_name = $twig->att('name');
	die schema_parsing_error_at($twig)."global element declaration must include a name\n"
		unless $element_name;

	# save substitution group information
	if($twig->att('substitutionGroup')) {
		my $info = extract_type_info($xsd, $das, $twig, 'substitutionGroup');
		$xsd->{substitutionGroup}->{$info->{namespace}}->{$info->{local_name}}->{$element_name} = $ns_info->{targetNamespace};
	}

	$nillable = $twig->att('nillable') if ($twig->att('nillable'));
	
	my $form=XSDFORMILLEGAL;
	warn "meaningless 'form' attribute in global element definition\n" if $twig->att('form');
	if ("" eq $ns_info->{targetNamespace}) {
		$form=XSDUNQUALIFIED;
	} else {
		$form=XSDQUALIFIED;
	}
	return if exists $xsd->{element}->{$ns_info->{targetNamespace}}->{$element_name};
	$xsd->{element}->{$ns_info->{targetNamespace}}->{$element_name}->{twig}=\$twig;
	if($twig->att('abstract') and $twig->att('abstract') eq 'true') {
		$min = $max = 0;
	}
	
	$xsd->{next_container_i}++ if $xsd->{container}->[$xsd->{next_container_i}];
	my $id = $xsd->{next_container_i};
	$xsd->{element}->{$ns_info->{targetNamespace}}->{$element_name}->{id}=$id;
	$xsd->{container}->[$id]= {
		id                => $id,
		name              => $twig->att('name'),
		container_type    => 'element',
		parent_id         => -1,
		element_parent_id => -1,
		id_in_parent      => -1,
		num_children      => 0,
		children          => [],
		num_attributes    => 0,
		attributes        => [],
		max_self_count    => $max,
		min_self_count    => $min,
		parser_mode       => XSD__STRICT,
		explicit_form     => undef,
		form              => $form,
		mixed             => 'false',
		nillable          => $nillable,
		namespace_uri     => $ns_info->{targetNamespace},
		ns_uri_id         => $das->{namespace_uris2id}->{$ns_info->{targetNamespace}},
	};
	++$xsd->{next_container_i};
}

sub handle_global_attribute_declaration {
	my ($das, $xsd, $twig) = @_;
	my $ns_info = get_namespace_information($twig);

	die schema_parsing_error_at($twig)."global attribute declaration must have a name\n" unless $twig->att('name');
	my $attribute_name = $twig->att('name');
	
	warn "meaningless 'form' attribute in global element definition\n" if $twig->att('form');
	my $form=XSDFORMILLEGAL;
	if ("" eq $ns_info->{+XSDTARGETNAMESPACE}) {
		$form=XSDUNQUALIFIED;
	} else {
		$form=XSDQUALIFIED;
	}
	
	my $namespace = '';
	$namespace=$ns_info->{targetNamespace} if $ns_info->{attributeFormDefault} eq XSDQUALIFIED;
	$namespace=$ns_info->{targetNamespace} if $form eq XSDQUALIFIED;
	$namespace='' if $form eq XSDUNQUALIFIED;
	
	$xsd->{attribute}->{$namespace}->{$attribute_name}->{twig}=\$twig;
}

sub handle_global_definition {
	my ($das, $xsd, $twig) = @_;
	my $tag = $twig->local_name;
	die schema_parsing_error_at($twig)."global/named ".$tag." definitions must include a name\n"
		unless defined $twig->att('name');
	$xsd->{$tag}->{get_target_namespace($twig)}->{$twig->att('name')}->{twig}=\$twig;
}

sub handle_named_simple_type {
	my ($das, $xsd, $twig)=@_;
	my $ns_info = get_namespace_information($twig);
	my $name = get_local_name($twig->att('name'));
							  
	my $simpleType = get_simpleType_from_twig_simple_type($das, $xsd, $twig, $name, $ns_info->{targetNamespace});
	$xsd->{simpleType}->{$ns_info->{targetNamespace}}->{$name} = $simpleType;
	return $simpleType;
}

sub get_schemaLocation_mappings_from_compose_ini {
	my $cwd = Cwd::getcwd;
	my $main_schema_dir = File::Basename::dirname(shift); # dirname of main schema document
	my $compose_ini = $main_schema_dir."/compose.ini"; # ini expected at same dir as main schema

	my $abs_base_dir = 
		URI->new_abs(
			URI->new($main_schema_dir,"file"),
			URI->new($cwd."/","file")
		)->as_string."/";
	$abs_base_dir =~ s,/+,/,g;

	open my $conf_fh, $compose_ini or return {};
	my $schemaLocation_mappings={};
	my $config;
	while (<$conf_fh>) { 
		$config->{$1} = $2
			if /^\s*([^=]+?)\s*=\s*(.*?)\s*$/;
	}
	
	foreach my $key (keys %{$config}) {
		next unless $key=~/^urn_(\S+)$/;
		$schemaLocation_mappings->{$config->{$&}} =
			URI->new_abs(
				URI->new($config->{"file_$1"},"file"), 
				$abs_base_dir
			)->as_string 
				if exists $config->{"file_$1"};
	}
	return $schemaLocation_mappings;
}

sub get_canonical_filename {
	my ($schema_document) = @_;
	return URI->new_abs(
			URI->new($schema_document,"file"),
			URI->new_abs(
				URI->new(Cwd::getcwd."/","file"),
				URI->new($schema_document,"file")
			)
		)->as_string;

}

sub get_context_as_string {
	my ($twig) = @_;
	map { $_->delete } $twig->children;
	return $twig->toString;
	
}

sub schema_parsing_error_at {
	my ($twig) = @_;
	return "Schema parsing error at: ".get_context_as_string($twig).": ";
}

sub handle_schema_composition { ## TODO: expand to support include and import similarly
	my ($das, $main_xsd, $main_twig, $schemaLocation_mappings) = @_;

	foreach my $child ($main_twig->root->children) {
		if (is_include($child) or is_redefine($child)) {
			if (not $child->att('schemaLocation')) {
#				warn get_context_as_string($child).": Must have schemaLocation attribute\n";
				next;
			}
			my $schemaLocation=$child->att('schemaLocation');
			$schemaLocation=$schemaLocation_mappings->{$schemaLocation}
				if exists $schemaLocation_mappings->{$schemaLocation};
			my $original_schema_location = $schemaLocation;
			$schemaLocation = 
				abs_schemaLocation_dir(
					$main_xsd->{current_schema_document_fullpath},
					$schemaLocation
				);
			$das->{self}->{original_schemaLocation_value}->{$schemaLocation} = $original_schema_location;
			if (exists $das->{dependency}->{$schemaLocation}->{include}->{get_target_namespace($main_twig)}) {
#				warn "Circular dependency -- skipping include of $schemaLocation while processing $main_xsd->{schemaLocation}\n";
				next;
			}
			$das->{dependency}->{$schemaLocation}->{include}->{get_target_namespace($main_twig)}++;
			# if processing same schema file (with a name different from seen name skip it
			my $schema_document_digest = get_md5_digest_of_schema_document($schemaLocation);
			if (exists $das->{digest}->{$schema_document_digest} and
				exists $das->{digest}->{$schema_document_digest}->{get_target_namespace($main_twig)}) {
				next;
			} else {
				$das->{digest}->{$schema_document_digest}->{get_target_namespace($main_twig)} = $schemaLocation;
			}
			my $twig=XML::Twig->new( %$twig_params, no_xxe => 1 );
			$das->{self}->parse_xsd($twig,$schemaLocation,$das) or return 1;
			sanity_check_targetNamespace_for_include($twig, $main_twig);
			my $xsd = new_xsd();
			$xsd->{schemaLocation}=$schemaLocation;
			# don't override twig if exists because it will blind
			# all children of that twig that were harvested before
			$das->{schema}->{$schemaLocation}->{get_target_namespace($main_twig)}->{twig} = $twig
				unless $das->{schema}->{$schemaLocation}->{get_target_namespace($main_twig)};
			my $include_twig = $das->{schema}->{$schemaLocation}->{get_target_namespace($main_twig)}->{twig};
			$include_twig->{twig_root}->set_att('targetNamespace' => $child->root->att('targetNamespace'))
				if $child->root->att('targetNamespace');
			$include_twig->{twig_root}->set_att('xmlns' => $child->root->att('targetNamespace'))
				unless $include_twig->{twig_root}->att('xmlns');
			$xsd->{current_schema_document_fullpath} = $schemaLocation;
			return 1 if handle_schema_composition($das, $xsd, $twig, $schemaLocation_mappings);
			#back from recursion:
			handle_top_level($das, $xsd, $twig->root);
			include_schema($das, $main_twig, $main_xsd, $twig, $xsd);
			if(is_redefine($child)) {
				handle_redefine_elements($das, $main_xsd, $child);
			}
		} elsif (is_import($child)) {
			my $schemaLocation = $child->att('schemaLocation');
			if(!$child->att('schemaLocation') && $child->att('namespace')){
#				warn "Schema Location not defined -- skipping import of ".$child->att('namespace')." while processing ".$main_xsd->{schemaLocation}."\n";
				next;
			}
			if ($child->att('namespace')) {
 				$das->{seen_import_namespace}->{$child->att('namespace')}++;
                                if($child->att('schemaLocation') =~ /wsdl_actions/) {
                                    $das->{seen_import_namespace}->{$child->att('namespace')}--; # wsdl converter created file doesn't count
                                }
				if ($das->{seen_import_namespace}->{$child->att('namespace')}>1) {
#					warn "already imported ".$child->att('namespace')."-- skipping ".get_context_as_string($child)."\n";
					next;
				}
			}
			$schemaLocation=$schemaLocation_mappings->{$schemaLocation}
				if exists $schemaLocation_mappings->{$schemaLocation};

			my $original_schema_location = $schemaLocation;
			$schemaLocation = 
				abs_schemaLocation_dir(
					$main_xsd->{current_schema_document_fullpath},
					$schemaLocation
				);
			$das->{self}->{original_schemaLocation_value}->{$schemaLocation} = $original_schema_location;
			if (exists $das->{dependency}->{$schemaLocation}->{import}) {
#				warn "Circular dependency -- skipping import of $schemaLocation while processing $main_xsd->{schemaLocation}\n";
				next;
			}
			
			$das->{dependency}->{$schemaLocation}->{import}++;
			# if processing same schema file (with a name different from seen name skip it
			my $schema_document_digest = get_md5_digest_of_schema_document($schemaLocation);
			my $imported_target_namespace = $child->att('namespace');
			my $twig=XML::Twig->new( %$twig_params, no_xxe => 1 );
			$das->{self}->parse_xsd($twig,$schemaLocation,$das) or return 1;
			sanity_check_targetNamespace_for_import($das, $main_twig, $twig->root, $child->att('namespace'));
			$imported_target_namespace = $twig->root->att('targetNamespace') unless $imported_target_namespace;
			$imported_target_namespace = "" if not defined $imported_target_namespace;
			if (exists $das->{digest} and
				exists $das->{digest}->{$schema_document_digest} and
				exists $das->{digest}->{$schema_document_digest}->{$imported_target_namespace}) {
				next;
			} else {
				$das->{digest}->{$schema_document_digest}->{$imported_target_namespace} = $schemaLocation;
			}
			my $xsd = new_xsd();
			$xsd->{schemaLocation}=$schemaLocation;
			my $ns_info = get_namespace_information($twig);
			# don't override twig if exists because it will blind
			# all children of that twig that were harvested before
			$child->set_att( 'namespace' => get_target_namespace($twig) ) unless $child->att('namespace');

			$das->{schema}->{$schemaLocation}->{$child->att('namespace') || get_target_namespace($twig)}->{twig} = $twig
				unless exists $das->{schema}->{$schemaLocation}->{$child->att('namespace') || get_target_namespace($twig)};
			$xsd->{current_schema_document_fullpath} =	$schemaLocation;
			return 1 if handle_schema_composition($das, $xsd, $twig, $schemaLocation_mappings);
			#back from recursion:
			handle_top_level($das, $xsd, $twig->root);
			import_schema($das, $main_twig, $main_xsd, $twig, $xsd);
		}
	}
	return 0;
}

sub include_schema { # merge included schema into including schema
	my ($das, $main_twig, $main_xsd, $twig, $xsd)=@_;
	my $element_ns;
	
	my $xsd_ns = get_target_namespace($twig);
	my $main_xsd_ns = get_target_namespace($main_twig);
	
	foreach my $namespace (keys %{$xsd->{element}}) {
		$element_ns = $namespace;
		$element_ns = $main_xsd_ns if(not $element_ns and not $xsd_ns);
		foreach my $top_level_elt (keys %{$xsd->{element}->{$namespace}}) {
			if (exists $main_xsd->{element}->{$element_ns}->{$top_level_elt}) {
				next;
			}
			$main_xsd->{next_container_i}++ if $main_xsd->{container}->[$main_xsd->{next_container_i}];
			$main_xsd->{element}->{$element_ns}->{$top_level_elt}->{id}   = $main_xsd->{next_container_i};
			$main_xsd->{element}->{$element_ns}->{$top_level_elt}->{twig} =
				$xsd->{element}->{$namespace}->{$top_level_elt}->{twig};

			move_container_include(
				$das,
				$xsd->{element}->{$namespace}->{$top_level_elt}->{id},
				$xsd,
				$main_xsd->{next_container_i},
				$main_xsd,
				$main_xsd_ns,
				-1,
				-1,
			);
		}
	}

	foreach my $namespace (keys %{$xsd->{attribute}}) {
		$element_ns = $namespace;
		$element_ns = $main_xsd_ns if(not $element_ns and not $xsd_ns);
		foreach my $item (keys %{$xsd->{attribute}->{$namespace}}) {
			if (exists $main_xsd->{attribute}->{$element_ns}->{$item}) {
				next;
			}
			$main_xsd->{attribute}->{$element_ns}->{$item}->{twig} =
				$xsd->{attribute}->{$namespace}->{$item}->{twig};
		}
	}

	foreach my $top_level ( qw/complexType group attributeGroup/ ) {
		foreach my $namespace (keys %{$xsd->{$top_level}}) {
			$element_ns = $namespace;
			$element_ns = $main_xsd_ns if(not $element_ns and not $xsd_ns);
			foreach my $item (keys %{$xsd->{$top_level}->{$namespace}}) {
				if (exists $main_xsd->{$top_level}->{$element_ns} and
					exists $main_xsd->{$top_level}->{$element_ns}->{$item}) {
					next;
				}
				$main_xsd->{$top_level}->{$element_ns}->{$item}->{twig} =
					$xsd->{$top_level}->{$namespace}->{$item}->{twig};
			}
		}
	}
	
	foreach my $namespace (keys %{$xsd->{substitutionGroup}}) {
		$element_ns = $namespace;
		$element_ns = $main_xsd_ns if(not $element_ns and not $xsd_ns);
		foreach my $item (keys %{$xsd->{substitutionGroup}->{$namespace}}) {
			foreach my $member (keys %{$xsd->{substitutionGroup}->{$namespace}->{$item}}) {
				$main_xsd->{substitutionGroup}->{$element_ns}->{$item}->{$member} =
					$xsd->{substitutionGroup}->{$namespace}->{$item}->{$member};
			}
		}
	}

	foreach my $namespace (keys %{$xsd->{simpleType}}) {
		next if $namespace eq +XMLSCHEMAURL;
		$element_ns = $namespace;
		$element_ns = $main_xsd_ns if(not $element_ns and not $xsd_ns);
		foreach my $item (keys %{$xsd->{simpleType}->{$namespace}}) {
			if (exists $main_xsd->{simpleType}->{$element_ns}->{$item}) {
				next;
			}
			# clone simple type from included schema to main schema
			if(exists $xsd->{simpleType}->{$element_ns}->{$item}->{twig}) {
				$main_xsd->{simpleType}->{$element_ns}->{$item}->{twig} =
					$xsd->{simpleType}->{$namespace}->{$item}->{twig};
			}
			else {	
				$main_xsd->{simpleType}->{$element_ns}->{$item} =
					Clone::clone($xsd->{simpleType}->{$namespace}->{$item});
			}
		}
	}
}

sub move_container_include {
	my ($das,
		$from_container_i,$from_xsd,
		$to_container_i,$to_xsd, $to_xsd_ns,
		$parent_container_id,$parent_element_id,
	) = @_;

	$to_xsd->{container}->[$to_container_i]=Clone::clone($from_xsd->{container}->[$from_container_i]);
	$to_xsd->{container}->[$to_container_i]->{id}=$to_container_i;
	$to_xsd->{container}->[$to_container_i]->{parent_id}=$parent_container_id;
	$to_xsd->{container}->[$to_container_i]->{element_parent_id}=$parent_element_id;
	$to_xsd->{container}->[$to_container_i]->{children}=[];
	$to_xsd->{container}->[$to_container_i]->{attributes}=[];
	foreach my $child_container_id (@{$from_xsd->{container}->[$from_container_i]->{children}}) {
		$to_xsd->{next_container_i}++
			if $to_xsd->{container}->[$to_xsd->{next_container_i}];
		push @{$to_xsd->{container}->[$to_container_i]->{children}},$to_xsd->{next_container_i};
		move_container_include($das,
			$child_container_id, $from_xsd,
			$to_xsd->{next_container_i}, $to_xsd, $to_xsd_ns,
			$to_container_i,
				$to_xsd->{container}->[$to_container_i]->{container_type} eq 'element' ? 
					$to_container_i:$parent_element_id,
		); 
	}
	foreach my $attribute_container_id (@{$from_xsd->{container}->[$from_container_i]->{attributes}}) {
		$to_xsd->{next_container_i}++
			if $to_xsd->{container}->[$to_xsd->{next_container_i}];
		push @{$to_xsd->{container}->[$to_container_i]->{attributes}},$to_xsd->{next_container_i};
		move_container_include($das,
			$attribute_container_id, $from_xsd,
			$to_xsd->{next_container_i}, $to_xsd, $to_xsd_ns,
			$to_container_i,
				$to_xsd->{container}->[$to_container_i]->{container_type} eq 'element' ?
					$to_container_i:$parent_element_id,
		); 
	}
	## target namespace update
	if (not $from_xsd->{container}->[$from_container_i]->{namespace_uri}) {
		$to_xsd->{container}->[$to_container_i]->{namespace_uri} = $to_xsd_ns;
		$to_xsd->{container}->[$to_container_i]->{ns_uri_id} = $das->{namespace_uris2id}->{$to_xsd_ns};
	}

	#top level names from included schema should become qualified if including schema has tns
	if (
		($to_xsd_ns ne '')  #including schema has target namespace
		and
		$to_xsd->{container}->[$to_container_i]->{container_type} eq 'element' 
		and
		exists $to_xsd->{element}->{$to_xsd_ns}->{$to_xsd->{container}->[$to_container_i]->{name}} #is top level element
	) {
		$to_xsd->{container}->[$to_container_i]->{form}=XSDQUALIFIED;
	}
}

sub import_schema {
	my ($das,$main_twig,$main_xsd,$twig,$xsd)=@_;
	my $main_xsd_ns = get_target_namespace($main_twig);

	foreach my $namespace (keys %{$xsd->{element}}) {
		foreach my $top_level_elt (keys %{$xsd->{element}->{$namespace}}) {
			if (exists $main_xsd->{element}->{$namespace}->{$top_level_elt}) {
				next;
			}
			$main_xsd->{next_container_i}++ 
				if $main_xsd->{container}->[$main_xsd->{next_container_i}];
			$main_xsd->{element}->{$namespace}->{$top_level_elt}->{id} = $main_xsd->{next_container_i};
			$main_xsd->{element}->{$namespace}->{$top_level_elt}->{twig} = $xsd->{element}->{$namespace}->{$top_level_elt}->{twig};
			move_container_import($das,
				$xsd->{element}->{$namespace}->{$top_level_elt}->{id},
				$xsd,
				$main_xsd->{next_container_i},
				$main_xsd,
				-1,
				-1,
				$namespace
			);
		}
	}

	foreach my $namespace (keys %{$xsd->{simpleType}}) {
		next if $namespace eq +XMLSCHEMAURL;
		foreach my $item (keys %{$xsd->{simpleType}->{$namespace}}) {
			if (exists $main_xsd->{simpleType}->{$namespace}->{$item}) {
				next;
			}
			# clone simple type from included schema to main schema
			if(exists $xsd->{simpleType}->{$namespace}->{$item}->{twig}) {
				$main_xsd->{simpleType}->{$namespace}->{$item}->{twig} = $xsd->{simpleType}->{$namespace}->{$item}->{twig};
			}
			else {	
				$main_xsd->{simpleType}->{$namespace}->{$item} = Clone::clone($xsd->{simpleType}->{$namespace}->{$item});
			}
		}
	}

	foreach my $namespace (keys %{$xsd->{substitutionGroup}}) {
		foreach my $item (keys %{$xsd->{substitutionGroup}->{$namespace}}) {
			foreach my $member (keys %{$xsd->{substitutionGroup}->{$namespace}->{$item}}) {
				$main_xsd->{substitutionGroup}->{$namespace}->{$item}->{$member} =
					$xsd->{substitutionGroup}->{$namespace}->{$item}->{$member};
			}
		}
	}
	
	foreach my $top_level ( qw/complexType attribute group attributeGroup/ ) {
		foreach my $namespace (keys %{$xsd->{$top_level}}) {
			foreach my $item (keys %{$xsd->{$top_level}->{$namespace}}) {
				if (exists $main_xsd->{$top_level}->{$namespace}->{$item}) {
					next;
				}
				$main_xsd->{$top_level}->{$namespace}->{$item}->{twig} = $xsd->{$top_level}->{$namespace}->{$item}->{twig};
				if($namespace eq "") {
					$main_xsd->{$top_level}->{$main_xsd_ns}->{$item}->{twig} = $xsd->{$top_level}->{$namespace}->{$item}->{twig};
				}
			}
		}
	}
}

sub move_container_import {
	my ( $das,
		$from_container_i,$from_xsd,
		$to_container_i,$to_xsd,
		$parent_container_id,$parent_element_id,
		$namespace
	) = @_;

	$to_xsd->{container}->[$to_container_i]=Clone::clone($from_xsd->{container}->[$from_container_i]);
	$to_xsd->{container}->[$to_container_i]->{id}=$to_container_i;
	$to_xsd->{container}->[$to_container_i]->{parent_id}=$parent_container_id;
	$to_xsd->{container}->[$to_container_i]->{element_parent_id}=$parent_element_id;
	$to_xsd->{container}->[$to_container_i]->{children}=[];
	$to_xsd->{container}->[$to_container_i]->{attributes}=[];
	foreach my $child_container_id (@{$from_xsd->{container}->[$from_container_i]->{children}}) {
		$to_xsd->{next_container_i}++
			if $to_xsd->{container}->[$to_xsd->{next_container_i}];
		push @{$to_xsd->{container}->[$to_container_i]->{children}},$to_xsd->{next_container_i};
		move_container_include($das,
			$child_container_id, $from_xsd,
			$to_xsd->{next_container_i}, $to_xsd, $namespace,
			$to_container_i,
				$to_xsd->{container}->[$to_container_i]->{container_type} eq 'element' ? 
					$to_container_i:$parent_element_id,
			$from_xsd->{container}->[$from_container_i]->{namespace_uri}
		);
	}
	foreach my $attribute_container_id (@{$from_xsd->{container}->[$from_container_i]->{attributes}}) {
		$to_xsd->{next_container_i}++
			if $to_xsd->{container}->[$to_xsd->{next_container_i}];
		push @{$to_xsd->{container}->[$to_container_i]->{attributes}},$to_xsd->{next_container_i};
		move_container_include($das,
			$attribute_container_id, $from_xsd,
			$to_xsd->{next_container_i}, $to_xsd, $namespace,
			$to_container_i,
				$to_xsd->{container}->[$to_container_i]->{container_type} eq 'element'? 
					$to_container_i:$parent_element_id,
			$from_xsd->{container}->[$from_container_i]->{namespace_uri}
		);
	}

	$to_xsd->{container}->[$to_container_i]->{namespace_uri} = $from_xsd->{container}->[$from_container_i]->{namespace_uri};
	$to_xsd->{container}->[$to_container_i]->{ns_uri_id} = $das->{namespace_uris2id}->{$from_xsd->{container}->[$from_container_i]->{namespace_uri}};

}

sub sanity_check_targetNamespace_for_include {
	my ($twig,$main_twig) = @_;
	die schema_parsing_error_at($twig->root)."conflicting target namespaces in includes\n"
		if($twig->root->att('targetNamespace') &&
		   $main_twig->root->att('targetNamespace') ne $twig->root->att('targetNamespace'));

	# returning means that either no targetNamespace in include or it is
	# similar to the targetNamespace of the "main" schema
}

sub sanity_check_targetNamespace_for_import {
	my ($das, $main_twig, $twig, $namespace) = @_;
	if ($namespace) {
		die schema_parsing_error_at($twig->root).
			"imported namespace must not be equal to the target namespace of the importing schema document\n"
				if $namespace eq get_target_namespace($main_twig);
		die schema_parsing_error_at($twig->root).
			"\n\ntarget namespace (".$twig->att('targetNamespace').") must be equal to the namespace attribute ($namespace) in import"
				if $twig->att('targetNamespace') and $namespace ne $twig->att('targetNamespace');
		if (not exists $das->{namespace_uris2id}->{$namespace}) {
			push @{$das->{namespace_uris}},$namespace;
			$das->{namespace_uris2id}->{$namespace}=@{$das->{namespace_uris}}-1;
		}
	} else {
		die schema_parsing_error_at($twig->root).
			"imported schema document must not have a target namespace if no namespace attribute in import"
				if $twig->att('targetNamespace');
		if (not exists $das->{namespace_uris2id}->{''}) {
			push @{$das->{namespace_uris}},'';
			$das->{namespace_uris2id}->{''}=@{$das->{namespace_uris}}-1;
		}
	}
}

sub extract_type_info {
	my ($xsd, $das, $twig, $att_name) = @_;
	
	return undef unless $twig->att($att_name);
	return extract_type_info_from_named_type($xsd, $das, $twig, $twig->att($att_name));
}

sub extract_type_info_from_named_type {
	my ($xsd,$das,$twig,$type_name) = @_;
	my $ns_info = get_namespace_information($twig);
	my $type_local_name = get_local_name($type_name);
	my $type_prefix     = get_prefix($type_name);
	my $type_namespace  = get_namespace_of_prefix($twig, $type_prefix);
	if($type_local_name =~ /anySimpleType/) {
		$type_local_name =~ s/anySimpleType/string/;
	}
	if(not $type_prefix) {
		$type_namespace = $ns_info->{defaultNamespaceURI};
		if(not defined $type_namespace) {
			$type_namespace = "";
		}
		
	}
	my $type_ns_uri_id  = $das->{namespace_uris2id}->{$type_namespace};

	if (not defined $type_ns_uri_id) {
		push @{$das->{namespace_uris}},$type_namespace;
		$das->{namespace_uris2id}->{$type_namespace} = @{$das->{namespace_uris}}-1;
	}
	
	my $type_info= {
		local_name => $type_local_name,
		prefix     => $type_prefix,
		namespace  => $type_namespace,
		ns_uri_id  => $type_ns_uri_id,
	};
	
	if (exists $xsd->{simpleType}->{$type_namespace}
	    and
	    exists $xsd->{simpleType}->{$type_namespace}->{$type_local_name}
	    and
	    exists $xsd->{simpleType}->{$type_namespace}->{$type_local_name}->{builtin_type}) {
		$type_info->{builtin_type} = $xsd->{simpleType}->{$type_namespace}->{$type_local_name}->{builtin_type};
	}
	
	return $type_info;
}

sub simpletype2key {
	my($builtin_type,$facet_id_list) = @_;
	my $r='';
	$r.=$builtin_type;
	$r.= '[';
	$r.=join('|',sort {$a<=>$b} @$facet_id_list);
	$r.=']';
	return $r;
}

sub metatype2key {
	my($is_list,$simpletype_id_list, $facets_id_list) = @_;
	if(defined $facets_id_list) {
		return
			($is_list?'true':'false') .
			'[' .
			join('|',sort {$a<=>$b} @$simpletype_id_list) .
			']' .
			'{' .
			join('|',sort {$a<=>$b} @$facets_id_list) .
			'}'
		;
	} else {
		return
			($is_list?'true':'false') .
			'[' .
			join('|',sort {$a<=>$b} @$simpletype_id_list) .
			']'
		;	
	}
}

sub add_namespace_unless_in_cache_and_get_namespace_id {
	my ($das, $namespace) = @_;
	my $schema_h = $F5::XML::Schema::SchemaDefinitions::schema_h;
	
	if (not exists $das->{namespace_uris2id}->{$namespace}) {
		push @{$das->{namespace_uris}},$namespace;
		$das->{namespace_uris2id}->{$namespace}=@{$das->{namespace_uris}}-1;
	}
	return $das->{namespace_uris2id}->{$namespace};
}

sub add_facet_unless_in_cache_and_get_facet_id {
	my ($das,$facet_name,$facet_value)=@_;

	if (exists $das->{facet2id}->{$facet_name}->{$facet_value}) {
		return $das->{facet2id}->{$facet_name}->{$facet_value};
	}

	my $facet_id = $das->{next_facet_i}++;
	$das->{facet}->[$facet_id] = {
		id    => $facet_id,
		type  => $facet_name,
		value => $facet_value,
	};
	$das->{facet2id}->{$facet_name}->{$facet_value}=$facet_id;

	return $facet_id;
}

sub process_simpleType_header {
	my ($das, $xsd, $name, $namespace) = @_;
	
	return -1 if(not $name or $name eq 'anySimpleType');
	
	return $das->{simpleType}->{$namespace}->{$name} if (defined $das->{simpleType}->{$namespace}->{$name});

	my $simpleType = $xsd->{simpleType}->{$namespace}->{$name};
	
	if(exists $simpleType->{twig}) {
		my $simpleTypeTwig = ${$simpleType->{twig}};
		$simpleType = handle_named_simple_type($das, $xsd, $simpleTypeTwig);
		$xsd->{simpleType}->{$namespace}->{$name} = $simpleType;
	}
	

	get_metatype_id_list_from_simpleType_template($das, $xsd, $simpleType);
	
	# simple type exist , not twig and do not have id
	
	my $stype_header_id = $das->{next_stype_header_i}++;
	
	my $stype_header_parent_id;
	if($name =~ /(.*)\@fixed/ or $name =~ /(.*)\@default/) {
		##
		# if current type was forged because of 'fixed' or 'default'
		# attribute we create special derivation where the original
		# type derive from current and current derive from base of original
		my $orig_id = process_simpleType_header($das, $xsd, $simpleType->{parent}, $simpleType->{parent_namespace});
		my $orig_stype_header = $das->{simpleType_ids}->[$orig_id];
		$stype_header_parent_id = $orig_stype_header->{parent_id};
		$orig_stype_header->{parent_id} = $stype_header_id;
	}
	elsif($simpleType->{parent_namespace} =~ /\@local.*/) {
		##
		# if base type for current simpleType was locally defined type
		# we will not create link to base local simpleType definition
		# because it can not be used in xsi:type parameter
		$stype_header_parent_id = -1;
	}
	else {
		$stype_header_parent_id =
			process_simpleType_header($das, $xsd, $simpleType->{parent}, $simpleType->{parent_namespace});
	}
	
	$das->{simpleType}->{$namespace}->{$name} = $stype_header_id;
	$das->{simpleType_ids}->[$stype_header_id] = {
		id           => $stype_header_id,
		name         => $name,
		namespace    => $namespace,
		namespace_id => $das->{namespace_uris2id}->{$namespace},
		metatype_id_list => [@{$xsd->{simpleType}->{$namespace}->{$name}->{metatype_id_list}}],
		parent_id => $stype_header_parent_id,
	};
	
	return $stype_header_id;
}

sub add_simpletype_unless_in_cache_and_get_simpletype_id {
	my ($das,$builtin_simple_type,$facet_ids)=@_;
	my $simpletype_key = simpletype2key($builtin_simple_type,$facet_ids);
	if (exists $das->{simpletype2id}->{$simpletype_key}) {
		return $das->{simpletype2id}->{$simpletype_key};
	}
	my $simpletype_id = $das->{next_simpletype_i}++;
	$das->{simpletype}->[$simpletype_id] = {
		id            => $simpletype_id,
		type          => $builtin_simple_type,
		facet_id_list => [@$facet_ids],
	};
	$das->{simpletype2id}->{$simpletype_key}=$simpletype_id;
	return $simpletype_id;
}

sub add_metatype_unless_in_cache_and_get_metatype_id {
	my ($das,$builtin_simple_type,$simpletype_ids,$facet_ids)=@_;

	my $is_list = 0;
	if ($builtin_simple_type=~/list|NMTOKENS|IDREFS|ENTITIES/ ){
		$is_list = 1 
	}

	my $metatype_key = metatype2key($is_list, $simpletype_ids, $facet_ids);

	if (exists $das->{metatype2id}->{$metatype_key}) {
		return $das->{metatype2id}->{$metatype_key};
	}

	my $metatype_id = $das->{next_metatype_i}++;
	
	if(defined $facet_ids) {
		$das->{metatype}->[$metatype_id] = {
			id                 => $metatype_id,
			is_list            => $is_list,
			simpletype_id_list => [@$simpletype_ids],
			facet_id_list      => [@$facet_ids],
		};
	}
	else {
		$das->{metatype}->[$metatype_id] = {
			id                 => $metatype_id,
			is_list            => $is_list,
			simpletype_id_list => [@$simpletype_ids],
		};		
	}
	
	$das->{metatype2id}->{$metatype_key}=$metatype_id;
	return $metatype_id;
}

## The simple type is defined in xsd->{simpleType}, i.e., has a template.
## So the purpose of this sub is to translate this to a list of metatype ids
sub get_metatype_id_list_from_simpleType_template {
	my ($das, $xsd, $simpleType) = @_;

	# process parent built-in type too
	if(defined $simpleType->{parent} and $simpleType->{parent} and $simpleType->{parent_namespace} ne "\@local" ) {
		my $parent_simpleType = $xsd->{simpleType}->{$simpleType->{parent_namespace}}->{$simpleType->{parent}};
		get_metatype_id_list_from_simpleType_template($das, $xsd, $parent_simpleType);
	}
	
	
	# this condition is expected to apply for all cases 
	# other than:
	# explicit built in simple types with no additional facets
	return $simpleType->{metatype_id_list}
		if $simpleType->{metatype_id_list};

	#
	# need to construct metatype(s) from template
	# 
	my $listSimpleType;
	if(not exists $simpleType->{is_atom}) {
		warn %{$simpleType};
	}
	if($simpleType->{is_atom} == 0) {
		$listSimpleType = $simpleType;
		$simpleType = $xsd->{simpleType}->{+XMLSCHEMAURL}->{$listSimpleType->{builtin_type}};
	}
	
	# facets
	my $facet_ids = get_facet_ids_from_simpleType($das,$simpleType);

	# simpletypes
	my @simpletype_ids=();
	push 
		@simpletype_ids, 
		add_simpletype_unless_in_cache_and_get_simpletype_id($das,$simpleType->{builtin_type},$facet_ids);


	# metatypes
	my @metatype_ids=();
	push 
		@metatype_ids,
		add_metatype_unless_in_cache_and_get_metatype_id($das,($listSimpleType)?'list':$simpleType->{builtin_type},\@simpletype_ids);
	
	# in case of builtin list
	if($listSimpleType) {
		# facets
		my $facet_ids = get_facet_ids_from_simpleType($das,$listSimpleType);
		$das->{metatype}->[$metatype_ids[0]]->{facet_id_list} = [@$facet_ids];
		$das->{metatype}->[$metatype_ids[0]]->{is_list} = 1;
		$simpleType = $listSimpleType;
	}
	
	$simpleType->{metatype_id_list} = [@metatype_ids];

	return $simpleType->{metatype_id_list};
}


sub get_facet_ids_from_simpleType {
	my ($das,$simpleType) = @_;	
	my @facet_ids=();
	if(defined $simpleType->{facet}->{+XSDWHITESPACE}) {
		push 
			@facet_ids,
			add_facet_unless_in_cache_and_get_facet_id($das, XSDWHITESPACE, $simpleType->{facet}->{+XSDWHITESPACE});
	}
	foreach my $facet (keys %{$simpleType->{facet}}) { #foreach possible facet
		next unless defined $simpleType->{facet}->{$facet}; #skip unrestricted facets
		next if XSDWHITESPACE eq $facet; #skip whitespace facet that was added early
		if(XSDENUMERATION eq $facet) {
			foreach my $enum_facet (keys %{$simpleType->{facet}->{$facet}}) {
				push 
					@facet_ids,
					add_facet_unless_in_cache_and_get_facet_id($das,$facet, $enum_facet);
			}
		} else {
				push 
					@facet_ids,
					add_facet_unless_in_cache_and_get_facet_id($das,$facet, $simpleType->{facet}->{$facet});
		}
	}
	return \@facet_ids;
}

sub get_xsd_children {
	my ($xsd,$twig,$pattern) = @_;
	my @children=();
	foreach my $child ($twig->children) {
		next unless $child->local_name =~/$pattern/;
		my $namespace = get_namespace_of_prefix($child, get_prefix($child->gi));
		next unless $namespace eq XMLSCHEMAURL;
		push @children,$child;
	}
	return @children;
}

sub get_namespace_of_prefix {
	my ($twig, $prefix) = @_;
	return XMLSCHEMAURL if $prefix eq 'xsd' && not defined $twig;
	return '' unless defined $twig;
	my $xmlns = 'xmlns';
	if (
		defined $prefix and
		$prefix ne ''   and 
		$prefix ne '#default' # deal with XML::Parser::Expat conventions
	) {
		$prefix = $twig->{att}->{'#original_gi'}->{$prefix} if(exists $twig->{att}->{'#original_gi'} and exists $twig->{att}->{'#original_gi'}->{$prefix});
		
		$xmlns .= ':'.$prefix unless $prefix eq '#default';
	}
	return $twig->att($xmlns) if defined $twig->att($xmlns);
	if($twig eq $twig->root){
		if($twig->att('targetNamespace')) {
			return $twig->att('targetNamespace');
		}
		else {
			return '';
		}
	}
	return get_namespace_of_prefix($twig->parent,$prefix);
}

# get twig of attribute
# return simpleType structure, which have ->{metatype_ids_list}
sub get_attribute_simple_type {
	my ($das, $xsd, $twig, $id) = @_;
	my $type_info;
	my $simpleType;
	
	## if attribute refference - extract type info from reffered attribute
    if($twig->att('ref')) {
		my $attribute_twig;
        $type_info = extract_type_info($xsd, $das, $twig, 'ref');
		if(exists $xsd->{attribute}->{$type_info->{namespace}}
		   and
		   exists $xsd->{attribute}->{$type_info->{namespace}}->{$type_info->{local_name}}) {
			$attribute_twig = ${$xsd->{attribute}->{$type_info->{namespace}}->{$type_info->{local_name}}->{twig}};
			## call processing of simple type from reffered attribute
			$simpleType = get_attribute_simple_type($das, $xsd, $attribute_twig, $id);
			}else{
				die schema_parsing_error_at($twig) .  " unknown refferenced attribute\n";
			}
		}
	elsif($twig->att('type')) {
		$type_info = extract_type_info($xsd, $das, $twig, 'type');
		$simpleType = get_named_simple_type($das, $xsd, $twig, $type_info);
	}
	else {
		my @simpleType_children = get_xsd_children($xsd,$twig,'simpleType');
		if(scalar @simpleType_children > 0) {
			my $namespace = get_namespace_information($twig);
			my $name = get_local_simpleType_name($das, $xsd, $twig->att('name'), "\@local{$namespace}", $id);
			$simpleType = get_simpleType_from_twig_simple_type($das, $xsd, $simpleType_children[0], $name, "\@local");
		}
		else {
			$simpleType = $xsd->{simpleType}->{+XMLSCHEMAURL}->{anySimpleType};
		}
	}
	if(not exists $xsd->{simpleType}->{$simpleType->{namespace_uri}}->{$simpleType->{name}}) {
		$xsd->{simpleType}->{$simpleType->{namespace_uri}}->{$simpleType->{name}} = $simpleType;
	}
	return $simpleType;
}

sub get_simpleType_from_twig_simple_type {
	
	my ($das, $xsd, $twig, $name, $namespace) = @_;
	my $simpleType;
	

	my @children = get_xsd_children($xsd,$twig,'restriction|list|union');

	die schema_parsing_error_at($twig) . " missing expected content\n"
		unless @children>0;

	my $child = $children[0];

	# the functionality still lacks a lot of cases
	if(is_restriction($child)) {
		$simpleType = get_simpleType_from_twig_restriction($das, $xsd, $child, $name, $namespace);
	}
	elsif(is_list($child)) {
		$simpleType = get_simpleType_from_twig_list($das, $xsd, $child, $name, $namespace);
	}
	elsif(is_union($child)) {
		$simpleType = get_simpleType_from_twig_union($das, $xsd, $child, $name, $namespace);
	}
	
	
	return $simpleType;

}

## the assumtion is that the base type should have already been added as a
## simpleType template
## if the base is a builtin type -- it is already prepared
## if the base type is of a named simple type then is should already have been
## processed in the handling of global definitions
## if the base is of a parent anonymous simpletype then the parent was already
## processed
sub get_base_info {
	my ($das, $xsd, $twig, $name, $namespace)=@_;
	my $simpleType;
	if ($twig->att('base')) {
		my $type_info = extract_type_info($xsd, $das, $twig, 'base');
		$simpleType = get_named_simple_type($das, $xsd, $twig, $type_info);
	} else {
		my @simpleType_children = get_xsd_children($xsd, $twig, 'simpleType');
		die schema_parsing_error_at($twig) . " missing simple type definition\n"
			unless @simpleType_children>0;
		$simpleType = get_simpleType_from_twig_simple_type($das, $xsd, $simpleType_children[0], $name."_base", $namespace);
		$xsd->{simpleType}->{$namespace}->{$name."_base"} = $simpleType;
	}
	return $simpleType;
}


sub get_named_xsd_child {
	my ($twig, $tag_name, $att_name, $att_value) = @_;
	foreach my $child ($twig->root->children()) {
		next unless $tag_name eq $child->local_name;
		return $child if($child->att($att_name) eq $att_value);
	}
}

sub get_named_simple_type {
	my ($das, $xsd, $twig, $type_info) = @_;
	my $simpleType;
	my $simpleTypeTwig;
	
	if(exists $xsd->{simpleType}->{$type_info->{namespace}}
	   and
	   exists $xsd->{simpleType}->{$type_info->{namespace}}->{$type_info->{local_name}}) {
		$simpleType=$xsd->{simpleType}->{$type_info->{namespace}}->{$type_info->{local_name}};
	} else {
		die schema_parsing_error_at($twig) .  " unknown base simple type\n";
	}
	
	if(exists $simpleType->{twig}) {
		my $simpleTypeTwig = ${$simpleType->{twig}};
		$simpleType = handle_named_simple_type($das, $xsd, $simpleTypeTwig);
	}
	
	# check/process metatype ids list for found simpleType
	get_metatype_id_list_from_simpleType_template($das, $xsd, $simpleType);

	return $simpleType;
}

sub consolidate_facets {
	my ($xsd, $das, $builtin_type, $base_facet_id_list, $facets)=@_;
	my %facet_ids;
	
	#hash new facets for easy lookup
	my $new_facets;
	foreach my $f (@$facets) {
		push @{$new_facets->{$f->{facet_name}}},$f->{facet_value};
	}
	
	# hash of enumeration facets ids of base type
	# we will use it to check restriction of derived type enumeration facets
	my $base_enumeration_ids;
	foreach my $id (@$base_facet_id_list) {
		my $base_facet = $das->{facet}->[$id];
		$base_enumeration_ids->{$id}++ if ('enumeration' eq $base_facet->{type});
	}
	
	foreach my $id (@$base_facet_id_list) {
		my $base_facet = $das->{facet}->[$id];

		if(exists $new_facets->{$base_facet->{type}}) {

			my $facet_name = $base_facet->{type};
			foreach my $facet_value (@{$new_facets->{$base_facet->{type}}}){
				if ($facet_name=~/minInclusive|minExclusive|minLength/) {
					$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$facet_name,$facet_value)}++;
				}elsif ($facet_name=~/maxInclusive|maxExclusive/) {
					$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$facet_name,$facet_value)}++;
				}elsif ($facet_name=~/length|maxLength|totalDigits|fractionDigits/) {
					if ($facet_value<0) {
						die "Encounter illegal value for facet (must be positive) " .
							"$facet_name = '$facet_value' ... \n";					
					}elsif ($facet_value<=$base_facet->{value}) {
						$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$facet_name,$facet_value)}++;
					}else{
						die "facet $facet_name = '$facet_value' was not restricting\n";
					}
				}elsif(XSDENUMERATION eq $facet_name) { # only new more restrictive
					#die "Facet '$facet_name' is illegal for $builtin_type\n"
					if('boolean' eq $builtin_type) {
						if($facet_value !~ /true|false|1|0/) {
							die "Enumeration facet '$facet_value' is illegal for $builtin_type\n"
						}
					}
					# when in base type exists enumeration facets
					# only new facets, which exist in base facets list included in resulting facets list
					if (exists $das->{facet2id}->{$facet_name}->{$facet_value}
					    and
					    exists $base_enumeration_ids->{$das->{facet2id}->{$facet_name}->{$facet_value}}){
						$facet_ids{$das->{facet2id}->{$facet_name}->{$facet_value}}++;
					} else {
##						# new facet does not exist in base enumeration facets
##						warn "Illegal attempt to extend an enumeration facets with " .
##							"$facet_name = '$facet_value'\n";
					}
					
					##$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$facet_name,$facet_value)}++;
				}elsif($facet_name eq 'pattern') { # both and parser with AND them
					$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$facet_name,$facet_value)}++;
					$facet_ids{$base_facet->{id}}++;
				}elsif($facet_name eq 'whiteSpace') { 
					my $whitespace_facet = get_consolidated_whitespace_facet($builtin_type,
												 $base_facet->{value},
												 $facet_value);
					$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$facet_name,$whitespace_facet)}++;
				}else{
					die "Unsupported facet ".$facet_name."\n";
				}
			}

			##delete $new_facets->{$base_facet->{type}};
		}else{
		    ##if(XSDENUMERATION ne $base_facet->{type}) {
			$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$base_facet->{type},$base_facet->{value})}++;
		    ##}
		}
	}
	my ($k, $v);
	foreach $k (keys %$new_facets) { 
		foreach my $v (@{$new_facets->{$k}}) {
			$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$k,$v)}++;
		}
	}
	return [map {int $_} keys %facet_ids]; # return list of indices
}

##
# whitespace facet fixed for most builtin types
# whitespace facet alowed to change only for 'string' based types
sub get_consolidated_whitespace_facet {
	my ($builtin_type, $old_facet_value, $new_facet_value) = @_;
	
	return $old_facet_value if($builtin_type !~ /^(string|normalizedString)$/);
	return $new_facet_value;
}
	
sub fix_int_type_range_facets {
	my ($facet_name, $facet_value) = @_;
	return ('minInclusive', ++$facet_value ) if($facet_name eq 'minExclusive');
	return ('maxInclusive', --$facet_value ) if($facet_name eq 'maxExclusive');
	return ($facet_name, $facet_value);
	# TODO may be we need to do sanity checks here
}

sub process_optional_flag_for_all_containers {
	my ($xsd) =@_;
	
	foreach my $cont (@{$xsd->{container}}) {
		if(not exists $cont->{is_optional}) { 
			$cont->{is_optional} = get_optional_flag_for_container($xsd, $cont);
		}
	}
}

sub get_optional_flag_for_container {
	my ($xsd, $cont) = @_;
	my $is_optional;

	if(not exists $cont->{is_optional}) {
		if($cont->{min_self_count} == 0) {
			return 1;
		}
		else { ## min occure not 0
			if($cont->{container_type} eq 'element') {
				return 0;
			}
			if($cont->{num_children} == 0) {
				return 0;
			}
			
			if($cont->{container_type} ne 'choice') {
				$is_optional = 1;
			}
			else {
				$is_optional = 0;
			}
			
			foreach my $child (@{$cont->{children}}) {
				if($cont->{container_type} ne 'choice') {
					$is_optional = ($is_optional and get_optional_flag_for_container($xsd, $xsd->{container}->[$child]));
				}
				else {
					$is_optional = ($is_optional or get_optional_flag_for_container($xsd, $xsd->{container}->[$child]));
				}
			}
		}
	}
	else {
		$is_optional = $cont->{is_optional};
	}
	
	return $is_optional;
}

sub get_consolidated_metatype_id_list {
	my ($das, $xsd, $twig, $base_metatype_id_list, $builtin_type)=@_;

	my $facets = [];
	my @patterns = ();
	foreach my $child (get_xsd_children($xsd,$twig,'minExclusive|maxExclusive|minInclusive|maxInclusive|length|minLength|maxLength|totalDigits|fractionDigits|enumeration|pattern|whiteSpace')) {
		my $facet_name = $child->local_name;
		my $facet_value = $child->att('value');
		($facet_name, $facet_value) = fix_int_type_range_facets($facet_name, $facet_value)
			if($builtin_type eq 'integer');
		if($facet_name eq "pattern") {
			push @patterns, "($facet_value)";
		}
		else {
			push @$facets, { 
				facet_name => $facet_name,
				facet_value => $facet_value,
			};
		}
	}
	if(scalar @patterns > 0) {
		push @$facets, { 
			facet_name => "pattern",
			facet_value => join( '|', @patterns),
		};
	}

	# handle facets on base type that is a union or a list
	#
	if($builtin_type =~ /union|list/) {
		my $simpletype_id_list;
		my $metatype_id_list;
		foreach my $metatype_id (@{$base_metatype_id_list}) {
			if($das->{metatype}->[$metatype_id]->{is_list}) {
				# handle facets on base type that is a list
				my %facet_ids;
				foreach my $f (@$facets) {
					# TODO pront warning message for illegal restriction facet on list
					next unless $f->{facet_name} =~ /length|minLength|maxLength|enumeration|pattern/;
					$facet_ids{add_facet_unless_in_cache_and_get_facet_id($das,$f->{facet_name},$f->{facet_value})}++;
				}
				if(scalar(keys %facet_ids) > 0) {
					$metatype_id = add_metatype_unless_in_cache_and_get_metatype_id($das,
																					'list',
																					$das->{metatype}->[$metatype_id]->{simpletype_id_list},
																					[map {int $_} keys %facet_ids]);
				}
				#$das->{metatype}->[$metatype_id]->{facet_id_list} = [map {int $_} keys %facet_ids];
				$metatype_id_list->{$metatype_id}++;
			} else {
				# handle facets on base type that is a union
				foreach my $simpletype_id (@{$das->{metatype}->[$metatype_id]->{simpletype_id_list}}) {
					my $base_facet_id_list = $das->{simpletype}->[$simpletype_id]->{facet_id_list};
					my $simpletype_id = get_simpletype_id_with_consolidated_facets(
							$xsd,
							$das,
							$das->{simpletype}->[$simpletype_id]->{type},
							$base_facet_id_list,
							$facets);
					$simpletype_id_list->{$simpletype_id}++;
				}
				$metatype_id_list->{add_metatype_unless_in_cache_and_get_metatype_id($das,"",[keys %{$simpletype_id_list}])}++;
			}
		}
		return [keys %{$metatype_id_list}];
	}
	
	# TODO
	# handle only one bound-range and not several
	# need to support more than one
	#
	
	# TODO
	# need to take into consideration "fixed" facets
	else {
		my $base_simpletype_id_list = $das->{metatype}->[$base_metatype_id_list->[0]]->{simpletype_id_list};
		my $base_facet_id_list = $das->{simpletype}->[$base_simpletype_id_list->[0]]->{facet_id_list};
	
		my $simpletype_id = get_simpletype_id_with_consolidated_facets($xsd,$das,$builtin_type, $base_facet_id_list,$facets);
	
		# TODO: there should be, in principal, a list of metatypes
		my $metatype_id = add_metatype_unless_in_cache_and_get_metatype_id(
					$das,
	#				$das->{simpletype}->[$base_simpletype_id_list->[0]]->{builtin_type},[$simpletype_id]);
					# TODO: note hard-coded 'string'
					$builtin_type || 'string',[$simpletype_id]);
		
		return [$metatype_id];
	}
}

sub get_simpletype_id_with_consolidated_facets {
	my ($xsd, $das, $builtin_type, $base_facet_id_list, $facets) = @_;
	
	my $facet_id_list = consolidate_facets($xsd, $das, $builtin_type, $base_facet_id_list, $facets);

	# make sure that whiteSpace facet, if exists, appears first
	for(my $i=0;$i<@$facet_id_list;++$i){
		my $fid=$facet_id_list->[$i];
		if ($das->{facet}->[$fid]->{type} eq 'whiteSpace') {
			($facet_id_list->[$i],$facet_id_list->[0])=($facet_id_list->[0],$facet_id_list->[$i]);
		}
	}

	return add_simpletype_unless_in_cache_and_get_simpletype_id($das,
		$builtin_type || 'string',$facet_id_list);

}

##
# process xsd:restriction and create
# simpleType structure of derived type
#
sub get_simpleType_from_twig_restriction {
	my ($das, $xsd, $twig, $name, $namespace)=@_;
	my $simpleType;

	my $base_info = get_base_info($das, $xsd, $twig, $name, $namespace);
	my $base_metatype_id_list = $base_info->{metatype_id_list};
	my $builtin_type = $base_info->{builtin_type};
	
	$simpleType = {
		name             => $name,
		builtin_type     => $builtin_type,
		namespace_uri    => $namespace,
		parent           => $base_info->{name},
		parent_namespace => $base_info->{namespace_uri},
		metatype_id_list => get_consolidated_metatype_id_list($das, $xsd, $twig, $base_metatype_id_list, $builtin_type),
	};
	
	return $simpleType;
}

##
# process xsd:list and create
# simpleType structure of list type
#
sub get_simpleType_from_twig_list {
	my ($das, $xsd, $twig, $name, $namespace) = @_;
	my $simpleType;
	my $metatype_id_list;

	if($twig->att('itemType')) {
		my $member_type = $twig->att('itemType');
		if($member_type =~ /anySimpleType/) {
			$member_type =~ s/anySimpleType/string/;
		}
		my $type_info = extract_type_info_from_named_type($xsd, $das, $twig, $member_type);
		my $simpleType = get_named_simple_type($das, $xsd, $twig, $type_info);
		foreach my $metatype_id (@{$simpleType->{metatype_id_list}}) {
			$metatype_id_list->{$metatype_id}++;
		}
	} else {
		# process body of list tag
		my @simpleType_children = get_xsd_children($xsd,$twig,'simpleType');
		
		if(scalar(@simpleType_children) == 0) {
			die schema_parsing_error_at($twig)."Empty List definition not allowed\n";
		}
		
		my $anonimous_metatype_id_list;
		$simpleType = get_simpleType_from_twig_simple_type($das,
														   $xsd,
														   $simpleType_children[0],
														   $name."_itemType",
														   "\@local");
		foreach my $metatype_id (@{$simpleType->{metatype_id_list}}) {
			$anonimous_metatype_id_list->{$metatype_id}++;
		}
			
		foreach my $metatype_id (keys %{$anonimous_metatype_id_list}) {
			$metatype_id_list->{$metatype_id}++;
		}
	}
	
	# simplify list metatype ids list
	# if the schema is correct here should be only 1 metatype !!
	# otherwise we have list of lists which is prohibited
	my $new_metatype_id_list;
	foreach my $metatype_id (keys %{$metatype_id_list}) {
		if($das->{metatype}->[$metatype_id]->{is_list}) {
			die schema_parsing_error_at($twig)."List of lists not allowed\n";
		} else {
			my $list_metatype_id = add_metatype_unless_in_cache_and_get_metatype_id($das, "list", $das->{metatype}->[$metatype_id]->{simpletype_id_list});
			$das->{metatype}->[$list_metatype_id]->{is_list} = 1;
			$simpleType = {
				name             => $name,
				builtin_type     => "list",
				namespace_uri    => $namespace,
				parent           => '', # list is not derivation of its item type
				parent_namespace => '', # unable to use it in xsi:type as substitute from its item type.
				metatype_id_list => [($list_metatype_id)],
			};

			return $simpleType;
		}
	}
	
	# should not get here
	die schema_parsing_error_at($twig)."Incorrect list definition.\n";
}

##
# process xsd:union and create
# simpleType structure of union type
#
sub get_simpleType_from_twig_union {
	my ($das, $xsd, $twig, $name, $namespace)=@_;
	my $simpleType;
	my $metatype_id_list;
	
	if($twig->att('memberTypes')) {
		my @member_types = split ' ', $twig->att('memberTypes');
		foreach my $member_type (@member_types) {
			my $type_info = extract_type_info_from_named_type($xsd, $das, $twig, $member_type);
			my $simpleType = get_named_simple_type($das, $xsd, $twig, $type_info);
			foreach my $metatype_id (@{$simpleType->{metatype_id_list}}) {
				$metatype_id_list->{$metatype_id}++;
			}
		}
	}
	
	# process body of union tag
	my @simpleType_children = get_xsd_children($xsd,$twig,'simpleType');
	my $anonimous_metatype_id_list;
	foreach my $simpleType_child (@simpleType_children) {
		$simpleType = get_simpleType_from_twig_simple_type($das, $xsd, $simpleType_child, $name."_memberType", "\@local");
		foreach my $metatype_id (@{$simpleType->{metatype_id_list}}) {
			$anonimous_metatype_id_list->{$metatype_id}++;
		}
	}
	
	foreach my $metatype_id (keys %{$anonimous_metatype_id_list}) {
		$metatype_id_list->{$metatype_id}++;
	}
	
	# simplify union metatype ids list
	my $new_metatype_id_list;
	my $simpletype_id_list;
	foreach my $metatype_id (keys %{$metatype_id_list}) {
		if($das->{metatype}->[$metatype_id]->{is_list}) {
			# this is list metatype - can't simplify it
			$new_metatype_id_list->{$metatype_id}++;
		} else {
			foreach my $simpletype_id (@{$das->{metatype}->[$metatype_id]->{simpletype_id_list}}) {
				$simpletype_id_list->{$simpletype_id}++;
			}
		}
	}
	if(scalar keys %{$simpletype_id_list} > 0) {
		$new_metatype_id_list->{add_metatype_unless_in_cache_and_get_metatype_id($das,"",[keys %{$simpletype_id_list}])}++;
	}
	
	$simpleType = {
		name             => $name,
		builtin_type     => "union",
		namespace_uri    => $namespace,
		parent           => '', # list is not derivation of its item type
		parent_namespace => '', # unable to use it in xsi:type as substitute from its item type.
		metatype_id_list => [map {int $_} keys %{$new_metatype_id_list}],
	};
	
	return $simpleType;
}

# checks if the element is of type:
# 	* simpleType
# 	* complexType
# 	* anyType
sub get_element_type_info {
	my ($das, $xsd, $twig, $id)=@_;
	
	if ($twig->att('type')) {
		my $type_info = extract_type_info($xsd, $das, $twig, 'type');
		return { type => 'anyType', }
			if (($type_info->{namespace} eq XMLSCHEMAURL) and ($type_info->{local_name} eq 'anyType'));
		if ($xsd->{simpleType}->{$type_info->{namespace}}
		    and
		    $xsd->{simpleType}->{$type_info->{namespace}}->{$type_info->{local_name}}) {
			my $simpleType = get_named_simple_type($das, $xsd, $twig, $type_info);
			return {
				type => 'simpleType',
				att_type => 1,
				simpleType => $simpleType,
				local_name => $type_info->{local_name},
				namespace => $type_info->{namespace},
			};
		} elsif ($xsd->{complexType}->{$type_info->{namespace}}->{$type_info->{local_name}}) {
			# the complexType must have been already defined!!
			return {
			 	type => 'complexType',
				att_type => 1,
			 	complexType => 1, 
			 	local_name => $type_info->{local_name},
			 	namespace => $type_info->{namespace},
			};
                } else {
			die schema_parsing_error_at($twig) . " unknown type for element\n";
		}
	}elsif(my @simpleType_children = get_xsd_children($xsd,$twig,'simpleType')){
		die schema_parsing_error_at($twig) . " missing simple type definition\n"
			unless @simpleType_children>0;
		my $namespace = get_namespace_information($twig);
		my $name = get_local_simpleType_name($das, $xsd, $twig->att('name'), "\@local{$namespace}", $id);
		my $simpleType = get_simpleType_from_twig_simple_type(
				$das, $xsd, $simpleType_children[0], $name, "\@local");
		$xsd->{simpleType}->{"\@local"}->{$name} = $simpleType;
		return {
			type => 'simpleType',
			att_type => 0,
			simpleType => $simpleType,
		};
	}elsif(my @complexType_children = get_xsd_children($xsd,$twig,'complexType')){
		die schema_parsing_error_at($twig) . " missing complex type definition\n"
			unless @complexType_children>0;
		return {
			type => 'complexType',
			att_type => 0,
			complexType_twig => 1, 
		};
	}elsif($twig->att('substitutionGroup')){
		# top element without type defined, but member of substitution group
		# take its type from head of substitution group
		my $info = extract_type_info($xsd, $das, $twig, 'substitutionGroup');

		if($xsd->{substitutionGroup}->{$info->{namespace}}->{$info->{local_name}}
		   and
		   $xsd->{element}->{$info->{namespace}}
		   and
		   $xsd->{element}->{$info->{namespace}}->{$info->{local_name}}) {
			my $head_twig  = ${$xsd->{element}->{$info->{namespace}}->{$info->{local_name}}->{twig}};
			return get_element_type_info ($das, $xsd, $head_twig, $id);
		}
		else {
			return { type => 'anyType', };
		}
	}else{
		return { type => 'anyType', };
	}
}


######
## Below is the OO-ish part of this module.
## All methods that are to be part of the API should be
## documented in the pod!
######

=pod

=head1 NAME

F5::XML::Schema::Processor - a Perl module for processing XML Schema documents
into a data structure that is consumed by F5's XML Instance processor.

=head1 SYNOPSIS

	use F5::XML::Schema::Processor;

	my $xsp = F5::XML::Schema::Processor->new(
		allow_external_references => 0, 
		schema_document => "/path/to/schema/document.xsd",
		schema_composition_mappings => {
			"schema location URN" => "/path/to/file/on/disk",
			# ...
		},
		blob_fh => $an_instantiated_IO_FILE_file_handle,
	);


	eval {
		$xsp->xparse;
	};

	die "\n\nException caught by schema processor:\n$@\n" if $@;

	if (not $xsp->OK) {
		my $error_message = "\n\nProblems parsing XML Schema document: ".$ARGV[0]."\n";
		$error_message .= join("\n",$xsp->errors);
		$error_message .= "Missing files: ".join(", ",$xsp->missing_files)."\n"
			if $xsp->missing_files;
		die $error_message;
	}

	$xsp->print_binary;

=head1 DESCRIPTION

F5::XML::Schema::Processor provides an OO-ish interface for feeding
configuration into the Schema Compiler, controlling the output stream (via a
file handle) where the output binary is to be printed into, and querying for
missing schema compisition files/mappings and other errors.

Errors are usually to be caught via trapping exceptions using eval(). The more
organized the error handling becomes the less errors will require exception
handling and will be handled via the C<< $xsp->OK >> and C<< $xsp->errors >>
methods.

=head1 METHODS

=head2 new

The constructor is used for generating a new object,
initializing its attributes.

It returns an instance of F5::XML::Schema::Processor.

Example usage: 

=begin

	my $xsp = F5::XML::Schema::Processor->new();

=end

An $xsp object can be instantiated with any number of the available named
parameters, including none. Any required parameter can be mutated and accessed
anytime using accessors and mutators of the same name (see below in the METHODS
section).

Available named parameters (via the hash argument) are:

=head3 allow_external_references

This argument takes a true or false value. It is recommended to use 0 for false
and 1 for true, but any other value that has the Perl semantics of true and
false will probably also work.

The meaning of this argument is whether or not to allow the Schema Processor to
access resources (e.g., schema documents) that were not explicitly available in
the schema document itself. As for schema composition, i.e., schema documents
that are referenced via I<xsd:import>, I<xsd:include> and I<xsd:redefine>, those will be
accessed if allow_external_references is true. If allow_external_references is
false the additional schemas will be referenced only if a mapping from
I<xsi:schemaLocation> to a local filename is provided via the
C<schema_composition_mappings> argument (see below).

=head3 schema_document

This is the representation of the XML Schema Document.  The representation can
be a full path on the disk or a URL (risky, as no one guarantees that the file
is accessible and moreover, this is a security problem). To fully understand
the implications of the representation see C<XML::Twig>'s documentation on its
C<xparse> method.

=head3 schema_composition_mappings 

This is a hash reference that maps I<xsi:schemaLocation> or I<namespace>
attributes of I<xsd:import>, I<xsd:redefine> and I<xsd:include> into a full
path filename of the schema document that should be available on disk and
accessible to the schema processor.

If C<allow_external_references> is false and the schema processor is expected
to provide schema composition functionality, i.e., follow and process
additional schema documents that are references from a schema document (this
can happen recursively) then every additional schema document should be mapped
using this hash or otherwise it will not be recognized by the schema processor.

=head3 blob_fh

Thie is a file handle scalar that should be instantiated and valid and will be
used by the schema processor as the output stream into which the binary output
will be written.

C<STDOUT> will be used if no such file handle is supplied.

=head3 debug

The expected value is of the form C<--nobin> or C<--nobin:foo> or
C<--nobin:foo:foo1> where C<foo> is a hash key of the C<$xsd> data structure and
C<foo1> is a hash key of the C<$xsd->{foo}> data structure. This argument should
be used by authors and maintainers of this module, and them alone.
C<foo> and C<foo1> can be equal 'list' in order to get listing of C<$xsd> or
C<$xsd->{foo}> hash members. If you don't know what you're doing, you shouldn't
use this argument!

Debug information goes to the standard output.

=head3 errors

A list reference containing errors. This shouldn't be used unless the user
wants to add error messages to the error messages stack. This is yet another
argument best used by authors and maintainers of this module. If you don't know
what you're doing, you shouldn't use this argument!

=cut

sub new {
	my ($class, %arguments) = @_;
	my $follow_schema_links = 1;
	if (defined $arguments{allow_external_references} and 0==$arguments{allow_external_references}) {
		$follow_schema_links=0;
	}

	bless {
		allow_external_references => $follow_schema_links,
		schema_document => $arguments{schema_document},
		schema_composition_mappings => $arguments{schema_composition_mappings},
		blob_fh => $arguments{blob_fh},
		debug => $arguments{debug},
		errors => $arguments{errors} || [],
		missing_files => [],
	}, $class;
}

=pod

=head2 xparse

This method does not expect any arguments.

This method throws exceptions (i.e., C<die>).

Example usage:

=begin

	eval {
		$xsp->xparse;
	};

	die "\n\nException caught by schema processor:\n$@\n" if $@;

=end

This method is where all the fun happens: it actually performs the schema
processing and the generation of binary representation for the consumption of
the XML Instance Processor.

Non fatal errors are to be pushed into an errors stack which is accessible via
the C< $xsp->errors > method. After execution of the C<< $xsd->xparse >> method
probine for errors can be done by checking if there are any items in the errors
stack via C<< $xsp->errors >> or by using a syntactic sugar that does the same,
the C<< $xsp->OK >> method.

=cut

sub xparse {
	my ($self) = shift;
	$self->schema_composition_mappings(get_schemaLocation_mappings_from_compose_ini($self->schema_document))
		unless $self->schema_composition_mappings;


	my $twig=XML::Twig->new( %$twig_params, no_xxe => 1 );
	
	my %builtin_stypes = %{$F5::XML::Schema::SchemaDefinitions::builtin_simple_types->{+XMLSCHEMAURL}};
	foreach my $type (keys %builtin_stypes) {
		next if $type eq 'anySimpleType';
		delete $builtin_stypes{$type}->{metatype_id_list};
	}
	
	my $xsd = new_xsd();
	$xsd->{schemaLocation}=get_canonical_filename($self->schema_document);
	$xsd->{current_schema_document_fullpath}=$xsd->{schemaLocation};

	$self->parse_xsd($twig,$xsd->{schemaLocation},$self) or return;

	my %file_dependency = (
		$xsd->{schemaLocation} => {
			main => 1,
			include => 1,
			import => 1,
		},
	);
	my $target_namespace = get_target_namespace($twig);
	my $das = {
		allow_external_references => $self->{allow_external_references},
		dependency => { %file_dependency },
		namespace_uris2id => {
			XMLSCHEMAURL,0,
			XLINKURL,1,
			XMLBASEURL,2,	
			$target_namespace => 3, # possibly '' if no tns
		},
		namespace_uris => [
			XMLSCHEMAURL,
			XLINKURL,
			XMLBASEURL,	
			$target_namespace
		],
		targetNamespace => $target_namespace,
		next_stype_header_i => 0,
		metatype => [], # list of ids to simple types (semantics: union)
		next_metatype_i => 0,
		metatype2id => {},
		simpletype => [], # list of ids to  (semantics: union)
		next_simpletype_i => 0,
		simpletype2id => {},
		facet => [], # list of facet {type=>$type,value=>$value} structures
		next_facet_i => 0,
		schema => {},
		facet2id => {},
		seen_import_namespace => {},
		self => $self,
	};
        weaken($das->{self});

	if ($target_namespace ne '') {
		push @{$das->{namespace_uris}},'';
		$das->{namespace_uris2id}->{''}=@{$das->{namespace_uris}}-1;
	}

	#$self->find_missing_files($xsd->{schemaLocation},$das,$self->schema_composition_mappings);
	#	return unless $self->OK;

	schema_sanity_check($xsd->{schemaLocation}) 
		if (defined $self->{debug} and $self->{debug}=~/sanity/);

	# This has to be done just after find_missing_files() return from its
	# recursion.
	# reset dependency graph so that handle_schema_composition can re-built
	# it and not get stuck in circles. Otherwise, it will return and report
	# circular dependencies even when there aren't any
	$das->{dependency} = { %file_dependency }; 

	$das->{digest}->{get_md5_digest_of_schema_document($xsd->{schemaLocation})}->{$target_namespace} = $xsd->{schemaLocation}; 

    # fill simple types derivation tree
	#handle_builtin_simple_types_headers($das, $xsd);

	return 1 if handle_schema_composition($das, $xsd, $twig, $self->schema_composition_mappings);
	handle_top_level($das, $xsd, $twig->root);
	

	foreach my $namespace (keys %{$xsd->{element}}) {
		foreach my $top_level_elt (keys %{$xsd->{element}->{$namespace}}) {
			my $elt_twig = ${$xsd->{element}->{$namespace}->{$top_level_elt}->{twig}};
			my $id       = $xsd->{element}->{$namespace}->{$top_level_elt}->{id};
			traverse($das, $elt_twig, $xsd, $id);
		}
	}

	# corrections
	handle_all_children_have_minOccurs_0($xsd);

	# handle global complex types
	foreach my $namespace (keys %{$xsd->{complexType}}) {
		foreach my $complexType (keys %{$xsd->{complexType}->{$namespace}}) {
			process_named_complexType($das, $xsd, $complexType, $namespace);
		}
	}
	foreach my $namespace (keys %{$xsd->{complexType}}) {
		foreach my $complexType (keys %{$xsd->{complexType}->{$namespace}}) {
			process_named_complexType_derivation($das, $xsd, $complexType, $namespace);
		}
	}
	# handle simpleTypes headers
	foreach my $namespace (keys %{$xsd->{simpleType}}) {
		foreach my $simpleType (keys %{$xsd->{simpleType}->{$namespace}}) {
			process_simpleType_header($das, $xsd, $simpleType, $namespace);
		}
	}
	
	# update simpleType/complexType index in containers
	foreach my $container (@{$xsd->{container}}) {
		if(exists $container->{simpleType}) {
			$container->{simpleType_id} = $das->{simpleType}->{$container->{simpleType}->{namespace_uri}}->{$container->{simpleType}->{name}};
			
		}
		if(exists $container->{complexType}) {
			$container->{complexType_id} = $xsd->{complexType}->{$container->{complexType}->{namespace_uri}}->{$container->{complexType}->{name}}->{id};
			
		}
		if($container->{container_type} eq 'complex_type') {
			$container->{complexType_id} = $xsd->{complexType}->{$container->{namespace_uri}}->{$container->{name}}->{content_id};
			if(exists $xsd->{complexType}->{$container->{namespace_uri}}->{$container->{name}}->{base_id}) {
				$container->{parent_id} = $xsd->{complexType}->{$container->{namespace_uri}}->{$container->{name}}->{base_id};
			}
		}
		if(exists $container->{element_type_info}) {
			delete $container->{element_type_info};
		}
	}	
	
	process_optional_flag_for_all_containers($xsd);

	#optimizations
	
	#output formats
	$xsd->{das}=$das;

	$self->{xsd}=$xsd;

    $self->blob(blobber($xsd));
	
	return 1;
}

sub schema_sanity_check {
	my ($schema_filename) = @_;

	my $cmd = qq{/usr/bin/xmllint --schema '$schema_filename' $DUMMY_XML_LOCATION 2>&1};
	my $output='';
	eval { $output = qx{$cmd}; };
	if ($output=~/WXS schema.*failed to compile/s) {
		## remove io warning that only appears in unit testing and not on bigip
		$output=~s,(?:I/O)?\s*warning\s*: failed to load external entity "$DUMMY_XML_LOCATION",,s;

		die $output;
	}
}

=pod

=head2 schema_composition_mappings($map)

Accessor and mutator for C<schema_composition_mappings> (see above in the
description of the possible arguments for the constructor C<new>).

Example usage:

=begin

	# to access the mappings, by getting a copy of them
	my $copy_of_the_mappings = $xsp->schema_composition_mappings;

	## OR

	# to change the mappings, by replacing them
	my $copy_of_the_new_mappings =
		$xsp->schema_composition_mappings($new_map);

=end

=cut

sub schema_composition_mappings {
	my ($self,$map) = @_;
	$self->{schema_composition_mappings} = { %$map }
		if defined $map;

	return undef unless $self->{schema_composition_mappings};

	return { %{$self->{schema_composition_mappings}} };
}

=pod

=head2 schema_document

Accessor and mutator for C<schema_document> (see above in the description of
the possible arguments for the constructor C<new>).

=begin

	# to get a copy of the schema document
	my $copy_of_schema_doc = $xsp->schema_document;

	## OR

	# to mutate the schema document in the object
	my $copy_of_the_schema_doc =
		$xsp->schema_document($new_schema_document);

=end

As explained above, a schema document can be a full path filename, a URI, or
the contents of the schema document itself.

=cut

sub schema_document {
	my ($self,$schema_document) = @_;
	$self->{schema_document} = $schema_document
		if defined $schema_document;
	return $self->{schema_document};
}

=pod

=head2 OK

This method returns a true value if there are no error messages on the error
messages stack, and returns false if these are error messages on the error
messages stack.

Example usage:

=begin

	warn "Error messages from the Schema Processor:\n" .  join(",\n",$xsd->errors) 
		unless $xsp->OK;

=end

=cut

sub OK {
	my ($self) = @_;
	return $self->errors < 1 ;
}

=pod

=head2 errors

This method returns a list of the error messages. If no error occurs the list
is expected to be empty. If the only error that occured is a fatal error, this
list might also be empty and the relevant error message should be extracted
from the thrown exception.

Example usage is available in the example usage of the method C<OK>.

=cut

sub errors {
	my ($self) = @_;
	return @{$self->{errors}};
}

=pod

=head2 push_error($error_message)

This method is to be used by the authors and maintainers of this module.
Don't use this method unless you know what you're doing!

This method is used to push an error message to the errors messages stack.

=cut

sub push_error {
	my ($self,$error_message) = @_;
	push @{$self->{errors}}, $error_message;
}

=pod

=head2 missing_files

This method is to be used in order to get a list of the missing files.
This is useful when schema composition is used and the schema processor cannot
access additional schema documents that are references by the schema document
or by schema documents that are references by it and are accessible to the
schema processor.

=cut

sub missing_files {
	my ($self) = @_;
	return @{$self->{missing_files}};
}

=pod

=head2 push_missing_files($missing_filename)

This method should not be used by anyone other than the authors or the
maintainers of this module.
Don't use it unless you know what you're doing!

The method is used to push another value of I<xsi:schemaLocation> or
I<namespace> attrubytes of I<xsd:import>, I<xsd:include> and I<xsd:redefine>

=cut

sub push_missing_files {
	my ($self,$missing_filename) = @_;
	push @{$self->{missing_files}},$missing_filename;
}

=pod

=head2 print_binary

This method is used to print the binary output (the result of the schema
processing) into the given file-handle. If no file handle was given either in
the constructor or before calling this method using C<< $xsp->blob_fh >> then
the binary will be printed to the standard output.

No binary will be printed if debug was invoked, Since debug should not be
invoked by anyone other than the authors and the maintainers of this module,
the behavior should be predictable.

=cut

sub print_binary {
	my ($self) = @_;

	if (defined $self->{debug}) {
		$Data::Dumper::Indent = 3; # pretty print with array indices
		$Data::Dumper::Sortkeys = 1; #sort hashes
		
		## drop twigs to beautify debug print
		undef $self->{xsd}->{das}->{schema};
		foreach my $top_level ( qw/element attributeGroup complexType group simpleType/ ) {
			foreach my $namespace (keys %{$self->{xsd}->{$top_level}}) {
				foreach my $top_level_elt (keys %{$self->{xsd}->{$top_level}->{$namespace}}) {
					delete $self->{xsd}->{$top_level}->{$namespace}->{$top_level_elt}->{twig};
				}
			}
		}
		
		if($self->{debug}=~/^--nobin$/) {
			print Data::Dumper::Dumper($self->{xsd});
			return;
		}elsif($self->{debug}=~/^--nobin:(\S+):(\S+)$/) {
			if($2 eq "list") {
				print "\t";
				print join "\n\t",(keys %{$self->{xsd}->{$1}});
				print "\n";
			}
			print Data::Dumper::Dumper($self->{xsd}->{$1}->{$2}) 
				if( exists $self->{xsd}->{$1} and exists $self->{xsd}->{$1}->{$2});
			return;
		}elsif($self->{debug}=~/^--nobin:(\S+)$/) {
			if($1 eq "list") {
				print "\t";
				print join "\n\t",(keys %{$self->{xsd}});
				print "\n";
			}
			print Data::Dumper::Dumper($self->{xsd}->{$1}) 
				if exists $self->{xsd}->{$1};
			return;
		}
	}

	if (defined $self->{blob_fh}) {
		binmode $self->{blob_fh};
		print {$self->{blob_fh}} $self->{blob};
	} else {
		binmode STDOUT;
		print $self->{blob};
	}
}

=pod

=head2 blob($blob)

This is an Accessor and mutator for C<blob> (see above in the description of
the possible arguments for the constructor C<new>).

There should be no need, in normal usage, to mutate the C<blob>. So don't mutate
the C<blob> unless you know what you're doing!!

If an argument is given then it mutates the C<blob>. Anyway, it returns the value
of the C<blob>.

=cut

sub blob {
	my ($self,$blob) = @_;
	$self->{blob}=$blob if defined $blob;
	return $self->{blob};
}

=pod

=head2 blob_fh($blob_fh)

This is an Accessor and mutator for C<blob_fh> (see above in the description of
the possible arguments for the constructor C<new>).

If an argument is given then it mutates the C<blob_fh>. Anyway, it returns the value
of the C<blob_fh>.

=cut

sub blob_fh {
	my ($self,$blob_fh) = @_;
	$self->{blob_fh}=$blob_fh if defined $blob_fh;
	return $self->{blob_fh};
}

## following method is not in the API (and shouldn't be)
sub parse_xsd {
	my ($self,$twig,$what_to_parse) = @_;
	eval {
		twig_parse($twig,$what_to_parse,$self);
	};

	if ($@) { # catch exception
		if (
			$@=~/No such file or directory at (.*) line \d+/ &&
			$@=~/Couldn't open (.*):/s
		) {
			my $missing_filename = $1;
			$missing_filename=$self->{original_schemaLocation_value}->{$missing_filename}
				if exists $self->{original_schemaLocation_value}->{$missing_filename};
			$self->push_missing_files($missing_filename);
			$self->push_error("Couldn't find or open schema document: |$missing_filename|");
		} elsif (
			$@=~/no element found at line 1, column 0, byte -1/
		) {
			$self->push_missing_files($what_to_parse);
			$self->push_error("Couldn't find or open schema document: |$what_to_parse|");
		} else {
			$self->push_error("===>Exception: ".$@);
		}
		return undef;
	};
	return 1; # OK
}

sub find_missing_files {
	my ($self,$schema_document,$das,$schemaLocation_mappings)=@_;

	my $twig=XML::Twig->new( %$twig_params, no_xxe => 1 );

	$self->parse_xsd($twig,$schema_document, $das)
		or return;

	foreach my $child ($twig->root->children) {
		next unless (
			is_include($child) or 
			is_redefine($child) or 
			is_import($child)
		);

		my $original_schemaLocation=$child->att('schemaLocation');
		if (not $original_schemaLocation) {
#			warn "Ignoring schema composition operation with no schemaLocation: ". get_context_as_string($child);
			next;
		}

		my $schemaLocation=$original_schemaLocation;

		$schemaLocation=$schemaLocation_mappings->{$original_schemaLocation}
			if exists $schemaLocation_mappings->{$schemaLocation};

		next unless defined $schemaLocation; # no schema location at all. Warn?


		# Circular dependency
		next if exists $das->{dependency}->{$schemaLocation}->{$child->local_name};

		$das->{dependency}->{$schemaLocation}->{$child->local_name}++;
	
		my $abs_base_dir = 
			URI->new_abs(
				URI->new(File::Basename::dirname($schema_document),"file"),
				URI->new(scalar(Cwd::getcwd)."/","file")
			)->as_string."/";
 
		#$abs_base_dir =~ s,/+,/,g;
	    $abs_base_dir =~ s,/+$,/,;

		$schemaLocation = 
			URI->new_abs(
				URI->new($schemaLocation,"file"), 
				$abs_base_dir
			)->as_string;

		$self->{original_schemaLocation_value}->{$schemaLocation} = $original_schemaLocation;

		$self->find_missing_files($schemaLocation,$das,$schemaLocation_mappings);

	}
}

sub abs_schemaLocation_dir {
	my ($schema_document_fullpath,$schemaLocation) = @_;
		my $abs_base_dir = 
			URI->new(
				File::Basename::dirname($schema_document_fullpath),"file"
			)->as_string."/";
		#$abs_base_dir =~ s,/+,/,g;
	    $abs_base_dir =~ s,/+$,/,;
	
		return	URI->new_abs(
				URI->new($schemaLocation, "file"), 
				$abs_base_dir
			)->as_string;
}

sub get_md5_digest_of_schema_document {
	my($schemaLocation) = @_;
	open my $fh, "<", $schemaLocation
		#or die "Cannot open $schemaLocation for reading: !$\n";
		# if file is remote just calculate digest of its full path name
		or return Digest::MD5::md5($schemaLocation);
	my $digest = Digest::MD5::md5(join '',<$fh>);
	close $fh;
	return $digest;
}

sub twig_parse {
	my ($twig,$what,$das) = @_;
	my $parsed_stuff;
	my $timeout_triggered = 0;
	if ($das->{allow_external_references}) {
		# Timeout to parse external file: 2 mins
		eval {
			local $SIG{ALRM} = sub { $timeout_triggered = 1; die "Timeout triggered\n"; };
			alarm 120;
			$parsed_stuff=$twig->xparse($what);
		};
		alarm 0;
	}else{
		eval { $parsed_stuff=$twig->parsefile($what); };
	}
	
	my $schema_ns = get_namespace_of_prefix($twig->root, defined($twig->root)?get_prefix($twig->root->gi):'');
	
#        die "\n\tnot a schema document '$what'\n\tschema namespace '$schema_ns'  - expecting '".XMLSCHEMAURL."'\n"
        die "\n\tThis is not a schema or WSDL document.\n"
		unless ($schema_ns eq '' or $schema_ns eq XMLSCHEMAURL);

	if ($@) {
		my $error_message;
		if ($@=~/cannot parse HTML/) {
			$error_message = "Unanticipated HTML file instead of a proper XML Schema document";
		} elsif ($timeout_triggered) {
			$error_message = "Fail to parse external XSD resource: $what";
		} else {
			$error_message = "The schema or WSDL document cannot be parsed (it is too large or contains bad syntax): $@";
			## remove report of which Perl modules failed and where. The important information to report
			## is the ill-formedness of the schema document and the location of the problem in the document.
			$error_message=~s/\s+(.*? at line \d+, column \d+, byte \d+).*?at.*/ $1/s;

		}
		die $error_message."\n";
	}

	return $parsed_stuff;
}

sub fake_twig_for_special_atts {
	my ($das,$xsd,$twig) = @_;
	# workaround to support xml:* attributes
	my $fake_twig = $twig;
	if ($twig->att('ref') and $twig->att('ref')=~/(xml|xlink):/) {
		$fake_twig = parse XML::Twig::Elt(
			'<xsd:attribute name="'.$twig->att('ref').'" type="xsd:string"'.
			' xmlns:xsd="'.XMLSCHEMAURL.'"/>'
		);
	}
	return $fake_twig;
}

1;
