package F5::XML::Schema::SchemaDefinitions;

use strict;
use warnings;

use File::Basename;
use Data::Dumper;
use Math::BigInt;

use F5::XML::Schema::SchemaH;

use constant XMLSCHEMAURL => 'http://www.w3.org/2001/XMLSchema';
use constant XSDATTRIBUTE => 'attribute';
use constant XSDELEMENT => 'element';
use constant XSDCOMPLEXTYPE => 'complexType';
use constant XSDSIMPLETYPE => 'simpleType';
use constant XSDSEQUENCE => 'sequence';
use constant XSDCHOICE => 'choice';
use constant XSDANY => 'any';
use constant XSDGROUP => 'group';
use constant XSDINCLUDE => 'include';
use constant XSDALL => 'all';
use constant XSDMINOCCURS => 'minOccurs';
use constant XSDMAXOCCURS => 'maxOccurs';
use constant DEFAULTNSPREFIX => '';

use constant PARSER_MODE_CHECK_NAMESPACES => 0x00000001;
use constant PARSER_MODE_CHECK_ELEMENTS => 0x00000002;
use constant PARSER_MODE_CHECK_SCHEMA => 0x00000004;
use constant PARSER_MODE_RESTRICT_NAMESPACE_DECLARATION	=> 0x00000008;
use constant PARSER_MODE_SKIP => 0x00000010;
use constant PARSER_MODE_LAX => 0x00000020;
use constant PARSER_MODE_STRICT => 0x00000040;

# no facet -- facet not applicable
# facet undef -- facet applicable but not restricting
# facet with defined value -- facet with restrictive value
our $builtin_simple_types = {
	XMLSCHEMAURL, {
		'anySimpleType' => {
			name => 'anySimpleType',
			builtin_type => 'anySimpleType',
			is_atom => 1,
			parent => undef,
			parent_namespace => undef,
			children => {
				string => 1,
				QName => 1,
				NOTATION => 1,
				float => 1,
				double => 1,
				decimal => 1,
				boolean => 1,
				base64Binary => 1,
				hexBinary => 1,
				anyURI  => 1,
			},
			is_a => {
				anySimpleType => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
				'totalDigits' => undef,
				'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				string => 1,
				legacy => 1,
				numeric => 1,
				other => 1,
				'date and time' => 1,
			},
			metatype_id_list => [],
		},
		'string' => {
			name => 'string',
			builtin_type => 'string',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {
				normalizedString => 1,
			},
			is_a => {
				anySimpleType => 1,
				string => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'preserve',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				string => 1,
			},
		},
		'normalizedString' => {
			name => 'normalizedString',
			builtin_type => 'normalizedString',
			is_atom => 1,
			parent => 'string',
			parent_namespace => XMLSCHEMAURL,
			children => {
				token => 1,
			},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'replace',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				string => 1,
			}
		},
		'token' => {
			name => 'token',
			builtin_type => 'token',
			is_atom => 1,
			parent => 'normalizedString',
			parent_namespace => XMLSCHEMAURL,
			children => {
				'language' => 1,
				'Name' => 1,
				'NMTOKEN' => 1,
			},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				string => 1,
			}
		},
		'language' => {
			name => 'language',
			builtin_type => 'language',
			is_atom => 1,
			parent => 'token',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				language => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => '[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*',
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				string => 1,
			}
		},
		'Name' => {
			name => 'Name',
			builtin_type => 'Name',
			is_atom => 1,
			parent => 'token',
			parent_namespace => XMLSCHEMAURL,
			children => {
				NCName => 1,
			},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				Name => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => '\i\c*',
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				string => 1,
			}
		},
		'NCName' => {
			name => 'NCName',
			builtin_type => 'NCName',
			is_atom => 1,
			parent => 'Name',
			parent_namespace => XMLSCHEMAURL,
			children => {
				ID => 1,
				IDREF => 1,
				ENTITY => 1,
			},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				Name => 1,
				NCName => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => '[\i-[:]][\c-[:]]*',
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				string => 1,
			}
			
		},
		'ID' => {
			name => 'ID',
			builtin_type => 'ID',
			is_atom => 1,
			parent => 'NCName',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				Name => 1,
				NCName => 1,
				ID => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => '[\i-[:]][\c-[:]]*',
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				legacy => 1,
			}
			
		},
		'IDREF' => {
			name => 'IDREF',
			builtin_type => 'IDREF',
			is_atom => 1,
			parent => 'NCName',
			parent_namespace => XMLSCHEMAURL,
			children => {
				IDREFS => 1,
			},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				Name => 1,
				NCName => 1,
				IDREF => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => '[\i-[:]][\c-[:]]*',
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				legacy => 1,
			}
		},
		'IDREFS' => {
			name => 'IDREFS',
			builtin_type => 'IDREF', # this is alist of IDREF
			is_atom => 0,
			parent => 'IDREF',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				Name => 1,
				NCName => 1,
				IDREF => 1,
				IDREFS => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => 1,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
	# (no need)		'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				legacy => 1,
			}
		},
		'ENTITY' => {
			name => 'ENTITY',
			builtin_type => 'ENTITY',
			is_atom => 1,
			parent => 'NCName',
			parent_namespace => XMLSCHEMAURL,
			children => {
				ENTITIES => 1,
			},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				Name => 1,
				NCName => 1,
				ENTITY => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => '[\i-[:]][\c-[:]]*',
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				legacy => 1,
			}
		},
		'ENTITIES' => {
			name => 'ENTITIES',
			builtin_type => 'ENTITY', # this is a list of ENTITY
			is_atom => 0,
			parent => 'ENTITY',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				Name => 1,
				NCName => 1,
				ENTITY => 1,
				ENTITIES => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => 1,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
	# (no need)		'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				legacy => 1,
			}
		},
		'NMTOKEN' => {
			name => 'NMTOKEN',
			builtin_type => 'NMTOKEN',
			is_atom => 1,
			parent => 'token',
			parent_namespace => XMLSCHEMAURL,
			children => {
				NMTOKENS => 1,
			},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				NMTOKEN => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => '\c+',
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				legacy => 1,
			},
		},
		'NMTOKENS' => {
			name => 'NMTOKENS',
			builtin_type => 'NMTOKEN', # this is a list of NMTOKEN
			is_atom => 0,
			parent => 'NMTOKEN',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				string => 1,
				normalizedString => 1,
				token => 1,
				NMTOKEN => 1,
				NMTOKENS => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => 1,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
	# (no need)		'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				legacy => 1,
			},
		},
		'QName' => {
			name => 'QName',
			builtin_type => 'QName',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				QName => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				string => 1,
			},
		},
		'NOTATION' => {
			name => 'NOTATION',
			builtin_type => 'NOTATION',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				NOTATION => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				legacy => 1,
			},
		},
		'date' => {
			name => 'date',
			builtin_type => 'date',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				date => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'time' => {
			name => 'time',
			builtin_type => 'time',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				time => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'dateTime' => {
			name => 'dateTime',
			builtin_type => 'dateTime',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				dateTime => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'gYear' => {
			name => 'gYear',
			builtin_type => 'gYear',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				gYear => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'gYearMonth' => {
			name => 'gYearMonth',
			builtin_type => 'gYearMonth',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				gYearMonth => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'gMonth' => {
			name => 'gMonth',
			builtin_type => 'gMonth',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				gMonth => 1,
			},
			face => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'gMonthDay' => {
			name => 'gMonthDay',
			builtin_type => 'gMonthDay',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				gMonthDay => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'gDay' => {
			name => 'gDay',
			builtin_type => 'gDay',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				gDay => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'duration' => {
			name => 'duration',
			builtin_type => 'duration',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				duration => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				'date and time' => 1,
			},
		},
		'float' => {
			name => 'float',
			builtin_type => 'float',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				float => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			},
		},
		'double' => {
			name => 'double',
			builtin_type => 'double',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				double => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			},
		},
		'decimal' => {
			name => 'decimal',
			builtin_type => 'decimal',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {
				integer => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	#			'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'integer' => {
			name => 'integer',
			builtin_type => 'integer',
			is_atom => 1,
			parent => 'decimal',
			parent_namespace => XMLSCHEMAURL,
			children => {
				long => 1,
				nonPositiveInteger => 1,
				nonNegativeInteger => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'long' => {
			name => 'long',
			builtin_type => 'long',
			is_atom => 1,
			parent => 'integer',
			parent_namespace => XMLSCHEMAURL,
			children => {
				int => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				long => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => Math::BigInt->new('-9223372036854775808'), # requires Math::BigInt
				'maxInclusive' => Math::BigInt->new('9223372036854775807'), # requires Math::BigInt
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'int' => {
			name => 'int',
			builtin_type => 'int',
			is_atom => 1,
			parent => 'long',
			parent_namespace => XMLSCHEMAURL,
			children => {
				short => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				long => 1,
				int => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => -2147483648,
				'maxInclusive' => 2147483647,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'short' => {
			name => 'short',
			builtin_type => 'short',
			is_atom => 1,
			parent => 'int',
			parent_namespace => XMLSCHEMAURL,
			children => {
				byte => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				long => 1,
				int => 1,
				short => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => -32768,
				'maxInclusive' => 32767,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'byte' => {
			name => 'byte',
			builtin_type => 'byte',
			is_atom => 1,
			parent => 'short',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				long => 1,
				int => 1,
				short => 1,
				byte => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => -128,
				'maxInclusive' => 127,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'nonPositiveInteger' => {
			name => 'nonPositiveInteger',
			builtin_type => 'nonPositiveInteger',
			is_atom => 1,
			parent => 'integer',
			parent_namespace => XMLSCHEMAURL,
			children => {
				negativeInteger => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				nonPositiveInteger => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => 0,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'negativeInteger' => {
			name => 'negativeInteger',
			builtin_type => 'negativeInteger',
			is_atom => 1,
			parent => 'nonPositiveInteger',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				nonPositiveInteger => 1,
				negativeInteger => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => undef,
				'maxInclusive' => -1,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'nonNegativeInteger' => {
			name => 'nonNegativeInteger',
			builtin_type => 'nonNegativeInteger',
			is_atom => 1,
			parent => 'integer',
			parent_namespace => XMLSCHEMAURL,
			children => {
				positiveInteger => 1,
				unsignedLong => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				nonNegativeInteger => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => 0,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'positiveInteger' => {
			name => 'positiveInteger',
			builtin_type => 'positiveInteger',
			is_atom => 1,
			parent => 'nonNegativeInteger',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				nonNegativeInteger => 1,
				positiveInteger => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => 1,
				'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'unsignedLong' => {
			name => 'unsignedLong',
			builtin_type => 'unsignedLong',
			is_atom => 1,
			parent => 'nonNegativeInteger',
			parent_namespace => XMLSCHEMAURL,
			children => {
				unsignedInt => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				nonNegativeInteger => 1,
				unsignedLong => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => Math::BigInt->new(0),
				'maxInclusive' => Math::BigInt->new('18446744073709551615'), # requires Math::BigInt
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'unsignedInt' => {
			name => 'unsignedInt',
			builtin_type => 'unsignedInt',
			is_atom => 1,
			parent => 'unsignedLong',
			parent_namespace => XMLSCHEMAURL,
			children => {
				unsignedShort => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				nonNegativeInteger => 1,
				unsignedLong => 1,
				unsignedInt => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => 0,
				'maxInclusive' => 4294967295,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'unsignedShort' => {
			name => 'unsignedShort',
			builtin_type => 'unsignedShort',
			is_atom => 1,
			parent => 'unsignedInt',
			parent_namespace => XMLSCHEMAURL,
			children => {
				unsignedByte => 1,
			},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				nonNegativeInteger => 1,
				unsignedLong => 1,
				unsignedInt => 1,
				unsignedShort => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => 0,
				'maxInclusive' => 65535,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'unsignedByte' => {
			name => 'unsignedByte',
			builtin_type => 'unsignedByte',
			is_atom => 1,
			parent => 'unsignedShort',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				decimal => 1,
				integer => 1,
				nonNegativeInteger => 1,
				unsignedLong => 1,
				unsignedInt => 1,
				unsignedShort => 1,
				unsignedByte => 1,
			},
			facet => {
				'minExclusive' => undef,
				'maxExclusive' => undef,
				'minInclusive' => 0,
				'maxInclusive' => 255,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
				'totalDigits' => undef,
	# (no need)	'fractionDigits' => 0,
	# (no need)	'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				numeric => 1,
			}
		},
		'boolean' => {
			name => 'boolean',
			builtin_type => 'boolean',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				boolean => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
	#			'length' => undef,
	#			'minLength' => undef,
	#			'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
	## we replace boolean definition with token enomeration {true, false, 1, 0}
				'whiteSpace' => 'collapse',
				'enumeration' => {'true'  => 1,
								  'false' => 1,
								  '1'     => 1,
								  '0'     => 1,
								 },
	#			'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				other => 1,
			}
		},
		'base64Binary' => {
			name => 'base64Binary',
			builtin_type => 'base64Binary',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				base64Binary => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				other => 1,
			}
		},
		'hexBinary' => {
			name => 'hexBinary',
			builtin_type => 'hexBinary',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				hexBinary => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => undef,
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				other => 1,
			}
		},
		'anyURI' => {
			name => 'anyURI',
			builtin_type => 'anyURI',
			is_atom => 1,
			parent => 'anySimpleType',
			parent_namespace => XMLSCHEMAURL,
			children => {},
			is_a => {
				anySimpleType => 1,
				anyURI => 1,
			},
			facet => {
	#			'minExclusive' => undef,
	#			'maxExclusive' => undef,
	#			'minInclusive' => undef,
	#			'maxInclusive' => undef,
				'length' => undef,
				'minLength' => undef,
				'maxLength' => undef,
	#			'totalDigits' => undef,
	#			'fractionDigits' => undef,
				'whiteSpace' => 'collapse',
				'enumeration' => undef,
				'pattern' => undef,
			},
			namespace_uri => XMLSCHEMAURL,
			ns_uri_id => 0,
			type => {
				other => 1,
			}
		},
	},
};

our $facet2simple_type = {
	XMLSCHEMAURL, {
		minExclusive => {
			nonNegativeInteger => 1,
			date => 1,
			dateTime => 1,
			double => 1,
			integer => 1,
			unsignedLong => 1,
			gDay => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			gYearMonth => 1,
			duration => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			int => 1,
			unsignedByte => 1,
			time => 1,
			gYear => 1,
			short => 1,
			float => 1,
			long => 1,
			gMonthDay => 1,
			unsignedShort => 1,
			decimal => 1,
		},
		maxExclusive => {
			nonNegativeInteger => 1,
			date => 1,
			dateTime => 1,
			double => 1,
			integer => 1,
			unsignedLong => 1,
			gDay => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			gYearMonth => 1,
			duration => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			int => 1,
			unsignedByte => 1,
			time => 1,
			gYear => 1,
			short => 1,
			float => 1,
			long => 1,
			gMonthDay => 1,
			unsignedShort => 1,
			decimal => 1,
		},
		minInclusive => {
			nonNegativeInteger => 1,
			date => 1,
			dateTime => 1,
			double => 1,
			integer => 1,
			unsignedLong => 1,
			gDay => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			gYearMonth => 1,
			duration => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			int => 1,
			unsignedByte => 1,
			time => 1,
			gYear => 1,
			short => 1,
			float => 1,
			long => 1,
			gMonthDay => 1,
			unsignedShort => 1,
			decimal => 1,
		},
		maxInclusive => {
			nonNegativeInteger => 1,
			date => 1,
			dateTime => 1,
			double => 1,
			integer => 1,
			unsignedLong => 1,
			gDay => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			gYearMonth => 1,
			duration => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			int => 1,
			unsignedByte => 1,
			time => 1,
			gYear => 1,
			short => 1,
			float => 1,
			long => 1,
			gMonthDay => 1,
			unsignedShort => 1,
			decimal => 1,
		},
		length => {
			ENTITY => 1,
			ENTITIES => 1,
			string => 1,
			NOTATION => 1,
			anyURI => 1,
			IDREF => 1,
			normalizedString => 1,
			base64Binary => 1,
			anySimpleType => 1,
			ID => 1,
			NMTOKENS => 1,
			NCName => 1,
			QName => 1,
			token => 1,
			language => 1,
			NMTOKEN => 1,
			hexBinary => 1,
			IDREFS => 1,
			Name => 1,
		},
		minLength => {
			ENTITY => 1,
			ENTITIES => 1,
			string => 1,
			NOTATION => 1,
			anyURI => 1,
			IDREF => 1,
			normalizedString => 1,
			base64Binary => 1,
			anySimpleType => 1,
			ID => 1,
			NMTOKENS => 1,
			NCName => 1,
			QName => 1,
			token => 1,
			language => 1,
			NMTOKEN => 1,
			hexBinary => 1,
			IDREFS => 1,
			Name => 1,
		},
		maxLength => {
			ENTITY => 1,
			ENTITIES => 1,
			string => 1,
			NOTATION => 1,
			anyURI => 1,
			IDREF => 1,
			normalizedString => 1,
			base64Binary => 1,
			anySimpleType => 1,
			ID => 1,
			NMTOKENS => 1,
			NCName => 1,
			QName => 1,
			token => 1,
			language => 1,
			NMTOKEN => 1,
			hexBinary => 1,
			IDREFS => 1,
			Name => 1,
		},
		totalDigits => {
			nonNegativeInteger => 1,
			integer => 1,
			unsignedLong => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			int => 1,
			unsignedByte => 1,
			short => 1,
			long => 1,
			unsignedShort => 1,
			decimal => 1,
		},
		fractionDigits => {
			nonNegativeInteger => 1,
			integer => 1,
			unsignedLong => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			int => 1,
			unsignedByte => 1,
			short => 1,
			long => 1,
			unsignedShort => 1,
			decimal => 1,
		},
		whiteSpace => {
			ENTITY => 1,
			nonNegativeInteger => 1,
			date => 1,
			ENTITIES => 1,
			dateTime => 1,
			string => 1,
			NOTATION => 1,
			double => 1,
			integer => 1,
			anyURI => 1,
			unsignedLong => 1,
			gDay => 1,
			IDREF => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			gYearMonth => 1,
			duration => 1,
			normalizedString => 1,
			base64Binary => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			ID => 1,
			int => 1,
			unsignedByte => 1,
			NMTOKENS => 1,
			time => 1,
			gYear => 1,
			NCName => 1,
			boolean => 1,
			short => 1,
			float => 1,
			long => 1,
			QName => 1,
			token => 1,
			gMonthDay => 1,
			language => 1,
			unsignedShort => 1,
			NMTOKEN => 1,
			hexBinary => 1,
			IDREFS => 1,
			decimal => 1,
			Name => 1,
		},
		enumeration => {
			ENTITY => 1,
			nonNegativeInteger => 1,
			date => 1,
			ENTITIES => 1,
			dateTime => 1,
			string => 1,
			NOTATION => 1,
			double => 1,
			integer => 1,
			anyURI => 1,
			unsignedLong => 1,
			gDay => 1,
			IDREF => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			gYearMonth => 1,
			duration => 1,
			normalizedString => 1,
			base64Binary => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			ID => 1,
			int => 1,
			unsignedByte => 1,
			NMTOKENS => 1,
			time => 1,
			gYear => 1,
			NCName => 1,
			short => 1,
			float => 1,
			long => 1,
			QName => 1,
			token => 1,
			gMonthDay => 1,
			language => 1,
			unsignedShort => 1,
			NMTOKEN => 1,
			hexBinary => 1,
			IDREFS => 1,
			decimal => 1,
			Name => 1,
		},
		pattern => {
			ENTITY => 1,
			nonNegativeInteger => 1,
			date => 1,
			ENTITIES => 1,
			dateTime => 1,
			string => 1,
			NOTATION => 1,
			double => 1,
			integer => 1,
			anyURI => 1,
			unsignedLong => 1,
			gDay => 1,
			IDREF => 1,
			nonPositiveInteger => 1,
			unsignedInt => 1,
			gYearMonth => 1,
			duration => 1,
			normalizedString => 1,
			base64Binary => 1,
			positiveInteger => 1,
			anySimpleType => 1,
			byte => 1,
			negativeInteger => 1,
			ID => 1,
			int => 1,
			unsignedByte => 1,
			NMTOKENS => 1,
			time => 1,
			gYear => 1,
			NCName => 1,
			boolean => 1,
			short => 1,
			float => 1,
			long => 1,
			QName => 1,
			token => 1,
			gMonthDay => 1,
			language => 1,
			unsignedShort => 1,
			NMTOKEN => 1,
			hexBinary => 1,
			IDREFS => 1,
			decimal => 1,
			Name => 1,
		},
	},
};

our $schema_h = F5::XML::Schema::SchemaH->new();

our $container_type2id = {
	element      => $schema_h->XSD_CONT_TYPE__ELEMENT,
	'ref'        => $schema_h->XSD_CONT_TYPE__ELEMENT_REF,
	any          => $schema_h->XSD_CONT_TYPE__ANY,
	'any_ref'    => $schema_h->XSD_CONT_TYPE__ANY_REF,
	attribute    => $schema_h->XSD_CONT_TYPE__ATTRIBUTE,
	anyAttribute => $schema_h->XSD_CONT_TYPE__ANY_ATTRIBUTE,
	sequence     => $schema_h->XSD_CONT_TYPE__SEQUENCE,
	choice       => $schema_h->XSD_CONT_TYPE__CHOICE,
	all          => $schema_h->XSD_CONT_TYPE__ALL,
	complex_type => $schema_h->XSD_CONT_TYPE__COMPLEX_TYPE,
	''           => $schema_h->XSD_CONT_TYPE__ILLEGAL,
};

our $parser_mode2int = {
	skip   => PARSER_MODE_SKIP,
	lax    => PARSER_MODE_LAX,
	strict => PARSER_MODE_STRICT,
};

our $form2int = {
	qualified          => $schema_h->ELEMENT_FORM_QUALIFIED,
	unqualified        => $schema_h->ELEMENT_FORM_UNQUALIFIED,
	'illegal form'     => $schema_h->ELEMENT_FORM_ILLEGAL,
};

our $facet2int = {
	minExclusive       => $schema_h->XS_FACET__MIN_EXCLUSIVE,
	maxExclusive       => $schema_h->XS_FACET__MAX_EXCLUSIVE,
	minInclusive       => $schema_h->XS_FACET__MIN_INCLUSIVE,
	maxInclusive       => $schema_h->XS_FACET__MAX_INCLUSIVE,
	length             => $schema_h->XS_FACET__LENGTH,
	minLength          => $schema_h->XS_FACET__MIN_LENGTH,
	maxLength          => $schema_h->XS_FACET__MAX_LENGTH,
	totalDigits        => $schema_h->XS_FACET__TOTAL_DIGITS,
	fractionDigits     => $schema_h->XS_FACET__FRACTION_DIGITS,
	whiteSpace         => $schema_h->XS_FACET__WHITESPACE,
	enumeration        => $schema_h->XS_FACET__ENUMERATION,
	pattern            => $schema_h->XS_FACET__PATTERN,
};

our $simpleType2int = {
	anySimpleType      => $schema_h->XS_SIMPLE_TYPE__ANY_SIMPLE_TYPE,
	string             => $schema_h->XS_SIMPLE_TYPE__STRING,
	QName              => $schema_h->XS_SIMPLE_TYPE__QNAME,
	normalizedString   => $schema_h->XS_SIMPLE_TYPE__NORMALIZED_STRING,
	token              => $schema_h->XS_SIMPLE_TYPE__TOKEN,
	language           => $schema_h->XS_SIMPLE_TYPE__LANGUAGE,
	Name               => $schema_h->XS_SIMPLE_TYPE__NAME,
	NCName             => $schema_h->XS_SIMPLE_TYPE__NC_NAME,
	decimal            => $schema_h->XS_SIMPLE_TYPE__DECIMAL,
	integer            => $schema_h->XS_SIMPLE_TYPE__INTEGER,
	long               => $schema_h->XS_SIMPLE_TYPE__LONG,
	'int'              => $schema_h->XS_SIMPLE_TYPE__INT,
	short              => $schema_h->XS_SIMPLE_TYPE__SHORT,
	byte               => $schema_h->XS_SIMPLE_TYPE__BYTE,
	date               => $schema_h->XS_SIMPLE_TYPE__DATE,
	boolean            => $schema_h->XS_SIMPLE_TYPE__BOOLEAN,
	float              => $schema_h->XS_SIMPLE_TYPE__FLOAT,
	double             => $schema_h->XS_SIMPLE_TYPE__DOUBLE,
	duration           => $schema_h->XS_SIMPLE_TYPE__DURATION,
	dateTime           => $schema_h->XS_SIMPLE_TYPE__DATE_TIME,
	'time'             => $schema_h->XS_SIMPLE_TYPE__TIME,
	gYearMonth         => $schema_h->XS_SIMPLE_TYPE__G_YEAR_MONTH,
	gYear              => $schema_h->XS_SIMPLE_TYPE__G_YEAR,
	gMonthDay          => $schema_h->XS_SIMPLE_TYPE__G_MONTH_DAY,
	gDay               => $schema_h->XS_SIMPLE_TYPE__G_DAY,
	gMonth             => $schema_h->XS_SIMPLE_TYPE__G_MONTH,
	hexBinary          => $schema_h->XS_SIMPLE_TYPE__HEX_BINARY,
	base64Binary       => $schema_h->XS_SIMPLE_TYPE__BASE_64_BINARY,
	anyURI             => $schema_h->XS_SIMPLE_TYPE__ANY_URI,
	NOTATION           => $schema_h->XS_SIMPLE_TYPE__NOTATION,
	NMTOKEN            => $schema_h->XS_SIMPLE_TYPE__NM_TOKEN,
	ID                 => $schema_h->XS_SIMPLE_TYPE__ID,
	IDREF              => $schema_h->XS_SIMPLE_TYPE__ID_REF,
	ENTITY             => $schema_h->XS_SIMPLE_TYPE__ENTITY,
	nonPositiveInteger => $schema_h->XS_SIMPLE_TYPE__NON_POSITIVE_INTEGER,
	negativeInteger    => $schema_h->XS_SIMPLE_TYPE__NEGATIVE_INTEGER,
	nonNegativeInteger => $schema_h->XS_SIMPLE_TYPE__NON_NEGATIVE_INTEGER,
	unsignedLong       => $schema_h->XS_SIMPLE_TYPE__UNSIGNED_LONG,
	unsignedInt        => $schema_h->XS_SIMPLE_TYPE__UNSIGNED_INT,
	unsignedShort      => $schema_h->XS_SIMPLE_TYPE__UNSIGNED_SHORT,
	unsignedByte       => $schema_h->XS_SIMPLE_TYPE__UNSIGNED_BYTE,
	positiveInteger    => $schema_h->XS_SIMPLE_TYPE__POSITIVE_INTEGER,
};

our $facet_value2blob = {
	whiteSpace => {
		preserve => eval($schema_h->XS_WHITESPACE_FACET_PRESERVE),
		replace  => eval($schema_h->XS_WHITESPACE_FACET_REPLACE),
		collapse => eval($schema_h->XS_WHITESPACE_FACET_COLLAPSE),
	},
};


1;
