package F5::XML::WSDL::Convert;

use strict;
use warnings;

use F5::Deploy;
use F5::XML::Twig;
use XML::Parser;
use File::Copy;
use File::Basename;
use Cwd 'abs_path';
use URI;

my $GENERIC_SCHEMAS_ROOT = F5::Deploy::map_path("/ts/share/xml/generic_schemas");

my $XML_SCHEMA_URI = "http://www.w3.org/2001/XMLSchema";
my $ARBITRARY_DIVIDER = '__//ARBITRARY_DIVIDER//__';


#----------------------------------------------------------------------

sub convert_wsdl {
  my(%params) = @_;

  my @soap_objects;

  my $schema_dir = $params{schema_dir};
  my $flg_follow_schema_links = (exists $params{flg_follow_schema_links} ? $params{flg_follow_schema_links} : 1);

  my %schema_composition_mappings;
  if ($params{schema_composition_mappings}) {
    # Make a local copy, in case we need to add newly discovered mapping
    %schema_composition_mappings = %{$params{schema_composition_mappings}};
  }

  my @wsdls_remaining;
  push(@wsdls_remaining, abs_path($params{wsdl_filename}) || $params{wsdl_filename});
  my %wsdls_seen = ($wsdls_remaining[0] => 1);
  my @missing_composition_wsdls;

  my $target_namespace;

  my %globally_allowed_namespaces;
  my $existing_global_elements = {};
  my @internal_types_schemas;
  my @imported_schemas;
  my %converted_schemas;
  my %all_methods = ();

  while (my $current_wsdl_filename = shift(@wsdls_remaining)) {
    my $current_wsdl = { xmlns_declarations => {} };

    my $twig = F5::XML::Twig->new(start_tag_handlers => {
                                                     'wsdl:definitions' => sub {
                                                       my($twig, $elm) = @_;
                                                       $current_wsdl->{namespace} = $elm->att('targetNamespace');

                                                       for my $att_name ($elm->att_names) {
                                                         if ($att_name =~ /^xmlns:?(.*)$/) {
                                                           $current_wsdl->{xmlns_declarations}->{$att_name} = $elm->att($att_name); 
                                                         }
                                                       }

                                                     },
                                                    },
                              twig_handlers => {

                                                'wsdl:import' => sub {
                                                  my($twig, $elm) = @_;
                                                  my $location = $elm->att('location');
                                                  my $local_filename = $location;

                                                  unless ($location) {
                                                    die "Missing location on WSDL import\n";
                                                  }

                                                  if ($schema_composition_mappings{$location}) {
                                                    $local_filename = $schema_composition_mappings{$location};
                                                  }
                                                  $local_filename = URI->new_abs($local_filename, $current_wsdl_filename)->as_string;

                                                  my $is_wsdl;
                                                  eval {
                                                    $is_wsdl = is_wsdl(xml_filename => $local_filename,
                                                                       die_if_missing => 1,
                                                                       flg_follow_schema_links => $flg_follow_schema_links);
                                                  };

                                                  if ($@) {
                                                    push(@missing_composition_wsdls, $location);
                                                  }
                                                  elsif ($is_wsdl) {
                                                    unless($wsdls_seen{$local_filename}) {
                                                      push(@wsdls_remaining, $local_filename);
                                                      $wsdls_seen{$local_filename} = 1;
                                                    }
                                                  }
                                                  else {
                                                    handle_schema_import(imported_schemas => \@imported_schemas,
                                                                         flg_follow_schema_links => $flg_follow_schema_links,
                                                                         import_filename => $local_filename,
                                                                         original_location => $location,
                                                                         import_namespace => $elm->att('namespace'),
                                                                         existing_global_elements => $existing_global_elements,
                                                                        );

                                                  }
                                                },

                                                'wsdl:types' => sub {
                                                  my($twig, $elm) = @_;

                                                  my @schemas = ($elm->children("xsd:schema"),
                                                                );
                                                  for my $schema (@schemas) {

                                                    my $types_schema;
                                                    my %schema_namespace_prefixes;
                                                    my $xml_schema_prefix;

                                                    if ($schema->has_children) {
                                                      $types_schema->{namespace} = $schema->att('targetNamespace');
                                                      unless ($types_schema->{namespace}) {
                                                        $types_schema->{namespace} = $current_wsdl->{namespace};
                                                        if ($types_schema->{namespace}) {
                                                            $schema->set_att('targetNamespace' => $types_schema->{namespace});
                                                        }
                                                      }

                                                      for my $att_name ($schema->att_names) {
                                                        if ($att_name =~ /^xmlns:?(.*)$/) {
                                                          $schema_namespace_prefixes{$att_name} = $schema->att($att_name); 
                                                        }
                                                      }

                                                      $xml_schema_prefix = find_next_available_name(target => 'xmlns:xsd',
                                                                                                    reserved => [ keys %schema_namespace_prefixes ]);
                                                      $schema->set_att($xml_schema_prefix => $XML_SCHEMA_URI);
                                                      $xml_schema_prefix =~ s/^xmlns://;

                                                      for my $prefix (keys %{$current_wsdl->{xmlns_declarations}}) {
                                                        unless (exists $schema_namespace_prefixes{$prefix}) {
                                                          $schema->set_att($prefix => $current_wsdl->{xmlns_declarations}->{$prefix});
                                                        }
                                                      }
                                                    }

                                                    # Strip imports from imported schemas and apply to parent wsdls
                                                    for my $child_import ($schema->children("xsd:import")) {
                                                      my $original_location = $child_import->att('schemaLocation');
                                                      my $import_filename = $original_location;
                                                      my $import_namespace = $child_import->att('namespace');

                                                      if ($import_filename) {
                                                        if ($schema_composition_mappings{$import_filename}) {
                                                          $import_filename = $schema_composition_mappings{$import_filename};
                                                        }

                                                        $import_filename = URI->new_abs($import_filename, $current_wsdl_filename)->as_string;

                                                        handle_schema_import(imported_schemas => \@imported_schemas,
                                                                             flg_follow_schema_links => $flg_follow_schema_links,
                                                                             import_filename => $import_filename,
                                                                             original_location => $original_location,
                                                                             import_namespace => $import_namespace,
                                                                             existing_global_elements => $existing_global_elements,
                                                                            );

                                                        if ($import_namespace eq $types_schema->{namespace}) {
                                                          $child_import->delete;
                                                        }
                                                      }
                                                    }

                                                    next unless $schema->has_children;


                                                    # Discover all global element declarations
                                                    for my $element ($schema->findnodes("./xsd:element")) {
                                                      $existing_global_elements->{$types_schema->{namespace}}->{$element->att('name')} = 1;
                                                      $types_schema->{elements}->{$element->att('name')} = 1;
                                                    }

                                                    # Discover all global complex types
                                                    for my $complex_type ($schema->findnodes("./xsd:complexType")) {
                                                      $types_schema->{complex_types}->{$complex_type->att('name')} = 1;
                                                    }

                                                    for my $soap_encoding_element ($schema->findnodes('//*[@wsdl:arrayType]')) {
                                                      # Change parent restriction to extension
                                                      my $complex_content_tag = $soap_encoding_element->parent->parent;
                                                      if ($complex_content_tag->name eq "xsd:complexContent") {
                                                        my $base = $complex_content_tag->first_child->att('base');
                                                        $complex_content_tag->first_child->delete;

                                                        XML::Twig::Elt->new("xsd:extension",
                                                                            { base => $base})->paste($complex_content_tag);
                                                      }
                                                    }

                                                    my $import = XML::Twig::Elt->new($xml_schema_prefix . ':import',
                                                                                     { namespace => "http://schemas.xmlsoap.org/soap/encoding/",
                                                                                       schemaLocation => 'soap-encoding.xsd'});
                                                    $import->paste(first_child => $schema);

                                                    # If another internal schema used the same namespace, include it in this schema and
                                                    # mark it so that it won't be imported elsewhere.
                                                    for my $other_internal_types_schema (@internal_types_schemas) {
                                                      if ($other_internal_types_schema->{namespace} eq $types_schema->{namespace}) {
                                                        $other_internal_types_schema->{no_import} = 1;

                                                        my $include = XML::Twig::Elt->new($xml_schema_prefix . ':include',
                                                                                          {schemaLocation => $other_internal_types_schema->{filename} });
                                                        $include->paste(first_child => $schema);

                                                        # Remove all global elements and complex types defined in the predecessor.
                                                        for my $element ($schema->findnodes("./xsd:element")) {
                                                          if ($other_internal_types_schema->{elements}->{$element->att('name')}) {
                                                            $element->delete;
                                                          }
                                                        }

                                                        for my $complex_type ($schema->findnodes("./xsd:complexType")) {
                                                          if ($other_internal_types_schema->{complex_types}->{$complex_type->att('name')}) {
                                                            $complex_type->delete;
                                                          }
                                                        }
                                                      }
                                                    }

                                                    my $external_schema_count = $#internal_types_schemas + 1;
                                                    $types_schema->{filename} = "internal_types$external_schema_count.xsd";
                                                    push(@internal_types_schemas, $types_schema);

                                                    open(my $outf, '>', $schema_dir . "/" . $internal_types_schemas[-1]->{filename});
                                                    print { $outf } xml_header();
                                                    $schema->print($outf, 'indented');
                                                    close $outf;
                                                  }

                                                },

                                                'wsdl:message' => sub {
                                                  my($twig, $elm) = @_;
                                                  extract_message(elm => $elm, 
                                                                  xml_schema_prefix => 'xsd',
                                                                  current_wsdl => $current_wsdl,
                                                                  globally_allowed_namespaces => \%globally_allowed_namespaces,
                                                                 );
                                                },

                                                'wsdl:portType/wsdl:operation' => sub {
                                                  my($twig, $elm) = @_;

                                                  my $method_name = $elm->att('name');

                                                  for my $method_direction (qw(input output)) {
                                                      if (my $method_elm = $elm->first_child("wsdl:$method_direction")) {
                                                          my $method_type = $method_elm->att('message');
                                                          my $method_type_namespace = '';
                                                          if ($method_type =~ s/^(.*)://) {
                                                              $method_type_namespace = get_namespace_of_prefix($method_elm, $1);
                                                          }

                                                          if ($method_direction eq 'output') {
                                                              $method_name = $method_type;
                                                          }
                                                          if ($method_name =~ /^[.\w]+$/) {
                                                              $method_type =~ s/ /__/g;
                                                              $method_type .= "__method";
                                                              $current_wsdl->{methods}->{$method_name}->{method_name} = $method_name;
                                                              if ($method_direction eq 'output') {
                                                                  $current_wsdl->{methods}->{$method_name}->{is_output} = 1;
                                                              }
                                                              $current_wsdl->{methods}->{$method_name}->{method_type} = $method_type;
                                                              $current_wsdl->{methods}->{$method_name}->{method_type_namespace} = $method_type_namespace;
                                                              $current_wsdl->{methods}->{$method_name}->{namespace} ||= $current_wsdl->{namespace};

                                                              for my $action_xmlns_prefix (qw(wsa wsa2)) {
                                                                  if ($method_elm->att("$action_xmlns_prefix:Action")) {
                                                                      $current_wsdl->{methods}->{$method_name}->{soap_action} = $method_elm->att("$action_xmlns_prefix:Action");
                                                                  }
                                                              }
                                                          }
                                                      }
                                                  }
                                                },

                                                'wsdl:binding/wsdl:operation' => sub {
                                                  my($twig, $elm) = @_;
                                                  my $method_name = $elm->att('name');

                                                  my $soap_operation_node = $elm->first_child('soap:operation');
                                                  unless ($soap_operation_node) {
                                                    $soap_operation_node = $elm->first_child('soap12:operation');
                                                  }
                                                  if ($soap_operation_node) {
                                                    $current_wsdl->{methods}->{$method_name}->{soap_action} = $soap_operation_node->att('soapAction');
                                                  }

                                                  my @nodes = $elm->findnodes('wsdl:input/soap:body');
                                                  unless (@nodes) {
                                                    @nodes = $elm->findnodes('wsdl:input/soap12:body');
                                                  }
                                                  if (@nodes) {
                                                    my $soap_body_node = $nodes[0];

                                                    if ($soap_body_node->att('namespace')) {
                                                      $current_wsdl->{methods}->{$method_name}->{namespace} = $soap_body_node->att('namespace');
                                                    }

                                                    if ($soap_body_node->att('parts')) {
                                                      $current_wsdl->{methods}->{$method_name}->{parts} = [ split(/\s+/, $soap_body_node->att('parts')) ];
                                                    }

                                                    if ($current_wsdl->{methods}->{$method_name} &&
                                                        ! $current_wsdl->{methods}->{$method_name}->{method_name}) {

                                                        $current_wsdl->{methods}->{$method_name}->{method_name} = $method_name;
                                                        $current_wsdl->{methods}->{$method_name}->{namespace} ||= $current_wsdl->{namespace};
                                                    }

                                                  }
                                                },

                                                'wsdl:service/wsdl:port/soap:address' => sub {
                                                  my($twig, $elm) = @_;
                                                  push(@soap_objects, $elm->att('location')); #todo: tie to specific wsdl?
                                                },

                                                'wsdl:service/wsdl:port/soap12:address' => sub {
                                                  my($twig, $elm) = @_;
                                                  push(@soap_objects, $elm->att('location')); #todo: tie to specific wsdl?
                                                }
                                               },
                              map_xmlns => {
                                            "http://schemas.xmlsoap.org/wsdl/" => 'wsdl',
                                            'http://schemas.xmlsoap.org/wsdl/soap/' => 'soap',
                                            'http://schemas.xmlsoap.org/wsdl/soap12/' => 'soap12',
                                            'http://www.w3.org/2005/03/addressing' => 'wsa',
                                            'http://www.w3.org/2006/05/addressing/wsdl' => 'wsa2',
                                            $XML_SCHEMA_URI => 'xsd',
                                           },
                              keep_original_prefix => 1,
                              output_filter => 'safe',
                             );

    if ($flg_follow_schema_links) {
      eval {
        $twig->xparse($current_wsdl_filename);
      };
      if ($@) {
        if ($current_wsdl_filename =~ m{^\w+://}) {
          # if the current filename looks like a URL
          die "Couldn't access WSDL: $current_wsdl_filename\n";
        }
        else {
          die "$@\n";
        }
      }
    }
    else {
      $twig->safe_parsefile($current_wsdl_filename);
    }

    unless ($converted_schemas{$current_wsdl->{namespace}}) {
      $converted_schemas{$current_wsdl->{namespace}} = { namespace => $current_wsdl->{namespace},
                                                         xmlns_declarations => { %{$current_wsdl->{xmlns_declarations}} } };
    }

    my %referenced_namespaces = ($current_wsdl->{namespace} => 1);
    if ($current_wsdl->{methods}) {

        for my $method (values %{$current_wsdl->{methods}}) {
            next unless ($method->{method_name});

            unless ($method->{is_output}) {
                my $method_key = $method->{namespace} . $ARBITRARY_DIVIDER . $method->{method_name};
                if (! $all_methods{$method_key}) {
                    $all_methods{$method_key} = {};
                }
                $all_methods{$method_key} = { %{$all_methods{$method_key}},
                                              %$method };
            }

            unless ($converted_schemas{$method->{namespace}}) {
                $converted_schemas{$method->{namespace}} = { namespace => $method->{namespace},
                                                             xmlns_declarations => {},
                                                           };
            }
            $referenced_namespaces{$method->{namespace}} = 1;
            $converted_schemas{$method->{namespace}}->{methods}->{$method->{method_name}} =  $method;
        }
    }

    for my $namespace (keys %referenced_namespaces) {
        push(@{$converted_schemas{$namespace}->{complex_types}}, @{$current_wsdl->{complex_types}}) if ($current_wsdl->{complex_types});
        %{$converted_schemas{$namespace}->{xmlns_declarations}} = (%{$converted_schemas{$namespace}->{xmlns_declarations}},
                                                                   %{$current_wsdl->{xmlns_declarations}});
    }

  }

  if (@missing_composition_wsdls) {
    return({
            completed_successfully => 0,
            missing_files => \@missing_composition_wsdls,
            error_msg => [ "Missing imports required - ". join(', ', @missing_composition_wsdls) ],
           });
  }


  my $converted_schema_count = 0;
  for my $converted_schema (values %converted_schemas) {
    $converted_schema->{filename} = "wsdl_actions$converted_schema_count.xsd";
    $converted_schema_count++;
  }

  for my $converted_schema (values %converted_schemas) {
    my $xml_schema_prefix = get_custom_namespace_prefix(desired_prefix => 'xsd',
                                                        namespace => $XML_SCHEMA_URI,
                                                        converted_schema => $converted_schema);

    my $target_prefix_no_xmlns;
    if ($converted_schema->{namespace}) {
        $target_prefix_no_xmlns = get_custom_namespace_prefix(desired_prefix => 'tns',
                                                              namespace => $converted_schema->{namespace},
                                                              converted_schema => $converted_schema);
    }

    for my $complex_type (@{$converted_schema->{complex_types}}) {

      my %allowed_parts;
      for my $method (values %{$converted_schema->{methods}}) {
        if ($complex_type->att('name') eq $method->{method_type} &&
            $method->{parts}) {
          %allowed_parts = map { ($_ => 1) } @{$method->{parts}};
        }        
      }

      for my $element ($complex_type->findnodes(".//xsd:element")) {

        # Add namespace prefix for unprefixed complex types
        if ($element->att('type') && $element->att('type') !~ /:/ && $target_prefix_no_xmlns) {
          $element->set_att(type => "$target_prefix_no_xmlns:" . $element->att('type'));
        }

        # Remove all elements that aren't listed in the method's parts
        my $check_name = $element->att('name');
        if (!$check_name) {
          ($check_name) = ($element->att('ref') =~ /:(.*)$/);
        }
        if ($check_name && %allowed_parts && ! exists $allowed_parts{$check_name}) {
          $element->delete;
        }
      }
    }

    my @method_tags;
    for my $method (values %{$converted_schema->{methods}}) {
      if ($existing_global_elements->{$method->{namespace}} &&
          $existing_global_elements->{$method->{namespace}}->{$method->{method_name}}) {
        next;
      }

      next unless $method->{method_type};

      my $method_type = $method->{method_type};
      if ($method->{method_type_namespace}) {
          my $method_type_prefix = get_custom_namespace_prefix(desired_prefix => 'tns',
                                                               namespace => $method->{method_type_namespace},
                                                               converted_schema => $converted_schema);

        $method_type = "$method_type_prefix:$method_type";
      }
      elsif ($method_type !~ /:/ && $target_prefix_no_xmlns) {
        $method_type = "$target_prefix_no_xmlns:$method_type";
      }

      push(@method_tags,
           XML::Twig::Elt->new("$xml_schema_prefix:element",
                               {
                                name => $method->{method_name},
                                type => $method_type,
                               }));
    }

    my @imports_includes = ();
    for my $other_schema ((@internal_types_schemas, values %converted_schemas,  @imported_schemas)) {
      next if $other_schema eq $converted_schema;
      next if $other_schema->{no_import};

      if (!$other_schema->{namespace} || $other_schema->{namespace} eq $converted_schema->{namespace}) {
        push(@imports_includes,
             XML::Twig::Elt->new("$xml_schema_prefix:include",
                                 { schemaLocation => $other_schema->{filename} }));
      }
      else {
        next unless $other_schema->{filename};

        # Skip if this schema will be included by any other converted_schema
        if (!grep($other_schema eq $_, values %converted_schemas) &&
            grep($other_schema->{namespace} eq $_->{namespace}, values %converted_schemas)) {
          next;
        }

        my %namespace_opts;
        if ($other_schema->{namespace}) {
            $namespace_opts{namespace} = $other_schema->{namespace};
        }
        push(@imports_includes,
             XML::Twig::Elt->new("$xml_schema_prefix:import",
                                 { schemaLocation => $other_schema->{filename},
                                   %namespace_opts,
                                 }));
      }
    }

    my %target_namespace_opts;
    if ($converted_schema->{namespace}) {
        $target_namespace_opts{targetNamespace} = $converted_schema->{namespace};
    }
    open(my $outf, '>', $schema_dir . "/" . $converted_schema->{filename});
    print { $outf } xml_header();
    my $wsdl_actions = XML::Twig::Elt->new("$xml_schema_prefix:schema",
                                           { %target_namespace_opts, 
                                             %{$converted_schema->{xmlns_declarations}},
                                           },
                                           @imports_includes,
                                           @{$converted_schema->{complex_types}},
                                           @method_tags);
    $wsdl_actions->print($outf, 'indented');
    close $outf;

  }

  create_modified_soapenv_schema(converted_schemas => \%converted_schemas,
                                 globally_allowed_namespaces => \%globally_allowed_namespaces,
                                 schema_dir => $schema_dir);

  copy("$GENERIC_SCHEMAS_ROOT/soap-encoding.xsd",
       "$schema_dir/soap-encoding.xsd");

  my %schema_mappings = ();
  for my $schema ((values %converted_schemas, @internal_types_schemas)) {
    if ($schema->{filename}) {
      $schema_mappings{$schema->{filename}} = $schema->{filename};
    }
  }

  return {
          completed_successfully => 1,
          methods => \%all_methods,
          soap_address => \@soap_objects,
          schema_document => "$schema_dir/base.xsd",
          schema_composition_mappings => {
                                          'soap-encoding.xsd' => "$schema_dir/soap-encoding.xsd",
                                          'wsdl_actions.xsd' => "$schema_dir/wsdl_actions.xsd",
                                          'xml.xsd' => "$schema_dir/xml.xsd",
                                          %schema_mappings,
                                         },

         };
}

#----------------------------------------------------------------------
# Convert WSDL message into schema complex type

sub extract_message {
  my(%params) = @_;

  my $elm = $params{elm};
  my $xml_schema_prefix = $params{xml_schema_prefix};
  my $current_wsdl = $params{current_wsdl};
  my $globally_allowed_namespaces = $params{globally_allowed_namespaces};

  my $sequence = XML::Twig::Elt->new("$xml_schema_prefix:sequence");
  for my $part ($elm->children('wsdl:part')) {
    my %tag_ns;
    for my $att_name ($part->att_names) {
      if ($att_name =~ /^xmlns/) {
        $tag_ns{$att_name} = $part->att($att_name);
      }
    }

    if ($part->att('type')) {
      my $part_type = $part->att('type');
      my $part_as_element = XML::Twig::Elt->new("$xml_schema_prefix:element",
                                                { name => $part->att('name'),
                                                  type => $part_type,
                                                  %tag_ns,
                                                });
      $part_as_element->paste(last_child => $sequence);
    }
    if ($part->att('element')) {
      my($prefix, $element_name) = split(/:/, $part->att('element'));
      unless ($element_name) {
        $element_name = $prefix;
        $prefix = '';
      }

      if ($prefix) {
        my $element_namespace = get_namespace_of_prefix($part, $prefix);
        $current_wsdl->{methods}->{$element_name}->{namespace} = $element_namespace;

        if (!$current_wsdl->{xmlns_declarations}->{"xmlns:$prefix"}) {
          $current_wsdl->{xmlns_declarations}->{"xmlns:$prefix"} = $element_namespace;
        }
        $globally_allowed_namespaces->{$current_wsdl->{xmlns_declarations}->{"xmlns:$prefix"}} = 1;
      }
      my $part_as_element = XML::Twig::Elt->new("$xml_schema_prefix:element",
                                                { ref => $part->att('element'),
                                                  %tag_ns,
                                                });
      $part_as_element->paste(last_child => $sequence);
    }
  }

  my $complex_type_name = $elm->att('name');
  $complex_type_name =~ s/ /__/g;
  $complex_type_name .= '__method';

  push(@{$current_wsdl->{complex_types}},
       XML::Twig::Elt->new("$xml_schema_prefix:complexType",
                           {
                            name => $complex_type_name},
                           $sequence,
                           XML::Twig::Elt->new("$xml_schema_prefix:anyAttribute",
                                               {
                                                namespace => "http://schemas.xmlsoap.org/soap/envelope/ http://www.w3.org/2003/05/soap-envelope"})));
}

#----------------------------------------------------------------------

sub xml_header {
  return '<?xml version="1.0" encoding="utf-8"?>';
}

#----------------------------------------------------------------------

sub create_modified_soapenv_schema {
  my(%params) = @_;

  my $converted_schemas = $params{converted_schemas};
  my $schema_dir = $params{schema_dir};
  my $globally_allowed_namespaces = $params{globally_allowed_namespaces};

  my $twig = F5::XML::Twig->new( 
                            keep_spaces => 1,
                            twig_handlers   => { 'xs:complexType[@name = "Body"]/xs:sequence' => 
                                                 sub {
                                                   my($twig, $elm) = @_;

                                                   $elm = $elm->parent;
                                                   $elm->cut_children;
                                                   $elm = XML::Twig::Elt->new('xs:choice')->paste(first_child => $elm);

                                                   my %allowed_namespaces = map { ($_ => 1) } (keys %$globally_allowed_namespaces, 
                                                                                               map { $_->{namespace} } values %$converted_schemas);
                                                   for my $allowed_namespace (keys %allowed_namespaces) {
                                                     my $any_tag = XML::Twig::Elt->new('xs:any',
                                                                                       { namespace => $allowed_namespace,
                                                                                         minOccurs => 1,
                                                                                         maxOccurs => 1,
                                                                                         processContents => 'strict' });
                                                     $any_tag->paste(last_child => $elm);
                                                   }
                                                 },
                                                 'xs:schema' => sub {
                                                   my($twig, $elm) = @_;

                                                   for my $converted_schema (sort { $b->{filename} cmp $a->{filename} } values %$converted_schemas) {
                                                       my %namespace_opts;
                                                       if ($converted_schema->{namespace}) {
                                                           $namespace_opts{namespace} = $converted_schema->{namespace};
                                                       }

                                                     my $import = XML::Twig::Elt->new('xs:import',
                                                                                      {
                                                                                       schemaLocation => $converted_schema->{filename},
                                                                                       %namespace_opts,
                                                                                      });
                                                     $import->paste(first_child => $elm);
                                                   }
                                                 }
                                               },
                           );

  my @soap_envelope_schemas = (
                               { src => "$GENERIC_SCHEMAS_ROOT/soap-envelope.xsd",
                                 dest => "$schema_dir/base.xsd" },
                               { src => "$GENERIC_SCHEMAS_ROOT/soap-envelope-12.xsd",
                                 dest => "$schema_dir/soap-envelope-12.xsd" },
                              );

  for my $soap_envelope_schema (@soap_envelope_schemas) {
    $twig->parsefile($soap_envelope_schema->{src});
    open(my $soapenv_xsd_fh, '>', $soap_envelope_schema->{dest});
    $twig->print($soapenv_xsd_fh);
    close $soapenv_xsd_fh;
  }

  copy("$GENERIC_SCHEMAS_ROOT/xml.xsd", "$schema_dir/xml.xsd");
}

#----------------------------------------------------------------------

sub find_next_available_name {
  my(%params) = @_;

  my $target = $params{target};
  my %reserved = map { ($_ => 1) } @{$params{reserved}};

  return $target if (!exists $reserved{$target});

  my $target_count = 1;
  while (1) {
    if (!exists $reserved{"$target$target_count"}) {
      return "$target$target_count";
    }
    $target_count++;
  }
}

#----------------------------------------------------------------------

sub is_xml {
  my(%params) = @_;
  my $xml_contents = $params{xml_contents};
  my $xml_filename = $params{xml_filename};
  my $die_if_missing = $params{die_if_missing};

  my $p = XML::Parser->new();
  my $is_xml = 0;
  my $err;

  # parsefile/parse will
  #   return 1 for well-formed xml
  #   die in any other case
  if ($xml_contents) {
    eval {
      $is_xml = $p->parse($xml_contents);
    };
    $err = $@;
  }
  elsif ($xml_filename) {
    eval {
      $is_xml = $p->parsefile($xml_filename);
    };
    $err = $@;
  }

  if ($err) {
    die $err if $die_if_missing;
  }

  return $is_xml;
}

#----------------------------------------------------------------------

sub is_wsdl {
  my(%params) = @_;

  my $flg_follow_schema_links = (exists $params{flg_follow_schema_links} ? $params{flg_follow_schema_links} : 1);
  my $xml_contents = $params{xml_contents};
  my $xml_filename = $params{xml_filename};
  my $die_if_missing = $params{die_if_missing};

  my $is_wsdl = 0;
  my $twig = F5::XML::Twig->new( start_tag_handlers => {
                                                    'wsdl:definitions' => sub { 
                                                      my($twig, $elm) = @_;
                                                      $is_wsdl = 1;
                                                      $twig->ignore;
                                                    }
                                                   },
                             map_xmlns => {
                                           "http://schemas.xmlsoap.org/wsdl/" => 'wsdl',
                                           'http://schemas.xmlsoap.org/wsdl/soap/' => 'soap',
                                          },
                             keep_original_prefix => 1,
                           );

  my $err;
  if ($flg_follow_schema_links) {
    my $xml_document = $xml_contents || $xml_filename;
    eval {
      $twig->xparse($xml_document);
    };
    if ($@) {
      $err = $@;
    }
  }
  else {
    if ($xml_contents) {
      $twig->safe_parse($xml_contents);
    }
    elsif ($xml_filename) {
      if (!-e $xml_filename) {
        $err = "File is missing: $xml_filename\n";
      }
      else {
        $twig->safe_parsefile($xml_filename);
      }
    }
  }

  if ($err) {
    die $err if $die_if_missing;
  }

  return $is_wsdl;
}

#----------------------------------------------------------------------

sub handle_schema_import {
  my(%params) = @_;

  my $imported_schemas = $params{imported_schemas};
  my $flg_follow_schema_links = $params{flg_follow_schema_links};
  my $import_filename = $params{import_filename};
  my $original_location = $params{original_location};
  my $import_namespace = $params{import_namespace};
  my $existing_global_elements = $params{existing_global_elements};

  if (!grep($_->{namespace} eq $import_namespace, @$imported_schemas)) {
    push(@$imported_schemas,
         { namespace => $import_namespace,
           filename => $original_location });
  }

  $import_filename =~ s|^file:/+|/|;

  if ($flg_follow_schema_links || -e $import_filename) {
    my $imported_schema_twig = F5::XML::Twig->new(
                                              twig_roots => {
                                                             '/xsd:schema/xsd:element' => sub {
                                                               # Discover all global element declarations in imported schema
                                                               my($twig, $elm) = @_;
                                                               $existing_global_elements->{$import_namespace}->{$elm->att('name')} = 1;
                                                             }
                                                            },
                                              map_xmlns => {
                                                            "http://schemas.xmlsoap.org/wsdl/" => 'wsdl',
                                                            'http://schemas.xmlsoap.org/wsdl/soap/' => 'soap',
                                                            $XML_SCHEMA_URI => 'xsd',
                                                           },
                                              keep_original_prefix => 1,
                                             );
    if ($flg_follow_schema_links) {
      eval {
        $imported_schema_twig->xparse($import_filename);
      };
    }
    else {
      $imported_schema_twig->safe_parsefile($import_filename);
    }
  }
}

#----------------------------------------------------------------------

# Copied from F5::XML::Schema::Processor, to avoid requiring the include
sub get_namespace_of_prefix {
	my ($twig, $prefix) = @_;
#	return XMLSCHEMAURL if $prefix eq 'xsd' && not defined $twig;
	return '' unless defined $twig;
	my $xmlns = 'xmlns';
	if (
		defined $prefix and
		$prefix ne ''   and 
		$prefix ne '#default' # deal with XML::Parser::Expat conventions
	) {
		$xmlns .= ':'.$prefix ;
	}
	return $twig->att($xmlns) if defined $twig->att($xmlns);
	return get_namespace_of_prefix($twig->parent,$prefix);
}

#----------------------------------------------------------------------

sub get_custom_namespace_prefix {
    my(%params) = @_;

    my $desired_prefix = $params{desired_prefix};
    my $namespace = $params{namespace};
    my $converted_schema = $params{converted_schema};

    if (!$converted_schema->{added_custom_namespaces}->{$namespace}) {
        my $actual_prefix = find_next_available_name(target => "xmlns:$desired_prefix",
                                                     reserved => [ keys %{$converted_schema->{xmlns_declarations}} ]);
        $converted_schema->{xmlns_declarations}->{$actual_prefix} = $namespace;
        my $prefix = $actual_prefix;
        $prefix =~ s/^xmlns://;

        $converted_schema->{added_custom_namespaces}->{$namespace} = $prefix;
    }

    return $converted_schema->{added_custom_namespaces}->{$namespace};
}

#----------------------------------------------------------------------

1;
