#!/usr/bin/perl -w

package F5::XML::Xpath::xpathprocessor;
use strict;
use warnings;

use File::Basename;
use Data::Dumper;
use MIME::Base64;


use constant XP_VERSION   => 2;

use constant DEBUG         => 0; # set 1 to enable debug print

use constant XP_NO_VALUE   => -1;
use constant XP_ANY_VALUE  => -2;

use constant XP_QUERY_TYPE_ORIG     => 1;
use constant XP_QUERY_TYPE_COND     => 2;
use constant XP_QUERY_TYPE_COND_EPS => 3;

# state flags
use constant XP_ACCEPTING                => 0x1;
use constant XP_CHILD_ELEMENT_ASTERISK   => 0x2;
use constant XP_CHILD_ATTRIBUTE_ASTERISK => 0x4;
use constant XP_CHILD_EPSILON            => 0x8;
use constant XP_SELF_EPSILON             => 0x10;
use constant XP_ONLY_NS                  => 0x20;
use constant XP_CONDITIONS               => 0x40;

# trans flags
use constant XP_ATTRIBUTE  => 0x1;
use constant XP_NO_NS      => 0x2;
use constant XP_ANY_NS     => 0x4;

sub read_namespaces_from_hash {
  my ($xpc, $xpd) = @_;
  my ($prefix, $ns);
  foreach $prefix (keys %{$xpc->{xpath_namespaces_hash}}) {
    $ns = $xpc->{xpath_namespaces_hash}->{$prefix};
    $xpd->{ns}->{$prefix} = $ns;
  }
  #add empty namespace
  $xpd->{ns}->{''} = '';
}

# read queries
sub read_queries_list {
  my ($xpc, $xpd) = @_;
  my $query;

  foreach (@{$xpc->{xpath_queries_list}}) {
    chomp;
    next unless /\S/;
    s|^([^/])|//$1|;        # make all relative queries absolute
    s|^\*:\*$|*|;            # *:* -> *
    $query = $_;
    $xpd->{queries}->[$xpd->{query_num}]->{id} = $xpd->{query_num};
    $xpd->{queries}->[$xpd->{query_num}]->{scope} = XP_NO_VALUE;
    $xpd->{queries}->[$xpd->{query_num}]->{orig_query} = $query;
    $xpd->{queries}->[$xpd->{query_num}]->{type}  = XP_QUERY_TYPE_ORIG;
    $query =~ s/\[(.*?)\]/f($1, $xpd, $xpd->{query_num})/ge;
    $xpd->{queries}->[$xpd->{query_num}]->{query} = $query;
    $xpd->{query_num}++;
    #shift @queriesdefs;
  }
}

sub find_index {
    my($string, $hash) = @_;

    if ($string) {
	return XP_ANY_VALUE if $string eq '*';
	if (exists $$hash{$string}) {
	    return $$hash{$string};
	} else {
	    return $$hash{''}; # 'no-namespace': this should come out to 0
	}
    } else {
	return XP_NO_VALUE; # no such key
    }
}

sub f {
  my ($arg, $xpd, $orig_query_num) = @_;
  $xpd->{cond_part}->[$xpd->{cond_part_num}]->{pquery} = $arg;
  $xpd->{cond_part}->[$xpd->{cond_part_num}]->{orig}   = $orig_query_num;
  $xpd->{cond_part_num}++;
  return "[".($xpd->{cond_part_num}-1)."]";
}

sub process_query{
  my ($xpc, $xpd, $parent, $cur, $query, $id, $orig_id, $norm_query) = @_;
  my ($selfq, $op, $value);
  
  if($query =~ m|^\s*(.*?)\s+(.*?)\s+["']?(.*?)["']?\s*$|) { #'
    ($selfq, $op, $value) = ($1, $2, $3);
    if($op !~ /=|!=|<|>|<=|>=/) {
      push_error($xpc, "unsupported operator ' $op ' in query $orig_id");
      return;
    }
    # value predicate
    if(defined $selfq and $selfq eq ".") {
      #$xpd->{queries}->[$orig_id]->{op}    = $op;
      #$xpd->{queries}->[$orig_id]->{value} = $value;
      $xpd->{queries}->[$id]->{op}    = $op;
      $xpd->{queries}->[$id]->{value} = $value;
      
      $xpd->{values}->{$value} = 1; # collect literals
      $xpd->{operators}->{$op} = 1; # collect literals
      push @{$xpd->{state}->[$parent]->{query}}, { id => $id, query => $norm_query };
      $xpd->{state}->[$parent]->{sflags} |= XP_ACCEPTING;
      return $norm_query;
    }
    $query = $selfq;
  }
  
  $query =~ s/(.*)\s*/$1/;
  
  my @parts = split(/\//, $query);
  # delete empty first part (when query begins with a slash)
  shift @parts unless $parts[0];
  
  if($id != $orig_id and not $parts[0]) {
    $xpd->{queries}->[$id]->{starts_with_epsilon} = 1;
  }
  
  foreach my $part (@parts) {
    my ($attribute, $prefix, $name, $predicate);
    if($part =~ m|(.*?)\[(.*)\]|) {
      ($part, $predicate) = ($1, $2);
    }
    $part =~ s|(\*:\*)|*|;
    if($part ne ".") {
      ($attribute, $prefix, $name) = ($part =~ m|^(\@)?(?:(.*):)?(.*)$|);
      $norm_query .= "/".$part;
      if (exists $cur->{$part}) {
	$parent = $xpd->{nodestate}->{$cur->{$part}};
      } else {
	$xpd->{num_states}++;
	$cur->{$part} = {};
	$xpd->{nodestate}->{$cur->{$part}} = $xpd->{num_states};
	push @{$xpd->{state}->[$parent]->{child}}, $xpd->{num_states};
	$xpd->{state}->[$xpd->{num_states}]->{name}    = $name;
	$xpd->{state}->[$xpd->{num_states}]->{prefix}  = $prefix;
	$xpd->{state}->[$xpd->{num_states}]->{tflags} |= XP_ATTRIBUTE if $attribute;
	if ($name eq '*') {
      if(defined $prefix) {
	  $xpd->{state}->[$parent]->{sflags} |= XP_ONLY_NS;
      }
	      else {
	  $xpd->{state}->[$parent]->{sflags} |= ($attribute ? XP_CHILD_ATTRIBUTE_ASTERISK : XP_CHILD_ELEMENT_ASTERISK);
      }
	      $xpd->{state}->[$parent]->{asterisk_child} = $xpd->{num_states};
	}
	if ($name eq '') {
	      $xpd->{state}->[$parent]->{sflags} |= XP_CHILD_EPSILON;
	      $xpd->{state}->[$xpd->{num_states}]->{sflags} |= XP_SELF_EPSILON;
	}
	$xpd->{strings}->{$name} = 1; # collect literals
	$parent = $xpd->{num_states};
      }
      $cur = $cur->{$part};
    }
    # predicate
    if(defined $predicate) {
      my @conditions;
      my $operator = "and";
      if($xpd->{cond_part}->[$predicate]->{pquery} =~ /and/) {
	@conditions = split "and", $xpd->{cond_part}->[$predicate]->{pquery};
      }
      elsif($xpd->{cond_part}->[$predicate]->{pquery} =~ /or/) {
	@conditions = split "or", $xpd->{cond_part}->[$predicate]->{pquery};
	$operator = "or";
      }
      else {
	@conditions = ($xpd->{cond_part}->[$predicate]->{pquery});
      }
      foreach my $condition (@conditions) {
	my $cond_query = process_query($xpc, $xpd, $parent, $cur, $condition, $xpd->{query_num}, $id, $norm_query);
	if(defined $cond_query) {
	  $xpd->{queries}->[$id]->{scope} = $parent unless (defined $xpd->{queries}->[$id]->{scope} and $xpd->{queries}->[$id]->{scope} ne -1);
	  $xpd->{state}->[$parent]->{condition_num} = 0 unless defined $xpd->{state}->[$parent]->{condition_num};
	  $xpd->{state}->[$parent]->{sflags} |= XP_CONDITIONS;
	  push @{$xpd->{state}->[$parent]->{condition}}, { orig=>$id, cond=>$xpd->{query_num}, op=>$operator };
	  $xpd->{state}->[$parent]->{condition_num}++;
	  $xpd->{queries}->[$xpd->{query_num}]->{id}  = $xpd->{query_num};
	  $xpd->{queries}->[$xpd->{query_num}]->{scope}  = $parent;
	  $xpd->{queries}->[$xpd->{query_num}]->{type}  = XP_QUERY_TYPE_COND;
	  if($xpd->{queries}->[$xpd->{query_num}]->{starts_with_epsilon}) {
	    $xpd->{queries}->[$xpd->{query_num}]->{type}  = XP_QUERY_TYPE_COND_EPS;    
	  }
	  $xpd->{queries}->[$xpd->{query_num}]->{query} = $cond_query;
	  $xpd->{queries}->[$xpd->{query_num}]->{cond_for} = $xpd->{cond_part}->[$predicate]->{orig};
	  $xpd->{query_num}++;
	}
      }
    }
  }
  
  # value predicate
  if(defined $op and $op ne "") {
    $xpd->{queries}->[$id]->{op}    = $op;
    $xpd->{queries}->[$id]->{value} = $value;
    $xpd->{values}->{$value} = 1; # collect literals
    $xpd->{operators}->{$op} = 1; # collect literals
  }
  push @{$xpd->{state}->[$parent]->{query}}, { id => $id, query => $norm_query };
  $xpd->{state}->[$parent]->{sflags} |= XP_ACCEPTING;
  return $norm_query;
}

sub process_blob{
  my ($xpc, $xpd) = @_;
  
  # process blob
  my (@sorted_string, %string_index);
  my (@sorted_value, %value_index);
  my (@sorted_op, %op_index);
  my (@sorted_ns, %ns_index);
  my (@orig_queries_list, @queries_list, %qs_index);
  
  @sorted_string = sort keys %{$xpd->{strings}};
  @string_index{@sorted_string} = (0..$#sorted_string);
  
  @sorted_value = sort keys %{$xpd->{values}};
  @value_index{@sorted_value} = (0..$#sorted_value);
  
  @sorted_op = sort keys %{$xpd->{operators}};
  @op_index{@sorted_op} = (0..$#sorted_op);
  
  @sorted_ns = sort keys %{$xpd->{ns}};
  @ns_index{@sorted_ns} = (0..$#sorted_ns);
  
  foreach my $query (@{$xpd->{queries}}) {
    if(exists $query->{orig_query}) {
      push @orig_queries_list, $query->{orig_query};
    }
    else {
      push @orig_queries_list, $query->{query};
    }
    push @queries_list, $query->{query};
  }
  @qs_index{@queries_list} = (0..$#queries_list);
  
  my ($query_blob, $pre_blob, $state_blob, $rest_blob) = ('','','','');
  my $num_trans = 0;
  
  my (@state_entry, @trans_entry);
  
  #print STDERR Data::Dumper::Dumper(\@{$xpd->{state}}) if DEBUG;
  #print STDERR Data::Dumper::Dumper($xpd->{num_states}+1) if DEBUG;
  my (@acceptig_queries_list) = ();
  my (@condition_tables_list) = ();
  
  for(my $s=0; $s<=$xpd->{num_states}; $s++) {
      my $srec = $xpd->{state}->[$s];
      my $sflags = $srec->{sflags} || 0;
      my ($query_num, $first_query) = (0, XP_NO_VALUE);
      my ($cond_num, $first_cond) = (0, XP_NO_VALUE);
      
      if($sflags & XP_ACCEPTING) {
          $first_query = scalar(@acceptig_queries_list);
          foreach my $q_num (@{$srec->{query}}) {
              push @acceptig_queries_list, $q_num->{id};
              $query_num++;
          }
      }
      
      if($sflags & XP_CONDITIONS) {
          $first_cond = scalar(@condition_tables_list);
          foreach my $cond (@{$srec->{condition}}) {
              push @condition_tables_list, $cond;
              $cond_num++;
          }
      }
      
      my $first_trans = $num_trans;
      my $asterisk_state = $srec->{asterisk_child} || XP_NO_VALUE;
      my $epsilon_state = XP_NO_VALUE;
      foreach my $trans (@{$srec->{child}}) {
        my $t = $xpd->{state}->[$trans];
        my $tflags = $t->{tflags} || 0;
        my $ns_index = find_index($t->{prefix}, \%ns_index);
        my $name_index = find_index($t->{name}, \%string_index);
        if ($ns_index == XP_NO_VALUE) {
	  if($name_index == XP_ANY_VALUE) {
	    $tflags |= XP_ANY_NS;
	  }
          else {
	    $tflags |= XP_NO_NS;
	  }
        } elsif ($ns_index == XP_ANY_VALUE) {
            $tflags |= XP_ANY_NS;
            $ns_index = XP_NO_VALUE;
        } #elsif ($name_index == XP_ANY_VALUE) { # have ns and name *
          #  $sflags |= XP_ONLY_NS;
          #  $sflags &= (~0 & ~(XP_CHILD_ATTRIBUTE_ASTERISK | XP_CHILD_ELEMENT_ASTERISK));
        #} 
        $epsilon_state = $trans if exists $t->{sflags} && $t->{sflags} & XP_SELF_EPSILON;
        push @trans_entry, [$tflags, $ns_index, $name_index, XP_NO_VALUE, XP_NO_VALUE, $s, $trans, XP_NO_VALUE];
        $num_trans++;
      }
      push @state_entry, [$sflags, scalar(@{$srec->{child}}), $first_trans, $asterisk_state, $epsilon_state, $query_num, $first_query, $cond_num, $first_cond, XP_NO_VALUE, XP_NO_VALUE, XP_NO_VALUE, XP_NO_VALUE];
  }
  
  print STDERR "\n1)----------queries----------------(".scalar(@{$xpd->{queries}}).")\n" if DEBUG;
  print STDERR "  [name placeholder, type, scope, op, value, predicate placeholder]\n" if DEBUG;
  for(my $i=0; $i < scalar(@{$xpd->{queries}}); $i++) {
      my $entry = $xpd->{queries}->[$i];
      print STDERR "query $i = [" if DEBUG;
      my $value = (exists $entry->{value}) ? $value_index{$entry->{value}} : XP_NO_VALUE;
      my $op    = (exists $entry->{op}) ? $op_index{$entry->{op}} : XP_NO_VALUE;
      $query_blob .= pack('I', XP_NO_VALUE)    ; print STDERR "-1 " if DEBUG;              # name placeholder
      $query_blob .= pack('I', XP_NO_VALUE)    ; print STDERR "-1 " if DEBUG;              # name placeholder
      $query_blob .= pack('I', $entry->{type}) ; print STDERR $entry->{type}." " if DEBUG;  # type
      $query_blob .= pack('I', $entry->{scope}); print STDERR $entry->{scope}." " if DEBUG; # scope
      $query_blob .= pack('I', $op)            ; print STDERR "$op " if DEBUG;              # operator
      $query_blob .= pack('I', $value)         ; print STDERR "$value " if DEBUG;           # value
      $query_blob .= pack('I', XP_NO_VALUE)    ; print STDERR "-1]\n" if DEBUG;            # predicate placeholder
      $query_blob .= pack('I', XP_NO_VALUE)    ; print STDERR "-1]\n" if DEBUG;            # predicate placeholder
  }
  
  print STDERR "\n2)----------states----------------(".scalar(@state_entry).")\n" if DEBUG;
  print STDERR "  [sflags, num_child, first_trans, * state, eps state, num_queries, first_query, cond_num, first_cond, queries list *, conditions *]\n" if DEBUG;
  for(my $i=0; $i < scalar(@state_entry); $i++) {
      my $entry = $state_entry[$i];
      print STDERR "state $i = [".join(", ",@$entry) ."]\n" if DEBUG;
      foreach my $number (@$entry) {
          $state_blob .= pack('I', $number);
      }
  }
  
  print STDERR "\n3a)-------accepting queries--------(".scalar(@acceptig_queries_list).")\n" if DEBUG;
  my $accqueries_store_size = 0;
  print STDERR "accepting queries [".join(", ",@acceptig_queries_list) ."]\n" if DEBUG;
  foreach my $i (@acceptig_queries_list) {
      $state_blob.= pack('I', $i);
      $accqueries_store_size++;
  }
  
  print STDERR "\n3b)-------condition tables---------(".scalar(@condition_tables_list).")\n" if DEBUG;
  my $cond_table_store_size = 0;
  print STDERR "conditions :\n" if DEBUG;
  foreach my $i (@condition_tables_list) {
      print STDERR "  [$i->{orig}, $i->{cond}, -1]\n" if DEBUG;
      $state_blob.= pack('I', $i->{orig});
      $state_blob.= pack('I', $i->{cond});
      $state_blob.= pack('I', ($i->{op} eq "and")?0:1); # and = 0, or =1
      $state_blob.= pack('I', XP_NO_VALUE);  # depth placeholder
      $state_blob.= pack('I', XP_NO_VALUE);  # flag placeholder
      $cond_table_store_size++;
  }
  
  print STDERR "\n4)---------transitions------------(".scalar(@trans_entry).")\n" if DEBUG;
  print STDERR "  [tflags, ns_index, name_index, -1, -1, s, trans, -1]\n" if DEBUG;
  for(my $i=0; $i < scalar(@trans_entry); $i++) {
    my $entry = $trans_entry[$i];
    print STDERR "trans $i = [".join(", ",@$entry) ."]\n" if DEBUG;
    foreach my $number (@$entry) {
      $rest_blob .= pack('I', $number);
    }
  }
  
  print STDERR "\n5)--------step strings------------(".scalar(@sorted_string).")\n" if DEBUG;
  print STDERR " (".join(", ", @sorted_string).")\n" if DEBUG;
  my $string_store_size = 0;
  foreach my $s (@sorted_string) {
      $rest_blob.= pack('Z*', $s);
      $string_store_size += length($s) + 1;
  }
  
  print STDERR "\n6)------values strings------------(".scalar(@sorted_value).")\n" if DEBUG;
  print STDERR " (".join(", ", @sorted_value).")\n" if DEBUG;
  my $value_store_size = 0;
  foreach my $s (@sorted_value) {
      $rest_blob.= pack('Z*', $s);
      $value_store_size += length($s) + 1;
  }
  
  print STDERR "\n7)---operators strings------------(".scalar(@sorted_op).")\n" if DEBUG;
  print STDERR " (".join(", ", @sorted_op).")\n" if DEBUG;
  my $op_store_size = 0;
  foreach my $s (@sorted_op) {
      $rest_blob.= pack('Z*', $s);
      $op_store_size += length($s) + 1;
  }
  
  print STDERR "\n8)----------ns strings------------(".scalar(@sorted_ns).")\n" if DEBUG;
  print STDERR " (".join(", ", @sorted_ns).")\n" if DEBUG;
  my $ns_store_size = 0;
  foreach my $ns (@sorted_ns) {
      $rest_blob.= pack('Z*', $xpd->{ns}->{$ns});
      $ns_store_size += length($xpd->{ns}->{$ns}) + 1;
  }
  
  print STDERR "\n9)-----queries strings------------(".scalar(@orig_queries_list).")\n" if DEBUG;
  print STDERR " (".join(", ", @orig_queries_list).")\n" if DEBUG;
  my $queries_store_size = 0;
  foreach my $query (@orig_queries_list) {
      $rest_blob.= pack('Z*', $query);
      $queries_store_size += length($query) + 1;
  }
  
  
  print STDERR "sizes of structures:\n" if DEBUG;
  print STDERR " queries:".scalar(@{$xpd->{queries}})."\n" if DEBUG;
  print STDERR " states:".scalar(@state_entry)."\n"        if DEBUG;
  print STDERR " accepting queries list:".$accqueries_store_size."\n" if DEBUG;
  print STDERR " conditions:".$cond_table_store_size."\n" if DEBUG;
  print STDERR " transitions:".scalar(@trans_entry)."\n"   if DEBUG;
  print STDERR "sizes of strings:\n" if DEBUG;
  print STDERR " step keywords:".$string_store_size."\n" if DEBUG;
  print STDERR " values:".$value_store_size."\n"    if DEBUG;
  print STDERR " operators:".$op_store_size."\n"    if DEBUG;
  print STDERR " namespaces:".$ns_store_size."\n"   if DEBUG;
  print STDERR " queries:".$queries_store_size."\n" if DEBUG;
  
  $pre_blob.= pack('I', XP_VERSION);
  $pre_blob.= pack('I', scalar(@{$xpd->{queries}}));
  $pre_blob.= pack('I', scalar(@state_entry));
  $pre_blob.= pack('I', $accqueries_store_size);
  $pre_blob.= pack('I', $cond_table_store_size);
  $pre_blob.= pack('I', scalar(@trans_entry));
  $pre_blob.= pack('I', $string_store_size);
  $pre_blob.= pack('I', $value_store_size);
  $pre_blob.= pack('I', $op_store_size);
  $pre_blob.= pack('I', $ns_store_size);
  $pre_blob.= pack('I', $queries_store_size);
  
  $xpc->{blob} = $pre_blob . $query_blob . $state_blob . $rest_blob;
}

######
## Below is the OO-ish part of this module.
## All methods that are to be part of the API should be
## documented in the pod!
######

=pod

=head1 NAME

F5::XML::Xpath::xpathprocessor - A Perl module for compiling XPath data for use by XPath parser.
Receives a file with XPath queries and a file with prefix-namespace mappings and produces
a binary data-structure representation of the data, loaded later by XPath parser.

=head1 SYNOPSIS

  use F5::XML::Xpath::xpathprocessor;

  my $xpc = F5::XML::Xpath::xpathprocessor->new(
    xpath_namespaces_hash     => \%prefix_namespace_url_mappings,
    xpath_queries_list        => \@xpath_queries,
  );

  eval {
    $xpc->xparse;
  };

  $xpc->print_blob("/path/to/output/binary/file.bin");
  
=head1 METHODS

=head2 new

The constructor is used for generating a new object,
initializing its attributes.

It returns an instance of F5::XML::Xpath::xpathprocessor.

Example usage: 

=begin roff

  my $xpc = F5::XML::Schema::Processor->new(
    xpath_namespaces_hash     => \%prefix_namespace_url_mappings,
    xpath_queries_list        => \@xpath_queries,
  );

=end roff

Detailed description of named parameters (via the hash argument) are:
C<xpath_queries_buf> is required

  
  
=head3 xpath_queries_list

  This parameter is a list reference of queries.
  Each list item in the list should contain a string that is a valid xpath query expression.
  
=head3 xpath_namespaces_hash

  This parameter gets the hash reference with prefix-namespace mappings:  prefix => namespace.
  
=cut

sub new {
  my ($class, %arguments) = @_;

  bless {
    xpath_queries_list    => $arguments{xpath_queries_list},
    xpath_namespaces_hash => $arguments{xpath_namespaces_hash},
    errors                => [],
    blob                  => "",
  }, $class;
}

=pod

=head2 xparse

This method does not expect any arguments.

Example usage:

=begin roff

    $xsp->xparse;

=end roff

This method extruct data from input files and build binary representation
of compiled queries data structures.

=cut

sub xparse {
  $Data::Dumper::Indent   = 3; # pretty print with array indices
  my ($self) = shift;
  
  my $xpd = {
    ns            => {},
    cond_part_num => 0,
    cond_part     => [],
    query_num     => 0,
    queries       => [],
    root          => {},
    num_states    => 0,
    nodestate     => {},
    strings       => {},
    values        => {},
    operators     => {},
  }; #xpath data
  
  if($self->{xpath_namespaces_hash}) {
    read_namespaces_from_hash($self, $xpd);
  }
  
#foreach my $k (sort keys %{$xpd->{ns}}){print $k,"\t",$xpd->{ns}->{$k},"\n";}die;

  if($self->{xpath_queries_list}) {
    read_queries_list($self, $xpd);
  }
  else {
     push_error("Parameter 'xpath_queries_list' is required!");
     return;
  }
  
  # process queries
  my $orig_query_num = $xpd->{query_num};
  
  for( my $i = 0; $i < $orig_query_num; $i++) {
      my $query_struct = $xpd->{queries}->[$i];
      my $query = $query_struct->{query};
      my $id = $query_struct->{id};
          my $norm_query = "";
          $norm_query = process_query($self, $xpd, 0, $xpd->{root}, $query, $id, $id, $norm_query);
          $query_struct->{query} = $norm_query;
  }
  
  
  print STDERR Data::Dumper::Dumper($xpd->{root}) if DEBUG;
  print STDERR Data::Dumper::Dumper($xpd->{queries}) if DEBUG;
  print STDERR Data::Dumper::Dumper($xpd->{state}) if DEBUG;
  
  return if not $self->OK;
  
  process_blob($self, $xpd);
}

=pod

=head2 print_blob

This method is used to print the binary output into the given file.

Example:

  $xpc->print_blob("/path/to/output/binary/file.bin");

=cut

sub print_blob {
  my ($self, $file) = @_;
  my $out_fh;
  
  eval {
    open( $out_fh, ">", $file);
  };
  if ($@) { # catch exception
    $self->push_error("can't open file $file!");
    return;
  }
  
  binmode $out_fh;
  print $out_fh $self->{blob};
  return 1;
}
  
sub print_blob_to_stdout {
  my ($self, $use_base64) = @_;
 
  if($use_base64) {
    print MIME::Base64::encode_base64($self->{blob});

  } else {
  
    binmode STDOUT;
    print $self->{blob};
  }
}

=pod

=head2 errors

This method returns a list of the error messages. If no error occurs the list
is expected to be empty. If the only error that occured is a fatal error, this
list might also be empty and the relevant error message should be extracted
from the thrown exception.

Example usage is available in the example usage of the method C<OK>.

=cut

sub errors {
	my ($self) = @_;
	return @{$self->{errors}};
}

=pod

=head2 push_error($error_message)

This method is to be used by the authors and maintainers of this module.
Don't use this method unless you know what you're doing!

This method is used to push an error message to the errors messages stack.

=cut

sub push_error {
	my ($self,$error_message) = @_;
	push @{$self->{errors}}, $error_message;
}

=pod

=head2 reset_errors

This methor erase previously accumulated errors

=cut

sub reset_errors {
	my ($self) = shift;
	$self->{errors} = [];
}

=pod

=head2 OK

This method returns a true value if there are no error messages on the error
messages stack, and returns false if these are error messages on the error
messages stack.

Example usage:

=begin roff

	warn "Error messages from the Schema Processor:\n" .  join(",\n",$xsd->errors) 
		unless $xsp->OK;

=end roff

=cut

sub OK {
	my ($self) = @_;
	return $self->errors < 1 ;
}

1;
