package F5::XMLHash;

# TODO: this package needs to go OO

use strict;
use warnings;

###############################################################################
# INFO:
#   every element has the following LIST structure -
#      [
#        'text',
#        { '-att1' => 'att-val' } , ... , { '-attN' => 'att-val' },
#        {
#          'value1' => [
#                       'text',
#                       { '-att1' => 'att-val' } , ... , { '-attN' => 'att-val' },
#                       { 'sub_value1' => [ ... ] },
#                       < ... more sub_values ... >
#                       { 'sub_ valueN' => [ ... ] },
#                      ],
#        },
#        < ... more values ... >
#        { 'valueN' => 'val' },
#      ]
#
#   note that:
#    - 'text', '-att' and 'val' are all optional
#    - 'text' can be only one and if it exists it's always the first
#    - all '-att' if they exist follow the 'text' and are all before the 'values'
#    - all 'values' are always last in the list
#
#   the methods below maintain this order and rely on it in order to achieve better performance and faster searches
###############################################################################

####################################################
#  twig like methods for policy hash manipulation  #
####################################################
###############################################################################
sub delete_elt {
    my (%params) = @_;

    get_xpath( %params, remove => 1 );
}

#-----------------------------------------------------------
# elt_ref not required if xpath is absolute
sub get_xpath {
    my (%params)    = @_;
    my $xpath       = $params{xpath};
    my $curr_e      = $params{elt_ref};
    my $export_base = $params{base};
    my $remove = $params{remove} || 0;
    my $list_element_id = $params{list_element_id};

    return $curr_e if $xpath eq '';
    return $curr_e if $xpath eq '.';

    while ( $xpath =~ s/\/\//\//g ) { }
    $xpath =~ s/\/$//;

    my $is_base = 0;
    if ( $xpath =~ s/^\/// ) {
        $curr_e  = $export_base;
        $is_base = 1;
    }

    my @paths = split( '/', $xpath );
    if ( !$is_base ) {
        my @keys = keys %$curr_e;
        unshift( @paths, $keys[0] );
    }

    while ( my $curr_p = shift(@paths) ) {
        my $next_p = $paths[0];
        last unless $next_p;
        my $is_before_last = ( defined $paths[1] ) ? 0 : 1;
        my $index     = 0;
        my $found     = 0;
        for my $next_e ( @{ $curr_e->{$curr_p} } ) {
            if ( _is_value($next_e) && exists $next_e->{$next_p} ) {
                if ( $remove && $is_before_last ) {
                    splice( @{ $curr_e->{$curr_p} }, $index, 1 );
                    return;
                }
                if ( $is_before_last && defined $list_element_id && $list_element_id ) {
                    $list_element_id--;
                    next;
                }
                $curr_e = $next_e;
                $found  = 1;
                last;
            }
            $index++;
        }
        if ( !$found ) {
            return undef;
        }
    }

    return $curr_e;
}

#-----------------------------------------------------------
sub set_att {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};
    my $att     = $params{att_name};
    my $val     = $params{att_val};

    my ( $exist, $same ) = _has_att(
        elt_ref  => $elt_ref,
        att_name => $att,
        att_val  => $val,
    );

    return if $exist && $same;

    if ($exist) {
        del_att(
            elt_ref  => $elt_ref,
            att_name => $att,
        );
    }

    my $elt = _get_element($elt_ref);

    if ( _has_text( elt_ref => $elt_ref ) ) {
        splice @{$elt}, 1, 0, { '-' . $att => $val };
    }
    else {
        unshift( @{$elt}, { '-' . $att => $val } );
    }
}

#-----------------------------------------------------------
sub set_text {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};
    my $txt     = $params{text};

    _del_text( elt_ref => $elt_ref );

    my $elt = _get_element($elt_ref);

    unshift( @{$elt}, $txt );
}

#-----------------------------------------------------------
sub set_tag {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};
    my $new_tag = $params{new_tag};

    my @keys    = keys %$elt_ref;
    my $old_tag = $keys[0];

    $elt_ref->{$new_tag} = delete $elt_ref->{$old_tag};
}

#-----------------------------------------------------------
sub del_att {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};
    my $att     = $params{att_name};

    my ( $exist, $same ) = _has_att(
        elt_ref  => $elt_ref,
        att_name => $att,
    );

    return unless $exist;

    my $elt = _get_element($elt_ref);

    my $index = 0;
    for my $sub_elt ( @{$elt} ) {
        my ( $is_att, $name ) = _is_attribute($sub_elt);

        if ( $is_att && $name eq "-$att" ) {
            last;
        }

        $index++;
    }

    splice( @{$elt}, $index, 1 );
}

#-----------------------------------------------------------
sub new_elt {
    my (%params) = @_;

    my $elt = $params{elt_name};

    return { $elt => [] };
}

#-----------------------------------------------------------
sub paste {
    my (%params) = @_;

    my $parent = $params{parent};
    my $child  = $params{child};

    my $elements = _get_element($parent);

    push( @{$elements}, $child );
}

#-----------------------------------------------------------
sub first_child {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};

    my $skip_text = 1;
    for my $child ( $elt_ref->{sets} ) {

        # this is tested only on the first iteration
        next if $skip_text && _is_text($child);
        $skip_text = 0;

        next if _is_attribute($child);
        return $child;
    }
}

#-----------------------------------------------------------
# children are all but attributes
sub purge_children {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};

    my $elements = _get_element($elt_ref);

    my @children = ();
    for my $index ( 0 .. $#{$elements} ) {
        if ( !_is_attribute( $elements->[$index] ) ) {
            push @children, $index;
        }
    }

    for my $index ( reverse @children ) {
        splice @{$elements}, $index, 1;
    }
}

#-----------------------------------------------------------
###############################################################################

####################################################
#  helper methods for the twig like methods above  #
####################################################
###############################################################################
sub _has_att {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};
    my $att     = $params{att_name};
    my $val     = $params{att_val} || '';

    my $elt = _get_element($elt_ref);

    my $has       = 0;
    my $same      = 0;
    my $skip_text = 1;
    for my $sub_elt ( @{$elt} ) {

        # this is tested only on the first iteration
        next if $skip_text && _is_text($sub_elt);
        $skip_text = 0;

        my ( $is_att, $name ) = _is_attribute($sub_elt);

        # give up on the first non-attribute
        if ( !$is_att ) {
            last;
        }
        elsif ( $name eq "-$att" ) {
            $has = 1;
            if ( $sub_elt->{$name} eq $val ) {
                $same = 1;
            }
            last;
        }
    }

    return (wantarray) ? ( $has, $same ) : $has;
}

#-----------------------------------------------------------
sub _del_text {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};
    return unless _has_text( elt_ref => $elt_ref );

    my $elt = _get_element($elt_ref);
    shift( @{$elt} );
}

#-----------------------------------------------------------
sub _has_text {
    my (%params) = @_;

    my $elt_ref = $params{elt_ref};
    my $elt     = _get_element($elt_ref);

    my $has = 0;
    if ( scalar @{$elt} && _is_text( $elt->[0] ) ) {
        $has = 1;
    }

    return $has;
}

#-----------------------------------------------------------
sub _get_element {
    my $elt_ref = shift;

    my @values = values %$elt_ref;

    return $values[0];
}

#-----------------------------------------------------------
sub _is_value {
    my $elt = shift;

    if ( _is_text($elt) || _is_attribute($elt) ) {
        return 0;
    }

    return 1;
}

#-----------------------------------------------------------
sub _is_text {
    my $elt = shift;

    if ( ref($elt) ) {
        return 0;
    }
    else {
        return 1;
    }

}

#-----------------------------------------------------------
sub _is_attribute {
    my $elt = shift;

    if ( ref($elt) ) {
        my @keys = keys %$elt;
        my $rv = ( $keys[0] =~ /^-/ ) ? 1 : 0;
        return (wantarray) ? ( $rv, $keys[0] ) : $rv;
    }
    else {
        # this is just the 'text'
        return (wantarray) ? ( 0, '' ) : 0;
    }
}
###############################################################################

1;
