package F5::YAML::Schema::Utils;

use strict;
use warnings;

use File::Slurp;
use YAML::XS;
use JSON;
use F5::Deploy;
use F5::ASMConfig::Entity::Base;

my $entity_references = {};
my $entity_kind_map = {};
my $reference_name_map = {};
my %entity_to_ref_map;

my $YAML_TO_JSON_APP = F5::Deploy::map_path('/usr/share/ts/bin/yaml2json');

#----------------------------------------------------------------------

sub parse_json {
    my(%params) = @_;

    my $rv;

    return if !($params{content} || $params{filename});

    my $content = $params{content} || read_file($params{filename});
    eval {
        $rv->{data} = decode_json($content);
        $rv->{size} = length($content);
        $rv->{completed_successfully} = 1;
    };
    if (my $err = $@) {
        $err =~ s/ at \S+ line \d+.*$//s;
        $rv->{completed_successfully} = 0;
        $rv->{error_message} = $err;
    }

    return $rv;
}

#----------------------------------------------------------------------

sub parse_yaml {
    my(%params) = @_;

    my $rv;

    return if !($params{content} || $params{filename});

    my $content = $params{content} || read_file($params{filename});
    eval {
        $rv->{data} = Load($content);
        $rv->{completed_successfully} = 1;
    };
    if (my $err = $@) {
        $rv->{completed_successfully} = 0;
        $rv->{error_message} = $err;
    }

    return $rv;
}

#----------------------------------------------------------------------

sub get_entity_structs {
    my(%params) = @_;

    my $entity_type = $params{entity_type};

    if (!$entity_references->{$entity_type}) {
        if (!%entity_to_ref_map) {
            my $policy_refs_map = $F5::ASMConfig::Handler::Rest::path_mapper->{references}->{policies}->{references};
            %entity_to_ref_map = map { $policy_refs_map->{"$_"}->{entity_type} => "$_"} grep { $policy_refs_map->{$_}->{entity_type} } keys %$policy_refs_map;
        }

        my $entity_field_map = $entity_type->rest_field_map();
        my $fields = $entity_type->rest_fields();

        my @rest_external_structs = @{$entity_type->rest_external_struct_fields};
        map { $_ =~ s/__/\//g } @rest_external_structs;
        my $entity_external_structs = [ @rest_external_structs,
                                        keys %{$entity_type->all_external_structs} ];

        for my $external_struct (@$entity_external_structs) {
            my ($field, $sub_field) = split('/', $external_struct, 2);
            next if $fields->{$field}->{exclude_from_declarative};

            my $field_def = $sub_field ? $fields->{$field}->{fields}->{$sub_field} : $fields->{$field};

            next if $field_def->{rest_value_func} && $field_def->{rest_value_func} eq 'map_rest_external_field';

            my $external_struct_entity = $field_def->{rest_entity_type} || $field_def->{entity_type};
            my $unique_record_parts = $external_struct_entity->declarative_unique_record_parts;
            my $unique_record_parts_count = scalar @$unique_record_parts;
            my $unique_field;
            if ( ( !$entity_to_ref_map{$external_struct_entity} || $field_def->{declarative_flat_reference} ) && $unique_record_parts_count == 1) {
                $unique_field = $unique_record_parts->[0];
            }
            elsif ($external_struct_entity->id_field && grep { $_ eq $external_struct_entity->id_field } @$unique_record_parts) {
                $unique_field = $external_struct_entity->id_field;
            }

            my $rest_unique_field_name = $unique_field ? $external_struct_entity->rest_field_map->{$unique_field} : undef;
            my $rest_unique_field_expand_name = $unique_field ? $external_struct_entity->fields->{$unique_field}->{rest_expand_name} : undef;
            my $rest_sub_field_name = $sub_field ? $field_def->{rest_field_name} : undef;
            my $rest_sub_field_expand_name = $sub_field ? $field_def->{rest_expand_name} : undef;

            $entity_references->{$entity_type}->{$external_struct} = { field => $field,
                                                                       rest_field_name => $entity_field_map->{$field},
                                                                       sub_field => $sub_field,
                                                                       rest_sub_field_name => $rest_sub_field_name,
                                                                       rest_unique_field_name => $rest_unique_field_name,
                                                                       rest_unique_field_expand_name => $rest_unique_field_expand_name,
                                                                       rest_group => $fields->{$field}->{rest_group},
                                                                       policy_reference => $entity_to_ref_map{$external_struct_entity},
                                                                       nested_struct => $fields->{$field}->{nested_struct},
                                                                       map_rest_nested_as_single => $fields->{$field}->{map_rest_nested_as_single},
                                                                       rest_expand_name => $fields->{$field}->{rest_expand_name},
                                                                       rest_sub_field_expand_name => $rest_sub_field_expand_name,
                                                                       skip_declarative_fixup => $field_def->{skip_declarative_fixup},
                                                                       declarative_fixup => $field_def->{declarative_fixup},
                                                                     };
        }
    }

    return $entity_references->{$entity_type};
}

#----------------------------------------------------------------------

sub populate_entity_kind_map {
     my(%params) = @_;
     my $refs_map = $params{refs_map};

     for my $reference (values %$refs_map) {
        if ($reference->{entity_type}) {
            $entity_kind_map->{$reference->{kind_state}} = $reference->{entity_type};
            $entity_kind_map->{$reference->{element_rest_name}} = $reference->{entity_type};
        }
        if ($reference->{references}) {
            populate_entity_kind_map(refs_map => $reference->{references});
        }
    }
}

#----------------------------------------------------------------------

sub get_entity_type {
    my(%params) = @_;

    my $kind = $params{entity_kind};

    if (!%$entity_kind_map) {
        populate_entity_kind_map(refs_map => $F5::ASMConfig::Handler::Rest::path_mapper->{references}->{policies}->{references});
    }

    if ( $entity_kind_map->{$kind} ) {
        return $entity_kind_map->{$kind};
    }
    else{
        die "Entity Type/Kind '$kind' has no matching entity type\n";
    }
}

#----------------------------------------------------------------------

sub populate_reference_name_map {
    my (%params) = @_;
    my $refs_map = $params{refs_map};
    for my $reference (sort keys %$refs_map) {
        if ($refs_map->{$reference}->{reference_name}) {
            $reference_name_map->{$refs_map->{$reference}->{reference_name}} = $reference;
        }
        if ($refs_map->{$reference}->{references}) {
            populate_reference_name_map(refs_map => $refs_map->{$reference}->{references});
        }
    }
}

#----------------------------------------------------------------------

sub get_reference_name {
    my(%params) = @_;

    my $reference_name = $params{reference_name};

    if (!%$reference_name_map) {
        populate_reference_name_map(refs_map => $F5::ASMConfig::Handler::Rest::path_mapper->{references}->{policies}->{references});
    }

    if ( $reference_name_map->{$reference_name} ) {
        return $reference_name_map->{$reference_name};
    }
    else {
        return;
    }
}

#----------------------------------------------------------------------

sub item_fixup {
    my(%params) = @_;

    my $item = $params{item};
    my $entity_type = $params{entity_type};
    my $direction = $params{direction};

    delete $item->{selfLink};

    my $fields = $params{fields} || $entity_type->example_fields;
    my $rest_field_map = $entity_type ? $entity_type->rest_field_map : undef;

    my $groups = {};

    if ( !$params{group} && $entity_type ) {
        item_references_fixup(item => $item,
                              entity_type => $entity_type,
                              direction => $direction, );
    }

    for my $field (keys %$fields) {
        my $field_def = $fields->{$field};
        my $rest_field = $rest_field_map ? $rest_field_map->{$field} : $field_def->{rest_field_name};

        if ($field_def->{rest_group} && !$params{group}) {
            $groups->{$field_def->{rest_group}}->{$field} = $field_def;
            next;
        }

        next if !defined $item || !defined $rest_field;
        if (!defined $item->{$rest_field}) {
            if (defined $field_def->{rest_expand_name} && defined $item->{$field_def->{rest_expand_name}}) {
                if ($direction eq 'to_declarative') {
                    $item->{$rest_field} = $item->{$field_def->{rest_expand_name}};
                    delete $item->{$field_def->{rest_expand_name}};
                }
                else {
                    $rest_field = $field_def->{rest_expand_name};
                }
            }
            else {
                next;
            }
        }

        my $field_value = $field_def->{map_rest_nested_as_single} ? [ $item->{$rest_field} ] : $item->{$rest_field};
        my $nested_struct = $field_def->{nested_struct}
                                && (!$field_def->{rest_value_func} || $field_def->{rest_value_func} ne 'map_rest_defense_attributes');
        my $external_struct = ($field_def->{external_struct} || $field_def->{diff_struct})
                                  && (!$field_def->{rest_value_func} || $field_def->{rest_value_func} ne 'map_rest_external_field');
        if ($entity_type) {
            my $delete_field = !exists $entity_type->declarative_fields->{$field};
            if (!$delete_field) {
                if ($nested_struct) {
                    $delete_field = !@$field_value;
                }
                elsif ($external_struct) {
                    $delete_field = !%$field_value;
                }
                else {
                    $delete_field = !defined $field_value;
                }
            }

            if ($delete_field) {
                delete $item->{$rest_field};
                next;
            }
        }

        my $nested_entity_type = $field_def->{rest_entity_type} || $field_def->{entity_type};
        if ($nested_entity_type && $nested_entity_type !~ /::Entity::/) {
            $nested_entity_type = undef;
        }
        if ($nested_struct) {
            my $fields;
            if (!$nested_entity_type) {
                $fields = $field_def->{fields};
            }
            for my $struct (@$field_value) {
                last if !(ref($struct) eq 'HASH');
                item_fixup(item => $struct,
                           entity_type => $nested_entity_type,
                           fields => $fields,
                           direction => $direction,
                           sub_struct => 1, );
            }
        }
        elsif ($nested_entity_type
                 && (!$field_def->{rest_value_func} || $field_def->{rest_value_func} ne 'map_rest_external_field') ) {
            item_fixup(item => $field_value,
                       entity_type => $nested_entity_type,
                       direction => $direction,
                       sub_struct => 1, );
        }
    }

    for my $group (keys %$groups) {
        item_fixup(item => $item->{$group},
                   entity_type => $entity_type,
                   fields => $groups->{$group},
                   direction => $direction,
                   group => 1, );
    }
}

#----------------------------------------------------------------------

sub item_references_fixup {
    my(%params) = @_;

    my $item = $params{item};
    my $entity_type = $params{entity_type};
    my $direction = $params{direction};

    if ($direction eq 'to_declarative') {
        delete $item->{link};
    }

    my @policy_references;

    my $entity_external_structs = get_entity_structs(entity_type => $entity_type);
    for my $external_struct (values %$entity_external_structs) {
        my $rest_field_name = $external_struct->{rest_field_name};
        if ((my $expand_name = $external_struct->{rest_expand_name}) && $direction eq 'to_declarative' && $external_struct->{nested_struct}) {
            if (my $rest_group = $external_struct->{rest_group}) {
                if (defined $item->{$rest_group}->{$expand_name}) {
                    $item->{$rest_group}->{$rest_field_name} = $item->{$rest_group}->{$expand_name};
                    delete $item->{$rest_group}->{$expand_name};
                }
            }
            elsif (defined $item->{$expand_name}) {
                $item->{$rest_field_name} = $item->{$expand_name};
                delete $item->{$expand_name};
            }
        }

        my $values;
        if (my $rest_group = $external_struct->{rest_group}) {
            $values = ($external_struct->{nested_struct} && !$external_struct->{map_rest_nested_as_single}) ? $item->{$rest_group}->{$rest_field_name} : [ $item->{$rest_group} ];
        }
        else {
            $values = ($external_struct->{nested_struct} && !$external_struct->{map_rest_nested_as_single} && defined $rest_field_name) ? $item->{$rest_field_name} : [ $item ];
        }

        for my $value (@$values) {
            if ($direction eq 'from_declarative') {
                my $rv = from_declarative(value => $value,
                                          field_def => $external_struct,
                                          entity_type => $entity_type);
                if ($rv && $rv->{policy_reference}) {
                    push (@policy_references, $rv->{policy_reference});
                }
            }
            else {
                to_declarative(value => $value,
                               field_def => $external_struct);
            }
        }
    }

    return { policy_references => \@policy_references };
}

#----------------------------------------------------------------------

sub to_declarative {
    my(%params) = @_;

    my $value = $params{value};
    my $field_def = $params{field_def};

    my $unique_field = $field_def->{rest_unique_field_name};
    my $rest_unique_field_expand_name = $field_def->{rest_unique_field_expand_name};
    my $rest_expand_name = $field_def->{rest_expand_name};
    my $rest_sub_field_expand_name = $field_def->{rest_sub_field_expand_name};
    if ($rest_expand_name && defined $value->{$rest_expand_name}) {
        my $rest_field_name = $field_def->{rest_field_name};
        if ($unique_field) {
            if (defined $value->{$rest_expand_name}->{$unique_field}) {
                $value->{$unique_field} = $value->{$rest_expand_name}->{$unique_field};
            }
            delete $value->{$rest_expand_name};
        }
        else {
            $value->{$rest_field_name} = $value->{$rest_expand_name};
            delete $value->{$rest_expand_name};
            delete $value->{$rest_field_name}->{link};
        }
    }
    elsif ($rest_sub_field_expand_name && defined $value->{$rest_sub_field_expand_name}) {
        my $rest_sub_field_name = $field_def->{rest_sub_field_name};
        if ($unique_field) {
            $value->{$unique_field} = $rest_unique_field_expand_name ? $value->{$rest_sub_field_expand_name}->{$rest_unique_field_expand_name}
                                                                     : $value->{$rest_sub_field_expand_name}->{$unique_field};
            delete $value->{$unique_field}->{link} if $rest_unique_field_expand_name;
            if (exists $value->{$rest_sub_field_expand_name}->{signatureId}) {
               $value->{name} = $value->{$rest_sub_field_expand_name}->{name};
               if ($value->{$rest_sub_field_expand_name}->{isUserDefined}) {
                  $value->{tag} = $value->{$rest_sub_field_expand_name}->{tag};
                  delete $value->{$unique_field};
               }
            }
            delete $value->{$rest_sub_field_expand_name};
        }
        else {
            $value->{$rest_sub_field_name} = $value->{$rest_sub_field_expand_name};
            delete $value->{$rest_sub_field_expand_name};
            delete $value->{$rest_sub_field_name}->{link};
            if( !keys %{$value->{$rest_sub_field_name}}) { #if the resulted hash is empty
                delete $value->{$rest_sub_field_name};
            }
        }
    }
    else {
        delete $value->{link};
    }
}

#----------------------------------------------------------------------

sub from_declarative {
    my(%params) = @_;

    my $value = $params{value};
    my $field_def = $params{field_def};
    my $entity_type = $params{entity_type};

    return if $field_def->{skip_declarative_fixup};

    if($field_def->{rest_field_name} eq 'signatureOverrides' ||
       $field_def->{rest_field_name} eq 'signature'){
        if ( $value && !defined $value->{signatureId} && defined $value->{name} ) {

             $value->{signature}->{name} = $value->{name};
             $value->{signature}->{tag} = $value->{tag} || '';

             delete $value->{name};
             delete $value->{tag};
        }
    }

    if ($field_def->{declarative_fixup}) {
        return $entity_type->from_declarative(object => $value,
                                              field_name => $field_def->{field},
                                              field_def => $field_def,);
    }

    my $unique_field = $field_def->{rest_unique_field_name};
    my $rest_field_name = $field_def->{rest_field_name};

    if (my $rest_sub_field_name = $field_def->{rest_sub_field_name}) {
        if ($unique_field) {
            return if !defined $value->{$unique_field};
            $value->{$rest_sub_field_name}->{$unique_field} = $value->{$unique_field};
            if (!($entity_type->reverse_rest_field_map->{$unique_field})) {
                delete $value->{$unique_field};
            }
        }
        else {
            return if !defined $value->{$rest_sub_field_name};
        }
    }
    elsif ($unique_field) {
        return if !defined $value->{$unique_field};
        $value->{$rest_field_name}->{$unique_field} = $value->{$unique_field};
        if (!($entity_type->reverse_rest_field_map->{$unique_field})) {
            delete $value->{$unique_field};
        }
    }
    elsif (!$value->{$rest_field_name}) {
        return;
    }

    return { policy_reference => $field_def->{policy_reference} };
}

#----------------------------------------------------------------------

sub yaml_to_json {
    my(%params) = @_;

    my $encoding = $params{encoding};
    my $filename = $params{filename};

    my $python_exe = F5::SystemUtils::get_python_exe(required_modules => [ 'yaml' ]);
    open(my $contFH, '-|', "$python_exe $YAML_TO_JSON_APP --encoding $encoding $filename 2>&1");

    my $json_cont = '';
    my $line;
    while (defined($line = <$contFH>)) {
        $json_cont .= $line;
    }
    close $contFH;

    my $rv = {};
    my $exitcode = $? >> 8;
    if ($exitcode) {
        $rv->{completed_successfully} = 0;
        $rv->{error_message} = $json_cont;
    }
    else {
        $rv->{completed_successfully} = 1;
        $rv->{data} = $json_cont;
    }

    return $rv;
}

#----------------------------------------------------------------------

our %declarative_schema_excludes = (
                                    %$F5::ASMConfig::Entity::Base::common_diff_excludes,
                                    rest_uuid => 1,
                                    flg_in_classification => 1,
                                    is_default => 1,
                                    is_referenced => 1,
                                    wildcard_sequence => 0,
                                   );

sub is_supported_field {
    my(%params) = @_;
    my $field = $params{field};
    my $field_def = $params{field_def};

    my $rv = 1;
    if ($declarative_schema_excludes{$field}) {
        $rv = 0;
    }
    elsif ($field_def->{extra_field} || $field_def->{exclude_from_declarative}
         || $field_def->{exclude_from_rest} || $field_def->{exclude_from_rest_example}) {
        $rv = 0;
    }
    elsif (my $nested_entity_type = $field_def->{rest_entity_type} || $field_def->{entity_type}) {
        if ($nested_entity_type =~ /::Entity::/) {
            $rv = is_supported_entity(entity_type => $nested_entity_type);
        }
    }

    my $rest_deprecated_version = $field_def->{rest_deprecated_version};
    if ($rv && $rest_deprecated_version) {
        my $ver = $params{ver} || F5::BigipVersionUtils::bigip_version();
        $rv = (F5::BigipVersionUtils::compare_versions(version1 => $rest_deprecated_version, version2 => $ver) > 0);
    }

    return $rv;
}

#----------------------------------------------------------------------

sub is_supported_entity {
    my(%params) = @_;
    my $entity_type = $params{entity_type};

    my $entity_config = $entity_type->entity_config;
    my $rest_config = $entity_config->{rest_paths}->[0]->{rest_path_properties};
    my $rv = 1;

    if ($rv && $entity_config->{declarative_handler_func}) {
        $rv = $entity_config->{declarative_handler_func} ne 'handle_excluded_entity';
    }

    my $rest_deprecated_version = $rest_config->{rest_deprecated_version};
    if ($rv && $rest_deprecated_version) {
        my $ver = $params{ver} || F5::BigipVersionUtils::bigip_version();
        $rv = (F5::BigipVersionUtils::compare_versions(version1 => $rest_deprecated_version, version2 => $ver) > 0);
    }

    return $rv;
}

#----------------------------------------------------------------------

sub is_read_only_entity {
    my(%params) = @_;
    my $entity_type = $params{entity_type};

    my $entity_config = $entity_type->entity_config;
    my $rest_config = $entity_config->{rest_paths}->[0]->{rest_path_properties};
    my $rv = 0;

    my $resource_methods = $rest_config->{allowed_methods};
    my $collection_methods = $rest_config->{collection_allowed_methods} || $resource_methods;
    if ( $collection_methods && !($collection_methods->{PATCH} || $collection_methods->{POST}) ) {
        $rv = !($rest_config->{singleton} || $resource_methods->{PATCH});
    }

    return $rv;
}

#----------------------------------------------------------------------

1;
