package F5::iControl;

use strict;
use warnings;

use F5::Logger;
use F5::ASMConfig::ASMConfig;

#use SOAP::Lite + trace => qw(method debug);
use SOAP::Lite;

use constant DEFAULT_PROXY_URI => 'iControl/iControlPortal.cgi';
use constant DEFAULT_HOST => '127.0.0.1';
use constant DEFAULT_HTTP_PORT => '80';
use constant DEFAULT_HTTPS_PORT => '443';
use constant DEFAULT_PROTOCOL => 'http';
use constant ERROR_PREFIX => "iControl Error:\n";
use constant TRUE => 1;
use constant FALSE => 0;
use constant MAX_EXTENSION => 100;
my %ICONTROL_TYPECAST_TYPES = (
	'{urn:iControl}Common.EnabledState' => 1,
);

=pod 

=head1 NAME

iControl - a Perl module for iControl functionality. It provides methods that
use iControl via SOAP::Lite to query bigip and set bigip db variables.

=head1 SYNOPSIS

	use F5::iControl;

	my $iControl = iControl->new;

        eval {
          my @bigip_users = $iControl->get_bigip_users;

          my $bigip_user_role = $iControl->get_bigip_user_role($user_name);

          my @bigip_httpClassifiers = $iControl->get_bigip_httpClassifiers;
        };
        if (my $iControl_error = $@) {
           my $display_message =  $iControl_error->{display_message};
           my $error_message =  $iControl_error->{error_message};
           my $error_level   =  $iControl_error->{error_level};
           my $error_code    =  $iControl_error->{error_code};
          ...
        }


=head1 DESCRIPTION

iControl provides an Object Oriented Perl interface to common
functionalities of ASM that are dependent upon iControl.
Note that this module DOES NOT necessarily implement any of the interfaces
provided by iControl.

=head1 METHODS

=head2 new

The constructor is used for generating a new object,
initializing its attributes.

=cut

sub new {
	my ($class, %arguments) = @_;
	my $error_message = '';
	
        my $logger = $arguments{logger} || F5::Logger->get_logger(logger_name => 'iControl');
	# Set proxy from parameters
	my $protocol = $arguments{protocol} || DEFAULT_PROTOCOL;
	my $authentication = '';
	if ($arguments{username} && $arguments{password}) {
	  $authentication .= "$arguments{username}:$arguments{password}\@";
	}
	my $host .= $arguments{host} || DEFAULT_HOST;
	my $port .= $arguments{port} || ($arguments{protocol} && $arguments{protocol} eq 'https' ? 
					 DEFAULT_HTTPS_PORT : DEFAULT_HTTP_PORT);
	my $proxy_uri .= $arguments{proxy_uri} || DEFAULT_PROXY_URI;
	my $proxy = "$protocol://$authentication$host:$port/$proxy_uri";
	
	bless {
		_proxy    => $proxy,
                logger    => $logger,
	} , $class;
}

sub check_error {
    my ($self, %params) = @_;
    my $display_message = $params{display_message};
    my $internal_error = $params{internal_error};
    if (defined($params{fault_obj})) {
        return if !$params{fault_obj}->fault;  #no error
        $display_message = $internal_error = $params{fault_obj}->faultcode . ": " . $params{fault_obj}->faultstring;
        $display_message =~ s/.*error_string[^A-Za-z]*//xms;       
    }
    my $error_message = ERROR_PREFIX.$internal_error;
    $error_message =~ s/\n/ /g;
    $self->{logger}->error($error_message);
    die { error_message  => $error_message,
          display_message => $display_message,
          error_level    => "info",
          error_code     => F5::ASMConfig::ErrorCode::ICONTROL_ERROR(),
        };
}

#Implement Typecast for iControl enumeration Elements
no warnings 'redefine';
sub SOAP::Deserializer::typecast {
	my ($self, $value, $name, $attrs, $children, $type) = @_;
	return unless ( $type && 
                        ($ICONTROL_TYPECAST_TYPES{$type} ||
                         $type =~ /^\Q{urn:com.f5.license.v5b}\E/));

	return $value;
}

use warnings;


1;
