# bd-cfg Table Definitions

bd-cfg tables (aka DCC tables) are each defined in a distinct YAML file under `tables/`. These definitions include:

- Table attributes, such as table enum value (`dcc_table_id`)
- DCC database table name and fields names
- C struct names and types
- Protobuf element name and index (for the entire table)
- Protobuf field names and types

## Workflow

**IMPORTANT**: Before adding a new table definition here, new table enums must first be added to [the WAF enum list](https://gitswarm.f5net.com/waf/waf-doc/-/tree/master/dev/enums).

## Table attributes

Example table defintion:

```yaml
name: account_custom_response_headers
c_struct_name_prefix: policy_custom_response_headers
dcc_table_id: 230
enum_value: 43
proto_index: 46
```

| Field name                | Description                                                                                                                                                  | Default                                     |
|---------------------------|--------------------------------------------------------------------------------------------------------------------------------------------------------------|---------------------------------------------|
| `name`                    | Base name for table. This should be defined according to the BD enum name, such as `CONFIG_TYPE_ACCOUNTS` (taken from [`common/include/common_config.h`][1]) |                                             |
| `dcc_table_id`            | BD enum number for this table (corresponding to numeric value in [`common/include/common_config.h`][1].                                                      |                                             |
| `enum_value`              | Protobuf enum values within enum BdTable. Beginning with `dcc_table_id` 252, `enum_value` will always be the same value as `dcc_table_id`.                   |                                             |
| `proto_index`             | Protobuf field index within message BdConfigElement. Beginning with `dcc_table_id` 252, `proto_index` will always be the same value as `dcc_table_id`.       |                                             |
| `table`                   | DCC database table name                                                                                                                                      | Taken from `name`                           |
| `output_file`             | Output file from `F5::PrepareFiles`                                                                                                                          | Taken from `name`                           |
| `c_struct_name_prefix`    | Prefix for `_key_t` and `_data_t` C structures in BD                                                                                                         | Taken from `name`                           |
| `c_table_class`           | C class used for internal BD table persistence (see below)                                                                                                   |                                             |
| `auto_generate_c_struct`  | Boolean, used to disable C struct generation in rare circumstances                                                                                           | true                                        |
| `split_adds_and_deletes`  | Boolean, used to output adds and deletes in separate delta files in rare circumstances                                                                       | false                                       |
| `split_config_by_account` | Boolean, used to disable splitting files by `policy_name_crc` in rare circumstances                                                                          | true if `policy_name_crc` exists as a field |
| `order_by_fields`         | Only used for 1 legacy table. Don't use this.                                                                                                                | Derived from `is_primary_key` fields        |
| `legacy_query`            | Only used for 2 legacy tables. Don't use this.                                                                                                               |                                             |

[1]: https://gitswarm.f5net.com/waf/waf-general/-/blob/master/common/include/common_config.h

Automatically derived fields (such as `output_file`) should only be specified if they differ from the derived value.

### HashMapMultiValueTable tables

HashMapMultiValueTable tables are indicated as the `c_table_class` used for BD persistence.

HashMapMultiValueTable tables may receive multiple values for a particular key value (i.e. `c_location`="key"). Each key acts as an accumulator for multiple values which can be iterated over, or even retrieved randomly. 

However, the unique identifier for adding/removing these rows is the entire record (both "key" and "data" fields). Thus, these table should not use `is_primary_key` to mark key fields. Alternatively, all fields **must** be marked as `is_primary_key` (this approach is error prone, as new fields may not be marked).

## Field attributes

Example field defintion:

```yaml
message_fields:
- name: custom_response_id
  c_location: key
  is_primary_key: true
  proto_index: 1
  proto_type: uint32
```

| Field name               | Description                                                                                       | Default                   |
|--------------------------|---------------------------------------------------------------------------------------------------|---------------------------|
| `name`                   | Field name                                                                                        |                           |
| `is_primary_key`         | Boolean, whether this field is part of the record's unique key                                    | false                     |
| `c_location`             | C structures field is included in. Possible values: "key", "data", "both"                         |                           |
| `c_type`                 | C field type. May be BD `.h` file defined type.                                                   |                           |
| `c_fill_macro`           | C function used for copying protobuf value to C struct                                            | Derived from `proto_type` |
| `proto_index`            | Protobuf field index. Should be in ascending numerical order, and should be considered immutable. |                           |
| `proto_type`             | Protobuf field type. May be "skip" if field is not included within the protobuf.                  |                           |
| `read_from_dcc_database` | Field is present in the  DCC database.                                                            | true                      |

**Note:** Field order is significant for binary tables.

## Table order file (`tables.yaml`)

`tables.yaml` determines the order of tables to be loaded, since for some tables, load order is significant. 

However, for most tables, load order is not significant. Thus, most new tables should **not** be added to `tables.yaml`. Any table not listed in `tables.yaml` will be loaded afterwards, sorted by `dcc_table_id` order.

## Deprecated sources

DCC tables were formerly defined in:

- [`mng/db/dcc_tables_cfg.yaml`](https://gitswarm.f5net.com/waf/waf-general/-/blob/master/mng/db/dcc_tables_cfg.yaml)
- [`common/asm_proto_msg/bd_config_auto/bd_config_auto.json`](https://gitswarm.f5net.com/waf/waf-general/-/blob/master/common/asm_proto_msg/bd_config_auto/bd_config_auto.json)

The files are no longer used in the build or product runtime, and should not be modified.
