package F5::ASMConfig::Entity::Base;

use strict;
use warnings;

use base qw/F5::ASMConfig::Validations/;
use File::Basename;
use YAML::XS;
use F5::XML::Twig;
use MIME::Base64;
use Scalar::Util;
use F5::ModuleLoader;

use F5::Deploy;
use F5::BigipVersionUtils;
use F5::BigipUtils;
use F5::DateTimeUtils;
use F5::ASM::COAPI;
use F5::IpUtils;
use F5::IpAddr;
use F5::DbUtils;
use F5::ObjUtils qw(freeze dclone);
use F5::Utils::Rest;
use F5::Utils::Encode;
use F5::ASMConfig::Entity::Iterator;
use F5::ASMConfig::Entity::Policy;
use F5::ASMConfig::Entity::Violation;
use F5::ASMConfig::Cache;
use F5::ASMConfig::Utils;
use F5::DateTimeUtils;
use F5::ASMConfig::Entity::JsonProfile;
use F5::ASMConfig::Entity::PlainTextProfile;
use F5::Logger;
use Digest::MD5 qw(md5_hex);
use Date::Parse;

use List::Util qw(max first);
use Scalar::Util qw(reftype);
use File::Temp qw(tempfile);
use POSIX;
use Time::Local;
use Socket qw(:DEFAULT inet_pton inet_ntop);
use JSON -convert_blessed_universally;
use Storable qw(thaw);
use Encode qw(:all);
eval {
    require Encode::HanExtra;
};
use utf8;

our $IMPORT_RESULT_ADDED = 'add';
our $IMPORT_RESULT_UPDATE = 'update';
our $IMPORT_RESULT_UNCHANGED = 'unchanged';

our $UNLIKELY_DELIMITER = '###UNLIKELY_DELIMITER###';


our %UNLEARNABLE_HTTP_CHECK_CODES = (  8192 => 1, # Bad HTTP Version
                                       16384 => 1, # Null in Request
                                       65536 => 1, # Unparsable request content
                         );

my %COMPARISON_MAPPING = (
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() => '=',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() => '=BINARY',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE() => '!=',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_NE() => '!=BINARY',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE() => 'LIKE',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE() => 'NOT LIKE',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP() => 'REGEXP BINARY',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP() => 'NOT REGEXP BINARY',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GT() => '>',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GE() => '>=',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT() => '<',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE() => '<=',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN() => 'IN',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN() => 'NOT IN',
#The ALL_NESTED comparisons are the opposite of a standard IN query.  
#They check that there is no existence of the opposite of their condition.
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_IN() => 'NOT IN',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_NOT_IN() => 'IN',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP() => 'REGEXP',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP() => 'NOT REGEXP',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NULL() => 'IS NULL',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NOT_NULL() => 'IS NOT NULL',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EXISTS() => '!=',
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS() => '=',
                         );

my %COMPARISON_MAPPING_BY_DB = ('mysql' => {  %COMPARISON_MAPPING },
                                'sqlite' => { %COMPARISON_MAPPING,
                                              F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() => '= ',
                                              F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_NE() => '!=',
                                              F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP() => 'REGEXP',
                                              F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP() => 'NOT REGEXP',
                                            },
                               );

my %COMPARISON_NOT_MAPPING = (
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_NE(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_NE() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GT() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GE() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GE(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GT(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_IN() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_NOT_IN(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_NOT_IN() =>  F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_IN(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NULL() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NOT_NULL(),
                          F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NOT_NULL() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NULL(),
                         );
 

my %DIRECTION_MAPPING = (
                         F5::ASMConfig::SortDirection::SORT_DIRECTION_ASC() => 'ASC',
                         F5::ASMConfig::SortDirection::SORT_DIRECTION_BINARY_ASC() => 'ASC',
                         F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC() => 'DESC',
                         F5::ASMConfig::SortDirection::SORT_DIRECTION_BINARY_DESC() => 'DESC',
                        );

my %BINARY_SORT_MAPPING = (
                            F5::ASMConfig::SortDirection::SORT_DIRECTION_BINARY_ASC() => 'BINARY', 
                            F5::ASMConfig::SortDirection::SORT_DIRECTION_BINARY_DESC() => 'BINARY', 
                          );

my %filter_func = (
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() => sub {
                       my(%params) = @_;
                       return (defined $params{field_value} && $params{field_value} eq $params{search_filter}->{field_value});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE() => sub {
                       my(%params) = @_;
                       return (! defined $params{field_value} || $params{field_value} ne $params{search_filter}->{field_value});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() => sub {
                       my(%params) = @_;
                       return (defined $params{field_value} && $params{field_value} eq $params{search_filter}->{field_value});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_NE() => sub {
                       my(%params) = @_;
                       return (! defined $params{field_value} || $params{field_value} ne $params{search_filter}->{field_value});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE() => sub {
                       my(%params) = @_;
                       my $regexp = $params{search_filter}->{field_value};
                       my $special_regex_characters = '.^$*+?()[]{}|';
                       $regexp =~ s/([\Q$special_regex_characters\E])/\\$1/g;
                       $regexp =~ s/(?:^%|(?<!\\)((\\\\)*)%)/$1.*/g; #Translate  non-escaped SQL wildcards to regex wildcard
                       $regexp =~ s/(?:^_|(?<!\\)((\\\\)*)_)/$1./g;
                       $regexp = '^' . $regexp . '$';
                       return (defined $params{field_value} && $params{field_value} =~ /$regexp/s);
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE() => sub {
                       my(%params) = @_;
                       my $regexp = $params{search_filter}->{field_value};
                       my $special_regex_characters = '.^$*+?()[]{}|';
                       $regexp =~ s/([\Q$special_regex_characters\E])/\\$1/g;
                       $regexp =~ s/(?:^%|(?<!\\)((\\\\)*)%)/$1.*/g; #Translate non-escaped SQL wildcards to regex wildcards
                       $regexp =~ s/(?:^_|(?<!\\)((\\\\)*)_)/$1./g;
                       $regexp = '^' . $regexp . '$';
                       return (! defined $params{field_value} || $params{field_value} !~ /$regexp/s);
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GT() => sub {
                       my(%params) = @_;
                       return (defined $params{field_value} && $params{field_value} gt $params{search_filter}->{field_value});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GE() => sub {
                       my(%params) = @_;
                       return (defined $params{field_value} && $params{field_value} ge $params{search_filter}->{field_value});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT() => sub {
                       my(%params) = @_;
                       return (defined $params{field_value} && $params{field_value} lt $params{search_filter}->{field_value});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE() => sub {
                       my(%params) = @_;
                       return (defined $params{field_value} && $params{field_value} le $params{search_filter}->{field_value});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN() => sub {
                       my(%params) = @_;
                       return scalar grep { defined $params{field_value} && $params{field_value} eq $_ } @{$params{search_filter}->{field_value_list}};
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN() => sub {
                       my(%params) = @_;
                       return ! (scalar grep { defined $params{field_value} && $params{field_value} eq $_ } @{$params{search_filter}->{field_value_list}});
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP() => sub {
                       my(%params) = @_;
                       return (defined $params{field_value} && $params{field_value} =~ /$params{search_filter}->{field_value}/s);
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP() => sub {
                       my(%params) = @_;
                       return (! defined $params{field_value} || $params{field_value} !~ /$params{search_filter}->{field_value}/s)
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EXISTS() => sub {
                       my(%params) = @_;
                       return (defined ($params{field_value}));
                   },
                   F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS() => sub {
                       my(%params) = @_;
                       return (! defined ($params{field_value}));
                   },
                  );


our $common_diff_excludes = { 'last_tightening_time' => 1,
                              'staged_since' => 1,
                              'last_updated' => 1,
                              'last_update'  => 1,
                              'obj_crc' => 1,
                              'user_config_level' => 1,
                              'created_by' => 1,
                              'modified_time' => 1,
                              'created_time' => 1,
                              'policy_id' => 1,
                              'wildcard_sequence' => 1,
                              'is_inherited' => 1,
                              'has_parent' => 1,
                           };

our $common_inheritance_excludes = { %$common_diff_excludes,
                                    in_staging => 1,
                                    flg_quarantined => 1,
                                    policy_type => 1,
                                    is_inherited => 0,  #Enforce validation on changing is_inherited flag
                                  };

my $common_default_table_excludes = { 'is_inherited' => 1,
                                      'has_parent' => 1,
                                    };

my %DISPLAY_STRINGS = (
                       value_was_updated => '$FIELD was set to $VALUE.',
                       value_will_be_updated => 'Set $FIELD to $VALUE.',
                       hidden_value_was_updated => '$FIELD was updated.',
                       value_list_was_updated => '$FIELD were set to $VALUE.',
                       value_list_was_appended => 'Added $VALUE to $FIELD.',
                       value_list_was_removed => 'Removed $VALUE from $FIELD.',
                       value_list_will_be_updated => 'Set $FIELD to $VALUE.',
                       value_list_will_be_appended => 'Add $VALUE to $FIELD.',
                       value_list_will_be_removed => 'Remove $VALUE from $FIELD.',
                      );

my $UTF8_ENCODE_VERSION = "12.0.0";

my $cached_defaults = {};
my $element_type_cache = {};
my $mcp_type_to_entity_map_cache = {};
my $field_display_value_db_map_cache = {};

our $POLICY_BUILDER_USER_NAME = "Policy Builder";
our $BIGIQ_USER_NAME = "BIGIQ";
our $verbose_logger = F5::Logger->get_logger(logger_name => 'asmconfigverbose');

#----------------------------------------------------------------------

sub new {
  my($class, %params) = @_;
  my $self = {};
  bless $self, $class;
  my $rv = $self->init(%params);

  if ($rv) {
    return $rv;
  }
  return;
}

#----------------------------------------------------------------------
sub init {
  my($self, %params) = @_;

  $self->{dbh} = $params{dbh};
  $self->{logger} = $params{logger};
  $self->{policy} = $params{policy};

  if ($params{get_criteria}) {
    $self->store_virtual_fields(%params, store_obj => $params{get_criteria});
  }

  if ($params{preload_default}) {
    my $all_fields = $params{all_fields} || 0;
    my $class = ref($self);
    unless ($cached_defaults->{$all_fields}->{$class}) {
      $cached_defaults->{$all_fields}->{$class} = $self->get_default_obj(%params);
    }
    my $default_obj = $cached_defaults->{$all_fields}->{$class};
    $self->update_default_object_by_pb_settings(%params,
                                                default_obj => $default_obj);

    $self->from_object(object => $default_obj);
  }

  if ($params{object}) {
    $self->from_object(object => $params{object},
                       instantiate_entities => $params{instantiate_entities},
                       full_retrieve_on_instantiate  => $params{full_retrieve_on_instantiate}, 
                      );
  }

  if ($params{xml}) {
      my $object = $self->from_xml(xml => $params{xml});
      $self->from_object(object => $object);
  }

  if ($params{rest_object}) {
    my $object = $self->from_rest(%params);
    $self->from_object(object => $object);
  }

  if ($params{mcp_object}) {
      $self->load_mcp_fields(%params);
  }

  if ($params{id}) {
    return $self->retrieve(%params,
                           get_criteria => { $self->id_field => $params{id}},
                          );
  }

  if ($params{get_criteria}) {
    return $self->retrieve(%params,
                          );
  }

  $self->{plc_dbname} = 'PLC';

  $self->{dcc_dbname} = 'DCC';
  
  return $self;
}

#----------------------------------------------------------------------

sub retrieve {
  my($self, %params) = @_;

  my $iterator = ref($self)->retrieve_all(dbh => $self->{dbh},
                                          logger => $self->{logger},
                                          policy => $self->{policy},
                                          find_single_record => 1,
                                          %params);
  my $new_self = $iterator->get_next;
  return unless $new_self;

  for my $k (keys %$new_self) {
    $self->{$k} = $new_self->{$k};
  }

  if ($params{recast} && (my $recast_field = $self->recast_field)) {
      my $new_class = $self->fields->{$recast_field}->{recast_map}->{$self->{$recast_field}};
      return $new_class->new(dbh => $self->{dbh},
                             logger => $self->{logger},
                             policy => $self->{policy},
                             %params);
  }

  return $self;
}

#----------------------------------------------------------------------

sub set_element_type_cache {
    my ($class, $set_element_type_cache) = @_;
    $element_type_cache = $set_element_type_cache;
    return $element_type_cache;
}

#----------------------------------------------------------------------

sub get_element_type_cache {
    my ($class) = @_;
    if (!%$element_type_cache) {
        $class->load_element_type_cache();
    }
    return $element_type_cache;
}

#----------------------------------------------------------------------

sub get_cached_element_type {
    my ($class, %params) = @_;
    my $element_type = $params{element_type};
    if ($element_type_cache->{$element_type}) {
        return $element_type_cache->{$element_type};
    }

    $class->load_element_type_cache();
    return $element_type_cache->{$element_type};
}


#----------------------------------------------------------------------

sub load_element_type_cache {
    my @entity_classes = F5::ModuleLoader::load_from_base_class('F5::ASMConfig::Entity::Base');
    for my $class (@entity_classes) {
        my $entity_function_name = $class;
        $entity_function_name =~ s/.*:://;
        $entity_function_name =~ s/([a-z])([A-Z])/${1}_${2}/g;
        $entity_function_name = lc $entity_function_name;

        if ($class->can('element_type') && $class->element_type) {
            $element_type_cache->{$class->element_type} =  { element_type_name => $class->element_type_name,
                                                             class => $class,
                                                             entity_function_name => $entity_function_name,
                                                           };
        }
    }
}

#----------------------------------------------------------------------

sub get_cached_entity_by_mcp_type {
    my ($class, $mcp_type) = @_;
    if ($mcp_type_to_entity_map_cache->{$mcp_type}) {
        return $mcp_type_to_entity_map_cache->{$mcp_type};
    }
    if (!%$mcp_type_to_entity_map_cache) {
        $class->load_mcp_type_to_entity_map_cache();
        return $mcp_type_to_entity_map_cache->{$mcp_type};
    }

    return;
}

#----------------------------------------------------------------------

sub load_mcp_type_to_entity_map_cache {
    my ($class) = @_;
    my $schema = $class->full_schema();
    my %cache_map = map { $_->{mcp_element_type} => $_->{entity_type} } grep { $_->{mcp_element_type} } values %{$schema};
    $mcp_type_to_entity_map_cache = \%cache_map;
}

#----------------------------------------------------------------------

sub retrieve_nested_structures {
  my($self, %params) = @_;
  my $dbh = $self->{dbh};

  my $fields = $params{fields} || $self->fields;
  my $nested_struct_fields = $params{nested_struct_fields} || $self->nested_struct_fields(%params);
  for my $field_name (@$nested_struct_fields) {
    next unless $fields->{$field_name} && $fields->{$field_name}->{db_table_name};
    next if ($fields->{$field_name}->{extra_field} && !$params{extra_fields_hash}->{$field_name});
    if (!$self->{$field_name . '__exists'}) {
          $self->{$field_name} = [];
          next;
    }

    $self->{$field_name} = $self->load_nested_struct_records(nested_def => $fields->{$field_name});
  }
  $self->after_nested_struct_load(%params);
}

#----------------------------------------------------------------------

sub after_nested_struct_load {
    return;
}

#----------------------------------------------------------------------

sub load_nested_struct_records {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $nested_def = $params{nested_def};
  my @nested_entities;
  return unless  $nested_def->{db_table_name};

  local $" = ',';

  my $nested_fields = $nested_def->{fields};
  my @nested_field_names = grep { $nested_fields->{$_}->{db_field_name} } keys %$nested_fields;
  my @serialized_field_names = grep { !$nested_fields->{$_}->{db_field_name} && $nested_fields->{$_}->{serialized_field} } keys %$nested_fields;
  my @nested_db_fields = map { $nested_fields->{$_}->{db_field_name} } @nested_field_names;

  my $foreign_id = $self->get_foreign_id(nested_def => $nested_def,
                                         entity => $self);
  my $table_name = $nested_def->{db_table_name};

my $sql = <<EndQuery;
SELECT @nested_db_fields
  FROM $table_name
 WHERE $foreign_id->{field_name} = ?
EndQuery

  #If there is only one column, and is an external entity, be more efficient and fetch them all in one query instead of a loop.
  if (scalar @nested_field_names == 1 &&
      $nested_def->{entity_type} && $nested_def->{entity_type} =~ /::Entity::/) {
      my $values = $dbh->selectcol_arrayref($sql, undef, $foreign_id->{field_value});

      @nested_entities = $nested_def->{entity_type}->retrieve_all(dbh => $dbh,
                                                         logger => $self->{logger},
                                                         policy => $self->{policy},
                                                         minimal_load => 1,
                                                         search_filters => [ F5::ASMConfig::SearchFilter->new(field_name => $nested_field_names[0],
                                                                                                              field_value_list => $values,
                                                                                                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN())],
                                                         selected_fields => [keys %$nested_fields],
                                                         );
      return \@nested_entities;
  }

  my $nested_sth = $dbh->prepare($sql);
  $nested_sth->execute($foreign_id->{field_value});
  while (my $nested_row = $nested_sth->fetchrow_hashref) {
    if ($nested_def->{entity_type}) {
      my %nested_db_row = map { ($_ => $nested_row->{$nested_fields->{$_}->{db_field_name}}) } @nested_field_names;
      my $nested_entity;
      if ($nested_def->{entity_type} =~ /::Entity::/) {
        $nested_entity = $nested_def->{entity_type}->new(dbh => $dbh,
                                                         logger => $self->{logger},
                                                         policy => $self->{policy},
                                                         minimal_load => 1,
                                                         selected_fields => [keys %$nested_fields],
                                                         get_criteria => \%nested_db_row);
      }
      else {
        $nested_entity = $nested_def->{entity_type}->new(\%nested_db_row);
        for my $field_name ((keys %nested_db_row), @serialized_field_names) {
            my $field_def  = $nested_def->{fields}->{$field_name};
            if ($field_def->{no_mapping}->{'to_entity'}) {
                $nested_entity->{$field_name} = $nested_db_row{$field_name};
            }
            else {
                $nested_entity->{$field_name} = $self->map_field_value(direction => 'to_entity',
                                                                       field_name => $field_name,
                                                                       field_value => $nested_db_row{$field_name},
                                                                       field_def => $field_def,
                                                                       dbh => $dbh,
                                                                       default_rd => $params{default_rd},
                                                                      );
            }
            if ($field_def->{serialized_field} && $nested_db_row{$field_def->{serialized_field}}) {
                eval {
                    $nested_entity->{$field_name} = thaw($nested_db_row{$field_def->{serialized_field}});
                };
                if ($@) {
                    $self->{logger}->info("Could not thaw value for $field_name: $@") if $self->{logger};
                }
            }
        }

      }

      push(@nested_entities, $nested_entity);
    }
    else {
      push(@nested_entities, $nested_row->{$nested_db_fields[0]});
    }
  }
  return \@nested_entities;
}

#----------------------------------------------------------------------

sub get_default_obj {
  my($self, %params) = @_;

  my $logger = $params{logger};
  my $dbh = $self->{dbh};
  my %default_values;

  my $fields = $params{all_fields} ? $self->all_fields : $self->fields;
  my %db_fields;
  for my $field_name (keys %$fields) {
    if ($params{rest_call} && exists $fields->{$field_name}->{rest_default_value}) {
        $default_values{$field_name} = $fields->{$field_name}->{rest_default_value};
    }
    elsif (exists $fields->{$field_name}->{default_value}) {
      $default_values{$field_name} = $fields->{$field_name}->{default_value};
    }
    elsif (exists $fields->{$field_name}->{const_value}) {
      $default_values{$field_name} = $fields->{$field_name}->{const_value};
    }
    elsif (my $func = $fields->{$field_name}->{default_value_func}) {
        $default_values{$field_name} = $self->$func(%params, field_def => $fields->{$field_name});
    }
    elsif ($fields->{$field_name}->{db_field_name}) {
      $db_fields{$fields->{$field_name}->{db_field_name}} = $field_name;
    }
    elsif ($fields->{$field_name}->{bitmask_field}) {
      $db_fields{$fields->{$field_name}->{bitmask_field}} = $fields->{$field_name}->{bitmask_field};
    }
  }

  if ($self->db_table) {
      my($database_name, $table_name) = split(/\./, $self->db_table);
      my $sth = $dbh->column_info(undef, $database_name, $table_name, '%'); 
      $sth->execute;
      my @primary_key;

      my $is_sqlite = F5::DbUtils::is_sqlite($dbh);
      while (my $row = $sth->fetchrow_hashref) {
        my $column_name = $row->{COLUMN_NAME};
        my $column_value = $row->{COLUMN_DEF};
        if ($is_sqlite && $column_value) {
            $column_value =~ s/'//g;
            if ($column_value eq 'NULL') {
                undef $column_value;
            }
        }

        if ($db_fields{$column_name}) {
          my $field_name = $db_fields{$column_name};

          next if ($fields->{$field_name}->{nested_struct});
          next if ($fields->{$field_name}->{external_struct});

          if (exists $fields->{$field_name}->{default_value}) {
            $default_values{$field_name} = $fields->{$field_name}->{default_value};
          }
          elsif (exists $fields->{$field_name}->{const_value}) {
              $default_values{$field_name} = $fields->{$field_name}->{const_value};
          }
          else {
            $default_values{$field_name} = $column_value;
          }
        }
      }
  }

  $self->map_fields_to_entity(rec => \%default_values,
                              for_default_object => 1,
                              dbh => $dbh, 
                              logger => $logger );

  my $nested_struct_values = $self->retrieve_default_nested_structures(%params);
  %default_values = (%default_values,
                     %$nested_struct_values);
  return \%default_values;
}

#----------------------------------------------------------------------

sub update_default_object_by_pb_settings {
    my ($self, %params) = @_;
    my $logger = $params{logger};
    my $dbh = $self->{dbh};
    my $policy = $params{policy};
    return unless ($policy && $policy->{policy_id});

    my $violation_toggle_fields = $self->violation_toggle_fields(%params);
    return unless %$violation_toggle_fields;

    my($flg_learning) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy->{policy_id});
SELECT flg_learning
  FROM PLC.PL_POLICY_BUILDER_CONFIG
 WHERE policy_id = ?
EndQuery
    return unless $flg_learning;

    my $default_obj = $params{default_obj};

    my @related_violations = map { $_->{violation} } values %$violation_toggle_fields;
    my $viol_index_clause = 'viol_index IN (' . join(',', ('?') x scalar @related_violations) . ')';
    my $violation_blocking_settings = $self->{dbh}->selectall_hashref(<<EndQuery, 'viol_index', undef, $policy->{policy_id}, @related_violations);
SELECT viol_index, flg_learn
  FROM PLC.PL_BLOCKING
 WHERE policy_id = ?
   AND $viol_index_clause
EndQuery

    for my $field (keys %$violation_toggle_fields) {
        if ($violation_toggle_fields->{$field}->{violation} == F5::Violation::VIOL_ATTACK_SIGNATURE()) {
            my($are_signatures_in_learning) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy->{policy_id});
SELECT flg_learn
  FROM PLC.PL_POLICY_NEGSIG_SETS
 WHERE policy_id = ?
   AND flg_learn = 1
 LIMIT 1
EndQuery
            $default_obj->{$field} = int(($are_signatures_in_learning || 0) == $violation_toggle_fields->{$field}->{value});
        }
        else {
            if ( $violation_toggle_fields->{$field}->{non_boolean} ) {
                if ($violation_blocking_settings->{$violation_toggle_fields->{$field}->{violation}}->{flg_learn} != $violation_toggle_fields->{$field}->{value}) {
                    $default_obj->{$field} = 0;
                }
            }
            else {
                $default_obj->{$field} = int($violation_blocking_settings->{$violation_toggle_fields->{$field}->{violation}}->{flg_learn} == $violation_toggle_fields->{$field}->{value});
            }
        }
    }
}

#----------------------------------------------------------------------

sub retrieve_default_nested_structures {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my %default_values;

  my $fields = $params{all_fields} ? $self->all_fields : $self->fields;
  my @nested_fields = @{$self->nested_struct_fields};
  if ($params{all_fields}) {
      my $diff_structs = $self->diff_structs;
      push (@nested_fields, grep { $diff_structs->{$_}->{nested_struct} } keys (%$diff_structs));
  }

  local $" = ',';

  for my $field_name (@nested_fields) {
    my $field_rec = $fields->{$field_name};
    if ($field_rec->{defaults_table}){
      my $nested_fields = $field_rec->{fields};
      my @nested_field_names = grep { $nested_fields->{$_}->{db_field_name} } keys %$nested_fields;
      my @nested_db_fields = map { $nested_fields->{$_}->{db_field_name} } @nested_field_names;
      
      my $table_name = $field_rec->{defaults_table};
      
      my @defaults_where_clauses = (1);
      my @bind_params;
      if ($field_rec->{defaults_table_where}) {
          for my $where_field (keys %{$field_rec->{defaults_table_where}}) {
              push (@defaults_where_clauses, "$where_field = ?");
              push (@bind_params, $field_rec->{defaults_table_where}->{$where_field});
          }
      }
      my $defaults_where = join (" AND ", @defaults_where_clauses);

      my $nested_sth = $dbh->prepare(<<EndQuery);
SELECT *
  FROM $table_name
  WHERE $defaults_where
EndQuery

      $nested_sth->execute(@bind_params);
      while(my $nested_row = $nested_sth->fetchrow_hashref) {
        if ($field_rec->{entity_type}) {
          my %nested_db_row = map { ($_ => $nested_row->{$nested_fields->{$_}->{db_field_name}}) } keys %$nested_fields;
          my $nested_entity;
          if ($field_rec->{entity_type} =~ /::Entity::/) {
            $nested_entity = $field_rec->{entity_type}->new(dbh => $dbh,
                                                            logger => $self->{logger},
                                                            minimal_load => 1,
                                                            get_criteria => \%nested_db_row);
          }
          else {
            $nested_entity = $field_rec->{entity_type}->new(\%nested_db_row);
          }
          push(@{$default_values{$field_name}}, $nested_entity);
        }
        else {
          push(@{$default_values{$field_name}}, $nested_row->{$nested_db_fields[0]});
        }
      }
    }
    elsif ($field_rec->{external_struct}) {
        my $entity_type = $field_rec->{entity_type};
        my @defaults;
        if ($entity_type->db_defaults_table) {
           @defaults = $entity_type->retrieve_all(dbh => $dbh,
                                                  logger => $self->{logger},
                                                  default_table => $entity_type->db_defaults_table,
                                                  find_single_record => 0,
                                                  search_filters => $params{search_filters},
                                                 );
       }
       $default_values{$field_name} = \@defaults;
    }
  }

  return \%default_values;
}

#----------------------------------------------------------------------

sub retrieve_all {
    my($class, %params) = @_;

    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $get_criteria = $params{get_criteria} || {};
    my $default_table = $params{default_table};
    my $default_rd = defined $params{default_rd} ? $params{default_rd} : 0;
    my $call_flags = $params{call_flags} || {};

    my $find_single_record = (defined($params{find_single_record}) ? $params{find_single_record} : 0);
    my $minimal_load = $params{minimal_load};
    my $full_policy_populate = $params{full_policy_populate} || $call_flags->{full_policy_populate};
    my $populate_diff_structs = $params{populate_diff_structs};
    local $" = ',';
    # Handle casting bind variables as numbers so they match the results of functions like COUNT or CASE evals
    # See: https://metacpan.org/pod/DBD::SQLite#Functions-And-Bind-Parameters
    local $dbh->{sqlite_see_if_its_a_number} = 1;

    my @search_filters;
    if ($params{search_filters}) {
        @search_filters = @{$params{search_filters}};
    }

    my @grouping_fields;
    if ($params{grouping_fields}) {
        @grouping_fields = @{$params{grouping_fields}};
    }
    my @grouping_aggregation_fields;
    if ($params{grouping_aggregation_fields}) {
        @grouping_aggregation_fields = @{$params{grouping_aggregation_fields}};
    }

    my $aggregate_field_aliases;
    if ($call_flags->{aggregate_fields}) {
        my ($aliases) = values %{$call_flags->{aggregate_fields}};
        $aggregate_field_aliases = { map { ($_ => 1) } keys (%$aliases) };
    }

    my $extra_fields = [];
    if ( $params{extra_fields} ) {
        $extra_fields = $params{extra_fields};
    }
    elsif( $params{call_flags} && $params{call_flags}->{extra_fields} ){
        $extra_fields = $params{call_flags}->{extra_fields};
    }

    my $selected_fields_hash;
    my $selected_fields = $params{selected_fields} || $call_flags->{selected_fields};
    my $rest_selected_fields = $params{rest_selected_fields} || $call_flags->{rest_selected_fields};
    if ($selected_fields) {
        $selected_fields_hash = { map { ($_ => 1) } @$selected_fields };
    }
    elsif ($rest_selected_fields) {
        $selected_fields_hash = $class->rest_selected_fields_to_selected_fields_hash(rest_selected_fields => $rest_selected_fields);
        $extra_fields = $class->rest_selected_fields_to_extra_fields(rest_selected_fields => $rest_selected_fields,
                                                                     extra_fields => $extra_fields);
    }
    my $extra_fields_hash = { map { ( $_ => 1 ) } @{$extra_fields} };

    my $fields = $class->fields;
    my @extern_db_prefixes = $class->extern_db_prefixes(get_criteria => $get_criteria,
                                                        extra_fields => $extra_fields,);
    my @db_fields;
    my %already_added_bitmask_fields;

    my $additional_field_mapping = $class->additional_field_mapping(dbh => $dbh,
                                                                    get_criteria => $get_criteria,
                                                                    extra_fields => $extra_fields,
                                                                    extra_fields_hash => $extra_fields_hash,);

    $fields = {%$fields, %$additional_field_mapping};

    my $policy = F5::ASMConfig::Cache::get_cached_policy(%params);

    if ($get_criteria) {
        for my $field_name (keys %$fields) {
            if (exists ($fields->{$field_name}->{const_value})) {
                # Populate get_criteria to fulfill unique_record_parts on update entity call without store_obj
                $get_criteria->{$field_name} = $fields->{$field_name}->{const_value};

                push(@search_filters,
                     {
                      field_name => $field_name,
                      field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                      field_value => $fields->{$field_name}->{const_value},
                     } );
            }
        }
    }

    my $policy_id_field = $class->policy_id_field;
    if ($policy_id_field && ! defined $default_table) {
        push(@db_fields, "main_table.$policy_id_field");
    }

    my $db_table = (defined $default_table) ? $default_table:  $class->db_table;


    my $id_field = $class->id_field;
    my @where_clauses;
    my @where_values;
    my $external_field_structs = {};

    if ( not defined $default_table) {
        if ($id_field && $get_criteria->{$id_field}) {
            push(@where_clauses, 'main_table.' . $class->id_db_field . ' = ?');
            push(@where_values, $get_criteria->{$id_field});
        }
        else {
            if ($policy && $policy->{policy_id} && $policy_id_field) {
                push(@where_clauses, 'main_table.' . $policy_id_field . ' = ?');
                push(@where_values, $policy->{policy_id});
            }

            my @filter_fields;
            my @unique_record_parts = @{$class->unique_record_parts(get_criteria => $get_criteria)};
            if (($find_single_record || scalar @unique_record_parts) && 
                scalar @unique_record_parts == scalar grep { defined $get_criteria->{$_} } @unique_record_parts) {
                # Getting a single record - only filter on unique record parts
                @filter_fields = @unique_record_parts;
            }
            else {
                @filter_fields = grep { defined $get_criteria->{$_} } keys %$get_criteria;
            }

            for my $filter_field (@filter_fields) {
                if (ref($get_criteria->{$filter_field})) {

                    if (exists($fields->{$filter_field}) && $fields->{$filter_field}->{external_struct} && !$fields->{$filter_field}->{populate_from_join_row}) {
                        if (ref($get_criteria->{$filter_field}) =~ /::Entity::/) {
                            my $thrift_class = ref($get_criteria->{$filter_field});
                            $thrift_class =~ s/::Entity::/::/;
                            $get_criteria->{$filter_field} = $thrift_class->new($get_criteria->{$filter_field});
                        }

                        # Do not convert external_struct obj into individual search_filters -- pass it instead as get_criteria
                        $external_field_structs->{$filter_field}->{get_criteria} = $get_criteria->{$filter_field};
                    }
                    elsif (exists($fields->{$filter_field})) {
                        my @additional_search_filters = $class->get_additional_search_filters(filter_field => $filter_field,
                                                                                              filter_value => $get_criteria->{$filter_field});

                        for my $additional_search_filter (@additional_search_filters) {
                            push(@search_filters,
                                 {
                                  field_name => $additional_search_filter->{field_name},
                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                  field_value => $additional_search_filter->{field_value} } );
                        }
                    }
                }
                else {
                    my $field_def = $fields->{$filter_field};
                    push(@search_filters,
                         { field_name => $filter_field,
                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                           field_value => $get_criteria->{$filter_field} });
                }
            }
        }
    }

    my @having_clauses;
    my @having_values;
    my @subselect_values;
    my $join_tables = '';
    my @join_tables_bind_values;

    for my $search_filter_index (reverse 0..$#search_filters) {
        my $search_filter = $search_filters[$search_filter_index];
        my $filter_field = $search_filter->{field_name};
        next if (! $filter_field);

        my $nested_subfield;
        my $struct_name;
        if ($filter_field =~ /^(.*?)__(.*)$/) {
            $struct_name = $1;
            $nested_subfield = $2;
            if ($fields->{$struct_name} && $fields->{$struct_name}->{external_struct} && !$fields->{$struct_name}->{populate_from_join_row}) {
                $search_filter = dclone($search_filter);
                $search_filter->{field_name} = $nested_subfield;
                push @{$external_field_structs->{$struct_name}->{search_filters}}, $search_filter;
                splice(@search_filters, $search_filter_index, 1);
            }
        }
    }

    for my $struct_name (keys %$external_field_structs) {
        push (@search_filters, { field_name => $struct_name,
                                 %{$external_field_structs->{$struct_name}} });
    }

    my %collection_nested; # support multiple nested structs
    
    for my $i (0..$#search_filters) {
        my $search_filter = $search_filters[$i];
        my $filter_field = $search_filter->{field_name};
        next if (! $filter_field);

        my $nested_subfield;
        if ( !( exists($fields->{$filter_field}) && $fields->{$filter_field}->{full_db_field_name}) && $filter_field =~ s/__(.*)$//) {
            $nested_subfield = $1;
        }

        if (exists ($fields->{$filter_field}) && $fields->{$filter_field}->{external_struct} && !$fields->{$filter_field}->{populate_from_join_row}
              && ($search_filter->{get_criteria} || $search_filter->{search_filters})) {
            my @external_structs = $fields->{$filter_field}->{entity_type}->retrieve_all(dbh => $dbh,
                                                                                         get_criteria => $search_filter->{get_criteria},
                                                                                         search_filters => $search_filter->{search_filters},
                                                                                         minimal_load => 1,
                                                                                         selected_fields => [],
                                                                                         policy => $policy,
                                                                                         logger => $params{logger},
                                                                                        );

            if ( ! @external_structs) {
                return return_empty_result_set(want_array => wantarray);
            }

            my $ext_id_field = $fields->{$filter_field}->{foreign_field} || $fields->{$filter_field}->{entity_type}->id_field();
            $search_filter = {
                              field_name => $search_filter->{field_name},
                              field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                              field_value_list => [ map { $_->{$ext_id_field} } @external_structs ],
                             };
        }

        my $comparison_mapping = get_comparison_mapping(dbh => $dbh);
        my @search_field_where_values;
        my $compare_clause = '';
        if ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN() ||
            $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN() ||
            $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_IN() || 
            $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_NOT_IN()) {

            if (! $search_filter->{field_value_list} || ! @{$search_filter->{field_value_list}}) {
                return return_empty_result_set(want_array => wantarray);
            }

            @search_field_where_values = @{$search_filter->{field_value_list}};
            my @placeholders = ('?') x scalar @search_field_where_values;
            $compare_clause = "$comparison_mapping->{$search_filter->{field_comparison}} (@placeholders)";
        }
        elsif ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NULL() ||
               $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NOT_NULL()) {

            $compare_clause = "$comparison_mapping->{$search_filter->{field_comparison}}";
        }
        else {
            @search_field_where_values = ($search_filter->{field_value});
            if ($comparison_mapping->{$search_filter->{field_comparison}}) {
                $compare_clause = "$comparison_mapping->{$search_filter->{field_comparison}} ?";
            }
        }

        my $filter_on_policy_id = ($filter_field eq 'policy_id' && $policy_id_field );
        if (($fields->{$filter_field} && ! $fields->{$filter_field}->{use_complex_filter})|| $filter_on_policy_id) {
            if ($selected_fields_hash) {
                $selected_fields_hash->{$filter_field} = 1;
            }
            if ($fields->{$filter_field}->{extra_field}) {
                $extra_fields_hash->{$filter_field} = 1;
            }

            my $db_field_name = $fields->{$filter_field}->{db_field_name} || ($filter_on_policy_id && $policy_id_field);
            if ($db_field_name) {
                push(@where_clauses, "main_table.$db_field_name $compare_clause");
                
                if ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP()) {
                    
                    # no need to change or validate values of patterns
                    push(@where_values, @search_field_where_values);
                }
                elsif ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EXISTS() ||
                       $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS()) {
                    push(@where_values, "");
                }
                else {
                    push(@where_values,  map {  $fields->{$filter_field}->{no_mapping}->{'to_db'} ? $_
                                                  : $class->map_field_value(
                                                                        field_value => $_,
                                                                        field_name => $filter_field,
                                                                        field_def => $fields->{$filter_field},
                                                                        policy => $policy,
                                                                        direction => 'to_db',
                                                                        dbh => $dbh,
                                                                        default_rd => $default_rd,
                                                                        validate_value => 0,
                                                                        call_flags => $call_flags,
                                                                       )
                                            } @search_field_where_values);
                }
            }

            elsif ($fields->{$filter_field}->{filter_using_having}) {
                push(@having_clauses, "$filter_field $compare_clause");
                push(@having_values, @search_field_where_values);
            }
            elsif ($fields->{$filter_field}->{full_db_field_name}) {
                push(@where_clauses, "$fields->{$filter_field}->{full_db_field_name} $compare_clause");
                push(@where_values, @search_field_where_values);
            }

            elsif ($fields->{$filter_field}->{extern_db_field_name}) {
                my @or_clauses;
                my @prefixes = $class->get_extern_db_prefix(field_def => $fields->{$filter_field}) || @extern_db_prefixes;
                for my $extern_db_prefix (@prefixes) {
                    push(@or_clauses, "$extern_db_prefix.$fields->{$filter_field}->{extern_db_field_name} $compare_clause");
                    push(@where_values, map {  $fields->{$filter_field}->{no_mapping}->{'to_db'} ? $_
                                                 : $class->map_field_value(
                                                                       field_value => $_,
                                                                       field_name => $filter_field,
                                                                       field_def => $fields->{$filter_field},
                                                                       policy => $policy,
                                                                       direction => 'to_db',
                                                                       dbh => $dbh,
                                                                       default_rd => $default_rd,
                                                                       call_flags => $call_flags,
                                                                      )
                                           } @search_field_where_values);
                    if (defined($fields->{$filter_field}->{default_value}) &&
                        $filter_func{$search_filter->{field_comparison}}->(search_filter => $search_filter,
                                                                           field_value => $fields->{$filter_field}->{default_value})) {
                        push(@or_clauses, "$extern_db_prefix.$fields->{$filter_field}->{extern_db_field_name} IS NULL");
                    }
                }

                if (@or_clauses) {
                    push(@where_clauses, '(' . join(' OR ', @or_clauses) . ')');
                }
            }

            elsif ($fields->{$filter_field}->{nested_struct}) {
                my @nested_field_names = keys %{$fields->{$filter_field}->{fields}};

                if (!$nested_subfield) {
                    # Use key field for search
                    ($nested_subfield) = grep { $fields->{$filter_field}->{fields}->{$_}->{unique_record_part} } @nested_field_names;
                }
                my $nested_subfield_db_field_name = $fields->{$filter_field}->{fields}->{$nested_subfield}->{db_field_name};
    
                my $nested_db_table_name = $fields->{$filter_field}->{db_table_name};
                my $foreign_id = $class->get_foreign_id(nested_def => $fields->{$filter_field}); 
                next unless $nested_db_table_name && $foreign_id->{parent_table_alias} && $foreign_id->{parent_db_field_name} && $foreign_id->{field_name};

                my $compare_value;
                if ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EXISTS() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN() ||
                    $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN()) {
                    $compare_value = 1;
                }
                elsif ($search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS() ||
                       $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE() ||
                       $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_NE() ||
                       $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_IN() ||
                       $search_filter->{field_comparison} == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_ALL_NESTED_NOT_IN() ) {
                    $compare_value = 0;
                }

                my $exist_field_name = "${filter_field}__exists";
                if (@search_field_where_values) {
                    my $subfield_compare_clause = '';

                    if ( $search_filter->{field_comparison} != F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EXISTS() &&
                         $search_filter->{field_comparison} != F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS()) {

                        $subfield_compare_clause = "$nested_db_table_name.$nested_subfield_db_field_name $compare_clause";
                        
                        if (!$search_filter->{collective_nested_field}) {
                            push(@subselect_values, @search_field_where_values);
                        }
                    }

                    if ($search_filter->{collective_nested_field}) {
                        if (!$collection_nested{$nested_db_table_name}) { # first time
                            $collection_nested{$nested_db_table_name} = { 
                                              where_clauses => [ "$foreign_id->{parent_table_alias}.$foreign_id->{parent_db_field_name} = $nested_db_table_name.$foreign_id->{field_name}" ],
                                              where_values => [],
                                              exist_field_name => "${filter_field}__collective_nested__exists", # without collective_nested - already in use
                                             };
                        }
                        
                        push(@{$collection_nested{$nested_db_table_name}->{where_clauses}}, $subfield_compare_clause);
                        push(@{$collection_nested{$nested_db_table_name}->{where_values}}, @search_field_where_values);
                        $exist_field_name = undef; # no need for another having clause
             	    }
                    else {
                        $exist_field_name = "${filter_field}__${nested_subfield}__exists";

                        $subfield_compare_clause ||= '1';
                        my $subselect = <<EndQuery;
EXISTS (SELECT *
          FROM $nested_db_table_name
         WHERE $foreign_id->{parent_table_alias}.$foreign_id->{parent_db_field_name} = $nested_db_table_name.$foreign_id->{field_name}
           AND $subfield_compare_clause )
EndQuery
                        push(@db_fields, "$subselect AS $exist_field_name");
                    }
                }

                if (defined $exist_field_name) {
                    push(@having_clauses, "$exist_field_name = $compare_value");
                }
            }
            elsif ($fields->{$filter_field}->{defaults_db_field_name}) {
                push(@where_clauses, "main_table.$fields->{$filter_field}->{defaults_db_field_name} $compare_clause");
                push(@where_values,  @search_field_where_values);
            }
        }
        else {

            if (my $complex_filter = $class->get_complex_filter(%params,
                                                                extra_fields => $extra_fields_hash,
                                                                filter_field => $filter_field,
                                                                field_comparison => $search_filter->{field_comparison},
                                                                field_value => $search_filter->{field_value},
                                                                field_value_list => $search_filter->{field_value_list})) {

                if ($complex_filter->{where_clause}) {
                    push(@where_clauses, $complex_filter->{where_clause});
                }
                if ($complex_filter->{where_values}) {
                    push(@where_values, @{$complex_filter->{where_values}});
                }
                if ($complex_filter->{having_clause}) {
                    push(@having_clauses, @{$complex_filter->{having_clauses}});
                }
                if ($complex_filter->{having_values}) {
                    push(@having_values, @{$complex_filter->{having_values}});
                }
                if ($complex_filter->{join_clause}) {
                    $join_tables .= $complex_filter->{join_clause};
                }
                if ($complex_filter->{join_values}) {
                    push(@join_tables_bind_values, @{$complex_filter->{join_values}});
                }
            }
        }
    }

    for my $collective_nested_db_table_name (keys %collection_nested) {
        my $collective_nested_where_clause = join(' AND ', @{$collection_nested{$collective_nested_db_table_name}->{where_clauses}});
        
        my $subselect = <<EndQuery;
EXISTS (SELECT *
          FROM $collective_nested_db_table_name
         WHERE $collective_nested_where_clause)
            AS $collection_nested{$collective_nested_db_table_name}->{exist_field_name}
EndQuery
        
        push(@db_fields, $subselect);
        push(@subselect_values, @{$collection_nested{$collective_nested_db_table_name}->{where_values}});
        push(@having_clauses, "$collection_nested{$collective_nested_db_table_name}->{exist_field_name} = 1");
    }

    push(@where_clauses, $class->additional_where_clauses(%params));

    my $where_clause = '';
    if (@where_clauses) {
        $where_clause = 'WHERE ' . join(' AND ', @where_clauses);
    }

    my $drop_grouping_clause_for_sqlite_count = 0;
    if (@having_clauses && !@grouping_fields && $id_field && F5::DbUtils::is_sqlite($dbh)) {
        push (@grouping_fields, ("main_table." . $class->id_db_field));
        $drop_grouping_clause_for_sqlite_count = 1;
    }

    my $grouping_clause = '';
    if (@grouping_fields) {
	if (@grouping_aggregation_fields) {
	    my $aggregation_fields = join(',', @grouping_aggregation_fields);
	    push(@db_fields, "COUNT(DISTINCT $aggregation_fields) AS grouping_count");
	}
	else {
	    push(@db_fields, 'COUNT(*) AS grouping_count');
	}
        $grouping_clause = 'GROUP BY  ' . join(', ', @grouping_fields);
    }

    my $having_clause = '';
    if (@having_clauses) {
        $having_clause = 'HAVING ' . join(' AND ', @having_clauses);
    }

    #Using bind values on a field name that can have a "having" filter breaks the number of bind variables for the count query
    #So these fields will be interpolating their values here (Only numeric arguments)
    if ($extra_fields_hash->{extra_has_suggestions} && $class->fields->{extra_has_suggestions}) {
        my $staging_subquery = $class->get_staging_subquery(%params,
                                                             policy => $policy,
                                                             has_suggestions => 1,);
        push(@db_fields, "$staging_subquery->{db_field} AS extra_has_suggestions");
    }
    if ($extra_fields_hash->{extra_has_tightening_suggestions} && $class->fields->{extra_has_tightening_suggestions}) {
        my $tightening_subquery = $class->get_tightening_subquery(%params,
                                                                  policy => $policy,
                                                                  has_suggestions => 1,);
        push(@db_fields, "$tightening_subquery->{db_field} AS extra_has_tightening_suggestions");
    }

    if ($extra_fields_hash->{extra_updated_in_enforcement_readiness_period} && $class->fields->{extra_updated_in_enforcement_readiness_period}) {
        my $query = $class->get_updated_in_enforcement_readiness_period_subquery(%params,
                                                                              policy => $policy,);
        push(@db_fields, $query);
    }
    if ($extra_fields_hash->{extra_learned_in_enforcement_readiness_period} && $class->fields->{extra_learned_in_enforcement_readiness_period}) {
        my $query = $class->get_learned_in_enforcement_readiness_period_subquery(%params,
                                                                              policy => $policy,);
        push(@db_fields, $query);
    }

    my $join_tables_rec = $class->get_retrieve_join_tables(get_criteria => $get_criteria,
                                                           extra_fields => $extra_fields,
                                                           search_filters => \@search_filters,
                                                           selected_fields_hash => $selected_fields_hash,
                                                           policy => $policy,
                                                          );
    if (ref($join_tables_rec)) {
        $join_tables .= $join_tables_rec->{query};
        push(@join_tables_bind_values, @{$join_tables_rec->{bind_values}});
    }
    else {
        $join_tables .= $join_tables_rec;
    }

    my $orderby_clause = '';

    if (!$params{sort_orders} && !$default_table) {
        $params{sort_orders} = $class->default_sort_orders(%params);
    }
    
    if ($params{sort_orders} && @{$params{sort_orders}}) {
        my @orderby_clauses;
        my %order_joins;
        for my $sort_order (@{$params{sort_orders}}) {
            my $sort_direction = $DIRECTION_MAPPING{$sort_order->{sort_direction}} || 'ASC';
            my $binary_sort = (F5::DbUtils::is_sqlite($dbh) ? '' : $BINARY_SORT_MAPPING{$sort_order->{sort_direction}}) || '';

            if ($fields->{$sort_order->{field_name}} && $fields->{$sort_order->{field_name}}->{db_field_name}) {
                my $field_name = $fields->{$sort_order->{field_name}}->{db_field_name};
                push(@orderby_clauses, "$binary_sort main_table.$field_name $sort_direction");
            }

            elsif ($fields->{$sort_order->{field_name}} && $fields->{$sort_order->{field_name}}->{full_db_field_name}) {
                push(@orderby_clauses, "$binary_sort $sort_order->{field_name} $sort_direction");
            }

            elsif ($fields->{$sort_order->{field_name}} && $fields->{$sort_order->{field_name}}->{extern_db_field_name}) {

                my @coalesce_orderby_clauses;
                my @prefixes = $class->get_extern_db_prefix(field_def => $fields->{$sort_order->{field_name}}) || @extern_db_prefixes;
                for my $extern_db_prefix (@prefixes) {
                    my $field_name = "$binary_sort $extern_db_prefix.$fields->{$sort_order->{field_name}}->{extern_db_field_name}";
                    push(@coalesce_orderby_clauses, $field_name);
                }
                push(@orderby_clauses, 'COALESCE(' . join(',', @coalesce_orderby_clauses) . ") $sort_direction");
            }

            elsif ($sort_order->{field_name} eq 'grouping_count' && @grouping_fields) {
                push(@orderby_clauses, "$binary_sort $sort_order->{field_name} $sort_direction");
            }

            elsif ($aggregate_field_aliases->{$sort_order->{field_name}}) {
                push(@orderby_clauses, "$binary_sort $sort_order->{field_name} $sort_direction");
            }

            elsif ($sort_order->{field_name} =~ /^(.*?)__(.*)$/) {
                my ($field_name, $sub_field) = ($1, $2);
                if ($fields->{$field_name} && ($fields->{$field_name}->{external_struct} || $fields->{$field_name}->{rest_entity_type})
                    && !$fields->{$field_name}->{populate_from_join_row}) {

                    my $entity_type = $fields->{$field_name}->{entity_type} || $fields->{$field_name}->{rest_entity_type};
                    my $ext_table = $entity_type->db_table;
                    my $ext_alias = "sort_" . $entity_type->element_type_name;
                    my $ext_id_field = $fields->{$field_name}->{foreign_field} || $entity_type->id_field();
                    my $ext_id_db_field = $entity_type->fields->{$ext_id_field}->{db_field_name};
                    my $sub_field_db_field = $entity_type->fields->{$sub_field}->{db_field_name};
                    $order_joins{$ext_alias} = <<EndQuery;
LEFT JOIN $ext_table $ext_alias 
  ON main_table.$fields->{$field_name}->{db_field_name} = $ext_alias.$ext_id_db_field
EndQuery
                    push (@orderby_clauses, "$binary_sort $ext_alias.$sub_field_db_field $sort_direction");
                }
            }
        }
        if (my @orderby_joins = values %order_joins) {
            $join_tables .= join ("", @orderby_joins);
        }
        if (@orderby_clauses) {
            $orderby_clause = 'ORDER BY ' . join(',', @orderby_clauses);
        }
    }

    my $limit_clause = '';
    my $expected_total_records = 0;
    if ($params{page_size} && !$find_single_record) {
        $params{page_size} = int($params{page_size});
        $params{start_page} = (defined $params{start_page}) && ($params{start_page} > 0) ? $params{start_page} : 1;
        my $skip_record_count = ($params{start_page} - 1) * $params{page_size};
        $limit_clause = "LIMIT $skip_record_count, $params{page_size}";
        $expected_total_records = $skip_record_count + $params{page_size};
    }

    if ($find_single_record) {
        $limit_clause = 'LIMIT 1';
    }

    my $exclude_from_default_table_fields = defined($default_table) ? $class->exclude_from_default_table_fields(%params) : {};

    for my $field_name (keys %$fields) {
        next if (( $fields->{$field_name}->{extra_field} && !$extra_fields_hash->{$field_name} )
                   || ( defined($default_table) && $exclude_from_default_table_fields->{$field_name} ));
        if ($selected_fields_hash && !(
                                       $fields->{$field_name}->{id_field} ||
                                       $fields->{$field_name}->{unique_record_part} ||
                                       $fields->{$field_name}->{pb_unique_record_part} ||
                                       $fields->{$field_name}->{always_select_field} ||
                                       $fields->{$field_name}->{rest_reference_field} ||
                                       exists ($fields->{$field_name}->{const_value}) ||
                                       # Extern fields with double underscore are needed for joining tables that
                                       # are part of unique record parts (e.g. for extractions)
                                       ($fields->{$field_name}->{extern_db_field_name} && $field_name =~ /__/ ) ||
                                       $field_name eq 'rest_uuid' ||
                                       $selected_fields_hash->{$field_name})) {
            next;
        }
        if ($fields->{$field_name}->{db_field_name}) {
            if ($fields->{$field_name}->{external_struct}) {
                push(@db_fields,
                     "main_table.$fields->{$field_name}->{db_field_name} as $fields->{$field_name}->{db_field_name}");
            }
            else {
                if ( not ( defined $default_table && $fields->{$field_name}->{id_field} )) {
                    my $db_field_name = "main_table.$fields->{$field_name}->{db_field_name}";
                    push(@db_fields,
                         "$db_field_name as $field_name");

                    if ($call_flags->{aggregate_fields}->{$field_name}) {
                        for my $alias (keys %{$call_flags->{aggregate_fields}->{$field_name}}) {
                            my $func = $call_flags->{aggregate_fields}->{$field_name}->{$alias};
                            push (@db_fields, "$func($db_field_name) as $alias");
                        }
                    }
                }
            }
        }

        elsif (my $bitmask_field = $fields->{$field_name}->{bitmask_field}) {
            # Add bitmask field to db_fields only once
            if ( ! $already_added_bitmask_fields{$bitmask_field}) {
                push(@db_fields,
                     "main_table.$bitmask_field as $bitmask_field");

                $already_added_bitmask_fields{$bitmask_field} = 1;
            }
        }

        elsif ($fields->{$field_name}->{full_db_field_name}) {
            my $full_db_field_name = $fields->{$field_name}->{full_db_field_name};
            push(@db_fields, "$full_db_field_name AS $field_name");
        }
        elsif ($fields->{$field_name}->{extern_db_field_name}) {
            my @prefixes = $class->get_extern_db_prefix(field_def => $fields->{$field_name}) || @extern_db_prefixes;
            for my $extern_db_prefix (@prefixes) {
                push(@db_fields, "$extern_db_prefix.$fields->{$field_name}->{extern_db_field_name} AS ${extern_db_prefix}__$field_name");
            }
        }
        elsif ($fields->{$field_name}->{referenced_tables}) {
            my @references;
            for my $reference (@{$fields->{$field_name}->{referenced_tables}}) {
                my $reference_table = $reference->{db_table_name};
                if ($reference->{join_db_table_name}) {
                    $reference_table .= " JOIN $reference->{join_db_table_name}"
                                     .  " ON $reference->{db_table_name}.$reference->{local_join_field} = $reference->{join_db_table_name}.$reference->{foreign_join_field}";
                }
                my $parent_id_field = $reference->{parent_id_field} || $class->id_db_field;
                my @reference_clauses = ("$reference->{foreign_id_field} = main_table.$parent_id_field");
                #Note: Where values only work for numeric values, if strings are needed this will have to be quoted in yaml or code.
                if ($reference->{where_values}) {
                    for my $where_field (keys %{$reference->{where_values}}) {
                        push (@reference_clauses, "$where_field = $reference->{where_values}->{$where_field}");
                    }
                }
                if (!$class->id_field && $class->policy_id_field) {
                    push (@reference_clauses, "$reference->{db_table_name}.policy_id = main_table.policy_id");
                }
                my $reference_where = join(' AND ', @reference_clauses);
                if ($reference->{count}) {
                    push (@references, "SELECT COUNT(*) FROM $reference_table WHERE $reference_where");
                }
                else {
                    push (@references, "EXISTS(SELECT * FROM $reference_table WHERE $reference_where)");
               }
            }
            my $full_db_field_name = join (' OR ', @references) ;
            push (@db_fields, "($full_db_field_name) AS $field_name");
        }

    }

    # Add existence check for nested struct
    my @nested_struct_fields;
    unless ($minimal_load) {
        push (@nested_struct_fields, @{$class->nested_struct_fields(%params)});
        push (@nested_struct_fields, grep {$additional_field_mapping->{$_}->{nested_struct}} (keys %$additional_field_mapping));

        for my $field_name (@nested_struct_fields) {
            next if ($selected_fields_hash && !$selected_fields_hash->{$field_name});
            next if ($fields->{$field_name}->{extra_field} && !$extra_fields_hash->{$field_name});
            next unless $fields->{$field_name}->{db_table_name};
            my $nested_db_table_name = $fields->{$field_name}->{db_table_name};
            my $foreign_id = $class->get_foreign_id(nested_def => $fields->{$field_name});

            my $subselect = <<EndQuery;
EXISTS (SELECT *
          FROM $nested_db_table_name
         WHERE $foreign_id->{parent_table_alias}.$foreign_id->{parent_db_field_name} = $nested_db_table_name.$foreign_id->{field_name}) AS ${field_name}__exists
EndQuery
            push(@db_fields, $subselect);
        }
    }

    # If no fields were selected, add placeholder to ensure legal SQL.
    # In this case, the number of rows returned will still reflect WHERE conditions.
    if (! @db_fields) {
        push(@db_fields, "1");
    }

    my $sql = <<EndQuery;
SELECT @db_fields 
FROM $db_table AS main_table
$join_tables
$where_clause
$grouping_clause
$having_clause 
$orderby_clause
$limit_clause
EndQuery

    my @bind_values = (@subselect_values, @join_tables_bind_values, @where_values, @having_values);

    if ($verbose_logger && $verbose_logger->is_trace()) {
        $verbose_logger->trace($sql . "\n VALUES (" . join (",", @bind_values) . ")")
    }

    my $sth;
    if (F5::DbUtils::is_sqlite($dbh)) {
        $sth = $dbh->prepare($sql);
    }
    else {
        $sth = $dbh->prepare_cached($sql, undef , 3);
    }

    $sth->execute(@bind_values);

    my $total_record_count;
    my $found_rows;
    if (!$find_single_record && !$params{skip_total_record_count} && !$call_flags->{skip_total_record_count}) { 
        my $get_total_record_count;
        if (F5::DbUtils::is_sqlite($dbh)) {
            #Force getting the true count since sqlite doesn't support FOUND_ROWS
            $get_total_record_count = 1;
        }
        else {
            ($total_record_count) = $dbh->selectrow_array('SELECT FOUND_ROWS()');
            $found_rows = $total_record_count;
            $get_total_record_count = ($expected_total_records && $expected_total_records == $total_record_count);
        }
        if ($get_total_record_count) {
            # Field aliases in HAVING clause need to be included even in the SELECT COUNT field list
            my @where_having_clauses = ('1');
            for my $having_clause (@having_clauses) {
                # Grab field name/alias from beginning of clause
                if ($having_clause =~ /^\s*(\S+)\s(.*)$/) {
                    my ($having_clause_field, $comparison) = ($1, $2);
                    my($having_field) = grep { /\s+as\s+\Q$having_clause_field\E\b/is } @db_fields;
                    if ($having_field) {
                        $having_field =~ /^\s*(.*)\s+as\s+\Q$having_clause_field\E\b/is;
                        my $field_sql = $1;

                        push(@where_having_clauses, "$field_sql $comparison");
                    }
                }
            }
            my $where_having_sql = " WHERE " . join(' AND ', @where_clauses, @where_having_clauses);

            my $join_tables_for_count = '';
            my @join_tables_for_count_bind_values;
            if ($join_tables) {
                my $join_tables_for_count_rec = $class->get_retrieve_join_tables(get_criteria => $get_criteria,
                                                                                 extra_fields => $extra_fields,
                                                                                 search_filters => \@search_filters,
                                                                                 selected_fields_hash => $selected_fields_hash,
                                                                                 only_if_needed_for_count => 1,
                                                                                 policy => $policy,
                                                                                );
                if (ref($join_tables_for_count_rec)) {
                    $join_tables_for_count = $join_tables_for_count_rec->{query};
                    @join_tables_for_count_bind_values = @{$join_tables_for_count_rec->{bind_values}};
                } else {
                    $join_tables_for_count = $join_tables_for_count_rec;
                }
            }

            if ($drop_grouping_clause_for_sqlite_count) {
                $grouping_clause = '';
            }

            my @count_bind_values = (@join_tables_for_count_bind_values, @where_values, @subselect_values, @having_values);

            eval {
                ($total_record_count) = $dbh->selectrow_array(<<EndQuery, undef, @count_bind_values);
SELECT COUNT(*)
FROM $db_table AS main_table
$join_tables_for_count
$where_having_sql
$grouping_clause
EndQuery
                if ($grouping_clause && !F5::DbUtils::is_sqlite($dbh)) {
                    ($total_record_count) = $dbh->selectrow_array('SELECT FOUND_ROWS()');
                }
            };
            if (my $err = $@) {
                if ($params{tolerate_total_record_count_failure}) {
                    $logger->info("Total record count could not be retrieved (database error encountered = $err)") if $logger;
                }
                else {
                    die $err;
                }
            }
        }
    }


    my $object_constructor;
    if ($params{object_constructor}) {
        $object_constructor = $params{object_constructor};
    }
    elsif ($params{return_as_raw_db_hashref}) {
        $object_constructor = sub {
              my(%params) = @_;
              return $params{rec};
          };
    }
    else {
          my $return_class = $class->return_class(%params);
          $object_constructor = sub {
              my(%local_params) = @_;
              my $rec = $local_params{rec};

              my $row_policy;
              if ($policy_id_field && $rec->{$policy_id_field}) {
                  if ($policy && $policy->{policy_id} && $policy->{policy_id} == $rec->{$policy_id_field}) {
                      $row_policy = $policy;
                  }
                  else {
                      $row_policy = { policy_id => $rec->{$policy_id_field} };
                      if ($full_policy_populate) {
                          $row_policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $dbh, policy => $row_policy);
                      }
                  }
              }
              $class->map_fields_to_entity(rec => $rec, 
                                           dbh => $dbh,
                                           policy => $row_policy,
                                           logger => $logger,
                                           minimal_load => $minimal_load,
                                           default_rd => $default_rd,
                                           extra_fields => $extra_fields,
                                           extra_fields_hash => $extra_fields_hash,
                                           selected_fields_hash => $selected_fields_hash,
                                           call_flags => $call_flags,
                                           found_rows => $found_rows,
                                          ); 
              my $obj = $return_class->new(dbh => $dbh,
                                           logger => $logger,
                                           policy => $row_policy,
                                           object => $rec);
              $obj->load_mcp_fields();
              if ($obj && !$minimal_load) {
                  $obj->retrieve_nested_structures(fields => $fields,
                                                   extra_fields_hash => $extra_fields_hash,
                                                   selected_fields_hash => $selected_fields_hash,
                                                   nested_struct_fields => \@nested_struct_fields,
                                                   call_flags => $call_flags,);
              }
              if ($obj && $populate_diff_structs) {
                  $obj->populate_diff_structs(selected_fields_hash => $selected_fields_hash,
                                              call_flags => $call_flags,);
              }

              if ($obj) {
                  for my $field_name (keys %{$obj->virtual_fields}) {
                      my $virtual_value_func = $obj->virtual_fields->{$field_name};
                      $obj->$virtual_value_func(%params, field_name => $field_name);
                  }
              }

              return $obj;
          };
    }

    my $iterator = F5::ASMConfig::Entity::Iterator->new(sth => $sth,
                                                        total_record_count => $total_record_count,
                                                        object_constructor => $object_constructor,
                                                       );

    if (wantarray) {
        my @retval;
        while (my $rec = $iterator->get_next) {
            push(@retval, $rec);
        }
        return @retval;
    }
    else {
        return $iterator;
    }
}

#----------------------------------------------------------------------

sub return_class {
    my ($class) = @_;
    return $class;
}

#----------------------------------------------------------------------

sub get_additional_search_filters {
  my($class, %params) = @_;

  my $filter_field = $params{filter_field};
  my $filter_value = $params{filter_value};
  my $suppress_appending_field_name = $params{suppress_appending_field_name};

  my @search_filters;
  if (ref($filter_value)) {
    if (reftype($filter_value) eq "ARRAY") {
      push(@search_filters,
               $class->get_additional_search_filters(filter_field => $filter_field,
                                                     suppress_appending_field_name => 1,
                                                     filter_value => $filter_value->[0]));
    } 
    elsif (reftype($filter_value) eq "HASH") {

      for my $k (keys %$filter_value) {
  
        if (defined $filter_value->{$k}) {
          if (ref($filter_value->{$k})) {
            push(@search_filters,
                 $class->get_additional_search_filters(filter_field => "${filter_field}__$k",
                                                       filter_value => $filter_value->{$k}));
          }
          elsif (defined $filter_value->{$k}) {
            push(@search_filters,
                 {
                  field_name => $filter_field . ($suppress_appending_field_name ? "" : "__$k"),
                  field_value => $filter_value->{$k} });
          }
        }
      }
    }
  }

  return @search_filters;
}

#----------------------------------------------------------------------

sub get_complex_filter {
    my($class, %params) = @_;

    my $filter_field = $params{filter_field};
    my $field_value = $params{field_value};
    my $extra_fields = $params{extra_fields};
    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $where_clause;
    my @where_values;
    my $join_clause;
    my @join_values;
    
    my $staging_field = $class->fields->{flg_quarantined} ? 'flg_quarantined'
                                                          : 'in_staging'; 
    my $staging_db_field;
    if ($class->fields->{$staging_field}) {
        $staging_db_field = $class->fields->{$staging_field}->{db_field_name};
    }

    my $staging_subquery;
    my $tightening_subquery;
    if ($filter_field eq 'staging_or_tightening') {
        my @where_or_clauses;

        if ($class->fields->{$staging_field}) {
            push(@where_or_clauses, "main_table.$staging_db_field = ?");
            push(@where_values, 1);
        }

        if ($class->fields->{has_tightening}) {
            push(@where_or_clauses, 'main_table.flg_tightening = ?');
            push(@where_values, F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS());
        }
                 
        $where_clause = '(' . join(' OR ', @where_or_clauses) . ')';
    }

    if ($filter_field eq 'staging_or_tightening_past_period') {
        my $plc_dbname = 'PLC';
        my($staging_period) = $dbh->selectrow_array(<<EndQuery, undef, $policy->{policy_id});
SELECT quarantined_period 
  FROM $plc_dbname.PL_POLICY_GENERAL
 WHERE policy_id = ?
EndQuery
        my $staging_start_check_time = time - ($staging_period * 60*60*24);

        my @where_clauses;

        if ($class->fields->{has_tightening}) {
            my $tightening_subquery = $class->get_tightening_subquery(%params, policy => $policy);

            push(@where_clauses, <<EndQuery);
 (    main_table.flg_tightening = ? 
  AND main_table.last_tightening_time < ?
  AND $tightening_subquery->{db_field})
EndQuery
            push(@where_values, F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS(), $staging_start_check_time);
        }

        if ($class->fields->{$staging_field}) {
            my $staging_subquery = $class->get_staging_subquery(%params, policy => $policy);

            push(@where_clauses, <<EndQuery);
 (    main_table.$staging_db_field = ? 
  AND main_table.last_updated < ?
  AND $staging_subquery->{db_field})
EndQuery
            push(@where_values, 1, $staging_start_check_time);
        }
                 
        $where_clause = '(' . join(' OR ', @where_clauses) . ')';
    }

    if ($filter_field eq 'staging_or_tightening_has_suggestions') {
        my $plc_dbname = 'PLC';
        my @where_clauses;

        if ($class->fields->{has_tightening}) {
            my $tightening_subquery = $class->get_tightening_subquery(%params, 
                                                                      policy => $policy, 
                                                                      has_suggestions => 1,);
            push(@where_clauses, <<EndQuery);
 (    main_table.flg_tightening = ?
  AND $tightening_subquery->{db_field})
EndQuery
            push(@where_values, F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS());
        }

        if ($class->fields->{$staging_field}) {
            my $staging_subquery = $class->get_staging_subquery(%params, 
                                                                policy => $policy,
                                                                has_suggestions => 1,);
            push(@where_clauses, <<EndQuery);
 (    main_table.$staging_db_field = ?
  AND $staging_subquery->{db_field})
EndQuery
            push(@where_values, 1);
        }

        $where_clause = '(' . join(' OR ', @where_clauses) . ')';
    }

    if ($tightening_subquery) {
        #avoid duplicates in join
        if ( !($extra_fields && $extra_fields->{extra_has_tightening_suggestions} && $class->fields->{extra_has_tightening_suggestions})) {
            $join_clause .= $tightening_subquery->{query};
            push(@join_values, @{$tightening_subquery->{bind_values}});
        }
    }
    if ($staging_subquery) {
        #avoid duplicates in join
        if (! ($extra_fields && $extra_fields->{extra_has_suggestions} && $class->fields->{extra_has_suggestions})) {
            $join_clause .= $staging_subquery->{query};
            push(@join_values, @{$staging_subquery->{bind_values}});
        }
    }

    my $rv;
    if ($where_clause) {
        $rv->{where_clause} = $where_clause;
        $rv->{where_values} = \@where_values;
    }
    if ($join_clause) {
        $rv->{join_clause} = $join_clause;
        $rv->{join_values} = \@join_values;
    }

    if ($rv) {
        return $rv;
    }

    return;
}

#----------------------------------------------------------------------

sub get_next_staging_or_tightening_past_period_time {
    my ($class, %params) = @_;
    my $policy = $params{policy};
    my $db_table = $class->db_table;

    my (@fields, @where_and_clauses, @where_and_values, @where_or_clauses, @where_or_values);
    my ($join_clause, @join_values);
    if ($class->policy_id_field) {
        push (@where_and_clauses, "main_table." . $class->policy_id_field . " = ? ");
        push (@where_and_values, $policy->{policy_id});
    }

    if ($class->fields->{has_tightening}) {
        push @fields, "MIN(CASE WHEN (main_table.flg_tightening = 1) THEN main_table.last_tightening_time ELSE NULL END)";
        my $tightening_subquery = $class->get_tightening_subquery(%params);
        push (@where_or_clauses, <<EndQuery);
 (    main_table.flg_tightening = ?
  AND $tightening_subquery->{db_field})
EndQuery
        $join_clause .= $tightening_subquery->{query};
        push(@where_or_values, F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS());
        push(@join_values, @{$tightening_subquery->{bind_values}});
    }

    if ($class->fields->{in_staging}) {
        push @fields, "MIN(CASE WHEN (main_table.flg_is_staging = 1) THEN main_table.last_updated ELSE NULL END)";
        my $staging_subquery = $class->get_staging_subquery(%params);
        push (@where_or_clauses, <<EndQuery);
 (    main_table.flg_is_staging = ?
  AND $staging_subquery->{db_field})
EndQuery
        $join_clause .= $staging_subquery->{query};
        push(@where_or_values, 1);
        push(@join_values, @{$staging_subquery->{bind_values}});
    }

    return unless @fields;

    my $fields = join(",", @fields);
    if (@where_or_clauses) {
        push (@where_and_clauses, '(' . join(' OR ', @where_or_clauses) . ')');
        push (@where_and_values, @where_or_values);
    }
    my $where_clause .= '(' . join(' AND', @where_and_clauses) . ')';

    my @mins = $params{dbh}->selectrow_array(<<EndQuery, undef, @join_values, @where_and_values);
SELECT $fields
  FROM $db_table as main_table
  $join_clause
 WHERE $where_clause
EndQuery

    my $min_time;
    for my $min (@mins) {
        if ($min && (!$min_time || $min < $min_time)) {
            $min_time = $min;
        }
    }
    return $min_time;
}

#----------------------------------------------------------------------

sub get_tightening_subquery {
    my($class, %params) = @_;
    my $policy = $params{policy};
    my $want_suggestions = $params{has_suggestions} ? 1 : 0;

    my $plc_dbname = 'PLC';
    my $query = <<EndQuery;
LEFT JOIN (
       SELECT DISTINCT matched_wildcard_element_rest_uuid, 1 as matched
       FROM  PLC.PL_SUGGESTIONS
       WHERE policy_id = ? AND suggestion_status = ? AND element_type = ?
    ) tightening_suggestions
    ON (tightening_suggestions.matched_wildcard_element_rest_uuid = main_table.rest_uuid)
EndQuery

    return { query => $query,
             bind_values => [ $policy->{policy_id}, F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING(), $class->element_type],
             db_field => "CASE WHEN (tightening_suggestions.matched) THEN $want_suggestions ELSE NOT $want_suggestions END"};

}

#----------------------------------------------------------------------

sub get_staging_subquery {
    my($class, %params) = @_;
    my $policy = $params{policy};
    my $want_suggestions = $params{has_suggestions} ? 1 : 0;

    my $plc_dbname = 'PLC';
    my $db_table = $class->db_table;
    my $query = <<EndQuery;
LEFT JOIN (SELECT DISTINCT T.rest_uuid AS union_rest_uuids, 1 AS matched FROM $db_table AS T
                JOIN PLC.PL_SUGGESTIONS AS S ON (S.policy_id = ? AND S.suggestion_status = ? AND T.rest_uuid = S.element_rest_uuid AND S.element_type = ?)) suggestions
        ON (suggestions.union_rest_uuids = main_table.rest_uuid)
EndQuery

    return { query => $query,
             bind_values => [ $policy->{policy_id}, F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_PENDING(), $class->element_type],
             db_field => "CASE WHEN (suggestions.matched) THEN $want_suggestions ELSE NOT $want_suggestions END"};

}

#----------------------------------------------------------------------

sub get_updated_in_enforcement_readiness_period_subquery {
    my($class, %params) = @_;
    my $policy = $params{policy};

    my $staging_start_time = $class->get_staging_period_start_time(%params,
                                                                   policy => $policy,);
    my $query = <<EndQuery;
(main_table.last_updated > $staging_start_time)
   AS extra_updated_in_enforcement_readiness_period
EndQuery

    return $query;
}

#----------------------------------------------------------------------

sub get_learned_in_enforcement_readiness_period_subquery {
    my($class, %params) = @_;
    my $policy = $params{policy};

    my $staging_start_time = $class->get_staging_period_start_time(%params,
                                                                policy => $policy,
                                                                );
    my $query = <<EndQuery;
(IFNULL(
  main_table.last_tightening_time, main_table.last_updated
  ) > $staging_start_time)
  AS extra_learned_in_enforcement_readiness_period
EndQuery

    return $query;
}

#----------------------------------------------------------------------

sub get_staging_period_start_time {
    my($class, %params) = @_;
    my $policy = $params{policy};

    my $staging_period = $params{dbh}->selectrow_array(<<EndQuery, undef, $params{policy}->{policy_id});
SELECT quarantined_period
  FROM PLC.PL_POLICY_GENERAL
  WHERE policy_id = ?
EndQuery

    my $now = time();
    my $staging_start_time = $now - ($staging_period * 60*60*24);

    return $staging_start_time;
}

#----------------------------------------------------------------------

sub additional_field_mapping {
  return {};
}

#----------------------------------------------------------------------

sub additional_where_clauses {
    return;
}


sub get_retrieve_join_tables {
    my($class, %params) = @_;

    my $extra_fields = $params{extra_fields} ? { map { ( $_ => 1 ) } @{$params{extra_fields}}} : {};
    my $search_filters = $params{search_filters} ? { map { ( $_->{field_name} => 1 ) } @{$params{search_filters}}} : {};

    my $join_table = '';
    my $extern_db_tables = $class->extern_db_tables || {};
    for my $table_name (keys %$extern_db_tables) {
        my $prefix = lc $table_name;
        $prefix =~ s/\./__/g;

        $join_table .= <<EndQuery;
  LEFT JOIN $table_name $prefix
       ON (main_table.$extern_db_tables->{$table_name}->{local_join_field} = $prefix.$extern_db_tables->{$table_name}->{foreign_join_field})
EndQuery
    }

    if ( (($extra_fields && $extra_fields->{extra_has_suggestions})
           || ($search_filters && $search_filters->{extra_has_suggestions}))
           && $class->fields->{extra_has_suggestions} ) {
        my $staging_subquery = $class->get_staging_subquery(%params,
                                                             policy => $params{policy},
                                                             has_suggestions => 1,);
        my $query = $staging_subquery->{query};
        $query =~ s/\?/shift(@{$staging_subquery->{bind_values}})/gexms;
        $join_table .= $query;
    }

    if ( (($extra_fields && $extra_fields->{extra_has_tightening_suggestions})
          || ($search_filters && $search_filters->{extra_has_tightening_suggestions}))
          && $class->fields->{extra_has_tightening_suggestions} ) {
        my $tightening_subquery = $class->get_tightening_subquery(%params,
                                                                  policy => $params{policy},
                                                                  has_suggestions => 1,);
        my $query = $tightening_subquery->{query};
        $query =~ s/\?/shift(@{$tightening_subquery->{bind_values}})/gexms;
        $join_table .= $query;
    }

    return $join_table;
}

#----------------------------------------------------------------------

sub get_extern_db_prefix {
    my($class, %params) = @_;
    my $field_def = $params{field_def};
    if ($field_def->{extern_db_table}) {
        my $prefix = lc $field_def->{extern_db_table};
        $prefix =~ s/\./__/g;
        return ($prefix);
    }
    return;
}

#----------------------------------------------------------------------

sub extern_db_prefixes {
  return;
}
#----------------------------------------------------------------------

sub map_fields_to_entity {
    my($class, %params) = @_;
    my $rec = $params{rec};
    my $dbh = $params{dbh};
    my $fields = $class->fields;

    for my $field_name (keys %$fields) {
        my $field_def = $fields->{$field_name};
        if ($params{for_default_object} && $field_def->{exclude_from_default_object}) {
            delete $rec->{$field_name};
            next;
        }
        if ($field_def->{extern_db_field_name} && $field_def->{extern_db_table}) {
            $rec->{$field_name} = $rec->{$class->get_extern_db_prefix(field_def => $field_def) . '__' . $field_name};
            if (!defined($rec->{$field_name}) && defined($field_def->{default_value})) {
                $rec->{$field_name} = $field_def->{default_value};
            }
        }

        if (! $field_def->{no_mapping}->{'to_entity'}) {
            $rec->{$field_name} = $class->map_field_value(direction => 'to_entity',
                                                          field_name => $field_name,
                                                          field_value => $rec->{$field_name},
                                                          field_def => $field_def,
                                                          dbh => $dbh,
                                                          default_rd => $params{default_rd},
                                                          call_flags => $params{call_flags},                                                    
                                                         );
        }

        if ($field_def->{bitmask_field} && !$field_def->{db_field_name}) {
            $rec->{$field_name} = (($rec->{$field_def->{bitmask_field}} || 0) & $field_def->{bitmask_value} ? 1 : 0);
        }

        if ($field_def->{external_struct} && (! $params{minimal_load} || ! $field_def->{skip_in_minimal_load})) {
            if ($field_def->{db_field_name} && defined $rec->{$field_def->{db_field_name}}) {
                $rec->{$field_name} = $rec->{$field_def->{db_field_name}};
            }
            if ($field_def->{populate_from_join_row}) {
                $class->populate_from_join_row(%params,
                                               field_name => $field_name,);
            }
            elsif (defined $rec->{$field_name}) {
                my $external_call_flags = $params{call_flags};
                my $rest_selected_fields = $external_call_flags->{rest_selected_fields};
                if ($rest_selected_fields && %$rest_selected_fields) {
                    my $rest_field_name     = $field_def->{rest_field_name} || $class->rest_field_map->{$field_name};
                    my $expand_name         = $field_def->{rest_expand_name} || $rest_field_name;
                    my $nested_selected_fields = $class->get_selected_rest_fields(%params,
                                                    rest_selected_fields => $rest_selected_fields,
                                                    field_name => $field_name,
                                                    field_def => $field_def,
                                                    rest_field_name => $rest_field_name,
                                                    rest_expand_name => $expand_name,
                                                 );
                    $external_call_flags = {%$external_call_flags, rest_selected_fields => $nested_selected_fields};
                    if (!$nested_selected_fields) {
                        $external_call_flags->{selected_fields} = [];
                    }
                }
                if ($field_def->{entity_type} eq 'F5::ASMConfig::Entity::Policy') {
                    $rec->{$field_name} = F5::ASMConfig::Cache::get_cached_policy(policy => { policy_id => $rec->{$field_name} },
                                                                                  dbh => $dbh,
                                                                                  call_flags => $external_call_flags,
                                                                                  context => 'field');
                }
                else {
                    my $foreign_field = $field_def->{foreign_field} || $field_def->{entity_type}->id_field();
                    $rec->{$field_name} = $field_def->{entity_type}->new(dbh => $dbh,
                                                                         get_criteria => {$foreign_field => $rec->{$field_name}},
                                                                         minimal_load => 1,
                                                                         policy => $params{policy},
                                                                         logger => $params{logger},
                                                                         call_flags => $external_call_flags,
                                                                        );
                }
            }
        }

        if ($field_def->{populate_from_policy}) {
            $rec->{$field_name} = $params{policy}->{$field_name};
        }

        if ($field_def->{serialized_field} && $rec->{$field_def->{serialized_field}}) {
            eval {
                $rec->{$field_name} = thaw($rec->{$field_def->{serialized_field}});
            };
            if ($@) {
                $params{logger}->info("Could not thaw value for $field_name: $@") if $params{logger};
            }
        }
        if ($field_def->{superseded_fields} && !$params{for_default_object}) {
            for my $superseded_field (@{$field_def->{superseded_fields}}) {
                $rec->{$superseded_field} = $rec->{$field_name};
            }
        }
    }
}

#----------------------------------------------------------------------

my $FIELD_DEF_MAPPING = { to_entity => [qw(db_value_map db_format format_field_func entity_type const_value)],
                          to_db     => [qw(db_value_map db_format format_field_func entity_type const_value normalize_func validate_func is_affected_by_insensitive_flag is_affected_by_protocol_free_flag)],
                        };

sub requires_mapping {
    my ($class_or_self, %params) = @_;
    my $field_def = $params{field_def};
    my $direction = $params{direction};
    my $mapping_required = first { $field_def->{$_} } @{$FIELD_DEF_MAPPING->{$params{direction}}};
    return $mapping_required ? 1 : 0;
}

sub map_field_value {
    my($class_or_self, %params) = @_;
    my $field_name = $params{field_name};
    my $field_value = $params{field_value};
    my $field_def = $params{field_def};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $policy = $params{policy};
    my $dbh = $params{dbh};
    my $default_rd = defined $params{default_rd} ?  $params{default_rd} : 0;
    my $validate_value = defined $params{validate_value} ? $params{validate_value} : 1;

    if (!exists($field_def->{no_mapping}->{$params{direction}}) ) {
        my $mapping_required = grep { $field_def->{$_} } @{$FIELD_DEF_MAPPING->{$params{direction}}};
        $field_def->{no_mapping}->{$params{direction}} = $mapping_required ? 0 : 1;
        return $field_value unless $mapping_required;
    }

    if ($params{direction} eq 'to_entity') {
        if ($field_def->{db_value_map}) {
            my %db_to_entity_map = reverse %{$field_def->{db_value_map}};
            if ( defined $field_value) {
                $field_value =  $db_to_entity_map{$field_value};
            }
        }
        elsif ($field_def->{db_format} &&
               $field_def->{db_format} eq 'datetime' &&
               $field_def->{thrift_format} eq 'unixtime' &&
               $field_value  ) {

            if ($field_value =~ /[1-9]/) {
                my @time_parts = reverse split(/[- :\/]/, $field_value);
                # Month range in timelocal is 0..11
                $time_parts[4]--;
                $field_value = timelocal(@time_parts);
            }
            else {
                $field_value = 0;
            }
        }
        elsif ($field_def->{db_format} &&
               $field_def->{db_format} eq 'unixtime' &&
               $field_def->{thrift_format} eq 'datetime' &&
               $field_value ) {
            $field_value = $class_or_self->mysqltime($field_value);
        }
        elsif($field_def->{format_field_func}){
           my $format_function_name = $field_def->{format_field_func};
           $field_value = $class_or_self->$format_function_name( direction => 'to_entity',
                                                                 field_value  => $field_value,
                                                                 store_obj    => $store_obj,
                                                                 existing_obj => $existing_obj,
                                                                 field_def    => $field_def,
                                                                 dbh          => $dbh,
                                                                 default_rd  => $default_rd,
                                                                 policy => $policy,
                                                               );
        }
        elsif ($field_def->{entity_type} && defined($field_value) && !ref($field_value) &&
               (my $unique_id_field = $class_or_self->nested_unique_id_field(field_def => $field_def))) {
            $field_value = $field_def->{entity_type}->new(dbh => $dbh,
                                                          logger => $params{logger},
                                                          policy => $policy,
                                                          get_criteria => {$unique_id_field => $field_value});
                                                          
        }
        elsif (exists ($field_def->{const_value})) {
            $field_value = $field_def->{const_value};
        }
    }

    elsif ($params{direction} eq 'to_db') {

	if ($field_def->{is_affected_by_insensitive_flag} &&
	    defined $policy &&
	    $policy->{flg_is_case_insensitive}) {
	    $field_value =~ y/A-Z/a-z/;
	}

        if ($field_def->{is_affected_by_protocol_free_flag} &&
            defined $policy &&
            $policy->{flg_is_protocol_free}) {
            $field_value = 0;
        }


	if ($field_def->{db_value_map}) {
            if (exists $field_def->{db_value_map}->{$field_value}) {
                $field_value = $field_def->{db_value_map}->{$field_value};
            }
	} 
        elsif ($field_def->{db_format} &&
		 $field_def->{db_format} eq 'datetime' &&
		 $field_def->{thrift_format} eq 'unixtime' &&
		 $field_value  ) {
	    $field_value = $class_or_self->mysqltime($field_value);
        }
        elsif ($field_def->{db_format} &&
                 $field_def->{db_format} eq 'unixtime' &&
                 $field_def->{thrift_format} eq 'datetime' &&
                 $field_value  ){
            if ($field_value =~ /[1-9]/) {
                my @time_parts = reverse split(/[- :]/, $field_value);
                # Month range in timelocal is 0..11
                $time_parts[4]--;
                $field_value = timelocal(@time_parts);
            }
            else {
                $field_value = 0;
            }
	} 
        elsif (exists ($field_def->{const_value})) {
	    $field_value = $field_def->{const_value};
	}

        if ($field_def->{format_field_func}){
           my $format_function_name = $field_def->{format_field_func};
           $field_value = $class_or_self->$format_function_name( direction => 'to_db',
                                                                 field_value  => $field_value,
                                                                 store_obj    => $store_obj,
                                                                 existing_obj => $existing_obj,
                                                                 field_def    => $field_def,
                                                                 dbh          => $dbh,
                                                                 default_rd  => $default_rd,
                                                               );
        }
        elsif ($field_def->{entity_type} && defined($field_value) && ref($field_value) &&
               (my $unique_id_field = $class_or_self->nested_unique_id_field(field_def => $field_def))) {
            $field_value = $field_value->{$unique_id_field};
        }

        if ($validate_value) {
            if ($field_def->{normalize_func}) {
                my $rv = $class_or_self->validate_field(%params,
                                                        field_value  => $field_value,
                                                        validate_func => $field_def->{normalize_func},
                                                        normalize_error_message => $field_def->{normalize_error_message},
                                                       );
                $field_value = $rv->{normalized_value};
            }

            if ($field_def->{validate_func}) {
                $class_or_self->validate_field(%params,
                                               field_value  => $field_value,
                                              );
            }
	}

       if ($field_def->{rest_unique_record_part} || ($field_def->{unique_record_part} && !$field_def->{external_struct}) ){
            $store_obj->{$field_name} = $class_or_self->map_field_value(direction => 'to_entity',
                                                             field_name => $field_name,
                                                             field_value => $field_value,
                                                             field_def => $field_def,
                                                             dbh => $dbh,
                                                             default_rd => $params{default_rd},
                                                             call_flags => $params{call_flags},
                                                           );

       }
    }

    return $field_value;
}

#----------------------------------------------------------------------

#For now this will just return the hardcoded values in the Entities, but it should utilize the xml value mapping eventually
my %common_rest_types = ( last_updated => 'last_update_field',
                          staged_since => 'timestamp',
                          last_tightening_time => 'timestamp',
                          wildcard_sequence => 'int',
                          is_wildcard => 'is_wildcard',
                          in_staging  => 'boolean',
                          has_tightening => 'tightening_mode',
                          selfLink       => 'selfLink',
                          metachar       => 'hex',
                         );

sub field_def_to_rest_value_type {
    my (%params) = @_;

    my $field_def = $params{field_def};
    my $field = $params{field_name};
    my $type;
    if (defined($field_def->{rest_value_type})) {
        $type = $field_def->{rest_value_type};
    }
    elsif ($field_def->{rest_value_func}) {
        $type = '';
    }
    elsif ($field_def->{rest_value_map} || $field_def->{db_value_map}) {
        $type = 'value_map';
    }
    elsif ($common_rest_types{$field}) {
        $type = $common_rest_types{$field}
    }
    elsif ($field =~ /^(flg_|is_)/) {
        $type = 'boolean';
    }
    elsif ($field =~ /^(rule|max|min)(_|imum)/) {
        $type = 'int';
    }
    elsif ($field_def->{rest_value_db_map} || $field_def->{display_value_db_map}) {
        $type = 'value_db_map';
    }
    elsif ($field_def->{display_value_type}) {
        $type = $field_def->{display_value_type};
    }
    $field_def->{rest_value_type} = $type;    
    return $type;
}

#----------------------------------------------------------------------

sub field_value_types {
    my ($class) = @_;

    my %field_value_types = %common_rest_types;
    my $fields = $class->all_fields;
    for my $field (keys %$fields) {
        my $field_def = $fields->{$field};        
        $field_value_types{$field} = field_def_to_rest_value_type(field_def => $field_def,
                                                                  field_name => $field);
        if ($field_def->{fields}) {
            for my $nested_field (keys %{$field_def->{fields}}) {
                $field_value_types{"${field}__${nested_field}"} = field_def_to_rest_value_type(field_def => $field_def->{fields}->{$nested_field},
                                                                                               field_name => $nested_field);
            }
        }
    }
    return $class->install_accessor('field_value_types', \%field_value_types);
}
#----------------------------------------------------------------------
sub map_xml_field_value {
    my ($self, %params) = @_;
    my $field_name   = $params{field_name};
    my $field_def    = $params{field_def};
    my $direction    = $params{direction};
    my $dbh          = $params{dbh} || $self->{dbh};
    my $field_value  = $params{field_value};

    my $xml_field_name = $field_def->{xml_field_name} || $field_name;

    if ( $direction eq 'from_xml' ) {
        #XML::Twig problem for empty hash
        if (ref($field_value) eq 'HASH' && !scalar(keys %$field_value)) {
            $field_value = undef;
        }

        if (! defined $field_value) {
            if ($field_def->{xml_default_value}) {
                $field_value = $field_def->{xml_default_value};
            }
            else {
                return;
            }
        }
    }
    else {
        $field_value = defined $field_value ? $field_value :  $self->{$field_name};
    }
 
    if ($field_def->{xml_value_type} && $field_def->{xml_value_type} eq 'date') {
            return $self->convert_by_value_type_date(mode => $direction,
                                                     field_value => $field_value);
    }
    elsif ($field_def->{xml_value_type} && $field_def->{xml_value_type} eq 'timestamp') {
            return $self->convert_by_value_type_timestamp(mode => $direction,
                                                     field_value => $field_value);
    }
    elsif ($field_def->{xml_value_db_map} || $field_def->{display_value_db_map}) {
            return  $self->convert_by_value_db_map(mode => $direction,
                                                   map_def => ($field_def->{xml_value_db_map} || $field_def->{display_value_db_map}),
                                                   field_value => $field_value);
    }

    # Currently nested_struct is handled only in from_xml direction
    if ( $field_def->{nested_struct} && $direction eq 'from_xml' ) {
        my  @values = ();
        my  @records = ref($field_value) eq 'ARRAY' ? @{$field_value} : $field_value;
        for my $nested_xml_rec ( @records ) {
            my $nested_rec;
            for my $nested_field_key ( keys  %{$field_def->{fields}} ) {
                my $nested_field_xml_name = $field_def->{fields}->{$nested_field_key}->{xml_field_name} || $nested_field_key;
                $nested_rec->{$nested_field_key} =  $self->map_xml_field_value ( field_value => $nested_xml_rec->{$nested_field_xml_name},
                                                                                 field_name => $nested_field_key,
                                                                                 field_def => $field_def->{fields}->{$nested_field_key},
                                                                                 direction => 'from_xml',
                                                                                 dbh => $dbh,
                                                                               );
            }
            push @values, ($nested_rec); 
        }
        $field_value = \@values;
    }

    if( $field_def->{xml_display_value_map} ) {
        my %xml_display_value_map = %{$field_def->{xml_display_value_map}};
        if ($direction eq 'from_xml') {
             %xml_display_value_map = reverse %{$field_def->{xml_display_value_map}};
        }
        $field_value = defined $xml_display_value_map{$field_value} ? $xml_display_value_map{$field_value} : $field_value;
    }

    return $field_value;

}
#----------------------------------------------------------------------

my %rest_map = ( boolean     => { 1 => JSON::true,
                                  0 => JSON::false,},
                 boolean_with_undef => { 1 => JSON::true,
                                         0 => JSON::false,},
                 reverse_boolean => { 0 => JSON::true,
                                      1 => JSON::false,},
                 enabled_flg => { 1 => 'enabled',
                                  0 => 'disabled'},
                 allowed_flg => { 1 => 'allowed',
                                  0 => 'disallowed'},
                 tightening_mode => { 0 => 'never',
                                      1 => 'always',
                                      2 => 'selective',
                                      3 => 'compact',},
                 is_wildcard     => { 0 => 'explicit',
                                      1 => 'wildcard', },
                );


sub map_rest_field_value {
    my ($self, %params) = @_;
    my $field_name = $params{field_name};
    my $field_value = $params{field_value};
    my $field_def = $params{field_def};
    my $direction = $params{direction};
    my $dbh = $params{dbh} || $self->{dbh};
    my $policy = $params{policy};

    my $value_type = defined($field_def->{rest_value_type}) ? $field_def->{rest_value_type} 
                                                            : $self->field_value_types->{$field_name};

    if (my $func = $field_def->{rest_value_func}) {
         return $self->$func(%params,
                             value_type => $value_type);
    }

    my $rv = $field_value;

    if (defined($field_value) && ref $field_value && ( ref $field_value ne 'JSON::XS::Boolean' ) && ( ref($field_value) ne 'JSON::PP::Boolean' ) && (!$value_type || $value_type !~ /boolean/)) {
        my %nested_args;
        my $entity_type = $field_def->{entity_type};

        if ($field_def->{dynamic_entity_type}) {
            $entity_type = ref $field_value;
            $nested_args{skip_populate_diff_structs} = 1;
        }
        my $is_full_entity = ($entity_type && $entity_type =~ /ASMConfig::Entity::/);

        if (!$field_def->{rest_expand_name} && $field_def->{fields} && scalar keys %{$field_def->{fields}}) {
            my $nested_fields = $field_def->{fields};
            my %nested_field_map = map { ( $_ => ($nested_fields->{$_}->{rest_field_name} || '') ) } keys %$nested_fields;
            $nested_args{fields_def} = $nested_fields;
	    $nested_args{rest_field_map} = \%nested_field_map;
            $nested_args{nested_struct_call} = 1;
        }
        elsif ($is_full_entity) {
           $nested_args{fields_def} = {%{$entity_type->fields}, %{$entity_type->diff_structs} };
           $nested_args{rest_field_map} = $entity_type->rest_field_map();
           $nested_args{dynamic_entity_type} = $field_def->{dynamic_entity_type};
        }

        $rv = [];
        my @values;
        if (!$field_def->{nested_struct} || ($field_def->{map_rest_nested_as_single} && $direction eq 'from_rest')) {
            @values = ($field_value);
        }
        else {
             @values = ref($field_value) eq 'ARRAY' ? @$field_value : ($field_value);
        }

        for my $nested_value (@values) {
            my $mapped_value = $nested_value;
            if ($entity_type) {
                if (!ref($nested_value)) {
                    my $field_display_name = ($params{rest_field_name}) ? $params{rest_field_name} : $self->get_field_display_name_string(%params);
                    my $error_message = "Unexpected format for '$field_display_name'. Expected Object. Got scalar or array.";
                    die { error_message   => $error_message,
                          display_message => $error_message,
                          error_level     => 'info',
                          error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                        };
                }
                next if $nested_value->{flg_is_hidden};
                if ($is_full_entity) {
                    if ($direction eq "to_rest") {
                        if ($field_def->{cross_policy_entity}) {
                            $nested_args{policy} = $nested_value->{policy};
                        }
                        $mapped_value = $nested_value->recast(dbh => $dbh)->to_rest(%params,
                                                                                    %nested_args,
                                                                                    );
                    }
                    else {
                        $mapped_value = $entity_type->new(dbh => $self->{dbh},
                                                          logger => $self->{logger},
                                                          object => {})->from_rest(object => $nested_value,
                                                                                   policy => $policy,
                                                                                   skip_retrieve => $params{skip_retrieve});
                    }
                }
                else {
                    $mapped_value = $self->$direction(%params,
                                                      object => $nested_value,
                                                      %nested_args,
                                                     );
                }
            }
            elsif ($value_type) {
                $mapped_value = $self->map_rest_value_type(%params,
                                                           field_value => $nested_value,
                                                           value_type => $value_type);
            }
            push (@$rv, $mapped_value);
        }                
        if (!$field_def->{nested_struct} || ($field_def->{map_rest_nested_as_single} && $direction eq 'to_rest')) {
            $rv = $rv->[0];
        }
    }
    elsif ($value_type) {
        $rv = $self->map_rest_value_type(%params,
                                         value_type => $value_type);
    }

    return $rv;
    
}

#----------------------------------------------------------------------

sub get_rest_map {
    my ($self, %params) = @_;
    my $field_def = $params{field_def};
    my $value_type = $params{value_type};   
    my $dbh = $params{dbh} || $self->{dbh};
    my $ver = $params{ver} || $params{call_flags}->{rest_version} || F5::BigipVersionUtils::bigip_version();
    my $reload_db_cache = $params{reload_db_cache};

    my %map;

    if ($rest_map{$value_type}) {
        %map = %{$rest_map{$value_type}};
    }

    if ($value_type eq 'value_map' || $field_def->{rest_value_map}) {
        %map = %{$field_def->{rest_value_map} || $field_def->{db_value_map}};

        if ($ver && $field_def->{rest_version_value_map}) {
            if (! $field_def->{cache_version_values}->{$ver}) {
                my %version_entry_cache = ();
                my @version_values = sort
                                        { F5::BigipVersionUtils::compare_versions(version1 => $a->{version}, version2 => $b->{version}) }
                                        @{$field_def->{rest_version_value_map}};
                for my $value_version_set (@version_values) {
                    my $compare_res = F5::BigipVersionUtils::compare_versions(version1 => $ver, version2 => $value_version_set->{version});
                    if ( $compare_res >= 0 ) {
                        %version_entry_cache = (%version_entry_cache, %{$value_version_set->{values}})
                    }
                    else {
                        last;
                    }
                }

                $field_def->{cache_version_values}->{$ver} = \%version_entry_cache;
            }

            %map = (%map, %{$field_def->{cache_version_values}->{$ver}});
        }
    }

    if ($value_type eq 'value_db_map' || $value_type eq 'db_enum') {
        my $value_db_map = $field_def->{rest_value_db_map} || $field_def->{display_value_db_map};
        my $map_cache = $field_def->{rest_map_cache};
        if ( ! $map_cache || $reload_db_cache ) {
            my $db_map = $self->get_value_db_map_cache(dbh => $params{dbh},
                                                       value_db_map => $value_db_map);
            for my $map_key (keys %$db_map) {
                my $map_value = $db_map->{$map_key}->{$value_db_map->{value_field}};
                if ($value_type eq "db_enum") {
                    $map_value = lc $map_value;
                    $map_value =~ s/ /-/g;
                }
                $map_cache->{$map_key} = $map_value;
            }
            $field_def->{rest_map_cache} = $map_cache;
        }
        if ($map_cache) { 
            %map = (%$map_cache, %map);
        }
    }
    elsif ($field_def->{validate_foreign_value}) {
        my $values = F5::DbUtils::get_column_values(dbh => $params{dbh},
                                                    table_name => $field_def->{validate_foreign_value},
                                                    field_name => $field_def->{db_field_name});
        %map = map { ( $_ => $_ ) } @$values;
    }

    if ($value_type eq 'mcp_enum' && $field_def->{mcp_enum_class}) {
        my $packageName = 'F5::Mcp::' . $field_def->{mcp_enum_class};
        eval "require $packageName";
        no strict 'refs';
        %map = %{"${packageName}::tmsh_name"};
    }

    if ($field_def->{rest_additional_mapping_values}) {
        %map = (%map, %{$field_def->{rest_additional_mapping_values}});
    }

    if ($field_def->{rest_feature_value_map}) {
        for my $feature_values (@{$field_def->{rest_feature_value_map}}) {
            if ($self->is_feature_enabled(feature_name => $feature_values->{feature})) {
                %map = (%map, %{$feature_values->{values}});
            }
        }
    }

    return \%map;
}

#----------------------------------------------------------------------

sub map_rest_value_type {
    my ($self, %params) = @_;
    my $rest_field_name = $params{rest_field_name};
    my $field_name = $params{field_name};
    my $field_value = $params{field_value};
    my $field_def = $params{field_def};
    my $direction = $params{direction};
    my $value_type = $params{value_type};
    my $call_flags = $params{call_flags};
    my $policy_encoding = $params{policy_encoding};
    my $dbh = $params{dbh} || $self->{dbh};
    my $server_time = $call_flags ? $call_flags->{server_time} 
                                  : 0;

    my $rv = $field_value;

    if ($field_def->{significant_null_value} && $direction eq 'to_rest' && !defined($field_value)) {
        $field_value = '';
    }
    my $rest_tolerant_mapping = $field_def->{rest_tolerant_mapping};
    my %map = %{$self->get_rest_map(%params)};
    if (%map && $direction ne 'to_rest') {
        %map = reverse %map;
    }

    if (defined ($field_value) && !exists($map{$field_value}) && $field_def->{reload_db_cache_on_missing_entry}) {
        %map = %{$self->get_rest_map(%params, reload_db_cache => 1)};
        if (%map && $direction ne 'to_rest') {
            %map = reverse %map;
        }
    }

    if (%map && defined($field_value) && exists($map{$field_value}) && $direction ne 'to_rest')  {
        #No conversion needed
    }
    elsif (defined($rv)) {
        if ($value_type eq 'int') {
            if ($direction eq 'to_rest') {
                $rv ||= 0;
                $rv += 0;
            }
            elsif ($rv !~ /^-?\d+$/) {
                my $message = "Field value for $rest_field_name must be an integer.";
                die { error_message => $message,
                      display_message => $message,
                      error_level     => 'info',
                      error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                    };
            }
            elsif (defined $rv && $field_def->{rest_value_map}->{$rv}) {
                my $message = "Illegal field value for $field_name - $rv, must be an \"". $field_def->{rest_value_map}->{$rv}."\"";
                die { error_message => $message,
                      display_message => $message,
                      error_level     => 'info',
                      error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
                    };
            }

            $rest_tolerant_mapping = 1;
        }
        elsif ($value_type eq 'float') {
            if ($direction eq 'to_rest') {
                $rv += 0;
            }
            elsif ($rv !~ /^-?(?:\d+(?:\.\d*)?|\.\d+)$/) {
                my $message = "Field value for $rest_field_name must be a float.";
                die { error_message => $message,
                      display_message => $message,
                      error_level     => 'info',
                      error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                    };
            }
        }
        elsif ($value_type eq 'hex') {
            if ($direction eq 'to_rest') {
                $rv = sprintf('0x%x', $field_value);
            }
            else {
                if( $field_value =~ /^(0x)?[A-Fa-f0-9]{1,8}$/ ) {
                    $rv = hex $field_value;
                }
                else {
                    my $message = "Field value for $rest_field_name must be in hex.";
                    die { error_message => $message,
                          display_message => $message,
                          error_level     => 'info',
                          error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                        }; 
                }
            }
        }
        elsif ($value_type eq 'timestamp') {
            if ($direction eq 'to_rest') {
                $rv = F5::DateTimeUtils::convert_epoch_to_rfc3339($field_value, $server_time);
            }
            else {
                $rv = F5::DateTimeUtils::convert_rfc3339_to_epoch($field_value);
            }
        }
        elsif ($value_type eq 'local_timestamp') {
            if ($direction eq 'to_rest') {
                $rv = F5::DateTimeUtils::convert_epoch_to_datetime($field_value);
            }
            else {
                $rv = F5::DateTimeUtils::convert_datetime_to_epoch($field_value);
            }
        }
        elsif ($value_type eq 'datetime') {
            if ($direction eq 'to_rest') {
                $rv = F5::DateTimeUtils::convert_epoch_to_rfc3339(F5::DateTimeUtils::convert_datetime_to_epoch($field_value), $server_time);
            }
            else {
                $rv = F5::DateTimeUtils::convert_epoch_to_datetime(F5::DateTimeUtils::convert_rfc3339_to_epoch($field_value));
            }
        }
        elsif ($value_type eq 'last_update_field') {
            if ($direction eq 'to_rest') {
                $rv = $field_value * 1000000;
            }
            else {
                $rv = $field_value / 1000000;
            }
        }
        elsif ($value_type eq 'last_update_field_from_datetime') {
            if ($direction eq 'to_rest') {
                $rv = F5::DateTimeUtils::convert_datetime_to_epoch($field_value);
                $rv = $rv * 1000000;
            }
            else {
                $rv = $field_value / 1000000;
                $rv = F5::DateTimeUtils::convert_epoch_to_datetime($rv);
            }
        }
        elsif ($value_type eq 'date') {
            if ( $field_value !~ /^(\d\d\d\d)-(\d\d)-(\d\d)$/ || !defined(Date::Parse::str2time($field_value)) ){
                my $message = "Field value for $rest_field_name must be a valid date in the format of YYYY-MM-DD.\n";
                die { error_message => $message,
                      display_message => $message,
                      error_level     => 'info',
                      error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                    };
            }
        }
        elsif ($value_type eq 'native_encoded_string') {
            if ($policy_encoding && $policy_encoding ne 'utf-8') {
                if ($direction eq 'from_rest') {
                    $rv = $self->safe_encode(value => $field_value,
                                             from => 'utf-8',
                                             to => $policy_encoding);
                }
                else {
                    Encode::from_to($rv, $policy_encoding, 'utf-8');
                }
            }
        }
    }
    elsif ($value_type eq 'boolean_with_undef' && $direction eq 'to_rest') {
        #undef is 0
        $field_value = 0;
    }

    if (%map && defined($field_value)) {
        if (exists $map{$field_value}) {
            $rv = $map{$field_value};
        }
        elsif($rest_tolerant_mapping) {
            $rv = "$rv";  #Force stringification for enums
        }
        elsif($direction eq 'to_rest') {
            #Do not throw exceptions when converting to REST from our objects
            $rv = undef;
        }
        else {
            my $message = "Invalid value '$field_value' for field '$rest_field_name'.";
            die { error_message => $message,
                  display_message => $message,
                  error_level     => 'info',
                  error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                };
        }
    }

    if ($field_def->{significant_null_value} && $direction ne 'to_rest' && defined($rv) && !length($rv)) {
        $rv = undef;
    }

    return $rv;
}

#----------------------------------------------------------------------

sub get_selected_rest_fields {
    my ($class, %params) = @_;
    my $rest_selected_fields = $params{rest_selected_fields};
    my $rest_field_name = $params{rest_field_name};
    return {} unless ($rest_selected_fields && %$rest_selected_fields && $rest_field_name);

    my $field_name      = $params{field_name};
    my $field_def       = $params{field_def} || $class->fields->{$field_name};
    my $rest_group_name = $params{rest_group_name} || $field_def->{rest_group};
    my $expand_name     = $params{rest_expand_name} || $field_def->{rest_expand_name} || $rest_field_name;

    if ($rest_group_name) {
        if ( ! exists($rest_selected_fields->{$rest_group_name}) && ! exists($rest_selected_fields->{'*'})) {
            return;
        }
        $rest_selected_fields = $rest_selected_fields->{$rest_group_name} || $rest_selected_fields->{'*'};
     }

     if ($rest_selected_fields && ! exists($rest_selected_fields->{$rest_field_name}) && ! exists($rest_selected_fields->{$expand_name}) 
         && (! exists($rest_selected_fields->{'*'}) || $field_def->{extra_field}) ) {
         return;
     }

     return $rest_selected_fields->{$rest_field_name} || $rest_selected_fields->{$expand_name} || $rest_selected_fields->{'*'};
}

#----------------------------------------------------------------------

#Pre-Caching for rest external structs does not cover the following exceptions:
# 1) Collections that can have different policies per item (RequestLog) 
# 2) Fields that are dynamically typed (eg. some ConfigLog and Suggestion fields)

my $UNIQUE_VALUES_CACHING_THRESHOLD = 100;

sub prepare_rest_reference_cache {
    my ($class, %params) = @_;

    my $call_flags = $params{call_flags};
    my $external_struct_fields = $params{external_struct_fields} || $class->rest_external_struct_fields(); #Excludes dynamically types fields
    for my $external_struct_field ( @$external_struct_fields ) {
        my $expand_fields        = $params{expand_fields};
        my $rest_selected_fields = $params{rest_selected_fields};
        my ($field, $subfield)   = split (/__/, $external_struct_field);
        my $field_def            = $params{field_def} || $class->fields->{$field};
        my $rest_field_name      = $field_def->{rest_field_name} || $class->rest_field_map->{$field};
        my $expand_name          = $params{rest_expand_name} || $field_def->{rest_expand_name} || $rest_field_name;

        if ($rest_selected_fields && %$rest_selected_fields) {
            $rest_selected_fields = $class->get_selected_rest_fields(%params,
                                              rest_selected_fields => $rest_selected_fields,
                                              field_name => $field,
                                              field_def => $field_def,
                                              rest_field_name => $rest_field_name,
                                              rest_expand_name => $expand_name,
                                              );
            next unless $rest_selected_fields;
        }
        elsif ($field_def->{extra_field}) {
            next;
        }
        my $is_expanded  = 0;
        if ($expand_fields && (exists ($expand_fields->{$expand_name}) || $expand_fields->{$rest_field_name} || exists ($expand_fields->{'*'}))) {
            $expand_fields = $expand_fields->{$expand_name} || $expand_fields->{$rest_field_name} || $expand_fields->{'*'};
            $is_expanded  = 1;
        }

        my @objects = map { $field_def->{nested_struct} ? @{$_->{$field}}
                                                        : $_->{$field}   
                          } @{$params{objects}};

        if ($subfield) {
            $class->prepare_rest_reference_cache(%params,
                                                objects => \@objects,
                                                field_def => $field_def->{fields}->{$subfield},
                                                rest_selected_fields => $rest_selected_fields,
                                                expand_fields => $expand_fields,
                                                external_struct_fields => [ $subfield ],
                                               );                                                                  
        }
        else {                                         
            my $rest_group_name = $field_def->{rest_group};
            my $expand_name = $field_def->{rest_expand_name} || $rest_field_name;

            my $entity_type = $field_def->{rest_entity_type};
            my $field_name = $field;
            if ($field_def->{rest_external_struct_field_name}) {
                $field_name = $field_def->{rest_external_struct_field_name};
            }
            elsif ($field_def->{fields}) {
                ($field_name) = keys %{$field_def->{fields}};
            }

            #Shortcut if values are not expanded and do not need to be fetched
            my @unique_record_parts = @{$entity_type->unique_record_parts(%params)};
            if (!$field_def->{rest_confirm_references} &&
                ($field_name eq 'rest_uuid' || (scalar @unique_record_parts == 1 && $entity_type->all_fields->{$field_name}->{unique_record_part}))
                && !$is_expanded) {
                next;
            }
                
            my %unique_values = $field_def->{entity_type} ? map { ($_->{$field_name} => 1) } @objects 
                                                          : map { ($_ => 1) }  @objects;
            my @values = keys (%unique_values);
            next unless @values;
           
            if ((scalar @values > $UNIQUE_VALUES_CACHING_THRESHOLD) && !$field_def->{rest_confirm_references}) { #Never full cache a list that needs to confirm references (Suggestions, Requests)
                @values = (); #Cache the full collection
            }
            #Populate the cache with the needed values without using the return value.

            F5::ASMConfig::Cache::get_cached_rest_external_structs(entity_type => $entity_type,
                                                                   field_name => $field_name,
                                                                   field_value_list => \@values,
                                                                   dbh => $params{dbh},
                                                                   policy => $params{policy},
                                                                   rest_selected_fields => $rest_selected_fields,
                                                                   is_expanded => $is_expanded,
                                                                   call_flags => $call_flags,
                                                                  );

        }
    }
}

#----------------------------------------------------------------------

sub map_rest_external_struct {
    my ($self, %params) = @_;
    my $field_name = $params{field_name};
    my $field_value = $params{field_value};
    my $direction = $params{direction};
    my $field_def = $params{field_def};
    my $rest_entity_type = $field_def->{rest_entity_type};
    my $policy = $params{policy} || $self->{policy};
    my $dbh = $params{dbh} || $self->{dbh};

    return unless $field_value; 

    my @rvs;
    my @values = ($field_def->{nested_struct} && ref($field_value) eq 'ARRAY') ? @$field_value : ($field_value);
    if ($field_def->{rest_external_struct_field_name}) {
        $field_name = $field_def->{rest_external_struct_field_name};
    }
    elsif ($field_def->{fields}) {
        ($field_name) = keys %{$field_def->{fields}};
    }

    if ($field_def->{dynamic_entity_type}) {
        $rest_entity_type = ref($values[0]);
    }

    my $new_field_def =  {%$field_def,
                          rest_value_func => '',
                          nested_struct => '',
                          entity_type => $rest_entity_type,
                          external_struct => 1,
                          fields => '',
                         };
    
    if ($direction eq "to_rest") {
        if ($field_def->{entity_type} || $field_def->{dynamic_entity_type}) {
            @values = map { $_->{$field_name} } @values;
        }
        if (@values) {
            @values = @{F5::ASMConfig::Cache::get_cached_rest_external_structs(entity_type => $rest_entity_type,
                                                                               field_name => $field_name,
                                                                               field_value_list => \@values,
                                                                               dbh => $self->{dbh},
                                                                               policy => $policy,
                                                                               rest_selected_fields => $params{rest_selected_fields},
                                                                               populate_diff_structs => 1,
                                                                               is_expanded => 1,
                                                                               call_flags => $params{call_flags},
                                                                              )};
        }
    }
    elsif ($field_def->{entity_type}) {
        @values = map { $field_def->{entity_type}->new($_) } @values;
    }

    for my $value (@values) {
        my $rv = $self->map_rest_field_value(%params,
                                             field_def => $new_field_def,
                                             field_value => $value,
                                             dbh => $dbh,
                                            );
        if ($direction eq "to_rest" || $params{skip_retrieve}) {
            push (@rvs, $rv);
        }
        else {           
            if (defined($rv) && !exists($rv->{$field_name})) {
                $rv = $rest_entity_type->new(dbh => $self->{dbh},
                                             policy => $policy,
                                             get_criteria => $rv);
            }
            if (!defined($rv->{$field_name})) {
                my $rest_field_name = $field_def->{rest_field_name} || $self->rest_field_map->{$field_name};
                die "Invalid value for $rest_field_name ('" . encode_json($value) . "')\n";
            }

            if (!$field_def->{entity_type}) {
                $rv = $rv->{$field_name};
            }
            push (@rvs, $rv);
        }
    }
    return $field_def->{nested_struct} ? \@rvs : $rvs[0];
}

#----------------------------------------------------------------------

sub map_rest_external_field {
    my ($self, %params) = @_;
    my $field_value = $params{field_value};
    my $direction = $params{direction};
    my $field_def  = $params{field_def};
    my $entity_type  = $params{entity_type} || $params{field_def}->{entity_type};

    my $name = $field_def->{rest_external_field};

    return if !defined($field_value);
    my @objects = $field_def->{nested_struct} ? @{$field_value} 
                                              : ($field_value);
    my @rv;
    for my $obj (@objects) {
        my $rv;      
        if ($direction eq 'to_rest') {
            $rv = ref $obj ? $obj->{$name} 
                           : $obj;
            if (my $value_type = $field_def->{rest_value_type}) {
                $rv = $self->map_rest_value_type(%params,
                                                 field_value => $rv,
                                                 field_name => $name,
                                                 field_def => $field_def,
                                                 value_type => $value_type);
            }
        }
        else {
            if (my $value_type = $field_def->{rest_value_type}) {
                $obj = $self->map_rest_value_type(%params,
                                                  field_value => $obj,
                                                  field_name => $name,
                                                  field_def => $field_def,
                                                  value_type => $value_type);
            }
            my %args = (dbh => $self->{dbh},
                        policy => $self->{policy},
                        object => {$name => $obj});
            if (my $foreign_field = $field_def->{foreign_field} ) {
                $args{$foreign_field} = $self->{$foreign_field};
            }
            $rv =  $entity_type->new(%args);
        }
        push (@rv, $rv);
    }
    
    return $field_def->{nested_struct} ? \@rv 
                                       : $rv[0];
}


#----------------------------------------------------------------------

sub map_rest_defense_attributes {
    my ($self, %params) = @_;
    my $field_value = $params{field_value};
    my $direction = $params{direction};
    my $entity_type  = $params{entity_type};
    my $field_def  = $params{field_def};

    my $rv;
    return if !defined($field_value);

    my $defense_attributes = $self->get_defense_attributes();
    my $map_cache = $field_def->{rest_map_cache};
    if (!defined($map_cache)) {
        %{$map_cache} = map { ( $_ => F5::Utils::Rest::to_camel_case($defense_attributes->{$_}->{def_attrib_name}) ) } keys %$defense_attributes;
        $field_def->{rest_map_cache} = $map_cache;
        %{$field_def->{rest_map_cache_reverse}} = reverse %$map_cache;
    }

    if ($direction eq 'to_rest') {
        for my $obj (@{$field_value}) {
            my $value = $obj->{def_attrib_value};
            my $def_attrib_id = $obj->{def_attrib_id};
            if (defined($value)) {
                if ($defense_attributes->{$def_attrib_id}->{value_type} eq F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_BOOLEAN()) {
                    $value = $value ? JSON::true : JSON::false;
                }
                elsif ($defense_attributes->{$def_attrib_id}->{value_type} eq F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER()) {
                    $value = ($value || 0) + 0;
                }
            }
            if ((defined $obj->{flg_check_attrib} && $defense_attributes->{$def_attrib_id}->{value_type} == F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER())) {
                $value = $obj->{flg_check_attrib} ? $value : 'any';
            }

            $rv->{$map_cache->{$obj->{def_attrib_id}}} = $value;
        }
    }
    else {
        my $reverse_map = $field_def->{rest_map_cache_reverse};
        for my $name (keys %$field_value) {
            my $value = $field_value->{$name};
            my $def_attrib_id = $reverse_map->{$name};
            if (!$def_attrib_id) {
                die "Invalid defense attribute '$name'\n";
            }
            my $is_integer = $defense_attributes->{$def_attrib_id}->{value_type} == F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER(); 
            my $is_boolean = $defense_attributes->{$def_attrib_id}->{value_type} == F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_BOOLEAN();
            my %args = ( flg_check_attrib => ($value eq 'any' && $is_integer ? 0 : 1),
                         def_attrib_id => $def_attrib_id );


            if ($is_boolean &&  ref($value) && ( ref($value) eq 'JSON::XS::Boolean' || ref($value) eq 'JSON::PP::Boolean' ) ){
                $args{def_attrib_value} = $value ? 1 : 0;
            }
            elsif ( !$is_integer || $value ne 'any' ){
                $args{def_attrib_value} = $value;
            }

            push (@$rv, $params{field_def}->{entity_type}->new(%args));
        }
    }

    return $rv;
}

#----------------------------------------------------------------------

sub map_rest_learning_mode {
    my ($self, %params) = @_;
    my $field_value = $params{field_value};
    my $field_name = $params{field_name};
    my $direction = $params{direction};
    my $rv;

    if ($direction eq "to_rest" && ( defined $self->{flg_enabled} || defined $self->{is_pb_enabled} ) ) {
         $field_value = ($self->{flg_enabled} || $self->{is_pb_enabled})  ? F5::ASMConfig::LearningMode::LEARNING_MODE_AUTOMATIC()
                         : ($self->{flg_learning} || $self->{is_learning_enabled}) ? F5::ASMConfig::LearningMode::LEARNING_MODE_MANUAL()
                         : F5::ASMConfig::LearningMode::LEARNING_MODE_NONE();
    }
    return $self->map_rest_value_type(%params,
                                      field_value => $field_value,);
}

#----------------------------------------------------------------------

sub map_rest_policy_encoding {
    my ($self, %params) = @_;
    my $field_value = $params{field_value};
    my $field_name = $params{field_name};
    my $direction = $params{direction};
    my $rv;

    if ($direction eq "to_rest") {
        if (defined($self->{flg_encoding_selected})) {
            $rv = $self->{flg_encoding_selected} ? $self->{encoding}
                : (exists $self->{is_pb_enabled} && !$self->{is_pb_enabled}) ? 'undefined'
                : 'auto-detect';
        }
    }
    else {
        if ($field_name eq 'flg_encoding_selected') {
            $rv = ($field_value ne 'undefined' && $field_value ne 'auto-detect') ? 1 : 0;
        }
        elsif ($field_value eq 'undefined' || $field_value eq 'auto-detect') {
            $rv = 'utf-8';
        }
        else {
            $rv = $field_value;
        }
    }
    return $rv;
}

#----------------------------------------------------------------------

sub recast {
   my ($self, %params) = @_;
   my $rv = $self;
   my $dbh = $params{dbh} || $self->{dbh};
   my $do_retrieve = defined($params{do_retrieve}) ? $params{do_retrieve}
                                             : 1;
   if (my $recast_field = $self->recast_field) {
       my $new_class = $self->fields->{$recast_field}->{recast_map}->{$self->{$recast_field}};
       $rv = $new_class->new(dbh => $dbh,
                             logger => $self->{logger},
                             policy => $self->{policy},
                             get_criteria => ($do_retrieve ? $self : undef),
                             object => ($do_retrieve ? undef : $self),
                            );
   }
   return $rv;
}

#----------------------------------------------------------------------

sub uncast {
   my ($self, %params) = @_;
   my $rv = $self;
   my $dbh = $params{dbh} || $self->{dbh};
   if (my $uncast_field = $self->uncast_field) {
       my $new_class = $self->fields->{$uncast_field}->{uncast_class};
       my $uncast_value = defined($self->fields->{$uncast_field}->{uncast_value}) ? $self->fields->{$uncast_field}->{uncast_value}
                                                                                  : $self->fields->{$uncast_field}->{const_value};
       $rv = $new_class->new(dbh => $dbh,
                             logger => $self->{logger},
                             policy => $self->{policy},
                             object => {%$self, $uncast_field => $uncast_value},
                            );
   }
   return $rv;
}

#----------------------------------------------------------------------
my %non_data_fields = (dbh => 1,
                       policy => 1,
                       );

sub to_rest {
    my ($self, %params) = @_;
    my %rest_field_map = %{$params{rest_field_map} || $self->rest_field_map()};
    my $object = $params{object} || $self;
    my $policy = $params{policy} || $self->{policy};
    my $call_flags = $params{call_flags};
    my $dbh    = $params{dbh} || $self->{dbh};
    my $ver = $params{ver} || $params{call_flags}->{rest_version} || F5::BigipVersionUtils::bigip_version();
    $self->{dbh} = $dbh;
    my $rv;

    my $fields = $params{fields_def} || {%{$self->fields}, %{$self->diff_structs} };

    if (!$policy && $params{extract_policy_field} && $self->{$params{extract_policy_field}}) {
        if ($fields->{$params{extract_policy_field}} && $fields->{$params{extract_policy_field}}->{nested_struct}) {
            $policy = $self->{$params{extract_policy_field}}->[0]->policy();
        }
        else {
            $policy = $self->{$params{extract_policy_field}}->policy();
        }
    }

    unless ($params{skip_populate_diff_structs}) {
        $self->populate_diff_structs(include_rest_fields_only => 1, 
                                     policy => $policy,
                                     ver => $ver,
                                     rest_selected_fields => $params{rest_selected_fields},
                                     call_flags => $call_flags,);
    }

    my $entity_config = $self->entity_config();
    if(!$params{nested_struct_call} && $entity_config->{rest_paths} && $entity_config->{rest_paths}->[0]->{rest_path_properties}->{singleton}) {
	$rv = {};
    }

    my $selfLink;
    if ($params{to_rest_reference} || (defined($object->{rest_uuid}) && length($object->{rest_uuid}))) {
        $selfLink = $object->self_link(base_link => $params{base_link},
                                       policy => $policy,
                                       ver => $ver,
                                       call_flags => $call_flags,);
    }

    my $convert_to_utf8 = (F5::BigipVersionUtils::compare_versions(version1 => $UTF8_ENCODE_VERSION,
                                                                   version2 => $ver) <= 0);

    if ($object->{grouping_count} && !$params{is_in_declarative_export}) {
        $rv->{'__metadata'}->{'groupbyCount'} = $object->{grouping_count};
    }
    if ($call_flags->{aggregate_fields}) {
        for my $aggregate_field (keys %{$call_flags->{aggregate_fields}}) {
            for my $alias (keys %{$call_flags->{aggregate_fields}->{$aggregate_field}}) {
                if (defined $object->{$alias}) {
                    $rv->{'__metadata'}->{$alias} = $object->{$alias};
                }
            }
        }
    }

    my @rest_fields = grep { $params{nested_struct_call} || ($params{dynamic_entity_type} || $call_flags->{show_all_rest_fields}) || !(exists($entity_config->{fields}->{$_}) && $entity_config->{fields}->{$_}->{is_dependent})} (keys %$fields);
    while (my $field = shift @rest_fields) {

        my $field_def = $fields->{$field};
        next if $params{to_rest_reference} && !$field_def;

        my $value = $object->{$field};

        if ($field_def->{rest_composite_field} && $object->{$field_def->{rest_composite_field}}) {
            $value = $object->{$field_def->{rest_composite_field}}->{$field};
        }

        if (!$params{dynamic_entity_type} && !$call_flags->{show_all_rest_fields} && defined($value) && $entity_config->{dependent_fields}->{$field} &&
          ref ($entity_config->{dependent_fields}->{$field}->{$value})) {
            push (@rest_fields, @{$entity_config->{dependent_fields}->{$field}->{$value}});
        }

        next if $non_data_fields{$field};

        if (my $rest_field_name = $rest_field_map{$field}) {

            if ( !$call_flags->{show_all_rest_fields}
                  && (my $func = $field_def->{exclude_from_rest_func}) ) {

                if ( ref($object) =~ /::Entity::/ ) {
                    next if $object->$func(%params, field_def => $field_def, direction => 'to_rest');
                }
                else {
                    next if $self->$func(%params, object => $object, field_def => $field_def);
                }
            }
            my $rest_deprecated_func = $field_def->{rest_deprecated_func};
            if ($field_def->{exclude_from_to_rest} ||
                ($rest_deprecated_func && $self->$rest_deprecated_func(%params, field_def => $field_def)) ||
                ($field_def->{rest_deprecated_version} && 
                     !($call_flags->{show_rest_inheritance_fields} && ($field_def->{rest_composite_field} || $field_def->{rest_deprecated_inheritance_field})) &&
                     F5::BigipVersionUtils::compare_versions(version1 => $field_def->{rest_deprecated_version},
                                                             version2 => $ver) <= 0)
               ) {
                next;
            }
            if ($params{nested_struct_call} && $field eq 'is_inherited') {
                next unless $self->{is_inherited}; #Field is only relevent if parent entity is inherited.
            }

            my $nested_selected_fields = $params{rest_selected_fields};
            my $rest_group_name = $field_def->{rest_group};
            my $expand_name = $field_def->{rest_expand_name} || $rest_field_name;
            if ($nested_selected_fields && %$nested_selected_fields) {
                $nested_selected_fields = $self->get_selected_rest_fields(%params,
                                                                       field_name => $field,
                                                                       rest_field_name => $rest_field_name,
                                                                       rest_group_name => $rest_group_name,
                                                                       rest_expand_name => $expand_name,
                                                                       field_def => $field_def);
                next unless $nested_selected_fields;
            }
            my $expand_fields = $params{expand_fields};

            if ($field_def->{rest_expand_name} && (!$expand_fields || !(exists($expand_fields->{$expand_name}) || exists($expand_fields->{$rest_field_name}) || exists($expand_fields->{'*'})) ) ) {
                next unless defined($value);

                my @nested_values;
                if ($field_def->{nested_struct}) {
                    @nested_values = @$value;
                }
                else {
                    @nested_values = ($value);
                }

                my @rvs;
                if ($field_def->{rest_value_func} && $field_def->{rest_value_func} eq 'map_rest_external_struct') {
                     my $entity_type = $field_def->{rest_entity_type};
                     my $field_name = $field;
                     if ($field_def->{rest_external_struct_field_name}) {
                         $field_name = $field_def->{rest_external_struct_field_name};
                     }
                     elsif ($field_def->{fields}) {
                         ($field_name) = keys %{$field_def->{fields}};
                     }

                     if ($field_def->{dynamic_entity_type}) {
                         $entity_type = ref($nested_values[0]);
                     }

                     if ($field_def->{entity_type} || $field_def->{dynamic_entity_type}) {
                         @nested_values = map { $_->{$field_name} } @nested_values;
                     }

                     my @unique_record_parts = @{$entity_type->unique_record_parts(%params)};
                     if (!$field_def->{rest_confirm_references} &&
                         ($field_name eq 'rest_uuid' || (scalar @unique_record_parts == 1 && $entity_type->all_fields->{$field_name}->{unique_record_part}))) {
                         @nested_values = map { $entity_type->new(dbh => $dbh, policy => $policy, object => { $field_name => $_ }) } @nested_values;
                     }
                     elsif (scalar @nested_values) {
                         @nested_values = @{F5::ASMConfig::Cache::get_cached_rest_external_structs(entity_type => $entity_type,
                                                                                                   field_name => $field_name,
                                                                                                   field_value_list => \@nested_values,
                                                                                                   dbh => $self->{dbh},
                                                                                                   policy => $policy,
                                                                                                   call_flags => $call_flags,
                                                                                                   )};
                     }

                     if (!$field_def->{nested_struct} && !@nested_values) {
                         next;
                     }
                }

                for my $nested_value (@nested_values) {
                    if ($nested_value) {
                        my $nested_entity = $nested_value->recast(dbh => $dbh);
                        if (defined($call_flags->{verbose_references}) && !$call_flags->{verbose_references}) {
                            push (@rvs, { 'link' => $nested_entity->self_link(base_link => $params{base_link},
                                                                              call_flags => $call_flags,), })
                        }
                        else {
                            push (@rvs, $nested_entity->to_rest(base_link => $params{base_link},
                                                                skip_populate_diff_structs => 1,
                                                                ver => $ver,
                                                                dbh => $dbh,
                                                                fields_def => $nested_entity->rest_reference_fields(),
                                                                to_rest_reference => 1));
                        }
                    }
                }
           
                $rest_field_name = $field_def->{rest_expand_name};
                my $is_nested_struct = $field_def->{nested_struct} && !$field_def->{map_rest_nested_as_single};
                if ($rest_group_name) {
                    $rv->{$rest_group_name}->{$rest_field_name} = $is_nested_struct ? \@rvs : $rvs[0];
                }
                else {
                    $rv->{$rest_field_name} = $is_nested_struct ? \@rvs : $rvs[0];
                }

                next;
            }

            if ($expand_fields && (exists ($expand_fields->{$expand_name}) || $expand_fields->{$rest_field_name} || exists ($expand_fields->{'*'}))) {
                $expand_fields = $expand_fields->{$expand_name} || $expand_fields->{$rest_field_name} || $expand_fields->{'*'};

                if ($field_def->{rest_expand_name} && !$field_def->{rest_value_func} && ($field_def->{fields} || $field_def->{populate_from_join_row})) {
                    next unless defined $value;
                    #Need to flesh out full fields for REST as they won't be fetched by map_rest_external_struct
                    my $entity_type = $field_def->{rest_entity_type} || $field_def->{entity_type};
                    my @nested_values = $field_def->{nested_struct} ? @$value : ($value);
                    @nested_values = map { $entity_type->new(dbh => $dbh, policy => $policy, get_criteria => $_) } @nested_values;
                    $value = $field_def->{nested_struct} ? \@nested_values : $nested_values[0];
                }
            }
            my $policy_encoding;
            my $check_policy = $policy || $self->policy();
            if ($convert_to_utf8 && $check_policy && $check_policy->{encoding}) {
                $policy_encoding = $check_policy->{encoding};
            }

            my $rest_value = $self->map_rest_field_value( field_value => $value,
                                                          field_name => $field,
                                                          field_def => $field_def,
                                                          rest_field_name => $rest_field_name,
                                                          policy => $policy,
                                                          direction => 'to_rest',
                                                          rest_selected_fields => $nested_selected_fields,
                                                          expand_fields => $expand_fields,
                                                          rest_field_name => $rest_field_name,
                                                          base_link => $params{base_link},
                                                          call_flags => $call_flags,
                                                          skip_populate_diff_structs => $params{skip_populate_diff_structs},
                                                          dbh => $dbh,
                                                          policy_encoding => $policy_encoding,
                                                        );
            if (defined($rest_value)) {
                if ($field_def->{populate_kind}) {
                   $rest_value->{kind} = $value->get_kind(%params);
                   $rest_value->{selfLink} = $value->self_link(%params);
                }
                my $base64_encoded = 0;
                if ($field_def->{check_base64}) {
                    if ($convert_to_utf8) {
                        my $check_policy = $policy || $self->policy();
                        if ($check_policy && $check_policy->{encoding} && $check_policy->{encoding} ne 'utf-8') {
                            Encode::from_to($rest_value, $check_policy->{encoding}, 'utf-8');
                        }
                    }
                    else {                          
                        if ($rest_value =~ /[^\x00-\x7F]/) {
                            $base64_encoded = 1;
                            $rest_value = MIME::Base64::encode_base64($rest_value, '');
                        }
                    }
                }

		$rv ||= {};
                my $rv_field_level = $rv;
                if ($rest_group_name) {
                    $rv_field_level->{$rest_group_name} ||= {};
                    $rv_field_level = $rv_field_level->{$rest_group_name};
                }
                $rv_field_level->{$rest_field_name} = $rest_value;
                if ($base64_encoded) {
                    $rv_field_level->{$rest_field_name . 'Base64Encoded'} = $rest_map{boolean}->{$base64_encoded};
                }
                if ($call_flags->{show_rest_inheritance_fields} && $object->{has_parent} && ($field_def->{rest_composite_field} || $field_def->{rest_deprecated_inheritance_field})) {
                    my $is_inherited;
                    if ($field_def->{rest_composite_field}) {
                        $is_inherited = $object->{$field_def->{rest_composite_field}}->{is_inherited};
                    }
                    else {
                        my ($inheritance_object) = $field_def->{nested_struct} ? @$value : $value;
                        $is_inherited = $inheritance_object ? $inheritance_object->{is_inherited}
                                                            : $is_inherited = F5::ASMConfig::Entity::Section->new(dbh => $dbh,
                                                                                                                  policy => $policy,
                                                                                                                  get_criteria => {section_type => $field_def->{entity_type}->get_section_type}
                                                                                                                 )->is_inherited();
                    }
                    $rv->{inheritanceState} ||= {};
                    $rv_field_level = $rv->{inheritanceState};
                    if ($rest_group_name) {
                        $rv_field_level->{$rest_group_name} ||= {};
                        $rv_field_level = $rv_field_level->{$rest_group_name};
                    }
                    $rv_field_level->{$rest_field_name} = $is_inherited ? JSON::true : JSON::false;
                }
            }
        }
    }
    if ($selfLink) {
        my $link_key = $params{to_rest_reference} ? 'link' : 'selfLink';
        $rv->{$link_key} = $selfLink;
    }
    return $rv;
}

#----------------------------------------------------------------------

sub validate_field {
    my($class_or_self, %params) = @_;
    my $field_value = $params{field_value};
    my $field_def = $params{field_def};
    my $field_name = $params{field_name};

    # Validation should only be done if there is a store_obj
    my $function_name = $params{validate_func} || $field_def->{validate_func};
    return unless $function_name;

    if (my $exclude_from_validate_func = $field_def->{exclude_from_validate_func}) {
        if ($class_or_self->$exclude_from_validate_func(%params)) {
            return;
        }
    }

    my $rv = $class_or_self->$function_name(%params);

    if (! $rv->{completed_successfully} ) {
        if($params{call_flags}->{rest_call}) {
            $field_value = $class_or_self->map_rest_field_value( %params,
                                                                 field_value => $field_value,
                                                                 field_name => $field_name,
                                                                 field_def => $field_def,
                                                                 direction => 'to_rest',
                                                        );
        }
        my $field_display_name = $class_or_self->get_field_display_name_string(%params);

        my $error_message;
        if ($field_def->{validate_error_message}) {
            $error_message = $field_def->{validate_error_message};
            $error_message =~ s/\$FIELD_DISPLAY_NAME/$field_display_name/g;
            $error_message =~ s/\$FIELD_VALUE/$field_value/g;
            $error_message =~ s/\$ERROR_MSG/$rv->{error_message}/g;
        }
        else {
            $error_message = "Failed validating value '$field_value' for $field_display_name";
            if (defined $rv->{error_message} ) {
                $error_message .= ": $rv->{error_message}";
            }
        }
        my $display_message = $rv->{display_message} || $error_message;
        $error_message =~ s/\0/{NULL}/g;
        $display_message =~ s/\0/{NULL}/g;
        my $error_code = $field_def->{error_code} || $rv->{error_code} || F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT();
        die { error_message => $error_message,
              display_message => $display_message,
              error_level     => 'info',
              error_code      => $error_code,
            };
    }

    return $rv;
}

#----------------------------------------------------------------------
sub validate_defense_attributes{
    my($class_or_self, %params) = @_;
    my $defense_settings = $params{defense_settings};

    for my $defense (@$defense_settings) {
        if (defined $defense->{def_attrib_value} &&
            $defense->{value_type} eq F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_BOOLEAN) {
            if ( $defense->{def_attrib_value} ne '0' && $defense->{def_attrib_value} ne '1' ){
                die  { error_message => "'$defense->{def_attrib_name}' must be a boolean value (true/false)",
                       display_message => "'$defense->{def_attrib_name}' must be a boolean value (true/false)",
                       error_level     => 'info',
                       error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                      };
            }
        }

        if (defined $defense->{def_attrib_value} &&
            $defense->{value_type} eq F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER) {
            if ( $defense->{def_attrib_value} =~ /\D/g){
                die  { error_message => "'$defense->{def_attrib_name}' must be a positive integer",
                       display_message => "'$defense->{def_attrib_name}' must be a positive integer",
                       error_level     => 'info',
                       error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                      };
            }

            if ( defined $defense->{min_value}  &&  ($defense->{def_attrib_value} <  $defense->{min_value}) ) {
                my $err_message = "Failed validating value '$defense->{def_attrib_value}' for '$defense->{def_attrib_name}': The value is out of range (minimum value = $defense->{min_value}";
                die  { error_message => $err_message,
                       display_message => $err_message,
                       error_level     => 'info',
                       error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                      };
            }

             if ( defined $defense->{max_value}  &&  ($defense->{def_attrib_value} >  $defense->{max_value}) ) {
                my $err_message = "Failed validating value '$defense->{def_attrib_value}' for '$defense->{def_attrib_name}': The value is out of range (maximum value = $defense->{max_value})";
                die  { error_message => $err_message,
                       display_message => $err_message,
                       error_level     => 'info',
                       error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                      };
            }

        }
    }
}
#----------------------------------------------------------------------

sub validate_tightening {
    my ($self, %params) = @_;

    my $field_value = $params{field_value};
    my $field_def = $params{field_def};

    if ( defined ($field_value) && (! grep { $field_value eq $_ } @{$field_def->{validate_allowed_values}}) ) {
        my $field_display_value = $self->get_tightening_mode_display_values(%params);
        my $field_display_name = $self->get_field_display_name_string(%params);

        my $error_message =  "Illegal value. The legal values are " . join (',', map { "'$field_display_value->{$_}'" } @{$field_def->{validate_allowed_values}} ) ;
        my $display_message = "Failed validating value '$field_display_value->{$field_value}' for $field_display_name. $error_message";
        return {
                 completed_successfully => 0,
                 display_message => $display_message,
                 error_message => $error_message,
               };
    }
    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

sub validate_metachar_override {
    my ($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $entity_class = $params{entity_class};
    my $field_def = $params{field_def};

    my $policy = $params{policy};
    my $logger = $params{logger};
    my $dbh = $params{dbh};

    if ( ref($self) ){
        $logger = $self->{logger} if !$logger;
        $dbh = $self->{dbh} if !$dbh;
        $policy = $self->{policy} if !$policy;
        $entity_class = ref($self) if !$entity_class;
    }

    my $min_val = defined($field_def->{validate_min_val}) ? $field_def->{validate_min_val} 
                                                          : 1;

    # for HEADER type charset it's 255 regardless of encoding_type
    if ($entity_class && $entity_class eq "F5::ASMConfig::Entity::Charset" &&
        $self->get_new_value(%params, field_name => 'charset_type') == F5::ASMConfig::CharsetType::CHARSET_TYPE_HEADER()
       ) {

        return $self->general_validate(%params,
                                       field_def => { validate_min_val => $min_val,
                                                      validate_max_val => 255 });
    }

    my $encoding_type = F5::ASMConfig::EncodingType::ENCODING_TYPE_MULTI_BYTE();
    my $max_val = 127;
    if($policy) {
        $policy = F5::ASMConfig::Cache::get_cached_policy(%params, policy => $policy, dbh => $dbh);
        $encoding_type = F5::ASMConfig::Cache::get_cached_encoding_type(%params,
                                                                        encoding_name => $policy->{encoding},
                                                                        dbh => $dbh,
                                                                        logger => $logger,
                                                                       );
    }

    if($encoding_type != F5::ASMConfig::EncodingType::ENCODING_TYPE_MULTI_BYTE()) {
        $max_val = 255;
    }

    return $self->general_validate(%params,
                                   field_def => { validate_min_val => $min_val,
                                                  validate_max_val => $max_val });
}

#----------------------------------------------------------------------

sub check_crossfield_validations {
    my ($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};

    my $validations = $self->crossfield_validations() || [];
    my $fields = $self->all_fields();
    for my $validation (@$validations) {
        my $field_count = 0;
        for my $field (@{$validation->{fields}}) {
            my $field_def = $fields->{$field};
            my $value = $self->get_new_value( store_obj => $store_obj,
                                              existing_obj => $existing_obj,
                                              field_name => $field,);
            
            if ($value &&
                (!$field_def->{nested_struct} || @$value)) {
                $field_count++;
            }
        }

        my $error_message;
        my $error_code;
        if (defined($validation->{min_count}) && $validation->{min_count} > $field_count) {
            $error_message = $validation->{error_message} || "Failed Required Fields: Must have at least $validation->{min_count} of ";
            $error_code = F5::ASMConfig::ErrorCode::MISSING_FIELD();
        }
        elsif (defined($validation->{max_count}) && $validation->{max_count} < $field_count) {
            $error_message = $validation->{error_message} || "Failed Exclusive Fields: Must not have more than $validation->{max_count} of ";
            $error_code = F5::ASMConfig::ErrorCode::MUTUALLY_EXCLUSIVE_FIELDS();
        }

       if ($error_code) {
           if (!$validation->{error_message}) {
               my @field_names;
               for my $field (@{$validation->{fields}}) {
                   my $field_display_name = $self->get_field_display_name_string(field_name => $field, call_flags => $params{call_flags});
                   if($field_display_name ne "") {
                       push (@field_names, $field_display_name);
                   }
               }
               $error_message .= "(" . join(", ", @field_names) . ")";
           }
           die { error_message => $error_message,
                 display_message => $error_message,
                 error_level   => 'info',
                 error_code    => $error_code,
               };
       }
    }
}

#----------------------------------------------------------------------

sub has_feature_flag {
    my ($class_or_self, %params) = @_;
    my $flag = $params{feature_flag} || $params{field_def}->{feature_flag};
    return F5::ASM::COAPI::is_fflag_enabled($flag);
}

#----------------------------------------------------------------------

sub is_missing_feature_flag {
    my ($class_or_self, %params) = @_;
    return $class_or_self->has_feature_flag(%params) ? 0 : 1;
}

#----------------------------------------------------------------------

sub is_feature_enabled {
    my ($class_or_self, %params) = @_;
    if ($ENV{F5WAF_ENABLE_ALL_FEATURES}) {
        return 1;
    }
    my $feature_name = $params{feature_name} || $params{field_def}->{feature_name};
    return F5::BigipUtils::is_feature_enabled($feature_name);
}

#----------------------------------------------------------------------

sub is_feature_disabled {
    my ($class_or_self, %params) = @_;
    return $class_or_self->is_feature_enabled(%params) ? 0 : 1;
}

#----------------------------------------------------------------------

sub fill_external_struct_field {
    my ($class_or_self, %params) = @_;
 
    my $fields = $class_or_self->fields;
    my $dbh = (ref($class_or_self) && $class_or_self->{dbh}) ? $class_or_self->{dbh} : $params{dbh};
    my $field_name = $params{field_name};
    my $criteria = $params{criteria};
    my $entity_type = $fields->{$field_name}->{entity_type};

    my $get_criteria;
    if (defined $criteria) {
        $get_criteria = $criteria;
    }
    else {
        $get_criteria = $class_or_self->{$field_name};
    }

    return unless $get_criteria;
    my @unique_record_parts = @{$entity_type->unique_record_parts(get_criteria => $get_criteria)};
    my @unique_value_fields = @{$entity_type->unique_value_fields(get_criteria => $get_criteria)};
    my $id_field = $entity_type->id_field();

    if ( (!$id_field || !defined $get_criteria->{$id_field})
         && !(grep {$get_criteria->{$_}} @unique_value_fields) ) {

        for my $unique_record_part ( @unique_record_parts ) {
            return if not defined $get_criteria->{$unique_record_part};
        }
    }

    my $external_struct_entity = $entity_type->new(dbh => $dbh,
                                                   get_criteria => $get_criteria,
                                                   minimal_load => 1,
                                                   policy => $params{policy},
                                                   logger => $params{logger},
                                                   );
    if (ref($class_or_self)) {
        $class_or_self->{$field_name} = $external_struct_entity;
    } 
    
    return $external_struct_entity;
}

#----------------------------------------------------------------------

sub store_add {
  my($self, %params) = @_;

  my $fields = $self->fields;
  my $dbh = $self->{dbh};
  my $db_table = $self->db_table;
  my $id_field = $self->id_field;
  my $default_rd = defined $params{session}->{default_rd}? $params{session}->{default_rd} : 0;
  my $validate_value = defined $params{validate_value} ? $params{validate_value} : 1;
  my $call_flags = $params{call_flags} || {};

  my @modified_entities;
  my $db_rec = {};

  my %set_fields_to_null = $self->get_fields_to_set_null(%params);

  my $policy = F5::ASMConfig::Cache::get_cached_policy(policy => ($params{policy} || $self->{policy}), 
                                                       dbh => $dbh);

  for my $field_name (keys %$fields) {
      my $field_def = $fields->{$field_name};

      if ((my $serialized_field = $field_def->{serialized_field}) && defined($self->{$field_name})) {
          $self->{$serialized_field} = freeze($self->{$field_name});
      }

      if ($field_def->{superseded_fields}) {
          for my $superseded_field (@{$field_def->{superseded_fields}}) {
              if (defined($self->{$superseded_field})) {
                  $self->{$field_name} = $self->{$superseded_field};
                  last;
              }
          }
      }

      if (defined $self->{$field_name} &&
          $field_def->{is_affected_by_insensitive_flag} &&
          $policy->{flg_is_case_insensitive} && $self->{$field_name}) {
          $self->{$field_name} =~ y/A-Z/a-z/;
      }
      elsif (defined $self->{$field_name} &&
          $field_def->{is_affected_by_protocol_free_flag} &&
          $policy->{flg_is_protocol_free} && $self->{$field_name}) {
          $self->{$field_name} = 0;
      }
      elsif (!defined $self->{$field_name} && 
             ($field_def->{validate_non_empty} || $field_def->{is_name_field})) {
          $self->{$field_name} = '';  #Set to empty string and let standard validation sort out if that's legal
      }

      if (defined $self->{$field_name} && $field_def->{where_values}) {
          for my $where_field (keys %{$field_def->{where_values}}) {
              if (!defined($self->{$where_field}) || (defined($params{thrift_obj}) && !defined($params{thrift_obj}->{$where_field})) ) {
                 $self->{$where_field} = $field_def->{where_values}->{$where_field};                 
              }
              elsif ($self->{$where_field} ne $field_def->{where_values}->{$where_field}) {
                  my $field_display_name = $self->get_field_display_name_string(%params, field_name => $field_name);
                  my $where_field_display_name = $self->get_field_display_name_string(%params, field_name => $where_field);
                  my $error_message = "Invalid value for $where_field_display_name to set $field_display_name.";
                  die { error_message => $error_message,
                        display_message => $error_message,
                        error_level     => 'info',
                        error_code      => F5::ASMConfig::ErrorCode::MUTUALLY_EXCLUSIVE_FIELDS(),
                  };
              }
          }
      }
  }

  if (!$call_flags->{inherited_call}) {
      delete $self->{is_inherited};
  }

  for my $field_name (@{$self->non_nested_struct_fields}) {
      if ($fields->{$field_name}->{external_struct}) {
          $self->fill_external_struct_field(field_name => $field_name, policy => $policy, logger => $params{logger},);
      }
  }

  $self->instantiate_nested_struct_entities(store_obj => $self,
                                            policy => $policy,
                                            validate_value => $validate_value,
                                            existing_obj => {},
                                            call_flags => $call_flags,
                                           );

  my $before_store_rv = $self->before_store(%params,
                                            db_rec => $db_rec,
                                            store_mode => 'add',
                                            store_obj => $self,
                                            existing_obj => {},
                                            set_fields_to_null => \%set_fields_to_null,
                                            validate_value => $validate_value,
                                            policy => $policy);
  if (reftype($before_store_rv) && reftype($before_store_rv) eq 'HASH' && $before_store_rv->{modified_entities}) {
      push (@modified_entities, @{$before_store_rv->{modified_entities}});
  }

  if ($validate_value) {
      $self->check_crossfield_validations(existing_obj => {},
                                          store_obj    => $self,
                                          call_flags => $call_flags);
      $self->validate_inheritance(existing_obj => {},
                                  store_obj    => $self,
                                  policy => $policy,
                                  store_mode => 'add',
                                  validate_value => $validate_value,
                                  call_flags => $call_flags,);
  }

  for my $field_name (@{$self->non_nested_struct_fields}) {
    if (defined $self->{$field_name} ||
        $set_fields_to_null{$field_name} ||
        exists ($fields->{$field_name}->{const_value}) ||
        exists ($fields->{$field_name}->{default_value}) ) {

      if (!defined($self->{$field_name}) && !$set_fields_to_null{$field_name}) {
          $self->{$field_name} = exists ($fields->{$field_name}->{const_value}) ? $fields->{$field_name}->{const_value}
                                                                                : $fields->{$field_name}->{default_value};
      }

      if ( ( $fields->{$field_name}->{unique_record_part} || $fields->{$field_name}->{rest_unique_record_part} ) &&
             $fields->{$field_name}->{check_base64} &&
             $self->contains_control_chars(policy => $policy, value => $self->{$field_name}) ){
           my $field_display_name = $self->get_field_display_name_string(%params, field_name => $field_name);
           my $message = "The value for '$field_display_name' contains an illegal control character";
           die { error_message => $message,
                 display_message => $message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
               };
      }

      $self->set_field_in_db_rec(db_rec => $db_rec,
                                 field_name => $field_name,
                                 field_value => $self->{$field_name},
                                 existing_obj => {},
                                 store_obj => $self,
                                 default_rd => $default_rd,
                                 call_flags => $call_flags,
                                );
    }
  }

  $self->update_staging(db_rec => $db_rec,
                        store_obj => $self,
                        existing_obj => {},
                       );

  if ($self->policy_id_field) {
      if ($policy->{_failed_to_load}) {
          my $error_msg = "Missing policy";
          die {
                error_message => $error_msg,
                display_message => $error_msg,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::MISSING_POLICY(),
              };
      }
      $db_rec->{$self->policy_id_field} = $policy->{policy_id};
  }

  if (exists $fields->{rest_uuid}) {
      $self->set_rest_uuid(store_mode => 'add');
      $self->set_field_in_db_rec(db_rec => $db_rec,
                                 field_name => 'rest_uuid',
                                 field_value => $self->{'rest_uuid'},
                                 existing_obj => {},
                                 store_obj => $self,
                                 call_flags => $call_flags,
                               );
  }

  if (exists $fields->{created_by} && $params{session} && $params{session}->{user_name} 
      && $params{session}->{user_name} eq $POLICY_BUILDER_USER_NAME) {
      $self->{created_by} = 'PolicyBuilder';
      $self->set_field_in_db_rec(db_rec => $db_rec,
                                 field_name => 'created_by',
                                 field_value => $self->{'created_by'},
                                 existing_obj => {},
                                 store_obj => $self,
                                 call_flags => $call_flags,
                               );
  }

  if ($db_table) {
     if ($params{batched_insertor}) {
       $params{batched_insertor}->insert_row(table_name => $db_table,
                                             rec => $db_rec,
                                            );
       $self->{logger}->debug("Inserted $db_table (batch insert)") if $self->{logger};
     }
     else {
       my $insert_id;
       my $original_id = $self->get_id;
       my $rv = F5::DbUtils::insert_row(dbh => $dbh,
                                        table_name => $db_table,
                                        rec => $db_rec,
                                        last_insert_id => \$insert_id);
       if ($rv && $id_field) {
           if (! $original_id && $insert_id) {
               $self->set_id( $insert_id);
           }
           $self->{logger}->debug("Inserted $db_table - id = " . $self->get_id) if $self->{logger};
       }
     }
  }
  $self->after_store(%params,
                     db_rec => $db_rec,
                     store_mode => 'add',
                     store_obj => $self,
                     existing_obj => {},
                     validate_value => $validate_value,
                     policy => $policy,);

  # insert nested structs
  for my $field_name (@{$self->nested_struct_fields(%params)}) {
    $self->insert_nested_struct(nested_def => $fields->{$field_name},
                                rec => $self->{$field_name},
                                updated_entities => [ $self ],
                                policy => $policy,
                                nested_parent_field_name => $field_name,
                                call_flags => $call_flags,
                                batched_insertor => $params{batched_insertor},
                               );
  }

  $self->after_nested_struct_store(db_rec => $db_rec,
                                   store_mode => 'add',
                                   store_obj => $self,
                                   existing_obj => {},
                                   %params);

  my $diff_struct_store_rv = $self->diff_struct_store(store_obj => $self,
                             existing_obj => {},
                             store_mode => 'add',
                             %params);
  if ($diff_struct_store_rv->{modified_entities}) {
      push (@modified_entities, @{$diff_struct_store_rv->{modified_entities}});
  }

  my %extern_db_tables;
  for my $field_name (@{$self->extern_db_fields(%params)}) {
      $extern_db_tables{$fields->{$field_name}->{extern_db_table}}->{$field_name}->{field_value} = $self->{$field_name};
      $extern_db_tables{$fields->{$field_name}->{extern_db_table}}->{$field_name}->{db_field_name} = $fields->{$field_name}->{extern_db_field_name};
  }

   for my $extern_db_table (keys %extern_db_tables){
      my $foreign_key_value = $self->get_foreign_key_value(extern_db_table => $extern_db_table);
      $self->update_extern_db_table( table_name => $extern_db_table,
                                     rec => $extern_db_tables{$extern_db_table},
                                     store_mode => 'add',
                                     foreign_key_name => $self->extern_db_tables->{$extern_db_table}->{foreign_join_field},
                                     foreign_key_value => $foreign_key_value,
                                     store_obj => $self,
                                     existing_obj => {},
                                     validate_value => $validate_value,
                                     call_flags => $call_flags,
                                    );
  }
  $self->mcp_store(store_obj => $self,
                   existing_obj => {},
                   %params);

  $self->update_parent_staging(update_ids => [ $self->get_id ],
                               dbh => $dbh,);

  #Refresh full fields to be returned
  $self->retrieve(%params,
                  policy => $policy,
                  get_criteria => $self,
                  );


  if (! $params{suppress_events} && ! $call_flags->{suppress_events} && ! $self->suppress_events(%params, store_mode => 'add')) {
       $self->add_change_log_entry(dbh => $dbh,
                                   table => $db_table,
                                   action => 'add',
                                   session => $params{session});
  }

  my $retval = { db_change_made => 1,
                 modified_entities => \@modified_entities,
               };
  return $retval;
}

#----------------------------------------------------------------------

sub validate_inheritance {
    my ($self, %params) = @_;
    my $call_flags = $params{call_flags};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $policy = $params{policy};
    
    return unless $policy;

    my $entity_config = $self->entity_config();
    if ($policy->{policy_type} && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_PARENT()) {
        my $no_inheritance = $entity_config->{no_inheritance};
        my $error = 'Invalid entity for Parent Policy';
        if ($entity_config->{no_inheritance_conditions}) {
            for my $condition (@{$entity_config->{no_inheritance_conditions}}) {
                $no_inheritance ||= $filter_func{$condition->{field_comparison}}->(search_filter => $condition,
                                                                                   field_value => $self->get_new_value(%params, field_name => $condition->{field_name}));

                my $field_display_name = $self->get_field_display_name_string(%params, field_name => $condition->{field_name});
                $error = "Invalid value for '$field_display_name' in Parent Policy";
            }
        }
        if ($no_inheritance && $params{validate_value}) {
            die { error_message   => $error,
                  error_level     => 'info',
                  error_code      => F5::ASMConfig::ErrorCode::INVALID_ACTION_FOR_POLICY_TYPE(),
            };
        }
    }
    elsif($policy->{has_parent} && !$call_flags->{inherited_call} ) {
        if ($self->strict_inheritance() && $params{store_mode} eq 'add') {
            my $section = F5::ASMConfig::Entity::Section->new(dbh => $self->{dbh},
                                                               policy => $policy,
                                                             get_criteria => {section_type => $self->get_section_type()});
           if ( $section->is_inherited() ){
               my $error = "Cannot add elements to a strictly inherited collection.";
               die { error_message   => $error,
                     error_level     => 'info',
                     error_code      => F5::ASMConfig::ErrorCode::MODIFIED_INHERITED_VALUE(),
               };
           }
        }

        if($existing_obj->{is_inherited}) {
            if ($call_flags->{ignore_inherited_fields}) {
                my @store_obj_inherited_field = grep { defined($store_obj->{$_}) } @{$self->restricted_inheritance_fields()};
                delete @$store_obj{@store_obj_inherited_field};
            }
            else {
                my $rv  = $self->has_value_changed( existing_obj => $existing_obj,
                                                    store_obj => $store_obj,
                                                    check_fields => $self->restricted_inheritance_fields(),
                                                  );

                if ($rv){
                    my $error = "Inherited values may not be changed.";
                    die { error_message   => $error,
                          error_level     => 'info',
                          error_code      => F5::ASMConfig::ErrorCode::MODIFIED_INHERITED_VALUE(),
                    };
                }
            }
       }

    }
}

#----------------------------------------------------------------------

sub diff_struct_store {
  my ($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $policy = $self->policy;
  my $store_obj = $params{store_obj};
  my $store_mode = $params{store_mode};
  my $existing_obj = $params{existing_obj};
  my $call_flags = $params{call_flags} || {};
  my $diff_structs = $self->diff_structs;

  my $change_made = 0;
  my $self_clone;

  my $retval;
  my @modified_entities = ();

  for my $field_name (keys %$diff_structs) {
      next unless defined $store_obj->{$field_name};
      my $field_def = $diff_structs->{$field_name};

      my @store_entities = ();
      
      if (!%$existing_obj) {
          $existing_obj = ref($self)->new(dbh => $dbh,
                                          policy => $policy,
                                          get_criteria => $self);
      }

      $existing_obj->populate_diff_structs(policy => $policy,
                                           call_flags => $call_flags,);

      if ($store_obj->{$field_name}) {
          @store_entities = $field_def->{nested_struct} ? @{$store_obj->{$field_name}}
                                                        : ($store_obj->{$field_name});
          for my $entity (@store_entities) {
              $entity = $field_def->{entity_type}->new(dbh => $self->{dbh},
                                                       policy => $self->policy,
                                                       object => $entity);
              if ($field_def->{foreign_field}) {
                  if (!$self_clone) {
                      $self_clone = ref($self)->new(dbh => $dbh, object => {%$self});
                  }
                  $entity->{$field_def->{foreign_field}} = $self_clone;
                  if ($entity->id_field && !$entity->fields->{$entity->id_field}->{unique_record_part} && exists ($entity->{$entity->id_field})) {
                      delete $entity->{$entity->id_field};
                  }
              }
          }
      }

      if ($field_def->{read_only} || ($call_flags->{rest_call} && $field_def->{rest_read_only})){
          delete $store_obj->{$field_name};
          next;
      }

      my $field_changed;
      my $rv;

      if ($field_def->{nested_struct}) {
          my @existing_entities = ();

          if ($existing_obj->{$field_name}) {
              @existing_entities = @{$existing_obj->{$field_name}};
          }

          my $existing_nested_records = $self->hashify_nested_struct_list(nested_def => $field_def,
                                                                          nested_recs => \@existing_entities);
          my $store_nested_records = $self->hashify_nested_struct_list(nested_def => $field_def,
                                                                       nested_recs => \@store_entities);
          if (!$call_flags->{append_nested_struct_on_update}) {
              for my $key (keys %$existing_nested_records) {
                  #Remove entry if it does not exist and we are not in "remove nested" mode. OR if it does exist and we are in "remove nested mode"
                  if (!$store_nested_records->{$key} ^ ($call_flags->{remove_nested_struct_on_update} || 0)) {
                      if ( !$field_def->{nested_struct_update_only} ){
                          push (@modified_entities, {event_type => F5::ASMConfig::EventType::DELETE, existing_entity => $existing_nested_records->{$key}});
                          $rv = $existing_nested_records->{$key}->delete(call_flags => $call_flags);
                          if (ref($rv) && $rv->{deleted_entities}) {
                              for my $entity (@{$rv->{deleted_entities}}) {
                                  push (@modified_entities, {event_type => F5::ASMConfig::EventType::DELETE, existing_entity => $entity});
                              }
                          }
                          $field_changed = 1;
                      }
                  }
              }
          }

          if (!$call_flags->{remove_nested_struct_on_update}) {
              for my $key (keys %$store_nested_records) {
                  if (!$existing_nested_records->{$key}) {
                      if (  $field_def->{nested_struct_update_only} ) {
                          die "Invalid value for $field_def->{rest_field_name}\n";
                      }
                      else {
                          push (@modified_entities, {event_type => F5::ASMConfig::EventType::ADD, entity => $store_nested_records->{$key}, existing_entity => $store_nested_records->{$key}});
                          $rv = $store_nested_records->{$key}->store_add(call_flags => $call_flags);
                      }
                  }
                  else {
                      $rv = $existing_nested_records->{$key}->store_update(store_obj => $store_nested_records->{$key},
                                                                           call_flags => $call_flags);
                      my $value_changed = $rv->{db_change_made};
                      if($value_changed) {
                          push (@modified_entities, {event_type => F5::ASMConfig::EventType::UPDATE, entity => $store_nested_records->{$key}, existing_entity => $existing_nested_records->{$key}});
                      }
                  }
                  my $sub_entity_changed = $rv->{modified_entities} && @{$rv->{modified_entities}};
                  if($sub_entity_changed) {
                      push (@modified_entities, @{$rv->{modified_entities}});
                  }
                  $field_changed = $rv->{db_change_made} || $sub_entity_changed || $field_changed;
              }
          }
      }
      else {
         my ($store_entity) = @store_entities; 
         if  ($existing_obj->{$field_name}) {
             $rv = $existing_obj->{$field_name}->store_update(store_obj => $store_entity,
                                                              call_flags => $call_flags);
             if ($rv->{db_change_made}) {
                 push (@modified_entities, {event_type => F5::ASMConfig::EventType::UPDATE, entity => $store_entity, existing_entity => $existing_obj->{$field_name}});
             }
         }
         else {
             push (@modified_entities, {event_type => F5::ASMConfig::EventType::ADD, entity => $store_entity, existing_entity => $store_entity});
             $rv = $store_entity->store_add(call_flags => $call_flags);
         }
         my $sub_entity_changed = $rv->{modified_entities} && @{$rv->{modified_entities}};
         if ($sub_entity_changed) {
             push (@modified_entities, @{$rv->{modified_entities}});
         }
         $field_changed = $rv->{db_change_made} || $sub_entity_changed;
      }

      delete $store_obj->{$field_name};
      if ($field_changed) {
          delete $existing_obj->{$field_name};
          $change_made = 1;
      }
  }
  
  $self->populate_diff_structs(policy => $policy);
 
  return {db_change_made => $change_made,
          modified_entities => \@modified_entities };
}

#----------------------------------------------------------------------

sub store_update {
  my($self, %params) = @_;

  my $fields = $self->fields;
  my $dbh = $self->{dbh};
  my $db_table = $self->db_table;
  my $store_obj = $params{store_obj};
  my $call_flags = $params{call_flags} || {};
  my $default_rd = defined $params{session}->{default_rd}? $params{session}->{default_rd} : 0;
  my $validate_value = defined $params{validate_value} ? $params{validate_value} :  1;

  $self->store_virtual_fields(%params);

  my @modified_entities;
  my $db_rec = {};

  my %set_fields_to_null = $self->get_fields_to_set_null(%params);

  my $policy = F5::ASMConfig::Cache::get_cached_policy(policy => ($params{policy} || $self->{policy}), 
                                                       dbh => $dbh);

  for my $field_name (keys %$fields) {
      my $field_def = $fields->{$field_name};

      if ((my $serialized_field = $field_def->{serialized_field}) && defined($store_obj->{$field_name})) {
          $store_obj->{$serialized_field} = freeze($store_obj->{$field_name});
      }

     if ($field_def->{superseded_fields} && !defined($store_obj->{$field_name})) {
          for my $superseded_field (@{$field_def->{superseded_fields}}) {
              if (defined($store_obj->{$superseded_field})) {
                  $store_obj->{$field_name} = $store_obj->{$superseded_field};
                  last;
              }
          }
      }

      if (defined $self->{$field_name} &&
          $field_def->{is_affected_by_insensitive_flag} &&
          $policy->{flg_is_case_insensitive} && $self->{$field_name} &&
          defined $store_obj->{$field_name}) {
          $store_obj->{$field_name} =~ y/A-Z/a-z/;
      }
      elsif (defined $self->{$field_name} &&
          $field_def->{is_affected_by_protocol_free_flag} &&
          $policy->{flg_is_protocol_free} && 
          defined $store_obj->{$field_name}) {
          $store_obj->{$field_name} = 0;
      }
  }

  if (!$call_flags->{inherited_call}) {
      delete $store_obj->{is_inherited};
  }

  for my $field_name (@{$self->non_nested_struct_fields}) {
      if ($fields->{$field_name}->{external_struct}) {
          if ($self->has_value_changed( existing_obj => $self, store_obj => $store_obj, check_fields => $field_name ) && 
              !$fields->{$field_name}->{read_only} && !($call_flags->{rest_call} && $fields->{$field_name}->{rest_read_only})) {

              $store_obj->{$field_name} = ref($self)->fill_external_struct_field(criteria => $store_obj->{$field_name},
                                                                                 field_name => $field_name,
                                                                                 policy => $policy,
                                                                                 logger => $params{logger},
                                                                                 dbh => $dbh,
                                                                                );
          }
          else {
              $store_obj->{$field_name} = $self->{$field_name};
          }
      }
      elsif ( ( $fields->{$field_name}->{unique_record_part} || $fields->{$field_name}->{rest_unique_record_part} ) &&
                $fields->{$field_name}->{check_base64} &&
                 $self->has_value_changed( existing_obj => $self, store_obj => $store_obj, check_fields => $field_name )){
         if ( $self->contains_control_chars(policy => $policy, value => $store_obj->{$field_name}) ){
           my $field_display_name = $self->get_field_display_name_string(%params, field_name => $field_name);
           my $message = "The value for '$field_display_name' contains an illegal control character";
           die { error_message => $message,
                 display_message => $message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
               };
         }
      }      
  }

  $self->instantiate_nested_struct_entities(store_obj => $store_obj,
                                            policy => $policy,
                                            validate_value => $validate_value,
                                            existing_obj => $self,
                                            call_flags => $call_flags,
                                           );

  my $before_store_rv = $self->before_store(%params,
                                            db_rec => $db_rec,
                                            store_mode => 'update',
                                            existing_obj => $self,
                                            set_fields_to_null => \%set_fields_to_null,
                                            validate_value => $validate_value,
                                            policy => $policy);
  if (reftype($before_store_rv) && reftype($before_store_rv) eq 'HASH' && $before_store_rv->{modified_entities}) {
      push (@modified_entities, @{$before_store_rv->{modified_entities}});
  }

  if ($validate_value) {
      $self->check_crossfield_validations(existing_obj => $self,
                                          store_obj    => $store_obj,
                                          call_flags => $call_flags);
      $self->validate_inheritance(existing_obj => $self,
                                  store_obj    => $store_obj,
                                  policy => $policy,
                                  store_mode => 'update',
                                  validate_value => $validate_value,
                                  call_flags => $params{call_flags},);
  }

  # Check non-nested fields for changes
  my %extern_db_tables;
  for my $field_name (@{$self->non_nested_struct_fields}) {
    if (defined $store_obj->{$field_name} || $set_fields_to_null{$field_name}) {
      if (defined $self->{$field_name} && defined $store_obj->{$field_name} && !$set_fields_to_null{$field_name} && $self->{$field_name} eq $store_obj->{$field_name}) {
        # No change to value -- discard updated value
        delete $store_obj->{$field_name};
      }
      elsif ($fields->{$field_name}->{external_struct} && 
             !( $self->has_value_changed( existing_obj => $self, store_obj => $store_obj, check_fields => $field_name)) &&
             ! $set_fields_to_null{$field_name}) {

        delete $store_obj->{$field_name};
      }
      elsif (!$call_flags->{force_update_read_only} && $call_flags->{rest_call} && !$params{for_populate_uuids} &&
             ($field_name eq 'rest_uuid' || $fields->{$field_name}->{unique_record_part} || $fields->{$field_name}->{rest_unique_record_part} || $fields->{$field_name}->{read_only} || $fields->{$field_name}->{rest_read_only})) {
        delete $store_obj->{$field_name};
      }
      elsif ($fields->{$field_name}->{extern_db_updatable} &&
             $self->has_value_changed( existing_obj => $self, store_obj => $store_obj, check_fields => $field_name)){
        $extern_db_tables{$fields->{$field_name}->{extern_db_table}}->{$field_name}->{field_value} = $store_obj->{$field_name};
        $extern_db_tables{$fields->{$field_name}->{extern_db_table}}->{$field_name}->{db_field_name} = $fields->{$field_name}->{extern_db_field_name};
      }
      else {
        $self->set_field_in_db_rec(db_rec => $db_rec,
                                   field_name => $field_name,
                                   field_value => $store_obj->{$field_name},
                                   existing_obj => $self,
                                   store_obj    => $store_obj,
                                   default_rd => $default_rd,
                                   validate_value => $validate_value,
                                   call_flags => $params{call_flags},
                                  );
      }
    }
  }


  if (exists $fields->{rest_uuid} &&
        ( $params{for_populate_uuids} ||
         (!$call_flags->{rest_call} && $self->has_value_changed( existing_obj => $self, store_obj => $store_obj, check_fields => $self->rest_unique_record_parts(get_criteria => $self))) )
     ) {
      $self->set_rest_uuid(object => $store_obj,
                           store_mode => 'update');
      if (defined $store_obj->{rest_uuid}) {
          $self->set_field_in_db_rec(db_rec => $db_rec,
                                     field_name => 'rest_uuid',
                                     field_value => $store_obj->{'rest_uuid'},
                                     existing_obj => $self,
                                     store_obj => $store_obj,
                                     call_flags => $params{call_flags},
                                   );
      }
  }

  my $db_change_made = 0;
  for my $extern_db_table (keys %extern_db_tables){
      my $foreign_key_value = $self->get_foreign_key_value(extern_db_table => $extern_db_table);
      $self->update_extern_db_table( table_name => $extern_db_table,
                                     rec => $extern_db_tables{$extern_db_table},
                                     store_mode => 'update',
                                     foreign_key_name => $self->extern_db_tables->{$extern_db_table}->{foreign_join_field},
                                     foreign_key_value => $foreign_key_value,
                                     existing_obj => $self,
                                     store_obj    => $store_obj, 
                                     validate_value => $validate_value,
                                     call_flags => $params{call_flags},
                                    );
      $db_change_made = 1;
  }

  my %save_nested;
  for my $nested_struct_field (@{$self->nested_struct_fields(%params)}) {
      my $nested_def = $fields->{$nested_struct_field};
      if ($call_flags->{disinherit} && $nested_def->{fields} && $nested_def->{fields}->{is_inherited} && $self->{$nested_struct_field}) {
          my @disinherited_values = map { {%$_, is_inherited => 0} } grep { $_->{is_inherited} } @{$self->{$nested_struct_field}};
          $store_obj->{$nested_struct_field} = \@disinherited_values;
      }
      $save_nested{$nested_struct_field} = $store_obj->{$nested_struct_field};
  }

  my $rv = $self->update_nested_structs(db_rec => $db_rec,
                                        call_flags => $call_flags,
                                        store_obj => $store_obj,
                                        policy => $policy,
                                        validate_value => $validate_value,
                                       );
  $db_change_made ||= $rv;

  for my $nested_struct_field (@{$self->nested_struct_fields(%params)}) {
    unless ($store_obj->{$nested_struct_field}) { delete $save_nested{$nested_struct_field}; }
  }

  $self->perform_nested_struct_db_actions(store_obj => \%save_nested, 
                                          call_flags => $call_flags, 
                                          updated_entities => [ $self ],
                                          dbh => $dbh,
                                          policy => $policy,
                                         ) if $db_change_made;

  # Base record is updated after nested and diff  structures to update staging if only a nested change was made
  my $diff_struct_store_rv = $self->diff_struct_store(store_obj => $store_obj,
                                                      existing_obj => $self,
                                                      store_mode => 'update',
                                                      %params);
  if ($diff_struct_store_rv->{modified_entities}) {
      push (@modified_entities, @{$diff_struct_store_rv->{modified_entities}});
  }

  $db_change_made ||= $diff_struct_store_rv->{db_change_made};

  if (%$db_rec || $db_change_made) {
    $self->update_staging(db_rec => $db_rec,
                          store_obj => $store_obj,
                          existing_obj => $self,
                          for_populate_uuids => $params{for_populate_uuids},
                         );
    $self->update_parent_staging(update_ids => [ $self->get_id ],
                                 dbh => $dbh,
                                 validate_value => $validate_value,);
  }

  if (%$db_rec) {
      my %where_fields;
      if ($self->get_id) {
          $where_fields{$self->id_db_field} = $self->get_id;
          $self->{logger}->debug("Updating $db_table - id = " . $self->get_id) if $self->{logger};
      }
      else {
          for my $field (grep { defined($self->{$_}) } keys %$fields) {
              if ($fields->{$field}->{unique_record_part}) {
                  if ($fields->{$field}->{no_mapping}->{'to_db'}) {
                     $where_fields{$fields->{$field}->{db_field_name}} = $self->{$field};
                  }
                  else {
                      $where_fields{$fields->{$field}->{db_field_name}} = $self->map_field_value(direction => 'to_db',
                                                                                                field_name => $field,
                                                                                                 field_def => $fields->{$field},
                                                                                                 field_value => $self->{$field},
                                                                                                 store_obj    => $store_obj,
                                                                                                 existing_obj => $self,
                                                                                                 dbh => $dbh,
                                                                                                 default_rd => $params{default_rd},
                                                                                                 validate_value => $validate_value,
                                                                                                 call_flags => $params{call_flags},
                                                                                                );
                  }
              }
          }

          if ($self->policy_id_field) {
              $where_fields{$self->policy_id_field} = $policy->{policy_id};
          }
      }

      my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                       table_name => $db_table,
                                       rec => $db_rec,
                                       where_fields => \%where_fields);
      $db_change_made ||= $rv;
  }

  my $after_store_rv = $self->after_store(%params,
                                          db_rec => $db_rec,
                                          store_mode => 'update',
                                          existing_obj => $self,
                                          db_change_made => $db_change_made,
                                          validate_value => $validate_value,
                                          policy => $policy);

  my $after_nested_struct_store_rv = $self->after_nested_struct_store(db_rec => $db_rec,
                                                                      store_mode => 'update',
                                                                      store_obj => $store_obj,
                                                                      existing_obj => $self,
                                                                      %params);
  $db_change_made ||= $after_store_rv;
  $db_change_made ||= $after_nested_struct_store_rv;

  $self->from_object(object => $store_obj);
  for my $field_name (@{$self->external_struct_fields}) {
      $self->fill_external_struct_field(criteria => $store_obj->{$field_name}, field_name => $field_name, policy => $policy, logger => $params{logger},);
  }
  for my $null_field (keys %set_fields_to_null) {
      $self->{$null_field} = undef;
  }

  $self->mcp_store(store_obj => $store_obj,
                   existing_obj => $self,
                   %params);

  if (! $params{for_populate_uuids} && !$call_flags->{inherited_call} && $db_change_made) {
      #Refresh full fields to be returned
      $self->retrieve(%params,
                      policy => $policy,
                      get_criteria => $self,
                     );
  }

  if ($db_change_made && ! $params{suppress_events} && ! $params{call_flags}->{suppress_events} && ! $self->suppress_events(%params, store_mode => 'update')) {
    $self->add_change_log_entry(dbh => $dbh,
                                table => $db_table,
                                action => 'update',
                                session => $params{session});
  }

  my $retval = {  db_change_made => $db_change_made,
                  modified_entities => \@modified_entities,
               };

  return $retval;
}

#----------------------------------------------------------------------

sub get_foreign_key_value {
    my($self, %params) = @_;

    my $extern_db_table = $params{extern_db_table};

    my $local_field_name = $self->get_entity_field_by_db_field(db_field_name => $self->extern_db_tables->{$extern_db_table}->{local_join_field});
    my $foreign_key_value;
    if ($local_field_name) {
        $foreign_key_value = $self->{$local_field_name};
    }
    else {
        $foreign_key_value = $self->{$self->extern_db_tables->{$extern_db_table}->{foreign_join_field}};
    }

    return $foreign_key_value;
}

#----------------------------------------------------------------------

sub store_virtual_fields {
    my($self, %params) = @_;

    my $store_obj = $params{store_obj};

    my $virtual_fields = $self->virtual_fields;
    return unless keys %$virtual_fields;
    if (ref($store_obj) =~ /Entity/ && !$params{call_flags}->{force_virtual_field_update} ) {
        return;
    }

    my $fields = $self->fields;
    for my $field_name (keys %$virtual_fields) {
        my $virtual_value_func = $virtual_fields->{$field_name};
        $self->$virtual_value_func(%params, field_name => $field_name, handle_virtual_value =>1);
    }
}

#----------------------------------------------------------------------

sub instantiate_nested_struct_entities {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $store_obj = $params{store_obj};
    my $policy = $params{policy};
    my $validate_value = defined $params{validate_value} ? $params{validate_value} : 1;
    my $fields = $self->fields;

    for my $field_name (@{$self->nested_struct_fields(%params)}) {
        next unless $store_obj->{$field_name};
        my $nested_def = $fields->{$field_name};
        my @nested_fields = keys %{$nested_def->{fields}};
        for my $nested_field_name (@nested_fields) {
            if ($nested_def->{fields}->{$nested_field_name}->{is_affected_by_insensitive_flag} &&
                $policy->{flg_is_case_insensitive}) {
                for my $affected_nested_field (@{$store_obj->{$field_name}}) {
                    $affected_nested_field =~ y/A-Z/a-z/;
                }
            }
            elsif ($nested_def->{fields}->{$nested_field_name}->{is_affected_by_protocol_free_flag} &&
                $policy->{flg_is_protocol_free}) {
                for my $affected_nested_field (@{$store_obj->{$field_name}}) {
                    $affected_nested_field = 0;
                }
            }

            if ( $validate_value &&
                 (defined $nested_def->{fields}->{$nested_field_name}->{required_field} ||
                  defined $nested_def->{fields}->{$nested_field_name}->{normalize_func} ||
                  defined $nested_def->{fields}->{$nested_field_name}->{validate_func}) 
               ) {

                for my $nested_field_val ( @{$store_obj->{$field_name}}){
                    if ( defined $nested_def->{fields}->{$nested_field_name}->{required_field} ){
                        if (!defined $nested_field_val->{$nested_field_name} ) {
                            my $field_display_name_string =  $self->get_field_display_name_string(%params, field_name => $field_name);
                            my $error_message = "Invalid '$field_display_name_string'.";
                            die { error_message   => $error_message,
                                  display_message => $error_message,
                                  error_level     => 'info',
                                  error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                            };
                        }
                    }

                    if ( defined $nested_def->{fields}->{$nested_field_name}->{normalize_func} ){
                        my $val = $nested_field_val;
                        if (ref($nested_field_val)) {
                            $val = $nested_field_val->{$nested_field_name};
                        }
                        if (!ref($val)) {
                            my $rv = $self->validate_field(%params,
                                                           field_def    => $nested_def->{fields}->{$nested_field_name},
                                                           field_value  => $val,
                                                           validate_func => $nested_def->{fields}->{$nested_field_name}->{normalize_func},
                                                           normalize_error_message => $nested_def->{fields}->{$nested_field_name}->{normalize_error_message},
                                                           );
                            if (ref($nested_field_val)) {
                                $nested_field_val->{$nested_field_name} = $rv->{normalized_value};
                            }
                            else {
                                $nested_field_val = $rv->{normalized_value};
                            }
                        }
                    }

                    if ( defined $nested_def->{fields}->{$nested_field_name}->{validate_func} ){
                        my $val = $nested_field_val;
                        if (ref($nested_field_val)) {
                            $val = $nested_field_val->{$nested_field_name};
                        }
                        if (!ref($val)) {
                            $self->validate_field(%params,
                                                  field_value => $val,
                                                  field_name  => $nested_field_name,
                                                  field_def   => $nested_def->{fields}->{$nested_field_name},
                                                  nested_parent_field_name => $field_name,
                                                 );
                        }
                    }
                }
            }
        }

        if ($store_obj->{$field_name}) {
            my @nested_entity_fields = grep { $nested_def->{fields}->{$_}->{entity_type} && $nested_def->{fields}->{$_}->{entity_type}  =~ /::Entity::/ } @nested_fields;

            if (($nested_def->{entity_type} && $nested_def->{entity_type} =~ /::Entity::/) || @nested_entity_fields) {
               my @unfound_nested_rec_indexes;
               for my $i (0..$#{$store_obj->{$field_name}}) {
                   my $nested_rec = $store_obj->{$field_name}->[$i];

                   if ($nested_def->{entity_type} && $nested_def->{entity_type} =~ /::Entity::/) {
                       $nested_rec = $nested_def->{entity_type}->new(dbh => $dbh,
                                                                     logger => $self->{logger},
                                                                     policy => $policy,
                                                                     minimal_load => 1,
                                                                     get_criteria => $nested_rec);

                       unless ($nested_rec) {
                           if ($nested_def->{if_missing} && $nested_def->{if_missing} eq 'skip') {
                               push(@unfound_nested_rec_indexes, $i);
                           }
                           else {
                               die "Associated record for $field_name not found -- aborting\n";
                           }
                       }
                   }
                   else {
                       for my $nested_entity_field (@nested_entity_fields) {
                           if ($nested_rec->{$nested_entity_field}) {
                               $nested_rec->{$nested_entity_field} = $nested_def->{fields}->{$nested_entity_field}->{entity_type}->new(dbh => $dbh,
                                                                                                                                       logger => $self->{logger},
                                                                                                                                       policy => $policy,
                                                                                                                                       minimal_load => 1,
                                                                                                                                       get_criteria => $nested_rec->{$nested_entity_field});
                               unless ($nested_rec->{$nested_entity_field}) {
                                   if ($nested_def->{if_missing} && $nested_def->{if_missing} eq 'skip') {
                                       push(@unfound_nested_rec_indexes, $i);
                                   }
                                   else {
                                       die "Associated record for $field_name not found -- aborting\n";
                                   }
                               }
                           }
                       }
                   }
                   $store_obj->{$field_name}->[$i] = $nested_rec;
               }

               for my $i (reverse @unfound_nested_rec_indexes) {
                   splice(@{$store_obj->{$field_name}}, $i, 1);
               }
            }
        }
    }
}

#----------------------------------------------------------------------

sub update_extern_db_table{
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $table_name = $params{table_name};
  my $rec = $params{rec};
  my $store_mode = $params{store_mode};
  my $foreign_key_name = $params{foreign_key_name};
  my $foreign_key_value = $params{foreign_key_value};
  my $validate_value = $params{validate_value};

  my $fields = $self->fields;

  my $attempt_insert = 0;
  my $delete_row = 0;
  my %update_rec;
  for my $field_name ( keys %$rec ){
      if ($validate_value) {
          $self->validate_field(%params,
                                field_value => $rec->{$field_name}->{field_value},
                                field_name  => $field_name,
                                field_def   => $fields->{$field_name},
                               );
      }
      if (!defined($rec->{$field_name}->{field_value}) && $fields->{$field_name}->{extern_db_optional}) {
          $delete_row = 1;
      }
      else {
          $update_rec{$rec->{$field_name}->{db_field_name}} = $rec->{$field_name}->{field_value};      
          if ((defined($fields->{$field_name}->{default_value}) || $fields->{$field_name}->{extern_db_optional}) && $store_mode eq 'update') {
              $attempt_insert = 1;
          }
      }
  }


  if ( $store_mode eq 'update' ) {
      eval {
          if ($delete_row) {
              my $rv = F5::DbUtils::delete_row(dbh => $dbh,
                                               table_name => $table_name,
                                               rec => {  $foreign_key_name => $foreign_key_value },
                                              );
          }
          else {
              my $rv = F5::DbUtils::update_row(dbh => $dbh,
                                               table_name => $table_name,
                                               rec => \%update_rec,
                                               where_fields => {  $foreign_key_name => $foreign_key_value },
                                              );
              if ($rv) {
                  $attempt_insert = 0;
              }
          }
      };
      if ($@ && !$attempt_insert) {
          die $@;
      }
  }

  if ( !$delete_row && ($store_mode eq 'add' || $attempt_insert) ) {
      $update_rec{$foreign_key_name} = $foreign_key_value;
      F5::DbUtils::insert_row(dbh => $dbh,
                              table_name => $table_name,
                              rec => \%update_rec,
                              );
  }
}

#----------------------------------------------------------------------

sub update_nested_structs {
  my($self, %params) = @_;

  my $dbh = $params{dbh} || $self->{dbh};
  my $store_obj = $params{store_obj};
  my $call_flags = $params{call_flags} || {};
  my $policy = $params{policy};
  my $validate_value = $params{validate_value};
  my $fields = $self->fields;
  my $db_change_made = 0;

  # Update nested structures
  for my $field_name (@{$self->nested_struct_fields(%params)}) {
    my $nested_def = $fields->{$field_name};

    if ($nested_def->{read_only} || ($call_flags->{rest_call} && $nested_def->{rest_read_only})) {
        delete $store_obj->{$field_name}
    }
    elsif ($store_obj->{$field_name}) {

      unless ($self->{$field_name}) {
        $self->{$field_name} = [];
      }

      my $check_inheritance = $policy->{has_parent} && $self->{is_inherited} && ! $call_flags->{inherited_call} 
                              && $nested_def->{fields} && $nested_def->{fields}->{is_inherited};

      if ($call_flags->{append_nested_struct_on_update}) {
        my @preupdate_nested_recs = (@{$self->{$field_name}},
                                     @{$store_obj->{$field_name}});

        # Convert to hash to reduce duplicate values
        my $nested_as_hash = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                               nested_recs => \@preupdate_nested_recs);

        if ($self->check_nested_structs_differ(nested_def => $fields->{$field_name},
                                               old_nested_records => $self->{$field_name},
                                               new_nested_records => $nested_as_hash,
                                               check_inheritance => $check_inheritance,)) {
          if ($nested_def->{db_table_name}) {
              $db_change_made = 1;
          }
        }
        else {
          delete $store_obj->{$field_name};
        }
      }
      elsif ($call_flags->{remove_nested_struct_on_update}) {
        my $existing_nested_structs = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                                        nested_recs => $self->{$field_name});
        my $store_nested_structs = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                                     nested_recs => $store_obj->{$field_name});
        my @nested_structs_to_delete = grep { exists $existing_nested_structs->{$_} } keys %$store_nested_structs;

        #Fail if the object is inherited, and this is not an inherited call, 
        #and this is either not a non-strict list, or the user is trying to remove an inherited entry in that list
        if ($check_inheritance && 
            (grep { $existing_nested_structs->{$_} && $existing_nested_structs->{$_}->{is_inherited} } keys %$store_nested_structs)
           ) {
            die { error_message   => 'Inherited values cannot be modified',
                  error_level     => 'info',
                  error_code      => F5::ASMConfig::ErrorCode::MODIFIED_INHERITED_VALUE(),
            };
        }

        if (@nested_structs_to_delete) {
          my @remaining_nested_structs = grep { ! exists $store_nested_structs->{$_} } keys %$existing_nested_structs;
          $store_obj->{$field_name} = [ map { $existing_nested_structs->{$_} } @remaining_nested_structs ];
          if ($nested_def->{db_table_name}) {
              $db_change_made = 1;
          }
        }
        else {
          delete $store_obj->{$field_name};
        }
      }
      else {
        if ($self->check_nested_structs_differ(nested_def => $nested_def,
                                               old_nested_records => $self->{$field_name},
                                               new_nested_records => $store_obj->{$field_name},
                                               check_inheritance => $check_inheritance,)) {
            if ($nested_def->{db_table_name}) {
                $db_change_made = 1;
            }
        }
        else {
          delete $store_obj->{$field_name};
        }
      }
    }
  }
  return $db_change_made;
}

#----------------------------------------------------------------------

sub perform_nested_struct_db_actions {
    my ($class_or_self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $call_flags = $params{call_flags};
    my $updated_entities = $params{updated_entities};
    my $policy = $params{policy};

    my $fields = $class_or_self->fields;

    for my $field_name (@{$class_or_self->nested_struct_fields(%params)}) {
        my $nested_def = $fields->{$field_name};
        next if $nested_def->{read_only};

        next unless $store_obj->{$field_name};

        if ($call_flags->{append_nested_struct_on_update} || $nested_def->{nested_struct_update_only} ) {

	    $class_or_self->insert_nested_struct(nested_def => $fields->{$field_name}, 
                                                 rec => $store_obj->{$field_name}, 
                                                 updated_entities => $updated_entities,
                                                 policy => $policy,
                                                 dbh => $params{dbh},
                                                 nested_parent_field_name => $field_name,
                                                 call_flags => $call_flags,
                                                );
	}
        elsif ($call_flags->{remove_nested_struct_on_update}) {
            $class_or_self->delete_nested_struct(nested_def => $fields->{$field_name}, 
                                                 rec => $store_obj->{$field_name},
                                                 updated_entities => $updated_entities,
                                                 call_flags => $call_flags,
                                                 dbh => $params{dbh},);
        }
        else {
	    $class_or_self->insert_nested_struct(nested_def => $fields->{$field_name}, 
                                                 rec => $store_obj->{$field_name},  
                                                 delete_existing => 1, 
                                                 updated_entities => $updated_entities,
                                                 policy => $policy,
                                                 dbh => $params{dbh},
                                                 call_flags => $call_flags,
                                                 nested_parent_field_name => $field_name,
                                                );
        }
    }
}

#----------------------------------------------------------------------

sub before_store {
  # abstract placeholder -- may be overriden in subclass
  return {};
}

#----------------------------------------------------------------------

sub after_store {
  # abstract placeholder -- may be overriden in subclass
  return;
}

#----------------------------------------------------------------------

sub mcp_store {
    my ($self, %params) = @_;
    my $call_flags = $params{call_flags} || {};

    if ( ! $call_flags->{skip_mcp} && ! $params{skip_mcp} && $self->is_mcp_object()) {
        my $obj;
        my %refresh_mcp_obj;
        my $query = F5::ASM::COAPI::Query->new($self->mcp_element_type, { $self->mcp_id_field_name => $self->{$self->mcp_id_field} } );
        if ($query && @$query) {
            ($obj) = @$query;
        }
        else {
            $obj = F5::ASM::COAPI::Object->new($self->mcp_element_type);
        }

        my $mcp_id = $self->{$self->mcp_id_field};
        %refresh_mcp_obj = ( success_action => { func => sub  {
                                                     F5::ASM::COAPI::Query->clear_cache(table => $self->mcp_element_type);
                                                     $self->{$self->mcp_id_field} = $mcp_id;
                                                     $self->load_mcp_fields(); 
                                             } }, 
                           );

        my $mcp_change_made = 0;
        for my $field (@{$self->mcp_fields}) {
            my $mcp_field_name = $self->fields->{$field}->{mcp_field_name};
            my $field_value = $self->get_new_value(%params, field_name => $field);
            if (defined($field_value) &&
                (!defined($obj->{$mcp_field_name}) || ($field_value ne $obj->{$mcp_field_name} ) )
               ) {
                $mcp_change_made = 1;
                $obj->{$mcp_field_name} = $field_value;
            }
            elsif (defined ($obj->{$mcp_field_name})) {
                $self->{$field} = $obj->{$mcp_field_name};
            }
        }
        if ($mcp_change_made) {
            $obj->transactional_action(action => 'save',
                                       dbh => $self->{dbh},
                                       %refresh_mcp_obj);
        }
    }
}

#----------------------------------------------------------------------

sub after_nested_struct_store {
  # abstract placeholder -- may be overriden in subclass
}

#----------------------------------------------------------------------

sub add_post_commit_hook {
  # abstract placeholder -- may be overriden in subclass
}

#----------------------------------------------------------------------

sub update_parent_staging {
  # abstract placeholder -- may be overriden in subclass
}

#----------------------------------------------------------------------

sub mysqltime {
    my($class, $time) = @_;
    return POSIX::strftime('%Y-%m-%d %H:%M:%S', localtime($time));
}

#----------------------------------------------------------------------

sub update_content_profile_staging {
    my ($self, %params) = @_;
    ## Update any Parameters or URLs that are using these content_profiles
    my $updated_time = time();
    my @update_ids = @{$params{update_ids}};
    my $search_filter = {field_name => "content_profile__profile_id", 
                         field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(), 
                         field_value_list => \@update_ids };
    my @parameter_content_profiles = F5::ASMConfig::Entity::ParameterContentProfile->retrieve_all(search_filters => [$search_filter],
                                                                                                   dbh => $params{dbh});
    my @parameter_ids = map { $_->{parameter}->{param_id}} (grep { $_->{parameter}->{in_staging} || $_->{parameter}->{has_tightening}} @parameter_content_profiles); 
    if (@parameter_ids) {
        my $param_search_filter = {field_name => "param_id",
                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                  field_value_list => \@parameter_ids };

        F5::ASMConfig::Entity::Parameter->update_many(call_flags => {search_filters => [$param_search_filter]},
                                                      store_obj  => {last_updated => $updated_time},
                                                      dbh        => $params{dbh},
                                                      validate_value => $params{validate_value},);
    }
    
    my @url_content_profiles  = F5::ASMConfig::Entity::UrlContentProfile->retrieve_all(search_filters => [$search_filter],
                                                                                         dbh => $params{dbh});
    my @url_ids =  map { $_->{url}->{url_id}} (grep { $_->{url}->{in_staging} || $_->{url}->{has_tightening}} @url_content_profiles);
    if (@url_ids) {
        my $url_search_filter = {field_name => "url_id",
                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                  field_value_list => \@url_ids };

        F5::ASMConfig::Entity::URL->update_many(call_flags => {search_filters => [$url_search_filter]},
                                                store_obj  => {last_updated => $updated_time},
                                                dbh        => $params{dbh},
                                                validate_value => $params{validate_value},);
    }

}

#----------------------------------------------------------------------

sub update_staging {
  my($class_or_self,%params) = @_;

  my $store_obj = $params{store_obj};
  my $db_rec = $params{db_rec};

  my $now_unixtime = time;
  my $fields = $class_or_self->fields;

  if ($fields->{last_updated}) {

    $db_rec->{last_updated} = $now_unixtime;
    $store_obj->{last_updated} = $now_unixtime;

    if (defined $store_obj->{in_staging} || $params{for_populate_uuids}) {
      if ($class_or_self->get_new_value(%params, field_name => 'in_staging')) {
        if ($store_obj->{staged_since}) {
          $db_rec->{staged_since} = $store_obj->{staged_since};
        }
        else {
          $db_rec->{staged_since} = $store_obj->{staged_since} = $now_unixtime;
        }
      }
      else {
        # Staging has been disabled
        $db_rec->{staged_since} = undef;
        $store_obj->{staged_since} = 0;
      }
    }
  }

  if ( $fields->{has_tightening} && $fields->{last_tightening_time}) {
      if (defined $store_obj->{has_tightening}) {

          if ($store_obj->{has_tightening}) {
              if (not defined $store_obj->{last_tightening_time}) {
                  $db_rec->{last_tightening_time} = $store_obj->{last_tightening_time} = $now_unixtime;
              }
          }
          else {
              # Tightening has been disabled
              $db_rec->{last_tightening_time} = undef;
              $store_obj->{last_tightening_time} = 0;
          }
      }
  }

}

#----------------------------------------------------------------------

sub set_field_in_db_rec {
    my($class_or_self, %params) = @_;

    my $db_rec = $params{db_rec};
    my $field_name = $params{field_name};
    my $field_value = $params{field_value};
    my $existing_obj = $params{existing_obj} || {};
    my $store_obj    = $params{store_obj} || {};
    my $dbh = ref($class_or_self) ? $class_or_self->{dbh} : $params{dbh};
    my $validate_value = defined $params{validate_value} ? $params{validate_value} : 1;

    my $fields = $class_or_self->fields;
    my $field_def = $fields->{$field_name};

    if ($field_def->{db_field_name}) {
        my %set_fields_to_null = $class_or_self->get_fields_to_set_null(%params);
        if ($set_fields_to_null{$field_name}) {
            $db_rec->{$field_def->{db_field_name}} = undef;
        }
        elsif ($field_def->{external_struct}) {
            if (ref($field_value)) {
                $db_rec->{$field_def->{db_field_name}} = $field_value->get_id;
            }
            else {
                if ($validate_value && $field_def->{validate_func}) {
                    $class_or_self->validate_field(%params,
                                                   field_def => $field_def,
                                                   field_value  => $field_value,
                                                  );
                }
                $db_rec->{$field_def->{db_field_name}} = $field_value;
            }
        }
        else {
            if ($field_def->{no_mapping}->{'to_db'}) {
                $db_rec->{$field_def->{db_field_name}} = $field_value;
            }
            else {
                $db_rec->{$field_def->{db_field_name}} = $class_or_self->map_field_value(direction => 'to_db',
                                                                                         field_name => $field_name,
                                                                                         field_def => $field_def,
                                                                                         field_value => $field_value,
                                                                                         store_obj   => $store_obj,
                                                                                         existing_obj => $existing_obj,
                                                                                         dbh => $dbh,
                                                                                         default_rd => $params{default_rd},
                                                                                         validate_value => $validate_value,
                                                                                         call_flags => $params{call_flags},
                                                                                        );
            }
        }
    }
    elsif ($field_def->{bitmask_field}) {
        if (! defined $db_rec->{$field_def->{bitmask_field}}) {
            $db_rec->{$field_def->{bitmask_field}} = $existing_obj->{$field_def->{bitmask_field}} || 0;
        }

        if ($field_value) {
            $db_rec->{$field_def->{bitmask_field}} |= $field_def->{bitmask_value};
        }
        else {
            $db_rec->{$field_def->{bitmask_field}} &= ~$field_def->{bitmask_value};
        }
    }
    elsif ($field_def->{validate_func} && $validate_value) {
        #Always validate value (via map_field_value) if it has a validate_func
        $class_or_self->map_field_value(direction => 'to_db',
                                        field_name => $field_name,
                                        field_def => $field_def,
                                        field_value => $field_value,
                                        store_obj   => $store_obj,
                                        existing_obj => $existing_obj,
                                        dbh => $dbh,
                                        default_rd => $params{default_rd},
                                        validate_value => $validate_value,
                                        call_flags => $params{call_flags},
                                       );
    }
}

#----------------------------------------------------------------------

sub hashify_nested_struct_list {
  my($self, %params) = @_;
  my $clone = $params{clone} || 0;
  my $nested_def = $params{nested_def};
  my $nested_recs = $params{nested_recs};
  if (ref ($params{nested_recs}) ne 'ARRAY') {
      return {};
  }
  my %nested_rec_hash;
  my $fields = $nested_def->{fields} || $nested_def->{entity_type}->fields;

  my @nested_field_names = keys %{$fields};
  my @nested_key_fields = grep { $fields->{$_}->{unique_record_part} } @nested_field_names;
  unless (@nested_key_fields) {
    @nested_key_fields = @nested_field_names;
  }
  for my $nested_rec (@$nested_recs) {
    my $hash_key;
    if (ref($nested_rec)) {
      $hash_key = join($UNLIKELY_DELIMITER, map { defined $nested_rec->{$_} ? (ref($nested_rec->{$_}) =~ /::Entity::/ ? $nested_rec->{$_}->unique_record_parts_str
                                                                                                                      : $nested_rec->{$_})
                                                                            : defined $fields->{$_}->{default_value} ? $fields->{$_}->{default_value}
                                                                                                                     : ''
                                                } @nested_key_fields);
    }
    else {
      $hash_key = $nested_rec;
    }

    $nested_rec_hash{$hash_key} = $nested_rec;
  }

  my $rv = \%nested_rec_hash;
  return ($clone) ? dclone($rv) : $rv;
}
#----------------------------------------------------------------------

sub check_nested_structs_differ {
  my($self, %params) = @_;

  my $check_inheritance = $params{check_inheritance};
  my $nested_def = $params{nested_def};
  my $nested_fields = $nested_def->{fields};
  my @nested_field_names = keys %$nested_fields;

  my $new_nested_records = $params{new_nested_records};
  if (ref($new_nested_records) eq 'ARRAY') {
    $new_nested_records = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                            nested_recs => $new_nested_records);
  }
  my $old_nested_records = $params{old_nested_records};
  if (ref($old_nested_records) eq 'ARRAY') {
    $old_nested_records = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                            nested_recs => $old_nested_records);
  }

  if ($nested_def->{nested_struct_update_only}) {
      for my $key (keys %$new_nested_records) {
          if ( ! exists $old_nested_records->{$key} ) {
              return 1;
          }
      }
  } else {
      my @only_in_new = grep { !$old_nested_records->{$_} } keys %$new_nested_records;
      my @only_in_old = grep { !$new_nested_records->{$_} } keys %$old_nested_records;

      if (@only_in_new || @only_in_old) {
          return 1;
      }
  }

  my $rv = 0;
  if ($nested_def->{entity_type}) {
      my @non_unique_fields = grep { !($nested_fields->{$_}->{unique_record_part}) && 
                                     !($check_inheritance && $_ eq 'is_inherited')} @nested_field_names;
      if (@non_unique_fields) {
          for my $k (keys %$new_nested_records) {
              if ($self->has_value_changed( existing_obj => $old_nested_records->{$k},
                                            store_obj => $new_nested_records->{$k},
                                            check_fields => \@non_unique_fields,
                                            fields => $nested_fields)) {
                  $rv = 1;
                  last unless $check_inheritance;
                  if ($old_nested_records->{$k}->{is_inherited}) {
                       die { error_message   => 'Inherited values cannot be modified',
                             error_level     => 'info',
                             error_code      => F5::ASMConfig::ErrorCode::MODIFIED_INHERITED_VALUE(),
                           };
                  }
              }
          }
      }
  }
  return $rv;
}

#----------------------------------------------------------------------

sub insert_nested_struct {
  my($class_or_self, %params) = @_;

  my $dbh = ref($class_or_self) ? $class_or_self->{dbh} : $params{dbh};
  my $nested_def = $params{nested_def};
  my $rec = $params{rec};
  my $updated_entities = $params{updated_entities};
  my $policy = $params{policy};
  my $call_flags = $params{call_flags};

  return unless $nested_def->{db_table_name};

  if ($params{delete_existing}) {
    my @delete_bind_values;
    my $foreign_id_field;
    my $table_name = $nested_def->{db_table_name};

    for my $entity (@$updated_entities) {
       my $foreign_id = $class_or_self->get_foreign_id(nested_def => $nested_def,
                                                      entity => $entity);
       $foreign_id_field = $foreign_id->{field_name};

       my $where_fields = {$foreign_id_field => $foreign_id->{field_value}};
       if ($nested_def->{fields} && $nested_def->{fields}->{is_inherited}) {
           $where_fields->{is_inherited} = $call_flags->{inherited_call} ? 1 : 0;
       }

       my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);
       my $where_clause = $constructed_where_clause->{where_clause};
       my @where_values = @{$constructed_where_clause->{where_values}};

       my($exists) = $dbh->selectrow_array(<<EndQuery, undef, @where_values);
SELECT count($foreign_id_field)
  FROM $table_name
 WHERE $where_clause
EndQuery

      if ($exists) {
          push (@delete_bind_values, $foreign_id->{field_value});
      }
    }

    if (@delete_bind_values) {
       my $where_fields = {$foreign_id_field => \@delete_bind_values};
       if ($nested_def->{fields} && $nested_def->{fields}->{is_inherited}) {
           $where_fields->{is_inherited} = $call_flags->{inherited_call} ? 1 : 0;
       }

       my $constructed_where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);
       my $where_clause = $constructed_where_clause->{where_clause};
       my @where_values = @{$constructed_where_clause->{where_values}};

      # Delete existing records
         $dbh->do(<<EndQuery, undef, @where_values);
 DELETE FROM $table_name
     WHERE $where_clause
EndQuery
      }
  }


  if (!$rec) {
    return;
  }
  if (ref($rec) ne 'ARRAY') {
    $rec = [ $rec ];
  }
  unless (@$rec) {
    return;
  }
  my @insert_rows;
  my @nested_field_names = grep { $nested_def->{fields}->{$_}->{db_field_name} } keys %{$nested_def->{fields}};
  my @serialized_field_names = grep { !$nested_def->{fields}->{$_}->{db_field_name} && $nested_def->{fields}->{$_}->{serialized_field} } keys %{$nested_def->{fields}};

  my $batched_insertor = $params{batched_insertor} || F5::BatchInsert->new(dbh => $dbh);

  for my $single_rec (@$rec) {
    for my $entity (@$updated_entities) {
      my $foreign_id = $class_or_self->get_foreign_id(nested_def => $nested_def,
                                                      entity => $entity);
      my %insert_row = (
                        policy_id => $policy->{policy_id},
                        $foreign_id->{field_name} => $foreign_id->{field_value},
                       );
  
      if (ref($single_rec)) {
          for my $nested_field_name (@serialized_field_names, @nested_field_names) {
              my $nested_field_def = $nested_def->{fields}->{$nested_field_name};

              if (!defined $single_rec->{$nested_field_name} ){
                  if ( defined $nested_field_def->{default_value} ){
                      $single_rec->{$nested_field_name} = $nested_field_def->{default_value};
                  }
                  else {
                      next;
                  }
              }
 
              if (!$call_flags->{inherited_call} && $nested_field_name eq 'is_inherited') {
                next; #never update is_inherited in not inherited calls
              }

              if ($nested_field_def->{no_mapping}->{'to_db'}) {
                  $insert_row{$nested_field_def->{db_field_name}} = $single_rec->{$nested_field_name};
              }
              else {
                  $insert_row{$nested_field_def->{db_field_name}} = $class_or_self->map_field_value(direction => 'to_db',
                                                                                                    field_name => $nested_field_name,
                                                                                                    field_def => $nested_field_def,
                                                                                                    field_value => $single_rec->{$nested_field_name},
                                                                                                    dbh => $dbh,
                                                                                                    default_rd => $params{default_rd},
                                                                                                    nested_parent_field_name => $params{nested_parent_field_name},
                                                                                                    call_flags => $params{call_flags},
                                                                                                    policy => $policy,
                                                                                                   );
              }
              if ((my $serialized_field = $nested_field_def->{serialized_field}) && defined($single_rec->{$nested_field_name})) {
                  $single_rec->{$serialized_field} = freeze($single_rec->{$nested_field_name});
              }
          }
      }
      else {
        $insert_row{$nested_def->{fields}->{$nested_field_names[0]}->{db_field_name}} = $single_rec;
      }

      $batched_insertor->insert_row(table_name => $nested_def->{db_table_name},
  				    rec => \%insert_row,
                                    do_replace =>        ($call_flags->{inherited_call} || !$params{delete_existing}) ? 1 : 0,
                                    do_insert_ignore =>  ($call_flags->{inherited_call} || !$params{delete_existing}) ? 0 : 1 ,
  				   );

    }
  }

  if (!$params{batched_insertor} ) {
      $batched_insertor->flush;
  }
}

#----------------------------------------------------------------------

sub delete_nested_struct {
  my($class_or_self, %params) = @_;

  my $dbh = ref($class_or_self) ? $class_or_self->{dbh} : $params{dbh};
  my $nested_def = $params{nested_def};
  my $rec = $params{rec};
  my $updated_entities = $params{updated_entities};

  return unless $nested_def->{db_table_name};
  return unless $rec;
  if (ref($rec) ne 'ARRAY') {
    $rec = [ $rec ];
  }
  return unless @$rec;

  my @foreign_ids;
  my $foreign_id_field;
  for my $entity (@$updated_entities) {
    my $foreign_id = $class_or_self->get_foreign_id(nested_def => $nested_def,
                                                    entity => $entity);
    push (@foreign_ids, $foreign_id->{field_value});
    $foreign_id_field = $foreign_id->{field_name};
  }

  my $foreign_id_closure = join (',', map {"?"} @foreign_ids);
  
  my @nested_field_names = grep { $nested_def->{fields}->{$_}->{db_field_name} } keys %{$nested_def->{fields}};
  my @nested_key_fields = grep { $nested_def->{fields}->{$_}->{unique_record_part} } @nested_field_names;
  unless (@nested_key_fields) {
    @nested_key_fields = @nested_field_names;
  }
  my $where_clauses;
  my @where_bind_values;

  if (scalar @nested_key_fields == 1) {
    for my $single_rec (@$rec) {
      if (ref($single_rec)) {
        push(@where_bind_values, $single_rec->{$nested_key_fields[0]});
      }
      else {
        push(@where_bind_values, $single_rec);
      }
    }
    my $db_field_name = $nested_def->{fields}->{$nested_key_fields[0]}->{db_field_name};
    $where_clauses = "$db_field_name IN (" . join(',', ('?') x scalar @where_bind_values) . ')';
  }
  else {
    my @where_clauses;
    for my $single_rec (@$rec) {

      my @where_clause_group;

      for my $nested_key_field (@nested_key_fields) {
        my $db_field_name = $nested_def->{fields}->{$nested_key_field}->{db_field_name};
        push(@where_clause_group, "$db_field_name = ?");
        push(@where_bind_values, $single_rec->{$nested_key_field});
      }

      push(@where_clauses, '(' . join(' AND ', @where_clause_group) . ')');
    }
    $where_clauses = join(' OR ', @where_clauses);
  }

  my $table_name = $nested_def->{db_table_name};
  $dbh->do(<<EndQuery, undef, @foreign_ids, @where_bind_values);
DELETE FROM $table_name
      WHERE $foreign_id_field IN ($foreign_id_closure)
        AND ($where_clauses)
EndQuery

}

#----------------------------------------------------------------------

sub is_deletable {
    my ($self, %params) = @_;
    my $policy = $params{policy} || $self->{policy};
    my $call_flags = $params{call_flags} || {};
    my $fields = $self->fields;

    if ($policy && $policy->{policy_type} && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY() &&
         $policy->{has_parent} && !$call_flags->{inherited_call} && $self->{is_inherited} ){

        return { error_message => "Cannot delete inherited elements.",
                 display_message => "Cannot delete inherited elements.",
                 error_level     => 'info',
                 error_code      =>  F5::ASMConfig::ErrorCode::MODIFIED_INHERITED_VALUE(),
                };
    }
    elsif ( exists $fields->{flg_is_user_defined} && !$self->{flg_is_user_defined}) {
        return { error_message => "User may not delete factory entity",
                 display_message => "User may not delete factory entity",
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED(),
               };
    }
    elsif ($self->{is_referenced} && !$call_flags->{force_delete}  && !$params{is_in_declarative_export}) {
        my $element_type_display_name = $self->get_element_type_display_name;
        my $entity_display_name = $self->get_display_name();
        my $error_message =  "$element_type_display_name \"$entity_display_name\" cannot be deleted while still referenced by other entities";
        return { error_message => $error_message,
                 display_message => $error_message,
                 error_level     => 'info',
                 error_code      => F5::ASMConfig::ErrorCode::FOREIGN_KEY_RESTRICTION(),
               };
    }
    return {};
}

#----------------------------------------------------------------------

sub delete {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $policy = $params{policy} || $self->{policy};

  my $call_flags = $params{call_flags} || {};
  my $fields = $self->fields;
  my $db_table = $self->db_table;
  my @deleted_entities;

  my $is_deletable_rv = $self->is_deletable(%params);
  if ($is_deletable_rv->{error_message}) {
      die $is_deletable_rv;
  }

  if ( ! $call_flags->{skip_mcp} && $self->is_mcp_object()) {
       if (my ($obj) = @{F5::ASM::COAPI::Query->new($self->mcp_element_type, { $self->mcp_id_field_name => $self->{$self->mcp_id_field} })}) {
           $obj->transactional_action(action => 'delete', dbh => $self->{dbh},
                                      success_action => { func => sub { F5::ASM::COAPI::Query->clear_cache(table => $self->mcp_element_type) }});
       }
  }

  if ($db_table) {
      my @where_clauses;
      my @where_values;

      @deleted_entities = $self->delete_linked_entities(dbh => $dbh,
                                                        policy => $policy,
                                                        entities => [ $self ]);

      $self->update_parent_staging(update_ids => [ $self->get_id ],
                                   dbh => $dbh,);

      $self->{logger}->debug("Deleting " . $self->get_debug_string) if $self->{logger};
      if ($self->id_db_field) {
          push(@where_clauses, $self->id_db_field . ' = ?');
          push(@where_values, $self->get_id);
      }
      else {
          my $policy_id_field = $self->policy_id_field;
          if ($policy_id_field && $policy && $policy->{policy_id}) {
              push(@where_clauses, $policy_id_field . ' = ?');
              push(@where_values, $policy->{policy_id});
          }

          my @unique_record_parts = grep { $fields->{$_}->{unique_record_part} } keys %$fields;
          for my $unique_record_part (@unique_record_parts) {
              push(@where_clauses, $fields->{$unique_record_part}->{db_field_name} . ' = ?');
              if ( ref($self->{$unique_record_part}) && $fields->{$unique_record_part}->{entity_type} ){
                  my $unique_id_field = $fields->{$unique_record_part}->{entity_type}->unique_id_field;
                  push(@where_values, $self->{$unique_record_part}->{$unique_id_field});
              }
              else {
                  push(@where_values, $self->{$unique_record_part});
              }
          }
      }

      my $where_clause = 'WHERE ' . join(' AND ', @where_clauses);
      my $rv = $dbh->do(<<EndQuery, undef, @where_values);
DELETE FROM $db_table 
$where_clause
EndQuery

      if ($rv && !$params{suppress_events} && ! $params{call_flags}->{suppress_events} && ! $self->suppress_events(%params, store_mode => 'delete')) {

        $self->add_change_log_entry(dbh => $dbh,
                                    table => $db_table,
                                    action => 'delete',
                                    session => $params{session});
      }
  }

  return { deleted_entities => \@deleted_entities };
}

#----------------------------------------------------------------------

sub delete_linked_entities {
    my ($class_or_self, %params) = @_;

    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $entities = $params{entities};

    if (ref ($class_or_self)) {
        $dbh      ||= $class_or_self->{dbh};
        $policy   ||= $class_or_self->{policy};
        $entities ||= [ $class_or_self ];
    }

    # Delete parameters and extractions tied to this URL
    my @linked_deleted_entities;

    my %entity_call_params = (dbh => $dbh,
                              policy => $policy,
                             );

    for my $entity_filter ($class_or_self->linked_entity_filters(entities => $entities)) {
        %entity_call_params = (%entity_call_params,
                               call_flags => { search_filters => [ $entity_filter->{filter} ]} ,
                              );

        for my $entity_type (@{$entity_filter->{entity_types}}) {
            if ($entity_filter->{cascade_deleted}) {
                push(@linked_deleted_entities,
                    $entity_type->retrieve_all(%entity_call_params,
                                               search_filters => [ $entity_filter->{filter} ],
                                              ));
            }
            else {
                push(@linked_deleted_entities,
                    $entity_type->delete_many(%entity_call_params,
                                             ));
            }
        }
    }
    return @linked_deleted_entities;
}

#----------------------------------------------------------------------

sub case_insensitive_search_filters {
    # abstract placeholder -- may be overriden in subclass
    return [];
}

#----------------------------------------------------------------------

sub linked_entity_filters {
    # abstract placeholder -- may be overriden in subclass
    return;
}

#----------------------------------------------------------------------

sub get_foreign_id {
  my($class, %params) = @_;
  my $nested_def = $params{nested_def};
  my $entity = $params{entity};

  my %retval;
  $retval{parent_table_alias} = $nested_def->{parent_table_alias} || 'main_table';

  if ($nested_def->{foreign_id_field}) {
    $retval{field_name} = $nested_def->{foreign_id_field};

    if ($nested_def->{parent_db_field_name}) {
      $retval{parent_db_field_name} = $nested_def->{parent_db_field_name};
      if ($entity) {
        $retval{field_value} = $entity->get_complex_field(field_name => $nested_def->{parent_entity_field_name});
      }
    }
    else {
      $retval{parent_db_field_name} = $class->id_db_field;
      if ($entity) {
        $retval{field_value} = $entity->{$class->id_field};
      }
    }
  }
  else {
    $retval{field_name} = 'policy_id';
    $retval{parent_db_field_name} = 'policy_id';
    if ($entity) {
      $retval{field_value} = $entity->{policy}->{policy_id};
    }
  }

  return \%retval;
}

#----------------------------------------------------------------------

sub get_complex_field {
    my($self, %params) = @_;
    my $field_name = $params{field_name};
    my @field_name_parts = split(/\./, $field_name);

    # Traverse to nested values
    my $current_point = $self;
    for my $i (0..($#field_name_parts - 1)) {
        if (! $current_point->{$field_name_parts[$i]}) {
            # Can't traverse further -- bail
            return;
        }
        $current_point = $current_point->{$field_name_parts[$i]};
    }
    return $current_point->{$field_name_parts[-1]};
}

#----------------------------------------------------------------------

sub from_object {
  my($self, %params) = @_;

  my $fields = {%{$self->external_structs}, %{$self->fields}};
  my $call_flags = $params{call_flags};

  for my $k (keys %{$params{object}}) {
    if (defined $params{object}->{$k}) {
       my $field_def = $fields->{$k};

       if ($field_def && $field_def->{nested_struct} &&
           ( ($field_def->{nested_struct_update_only} && defined $self->{$k})
            || ($call_flags->{append_nested_struct_on_update} || $call_flags->{remove_nested_struct_on_update})
          )) {
            my $existing_values = $self->hashify_nested_struct_list(nested_def => $field_def,
                                                                    nested_recs => $self->{$k});

            my $store_values = $self->hashify_nested_struct_list(nested_def => $field_def,
                                                                    nested_recs => $params{object}->{$k});
            if ($field_def->{nested_struct_update_only}) {
                # Remove illegal keys from store values
                for my $store_value_key (keys %$store_values) {
                    if (! defined $existing_values->{$store_value_key}) {
                        delete $store_values->{$store_value_key};
                    }
                }
                %$store_values = (%$existing_values, %$store_values);
            }
            elsif ($call_flags->{append_nested_struct_on_update}) {
                %$store_values = (%$existing_values, %$store_values);
            }
            elsif ($call_flags->{remove_nested_struct_on_update}) {
                for my $store_value_key (keys %$store_values) {
                    delete $existing_values->{$store_value_key};
                }
                %$store_values = %$existing_values;
            }

            $self->{$k} = [ values %$store_values ];
       }
       else {
            my @values = $field_def->{nested_struct} ? @{$params{object}->{$k}}
                                                     : ($params{object}->{$k});
            if ($params{instantiate_entities} && $field_def && $field_def->{entity_type} && ($field_def->{entity_type} =~ /::Entity::/)) {
                for my $value (@values) {
                    if (ref($value) !~ /::Entity::/ ) {
                        my %instantiate_params = (%params,
                                                  policy => $self->{policy},
                                                  dbh => $self->{dbh});
                        if ($field_def->{entity_type}->entity_config->{full_retrieve_on_instantiate} || $params{full_retrieve_on_instantiate} ) {
                            $instantiate_params{get_criteria} = $value;
                            delete $instantiate_params{object};
                        }
                        else {
                            $instantiate_params{object} = $value;
                        }
                        $value = $field_def->{entity_type}->new(%instantiate_params);
                    }
                }
            }
            $self->{$k} = $field_def->{nested_struct} ? \@values
                                                      : $values[0];
       }
    }   
  }

  for my $field_name (keys %{$self->virtual_fields}) {
      my $virtual_value_func = $self->virtual_fields->{$field_name};
      $self->$virtual_value_func(%params, field_name => $field_name, handle_virtual_value =>1);
  }
}

#----------------------------------------------------------------------

sub validate_rest_object {
    my ($self, %params) = @_;
    my $obj = $params{object};
    my $valid_fields = $params{valid_fields};
    my $references = $params{references} || {};
    my $nested_field = $params{nested_field} || '';
 
   for my $reference (keys %$references) {
       $valid_fields->{$reference} = 1;
       if (my $ref_name = $references->{$reference}->{reference_name}) {
           $valid_fields->{$ref_name} = 1;
       }
   }


    for my $rest_field (keys %$obj) {
        if ($valid_fields->{$rest_field}) {
            if (defined($obj->{$rest_field}) && ref $obj->{$rest_field} && ref $valid_fields->{$rest_field}) {
                $self->validate_rest_object(object => $obj->{$rest_field},
                                            valid_fields => $valid_fields->{$rest_field},
                                            nested_field => $rest_field);
            }
        }
        else {
            my $error_message = "Unknown field '" . ($nested_field ? "$nested_field/" : "") . "$rest_field' in " . $self->get_element_type_display_name;
            my $logger = $self->{logger} ? $self->{logger}
                                         : F5::Logger->new('/var/log/ts/asm_config_server.log');
            $logger->info($error_message);
        }
    }
}

#----------------------------------------------------------------------

sub from_rest {
    my($self, %params) = @_;
    my $obj;
    eval {
        $obj = $params{object};
        if (!$obj) {
            if ($params{rest_object}) {
                $obj = decode_json($params{rest_object});
                delete $obj->{selfLink};
            }
            else {
                $obj = $self;
            }
        }
    };
    if (my $err = $@) {
        die { error_message => $err,
              display_message => $err,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
            };
    }

    my $ver = $params{ver} || $params{call_flags}->{rest_version} || F5::BigipVersionUtils::bigip_version();
    #field translation here
    my %rest_field_map = $params{rest_field_map} ? %{$params{rest_field_map}}
                                                 : %{$self->rest_field_map()};
 
    my $return_type = ($params{field_def} && $params{field_def}->{entity_type}) ? $params{field_def}->{entity_type}
                                                                                : ref $self;
    my $fields = $params{fields_def} || {%{$self->fields}, %{$self->diff_structs}};

    my $valid_fields = { kind => 1,
                         selfLink => 1,
                         generation => 1,
                         lastUpdateMicros => 1,
                         __short_class => 1,
                       };

    my $extract_policy_field = $params{extract_policy_field};
    my @check_fields;
    my %rest_composite_fields;
    my %rv_params = (dbh => $self->{dbh});
    my $policy = $params{policy} || $self->{policy};
    if (my $self_link = $obj->{selfLink}) {
        my $rest_uuid = F5::Utils::Rest::extract_rest_uuid_from_link('link' => $self_link); #keep only the rest_uuid on the end
        $rv_params{get_criteria} = { rest_uuid => $rest_uuid };
        $policy = F5::Utils::Rest::extract_policy_from_link(dbh => $self->{dbh}, 'link' => $self_link) || $policy;
    }
    elsif (defined(my $rest_uuid = $obj->{id})) {
        $rv_params{get_criteria} = { rest_uuid => $rest_uuid };
    }

    if ($return_type ne 'F5::ASMConfig::Entity::Policy') {
        $rv_params{policy} = $policy;
    }

    if ($params{skip_retrieve}) {
        delete $rv_params{get_criteria};
    }

    my $rv = $return_type->new(%rv_params,
                               selected_fields => []);
    if (!$rv) {
        delete $rv_params{get_criteria};
        $rv = $return_type->new(%rv_params,
                                 selected_fields => []);
    }

    my $convert_to_utf8 = (F5::BigipVersionUtils::compare_versions(version1 => $UTF8_ENCODE_VERSION,
                                                                   version2 => $ver) <= 0);

    my @fields = keys %$fields;
    if ($extract_policy_field) {
        @fields = ($extract_policy_field, grep {$_ ne $extract_policy_field } @fields);
    }
    for my $entity_field (@fields) {
        if (my $rest_field_name = $rest_field_map{$entity_field}) {
            my $field_def = $fields->{$entity_field};
            my $value;
            my $exists;
            my $rest_group_name = $field_def->{rest_group} || '';
            my $rest_expand_name = $field_def->{rest_expand_name};
            if ($rest_expand_name && (( !$rest_group_name && $obj->{$rest_expand_name}) 
                                      || ( $rest_group_name && $obj->{$rest_group_name} && $obj->{$rest_group_name}->{$rest_expand_name}))
                ) {
                my $rest_value;
                if ( $rest_group_name) {
                     $rest_value = $obj->{$rest_group_name}->{$rest_expand_name};
                     $valid_fields->{$rest_group_name}->{$rest_expand_name} = 1;
                }
                else {
                     $rest_value = $obj->{$rest_expand_name};
                     $valid_fields->{$rest_expand_name} = 1;
                }

                next unless defined($rest_value) && ref($rest_value);

                my @rest_values = ($field_def->{nested_struct} && !$field_def->{map_rest_nested_as_single}) ? @$rest_value : ($rest_value);
                my @values;
                for my $rest_value (@rest_values) {
                    if (my $link = $rest_value->{'link'}) {
                        my $entity_type = $field_def->{rest_entity_type} || $field_def->{entity_type};
                        my $local_policy;
                        if ($entity_type ne "F5::ASMConfig::Entity::Policy") {
                            $local_policy = F5::Utils::Rest::extract_policy_from_link(dbh => $self->{dbh}, 'link' => $link) || $policy;
                        }
                        my $rest_uuid = $link;
                        if ($link =~ m|/mgmt/tm/asm/|) {
                            $rest_uuid =  F5::Utils::Rest::extract_rest_uuid_from_link('link' => $link); #keep only the rest_uuid on the end
                        }

                        if (!defined($rest_uuid) || !length($rest_uuid)) {
                            die "Invalid reference for $rest_expand_name (uuid is not defined)\n";
                        }

                        my $get_criteria = { rest_uuid => $rest_uuid };  #May need to add on foreign relations here
                        my $local_value = $entity_type->new(dbh => $self->{dbh},
                                                   policy => $local_policy,
                                                   get_criteria => $get_criteria,
                                                  );

                        if (!$local_value) {
                            die "Invalid reference for $rest_expand_name ('$rest_uuid')\n";
                        }

                        if ($link =~ m|([^\/]+)/[^\/]+$|) {
                            my $link_rest_path_name = $1;
                            if ($local_value->element_rest_path_name eq 'content-profiles') {
                                if ( $link_rest_path_name !~ /\-profiles$/) {
                                    die "Invalid reference for $rest_expand_name (wrong profile path '$link_rest_path_name')\n";
                                }
                            }
                            else {
                                my $element_rest_path_name = $local_value->element_rest_path_name;
                                if ( ($link_rest_path_name ne $element_rest_path_name)
                                    && !($element_rest_path_name =~ /\// && $link =~ /$element_rest_path_name\/[^\/]+$/ )) {
                                    die "Invalid reference for $rest_expand_name (wrong path '$link_rest_path_name')\n";
                                }
                            }
                        }

                        if ($field_def->{rest_value_func} && $field_def->{rest_value_func} eq 'map_rest_external_struct') {
                             my $field_name = $entity_field;
                             if ($field_def->{rest_external_struct_field_name}) {
                                 $field_name = $field_def->{rest_external_struct_field_name};
                             }
                             elsif ($field_def->{fields}) {
                                 ($field_name) = keys %{$field_def->{fields}};
                             }
                             if ($field_def->{entity_type}) {
                                 $local_value = $field_def->{entity_type}->new($local_value);
                             }
                             else {
                                 $local_value = $local_value->{$field_name};
                             }
                            
                        }
                        push (@values, $local_value);
                    }
                    else {
                        die "Invalid reference for $rest_expand_name\n";
                    }
                }
                $value = $field_def->{nested_struct} ? \@values : $values[0];
            }
            else {
                my $rest_value;
                if ( ! $rest_group_name) {
                    $exists = exists ($obj->{$rest_field_name});
                    $rest_value = $obj->{$rest_field_name};
                    $valid_fields->{$rest_field_name} = 1;
                    if ($field_def->{check_base64}) {
                        $valid_fields->{$rest_field_name . 'Base64Encoded'} = 1;
                        if (($obj->{$rest_field_name . 'Base64Encoded'} || '') eq JSON::true) {
                            $rest_value = F5::Utils::Encode::decode_base64($rest_value);
                        }
                        elsif ($convert_to_utf8 && $policy && $policy->{encoding} ne 'utf-8') {
                            $rest_value = $self->safe_encode(value => $rest_value,
                                                             from => 'utf-8',
                                                             to => $policy->{encoding});
                        }
                    }
                }
                elsif (exists $obj->{$rest_group_name}) {
                    if ( ref($obj->{$rest_group_name}) ne 'HASH' ){
                        die "'$rest_group_name' value must be an object.\n";
                    }
                    $exists = exists ($obj->{$rest_group_name}->{$rest_field_name});
                    $rest_value = $obj->{$rest_group_name}->{$rest_field_name};
                    $valid_fields->{$rest_group_name}->{$rest_field_name} = 1;
                    if ($field_def->{check_base64}) {
                        $valid_fields->{$rest_group_name}->{$rest_field_name . 'Base64Encoded'} = 1;
                        if ($obj->{$rest_group_name}->{$rest_field_name . 'Base64Encoded'} eq JSON::true) {
                            $rest_value = F5::Utils::Encode::decode_base64($rest_value);
                        }
                        elsif ($convert_to_utf8 && $policy && $policy->{encoding} ne 'utf-8') {
                            $rest_value = $self->safe_encode(value => $rest_value,
                                                             from => 'utf-8',
                                                             to => $policy->{encoding});
                        }
                    }
                }

                if (defined $rest_value ) {
                    my $policy_encoding;
                    if ($convert_to_utf8 && $policy && $policy->{encoding}) {
                        $policy_encoding = $policy->{encoding};
                    }
                    $value = $self->map_rest_field_value( field_value => $rest_value,
                                                          field_name => $entity_field,
                                                          field_def => $field_def,
                                                          rest_field_name => $rest_field_name,
                                                          policy => $policy,
                                                          direction => 'from_rest',
                                                          skip_retrieve => $params{skip_retrieve},
                                                          call_flags => $params{call_flags},
                                                          dbh => $self->{dbh},
                                                          policy_encoding => $policy_encoding,
                                                        );
                }
            }
            if (defined($value) || ($field_def->{significant_null_value} && $exists)) {
                if (utf8::is_utf8($value)) {
                    $value = $self->safe_encode(value => $value,
                                                from => 'utf-8');
                }
                $rv->{$entity_field} = $value;
                if ($field_def->{rest_composite_field}) {
                    $rest_composite_fields{$field_def->{rest_composite_field}}->{$entity_field} = $value;
                }
                if ($fields->{$entity_field}->{exclude_from_rest_func} && !$params{call_flags}->{show_all_rest_fields}) {
                    push(@check_fields, $entity_field);
                }
                if ($extract_policy_field && $extract_policy_field eq $entity_field) {
                    if ($field_def->{nested_struct}) {
                        $policy = $value->[0]->policy();
                    }
                    else {
                        $policy = $value->policy();
                    }
                    $rv->{policy} = $policy;
                }
            }
            elsif ($exists && !$field_def->{validate_non_empty}) {
                $params{call_flags}->{set_fields_to_null} ||= [];
                push @{$params{call_flags}->{set_fields_to_null}}, $entity_field;
            }
        }
    }

    for my $rest_composite_field (keys %rest_composite_fields) {
        my $field_def = $fields->{$rest_composite_field};
        my $entity_type = $field_def->{entity_type};
        $rv->{$rest_composite_field} = $entity_type->new(dbh => $self->{dbh},
                                                         policy => $policy,
                                                         object => $rest_composite_fields{$rest_composite_field});
    }

    $self->validate_rest_object(object => $obj,
                                valid_fields => $valid_fields,
                                references => $params{references},);

    for my $field (@check_fields) {
        my $field_def = $fields->{$field};
        my $func = $field_def->{exclude_from_rest_func};

        if ( ref($rv) =~ /::Entity::/ ) {
             if ($rv->$func(%params, field_def => $field_def, direction => 'from_rest')) {
                  delete $rv->{$field};
             }
        }
        else {
            if ( $self->$func(%params, object => $rv, field_def => $field_def, direction => 'from_rest') ){
                delete $rv->{$field};
            }
        }
    }
    return $rv;
}

#----------------------------------------------------------------------

sub get_id {
  my($self) = @_;
  if (my $id_field = $self->id_field) {
    return $self->{$id_field};
  }
  return;
}

#----------------------------------------------------------------------

sub get_rest_unique_id {
  my($self) = @_;

  my $rest_unique_id_field = $self->rest_unique_id_field;
  if ($rest_unique_id_field) {
    return $self->{$rest_unique_id_field};
  }
  return;
}

#----------------------------------------------------------------------

sub get_unique_id {
    my ($self, %params) = @_;
    if (my $unique_id_field = $self->unique_id_field) {
        if ($self->fields->{$unique_id_field}->{entity_type} && $self->fields->{$unique_id_field}->{entity_type} =~ /Entity/){
            return $self->{$unique_id_field}->{$self->fields->{$unique_id_field}->{entity_type}->unique_id_field()};
        }
        else{
            return $self->{$unique_id_field};
        }
    }
    return $params{default};
}

#----------------------------------------------------------------------

sub set_id {
  my($self, $new_val) = @_;
  if (my $id_field = $self->id_field) {
    $self->{$id_field} = $new_val;
    return $new_val;
  }
  return;
}

#----------------------------------------------------------------------

sub add_change_log_entry {
  my($self, %params) = @_;

  return 1 if $self->suppress_modified_flag_update;

  my $dbh = $params{dbh};
  my $policy_id;
  if ($params{policy_id}) {
    $policy_id = $params{policy_id};
  }
  elsif (ref($self) && $self->{policy}) {
    $policy_id = $self->{policy}->{policy_id};
  }
  return unless $policy_id;

  my $plc_dbname = 'PLC';
  my $table_name = "PL_POLICY_CHANGES_LOG";
  my $user_name = 'config layer';
  if ($params{session} && ( defined $params{session}->{gui_user_name} || defined $params{session}->{user_name})) {
      $user_name = $params{session}->{gui_user_name} || $params{session}->{user_name}
  }

  F5::DbUtils::insert_row(dbh => $dbh,
                          table_name => "$plc_dbname.$table_name",
                          rec => { policy_id => $policy_id,
                                   modified_table => $params{table},
                                   modify_action => $params{action},
                                   user_name => $user_name,
                                   update_time => time(), 
                                 },
                         );
  
  return 1;
}

#----------------------------------------------------------------------

sub get_last_updated {
    my ($self, %params) = @_;
    if ($self->{last_updated}) {
        return $self->{last_updated};
    }
    my @search_filters = ({ field_name  => 'element_type',
                            field_value => $self->element_type,
                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ()});
    if (my $id = $self->get_id) {
        push (@search_filters,  { field_name => 'element_id',
                                  field_value => $id,
                                  field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ()});
    }
    my $config_log = F5::ASMConfig::Entity::ConfigLog->new(dbh => $self->{dbh},
                                                           policy => $self->{policy},
                                                           get_criteria => {},
                                                           search_filters => \@search_filters,
                                                           sort_orders => [ { field_name => 'event_time',
                                                                              sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC()} ],);
    my $last_updated_time = $config_log ? $config_log->{event_time} 
                                        : 0;
    return $last_updated_time;
}

#----------------------------------------------------------------------

sub get_debug_string {
  my($self) = @_;

  my $fields = $self->fields;
  my @display_parts;
  for my $field (keys %$fields) {
    my $field_rec = $fields->{$field};
    if ($field_rec->{id_field} || $field_rec->{unique_record_part}) {
      if (ref($self->{$field}) && ref($self->{$field}) =~ /::Entity::/) {
          push(@display_parts,  "$field = '" . $self->{$field}->get_debug_string . "'");
      }
      else {
          my $print_field_value = $self->{$field} || '';
          $print_field_value =~ s/\0/{NULL}/g;

          push(@display_parts, "$field = '$print_field_value'");
      }
    }
  }

  my $debug_string = $self->element_type_name;
  if (@display_parts) {
    $debug_string .= " [" . join(', ', @display_parts) . ']';
  }
  return $debug_string;
}

#----------------------------------------------------------------------

sub sync_from_pb_config {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $table_name = $params{table_name} || 'PLC.PL_POLICY_BUILDER_CONFIG';
    my $is_pure_wildcard = $self->is_pure_wildcard(%params); 

    if ($is_pure_wildcard ) {
        my $pb_config_field = $params{pb_config_field};        
        my ($pb_config_value) = $dbh->selectrow_array(<<EndQuery, undef, $self->{policy}->{policy_id});
SELECT $pb_config_field
  FROM $table_name
 WHERE policy_id = ?
EndQuery
        $store_obj->{has_tightening} = $pb_config_value;
    }
}

#----------------------------------------------------------------------

sub adjust_wildcards {
    my($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $where_values = $params{where_values} || {};
    my $name_field = $params{name_field} || $self->name_field();
    my $add_default_wildcard_as_last = defined $params{add_default_wildcard_as_last} ? $params{add_default_wildcard_as_last} : 1;

    if ($store_obj->{is_wildcard} && !$existing_obj->{is_wildcard} && ! $store_obj->{wildcard_sequence}) {
        # Wildcard default for update to become wildcard
        my $pure_wildcard_flg = 0;
        if ($self->can("is_pure_wildcard")) {
            $pure_wildcard_flg = $self->is_pure_wildcard(%params);
        }
        elsif ($self->get_new_value(%params, field_name => $name_field) eq '*') {
            $pure_wildcard_flg = 1;
        }

        if ($pure_wildcard_flg && $add_default_wildcard_as_last) {
            $store_obj->{wildcard_sequence} = -1;
        }
        else {
            $store_obj->{wildcard_sequence} = 1;
        }
    }
    else {
        unless ($self->has_value_changed(%params, check_fields => ['wildcard_sequence', 'is_wildcard'])) {
            # Return unless wildcard_sequence differs
            return $existing_obj->{wildcard_sequence};
        }
    }

    return $self->order_field_adjust(%params,
                                     order_field_name => 'wildcard_sequence',
                                     name_field       => $name_field,
                                     where_values => {
                                                      %$where_values,
                                                      flg_is_pattern => 1,
                                                     },
                                    );
}

#----------------------------------------------------------------------

sub order_field_adjust {
    my($self, %params) = @_;
    my $dbh = $self->{dbh};
    my $store_obj = $params{store_obj};
    my $order_field_name = $params{order_field_name};
    my $name_field = $params{name_field};
    my $where_values = $params{where_values} || {};
    my $db_table_name = $params{db_table_name} || $self->db_table;

    my $order_db_field_name = $self->fields->{$order_field_name}->{db_field_name} || $self->fields->{$order_field_name}->{extern_db_field_name};
    my $name_db_field_name;
    if ($name_field) {
        $name_db_field_name= $self->fields->{$name_field}->{db_field_name} || $self->fields->{$name_field}->{extern_db_field_name};
    }

    $where_values->{policy_id} = $self->{policy}->{policy_id};
    my @where_clauses;
    my @where_bind_values;
    for my $k (keys %$where_values) {
        push(@where_clauses, "$k = ?");
        push(@where_bind_values, $where_values->{$k});
    }
    my $where_clauses = join(' AND ', @where_clauses);
    if ( defined $store_obj->{$order_field_name} && $store_obj->{$order_field_name} == -2 && $name_db_field_name){ # -2 indicates that wildcard_sequence should be last right before the pure wildcard
      my $pure_wildcard_pattern_order = $self->{dbh}->selectrow_array(<<EndQuery, undef,'*',@where_bind_values);
SELECT MIN($order_db_field_name)
  FROM $db_table_name
 WHERE $name_db_field_name = ?
   AND $where_clauses
EndQuery

     $store_obj->{$order_field_name} =  $pure_wildcard_pattern_order || -1;
    }

    if (defined($store_obj->{$order_field_name}) && $store_obj->{$order_field_name} == -1) {
        # determine max and set wildcard_sequence 1 above
        my($max_order) = $dbh->selectrow_array(<<EndQuery, undef, @where_bind_values);
SELECT MAX($order_db_field_name)
  FROM $db_table_name
 WHERE $where_clauses
EndQuery

        $max_order ||= 0;
        $store_obj->{$order_field_name} = $max_order + 1;
    }
    else {
        my $existing_order_row = $dbh->selectrow_arrayref(<<EndQuery, undef, $store_obj->{$order_field_name}, @where_bind_values);
SELECT $order_db_field_name
  FROM $db_table_name
 WHERE $order_db_field_name = ?
   AND $where_clauses
EndQuery

        if ($existing_order_row && @$existing_order_row) {
            if ($params{call_flags} && $params{call_flags}->{displace_only_target_wildcard_order}) {
                my($max_order) = $dbh->selectrow_array(<<EndQuery, undef, @where_bind_values);
SELECT MAX($order_db_field_name)
  FROM $db_table_name
 WHERE $where_clauses
EndQuery

                $dbh->do(<<EndQuery, undef, $max_order + 1, $store_obj->{$order_field_name}, @where_bind_values);
UPDATE $db_table_name
   SET $order_db_field_name = ?
 WHERE $order_db_field_name = ?
   AND $where_clauses
EndQuery
            }
            else {
                my $next_available_order_value = F5::DbUtils::find_lowest_unused_id(dbh => $dbh,
                                                                                    table_name => $db_table_name,
                                                                                    id_name => $order_db_field_name,
                                                                                    where_values => $where_values,
                                                                                    min_id_value => $store_obj->{$order_field_name},
                                                                                   );
                if ($params{existing_obj} && $params{existing_obj}->{is_wildcard} &&
                    $params{existing_obj}->{$order_field_name} &&
                    $params{existing_obj}->{$order_field_name} > $store_obj->{$order_field_name} &&
                    $params{existing_obj}->{$order_field_name} < $next_available_order_value) {
                    # The current object is holding this slot, but will free it up after update
                    $next_available_order_value = $params{existing_obj}->{$order_field_name};
                }

                $dbh->do(<<EndQuery, undef, $store_obj->{$order_field_name}, $next_available_order_value, @where_bind_values);
UPDATE $db_table_name
   SET $order_db_field_name = $order_db_field_name + 1
 WHERE $order_db_field_name >= ?
   AND $order_db_field_name < ?
   AND $where_clauses
EndQuery
            }
        }
    }

    return $store_obj->{$order_field_name};
}

#----------------------------------------------------------------------

sub has_value_changed {
    my($class_or_self, %params) = @_;
    my $store_obj = $params{store_obj};
    my $existing_obj = $params{existing_obj};
    my $check_fields = $params{check_fields};
    my $fields = $params{fields} || $class_or_self->fields;

    if (!ref($check_fields)) {
        $check_fields = [ $check_fields ];
    }

    for my $check_field (@$check_fields) {
        my $field_def = $fields->{$check_field};
        my $existing_obj_check_field = undef;
        if ($existing_obj && defined $existing_obj->{$check_field}) {
            $existing_obj_check_field = $existing_obj->{$check_field};
        }

        if (defined $store_obj->{$check_field} || $field_def->{significant_null_value}) {
            if (ref($existing_obj_check_field) && ref($existing_obj_check_field) =~ /::Entity::/) {
                if ($existing_obj_check_field->differs_from($store_obj->{$check_field})) {
                    return 1;
                }
            }
            elsif (ref($store_obj->{$check_field}) && ref($store_obj->{$check_field}) =~ /::Entity::/) {
                if ($store_obj->{$check_field}->differs_from( $existing_obj_check_field)) {

                    return 1;
                }
            }
            elsif (ref($existing_obj_check_field) && ref($existing_obj_check_field) eq "ARRAY") {
                if ($class_or_self->check_nested_structs_differ(nested_def => $field_def,
                                                                old_nested_records => $existing_obj_check_field,
                                                                new_nested_records => $store_obj->{$check_field})) {
                    return 1;
                }
            }
            elsif ( ( defined($existing_obj_check_field) != defined($store_obj->{$check_field}) ) ||
                   (defined($existing_obj_check_field) && ($store_obj->{$check_field} ne $existing_obj_check_field) )) {
                return 1;
            }
        }
    }
    return;
}

#----------------------------------------------------------------------

sub get_new_value {
  my($self, %params) = @_;
  my $store_obj = $params{store_obj} || $self;
  my $existing_obj = $params{existing_obj} || {}; 
  my $field_name = $params{field_name};
  my $call_flags = $params{call_flags};

  if ($self->all_fields->{$field_name}->{nested_struct} && defined $store_obj->{$field_name}
      && $call_flags && ($call_flags->{append_nested_struct_on_update} || $call_flags->{remove_nested_struct_on_update})) {

      my $nested_def = $self->all_fields->{$field_name};
      my $store_nested_structs = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                                   nested_recs => $call_flags->{append_nested_struct_on_update} ? [@{$existing_obj->{$field_name} || []}, @{$store_obj->{$field_name}}]
                                                                                                                                : $store_obj->{$field_name});
      my @nested_structs_to_return;
      if ($call_flags->{append_nested_struct_on_update}) {
          @nested_structs_to_return = (values %$store_nested_structs);
      }
      else {
          my $existing_nested_structs = $self->hashify_nested_struct_list(nested_def => $nested_def,
                                                                          nested_recs => $existing_obj->{$field_name});
          @nested_structs_to_return = map { $existing_nested_structs->{$_} } grep { ! exists $store_nested_structs->{$_} } keys %$existing_nested_structs;
      }
      return \@nested_structs_to_return;
  }
  return (defined $store_obj->{$field_name} ? $store_obj->{$field_name} : $existing_obj->{$field_name});
}

#----------------------------------------------------------------------

sub differs_from {
    my($self, $other_obj) = @_;

    return 1 unless (defined $other_obj);

    my $id_field = $self->id_field;

    if ($self->{$id_field} && $other_obj->{$id_field}) {
        return ($self->{$id_field} ne $other_obj->{$id_field});
    }
    else {
        my @unique_record_parts = @{$self->unique_record_parts};
        for my $field (@unique_record_parts) {
            if (!defined $other_obj->{$field} ||
                $self->{$field} ne $other_obj->{$field}) {

                return 1;
            }
        }
    }
    return;
}

#----------------------------------------------------------------------

sub reset_defaults {
    my($class, %params) = @_;

    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $fields = $class->fields;
    my ($code_field_name) = grep { $fields->{$_}->{code_field} } keys %$fields;

    my %things_old;

    my $iterator = $class->retrieve_all(dbh => $dbh, policy => $policy);

    while(my $thing = $iterator->get_next) {
        $things_old{$thing->{$code_field_name}} = $thing->{flg_is_enabled};
    }

    my $db_table = $class->db_table;
    my $db_defaults_table = $class->db_defaults_table;


   $dbh->do(<<EndQuery, undef, $policy->{policy_id}, $policy->{policy_id});
UPDATE $db_table AS A
   SET flg_is_enabled = ( SELECT flg_is_enabled FROM $db_defaults_table AS B
        WHERE A.$code_field_name = B.$code_field_name
        AND A.policy_id = ? )
   WHERE EXISTS ( SELECT flg_is_enabled FROM $db_defaults_table AS B
        WHERE A.$code_field_name = B.$code_field_name
        AND A.policy_id = ? )
EndQuery

    my @entities_changed = ();

    $iterator = $class->retrieve_all(dbh => $dbh, policy => $policy);
    while(my $thing = $iterator->get_next) {
        if ($things_old{$thing->{$code_field_name}} != $thing->{flg_is_enabled}) {
            push(@entities_changed, $thing);
        }
    }

    $class->add_change_log_entry(dbh => $dbh,
        policy_id => $policy->{policy_id},
        table => $db_table,
        action => 'update',
        session => $params{session}) if scalar @entities_changed;

    return @entities_changed;
}

#----------------------------------------------------------------------

sub update_many_as_iterative {
    my($class, %params) = @_;

    my $call_flags = $params{call_flags} || {};
    if ($call_flags->{inherited_call}) { #includes 'disinherit' calls
        return 1;
    }

    my $id_field = $class->id_field;
    if (!$id_field) {
        my @unique_record_parts = @{$class->unique_record_parts(%params)};
        if (scalar @unique_record_parts != 1) {
            return 1;
        }
    }

    my $store_obj = $params{store_obj};
    my @force_iterative_fields = @{$class->force_iterative_fields(%params)};
    for my $field (@force_iterative_fields) {
        if (defined($store_obj->{$field})) {
            return 1;
        }
    }
    return 0;
}

#----------------------------------------------------------------------

sub update_many {
    my($class, %params) = @_;

    my $call_flags = $params{call_flags} || {};
    my $policy = $params{policy};
    my $search_criteria = $params{search_criteria};
    my $store_obj = $params{store_obj};
    my $dbh = $params{dbh};
    my $logger = $params{logger} || F5::Logger->get_logger(logger_name => 'asm_config_server.pl');

    if ($policy && $policy->{has_parent} && !$call_flags->{inherited_call} && $class->strict_inheritance()) {
        my $is_inherited = F5::ASMConfig::Entity::Section->new(dbh => $dbh,
                                                               policy => $policy,
                                                               get_criteria => {section_type => $class->get_section_type()},
                                                              )->is_inherited();
        if ($is_inherited) {
            die { error_message => "Cannot modify elements from strictly inherited collection.",
                  display_message => "Cannot modify elements from strictly inherited collection.",
                  error_level     => 'info',
                  error_code      =>  F5::ASMConfig::ErrorCode::MODIFIED_INHERITED_VALUE(),
                };
        }
    }

    if ($search_criteria) {
      $class->store_virtual_fields(%params, store_obj => $search_criteria);
    }
    $class->store_virtual_fields(%params);

    my @updated_entities = ();
    my $id_db_field = $class->id_db_field;
    my $id_field = $class->id_field;
    my $fields = $class->fields;
    my $db_table = $class->db_table;

    my @unique_record_parts = @{$class->unique_record_parts(get_criteria => $search_criteria)};
    my @read_only_fields = @{$class->read_only_fields()};
    my @rest_read_only_fields = @{$class->rest_read_only_fields()};
    my @rest_unique_record_parts = @{$class->rest_unique_record_parts(get_criteria => $search_criteria)};
    for my $field (@unique_record_parts, @read_only_fields, @rest_unique_record_parts, $id_field, 'rest_uuid','wildcard_sequence') {
        if (defined($field)) {
            delete $store_obj->{$field};
        }
    }
    if ($call_flags->{rest_call}) {
        for my $field (@rest_read_only_fields) {
            delete $store_obj->{$field};
        }
    }

    my $iterator = $class->retrieve_all(dbh => $dbh, 
                                        policy => $policy, 
                                        get_criteria => $search_criteria,
                                        search_filters => $call_flags->{search_filters},
                                        page_size => $params{page_size},
                                        start_page => $params{start_page},
                                        default_rd => $params{default_rd},
                                        logger => $params{logger},
                                       );

    my $ret_val;
    my @modified_entities;
    if ($class->update_many_as_iterative(%params)) {
        while (my $obj = $iterator->get_next) {
            if ($policy && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY() &&
                  $policy->{has_parent} && !$call_flags->{inherited_call} && $obj->{is_inherited} ){
                my $element_type_display_name = $obj->get_element_type_display_name;
                my $entity_display_name = $obj->get_display_name();
                $logger->debug("Failed to update $element_type_display_name \"$entity_display_name\". Could not modify inherited elements.");
                next;
            }
            my $store_obj_clone = { %$store_obj };
            eval {
                $ret_val = $obj->store_update(%params, store_obj => $store_obj_clone);
            };
            if (my $err = $@) {
                if ($err->{error_code} == F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED()) {
                    $logger->debug($err->{error_message});
                    next;
                }
                else {
                   die $err;
                }
            }
            if( $ret_val->{modified_entities}) {
                push(@modified_entities, @{$ret_val->{modified_entities}});
            }
            if( $ret_val->{db_change_made} || ($ret_val->{modified_entities} && @{$ret_val->{modified_entities}})) {
                $obj->{__store_obj} = $store_obj_clone;
                $obj->{__suppress_events} = !$ret_val->{db_change_made};
                push(@updated_entities, $obj);
            }
        }
    }
    else {
        my $db_rec = {};
        my @check_fields = grep { ($_ ne 'last_updated') && ($_ ne 'staged_since')} @{$class->non_nested_struct_fields};
        my @check_fields_nested = @{$class->nested_struct_fields(%params)};
    
        if (!$call_flags->{inherited_call}) {
            delete $store_obj->{is_inherited};
        }

        for my $field_name (@check_fields) {
            if (defined $store_obj->{$field_name}) {
                if ($fields->{$field_name}->{external_struct}) {
                    $store_obj->{$field_name} = $class->fill_external_struct_field(field_name => $field_name,
                                                                                   policy => $policy,
                                                                                   criteria => $store_obj->{$field_name},
                                                                                   logger => $params{logger},
                                                                                   dbh => $dbh);
                }
    
                $class->set_field_in_db_rec(db_rec => $db_rec,
                                            field_name => $field_name,
                                            field_value => $store_obj->{$field_name},
                                            default_rd => $params{default_rd},
                                            call_flags => $params{call_flags},
                                            policy => $policy,
                                            dbh => $dbh,
                                           );
            }
       }
    
       push(@check_fields, @check_fields_nested);
       my @store_obj_fields = grep { $fields->{$_} } keys %$store_obj;
       $class->update_staging(db_rec => $db_rec,
                               store_obj => $store_obj,
                             );
    
        my %nested_changes = ();
        my $db_change_made = 0;
        my %restricted_inheritance_fields = map { ($_ => 1) } @{$class->restricted_inheritance_fields()};
 
        while(my $rec = $iterator->get_next) {
            my $store_obj_clone = { %$store_obj };
            my $entity_change_made;
            eval {
                $rec->instantiate_nested_struct_entities(store_obj => $store_obj_clone,
                                                         policy => $policy,
                                                        );
                $entity_change_made = $rec->update_nested_structs(db_rec => $db_rec,
                                                                  call_flags => $call_flags,
                                                                  store_obj => $store_obj_clone,
                                                                  policy => $policy,
                                                                  call_flags => $params{call_flags},
                                                                  dbh => $dbh,
                                                                 );
            };
            if (my $err = $@) {
                if ($err->{error_code} == F5::ASMConfig::ErrorCode::MODIFIED_INHERITED_VALUE()) {
                    my $element_type_display_name = $rec->get_element_type_display_name;
                    my $entity_display_name = $rec->get_display_name();
                    $logger->debug("Failed to update $element_type_display_name \"$entity_display_name\". Could not modify inherited elements.");
                    next;
                }
                else {
                    die $err;
                }
            }


            for my $field (@store_obj_fields) {
                if ( $class->has_value_changed( existing_obj => $rec, store_obj => $store_obj_clone, check_fields => $field)) {
                    $entity_change_made = 1;
                    if (!$call_flags->{inherited_call} && $rec->{is_inherited} && $restricted_inheritance_fields{$field} ){
                        $entity_change_made = 0;
                        my $element_type_display_name = $rec->get_element_type_display_name;
                        my $entity_display_name = $rec->get_display_name();
                        $logger->debug("Failed to update $element_type_display_name \"$entity_display_name\". Could not modify inherited elements.");
                        last;
                    }
                }
                else {
                    delete $store_obj_clone->{$field};
                }
            }
            if ($entity_change_made) {
                $db_change_made = 1;
                $rec->from_object(object => $store_obj_clone);
                $rec->{__store_obj} = $store_obj_clone;
                push(@updated_entities, $rec);
            }
        }
    
        if (! scalar @updated_entities) {
            return { updated_entities => [],
                     modified_entities => [],
                   };
        }
    
        local $" = ',';
    
        my @where_clauses;
        my @bind_values;
        if ($id_field) {
            push(@bind_values, map { $_->{$id_field} } @updated_entities);
            push(@where_clauses, $class->id_db_field . " IN (" . join(',', ('?') x scalar @updated_entities) . ')');
        }
        else {
            my $policy_id_field = $class->policy_id_field;
            if ($policy && $policy->{policy_id} && $policy_id_field) {
                push(@where_clauses, $policy_id_field . ' = ?');
                push(@bind_values, $policy->{policy_id});
            }
    
            my @unique_record_parts = @{$class->unique_record_parts};
            my $key_field = $unique_record_parts[0];
            if ($fields->{$key_field}->{external_struct}) {
                my $ext_id_field = $fields->{$key_field}->{foreign_field} || $fields->{$key_field}->{entity_type}->unique_id_field();
                push (@bind_values, map { $_->{$key_field}->{$ext_id_field} } @updated_entities);
            }
            else {
                push (@bind_values, map { $_->{$key_field} } @updated_entities);
            }
            push(@where_clauses, $fields->{$key_field}->{db_field_name} . " IN (" . join(',', ('?') x scalar @updated_entities) . ')');
        }
        my @update_fields = keys %$db_rec;

        if (@update_fields) {    
            my @update_clauses = map { "$_ = ?" } @update_fields;
            my @update_values = @{$db_rec}{@update_fields};
            my $where_clause = join(' AND ', @where_clauses);

            my $rv = $dbh->do(<<EndQuery, undef, @update_values, @bind_values);
    UPDATE $db_table
       SET @update_clauses
     WHERE $where_clause
EndQuery
        }
 
        $class->perform_nested_struct_db_actions(call_flags => $call_flags,
                                                 store_obj => $store_obj,
                                                 updated_entities => \@updated_entities,
                                                 dbh => $dbh,
                                                 policy => $policy,
                                                ) if $db_change_made;
    
        $class->update_parent_staging(update_ids => \@bind_values,
                                      dbh => $dbh);
        if (! $params{suppress_events} && ! $call_flags->{suppress_events}) { 
            $class->add_change_log_entry(
                dbh => $dbh,
                policy_id => $policy->{policy_id},
                table => $db_table,
                action => 'update',
                session => $params{session});
        }
    }

    return { updated_entities => \@updated_entities,
             modified_entities => \@modified_entities };

}

#----------------------------------------------------------------------

sub delete_many_as_iterative {
    my($class, %params) = @_;

    my $call_flags = $params{call_flags} || {};
    if ($call_flags->{inherited_call}) {
        return 1;
    }

    my $id_field = $class->id_field;
    if (!$id_field) {
        my @unique_record_parts = @{$class->unique_record_parts(%params)};
        if (scalar @unique_record_parts != 1) {
            return 1;
        }
    }
    return 0;
}

#----------------------------------------------------------------------
#

sub delete_many {
    my($class, %params) = @_;

    my $call_flags = $params{call_flags} || {};
    my $policy = $params{policy};
    my $search_criteria = $params{search_criteria};
    my $dbh = $params{dbh};
    my $logger = $params{logger} || F5::Logger->get_logger(logger_name => 'asm_config_server.pl');

    if ($policy && $policy->{has_parent} && !$call_flags->{inherited_call} && $class->strict_inheritance()) {
        my $is_inherited = F5::ASMConfig::Entity::Section->new(dbh => $dbh,
                                                               policy => $policy,
                                                               get_criteria => {section_type => $class->get_section_type()},
                                                              )->is_inherited();
        if ($is_inherited) {
            die { error_message => "Cannot delete elements from strictly inherited collection.",
                  display_message => "Cannot delete elements from strictly inherited collection.",
                  error_level     => 'info',
                  error_code      =>  F5::ASMConfig::ErrorCode::MODIFIED_INHERITED_VALUE(),
                };
        }
    }

    my @deleted_entities = ();

    my $iterator = $class->retrieve_all(dbh => $dbh, 
                                        policy => $policy, 
                                        get_criteria => $search_criteria,
                                        search_filters => $call_flags->{search_filters},
                                        sort_orders => $call_flags->{sort_orders},
                                        page_size => $params{page_size},
                                        start_page => $params{start_page},
                                        default_rd => $params{default_rd},                                        
                                       );
    if ($class->delete_many_as_iterative(%params)) {
        while(my $obj = $iterator->get_next) {
            if ($policy && $policy->{policy_type} == F5::ASMConfig::PolicyType::POLICY_TYPE_SECURITY() &&
                  $policy->{has_parent} && !$call_flags->{inherited_call} && $obj->{is_inherited} ){
                my $element_type_display_name = $obj->get_element_type_display_name;
                my $entity_display_name = $obj->get_display_name();
                $logger->debug("Failed to delete $element_type_display_name \"$entity_display_name\". Could not delete inherited elements.");
                next;
            }
            my $rv;
            eval {
                $rv = $obj->delete(%params);
            };
            if (my $err = $@) {
                if ($err->{error_code} == F5::ASMConfig::ErrorCode::ACTION_NOT_ALLOWED()) {
                    $logger->debug($err->{error_message});
                    next;
                }
                else {
                   die $err;
                }
            }
            push (@deleted_entities, @{$rv->{deleted_entities}}, $obj);
        }
    }
    else {
        while(my $rec = $iterator->get_next) {
            my $is_deletable_rv = $rec->is_deletable(%params);
            if ($is_deletable_rv->{error_message}) {
                die $is_deletable_rv;
            }
            push (@deleted_entities, $rec);
        }

        return unless scalar @deleted_entities;
        my @linked_deleted_entities = $class->delete_linked_entities(dbh => $dbh,
                                                                     policy => $policy,
                                                                     entities => \@deleted_entities);
    
        my $db_table = $class->db_table;
        my $id_field = $class->id_field;
        my $fields = $class->fields;

        my @where_clauses;
        my @bind_values;
        if ($id_field) {
          push(@bind_values, map { $_->{$id_field} } @deleted_entities);
          push(@where_clauses, $class->id_db_field . " IN (" . join(',', ('?') x scalar @deleted_entities) . ')');
        }
        else {
          my $policy_id_field = $class->policy_id_field;
          if ($policy && $policy->{policy_id} && $policy_id_field) {
            push(@where_clauses, $policy_id_field . ' = ?');
            push(@bind_values, $policy->{policy_id});
          }
    
          my @unique_record_parts = @{$class->unique_record_parts};
          my $key_field = $unique_record_parts[0];
          if ($fields->{$key_field}->{external_struct}) {
              my $ext_id_field = $fields->{$key_field}->{foreign_field} || $fields->{$key_field}->{entity_type}->unique_id_field();
              push (@bind_values, map { $_->{$key_field}->{$ext_id_field} } @deleted_entities);
          }
          else {
              push (@bind_values, map { $_->{$key_field} } @deleted_entities);
          }

          push(@where_clauses, $fields->{$key_field}->{db_field_name} . " IN (" . join(',', ('?') x scalar @deleted_entities) . ')');
        }
        my $where_clause = join(' AND ', @where_clauses);
        my $rv = $dbh->do(<<EndQuery, undef, @bind_values);
    DELETE
      FROM $db_table
     WHERE $where_clause
EndQuery
    
        my %policy_ids;
        if ($policy && $policy->{policy_id}) {
            $policy_ids{$policy->{policy_id}} = 1;
        }
        else {
            for my $deleted_entity (@deleted_entities) {
                if ($deleted_entity->{policy}) {
                    $policy_ids{$deleted_entity->{policy}->{policy_id}} = 1;
                }
            }
        }

        if (! $params{suppress_events} && ! $call_flags->{suppress_events}) {
            for my $policy_id (keys %policy_ids) {
                $class->add_change_log_entry(
                    dbh => $dbh,
                    policy_id => $policy_id,
                    table => $db_table,
                    action => 'delete',
                    session => $params{session});
            }
        }
        push (@deleted_entities, @linked_deleted_entities);
    }
    return @deleted_entities;
}

#======================================================================
# accessors for specific entity implementations

sub install_accessor {
  my($class_or_self, $field, $val) = @_;

  no strict 'refs';
  no warnings 'redefine';
  my $class = ref($class_or_self) || $class_or_self;

  if (!defined $val) {
    my $entity_config = $class->entity_config;
    $val = $entity_config->{$field};
  }
  *{$class . '::' . $field} = sub {
    return $val;
  };
  return $val;
}

#----------------------------------------------------------------------

sub get_value_db_map_cache {
    my ($self, %params) = @_;
    my $value_db_map = $params{value_db_map};
    my $dbh = $params{dbh};
    my $table_name = $value_db_map->{table_name};
    my $comparison_mapping = get_comparison_mapping(dbh => $dbh);

    my $where_clause = 1;
    if (defined $value_db_map->{db_map_filter_field} &&
        defined $value_db_map->{db_map_filter_value} &&
        defined $value_db_map->{db_map_filter_condition}) {
        my $condition_str = $comparison_mapping->{ $value_db_map->{db_map_filter_condition} };
        $where_clause = "$value_db_map->{db_map_filter_field} $condition_str $value_db_map->{db_map_filter_value}";
    }
    my $map_cache =  $dbh->selectall_hashref(<<"EndQuery", $value_db_map->{lookup_field});
SELECT $value_db_map->{lookup_field}, $value_db_map->{value_field}
  FROM $table_name
  WHERE $where_clause
EndQuery
   return $map_cache;
}

#----------------------------------------------------------------------

sub element_type {
  my ($class) = @_;
  return $class->install_accessor('element_type');
}

#----------------------------------------------------------------------

sub element_rest_name {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    my $rest_name = $entity_config->{element_rest_name} || $class->element_type_name;
    $rest_name =~ s/_/-/g;
    return $class->install_accessor('element_rest_name', $rest_name);
}

#----------------------------------------------------------------------

sub element_rest_name_plural {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    my $rest_name = $entity_config->{element_rest_name_plural} || $entity_config->{element_type_name_plural};
    if (!$rest_name && $entity_config->{rest_paths}) {
        my $rest_path_properties = $entity_config->{rest_paths}->[0]->{rest_path_properties};
        $rest_name = $class->element_rest_name() . ($rest_path_properties->{singleton} ? '' : 's');
    }
    $rest_name =~ s/_/-/g;
    return $class->install_accessor('element_rest_name_plural', $rest_name);
}

#----------------------------------------------------------------------

sub element_rest_path_name {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    my $path_name = $entity_config->{element_rest_path_name} || ($class->element_rest_name_plural);
    $path_name =~ s/_/-/g;
    return $class->install_accessor('element_rest_path_name', $path_name);
}

#----------------------------------------------------------------------

sub element_type_name_plural {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    my $plural_name = $entity_config->{element_type_name_plural} || ($class->element_type_name() . 's');
    return $class->install_accessor('element_type_name_plural', $plural_name);
}

#----------------------------------------------------------------------

sub element_type_name {
  my ($class) = @_;
  return $class->install_accessor('element_type_name');
}

#----------------------------------------------------------------------

sub suppress_modified_flag_update {
  my ($class) = @_;
  return $class->install_accessor('suppress_modified_flag_update');
}

#----------------------------------------------------------------------

sub db_table {
  my ($class) = @_;
  return $class->install_accessor('db_table');
}

#----------------------------------------------------------------------

sub db_defaults_table {
  my ($class) = @_;
  return $class->install_accessor('db_defaults_table');
}

#----------------------------------------------------------------------

sub feature_name {
  my ($class) = @_;
  return $class->install_accessor('feature_name');
}

#----------------------------------------------------------------------

sub example_fields {
  my ($class) = @_;
  my $entity_config = $class->entity_config;
  my $fields = $entity_config->{example_fields} || $entity_config->{fields};
  return $class->install_accessor('example_fields', $fields);
}

#----------------------------------------------------------------------

sub all_fields {
  my ($class) = @_;
  my $entity_config = $class->entity_config;
  my $fields = $entity_config->{fields};
  for my $field_name (keys %$fields) {
      if ($fields->{$field_name}->{interpolated_db_field_name}) {
          my $interpolated_db_field_name = delete $fields->{$field_name}->{interpolated_db_field_name};
          for my $placeholder (keys %{$fields->{$field_name}->{interpolated_fields}}) {
              $interpolated_db_field_name =~ s/\Q$placeholder\E/$fields->{$field_name}->{interpolated_fields}->{$placeholder}/g;
          }
          $fields->{$field_name}->{full_db_field_name} = $interpolated_db_field_name;
     }
     elsif ($fields->{$field_name}->{sqlite_full_db_field_name} && F5::BigipUtils::is_docker()) {
          $fields->{$field_name}->{full_db_field_name} = $fields->{$field_name}->{sqlite_full_db_field_name};
     }
  }
  return $class->install_accessor('all_fields', $fields);
}

#----------------------------------------------------------------------

sub reverse_dependent_fields {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    my $fields = $entity_config->{fields};

    my %rv;
    if (my $dependent_field_config = $class->entity_config->{dependent_fields}) {
      my $reverse_dependent_fields;
      for my $field (keys %$dependent_field_config) {
          for my $dependent_field_list (values %{$dependent_field_config->{$field}}) {
              for my $dependent_field (@$dependent_field_list) {
                  $reverse_dependent_fields->{$dependent_field}->{$field} = 1;
              }
          }
      }
      %rv = map { $_ => [ keys %{$reverse_dependent_fields->{$_}} ] } (keys %$reverse_dependent_fields);
    }
    return $class->install_accessor('reverse_dependent_fields', \%rv);
}

#----------------------------------------------------------------------

sub fields {
  my ($class) = @_;
  my $all_fields = $class->all_fields;
  my %fields = map { $_ => $all_fields->{$_} } grep { ! $all_fields->{$_}->{diff_struct} } keys %{$all_fields};
  return $class->install_accessor('fields', \%fields);
}

#----------------------------------------------------------------------

sub can_delete_on_disinherit {
  my ($class) = @_;
  return $class->install_accessor('can_delete_on_disinherit');
}

#----------------------------------------------------------------------

sub crossfield_validations {
  my ($class) = @_;
  return $class->install_accessor('crossfield_validations');
}

#----------------------------------------------------------------------

sub diff_structs {
    my ($class) = @_;
    my $all_fields = $class->all_fields;
    my %diff_structs = map { $_ => $all_fields->{$_} } grep { $all_fields->{$_}->{diff_struct} } keys %{$all_fields};
    return $class->install_accessor('diff_structs', \%diff_structs);
}

#----------------------------------------------------------------------

sub webhook_fields {
    my ($class) = @_;
    my $all_fields = $class->all_fields;
    my %webhook_fields = map { $_ => $all_fields->{$_} } grep { $all_fields->{$_}->{webhook_field} } keys %{$all_fields};

    for my $field (keys %webhook_fields) {
        $webhook_fields{$field}->{webhook_field_name} = $webhook_fields{$field}->{rest_field_name} || $class->rest_field_map->{$field};
    }

    return $class->install_accessor('webhook_fields', \%webhook_fields);
}
#----------------------------------------------------------------------

sub reverse_webhook_field_map {
    my($class, %params) = @_;
    my $webhook_fields = $class->webhook_fields(%params);
    my %reverse_webhook_field_map = map { $webhook_fields->{$_}->{webhook_field_name} => $_ } keys %$webhook_fields;
    return $class->install_accessor('reverse_webhook_field_map', \%reverse_webhook_field_map);
}

#----------------------------------------------------------------------

sub after_update_func {
    my ($class) = @_;
    return $class->install_accessor('after_update_func');
}

#----------------------------------------------------------------------

#For overriding
sub populate_for_event {
    return;
}

#----------------------------------------------------------------------

sub populate_diff_structs {
    my ($self, %params) = @_;
    my $diff_structs = $self->diff_structs;
    my @fields = $params{fields} ? @{$params{fields}}
                                 : keys %$diff_structs;
    my $policy = $params{policy} || $self->policy;
    my $rest_selected_fields = $params{rest_selected_fields};
    my $call_flags = $params{call_flags};

    if ($params{include_rest_fields_only}) {
        my $rest_field_map = $self->rest_field_map();
        my $selected_fields_hash;
        if ($rest_selected_fields) {
            $selected_fields_hash = $self->rest_selected_fields_to_selected_fields_hash(rest_selected_fields => $params{rest_selected_fields},
                                                                                        fields => $diff_structs);
        }
        @fields = grep {  ($rest_field_map->{$_} && (!$selected_fields_hash || $selected_fields_hash->{$_}))
                        || ($diff_structs->{$_}->{rest_deprecated_version} 
                           && ($call_flags->{show_rest_inheritance_fields} || 
                               F5::BigipVersionUtils::compare_versions(version1 => $diff_structs->{$_}->{rest_deprecated_version},
                                                                      version2 => $params{ver}) > 0 ) )
                       } @fields;
    }
    elsif ($params{selected_fields_hash}) {
        @fields = grep { $params{selected_fields_hash}->{$_} } @fields;
    }

    for my $diff_struct_field (@fields) {
        next if $self->{$diff_struct_field};
        my $diff_struct_def = $diff_structs->{$diff_struct_field};
        my $get_criteria = {};

        if ($diff_struct_def->{where_values}) {
            next if grep { $self->{$_} ne $diff_struct_def->{where_values}->{$_} } keys %{$diff_struct_def->{where_values}};                
        }

        if ($diff_struct_def->{foreign_field}) {
            my $parent_field = $diff_struct_def->{parent_field} ? $self->{$diff_struct_def->{parent_field}} 
                                                                : $self;
            $get_criteria->{$diff_struct_def->{foreign_field}} = $parent_field;
        }

        if ($diff_struct_def->{diff_search_criteria}) {
            for my $diff_search_criteria_field ( keys %{$diff_struct_def->{diff_search_criteria}} ){
                $get_criteria->{$diff_search_criteria_field} = $diff_struct_def->{diff_search_criteria}->{$diff_search_criteria_field};
            }
        }
        my $sort_orders = $diff_struct_def->{nested_struct} ? $diff_struct_def->{entity_type}->natural_sort_orders
                                                            : [];

        my $nested_selected_fields;
        if ($rest_selected_fields && %$rest_selected_fields) {
            my $rest_field_name     = $diff_struct_def->{rest_field_name} || $self->rest_field_map->{$diff_struct_field};
            my $expand_name         = $diff_struct_def->{rest_expand_name} || $rest_field_name;
            $nested_selected_fields = $self->get_selected_rest_fields(%params,
                                                  rest_selected_fields => $rest_selected_fields,
                                                  field_name => $diff_struct_field,
                                                  field_def => $diff_struct_def,
                                                  rest_field_name => $rest_field_name,
                                                  rest_expand_name => $expand_name,
                                              );
            next unless $nested_selected_fields;
        }


        my @structs = $diff_struct_def->{entity_type}->retrieve_all(dbh => $self->{dbh},
                                                                    policy => $policy,
                                                                    get_criteria => $get_criteria,
                                                                    sort_orders => $sort_orders,
                                                                    search_filters => $diff_struct_def->{entity_type}->diff_search_filters,
                                                                    rest_selected_fields => $nested_selected_fields,
                                                                   );

        ($self->{$diff_struct_field}) = $diff_struct_def->{nested_struct} ? \@structs 
                                                                          : @structs;
    }
}

#----------------------------------------------------------------------

sub get_missing_dependent_external_entities {
    return [];
}

#----------------------------------------------------------------------

sub get_missing_external_entities {
    my ($self, %params) = @_;
    my $fields = $self->external_structs();
    my @external_entities;

    for my $field (keys %$fields) {
        my $field_def = $fields->{$field};
        if ($self->{$field}) {
            my @values = $field_def->{nested_struct} ? @{$self->{$field}}
                                                     : ($self->{$field});
            for my $value (@values) {            
                my $field_class = $fields->{$field}->{entity_type} || ref $value;
                my $entity = $field_class->new(dbh => $self->{dbh},
                                               get_criteria => $value,
                                               policy => $self->{policy},
                                               recast => 1);
                if ( !$entity ) {
                    my $entity = $field_class->new(dbh => $self->{dbh},
                                                   policy => $self->{policy},
                                                   object => $value)->recast(do_retrieve => 0);
                    if (my @chained_entities = @{$entity->get_missing_external_entities(%params)}) {
                         push (@external_entities, @chained_entities);
                    }
                    push (@external_entities, $entity);
                }
            }
        }
    }
    return \@external_entities;
}

#----------------------------------------------------------------------

sub get_external_entities {
    my ($self, %params) = @_;
    my $within_policy = $params{within_policy};
    my @external_structs = keys %{$self->external_structs};
    my @external_entities;

    if ($params{exclude_fields}) {
        my %excluded_fields = map { ($_ => 1) } @{$params{exclude_fields}};
        @external_structs = grep { !$excluded_fields{$_} } @external_structs;
    }

    if ($params{fields}) {
        my %fields =  map { ($_ => 1) } @{$params{fields}};
        @external_structs = grep { $fields{$_} } @external_structs;
    }

    for my $field (@external_structs) {
        my @external_objects;
        if ($self->external_structs->{$field}->{nested_external_struct}) {
            my ($nested_field, $nested_subfield) = split('/', $field);
            if ($self->{$nested_field}) {
                for my $nested_object (@{$self->{$nested_field}}) {
                    if ($nested_object->{$nested_subfield}) {
                        push (@external_objects, $nested_object->{$nested_subfield});
                    }
                }
            }
        }
        elsif ($self->{$field}) {
            @external_objects = ($self->{$field});
        }

        for my $external_object (@external_objects) {
            my $field_class = $self->external_structs->{$field}->{entity_type} || ref $external_object;
            next if ($within_policy && !$field_class->policy_id_field());

            my $entity = $field_class->new(dbh => $self->{dbh}, 
                                           get_criteria => $external_object,
                                           policy => $self->{policy},
                                           recast => 1);
            if ($entity) {
                if (my @chained_entities = @{$entity->get_external_entities()}) {
                    push (@external_entities, @chained_entities);
                }
                push (@external_entities, $entity);
            }
        }
    }
    return \@external_entities;
}

#----------------------------------------------------------------------

sub get_diff_entities {
    my ($self, %params) = @_;

    my $diff_structs = $self->diff_structs;
    $self->populate_diff_structs();

    my @diff_entities;  
    for my $field (keys %$diff_structs) {
        if ($self->{$field}) {
            my @entities = $diff_structs->{$field}->{nested_struct} ? @{$self->{$field}}
                                                                    : ($self->{$field});

            my $exclude_fields = $diff_structs->{$field}->{foreign_field} ? [$diff_structs->{$field}->{foreign_field}]
                                                                          : [];

            for my $entity (@entities) {
                if (my @chained_entities =  @{$entity->get_external_entities(exclude_fields => $exclude_fields)}) {
                    push (@diff_entities, @chained_entities);
                }
                push (@diff_entities, $entity);
            }
            delete $self->{$field};
        }
    }
    return \@diff_entities;
}

#----------------------------------------------------------------------

sub get_linked_entities {
    my ($self, %params) = @_;

    my @linked_entities;
    for my $entity_filter ($self->linked_entity_filters()) {
        if (defined($params{automatically_created}) && 
            (!defined($entity_filter->{automatically_created}) || $entity_filter->{automatically_created} != $params{automatically_created})) {
            next;
        }
        for my $entity_type (@{$entity_filter->{entity_types}}) {
            my @entities = $entity_type->retrieve_all(dbh => $self->{dbh},
                                                      policy => $self->{policy},
                                                      search_filters => [ $entity_filter->{filter} ] ,);
            push(@linked_entities, @entities);
        }
    }
    return \@linked_entities;
}

#----------------------------------------------------------------------

sub extern_db_tables {
  my ($class) = @_;
  return $class->install_accessor('extern_db_tables');
}

#----------------------------------------------------------------------

sub all_external_structs {
  my ($class) = @_;
  my $fields = $class->all_fields;

  my %external_structs;
  for my $field (keys %{$fields}) {
      if ($fields->{$field}->{external_struct}) {
          $external_structs{$field} = $fields->{$field};
      }
      elsif (my $entity_type = $fields->{$field}->{rest_entity_type} || $fields->{$field}->{entity_type} ) {
          if ($entity_type =~ /::Entity::/) {
              $external_structs{$field} = $fields->{$field};
          }
      }

      next if !($fields->{$field}->{diff_struct} || $fields->{$field}->{nested_struct});
      my $nested_fields = $fields->{$field}->{fields};
      for my $nested_field (keys %$nested_fields) {
          if ($nested_fields->{$nested_field}->{external_struct}) {
              $external_structs{"$field/$nested_field"} = {%{$nested_fields->{$nested_field}}, nested_external_struct => 1 };
          }
          elsif (my $entity_type = $nested_fields->{$nested_field}->{rest_entity_type} || $nested_fields->{$nested_field}->{entity_type} ) {
              if ($entity_type =~ /::Entity::/) {
                  $external_structs{"$field/$nested_field"} = {%{$nested_fields->{$nested_field}}, nested_external_struct => 1 };
              }
          }
      }
  }

  return $class->install_accessor('all_external_structs', \%external_structs);
}

#----------------------------------------------------------------------

sub external_structs {
  my ($class) = @_;
  my $fields = $class->fields;
  my %external_structs = map { $_ => $fields->{$_} } grep { $fields->{$_}->{external_struct} } keys %{$fields}; 
  for my $field (@{$class->nested_struct_fields}) {
      my $nested_fields = $fields->{$field}->{fields};
      for my $nested_field (keys %$nested_fields) {
          if ($nested_fields->{$nested_field}->{external_struct}) {
              $external_structs{"$field/$nested_field"} = {%{$nested_fields->{$nested_field}}, nested_external_struct => 1 };
          }
      }
  }
  return $class->install_accessor('external_structs', \%external_structs);
}

#----------------------------------------------------------------------

sub recast_field {
  my ($class) = @_;
  my $fields = $class->fields;
  my($id_field) = grep { $fields->{$_}->{recast_map} } keys %$fields;
  return $class->install_accessor('recast_field', $id_field);
}

#----------------------------------------------------------------------

sub uncast_field {
  my ($class) = @_;
  my $fields = $class->fields;
  my ($uncast_field) = grep { $fields->{$_}->{uncast_class} } keys %$fields;
  return $class->install_accessor('uncast_field', $uncast_field);
}

#----------------------------------------------------------------------

sub id_field {
  my ($class) = @_;
  my $fields = $class->fields;
  my ($id_field) = grep { $fields->{$_}->{id_field} } keys %$fields;
  return $class->install_accessor('id_field', $id_field);
}

#----------------------------------------------------------------------

sub rest_unique_id_field {
  my ($class) = @_;
  my $fields = $class->fields;
  my ($rest_unique_id_field) = grep { $fields->{$_}->{rest_unique_id_field} } keys %$fields;
  if (! $rest_unique_id_field) {
      $rest_unique_id_field = $class->unique_id_field;
  }
  return $class->install_accessor('rest_unique_id_field', $rest_unique_id_field);
}

#----------------------------------------------------------------------

sub unique_id_field {
    my ($class) = @_;
    my $unique_id_field = $class->id_field;
    if (!$unique_id_field) {
        my @unique_record_parts = @{$class->unique_record_parts};
        if (scalar @unique_record_parts == 1) {
            $unique_id_field = $unique_record_parts[0];
        }
    }
    return $class->install_accessor('unique_id_field', $unique_id_field);
}

#----------------------------------------------------------------------

sub unique_value_fields {
    my ($class) = @_;
    
    my $unique_value_fields = [ grep { $class->fields->{$_}->{validate_unique_value} } keys %{$class->fields} ];
    return $class->install_accessor('unique_value_fields', $unique_value_fields);
}

#----------------------------------------------------------------------

sub nested_unique_id_field {
    my ($class, %params) = @_;
    my $field_def = $params{field_def};
    my $unique_id_field;
    if ($field_def->{fields}) {
        ($unique_id_field) = grep { $field_def->{fields}->{$_}->{id_field} || $field_def->{fields}->{$_}->{unique_record_part}} (keys %{$field_def->{fields}});
    }
    elsif ($field_def->{entity_type}) {
        ($unique_id_field) = $field_def->{entity_type}->id_field();
    }
    return $unique_id_field;
}

#----------------------------------------------------------------------

sub name_field {
    my ($class) = @_;
    my $fields = $class->fields;
    my ($name_field) = grep { $fields->{$_}->{is_name_field} } keys %$fields;
    return $class->install_accessor('name_field', $name_field);
}

#----------------------------------------------------------------------

sub id_db_field {
  my ($class) = @_;

  my $id_db_field;
  if (my $id_field = $class->id_field) {
    $id_db_field = $class->fields->{$id_field}->{db_field_name};
  }
  return $class->install_accessor('id_db_field', $id_db_field);
}

#----------------------------------------------------------------------

sub get_module_path {
  return dirname(__FILE__);
}

#----------------------------------------------------------------------

sub policy {
    my ($self) = @_;
    return $self->{policy};
}


#----------------------------------------------------------------------

sub policy_id_field {
  my ($class) = @_;
  return $class->install_accessor('policy_id_field');
}

#----------------------------------------------------------------------

sub strict_inheritance {
  my ($class) = @_;
  return $class->install_accessor('strict_inheritance');
}

#----------------------------------------------------------------------

sub get_element_type_display_name {
    my ($class) = @_;
    return $class->element_type_display_name();
}

#----------------------------------------------------------------------

sub element_type_display_name {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    my $display_string = $entity_config->{element_type_display_name};

    return $display_string || '';
}

#----------------------------------------------------------------------

sub element_type_display_name_plural {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    my $display_string = $entity_config->{element_type_display_name_plural} || $entity_config->{element_type_display_name} . 's';

    return $display_string || '';
}

#----------------------------------------------------------------------

sub get_display_name {
  my ($self, %params) = @_;
  my $display_type_precedence = $params{display_type_precedence} || ['display_name'];
  my $entity_config = $self->entity_config;
  my $display_string;
  for my $type (@$display_type_precedence) {
      $display_string = $entity_config->{$type};
      last if $display_string;
  }

  return '' unless $display_string;
  my $empty_string_value = defined($params{empty_string_value}) ? $params{empty_string_value} :  "UNNAMED";

# I prophetically foresee a situation in which we will support "mixed" style strings like 'foo $bar baz'
# we might even use split and join to replace and manipulate the variable. In this case a convertion to binary
# will be a must:  pack("H*", unpack("H*", $pre) . unpack("H*", $pattern) . unpack("H*", $post))

  if ($display_string =~ /^\$(\w+)$/) {
     $display_string = $self->get_field_display_string(field_name => $1, empty_string_value => $empty_string_value, dbh => $params{dbh});
  }
  else {
      $display_string =~ s/\$(\w+)/$self->get_field_display_string(field_name => $1, empty_string_value => $empty_string_value, dbh => $params{dbh})/eg;
  }

  $display_string =~ s/\0/{NULL}/g;

  return $display_string;
}

#----------------------------------------------------------------------

sub get_full_display_name {
    my($self, %params) = @_;
    my $precedence = $params{display_type_precedence} || ['full_display_name', 'display_name', 'element_type_display_name'];
    my $display_name = $self->get_display_name(dbh => $self->{dbh},
                                   %params,
                                   display_type_precedence => $precedence,
                                  );

    return $display_name;
}

#----------------------------------------------------------------------

sub get_rest_display_name {
    my($self, %params) = @_;
    my $display_name = $self->get_full_display_name(dbh => $self->{dbh},
                                                    %params,
                                                    display_type_precedence => ['rest_display_name', 'full_display_name', 'display_name', 'element_type_display_name'],
                                                   );

    return $display_name;
}

#----------------------------------------------------------------------

sub get_tightening_mode_display_values {
    my ($class, %params) = @_;
    my %field_display_value;
    if ($params{call_flags} && $params{call_flags}->{rest_call}) {
       %field_display_value = (F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER => 'never',
                               F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION => 'selective',
                               F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS => 'always',
                               F5::ASMConfig::TighteningMode::TIGHTENING_MODE_COMPACT => 'compact',
                              );
    }
    else {
       %field_display_value = (F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER => 'Never',
                               F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION => 'Selective',
                               F5::ASMConfig::TighteningMode::TIGHTENING_MODE_ALWAYS => 'Add All Entities',
                               F5::ASMConfig::TighteningMode::TIGHTENING_MODE_COMPACT => 'Compact',
                              );
    }
    return \%field_display_value;
}

#----------------------------------------------------------------------

sub get_entity_changes_string {
    my ($class, %params) = @_;
    my @changes = $class->get_entity_changes(%params);
    my @update_strings;
    for my $change (@changes) {
        my $update_string = $change;
        if (ref ($change)) {
            $update_string = $change->{update_string};
            $update_string =~ s/\$FIELD/$change->{field_display_name}/g;
            $update_string =~ s/\$VALUE/$change->{field_display_value}/g;
        }
        push (@update_strings, $update_string);
    }
    return join("\n", @update_strings);
}

#----------------------------------------------------------------------

sub get_entity_changes {
  my ($class, %params) = @_;

  my $entity = $params{entity};
  my $call_flags = $params{call_flags};

  my %set_fields_to_null = $class->get_fields_to_set_null(%params);

  if (ref($class)) {
    $entity = $class;
    $class = ref($class);
  }
  my $full_thrift_obj = $params{full_thrift_obj} || $entity;
  my $dbh = $params{dbh} || $entity->{dbh};

  my @changes;  #({field, field_display_name, field_display_value, update_string})

  my $entity_config = $class->entity_config;

  my $id_field = $class->id_field;
  my $fields = $class->fields;
  my $pb_unique_record_parts = $class->pb_unique_record_parts(all_parts => 1);
  my %pb_unique_record_fields = map { $_ => 1 } (@$pb_unique_record_parts);

  my $change_fields = $params{fields} || [keys %{$entity_config->{fields}}];
  my %seen_composite_field;
  for my $field (@$change_fields) {
    if (defined $entity->{$field} || $set_fields_to_null{$field} ||
        $entity_config->{fields}->{$field}->{display_when_unchanged}) {

      next if ($id_field && $field eq $id_field);
      next if (($params{ignore_pb_unique_record_parts} || $call_flags->{ignore_pb_unique_record_parts} || $params{for_suggestion}) && $pb_unique_record_fields{$field});
      next if ($params{for_suggestion} && $entity_config->{fields}->{$field}->{exclude_from_suggestion_description});

      if (defined $entity_config->{fields}->{$field}->{composite_field}) {
          next if $seen_composite_field{$entity_config->{fields}->{$field}->{composite_field}}++;
          
      }

      my $field_display_config = $class->get_field_display_config(field => $field,
                                                                  full_thrift_obj => $full_thrift_obj,
                                                                  entity_config => $entity_config,
                                                                  field_values => $params{field_values},);

      if ($field_display_config->{external_field_display_config}) {
          $field_display_config->{fields} = $class->full_schema->{$field_display_config->{external_field_display_config}}->{fields};
      }
      if ($field_display_config->{field}) {
          $field = $field_display_config->{field};
      }

      my $field_value = (defined $params{field_values} && exists $params{field_values}->{$field}) ? $params{field_values}->{$field}
                                                                                                  : $full_thrift_obj->{$field};

      my @field_values = ($field_value);
      if ($field_display_config->{display_dynamic_field} && ref($field_value) eq "ARRAY") {
          @field_values = @$field_value;
      }
      for my $value (@field_values) {
          my $update_string;
          if ($field_display_config->{display_value_map_override_phrase}) {
              $update_string = $field_display_config->{display_value_map_override_phrase}->{$value};
          }
          elsif ($field_display_config->{display_override_phrase}) {
              $update_string = $field_display_config->{display_override_phrase};
          }
          elsif (ref($value) && ref($value) eq "ARRAY") {              
              if ($call_flags->{append_nested_struct_on_update}) {
                  $update_string = $params{future_tense} ? $DISPLAY_STRINGS{value_list_will_be_appended}
                                                         : $DISPLAY_STRINGS{value_list_was_appended};
              }
              elsif ($call_flags->{remove_nested_struct_on_update}) {
                  $update_string = $params{future_tense} ? $DISPLAY_STRINGS{value_list_will_be_removed}
                                                         : $DISPLAY_STRINGS{value_list_was_removed};
              }
              else {
                  $update_string = $params{future_tense} ? $DISPLAY_STRINGS{value_list_will_be_updated}
                                                         : $DISPLAY_STRINGS{value_list_was_updated};
              }
          }
          elsif ($params{future_tense}) {
              $update_string = $DISPLAY_STRINGS{value_will_be_updated};
          }
          elsif ($field_display_config->{hidden_value}) {
              $update_string = $DISPLAY_STRINGS{hidden_value_was_updated};
          }
          else {
              $update_string = $DISPLAY_STRINGS{value_was_updated};
          }

          my $field_display_name = $class->get_field_display_name_string(dbh => $dbh,
                                                                         field_name           => $field,
                                                                         field_display_config => $field_display_config,
                                                                         field_value          => $value,
                                                                         full_thrift_obj      => $full_thrift_obj,
                                                                         entity_config        => $entity_config,
                                                                         get_rest_display_name  => $params{get_rest_display_name},
                                                                         field_values         => $params{field_values});

          if ($field_display_name) {
              my $field_display_value = $class->get_field_display_string(dbh => $dbh,
                                                                         field_display_config => $field_display_config,
                                                                         field_value => $value,
                                                                         field_name => $field,
                                                                         get_rest_display_name => $params{get_rest_display_name},
                                                                       );
              if (!defined($field_display_value)) {
                  $update_string = '';
              }

              push(@changes, { field               => $field,
                               update_string       => $update_string,
                               field_display_name  => $field_display_name,
                               field_display_value => $field_display_value, });
         }
      }
    }
  }

  return @changes;
}

#----------------------------------------------------------------------

sub get_field_display_config {
    my ($class_or_self, %params) = @_;
    my $field = $params{field};
    my $entity_config = $params{entity_config} || $class_or_self->entity_config;
    my $full_thrift_obj = $params{full_thrift_obj} || $class_or_self;
    my %field_display_config;

    return {} unless defined $field;

    if (!ref($full_thrift_obj)) {
        $full_thrift_obj = {};
    }
   
    if ($entity_config->{fields}->{$field}) {
        if ($params{nested_field_name}) {
            %field_display_config = %{$entity_config->{fields}->{$field}->{fields}->{$params{nested_field_name}}};
        }
        else {
            %field_display_config = %{$entity_config->{fields}->{$field}};
        }

        if (my $composite_field = $field_display_config{composite_field}) {
            if ($entity_config->{fields}->{$composite_field}) {
                #Take base config
                %field_display_config = (%field_display_config, %{$entity_config->{fields}->{$composite_field}});
                $field_display_config{field} = $composite_field;
            }
            my $composite_field_def = $entity_config->{composite_fields}->{$composite_field};
            while ($composite_field_def) {
                #Inherit field def
                if ($composite_field_def->{field} && $entity_config->{fields}->{$composite_field_def->{field}}) {
                    %field_display_config = (%field_display_config, %{$entity_config->{fields}->{$composite_field_def->{field}}});
                }
                #take overrides
                %field_display_config = (%field_display_config, %$composite_field_def);
                if ($composite_field_def->{check}) {
                    my $check_field = (keys %{$composite_field_def->{check}})[0];
                    my $check_value = (defined $params{field_values} && exists $params{field_values}->{$check_field}) ? $params{field_values}->{$check_field}
                                                                                                                       : $full_thrift_obj->{$check_field};
                    if (!defined($check_value)) {
                        $check_value = '';
                    }
                    $composite_field_def = $composite_field_def->{check}->{$check_field}->{$check_value};
                }
                else {
                    undef $composite_field_def;
                }
            }
        }
    }
    return \%field_display_config;
}

#----------------------------------------------------------------------

sub get_field_display_string {
  my($class_or_self, %params) = @_;

  my $dbh = $params{dbh};
  my $policy = $params{policy};
  my $field_display_config = $params{field_display_config};
  my $field_value = $params{field_value};
  my $field_name = $params{field_name};
  my $empty_string_value = defined($params{empty_string_value}) ? $params{empty_string_value} : "empty value";
  my $full_thrift_obj = $params{full_thrift_obj};

  if (ref($class_or_self)) {
    $dbh ||= $class_or_self->{dbh};
    $policy ||= $class_or_self->{policy};
    $full_thrift_obj ||= $class_or_self;
  }
  if (!$field_display_config) {
      $field_display_config = $class_or_self->get_field_display_config(field => $field_name,
                                                                       full_thrift_obj => $full_thrift_obj,);
      if ($field_display_config->{field}) {
          $field_name = $field_display_config->{field};
      }
      $field_value ||= $full_thrift_obj->{$field_name};            

      if (! $field_value && $field_display_config->{display_field_value_entry}) {
	  my $curr_value = $full_thrift_obj->{ $field_display_config->{display_field_value_entry} };
	  if ($curr_value) {
	      $field_value = $curr_value;
	  }
      }
  }

  if ( ! defined $field_value && defined $field_display_config->{default_value} ){
      $field_value = $field_display_config->{default_value};
  }

  my $field_display_value = $field_value;
  if (ref($field_value)) {
      if (ref($field_value) eq 'ARRAY') {
          my @nested_values = @{$field_value};
          my @nested_display_values;

          for my $nested_value (@nested_values) {
              my $display_string;
              if ($field_display_config->{display_format_string}) {
                  $display_string = $field_display_config->{display_format_string};
                  $display_string =~ s/\$(\w+)/$class_or_self->get_field_display_string(%params,
                                                                                        field_name => $1,
                                                                                        field_value => $nested_value->{$1},
                                                                                        field_display_config => $field_display_config->{fields}->{$1})/eg;
              }
              else {
                  #Recur for simple array elements that may also have value lookups
                  $display_string = $class_or_self->get_field_display_string(%params,
                                                                             field_value => $nested_value,
                                                                             field_display_config => $field_display_config);
              }

              push(@nested_display_values, $display_string);
          }
          return join(', ', @nested_display_values) || 'an empty list';
      }
      elsif (ref($field_value) =~ /^F5::ASMConfig::/ && !$field_display_config->{display_value_func}) {
          my $external_struct;
          if (ref($field_value) =~ /::Entity::/) {
              $external_struct = $field_value;
          }
          else {
              my $entity_class = ref($field_value);
              $entity_class =~ s/ASMConfig::/ASMConfig::Entity::/;
              $external_struct = $entity_class->new(object => $field_value, dbh => $dbh, policy => $policy);
          }
          my $display_name;
          if ($field_display_config->{display_field}) {
              $field_name = $field_display_config->{display_field}
          }
          if ($field_display_config->{display_format_string}) {
              $display_name = $field_display_config->{display_format_string};
              $display_name =~ s/\$(\w+)/$class_or_self->get_field_display_string(%params,
                                                                                  field_value => $field_value->{$1},
                                                                                  field_display_config => $field_display_config->{fields}->{$1})/eg;
          }
          elsif ($field_name && exists($external_struct->{$field_name})) {
               $display_name = $external_struct->get_field_display_string(field_name => $field_name);
          }
          elsif ($field_display_config->{display_value_db_map} && exists($external_struct->{$field_display_config->{display_value_db_map}->{lookup_field}})) {
              #Fall through to display_value_db_map handling
              $field_display_value = $external_struct->{$field_display_config->{display_value_db_map}->{lookup_field}};
          }
          elsif ($params{get_rest_display_name})  {
              $display_name = $external_struct->get_rest_display_name;
          }
          else {
              $display_name = $external_struct->get_full_display_name();
          }

          if (defined($display_name)) {
              return $display_name;
          }
      }
  }  
  if ($field_display_config->{display_value_type}) {
      if ($field_display_config->{display_value_type} eq 'boolean') {
          $field_display_value = ($field_display_value ? 'true' : 'false');
      }
      elsif ($field_display_config->{display_value_type} eq 'enabled_flg') {
          $field_display_value = ($field_display_value ? 'enabled' : 'disabled');
      }
      elsif ($field_display_config->{display_value_type} eq 'allowed_flg') {
          $field_display_value = ($field_display_value ? 'allowed' : 'disallowed');
      }
      elsif ($field_display_config->{display_value_type} eq 'hex') {
          $field_display_value = sprintf('0x%x', $field_display_value);
      }
      elsif ($field_display_config->{display_value_type} eq 'date') {
          $field_display_value = F5::DateTimeUtils::convert_epoch_to_datetime($field_display_value);
      }
      elsif ($field_display_config->{display_value_type} eq 'percent') {
          $field_display_value = $field_display_value . '%';
      }
  }
  elsif ($field_display_config->{display_value_func}) {
      my $function_name = $field_display_config->{display_value_func};
      $field_display_value = $class_or_self->$function_name(%params,
                                                            direction => 'to_display',
                                                            field_value => $field_display_value);
  }

  if ($field_display_config->{display_value_db_map}) {
      my $value_db_map = $field_display_config->{display_value_db_map};
      my $db_map_cache_key = join($UNLIKELY_DELIMITER,
                                  $value_db_map->{table_name},
                                  $value_db_map->{lookup_field},
                                  $value_db_map->{value_field});
      my $map_cache = $field_display_value_db_map_cache->{$db_map_cache_key};

      if ( ! $map_cache || ! $map_cache->{$field_display_value} ) {
          $map_cache = $class_or_self->get_value_db_map_cache(dbh => $dbh,
                                                              value_db_map => $value_db_map);

          $field_display_value_db_map_cache->{$db_map_cache_key} = $map_cache;
      }

      my $cache_display_value;
      if ( $map_cache->{$field_display_value} ) {
          $cache_display_value = $map_cache->{$field_display_value}->{$value_db_map->{value_field}};
      }
      elsif ($value_db_map->{tolerant_mapping} ) {
          $cache_display_value = $field_display_value; #Retain value on a mapping miss
      }
      $field_display_value = defined( $cache_display_value ) ? $cache_display_value : '';
  }

  if ($field_display_config->{display_value_map} && defined $field_display_value && defined $field_display_config->{display_value_map}->{$field_display_value}) {
      $field_display_value = $field_display_config->{display_value_map}->{$field_display_value};
  }

  if ($field_display_config->{display_value_format} && defined $field_display_value) {
      my $value_string = $field_display_config->{display_value_format};
      $value_string =~ s/\$VALUE/$field_display_value/g;
      $field_display_value = $value_string;
  }

  if (! defined($field_display_value) || $field_display_value eq '') {
      $field_display_value = $field_display_config->{empty_string_value} ? $field_display_config->{empty_string_value}
                                                                         : $empty_string_value;
  }

  return defined $field_display_value ? $field_display_value : '';
}

#----------------------------------------------------------------------
my $SECONDS_IN_A_DAY = 60 * 60 * 24;
my $SECONDS_IN_A_MINUTE = 60;
my $SECONDS_IN_AN_HOUR = 60 * 60;

sub convert_seconds_to_days {
    my($self, %params) = @_;

    return if !defined $params{field_value};

    if (!$params{direction} || ($params{direction} =~ /^to_/)) {
        return ($params{field_value} / $SECONDS_IN_A_DAY);
    }
    else {
        return ($params{field_value} * $SECONDS_IN_A_DAY);
    }
}

#----------------------------------------------------------------------

sub convert_seconds_to_minutes {
    my($self, %params) = @_;

    return if !defined $params{field_value};

    if (!$params{direction} || ($params{direction} =~ /^to_/)) {
        return ($params{field_value} / $SECONDS_IN_A_MINUTE);
    }
    else {
        return ($params{field_value} * $SECONDS_IN_A_MINUTE);
    }
}

#----------------------------------------------------------------------

sub convert_seconds_to_hours {
    my($self, %params) = @_;

    return if !defined $params{field_value};

    if (!$params{direction} || ($params{direction} =~ /^to_/)) {
        return ($params{field_value} / ($SECONDS_IN_AN_HOUR));
    }
    else {
        return ($params{field_value} * ($SECONDS_IN_AN_HOUR));
    }
}

#----------------------------------------------------------------------

sub display_defense_attribute {
    my($class_or_self, %params) = @_;
    my $defense_attribute = $params{field_value};
    my $attribute = $class_or_self->get_defense_attributes(%params)->{$defense_attribute->{def_attrib_id}};
    my $value = $defense_attribute->{def_attrib_value};
    if ($attribute->{value_type} eq F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_BOOLEAN()) {
        $value = $value ? "true" : "false";
    }
    elsif ($attribute->{value_type} eq F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER()) {
       $value = ($value || 0) + 0;
    }
    if ((defined $defense_attribute->{flg_check_attrib} && $attribute->{value_type} == F5::ASMConfig::DefenseAttributeValueType::DEFENSE_ATTRIBUTE_VALUE_TYPE_INTEGER())) {
        $value = $defense_attribute->{flg_check_attrib} ? $value : 'any';
    }
    return $value;
}

#----------------------------------------------------------------------

sub get_policy_entities {
    my ($class, %params) = @_;
    my $schema = $class->full_schema(%params);
    my @entities_with_policy = map { $_->{entity_type} } grep { $_->{policy_id_field} } values %{$schema};
    return @entities_with_policy;
}

#----------------------------------------------------------------------

sub extern_unique_record_part_db_fields {
    return [];
}

#----------------------------------------------------------------------

sub get_case_insensitive_fields {
    my ($class, %params) = @_;
    my $entity_fields = $class->fields;
    my @case_insensitive_fields = $class->retrieve_case_insensitive_fields(%params, fields => $entity_fields, db_table_name => $class->db_table);
    return $class->install_accessor('get_case_insensitive_fields', \@case_insensitive_fields);
}

#----------------------------------------------------------------------

sub retrieve_case_insensitive_fields {
    my ($class, %params) = @_;
    my @case_insensitive_fields;
    my @nested_fields;

    my $fields = $params{fields} || $class->fields;

    my $field_names = $params{field_names} || [ grep { $fields->{$_}->{is_affected_by_insensitive_flag} } keys %$fields ];
    return @case_insensitive_fields unless $field_names && @$field_names;

    my $db_table_name = $params{db_table_name} || $class->db_table;
    my @unique_record_parts = map { $fields->{$_}->{db_field_name} } grep { $fields->{$_}->{unique_record_part} } keys %{$fields};

    for my $field (@$field_names) {
        my $case_insensitive_field_rec = { db_field_name => $fields->{$field}->{db_field_name},
                                           is_unique_record_part => $fields->{$field}->{unique_record_part},
                                           db_table => $db_table_name,
                                           extern_db_field_name => $fields->{$field}->{extern_db_field_name},
                                           unique_record_parts => \@unique_record_parts,
                                           field_name => $field
                                         };
        push(@case_insensitive_fields, $case_insensitive_field_rec);

        if ( $fields->{$field}->{fields} && $fields->{$field}->{db_table_name} ) {
            my $nested_fields_rec = { db_table => $fields->{$field}->{db_table_name},
                                      fields => $fields->{$field}->{fields}
                                    };
            push(@nested_fields, $nested_fields_rec);
        }
    }

    if (@nested_fields) {
        for my $nested_fields_hash (@nested_fields) {
            my @affected_fields = $class->retrieve_case_insensitive_fields(%params, fields => $nested_fields_hash->{fields}, db_table_name => $nested_fields_hash->{db_table});
            push(@case_insensitive_fields, @affected_fields);
        }
    }
    return @case_insensitive_fields;
}

#----------------------------------------------------------------------

sub full_schema {
    my ($class) = @_;
    return $class->install_accessor('full_schema', F5::ASMConfig::Utils::get_entity_schema);
}

#----------------------------------------------------------------------

sub entity_config {
   my ($class) = @_;
   my $class_name = ref $class || $class; 
   $class_name =~ s/.*:://g;
   my $entity_config = $class->full_schema->{$class_name} || {};
   return $class->install_accessor('entity_config', $entity_config);
}

#----------------------------------------------------------------------

sub signature_fields {
  my ($class, %params) = @_;
  my $signature_fields = [ grep { $class->fields->{$_}->{is_signature_field} } keys %{$class->all_fields} ];
  return $class->install_accessor('signature_fields', $signature_fields);
}

#----------------------------------------------------------------------

sub metachar_fields {
  my ($class, %params) = @_;
  my $metachar_fields = [ grep { $class->fields->{$_}->{is_metachar_field} } keys %{$class->all_fields} ];
  return $class->install_accessor('metachar_fields', $metachar_fields);
}

#----------------------------------------------------------------------

sub pb_unique_record_parts {
  my($class, %params) = @_;
  my @unique_record_parts = @{$class->unique_record_parts(%params)};
  my $pb_unique_record_parts = [ @unique_record_parts, grep { $class->fields->{$_}->{pb_unique_record_part} } keys %{$class->fields} ];
  return $class->install_accessor('pb_unique_record_parts', $pb_unique_record_parts);
}

#----------------------------------------------------------------------

sub unique_record_parts {
  my($class) = @_;

  my $unique_record_parts = [ grep { $class->fields->{$_}->{unique_record_part} } keys %{$class->fields} ];
  return $class->install_accessor('unique_record_parts', $unique_record_parts);
}

#----------------------------------------------------------------------

sub read_only_fields {
  my($class) = @_;
  my $read_only_fields = [ grep { $class->fields->{$_}->{read_only} } keys %{$class->fields} ];
  return $class->install_accessor('read_only_fields', $read_only_fields);
}

#----------------------------------------------------------------------

sub rest_read_only_fields {
  my($class) = @_;
  my $rest_read_only_fields = [ grep { $class->fields->{$_}->{rest_read_only} } keys %{$class->fields} ];
  return $class->install_accessor('rest_read_only_fields', $rest_read_only_fields);
}

#----------------------------------------------------------------------

sub extra_fields {
    my ($class) = @_;
    my $extra_fields = [ grep { $class->all_fields->{$_}->{extra_field} } keys %{$class->all_fields} ];
    return $class->install_accessor('extra_fields', $extra_fields);
}

#----------------------------------------------------------------------

sub export_fields {
  my($class) = @_;
  my $export_fields = [ sort { $class->fields->{$a}->{export_display_name} cmp $class->fields->{$b}->{export_display_name} } grep { $class->fields->{$_}->{export_display_name} } keys %{$class->fields} ];
  return $class->install_accessor('export_fields', $export_fields);
}

#----------------------------------------------------------------------

sub default_sort_orders {
    my ($class, %params) = @_;
    my @sort_order_fields = sort { $class->fields->{$a}->{default_sort_order} <=> 
                                   $class->fields->{$b}->{default_sort_order} } grep { $class->fields->{$_}->{default_sort_order} } keys %{$class->fields};
    if (!@sort_order_fields && $class->unique_id_field ) {
       @sort_order_fields = ( $class->unique_id_field );
    }

    my @sort_orders;
    for my $field (@sort_order_fields) {
        push (@sort_orders, F5::ASMConfig::SortOrder->new(field_name => $field,
                                                          sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC()));
    }

    return $class->install_accessor('default_sort_orders',\@sort_orders);
}

#----------------------------------------------------------------------

sub force_iterative_fields {
  my($class, %params) = @_;
  my $force_iterative_fields = [ keys %{$class->diff_structs}, 
                                 grep { $class->fields->{$_}->{force_iterative} || $class->fields->{$_}->{extern_db_field_name}} keys %{$class->fields} 
                               ];
  return $class->install_accessor('force_iterative_fields', $force_iterative_fields);
}

#----------------------------------------------------------------------

sub rest_reference_fields {
    my ($class) = @_;
    my $all_fields = $class->all_fields;
    my %rest_reference_fields = map { $_ => $all_fields->{$_} } grep { $all_fields->{$_}->{rest_unique_record_part}
                                          || $all_fields->{$_}->{unique_record_part}
                                          || $all_fields->{$_}->{pb_unique_record_part}
                                          || $all_fields->{$_}->{rest_reference_field}
                                        } keys %{$all_fields};
    return $class->install_accessor('rest_reference_fields', \%rest_reference_fields);
}

#----------------------------------------------------------------------

sub rest_unique_record_parts {
  my($class) = @_;
  my $rest_unique_record_parts = [ grep { $class->fields->{$_}->{rest_unique_record_part}
                                          || ($class->fields->{$_}->{unique_record_part} && !$class->fields->{$_}->{external_struct})
                                        } sort { $a cmp $b } keys %{$class->fields} 
                                 ];
  return $class->install_accessor('rest_unique_record_parts', $rest_unique_record_parts);
}

#----------------------------------------------------------------------

sub mcp_unique_record_parts {
  my($class) = @_;

  my $mcp_unique_record_parts = [ grep { $class->fields->{$_}->{mcp_unique_record_part} } keys %{$class->fields} ];
  return $class->install_accessor('mcp_unique_record_parts', $mcp_unique_record_parts);
}

#----------------------------------------------------------------------

sub unique_record_parts_str {
    my($self, %params) = @_;
    my $object = $params{object} || $self;
    my $unique_record_parts = $params{unique_record_parts} || $self->unique_record_parts(get_criteria => $self);
    return join($UNLIKELY_DELIMITER, map { defined $object->{$_} ? (ref($object->{$_}) =~ /::Entity::/ ? $object->{$_}->unique_record_parts_str
                                                                                                       : $object->{$_})
                                                                 : '' } @$unique_record_parts);
}

#----------------------------------------------------------------------

sub get_rest_uuid {
    my($self, %params) = @_;
    my $unique_record_parts = $self->rest_unique_record_parts(%params);
    my $fields = $self->fields();
    my $object = $params{object} || $self;
    my $rest_uuid;
    my @values;
    for my $unique_record_part (@$unique_record_parts) {
        my $field_def = $fields->{$unique_record_part};
        my $value = defined $object->{$unique_record_part} ? $object->{$unique_record_part} : $self->{$unique_record_part};
        if ($field_def && (defined($field_def->{const_rest_uuid_value}) || defined($field_def->{const_value}))) {
            $value = defined($field_def->{const_rest_uuid_value}) ? $field_def->{const_rest_uuid_value}
                                                                  : $field_def->{const_value};
        }

        if ($field_def && $field_def->{rest_unique_record_part_filter}) {
            my $search_filter = F5::ASMConfig::SearchFilter->new(%{$field_def->{rest_unique_record_part_filter}}, field_name => $unique_record_part);
            my $match = $filter_func{$field_def->{rest_unique_record_part_filter}->{field_comparison}}->(search_filter => $search_filter,
                                                                                                         field_value => $value);
            next unless $match;
 
        }
        if (defined($value)) {
            if (ref($value) && ref($value) =~ /::Entity::/) {
                push (@values, ($value->get_rest_uuid));
            }
            else {
                push (@values, $value);
            }
        }
    }
    if (@$unique_record_parts) {
        $rest_uuid = F5::Utils::Rest::generate_uuid(join($UNLIKELY_DELIMITER, @values));        
    }

    return $rest_uuid;
}

#----------------------------------------------------------------------

sub set_rest_uuid {
    my($self, %params) = @_;
    my $object = $params{object} || $self;
    my $rest_uuid = $self->get_rest_uuid(%params);
    if (defined($rest_uuid) && length($rest_uuid)) {
        $object->{rest_uuid} = $rest_uuid;
    }
    return $object->{rest_uuid};
}

#----------------------------------------------------------------------

sub unique_record_parts_str_for_diff {
    my ($self) = @_;

    my @unique_parts;
    for my $part (@{$self->unique_record_parts(get_criteria => $self)}) {
        if (ref($self->{$part})) {
            if (ref($self->{$part}) =~ /::Entity::/) {
                push (@unique_parts, $self->{$part}->unique_record_parts_str_for_diff);
            }
            else {
               my @sub_fields = keys %{$self->fields->{$part}->{fields}};
               my @unique_sub_fields = grep {$self->fields->{$part}->{fields}->{$_}->{unique_record_part}} @sub_fields;
##TODO: Remove this "if/die" if/when we find there are no issues
               if (@unique_sub_fields) {
                   push (@unique_parts, $self->unique_record_parts_str(object => $self->{$part},
                                                                       unique_record_parts => \@unique_sub_fields));
               }
               else {
                   die "No unique record parts for entity: '". ref($self) . "' part: '$part' (" . ref($self->{$part}) . ")";
               }
            }
        }
        else {
            push (@unique_parts, defined $self->{$part} ? $self->{$part} : '');
        }
    }
    return join($UNLIKELY_DELIMITER, @unique_parts);
}

#----------------------------------------------------------------------

sub natural_sort_orders {
    my ($class, %params) = @_;

    my $external_structs = $class->external_structs(%params);
    my $prefix = $params{prefix} || "";
    my @sort_orders;
    for my $field (@{$class->unique_record_parts(all_parts => 1)}) {
        if ($external_structs->{$field}) {
            my @external_sort_orders = @{$external_structs->{$field}->{entity_type}->natural_sort_orders(prefix => "$prefix${field}__")};
            push (@sort_orders, @external_sort_orders) if @external_sort_orders;
        }
        else {
            push (@sort_orders, F5::ASMConfig::SortOrder->new(field_name => $prefix . $field,
                                                              sort_direction => F5::ASMConfig::SortDirection::SORT_DIRECTION_BINARY_ASC()));
        }
    }
    return \@sort_orders;
}

#----------------------------------------------------------------------

sub nested_struct_fields {
  my($class) = @_;

  my @nested_struct_fields = grep { $class->fields->{$_}->{nested_struct} } keys %{$class->fields};
  return $class->install_accessor('nested_struct_fields', \@nested_struct_fields);
}
#----------------------------------------------------------------------

sub nested_struct_fields_with_is_inherited {
    my ($class) = @_;
    my $fields = $class->fields;
    my @nested_struct_fields_with_is_inherited = grep { $fields->{$_}->{fields} && $fields->{$_}->{fields}->{is_inherited} } @{$class->nested_struct_fields()};
    return $class->install_accessor('nested_struct_fields_with_is_inherited', \@nested_struct_fields_with_is_inherited);
}

#----------------------------------------------------------------------

sub extern_db_fields {
  my($class) = @_;

  my @extern_db_fields = grep { $class->fields->{$_}->{extern_db_updatable} } keys %{$class->fields};
  return $class->install_accessor('extern_db_fields', \@extern_db_fields);
}

#----------------------------------------------------------------------

sub non_nested_struct_fields {
  my($class) = @_;

  my @non_nested_struct_fields = grep { ! $class->fields->{$_}->{nested_struct} } keys %{$class->fields};
  return $class->install_accessor('non_nested_struct_fields', \@non_nested_struct_fields);
}

#----------------------------------------------------------------------

sub mcp_index_field {
  my ($class) = @_;
  my $fields = $class->fields;
  my ($mcp_index_field) = grep { $fields->{$_}->{mcp_index_field} } keys %$fields;
  return $class->install_accessor('mcp_index_field', $mcp_index_field);
}

#----------------------------------------------------------------------

sub mcp_index_field_name {
  my ($class) = @_;

  my $mcp_index_field_name;
  if (my $index_field = $class->mcp_index_field) {
    $mcp_index_field_name = $class->fields->{$index_field}->{mcp_field_name};
  }
  return $class->install_accessor('mcp_index_field_name', $mcp_index_field_name);
}

#----------------------------------------------------------------------

sub mcp_id_field {
  my($self) = @_;
  my $fields = $self->fields;
  my($mcp_id_field) = grep { $fields->{$_}->{mcp_id_field} } keys %$fields;
  return $self->install_accessor('mcp_id_field', $mcp_id_field);
}

#----------------------------------------------------------------------

sub mcp_id_field_name {
  my($self) = @_;

  my $mcp_id_field_name;
  if (my $id_field = $self->mcp_id_field) {
    $mcp_id_field_name = $self->fields->{$id_field}->{mcp_field_name};
  }
  return $self->install_accessor('mcp_id_field_name', $mcp_id_field_name);
}


#----------------------------------------------------------------------

sub mcp_folder_field {
    my($self) = @_;
    my $fields = $self->fields;
    my($mcp_folder_field) = grep { $fields->{$_}->{mcp_field_name} && $fields->{$_}->{mcp_field_name} eq "folder_name"} keys %$fields;
    return $self->install_accessor('mcp_folder_field', $mcp_folder_field);
}

#----------------------------------------------------------------#

sub mcp_folder {
    my ($class_or_self, %params) = @_;
    my $obj = $params{obj} || $class_or_self;

    if ($class_or_self->mcp_folder_field() && $obj->{$class_or_self->mcp_folder_field()}) {
        return $obj->{$class_or_self->mcp_folder_field()};
    }

    if ($class_or_self->mcp_id_field() && $obj->{$class_or_self->mcp_id_field()}) {
        return dirname($obj->{$class_or_self->mcp_id_field()});
    }

    return;
}


#----------------------------------------------------------------------

sub rest_fields {
    my ($class, %params) = @_;
    my $fields = {%{$class->fields}, %{$class->diff_structs} };
    return $class->install_accessor('rest_fields', $fields);
}

#----------------------------------------------------------------------

sub rest_external_struct_fields {
    my ($class, %params) = @_;
    my @rest_external_struct_fields;
    my $fields = $class->rest_fields(%params);
    for my $field (keys %$fields) {
        my $field_def = $fields->{$field};
        if ($field_def->{rest_value_func} 
            && $field_def->{rest_value_func} eq 'map_rest_external_struct'
            && !$field_def->{dynamic_entity_type}) {
            push (@rest_external_struct_fields, $field);
        }
        elsif ($field_def->{fields}) {
            for my $sub_field (keys %{$field_def->{fields}}) {
                if ($field_def->{fields}->{$sub_field}->{rest_value_func} 
                    && $field_def->{fields}->{$sub_field}->{rest_value_func} eq 'map_rest_external_struct' 
                    && !$field_def->{fields}->{$sub_field}->{dynamic_entity_type}) {
                     push (@rest_external_struct_fields, "${field}__${sub_field}");
                }
            }
        }
    }
    return $class->install_accessor('rest_external_struct_fields', \@rest_external_struct_fields);
}

#----------------------------------------------------------------------

my %common_rest_names = (last_updated => 'lastUpdateMicros',
                         description => 'description',
                         rest_uuid   => 'id',
                        );

sub rest_field_map {
  my($class, %params) = @_;
  my $fields = {%{$class->fields}, %{$class->diff_structs} };
  my @rest_fields = grep { exists($fields->{$_}->{exclude_from_rest}) ? !$fields->{$_}->{exclude_from_rest} 
                                                                      : !$fields->{$_}->{id_field}} keys %{$fields};
  my %rest_field_map;
  for my $field (@rest_fields) {
      my $rest_field_name;
      if ($fields->{$field}->{rest_field_name}) {
          $rest_field_name = $fields->{$field}->{rest_field_name};
      }
      elsif ($common_rest_names{$field}) {
          $rest_field_name = $common_rest_names{$field};
      }
      elsif ($fields->{$field}->{is_name_field}) {
          $rest_field_name = 'name';
      }
      else {
          my $field_display_name = $class->get_field_display_name_string(
                                                                         dbh => $params{dbh},
                                                                         field_name           => $field,
                                                                         full_thrift_obj      => {},
                                                                         );
         #toCamelCase
         $field_display_name = lc($field_display_name);
         $field_display_name = F5::Utils::Rest::to_camel_case($field_display_name);
         $rest_field_name = $field_display_name;
      }
      if (length($rest_field_name)) {
          $rest_field_map{$field} = $rest_field_name;
          if ($fields->{$field}->{rest_group}) {
              $rest_field_map{rest_groups}->{$field} = "$fields->{$field}->{rest_group}/$rest_field_name";
          }
      }
  }
  return $class->install_accessor('rest_field_map', \%rest_field_map);
}


#----------------------------------------------------------------------

sub reverse_rest_field_map {
    my($class, %params) = @_;
    my $rest_field_map = $class->rest_field_map(%params);
    my $rest_groups = delete $rest_field_map->{rest_groups};
    my %reverse_rest_field_map = reverse %$rest_field_map;
    if ($rest_groups) {
        %reverse_rest_field_map = (%reverse_rest_field_map, reverse %$rest_groups);
    }
    return $class->install_accessor('reverse_rest_field_map', \%reverse_rest_field_map);
}

#----------------------------------------------------------------------

sub live_update_fields {
  my($class) = @_;
  my @live_update_fields = grep { $class->fields->{$_}->{live_update_field_name} } keys %{$class->fields};
  return $class->install_accessor('live_update_fields', \@live_update_fields);
}

#----------------------------------------------------------------------

sub live_update_type {
  my($self) = @_;
  return $self->install_accessor('live_update_type');
}

#----------------------------------------------------------------------

sub live_update_write_types {
  my($self) = @_;
  my $write_types = $self->entity_config->{live_update_write_types} || [ $self->live_update_type ];
  return $self->install_accessor('live_update_write_types', $write_types);
}

#----------------------------------------------------------------------

sub live_update_rest_path {
  my($self) = @_;
  return $self->install_accessor('live_update_rest_path');
}

#----------------------------------------------------------------------

sub mcp_fields {
  my($class) = @_;
  my @mcp_fields = grep { $class->fields->{$_}->{mcp_field_name} } keys %{$class->fields};
  return $class->install_accessor('mcp_fields', \@mcp_fields);
}

#----------------------------------------------------------------------

sub mcp_element_type {
  my($self) = @_;
  return $self->install_accessor('mcp_element_type');
}

#----------------------------------------------------------------------

sub is_mcp_object {
    return;
}


#----------------------------------------------------------------------

sub load_mcp_fields {
    my ($self, %params) = @_;
    return unless $self->is_mcp_object();
    return if F5::BigipUtils::is_docker();

    my $obj = $params{mcp_object};
    if (!$obj) {
        my $query = F5::ASM::COAPI::Query->new($self->mcp_element_type, { $self->mcp_id_field_name => $self->{$self->mcp_id_field}},
                                               do_cache => 0 );
        if ($query && @$query) {
            ($obj) = @$query;
        }
    }

    return unless $obj;

    for my $field (@{$self->mcp_fields}) {
        my $mcp_field_name = $self->fields->{$field}->{mcp_field_name};
        if ((!defined($self->{$field}) || !$self->db_table) && defined($obj->{$mcp_field_name})) {
            $self->{$field} = $obj->{$mcp_field_name};
        }
    }
}

#----------------------------------------------------------------------

sub convert_search_criteria_to_unique_record_parts {
    my($self, %params) = @_;
    my $search_criteria = $params{search_criteria};

    my @unique_record_parts = @{$self->unique_record_parts(get_criteria => $self)};
    return unless @unique_record_parts;

    for my $k (keys %$search_criteria) {
        delete $search_criteria->{$k};
    }
    for my $unique_record_part (@unique_record_parts) {
        $search_criteria->{$unique_record_part} = $self->{$unique_record_part};
        if (ref $search_criteria->{$unique_record_part}
            && $self->external_structs->{$unique_record_part}
           ) {
            no strict 'refs';
            my $external_search_criteria = bless( { %{$search_criteria->{$unique_record_part}} }, ref $search_criteria->{$unique_record_part} );
            my $entity = $self->external_structs->{$unique_record_part}->{entity_type}->new(object => $external_search_criteria, dbh => $self->{dbh});
            $entity->convert_search_criteria_to_unique_record_parts(search_criteria => $external_search_criteria);
            $search_criteria->{$unique_record_part} = $external_search_criteria;
        }
    }
}

#----------------------------------------------------------------------

sub populate_unique_record_parts_from_id {
    my ($class_or_self, %params) = @_;
    my $object = $params{object} || $class_or_self;
    my @unique_record_parts = $params{unique_record_parts} ? @{$params{unique_record_parts}}
                                                           : @{$class_or_self->unique_record_parts(get_criteria => $object)};
    if ($class_or_self->id_field && exists ($object->{$class_or_self->id_field}) &&
        (scalar @unique_record_parts != scalar grep { defined $object->{$_} } @unique_record_parts)) {

        my $class = ref($class_or_self) ? ref($class_or_self) : $class_or_self;
        if ( my $retrieved_object = $class->new(dbh => $params{dbh}, get_criteria => $object) ) {
            for my $unique_record_part (@unique_record_parts) {
                if (!defined($object->{$unique_record_part})) {
                    $object->{$unique_record_part} = $retrieved_object->{$unique_record_part};
                }
            }
        }
    }
}

#----------------------------------------------------------------------
sub remove_non_unique_record_parts {
    my ($class_or_self, %params) = @_;
    my $object = $params{object} || $class_or_self;
    my $field_exclusions = $params{field_exclusions} || [];
    my @unique_record_parts = $params{unique_record_parts} ? @{$params{unique_record_parts}}
                                                           : @{$class_or_self->unique_record_parts(get_criteria => $object)};
    return unless @unique_record_parts || @$field_exclusions;
    $class_or_self->populate_unique_record_parts_from_id(%params);

    my %unique_fields = map { ($_ => 1) } @unique_record_parts;
    my %excluded_fields = map { ($_ => 1) } @$field_exclusions;
    for my $k (keys %$object) {
        next if $excluded_fields{$k};
        if (!$unique_fields{$k}) {
            delete $object->{$k};
        }
        elsif (ref $object->{$k} && ((ref $object->{$k}) =~ /::Entity::/)) {
            $object->{$k}->remove_non_unique_record_parts(dbh => $params{dbh});
        }
    }
    return;
}
#----------------------------------------------------------------------
sub purify_store_object {
    my ($class_or_self, %params) = @_;
    my $object = $params{object} || $class_or_self;    
    my $skip_purify_diffs = $params{skip_purify_diffs};

    $class_or_self->populate_unique_record_parts_from_id(%params);

    for my $nested_field (@{$class_or_self->nested_struct_fields(%params)}) {
        if (exists ($object->{$nested_field}) && defined ($object->{$nested_field}) 
            && defined($class_or_self->fields->{$nested_field}->{entity_type})
            && $class_or_self->fields->{$nested_field}->{entity_type} =~ /::Entity::/) {
            for my $nested_object (@{$object->{$nested_field}}) {
                $class_or_self->fields->{$nested_field}->{entity_type}->remove_non_unique_record_parts(object => $nested_object,
                                                                                                       dbh => $params{dbh},);
            }
        }
    }
    for my $external_field (keys %{$class_or_self->external_structs}) {
        my @external_objects;
        if ($class_or_self->external_structs->{$external_field}->{nested_external_struct}) {
            my ($field, $subfield) = split ('/', $external_field);
            if (exists $object->{$field} && defined($object->{$field})) {
                for my $nested_object (@{$object->{$field}}) {
                    if (exists $nested_object->{$subfield} && defined($nested_object->{$subfield})) {
                        push (@external_objects, $nested_object->{$subfield});
                    }
                }
            }
        }
        elsif (exists $object->{$external_field} && defined($object->{$external_field})) {
            @external_objects = ($object->{$external_field});
        }

        for my $external_object (@external_objects) {
            $class_or_self->external_structs->{$external_field}->{entity_type}->remove_non_unique_record_parts(object => $external_object,
                                                                                                               dbh => $params{dbh},);
            $class_or_self->external_structs->{$external_field}->{entity_type}->purify_store_object(object => $external_object,
                                                                                                    dbh => $params{dbh},
                                                                                                    skip_purify_diffs => $skip_purify_diffs,);
        }
    }

    for my $diff_field (keys %{$class_or_self->diff_structs}) {
        if (exists $object->{$diff_field} && defined($object->{$diff_field})) {

             my @diff_objects = $class_or_self->diff_structs->{$diff_field}->{nested_struct} ? @{$object->{$diff_field}} 
                               : ($object->{$diff_field});

            for my $diff_object (@diff_objects) {
                    if (! $skip_purify_diffs) {
                        $class_or_self->diff_structs->{$diff_field}->{entity_type}->remove_non_unique_record_parts(object => $diff_object,
                                                                                                                   dbh => $params{dbh},);
                    }
                    $class_or_self->diff_structs->{$diff_field}->{entity_type}->purify_store_object(object => $diff_object,
                                                                                                    dbh => $params{dbh},
                                                                                                    skip_purify_diffs => $skip_purify_diffs,);
            } 
        }
    }

    if ($class_or_self->id_field && !$class_or_self->fields->{$class_or_self->id_field}->{unique_record_part} && exists ($object->{$class_or_self->id_field})) {
        delete $object->{$class_or_self->id_field};
    }
   
    return $class_or_self;
}
#----------------------------------------------------------------------

sub get_differing_fields {
    my ($class_or_self, %params) = @_;
    my $obj1 = $params{obj1} || $class_or_self;
    my $obj2 = $params{obj2};

    my $entity_config = $class_or_self->entity_config;
    my $diff_structs = $class_or_self->diff_structs;
    my @diff_fields;
    my %specified_diff_fields;

    if ($params{diff_fields}) {
        @diff_fields = @{$params{diff_fields}};
        %specified_diff_fields = map { ( $_ => 1 ) } @diff_fields;
    }
                                        
    if (!@diff_fields) {
        @diff_fields = (@{$obj1->diff_fields()}, keys (%$diff_structs));
    }
    @diff_fields = grep { !(exists($entity_config->{fields}->{$_}) && $entity_config->{fields}->{$_}->{is_dependent})} @diff_fields;    

    if (!$obj2 || !(ref $obj2) || (ref $obj2 ne ref $obj1)) {  
        return @diff_fields;
    }

    $obj1->populate_diff_structs();
    $obj2->populate_diff_structs();

    my %seen_composite_field;
    my @diff_rv;
    while (my $field = shift @diff_fields) {
        my @diff;
        my $field_def = $diff_structs->{$field} || $class_or_self->fields->{$field};

        next if !($specified_diff_fields{$field}) && ($common_diff_excludes->{$field} || $field_def->{exclude_from_diff});
        if (my $exclude_from_diff_func = $field_def->{exclude_from_diff_func}) {
            next if $obj1->$exclude_from_diff_func(field_def => $field_def);
        }

        my $composite_field;
        if ($composite_field = $entity_config->{fields}->{$field}->{composite_field}) {
            if (!exists($seen_composite_field{$composite_field})) {
                $seen_composite_field{$composite_field} = ($obj1->get_field_display_string(field_name => $composite_field) eq $obj2->get_field_display_string(field_name => $composite_field)) ? 1 : 0;
                if (!$seen_composite_field{$composite_field}) {
                    push (@diff_rv, $composite_field); #If it's different, only add it once
                }
            }
            if ($seen_composite_field{$composite_field}) { #If equal, add dependent fields
                 if ($entity_config->{dependent_fields}->{$field} &&
                    ref ($entity_config->{dependent_fields}->{$field}->{$obj1->{$field}})) {
                    push (@diff_fields, @{$entity_config->{dependent_fields}->{$field}->{$obj1->{$field}}});
                }
                next;
            }
        }

        if (defined ($obj1->{$field}) != defined($obj2->{$field})) {          
             if ( (defined ($obj1->{$field}) && length($obj1->{$field})) ||
                  (defined ($obj2->{$field}) && length($obj2->{$field}))) {
                 push (@diff_rv, $field) if !$composite_field;
             }
             next;
        }
        next if (!defined $obj1->{$field});

        if ($field_def->{nested_struct}) {     
                push (@diff, $field);
        }
        elsif ($diff_structs->{$field} && $field_def->{external_struct}) {
            if (defined($obj1->{$field}) ne defined($obj2->{$field})) {
                push (@diff, $field);
            }
            elsif (my @sub_diff_fields = $obj1->{$field}->get_differing_fields(obj2 => $obj2->{$field},
                                                                               diff_fields => $field_def->{diff_fields},
                                                                               )) {
                push (@diff, map { "$field.$_" } @sub_diff_fields);
            }
        }
        elsif (ref($obj1->{$field}) && ref($obj1->{$field}) =~ /::Entity::/) {
            #External structs
            if (my $id_field = $obj1->{$field}->id_field) {
                delete $obj1->{$field}->{$id_field};
                delete $obj2->{$field}->{$id_field};
            }
            if ($obj1->{$field}->differs_from($obj2->{$field})) {
                 push (@diff, $field);
            }
        }
        elsif ($obj1->{$field} ne $obj2->{$field}) {
            push (@diff, $field);
        }
        else {
            #fields match, so check dependent fields
            if ($entity_config->{dependent_fields}->{$field} &&
                ref ($entity_config->{dependent_fields}->{$field}->{$obj1->{$field}})) {
                push (@diff_fields, @{$entity_config->{dependent_fields}->{$field}->{$obj1->{$field}}});
            }
        }

        if (!$composite_field) {
            push (@diff_rv, @diff);
        }
    }

    return @diff_rv;
}


#----------------------------------------------------------------------

sub diff_search_filters {
    return [];
}

#----------------------------------------------------------------------

sub rest_search_filters {
    my ($class) = @_;
    my $search_filters = $class->diff_search_filters();
    if ($class->fields->{feature_name} && !$ENV{F5WAF_ENABLE_ALL_FEATURES}) {
        #Always allow empty string for elements that are not tied to any feature
        my @enabled_features = ('',F5::BigipUtils::waf_enabled_features()); 
        push (@$search_filters, F5::ASMConfig::SearchFilter->new(field_name => 'feature_name',
                                                                 field_value_list => \@enabled_features,
                                                                 field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()));
    }
    return $search_filters;
}

#----------------------------------------------------------------------

sub is_valid_diff_type { 
    return 1;
}

#----------------------------------------------------------------------

sub get_modified_entities {
    my ($class, %params) = @_;
    my $target = $params{target};
    my $source = $params{source};
    my $get_linked_entities = defined($params{get_linked_entities}) ? $params{get_linked_entities}
                                                                    : 1;
    my @fields = $params{fields} ? @{$params{fields}} 
                                 : ();

    my @modified_entities;
    if ($source && !$target) {
        for my $entity (@{$source->get_external_entities()}) {            
            push (@modified_entities, @{$entity->get_modified_entities(source => $entity, get_linked_entities => 0)});
        }

        push (@modified_entities, {type => 'add', entity => $source});

        for my $entity (@{$source->get_diff_entities()}) {
            push (@modified_entities, {type => 'add', entity => $entity});
        }

        if ($get_linked_entities) {
            for my $entity (@{$source->get_linked_entities()}) {
                push (@modified_entities, @{$entity->get_modified_entities(source => $entity)});
            }
        }
    }
    elsif ($target && !$source) {
        for my $entity (@{$target->get_linked_entities()}) {
            push (@modified_entities, @{$entity->get_modified_entities(target => $entity)});
        }
        push (@modified_entities, {type => 'delete', entity => $target});
    }
    elsif ($target && $source) {
        my $entity_config = $class->entity_config;
        my $diff_structs  = $class->diff_structs;

        $source->populate_diff_structs();
        $target->populate_diff_structs();
            
        if (!@fields) {
            @fields = @{$class->diff_fields};
            @fields = grep { !(exists $entity_config->{fields}->{$_} && 
                             $entity_config->{fields}->{$_}->{is_dependent})} (@fields, keys (%$diff_structs));
        }

        my @set_fields_to_null;
        my @external_modified_entities;
        my $self_change_was_made = 0;
        my %composite_field_references;

        if ($entity_config->{get_external_entities_on_diff_conflict}) {
            for my $entity (@{$source->get_external_entities()}) {
                push(@external_modified_entities, @{$entity->get_modified_entities(source => $entity, get_linked_entities => 0)});
            }
        }
    
        while ( my $field = shift @fields) {
            if ($field =~ /\./) {
                my ($struct_name, $sub_field) = split(/\./, $field, 2);
                my $field_def = $diff_structs->{$struct_name} || $target->fields->{$struct_name};

                my @grouped_fields = grep {$_ =~ /^$struct_name\./} @fields;
                @fields = grep {$_ !~ /^$struct_name\./} @fields;

                #Strip out he struct name.
                for my $grouped_field (@grouped_fields) {
                    $grouped_field =~ s/^.*?\.//;
                }
                push (@grouped_fields, $sub_field);  
    
                if ($field_def->{nested_struct}) {
                    my $source_nested_structs = $source->hashify_nested_struct_list(nested_def => $field_def,
                                                                                    nested_recs => $source->{$struct_name});

                    my $target_nested_structs = $target->hashify_nested_struct_list(nested_def => $field_def,
                                                                                    nested_recs => $target->{$struct_name});

                    my @nested_field_names = $field_def->{fields} ? grep { !$field_def->{fields}->{$_}->{unique_record_part}} keys %{$field_def->{fields}}
                                                                  : @{$field_def->{entity_type}->diff_fields};

                    for my $nested_key (@grouped_fields) {
                        if ($diff_structs->{$struct_name}) {
                            push (@modified_entities, @{$field_def->{entity_type}->get_modified_entities(source => $source_nested_structs->{$nested_key},
                                                                                                         target => $target_nested_structs->{$nested_key},
                                                                                                         fields => \@nested_field_names)});
                        }
                        elsif (!$source_nested_structs->{$nested_key}) {
                            delete $target_nested_structs->{$nested_key};
                            $self_change_was_made = 1;
                        }
                        else {
                            if ($field_def->{entity_type} && $field_def->{entity_type} =~ /::Entity::/) {
                                #populate full entity if it only exists in source
                                if ($source_nested_structs->{$nested_key} && !$target_nested_structs->{$nested_key}) {
                                    $source_nested_structs->{$nested_key} = $field_def->{entity_type}->new(dbh => $source_nested_structs->{$nested_key}->{dbh},
                                                                                                           policy => $source_nested_structs->{$nested_key}->{policy}, 
                                                                                                           get_criteria => $source_nested_structs->{$nested_key});
                                }
                                push (@external_modified_entities, @{$field_def->{entity_type}->get_modified_entities(source => $source_nested_structs->{$nested_key},
                                                                                                                      target => $target_nested_structs->{$nested_key},
                                                                                                                      fields => \@nested_field_names)});
                            }
                            elsif (my @external_fields = grep { $field_def->{fields}->{$_}->{external_struct} } keys %{$field_def->{fields}}) {
                                for my $external_field (@external_fields) {
                                    $source_nested_structs->{$nested_key}->{$external_field} = 
                                        $field_def->{fields}->{$external_field}->{entity_type}->new(dbh => $source->{dbh},
                                                                                                    policy => $source->{policy},
                                                                                                    get_criteria => $source_nested_structs->{$nested_key}->{$external_field});
                                     push (@external_modified_entities, @{$field_def->{fields}->{$external_field}->{entity_type}->get_modified_entities(
                                                                                                                    source => $source_nested_structs->{$nested_key}->{$external_field},
                                                                                                                    target => $target_nested_structs->{$nested_key}->{$external_field},
                                                                                                                   )});
                                }
                            }
                            $target_nested_structs->{$nested_key} = $source_nested_structs->{$nested_key};
                            $self_change_was_made = 1;
                        }
                    }

                    $target->{$struct_name} = [ values %$target_nested_structs ];
               }
               elsif ($diff_structs->{$struct_name}) {
                    push (@modified_entities, @{$diff_structs->{$struct_name}->{entity_type}->get_modified_entities(source => $source->{$struct_name},
                                                                                                                    target => $target->{$struct_name}, 
                                                                                                                    fields => \@grouped_fields       ,)});
               }
            }
            elsif ($diff_structs->{$field}) {
                my $field_def = $diff_structs->{$field};

                if ($field_def->{nested_struct}) {
                    my $source_nested_structs = $source->hashify_nested_struct_list(nested_def => $field_def, nested_recs => $source->{$field});
                    my $target_nested_structs = $target->hashify_nested_struct_list(nested_def => $field_def, nested_recs => $target->{$field});
                    my %combined = (%$source_nested_structs, %$target_nested_structs);
                    for my $nested_key (keys %combined) {
                        push (@modified_entities, @{$field_def->{entity_type}->get_modified_entities(source => $source_nested_structs->{$nested_key},
                                                                                                     target => $target_nested_structs->{$nested_key},
                                                                                                     )});
                    }
                }
                else {
                    push (@modified_entities, @{$field_def->{entity_type}->get_modified_entities(source => $source->{$field},
                                                                                                 target => $target->{$field},)});
                }

            }
            elsif (defined $entity_config->{fields}->{$field}->{composite_field}) {
                $self_change_was_made = 1;
                $composite_field_references{$entity_config->{fields}->{$field}->{composite_field}} = 1;
            }
            else {
                if ($entity_config->{dependent_fields}->{$field}) {                 
                    if (ref ($entity_config->{dependent_fields}->{$field}->{$source->{$field}})) {
                        push (@fields, @{$entity_config->{dependent_fields}->{$field}->{$source->{$field}}});
                    }
                    if (ref ($entity_config->{dependent_fields}->{$field}->{$target->{$field}}) &&
                        $target->{$field} ne $source->{$field}) {
                        push (@fields, @{$entity_config->{dependent_fields}->{$field}->{$target->{$field}}});
                    }
                }
                $target->{$field} = $source->{$field};
                if (!defined($target->{$field})) {
                    push (@set_fields_to_null, $field);
                }
                $self_change_was_made = 1;
            }
        }
    
        for my $composite_field (keys %composite_field_references) {
            my @affecting_fields = grep { exists  ($entity_config->{fields}->{$_}) && 
                                          defined ($entity_config->{fields}->{$_}->{composite_field}) && 
                                                   $entity_config->{fields}->{$_}->{composite_field} eq $composite_field } keys %{$entity_config->{fields}};
            for my $field (@affecting_fields) {
                $target->{$field} = $source->{$field};
                if (!defined($target->{$field})) {
                    push (@set_fields_to_null, $field);
                }
            }    
        }
        if ($self_change_was_made) {
            unshift (@modified_entities, {type => 'update', entity => $target, set_fields_to_null => \@set_fields_to_null});
        }
        if (@external_modified_entities) {
            unshift (@modified_entities, @external_modified_entities);
        }
    }
    return \@modified_entities;
}

#----------------------------------------------------------------------

sub external_struct_fields {
  my($class) = @_;

  my @external_struct_fields = grep { $class->fields->{$_}->{external_struct} } keys %{$class->fields};
  return $class->install_accessor('external_struct_fields', \@external_struct_fields);
}

#----------------------------------------------------------------------

sub diff_fields {
  my($class) = @_;

  my @diff_fields = grep { ! $common_diff_excludes->{$_} && 
                           ! $class->fields->{$_}->{exclude_from_diff}  &&
                           ! $class->fields->{$_}->{id_field} } keys %{$class->fields};
  return $class->install_accessor('diff_fields', \@diff_fields);
}

#----------------------------------------------------------------------

sub non_diff_fields {
  my($class) = @_;

  my @non_diff_fields = grep { $common_diff_excludes->{$_} ||
                           $class->all_fields->{$_}->{exclude_from_diff}  ||
                           $class->all_fields->{$_}->{id_field} } keys %{$class->all_fields};
  return $class->install_accessor('non_diff_fields', \@non_diff_fields);
}

#----------------------------------------------------------------------

sub violation_toggle_fields {
   my($class) = @_;

   my %violation_toggle_fields = map { $_ => $class->fields->{$_}->{violation_toggle} } grep { $class->fields->{$_}->{violation_toggle} } keys %{$class->fields};
   return $class->install_accessor('violation_toggle_fields', \%violation_toggle_fields);
}

#----------------------------------------------------------------------

sub virtual_fields {
    my($class) = @_;
    my %virtual_fields = map { $_ => $class->fields->{$_}->{virtual_value_func} } grep { $class->fields->{$_}->{virtual_value_func} } keys %{$class->fields};
    return $class->install_accessor('virtual_fields', \%virtual_fields);
}

#----------------------------------------------------------------------

sub add_pbconfig_event {
    my ($class) = @_;
    return $class->install_accessor('add_pbconfig_event');
}

#----------------------------------------------------------------------

sub remove_non_diff_fields {
    my ($self) = @_;
    
    my @non_diff_fields = @{$self->non_diff_fields()};
    for my $field (@non_diff_fields) {
        delete $self->{$field};
    }
    $self->purify_store_object(dbh => $self->{dbh});
}

#----------------------------------------------------------------------

sub exclude_from_inheritance_fields {
    my($class) = @_;
    my @exclude_from_inheritance_fields = grep { $common_inheritance_excludes->{$_} 
                                              || $class->all_fields->{$_}->{exclude_from_inheritance}
                                              || $class->all_fields->{$_}->{id_field} 
                                              || $class->all_fields->{$_}->{rest_composite_field}
                                              || $class->all_fields->{$_}->{diff_struct} } keys %{$class->all_fields};
    push (@exclude_from_inheritance_fields, qw(policy policy_id __store_obj));

    return $class->install_accessor('exclude_from_inheritance_fields', \@exclude_from_inheritance_fields);
}

#----------------------------------------------------------------------

sub exclude_from_default_table_fields {
    my($class, %params) = @_;
    my $dbh = $params{dbh};
    my $default_table = $params{default_table};

    my %exclude_from_default_table_fields;
    if ( defined ($default_table) ) {
        my $default_table_fields = F5::DbUtils::get_db_fields($dbh, $default_table, undef, 1);
        %exclude_from_default_table_fields = map { ($_ => 1) } grep { !$default_table_fields->{$_} } keys %{$common_default_table_excludes};
    }
    return $class->install_accessor('exclude_from_default_table_fields', \%exclude_from_default_table_fields);
}

#----------------------------------------------------------------------

sub restricted_inheritance_fields {
    my($class) = @_;

    my %exclude_from_inheritance_fields = map { ($_ => 1) } @{$class->exclude_from_inheritance_fields()};
    my @restricted_inheritance_fields = grep { !$exclude_from_inheritance_fields{$_} && !( $class->all_fields->{$_}->{nested_struct}
                                                                                        && $class->all_fields->{$_}->{fields}
                                                                                        && $class->all_fields->{$_}->{fields}->{is_inherited} )}  keys %{$class->all_fields()};

    return $class->install_accessor('restricted_inheritance_fields', \@restricted_inheritance_fields );
}

#----------------------------------------------------------------------

sub remove_non_inheritance_fields {
    my ($self) = @_;

    for my $field (@{$self->exclude_from_inheritance_fields()}) {
        delete $self->{$field};
    }
    $self->purify_store_object(dbh => $self->{dbh});
}

#----------------------------------------------------------------------

sub set_inherited {
    my ($self, %params) = @_;
    $self->{is_inherited} = 1;
    for my $field_name (@{$self->nested_struct_fields_with_is_inherited()}) {
        if (ref $self->{$field_name}) {
            for my $i (0..$#{$self->{$field_name}}) {
                $self->{$field_name}->[$i]->{is_inherited} = 1;
            }
        }
    }
    return $self;
}

#----------------------------------------------------------------------

sub diffable {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    if (defined $entity_config->{diffable}) {
        return $entity_config->{diffable};
    }

    return $class->policy_id_field && 
           ! $class->suppress_events;
}

#----------------------------------------------------------------------

sub relayable {
    my ($class) = @_;
    my $entity_config = $class->entity_config;
    my $relayable = F5::ASMConfig::RelayabilityType::RELAYABILITY_TYPE_IS_RELAYABLE();
    if (defined $entity_config->{relayable}) {
        $relayable = $entity_config->{relayable};
    }

    return $relayable;
}


#----------------------------------------------------------------------

sub get_section_type {
    my ($self) = @_;
    my $section_type;
    my $entity_config = $self->entity_config;
    if (defined $entity_config->{section_type}) {
        $section_type = $entity_config->{section_type};
    }
    elsif ($self->fields->{section_type} && defined($self->{section_type})) {
        $section_type = $self->{section_type};
    }
    elsif ($self->section_type_field() && defined($self->{$self->section_type_field()})) {
        $section_type = $self->{$self->section_type_field()}->{section_type};
    }
    return $section_type;
}

#----------------------------------------------------------------------

sub section_type_field {
    my ($class) = @_;
    my $fields = $class->fields;
    my ($section_type_field) = grep { $fields->{$_}->{section_type_field} } keys %$fields;
    return $class->install_accessor('section_type_field', $section_type_field);
}

#----------------------------------------------------------------------

sub skip_inheritance {
    my ($class) = @_;
    my $entity_config = $class->entity_config();
    my $skip_inheritance = $entity_config->{skip_inheritance} || $entity_config->{no_inheritance};
    return $class->install_accessor('skip_inheritance', $skip_inheritance);
}

#----------------------------------------------------------------------

sub can_update_many_cross_policy {
    my ($class) = @_;
    my $entity_config = $class->entity_config();
    my $can_update_many_cross_policy = defined($entity_config->{can_update_many_cross_policy}) ? $entity_config->{can_update_many_cross_policy}
                                                                                               : 1;
    return $class->install_accessor('can_update_many_cross_policy', $can_update_many_cross_policy);
}

#----------------------------------------------------------------------

sub suppress_events {
    my ($class) = @_;
    return $class->install_accessor('suppress_events');
}

#----------------------------------------------------------------------

#----------------------------------------------------------------------

sub general_validate {
  my ($self, %params) = @_;
  my $field_value = $params{field_value};
  my $field_def = $params{field_def};
  my $dbh = ref($self) ? $self->{dbh} : $params{dbh};
  my $policy = $params{policy};

  if ( !$policy && ref($self) ){
        $policy = $self->{policy};
  }

  if (defined($field_def)) {

      if ($field_def->{validate_non_empty} && ( !length($field_value) ||  !($field_value =~ /\S/))) {
          return {
                  completed_successfully => 0,
                  error_message => 'The value is empty',
                 };
      }

      if (defined $field_value) {
          if ($field_def->{validate_max_len} && length($field_value) > $field_def->{validate_max_len}) {
              return {
                      completed_successfully => 0,
                      error_message => "The value is too long (maximum length = $field_def->{validate_max_len})",
                     };
          }

          if ($field_def->{validate_legal_chars} && $field_value =~ /([^$field_def->{validate_legal_chars}])/) {
              return {
                      completed_successfully => 0,
                      error_message => "The value contains an illegal character (\"$1\")",
                     };
          }

          if ($field_def->{validate_illegal_chars} && $field_value =~ /([$field_def->{validate_illegal_chars}])/) {
              return {
                      completed_successfully => 0,
                      error_message => "The value contains an illegal character (\"$1\")",
                     };
          }

          if ($field_def->{validate_max_val} && $field_value > $field_def->{validate_max_val}) {
              my $validate_max_val = $field_def->{validate_max_val};
              if($params{call_flags}->{rest_call}) {
                  $validate_max_val = $self->map_rest_field_value( %params,
                                                                   field_value => $field_def->{validate_max_val},
                                                                   field_def => $field_def,
                                                                   direction => 'to_rest',
                                                                  );
              }
              return {
                      completed_successfully => 0,
                      error_message => "The value is out of range (maximum value = $validate_max_val)",
                     };
          }

          if (defined $field_def->{validate_min_val} && $field_value < $field_def->{validate_min_val} && 
              ( $field_value || !$field_def->{zero_is_unlimited} ) && ( length($field_value) || !$field_def->{allow_empty_value})) {
              my $validate_min_val = $field_def->{validate_min_val};
              if($params{call_flags}->{rest_call}) {
                  $validate_min_val = $self->map_rest_field_value( %params,
                                                                   field_value => $field_def->{validate_min_val},
                                                                   field_def => $field_def,
                                                                   direction => 'to_rest',
                                                                  );
              }
              return {
                      completed_successfully => 0,
                      error_message => "The value is out of range (minimum value = $validate_min_val)",
                     };
          }

          if ($field_def->{validate_mod_val}) {
              use integer;
              my $remainder = $field_value % $field_def->{validate_mod_val};
              if ($remainder) {
                  return {
                          completed_successfully => 0,
                          error_message => "The value $field_value is illegal (value must be divisible by $field_def->{validate_mod_val})",
                         };
                  }
          }

          if ($field_def->{validate_is_a_flag} && ! ( $field_value == 1 || $field_value == 0 ) ) {
              return {
                      completed_successfully => 0,
                      error_message => "The value is invalid for the flag (0/1)",
                     };
          }

          if ($field_def->{validate_foreign_value}) {
              my $validation_exclude_flg = 0;
              if ($field_def->{validate_foreign_value_excludes} && ref($field_def->{validate_foreign_value_excludes})) {
                  for my $exclude_value (@{$field_def->{validate_foreign_value_excludes}}) {
                      if ($field_value eq $exclude_value) {
                          $validation_exclude_flg = 1;
                          last;
                      }
                  }
              }

              if (! $validation_exclude_flg && ! F5::DbUtils::exists_in_table( dbh => $dbh,
                                                                               field_name => $field_def->{db_field_name},
                                                                               field_value => $field_value, 
                                                                               table_name => $field_def->{validate_foreign_value})) {
                  return {
                          completed_successfully => 0,
                          error_message => "The value is invalid",
                         }; 
              }
          } 

          if ($field_def->{validate_does_not_contain_null} &&  ( $field_value =~ /\0/ ) ) { 
              return {
                      completed_successfully => 0,
                      error_message => "The value must not contain null characters",
                     };

          }
          if ($field_def->{validate_unique_value}  &&  F5::DbUtils::exists_in_table( dbh => $dbh,
                                                                                     policy => $policy,
                                                                                     field_name => $field_def->{db_field_name},
                                                                                     field_value => $field_value,
                                                                                     table_name => $field_def->{validate_unique_value})) {
              return {
                      completed_successfully => 0,
                      error_message => "The value '$field_value' already exists, while the field is a unique field.",                      
                      error_code => F5::ASMConfig::ErrorCode::DUPLICATE_RECORD(),
                     };
          }

          if ($field_def->{validate_unique_host}  &&  !$self->validate_unique_host( dbh => $dbh,
                                                                                   policy => $policy,
                                                                                   field_name => $field_def->{db_field_name},
                                                                                   field_value => $field_value,
                                                                                   table_name => $field_def->{validate_unique_host})) {
              return {
                      completed_successfully => 0,
                      error_message => "The host '$field_value' already exists, while the host must be unique IP or Domain.",
                      error_code => F5::ASMConfig::ErrorCode::DUPLICATE_RECORD(),
                     };
          }

          if ($field_def->{validate_regexp} && !($field_value =~ /$field_def->{validate_regexp}/)) {
              return {
                      completed_successfully => 0,
                      error_message => "The value '$field_value' is not valid for $field_def->{display_field_name}",
                      error_code => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                     };
          }

          if ($field_def->{validate_illegal_regexp} && ($field_value =~ /$field_def->{validate_illegal_regexp}/)) {
              return {
                      completed_successfully => 0,
                      error_message => "The value '$field_value' is not valid for $field_def->{display_field_name}",
                      error_code => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
                     };
          }

          if ($field_def->{validate_valid_lookbehind_regexp} || $field_def->{validate_valid_regexp}) {

              my $pattern = $field_value;
              my $regexp_type_display = 'regular expression';
              if ($field_def->{validate_valid_lookbehind_regexp}) {
                  $pattern = '(?<=' . $pattern . ')';
                  $regexp_type_display = 'look-behind regular expression';
              }
              $pattern =~ s/(?<!\\)\//\\\//g;

              my $regexp_error;
              no warnings 'all';
              no strict;
              eval "my \$temp = qr/$pattern/;";
              if ($@) {
                  $regexp_error = $@;
                  $regexp_error =~ s/ at \(eval .*$//s;

                  return {
                          completed_successfully => 0,
                          error_message => "The value is not valid as a $regexp_type_display ($regexp_error)",
                         };
              }

              if (!F5::BigipUtils::is_docker()) {
                  # PCRE validation
                  my ($tmp_fh, $tmp_filename) = tempfile();
                  print $tmp_fh '"/'.$pattern.'/"';
                  close $tmp_fh;

                  my $res =  F5::SystemUtils::run_system_tool( command_line_params => ["pcretest $tmp_filename"], );
                  unlink($tmp_filename);
    
                  if ($res && $res->{completed_successfully}) {
                      my @regexp_errors = map { substr $_, index($_, 'Failed') } grep { $_ =~ m/Failed/ } (split(/\n/,$res->{standard_output}));
                      if (@regexp_errors) {
                          my $regexp_error = $regexp_errors[0];
                          return {
                                  completed_successfully => 0,
                                  error_message => "PCRE validation failure, The value is not valid as a $regexp_type_display ($regexp_error)",
                                 };
                      }
                  }
              }
          }

          if ($field_def->{validate_mcp_path}) {
                if ($field_value =~ /\/\//) {
                    return {
                              completed_successfully => 0,
                              error_message => "The value contains an invalid path: more than one '/' in a row",
                    };
                }

		if ($field_value =~ /\/$/) {
		    return {
                              completed_successfully => 0,
                              error_message => "The value contains an invalid path: path can not end with '/'",
                    };
                }

                if (!$params{call_flags}->{skip_mcp} && !F5::BigipUtils::is_docker()) { #Don't check this if the call is not going to write to MCP
                    my %mcp_folders = map { ( $_->{name} => $_ ) } @{F5::ASM::COAPI::Query->new('folder')};
                    my $folder = dirname($field_value);
                    if (!$mcp_folders{$folder}) {
                            return {
                                  completed_successfully => 0,
                                  error_message => "The value contains an invalid path: folder does not exist ($folder)",
                            };
                    }
               }
          }
      }
  }

  return {
            completed_successfully => 1,
         };
}

#----------------------------------------------------------------------

sub get_fields_to_set_null {
    my($class_or_self, %params) = @_;
    my %set_fields_to_null;
    if ($params{call_flags} && $params{call_flags}->{set_fields_to_null}) {
        %set_fields_to_null = map { ($_ => 1) } @{$params{call_flags}->{set_fields_to_null}};
    }
    return %set_fields_to_null;
}

#----------------------------------------------------------------------
sub get_field_display_name_string{
    my($class_or_self, %params) = @_;

    my $field_name = $params{field_name};
    my $nested_parent_field_name = $params{nested_parent_field_name};
    my $full_thrift_obj = $params{full_thrift_obj};

    my $field_display_config = $params{field_display_config};
    if (!$field_display_config) {
        if ($nested_parent_field_name) {
            $field_display_config =  $class_or_self->get_field_display_config(field => $nested_parent_field_name,
                                                                              nested_field_name => $field_name,
                                                                              full_thrift_obj   => $full_thrift_obj);

        }
        else {
            $field_display_config = $class_or_self->get_field_display_config(field => $field_name,
                                                                             full_thrift_obj => $full_thrift_obj);
        }
    }

    if ($params{call_flags} && $params{call_flags}->{rest_call} && ($params{call_flags}->{display_rest_names} || !$params{call_flags}->{partition_was_defined}) ) {
        my $rest_field_map = $class_or_self->rest_field_map();
        $nested_parent_field_name ||= $field_name;
        if (my $rest_field_name = ($rest_field_map->{$nested_parent_field_name} || $rest_field_map->{$field_name} )) {
            return $rest_field_name;
        }
    }

    my $display_name_string = $field_display_config->{display_field_name} || '';
    $display_name_string =~ s/\$(\w+)/$class_or_self->get_field_display_string(field_name => $1,
                                                                               field_display_config => $field_display_config->{fields}->{$1},
                                                                               dbh                  => $params{dbh},
                                                                               field_value          => $params{field_value}->{$1},
                                                                               full_thrift_obj      => $full_thrift_obj)/eg;


    return $display_name_string;
}

#----------------------------------------------------------------------

sub apply_sort_order {
    my($class, %params) = @_;
    my $result_set = $params{result_set};

    @$result_set = sort { sort_compare(val1 => $a, val2 => $b, sort_orders => $params{sort_orders}) } @$result_set;
}

#----------------------------------------------------------------------

sub negate_search_filter {
    my($class, %params) = @_;
    my $search_filter = $params{search_filter};
    $search_filter->{field_comparison} = $COMPARISON_NOT_MAPPING{$search_filter->{field_comparison}};
    return $search_filter;
}

#----------------------------------------------------------------------

sub sort_compare {
    my(%params) = @_;
    my $val1 = $params{val1};
    my $val2 = $params{val2};
    my $sort_orders = $params{sort_orders};

    for my $sort_order (@$sort_orders) {
        my $compare_result = 0;

        if (defined $val1->{$sort_order->{field_name}}) {
            if (defined $val2->{$sort_order->{field_name}}) {
                my $binary_sort = $BINARY_SORT_MAPPING{$sort_order->{sort_direction}} || '';
                if ( !$binary_sort &&
                    $val1->{$sort_order->{field_name}} =~ /^-?\d+(\.\d+)?$/ && 
                    $val2->{$sort_order->{field_name}} =~ /^-?\d+(\.\d+)?$/) {

                    $compare_result = ($val1->{$sort_order->{field_name}} <=> $val2->{$sort_order->{field_name}});
                }
                else {
                    $compare_result = ($val1->{$sort_order->{field_name}} cmp $val2->{$sort_order->{field_name}});
                }
            }
            else {
                $compare_result = 1;
            }
        }
        elsif (defined $val2->{$sort_order->{field_name}}) {
            $compare_result = -1;
        }

        if ($sort_order->{sort_direction} == F5::ASMConfig::SortDirection::SORT_DIRECTION_DESC() ||
            $sort_order->{sort_direction} == F5::ASMConfig::SortDirection::SORT_DIRECTION_BINARY_DESC()) {
            $compare_result *= -1;
        }

        if ($compare_result) {
            return $compare_result;
        }
    }

    return 0;
}

#----------------------------------------------------------------------

sub apply_search_filters {
    my($class, %params) = @_;
    my $result_set = $params{result_set};
    my $search_filters = $params{search_filters};
    my $ignore_nonexistent_fields = $params{ignore_nonexistent_fields};

    my $i = 0;
    while($i <= $#$result_set) {
        my $filter_nonmatch_found = 0;

        for my $search_filter (@$search_filters) {
            my $result_ref = $result_set->[$i];
            my $field_name = $search_filter->{field_name};
            while ($field_name =~ /^(.*?)__(.*)$/) {
                $result_ref = $result_ref->{$1};
                $field_name = $2;
            }
            next if ($ignore_nonexistent_fields && !exists($result_ref->{$field_name})); 
            my $rv = $filter_func{$search_filter->{field_comparison}}->(search_filter => $search_filter,
                                                                        field_value => $result_ref->{$field_name});
            if (! $rv) {
                $filter_nonmatch_found = 1;
                last;
            }
        }
        if ($filter_nonmatch_found) {
            splice(@$result_set, $i, 1);
        }
        else {
            $i++;
        }
    }
}

#----------------------------------------------------------------------

sub apply_paging {
    my($class, %params) = @_;
    my $result_set = $params{result_set};
    if ($params{page_size}) {
        $params{start_page} = (defined $params{start_page}) && ($params{start_page} > 0) ? $params{start_page} : 1;
        my $first_record_index = ($params{start_page} - 1) * $params{page_size};
        my $last_record_index = $first_record_index + $params{page_size} - 1;
        if ($last_record_index > $#$result_set) {
            $last_record_index = $#$result_set;
        }
        @$result_set = @{$result_set}[$first_record_index..$last_record_index];
    }
}

#----------------------------------------------------------------------

sub return_empty_result_set {
    my(%params) = @_;
    if ($params{want_array}) {
        return;
    }
    else {
        my $iterator = F5::ASMConfig::Entity::Iterator->new(
                                                            get_next_function => sub {return;},
                                                            get_all_function => sub {return;},
                                                            total_record_count => 0,
                                                           );

        return $iterator;
    }
}

#----------------------------------------------------------------------

sub affects_policy {
    return;
}

#----------------------------------------------------------------------
sub format_ip_with_route_domain{
  my ($self,%params) = @_;
  my $ip_address = $params{field_value};
  my $default_rd = defined $params{default_rd} ? $params{default_rd} : 0;
  my $direction =  $params{direction};

  return if not defined $ip_address;

  if ( $direction eq 'to_db' ){
      if (  $default_rd && ($ip_address !~ /\%/g) ){
         $ip_address = $ip_address."%".$default_rd;
      }
      else
      {
          if ( !$default_rd && ($ip_address =~  /^(.*)\%0$/g ) ){
             $ip_address = $1;
          }
      }
  }
  elsif (  $direction eq 'to_entity' ){
          if( $ip_address  =~ /^(.*)\%([0-9]+)$/g ){
               if ( $2 == $default_rd) {
                   $ip_address = $1;
               }
          }
          else
          {
             if (  $default_rd && ($ip_address !~ /\%/g) ){
                     $ip_address = $ip_address."%0";
              }
          }
  }

  return  $ip_address;
}
#----------------------------------------------------------------------

sub format_host_with_route_domain{
  my ($self,%params) = @_;
  my $host = $params{field_value};
  my $default_rd = defined $params{default_rd} ? $params{default_rd} : 0;
  my $direction =  $params{direction};

  my $rv = $self->normalize_ip(field_value => $host,
                               remove_brackets => 1,
                               field_def => $params{field_def},);

  if ( $rv->{completed_successfully}){
     return $self->format_ip_with_route_domain( field_value => $host,
                                                default_rd => $default_rd,
                                                direction => $direction,
                                              );
  }

  return $host;
}

#----------------------------------------------------------------------

sub format_md5_hex {
    my ($self, %params) = @_;
    return unless $params{field_value};
    return md5_hex($params{field_value});
}

#----------------------------------------------------------------------

sub convert_json_string_to_struct {
    my ($self,%params) = @_;

    if ( $params{direction} eq 'to_entity' ) {
        my $json_string = $params{field_value};
        if ($json_string) {
            return decode_json($json_string);
        }
        else {
            return undef;
        }
    }

    return $params{field_value};
}

#----------------------------------------------------------------------

sub store_as_json {
    my ($self, %params) = @_;

    my $field_value = $params{field_value};
    my $direction = $params{direction};
    my $dbh       = $params{dbh};
    my $policy    = $params{policy};

    return unless defined($field_value);
    if ( $direction eq 'to_entity' ) {
        return JSON::decode_json($params{field_value});
    } else {
        return JSON::encode_json($params{field_value});
    }
}

#----------------------------------------------------------------------

sub rest_selected_fields_to_selected_fields_hash {
    my ($class, %params) = @_;
    my $rest_selected_fields = $params{rest_selected_fields};

    if (!$rest_selected_fields || !%$rest_selected_fields || $rest_selected_fields->{'*'}) {
        return;
    }

    my %rest_field_map = %{$class->rest_field_map()};
    my $reverse_dependent_fields = $class->reverse_dependent_fields();
    my $field_defs = $params{fields} || $class->all_fields;
    my $selected_fields_hash;
    for my $field (keys %rest_field_map) {
       next if $selected_fields_hash->{$field};

       my $field_def = $field_defs->{$field};
       my $rest_expand_name = $field_def->{rest_expand_name} || '';
       my $rest_field_name = $rest_field_map{$field};
       my @fields_to_select;
       if ($rest_selected_fields->{$rest_field_name} || $rest_selected_fields->{$rest_expand_name}) {
           push @fields_to_select, $field;
       }
       elsif ($field_def->{rest_group} && (my $selected_subfields = $rest_selected_fields->{$field_def->{rest_group}})) {
           if (!%$selected_subfields || $selected_subfields->{'*'} || $selected_subfields->{$rest_field_name} || $selected_subfields->{$rest_expand_name}) {
               push @fields_to_select, $field;
           }
       }
       while (my $field_to_select = shift @fields_to_select) {
           next if $selected_fields_hash->{$field_to_select};
           $selected_fields_hash->{$field_to_select} = 1;
           if ($reverse_dependent_fields->{$field_to_select} ) {
               push @fields_to_select, @{$reverse_dependent_fields->{$field_to_select}};
           }
       }
    }
    return $selected_fields_hash;
}


#----------------------------------------------------------------------

sub rest_selected_fields_to_extra_fields {
    my ($class, %params) = @_;
    my $rest_selected_fields = $params{rest_selected_fields};
    my $extra_fields = $params{extra_fields};
    my $field_defs = $params{fields} || $class->fields;
    my %rest_field_map = %{$class->rest_field_map()};

    for my $extra_field (@{$class->extra_fields()}) {
        my $field_def = $field_defs->{$extra_field};
        my $rest_field_name = $rest_field_map{$extra_field};
        next unless $rest_field_name;
        my $expand_name     = $field_def->{rest_expand_name} || $rest_field_name;
        if ($rest_selected_fields->{$rest_field_name} || $rest_selected_fields->{$expand_name}) {
            push (@$extra_fields, $extra_field);
        }
        elsif ($field_def->{rest_group} && $rest_selected_fields->{$field_def->{rest_group}}
               && ($rest_selected_fields->{$field_def->{rest_group}}->{$rest_field_name}
                   || $rest_selected_fields->{$field_def->{rest_group}}->{$expand_name}) ) {

            push (@$extra_fields, $extra_field);
        }
    }
    return $extra_fields;
}

#----------------------------------------------------------------------

sub convert_rest_orders {
    my ($class, %params) = @_;

    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $rest_orders = $params{rest_orders};

    my @sort_orders;
    my $fields = {%{$class->fields}, %{$class->diff_structs} };
    my %reverse_rest_field_map = %{$params{reverse_rest_field_map} || $class->reverse_rest_field_map()};

    for my $order (@$rest_orders) {
        my $field = $order->{rest_field};
        my $sort_field;
        if ($sort_field = $reverse_rest_field_map{$order->{full_field}}) {
            push (@sort_orders, F5::ASMConfig::SortOrder->new({field_name => $sort_field,
                                                               sort_direction => $order->{direction}}));
        }
        elsif ( ($sort_field = $reverse_rest_field_map{$field}) && $fields->{$sort_field}
                 && ($fields->{$sort_field}->{external_struct} || $fields->{$sort_field}->{rest_entity_type}) ) {
            my $nested_field = $order->{nested_fields};
            my $nested_class = $fields->{$sort_field}->{rest_entity_type} || $fields->{$sort_field}->{entity_type};
            my %nested_fields = %{$nested_class->fields};
            my %nested_reverse_rest_field_map = %{$nested_class->reverse_rest_field_map()};
            if (my $nested_entity_field = $nested_reverse_rest_field_map{$nested_field}) {
                push (@sort_orders, F5::ASMConfig::SortOrder->new({field_name => "${sort_field}__$nested_entity_field",
                                                                   sort_direction => $order->{direction}}));
            }
        }
    }
    return \@sort_orders;
}

#----------------------------------------------------------------------

my %rest_filter_comparison_map = (
                 'eq' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_EQ(),
                 'ne' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_BINARY_NE(),
                 'gt' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GT(),
                 'ge' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_GE(),
                 'lt' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LT(),
                 'le' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LE(),
                 'in' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                 'like' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE(),
                 'not like' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE(),
                 'not in' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN(),
                 'exists' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EXISTS(),
                 'not exists' => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_EXISTS(),
             );

my %complex_filters = ( 'staging_or_tightening' => 1,
                        'staging_or_tightening_past_period' => 1,
                        'staging_or_tightening_has_suggestions' => 1,
                      );

sub convert_rest_filter_string {
    my ($class_or_self, %params) = @_;
    my $call_flags = $params{call_flags};
    my $filter_string = $params{filter_string};

    my $filters;
    eval {
        $filters = F5::Utils::Rest::parse_odata_filter(filter => $filter_string);
    };
    if ($@) {
        die {
             error_code   => F5::ASMConfig::ErrorCode::INVALID_FIELD_FORMAT(),
             error_message => $@,
             error_level  => "info",
            };
    }

    my $field_filters;
    for my $filter (@$filters) {
        my $value = defined($filter->{value_list}) ? $filter->{value_list} : $filter->{value};
        push (@{$field_filters->{$filter->{field}}}, { full_field => $filter->{field},
                                                       'cmp' => $filter->{'cmp'},
                                                       value => $value,
                                                       rest_field => $filter->{field}, });
    }

    return $class_or_self->rest_filters_to_search_filters(%params, rest_filters_by_field => $field_filters);
}

sub rest_filters_to_search_filters {
    my ($class, %params) = @_;

    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $rest_filters_by_field = $params{rest_filters_by_field};

    my @search_filters;
    
    my $fields = {%{$class->fields}, %{$class->diff_structs} };
    my %reverse_rest_field_map = %{$class->reverse_rest_field_map()};

    for my $field (keys %$rest_filters_by_field) {
        next if $field =~ /Base64Encoded$/;

        my $nested_entity_type;
        my $group_by;
        my $foreign_grouping_field;
        my @nested_search_filters;
        my $is_external_struct;
        my $nested_def;
        my $entity_field;

        my @filters = @{$rest_filters_by_field->{$field}};
        my $base64encoded;
        my %extra_fields;
        if ($rest_filters_by_field->{$field . "Base64Encoded"}) {
            my ($filter) = @{$rest_filters_by_field->{$field . "Base64Encoded"}};
            #Assuming if they are sending in xBase64Encoded it's only eq/ne
            if (($filter->{value} eq JSON::true && $filter->{'cmp'} eq 'eq')
                || ($filter->{value} eq JSON::false && $filter->{'cmp'} eq 'ne')) {
                $base64encoded = 1;
            }
        }
        for my $filter (@filters) {
            if ($complex_filters{$field}) {
                push (@search_filters, F5::ASMConfig::SearchFilter->new(field_comparison => $rest_filter_comparison_map{$filter->{'cmp'}},
                                                                        field_value => (($filter->{value} eq JSON::true) ? 1 : 0),
                                                                        field_name => $field,));
                next;
            }
            $entity_field = $reverse_rest_field_map{$field};

            if ($filter->{'cmp'} =~ /exists/) {
                if ($entity_field) {
                    my $field_comparison = $rest_filter_comparison_map{$filter->{'cmp'}};
                    if (! $fields->{$entity_field}->{nested_struct}) {
                        $field_comparison = ($filter->{'cmp'} eq 'exists') ? F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NOT_NULL()
                                                                           : F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IS_NULL();
                    }
                    elsif ($fields->{$entity_field}->{diff_struct} && $fields->{$entity_field}->{group_by}) {
                        my $nested_entity_type = $fields->{$entity_field}->{entity_type};
                        my $foreign_grouping_field = $fields->{$entity_field}->{group_by};
                        my $group_by = $class->unique_id_field;
                        $field_comparison = ($filter->{'cmp'} eq 'exists') ? F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN()
                                                                           : F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN();

                        my @nested_entities = $nested_entity_type->retrieve_all(dbh => $dbh,
                                                                                policy => $policy,
                                                                                grouping_fields => [$foreign_grouping_field],
                                                                                selected_fields => [$foreign_grouping_field],
                                                                                return_as_raw_db_hashref => 1,
                                                                               );

                        my @values = map { $_->{$foreign_grouping_field} } @nested_entities;
                        next if (!@values && $field_comparison == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN());
                        push (@search_filters, F5::ASMConfig::SearchFilter->new(field_name => $group_by,
                                                                                field_comparison => $field_comparison,
                                                                                field_value_list => \@values,
                                                                                derived_filter => 1,));
                        next;
                    }

                    push (@search_filters, F5::ASMConfig::SearchFilter->new(field_comparison => $field_comparison,
                                                                            field_name => $entity_field,));
                }
                next;
            }

            if ($base64encoded) {
                my $is_array = ref $filter->{value} eq 'ARRAY';
                my @filter_values = $is_array ? @{$filter->{value}} : ($filter->{value});
                my @values;
                for my $value (@filter_values) {
                    if ($filter->{'cmp'} eq 'like' || $filter->{'cmp'} eq 'not like') {
                        $value =~ s/([^%_]+)/F5::Utils::Encode::decode_base64($1)/eg;
                    }
                    else {
                        $value = F5::Utils::Encode::decode_base64($value);
                    }
                    push (@values, $value);
                }
                
                ($filter->{value}) = $is_array ? \@values : @values;
            }

            my $leaf_field = $field;
            my $filter_field;
            if (!$entity_field && $filter->{nested_fields}) {
                my ($group_field, $remainder) = split('/', $filter->{nested_fields}, 2);
                if ($entity_field = $reverse_rest_field_map{"$field/$group_field"}) {
                    $leaf_field = $group_field;
                }
            }

            my ($search_value_rest_obj) = (ref $filter->{value} eq 'ARRAY') ? @{$filter->{value}} 
                                                                            : $filter->{value};
            if ($filter->{nested_fields}) {
                for my $part (reverse (split('/', $filter->{nested_fields}))) {
                    if (!$filter_field) {
                        $filter_field = $part;
                    }

                    if ($part eq $leaf_field && $entity_field && $fields->{$entity_field}->{nested_struct} && !$fields->{$entity_field}->{map_rest_nested_as_single} ) {
                        $search_value_rest_obj = [$search_value_rest_obj];
                    }
                    $search_value_rest_obj = { $part => $search_value_rest_obj };
                }                
            }
            else {
                $filter_field = $field;
            }

            my $nested_search_value_rest_obj = $search_value_rest_obj;
            if ($leaf_field eq $field && $entity_field && $fields->{$entity_field}->{nested_struct} && !$fields->{$entity_field}->{map_rest_nested_as_single} ) {
                $search_value_rest_obj = [$search_value_rest_obj];
            }

            $search_value_rest_obj = { %extra_fields, $filter->{rest_field} => $search_value_rest_obj };

            if ( $entity_field && $fields->{$entity_field}->{nested_struct} && ( (scalar @filters > 1 && $filter->{is_grouped}) || $fields->{$entity_field}->{map_rest_nested_as_single} ) ) {
                $nested_def = $fields->{$entity_field};
                $nested_entity_type = $nested_def->{rest_entity_type} || $nested_def->{entity_type};
                $is_external_struct = ($nested_entity_type =~ /::Entity::/);
                $nested_entity_type = $class if !$is_external_struct;
            }

            my $entity_type = $nested_entity_type ? $nested_entity_type : $class;
            my $entity = $entity_type->new(dbh => $dbh,
                                           policy => $policy,
                                           object => {},)->from_rest(object => ( $nested_entity_type && $is_external_struct ) ? $nested_search_value_rest_obj
                                                                                                                              : $search_value_rest_obj,
                                                                     skip_retrieve => 1,
                                                                    );            
            my @filters_rv = @{ $entity->to_search_filters(field_comparison => $rest_filter_comparison_map{$filter->{'cmp'}},
                                                           filter_value => $filter->{value},
                                                           collective_nested_field => ( $nested_entity_type && !$is_external_struct ),
                                                           filter_field => $filter_field,)
                              };
            
            if ( $nested_entity_type ) {
                push (@nested_search_filters, @filters_rv);
            }
            else {
                push (@search_filters, @filters_rv);
            }
        }

        if ($nested_entity_type) {
            # TODO this can be an issue in case of rest group with several nested fields that have different grouping.
            # in that case we'll have to calculate grouping for each filter.
            if ( $is_external_struct ) {
                if ( $nested_def->{foreign_id_field} || ($nested_def->{db_table_name} && $nested_def->{fields})) {
                    $foreign_grouping_field = $nested_entity_type->unique_id_field;
                    if (! $nested_def->{fields}->{$foreign_grouping_field}) {
                        ($foreign_grouping_field) = grep { $nested_def->{fields}->{$_}->{unique_record_part} } keys %{$nested_def->{fields}};
                    }
                    if (!$foreign_grouping_field) { # like Suggestion - support_id_list
                        ($foreign_grouping_field) = grep { $nested_def->{fields}->{$_}->{db_field_name} } keys %{$nested_def->{fields}};
                    }
                    $group_by = "${entity_field}__$foreign_grouping_field";
                }
                elsif ( $nested_def->{group_by} ) {
                    $foreign_grouping_field = $nested_def->{group_by};
                    $group_by = $class->unique_id_field;
                }
                else {
                    # in this case there is no junction table and the referenced entities are globals(policy level or higher)
                    $foreign_grouping_field = $class->policy_id_field() || $nested_entity_type->unique_id_field;
                    $group_by = $foreign_grouping_field;
                }
            }
            else {
                $group_by = $nested_def->{parent_group_by} || (($nested_def->{foreign_field} || $nested_def->{foreign_id_field}) ? $class->id_field : 'policy_id') || $field;
                $foreign_grouping_field = $group_by;
            }

            my @nested_entities = $nested_entity_type->retrieve_all(dbh => $dbh,
                                                                    policy => $policy,
                                                                    search_filters => \@nested_search_filters,
                                                                    grouping_fields => [$foreign_grouping_field],
                                                                    selected_fields => [$foreign_grouping_field],
                                                                    return_as_raw_db_hashref => 1,
                                                                   );

            my @values = map { $_->{$foreign_grouping_field} } @nested_entities;
            push (@search_filters, F5::ASMConfig::SearchFilter->new(field_name => $group_by,
                                                                    field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                                    field_value_list => \@values,
                                                                    derived_filter => 1,));
        }
    }
    return \@search_filters;
}

#----------------------------------------------------------------------

sub to_search_filters {
    my ($self, %params) = @_;

    my $field_comparison = $params{field_comparison} || F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ();
    my $fields = {%{$self->fields}, %{$self->diff_structs} };
    my @search_filters;
    my %reverse_rest_field_map = %{$self->reverse_rest_field_map()};
    my $collective_nested_field =  $params{collective_nested_field};
    for my $field (keys %$fields) {
        if (defined($self->{$field})) {

            my $field_def = $fields->{$field};
            my $is_nested_struct = $field_def->{nested_struct};
            my $value;
            if ($is_nested_struct) {
                ($value) = @{$self->{$field}};
            }
            else {
                $value = $self->{$field};
            }

            my $is_external_struct = $field_def->{external_struct} || ($field_def->{rest_entity_type} && ref $value);

            if ($is_external_struct || ( ref($field_def->{fields}) eq 'HASH' && %{$field_def->{fields}}) ) {
                my $group_by = $field_def->{parent_group_by} || $params{group_by} || (($field_def->{foreign_field} || $field_def->{foreign_id_field}) ? $self->id_field : 'policy_id');
                $group_by ||= $field; #in case of no id_field (PolicySignatureSet)
                my $foreign_grouping_field = 'policy_id';
                my $entity_type;
                my $nested_search_filters = [];
                if ($is_external_struct || ($field_def->{entity_type} && $field_def->{entity_type} =~ /::Entity::/)) {
                    my %nested_reverse_rest_field_map = %{$value->reverse_rest_field_map()};
                    my $filter_field = $nested_reverse_rest_field_map{$params{filter_field}} || $reverse_rest_field_map{$params{filter_field}} || $params{filter_field};
                    $nested_search_filters = $value->to_search_filters(%params,
                                                                       filter_field => $filter_field,
                                                                       is_nested_struct_filter => $is_nested_struct);
                    if ($field_def->{group_by} || $field_def->{foreign_id_field} || $field_def->{foreign_field} || $params{group_by}) {
                       $foreign_grouping_field = $field_def->{group_by} || $field_def->{foreign_id_field} || $field_def->{foreign_field} || $params{group_by};
                    }
                    elsif ($field_def->{rest_entity_type}) {
                        $foreign_grouping_field = $group_by = $field;
                        if ($field_def->{fields}) {
                            ($foreign_grouping_field) = keys %{$field_def->{fields}};
                        }
                    }

                    $entity_type = $field_def->{rest_entity_type} || $field_def->{entity_type};

                    if ($is_nested_struct) {
                        if ( $field_def->{foreign_id_field} || ($field_def->{db_table_name} && $field_def->{fields}) ) {
                            $foreign_grouping_field = $entity_type->unique_id_field;
                            if (! $field_def->{fields}->{$foreign_grouping_field}) {
                                ($foreign_grouping_field) = grep { $field_def->{fields}->{$_}->{unique_record_part} } keys %{$field_def->{fields}};
                            }
                            if (!$foreign_grouping_field) {
                                ($foreign_grouping_field) = grep { $field_def->{fields}->{$_}->{db_field_name} } keys %{$field_def->{fields}};
                            }
                            $group_by = "${field}__$foreign_grouping_field";
                        }
                        elsif ( $field_def->{group_by} || $params{group_by} ) {
                            $foreign_grouping_field = $field_def->{group_by} || $params{group_by};
                            $group_by = $self->unique_id_field;
                        }
                        else {
                            # in this case there is no junction table and the referenced entities are globals(policy level or higher)
                            $foreign_grouping_field = $self->policy_id_field() || $entity_type->unique_id_field;
                            $group_by = $foreign_grouping_field;
                        }
                    }
                }
                else {
                    for my $nested_field (keys %{$field_def->{fields}}) {
                        if (defined $value && ( ! ref($value) || ! (reftype($value) eq 'HASH')  || defined $value->{$nested_field})) {
                            my $external_entity_type = $field_def->{fields}->{$nested_field}->{rest_entity_type} || $field_def->{fields}->{$nested_field}->{entity_type};
                            if ($external_entity_type && $external_entity_type =~ /::Entity::/) {
                               my %nested_reverse_rest_field_map = %{$value->{$nested_field}->reverse_rest_field_map()};
                               my $filter_field = $nested_reverse_rest_field_map{$params{filter_field}} || $reverse_rest_field_map{$params{filter_field}} || $params{filter_field};

                               my $external_search_filters = $value->{$nested_field}->to_search_filters(%params,
                                                                                                        filter_field => $filter_field,
                                                                                                        is_nested_struct_filter => $is_nested_struct);

                                my @external_entities = $external_entity_type->retrieve_all(%params,
                                                                                            dbh => $self->{dbh},
                                                                                            policy => $self->{policy},
                                                                                            search_filters => $external_search_filters,
                                                                                            minimal_load => 1, );

                                my @external_values = map { $_->{$external_entity_type->id_field} } @external_entities;



                                my $nested_field_comparison = F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN();
                                if ($field_comparison == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE() || $field_comparison == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN()) {
                                    next unless @external_values;
                                    $nested_field_comparison = F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN();
                                }

                                push (@$nested_search_filters, F5::ASMConfig::SearchFilter->new(field_name => "${field}__$nested_field",
                                                                                                collective_nested_field => $collective_nested_field,
                                                                                                field_comparison => $nested_field_comparison,
                                                                                                field_value_list => \@external_values,
                                                                                                derived_filter => 1,));
                            }
                            else {
                                my ($filter_field) = grep {  $field_def->{fields}->{$_}->{rest_field_name} && $field_def->{fields}->{$_}->{rest_field_name} eq $params{filter_field} } keys %{$field_def->{fields}};
                                $filter_field ||= $params{filter_field};

                                if ($filter_field && $nested_field eq $filter_field && (ref($params{filter_value}) eq 'ARRAY')) {
                                    my @values = map { $self->map_rest_field_value( field_value => $_,
                                                                                    field_name => $nested_field,
                                                                                    field_def => $field_def->{fields}->{$nested_field},
                                                                                    rest_field_name => $params{filter_field},
                                                                                    policy => $self->policy,
                                                                                    direction => 'from_rest',
                                                                                    dbh => $self->{dbh},) } @{$params{filter_value}};
                                    push (@$nested_search_filters, F5::ASMConfig::SearchFilter->new(field_name => "${field}__$nested_field",
                                                                                                    field_comparison => $field_comparison,
                                                                                                    collective_nested_field => $collective_nested_field,
                                                                                                    field_value_list => \@values,
                                                                                                    derived_filter => 1,));
                                }
                                else {
                                    push (@$nested_search_filters, F5::ASMConfig::SearchFilter->new(field_name => "${field}__$nested_field",
                                                                                                    collective_nested_field => $collective_nested_field,
                                                                                                    field_comparison => $field_comparison,
                                                                                                    field_value => (ref($value) && reftype($value) eq 'HASH') ? $value->{$nested_field} : $value,
                                                                                                    derived_filter => 1,));
                                }
                            }
                        }
                        $foreign_grouping_field = $group_by;
                    }

                    return $nested_search_filters if $collective_nested_field;

                    $entity_type = ref $self;
                }

                # A field with '__' in the name indicates it is a nested or external struct field and we cannot do a minimal load
                my $minimal_load = (index($foreign_grouping_field, '__') > 0) ? 1 : 0;
                my @nested_entities = $entity_type->retrieve_all(%params,
                                                                 dbh => $self->{dbh},
                                                                 policy => $self->{policy},
                                                                 grouping_fields => [$foreign_grouping_field],
                                                                 search_filters => $nested_search_filters,
                                                                 selected_fields => [$foreign_grouping_field],
                                                                 minimal_load => $minimal_load,
                                                                 return_as_raw_db_hashref => 1,
                                                                );

                my @values = map { $_->{$foreign_grouping_field} } @nested_entities;
                my $nested_field_comparison = F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN();

                if ($field_comparison == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE() && $is_nested_struct) {
                    next unless @values;
                    $nested_field_comparison = F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN();
                }

                push (@search_filters, F5::ASMConfig::SearchFilter->new(field_name => $group_by,
                                                                        field_comparison => $nested_field_comparison,
                                                                        field_value_list => \@values,
                                                                        derived_filter => 1,));
            }
            else {
                my $search_field_comparison = $field_comparison;
                my %override_args;                
                if (($is_nested_struct || $params{is_nested_struct_filter}) && $field_comparison == F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE()) {
                     $search_field_comparison = F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ();
                }

                my $filter_field = $reverse_rest_field_map{$params{filter_field}} || $params{filter_field};
                if ($filter_field && $field eq $filter_field && (ref($params{filter_value}) eq 'ARRAY')) {
                    my @values = map { $self->map_rest_field_value( field_value => $_,
                                                                    field_name => $field,
                                                                    field_def => $field_def,
                                                                    rest_field_name => $params{filter_field},
                                                                    policy => $self->policy,
                                                                    direction => 'from_rest',
                                                                    dbh => $self->{dbh},) } @{$params{filter_value}};

                    push (@search_filters, F5::ASMConfig::SearchFilter->new(field_name => $field,
                                                                            field_comparison => $field_comparison,
                                                                            field_value_list => \@values ));
                }
                else {
                    push (@search_filters, F5::ASMConfig::SearchFilter->new(field_name => $field,
                                                                            field_comparison => $search_field_comparison,
                                                                            field_value => $value,
                                                                            %override_args));
                }
            }

            if ($field_def->{where_values}) {
                for my $where_field (keys %{$field_def->{where_values}}) {
                    push (@search_filters, F5::ASMConfig::SearchFilter->new(field_name => $where_field,
                                                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                                                            field_value => $field_def->{where_values}->{$where_field},));
                }
            }
        }
    }
    return \@search_filters;
}

#----------------------------------------------------------------------

sub remove_undefined_fields {
    my ($self) = @_;
    my $fields = $self->fields;
    for my $field (keys %{$self}) {
        next if $non_data_fields{$field};
        if (!$fields->{$field} || !defined($self->{$field})) {
            delete $self->{$field};
        } 
        elsif (ref $self->{$field} && (ref $self->{$field} =~ /ASMConfig/)) {
            $self->{$field}->remove_undefined_fields();
        }
    }
    return $self;
}

#----------------------------------------------------------------------

sub self_link {
    my ($self, %params) = @_;
    my $link = $params{base_link} || '';
    my $policy = $params{policy} || $self->{policy};
    my $dbh = $params{dbh} || $self->{dbh};
    my @unique_record_parts = @{$self->unique_record_parts(get_criteria => $self)};
    if ($policy && $self->policy_id_field()) {
        if (!$policy->{rest_uuid} || $policy->{rest_uuid} ne 'example') {
            $policy = F5::ASMConfig::Cache::get_cached_policy(policy => $policy,
                                                              dbh => $dbh);
        }
        if ($policy->{rest_uuid}) {
            $link .= "/policies/$policy->{rest_uuid}";
        }
    }

    for my $unique_record_part (@unique_record_parts) {
        if (ref $self->{$unique_record_part}
           && $self->external_structs->{$unique_record_part}
           && !$self->external_structs->{$unique_record_part}->{rest_unique_record_part}
          ) {
            $link = $self->{$unique_record_part}->self_link(%params,
                                                            ver => '',
                                                            policy => $policy);
        }
    }

    $link .= '/' . $self->element_rest_path_name();
    if ($self->fields->{rest_uuid}) {
        $self->{rest_uuid} ||= $self->get_rest_uuid(%params);
        $link .= "/" . $self->{rest_uuid};
    }
    $link = $self->add_ver_to_self_link(%params, 'link' => $link);
    return $link;
}
#----------------------------------------------------------------------

sub add_ver_to_self_link {
    my ($self, %params) = @_;

    my $link = $params{'link'};
    my $ver = defined($params{ver}) ? $params{ver}
                                    : $params{call_flags}->{rest_version};
    if ($ver) {
        $link .= "?ver=$ver";
    }   
    return $link;
}

#----------------------------------------------------------------------

sub rest_path_to_kind {
    my ($self, %params) = @_;
    my $rest_path = $params{rest_path};
    my $kind = "tm:asm:$rest_path:";
    $kind =~ s/\//:/g;
    $kind .= $self->element_rest_name() . "state";
    return $kind;
}

#----------------------------------------------------------------------

sub get_kind {
    my ($self, %params) = @_;
    my $entity_config = $self->entity_config;
    my $kind;
    if ($entity_config->{rest_paths}) {
        #Take the first path as cannonical (all entites only have one endpoint now except for Parameters currently)
        my ($rest_paths) = @{$entity_config->{rest_paths}};
        $kind = $self->rest_path_to_kind(rest_path => $rest_paths->{rest_path});
    }
    return $self->install_accessor('get_kind', $kind);
}

#----------------------------------------------------------------------

sub safe_encode {
    my($self, %params) = @_;
    my $value = $params{value};
    my $from = $params{from};
    my $to = $params{to};

    if ($from eq 'utf-8' && $value =~ /\w/) {
        $value = Encode::encode('utf-8', $value);
    }
    if ($from && $to) {
        Encode::from_to($value, $from, $to);
    }
    return $value;
}

#----------------------------------------------------------------------

sub get_sig_ids {
    my ($self, %params) = @_;
    my @sig_ids;
    #For suggestion grouping
    my $fields = $self->all_fields(%params);
    my $signature_fields = $self->signature_fields(%params);
    for my $signature_field (@$signature_fields) {
        if ($self->{$signature_field}) {            
            if ($fields->{$signature_field}->{nested_struct}) {
                push (@sig_ids, map { $_->{sig_id} } @{$self->{$signature_field}});
            }
            else {
                push (@sig_ids, $self->{$signature_field});
            }
        }
    }
    return @sig_ids;
}

#----------------------------------------------------------------------

sub get_metachars {
    my ($self, %params) = @_;
    #For suggestion grouping
    my @metachars;
    #For suggestion grouping
    my $fields = $self->all_fields(%params);
    my $metachar_fields = $self->metachar_fields(%params);
    for my $metachar_field (@$metachar_fields) {
        if ($self->{$metachar_field}) {
            if ($fields->{$metachar_field}->{nested_struct}) {
                push (@metachars, map { $_->{metachar} } @{$self->{$metachar_field}});
            }
            else {
                push (@metachars, $self->{$metachar_field});
            }
        }
    }
    return @metachars;
}

#----------------------------------------------------------------------

sub TO_JSON {
    my ($self) = @_;
    my $rv = {%$self};
    for my $field (keys %non_data_fields) {
        delete $rv->{$field};
    }
    return $rv;
}

#----------------------------------------------------------------------

sub to_xml {
    my($self, %params) = @_;

    my $twig = XML::Twig::Elt->new($self->element_type_name);

    my $fields = $self->fields;

    for my $field_name (sort keys %$fields) {
        my $field = $fields->{$field_name};

        my $xml_field_name = $field->{xml_field_name};
        if (! $xml_field_name) {
            next;
        }

        my $field_value = $self->map_xml_field_value (
                                                      field_value => $self->{$field_name},
                                                      field_name => $field_name,
                                                      field_def => $fields->{$field_name},
                                                      direction => 'to_xml',
                                                      dbh => $self->{dbh},
                                                    );


        if (! defined $field_value) {
            next;
        }

        if ($xml_field_name =~ /^\@(.*)$/) {
            my $attrib_name = $1;
            $twig->set_att($attrib_name => $field_value);
        }
        else {
            my $elt = XML::Twig::Elt->new($xml_field_name);
            if (defined $field_value) {
                $elt->set_text($field_value);
                $elt->paste(first_child => $twig);
            }
        }
    }

    $twig->set_pretty_print('indented');
    return $twig;
}

#----------------------------------------------------------------------

sub convert_by_value_db_map {
    my($self, %params) = @_;

    my $map_def = $params{map_def};
    my $field_value = $params{field_value};

    if (! $map_def->{xml_map_cache}) {
        my $value_map = $self->{dbh}->selectcol_arrayref(<<EndQuery, { Columns => [1,2] });
SELECT $map_def->{lookup_field}, $map_def->{value_field}
  FROM $map_def->{table_name}
EndQuery
        $map_def->{xml_map_cache} = { @$value_map };
    }

    if (! $map_def->{reverse_xml_map_cache}) {
        $map_def->{reverse_xml_map_cache} = { reverse %{$map_def->{xml_map_cache}} };
    }

    if ($params{mode} eq 'to_xml') {
        if (exists $map_def->{xml_map_cache}->{$field_value}) {
            return $map_def->{xml_map_cache}->{$field_value};
        }
    }
    elsif ($params{mode} eq 'from_xml') {
        if (exists $map_def->{reverse_xml_map_cache}->{$field_value}) {
            return $map_def->{reverse_xml_map_cache}->{$field_value};
        }
    }

    return $field_value;
}

#----------------------------------------------------------------------

sub convert_by_value_type_date {
    my($self, %params) = @_;
    my $field_value = $params{field_value};

    if ($params{mode} eq 'to_xml') {
        return F5::DateTimeUtils::convert_epoch_to_datetime($field_value);
    }
    elsif ($params{mode} eq 'from_xml') {
        return F5::DateTimeUtils::convert_datetime_to_epoch($field_value);
    }
}

#----------------------------------------------------------------------

sub convert_by_value_type_timestamp {
    my($self, %params) = @_;
    my $field_value = $params{field_value};

    if ($params{mode} eq 'to_xml') {
        return F5::DateTimeUtils::convert_epoch_to_rfc3339($field_value);
    }
    elsif ($params{mode} eq 'from_xml') {
        return F5::DateTimeUtils::convert_rfc3339_to_epoch($field_value);
    }
}

#----------------------------------------------------------------------

sub from_xml {
    my($self, %params) = @_;

    # Assume xml is a twig for now
    my $twig = $params{xml};
    my $xml_rec = $twig->simplify;
    for my $k (keys %{$twig->{att}}) {
        $xml_rec->{'@' . $k} = $twig->{att}->{$k};
    }
    my $rec = {};

    my $fields = $self->fields;

    for my $field_name (keys %$fields) {
        my $field_def = $fields->{$field_name};
        next if $field_def->{exclude_from_xml};
        my $xml_field_name = $field_def->{xml_field_name} || $field_name;
        my $field_value = get_value_from_xpath( xpath => $xml_field_name, rec => $xml_rec);

        $field_value = $self->map_xml_field_value ( 
                                                    field_value => $field_value,
                                                    field_name => $field_name,
                                                    field_def => $field_def,
                                                    direction => 'from_xml',
                                                    dbh => $self->{dbh},
                                                  );

        $rec->{$field_name} = $field_value;
    }

    return $rec;
}
#----------------------------------------------------------------------

sub get_value_from_xpath {
    my %params = @_;
    my $xpath = $params{xpath};
    my $rec = $params{rec};

    my @path_entries = split(/\//, $xpath);
    my $curr_value = $rec;
    for my $path (@path_entries) {
        if (ref($curr_value) ne 'HASH') {
            return;
        }
        $curr_value = $curr_value->{$path};
    }

    return $curr_value;
}
#----------------------------------------------------------------------

# Class method for exporting a list of this entity as XML
sub export_xml {
    my($class, %params) = @_;

    my $element_type_name_plural = $class->element_type_name_plural();
    my $bigip_version = F5::BigipVersionUtils::bigip_version();

    my $full_xml = <<EndXML;
<?xml version="1.0" encoding="utf-8"?>
<$element_type_name_plural export_version="$bigip_version">
EndXML

    my $iterator = $class->retrieve_all(%params);
    while(my $obj = $iterator->get_next) {
        $full_xml .= $obj->to_xml->sprint;
    }

    $full_xml .= <<EndXML;
</$element_type_name_plural>
EndXML

    return $full_xml;
}

#----------------------------------------------------------------------

# Class method for importing a list of this entity as XML
sub import_xml {
    my($class, %params) = @_;

    my @entity_rvs;
    my $element_type_name_plural = $class->element_type_name_plural();
    my $element_type_name = $class->element_type_name;
    my $entity_root_tag_found = 0;
    my $twig = F5::XML::Twig->new(twig_handlers => {
                                                $element_type_name_plural => sub {
                                                    $entity_root_tag_found = 1;
                                                },
                                                $element_type_name => sub {
                                                    my($twig, $elm) = @_;

                                                    if ( ! $entity_root_tag_found ) {
                                                        my $root_name = $twig->root->name;
                                                        if ($element_type_name_plural eq $root_name) {
                                                            $entity_root_tag_found = 1;
                                                        }
                                                        else {
                                                            $twig->finish_now;
                                                            return;
                                                        }
                                                    }

                                                    my $rv = $class->import_xml_entity(%params,
                                                                                       twig => $twig,
                                                                                       element => $elm);

                                                    push(@entity_rvs, $rv);
                                                }
                                               }
                             );

    if ($params{xml_filename}) {
        $twig->parsefile($params{xml_filename});
    }
    elsif ($params{xml_content}) {
        $twig->parse($params{xml_content});
    }

    if ( ! $entity_root_tag_found ) {
        my $message = "XML file for $class entity should have $element_type_name_plural root tag";
        die { error_message   => $message,
              display_message => $message,
              error_level     => 'info',
              error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
        };
    }

    return { entity_rvs => \@entity_rvs };
}

#----------------------------------------------------------------------

sub import_xml_entity {
    my($class, %params) = @_;
    my $elm = $params{element};

    my $import_result;
    my $new_obj = $class->new(%params,
                              xml => $elm);

    # Remove id from XML import
    my $old_id = delete $new_obj->{$class->id_field};

    my $existing_obj = $class->new(%params,
                                   get_criteria => $new_obj);

    my $import_obj;
    if ($existing_obj) {
        if ($params{store_values} && %{$params{store_values}}) { 
            for my $store_value ( keys %{$params{store_values}} ) {
                $new_obj->{$store_value} = $params{store_values}->{$store_value};
            }
        }
        my $rv = $existing_obj->store_update(store_obj => $new_obj);
        $rv = $rv->{db_change_made};
        if ($rv) {
            $import_result = $IMPORT_RESULT_UPDATE;
        }
        else {
            $import_result = $IMPORT_RESULT_UNCHANGED;
        }
        $import_obj = $existing_obj;
    }
    else {
        $new_obj->store_add;
        $import_result = $IMPORT_RESULT_ADDED;
        $import_obj = $new_obj;
    }

    return { entity => $import_obj,
             import_result => $import_result,
             old_id => $old_id,
           };
}

#----------------------------------------------------------------------

my %negative_comparison_equiv = (
                                 F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NE() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_EQ(),
                                 F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_LIKE() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_LIKE(),
                                 F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_REGEXP() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_REGEXP(),
                                 F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN() => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                );

sub is_filter_match {
    my($class, %params) = @_;
    my $search_filter = $params{search_filter};
    my $field_comparison = $search_filter->{field_comparison};
    if ($params{match_as_positive_equivalent} &&
        $negative_comparison_equiv{$field_comparison}) {

        $field_comparison = $negative_comparison_equiv{$field_comparison};
    }

    return $filter_func{$field_comparison}->(%params);
}

#----------------------------------------------------------------------

sub is_positive_match {
    my($class, %params) = @_;
    my $search_filter = $params{search_filter};

    return (! exists $negative_comparison_equiv{$search_filter->{field_comparison}});
}

#----------------------------------------------------------------------

sub is_from_learning {
    my ($self, %params) = @_;
    my $call_flags = $params{call_flags};
    my $session = $params{session};

    return $call_flags->{suggestion_rest_uuid} || ($session->{user_name} && $session->{user_name} eq $POLICY_BUILDER_USER_NAME);
}

#----------------------------------------------------------------------

sub parent_diff_struct {
    my ($self, %params) = @_;
    my $entity_config = $self->entity_config;
    return $entity_config->{parent_diff_struct}; 
}

#----------------------------------------------------------------------

sub get_entity_field_by_db_field {
    my ($self, %params) = @_;

    my $db_field_name = $params{db_field_name};
    return undef if ! $db_field_name;

    my $fields = $self->all_fields;
    for my $field (keys %$fields) {
        my $field_def = $fields->{$field};        
        if ($field_def->{db_field_name} eq $db_field_name) {
            return $field;
        }
    }

    return undef;
}

#----------------------------------------------------------------------
#
sub get_diff_parent_entity {
    my ($self, %params) = @_;
    my $diff_parent_entity = $self;
    my $policy = $params{policy} || $self->{policy};
    my $dbh = $params{dbh} || $self->{dbh};
    if (my $parent_diff_struct = ($params{parent_diff_struct} || $self->parent_diff_struct(%params))) {
        my $get_criteria = $parent_diff_struct->{parent_field} ? $self->{$parent_diff_struct->{parent_field}}
                                                                      : {};
        my $diff_parent_entity_type = $parent_diff_struct->{entity_type};
        my $subfield = $parent_diff_struct->{subfield};
        my $subfield_config = $diff_parent_entity_type->diff_structs()->{$subfield};
        if ($subfield_config->{where_values}) {
            for my $where_field (keys %{$subfield_config->{where_values}}) {
                $get_criteria->{$where_field} = $subfield_config->{where_values}->{$where_field};
            }
        }

        $diff_parent_entity = $diff_parent_entity_type->new(dbh => $dbh, policy => $policy, get_criteria => $get_criteria) ||
                              $diff_parent_entity_type->new(dbh => $dbh, policy => $policy, object => $get_criteria); #Fall back to stub in case parent struct doesn't actually exist yet.

        if ($subfield_config->{nested_struct}) {
            $diff_parent_entity->{$subfield} = [ $self ];
        }
        else {
            $diff_parent_entity->{$subfield} = $self;
        }

        if ($params{remove_non_unique_record_parts}) {
            my @field_exclusions = ($subfield, 'dbh', 'policy');
            $diff_parent_entity->remove_non_unique_record_parts(dbh => $dbh,
                                                                field_exclusions => \@field_exclusions);
        }
        
    }
    return $diff_parent_entity;
}

#----------------------------------------------------------------------

sub remove_invalid_fields {
    my ($self, %params) = @_;
    my $object = $params{object} || $self;
    my $all_fields = {%{$self->fields}, %{$self->diff_structs}};
    my $entity_config = $self->entity_config();
    my @fields = grep { !($entity_config->{fields}->{$_}->{is_dependent})} (keys %$all_fields);
    my $valid_fields = {};
    while (my $field = shift @fields) {
        $valid_fields->{$field} = 1;
        my $value = $object->{$field};
        if (defined($value) && $entity_config->{dependent_fields}->{$field} &&
              ref ($entity_config->{dependent_fields}->{$field}->{$value})) {
            push (@fields, @{$entity_config->{dependent_fields}->{$field}->{$value}});
        }
    }

    for my $field (keys %$all_fields) {
        if (!$valid_fields->{$field}) {
            delete $object->{$field};
        }
    }
    return $object;
}

#----------------------------------------------------------------------

sub clean_from_parent {
    my ($class, %params) = @_;
    my $policy = $params{policy};
    my $dbh    = $params{dbh};
    my @search_filters;
    my $entity_config = $class->entity_config();
    return if ($entity_config->{rest_paths} &&
               $entity_config->{rest_paths}->[0]->{rest_path_properties}->{singleton});

    if ($entity_config->{no_inheritance_conditions}) {
        push (@search_filters, map { $_ } @{$entity_config->{no_inheritance_conditions}});
    }
    my $call_flags = { force_delete => 1,
                       search_filters => \@search_filters,
                       suppress_events => 1 };
    $class->delete_many(dbh => $dbh,
                        policy => $policy,
                        call_flags => $call_flags,);
                        
}

#----------------------------------------------------------------------

sub clone_to_explicit {
    my ($self, %params) = @_;
    my $class = ref $self;
    my $explicit_object = $params{object};
    my $call_flags = $params{call_flags};

    my $explicit_entity = $class->new(policy => $self->{policy}, dbh => $self->{dbh}, object => $self);
    $explicit_entity->populate_diff_structs(%params);
    $explicit_entity->remove_invalid_fields();  #Remove fields that were not valid while it was a wildcard but remained.
    $explicit_entity->from_object(object => $explicit_object, call_flags => $call_flags);
    if (!defined($explicit_object->{is_wildcard})) {
        $explicit_entity->{is_wildcard} = 0;
    }
    if (!defined($explicit_object->{has_tightening})) {
        $explicit_entity->{has_tightening} = 0;
    }
    delete $explicit_entity->{in_staging};
    delete $explicit_entity->{flg_in_classification};
    if (defined($explicit_entity->{description})) {
        $explicit_entity->{description} .= " (Cloned from " . $self->get_full_display_name(%params) . ")";
    }
    $explicit_entity->remove_invalid_fields(); #Remove fields that are now not valid that it is an explicit.
    $explicit_entity->purify_store_object(%params, object => $explicit_entity, skip_purify_diffs => 1);
    return $explicit_entity;
}

#----------------------------------------------------------------------
sub contains_control_chars {
    my ($self, %params) = @_;
    my $policy = $params{policy};
    my $value = $params{value};

    if ($policy && $policy->{encoding} && $policy->{encoding} ne 'utf-8') {
        eval {
            Encode::from_to($value, $policy->{encoding}, 'utf-8');
        };
        if ($@){
            die "$@";
        }
    }

    if($value  =~ /[\cA-\cZ]/){
        return 1;
    }

    return 0;
}

#----------------------------------------------------------------------
sub catchup_search_filters {
  return [];
}

#----------------------------------------------------------------------

sub is_unlearnable_http_check {
    my ($self, %params) = @_;
    my $code;
    if ($params{object}) {
        $code = $params{object}->{http_check_code};
    }
    else {
        $code = $self->get_new_value(%params, field_name => 'http_check_code');
    }
    return ($code && $UNLEARNABLE_HTTP_CHECK_CODES{$code}) ? $UNLEARNABLE_HTTP_CHECK_CODES{$code}
                                     : 0;
}

#----------------------------------------------------------------------

sub get_comparison_mapping {
    my(%params) = @_;
    my $dbh = $params{dbh};
    if ($dbh) {
        my $db_type = F5::DbUtils::get_db_type($dbh);
        if ($COMPARISON_MAPPING_BY_DB{$db_type}) {
            return $COMPARISON_MAPPING_BY_DB{$db_type};
        }
    }
    return \%COMPARISON_MAPPING;
}

sub export_declarative {
    my ($class, %params) = @_;
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $policy = $params{policy};
    my $search_filters = $params{search_filters} || [];
    my $minimal_export = $params{minimal_export};

    my @entities;
    if ($minimal_export) {
        @entities = @{$class->extract_diff_entities(%params, is_in_declarative_export => 1)};
    }
    else {
        push (@$search_filters, @{$class->declarative_search_filters(%params)});
        @entities = $class->retrieve_all(%params,
                                         search_filters => $search_filters);
    }

    my @warnings;
    my $i = 0;
    while ($i <= $#entities) {
        my $entity = $entities[$i];
        if ($entity->{'$action'} && $entity->{'$action'} eq 'warn') {
            push @warnings, $entity->{'$warning'};
            splice (@entities, $i, 1);
        }
        else {
            $entities[$i] = $entity->to_declarative(%params,
                                                    is_in_declarative_export => 1);
            $i++;
        }
    }

    return { items => \@entities,
             warnings => \@warnings };
}

#----------------------------------------------------------------------

sub to_declarative {
    my ($self, %params) = @_;

    my $delete = defined $self->{'$action'} && $self->{'$action'} eq 'delete';
    my $entity = $self->to_rest(skip_populate_diff_structs => $params{minimal_export},
                                is_in_declarative_export => $params{is_in_declarative_export});
    F5::YAML::Schema::Utils::item_fixup(item => $entity,
                                        entity_type => $self->entity_config->{entity_type},
                                        direction => 'to_declarative',
                                       );

    if ($delete) {
        $entity->{'$action'} = 'delete';
    }
    return $entity;
}

#----------------------------------------------------------------------

sub extract_diff_entities {
    my ($class, %params) = @_;
    my $dbh = $params{dbh};
    my $policy = $params{policy};
    my $diff_items = $params{diff_items};

    my @items;
    for my $diff_item (@$diff_items) {
        my $diff_type = $diff_item->{diff_type};
        my $item;
        my $get_criteria;

        if ($diff_type == F5::ASMConfig::DiffType::DIFF_TYPE_ONLY_SECOND()) {
            $item = $diff_item->{second_element};
            $item = $class->new(dbh => $dbh, object => $item, policy => $policy);
            my $is_deletable_rv = $item->is_deletable(%params);
            if ($is_deletable_rv->{error_message}) {
                $item->{'$action'} = 'warn';
                $item->{'$warning'} = $is_deletable_rv->{error_message};
            }
            else {
                $item = { map {$_ => $item->{$_} } @{$class->declarative_unique_record_parts}};
                $item = $class->new(dbh => $dbh, object => $item);
                $item->{'$action'} = 'delete';
            }
        }
        else {
            if ($class->entity_config->{rest_paths}->[0]->{rest_path_properties}->{singleton}) {
                $get_criteria = { policy_id => $policy->{policy_id} };
            }
            else {
                $get_criteria = { $class->unique_id_field => $diff_item->{first_element_id} };
            }
        }

        if (!$item) {
            $item = $class->new(dbh => $dbh,
                                policy => $policy,
                                get_criteria => $get_criteria,
                                populate_diff_structs => 1,);
        }

        if ($diff_type == F5::ASMConfig::DiffType::DIFF_TYPE_CONFLICT()) {
            my $object;
            my %details_fields;
            for my $detail (@{$diff_item->{diff_item_details}}) {
               my ($field, $value) = split(/\./, $detail->{field_name});
               next if defined $details_fields{$field};
               $details_fields{$field} = 1;
            }

            my @fields = keys %details_fields;
            push (@fields, @{$class->declarative_unique_record_parts});
            my $dependent_fields = $class->entity_config->{dependent_fields};
            my $reversed_dependent_fields = $class->reverse_dependent_fields();

            # fields that must be defined in the object in order for the diff fields to be properly displayed
            # we should not populate the dependencies of these fields
            my %required_fields;

            while (my $field = shift @fields) {
                next if defined $object->{$field};
                $object->{$field} = $item->{$field};

                next if !defined $dependent_fields->{$field} && !defined $reversed_dependent_fields->{$field};
                if ( (defined $details_fields{$field} || defined $required_fields{$field}) && defined $reversed_dependent_fields->{$field}) {
                    push (@fields, @{$reversed_dependent_fields->{$field}});
                    map { $required_fields{$_} = 1 } @{$reversed_dependent_fields->{$field}};
                }
                if (!defined $required_fields{$field} && ref ($dependent_fields->{$field}->{$item->{$field}})) {
                    push (@fields, @{$dependent_fields->{$field}->{$item->{$field}}});
                }
            }
            $item = $class->new(dbh => $dbh,
                                policy => $policy,
                                object => $object,);
        }
        push (@items, $item);
    }
    return \@items;
}

#----------------------------------------------------------------------

sub declarative_search_filters {
    my ($class, %params) = @_;
    my $search_filters = $params{search_filters} || [];

    push (@$search_filters, @{$class->rest_search_filters()});
    if (my $fields = $class->entity_config->{declarative_excluded_values}) {
        for my $field (keys %$fields) {
            my $values = $class->entity_config->{declarative_excluded_values}->{$field};
            push (@$search_filters,
                  F5::ASMConfig::SearchFilter->new(field_name => $field,
                                                   field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_NOT_IN(),
                                                   field_value_list => $values) );
        }
    }

    return $search_filters;
}

#----------------------------------------------------------------------

sub declarative_fields {
    my ($class, %params) = @_;
    my $rest_field_map = $class->rest_field_map;
    my $declarative_fields = {};
    for my $rest_field (keys %{$rest_field_map}) {
        my $field_def = $class->fields->{$rest_field};

        if (!$field_def->{declarative_required_field}) {
            if (!$class->policy_id_field
                  && !($field_def->{unique_record_part} || $field_def->{declarative_unique_record_part})) {
                next;
            }
            elsif ($field_def->{full_db_field_name}) {
                next;
            }
            elsif (!F5::YAML::Schema::Utils::is_supported_field(field => $rest_field,
                                                                field_def => $field_def)) {
                next;
            }
        }
        if (!defined $field_def->{rest_field_name}) {
            $field_def->{rest_field_name} = $rest_field_map->{$rest_field};
        }
        $declarative_fields->{$rest_field} = $field_def;

    }
    return $class->install_accessor('declarative_fields', $declarative_fields);
}

#----------------------------------------------------------------------

sub map_declarative_field_value {
    my ($self, %params) = @_;
    my $dbh = $params{dbh} || $self->{dbh};
    my $field_name = $params{field_name};
    my $field_value = $params{field_value} || $self->{$field_name};
    my $field_def = $self->declarative_fields->{$field_name};

    if (!$field_def || !defined $field_value) {
        return;
    }

    return $self->map_rest_field_value(field_name => $field_name,
                                       field_value => $field_value,
                                       field_def => $field_def,
                                       direction => 'to_rest',
                                       dbh => $dbh,
                                      );
}

#----------------------------------------------------------------------

sub declarative_unique_record_parts {
    my($class) = @_;
    my $fields = $class->declarative_fields;
    my @declarative_unique_record_parts = grep { $fields->{$_}->{declarative_unique_record_part}
                                                 || $fields->{$_}->{unique_record_part}
                                                 || $fields->{$_}->{pb_unique_record_part} } keys %$fields;
    return $class->install_accessor('declarative_unique_record_parts', \@declarative_unique_record_parts);
}

#----------------------------------------------------------------------

sub before_declarative_import {
    return;
}

#----------------------------------------------------------------------

sub validate_unique_host {
    my ($self, %params) = @_;
    my $dbh = $params{dbh} || $self->{dbh};
    my $table_name = $params{table_name};
    my $policy_id = $params{policy}->{policy_id};

    if ($table_name eq '' || !$policy_id) {
        my $error_message = "Non existent table or invalid id";
        die { error_message =>  $error_message,
            display_message =>  $error_message,
            error_level     => 'info',
            error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
        };
    }

    my $new_host = $params{field_value};
    my $new_host_prefix_length = 0;

    if ($new_host =~ /^\*\.(.*)/g) {
        $new_host = $1;
    }
    elsif ($new_host =~ /(.*)\/(\d+)/g) {
        $new_host = $1;
        $new_host_prefix_length = $2;
    }

    if (!$new_host_prefix_length) {
        $new_host_prefix_length = (defined inet_pton(AF_INET, $new_host)) ? 32 : 128;
    }

    my $current_deny_list = $dbh->prepare(<<EndQuery);
SELECT *
 FROM $table_name
 WHERE policy_id  = ?
EndQuery
    $current_deny_list->execute($policy_id);

    while (my $current_host_row = $current_deny_list->fetchrow_hashref) {
        my $current_prefix_length = 0;
        my $current_host = $current_host_row->{host};

        if ($current_host =~ /(.*)\/(\d+)/g) {
            $current_host = $1;
            $current_prefix_length = $2;
        }

        if (defined inet_pton(AF_INET, $current_host)) {
            $current_prefix_length = $current_prefix_length ? $current_prefix_length : 32;
            if ($new_host_prefix_length > $current_prefix_length) {
                my $netmask = F5::IpUtils::cidr_to_netmask_ipv4($current_prefix_length);
                if (F5::IpAddr::ip_addr_maskcmp($new_host, $current_host, $netmask)) {
                    return 0;
                }
            }
        }
        elsif (defined inet_pton(AF_INET6, $current_host)) {
            $current_prefix_length = $current_prefix_length ? $current_prefix_length : 128;
            if ($new_host_prefix_length > $current_prefix_length) {
                my $netmask = F5::IpUtils::cidr_to_netmask_ipv6($current_prefix_length);
                if (F5::IpAddr::ip_addr_maskcmp($new_host, $current_host, $netmask)) {
                    return 0;
                }
            }
        }
        elsif ($current_host =~ /^\*\.(.*)/g) {
            $current_host = $1;
            if ($new_host =~ /.*\.$current_host$/gi) {
                return 0;
            }
        }
    }

    return 1;
}

#----------------------------------------------------------------------

sub declarative_required_fields {
    my($class) = @_;
    my $fields = $class->fields;

    my @required_fields = grep { $fields->{$_}->{declarative_required_field} } keys %$fields;
    if (!@required_fields) {
        my @unique_record_parts = grep { !$fields->{$_}->{exclude_from_rest} } @{$class->unique_record_parts};
        if (my ($name_field) = grep { $fields->{$_}->{is_name_field} } @unique_record_parts) {
            @required_fields = ( $name_field );
        }
        else {
            @required_fields = @unique_record_parts;
        }
    }

    my @required_fields_rv;
    for my $field (@required_fields) {
        my $rest_field_map = $class->rest_field_map();
        if ($fields->{$field}->{declarative_flat_reference}) {
            $rest_field_map = $fields->{$field}->{entity_type}->rest_field_map();
            ($field) = @{$fields->{$field}->{entity_type}->unique_record_parts};
        }
        push (@required_fields_rv, $rest_field_map->{$field});
    }

    return $class->install_accessor('declarative_required_fields', \@required_fields_rv);
}

#----------------------------------------------------------------------

sub is_valid_host {
    my ($self, %params) = @_;

    my $store_obj = $params{store_obj};
    my $entity_class = $params{entity_class};
    my $field_def = $params{field_def};
    my $policy_id = $field_def->{policy}->{policy_id};
    my $allow_wildcard = $field_def->{allow_wildcard} || 0;

    my $input_host = $params{field_value};
    my $host = $input_host;
    my $prefix_length = 0;
    my $found_prefix = 0;
    my $is_wildcard_flg = 0;
    my $error_message = "";
    my $host_action = $store_obj->{action_type};

    my $validate_ret = $self->general_validate(%params);
    if (!$validate_ret->{completed_successfully}) {
        die { error_message =>  $validate_ret->{error_message},
            display_message =>  $validate_ret->{error_message},
            error_level     => 'info',
            error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
        };
    }

    if ($allow_wildcard && $input_host =~ /^\*\.(.*)/g) {
        $host = $1;
        $is_wildcard_flg = 1;
    }
    elsif ($input_host =~ /(.*)\/(\d+)/g) {
        $host = $1;
        $prefix_length = $2;
        $found_prefix = 1;
    }

    if (!$is_wildcard_flg && $prefix_length < 32 && defined inet_pton(AF_INET, $host)) {
        if ($host_action && $host_action == F5::ASMConfig::SsrfActionType::SSRF_ACTION_RESOLVE()) {
            my $action_display_value = $self->get_field_display_string('field_name' => 'action_type');
            $error_message = "'$action_display_value' is not applicable for '$input_host'";
            die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
            };
        }

        my $netmask = F5::IpUtils::cidr_to_netmask_ipv4($prefix_length);

        if (!F5::IpUtils::is_netmask_valid_for_address(ip_address => $host, mask => $netmask)) {
            $error_message =  "Invalid IP/Mask combination";
        }
    }
    elsif (!$is_wildcard_flg && $prefix_length < 128 && defined inet_pton(AF_INET6, $host)) {
        if ($host_action && $host_action == F5::ASMConfig::SsrfActionType::SSRF_ACTION_RESOLVE()) {
            my $action_display_value = $self->get_field_display_string('field_name' => 'action_type');
            $error_message = "'$action_display_value' is not applicable for '$input_host'";
            die { error_message => $error_message,
                display_message => $error_message,
                error_level     => 'info',
                error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
            };
        }

        my $netmask = F5::IpUtils::cidr_to_netmask_ipv6($prefix_length);

        if (!F5::IpUtils::is_netmask_valid_for_address(ip_address => $host, mask => $netmask)) {
            $error_message =  "Invalid IP/Mask combination";
        }
    }
    elsif ((!$found_prefix && $self->is_valid_host_name(field_value => $host)->{completed_successfully}) || $host eq "*") {
    }
    else {
        $error_message = "Invalid host name $input_host";
    }

    if ($error_message ne '') {
        die { error_message => $error_message,
            display_message => $error_message,
            error_level     => 'info',
            error_code      => F5::ASMConfig::ErrorCode::KNOWN_VALIDATION_ERRORS(),
        };
    }

    return { completed_successfully => 1 };
}

#----------------------------------------------------------------------

1;
