package F5::PrepareConf::Policy;

use strict;
use warnings;
use base 'F5::PrepareConf::Base';

use Math::BigInt;
use Cwd;
use File::Basename;

use Text::Iconv;
use List::Util qw(max);
use DBI;
use Digest::MD5;
use Socket qw(:DEFAULT inet_pton inet_ntop);
use F5::IpAddr;
use File::Copy;
use Memoize;
use File::Slurp;
use File::Path qw(make_path);
use JSON;

# tsdev includes

use F5::Deploy;
use F5::Cfg;
use F5::Text::Glob;
use F5::PrepareConf::Constants qw(ERRMSG_DBH_LOST RET_OK RET_NOT_OK FALSE TRUE :KEY_UI_TYPE);
use F5::PrepareConf::Checksums;
use F5::PrepareConf::Base;
use F5::ASMConfig::Cache;
use F5::ASMConfig::Handler;
use F5::ASMConfig::Entity::URL;
use F5::ASMConfig::Entity::SessionAwarenessDataPoint;
use F5::ASMConfig::Entity::Suggestion;
use F5::ASMConfig::Entity::SsrfHost;
use F5::Violation;
use F5::Violation::Utils;

use F5::BatchInsert;
use F5::LazyInsert;

use F5::DbUtils;
use F5::IpUtils;
use F5::BigipUtils;
use F5::ObjUtils qw(dclone);
use F5::Utils::File;
use F5::CRC;

use F5::NegativeSignatures;
use F5::NegativeSignatures::Collection;
use F5::ThreatCampaigns;
use F5::BotSignatures;
use F5::License;
use F5::ASMConfig::Cache;
use F5::GraphqlProfile;

# input bitmap flags
use constant  INPUT_HAS_NO_USER_DATA        =>  1;
use constant  INPUT_HAS_USER_DATA           =>  2;
use constant  INPUT_CHECK_USER_DATA         =>  4;

# dynamic params
use constant  DYNAMIC_CONTENT_VALUE         =>  2;
use constant  FLG_SINGLE_MULTIPLE           =>  10;
use constant  FORM_DYNAMIC_PARAM            =>  1;
use constant  URL_DYNAMIC_PARAM             =>  2;
use constant  XML_DYNAMIC_PARAM             =>  4;
use constant  REGEXP_DYNAMIC_PARAM          =>  8;
use constant  DYNAMIC_PARAMETER_NAME        => 16;
use constant  NAME_REGEXP_DYNAMIC_PARAM     => 64;
use constant  LINK_DYNAMIC_PARAM            => 128;
use constant  FORM_NAME_DYNAMIC_PARAM       => 256;

# reg exp types
use constant REG_EXP_TYPE_URI               => 1;
use constant REG_EXP_TYPE_RESPONSE          => 2;
use constant REG_EXP_TYPE_HEADER_VALUE      => 3;
use constant REG_EXP_TYPE_KEY_VALUE_PAIRS   => 4;
use constant COMPUTE_ADDITIONAL_DO_REPLACE  => 1;
use constant NEGSIG_BLOCKING_MASK_REJECT    => 1;
use constant NEGSIG_BLOCKING_MASK_ALARM     => 2;
use constant NEGSIG_BLOCKING_MASK_LEARN     => 4;

use constant HTTP_RESPONSE_SPLITTING_ATTACK_TYPE_ID => 32;
use constant RESPONSE_SPLITTING_BITMAP_VALUE => 4;

use constant CORS_ALLOWED_HEADER_TYPE => 0;
use constant CORS_EXPOSED_HEADER_TYPE => 1;

use constant REDIRECT_ACTION_NONE    => 0;
use constant REDIRECT_ACTION_ENFORCE => 1;
use constant REDIRECT_ACTION_TIGHTEN => 2;

use constant SIG_TYPE_ATTACK_SIGNATURE  => 1;
use constant SIG_TYPE_THREAT_CAMPAIGN   => 2;
use constant SIG_TYPE_BOT_SIGNATURE     => 3;
use constant SIG_TYPE_BROWSER_SIGNATURE => 4;


my $ANY_URL_METHOD_NAME = $F5::ASMConfig::Entity::URL::ANY_URL_METHOD_NAME;

my $ACCOUNT_DATA_DIR                       = $F5::PrepareConf::Base::ACCOUNT_DATA_DIR;
my $MULTIPLE_DECODING_NORMALIZATION_CODE = 16;
my $MAX_HEADERS_INDEX = 131072;
my $MAX_PARAMS_INDEX = 524288;
my $MAX_COOKIES_INDEX = 4194304;
my $QUERY_AS_VALUE_BIT = 22;

my %response_types = (
                   'default'  => F5::ASMConfig::ResponseType::RESPONSE_TYPE_DEFAULT(),
                   'persistent flow' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_PERSISTENT_FLOW(),
                   'XML' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_XML(),
                   'Ajax' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_AJAX(),
                   'Ajax_login' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_AJAX_LOGIN(),
                   'Hijack' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_HIJACK(),
                   'Mobile' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_MOBILE(),
                   'Captcha' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_CAPTCHA(),
                   'Captcha_fail' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_CAPTCHA_FAIL(),
                   'Failed_login_honeypot' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_FAILED_LOGIN_HONEYPOT(),
                   'Failed_login_honeypot_ajax' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_FAILED_LOGIN_HONEYPOT_AJAX(),
                   'Leaked_credentials' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_LEAKED_CREDENTIALS(),
                   'Leaked_credentials_ajax' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_LEAKED_CREDENTIALS_AJAX(),
                   'graphql' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_GRAPHQL(),
                   'grpc' => F5::ASMConfig::ResponseType::RESPONSE_TYPE_GRPC(),
               );

my %ajax_action_types = (
                   redirect  => F5::ASMConfig::AjaxActionType::AJAX_ACTION_TYPE_REDIRECT(),
                   alert_popup => F5::ASMConfig::AjaxActionType::AJAX_ACTION_TYPE_ALERT_POPUP(),
                   custom  => F5::ASMConfig::AjaxActionType::AJAX_ACTION_TYPE_CUSTOM(),
               );

my %enforcement_type_id = (
                             'http'     => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_HTTP(),
                             'xml'      => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_XML(),
                             'json'     => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_JSON(),
                             'apply_value_signatures'    => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_APPLY_VALUE_SIGNATURES(),
                             'disallow' =>  F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_DISALLOW(),
                             'dont_check' => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_DONT_CHECK(),
                             'gwt'     => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GWT(),
                             'do_nothing' => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_DO_NOTHING(),
                             'graphql' => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRAPHQL(),
                             'grpc' => F5::ASMConfig::EnforcementType::ENFORCEMENT_TYPE_GRPC(),
              );

my %content_profile_types = (
                              F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML() => 'xml',
                              F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON() => 'json',
                              F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GWT() => 'gwt',
                              F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_PLAIN_TEXT() => 'plain_text',
                              F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRAPHQL() => 'graphql',
                              F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRPC() => 'grpc',
                            );

my %captcha_token_to_replacement = (
                                     #TODO: change --> refresh
                                     change => '<a href="#" id="bottle"><img src="data:image/png;base64,PLACE_HOLDER_FOR_CAPTCHA_CHANGE" alt="bottle" class="thumbnails"/></a>',
                                     solution => '<input type="text" id="ans" name="answer" value=""/>',
                                     submit => '<button id="jar" type="button">submit</button>',
                                     image => '<img src="data:image/png;base64,PLACE_HOLDER_FOR_CAPTCHA_IMAGE" alt="Red dot"/>',
                                     audio => '<audio id="captcha_audio" src="data:audio/mpeg;base64,PLACE_HOLDER_FOR_CAPTCHA_AUDIO" alt="Red dot">Audio is not supported in your browser.</audio><img onclick="document.getElementById(\'captcha_audio\').play()" src="data:image/png;base64,PLACE_HOLDER_FOR_CAPTCHA_AUDIO_ICON" style="cursor: pointer;" />',
                                     support_id => 'PLACE_HOLDER_FOR_CAPTCHA_SUPPORT_ID',
                                    );

my $CAPTCHA_REFRESH_BASE64_ENCODED_IMAGE    = F5::Deploy::map_path('/var/ts/share/alternate_responses/captcha_refresh_base64_encoded_image');
my $CAPTCHA_AUDIO_ICON_BASE64_ENCODED_IMAGE = F5::Deploy::map_path('/var/ts/share/alternate_responses/captcha_audio_icon_base64_encoded_image');

#----------------------------------------------------------------------

my $policy_parts = [
                    {
                     policy_part => 'charsets',
                     prepare_function => 'preload_charset_templates',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_CHARSETS
                                                    ) ],
                               },
                     has_readonly_mode => 1,
                    },
                    {
                     policy_part => 'allowed_methods',
                     prepare_function => 'preload_allowed_methods',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_ALLOWED_METHODS
                                                       PLC.PL_OBJECT_METHOD_OVERRIDES
                                                    ) ],
                               },
                     has_readonly_mode => 1,
                    },
                    {
                     policy_part => 'account_attributes',
                     prepare_function => 'prepare_account_attributes',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_POLICIES
                                                       PLC.PL_POLICY_GENERAL
                                                       PLC.PL_POLICY_ANTIVIRUS
                                                       PLC.PL_POLICY_NEGSIG_SIGNATURES
                                                       PLC.PL_POLICY_TEMPLATES_SETTINGS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'blocking',
                     prepare_function => 'prepare_blocking_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_ANTIVIRUS
                                                       PLC.PL_BLOCKING
                                                       PLC.PL_SESSION_AWARENESS_VIOLATIONS
                                                       PLC.PL_BEHAVIORAL_ENFORCEMENT
                                                       PLC.PL_BEHAVIORAL_ENFORCEMENT_VIOLATIONS
                                                       PLC.PL_DECEPTION_SETTINGS
                                                    ) ],
                               },
                     depends => [ 'charsets','account_attributes' ],
                     has_readonly_mode => 1,
                    },
                    {
                     policy_part => 'xml_profiles',
                     prepare_function => 'prepare_xml_profiles',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_CONTENT_PROFILES
                                                       PLC.PL_OBJECT_CONTENT_PROFILES
                                                       PLC.PL_XML_PROFILE_VALIDATION_FILES
                                                       PLC.PL_PARAM_CONTENT_PROFILES
                                                       PLC.PL_XML_PROFILE_SOAP_METHODS
                                                       PLC.PL_XML_PROFILE_DEFENSE_ATTRIBUTES
                                                       PLC.PL_XML_SOAP_SECURITY
                                                       PLC.PL_XML_ROLE_ACTOR
                                                       PLC.PL_XML_SOAP_SPECIFIC_ELEMENTS
                                                       PLC.PL_XML_SOAP_PREFIX_NAMESPACE_MAPPING
                                                    )
                                                 ],
                               },
                     depends => [ 'objects', 'parameters' ],
                     has_readonly_mode => 1,
                    },
                    {
			policy_part => 'json_profiles',
                        prepare_function => 'prepare_json_profiles',
                        tables => {
				  policy_tables => [ qw(
                                                         PLC.PL_CONTENT_PROFILES
                                                         PLC.PL_OBJECT_CONTENT_PROFILES
                                                         PLC.PL_JSON_PROFILE_VALIDATION_FILES
                                                         PLC.PL_PARAM_CONTENT_PROFILES
                                                     )
                                                   ],
                                  },
                        depends => [ 'objects', 'parameters' ],
                        has_readonly_mode => 1,
                    },
                    {
                        policy_part => 'graphql_profiles',
                        prepare_function => 'prepare_graphql_profiles',
                        tables => {
                                  policy_tables => [ qw(
                                                         PLC.PL_CONTENT_PROFILES
                                                         PLC.PL_OBJECT_CONTENT_PROFILES
                                                         PLC.PL_GRAPHQL_PROFILE_DEFENSE_ATTRIBUTES
                                                         PLC.PL_CONTENT_PROFILE_VALUE_METACHARS
                                                         PLC.PL_CONTENT_SENSITIVE_PARAMS
                                                         PLC.PL_GRAPHQL_PROFILE_IDL_FILES
                                                         PLC.PL_CONTENT_PROFILE_DISALLOWED_RESPONSE_PATTERNS
                                                      )
                                                   ],
                                  },
                        depends => [ 'objects' ],
                        has_readonly_mode => 1,
                    },
                    {
                        policy_part => 'grpc_profiles',
                        prepare_function => 'prepare_grpc_profiles',
                        tables => {
                                  policy_tables => [ qw(
                                                         PLC.PL_CONTENT_PROFILES
                                                         PLC.PL_OBJECT_CONTENT_PROFILES
                                                         PLC.PL_GRPC_PROFILE_IDL_FILES
                                                         PLC.PL_PARAM_CONTENT_PROFILES
                                                     )
                                                   ],
                                  },
                        depends => [ 'objects', 'parameters' ],
                        has_readonly_mode => 1,
                    },
                    {
                        policy_part => 'gwt_profiles',
                        prepare_function => 'prepare_gwt_profiles',
                        tables => {
                                  policy_tables => [ qw(
                                                         PLC.PL_CONTENT_PROFILES
                                                         PLC.PL_OBJECT_CONTENT_PROFILES
                                                         PLC.PL_PARAM_CONTENT_PROFILES
                                                     )
                                                   ],
                                  },
                        depends => [ 'objects', 'parameters' ],
                        has_readonly_mode => 1,
                    },
                    {
                        policy_part => 'plain_text_profiles',
                        prepare_function => 'prepare_plain_text_profiles',
                        tables => {
                                  policy_tables => [ qw(
                                                         PLC.PL_CONTENT_PROFILES
                                                         PLC.PL_OBJECT_CONTENT_PROFILES
                                                         PLC.PL_PARAM_CONTENT_PROFILES
                                                     )
                                                   ],
                                  },
                        depends => [ 'objects', 'parameters' ],
                        has_readonly_mode => 1,
                    },

                   {
                     policy_part => 'soap_security',
                     prepare_function => 'prepare_wss_checks',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_XML_WS_SUBVIOLATIONS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'file_types',
                     prepare_function => 'prepare_file_type_attrs_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_POLICIES
                                                       PLC.PL_FILETYPES
                                                    )
                                                 ],
                                tables => [ qw (DCC.ACCOUNTS) ],
                               },
                    },
                    {
                     policy_part => 'disallowed_file_types',
                     prepare_function => 'prepare_disallowed_file_types',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_FILETYPES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'session_awareness_settings',
                     row_handler => 'session_awareness_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                         PLC.PL_SESSION_AWARENESS_SETTINGS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_SESSION_AWARENESS_SETTINGS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_SESSION_AWARENESS_SETTINGS
                                                       PLC.PL_SESSION_AWARENESS_VIOLATIONS 
                                                       PLC.PL_SESSION_AWARENESS_LOGIN_OBJECTS
                                                       PLC.PL_OBJECTS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'ip_address_allowlists',
                     prepare_function => 'prepare_ip_address_allowlist',
                     tables => {
                                 source_tables => [ qw(
                                                         PLC.PL_IP_WHITELIST
                                                         PLC.PL_SSRF_HOSTS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_IP_WHITELIST
                                                    ) ],

                                policy_tables => [ qw(
                                                       PLC.PL_IP_WHITELIST
                                                       PLC.PL_SSRF_HOSTS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'ip_reputation',
                     prepare_function => 'prepare_ip_reputation_categories',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_IP_REPUTATION_CATEGORIES
                                                    ) ],
                               },
                    },

                    {
                     policy_part => 'allowed_response_codes',
                     prepare_function => 'prepare_allowed_response_codes',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_ALLOWED_RESPONSE_CODES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'http_length',
                     row_handler => 'http_length_row_handler',
                     multiple_source_table_handling => 'join',
                     join_field_name => 'policy_id',
                     tables => {
                                source_tables => [ qw(
                                                       PLC.PL_POLICY_COOKIE_SETTINGS
                                                       PLC.PL_POLICY_HEADER_SETTINGS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_GLOBAL_LENGTHS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_COOKIE_SETTINGS
                                                       PLC.PL_POLICY_HEADER_SETTINGS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'enforcer_settings',
                     tables => {
                                source_tables => [ qw(
                                                       PLC.PL_POLICY_ENFORCER_SETTINGS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_ENFORCER_SETTINGS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_ENFORCER_SETTINGS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'threat_rating_settings',
                     required_features => [ qw( threat_rating )],
                     tables => {
                                source_tables => [ qw(
                                                       PLC.PL_THREAT_RATING_SETTINGS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_THREAT_RATING_SETTINGS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_THREAT_RATING_SETTINGS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'http_checks',
                     prepare_function => 'prepare_http_checks',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_HTTP_CHECKS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'objects',
                     prepare_function => 'prepare_objects_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_OBJECTS
                                                       PLC.PL_OBJECT_METACHARS
                                                       PLC.PL_LOGIN_OBJECTS
                                                       PLC.PL_REVOKE_OBJECTS
                                                       PLC.PL_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES
                                                       PLC.PL_CHARSETS
                                                       PLC.PL_OBJECT_CONTENT_PROFILES
                                                    ) ],
                                tables => [ qw (DCC.ACCOUNTS) ],
                               },
                     depends => [ 'charsets' ],
                     has_readonly_mode => 1,
                    },
                    {
                     policy_part => 'websocket_urls',
                     prepare_function => 'prepare_websocket_urls_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_WEBSOCKET_URLS
                                                       PLC.PL_WEBSOCKET_URL_METACHARS
                                                    ) ],
                                tables => [ qw (DCC.ACCOUNTS) ],
                               },
                     has_readonly_mode => 1,
                    },
                    {
                     policy_part => 'flows',
                     prepare_function => 'prepare_flow_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_FLOWS
                                                       PLC.PL_FLOW_PARAMS
                                                       PLC.PL_PARAM_ATTRIBUTES
                                                    ) ],
                               },
                     depends => [ 'objects' ],
                     has_readonly_mode => 1,
                    },
                    {
                     policy_part => 'parameters',
                     prepare_function => 'prepare_parameters',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_FLOW_PARAMS
                                                       PLC.PL_OBJECT_PARAMS
                                                       PLC.PL_GLOBAL_PARAMS
                                                       PLC.PL_PARAM_ATTRIBUTES
                                                       PLC.PL_PARAM_NAME_METACHARS
                                                       PLC.PL_PARAM_VALUE_METACHARS
                                                       PLC.PL_CHARSETS
                                                    ) ],
                                tables => [ qw ( DCC.ACCOUNTS ) ],
                               },
                     depends => [ 'objects', 'flows', 'charsets' ],
                     has_readonly_mode => 1,
                    },
                    {
                     policy_part => 'headers',
                     prepare_function => 'prepare_headers',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_HEADERS
                                                       PLC.PL_PARAM_ATTRIBUTES
                                                    ) ],
                                tables => [  qw(
                                                DCC.ACCOUNT_WILDCARD_HEADERS
                                                DCC.ACCOUNT_HEADERS
                                               ) ],
                               },
                     has_readonly_mode => 1,
                    },
                    {
                     depends => [ 'objects', 'parameters'],
                     policy_part => 'positional_parameters',
                     row_handler => 'positional_parameters_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                       PLC.PL_OBJECT_POSITIONAL_PARAMETERS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_OBJECT_POSITIONAL_PARAMETERS
                                                  ) ],
                               },
                    },
                    {
                     policy_part => 'extracts',
                     prepare_function => 'prepare_extracts',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_OBJECT_EXTRACTS
                                                       PLC.PL_GLOBAL_EXTRACTS
                                                       PLC.PL_GLOBAL_EXTRACT_OBJECT_TYPES
                                                       PLC.PL_GLOBAL_EXTRACT_OBJECTS
                                                       PLC.PL_FILETYPES
                                                       PLC.PL_EXTRACT_ATTRIBUTES
                                                       PLC.PL_DYNAMIC_PARAM_EXTRACTS
                                                       PLC.PL_PARAM_ATTRIBUTES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'static_params',
                     prepare_function => 'prepare_static_params_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_STATIC_PARAMS
                                                       PLC.PL_PARAM_ATTRIBUTES
                                                    ) ],
                               },
                     depends => [ 'parameters' ],
                    },
                    {
                     policy_part => 'alternate_response',
                     prepare_function => 'prepare_alternate_response_file_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_ALTERNATE_RESPONSES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'alternate_response_settings',
                     prepare_function => 'prepare_alternate_response_settings',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_ALTERNATE_RESPONSE_SETTINGS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'allowed_cookies',
                     prepare_function => 'prepare_allowed_cookies_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_ALLOWED_COOKIES
                                                    ) ],
                                tables => [  qw(
                                                 DCC.ACCOUNTS
                                                 DCC.ACCOUNT_ALLOWED_COOKIES
                                                 DCC.ACCOUNT_WILDCARD_ALLOWED_COOKIES
                                               ) ],
                               },
                    },
                    {
                     policy_part => 'valid_host_names',
                     prepare_function => 'prepare_valid_host_names_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_VALID_HOST_NAMES
                                                    ) ],
                                tables => [ qw (DCC.ACCOUNTS) ],
                               },
                    },
                    {
                     policy_part => 'navigation_params',
                     prepare_function => 'prepare_navigation_params_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_CRAWLER_NAVIGATION_PARAMS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'dynamic_flows',
                     row_handler => 'dynamic_flows_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                         PLC.PL_DYNAMIC_FLOWS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_REGEXP_DYNAMIC_FLOWS
                                                  ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_DYNAMIC_FLOWS
                                                    ) ],
                               },
                     depends => [ 'objects' ],
                    },
                    {
                     policy_part => 'prerequisite_objects',
                     prepare_function => 'prepare_preprequisite_obj_attrs_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_LOGIN_OBJECTS
                                                       PLC.PL_LOGIN_OBJECTS_OMITS_RESPONSE_CODES
                                                       PLC.PL_PREREQUISITE_OBJECT_TARGETS
                                                       PLC.PL_SESSION_AWARENESS_LOGIN_OBJECTS
                                                       PLC.PL_SESSION_AWARENESS_SETTINGS
                                                    ) ],
                               },
                     depends => [ 'objects' ],
                    },
                    {
                     policy_part => 'logout_objects',
                     prepare_function => 'prepare_logout_objects_attrs_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_REVOKE_OBJECTS
                                                    ) ],
                               },
                     depends => [ 'objects' ],
                    },
                    {
                     policy_part => 'brute_force',
                     prepare_function => 'prepare_login_brute_force_object_attr',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES
                                                       PLC.PL_LOGIN_OBJECTS
                                                    ) ],
                               },
                     depends => [ 'objects' ],
                    },
                    {
                     policy_part => 'prerequisite_targets',
                     prepare_function => 'prepare_preprequisite_targets_tbl',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_PREREQUISITE_OBJECT_TARGETS
                                                    ) ],
                               },
                     depends => [ 'objects' ],
                    },
                    {
                     policy_part => 'objects_web_profiles',
                     prepare_function => 'prepare_objects_header_web_profiles',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_OBJECT_CONTENT_PROFILES
                                                    ) ],
                               },
                     depends => [ 'xml_profiles', 'json_profiles', 'gwt_profiles', 'plain_text_profiles', 'graphql_profiles', 'grpc_profiles', 'objects' ],
                    },

                    {
                     policy_part => 'sensitive_parameters',
                     prepare_function => 'prepare_sensitive_params',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_HIDE_PARAM_VALUES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'bot_signature_names',
                     prepare_function => 'prepare_bot_signature_names',
                     # prepare_bot_signature_names handles checksum on its own, since it is policy/account agnostic
                     do_checksums => 0, 
                     tables => {
                                   tables => [ qw(
                                                  PLC.BOT_SIGNATURES
                                                 )],
                               },
                    },
                    {
                     policy_part => 'bot_anomalies',
                     do_checksums => 1, 
                     policy_independent => 1,
                     tables => {
                                source_tables => [ qw(
                                                         PLC.BOT_ANOMALIES
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.BOT_ANOMALIES
                                                  ) ],
                                policy_tables => [ qw(
                                                       PLC.BOT_ANOMALIES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'policy_bot_anomalies',
                     prepare_function => 'prepare_bot_anomalies',
                     do_checksums => 1, 
                     tables => {
                                source_tables => [ qw(
                                                         PLC.PL_BOT_ANOMALIES
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_BOT_ANOMALIES
                                                  ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_BOT_ANOMALIES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'negsig_collections',
                     prepare_function => 'prepare_negsig_collections',
                     # Hardcoded -- only call negsig for now, until MySQL fixes CHECKSUM for adjacent string values
                     do_checksums => 1,
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_PARAM_NEGSIG_SIGNATURES
                                                       PLC.PL_OBJECT_NEGSIG_SIGNATURES
                                                       PLC.PL_HEADER_NEGSIG_SIGNATURES
                                                       PLC.PL_CONTENT_PROFILE_NEGSIG_SIGNATURES
                                                       PLC.PL_POLICY_NEGSIG_ATTRIBUTES
                                                       PLC.PL_POLICY_NEGSIG_SETS
                                                       PLC.PL_POLICY_NEGSIG_SIGNATURES
                                                       PLC.PL_COOKIE_NEGSIG_SIGNATURES
                                                       PLC.PL_POLICIES
                                                       PLC.PL_OBJECTS
                                                       PLC.PL_PARAM_ATTRIBUTES
                                                       PLC.PL_THREAT_CAMPAIGNS
                                                       PLC.PL_POLICY_THREAT_CAMPAIGN_ATTRIBUTES
                                                       PLC.PL_BROWSER_DEFINITIONS
                                                       PLC.PL_BROWSERS
                                                    ) ],
                                tables => [ qw(
                                                DCC.CONTENT_PROFILE_TEMPLATES
                                                PLC.NEGSIG_SIGNATURES
                                                PLC.NEGSIG_SIGNATURE_SYSTEMS
                                                PLC.NEGSIG_SETS
                                                PLC.NEGSIG_SET_FILTERS
                                                PLC.NEGSIG_SET_FILTER_SYSTEMS
                                                PLC.NEGSIG_SET_SIGNATURES
                                                PLC.THREAT_CAMPAIGNS
                                                PLC.THREAT_CAMPAIGNS_MATCH_RULES
                                             ) ],
                               },
                     depends => [ 'xml_profiles', 'parameters', 'objects', 'headers' ],
                    },
                    {
                     policy_part => 'normalization',
                     prepare_function => 'prepare_normalization',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_NORMALIZATION_ATTRIBUTES
                                                       PLC.PL_NORMALIZATION
                                                       PLC.PL_BLOCKING
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'response_scrubbing',
                     prepare_function => 'prepare_response_scrubbing',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_RESPONSE_SCRUBBING
                                                       PLC.PL_RESPONSE_SCRUBBING_PATTERNS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'cleanup_charset_templates',
                     prepare_function => 'cleanup_charset_templates',
                    },
                    {
                     policy_part => 'cleanup_method_templates',
                     prepare_function => 'cleanup_method_templates',
                    },
                    {
                     policy_part => 'negsig_enforcer_names',
                     do_checksums => 0, 
                     # negsig_enforcer_names handles checksum on its own, since it is policy/account agnostic
                     prepare_function => 'prepare_negsig_enforcer_names',
                     tables => {
                                   tables => [ qw(
                                                    PLC.NEGSIG_SIGNATURES
                                                    PLC.NEGSIG_ATTACK_TYPES
                                                    PLC.DECEPTION_TRAP_CONDITIONS
                                                 )],
                               },
                    },
                    {
                     policy_part => 'data_protection',
                     do_checksums => 0,
                     prepare_function => 'prepare_data_protection',
                     tables => {
                                tables => [ qw(
                                                  PLC.DATA_PROTECTION
                                                  PLC.DATA_PROTECTION_CONFIGURATION
                                             )],
                               },
                    },
		    {
		     policy_part => 'credential_stuffing_file',
		     do_checksums => 0,
		     prepare_function => 'prepare_credential_stuffing_file',
		    },
                    {
                     policy_part => 'server_technologies_fingerprints',
                     do_checksums => 0,
                     prepare_function => 'prepare_server_technologies_fingerprints',
                     tables => {
                                   tables => [ qw(
                                                    PLC.SERVER_TECHNOLOGIES
                                                    PLC.SERVER_TECHNOLOGIES_FINGERPRINTS
                                                 )],
                               },
                    },
                    {
                     policy_part => 'threat_campaigns',
                     tables => {
                                source_tables => [ qw( PLC.THREAT_CAMPAIGNS ) ],
                                dest_tables => [ qw( DCC.THREAT_CAMPAIGNS ) ],
                               },
                     row_handler => 'threat_campaigns_row_handler',
                    },
                    {
                     policy_part => 'xff',
                     prepare_function => 'prepare_xff',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_XFF
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'csrf',
                     row_handler => 'csrf_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                         PLC.PL_POLICY_CSRF
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_CSRF
                                                  ) ],

                                policy_tables => [ qw(
                                                       PLC.PL_IP_WHITELIST
                                                       PLC.PL_BLOCKING
                                                       PLC.PL_POLICIES
                                                    ) ],
                               },
                    },
                   {
                     policy_part => 'csrf_urls',
                     row_handler => 'csrf_urls_row_handler',
                     order_by => 'pattern_order, last_updated, target_pattern',
                     tables => {
                                source_tables => [ qw(
                                                         PLC.PL_POLICY_CSRF_URLS
                                                    ) ],
                                dest_tables => [ qw(
                                                      DCC.ACCOUNT_CSRF_URLS
                                                  ) ],

                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_CSRF
                                                    ) ],
                               },
                   },

                   {
                     policy_part => 'csrf_urls_parameters',
                     row_handler => 'csrf_urls_parameters_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                         PLC.PL_POLICY_CSRF_URLS_PARAMETERS
                                                    ) ],
                                dest_tables => [ qw(
                                                      DCC.ACCOUNT_CSRF_URLS_PARAMETERS
                                                  ) ],

                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_CSRF
                                                       PLC.PL_POLICY_CSRF_URLS
                                                       PLC.PL_POLICIES
                                                    ) ],
                               },
                   },
                   {
                     policy_part => 'prerequisite_attributes',
                     prepare_function => 'prepare_prerequisite_attributes',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_PREREQUISITE_ATTRIBUTES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'custom_headers',
                     prepare_function => 'prepare_custom_headers',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_CUSTOM_HEADERS
                                                    ) ],
                               },
                    },
                   {
                     policy_part => 'response_scrubbing_urls',
                     prepare_function => 'prepare_response_scrubbing_urls',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_RESPONSE_SCRUBBING_URLS
                                                    ) ],
                               },
                   },
                   {
                     policy_part => 'case_insensitive',
                     prepare_function => 'prepare_case_insensitive',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_POLICIES
                                                    ) ],
                               },
                   },
                   {
                     policy_part => 'geolocation_enforcement',
                     prepare_function => 'prepare_geolocation_enforcement',
                     tables => {
                                policy_tables => [ qw(
                                                       PLC.PL_GEOLOCATION_ENFORCEMENT
                                                       PLC.PL_GEOLOCATION_ENFORCEMENT_LIST
                                                    ) ],
                               },
                   },
                   {
                     policy_part => 'database_protection',
                     tables => {
                                source_tables => [ qw(
                                                         PLC.PL_DATABASE_PROTECTION
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_DATABASE_PROTECTION
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_DATABASE_PROTECTION
                                                    ) ],
                               },
                    },
                   {
                    policy_part => 'violations',
                    do_checksums => 0, 
                    # violations handles checksum on its own, since it is policy/account agnostic
                    prepare_function => 'prepare_violations',

                    tables => {
                               tables => [ qw(
                                                 PLC.VIOLATIONS
                                                 PLC.VIOLATIONS_PROPERTIES
                                            ) ],
                              },
                   },
                   {
                     policy_part => 'object_allowed_origins',
                     row_handler => 'object_allowed_origins_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                        PLC.PL_OBJECT_ALLOWED_ORIGINS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_OBJECT_CORS_ALLOWED_ORIGINS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_OBJECT_ALLOWED_ORIGINS
                                                       PLC.PL_OBJECTS
                                                    ) ],
                               },
                    },
                   {
                     policy_part => 'object_cors_headers',
                     row_handler => 'object_cors_headers_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                        PLC.PL_OBJECT_EXPOSED_HEADERS
                                                        PLC.PL_OBJECT_ALLOWED_HEADERS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_OBJECT_CORS_HEADERS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_OBJECT_EXPOSED_HEADERS
                                                       PLC.PL_OBJECT_ALLOWED_HEADERS
                                                       PLC.PL_OBJECTS
                                                    ) ],
                               },
                    },
                   {
                     policy_part => 'object_allowed_methods',
                     row_handler => 'object_allowed_methods_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                        PLC.PL_OBJECT_ALLOWED_METHODS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_OBJECT_CORS_ALLOWED_METHODS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_OBJECT_ALLOWED_METHODS
                                                       PLC.PL_OBJECTS
                                                    ) ],
                               },
                    },
                   {
                     policy_part => 'websocket_url_allowed_origins',
                     row_handler => 'websocket_url_allowed_origins_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                       PLC.PL_WEBSOCKET_URL_ALLOWED_ORIGINS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_WEBSOCKET_URL_CORS_ALLOWED_ORIGINS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_WEBSOCKET_URL_ALLOWED_ORIGINS
                                                       PLC.PL_WEBSOCKET_URLS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'redirection_domains',
                     row_handler => 'redirection_domains_row_handler',
                     tables => {
                                source_tables => [ qw(
                                                        PLC.PL_REDIRECTION_DOMAINS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_REDIRECT_DOMAINS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_REDIRECTION_DOMAINS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'policy_builder_configuration',
                     row_handler => 'pb_configuration_row_handler',
                     multiple_source_table_handling => 'join',
                     join_field_name => 'policy_id',
                     tables => {
                                source_tables => [ qw(
                                                        PLC.PL_POLICY_BUILDER_CONFIG
                                                        PLC.PL_POLICY_BUILDER_CONFIG_SESSIONS_AND_LOGINS
                                                        PLC.PL_POLICY_BUILDER_CONFIG_SERVER_TECHNOLOGIES
                                                        PLC.PL_POLICY_BUILDER_CONFIG_COOKIE
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_PB_SETTINGS
                                                    ) ],
                                policy_tables => [ qw(
                                                       PLC.PL_POLICY_BUILDER_CONFIG
                                                       PLC.PL_POLICY_BUILDER_CONFIG_SESSIONS_AND_LOGINS
                                                       PLC.PL_POLICY_BUILDER_CONFIG_SERVER_TECHNOLOGIES
                                                       PLC.PL_POLICY_BUILDER_CONFIG_COOKIE
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'ignored_login_urls',
                     prepare_function => 'prepare_ignored_login_urls',
                     tables => {
                                source_tables => [ qw(
                                                        PLC.PL_SUGGESTIONS
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_LOGIN_IGNORE_URLS
                                                    ) ],
                                policy_tables => [ qw(
                                                        PLC.PL_SUGGESTIONS
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'enforcer_accounts',
                     do_checksums => 0, 
                     prepare_function => 'prepare_enforcer_accounts',
                     tables => {
                                   tables => [ qw(
                                                    DCC.ACCOUNTS
                                                    DCC.LANGUAGE_ENCODINGS
                                                    DCC.ACCOUNT_XFF
                                                    DCC.ACCOUNT_ALTERNATE_RESPONSE_SETTINGS
                                                    DCC.ENCODINGS
                                                 )],
                               },
                    },
                    {
                     policy_part => 'microservices_configuration',
                     prepare_function => 'prepare_microservices',
                     depends => [ 'blocking', 'normalization' ],
                     tables => {
                                source_tables => [ qw(
                                                        PLC.PL_MICROSERVICES
                                                    ) ],
                                dest_tables => [ qw(
                                                       DCC.ACCOUNT_MICROSERVICES
                                                    ) ],
                                policy_tables => [ qw(
                                                        PLC.PL_MICROSERVICES
                                                        PLC.PL_MICROSERVICE_BLOCKING_OVERRIDES
                                                        PLC.PLC.PL_MICROSERVICE_HTTP_CHECKS_OVERRIDES
                                                        PLC.PL_MICROSERVICE_NORMALIZATION_OVERRIDES
                                                    ) ],
                               },
                    },
                    {
                     policy_part => 'server_technology_responses',
                     prepare_function => 'prepare_server_technology_responses',
                     depends => [ 'blocking' ],
                     tables => {
                                policy_tables => [ qw(
                                                        PLC.PL_DECEPTION_RESPONSE_CODES
                                                        PLC.DECEPTION_RESPONSES
                                                        PLC.DECEPTION_RESPONSES_SIGNATURES
                                                    ) ],
                                tables => [ qw(
                                                DCC_ACCOUNT_SERVER_TECHNOLOGY_RESPONSES
                                              ) ],
                               },
                    },
                    {
                     policy_part => 'signature_responses',
                     prepare_function => 'prepare_signature_responses',
                     depends => [ 'blocking' ],
                     tables => {
                                policy_tables => [ qw(
                                                        PLC.PL_SERVER_TECHNOLOGIES
                                                    ) ],
                                tables => [ qw(
                                                DCC.ACCOUNT_SIGNATURE_RESPONSES
                                              ) ],
                               },
                    },
                    {
                     policy_part => 'custom_deception_responses',
                     prepare_function => 'prepare_custom_deception_responses',
                     depends => [ 'blocking' ],
                     tables => {
                                policy_tables => [ qw(
                                                        PLC.PL_DECEPTION_RESPONSES
                                                    ) ],
                                tables => [ qw(
                                                DCC.ACCOUNT_CUSTOM_RESPONSES
                                                DCC.ACCOUNT_CUSTOM_RESPONSE_HEADERS
                                              ) ],
                               },
                    },
                    {
                     policy_part => 'deception_traps',
                     prepare_function => 'prepare_account_deception_traps',
                     tables => {
                                tables => [ qw(
                                                DCC.ACCOUNT_DECEPTION_TRAPS
                                                DCC.ACCOUNT_GLOBAL_PARAMETERS
                                                DCC.ACCOUNT_APPLICATION_OBJECT_ATTRIBUTES
                                                DCC.ACCOUNT_ALLOWED_COOKIES
                                                DCC.NEGSIG_SIGNATURES_NAMES
                                                DCC.ACCOUNT_HEADERS
                                              ) ],
                               },
                    },
                    {
                     policy_part => 'browsers',
                     prepare_function => 'prepare_account_browsers',
                     tables => {
                                policy_tables => [ qw(
                                                        PLC.PL_BROWSERS
                                                    ) ],
                                tables => [ qw(
                                                DCC.ACCOUNT_FACTORY_BROWSER_PROPERTIES
                                              ) ],
                               },
                    },
                    {
                     policy_part => 'negsig_attributes',
                     tables => {
                                source_tables => [ qw(
                                                        PLC.PL_POLICY_NEGSIG_ATTRIBUTES
                                                    ) ],
                                dest_tables => [ qw(
                                                        DCC.ACCOUNT_NEGSIG_ATTRIBUTES
                                                    ) ],
                                policy_tables => [ qw(
                                                        PLC.PL_POLICY_NEGSIG_ATTRIBUTES
                                                    ) ],

                               },
                    },
                    {
                     policy_part => 'bot_defense_class',
                     prepare_function => 'prepare_bot_defense_class',
                     tables => {
                               policy_tables => [ qw(
                                                     PLC.PL_BOT_CLASS
                                                    )
                                                ],
                               },
                    },
                    {
                     policy_part => 'bot_defense',
                     tables => {
                               source_tables => [ qw(
                                                        PLC.PL_BOT_DEFENSE
                                                    ) ],
                               dest_tables => [ qw(
                                                        DCC.ACCOUNT_BOT_DEFENSE
                                                    ) ],
                               policy_tables => [ qw(
                                                        PLC.PL_BOT_DEFENSE
                                                    ) ],
                               }
                    }
                  ];

#--------------------------------------------------------------------------------

# inits the enviroment and reads from config files
sub init
{
    my ($self, %params) = @_;

    $self = $self->SUPER::init(%params);
    $self->reset_object_data();
    my %defaults = ( prepare_negsig_exclusions => 1,
                     generate_negsig_collections => 1,
                     unescape => 0,
                     policy_name => undef,
                     force_policy_name_flg => 0,
                     negative_signatures => undef,
		     refresh_negsig_exclusions => {},
                   );

    for my $k (keys %defaults) {
        if (exists $params{$k}) {
            $self->{$k} = $params{$k};
        }
        else {
            $self->{$k} = $defaults{$k};
        }
    };

    $self->{checksums} = F5::PrepareConf::Checksums->new(dbh => $self->{dbh},
                                                         logger => $self->{logger});

    $self->{captcha_refresh_base64_encoded_image} = File::Slurp::read_file($CAPTCHA_REFRESH_BASE64_ENCODED_IMAGE);
    $self->{captcha_audio_icon_base64_encoded_image} = File::Slurp::read_file($CAPTCHA_AUDIO_ICON_BASE64_ENCODED_IMAGE);

    return $self;
}

sub reset_object_data {
    my ($self) = @_;

    delete $self->{is_policy_transparent};
    delete $self->{is_policy_case_insensitive};
    delete $self->{is_learning_enabled};
    delete $self->{global_subviolation_attribute}->{max_decoding_passes};
    delete $self->{global_subviolation_attribute}->{max_headers};
    delete $self->{global_subviolation_attribute}->{max_params};
    delete $self->{global_subviolation_attribute}->{max_cookies};
    delete $self->{is_policy_passive};
    delete $self->{is_bot_defense_enabled};
    delete $self->{bot_action_blocking_mask};
    delete $self->{max_profile_templ_id};
    $self->{h_objects} = {};
    $self->{h_headers} = {};
    $self->{h_websocket_urls} = {};
    $self->{h_flows} = {};
    $self->{h_keys} = {};
    $self->{h_xml_profiles} = {};
    $self->{h_json_profiles} = {};
    $self->{h_gwt_profiles} = {};
    $self->{h_grpc_profiles} = {};
    $self->{h_plain_text_profiles} = {};
    $self->{h_header_parameters} = {};
    $self->{charset_templates} = {};
    $self->{base_charsets} = {};
    $self->{viols} = 0;
    $self->{http_check_viols} = 0;
    $self->{normalization_viols} = 0;
    $self->{redirect_action} = {};
    $self->{allowed_methods_templates} = {};
    $self->{csrf_urls} = {};
    $self->{csrf_urls_dcc_pattern_order} = 1;
    $self->{deception_settings} = {};
    $self->{h_cookie_parameters} = {};
}

#--------------------------------------------------------------------------------

sub prepare {
  my ($self, %params) = @_;

  $self->reset_object_data();

  my $account_id = $params{account_id};
  my $policy_name_crc = $params{policy_name_crc};
  my $plc_id = $params{policy_id};
  my $ucs_upgrade_real_policy_id =  $params{ucs_upgrade_real_policy_id};
  my $extra_prepare_params = $params{extra_prepare_params} || {};

  # Table checksums in MariaDB 5.5 are significantly slower than in MySQL 5.1.
  # Coupled with other performance improvements in signature handling, checksums
  # should never be used. See https://jira.pdsea.f5net.com/browse/ASMMNG-1115
  my $do_table_checksums = 0;
  my $policy_parts_to_prepare;

  if (exists $params{policy_parts_to_prepare}) {
    my %policy_parts_to_prepare = map { ($_ => 1) } @{$params{policy_parts_to_prepare}};
    $policy_parts_to_prepare = [ map { dclone($_) } grep { $policy_parts_to_prepare{$_->{policy_part}} } @$policy_parts ];
  }
  else {
    $policy_parts_to_prepare = $policy_parts;
  }

  if (! $self->{dbh}) {
    $self->{logger}->error(ERRMSG_DBH_LOST);
    return undef;
  }

    if ( !$policy_name_crc && $account_id ){
        ($policy_name_crc) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $account_id);
SELECT policy_name_crc
  FROM DCC.ACCOUNTS
 WHERE account_id = ?
EndQuery
    }

  # chdir to work directory
  chdir(dirname($0));

  $self->{logger}->info("Start");

  for my $policy_part (@$policy_parts_to_prepare) {
    $self->{logger}->debug("Preparing $policy_part->{policy_part}") if $self->{logger};

    # Check tables with checksum
    my $have_tables_changed = 1;
    if ($do_table_checksums && $policy_part->{do_checksums} && 
        $self->get_flg_encoding_selected($policy_name_crc)) {
      $have_tables_changed = $self->{checksums}->have_tables_changed(policy_id => $params{policy_id},
                                                                     policy_tables => $policy_part->{policy_tables},
                                                                     tables => $policy_part->{tables});
    }

    my $prepare_function = $policy_part->{prepare_function} || 'prepare_generic';

    my $flg_encoding_selected = $self->get_flg_encoding_selected($policy_name_crc);
    my %function_params = ( flg_encoding_selected => $flg_encoding_selected,
                            policy_part => $policy_part,
                            %$extra_prepare_params,
                           );
    if ($params{active_policy_was_changed} || $have_tables_changed) {
      # If checksum differs, call prepare_function
      $self->$prepare_function($policy_name_crc, $params{policy_id}, {%function_params});
    }
    elsif ($policy_part->{has_readonly_mode}) {
      $self->$prepare_function($policy_name_crc, $params{policy_id},
                               {%function_params, 
                                readonly_mode => 1,});
    }
  }

  if ($do_table_checksums) {
    $self->{checksums}->store(policy_id => $params{policy_id});
  }

  $self->{logger}->info("End");

  return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_websocket_urls_tbl {
    my ($self, $policy_name_crc, $plc_id, $params) = @_;
    my $dbh = $self->{dbh};
    my $readonly_mode = $params->{readonly_mode};

    my $ws_dcc_table = 'DCC.ACCOUNT_WEBSOCKET_URL';
    my $ws_wildcard_dcc_table = 'DCC.ACCOUNT_WEBSOCKET_URL_WILDCARD';
    local $" = ',';
    unless ($readonly_mode) {
        $self->empty_table($ws_dcc_table, $policy_name_crc);
        $self->empty_table($ws_wildcard_dcc_table, $policy_name_crc);
    }

    if ( $plc_id == 0 || $readonly_mode ) {
        return RET_OK;
    }

    my @plc_fields = qw(websocket_url_id protocol uri flg_is_pattern message_max_binary_size message_max_frame_size message_max_frames
                        websocket_json_profile_id websocket_text_profile_id flg_is_staging websocket_extension flg_in_classification
                        flg_wildcard_includes_slash pattern_order flg_tightening
                        flg_websocket_message_format_text flg_websocket_message_format_json flg_websocket_message_format_binary
                        websocket_json_profile_id websocket_text_profile_id flg_enable_cors_enforcement
                        flg_check_websocket_url_metachar flg_check_payload
                        flg_check_message_max_binary_size flg_check_message_max_frame_size flg_check_message_max_frames flg_allow_cors);

    my $plc_sth = $dbh->prepare("SELECT CASE WHEN (flg_is_allowed = 1) THEN 0 ELSE 1 END as flg_is_disallowed, @plc_fields from PLC.PL_WEBSOCKET_URLS
                                 WHERE policy_id = ? ORDER BY pattern_order,last_updated,uri")
        or $self->{logger}->error("DBI->prepare failed: $DBI::errstr");

    $plc_sth->execute($plc_id) or $self->{logger}->error("DBI->execute failed: $DBI::errstr");
    my $dcc_pattern_order = 1;

    while (my $plc_row = $plc_sth->fetchrow_hashref) {
        my $flg_is_pattern = $plc_row->{flg_is_pattern};
        $plc_row->{policy_name_crc} = $policy_name_crc;
        $plc_row->{websocket_url_crc} = F5::CRC::get_crc_any("get_crc_obj", $plc_row->{protocol}, $policy_name_crc, $plc_row->{uri}, $self->{unescape});

        $plc_row->{message_max_binary_size} = $plc_row->{flg_check_message_max_binary_size} ? $plc_row->{message_max_binary_size} : 0;
        $plc_row->{message_max_frame_size} = $plc_row->{flg_check_message_max_frame_size} ? $plc_row->{message_max_frame_size} : 0;
        $plc_row->{message_max_frames} = $plc_row->{flg_check_message_max_frames} ? $plc_row->{message_max_frames} : 0;

        for my $row_key (keys %$plc_row) {
            $self->{h_websocket_urls}->{ $plc_row->{websocket_url_id} }->{$row_key} = $plc_row->{$row_key};
        }

        my @dcc_fields = qw(policy_name_crc protocol websocket_url_crc flg_is_disallowed flg_is_staging
                            message_max_binary_size message_max_frame_size message_max_frames websocket_text_profile_id
                            websocket_json_profile_id websocket_extension flg_in_classification
                            flg_websocket_message_format_text flg_websocket_message_format_json flg_websocket_message_format_binary
                            flg_enable_cors_enforcement flg_check_payload flg_allow_cors);
        my $dcc_table_name;
        if ($flg_is_pattern) {
            if ($plc_row->{flg_check_websocket_url_metachar}) {
                $plc_row->{websocket_url_charset_templ_id} = $self->get_charset_template(
                                                                               policy_name_crc => $policy_name_crc,
                                                                               base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_URI(),
                                                                               override_table => 'PLC.PL_WEBSOCKET_URL_METACHARS',
                                                                               override_id_field => 'websocket_url_id',
                                                                               override_id_value => $plc_row->{websocket_url_id});
            }
            else {
                $plc_row->{websocket_url_charset_templ_id} = 0;
            }

            push @dcc_fields, qw(flg_tightening pattern pattern_order flg_wildcard_includes_slash websocket_url_charset_templ_id flg_check_websocket_url_metachar);
            $plc_row->{pattern} = $plc_row->{uri};
            $plc_row->{pattern_order} = $dcc_pattern_order++;
            $dcc_table_name = $ws_wildcard_dcc_table;
        }
        else {
            $dcc_table_name = $ws_dcc_table;
        }

        $plc_row->{websocket_text_profile_id} = $self->{h_websocket_urls}->{$plc_row->{websocket_url_id}}->{websocket_text_profile_templ_id} || 0;
        $plc_row->{websocket_json_profile_id} = $self->{h_websocket_urls}->{$plc_row->{websocket_url_id}}->{websocket_json_profile_templ_id} || 0;

        if ($plc_row->{flg_tightening} == F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION) {
           $plc_row->{flg_tightening} = 0;
        }

        $self->insert_row(table_name => $dcc_table_name,
                          fields => \@dcc_fields,
                          rec => $plc_row);
    }

    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_objects_tbl {
  my ($self, $policy_name_crc, $plc_id, $params) = @_;
  my $dbh = $self->{dbh};

  my $readonly_mode = $params->{readonly_mode};
  my $flg_encoding_selected = defined $params->{flg_encoding_selected} ? $params->{flg_encoding_selected} : $self->get_flg_encoding_selected($policy_name_crc);
  my $is_policy_passive =  $self->is_policy_passive(policy_id => $plc_id);

  local $" = ',';
  unless ($readonly_mode) {
    $self->empty_table('DCC.ACCOUNT_APPLICATION_OBJECT_ATTRIBUTES', $policy_name_crc);
    $self->empty_table('DCC.ACCOUNT_APPLICATION_WILDCARD_OBJECT_ATTRIBUTES', $policy_name_crc);
  }

  return RET_OK if $plc_id == 0;

  my %urls_with_extra_content_profile =  map { ($_ => 1) } @{$dbh->selectcol_arrayref(<<EndQuery,undef, $plc_id)};
SELECT DISTINCT obj_id
  FROM PLC.PL_OBJECT_CONTENT_PROFILES
 WHERE policy_id = ?
   AND header_order > 3
EndQuery

  my $default_content_profiles = $dbh->selectall_hashref(<<EndQuery,'content_profile_type',undef,$plc_id);
SELECT * from PLC.PL_CONTENT_PROFILES
WHERE policy_id = ?
AND profile_name = 'Default'
EndQuery

  my $content_type_json = F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON();
  my $content_type_xml = F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML();
  my $default_xml_profile_id = (exists $default_content_profiles->{$content_type_xml} ) ? $default_content_profiles->{$content_type_xml}->{id} : 0;
  my $default_json_profile_id = (exists $default_content_profiles->{$content_type_json}) ? $default_content_profiles->{$content_type_json}->{id} : 0;

  my $default_url_content_profiles = F5::ASMConfig::Entity::URL->get_default_url_content_profiles_db_rec( policy_id => $plc_id,
                                                                                                          dbh => $dbh,
                                                                                                          default_xml_profile_id => $default_xml_profile_id,
                                                                                                          default_json_profile_id => $default_json_profile_id,
                                                                                                         );

  my %url_ids;
  for my $default_url_content_profile ( @$default_url_content_profiles ){
     $default_url_content_profile->{policy_id} = $plc_id;
     my $where_clause = F5::DbUtils::construct_where_clause(where_fields => $default_url_content_profile);
     $url_ids{$default_url_content_profile->{enforcement_type}} = $dbh->selectall_hashref(<<EndQuery,'obj_id', undef, @{$where_clause->{where_values}});
SELECT obj_id
  FROM PLC.PL_OBJECT_CONTENT_PROFILES
 WHERE $where_clause->{where_clause}
EndQuery
  }

  my %urls_with_defaults;
  my @urls_with_non_default_configuration;

  for my $url_id ( keys %{$url_ids{apply_value_signatures}}){
      if ( $url_ids{http}->{$url_id} &&
           $url_ids{xml}->{$url_id} &&
           $url_ids{json}->{$url_id} &&
           !$urls_with_extra_content_profile{$url_id} ) {

             $urls_with_defaults{$url_id} = 1 ;
      }
  }

  my $method_codes =  $dbh->selectall_hashref(<<EndQuery, 'method_name');
SELECT method_name, method_code
  FROM DCC.METHOD_POOL
EndQuery

  my @plc_fields = qw(OBJ.id uri flg_check_flow flg_is_referrer flg_cookie_change 
		      flg_entry_point flg_parse_dynamic_params app_id
		      flg_is_pattern pattern_order flg_tightening
		      flg_check_obj_metachar flg_check_obj_methods flg_is_staging last_updated
                      flg_clickjacking_protection clickjacking_type clickjacking_from flg_wildcard_includes_slash
                      flg_enable_cors_enforcement flg_check_credentials flg_allow_credentials flg_check_maximum_age maximum_age
                      allowed_origins_enforcement_mode flg_allowed_methods_check flg_allowed_headers_check flg_exposed_headers_check flg_allow_cors
                      OCP.enforcement_type OCP.profile_id flg_in_classification flg_check_signatures method_name operation_id flg_mandatory_body flg_disallow_file_upload_of_executables);

  my $plc_sth = $dbh->prepare("SELECT 
  EXISTS(   select 1
            from PLC.PL_OBJECT_POSITIONAL_PARAMETERS
            where object_id = OBJ.id) as flg_has_positional_params,
        (   select count(*) 
            from PLC.PL_LOGIN_OBJECTS
            where object_id = OBJ.id) as flg_is_prerequisite, 
        (   select count(*)
            from PLC.PL_SESSION_AWARENESS_LOGIN_OBJECTS
            where object_id = OBJ.id ) as flg_is_session_awareness,
        (   select count(*)
            from PLC.PL_SESSION_AWARENESS_SETTINGS AS SA, PLC.PL_LOGIN_OBJECTS as LP
            where LP.policy_id = SA.policy_id AND LP.object_id = OBJ.id
            AND SA.flg_session_awareness_on = ? AND SA.application_username_type = ?) as flg_is_default_session_awareness,
        (   select count(*)
            from PLC.PL_REVOKE_OBJECTS
            where revoke_object_id = OBJ.id) as flg_is_revoke,
        (   select count(*) 
            from PLC.PL_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES B, PLC.PL_LOGIN_OBJECTS C
            where B.object_id = OBJ.id AND B.object_id = C.object_id) as flg_is_brute_force,
        (   select count(*)
            from PLC.PL_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES B, PLC.PL_LOGIN_OBJECTS C
            where B.flg_is_default = 1 AND OBJ.id = C.object_id and B.flg_enabled = 1 and C.auth_type != ? and B.policy_id = ?) as flg_is_default_brute_force,
        (   select max(dynamic_param_type)
            from PLC.PL_EXTRACT_ATTRIBUTES A, PLC.PL_OBJECT_EXTRACTS B
            where A.id = B.extract_attrib_id AND B.obj_id = OBJ.id) as dynamic_param_type,
        (CASE WHEN (flg_is_allowed = 1) THEN 0 ELSE 1 END) as flg_is_disallowed,
        @plc_fields 
        FROM PLC.PL_OBJECTS as OBJ
   LEFT JOIN PLC.PL_OBJECT_CONTENT_PROFILES as OCP
          ON (    OCP.obj_id =  OBJ.id
              AND OCP.policy_id = OBJ.policy_id
              )
     WHERE OBJ.policy_id = ?
       AND OCP.header_order = ? 
     ORDER BY OBJ.pattern_order,OBJ.last_updated,OBJ.uri")
    or $self->{logger}->error("DBI->prepare failed: $DBI::errstr",);

  $plc_sth->execute(1, F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_ALL_LOGIN_PAGES(),
                    F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_NONE(), $plc_id, $plc_id, 0) 
                    or $self->{logger}->error("DBI->execute failed: $DBI::errstr");

  my $exclude_negsig_urls = $dbh->selectall_hashref(<<EndQuery, 'obj_id', undef, $plc_id);
SELECT DISTINCT obj_id
    FROM PLC.PL_OBJECT_NEGSIG_SIGNATURES
    WHERE policy_id = ?
EndQuery

  my $dcc_pattern_order = 1;
  while (my $plc_row = $plc_sth->fetchrow_hashref) {
    my $obj_id = $plc_row->{id};

    my ($max_param_index) = 0;
    # first handle positional_params relalted stuff
    #     cause it might result in rewriting the url
    if ($plc_row->{flg_is_pattern} && $plc_row->{flg_has_positional_params}) {
         ($max_param_index) = $dbh->selectrow_array(<<EndQuery, undef, $obj_id);
SELECT max(url_segment_index)
  FROM PLC.PL_OBJECT_POSITIONAL_PARAMETERS
 WHERE object_id = ?
EndQuery
    }

    my $method_code;
    if ($plc_row->{method_name} && $plc_row->{method_name} ne $ANY_URL_METHOD_NAME) {
        $method_code = $method_codes->{$plc_row->{method_name}}->{method_code};
        $plc_row->{method_code} = $method_code;
    }

    my $obj_crc = get_url_crc(policy_name_crc => $policy_name_crc,
                          app_id => $plc_row->{app_id},
                          unescape => $self->{unescape},
                          url_name => $plc_row->{uri},
                          method_code => $method_code, );

    $self->{h_objects}->{$obj_id}->{flg_is_pattern} = $plc_row->{flg_is_pattern};
    $self->{h_objects}->{$obj_id}->{flg_has_positional_params} = $plc_row->{flg_has_positional_params};
    $self->{h_objects}->{$obj_id}->{objcrc} = $obj_crc;
    $self->{h_objects}->{$obj_id}->{obj} = $plc_row->{uri};
    $self->{h_objects}->{$obj_id}->{app_id} = $plc_row->{app_id};
    $self->{h_objects}->{$obj_id}->{flg_enable_cors_enforcement} = $plc_row->{flg_enable_cors_enforcement};
    $self->{h_objects}->{$obj_id}->{allowed_origins_enforcement_mode} = $plc_row->{allowed_origins_enforcement_mode};
    $self->{h_objects}->{$obj_id}->{flg_allowed_methods_check}  = $plc_row->{flg_allowed_methods_check};
    $self->{h_objects}->{$obj_id}->{flg_allowed_headers_check}  = $plc_row->{flg_allowed_headers_check};
    $self->{h_objects}->{$obj_id}->{flg_exposed_headers_check}  = $plc_row->{flg_exposed_headers_check};
    $self->{h_objects}->{$obj_id}->{flg_in_classification}  = $plc_row->{flg_in_classification};

    $plc_row->{flg_disallow_cors} = !$plc_row->{flg_allow_cors}; #BD wants to get flg 'disallow_cors'
    $self->{h_objects}->{$obj_id}->{flg_disallow_cors}  = $plc_row->{flg_disallow_cors};
    $plc_row->{excluded_signatures_exist} = exists $exclude_negsig_urls->{$obj_id} ? 1 : 0;
    if ( $urls_with_defaults{$obj_id} ){
        $plc_row->{default_xml_profile_id} = $self->get_content_profile_id( enforcement_type => 'xml', profile_id => $default_xml_profile_id);
        $plc_row->{default_json_profile_id} = $self->get_content_profile_id( enforcement_type => 'json', profile_id => $default_json_profile_id);
    }
    else{
        push(@urls_with_non_default_configuration,$obj_id);
    }

    if ($readonly_mode) {
      next;
    }

    $plc_row->{flg_is_brute_force} = $plc_row->{flg_is_brute_force} || $plc_row->{flg_is_default_brute_force};
    $plc_row->{flg_is_session_awareness} = $plc_row->{flg_is_session_awareness} || $plc_row->{flg_is_default_session_awareness};

    if (! $flg_encoding_selected) {
        $plc_row->{flg_tightening} = 0;
        $plc_row->{flg_is_staging} = 0;
    }

    if ($plc_row->{flg_tightening} == F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION) {
        $plc_row->{flg_tightening} = 0;
    }

    if (! $plc_row->{flg_clickjacking_protection}) {
        $plc_row->{clickjacking_type} = 0;
    }

    if ($plc_row->{clickjacking_type} == F5::ASMConfig::ClickjackingType::CLICKJACKING_ONLY_FROM) {
         $plc_row->{clickjacking_from} = "ALLOW-FROM " . $plc_row->{clickjacking_from};
    }
    else {
        $plc_row->{clickjacking_from} = '';
    }

    $plc_row->{policy_name_crc} = $policy_name_crc;
    $plc_row->{object_crc} = $obj_crc;
    $plc_row->{content_profile_id} = $self->get_content_profile_id(%$plc_row);

    $plc_row->{type_id} = $enforcement_type_id{$plc_row->{enforcement_type}};

    if ( $is_policy_passive ) {
        $plc_row->{flg_check_flow} = 0;
        $plc_row->{flg_entry_point} = 0;
        $plc_row->{flg_is_referrer} = 0;
        $plc_row->{flg_enable_cors_enforcement} = 0;
        $plc_row->{flg_disallow_cors} = 0;
    }

    my $object_settings = {
        flg_is_prerequisite                 => 2**0, 
        flg_is_revoke                       => 2**1, 
        flg_entry_point                     => 2**2, 
        flg_is_referrer                     => 2**3, 
        flg_check_flow                      => 2**4,
        flg_cookie_change                   => 2**5, 
        flg_is_applies_to_renewal_needed    => 2**6, 
        flg_is_brute_force                  => 2**7, 
        flg_is_appies_to                    => 2**8,
        flg_is_disallowed                   => 2**9, 
        flg_csrf_renewal_needed             => 2**10, 
        flg_csrf_entry_policy               => 2**11, 
        flg_is_session_awareness            => 2**12,
        flg_enable_cors_enforcement         => 2**13,
        flg_check_credentials               => 2**14, 
        flg_allow_credentials               => 2**15,
        flg_check_maximum_age               => 2**16,
        flg_allowed_methods_check           => 2**17,
        flg_allowed_headers_check           => 2**18,
        flg_exposed_headers_check           => 2**19,
        flg_disallow_cors                   => 2**20,
        flg_mandatory_body                  => 2**21,
        placeholder_query_as_value          => 2**$QUERY_AS_VALUE_BIT,  # 22 - this bit updated by parameter entity in case param value is whole query string
        flg_disallow_file_upload_of_executables => 2**23,
    };

    $plc_row->{object_settings_bitmap} = 0;

    for my $k (keys %$object_settings) {
       if ($plc_row->{$k}) {
          $plc_row->{object_settings_bitmap} |= $object_settings->{$k};
       }
    }

    $plc_row->{profile_templ_id} = $self->{h_objects}->{$obj_id}->{profile_templ_id} || 0;

    my $dcc_table_name;
    $plc_row->{obj_methods_templ_id} = 0;
    if ($plc_row->{flg_check_obj_methods} || $method_code) {

        $plc_row->{obj_methods_templ_id} = $self->find_or_add_allowed_methods_template(
                                                                       policy_name_crc => $policy_name_crc,
                                                                       obj_id => $obj_id,
                                                                       policy_id => $plc_id,
                                                                       method_code => $method_code,
                                                                       );
    }

    if ($plc_row->{flg_is_pattern}) {
      $dcc_table_name = 'DCC.ACCOUNT_APPLICATION_WILDCARD_OBJECT_ATTRIBUTES';
      $plc_row->{pattern_order} = $dcc_pattern_order++; 
      if ($plc_row->{flg_check_obj_metachar}) {
        $plc_row->{obj_charset_templ_id} = $self->get_charset_template(
                                                                       policy_name_crc => $policy_name_crc,
                                                                       base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_URI(),
                                                                       override_table => 'PLC.PL_OBJECT_METACHARS',
                                                                       override_id_field => 'obj_id',
                                                                       override_id_value => $plc_row->{id});
      }
      else {
        $plc_row->{obj_charset_templ_id} = 0;
      }
      $plc_row->{negative_check_bitmap} = get_negative_check_bitmap(name_metachar_flag => $plc_row->{flg_check_obj_metachar});
      $plc_row->{pattern} = $plc_row->{uri};
      $plc_row->{positional_param_last_segment} = $max_param_index;
    }
    else {
      $dcc_table_name = 'DCC.ACCOUNT_APPLICATION_OBJECT_ATTRIBUTES';
      $plc_row->{flg_parse_dynamic_params} = $plc_row->{dynamic_param_type} if $plc_row->{dynamic_param_type};
    }

    $self->insert_row(
                      table_name => $dcc_table_name,
                      rec => $plc_row);

 }

  $self->{h_objects}->{urls_with_non_default_configuration} = \@urls_with_non_default_configuration; 
  return RET_OK;
}

#--------------------------------------------------------------------------------
sub prepare_flow_tbl {
  my ($self, $policy_name_crc, $plc_id, $params) = @_;

  my $dbh = $self->{dbh};
  my $readonly_mode = $params->{readonly_mode};

  unless ($readonly_mode) {
    $self->empty_table('DCC.ACCOUNT_APPLICATION_FLOW', $policy_name_crc);
  }
  return RET_OK if $plc_id == 0;

  my @pl_flows_fields = qw(id from_id to_id method flg_pd flg_qs flg_check_pd
		      flg_check_qs flg_unicode_encoding frame_index);
  local $" = ',';
  my $pl_flows_query = <<"EndQuery";

SELECT @pl_flows_fields, 
       IFNULL(mandatory_count,0) AS calc_num_mandatory_params,
       IFNULL(dynamic_param_name_count,0) AS calc_num_dynamic_param_name
  FROM PLC.PL_FLOWS 
       LEFT JOIN 
          (SELECT flow_id, COUNT(*) as mandatory_count
             FROM PLC.PL_FLOW_PARAMS flow_params,
                  PLC.PL_PARAM_ATTRIBUTES param_attrib
            WHERE flg_mandatory = 1
              AND flow_params.policy_id = ?
              AND flow_params.param_attrib_id = param_attrib.id
         GROUP BY flow_id) mandatory_params
       ON PL_FLOWS.id = mandatory_params.flow_id
       LEFT JOIN 
          (SELECT flow_id, COUNT(*) as dynamic_param_name_count
             FROM PLC.PL_FLOW_PARAMS flow_params,
                  PLC.PL_PARAM_ATTRIBUTES param_attrib
            WHERE param_attrib.key_input_type = ?
              AND flow_params.policy_id = ?
              AND flow_params.param_attrib_id = param_attrib.id
         GROUP BY flow_id) dynamic_param_name_params
       ON PL_FLOWS.id = dynamic_param_name_params.flow_id
 WHERE PL_FLOWS.policy_id = ?
EndQuery

  my $plc_sth = $dbh->prepare($pl_flows_query)
    or $self->{logger}->error('DBI->prepare',$DBI::errstr);

  $plc_sth->execute($plc_id, DYNAMIC_PARAMETER_NAME, $plc_id, $plc_id)
    or $self->{logger}->error('DBI->execute',$DBI::errstr);

  while (my $plc_row = $plc_sth->fetchrow_hashref) {
    if ( ($plc_row->{from_id} != 0) and 
	 (
	  (not exists $self->{h_objects}->{$plc_row->{from_id}}) or 
	  (not defined $self->{h_objects}->{$plc_row->{from_id}}->{objcrc})
	 )
       ) {
      # if the object crc is undefined and the object isn't
      # an entry point then this is an undefined object and
      # so there is probably a problem in the policy or a bug
      # in export policy, import policy or prepare policy.
      $self->{logger}->error("PLC.PL_FLOWS.from_id has no PLC.PL_OBJECTS.id. Undefined obj_id $plc_row->{from_id}!");
      next;
    }
    # calculate the arch crc
    my $from_crc = ($plc_row->{from_id}) ? $self->{h_objects}->{$plc_row->{from_id}}->{objcrc} : 0; # if the from_obj_id==0 -> from_objcrc=0 -> virtual flow!
    my $to_crc = $self->{h_objects}->{$plc_row->{to_id}}->{objcrc};
    my $to_obj = $self->{h_objects}->{$plc_row->{to_id}}->{obj};
    if( (not defined $to_obj) or ($to_obj eq '')) {
      $self->{logger}->error("obj_id2obj. Undefined obj_id $plc_row->{to_id}!");
      next;
    }

    my $flowcrc = F5::CRC::get_crc_any("get_crc_flow", $from_crc, $to_crc, $plc_row->{method}, $self->{unescape});
    $self->{h_flows}->{$plc_row->{id}}->{flowcrc} = $flowcrc;
    $self->{h_flows}->{$plc_row->{id}}->{from_id} = $plc_row->{from_id};
    $self->{h_flows}->{$plc_row->{id}}->{to_id} = $plc_row->{to_id};
    if ($readonly_mode) {
      next;
    }

    my $input_bitmap = input_flags2bitmap($plc_row->{flg_pd}, $plc_row->{flg_qs},
					  $plc_row->{flg_check_pd}, $plc_row->{flg_check_qs});

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_APPLICATION_FLOW',
                      rec => { policy_name_crc => $policy_name_crc,
                               arch_crc => $flowcrc,
                               input_attributes => $input_bitmap,
                               encoding => $plc_row->{flg_unicode_encoding},
                               num_mandatory_params => $plc_row->{calc_num_mandatory_params},
                               frame_index => $plc_row->{frame_index},
                               flg_has_dynamic_param_name => $plc_row->{calc_num_dynamic_param_name} });
  }

  return RET_OK;
}

#--------------------------------------------------------------------------------

sub get_url_crc {
    my (%params) = @_;

    my $policy_name_crc = $params{policy_name_crc};
    my $app_id = $params{app_id};
    my $url_name = $params{url_name};
    my $method_code = $params{method_code};
    my $unescape = $params{unescape};

    my $obj_crc;
    if ($unescape) {
        $obj_crc = F5::CRC::get_crc_any("get_crc_obj", $app_id, $policy_name_crc, $url_name, $unescape);
    }
    else {
        $obj_crc = F5::CRC::get_crc_obj($app_id, $policy_name_crc, $url_name);
    }

    if ($method_code) {
        $obj_crc = F5::CRC::get_crc_int($obj_crc, $method_code);
    }

    return $obj_crc;
}

#----------------------------------------------------------------------

sub prepare_parameters {
  my ($self, $policy_name_crc, $policy_id, $params) = @_;
  my $dbh = $self->{dbh};
  my $readonly_mode = $params->{readonly_mode};
  my $flg_encoding_selected = defined $params->{flg_encoding_selected} ? $params->{flg_encoding_selected} : $self->get_flg_encoding_selected($policy_name_crc);

  my @parameter_types = ( { plc_table => 'PL_FLOW_PARAMS',
			    extra_plc_fields => 'flow_id',
			    dcc_table => 'DCC.ACCOUNT_FLOW_PARAMETERS',
			    dcc_wildcard_table => 'DCC.ACCOUNT_WILDCARD_FLOW_PARAMETERS',
			    wildcard_crc_field => 'flow_crc',
			    get_wildcard_crc => sub {
			      my $param = shift;
			      return $self->{h_flows}->{$param->{flow_id}}->{flowcrc};
			    },
			    compute_keycrc => sub { 
			      my ($self,$param) = @_;
			      my $flow_crc = $self->{h_flows}->{$param->{flow_id}}->{flowcrc};
			      if (!$flow_crc) {
				$self->{logger}->error("Undefined flow id $param->{flow_id}!");
				return;
			      }

			      return F5::CRC::get_crc_any("get_crc_str", $flow_crc, $param->{name}, $self->{unescape});
			    } },
			  { plc_table => 'PL_OBJECT_PARAMS',
			    extra_plc_fields => 'obj_id',
			    dcc_table => 'DCC.ACCOUNT_OBJECT_PARAMETERS',
			    dcc_wildcard_table => 'DCC.ACCOUNT_WILDCARD_OBJECT_PARAMETERS',
			    dcc_mandatory_parameters_table => 'DCC.ACCOUNT_OBJECT_MANDATORY_PARAMETERS',
			    dcc_query_as_value_parameters_table => 'DCC.ACCOUNT_OBJECT_QUERY_AS_VALUE_PARAMETERS',
			    dcc_mandatory_crc_fields => {
			        obj_crc => sub {
			            my $param = shift;
			            return $self->{h_objects}->{$param->{obj_id}}->{objcrc};
			        },
			        object_param_key_crc => sub {
			            my $param = shift;
			            return $param->{key_crc};
			        },
			    },
			    wildcard_crc_field => 'object_crc',
			    get_wildcard_crc => sub {
			      my $param = shift;
			      return $self->{h_objects}->{$param->{obj_id}}->{objcrc};
			    },
			    compute_keycrc => sub { 
			      my ($self,$param) = @_;
			      my $obj_crc = $self->{h_objects}->{$param->{obj_id}}->{objcrc};
			      if (!$obj_crc) {
				$self->{logger}->error("Undefined obj id $param->{obj_id}!");
				return;
			      }

			      return F5::CRC::get_crc_any("get_crc_str", $obj_crc, $param->{name}, $self->{unescape});
			    } },
			  { plc_table => 'PL_GLOBAL_PARAMS',
			    dcc_table => 'DCC.ACCOUNT_GLOBAL_PARAMETERS',
			    dcc_wildcard_table => 'DCC.ACCOUNT_WILDCARD_GLOBAL_PARAMETERS',
			    dcc_mandatory_parameters_table => 'DCC.ACCOUNT_GLOBAL_MANDATORY_PARAMETERS',
			    dcc_mandatory_crc_fields => {
			        global_param_key_crc => sub {
			            my $param = shift;
			            return $param->{key_crc};
			        },
			    },
			    compute_keycrc => sub { 
			      my ($self,$param) = @_;
			      return F5::CRC::get_crc_any("get_crc_str", $policy_name_crc, $param->{name}, $self->{unescape});
			    } },
			);

  unless ($readonly_mode) {
    $self->empty_table('DCC.ACCOUNT_PARAM_VALUE_RANGES', $policy_name_crc);
    $self->empty_table('DCC.ACCOUNT_PARAM_REG_EXPS', $policy_name_crc);
    $self->empty_table('DCC.ACCOUNT_PARAMETER_NAMES', $policy_name_crc);
    $self->empty_table('DCC.ACCOUNT_OBJECT_MANDATORY_PARAMETERS', $policy_name_crc);
    $self->empty_table('DCC.ACCOUNT_GLOBAL_MANDATORY_PARAMETERS', $policy_name_crc);
    $self->empty_table('DCC.ACCOUNT_OBJECT_QUERY_AS_VALUE_PARAMETERS', $policy_name_crc);
  }

  return RET_OK if $policy_id == 0;
  my %regexp_crcs = ();	# keep track of regexp crcs, to only add each one once
  my %positional_param_attrib_ids =  map { ($_ => 1) } @{$dbh->selectcol_arrayref(<<EndQuery,undef, $policy_id)};
SELECT DISTINCT param_attrib_id
  FROM PLC.PL_OBJECT_POSITIONAL_PARAMETERS
 WHERE policy_id = ?
EndQuery

  local $" = ',';

  for my $param_type (@parameter_types) {
    unless ($readonly_mode) {
      $self->empty_table($param_type->{dcc_table}, $policy_name_crc);
      if ($param_type->{dcc_wildcard_table}) {
        $self->empty_table($param_type->{dcc_wildcard_table}, $policy_name_crc);
      }
    }

    my @param_fields = qw(param_attrib_id param_t.name key_input_type max_len key_ui_type regexpr_values
			  min_value max_value flg_check_min flg_check_max flg_mandatory flg_null_value
			  flg_regexpr key_type flg_is_sensitive 
			  param_t.flg_is_pattern param_t.pattern_order
			  flg_check_param_name_metachar flg_tightening flg_is_staging last_updated
                          flg_check_max_len flg_check_param_value_metachar flg_check_signatures flg_allow_repeated_parameter_name
                          flg_in_classification flg_disallow_upload_of_exe flg_is_base64 host_name_representation flg_detect_base64
              parameter_location flg_check_min_items_in_array  flg_check_max_items_in_array  flg_check_array_unique_items
              min_items_in_array max_items_in_array array_serialization_format
              flg_exclusive_min flg_exclusive_max flg_check_multiple_of_value multiple_of_value min_len flg_check_min_len flg_is_header
              object_serialization_style flg_object_serialization_explode
              flg_is_cookie
			 );
    if (defined ($param_type->{extra_plc_fields}) ) {
        push(@param_fields, split(/,/, $param_type->{extra_plc_fields}));
    }

    my $param_query = "SELECT @param_fields FROM PLC.PL_PARAM_ATTRIBUTES AS param_attrib_t, PLC.$param_type->{plc_table} AS param_t WHERE param_t.policy_id = $policy_id AND param_t.param_attrib_id = param_attrib_t.id";
    $param_query.= " AND $param_type->{extra_where_condition}" if exists ($param_type->{extra_where_condition});
    $param_query.= " ORDER BY param_t.pattern_order, param_attrib_t.last_updated, param_t.name";
    my $param_sth = $dbh->prepare($param_query) or $self->{logger}->error("ERR: $DBI::errstr on retrieving parameters for policy $policy_id");
    $param_sth->execute or $self->{logger}->error("Select from table $param_type->{plc_table} failed ($DBI::errstr)");
    my %query_as_value_params = ();
     my $dcc_pattern_order = 1;
    while (my $param = $param_sth->fetchrow_hashref()) {
      next if $param->{name} =~ /\0/;  #Do not prepare parameters with null characters in them
      if (! $flg_encoding_selected) {
          $param->{flg_tightening} = 0;
          $param->{flg_is_staging} = 0;
      }
      if ($param->{flg_tightening} == F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION) {
          $param->{flg_tightening} = 0;
      }
      $param->{policy_name_crc} = $policy_name_crc;
      $param->{min_user_input_len} = ($param->{flg_check_min_len} ? $param->{min_len} : 0) || 0;
      $param->{max_user_input_len} = ($param->{flg_check_max_len} ? $param->{max_len} : 0) || 0;
      $param->{ui_type} = $param->{key_ui_type};
      $param->{flg_null_value_allow} = $param->{flg_null_value};
      $param->{flg_allow_repeated_parameter_name} = $param->{flg_allow_repeated_parameter_name};
      $param->{param_name} = $param->{name};
      if ($param->{flg_is_header}) {
        $self->{h_header_parameters}->{$param->{param_name}} = 1;
        $param->{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_HEADER();
      }
      if ($param->{flg_is_cookie}) {
        $self->{h_cookie_parameters}->{$param->{param_name}} = 1;
        $param->{parameter_location} = F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE();
      }
      $param->{flg_is_base64} =  $param->{flg_is_base64} || 0;
      $param->{flags_array_type_bitmask} = 0;
      $param->{min_values_items_in_array} = $param->{flg_check_min_items_in_array} ? $param->{min_items_in_array} : 0;
      $param->{max_values_items_in_array} = $param->{flg_check_max_items_in_array} ? $param->{max_items_in_array} : 0;
      if ($param->{key_input_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_ARRAY_VALUE) {
          my %check_array_flags_bitmask_values = ( flg_check_min_items_in_array => 1,
                                                   flg_check_max_items_in_array => 2,
                                                   flg_check_array_unique_items => 4,
                                                   flg_check_array_enum_values => 8,);
          for my $check_array_flag (keys %check_array_flags_bitmask_values) {
              if ($param->{$check_array_flag}) {
                  $param->{flags_array_type_bitmask} |= $check_array_flags_bitmask_values{$check_array_flag};
              }
          }

          my $array_param_enum_values_count = $self->check_param_enums(policy_id => $policy_id, param_attrib_id => $param->{param_attrib_id});
          if ($array_param_enum_values_count > 0) {
              $param->{flags_array_type_bitmask} |= $check_array_flags_bitmask_values{flg_check_array_enum_values};
          }

          if (($param->{array_serialization_format} == F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MULTI())
               ||
              ($param->{array_serialization_format} == F5::ASMConfig::ArraySerializationFormat::ARRAY_SERIALIZATION_FORMAT_MULTIPART())) {
              $param->{flg_allow_repeated_parameter_name} = 1;
          }
      }

      if (($param->{key_input_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_OBJECT_VALUE() ||
           $param->{key_input_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_OPENAPI_ARRAY_VALUE() ) &&
          ($param->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_MATRIX() ||
           $param->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_FORM() ||
           $param->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_SPACE_DELIMITED() ||
           $param->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_PIPE_DELIMITED()) &&
          $param->{flg_object_serialization_explode}) {
              $param->{flg_allow_repeated_parameter_name} = 1;
      }

      if ($param->{key_input_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT &&
             ($param->{key_ui_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_INTEGER ||
              $param->{key_ui_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_DECIMAL ||
              $param->{key_ui_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_BOOLEAN ||
              $param->{key_ui_type} == F5::ASMConfig::ParamDataType::PARAM_DATA_ALPHA_NUMERIC) ) {
          my $param_enum_values_count = $self->check_param_enums(policy_id => $policy_id, param_attrib_id => $param->{param_attrib_id});
          if ($param_enum_values_count > 0) {
              $param->{flg_check_enum_values} = 1;
          }
      }

      $param->{regexp_crc} = 0;

      if ($param->{key_ui_type} == KEY_UI_TYPE_INTEGER &&
         ($param->{flg_check_max} || $param->{flg_check_min} || $param->{flg_check_multiple_of_value})) {
	$param->{ui_type} |= KEY_UI_TYPE_INTEGER_CHECK_RANGE;
      }

      if ($param->{key_ui_type} == KEY_UI_TYPE_DECIMAL &&
         ($param->{flg_check_max} || $param->{flg_check_min} || $param->{flg_check_multiple_of_value})) {
	$param->{ui_type} |= KEY_UI_TYPE_DECIMAL_CHECK_RANGE;
      }

      $param->{profile_templ_id} = $self->{h_keys}->{$param->{param_attrib_id}}->{profile_templ_id} || 0;

      if ($param->{flg_regexpr}) {
	my $regexp_crc = F5::CRC::get_crc_any("get_crc32", $param->{regexpr_values}, 0);
 	$param->{regexp_crc} = $regexp_crc;

 	if (not exists $regexp_crcs{$regexp_crc}) {
          $self->insert_row(
                            table_name => 'DCC.ACCOUNT_PARAM_REG_EXPS',
                            ignore => 1,
                            rec => { policy_name_crc => $policy_name_crc,
                                     reg_exp_crc => $regexp_crc,
                                     reg_exp => $param->{regexpr_values} });
 	  $regexp_crcs{$regexp_crc} = $param->{regexpr_values};
	}
      }

      # Parameters
      my $dcc_table = '';

      $param->{negative_check_bitmap} = get_negative_check_bitmap(name_metachar_flag => $param->{flg_check_param_name_metachar},
                                                                  value_metachar_flag => $param->{flg_check_param_value_metachar},
                                                                  signature_flag => $param->{flg_check_signatures},
                                                                 );
      if ($param->{flg_check_param_value_metachar}) {
        $param->{param_value_charset_templ_id} = $self->get_charset_template(
                                                                             policy_name_crc => $policy_name_crc,
                                                                             base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_PARAM_VALUE(),
                                                                             override_table => 'PLC.PL_PARAM_VALUE_METACHARS',
                                                                             override_id_field => 'param_attrib_id',
                                                                             override_id_value => $param->{param_attrib_id});
      }
      else {
        $param->{param_value_charset_templ_id} = 0;
      }

      my $keycrc = $param_type->{compute_keycrc}->($self,$param);
      next unless $keycrc;
      $self->{h_keys}->{$param->{param_attrib_id}}->{keycrc} = $keycrc;
      $self->{h_keys}->{$param->{param_attrib_id}}->{param_name} = $param->{param_name};
      if ($readonly_mode) {
        next;
      }

      $param->{key_crc} = $keycrc;

      my $is_query_as_value_parameter = 0;
      if ( $param_type->{dcc_query_as_value_parameters_table} &&
           $param->{key_input_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_OBJECT_VALUE() &&
          ($param->{parameter_location} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_QUERY() ||
           $param->{parameter_location} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_ANY() ||
           $param->{parameter_location} == F5::ASMConfig::ParameterLocation::PARAMETER_LOCATION_COOKIE()) ) {
               if ($query_as_value_params{$self->{h_objects}->{$param->{obj_id}}->{objcrc}} ) {
                   next; # only 1 "query as value" parameter allowed per URL
               }

               if ($param->{object_serialization_style} == F5::ASMConfig::ObjectSerializationStyle::OBJECT_SERIALIZATION_STYLE_FORM() &&
                   $param->{flg_object_serialization_explode} ) {
                       $is_query_as_value_parameter = 1;
               }
       }

      if ($param->{flg_is_pattern}) {
	$dcc_table = $param_type->{dcc_wildcard_table};

	$param->{pattern} = $param->{name};
        $param->{pattern_order} = $dcc_pattern_order++;
        if ($param->{flg_check_param_name_metachar}) {
          $param->{param_name_charset_templ_id} = $self->get_charset_template(
                                                                              policy_name_crc => $policy_name_crc,
                                                                              base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_PARAM_NAME(),
                                                                              override_table => 'PLC.PL_PARAM_NAME_METACHARS',
                                                                              override_id_field => 'param_attrib_id',
                                                                              override_id_value => $param->{param_attrib_id});
        }
        else {
          $param->{param_name_charset_templ_id} = 0;
        }

	if (exists $param_type->{wildcard_crc_field}) {
	  $param->{$param_type->{wildcard_crc_field}} = $param_type->{get_wildcard_crc}->($param);
	}
      }
      else {
	# Explicit parameter
	$dcc_table = $param_type->{dcc_table};

      }

      if (($param->{key_input_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_OBJECT_VALUE) ||
          ($param->{key_input_type} == F5::ASMConfig::ParamValueType::PARAM_VALUE_OPENAPI_ARRAY_VALUE)) {
          # in case "value type" is "OBJECT" or "OPENAPI_ARRAY", "JSON" bit should be also switched ON
          $param->{key_input_type} += F5::ASMConfig::ParamValueType::PARAM_VALUE_JSON_VALUE;
      }

      $self->insert_row(
                        table_name => $dcc_table,
                        rec => $param);

      if ($is_query_as_value_parameter) {
          my %rec = (policy_name_crc => $policy_name_crc,
                     obj_crc => $self->{h_objects}->{$param->{obj_id}}->{objcrc},
                     object_param_key_crc => $param->{key_crc},
                     parameter_name => $param->{name},);
          $self->insert_row(table_name => $param_type->{dcc_query_as_value_parameters_table}, rec => \%rec,);

          my $object_table_name = 'DCC.ACCOUNT_APPLICATION_OBJECT_ATTRIBUTES';
          if ($self->{h_objects}->{$param->{obj_id}}->{flg_is_pattern}) {
              $object_table_name = 'DCC.ACCOUNT_APPLICATION_WILDCARD_OBJECT_ATTRIBUTES';
          }

          $dbh->do(<<EndQuery, undef, $policy_name_crc, $rec{obj_crc});
UPDATE $object_table_name
  SET object_settings_bitmap = object_settings_bitmap | $QUERY_AS_VALUE_BIT
  WHERE policy_name_crc = ?
    AND object_crc = ?
EndQuery

          $query_as_value_params{$self->{h_objects}->{$param->{obj_id}}->{objcrc}} = 1;
      }

      # Parameter values ranges
      if (($param->{flg_check_min} || $param->{flg_check_max} || $param->{flg_check_multiple_of_value}) &&
          ($param->{ui_type} & (KEY_UI_TYPE_INTEGER_CHECK_RANGE | KEY_UI_TYPE_DECIMAL_CHECK_RANGE))) {
        my @param_value_range_fields = ('policy_name_crc', 'key_crc', 'min_value', 'max_value', 'flag_check_min', 'flag_check_max',
                                        'flg_exclusive_min', 'flg_exclusive_max', 'flg_check_multiple_of_value', 'multiple_of_value');
        $param->{flag_check_min} = $param->{flg_check_min};
        $param->{flag_check_max} = $param->{flg_check_max};
        $param->{min_value} ||= 0;
        $param->{max_value} ||= 0;

        $self->insert_row(
                          table_name => 'DCC.ACCOUNT_PARAM_VALUE_RANGES',
                          fields => \@param_value_range_fields,
                          rec => { policy_name_crc => $policy_name_crc,
                                   %$param } );

      }

      if ($positional_param_attrib_ids{$param->{param_attrib_id}}) {
          $self->insert_row(table_name => 'DCC.ACCOUNT_PARAMETER_NAMES',
                            rec => {
                                    policy_name_crc => $policy_name_crc,
                                    param_name_crc => $param->{key_crc},
                                    param_name => $param->{name},
                                   },
                           );
      }

      if ($param->{flg_mandatory} && (! $param->{flg_is_pattern} || $param_type->{plc_table} eq 'PL_FLOW_PARAMS') && $param_type->{dcc_mandatory_parameters_table}) {
          my $dcc_mandatory_parameters_table = $param_type->{dcc_mandatory_parameters_table};
          my %rec = (policy_name_crc => $policy_name_crc, parameter_name => $param->{name},);
          if ($param_type->{dcc_mandatory_crc_fields}) {
	          for my $crc_field_name (keys %{$param_type->{dcc_mandatory_crc_fields}}) {
	              $rec{$crc_field_name} = $param_type->{dcc_mandatory_crc_fields}->{$crc_field_name}->($param);
      }
    }
          $self->insert_row(table_name => $dcc_mandatory_parameters_table, rec => \%rec,);
      }

      }
    }

  my ($scanner_type) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT scanner_type
  FROM PLC.PL_SCANNER_CONFIG
 WHERE policy_id = ?
EndQuery

  if ((defined $scanner_type) && ($scanner_type =~ /^sentinel/)) {
      $self->whitehat_settings($policy_id, $policy_name_crc);
  }

  if (%regexp_crcs) {
      # Parameter regexps needs to be copied to "pool" table, if this policy had any
      $dbh->do('DELETE FROM DCC.GLOBAL_PARAM_REG_EXPS');
      $dbh->do(<<EndQuery);
INSERT INTO DCC.GLOBAL_PARAM_REG_EXPS (reg_exp_crc, reg_exp)
     SELECT DISTINCT reg_exp_crc, reg_exp
       FROM DCC.ACCOUNT_PARAM_REG_EXPS
EndQuery
  }

  return RET_OK;
}

#--------------------------------------------------------------------------------

sub check_param_enums {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $param_attrib_id = $params{param_attrib_id};
    my $policy_id = $params{policy_id};

    my ($param_enum_values_count) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id, $param_attrib_id);
SELECT count(*)
  FROM PLC.PL_STATIC_PARAMS
  WHERE policy_id = ?
  AND param_attrib_id = ?
EndQuery

    return $param_enum_values_count;
}

#--------------------------------------------------------------------------------

sub whitehat_settings {
    my ($self, $policy_id, $policy_name_crc) = @_;
    my $dbh = $self->{dbh};

    $self->empty_table('DCC.WH_PARAMETERS', $policy_name_crc);

    my $attack_types = $dbh->selectall_hashref(<<"EndQuery", 'id', undef, $policy_id);
SELECT params.id id, sv.attack_type_id attack_type_id
  FROM PLC.PL_PARAM_ATTRIBUTES params,
       PLC.PL_SCANNER_VULNS sv 
 WHERE id = sv.param_attrib_id
   AND params.policy_id = ?
EndQuery

    my %attack_type_map = (
                           18 => 1,
                           19 => 2,
                           17 => 4,
                           14 => 8,
                           32 => 16,
                          );

    my %param_attacks;
    for my $param_id (keys %$attack_types) {
        $param_attacks{$param_id} += $attack_type_map{$attack_types->{$param_id}->{attack_type_id}};
    }

    for my $param_id (keys %$attack_types) {
        next unless $self->{h_keys}->{$param_id}->{keycrc};
        $dbh->do(<<EndQuery, undef, $policy_name_crc, $self->{h_keys}->{$param_id}->{keycrc}, $param_attacks{$param_id});
INSERT INTO DCC.WH_PARAMETERS
(policy_name_crc, param_crc, attack_type_bitmap)
VALUES
(?,?,?)
EndQuery
    }
}

#--------------------------------------------------------------------------------

sub prepare_extracts {
  # See comment at bottom regarding affected tables
  my ($self, $policy_name_crc, $policy_id) = @_;
  my $dbh = $self->{dbh};

  my $saved_extract_regexps = {};

  my @scopes = ( { plc_tables => [ 'PLC.PL_OBJECT_EXTRACTS AS extract_t' ],
		   plc_fields => [ 'obj_id' ],
		   dcc_fields => [ 'object_crc' ],
		   dcc_table_prefix => 'ACCOUNT_OBJECT',
		   compute_additional => \&add_object_crc,
		 },
		 { plc_tables => [ 'PLC.PL_GLOBAL_EXTRACTS AS extract_t' ],
		   plc_conditions => [ 'extract_t.flg_applies_to_all = 1' ],
		   dcc_table_prefix => 'ACCOUNT_GLOBAL',
		 },
		 { plc_tables => [ 'PLC.PL_GLOBAL_EXTRACTS AS extract_t' ],
		   plc_conditions => [ "extract_t.applies_to_regexp != ''"],
		   plc_fields => [ 'applies_to_regexp' ],
		   dcc_fields => [ 'object_regexp' ],
		   dcc_table_prefix => 'ACCOUNT_GLOBAL_BYREGEXP',
		   compute_additional => \&dcc_plc_field_rename,
		 },
		 { plc_tables => [ 'PLC.PL_GLOBAL_EXTRACTS AS extract_t', 
				   'PLC.PL_GLOBAL_EXTRACT_OBJECT_TYPES AS extract_filetype_t',
				   'PLC.PL_FILETYPES AS filetype_t' ],
		   plc_conditions => [ 'extract_t.id = extract_filetype_t.global_extract_id',
				       'extract_filetype_t.obj_type_id = filetype_t.id' ],
		   plc_fields => [ 'filetype_t.name AS filetype_name' ],
		   dcc_fields => [ 'filetype_crc' ],
		   dcc_table_prefix => 'ACCOUNT_GLOBAL_BYFILETYPE',
		   compute_additional => sub {
		     my($self, $extract) = @_;
		     $extract->{filetype_crc} = F5::CRC::get_crc_any("get_crc32", $extract->{filetype_name});
		   }
		 },
		 { plc_tables => [ 'PLC.PL_GLOBAL_EXTRACTS AS extract_t', 
				   'PLC.PL_GLOBAL_EXTRACT_OBJECTS AS extract_objects_t' ],
		   plc_conditions => [ 'extract_t.id = extract_objects_t.global_extract_id' ],
		   plc_fields => [ 'obj_id' ],
		   dcc_fields => [ 'object_crc' ],
		   dcc_table_prefix => 'ACCOUNT_GLOBAL_BYOBJECT',
		   compute_additional => \&add_object_crc
		 },
	       );

  my @methods = ( { dynamic_param_type_val => FORM_DYNAMIC_PARAM,
		    dcc_table_suffix => 'EXTRACT_FORM_PARAM_INDEX',
		    dcc_fields => [ 'form_idx', 'param_idx', 'param_name_crc' ]
		  },
		  { dynamic_param_type_val => FORM_NAME_DYNAMIC_PARAM,
		    dcc_table_suffix => 'EXTRACT_FORM_PARAM_NAME',
		    dcc_fields => [ 'param_name_crc' ]
		  },
		  { dynamic_param_type_val => LINK_DYNAMIC_PARAM,
		    dcc_table_suffix => 'EXTRACT_HREF',
		    dcc_fields => [ 'param_name' ]
		  },
		  { dynamic_param_type_val => XML_DYNAMIC_PARAM,
		    dcc_table_suffix => 'EXTRACT_XPATH',
		    dcc_fields => [ 'param_name', 'xpath' ]
		  },
		  { dynamic_param_type_val => REGEXP_DYNAMIC_PARAM,
		    dcc_table_suffix => 'EXTRACT_REGEXP',
		    dcc_fields => [ qw(param_name reg_exp regexp_num_occurrences) ],
		    compute_additional => sub {
		      my($self, $extract, $scope, $method) = @_;

		      plc_regexp_merge($self, $extract, $scope, $method);

		      if (check_duplicate_dcc_record($extract, $scope, $method, $saved_extract_regexps)) {
			return COMPUTE_ADDITIONAL_DO_REPLACE;
		      }

		      return;
		    }
		  },
		  { dynamic_param_type_val => NAME_REGEXP_DYNAMIC_PARAM,
		    dcc_table_suffix => 'EXTRACT_DYNAMIC_PARAM_NAME',
		    dcc_fields => [ qw(reg_exp flg_check_dynamic_value) ],
		    compute_additional => \&plc_regexp_merge
		  },
		);

  # LINK_DYNAMIC_PARAM are handled **for objects only** (for now) as regexps.
  my $method_for_object_extract_href = { dynamic_param_type_val => LINK_DYNAMIC_PARAM,
					 dcc_table_suffix => 'EXTRACT_REGEXP',
					 dcc_fields => [ qw(param_name reg_exp regexp_num_occurrences) ],
					 compute_additional => sub {
					   my($self, $extract, $scope, $method) = @_;

					   $extract->{reg_exp} = q{(?<=[?&;]} . $extract->{param_name} . q{=)[^&;'">]+?(?=[&;'">])};

					   if (check_duplicate_dcc_record($extract, $scope, $method, $saved_extract_regexps)) {
					     return COMPUTE_ADDITIONAL_DO_REPLACE;
					   }

					   return;
					 }
				       };

  return RET_OK if $policy_id == 0;

  for my $scope (@scopes) {
    my @scope_methods = @methods;
    if ($scope->{dcc_table_prefix} eq 'ACCOUNT_OBJECT') {
      @scope_methods = grep($_->{dynamic_param_type_val} != LINK_DYNAMIC_PARAM, @methods);
      push(@scope_methods, $method_for_object_extract_href);
    }

    for my $method (@scope_methods) {
      my $dcc_table = "DCC." . $scope->{dcc_table_prefix} . "_" . $method->{dcc_table_suffix};
      $self->empty_table($dcc_table, $policy_name_crc);
    }

    my @extract_fields = qw(extract_t.name form_idx param_idx xpath dynamic_param_type regexp_prefix
			    regexp_body regexp_suffix flg_check_dynamic_value regexp_num_occurrences param_t.key_input_type);

    push(@extract_fields, @{$scope->{plc_fields}}) if ($scope->{plc_fields});

    my $base_table = 'PLC.PL_EXTRACT_ATTRIBUTES AS extract_attrib_t ' . 
                     'LEFT JOIN PLC.PL_DYNAMIC_PARAM_EXTRACTS AS dynamic_extract_t ' .
                     'ON extract_attrib_t.id = dynamic_extract_t.extract_attrib_id ' .
                     'LEFT JOIN PLC.PL_PARAM_ATTRIBUTES AS param_t ' .
                     'ON param_t.id = dynamic_extract_t.param_attrib_id ';

    my @plc_tables = (@{$scope->{plc_tables}}, $base_table);
    my @plc_conditions = ("extract_t.policy_id = $policy_id",
			  "extract_t.extract_attrib_id = extract_attrib_t.id");
    push(@plc_conditions, @{$scope->{plc_conditions}}) if ($scope->{plc_conditions});
    local $" = ',';
    my $get_plc_extract_sql = "SELECT @extract_fields FROM @plc_tables WHERE " . join(" AND ", @plc_conditions);

    my $extract_sth = $dbh->prepare($get_plc_extract_sql)
      or $self->{logger}->error("ERR: $DBI::errstr on retrieving extracts for policy $policy_id");
    $extract_sth->execute or $self->{logger}->error("Select from table $scope->{plc_table} failed ($DBI::errstr)");

    while (my $extract = $extract_sth->fetchrow_hashref()) {
      my $insert_style = 0;
      $extract->{policy_name_crc} = $policy_name_crc;
      $extract->{param_name} = $extract->{name};
      if (defined $extract->{key_input_type} && $extract->{key_input_type} == DYNAMIC_PARAMETER_NAME) {
	$extract->{param_name_crc} = 0;
      }
      else {
	$extract->{param_name_crc} = F5::CRC::get_crc_any("get_crc16", $extract->{param_name}, $self->{unescape});
      }

      if ($scope->{compute_additional}) {
	my $rv = $scope->{compute_additional}->($self, $extract, $scope);
	$insert_style = $rv if $rv;
      }

      for my $method (@scope_methods) {
	if ($extract->{dynamic_param_type} & $method->{dynamic_param_type_val}) {
	  if ($method->{compute_additional}) {
	    my $rv = $method->{compute_additional}->($self, $extract, $scope, $method);
	    $insert_style = $rv if $rv;
	  }

	  my $dcc_table = "DCC." . $scope->{dcc_table_prefix} . "_" . $method->{dcc_table_suffix};
	  my @dcc_fields = ( 'policy_name_crc', @{$method->{dcc_fields}});
	  push(@dcc_fields, @{$scope->{dcc_fields}}) if ($scope->{dcc_fields});

          $self->insert_row(
                            do_replace => ($insert_style == COMPUTE_ADDITIONAL_DO_REPLACE),
                            table_name => $dcc_table,
                            fields => \@dcc_fields,
                            rec => $extract);
	}
      }
    }
  }

  return RET_OK;
}

#----------------------------------------------------------------------

sub check_duplicate_dcc_record {
  my($extract, $scope, $method, $saved_hash) = @_;

  my $extract_key = '';
  if ($scope->{plc_fields}) {
    $extract_key .= join(',', @{$extract}{@{$scope->{plc_fields}}});
  }

  my $dcc_table = "DCC." . $scope->{dcc_table_prefix} . "_" . $method->{dcc_table_suffix};
  if ($saved_hash->{$dcc_table}->{$extract_key}->{$extract->{param_name}}) {
    return 1;
  }
  $saved_hash->{$dcc_table}->{$extract_key}->{$extract->{param_name}} = 1;

  return;
}

#--------------------------------------------------------------------------------
sub prepare_static_params_tbl {
  my ($self, $policy_name_crc, $plc_id) = @_;
  my $dbh = $self->{dbh};

  $self->empty_table("DCC.ACCOUNT_APPLICATION_STATIC_PARAMS", $policy_name_crc);

  return RET_OK if $plc_id == 0;

  my $plc_sql = <<EndQuery;
SELECT DISTINCT param_attrib_id, value, key_input_type, key_ui_type
   FROM PLC.PL_STATIC_PARAMS, PLC.PL_PARAM_ATTRIBUTES
  WHERE PLC.PL_STATIC_PARAMS.policy_id = ?
    AND PLC.PL_STATIC_PARAMS.param_attrib_id= PLC.PL_PARAM_ATTRIBUTES.id
    AND PLC.PL_PARAM_ATTRIBUTES.key_input_type in (?,?,?)
EndQuery

  my $plc_sth = $dbh->prepare($plc_sql)
    or $self->{logger}->error("Prepare failed on SQL: '$plc_sql': $DBI::errstr");	

  $plc_sth->execute($plc_id,
                       F5::ASMConfig::ParamValueType::PARAM_VALUE_STATIC_CONTENT_VALUE(),
                       F5::ASMConfig::ParamValueType::PARAM_VALUE_ARRAY_VALUE(),
                       F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT(),)
    or $self->{logger}->error("Execute failed on SQL: '$plc_sql': $DBI::errstr");

    my %enum_data_types = map { $_ => 1 } ( F5::ASMConfig::ParamDataType::PARAM_DATA_INTEGER(),
                                            F5::ASMConfig::ParamDataType::PARAM_DATA_DECIMAL(),
                                            F5::ASMConfig::ParamDataType::PARAM_DATA_ALPHA_NUMERIC(),
                                            F5::ASMConfig::ParamDataType::PARAM_DATA_BOOLEAN(),);
  while(my @plc_vals = $plc_sth->fetchrow_array()) {
      my ($param_attrib_id, $value, $key_input_type, $key_ui_type) = @plc_vals;
      if ($key_input_type == F5::ASMConfig::ParamValueType::PARAM_VALUE_USER_INPUT() && ! $enum_data_types{$key_ui_type}) {
          next;
      }
      my $keycrc = $self->{h_keys}->{$plc_vals[0]}->{keycrc};
      next unless $keycrc;
      if (not defined $keycrc or $keycrc eq '') {
          $self->{logger}->info("Parameter attributes are not associated with any parameter (id =  $plc_vals[0])!");
          next;
      }
		
      my $key_value_crc = F5::CRC::get_crc_any("get_crc_str", $keycrc, $plc_vals[1], $self->{unescape});		

      $self->insert_row(
                        table_name => 'DCC.ACCOUNT_APPLICATION_STATIC_PARAMS',
                        rec => { policy_name_crc => $policy_name_crc,
                        key_value_crc => $key_value_crc });
  }

  return RET_OK;
}



#--------------------------------------------------------------------------------
sub http_length_row_handler {
    my ($self, %params) = @_;

    my $row = $params{row};
    $row->{max_cookie_len} = $row->{cookie_len};
    $row->{max_header_len} = $row->{http_len};

}

#--------------------------------------------------------------------------------
sub prepare_file_type_attrs_tbl {
  my ($self, $policy_name_crc, $policy_id, $params) = @_;

  my $flg_encoding_selected = defined $params->{flg_encoding_selected} ? $params->{flg_encoding_selected} : $self->get_flg_encoding_selected($policy_name_crc);

  my $dbh = $self->{dbh};
  local $" = ',';
  $self->empty_table('DCC.ACCOUNT_APPLICATION_FILE_TYPE_ATTRIBUTES', $policy_name_crc);
  $self->empty_table('DCC.ACCOUNT_APPLICATION_WILDCARD_FILE_TYPE_ATTRIBUTES', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @plc_fields = qw(id name uri_len req_len qs_len pd_len flg_flow_check 
		      flg_is_referrer flg_filter_app_error
		      flg_is_pattern pattern_order flg_tightening flg_is_staging
                      flg_check_uri_len flg_check_req_len flg_check_qs_len flg_check_pd_len
                    );

  my $plc_sth = $dbh->prepare("SELECT @plc_fields from PLC.PL_FILETYPES WHERE policy_id = ? and flg_is_allowed = ? ORDER BY pattern_order,last_updated,name");

  $plc_sth->execute($policy_id,TRUE);
  my $dcc_pattern_order = 1;
  while(my $plc_row = $plc_sth->fetchrow_hashref) {
    $plc_row->{policy_name_crc} = $policy_name_crc;
    $plc_row->{validity} = 1;
    $plc_row->{max_request_len} = $plc_row->{req_len};
    $plc_row->{max_object_name_len} = $plc_row->{uri_len};
    $plc_row->{max_user_input_len} = $plc_row->{qs_len};
    $plc_row->{max_post_data_len} = $plc_row->{pd_len};
    if (! $flg_encoding_selected) {
        $plc_row->{flg_tightening} = 0;
        $plc_row->{flg_is_staging} = 0;
    }

    if ($plc_row->{flg_tightening} == F5::ASMConfig::TighteningMode::TIGHTENING_MODE_EXCEPTION) {
        $plc_row->{flg_tightening} = 0;
    }


    my $dcc_table_name;
    my @dcc_fields = qw(policy_name_crc validity
			max_request_len max_object_name_len max_user_input_len
			max_post_data_len flg_filter_app_error flg_is_staging
                        flg_check_uri_len flg_check_req_len flg_check_qs_len flg_check_pd_len
                      );
    if ($plc_row->{flg_is_pattern}) {
      $dcc_table_name = 'DCC.ACCOUNT_APPLICATION_WILDCARD_FILE_TYPE_ATTRIBUTES';
      $plc_row->{pattern} = $plc_row->{name};
      $plc_row->{pattern_order} = $dcc_pattern_order++;
      push(@dcc_fields, qw(pattern flg_tightening pattern_order));
    }
    else { 
      # Explicit filetype
      $dcc_table_name = 'DCC.ACCOUNT_APPLICATION_FILE_TYPE_ATTRIBUTES';
      $plc_row->{filetype_crc} = F5::CRC::get_crc_any("get_crc32", $plc_row->{name});
      push(@dcc_fields, qw(filetype_crc));
    }

    $self->insert_row(
                      table_name => $dcc_table_name,
                      fields => \@dcc_fields,
                      rec => $plc_row);

  }

  return RET_OK;
}

#--------------------------------------------------------------------------------
sub prepare_account_attributes {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  return RET_OK if $policy_id == 0;

  my %table_db_fields = ( 'PLC.PL_POLICIES' => { fields => [
                                                  qw( name
                                                      encoding
                                                      flg_encoding_selected
                                                      flg_is_protocol_free
                                                      flg_is_owa
                                                      flg_passive_mode
                                                   )],
                                                 policy_id_field => 'id',},

                          'PLC.PL_POLICY_GENERAL' => { fields => [
                                                       qw( flg_use_dynamic_session_in_url
                                                           flg_irules
                                                           dynamic_session_pattern
                                                           dynamic_session_description
                                                           path_parameter_handling
                                                           flg_mask_sensitive
                                                           flg_enable_correlation )],
                                                     policy_id_field => 'policy_id',},
                          'PLC.PL_POLICY_ANTIVIRUS' => {fields => [ qw( flg_inspect_http_uploads ) ],
                                                        policy_id_field => 'policy_id',},
                         );
                           

  my $row = {};
  for my $table_name (keys %table_db_fields) {
      my @db_fields = @{$table_db_fields{$table_name}->{fields}};
      my $table_row = $dbh->selectrow_hashref(<<EndQuery, undef, $policy_id);
SELECT @db_fields
  FROM $table_name
 WHERE $table_db_fields{$table_name}->{policy_id_field} = ?
EndQuery
      $row = {%$row, %$table_row};
  }

  my $policy_attributes = { max_headers => $MAX_HEADERS_INDEX, max_params => $MAX_PARAMS_INDEX, max_cookies => $MAX_COOKIES_INDEX, };

  for my $attribute ( keys %{$policy_attributes}) {
  ($row->{$attribute}) =  $dbh->selectrow_array(<<EndQuery, undef, $policy_id, $policy_attributes->{$attribute} );
SELECT $attribute
  FROM PLC.PL_HTTP_CHECKS
 WHERE policy_id = ?
   AND http_check_code = ?
EndQuery

  $self->{global_subviolation_attribute}->{$attribute} = $row->{$attribute};
  }

  $row->{policy_name} = $row->{name};

  delete $row->{name};

  ($row->{flg_is_whitehat}) = $dbh->selectrow_array(<<EndQuery, undef, 'sentinel%', $policy_id);
SELECT scanner_type LIKE ?
  FROM PLC.PL_SCANNER_CONFIG
 WHERE policy_id = ?
EndQuery


  if ($row->{flg_is_whitehat}) {
      my ($negsig_count) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id, HTTP_RESPONSE_SPLITTING_ATTACK_TYPE_ID);
SELECT COUNT(pns.sig_id)
  FROM PLC.PL_POLICY_NEGSIG_SIGNATURES AS pns, PLC.NEGSIG_SIGNATURES AS ns
 WHERE pns.policy_id = ?
   AND ns.attack_type_id = ?
   AND pns.flg_disabled = 0
   AND ns.sig_id = pns.sig_id
EndQuery

      $row->{whitehat_attacks_bitmap} = $negsig_count > 0 ? RESPONSE_SPLITTING_BITMAP_VALUE : 0;
  }
  else {      
      $row->{whitehat_attacks_bitmap} = $row->{flg_is_whitehat} = 0;
  }

  ($row->{lang_set_id}) = $dbh->selectrow_array(<<EndQuery, undef, $row->{encoding});
SELECT lang_set_id
  FROM DCC.LANGUAGE_ENCODINGS le JOIN DCC.ENCODINGS e ON le.encoding_id = e.encoding_id
 WHERE le.flg_visible = 1 AND e.encoding = ?
EndQuery
  delete $row->{encoding};

  $row->{policy_template_system_id} = $row->{flg_is_owa} ? 1 : 0;
  $row->{policy_name} = $self->{policy_name} if defined $self->{policy_name};
  delete $row->{flg_is_owa};

  my ($flg_is_redirection_enabled) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT flg_enabled
  FROM PLC.PL_REDIRECTION_PROTECTION
 WHERE policy_id = ?
EndQuery

  my ($wildcard_redirection_tightening) = $dbh->selectrow_array(<<EndQuery,undef,$policy_id);
SELECT flg_tightening
  FROM PLC.PL_REDIRECTION_DOMAINS
 WHERE domain_name = '*'
   AND flg_is_pattern = 1
   AND policy_id = ?
EndQuery

  $row->{redirect_action} = REDIRECT_ACTION_NONE;
  $row->{flg_virus_check_file_upload} = $row->{flg_inspect_http_uploads} ;
  delete $row->{flg_inspect_http_uploads} ;
  $row->{flg_use_violation_asm_irule} = $row->{flg_irules};
  delete $row->{flg_irules};

  if ($flg_is_redirection_enabled){
      $row->{redirect_action} =  defined $wildcard_redirection_tightening ?
                        ( $wildcard_redirection_tightening ? REDIRECT_ACTION_TIGHTEN : REDIRECT_ACTION_NONE ) : REDIRECT_ACTION_ENFORCE ;
  }

  $self->{redirect_action}->{$policy_id} =  $row->{redirect_action};
  my @update_fields = keys %$row;
  my @placeholders = map { "$_ = ?" } @update_fields;
  my $unix_timestamp = F5::DbUtils::get_unix_timestamp_string(dbh => $dbh);

  $dbh->do(<<EndQuery, undef, @{$row}{@update_fields}, $policy_name_crc);
UPDATE DCC.ACCOUNTS
   SET @placeholders,
       policy_apply_date = $unix_timestamp
 WHERE policy_name_crc = ?
EndQuery

  return RET_OK;
}

#--------------------------------------------------------------------------------
sub prepare_http_checks {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  return RET_OK if $policy_id == 0;

  my $http_check_viols = $dbh->selectall_hashref(<<EndQuery,'http_check_code', undef, $policy_id);
SELECT http_check_code, flg_is_enabled, flg_pb_enabled
  FROM PLC.PL_HTTP_CHECKS
WHERE policy_id = ?
EndQuery

  my $blocking_flags = $dbh->selectrow_hashref(<<EndQuery,undef,$policy_id, F5::Violation::VIOL_HTTP_PROTOCOL());
SELECT flg_alarm,flg_reject,flg_learn
  FROM PLC.PL_BLOCKING
 WHERE policy_id = ?
   AND viol_index = ?
EndQuery

  $self->{http_check_viols} = $http_check_viols;
  my $dcc_row = $self->calculate_http_check_viols( http_check_viols       => $http_check_viols,
                                                   blocking_flags         => $blocking_flags,
                                                   policy_name_crc        => $policy_name_crc,
                                                   policy_id              => $policy_id,
                                                   is_learning_enabled    => $self->is_learning_enabled(policy_id => $policy_id),
                                                  );

  # update mask in DCC.ACCOUNTS table
  $self->update_row(table_name => 'DCC.ACCOUNTS',
                    rec => $dcc_row,
                    where_fields => { policy_name_crc => $policy_name_crc },
                   );

  return RET_OK;
}

#--------------------------------------------------------------------------------

sub calculate_http_check_viols {
    my($self, %params) = @_;
    my $http_check_viols = $params{http_check_viols};
    my $policy_name_crc = $params{policy_name_crc};
    my $policy_id = $params{policy_id};
    my $is_learning_enabled = $params{is_learning_enabled};
    my $dbh = $self->{dbh};
    my $blocking_flags = $params{blocking_flags};

    if ( !$is_learning_enabled ) {
        $blocking_flags->{flg_learn} = 0;
    }

    my %dcc_row = (
                   http_protocol_checks_reject_mask => 0,
                   http_protocol_checks_alarm_mask => 0,
                   http_protocol_checks_learn_mask  => 0,
                  );

  # create http-checks mask
    for my $http_check_row ( values %$http_check_viols) {
        if ($http_check_row->{flg_is_enabled}) {
            for my $blocktype (qw(reject alarm)) {
                if ($blocking_flags->{"flg_$blocktype"}) {
                    $dcc_row{"http_protocol_checks_${blocktype}_mask"} |= $http_check_row->{http_check_code};
                }
            }
        }
        if ($http_check_row->{flg_pb_enabled}) {
            if ($blocking_flags->{flg_learn}) {
                $dcc_row{http_protocol_checks_learn_mask} |= $http_check_row->{http_check_code};
            }
        }
    }

  $dcc_row{http_protocol_checks_mask} = $dcc_row{http_protocol_checks_alarm_mask}; ## will be removed after BD change

  return \%dcc_row;
}

#--------------------------------------------------------------------------------

sub prepare_xff {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  $self->empty_table('DCC.ACCOUNT_XFF', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @plc_fields = qw(flg_is_trust_xff);
  my @dcc_fields = qw(policy_name_crc flg_is_trust_xff);

  my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields 
  FROM PLC.PL_POLICY_XFF
 WHERE policy_id = ?
EndQuery

  $plc_sth->execute($policy_id);

  while(my $plc_row = $plc_sth->fetchrow_hashref) {

    $plc_row->{policy_name_crc} = $policy_name_crc;

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_XFF',
                      fields => \@dcc_fields,
                      rec => $plc_row,);
  }

  return RET_OK;
}

#--------------------------------------------------------------------------------

sub threat_campaigns_row_handler {
    my ($self, %params) = @_;

    my $row = $params{row};

    $row->{threat_campaign_name} = $row->{name};
    $row->{threat_campaign_display_name} = $row->{display_name};
}

#--------------------------------------------------------------------------------

sub csrf_row_handler {
  my ($self, %params) = @_;

  my $row = $params{row};

  $row->{flg_is_enabled} = $self->{viols}->{F5::Violation::VIOL_CSRF()}->{is_enabled} || 
    $self->{viols}->{F5::Violation::VIOL_CSRF_EXPIRED()}->{is_enabled} || 0;

}

#--------------------------------------------------------------------------------

sub csrf_urls_parameters_row_handler {
  my ($self, %params) = @_;

  my $row = $params{row};

  $row->{csrf_url_crc} = $self->{csrf_urls}->{$params{policy_name_crc}}->{$row->{csrf_url_id}}->{csrf_url_crc};

}
#--------------------------------------------------------------------------------

sub prepare_case_insensitive {
    my ($self, $policy_name_crc, $policy_id) = @_;

    my $dbh = $self->{dbh};

    $self->empty_table('DCC.ACCOUNT_CASE_INSENSITIVE', $policy_name_crc);
    return RET_OK if $policy_id == 0;

    my ($flg_is_case_insensitive) = @{$dbh->selectcol_arrayref(<<EndQuery, undef, $policy_id)};
SELECT flg_is_case_insensitive
  FROM PLC.PL_POLICIES
 WHERE id = ?
EndQuery

    if ( ! defined $flg_is_case_insensitive ) {
        return RET_NOT_OK;
    }

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_CASE_INSENSITIVE',
                      fields => [qw/policy_name_crc flg_is_case_insensitive/],
                      rec => {flg_is_case_insensitive => $flg_is_case_insensitive, policy_name_crc => $policy_name_crc});
}

#--------------------------------------------------------------------------------

sub prepare_response_scrubbing_urls {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';
  $self->empty_table('DCC.ACCOUNT_RESPONSE_SCRUBBING_URLS', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @plc_fields = qw( policy_id target_pattern);
  my $plc_sth = $dbh->prepare("SELECT @plc_fields from PLC.PL_POLICY_RESPONSE_SCRUBBING_URLS WHERE policy_id = ?");
  $plc_sth->execute($policy_id);

  while(my $plc_row = $plc_sth->fetchrow_hashref) {
    $plc_row->{policy_name_crc} = $policy_name_crc;

    my @dcc_fields = qw(policy_name_crc target_pattern);
    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_RESPONSE_SCRUBBING_URLS',
                      fields => \@dcc_fields,
                      rec => $plc_row);
  }
  return RET_OK;
}


#--------------------------------------------------------------------------------

sub csrf_urls_row_handler {
    my ($self, %params) = @_;

    my $row = $params{row};
    $row->{pattern} = $row->{target_pattern};
    $row->{method} = $row->{csrf_url_method};
    $row->{csrf_url_crc} =  F5::CRC::get_crc_any("get_crc_obj", $row->{method} ,$params{policy_name_crc}, $row->{target_pattern}); 
    $self->{csrf_urls}->{$params{policy_name_crc}}->{$row->{csrf_url_id}}->{csrf_url_crc} = $row->{csrf_url_crc};
    my $target_pattern = $row->{target_pattern};
    my $regexp_flag = $self->is_policy_case_insensitive(policy_id => $row->{policy_id}) ? 'i' : '';
    $target_pattern =~ s|^/||;
    $row->{url_regexp} =  "/^\\/(\\.\\/)*" . F5::Text::Glob::glob_to_regex_string($target_pattern) . "\$/$regexp_flag";
    if($row->{flg_is_pattern}){
        $row->{pattern_order} = $self->{csrf_urls_dcc_pattern_order}++;
    }
}
#--------------------------------------------------------------------------------

sub prepare_prerequisite_attributes {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  $self->empty_table('DCC.ACCOUNT_PREREQUISITE_ATTRIBUTES', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @plc_fields = qw(expiration_period);
  my @dcc_fields = qw(policy_name_crc expiration_period);

  my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields 
  FROM PLC.PL_POLICY_PREREQUISITE_ATTRIBUTES
 WHERE policy_id = ?
EndQuery

  $plc_sth->execute($policy_id);

  while(my $plc_row = $plc_sth->fetchrow_hashref) {

    $plc_row->{policy_name_crc} = $policy_name_crc;

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_PREREQUISITE_ATTRIBUTES',
                      fields => \@dcc_fields,
                      rec => $plc_row,);
  }

  return RET_OK;
}


#--------------------------------------------------------------------------------

sub prepare_login_brute_force_object_attr {
    my ($self, $policy_name_crc, $policy_id) = @_;
    my $dbh = $self->{dbh};
    local $" = ',';

    $self->empty_table('DCC.ACCOUNT_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES', $policy_name_crc);

    return RET_OK if $policy_id == 0;

    my $default_brute_force = $dbh->selectrow_hashref(<<EndQuery, undef, $policy_id, 0);
SELECT *
  FROM PLC.PL_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES brute_force
 WHERE policy_id = ?
   AND brute_force.object_id = ?
EndQuery

    my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT brute_force.*, login_obj.*, brute_force.object_id AS brute_force_object_id
  FROM PLC.PL_LOGIN_OBJECTS login_obj
LEFT JOIN PLC.PL_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES brute_force
       ON (login_obj.object_id = brute_force.object_id)
 WHERE login_obj.policy_id = ?
   AND auth_type != ?
EndQuery

    $plc_sth->execute($policy_id, F5::ASMConfig::LoginAuthType::LOGIN_AUTH_TYPE_NONE());

    while (my $plc_row = $plc_sth->fetchrow_hashref) {
        $plc_row->{policy_name_crc} = $policy_name_crc;
        $plc_row->{obj_crc} = $self->{h_objects}->{$plc_row->{object_id}}->{objcrc};
        $plc_row->{obj} = $self->{h_objects}->{$plc_row->{object_id}}->{obj};
        $plc_row->{protocol} = $self->{h_objects}->{$plc_row->{object_id}}->{app_id};
        $plc_row->{login_object_rest_uuid} = $plc_row->{rest_uuid};

        if (! $plc_row->{brute_force_object_id}) {
            # Take from default brute force, if enabled
            if (! $default_brute_force->{flg_enabled}) {
                next;
            }

            for my $k (keys %$default_brute_force) {
                $plc_row->{$k} = $default_brute_force->{$k};
            }
        }

        $self->insert_row(
                          table_name => 'DCC.ACCOUNT_LOGIN_BRUTE_FORCE_OBJECT_ATTRIBUTES',
                          rec => $plc_row);
    }

    return RET_OK;
}

#--------------------------------------------------------------------------------
sub prepare_custom_headers {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  $self->empty_table('DCC.ACCOUNT_CUSTOM_HEADERS_POOL', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @plc_fields = qw(custom_header_id header_name);

  my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields
FROM PLC.PL_POLICY_CUSTOM_HEADERS 
WHERE policy_id = ?
EndQuery

  $plc_sth->execute($policy_id);

  my $header_code = 0;
  while(my $plc_row = $plc_sth->fetchrow_hashref) {
      $self->insert_row(
          table_name => 'DCC.ACCOUNT_CUSTOM_HEADERS_POOL',
          rec => { policy_name_crc => $policy_name_crc,
                   header_code => $header_code,
                   header_name => $plc_row->{header_name} });
      $header_code++;
  }

  return RET_OK;
}


#--------------------------------------------------------------------------------
sub prepare_headers {
  my ($self, $policy_name_crc, $policy_id, $params) = @_;

  my $dbh = $self->{dbh};
  
  my $readonly_mode = $params->{readonly_mode};

  my %headers;
  my $header_code = 1;
  my $header_mask = 0;
  my $shifter = 1;

  my $mandatory_header_mask = Math::BigInt->new($header_mask);
  my $excluded_header_mask = Math::BigInt->new($header_mask);
  $header_code = Math::BigInt->new($header_code);
  $shifter = Math::BigInt->new($shifter);
  my $processed_header_name = {};

  unless ($readonly_mode) {
      $self->empty_table('DCC.ACCOUNT_HEADERS', $policy_name_crc);
      $self->empty_table('DCC.ACCOUNT_WILDCARD_HEADERS', $policy_name_crc);
  }

  return RET_OK if $policy_id == 0;

  my $plc_headers = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id);
  SELECT header.*,
         EXISTS( SELECT 1
                   FROM PLC.PL_HEADER_NEGSIG_SIGNATURES hns
                  WHERE hns.header_id = header.header_id) as excluded_signatures_exist
    FROM PLC.PL_HEADERS header
   WHERE policy_id = ?
ORDER BY pattern_order, last_updated, header_name
EndQuery

  my %dcc_headers;

  for my $plc_header (@$plc_headers) {
      my $header_name = lc($plc_header->{header_name});
      $plc_header->{header_name} = $header_name;
      if (! $dcc_headers{$header_name}) {
          $dcc_headers{$header_name} = $plc_header;
      }
  }

  for my $header_param_name (keys %{$self->{h_header_parameters}}) {
      my $header_name = lc $header_param_name;
      my $dcc_header = $dcc_headers{$header_name};
      if (! $dcc_header) {
          if ($dcc_headers{'*'}) {
              $dcc_header = dclone($dcc_headers{'*'});
              delete $dcc_header->{header_id};
          }
          else {
              $dcc_header = {};
          }
          $dcc_header->{header_name} = $header_name;
          $dcc_headers{$header_param_name} = $dcc_header;
      }
      $dcc_header->{flg_enforced_as_param} = 1;
      $dcc_header->{flg_is_pattern} = 0;
  }

  my $dcc_pattern_order = 1;
  for my $dcc_header (sort { ($a->{pattern_order} || 0) <=> ($b->{pattern_order} || 0) || ($a->{header_name} cmp $b->{header_name}) } values %dcc_headers) {
      $mandatory_header_mask |= $header_code if ($dcc_header->{flg_is_mandatory});
      $excluded_header_mask |= $header_code if (!$dcc_header->{flg_check_signatures});

      $dcc_header->{normalization_mask} = calc_mask(
                                                    1 => $dcc_header->{flg_percent_normalization},
                                                    2 => $dcc_header->{flg_uri_normalization},
                                                    3 => $dcc_header->{flg_is_base64},
                                                    4 => $dcc_header->{flg_html_normalization},
                                                    5 => $dcc_header->{flg_detect_base64}
                                         );

      $dcc_header->{header_name_crc} = F5::CRC::get_crc_any("get_crc32", $dcc_header->{header_name});
      if ($dcc_header->{header_id}) {
          $self->{h_headers}->{$dcc_header->{header_id}}->{header_name_crc} = $dcc_header->{header_name_crc};
      }

      $dcc_header->{policy_name_crc} = $policy_name_crc;
      $dcc_header->{header_code} = $header_code;

      $header_code <<= $shifter;
      next if ($readonly_mode);

      if ($dcc_header->{flg_is_pattern}) {
          $dcc_header->{pattern_order} = $dcc_pattern_order;
           $self->insert_row(
                               table_name => 'DCC.ACCOUNT_WILDCARD_HEADERS',
                            rec => $dcc_header,
                             );
          $dcc_pattern_order++;
      } else {
           $self->insert_row(
                               table_name => 'DCC.ACCOUNT_HEADERS',
                            rec => $dcc_header,
                             );
      }
  }
  
  if (!$readonly_mode) {
      # add header mask to DCC.ACCOUNTS table
      $dbh->do("UPDATE DCC.ACCOUNTS SET negsig_excluded_headers_mask= ?, mandatory_headers_mask= ?  WHERE policy_name_crc = ?",
                undef, $excluded_header_mask, $mandatory_header_mask, $policy_name_crc );
  }

  return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_ip_address_allowlist {
   my ($self, $policy_name_crc, $policy_id) = @_;
   my $logger = $self->{logger};
   my $dbh = $self->{dbh};

   my %ignore_flags_bitmask_values = ( flg_ignore_anomaly  => 1,
                                      #flg_never_block     => 2,
                                      flg_never_log        => 4,
                                      flg_ignore_ip_reputation => 8,
                                      #flg_always_block => 16,
                                      # flg_ssrf_ip_disallow => 32, flag to identify SSRF disallowed IPs
                                      # flg_ssrf_ip_allow => 64, flag to identify SSRF allowed IPs
                                    );

   return RET_OK if $policy_id == 0;

   $self->empty_table('DCC.ACCOUNT_IP_WHITELIST', $policy_name_crc);
   $self->empty_table('DCC.ACCOUNT_SSRF_DISALLOWED_HOSTS', $policy_name_crc);

   my $ssrf_denylist_configuration = $dbh->prepare(<<EndQuery);
SELECT *
  FROM PLC.PL_SSRF_HOSTS
 WHERE policy_id  = ?
EndQuery
    $ssrf_denylist_configuration->execute($policy_id);

    my %ssrf_denylist = ();
    my (@explicit_domains, @wildcard_domains, $pure_wildcard);
    while (my $plc_row = $ssrf_denylist_configuration->fetchrow_hashref) {
        my $host = $plc_row->{host};
        my $prefix_length = 0;

        if ($host =~ /(.*)\/(\d+)/) {
            $host = $1;
            $prefix_length = $2;
        }

        my $is_valid_ipv4 = (defined inet_pton(AF_INET, $host)) ? 1 : 0;
        my $is_valid_ipv6 = (defined inet_pton(AF_INET6, $host)) ? 1 : 0;

        if ($is_valid_ipv4 || $is_valid_ipv6) {
            my $split_ip_route = F5::IpUtils::split_ip_route( ip_address => $host , die_on_error => 0 );
            if ( $split_ip_route->{completed_successfully} ) {
                $ssrf_denylist{$host}->{ip_address} = $split_ip_route->{ip_address};
                $ssrf_denylist{$host}->{route_domain} = $split_ip_route->{route_domain};

                if ($is_valid_ipv4) {
                    $ssrf_denylist{$host}->{ip_mask} = F5::IpUtils::cidr_to_netmask_ipv4($prefix_length);
                }
                else {
                    $ssrf_denylist{$host}->{ip_mask} = F5::IpUtils::cidr_to_netmask_ipv6($prefix_length);
                }

                $ssrf_denylist{$host}->{ip_address_int} =  F5::IpAddr::ip_pton($ssrf_denylist{$host}->{ip_address});
                $ssrf_denylist{$host}->{ip_mask_int} =  F5::IpAddr::ip_pton($ssrf_denylist{$host}->{ip_mask});
                # ignore_flags_mask => 32 for enabling SSRF denylist checking.

                if ($plc_row->{action_type} == F5::ASMConfig::SsrfActionType::SSRF_ACTION_DISALLOW()) {
                    $ssrf_denylist{$host}->{ignore_flags_mask} |= 32; #deny
                }
                else {
                    $ssrf_denylist{$host}->{ignore_flags_mask} |= 64; #allow
                }
            }
            else {
                $logger->info("Could not split [$host] into ip and route") if $logger;
            }
        }
        else {
            if ($plc_row->{flg_is_pattern} && $plc_row->{host} ne '*') {
                push @wildcard_domains, $plc_row;
            }
            elsif ($plc_row->{host} eq "*") {
                $pure_wildcard = $plc_row;
            }
            else {
                push @explicit_domains, $plc_row;
            }
        }
    }
    my @ordered_wildcard_domains = F5::ASMConfig::Entity::SsrfHost::sort_domains(\@wildcard_domains);
    my @all_domains = (@explicit_domains, @ordered_wildcard_domains, $pure_wildcard);
    my $dcc_pattern_order = 1;
    for my $rec (@all_domains) {
        my $row;
        my $host = $rec->{host};
        $row->{policy_name_crc} = $policy_name_crc;

        if ($host =~ m/^\*\.(.*)/g) {
            $row->{domain_name} = $1;
            $row->{flg_include_sub_domains} = 1;
        }
        else {
            $row->{domain_name} = $host;
            $row->{flg_include_sub_domains} = 0;
        }
        $row->{action_type} = $rec->{action_type};
        $row->{pattern_order} = $dcc_pattern_order++;
        $self->insert_row(
             table_name => 'DCC.ACCOUNT_SSRF_DISALLOWED_HOSTS',
             rec => $row);

    }

   my $ip_acceptlist_configuration = $dbh->prepare(<<EndQuery);
SELECT *
  FROM PLC.PL_IP_WHITELIST
 WHERE policy_id  = ?
EndQuery
    $ip_acceptlist_configuration->execute($policy_id);

    while (my $plc_row = $ip_acceptlist_configuration->fetchrow_hashref) {
        my $row;
        my $ip = $plc_row->{ip_address};

        if  ($ssrf_denylist{$ip} && ($ssrf_denylist{$ip}->{ip_mask} eq $plc_row->{mask})) {
            $row = $ssrf_denylist{$ip};
            $row->{policy_name_crc} = $policy_name_crc;
            $ssrf_denylist{$ip}->{__added_as_ip_allowlist} = 1;
        }
        else {
            $row->{policy_name_crc} = $policy_name_crc;
            my $split_ip_route = F5::IpUtils::split_ip_route( ip_address => $plc_row->{ip_address} , die_on_error => 0 );
            if (! $split_ip_route->{completed_successfully}) {
                $logger->info("Could not split [$row->{ip_address}] into ip and route") if $logger;
                next;
            }

            $row->{ip_address} = $split_ip_route->{ip_address};
            $row->{route_domain} = $split_ip_route->{route_domain};
            $row->{ip_mask} =  $plc_row->{mask};
            $row->{ip_address_int} =  F5::IpAddr::ip_pton($split_ip_route->{ip_address});
            $row->{ip_mask_int} =  F5::IpAddr::ip_pton($plc_row->{mask});
        }

        for my $ignore_flags_bitmask_field ( keys  %ignore_flags_bitmask_values ) {
            if ($plc_row->{$ignore_flags_bitmask_field}){
                $row->{ignore_flags_mask} |= $ignore_flags_bitmask_values{$ignore_flags_bitmask_field};
            }
        }

        if ($plc_row->{block_ip} == 2){ #always block
            $row->{ignore_flags_mask} |= 16;
        }
        if ($plc_row->{block_ip} == 3){ #never block
            $row->{ignore_flags_mask} |= 2;
        }

        $self->insert_row(
                table_name => 'DCC.ACCOUNT_IP_WHITELIST',
                rec => $row);
    }

    for my $ssrf_denylist_rec (values %ssrf_denylist) {
        if (!$ssrf_denylist_rec->{__added_as_ip_allowlist}) {
            $ssrf_denylist_rec->{policy_name_crc} = $policy_name_crc;
            $self->insert_row(
                    table_name => 'DCC.ACCOUNT_IP_WHITELIST',
                    rec => $ssrf_denylist_rec);
        }
    }
}

#--------------------------------------------------------------------------------

sub prepare_ip_reputation_categories{
   my ($self, $policy_name_crc, $policy_id) = @_;

   my $dbh = $self->{dbh};
   local $" = ',';

   $self->empty_table('DCC.ACCOUNT_IP_REPUTATION', $policy_name_crc);

   return RET_OK if $policy_id == 0;


   my $ip_rep_configuration = $dbh->selectrow_hashref(<<EndQuery,undef,$policy_id);
SELECT flg_enabled
  FROM PLC.PL_IP_REPUTATION
 WHERE policy_id  = ?
EndQuery

   # Need to re-determine whether violation is enabled, since it is forced OFF for ACCOUNT_BLOCKING (as per bd requirements)
   my $blocking_flags = $dbh->selectrow_hashref(<<EndQuery,undef,$policy_id, F5::Violation::VIOL_MALICIOUS_IP());
SELECT flg_alarm,flg_reject,flg_learn
  FROM PLC.PL_BLOCKING
 WHERE policy_id = ?
   AND viol_index = ?
EndQuery

   my $is_learning_enabled = $self->is_learning_enabled(policy_id => $policy_id);

   my $bitmask_blocking_flags =  $dbh->selectall_hashref(<<EndQuery,'category_bitmask_val',undef,$policy_id);
SELECT b.category_bitmask_val,
       a.flg_alarm,a.flg_block
  FROM PLC.PL_IP_REPUTATION_CATEGORIES as a,
       PLC.IP_REPUTATION_CATEGORIES as b 
 WHERE a.category_id=b.category_id 
   AND policy_id = ?
EndQuery


   my $alarm_mask = 0;
   my $block_mask = 0;
   my $learn_mask = 0;

   if ( $ip_rep_configuration->{flg_enabled} ) {
      for my $plc_row ( values %{$bitmask_blocking_flags} ) {
        if ( $blocking_flags->{flg_alarm} && $plc_row->{flg_alarm} ){
             $alarm_mask |= $plc_row->{category_bitmask_val};
        }

        if ( $blocking_flags->{flg_reject} && $plc_row->{flg_block} ){
             $block_mask |= $plc_row->{category_bitmask_val};
        }

        if ( $is_learning_enabled && $blocking_flags->{flg_learn} ) {
             $learn_mask |= $plc_row->{category_bitmask_val};
        }
      }
   }

   $self->insert_row(
                     table_name => 'DCC.ACCOUNT_IP_REPUTATION',
                     rec => { policy_name_crc => $policy_name_crc,
                              categories_alarm_mask  =>  $alarm_mask,
                              categories_block_mask =>  $block_mask,
                              categories_learn_mask => $learn_mask,
                            }
                     );

    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_allowed_response_codes  {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  $self->empty_table('DCC.ACCOUNT_ALLOWED_RESPONSE_CODES', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @plc_fields = qw(response_code);

  my $plc_sth = $dbh->prepare("SELECT @plc_fields from PLC.PL_ALLOWED_RESPONSE_CODES WHERE policy_id = ?");

  $plc_sth->execute($policy_id);

  while(my $plc_row = $plc_sth->fetchrow_hashref) {

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_ALLOWED_RESPONSE_CODES',
                      rec => { policy_name_crc => $policy_name_crc,
                               allowed_response_code => $plc_row->{response_code} });
  }

  return RET_OK;
}


#--------------------------------------------------------------------------------

sub positional_parameters_row_handler {
    my ($self, %params) = @_;
    my $row = $params{row};

    my $object_crc = $self->{h_objects}->{$row->{object_id}}->{objcrc};
    my $param_name_crc = $self->{h_keys}->{$row->{param_attrib_id}}->{keycrc};
    my $flg_has_positional_params = $self->{h_objects}->{$row->{object_id}}->{flg_has_positional_params};
    my $flg_is_pattern = $self->{h_objects}->{$row->{object_id}}->{flg_is_pattern};

    if (!($object_crc && $param_name_crc && $flg_has_positional_params && $flg_is_pattern && $row->{url_segment_index})) {
        $row->{__skip_row} = 1;
        return;
    }

    $row->{object_crc} = $object_crc;
    $row->{param_name_crc} = $param_name_crc;
    $row->{url_segment_index} = $row->{url_segment_index};
}

#--------------------------------------------------------------------------------

sub session_awareness_row_handler {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $row = $params{row};
    my $policy_id = $params{policy_id};
    my $policy_name_crc = $params{policy_name_crc};

    $row->{viol_set} = F5::Violation::Utils::get_viol_set(viols => $self->{viols},
                                                          field_name => 'flg_session_awareness');

    $row->{flg_use_apm_username} = ( $row->{application_username_type} == F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_APM() ) ? 1 : 0; 
    my $sa_login_page_table = $row->{application_username_type} == F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_ALL_LOGIN_PAGES() ? 'PLC.PL_LOGIN_OBJECTS' : 'PLC.PL_SESSION_AWARENESS_LOGIN_OBJECTS' ;

    my ($login_page_count) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT count(*)
  FROM $sa_login_page_table
 WHERE policy_id = ?
EndQuery

    # If flg_.._enabled flag is disabled, change corresponding threshold to 0
    for my $field (keys %$row) {
        if ($field =~ /^flg_(.*)_enabled$/) {
            my $threshold_field = $1;

            # Disable Force Logout if there are no Login Pages nor APM
            if ($threshold_field =~ /block_all/ && $row->{flg_block_only_authenticated} && ! $row->{flg_use_apm_username} &&
                $login_page_count == 0) {

                $row->{$field} = 0;
            }

            if ($row->{flg_violations_actions_enabled} == 0 || $row->{$field} == 0) {
                $row->{$threshold_field} = 0;
            }
        }
    }

    if ($row->{flg_check_block_all_period} == 0) {
        $row->{block_all_period} = 0;
    }

}
#--------------------------------------------------------------------------------
sub redirection_domains_row_handler{
    my ($self, %params) = @_;

    my $row = $params{row};
    if ($self->{redirect_action}->{$row->{policy_id}} == REDIRECT_ACTION_NONE) {
        $row->{__skip_row} = 1;
    }

    if ($row->{flg_is_pattern} ) {
        $row->{__skip_row} = 1;
    }
}
#--------------------------------------------------------------------------------

sub pb_configuration_row_handler{
    my ($self, %params) = @_;

    my $row = $params{row};
    #0 - NONE
    #1 - MANUAL
    #2 - AUTO
    $row->{pb_learning_mode} = $row->{flg_learning} + $row->{flg_enabled};
    $row->{pb_learn_from_response} = $row->{flg_learn_from_responses};
    $row->{pb_learn_login_forms} = $row->{flg_learn_login_pages};
    $row->{pb_learning_progress} = $row->{learning_progress};
    $row->{pb_server_technologies} = $row->{flg_server_technologies};
    $row->{pb_cookie_learning} = ($row->{flg_cookies} != F5::ASMConfig::TighteningMode::TIGHTENING_MODE_NEVER);
}

#--------------------------------------------------------------------------------

sub prepare_ignored_login_urls{
  my ($self, $policy_name_crc, $policy_id) = @_;

  $self->empty_table('DCC.ACCOUNT_LOGIN_IGNORE_URLS', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @ignored_login_urls_suggestions = F5::ASMConfig::Entity::Suggestion->retrieve_all(dbh => $self->{dbh},
                                                                                       policy => F5::ASMConfig::Policy->new( policy_id =>  $policy_id ), 
                                                                                       get_criteria => { element_type => F5::ASMConfig::ElementType::LOGIN_OBJECT,
                                                                                                         suggestion_status => F5::ASMConfig::SuggestionStatus::SUGGESTION_STATUS_IGNORED,
                                                                                                       },
                                                                                      );

  for my $ignored_login_urls_suggestion (@ignored_login_urls_suggestions ){
      my $ignored_url = $ignored_login_urls_suggestion->{element}->{url};
      $self->insert_row(
                      table_name => 'DCC.ACCOUNT_LOGIN_IGNORE_URLS',
                      rec => { policy_name_crc => $policy_name_crc,
                               object_crc =>   F5::CRC::get_crc_any("get_crc_obj", 
                                                                    $ignored_url->{app_id}, 
                                                                    $policy_name_crc, 
                                                                    $ignored_url->{url_name}, 
                                                                    $self->{unescape}),
                               },
                        );
  }
}

#--------------------------------------------------------------------------------

sub prepare_disallowed_file_types  {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  $self->empty_table('DCC.ACCOUNT_APPLICATION_DISALLOWED_FILE_TYPES', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @plc_fields = qw(name);

  my $plc_sth = $dbh->prepare("SELECT @plc_fields from PLC.PL_FILETYPES WHERE policy_id = ? and flg_is_allowed = ?");

  $plc_sth->execute($policy_id,FALSE);

  my @dcc_fields = qw(policy_name_crc filetype_crc); 
  while(my $plc_row = $plc_sth->fetchrow_hashref) {
    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_APPLICATION_DISALLOWED_FILE_TYPES',
                      rec => { policy_name_crc => $policy_name_crc,
                               filetype_crc => F5::CRC::get_crc_any("get_crc32", $plc_row->{name}) });
  }

  return RET_OK;
}

#--------------------------------------------------------------------------------

sub get_criteria_bitmap {
  my ($self) = $_[0];
  my (@criteria) = @_[1..8];
  my ($dbh, $policy_name_crc, $object_crc) = @_[9..11];

  my $bitmap = 0;
  for (0..7) {
    if ($criteria[$_]) {
      $bitmap |= 2**$_;
      $self->insert_row(
                        table_name => 'DCC.ACCOUNT_PREREQUISITE_VALID_CRITERIA',
                        ignore => 1,
                        rec => { policy_name_crc => $policy_name_crc,
                                 object_crc => $object_crc,
                                 criterium_type => 2**$_,
                                 criterium_match => $criteria[$_] });
    }
  }
  return $bitmap;
}

#--------------------------------------------------------------------------------

sub prepare_preprequisite_obj_attrs_tbl {
  my ($self, $policy_name_crc, $policy_id) = @_;
  my $dbh = $self->{dbh};
  local $" = ',';
  $self->empty_table('DCC.ACCOUNT_LOGIN_OBJECT_ATTRIBUTES', $policy_name_crc);
  $self->empty_table('DCC.ACCOUNT_PREREQUISITE_VALID_CRITERIA', $policy_name_crc);


  return RET_OK if $policy_id == 0;

  my @plc_fields = qw( id object_id policy_id resp_contains
                        resp_omits resp_http_status
                        header_contains header_omits cookie_contains auth_type username_param password_param
                        param_contains);

  my ($application_username_type) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id );
SELECT application_username_type 
  FROM PLC.PL_SESSION_AWARENESS_SETTINGS
 WHERE policy_id = ?
EndQuery

  my $sa_login_page_table = $application_username_type == F5::ASMConfig::ApplicationUsernameType::APPLICATION_USERNAME_ALL_LOGIN_PAGES() ? 'PLC.PL_LOGIN_OBJECTS' : 'PLC.PL_SESSION_AWARENESS_LOGIN_OBJECTS'; 

  my $sa_login_pages = $dbh->selectall_hashref(<<EndQuery, 'object_id', undef, $policy_id);
SELECT object_id, policy_id 
  FROM $sa_login_page_table
 WHERE policy_id = ?
EndQuery

  my %login_objects_response_codes_omits;

  my $resp_http_status_omits = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id);
SELECT login_object_id, response_code
  FROM PLC.PL_LOGIN_OBJECTS_OMITS_RESPONSE_CODES
 WHERE policy_id = ?
EndQuery

  for my $resp_http_status_omit ( @$resp_http_status_omits ) {
      my $login_object_id = $resp_http_status_omit->{login_object_id};
      my $omit_response_code = ( $resp_http_status_omit->{response_code} =~ /^([1-5])xx$/) ? $1 :  $resp_http_status_omit->{response_code};
      push @{$login_objects_response_codes_omits{$login_object_id}}, $omit_response_code;
  }

  my $plc_sth = $dbh->prepare("SELECT @plc_fields from PLC.PL_LOGIN_OBJECTS WHERE policy_id = ?");
  $plc_sth->execute($policy_id);
  while(my $plc_row = $plc_sth->fetchrow_hashref) {
    $plc_row->{policy_name_crc} = $policy_name_crc;
    $plc_row->{flg_has_session_awareness} = exists($sa_login_pages->{$plc_row->{object_id}}) ? 1 : 0;
    $plc_row->{object_crc} = $self->{h_objects}->{$plc_row->{object_id}}->{objcrc};
    if ( defined $plc_row->{resp_http_status} ){
         $plc_row->{resp_http_status}= (  $plc_row->{resp_http_status} =~ /^([1-5])xx$/) ? $1 :  $plc_row->{resp_http_status};
    }
    if ( defined $login_objects_response_codes_omits{$plc_row->{id}} ){
        $plc_row->{resp_http_status_omits} = join (",", @{$login_objects_response_codes_omits{$plc_row->{id}}});
    }

    $plc_row->{criteria_bitmap} = $self->get_criteria_bitmap(
            $plc_row->{resp_omits},
            $plc_row->{resp_contains}, 
            $plc_row->{resp_http_status},
            $plc_row->{header_contains},
            $plc_row->{header_omits},
            $plc_row->{cookie_contains},
            $plc_row->{param_contains},
            $plc_row->{resp_http_status_omits},
            $dbh, $policy_name_crc, $plc_row->{object_crc}
    );

    $plc_row->{username_param_crc} = F5::CRC::get_crc_any('get_crc', $plc_row->{username_param});
    $plc_row->{password_param_crc} = F5::CRC::get_crc_any('get_crc', $plc_row->{password_param});

    my @dcc_fields = qw(policy_name_crc object_crc criteria_bitmap auth_type username_param_crc password_param_crc);

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_LOGIN_OBJECT_ATTRIBUTES',
                      fields => \@dcc_fields,
                      rec => $plc_row);
  }

  return RET_OK;
}

#--------------------------------------------------------------------------------
sub prepare_logout_objects_attrs_tbl {
  my ($self, $policy_name_crc, $policy_id) = @_;
  my $dbh = $self->{dbh};
  local $" = ',';
  $self->empty_table('DCC.ACCOUNT_LOGOUT_OBJECT_ATTRIBUTES', $policy_name_crc);
  $self->empty_table('DCC.ACCOUNT_LOGOUT_OBJECT_VALID_CRITERIA', $policy_name_crc);


  return RET_OK if $policy_id == 0;

  my $plc_sth = $dbh->prepare("SELECT * from PLC.PL_REVOKE_OBJECTS WHERE policy_id = ?");
  $plc_sth->execute($policy_id);
  while(my $plc_row = $plc_sth->fetchrow_hashref) {
    $plc_row->{policy_name_crc} = $policy_name_crc;
    $plc_row->{object_crc} = $self->{h_objects}->{$plc_row->{revoke_object_id}}->{objcrc};
    $plc_row->{criteria_bitmap} = 0;
    my $index = 0;
    for my $criteria ( qw(request_omits request_contains) ){
        if ( $plc_row->{$criteria} ){    
           $plc_row->{criteria_bitmap} |= 2**$index;
           $plc_row->{criterium_type} = 2**$index;
           $plc_row->{criterium_match} = $plc_row->{$criteria};

           $self->insert_row(
                              table_name => 'DCC.ACCOUNT_LOGOUT_OBJECT_VALID_CRITERIA',
                              ignore => 1,
                              rec => $plc_row, 
                            );
        }
        $index++;
    }

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_LOGOUT_OBJECT_ATTRIBUTES',
                      rec => $plc_row);
  }

  return RET_OK;
}
#--------------------------------------------------------------------------------

sub prepare_preprequisite_targets_tbl {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';
  $self->empty_table('DCC.ACCOUNT_PREREQUISITE_OBJECT_TARGETS', $policy_name_crc);

  return RET_OK if $policy_id == 0;

  my @plc_fields = qw( policy_id target_pattern);
  my $plc_sth = $dbh->prepare("SELECT @plc_fields from PLC.PL_PREREQUISITE_OBJECT_TARGETS WHERE policy_id = ?");
  $plc_sth->execute($policy_id);

  while(my $plc_row = $plc_sth->fetchrow_hashref) {
    $plc_row->{policy_name_crc} = $policy_name_crc;
    $plc_row->{restricted_target_pattern} = $plc_row->{target_pattern};

    my @dcc_fields = qw(policy_name_crc restricted_target_pattern);
    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_PREREQUISITE_OBJECT_TARGETS',
                      fields => \@dcc_fields,
                      rec => $plc_row);
  }
  return RET_OK;
}

#--------------------------------------------------------------------------------
sub prepare_objects_header_web_profiles{
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';
  $self->empty_table('DCC.ACCOUNTS_OBJECTS_HEADER_WEB_PROFILES', $policy_name_crc);
  
  return RET_OK if $policy_id == 0;
  my @urls_with_non_default_configuration;

  if ( defined $self->{h_objects}->{urls_with_non_default_configuration} ){
      @urls_with_non_default_configuration = @{$self->{h_objects}->{urls_with_non_default_configuration}} ;
  }

  return RET_OK unless @urls_with_non_default_configuration;

  my @plc_fields = qw( policy_id obj_id profile_id header_name header_value header_order enforcement_type );
  my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields
  FROM PLC.PL_OBJECT_CONTENT_PROFILES
 WHERE policy_id = ?
   AND header_order != ?
   AND obj_id IN ( @urls_with_non_default_configuration )
   ORDER BY obj_id, header_order
EndQuery
  $plc_sth->execute($policy_id,0);

  my $dcc_pattern_order;
  while(my $plc_row = $plc_sth->fetchrow_hashref) {
    $plc_row->{policy_name_crc} = $policy_name_crc;
    my $obj_id = $plc_row->{obj_id};

    if (! $obj_id || ! exists $self->{h_objects}->{$plc_row->{obj_id}}) {
        # Invalid record in PLC -- skip
        next;
    }

    $dcc_pattern_order->{$obj_id}++;
    $plc_row->{header_order} = $dcc_pattern_order->{$obj_id};

    $plc_row->{object_crc} = $self->{h_objects}->{$plc_row->{obj_id}}->{objcrc};
    $plc_row->{flg_in_classification} = ($plc_row->{header_name} eq '*' && $plc_row->{header_value} eq '*') ? $self->{h_objects}->{$plc_row->{obj_id}}->{flg_in_classification} : 0;
    $plc_row->{content_profile_id} = $self->get_content_profile_id(%$plc_row);
    $plc_row->{type_id} = $enforcement_type_id{$plc_row->{enforcement_type}};

    $self->insert_row(
                     table_name => 'DCC.ACCOUNTS_OBJECTS_HEADER_WEB_PROFILES',
                      rec => $plc_row);
  }

  return RET_OK;
}

#--------------------------------------------------------------------------------

sub get_content_profile_id {
    my($self, %params) = @_;
    my $enforcement_type = $params{enforcement_type};
    my $profile_id = $params{profile_id};

    my $profile_cache = $self->{"h_${enforcement_type}_profiles"};

    if ($profile_cache && $profile_id) {
         return $profile_cache->{$profile_id}->{profile_templ_id}->{$enforcement_type} || 0;
    }
    return 0;
}

#----------------------------------------------------------------------

sub prepare_blocking_tbl {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    my $flg_encoding_selected = defined $params->{flg_encoding_selected} ? $params->{flg_encoding_selected} : $self->get_flg_encoding_selected($policy_name_crc);
    my $is_policy_transparent = $self->is_policy_transparent(policy_id => $policy_id);
    my $is_learning_enabled = $self->is_learning_enabled(policy_id => $policy_id);
    my $is_policy_passive =  $self->is_policy_passive(policy_id => $policy_id);

    if ( $is_policy_passive ) {
        $is_policy_transparent = 1;
    }

    my $dbh = $self->{dbh};
    my $readonly_mode = $params->{readonly_mode};

    return RET_OK if (! $policy_id);

    if (! $readonly_mode) {
        $self->empty_table("DCC.ACCOUNT_BLOCKING", $policy_name_crc);
    }

    my $behavioral_enforcement;
    my $deception_settings;
    my $policy_settings;

    $behavioral_enforcement = $dbh->selectrow_hashref(<<EndQuery, undef, $policy_id);
SELECT *
  FROM PLC.PL_BEHAVIORAL_ENFORCEMENT
 WHERE policy_id = ?
EndQuery

    $deception_settings = $dbh->selectrow_hashref(<<EndQuery, undef, $policy_id);
SELECT *
  FROM PLC.PL_DECEPTION_SETTINGS
 WHERE policy_id = ?
EndQuery

    $policy_settings = $dbh->selectrow_hashref(<<EndQuery, undef, $policy_name_crc);
SELECT *
  FROM PLC.PL_POLICIES
 WHERE policy_name_crc = ?
EndQuery

    my $viols = $dbh->selectall_hashref('SELECT * FROM PLC.PL_BLOCKING WHERE policy_id = ?', 'viol_index', undef, $policy_id);
    my $session_awareness_viols = $dbh->selectall_hashref('SELECT * FROM PLC.PL_SESSION_AWARENESS_VIOLATIONS WHERE policy_id = ?', 'viol_index', undef, $policy_id);

    my $behavioral_enforcement_viols = {};
    if ($behavioral_enforcement && $behavioral_enforcement->{flg_enabled} && $behavioral_enforcement->{flg_block_violations}) {
        $behavioral_enforcement_viols = $dbh->selectall_hashref('SELECT * FROM PLC.PL_BEHAVIORAL_ENFORCEMENT_VIOLATIONS WHERE policy_id = ?', 'viol_index', undef, $policy_id);
    }


    my $rv  = $self->calculate_viols( viols =>$viols,
                                      session_awareness_viols => $session_awareness_viols,
                                      behavioral_enforcement_viols => $behavioral_enforcement_viols,
                                      policy_name_crc => $policy_name_crc,
                                      policy_id => $policy_id,
                                      flg_encoding_selected => $flg_encoding_selected,
                                      is_learning_enabled => $is_learning_enabled,
                                     );

    $self->{viols} =  $rv->{viols};
    $rv->{dcc_rec}->{is_policy_transparent} = $is_policy_transparent;

    if ($deception_settings) {
        $self->{deception_settings} = $deception_settings;
        $rv->{dcc_rec} = { %{$rv->{dcc_rec}}, %$deception_settings };
        $rv->{dcc_rec}->{flg_use_custom_deception_pages} = $deception_settings->{flg_user_defined_responses} || 0;
        if ($deception_settings->{flg_user_defined_responses}) {
            $rv->{dcc_rec}->{server_technology_id} = 0;
        }
    }
    if ($behavioral_enforcement) {
        $rv->{dcc_rec}->{flg_behavioral_enforcement} = $behavioral_enforcement->{flg_enabled} || 0;
        if ($behavioral_enforcement->{flg_enabled}) {
            $rv->{dcc_rec} = { %{$rv->{dcc_rec}}, %$behavioral_enforcement };
        }
    }
    if ($policy_settings) {
        $rv->{dcc_rec}->{flg_managed_by_bewaf} = $policy_settings->{flg_managed_by_bewaf};
    }

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_BLOCKING',
                      rec => $rv->{dcc_rec},
                     );
    return RET_OK;
    }

#----------------------------------------------------------------------
sub calculate_viols {
    my($self, %params) = @_;
    my $viols = $params{viols};
    my $session_awareness_viols = $params{session_awareness_viols};
    my $behavioral_enforcement_viols = $params{behavioral_enforcement_viols};
    my $policy_name_crc = $params{policy_name_crc};
    my $policy_id = $params{policy_id};
    my $flg_encoding_selected = $params{flg_encoding_selected};
    my $is_learning_enabled = $params{is_learning_enabled};
    my $dbh = $self->{dbh};

    my %blocking_considerations = (
                                   F5::Violation::VIOL_CSRF() => { feature_enabled_table => "PLC.PL_POLICY_CSRF" },
                                   F5::Violation::VIOL_CSRF_EXPIRED() => { feature_enabled_table => "PLC.PL_POLICY_CSRF" },
                                   F5::Violation::VIOL_DATA_GUARD() => { feature_enabled_table => "PLC.PL_RESPONSE_SCRUBBING" },
                                   F5::Violation::VIOL_EVASION() => { ignore => 1 },
                                   F5::Violation::VIOL_ATTACK_SIGNATURE() => { ignore => 1,
                                                                               is_behavioral_enabled => sub {
                                                                                   my($behavioral_enabled_count) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT COUNT(*)
  FROM PLC.PL_POLICY_NEGSIG_SETS
 WHERE policy_id = ?
   AND (flg_alarm = 1 OR flg_reject = 1)
EndQuery
                                                                                   return $behavioral_enabled_count ? 1 : 0;
                                                                               },
                                                                          },
                                   F5::Violation::VIOL_MALICIOUS_IP() => { ignore => 1 },
                                   F5::Violation::VIOL_VIRUS() => { is_feature_enabled  => sub { 
                                                                        my($flg_inspect_http_uploads) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT flg_inspect_http_uploads
  FROM PLC.PL_POLICY_ANTIVIRUS
 WHERE policy_id = ?
EndQuery

                                                                        my($xml_profile_with_inspect_exists) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id, F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML());
SELECT flg_inspect_soap_attachments
  FROM PLC.PL_CONTENT_PROFILES xml_prof
 WHERE xml_prof.policy_id = ?
   AND xml_prof.content_profile_type = ?
   AND xml_prof.flg_inspect_soap_attachments = 1
EndQuery

                                                                        return ($flg_inspect_http_uploads || $xml_profile_with_inspect_exists);
                                                                    },
                                                                  },

                                   F5::Violation::VIOL_HTTP_PROTOCOL() => { is_learn_flag_set => sub {
                                      my ($sub_viol_learn_flg_set) = $dbh->selectrow_array("SELECT MAX(flg_pb_enabled) FROM PLC.PL_HTTP_CHECKS WHERE policy_id = ?", undef, $policy_id);
                                                                        return $sub_viol_learn_flg_set;
                                                                                                     }
                                                                          },
                               );

    for my $viol_index (keys %$viols) {
        my $viol = $viols->{$viol_index};

        $viol->{flg_session_awareness} = exists($session_awareness_viols->{$viol_index}) ? 1 : 0;
        my $is_behavioral_enabled = 1;
        if ($blocking_considerations{$viol_index}->{is_learn_flag_set}) {
            $viol->{flg_learn} &&= $blocking_considerations{$viol_index}->{is_learn_flag_set}->();
        }
        if (!$is_learning_enabled) {
            $viol->{flg_learn} = 0;
        }

        my $feature_is_enabled = 1;
        if (! $flg_encoding_selected) {
            # Disable all violations unless policy is configured
            $feature_is_enabled = 0;
            $is_behavioral_enabled = 0;
        }

        if ($feature_is_enabled && $blocking_considerations{$viol_index}) {

            if ($blocking_considerations{$viol_index}->{ignore}) {
                $feature_is_enabled = 0;
            }
            if ($blocking_considerations{$viol_index}->{feature_enabled_table}) {

                ($feature_is_enabled) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT flg_is_enabled
  FROM $blocking_considerations{$viol_index}->{feature_enabled_table}
 WHERE policy_id = ?
EndQuery
                $is_behavioral_enabled = $feature_is_enabled;
            }
            if ($blocking_considerations{$viol_index}->{is_feature_enabled}) {
                $feature_is_enabled = $blocking_considerations{$viol_index}->{is_feature_enabled}->();
                $is_behavioral_enabled = $feature_is_enabled;
            }
        }

        $viol->{flg_behavioral} = 0;
        if (exists($behavioral_enforcement_viols->{$viol_index})) {
            if ($blocking_considerations{$viol_index} && $blocking_considerations{$viol_index}->{is_behavioral_enabled}) {
                $viol->{flg_behavioral} = $blocking_considerations{$viol_index}->{is_behavioral_enabled}->();
            }
            else {
                $viol->{flg_behavioral} = ($viol->{flg_reject} || $viol->{flg_alarm}) && $is_behavioral_enabled;
            }
        }

        if (! $feature_is_enabled) {
            $viol->{flg_reject} = 0;
            $viol->{flg_alarm} = 0;
            $viol->{flg_learn} = 0;
        }

        $viol->{is_enabled} = $viol->{flg_reject} || $viol->{flg_alarm} || $viol->{flg_learn} || 0;
    }

    my $rv;
    $rv->{viols} = $viols;

    my %viol_sets_to_build = ( 'viol_set_block' => 'flg_reject',
                               'viol_set_alarm' => 'flg_alarm',
                               'viol_set_learn' => 'flg_learn',
                               'viol_set_behavioral' => 'flg_behavioral',
                             );

    my $dcc_rec = { policy_name_crc => $policy_name_crc };
    for my $viol_set (keys %viol_sets_to_build) {
        $dcc_rec->{$viol_set} = F5::Violation::Utils::get_viol_set(viols => $viols,
                                                                   field_name => $viol_sets_to_build{$viol_set});
    }

    $rv->{dcc_rec} = $dcc_rec;

    return $rv;
}
#----------------------------------------------------------------------

sub is_policy_transparent {
    my($self, %params) = @_;
    if (!defined($self->{is_policy_transparent})) {

        ($self->{is_policy_transparent}) = $self->{dbh}->selectrow_array('SELECT flg_disable_rejects from PLC.PL_POLICIES WHERE id = ?', 
                                                                         undef, $params{policy_id});
    }
    return $self->{is_policy_transparent};
}

#----------------------------------------------------------------------

sub is_policy_passive {
    my($self, %params) = @_;
    if (!defined($self->{is_policy_passive})) {

        ($self->{is_policy_passive}) = $self->{dbh}->selectrow_array('SELECT flg_passive_mode from PLC.PL_POLICIES WHERE id = ?',
                                                                         undef, $params{policy_id});
    }
    return $self->{is_policy_passive};
}

#----------------------------------------------------------------------

sub is_policy_case_insensitive {
    my($self, %params) = @_;
    if (!defined($self->{is_policy_case_insensitive})) {

        ($self->{is_policy_case_insensitive}) = $self->{dbh}->selectrow_array('SELECT flg_is_case_insensitive from PLC.PL_POLICIES WHERE id = ?',
                                                                         undef, $params{policy_id});
    }
    return $self->{is_policy_case_insensitive};
}
#--------------------------------------------------------------------------------

sub is_learning_enabled {
    my($self, %params) = @_;
    if (!defined($self->{is_learning_enabled})) {

        ($self->{is_learning_enabled}) = $self->{dbh}->selectrow_array('SELECT flg_learning from PLC.PL_POLICY_BUILDER_CONFIG WHERE policy_id = ?',
                                                                         undef, $params{policy_id});
    }
    return $self->{is_learning_enabled};
}

#--------------------------------------------------------------------------------

sub is_bot_defense_enabled {
    my($self, %params) = @_;
    if (!defined($self->{is_bot_defense_enabled})) {

        ($self->{is_bot_defense_enabled}) = $self->{dbh}->selectrow_array('SELECT flg_enabled from PLC.PL_BOT_DEFENSE WHERE policy_id = ?',
                                                                         undef, $params{policy_id});
    }
    return $self->{is_bot_defense_enabled};
}

#--------------------------------------------------------------------------------

sub get_bitmask_pos {
    my($code) = @_;
    my $pos = 0;

    # Trim trailing zeros
    $code =~ s/^(\d)(0*)$/$1/;
    $pos = 4 * length($2);

    $pos += log($code)/log(2);
    return $pos;
}

#--------------------------------------------------------------------------------

sub get_encoding {
    my ($self, %params) = @_;

    my $policy_id = $params{policy_id} || 0;
    my($encoding) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_id);
SELECT enc.encoding
  FROM PLC.PL_POLICIES pol,
       DCC.ACCOUNTS acc,
       DCC.LANGUAGE_ENCODINGS lang,
       DCC.ENCODINGS enc
 WHERE pol.id = ?
   AND pol.name = acc.policy_name
   AND acc.flg_encoding_selected = 1
   AND acc.lang_set_id = lang.lang_set_id
   AND lang.encoding_id = enc.encoding_id
EndQuery

    return $encoding;
}

#--------------------------------------------------------------------------------

sub prepare_custom_deception_responses {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    $self->empty_table("DCC.ACCOUNT_CUSTOM_RESPONSE_HEADERS", $policy_name_crc);
    $self->empty_table("DCC.ACCOUNT_CUSTOM_RESPONSES", $policy_name_crc);

    return RET_OK unless F5::License::is_waf_licensed();

    if (!$self->{deception_settings}->{flg_user_defined_responses}) {
        return RET_OK;
    }

    my $dbh = $self->{dbh};

    my $custom_responses = $dbh->selectall_hashref(<<EndQuery, 'id', undef, $policy_id);
SELECT *
  FROM PLC.PL_DECEPTION_RESPONSES
 WHERE policy_id = ?
EndQuery

    my $custom_response_id = 0;
    for my $response_page (values %$custom_responses) {
        $custom_response_id++;
        $self->insert_row(table_name => 'DCC.ACCOUNT_CUSTOM_RESPONSES',
                          rec => { policy_name_crc => $policy_name_crc,
                                   custom_response_id => $custom_response_id,
                                   response_code => $response_page->{response_code},
                                   response_body => $response_page->{response_body} });

        next if !$response_page->{response_headers};

        my $parsed_response;
        eval {
            $parsed_response = $self->parse_response_page_headers(policy_id => $policy_id,
                                                                  response_headers => $response_page->{response_headers});
        };
        if (my $err = $@) {
            $self->{logger}->error($err->{error_message});
            next;
        }

        my $headers = $parsed_response->{response_headers};
        for my $header_name (keys %$headers) {
            $self->insert_row(
                              table_name => 'DCC.ACCOUNT_CUSTOM_RESPONSE_HEADERS',
                              rec => { policy_name_crc => $policy_name_crc,
                                       custom_response_id => $custom_response_id,
                                       header_name => $header_name,
                                       header_value => $headers->{$header_name} });
        }
    }

    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_server_technology_responses {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    $self->empty_table("DCC.ACCOUNT_SERVER_TECHNOLOGY_RESPONSES", $policy_name_crc);

    return RET_OK unless F5::License::is_waf_licensed();

    if ($self->{deception_settings}->{flg_user_defined_responses}) {
        return RET_OK;
    }

    my $dbh = $self->{dbh};

    my $server_technology_id = $self->{deception_settings}->{server_technology_id};
    my $where_fields = { server_technology_id => $server_technology_id };

    # all response codes are used if none are specified by user
    my $response_codes = $self->{dbh}->selectcol_arrayref("SELECT response_code FROM PLC.PL_DECEPTION_RESPONSE_CODES WHERE policy_id = ?",
                                                          undef, $policy_id);
    if (@$response_codes) {
        $where_fields->{response_code} = $response_codes;
    }

    my $where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_fields);

    my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT responses.deception_response_id
  FROM PLC.DECEPTION_RESPONSES as responses
 LEFT JOIN PLC.DECEPTION_RESPONSES_SIGNATURES sigs
    ON sigs.deception_response_id = responses.deception_response_id
 WHERE $where_clause->{where_clause} AND sig_id IS NULL
EndQuery

    $plc_sth->execute(@{$where_clause->{where_values}});
    my $responses = $plc_sth->fetchall_arrayref({});

    for my $response (@$responses) {
        $self->insert_row(table_name => 'DCC.ACCOUNT_SERVER_TECHNOLOGY_RESPONSES',
                          rec => { policy_name_crc => $policy_name_crc,
                                   deception_response_id => $response->{deception_response_id}, });
    }
}

#--------------------------------------------------------------------------------

sub prepare_signature_responses {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    $self->empty_table("DCC.ACCOUNT_SIGNATURE_RESPONSES", $policy_name_crc);

    return RET_OK unless F5::License::is_waf_licensed();

    if (!$self->{deception_settings}->{flg_response_by_attack_type}) {
        return RET_OK;
    }

    my $dbh = $self->{dbh};

    my $policy_server_technologies = $dbh->selectcol_arrayref('SELECT server_technology_id FROM PLC.PL_SERVER_TECHNOLOGIES WHERE policy_id = ?',
                                                               undef, $policy_id);

    # add value to retrieve default pages
    push (@$policy_server_technologies, 0);

    my $where_clause = F5::DbUtils::construct_where_clause(where_fields => { server_technology_id => $policy_server_technologies});

    # For each signature retrieve one response page that corresponds to the policy server technologies
    # If no suitable page is found, a default page for the signature will be returned

    my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT sigs.sig_id, responses.deception_response_id
  FROM PLC.DECEPTION_RESPONSES responses
  JOIN PLC.DECEPTION_RESPONSES_SIGNATURES sigs
    ON sigs.deception_response_id = responses.deception_response_id
  JOIN (SELECT max(server_technology_id) max_id, sig_id
        FROM PLC.DECEPTION_RESPONSES responses1
          JOIN PLC.DECEPTION_RESPONSES_SIGNATURES sigs1
          ON responses1.deception_response_id = sigs1.deception_response_id
        WHERE $where_clause->{where_clause}
        GROUP BY sig_id) max_ids
    ON max_ids.sig_id = sigs.sig_id
  WHERE responses.server_technology_id = max_id;
EndQuery

    $plc_sth->execute(@{$where_clause->{where_values}});
    my $responses = $plc_sth->fetchall_arrayref({});
    for my $response (@$responses) {
        $self->insert_row(table_name => 'DCC.ACCOUNT_SIGNATURE_RESPONSES',
                          rec => { policy_name_crc => $policy_name_crc,
                                   sig_id => $response->{sig_id},
                                   deception_response_id => $response->{deception_response_id}, });
    }
}

#--------------------------------------------------------------------------------

sub prepare_account_deception_traps {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    $self->empty_table("DCC.ACCOUNT_DECEPTION_TRAPS", $policy_name_crc);

    my $tables_map = {
        F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_PARAM_NAME() => {
            dcc_table => 'DCC.ACCOUNT_GLOBAL_PARAMETERS',
            plc_table => 'PLC.PL_GLOBAL_PARAMS',
            where_field => 'name',
        },
        F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_COOKIE_NAME() => {
            dcc_table => 'DCC.ACCOUNT_ALLOWED_COOKIES',
            plc_table => 'PLC.PL_ALLOWED_COOKIES',
            where_field => 'cookie_name',
        },
        F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_HEADER_NAME() => {
            dcc_table => 'DCC.ACCOUNT_HEADERS',
            plc_table => 'PLC.PL_HEADERS',
            where_field => 'header_name',
        },
        F5::ASMConfig::DeceptionTrapSource::DECEPTION_TRAP_SOURCE_URL() => {
            dcc_table => 'DCC.ACCOUNT_APPLICATION_OBJECT_ATTRIBUTES',
            plc_table => 'PLC.PL_OBJECTS',
            where_field => 'uri',
        },
    };

    # clean conviction dummy entities
    for my $table (values %$tables_map) {
        F5::DbUtils::delete_rows(dbh => $self->{dbh},
                                 table_name => $table->{dcc_table},
                                 where_fields => { policy_name_crc => $policy_name_crc,
                                                   flg_conviction => 1 },
                                );
    }

    return RET_OK unless F5::License::is_waf_licensed();

    my $conditions = $self->{dbh}->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT *
  FROM PLC.DECEPTION_TRAP_CONDITIONS
EndQuery

    my %disallowed_trap_ids;

    my ($last_header_code) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_name_crc);
SELECT MAX(header_code)
  FROM DCC.ACCOUNT_HEADERS
 WHERE policy_name_crc = ?
EndQuery

    my $header_code = $last_header_code || 1;
    $header_code = Math::BigInt->new($header_code);
    my $shifter = Math::BigInt->new(1);
    if ($last_header_code) {
        $header_code <<= $shifter;
    }

    for my $condition (@$conditions) {

        next if ($disallowed_trap_ids{$condition->{deception_trap_id}});

        next if (!defined $tables_map->{$condition->{condition_source}});

        my $plc_table = $tables_map->{$condition->{condition_source}}->{plc_table};
        my $dcc_table = $tables_map->{$condition->{condition_source}}->{dcc_table};
        my $where_values = {};
        my $recs = [];

        my $crc;
        my $condition_value = $condition->{condition_value};
        if ($dcc_table eq 'DCC.ACCOUNT_GLOBAL_PARAMETERS') {
            if ($self->is_policy_case_insensitive(policy_id => $policy_id)) {
                $condition_value = lc($condition->{condition_value});
            }
            $crc = F5::CRC::get_crc_any("get_crc_str", $policy_name_crc, $condition_value, $self->{unescape});
            push (@$recs, { key_crc => $crc, policy_name_crc => $policy_name_crc });
        }
        elsif ($dcc_table eq 'DCC.ACCOUNT_ALLOWED_COOKIES') {
            $crc = F5::CRC::get_crc_any("get_crc32", $condition_value, 0);
            push (@$recs, { cookie_name_crc => $crc, policy_name_crc => $policy_name_crc });
        }
        elsif ($dcc_table eq 'DCC.ACCOUNT_APPLICATION_OBJECT_ATTRIBUTES') {
            if ($self->is_policy_case_insensitive(policy_id => $policy_id)) {
                $condition_value = lc($condition->{condition_value});
            }
            $crc = F5::CRC::get_crc_any("get_crc_obj", 0, $policy_name_crc, $condition_value, $self->{unescape});
            push (@$recs, { object_crc => $crc, app_id => 0, policy_name_crc => $policy_name_crc });
            $crc = F5::CRC::get_crc_any("get_crc_obj", 1, $policy_name_crc, $condition_value, $self->{unescape});
            push (@$recs, { object_crc => $crc, app_id => 1, policy_name_crc => $policy_name_crc });
        }
        elsif ($dcc_table eq 'DCC.ACCOUNT_HEADERS') {
            $crc = F5::CRC::get_crc_any("get_crc32", $condition_value);
            push (@$recs, { header_name_crc => $crc,
                            header_name => $condition_value,
                            header_code => $header_code,
                            policy_name_crc => $policy_name_crc, });

            $header_code <<= $shifter;
        }

        if (!defined $tables_map->{$condition->{condition_source}}->{seen_keys}) {
            $tables_map->{$condition->{condition_source}}->{seen_keys} = {};
        }

        if (defined $crc) {
            # Avoid duplicates in entities keys
            next if $tables_map->{$condition->{condition_source}}->{seen_keys}->{$crc};
            $tables_map->{$condition->{condition_source}}->{seen_keys}->{$crc} = 1;
        }

        $where_values->{policy_id} = $policy_id;
        $where_values->{$tables_map->{$condition->{condition_source}}->{where_field}} = $condition_value;
        my $where_clause = F5::DbUtils::construct_where_clause(where_fields => $where_values);
        my ($exists) = $self->{dbh}->selectrow_array(<<EndQuery, undef, @{$where_clause->{where_values}});
SELECT COUNT(*)
  FROM $plc_table
 WHERE $where_clause->{where_clause}
EndQuery

        if ($exists) {
            $disallowed_trap_ids{$condition->{deception_trap_id}} = 1;
            next;
        }

        for my $rec (@$recs) {
           $self->insert_row(table_name => $dcc_table,
                             rec => {%$rec,
                                     flg_conviction => 1,
                                    },
                            );
        }
    }

    my $where_clause = F5::DbUtils::construct_where_clause(where_fields => {
                                                               deception_trap_id => {
                                                                   values => [keys %disallowed_trap_ids],
                                                                   comparison => 'ne',
                                                               }
                                                          });

    my $injections = $self->{dbh}->selectall_arrayref(<<EndQuery, { Slice => {} }, @{$where_clause->{where_values}});
SELECT *
  FROM PLC.DECEPTION_TRAP_INJECTIONS
 WHERE $where_clause->{where_clause}
EndQuery

    for my $injection (@$injections) {
        $self->insert_row(table_name => 'DCC.ACCOUNT_DECEPTION_TRAPS',
                          rec => { policy_name_crc => $policy_name_crc,
                                   injection_type => $injection->{injection_type},
                                   deception_trap_id => $injection->{deception_trap_id}, });
    }

    return RET_OK;
}

#--------------------------------------------------------------------------------

sub prepare_alternate_response_file_tbl {
    my ($self, $policy_name_crc,$plc_id) = @_;

    my $dbh = $self->{dbh};
    
    $self->empty_table("DCC.ACCOUNT_ALTERNATE_RESPONSE_FILE", $policy_name_crc);
    $self->empty_table("DCC.ACCOUNT_ALTERNATE_RESPONSE_FILE_HEADERS", $policy_name_crc);
    
    return RET_OK if $plc_id == 0;

    my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT * FROM PLC.PL_ALTERNATE_RESPONSES WHERE policy_id = ?
EndQuery

    $plc_sth->execute($plc_id);

    while (my $plc_data = $plc_sth->fetchrow_hashref()) {
        if ( $plc_data->{response_type} eq 'redirect' && $plc_data->{redirect_url} ) { 
             $plc_data->{alternate_response_header} = <<END;
HTTP/1.0 302 Found
Location: $plc_data->{redirect_url}
Content-Length: 0
Connection: close
END
             $plc_data->{alternate_response_header}  =~ s/\n/\r\n/g ;
        }

        if ($plc_data->{cause} eq 'grpc') {
            my $additional_headers = <<END;
grpc-status: $plc_data->{grpc_status_code}
grpc-message: $plc_data->{grpc_status_message}
END
            $additional_headers  =~ s/\n/\r\n/g;
            $plc_data->{alternate_response_header} .= "\r\n$additional_headers";
        }

        $plc_data->{policy_name_crc} = $policy_name_crc;
        if (! $plc_data->{cause}) {
            $self->{logger}->debug("Policy has empty response page cause -- skipping") if $self->{logger};
            next;
        }
        if (! $plc_data->{alternate_response_header}) {
            $self->{logger}->info("Blocking response page header for cause $plc_data->{cause} is empty -- skipping") if $self->{logger};
            next;
        }

        my $parsed_response;
        eval {
            $parsed_response = $self->parse_response_page_headers(policy_id => $plc_id,
                                                                  response_headers => $plc_data->{alternate_response_header});
        };
        if (my $err = $@) {
            $self->{logger}->error("Blocking response page header for cause $plc_data->{cause} cannot be parsed ($err->{error_message})") if $self->{logger};
            next;
        }

        $plc_data->{cause} = $response_types{$plc_data->{cause}};
        $plc_data->{response_code} = $parsed_response->{response_code};

        if ($plc_data->{response_type} eq 'redirect') {
            $plc_data->{alternate_response_content} = '';
        }

        my $headers = $parsed_response->{response_headers};
        for my $header_name (keys %$headers) {
            $self->insert_row(
                              table_name => 'DCC.ACCOUNT_ALTERNATE_RESPONSE_FILE_HEADERS',
                              rec => 
                              {
                               policy_name_crc => $plc_data->{policy_name_crc},
                               cause => $plc_data->{cause},
                               header_name => $header_name,
                               header_value => $headers->{$header_name},
                              });
        }
        if ($plc_data->{cause} == $response_types{Ajax} || $plc_data->{cause} == $response_types{Ajax_login} || 
            $plc_data->{cause} == $response_types{Failed_login_honeypot_ajax} || $plc_data->{cause} == $response_types{Leaked_credentials_ajax}) {
            my $content;
            if ($plc_data->{ajax_action_type} eq 'alert_popup') {
                $content = $plc_data->{ajax_popup_message};
            }
            elsif ($plc_data->{ajax_action_type} eq 'redirect') {
                $content = $plc_data->{ajax_redirect_url};
            }
            else {
                $content = $plc_data->{ajax_custom_content};
            }
            $self->insert_row(
                              table_name => 'DCC.ACCOUNT_ALTERNATE_RESPONSE_FILE',
                              rec => {
                                      policy_name_crc => $policy_name_crc,
                                      cause => $plc_data->{cause},
                                      alternate_response_content => $content,
                                      response_code => '200',
                                     });
            $self->insert_row(
                              table_name => 'DCC.ACCOUNT_ALTERNATE_RESPONSE_FILE_HEADERS',
                              rec => {
                                      policy_name_crc => $policy_name_crc,
                                      cause => $plc_data->{cause},
                                      header_name => '<%X-TS-BP-Action%>',
                                      header_value => $ajax_action_types{$plc_data->{ajax_action_type}}
                                     });
        }
        elsif (($plc_data->{cause} == $response_types{Captcha} || $plc_data->{cause} == $response_types{Captcha_fail})) {
            # replace captcha tokens
            for my $captcha_token (qw(change solution submit image support_id audio)) {
                my $pattern = '%ASM.captcha.' . $captcha_token . '%';
                my $regex = qr/$pattern/;
                my $replacement = $captcha_token_to_replacement{$captcha_token};
                $plc_data->{alternate_response_content} =~ s/$regex/$replacement/;
            }

            # these place holders are replaced by a chunk of base64 encoded strings, representing a refresh button image or audio play image, that is being read from a file on init
            $plc_data->{alternate_response_content} =~ s/PLACE_HOLDER_FOR_CAPTCHA_CHANGE/$self->{captcha_refresh_base64_encoded_image}/;
            $plc_data->{alternate_response_content} =~ s/PLACE_HOLDER_FOR_CAPTCHA_AUDIO_ICON/$self->{captcha_audio_icon_base64_encoded_image}/;

            # remove the remaining three place holders, PLACE_HOLDER_FOR_CAPTCHA_[IMAGE|SUPPORT_ID|AUDIO], from response content and populate the respectful offsets
            $plc_data->{offset_to_image} = -1;
            $plc_data->{offset_to_support_id} = -1;
            $plc_data->{offset_to_audio} = -1;
            my $regex = qr/PLACE_HOLDER_FOR_CAPTCHA_(SUPPORT_ID|IMAGE|AUDIO)/;
            while ($plc_data->{alternate_response_content} =~ s/$regex//) {
                $plc_data->{'offset_to_' . lc($1)} = $-[0]; # LAST_MATCH_START perlvar
            }

            $self->insert_row(
                              table_name => 'DCC.ACCOUNT_ALTERNATE_RESPONSE_FILE',
                              rec => $plc_data);
        }
        else {
            $self->insert_row(
                              table_name => 'DCC.ACCOUNT_ALTERNATE_RESPONSE_FILE',
                              rec => $plc_data);
        }
    }
    return RET_OK;
}
#--------------------------------------------------------------------------------
sub prepare_alternate_response_settings{
    my ($self, $policy_name_crc, $plc_id) = @_;

    my $dbh = $self->{dbh};
    my $plc_fields = 'flg_ajax_enabled';

    $self->empty_table("DCC.ACCOUNT_ALTERNATE_RESPONSE_SETTINGS", $policy_name_crc);

    return RET_OK if $plc_id == 0;

    my $plc_sql = "SELECT $plc_fields from PLC.PL_ALTERNATE_RESPONSE_SETTINGS WHERE policy_id=$plc_id";

    my $plc_sth = $dbh->prepare($plc_sql)
               or $self->{logger}->error("prepare failed on SQL: '$plc_sql': $DBI::errstr");

    $plc_sth->execute
               or $self->{logger}->error("execute failed: $DBI::errstr");

     while(my $plc_row = $plc_sth->fetchrow_hashref) {
          $self->insert_row(
                            table_name => 'DCC.ACCOUNT_ALTERNATE_RESPONSE_SETTINGS',
                            rec => { policy_name_crc => $policy_name_crc,
                                     %$plc_row });
     }

     return RET_OK;
}
#--------------------------------------------------------------------------------

sub prepare_allowed_cookies_tbl {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    my $flg_encoding_selected = defined $params->{flg_encoding_selected} ? $params->{flg_encoding_selected} : $self->get_flg_encoding_selected($policy_name_crc);

    my $dbh = $self->{dbh};
    
    $self->empty_table("DCC.ACCOUNT_ALLOWED_COOKIES", $policy_name_crc);
    $self->empty_table("DCC.ACCOUNT_WILDCARD_ALLOWED_COOKIES", $policy_name_crc);
    
    return RET_OK if $policy_id == 0;

    my $plc_cookies = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id);
SELECT *
    FROM PLC.PL_ALLOWED_COOKIES
  WHERE policy_id = ?
 ORDER BY pattern_order,last_updated,cookie_name
EndQuery

    my %dcc_cookies;
    # We want dcc_cookies to include both PL COOKIES and parameters cookies.
    # First get PL cookies
    for my $plc_cookie (@$plc_cookies) {
          my $cookie_name = $plc_cookie->{cookie_name};
          if (! $dcc_cookies{$cookie_name}) {
              $dcc_cookies{$cookie_name} = $plc_cookie;
          }
    }


    # Now add the cookie parameters (only if no name collision)
    for my $cookie_param_name (keys %{$self->{h_cookie_parameters}}) {
          my $cookie_name = $cookie_param_name;
          my $dcc_cookie = $dcc_cookies{$cookie_name};
          if (! $dcc_cookie) {
              if ($dcc_cookies{'*'}) {
                  $dcc_cookie = dclone($dcc_cookies{'*'});
                  delete $dcc_cookie->{id};
              }
              else {
                  $dcc_cookie = {};
              }
              $dcc_cookie->{cookie_name} = $cookie_name;
              $dcc_cookies{$cookie_param_name} = $dcc_cookie;
          }
          $dcc_cookie->{flg_enforced_as_param} = 1;
          $dcc_cookie->{flg_is_pattern} = 0;
    }



    my $dcc_pattern_order = 1;

    for my $plc_row (sort { ($a->{pattern_order} || 0) <=> ($b->{pattern_order} || 0) } values %dcc_cookies) {

        my $dcc_table_name;
        $plc_row->{policy_name_crc} = $policy_name_crc;
        if (! $flg_encoding_selected) { 
            $plc_row->{flg_tightening} = 0;
        }
        #flg_tightening:  0 - Never, 2 - Selective, cookie_enforcement_mode: allow/enforce
        #  So if it's Selective (2), add the mapped value for the tightening based on cookie_enforcement_mode  (0 - off,  1 - add_allow 2 - add_enforce)
        if ($plc_row->{flg_tightening}) {
            $plc_row->{flg_tightening} = ( $plc_row->{cookie_enforcement_mode} eq 'enforce')  ? 1 : $plc_row->{flg_tightening} ;
        }

         $plc_row->{cookie_enforcement_mode} = ( $plc_row->{cookie_enforcement_mode} eq 'enforce')  ? 1 : 0 ;

         my $cookie_name_crc = F5::CRC::get_crc_any("get_crc32", $plc_row->{cookie_name}, 0);
         $plc_row->{cookie_name_crc} = $cookie_name_crc;

        if ($plc_row->{flg_is_pattern}) { # wildcard
            $dcc_table_name = "DCC.ACCOUNT_WILDCARD_ALLOWED_COOKIES";
            $plc_row->{pattern} = $plc_row->{cookie_name};
            $plc_row->{pattern_order} = $dcc_pattern_order++;
            $self->{logger}->error("Missing pattern_order on wildcard cookie ($plc_row->{cookie_name})")  if !defined($plc_row->{pattern_order});
        }
        else {
            $dcc_table_name = "DCC.ACCOUNT_ALLOWED_COOKIES";
        }
        $self->insert_row(
                          table_name => $dcc_table_name,
                          rec => $plc_row);
    }

}
#--------------------------------------------------------------------------------

sub prepare_valid_host_names_tbl {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    my $flg_encoding_selected = defined $params->{flg_encoding_selected} ? $params->{flg_encoding_selected} : $self->get_flg_encoding_selected($policy_name_crc);

    my $dbh = $self->{dbh};
    local $" = ',';
    my $domain_dcc_table_name = "DCC.ACCOUNT_VALID_HOST_NAME_DOMAINS";
    my $dcc_table_name = "DCC.ACCOUNT_VALID_HOST_NAMES_REGEXES";
    my $data_dir       = "$ACCOUNT_DATA_DIR/valid_host_names";
 
    $self->empty_table($dcc_table_name, $policy_name_crc);
    $self->empty_table($domain_dcc_table_name, $policy_name_crc);


    return RET_OK if $policy_id == 0;

    my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT *
    FROM PLC.PL_VALID_HOST_NAMES 
  WHERE policy_id = ?
EndQuery

    $plc_sth->execute($policy_id);

    #First build list of regex strings
    my @regexes;
    while (my $plc_row = $plc_sth->fetchrow_hashref) {
        # Escape qw(. [ ])
        my $regex = $plc_row->{host_name};
        $regex =~ s/([\[\]\.])/\\$1/g;
        if ($plc_row->{flg_include_subdomains}) {
            push @regexes,  '/^(?:.*\.)?' . $regex . '$/i';
        }
        else {
            push @regexes,  '/^' . $regex. '$/i';
        }

        $self->insert_row(table_name => $domain_dcc_table_name,
                          rec        => { policy_name_crc => $policy_name_crc, domain_name => $plc_row->{host_name}, flg_include_sub_domains => $plc_row->{flg_include_subdomains},
                                        },
                          );
    }
    my $regex_string =  join (",", @regexes);
   
    #Then write them to a file
    my $md5_sum = Digest::MD5->new->add($regex_string)->hexdigest;
    my $filename = "$data_dir/regexes_$md5_sum";
    F5::Utils::File::write_file(filename => $filename,
                                contents => $regex_string,
                                overwrite => 0,
                                permissions_mode => 0664,
                               );

    #Then add that filename to the DCC table
    $self->insert_row(
                      table_name => $dcc_table_name,
                      rec        => {
                                      policy_name_crc       => $policy_name_crc,
                                      regexes_filename => $filename, 
                                    }
                     );

    F5::DbUtils::delete_unreferenced_files(dbh          => $dbh,
                                           logger       => $self->{logger},
                                           db_file_refs => [
                                                            { table_name => $dcc_table_name,
                                                              field      => 'regexes_filename',
                                                            },
                                                           ],
                                           dir          => "$data_dir");

    return RET_OK;
}


#--------------------------------------------------------------------------------

sub prepare_navigation_params_tbl {
	my ($self, $policy_name_crc, $plc_id)  = @_;

    my $dbh = $self->{dbh};
	my $plc_fields = 'object_path, param_name';

	$self->empty_table("DCC.ACCOUNT_NAVIGATION_PARAMS", $policy_name_crc);
	return RET_OK if $plc_id == 0;

	my $plc_sql = "SELECT $plc_fields from PLC.PL_CRAWLER_NAVIGATION_PARAMS WHERE policy_id=$plc_id";

	my $plc_sth = $dbh->prepare($plc_sql)
		or $self->{logger}->error("prepare failed on SQL: '$plc_sql': $DBI::errstr");	

	$plc_sth->execute
		or $self->{logger}->error("execute failed: $DBI::errstr");

	while(my $plc_row = $plc_sth->fetchrow_hashref) {
          my $object_path_crc = ($plc_row->{object_path}) ? F5::CRC::get_crc_any('get_crc', $plc_row->{object_path}, $self->{unescape}) : '0';

          $self->insert_row(
                            table_name => 'DCC.ACCOUNT_NAVIGATION_PARAMS',
                            rec => { policy_name_crc => $policy_name_crc,
                                     object_path_crc => $object_path_crc,
                                     param_name => $plc_row->{param_name} });
	}

	return RET_OK;
}
#--------------------------------------------------------------------------------
sub dynamic_flows_row_handler {
    my ($self, %params)  = @_;
    my $row = $params{row};

    $self->plc_regexp_merge($row);
}

#--------------------------------------------------------------------------------

sub prepare_sensitive_params {
  my($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};

  $self->empty_table('DCC.ACCOUNT_SENSITIVE_PARAMS', $policy_name_crc);

  my %seen_param_names;
  my $get_plc_sth = $dbh->prepare("SELECT param_name FROM PLC.PL_HIDE_PARAM_VALUES WHERE policy_id = ?");
  $get_plc_sth->execute($policy_id);

  while(my $row = $get_plc_sth->fetchrow_hashref) {
    $seen_param_names{$row->{param_name}} = 1;
    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_SENSITIVE_PARAMS',
                      rec => { policy_name_crc => $policy_name_crc,
                               from_params_tables => 0,
                               param_name => $row->{param_name} });
  }

  my @param_tables = qw(PLC.PL_GLOBAL_PARAMS
                        PLC.PL_OBJECT_PARAMS
                        PLC.PL_FLOW_PARAMS);

  for my $param_table (@param_tables) {
      # Adding also flow/object/global parameters names with flg_is_sensitive checked
      my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT DISTINCT param.name
   FROM $param_table param, PLC.PL_PARAM_ATTRIBUTES param_attrib
  WHERE param.param_attrib_id = param_attrib.id
    AND param.policy_id = ?
    AND param_attrib.flg_is_sensitive = 1
    AND param.flg_is_pattern = 0
EndQuery

      $plc_sth->execute($policy_id);
      while (my $plc_row = $plc_sth->fetchrow_hashref) {
          if ($seen_param_names{$plc_row->{name}}) {
              next;
          }
          $seen_param_names{$plc_row->{name}} = 1;
          $self->insert_row(
                            table_name => 'DCC.ACCOUNT_SENSITIVE_PARAMS',
                            rec => { policy_name_crc => $policy_name_crc,
                                     from_params_tables => 1,
                                     param_name => $plc_row->{name} });
      }
  }

  return RET_OK;
}

#--------------------------------------------------------------------------------
sub input_flags2bitmap
{
    my ($flg_pd, $flg_qs, $flg_check_pd, $flg_check_qs) = @_;
    my $bitmap = 0;
    
    $bitmap|= INPUT_HAS_NO_USER_DATA    if(!$flg_pd && !$flg_qs);
    $bitmap|= INPUT_HAS_USER_DATA       if($flg_pd || $flg_qs);
    $bitmap|= INPUT_CHECK_USER_DATA     if($flg_check_pd || $flg_check_qs);

    return $bitmap;
}

#----------------------------------------------------------------------

sub add_object_crc {
  my ($self, $extract) = @_;

  if (!$self->{h_objects}->{$extract->{obj_id}}) {
    $self->{logger}->error("Undefined object id $extract->{obj_id}!");
    return;
  }
  $extract->{object_crc} = $self->{h_objects}->{$extract->{obj_id}}->{objcrc};
  return;
}

#----------------------------------------------------------------------

sub dcc_plc_field_rename {
  my($self, $extract, $scope) = @_;

  for my $i (0..$#{$scope->{plc_fields}}) {
    my $dcc_name = $scope->{dcc_fields}->[$i];
    my $plc_name = $scope->{plc_fields}->[$i];
    $extract->{$dcc_name} = $extract->{$plc_name};
  }
  return;
}

#----------------------------------------------------------------------

sub plc_regexp_merge {
  my($self, $extract, $scope, $method) = @_;

  my $full_regexp = '';
  if ($extract->{regexp_prefix}) {
    $full_regexp .= "(?<=$extract->{regexp_prefix})";
  }
  $full_regexp .= $extract->{regexp_body};
  if ($extract->{regexp_suffix}) {
    $full_regexp .= "(?=$extract->{regexp_suffix})";
  }
  $extract->{reg_exp} = $full_regexp;
  return;
}

#----------------------------------------------------------------------

sub get_negative_check_bitmap {
  my(%params) = @_;

  my %bit_values = ( name_metachar_flag => 1,
                     # 2 is used by BD for regexp
                     value_metachar_flag => 4,
                     signature_flag => 8,
                   );

  my $bitmap = 0;
  for my $bit_value_param (keys %bit_values) {
    if ($params{$bit_value_param}) {
      $bitmap |= $bit_values{$bit_value_param};
    }
  }

  return $bitmap;
}

#----------------------------------------------------------------------

sub preload_charset_templates {
  my($self, $policy_name_crc, $policy_id) = @_;
  my $dbh = $self->{dbh};

  %{$self->{base_charsets}} = @{$dbh->selectcol_arrayref('SELECT charset_type, charset FROM PLC.PL_CHARSETS WHERE policy_id = ?',
					      { Columns => [1,2] },
					      $policy_id)};

  my $load_charset_templates_sth = $dbh->prepare('SELECT charset_templ_id, charset FROM DCC.ACCOUNT_CHARSET_TEMPLATES WHERE policy_name_crc = ?');
  $load_charset_templates_sth->execute($policy_name_crc);

  while(my $charset_templ_row = $load_charset_templates_sth->fetchrow_hashref) {
    $self->{charset_templates}->{$charset_templ_row->{charset}} = { charset_templ_id => $charset_templ_row->{charset_templ_id} };
  }

  $self->empty_table('DCC.ACCOUNT_CHARSETS', $policy_name_crc);
  for my $base_charset_type (sort keys %{$self->{base_charsets}}) {
    my $charset_templ_id = $self->find_or_add_charset_template($policy_name_crc, $self->{base_charsets}->{$base_charset_type});
    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_CHARSETS',
                      rec => { policy_name_crc => $policy_name_crc,
                               charset_type => $base_charset_type,
                               charset_templ_id => $charset_templ_id });
  }
}

#----------------------------------------------------------------------

sub find_or_add_charset_template {
  my($self, $policy_name_crc, $target_charset) = @_;
  my $dbh = $self->{dbh};

  if (exists $self->{charset_templates}->{$target_charset}) {
    $self->{charset_templates}->{$target_charset}->{flg_in_use} = 1;
    return $self->{charset_templates}->{$target_charset}->{charset_templ_id};
  }

  # New charset needs to be added
  my $new_charset_templ_id;
  my ($max_charset_templ_id) = $dbh->selectrow_array(<<EndQuery, undef, $policy_name_crc);
SELECT MAX(charset_templ_id) 
  FROM DCC.ACCOUNT_CHARSET_TEMPLATES
  WHERE policy_name_crc = ?
EndQuery

  $max_charset_templ_id ||= 0;
  my @used_charset_templ_ids = map { $_->{charset_templ_id} } values %{$self->{charset_templates}};

  $new_charset_templ_id = max( $max_charset_templ_id, @used_charset_templ_ids) + 1;

  $self->{charset_templates}->{$target_charset} =  { charset_templ_id => $new_charset_templ_id,
					   flg_in_use => 1 };

  $self->insert_row(
                    table_name => 'DCC.ACCOUNT_CHARSET_TEMPLATES',
                    rec => { policy_name_crc => $policy_name_crc,
                             charset_templ_id => $new_charset_templ_id,
                             charset => $target_charset });
  return $new_charset_templ_id;
}

#----------------------------------------------------------------------

sub cleanup_charset_templates {
  my($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  my @in_use_ids = map { $_->{charset_templ_id} } grep { $_->{flg_in_use} } values %{$self->{charset_templates}};

  if (@in_use_ids) {
    $dbh->do("DELETE FROM DCC.ACCOUNT_CHARSET_TEMPLATES WHERE policy_name_crc = ? AND charset_templ_id NOT IN (@in_use_ids)",
	     undef, $policy_name_crc);
  }

}

#----------------------------------------------------------------------

sub get_charset_template {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $policy_name_crc = $params{policy_name_crc};
  my $base_type = $params{base_type};
  my $override_table = $params{override_table};
  my $override_id_field = $params{override_id_field};
  my $override_id_value = $params{override_id_value};

  my @charset;
  if (defined ($self->{base_charsets}->{$params{base_type}}) ) {
      @charset = split(//, $self->{base_charsets}->{$params{base_type}});
  }

  # Load overrides, if any
  my $load_overrides_sth = $dbh->prepare("SELECT metachar, flg_is_allowed FROM $override_table WHERE $override_id_field = ?");
  $load_overrides_sth->execute($override_id_value);
  while(my $override_row = $load_overrides_sth->fetchrow_hashref) {
    $charset[$override_row->{metachar}] = $override_row->{flg_is_allowed};
  }

  return $self->find_or_add_charset_template($policy_name_crc, join('', @charset));
}

#----------------------------------------------------------------------
sub preload_allowed_methods {
  my($self, $policy_name_crc, $policy_id) = @_;
  my $dbh = $self->{dbh};

  my $load_method_templates =  $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc);
SELECT *
  FROM DCC.ACCOUNT_ALLOWED_METHODS
 WHERE policy_name_crc = ?
EndQuery

  my %allowed_methods_refs;
  for my $load_method_template ( @$load_method_templates ){
     push @{$allowed_methods_refs{$load_method_template->{obj_methods_templ_id}}}, $load_method_template->{method_code};
  }

  my %last_used_methods_templates;
  for my $allowed_methods_ref ( keys %allowed_methods_refs){
       my @sorted_method_codes = sort @{$allowed_methods_refs{$allowed_methods_ref}};
       my $ref_id_key = join "_", @sorted_method_codes;
       $last_used_methods_templates{$ref_id_key} = $allowed_methods_ref;
  }

  $self->{allowed_methods_templates}->{last_used_methods_templates} = \%last_used_methods_templates;

  my $allowed_methods = $dbh->selectall_hashref(<<EndQuery,'method_code', undef, $policy_id);
SELECT policy_methods.method_code, policy_methods.act_as_method_code 
  FROM PLC.PL_ALLOWED_METHODS policy_methods,
       DCC.METHOD_POOL method_pool
 WHERE policy_id = ?
   AND policy_methods.method_code = method_pool.method_code
EndQuery

  my @sorted_method_codes = sort keys %$allowed_methods;
  my $ref_id_key = join "_", @sorted_method_codes;

  my $max_ref_id = $dbh->selectrow_array(<<EndQuery, undef, $policy_name_crc);
SELECT MAX(obj_methods_templ_id) 
  FROM DCC.ACCOUNT_ALLOWED_METHODS
  WHERE policy_name_crc = ?
EndQuery

  $self->{allowed_methods_templates}->{current_max_ref_id} = $max_ref_id || 0;
  $self->{allowed_methods_templates}->{policy_allowed_methods} = $allowed_methods;
 
  require F5::ASM::COAPI;
  my $found_methods = F5::ASM::COAPI::Query->new('asm_http_method');
  my %default_act_as;
  for my $method ( @$found_methods){
      $default_act_as{$method->{leaf_name}} =  ( defined $allowed_methods->{$method->{code}} ) ? $allowed_methods->{$method->{code}}->{act_as_method_code} :
                                  $method->{default_act_as} =~ /POST/ ? 2 : 1;
  }
  
  $self->{allowed_methods_templates}->{default_act_as} = \%default_act_as;

  my $obj_methods_templ_id = $self->find_or_add_allowed_methods_template(ref_id_key => $ref_id_key,
                                                                         policy_name_crc => $policy_name_crc,
                                                                         policy_id => $policy_id,
                                                                        );
  $dbh->do("UPDATE DCC.ACCOUNTS SET obj_methods_templ_id = ? WHERE policy_name_crc = ?",
             undef, $obj_methods_templ_id, $policy_name_crc);  

}
#----------------------------------------------------------------------
sub find_or_add_allowed_methods_template {
  my($self, %params) = @_;

  my $dbh = $self->{dbh};
  my $policy_name_crc = $params{policy_name_crc};
  my $policy_id = $params{policy_id};
  my $obj_id = $params{obj_id};
  my $method_code = $params{method_code};
  my $ref_id_key = $params{ref_id_key}; 

  if (!exists $self->{allowed_methods_templates}->{policy_allowed_methods}) {
	 $self->{logger}->trace("$policy_id $policy_name_crc do preload_methods") if $self->{logger};
	 $self->preload_allowed_methods($policy_name_crc, $policy_id);
  }

  my $base_methods = $self->{allowed_methods_templates}->{policy_allowed_methods};
  my $url_allowed_methods;

  if ( $obj_id ) {
      my $url_override_methods = {};
      if ($method_code) {
          $url_override_methods->{$method_code} = {method_code => $method_code, flg_is_allowed => 1};
      }
      else {
          $url_override_methods = $dbh->selectall_hashref(<<EndQuery,'method_code', undef, $policy_id, $obj_id);
SELECT A.flg_is_allowed, method_code 
  FROM PLC.PL_OBJECT_METHOD_OVERRIDES as A, DCC.METHOD_POOL as B 
 WHERE policy_id = ? 
   AND obj_id = ? 
   AND A.method_name = B.method_name
EndQuery
      }

      for my $base_method ( values %$base_methods ){
         my $base_method_code = $base_method->{method_code};
         if (!exists $url_override_methods->{$base_method_code} ||  $url_override_methods->{$base_method_code}->{flg_is_allowed}){
            $url_allowed_methods->{$base_method_code} = $base_method;
         }
      }

      if ( $url_override_methods && %$url_override_methods ){
        for my $override_method ( values %$url_override_methods  ){
            my $override_method_code = $override_method->{method_code};
            $override_method->{act_as_method_code} =  $self->{allowed_methods_templates}->{$override_method_code} || 1 ;
            if (!exists $base_methods->{$override_method_code} && $override_method->{flg_is_allowed}){
                $url_allowed_methods->{$override_method_code} = $override_method;
            }
        }
      }
  
      my @sorted_method_codes = sort keys %$url_allowed_methods;
      $ref_id_key = join "_", @sorted_method_codes;
  }
  else{
      $url_allowed_methods = $base_methods;
  }

  if (exists $self->{allowed_methods_templates}->{$ref_id_key}->{obj_methods_templ_id} ){
      return $self->{allowed_methods_templates}->{$ref_id_key}->{obj_methods_templ_id};
  }
  elsif ( exists  $self->{allowed_methods_templates}->{last_used_methods_templates}->{$ref_id_key} ){
     my $found_obj_methods_templ_id = $self->{allowed_methods_templates}->{last_used_methods_templates}->{$ref_id_key};
     push @{$self->{allowed_methods_templates}->{used_method_templ_ids}}, $found_obj_methods_templ_id; 
     $self->{allowed_methods_templates}->{$ref_id_key}->{obj_methods_templ_id} = $found_obj_methods_templ_id;
     return $found_obj_methods_templ_id;
  }
  my $obj_methods_templ_id = ++$self->{allowed_methods_templates}->{current_max_ref_id};

  for my $allowed_method ( values %$url_allowed_methods ){
      $self->insert_row(
                      table_name => 'DCC.ACCOUNT_ALLOWED_METHODS',
                      rec => { policy_name_crc => $policy_name_crc,
                               method_code => $allowed_method->{method_code},
                               act_as_method_code => $allowed_method->{act_as_method_code},
                               obj_methods_templ_id => $obj_methods_templ_id });
  }

  push @{$self->{allowed_methods_templates}->{used_method_templ_ids}}, $obj_methods_templ_id;
  $self->{allowed_methods_templates}->{$ref_id_key}->{obj_methods_templ_id} = $obj_methods_templ_id;
  return $obj_methods_templ_id;
}
#----------------------------------------------------------------------
sub cleanup_method_templates {
  my($self, $policy_name_crc, $policy_id) = @_;
  local $" = ',';

  my @in_use_ids = @{$self->{allowed_methods_templates}->{used_method_templ_ids}};

  if (@in_use_ids) {
    $self->{dbh}->do("DELETE FROM DCC.ACCOUNT_ALLOWED_METHODS WHERE policy_name_crc = ? AND obj_methods_templ_id NOT IN (@in_use_ids)",
	     undef, $policy_name_crc);
  }

}
#----------------------------------------------------------------------

sub load_max_profile_templ_id {
   my ($self, %params) = @_;
   my $policy_name_crc = $params{policy_name_crc};

   ($self->{max_profile_templ_id}) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_name_crc);
SELECT MAX(profile_templ_id)
  FROM DCC.CONTENT_PROFILE_TEMPLATES
  WHERE policy_name_crc = ?
EndQuery

   $self->{max_profile_templ_id} ||= 0;
   return $self->{max_profile_templ_id};
}

#----------------------------------------------------------------------

sub prepare_json_profiles {
  my($self, $policy_name_crc, $policy_id, $params) = @_;
  my $dbh = $self->{dbh};

  local $" = ',';

  if ($params->{readonly_mode}) {
      my $existing_json_profiles_sth = $dbh->prepare(<<EndQuery);
SELECT profile_templ_id, profile_id
  FROM DCC.CONTENT_PROFILE_TEMPLATES   
   WHERE policy_name_crc = ? AND flg_is_json_security = ?
EndQuery

      $existing_json_profiles_sth->execute($policy_name_crc, 1);

      while (my $existing_json_profile = $existing_json_profiles_sth->fetchrow_hashref) {   
          $existing_json_profile->{content_type} = 'json';
          $self->{h_json_profiles}->{$existing_json_profile->{profile_id}}->{profile_templ_id}->{$existing_json_profile->{content_type}} = $existing_json_profile->{profile_templ_id};
      }
      return 1;
   }
   require F5::JSONProfile;

   $self->{logger}->debug('Preparing JSON profiles') if $self->{logger};

   if ( !defined($self->{max_profile_templ_id}) ) {
       $self->load_max_profile_templ_id(policy_name_crc => $policy_name_crc);
   }

   my @compare_fields = qw(profile_id
                           defense_config_filename
                           json_schemas_md5
                           charset_templ_id
                           flg_check_element_metachars
                           flg_check_signatures
                           profile_name
                           flg_handle_json_parameters
                        );

   my $existing_json_templates = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, 1);
SELECT profile_templ_id, @compare_fields
  FROM DCC.CONTENT_PROFILE_TEMPLATES
  WHERE policy_name_crc = ? AND flg_is_json_security = ?
EndQuery

 $self->{h_json_profiles} = $dbh->selectall_hashref(<<EndQuery, 'profile_id', undef, 1, $policy_id, F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON());
SELECT json_prof.id as profile_id, json_prof.*
  FROM           PLC.PL_CONTENT_PROFILES json_prof
        LEFT JOIN PLC.PL_OBJECT_CONTENT_PROFILES obj_json_prof
              ON json_prof.id = obj_json_prof.profile_id
        LEFT JOIN PLC.PL_PARAM_CONTENT_PROFILES param_json_prof
              ON json_prof.id = param_json_prof.profile_id
        LEFT JOIN PLC.PL_WEBSOCKET_URLS websocket_json_prof
              ON (json_prof.id = websocket_json_prof.websocket_json_profile_id AND websocket_json_prof.flg_websocket_message_format_json = ?)
 WHERE json_prof.policy_id = ?
   AND json_prof.content_profile_type = ?
   AND (param_json_prof.profile_id IS NOT NULL OR obj_json_prof.profile_id IS NOT NULL OR websocket_json_prof.websocket_json_profile_id IS NOT NULL)
EndQuery

   for my $json_profile (values %{$self->{h_json_profiles}}) {
        my $validation_file_result;
        if (F5::License::is_waf_licensed()) {
            $validation_file_result = F5::JSONProfile::handle_validation_files(dbh => $dbh,
                                                                               logger => $self->{logger},
                                                                               policy_name_crc => $policy_name_crc,
                                                                               profile_id => $json_profile->{profile_id},
                                                                               profile_rest_uuid => $json_profile->{rest_uuid},
                                                                              );
            if ($validation_file_result && ! $validation_file_result->{completed_successfully}) {
                my $error_msg = 'Failed to validate JSON schema ';
                if ($validation_file_result->{messages}) {
                    $error_msg .= join('; ', @{$validation_file_result->{messages}});
                }
                if ($validation_file_result->{missing_files}) {
                    $error_msg .= 'Missing JSON schema files: ' . join('; ', @{$validation_file_result->{missing_files}});
                }
                $self->{logger}->error($error_msg) if $self->{logger};
            }
            else {
                $json_profile->{json_schemas_md5} = $validation_file_result->{json_md5};
            }
        }

        $json_profile->{schema_filename} = '';
        $json_profile->{defense_config_filename} = '';
        $json_profile->{json_schemas_md5} ||= '';
        $json_profile->{graphql_schemas_md5} = '';
        $json_profile->{flg_is_json_security} = 1;
        $json_profile->{charset_templ_id} ||= 0 ;

        if ( $json_profile->{flg_check_element_metachars}) {
            if (! ref($self->{base_charsets}) || ! scalar( keys %{$self->{base_charsets}})) {
                $self->preload_charset_templates($policy_name_crc, $policy_id);
            }

            $json_profile->{charset_templ_id} = $self->get_charset_template(
                                                          policy_name_crc => $policy_name_crc,
                                                          base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_JSON_PROFILE(),
                                                          override_table => 'PLC.PL_CONTENT_PROFILE_VALUE_METACHARS',
                                                          override_id_field => 'profile_id',
                                                          override_id_value => $json_profile->{profile_id});
        }

        my $defense_config_result = F5::JSONProfile::create_defense_config(dbh => $dbh,
                                                                           policy_name_crc => $policy_name_crc,
                                                                           profile_rest_uuid => $json_profile->{rest_uuid},
									   profile_id => $json_profile->{profile_id});

        $self->{logger}->debug("Defense config file for profile_id $json_profile->{profile_id} is $defense_config_result->{defense_config_filename}") if $self->{logger};
        if ($defense_config_result && $defense_config_result->{defense_config_filename}) {
          $json_profile->{defense_config_filename} = copy_policy_file_for_account(policy_file => $defense_config_result->{defense_config_filename},
                                                                             account_template => $ACCOUNT_DATA_DIR . '/json/defense_config_%MD5%');
        }

        my $prev_schema_filename;
        my ($existing_json_template) = grep(find_matching_hash($_, $json_profile, \@compare_fields), @$existing_json_templates);
        $json_profile->{content_type} = 'json';
        if ($existing_json_template) {
           $json_profile->{profile_templ_id}->{$json_profile->{content_type}} = $existing_json_template->{profile_templ_id};
           $prev_schema_filename = $existing_json_template->{json_schemas_md5};
        }
        else {
           $prev_schema_filename = $json_profile->{json_schemas_md5};
           $json_profile->{policy_name_crc} = $policy_name_crc;
           if ($self->{ucs_upgrade_real_policy_id}) {
              $self->ucs_upgrade_profile_id_fix(profile => $json_profile);
           }
           $self->{max_profile_templ_id}++;
           $self->insert_row(
                            table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                            rec => { %$json_profile, profile_templ_id => $self->{max_profile_templ_id} },
                            set_date_fields_to_now => [ 'modify_time' ],
                            );
           $json_profile->{profile_templ_id}->{$json_profile->{content_type}} = $self->{max_profile_templ_id};
       }

       my $profile_templ_id = $json_profile->{profile_templ_id}->{$json_profile->{content_type}};
       if (! F5::License::is_waf_licensed()) {
           $self->empty_table('DCC.JSON_VALIDATION_FILES', $policy_name_crc);
 
           $dbh->do(<<EndQuery, undef, '', $profile_templ_id, $policy_name_crc);
UPDATE DCC.CONTENT_PROFILE_TEMPLATES 
  SET json_schemas_md5 = ?
  WHERE profile_templ_id = ?
    AND policy_name_crc = ? 
EndQuery

       }
       else {
           if ($validation_file_result && $validation_file_result->{completed_successfully}) {
               my $json_md5 = $validation_file_result->{json_md5} || '';
               if (!$prev_schema_filename  || ( $prev_schema_filename ne $json_md5 ) ) {
                  $dbh->do(<<EndQuery, undef, $json_md5, $profile_templ_id, $policy_name_crc);
UPDATE DCC.CONTENT_PROFILE_TEMPLATES 
  SET json_schemas_md5 = ? 
  WHERE profile_templ_id = ?
    AND policy_name_crc = ?
EndQuery
               }
               elsif ($existing_json_template) {
                  next;
               }

               $dbh->do(<<EndQuery, undef, $profile_templ_id, $policy_name_crc);
DELETE FROM DCC.JSON_VALIDATION_FILES
      WHERE profile_templ_id = ?
        AND policy_name_crc = ?
EndQuery

               my $file_id = 0;
               for my $validation_file (@{$validation_file_result->{profile_validation_files} || [] }) {

                   $file_id++;
                   $self->insert_row(table_name => 'DCC.JSON_VALIDATION_FILES',
                                     rec => {file_id => $file_id,
                                             policy_name_crc => $policy_name_crc,
                                             profile_templ_id => $profile_templ_id,
                                             schema_filename => copy_policy_file_for_account(policy_file => $validation_file->{local_filename},
                                                                                             account_template => $ACCOUNT_DATA_DIR . '/json/schema_%MD5%'),
                                            },
                                    );
               }
           }
       }
   }

   # Mark unused JSON templates for deletion
   my %newly_used_json_templates = map { ($_ => 1) } map { values %{$_->{profile_templ_id}} } values %{$self->{h_json_profiles}};
   my @unused_json_templates = grep { !$newly_used_json_templates{$_->{profile_templ_id}} } @$existing_json_templates;
   if (@unused_json_templates) {
     my @placeholders = ('?') x scalar @unused_json_templates;
     $dbh->do("DELETE FROM DCC.CONTENT_PROFILE_TEMPLATES WHERE profile_templ_id IN (@placeholders) AND policy_name_crc = ?",
             undef, (map { $_->{profile_templ_id} } @unused_json_templates), $policy_name_crc);
     $dbh->do("DELETE FROM DCC.JSON_VALIDATION_FILES WHERE profile_templ_id IN (@placeholders) AND policy_name_crc = ?",
             undef, (map { $_->{profile_templ_id} } @unused_json_templates), $policy_name_crc);
   }

   my @json_entities = (
                      {
                       table => 'PLC.PL_OBJECT_CONTENT_PROFILES',
                       fields => [ qw(A.id A.profile_id A.obj_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),
                      save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_objects}->{$plc_row->{obj_id}}->{profile_templ_id} = $profile_templ_id;
                       }
                      },
                      {
                       table => 'PLC.PL_WEBSOCKET_URLS',
                       fields => [ qw(A.websocket_json_profile_id A.websocket_url_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),
                       profile_id_field => 'websocket_json_profile_id',
                       save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_websocket_urls}->{$plc_row->{websocket_url_id}}->{websocket_json_profile_templ_id} = $profile_templ_id;
                       }
                      },
                      {
                       table => 'PLC.PL_PARAM_CONTENT_PROFILES',
                       fields => [ qw(A.id A.profile_id A.param_attrib_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_JSON(),
                       save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_keys}->{$plc_row->{param_attrib_id}}->{profile_templ_id} = $profile_templ_id;
                       }
                      },
                     );
  for my $json_entity (@json_entities) {
    my @plc_fields = @{$json_entity->{fields}};
    my $profile_id_field = $json_entity->{profile_id_field} || 'profile_id';
    my $get_json_profile_entities_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields 
  FROM $json_entity->{table} A, PLC.PL_CONTENT_PROFILES B
 WHERE A.policy_id = ? AND A.$profile_id_field = B.id AND B.content_profile_type=?
EndQuery
    $get_json_profile_entities_sth->execute($policy_id, $json_entity->{profile_type});
    while(my $plc_row = $get_json_profile_entities_sth->fetchrow_hashref) {
      $plc_row->{content_type}='json';
      next if not exists $self->{h_json_profiles}->{$plc_row->{$profile_id_field}};
      my $profile_templ_id = $self->{h_json_profiles}->{$plc_row->{$profile_id_field}}->{profile_templ_id}->{$plc_row->{content_type}};
      $json_entity->{save_profile_templ_id}->($self, $plc_row, $profile_templ_id);
    }
  }

  F5::DbUtils::delete_unreferenced_files(dbh => $dbh,
                                         logger => $self->{logger},
                                         db_file_refs => [
                                                           { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                             field => 'defense_config_filename'
                                                           },
                                                           { table_name => 'DCC.JSON_VALIDATION_FILES',
                                                             field => 'schema_filename'
                                                           },
                                                         ],
                                         dir => "$ACCOUNT_DATA_DIR/json");


  return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_grpc_profiles {
  my($self, $policy_name_crc, $policy_id, $params) = @_;
  if (!F5::License::is_waf_licensed() || !F5::BigipUtils::is_feature_enabled('grpc_enforcement')) {
      return 1;
  }
  require F5::GRPCProfile;

  my $dbh = $self->{dbh};

  local $" = ',';

  if ($params->{readonly_mode}) {
      my $existing_grpc_profiles_sth = $dbh->prepare(<<EndQuery);
SELECT profile_templ_id, profile_id
  FROM DCC.CONTENT_PROFILE_TEMPLATES
   WHERE policy_name_crc = ? AND flg_is_grpc_security = ?
EndQuery

      $existing_grpc_profiles_sth->execute($policy_name_crc, 1);


      while (my $existing_grpc_profile = $existing_grpc_profiles_sth->fetchrow_hashref) {
          $existing_grpc_profile->{content_type} = 'grpc';
          $self->{h_grpc_profiles}->{$existing_grpc_profile->{profile_id}}->{profile_templ_id}->{$existing_grpc_profile->{content_type}} = $existing_grpc_profile->{profile_templ_id};
      }
      return 1;
   }

   $self->{logger}->debug('Preparing gRPC profiles') if $self->{logger};
   my $account_dir = "$ACCOUNT_DATA_DIR/grpc";

   if ( !defined($self->{max_profile_templ_id}) ) {
       $self->load_max_profile_templ_id(policy_name_crc => $policy_name_crc);
   }

   my @compare_fields = qw(profile_id
                           charset_templ_id
                           flg_is_grpc_security
                           flg_check_element_metachars
                           flg_check_signatures
                           flg_tolerate
                           flg_check_document_length
                           flg_detect_string_base64
                           max_document_length
                           profile_name
                           schema_filename
                        );

   my $existing_grpc_templates = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, 1);
SELECT profile_templ_id, @compare_fields
  FROM DCC.CONTENT_PROFILE_TEMPLATES
  WHERE policy_name_crc = ? AND flg_is_grpc_security = ?
EndQuery

 $self->{h_grpc_profiles} = $dbh->selectall_hashref(<<EndQuery, 'profile_id', undef, $policy_id, F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRPC());
SELECT grpc_prof.id as profile_id, grpc_prof.*
  FROM PLC.PL_CONTENT_PROFILES grpc_prof
       JOIN PLC.PL_OBJECT_CONTENT_PROFILES obj_grpc_prof
         ON grpc_prof.id = obj_grpc_prof.profile_id
 WHERE grpc_prof.policy_id = ?
   AND grpc_prof.content_profile_type = ?
EndQuery

   for my $grpc_profile (values %{$self->{h_grpc_profiles}}) {
        $grpc_profile->{content_type} = 'grpc';
        $grpc_profile->{flg_is_grpc_security} = 1;
        $grpc_profile->{charset_templ_id} ||= 0 ;

        if ( $grpc_profile->{flg_check_element_metachars}) {
            if (! ref($self->{base_charsets}) || ! scalar( keys %{$self->{base_charsets}})) {
                $self->preload_charset_templates($policy_name_crc, $policy_id);
            }

            $grpc_profile->{charset_templ_id} = $self->get_charset_template(
                                                          policy_name_crc => $policy_name_crc,
                                                          base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_PARAM_VALUE(),
                                                          override_table => 'PLC.PL_CONTENT_PROFILE_VALUE_METACHARS',
                                                          override_id_field => 'profile_id',
                                                          override_id_value => $grpc_profile->{profile_id});
        }

        my $res = $dbh->selectall_hashref(<<EndQuery, 'def_attrib_id',undef, $grpc_profile->{profile_id});
SELECT def_attrib_id,def_attrib_value, flg_check_attrib
  FROM PLC.PL_GRPC_PROFILE_DEFENSE_ATTRIBUTES
 WHERE profile_id = ?
EndQuery

        my %grpc_defs = (
                          1 => { def_attrib => 'GRPC_DEFENSE_MAX_DATA_LENGTH',
                                 flg_check_attrib_field => 'flg_check_document_length',
                                 attrib_value_field => 'max_document_length',
                               },
                          2 => { def_attrib => 'GRPC_DEFENSE_ALLOW_UNKNOWN_FIELDS',
                                 attrib_value_field => 'flg_tolerate',
                               },

                        );


        for my $def_attrib_id (keys %grpc_defs) {
            my $grpc_def = $grpc_defs{$def_attrib_id};
            my $plc_rec = $res->{$def_attrib_id};

            if (exists $grpc_def->{flg_check_attrib_field}) {
                $grpc_profile->{$grpc_def->{flg_check_attrib_field}} = $plc_rec->{flg_check_attrib};
                $grpc_profile->{$grpc_def->{attrib_value_field}} = $plc_rec->{def_attrib_value};
            }
            else {
               $grpc_profile->{$grpc_def->{attrib_value_field}} = ( $plc_rec->{flg_check_attrib} && $plc_rec->{def_attrib_value});
            }
        }

        my $idl_file_result = F5::GRPCProfile::handle_idl_files(dbh => $dbh,
                                                                policy_name_crc => $policy_name_crc,
                                                                profile_rest_uuid => $grpc_profile->{rest_uuid},
                                                                profile_id => $grpc_profile->{profile_id},
                                                                base_dir => $account_dir,
                                                               );

        if ($idl_file_result && ! $idl_file_result->{completed_successfully}) {
           $self->{logger}->error( join('; ', @{$idl_file_result->{messages}}) ) if $self->{logger};
           next;
        }

        if ($idl_file_result && $idl_file_result->{idl_file}) {
            $grpc_profile->{schema_filename} = $idl_file_result->{proto_dir} . "/$idl_file_result->{idl_file}";
        }

        my ($existing_grpc_template) = grep(find_matching_hash($_, $grpc_profile, \@compare_fields), @$existing_grpc_templates);
        $grpc_profile->{content_type} = 'grpc';
        if ($existing_grpc_template) {
            $grpc_profile->{profile_templ_id}->{$grpc_profile->{content_type}} = $existing_grpc_template->{profile_templ_id};
        }
        else {
            $grpc_profile->{policy_name_crc} = $policy_name_crc;
            if ($self->{ucs_upgrade_real_policy_id}) {
                $self->ucs_upgrade_profile_id_fix(profile => $grpc_profile);
            }
            $self->{max_profile_templ_id}++;
            $self->insert_row(
                             table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                             rec => { %$grpc_profile, profile_templ_id => $self->{max_profile_templ_id} },
                             set_date_fields_to_now => [ 'modify_time' ],
                            );
            $grpc_profile->{profile_templ_id}->{$grpc_profile->{content_type}} = $self->{max_profile_templ_id};
       }
   }

   # Mark unused GRPC templates for deletion
   my %newly_used_grpc_templates = map { ($_ => 1) } map { values %{$_->{profile_templ_id}} } values %{$self->{h_grpc_profiles}};
   my @unused_grpc_templates = grep { !$newly_used_grpc_templates{$_->{profile_templ_id}} } @$existing_grpc_templates;
   if (@unused_grpc_templates) {
     my @placeholders = ('?') x scalar @unused_grpc_templates;
     $dbh->do("DELETE FROM DCC.CONTENT_PROFILE_TEMPLATES WHERE profile_templ_id IN (@placeholders) AND policy_name_crc = ?",
             undef, (map { $_->{profile_templ_id} } @unused_grpc_templates), $policy_name_crc);
   }

   my @grpc_entities = (
                      {
                       table => 'PLC.PL_OBJECT_CONTENT_PROFILES',
                       fields => [ qw(A.id A.profile_id A.obj_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRPC(),
                       save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_objects}->{$plc_row->{obj_id}}->{profile_templ_id} = $profile_templ_id;
                       }
                      },
                     );

  for my $grpc_entity (@grpc_entities) {
    my @plc_fields = @{$grpc_entity->{fields}};
    my $profile_id_field = $grpc_entity->{profile_id_field} || 'profile_id';
    my $get_grpc_profile_entities_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields
  FROM $grpc_entity->{table} A, PLC.PL_CONTENT_PROFILES B
 WHERE A.policy_id = ? AND A.$profile_id_field = B.id AND B.content_profile_type=?
EndQuery

    $get_grpc_profile_entities_sth->execute($policy_id, $grpc_entity->{profile_type});
    while (my $plc_row = $get_grpc_profile_entities_sth->fetchrow_hashref) {
      $plc_row->{content_type} = 'grpc';
      next if not exists $self->{h_grpc_profiles}->{$plc_row->{$profile_id_field}};
      my $profile_templ_id = $self->{h_grpc_profiles}->{$plc_row->{$profile_id_field}}->{profile_templ_id}->{$plc_row->{content_type}};
      $grpc_entity->{save_profile_templ_id}->($self, $plc_row, $profile_templ_id);
    }
  }

  F5::DbUtils::delete_unreferenced_files(dbh => $dbh,
                                          logger => $self->{logger},
                                          db_file_refs => [
                                                      { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                        field => 'schema_filename' },
                                                     ],
                                          db_value_to_filename_conversion_func => sub {
                                             my $filename = shift;
                                             $filename =~ s|^$account_dir/([^/]*).*$|$account_dir/$1|;  #Take directory name of to clean.
                                             return $filename;
                                          },
                                          dir => $account_dir);

  return RET_OK;
}

#----------------------------------------------------------------------

my %cost_by_field_type = (
                          F5::ASMConfig::GraphqlSchemaType::GRAPHQL_SCHEMA_TYPE_SCALAR() => 1,
                          F5::ASMConfig::GraphqlSchemaType::GRAPHQL_SCHEMA_TYPE_OBJECT() => 1,
                          F5::ASMConfig::GraphqlSchemaType::GRAPHQL_SCHEMA_TYPE_ARRAY() => 10
                         );

sub prepare_graphql_profiles {
    my($self, $policy_name_crc, $policy_id, $params) = @_;
    my $dbh = $self->{dbh};

    local $" = ',';

    if ( !F5::License::is_waf_licensed() ) {
        $self->empty_table("DCC.CONTENT_PROFILE_TEMPLATES", undef, {"flg_is_graphql_security" => 1});
        return RET_OK;
    }

    if ($params->{readonly_mode}) {
        my $existing_graphql_profiles_sth = $dbh->prepare(<<EndQuery);
SELECT profile_templ_id, profile_id
  FROM DCC.CONTENT_PROFILE_TEMPLATES
   WHERE policy_name_crc = ? AND flg_is_graphql_security = ?
EndQuery

        $existing_graphql_profiles_sth->execute($policy_name_crc, 1);
        while (my $existing_graphql_profile = $existing_graphql_profiles_sth->fetchrow_hashref) {
            $existing_graphql_profile->{content_type} = 'graphql';
            $self->{h_graphql_profiles}->{$existing_graphql_profile->{profile_id}}->{profile_templ_id}->{$existing_graphql_profile->{content_type}} = $existing_graphql_profile->{profile_templ_id};
        }
        return 1;
     }

     $self->{logger}->debug('Preparing GraphQL profiles') if $self->{logger};

     if ( !defined($self->{max_profile_templ_id}) ) {
         $self->load_max_profile_templ_id(policy_name_crc => $policy_name_crc);
     }

     my @compare_fields = qw(profile_id
                             charset_templ_id
                             graphql_schemas_md5
                             flg_check_param_value_length
                             max_param_value_length
                             flg_check_document_length
                             max_document_length
                             flg_check_element_metachars
                             flg_check_signatures
                             profile_name
                             flg_tolerate
                             flg_is_graphql_security
                             flg_check_max_batched_queries
                             max_batched_queries
                             flg_check_max_depth
                             max_depth
                             flg_allow_introspection_query
                             flg_check_max_query_cost
                             max_query_cost
                             block_disallowed_response_patterns
                             all_disallowed_response_patterns
                          );

    my $existing_graphql_templates = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, 1);
SELECT profile_templ_id, @compare_fields
  FROM DCC.CONTENT_PROFILE_TEMPLATES
  WHERE policy_name_crc = ? AND flg_is_graphql_security = ?
EndQuery

    $self->{h_graphql_profiles} = $dbh->selectall_hashref(<<EndQuery, 'profile_id', undef, $policy_id, F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRAPHQL());
SELECT graphql_prof.id as profile_id, graphql_prof.*
  FROM           PLC.PL_CONTENT_PROFILES graphql_prof
       LEFT JOIN PLC.PL_OBJECT_CONTENT_PROFILES obj_graphql_prof
              ON graphql_prof.id = obj_graphql_prof.profile_id
 WHERE graphql_prof.policy_id = ?
   AND graphql_prof.content_profile_type = ?
EndQuery

    for my $graphql_profile (values %{$self->{h_graphql_profiles}}) {
        my $idl_file_result = F5::GraphqlProfile::handle_idl_files(dbh => $dbh,
                                                                   logger => $self->{logger},
                                                                   profile_id => $graphql_profile->{profile_id},
                                                                  );

        if ($idl_file_result && ! $idl_file_result->{completed_successfully}) {
             my $error_msg = 'Failed to validate GraphQL schema ';
             if ($idl_file_result->{messages}) {
                 $error_msg .= join('; ', @{$idl_file_result->{messages}});
             }
             $self->{logger}->error($error_msg) if $self->{logger};
        }
        else {
             $graphql_profile->{graphql_schemas_md5} = $idl_file_result->{graphql_md5};
        }

        $graphql_profile->{schema_filename} = '';
        $graphql_profile->{defense_config_filename} = '';
        $graphql_profile->{json_schemas_md5} ||= '';
        $graphql_profile->{graphql_schemas_md5} ||= '';
        $graphql_profile->{request_xpath_filename} = '';
        $graphql_profile->{response_xpath_filename} = '';
        $graphql_profile->{content_type}= 'graphql';
        $graphql_profile->{flg_is_graphql_security} = 1;


        $graphql_profile->{all_disallowed_response_patterns} = '';

        if ($graphql_profile->{block_disallowed_response_patterns}) {
            my $disallowed_response_patterns = $dbh->selectcol_arrayref(<<EndQuery, undef, $graphql_profile->{profile_id});
SELECT pattern
  FROM PLC.PL_CONTENT_PROFILE_DISALLOWED_RESPONSE_PATTERNS
 WHERE profile_id = ?
EndQuery
            $graphql_profile->{all_disallowed_response_patterns} = join('|', @$disallowed_response_patterns);
        }

        if ($graphql_profile->{flg_check_element_metachars}) {
            if (! ref($self->{base_charsets}) || ! scalar( keys %{$self->{base_charsets}})) {
                $self->preload_charset_templates($policy_name_crc, $policy_id);
            }

            $graphql_profile->{charset_templ_id} = $self->get_charset_template(
                                                          policy_name_crc => $policy_name_crc,
                                                          base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_JSON_PROFILE(),
                                                          override_table => 'PLC.PL_CONTENT_PROFILE_VALUE_METACHARS',
                                                          override_id_field => 'profile_id',
                                                          override_id_value => $graphql_profile->{profile_id});
        }

        my $res = $dbh->selectall_hashref(<<EndQuery, 'def_attrib_id',undef, $graphql_profile->{profile_id});
SELECT def_attrib_id,def_attrib_value, flg_check_attrib
  FROM PLC.PL_GRAPHQL_PROFILE_DEFENSE_ATTRIBUTES
 WHERE profile_id = ?
EndQuery

        my %graphql_defs = (
                          1 => { def_attrib => 'GRAPHQL_DEFENSE_MAX_TOTAL_LENGTH',
                                 flg_check_attrib_field => 'flg_check_document_length',
                                 attrib_value_field => 'max_document_length',
                               },
                          2 => { def_attrib => 'GRAPHQL_DEFENSE_MAX_VALUE_LENGTH',
                                 flg_check_attrib_field => 'flg_check_param_value_length',
                                 attrib_value_field => 'max_param_value_length',
                               },
                          3 => { def_attrib => 'GRAPHQL_DEFENSE_MAX_DEPTH',
                                 flg_check_attrib_field => 'flg_check_max_depth',
                                 attrib_value_field => 'max_depth',
                               },
                          4 => { def_attrib => 'GRAPHQL_DEFENSE_MAX_BATCHED_QUERY',
                                 flg_check_attrib_field => 'flg_check_max_batched_queries',
                                 attrib_value_field => 'max_batched_queries',
                               },
                          5 => { def_attrib => 'GRAPHQL_DEFENSE_TOLERATE_GRAPHQL_PARSING_WARNINGS',
                                 attrib_value_field => 'flg_tolerate',
                               },
                          6 => { def_attrib => 'GRAPHQL_DEFENSE_ALLOW_INTROSPECTION_QUERY',
                                 attrib_value_field => 'flg_allow_introspection_query',
                               },
                          7 => { def_attrib => 'GRAPHQL_DEFENSE_MAX_QUERY_COST',
                                 flg_check_attrib_field => 'flg_check_max_query_cost',
                                 attrib_value_field => 'max_query_cost',
                               },
                        );

        for my $def_attrib_id (keys %graphql_defs) {
            my $graphql_def = $graphql_defs{$def_attrib_id};
            my $plc_rec = $res->{$def_attrib_id};

            if (exists $graphql_def->{flg_check_attrib_field}) {
                $graphql_profile->{$graphql_def->{flg_check_attrib_field}} = $plc_rec->{flg_check_attrib};
                $graphql_profile->{$graphql_def->{attrib_value_field}} = $plc_rec->{def_attrib_value} || 0;
            }
            else {
               $graphql_profile->{$graphql_def->{attrib_value_field}} = ( $plc_rec->{flg_check_attrib} && $plc_rec->{def_attrib_value});
            }
        }

        my $prev_schema_filename;
        my($existing_graphql_template) = grep(find_matching_hash($_, $graphql_profile, \@compare_fields), @$existing_graphql_templates);
        $graphql_profile->{content_type}='graphql';
        if ($existing_graphql_template) {
            $graphql_profile->{profile_templ_id}->{$graphql_profile->{content_type}} = $existing_graphql_template->{profile_templ_id};
            $prev_schema_filename = $existing_graphql_template->{graphql_schemas_md5};
        }
        else {
            $prev_schema_filename = $graphql_profile->{graphql_schemas_md5};
            $graphql_profile->{policy_name_crc} = $policy_name_crc;
            if ($self->{ucs_upgrade_real_policy_id}) {
                $self->ucs_upgrade_profile_id_fix(profile => $graphql_profile);
            }
            $self->{max_profile_templ_id}++;
            $self->insert_row(
                             table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                             rec => { %$graphql_profile, profile_templ_id => $self->{max_profile_templ_id} },
                             set_date_fields_to_now => [ 'modify_time' ],
                            );
            $graphql_profile->{profile_templ_id}->{$graphql_profile->{content_type}} = $self->{max_profile_templ_id};

       }

       my $profile_templ_id = $graphql_profile->{profile_templ_id}->{$graphql_profile->{content_type}};

       if ($idl_file_result && $idl_file_result->{completed_successfully}) {
           if (!$prev_schema_filename || ( $prev_schema_filename ne $idl_file_result->{graphql_md5}) ) {
               $self->update_row(table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                 rec => {
                                     graphql_schemas_md5 => $idl_file_result->{graphql_md5} || '',
                                 },
                                 where_fields => {
                                     profile_templ_id => $profile_templ_id,
                                     policy_name_crc => $policy_name_crc,
                                 },
                   );

           }
           elsif ($existing_graphql_template) {
               next;
           }

           $self->empty_table("DCC.CONTENT_PROFILE_FIELDS", undef, {"profile_templ_id" => $profile_templ_id, "policy_name_crc" => $policy_name_crc});

           for my $idl_file (@{$idl_file_result->{profile_idl_files}}) {
               next if !defined $idl_file->{parsed_filename};
               my $fields = from_json(File::Slurp::read_file($idl_file->{parsed_filename}));
               for my $field (keys %$fields) {
                   my $field_type = $fields->{$field}->{type};
                   my $cost = defined $fields->{$field}->{cost} ? $fields->{$field}->{cost} : $cost_by_field_type{$field_type};
                   my $rec = { profile_templ_id => $profile_templ_id,
                               policy_name_crc => $policy_name_crc,
                               field_name_crc => F5::CRC::get_crc_any('get_crc32', $field),
                               field_name => $field,
                               field_type => $field_type,
                               cost => $cost, };
                   $self->insert_row(table_name => 'DCC.CONTENT_PROFILE_FIELDS',
                                     rec => $rec,
                                    );
               }
           }
       }

       $self->empty_table("DCC.CONTENT_PROFILE_SENSITIVE_PARAMS", undef, {"profile_templ_id" => $profile_templ_id,
                                                                          policy_name_crc => $policy_name_crc });
       my $sensitive_params = $dbh->selectcol_arrayref(<<EndQuery, undef, $graphql_profile->{profile_id});
SELECT sensitive_param_name
  FROM PLC.PL_CONTENT_SENSITIVE_PARAMS
 WHERE profile_id = ?
EndQuery

       for my $param (@$sensitive_params){
           $self->insert_row(table_name => 'DCC.CONTENT_PROFILE_SENSITIVE_PARAMS',
                             rec => { profile_templ_id => $profile_templ_id,
                                      policy_name_crc => $policy_name_crc,
                                      param_name => $param },
                            );
       }
    }

   # Mark unused GraphQL templates for deletion
    my %newly_used_graphql_templates = map { ($_ => 1) } map { values %{$_->{profile_templ_id}} } values %{$self->{h_graphql_profiles}};
    my @unused_graphql_templates = grep { !$newly_used_graphql_templates{$_->{profile_templ_id}} } @$existing_graphql_templates;
    if (@unused_graphql_templates) {
        my @placeholders = ('?') x scalar @unused_graphql_templates;
        $dbh->do("DELETE FROM DCC.CONTENT_PROFILE_TEMPLATES WHERE profile_templ_id IN (@placeholders)",
                  undef, map { $_->{profile_templ_id} } @unused_graphql_templates);
    }

   my @graphql_entities = (
                      {
                       table => 'PLC.PL_OBJECT_CONTENT_PROFILES',
                       fields => [ qw(A.id A.profile_id A.obj_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GRAPHQL(),
                       save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_objects}->{$plc_row->{obj_id}}->{profile_templ_id} = $profile_templ_id;
                       }
                      },
                     );

    for my $graphql_entity (@graphql_entities) {
        my @plc_fields = @{$graphql_entity->{fields}};
        my $profile_id_field = $graphql_entity->{profile_id_field} || 'profile_id';
        my $get_graphql_profile_entities_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields
  FROM $graphql_entity->{table} A, PLC.PL_CONTENT_PROFILES B
 WHERE A.policy_id = ? AND A.$profile_id_field = B.id AND B.content_profile_type=?
EndQuery
        $get_graphql_profile_entities_sth->execute($policy_id, $graphql_entity->{profile_type});
        while(my $plc_row = $get_graphql_profile_entities_sth->fetchrow_hashref) {
            $plc_row->{content_type}='graphql';
            next if not exists $self->{h_graphql_profiles}->{$plc_row->{$profile_id_field}};
            my $profile_templ_id = $self->{h_graphql_profiles}->{$plc_row->{$profile_id_field}}->{profile_templ_id}->{$plc_row->{content_type}};
            $graphql_entity->{save_profile_templ_id}->($self, $plc_row, $profile_templ_id);
        }
    }


    return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_gwt_profiles {
  my($self, $policy_name_crc, $policy_id, $params) = @_;
  my $dbh = $self->{dbh};

  local $" = ',';

     if ($params->{readonly_mode}) {
       my $existing_gwt_profiles_sth = $dbh->prepare(<<EndQuery);
SELECT profile_templ_id, profile_id
  FROM DCC.CONTENT_PROFILE_TEMPLATES
   WHERE policy_name_crc = ? AND flg_is_gwt_security = ?
EndQuery

   $existing_gwt_profiles_sth->execute($policy_name_crc, 1);


     while (my $existing_gwt_profile = $existing_gwt_profiles_sth->fetchrow_hashref) {
      $existing_gwt_profile->{content_type} = 'gwt';
      $self->{h_gwt_profiles}->{$existing_gwt_profile->{profile_id}}->{profile_templ_id}->{$existing_gwt_profile->{content_type}} = $existing_gwt_profile->{profile_templ_id};
     }
   return 1;
   }

   if ( !defined($self->{max_profile_templ_id}) ) {
       $self->load_max_profile_templ_id(policy_name_crc => $policy_name_crc);
   }

   my @compare_fields = qw(profile_id
                           charset_templ_id
                           flg_is_gwt_security
                           flg_check_param_value_length
                           max_param_value_length
                           flg_check_document_length
                           max_document_length
                           flg_check_element_metachars
                           flg_check_signatures
                           flg_tolerate
                           profile_name
                        );

   my $existing_gwt_templates = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, 1);
SELECT profile_templ_id, @compare_fields
  FROM DCC.CONTENT_PROFILE_TEMPLATES
  WHERE policy_name_crc = ? AND flg_is_gwt_security = ?
EndQuery

 $self->{h_gwt_profiles} = $dbh->selectall_hashref(<<EndQuery, 'profile_id', undef, $policy_id, F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GWT());
SELECT gwt_prof.id as profile_id, gwt_prof.*
  FROM PLC.PL_CONTENT_PROFILES gwt_prof
         JOIN PLC.PL_OBJECT_CONTENT_PROFILES obj_gwt_prof
         ON gwt_prof.id = obj_gwt_prof.profile_id
 WHERE gwt_prof.policy_id = ?
   AND gwt_prof.content_profile_type = ?
EndQuery

   for my $gwt_profile (values %{$self->{h_gwt_profiles}}) {
        $gwt_profile->{flg_is_gwt_security} = 1;
        $gwt_profile->{charset_templ_id} ||= 0 ;

        my $res = $dbh->selectall_hashref(<<EndQuery, 'def_attrib_id',undef, $gwt_profile->{profile_id});
SELECT def_attrib_id,def_attrib_value, flg_check_attrib
  FROM PLC.PL_GWT_PROFILE_DEFENSE_ATTRIBUTES
 WHERE  profile_id = ?
EndQuery

        my %gwt_defs = (
                          1 => { def_attrib => 'GWT_DEFENSE_MAX_TOTAL_LENGTH',
                                 flg_check_attrib_field => 'flg_check_document_length',
                                 attrib_value_field => 'max_document_length',
                               },
                          2 => { def_attrib => 'GWT_DEFENSE_MAX_VALUE_LENGTH',
                                 flg_check_attrib_field => 'flg_check_param_value_length',
                                 attrib_value_field => 'max_param_value_length',
                                },
                          3 => { def_attrib => 'GWT_DEFENSE_TOLERATE_GWT_PARSING_WARNINGS',
                                 attrib_value_field => 'flg_tolerate',
                               },
                        );


        for my $def_attrib_id (keys %gwt_defs) {
            my $gwt_def = $gwt_defs{$def_attrib_id};
            my $plc_rec = $res->{$def_attrib_id};

            if (exists $gwt_def->{flg_check_attrib_field}) {
                $gwt_profile->{$gwt_def->{flg_check_attrib_field}} = $plc_rec->{flg_check_attrib};
                $gwt_profile->{$gwt_def->{attrib_value_field}} = $plc_rec->{def_attrib_value};
            }
            else {
               $gwt_profile->{$gwt_def->{attrib_value_field}} = ( $plc_rec->{flg_check_attrib} && $plc_rec->{def_attrib_value});
            }
        }


        if ( $gwt_profile->{flg_check_element_metachars}) {
            $gwt_profile->{charset_templ_id} = $self->get_charset_template(
                                                          policy_name_crc => $policy_name_crc,
                                                          base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_GWT_PROFILE(),
                                                          override_table => 'PLC.PL_CONTENT_PROFILE_VALUE_METACHARS',
                                                          override_id_field => 'profile_id',
                                                          override_id_value => $gwt_profile->{profile_id});

        }

        my ($existing_gwt_template) = grep(find_matching_hash($_, $gwt_profile, \@compare_fields), @$existing_gwt_templates);
        $gwt_profile->{content_type} = 'gwt';
        if ($existing_gwt_template) {
           $gwt_profile->{profile_templ_id}->{$gwt_profile->{content_type}} = $existing_gwt_template->{profile_templ_id};
        }
        else {
           $gwt_profile->{policy_name_crc} = $policy_name_crc;

           if ($self->{ucs_upgrade_real_policy_id}) {
               $self->ucs_upgrade_profile_id_fix(profile => $gwt_profile);
           }
           $self->{max_profile_templ_id}++;
           $self->insert_row(
                             table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                             rec => { %$gwt_profile, profile_templ_id => $self->{max_profile_templ_id} },
                             set_date_fields_to_now => [ 'modify_time' ],
                            );
           $gwt_profile->{profile_templ_id}->{$gwt_profile->{content_type}} = $self->{max_profile_templ_id};
       }
   }

   # Mark unused GWT templates for deletion
   my %newly_used_gwt_templates = map { ($_ => 1) } map { values %{$_->{profile_templ_id}} } values %{$self->{h_gwt_profiles}};
   my @unused_gwt_templates = grep { !$newly_used_gwt_templates{$_->{profile_templ_id}} } @$existing_gwt_templates;
   if (@unused_gwt_templates) {
     my @placeholders = ('?') x scalar @unused_gwt_templates;
     $dbh->do("DELETE FROM DCC.CONTENT_PROFILE_TEMPLATES WHERE profile_templ_id IN (@placeholders) AND policy_name_crc = ?",
             undef, (map { $_->{profile_templ_id} } @unused_gwt_templates), $policy_name_crc);
   }


   my @gwt_entities = (
                      {
                       table => 'PLC.PL_OBJECT_CONTENT_PROFILES',
                       fields => [ qw(A.id A.profile_id A.obj_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GWT(),
                      save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_objects}->{$plc_row->{obj_id}}->{profile_templ_id} = $profile_templ_id;
                       }
                      },
                      {
                       table => 'PLC.PL_PARAM_CONTENT_PROFILES',
                       fields => [ qw(A.id A.profile_id A.param_attrib_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_GWT(),
                       save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_keys}->{$plc_row->{param_attrib_id}}->{profile_templ_id} = $profile_templ_id;
                       }
                      },
                     );
  for my $gwt_entity (@gwt_entities) {
    my @plc_fields = @{$gwt_entity->{fields}};
    my $get_gwt_profile_entities_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields
  FROM $gwt_entity->{table} A, PLC.PL_CONTENT_PROFILES B
 WHERE A.policy_id = ? AND A.profile_id=B.id AND B.content_profile_type= ?
EndQuery
    $get_gwt_profile_entities_sth->execute($policy_id, $gwt_entity->{profile_type});
    while(my $plc_row = $get_gwt_profile_entities_sth->fetchrow_hashref) {
      $plc_row->{content_type}='gwt';
      next if not exists $self->{h_gwt_profiles}->{$plc_row->{profile_id}};
      my $profile_templ_id = $self->{h_gwt_profiles}->{$plc_row->{profile_id}}->{profile_templ_id}->{$plc_row->{content_type}};
      $gwt_entity->{save_profile_templ_id}->($self, $plc_row, $profile_templ_id);
    }
  }

  return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_plain_text_profiles {
    my($self, $policy_name_crc, $policy_id, $params) = @_;
    my $dbh = $self->{dbh};

    local $" = ',';

    if ($params->{readonly_mode}) {
        my $existing_plain_text_profiles_sth = $dbh->prepare(<<EndQuery);
SELECT profile_templ_id, profile_id
  FROM DCC.CONTENT_PROFILE_TEMPLATES
   WHERE policy_name_crc = ? AND flg_is_plain_text_security = ?
EndQuery

        $existing_plain_text_profiles_sth->execute($policy_name_crc, 1);
        while (my $existing_plain_text_profile = $existing_plain_text_profiles_sth->fetchrow_hashref) {
            $existing_plain_text_profile->{content_type} = 'plain_text';
            $self->{h_plain_text_profiles}->{$existing_plain_text_profile->{profile_id}}->{profile_templ_id}->{$existing_plain_text_profile->{content_type}} = $existing_plain_text_profile->{profile_templ_id};
        }
        return 1;
   }

   if ( !defined($self->{max_profile_templ_id}) ) {
       $self->load_max_profile_templ_id(policy_name_crc => $policy_name_crc);
   }

   my @compare_fields = qw(profile_id
                           charset_templ_id
                           flg_check_param_value_length
                           max_param_value_length
                           flg_check_document_length
                           max_document_length
                           flg_check_element_metachars
                           flg_check_signatures
                           flg_tolerate
                           profile_name
                           flg_is_plain_text_security
                           flg_perform_percent_decoding
                        );

   my $existing_plain_text_templates = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, 1);
SELECT profile_templ_id, @compare_fields
  FROM DCC.CONTENT_PROFILE_TEMPLATES
  WHERE policy_name_crc = ? AND flg_is_plain_text_security = ?
EndQuery

   $self->{h_plain_text_profiles} = $dbh->selectall_hashref(<<EndQuery, 'profile_id', undef, $policy_id, F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_PLAIN_TEXT(), 1);
SELECT plain_text_prof.id as profile_id, plain_text_prof.*
  FROM PLC.PL_CONTENT_PROFILES plain_text_prof, PLC.PL_WEBSOCKET_URLS websocket_url_plain_text_prof
  WHERE plain_text_prof.policy_id = ?
    AND websocket_url_plain_text_prof.websocket_text_profile_id = plain_text_prof.id
    AND plain_text_prof.content_profile_type = ?
    AND websocket_url_plain_text_prof.flg_websocket_message_format_text = ?;
EndQuery

    for my $plain_text_profile (values %{$self->{h_plain_text_profiles}}) {
        $plain_text_profile->{defense_config_filename} = '';
        $plain_text_profile->{schema_filename} = '';
        $plain_text_profile->{request_xpath_filename} = '';
        $plain_text_profile->{json_schemas_md5} = '';
        $plain_text_profile->{graphql_schemas_md5} = '';
        $plain_text_profile->{response_xpath_filename} = '';

        $plain_text_profile->{flg_is_wss_security} = 0;
        $plain_text_profile->{flg_is_xml_security} = 0;
        $plain_text_profile->{flg_is_json_security} = 0;
        $plain_text_profile->{flg_is_grpc_security} = 0;
        $plain_text_profile->{flg_is_soap_traffic} = 0;
        $plain_text_profile->{flg_is_gwt_security} = 0;
        $plain_text_profile->{charset_templ_id} ||= 0 ;
        $plain_text_profile->{flg_is_plain_text_security} = 1;

        my $res = $dbh->selectall_hashref(<<EndQuery, 'def_attrib_id',undef, $plain_text_profile->{profile_id});
SELECT def_attrib_id,def_attrib_value, flg_check_attrib
  FROM PLC.PL_PLAIN_TEXT_PROFILE_DEFENSE_ATTRIBUTES
 WHERE  profile_id = ?
EndQuery

        my %plain_text_defs = (
                          1 => { def_attrib => 'PLAIN_TEXT_DEFENSE_MAX_TOTAL_LENGTH',
                                 flg_check_attrib_field => 'flg_check_document_length',
                                 attrib_value_field => 'max_document_length',
                               },
                          2 => { def_attrib => 'PLAIN_TEXT_DEFENSE_MAX_LINE_LENGTH',
                                 flg_check_attrib_field => 'flg_check_param_value_length',
                                 attrib_value_field => 'max_param_value_length',
                                },
                          3 => { def_attrib => 'PLAIN_TEXT_DEFENSE_PERCENT_DECODING',
                                 attrib_value_field => 'flg_perform_percent_decoding',
                               },
                        );

        for my $def_attrib_id (keys %plain_text_defs) {
            my $plain_text_def = $plain_text_defs{$def_attrib_id};
            my $plc_rec = $res->{$def_attrib_id};

            if (exists $plain_text_def->{flg_check_attrib_field}) {
                $plain_text_profile->{$plain_text_def->{flg_check_attrib_field}} = $plc_rec->{flg_check_attrib};
                $plain_text_profile->{$plain_text_def->{attrib_value_field}} = $plc_rec->{def_attrib_value};
            }
            else {
                $plain_text_profile->{$plain_text_def->{attrib_value_field}} = ( $plc_rec->{flg_check_attrib} && $plc_rec->{def_attrib_value});
            }

            $plain_text_profile->{$plain_text_def->{attrib_value_field}} = ( $plc_rec->{flg_check_attrib} && $plc_rec->{def_attrib_value});
        }

        if ($plain_text_profile->{flg_check_element_metachars}) {
            $plain_text_profile->{charset_templ_id} = $self->get_charset_template(
                                                          policy_name_crc => $policy_name_crc,
                                                          base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_PLAIN_TEXT_PROFILE(),
                                                          override_table => 'PLC.PL_CONTENT_PROFILE_VALUE_METACHARS',
                                                          override_id_field => 'profile_id',
                                                          override_id_value => $plain_text_profile->{profile_id});
        }

        my($existing_plain_text_template) = grep(find_matching_hash($_, $plain_text_profile, \@compare_fields), @$existing_plain_text_templates);
        $plain_text_profile->{content_type}='plain_text';
        if ($existing_plain_text_template) {
           $plain_text_profile->{profile_templ_id}->{$plain_text_profile->{content_type}} = $existing_plain_text_template->{profile_templ_id};
        }
        else {
            $plain_text_profile->{policy_name_crc} = $policy_name_crc;

            if ($self->{ucs_upgrade_real_policy_id}) {
                $self->ucs_upgrade_profile_id_fix(profile => $plain_text_profile);
            }
            $self->{max_profile_templ_id}++;
            $self->insert_row(
                            table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                            rec => { %$plain_text_profile, profile_templ_id => $self->{max_profile_templ_id} },
                            set_date_fields_to_now => [ 'modify_time' ],
                            );
            $plain_text_profile->{profile_templ_id}->{$plain_text_profile->{content_type}} = $self->{max_profile_templ_id};
        }
    }


    my %newly_used_plain_text_templates = map { ($_ => 1) } map { values %{$_->{profile_templ_id}} } values %{$self->{h_plain_text_profiles}};
    my @unused_plain_text_templates = grep { !$newly_used_plain_text_templates{$_->{profile_templ_id}} } @$existing_plain_text_templates;
    if (@unused_plain_text_templates) {
        my @placeholders = ('?') x scalar @unused_plain_text_templates;
        $dbh->do("DELETE FROM DCC.CONTENT_PROFILE_TEMPLATES WHERE profile_templ_id IN (@placeholders) AND policy_name_crc = ?",
                undef, (map { $_->{profile_templ_id} } @unused_plain_text_templates), $policy_name_crc);
    }

    my @plain_text_entities = (
                        {
                            table => 'PLC.PL_WEBSOCKET_URLS',
                            fields => [ qw(A.websocket_text_profile_id A.websocket_url_id) ],
                            profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_PLAIN_TEXT(),
                            profile_id_field => 'websocket_text_profile_id',
                            save_profile_templ_id => sub {
                                my($self, $plc_row, $profile_templ_id) = @_;
                                $self->{h_websocket_urls}->{$plc_row->{websocket_url_id}}->{websocket_text_profile_templ_id} = $profile_templ_id;
                            }
                        },
                        {
                            table => 'PLC.PL_PARAM_CONTENT_PROFILES',
                            fields => [ qw(A.id A.profile_id A.param_attrib_id) ],
                            profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_PLAIN_TEXT(),
                            save_profile_templ_id => sub {
                                my($self, $plc_row, $profile_templ_id) = @_;
                                $self->{h_keys}->{$plc_row->{param_attrib_id}}->{profile_templ_id} = $profile_templ_id;
                            }
                        },
                     );
    for my $plain_text_entity (@plain_text_entities) {
        my @plc_fields = @{$plain_text_entity->{fields}};
        my $profile_id_field = $plain_text_entity->{profile_id_field} || 'profile_id';
        my $get_plain_text_profile_entities_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields
  FROM $plain_text_entity->{table} A, PLC.PL_CONTENT_PROFILES B
 WHERE A.policy_id = ? AND A.$profile_id_field = B.id AND B.content_profile_type= ?
EndQuery
        $get_plain_text_profile_entities_sth->execute($policy_id, $plain_text_entity->{profile_type});
        while(my $plc_row = $get_plain_text_profile_entities_sth->fetchrow_hashref) {
            $plc_row->{content_type} = 'plain_text';
            next if not exists $self->{h_plain_text_profiles}->{$plc_row->{$profile_id_field}};
            my $profile_templ_id = $self->{h_plain_text_profiles}->{$plc_row->{$profile_id_field}}->{profile_templ_id}->{$plc_row->{content_type}};
            $plain_text_entity->{save_profile_templ_id}->($self, $plc_row, $profile_templ_id);
        }
    }

    return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_xml_profiles {
  my($self, $policy_name_crc, $policy_id, $params) = @_;

  require F5::XMLProfile;
  my $dbh = $self->{dbh};
  local $" = ',';

  if ($params->{readonly_mode}) {
    my $existing_xml_profiles_sth = $dbh->prepare(<<EndQuery);
SELECT profile_templ_id, profile_id
  FROM DCC.CONTENT_PROFILE_TEMPLATES
 WHERE policy_name_crc = ? AND flg_is_xml_security = ?
EndQuery
    $existing_xml_profiles_sth->execute($policy_name_crc,1);

    while (my $existing_xml_profile = $existing_xml_profiles_sth->fetchrow_hashref) {
      $self->{h_xml_profiles}->{$existing_xml_profile->{profile_id}}->{profile_templ_id}->{xml} = $existing_xml_profile->{profile_templ_id};
      
    }
    return 1;
  }

  $self->{logger}->debug('Preparing XML profiles') if $self->{logger};

  if ( !defined($self->{max_profile_templ_id}) ) {
      $self->load_max_profile_templ_id(policy_name_crc => $policy_name_crc);
  }

  my @compare_fields = qw(profile_id 
                          defense_config_filename 
                          schema_filename 
                          request_xpath_filename 
                          response_xpath_filename
                          flg_validation_soapaction_header
                          flg_allow_soap_attachment
                          flg_is_wss_security
                          flg_is_xml_security
                          flg_is_soap_traffic
                          charset_templ_id
                          flg_check_param_value_length
                          max_param_value_length
                          flg_use_xml_blocking_page
                          flg_check_element_metachars
                          flg_check_attribute_metachars
                          flg_check_signatures
                          flg_inspect_soap_attachments
                          profile_name
                        );

  my $existing_xml_templates = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc,1);
SELECT profile_templ_id, @compare_fields
  FROM DCC.CONTENT_PROFILE_TEMPLATES
 WHERE policy_name_crc = ?  AND flg_is_xml_security = ?
EndQuery

  local $" = ',';

  my @profiles_in_use;
  for my $profile_link_table (qw(PLC.PL_OBJECT_CONTENT_PROFILES PLC.PL_PARAM_CONTENT_PROFILES)) {

      my $profile_ids = $dbh->selectcol_arrayref(<<EndQuery, undef, $policy_id);
SELECT profile_id
  FROM $profile_link_table
 WHERE policy_id = ?
EndQuery

      if ($profile_ids && @$profile_ids) {
          push(@profiles_in_use, @$profile_ids);
      }
  }

  $self->{h_xml_profiles} = {};

  if (@profiles_in_use) {
      my @placeholders = ('?') x scalar @profiles_in_use;

      $self->{h_xml_profiles} = $dbh->selectall_hashref(<<EndQuery, 'profile_id', undef, $policy_id, F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(), @profiles_in_use);
SELECT id as profile_id, *
  FROM PLC.PL_CONTENT_PROFILES
 WHERE policy_id = ?
   AND content_profile_type = ?
   AND id IN (@placeholders)
EndQuery
  }

  for my $xml_profile (values %{$self->{h_xml_profiles}}) {

    $xml_profile->{schema_filename} = '';
    $xml_profile->{request_xpath_filename} = '';
    $xml_profile->{response_xpath_filename} = '';

    $xml_profile->{flg_is_xml_security} = 1 ;

    if ($xml_profile->{flg_check_element_metachars} || $xml_profile->{flg_check_attribute_metachars}){
         $xml_profile->{charset_templ_id} = $self->get_charset_template(
                                                          policy_name_crc => $policy_name_crc,
                                                          base_type => F5::ASMConfig::CharsetType::CHARSET_TYPE_XML_PROFILE(),
                                                          override_table => 'PLC.PL_CONTENT_PROFILE_VALUE_METACHARS',
                                                          override_id_field => 'profile_id',
                                                          override_id_value => $xml_profile->{profile_id});
    }else{
         $xml_profile->{charset_templ_id} = 0;
    }


    ($xml_profile->{flg_is_soap_traffic}) = $dbh->selectrow_array(<<EndQuery, undef, F5::ASMConfig::XmlValidationFileType::XML_VALIDATION_FILE_TYPE_WSDL(), $xml_profile->{profile_id});
SELECT count(*) 
  FROM PLC.PL_XML_PROFILE_VALIDATION_FILES profile_files
  JOIN PLC.PL_XML_VALIDATION_FILES validation_files ON profile_files.xml_validation_file_id = validation_files.xml_validation_file_id
 WHERE file_type = ?
   AND profile_id = ?
EndQuery

    my $validation_file_result = F5::XMLProfile::handle_validation_files(dbh => $dbh,
                                                                         policy_name_crc => $policy_name_crc,
                                                                         profile_rest_uuid => $xml_profile->{rest_uuid},
                                                                         profile_id => $xml_profile->{profile_id},
                                                                         );

    if ($validation_file_result && ! $validation_file_result->{completed_successfully}) {
        $self->{logger}->error( join('; ', @{$validation_file_result->{messages}}) ) if $self->{logger};
    }

    if ($validation_file_result && $validation_file_result->{compiled_schema_filename}) {
      $xml_profile->{schema_filename} = copy_policy_file_for_account(policy_file => $validation_file_result->{compiled_schema_filename},
                                                                     account_template => $ACCOUNT_DATA_DIR . '/xml/compiled_schema_%MD5%');
    }

   if($xml_profile->{flg_is_wss_security}){
        my $xpath_result =  F5::XMLProfile::create_xpath_files(dbh => $dbh,
                                                               policy_name_crc => $policy_name_crc,
                                                               profile_rest_uuid => $xml_profile->{rest_uuid},
                                                               profile_id => $xml_profile->{profile_id});

       if($xpath_result && $xpath_result->{request_xpath_filename}){
           $xml_profile->{request_xpath_filename} = copy_policy_file_for_account(policy_file => $xpath_result->{request_xpath_filename},
                                                                                 account_template => $ACCOUNT_DATA_DIR . '/xml/request_xpath__%MD5%');
       }

       if($xpath_result && $xpath_result->{response_xpath_filename}){
           $xml_profile->{response_xpath_filename} = copy_policy_file_for_account(policy_file => $xpath_result->{response_xpath_filename},
                                                                                  account_template => $ACCOUNT_DATA_DIR . '/xml/response_xpath__%MD5%');
       }

    }

        my $def_attrib = $dbh->selectrow_hashref(<<EndQuery,undef, F5::XMLProfile::XML_DEFENSE_MAX_VALUE_LENGTH(), $xml_profile->{profile_id});
SELECT def_attrib_value, flg_check_attrib
  FROM PLC.PL_XML_PROFILE_DEFENSE_ATTRIBUTES
 WHERE def_attrib_id = ? AND profile_id = ?
EndQuery


        $xml_profile->{max_param_value_length} = $def_attrib->{def_attrib_value};
        $xml_profile->{flg_check_param_value_length} = $def_attrib->{flg_check_attrib};


    $xml_profile->{defense_config_filename} = '';
    my $defense_config_result = F5::XMLProfile::create_defense_config(dbh => $dbh,
                                                                      policy_name_crc => $policy_name_crc,
                                                                      profile_rest_uuid => $xml_profile->{rest_uuid},
                                                                      profile_id => $xml_profile->{profile_id});

    $self->{logger}->debug("Defense config file for profile_id $xml_profile->{profile_id} is $defense_config_result->{defense_config_filename}") if $self->{logger};
    if ($defense_config_result && $defense_config_result->{defense_config_filename}) {
      $xml_profile->{defense_config_filename} = copy_policy_file_for_account(policy_file => $defense_config_result->{defense_config_filename},
                                                                             account_template => $ACCOUNT_DATA_DIR . '/xml/defense_config_%MD5%');
    }

    
    my($existing_xml_template) = grep(find_matching_hash($_, $xml_profile, \@compare_fields), @$existing_xml_templates);
    if ($existing_xml_template) {
 	$xml_profile->{profile_templ_id}->{xml} = $existing_xml_template->{profile_templ_id};
    }
    else {
      $xml_profile->{policy_name_crc} = $policy_name_crc;
      if ($self->{ucs_upgrade_real_policy_id}) {
          $self->ucs_upgrade_profile_id_fix(profile => $xml_profile);
      }
      $self->{max_profile_templ_id}++;
      $self->insert_row(
                        table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                        rec => { %$xml_profile, profile_templ_id => $self->{max_profile_templ_id} },
                        set_date_fields_to_now => [ 'modify_time' ],
                       );
      $xml_profile->{profile_templ_id}->{xml} = $self->{max_profile_templ_id};
    }
  }

  # Mark unused XML templates for deletion
  my %newly_used_xml_templates = map { ($_ => 1) } map { values %{$_->{profile_templ_id}} } values %{$self->{h_xml_profiles}};
  my @unused_xml_templates = grep { !$newly_used_xml_templates{$_->{profile_templ_id}} } @$existing_xml_templates;
  if (@unused_xml_templates) {
    my @placeholders = ('?') x scalar @unused_xml_templates;
    $dbh->do("DELETE FROM DCC.CONTENT_PROFILE_TEMPLATES WHERE profile_templ_id IN (@placeholders) AND policy_name_crc = ?",
             undef, (map { $_->{profile_templ_id} } @unused_xml_templates), $policy_name_crc);
    $dbh->do("DELETE FROM DCC.XML_SOAP_METHODS WHERE profile_templ_id IN (@placeholders) AND policy_name_crc = ?",
             undef, (map { $_->{profile_templ_id} } @unused_xml_templates), $policy_name_crc);
  }

  my @xml_entities = (
                      {
                       table => 'PLC.PL_OBJECT_CONTENT_PROFILES',
                       fields => [ qw(A.id A.profile_id A.obj_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(),
                       save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_objects}->{$plc_row->{obj_id}}->{profile_templ_id} = $profile_templ_id;
                       }
                      },
                      {
                       table => 'PLC.PL_PARAM_CONTENT_PROFILES',
                       fields => [ qw(A.id A.profile_id A.param_attrib_id) ],
                       profile_type => F5::ASMConfig::ContentProfileType::CONTENT_PROFILE_TYPE_XML(),
                       save_profile_templ_id => sub {
                         my($self, $plc_row, $profile_templ_id) = @_;
                         $self->{h_keys}->{$plc_row->{param_attrib_id}}->{profile_templ_id} = $profile_templ_id;
                       }
                      },
                     );

  for my $xml_entity (@xml_entities) {

    my @plc_fields = @{$xml_entity->{fields}};
    my $get_xml_profile_entities_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields
  FROM $xml_entity->{table} A, PLC.PL_CONTENT_PROFILES B
 WHERE A.policy_id = ? AND A.profile_id=B.id AND B.content_profile_type= ?
EndQuery

    $get_xml_profile_entities_sth->execute($policy_id, $xml_entity->{profile_type});

    while(my $plc_row = $get_xml_profile_entities_sth->fetchrow_hashref) {
      next if not exists $self->{h_xml_profiles}->{$plc_row->{profile_id}};
      my $profile_templ_id = $self->{h_xml_profiles}->{$plc_row->{profile_id}}->{profile_templ_id}->{xml};
      $xml_entity->{save_profile_templ_id}->($self, $plc_row, $profile_templ_id);
    }
  }
  $self->prepare_soap_method_data($policy_name_crc, $policy_id);

   F5::DbUtils::delete_unreferenced_files(dbh => $dbh,
                                          logger => $self->{logger},
                                          db_file_refs => [
                                                      { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                        field => 'schema_filename' },
                                                      { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                        field => 'defense_config_filename' },
                                                      { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                        field => 'request_xpath_filename' },
                                                      { table_name => 'DCC.CONTENT_PROFILE_TEMPLATES',
                                                        field => 'response_xpath_filename' },
                                                     ],
                                          dir => "$ACCOUNT_DATA_DIR/xml");

  return RET_OK;
}

#----------------------------------------------------------------------
sub prepare_wss_checks {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  return RET_OK if $policy_id == 0;

  # create wss mask
  my ($wss_mask) = $dbh->selectrow_array("SELECT BIT_OR(subviolation_code) FROM PLC.PL_XML_WS_SUBVIOLATIONS WHERE policy_id =? AND flg_is_enabled = ?",
                                            undef, $policy_id, TRUE) || 0;

  # insert mask to DCC.ACCOUNTS table
  $dbh->do("UPDATE DCC.ACCOUNTS SET wss_sub_violations_mask = ? WHERE policy_name_crc = ?",
              undef, $wss_mask, $policy_name_crc);

  return RET_OK;
}
#----------------------------------------------------------------------

sub copy_policy_file_for_account {
  my(%params) = @_;

  my $policy_file = $params{policy_file};
  my $account_template = $params{account_template};
  my $do_md5 = (index($account_template, '%MD5%') != -1);

  my $account_filename = $account_template;

  if ($do_md5) {
      open(my $md5_fh, '<', $policy_file);
      my $md5_sum = Digest::MD5->new->addfile( $md5_fh )->hexdigest;
      close($md5_fh);
      $account_filename =~ s/\%MD5\%/$md5_sum/;
  }

  unless (-e $account_filename) {
    make_path(dirname($account_filename));
    copy($policy_file, $account_filename);
    chmod(0664, $account_filename);
  }

  return $account_filename;
}

#----------------------------------------------------------------------

sub prepare_negsig_collections {
    my($self, $policy_name_crc, $policy_id, $params) = @_;

    $params ||= {};
    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $single_sig_id = $params->{single_sig_id};

    my $is_learning_enabled = $self->is_learning_enabled(policy_id => $policy_id);
    my $sig_result = {};
    my $buffer_type_mask = { Request => 0, Response => 0 };
    my %always_disabled_sigs;
    my $negative_signatures = $self->{negative_signatures} || F5::NegativeSignatures->new($dbh);
    
    my $collection = F5::NegativeSignatures::Collection->new(dbh => $dbh,
                                                             logger => $logger,
                                                             negative_signatures => $negative_signatures,
                                                            );
    my $threat_campaign_rv;
    my $bot_signatures_rv;
    my $browsers_rv;

    if ($self->get_flg_encoding_selected($policy_name_crc)) {

        my %lazy_insert_limit = ( policy_name_crc => $policy_name_crc );
        if ($single_sig_id) {
            $lazy_insert_limit{sig_id} = $single_sig_id;
        }
        my $lazy_insert = F5::LazyInsert->new(dbh => $dbh,
                                              logger => $logger,
                                              batch_insert => F5::BatchInsert->new(dbh => $dbh),
                                              default_limit => \%lazy_insert_limit,
                                             );

        $lazy_insert->handle_table(table_name => 'DCC.ACCOUNT_NEGSIG_SIGNATURE_PROPERTIES',
                                   unique_key => 'sig_id');

        my %set_names_templates;
        my $set_names_id = 1;
        my %additional_table_options;
        if ($single_sig_id) {
            $additional_table_options{delete_missing_on_flush} = 0;
        }
        $lazy_insert->handle_table(table_name => 'DCC.ACCOUNT_NEGSIG_SET_NAMES_TEMPLATES',
                                   unique_key => 'negsig_set_names_id',
                                   %additional_table_options,
                                  );

        if ($self->{prepare_negsig_exclusions}) {
            $lazy_insert->handle_table(table_name => 'DCC.ACCOUNT_NEGSIG_PARAM_EXCLUSIONS');
            $lazy_insert->handle_table(table_name => 'DCC.ACCOUNT_NEGSIG_URL_EXCLUSIONS');
            $lazy_insert->handle_table(table_name => 'DCC.ACCOUNT_NEGSIG_XML_EXCLUSIONS');
            $lazy_insert->handle_table(table_name => 'DCC.ACCOUNT_NEGSIG_HEADER_EXCLUSIONS');
            $lazy_insert->handle_table(table_name => 'DCC.ACCOUNT_NEGSIG_COOKIE_EXCLUSIONS');
        }

        my $signature_filter = undef;
        if ($single_sig_id) {
            $signature_filter = { sig_id => $single_sig_id };
        }
        $sig_result = $negative_signatures->get_signatures_for_policy($policy_id, $signature_filter, undef,
                                                                      get_as_composite_result => 1,
                                                                      include_previous_enforced => 1,
                                                                      learning_disabled => !$is_learning_enabled,
                                                                     );

        my %rule_id_map;
        for my $signature_key (sort keys %{$sig_result->{signatures}}) {
            my $signature = $sig_result->{signatures}->{$signature_key};
            $buffer_type_mask->{$signature->{apply_to}} |= $signature->{buffer_type_mask};

            my %dcc_row;
            $dcc_row{policy_name_crc} = $policy_name_crc;
            # Is this the signature with the previous enforced rule ?
            if(exists $signature->{sig_id_orig}) {
                # We need to generate a sig id based on the previous enforced rule
                $signature->{sig_id} = $collection->get_rule_id($signature->{rule});
                next if defined $rule_id_map{$signature->{sig_id}};
                $rule_id_map{$signature->{sig_id}} = 1;
            }
            else {
                $signature->{sig_id_orig} = $signature->{sig_id};
            }

            $dcc_row{sig_id} = $signature->{sig_id};
            $dcc_row{sig_id_orig} = $signature->{sig_id_orig};

            my @all_signature_sets = @{$signature->{all_sets}};
            my %all_set_names = map {$_->{set_name} => 1} @all_signature_sets;
            my $set_template = join(';', sort keys %all_set_names);
            if (! $set_names_templates{$set_template}) {
                $set_names_templates{$set_template} = $set_names_id;
                $lazy_insert->insert_row(table_name => 'DCC.ACCOUNT_NEGSIG_SET_NAMES_TEMPLATES',
                                         rec => { negsig_set_names_id => $set_names_id,
                                                  policy_name_crc => $policy_name_crc,
                                                  negsig_set_names_text => $set_template,
                                                },
                                         );
                $set_names_id++;
            }
            $dcc_row{negsig_set_names_id} = $set_names_templates{$set_template};

            $dcc_row{blocking_mask} = get_blocking_mask(signature => $signature,
                                                       );

            my $override_blocking_mask = 0;
            my $cookie_override_blocking_mask  = 0;

            if ($signature->{flg_has_value_override_enable}) {
                $override_blocking_mask = get_blocking_mask(signature => $signature,
                                                            prefix => 'predisabled_',
                                                           );
            }

            if ($signature->{flg_has_cookie_override_enable}) {
                $cookie_override_blocking_mask = get_blocking_mask(signature => $signature,
                                                                   prefix => 'predisabled_',
                                                                  );
            }

            if (! $dcc_row{blocking_mask} && ! $override_blocking_mask && ! $cookie_override_blocking_mask) {
                $always_disabled_sigs{$signature->{sig_id}} = 1;
                next;
            }

            $dcc_row{sig_crc} = F5::CRC::get_crc_policy_signature($policy_name_crc, $signature->{sig_id});
            $dcc_row{threshold} = $signature->{threshold_in_seconds} || 0;
            $dcc_row{last_enforcement_change} = $signature->{last_enforcement_change};
            $dcc_row{flg_quarantined} = $signature->{flg_quarantined};
            $dcc_row{sig_type} = SIG_TYPE_ATTACK_SIGNATURE;

            $lazy_insert->insert_row(table_name => 'DCC.ACCOUNT_NEGSIG_SIGNATURE_PROPERTIES',
                                     rec => \%dcc_row);

            # Prepare negsig overrides only for the signature with the previous enforced rule
            my $refresh_sig = exists $self->{refresh_negsig_exclusions}->{$policy_id} &&
       				     $self->{refresh_negsig_exclusions}->{$policy_id}->{$signature->{sig_id_orig}} &&
                                     $signature->{sig_id} != $signature->{sig_id_orig};
                                     
            my $prepare_negsig_exclusions = $self->{prepare_negsig_exclusions} &&
					   ($signature->{flg_has_value_override} || $signature->{flg_has_cookie_override});
                                                                     
            if ($refresh_sig || $prepare_negsig_exclusions) {
                $self->prepare_negsig_overrides(policy_name_crc => $policy_name_crc,
                                                policy_id => $policy_id,
                                                signature => $signature,
                                                lazy_insert => $lazy_insert,
                                                refresh_sig => $refresh_sig,
                                               );
            }
        }

        if (! $single_sig_id) {
            $threat_campaign_rv = $self->prepare_threat_campaign_rules(policy_name_crc => $policy_name_crc,
                                                                       policy_id => $policy_id,
                                                                       lazy_insert => $lazy_insert,);

            $bot_signatures_rv = $self->prepare_bot_signatures(policy_name_crc => $policy_name_crc,
                                                               policy_id => $policy_id,
                                                               lazy_insert => $lazy_insert);

            $browsers_rv = $self->prepare_browsers(policy_name_crc => $policy_name_crc,
                                                   policy_id => $policy_id,
                                                   lazy_insert => $lazy_insert);
        }

        $lazy_insert->flush();
        $self->add_modified_dcc_tables($lazy_insert->get_modified_tables());
    }

    if ($threat_campaign_rv) {
        $buffer_type_mask->{Request} |= $threat_campaign_rv->{threat_campaigns_request_buffer_type_mask};
        $buffer_type_mask->{Response} |= $threat_campaign_rv->{threat_campaigns_response_buffer_type_mask};
    }
    if ($bot_signatures_rv) {
        $buffer_type_mask->{Request} |= $bot_signatures_rv->{bot_signatures_request_buffer_type_mask};
    }
    if ($browsers_rv) {
        $buffer_type_mask->{Request} |= $browsers_rv->{browser_signatures_request_buffer_type_mask};
    }

    my $max_last_enforcement_change = List::Util::max((map { $_->{last_enforcement_change} } values %{$sig_result->{signatures}}),
                                                      $threat_campaign_rv->{max_last_enforcement_change} || 0,
                                                      $bot_signatures_rv->{max_last_enforcement_change} || 0,
                                                      $browsers_rv->{max_last_enforcement_change} || 0,
                                                     );

    $collection->add_for_account(policy_name_crc => $policy_name_crc,
                                 sets => [ sort map { $_->{set_id} } values %{$sig_result->{sets}} ],
                                 always_disabled_sigs => [ sort keys %always_disabled_sigs ],
                                 max_last_enforcement_change => $max_last_enforcement_change,
                                 buffer_type_mask => $buffer_type_mask,
                                 single_sig_id => $single_sig_id,
                                );

    if ($self->{generate_negsig_collections}) {
        $collection->generate(policy_name_crc => $policy_name_crc,);
    }
    $self->add_modified_dcc_tables($collection->get_modified_tables());

    return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_threat_campaign_rules {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $policy_name_crc = $params{policy_name_crc};
    my $policy_id = $params{policy_id};
    my $lazy_insert = $params{lazy_insert};

    $logger->debug("prepare_threat_campaign_rules, start") if $logger;

    my ($policy_threat_campaign_reject_flg, $policy_threat_campaign_alarm_flg, $policy_threat_campaign_staging_flg) = $self->{dbh}->selectrow_array(<<EndQuery, undef, $policy_id, F5::Violation::VIOL_THREAT_CAMPAIGN());
SELECT pb.flg_reject, pb.flg_alarm, ptca.flg_quarantine_threat_campaigns
  FROM PLC.PL_BLOCKING pb, PLC.PL_POLICY_THREAT_CAMPAIGN_ATTRIBUTES ptca
    WHERE pb.policy_id = ptca.policy_id
    AND pb.policy_id = ?
    AND pb.viol_index = ?;
EndQuery

    my ($max_last_enforcement_change, $threat_campaigns_request_buffer_type_mask, $threat_campaigns_response_buffer_type_mask) = (0, 0, 0);
    my $campaign_rules = F5::ThreatCampaigns::get_campaign_rules_for_policy(dbh => $dbh,
                                                                            logger => $logger,
                                                                            policy_id => $policy_id);

    for my $threat_campaign_id (keys %$campaign_rules) {
        my $rules = $campaign_rules->{$threat_campaign_id};
        for my $match_rule (@$rules) {
            next unless $match_rule->{flg_enabled};

            my $sig_crc = F5::CRC::get_crc_policy_signature($policy_name_crc, $match_rule->{sig_id});

            $match_rule->{policy_name_crc} = $policy_name_crc;
            $match_rule->{sig_id_orig} = $match_rule->{sig_id};
            $match_rule->{sig_crc} = $sig_crc;
            $match_rule->{flg_reject} = $policy_threat_campaign_reject_flg;
            $match_rule->{flg_alarm} = $policy_threat_campaign_alarm_flg;
            $match_rule->{flg_learn} = 0;

            $match_rule->{flg_quarantined} = $policy_threat_campaign_staging_flg && $match_rule->{flg_is_staging};
            if ($match_rule->{flg_quarantined}) {
                $match_rule->{flg_reject} = 0;
                $match_rule->{flg_alarm} = 0;
                $match_rule->{flg_learn} = 1;
            }

            $match_rule->{blocking_mask} = get_blocking_mask(signature => $match_rule);
            $match_rule->{last_enforcement_change} = $match_rule->{last_updated};
            $match_rule->{threshold} = 0;
            $match_rule->{sig_type} = SIG_TYPE_THREAT_CAMPAIGN;
            if ($match_rule->{last_enforcement_change} > $max_last_enforcement_change) {
                $max_last_enforcement_change = $match_rule->{last_enforcement_change};
            }

            $lazy_insert->insert_row(table_name => 'DCC.ACCOUNT_NEGSIG_SIGNATURE_PROPERTIES',
                                     rec => $match_rule,);

            if ($match_rule->{apply_to} == F5::ASMConfig::ApplyTo::APPLY_TO_REQUEST()) {
                $threat_campaigns_request_buffer_type_mask |= $match_rule->{buffer_type_mask};
            }
            elsif ($match_rule->{apply_to} == F5::ASMConfig::ApplyTo::APPLY_TO_RESPONSE()) {
                $threat_campaigns_response_buffer_type_mask |= $match_rule->{buffer_type_mask};
            }
        }
    }

    $logger->debug("prepare_threat_campaign_rules, finish") if $logger;
    return { max_last_enforcement_change => $max_last_enforcement_change,
             threat_campaigns_request_buffer_type_mask => $threat_campaigns_request_buffer_type_mask,
             threat_campaigns_response_buffer_type_mask => $threat_campaigns_response_buffer_type_mask,
           };
}

#----------------------------------------------------------------------

sub prepare_browsers {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $policy_name_crc = $params{policy_name_crc};
    my $policy_id = $params{policy_id};
    my $lazy_insert = $params{lazy_insert};

    $logger->debug("prepare_browsers, start") if $logger;

    my ($max_last_enforcement_change, $browser_signatures_request_buffer_type_mask) = (0, 0);
    my $is_bot_defense_enabled = $self->is_bot_defense_enabled(policy_id => $policy_id);

    if (! $is_bot_defense_enabled) {
        $logger->debug("Bot Defense is disabled -- skipping browsers") if $logger;
        return { max_last_enforcement_change => 0,
                 browser_signatures_request_buffer_type_mask => 0,
               };
    }

    my $browsers = F5::Browsers::get_user_defined_browsers(dbh => $dbh,
                                                           policy_id => $policy_id);

    my ($browser_default_action) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id, F5::ASMConfig::BotClass::BOT_CLASS_BROWSER());
SELECT action
  FROM PLC.PL_BOT_CLASSES
 WHERE policy_id = ? AND bot_class = ?
EndQuery

    for my $browser (values @$browsers) {
        my $signature = {};
        $signature->{sig_id} = F5::Browsers::get_browser_sig_id(dbh => $dbh, browser => $browser);
        my $sig_crc = F5::CRC::get_crc_policy_signature($policy_name_crc, $signature->{sig_id});
        $signature->{policy_name_crc} = $policy_name_crc;
        $signature->{sig_id_orig} = $signature->{sig_id};
        $signature->{sig_crc} = $sig_crc;
        $signature->{blocking_mask} = $self->get_bot_action_blocking_mask(policy_id => $policy_id,
                                                                          action => $browser->{action} || $browser_default_action);
        $signature->{last_enforcement_change} = $browser->{last_updated};
        $signature->{threshold} = 0;
        $signature->{sig_type} = SIG_TYPE_BROWSER_SIGNATURE;
        $signature->{flg_quarantined} = 0;
        if ($signature->{last_enforcement_change} > $max_last_enforcement_change) {
            $max_last_enforcement_change = $signature->{last_enforcement_change};
        }

        $lazy_insert->insert_row(table_name => 'DCC.ACCOUNT_NEGSIG_SIGNATURE_PROPERTIES',
                                 rec => $signature,);

        $browser_signatures_request_buffer_type_mask |= $signature->{buffer_type_mask};
    }

    $logger->debug("prepare_browsers, finish") if $logger;

    return { max_last_enforcement_change => $max_last_enforcement_change,
             browser_signatures_request_buffer_type_mask => $browser_signatures_request_buffer_type_mask,
           };
}

#----------------------------------------------------------------------

sub prepare_bot_signatures {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};
    my $policy_name_crc = $params{policy_name_crc};
    my $policy_id = $params{policy_id};
    my $lazy_insert = $params{lazy_insert};

    $logger->debug("prepare_bot_signatures, start") if $logger;

    my ($max_last_enforcement_change, $bot_signatures_request_buffer_type_mask) = (0, 0);
    my $is_bot_defense_enabled = $self->is_bot_defense_enabled(policy_id => $policy_id);

    if (! $is_bot_defense_enabled) {
        $logger->debug("Bot Defense is disabled -- skipping bot signatures") if $logger;
        return { max_last_enforcement_change => 0,
                 bot_signatures_request_buffer_type_mask => 0,
               };
    }

    my $bot_signatures = F5::BotSignatures::get_bot_signatures_for_policy(dbh => $dbh,
                                                                          logger => $logger,
                                                                          policy_id => $policy_id);

    for my $bot_signature (values @$bot_signatures) {
        if ($bot_signature->{action} eq F5::ASMConfig::BotAction::BOT_ACTION_IGNORE()) {
            next;
        }

        my $sig_crc = F5::CRC::get_crc_policy_signature($policy_name_crc, $bot_signature->{sig_id});

        $bot_signature->{policy_name_crc} = $policy_name_crc;
        $bot_signature->{sig_id_orig} = $bot_signature->{sig_id};
        $bot_signature->{sig_crc} = $sig_crc;
        $bot_signature->{blocking_mask} = $self->get_bot_action_blocking_mask(policy_id => $policy_id,
                                                                              action => $bot_signature->{action});
        $bot_signature->{last_enforcement_change} = $bot_signature->{last_updated};
        $bot_signature->{threshold} = 0;
        $bot_signature->{sig_type} = SIG_TYPE_BOT_SIGNATURE;
        $bot_signature->{flg_quarantined} = 0;
        # todo: consider renaming to unify bot_sig_id (used in DCC) and bot_signature_id (used in PLC)
        $bot_signature->{bot_sig_id} = $bot_signature->{bot_signature_id};
        if ($bot_signature->{last_enforcement_change} > $max_last_enforcement_change) {
            $max_last_enforcement_change = $bot_signature->{last_enforcement_change};
        }

        $lazy_insert->insert_row(table_name => 'DCC.ACCOUNT_NEGSIG_SIGNATURE_PROPERTIES',
                                 rec => $bot_signature,);

        $bot_signatures_request_buffer_type_mask |= $bot_signature->{buffer_type_mask};
    }

    $logger->debug("prepare_bot_signatures, finish") if $logger;
    return { max_last_enforcement_change => $max_last_enforcement_change,
             bot_signatures_request_buffer_type_mask => $bot_signatures_request_buffer_type_mask,
           };
}

#----------------------------------------------------------------------

sub get_bot_action_blocking_mask {
    my ($self, %params) = @_;
    my $policy_id = $params{policy_id};
    my $action = $params{action};

    if (! defined $self->{bot_action_blocking_mask}) {
        $self->{bot_action_blocking_mask} = {
            F5::ASMConfig::BotAction::BOT_ACTION_BLOCK() => get_blocking_mask(signature => { flg_reject => 1,
                                                                                             flg_alarm => 1,
                                                                                             flg_learn => 1}),
            F5::ASMConfig::BotAction::BOT_ACTION_ALARM() => get_blocking_mask(signature => { flg_reject => 0,
                                                                                             flg_alarm => 1,
                                                                                             flg_learn => 1}),
            F5::ASMConfig::BotAction::BOT_ACTION_DETECT() => get_blocking_mask(signature => { flg_reject => 0,
                                                                                              flg_alarm => 0,
                                                                                              flg_learn => 1}),
            F5::ASMConfig::BotAction::BOT_ACTION_IGNORE() => get_blocking_mask(signature => { flg_reject => 0,
                                                                                              flg_alarm => 0,
                                                                                              flg_learn => 0}),

        };
    }

    return $self->{bot_action_blocking_mask}->{$action};
}

#----------------------------------------------------------------------

sub prepare_bot_anomalies {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $logger->debug("prepare_bot_anomalies, start") if $logger;

    $self->empty_table('DCC.ACCOUNT_BOT_ANOMALIES', $policy_name_crc);

    my $is_bot_defense_enabled = $self->is_bot_defense_enabled(policy_id => $policy_id);

    if (! $is_bot_defense_enabled) {
        $logger->debug("Bot Defense is disabled -- skipping bot anomalies") if $logger;
        return RET_OK;
    }

    my $policy_bot_classes = $dbh->selectall_hashref(<<EndQuery, 'bot_class', undef, $policy_id);
SELECT *
  FROM PLC.PL_BOT_CLASSES
 WHERE policy_id = ?
EndQuery

    my $policy_bot_anomalies = $dbh->selectall_hashref(<<EndQuery, 'bot_anomaly_id', undef, $policy_id);
SELECT *
  FROM PLC.PL_BOT_ANOMALIES
 WHERE policy_id = ?
EndQuery

    my $bot_anomalies = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} });
SELECT *
  FROM PLC.BOT_ANOMALIES
EndQuery

    for my $bot_anomaly (values @$bot_anomalies) {

        $bot_anomaly->{action} = $policy_bot_classes->{$bot_anomaly->{bot_class}}->{action};
        $bot_anomaly->{policy_name_crc} = $policy_name_crc;

        if ($policy_bot_anomalies->{$bot_anomaly->{bot_anomaly_id}}) {
            if ( $policy_bot_anomalies->{$bot_anomaly->{bot_anomaly_id}}->{action} ne F5::ASMConfig::BotAction::BOT_ACTION_DEFAULT()) {
                $bot_anomaly->{action} = $policy_bot_anomalies->{$bot_anomaly->{bot_anomaly_id}}->{action};
            }
            if ( $policy_bot_anomalies->{$bot_anomaly->{bot_anomaly_id}}->{score_threshold} != 0 ) {
                $bot_anomaly->{score_threshold} = $policy_bot_anomalies->{$bot_anomaly->{bot_anomaly_id}}->{score_threshold};
            }
        }

        $bot_anomaly->{blocking_mask} = $self->get_bot_action_blocking_mask(policy_id => $policy_id,
                                                                            action => $bot_anomaly->{action});

        if ($bot_anomaly->{action} eq F5::ASMConfig::BotAction::BOT_ACTION_IGNORE()) {
            next; 
        }

        $self->insert_row(table_name => 'DCC.ACCOUNT_BOT_ANOMALIES',
                                 rec => $bot_anomaly,);
    }

    $logger->debug("prepare_bot_anomalies, finish") if $logger;
    return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_negsig_overrides {
    my ($self, %params) = @_;

    my $dbh = $self->{dbh};
    my $policy_name_crc = $params{policy_name_crc};
    my $policy_id = $params{policy_id};
    my $signature = $params{signature};
    my $refresh_sig = $params{refresh_sig};

    if ($signature->{flg_has_param_override} || $refresh_sig) {
    	my $param_overrides;
    	if($refresh_sig) {
            $param_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, $signature->{sig_id_orig});
SELECT flg_is_enabled, keycrc, blocking_mask
  FROM DCC.ACCOUNT_NEGSIG_PARAM_EXCLUSIONS
 WHERE policy_name_crc = ?
   AND sig_id = ?
EndQuery
    	}
    	else {
            $param_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id, $signature->{sig_id_orig});
SELECT param_attrib_id, flg_is_enabled
  FROM PLC.PL_PARAM_NEGSIG_SIGNATURES
 WHERE policy_id = ? 
   AND sig_id = ?
EndQuery
    	}
 
        for my $param_override (@$param_overrides) {        	
	        my $keycrc =  $refresh_sig ? $param_override->{keycrc} :
	        	      $self->{h_keys}->{$param_override->{param_attrib_id}}->{keycrc}; #Skip if we have no keycrc for this param (ie. param with null character)
            next unless $keycrc;
            
            $self->add_negsig_exclusion(%params,
                                        dcc_table => 'DCC.ACCOUNT_NEGSIG_PARAM_EXCLUSIONS',
                                        flg_is_enabled => $param_override->{flg_is_enabled},
                                        rec => {
                                                param_sig_crc => F5::CRC::get_crc_crc($keycrc,
                                                                                      F5::CRC::get_crc_policy_signature($policy_name_crc,
                                                                                                                         $signature->{sig_id})),
                                                flg_is_enabled => $param_override->{flg_is_enabled},
                                                keycrc => $keycrc,
                                                sig_id => $signature->{sig_id},                                                
                                               },
					 regular_insert => $refresh_sig,
					 blocking_mask => $refresh_sig ? $param_override->{blocking_mask} : undef,
                                       );
        }
    }

    if ($signature->{flg_has_url_override} || $refresh_sig) {    	
    	my $url_overrides;    	    	
    	if($refresh_sig) {
    		$url_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, $signature->{sig_id_orig});
SELECT flg_is_enabled, objcrc, blocking_mask
  FROM DCC.ACCOUNT_NEGSIG_URL_EXCLUSIONS
 WHERE policy_name_crc = ? 
   AND sig_id = ?
EndQuery
        }
    	else {
    		$url_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id, $signature->{sig_id_orig});
SELECT obj_id, flg_is_enabled
  FROM PLC.PL_OBJECT_NEGSIG_SIGNATURES
 WHERE policy_id = ? 
   AND sig_id = ?
EndQuery
    	}
        
        for my $url_override (@$url_overrides) {
            my $objcrc = $refresh_sig ? $url_override->{objcrc} :
            				$self->{h_objects}->{$url_override->{obj_id}}->{objcrc};
            next unless $objcrc;
                                                                                               
            $self->add_negsig_exclusion(%params,
                                        dcc_table => 'DCC.ACCOUNT_NEGSIG_URL_EXCLUSIONS',
                                        flg_is_enabled => $url_override->{flg_is_enabled},
                                        rec => {
                                                url_sig_crc => F5::CRC::get_crc_crc($objcrc,
                                                                                      F5::CRC::get_crc_policy_signature($policy_name_crc,
                                                                                                                         $signature->{sig_id})),
                                                flg_is_enabled => $url_override->{flg_is_enabled},
                                                objcrc => $objcrc,
                                                sig_id => $signature->{sig_id},                                                                                                
                                               },
                                        regular_insert => $refresh_sig,
                                        blocking_mask => $refresh_sig ? $url_override->{blocking_mask} : undef,
                                       );
        }
    }

    if ($signature->{flg_has_header_override} || $refresh_sig) {
    	my $header_overrides;
    	if($refresh_sig) {
            $header_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, $signature->{sig_id_orig});
SELECT flg_is_enabled, header_name_crc, blocking_mask
  FROM DCC.ACCOUNT_NEGSIG_HEADER_EXCLUSIONS
 WHERE policy_name_crc = ? 
   AND sig_id = ?
EndQuery
    	}
    	else {
    	    $header_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id, $signature->{sig_id_orig});
SELECT header_id, flg_is_enabled
  FROM PLC.PL_HEADER_NEGSIG_SIGNATURES
 WHERE policy_id = ?
   AND sig_id = ?
EndQuery
    	}
    		
        for my $header_override (@$header_overrides) {
	        my $header_name_crc = $refresh_sig ? $header_override->{header_name_crc} :  
	        				     $self->{h_headers}->{$header_override->{header_id}}->{header_name_crc};
            next unless $header_name_crc;
            
            $self->add_negsig_exclusion(%params,
                                        dcc_table => 'DCC.ACCOUNT_NEGSIG_HEADER_EXCLUSIONS',
                                        flg_is_enabled => $header_override->{flg_is_enabled},
                                        rec => {
                                                header_sig_crc => F5::CRC::get_crc_crc($header_name_crc,
                                                                                      F5::CRC::get_crc_policy_signature($policy_name_crc,
                                                                                                                         $signature->{sig_id})),
												flg_is_enabled => $header_override->{flg_is_enabled},
												header_name_crc => $header_name_crc,
                                                sig_id => $signature->{sig_id},
                                               },
					regular_insert => $refresh_sig,
					blocking_mask => $refresh_sig ? $header_override->{blocking_mask} : undef,
                                       );
        }
    }

    if ($signature->{flg_has_content_profile_override} || $refresh_sig) {
    	my $content_profile_overrides;
    	if($refresh_sig) {
    		$content_profile_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, $signature->{sig_id_orig});
SELECT flg_is_enabled, profile_templ_id, blocking_mask
  FROM DCC.ACCOUNT_NEGSIG_XML_EXCLUSIONS
 WHERE policy_name_crc = ?
   AND sig_id = ?
EndQuery
    	}
    	else {
    	    $content_profile_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id, $signature->{sig_id_orig});
SELECT profile_sig.profile_id, profile_sig.flg_is_enabled, profile.content_profile_type
  FROM PLC.PL_CONTENT_PROFILES profile,
       PLC.PL_CONTENT_PROFILE_NEGSIG_SIGNATURES profile_sig
 WHERE profile.policy_id = ? 
   AND profile_sig.sig_id = ?
   AND profile.id = profile_sig.profile_id
EndQuery
    	}
    	
        for my $content_profile_override (@$content_profile_overrides) {            
            my $profile_templ_id;
            if($refresh_sig) {
            	$profile_templ_id = $content_profile_override->{profile_templ_id};
            }
            else {
            	my $content_profile_type = $content_profile_types{$content_profile_override->{content_profile_type}};
	        next if !$self->{"h_${content_profile_type}_profiles"}->{$content_profile_override->{profile_id}};
		$profile_templ_id =  $self->{"h_${content_profile_type}_profiles"}->{$content_profile_override->{profile_id}}->{profile_templ_id};
		if (ref($profile_templ_id) eq "HASH") {
                    $profile_templ_id = $profile_templ_id->{$content_profile_type};
            	}
            }
                                    
            $self->add_negsig_exclusion(%params,
                                        dcc_table => 'DCC.ACCOUNT_NEGSIG_XML_EXCLUSIONS',
                                        flg_is_enabled => $content_profile_override->{flg_is_enabled},
                                        rec => {
                                                xml_sig_crc => F5::CRC::get_crc_xml_policy_signature($profile_templ_id, 
                                                                                                      $policy_name_crc, 
                                                                                                      $signature->{sig_id}),
                                                flg_is_enabled => $content_profile_override->{flg_is_enabled},
                                                profile_templ_id => $profile_templ_id,
                                                sig_id => $signature->{sig_id},
                                               },
					regular_insert => $refresh_sig,
					blocking_mask => $refresh_sig ? $content_profile_override->{blocking_mask} : undef,
                                       );
        }

    }

    if ($signature->{flg_has_cookie_override} || $refresh_sig) {
    	my $cookie_overrides;
    	if($refresh_sig) {
    	    $cookie_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_name_crc, $signature->{sig_id_orig});
SELECT flg_is_enabled, cookie_name, blocking_mask
  FROM DCC.ACCOUNT_NEGSIG_COOKIE_EXCLUSIONS
 WHERE policy_name_crc = ?
   AND sig_id = ?
EndQuery
    	}
    	else {    		    
            $cookie_overrides = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id, $signature->{sig_id_orig});
SELECT cookie_sig.flg_is_enabled, cookie.cookie_name
  FROM PLC.PL_ALLOWED_COOKIES cookie,
       PLC.PL_COOKIE_NEGSIG_SIGNATURES cookie_sig
 WHERE cookie.policy_id = ? 
   AND cookie_sig.sig_id = ?
   AND cookie.id = cookie_sig.cookie_id
EndQuery
    	}

        for my $cookie_override (@$cookie_overrides) {
            $self->add_negsig_exclusion(%params,
                                        dcc_table => 'DCC.ACCOUNT_NEGSIG_COOKIE_EXCLUSIONS',
                                        flg_is_enabled => $cookie_override->{flg_is_enabled},
                                        rec => {
                                                cookie_sig_crc => F5::CRC::get_crc_crc(F5::CRC::get_crc_any("get_crc32", $cookie_override->{cookie_name}, 0),
                                                                                       F5::CRC::get_crc_policy_signature($policy_name_crc, $signature->{sig_id})),
                                                flg_is_enabled => $cookie_override->{flg_is_enabled},
                                                cookie_name => $cookie_override->{cookie_name},
                                                sig_id => $signature->{sig_id},
                                               },
					regular_insert => $refresh_sig,
					blocking_mask => $refresh_sig ? $cookie_override->{blocking_mask} : undef,
                                   );
        }
    }

}

#----------------------------------------------------------------------

sub add_negsig_exclusion {
    my($self, %params) = @_;
    my $policy_name_crc = $params{policy_name_crc};
    my $dcc_table = $params{dcc_table};
    my $sig_crc = $params{sig_crc};
    my $signature = $params{signature};
    my $flg_is_enabled = $params{flg_is_enabled};
    my $rec = $params{rec};
    my $lazy_insert = $params{lazy_insert};
    my $regular_insert = $params{regular_insert};

    if ($regular_insert) {
        F5::DbUtils::insert_row(dbh => $self->{dbh},
                                table_name => $dcc_table,
		        	ignore => 1,
                                rec => { %$rec,
                                         policy_name_crc => $policy_name_crc,
                                         blocking_mask => $params{blocking_mask} } );
    }
    else {
        my $override_blocking_mask = 0;
        if ($flg_is_enabled) {
            $override_blocking_mask = get_blocking_mask(signature => $signature, 
                                                        prefix => 'predisabled_',
                                                       );
        }

        my $sig_blocking_mask = get_blocking_mask(signature => $signature,);

        if ($override_blocking_mask != $sig_blocking_mask) {
            $lazy_insert->insert_row(
                                    table_name => $dcc_table,
                                    rec => { %$rec,
                                             policy_name_crc => $policy_name_crc,
                                             blocking_mask => $override_blocking_mask } );
        }
    }
}

#----------------------------------------------------------------------

sub get_blocking_mask {
  my (%params) = @_;

  my $signature = $params{signature};
  my $prefix = $params{prefix} || '';

  if (! exists $signature->{$prefix . 'flg_reject'}) {
    $prefix = '';
  }

  my $blocking_mask = 0;
  $blocking_mask |= ($signature->{$prefix . 'flg_reject'}? NEGSIG_BLOCKING_MASK_REJECT : 0);
  $blocking_mask |= ($signature->{$prefix . 'flg_alarm'} ? NEGSIG_BLOCKING_MASK_ALARM : 0);
  $blocking_mask |= ($signature->{$prefix . 'flg_learn'} ? NEGSIG_BLOCKING_MASK_LEARN : 0);

  return $blocking_mask;
}

#----------------------------------------------------------------------

sub prepare_soap_method_data {
  my ($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  local $" = ',';

  # Erase methods from all XML templates that are about to be populated
  # Note: old XML templates may exist in the DCC (that no longer have a PLC XML profile), 
  # and need to be preserved until the template is deleted.
  for my $xml_profile (values %{$self->{h_xml_profiles}}) {
    for my $profile_templ_id (values %{$xml_profile->{profile_templ_id}}) {
      $dbh->do(<<EndQuery, undef, $profile_templ_id, $policy_name_crc);
DELETE FROM DCC.XML_SOAP_METHODS
      WHERE profile_templ_id = ?
        AND policy_name_crc = ?
EndQuery
    }
  }

  my %xml_templ_has_disallowed_methods;
  my @plc_fields = qw(profile_id soap_method namespace allowed_soap_action flg_allowed);

  my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields 
  FROM PLC.PL_XML_PROFILE_SOAP_METHODS
 WHERE policy_id = ?
EndQuery
  $plc_sth->execute($policy_id);

  while (my $plc_row = $plc_sth->fetchrow_hashref) {
    my $dcc_row = {};
    $dcc_row->{policy_name_crc} = $policy_name_crc;
    $dcc_row->{allowed_soap_action} = $plc_row->{allowed_soap_action};
    $dcc_row->{flg_allowed} = $plc_row->{flg_allowed};

    for my $profile_templ_id (values %{$self->{h_xml_profiles}->{$plc_row->{profile_id}}->{profile_templ_id}}) {
      $dcc_row->{soap_method_crc} = F5::CRC::get_crc_soap_method($profile_templ_id, $plc_row->{namespace}, $plc_row->{soap_method});
      $dcc_row->{profile_templ_id} = $profile_templ_id;

      my @dcc_fields = qw(policy_name_crc soap_method_crc allowed_soap_action flg_allowed profile_templ_id);
      $self->insert_row(
                        table_name => 'DCC.XML_SOAP_METHODS',
                        fields => \@dcc_fields,
                        rec => $dcc_row); 

      if (!$dcc_row->{flg_allowed}) {
        $xml_templ_has_disallowed_methods{$profile_templ_id} = 1;
      }
    }
  }

  for my $profile_templ_id (keys %xml_templ_has_disallowed_methods) {
    $dbh->do(<<EndQuery, undef, $profile_templ_id, $policy_name_crc);
UPDATE DCC.CONTENT_PROFILE_TEMPLATES
   SET flg_has_disallowed_methods = 1
 WHERE profile_templ_id = ?
   AND policy_name_crc = ?
EndQuery
  }
}

#----------------------------------------------------------------------

sub prepare_normalization {
    my($self, $policy_name_crc, $policy_id) = @_;

    my $dbh = $self->{dbh};
    return RET_OK unless $policy_id;

    $self->empty_table('DCC.ACCOUNT_NORMALIZATION_PROFILE', $policy_name_crc);


    my $normalization_viols = $dbh->selectall_hashref(<<EndQuery,'normalization_code', undef, $policy_id);
SELECT normalization_code, flg_is_enabled, flg_pb_enabled, max_decoding_passes
  FROM PLC.PL_NORMALIZATION
WHERE policy_id = ?
EndQuery

    $self->{normalization_viols} = $normalization_viols;

    # Need to re-determine whether violation is enabled, since it is forced OFF for ACCOUNT_BLOCKING (as per bd requirements)
    my $blocking_flags = $dbh->selectrow_hashref(<<EndQuery,undef,$policy_id, F5::Violation::VIOL_EVASION());
SELECT flg_alarm,flg_reject,flg_learn
  FROM PLC.PL_BLOCKING
 WHERE policy_id = ?
   AND viol_index = ?
EndQuery

    my $dcc_row = $self->calculate_normalization_viols ( normalization_viols => $normalization_viols,
                                                         blocking_flags         => $blocking_flags,
                                                         policy_name_crc             => $policy_name_crc,
                                                         policy_id              => $policy_id,
                                                         is_learning_enabled    => $self->is_learning_enabled(policy_id => $policy_id),
                                                       );


    ($dcc_row->{max_decoding_passes}) = $dbh->selectrow_array(<<EndQuery, undef, $policy_id,$MULTIPLE_DECODING_NORMALIZATION_CODE);
SELECT max_decoding_passes
  FROM PLC.PL_NORMALIZATION
 WHERE policy_id = ?
   AND normalization_code = ?
EndQuery

    if ( ! defined $dcc_row->{max_decoding_passes} ) {
        return RET_NOT_OK;
    }

    $self->{global_subviolation_attribute}->{max_decoding_passes} = $dcc_row->{max_decoding_passes};

    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_NORMALIZATION_PROFILE',
                      rec => $dcc_row);

    return RET_OK;
}
#----------------------------------------------------------------------

sub calculate_normalization_viols {
    my($self, %params) = @_;
    my $normalization_viols = $params{normalization_viols};
    my $policy_name_crc = $params{policy_name_crc};
    my $policy_id = $params{policy_id};
    my $is_learning_enabled = $params{is_learning_enabled};
    my $dbh = $self->{dbh};
    my $blocking_flags = $params{blocking_flags};

    if ( ! $is_learning_enabled) {
        $blocking_flags->{flg_learn} = 0;
    }

    my %dcc_row = (
                   policy_name_crc => $policy_name_crc,
                   reject_bitmap => 0,
                   alarm_bitmap => 0,
                   learn_bitmap => 0,
                   user_mask_bitmap => 0,
                  );

    for my $normalization_row ( values %$normalization_viols ) {
        if ($normalization_row->{flg_is_enabled}) {
            for my $blocktype (qw(reject alarm)) {
                if ($blocking_flags->{"flg_$blocktype"}) {
                    $dcc_row{"${blocktype}_bitmap"} |= $normalization_row->{normalization_code};
                }
            }
            $dcc_row{user_mask_bitmap} |= $normalization_row->{normalization_code};
        }
        if ($normalization_row->{flg_pb_enabled}) {
            if ($blocking_flags->{flg_learn}) {
                $dcc_row{learn_bitmap} |= $normalization_row->{normalization_code};
            }
        }
    }

    return \%dcc_row;
    }

#----------------------------------------------------------------------

sub prepare_response_scrubbing {
  my($self, $policy_name_crc, $policy_id) = @_;

  my $dbh = $self->{dbh};
  return RET_OK unless $policy_id;
  $self->empty_table('DCC.ACCOUNT_RESPONSE_SCRUBBING', $policy_name_crc);
  $self->empty_table('DCC.ACCOUNT_RESPONSE_SCRUBBING_FORMATS', $policy_name_crc);
  
  my $dcc_row = $dbh->selectrow_hashref(<<EndQuery, undef, $policy_id);
SELECT *
  FROM PLC.PL_RESPONSE_SCRUBBING
 WHERE policy_id = ?
EndQuery

  my @scrubbing_active_fields = qw(flg_check_for_ccn flg_check_for_ssn flg_check_file_content all_patterns all_exceptions flg_enforcement_mode ccn_exposed_len ssn_exposed_len);
  my @dcc_fields = (qw(policy_name_crc  flg_mask_data), @scrubbing_active_fields);

  $dcc_row->{policy_name_crc} = $policy_name_crc;

  # flg_mask_data works distinctly from Data Guard violation, and needs to take flg_is_enabled into consideration on its own
  $dcc_row->{flg_mask_data} = ($dcc_row->{flg_mask_data} && $dcc_row->{flg_is_enabled} ? 1 : 0);

  $dcc_row->{flg_enforcement_mode} = (defined $dcc_row->{enforcement_mode} && $dcc_row->{enforcement_mode} eq 'enforce_all_except_url_list') ? 0 : 1;
  delete $dcc_row->{enforcement_mode};

  my $patterns = $dbh->selectcol_arrayref("SELECT pattern FROM PLC.PL_RESPONSE_SCRUBBING_PATTERNS A, PLC.PL_RESPONSE_SCRUBBING B WHERE 
                                           A.policy_id=B.policy_id and A.policy_id = ? and B.flg_check_custom_patterns = ?", undef, $policy_id, TRUE);

  $dcc_row->{all_patterns} = join('|', @$patterns);

  my $exceptions = $dbh->selectcol_arrayref("SELECT pattern FROM PLC.PL_RESPONSE_SCRUBBING_EXCEPTIONS A, PLC.PL_RESPONSE_SCRUBBING B WHERE
                                           A.policy_id=B.policy_id and A.policy_id = ? and B.flg_check_exception_patterns = ?", undef, $policy_id, TRUE);

  $dcc_row->{all_exceptions} = join('|', @$exceptions);

  $dcc_row->{flg_check_for_ccn} = $dcc_row->{flg_check_for_credit_card};
  delete $dcc_row->{flg_check_for_credit_card};

  if (grep($dcc_row->{$_}, @scrubbing_active_fields) && $dcc_row->{flg_is_enabled}) {
    $self->insert_row(
                      table_name => 'DCC.ACCOUNT_RESPONSE_SCRUBBING',
                      fields => \@dcc_fields,
                      rec => $dcc_row);
  }

  my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT A.format_id,B.format_name
  FROM PLC.PL_RESPONSE_SCRUBBING_FORMATS A,PLC.RESPONSE_SCRUBBING_FORMATS B
 WHERE A.format_id = B.format_id
   AND policy_id = ?
EndQuery
  $plc_sth->execute($policy_id);

  while (my $plc_row = $plc_sth->fetchrow_hashref) {
    my $dcc_row = {};
    $dcc_row->{policy_name_crc} = $policy_name_crc;
    $dcc_row->{format_id} = $plc_row->{format_id};
    $dcc_row->{format_name} = $plc_row->{format_name};


    my @dcc_fields = qw(policy_name_crc format_id format_name);
    $self->insert_row(
                       table_name => 'DCC.ACCOUNT_RESPONSE_SCRUBBING_FORMATS',
                       fields => \@dcc_fields,
                       rec => $dcc_row);

  }
  
  return 1;
}
#--------------------------------------------------------------------------------
sub prepare_geolocation_enforcement {
  my ($self, $policy_name_crc, $policy_id) = @_;

  return RET_OK if $policy_id == 0;
  my $dbh = $self->{dbh};

  $self->empty_table('DCC.ACCOUNT_GEOLOCATION_ENFORCEMENT', $policy_name_crc);

  my @plc_fields = qw(country_code);

  my $plc_sth = $dbh->prepare(<<EndQuery);
SELECT @plc_fields
  FROM PLC.PL_GEOLOCATION_ENFORCEMENT_LIST
 WHERE policy_id = ?
EndQuery
  $plc_sth->execute($policy_id);

  my $countries_defined = 0;
  my @dcc_fields = qw(policy_name_crc country_code);
  while (my $plc_row = $plc_sth->fetchrow_hashref) {
      $plc_row->{policy_name_crc} = $policy_name_crc;
      $self->insert_row(
                        table_name => 'DCC.ACCOUNT_GEOLOCATION_ENFORCEMENT',
                        fields => \@dcc_fields,
                        rec => $plc_row);
      $countries_defined = 1;
  }

  return RET_OK;
}

#----------------------------------------------------------------------

sub find_matching_hash {
  my($hash_a, $hash_b, $compare_fields) = @_;

  for my $field (@$compare_fields) {
    if ((exists $hash_a->{$field} && exists $hash_b->{$field} &&
         $hash_a->{$field} ne $hash_b->{$field}) || 
        (exists $hash_a->{$field} xor exists $hash_b->{$field})) {
      return;
    }
  }
  return $hash_a;
}

#----------------------------------------------------------------------

sub get_flg_encoding_selected {
    my($self, $policy_name_crc) = @_;

    my $dbh = $self->{dbh};
    my ($flg_encoding_selected) = $dbh->selectrow_array(<<EndQuery, undef, $policy_name_crc);
SELECT flg_encoding_selected
  FROM DCC.ACCOUNTS
 WHERE policy_name_crc = ?
EndQuery

    return $flg_encoding_selected;
}
 
#----------------------------------------------------------------------
sub ucs_upgrade_profile_id_fix {
    my ($self, %params) = @_;
    my $profile = $params{profile};
    my $dbh = $self->{dbh};
    my ($real_profile_id) = @{$dbh->selectcol_arrayref(<<EndQuery, undef, $profile->{profile_id}, $self->{ucs_upgrade_real_policy_id})};
SELECT current_policy.id
  FROM PLC.PL_CONTENT_PROFILES policy_from_history JOIN PLC.PL_CONTENT_PROFILES current_policy
     ON policy_from_history.profile_name = current_policy.profile_name
    AND policy_from_history.content_profile_type = current_policy.content_profile_type
  WHERE policy_from_history.id = ?
    AND current_policy.policy_id = ?
EndQuery
   if ($real_profile_id) {
       $profile->{profile_id} = $real_profile_id;
   }
   return;
}
#----------------------------------------------------------------------
sub calc_mask{
  my %params = @_;
  my $mask = 0;

  for my $flg_index (1 .. scalar keys %params ){
     if( $params{$flg_index} ){
       $mask |= (2**($flg_index-1));
     }
  }
  return $mask;
}
#----------------------------------------------------------------------

sub object_allowed_origins_row_handler {
  my ($self, %params) = @_;

  my $row = $params{row};
  my $obj_id = $row->{obj_id};

  if ( !$self->{h_objects}->{$obj_id}->{flg_enable_cors_enforcement} 
         || $self->{h_objects}->{$obj_id}->{flg_disallow_cors} 
         || $self->{h_objects}->{$obj_id}->{allowed_origins_enforcement_mode} == F5::ASMConfig::CorsOriginEnforcementMode::CORS_ORIGIN_ENFORCEMENT_MODE_UNMODIFIED() ) {
     $row->{__skip_row} = 1;
     return;
  }

  $row->{object_crc} =  $self->{h_objects}->{$row->{obj_id}}->{objcrc};

}

#--------------------------------------------------------------------------------

sub object_cors_headers_row_handler {
  my ($self, %params) = @_;

  my $row = $params{row};
  my $obj_id = $row->{obj_id};

  if ( !$self->{h_objects}->{$obj_id}->{flg_enable_cors_enforcement} || $self->{h_objects}->{$obj_id}->{flg_disallow_cors} ) {
     $row->{__skip_row} = 1;
     return;
  }

  $row->{object_crc} =  $self->{h_objects}->{$row->{obj_id}}->{objcrc};

  if ($row->{allowed_header_name}) {
       if ( !$self->{h_objects}->{$obj_id}->{flg_allowed_headers_check} ) {
            $row->{__skip_row} = 1;
            return;
       }

      $row->{header_name} = $row->{allowed_header_name} ;
      $row->{header_type} = CORS_ALLOWED_HEADER_TYPE;
  }

  if ($row->{exposed_header_name}) {
       if ( !$self->{h_objects}->{$obj_id}->{flg_exposed_headers_check} ) {
            $row->{__skip_row} = 1;
            return;
       }

      $row->{header_name} = $row->{exposed_header_name} ;
      $row->{header_type} = CORS_EXPOSED_HEADER_TYPE;
  }

}
#--------------------------------------------------------------------------------
#
sub object_allowed_methods_row_handler {
  my ($self, %params) = @_;

  my $row = $params{row};
  my $obj_id = $row->{obj_id};

  if ( !$self->{h_objects}->{$obj_id}->{flg_enable_cors_enforcement} 
        || $self->{h_objects}->{$obj_id}->{flg_disallow_cors} 
         || !$self->{h_objects}->{$obj_id}->{flg_allowed_methods_check} ) {
     $row->{__skip_row} = 1;
     return;
  }

  $row->{object_crc} =  $self->{h_objects}->{$row->{obj_id}}->{objcrc};

}

#----------------------------------------------------------------------

sub websocket_url_allowed_origins_row_handler {
  my ($self, %params) = @_;

  my $row = $params{row};
  my $websocket_url_id = $row->{websocket_url_id};

  if ( !$self->{h_websocket_urls}->{$websocket_url_id}->{flg_enable_cors_enforcement}
         || !$self->{h_websocket_urls}->{$websocket_url_id}->{flg_allow_cors} ) {
     $row->{__skip_row} = 1;
     return;
  }

  $row->{websocket_url_crc} = $self->{h_websocket_urls}->{$row->{websocket_url_id}}->{websocket_url_crc};

}
#----------------------------------------------------------------------

sub prepare_account_browsers {
    my ($self, $policy_name_crc, $policy_id, $params) = @_;
    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $self->empty_table('DCC.ACCOUNT_FACTORY_BROWSER_PROPERTIES', $policy_name_crc);

    my $is_bot_defense_enabled = $self->is_bot_defense_enabled(policy_id => $policy_id);
    if (! $is_bot_defense_enabled) {
        $logger->debug("Bot Defense is disabled -- skipping browsers") if $logger;
        return RET_OK;
    }

    my $factory_browsers = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id);
SELECT browsers.*, factory.factory_browser_id
  FROM PLC.PL_BROWSERS browsers
LEFT JOIN PLC.PL_BROWSER_DEFINITIONS defs
       ON (defs.browser_definition_id = browsers.browser_definition_id)
LEFT JOIN PLC.FACTORY_BROWSERS factory
       ON (defs.browser_name = factory.browser_name)
 WHERE browsers.policy_id = ? AND factory.factory_browser_id IS NOT NULL
EndQuery

    for my $browser (@$factory_browsers) {
        $browser->{action_mask} = $self->get_bot_action_blocking_mask(policy_id => $policy_id,
                                                                      action => $browser->{action});
        $browser->{policy_name_crc} = $policy_name_crc;
        $self->insert_row(table_name => 'DCC.ACCOUNT_FACTORY_BROWSER_PROPERTIES',
                          rec => $browser);
    }

    return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_microservices {
   my ($self, $policy_name_crc, $policy_id, $params) = @_;
   my $dbh = $self->{dbh};

   my $readonly_mode = $params->{readonly_mode};
   unless ($readonly_mode) {
       $self->empty_table('DCC.ACCOUNT_MICROSERVICES', $policy_name_crc);
   }

   my $session_awareness_viols = $dbh->selectall_hashref('SELECT * FROM PLC.PL_SESSION_AWARENESS_VIOLATIONS WHERE policy_id = ?', 'viol_index', undef, $policy_id);
   my $flg_encoding_selected = defined $params->{flg_encoding_selected} ? $params->{flg_encoding_selected} : $self->get_flg_encoding_selected($policy_name_crc);
   my $is_learning_enabled = $self->is_learning_enabled(policy_id => $policy_id);
   my $is_policy_transparent = $self->is_policy_transparent(policy_id => $policy_id);

   my $microservices = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id);
SELECT *
  FROM PLC.PL_MICROSERVICES
WHERE policy_id = ?
ORDER BY pattern_order
EndQuery
  my $original_viols;
  my $http_check_blocking_flags;
  my $evasion_blocking_flags;
  if ( @$microservices ) {
      my $viols = $dbh->selectall_hashref('SELECT * FROM PLC.PL_BLOCKING WHERE policy_id = ?', 'viol_index', undef, $policy_id);
      $http_check_blocking_flags = $viols->{F5::Violation::VIOL_HTTP_PROTOCOL()};
      $evasion_blocking_flags = $viols->{F5::Violation::VIOL_EVASION()};
      $original_viols = dclone($self->{viols});
  }

  my $dcc_pattern_order = 1;
  my $is_awaf_licensed =  F5::License::is_waf_licensed();
  for my $microservice ( @$microservices ) {
      my $rec = {};
      $rec->{policy_name_crc} = $policy_name_crc;
      $rec->{pattern_order} = $dcc_pattern_order++;

      $rec->{object_pattern} = $microservice->{microservice_uri};
      $rec->{flg_is_pattern_object} = $microservice->{flg_is_pattern_uri};
      $rec->{host_pattern} = $microservice->{microservice_host_name};
      $rec->{flg_is_pattern_host} = $microservice->{flg_is_pattern_host};
      $rec->{flg_wildcard_includes_slash} = $microservice->{flg_wildcard_uri_includes_slash};
      my $microservice_http_check_blocking_flags = dclone($http_check_blocking_flags);
      my $microservice_evasion_blocking_flags = dclone ($evasion_blocking_flags);

      if ( $microservice->{transparency_enforcement_mode} == F5::ASMConfig::TransparencyEnforcementMode::TRANSPARENCY_ENFORCEMENT_BLOCKING() ) {
          $rec->{is_transparent} = 0;
      }
      elsif (  $microservice->{transparency_enforcement_mode} == F5::ASMConfig::TransparencyEnforcementMode::TRANSPARENCY_ENFORCEMENT_TRANSPARENT() ) {
          $rec->{is_transparent} = 1;
      }
      else{
         $rec->{is_transparent} = $is_policy_transparent;
      }

      my $global_blocking_overrides = {};
      if ( $is_awaf_licensed ) {
          $global_blocking_overrides =  $dbh->selectall_hashref(<<EndQuery, 'viol_index', undef, $policy_id, $microservice->{microservice_id});
SELECT viol_index, flg_learn, flg_alarm, flg_reject
  FROM PLC.PL_MICROSERVICE_BLOCKING_OVERRIDES
WHERE policy_id = ?
   AND microservice_id = ?
EndQuery
      }

      my $microservice_viols = dclone($original_viols);

      for my $global_blocking_override ( values %$global_blocking_overrides) {
          my $viol_index = $global_blocking_override->{viol_index};
          $microservice_viols->{$viol_index}->{flg_learn} = $global_blocking_override->{flg_learn};
          $microservice_viols->{$viol_index}->{flg_alarm} = $global_blocking_override->{flg_alarm};
          $microservice_viols->{$viol_index}->{flg_reject} = $global_blocking_override->{flg_reject};

          if ( $viol_index == F5::Violation::VIOL_HTTP_PROTOCOL() ) {
              $microservice_http_check_blocking_flags = dclone($microservice_viols->{F5::Violation::VIOL_HTTP_PROTOCOL()});
          }
          elsif ( $viol_index == F5::Violation::VIOL_EVASION() ) {
              $microservice_evasion_blocking_flags = dclone($microservice_viols->{F5::Violation::VIOL_EVASION()});
          }
      }

      if ( $self->is_policy_passive(policy_id => $policy_id) ){
          $rec->{is_transparent} = 1;
      }

      my $rv =  $self->calculate_viols( viols =>$microservice_viols,
                                      session_awareness_viols => $session_awareness_viols,
                                      policy_name_crc => $policy_name_crc,
                                      policy_id => $policy_id,
                                      flg_encoding_selected => $flg_encoding_selected,
                                      is_learning_enabled => $is_learning_enabled,
                                     );

      $rec->{viol_set_block} = $rv->{dcc_rec}->{viol_set_block};
      $rec->{viol_set_alarm} = $rv->{dcc_rec}->{viol_set_alarm};
      $rec->{viol_set_learn} = $rv->{dcc_rec}->{viol_set_learn};

      my $http_checks_overrides = {};
      if ( $is_awaf_licensed ) {
          $http_checks_overrides =  $dbh->selectall_hashref(<<EndQuery, 'http_check_code', undef, $policy_id, $microservice->{microservice_id});
SELECT *
  FROM PLC.PL_MICROSERVICE_HTTP_CHECKS_OVERRIDES
WHERE policy_id = ?
   AND microservice_id = ?
EndQuery
      }

      my $microservice_http_check_viols = dclone($self->{http_check_viols});

      for my $http_check_override ( values %$http_checks_overrides) {
          my $http_check_code = $http_check_override->{http_check_code};
          $microservice_http_check_viols->{$http_check_code}->{flg_is_enabled} = $http_check_override->{flg_is_enabled};
          $microservice_http_check_viols->{$http_check_code}->{flg_pb_enabled} = $http_check_override->{flg_pb_enabled};
      }

      $rv  = $self->calculate_http_check_viols( http_check_viols       => $microservice_http_check_viols,
                                                blocking_flags         => $microservice_http_check_blocking_flags,
                                                policy_name_crc        => $policy_name_crc,
                                                policy_id              => $policy_id,
                                                is_learning_enabled    => $is_learning_enabled,
                                              );


      $rec->{http_protocol_checks_learn_mask} = $rv->{http_protocol_checks_learn_mask};
      $rec->{http_protocol_checks_alarm_mask} = $rv->{http_protocol_checks_alarm_mask};
      $rec->{http_protocol_checks_reject_mask} = $rv->{http_protocol_checks_reject_mask};

      my $normalizations_overrides = {};
      if ( $is_awaf_licensed ) {
          $normalizations_overrides =  $dbh->selectall_hashref(<<EndQuery, 'normalization_code', undef, $policy_id, $microservice->{microservice_id});
SELECT *
  FROM PLC.PL_MICROSERVICE_NORMALIZATION_OVERRIDES
WHERE policy_id = ?
   AND microservice_id = ?
EndQuery
      }

      my $microservice_normalization_viols =  dclone($self->{normalization_viols});

      for my $normalization_override ( values %$normalizations_overrides) {
          my $normalization_code = $normalization_override->{normalization_code};
          $microservice_normalization_viols->{$normalization_code}->{flg_is_enabled} = $normalization_override->{flg_is_enabled};
          $microservice_normalization_viols->{$normalization_code}->{flg_pb_enabled} = $normalization_override->{flg_pb_enabled};
      }

      $rv  = $self->calculate_normalization_viols( normalization_viols    => $microservice_normalization_viols,
                                                   blocking_flags         => $microservice_evasion_blocking_flags,
                                                   policy_name_crc        => $policy_name_crc,
                                                   policy_id              => $policy_id,
                                                   is_learning_enabled    => $is_learning_enabled,
                                                 );

      $rec->{evasion_alarm_mask} = $rv->{alarm_bitmap};
      $rec->{evasion_learn_mask} = $rv->{learn_bitmap};
      $rec->{evasion_reject_mask} = $rv->{reject_bitmap};

      if ( defined $normalizations_overrides->{$MULTIPLE_DECODING_NORMALIZATION_CODE} ) {
          $rec->{evasion_max_decoding_passes} = $normalizations_overrides->{$MULTIPLE_DECODING_NORMALIZATION_CODE}->{max_decoding_passes};
      }
      else {
          $rec->{evasion_max_decoding_passes} =  $self->{global_subviolation_attribute}->{max_decoding_passes};
      }

      if ( defined $http_checks_overrides &&
           defined $http_checks_overrides->{$MAX_HEADERS_INDEX} &&
           defined $http_checks_overrides->{$MAX_HEADERS_INDEX}->{max_headers} ) {
          $rec->{max_headers} = $http_checks_overrides->{$MAX_HEADERS_INDEX}->{max_headers};
      }
      else {
          $rec->{max_headers} = $self->{global_subviolation_attribute}->{max_headers};
      }


      if ( defined $http_checks_overrides &&
           defined $http_checks_overrides->{$MAX_PARAMS_INDEX} &&
           defined $http_checks_overrides->{$MAX_PARAMS_INDEX}->{max_params} ) {
          $rec->{max_params} = $http_checks_overrides->{$MAX_PARAMS_INDEX}->{max_params};
      }
      else {
          $rec->{max_params} = $self->{global_subviolation_attribute}->{max_params};;
      }


      if ( defined $http_checks_overrides &&
           defined $http_checks_overrides->{$MAX_COOKIES_INDEX} &&
           defined $http_checks_overrides->{$MAX_COOKIES_INDEX}->{max_cookies} ) {
          $rec->{max_cookies} = $http_checks_overrides->{$MAX_COOKIES_INDEX}->{max_cookies};
      }
      else {
          $rec->{max_cookies} = $self->{global_subviolation_attribute}->{max_cookies};;
      }


     $self->insert_row(
                       table_name => 'DCC.ACCOUNT_MICROSERVICES',
                       rec => $rec);


  }
  return RET_OK;
}

#----------------------------------------------------------------------

sub prepare_bot_defense_class {
    my($self, $policy_name_crc, $policy_id, $params) = @_;
    my $logger = $self->{logger};
    my $dbh = $self->{dbh};

    $logger->debug("prepare_bot_defense_class, start") if $logger;

    $self->empty_table("DCC.ACCOUNT_BOT_DEFENSE_CLASS", $policy_name_crc);

    my $is_bot_defense_enabled = $self->is_bot_defense_enabled(policy_id => $policy_id);
    if (! $is_bot_defense_enabled) {
        $logger->debug("Bot Defense is disabled -- skipping bot class") if $logger;
        return RET_OK;
    }

    my $classes = $dbh->selectall_arrayref(<<EndQuery, { Slice => {} }, $policy_id);
SELECT *
  FROM PLC.PL_BOT_CLASSES
 WHERE policy_id = ?
EndQuery

    for my $class (@$classes) {
        my $row = { policy_name_crc => $policy_name_crc,
                    bot_class => $class->{bot_class},
                    action => $self->get_bot_action_blocking_mask(policy_id => $policy_id,
                                                                  action => $class->{action}), };
        $self->insert_row(table_name => 'DCC.ACCOUNT_BOT_DEFENSE_CLASS',
                          rec => $row);
    }

    $logger->debug("prepare_bot_defense_class, finish") if $logger;

    return RET_OK;
}

#----------------------------------------------------------------------
#
# The following tables are written by prepare_extracts. The names are
# constructed dynamically at run-time, and they are listed here for convenience,
# so that they can be found by a source grep.
#
#  ACCOUNT_OBJECT_EXTRACT_FORM_PARAM_INDEX
#  ACCOUNT_GLOBAL_EXTRACT_FORM_PARAM_INDEX
#  ACCOUNT_GLOBAL_BYOBJECT_EXTRACT_FORM_PARAM_INDEX
#  ACCOUNT_GLOBAL_BYFILETYPE_EXTRACT_FORM_PARAM_INDEX
#  ACCOUNT_GLOBAL_BYREGEXP_EXTRACT_FORM_PARAM_INDEX
#  ACCOUNT_OBJECT_EXTRACT_FORM_PARAM_NAME
#  ACCOUNT_GLOBAL_EXTRACT_FORM_PARAM_NAME
#  ACCOUNT_GLOBAL_BYOBJECT_EXTRACT_FORM_PARAM_NAME
#  ACCOUNT_GLOBAL_BYFILETYPE_EXTRACT_FORM_PARAM_NAME
#  ACCOUNT_GLOBAL_BYREGEXP_EXTRACT_FORM_PARAM_NAME
#  ACCOUNT_OBJECT_EXTRACT_HREF
#  ACCOUNT_GLOBAL_EXTRACT_HREF
#  ACCOUNT_GLOBAL_BYOBJECT_EXTRACT_HREF
#  ACCOUNT_GLOBAL_BYFILETYPE_EXTRACT_HREF
#  ACCOUNT_GLOBAL_BYREGEXP_EXTRACT_HREF
#  ACCOUNT_OBJECT_EXTRACT_XPATH
#  ACCOUNT_GLOBAL_EXTRACT_XPATH
#  ACCOUNT_GLOBAL_BYOBJECT_EXTRACT_XPATH
#  ACCOUNT_GLOBAL_BYFILETYPE_EXTRACT_XPATH
#  ACCOUNT_GLOBAL_BYREGEXP_EXTRACT_XPATH
#  ACCOUNT_OBJECT_EXTRACT_REGEXP
#  ACCOUNT_GLOBAL_EXTRACT_REGEXP
#  ACCOUNT_GLOBAL_BYOBJECT_EXTRACT_REGEXP
#  ACCOUNT_GLOBAL_BYFILETYPE_EXTRACT_REGEXP
#  ACCOUNT_GLOBAL_BYREGEXP_EXTRACT_REGEXP
#  ACCOUNT_OBJECT_EXTRACT_DYNAMIC_PARAM_NAME
#  ACCOUNT_GLOBAL_EXTRACT_DYNAMIC_PARAM_NAME
#  ACCOUNT_GLOBAL_BYOBJECT_EXTRACT_DYNAMIC_PARAM_NAME
#  ACCOUNT_GLOBAL_BYFILETYPE_EXTRACT_DYNAMIC_PARAM_NAME
#  ACCOUNT_GLOBAL_BYREGEXP_EXTRACT_DYNAMIC_PARAM_NAME

1;
